/**************************************************************************
 *
 * Copyright 2016 VMware, Inc.
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 **************************************************************************/


#include <stdlib.h>

#include "glretrace_wgl.hpp"
#include "glproc.hpp"


enum Function {
    F_BEGIN,
    F_FRONTFACE,
    F_VERTEX3F,
    F_NORMAL3F,
    F_END,
    F_TRANSLATEF,
};


struct Call {
    enum Function function;
    float args[3];
};


struct CallList {
    const Call *calls;
    size_t length;
};


#define C_Begin(mode)          { F_BEGIN,      { mode } }
#define C_FrontFace(mode)      { F_FRONTFACE,  { mode } }
#define C_Vertex3f(x, y, z)    { F_VERTEX3F,   { x, y, z } }
#define C_Normal3f(nx, ny, nz) { F_NORMAL3F,   { nx, ny, nz } }
#define C_End()                { F_END }
#define C_Translatef(x, y, z)  { F_TRANSLATEF, { x, y, z } }


/*
 * This matches the calls done by wglUseFontOutlines for Arial font.
 */
static const Call char_0 [] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_1[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_2[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_3[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_4[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_5[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_6[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_7[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_8[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_9[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_10[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_11[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_12[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_13[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_14[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_15[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_16[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_17[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_18[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_19[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_20[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_21[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_22[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_23[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_24[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_25[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_26[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_27[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_28[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_29[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_30[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_31[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_32[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_33[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.997530401f, -0.0702363178f, 0.f),
    C_Vertex3f(0.101074219f, 0.158691406f, 0.f),
    C_Vertex3f(0.101074219f, 0.158691406f, -1.f),
    C_Vertex3f(0.0771484375f, 0.498497009f, -1.f),
    C_Vertex3f(0.0771484375f, 0.498497009f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0771484375f, 0.498497009f, 0.f),
    C_Vertex3f(0.0771484375f, 0.498497009f, -1.f),
    C_Vertex3f(0.0771484375f, 0.640625f, -1.f),
    C_Vertex3f(0.0771484375f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0771484375f, 0.640625f, 0.f),
    C_Vertex3f(0.0771484375f, 0.640625f, -1.f),
    C_Vertex3f(0.174804688f, 0.640625f, -1.f),
    C_Vertex3f(0.174804688f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.174804688f, 0.640625f, 0.f),
    C_Vertex3f(0.174804688f, 0.640625f, -1.f),
    C_Vertex3f(0.174804688f, 0.498497009f, -1.f),
    C_Vertex3f(0.174804688f, 0.498497009f, 0.f),
    C_Normal3f(0.997727215f, -0.06738282f, 0.f),
    C_Vertex3f(0.174804688f, 0.498497009f, 0.f),
    C_Vertex3f(0.174804688f, 0.498497009f, -1.f),
    C_Vertex3f(0.151855469f, 0.158691406f, -1.f),
    C_Vertex3f(0.151855469f, 0.158691406f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.151855469f, 0.158691406f, 0.f),
    C_Vertex3f(0.151855469f, 0.158691406f, -1.f),
    C_Vertex3f(0.101074219f, 0.158691406f, -1.f),
    C_Vertex3f(0.101074219f, 0.158691406f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0805664063f, 0.f, 0.f),
    C_Vertex3f(0.0805664063f, 0.f, -1.f),
    C_Vertex3f(0.0805664063f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0805664063f, 0.0893554688f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0805664063f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0805664063f, 0.0893554688f, -1.f),
    C_Vertex3f(0.171386719f, 0.0893554688f, -1.f),
    C_Vertex3f(0.171386719f, 0.0893554688f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.171386719f, 0.0893554688f, 0.f),
    C_Vertex3f(0.171386719f, 0.0893554688f, -1.f),
    C_Vertex3f(0.171386719f, 0.f, -1.f),
    C_Vertex3f(0.171386719f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.171386719f, 0.f, 0.f),
    C_Vertex3f(0.171386719f, 0.f, -1.f),
    C_Vertex3f(0.0805664063f, 0.f, -1.f),
    C_Vertex3f(0.0805664063f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.174804688f, 0.640625f, 0.f),
    C_Vertex3f(0.0771484375f, 0.640625f, 0.f),
    C_Vertex3f(0.101074219f, 0.158691406f, 0.f),
    C_Vertex3f(0.151855469f, 0.158691406f, 0.f),
    C_Vertex3f(0.174804688f, 0.498497009f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0805664063f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0805664063f, 0.f, 0.f),
    C_Vertex3f(0.171386719f, 0.f, 0.f),
    C_Vertex3f(0.171386719f, 0.0893554688f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.101074219f, 0.158691406f, 0.f),
    C_Vertex3f(0.0771484375f, 0.640625f, 0.f),
    C_Vertex3f(0.0771484375f, 0.498497009f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.174804688f, 0.640625f, -1.f),
    C_Vertex3f(0.0771484375f, 0.640625f, -1.f),
    C_Vertex3f(0.101074219f, 0.158691406f, -1.f),
    C_Vertex3f(0.151855469f, 0.158691406f, -1.f),
    C_Vertex3f(0.174804688f, 0.498497009f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0805664063f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0805664063f, 0.f, -1.f),
    C_Vertex3f(0.171386719f, 0.f, -1.f),
    C_Vertex3f(0.171386719f, 0.0893554688f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.101074219f, 0.158691406f, -1.f),
    C_Vertex3f(0.0771484375f, 0.640625f, -1.f),
    C_Vertex3f(0.0771484375f, 0.498497009f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_34[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.984425306f, -0.175803423f, 0.f),
    C_Vertex3f(0.0628128052f, 0.413574219f, 0.f),
    C_Vertex3f(0.0628128052f, 0.413574219f, -1.f),
    C_Vertex3f(0.041015625f, 0.535629272f, -1.f),
    C_Vertex3f(0.041015625f, 0.535629272f, 0.f),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.041015625f, 0.535629272f, 0.f),
    C_Vertex3f(0.041015625f, 0.535629272f, -1.f),
    C_Vertex3f(0.041015625f, 0.640625f, -1.f),
    C_Vertex3f(0.041015625f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.041015625f, 0.640625f, 0.f),
    C_Vertex3f(0.041015625f, 0.640625f, -1.f),
    C_Vertex3f(0.130371094f, 0.640625f, -1.f),
    C_Vertex3f(0.130371094f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.130371094f, 0.640625f, 0.f),
    C_Vertex3f(0.130371094f, 0.640625f, -1.f),
    C_Vertex3f(0.130371094f, 0.535629272f, -1.f),
    C_Vertex3f(0.130371094f, 0.535629272f, 0.f),
    C_Normal3f(0.98733145f, -0.158671662f, 0.f),
    C_Vertex3f(0.130371094f, 0.535629272f, 0.f),
    C_Vertex3f(0.130371094f, 0.535629272f, -1.f),
    C_Vertex3f(0.11075592f, 0.413574219f, -1.f),
    C_Vertex3f(0.11075592f, 0.413574219f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.11075592f, 0.413574219f, 0.f),
    C_Vertex3f(0.11075592f, 0.413574219f, -1.f),
    C_Vertex3f(0.0628128052f, 0.413574219f, -1.f),
    C_Vertex3f(0.0628128052f, 0.413574219f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.985037327f, -0.172341511f, 0.f),
    C_Vertex3f(0.207878113f, 0.413574219f, 0.f),
    C_Vertex3f(0.207878113f, 0.413574219f, -1.f),
    C_Vertex3f(0.186523438f, 0.535629272f, -1.f),
    C_Vertex3f(0.186523438f, 0.535629272f, 0.f),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.186523438f, 0.535629272f, 0.f),
    C_Vertex3f(0.186523438f, 0.535629272f, -1.f),
    C_Vertex3f(0.186523438f, 0.640625f, -1.f),
    C_Vertex3f(0.186523438f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.186523438f, 0.640625f, 0.f),
    C_Vertex3f(0.186523438f, 0.640625f, -1.f),
    C_Vertex3f(0.275878906f, 0.640625f, -1.f),
    C_Vertex3f(0.275878906f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.275878906f, 0.640625f, 0.f),
    C_Vertex3f(0.275878906f, 0.640625f, -1.f),
    C_Vertex3f(0.275878906f, 0.535629272f, -1.f),
    C_Vertex3f(0.275878906f, 0.535629272f, 0.f),
    C_Normal3f(0.985627592f, -0.168933049f, 0.f),
    C_Vertex3f(0.275878906f, 0.535629272f, 0.f),
    C_Vertex3f(0.275878906f, 0.535629272f, -1.f),
    C_Vertex3f(0.254959106f, 0.413574219f, -1.f),
    C_Vertex3f(0.254959106f, 0.413574219f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.254959106f, 0.413574219f, 0.f),
    C_Vertex3f(0.254959106f, 0.413574219f, -1.f),
    C_Vertex3f(0.207878113f, 0.413574219f, -1.f),
    C_Vertex3f(0.207878113f, 0.413574219f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130371094f, 0.640625f, 0.f),
    C_Vertex3f(0.041015625f, 0.640625f, 0.f),
    C_Vertex3f(0.0628128052f, 0.413574219f, 0.f),
    C_Vertex3f(0.11075592f, 0.413574219f, 0.f),
    C_Vertex3f(0.130371094f, 0.535629272f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.275878906f, 0.640625f, 0.f),
    C_Vertex3f(0.186523438f, 0.640625f, 0.f),
    C_Vertex3f(0.207878113f, 0.413574219f, 0.f),
    C_Vertex3f(0.254959106f, 0.413574219f, 0.f),
    C_Vertex3f(0.275878906f, 0.535629272f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.207878113f, 0.413574219f, 0.f),
    C_Vertex3f(0.186523438f, 0.640625f, 0.f),
    C_Vertex3f(0.186523438f, 0.535629272f, 0.f),
    C_Vertex3f(0.0628128052f, 0.413574219f, 0.f),
    C_Vertex3f(0.041015625f, 0.640625f, 0.f),
    C_Vertex3f(0.041015625f, 0.535629272f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130371094f, 0.640625f, -1.f),
    C_Vertex3f(0.041015625f, 0.640625f, -1.f),
    C_Vertex3f(0.0628128052f, 0.413574219f, -1.f),
    C_Vertex3f(0.11075592f, 0.413574219f, -1.f),
    C_Vertex3f(0.130371094f, 0.535629272f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.275878906f, 0.640625f, -1.f),
    C_Vertex3f(0.186523438f, 0.640625f, -1.f),
    C_Vertex3f(0.207878113f, 0.413574219f, -1.f),
    C_Vertex3f(0.254959106f, 0.413574219f, -1.f),
    C_Vertex3f(0.275878906f, 0.535629272f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.207878113f, 0.413574219f, -1.f),
    C_Vertex3f(0.186523438f, 0.640625f, -1.f),
    C_Vertex3f(0.186523438f, 0.535629272f, -1.f),
    C_Vertex3f(0.0628128052f, 0.413574219f, -1.f),
    C_Vertex3f(0.041015625f, 0.640625f, -1.f),
    C_Vertex3f(0.041015625f, 0.535629272f, -1.f),
    C_End(),
    C_Translatef(0.317871094f, 0.f, 0.f),
};
static const Call char_35[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.979767263f, 0.200140283f, 0.f),
    C_Vertex3f(0.044921875f, -0.0107421875f, 0.f),
    C_Vertex3f(0.044921875f, -0.0107421875f, -1.f),
    C_Vertex3f(0.0829238892f, 0.175292969f, -1.f),
    C_Vertex3f(0.0829238892f, 0.175292969f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0829238892f, 0.175292969f, 0.f),
    C_Vertex3f(0.0829238892f, 0.175292969f, -1.f),
    C_Vertex3f(0.00927734375f, 0.175292969f, -1.f),
    C_Vertex3f(0.00927734375f, 0.175292969f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.00927734375f, 0.175292969f, 0.f),
    C_Vertex3f(0.00927734375f, 0.175292969f, -1.f),
    C_Vertex3f(0.00927734375f, 0.240234375f, -1.f),
    C_Vertex3f(0.00927734375f, 0.240234375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.00927734375f, 0.240234375f, 0.f),
    C_Vertex3f(0.00927734375f, 0.240234375f, -1.f),
    C_Vertex3f(0.0961914063f, 0.240234375f, -1.f),
    C_Vertex3f(0.0961914063f, 0.240234375f, 0.f),
    C_Normal3f(-0.97976613f, 0.200145498f, 0.f),
    C_Vertex3f(0.0961914063f, 0.240234375f, 0.f),
    C_Vertex3f(0.0961914063f, 0.240234375f, -1.f),
    C_Vertex3f(0.128608704f, 0.398925781f, -1.f),
    C_Vertex3f(0.128608704f, 0.398925781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.128608704f, 0.398925781f, 0.f),
    C_Vertex3f(0.128608704f, 0.398925781f, -1.f),
    C_Vertex3f(0.00927734375f, 0.398925781f, -1.f),
    C_Vertex3f(0.00927734375f, 0.398925781f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.00927734375f, 0.398925781f, 0.f),
    C_Vertex3f(0.00927734375f, 0.398925781f, -1.f),
    C_Vertex3f(0.00927734375f, 0.463867188f, -1.f),
    C_Vertex3f(0.00927734375f, 0.463867188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.00927734375f, 0.463867188f, 0.f),
    C_Vertex3f(0.00927734375f, 0.463867188f, -1.f),
    C_Vertex3f(0.141876221f, 0.463867188f, -1.f),
    C_Vertex3f(0.141876221f, 0.463867188f, 0.f),
    C_Normal3f(-0.979768932f, 0.200131834f, 0.f),
    C_Vertex3f(0.141876221f, 0.463867188f, 0.f),
    C_Vertex3f(0.141876221f, 0.463867188f, -1.f),
    C_Vertex3f(0.180175781f, 0.651367188f, -1.f),
    C_Vertex3f(0.180175781f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.180175781f, 0.651367188f, 0.f),
    C_Vertex3f(0.180175781f, 0.651367188f, -1.f),
    C_Vertex3f(0.245735168f, 0.651367188f, -1.f),
    C_Vertex3f(0.245735168f, 0.651367188f, 0.f),
    C_Normal3f(0.979768932f, -0.200131834f, 0.f),
    C_Vertex3f(0.245735168f, 0.651367188f, 0.f),
    C_Vertex3f(0.245735168f, 0.651367188f, -1.f),
    C_Vertex3f(0.207435608f, 0.463867188f, -1.f),
    C_Vertex3f(0.207435608f, 0.463867188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.207435608f, 0.463867188f, 0.f),
    C_Vertex3f(0.207435608f, 0.463867188f, -1.f),
    C_Vertex3f(0.344863892f, 0.463867188f, -1.f),
    C_Vertex3f(0.344863892f, 0.463867188f, 0.f),
    C_Normal3f(-0.979768932f, 0.200131834f, 0.f),
    C_Vertex3f(0.344863892f, 0.463867188f, 0.f),
    C_Vertex3f(0.344863892f, 0.463867188f, -1.f),
    C_Vertex3f(0.383163452f, 0.651367188f, -1.f),
    C_Vertex3f(0.383163452f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.383163452f, 0.651367188f, 0.f),
    C_Vertex3f(0.383163452f, 0.651367188f, -1.f),
    C_Vertex3f(0.448730469f, 0.651367188f, -1.f),
    C_Vertex3f(0.448730469f, 0.651367188f, 0.f),
    C_Normal3f(0.979768932f, -0.200131834f, 0.f),
    C_Vertex3f(0.448730469f, 0.651367188f, 0.f),
    C_Vertex3f(0.448730469f, 0.651367188f, -1.f),
    C_Vertex3f(0.410430908f, 0.463867188f, -1.f),
    C_Vertex3f(0.410430908f, 0.463867188f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.410430908f, 0.463867188f, 0.f),
    C_Vertex3f(0.410430908f, 0.463867188f, -1.f),
    C_Vertex3f(0.486816406f, 0.463867188f, -1.f),
    C_Vertex3f(0.486816406f, 0.463867188f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.486816406f, 0.463867188f, 0.f),
    C_Vertex3f(0.486816406f, 0.463867188f, -1.f),
    C_Vertex3f(0.486816406f, 0.398925781f, -1.f),
    C_Vertex3f(0.486816406f, 0.398925781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.486816406f, 0.398925781f, 0.f),
    C_Vertex3f(0.486816406f, 0.398925781f, -1.f),
    C_Vertex3f(0.397163391f, 0.398925781f, -1.f),
    C_Vertex3f(0.397163391f, 0.398925781f, 0.f),
    C_Normal3f(0.97976613f, -0.200145498f, 0.f),
    C_Vertex3f(0.397163391f, 0.398925781f, 0.f),
    C_Vertex3f(0.397163391f, 0.398925781f, -1.f),
    C_Vertex3f(0.364746094f, 0.240234375f, -1.f),
    C_Vertex3f(0.364746094f, 0.240234375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.364746094f, 0.240234375f, 0.f),
    C_Vertex3f(0.364746094f, 0.240234375f, -1.f),
    C_Vertex3f(0.486816406f, 0.240234375f, -1.f),
    C_Vertex3f(0.486816406f, 0.240234375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.486816406f, 0.240234375f, 0.f),
    C_Vertex3f(0.486816406f, 0.240234375f, -1.f),
    C_Vertex3f(0.486816406f, 0.175292969f, -1.f),
    C_Vertex3f(0.486816406f, 0.175292969f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.486816406f, 0.175292969f, 0.f),
    C_Vertex3f(0.486816406f, 0.175292969f, -1.f),
    C_Vertex3f(0.351478577f, 0.175292969f, -1.f),
    C_Vertex3f(0.351478577f, 0.175292969f, 0.f),
    C_Normal3f(0.979767263f, -0.200140283f, 0.f),
    C_Vertex3f(0.351478577f, 0.175292969f, 0.f),
    C_Vertex3f(0.351478577f, 0.175292969f, -1.f),
    C_Vertex3f(0.313476563f, -0.0107421875f, -1.f),
    C_Vertex3f(0.313476563f, -0.0107421875f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.313476563f, -0.0107421875f, 0.f),
    C_Vertex3f(0.313476563f, -0.0107421875f, -1.f),
    C_Vertex3f(0.247917175f, -0.0107421875f, -1.f),
    C_Vertex3f(0.247917175f, -0.0107421875f, 0.f),
    C_Normal3f(-0.979767263f, 0.200140283f, 0.f),
    C_Vertex3f(0.247917175f, -0.0107421875f, 0.f),
    C_Vertex3f(0.247917175f, -0.0107421875f, -1.f),
    C_Vertex3f(0.285919189f, 0.175292969f, -1.f),
    C_Vertex3f(0.285919189f, 0.175292969f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.285919189f, 0.175292969f, 0.f),
    C_Vertex3f(0.285919189f, 0.175292969f, -1.f),
    C_Vertex3f(0.148490906f, 0.175292969f, -1.f),
    C_Vertex3f(0.148490906f, 0.175292969f, 0.f),
    C_Normal3f(0.979767263f, -0.200140283f, 0.f),
    C_Vertex3f(0.148490906f, 0.175292969f, 0.f),
    C_Vertex3f(0.148490906f, 0.175292969f, -1.f),
    C_Vertex3f(0.110488892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.110488892f, -0.0107421875f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.110488892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.110488892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.044921875f, -0.0107421875f, -1.f),
    C_Vertex3f(0.044921875f, -0.0107421875f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.161758423f, 0.240234375f, 0.f),
    C_Vertex3f(0.161758423f, 0.240234375f, -1.f),
    C_Vertex3f(0.299186707f, 0.240234375f, -1.f),
    C_Vertex3f(0.299186707f, 0.240234375f, 0.f),
    C_Normal3f(-0.979775369f, 0.200100288f, 0.f),
    C_Vertex3f(0.299186707f, 0.240234375f, 0.f),
    C_Vertex3f(0.299186707f, 0.240234375f, -1.f),
    C_Vertex3f(0.331596375f, 0.398925781f, -1.f),
    C_Vertex3f(0.331596375f, 0.398925781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.331596375f, 0.398925781f, 0.f),
    C_Vertex3f(0.331596375f, 0.398925781f, -1.f),
    C_Vertex3f(0.194168091f, 0.398925781f, -1.f),
    C_Vertex3f(0.194168091f, 0.398925781f, 0.f),
    C_Normal3f(0.979775369f, -0.200100288f, 0.f),
    C_Vertex3f(0.194168091f, 0.398925781f, 0.f),
    C_Vertex3f(0.194168091f, 0.398925781f, -1.f),
    C_Vertex3f(0.161758423f, 0.240234375f, -1.f),
    C_Vertex3f(0.161758423f, 0.240234375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.180175781f, 0.651367188f, 0.f),
    C_Vertex3f(0.141876221f, 0.463867188f, 0.f),
    C_Vertex3f(0.148490906f, 0.175292969f, 0.f),
    C_Vertex3f(0.128608704f, 0.398925781f, 0.f),
    C_Vertex3f(0.110488892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.0961914063f, 0.240234375f, 0.f),
    C_Vertex3f(0.0829238892f, 0.175292969f, 0.f),
    C_Vertex3f(0.00927734375f, 0.240234375f, 0.f),
    C_Vertex3f(0.00927734375f, 0.175292969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.00927734375f, 0.463867188f, 0.f),
    C_Vertex3f(0.00927734375f, 0.398925781f, 0.f),
    C_Vertex3f(0.128608704f, 0.398925781f, 0.f),
    C_Vertex3f(0.141876221f, 0.463867188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.245735168f, 0.651367188f, 0.f),
    C_Vertex3f(0.194168091f, 0.398925781f, 0.f),
    C_Vertex3f(0.207435608f, 0.463867188f, 0.f),
    C_Vertex3f(0.331596375f, 0.398925781f, 0.f),
    C_Vertex3f(0.344863892f, 0.463867188f, 0.f),
    C_Vertex3f(0.351478577f, 0.175292969f, 0.f),
    C_Vertex3f(0.383163452f, 0.651367188f, 0.f),
    C_Vertex3f(0.364746094f, 0.240234375f, 0.f),
    C_Vertex3f(0.397163391f, 0.398925781f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.180175781f, 0.651367188f, 0.f),
    C_Vertex3f(0.148490906f, 0.175292969f, 0.f),
    C_Vertex3f(0.161758423f, 0.240234375f, 0.f),
    C_Vertex3f(0.194168091f, 0.398925781f, 0.f),
    C_Vertex3f(0.245735168f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.351478577f, 0.175292969f, 0.f),
    C_Vertex3f(0.331596375f, 0.398925781f, 0.f),
    C_Vertex3f(0.313476563f, -0.0107421875f, 0.f),
    C_Vertex3f(0.299186707f, 0.240234375f, 0.f),
    C_Vertex3f(0.285919189f, 0.175292969f, 0.f),
    C_Vertex3f(0.161758423f, 0.240234375f, 0.f),
    C_Vertex3f(0.148490906f, 0.175292969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.397163391f, 0.398925781f, 0.f),
    C_Vertex3f(0.486816406f, 0.398925781f, 0.f),
    C_Vertex3f(0.410430908f, 0.463867188f, 0.f),
    C_Vertex3f(0.448730469f, 0.651367188f, 0.f),
    C_Vertex3f(0.383163452f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.364746094f, 0.240234375f, 0.f),
    C_Vertex3f(0.351478577f, 0.175292969f, 0.f),
    C_Vertex3f(0.486816406f, 0.175292969f, 0.f),
    C_Vertex3f(0.486816406f, 0.240234375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.410430908f, 0.463867188f, 0.f),
    C_Vertex3f(0.486816406f, 0.398925781f, 0.f),
    C_Vertex3f(0.486816406f, 0.463867188f, 0.f),
    C_Vertex3f(0.313476563f, -0.0107421875f, 0.f),
    C_Vertex3f(0.285919189f, 0.175292969f, 0.f),
    C_Vertex3f(0.247917175f, -0.0107421875f, 0.f),
    C_Vertex3f(0.110488892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.0829238892f, 0.175292969f, 0.f),
    C_Vertex3f(0.044921875f, -0.0107421875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.180175781f, 0.651367188f, -1.f),
    C_Vertex3f(0.141876221f, 0.463867188f, -1.f),
    C_Vertex3f(0.148490906f, 0.175292969f, -1.f),
    C_Vertex3f(0.128608704f, 0.398925781f, -1.f),
    C_Vertex3f(0.110488892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.0961914063f, 0.240234375f, -1.f),
    C_Vertex3f(0.0829238892f, 0.175292969f, -1.f),
    C_Vertex3f(0.00927734375f, 0.240234375f, -1.f),
    C_Vertex3f(0.00927734375f, 0.175292969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.00927734375f, 0.463867188f, -1.f),
    C_Vertex3f(0.00927734375f, 0.398925781f, -1.f),
    C_Vertex3f(0.128608704f, 0.398925781f, -1.f),
    C_Vertex3f(0.141876221f, 0.463867188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.245735168f, 0.651367188f, -1.f),
    C_Vertex3f(0.194168091f, 0.398925781f, -1.f),
    C_Vertex3f(0.207435608f, 0.463867188f, -1.f),
    C_Vertex3f(0.331596375f, 0.398925781f, -1.f),
    C_Vertex3f(0.344863892f, 0.463867188f, -1.f),
    C_Vertex3f(0.351478577f, 0.175292969f, -1.f),
    C_Vertex3f(0.383163452f, 0.651367188f, -1.f),
    C_Vertex3f(0.364746094f, 0.240234375f, -1.f),
    C_Vertex3f(0.397163391f, 0.398925781f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.180175781f, 0.651367188f, -1.f),
    C_Vertex3f(0.148490906f, 0.175292969f, -1.f),
    C_Vertex3f(0.161758423f, 0.240234375f, -1.f),
    C_Vertex3f(0.194168091f, 0.398925781f, -1.f),
    C_Vertex3f(0.245735168f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.351478577f, 0.175292969f, -1.f),
    C_Vertex3f(0.331596375f, 0.398925781f, -1.f),
    C_Vertex3f(0.313476563f, -0.0107421875f, -1.f),
    C_Vertex3f(0.299186707f, 0.240234375f, -1.f),
    C_Vertex3f(0.285919189f, 0.175292969f, -1.f),
    C_Vertex3f(0.161758423f, 0.240234375f, -1.f),
    C_Vertex3f(0.148490906f, 0.175292969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.397163391f, 0.398925781f, -1.f),
    C_Vertex3f(0.486816406f, 0.398925781f, -1.f),
    C_Vertex3f(0.410430908f, 0.463867188f, -1.f),
    C_Vertex3f(0.448730469f, 0.651367188f, -1.f),
    C_Vertex3f(0.383163452f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.364746094f, 0.240234375f, -1.f),
    C_Vertex3f(0.351478577f, 0.175292969f, -1.f),
    C_Vertex3f(0.486816406f, 0.175292969f, -1.f),
    C_Vertex3f(0.486816406f, 0.240234375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.410430908f, 0.463867188f, -1.f),
    C_Vertex3f(0.486816406f, 0.398925781f, -1.f),
    C_Vertex3f(0.486816406f, 0.463867188f, -1.f),
    C_Vertex3f(0.313476563f, -0.0107421875f, -1.f),
    C_Vertex3f(0.285919189f, 0.175292969f, -1.f),
    C_Vertex3f(0.247917175f, -0.0107421875f, -1.f),
    C_Vertex3f(0.110488892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.0829238892f, 0.175292969f, -1.f),
    C_Vertex3f(0.044921875f, -0.0107421875f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_36[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.222167969f, -0.0922851563f, 0.f),
    C_Vertex3f(0.222167969f, -0.0922851563f, -1.f),
    C_Vertex3f(0.222167969f, -0.013671875f, -1.f),
    C_Vertex3f(0.222167969f, -0.013671875f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.12531729f, -0.99211669f, 0.f),
    C_Vertex3f(0.222167969f, -0.013671875f, 0.f),
    C_Vertex3f(0.222167969f, -0.013671875f, -1.f),
    C_Normal3f(-0.167302385f, -0.985905647f, 0.f),
    C_Vertex3f(0.207460403f, -0.0118141174f, 0.f),
    C_Vertex3f(0.207460403f, -0.0118141174f, -1.f),
    C_Normal3f(-0.226709396f, -0.973962486f, 0.f),
    C_Vertex3f(0.180796623f, -0.00665140152f, 0.f),
    C_Vertex3f(0.180796623f, -0.00665140152f, -1.f),
    C_Normal3f(-0.309432268f, -0.950921476f, 0.f),
    C_Vertex3f(0.156884193f, -4.14848328e-005f, 0.f),
    C_Vertex3f(0.156884193f, -4.14848328e-005f, -1.f),
    C_Normal3f(-0.405644834f, -0.91403079f, 0.f),
    C_Vertex3f(0.135723114f, 0.00801563263f, 0.f),
    C_Vertex3f(0.135723114f, 0.00801563263f, -1.f),
    C_Normal3f(-0.524854243f, -0.851192117f, 0.f),
    C_Vertex3f(0.117313385f, 0.0175199509f, 0.f),
    C_Vertex3f(0.117313385f, 0.0175199509f, -1.f),
    C_Normal3f(-0.643237889f, -0.765666306f, 0.f),
    C_Vertex3f(0.10015583f, 0.029946804f, 0.f),
    C_Vertex3f(0.10015583f, 0.029946804f, -1.f),
    C_Normal3f(-0.740008473f, -0.672597587f, 0.f),
    C_Vertex3f(0.0842504501f, 0.0452961922f, 0.f),
    C_Vertex3f(0.0842504501f, 0.0452961922f, -1.f),
    C_Normal3f(-0.815060079f, -0.5793764f, 0.f),
    C_Vertex3f(0.0695972443f, 0.0635681152f, 0.f),
    C_Vertex3f(0.0695972443f, 0.0635681152f, -1.f),
    C_Normal3f(-0.881236613f, -0.472675413f, 0.f),
    C_Vertex3f(0.0561962128f, 0.0847625732f, 0.f),
    C_Vertex3f(0.0561962128f, 0.0847625732f, -1.f),
    C_Normal3f(-0.937281966f, -0.348571897f, 0.f),
    C_Vertex3f(0.0456008911f, 0.108305931f, 0.f),
    C_Vertex3f(0.0456008911f, 0.108305931f, -1.f),
    C_Normal3f(-0.973271728f, -0.229656667f, 0.f),
    C_Vertex3f(0.0378112793f, 0.134198189f, 0.f),
    C_Vertex3f(0.0378112793f, 0.134198189f, -1.f),
    C_Normal3f(-0.990410566f, -0.138155445f, 0.f),
    C_Vertex3f(0.0328273773f, 0.162439346f, 0.f),
    C_Vertex3f(0.0328273773f, 0.162439346f, -1.f),
    C_Normal3f(-0.997474432f, -0.0710260496f, 0.f),
    C_Vertex3f(0.0317382813f, 0.177734375f, 0.f),
    C_Vertex3f(0.0317382813f, 0.177734375f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.182089254f, 0.98328203f, 0.f),
    C_Vertex3f(0.0317382813f, 0.177734375f, 0.f),
    C_Vertex3f(0.0317382813f, 0.177734375f, -1.f),
    C_Vertex3f(0.110839844f, 0.192382813f, -1.f),
    C_Vertex3f(0.110839844f, 0.192382813f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.989745557f, 0.142841265f, 0.f),
    C_Vertex3f(0.110839844f, 0.192382813f, 0.f),
    C_Vertex3f(0.110839844f, 0.192382813f, -1.f),
    C_Normal3f(0.98582685f, 0.167765915f, 0.f),
    C_Vertex3f(0.111981392f, 0.184473038f, 0.f),
    C_Vertex3f(0.111981392f, 0.184473038f, -1.f),
    C_Normal3f(0.979420185f, 0.201831445f, 0.f),
    C_Vertex3f(0.114699364f, 0.169703484f, 0.f),
    C_Vertex3f(0.114699364f, 0.169703484f, -1.f),
    C_Normal3f(0.968891919f, 0.247483939f, 0.f),
    C_Vertex3f(0.117852211f, 0.155983925f, 0.f),
    C_Vertex3f(0.117852211f, 0.155983925f, -1.f),
    C_Normal3f(0.954256356f, 0.298989564f, 0.f),
    C_Vertex3f(0.121439934f, 0.143314362f, 0.f),
    C_Vertex3f(0.121439934f, 0.143314362f, -1.f),
    C_Normal3f(0.934108198f, 0.356990159f, 0.f),
    C_Vertex3f(0.125462532f, 0.131694794f, 0.f),
    C_Vertex3f(0.125462532f, 0.131694794f, -1.f),
    C_Normal3f(0.906616926f, 0.421954513f, 0.f),
    C_Vertex3f(0.129920006f, 0.121125221f, 0.f),
    C_Vertex3f(0.129920006f, 0.121125221f, -1.f),
    C_Normal3f(0.869488537f, 0.493953198f, 0.f),
    C_Vertex3f(0.134812355f, 0.111605644f, 0.f),
    C_Vertex3f(0.134812355f, 0.111605644f, -1.f),
    C_Normal3f(0.81595248f, 0.57811898f, 0.f),
    C_Vertex3f(0.14013958f, 0.103136063f, 0.f),
    C_Vertex3f(0.14013958f, 0.103136063f, -1.f),
    C_Normal3f(0.758622766f, 0.651530087f, 0.f),
    C_Vertex3f(0.147151947f, 0.0941896439f, 0.f),
    C_Vertex3f(0.147151947f, 0.0941896439f, -1.f),
    C_Normal3f(0.701474249f, 0.712694764f, 0.f),
    C_Vertex3f(0.155822277f, 0.0848755836f, 0.f),
    C_Vertex3f(0.155822277f, 0.0848755836f, -1.f),
    C_Normal3f(0.632287979f, 0.774733365f, 0.f),
    C_Vertex3f(0.164900303f, 0.0767207146f, 0.f),
    C_Vertex3f(0.164900303f, 0.0767207146f, -1.f),
    C_Normal3f(0.552098334f, 0.833779037f, 0.f),
    C_Vertex3f(0.174386024f, 0.0697250366f, 0.f),
    C_Vertex3f(0.174386024f, 0.0697250366f, -1.f),
    C_Normal3f(0.46178925f, 0.886989713f, 0.f),
    C_Vertex3f(0.184279442f, 0.0638885498f, 0.f),
    C_Vertex3f(0.184279442f, 0.0638885498f, -1.f),
    C_Normal3f(0.363404214f, 0.931631625f, 0.f),
    C_Vertex3f(0.194580555f, 0.0592112541f, 0.f),
    C_Vertex3f(0.194580555f, 0.0592112541f, -1.f),
    C_Normal3f(0.26001358f, 0.965604961f, 0.f),
    C_Vertex3f(0.205289364f, 0.0556931496f, 0.f),
    C_Vertex3f(0.205289364f, 0.0556931496f, -1.f),
    C_Normal3f(0.172664493f, 0.984980643f, 0.f),
    C_Vertex3f(0.216405869f, 0.0533342361f, 0.f),
    C_Vertex3f(0.216405869f, 0.0533342361f, -1.f),
    C_Normal3f(0.103545018f, 0.994624794f, 0.f),
    C_Vertex3f(0.222167969f, 0.052734375f, 0.f),
    C_Vertex3f(0.222167969f, 0.052734375f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.222167969f, 0.052734375f, 0.f),
    C_Vertex3f(0.222167969f, 0.052734375f, -1.f),
    C_Vertex3f(0.222167969f, 0.300125122f, -1.f),
    C_Vertex3f(0.222167969f, 0.300125122f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.248605967f, -0.968604743f, 0.f),
    C_Vertex3f(0.222167969f, 0.300125122f, 0.f),
    C_Vertex3f(0.222167969f, 0.300125122f, -1.f),
    C_Normal3f(-0.295930803f, -0.955209434f, 0.f),
    C_Vertex3f(0.21021843f, 0.303192139f, 0.f),
    C_Vertex3f(0.21021843f, 0.303192139f, -1.f),
    C_Normal3f(-0.350919276f, -0.936405778f, 0.f),
    C_Vertex3f(0.185955048f, 0.311344147f, 0.f),
    C_Vertex3f(0.185955048f, 0.311344147f, -1.f),
    C_Normal3f(-0.41080907f, -0.911721408f, 0.f),
    C_Vertex3f(0.161327362f, 0.32151413f, 0.f),
    C_Vertex3f(0.161327362f, 0.32151413f, -1.f),
    C_Normal3f(-0.462284803f, -0.886731565f, 0.f),
    C_Vertex3f(0.136335373f, 0.333702087f, 0.f),
    C_Vertex3f(0.136335373f, 0.333702087f, -1.f),
    C_Normal3f(-0.540422857f, -0.84139359f, 0.f),
    C_Vertex3f(0.11425972f, 0.34605217f, 0.f),
    C_Vertex3f(0.11425972f, 0.34605217f, -1.f),
    C_Normal3f(-0.656078458f, -0.754692614f, 0.f),
    C_Vertex3f(0.0976223946f, 0.358567238f, 0.f),
    C_Vertex3f(0.0976223946f, 0.358567238f, -1.f),
    C_Normal3f(-0.757439077f, -0.652905822f, 0.f),
    C_Vertex3f(0.0831427574f, 0.373103142f, 0.f),
    C_Vertex3f(0.0831427574f, 0.373103142f, -1.f),
    C_Normal3f(-0.842268705f, -0.53905791f, 0.f),
    C_Vertex3f(0.0708208084f, 0.389659882f, 0.f),
    C_Vertex3f(0.0708208084f, 0.389659882f, -1.f),
    C_Normal3f(-0.907430172f, -0.420202941f, 0.f),
    C_Vertex3f(0.0606565475f, 0.408237457f, 0.f),
    C_Vertex3f(0.0606565475f, 0.408237457f, -1.f),
    C_Normal3f(-0.95503664f, -0.296487749f, 0.f),
    C_Vertex3f(0.0530333519f, 0.428071976f, 0.f),
    C_Vertex3f(0.0530333519f, 0.428071976f, -1.f),
    C_Normal3f(-0.984948635f, -0.172847137f, 0.f),
    C_Vertex3f(0.0479512215f, 0.449163437f, 0.f),
    C_Vertex3f(0.0479512215f, 0.449163437f, -1.f),
    C_Normal3f(-0.998383999f, -0.0568283983f, 0.f),
    C_Vertex3f(0.0454101563f, 0.471511841f, 0.f),
    C_Vertex3f(0.0454101563f, 0.471511841f, -1.f),
    C_Normal3f(-0.999043703f, 0.0437230207f, 0.f),
    C_Vertex3f(0.0454101563f, 0.493805885f, 0.f),
    C_Vertex3f(0.0454101563f, 0.493805885f, -1.f),
    C_Normal3f(-0.990008652f, 0.141006738f, 0.f),
    C_Vertex3f(0.0472605228f, 0.513791561f, 0.f),
    C_Vertex3f(0.0472605228f, 0.513791561f, -1.f),
    C_Normal3f(-0.970100522f, 0.242703378f, 0.f),
    C_Vertex3f(0.050961256f, 0.53278017f, 0.f),
    C_Vertex3f(0.050961256f, 0.53278017f, -1.f),
    C_Normal3f(-0.937794328f, 0.347191334f, 0.f),
    C_Vertex3f(0.0565123558f, 0.550771713f, 0.f),
    C_Vertex3f(0.0565123558f, 0.550771713f, -1.f),
    C_Normal3f(-0.89271754f, 0.450616628f, 0.f),
    C_Vertex3f(0.0639138222f, 0.56776619f, 0.f),
    C_Vertex3f(0.0639138222f, 0.56776619f, -1.f),
    C_Normal3f(-0.835902452f, 0.548878014f, 0.f),
    C_Vertex3f(0.0731656551f, 0.583763599f, 0.f),
    C_Vertex3f(0.0731656551f, 0.583763599f, -1.f),
    C_Normal3f(-0.769719124f, 0.638382792f, 0.f),
    C_Vertex3f(0.0842678547f, 0.598763943f, 0.f),
    C_Vertex3f(0.0842678547f, 0.598763943f, -1.f),
    C_Normal3f(-0.7010445f, 0.71311754f, 0.f),
    C_Vertex3f(0.0972204208f, 0.61276722f, 0.f),
    C_Vertex3f(0.0972204208f, 0.61276722f, -1.f),
    C_Normal3f(-0.624821782f, 0.780767322f, 0.f),
    C_Vertex3f(0.109574318f, 0.623641968f, 0.f),
    C_Vertex3f(0.109574318f, 0.623641968f, -1.f),
    C_Normal3f(-0.538251579f, 0.842784226f, 0.f),
    C_Vertex3f(0.120676279f, 0.631538153f, 0.f),
    C_Vertex3f(0.120676279f, 0.631538153f, -1.f),
    C_Normal3f(-0.456937611f, 0.889498711f, 0.f),
    C_Vertex3f(0.13297534f, 0.638587236f, 0.f),
    C_Vertex3f(0.13297534f, 0.638587236f, -1.f),
    C_Normal3f(-0.379330844f, 0.92526108f, 0.f),
    C_Vertex3f(0.1464715f, 0.644789219f, 0.f),
    C_Vertex3f(0.1464715f, 0.644789219f, -1.f),
    C_Normal3f(-0.306917578f, 0.951736093f, 0.f),
    C_Vertex3f(0.161164761f, 0.6501441f, 0.f),
    C_Vertex3f(0.161164761f, 0.6501441f, -1.f),
    C_Normal3f(-0.240429789f, 0.970666587f, 0.f),
    C_Vertex3f(0.17705512f, 0.65465188f, 0.f),
    C_Vertex3f(0.17705512f, 0.65465188f, -1.f),
    C_Normal3f(-0.180042297f, 0.983658791f, 0.f),
    C_Vertex3f(0.19414258f, 0.658312559f, 0.f),
    C_Vertex3f(0.19414258f, 0.658312559f, -1.f),
    C_Normal3f(-0.134251282f, 0.990947306f, 0.f),
    C_Vertex3f(0.212427139f, 0.661126137f, 0.f),
    C_Vertex3f(0.212427139f, 0.661126137f, -1.f),
    C_Normal3f(-0.100429557f, 0.994944215f, 0.f),
    C_Vertex3f(0.222167969f, 0.662109375f, 0.f),
    C_Vertex3f(0.222167969f, 0.662109375f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.222167969f, 0.662109375f, 0.f),
    C_Vertex3f(0.222167969f, 0.662109375f, -1.f),
    C_Vertex3f(0.222167969f, 0.699707031f, -1.f),
    C_Vertex3f(0.222167969f, 0.699707031f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.222167969f, 0.699707031f, 0.f),
    C_Vertex3f(0.222167969f, 0.699707031f, -1.f),
    C_Vertex3f(0.268554688f, 0.699707031f, -1.f),
    C_Vertex3f(0.268554688f, 0.699707031f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.268554688f, 0.699707031f, 0.f),
    C_Vertex3f(0.268554688f, 0.699707031f, -1.f),
    C_Vertex3f(0.268554688f, 0.662109375f, -1.f),
    C_Vertex3f(0.268554688f, 0.662109375f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.0959435776f, 0.995386779f, 0.f),
    C_Vertex3f(0.268554688f, 0.662109375f, 0.f),
    C_Vertex3f(0.268554688f, 0.662109375f, -1.f),
    C_Normal3f(0.133008912f, 0.991114914f, 0.f),
    C_Vertex3f(0.277083397f, 0.661287308f, 0.f),
    C_Vertex3f(0.277083397f, 0.661287308f, -1.f),
    C_Normal3f(0.182337344f, 0.983235955f, 0.f),
    C_Vertex3f(0.293257952f, 0.658794165f, 0.f),
    C_Vertex3f(0.293257952f, 0.658794165f, -1.f),
    C_Normal3f(0.245857179f, 0.969306171f, 0.f),
    C_Vertex3f(0.308549643f, 0.655452013f, 0.f),
    C_Vertex3f(0.308549643f, 0.655452013f, -1.f),
    C_Normal3f(0.313771278f, 0.949498653f, 0.f),
    C_Vertex3f(0.322958469f, 0.651260853f, 0.f),
    C_Vertex3f(0.322958469f, 0.651260853f, -1.f),
    C_Normal3f(0.385383517f, 0.922756493f, 0.f),
    C_Vertex3f(0.336484432f, 0.646220684f, 0.f),
    C_Vertex3f(0.336484432f, 0.646220684f, -1.f),
    C_Normal3f(0.459565043f, 0.888144195f, 0.f),
    C_Vertex3f(0.349127531f, 0.640331507f, 0.f),
    C_Vertex3f(0.349127531f, 0.640331507f, -1.f),
    C_Normal3f(0.534738481f, 0.845017672f, 0.f),
    C_Vertex3f(0.360887766f, 0.633593321f, 0.f),
    C_Vertex3f(0.360887766f, 0.633593321f, -1.f),
    C_Normal3f(0.610403895f, 0.792090356f, 0.f),
    C_Vertex3f(0.371765137f, 0.626006126f, 0.f),
    C_Vertex3f(0.371765137f, 0.626006126f, -1.f),
    C_Normal3f(0.680616021f, 0.732640386f, 0.f),
    C_Vertex3f(0.383172989f, 0.616419792f, 0.f),
    C_Vertex3f(0.383172989f, 0.616419792f, -1.f),
    C_Normal3f(0.746510267f, 0.665373981f, 0.f),
    C_Vertex3f(0.394717455f, 0.604683638f, 0.f),
    C_Vertex3f(0.394717455f, 0.604683638f, -1.f),
    C_Normal3f(0.808636844f, 0.588308096f, 0.f),
    C_Vertex3f(0.404985189f, 0.591947794f, 0.f),
    C_Vertex3f(0.404985189f, 0.591947794f, -1.f),
    C_Normal3f(0.86249274f, 0.506069422f, 0.f),
    C_Vertex3f(0.413976192f, 0.578212261f, 0.f),
    C_Vertex3f(0.413976192f, 0.578212261f, -1.f),
    C_Normal3f(0.9069525f, 0.421232849f, 0.f),
    C_Vertex3f(0.421690464f, 0.563477039f, 0.f),
    C_Vertex3f(0.421690464f, 0.563477039f, -1.f),
    C_Normal3f(0.941722631f, 0.336390257f, 0.f),
    C_Vertex3f(0.428128004f, 0.547742128f, 0.f),
    C_Vertex3f(0.428128004f, 0.547742128f, -1.f),
    C_Normal3f(0.96725297f, 0.253814161f, 0.f),
    C_Vertex3f(0.433288813f, 0.531007528f, 0.f),
    C_Vertex3f(0.433288813f, 0.531007528f, -1.f),
    C_Normal3f(0.982167661f, 0.18800737f, 0.f),
    C_Vertex3f(0.43717289f, 0.513273239f, 0.f),
    C_Vertex3f(0.43717289f, 0.513273239f, -1.f),
    C_Normal3f(0.990453243f, 0.13784866f, 0.f),
    C_Vertex3f(0.438476563f, 0.50390625f, 0.f),
    C_Vertex3f(0.438476563f, 0.50390625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.14892301f, -0.988848746f, 0.f),
    C_Vertex3f(0.438476563f, 0.50390625f, 0.f),
    C_Vertex3f(0.438476563f, 0.50390625f, -1.f),
    C_Vertex3f(0.357421875f, 0.491699219f, -1.f),
    C_Vertex3f(0.357421875f, 0.491699219f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.989068091f, -0.14745988f, 0.f),
    C_Vertex3f(0.357421875f, 0.491699219f, 0.f),
    C_Vertex3f(0.357421875f, 0.491699219f, -1.f),
    C_Normal3f(-0.976688147f, -0.214662939f, 0.f),
    C_Vertex3f(0.355688095f, 0.503328323f, 0.f),
    C_Vertex3f(0.355688095f, 0.503328323f, -1.f),
    C_Normal3f(-0.950497031f, -0.310733527f, 0.f),
    C_Vertex3f(0.350350857f, 0.523871422f, 0.f),
    C_Vertex3f(0.350350857f, 0.523871422f, -1.f),
    C_Normal3f(-0.896452665f, -0.443139493f, 0.f),
    C_Vertex3f(0.34314394f, 0.541699409f, 0.f),
    C_Vertex3f(0.34314394f, 0.541699409f, -1.f),
    C_Normal3f(-0.808524549f, -0.588462472f, 0.f),
    C_Vertex3f(0.334067345f, 0.556812286f, 0.f),
    C_Vertex3f(0.334067345f, 0.556812286f, -1.f),
    C_Normal3f(-0.688031852f, -0.72568047f, 0.f),
    C_Vertex3f(0.323121071f, 0.569210052f, 0.f),
    C_Vertex3f(0.323121071f, 0.569210052f, -1.f),
    C_Normal3f(-0.550557852f, -0.834797025f, 0.f),
    C_Vertex3f(0.310142994f, 0.57949543f, 0.f),
    C_Vertex3f(0.310142994f, 0.57949543f, -1.f),
    C_Normal3f(-0.405863434f, -0.913933754f, 0.f),
    C_Vertex3f(0.295133114f, 0.587668419f, 0.f),
    C_Vertex3f(0.295133114f, 0.587668419f, -1.f),
    C_Normal3f(-0.289368659f, -0.957217753f, 0.f),
    C_Vertex3f(0.278091431f, 0.593729019f, 0.f),
    C_Vertex3f(0.278091431f, 0.593729019f, -1.f),
    C_Normal3f(-0.202702746f, -0.979240298f, 0.f),
    C_Vertex3f(0.268554688f, 0.595703125f, 0.f),
    C_Vertex3f(0.268554688f, 0.595703125f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.268554688f, 0.595703125f, 0.f),
    C_Vertex3f(0.268554688f, 0.595703125f, -1.f),
    C_Vertex3f(0.268554688f, 0.369354248f, -1.f),
    C_Vertex3f(0.268554688f, 0.369354248f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.251965344f, 0.967736304f, 0.f),
    C_Vertex3f(0.268554688f, 0.369354248f, 0.f),
    C_Vertex3f(0.268554688f, 0.369354248f, -1.f),
    C_Normal3f(0.25642693f, 0.966563582f, 0.f),
    C_Vertex3f(0.275960922f, 0.367425919f, 0.f),
    C_Vertex3f(0.275960922f, 0.367425919f, -1.f),
    C_Normal3f(0.262785554f, 0.96485424f, 0.f),
    C_Vertex3f(0.289520979f, 0.363791943f, 0.f),
    C_Vertex3f(0.289520979f, 0.363791943f, -1.f),
    C_Normal3f(0.27185598f, 0.962337911f, 0.f),
    C_Vertex3f(0.301828623f, 0.36038065f, 0.f),
    C_Vertex3f(0.301828623f, 0.36038065f, -1.f),
    C_Normal3f(0.28300488f, 0.959118426f, 0.f),
    C_Vertex3f(0.312883854f, 0.357192039f, 0.f),
    C_Vertex3f(0.312883854f, 0.357192039f, -1.f),
    C_Normal3f(0.297032356f, 0.954867363f, 0.f),
    C_Vertex3f(0.322686672f, 0.354226112f, 0.f),
    C_Vertex3f(0.322686672f, 0.354226112f, -1.f),
    C_Normal3f(0.315203935f, 0.949023962f, 0.f),
    C_Vertex3f(0.331237078f, 0.351482868f, 0.f),
    C_Vertex3f(0.331237078f, 0.351482868f, -1.f),
    C_Normal3f(0.339639902f, 0.940555513f, 0.f),
    C_Vertex3f(0.33853507f, 0.348962307f, 0.f),
    C_Vertex3f(0.33853507f, 0.348962307f, -1.f),
    C_Normal3f(0.386725634f, 0.922194779f, 0.f),
    C_Vertex3f(0.34458065f, 0.346664429f, 0.f),
    C_Vertex3f(0.34458065f, 0.346664429f, -1.f),
    C_Normal3f(0.446164489f, 0.894950986f, 0.f),
    C_Vertex3f(0.356126785f, 0.341585159f, 0.f),
    C_Vertex3f(0.356126785f, 0.341585159f, -1.f),
    C_Normal3f(0.511901081f, 0.859044433f, 0.f),
    C_Vertex3f(0.372737408f, 0.332627296f, 0.f),
    C_Vertex3f(0.372737408f, 0.332627296f, -1.f),
    C_Normal3f(0.589340925f, 0.807884455f, 0.f),
    C_Vertex3f(0.38765955f, 0.322794914f, 0.f),
    C_Vertex3f(0.38765955f, 0.322794914f, -1.f),
    C_Normal3f(0.668755233f, 0.743482649f, 0.f),
    C_Vertex3f(0.400893211f, 0.312088013f, 0.f),
    C_Vertex3f(0.400893211f, 0.312088013f, -1.f),
    C_Normal3f(0.744409382f, 0.667723477f, 0.f),
    C_Vertex3f(0.412438393f, 0.300506592f, 0.f),
    C_Vertex3f(0.412438393f, 0.300506592f, -1.f),
    C_Normal3f(0.810778737f, 0.585352838f, 0.f),
    C_Vertex3f(0.422649384f, 0.287833214f, 0.f),
    C_Vertex3f(0.422649384f, 0.287833214f, -1.f),
    C_Normal3f(0.867411375f, 0.497591794f, 0.f),
    C_Vertex3f(0.431526184f, 0.274067879f, 0.f),
    C_Vertex3f(0.431526184f, 0.274067879f, -1.f),
    C_Normal3f(0.913158715f, 0.407604098f, 0.f),
    C_Vertex3f(0.439068794f, 0.259210587f, 0.f),
    C_Vertex3f(0.439068794f, 0.259210587f, -1.f),
    C_Normal3f(0.948565185f, 0.316582084f, 0.f),
    C_Vertex3f(0.445277214f, 0.243261337f, 0.f),
    C_Vertex3f(0.445277214f, 0.243261337f, -1.f),
    C_Normal3f(0.97473222f, 0.223376706f, 0.f),
    C_Vertex3f(0.449933529f, 0.226656914f, 0.f),
    C_Vertex3f(0.449933529f, 0.226656914f, -1.f),
    C_Normal3f(0.991351604f, 0.1312332f, 0.f),
    C_Vertex3f(0.453037739f, 0.209397316f, 0.f),
    C_Vertex3f(0.453037739f, 0.209397316f, -1.f),
    C_Normal3f(0.99914068f, 0.0414483435f, 0.f),
    C_Vertex3f(0.454589844f, 0.191482544f, 0.f),
    C_Vertex3f(0.454589844f, 0.191482544f, -1.f),
    C_Normal3f(0.999141812f, -0.0414194837f, 0.f),
    C_Vertex3f(0.454589844f, 0.171982765f, 0.f),
    C_Vertex3f(0.454589844f, 0.171982765f, -1.f),
    C_Normal3f(0.992051125f, -0.125835374f, 0.f),
    C_Vertex3f(0.45296967f, 0.152399898f, 0.f),
    C_Vertex3f(0.45296967f, 0.152399898f, -1.f),
    C_Normal3f(0.976401925f, -0.21596159f, 0.f),
    C_Vertex3f(0.449729323f, 0.133663774f, 0.f),
    C_Vertex3f(0.449729323f, 0.133663774f, -1.f),
    C_Normal3f(0.951127589f, -0.308798134f, 0.f),
    C_Vertex3f(0.444868803f, 0.115774393f, 0.f),
    C_Vertex3f(0.444868803f, 0.115774393f, -1.f),
    C_Normal3f(0.915755212f, -0.401736796f, 0.f),
    C_Vertex3f(0.438388109f, 0.0987317562f, 0.f),
    C_Vertex3f(0.438388109f, 0.0987317562f, -1.f),
    C_Normal3f(0.870656192f, -0.49189201f, 0.f),
    C_Vertex3f(0.430287242f, 0.0825358629f, 0.f),
    C_Vertex3f(0.430287242f, 0.0825358629f, -1.f),
    C_Normal3f(0.817089498f, -0.576510906f, 0.f),
    C_Vertex3f(0.420566201f, 0.0671867132f, 0.f),
    C_Vertex3f(0.420566201f, 0.0671867132f, -1.f),
    C_Normal3f(0.757032156f, -0.653377593f, 0.f),
    C_Vertex3f(0.409224987f, 0.0526843071f, 0.f),
    C_Vertex3f(0.409224987f, 0.0526843071f, -1.f),
    C_Normal3f(0.691500723f, -0.72237581f, 0.f),
    C_Vertex3f(0.396263599f, 0.0390286446f, 0.f),
    C_Vertex3f(0.396263599f, 0.0390286446f, -1.f),
    C_Normal3f(0.616880596f, -0.787056804f, 0.f),
    C_Vertex3f(0.382349133f, 0.0269571543f, 0.f),
    C_Vertex3f(0.382349133f, 0.0269571543f, -1.f),
    C_Normal3f(0.534159601f, -0.845383584f, 0.f),
    C_Vertex3f(0.367481589f, 0.0164698362f, 0.f),
    C_Vertex3f(0.367481589f, 0.0164698362f, -1.f),
    C_Normal3f(0.445563942f, -0.895250142f, 0.f),
    C_Vertex3f(0.351660967f, 0.00756669044f, 0.f),
    C_Vertex3f(0.351660967f, 0.00756669044f, -1.f),
    C_Normal3f(0.35388118f, -0.935290396f, 0.f),
    C_Vertex3f(0.334887266f, 0.000247716904f, 0.f),
    C_Vertex3f(0.334887266f, 0.000247716904f, -1.f),
    C_Normal3f(0.262040257f, -0.965057015f, 0.f),
    C_Vertex3f(0.317160487f, -0.00548708439f, 0.f),
    C_Vertex3f(0.317160487f, -0.00548708439f, -1.f),
    C_Normal3f(0.17268832f, -0.98497647f, 0.f),
    C_Vertex3f(0.29848063f, -0.00963771343f, 0.f),
    C_Vertex3f(0.29848063f, -0.00963771343f, -1.f),
    C_Normal3f(0.101643033f, -0.994820893f, 0.f),
    C_Vertex3f(0.278847694f, -0.0122041702f, 0.f),
    C_Vertex3f(0.278847694f, -0.0122041702f, -1.f),
    C_Normal3f(0.0476618819f, -0.998863459f, 0.f),
    C_Vertex3f(0.268554688f, -0.0126953125f, 0.f),
    C_Vertex3f(0.268554688f, -0.0126953125f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.268554688f, -0.0126953125f, 0.f),
    C_Vertex3f(0.268554688f, -0.0126953125f, -1.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, -1.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, 0.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, -1.f),
    C_Vertex3f(0.222167969f, -0.0922851563f, -1.f),
    C_Vertex3f(0.222167969f, -0.0922851563f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.151808649f, -0.988409877f, 0.f),
    C_Vertex3f(0.222167969f, 0.596679688f, 0.f),
    C_Vertex3f(0.222167969f, 0.596679688f, -1.f),
    C_Normal3f(0.246894479f, -0.969042301f, 0.f),
    C_Vertex3f(0.210767746f, 0.594928741f, 0.f),
    C_Vertex3f(0.210767746f, 0.594928741f, -1.f),
    C_Normal3f(0.375525355f, -0.926812112f, 0.f),
    C_Vertex3f(0.190383434f, 0.588581562f, 0.f),
    C_Vertex3f(0.190383434f, 0.588581562f, -1.f),
    C_Normal3f(0.535065472f, -0.844810486f, 0.f),
    C_Vertex3f(0.172415257f, 0.579389095f, 0.f),
    C_Vertex3f(0.172415257f, 0.579389095f, -1.f),
    C_Normal3f(0.684290588f, -0.729209363f, 0.f),
    C_Vertex3f(0.156863213f, 0.567351341f, 0.f),
    C_Vertex3f(0.156863213f, 0.567351341f, -1.f),
    C_Normal3f(0.804467916f, -0.593996048f, 0.f),
    C_Vertex3f(0.143727303f, 0.5524683f, 0.f),
    C_Vertex3f(0.143727303f, 0.5524683f, -1.f),
    C_Normal3f(0.899820924f, -0.436259538f, 0.f),
    C_Vertex3f(0.13387537f, 0.536218166f, 0.f),
    C_Vertex3f(0.13387537f, 0.536218166f, -1.f),
    C_Normal3f(0.965631187f, -0.259916186f, 0.f),
    C_Vertex3f(0.127307415f, 0.518600941f, 0.f),
    C_Vertex3f(0.127307415f, 0.518600941f, -1.f),
    C_Normal3f(0.99651432f, -0.0834213868f, 0.f),
    C_Vertex3f(0.124023438f, 0.499616623f, 0.f),
    C_Vertex3f(0.124023438f, 0.499616623f, -1.f),
    C_Normal3f(0.997366369f, 0.0725276992f, 0.f),
    C_Vertex3f(0.124023438f, 0.479372025f, 0.f),
    C_Vertex3f(0.124023438f, 0.479372025f, -1.f),
    C_Normal3f(0.97258848f, 0.232533291f, 0.f),
    C_Vertex3f(0.126840591f, 0.460876465f, 0.f),
    C_Vertex3f(0.126840591f, 0.460876465f, -1.f),
    C_Normal3f(0.915552616f, 0.402198344f, 0.f),
    C_Vertex3f(0.132474899f, 0.444023132f, 0.f),
    C_Vertex3f(0.132474899f, 0.444023132f, -1.f),
    C_Normal3f(0.832386851f, 0.554194987f, 0.f),
    C_Vertex3f(0.140926361f, 0.428812027f, 0.f),
    C_Vertex3f(0.140926361f, 0.428812027f, -1.f),
    C_Normal3f(0.739837706f, 0.672785282f, 0.f),
    C_Vertex3f(0.149377823f, 0.418635368f, 0.f),
    C_Vertex3f(0.149377823f, 0.418635368f, -1.f),
    C_Normal3f(0.656612396f, 0.754228115f, 0.f),
    C_Vertex3f(0.155966282f, 0.412273169f, 0.f),
    C_Vertex3f(0.155966282f, 0.412273169f, -1.f),
    C_Normal3f(0.58127135f, 0.813709736f, 0.f),
    C_Vertex3f(0.163508892f, 0.406333208f, 0.f),
    C_Vertex3f(0.163508892f, 0.406333208f, -1.f),
    C_Normal3f(0.509005129f, 0.86076349f, 0.f),
    C_Vertex3f(0.172005653f, 0.400815487f, 0.f),
    C_Vertex3f(0.172005653f, 0.400815487f, -1.f),
    C_Normal3f(0.441446662f, 0.897287548f, 0.f),
    C_Vertex3f(0.181456566f, 0.395720005f, 0.f),
    C_Vertex3f(0.181456566f, 0.395720005f, -1.f),
    C_Normal3f(0.37938574f, 0.925238609f, 0.f),
    C_Vertex3f(0.191861629f, 0.391046762f, 0.f),
    C_Vertex3f(0.191861629f, 0.391046762f, -1.f),
    C_Normal3f(0.323016644f, 0.946393311f, 0.f),
    C_Vertex3f(0.203220844f, 0.386795759f, 0.f),
    C_Vertex3f(0.203220844f, 0.386795759f, -1.f),
    C_Normal3f(0.280269921f, 0.959921181f, 0.f),
    C_Vertex3f(0.21553421f, 0.382966995f, 0.f),
    C_Vertex3f(0.21553421f, 0.382966995f, -1.f),
    C_Normal3f(0.248690262f, 0.968583107f, 0.f),
    C_Vertex3f(0.222167969f, 0.381263733f, 0.f),
    C_Vertex3f(0.222167969f, 0.381263733f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.222167969f, 0.381263733f, 0.f),
    C_Vertex3f(0.222167969f, 0.381263733f, -1.f),
    C_Vertex3f(0.222167969f, 0.596679688f, -1.f),
    C_Vertex3f(0.222167969f, 0.596679688f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.123830169f, 0.992303491f, 0.f),
    C_Vertex3f(0.268554688f, 0.052734375f, 0.f),
    C_Vertex3f(0.268554688f, 0.052734375f, -1.f),
    C_Normal3f(-0.179115683f, 0.983828068f, 0.f),
    C_Vertex3f(0.274248123f, 0.0534448624f, 0.f),
    C_Vertex3f(0.274248123f, 0.0534448624f, -1.f),
    C_Normal3f(-0.250397712f, 0.968143046f, 0.f),
    C_Vertex3f(0.285140276f, 0.0557539463f, 0.f),
    C_Vertex3f(0.285140276f, 0.0557539463f, -1.f),
    C_Normal3f(-0.337670147f, 0.94126451f, 0.f),
    C_Vertex3f(0.29553771f, 0.0589511395f, 0.f),
    C_Vertex3f(0.29553771f, 0.0589511395f, -1.f),
    C_Normal3f(-0.424695879f, 0.905336022f, 0.f),
    C_Vertex3f(0.305440426f, 0.0630364418f, 0.f),
    C_Vertex3f(0.305440426f, 0.0630364418f, -1.f),
    C_Normal3f(-0.509033024f, 0.86074692f, 0.f),
    C_Vertex3f(0.314848423f, 0.0680098534f, 0.f),
    C_Vertex3f(0.314848423f, 0.0680098534f, -1.f),
    C_Normal3f(-0.58836019f, 0.808598995f, 0.f),
    C_Vertex3f(0.323761702f, 0.0738713741f, 0.f),
    C_Vertex3f(0.323761702f, 0.0738713741f, -1.f),
    C_Normal3f(-0.660784662f, 0.750575602f, 0.f),
    C_Vertex3f(0.332180262f, 0.0806210041f, 0.f),
    C_Vertex3f(0.332180262f, 0.0806210041f, -1.f),
    C_Normal3f(-0.725054145f, 0.688691795f, 0.f),
    C_Vertex3f(0.340104103f, 0.0882587433f, 0.f),
    C_Vertex3f(0.340104103f, 0.0882587433f, -1.f),
    C_Normal3f(-0.783389568f, 0.62153089f, 0.f),
    C_Vertex3f(0.347533226f, 0.0967845917f, 0.f),
    C_Vertex3f(0.347533226f, 0.0967845917f, -1.f),
    C_Normal3f(-0.838379622f, 0.545086682f, 0.f),
    C_Vertex3f(0.354033709f, 0.105815887f, 0.f),
    C_Vertex3f(0.354033709f, 0.105815887f, -1.f),
    C_Normal3f(-0.886585832f, 0.46256417f, 0.f),
    C_Vertex3f(0.359605551f, 0.115352631f, 0.f),
    C_Vertex3f(0.359605551f, 0.115352631f, -1.f),
    C_Normal3f(-0.926574111f, 0.376112252f, 0.f),
    C_Vertex3f(0.364248753f, 0.125394821f, 0.f),
    C_Vertex3f(0.364248753f, 0.125394821f, -1.f),
    C_Normal3f(-0.957578659f, 0.288171977f, 0.f),
    C_Vertex3f(0.367963314f, 0.135942459f, 0.f),
    C_Vertex3f(0.367963314f, 0.135942459f, -1.f),
    C_Normal3f(-0.979560733f, 0.201148733f, 0.f),
    C_Vertex3f(0.370749235f, 0.146995544f, 0.f),
    C_Vertex3f(0.370749235f, 0.146995544f, -1.f),
    C_Normal3f(-0.993117392f, 0.117123306f, 0.f),
    C_Vertex3f(0.372606516f, 0.158554077f, 0.f),
    C_Vertex3f(0.372606516f, 0.158554077f, -1.f),
    C_Normal3f(-0.999489784f, 0.0319392607f, 0.f),
    C_Vertex3f(0.373535156f, 0.170618057f, 0.f),
    C_Vertex3f(0.373535156f, 0.170618057f, -1.f),
    C_Normal3f(-0.997525215f, -0.0703087673f, 0.f),
    C_Vertex3f(0.373535156f, 0.187614441f, 0.f),
    C_Vertex3f(0.373535156f, 0.187614441f, -1.f),
    C_Normal3f(-0.978618085f, -0.205686048f, 0.f),
    C_Vertex3f(0.370975494f, 0.206933975f, 0.f),
    C_Vertex3f(0.370975494f, 0.206933975f, -1.f),
    C_Normal3f(-0.929787338f, -0.368096828f, 0.f),
    C_Vertex3f(0.365856171f, 0.224149704f, 0.f),
    C_Vertex3f(0.365856171f, 0.224149704f, -1.f),
    C_Normal3f(-0.8508237f, -0.525451183f, 0.f),
    C_Vertex3f(0.358177185f, 0.239261627f, 0.f),
    C_Vertex3f(0.358177185f, 0.239261627f, -1.f),
    C_Normal3f(-0.748062432f, -0.66362828f, 0.f),
    C_Vertex3f(0.350498199f, 0.249017715f, 0.f),
    C_Vertex3f(0.350498199f, 0.249017715f, -1.f),
    C_Normal3f(-0.645684659f, -0.763604224f, 0.f),
    C_Vertex3f(0.344017982f, 0.255222321f, 0.f),
    C_Vertex3f(0.344017982f, 0.255222321f, -1.f),
    C_Normal3f(-0.559698224f, -0.828696489f, 0.f),
    C_Vertex3f(0.336176872f, 0.261127472f, 0.f),
    C_Vertex3f(0.336176872f, 0.261127472f, -1.f),
    C_Normal3f(-0.483320773f, -0.875443399f, 0.f),
    C_Vertex3f(0.326974869f, 0.26673317f, 0.f),
    C_Vertex3f(0.326974869f, 0.26673317f, -1.f),
    C_Normal3f(-0.416876376f, -0.908963203f, 0.f),
    C_Vertex3f(0.316411972f, 0.272039413f, 0.f),
    C_Vertex3f(0.316411972f, 0.272039413f, -1.f),
    C_Normal3f(-0.359577477f, -0.933115184f, 0.f),
    C_Vertex3f(0.304488182f, 0.277046204f, 0.f),
    C_Vertex3f(0.304488182f, 0.277046204f, -1.f),
    C_Normal3f(-0.31025213f, -0.950654328f, 0.f),
    C_Vertex3f(0.291203499f, 0.28175354f, 0.f),
    C_Vertex3f(0.291203499f, 0.28175354f, -1.f),
    C_Normal3f(-0.274368137f, -0.961624682f, 0.f),
    C_Vertex3f(0.276557922f, 0.286161423f, 0.f),
    C_Vertex3f(0.276557922f, 0.286161423f, -1.f),
    C_Normal3f(-0.248614177f, -0.968602657f, 0.f),
    C_Vertex3f(0.268554688f, 0.288215637f, 0.f),
    C_Vertex3f(0.268554688f, 0.288215637f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.268554688f, 0.288215637f, 0.f),
    C_Vertex3f(0.268554688f, 0.288215637f, -1.f),
    C_Vertex3f(0.268554688f, 0.052734375f, -1.f),
    C_Vertex3f(0.268554688f, 0.052734375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.205289364f, 0.0556931496f, 0.f),
    C_Vertex3f(0.207460403f, -0.0118141174f, 0.f),
    C_Vertex3f(0.216405869f, 0.0533342361f, 0.f),
    C_Vertex3f(0.222167969f, -0.013671875f, 0.f),
    C_Vertex3f(0.222167969f, 0.052734375f, 0.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, 0.f),
    C_Vertex3f(0.222167969f, 0.300125122f, 0.f),
    C_Vertex3f(0.222167969f, 0.381263733f, 0.f),
    C_Vertex3f(0.21553421f, 0.382966995f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.205289364f, 0.0556931496f, 0.f),
    C_Vertex3f(0.194580555f, 0.0592112541f, 0.f),
    C_Vertex3f(0.207460403f, -0.0118141174f, 0.f),
    C_Vertex3f(0.184279442f, 0.0638885498f, 0.f),
    C_Vertex3f(0.180796623f, -0.00665140152f, 0.f),
    C_Vertex3f(0.174386024f, 0.0697250366f, 0.f),
    C_Vertex3f(0.164900303f, 0.0767207146f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156884193f, -4.14848328e-005f, 0.f),
    C_Vertex3f(0.180796623f, -0.00665140152f, 0.f),
    C_Vertex3f(0.164900303f, 0.0767207146f, 0.f),
    C_Vertex3f(0.155822277f, 0.0848755836f, 0.f),
    C_Vertex3f(0.147151947f, 0.0941896439f, 0.f),
    C_Vertex3f(0.14013958f, 0.103136063f, 0.f),
    C_Vertex3f(0.135723114f, 0.00801563263f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135723114f, 0.00801563263f, 0.f),
    C_Vertex3f(0.14013958f, 0.103136063f, 0.f),
    C_Vertex3f(0.134812355f, 0.111605644f, 0.f),
    C_Vertex3f(0.129920006f, 0.121125221f, 0.f),
    C_Vertex3f(0.125462532f, 0.131694794f, 0.f),
    C_Vertex3f(0.121439934f, 0.143314362f, 0.f),
    C_Vertex3f(0.117852211f, 0.155983925f, 0.f),
    C_Vertex3f(0.117313385f, 0.0175199509f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.117313385f, 0.0175199509f, 0.f),
    C_Vertex3f(0.117852211f, 0.155983925f, 0.f),
    C_Vertex3f(0.114699364f, 0.169703484f, 0.f),
    C_Vertex3f(0.111981392f, 0.184473038f, 0.f),
    C_Vertex3f(0.110839844f, 0.192382813f, 0.f),
    C_Vertex3f(0.10015583f, 0.029946804f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.110839844f, 0.192382813f, 0.f),
    C_Vertex3f(0.0317382813f, 0.177734375f, 0.f),
    C_Vertex3f(0.0328273773f, 0.162439346f, 0.f),
    C_Vertex3f(0.0378112793f, 0.134198189f, 0.f),
    C_Vertex3f(0.0456008911f, 0.108305931f, 0.f),
    C_Vertex3f(0.0561962128f, 0.0847625732f, 0.f),
    C_Vertex3f(0.0695972443f, 0.0635681152f, 0.f),
    C_Vertex3f(0.0842504501f, 0.0452961922f, 0.f),
    C_Vertex3f(0.10015583f, 0.029946804f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.21021843f, 0.303192139f, 0.f),
    C_Vertex3f(0.222167969f, 0.300125122f, 0.f),
    C_Vertex3f(0.21553421f, 0.382966995f, 0.f),
    C_Vertex3f(0.203220844f, 0.386795759f, 0.f),
    C_Vertex3f(0.191861629f, 0.391046762f, 0.f),
    C_Vertex3f(0.185955048f, 0.311344147f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.185955048f, 0.311344147f, 0.f),
    C_Vertex3f(0.191861629f, 0.391046762f, 0.f),
    C_Vertex3f(0.181456566f, 0.395720005f, 0.f),
    C_Vertex3f(0.172005653f, 0.400815487f, 0.f),
    C_Vertex3f(0.163508892f, 0.406333208f, 0.f),
    C_Vertex3f(0.161327362f, 0.32151413f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161327362f, 0.32151413f, 0.f),
    C_Vertex3f(0.163508892f, 0.406333208f, 0.f),
    C_Vertex3f(0.155966282f, 0.412273169f, 0.f),
    C_Vertex3f(0.149377823f, 0.418635368f, 0.f),
    C_Vertex3f(0.140926361f, 0.428812027f, 0.f),
    C_Vertex3f(0.136335373f, 0.333702087f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136335373f, 0.333702087f, 0.f),
    C_Vertex3f(0.140926361f, 0.428812027f, 0.f),
    C_Vertex3f(0.132474899f, 0.444023132f, 0.f),
    C_Vertex3f(0.126840591f, 0.460876465f, 0.f),
    C_Vertex3f(0.124023438f, 0.479372025f, 0.f),
    C_Vertex3f(0.11425972f, 0.34605217f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11425972f, 0.34605217f, 0.f),
    C_Vertex3f(0.124023438f, 0.479372025f, 0.f),
    C_Vertex3f(0.120676279f, 0.631538153f, 0.f),
    C_Vertex3f(0.124023438f, 0.499616623f, 0.f),
    C_Vertex3f(0.13297534f, 0.638587236f, 0.f),
    C_Vertex3f(0.127307415f, 0.518600941f, 0.f),
    C_Vertex3f(0.13387537f, 0.536218166f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.109574318f, 0.623641968f, 0.f),
    C_Vertex3f(0.0972204208f, 0.61276722f, 0.f),
    C_Vertex3f(0.0976223946f, 0.358567238f, 0.f),
    C_Vertex3f(0.11425972f, 0.34605217f, 0.f),
    C_Vertex3f(0.120676279f, 0.631538153f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0472605228f, 0.513791561f, 0.f),
    C_Vertex3f(0.0479512215f, 0.449163437f, 0.f),
    C_Vertex3f(0.050961256f, 0.53278017f, 0.f),
    C_Vertex3f(0.0530333519f, 0.428071976f, 0.f),
    C_Vertex3f(0.0565123558f, 0.550771713f, 0.f),
    C_Vertex3f(0.0606565475f, 0.408237457f, 0.f),
    C_Vertex3f(0.0639138222f, 0.56776619f, 0.f),
    C_Vertex3f(0.0708208084f, 0.389659882f, 0.f),
    C_Vertex3f(0.0731656551f, 0.583763599f, 0.f),
    C_Vertex3f(0.0831427574f, 0.373103142f, 0.f),
    C_Vertex3f(0.0842678547f, 0.598763943f, 0.f),
    C_Vertex3f(0.0976223946f, 0.358567238f, 0.f),
    C_Vertex3f(0.0972204208f, 0.61276722f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0454101563f, 0.493805885f, 0.f),
    C_Vertex3f(0.0454101563f, 0.471511841f, 0.f),
    C_Vertex3f(0.0479512215f, 0.449163437f, 0.f),
    C_Vertex3f(0.0472605228f, 0.513791561f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268554688f, 0.699707031f, 0.f),
    C_Vertex3f(0.222167969f, 0.699707031f, 0.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, 0.f),
    C_Vertex3f(0.222167969f, 0.662109375f, 0.f),
    C_Vertex3f(0.222167969f, 0.596679688f, 0.f),
    C_Vertex3f(0.212427139f, 0.661126137f, 0.f),
    C_Vertex3f(0.210767746f, 0.594928741f, 0.f),
    C_Vertex3f(0.19414258f, 0.658312559f, 0.f),
    C_Vertex3f(0.190383434f, 0.588581562f, 0.f),
    C_Vertex3f(0.17705512f, 0.65465188f, 0.f),
    C_Vertex3f(0.172415257f, 0.579389095f, 0.f),
    C_Vertex3f(0.161164761f, 0.6501441f, 0.f),
    C_Vertex3f(0.156863213f, 0.567351341f, 0.f),
    C_Vertex3f(0.1464715f, 0.644789219f, 0.f),
    C_Vertex3f(0.143727303f, 0.5524683f, 0.f),
    C_Vertex3f(0.13387537f, 0.536218166f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.293257952f, 0.658794165f, 0.f),
    C_Vertex3f(0.277083397f, 0.661287308f, 0.f),
    C_Vertex3f(0.278091431f, 0.593729019f, 0.f),
    C_Vertex3f(0.268554688f, 0.662109375f, 0.f),
    C_Vertex3f(0.268554688f, 0.595703125f, 0.f),
    C_Vertex3f(0.268554688f, 0.699707031f, 0.f),
    C_Vertex3f(0.268554688f, 0.369354248f, 0.f),
    C_Vertex3f(0.268554688f, 0.288215637f, 0.f),
    C_Vertex3f(0.276557922f, 0.286161423f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.268554688f, 0.052734375f, 0.f),
    C_Vertex3f(0.268554688f, 0.288215637f, 0.f),
    C_Vertex3f(0.268554688f, 0.699707031f, 0.f),
    C_Vertex3f(0.268554688f, -0.0126953125f, 0.f),
    C_Vertex3f(0.278847694f, -0.0122041702f, 0.f),
    C_Vertex3f(0.274248123f, 0.0534448624f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.438476563f, 0.50390625f, 0.f),
    C_Vertex3f(0.43717289f, 0.513273239f, 0.f),
    C_Vertex3f(0.433288813f, 0.531007528f, 0.f),
    C_Vertex3f(0.428128004f, 0.547742128f, 0.f),
    C_Vertex3f(0.421690464f, 0.563477039f, 0.f),
    C_Vertex3f(0.413976192f, 0.578212261f, 0.f),
    C_Vertex3f(0.404985189f, 0.591947794f, 0.f),
    C_Vertex3f(0.394717455f, 0.604683638f, 0.f),
    C_Vertex3f(0.383172989f, 0.616419792f, 0.f),
    C_Vertex3f(0.371765137f, 0.626006126f, 0.f),
    C_Vertex3f(0.360887766f, 0.633593321f, 0.f),
    C_Vertex3f(0.357421875f, 0.491699219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323121071f, 0.569210052f, 0.f),
    C_Vertex3f(0.334067345f, 0.556812286f, 0.f),
    C_Vertex3f(0.336484432f, 0.646220684f, 0.f),
    C_Vertex3f(0.34314394f, 0.541699409f, 0.f),
    C_Vertex3f(0.349127531f, 0.640331507f, 0.f),
    C_Vertex3f(0.350350857f, 0.523871422f, 0.f),
    C_Vertex3f(0.360887766f, 0.633593321f, 0.f),
    C_Vertex3f(0.355688095f, 0.503328323f, 0.f),
    C_Vertex3f(0.357421875f, 0.491699219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.293257952f, 0.658794165f, 0.f),
    C_Vertex3f(0.295133114f, 0.587668419f, 0.f),
    C_Vertex3f(0.308549643f, 0.655452013f, 0.f),
    C_Vertex3f(0.310142994f, 0.57949543f, 0.f),
    C_Vertex3f(0.322958469f, 0.651260853f, 0.f),
    C_Vertex3f(0.323121071f, 0.569210052f, 0.f),
    C_Vertex3f(0.336484432f, 0.646220684f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.373535156f, 0.187614441f, 0.f),
    C_Vertex3f(0.382349133f, 0.0269571543f, 0.f),
    C_Vertex3f(0.38765955f, 0.322794914f, 0.f),
    C_Vertex3f(0.396263599f, 0.0390286446f, 0.f),
    C_Vertex3f(0.400893211f, 0.312088013f, 0.f),
    C_Vertex3f(0.409224987f, 0.0526843071f, 0.f),
    C_Vertex3f(0.412438393f, 0.300506592f, 0.f),
    C_Vertex3f(0.420566201f, 0.0671867132f, 0.f),
    C_Vertex3f(0.422649384f, 0.287833214f, 0.f),
    C_Vertex3f(0.430287242f, 0.0825358629f, 0.f),
    C_Vertex3f(0.431526184f, 0.274067879f, 0.f),
    C_Vertex3f(0.438388109f, 0.0987317562f, 0.f),
    C_Vertex3f(0.439068794f, 0.259210587f, 0.f),
    C_Vertex3f(0.444868803f, 0.115774393f, 0.f),
    C_Vertex3f(0.445277214f, 0.243261337f, 0.f),
    C_Vertex3f(0.449729323f, 0.133663774f, 0.f),
    C_Vertex3f(0.449933529f, 0.226656914f, 0.f),
    C_Vertex3f(0.45296967f, 0.152399898f, 0.f),
    C_Vertex3f(0.453037739f, 0.209397316f, 0.f),
    C_Vertex3f(0.454589844f, 0.171982765f, 0.f),
    C_Vertex3f(0.454589844f, 0.191482544f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.372737408f, 0.332627296f, 0.f),
    C_Vertex3f(0.356126785f, 0.341585159f, 0.f),
    C_Vertex3f(0.358177185f, 0.239261627f, 0.f),
    C_Vertex3f(0.365856171f, 0.224149704f, 0.f),
    C_Vertex3f(0.370975494f, 0.206933975f, 0.f),
    C_Vertex3f(0.373535156f, 0.187614441f, 0.f),
    C_Vertex3f(0.38765955f, 0.322794914f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.358177185f, 0.239261627f, 0.f),
    C_Vertex3f(0.356126785f, 0.341585159f, 0.f),
    C_Vertex3f(0.350498199f, 0.249017715f, 0.f),
    C_Vertex3f(0.34458065f, 0.346664429f, 0.f),
    C_Vertex3f(0.344017982f, 0.255222321f, 0.f),
    C_Vertex3f(0.33853507f, 0.348962307f, 0.f),
    C_Vertex3f(0.336176872f, 0.261127472f, 0.f),
    C_Vertex3f(0.331237078f, 0.351482868f, 0.f),
    C_Vertex3f(0.326974869f, 0.26673317f, 0.f),
    C_Vertex3f(0.322686672f, 0.354226112f, 0.f),
    C_Vertex3f(0.316411972f, 0.272039413f, 0.f),
    C_Vertex3f(0.312883854f, 0.357192039f, 0.f),
    C_Vertex3f(0.304488182f, 0.277046204f, 0.f),
    C_Vertex3f(0.301828623f, 0.36038065f, 0.f),
    C_Vertex3f(0.291203499f, 0.28175354f, 0.f),
    C_Vertex3f(0.289520979f, 0.363791943f, 0.f),
    C_Vertex3f(0.276557922f, 0.286161423f, 0.f),
    C_Vertex3f(0.275960922f, 0.367425919f, 0.f),
    C_Vertex3f(0.268554688f, 0.369354248f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.382349133f, 0.0269571543f, 0.f),
    C_Vertex3f(0.373535156f, 0.187614441f, 0.f),
    C_Vertex3f(0.373535156f, 0.170618057f, 0.f),
    C_Vertex3f(0.372606516f, 0.158554077f, 0.f),
    C_Vertex3f(0.370749235f, 0.146995544f, 0.f),
    C_Vertex3f(0.367963314f, 0.135942459f, 0.f),
    C_Vertex3f(0.367481589f, 0.0164698362f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.367481589f, 0.0164698362f, 0.f),
    C_Vertex3f(0.367963314f, 0.135942459f, 0.f),
    C_Vertex3f(0.364248753f, 0.125394821f, 0.f),
    C_Vertex3f(0.359605551f, 0.115352631f, 0.f),
    C_Vertex3f(0.354033709f, 0.105815887f, 0.f),
    C_Vertex3f(0.351660967f, 0.00756669044f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.354033709f, 0.105815887f, 0.f),
    C_Vertex3f(0.347533226f, 0.0967845917f, 0.f),
    C_Vertex3f(0.351660967f, 0.00756669044f, 0.f),
    C_Vertex3f(0.340104103f, 0.0882587433f, 0.f),
    C_Vertex3f(0.334887266f, 0.000247716904f, 0.f),
    C_Vertex3f(0.332180262f, 0.0806210041f, 0.f),
    C_Vertex3f(0.323761702f, 0.0738713741f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.317160487f, -0.00548708439f, 0.f),
    C_Vertex3f(0.334887266f, 0.000247716904f, 0.f),
    C_Vertex3f(0.323761702f, 0.0738713741f, 0.f),
    C_Vertex3f(0.314848423f, 0.0680098534f, 0.f),
    C_Vertex3f(0.305440426f, 0.0630364418f, 0.f),
    C_Vertex3f(0.29848063f, -0.00963771343f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.274248123f, 0.0534448624f, 0.f),
    C_Vertex3f(0.278847694f, -0.0122041702f, 0.f),
    C_Vertex3f(0.285140276f, 0.0557539463f, 0.f),
    C_Vertex3f(0.29848063f, -0.00963771343f, 0.f),
    C_Vertex3f(0.29553771f, 0.0589511395f, 0.f),
    C_Vertex3f(0.305440426f, 0.0630364418f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.293257952f, 0.658794165f, 0.f),
    C_Vertex3f(0.278091431f, 0.593729019f, 0.f),
    C_Vertex3f(0.295133114f, 0.587668419f, 0.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, 0.f),
    C_Vertex3f(0.222167969f, -0.013671875f, 0.f),
    C_Vertex3f(0.222167969f, -0.0922851563f, 0.f),
    C_Vertex3f(0.222167969f, 0.381263733f, 0.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, 0.f),
    C_Vertex3f(0.222167969f, 0.596679688f, 0.f),
    C_Vertex3f(0.268554688f, 0.699707031f, 0.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, 0.f),
    C_Vertex3f(0.268554688f, -0.0126953125f, 0.f),
    C_Vertex3f(0.13297534f, 0.638587236f, 0.f),
    C_Vertex3f(0.13387537f, 0.536218166f, 0.f),
    C_Vertex3f(0.1464715f, 0.644789219f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.205289364f, 0.0556931496f, -1.f),
    C_Vertex3f(0.207460403f, -0.0118141174f, -1.f),
    C_Vertex3f(0.216405869f, 0.0533342361f, -1.f),
    C_Vertex3f(0.222167969f, -0.013671875f, -1.f),
    C_Vertex3f(0.222167969f, 0.052734375f, -1.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, -1.f),
    C_Vertex3f(0.222167969f, 0.300125122f, -1.f),
    C_Vertex3f(0.222167969f, 0.381263733f, -1.f),
    C_Vertex3f(0.21553421f, 0.382966995f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.205289364f, 0.0556931496f, -1.f),
    C_Vertex3f(0.194580555f, 0.0592112541f, -1.f),
    C_Vertex3f(0.207460403f, -0.0118141174f, -1.f),
    C_Vertex3f(0.184279442f, 0.0638885498f, -1.f),
    C_Vertex3f(0.180796623f, -0.00665140152f, -1.f),
    C_Vertex3f(0.174386024f, 0.0697250366f, -1.f),
    C_Vertex3f(0.164900303f, 0.0767207146f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156884193f, -4.14848328e-005f, -1.f),
    C_Vertex3f(0.180796623f, -0.00665140152f, -1.f),
    C_Vertex3f(0.164900303f, 0.0767207146f, -1.f),
    C_Vertex3f(0.155822277f, 0.0848755836f, -1.f),
    C_Vertex3f(0.147151947f, 0.0941896439f, -1.f),
    C_Vertex3f(0.14013958f, 0.103136063f, -1.f),
    C_Vertex3f(0.135723114f, 0.00801563263f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135723114f, 0.00801563263f, -1.f),
    C_Vertex3f(0.14013958f, 0.103136063f, -1.f),
    C_Vertex3f(0.134812355f, 0.111605644f, -1.f),
    C_Vertex3f(0.129920006f, 0.121125221f, -1.f),
    C_Vertex3f(0.125462532f, 0.131694794f, -1.f),
    C_Vertex3f(0.121439934f, 0.143314362f, -1.f),
    C_Vertex3f(0.117852211f, 0.155983925f, -1.f),
    C_Vertex3f(0.117313385f, 0.0175199509f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.117313385f, 0.0175199509f, -1.f),
    C_Vertex3f(0.117852211f, 0.155983925f, -1.f),
    C_Vertex3f(0.114699364f, 0.169703484f, -1.f),
    C_Vertex3f(0.111981392f, 0.184473038f, -1.f),
    C_Vertex3f(0.110839844f, 0.192382813f, -1.f),
    C_Vertex3f(0.10015583f, 0.029946804f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.110839844f, 0.192382813f, -1.f),
    C_Vertex3f(0.0317382813f, 0.177734375f, -1.f),
    C_Vertex3f(0.0328273773f, 0.162439346f, -1.f),
    C_Vertex3f(0.0378112793f, 0.134198189f, -1.f),
    C_Vertex3f(0.0456008911f, 0.108305931f, -1.f),
    C_Vertex3f(0.0561962128f, 0.0847625732f, -1.f),
    C_Vertex3f(0.0695972443f, 0.0635681152f, -1.f),
    C_Vertex3f(0.0842504501f, 0.0452961922f, -1.f),
    C_Vertex3f(0.10015583f, 0.029946804f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.21021843f, 0.303192139f, -1.f),
    C_Vertex3f(0.222167969f, 0.300125122f, -1.f),
    C_Vertex3f(0.21553421f, 0.382966995f, -1.f),
    C_Vertex3f(0.203220844f, 0.386795759f, -1.f),
    C_Vertex3f(0.191861629f, 0.391046762f, -1.f),
    C_Vertex3f(0.185955048f, 0.311344147f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.185955048f, 0.311344147f, -1.f),
    C_Vertex3f(0.191861629f, 0.391046762f, -1.f),
    C_Vertex3f(0.181456566f, 0.395720005f, -1.f),
    C_Vertex3f(0.172005653f, 0.400815487f, -1.f),
    C_Vertex3f(0.163508892f, 0.406333208f, -1.f),
    C_Vertex3f(0.161327362f, 0.32151413f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161327362f, 0.32151413f, -1.f),
    C_Vertex3f(0.163508892f, 0.406333208f, -1.f),
    C_Vertex3f(0.155966282f, 0.412273169f, -1.f),
    C_Vertex3f(0.149377823f, 0.418635368f, -1.f),
    C_Vertex3f(0.140926361f, 0.428812027f, -1.f),
    C_Vertex3f(0.136335373f, 0.333702087f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136335373f, 0.333702087f, -1.f),
    C_Vertex3f(0.140926361f, 0.428812027f, -1.f),
    C_Vertex3f(0.132474899f, 0.444023132f, -1.f),
    C_Vertex3f(0.126840591f, 0.460876465f, -1.f),
    C_Vertex3f(0.124023438f, 0.479372025f, -1.f),
    C_Vertex3f(0.11425972f, 0.34605217f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11425972f, 0.34605217f, -1.f),
    C_Vertex3f(0.124023438f, 0.479372025f, -1.f),
    C_Vertex3f(0.120676279f, 0.631538153f, -1.f),
    C_Vertex3f(0.124023438f, 0.499616623f, -1.f),
    C_Vertex3f(0.13297534f, 0.638587236f, -1.f),
    C_Vertex3f(0.127307415f, 0.518600941f, -1.f),
    C_Vertex3f(0.13387537f, 0.536218166f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.109574318f, 0.623641968f, -1.f),
    C_Vertex3f(0.0972204208f, 0.61276722f, -1.f),
    C_Vertex3f(0.0976223946f, 0.358567238f, -1.f),
    C_Vertex3f(0.11425972f, 0.34605217f, -1.f),
    C_Vertex3f(0.120676279f, 0.631538153f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0472605228f, 0.513791561f, -1.f),
    C_Vertex3f(0.0479512215f, 0.449163437f, -1.f),
    C_Vertex3f(0.050961256f, 0.53278017f, -1.f),
    C_Vertex3f(0.0530333519f, 0.428071976f, -1.f),
    C_Vertex3f(0.0565123558f, 0.550771713f, -1.f),
    C_Vertex3f(0.0606565475f, 0.408237457f, -1.f),
    C_Vertex3f(0.0639138222f, 0.56776619f, -1.f),
    C_Vertex3f(0.0708208084f, 0.389659882f, -1.f),
    C_Vertex3f(0.0731656551f, 0.583763599f, -1.f),
    C_Vertex3f(0.0831427574f, 0.373103142f, -1.f),
    C_Vertex3f(0.0842678547f, 0.598763943f, -1.f),
    C_Vertex3f(0.0976223946f, 0.358567238f, -1.f),
    C_Vertex3f(0.0972204208f, 0.61276722f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0454101563f, 0.493805885f, -1.f),
    C_Vertex3f(0.0454101563f, 0.471511841f, -1.f),
    C_Vertex3f(0.0479512215f, 0.449163437f, -1.f),
    C_Vertex3f(0.0472605228f, 0.513791561f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268554688f, 0.699707031f, -1.f),
    C_Vertex3f(0.222167969f, 0.699707031f, -1.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, -1.f),
    C_Vertex3f(0.222167969f, 0.662109375f, -1.f),
    C_Vertex3f(0.222167969f, 0.596679688f, -1.f),
    C_Vertex3f(0.212427139f, 0.661126137f, -1.f),
    C_Vertex3f(0.210767746f, 0.594928741f, -1.f),
    C_Vertex3f(0.19414258f, 0.658312559f, -1.f),
    C_Vertex3f(0.190383434f, 0.588581562f, -1.f),
    C_Vertex3f(0.17705512f, 0.65465188f, -1.f),
    C_Vertex3f(0.172415257f, 0.579389095f, -1.f),
    C_Vertex3f(0.161164761f, 0.6501441f, -1.f),
    C_Vertex3f(0.156863213f, 0.567351341f, -1.f),
    C_Vertex3f(0.1464715f, 0.644789219f, -1.f),
    C_Vertex3f(0.143727303f, 0.5524683f, -1.f),
    C_Vertex3f(0.13387537f, 0.536218166f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.293257952f, 0.658794165f, -1.f),
    C_Vertex3f(0.277083397f, 0.661287308f, -1.f),
    C_Vertex3f(0.278091431f, 0.593729019f, -1.f),
    C_Vertex3f(0.268554688f, 0.662109375f, -1.f),
    C_Vertex3f(0.268554688f, 0.595703125f, -1.f),
    C_Vertex3f(0.268554688f, 0.699707031f, -1.f),
    C_Vertex3f(0.268554688f, 0.369354248f, -1.f),
    C_Vertex3f(0.268554688f, 0.288215637f, -1.f),
    C_Vertex3f(0.276557922f, 0.286161423f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.268554688f, 0.052734375f, -1.f),
    C_Vertex3f(0.268554688f, 0.288215637f, -1.f),
    C_Vertex3f(0.268554688f, 0.699707031f, -1.f),
    C_Vertex3f(0.268554688f, -0.0126953125f, -1.f),
    C_Vertex3f(0.278847694f, -0.0122041702f, -1.f),
    C_Vertex3f(0.274248123f, 0.0534448624f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.438476563f, 0.50390625f, -1.f),
    C_Vertex3f(0.43717289f, 0.513273239f, -1.f),
    C_Vertex3f(0.433288813f, 0.531007528f, -1.f),
    C_Vertex3f(0.428128004f, 0.547742128f, -1.f),
    C_Vertex3f(0.421690464f, 0.563477039f, -1.f),
    C_Vertex3f(0.413976192f, 0.578212261f, -1.f),
    C_Vertex3f(0.404985189f, 0.591947794f, -1.f),
    C_Vertex3f(0.394717455f, 0.604683638f, -1.f),
    C_Vertex3f(0.383172989f, 0.616419792f, -1.f),
    C_Vertex3f(0.371765137f, 0.626006126f, -1.f),
    C_Vertex3f(0.360887766f, 0.633593321f, -1.f),
    C_Vertex3f(0.357421875f, 0.491699219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323121071f, 0.569210052f, -1.f),
    C_Vertex3f(0.334067345f, 0.556812286f, -1.f),
    C_Vertex3f(0.336484432f, 0.646220684f, -1.f),
    C_Vertex3f(0.34314394f, 0.541699409f, -1.f),
    C_Vertex3f(0.349127531f, 0.640331507f, -1.f),
    C_Vertex3f(0.350350857f, 0.523871422f, -1.f),
    C_Vertex3f(0.360887766f, 0.633593321f, -1.f),
    C_Vertex3f(0.355688095f, 0.503328323f, -1.f),
    C_Vertex3f(0.357421875f, 0.491699219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.293257952f, 0.658794165f, -1.f),
    C_Vertex3f(0.295133114f, 0.587668419f, -1.f),
    C_Vertex3f(0.308549643f, 0.655452013f, -1.f),
    C_Vertex3f(0.310142994f, 0.57949543f, -1.f),
    C_Vertex3f(0.322958469f, 0.651260853f, -1.f),
    C_Vertex3f(0.323121071f, 0.569210052f, -1.f),
    C_Vertex3f(0.336484432f, 0.646220684f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.373535156f, 0.187614441f, -1.f),
    C_Vertex3f(0.382349133f, 0.0269571543f, -1.f),
    C_Vertex3f(0.38765955f, 0.322794914f, -1.f),
    C_Vertex3f(0.396263599f, 0.0390286446f, -1.f),
    C_Vertex3f(0.400893211f, 0.312088013f, -1.f),
    C_Vertex3f(0.409224987f, 0.0526843071f, -1.f),
    C_Vertex3f(0.412438393f, 0.300506592f, -1.f),
    C_Vertex3f(0.420566201f, 0.0671867132f, -1.f),
    C_Vertex3f(0.422649384f, 0.287833214f, -1.f),
    C_Vertex3f(0.430287242f, 0.0825358629f, -1.f),
    C_Vertex3f(0.431526184f, 0.274067879f, -1.f),
    C_Vertex3f(0.438388109f, 0.0987317562f, -1.f),
    C_Vertex3f(0.439068794f, 0.259210587f, -1.f),
    C_Vertex3f(0.444868803f, 0.115774393f, -1.f),
    C_Vertex3f(0.445277214f, 0.243261337f, -1.f),
    C_Vertex3f(0.449729323f, 0.133663774f, -1.f),
    C_Vertex3f(0.449933529f, 0.226656914f, -1.f),
    C_Vertex3f(0.45296967f, 0.152399898f, -1.f),
    C_Vertex3f(0.453037739f, 0.209397316f, -1.f),
    C_Vertex3f(0.454589844f, 0.171982765f, -1.f),
    C_Vertex3f(0.454589844f, 0.191482544f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.372737408f, 0.332627296f, -1.f),
    C_Vertex3f(0.356126785f, 0.341585159f, -1.f),
    C_Vertex3f(0.358177185f, 0.239261627f, -1.f),
    C_Vertex3f(0.365856171f, 0.224149704f, -1.f),
    C_Vertex3f(0.370975494f, 0.206933975f, -1.f),
    C_Vertex3f(0.373535156f, 0.187614441f, -1.f),
    C_Vertex3f(0.38765955f, 0.322794914f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.358177185f, 0.239261627f, -1.f),
    C_Vertex3f(0.356126785f, 0.341585159f, -1.f),
    C_Vertex3f(0.350498199f, 0.249017715f, -1.f),
    C_Vertex3f(0.34458065f, 0.346664429f, -1.f),
    C_Vertex3f(0.344017982f, 0.255222321f, -1.f),
    C_Vertex3f(0.33853507f, 0.348962307f, -1.f),
    C_Vertex3f(0.336176872f, 0.261127472f, -1.f),
    C_Vertex3f(0.331237078f, 0.351482868f, -1.f),
    C_Vertex3f(0.326974869f, 0.26673317f, -1.f),
    C_Vertex3f(0.322686672f, 0.354226112f, -1.f),
    C_Vertex3f(0.316411972f, 0.272039413f, -1.f),
    C_Vertex3f(0.312883854f, 0.357192039f, -1.f),
    C_Vertex3f(0.304488182f, 0.277046204f, -1.f),
    C_Vertex3f(0.301828623f, 0.36038065f, -1.f),
    C_Vertex3f(0.291203499f, 0.28175354f, -1.f),
    C_Vertex3f(0.289520979f, 0.363791943f, -1.f),
    C_Vertex3f(0.276557922f, 0.286161423f, -1.f),
    C_Vertex3f(0.275960922f, 0.367425919f, -1.f),
    C_Vertex3f(0.268554688f, 0.369354248f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.382349133f, 0.0269571543f, -1.f),
    C_Vertex3f(0.373535156f, 0.187614441f, -1.f),
    C_Vertex3f(0.373535156f, 0.170618057f, -1.f),
    C_Vertex3f(0.372606516f, 0.158554077f, -1.f),
    C_Vertex3f(0.370749235f, 0.146995544f, -1.f),
    C_Vertex3f(0.367963314f, 0.135942459f, -1.f),
    C_Vertex3f(0.367481589f, 0.0164698362f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.367481589f, 0.0164698362f, -1.f),
    C_Vertex3f(0.367963314f, 0.135942459f, -1.f),
    C_Vertex3f(0.364248753f, 0.125394821f, -1.f),
    C_Vertex3f(0.359605551f, 0.115352631f, -1.f),
    C_Vertex3f(0.354033709f, 0.105815887f, -1.f),
    C_Vertex3f(0.351660967f, 0.00756669044f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.354033709f, 0.105815887f, -1.f),
    C_Vertex3f(0.347533226f, 0.0967845917f, -1.f),
    C_Vertex3f(0.351660967f, 0.00756669044f, -1.f),
    C_Vertex3f(0.340104103f, 0.0882587433f, -1.f),
    C_Vertex3f(0.334887266f, 0.000247716904f, -1.f),
    C_Vertex3f(0.332180262f, 0.0806210041f, -1.f),
    C_Vertex3f(0.323761702f, 0.0738713741f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.317160487f, -0.00548708439f, -1.f),
    C_Vertex3f(0.334887266f, 0.000247716904f, -1.f),
    C_Vertex3f(0.323761702f, 0.0738713741f, -1.f),
    C_Vertex3f(0.314848423f, 0.0680098534f, -1.f),
    C_Vertex3f(0.305440426f, 0.0630364418f, -1.f),
    C_Vertex3f(0.29848063f, -0.00963771343f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.274248123f, 0.0534448624f, -1.f),
    C_Vertex3f(0.278847694f, -0.0122041702f, -1.f),
    C_Vertex3f(0.285140276f, 0.0557539463f, -1.f),
    C_Vertex3f(0.29848063f, -0.00963771343f, -1.f),
    C_Vertex3f(0.29553771f, 0.0589511395f, -1.f),
    C_Vertex3f(0.305440426f, 0.0630364418f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.293257952f, 0.658794165f, -1.f),
    C_Vertex3f(0.278091431f, 0.593729019f, -1.f),
    C_Vertex3f(0.295133114f, 0.587668419f, -1.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, -1.f),
    C_Vertex3f(0.222167969f, -0.013671875f, -1.f),
    C_Vertex3f(0.222167969f, -0.0922851563f, -1.f),
    C_Vertex3f(0.222167969f, 0.381263733f, -1.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, -1.f),
    C_Vertex3f(0.222167969f, 0.596679688f, -1.f),
    C_Vertex3f(0.268554688f, 0.699707031f, -1.f),
    C_Vertex3f(0.268554688f, -0.0922851563f, -1.f),
    C_Vertex3f(0.268554688f, -0.0126953125f, -1.f),
    C_Vertex3f(0.13297534f, 0.638587236f, -1.f),
    C_Vertex3f(0.13387537f, 0.536218166f, -1.f),
    C_Vertex3f(0.1464715f, 0.644789219f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_37[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999542296f, -0.0302517843f, 0.f),
    C_Vertex3f(0.0522460938f, 0.475595474f, 0.f),
    C_Vertex3f(0.0522460938f, 0.475595474f, -1.f),
    C_Normal3f(-0.9995507f, 0.029973954f, 0.f),
    C_Vertex3f(0.0522460938f, 0.495107651f, 0.f),
    C_Vertex3f(0.0522460938f, 0.495107651f, -1.f),
    C_Normal3f(-0.995031834f, 0.0995573699f, 0.f),
    C_Vertex3f(0.0533266068f, 0.511627674f, 0.f),
    C_Vertex3f(0.0533266068f, 0.511627674f, -1.f),
    C_Normal3f(-0.985256791f, 0.171082452f, 0.f),
    C_Vertex3f(0.0554876328f, 0.527505398f, 0.f),
    C_Vertex3f(0.0554876328f, 0.527505398f, -1.f),
    C_Normal3f(-0.969323039f, 0.245790139f, 0.f),
    C_Vertex3f(0.0587291718f, 0.542740822f, 0.f),
    C_Vertex3f(0.0587291718f, 0.542740822f, -1.f),
    C_Normal3f(-0.946573734f, 0.322487503f, 0.f),
    C_Vertex3f(0.0630512238f, 0.557333946f, 0.f),
    C_Vertex3f(0.0630512238f, 0.557333946f, -1.f),
    C_Normal3f(-0.916654587f, 0.399680465f, 0.f),
    C_Vertex3f(0.0684537888f, 0.571284771f, 0.f),
    C_Vertex3f(0.0684537888f, 0.571284771f, -1.f),
    C_Normal3f(-0.879617751f, 0.475681156f, 0.f),
    C_Vertex3f(0.0749368668f, 0.584593296f, 0.f),
    C_Vertex3f(0.0749368668f, 0.584593296f, -1.f),
    C_Normal3f(-0.83597362f, 0.548769653f, 0.f),
    C_Vertex3f(0.0825004578f, 0.597259521f, 0.f),
    C_Vertex3f(0.0825004578f, 0.597259521f, -1.f),
    C_Normal3f(-0.77710247f, 0.629374087f, 0.f),
    C_Vertex3f(0.0911445618f, 0.609283447f, 0.f),
    C_Vertex3f(0.0911445618f, 0.609283447f, -1.f),
    C_Normal3f(-0.695233703f, 0.718783677f, 0.f),
    C_Vertex3f(0.100759506f, 0.619804382f, 0.f),
    C_Vertex3f(0.100759506f, 0.619804382f, -1.f),
    C_Normal3f(-0.598286867f, 0.801282048f, 0.f),
    C_Vertex3f(0.111345291f, 0.628822327f, 0.f),
    C_Vertex3f(0.111345291f, 0.628822327f, -1.f),
    C_Normal3f(-0.489700556f, 0.871890724f, 0.f),
    C_Vertex3f(0.122901917f, 0.63633728f, 0.f),
    C_Vertex3f(0.122901917f, 0.63633728f, -1.f),
    C_Normal3f(-0.374785423f, 0.927111566f, 0.f),
    C_Vertex3f(0.135429382f, 0.642349243f, 0.f),
    C_Vertex3f(0.135429382f, 0.642349243f, -1.f),
    C_Normal3f(-0.259498686f, 0.965743423f, 0.f),
    C_Vertex3f(0.148927689f, 0.646858215f, 0.f),
    C_Vertex3f(0.148927689f, 0.646858215f, -1.f),
    C_Normal3f(-0.149071217f, 0.988826454f, 0.f),
    C_Vertex3f(0.163396835f, 0.649864197f, 0.f),
    C_Vertex3f(0.163396835f, 0.649864197f, -1.f),
    C_Normal3f(-0.0481236875f, 0.998841345f, 0.f),
    C_Vertex3f(0.178836823f, 0.651367188f, 0.f),
    C_Vertex3f(0.178836823f, 0.651367188f, -1.f),
    C_Normal3f(0.0445118397f, 0.999008954f, 0.f),
    C_Vertex3f(0.194592476f, 0.651367188f, 0.f),
    C_Vertex3f(0.194592476f, 0.651367188f, -1.f),
    C_Normal3f(0.141479254f, 0.989941239f, 0.f),
    C_Vertex3f(0.209043145f, 0.650021315f, 0.f),
    C_Vertex3f(0.209043145f, 0.650021315f, -1.f),
    C_Normal3f(0.242243826f, 0.97021544f, 0.f),
    C_Vertex3f(0.222844005f, 0.647329569f, 0.f),
    C_Vertex3f(0.222844005f, 0.647329569f, -1.f),
    C_Normal3f(0.344747186f, 0.93869555f, 0.f),
    C_Vertex3f(0.235995054f, 0.64329195f, 0.f),
    C_Vertex3f(0.235995054f, 0.64329195f, -1.f),
    C_Normal3f(0.445345551f, 0.89535886f, 0.f),
    C_Vertex3f(0.248496294f, 0.637908459f, 0.f),
    C_Vertex3f(0.248496294f, 0.637908459f, -1.f),
    C_Normal3f(0.540364563f, 0.841431022f, 0.f),
    C_Vertex3f(0.260347724f, 0.631179094f, 0.f),
    C_Vertex3f(0.260347724f, 0.631179094f, -1.f),
    C_Normal3f(0.626737356f, 0.779230535f, 0.f),
    C_Vertex3f(0.271549344f, 0.623103857f, 0.f),
    C_Vertex3f(0.271549344f, 0.623103857f, -1.f),
    C_Normal3f(0.702468693f, 0.711714625f, 0.f),
    C_Vertex3f(0.282101154f, 0.613682747f, 0.f),
    C_Vertex3f(0.282101154f, 0.613682747f, -1.f),
    C_Normal3f(0.775400639f, 0.631469607f, 0.f),
    C_Vertex3f(0.292003155f, 0.602915764f, 0.f),
    C_Vertex3f(0.292003155f, 0.602915764f, -1.f),
    C_Normal3f(0.84405303f, 0.536259651f, 0.f),
    C_Vertex3f(0.300667405f, 0.590884686f, 0.f),
    C_Vertex3f(0.300667405f, 0.590884686f, -1.f),
    C_Normal3f(0.898415864f, 0.439145625f, 0.f),
    C_Vertex3f(0.308093905f, 0.577589512f, 0.f),
    C_Vertex3f(0.308093905f, 0.577589512f, -1.f),
    C_Normal3f(0.938881755f, 0.344239712f, 0.f),
    C_Vertex3f(0.314282656f, 0.563030243f, 0.f),
    C_Vertex3f(0.314282656f, 0.563030243f, -1.f),
    C_Normal3f(0.967049181f, 0.254589677f, 0.f),
    C_Vertex3f(0.319233656f, 0.547206879f, 0.f),
    C_Vertex3f(0.319233656f, 0.547206879f, -1.f),
    C_Normal3f(0.985092163f, 0.172027618f, 0.f),
    C_Vertex3f(0.322946906f, 0.530119419f, 0.f),
    C_Vertex3f(0.322946906f, 0.530119419f, -1.f),
    C_Normal3f(0.995251417f, 0.0973370969f, 0.f),
    C_Vertex3f(0.325422406f, 0.511767864f, 0.f),
    C_Vertex3f(0.325422406f, 0.511767864f, -1.f),
    C_Normal3f(0.999526858f, 0.0307593662f, 0.f),
    C_Vertex3f(0.326660156f, 0.492152214f, 0.f),
    C_Vertex3f(0.326660156f, 0.492152214f, -1.f),
    C_Normal3f(0.999505162f, -0.0314538181f, 0.f),
    C_Vertex3f(0.326660156f, 0.471547127f, 0.f),
    C_Vertex3f(0.326660156f, 0.471547127f, -1.f),
    C_Normal3f(0.994934082f, -0.100529537f, 0.f),
    C_Vertex3f(0.325408697f, 0.45238471f, 0.f),
    C_Vertex3f(0.325408697f, 0.45238471f, -1.f),
    C_Normal3f(0.98423481f, -0.176866949f, 0.f),
    C_Vertex3f(0.322905779f, 0.434390306f, 0.f),
    C_Vertex3f(0.322905779f, 0.434390306f, -1.f),
    C_Normal3f(0.965508878f, -0.260370135f, 0.f),
    C_Vertex3f(0.319151402f, 0.417563915f, 0.f),
    C_Vertex3f(0.319151402f, 0.417563915f, -1.f),
    C_Normal3f(0.936764181f, -0.34996146f, 0.f),
    C_Vertex3f(0.314145565f, 0.401905537f, 0.f),
    C_Vertex3f(0.314145565f, 0.401905537f, -1.f),
    C_Normal3f(0.896228433f, -0.443592936f, 0.f),
    C_Vertex3f(0.307888269f, 0.387415171f, 0.f),
    C_Vertex3f(0.307888269f, 0.387415171f, -1.f),
    C_Normal3f(0.842813432f, -0.538205802f, 0.f),
    C_Vertex3f(0.300379515f, 0.374092817f, 0.f),
    C_Vertex3f(0.300379515f, 0.374092817f, -1.f),
    C_Normal3f(0.77660501f, -0.629987895f, 0.f),
    C_Vertex3f(0.291619301f, 0.361938477f, 0.f),
    C_Vertex3f(0.291619301f, 0.361938477f, -1.f),
    C_Normal3f(0.707016706f, -0.707196772f, 0.f),
    C_Vertex3f(0.281607628f, 0.350952148f, 0.f),
    C_Vertex3f(0.281607628f, 0.350952148f, -1.f),
    C_Normal3f(0.63415879f, -0.773202837f, 0.f),
    C_Vertex3f(0.27101469f, 0.341339111f, 0.f),
    C_Vertex3f(0.27101469f, 0.341339111f, -1.f),
    C_Normal3f(0.550147414f, -0.835067511f, 0.f),
    C_Vertex3f(0.259840488f, 0.333099365f, 0.f),
    C_Vertex3f(0.259840488f, 0.333099365f, -1.f),
    C_Normal3f(0.456452489f, -0.889747739f, 0.f),
    C_Vertex3f(0.248085022f, 0.32623291f, 0.f),
    C_Vertex3f(0.248085022f, 0.32623291f, -1.f),
    C_Normal3f(0.355742872f, -0.934583843f, 0.f),
    C_Vertex3f(0.235748291f, 0.320739746f, 0.f),
    C_Vertex3f(0.235748291f, 0.320739746f, -1.f),
    C_Normal3f(0.251564145f, -0.967840672f, 0.f),
    C_Vertex3f(0.222830296f, 0.316619873f, 0.f),
    C_Vertex3f(0.222830296f, 0.316619873f, -1.f),
    C_Normal3f(0.147740901f, -0.989026129f, 0.f),
    C_Vertex3f(0.209331036f, 0.313873291f, 0.f),
    C_Vertex3f(0.209331036f, 0.313873291f, -1.f),
    C_Normal3f(0.0478167869f, -0.998856187f, 0.f),
    C_Vertex3f(0.195250511f, 0.3125f, 0.f),
    C_Vertex3f(0.195250511f, 0.3125f, -1.f),
    C_Normal3f(-0.0473313145f, -0.998879313f, 0.f),
    C_Vertex3f(0.180644035f, 0.3125f, 0.f),
    C_Vertex3f(0.180644035f, 0.3125f, -1.f),
    C_Normal3f(-0.147163376f, -0.989112198f, 0.f),
    C_Vertex3f(0.166701317f, 0.313852787f, 0.f),
    C_Vertex3f(0.166701317f, 0.313852787f, -1.f),
    C_Normal3f(-0.251227051f, -0.967928112f, 0.f),
    C_Vertex3f(0.153367043f, 0.316558361f, 0.f),
    C_Vertex3f(0.153367043f, 0.316558361f, -1.f),
    C_Normal3f(-0.356173575f, -0.934419811f, 0.f),
    C_Vertex3f(0.140641212f, 0.320616722f, 0.f),
    C_Vertex3f(0.140641212f, 0.320616722f, -1.f),
    C_Normal3f(-0.458073854f, -0.888914049f, 0.f),
    C_Vertex3f(0.128523827f, 0.32602787f, 0.f),
    C_Vertex3f(0.128523827f, 0.32602787f, -1.f),
    C_Normal3f(-0.553181589f, -0.833060622f, 0.f),
    C_Vertex3f(0.117014885f, 0.332791805f, 0.f),
    C_Vertex3f(0.117014885f, 0.332791805f, -1.f),
    C_Normal3f(-0.638592303f, -0.769545197f, 0.f),
    C_Vertex3f(0.106114388f, 0.340908527f, 0.f),
    C_Vertex3f(0.106114388f, 0.340908527f, -1.f),
    C_Normal3f(-0.712636352f, -0.701533556f, 0.f),
    C_Vertex3f(0.0958223343f, 0.350378036f, 0.f),
    C_Vertex3f(0.0958223343f, 0.350378036f, -1.f),
    C_Normal3f(-0.785222948f, -0.619213164f, 0.f),
    C_Vertex3f(0.0861387253f, 0.361200333f, 0.f),
    C_Vertex3f(0.0861387253f, 0.361200333f, -1.f),
    C_Normal3f(-0.853595793f, -0.520935893f, 0.f),
    C_Vertex3f(0.0776655674f, 0.373402596f, 0.f),
    C_Vertex3f(0.0776655674f, 0.373402596f, -1.f),
    C_Normal3f(-0.906253219f, -0.422735006f, 0.f),
    C_Vertex3f(0.0704028606f, 0.386984825f, 0.f),
    C_Vertex3f(0.0704028606f, 0.386984825f, -1.f),
    C_Normal3f(-0.944444418f, -0.328670979f, 0.f),
    C_Vertex3f(0.064350605f, 0.401947021f, 0.f),
    C_Vertex3f(0.064350605f, 0.401947021f, -1.f),
    C_Normal3f(-0.97042954f, -0.241384804f, 0.f),
    C_Vertex3f(0.0595088005f, 0.418289185f, 0.f),
    C_Vertex3f(0.0595088005f, 0.418289185f, -1.f),
    C_Normal3f(-0.986761332f, -0.162179425f, 0.f),
    C_Vertex3f(0.0558774471f, 0.436011314f, 0.f),
    C_Vertex3f(0.0558774471f, 0.436011314f, -1.f),
    C_Normal3f(-0.995818496f, -0.091353938f, 0.f),
    C_Vertex3f(0.0534565449f, 0.455113411f, 0.f),
    C_Vertex3f(0.0534565449f, 0.455113411f, -1.f),
    C_Normal3f(-0.999542296f, -0.0302517843f, 0.f),
    C_Vertex3f(0.0522460938f, 0.475595474f, 0.f),
    C_Vertex3f(0.0522460938f, 0.475595474f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.111304335f, -0.993786335f, 0.f),
    C_Vertex3f(0.196548462f, 0.597167969f, 0.f),
    C_Vertex3f(0.196548462f, 0.597167969f, -1.f),
    C_Normal3f(0.112364031f, -0.993667126f, 0.f),
    C_Vertex3f(0.181915283f, 0.597167969f, 0.f),
    C_Vertex3f(0.181915283f, 0.597167969f, -1.f),
    C_Normal3f(0.348519921f, -0.937301397f, 0.f),
    C_Vertex3f(0.168510914f, 0.593997478f, 0.f),
    C_Vertex3f(0.168510914f, 0.593997478f, -1.f),
    C_Normal3f(0.565459251f, -0.824776232f, 0.f),
    C_Vertex3f(0.156335354f, 0.587656498f, 0.f),
    C_Vertex3f(0.156335354f, 0.587656498f, -1.f),
    C_Normal3f(0.721895516f, -0.692002058f, 0.f),
    C_Vertex3f(0.145388603f, 0.578145027f, 0.f),
    C_Vertex3f(0.145388603f, 0.578145027f, -1.f),
    C_Normal3f(0.830083966f, -0.557638586f, 0.f),
    C_Vertex3f(0.138100147f, 0.568633556f, 0.f),
    C_Vertex3f(0.138100147f, 0.568633556f, -1.f),
    C_Normal3f(0.903640091f, -0.428292572f, 0.f),
    C_Vertex3f(0.133848548f, 0.560966849f, 0.f),
    C_Vertex3f(0.133848548f, 0.560966849f, -1.f),
    C_Normal3f(0.945036769f, -0.32696408f, 0.f),
    C_Vertex3f(0.13020432f, 0.551974416f, 0.f),
    C_Vertex3f(0.13020432f, 0.551974416f, -1.f),
    C_Normal3f(0.9703933f, -0.241530389f, 0.f),
    C_Vertex3f(0.127167463f, 0.541656256f, 0.f),
    C_Vertex3f(0.127167463f, 0.541656256f, -1.f),
    C_Normal3f(0.985407174f, -0.170213774f, 0.f),
    C_Vertex3f(0.124737978f, 0.530012369f, 0.f),
    C_Vertex3f(0.124737978f, 0.530012369f, -1.f),
    C_Normal3f(0.993854046f, -0.110698603f, 0.f),
    C_Vertex3f(0.122915864f, 0.517042756f, 0.f),
    C_Vertex3f(0.122915864f, 0.517042756f, -1.f),
    C_Normal3f(0.998150289f, -0.0607941896f, 0.f),
    C_Vertex3f(0.121701121f, 0.502747416f, 0.f),
    C_Vertex3f(0.121701121f, 0.502747416f, -1.f),
    C_Normal3f(0.999817669f, -0.0190950092f, 0.f),
    C_Vertex3f(0.12109375f, 0.48712635f, 0.f),
    C_Vertex3f(0.12109375f, 0.48712635f, -1.f),
    C_Normal3f(0.999796689f, 0.0201660916f, 0.f),
    C_Vertex3f(0.12109375f, 0.470945358f, 0.f),
    C_Vertex3f(0.12109375f, 0.470945358f, -1.f),
    C_Normal3f(0.997747064f, 0.067088455f, 0.f),
    C_Vertex3f(0.121708035f, 0.456671238f, 0.f),
    C_Vertex3f(0.121708035f, 0.456671238f, -1.f),
    C_Normal3f(0.992610693f, 0.121342421f, 0.f),
    C_Vertex3f(0.122936606f, 0.443538189f, 0.f),
    C_Vertex3f(0.122936606f, 0.443538189f, -1.f),
    C_Normal3f(0.982739747f, 0.184993237f, 0.f),
    C_Vertex3f(0.124779463f, 0.431546211f, 0.f),
    C_Vertex3f(0.124779463f, 0.431546211f, -1.f),
    C_Normal3f(0.965698421f, 0.259666264f, 0.f),
    C_Vertex3f(0.127236605f, 0.420695305f, 0.f),
    C_Vertex3f(0.127236605f, 0.420695305f, -1.f),
    C_Normal3f(0.937961102f, 0.346740723f, 0.f),
    C_Vertex3f(0.130308032f, 0.41098547f, 0.f),
    C_Vertex3f(0.130308032f, 0.41098547f, -1.f),
    C_Normal3f(0.894707084f, 0.446653426f, 0.f),
    C_Vertex3f(0.133993745f, 0.402416706f, 0.f),
    C_Vertex3f(0.133993745f, 0.402416706f, -1.f),
    C_Normal3f(0.822174847f, 0.569235086f, 0.f),
    C_Vertex3f(0.138293743f, 0.394989014f, 0.f),
    C_Vertex3f(0.138293743f, 0.394989014f, -1.f),
    C_Normal3f(0.716807783f, 0.697270989f, 0.f),
    C_Vertex3f(0.145665169f, 0.385559082f, 0.f),
    C_Vertex3f(0.145665169f, 0.385559082f, -1.f),
    C_Normal3f(0.562608302f, 0.826723576f, 0.f),
    C_Vertex3f(0.156639576f, 0.37612915f, 0.f),
    C_Vertex3f(0.156639576f, 0.37612915f, -1.f),
    C_Normal3f(0.348209977f, 0.937416613f, 0.f),
    C_Vertex3f(0.168759823f, 0.369842529f, 0.f),
    C_Vertex3f(0.168759823f, 0.369842529f, -1.f),
    C_Normal3f(0.111958086f, 0.993712962f, 0.f),
    C_Vertex3f(0.182025909f, 0.366699219f, 0.f),
    C_Vertex3f(0.182025909f, 0.366699219f, -1.f),
    C_Normal3f(-0.111711197f, 0.993740737f, 0.f),
    C_Vertex3f(0.196659088f, 0.366699219f, 0.f),
    C_Vertex3f(0.196659088f, 0.366699219f, -1.f),
    C_Normal3f(-0.345238507f, 0.938515007f, 0.f),
    C_Vertex3f(0.210229397f, 0.369869709f, 0.f),
    C_Vertex3f(0.210229397f, 0.369869709f, -1.f),
    C_Normal3f(-0.562709033f, 0.82665503f, 0.f),
    C_Vertex3f(0.222515583f, 0.37621069f, 0.f),
    C_Vertex3f(0.222515583f, 0.37621069f, -1.f),
    C_Normal3f(-0.720847607f, 0.693093479f, 0.f),
    C_Vertex3f(0.233517647f, 0.38572216f, 0.f),
    C_Vertex3f(0.233517647f, 0.38572216f, -1.f),
    C_Normal3f(-0.829879522f, 0.557942629f, 0.f),
    C_Vertex3f(0.240806103f, 0.395233631f, 0.f),
    C_Vertex3f(0.240806103f, 0.395233631f, -1.f),
    C_Normal3f(-0.90323323f, 0.429149985f, 0.f),
    C_Vertex3f(0.245057702f, 0.402886748f, 0.f),
    C_Vertex3f(0.245057702f, 0.402886748f, -1.f),
    C_Normal3f(-0.944679618f, 0.327994645f, 0.f),
    C_Vertex3f(0.24870193f, 0.411852002f, 0.f),
    C_Vertex3f(0.24870193f, 0.411852002f, -1.f),
    C_Normal3f(-0.970146537f, 0.242519438f, 0.f),
    C_Vertex3f(0.251738787f, 0.422129393f, 0.f),
    C_Vertex3f(0.251738787f, 0.422129393f, -1.f),
    C_Normal3f(-0.985264242f, 0.171039f, 0.f),
    C_Vertex3f(0.254168272f, 0.43371892f, 0.f),
    C_Vertex3f(0.254168272f, 0.43371892f, -1.f),
    C_Normal3f(-0.993786693f, 0.111301348f, 0.f),
    C_Vertex3f(0.255990386f, 0.446620584f, 0.f),
    C_Vertex3f(0.255990386f, 0.446620584f, -1.f),
    C_Normal3f(-0.998128355f, 0.061153993f, 0.f),
    C_Vertex3f(0.257205129f, 0.460834384f, 0.f),
    C_Vertex3f(0.257205129f, 0.460834384f, -1.f),
    C_Normal3f(-0.999816597f, 0.0191522781f, 0.f),
    C_Vertex3f(0.2578125f, 0.476360321f, 0.f),
    C_Vertex3f(0.2578125f, 0.476360321f, -1.f),
    C_Normal3f(-0.999798f, -0.0201033223f, 0.f),
    C_Vertex3f(0.2578125f, 0.492541313f, 0.f),
    C_Vertex3f(0.2578125f, 0.492541313f, -1.f),
    C_Normal3f(-0.997775733f, -0.0666606799f, 0.f),
    C_Vertex3f(0.257198215f, 0.506910563f, 0.f),
    C_Vertex3f(0.257198215f, 0.506910563f, -1.f),
    C_Normal3f(-0.992696941f, -0.120635211f, 0.f),
    C_Vertex3f(0.255969644f, 0.520125151f, 0.f),
    C_Vertex3f(0.255969644f, 0.520125151f, -1.f),
    C_Normal3f(-0.982918441f, -0.18404147f, 0.f),
    C_Vertex3f(0.254126787f, 0.532185078f, 0.f),
    C_Vertex3f(0.254126787f, 0.532185078f, -1.f),
    C_Normal3f(-0.965997696f, -0.258550465f, 0.f),
    C_Vertex3f(0.251669645f, 0.543090343f, 0.f),
    C_Vertex3f(0.251669645f, 0.543090343f, -1.f),
    C_Normal3f(-0.938378155f, -0.345610112f, 0.f),
    C_Vertex3f(0.248598218f, 0.552840948f, 0.f),
    C_Vertex3f(0.248598218f, 0.552840948f, -1.f),
    C_Normal3f(-0.895160556f, -0.445743769f, 0.f),
    C_Vertex3f(0.244912505f, 0.561436892f, 0.f),
    C_Vertex3f(0.244912505f, 0.561436892f, -1.f),
    C_Normal3f(-0.822389424f, -0.568925023f, 0.f),
    C_Vertex3f(0.240612507f, 0.568878174f, 0.f),
    C_Vertex3f(0.240612507f, 0.568878174f, -1.f),
    C_Normal3f(-0.715757668f, -0.698348701f, 0.f),
    C_Vertex3f(0.233241081f, 0.578308105f, 0.f),
    C_Vertex3f(0.233241081f, 0.578308105f, -1.f),
    C_Normal3f(-0.559855759f, -0.828589976f, 0.f),
    C_Vertex3f(0.222211361f, 0.587738037f, 0.f),
    C_Vertex3f(0.222211361f, 0.587738037f, -1.f),
    C_Normal3f(-0.344905913f, -0.938637316f, 0.f),
    C_Vertex3f(0.209980488f, 0.594024658f, 0.f),
    C_Vertex3f(0.209980488f, 0.594024658f, -1.f),
    C_Normal3f(-0.111304335f, -0.993786335f, 0.f),
    C_Vertex3f(0.196548462f, 0.597167969f, 0.f),
    C_Vertex3f(0.196548462f, 0.597167969f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.887815773f, 0.460199118f, 0.f),
    C_Vertex3f(0.189674377f, -0.0234375f, 0.f),
    C_Vertex3f(0.189674377f, -0.0234375f, -1.f),
    C_Vertex3f(0.539459229f, 0.651367188f, -1.f),
    C_Vertex3f(0.539459229f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.539459229f, 0.651367188f, 0.f),
    C_Vertex3f(0.539459229f, 0.651367188f, -1.f),
    C_Vertex3f(0.60345459f, 0.651367188f, -1.f),
    C_Vertex3f(0.60345459f, 0.651367188f, 0.f),
    C_Normal3f(0.888405979f, -0.459058642f, 0.f),
    C_Vertex3f(0.60345459f, 0.651367188f, 0.f),
    C_Vertex3f(0.60345459f, 0.651367188f, -1.f),
    C_Vertex3f(0.254768372f, -0.0234375f, -1.f),
    C_Vertex3f(0.254768372f, -0.0234375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.254768372f, -0.0234375f, 0.f),
    C_Vertex3f(0.254768372f, -0.0234375f, -1.f),
    C_Vertex3f(0.189674377f, -0.0234375f, -1.f),
    C_Vertex3f(0.189674377f, -0.0234375f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999541819f, -0.0302679744f, 0.f),
    C_Vertex3f(0.465820313f, 0.139712334f, 0.f),
    C_Vertex3f(0.465820313f, 0.139712334f, -1.f),
    C_Normal3f(-0.999554336f, 0.0298512783f, 0.f),
    C_Vertex3f(0.465820313f, 0.15922451f, 0.f),
    C_Vertex3f(0.465820313f, 0.15922451f, -1.f),
    C_Normal3f(-0.995095313f, 0.0989217609f, 0.f),
    C_Vertex3f(0.466899037f, 0.175832868f, 0.f),
    C_Vertex3f(0.466899037f, 0.175832868f, -1.f),
    C_Normal3f(-0.985410571f, 0.170194119f, 0.f),
    C_Vertex3f(0.469056487f, 0.191778541f, 0.f),
    C_Vertex3f(0.469056487f, 0.191778541f, -1.f),
    C_Normal3f(-0.96956563f, 0.244831294f, 0.f),
    C_Vertex3f(0.472292662f, 0.207061529f, 0.f),
    C_Vertex3f(0.472292662f, 0.207061529f, -1.f),
    C_Normal3f(-0.94685328f, 0.321666032f, 0.f),
    C_Vertex3f(0.476607561f, 0.221681833f, 0.f),
    C_Vertex3f(0.476607561f, 0.221681833f, -1.f),
    C_Normal3f(-0.916860342f, 0.399208099f, 0.f),
    C_Vertex3f(0.482001185f, 0.235639453f, 0.f),
    C_Vertex3f(0.482001185f, 0.235639453f, -1.f),
    C_Normal3f(-0.879582644f, 0.475746185f, 0.f),
    C_Vertex3f(0.488473535f, 0.248934388f, 0.f),
    C_Vertex3f(0.488473535f, 0.248934388f, -1.f),
    C_Normal3f(-0.835487843f, 0.549509108f, 0.f),
    C_Vertex3f(0.496024609f, 0.261566639f, 0.f),
    C_Vertex3f(0.496024609f, 0.261566639f, -1.f),
    C_Normal3f(-0.775977373f, 0.63076061f, 0.f),
    C_Vertex3f(0.504654408f, 0.273536205f, 0.f),
    C_Vertex3f(0.504654408f, 0.273536205f, -1.f),
    C_Normal3f(-0.693445206f, 0.720509291f, 0.f),
    C_Vertex3f(0.514267564f, 0.284009576f, 0.f),
    C_Vertex3f(0.514267564f, 0.284009576f, -1.f),
    C_Normal3f(-0.595956802f, 0.803016424f, 0.f),
    C_Vertex3f(0.524864078f, 0.292986751f, 0.f),
    C_Vertex3f(0.524864078f, 0.292986751f, -1.f),
    C_Normal3f(-0.48710534f, 0.873343289f, 0.f),
    C_Vertex3f(0.536443949f, 0.30046773f, 0.f),
    C_Vertex3f(0.536443949f, 0.30046773f, -1.f),
    C_Normal3f(-0.372292787f, 0.928115368f, 0.f),
    C_Vertex3f(0.549007177f, 0.306452513f, 0.f),
    C_Vertex3f(0.549007177f, 0.306452513f, -1.f),
    C_Normal3f(-0.257466793f, 0.966287196f, 0.f),
    C_Vertex3f(0.562553763f, 0.3109411f, 0.f),
    C_Vertex3f(0.562553763f, 0.3109411f, -1.f),
    C_Normal3f(-0.147764146f, 0.989022613f, 0.f),
    C_Vertex3f(0.577083707f, 0.313933492f, 0.f),
    C_Vertex3f(0.577083707f, 0.313933492f, -1.f),
    C_Normal3f(-0.0477495156f, 0.998859286f, 0.f),
    C_Vertex3f(0.592597008f, 0.315429688f, 0.f),
    C_Vertex3f(0.592597008f, 0.315429688f, -1.f),
    C_Normal3f(0.0445055254f, 0.999009132f, 0.f),
    C_Vertex3f(0.608382225f, 0.315429688f, 0.f),
    C_Vertex3f(0.608382225f, 0.315429688f, -1.f),
    C_Normal3f(0.141720682f, 0.989906669f, 0.f),
    C_Vertex3f(0.622807622f, 0.314083815f, 0.f),
    C_Vertex3f(0.622807622f, 0.314083815f, -1.f),
    C_Normal3f(0.242635757f, 0.970117509f, 0.f),
    C_Vertex3f(0.63658464f, 0.311392069f, 0.f),
    C_Vertex3f(0.63658464f, 0.311392069f, -1.f),
    C_Normal3f(0.345261753f, 0.938506424f, 0.f),
    C_Vertex3f(0.649713278f, 0.30735445f, 0.f),
    C_Vertex3f(0.649713278f, 0.30735445f, -1.f),
    C_Normal3f(0.445940405f, 0.895062685f, 0.f),
    C_Vertex3f(0.662193537f, 0.301970959f, 0.f),
    C_Vertex3f(0.662193537f, 0.301970959f, -1.f),
    C_Normal3f(0.540990233f, 0.84102881f, 0.f),
    C_Vertex3f(0.674025416f, 0.295241594f, 0.f),
    C_Vertex3f(0.674025416f, 0.295241594f, -1.f),
    C_Normal3f(0.627346754f, 0.778740048f, 0.f),
    C_Vertex3f(0.685208917f, 0.287166357f, 0.f),
    C_Vertex3f(0.685208917f, 0.287166357f, -1.f),
    C_Normal3f(0.703024745f, 0.711165428f, 0.f),
    C_Vertex3f(0.695744038f, 0.277745247f, 0.f),
    C_Vertex3f(0.695744038f, 0.277745247f, -1.f),
    C_Normal3f(0.775877118f, 0.630883992f, 0.f),
    C_Vertex3f(0.705630779f, 0.266978264f, 0.f),
    C_Vertex3f(0.705630779f, 0.266978264f, -1.f),
    C_Normal3f(0.84442699f, 0.535670519f, 0.f),
    C_Vertex3f(0.714281678f, 0.254947186f, 0.f),
    C_Vertex3f(0.714281678f, 0.254947186f, -1.f),
    C_Normal3f(0.898682773f, 0.438599199f, 0.f),
    C_Vertex3f(0.721696734f, 0.241652012f, 0.f),
    C_Vertex3f(0.721696734f, 0.241652012f, -1.f),
    C_Normal3f(0.939053118f, 0.343771964f, 0.f),
    C_Vertex3f(0.727875948f, 0.227092743f, 0.f),
    C_Vertex3f(0.727875948f, 0.227092743f, -1.f),
    C_Normal3f(0.967145681f, 0.254222751f, 0.f),
    C_Vertex3f(0.732819319f, 0.211269379f, 0.f),
    C_Vertex3f(0.732819319f, 0.211269379f, -1.f),
    C_Normal3f(0.985137045f, 0.171770364f, 0.f),
    C_Vertex3f(0.736526847f, 0.194181919f, 0.f),
    C_Vertex3f(0.736526847f, 0.194181919f, -1.f),
    C_Normal3f(0.99526602f, 0.0971885324f, 0.f),
    C_Vertex3f(0.738998532f, 0.175830364f, 0.f),
    C_Vertex3f(0.738998532f, 0.175830364f, -1.f),
    C_Normal3f(0.999528289f, 0.0307120122f, 0.f),
    C_Vertex3f(0.740234375f, 0.156214714f, 0.f),
    C_Vertex3f(0.740234375f, 0.156214714f, -1.f),
    C_Normal3f(0.999506831f, -0.0314029381f, 0.f),
    C_Vertex3f(0.740234375f, 0.135609627f, 0.f),
    C_Vertex3f(0.740234375f, 0.135609627f, -1.f),
    C_Normal3f(0.994950294f, -0.100368381f, 0.f),
    C_Vertex3f(0.738984942f, 0.11644721f, 0.f),
    C_Vertex3f(0.738984942f, 0.11644721f, -1.f),
    C_Normal3f(0.984284639f, -0.176589489f, 0.f),
    C_Vertex3f(0.736486077f, 0.0984528065f, 0.f),
    C_Vertex3f(0.736486077f, 0.0984528065f, -1.f),
    C_Normal3f(0.965614736f, -0.259977013f, 0.f),
    C_Vertex3f(0.73273778f, 0.0816264153f, 0.f),
    C_Vertex3f(0.73273778f, 0.0816264153f, -1.f),
    C_Normal3f(0.93694979f, -0.349463969f, 0.f),
    C_Vertex3f(0.727740049f, 0.0659680367f, 0.f),
    C_Vertex3f(0.727740049f, 0.0659680367f, -1.f),
    C_Normal3f(0.89651382f, -0.443015635f, 0.f),
    C_Vertex3f(0.721492887f, 0.0514776707f, 0.f),
    C_Vertex3f(0.721492887f, 0.0514776707f, -1.f),
    C_Normal3f(0.843208671f, -0.537586272f, 0.f),
    C_Vertex3f(0.713996291f, 0.0381553173f, 0.f),
    C_Vertex3f(0.713996291f, 0.0381553173f, -1.f),
    C_Normal3f(0.777104139f, -0.629371941f, 0.f),
    C_Vertex3f(0.705250263f, 0.0260009766f, 0.f),
    C_Vertex3f(0.705250263f, 0.0260009766f, -1.f),
    C_Normal3f(0.707354367f, -0.706859052f, 0.f),
    C_Vertex3f(0.695254803f, 0.0150146484f, 0.f),
    C_Vertex3f(0.695254803f, 0.0150146484f, -1.f),
    C_Normal3f(0.634057045f, -0.773286223f, 0.f),
    C_Vertex3f(0.684665322f, 0.00540161133f, 0.f),
    C_Vertex3f(0.684665322f, 0.00540161133f, -1.f),
    C_Normal3f(0.549623251f, -0.835412621f, 0.f),
    C_Vertex3f(0.673481822f, -0.00283813477f, 0.f),
    C_Vertex3f(0.673481822f, -0.00283813477f, -1.f),
    C_Normal3f(0.455601037f, -0.890184104f, 0.f),
    C_Vertex3f(0.661704302f, -0.00970458984f, 0.f),
    C_Vertex3f(0.661704302f, -0.00970458984f, -1.f),
    C_Normal3f(0.354732096f, -0.934968054f, 0.f),
    C_Vertex3f(0.649332762f, -0.0151977539f, 0.f),
    C_Vertex3f(0.649332762f, -0.0151977539f, -1.f),
    C_Normal3f(0.250605345f, -0.968089283f, 0.f),
    C_Vertex3f(0.636367202f, -0.019317627f, 0.f),
    C_Vertex3f(0.636367202f, -0.019317627f, -1.f),
    C_Normal3f(0.147045165f, -0.989129722f, 0.f),
    C_Vertex3f(0.622807622f, -0.022064209f, 0.f),
    C_Vertex3f(0.622807622f, -0.022064209f, -1.f),
    C_Normal3f(0.0476419777f, -0.998864532f, 0.f),
    C_Vertex3f(0.608654022f, -0.0234375f, 0.f),
    C_Vertex3f(0.608654022f, -0.0234375f, -1.f),
    C_Normal3f(-0.0475485139f, -0.998869002f, 0.f),
    C_Vertex3f(0.594015121f, -0.0234375f, 0.f),
    C_Vertex3f(0.594015121f, -0.0234375f, -1.f),
    C_Normal3f(-0.14809978f, -0.988972366f, 0.f),
    C_Vertex3f(0.580092788f, -0.0220779181f, 0.f),
    C_Vertex3f(0.580092788f, -0.0220779181f, -1.f),
    C_Normal3f(-0.252762735f, -0.967528284f, 0.f),
    C_Vertex3f(0.566778302f, -0.0193587542f, 0.f),
    C_Vertex3f(0.566778302f, -0.0193587542f, -1.f),
    C_Normal3f(-0.358209252f, -0.933641315f, 0.f),
    C_Vertex3f(0.554071665f, -0.0152800083f, 0.f),
    C_Vertex3f(0.554071665f, -0.0152800083f, -1.f),
    C_Normal3f(-0.460451365f, -0.887684882f, 0.f),
    C_Vertex3f(0.541972876f, -0.00984168053f, 0.f),
    C_Vertex3f(0.541972876f, -0.00984168053f, -1.f),
    C_Normal3f(-0.555712759f, -0.831374288f, 0.f),
    C_Vertex3f(0.530481935f, -0.00304377079f, 0.f),
    C_Vertex3f(0.530481935f, -0.00304377079f, -1.f),
    C_Normal3f(-0.641096294f, -0.767460525f, 0.f),
    C_Vertex3f(0.519598842f, 0.00511372089f, 0.f),
    C_Vertex3f(0.519598842f, 0.00511372089f, -1.f),
    C_Normal3f(-0.714969695f, -0.699155331f, 0.f),
    C_Vertex3f(0.509323597f, 0.0146307945f, 0.f),
    C_Vertex3f(0.509323597f, 0.0146307945f, -1.f),
    C_Normal3f(-0.78687644f, -0.61711067f, 0.f),
    C_Vertex3f(0.4996562f, 0.0255074501f, 0.f),
    C_Vertex3f(0.4996562f, 0.0255074501f, -1.f),
    C_Normal3f(-0.854409218f, -0.519600809f, 0.f),
    C_Vertex3f(0.491197228f, 0.0377436876f, 0.f),
    C_Vertex3f(0.491197228f, 0.0377436876f, -1.f),
    C_Normal3f(-0.906562805f, -0.42207092f, 0.f),
    C_Vertex3f(0.483946681f, 0.0513395071f, 0.f),
    C_Vertex3f(0.483946681f, 0.0513395071f, -1.f),
    C_Normal3f(-0.944504619f, -0.328498125f, 0.f),
    C_Vertex3f(0.477904558f, 0.0662949085f, 0.f),
    C_Vertex3f(0.477904558f, 0.0662949085f, -1.f),
    C_Normal3f(-0.970400035f, -0.241503254f, 0.f),
    C_Vertex3f(0.47307086f, 0.0826098919f, 0.f),
    C_Vertex3f(0.47307086f, 0.0826098919f, -1.f),
    C_Normal3f(-0.986723185f, -0.162411109f, 0.f),
    C_Vertex3f(0.469445586f, 0.100284457f, 0.f),
    C_Vertex3f(0.469445586f, 0.100284457f, -1.f),
    C_Normal3f(-0.995799482f, -0.0915607437f, 0.f),
    C_Vertex3f(0.467028737f, 0.119318604f, 0.f),
    C_Vertex3f(0.467028737f, 0.119318604f, -1.f),
    C_Normal3f(-0.999541819f, -0.0302679744f, 0.f),
    C_Vertex3f(0.465820313f, 0.139712334f, 0.f),
    C_Vertex3f(0.465820313f, 0.139712334f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.111224085f, -0.993795276f, 0.f),
    C_Vertex3f(0.610321045f, 0.261230469f, 0.f),
    C_Vertex3f(0.610321045f, 0.261230469f, -1.f),
    C_Normal3f(0.111640364f, -0.993748665f, 0.f),
    C_Vertex3f(0.595624924f, 0.261230469f, 0.f),
    C_Vertex3f(0.595624924f, 0.261230469f, -1.f),
    C_Normal3f(0.346239418f, -0.938146234f, 0.f),
    C_Vertex3f(0.582099438f, 0.258059978f, 0.f),
    C_Vertex3f(0.582099438f, 0.258059978f, -1.f),
    C_Normal3f(0.563955128f, -0.825805366f, 0.f),
    C_Vertex3f(0.569853306f, 0.251718998f, 0.f),
    C_Vertex3f(0.569853306f, 0.251718998f, -1.f),
    C_Normal3f(0.721949756f, -0.691945434f, 0.f),
    C_Vertex3f(0.558886528f, 0.242207527f, 0.f),
    C_Vertex3f(0.558886528f, 0.242207527f, -1.f),
    C_Normal3f(0.830894351f, -0.556430101f, 0.f),
    C_Vertex3f(0.55162096f, 0.232696056f, 0.f),
    C_Vertex3f(0.55162096f, 0.232696056f, -1.f),
    C_Normal3f(0.904160261f, -0.427193344f, 0.f),
    C_Vertex3f(0.547382712f, 0.225029349f, 0.f),
    C_Vertex3f(0.547382712f, 0.225029349f, -1.f),
    C_Normal3f(0.945353687f, -0.326046616f, 0.f),
    C_Vertex3f(0.543749928f, 0.216036916f, 0.f),
    C_Vertex3f(0.543749928f, 0.216036916f, -1.f),
    C_Normal3f(0.970570743f, -0.240815938f, 0.f),
    C_Vertex3f(0.540722609f, 0.205718756f, 0.f),
    C_Vertex3f(0.540722609f, 0.205718756f, -1.f),
    C_Normal3f(0.98549664f, -0.169694647f, 0.f),
    C_Vertex3f(0.538300753f, 0.194074869f, 0.f),
    C_Vertex3f(0.538300753f, 0.194074869f, -1.f),
    C_Normal3f(0.993892193f, -0.110355213f, 0.f),
    C_Vertex3f(0.536484361f, 0.181105256f, 0.f),
    C_Vertex3f(0.536484361f, 0.181105256f, -1.f),
    C_Normal3f(0.998161852f, -0.0606039762f, 0.f),
    C_Vertex3f(0.535273433f, 0.166809916f, 0.f),
    C_Vertex3f(0.535273433f, 0.166809916f, -1.f),
    C_Normal3f(0.999818146f, -0.0190676469f, 0.f),
    C_Vertex3f(0.534667969f, 0.15118885f, 0.f),
    C_Vertex3f(0.534667969f, 0.15118885f, -1.f),
    C_Normal3f(0.999796093f, 0.0201941673f, 0.f),
    C_Vertex3f(0.534667969f, 0.135062218f, 0.f),
    C_Vertex3f(0.534667969f, 0.135062218f, -1.f),
    C_Normal3f(0.997736156f, 0.0672498718f, 0.f),
    C_Vertex3f(0.535280228f, 0.120876431f, 0.f),
    C_Vertex3f(0.535280228f, 0.120876431f, -1.f),
    C_Normal3f(0.99259156f, 0.121498376f, 0.f),
    C_Vertex3f(0.536504745f, 0.107811332f, 0.f),
    C_Vertex3f(0.536504745f, 0.107811332f, -1.f),
    C_Normal3f(0.98274076f, 0.184988424f, 0.f),
    C_Vertex3f(0.538341522f, 0.0958669186f, 0.f),
    C_Vertex3f(0.538341522f, 0.0958669186f, -1.f),
    C_Normal3f(0.965805709f, 0.259266764f, 0.f),
    C_Vertex3f(0.540790558f, 0.0850431919f, 0.f),
    C_Vertex3f(0.540790558f, 0.0850431919f, -1.f),
    C_Normal3f(0.938374341f, 0.345620573f, 0.f),
    C_Vertex3f(0.543851852f, 0.0753401518f, 0.f),
    C_Vertex3f(0.543851852f, 0.0753401518f, -1.f),
    C_Normal3f(0.895821929f, 0.444413185f, 0.f),
    C_Vertex3f(0.547525406f, 0.0667577982f, 0.f),
    C_Vertex3f(0.547525406f, 0.0667577982f, -1.f),
    C_Normal3f(0.824860394f, 0.565336525f, 0.f),
    C_Vertex3f(0.551811218f, 0.0592961311f, 0.f),
    C_Vertex3f(0.551811218f, 0.0592961311f, -1.f),
    C_Normal3f(0.720919788f, 0.693018377f, 0.f),
    C_Vertex3f(0.559158325f, 0.0497846603f, 0.f),
    C_Vertex3f(0.559158325f, 0.0497846603f, -1.f),
    C_Normal3f(0.56712091f, 0.823634624f, 0.f),
    C_Vertex3f(0.570097923f, 0.0402731895f, 0.f),
    C_Vertex3f(0.570097923f, 0.0402731895f, -1.f),
    C_Normal3f(0.351784319f, 0.936081052f, 0.f),
    C_Vertex3f(0.582180977f, 0.033932209f, 0.f),
    C_Vertex3f(0.582180977f, 0.033932209f, -1.f),
    C_Normal3f(0.112820596f, 0.993615389f, 0.f),
    C_Vertex3f(0.595407486f, 0.0307617188f, 0.f),
    C_Vertex3f(0.595407486f, 0.0307617188f, -1.f),
    C_Normal3f(-0.111006893f, 0.993819714f, 0.f),
    C_Vertex3f(0.610103607f, 0.0307617188f, 0.f),
    C_Vertex3f(0.610103607f, 0.0307617188f, -1.f),
    C_Normal3f(-0.343053162f, 0.939315915f, 0.f),
    C_Vertex3f(0.623792171f, 0.033932209f, 0.f),
    C_Vertex3f(0.623792171f, 0.033932209f, -1.f),
    C_Normal3f(-0.561263204f, 0.827637434f, 0.f),
    C_Vertex3f(0.636147022f, 0.0402731895f, 0.f),
    C_Vertex3f(0.636147022f, 0.0402731895f, -1.f),
    C_Normal3f(-0.720919788f, 0.693018377f, 0.f),
    C_Vertex3f(0.647168159f, 0.0497846603f, 0.f),
    C_Vertex3f(0.647168159f, 0.0497846603f, -1.f),
    C_Normal3f(-0.830690742f, 0.556734204f, 0.f),
    C_Vertex3f(0.654433727f, 0.0592961311f, 0.f),
    C_Vertex3f(0.654433727f, 0.0592961311f, -1.f),
    C_Normal3f(-0.903755248f, 0.428049564f, 0.f),
    C_Vertex3f(0.658671975f, 0.0669492483f, 0.f),
    C_Vertex3f(0.658671975f, 0.0669492483f, -1.f),
    C_Normal3f(-0.944998443f, 0.327074975f, 0.f),
    C_Vertex3f(0.662304759f, 0.0759145021f, 0.f),
    C_Vertex3f(0.662304759f, 0.0759145021f, -1.f),
    C_Normal3f(-0.97032553f, 0.241802454f, 0.f),
    C_Vertex3f(0.665332079f, 0.0861918926f, 0.f),
    C_Vertex3f(0.665332079f, 0.0861918926f, -1.f),
    C_Normal3f(-0.985354662f, 0.170517534f, 0.f),
    C_Vertex3f(0.667753935f, 0.0977814198f, 0.f),
    C_Vertex3f(0.667753935f, 0.0977814198f, -1.f),
    C_Normal3f(-0.993825257f, 0.110956132f, 0.f),
    C_Vertex3f(0.669570327f, 0.110683084f, 0.f),
    C_Vertex3f(0.669570327f, 0.110683084f, -1.f),
    C_Normal3f(-0.998140037f, 0.0609626621f, 0.f),
    C_Vertex3f(0.670781255f, 0.124896884f, 0.f),
    C_Vertex3f(0.670781255f, 0.124896884f, -1.f),
    C_Normal3f(-0.999817848f, 0.0190921575f, 0.f),
    C_Vertex3f(0.671386719f, 0.140422821f, 0.f),
    C_Vertex3f(0.671386719f, 0.140422821f, -1.f),
    C_Normal3f(-0.999799252f, -0.0200370271f, 0.f),
    C_Vertex3f(0.671386719f, 0.156603813f, 0.f),
    C_Vertex3f(0.671386719f, 0.156603813f, -1.f),
    C_Normal3f(-0.997790277f, -0.0664417297f, 0.f),
    C_Vertex3f(0.67077446f, 0.170973063f, 0.f),
    C_Vertex3f(0.67077446f, 0.170973063f, -1.f),
    C_Normal3f(-0.992744446f, -0.120242991f, 0.f),
    C_Vertex3f(0.669549942f, 0.184187651f, 0.f),
    C_Vertex3f(0.669549942f, 0.184187651f, -1.f),
    C_Normal3f(-0.983028173f, -0.183454767f, 0.f),
    C_Vertex3f(0.667713165f, 0.196247578f, 0.f),
    C_Vertex3f(0.667713165f, 0.196247578f, -1.f),
    C_Normal3f(-0.966210544f, -0.257754266f, 0.f),
    C_Vertex3f(0.66526413f, 0.207152843f, 0.f),
    C_Vertex3f(0.66526413f, 0.207152843f, -1.f),
    C_Normal3f(-0.938747525f, -0.344605535f, 0.f),
    C_Vertex3f(0.662202835f, 0.216903448f, 0.f),
    C_Vertex3f(0.662202835f, 0.216903448f, -1.f),
    C_Normal3f(-0.895747006f, -0.444564313f, 0.f),
    C_Vertex3f(0.658529282f, 0.225499392f, 0.f),
    C_Vertex3f(0.658529282f, 0.225499392f, -1.f),
    C_Normal3f(-0.823267579f, -0.567653596f, 0.f),
    C_Vertex3f(0.654243469f, 0.232940674f, 0.f),
    C_Vertex3f(0.654243469f, 0.232940674f, -1.f),
    C_Normal3f(-0.716898263f, -0.697177768f, 0.f),
    C_Vertex3f(0.646896362f, 0.242370605f, 0.f),
    C_Vertex3f(0.646896362f, 0.242370605f, -1.f),
    C_Normal3f(-0.561095059f, -0.827751338f, 0.f),
    C_Vertex3f(0.635902405f, 0.251800537f, 0.f),
    C_Vertex3f(0.635902405f, 0.251800537f, -1.f),
    C_Normal3f(-0.345873922f, -0.93828094f, 0.f),
    C_Vertex3f(0.623710632f, 0.258087158f, 0.f),
    C_Vertex3f(0.623710632f, 0.258087158f, -1.f),
    C_Normal3f(-0.111224085f, -0.993795276f, 0.f),
    C_Vertex3f(0.610321045f, 0.261230469f, 0.f),
    C_Vertex3f(0.610321045f, 0.261230469f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.2578125f, 0.492541313f, 0.f),
    C_Vertex3f(0.259840488f, 0.333099365f, 0.f),
    C_Vertex3f(0.260347724f, 0.631179094f, 0.f),
    C_Vertex3f(0.27101469f, 0.341339111f, 0.f),
    C_Vertex3f(0.271549344f, 0.623103857f, 0.f),
    C_Vertex3f(0.281607628f, 0.350952148f, 0.f),
    C_Vertex3f(0.282101154f, 0.613682747f, 0.f),
    C_Vertex3f(0.291619301f, 0.361938477f, 0.f),
    C_Vertex3f(0.292003155f, 0.602915764f, 0.f),
    C_Vertex3f(0.300379515f, 0.374092817f, 0.f),
    C_Vertex3f(0.300667405f, 0.590884686f, 0.f),
    C_Vertex3f(0.307888269f, 0.387415171f, 0.f),
    C_Vertex3f(0.308093905f, 0.577589512f, 0.f),
    C_Vertex3f(0.314145565f, 0.401905537f, 0.f),
    C_Vertex3f(0.314282656f, 0.563030243f, 0.f),
    C_Vertex3f(0.319151402f, 0.417563915f, 0.f),
    C_Vertex3f(0.319233656f, 0.547206879f, 0.f),
    C_Vertex3f(0.322905779f, 0.434390306f, 0.f),
    C_Vertex3f(0.322946906f, 0.530119419f, 0.f),
    C_Vertex3f(0.325408697f, 0.45238471f, 0.f),
    C_Vertex3f(0.325422406f, 0.511767864f, 0.f),
    C_Vertex3f(0.326660156f, 0.471547127f, 0.f),
    C_Vertex3f(0.326660156f, 0.492152214f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.260347724f, 0.631179094f, 0.f),
    C_Vertex3f(0.248496294f, 0.637908459f, 0.f),
    C_Vertex3f(0.248598218f, 0.552840948f, 0.f),
    C_Vertex3f(0.251669645f, 0.543090343f, 0.f),
    C_Vertex3f(0.254126787f, 0.532185078f, 0.f),
    C_Vertex3f(0.255969644f, 0.520125151f, 0.f),
    C_Vertex3f(0.257198215f, 0.506910563f, 0.f),
    C_Vertex3f(0.2578125f, 0.492541313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.209980488f, 0.594024658f, 0.f),
    C_Vertex3f(0.222211361f, 0.587738037f, 0.f),
    C_Vertex3f(0.222844005f, 0.647329569f, 0.f),
    C_Vertex3f(0.233241081f, 0.578308105f, 0.f),
    C_Vertex3f(0.235995054f, 0.64329195f, 0.f),
    C_Vertex3f(0.240612507f, 0.568878174f, 0.f),
    C_Vertex3f(0.248496294f, 0.637908459f, 0.f),
    C_Vertex3f(0.244912505f, 0.561436892f, 0.f),
    C_Vertex3f(0.248598218f, 0.552840948f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.138100147f, 0.568633556f, 0.f),
    C_Vertex3f(0.145388603f, 0.578145027f, 0.f),
    C_Vertex3f(0.148927689f, 0.646858215f, 0.f),
    C_Vertex3f(0.156335354f, 0.587656498f, 0.f),
    C_Vertex3f(0.163396835f, 0.649864197f, 0.f),
    C_Vertex3f(0.168510914f, 0.593997478f, 0.f),
    C_Vertex3f(0.178836823f, 0.651367188f, 0.f),
    C_Vertex3f(0.181915283f, 0.597167969f, 0.f),
    C_Vertex3f(0.194592476f, 0.651367188f, 0.f),
    C_Vertex3f(0.196548462f, 0.597167969f, 0.f),
    C_Vertex3f(0.209043145f, 0.650021315f, 0.f),
    C_Vertex3f(0.209980488f, 0.594024658f, 0.f),
    C_Vertex3f(0.222844005f, 0.647329569f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135429382f, 0.642349243f, 0.f),
    C_Vertex3f(0.122901917f, 0.63633728f, 0.f),
    C_Vertex3f(0.122915864f, 0.517042756f, 0.f),
    C_Vertex3f(0.124737978f, 0.530012369f, 0.f),
    C_Vertex3f(0.127167463f, 0.541656256f, 0.f),
    C_Vertex3f(0.13020432f, 0.551974416f, 0.f),
    C_Vertex3f(0.133848548f, 0.560966849f, 0.f),
    C_Vertex3f(0.138100147f, 0.568633556f, 0.f),
    C_Vertex3f(0.148927689f, 0.646858215f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.121708035f, 0.456671238f, 0.f),
    C_Vertex3f(0.12109375f, 0.470945358f, 0.f),
    C_Vertex3f(0.117014885f, 0.332791805f, 0.f),
    C_Vertex3f(0.12109375f, 0.48712635f, 0.f),
    C_Vertex3f(0.122901917f, 0.63633728f, 0.f),
    C_Vertex3f(0.121701121f, 0.502747416f, 0.f),
    C_Vertex3f(0.122915864f, 0.517042756f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0533266068f, 0.511627674f, 0.f),
    C_Vertex3f(0.0534565449f, 0.455113411f, 0.f),
    C_Vertex3f(0.0554876328f, 0.527505398f, 0.f),
    C_Vertex3f(0.0558774471f, 0.436011314f, 0.f),
    C_Vertex3f(0.0587291718f, 0.542740822f, 0.f),
    C_Vertex3f(0.0595088005f, 0.418289185f, 0.f),
    C_Vertex3f(0.0630512238f, 0.557333946f, 0.f),
    C_Vertex3f(0.064350605f, 0.401947021f, 0.f),
    C_Vertex3f(0.0684537888f, 0.571284771f, 0.f),
    C_Vertex3f(0.0704028606f, 0.386984825f, 0.f),
    C_Vertex3f(0.0749368668f, 0.584593296f, 0.f),
    C_Vertex3f(0.0776655674f, 0.373402596f, 0.f),
    C_Vertex3f(0.0825004578f, 0.597259521f, 0.f),
    C_Vertex3f(0.0861387253f, 0.361200333f, 0.f),
    C_Vertex3f(0.0911445618f, 0.609283447f, 0.f),
    C_Vertex3f(0.0958223343f, 0.350378036f, 0.f),
    C_Vertex3f(0.100759506f, 0.619804382f, 0.f),
    C_Vertex3f(0.106114388f, 0.340908527f, 0.f),
    C_Vertex3f(0.111345291f, 0.628822327f, 0.f),
    C_Vertex3f(0.117014885f, 0.332791805f, 0.f),
    C_Vertex3f(0.122901917f, 0.63633728f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0522460938f, 0.495107651f, 0.f),
    C_Vertex3f(0.0522460938f, 0.475595474f, 0.f),
    C_Vertex3f(0.0534565449f, 0.455113411f, 0.f),
    C_Vertex3f(0.0533266068f, 0.511627674f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259840488f, 0.333099365f, 0.f),
    C_Vertex3f(0.2578125f, 0.492541313f, 0.f),
    C_Vertex3f(0.2578125f, 0.476360321f, 0.f),
    C_Vertex3f(0.257205129f, 0.460834384f, 0.f),
    C_Vertex3f(0.255990386f, 0.446620584f, 0.f),
    C_Vertex3f(0.254168272f, 0.43371892f, 0.f),
    C_Vertex3f(0.251738787f, 0.422129393f, 0.f),
    C_Vertex3f(0.24870193f, 0.411852002f, 0.f),
    C_Vertex3f(0.248085022f, 0.32623291f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.24870193f, 0.411852002f, 0.f),
    C_Vertex3f(0.245057702f, 0.402886748f, 0.f),
    C_Vertex3f(0.248085022f, 0.32623291f, 0.f),
    C_Vertex3f(0.240806103f, 0.395233631f, 0.f),
    C_Vertex3f(0.235748291f, 0.320739746f, 0.f),
    C_Vertex3f(0.233517647f, 0.38572216f, 0.f),
    C_Vertex3f(0.222830296f, 0.316619873f, 0.f),
    C_Vertex3f(0.222515583f, 0.37621069f, 0.f),
    C_Vertex3f(0.210229397f, 0.369869709f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.222830296f, 0.316619873f, 0.f),
    C_Vertex3f(0.210229397f, 0.369869709f, 0.f),
    C_Vertex3f(0.209331036f, 0.313873291f, 0.f),
    C_Vertex3f(0.196659088f, 0.366699219f, 0.f),
    C_Vertex3f(0.195250511f, 0.3125f, 0.f),
    C_Vertex3f(0.182025909f, 0.366699219f, 0.f),
    C_Vertex3f(0.180644035f, 0.3125f, 0.f),
    C_Vertex3f(0.168759823f, 0.369842529f, 0.f),
    C_Vertex3f(0.166701317f, 0.313852787f, 0.f),
    C_Vertex3f(0.156639576f, 0.37612915f, 0.f),
    C_Vertex3f(0.153367043f, 0.316558361f, 0.f),
    C_Vertex3f(0.145665169f, 0.385559082f, 0.f),
    C_Vertex3f(0.140641212f, 0.320616722f, 0.f),
    C_Vertex3f(0.138293743f, 0.394989014f, 0.f),
    C_Vertex3f(0.133993745f, 0.402416706f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130308032f, 0.41098547f, 0.f),
    C_Vertex3f(0.127236605f, 0.420695305f, 0.f),
    C_Vertex3f(0.128523827f, 0.32602787f, 0.f),
    C_Vertex3f(0.140641212f, 0.320616722f, 0.f),
    C_Vertex3f(0.133993745f, 0.402416706f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.128523827f, 0.32602787f, 0.f),
    C_Vertex3f(0.127236605f, 0.420695305f, 0.f),
    C_Vertex3f(0.124779463f, 0.431546211f, 0.f),
    C_Vertex3f(0.122936606f, 0.443538189f, 0.f),
    C_Vertex3f(0.121708035f, 0.456671238f, 0.f),
    C_Vertex3f(0.117014885f, 0.332791805f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.539459229f, 0.651367188f, 0.f),
    C_Vertex3f(0.189674377f, -0.0234375f, 0.f),
    C_Vertex3f(0.254768372f, -0.0234375f, 0.f),
    C_Vertex3f(0.60345459f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.671386719f, 0.156603813f, 0.f),
    C_Vertex3f(0.673481822f, -0.00283813477f, 0.f),
    C_Vertex3f(0.674025416f, 0.295241594f, 0.f),
    C_Vertex3f(0.684665322f, 0.00540161133f, 0.f),
    C_Vertex3f(0.685208917f, 0.287166357f, 0.f),
    C_Vertex3f(0.695254803f, 0.0150146484f, 0.f),
    C_Vertex3f(0.695744038f, 0.277745247f, 0.f),
    C_Vertex3f(0.705250263f, 0.0260009766f, 0.f),
    C_Vertex3f(0.705630779f, 0.266978264f, 0.f),
    C_Vertex3f(0.713996291f, 0.0381553173f, 0.f),
    C_Vertex3f(0.714281678f, 0.254947186f, 0.f),
    C_Vertex3f(0.721492887f, 0.0514776707f, 0.f),
    C_Vertex3f(0.721696734f, 0.241652012f, 0.f),
    C_Vertex3f(0.727740049f, 0.0659680367f, 0.f),
    C_Vertex3f(0.727875948f, 0.227092743f, 0.f),
    C_Vertex3f(0.73273778f, 0.0816264153f, 0.f),
    C_Vertex3f(0.732819319f, 0.211269379f, 0.f),
    C_Vertex3f(0.736486077f, 0.0984528065f, 0.f),
    C_Vertex3f(0.736526847f, 0.194181919f, 0.f),
    C_Vertex3f(0.738984942f, 0.11644721f, 0.f),
    C_Vertex3f(0.738998532f, 0.175830364f, 0.f),
    C_Vertex3f(0.740234375f, 0.135609627f, 0.f),
    C_Vertex3f(0.740234375f, 0.156214714f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.674025416f, 0.295241594f, 0.f),
    C_Vertex3f(0.662193537f, 0.301970959f, 0.f),
    C_Vertex3f(0.662202835f, 0.216903448f, 0.f),
    C_Vertex3f(0.66526413f, 0.207152843f, 0.f),
    C_Vertex3f(0.667713165f, 0.196247578f, 0.f),
    C_Vertex3f(0.669549942f, 0.184187651f, 0.f),
    C_Vertex3f(0.67077446f, 0.170973063f, 0.f),
    C_Vertex3f(0.671386719f, 0.156603813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.623710632f, 0.258087158f, 0.f),
    C_Vertex3f(0.635902405f, 0.251800537f, 0.f),
    C_Vertex3f(0.63658464f, 0.311392069f, 0.f),
    C_Vertex3f(0.646896362f, 0.242370605f, 0.f),
    C_Vertex3f(0.649713278f, 0.30735445f, 0.f),
    C_Vertex3f(0.654243469f, 0.232940674f, 0.f),
    C_Vertex3f(0.662193537f, 0.301970959f, 0.f),
    C_Vertex3f(0.658529282f, 0.225499392f, 0.f),
    C_Vertex3f(0.662202835f, 0.216903448f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.55162096f, 0.232696056f, 0.f),
    C_Vertex3f(0.558886528f, 0.242207527f, 0.f),
    C_Vertex3f(0.562553763f, 0.3109411f, 0.f),
    C_Vertex3f(0.569853306f, 0.251718998f, 0.f),
    C_Vertex3f(0.577083707f, 0.313933492f, 0.f),
    C_Vertex3f(0.582099438f, 0.258059978f, 0.f),
    C_Vertex3f(0.592597008f, 0.315429688f, 0.f),
    C_Vertex3f(0.595624924f, 0.261230469f, 0.f),
    C_Vertex3f(0.608382225f, 0.315429688f, 0.f),
    C_Vertex3f(0.610321045f, 0.261230469f, 0.f),
    C_Vertex3f(0.622807622f, 0.314083815f, 0.f),
    C_Vertex3f(0.623710632f, 0.258087158f, 0.f),
    C_Vertex3f(0.63658464f, 0.311392069f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.549007177f, 0.306452513f, 0.f),
    C_Vertex3f(0.536443949f, 0.30046773f, 0.f),
    C_Vertex3f(0.536484361f, 0.181105256f, 0.f),
    C_Vertex3f(0.538300753f, 0.194074869f, 0.f),
    C_Vertex3f(0.540722609f, 0.205718756f, 0.f),
    C_Vertex3f(0.543749928f, 0.216036916f, 0.f),
    C_Vertex3f(0.547382712f, 0.225029349f, 0.f),
    C_Vertex3f(0.55162096f, 0.232696056f, 0.f),
    C_Vertex3f(0.562553763f, 0.3109411f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.535280228f, 0.120876431f, 0.f),
    C_Vertex3f(0.534667969f, 0.135062218f, 0.f),
    C_Vertex3f(0.530481935f, -0.00304377079f, 0.f),
    C_Vertex3f(0.534667969f, 0.15118885f, 0.f),
    C_Vertex3f(0.536443949f, 0.30046773f, 0.f),
    C_Vertex3f(0.535273433f, 0.166809916f, 0.f),
    C_Vertex3f(0.536484361f, 0.181105256f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466899037f, 0.175832868f, 0.f),
    C_Vertex3f(0.467028737f, 0.119318604f, 0.f),
    C_Vertex3f(0.469056487f, 0.191778541f, 0.f),
    C_Vertex3f(0.469445586f, 0.100284457f, 0.f),
    C_Vertex3f(0.472292662f, 0.207061529f, 0.f),
    C_Vertex3f(0.47307086f, 0.0826098919f, 0.f),
    C_Vertex3f(0.476607561f, 0.221681833f, 0.f),
    C_Vertex3f(0.477904558f, 0.0662949085f, 0.f),
    C_Vertex3f(0.482001185f, 0.235639453f, 0.f),
    C_Vertex3f(0.483946681f, 0.0513395071f, 0.f),
    C_Vertex3f(0.488473535f, 0.248934388f, 0.f),
    C_Vertex3f(0.491197228f, 0.0377436876f, 0.f),
    C_Vertex3f(0.496024609f, 0.261566639f, 0.f),
    C_Vertex3f(0.4996562f, 0.0255074501f, 0.f),
    C_Vertex3f(0.504654408f, 0.273536205f, 0.f),
    C_Vertex3f(0.509323597f, 0.0146307945f, 0.f),
    C_Vertex3f(0.514267564f, 0.284009576f, 0.f),
    C_Vertex3f(0.519598842f, 0.00511372089f, 0.f),
    C_Vertex3f(0.524864078f, 0.292986751f, 0.f),
    C_Vertex3f(0.530481935f, -0.00304377079f, 0.f),
    C_Vertex3f(0.536443949f, 0.30046773f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.465820313f, 0.15922451f, 0.f),
    C_Vertex3f(0.465820313f, 0.139712334f, 0.f),
    C_Vertex3f(0.467028737f, 0.119318604f, 0.f),
    C_Vertex3f(0.466899037f, 0.175832868f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.673481822f, -0.00283813477f, 0.f),
    C_Vertex3f(0.671386719f, 0.156603813f, 0.f),
    C_Vertex3f(0.671386719f, 0.140422821f, 0.f),
    C_Vertex3f(0.670781255f, 0.124896884f, 0.f),
    C_Vertex3f(0.669570327f, 0.110683084f, 0.f),
    C_Vertex3f(0.667753935f, 0.0977814198f, 0.f),
    C_Vertex3f(0.665332079f, 0.0861918926f, 0.f),
    C_Vertex3f(0.662304759f, 0.0759145021f, 0.f),
    C_Vertex3f(0.661704302f, -0.00970458984f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.662304759f, 0.0759145021f, 0.f),
    C_Vertex3f(0.658671975f, 0.0669492483f, 0.f),
    C_Vertex3f(0.661704302f, -0.00970458984f, 0.f),
    C_Vertex3f(0.654433727f, 0.0592961311f, 0.f),
    C_Vertex3f(0.649332762f, -0.0151977539f, 0.f),
    C_Vertex3f(0.647168159f, 0.0497846603f, 0.f),
    C_Vertex3f(0.636367202f, -0.019317627f, 0.f),
    C_Vertex3f(0.636147022f, 0.0402731895f, 0.f),
    C_Vertex3f(0.623792171f, 0.033932209f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.636367202f, -0.019317627f, 0.f),
    C_Vertex3f(0.623792171f, 0.033932209f, 0.f),
    C_Vertex3f(0.622807622f, -0.022064209f, 0.f),
    C_Vertex3f(0.610103607f, 0.0307617188f, 0.f),
    C_Vertex3f(0.608654022f, -0.0234375f, 0.f),
    C_Vertex3f(0.595407486f, 0.0307617188f, 0.f),
    C_Vertex3f(0.594015121f, -0.0234375f, 0.f),
    C_Vertex3f(0.582180977f, 0.033932209f, 0.f),
    C_Vertex3f(0.580092788f, -0.0220779181f, 0.f),
    C_Vertex3f(0.570097923f, 0.0402731895f, 0.f),
    C_Vertex3f(0.566778302f, -0.0193587542f, 0.f),
    C_Vertex3f(0.559158325f, 0.0497846603f, 0.f),
    C_Vertex3f(0.554071665f, -0.0152800083f, 0.f),
    C_Vertex3f(0.551811218f, 0.0592961311f, 0.f),
    C_Vertex3f(0.547525406f, 0.0667577982f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.543851852f, 0.0753401518f, 0.f),
    C_Vertex3f(0.540790558f, 0.0850431919f, 0.f),
    C_Vertex3f(0.541972876f, -0.00984168053f, 0.f),
    C_Vertex3f(0.554071665f, -0.0152800083f, 0.f),
    C_Vertex3f(0.547525406f, 0.0667577982f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.541972876f, -0.00984168053f, 0.f),
    C_Vertex3f(0.540790558f, 0.0850431919f, 0.f),
    C_Vertex3f(0.538341522f, 0.0958669186f, 0.f),
    C_Vertex3f(0.536504745f, 0.107811332f, 0.f),
    C_Vertex3f(0.535280228f, 0.120876431f, 0.f),
    C_Vertex3f(0.530481935f, -0.00304377079f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.2578125f, 0.492541313f, -1.f),
    C_Vertex3f(0.259840488f, 0.333099365f, -1.f),
    C_Vertex3f(0.260347724f, 0.631179094f, -1.f),
    C_Vertex3f(0.27101469f, 0.341339111f, -1.f),
    C_Vertex3f(0.271549344f, 0.623103857f, -1.f),
    C_Vertex3f(0.281607628f, 0.350952148f, -1.f),
    C_Vertex3f(0.282101154f, 0.613682747f, -1.f),
    C_Vertex3f(0.291619301f, 0.361938477f, -1.f),
    C_Vertex3f(0.292003155f, 0.602915764f, -1.f),
    C_Vertex3f(0.300379515f, 0.374092817f, -1.f),
    C_Vertex3f(0.300667405f, 0.590884686f, -1.f),
    C_Vertex3f(0.307888269f, 0.387415171f, -1.f),
    C_Vertex3f(0.308093905f, 0.577589512f, -1.f),
    C_Vertex3f(0.314145565f, 0.401905537f, -1.f),
    C_Vertex3f(0.314282656f, 0.563030243f, -1.f),
    C_Vertex3f(0.319151402f, 0.417563915f, -1.f),
    C_Vertex3f(0.319233656f, 0.547206879f, -1.f),
    C_Vertex3f(0.322905779f, 0.434390306f, -1.f),
    C_Vertex3f(0.322946906f, 0.530119419f, -1.f),
    C_Vertex3f(0.325408697f, 0.45238471f, -1.f),
    C_Vertex3f(0.325422406f, 0.511767864f, -1.f),
    C_Vertex3f(0.326660156f, 0.471547127f, -1.f),
    C_Vertex3f(0.326660156f, 0.492152214f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.260347724f, 0.631179094f, -1.f),
    C_Vertex3f(0.248496294f, 0.637908459f, -1.f),
    C_Vertex3f(0.248598218f, 0.552840948f, -1.f),
    C_Vertex3f(0.251669645f, 0.543090343f, -1.f),
    C_Vertex3f(0.254126787f, 0.532185078f, -1.f),
    C_Vertex3f(0.255969644f, 0.520125151f, -1.f),
    C_Vertex3f(0.257198215f, 0.506910563f, -1.f),
    C_Vertex3f(0.2578125f, 0.492541313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.209980488f, 0.594024658f, -1.f),
    C_Vertex3f(0.222211361f, 0.587738037f, -1.f),
    C_Vertex3f(0.222844005f, 0.647329569f, -1.f),
    C_Vertex3f(0.233241081f, 0.578308105f, -1.f),
    C_Vertex3f(0.235995054f, 0.64329195f, -1.f),
    C_Vertex3f(0.240612507f, 0.568878174f, -1.f),
    C_Vertex3f(0.248496294f, 0.637908459f, -1.f),
    C_Vertex3f(0.244912505f, 0.561436892f, -1.f),
    C_Vertex3f(0.248598218f, 0.552840948f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.138100147f, 0.568633556f, -1.f),
    C_Vertex3f(0.145388603f, 0.578145027f, -1.f),
    C_Vertex3f(0.148927689f, 0.646858215f, -1.f),
    C_Vertex3f(0.156335354f, 0.587656498f, -1.f),
    C_Vertex3f(0.163396835f, 0.649864197f, -1.f),
    C_Vertex3f(0.168510914f, 0.593997478f, -1.f),
    C_Vertex3f(0.178836823f, 0.651367188f, -1.f),
    C_Vertex3f(0.181915283f, 0.597167969f, -1.f),
    C_Vertex3f(0.194592476f, 0.651367188f, -1.f),
    C_Vertex3f(0.196548462f, 0.597167969f, -1.f),
    C_Vertex3f(0.209043145f, 0.650021315f, -1.f),
    C_Vertex3f(0.209980488f, 0.594024658f, -1.f),
    C_Vertex3f(0.222844005f, 0.647329569f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135429382f, 0.642349243f, -1.f),
    C_Vertex3f(0.122901917f, 0.63633728f, -1.f),
    C_Vertex3f(0.122915864f, 0.517042756f, -1.f),
    C_Vertex3f(0.124737978f, 0.530012369f, -1.f),
    C_Vertex3f(0.127167463f, 0.541656256f, -1.f),
    C_Vertex3f(0.13020432f, 0.551974416f, -1.f),
    C_Vertex3f(0.133848548f, 0.560966849f, -1.f),
    C_Vertex3f(0.138100147f, 0.568633556f, -1.f),
    C_Vertex3f(0.148927689f, 0.646858215f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.121708035f, 0.456671238f, -1.f),
    C_Vertex3f(0.12109375f, 0.470945358f, -1.f),
    C_Vertex3f(0.117014885f, 0.332791805f, -1.f),
    C_Vertex3f(0.12109375f, 0.48712635f, -1.f),
    C_Vertex3f(0.122901917f, 0.63633728f, -1.f),
    C_Vertex3f(0.121701121f, 0.502747416f, -1.f),
    C_Vertex3f(0.122915864f, 0.517042756f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0533266068f, 0.511627674f, -1.f),
    C_Vertex3f(0.0534565449f, 0.455113411f, -1.f),
    C_Vertex3f(0.0554876328f, 0.527505398f, -1.f),
    C_Vertex3f(0.0558774471f, 0.436011314f, -1.f),
    C_Vertex3f(0.0587291718f, 0.542740822f, -1.f),
    C_Vertex3f(0.0595088005f, 0.418289185f, -1.f),
    C_Vertex3f(0.0630512238f, 0.557333946f, -1.f),
    C_Vertex3f(0.064350605f, 0.401947021f, -1.f),
    C_Vertex3f(0.0684537888f, 0.571284771f, -1.f),
    C_Vertex3f(0.0704028606f, 0.386984825f, -1.f),
    C_Vertex3f(0.0749368668f, 0.584593296f, -1.f),
    C_Vertex3f(0.0776655674f, 0.373402596f, -1.f),
    C_Vertex3f(0.0825004578f, 0.597259521f, -1.f),
    C_Vertex3f(0.0861387253f, 0.361200333f, -1.f),
    C_Vertex3f(0.0911445618f, 0.609283447f, -1.f),
    C_Vertex3f(0.0958223343f, 0.350378036f, -1.f),
    C_Vertex3f(0.100759506f, 0.619804382f, -1.f),
    C_Vertex3f(0.106114388f, 0.340908527f, -1.f),
    C_Vertex3f(0.111345291f, 0.628822327f, -1.f),
    C_Vertex3f(0.117014885f, 0.332791805f, -1.f),
    C_Vertex3f(0.122901917f, 0.63633728f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0522460938f, 0.495107651f, -1.f),
    C_Vertex3f(0.0522460938f, 0.475595474f, -1.f),
    C_Vertex3f(0.0534565449f, 0.455113411f, -1.f),
    C_Vertex3f(0.0533266068f, 0.511627674f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259840488f, 0.333099365f, -1.f),
    C_Vertex3f(0.2578125f, 0.492541313f, -1.f),
    C_Vertex3f(0.2578125f, 0.476360321f, -1.f),
    C_Vertex3f(0.257205129f, 0.460834384f, -1.f),
    C_Vertex3f(0.255990386f, 0.446620584f, -1.f),
    C_Vertex3f(0.254168272f, 0.43371892f, -1.f),
    C_Vertex3f(0.251738787f, 0.422129393f, -1.f),
    C_Vertex3f(0.24870193f, 0.411852002f, -1.f),
    C_Vertex3f(0.248085022f, 0.32623291f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.24870193f, 0.411852002f, -1.f),
    C_Vertex3f(0.245057702f, 0.402886748f, -1.f),
    C_Vertex3f(0.248085022f, 0.32623291f, -1.f),
    C_Vertex3f(0.240806103f, 0.395233631f, -1.f),
    C_Vertex3f(0.235748291f, 0.320739746f, -1.f),
    C_Vertex3f(0.233517647f, 0.38572216f, -1.f),
    C_Vertex3f(0.222830296f, 0.316619873f, -1.f),
    C_Vertex3f(0.222515583f, 0.37621069f, -1.f),
    C_Vertex3f(0.210229397f, 0.369869709f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.222830296f, 0.316619873f, -1.f),
    C_Vertex3f(0.210229397f, 0.369869709f, -1.f),
    C_Vertex3f(0.209331036f, 0.313873291f, -1.f),
    C_Vertex3f(0.196659088f, 0.366699219f, -1.f),
    C_Vertex3f(0.195250511f, 0.3125f, -1.f),
    C_Vertex3f(0.182025909f, 0.366699219f, -1.f),
    C_Vertex3f(0.180644035f, 0.3125f, -1.f),
    C_Vertex3f(0.168759823f, 0.369842529f, -1.f),
    C_Vertex3f(0.166701317f, 0.313852787f, -1.f),
    C_Vertex3f(0.156639576f, 0.37612915f, -1.f),
    C_Vertex3f(0.153367043f, 0.316558361f, -1.f),
    C_Vertex3f(0.145665169f, 0.385559082f, -1.f),
    C_Vertex3f(0.140641212f, 0.320616722f, -1.f),
    C_Vertex3f(0.138293743f, 0.394989014f, -1.f),
    C_Vertex3f(0.133993745f, 0.402416706f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130308032f, 0.41098547f, -1.f),
    C_Vertex3f(0.127236605f, 0.420695305f, -1.f),
    C_Vertex3f(0.128523827f, 0.32602787f, -1.f),
    C_Vertex3f(0.140641212f, 0.320616722f, -1.f),
    C_Vertex3f(0.133993745f, 0.402416706f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.128523827f, 0.32602787f, -1.f),
    C_Vertex3f(0.127236605f, 0.420695305f, -1.f),
    C_Vertex3f(0.124779463f, 0.431546211f, -1.f),
    C_Vertex3f(0.122936606f, 0.443538189f, -1.f),
    C_Vertex3f(0.121708035f, 0.456671238f, -1.f),
    C_Vertex3f(0.117014885f, 0.332791805f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.539459229f, 0.651367188f, -1.f),
    C_Vertex3f(0.189674377f, -0.0234375f, -1.f),
    C_Vertex3f(0.254768372f, -0.0234375f, -1.f),
    C_Vertex3f(0.60345459f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.671386719f, 0.156603813f, -1.f),
    C_Vertex3f(0.673481822f, -0.00283813477f, -1.f),
    C_Vertex3f(0.674025416f, 0.295241594f, -1.f),
    C_Vertex3f(0.684665322f, 0.00540161133f, -1.f),
    C_Vertex3f(0.685208917f, 0.287166357f, -1.f),
    C_Vertex3f(0.695254803f, 0.0150146484f, -1.f),
    C_Vertex3f(0.695744038f, 0.277745247f, -1.f),
    C_Vertex3f(0.705250263f, 0.0260009766f, -1.f),
    C_Vertex3f(0.705630779f, 0.266978264f, -1.f),
    C_Vertex3f(0.713996291f, 0.0381553173f, -1.f),
    C_Vertex3f(0.714281678f, 0.254947186f, -1.f),
    C_Vertex3f(0.721492887f, 0.0514776707f, -1.f),
    C_Vertex3f(0.721696734f, 0.241652012f, -1.f),
    C_Vertex3f(0.727740049f, 0.0659680367f, -1.f),
    C_Vertex3f(0.727875948f, 0.227092743f, -1.f),
    C_Vertex3f(0.73273778f, 0.0816264153f, -1.f),
    C_Vertex3f(0.732819319f, 0.211269379f, -1.f),
    C_Vertex3f(0.736486077f, 0.0984528065f, -1.f),
    C_Vertex3f(0.736526847f, 0.194181919f, -1.f),
    C_Vertex3f(0.738984942f, 0.11644721f, -1.f),
    C_Vertex3f(0.738998532f, 0.175830364f, -1.f),
    C_Vertex3f(0.740234375f, 0.135609627f, -1.f),
    C_Vertex3f(0.740234375f, 0.156214714f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.674025416f, 0.295241594f, -1.f),
    C_Vertex3f(0.662193537f, 0.301970959f, -1.f),
    C_Vertex3f(0.662202835f, 0.216903448f, -1.f),
    C_Vertex3f(0.66526413f, 0.207152843f, -1.f),
    C_Vertex3f(0.667713165f, 0.196247578f, -1.f),
    C_Vertex3f(0.669549942f, 0.184187651f, -1.f),
    C_Vertex3f(0.67077446f, 0.170973063f, -1.f),
    C_Vertex3f(0.671386719f, 0.156603813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.623710632f, 0.258087158f, -1.f),
    C_Vertex3f(0.635902405f, 0.251800537f, -1.f),
    C_Vertex3f(0.63658464f, 0.311392069f, -1.f),
    C_Vertex3f(0.646896362f, 0.242370605f, -1.f),
    C_Vertex3f(0.649713278f, 0.30735445f, -1.f),
    C_Vertex3f(0.654243469f, 0.232940674f, -1.f),
    C_Vertex3f(0.662193537f, 0.301970959f, -1.f),
    C_Vertex3f(0.658529282f, 0.225499392f, -1.f),
    C_Vertex3f(0.662202835f, 0.216903448f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.55162096f, 0.232696056f, -1.f),
    C_Vertex3f(0.558886528f, 0.242207527f, -1.f),
    C_Vertex3f(0.562553763f, 0.3109411f, -1.f),
    C_Vertex3f(0.569853306f, 0.251718998f, -1.f),
    C_Vertex3f(0.577083707f, 0.313933492f, -1.f),
    C_Vertex3f(0.582099438f, 0.258059978f, -1.f),
    C_Vertex3f(0.592597008f, 0.315429688f, -1.f),
    C_Vertex3f(0.595624924f, 0.261230469f, -1.f),
    C_Vertex3f(0.608382225f, 0.315429688f, -1.f),
    C_Vertex3f(0.610321045f, 0.261230469f, -1.f),
    C_Vertex3f(0.622807622f, 0.314083815f, -1.f),
    C_Vertex3f(0.623710632f, 0.258087158f, -1.f),
    C_Vertex3f(0.63658464f, 0.311392069f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.549007177f, 0.306452513f, -1.f),
    C_Vertex3f(0.536443949f, 0.30046773f, -1.f),
    C_Vertex3f(0.536484361f, 0.181105256f, -1.f),
    C_Vertex3f(0.538300753f, 0.194074869f, -1.f),
    C_Vertex3f(0.540722609f, 0.205718756f, -1.f),
    C_Vertex3f(0.543749928f, 0.216036916f, -1.f),
    C_Vertex3f(0.547382712f, 0.225029349f, -1.f),
    C_Vertex3f(0.55162096f, 0.232696056f, -1.f),
    C_Vertex3f(0.562553763f, 0.3109411f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.535280228f, 0.120876431f, -1.f),
    C_Vertex3f(0.534667969f, 0.135062218f, -1.f),
    C_Vertex3f(0.530481935f, -0.00304377079f, -1.f),
    C_Vertex3f(0.534667969f, 0.15118885f, -1.f),
    C_Vertex3f(0.536443949f, 0.30046773f, -1.f),
    C_Vertex3f(0.535273433f, 0.166809916f, -1.f),
    C_Vertex3f(0.536484361f, 0.181105256f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466899037f, 0.175832868f, -1.f),
    C_Vertex3f(0.467028737f, 0.119318604f, -1.f),
    C_Vertex3f(0.469056487f, 0.191778541f, -1.f),
    C_Vertex3f(0.469445586f, 0.100284457f, -1.f),
    C_Vertex3f(0.472292662f, 0.207061529f, -1.f),
    C_Vertex3f(0.47307086f, 0.0826098919f, -1.f),
    C_Vertex3f(0.476607561f, 0.221681833f, -1.f),
    C_Vertex3f(0.477904558f, 0.0662949085f, -1.f),
    C_Vertex3f(0.482001185f, 0.235639453f, -1.f),
    C_Vertex3f(0.483946681f, 0.0513395071f, -1.f),
    C_Vertex3f(0.488473535f, 0.248934388f, -1.f),
    C_Vertex3f(0.491197228f, 0.0377436876f, -1.f),
    C_Vertex3f(0.496024609f, 0.261566639f, -1.f),
    C_Vertex3f(0.4996562f, 0.0255074501f, -1.f),
    C_Vertex3f(0.504654408f, 0.273536205f, -1.f),
    C_Vertex3f(0.509323597f, 0.0146307945f, -1.f),
    C_Vertex3f(0.514267564f, 0.284009576f, -1.f),
    C_Vertex3f(0.519598842f, 0.00511372089f, -1.f),
    C_Vertex3f(0.524864078f, 0.292986751f, -1.f),
    C_Vertex3f(0.530481935f, -0.00304377079f, -1.f),
    C_Vertex3f(0.536443949f, 0.30046773f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.465820313f, 0.15922451f, -1.f),
    C_Vertex3f(0.465820313f, 0.139712334f, -1.f),
    C_Vertex3f(0.467028737f, 0.119318604f, -1.f),
    C_Vertex3f(0.466899037f, 0.175832868f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.673481822f, -0.00283813477f, -1.f),
    C_Vertex3f(0.671386719f, 0.156603813f, -1.f),
    C_Vertex3f(0.671386719f, 0.140422821f, -1.f),
    C_Vertex3f(0.670781255f, 0.124896884f, -1.f),
    C_Vertex3f(0.669570327f, 0.110683084f, -1.f),
    C_Vertex3f(0.667753935f, 0.0977814198f, -1.f),
    C_Vertex3f(0.665332079f, 0.0861918926f, -1.f),
    C_Vertex3f(0.662304759f, 0.0759145021f, -1.f),
    C_Vertex3f(0.661704302f, -0.00970458984f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.662304759f, 0.0759145021f, -1.f),
    C_Vertex3f(0.658671975f, 0.0669492483f, -1.f),
    C_Vertex3f(0.661704302f, -0.00970458984f, -1.f),
    C_Vertex3f(0.654433727f, 0.0592961311f, -1.f),
    C_Vertex3f(0.649332762f, -0.0151977539f, -1.f),
    C_Vertex3f(0.647168159f, 0.0497846603f, -1.f),
    C_Vertex3f(0.636367202f, -0.019317627f, -1.f),
    C_Vertex3f(0.636147022f, 0.0402731895f, -1.f),
    C_Vertex3f(0.623792171f, 0.033932209f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.636367202f, -0.019317627f, -1.f),
    C_Vertex3f(0.623792171f, 0.033932209f, -1.f),
    C_Vertex3f(0.622807622f, -0.022064209f, -1.f),
    C_Vertex3f(0.610103607f, 0.0307617188f, -1.f),
    C_Vertex3f(0.608654022f, -0.0234375f, -1.f),
    C_Vertex3f(0.595407486f, 0.0307617188f, -1.f),
    C_Vertex3f(0.594015121f, -0.0234375f, -1.f),
    C_Vertex3f(0.582180977f, 0.033932209f, -1.f),
    C_Vertex3f(0.580092788f, -0.0220779181f, -1.f),
    C_Vertex3f(0.570097923f, 0.0402731895f, -1.f),
    C_Vertex3f(0.566778302f, -0.0193587542f, -1.f),
    C_Vertex3f(0.559158325f, 0.0497846603f, -1.f),
    C_Vertex3f(0.554071665f, -0.0152800083f, -1.f),
    C_Vertex3f(0.551811218f, 0.0592961311f, -1.f),
    C_Vertex3f(0.547525406f, 0.0667577982f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.543851852f, 0.0753401518f, -1.f),
    C_Vertex3f(0.540790558f, 0.0850431919f, -1.f),
    C_Vertex3f(0.541972876f, -0.00984168053f, -1.f),
    C_Vertex3f(0.554071665f, -0.0152800083f, -1.f),
    C_Vertex3f(0.547525406f, 0.0667577982f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.541972876f, -0.00984168053f, -1.f),
    C_Vertex3f(0.540790558f, 0.0850431919f, -1.f),
    C_Vertex3f(0.538341522f, 0.0958669186f, -1.f),
    C_Vertex3f(0.536504745f, 0.107811332f, -1.f),
    C_Vertex3f(0.535280228f, 0.120876431f, -1.f),
    C_Vertex3f(0.530481935f, -0.00304377079f, -1.f),
    C_End(),
    C_Translatef(0.795898438f, 0.f, 0.f),
};
static const Call char_38[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.742816985f, -0.669494569f, 0.f),
    C_Vertex3f(0.425132751f, 0.0755157471f, 0.f),
    C_Vertex3f(0.425132751f, 0.0755157471f, -1.f),
    C_Normal3f(0.702629447f, -0.711555958f, 0.f),
    C_Vertex3f(0.41541481f, 0.0647335052f, 0.f),
    C_Vertex3f(0.41541481f, 0.0647335052f, -1.f),
    C_Normal3f(0.645213008f, -0.764002681f, 0.f),
    C_Vertex3f(0.395104885f, 0.0458645821f, 0.f),
    C_Vertex3f(0.395104885f, 0.0458645821f, -1.f),
    C_Normal3f(0.565522134f, -0.824733019f, 0.f),
    C_Vertex3f(0.373920918f, 0.0296912193f, 0.f),
    C_Vertex3f(0.373920918f, 0.0296912193f, -1.f),
    C_Normal3f(0.474624872f, -0.880188107f, 0.f),
    C_Vertex3f(0.351862907f, 0.0162134171f, 0.f),
    C_Vertex3f(0.351862907f, 0.0162134171f, -1.f),
    C_Normal3f(0.373984993f, -0.927434802f, 0.f),
    C_Vertex3f(0.328930855f, 0.00543117523f, 0.f),
    C_Vertex3f(0.328930855f, 0.00543117523f, -1.f),
    C_Normal3f(0.266983688f, -0.963701069f, 0.f),
    C_Vertex3f(0.3050704f, -0.00265550613f, 0.f),
    C_Vertex3f(0.3050704f, -0.00265550613f, -1.f),
    C_Normal3f(0.158099204f, -0.987423241f, 0.f),
    C_Vertex3f(0.280281544f, -0.00804662704f, 0.f),
    C_Vertex3f(0.280281544f, -0.00804662704f, -1.f),
    C_Normal3f(0.0524478033f, -0.998623669f, 0.f),
    C_Vertex3f(0.254564285f, -0.0107421875f, 0.f),
    C_Vertex3f(0.254564285f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0423890352f, -0.999101222f, 0.f),
    C_Vertex3f(0.228957176f, -0.0107421875f, 0.f),
    C_Vertex3f(0.228957176f, -0.0107421875f, -1.f),
    C_Normal3f(-0.136049315f, -0.990702093f, 0.f),
    C_Vertex3f(0.205657959f, -0.00866723061f, 0.f),
    C_Vertex3f(0.205657959f, -0.00866723061f, -1.f),
    C_Normal3f(-0.235628992f, -0.971843064f, 0.f),
    C_Vertex3f(0.183628082f, -0.00451731682f, 0.f),
    C_Vertex3f(0.183628082f, -0.00451731682f, -1.f),
    C_Normal3f(-0.339424312f, -0.940633416f, 0.f),
    C_Vertex3f(0.162867546f, 0.00170755386f, 0.f),
    C_Vertex3f(0.162867546f, 0.00170755386f, -1.f),
    C_Normal3f(-0.44375211f, -0.896149576f, 0.f),
    C_Vertex3f(0.14337635f, 0.0100073814f, 0.f),
    C_Vertex3f(0.14337635f, 0.0100073814f, -1.f),
    C_Normal3f(-0.544337451f, -0.838866353f, 0.f),
    C_Vertex3f(0.125154495f, 0.0203821659f, 0.f),
    C_Vertex3f(0.125154495f, 0.0203821659f, -1.f),
    C_Normal3f(-0.637085319f, -0.770793319f, 0.f),
    C_Vertex3f(0.108201981f, 0.0328319073f, 0.f),
    C_Vertex3f(0.108201981f, 0.0328319073f, -1.f),
    C_Normal3f(-0.717482865f, -0.696576118f, 0.f),
    C_Vertex3f(0.0925188065f, 0.0473566055f, 0.f),
    C_Vertex3f(0.0925188065f, 0.0473566055f, -1.f),
    C_Normal3f(-0.77997607f, -0.625809312f, 0.f),
    C_Vertex3f(0.0794696808f, 0.0624265671f, 0.f),
    C_Vertex3f(0.0794696808f, 0.0624265671f, -1.f),
    C_Normal3f(-0.830128014f, -0.557572782f, 0.f),
    C_Vertex3f(0.0692458153f, 0.0763628483f, 0.f),
    C_Vertex3f(0.0692458153f, 0.0763628483f, -1.f),
    C_Normal3f(-0.875161648f, -0.48383072f, 0.f),
    C_Vertex3f(0.060482502f, 0.0906951427f, 0.f),
    C_Vertex3f(0.060482502f, 0.0906951427f, -1.f),
    C_Normal3f(-0.915204108f, -0.402990639f, 0.f),
    C_Vertex3f(0.0531797409f, 0.10542345f, 0.f),
    C_Vertex3f(0.0531797409f, 0.10542345f, -1.f),
    C_Normal3f(-0.948599875f, -0.316477954f, 0.f),
    C_Vertex3f(0.047337532f, 0.120547771f, 0.f),
    C_Vertex3f(0.047337532f, 0.120547771f, -1.f),
    C_Normal3f(-0.974063396f, -0.226275548f, 0.f),
    C_Vertex3f(0.0429558754f, 0.136068106f, 0.f),
    C_Vertex3f(0.0429558754f, 0.136068106f, -1.f),
    C_Normal3f(-0.990884304f, -0.134715766f, 0.f),
    C_Vertex3f(0.040034771f, 0.151984453f, 0.f),
    C_Vertex3f(0.040034771f, 0.151984453f, -1.f),
    C_Normal3f(-0.998965979f, -0.0454645529f, 0.f),
    C_Vertex3f(0.0385742188f, 0.168296814f, 0.f),
    C_Vertex3f(0.0385742188f, 0.168296814f, -1.f),
    C_Normal3f(-0.999222755f, 0.0394201688f, 0.f),
    C_Vertex3f(0.0385742188f, 0.184076309f, 0.f),
    C_Vertex3f(0.0385742188f, 0.184076309f, -1.f),
    C_Normal3f(-0.99225527f, 0.124215335f, 0.f),
    C_Vertex3f(0.0397683382f, 0.198565364f, 0.f),
    C_Vertex3f(0.0397683382f, 0.198565364f, -1.f),
    C_Normal3f(-0.977849543f, 0.209309042f, 0.f),
    C_Vertex3f(0.0421565771f, 0.212692857f, 0.f),
    C_Vertex3f(0.0421565771f, 0.212692857f, -1.f),
    C_Normal3f(-0.955791473f, 0.294045478f, 0.f),
    C_Vertex3f(0.0457389355f, 0.226458788f, 0.f),
    C_Vertex3f(0.0457389355f, 0.226458788f, -1.f),
    C_Normal3f(-0.92643106f, 0.376464546f, 0.f),
    C_Vertex3f(0.0505154133f, 0.239863157f, 0.f),
    C_Vertex3f(0.0505154133f, 0.239863157f, -1.f),
    C_Normal3f(-0.890610576f, 0.454766959f, 0.f),
    C_Vertex3f(0.0564860106f, 0.252905965f, 0.f),
    C_Vertex3f(0.0564860106f, 0.252905965f, -1.f),
    C_Normal3f(-0.84955287f, 0.52750355f, 0.f),
    C_Vertex3f(0.0636507273f, 0.265587211f, 0.f),
    C_Vertex3f(0.0636507273f, 0.265587211f, -1.f),
    C_Normal3f(-0.80469352f, 0.593690515f, 0.f),
    C_Vertex3f(0.0720095634f, 0.277906895f, 0.f),
    C_Vertex3f(0.0720095634f, 0.277906895f, -1.f),
    C_Normal3f(-0.757717788f, 0.652582407f, 0.f),
    C_Vertex3f(0.0815625191f, 0.289865017f, 0.f),
    C_Vertex3f(0.0815625191f, 0.289865017f, -1.f),
    C_Normal3f(-0.709962845f, 0.70423919f, 0.f),
    C_Vertex3f(0.0922957659f, 0.301461339f, 0.f),
    C_Vertex3f(0.0922957659f, 0.301461339f, -1.f),
    C_Normal3f(-0.662327468f, 0.74921447f, 0.f),
    C_Vertex3f(0.104209304f, 0.312695861f, 0.f),
    C_Vertex3f(0.104209304f, 0.312695861f, -1.f),
    C_Normal3f(-0.61568749f, 0.787990391f, 0.f),
    C_Vertex3f(0.117303133f, 0.323568583f, 0.f),
    C_Vertex3f(0.117303133f, 0.323568583f, -1.f),
    C_Normal3f(-0.570679605f, 0.821172833f, 0.f),
    C_Vertex3f(0.131577253f, 0.334079504f, 0.f),
    C_Vertex3f(0.131577253f, 0.334079504f, -1.f),
    C_Normal3f(-0.527727008f, 0.849413931f, 0.f),
    C_Vertex3f(0.147031665f, 0.344228625f, 0.f),
    C_Vertex3f(0.147031665f, 0.344228625f, -1.f),
    C_Normal3f(-0.48707819f, 0.873358369f, 0.f),
    C_Vertex3f(0.163666368f, 0.354015946f, 0.f),
    C_Vertex3f(0.163666368f, 0.354015946f, -1.f),
    C_Normal3f(-0.45504877f, 0.89046663f, 0.f),
    C_Vertex3f(0.181481361f, 0.363441467f, 0.f),
    C_Vertex3f(0.181481361f, 0.363441467f, -1.f),
    C_Normal3f(-0.430644006f, 0.902521849f, 0.f),
    C_Vertex3f(0.190979004f, 0.367973328f, 0.f),
    C_Vertex3f(0.190979004f, 0.367973328f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.755975723f, -0.654599667f, 0.f),
    C_Vertex3f(0.190979004f, 0.367973328f, 0.f),
    C_Vertex3f(0.190979004f, 0.367973328f, -1.f),
    C_Normal3f(-0.764071584f, -0.645131409f, 0.f),
    C_Vertex3f(0.185590744f, 0.374196053f, 0.f),
    C_Vertex3f(0.185590744f, 0.374196053f, -1.f),
    C_Normal3f(-0.775041997f, -0.631909788f, 0.f),
    C_Vertex3f(0.175712347f, 0.386054635f, 0.f),
    C_Vertex3f(0.175712347f, 0.386054635f, -1.f),
    C_Normal3f(-0.789611816f, -0.613606632f, 0.f),
    C_Vertex3f(0.166732073f, 0.39732635f, 0.f),
    C_Vertex3f(0.166732073f, 0.39732635f, -1.f),
    C_Normal3f(-0.80593425f, -0.592004955f, 0.f),
    C_Vertex3f(0.158649921f, 0.408011198f, 0.f),
    C_Vertex3f(0.158649921f, 0.408011198f, -1.f),
    C_Normal3f(-0.82426399f, -0.56620574f, 0.f),
    C_Vertex3f(0.151465893f, 0.418109179f, 0.f),
    C_Vertex3f(0.151465893f, 0.418109179f, -1.f),
    C_Normal3f(-0.844861329f, -0.534985423f, 0.f),
    C_Vertex3f(0.145179987f, 0.427620292f, 0.f),
    C_Vertex3f(0.145179987f, 0.427620292f, -1.f),
    C_Normal3f(-0.867947102f, -0.496656537f, 0.f),
    C_Vertex3f(0.139792204f, 0.436544538f, 0.f),
    C_Vertex3f(0.139792204f, 0.436544538f, -1.f),
    C_Normal3f(-0.896297872f, -0.443452388f, 0.f),
    C_Vertex3f(0.135302544f, 0.444881916f, 0.f),
    C_Vertex3f(0.135302544f, 0.444881916f, -1.f),
    C_Normal3f(-0.928080618f, -0.372379243f, 0.f),
    C_Vertex3f(0.129915237f, 0.456507683f, 0.f),
    C_Vertex3f(0.129915237f, 0.456507683f, -1.f),
    C_Normal3f(-0.958486795f, -0.285136849f, 0.f),
    C_Vertex3f(0.124527931f, 0.471735477f, 0.f),
    C_Vertex3f(0.124527931f, 0.471735477f, -1.f),
    C_Normal3f(-0.98387593f, -0.178851992f, 0.f),
    C_Vertex3f(0.120936394f, 0.486690044f, 0.f),
    C_Vertex3f(0.120936394f, 0.486690044f, -1.f),
    C_Normal3f(-0.998100042f, -0.0616133362f, 0.f),
    C_Vertex3f(0.119140625f, 0.501371384f, 0.f),
    C_Vertex3f(0.119140625f, 0.501371384f, -1.f),
    C_Normal3f(-0.99875921f, 0.0497987345f, 0.f),
    C_Vertex3f(0.119140625f, 0.515780449f, 0.f),
    C_Vertex3f(0.119140625f, 0.515780449f, -1.f),
    C_Normal3f(-0.988242269f, 0.152895957f, 0.f),
    C_Vertex3f(0.12055409f, 0.52971971f, 0.f),
    C_Vertex3f(0.12055409f, 0.52971971f, -1.f),
    C_Normal3f(-0.966147184f, 0.257991731f, 0.f),
    C_Vertex3f(0.123381019f, 0.543188214f, 0.f),
    C_Vertex3f(0.123381019f, 0.543188214f, -1.f),
    C_Normal3f(-0.932399392f, 0.361429572f, 0.f),
    C_Vertex3f(0.127621412f, 0.556185961f, 0.f),
    C_Vertex3f(0.127621412f, 0.556185961f, -1.f),
    C_Normal3f(-0.888133287f, 0.459585994f, 0.f),
    C_Vertex3f(0.13327527f, 0.56871295f, 0.f),
    C_Vertex3f(0.13327527f, 0.56871295f, -1.f),
    C_Normal3f(-0.835508585f, 0.549477279f, 0.f),
    C_Vertex3f(0.140342593f, 0.580769181f, 0.f),
    C_Vertex3f(0.140342593f, 0.580769181f, -1.f),
    C_Normal3f(-0.777266681f, 0.629171193f, 0.f),
    C_Vertex3f(0.14882338f, 0.592354655f, 0.f),
    C_Vertex3f(0.14882338f, 0.592354655f, -1.f),
    C_Normal3f(-0.71620971f, 0.697885096f, 0.f),
    C_Vertex3f(0.158717632f, 0.603469372f, 0.f),
    C_Vertex3f(0.158717632f, 0.603469372f, -1.f),
    C_Normal3f(-0.649783075f, 0.760119855f, 0.f),
    C_Vertex3f(0.170025349f, 0.614113331f, 0.f),
    C_Vertex3f(0.170025349f, 0.614113331f, -1.f),
    C_Normal3f(-0.57192564f, 0.820305467f, 0.f),
    C_Vertex3f(0.182063937f, 0.623426795f, 0.f),
    C_Vertex3f(0.182063937f, 0.623426795f, -1.f),
    C_Normal3f(-0.486687332f, 0.873576283f, 0.f),
    C_Vertex3f(0.194833398f, 0.631409764f, 0.f),
    C_Vertex3f(0.194833398f, 0.631409764f, -1.f),
    C_Normal3f(-0.396421283f, 0.918068707f, 0.f),
    C_Vertex3f(0.208333731f, 0.638062239f, 0.f),
    C_Vertex3f(0.208333731f, 0.638062239f, -1.f),
    C_Normal3f(-0.303935289f, 0.952692628f, 0.f),
    C_Vertex3f(0.222564936f, 0.643384218f, 0.f),
    C_Vertex3f(0.222564936f, 0.643384218f, -1.f),
    C_Normal3f(-0.212074637f, 0.977253437f, 0.f),
    C_Vertex3f(0.237527013f, 0.647375703f, 0.f),
    C_Vertex3f(0.237527013f, 0.647375703f, -1.f),
    C_Normal3f(-0.123331577f, 0.992365539f, 0.f),
    C_Vertex3f(0.253219962f, 0.650036693f, 0.f),
    C_Vertex3f(0.253219962f, 0.650036693f, -1.f),
    C_Normal3f(-0.04004886f, 0.999197662f, 0.f),
    C_Vertex3f(0.269643784f, 0.651367188f, 0.f),
    C_Vertex3f(0.269643784f, 0.651367188f, -1.f),
    C_Normal3f(0.0387290753f, 0.999249756f, 0.f),
    C_Vertex3f(0.2864151f, 0.651367188f, 0.f),
    C_Vertex3f(0.2864151f, 0.651367188f, -1.f),
    C_Normal3f(0.122471318f, 0.992472053f, 0.f),
    C_Vertex3f(0.302058816f, 0.650110841f, 0.f),
    C_Vertex3f(0.302058816f, 0.650110841f, -1.f),
    C_Normal3f(0.211320668f, 0.977416754f, 0.f),
    C_Vertex3f(0.316958308f, 0.647598147f, 0.f),
    C_Vertex3f(0.316958308f, 0.647598147f, -1.f),
    C_Normal3f(0.303935647f, 0.952692568f, 0.f),
    C_Vertex3f(0.331113577f, 0.643829107f, 0.f),
    C_Vertex3f(0.331113577f, 0.643829107f, -1.f),
    C_Normal3f(0.397806227f, 0.917469501f, 0.f),
    C_Vertex3f(0.344524622f, 0.63880372f, 0.f),
    C_Vertex3f(0.344524622f, 0.63880372f, -1.f),
    C_Normal3f(0.489945203f, 0.871753275f, 0.f),
    C_Vertex3f(0.357191443f, 0.632521987f, 0.f),
    C_Vertex3f(0.357191443f, 0.632521987f, -1.f),
    C_Normal3f(0.577295005f, 0.816535711f, 0.f),
    C_Vertex3f(0.369114041f, 0.624983907f, 0.f),
    C_Vertex3f(0.369114041f, 0.624983907f, -1.f),
    C_Normal3f(0.657204807f, 0.753712058f, 0.f),
    C_Vertex3f(0.380292416f, 0.61618948f, 0.f),
    C_Vertex3f(0.380292416f, 0.61618948f, -1.f),
    C_Normal3f(0.725219011f, 0.688518226f, 0.f),
    C_Vertex3f(0.390726566f, 0.606138706f, 0.f),
    C_Vertex3f(0.390726566f, 0.606138706f, -1.f),
    C_Normal3f(0.786843419f, 0.61715275f, 0.f),
    C_Vertex3f(0.399856448f, 0.595582604f, 0.f),
    C_Vertex3f(0.399856448f, 0.595582604f, -1.f),
    C_Normal3f(0.844552875f, 0.535472095f, 0.f),
    C_Vertex3f(0.407682061f, 0.584521174f, 0.f),
    C_Vertex3f(0.407682061f, 0.584521174f, -1.f),
    C_Normal3f(0.89565146f, 0.444756627f, 0.f),
    C_Vertex3f(0.414203405f, 0.572954416f, 0.f),
    C_Vertex3f(0.414203405f, 0.572954416f, -1.f),
    C_Normal3f(0.937743306f, 0.347328961f, 0.f),
    C_Vertex3f(0.419420481f, 0.56088233f, 0.f),
    C_Vertex3f(0.419420481f, 0.56088233f, -1.f),
    C_Normal3f(0.969190419f, 0.246312767f, 0.f),
    C_Vertex3f(0.423333287f, 0.548304915f, 0.f),
    C_Vertex3f(0.423333287f, 0.548304915f, -1.f),
    C_Normal3f(0.989409089f, 0.145153672f, 0.f),
    C_Vertex3f(0.425941825f, 0.535222173f, 0.f),
    C_Vertex3f(0.425941825f, 0.535222173f, -1.f),
    C_Normal3f(0.999173045f, 0.0406596325f, 0.f),
    C_Vertex3f(0.427246094f, 0.521634102f, 0.f),
    C_Vertex3f(0.427246094f, 0.521634102f, -1.f),
    C_Normal3f(0.995619297f, -0.0934995413f, 0.f),
    C_Vertex3f(0.427246094f, 0.503170967f, 0.f),
    C_Vertex3f(0.427246094f, 0.503170967f, -1.f),
    C_Normal3f(0.966853142f, -0.255333304f, 0.f),
    C_Vertex3f(0.423446178f, 0.481171131f, 0.f),
    C_Vertex3f(0.423446178f, 0.481171131f, -1.f),
    C_Normal3f(0.909243345f, -0.4162651f, 0.f),
    C_Vertex3f(0.415846348f, 0.46000433f, 0.f),
    C_Vertex3f(0.415846348f, 0.46000433f, -1.f),
    C_Normal3f(0.831635058f, -0.555322647f, 0.f),
    C_Vertex3f(0.404446602f, 0.439670563f, 0.f),
    C_Vertex3f(0.404446602f, 0.439670563f, -1.f),
    C_Normal3f(0.744770467f, -0.667320669f, 0.f),
    C_Vertex3f(0.389246941f, 0.42016983f, 0.f),
    C_Vertex3f(0.389246941f, 0.42016983f, -1.f),
    C_Normal3f(0.65777123f, -0.753217697f, 0.f),
    C_Vertex3f(0.370247364f, 0.401502132f, 0.f),
    C_Vertex3f(0.370247364f, 0.401502132f, -1.f),
    C_Normal3f(0.576313019f, -0.817229033f, 0.f),
    C_Vertex3f(0.347447872f, 0.383667469f, 0.f),
    C_Vertex3f(0.347447872f, 0.383667469f, -1.f),
    C_Normal3f(0.514593601f, -0.857434273f, 0.f),
    C_Vertex3f(0.320848465f, 0.36666584f, 0.f),
    C_Vertex3f(0.320848465f, 0.36666584f, -1.f),
    C_Normal3f(0.469584495f, -0.882887542f, 0.f),
    C_Vertex3f(0.305648804f, 0.358581543f, 0.f),
    C_Vertex3f(0.305648804f, 0.358581543f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.783489406f, 0.621405125f, 0.f),
    C_Vertex3f(0.305648804f, 0.358581543f, 0.f),
    C_Vertex3f(0.305648804f, 0.358581543f, -1.f),
    C_Vertex3f(0.421302795f, 0.212760925f, -1.f),
    C_Vertex3f(0.421302795f, 0.212760925f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.891721725f, 0.452584147f, 0.f),
    C_Vertex3f(0.421302795f, 0.212760925f, 0.f),
    C_Vertex3f(0.421302795f, 0.212760925f, -1.f),
    C_Normal3f(-0.913196743f, 0.407518893f, 0.f),
    C_Vertex3f(0.426181793f, 0.222373962f, 0.f),
    C_Vertex3f(0.426181793f, 0.222373962f, -1.f),
    C_Normal3f(-0.935458302f, 0.353436828f, 0.f),
    C_Vertex3f(0.434855461f, 0.243130684f, 0.f),
    C_Vertex3f(0.434855461f, 0.243130684f, -1.f),
    C_Normal3f(-0.956313372f, 0.292343676f, 0.f),
    C_Vertex3f(0.442444801f, 0.265418053f, 0.f),
    C_Vertex3f(0.442444801f, 0.265418053f, -1.f),
    C_Normal3f(-0.969562948f, 0.244842201f, 0.f),
    C_Vertex3f(0.448949814f, 0.289236069f, 0.f),
    C_Vertex3f(0.448949814f, 0.289236069f, -1.f),
    C_Normal3f(-0.97789067f, 0.209117025f, 0.f),
    C_Vertex3f(0.451660156f, 0.3019104f, 0.f),
    C_Vertex3f(0.451660156f, 0.3019104f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.209591568f, 0.977789044f, 0.f),
    C_Vertex3f(0.451660156f, 0.3019104f, 0.f),
    C_Vertex3f(0.451660156f, 0.3019104f, -1.f),
    C_Vertex3f(0.533203125f, 0.284431458f, -1.f),
    C_Vertex3f(0.533203125f, 0.284431458f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.970762491f, -0.24004212f, 0.f),
    C_Vertex3f(0.533203125f, 0.284431458f, 0.f),
    C_Vertex3f(0.533203125f, 0.284431458f, -1.f),
    C_Normal3f(0.965468407f, -0.260519862f, 0.f),
    C_Vertex3f(0.530610085f, 0.273944855f, 0.f),
    C_Vertex3f(0.530610085f, 0.273944855f, -1.f),
    C_Normal3f(0.957811654f, -0.28739658f, 0.f),
    C_Vertex3f(0.524965048f, 0.25390172f, 0.f),
    C_Vertex3f(0.524965048f, 0.25390172f, -1.f),
    C_Normal3f(0.94687885f, -0.321590513f, 0.f),
    C_Vertex3f(0.518861055f, 0.234788656f, 0.f),
    C_Vertex3f(0.518861055f, 0.234788656f, -1.f),
    C_Normal3f(0.933736503f, -0.35796091f, 0.f),
    C_Vertex3f(0.512298107f, 0.216605663f, 0.f),
    C_Vertex3f(0.512298107f, 0.216605663f, -1.f),
    C_Normal3f(0.918021262f, -0.396531314f, 0.f),
    C_Vertex3f(0.505276203f, 0.199352741f, 0.f),
    C_Vertex3f(0.505276203f, 0.199352741f, -1.f),
    C_Normal3f(0.899330497f, -0.437269658f, 0.f),
    C_Vertex3f(0.497795343f, 0.18302989f, 0.f),
    C_Vertex3f(0.497795343f, 0.18302989f, -1.f),
    C_Normal3f(0.877230525f, -0.480069399f, 0.f),
    C_Vertex3f(0.489855528f, 0.16763711f, 0.f),
    C_Vertex3f(0.489855528f, 0.16763711f, -1.f),
    C_Normal3f(0.855884135f, -0.517167747f, 0.f),
    C_Vertex3f(0.481456757f, 0.1531744f, 0.f),
    C_Vertex3f(0.481456757f, 0.1531744f, -1.f),
    C_Normal3f(0.956226945f, 0.292626143f, 0.f),
    C_Vertex3f(0.477027893f, 0.146408081f, 0.f),
    C_Vertex3f(0.477027893f, 0.146408081f, -1.f),
    C_Normal3f(0.770648122f, 0.637260914f, 0.f),
    C_Vertex3f(0.488014221f, 0.131746292f, 0.f),
    C_Vertex3f(0.488014221f, 0.131746292f, -1.f),
    C_Normal3f(0.729652047f, 0.683818638f, 0.f),
    C_Vertex3f(0.511454582f, 0.104775429f, 0.f),
    C_Vertex3f(0.511454582f, 0.104775429f, -1.f),
    C_Normal3f(0.674733937f, 0.738061011f, 0.f),
    C_Vertex3f(0.536362648f, 0.08015728f, 0.f),
    C_Vertex3f(0.536362648f, 0.08015728f, -1.f),
    C_Normal3f(0.624503434f, 0.781022072f, 0.f),
    C_Vertex3f(0.562738419f, 0.0578918457f, 0.f),
    C_Vertex3f(0.562738419f, 0.0578918457f, -1.f),
    C_Normal3f(0.581712425f, 0.813394427f, 0.f),
    C_Vertex3f(0.576660156f, 0.0479354858f, 0.f),
    C_Vertex3f(0.576660156f, 0.0479354858f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.764717758f, -0.64436537f, 0.f),
    C_Vertex3f(0.576660156f, 0.0479354858f, 0.f),
    C_Vertex3f(0.576660156f, 0.0479354858f, -1.f),
    C_Vertex3f(0.523925781f, -0.0146484375f, -1.f),
    C_Vertex3f(0.523925781f, -0.0146484375f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.53726691f, -0.84341228f, 0.f),
    C_Vertex3f(0.523925781f, -0.0146484375f, 0.f),
    C_Vertex3f(0.523925781f, -0.0146484375f, -1.f),
    C_Normal3f(-0.590055466f, -0.807362795f, 0.f),
    C_Vertex3f(0.511960983f, -0.00702667236f, 0.f),
    C_Vertex3f(0.511960983f, -0.00702667236f, -1.f),
    C_Normal3f(-0.645592511f, -0.763682187f, 0.f),
    C_Vertex3f(0.487647057f, 0.0118656158f, 0.f),
    C_Vertex3f(0.487647057f, 0.0118656158f, -1.f),
    C_Normal3f(-0.699530303f, -0.714602947f, 0.f),
    C_Vertex3f(0.462948799f, 0.034406662f, 0.f),
    C_Vertex3f(0.462948799f, 0.034406662f, -1.f),
    C_Normal3f(-0.735969543f, -0.677014768f, 0.f),
    C_Vertex3f(0.437866211f, 0.0605964661f, 0.f),
    C_Vertex3f(0.437866211f, 0.0605964661f, -1.f),
    C_Normal3f(-0.760627747f, -0.64918828f, 0.f),
    C_Vertex3f(0.425132751f, 0.0755157471f, 0.f),
    C_Vertex3f(0.425132751f, 0.0755157471f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.509632707f, 0.860392034f, 0.f),
    C_Vertex3f(0.265274048f, 0.409492493f, 0.f),
    C_Vertex3f(0.265274048f, 0.409492493f, -1.f),
    C_Normal3f(-0.524717748f, 0.851276219f, 0.f),
    C_Vertex3f(0.271636963f, 0.413261414f, 0.f),
    C_Vertex3f(0.271636963f, 0.413261414f, -1.f),
    C_Normal3f(-0.545607924f, 0.838040531f, 0.f),
    C_Vertex3f(0.283241034f, 0.420567155f, 0.f),
    C_Vertex3f(0.283241034f, 0.420567155f, -1.f),
    C_Normal3f(-0.574159384f, 0.818743587f, 0.f),
    C_Vertex3f(0.293723345f, 0.427640796f, 0.f),
    C_Vertex3f(0.293723345f, 0.427640796f, -1.f),
    C_Normal3f(-0.607240975f, 0.794517756f, 0.f),
    C_Vertex3f(0.303083897f, 0.434482336f, 0.f),
    C_Vertex3f(0.303083897f, 0.434482336f, -1.f),
    C_Normal3f(-0.645750642f, 0.763548374f, 0.f),
    C_Vertex3f(0.311322689f, 0.441091776f, 0.f),
    C_Vertex3f(0.311322689f, 0.441091776f, -1.f),
    C_Normal3f(-0.690657079f, 0.723182321f, 0.f),
    C_Vertex3f(0.318439722f, 0.447469115f, 0.f),
    C_Vertex3f(0.318439722f, 0.447469115f, -1.f),
    C_Normal3f(-0.742796302f, 0.669517577f, 0.f),
    C_Vertex3f(0.324434996f, 0.453614354f, 0.f),
    C_Vertex3f(0.324434996f, 0.453614354f, -1.f),
    C_Normal3f(-0.808651209f, 0.588288426f, 0.f),
    C_Vertex3f(0.32930851f, 0.459527493f, 0.f),
    C_Vertex3f(0.32930851f, 0.459527493f, -1.f),
    C_Normal3f(-0.876422584f, 0.481542796f, 0.f),
    C_Vertex3f(0.334936142f, 0.468049049f, 0.f),
    C_Vertex3f(0.334936142f, 0.468049049f, -1.f),
    C_Normal3f(-0.934038818f, 0.357171476f, 0.f),
    C_Vertex3f(0.340563774f, 0.480012417f, 0.f),
    C_Vertex3f(0.340563774f, 0.480012417f, -1.f),
    C_Normal3f(-0.976907492f, 0.213662863f, 0.f),
    C_Vertex3f(0.344315529f, 0.492577076f, 0.f),
    C_Vertex3f(0.344315529f, 0.492577076f, -1.f),
    C_Normal3f(-0.997801602f, 0.0662719086f, 0.f),
    C_Vertex3f(0.346191406f, 0.505743027f, 0.f),
    C_Vertex3f(0.346191406f, 0.505743027f, -1.f),
    C_Normal3f(-0.996314168f, -0.0857784376f, 0.f),
    C_Vertex3f(0.346191406f, 0.520820618f, 0.f),
    C_Vertex3f(0.346191406f, 0.520820618f, -1.f),
    C_Normal3f(-0.964507639f, -0.264054775f, 0.f),
    C_Vertex3f(0.343611717f, 0.535706043f, 0.f),
    C_Vertex3f(0.343611717f, 0.535706043f, -1.f),
    C_Normal3f(-0.890633345f, -0.454722106f, 0.f),
    C_Vertex3f(0.338452339f, 0.549088955f, 0.f),
    C_Vertex3f(0.338452339f, 0.549088955f, -1.f),
    C_Normal3f(-0.776575327f, -0.630024374f, 0.f),
    C_Vertex3f(0.330713272f, 0.560969353f, 0.f),
    C_Vertex3f(0.330713272f, 0.560969353f, -1.f),
    C_Normal3f(-0.638092518f, -0.769959688f, 0.f),
    C_Vertex3f(0.320394516f, 0.571347237f, 0.f),
    C_Vertex3f(0.320394516f, 0.571347237f, -1.f),
    C_Normal3f(-0.468408972f, -0.883511782f, 0.f),
    C_Vertex3f(0.30879879f, 0.57913065f, 0.f),
    C_Vertex3f(0.30879879f, 0.57913065f, -1.f),
    C_Normal3f(-0.276783139f, -0.960932434f, 0.f),
    C_Vertex3f(0.295926094f, 0.584319592f, 0.f),
    C_Vertex3f(0.295926094f, 0.584319592f, -1.f),
    C_Normal3f(-0.0867527723f, -0.996229887f, 0.f),
    C_Vertex3f(0.281776428f, 0.586914063f, 0.f),
    C_Vertex3f(0.281776428f, 0.586914063f, -1.f),
    C_Normal3f(0.0847758874f, -0.996400058f, 0.f),
    C_Vertex3f(0.266132355f, 0.586914063f, 0.f),
    C_Vertex3f(0.266132355f, 0.586914063f, -1.f),
    C_Normal3f(0.267623365f, -0.963523626f, 0.f),
    C_Vertex3f(0.251602173f, 0.584346771f, 0.f),
    C_Vertex3f(0.251602173f, 0.584346771f, -1.f),
    C_Normal3f(0.455807865f, -0.890078247f, 0.f),
    C_Vertex3f(0.23840332f, 0.579212189f, 0.f),
    C_Vertex3f(0.23840332f, 0.579212189f, -1.f),
    C_Normal3f(0.625714958f, -0.780051768f, 0.f),
    C_Vertex3f(0.226535797f, 0.571510315f, 0.f),
    C_Vertex3f(0.226535797f, 0.571510315f, -1.f),
    C_Normal3f(0.761842608f, -0.647762239f, 0.f),
    C_Vertex3f(0.215999603f, 0.56124115f, 0.f),
    C_Vertex3f(0.215999603f, 0.56124115f, -1.f),
    C_Normal3f(0.876506507f, -0.481390119f, 0.f),
    C_Vertex3f(0.208097458f, 0.549824715f, 0.f),
    C_Vertex3f(0.208097458f, 0.549824715f, -1.f),
    C_Normal3f(0.957628071f, -0.288007915f, 0.f),
    C_Vertex3f(0.202829361f, 0.537261009f, 0.f),
    C_Vertex3f(0.202829361f, 0.537261009f, -1.f),
    C_Normal3f(0.99577713f, -0.0918033496f, 0.f),
    C_Vertex3f(0.200195313f, 0.523550034f, 0.f),
    C_Vertex3f(0.200195313f, 0.523550034f, -1.f),
    C_Normal3f(0.992336988f, 0.123560913f, 0.f),
    C_Vertex3f(0.200195313f, 0.50868988f, 0.f),
    C_Vertex3f(0.200195313f, 0.50868988f, -1.f),
    C_Normal3f(0.946780384f, 0.321880102f, 0.f),
    C_Vertex3f(0.203990936f, 0.493066788f, 0.f),
    C_Vertex3f(0.203990936f, 0.493066788f, -1.f),
    C_Normal3f(0.893120766f, 0.449817091f, 0.f),
    C_Vertex3f(0.209684372f, 0.480778694f, 0.f),
    C_Vertex3f(0.209684372f, 0.480778694f, -1.f),
    C_Normal3f(0.853806198f, 0.520590901f, 0.f),
    C_Vertex3f(0.214429855f, 0.472340107f, 0.f),
    C_Vertex3f(0.214429855f, 0.472340107f, -1.f),
    C_Normal3f(0.819039226f, 0.573737621f, 0.f),
    C_Vertex3f(0.220125198f, 0.463654995f, 0.f),
    C_Vertex3f(0.220125198f, 0.463654995f, -1.f),
    C_Normal3f(0.791429281f, 0.611260831f, 0.f),
    C_Vertex3f(0.226770401f, 0.454723358f, 0.f),
    C_Vertex3f(0.226770401f, 0.454723358f, -1.f),
    C_Normal3f(0.760454476f, 0.649391294f, 0.f),
    C_Vertex3f(0.230567932f, 0.450134277f, 0.f),
    C_Vertex3f(0.230567932f, 0.450134277f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.760454476f, 0.649391294f, 0.f),
    C_Vertex3f(0.230567932f, 0.450134277f, 0.f),
    C_Vertex3f(0.230567932f, 0.450134277f, -1.f),
    C_Vertex3f(0.265274048f, 0.409492493f, -1.f),
    C_Vertex3f(0.265274048f, 0.409492493f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.780421376f, -0.625253975f, 0.f),
    C_Vertex3f(0.37563324f, 0.137496948f, 0.f),
    C_Vertex3f(0.37563324f, 0.137496948f, -1.f),
    C_Vertex3f(0.232032776f, 0.316734314f, -1.f),
    C_Vertex3f(0.232032776f, 0.316734314f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.51191318f, -0.85903728f, 0.f),
    C_Vertex3f(0.232032776f, 0.316734314f, 0.f),
    C_Vertex3f(0.232032776f, 0.316734314f, -1.f),
    C_Normal3f(0.528495848f, -0.848935843f, 0.f),
    C_Vertex3f(0.224048615f, 0.311976433f, 0.f),
    C_Vertex3f(0.224048615f, 0.311976433f, -1.f),
    C_Normal3f(0.551100314f, -0.834439039f, 0.f),
    C_Vertex3f(0.209379077f, 0.302631497f, 0.f),
    C_Vertex3f(0.209379077f, 0.302631497f, -1.f),
    C_Normal3f(0.581323326f, -0.813672721f, 0.f),
    C_Vertex3f(0.196008325f, 0.293457389f, 0.f),
    C_Vertex3f(0.196008325f, 0.293457389f, -1.f),
    C_Normal3f(0.615386128f, -0.78822577f, 0.f),
    C_Vertex3f(0.183936357f, 0.284454107f, 0.f),
    C_Vertex3f(0.183936357f, 0.284454107f, -1.f),
    C_Normal3f(0.653781176f, -0.756683648f, 0.f),
    C_Vertex3f(0.173163176f, 0.275621653f, 0.f),
    C_Vertex3f(0.173163176f, 0.275621653f, -1.f),
    C_Normal3f(0.696925879f, -0.717143238f, 0.f),
    C_Vertex3f(0.163688779f, 0.266960025f, 0.f),
    C_Vertex3f(0.163688779f, 0.266960025f, -1.f),
    C_Normal3f(0.744994223f, -0.667070925f, 0.f),
    C_Vertex3f(0.155513167f, 0.258469224f, 0.f),
    C_Vertex3f(0.155513167f, 0.258469224f, -1.f),
    C_Normal3f(0.803064525f, -0.595892131f, 0.f),
    C_Vertex3f(0.148636341f, 0.25014925f, 0.f),
    C_Vertex3f(0.148636341f, 0.25014925f, -1.f),
    C_Normal3f(0.861868143f, -0.50713253f, 0.f),
    C_Vertex3f(0.140269279f, 0.237925529f, 0.f),
    C_Vertex3f(0.140269279f, 0.237925529f, -1.f),
    C_Normal3f(0.918308377f, -0.395865768f, 0.f),
    C_Vertex3f(0.131902218f, 0.221709728f, 0.f),
    C_Vertex3f(0.131902218f, 0.221709728f, -1.f),
    C_Normal3f(0.967828393f, -0.251611024f, 0.f),
    C_Vertex3f(0.126324177f, 0.20557642f, 0.f),
    C_Vertex3f(0.126324177f, 0.20557642f, -1.f),
    C_Normal3f(0.996606946f, -0.08230827f, 0.f),
    C_Vertex3f(0.123535156f, 0.189525604f, 0.f),
    C_Vertex3f(0.123535156f, 0.189525604f, -1.f),
    C_Normal3f(0.994734704f, 0.102484375f, 0.f),
    C_Vertex3f(0.123535156f, 0.171806335f, 0.f),
    C_Vertex3f(0.123535156f, 0.171806335f, -1.f),
    C_Normal3f(0.960512042f, 0.278238088f, 0.f),
    C_Vertex3f(0.12740612f, 0.15195322f, 0.f),
    C_Vertex3f(0.12740612f, 0.15195322f, -1.f),
    C_Normal3f(0.903716564f, 0.428131223f, 0.f),
    C_Vertex3f(0.135148048f, 0.131717205f, 0.f),
    C_Vertex3f(0.135148048f, 0.131717205f, -1.f),
    C_Normal3f(0.842809618f, 0.538211882f, 0.f),
    C_Vertex3f(0.146760941f, 0.111098289f, 0.f),
    C_Vertex3f(0.146760941f, 0.111098289f, -1.f),
    C_Normal3f(0.777705848f, 0.628628433f, 0.f),
    C_Vertex3f(0.158373833f, 0.0953469276f, 0.f),
    C_Vertex3f(0.158373833f, 0.0953469276f, -1.f),
    C_Normal3f(0.690074325f, 0.723738432f, 0.f),
    C_Vertex3f(0.166919947f, 0.0861586332f, 0.f),
    C_Vertex3f(0.166919947f, 0.0861586332f, -1.f),
    C_Normal3f(0.59498167f, 0.803739309f, 0.f),
    C_Vertex3f(0.176270247f, 0.0782829523f, 0.f),
    C_Vertex3f(0.176270247f, 0.0782829523f, -1.f),
    C_Normal3f(0.488293499f, 0.872679472f, 0.f),
    C_Vertex3f(0.186424732f, 0.0717198849f, 0.f),
    C_Vertex3f(0.186424732f, 0.0717198849f, -1.f),
    C_Normal3f(0.374894232f, 0.927067578f, 0.f),
    C_Vertex3f(0.197383404f, 0.0664694309f, 0.f),
    C_Vertex3f(0.197383404f, 0.0664694309f, -1.f),
    C_Normal3f(0.260443687f, 0.96548897f, 0.f),
    C_Vertex3f(0.209146261f, 0.0625315905f, 0.f),
    C_Vertex3f(0.209146261f, 0.0625315905f, -1.f),
    C_Normal3f(0.150095984f, 0.988671422f, 0.f),
    C_Vertex3f(0.221713305f, 0.0599063635f, 0.f),
    C_Vertex3f(0.221713305f, 0.0599063635f, -1.f),
    C_Normal3f(0.0447695181f, 0.998997331f, 0.f),
    C_Vertex3f(0.235084534f, 0.05859375f, 0.f),
    C_Vertex3f(0.235084534f, 0.05859375f, -1.f),
    C_Normal3f(-0.0809533373f, 0.99671787f, 0.f),
    C_Vertex3f(0.251003265f, 0.05859375f, 0.f),
    C_Vertex3f(0.251003265f, 0.05859375f, -1.f),
    C_Normal3f(-0.222195745f, 0.97500205f, 0.f),
    C_Vertex3f(0.268965721f, 0.0613455772f, 0.f),
    C_Vertex3f(0.268965721f, 0.0613455772f, -1.f),
    C_Normal3f(-0.349993318f, 0.9367522f, 0.f),
    C_Vertex3f(0.287228584f, 0.0668492317f, 0.f),
    C_Vertex3f(0.287228584f, 0.0668492317f, -1.f),
    C_Normal3f(-0.457623184f, 0.889146328f, 0.f),
    C_Vertex3f(0.305791855f, 0.0751047134f, 0.f),
    C_Vertex3f(0.305791855f, 0.0751047134f, -1.f),
    C_Normal3f(-0.550245106f, 0.835003257f, 0.f),
    C_Vertex3f(0.324655533f, 0.0861120224f, 0.f),
    C_Vertex3f(0.324655533f, 0.0861120224f, -1.f),
    C_Normal3f(-0.641580462f, 0.76705575f, 0.f),
    C_Vertex3f(0.341638565f, 0.0987267494f, 0.f),
    C_Vertex3f(0.341638565f, 0.0987267494f, -1.f),
    C_Normal3f(-0.72771436f, 0.685880303f, 0.f),
    C_Vertex3f(0.356740952f, 0.112948895f, 0.f),
    C_Vertex3f(0.356740952f, 0.112948895f, -1.f),
    C_Normal3f(-0.792480886f, 0.609896719f, 0.f),
    C_Vertex3f(0.369962692f, 0.128778458f, 0.f),
    C_Vertex3f(0.369962692f, 0.128778458f, -1.f),
    C_Normal3f(-0.838288546f, 0.545226812f, 0.f),
    C_Vertex3f(0.37563324f, 0.137496948f, 0.f),
    C_Vertex3f(0.37563324f, 0.137496948f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151465893f, 0.418109179f, 0.f),
    C_Vertex3f(0.158649921f, 0.408011198f, 0.f),
    C_Vertex3f(0.158717632f, 0.603469372f, 0.f),
    C_Vertex3f(0.166732073f, 0.39732635f, 0.f),
    C_Vertex3f(0.170025349f, 0.614113331f, 0.f),
    C_Vertex3f(0.175712347f, 0.386054635f, 0.f),
    C_Vertex3f(0.182063937f, 0.623426795f, 0.f),
    C_Vertex3f(0.185590744f, 0.374196053f, 0.f),
    C_Vertex3f(0.194833398f, 0.631409764f, 0.f),
    C_Vertex3f(0.190979004f, 0.367973328f, 0.f),
    C_Vertex3f(0.196008325f, 0.293457389f, 0.f),
    C_Vertex3f(0.183936357f, 0.284454107f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.135302544f, 0.444881916f, 0.f),
    C_Vertex3f(0.139792204f, 0.436544538f, 0.f),
    C_Vertex3f(0.140342593f, 0.580769181f, 0.f),
    C_Vertex3f(0.145179987f, 0.427620292f, 0.f),
    C_Vertex3f(0.14882338f, 0.592354655f, 0.f),
    C_Vertex3f(0.151465893f, 0.418109179f, 0.f),
    C_Vertex3f(0.158717632f, 0.603469372f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12055409f, 0.52971971f, 0.f),
    C_Vertex3f(0.120936394f, 0.486690044f, 0.f),
    C_Vertex3f(0.123381019f, 0.543188214f, 0.f),
    C_Vertex3f(0.124527931f, 0.471735477f, 0.f),
    C_Vertex3f(0.127621412f, 0.556185961f, 0.f),
    C_Vertex3f(0.129915237f, 0.456507683f, 0.f),
    C_Vertex3f(0.13327527f, 0.56871295f, 0.f),
    C_Vertex3f(0.135302544f, 0.444881916f, 0.f),
    C_Vertex3f(0.140342593f, 0.580769181f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.119140625f, 0.515780449f, 0.f),
    C_Vertex3f(0.119140625f, 0.501371384f, 0.f),
    C_Vertex3f(0.120936394f, 0.486690044f, 0.f),
    C_Vertex3f(0.12055409f, 0.52971971f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.533203125f, 0.284431458f, 0.f),
    C_Vertex3f(0.451660156f, 0.3019104f, 0.f),
    C_Vertex3f(0.477027893f, 0.146408081f, 0.f),
    C_Vertex3f(0.481456757f, 0.1531744f, 0.f),
    C_Vertex3f(0.489855528f, 0.16763711f, 0.f),
    C_Vertex3f(0.497795343f, 0.18302989f, 0.f),
    C_Vertex3f(0.505276203f, 0.199352741f, 0.f),
    C_Vertex3f(0.512298107f, 0.216605663f, 0.f),
    C_Vertex3f(0.518861055f, 0.234788656f, 0.f),
    C_Vertex3f(0.524965048f, 0.25390172f, 0.f),
    C_Vertex3f(0.530610085f, 0.273944855f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.536362648f, 0.08015728f, 0.f),
    C_Vertex3f(0.511454582f, 0.104775429f, 0.f),
    C_Vertex3f(0.511960983f, -0.00702667236f, 0.f),
    C_Vertex3f(0.488014221f, 0.131746292f, 0.f),
    C_Vertex3f(0.487647057f, 0.0118656158f, 0.f),
    C_Vertex3f(0.477027893f, 0.146408081f, 0.f),
    C_Vertex3f(0.462948799f, 0.034406662f, 0.f),
    C_Vertex3f(0.451660156f, 0.3019104f, 0.f),
    C_Vertex3f(0.448949814f, 0.289236069f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.442444801f, 0.265418053f, 0.f),
    C_Vertex3f(0.434855461f, 0.243130684f, 0.f),
    C_Vertex3f(0.437866211f, 0.0605964661f, 0.f),
    C_Vertex3f(0.462948799f, 0.034406662f, 0.f),
    C_Vertex3f(0.448949814f, 0.289236069f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.395104885f, 0.0458645821f, 0.f),
    C_Vertex3f(0.41541481f, 0.0647335052f, 0.f),
    C_Vertex3f(0.421302795f, 0.212760925f, 0.f),
    C_Vertex3f(0.425132751f, 0.0755157471f, 0.f),
    C_Vertex3f(0.426181793f, 0.222373962f, 0.f),
    C_Vertex3f(0.437866211f, 0.0605964661f, 0.f),
    C_Vertex3f(0.434855461f, 0.243130684f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37563324f, 0.137496948f, 0.f),
    C_Vertex3f(0.369962692f, 0.128778458f, 0.f),
    C_Vertex3f(0.373920918f, 0.0296912193f, 0.f),
    C_Vertex3f(0.395104885f, 0.0458645821f, 0.f),
    C_Vertex3f(0.421302795f, 0.212760925f, 0.f),
    C_Vertex3f(0.305648804f, 0.358581543f, 0.f),
    C_Vertex3f(0.283241034f, 0.420567155f, 0.f),
    C_Vertex3f(0.271636963f, 0.413261414f, 0.f),
    C_Vertex3f(0.265274048f, 0.409492493f, 0.f),
    C_Vertex3f(0.232032776f, 0.316734314f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.324434996f, 0.453614354f, 0.f),
    C_Vertex3f(0.318439722f, 0.447469115f, 0.f),
    C_Vertex3f(0.320848465f, 0.36666584f, 0.f),
    C_Vertex3f(0.311322689f, 0.441091776f, 0.f),
    C_Vertex3f(0.305648804f, 0.358581543f, 0.f),
    C_Vertex3f(0.303083897f, 0.434482336f, 0.f),
    C_Vertex3f(0.293723345f, 0.427640796f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.265274048f, 0.409492493f, 0.f),
    C_Vertex3f(0.230567932f, 0.450134277f, 0.f),
    C_Vertex3f(0.232032776f, 0.316734314f, 0.f),
    C_Vertex3f(0.226770401f, 0.454723358f, 0.f),
    C_Vertex3f(0.224048615f, 0.311976433f, 0.f),
    C_Vertex3f(0.220125198f, 0.463654995f, 0.f),
    C_Vertex3f(0.214429855f, 0.472340107f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.209684372f, 0.480778694f, 0.f),
    C_Vertex3f(0.203990936f, 0.493066788f, 0.f),
    C_Vertex3f(0.209379077f, 0.302631497f, 0.f),
    C_Vertex3f(0.224048615f, 0.311976433f, 0.f),
    C_Vertex3f(0.214429855f, 0.472340107f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.208333731f, 0.638062239f, 0.f),
    C_Vertex3f(0.194833398f, 0.631409764f, 0.f),
    C_Vertex3f(0.200195313f, 0.523550034f, 0.f),
    C_Vertex3f(0.196008325f, 0.293457389f, 0.f),
    C_Vertex3f(0.200195313f, 0.50868988f, 0.f),
    C_Vertex3f(0.209379077f, 0.302631497f, 0.f),
    C_Vertex3f(0.203990936f, 0.493066788f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.523925781f, -0.0146484375f, 0.f),
    C_Vertex3f(0.576660156f, 0.0479354858f, 0.f),
    C_Vertex3f(0.562738419f, 0.0578918457f, 0.f),
    C_Vertex3f(0.536362648f, 0.08015728f, 0.f),
    C_Vertex3f(0.511960983f, -0.00702667236f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.324655533f, 0.0861120224f, 0.f),
    C_Vertex3f(0.328930855f, 0.00543117523f, 0.f),
    C_Vertex3f(0.341638565f, 0.0987267494f, 0.f),
    C_Vertex3f(0.351862907f, 0.0162134171f, 0.f),
    C_Vertex3f(0.356740952f, 0.112948895f, 0.f),
    C_Vertex3f(0.373920918f, 0.0296912193f, 0.f),
    C_Vertex3f(0.369962692f, 0.128778458f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.251003265f, 0.05859375f, 0.f),
    C_Vertex3f(0.254564285f, -0.0107421875f, 0.f),
    C_Vertex3f(0.268965721f, 0.0613455772f, 0.f),
    C_Vertex3f(0.280281544f, -0.00804662704f, 0.f),
    C_Vertex3f(0.287228584f, 0.0668492317f, 0.f),
    C_Vertex3f(0.3050704f, -0.00265550613f, 0.f),
    C_Vertex3f(0.305791855f, 0.0751047134f, 0.f),
    C_Vertex3f(0.328930855f, 0.00543117523f, 0.f),
    C_Vertex3f(0.324655533f, 0.0861120224f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235084534f, 0.05859375f, 0.f),
    C_Vertex3f(0.221713305f, 0.0599063635f, 0.f),
    C_Vertex3f(0.228957176f, -0.0107421875f, 0.f),
    C_Vertex3f(0.254564285f, -0.0107421875f, 0.f),
    C_Vertex3f(0.251003265f, 0.05859375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.209146261f, 0.0625315905f, 0.f),
    C_Vertex3f(0.197383404f, 0.0664694309f, 0.f),
    C_Vertex3f(0.205657959f, -0.00866723061f, 0.f),
    C_Vertex3f(0.228957176f, -0.0107421875f, 0.f),
    C_Vertex3f(0.221713305f, 0.0599063635f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.186424732f, 0.0717198849f, 0.f),
    C_Vertex3f(0.176270247f, 0.0782829523f, 0.f),
    C_Vertex3f(0.183628082f, -0.00451731682f, 0.f),
    C_Vertex3f(0.205657959f, -0.00866723061f, 0.f),
    C_Vertex3f(0.197383404f, 0.0664694309f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.166919947f, 0.0861586332f, 0.f),
    C_Vertex3f(0.158373833f, 0.0953469276f, 0.f),
    C_Vertex3f(0.162867546f, 0.00170755386f, 0.f),
    C_Vertex3f(0.183628082f, -0.00451731682f, 0.f),
    C_Vertex3f(0.176270247f, 0.0782829523f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.146760941f, 0.111098289f, 0.f),
    C_Vertex3f(0.135148048f, 0.131717205f, 0.f),
    C_Vertex3f(0.14337635f, 0.0100073814f, 0.f),
    C_Vertex3f(0.162867546f, 0.00170755386f, 0.f),
    C_Vertex3f(0.158373833f, 0.0953469276f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.126324177f, 0.20557642f, 0.f),
    C_Vertex3f(0.131577253f, 0.334079504f, 0.f),
    C_Vertex3f(0.123535156f, 0.189525604f, 0.f),
    C_Vertex3f(0.117303133f, 0.323568583f, 0.f),
    C_Vertex3f(0.123535156f, 0.171806335f, 0.f),
    C_Vertex3f(0.125154495f, 0.0203821659f, 0.f),
    C_Vertex3f(0.12740612f, 0.15195322f, 0.f),
    C_Vertex3f(0.14337635f, 0.0100073814f, 0.f),
    C_Vertex3f(0.135148048f, 0.131717205f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.125154495f, 0.0203821659f, 0.f),
    C_Vertex3f(0.117303133f, 0.323568583f, 0.f),
    C_Vertex3f(0.108201981f, 0.0328319073f, 0.f),
    C_Vertex3f(0.104209304f, 0.312695861f, 0.f),
    C_Vertex3f(0.0925188065f, 0.0473566055f, 0.f),
    C_Vertex3f(0.0922957659f, 0.301461339f, 0.f),
    C_Vertex3f(0.0815625191f, 0.289865017f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0925188065f, 0.0473566055f, 0.f),
    C_Vertex3f(0.0815625191f, 0.289865017f, 0.f),
    C_Vertex3f(0.0794696808f, 0.0624265671f, 0.f),
    C_Vertex3f(0.0720095634f, 0.277906895f, 0.f),
    C_Vertex3f(0.0692458153f, 0.0763628483f, 0.f),
    C_Vertex3f(0.0636507273f, 0.265587211f, 0.f),
    C_Vertex3f(0.060482502f, 0.0906951427f, 0.f),
    C_Vertex3f(0.0564860106f, 0.252905965f, 0.f),
    C_Vertex3f(0.0531797409f, 0.10542345f, 0.f),
    C_Vertex3f(0.0505154133f, 0.239863157f, 0.f),
    C_Vertex3f(0.047337532f, 0.120547771f, 0.f),
    C_Vertex3f(0.0457389355f, 0.226458788f, 0.f),
    C_Vertex3f(0.0429558754f, 0.136068106f, 0.f),
    C_Vertex3f(0.0421565771f, 0.212692857f, 0.f),
    C_Vertex3f(0.040034771f, 0.151984453f, 0.f),
    C_Vertex3f(0.0397683382f, 0.198565364f, 0.f),
    C_Vertex3f(0.0385742188f, 0.184076309f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.425941825f, 0.535222173f, 0.f),
    C_Vertex3f(0.423333287f, 0.548304915f, 0.f),
    C_Vertex3f(0.423446178f, 0.481171131f, 0.f),
    C_Vertex3f(0.427246094f, 0.503170967f, 0.f),
    C_Vertex3f(0.427246094f, 0.521634102f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.419420481f, 0.56088233f, 0.f),
    C_Vertex3f(0.414203405f, 0.572954416f, 0.f),
    C_Vertex3f(0.415846348f, 0.46000433f, 0.f),
    C_Vertex3f(0.423446178f, 0.481171131f, 0.f),
    C_Vertex3f(0.423333287f, 0.548304915f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.407682061f, 0.584521174f, 0.f),
    C_Vertex3f(0.399856448f, 0.595582604f, 0.f),
    C_Vertex3f(0.404446602f, 0.439670563f, 0.f),
    C_Vertex3f(0.415846348f, 0.46000433f, 0.f),
    C_Vertex3f(0.414203405f, 0.572954416f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369114041f, 0.624983907f, 0.f),
    C_Vertex3f(0.370247364f, 0.401502132f, 0.f),
    C_Vertex3f(0.380292416f, 0.61618948f, 0.f),
    C_Vertex3f(0.389246941f, 0.42016983f, 0.f),
    C_Vertex3f(0.390726566f, 0.606138706f, 0.f),
    C_Vertex3f(0.404446602f, 0.439670563f, 0.f),
    C_Vertex3f(0.399856448f, 0.595582604f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.357191443f, 0.632521987f, 0.f),
    C_Vertex3f(0.344524622f, 0.63880372f, 0.f),
    C_Vertex3f(0.346191406f, 0.520820618f, 0.f),
    C_Vertex3f(0.347447872f, 0.383667469f, 0.f),
    C_Vertex3f(0.370247364f, 0.401502132f, 0.f),
    C_Vertex3f(0.369114041f, 0.624983907f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320394516f, 0.571347237f, 0.f),
    C_Vertex3f(0.330713272f, 0.560969353f, 0.f),
    C_Vertex3f(0.331113577f, 0.643829107f, 0.f),
    C_Vertex3f(0.338452339f, 0.549088955f, 0.f),
    C_Vertex3f(0.344524622f, 0.63880372f, 0.f),
    C_Vertex3f(0.343611717f, 0.535706043f, 0.f),
    C_Vertex3f(0.346191406f, 0.520820618f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.23840332f, 0.579212189f, 0.f),
    C_Vertex3f(0.251602173f, 0.584346771f, 0.f),
    C_Vertex3f(0.253219962f, 0.650036693f, 0.f),
    C_Vertex3f(0.266132355f, 0.586914063f, 0.f),
    C_Vertex3f(0.269643784f, 0.651367188f, 0.f),
    C_Vertex3f(0.281776428f, 0.586914063f, 0.f),
    C_Vertex3f(0.2864151f, 0.651367188f, 0.f),
    C_Vertex3f(0.295926094f, 0.584319592f, 0.f),
    C_Vertex3f(0.302058816f, 0.650110841f, 0.f),
    C_Vertex3f(0.30879879f, 0.57913065f, 0.f),
    C_Vertex3f(0.316958308f, 0.647598147f, 0.f),
    C_Vertex3f(0.320394516f, 0.571347237f, 0.f),
    C_Vertex3f(0.331113577f, 0.643829107f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.202829361f, 0.537261009f, 0.f),
    C_Vertex3f(0.208097458f, 0.549824715f, 0.f),
    C_Vertex3f(0.208333731f, 0.638062239f, 0.f),
    C_Vertex3f(0.215999603f, 0.56124115f, 0.f),
    C_Vertex3f(0.222564936f, 0.643384218f, 0.f),
    C_Vertex3f(0.226535797f, 0.571510315f, 0.f),
    C_Vertex3f(0.237527013f, 0.647375703f, 0.f),
    C_Vertex3f(0.23840332f, 0.579212189f, 0.f),
    C_Vertex3f(0.253219962f, 0.650036693f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.181481361f, 0.363441467f, 0.f),
    C_Vertex3f(0.163666368f, 0.354015946f, 0.f),
    C_Vertex3f(0.163688779f, 0.266960025f, 0.f),
    C_Vertex3f(0.173163176f, 0.275621653f, 0.f),
    C_Vertex3f(0.183936357f, 0.284454107f, 0.f),
    C_Vertex3f(0.190979004f, 0.367973328f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.131902218f, 0.221709728f, 0.f),
    C_Vertex3f(0.140269279f, 0.237925529f, 0.f),
    C_Vertex3f(0.147031665f, 0.344228625f, 0.f),
    C_Vertex3f(0.148636341f, 0.25014925f, 0.f),
    C_Vertex3f(0.163666368f, 0.354015946f, 0.f),
    C_Vertex3f(0.155513167f, 0.258469224f, 0.f),
    C_Vertex3f(0.163688779f, 0.266960025f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.131577253f, 0.334079504f, 0.f),
    C_Vertex3f(0.126324177f, 0.20557642f, 0.f),
    C_Vertex3f(0.131902218f, 0.221709728f, 0.f),
    C_Vertex3f(0.147031665f, 0.344228625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.347447872f, 0.383667469f, 0.f),
    C_Vertex3f(0.346191406f, 0.520820618f, 0.f),
    C_Vertex3f(0.346191406f, 0.505743027f, 0.f),
    C_Vertex3f(0.344315529f, 0.492577076f, 0.f),
    C_Vertex3f(0.340563774f, 0.480012417f, 0.f),
    C_Vertex3f(0.334936142f, 0.468049049f, 0.f),
    C_Vertex3f(0.32930851f, 0.459527493f, 0.f),
    C_Vertex3f(0.324434996f, 0.453614354f, 0.f),
    C_Vertex3f(0.320848465f, 0.36666584f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.208333731f, 0.638062239f, 0.f),
    C_Vertex3f(0.200195313f, 0.523550034f, 0.f),
    C_Vertex3f(0.202829361f, 0.537261009f, 0.f),
    C_Vertex3f(0.040034771f, 0.151984453f, 0.f),
    C_Vertex3f(0.0385742188f, 0.184076309f, 0.f),
    C_Vertex3f(0.0385742188f, 0.168296814f, 0.f),
    C_Vertex3f(0.283241034f, 0.420567155f, 0.f),
    C_Vertex3f(0.305648804f, 0.358581543f, 0.f),
    C_Vertex3f(0.293723345f, 0.427640796f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151465893f, 0.418109179f, -1.f),
    C_Vertex3f(0.158649921f, 0.408011198f, -1.f),
    C_Vertex3f(0.158717632f, 0.603469372f, -1.f),
    C_Vertex3f(0.166732073f, 0.39732635f, -1.f),
    C_Vertex3f(0.170025349f, 0.614113331f, -1.f),
    C_Vertex3f(0.175712347f, 0.386054635f, -1.f),
    C_Vertex3f(0.182063937f, 0.623426795f, -1.f),
    C_Vertex3f(0.185590744f, 0.374196053f, -1.f),
    C_Vertex3f(0.194833398f, 0.631409764f, -1.f),
    C_Vertex3f(0.190979004f, 0.367973328f, -1.f),
    C_Vertex3f(0.196008325f, 0.293457389f, -1.f),
    C_Vertex3f(0.183936357f, 0.284454107f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.135302544f, 0.444881916f, -1.f),
    C_Vertex3f(0.139792204f, 0.436544538f, -1.f),
    C_Vertex3f(0.140342593f, 0.580769181f, -1.f),
    C_Vertex3f(0.145179987f, 0.427620292f, -1.f),
    C_Vertex3f(0.14882338f, 0.592354655f, -1.f),
    C_Vertex3f(0.151465893f, 0.418109179f, -1.f),
    C_Vertex3f(0.158717632f, 0.603469372f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12055409f, 0.52971971f, -1.f),
    C_Vertex3f(0.120936394f, 0.486690044f, -1.f),
    C_Vertex3f(0.123381019f, 0.543188214f, -1.f),
    C_Vertex3f(0.124527931f, 0.471735477f, -1.f),
    C_Vertex3f(0.127621412f, 0.556185961f, -1.f),
    C_Vertex3f(0.129915237f, 0.456507683f, -1.f),
    C_Vertex3f(0.13327527f, 0.56871295f, -1.f),
    C_Vertex3f(0.135302544f, 0.444881916f, -1.f),
    C_Vertex3f(0.140342593f, 0.580769181f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.119140625f, 0.515780449f, -1.f),
    C_Vertex3f(0.119140625f, 0.501371384f, -1.f),
    C_Vertex3f(0.120936394f, 0.486690044f, -1.f),
    C_Vertex3f(0.12055409f, 0.52971971f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.533203125f, 0.284431458f, -1.f),
    C_Vertex3f(0.451660156f, 0.3019104f, -1.f),
    C_Vertex3f(0.477027893f, 0.146408081f, -1.f),
    C_Vertex3f(0.481456757f, 0.1531744f, -1.f),
    C_Vertex3f(0.489855528f, 0.16763711f, -1.f),
    C_Vertex3f(0.497795343f, 0.18302989f, -1.f),
    C_Vertex3f(0.505276203f, 0.199352741f, -1.f),
    C_Vertex3f(0.512298107f, 0.216605663f, -1.f),
    C_Vertex3f(0.518861055f, 0.234788656f, -1.f),
    C_Vertex3f(0.524965048f, 0.25390172f, -1.f),
    C_Vertex3f(0.530610085f, 0.273944855f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.536362648f, 0.08015728f, -1.f),
    C_Vertex3f(0.511454582f, 0.104775429f, -1.f),
    C_Vertex3f(0.511960983f, -0.00702667236f, -1.f),
    C_Vertex3f(0.488014221f, 0.131746292f, -1.f),
    C_Vertex3f(0.487647057f, 0.0118656158f, -1.f),
    C_Vertex3f(0.477027893f, 0.146408081f, -1.f),
    C_Vertex3f(0.462948799f, 0.034406662f, -1.f),
    C_Vertex3f(0.451660156f, 0.3019104f, -1.f),
    C_Vertex3f(0.448949814f, 0.289236069f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.442444801f, 0.265418053f, -1.f),
    C_Vertex3f(0.434855461f, 0.243130684f, -1.f),
    C_Vertex3f(0.437866211f, 0.0605964661f, -1.f),
    C_Vertex3f(0.462948799f, 0.034406662f, -1.f),
    C_Vertex3f(0.448949814f, 0.289236069f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.395104885f, 0.0458645821f, -1.f),
    C_Vertex3f(0.41541481f, 0.0647335052f, -1.f),
    C_Vertex3f(0.421302795f, 0.212760925f, -1.f),
    C_Vertex3f(0.425132751f, 0.0755157471f, -1.f),
    C_Vertex3f(0.426181793f, 0.222373962f, -1.f),
    C_Vertex3f(0.437866211f, 0.0605964661f, -1.f),
    C_Vertex3f(0.434855461f, 0.243130684f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37563324f, 0.137496948f, -1.f),
    C_Vertex3f(0.369962692f, 0.128778458f, -1.f),
    C_Vertex3f(0.373920918f, 0.0296912193f, -1.f),
    C_Vertex3f(0.395104885f, 0.0458645821f, -1.f),
    C_Vertex3f(0.421302795f, 0.212760925f, -1.f),
    C_Vertex3f(0.305648804f, 0.358581543f, -1.f),
    C_Vertex3f(0.283241034f, 0.420567155f, -1.f),
    C_Vertex3f(0.271636963f, 0.413261414f, -1.f),
    C_Vertex3f(0.265274048f, 0.409492493f, -1.f),
    C_Vertex3f(0.232032776f, 0.316734314f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.324434996f, 0.453614354f, -1.f),
    C_Vertex3f(0.318439722f, 0.447469115f, -1.f),
    C_Vertex3f(0.320848465f, 0.36666584f, -1.f),
    C_Vertex3f(0.311322689f, 0.441091776f, -1.f),
    C_Vertex3f(0.305648804f, 0.358581543f, -1.f),
    C_Vertex3f(0.303083897f, 0.434482336f, -1.f),
    C_Vertex3f(0.293723345f, 0.427640796f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.265274048f, 0.409492493f, -1.f),
    C_Vertex3f(0.230567932f, 0.450134277f, -1.f),
    C_Vertex3f(0.232032776f, 0.316734314f, -1.f),
    C_Vertex3f(0.226770401f, 0.454723358f, -1.f),
    C_Vertex3f(0.224048615f, 0.311976433f, -1.f),
    C_Vertex3f(0.220125198f, 0.463654995f, -1.f),
    C_Vertex3f(0.214429855f, 0.472340107f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.209684372f, 0.480778694f, -1.f),
    C_Vertex3f(0.203990936f, 0.493066788f, -1.f),
    C_Vertex3f(0.209379077f, 0.302631497f, -1.f),
    C_Vertex3f(0.224048615f, 0.311976433f, -1.f),
    C_Vertex3f(0.214429855f, 0.472340107f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.208333731f, 0.638062239f, -1.f),
    C_Vertex3f(0.194833398f, 0.631409764f, -1.f),
    C_Vertex3f(0.200195313f, 0.523550034f, -1.f),
    C_Vertex3f(0.196008325f, 0.293457389f, -1.f),
    C_Vertex3f(0.200195313f, 0.50868988f, -1.f),
    C_Vertex3f(0.209379077f, 0.302631497f, -1.f),
    C_Vertex3f(0.203990936f, 0.493066788f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.523925781f, -0.0146484375f, -1.f),
    C_Vertex3f(0.576660156f, 0.0479354858f, -1.f),
    C_Vertex3f(0.562738419f, 0.0578918457f, -1.f),
    C_Vertex3f(0.536362648f, 0.08015728f, -1.f),
    C_Vertex3f(0.511960983f, -0.00702667236f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.324655533f, 0.0861120224f, -1.f),
    C_Vertex3f(0.328930855f, 0.00543117523f, -1.f),
    C_Vertex3f(0.341638565f, 0.0987267494f, -1.f),
    C_Vertex3f(0.351862907f, 0.0162134171f, -1.f),
    C_Vertex3f(0.356740952f, 0.112948895f, -1.f),
    C_Vertex3f(0.373920918f, 0.0296912193f, -1.f),
    C_Vertex3f(0.369962692f, 0.128778458f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.251003265f, 0.05859375f, -1.f),
    C_Vertex3f(0.254564285f, -0.0107421875f, -1.f),
    C_Vertex3f(0.268965721f, 0.0613455772f, -1.f),
    C_Vertex3f(0.280281544f, -0.00804662704f, -1.f),
    C_Vertex3f(0.287228584f, 0.0668492317f, -1.f),
    C_Vertex3f(0.3050704f, -0.00265550613f, -1.f),
    C_Vertex3f(0.305791855f, 0.0751047134f, -1.f),
    C_Vertex3f(0.328930855f, 0.00543117523f, -1.f),
    C_Vertex3f(0.324655533f, 0.0861120224f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235084534f, 0.05859375f, -1.f),
    C_Vertex3f(0.221713305f, 0.0599063635f, -1.f),
    C_Vertex3f(0.228957176f, -0.0107421875f, -1.f),
    C_Vertex3f(0.254564285f, -0.0107421875f, -1.f),
    C_Vertex3f(0.251003265f, 0.05859375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.209146261f, 0.0625315905f, -1.f),
    C_Vertex3f(0.197383404f, 0.0664694309f, -1.f),
    C_Vertex3f(0.205657959f, -0.00866723061f, -1.f),
    C_Vertex3f(0.228957176f, -0.0107421875f, -1.f),
    C_Vertex3f(0.221713305f, 0.0599063635f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.186424732f, 0.0717198849f, -1.f),
    C_Vertex3f(0.176270247f, 0.0782829523f, -1.f),
    C_Vertex3f(0.183628082f, -0.00451731682f, -1.f),
    C_Vertex3f(0.205657959f, -0.00866723061f, -1.f),
    C_Vertex3f(0.197383404f, 0.0664694309f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.166919947f, 0.0861586332f, -1.f),
    C_Vertex3f(0.158373833f, 0.0953469276f, -1.f),
    C_Vertex3f(0.162867546f, 0.00170755386f, -1.f),
    C_Vertex3f(0.183628082f, -0.00451731682f, -1.f),
    C_Vertex3f(0.176270247f, 0.0782829523f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.146760941f, 0.111098289f, -1.f),
    C_Vertex3f(0.135148048f, 0.131717205f, -1.f),
    C_Vertex3f(0.14337635f, 0.0100073814f, -1.f),
    C_Vertex3f(0.162867546f, 0.00170755386f, -1.f),
    C_Vertex3f(0.158373833f, 0.0953469276f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.126324177f, 0.20557642f, -1.f),
    C_Vertex3f(0.131577253f, 0.334079504f, -1.f),
    C_Vertex3f(0.123535156f, 0.189525604f, -1.f),
    C_Vertex3f(0.117303133f, 0.323568583f, -1.f),
    C_Vertex3f(0.123535156f, 0.171806335f, -1.f),
    C_Vertex3f(0.125154495f, 0.0203821659f, -1.f),
    C_Vertex3f(0.12740612f, 0.15195322f, -1.f),
    C_Vertex3f(0.14337635f, 0.0100073814f, -1.f),
    C_Vertex3f(0.135148048f, 0.131717205f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.125154495f, 0.0203821659f, -1.f),
    C_Vertex3f(0.117303133f, 0.323568583f, -1.f),
    C_Vertex3f(0.108201981f, 0.0328319073f, -1.f),
    C_Vertex3f(0.104209304f, 0.312695861f, -1.f),
    C_Vertex3f(0.0925188065f, 0.0473566055f, -1.f),
    C_Vertex3f(0.0922957659f, 0.301461339f, -1.f),
    C_Vertex3f(0.0815625191f, 0.289865017f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0925188065f, 0.0473566055f, -1.f),
    C_Vertex3f(0.0815625191f, 0.289865017f, -1.f),
    C_Vertex3f(0.0794696808f, 0.0624265671f, -1.f),
    C_Vertex3f(0.0720095634f, 0.277906895f, -1.f),
    C_Vertex3f(0.0692458153f, 0.0763628483f, -1.f),
    C_Vertex3f(0.0636507273f, 0.265587211f, -1.f),
    C_Vertex3f(0.060482502f, 0.0906951427f, -1.f),
    C_Vertex3f(0.0564860106f, 0.252905965f, -1.f),
    C_Vertex3f(0.0531797409f, 0.10542345f, -1.f),
    C_Vertex3f(0.0505154133f, 0.239863157f, -1.f),
    C_Vertex3f(0.047337532f, 0.120547771f, -1.f),
    C_Vertex3f(0.0457389355f, 0.226458788f, -1.f),
    C_Vertex3f(0.0429558754f, 0.136068106f, -1.f),
    C_Vertex3f(0.0421565771f, 0.212692857f, -1.f),
    C_Vertex3f(0.040034771f, 0.151984453f, -1.f),
    C_Vertex3f(0.0397683382f, 0.198565364f, -1.f),
    C_Vertex3f(0.0385742188f, 0.184076309f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.425941825f, 0.535222173f, -1.f),
    C_Vertex3f(0.423333287f, 0.548304915f, -1.f),
    C_Vertex3f(0.423446178f, 0.481171131f, -1.f),
    C_Vertex3f(0.427246094f, 0.503170967f, -1.f),
    C_Vertex3f(0.427246094f, 0.521634102f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.419420481f, 0.56088233f, -1.f),
    C_Vertex3f(0.414203405f, 0.572954416f, -1.f),
    C_Vertex3f(0.415846348f, 0.46000433f, -1.f),
    C_Vertex3f(0.423446178f, 0.481171131f, -1.f),
    C_Vertex3f(0.423333287f, 0.548304915f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.407682061f, 0.584521174f, -1.f),
    C_Vertex3f(0.399856448f, 0.595582604f, -1.f),
    C_Vertex3f(0.404446602f, 0.439670563f, -1.f),
    C_Vertex3f(0.415846348f, 0.46000433f, -1.f),
    C_Vertex3f(0.414203405f, 0.572954416f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369114041f, 0.624983907f, -1.f),
    C_Vertex3f(0.370247364f, 0.401502132f, -1.f),
    C_Vertex3f(0.380292416f, 0.61618948f, -1.f),
    C_Vertex3f(0.389246941f, 0.42016983f, -1.f),
    C_Vertex3f(0.390726566f, 0.606138706f, -1.f),
    C_Vertex3f(0.404446602f, 0.439670563f, -1.f),
    C_Vertex3f(0.399856448f, 0.595582604f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.357191443f, 0.632521987f, -1.f),
    C_Vertex3f(0.344524622f, 0.63880372f, -1.f),
    C_Vertex3f(0.346191406f, 0.520820618f, -1.f),
    C_Vertex3f(0.347447872f, 0.383667469f, -1.f),
    C_Vertex3f(0.370247364f, 0.401502132f, -1.f),
    C_Vertex3f(0.369114041f, 0.624983907f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320394516f, 0.571347237f, -1.f),
    C_Vertex3f(0.330713272f, 0.560969353f, -1.f),
    C_Vertex3f(0.331113577f, 0.643829107f, -1.f),
    C_Vertex3f(0.338452339f, 0.549088955f, -1.f),
    C_Vertex3f(0.344524622f, 0.63880372f, -1.f),
    C_Vertex3f(0.343611717f, 0.535706043f, -1.f),
    C_Vertex3f(0.346191406f, 0.520820618f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.23840332f, 0.579212189f, -1.f),
    C_Vertex3f(0.251602173f, 0.584346771f, -1.f),
    C_Vertex3f(0.253219962f, 0.650036693f, -1.f),
    C_Vertex3f(0.266132355f, 0.586914063f, -1.f),
    C_Vertex3f(0.269643784f, 0.651367188f, -1.f),
    C_Vertex3f(0.281776428f, 0.586914063f, -1.f),
    C_Vertex3f(0.2864151f, 0.651367188f, -1.f),
    C_Vertex3f(0.295926094f, 0.584319592f, -1.f),
    C_Vertex3f(0.302058816f, 0.650110841f, -1.f),
    C_Vertex3f(0.30879879f, 0.57913065f, -1.f),
    C_Vertex3f(0.316958308f, 0.647598147f, -1.f),
    C_Vertex3f(0.320394516f, 0.571347237f, -1.f),
    C_Vertex3f(0.331113577f, 0.643829107f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.202829361f, 0.537261009f, -1.f),
    C_Vertex3f(0.208097458f, 0.549824715f, -1.f),
    C_Vertex3f(0.208333731f, 0.638062239f, -1.f),
    C_Vertex3f(0.215999603f, 0.56124115f, -1.f),
    C_Vertex3f(0.222564936f, 0.643384218f, -1.f),
    C_Vertex3f(0.226535797f, 0.571510315f, -1.f),
    C_Vertex3f(0.237527013f, 0.647375703f, -1.f),
    C_Vertex3f(0.23840332f, 0.579212189f, -1.f),
    C_Vertex3f(0.253219962f, 0.650036693f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.181481361f, 0.363441467f, -1.f),
    C_Vertex3f(0.163666368f, 0.354015946f, -1.f),
    C_Vertex3f(0.163688779f, 0.266960025f, -1.f),
    C_Vertex3f(0.173163176f, 0.275621653f, -1.f),
    C_Vertex3f(0.183936357f, 0.284454107f, -1.f),
    C_Vertex3f(0.190979004f, 0.367973328f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.131902218f, 0.221709728f, -1.f),
    C_Vertex3f(0.140269279f, 0.237925529f, -1.f),
    C_Vertex3f(0.147031665f, 0.344228625f, -1.f),
    C_Vertex3f(0.148636341f, 0.25014925f, -1.f),
    C_Vertex3f(0.163666368f, 0.354015946f, -1.f),
    C_Vertex3f(0.155513167f, 0.258469224f, -1.f),
    C_Vertex3f(0.163688779f, 0.266960025f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.131577253f, 0.334079504f, -1.f),
    C_Vertex3f(0.126324177f, 0.20557642f, -1.f),
    C_Vertex3f(0.131902218f, 0.221709728f, -1.f),
    C_Vertex3f(0.147031665f, 0.344228625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.347447872f, 0.383667469f, -1.f),
    C_Vertex3f(0.346191406f, 0.520820618f, -1.f),
    C_Vertex3f(0.346191406f, 0.505743027f, -1.f),
    C_Vertex3f(0.344315529f, 0.492577076f, -1.f),
    C_Vertex3f(0.340563774f, 0.480012417f, -1.f),
    C_Vertex3f(0.334936142f, 0.468049049f, -1.f),
    C_Vertex3f(0.32930851f, 0.459527493f, -1.f),
    C_Vertex3f(0.324434996f, 0.453614354f, -1.f),
    C_Vertex3f(0.320848465f, 0.36666584f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.208333731f, 0.638062239f, -1.f),
    C_Vertex3f(0.200195313f, 0.523550034f, -1.f),
    C_Vertex3f(0.202829361f, 0.537261009f, -1.f),
    C_Vertex3f(0.040034771f, 0.151984453f, -1.f),
    C_Vertex3f(0.0385742188f, 0.184076309f, -1.f),
    C_Vertex3f(0.0385742188f, 0.168296814f, -1.f),
    C_Vertex3f(0.283241034f, 0.420567155f, -1.f),
    C_Vertex3f(0.305648804f, 0.358581543f, -1.f),
    C_Vertex3f(0.293723345f, 0.427640796f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_39[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.986297071f, -0.164979234f, 0.f),
    C_Vertex3f(0.0596008301f, 0.413574219f, 0.f),
    C_Vertex3f(0.0596008301f, 0.413574219f, -1.f),
    C_Vertex3f(0.0395507813f, 0.533439636f, -1.f),
    C_Vertex3f(0.0395507813f, 0.533439636f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0395507813f, 0.533439636f, 0.f),
    C_Vertex3f(0.0395507813f, 0.533439636f, -1.f),
    C_Vertex3f(0.0395507813f, 0.640625f, -1.f),
    C_Vertex3f(0.0395507813f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0395507813f, 0.640625f, 0.f),
    C_Vertex3f(0.0395507813f, 0.640625f, -1.f),
    C_Vertex3f(0.12890625f, 0.640625f, -1.f),
    C_Vertex3f(0.12890625f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.12890625f, 0.640625f, 0.f),
    C_Vertex3f(0.12890625f, 0.640625f, -1.f),
    C_Vertex3f(0.12890625f, 0.533439636f, -1.f),
    C_Vertex3f(0.12890625f, 0.533439636f, 0.f),
    C_Normal3f(0.985109389f, -0.171928585f, 0.f),
    C_Vertex3f(0.12890625f, 0.533439636f, 0.f),
    C_Vertex3f(0.12890625f, 0.533439636f, -1.f),
    C_Vertex3f(0.10798645f, 0.413574219f, -1.f),
    C_Vertex3f(0.10798645f, 0.413574219f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.10798645f, 0.413574219f, 0.f),
    C_Vertex3f(0.10798645f, 0.413574219f, -1.f),
    C_Vertex3f(0.0596008301f, 0.413574219f, -1.f),
    C_Vertex3f(0.0596008301f, 0.413574219f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0596008301f, 0.413574219f, 0.f),
    C_Vertex3f(0.10798645f, 0.413574219f, 0.f),
    C_Vertex3f(0.12890625f, 0.533439636f, 0.f),
    C_Vertex3f(0.12890625f, 0.640625f, 0.f),
    C_Vertex3f(0.0395507813f, 0.640625f, 0.f),
    C_Vertex3f(0.0395507813f, 0.533439636f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0596008301f, 0.413574219f, -1.f),
    C_Vertex3f(0.10798645f, 0.413574219f, -1.f),
    C_Vertex3f(0.12890625f, 0.533439636f, -1.f),
    C_Vertex3f(0.12890625f, 0.640625f, -1.f),
    C_Vertex3f(0.0395507813f, 0.640625f, -1.f),
    C_Vertex3f(0.0395507813f, 0.533439636f, -1.f),
    C_End(),
    C_Translatef(0.170898438f, 0.f, 0.f),
};
static const Call char_40[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.783403635f, -0.621513307f, 0.f),
    C_Vertex3f(0.209472656f, -0.188476563f, 0.f),
    C_Vertex3f(0.209472656f, -0.188476563f, -1.f),
    C_Normal3f(-0.799380183f, -0.600825608f, 0.f),
    C_Vertex3f(0.20132637f, -0.178208351f, 0.f),
    C_Vertex3f(0.20132637f, -0.178208351f, -1.f),
    C_Normal3f(-0.818240702f, -0.574875832f, 0.f),
    C_Vertex3f(0.185662508f, -0.156797886f, 0.f),
    C_Vertex3f(0.185662508f, -0.156797886f, -1.f),
    C_Normal3f(-0.839256108f, -0.543736279f, 0.f),
    C_Vertex3f(0.170627356f, -0.134513378f, 0.f),
    C_Vertex3f(0.170627356f, -0.134513378f, -1.f),
    C_Normal3f(-0.858538628f, -0.512748837f, 0.f),
    C_Vertex3f(0.156220913f, -0.111354828f, 0.f),
    C_Vertex3f(0.156220913f, -0.111354828f, -1.f),
    C_Normal3f(-0.876138926f, -0.482058674f, 0.f),
    C_Vertex3f(0.14244318f, -0.0873222351f, 0.f),
    C_Vertex3f(0.14244318f, -0.0873222351f, -1.f),
    C_Normal3f(-0.892122626f, -0.451793462f, 0.f),
    C_Vertex3f(0.129294157f, -0.0624155998f, 0.f),
    C_Vertex3f(0.129294157f, -0.0624155998f, -1.f),
    C_Normal3f(-0.90656656f, -0.422062725f, 0.f),
    C_Vertex3f(0.116773844f, -0.036634922f, 0.f),
    C_Vertex3f(0.116773844f, -0.036634922f, -1.f),
    C_Normal3f(-0.919556201f, -0.392958522f, 0.f),
    C_Vertex3f(0.10488224f, -0.00998020172f, 0.f),
    C_Vertex3f(0.10488224f, -0.00998020172f, -1.f),
    C_Normal3f(-0.934206545f, -0.356732637f, 0.f),
    C_Vertex3f(0.0936193466f, 0.0175485611f, 0.f),
    C_Vertex3f(0.0936193466f, 0.0175485611f, -1.f),
    C_Normal3f(-0.950185239f, -0.311685562f, 0.f),
    C_Vertex3f(0.0837643147f, 0.0453231335f, 0.f),
    C_Vertex3f(0.0837643147f, 0.0453231335f, -1.f),
    C_Normal3f(-0.964171767f, -0.265278637f, 0.f),
    C_Vertex3f(0.0753171444f, 0.0733435154f, 0.f),
    C_Vertex3f(0.0753171444f, 0.0733435154f, -1.f),
    C_Normal3f(-0.975992382f, -0.21780476f, 0.f),
    C_Vertex3f(0.0682778358f, 0.101609707f, 0.f),
    C_Vertex3f(0.0682778358f, 0.101609707f, -1.f),
    C_Normal3f(-0.985514998f, -0.169588149f, 0.f),
    C_Vertex3f(0.062646389f, 0.130121708f, 0.f),
    C_Vertex3f(0.062646389f, 0.130121708f, -1.f),
    C_Normal3f(-0.992655694f, -0.120973632f, 0.f),
    C_Vertex3f(0.0584228039f, 0.158879519f, 0.f),
    C_Vertex3f(0.0584228039f, 0.158879519f, -1.f),
    C_Normal3f(-0.997381926f, -0.0723142847f, 0.f),
    C_Vertex3f(0.0556070805f, 0.187883139f, 0.f),
    C_Vertex3f(0.0556070805f, 0.187883139f, -1.f),
    C_Normal3f(-0.999695003f, -0.0246932749f, 0.f),
    C_Vertex3f(0.0541992188f, 0.217132568f, 0.f),
    C_Vertex3f(0.0541992188f, 0.217132568f, -1.f),
    C_Normal3f(-0.999806404f, 0.0196766835f, 0.f),
    C_Vertex3f(0.0541992188f, 0.244879723f, 0.f),
    C_Vertex3f(0.0541992188f, 0.244879723f, -1.f),
    C_Normal3f(-0.998102665f, 0.0615709201f, 0.f),
    C_Vertex3f(0.0552515984f, 0.270605803f, 0.f),
    C_Vertex3f(0.0552515984f, 0.270605803f, -1.f),
    C_Normal3f(-0.994643509f, 0.10336525f, 0.f),
    C_Vertex3f(0.0573563576f, 0.296058893f, 0.f),
    C_Vertex3f(0.0573563576f, 0.296058893f, -1.f),
    C_Normal3f(-0.989356577f, 0.14551121f, 0.f),
    C_Vertex3f(0.0605134964f, 0.321238995f, 0.f),
    C_Vertex3f(0.0605134964f, 0.321238995f, -1.f),
    C_Normal3f(-0.982210875f, 0.187781513f, 0.f),
    C_Vertex3f(0.0647230148f, 0.346146107f, 0.f),
    C_Vertex3f(0.0647230148f, 0.346146107f, -1.f),
    C_Normal3f(-0.973204792f, 0.229940414f, 0.f),
    C_Vertex3f(0.0699849129f, 0.37078023f, 0.f),
    C_Vertex3f(0.0699849129f, 0.37078023f, -1.f),
    C_Normal3f(-0.962367892f, 0.271750033f, 0.f),
    C_Vertex3f(0.0762991905f, 0.395141363f, 0.f),
    C_Vertex3f(0.0762991905f, 0.395141363f, -1.f),
    C_Normal3f(-0.94929719f, 0.314380258f, 0.f),
    C_Vertex3f(0.0836658478f, 0.419229507f, 0.f),
    C_Vertex3f(0.0836658478f, 0.419229507f, -1.f),
    C_Normal3f(-0.934527814f, 0.355890036f, 0.f),
    C_Vertex3f(0.0927963257f, 0.444955826f, 0.f),
    C_Vertex3f(0.0927963257f, 0.444955826f, -1.f),
    C_Normal3f(-0.918856561f, 0.394591779f, 0.f),
    C_Vertex3f(0.10397768f, 0.472566128f, 0.f),
    C_Vertex3f(0.10397768f, 0.472566128f, -1.f),
    C_Normal3f(-0.902069449f, 0.431590706f, 0.f),
    C_Vertex3f(0.11649847f, 0.50014925f, 0.f),
    C_Vertex3f(0.11649847f, 0.50014925f, -1.f),
    C_Normal3f(-0.884468198f, 0.466600388f, 0.f),
    C_Vertex3f(0.130358696f, 0.527705193f, 0.f),
    C_Vertex3f(0.130358696f, 0.527705193f, -1.f),
    C_Normal3f(-0.866250396f, 0.499610037f, 0.f),
    C_Vertex3f(0.145558357f, 0.555233955f, 0.f),
    C_Vertex3f(0.145558357f, 0.555233955f, -1.f),
    C_Normal3f(-0.847598672f, 0.530637801f, 0.f),
    C_Vertex3f(0.162097454f, 0.582735538f, 0.f),
    C_Vertex3f(0.162097454f, 0.582735538f, -1.f),
    C_Normal3f(-0.828677893f, 0.559725761f, 0.f),
    C_Vertex3f(0.179975986f, 0.610209942f, 0.f),
    C_Vertex3f(0.179975986f, 0.610209942f, -1.f),
    C_Normal3f(-0.812811077f, 0.582527339f, 0.f),
    C_Vertex3f(0.199193954f, 0.637657166f, 0.f),
    C_Vertex3f(0.199193954f, 0.637657166f, -1.f),
    C_Normal3f(-0.800106883f, 0.599857569f, 0.f),
    C_Vertex3f(0.209472656f, 0.651367188f, 0.f),
    C_Vertex3f(0.209472656f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.209472656f, 0.651367188f, 0.f),
    C_Vertex3f(0.209472656f, 0.651367188f, -1.f),
    C_Vertex3f(0.265625f, 0.651367188f, -1.f),
    C_Vertex3f(0.265625f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.864859104f, -0.502014637f, 0.f),
    C_Vertex3f(0.265625f, 0.651367188f, 0.f),
    C_Vertex3f(0.265625f, 0.651367188f, -1.f),
    C_Normal3f(0.866869211f, -0.498535752f, 0.f),
    C_Vertex3f(0.259030342f, 0.640006065f, 0.f),
    C_Vertex3f(0.259030342f, 0.640006065f, -1.f),
    C_Normal3f(0.869675517f, -0.493623674f, 0.f),
    C_Vertex3f(0.246958256f, 0.618908882f, 0.f),
    C_Vertex3f(0.246958256f, 0.618908882f, -1.f),
    C_Normal3f(0.873564363f, -0.48670876f, 0.f),
    C_Vertex3f(0.236003399f, 0.59943676f, 0.f),
    C_Vertex3f(0.236003399f, 0.59943676f, -1.f),
    C_Normal3f(0.878167391f, -0.478353441f, 0.f),
    C_Vertex3f(0.226165771f, 0.581589699f, 0.f),
    C_Vertex3f(0.226165771f, 0.581589699f, -1.f),
    C_Normal3f(0.883697152f, -0.468059242f, 0.f),
    C_Vertex3f(0.217445374f, 0.565367699f, 0.f),
    C_Vertex3f(0.217445374f, 0.565367699f, -1.f),
    C_Normal3f(0.890455842f, -0.455069512f, 0.f),
    C_Vertex3f(0.209842205f, 0.55077076f, 0.f),
    C_Vertex3f(0.209842205f, 0.55077076f, -1.f),
    C_Normal3f(0.89888674f, -0.438180864f, 0.f),
    C_Vertex3f(0.203356266f, 0.537798882f, 0.f),
    C_Vertex3f(0.203356266f, 0.537798882f, -1.f),
    C_Normal3f(0.911201239f, -0.411961406f, 0.f),
    C_Vertex3f(0.197987556f, 0.526452065f, 0.f),
    C_Vertex3f(0.197987556f, 0.526452065f, -1.f),
    C_Normal3f(0.926197827f, -0.377038121f, 0.f),
    C_Vertex3f(0.189212799f, 0.506515503f, 0.f),
    C_Vertex3f(0.189212799f, 0.506515503f, -1.f),
    C_Normal3f(0.941219926f, -0.337794334f, 0.f),
    C_Vertex3f(0.177330971f, 0.475708961f, 0.f),
    C_Vertex3f(0.177330971f, 0.475708961f, -1.f),
    C_Normal3f(0.955980062f, -0.293431818f, 0.f),
    C_Vertex3f(0.166865349f, 0.444247246f, 0.f),
    C_Vertex3f(0.166865349f, 0.444247246f, -1.f),
    C_Normal3f(0.96907568f, -0.246763766f, 0.f),
    C_Vertex3f(0.157815933f, 0.412130356f, 0.f),
    C_Vertex3f(0.157815933f, 0.412130356f, -1.f),
    C_Normal3f(0.980637014f, -0.195834011f, 0.f),
    C_Vertex3f(0.149312973f, 0.37531662f, 0.f),
    C_Vertex3f(0.149312973f, 0.37531662f, -1.f),
    C_Normal3f(0.989956379f, -0.141373023f, 0.f),
    C_Vertex3f(0.14228344f, 0.33435154f, 0.f),
    C_Vertex3f(0.14228344f, 0.33435154f, -1.f),
    C_Normal3f(0.996368527f, -0.0851455852f, 0.f),
    C_Vertex3f(0.137597084f, 0.293276787f, 0.f),
    C_Vertex3f(0.137597084f, 0.293276787f, -1.f),
    C_Normal3f(0.999511838f, -0.0312405061f, 0.f),
    C_Vertex3f(0.135253906f, 0.252092361f, 0.f),
    C_Vertex3f(0.135253906f, 0.252092361f, -1.f),
    C_Normal3f(0.999856174f, 0.0169580523f, 0.f),
    C_Vertex3f(0.135253906f, 0.218308926f, 0.f),
    C_Vertex3f(0.135253906f, 0.218308926f, -1.f),
    C_Normal3f(0.998312831f, 0.0580642819f, 0.f),
    C_Vertex3f(0.13627243f, 0.192039609f, 0.f),
    C_Vertex3f(0.13627243f, 0.192039609f, -1.f),
    C_Normal3f(0.995333135f, 0.0964976624f, 0.f),
    C_Vertex3f(0.138309479f, 0.165773749f, 0.f),
    C_Vertex3f(0.138309479f, 0.165773749f, -1.f),
    C_Normal3f(0.990911663f, 0.134514302f, 0.f),
    C_Vertex3f(0.141365051f, 0.139511347f, 0.f),
    C_Vertex3f(0.141365051f, 0.139511347f, -1.f),
    C_Normal3f(0.98510468f, 0.171956077f, 0.f),
    C_Vertex3f(0.145439148f, 0.113252401f, 0.f),
    C_Vertex3f(0.145439148f, 0.113252401f, -1.f),
    C_Normal3f(0.977984607f, 0.208676979f, 0.f),
    C_Vertex3f(0.150531769f, 0.086996913f, 0.f),
    C_Vertex3f(0.150531769f, 0.086996913f, -1.f),
    C_Normal3f(0.969637811f, 0.244545639f, 0.f),
    C_Vertex3f(0.156642914f, 0.0607448816f, 0.f),
    C_Vertex3f(0.156642914f, 0.0607448816f, -1.f),
    C_Normal3f(0.96016115f, 0.279447109f, 0.f),
    C_Vertex3f(0.163772583f, 0.0344963074f, 0.f),
    C_Vertex3f(0.163772583f, 0.0344963074f, -1.f),
    C_Normal3f(0.949659526f, 0.313284039f, 0.f),
    C_Vertex3f(0.171920776f, 0.00825119019f, 0.f),
    C_Vertex3f(0.171920776f, 0.00825119019f, -1.f),
    C_Normal3f(0.938242912f, 0.345977247f, 0.f),
    C_Vertex3f(0.181087494f, -0.0179904699f, 0.f),
    C_Vertex3f(0.181087494f, -0.0179904699f, -1.f),
    C_Normal3f(0.926023602f, 0.377465457f, 0.f),
    C_Vertex3f(0.191272736f, -0.044228673f, 0.f),
    C_Vertex3f(0.191272736f, -0.044228673f, -1.f),
    C_Normal3f(0.913113832f, 0.40770483f, 0.f),
    C_Vertex3f(0.202476501f, -0.070463419f, 0.f),
    C_Vertex3f(0.202476501f, -0.070463419f, -1.f),
    C_Normal3f(0.899622977f, 0.436667502f, 0.f),
    C_Vertex3f(0.214698792f, -0.0966947079f, 0.f),
    C_Vertex3f(0.214698792f, -0.0966947079f, -1.f),
    C_Normal3f(0.885656595f, 0.464340657f, 0.f),
    C_Vertex3f(0.227939606f, -0.12292254f, 0.f),
    C_Vertex3f(0.227939606f, -0.12292254f, -1.f),
    C_Normal3f(0.871314704f, 0.490724653f, 0.f),
    C_Vertex3f(0.242198944f, -0.149146914f, 0.f),
    C_Vertex3f(0.242198944f, -0.149146914f, -1.f),
    C_Normal3f(0.859143555f, 0.511734664f, 0.f),
    C_Vertex3f(0.257476807f, -0.175367832f, 0.f),
    C_Vertex3f(0.257476807f, -0.175367832f, -1.f),
    C_Normal3f(0.849299133f, 0.527911842f, 0.f),
    C_Vertex3f(0.265625f, -0.188476563f, 0.f),
    C_Vertex3f(0.265625f, -0.188476563f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.265625f, -0.188476563f, 0.f),
    C_Vertex3f(0.265625f, -0.188476563f, -1.f),
    C_Vertex3f(0.209472656f, -0.188476563f, -1.f),
    C_Vertex3f(0.209472656f, -0.188476563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.209472656f, -0.188476563f, 0.f),
    C_Vertex3f(0.265625f, -0.188476563f, 0.f),
    C_Vertex3f(0.257476807f, -0.175367832f, 0.f),
    C_Vertex3f(0.242198944f, -0.149146914f, 0.f),
    C_Vertex3f(0.227939606f, -0.12292254f, 0.f),
    C_Vertex3f(0.214698792f, -0.0966947079f, 0.f),
    C_Vertex3f(0.202476501f, -0.070463419f, 0.f),
    C_Vertex3f(0.20132637f, -0.178208351f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191272736f, -0.044228673f, 0.f),
    C_Vertex3f(0.181087494f, -0.0179904699f, 0.f),
    C_Vertex3f(0.185662508f, -0.156797886f, 0.f),
    C_Vertex3f(0.20132637f, -0.178208351f, 0.f),
    C_Vertex3f(0.202476501f, -0.070463419f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.171920776f, 0.00825119019f, 0.f),
    C_Vertex3f(0.163772583f, 0.0344963074f, 0.f),
    C_Vertex3f(0.170627356f, -0.134513378f, 0.f),
    C_Vertex3f(0.185662508f, -0.156797886f, 0.f),
    C_Vertex3f(0.181087494f, -0.0179904699f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156642914f, 0.0607448816f, 0.f),
    C_Vertex3f(0.150531769f, 0.086996913f, 0.f),
    C_Vertex3f(0.156220913f, -0.111354828f, 0.f),
    C_Vertex3f(0.170627356f, -0.134513378f, 0.f),
    C_Vertex3f(0.163772583f, 0.0344963074f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145439148f, 0.113252401f, 0.f),
    C_Vertex3f(0.141365051f, 0.139511347f, 0.f),
    C_Vertex3f(0.14244318f, -0.0873222351f, 0.f),
    C_Vertex3f(0.156220913f, -0.111354828f, 0.f),
    C_Vertex3f(0.150531769f, 0.086996913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14244318f, -0.0873222351f, 0.f),
    C_Vertex3f(0.141365051f, 0.139511347f, 0.f),
    C_Vertex3f(0.138309479f, 0.165773749f, 0.f),
    C_Vertex3f(0.13627243f, 0.192039609f, 0.f),
    C_Vertex3f(0.135253906f, 0.218308926f, 0.f),
    C_Vertex3f(0.129294157f, -0.0624155998f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0836658478f, 0.419229507f, 0.f),
    C_Vertex3f(0.0837643147f, 0.0453231335f, 0.f),
    C_Vertex3f(0.0927963257f, 0.444955826f, 0.f),
    C_Vertex3f(0.0936193466f, 0.0175485611f, 0.f),
    C_Vertex3f(0.10397768f, 0.472566128f, 0.f),
    C_Vertex3f(0.10488224f, -0.00998020172f, 0.f),
    C_Vertex3f(0.11649847f, 0.50014925f, 0.f),
    C_Vertex3f(0.116773844f, -0.036634922f, 0.f),
    C_Vertex3f(0.130358696f, 0.527705193f, 0.f),
    C_Vertex3f(0.129294157f, -0.0624155998f, 0.f),
    C_Vertex3f(0.135253906f, 0.218308926f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0552515984f, 0.270605803f, 0.f),
    C_Vertex3f(0.0556070805f, 0.187883139f, 0.f),
    C_Vertex3f(0.0573563576f, 0.296058893f, 0.f),
    C_Vertex3f(0.0584228039f, 0.158879519f, 0.f),
    C_Vertex3f(0.0605134964f, 0.321238995f, 0.f),
    C_Vertex3f(0.062646389f, 0.130121708f, 0.f),
    C_Vertex3f(0.0647230148f, 0.346146107f, 0.f),
    C_Vertex3f(0.0682778358f, 0.101609707f, 0.f),
    C_Vertex3f(0.0699849129f, 0.37078023f, 0.f),
    C_Vertex3f(0.0753171444f, 0.0733435154f, 0.f),
    C_Vertex3f(0.0762991905f, 0.395141363f, 0.f),
    C_Vertex3f(0.0837643147f, 0.0453231335f, 0.f),
    C_Vertex3f(0.0836658478f, 0.419229507f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0541992188f, 0.244879723f, 0.f),
    C_Vertex3f(0.0541992188f, 0.217132568f, 0.f),
    C_Vertex3f(0.0556070805f, 0.187883139f, 0.f),
    C_Vertex3f(0.0552515984f, 0.270605803f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.209472656f, 0.651367188f, 0.f),
    C_Vertex3f(0.199193954f, 0.637657166f, 0.f),
    C_Vertex3f(0.203356266f, 0.537798882f, 0.f),
    C_Vertex3f(0.209842205f, 0.55077076f, 0.f),
    C_Vertex3f(0.217445374f, 0.565367699f, 0.f),
    C_Vertex3f(0.226165771f, 0.581589699f, 0.f),
    C_Vertex3f(0.236003399f, 0.59943676f, 0.f),
    C_Vertex3f(0.246958256f, 0.618908882f, 0.f),
    C_Vertex3f(0.259030342f, 0.640006065f, 0.f),
    C_Vertex3f(0.265625f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.166865349f, 0.444247246f, 0.f),
    C_Vertex3f(0.177330971f, 0.475708961f, 0.f),
    C_Vertex3f(0.179975986f, 0.610209942f, 0.f),
    C_Vertex3f(0.189212799f, 0.506515503f, 0.f),
    C_Vertex3f(0.199193954f, 0.637657166f, 0.f),
    C_Vertex3f(0.197987556f, 0.526452065f, 0.f),
    C_Vertex3f(0.203356266f, 0.537798882f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162097454f, 0.582735538f, 0.f),
    C_Vertex3f(0.145558357f, 0.555233955f, 0.f),
    C_Vertex3f(0.149312973f, 0.37531662f, 0.f),
    C_Vertex3f(0.157815933f, 0.412130356f, 0.f),
    C_Vertex3f(0.166865349f, 0.444247246f, 0.f),
    C_Vertex3f(0.179975986f, 0.610209942f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145558357f, 0.555233955f, 0.f),
    C_Vertex3f(0.130358696f, 0.527705193f, 0.f),
    C_Vertex3f(0.135253906f, 0.252092361f, 0.f),
    C_Vertex3f(0.137597084f, 0.293276787f, 0.f),
    C_Vertex3f(0.14228344f, 0.33435154f, 0.f),
    C_Vertex3f(0.149312973f, 0.37531662f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.135253906f, 0.252092361f, 0.f),
    C_Vertex3f(0.130358696f, 0.527705193f, 0.f),
    C_Vertex3f(0.135253906f, 0.218308926f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.209472656f, -0.188476563f, -1.f),
    C_Vertex3f(0.265625f, -0.188476563f, -1.f),
    C_Vertex3f(0.257476807f, -0.175367832f, -1.f),
    C_Vertex3f(0.242198944f, -0.149146914f, -1.f),
    C_Vertex3f(0.227939606f, -0.12292254f, -1.f),
    C_Vertex3f(0.214698792f, -0.0966947079f, -1.f),
    C_Vertex3f(0.202476501f, -0.070463419f, -1.f),
    C_Vertex3f(0.20132637f, -0.178208351f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191272736f, -0.044228673f, -1.f),
    C_Vertex3f(0.181087494f, -0.0179904699f, -1.f),
    C_Vertex3f(0.185662508f, -0.156797886f, -1.f),
    C_Vertex3f(0.20132637f, -0.178208351f, -1.f),
    C_Vertex3f(0.202476501f, -0.070463419f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.171920776f, 0.00825119019f, -1.f),
    C_Vertex3f(0.163772583f, 0.0344963074f, -1.f),
    C_Vertex3f(0.170627356f, -0.134513378f, -1.f),
    C_Vertex3f(0.185662508f, -0.156797886f, -1.f),
    C_Vertex3f(0.181087494f, -0.0179904699f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156642914f, 0.0607448816f, -1.f),
    C_Vertex3f(0.150531769f, 0.086996913f, -1.f),
    C_Vertex3f(0.156220913f, -0.111354828f, -1.f),
    C_Vertex3f(0.170627356f, -0.134513378f, -1.f),
    C_Vertex3f(0.163772583f, 0.0344963074f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145439148f, 0.113252401f, -1.f),
    C_Vertex3f(0.141365051f, 0.139511347f, -1.f),
    C_Vertex3f(0.14244318f, -0.0873222351f, -1.f),
    C_Vertex3f(0.156220913f, -0.111354828f, -1.f),
    C_Vertex3f(0.150531769f, 0.086996913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14244318f, -0.0873222351f, -1.f),
    C_Vertex3f(0.141365051f, 0.139511347f, -1.f),
    C_Vertex3f(0.138309479f, 0.165773749f, -1.f),
    C_Vertex3f(0.13627243f, 0.192039609f, -1.f),
    C_Vertex3f(0.135253906f, 0.218308926f, -1.f),
    C_Vertex3f(0.129294157f, -0.0624155998f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0836658478f, 0.419229507f, -1.f),
    C_Vertex3f(0.0837643147f, 0.0453231335f, -1.f),
    C_Vertex3f(0.0927963257f, 0.444955826f, -1.f),
    C_Vertex3f(0.0936193466f, 0.0175485611f, -1.f),
    C_Vertex3f(0.10397768f, 0.472566128f, -1.f),
    C_Vertex3f(0.10488224f, -0.00998020172f, -1.f),
    C_Vertex3f(0.11649847f, 0.50014925f, -1.f),
    C_Vertex3f(0.116773844f, -0.036634922f, -1.f),
    C_Vertex3f(0.130358696f, 0.527705193f, -1.f),
    C_Vertex3f(0.129294157f, -0.0624155998f, -1.f),
    C_Vertex3f(0.135253906f, 0.218308926f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0552515984f, 0.270605803f, -1.f),
    C_Vertex3f(0.0556070805f, 0.187883139f, -1.f),
    C_Vertex3f(0.0573563576f, 0.296058893f, -1.f),
    C_Vertex3f(0.0584228039f, 0.158879519f, -1.f),
    C_Vertex3f(0.0605134964f, 0.321238995f, -1.f),
    C_Vertex3f(0.062646389f, 0.130121708f, -1.f),
    C_Vertex3f(0.0647230148f, 0.346146107f, -1.f),
    C_Vertex3f(0.0682778358f, 0.101609707f, -1.f),
    C_Vertex3f(0.0699849129f, 0.37078023f, -1.f),
    C_Vertex3f(0.0753171444f, 0.0733435154f, -1.f),
    C_Vertex3f(0.0762991905f, 0.395141363f, -1.f),
    C_Vertex3f(0.0837643147f, 0.0453231335f, -1.f),
    C_Vertex3f(0.0836658478f, 0.419229507f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0541992188f, 0.244879723f, -1.f),
    C_Vertex3f(0.0541992188f, 0.217132568f, -1.f),
    C_Vertex3f(0.0556070805f, 0.187883139f, -1.f),
    C_Vertex3f(0.0552515984f, 0.270605803f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.209472656f, 0.651367188f, -1.f),
    C_Vertex3f(0.199193954f, 0.637657166f, -1.f),
    C_Vertex3f(0.203356266f, 0.537798882f, -1.f),
    C_Vertex3f(0.209842205f, 0.55077076f, -1.f),
    C_Vertex3f(0.217445374f, 0.565367699f, -1.f),
    C_Vertex3f(0.226165771f, 0.581589699f, -1.f),
    C_Vertex3f(0.236003399f, 0.59943676f, -1.f),
    C_Vertex3f(0.246958256f, 0.618908882f, -1.f),
    C_Vertex3f(0.259030342f, 0.640006065f, -1.f),
    C_Vertex3f(0.265625f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.166865349f, 0.444247246f, -1.f),
    C_Vertex3f(0.177330971f, 0.475708961f, -1.f),
    C_Vertex3f(0.179975986f, 0.610209942f, -1.f),
    C_Vertex3f(0.189212799f, 0.506515503f, -1.f),
    C_Vertex3f(0.199193954f, 0.637657166f, -1.f),
    C_Vertex3f(0.197987556f, 0.526452065f, -1.f),
    C_Vertex3f(0.203356266f, 0.537798882f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162097454f, 0.582735538f, -1.f),
    C_Vertex3f(0.145558357f, 0.555233955f, -1.f),
    C_Vertex3f(0.149312973f, 0.37531662f, -1.f),
    C_Vertex3f(0.157815933f, 0.412130356f, -1.f),
    C_Vertex3f(0.166865349f, 0.444247246f, -1.f),
    C_Vertex3f(0.179975986f, 0.610209942f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145558357f, 0.555233955f, -1.f),
    C_Vertex3f(0.130358696f, 0.527705193f, -1.f),
    C_Vertex3f(0.135253906f, 0.252092361f, -1.f),
    C_Vertex3f(0.137597084f, 0.293276787f, -1.f),
    C_Vertex3f(0.14228344f, 0.33435154f, -1.f),
    C_Vertex3f(0.149312973f, 0.37531662f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.135253906f, 0.252092361f, -1.f),
    C_Vertex3f(0.130358696f, 0.527705193f, -1.f),
    C_Vertex3f(0.135253906f, 0.218308926f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_41[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.110351563f, -0.188476563f, 0.f),
    C_Vertex3f(0.110351563f, -0.188476563f, -1.f),
    C_Vertex3f(0.0541992188f, -0.188476563f, -1.f),
    C_Vertex3f(0.0541992188f, -0.188476563f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.849299133f, 0.527911842f, 0.f),
    C_Vertex3f(0.0541992188f, -0.188476563f, 0.f),
    C_Vertex3f(0.0541992188f, -0.188476563f, -1.f),
    C_Normal3f(-0.859143555f, 0.511734664f, 0.f),
    C_Vertex3f(0.0623474121f, -0.175367832f, 0.f),
    C_Vertex3f(0.0623474121f, -0.175367832f, -1.f),
    C_Normal3f(-0.871314704f, 0.490724653f, 0.f),
    C_Vertex3f(0.0776252747f, -0.149146914f, 0.f),
    C_Vertex3f(0.0776252747f, -0.149146914f, -1.f),
    C_Normal3f(-0.885656595f, 0.464340657f, 0.f),
    C_Vertex3f(0.091884613f, -0.12292254f, 0.f),
    C_Vertex3f(0.091884613f, -0.12292254f, -1.f),
    C_Normal3f(-0.899622977f, 0.436667502f, 0.f),
    C_Vertex3f(0.105125427f, -0.0966947079f, 0.f),
    C_Vertex3f(0.105125427f, -0.0966947079f, -1.f),
    C_Normal3f(-0.913113832f, 0.40770483f, 0.f),
    C_Vertex3f(0.117347717f, -0.070463419f, 0.f),
    C_Vertex3f(0.117347717f, -0.070463419f, -1.f),
    C_Normal3f(-0.926023602f, 0.377465457f, 0.f),
    C_Vertex3f(0.128551483f, -0.044228673f, 0.f),
    C_Vertex3f(0.128551483f, -0.044228673f, -1.f),
    C_Normal3f(-0.938242912f, 0.345977247f, 0.f),
    C_Vertex3f(0.138736725f, -0.0179904699f, 0.f),
    C_Vertex3f(0.138736725f, -0.0179904699f, -1.f),
    C_Normal3f(-0.949659526f, 0.313284039f, 0.f),
    C_Vertex3f(0.147903442f, 0.00825119019f, 0.f),
    C_Vertex3f(0.147903442f, 0.00825119019f, -1.f),
    C_Normal3f(-0.96016115f, 0.279447109f, 0.f),
    C_Vertex3f(0.156051636f, 0.0344963074f, 0.f),
    C_Vertex3f(0.156051636f, 0.0344963074f, -1.f),
    C_Normal3f(-0.969637811f, 0.244545639f, 0.f),
    C_Vertex3f(0.163181305f, 0.0607448816f, 0.f),
    C_Vertex3f(0.163181305f, 0.0607448816f, -1.f),
    C_Normal3f(-0.977984607f, 0.208676979f, 0.f),
    C_Vertex3f(0.16929245f, 0.086996913f, 0.f),
    C_Vertex3f(0.16929245f, 0.086996913f, -1.f),
    C_Normal3f(-0.98510468f, 0.171956077f, 0.f),
    C_Vertex3f(0.174385071f, 0.113252401f, 0.f),
    C_Vertex3f(0.174385071f, 0.113252401f, -1.f),
    C_Normal3f(-0.990911663f, 0.134514302f, 0.f),
    C_Vertex3f(0.178459167f, 0.139511347f, 0.f),
    C_Vertex3f(0.178459167f, 0.139511347f, -1.f),
    C_Normal3f(-0.995333135f, 0.0964976624f, 0.f),
    C_Vertex3f(0.18151474f, 0.165773749f, 0.f),
    C_Vertex3f(0.18151474f, 0.165773749f, -1.f),
    C_Normal3f(-0.998312831f, 0.0580642819f, 0.f),
    C_Vertex3f(0.183551788f, 0.192039609f, 0.f),
    C_Vertex3f(0.183551788f, 0.192039609f, -1.f),
    C_Normal3f(-0.999855578f, 0.0169893391f, 0.f),
    C_Vertex3f(0.184570313f, 0.218308926f, 0.f),
    C_Vertex3f(0.184570313f, 0.218308926f, -1.f),
    C_Normal3f(-0.999506831f, -0.0314017348f, 0.f),
    C_Vertex3f(0.184570313f, 0.251981735f, 0.f),
    C_Vertex3f(0.184570313f, 0.251981735f, -1.f),
    C_Normal3f(-0.996314943f, -0.0857694522f, 0.f),
    C_Vertex3f(0.182227135f, 0.292891502f, 0.f),
    C_Vertex3f(0.182227135f, 0.292891502f, -1.f),
    C_Normal3f(-0.989783108f, -0.142581403f, 0.f),
    C_Vertex3f(0.177540779f, 0.333638191f, 0.f),
    C_Vertex3f(0.177540779f, 0.333638191f, -1.f),
    C_Normal3f(-0.98061049f, -0.195967004f, 0.f),
    C_Vertex3f(0.170511246f, 0.374221802f, 0.f),
    C_Vertex3f(0.170511246f, 0.374221802f, -1.f),
    C_Normal3f(-0.969801903f, -0.243894055f, 0.f),
    C_Vertex3f(0.162117004f, 0.4108181f, 0.f),
    C_Vertex3f(0.162117004f, 0.4108181f, -1.f),
    C_Normal3f(-0.957120419f, -0.289690197f, 0.f),
    C_Vertex3f(0.153230667f, 0.44293499f, 0.f),
    C_Vertex3f(0.153230667f, 0.44293499f, -1.f),
    C_Normal3f(-0.94200182f, -0.335607886f, 0.f),
    C_Vertex3f(0.142873764f, 0.474396706f, 0.f),
    C_Vertex3f(0.142873764f, 0.474396706f, -1.f),
    C_Normal3f(-0.926686645f, -0.375834703f, 0.f),
    C_Vertex3f(0.131046295f, 0.505203247f, 0.f),
    C_Vertex3f(0.131046295f, 0.505203247f, -1.f),
    C_Normal3f(-0.911933899f, -0.41033709f, 0.f),
    C_Vertex3f(0.122271538f, 0.525195122f, 0.f),
    C_Vertex3f(0.122271538f, 0.525195122f, -1.f),
    C_Normal3f(-0.900132418f, -0.435616374f, 0.f),
    C_Vertex3f(0.116889238f, 0.536665916f, 0.f),
    C_Vertex3f(0.116889238f, 0.536665916f, -1.f),
    C_Normal3f(-0.89148128f, -0.453057736f, 0.f),
    C_Vertex3f(0.11037612f, 0.549775124f, 0.f),
    C_Vertex3f(0.11037612f, 0.549775124f, -1.f),
    C_Normal3f(-0.884536624f, -0.466470718f, 0.f),
    C_Vertex3f(0.102732182f, 0.564522743f, 0.f),
    C_Vertex3f(0.102732182f, 0.564522743f, -1.f),
    C_Normal3f(-0.878849566f, -0.477099001f, 0.f),
    C_Vertex3f(0.0939574242f, 0.580908775f, 0.f),
    C_Vertex3f(0.0939574242f, 0.580908775f, -1.f),
    C_Normal3f(-0.874112189f, -0.485724181f, 0.f),
    C_Vertex3f(0.0840518475f, 0.59893322f, 0.f),
    C_Vertex3f(0.0840518475f, 0.59893322f, -1.f),
    C_Normal3f(-0.87010783f, -0.492861569f, 0.f),
    C_Vertex3f(0.0730154514f, 0.618596077f, 0.f),
    C_Vertex3f(0.0730154514f, 0.618596077f, -1.f),
    C_Normal3f(-0.867216706f, -0.497930974f, 0.f),
    C_Vertex3f(0.0608482361f, 0.639897346f, 0.f),
    C_Vertex3f(0.0608482361f, 0.639897346f, -1.f),
    C_Normal3f(-0.865145445f, -0.501521051f, 0.f),
    C_Vertex3f(0.0541992188f, 0.651367188f, 0.f),
    C_Vertex3f(0.0541992188f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0541992188f, 0.651367188f, 0.f),
    C_Vertex3f(0.0541992188f, 0.651367188f, -1.f),
    C_Vertex3f(0.110351563f, 0.651367188f, -1.f),
    C_Vertex3f(0.110351563f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.800106883f, 0.599857569f, 0.f),
    C_Vertex3f(0.110351563f, 0.651367188f, 0.f),
    C_Vertex3f(0.110351563f, 0.651367188f, -1.f),
    C_Normal3f(0.812811077f, 0.582527339f, 0.f),
    C_Vertex3f(0.120630264f, 0.637657166f, 0.f),
    C_Vertex3f(0.120630264f, 0.637657166f, -1.f),
    C_Normal3f(0.828677893f, 0.559725761f, 0.f),
    C_Vertex3f(0.139848232f, 0.610209942f, 0.f),
    C_Vertex3f(0.139848232f, 0.610209942f, -1.f),
    C_Normal3f(0.847598672f, 0.530637801f, 0.f),
    C_Vertex3f(0.157726765f, 0.582735538f, 0.f),
    C_Vertex3f(0.157726765f, 0.582735538f, -1.f),
    C_Normal3f(0.866250396f, 0.499610037f, 0.f),
    C_Vertex3f(0.174265862f, 0.555233955f, 0.f),
    C_Vertex3f(0.174265862f, 0.555233955f, -1.f),
    C_Normal3f(0.884468198f, 0.466600388f, 0.f),
    C_Vertex3f(0.189465523f, 0.527705193f, 0.f),
    C_Vertex3f(0.189465523f, 0.527705193f, -1.f),
    C_Normal3f(0.902069449f, 0.431590706f, 0.f),
    C_Vertex3f(0.203325748f, 0.50014925f, 0.f),
    C_Vertex3f(0.203325748f, 0.50014925f, -1.f),
    C_Normal3f(0.918856561f, 0.394591779f, 0.f),
    C_Vertex3f(0.215846539f, 0.472566128f, 0.f),
    C_Vertex3f(0.215846539f, 0.472566128f, -1.f),
    C_Normal3f(0.934527814f, 0.355890036f, 0.f),
    C_Vertex3f(0.227027893f, 0.444955826f, 0.f),
    C_Vertex3f(0.227027893f, 0.444955826f, -1.f),
    C_Normal3f(0.94929719f, 0.314380258f, 0.f),
    C_Vertex3f(0.236158371f, 0.419229507f, 0.f),
    C_Vertex3f(0.236158371f, 0.419229507f, -1.f),
    C_Normal3f(0.962367892f, 0.271750033f, 0.f),
    C_Vertex3f(0.243525028f, 0.395141363f, 0.f),
    C_Vertex3f(0.243525028f, 0.395141363f, -1.f),
    C_Normal3f(0.973204792f, 0.229940414f, 0.f),
    C_Vertex3f(0.249839306f, 0.37078023f, 0.f),
    C_Vertex3f(0.249839306f, 0.37078023f, -1.f),
    C_Normal3f(0.982210875f, 0.187781513f, 0.f),
    C_Vertex3f(0.255101204f, 0.346146107f, 0.f),
    C_Vertex3f(0.255101204f, 0.346146107f, -1.f),
    C_Normal3f(0.989356577f, 0.14551121f, 0.f),
    C_Vertex3f(0.259310722f, 0.321238995f, 0.f),
    C_Vertex3f(0.259310722f, 0.321238995f, -1.f),
    C_Normal3f(0.994643509f, 0.10336525f, 0.f),
    C_Vertex3f(0.262467861f, 0.296058893f, 0.f),
    C_Vertex3f(0.262467861f, 0.296058893f, -1.f),
    C_Normal3f(0.998102665f, 0.0615709201f, 0.f),
    C_Vertex3f(0.26457262f, 0.270605803f, 0.f),
    C_Vertex3f(0.26457262f, 0.270605803f, -1.f),
    C_Normal3f(0.999806404f, 0.0196766835f, 0.f),
    C_Vertex3f(0.265625f, 0.244879723f, 0.f),
    C_Vertex3f(0.265625f, 0.244879723f, -1.f),
    C_Normal3f(0.999692082f, -0.0248123817f, 0.f),
    C_Vertex3f(0.265625f, 0.217132568f, 0.f),
    C_Vertex3f(0.265625f, 0.217132568f, -1.f),
    C_Normal3f(0.997356713f, -0.0726614669f, 0.f),
    C_Vertex3f(0.264210343f, 0.187883139f, 0.f),
    C_Vertex3f(0.264210343f, 0.187883139f, -1.f),
    C_Normal3f(0.992585421f, -0.121548891f, 0.f),
    C_Vertex3f(0.26138103f, 0.158879519f, 0.f),
    C_Vertex3f(0.26138103f, 0.158879519f, -1.f),
    C_Normal3f(0.985377908f, -0.170382962f, 0.f),
    C_Vertex3f(0.25713706f, 0.130121708f, 0.f),
    C_Vertex3f(0.25713706f, 0.130121708f, -1.f),
    C_Normal3f(0.975768507f, -0.218805775f, 0.f),
    C_Vertex3f(0.251478434f, 0.101609707f, 0.f),
    C_Vertex3f(0.251478434f, 0.101609707f, -1.f),
    C_Normal3f(0.963843703f, -0.266468287f, 0.f),
    C_Vertex3f(0.24440515f, 0.0733435154f, 0.f),
    C_Vertex3f(0.24440515f, 0.0733435154f, -1.f),
    C_Normal3f(0.949739039f, -0.313042819f, 0.f),
    C_Vertex3f(0.235917211f, 0.0453231335f, 0.f),
    C_Vertex3f(0.235917211f, 0.0453231335f, -1.f),
    C_Normal3f(0.933631837f, -0.358233869f, 0.f),
    C_Vertex3f(0.226014614f, 0.0175485611f, 0.f),
    C_Vertex3f(0.226014614f, 0.0175485611f, -1.f),
    C_Normal3f(0.919013917f, -0.394225031f, 0.f),
    C_Vertex3f(0.214697361f, -0.00998020172f, 0.f),
    C_Vertex3f(0.214697361f, -0.00998020172f, -1.f),
    C_Normal3f(0.906251848f, -0.422738284f, 0.f),
    C_Vertex3f(0.202771783f, -0.036634922f, 0.f),
    C_Vertex3f(0.202771783f, -0.036634922f, -1.f),
    C_Normal3f(0.892074406f, -0.451888621f, 0.f),
    C_Vertex3f(0.19023788f, -0.0624155998f, 0.f),
    C_Vertex3f(0.19023788f, -0.0624155998f, -1.f),
    C_Normal3f(0.876395822f, -0.481591582f, 0.f),
    C_Vertex3f(0.177095652f, -0.0873222351f, 0.f),
    C_Vertex3f(0.177095652f, -0.0873222351f, -1.f),
    C_Normal3f(0.859137058f, -0.511745512f, 0.f),
    C_Vertex3f(0.163345098f, -0.111354828f, 0.f),
    C_Vertex3f(0.163345098f, -0.111354828f, -1.f),
    C_Normal3f(0.840229392f, -0.542231023f, 0.f),
    C_Vertex3f(0.14898622f, -0.134513378f, 0.f),
    C_Vertex3f(0.14898622f, -0.134513378f, -1.f),
    C_Normal3f(0.819617271f, -0.572911441f, 0.f),
    C_Vertex3f(0.134019017f, -0.156797886f, 0.f),
    C_Vertex3f(0.134019017f, -0.156797886f, -1.f),
    C_Normal3f(0.80110991f, -0.598517179f, 0.f),
    C_Vertex3f(0.118443489f, -0.178208351f, 0.f),
    C_Vertex3f(0.118443489f, -0.178208351f, -1.f),
    C_Normal3f(0.785423994f, -0.618958175f, 0.f),
    C_Vertex3f(0.110351563f, -0.188476563f, 0.f),
    C_Vertex3f(0.110351563f, -0.188476563f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.189465523f, 0.527705193f, 0.f),
    C_Vertex3f(0.174265862f, 0.555233955f, 0.f),
    C_Vertex3f(0.177540779f, 0.333638191f, 0.f),
    C_Vertex3f(0.182227135f, 0.292891502f, 0.f),
    C_Vertex3f(0.184570313f, 0.251981735f, 0.f),
    C_Vertex3f(0.19023788f, -0.0624155998f, 0.f),
    C_Vertex3f(0.203325748f, 0.50014925f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142873764f, 0.474396706f, 0.f),
    C_Vertex3f(0.153230667f, 0.44293499f, 0.f),
    C_Vertex3f(0.157726765f, 0.582735538f, 0.f),
    C_Vertex3f(0.162117004f, 0.4108181f, 0.f),
    C_Vertex3f(0.174265862f, 0.555233955f, 0.f),
    C_Vertex3f(0.170511246f, 0.374221802f, 0.f),
    C_Vertex3f(0.177540779f, 0.333638191f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.139848232f, 0.610209942f, 0.f),
    C_Vertex3f(0.120630264f, 0.637657166f, 0.f),
    C_Vertex3f(0.122271538f, 0.525195122f, 0.f),
    C_Vertex3f(0.131046295f, 0.505203247f, 0.f),
    C_Vertex3f(0.142873764f, 0.474396706f, 0.f),
    C_Vertex3f(0.157726765f, 0.582735538f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0939574242f, 0.580908775f, 0.f),
    C_Vertex3f(0.102732182f, 0.564522743f, 0.f),
    C_Vertex3f(0.110351563f, 0.651367188f, 0.f),
    C_Vertex3f(0.11037612f, 0.549775124f, 0.f),
    C_Vertex3f(0.120630264f, 0.637657166f, 0.f),
    C_Vertex3f(0.116889238f, 0.536665916f, 0.f),
    C_Vertex3f(0.122271538f, 0.525195122f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.110351563f, 0.651367188f, 0.f),
    C_Vertex3f(0.0541992188f, 0.651367188f, 0.f),
    C_Vertex3f(0.0608482361f, 0.639897346f, 0.f),
    C_Vertex3f(0.0730154514f, 0.618596077f, 0.f),
    C_Vertex3f(0.0840518475f, 0.59893322f, 0.f),
    C_Vertex3f(0.0939574242f, 0.580908775f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243525028f, 0.395141363f, 0.f),
    C_Vertex3f(0.24440515f, 0.0733435154f, 0.f),
    C_Vertex3f(0.249839306f, 0.37078023f, 0.f),
    C_Vertex3f(0.251478434f, 0.101609707f, 0.f),
    C_Vertex3f(0.255101204f, 0.346146107f, 0.f),
    C_Vertex3f(0.25713706f, 0.130121708f, 0.f),
    C_Vertex3f(0.259310722f, 0.321238995f, 0.f),
    C_Vertex3f(0.26138103f, 0.158879519f, 0.f),
    C_Vertex3f(0.262467861f, 0.296058893f, 0.f),
    C_Vertex3f(0.264210343f, 0.187883139f, 0.f),
    C_Vertex3f(0.26457262f, 0.270605803f, 0.f),
    C_Vertex3f(0.265625f, 0.217132568f, 0.f),
    C_Vertex3f(0.265625f, 0.244879723f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.19023788f, -0.0624155998f, 0.f),
    C_Vertex3f(0.202771783f, -0.036634922f, 0.f),
    C_Vertex3f(0.203325748f, 0.50014925f, 0.f),
    C_Vertex3f(0.214697361f, -0.00998020172f, 0.f),
    C_Vertex3f(0.215846539f, 0.472566128f, 0.f),
    C_Vertex3f(0.226014614f, 0.0175485611f, 0.f),
    C_Vertex3f(0.227027893f, 0.444955826f, 0.f),
    C_Vertex3f(0.235917211f, 0.0453231335f, 0.f),
    C_Vertex3f(0.236158371f, 0.419229507f, 0.f),
    C_Vertex3f(0.24440515f, 0.0733435154f, 0.f),
    C_Vertex3f(0.243525028f, 0.395141363f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.19023788f, -0.0624155998f, 0.f),
    C_Vertex3f(0.184570313f, 0.251981735f, 0.f),
    C_Vertex3f(0.184570313f, 0.218308926f, 0.f),
    C_Vertex3f(0.183551788f, 0.192039609f, 0.f),
    C_Vertex3f(0.18151474f, 0.165773749f, 0.f),
    C_Vertex3f(0.178459167f, 0.139511347f, 0.f),
    C_Vertex3f(0.177095652f, -0.0873222351f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.178459167f, 0.139511347f, 0.f),
    C_Vertex3f(0.174385071f, 0.113252401f, 0.f),
    C_Vertex3f(0.177095652f, -0.0873222351f, 0.f),
    C_Vertex3f(0.16929245f, 0.086996913f, 0.f),
    C_Vertex3f(0.163345098f, -0.111354828f, 0.f),
    C_Vertex3f(0.163181305f, 0.0607448816f, 0.f),
    C_Vertex3f(0.156051636f, 0.0344963074f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14898622f, -0.134513378f, 0.f),
    C_Vertex3f(0.163345098f, -0.111354828f, 0.f),
    C_Vertex3f(0.156051636f, 0.0344963074f, 0.f),
    C_Vertex3f(0.147903442f, 0.00825119019f, 0.f),
    C_Vertex3f(0.138736725f, -0.0179904699f, 0.f),
    C_Vertex3f(0.134019017f, -0.156797886f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.105125427f, -0.0966947079f, 0.f),
    C_Vertex3f(0.110351563f, -0.188476563f, 0.f),
    C_Vertex3f(0.117347717f, -0.070463419f, 0.f),
    C_Vertex3f(0.118443489f, -0.178208351f, 0.f),
    C_Vertex3f(0.128551483f, -0.044228673f, 0.f),
    C_Vertex3f(0.134019017f, -0.156797886f, 0.f),
    C_Vertex3f(0.138736725f, -0.0179904699f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.110351563f, -0.188476563f, 0.f),
    C_Vertex3f(0.105125427f, -0.0966947079f, 0.f),
    C_Vertex3f(0.091884613f, -0.12292254f, 0.f),
    C_Vertex3f(0.0776252747f, -0.149146914f, 0.f),
    C_Vertex3f(0.0623474121f, -0.175367832f, 0.f),
    C_Vertex3f(0.0541992188f, -0.188476563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.189465523f, 0.527705193f, -1.f),
    C_Vertex3f(0.174265862f, 0.555233955f, -1.f),
    C_Vertex3f(0.177540779f, 0.333638191f, -1.f),
    C_Vertex3f(0.182227135f, 0.292891502f, -1.f),
    C_Vertex3f(0.184570313f, 0.251981735f, -1.f),
    C_Vertex3f(0.19023788f, -0.0624155998f, -1.f),
    C_Vertex3f(0.203325748f, 0.50014925f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142873764f, 0.474396706f, -1.f),
    C_Vertex3f(0.153230667f, 0.44293499f, -1.f),
    C_Vertex3f(0.157726765f, 0.582735538f, -1.f),
    C_Vertex3f(0.162117004f, 0.4108181f, -1.f),
    C_Vertex3f(0.174265862f, 0.555233955f, -1.f),
    C_Vertex3f(0.170511246f, 0.374221802f, -1.f),
    C_Vertex3f(0.177540779f, 0.333638191f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.139848232f, 0.610209942f, -1.f),
    C_Vertex3f(0.120630264f, 0.637657166f, -1.f),
    C_Vertex3f(0.122271538f, 0.525195122f, -1.f),
    C_Vertex3f(0.131046295f, 0.505203247f, -1.f),
    C_Vertex3f(0.142873764f, 0.474396706f, -1.f),
    C_Vertex3f(0.157726765f, 0.582735538f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0939574242f, 0.580908775f, -1.f),
    C_Vertex3f(0.102732182f, 0.564522743f, -1.f),
    C_Vertex3f(0.110351563f, 0.651367188f, -1.f),
    C_Vertex3f(0.11037612f, 0.549775124f, -1.f),
    C_Vertex3f(0.120630264f, 0.637657166f, -1.f),
    C_Vertex3f(0.116889238f, 0.536665916f, -1.f),
    C_Vertex3f(0.122271538f, 0.525195122f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.110351563f, 0.651367188f, -1.f),
    C_Vertex3f(0.0541992188f, 0.651367188f, -1.f),
    C_Vertex3f(0.0608482361f, 0.639897346f, -1.f),
    C_Vertex3f(0.0730154514f, 0.618596077f, -1.f),
    C_Vertex3f(0.0840518475f, 0.59893322f, -1.f),
    C_Vertex3f(0.0939574242f, 0.580908775f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243525028f, 0.395141363f, -1.f),
    C_Vertex3f(0.24440515f, 0.0733435154f, -1.f),
    C_Vertex3f(0.249839306f, 0.37078023f, -1.f),
    C_Vertex3f(0.251478434f, 0.101609707f, -1.f),
    C_Vertex3f(0.255101204f, 0.346146107f, -1.f),
    C_Vertex3f(0.25713706f, 0.130121708f, -1.f),
    C_Vertex3f(0.259310722f, 0.321238995f, -1.f),
    C_Vertex3f(0.26138103f, 0.158879519f, -1.f),
    C_Vertex3f(0.262467861f, 0.296058893f, -1.f),
    C_Vertex3f(0.264210343f, 0.187883139f, -1.f),
    C_Vertex3f(0.26457262f, 0.270605803f, -1.f),
    C_Vertex3f(0.265625f, 0.217132568f, -1.f),
    C_Vertex3f(0.265625f, 0.244879723f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.19023788f, -0.0624155998f, -1.f),
    C_Vertex3f(0.202771783f, -0.036634922f, -1.f),
    C_Vertex3f(0.203325748f, 0.50014925f, -1.f),
    C_Vertex3f(0.214697361f, -0.00998020172f, -1.f),
    C_Vertex3f(0.215846539f, 0.472566128f, -1.f),
    C_Vertex3f(0.226014614f, 0.0175485611f, -1.f),
    C_Vertex3f(0.227027893f, 0.444955826f, -1.f),
    C_Vertex3f(0.235917211f, 0.0453231335f, -1.f),
    C_Vertex3f(0.236158371f, 0.419229507f, -1.f),
    C_Vertex3f(0.24440515f, 0.0733435154f, -1.f),
    C_Vertex3f(0.243525028f, 0.395141363f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.19023788f, -0.0624155998f, -1.f),
    C_Vertex3f(0.184570313f, 0.251981735f, -1.f),
    C_Vertex3f(0.184570313f, 0.218308926f, -1.f),
    C_Vertex3f(0.183551788f, 0.192039609f, -1.f),
    C_Vertex3f(0.18151474f, 0.165773749f, -1.f),
    C_Vertex3f(0.178459167f, 0.139511347f, -1.f),
    C_Vertex3f(0.177095652f, -0.0873222351f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.178459167f, 0.139511347f, -1.f),
    C_Vertex3f(0.174385071f, 0.113252401f, -1.f),
    C_Vertex3f(0.177095652f, -0.0873222351f, -1.f),
    C_Vertex3f(0.16929245f, 0.086996913f, -1.f),
    C_Vertex3f(0.163345098f, -0.111354828f, -1.f),
    C_Vertex3f(0.163181305f, 0.0607448816f, -1.f),
    C_Vertex3f(0.156051636f, 0.0344963074f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14898622f, -0.134513378f, -1.f),
    C_Vertex3f(0.163345098f, -0.111354828f, -1.f),
    C_Vertex3f(0.156051636f, 0.0344963074f, -1.f),
    C_Vertex3f(0.147903442f, 0.00825119019f, -1.f),
    C_Vertex3f(0.138736725f, -0.0179904699f, -1.f),
    C_Vertex3f(0.134019017f, -0.156797886f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.105125427f, -0.0966947079f, -1.f),
    C_Vertex3f(0.110351563f, -0.188476563f, -1.f),
    C_Vertex3f(0.117347717f, -0.070463419f, -1.f),
    C_Vertex3f(0.118443489f, -0.178208351f, -1.f),
    C_Vertex3f(0.128551483f, -0.044228673f, -1.f),
    C_Vertex3f(0.134019017f, -0.156797886f, -1.f),
    C_Vertex3f(0.138736725f, -0.0179904699f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.110351563f, -0.188476563f, -1.f),
    C_Vertex3f(0.105125427f, -0.0966947079f, -1.f),
    C_Vertex3f(0.091884613f, -0.12292254f, -1.f),
    C_Vertex3f(0.0776252747f, -0.149146914f, -1.f),
    C_Vertex3f(0.0623474121f, -0.175367832f, -1.f),
    C_Vertex3f(0.0541992188f, -0.188476563f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_42[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.951283157f, 0.308318704f, 0.f),
    C_Vertex3f(0.0278320313f, 0.522949219f, 0.f),
    C_Vertex3f(0.0278320313f, 0.522949219f, -1.f),
    C_Vertex3f(0.0479354858f, 0.584976196f, -1.f),
    C_Vertex3f(0.0479354858f, 0.584976196f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.332061499f, 0.943257689f, 0.f),
    C_Vertex3f(0.0479354858f, 0.584976196f, 0.f),
    C_Vertex3f(0.0479354858f, 0.584976196f, -1.f),
    C_Normal3f(0.33923775f, 0.94070065f, 0.f),
    C_Vertex3f(0.0566205978f, 0.581918716f, 0.f),
    C_Vertex3f(0.0566205978f, 0.581918716f, -1.f),
    C_Normal3f(0.349135071f, 0.937072396f, 0.f),
    C_Vertex3f(0.0728032589f, 0.57600832f, 0.f),
    C_Vertex3f(0.0728032589f, 0.57600832f, -1.f),
    C_Normal3f(0.362617135f, 0.931938171f, 0.f),
    C_Vertex3f(0.087798357f, 0.570302486f, 0.f),
    C_Vertex3f(0.087798357f, 0.570302486f, -1.f),
    C_Normal3f(0.378234446f, 0.925709784f, 0.f),
    C_Vertex3f(0.101605892f, 0.564801216f, 0.f),
    C_Vertex3f(0.101605892f, 0.564801216f, -1.f),
    C_Normal3f(0.396519512f, 0.918026328f, 0.f),
    C_Vertex3f(0.114225864f, 0.559504509f, 0.f),
    C_Vertex3f(0.114225864f, 0.559504509f, -1.f),
    C_Normal3f(0.418188751f, 0.908360124f, 0.f),
    C_Vertex3f(0.125658274f, 0.554412365f, 0.f),
    C_Vertex3f(0.125658274f, 0.554412365f, -1.f),
    C_Normal3f(0.444223225f, 0.895916164f, 0.f),
    C_Vertex3f(0.13590312f, 0.549524784f, 0.f),
    C_Vertex3f(0.13590312f, 0.549524784f, -1.f),
    C_Normal3f(0.470223367f, 0.882547379f, 0.f),
    C_Vertex3f(0.144960403f, 0.544841766f, 0.f),
    C_Vertex3f(0.144960403f, 0.544841766f, -1.f),
    C_Normal3f(0.494595587f, 0.86912322f, 0.f),
    C_Vertex3f(0.148895264f, 0.542602539f, 0.f),
    C_Vertex3f(0.148895264f, 0.542602539f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.994503856f, -0.104699768f, 0.f),
    C_Vertex3f(0.148895264f, 0.542602539f, 0.f),
    C_Vertex3f(0.148895264f, 0.542602539f, -1.f),
    C_Normal3f(-0.994809568f, -0.101754352f, 0.f),
    C_Vertex3f(0.147854805f, 0.552485466f, 0.f),
    C_Vertex3f(0.147854805f, 0.552485466f, -1.f),
    C_Normal3f(-0.995227695f, -0.0975798741f, 0.f),
    C_Vertex3f(0.146020412f, 0.570708752f, 0.f),
    C_Vertex3f(0.146020412f, 0.570708752f, -1.f),
    C_Normal3f(-0.995789289f, -0.091671288f, 0.f),
    C_Vertex3f(0.144432545f, 0.587389469f, 0.f),
    C_Vertex3f(0.144432545f, 0.587389469f, -1.f),
    C_Normal3f(-0.996425033f, -0.0844809189f, 0.f),
    C_Vertex3f(0.143091202f, 0.602527618f, 0.f),
    C_Vertex3f(0.143091202f, 0.602527618f, -1.f),
    C_Normal3f(-0.997142613f, -0.0755425096f, 0.f),
    C_Vertex3f(0.141996384f, 0.616123199f, 0.f),
    C_Vertex3f(0.141996384f, 0.616123199f, -1.f),
    C_Normal3f(-0.997941375f, -0.0641336367f, 0.f),
    C_Vertex3f(0.14114809f, 0.628176212f, 0.f),
    C_Vertex3f(0.14114809f, 0.628176212f, -1.f),
    C_Normal3f(-0.998795211f, -0.0490729772f, 0.f),
    C_Vertex3f(0.140546322f, 0.638686657f, 0.f),
    C_Vertex3f(0.140546322f, 0.638686657f, -1.f),
    C_Normal3f(-0.999478757f, -0.0322848968f, 0.f),
    C_Vertex3f(0.140191078f, 0.647654533f, 0.f),
    C_Vertex3f(0.140191078f, 0.647654533f, -1.f),
    C_Normal3f(-0.999892831f, -0.0146400956f, 0.f),
    C_Vertex3f(0.140136719f, 0.651367188f, 0.f),
    C_Vertex3f(0.140136719f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.140136719f, 0.651367188f, 0.f),
    C_Vertex3f(0.140136719f, 0.651367188f, -1.f),
    C_Vertex3f(0.203613281f, 0.651367188f, -1.f),
    C_Vertex3f(0.203613281f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999539852f, -0.030331837f, 0.f),
    C_Vertex3f(0.203613281f, 0.651367188f, 0.f),
    C_Vertex3f(0.203613281f, 0.651367188f, -1.f),
    C_Normal3f(0.998520195f, -0.0543810278f, 0.f),
    C_Vertex3f(0.203285217f, 0.640556335f, 0.f),
    C_Vertex3f(0.203285217f, 0.640556335f, -1.f),
    C_Normal3f(0.996830523f, -0.0795542076f, 0.f),
    C_Vertex3f(0.201698303f, 0.616205215f, 0.f),
    C_Vertex3f(0.201698303f, 0.616205215f, -1.f),
    C_Normal3f(0.994545937f, -0.104299456f, 0.f),
    C_Vertex3f(0.199180603f, 0.58912468f, 0.f),
    C_Vertex3f(0.199180603f, 0.58912468f, -1.f),
    C_Normal3f(0.99259752f, -0.12145032f, 0.f),
    C_Vertex3f(0.195732117f, 0.559314728f, 0.f),
    C_Vertex3f(0.195732117f, 0.559314728f, -1.f),
    C_Normal3f(0.991064847f, -0.133381292f, 0.f),
    C_Vertex3f(0.19354248f, 0.543045044f, 0.f),
    C_Vertex3f(0.19354248f, 0.543045044f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.45011425f, 0.892971039f, 0.f),
    C_Vertex3f(0.19354248f, 0.543045044f, 0.f),
    C_Vertex3f(0.19354248f, 0.543045044f, -1.f),
    C_Normal3f(-0.424185932f, 0.905575097f, 0.f),
    C_Vertex3f(0.204807281f, 0.548723221f, 0.f),
    C_Vertex3f(0.204807281f, 0.548723221f, -1.f),
    C_Normal3f(-0.393663138f, 0.919254839f, 0.f),
    C_Vertex3f(0.22897625f, 0.559642792f, 0.f),
    C_Vertex3f(0.22897625f, 0.559642792f, -1.f),
    C_Normal3f(-0.35967654f, 0.933077037f, 0.f),
    C_Vertex3f(0.254784584f, 0.57012558f, 0.f),
    C_Vertex3f(0.254784584f, 0.57012558f, -1.f),
    C_Normal3f(-0.333422601f, 0.942777455f, 0.f),
    C_Vertex3f(0.282232285f, 0.580171585f, 0.f),
    C_Vertex3f(0.282232285f, 0.580171585f, -1.f),
    C_Normal3f(-0.313686252f, 0.949526668f, 0.f),
    C_Vertex3f(0.296775818f, 0.584976196f, 0.f),
    C_Vertex3f(0.296775818f, 0.584976196f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.951214433f, 0.30853042f, 0.f),
    C_Vertex3f(0.296775818f, 0.584976196f, 0.f),
    C_Vertex3f(0.296775818f, 0.584976196f, -1.f),
    C_Vertex3f(0.316894531f, 0.522949219f, -1.f),
    C_Vertex3f(0.316894531f, 0.522949219f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.313409209f, -0.94961822f, 0.f),
    C_Vertex3f(0.316894531f, 0.522949219f, 0.f),
    C_Vertex3f(0.316894531f, 0.522949219f, -1.f),
    C_Normal3f(0.283221424f, -0.95905453f, 0.f),
    C_Vertex3f(0.303007126f, 0.51836586f, 0.f),
    C_Vertex3f(0.303007126f, 0.51836586f, -1.f),
    C_Normal3f(0.243680745f, -0.969855428f, 0.f),
    C_Vertex3f(0.275506973f, 0.510726929f, 0.f),
    C_Vertex3f(0.275506973f, 0.510726929f, -1.f),
    C_Normal3f(0.193664536f, -0.981067896f, 0.f),
    C_Vertex3f(0.248281479f, 0.504615784f, 0.f),
    C_Vertex3f(0.248281479f, 0.504615784f, -1.f),
    C_Normal3f(0.149967596f, -0.988690853f, 0.f),
    C_Vertex3f(0.221330643f, 0.500032425f, 0.f),
    C_Vertex3f(0.221330643f, 0.500032425f, -1.f),
    C_Normal3f(0.113799013f, -0.993503749f, 0.f),
    C_Vertex3f(0.207992554f, 0.498504639f, 0.f),
    C_Vertex3f(0.207992554f, 0.498504639f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.655193746f, 0.755460858f, 0.f),
    C_Vertex3f(0.207992554f, 0.498504639f, 0.f),
    C_Vertex3f(0.207992554f, 0.498504639f, -1.f),
    C_Normal3f(0.676153302f, 0.736761034f, 0.f),
    C_Vertex3f(0.211326599f, 0.495613098f, 0.f),
    C_Vertex3f(0.211326599f, 0.495613098f, -1.f),
    C_Normal3f(0.697145641f, 0.716929495f, 0.f),
    C_Vertex3f(0.218678474f, 0.488697767f, 0.f),
    C_Vertex3f(0.218678474f, 0.488697767f, -1.f),
    C_Normal3f(0.716865301f, 0.697211742f, 0.f),
    C_Vertex3f(0.226714134f, 0.480650187f, 0.f),
    C_Vertex3f(0.226714134f, 0.480650187f, -1.f),
    C_Normal3f(0.732361734f, 0.680915892f, 0.f),
    C_Vertex3f(0.235433578f, 0.471470356f, 0.f),
    C_Vertex3f(0.235433578f, 0.471470356f, -1.f),
    C_Normal3f(0.744833231f, 0.667250693f, 0.f),
    C_Vertex3f(0.244836807f, 0.461158276f, 0.f),
    C_Vertex3f(0.244836807f, 0.461158276f, -1.f),
    C_Normal3f(0.755071282f, 0.655642688f, 0.f),
    C_Vertex3f(0.25492382f, 0.449713945f, 0.f),
    C_Vertex3f(0.25492382f, 0.449713945f, -1.f),
    C_Normal3f(0.763617218f, 0.645669341f, 0.f),
    C_Vertex3f(0.265694618f, 0.437137365f, 0.f),
    C_Vertex3f(0.265694618f, 0.437137365f, -1.f),
    C_Normal3f(0.769724905f, 0.638375819f, 0.f),
    C_Vertex3f(0.2771492f, 0.423428535f, 0.f),
    C_Vertex3f(0.2771492f, 0.423428535f, -1.f),
    C_Normal3f(0.774068117f, 0.633102357f, 0.f),
    C_Vertex3f(0.283218384f, 0.416007996f, 0.f),
    C_Vertex3f(0.283218384f, 0.416007996f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.577200294f, -0.816602528f, 0.f),
    C_Vertex3f(0.283218384f, 0.416007996f, 0.f),
    C_Vertex3f(0.283218384f, 0.416007996f, -1.f),
    C_Vertex3f(0.230728149f, 0.37890625f, -1.f),
    C_Vertex3f(0.230728149f, 0.37890625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.805525959f, -0.592560589f, 0.f),
    C_Vertex3f(0.230728149f, 0.37890625f, 0.f),
    C_Vertex3f(0.230728149f, 0.37890625f, -1.f),
    C_Normal3f(-0.82062155f, -0.571471989f, 0.f),
    C_Vertex3f(0.224386215f, 0.387527466f, 0.f),
    C_Vertex3f(0.224386215f, 0.387527466f, -1.f),
    C_Normal3f(-0.835455954f, -0.549557388f, 0.f),
    C_Vertex3f(0.210551262f, 0.407879829f, 0.f),
    C_Vertex3f(0.210551262f, 0.407879829f, -1.f),
    C_Normal3f(-0.849128366f, -0.5281865f, 0.f),
    C_Vertex3f(0.195565224f, 0.431342125f, 0.f),
    C_Vertex3f(0.195565224f, 0.431342125f, -1.f),
    C_Normal3f(-0.85810256f, -0.513478279f, 0.f),
    C_Vertex3f(0.179428101f, 0.457914352f, 0.f),
    C_Vertex3f(0.179428101f, 0.457914352f, -1.f),
    C_Normal3f(-0.864113033f, -0.503297806f, 0.f),
    C_Vertex3f(0.170783997f, 0.472755432f, 0.f),
    C_Vertex3f(0.170783997f, 0.472755432f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.884908736f, -0.465764523f, 0.f),
    C_Vertex3f(0.170783997f, 0.472755432f, 0.f),
    C_Vertex3f(0.170783997f, 0.472755432f, -1.f),
    C_Normal3f(0.87683785f, -0.480786294f, 0.f),
    C_Vertex3f(0.162685394f, 0.457368851f, 0.f),
    C_Vertex3f(0.162685394f, 0.457368851f, -1.f),
    C_Normal3f(0.864375889f, -0.502846181f, 0.f),
    C_Vertex3f(0.147478104f, 0.430251122f, 0.f),
    C_Vertex3f(0.147478104f, 0.430251122f, -1.f),
    C_Normal3f(0.844456851f, -0.53562361f, 0.f),
    C_Vertex3f(0.133260727f, 0.406788826f, 0.f),
    C_Vertex3f(0.133260727f, 0.406788826f, -1.f),
    C_Normal3f(0.82159996f, -0.570064545f, 0.f),
    C_Vertex3f(0.120033264f, 0.386981964f, 0.f),
    C_Vertex3f(0.120033264f, 0.386981964f, -1.f),
    C_Normal3f(0.797053754f, -0.603908479f, 0.f),
    C_Vertex3f(0.11391449f, 0.37890625f, 0.f),
    C_Vertex3f(0.11391449f, 0.37890625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.584086359f, -0.811691523f, 0.f),
    C_Vertex3f(0.11391449f, 0.37890625f, 0.f),
    C_Vertex3f(0.11391449f, 0.37890625f, -1.f),
    C_Vertex3f(0.0623550415f, 0.416007996f, -1.f),
    C_Vertex3f(0.0623550415f, 0.416007996f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.776256204f, 0.630417585f, 0.f),
    C_Vertex3f(0.0623550415f, 0.416007996f, 0.f),
    C_Vertex3f(0.0623550415f, 0.416007996f, -1.f),
    C_Normal3f(-0.773287654f, 0.634055316f, 0.f),
    C_Vertex3f(0.0686912537f, 0.423810005f, 0.f),
    C_Vertex3f(0.0686912537f, 0.423810005f, -1.f),
    C_Normal3f(-0.769046605f, 0.6391927f, 0.f),
    C_Vertex3f(0.0804619789f, 0.43809104f, 0.f),
    C_Vertex3f(0.0804619789f, 0.43809104f, -1.f),
    C_Normal3f(-0.762976229f, 0.646426558f, 0.f),
    C_Vertex3f(0.0913310051f, 0.451049089f, 0.f),
    C_Vertex3f(0.0913310051f, 0.451049089f, -1.f),
    C_Normal3f(-0.755482972f, 0.655168295f, 0.f),
    C_Vertex3f(0.101298332f, 0.462684155f, 0.f),
    C_Vertex3f(0.101298332f, 0.462684155f, -1.f),
    C_Normal3f(-0.746007502f, 0.665937603f, 0.f),
    C_Vertex3f(0.11036396f, 0.472996235f, 0.f),
    C_Vertex3f(0.11036396f, 0.472996235f, -1.f),
    C_Normal3f(-0.733658016f, 0.679518759f, 0.f),
    C_Vertex3f(0.118527889f, 0.481985331f, 0.f),
    C_Vertex3f(0.118527889f, 0.481985331f, -1.f),
    C_Normal3f(-0.716926634f, 0.697148621f, 0.f),
    C_Vertex3f(0.125790119f, 0.489651442f, 0.f),
    C_Vertex3f(0.125790119f, 0.489651442f, -1.f),
    C_Normal3f(-0.697716177f, 0.716374338f, 0.f),
    C_Vertex3f(0.13215065f, 0.495994568f, 0.f),
    C_Vertex3f(0.13215065f, 0.495994568f, -1.f),
    C_Normal3f(-0.676911652f, 0.736064255f, 0.f),
    C_Vertex3f(0.134880066f, 0.498504639f, 0.f),
    C_Vertex3f(0.134880066f, 0.498504639f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.190054953f, -0.981773436f, 0.f),
    C_Vertex3f(0.134880066f, 0.498504639f, 0.f),
    C_Vertex3f(0.134880066f, 0.498504639f, -1.f),
    C_Normal3f(-0.211467117f, -0.977385163f, 0.f),
    C_Vertex3f(0.106700897f, 0.503959656f, 0.f),
    C_Vertex3f(0.106700897f, 0.503959656f, -1.f),
    C_Normal3f(-0.234084293f, -0.972216308f, 0.f),
    C_Vertex3f(0.0531768799f, 0.516181946f, 0.f),
    C_Vertex3f(0.0531768799f, 0.516181946f, -1.f),
    C_Normal3f(-0.257970363f, -0.966152847f, 0.f),
    C_Vertex3f(0.0278320313f, 0.522949219f, 0.f),
    C_Vertex3f(0.0278320313f, 0.522949219f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134880066f, 0.498504639f, 0.f),
    C_Vertex3f(0.13215065f, 0.495994568f, 0.f),
    C_Vertex3f(0.133260727f, 0.406788826f, 0.f),
    C_Vertex3f(0.147478104f, 0.430251122f, 0.f),
    C_Vertex3f(0.13590312f, 0.549524784f, 0.f),
    C_Vertex3f(0.125658274f, 0.554412365f, 0.f),
    C_Vertex3f(0.114225864f, 0.559504509f, 0.f),
    C_Vertex3f(0.106700897f, 0.503959656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.106700897f, 0.503959656f, 0.f),
    C_Vertex3f(0.114225864f, 0.559504509f, 0.f),
    C_Vertex3f(0.101605892f, 0.564801216f, 0.f),
    C_Vertex3f(0.087798357f, 0.570302486f, 0.f),
    C_Vertex3f(0.0728032589f, 0.57600832f, 0.f),
    C_Vertex3f(0.0566205978f, 0.581918716f, 0.f),
    C_Vertex3f(0.0531768799f, 0.516181946f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0479354858f, 0.584976196f, 0.f),
    C_Vertex3f(0.0278320313f, 0.522949219f, 0.f),
    C_Vertex3f(0.0531768799f, 0.516181946f, 0.f),
    C_Vertex3f(0.0566205978f, 0.581918716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.203613281f, 0.651367188f, 0.f),
    C_Vertex3f(0.140136719f, 0.651367188f, 0.f),
    C_Vertex3f(0.140191078f, 0.647654533f, 0.f),
    C_Vertex3f(0.140546322f, 0.638686657f, 0.f),
    C_Vertex3f(0.14114809f, 0.628176212f, 0.f),
    C_Vertex3f(0.141996384f, 0.616123199f, 0.f),
    C_Vertex3f(0.143091202f, 0.602527618f, 0.f),
    C_Vertex3f(0.144432545f, 0.587389469f, 0.f),
    C_Vertex3f(0.146020412f, 0.570708752f, 0.f),
    C_Vertex3f(0.147854805f, 0.552485466f, 0.f),
    C_Vertex3f(0.148895264f, 0.542602539f, 0.f),
    C_Vertex3f(0.162685394f, 0.457368851f, 0.f),
    C_Vertex3f(0.170783997f, 0.472755432f, 0.f),
    C_Vertex3f(0.179428101f, 0.457914352f, 0.f),
    C_Vertex3f(0.19354248f, 0.543045044f, 0.f),
    C_Vertex3f(0.195732117f, 0.559314728f, 0.f),
    C_Vertex3f(0.199180603f, 0.58912468f, 0.f),
    C_Vertex3f(0.201698303f, 0.616205215f, 0.f),
    C_Vertex3f(0.203285217f, 0.640556335f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.283218384f, 0.416007996f, 0.f),
    C_Vertex3f(0.2771492f, 0.423428535f, 0.f),
    C_Vertex3f(0.265694618f, 0.437137365f, 0.f),
    C_Vertex3f(0.25492382f, 0.449713945f, 0.f),
    C_Vertex3f(0.244836807f, 0.461158276f, 0.f),
    C_Vertex3f(0.235433578f, 0.471470356f, 0.f),
    C_Vertex3f(0.230728149f, 0.37890625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.226714134f, 0.480650187f, 0.f),
    C_Vertex3f(0.218678474f, 0.488697767f, 0.f),
    C_Vertex3f(0.224386215f, 0.387527466f, 0.f),
    C_Vertex3f(0.230728149f, 0.37890625f, 0.f),
    C_Vertex3f(0.235433578f, 0.471470356f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.22897625f, 0.559642792f, 0.f),
    C_Vertex3f(0.204807281f, 0.548723221f, 0.f),
    C_Vertex3f(0.207992554f, 0.498504639f, 0.f),
    C_Vertex3f(0.210551262f, 0.407879829f, 0.f),
    C_Vertex3f(0.211326599f, 0.495613098f, 0.f),
    C_Vertex3f(0.224386215f, 0.387527466f, 0.f),
    C_Vertex3f(0.218678474f, 0.488697767f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.195565224f, 0.431342125f, 0.f),
    C_Vertex3f(0.210551262f, 0.407879829f, 0.f),
    C_Vertex3f(0.204807281f, 0.548723221f, 0.f),
    C_Vertex3f(0.19354248f, 0.543045044f, 0.f),
    C_Vertex3f(0.179428101f, 0.457914352f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.207992554f, 0.498504639f, 0.f),
    C_Vertex3f(0.221330643f, 0.500032425f, 0.f),
    C_Vertex3f(0.22897625f, 0.559642792f, 0.f),
    C_Vertex3f(0.248281479f, 0.504615784f, 0.f),
    C_Vertex3f(0.254784584f, 0.57012558f, 0.f),
    C_Vertex3f(0.275506973f, 0.510726929f, 0.f),
    C_Vertex3f(0.282232285f, 0.580171585f, 0.f),
    C_Vertex3f(0.303007126f, 0.51836586f, 0.f),
    C_Vertex3f(0.296775818f, 0.584976196f, 0.f),
    C_Vertex3f(0.316894531f, 0.522949219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147478104f, 0.430251122f, 0.f),
    C_Vertex3f(0.162685394f, 0.457368851f, 0.f),
    C_Vertex3f(0.148895264f, 0.542602539f, 0.f),
    C_Vertex3f(0.144960403f, 0.544841766f, 0.f),
    C_Vertex3f(0.13590312f, 0.549524784f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11036396f, 0.472996235f, 0.f),
    C_Vertex3f(0.11391449f, 0.37890625f, 0.f),
    C_Vertex3f(0.118527889f, 0.481985331f, 0.f),
    C_Vertex3f(0.120033264f, 0.386981964f, 0.f),
    C_Vertex3f(0.125790119f, 0.489651442f, 0.f),
    C_Vertex3f(0.133260727f, 0.406788826f, 0.f),
    C_Vertex3f(0.13215065f, 0.495994568f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.11391449f, 0.37890625f, 0.f),
    C_Vertex3f(0.11036396f, 0.472996235f, 0.f),
    C_Vertex3f(0.101298332f, 0.462684155f, 0.f),
    C_Vertex3f(0.0913310051f, 0.451049089f, 0.f),
    C_Vertex3f(0.0804619789f, 0.43809104f, 0.f),
    C_Vertex3f(0.0686912537f, 0.423810005f, 0.f),
    C_Vertex3f(0.0623550415f, 0.416007996f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134880066f, 0.498504639f, -1.f),
    C_Vertex3f(0.13215065f, 0.495994568f, -1.f),
    C_Vertex3f(0.133260727f, 0.406788826f, -1.f),
    C_Vertex3f(0.147478104f, 0.430251122f, -1.f),
    C_Vertex3f(0.13590312f, 0.549524784f, -1.f),
    C_Vertex3f(0.125658274f, 0.554412365f, -1.f),
    C_Vertex3f(0.114225864f, 0.559504509f, -1.f),
    C_Vertex3f(0.106700897f, 0.503959656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.106700897f, 0.503959656f, -1.f),
    C_Vertex3f(0.114225864f, 0.559504509f, -1.f),
    C_Vertex3f(0.101605892f, 0.564801216f, -1.f),
    C_Vertex3f(0.087798357f, 0.570302486f, -1.f),
    C_Vertex3f(0.0728032589f, 0.57600832f, -1.f),
    C_Vertex3f(0.0566205978f, 0.581918716f, -1.f),
    C_Vertex3f(0.0531768799f, 0.516181946f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0479354858f, 0.584976196f, -1.f),
    C_Vertex3f(0.0278320313f, 0.522949219f, -1.f),
    C_Vertex3f(0.0531768799f, 0.516181946f, -1.f),
    C_Vertex3f(0.0566205978f, 0.581918716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.203613281f, 0.651367188f, -1.f),
    C_Vertex3f(0.140136719f, 0.651367188f, -1.f),
    C_Vertex3f(0.140191078f, 0.647654533f, -1.f),
    C_Vertex3f(0.140546322f, 0.638686657f, -1.f),
    C_Vertex3f(0.14114809f, 0.628176212f, -1.f),
    C_Vertex3f(0.141996384f, 0.616123199f, -1.f),
    C_Vertex3f(0.143091202f, 0.602527618f, -1.f),
    C_Vertex3f(0.144432545f, 0.587389469f, -1.f),
    C_Vertex3f(0.146020412f, 0.570708752f, -1.f),
    C_Vertex3f(0.147854805f, 0.552485466f, -1.f),
    C_Vertex3f(0.148895264f, 0.542602539f, -1.f),
    C_Vertex3f(0.162685394f, 0.457368851f, -1.f),
    C_Vertex3f(0.170783997f, 0.472755432f, -1.f),
    C_Vertex3f(0.179428101f, 0.457914352f, -1.f),
    C_Vertex3f(0.19354248f, 0.543045044f, -1.f),
    C_Vertex3f(0.195732117f, 0.559314728f, -1.f),
    C_Vertex3f(0.199180603f, 0.58912468f, -1.f),
    C_Vertex3f(0.201698303f, 0.616205215f, -1.f),
    C_Vertex3f(0.203285217f, 0.640556335f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.283218384f, 0.416007996f, -1.f),
    C_Vertex3f(0.2771492f, 0.423428535f, -1.f),
    C_Vertex3f(0.265694618f, 0.437137365f, -1.f),
    C_Vertex3f(0.25492382f, 0.449713945f, -1.f),
    C_Vertex3f(0.244836807f, 0.461158276f, -1.f),
    C_Vertex3f(0.235433578f, 0.471470356f, -1.f),
    C_Vertex3f(0.230728149f, 0.37890625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.226714134f, 0.480650187f, -1.f),
    C_Vertex3f(0.218678474f, 0.488697767f, -1.f),
    C_Vertex3f(0.224386215f, 0.387527466f, -1.f),
    C_Vertex3f(0.230728149f, 0.37890625f, -1.f),
    C_Vertex3f(0.235433578f, 0.471470356f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.22897625f, 0.559642792f, -1.f),
    C_Vertex3f(0.204807281f, 0.548723221f, -1.f),
    C_Vertex3f(0.207992554f, 0.498504639f, -1.f),
    C_Vertex3f(0.210551262f, 0.407879829f, -1.f),
    C_Vertex3f(0.211326599f, 0.495613098f, -1.f),
    C_Vertex3f(0.224386215f, 0.387527466f, -1.f),
    C_Vertex3f(0.218678474f, 0.488697767f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.195565224f, 0.431342125f, -1.f),
    C_Vertex3f(0.210551262f, 0.407879829f, -1.f),
    C_Vertex3f(0.204807281f, 0.548723221f, -1.f),
    C_Vertex3f(0.19354248f, 0.543045044f, -1.f),
    C_Vertex3f(0.179428101f, 0.457914352f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.207992554f, 0.498504639f, -1.f),
    C_Vertex3f(0.221330643f, 0.500032425f, -1.f),
    C_Vertex3f(0.22897625f, 0.559642792f, -1.f),
    C_Vertex3f(0.248281479f, 0.504615784f, -1.f),
    C_Vertex3f(0.254784584f, 0.57012558f, -1.f),
    C_Vertex3f(0.275506973f, 0.510726929f, -1.f),
    C_Vertex3f(0.282232285f, 0.580171585f, -1.f),
    C_Vertex3f(0.303007126f, 0.51836586f, -1.f),
    C_Vertex3f(0.296775818f, 0.584976196f, -1.f),
    C_Vertex3f(0.316894531f, 0.522949219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147478104f, 0.430251122f, -1.f),
    C_Vertex3f(0.162685394f, 0.457368851f, -1.f),
    C_Vertex3f(0.148895264f, 0.542602539f, -1.f),
    C_Vertex3f(0.144960403f, 0.544841766f, -1.f),
    C_Vertex3f(0.13590312f, 0.549524784f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11036396f, 0.472996235f, -1.f),
    C_Vertex3f(0.11391449f, 0.37890625f, -1.f),
    C_Vertex3f(0.118527889f, 0.481985331f, -1.f),
    C_Vertex3f(0.120033264f, 0.386981964f, -1.f),
    C_Vertex3f(0.125790119f, 0.489651442f, -1.f),
    C_Vertex3f(0.133260727f, 0.406788826f, -1.f),
    C_Vertex3f(0.13215065f, 0.495994568f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.11391449f, 0.37890625f, -1.f),
    C_Vertex3f(0.11036396f, 0.472996235f, -1.f),
    C_Vertex3f(0.101298332f, 0.462684155f, -1.f),
    C_Vertex3f(0.0913310051f, 0.451049089f, -1.f),
    C_Vertex3f(0.0804619789f, 0.43809104f, -1.f),
    C_Vertex3f(0.0686912537f, 0.423810005f, -1.f),
    C_Vertex3f(0.0623550415f, 0.416007996f, -1.f),
    C_End(),
    C_Translatef(0.348144531f, 0.f, 0.f),
};
static const Call char_43[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.224121094f, 0.103515625f, 0.f),
    C_Vertex3f(0.224121094f, 0.103515625f, -1.f),
    C_Vertex3f(0.224121094f, 0.279296875f, -1.f),
    C_Vertex3f(0.224121094f, 0.279296875f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.224121094f, 0.279296875f, 0.f),
    C_Vertex3f(0.224121094f, 0.279296875f, -1.f),
    C_Vertex3f(0.0498046875f, 0.279296875f, -1.f),
    C_Vertex3f(0.0498046875f, 0.279296875f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0498046875f, 0.279296875f, 0.f),
    C_Vertex3f(0.0498046875f, 0.279296875f, -1.f),
    C_Vertex3f(0.0498046875f, 0.352539063f, -1.f),
    C_Vertex3f(0.0498046875f, 0.352539063f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0498046875f, 0.352539063f, 0.f),
    C_Vertex3f(0.0498046875f, 0.352539063f, -1.f),
    C_Vertex3f(0.224121094f, 0.352539063f, -1.f),
    C_Vertex3f(0.224121094f, 0.352539063f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.224121094f, 0.352539063f, 0.f),
    C_Vertex3f(0.224121094f, 0.352539063f, -1.f),
    C_Vertex3f(0.224121094f, 0.526855469f, -1.f),
    C_Vertex3f(0.224121094f, 0.526855469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.224121094f, 0.526855469f, 0.f),
    C_Vertex3f(0.224121094f, 0.526855469f, -1.f),
    C_Vertex3f(0.298339844f, 0.526855469f, -1.f),
    C_Vertex3f(0.298339844f, 0.526855469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.298339844f, 0.526855469f, 0.f),
    C_Vertex3f(0.298339844f, 0.526855469f, -1.f),
    C_Vertex3f(0.298339844f, 0.352539063f, -1.f),
    C_Vertex3f(0.298339844f, 0.352539063f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.298339844f, 0.352539063f, 0.f),
    C_Vertex3f(0.298339844f, 0.352539063f, -1.f),
    C_Vertex3f(0.47265625f, 0.352539063f, -1.f),
    C_Vertex3f(0.47265625f, 0.352539063f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.47265625f, 0.352539063f, 0.f),
    C_Vertex3f(0.47265625f, 0.352539063f, -1.f),
    C_Vertex3f(0.47265625f, 0.279296875f, -1.f),
    C_Vertex3f(0.47265625f, 0.279296875f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.47265625f, 0.279296875f, 0.f),
    C_Vertex3f(0.47265625f, 0.279296875f, -1.f),
    C_Vertex3f(0.298339844f, 0.279296875f, -1.f),
    C_Vertex3f(0.298339844f, 0.279296875f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.298339844f, 0.279296875f, 0.f),
    C_Vertex3f(0.298339844f, 0.279296875f, -1.f),
    C_Vertex3f(0.298339844f, 0.103515625f, -1.f),
    C_Vertex3f(0.298339844f, 0.103515625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.298339844f, 0.103515625f, 0.f),
    C_Vertex3f(0.298339844f, 0.103515625f, -1.f),
    C_Vertex3f(0.224121094f, 0.103515625f, -1.f),
    C_Vertex3f(0.224121094f, 0.103515625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.298339844f, 0.526855469f, 0.f),
    C_Vertex3f(0.224121094f, 0.526855469f, 0.f),
    C_Vertex3f(0.298339844f, 0.103515625f, 0.f),
    C_Vertex3f(0.224121094f, 0.352539063f, 0.f),
    C_Vertex3f(0.224121094f, 0.279296875f, 0.f),
    C_Vertex3f(0.0498046875f, 0.352539063f, 0.f),
    C_Vertex3f(0.0498046875f, 0.279296875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.298339844f, 0.279296875f, 0.f),
    C_Vertex3f(0.47265625f, 0.279296875f, 0.f),
    C_Vertex3f(0.298339844f, 0.352539063f, 0.f),
    C_Vertex3f(0.298339844f, 0.526855469f, 0.f),
    C_Vertex3f(0.298339844f, 0.103515625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.298339844f, 0.352539063f, 0.f),
    C_Vertex3f(0.47265625f, 0.279296875f, 0.f),
    C_Vertex3f(0.47265625f, 0.352539063f, 0.f),
    C_Vertex3f(0.298339844f, 0.103515625f, 0.f),
    C_Vertex3f(0.224121094f, 0.279296875f, 0.f),
    C_Vertex3f(0.224121094f, 0.103515625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.298339844f, 0.526855469f, -1.f),
    C_Vertex3f(0.224121094f, 0.526855469f, -1.f),
    C_Vertex3f(0.298339844f, 0.103515625f, -1.f),
    C_Vertex3f(0.224121094f, 0.352539063f, -1.f),
    C_Vertex3f(0.224121094f, 0.279296875f, -1.f),
    C_Vertex3f(0.0498046875f, 0.352539063f, -1.f),
    C_Vertex3f(0.0498046875f, 0.279296875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.298339844f, 0.279296875f, -1.f),
    C_Vertex3f(0.47265625f, 0.279296875f, -1.f),
    C_Vertex3f(0.298339844f, 0.352539063f, -1.f),
    C_Vertex3f(0.298339844f, 0.526855469f, -1.f),
    C_Vertex3f(0.298339844f, 0.103515625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.298339844f, 0.352539063f, -1.f),
    C_Vertex3f(0.47265625f, 0.279296875f, -1.f),
    C_Vertex3f(0.47265625f, 0.352539063f, -1.f),
    C_Vertex3f(0.298339844f, 0.103515625f, -1.f),
    C_Vertex3f(0.224121094f, 0.279296875f, -1.f),
    C_Vertex3f(0.224121094f, 0.103515625f, -1.f),
    C_End(),
    C_Translatef(0.522460938f, 0.f, 0.f),
};
static const Call char_44[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0795898438f, 0.f, 0.f),
    C_Vertex3f(0.0795898438f, 0.f, -1.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, -1.f),
    C_Vertex3f(0.168945313f, 0.0893554688f, -1.f),
    C_Vertex3f(0.168945313f, 0.0893554688f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.168945313f, 0.0893554688f, 0.f),
    C_Vertex3f(0.168945313f, 0.0893554688f, -1.f),
    C_Normal3f(0.999845922f, -0.017552482f, 0.f),
    C_Vertex3f(0.168945313f, -0.0123672485f, 0.f),
    C_Vertex3f(0.168945313f, -0.0123672485f, -1.f),
    C_Normal3f(0.988292277f, -0.152572513f, 0.f),
    C_Vertex3f(0.16676712f, -0.0347213745f, 0.f),
    C_Vertex3f(0.16676712f, -0.0347213745f, -1.f),
    C_Normal3f(0.960453153f, -0.278441697f, 0.f),
    C_Vertex3f(0.162410736f, -0.0546951294f, 0.f),
    C_Vertex3f(0.162410736f, -0.0546951294f, -1.f),
    C_Normal3f(0.906842053f, -0.421470672f, 0.f),
    C_Vertex3f(0.15587616f, -0.0722885132f, 0.f),
    C_Vertex3f(0.15587616f, -0.0722885132f, -1.f),
    C_Normal3f(0.820653856f, -0.571425498f, 0.f),
    C_Vertex3f(0.147163391f, -0.0875015259f, 0.f),
    C_Vertex3f(0.147163391f, -0.0875015259f, -1.f),
    C_Normal3f(0.709192693f, -0.705014646f, 0.f),
    C_Vertex3f(0.135890961f, -0.100990295f, 0.f),
    C_Vertex3f(0.135890961f, -0.100990295f, -1.f),
    C_Normal3f(0.585075676f, -0.810978711f, 0.f),
    C_Vertex3f(0.122058868f, -0.112754822f, 0.f),
    C_Vertex3f(0.122058868f, -0.112754822f, -1.f),
    C_Normal3f(0.481169224f, -0.876627743f, 0.f),
    C_Vertex3f(0.105667114f, -0.122795105f, 0.f),
    C_Vertex3f(0.105667114f, -0.122795105f, -1.f),
    C_Normal3f(0.401824236f, -0.915716887f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.837610602f, -0.546267807f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, -1.f),
    C_Vertex3f(0.07421875f, -0.0932617188f, -1.f),
    C_Vertex3f(0.07421875f, -0.0932617188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.402259469f, 0.915525675f, 0.f),
    C_Vertex3f(0.07421875f, -0.0932617188f, 0.f),
    C_Vertex3f(0.07421875f, -0.0932617188f, -1.f),
    C_Normal3f(-0.486655027f, 0.873594284f, 0.f),
    C_Vertex3f(0.0804481506f, -0.0905246735f, 0.f),
    C_Vertex3f(0.0804481506f, -0.0905246735f, -1.f),
    C_Normal3f(-0.595042884f, 0.80369401f, 0.f),
    C_Vertex3f(0.0912671089f, -0.0837645531f, 0.f),
    C_Vertex3f(0.0912671089f, -0.0837645531f, -1.f),
    C_Normal3f(-0.720657885f, 0.693290889f, 0.f),
    C_Vertex3f(0.100446224f, -0.0757184029f, 0.f),
    C_Vertex3f(0.100446224f, -0.0757184029f, -1.f),
    C_Normal3f(-0.829384446f, 0.558678269f, 0.f),
    C_Vertex3f(0.107985497f, -0.0663862228f, 0.f),
    C_Vertex3f(0.107985497f, -0.0663862228f, -1.f),
    C_Normal3f(-0.913580418f, 0.406658173f, 0.f),
    C_Vertex3f(0.113884926f, -0.055768013f, 0.f),
    C_Vertex3f(0.113884926f, -0.055768013f, -1.f),
    C_Normal3f(-0.963258266f, 0.268576592f, 0.f),
    C_Vertex3f(0.11847353f, -0.0428242683f, 0.f),
    C_Vertex3f(0.11847353f, -0.0428242683f, -1.f),
    C_Normal3f(-0.987503946f, 0.157594293f, 0.f),
    C_Vertex3f(0.121751308f, -0.0275549889f, 0.f),
    C_Vertex3f(0.121751308f, -0.0275549889f, -1.f),
    C_Normal3f(-0.996549428f, 0.0830012336f, 0.f),
    C_Vertex3f(0.123718262f, -0.00996017456f, 0.f),
    C_Vertex3f(0.123718262f, -0.00996017456f, -1.f),
    C_Normal3f(-0.999457955f, 0.0329197198f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, -1.f),
    C_Vertex3f(0.0795898438f, 0.f, -1.f),
    C_Vertex3f(0.0795898438f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.168945313f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, 0.f),
    C_Vertex3f(0.135890961f, -0.100990295f, 0.f),
    C_Vertex3f(0.147163391f, -0.0875015259f, 0.f),
    C_Vertex3f(0.15587616f, -0.0722885132f, 0.f),
    C_Vertex3f(0.162410736f, -0.0546951294f, 0.f),
    C_Vertex3f(0.16676712f, -0.0347213745f, 0.f),
    C_Vertex3f(0.168945313f, -0.0123672485f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123718262f, -0.00996017456f, 0.f),
    C_Vertex3f(0.121751308f, -0.0275549889f, 0.f),
    C_Vertex3f(0.122058868f, -0.112754822f, 0.f),
    C_Vertex3f(0.135890961f, -0.100990295f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122058868f, -0.112754822f, 0.f),
    C_Vertex3f(0.121751308f, -0.0275549889f, 0.f),
    C_Vertex3f(0.11847353f, -0.0428242683f, 0.f),
    C_Vertex3f(0.113884926f, -0.055768013f, 0.f),
    C_Vertex3f(0.107985497f, -0.0663862228f, 0.f),
    C_Vertex3f(0.105667114f, -0.122795105f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.100446224f, -0.0757184029f, 0.f),
    C_Vertex3f(0.0912671089f, -0.0837645531f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, 0.f),
    C_Vertex3f(0.105667114f, -0.122795105f, 0.f),
    C_Vertex3f(0.107985497f, -0.0663862228f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0804481506f, -0.0905246735f, 0.f),
    C_Vertex3f(0.07421875f, -0.0932617188f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, 0.f),
    C_Vertex3f(0.0912671089f, -0.0837645531f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.124046326f, 0.f, 0.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0795898438f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.168945313f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, -1.f),
    C_Vertex3f(0.124046326f, 0.f, -1.f),
    C_Vertex3f(0.135890961f, -0.100990295f, -1.f),
    C_Vertex3f(0.147163391f, -0.0875015259f, -1.f),
    C_Vertex3f(0.15587616f, -0.0722885132f, -1.f),
    C_Vertex3f(0.162410736f, -0.0546951294f, -1.f),
    C_Vertex3f(0.16676712f, -0.0347213745f, -1.f),
    C_Vertex3f(0.168945313f, -0.0123672485f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123718262f, -0.00996017456f, -1.f),
    C_Vertex3f(0.121751308f, -0.0275549889f, -1.f),
    C_Vertex3f(0.122058868f, -0.112754822f, -1.f),
    C_Vertex3f(0.135890961f, -0.100990295f, -1.f),
    C_Vertex3f(0.124046326f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122058868f, -0.112754822f, -1.f),
    C_Vertex3f(0.121751308f, -0.0275549889f, -1.f),
    C_Vertex3f(0.11847353f, -0.0428242683f, -1.f),
    C_Vertex3f(0.113884926f, -0.055768013f, -1.f),
    C_Vertex3f(0.107985497f, -0.0663862228f, -1.f),
    C_Vertex3f(0.105667114f, -0.122795105f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.100446224f, -0.0757184029f, -1.f),
    C_Vertex3f(0.0912671089f, -0.0837645531f, -1.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, -1.f),
    C_Vertex3f(0.105667114f, -0.122795105f, -1.f),
    C_Vertex3f(0.107985497f, -0.0663862228f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0804481506f, -0.0905246735f, -1.f),
    C_Vertex3f(0.07421875f, -0.0932617188f, -1.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, -1.f),
    C_Vertex3f(0.0912671089f, -0.0837645531f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.124046326f, 0.f, -1.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0795898438f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_45[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0278320313f, 0.192382813f, 0.f),
    C_Vertex3f(0.0278320313f, 0.192382813f, -1.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, -1.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, 0.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, -1.f),
    C_Vertex3f(0.270019531f, 0.271484375f, -1.f),
    C_Vertex3f(0.270019531f, 0.271484375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.270019531f, 0.271484375f, 0.f),
    C_Vertex3f(0.270019531f, 0.271484375f, -1.f),
    C_Vertex3f(0.270019531f, 0.192382813f, -1.f),
    C_Vertex3f(0.270019531f, 0.192382813f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.270019531f, 0.192382813f, 0.f),
    C_Vertex3f(0.270019531f, 0.192382813f, -1.f),
    C_Vertex3f(0.0278320313f, 0.192382813f, -1.f),
    C_Vertex3f(0.0278320313f, 0.192382813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0278320313f, 0.192382813f, 0.f),
    C_Vertex3f(0.270019531f, 0.192382813f, 0.f),
    C_Vertex3f(0.270019531f, 0.271484375f, 0.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0278320313f, 0.192382813f, -1.f),
    C_Vertex3f(0.270019531f, 0.192382813f, -1.f),
    C_Vertex3f(0.270019531f, 0.271484375f, -1.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_46[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0810546875f, 0.f, 0.f),
    C_Vertex3f(0.0810546875f, 0.f, -1.f),
    C_Vertex3f(0.0810546875f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0810546875f, 0.0893554688f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0810546875f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0810546875f, 0.0893554688f, -1.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, -1.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, 0.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, -1.f),
    C_Vertex3f(0.170410156f, 0.f, -1.f),
    C_Vertex3f(0.170410156f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.170410156f, 0.f, 0.f),
    C_Vertex3f(0.170410156f, 0.f, -1.f),
    C_Vertex3f(0.0810546875f, 0.f, -1.f),
    C_Vertex3f(0.0810546875f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0810546875f, 0.f, 0.f),
    C_Vertex3f(0.170410156f, 0.f, 0.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0810546875f, 0.0893554688f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0810546875f, 0.f, -1.f),
    C_Vertex3f(0.170410156f, 0.f, -1.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0810546875f, 0.0893554688f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_47[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.96290499f, 0.269840628f, 0.f),
    C_Vertex3f(0.f, -0.0107421875f, 0.f),
    C_Vertex3f(0.f, -0.0107421875f, -1.f),
    C_Vertex3f(0.185546875f, 0.651367188f, -1.f),
    C_Vertex3f(0.185546875f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.185546875f, 0.651367188f, 0.f),
    C_Vertex3f(0.185546875f, 0.651367188f, -1.f),
    C_Vertex3f(0.248916626f, 0.651367188f, -1.f),
    C_Vertex3f(0.248916626f, 0.651367188f, 0.f),
    C_Normal3f(0.96290499f, -0.269840628f, 0.f),
    C_Vertex3f(0.248916626f, 0.651367188f, 0.f),
    C_Vertex3f(0.248916626f, 0.651367188f, -1.f),
    C_Vertex3f(0.063369751f, -0.0107421875f, -1.f),
    C_Vertex3f(0.063369751f, -0.0107421875f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.063369751f, -0.0107421875f, 0.f),
    C_Vertex3f(0.063369751f, -0.0107421875f, -1.f),
    C_Vertex3f(0.f, -0.0107421875f, -1.f),
    C_Vertex3f(0.f, -0.0107421875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.f, -0.0107421875f, 0.f),
    C_Vertex3f(0.063369751f, -0.0107421875f, 0.f),
    C_Vertex3f(0.248916626f, 0.651367188f, 0.f),
    C_Vertex3f(0.185546875f, 0.651367188f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.f, -0.0107421875f, -1.f),
    C_Vertex3f(0.063369751f, -0.0107421875f, -1.f),
    C_Vertex3f(0.248916626f, 0.651367188f, -1.f),
    C_Vertex3f(0.185546875f, 0.651367188f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_48[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999683976f, -0.0251354929f, 0.f),
    C_Vertex3f(0.037109375f, 0.293394089f, 0.f),
    C_Vertex3f(0.037109375f, 0.293394089f, -1.f),
    C_Normal3f(-0.999934494f, 0.0114529217f, 0.f),
    C_Vertex3f(0.037109375f, 0.330135345f, 0.f),
    C_Vertex3f(0.037109375f, 0.330135345f, -1.f),
    C_Normal3f(-0.999135077f, 0.041581504f, 0.f),
    C_Vertex3f(0.0378395319f, 0.357142806f, 0.f),
    C_Vertex3f(0.0378395319f, 0.357142806f, -1.f),
    C_Normal3f(-0.997331142f, 0.0730101019f, 0.f),
    C_Vertex3f(0.0392998457f, 0.38276875f, 0.f),
    C_Vertex3f(0.0392998457f, 0.38276875f, -1.f),
    C_Normal3f(-0.994165421f, 0.107865937f, 0.f),
    C_Vertex3f(0.0414903164f, 0.407013178f, 0.f),
    C_Vertex3f(0.0414903164f, 0.407013178f, -1.f),
    C_Normal3f(-0.989197373f, 0.146589816f, 0.f),
    C_Vertex3f(0.044410944f, 0.429876089f, 0.f),
    C_Vertex3f(0.044410944f, 0.429876089f, -1.f),
    C_Normal3f(-0.981851399f, 0.189651757f, 0.f),
    C_Vertex3f(0.0480617285f, 0.451357484f, 0.f),
    C_Vertex3f(0.0480617285f, 0.451357484f, -1.f),
    C_Normal3f(-0.971381009f, 0.237527177f, 0.f),
    C_Vertex3f(0.0524426699f, 0.471457362f, 0.f),
    C_Vertex3f(0.0524426699f, 0.471457362f, -1.f),
    C_Normal3f(-0.956828773f, 0.290652245f, 0.f),
    C_Vertex3f(0.0575537682f, 0.490175724f, 0.f),
    C_Vertex3f(0.0575537682f, 0.490175724f, -1.f),
    C_Normal3f(-0.938479424f, 0.345335066f, 0.f),
    C_Vertex3f(0.0633950233f, 0.507512569f, 0.f),
    C_Vertex3f(0.0633950233f, 0.507512569f, -1.f),
    C_Normal3f(-0.91668582f, 0.399608523f, 0.f),
    C_Vertex3f(0.0699458122f, 0.523852229f, 0.f),
    C_Vertex3f(0.0699458122f, 0.523852229f, -1.f),
    C_Normal3f(-0.889748216f, 0.456451535f, 0.f),
    C_Vertex3f(0.0772061348f, 0.539194703f, 0.f),
    C_Vertex3f(0.0772061348f, 0.539194703f, -1.f),
    C_Normal3f(-0.857039094f, 0.515251279f, 0.f),
    C_Vertex3f(0.0851759911f, 0.553539991f, 0.f),
    C_Vertex3f(0.0851759911f, 0.553539991f, -1.f),
    C_Normal3f(-0.818048716f, 0.575148821f, 0.f),
    C_Vertex3f(0.093855381f, 0.566888094f, 0.f),
    C_Vertex3f(0.093855381f, 0.566888094f, -1.f),
    C_Normal3f(-0.772475183f, 0.635044992f, 0.f),
    C_Vertex3f(0.103244305f, 0.579239011f, 0.f),
    C_Vertex3f(0.103244305f, 0.579239011f, -1.f),
    C_Normal3f(-0.720315933f, 0.693646073f, 0.f),
    C_Vertex3f(0.113342762f, 0.590592742f, 0.f),
    C_Vertex3f(0.113342762f, 0.590592742f, -1.f),
    C_Normal3f(-0.661940992f, 0.749555945f, 0.f),
    C_Vertex3f(0.124150753f, 0.600949287f, 0.f),
    C_Vertex3f(0.124150753f, 0.600949287f, -1.f),
    C_Normal3f(-0.593803763f, 0.804609954f, 0.f),
    C_Vertex3f(0.135668278f, 0.610308647f, 0.f),
    C_Vertex3f(0.135668278f, 0.610308647f, -1.f),
    C_Normal3f(-0.515721619f, 0.85675627f, 0.f),
    C_Vertex3f(0.147929549f, 0.618498087f, 0.f),
    C_Vertex3f(0.147929549f, 0.618498087f, -1.f),
    C_Normal3f(-0.433478385f, 0.901163995f, 0.f),
    C_Vertex3f(0.160934567f, 0.625517607f, 0.f),
    C_Vertex3f(0.160934567f, 0.625517607f, -1.f),
    C_Normal3f(-0.349342674f, 0.936995029f, 0.f),
    C_Vertex3f(0.174683332f, 0.631367207f, 0.f),
    C_Vertex3f(0.174683332f, 0.631367207f, -1.f),
    C_Normal3f(-0.265579373f, 0.964088917f, 0.f),
    C_Vertex3f(0.189175844f, 0.636046886f, 0.f),
    C_Vertex3f(0.189175844f, 0.636046886f, -1.f),
    C_Normal3f(-0.184183598f, 0.982891858f, 0.f),
    C_Vertex3f(0.204412103f, 0.639556646f, 0.f),
    C_Vertex3f(0.204412103f, 0.639556646f, -1.f),
    C_Normal3f(-0.106707044f, 0.99429059f, 0.f),
    C_Vertex3f(0.220392108f, 0.641896486f, 0.f),
    C_Vertex3f(0.220392108f, 0.641896486f, -1.f),
    C_Normal3f(-0.0305003598f, 0.999534786f, 0.f),
    C_Vertex3f(0.23711586f, 0.643066406f, 0.f),
    C_Vertex3f(0.23711586f, 0.643066406f, -1.f),
    C_Normal3f(0.0565099902f, 0.998402059f, 0.f),
    C_Vertex3f(0.258731842f, 0.643066406f, 0.f),
    C_Vertex3f(0.258731842f, 0.643066406f, -1.f),
    C_Normal3f(0.163879335f, 0.986480474f, 0.f),
    C_Vertex3f(0.282928944f, 0.640473366f, 0.f),
    C_Vertex3f(0.282928944f, 0.640473366f, -1.f),
    C_Normal3f(0.284479469f, 0.958682239f, 0.f),
    C_Vertex3f(0.305558681f, 0.635287285f, 0.f),
    C_Vertex3f(0.305558681f, 0.635287285f, -1.f),
    C_Normal3f(0.40850091f, 0.912757993f, 0.f),
    C_Vertex3f(0.326621056f, 0.627508163f, 0.f),
    C_Vertex3f(0.326621056f, 0.627508163f, -1.f),
    C_Normal3f(0.525819063f, 0.850596368f, 0.f),
    C_Vertex3f(0.346116066f, 0.617136002f, 0.f),
    C_Vertex3f(0.346116066f, 0.617136002f, -1.f),
    C_Normal3f(0.632587612f, 0.774488747f, 0.f),
    C_Vertex3f(0.363887787f, 0.60447073f, 0.f),
    C_Vertex3f(0.363887787f, 0.60447073f, -1.f),
    C_Normal3f(0.727541268f, 0.686063826f, 0.f),
    C_Vertex3f(0.379936218f, 0.589512348f, 0.f),
    C_Vertex3f(0.379936218f, 0.589512348f, -1.f),
    C_Normal3f(0.80699867f, 0.590553164f, 0.f),
    C_Vertex3f(0.39426136f, 0.572260857f, 0.f),
    C_Vertex3f(0.39426136f, 0.572260857f, -1.f),
    C_Normal3f(0.86550349f, 0.500902832f, 0.f),
    C_Vertex3f(0.406863213f, 0.552716255f, 0.f),
    C_Vertex3f(0.406863213f, 0.552716255f, -1.f),
    C_Normal3f(0.906683326f, 0.421811998f, 0.f),
    C_Vertex3f(0.418116093f, 0.531042576f, 0.f),
    C_Vertex3f(0.418116093f, 0.531042576f, -1.f),
    C_Normal3f(0.937511384f, 0.347954512f, 0.f),
    C_Vertex3f(0.42802f, 0.507239819f, 0.f),
    C_Vertex3f(0.42802f, 0.507239819f, -1.f),
    C_Normal3f(0.958574712f, 0.284841239f, 0.f),
    C_Vertex3f(0.436574936f, 0.481307983f, 0.f),
    C_Vertex3f(0.436574936f, 0.481307983f, -1.f),
    C_Normal3f(0.973318875f, 0.229456589f, 0.f),
    C_Vertex3f(0.441979408f, 0.460262299f, 0.f),
    C_Vertex3f(0.441979408f, 0.460262299f, -1.f),
    C_Normal3f(0.983394623f, 0.181480303f, 0.f),
    C_Vertex3f(0.445132017f, 0.445010185f, 0.f),
    C_Vertex3f(0.445132017f, 0.445010185f, -1.f),
    C_Normal3f(0.989652157f, 0.143486992f, 0.f),
    C_Vertex3f(0.447834253f, 0.428536415f, 0.f),
    C_Vertex3f(0.447834253f, 0.428536415f, -1.f),
    C_Normal3f(0.993927479f, 0.110037252f, 0.f),
    C_Vertex3f(0.450086117f, 0.410840988f, 0.f),
    C_Vertex3f(0.450086117f, 0.410840988f, -1.f),
    C_Normal3f(0.996757984f, 0.0804588869f, 0.f),
    C_Vertex3f(0.451887608f, 0.391923904f, 0.f),
    C_Vertex3f(0.451887608f, 0.391923904f, -1.f),
    C_Normal3f(0.998531044f, 0.0541831888f, 0.f),
    C_Vertex3f(0.453238726f, 0.371785164f, 0.f),
    C_Vertex3f(0.453238726f, 0.371785164f, -1.f),
    C_Normal3f(0.999527693f, 0.0307329223f, 0.f),
    C_Vertex3f(0.454139471f, 0.350424767f, 0.f),
    C_Vertex3f(0.454139471f, 0.350424767f, -1.f),
    C_Normal3f(0.999957085f, 0.00927249715f, 0.f),
    C_Vertex3f(0.454589844f, 0.327842712f, 0.f),
    C_Vertex3f(0.454589844f, 0.327842712f, -1.f),
    C_Normal3f(0.999906182f, -0.0136959227f, 0.f),
    C_Vertex3f(0.454589844f, 0.301856041f, 0.f),
    C_Vertex3f(0.454589844f, 0.301856041f, -1.f),
    C_Normal3f(0.999139726f, -0.041468773f, 0.f),
    C_Vertex3f(0.453866601f, 0.275040507f, 0.f),
    C_Vertex3f(0.453866601f, 0.275040507f, -1.f),
    C_Normal3f(0.99734956f, -0.0727601051f, 0.f),
    C_Vertex3f(0.452420115f, 0.249579072f, 0.f),
    C_Vertex3f(0.452420115f, 0.249579072f, -1.f),
    C_Normal3f(0.994214535f, -0.10741245f, 0.f),
    C_Vertex3f(0.450250387f, 0.225471735f, 0.f),
    C_Vertex3f(0.450250387f, 0.225471735f, -1.f),
    C_Normal3f(0.989306867f, -0.145849019f, 0.f),
    C_Vertex3f(0.447357416f, 0.202718496f, 0.f),
    C_Vertex3f(0.447357416f, 0.202718496f, -1.f),
    C_Normal3f(0.982069552f, -0.188519135f, 0.f),
    C_Vertex3f(0.443741202f, 0.181319356f, 0.f),
    C_Vertex3f(0.443741202f, 0.181319356f, -1.f),
    C_Normal3f(0.97178334f, -0.235875204f, 0.f),
    C_Vertex3f(0.439401746f, 0.161274314f, 0.f),
    C_Vertex3f(0.439401746f, 0.161274314f, -1.f),
    C_Normal3f(0.957530975f, -0.288330495f, 0.f),
    C_Vertex3f(0.434339046f, 0.14258337f, 0.f),
    C_Vertex3f(0.434339046f, 0.14258337f, -1.f),
    C_Normal3f(0.939434826f, -0.342727661f, 0.f),
    C_Vertex3f(0.428553104f, 0.125246525f, 0.f),
    C_Vertex3f(0.428553104f, 0.125246525f, -1.f),
    C_Normal3f(0.917726338f, -0.397213221f, 0.f),
    C_Vertex3f(0.422050595f, 0.10890007f, 0.f),
    C_Vertex3f(0.422050595f, 0.10890007f, -1.f),
    C_Normal3f(0.890876234f, -0.454245985f, 0.f),
    C_Vertex3f(0.414831519f, 0.0935440063f, 0.f),
    C_Vertex3f(0.414831519f, 0.0935440063f, -1.f),
    C_Normal3f(0.858266115f, -0.513204873f, 0.f),
    C_Vertex3f(0.406895876f, 0.0791783333f, 0.f),
    C_Vertex3f(0.406895876f, 0.0791783333f, -1.f),
    C_Normal3f(0.819398284f, -0.573224664f, 0.f),
    C_Vertex3f(0.398243666f, 0.065803051f, 0.f),
    C_Vertex3f(0.398243666f, 0.065803051f, -1.f),
    C_Normal3f(0.773985207f, -0.633203804f, 0.f),
    C_Vertex3f(0.388874888f, 0.0534181595f, 0.f),
    C_Vertex3f(0.388874888f, 0.0534181595f, -1.f),
    C_Normal3f(0.722039163f, -0.691852152f, 0.f),
    C_Vertex3f(0.378789544f, 0.0420236588f, 0.f),
    C_Vertex3f(0.378789544f, 0.0420236588f, -1.f),
    C_Normal3f(0.663942337f, -0.74778378f, 0.f),
    C_Vertex3f(0.367987633f, 0.0316195488f, 0.f),
    C_Vertex3f(0.367987633f, 0.0316195488f, -1.f),
    C_Normal3f(0.595781207f, -0.80314678f, 0.f),
    C_Vertex3f(0.356469154f, 0.0222058296f, 0.f),
    C_Vertex3f(0.356469154f, 0.0222058296f, -1.f),
    C_Normal3f(0.517280996f, -0.855815589f, 0.f),
    C_Vertex3f(0.344193459f, 0.0139688253f, 0.f),
    C_Vertex3f(0.344193459f, 0.0139688253f, -1.f),
    C_Normal3f(0.434607655f, -0.900619864f, 0.f),
    C_Vertex3f(0.331160545f, 0.00690853596f, 0.f),
    C_Vertex3f(0.331160545f, 0.00690853596f, -1.f),
    C_Normal3f(0.350076407f, -0.936721087f, 0.f),
    C_Vertex3f(0.317370415f, 0.00102496147f, 0.f),
    C_Vertex3f(0.317370415f, 0.00102496147f, -1.f),
    C_Normal3f(0.265988827f, -0.963976145f, 0.f),
    C_Vertex3f(0.302823067f, -0.00368189812f, 0.f),
    C_Vertex3f(0.302823067f, -0.00368189812f, -1.f),
    C_Normal3f(0.184360713f, -0.982858717f, 0.f),
    C_Vertex3f(0.287518501f, -0.00721204281f, 0.f),
    C_Vertex3f(0.287518501f, -0.00721204281f, -1.f),
    C_Normal3f(0.106748484f, -0.99428606f, 0.f),
    C_Vertex3f(0.271456718f, -0.0095654726f, 0.f),
    C_Vertex3f(0.271456718f, -0.0095654726f, -1.f),
    C_Normal3f(0.0316333435f, -0.999499559f, 0.f),
    C_Vertex3f(0.254637718f, -0.0107421875f, 0.f),
    C_Vertex3f(0.254637718f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0488444343f, -0.998806417f, 0.f),
    C_Vertex3f(0.234276772f, -0.0107421875f, 0.f),
    C_Vertex3f(0.234276772f, -0.0107421875f, -1.f),
    C_Normal3f(-0.144151539f, -0.989555657f, 0.f),
    C_Vertex3f(0.212373257f, -0.00867533684f, 0.f),
    C_Vertex3f(0.212373257f, -0.00867533684f, -1.f),
    C_Normal3f(-0.249671578f, -0.968330562f, 0.f),
    C_Vertex3f(0.191711903f, -0.00454163551f, 0.f),
    C_Vertex3f(0.191711903f, -0.00454163551f, -1.f),
    C_Normal3f(-0.359149218f, -0.93328017f, 0.f),
    C_Vertex3f(0.172292709f, 0.00165891647f, 0.f),
    C_Vertex3f(0.172292709f, 0.00165891647f, -1.f),
    C_Normal3f(-0.468143255f, -0.883652568f, 0.f),
    C_Vertex3f(0.154115677f, 0.00992631912f, 0.f),
    C_Vertex3f(0.154115677f, 0.00992631912f, -1.f),
    C_Normal3f(-0.571706712f, -0.820458114f, 0.f),
    C_Vertex3f(0.137180805f, 0.0202605724f, 0.f),
    C_Vertex3f(0.137180805f, 0.0202605724f, -1.f),
    C_Normal3f(-0.665401399f, -0.74648571f, 0.f),
    C_Vertex3f(0.121488094f, 0.0326616764f, 0.f),
    C_Vertex3f(0.121488094f, 0.0326616764f, -1.f),
    C_Normal3f(-0.74907285f, -0.662487686f, 0.f),
    C_Vertex3f(0.107037544f, 0.047129631f, 0.f),
    C_Vertex3f(0.107037544f, 0.047129631f, -1.f),
    C_Normal3f(-0.825281441f, -0.564721704f, 0.f),
    C_Vertex3f(0.0925178528f, 0.0654182434f, 0.f),
    C_Vertex3f(0.0925178528f, 0.0654182434f, -1.f),
    C_Normal3f(-0.887244105f, -0.461300254f, 0.f),
    C_Vertex3f(0.0786657333f, 0.0885920525f, 0.f),
    C_Vertex3f(0.0786657333f, 0.0885920525f, -1.f),
    C_Normal3f(-0.931494236f, -0.363756001f, 0.f),
    C_Vertex3f(0.0667924881f, 0.114897251f, 0.f),
    C_Vertex3f(0.0667924881f, 0.114897251f, -1.f),
    C_Normal3f(-0.961136937f, -0.276071966f, 0.f),
    C_Vertex3f(0.0568981171f, 0.144333839f, 0.f),
    C_Vertex3f(0.0568981171f, 0.144333839f, -1.f),
    C_Normal3f(-0.98002851f, -0.198857471f, 0.f),
    C_Vertex3f(0.0489826202f, 0.176901817f, 0.f),
    C_Vertex3f(0.0489826202f, 0.176901817f, -1.f),
    C_Normal3f(-0.991302729f, -0.131602049f, 0.f),
    C_Vertex3f(0.0430459976f, 0.212601185f, 0.f),
    C_Vertex3f(0.0430459976f, 0.212601185f, -1.f),
    C_Normal3f(-0.997311234f, -0.0732819363f, 0.f),
    C_Vertex3f(0.0390882492f, 0.251431942f, 0.f),
    C_Vertex3f(0.0390882492f, 0.251431942f, -1.f),
    C_Normal3f(-0.999683976f, -0.0251354929f, 0.f),
    C_Vertex3f(0.037109375f, 0.293394089f, 0.f),
    C_Vertex3f(0.037109375f, 0.293394089f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.99986136f, -0.016653141f, 0.f),
    C_Vertex3f(0.118164063f, 0.335638046f, 0.f),
    C_Vertex3f(0.118164063f, 0.335638046f, -1.f),
    C_Normal3f(0.999884427f, 0.0152069544f, 0.f),
    C_Vertex3f(0.118164063f, 0.296167374f, 0.f),
    C_Vertex3f(0.118164063f, 0.296167374f, -1.f),
    C_Normal3f(0.998753488f, 0.0499157831f, 0.f),
    C_Vertex3f(0.11931479f, 0.259975672f, 0.f),
    C_Vertex3f(0.11931479f, 0.259975672f, -1.f),
    C_Normal3f(0.995783329f, 0.0917361453f, 0.f),
    C_Vertex3f(0.121616244f, 0.227093458f, 0.f),
    C_Vertex3f(0.121616244f, 0.227093458f, -1.f),
    C_Normal3f(0.989753664f, 0.142785326f, 0.f),
    C_Vertex3f(0.125068426f, 0.197520733f, 0.f),
    C_Vertex3f(0.125068426f, 0.197520733f, -1.f),
    C_Normal3f(0.978569567f, 0.205916718f, 0.f),
    C_Vertex3f(0.129671335f, 0.171257496f, 0.f),
    C_Vertex3f(0.129671335f, 0.171257496f, -1.f),
    C_Normal3f(0.958574891f, 0.284840554f, 0.f),
    C_Vertex3f(0.135424972f, 0.148303747f, 0.f),
    C_Vertex3f(0.135424972f, 0.148303747f, -1.f),
    C_Normal3f(0.923366487f, 0.383919686f, 0.f),
    C_Vertex3f(0.142329335f, 0.128659487f, 0.f),
    C_Vertex3f(0.142329335f, 0.128659487f, -1.f),
    C_Normal3f(0.862058759f, 0.506808221f, 0.f),
    C_Vertex3f(0.150384426f, 0.112324715f, 0.f),
    C_Vertex3f(0.150384426f, 0.112324715f, -1.f),
    C_Normal3f(0.790066242f, 0.613021493f, 0.f),
    C_Vertex3f(0.159590244f, 0.0992994308f, 0.f),
    C_Vertex3f(0.159590244f, 0.0992994308f, -1.f),
    C_Normal3f(0.726406097f, 0.687265754f, 0.f),
    C_Vertex3f(0.169334054f, 0.0879023075f, 0.f),
    C_Vertex3f(0.169334054f, 0.0879023075f, -1.f),
    C_Normal3f(0.647090495f, 0.762413144f, 0.f),
    C_Vertex3f(0.179615855f, 0.0781333447f, 0.f),
    C_Vertex3f(0.179615855f, 0.0781333447f, -1.f),
    C_Normal3f(0.551267743f, 0.834328294f, 0.f),
    C_Vertex3f(0.190435648f, 0.0699925423f, 0.f),
    C_Vertex3f(0.190435648f, 0.0699925423f, -1.f),
    C_Normal3f(0.440104902f, 0.897946358f, 0.f),
    C_Vertex3f(0.201793432f, 0.0634799004f, 0.f),
    C_Vertex3f(0.201793432f, 0.0634799004f, -1.f),
    C_Normal3f(0.317313462f, 0.948320746f, 0.f),
    C_Vertex3f(0.213689208f, 0.0585954189f, 0.f),
    C_Vertex3f(0.213689208f, 0.0585954189f, -1.f),
    C_Normal3f(0.188802779f, 0.982015014f, 0.f),
    C_Vertex3f(0.226122975f, 0.055339098f, 0.f),
    C_Vertex3f(0.226122975f, 0.055339098f, -1.f),
    C_Normal3f(0.0613670424f, 0.998115301f, 0.f),
    C_Vertex3f(0.239094734f, 0.0537109375f, 0.f),
    C_Vertex3f(0.239094734f, 0.0537109375f, -1.f),
    C_Normal3f(-0.0616221726f, 0.998099446f, 0.f),
    C_Vertex3f(0.252604485f, 0.0537109375f, 0.f),
    C_Vertex3f(0.252604485f, 0.0537109375f, -1.f),
    C_Normal3f(-0.189562485f, 0.981868684f, 0.f),
    C_Vertex3f(0.265576243f, 0.0553458929f, 0.f),
    C_Vertex3f(0.265576243f, 0.0553458929f, -1.f),
    C_Normal3f(-0.318503648f, 0.947921693f, 0.f),
    C_Vertex3f(0.278010011f, 0.0586158037f, 0.f),
    C_Vertex3f(0.278010011f, 0.0586158037f, -1.f),
    C_Normal3f(-0.441584051f, 0.897219896f, 0.f),
    C_Vertex3f(0.289905787f, 0.0635206699f, 0.f),
    C_Vertex3f(0.289905787f, 0.0635206699f, -1.f),
    C_Normal3f(-0.55286628f, 0.833270013f, 0.f),
    C_Vertex3f(0.301263571f, 0.0700604916f, 0.f),
    C_Vertex3f(0.301263571f, 0.0700604916f, -1.f),
    C_Normal3f(-0.648656189f, 0.761081576f, 0.f),
    C_Vertex3f(0.312083364f, 0.0782352686f, 0.f),
    C_Vertex3f(0.312083364f, 0.0782352686f, -1.f),
    C_Normal3f(-0.727833271f, 0.68575412f, 0.f),
    C_Vertex3f(0.322365165f, 0.088045001f, 0.f),
    C_Vertex3f(0.322365165f, 0.088045001f, -1.f),
    C_Normal3f(-0.791300476f, 0.611427426f, 0.f),
    C_Vertex3f(0.332108974f, 0.0994896889f, 0.f),
    C_Vertex3f(0.332108974f, 0.0994896889f, -1.f),
    C_Normal3f(-0.862722695f, 0.505677164f, 0.f),
    C_Vertex3f(0.341314793f, 0.112569332f, 0.f),
    C_Vertex3f(0.341314793f, 0.112569332f, -1.f),
    C_Normal3f(-0.923546076f, 0.383487403f, 0.f),
    C_Vertex3f(0.349369884f, 0.128938079f, 0.f),
    C_Vertex3f(0.349369884f, 0.128938079f, -1.f),
    C_Normal3f(-0.958587348f, 0.284798831f, 0.f),
    C_Vertex3f(0.356274247f, 0.148595929f, 0.f),
    C_Vertex3f(0.356274247f, 0.148595929f, -1.f),
    C_Normal3f(-0.978540897f, 0.206052914f, 0.f),
    C_Vertex3f(0.362027884f, 0.171542883f, 0.f),
    C_Vertex3f(0.362027884f, 0.171542883f, -1.f),
    C_Normal3f(-0.989726603f, 0.142972812f, 0.f),
    C_Vertex3f(0.366630793f, 0.19777894f, 0.f),
    C_Vertex3f(0.366630793f, 0.19777894f, -1.f),
    C_Normal3f(-0.995767891f, 0.0919046998f, 0.f),
    C_Vertex3f(0.370082974f, 0.227304101f, 0.f),
    C_Vertex3f(0.370082974f, 0.227304101f, -1.f),
    C_Normal3f(-0.998747885f, 0.0500286967f, 0.f),
    C_Vertex3f(0.372384429f, 0.260118365f, 0.f),
    C_Vertex3f(0.372384429f, 0.260118365f, -1.f),
    C_Normal3f(-0.999884069f, 0.0152247241f, 0.f),
    C_Vertex3f(0.373535156f, 0.296221733f, 0.f),
    C_Vertex3f(0.373535156f, 0.296221733f, -1.f),
    C_Normal3f(-0.999884844f, -0.0151739856f, 0.f),
    C_Vertex3f(0.373535156f, 0.335692406f, 0.f),
    C_Vertex3f(0.373535156f, 0.335692406f, -1.f),
    C_Normal3f(-0.998759985f, -0.0497839674f, 0.f),
    C_Vertex3f(0.372386336f, 0.371922851f, 0.f),
    C_Vertex3f(0.372386336f, 0.371922851f, -1.f),
    C_Normal3f(-0.995804012f, -0.0915117562f, 0.f),
    C_Vertex3f(0.370088696f, 0.404834867f, 0.f),
    C_Vertex3f(0.370088696f, 0.404834867f, -1.f),
    C_Normal3f(-0.989798903f, -0.142471522f, 0.f),
    C_Vertex3f(0.366642237f, 0.434428453f, 0.f),
    C_Vertex3f(0.366642237f, 0.434428453f, -1.f),
    C_Normal3f(-0.978650808f, -0.205530241f, 0.f),
    C_Vertex3f(0.362046957f, 0.460703611f, 0.f),
    C_Vertex3f(0.362046957f, 0.460703611f, -1.f),
    C_Normal3f(-0.958698332f, -0.284424961f, 0.f),
    C_Vertex3f(0.356302857f, 0.48366034f, 0.f),
    C_Vertex3f(0.356302857f, 0.48366034f, -1.f),
    C_Normal3f(-0.923513174f, -0.383566618f, 0.f),
    C_Vertex3f(0.349409938f, 0.50329864f, 0.f),
    C_Vertex3f(0.349409938f, 0.50329864f, -1.f),
    C_Normal3f(-0.862133741f, -0.506680787f, 0.f),
    C_Vertex3f(0.341368198f, 0.519618511f, 0.f),
    C_Vertex3f(0.341368198f, 0.519618511f, -1.f),
    C_Normal3f(-0.789484382f, -0.613770545f, 0.f),
    C_Vertex3f(0.332177639f, 0.532619953f, 0.f),
    C_Vertex3f(0.332177639f, 0.532619953f, -1.f),
    C_Normal3f(-0.724608481f, -0.689160764f, 0.f),
    C_Vertex3f(0.322416186f, 0.543996215f, 0.f),
    C_Vertex3f(0.322416186f, 0.543996215f, -1.f),
    C_Normal3f(-0.644017398f, -0.765010834f, 0.f),
    C_Vertex3f(0.31208384f, 0.553747296f, 0.f),
    C_Vertex3f(0.31208384f, 0.553747296f, -1.f),
    C_Normal3f(-0.54712522f, -0.837050736f, 0.f),
    C_Vertex3f(0.301180601f, 0.561873198f, 0.f),
    C_Vertex3f(0.301180601f, 0.561873198f, -1.f),
    C_Normal3f(-0.43543759f, -0.900218904f, 0.f),
    C_Vertex3f(0.289706469f, 0.568373919f, 0.f),
    C_Vertex3f(0.289706469f, 0.568373919f, -1.f),
    C_Normal3f(-0.312953323f, -0.949768484f, 0.f),
    C_Vertex3f(0.277661443f, 0.573249459f, 0.f),
    C_Vertex3f(0.277661443f, 0.573249459f, -1.f),
    C_Normal3f(-0.185668975f, -0.982612371f, 0.f),
    C_Vertex3f(0.265045524f, 0.57649982f, 0.f),
    C_Vertex3f(0.265045524f, 0.57649982f, -1.f),
    C_Normal3f(-0.0604836755f, -0.998169184f, 0.f),
    C_Vertex3f(0.251858711f, 0.578125f, 0.f),
    C_Vertex3f(0.251858711f, 0.578125f, -1.f),
    C_Normal3f(0.0540969484f, -0.998535752f, 0.f),
    C_Vertex3f(0.238224983f, 0.578125f, 0.f),
    C_Vertex3f(0.238224983f, 0.578125f, -1.f),
    C_Normal3f(0.169648081f, -0.985504687f, 0.f),
    C_Vertex3f(0.225392342f, 0.576691151f, 0.f),
    C_Vertex3f(0.225392342f, 0.576691151f, -1.f),
    C_Normal3f(0.290283322f, -0.956940651f, 0.f),
    C_Vertex3f(0.213236809f, 0.573823452f, 0.f),
    C_Vertex3f(0.213236809f, 0.573823452f, -1.f),
    C_Normal3f(0.410741121f, -0.911751986f, 0.f),
    C_Vertex3f(0.201758385f, 0.569521904f, 0.f),
    C_Vertex3f(0.201758385f, 0.569521904f, -1.f),
    C_Normal3f(0.524905562f, -0.851160526f, 0.f),
    C_Vertex3f(0.190957069f, 0.563786507f, 0.f),
    C_Vertex3f(0.190957069f, 0.563786507f, -1.f),
    C_Normal3f(0.627488017f, -0.778626204f, 0.f),
    C_Vertex3f(0.180832863f, 0.55661726f, 0.f),
    C_Vertex3f(0.180832863f, 0.55661726f, -1.f),
    C_Normal3f(0.715174079f, -0.698946297f, 0.f),
    C_Vertex3f(0.171385765f, 0.548014164f, 0.f),
    C_Vertex3f(0.171385765f, 0.548014164f, -1.f),
    C_Normal3f(0.77608031f, -0.630634189f, 0.f),
    C_Vertex3f(0.162615776f, 0.537977219f, 0.f),
    C_Vertex3f(0.162615776f, 0.537977219f, -1.f),
    C_Normal3f(0.820468962f, -0.571691275f, 0.f),
    C_Vertex3f(0.158569336f, 0.532241821f, 0.f),
    C_Vertex3f(0.158569336f, 0.532241821f, -1.f),
    C_Normal3f(0.874653757f, -0.484748274f, 0.f),
    C_Vertex3f(0.153518677f, 0.524921417f, 0.f),
    C_Vertex3f(0.153518677f, 0.524921417f, -1.f),
    C_Normal3f(0.920180082f, -0.391495317f, 0.f),
    C_Vertex3f(0.144680023f, 0.507180691f, 0.f),
    C_Vertex3f(0.144680023f, 0.507180691f, -1.f),
    C_Normal3f(0.955114305f, -0.296237677f, 0.f),
    C_Vertex3f(0.137104034f, 0.486340046f, 0.f),
    C_Vertex3f(0.137104034f, 0.486340046f, -1.f),
    C_Normal3f(0.97604543f, -0.217566431f, 0.f),
    C_Vertex3f(0.13079071f, 0.462399483f, 0.f),
    C_Vertex3f(0.13079071f, 0.462399483f, -1.f),
    C_Normal3f(0.988263428f, -0.152759418f, 0.f),
    C_Vertex3f(0.125740051f, 0.435359001f, 0.f),
    C_Vertex3f(0.125740051f, 0.435359001f, -1.f),
    C_Normal3f(0.995075643f, -0.0991190225f, 0.f),
    C_Vertex3f(0.121952057f, 0.405218601f, 0.f),
    C_Vertex3f(0.121952057f, 0.405218601f, -1.f),
    C_Normal3f(0.998521388f, -0.0543599203f, 0.f),
    C_Vertex3f(0.119426727f, 0.371978283f, 0.f),
    C_Vertex3f(0.119426727f, 0.371978283f, -1.f),
    C_Normal3f(0.99986136f, -0.016653141f, 0.f),
    C_Vertex3f(0.118164063f, 0.335638046f, 0.f),
    C_Vertex3f(0.118164063f, 0.335638046f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.450086117f, 0.410840988f, 0.f),
    C_Vertex3f(0.450250387f, 0.225471735f, 0.f),
    C_Vertex3f(0.451887608f, 0.391923904f, 0.f),
    C_Vertex3f(0.452420115f, 0.249579072f, 0.f),
    C_Vertex3f(0.453238726f, 0.371785164f, 0.f),
    C_Vertex3f(0.453866601f, 0.275040507f, 0.f),
    C_Vertex3f(0.454139471f, 0.350424767f, 0.f),
    C_Vertex3f(0.454589844f, 0.301856041f, 0.f),
    C_Vertex3f(0.454589844f, 0.327842712f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.428553104f, 0.125246525f, 0.f),
    C_Vertex3f(0.434339046f, 0.14258337f, 0.f),
    C_Vertex3f(0.436574936f, 0.481307983f, 0.f),
    C_Vertex3f(0.439401746f, 0.161274314f, 0.f),
    C_Vertex3f(0.441979408f, 0.460262299f, 0.f),
    C_Vertex3f(0.443741202f, 0.181319356f, 0.f),
    C_Vertex3f(0.445132017f, 0.445010185f, 0.f),
    C_Vertex3f(0.447357416f, 0.202718496f, 0.f),
    C_Vertex3f(0.447834253f, 0.428536415f, 0.f),
    C_Vertex3f(0.450250387f, 0.225471735f, 0.f),
    C_Vertex3f(0.450086117f, 0.410840988f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.406895876f, 0.0791783333f, 0.f),
    C_Vertex3f(0.414831519f, 0.0935440063f, 0.f),
    C_Vertex3f(0.418116093f, 0.531042576f, 0.f),
    C_Vertex3f(0.422050595f, 0.10890007f, 0.f),
    C_Vertex3f(0.42802f, 0.507239819f, 0.f),
    C_Vertex3f(0.428553104f, 0.125246525f, 0.f),
    C_Vertex3f(0.436574936f, 0.481307983f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.373535156f, 0.335692406f, 0.f),
    C_Vertex3f(0.378789544f, 0.0420236588f, 0.f),
    C_Vertex3f(0.379936218f, 0.589512348f, 0.f),
    C_Vertex3f(0.388874888f, 0.0534181595f, 0.f),
    C_Vertex3f(0.39426136f, 0.572260857f, 0.f),
    C_Vertex3f(0.398243666f, 0.065803051f, 0.f),
    C_Vertex3f(0.406863213f, 0.552716255f, 0.f),
    C_Vertex3f(0.406895876f, 0.0791783333f, 0.f),
    C_Vertex3f(0.418116093f, 0.531042576f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.379936218f, 0.589512348f, 0.f),
    C_Vertex3f(0.363887787f, 0.60447073f, 0.f),
    C_Vertex3f(0.366642237f, 0.434428453f, 0.f),
    C_Vertex3f(0.370088696f, 0.404834867f, 0.f),
    C_Vertex3f(0.372386336f, 0.371922851f, 0.f),
    C_Vertex3f(0.373535156f, 0.335692406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.363887787f, 0.60447073f, 0.f),
    C_Vertex3f(0.346116066f, 0.617136002f, 0.f),
    C_Vertex3f(0.349409938f, 0.50329864f, 0.f),
    C_Vertex3f(0.356302857f, 0.48366034f, 0.f),
    C_Vertex3f(0.362046957f, 0.460703611f, 0.f),
    C_Vertex3f(0.366642237f, 0.434428453f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.31208384f, 0.553747296f, 0.f),
    C_Vertex3f(0.322416186f, 0.543996215f, 0.f),
    C_Vertex3f(0.326621056f, 0.627508163f, 0.f),
    C_Vertex3f(0.332177639f, 0.532619953f, 0.f),
    C_Vertex3f(0.346116066f, 0.617136002f, 0.f),
    C_Vertex3f(0.341368198f, 0.519618511f, 0.f),
    C_Vertex3f(0.349409938f, 0.50329864f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.305558681f, 0.635287285f, 0.f),
    C_Vertex3f(0.282928944f, 0.640473366f, 0.f),
    C_Vertex3f(0.289706469f, 0.568373919f, 0.f),
    C_Vertex3f(0.301180601f, 0.561873198f, 0.f),
    C_Vertex3f(0.31208384f, 0.553747296f, 0.f),
    C_Vertex3f(0.326621056f, 0.627508163f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.238224983f, 0.578125f, 0.f),
    C_Vertex3f(0.251858711f, 0.578125f, 0.f),
    C_Vertex3f(0.258731842f, 0.643066406f, 0.f),
    C_Vertex3f(0.265045524f, 0.57649982f, 0.f),
    C_Vertex3f(0.282928944f, 0.640473366f, 0.f),
    C_Vertex3f(0.277661443f, 0.573249459f, 0.f),
    C_Vertex3f(0.289706469f, 0.568373919f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.190957069f, 0.563786507f, 0.f),
    C_Vertex3f(0.201758385f, 0.569521904f, 0.f),
    C_Vertex3f(0.204412103f, 0.639556646f, 0.f),
    C_Vertex3f(0.213236809f, 0.573823452f, 0.f),
    C_Vertex3f(0.220392108f, 0.641896486f, 0.f),
    C_Vertex3f(0.225392342f, 0.576691151f, 0.f),
    C_Vertex3f(0.23711586f, 0.643066406f, 0.f),
    C_Vertex3f(0.238224983f, 0.578125f, 0.f),
    C_Vertex3f(0.258731842f, 0.643066406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.162615776f, 0.537977219f, 0.f),
    C_Vertex3f(0.171385765f, 0.548014164f, 0.f),
    C_Vertex3f(0.174683332f, 0.631367207f, 0.f),
    C_Vertex3f(0.180832863f, 0.55661726f, 0.f),
    C_Vertex3f(0.189175844f, 0.636046886f, 0.f),
    C_Vertex3f(0.190957069f, 0.563786507f, 0.f),
    C_Vertex3f(0.204412103f, 0.639556646f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.160934567f, 0.625517607f, 0.f),
    C_Vertex3f(0.147929549f, 0.618498087f, 0.f),
    C_Vertex3f(0.153518677f, 0.524921417f, 0.f),
    C_Vertex3f(0.158569336f, 0.532241821f, 0.f),
    C_Vertex3f(0.162615776f, 0.537977219f, 0.f),
    C_Vertex3f(0.174683332f, 0.631367207f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.125740051f, 0.435359001f, 0.f),
    C_Vertex3f(0.13079071f, 0.462399483f, 0.f),
    C_Vertex3f(0.135668278f, 0.610308647f, 0.f),
    C_Vertex3f(0.137104034f, 0.486340046f, 0.f),
    C_Vertex3f(0.147929549f, 0.618498087f, 0.f),
    C_Vertex3f(0.144680023f, 0.507180691f, 0.f),
    C_Vertex3f(0.153518677f, 0.524921417f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.124150753f, 0.600949287f, 0.f),
    C_Vertex3f(0.113342762f, 0.590592742f, 0.f),
    C_Vertex3f(0.118164063f, 0.335638046f, 0.f),
    C_Vertex3f(0.119426727f, 0.371978283f, 0.f),
    C_Vertex3f(0.121952057f, 0.405218601f, 0.f),
    C_Vertex3f(0.125740051f, 0.435359001f, 0.f),
    C_Vertex3f(0.135668278f, 0.610308647f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.113342762f, 0.590592742f, 0.f),
    C_Vertex3f(0.103244305f, 0.579239011f, 0.f),
    C_Vertex3f(0.107037544f, 0.047129631f, 0.f),
    C_Vertex3f(0.121488094f, 0.0326616764f, 0.f),
    C_Vertex3f(0.118164063f, 0.296167374f, 0.f),
    C_Vertex3f(0.118164063f, 0.335638046f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378789544f, 0.0420236588f, 0.f),
    C_Vertex3f(0.373535156f, 0.335692406f, 0.f),
    C_Vertex3f(0.373535156f, 0.296221733f, 0.f),
    C_Vertex3f(0.372384429f, 0.260118365f, 0.f),
    C_Vertex3f(0.370082974f, 0.227304101f, 0.f),
    C_Vertex3f(0.367987633f, 0.0316195488f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.370082974f, 0.227304101f, 0.f),
    C_Vertex3f(0.366630793f, 0.19777894f, 0.f),
    C_Vertex3f(0.367987633f, 0.0316195488f, 0.f),
    C_Vertex3f(0.362027884f, 0.171542883f, 0.f),
    C_Vertex3f(0.356469154f, 0.0222058296f, 0.f),
    C_Vertex3f(0.356274247f, 0.148595929f, 0.f),
    C_Vertex3f(0.349369884f, 0.128938079f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.344193459f, 0.0139688253f, 0.f),
    C_Vertex3f(0.356469154f, 0.0222058296f, 0.f),
    C_Vertex3f(0.349369884f, 0.128938079f, 0.f),
    C_Vertex3f(0.341314793f, 0.112569332f, 0.f),
    C_Vertex3f(0.332108974f, 0.0994896889f, 0.f),
    C_Vertex3f(0.331160545f, 0.00690853596f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.301263571f, 0.0700604916f, 0.f),
    C_Vertex3f(0.302823067f, -0.00368189812f, 0.f),
    C_Vertex3f(0.312083364f, 0.0782352686f, 0.f),
    C_Vertex3f(0.317370415f, 0.00102496147f, 0.f),
    C_Vertex3f(0.322365165f, 0.088045001f, 0.f),
    C_Vertex3f(0.331160545f, 0.00690853596f, 0.f),
    C_Vertex3f(0.332108974f, 0.0994896889f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.252604485f, 0.0537109375f, 0.f),
    C_Vertex3f(0.254637718f, -0.0107421875f, 0.f),
    C_Vertex3f(0.265576243f, 0.0553458929f, 0.f),
    C_Vertex3f(0.271456718f, -0.0095654726f, 0.f),
    C_Vertex3f(0.278010011f, 0.0586158037f, 0.f),
    C_Vertex3f(0.287518501f, -0.00721204281f, 0.f),
    C_Vertex3f(0.289905787f, 0.0635206699f, 0.f),
    C_Vertex3f(0.302823067f, -0.00368189812f, 0.f),
    C_Vertex3f(0.301263571f, 0.0700604916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.239094734f, 0.0537109375f, 0.f),
    C_Vertex3f(0.226122975f, 0.055339098f, 0.f),
    C_Vertex3f(0.234276772f, -0.0107421875f, 0.f),
    C_Vertex3f(0.254637718f, -0.0107421875f, 0.f),
    C_Vertex3f(0.252604485f, 0.0537109375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.190435648f, 0.0699925423f, 0.f),
    C_Vertex3f(0.191711903f, -0.00454163551f, 0.f),
    C_Vertex3f(0.201793432f, 0.0634799004f, 0.f),
    C_Vertex3f(0.212373257f, -0.00867533684f, 0.f),
    C_Vertex3f(0.213689208f, 0.0585954189f, 0.f),
    C_Vertex3f(0.234276772f, -0.0107421875f, 0.f),
    C_Vertex3f(0.226122975f, 0.055339098f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179615855f, 0.0781333447f, 0.f),
    C_Vertex3f(0.169334054f, 0.0879023075f, 0.f),
    C_Vertex3f(0.172292709f, 0.00165891647f, 0.f),
    C_Vertex3f(0.191711903f, -0.00454163551f, 0.f),
    C_Vertex3f(0.190435648f, 0.0699925423f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.159590244f, 0.0992994308f, 0.f),
    C_Vertex3f(0.150384426f, 0.112324715f, 0.f),
    C_Vertex3f(0.154115677f, 0.00992631912f, 0.f),
    C_Vertex3f(0.172292709f, 0.00165891647f, 0.f),
    C_Vertex3f(0.169334054f, 0.0879023075f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.142329335f, 0.128659487f, 0.f),
    C_Vertex3f(0.135424972f, 0.148303747f, 0.f),
    C_Vertex3f(0.137180805f, 0.0202605724f, 0.f),
    C_Vertex3f(0.154115677f, 0.00992631912f, 0.f),
    C_Vertex3f(0.150384426f, 0.112324715f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.137180805f, 0.0202605724f, 0.f),
    C_Vertex3f(0.135424972f, 0.148303747f, 0.f),
    C_Vertex3f(0.129671335f, 0.171257496f, 0.f),
    C_Vertex3f(0.125068426f, 0.197520733f, 0.f),
    C_Vertex3f(0.121616244f, 0.227093458f, 0.f),
    C_Vertex3f(0.121488094f, 0.0326616764f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.11931479f, 0.259975672f, 0.f),
    C_Vertex3f(0.118164063f, 0.296167374f, 0.f),
    C_Vertex3f(0.121488094f, 0.0326616764f, 0.f),
    C_Vertex3f(0.121616244f, 0.227093458f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0772061348f, 0.539194703f, 0.f),
    C_Vertex3f(0.0786657333f, 0.0885920525f, 0.f),
    C_Vertex3f(0.0851759911f, 0.553539991f, 0.f),
    C_Vertex3f(0.0925178528f, 0.0654182434f, 0.f),
    C_Vertex3f(0.093855381f, 0.566888094f, 0.f),
    C_Vertex3f(0.107037544f, 0.047129631f, 0.f),
    C_Vertex3f(0.103244305f, 0.579239011f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0699458122f, 0.523852229f, 0.f),
    C_Vertex3f(0.0633950233f, 0.507512569f, 0.f),
    C_Vertex3f(0.0667924881f, 0.114897251f, 0.f),
    C_Vertex3f(0.0786657333f, 0.0885920525f, 0.f),
    C_Vertex3f(0.0772061348f, 0.539194703f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0480617285f, 0.451357484f, 0.f),
    C_Vertex3f(0.0489826202f, 0.176901817f, 0.f),
    C_Vertex3f(0.0524426699f, 0.471457362f, 0.f),
    C_Vertex3f(0.0568981171f, 0.144333839f, 0.f),
    C_Vertex3f(0.0575537682f, 0.490175724f, 0.f),
    C_Vertex3f(0.0667924881f, 0.114897251f, 0.f),
    C_Vertex3f(0.0633950233f, 0.507512569f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.044410944f, 0.429876089f, 0.f),
    C_Vertex3f(0.0414903164f, 0.407013178f, 0.f),
    C_Vertex3f(0.0430459976f, 0.212601185f, 0.f),
    C_Vertex3f(0.0489826202f, 0.176901817f, 0.f),
    C_Vertex3f(0.0480617285f, 0.451357484f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0392998457f, 0.38276875f, 0.f),
    C_Vertex3f(0.0378395319f, 0.357142806f, 0.f),
    C_Vertex3f(0.0390882492f, 0.251431942f, 0.f),
    C_Vertex3f(0.0430459976f, 0.212601185f, 0.f),
    C_Vertex3f(0.0414903164f, 0.407013178f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.037109375f, 0.330135345f, 0.f),
    C_Vertex3f(0.037109375f, 0.293394089f, 0.f),
    C_Vertex3f(0.0390882492f, 0.251431942f, 0.f),
    C_Vertex3f(0.0378395319f, 0.357142806f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.450086117f, 0.410840988f, -1.f),
    C_Vertex3f(0.450250387f, 0.225471735f, -1.f),
    C_Vertex3f(0.451887608f, 0.391923904f, -1.f),
    C_Vertex3f(0.452420115f, 0.249579072f, -1.f),
    C_Vertex3f(0.453238726f, 0.371785164f, -1.f),
    C_Vertex3f(0.453866601f, 0.275040507f, -1.f),
    C_Vertex3f(0.454139471f, 0.350424767f, -1.f),
    C_Vertex3f(0.454589844f, 0.301856041f, -1.f),
    C_Vertex3f(0.454589844f, 0.327842712f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.428553104f, 0.125246525f, -1.f),
    C_Vertex3f(0.434339046f, 0.14258337f, -1.f),
    C_Vertex3f(0.436574936f, 0.481307983f, -1.f),
    C_Vertex3f(0.439401746f, 0.161274314f, -1.f),
    C_Vertex3f(0.441979408f, 0.460262299f, -1.f),
    C_Vertex3f(0.443741202f, 0.181319356f, -1.f),
    C_Vertex3f(0.445132017f, 0.445010185f, -1.f),
    C_Vertex3f(0.447357416f, 0.202718496f, -1.f),
    C_Vertex3f(0.447834253f, 0.428536415f, -1.f),
    C_Vertex3f(0.450250387f, 0.225471735f, -1.f),
    C_Vertex3f(0.450086117f, 0.410840988f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.406895876f, 0.0791783333f, -1.f),
    C_Vertex3f(0.414831519f, 0.0935440063f, -1.f),
    C_Vertex3f(0.418116093f, 0.531042576f, -1.f),
    C_Vertex3f(0.422050595f, 0.10890007f, -1.f),
    C_Vertex3f(0.42802f, 0.507239819f, -1.f),
    C_Vertex3f(0.428553104f, 0.125246525f, -1.f),
    C_Vertex3f(0.436574936f, 0.481307983f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.373535156f, 0.335692406f, -1.f),
    C_Vertex3f(0.378789544f, 0.0420236588f, -1.f),
    C_Vertex3f(0.379936218f, 0.589512348f, -1.f),
    C_Vertex3f(0.388874888f, 0.0534181595f, -1.f),
    C_Vertex3f(0.39426136f, 0.572260857f, -1.f),
    C_Vertex3f(0.398243666f, 0.065803051f, -1.f),
    C_Vertex3f(0.406863213f, 0.552716255f, -1.f),
    C_Vertex3f(0.406895876f, 0.0791783333f, -1.f),
    C_Vertex3f(0.418116093f, 0.531042576f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.379936218f, 0.589512348f, -1.f),
    C_Vertex3f(0.363887787f, 0.60447073f, -1.f),
    C_Vertex3f(0.366642237f, 0.434428453f, -1.f),
    C_Vertex3f(0.370088696f, 0.404834867f, -1.f),
    C_Vertex3f(0.372386336f, 0.371922851f, -1.f),
    C_Vertex3f(0.373535156f, 0.335692406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.363887787f, 0.60447073f, -1.f),
    C_Vertex3f(0.346116066f, 0.617136002f, -1.f),
    C_Vertex3f(0.349409938f, 0.50329864f, -1.f),
    C_Vertex3f(0.356302857f, 0.48366034f, -1.f),
    C_Vertex3f(0.362046957f, 0.460703611f, -1.f),
    C_Vertex3f(0.366642237f, 0.434428453f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.31208384f, 0.553747296f, -1.f),
    C_Vertex3f(0.322416186f, 0.543996215f, -1.f),
    C_Vertex3f(0.326621056f, 0.627508163f, -1.f),
    C_Vertex3f(0.332177639f, 0.532619953f, -1.f),
    C_Vertex3f(0.346116066f, 0.617136002f, -1.f),
    C_Vertex3f(0.341368198f, 0.519618511f, -1.f),
    C_Vertex3f(0.349409938f, 0.50329864f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.305558681f, 0.635287285f, -1.f),
    C_Vertex3f(0.282928944f, 0.640473366f, -1.f),
    C_Vertex3f(0.289706469f, 0.568373919f, -1.f),
    C_Vertex3f(0.301180601f, 0.561873198f, -1.f),
    C_Vertex3f(0.31208384f, 0.553747296f, -1.f),
    C_Vertex3f(0.326621056f, 0.627508163f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.238224983f, 0.578125f, -1.f),
    C_Vertex3f(0.251858711f, 0.578125f, -1.f),
    C_Vertex3f(0.258731842f, 0.643066406f, -1.f),
    C_Vertex3f(0.265045524f, 0.57649982f, -1.f),
    C_Vertex3f(0.282928944f, 0.640473366f, -1.f),
    C_Vertex3f(0.277661443f, 0.573249459f, -1.f),
    C_Vertex3f(0.289706469f, 0.568373919f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.190957069f, 0.563786507f, -1.f),
    C_Vertex3f(0.201758385f, 0.569521904f, -1.f),
    C_Vertex3f(0.204412103f, 0.639556646f, -1.f),
    C_Vertex3f(0.213236809f, 0.573823452f, -1.f),
    C_Vertex3f(0.220392108f, 0.641896486f, -1.f),
    C_Vertex3f(0.225392342f, 0.576691151f, -1.f),
    C_Vertex3f(0.23711586f, 0.643066406f, -1.f),
    C_Vertex3f(0.238224983f, 0.578125f, -1.f),
    C_Vertex3f(0.258731842f, 0.643066406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.162615776f, 0.537977219f, -1.f),
    C_Vertex3f(0.171385765f, 0.548014164f, -1.f),
    C_Vertex3f(0.174683332f, 0.631367207f, -1.f),
    C_Vertex3f(0.180832863f, 0.55661726f, -1.f),
    C_Vertex3f(0.189175844f, 0.636046886f, -1.f),
    C_Vertex3f(0.190957069f, 0.563786507f, -1.f),
    C_Vertex3f(0.204412103f, 0.639556646f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.160934567f, 0.625517607f, -1.f),
    C_Vertex3f(0.147929549f, 0.618498087f, -1.f),
    C_Vertex3f(0.153518677f, 0.524921417f, -1.f),
    C_Vertex3f(0.158569336f, 0.532241821f, -1.f),
    C_Vertex3f(0.162615776f, 0.537977219f, -1.f),
    C_Vertex3f(0.174683332f, 0.631367207f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.125740051f, 0.435359001f, -1.f),
    C_Vertex3f(0.13079071f, 0.462399483f, -1.f),
    C_Vertex3f(0.135668278f, 0.610308647f, -1.f),
    C_Vertex3f(0.137104034f, 0.486340046f, -1.f),
    C_Vertex3f(0.147929549f, 0.618498087f, -1.f),
    C_Vertex3f(0.144680023f, 0.507180691f, -1.f),
    C_Vertex3f(0.153518677f, 0.524921417f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.124150753f, 0.600949287f, -1.f),
    C_Vertex3f(0.113342762f, 0.590592742f, -1.f),
    C_Vertex3f(0.118164063f, 0.335638046f, -1.f),
    C_Vertex3f(0.119426727f, 0.371978283f, -1.f),
    C_Vertex3f(0.121952057f, 0.405218601f, -1.f),
    C_Vertex3f(0.125740051f, 0.435359001f, -1.f),
    C_Vertex3f(0.135668278f, 0.610308647f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.113342762f, 0.590592742f, -1.f),
    C_Vertex3f(0.103244305f, 0.579239011f, -1.f),
    C_Vertex3f(0.107037544f, 0.047129631f, -1.f),
    C_Vertex3f(0.121488094f, 0.0326616764f, -1.f),
    C_Vertex3f(0.118164063f, 0.296167374f, -1.f),
    C_Vertex3f(0.118164063f, 0.335638046f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378789544f, 0.0420236588f, -1.f),
    C_Vertex3f(0.373535156f, 0.335692406f, -1.f),
    C_Vertex3f(0.373535156f, 0.296221733f, -1.f),
    C_Vertex3f(0.372384429f, 0.260118365f, -1.f),
    C_Vertex3f(0.370082974f, 0.227304101f, -1.f),
    C_Vertex3f(0.367987633f, 0.0316195488f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.370082974f, 0.227304101f, -1.f),
    C_Vertex3f(0.366630793f, 0.19777894f, -1.f),
    C_Vertex3f(0.367987633f, 0.0316195488f, -1.f),
    C_Vertex3f(0.362027884f, 0.171542883f, -1.f),
    C_Vertex3f(0.356469154f, 0.0222058296f, -1.f),
    C_Vertex3f(0.356274247f, 0.148595929f, -1.f),
    C_Vertex3f(0.349369884f, 0.128938079f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.344193459f, 0.0139688253f, -1.f),
    C_Vertex3f(0.356469154f, 0.0222058296f, -1.f),
    C_Vertex3f(0.349369884f, 0.128938079f, -1.f),
    C_Vertex3f(0.341314793f, 0.112569332f, -1.f),
    C_Vertex3f(0.332108974f, 0.0994896889f, -1.f),
    C_Vertex3f(0.331160545f, 0.00690853596f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.301263571f, 0.0700604916f, -1.f),
    C_Vertex3f(0.302823067f, -0.00368189812f, -1.f),
    C_Vertex3f(0.312083364f, 0.0782352686f, -1.f),
    C_Vertex3f(0.317370415f, 0.00102496147f, -1.f),
    C_Vertex3f(0.322365165f, 0.088045001f, -1.f),
    C_Vertex3f(0.331160545f, 0.00690853596f, -1.f),
    C_Vertex3f(0.332108974f, 0.0994896889f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.252604485f, 0.0537109375f, -1.f),
    C_Vertex3f(0.254637718f, -0.0107421875f, -1.f),
    C_Vertex3f(0.265576243f, 0.0553458929f, -1.f),
    C_Vertex3f(0.271456718f, -0.0095654726f, -1.f),
    C_Vertex3f(0.278010011f, 0.0586158037f, -1.f),
    C_Vertex3f(0.287518501f, -0.00721204281f, -1.f),
    C_Vertex3f(0.289905787f, 0.0635206699f, -1.f),
    C_Vertex3f(0.302823067f, -0.00368189812f, -1.f),
    C_Vertex3f(0.301263571f, 0.0700604916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.239094734f, 0.0537109375f, -1.f),
    C_Vertex3f(0.226122975f, 0.055339098f, -1.f),
    C_Vertex3f(0.234276772f, -0.0107421875f, -1.f),
    C_Vertex3f(0.254637718f, -0.0107421875f, -1.f),
    C_Vertex3f(0.252604485f, 0.0537109375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.190435648f, 0.0699925423f, -1.f),
    C_Vertex3f(0.191711903f, -0.00454163551f, -1.f),
    C_Vertex3f(0.201793432f, 0.0634799004f, -1.f),
    C_Vertex3f(0.212373257f, -0.00867533684f, -1.f),
    C_Vertex3f(0.213689208f, 0.0585954189f, -1.f),
    C_Vertex3f(0.234276772f, -0.0107421875f, -1.f),
    C_Vertex3f(0.226122975f, 0.055339098f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179615855f, 0.0781333447f, -1.f),
    C_Vertex3f(0.169334054f, 0.0879023075f, -1.f),
    C_Vertex3f(0.172292709f, 0.00165891647f, -1.f),
    C_Vertex3f(0.191711903f, -0.00454163551f, -1.f),
    C_Vertex3f(0.190435648f, 0.0699925423f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.159590244f, 0.0992994308f, -1.f),
    C_Vertex3f(0.150384426f, 0.112324715f, -1.f),
    C_Vertex3f(0.154115677f, 0.00992631912f, -1.f),
    C_Vertex3f(0.172292709f, 0.00165891647f, -1.f),
    C_Vertex3f(0.169334054f, 0.0879023075f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.142329335f, 0.128659487f, -1.f),
    C_Vertex3f(0.135424972f, 0.148303747f, -1.f),
    C_Vertex3f(0.137180805f, 0.0202605724f, -1.f),
    C_Vertex3f(0.154115677f, 0.00992631912f, -1.f),
    C_Vertex3f(0.150384426f, 0.112324715f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.137180805f, 0.0202605724f, -1.f),
    C_Vertex3f(0.135424972f, 0.148303747f, -1.f),
    C_Vertex3f(0.129671335f, 0.171257496f, -1.f),
    C_Vertex3f(0.125068426f, 0.197520733f, -1.f),
    C_Vertex3f(0.121616244f, 0.227093458f, -1.f),
    C_Vertex3f(0.121488094f, 0.0326616764f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.11931479f, 0.259975672f, -1.f),
    C_Vertex3f(0.118164063f, 0.296167374f, -1.f),
    C_Vertex3f(0.121488094f, 0.0326616764f, -1.f),
    C_Vertex3f(0.121616244f, 0.227093458f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0772061348f, 0.539194703f, -1.f),
    C_Vertex3f(0.0786657333f, 0.0885920525f, -1.f),
    C_Vertex3f(0.0851759911f, 0.553539991f, -1.f),
    C_Vertex3f(0.0925178528f, 0.0654182434f, -1.f),
    C_Vertex3f(0.093855381f, 0.566888094f, -1.f),
    C_Vertex3f(0.107037544f, 0.047129631f, -1.f),
    C_Vertex3f(0.103244305f, 0.579239011f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0699458122f, 0.523852229f, -1.f),
    C_Vertex3f(0.0633950233f, 0.507512569f, -1.f),
    C_Vertex3f(0.0667924881f, 0.114897251f, -1.f),
    C_Vertex3f(0.0786657333f, 0.0885920525f, -1.f),
    C_Vertex3f(0.0772061348f, 0.539194703f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0480617285f, 0.451357484f, -1.f),
    C_Vertex3f(0.0489826202f, 0.176901817f, -1.f),
    C_Vertex3f(0.0524426699f, 0.471457362f, -1.f),
    C_Vertex3f(0.0568981171f, 0.144333839f, -1.f),
    C_Vertex3f(0.0575537682f, 0.490175724f, -1.f),
    C_Vertex3f(0.0667924881f, 0.114897251f, -1.f),
    C_Vertex3f(0.0633950233f, 0.507512569f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.044410944f, 0.429876089f, -1.f),
    C_Vertex3f(0.0414903164f, 0.407013178f, -1.f),
    C_Vertex3f(0.0430459976f, 0.212601185f, -1.f),
    C_Vertex3f(0.0489826202f, 0.176901817f, -1.f),
    C_Vertex3f(0.0480617285f, 0.451357484f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0392998457f, 0.38276875f, -1.f),
    C_Vertex3f(0.0378395319f, 0.357142806f, -1.f),
    C_Vertex3f(0.0390882492f, 0.251431942f, -1.f),
    C_Vertex3f(0.0430459976f, 0.212601185f, -1.f),
    C_Vertex3f(0.0414903164f, 0.407013178f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.037109375f, 0.330135345f, -1.f),
    C_Vertex3f(0.037109375f, 0.293394089f, -1.f),
    C_Vertex3f(0.0390882492f, 0.251431942f, -1.f),
    C_Vertex3f(0.0378395319f, 0.357142806f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_49[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.333007813f, 0.f, 0.f),
    C_Vertex3f(0.333007813f, 0.f, -1.f),
    C_Vertex3f(0.254394531f, 0.f, -1.f),
    C_Vertex3f(0.254394531f, 0.f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.254394531f, 0.f, 0.f),
    C_Vertex3f(0.254394531f, 0.f, -1.f),
    C_Vertex3f(0.254394531f, 0.501083374f, -1.f),
    C_Vertex3f(0.254394531f, 0.501083374f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.6902861f, -0.723536551f, 0.f),
    C_Vertex3f(0.254394531f, 0.501083374f, 0.f),
    C_Vertex3f(0.254394531f, 0.501083374f, -1.f),
    C_Normal3f(0.653910816f, -0.75657171f, 0.f),
    C_Vertex3f(0.247297287f, 0.494312286f, 0.f),
    C_Vertex3f(0.247297287f, 0.494312286f, -1.f),
    C_Normal3f(0.611769915f, -0.791035771f, 0.f),
    C_Vertex3f(0.230892181f, 0.480770111f, 0.f),
    C_Vertex3f(0.230892181f, 0.480770111f, -1.f),
    C_Normal3f(0.566073596f, -0.824354649f, 0.f),
    C_Vertex3f(0.212276459f, 0.467227936f, 0.f),
    C_Vertex3f(0.212276459f, 0.467227936f, -1.f),
    C_Normal3f(0.525384963f, -0.850864589f, 0.f),
    C_Vertex3f(0.191450119f, 0.45368576f, 0.f),
    C_Vertex3f(0.191450119f, 0.45368576f, -1.f),
    C_Normal3f(0.492245495f, -0.870456457f, 0.f),
    C_Vertex3f(0.168413162f, 0.440143585f, 0.f),
    C_Vertex3f(0.168413162f, 0.440143585f, -1.f),
    C_Normal3f(0.459338963f, -0.88826102f, 0.f),
    C_Vertex3f(0.146549225f, 0.428294182f, 0.f),
    C_Vertex3f(0.146549225f, 0.428294182f, -1.f),
    C_Normal3f(0.420222044f, -0.90742135f, 0.f),
    C_Vertex3f(0.125858307f, 0.41813755f, 0.f),
    C_Vertex3f(0.125858307f, 0.41813755f, -1.f),
    C_Normal3f(0.381733924f, -0.924272299f, 0.f),
    C_Vertex3f(0.106340408f, 0.409673691f, 0.f),
    C_Vertex3f(0.106340408f, 0.409673691f, -1.f),
    C_Normal3f(0.346265763f, -0.938136399f, 0.f),
    C_Vertex3f(0.0971679688f, 0.406288147f, 0.f),
    C_Vertex3f(0.0971679688f, 0.406288147f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0971679688f, 0.406288147f, 0.f),
    C_Vertex3f(0.0971679688f, 0.406288147f, -1.f),
    C_Vertex3f(0.0971679688f, 0.482299805f, -1.f),
    C_Vertex3f(0.0971679688f, 0.482299805f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.425645769f, 0.904889882f, 0.f),
    C_Vertex3f(0.0971679688f, 0.482299805f, 0.f),
    C_Vertex3f(0.0971679688f, 0.482299805f, -1.f),
    C_Normal3f(-0.465204477f, 0.885203242f, 0.f),
    C_Vertex3f(0.11365509f, 0.490055084f, 0.f),
    C_Vertex3f(0.11365509f, 0.490055084f, -1.f),
    C_Normal3f(-0.515538633f, 0.8568663f, 0.f),
    C_Vertex3f(0.144554615f, 0.507203102f, 0.f),
    C_Vertex3f(0.144554615f, 0.507203102f, -1.f),
    C_Normal3f(-0.576477826f, 0.817112803f, 0.f),
    C_Vertex3f(0.173379421f, 0.525988579f, 0.f),
    C_Vertex3f(0.173379421f, 0.525988579f, -1.f),
    C_Normal3f(-0.636894226f, 0.770951271f, 0.f),
    C_Vertex3f(0.200129509f, 0.546411514f, 0.f),
    C_Vertex3f(0.200129509f, 0.546411514f, -1.f),
    C_Normal3f(-0.691507041f, 0.722369671f, 0.f),
    C_Vertex3f(0.224804878f, 0.568471909f, 0.f),
    C_Vertex3f(0.224804878f, 0.568471909f, -1.f),
    C_Normal3f(-0.745715618f, 0.666264415f, 0.f),
    C_Vertex3f(0.245877743f, 0.590205193f, 0.f),
    C_Vertex3f(0.245877743f, 0.590205193f, -1.f),
    C_Normal3f(-0.804754436f, 0.593607783f, 0.f),
    C_Vertex3f(0.263348103f, 0.611611366f, 0.f),
    C_Vertex3f(0.263348103f, 0.611611366f, -1.f),
    C_Normal3f(-0.855958343f, 0.517044723f, 0.f),
    C_Vertex3f(0.277215958f, 0.63269043f, 0.f),
    C_Vertex3f(0.277215958f, 0.63269043f, -1.f),
    C_Normal3f(-0.896330357f, 0.443386942f, 0.f),
    C_Vertex3f(0.282348633f, 0.643066406f, 0.f),
    C_Vertex3f(0.282348633f, 0.643066406f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.282348633f, 0.643066406f, 0.f),
    C_Vertex3f(0.282348633f, 0.643066406f, -1.f),
    C_Vertex3f(0.333007813f, 0.643066406f, -1.f),
    C_Vertex3f(0.333007813f, 0.643066406f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.333007813f, 0.643066406f, 0.f),
    C_Vertex3f(0.333007813f, 0.643066406f, -1.f),
    C_Vertex3f(0.333007813f, 0.f, -1.f),
    C_Vertex3f(0.333007813f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.146549225f, 0.428294182f, 0.f),
    C_Vertex3f(0.168413162f, 0.440143585f, 0.f),
    C_Vertex3f(0.173379421f, 0.525988579f, 0.f),
    C_Vertex3f(0.191450119f, 0.45368576f, 0.f),
    C_Vertex3f(0.200129509f, 0.546411514f, 0.f),
    C_Vertex3f(0.212276459f, 0.467227936f, 0.f),
    C_Vertex3f(0.224804878f, 0.568471909f, 0.f),
    C_Vertex3f(0.230892181f, 0.480770111f, 0.f),
    C_Vertex3f(0.245877743f, 0.590205193f, 0.f),
    C_Vertex3f(0.247297287f, 0.494312286f, 0.f),
    C_Vertex3f(0.263348103f, 0.611611366f, 0.f),
    C_Vertex3f(0.254394531f, 0.501083374f, 0.f),
    C_Vertex3f(0.333007813f, 0.f, 0.f),
    C_Vertex3f(0.254394531f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0971679688f, 0.482299805f, 0.f),
    C_Vertex3f(0.106340408f, 0.409673691f, 0.f),
    C_Vertex3f(0.11365509f, 0.490055084f, 0.f),
    C_Vertex3f(0.125858307f, 0.41813755f, 0.f),
    C_Vertex3f(0.144554615f, 0.507203102f, 0.f),
    C_Vertex3f(0.146549225f, 0.428294182f, 0.f),
    C_Vertex3f(0.173379421f, 0.525988579f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.333007813f, 0.f, 0.f),
    C_Vertex3f(0.333007813f, 0.643066406f, 0.f),
    C_Vertex3f(0.282348633f, 0.643066406f, 0.f),
    C_Vertex3f(0.277215958f, 0.63269043f, 0.f),
    C_Vertex3f(0.263348103f, 0.611611366f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.106340408f, 0.409673691f, 0.f),
    C_Vertex3f(0.0971679688f, 0.482299805f, 0.f),
    C_Vertex3f(0.0971679688f, 0.406288147f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.146549225f, 0.428294182f, -1.f),
    C_Vertex3f(0.168413162f, 0.440143585f, -1.f),
    C_Vertex3f(0.173379421f, 0.525988579f, -1.f),
    C_Vertex3f(0.191450119f, 0.45368576f, -1.f),
    C_Vertex3f(0.200129509f, 0.546411514f, -1.f),
    C_Vertex3f(0.212276459f, 0.467227936f, -1.f),
    C_Vertex3f(0.224804878f, 0.568471909f, -1.f),
    C_Vertex3f(0.230892181f, 0.480770111f, -1.f),
    C_Vertex3f(0.245877743f, 0.590205193f, -1.f),
    C_Vertex3f(0.247297287f, 0.494312286f, -1.f),
    C_Vertex3f(0.263348103f, 0.611611366f, -1.f),
    C_Vertex3f(0.254394531f, 0.501083374f, -1.f),
    C_Vertex3f(0.333007813f, 0.f, -1.f),
    C_Vertex3f(0.254394531f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0971679688f, 0.482299805f, -1.f),
    C_Vertex3f(0.106340408f, 0.409673691f, -1.f),
    C_Vertex3f(0.11365509f, 0.490055084f, -1.f),
    C_Vertex3f(0.125858307f, 0.41813755f, -1.f),
    C_Vertex3f(0.144554615f, 0.507203102f, -1.f),
    C_Vertex3f(0.146549225f, 0.428294182f, -1.f),
    C_Vertex3f(0.173379421f, 0.525988579f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.333007813f, 0.f, -1.f),
    C_Vertex3f(0.333007813f, 0.643066406f, -1.f),
    C_Vertex3f(0.282348633f, 0.643066406f, -1.f),
    C_Vertex3f(0.277215958f, 0.63269043f, -1.f),
    C_Vertex3f(0.263348103f, 0.611611366f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.106340408f, 0.409673691f, -1.f),
    C_Vertex3f(0.0971679688f, 0.482299805f, -1.f),
    C_Vertex3f(0.0971679688f, 0.406288147f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_50[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.450195313f, 0.0756835938f, 0.f),
    C_Vertex3f(0.450195313f, 0.0756835938f, -1.f),
    C_Vertex3f(0.450195313f, 0.f, -1.f),
    C_Vertex3f(0.450195313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.450195313f, 0.f, 0.f),
    C_Vertex3f(0.450195313f, 0.f, -1.f),
    C_Vertex3f(0.0268554688f, 0.f, -1.f),
    C_Vertex3f(0.0268554688f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999532521f, -0.0305733047f, 0.f),
    C_Vertex3f(0.0268554688f, 0.f, 0.f),
    C_Vertex3f(0.0268554688f, 0.f, -1.f),
    C_Normal3f(-0.999426425f, 0.0338650495f, 0.f),
    C_Vertex3f(0.026638031f, 0.00710868835f, 0.f),
    C_Vertex3f(0.026638031f, 0.00710868835f, -1.f),
    C_Normal3f(-0.99323982f, 0.116080932f, 0.f),
    C_Vertex3f(0.0275688171f, 0.0210523605f, 0.f),
    C_Vertex3f(0.0275688171f, 0.0210523605f, -1.f),
    C_Normal3f(-0.976614296f, 0.214998975f, 0.f),
    C_Vertex3f(0.0298652649f, 0.0347223282f, 0.f),
    C_Vertex3f(0.0298652649f, 0.0347223282f, -1.f),
    C_Normal3f(-0.949056089f, 0.315107435f, 0.f),
    C_Vertex3f(0.0335273743f, 0.0481185913f, 0.f),
    C_Vertex3f(0.0335273743f, 0.0481185913f, -1.f),
    C_Normal3f(-0.915489316f, 0.40234229f, 0.f),
    C_Vertex3f(0.0400867462f, 0.0655078888f, 0.f),
    C_Vertex3f(0.0400867462f, 0.0655078888f, -1.f),
    C_Normal3f(-0.8766855f, 0.481063932f, 0.f),
    C_Vertex3f(0.050611496f, 0.086991787f, 0.f),
    C_Vertex3f(0.050611496f, 0.086991787f, -1.f),
    C_Normal3f(-0.831598759f, 0.555376887f, 0.f),
    C_Vertex3f(0.0635700226f, 0.108303547f, 0.f),
    C_Vertex3f(0.0635700226f, 0.108303547f, -1.f),
    C_Normal3f(-0.788429916f, 0.615124643f, 0.f),
    C_Vertex3f(0.078962326f, 0.129443169f, 0.f),
    C_Vertex3f(0.078962326f, 0.129443169f, -1.f),
    C_Normal3f(-0.752858877f, 0.658182085f, 0.f),
    C_Vertex3f(0.0923318863f, 0.145168781f, 0.f),
    C_Vertex3f(0.0923318863f, 0.145168781f, -1.f),
    C_Normal3f(-0.729794562f, 0.683666468f, 0.f),
    C_Vertex3f(0.102230549f, 0.156058431f, 0.f),
    C_Vertex3f(0.102230549f, 0.156058431f, -1.f),
    C_Normal3f(-0.710849822f, 0.703343868f, 0.f),
    C_Vertex3f(0.113114834f, 0.167353988f, 0.f),
    C_Vertex3f(0.113114834f, 0.167353988f, -1.f),
    C_Normal3f(-0.693628967f, 0.720332503f, 0.f),
    C_Vertex3f(0.124984741f, 0.179055452f, 0.f),
    C_Vertex3f(0.124984741f, 0.179055452f, -1.f),
    C_Normal3f(-0.677946448f, 0.735111356f, 0.f),
    C_Vertex3f(0.137840271f, 0.191162825f, 0.f),
    C_Vertex3f(0.137840271f, 0.191162825f, -1.f),
    C_Normal3f(-0.663632751f, 0.748058558f, 0.f),
    C_Vertex3f(0.151681423f, 0.203676105f, 0.f),
    C_Vertex3f(0.151681423f, 0.203676105f, -1.f),
    C_Normal3f(-0.65053606f, 0.75947535f, 0.f),
    C_Vertex3f(0.166508198f, 0.216595292f, 0.f),
    C_Vertex3f(0.166508198f, 0.216595292f, -1.f),
    C_Normal3f(-0.638614714f, 0.76952666f, 0.f),
    C_Vertex3f(0.182320595f, 0.229920387f, 0.f),
    C_Vertex3f(0.182320595f, 0.229920387f, -1.f),
    C_Normal3f(-0.641009688f, 0.767532766f, 0.f),
    C_Vertex3f(0.203754425f, 0.247505188f, 0.f),
    C_Vertex3f(0.203754425f, 0.247505188f, -1.f),
    C_Normal3f(-0.653400242f, 0.757012606f, 0.f),
    C_Vertex3f(0.227708578f, 0.267826438f, 0.f),
    C_Vertex3f(0.227708578f, 0.267826438f, -1.f),
    C_Normal3f(-0.66782409f, 0.744319141f, 0.f),
    C_Vertex3f(0.249547243f, 0.287030339f, 0.f),
    C_Vertex3f(0.249547243f, 0.287030339f, -1.f),
    C_Normal3f(-0.684544384f, 0.728971183f, 0.f),
    C_Vertex3f(0.26927042f, 0.305116892f, 0.f),
    C_Vertex3f(0.26927042f, 0.305116892f, -1.f),
    C_Normal3f(-0.704109609f, 0.710091352f, 0.f),
    C_Vertex3f(0.286878109f, 0.322086096f, 0.f),
    C_Vertex3f(0.286878109f, 0.322086096f, -1.f),
    C_Normal3f(-0.727229178f, 0.686394691f, 0.f),
    C_Vertex3f(0.30237031f, 0.337937951f, 0.f),
    C_Vertex3f(0.30237031f, 0.337937951f, -1.f),
    C_Normal3f(-0.754812241f, 0.65594089f, 0.f),
    C_Vertex3f(0.315747023f, 0.352672458f, 0.f),
    C_Vertex3f(0.315747023f, 0.352672458f, -1.f),
    C_Normal3f(-0.787979245f, 0.615701735f, 0.f),
    C_Vertex3f(0.327008247f, 0.366289616f, 0.f),
    C_Vertex3f(0.327008247f, 0.366289616f, -1.f),
    C_Normal3f(-0.822872043f, 0.568226695f, 0.f),
    C_Vertex3f(0.336153984f, 0.378789425f, 0.f),
    C_Vertex3f(0.336153984f, 0.378789425f, -1.f),
    C_Normal3f(-0.854991615f, 0.518641889f, 0.f),
    C_Vertex3f(0.344156504f, 0.391122699f, 0.f),
    C_Vertex3f(0.344156504f, 0.391122699f, -1.f),
    C_Normal3f(-0.887087941f, 0.461600602f, 0.f),
    C_Vertex3f(0.351015806f, 0.403289437f, 0.f),
    C_Vertex3f(0.351015806f, 0.403289437f, -1.f),
    C_Normal3f(-0.918103635f, 0.3963404f, 0.f),
    C_Vertex3f(0.356731892f, 0.41528964f, 0.f),
    C_Vertex3f(0.356731892f, 0.41528964f, -1.f),
    C_Normal3f(-0.946622252f, 0.322344899f, 0.f),
    C_Vertex3f(0.36130476f, 0.427123308f, 0.f),
    C_Vertex3f(0.36130476f, 0.427123308f, -1.f),
    C_Normal3f(-0.970885813f, 0.239542961f, 0.f),
    C_Vertex3f(0.364734411f, 0.438790441f, 0.f),
    C_Vertex3f(0.364734411f, 0.438790441f, -1.f),
    C_Normal3f(-0.988908052f, 0.148529023f, 0.f),
    C_Vertex3f(0.367020845f, 0.450291038f, 0.f),
    C_Vertex3f(0.367020845f, 0.450291038f, -1.f),
    C_Normal3f(-0.998740673f, 0.0501700975f, 0.f),
    C_Vertex3f(0.368164063f, 0.461625099f, 0.f),
    C_Vertex3f(0.368164063f, 0.461625099f, -1.f),
    C_Normal3f(-0.9989416f, -0.0459977314f, 0.f),
    C_Vertex3f(0.368164063f, 0.473049164f, 0.f),
    C_Vertex3f(0.368164063f, 0.473049164f, -1.f),
    C_Normal3f(-0.990030706f, -0.140851796f, 0.f),
    C_Vertex3f(0.36712122f, 0.484272718f, 0.f),
    C_Vertex3f(0.36712122f, 0.484272718f, -1.f),
    C_Normal3f(-0.970733225f, -0.240160361f, 0.f),
    C_Vertex3f(0.365035534f, 0.495039225f, 0.f),
    C_Vertex3f(0.365035534f, 0.495039225f, -1.f),
    C_Normal3f(-0.940267205f, -0.340437531f, 0.f),
    C_Vertex3f(0.361907005f, 0.505348682f, 0.f),
    C_Vertex3f(0.361907005f, 0.505348682f, -1.f),
    C_Normal3f(-0.898834109f, -0.438288927f, 0.f),
    C_Vertex3f(0.357735634f, 0.515201092f, 0.f),
    C_Vertex3f(0.357735634f, 0.515201092f, -1.f),
    C_Normal3f(-0.847733676f, -0.530421972f, 0.f),
    C_Vertex3f(0.35252142f, 0.524596453f, 0.f),
    C_Vertex3f(0.35252142f, 0.524596453f, -1.f),
    C_Normal3f(-0.789167821f, -0.614177644f, 0.f),
    C_Vertex3f(0.346264362f, 0.533534765f, 0.f),
    C_Vertex3f(0.346264362f, 0.533534765f, -1.f),
    C_Normal3f(-0.7258237f, -0.687880754f, 0.f),
    C_Vertex3f(0.338964462f, 0.542016029f, 0.f),
    C_Vertex3f(0.338964462f, 0.542016029f, -1.f),
    C_Normal3f(-0.655801713f, -0.754933178f, 0.f),
    C_Vertex3f(0.330621719f, 0.550040245f, 0.f),
    C_Vertex3f(0.330621719f, 0.550040245f, -1.f),
    C_Normal3f(-0.574280977f, -0.818658292f, 0.f),
    C_Vertex3f(0.321644783f, 0.557061434f, 0.f),
    C_Vertex3f(0.321644783f, 0.557061434f, -1.f),
    C_Normal3f(-0.485707998f, -0.87412113f, 0.f),
    C_Vertex3f(0.312033653f, 0.563079596f, 0.f),
    C_Vertex3f(0.312033653f, 0.563079596f, -1.f),
    C_Normal3f(-0.392953336f, -0.919558346f, 0.f),
    C_Vertex3f(0.30178833f, 0.56809473f, 0.f),
    C_Vertex3f(0.30178833f, 0.56809473f, -1.f),
    C_Normal3f(-0.299179584f, -0.95419687f, 0.f),
    C_Vertex3f(0.290908813f, 0.572106838f, 0.f),
    C_Vertex3f(0.290908813f, 0.572106838f, -1.f),
    C_Normal3f(-0.207346141f, -0.97826767f, 0.f),
    C_Vertex3f(0.279395103f, 0.575115919f, 0.f),
    C_Vertex3f(0.279395103f, 0.575115919f, -1.f),
    C_Normal3f(-0.119831458f, -0.992794275f, 0.f),
    C_Vertex3f(0.2672472f, 0.577121973f, 0.f),
    C_Vertex3f(0.2672472f, 0.577121973f, -1.f),
    C_Normal3f(-0.0377108715f, -0.999288738f, 0.f),
    C_Vertex3f(0.254465103f, 0.578125f, 0.f),
    C_Vertex3f(0.254465103f, 0.578125f, -1.f),
    C_Normal3f(0.0390183516f, -0.999238491f, 0.f),
    C_Vertex3f(0.240668297f, 0.578125f, 0.f),
    C_Vertex3f(0.240668297f, 0.578125f, -1.f),
    C_Normal3f(0.120883375f, -0.992666721f, 0.f),
    C_Vertex3f(0.227199793f, 0.577060342f, 0.f),
    C_Vertex3f(0.227199793f, 0.577060342f, -1.f),
    C_Normal3f(0.209783167f, -0.977747917f, 0.f),
    C_Vertex3f(0.214440107f, 0.574931026f, 0.f),
    C_Vertex3f(0.214440107f, 0.574931026f, -1.f),
    C_Normal3f(0.303551555f, -0.952815056f, 0.f),
    C_Vertex3f(0.20238924f, 0.571737051f, 0.f),
    C_Vertex3f(0.20238924f, 0.571737051f, -1.f),
    C_Normal3f(0.399689019f, -0.916650832f, 0.f),
    C_Vertex3f(0.191047192f, 0.567478418f, 0.f),
    C_Vertex3f(0.191047192f, 0.567478418f, -1.f),
    C_Normal3f(0.494991928f, -0.868897498f, 0.f),
    C_Vertex3f(0.180413961f, 0.562155128f, 0.f),
    C_Vertex3f(0.180413961f, 0.562155128f, -1.f),
    C_Normal3f(0.585969567f, -0.810333073f, 0.f),
    C_Vertex3f(0.17048955f, 0.555767179f, 0.f),
    C_Vertex3f(0.17048955f, 0.555767179f, -1.f),
    C_Normal3f(0.669421732f, -0.74288249f, 0.f),
    C_Vertex3f(0.161273956f, 0.548314571f, 0.f),
    C_Vertex3f(0.161273956f, 0.548314571f, -1.f),
    C_Normal3f(0.74545902f, -0.666551352f, 0.f),
    C_Vertex3f(0.152767181f, 0.539797306f, 0.f),
    C_Vertex3f(0.152767181f, 0.539797306f, -1.f),
    C_Normal3f(0.814887226f, -0.579619527f, 0.f),
    C_Vertex3f(0.145310163f, 0.530460954f, 0.f),
    C_Vertex3f(0.145310163f, 0.530460954f, -1.f),
    C_Normal3f(0.873702347f, -0.486460924f, 0.f),
    C_Vertex3f(0.138902903f, 0.520305514f, 0.f),
    C_Vertex3f(0.138902903f, 0.520305514f, -1.f),
    C_Normal3f(0.92049402f, -0.390756786f, 0.f),
    C_Vertex3f(0.133545399f, 0.509330988f, 0.f),
    C_Vertex3f(0.133545399f, 0.509330988f, -1.f),
    C_Normal3f(0.955159783f, -0.296090901f, 0.f),
    C_Vertex3f(0.129237652f, 0.497537374f, 0.f),
    C_Vertex3f(0.129237652f, 0.497537374f, -1.f),
    C_Normal3f(0.978677094f, -0.205404967f, 0.f),
    C_Vertex3f(0.125979662f, 0.484924674f, 0.f),
    C_Vertex3f(0.125979662f, 0.484924674f, -1.f),
    C_Normal3f(0.992685616f, -0.120728396f, 0.f),
    C_Vertex3f(0.123771429f, 0.471492887f, 0.f),
    C_Vertex3f(0.123771429f, 0.471492887f, -1.f),
    C_Normal3f(0.998453915f, -0.0555846989f, 0.f),
    C_Vertex3f(0.122612953f, 0.457242012f, 0.f),
    C_Vertex3f(0.122612953f, 0.457242012f, -1.f),
    C_Normal3f(0.999974012f, -0.00721408939f, 0.f),
    C_Vertex3f(0.122558594f, 0.449707031f, 0.f),
    C_Vertex3f(0.122558594f, 0.449707031f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.10187681f, -0.994797051f, 0.f),
    C_Vertex3f(0.122558594f, 0.449707031f, 0.f),
    C_Vertex3f(0.122558594f, 0.449707031f, -1.f),
    C_Vertex3f(0.0415039063f, 0.458007813f, -1.f),
    C_Vertex3f(0.0415039063f, 0.458007813f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99580586f, 0.0914910287f, 0.f),
    C_Vertex3f(0.0415039063f, 0.458007813f, 0.f),
    C_Vertex3f(0.0415039063f, 0.458007813f, -1.f),
    C_Normal3f(-0.990376651f, 0.138397574f, 0.f),
    C_Vertex3f(0.0425415039f, 0.469301224f, 0.f),
    C_Vertex3f(0.0425415039f, 0.469301224f, -1.f),
    C_Normal3f(-0.979543269f, 0.201233596f, 0.f),
    C_Vertex3f(0.0460506678f, 0.490544558f, 0.f),
    C_Vertex3f(0.0460506678f, 0.490544558f, -1.f),
    C_Normal3f(-0.959300697f, 0.282386392f, 0.f),
    C_Vertex3f(0.0509938002f, 0.510444403f, 0.f),
    C_Vertex3f(0.0509938002f, 0.510444403f, -1.f),
    C_Normal3f(-0.929543912f, 0.368711591f, 0.f),
    C_Vertex3f(0.0573709011f, 0.529000759f, 0.f),
    C_Vertex3f(0.0573709011f, 0.529000759f, -1.f),
    C_Normal3f(-0.888824284f, 0.458248258f, 0.f),
    C_Vertex3f(0.0651819706f, 0.546213627f, 0.f),
    C_Vertex3f(0.0651819706f, 0.546213627f, -1.f),
    C_Normal3f(-0.836336195f, 0.548216879f, 0.f),
    C_Vertex3f(0.0744270086f, 0.562083006f, 0.f),
    C_Vertex3f(0.0744270086f, 0.562083006f, -1.f),
    C_Normal3f(-0.772294104f, 0.635265172f, 0.f),
    C_Vertex3f(0.0851060152f, 0.576608896f, 0.f),
    C_Vertex3f(0.0851060152f, 0.576608896f, -1.f),
    C_Normal3f(-0.698141515f, 0.715959787f, 0.f),
    C_Vertex3f(0.0972189903f, 0.589791298f, 0.f),
    C_Vertex3f(0.0972189903f, 0.589791298f, -1.f),
    C_Normal3f(-0.617785335f, 0.786346734f, 0.f),
    C_Vertex3f(0.110765934f, 0.601630211f, 0.f),
    C_Vertex3f(0.110765934f, 0.601630211f, -1.f),
    C_Normal3f(-0.532547772f, 0.846399903f, 0.f),
    C_Vertex3f(0.125473619f, 0.61198926f, 0.f),
    C_Vertex3f(0.125473619f, 0.61198926f, -1.f),
    C_Normal3f(-0.443498164f, 0.896275282f, 0.f),
    C_Vertex3f(0.141342044f, 0.620868444f, 0.f),
    C_Vertex3f(0.141342044f, 0.620868444f, -1.f),
    C_Normal3f(-0.353721142f, 0.935350895f, 0.f),
    C_Vertex3f(0.15837121f, 0.628267765f, 0.f),
    C_Vertex3f(0.15837121f, 0.628267765f, -1.f),
    C_Normal3f(-0.266001403f, 0.963972628f, 0.f),
    C_Vertex3f(0.176561117f, 0.634187222f, 0.f),
    C_Vertex3f(0.176561117f, 0.634187222f, -1.f),
    C_Normal3f(-0.182505667f, 0.983204782f, 0.f),
    C_Vertex3f(0.195911765f, 0.638626814f, 0.f),
    C_Vertex3f(0.195911765f, 0.638626814f, -1.f),
    C_Normal3f(-0.104666635f, 0.994507313f, 0.f),
    C_Vertex3f(0.216423154f, 0.641586542f, 0.f),
    C_Vertex3f(0.216423154f, 0.641586542f, -1.f),
    C_Normal3f(-0.0331516638f, 0.999450326f, 0.f),
    C_Vertex3f(0.238095284f, 0.643066406f, 0.f),
    C_Vertex3f(0.238095284f, 0.643066406f, -1.f),
    C_Normal3f(0.0356260724f, 0.999365151f, 0.f),
    C_Vertex3f(0.261037827f, 0.643066406f, 0.f),
    C_Vertex3f(0.261037827f, 0.643066406f, -1.f),
    C_Normal3f(0.112009667f, 0.993707061f, 0.f),
    C_Vertex3f(0.282888174f, 0.641469598f, 0.f),
    C_Vertex3f(0.282888174f, 0.641469598f, -1.f),
    C_Normal3f(0.195293441f, 0.980744839f, 0.f),
    C_Vertex3f(0.303536654f, 0.638275981f, 0.f),
    C_Vertex3f(0.303536654f, 0.638275981f, -1.f),
    C_Normal3f(0.284318686f, 0.958729923f, 0.f),
    C_Vertex3f(0.322983265f, 0.633485556f, 0.f),
    C_Vertex3f(0.322983265f, 0.633485556f, -1.f),
    C_Normal3f(0.377180696f, 0.926139653f, 0.f),
    C_Vertex3f(0.341228008f, 0.627098322f, 0.f),
    C_Vertex3f(0.341228008f, 0.627098322f, -1.f),
    C_Normal3f(0.471148282f, 0.882054031f, 0.f),
    C_Vertex3f(0.358270884f, 0.61911428f, 0.f),
    C_Vertex3f(0.358270884f, 0.61911428f, -1.f),
    C_Normal3f(0.562903404f, 0.826522648f, 0.f),
    C_Vertex3f(0.374111891f, 0.609533429f, 0.f),
    C_Vertex3f(0.374111891f, 0.609533429f, -1.f),
    C_Normal3f(0.649018824f, 0.760772467f, 0.f),
    C_Vertex3f(0.38875103f, 0.59835577f, 0.f),
    C_Vertex3f(0.38875103f, 0.59835577f, -1.f),
    C_Normal3f(0.722285569f, 0.691594899f, 0.f),
    C_Vertex3f(0.402188301f, 0.585581303f, 0.f),
    C_Vertex3f(0.402188301f, 0.585581303f, -1.f),
    C_Normal3f(0.786860168f, 0.617131352f, 0.f),
    C_Vertex3f(0.413945913f, 0.572042823f, 0.f),
    C_Vertex3f(0.413945913f, 0.572042823f, -1.f),
    C_Normal3f(0.84643209f, 0.532496572f, 0.f),
    C_Vertex3f(0.424023867f, 0.557740331f, 0.f),
    C_Vertex3f(0.424023867f, 0.557740331f, -1.f),
    C_Normal3f(0.898249805f, 0.439485073f, 0.f),
    C_Vertex3f(0.432422161f, 0.542673826f, 0.f),
    C_Vertex3f(0.432422161f, 0.542673826f, -1.f),
    C_Normal3f(0.940103292f, 0.340889901f, 0.f),
    C_Vertex3f(0.439140797f, 0.526843309f, 0.f),
    C_Vertex3f(0.439140797f, 0.526843309f, -1.f),
    C_Normal3f(0.970744789f, 0.240113795f, 0.f),
    C_Vertex3f(0.444179773f, 0.51024878f, 0.f),
    C_Vertex3f(0.444179773f, 0.51024878f, -1.f),
    C_Normal3f(0.990065336f, 0.14060773f, 0.f),
    C_Vertex3f(0.447539091f, 0.492890239f, 0.f),
    C_Vertex3f(0.447539091f, 0.492890239f, -1.f),
    C_Normal3f(0.99898082f, 0.0451363623f, 0.f),
    C_Vertex3f(0.44921875f, 0.474767685f, 0.f),
    C_Vertex3f(0.44921875f, 0.474767685f, -1.f),
    C_Normal3f(0.998674393f, -0.0514739379f, 0.f),
    C_Vertex3f(0.44921875f, 0.455715179f, 0.f),
    C_Vertex3f(0.44921875f, 0.455715179f, -1.f),
    C_Normal3f(0.98815608f, -0.153452113f, 0.f),
    C_Vertex3f(0.447254658f, 0.436661243f, 0.f),
    C_Vertex3f(0.447254658f, 0.436661243f, -1.f),
    C_Normal3f(0.967566907f, -0.252615154f, 0.f),
    C_Vertex3f(0.443326473f, 0.417771816f, 0.f),
    C_Vertex3f(0.443326473f, 0.417771816f, -1.f),
    C_Normal3f(0.938245475f, -0.345970005f, 0.f),
    C_Vertex3f(0.437434196f, 0.399046898f, 0.f),
    C_Vertex3f(0.437434196f, 0.399046898f, -1.f),
    C_Normal3f(0.899177194f, -0.437584817f, 0.f),
    C_Vertex3f(0.429577827f, 0.380486488f, 0.f),
    C_Vertex3f(0.429577827f, 0.380486488f, -1.f),
    C_Normal3f(0.854257643f, -0.519849837f, 0.f),
    C_Vertex3f(0.419130325f, 0.361434937f, 0.f),
    C_Vertex3f(0.419130325f, 0.361434937f, -1.f),
    C_Normal3f(0.809849381f, -0.586637795f, 0.f),
    C_Vertex3f(0.40609169f, 0.341892242f, 0.f),
    C_Vertex3f(0.40609169f, 0.341892242f, -1.f),
    C_Normal3f(0.77064985f, -0.637258768f, 0.f),
    C_Vertex3f(0.390461922f, 0.321858406f, 0.f),
    C_Vertex3f(0.390461922f, 0.321858406f, -1.f),
    C_Normal3f(0.738712132f, -0.674021065f, 0.f),
    C_Vertex3f(0.376796246f, 0.306464672f, 0.f),
    C_Vertex3f(0.376796246f, 0.306464672f, -1.f),
    C_Normal3f(0.717080295f, -0.696990609f, 0.f),
    C_Vertex3f(0.366179824f, 0.295245767f, 0.f),
    C_Vertex3f(0.366179824f, 0.295245767f, -1.f),
    C_Normal3f(0.700935543f, -0.713224709f, 0.f),
    C_Vertex3f(0.354057431f, 0.283070445f, 0.f),
    C_Vertex3f(0.354057431f, 0.283070445f, -1.f),
    C_Normal3f(0.687358856f, -0.726317942f, 0.f),
    C_Vertex3f(0.340429068f, 0.269938707f, 0.f),
    C_Vertex3f(0.340429068f, 0.269938707f, -1.f),
    C_Normal3f(0.675799847f, -0.737085104f, 0.f),
    C_Vertex3f(0.325294733f, 0.255850554f, 0.f),
    C_Vertex3f(0.325294733f, 0.255850554f, -1.f),
    C_Normal3f(0.665849984f, -0.746085703f, 0.f),
    C_Vertex3f(0.308654428f, 0.240805984f, 0.f),
    C_Vertex3f(0.308654428f, 0.240805984f, -1.f),
    C_Normal3f(0.65720129f, -0.753715217f, 0.f),
    C_Vertex3f(0.290508151f, 0.224804997f, 0.f),
    C_Vertex3f(0.290508151f, 0.224804997f, -1.f),
    C_Normal3f(0.649056852f, -0.760739863f, 0.f),
    C_Vertex3f(0.270855904f, 0.207847595f, 0.f),
    C_Vertex3f(0.270855904f, 0.207847595f, -1.f),
    C_Normal3f(0.645620167f, -0.763658643f, 0.f),
    C_Vertex3f(0.251429558f, 0.19146347f, 0.f),
    C_Vertex3f(0.251429558f, 0.19146347f, -1.f),
    C_Normal3f(0.648851573f, -0.760914981f, 0.f),
    C_Vertex3f(0.235319018f, 0.177803636f, 0.f),
    C_Vertex3f(0.235319018f, 0.177803636f, -1.f),
    C_Normal3f(0.653815746f, -0.756653845f, 0.f),
    C_Vertex3f(0.220792413f, 0.165338397f, 0.f),
    C_Vertex3f(0.220792413f, 0.165338397f, -1.f),
    C_Normal3f(0.659955919f, -0.751304328f, 0.f),
    C_Vertex3f(0.207849741f, 0.154067755f, 0.f),
    C_Vertex3f(0.207849741f, 0.154067755f, -1.f),
    C_Normal3f(0.667743862f, -0.744391203f, 0.f),
    C_Vertex3f(0.196491003f, 0.143991709f, 0.f),
    C_Vertex3f(0.196491003f, 0.143991709f, -1.f),
    C_Normal3f(0.677939892f, -0.735117376f, 0.f),
    C_Vertex3f(0.186716199f, 0.135110259f, 0.f),
    C_Vertex3f(0.186716199f, 0.135110259f, -1.f),
    C_Normal3f(0.691851437f, -0.722039938f, 0.f),
    C_Vertex3f(0.178525329f, 0.127423406f, 0.f),
    C_Vertex3f(0.178525329f, 0.127423406f, -1.f),
    C_Normal3f(0.717426002f, -0.696634769f, 0.f),
    C_Vertex3f(0.171918392f, 0.120931149f, 0.f),
    C_Vertex3f(0.171918392f, 0.120931149f, -1.f),
    C_Normal3f(0.764170766f, -0.645013988f, 0.f),
    C_Vertex3f(0.159360886f, 0.107686996f, 0.f),
    C_Vertex3f(0.159360886f, 0.107686996f, -1.f),
    C_Normal3f(0.810201466f, -0.586151481f, 0.f),
    C_Vertex3f(0.14276123f, 0.0863876343f, 0.f),
    C_Vertex3f(0.14276123f, 0.0863876343f, -1.f),
    C_Normal3f(0.852845788f, -0.522162855f, 0.f),
    C_Vertex3f(0.136207581f, 0.0756835938f, 0.f),
    C_Vertex3f(0.136207581f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.136207581f, 0.0756835938f, 0.f),
    C_Vertex3f(0.136207581f, 0.0756835938f, -1.f),
    C_Vertex3f(0.450195313f, 0.0756835938f, -1.f),
    C_Vertex3f(0.450195313f, 0.0756835938f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136207581f, 0.0756835938f, 0.f),
    C_Vertex3f(0.14276123f, 0.0863876343f, 0.f),
    C_Vertex3f(0.137840271f, 0.191162825f, 0.f),
    C_Vertex3f(0.124984741f, 0.179055452f, 0.f),
    C_Vertex3f(0.113114834f, 0.167353988f, 0.f),
    C_Vertex3f(0.102230549f, 0.156058431f, 0.f),
    C_Vertex3f(0.0923318863f, 0.145168781f, 0.f),
    C_Vertex3f(0.078962326f, 0.129443169f, 0.f),
    C_Vertex3f(0.0635700226f, 0.108303547f, 0.f),
    C_Vertex3f(0.050611496f, 0.086991787f, 0.f),
    C_Vertex3f(0.450195313f, 0.f, 0.f),
    C_Vertex3f(0.450195313f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.450195313f, 0.f, 0.f),
    C_Vertex3f(0.050611496f, 0.086991787f, 0.f),
    C_Vertex3f(0.0400867462f, 0.0655078888f, 0.f),
    C_Vertex3f(0.0335273743f, 0.0481185913f, 0.f),
    C_Vertex3f(0.0298652649f, 0.0347223282f, 0.f),
    C_Vertex3f(0.0275688171f, 0.0210523605f, 0.f),
    C_Vertex3f(0.0268554688f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.36712122f, 0.484272718f, 0.f),
    C_Vertex3f(0.368164063f, 0.473049164f, 0.f),
    C_Vertex3f(0.374111891f, 0.609533429f, 0.f),
    C_Vertex3f(0.376796246f, 0.306464672f, 0.f),
    C_Vertex3f(0.38875103f, 0.59835577f, 0.f),
    C_Vertex3f(0.390461922f, 0.321858406f, 0.f),
    C_Vertex3f(0.402188301f, 0.585581303f, 0.f),
    C_Vertex3f(0.40609169f, 0.341892242f, 0.f),
    C_Vertex3f(0.413945913f, 0.572042823f, 0.f),
    C_Vertex3f(0.419130325f, 0.361434937f, 0.f),
    C_Vertex3f(0.424023867f, 0.557740331f, 0.f),
    C_Vertex3f(0.429577827f, 0.380486488f, 0.f),
    C_Vertex3f(0.432422161f, 0.542673826f, 0.f),
    C_Vertex3f(0.437434196f, 0.399046898f, 0.f),
    C_Vertex3f(0.439140797f, 0.526843309f, 0.f),
    C_Vertex3f(0.443326473f, 0.417771816f, 0.f),
    C_Vertex3f(0.444179773f, 0.51024878f, 0.f),
    C_Vertex3f(0.447254658f, 0.436661243f, 0.f),
    C_Vertex3f(0.447539091f, 0.492890239f, 0.f),
    C_Vertex3f(0.44921875f, 0.455715179f, 0.f),
    C_Vertex3f(0.44921875f, 0.474767685f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.35252142f, 0.524596453f, 0.f),
    C_Vertex3f(0.357735634f, 0.515201092f, 0.f),
    C_Vertex3f(0.358270884f, 0.61911428f, 0.f),
    C_Vertex3f(0.361907005f, 0.505348682f, 0.f),
    C_Vertex3f(0.374111891f, 0.609533429f, 0.f),
    C_Vertex3f(0.365035534f, 0.495039225f, 0.f),
    C_Vertex3f(0.36712122f, 0.484272718f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.330621719f, 0.550040245f, 0.f),
    C_Vertex3f(0.338964462f, 0.542016029f, 0.f),
    C_Vertex3f(0.341228008f, 0.627098322f, 0.f),
    C_Vertex3f(0.346264362f, 0.533534765f, 0.f),
    C_Vertex3f(0.358270884f, 0.61911428f, 0.f),
    C_Vertex3f(0.35252142f, 0.524596453f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.322983265f, 0.633485556f, 0.f),
    C_Vertex3f(0.303536654f, 0.638275981f, 0.f),
    C_Vertex3f(0.312033653f, 0.563079596f, 0.f),
    C_Vertex3f(0.321644783f, 0.557061434f, 0.f),
    C_Vertex3f(0.330621719f, 0.550040245f, 0.f),
    C_Vertex3f(0.341228008f, 0.627098322f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.2672472f, 0.577121973f, 0.f),
    C_Vertex3f(0.279395103f, 0.575115919f, 0.f),
    C_Vertex3f(0.282888174f, 0.641469598f, 0.f),
    C_Vertex3f(0.290908813f, 0.572106838f, 0.f),
    C_Vertex3f(0.303536654f, 0.638275981f, 0.f),
    C_Vertex3f(0.30178833f, 0.56809473f, 0.f),
    C_Vertex3f(0.312033653f, 0.563079596f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.261037827f, 0.643066406f, 0.f),
    C_Vertex3f(0.238095284f, 0.643066406f, 0.f),
    C_Vertex3f(0.240668297f, 0.578125f, 0.f),
    C_Vertex3f(0.254465103f, 0.578125f, 0.f),
    C_Vertex3f(0.2672472f, 0.577121973f, 0.f),
    C_Vertex3f(0.282888174f, 0.641469598f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.20238924f, 0.571737051f, 0.f),
    C_Vertex3f(0.214440107f, 0.574931026f, 0.f),
    C_Vertex3f(0.216423154f, 0.641586542f, 0.f),
    C_Vertex3f(0.227199793f, 0.577060342f, 0.f),
    C_Vertex3f(0.238095284f, 0.643066406f, 0.f),
    C_Vertex3f(0.240668297f, 0.578125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.195911765f, 0.638626814f, 0.f),
    C_Vertex3f(0.176561117f, 0.634187222f, 0.f),
    C_Vertex3f(0.180413961f, 0.562155128f, 0.f),
    C_Vertex3f(0.191047192f, 0.567478418f, 0.f),
    C_Vertex3f(0.20238924f, 0.571737051f, 0.f),
    C_Vertex3f(0.216423154f, 0.641586542f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145310163f, 0.530460954f, 0.f),
    C_Vertex3f(0.152767181f, 0.539797306f, 0.f),
    C_Vertex3f(0.15837121f, 0.628267765f, 0.f),
    C_Vertex3f(0.161273956f, 0.548314571f, 0.f),
    C_Vertex3f(0.176561117f, 0.634187222f, 0.f),
    C_Vertex3f(0.17048955f, 0.555767179f, 0.f),
    C_Vertex3f(0.180413961f, 0.562155128f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.141342044f, 0.620868444f, 0.f),
    C_Vertex3f(0.125473619f, 0.61198926f, 0.f),
    C_Vertex3f(0.125979662f, 0.484924674f, 0.f),
    C_Vertex3f(0.129237652f, 0.497537374f, 0.f),
    C_Vertex3f(0.133545399f, 0.509330988f, 0.f),
    C_Vertex3f(0.138902903f, 0.520305514f, 0.f),
    C_Vertex3f(0.145310163f, 0.530460954f, 0.f),
    C_Vertex3f(0.15837121f, 0.628267765f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0972189903f, 0.589791298f, 0.f),
    C_Vertex3f(0.122558594f, 0.449707031f, 0.f),
    C_Vertex3f(0.110765934f, 0.601630211f, 0.f),
    C_Vertex3f(0.122612953f, 0.457242012f, 0.f),
    C_Vertex3f(0.125473619f, 0.61198926f, 0.f),
    C_Vertex3f(0.123771429f, 0.471492887f, 0.f),
    C_Vertex3f(0.125979662f, 0.484924674f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122558594f, 0.449707031f, 0.f),
    C_Vertex3f(0.0972189903f, 0.589791298f, 0.f),
    C_Vertex3f(0.0851060152f, 0.576608896f, 0.f),
    C_Vertex3f(0.0744270086f, 0.562083006f, 0.f),
    C_Vertex3f(0.0651819706f, 0.546213627f, 0.f),
    C_Vertex3f(0.0573709011f, 0.529000759f, 0.f),
    C_Vertex3f(0.0509938002f, 0.510444403f, 0.f),
    C_Vertex3f(0.0460506678f, 0.490544558f, 0.f),
    C_Vertex3f(0.0425415039f, 0.469301224f, 0.f),
    C_Vertex3f(0.0415039063f, 0.458007813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.368164063f, 0.473049164f, 0.f),
    C_Vertex3f(0.368164063f, 0.461625099f, 0.f),
    C_Vertex3f(0.376796246f, 0.306464672f, 0.f),
    C_Vertex3f(0.367020845f, 0.450291038f, 0.f),
    C_Vertex3f(0.366179824f, 0.295245767f, 0.f),
    C_Vertex3f(0.364734411f, 0.438790441f, 0.f),
    C_Vertex3f(0.36130476f, 0.427123308f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356731892f, 0.41528964f, 0.f),
    C_Vertex3f(0.351015806f, 0.403289437f, 0.f),
    C_Vertex3f(0.354057431f, 0.283070445f, 0.f),
    C_Vertex3f(0.366179824f, 0.295245767f, 0.f),
    C_Vertex3f(0.36130476f, 0.427123308f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.344156504f, 0.391122699f, 0.f),
    C_Vertex3f(0.336153984f, 0.378789425f, 0.f),
    C_Vertex3f(0.340429068f, 0.269938707f, 0.f),
    C_Vertex3f(0.354057431f, 0.283070445f, 0.f),
    C_Vertex3f(0.351015806f, 0.403289437f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.207849741f, 0.154067755f, 0.f),
    C_Vertex3f(0.220792413f, 0.165338397f, 0.f),
    C_Vertex3f(0.227708578f, 0.267826438f, 0.f),
    C_Vertex3f(0.235319018f, 0.177803636f, 0.f),
    C_Vertex3f(0.249547243f, 0.287030339f, 0.f),
    C_Vertex3f(0.251429558f, 0.19146347f, 0.f),
    C_Vertex3f(0.26927042f, 0.305116892f, 0.f),
    C_Vertex3f(0.270855904f, 0.207847595f, 0.f),
    C_Vertex3f(0.286878109f, 0.322086096f, 0.f),
    C_Vertex3f(0.290508151f, 0.224804997f, 0.f),
    C_Vertex3f(0.30237031f, 0.337937951f, 0.f),
    C_Vertex3f(0.308654428f, 0.240805984f, 0.f),
    C_Vertex3f(0.315747023f, 0.352672458f, 0.f),
    C_Vertex3f(0.325294733f, 0.255850554f, 0.f),
    C_Vertex3f(0.327008247f, 0.366289616f, 0.f),
    C_Vertex3f(0.340429068f, 0.269938707f, 0.f),
    C_Vertex3f(0.336153984f, 0.378789425f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.203754425f, 0.247505188f, 0.f),
    C_Vertex3f(0.182320595f, 0.229920387f, 0.f),
    C_Vertex3f(0.186716199f, 0.135110259f, 0.f),
    C_Vertex3f(0.196491003f, 0.143991709f, 0.f),
    C_Vertex3f(0.207849741f, 0.154067755f, 0.f),
    C_Vertex3f(0.227708578f, 0.267826438f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.137840271f, 0.191162825f, 0.f),
    C_Vertex3f(0.14276123f, 0.0863876343f, 0.f),
    C_Vertex3f(0.151681423f, 0.203676105f, 0.f),
    C_Vertex3f(0.159360886f, 0.107686996f, 0.f),
    C_Vertex3f(0.166508198f, 0.216595292f, 0.f),
    C_Vertex3f(0.171918392f, 0.120931149f, 0.f),
    C_Vertex3f(0.182320595f, 0.229920387f, 0.f),
    C_Vertex3f(0.178525329f, 0.127423406f, 0.f),
    C_Vertex3f(0.186716199f, 0.135110259f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0268554688f, 0.f, 0.f),
    C_Vertex3f(0.0275688171f, 0.0210523605f, 0.f),
    C_Vertex3f(0.026638031f, 0.00710868835f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136207581f, 0.0756835938f, -1.f),
    C_Vertex3f(0.14276123f, 0.0863876343f, -1.f),
    C_Vertex3f(0.137840271f, 0.191162825f, -1.f),
    C_Vertex3f(0.124984741f, 0.179055452f, -1.f),
    C_Vertex3f(0.113114834f, 0.167353988f, -1.f),
    C_Vertex3f(0.102230549f, 0.156058431f, -1.f),
    C_Vertex3f(0.0923318863f, 0.145168781f, -1.f),
    C_Vertex3f(0.078962326f, 0.129443169f, -1.f),
    C_Vertex3f(0.0635700226f, 0.108303547f, -1.f),
    C_Vertex3f(0.050611496f, 0.086991787f, -1.f),
    C_Vertex3f(0.450195313f, 0.f, -1.f),
    C_Vertex3f(0.450195313f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.450195313f, 0.f, -1.f),
    C_Vertex3f(0.050611496f, 0.086991787f, -1.f),
    C_Vertex3f(0.0400867462f, 0.0655078888f, -1.f),
    C_Vertex3f(0.0335273743f, 0.0481185913f, -1.f),
    C_Vertex3f(0.0298652649f, 0.0347223282f, -1.f),
    C_Vertex3f(0.0275688171f, 0.0210523605f, -1.f),
    C_Vertex3f(0.0268554688f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.36712122f, 0.484272718f, -1.f),
    C_Vertex3f(0.368164063f, 0.473049164f, -1.f),
    C_Vertex3f(0.374111891f, 0.609533429f, -1.f),
    C_Vertex3f(0.376796246f, 0.306464672f, -1.f),
    C_Vertex3f(0.38875103f, 0.59835577f, -1.f),
    C_Vertex3f(0.390461922f, 0.321858406f, -1.f),
    C_Vertex3f(0.402188301f, 0.585581303f, -1.f),
    C_Vertex3f(0.40609169f, 0.341892242f, -1.f),
    C_Vertex3f(0.413945913f, 0.572042823f, -1.f),
    C_Vertex3f(0.419130325f, 0.361434937f, -1.f),
    C_Vertex3f(0.424023867f, 0.557740331f, -1.f),
    C_Vertex3f(0.429577827f, 0.380486488f, -1.f),
    C_Vertex3f(0.432422161f, 0.542673826f, -1.f),
    C_Vertex3f(0.437434196f, 0.399046898f, -1.f),
    C_Vertex3f(0.439140797f, 0.526843309f, -1.f),
    C_Vertex3f(0.443326473f, 0.417771816f, -1.f),
    C_Vertex3f(0.444179773f, 0.51024878f, -1.f),
    C_Vertex3f(0.447254658f, 0.436661243f, -1.f),
    C_Vertex3f(0.447539091f, 0.492890239f, -1.f),
    C_Vertex3f(0.44921875f, 0.455715179f, -1.f),
    C_Vertex3f(0.44921875f, 0.474767685f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.35252142f, 0.524596453f, -1.f),
    C_Vertex3f(0.357735634f, 0.515201092f, -1.f),
    C_Vertex3f(0.358270884f, 0.61911428f, -1.f),
    C_Vertex3f(0.361907005f, 0.505348682f, -1.f),
    C_Vertex3f(0.374111891f, 0.609533429f, -1.f),
    C_Vertex3f(0.365035534f, 0.495039225f, -1.f),
    C_Vertex3f(0.36712122f, 0.484272718f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.330621719f, 0.550040245f, -1.f),
    C_Vertex3f(0.338964462f, 0.542016029f, -1.f),
    C_Vertex3f(0.341228008f, 0.627098322f, -1.f),
    C_Vertex3f(0.346264362f, 0.533534765f, -1.f),
    C_Vertex3f(0.358270884f, 0.61911428f, -1.f),
    C_Vertex3f(0.35252142f, 0.524596453f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.322983265f, 0.633485556f, -1.f),
    C_Vertex3f(0.303536654f, 0.638275981f, -1.f),
    C_Vertex3f(0.312033653f, 0.563079596f, -1.f),
    C_Vertex3f(0.321644783f, 0.557061434f, -1.f),
    C_Vertex3f(0.330621719f, 0.550040245f, -1.f),
    C_Vertex3f(0.341228008f, 0.627098322f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.2672472f, 0.577121973f, -1.f),
    C_Vertex3f(0.279395103f, 0.575115919f, -1.f),
    C_Vertex3f(0.282888174f, 0.641469598f, -1.f),
    C_Vertex3f(0.290908813f, 0.572106838f, -1.f),
    C_Vertex3f(0.303536654f, 0.638275981f, -1.f),
    C_Vertex3f(0.30178833f, 0.56809473f, -1.f),
    C_Vertex3f(0.312033653f, 0.563079596f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.261037827f, 0.643066406f, -1.f),
    C_Vertex3f(0.238095284f, 0.643066406f, -1.f),
    C_Vertex3f(0.240668297f, 0.578125f, -1.f),
    C_Vertex3f(0.254465103f, 0.578125f, -1.f),
    C_Vertex3f(0.2672472f, 0.577121973f, -1.f),
    C_Vertex3f(0.282888174f, 0.641469598f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.20238924f, 0.571737051f, -1.f),
    C_Vertex3f(0.214440107f, 0.574931026f, -1.f),
    C_Vertex3f(0.216423154f, 0.641586542f, -1.f),
    C_Vertex3f(0.227199793f, 0.577060342f, -1.f),
    C_Vertex3f(0.238095284f, 0.643066406f, -1.f),
    C_Vertex3f(0.240668297f, 0.578125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.195911765f, 0.638626814f, -1.f),
    C_Vertex3f(0.176561117f, 0.634187222f, -1.f),
    C_Vertex3f(0.180413961f, 0.562155128f, -1.f),
    C_Vertex3f(0.191047192f, 0.567478418f, -1.f),
    C_Vertex3f(0.20238924f, 0.571737051f, -1.f),
    C_Vertex3f(0.216423154f, 0.641586542f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145310163f, 0.530460954f, -1.f),
    C_Vertex3f(0.152767181f, 0.539797306f, -1.f),
    C_Vertex3f(0.15837121f, 0.628267765f, -1.f),
    C_Vertex3f(0.161273956f, 0.548314571f, -1.f),
    C_Vertex3f(0.176561117f, 0.634187222f, -1.f),
    C_Vertex3f(0.17048955f, 0.555767179f, -1.f),
    C_Vertex3f(0.180413961f, 0.562155128f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.141342044f, 0.620868444f, -1.f),
    C_Vertex3f(0.125473619f, 0.61198926f, -1.f),
    C_Vertex3f(0.125979662f, 0.484924674f, -1.f),
    C_Vertex3f(0.129237652f, 0.497537374f, -1.f),
    C_Vertex3f(0.133545399f, 0.509330988f, -1.f),
    C_Vertex3f(0.138902903f, 0.520305514f, -1.f),
    C_Vertex3f(0.145310163f, 0.530460954f, -1.f),
    C_Vertex3f(0.15837121f, 0.628267765f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0972189903f, 0.589791298f, -1.f),
    C_Vertex3f(0.122558594f, 0.449707031f, -1.f),
    C_Vertex3f(0.110765934f, 0.601630211f, -1.f),
    C_Vertex3f(0.122612953f, 0.457242012f, -1.f),
    C_Vertex3f(0.125473619f, 0.61198926f, -1.f),
    C_Vertex3f(0.123771429f, 0.471492887f, -1.f),
    C_Vertex3f(0.125979662f, 0.484924674f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122558594f, 0.449707031f, -1.f),
    C_Vertex3f(0.0972189903f, 0.589791298f, -1.f),
    C_Vertex3f(0.0851060152f, 0.576608896f, -1.f),
    C_Vertex3f(0.0744270086f, 0.562083006f, -1.f),
    C_Vertex3f(0.0651819706f, 0.546213627f, -1.f),
    C_Vertex3f(0.0573709011f, 0.529000759f, -1.f),
    C_Vertex3f(0.0509938002f, 0.510444403f, -1.f),
    C_Vertex3f(0.0460506678f, 0.490544558f, -1.f),
    C_Vertex3f(0.0425415039f, 0.469301224f, -1.f),
    C_Vertex3f(0.0415039063f, 0.458007813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.368164063f, 0.473049164f, -1.f),
    C_Vertex3f(0.368164063f, 0.461625099f, -1.f),
    C_Vertex3f(0.376796246f, 0.306464672f, -1.f),
    C_Vertex3f(0.367020845f, 0.450291038f, -1.f),
    C_Vertex3f(0.366179824f, 0.295245767f, -1.f),
    C_Vertex3f(0.364734411f, 0.438790441f, -1.f),
    C_Vertex3f(0.36130476f, 0.427123308f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356731892f, 0.41528964f, -1.f),
    C_Vertex3f(0.351015806f, 0.403289437f, -1.f),
    C_Vertex3f(0.354057431f, 0.283070445f, -1.f),
    C_Vertex3f(0.366179824f, 0.295245767f, -1.f),
    C_Vertex3f(0.36130476f, 0.427123308f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.344156504f, 0.391122699f, -1.f),
    C_Vertex3f(0.336153984f, 0.378789425f, -1.f),
    C_Vertex3f(0.340429068f, 0.269938707f, -1.f),
    C_Vertex3f(0.354057431f, 0.283070445f, -1.f),
    C_Vertex3f(0.351015806f, 0.403289437f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.207849741f, 0.154067755f, -1.f),
    C_Vertex3f(0.220792413f, 0.165338397f, -1.f),
    C_Vertex3f(0.227708578f, 0.267826438f, -1.f),
    C_Vertex3f(0.235319018f, 0.177803636f, -1.f),
    C_Vertex3f(0.249547243f, 0.287030339f, -1.f),
    C_Vertex3f(0.251429558f, 0.19146347f, -1.f),
    C_Vertex3f(0.26927042f, 0.305116892f, -1.f),
    C_Vertex3f(0.270855904f, 0.207847595f, -1.f),
    C_Vertex3f(0.286878109f, 0.322086096f, -1.f),
    C_Vertex3f(0.290508151f, 0.224804997f, -1.f),
    C_Vertex3f(0.30237031f, 0.337937951f, -1.f),
    C_Vertex3f(0.308654428f, 0.240805984f, -1.f),
    C_Vertex3f(0.315747023f, 0.352672458f, -1.f),
    C_Vertex3f(0.325294733f, 0.255850554f, -1.f),
    C_Vertex3f(0.327008247f, 0.366289616f, -1.f),
    C_Vertex3f(0.340429068f, 0.269938707f, -1.f),
    C_Vertex3f(0.336153984f, 0.378789425f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.203754425f, 0.247505188f, -1.f),
    C_Vertex3f(0.182320595f, 0.229920387f, -1.f),
    C_Vertex3f(0.186716199f, 0.135110259f, -1.f),
    C_Vertex3f(0.196491003f, 0.143991709f, -1.f),
    C_Vertex3f(0.207849741f, 0.154067755f, -1.f),
    C_Vertex3f(0.227708578f, 0.267826438f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.137840271f, 0.191162825f, -1.f),
    C_Vertex3f(0.14276123f, 0.0863876343f, -1.f),
    C_Vertex3f(0.151681423f, 0.203676105f, -1.f),
    C_Vertex3f(0.159360886f, 0.107686996f, -1.f),
    C_Vertex3f(0.166508198f, 0.216595292f, -1.f),
    C_Vertex3f(0.171918392f, 0.120931149f, -1.f),
    C_Vertex3f(0.182320595f, 0.229920387f, -1.f),
    C_Vertex3f(0.178525329f, 0.127423406f, -1.f),
    C_Vertex3f(0.186716199f, 0.135110259f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0268554688f, 0.f, -1.f),
    C_Vertex3f(0.0275688171f, 0.0210523605f, -1.f),
    C_Vertex3f(0.026638031f, 0.00710868835f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_51[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.129339188f, 0.991600454f, 0.f),
    C_Vertex3f(0.0375976563f, 0.169921875f, 0.f),
    C_Vertex3f(0.0375976563f, 0.169921875f, -1.f),
    C_Vertex3f(0.116210938f, 0.180175781f, -1.f),
    C_Vertex3f(0.116210938f, 0.180175781f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.98013556f, 0.198328838f, 0.f),
    C_Vertex3f(0.116210938f, 0.180175781f, 0.f),
    C_Vertex3f(0.116210938f, 0.180175781f, -1.f),
    C_Normal3f(0.973069251f, 0.230512768f, 0.f),
    C_Vertex3f(0.117903709f, 0.17181015f, 0.f),
    C_Vertex3f(0.117903709f, 0.17181015f, -1.f),
    C_Normal3f(0.961673141f, 0.274198443f, 0.f),
    C_Vertex3f(0.121879339f, 0.156247616f, 0.f),
    C_Vertex3f(0.121879339f, 0.156247616f, -1.f),
    C_Normal3f(0.943272173f, 0.332020551f, 0.f),
    C_Vertex3f(0.126445055f, 0.141853809f, 0.f),
    C_Vertex3f(0.126445055f, 0.141853809f, -1.f),
    C_Normal3f(0.918264449f, 0.395967871f, 0.f),
    C_Vertex3f(0.131600857f, 0.128628731f, 0.f),
    C_Vertex3f(0.131600857f, 0.128628731f, -1.f),
    C_Normal3f(0.884821415f, 0.465930194f, 0.f),
    C_Vertex3f(0.137346745f, 0.11657238f, 0.f),
    C_Vertex3f(0.137346745f, 0.11657238f, -1.f),
    C_Normal3f(0.840900242f, 0.541190267f, 0.f),
    C_Vertex3f(0.143682718f, 0.105684757f, 0.f),
    C_Vertex3f(0.143682718f, 0.105684757f, -1.f),
    C_Normal3f(0.784480333f, 0.620153666f, 0.f),
    C_Vertex3f(0.150608778f, 0.0959658623f, 0.f),
    C_Vertex3f(0.150608778f, 0.0959658623f, -1.f),
    C_Normal3f(0.714002132f, 0.700143516f, 0.f),
    C_Vertex3f(0.158124924f, 0.0874156952f, 0.f),
    C_Vertex3f(0.158124924f, 0.0874156952f, -1.f),
    C_Normal3f(0.638830364f, 0.769347608f, 0.f),
    C_Vertex3f(0.166231155f, 0.080034256f, 0.f),
    C_Vertex3f(0.166231155f, 0.080034256f, -1.f),
    C_Normal3f(0.563639224f, 0.826021075f, 0.f),
    C_Vertex3f(0.174792767f, 0.0735754967f, 0.f),
    C_Vertex3f(0.174792767f, 0.0735754967f, -1.f),
    C_Normal3f(0.481202632f, 0.876609445f, 0.f),
    C_Vertex3f(0.183809757f, 0.0680394173f, 0.f),
    C_Vertex3f(0.183809757f, 0.0680394173f, -1.f),
    C_Normal3f(0.393510222f, 0.919320285f, 0.f),
    C_Vertex3f(0.193282127f, 0.0634260178f, 0.f),
    C_Vertex3f(0.193282127f, 0.0634260178f, -1.f),
    C_Normal3f(0.303042054f, 0.95297718f, 0.f),
    C_Vertex3f(0.203209877f, 0.0597352982f, 0.f),
    C_Vertex3f(0.203209877f, 0.0597352982f, -1.f),
    C_Normal3f(0.212429643f, 0.977176309f, 0.f),
    C_Vertex3f(0.213593006f, 0.0569672585f, 0.f),
    C_Vertex3f(0.213593006f, 0.0569672585f, -1.f),
    C_Normal3f(0.124100514f, 0.992269695f, 0.f),
    C_Vertex3f(0.224431515f, 0.0551218987f, 0.f),
    C_Vertex3f(0.224431515f, 0.0551218987f, -1.f),
    C_Normal3f(0.0382033736f, 0.999269962f, 0.f),
    C_Vertex3f(0.235725403f, 0.0541992188f, 0.f),
    C_Vertex3f(0.235725403f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0457790941f, 0.998951614f, 0.f),
    C_Vertex3f(0.248565674f, 0.0541992188f, 0.f),
    C_Vertex3f(0.248565674f, 0.0541992188f, -1.f),
    C_Normal3f(-0.136252552f, 0.990674138f, 0.f),
    C_Vertex3f(0.261946321f, 0.0554008484f, 0.f),
    C_Vertex3f(0.261946321f, 0.0554008484f, -1.f),
    C_Normal3f(-0.232709646f, 0.972546279f, 0.f),
    C_Vertex3f(0.27477634f, 0.0578041077f, 0.f),
    C_Vertex3f(0.27477634f, 0.0578041077f, -1.f),
    C_Normal3f(-0.330649942f, 0.943753481f, 0.f),
    C_Vertex3f(0.287055731f, 0.0614089966f, 0.f),
    C_Vertex3f(0.287055731f, 0.0614089966f, -1.f),
    C_Normal3f(-0.426926225f, 0.904286444f, 0.f),
    C_Vertex3f(0.298784494f, 0.0662155151f, 0.f),
    C_Vertex3f(0.298784494f, 0.0662155151f, -1.f),
    C_Normal3f(-0.518370092f, 0.855156422f, 0.f),
    C_Vertex3f(0.30996263f, 0.0722236633f, 0.f),
    C_Vertex3f(0.30996263f, 0.0722236633f, -1.f),
    C_Normal3f(-0.602289736f, 0.798277557f, 0.f),
    C_Vertex3f(0.320590138f, 0.0794334412f, 0.f),
    C_Vertex3f(0.320590138f, 0.0794334412f, -1.f),
    C_Normal3f(-0.676844239f, 0.736126304f, 0.f),
    C_Vertex3f(0.330667019f, 0.0878448486f, 0.f),
    C_Vertex3f(0.330667019f, 0.0878448486f, -1.f),
    C_Normal3f(-0.742510676f, 0.669834137f, 0.f),
    C_Vertex3f(0.340193272f, 0.0974578857f, 0.f),
    C_Vertex3f(0.340193272f, 0.0974578857f, -1.f),
    C_Normal3f(-0.804219961f, 0.594331741f, 0.f),
    C_Vertex3f(0.348528743f, 0.107644558f, 0.f),
    C_Vertex3f(0.348528743f, 0.107644558f, -1.f),
    C_Normal3f(-0.860194027f, 0.50996691f, 0.f),
    C_Vertex3f(0.355673432f, 0.118404865f, 0.f),
    C_Vertex3f(0.355673432f, 0.118404865f, -1.f),
    C_Normal3f(-0.908105314f, 0.418741792f, 0.f),
    C_Vertex3f(0.36162734f, 0.129738808f, 0.f),
    C_Vertex3f(0.36162734f, 0.129738808f, -1.f),
    C_Normal3f(-0.946259737f, 0.32340762f, 0.f),
    C_Vertex3f(0.366390467f, 0.141646385f, 0.f),
    C_Vertex3f(0.366390467f, 0.141646385f, -1.f),
    C_Normal3f(-0.973879278f, 0.227066651f, 0.f),
    C_Vertex3f(0.369962811f, 0.154127598f, 0.f),
    C_Vertex3f(0.369962811f, 0.154127598f, -1.f),
    C_Normal3f(-0.991156876f, 0.132695362f, 0.f),
    C_Vertex3f(0.372344375f, 0.167182446f, 0.f),
    C_Vertex3f(0.372344375f, 0.167182446f, -1.f),
    C_Normal3f(-0.999063969f, 0.0432566293f, 0.f),
    C_Vertex3f(0.373535156f, 0.180810928f, 0.f),
    C_Vertex3f(0.373535156f, 0.180810928f, -1.f),
    C_Normal3f(-0.999151349f, -0.0411898941f, 0.f),
    C_Vertex3f(0.373535156f, 0.194684982f, 0.f),
    C_Vertex3f(0.373535156f, 0.194684982f, -1.f),
    C_Normal3f(-0.991517961f, -0.129969835f, 0.f),
    C_Vertex3f(0.372429252f, 0.207637072f, 0.f),
    C_Vertex3f(0.372429252f, 0.207637072f, -1.f),
    C_Normal3f(-0.974719524f, -0.223432034f, 0.f),
    C_Vertex3f(0.370217443f, 0.21999526f, 0.f),
    C_Vertex3f(0.370217443f, 0.21999526f, -1.f),
    C_Normal3f(-0.947481275f, -0.319811255f, 0.f),
    C_Vertex3f(0.366899729f, 0.231759548f, 0.f),
    C_Vertex3f(0.366899729f, 0.231759548f, -1.f),
    C_Normal3f(-0.909289181f, -0.416164815f, 0.f),
    C_Vertex3f(0.36247611f, 0.242929935f, 0.f),
    C_Vertex3f(0.36247611f, 0.242929935f, -1.f),
    C_Normal3f(-0.860625505f, -0.509238303f, 0.f),
    C_Vertex3f(0.356946588f, 0.253506422f, 0.f),
    C_Vertex3f(0.356946588f, 0.253506422f, -1.f),
    C_Normal3f(-0.803008199f, -0.595967889f, 0.f),
    C_Vertex3f(0.35031116f, 0.263489008f, 0.f),
    C_Vertex3f(0.35031116f, 0.263489008f, -1.f),
    C_Normal3f(-0.738763273f, -0.673965156f, 0.f),
    C_Vertex3f(0.342569828f, 0.272877693f, 0.f),
    C_Vertex3f(0.342569828f, 0.272877693f, -1.f),
    C_Normal3f(-0.669515669f, -0.742797911f, 0.f),
    C_Vertex3f(0.333722591f, 0.281672478f, 0.f),
    C_Vertex3f(0.333722591f, 0.281672478f, -1.f),
    C_Normal3f(-0.591203511f, -0.806522369f, 0.f),
    C_Vertex3f(0.324274659f, 0.289367914f, 0.f),
    C_Vertex3f(0.324274659f, 0.289367914f, -1.f),
    C_Normal3f(-0.504463077f, -0.863433361f, 0.f),
    C_Vertex3f(0.314226031f, 0.295964003f, 0.f),
    C_Vertex3f(0.314226031f, 0.295964003f, -1.f),
    C_Normal3f(-0.41172865f, -0.911306441f, 0.f),
    C_Vertex3f(0.303576708f, 0.301460743f, 0.f),
    C_Vertex3f(0.303576708f, 0.301460743f, -1.f),
    C_Normal3f(-0.316050053f, -0.948742568f, 0.f),
    C_Vertex3f(0.292326689f, 0.305858135f, 0.f),
    C_Vertex3f(0.292326689f, 0.305858135f, -1.f),
    C_Normal3f(-0.220610797f, -0.975361943f, 0.f),
    C_Vertex3f(0.280475974f, 0.309156179f, 0.f),
    C_Vertex3f(0.280475974f, 0.309156179f, -1.f),
    C_Normal3f(-0.128249332f, -0.991742015f, 0.f),
    C_Vertex3f(0.268024564f, 0.311354876f, 0.f),
    C_Vertex3f(0.268024564f, 0.311354876f, -1.f),
    C_Normal3f(-0.0431595892f, -0.999068201f, 0.f),
    C_Vertex3f(0.254972458f, 0.312454224f, 0.f),
    C_Vertex3f(0.254972458f, 0.312454224f, -1.f),
    C_Normal3f(0.0438490137f, -0.99903816f, 0.f),
    C_Vertex3f(0.242576599f, 0.312454224f, 0.f),
    C_Vertex3f(0.242576599f, 0.312454224f, -1.f),
    C_Normal3f(0.123363428f, -0.992361546f, 0.f),
    C_Vertex3f(0.230072021f, 0.311361313f, 0.f),
    C_Vertex3f(0.230072021f, 0.311361313f, -1.f),
    C_Normal3f(0.184521854f, -0.982828438f, 0.f),
    C_Vertex3f(0.216201782f, 0.309175491f, 0.f),
    C_Vertex3f(0.216201782f, 0.309175491f, -1.f),
    C_Normal3f(0.226156473f, -0.974091053f, 0.f),
    C_Vertex3f(0.200965881f, 0.305896759f, 0.f),
    C_Vertex3f(0.200965881f, 0.305896759f, -1.f),
    C_Normal3f(0.25464642f, -0.967034161f, 0.f),
    C_Vertex3f(0.1926651f, 0.303710938f, 0.f),
    C_Vertex3f(0.1926651f, 0.303710938f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.992046177f, 0.125874653f, 0.f),
    C_Vertex3f(0.1926651f, 0.303710938f, 0.f),
    C_Vertex3f(0.1926651f, 0.303710938f, -1.f),
    C_Vertex3f(0.201400757f, 0.372558594f, -1.f),
    C_Vertex3f(0.201400757f, 0.372558594f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.110854723f, 0.993836641f, 0.f),
    C_Vertex3f(0.201400757f, 0.372558594f, 0.f),
    C_Vertex3f(0.201400757f, 0.372558594f, -1.f),
    C_Normal3f(0.0851281956f, 0.996369958f, 0.f),
    C_Vertex3f(0.20533371f, 0.372119904f, 0.f),
    C_Vertex3f(0.20533371f, 0.372119904f, -1.f),
    C_Normal3f(0.0205883458f, 0.999787986f, 0.f),
    C_Vertex3f(0.211669922f, 0.371681213f, 0.f),
    C_Vertex3f(0.211669922f, 0.371681213f, -1.f),
    C_Normal3f(-0.084013395f, 0.99646467f, 0.f),
    C_Vertex3f(0.226636887f, 0.371681213f, 0.f),
    C_Vertex3f(0.226636887f, 0.371681213f, -1.f),
    C_Normal3f(-0.206760541f, 0.978391588f, 0.f),
    C_Vertex3f(0.250507355f, 0.374955654f, 0.f),
    C_Vertex3f(0.250507355f, 0.374955654f, -1.f),
    C_Normal3f(-0.348944038f, 0.937143624f, 0.f),
    C_Vertex3f(0.27312088f, 0.381504536f, 0.f),
    C_Vertex3f(0.27312088f, 0.381504536f, -1.f),
    C_Normal3f(-0.474657565f, 0.880170584f, 0.f),
    C_Vertex3f(0.294477463f, 0.391327858f, 0.f),
    C_Vertex3f(0.294477463f, 0.391327858f, -1.f),
    C_Normal3f(-0.585953832f, 0.810344398f, 0.f),
    C_Vertex3f(0.309552193f, 0.40115118f, 0.f),
    C_Vertex3f(0.309552193f, 0.40115118f, -1.f),
    C_Normal3f(-0.694374084f, 0.719614208f, 0.f),
    C_Vertex3f(0.318345785f, 0.40858686f, 0.f),
    C_Vertex3f(0.318345785f, 0.40858686f, -1.f),
    C_Normal3f(-0.785369575f, 0.619027197f, 0.f),
    C_Vertex3f(0.32588315f, 0.416909337f, 0.f),
    C_Vertex3f(0.32588315f, 0.416909337f, -1.f),
    C_Normal3f(-0.862910032f, 0.505357444f, 0.f),
    C_Vertex3f(0.332164288f, 0.426118612f, 0.f),
    C_Vertex3f(0.332164288f, 0.426118612f, -1.f),
    C_Normal3f(-0.922910511f, 0.385014504f, 0.f),
    C_Vertex3f(0.337189198f, 0.436214685f, 0.f),
    C_Vertex3f(0.337189198f, 0.436214685f, -1.f),
    C_Normal3f(-0.964241147f, 0.265026599f, 0.f),
    C_Vertex3f(0.34095788f, 0.447197556f, 0.f),
    C_Vertex3f(0.34095788f, 0.447197556f, -1.f),
    C_Normal3f(-0.988491774f, 0.151274711f, 0.f),
    C_Vertex3f(0.343470335f, 0.459067225f, 0.f),
    C_Vertex3f(0.343470335f, 0.459067225f, -1.f),
    C_Normal3f(-0.998737872f, 0.0502264425f, 0.f),
    C_Vertex3f(0.344726563f, 0.471823692f, 0.f),
    C_Vertex3f(0.344726563f, 0.471823692f, -1.f),
    C_Normal3f(-0.999178648f, -0.040522296f, 0.f),
    C_Vertex3f(0.344726563f, 0.484046936f, 0.f),
    C_Vertex3f(0.344726563f, 0.484046936f, -1.f),
    C_Normal3f(-0.990905643f, -0.134558558f, 0.f),
    C_Vertex3f(0.343811512f, 0.494386554f, 0.f),
    C_Vertex3f(0.343811512f, 0.494386554f, -1.f),
    C_Normal3f(-0.973001659f, -0.230797961f, 0.f),
    C_Vertex3f(0.341981411f, 0.504262567f, 0.f),
    C_Vertex3f(0.341981411f, 0.504262567f, -1.f),
    C_Normal3f(-0.944195628f, -0.329384953f, 0.f),
    C_Vertex3f(0.339236259f, 0.513674974f, 0.f),
    C_Vertex3f(0.339236259f, 0.513674974f, -1.f),
    C_Normal3f(-0.904193938f, -0.427122146f, 0.f),
    C_Vertex3f(0.335576057f, 0.522623777f, 0.f),
    C_Vertex3f(0.335576057f, 0.522623777f, -1.f),
    C_Normal3f(-0.853786111f, -0.520623982f, 0.f),
    C_Vertex3f(0.331000805f, 0.531108975f, 0.f),
    C_Vertex3f(0.331000805f, 0.531108975f, -1.f),
    C_Normal3f(-0.794804215f, -0.606866002f, 0.f),
    C_Vertex3f(0.325510502f, 0.539130569f, 0.f),
    C_Vertex3f(0.325510502f, 0.539130569f, -1.f),
    C_Normal3f(-0.729806602f, -0.683653712f, 0.f),
    C_Vertex3f(0.319105148f, 0.546688557f, 0.f),
    C_Vertex3f(0.319105148f, 0.546688557f, -1.f),
    C_Normal3f(-0.659211993f, -0.751957178f, 0.f),
    C_Vertex3f(0.311784744f, 0.55378294f, 0.f),
    C_Vertex3f(0.311784744f, 0.55378294f, -1.f),
    C_Normal3f(-0.578885257f, -0.815408945f, 0.f),
    C_Vertex3f(0.303931713f, 0.559990525f, 0.f),
    C_Vertex3f(0.303931713f, 0.559990525f, -1.f),
    C_Normal3f(-0.491075903f, -0.871116757f, 0.f),
    C_Vertex3f(0.295546055f, 0.565311313f, 0.f),
    C_Vertex3f(0.295546055f, 0.565311313f, -1.f),
    C_Normal3f(-0.398498207f, -0.917169094f, 0.f),
    C_Vertex3f(0.286627769f, 0.569745302f, 0.f),
    C_Vertex3f(0.286627769f, 0.569745302f, -1.f),
    C_Normal3f(-0.304267049f, -0.95258671f, 0.f),
    C_Vertex3f(0.277176857f, 0.573292494f, 0.f),
    C_Vertex3f(0.277176857f, 0.573292494f, -1.f),
    C_Normal3f(-0.211406603f, -0.977398157f, 0.f),
    C_Vertex3f(0.267193317f, 0.575952888f, 0.f),
    C_Vertex3f(0.267193317f, 0.575952888f, -1.f),
    C_Normal3f(-0.122438304f, -0.992476106f, 0.f),
    C_Vertex3f(0.256677151f, 0.577726483f, 0.f),
    C_Vertex3f(0.256677151f, 0.577726483f, -1.f),
    C_Normal3f(-0.0392505415f, -0.999229431f, 0.f),
    C_Vertex3f(0.245628357f, 0.578613281f, 0.f),
    C_Vertex3f(0.245628357f, 0.578613281f, -1.f),
    C_Normal3f(0.0401106067f, -0.999195218f, 0.f),
    C_Vertex3f(0.234101295f, 0.578613281f, 0.f),
    C_Vertex3f(0.234101295f, 0.578613281f, -1.f),
    C_Normal3f(0.125096425f, -0.992144644f, 0.f),
    C_Vertex3f(0.223106742f, 0.577709198f, 0.f),
    C_Vertex3f(0.223106742f, 0.577709198f, -1.f),
    C_Normal3f(0.214788839f, -0.97666049f, 0.f),
    C_Vertex3f(0.212590337f, 0.575901031f, 0.f),
    C_Vertex3f(0.212590337f, 0.575901031f, -1.f),
    C_Normal3f(0.307289928f, -0.95161593f, 0.f),
    C_Vertex3f(0.20255208f, 0.573188782f, 0.f),
    C_Vertex3f(0.20255208f, 0.573188782f, -1.f),
    C_Normal3f(0.400028229f, -0.916502774f, 0.f),
    C_Vertex3f(0.192991972f, 0.569572449f, 0.f),
    C_Vertex3f(0.192991972f, 0.569572449f, -1.f),
    C_Normal3f(0.490136564f, -0.871645689f, 0.f),
    C_Vertex3f(0.183910012f, 0.565052032f, 0.f),
    C_Vertex3f(0.183910012f, 0.565052032f, -1.f),
    C_Normal3f(0.57485497f, -0.818255305f, 0.f),
    C_Vertex3f(0.175306201f, 0.559627533f, 0.f),
    C_Vertex3f(0.175306201f, 0.559627533f, -1.f),
    C_Normal3f(0.651937425f, -0.758272707f, 0.f),
    C_Vertex3f(0.167180538f, 0.55329895f, 0.f),
    C_Vertex3f(0.167180538f, 0.55329895f, -1.f),
    C_Normal3f(0.724731624f, -0.689031303f, 0.f),
    C_Vertex3f(0.159533024f, 0.546066284f, 0.f),
    C_Vertex3f(0.159533024f, 0.546066284f, -1.f),
    C_Normal3f(0.791880012f, -0.610676765f, 0.f),
    C_Vertex3f(0.152568221f, 0.537929535f, 0.f),
    C_Vertex3f(0.152568221f, 0.537929535f, -1.f),
    C_Normal3f(0.847689569f, -0.530492604f, 0.f),
    C_Vertex3f(0.14628613f, 0.528888702f, 0.f),
    C_Vertex3f(0.14628613f, 0.528888702f, -1.f),
    C_Normal3f(0.892373502f, -0.451297522f, 0.f),
    C_Vertex3f(0.14068675f, 0.518943787f, 0.f),
    C_Vertex3f(0.14068675f, 0.518943787f, -1.f),
    C_Normal3f(0.926915944f, -0.375268966f, 0.f),
    C_Vertex3f(0.135770082f, 0.508094788f, 0.f),
    C_Vertex3f(0.135770082f, 0.508094788f, -1.f),
    C_Normal3f(0.952719212f, -0.30385232f, 0.f),
    C_Vertex3f(0.131536126f, 0.496341705f, 0.f),
    C_Vertex3f(0.131536126f, 0.496341705f, -1.f),
    C_Normal3f(0.971306443f, -0.237831756f, 0.f),
    C_Vertex3f(0.127984881f, 0.48368454f, 0.f),
    C_Vertex3f(0.127984881f, 0.48368454f, -1.f),
    C_Normal3f(0.982332468f, -0.187143937f, 0.f),
    C_Vertex3f(0.125116348f, 0.470123291f, 0.f),
    C_Vertex3f(0.125116348f, 0.470123291f, -1.f),
    C_Normal3f(0.988775074f, -0.149411425f, 0.f),
    C_Vertex3f(0.124023438f, 0.462890625f, 0.f),
    C_Vertex3f(0.124023438f, 0.462890625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.177271411f, -0.984161973f, 0.f),
    C_Vertex3f(0.124023438f, 0.462890625f, 0.f),
    C_Vertex3f(0.124023438f, 0.462890625f, -1.f),
    C_Vertex3f(0.0454101563f, 0.477050781f, -1.f),
    C_Vertex3f(0.0454101563f, 0.477050781f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.983761907f, 0.179478407f, 0.f),
    C_Vertex3f(0.0454101563f, 0.477050781f, 0.f),
    C_Vertex3f(0.0454101563f, 0.477050781f, -1.f),
    C_Normal3f(-0.974604726f, 0.223932147f, 0.f),
    C_Vertex3f(0.0472135544f, 0.486935616f, 0.f),
    C_Vertex3f(0.0472135544f, 0.486935616f, -1.f),
    C_Normal3f(-0.95929718f, 0.282398403f, 0.f),
    C_Vertex3f(0.0519680977f, 0.505592465f, 0.f),
    C_Vertex3f(0.0519680977f, 0.505592465f, -1.f),
    C_Normal3f(-0.934401929f, 0.356220573f, 0.f),
    C_Vertex3f(0.057870388f, 0.523136497f, 0.f),
    C_Vertex3f(0.057870388f, 0.523136497f, -1.f),
    C_Normal3f(-0.901434362f, 0.432915926f, 0.f),
    C_Vertex3f(0.0649204254f, 0.539567709f, 0.f),
    C_Vertex3f(0.0649204254f, 0.539567709f, -1.f),
    C_Normal3f(-0.85968262f, 0.510828495f, 0.f),
    C_Vertex3f(0.0731182098f, 0.554886103f, 0.f),
    C_Vertex3f(0.0731182098f, 0.554886103f, -1.f),
    C_Normal3f(-0.808940768f, 0.587890208f, 0.f),
    C_Vertex3f(0.0824637413f, 0.569091678f, 0.f),
    C_Vertex3f(0.0824637413f, 0.569091678f, -1.f),
    C_Normal3f(-0.749669671f, 0.661812246f, 0.f),
    C_Vertex3f(0.0929570198f, 0.582184434f, 0.f),
    C_Vertex3f(0.0929570198f, 0.582184434f, -1.f),
    C_Normal3f(-0.683051467f, 0.730370224f, 0.f),
    C_Vertex3f(0.104598045f, 0.594164371f, 0.f),
    C_Vertex3f(0.104598045f, 0.594164371f, -1.f),
    C_Normal3f(-0.611629307f, 0.791144371f, 0.f),
    C_Vertex3f(0.117386818f, 0.60503149f, 0.f),
    C_Vertex3f(0.117386818f, 0.60503149f, -1.f),
    C_Normal3f(-0.53463006f, 0.845086157f, 0.f),
    C_Vertex3f(0.130954266f, 0.614540219f, 0.f),
    C_Vertex3f(0.130954266f, 0.614540219f, -1.f),
    C_Normal3f(-0.452214152f, 0.89190942f, 0.f),
    C_Vertex3f(0.145300388f, 0.622690558f, 0.f),
    C_Vertex3f(0.145300388f, 0.622690558f, -1.f),
    C_Normal3f(-0.36658293f, 0.930385351f, 0.f),
    C_Vertex3f(0.160425186f, 0.629482508f, 0.f),
    C_Vertex3f(0.160425186f, 0.629482508f, -1.f),
    C_Normal3f(-0.280124605f, 0.959963679f, 0.f),
    C_Vertex3f(0.176328659f, 0.634916067f, 0.f),
    C_Vertex3f(0.176328659f, 0.634916067f, -1.f),
    C_Normal3f(-0.195103869f, 0.980782568f, 0.f),
    C_Vertex3f(0.193010807f, 0.638991237f, 0.f),
    C_Vertex3f(0.193010807f, 0.638991237f, -1.f),
    C_Normal3f(-0.113412887f, 0.993547976f, 0.f),
    C_Vertex3f(0.21047163f, 0.641708016f, 0.f),
    C_Vertex3f(0.21047163f, 0.641708016f, -1.f),
    C_Normal3f(-0.0332222208f, 0.999448001f, 0.f),
    C_Vertex3f(0.228711128f, 0.643066406f, 0.f),
    C_Vertex3f(0.228711128f, 0.643066406f, -1.f),
    C_Normal3f(0.0586304106f, 0.99827981f, 0.f),
    C_Vertex3f(0.251337051f, 0.643066406f, 0.f),
    C_Vertex3f(0.251337051f, 0.643066406f, -1.f),
    C_Normal3f(0.168297589f, 0.985736191f, 0.f),
    C_Vertex3f(0.276531696f, 0.640257835f, 0.f),
    C_Vertex3f(0.276531696f, 0.640257835f, -1.f),
    C_Normal3f(0.284764469f, 0.958597481f, 0.f),
    C_Vertex3f(0.300687313f, 0.634640694f, 0.f),
    C_Vertex3f(0.300687313f, 0.634640694f, -1.f),
    C_Normal3f(0.398902982f, 0.916993141f, 0.f),
    C_Vertex3f(0.323803902f, 0.626214981f, 0.f),
    C_Vertex3f(0.323803902f, 0.626214981f, -1.f),
    C_Normal3f(0.508041561f, 0.861332476f, 0.f),
    C_Vertex3f(0.345881462f, 0.614980698f, 0.f),
    C_Vertex3f(0.345881462f, 0.614980698f, -1.f),
    C_Normal3f(0.618183315f, 0.786033928f, 0.f),
    C_Vertex3f(0.365363598f, 0.601701736f, 0.f),
    C_Vertex3f(0.365363598f, 0.601701736f, -1.f),
    C_Normal3f(0.723664582f, 0.690151811f, 0.f),
    C_Vertex3f(0.382250309f, 0.586378098f, 0.f),
    C_Vertex3f(0.382250309f, 0.586378098f, -1.f),
    C_Normal3f(0.816718042f, 0.577036917f, 0.f),
    C_Vertex3f(0.396541595f, 0.569009781f, 0.f),
    C_Vertex3f(0.396541595f, 0.569009781f, -1.f),
    C_Normal3f(0.887526929f, 0.460755736f, 0.f),
    C_Vertex3f(0.408237457f, 0.549596786f, 0.f),
    C_Vertex3f(0.408237457f, 0.549596786f, -1.f),
    C_Normal3f(0.940927088f, 0.338609159f, 0.f),
    C_Vertex3f(0.417009354f, 0.529583931f, 0.f),
    C_Vertex3f(0.417009354f, 0.529583931f, -1.f),
    C_Normal3f(0.978707194f, 0.205261379f, 0.f),
    C_Vertex3f(0.422857285f, 0.508971214f, 0.f),
    C_Vertex3f(0.422857285f, 0.508971214f, -1.f),
    C_Normal3f(0.997639477f, 0.0686698854f, 0.f),
    C_Vertex3f(0.42578125f, 0.487758636f, 0.f),
    C_Vertex3f(0.42578125f, 0.487758636f, -1.f),
    C_Normal3f(0.997705936f, -0.0676970482f, 0.f),
    C_Vertex3f(0.42578125f, 0.466491699f, 0.f),
    C_Vertex3f(0.42578125f, 0.466491699f, -1.f),
    C_Normal3f(0.97742027f, -0.211304471f, 0.f),
    C_Vertex3f(0.42299509f, 0.446696758f, 0.f),
    C_Vertex3f(0.42299509f, 0.446696758f, -1.f),
    C_Normal3f(0.935264945f, -0.353948534f, 0.f),
    C_Vertex3f(0.417422771f, 0.427828312f, 0.f),
    C_Vertex3f(0.417422771f, 0.427828312f, -1.f),
    C_Normal3f(0.873282731f, -0.48721388f, 0.f),
    C_Vertex3f(0.409064293f, 0.40988636f, 0.f),
    C_Vertex3f(0.409064293f, 0.40988636f, -1.f),
    C_Normal3f(0.791070402f, -0.611725152f, 0.f),
    C_Vertex3f(0.397919655f, 0.392870903f, 0.f),
    C_Vertex3f(0.397919655f, 0.392870903f, -1.f),
    C_Normal3f(0.688636422f, -0.725106716f, 0.f),
    C_Vertex3f(0.38409853f, 0.377601147f, 0.f),
    C_Vertex3f(0.38409853f, 0.377601147f, -1.f),
    C_Normal3f(0.578549325f, -0.815647423f, 0.f),
    C_Vertex3f(0.367600918f, 0.364077091f, 0.f),
    C_Vertex3f(0.367600918f, 0.364077091f, -1.f),
    C_Normal3f(0.487255991f, -0.873259127f, 0.f),
    C_Vertex3f(0.348426819f, 0.352298737f, 0.f),
    C_Vertex3f(0.348426819f, 0.352298737f, -1.f),
    C_Normal3f(0.417266458f, -0.908784151f, 0.f),
    C_Vertex3f(0.337501526f, 0.34728241f, 0.f),
    C_Vertex3f(0.337501526f, 0.34728241f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.224668384f, 0.974435329f, 0.f),
    C_Vertex3f(0.337501526f, 0.34728241f, 0.f),
    C_Vertex3f(0.337501526f, 0.34728241f, -1.f),
    C_Normal3f(0.271781117f, 0.962359071f, 0.f),
    C_Vertex3f(0.344603539f, 0.345644951f, 0.f),
    C_Vertex3f(0.344603539f, 0.345644951f, -1.f),
    C_Normal3f(0.333009481f, 0.942923486f, 0.f),
    C_Vertex3f(0.358011961f, 0.34149003f, 0.f),
    C_Vertex3f(0.358011961f, 0.34149003f, -1.f),
    C_Normal3f(0.409007847f, 0.912530899f, 0.f),
    C_Vertex3f(0.370624781f, 0.336455107f, 0.f),
    C_Vertex3f(0.370624781f, 0.336455107f, -1.f),
    C_Normal3f(0.486274809f, 0.87380594f, 0.f),
    C_Vertex3f(0.382441998f, 0.33054018f, 0.f),
    C_Vertex3f(0.382441998f, 0.33054018f, -1.f),
    C_Normal3f(0.562881172f, 0.826537848f, 0.f),
    C_Vertex3f(0.393463612f, 0.323745251f, 0.f),
    C_Vertex3f(0.393463612f, 0.323745251f, -1.f),
    C_Normal3f(0.636693895f, 0.771116674f, 0.f),
    C_Vertex3f(0.403689623f, 0.316070318f, 0.f),
    C_Vertex3f(0.403689623f, 0.316070318f, -1.f),
    C_Normal3f(0.70562768f, 0.708582759f, 0.f),
    C_Vertex3f(0.413120031f, 0.307515383f, 0.f),
    C_Vertex3f(0.413120031f, 0.307515383f, -1.f),
    C_Normal3f(0.767919481f, 0.640546322f, 0.f),
    C_Vertex3f(0.421754837f, 0.298080444f, 0.f),
    C_Vertex3f(0.421754837f, 0.298080444f, -1.f),
    C_Normal3f(0.822777271f, 0.568364024f, 0.f),
    C_Vertex3f(0.42959404f, 0.287765503f, 0.f),
    C_Vertex3f(0.42959404f, 0.287765503f, -1.f),
    C_Normal3f(0.870901167f, 0.491458237f, 0.f),
    C_Vertex3f(0.436453342f, 0.27680254f, 0.f),
    C_Vertex3f(0.436453342f, 0.27680254f, -1.f),
    C_Normal3f(0.911386251f, 0.411552072f, 0.f),
    C_Vertex3f(0.442332745f, 0.265191555f, 0.f),
    C_Vertex3f(0.442332745f, 0.265191555f, -1.f),
    C_Normal3f(0.943729758f, 0.330717593f, 0.f),
    C_Vertex3f(0.447232246f, 0.252932549f, 0.f),
    C_Vertex3f(0.447232246f, 0.252932549f, -1.f),
    C_Normal3f(0.968007922f, 0.250919789f, 0.f),
    C_Vertex3f(0.451151848f, 0.24002552f, 0.f),
    C_Vertex3f(0.451151848f, 0.24002552f, -1.f),
    C_Normal3f(0.984777331f, 0.173820063f, 0.f),
    C_Vertex3f(0.454091549f, 0.22647047f, 0.f),
    C_Vertex3f(0.454091549f, 0.22647047f, -1.f),
    C_Normal3f(0.994920254f, 0.100666054f, 0.f),
    C_Vertex3f(0.45605135f, 0.212267399f, 0.f),
    C_Vertex3f(0.45605135f, 0.212267399f, -1.f),
    C_Normal3f(0.999561548f, 0.0296094883f, 0.f),
    C_Vertex3f(0.45703125f, 0.197416306f, 0.f),
    C_Vertex3f(0.45703125f, 0.197416306f, -1.f),
    C_Normal3f(0.998762608f, -0.0497328117f, 0.f),
    C_Vertex3f(0.45703125f, 0.179187775f, 0.f),
    C_Vertex3f(0.45703125f, 0.179187775f, -1.f),
    C_Normal3f(0.989636242f, -0.143597081f, 0.f),
    C_Vertex3f(0.455119729f, 0.159028053f, 0.f),
    C_Vertex3f(0.455119729f, 0.159028053f, -1.f),
    C_Normal3f(0.969681084f, -0.244374037f, 0.f),
    C_Vertex3f(0.451296687f, 0.139666557f, 0.f),
    C_Vertex3f(0.451296687f, 0.139666557f, -1.f),
    C_Normal3f(0.938372254f, -0.345626473f, 0.f),
    C_Vertex3f(0.445562124f, 0.121103287f, 0.f),
    C_Vertex3f(0.445562124f, 0.121103287f, -1.f),
    C_Normal3f(0.896095276f, -0.443861693f, 0.f),
    C_Vertex3f(0.43791604f, 0.103338242f, 0.f),
    C_Vertex3f(0.43791604f, 0.103338242f, -1.f),
    C_Normal3f(0.844346881f, -0.535797f, 0.f),
    C_Vertex3f(0.428358436f, 0.0863714218f, 0.f),
    C_Vertex3f(0.428358436f, 0.0863714218f, -1.f),
    C_Normal3f(0.785480857f, -0.618885934f, 0.f),
    C_Vertex3f(0.41688931f, 0.0702028275f, 0.f),
    C_Vertex3f(0.41688931f, 0.0702028275f, -1.f),
    C_Normal3f(0.722254336f, -0.691627443f, 0.f),
    C_Vertex3f(0.403508663f, 0.0548324585f, 0.f),
    C_Vertex3f(0.403508663f, 0.0548324585f, -1.f),
    C_Normal3f(0.654120147f, -0.756390631f, 0.f),
    C_Vertex3f(0.388216496f, 0.0402603149f, 0.f),
    C_Vertex3f(0.388216496f, 0.0402603149f, -1.f),
    C_Normal3f(0.575892985f, -0.817525089f, 0.f),
    C_Vertex3f(0.371914029f, 0.0275096893f, 0.f),
    C_Vertex3f(0.371914029f, 0.0275096893f, -1.f),
    C_Normal3f(0.490104139f, -0.871663988f, 0.f),
    C_Vertex3f(0.354601264f, 0.0165805817f, 0.f),
    C_Vertex3f(0.354601264f, 0.0165805817f, -1.f),
    C_Normal3f(0.399162263f, -0.91688025f, 0.f),
    C_Vertex3f(0.3362782f, 0.00747299194f, 0.f),
    C_Vertex3f(0.3362782f, 0.00747299194f, -1.f),
    C_Normal3f(0.305950403f, -0.952047527f, 0.f),
    C_Vertex3f(0.316944838f, 0.000186920166f, 0.f),
    C_Vertex3f(0.316944838f, 0.000186920166f, -1.f),
    C_Normal3f(0.213389069f, -0.976967275f, 0.f),
    C_Vertex3f(0.296601176f, -0.00527763367f, 0.f),
    C_Vertex3f(0.296601176f, -0.00527763367f, -1.f),
    C_Normal3f(0.124029741f, -0.992278516f, 0.f),
    C_Vertex3f(0.275247216f, -0.00892066956f, 0.f),
    C_Vertex3f(0.275247216f, -0.00892066956f, -1.f),
    C_Normal3f(0.0408149101f, -0.999166667f, 0.f),
    C_Vertex3f(0.252882957f, -0.0107421875f, 0.f),
    C_Vertex3f(0.252882957f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0370388776f, -0.999313831f, 0.f),
    C_Vertex3f(0.23065567f, -0.0107421875f, 0.f),
    C_Vertex3f(0.23065567f, -0.0107421875f, -1.f),
    C_Normal3f(-0.118618131f, -0.992940009f, 0.f),
    C_Vertex3f(0.210469842f, -0.0091701746f, 0.f),
    C_Vertex3f(0.210469842f, -0.0091701746f, -1.f),
    C_Normal3f(-0.20415765f, -0.978938043f, 0.f),
    C_Vertex3f(0.191178203f, -0.0060261488f, 0.f),
    C_Vertex3f(0.191178203f, -0.0060261488f, -1.f),
    C_Normal3f(-0.293056995f, -0.95609498f, 0.f),
    C_Vertex3f(0.172780752f, -0.00131011009f, 0.f),
    C_Vertex3f(0.172780752f, -0.00131011009f, -1.f),
    C_Normal3f(-0.383106947f, -0.923704028f, 0.f),
    C_Vertex3f(0.155277491f, 0.00497794151f, 0.f),
    C_Vertex3f(0.155277491f, 0.00497794151f, -1.f),
    C_Normal3f(-0.471707076f, -0.881755412f, 0.f),
    C_Vertex3f(0.138668418f, 0.012838006f, 0.f),
    C_Vertex3f(0.138668418f, 0.012838006f, -1.f),
    C_Normal3f(-0.556190014f, -0.831055164f, 0.f),
    C_Vertex3f(0.122953534f, 0.0222700834f, 0.f),
    C_Vertex3f(0.122953534f, 0.0222700834f, -1.f),
    C_Normal3f(-0.634200037f, -0.773169041f, 0.f),
    C_Vertex3f(0.108132839f, 0.0332741737f, 0.f),
    C_Vertex3f(0.108132839f, 0.0332741737f, -1.f),
    C_Normal3f(-0.703374207f, -0.710819721f, 0.f),
    C_Vertex3f(0.0942063332f, 0.0458502769f, 0.f),
    C_Vertex3f(0.0942063332f, 0.0458502769f, -1.f),
    C_Normal3f(-0.767092049f, -0.641537011f, 0.f),
    C_Vertex3f(0.0817748308f, 0.0593560934f, 0.f),
    C_Vertex3f(0.0817748308f, 0.0593560934f, -1.f),
    C_Normal3f(-0.825462222f, -0.564457297f, 0.f),
    C_Vertex3f(0.0708383322f, 0.0737916231f, 0.f),
    C_Vertex3f(0.0708383322f, 0.0737916231f, -1.f),
    C_Normal3f(-0.876509428f, -0.481384665f, 0.f),
    C_Vertex3f(0.0613968372f, 0.0891568661f, 0.f),
    C_Vertex3f(0.0613968372f, 0.0891568661f, -1.f),
    C_Normal3f(-0.918823004f, -0.394670099f, 0.f),
    C_Vertex3f(0.053450346f, 0.105451822f, 0.f),
    C_Vertex3f(0.053450346f, 0.105451822f, -1.f),
    C_Normal3f(-0.951745033f, -0.30689007f, 0.f),
    C_Vertex3f(0.0469988585f, 0.122676492f, 0.f),
    C_Vertex3f(0.0469988585f, 0.122676492f, -1.f),
    C_Normal3f(-0.975388825f, -0.220492035f, 0.f),
    C_Vertex3f(0.0420423746f, 0.140830874f, 0.f),
    C_Vertex3f(0.0420423746f, 0.140830874f, -1.f),
    C_Normal3f(-0.988528609f, -0.151034042f, 0.f),
    C_Vertex3f(0.0385808945f, 0.15991497f, 0.f),
    C_Vertex3f(0.0385808945f, 0.15991497f, -1.f),
    C_Normal3f(-0.995207548f, -0.0977850929f, 0.f),
    C_Vertex3f(0.0375976563f, 0.169921875f, 0.f),
    C_Vertex3f(0.0375976563f, 0.169921875f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.421754837f, 0.298080444f, 0.f),
    C_Vertex3f(0.413120031f, 0.307515383f, 0.f),
    C_Vertex3f(0.41688931f, 0.0702028275f, 0.f),
    C_Vertex3f(0.403689623f, 0.316070318f, 0.f),
    C_Vertex3f(0.403508663f, 0.0548324585f, 0.f),
    C_Vertex3f(0.393463612f, 0.323745251f, 0.f),
    C_Vertex3f(0.388216496f, 0.0402603149f, 0.f),
    C_Vertex3f(0.382441998f, 0.33054018f, 0.f),
    C_Vertex3f(0.373535156f, 0.194684982f, 0.f),
    C_Vertex3f(0.372429252f, 0.207637072f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.370624781f, 0.336455107f, 0.f),
    C_Vertex3f(0.358011961f, 0.34149003f, 0.f),
    C_Vertex3f(0.36247611f, 0.242929935f, 0.f),
    C_Vertex3f(0.366899729f, 0.231759548f, 0.f),
    C_Vertex3f(0.370217443f, 0.21999526f, 0.f),
    C_Vertex3f(0.372429252f, 0.207637072f, 0.f),
    C_Vertex3f(0.382441998f, 0.33054018f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.337189198f, 0.436214685f, 0.f),
    C_Vertex3f(0.333722591f, 0.281672478f, 0.f),
    C_Vertex3f(0.337501526f, 0.34728241f, 0.f),
    C_Vertex3f(0.342569828f, 0.272877693f, 0.f),
    C_Vertex3f(0.344603539f, 0.345644951f, 0.f),
    C_Vertex3f(0.35031116f, 0.263489008f, 0.f),
    C_Vertex3f(0.358011961f, 0.34149003f, 0.f),
    C_Vertex3f(0.356946588f, 0.253506422f, 0.f),
    C_Vertex3f(0.36247611f, 0.242929935f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.280475974f, 0.309156179f, 0.f),
    C_Vertex3f(0.292326689f, 0.305858135f, 0.f),
    C_Vertex3f(0.294477463f, 0.391327858f, 0.f),
    C_Vertex3f(0.303576708f, 0.301460743f, 0.f),
    C_Vertex3f(0.309552193f, 0.40115118f, 0.f),
    C_Vertex3f(0.314226031f, 0.295964003f, 0.f),
    C_Vertex3f(0.318345785f, 0.40858686f, 0.f),
    C_Vertex3f(0.324274659f, 0.289367914f, 0.f),
    C_Vertex3f(0.32588315f, 0.416909337f, 0.f),
    C_Vertex3f(0.333722591f, 0.281672478f, 0.f),
    C_Vertex3f(0.332164288f, 0.426118612f, 0.f),
    C_Vertex3f(0.337189198f, 0.436214685f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.27312088f, 0.381504536f, 0.f),
    C_Vertex3f(0.250507355f, 0.374955654f, 0.f),
    C_Vertex3f(0.254972458f, 0.312454224f, 0.f),
    C_Vertex3f(0.268024564f, 0.311354876f, 0.f),
    C_Vertex3f(0.280475974f, 0.309156179f, 0.f),
    C_Vertex3f(0.294477463f, 0.391327858f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.211669922f, 0.371681213f, 0.f),
    C_Vertex3f(0.216201782f, 0.309175491f, 0.f),
    C_Vertex3f(0.226636887f, 0.371681213f, 0.f),
    C_Vertex3f(0.230072021f, 0.311361313f, 0.f),
    C_Vertex3f(0.250507355f, 0.374955654f, 0.f),
    C_Vertex3f(0.242576599f, 0.312454224f, 0.f),
    C_Vertex3f(0.254972458f, 0.312454224f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.1926651f, 0.303710938f, 0.f),
    C_Vertex3f(0.200965881f, 0.305896759f, 0.f),
    C_Vertex3f(0.201400757f, 0.372558594f, 0.f),
    C_Vertex3f(0.216201782f, 0.309175491f, 0.f),
    C_Vertex3f(0.20533371f, 0.372119904f, 0.f),
    C_Vertex3f(0.211669922f, 0.371681213f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.451151848f, 0.24002552f, 0.f),
    C_Vertex3f(0.451296687f, 0.139666557f, 0.f),
    C_Vertex3f(0.454091549f, 0.22647047f, 0.f),
    C_Vertex3f(0.455119729f, 0.159028053f, 0.f),
    C_Vertex3f(0.45605135f, 0.212267399f, 0.f),
    C_Vertex3f(0.45703125f, 0.179187775f, 0.f),
    C_Vertex3f(0.45703125f, 0.197416306f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.436453342f, 0.27680254f, 0.f),
    C_Vertex3f(0.43791604f, 0.103338242f, 0.f),
    C_Vertex3f(0.442332745f, 0.265191555f, 0.f),
    C_Vertex3f(0.445562124f, 0.121103287f, 0.f),
    C_Vertex3f(0.447232246f, 0.252932549f, 0.f),
    C_Vertex3f(0.451296687f, 0.139666557f, 0.f),
    C_Vertex3f(0.451151848f, 0.24002552f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.42959404f, 0.287765503f, 0.f),
    C_Vertex3f(0.421754837f, 0.298080444f, 0.f),
    C_Vertex3f(0.428358436f, 0.0863714218f, 0.f),
    C_Vertex3f(0.43791604f, 0.103338242f, 0.f),
    C_Vertex3f(0.436453342f, 0.27680254f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.373535156f, 0.194684982f, 0.f),
    C_Vertex3f(0.373535156f, 0.180810928f, 0.f),
    C_Vertex3f(0.388216496f, 0.0402603149f, 0.f),
    C_Vertex3f(0.372344375f, 0.167182446f, 0.f),
    C_Vertex3f(0.371914029f, 0.0275096893f, 0.f),
    C_Vertex3f(0.369962811f, 0.154127598f, 0.f),
    C_Vertex3f(0.366390467f, 0.141646385f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.366390467f, 0.141646385f, 0.f),
    C_Vertex3f(0.36162734f, 0.129738808f, 0.f),
    C_Vertex3f(0.371914029f, 0.0275096893f, 0.f),
    C_Vertex3f(0.355673432f, 0.118404865f, 0.f),
    C_Vertex3f(0.354601264f, 0.0165805817f, 0.f),
    C_Vertex3f(0.348528743f, 0.107644558f, 0.f),
    C_Vertex3f(0.340193272f, 0.0974578857f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.3362782f, 0.00747299194f, 0.f),
    C_Vertex3f(0.354601264f, 0.0165805817f, 0.f),
    C_Vertex3f(0.340193272f, 0.0974578857f, 0.f),
    C_Vertex3f(0.330667019f, 0.0878448486f, 0.f),
    C_Vertex3f(0.320590138f, 0.0794334412f, 0.f),
    C_Vertex3f(0.316944838f, 0.000186920166f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320590138f, 0.0794334412f, 0.f),
    C_Vertex3f(0.30996263f, 0.0722236633f, 0.f),
    C_Vertex3f(0.316944838f, 0.000186920166f, 0.f),
    C_Vertex3f(0.298784494f, 0.0662155151f, 0.f),
    C_Vertex3f(0.296601176f, -0.00527763367f, 0.f),
    C_Vertex3f(0.287055731f, 0.0614089966f, 0.f),
    C_Vertex3f(0.275247216f, -0.00892066956f, 0.f),
    C_Vertex3f(0.27477634f, 0.0578041077f, 0.f),
    C_Vertex3f(0.261946321f, 0.0554008484f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.252882957f, -0.0107421875f, 0.f),
    C_Vertex3f(0.275247216f, -0.00892066956f, 0.f),
    C_Vertex3f(0.261946321f, 0.0554008484f, 0.f),
    C_Vertex3f(0.248565674f, 0.0541992188f, 0.f),
    C_Vertex3f(0.235725403f, 0.0541992188f, 0.f),
    C_Vertex3f(0.23065567f, -0.0107421875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.235725403f, 0.0541992188f, 0.f),
    C_Vertex3f(0.224431515f, 0.0551218987f, 0.f),
    C_Vertex3f(0.23065567f, -0.0107421875f, 0.f),
    C_Vertex3f(0.213593006f, 0.0569672585f, 0.f),
    C_Vertex3f(0.210469842f, -0.0091701746f, 0.f),
    C_Vertex3f(0.203209877f, 0.0597352982f, 0.f),
    C_Vertex3f(0.193282127f, 0.0634260178f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191178203f, -0.0060261488f, 0.f),
    C_Vertex3f(0.210469842f, -0.0091701746f, 0.f),
    C_Vertex3f(0.193282127f, 0.0634260178f, 0.f),
    C_Vertex3f(0.183809757f, 0.0680394173f, 0.f),
    C_Vertex3f(0.174792767f, 0.0735754967f, 0.f),
    C_Vertex3f(0.172780752f, -0.00131011009f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.174792767f, 0.0735754967f, 0.f),
    C_Vertex3f(0.166231155f, 0.080034256f, 0.f),
    C_Vertex3f(0.172780752f, -0.00131011009f, 0.f),
    C_Vertex3f(0.158124924f, 0.0874156952f, 0.f),
    C_Vertex3f(0.155277491f, 0.00497794151f, 0.f),
    C_Vertex3f(0.150608778f, 0.0959658623f, 0.f),
    C_Vertex3f(0.143682718f, 0.105684757f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138668418f, 0.012838006f, 0.f),
    C_Vertex3f(0.155277491f, 0.00497794151f, 0.f),
    C_Vertex3f(0.143682718f, 0.105684757f, 0.f),
    C_Vertex3f(0.137346745f, 0.11657238f, 0.f),
    C_Vertex3f(0.131600857f, 0.128628731f, 0.f),
    C_Vertex3f(0.126445055f, 0.141853809f, 0.f),
    C_Vertex3f(0.122953534f, 0.0222700834f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122953534f, 0.0222700834f, 0.f),
    C_Vertex3f(0.126445055f, 0.141853809f, 0.f),
    C_Vertex3f(0.121879339f, 0.156247616f, 0.f),
    C_Vertex3f(0.117903709f, 0.17181015f, 0.f),
    C_Vertex3f(0.116210938f, 0.180175781f, 0.f),
    C_Vertex3f(0.108132839f, 0.0332741737f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.180175781f, 0.f),
    C_Vertex3f(0.0375976563f, 0.169921875f, 0.f),
    C_Vertex3f(0.0385808945f, 0.15991497f, 0.f),
    C_Vertex3f(0.0420423746f, 0.140830874f, 0.f),
    C_Vertex3f(0.0469988585f, 0.122676492f, 0.f),
    C_Vertex3f(0.053450346f, 0.105451822f, 0.f),
    C_Vertex3f(0.0613968372f, 0.0891568661f, 0.f),
    C_Vertex3f(0.0708383322f, 0.0737916231f, 0.f),
    C_Vertex3f(0.0817748308f, 0.0593560934f, 0.f),
    C_Vertex3f(0.0942063332f, 0.0458502769f, 0.f),
    C_Vertex3f(0.108132839f, 0.0332741737f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.42578125f, 0.466491699f, 0.f),
    C_Vertex3f(0.42578125f, 0.487758636f, 0.f),
    C_Vertex3f(0.42299509f, 0.446696758f, 0.f),
    C_Vertex3f(0.422857285f, 0.508971214f, 0.f),
    C_Vertex3f(0.417422771f, 0.427828312f, 0.f),
    C_Vertex3f(0.417009354f, 0.529583931f, 0.f),
    C_Vertex3f(0.409064293f, 0.40988636f, 0.f),
    C_Vertex3f(0.408237457f, 0.549596786f, 0.f),
    C_Vertex3f(0.397919655f, 0.392870903f, 0.f),
    C_Vertex3f(0.396541595f, 0.569009781f, 0.f),
    C_Vertex3f(0.38409853f, 0.377601147f, 0.f),
    C_Vertex3f(0.382250309f, 0.586378098f, 0.f),
    C_Vertex3f(0.367600918f, 0.364077091f, 0.f),
    C_Vertex3f(0.365363598f, 0.601701736f, 0.f),
    C_Vertex3f(0.348426819f, 0.352298737f, 0.f),
    C_Vertex3f(0.345881462f, 0.614980698f, 0.f),
    C_Vertex3f(0.344726563f, 0.484046936f, 0.f),
    C_Vertex3f(0.343811512f, 0.494386554f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.345881462f, 0.614980698f, 0.f),
    C_Vertex3f(0.323803902f, 0.626214981f, 0.f),
    C_Vertex3f(0.325510502f, 0.539130569f, 0.f),
    C_Vertex3f(0.331000805f, 0.531108975f, 0.f),
    C_Vertex3f(0.335576057f, 0.522623777f, 0.f),
    C_Vertex3f(0.339236259f, 0.513674974f, 0.f),
    C_Vertex3f(0.341981411f, 0.504262567f, 0.f),
    C_Vertex3f(0.343811512f, 0.494386554f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.323803902f, 0.626214981f, 0.f),
    C_Vertex3f(0.300687313f, 0.634640694f, 0.f),
    C_Vertex3f(0.303931713f, 0.559990525f, 0.f),
    C_Vertex3f(0.311784744f, 0.55378294f, 0.f),
    C_Vertex3f(0.319105148f, 0.546688557f, 0.f),
    C_Vertex3f(0.325510502f, 0.539130569f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.300687313f, 0.634640694f, 0.f),
    C_Vertex3f(0.276531696f, 0.640257835f, 0.f),
    C_Vertex3f(0.277176857f, 0.573292494f, 0.f),
    C_Vertex3f(0.286627769f, 0.569745302f, 0.f),
    C_Vertex3f(0.295546055f, 0.565311313f, 0.f),
    C_Vertex3f(0.303931713f, 0.559990525f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.234101295f, 0.578613281f, 0.f),
    C_Vertex3f(0.245628357f, 0.578613281f, 0.f),
    C_Vertex3f(0.251337051f, 0.643066406f, 0.f),
    C_Vertex3f(0.256677151f, 0.577726483f, 0.f),
    C_Vertex3f(0.276531696f, 0.640257835f, 0.f),
    C_Vertex3f(0.267193317f, 0.575952888f, 0.f),
    C_Vertex3f(0.277176857f, 0.573292494f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.228711128f, 0.643066406f, 0.f),
    C_Vertex3f(0.21047163f, 0.641708016f, 0.f),
    C_Vertex3f(0.212590337f, 0.575901031f, 0.f),
    C_Vertex3f(0.223106742f, 0.577709198f, 0.f),
    C_Vertex3f(0.234101295f, 0.578613281f, 0.f),
    C_Vertex3f(0.251337051f, 0.643066406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.183910012f, 0.565052032f, 0.f),
    C_Vertex3f(0.192991972f, 0.569572449f, 0.f),
    C_Vertex3f(0.193010807f, 0.638991237f, 0.f),
    C_Vertex3f(0.20255208f, 0.573188782f, 0.f),
    C_Vertex3f(0.21047163f, 0.641708016f, 0.f),
    C_Vertex3f(0.212590337f, 0.575901031f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.176328659f, 0.634916067f, 0.f),
    C_Vertex3f(0.160425186f, 0.629482508f, 0.f),
    C_Vertex3f(0.167180538f, 0.55329895f, 0.f),
    C_Vertex3f(0.175306201f, 0.559627533f, 0.f),
    C_Vertex3f(0.183910012f, 0.565052032f, 0.f),
    C_Vertex3f(0.193010807f, 0.638991237f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.160425186f, 0.629482508f, 0.f),
    C_Vertex3f(0.145300388f, 0.622690558f, 0.f),
    C_Vertex3f(0.14628613f, 0.528888702f, 0.f),
    C_Vertex3f(0.152568221f, 0.537929535f, 0.f),
    C_Vertex3f(0.159533024f, 0.546066284f, 0.f),
    C_Vertex3f(0.167180538f, 0.55329895f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145300388f, 0.622690558f, 0.f),
    C_Vertex3f(0.130954266f, 0.614540219f, 0.f),
    C_Vertex3f(0.131536126f, 0.496341705f, 0.f),
    C_Vertex3f(0.135770082f, 0.508094788f, 0.f),
    C_Vertex3f(0.14068675f, 0.518943787f, 0.f),
    C_Vertex3f(0.14628613f, 0.528888702f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130954266f, 0.614540219f, 0.f),
    C_Vertex3f(0.117386818f, 0.60503149f, 0.f),
    C_Vertex3f(0.124023438f, 0.462890625f, 0.f),
    C_Vertex3f(0.125116348f, 0.470123291f, 0.f),
    C_Vertex3f(0.127984881f, 0.48368454f, 0.f),
    C_Vertex3f(0.131536126f, 0.496341705f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.124023438f, 0.462890625f, 0.f),
    C_Vertex3f(0.117386818f, 0.60503149f, 0.f),
    C_Vertex3f(0.104598045f, 0.594164371f, 0.f),
    C_Vertex3f(0.0929570198f, 0.582184434f, 0.f),
    C_Vertex3f(0.0824637413f, 0.569091678f, 0.f),
    C_Vertex3f(0.0731182098f, 0.554886103f, 0.f),
    C_Vertex3f(0.0649204254f, 0.539567709f, 0.f),
    C_Vertex3f(0.057870388f, 0.523136497f, 0.f),
    C_Vertex3f(0.0519680977f, 0.505592465f, 0.f),
    C_Vertex3f(0.0472135544f, 0.486935616f, 0.f),
    C_Vertex3f(0.0454101563f, 0.477050781f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.348426819f, 0.352298737f, 0.f),
    C_Vertex3f(0.344726563f, 0.484046936f, 0.f),
    C_Vertex3f(0.344726563f, 0.471823692f, 0.f),
    C_Vertex3f(0.343470335f, 0.459067225f, 0.f),
    C_Vertex3f(0.34095788f, 0.447197556f, 0.f),
    C_Vertex3f(0.337501526f, 0.34728241f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.337501526f, 0.34728241f, 0.f),
    C_Vertex3f(0.34095788f, 0.447197556f, 0.f),
    C_Vertex3f(0.337189198f, 0.436214685f, 0.f),
    C_Vertex3f(0.421754837f, 0.298080444f, 0.f),
    C_Vertex3f(0.41688931f, 0.0702028275f, 0.f),
    C_Vertex3f(0.428358436f, 0.0863714218f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.421754837f, 0.298080444f, -1.f),
    C_Vertex3f(0.413120031f, 0.307515383f, -1.f),
    C_Vertex3f(0.41688931f, 0.0702028275f, -1.f),
    C_Vertex3f(0.403689623f, 0.316070318f, -1.f),
    C_Vertex3f(0.403508663f, 0.0548324585f, -1.f),
    C_Vertex3f(0.393463612f, 0.323745251f, -1.f),
    C_Vertex3f(0.388216496f, 0.0402603149f, -1.f),
    C_Vertex3f(0.382441998f, 0.33054018f, -1.f),
    C_Vertex3f(0.373535156f, 0.194684982f, -1.f),
    C_Vertex3f(0.372429252f, 0.207637072f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.370624781f, 0.336455107f, -1.f),
    C_Vertex3f(0.358011961f, 0.34149003f, -1.f),
    C_Vertex3f(0.36247611f, 0.242929935f, -1.f),
    C_Vertex3f(0.366899729f, 0.231759548f, -1.f),
    C_Vertex3f(0.370217443f, 0.21999526f, -1.f),
    C_Vertex3f(0.372429252f, 0.207637072f, -1.f),
    C_Vertex3f(0.382441998f, 0.33054018f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.337189198f, 0.436214685f, -1.f),
    C_Vertex3f(0.333722591f, 0.281672478f, -1.f),
    C_Vertex3f(0.337501526f, 0.34728241f, -1.f),
    C_Vertex3f(0.342569828f, 0.272877693f, -1.f),
    C_Vertex3f(0.344603539f, 0.345644951f, -1.f),
    C_Vertex3f(0.35031116f, 0.263489008f, -1.f),
    C_Vertex3f(0.358011961f, 0.34149003f, -1.f),
    C_Vertex3f(0.356946588f, 0.253506422f, -1.f),
    C_Vertex3f(0.36247611f, 0.242929935f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.280475974f, 0.309156179f, -1.f),
    C_Vertex3f(0.292326689f, 0.305858135f, -1.f),
    C_Vertex3f(0.294477463f, 0.391327858f, -1.f),
    C_Vertex3f(0.303576708f, 0.301460743f, -1.f),
    C_Vertex3f(0.309552193f, 0.40115118f, -1.f),
    C_Vertex3f(0.314226031f, 0.295964003f, -1.f),
    C_Vertex3f(0.318345785f, 0.40858686f, -1.f),
    C_Vertex3f(0.324274659f, 0.289367914f, -1.f),
    C_Vertex3f(0.32588315f, 0.416909337f, -1.f),
    C_Vertex3f(0.333722591f, 0.281672478f, -1.f),
    C_Vertex3f(0.332164288f, 0.426118612f, -1.f),
    C_Vertex3f(0.337189198f, 0.436214685f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.27312088f, 0.381504536f, -1.f),
    C_Vertex3f(0.250507355f, 0.374955654f, -1.f),
    C_Vertex3f(0.254972458f, 0.312454224f, -1.f),
    C_Vertex3f(0.268024564f, 0.311354876f, -1.f),
    C_Vertex3f(0.280475974f, 0.309156179f, -1.f),
    C_Vertex3f(0.294477463f, 0.391327858f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.211669922f, 0.371681213f, -1.f),
    C_Vertex3f(0.216201782f, 0.309175491f, -1.f),
    C_Vertex3f(0.226636887f, 0.371681213f, -1.f),
    C_Vertex3f(0.230072021f, 0.311361313f, -1.f),
    C_Vertex3f(0.250507355f, 0.374955654f, -1.f),
    C_Vertex3f(0.242576599f, 0.312454224f, -1.f),
    C_Vertex3f(0.254972458f, 0.312454224f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.1926651f, 0.303710938f, -1.f),
    C_Vertex3f(0.200965881f, 0.305896759f, -1.f),
    C_Vertex3f(0.201400757f, 0.372558594f, -1.f),
    C_Vertex3f(0.216201782f, 0.309175491f, -1.f),
    C_Vertex3f(0.20533371f, 0.372119904f, -1.f),
    C_Vertex3f(0.211669922f, 0.371681213f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.451151848f, 0.24002552f, -1.f),
    C_Vertex3f(0.451296687f, 0.139666557f, -1.f),
    C_Vertex3f(0.454091549f, 0.22647047f, -1.f),
    C_Vertex3f(0.455119729f, 0.159028053f, -1.f),
    C_Vertex3f(0.45605135f, 0.212267399f, -1.f),
    C_Vertex3f(0.45703125f, 0.179187775f, -1.f),
    C_Vertex3f(0.45703125f, 0.197416306f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.436453342f, 0.27680254f, -1.f),
    C_Vertex3f(0.43791604f, 0.103338242f, -1.f),
    C_Vertex3f(0.442332745f, 0.265191555f, -1.f),
    C_Vertex3f(0.445562124f, 0.121103287f, -1.f),
    C_Vertex3f(0.447232246f, 0.252932549f, -1.f),
    C_Vertex3f(0.451296687f, 0.139666557f, -1.f),
    C_Vertex3f(0.451151848f, 0.24002552f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.42959404f, 0.287765503f, -1.f),
    C_Vertex3f(0.421754837f, 0.298080444f, -1.f),
    C_Vertex3f(0.428358436f, 0.0863714218f, -1.f),
    C_Vertex3f(0.43791604f, 0.103338242f, -1.f),
    C_Vertex3f(0.436453342f, 0.27680254f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.373535156f, 0.194684982f, -1.f),
    C_Vertex3f(0.373535156f, 0.180810928f, -1.f),
    C_Vertex3f(0.388216496f, 0.0402603149f, -1.f),
    C_Vertex3f(0.372344375f, 0.167182446f, -1.f),
    C_Vertex3f(0.371914029f, 0.0275096893f, -1.f),
    C_Vertex3f(0.369962811f, 0.154127598f, -1.f),
    C_Vertex3f(0.366390467f, 0.141646385f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.366390467f, 0.141646385f, -1.f),
    C_Vertex3f(0.36162734f, 0.129738808f, -1.f),
    C_Vertex3f(0.371914029f, 0.0275096893f, -1.f),
    C_Vertex3f(0.355673432f, 0.118404865f, -1.f),
    C_Vertex3f(0.354601264f, 0.0165805817f, -1.f),
    C_Vertex3f(0.348528743f, 0.107644558f, -1.f),
    C_Vertex3f(0.340193272f, 0.0974578857f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.3362782f, 0.00747299194f, -1.f),
    C_Vertex3f(0.354601264f, 0.0165805817f, -1.f),
    C_Vertex3f(0.340193272f, 0.0974578857f, -1.f),
    C_Vertex3f(0.330667019f, 0.0878448486f, -1.f),
    C_Vertex3f(0.320590138f, 0.0794334412f, -1.f),
    C_Vertex3f(0.316944838f, 0.000186920166f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320590138f, 0.0794334412f, -1.f),
    C_Vertex3f(0.30996263f, 0.0722236633f, -1.f),
    C_Vertex3f(0.316944838f, 0.000186920166f, -1.f),
    C_Vertex3f(0.298784494f, 0.0662155151f, -1.f),
    C_Vertex3f(0.296601176f, -0.00527763367f, -1.f),
    C_Vertex3f(0.287055731f, 0.0614089966f, -1.f),
    C_Vertex3f(0.275247216f, -0.00892066956f, -1.f),
    C_Vertex3f(0.27477634f, 0.0578041077f, -1.f),
    C_Vertex3f(0.261946321f, 0.0554008484f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.252882957f, -0.0107421875f, -1.f),
    C_Vertex3f(0.275247216f, -0.00892066956f, -1.f),
    C_Vertex3f(0.261946321f, 0.0554008484f, -1.f),
    C_Vertex3f(0.248565674f, 0.0541992188f, -1.f),
    C_Vertex3f(0.235725403f, 0.0541992188f, -1.f),
    C_Vertex3f(0.23065567f, -0.0107421875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.235725403f, 0.0541992188f, -1.f),
    C_Vertex3f(0.224431515f, 0.0551218987f, -1.f),
    C_Vertex3f(0.23065567f, -0.0107421875f, -1.f),
    C_Vertex3f(0.213593006f, 0.0569672585f, -1.f),
    C_Vertex3f(0.210469842f, -0.0091701746f, -1.f),
    C_Vertex3f(0.203209877f, 0.0597352982f, -1.f),
    C_Vertex3f(0.193282127f, 0.0634260178f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191178203f, -0.0060261488f, -1.f),
    C_Vertex3f(0.210469842f, -0.0091701746f, -1.f),
    C_Vertex3f(0.193282127f, 0.0634260178f, -1.f),
    C_Vertex3f(0.183809757f, 0.0680394173f, -1.f),
    C_Vertex3f(0.174792767f, 0.0735754967f, -1.f),
    C_Vertex3f(0.172780752f, -0.00131011009f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.174792767f, 0.0735754967f, -1.f),
    C_Vertex3f(0.166231155f, 0.080034256f, -1.f),
    C_Vertex3f(0.172780752f, -0.00131011009f, -1.f),
    C_Vertex3f(0.158124924f, 0.0874156952f, -1.f),
    C_Vertex3f(0.155277491f, 0.00497794151f, -1.f),
    C_Vertex3f(0.150608778f, 0.0959658623f, -1.f),
    C_Vertex3f(0.143682718f, 0.105684757f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138668418f, 0.012838006f, -1.f),
    C_Vertex3f(0.155277491f, 0.00497794151f, -1.f),
    C_Vertex3f(0.143682718f, 0.105684757f, -1.f),
    C_Vertex3f(0.137346745f, 0.11657238f, -1.f),
    C_Vertex3f(0.131600857f, 0.128628731f, -1.f),
    C_Vertex3f(0.126445055f, 0.141853809f, -1.f),
    C_Vertex3f(0.122953534f, 0.0222700834f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122953534f, 0.0222700834f, -1.f),
    C_Vertex3f(0.126445055f, 0.141853809f, -1.f),
    C_Vertex3f(0.121879339f, 0.156247616f, -1.f),
    C_Vertex3f(0.117903709f, 0.17181015f, -1.f),
    C_Vertex3f(0.116210938f, 0.180175781f, -1.f),
    C_Vertex3f(0.108132839f, 0.0332741737f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.180175781f, -1.f),
    C_Vertex3f(0.0375976563f, 0.169921875f, -1.f),
    C_Vertex3f(0.0385808945f, 0.15991497f, -1.f),
    C_Vertex3f(0.0420423746f, 0.140830874f, -1.f),
    C_Vertex3f(0.0469988585f, 0.122676492f, -1.f),
    C_Vertex3f(0.053450346f, 0.105451822f, -1.f),
    C_Vertex3f(0.0613968372f, 0.0891568661f, -1.f),
    C_Vertex3f(0.0708383322f, 0.0737916231f, -1.f),
    C_Vertex3f(0.0817748308f, 0.0593560934f, -1.f),
    C_Vertex3f(0.0942063332f, 0.0458502769f, -1.f),
    C_Vertex3f(0.108132839f, 0.0332741737f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.42578125f, 0.466491699f, -1.f),
    C_Vertex3f(0.42578125f, 0.487758636f, -1.f),
    C_Vertex3f(0.42299509f, 0.446696758f, -1.f),
    C_Vertex3f(0.422857285f, 0.508971214f, -1.f),
    C_Vertex3f(0.417422771f, 0.427828312f, -1.f),
    C_Vertex3f(0.417009354f, 0.529583931f, -1.f),
    C_Vertex3f(0.409064293f, 0.40988636f, -1.f),
    C_Vertex3f(0.408237457f, 0.549596786f, -1.f),
    C_Vertex3f(0.397919655f, 0.392870903f, -1.f),
    C_Vertex3f(0.396541595f, 0.569009781f, -1.f),
    C_Vertex3f(0.38409853f, 0.377601147f, -1.f),
    C_Vertex3f(0.382250309f, 0.586378098f, -1.f),
    C_Vertex3f(0.367600918f, 0.364077091f, -1.f),
    C_Vertex3f(0.365363598f, 0.601701736f, -1.f),
    C_Vertex3f(0.348426819f, 0.352298737f, -1.f),
    C_Vertex3f(0.345881462f, 0.614980698f, -1.f),
    C_Vertex3f(0.344726563f, 0.484046936f, -1.f),
    C_Vertex3f(0.343811512f, 0.494386554f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.345881462f, 0.614980698f, -1.f),
    C_Vertex3f(0.323803902f, 0.626214981f, -1.f),
    C_Vertex3f(0.325510502f, 0.539130569f, -1.f),
    C_Vertex3f(0.331000805f, 0.531108975f, -1.f),
    C_Vertex3f(0.335576057f, 0.522623777f, -1.f),
    C_Vertex3f(0.339236259f, 0.513674974f, -1.f),
    C_Vertex3f(0.341981411f, 0.504262567f, -1.f),
    C_Vertex3f(0.343811512f, 0.494386554f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.323803902f, 0.626214981f, -1.f),
    C_Vertex3f(0.300687313f, 0.634640694f, -1.f),
    C_Vertex3f(0.303931713f, 0.559990525f, -1.f),
    C_Vertex3f(0.311784744f, 0.55378294f, -1.f),
    C_Vertex3f(0.319105148f, 0.546688557f, -1.f),
    C_Vertex3f(0.325510502f, 0.539130569f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.300687313f, 0.634640694f, -1.f),
    C_Vertex3f(0.276531696f, 0.640257835f, -1.f),
    C_Vertex3f(0.277176857f, 0.573292494f, -1.f),
    C_Vertex3f(0.286627769f, 0.569745302f, -1.f),
    C_Vertex3f(0.295546055f, 0.565311313f, -1.f),
    C_Vertex3f(0.303931713f, 0.559990525f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.234101295f, 0.578613281f, -1.f),
    C_Vertex3f(0.245628357f, 0.578613281f, -1.f),
    C_Vertex3f(0.251337051f, 0.643066406f, -1.f),
    C_Vertex3f(0.256677151f, 0.577726483f, -1.f),
    C_Vertex3f(0.276531696f, 0.640257835f, -1.f),
    C_Vertex3f(0.267193317f, 0.575952888f, -1.f),
    C_Vertex3f(0.277176857f, 0.573292494f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.228711128f, 0.643066406f, -1.f),
    C_Vertex3f(0.21047163f, 0.641708016f, -1.f),
    C_Vertex3f(0.212590337f, 0.575901031f, -1.f),
    C_Vertex3f(0.223106742f, 0.577709198f, -1.f),
    C_Vertex3f(0.234101295f, 0.578613281f, -1.f),
    C_Vertex3f(0.251337051f, 0.643066406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.183910012f, 0.565052032f, -1.f),
    C_Vertex3f(0.192991972f, 0.569572449f, -1.f),
    C_Vertex3f(0.193010807f, 0.638991237f, -1.f),
    C_Vertex3f(0.20255208f, 0.573188782f, -1.f),
    C_Vertex3f(0.21047163f, 0.641708016f, -1.f),
    C_Vertex3f(0.212590337f, 0.575901031f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.176328659f, 0.634916067f, -1.f),
    C_Vertex3f(0.160425186f, 0.629482508f, -1.f),
    C_Vertex3f(0.167180538f, 0.55329895f, -1.f),
    C_Vertex3f(0.175306201f, 0.559627533f, -1.f),
    C_Vertex3f(0.183910012f, 0.565052032f, -1.f),
    C_Vertex3f(0.193010807f, 0.638991237f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.160425186f, 0.629482508f, -1.f),
    C_Vertex3f(0.145300388f, 0.622690558f, -1.f),
    C_Vertex3f(0.14628613f, 0.528888702f, -1.f),
    C_Vertex3f(0.152568221f, 0.537929535f, -1.f),
    C_Vertex3f(0.159533024f, 0.546066284f, -1.f),
    C_Vertex3f(0.167180538f, 0.55329895f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145300388f, 0.622690558f, -1.f),
    C_Vertex3f(0.130954266f, 0.614540219f, -1.f),
    C_Vertex3f(0.131536126f, 0.496341705f, -1.f),
    C_Vertex3f(0.135770082f, 0.508094788f, -1.f),
    C_Vertex3f(0.14068675f, 0.518943787f, -1.f),
    C_Vertex3f(0.14628613f, 0.528888702f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130954266f, 0.614540219f, -1.f),
    C_Vertex3f(0.117386818f, 0.60503149f, -1.f),
    C_Vertex3f(0.124023438f, 0.462890625f, -1.f),
    C_Vertex3f(0.125116348f, 0.470123291f, -1.f),
    C_Vertex3f(0.127984881f, 0.48368454f, -1.f),
    C_Vertex3f(0.131536126f, 0.496341705f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.124023438f, 0.462890625f, -1.f),
    C_Vertex3f(0.117386818f, 0.60503149f, -1.f),
    C_Vertex3f(0.104598045f, 0.594164371f, -1.f),
    C_Vertex3f(0.0929570198f, 0.582184434f, -1.f),
    C_Vertex3f(0.0824637413f, 0.569091678f, -1.f),
    C_Vertex3f(0.0731182098f, 0.554886103f, -1.f),
    C_Vertex3f(0.0649204254f, 0.539567709f, -1.f),
    C_Vertex3f(0.057870388f, 0.523136497f, -1.f),
    C_Vertex3f(0.0519680977f, 0.505592465f, -1.f),
    C_Vertex3f(0.0472135544f, 0.486935616f, -1.f),
    C_Vertex3f(0.0454101563f, 0.477050781f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.348426819f, 0.352298737f, -1.f),
    C_Vertex3f(0.344726563f, 0.484046936f, -1.f),
    C_Vertex3f(0.344726563f, 0.471823692f, -1.f),
    C_Vertex3f(0.343470335f, 0.459067225f, -1.f),
    C_Vertex3f(0.34095788f, 0.447197556f, -1.f),
    C_Vertex3f(0.337501526f, 0.34728241f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.337501526f, 0.34728241f, -1.f),
    C_Vertex3f(0.34095788f, 0.447197556f, -1.f),
    C_Vertex3f(0.337189198f, 0.436214685f, -1.f),
    C_Vertex3f(0.421754837f, 0.298080444f, -1.f),
    C_Vertex3f(0.41688931f, 0.0702028275f, -1.f),
    C_Vertex3f(0.428358436f, 0.0863714218f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_52[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.2890625f, 0.f, 0.f),
    C_Vertex3f(0.2890625f, 0.f, -1.f),
    C_Vertex3f(0.2890625f, 0.153320313f, -1.f),
    C_Vertex3f(0.2890625f, 0.153320313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.2890625f, 0.153320313f, 0.f),
    C_Vertex3f(0.2890625f, 0.153320313f, -1.f),
    C_Vertex3f(0.0112304688f, 0.153320313f, -1.f),
    C_Vertex3f(0.0112304688f, 0.153320313f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0112304688f, 0.153320313f, 0.f),
    C_Vertex3f(0.0112304688f, 0.153320313f, -1.f),
    C_Vertex3f(0.0112304688f, 0.225585938f, -1.f),
    C_Vertex3f(0.0112304688f, 0.225585938f, 0.f),
    C_Normal3f(-0.817873478f, 0.575398088f, 0.f),
    C_Vertex3f(0.0112304688f, 0.225585938f, 0.f),
    C_Vertex3f(0.0112304688f, 0.225585938f, -1.f),
    C_Vertex3f(0.303222656f, 0.640625f, -1.f),
    C_Vertex3f(0.303222656f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.303222656f, 0.640625f, 0.f),
    C_Vertex3f(0.303222656f, 0.640625f, -1.f),
    C_Vertex3f(0.367675781f, 0.640625f, -1.f),
    C_Vertex3f(0.367675781f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.367675781f, 0.640625f, 0.f),
    C_Vertex3f(0.367675781f, 0.640625f, -1.f),
    C_Vertex3f(0.367675781f, 0.225585938f, -1.f),
    C_Vertex3f(0.367675781f, 0.225585938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.367675781f, 0.225585938f, 0.f),
    C_Vertex3f(0.367675781f, 0.225585938f, -1.f),
    C_Vertex3f(0.454101563f, 0.225585938f, -1.f),
    C_Vertex3f(0.454101563f, 0.225585938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.454101563f, 0.225585938f, 0.f),
    C_Vertex3f(0.454101563f, 0.225585938f, -1.f),
    C_Vertex3f(0.454101563f, 0.153320313f, -1.f),
    C_Vertex3f(0.454101563f, 0.153320313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.454101563f, 0.153320313f, 0.f),
    C_Vertex3f(0.454101563f, 0.153320313f, -1.f),
    C_Vertex3f(0.367675781f, 0.153320313f, -1.f),
    C_Vertex3f(0.367675781f, 0.153320313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.367675781f, 0.153320313f, 0.f),
    C_Vertex3f(0.367675781f, 0.153320313f, -1.f),
    C_Vertex3f(0.367675781f, 0.f, -1.f),
    C_Vertex3f(0.367675781f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.367675781f, 0.f, 0.f),
    C_Vertex3f(0.367675781f, 0.f, -1.f),
    C_Vertex3f(0.2890625f, 0.f, -1.f),
    C_Vertex3f(0.2890625f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.2890625f, 0.225585938f, 0.f),
    C_Vertex3f(0.2890625f, 0.225585938f, -1.f),
    C_Vertex3f(0.2890625f, 0.514831543f, -1.f),
    C_Vertex3f(0.2890625f, 0.514831543f, 0.f),
    C_Normal3f(0.821854949f, -0.569696844f, 0.f),
    C_Vertex3f(0.2890625f, 0.514831543f, 0.f),
    C_Vertex3f(0.2890625f, 0.514831543f, -1.f),
    C_Vertex3f(0.0885620117f, 0.225585938f, -1.f),
    C_Vertex3f(0.0885620117f, 0.225585938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0885620117f, 0.225585938f, 0.f),
    C_Vertex3f(0.0885620117f, 0.225585938f, -1.f),
    C_Vertex3f(0.2890625f, 0.225585938f, -1.f),
    C_Vertex3f(0.2890625f, 0.225585938f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0112304688f, 0.153320313f, 0.f),
    C_Vertex3f(0.2890625f, 0.153320313f, 0.f),
    C_Vertex3f(0.0112304688f, 0.225585938f, 0.f),
    C_Vertex3f(0.0885620117f, 0.225585938f, 0.f),
    C_Vertex3f(0.303222656f, 0.640625f, 0.f),
    C_Vertex3f(0.2890625f, 0.514831543f, 0.f),
    C_Vertex3f(0.367675781f, 0.f, 0.f),
    C_Vertex3f(0.2890625f, 0.225585938f, 0.f),
    C_Vertex3f(0.2890625f, 0.153320313f, 0.f),
    C_Vertex3f(0.0885620117f, 0.225585938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.303222656f, 0.640625f, 0.f),
    C_Vertex3f(0.367675781f, 0.f, 0.f),
    C_Vertex3f(0.367675781f, 0.640625f, 0.f),
    C_Vertex3f(0.367675781f, 0.153320313f, 0.f),
    C_Vertex3f(0.367675781f, 0.225585938f, 0.f),
    C_Vertex3f(0.454101563f, 0.153320313f, 0.f),
    C_Vertex3f(0.454101563f, 0.225585938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.367675781f, 0.f, 0.f),
    C_Vertex3f(0.2890625f, 0.153320313f, 0.f),
    C_Vertex3f(0.2890625f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0112304688f, 0.153320313f, -1.f),
    C_Vertex3f(0.2890625f, 0.153320313f, -1.f),
    C_Vertex3f(0.0112304688f, 0.225585938f, -1.f),
    C_Vertex3f(0.0885620117f, 0.225585938f, -1.f),
    C_Vertex3f(0.303222656f, 0.640625f, -1.f),
    C_Vertex3f(0.2890625f, 0.514831543f, -1.f),
    C_Vertex3f(0.367675781f, 0.f, -1.f),
    C_Vertex3f(0.2890625f, 0.225585938f, -1.f),
    C_Vertex3f(0.2890625f, 0.153320313f, -1.f),
    C_Vertex3f(0.0885620117f, 0.225585938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.303222656f, 0.640625f, -1.f),
    C_Vertex3f(0.367675781f, 0.f, -1.f),
    C_Vertex3f(0.367675781f, 0.640625f, -1.f),
    C_Vertex3f(0.367675781f, 0.153320313f, -1.f),
    C_Vertex3f(0.367675781f, 0.225585938f, -1.f),
    C_Vertex3f(0.454101563f, 0.153320313f, -1.f),
    C_Vertex3f(0.454101563f, 0.225585938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.367675781f, 0.f, -1.f),
    C_Vertex3f(0.2890625f, 0.153320313f, -1.f),
    C_Vertex3f(0.2890625f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_53[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.0825574473f, 0.996586323f, 0.f),
    C_Vertex3f(0.037109375f, 0.16796875f, 0.f),
    C_Vertex3f(0.037109375f, 0.16796875f, -1.f),
    C_Vertex3f(0.119628906f, 0.174804688f, -1.f),
    C_Vertex3f(0.119628906f, 0.174804688f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.988623321f, 0.150412783f, 0.f),
    C_Vertex3f(0.119628906f, 0.174804688f, 0.f),
    C_Vertex3f(0.119628906f, 0.174804688f, -1.f),
    C_Normal3f(0.981788993f, 0.189974412f, 0.f),
    C_Vertex3f(0.120776176f, 0.167263985f, 0.f),
    C_Vertex3f(0.120776176f, 0.167263985f, -1.f),
    C_Normal3f(0.970029175f, 0.242988139f, 0.f),
    C_Vertex3f(0.123825192f, 0.153118253f, 0.f),
    C_Vertex3f(0.123825192f, 0.153118253f, -1.f),
    C_Normal3f(0.950164378f, 0.311749429f, 0.f),
    C_Vertex3f(0.127628684f, 0.139908195f, 0.f),
    C_Vertex3f(0.127628684f, 0.139908195f, -1.f),
    C_Normal3f(0.922637641f, 0.385667861f, 0.f),
    C_Vertex3f(0.132186651f, 0.12763381f, 0.f),
    C_Vertex3f(0.132186651f, 0.12763381f, -1.f),
    C_Normal3f(0.885985076f, 0.463713616f, 0.f),
    C_Vertex3f(0.137499094f, 0.116295099f, 0.f),
    C_Vertex3f(0.137499094f, 0.116295099f, -1.f),
    C_Normal3f(0.838972747f, 0.54417336f, 0.f),
    C_Vertex3f(0.143566012f, 0.105892062f, 0.f),
    C_Vertex3f(0.143566012f, 0.105892062f, -1.f),
    C_Normal3f(0.780910671f, 0.62464267f, 0.f),
    C_Vertex3f(0.150387406f, 0.0964246988f, 0.f),
    C_Vertex3f(0.150387406f, 0.0964246988f, -1.f),
    C_Normal3f(0.711978853f, 0.70220083f, 0.f),
    C_Vertex3f(0.157963276f, 0.0878930092f, 0.f),
    C_Vertex3f(0.157963276f, 0.0878930092f, -1.f),
    C_Normal3f(0.639934123f, 0.768429756f, 0.f),
    C_Vertex3f(0.166293621f, 0.0802969933f, 0.f),
    C_Vertex3f(0.166293621f, 0.0802969933f, -1.f),
    C_Normal3f(0.565854907f, 0.824504912f, 0.f),
    C_Vertex3f(0.175065637f, 0.0736504793f, 0.f),
    C_Vertex3f(0.175065637f, 0.0736504793f, -1.f),
    C_Normal3f(0.48428601f, 0.874909699f, 0.f),
    C_Vertex3f(0.184279323f, 0.0679534674f, 0.f),
    C_Vertex3f(0.184279323f, 0.0679534674f, -1.f),
    C_Normal3f(0.397065639f, 0.917790174f, 0.f),
    C_Vertex3f(0.193934679f, 0.0632059574f, 0.f),
    C_Vertex3f(0.193934679f, 0.0632059574f, -1.f),
    C_Normal3f(0.306577802f, 0.951845586f, 0.f),
    C_Vertex3f(0.204031706f, 0.0594079494f, 0.f),
    C_Vertex3f(0.204031706f, 0.0594079494f, -1.f),
    C_Normal3f(0.215438113f, 0.976517558f, 0.f),
    C_Vertex3f(0.214570403f, 0.0565594435f, 0.f),
    C_Vertex3f(0.214570403f, 0.0565594435f, -1.f),
    C_Normal3f(0.126136214f, 0.992012918f, 0.f),
    C_Vertex3f(0.225550771f, 0.0546604395f, 0.f),
    C_Vertex3f(0.225550771f, 0.0546604395f, -1.f),
    C_Normal3f(0.0387219526f, 0.999250054f, 0.f),
    C_Vertex3f(0.236972809f, 0.0537109375f, 0.f),
    C_Vertex3f(0.236972809f, 0.0537109375f, -1.f),
    C_Normal3f(-0.0499646328f, 0.998750985f, 0.f),
    C_Vertex3f(0.250053406f, 0.0537109375f, 0.f),
    C_Vertex3f(0.250053406f, 0.0537109375f, -1.f),
    C_Normal3f(-0.147932872f, 0.9889974f, 0.f),
    C_Vertex3f(0.263787508f, 0.0550523996f, 0.f),
    C_Vertex3f(0.263787508f, 0.0550523996f, -1.f),
    C_Normal3f(-0.251810223f, 0.967776537f, 0.f),
    C_Vertex3f(0.276958227f, 0.0577353239f, 0.f),
    C_Vertex3f(0.276958227f, 0.0577353239f, -1.f),
    C_Normal3f(-0.355971098f, 0.934496999f, 0.f),
    C_Vertex3f(0.289565563f, 0.0617597103f, 0.f),
    C_Vertex3f(0.289565563f, 0.0617597103f, -1.f),
    C_Normal3f(-0.456595123f, 0.889674544f, 0.f),
    C_Vertex3f(0.301609516f, 0.0671255589f, 0.f),
    C_Vertex3f(0.301609516f, 0.0671255589f, -1.f),
    C_Normal3f(-0.550155461f, 0.835062265f, 0.f),
    C_Vertex3f(0.313090086f, 0.0738328695f, 0.f),
    C_Vertex3f(0.313090086f, 0.0738328695f, -1.f),
    C_Normal3f(-0.634009898f, 0.773324847f, 0.f),
    C_Vertex3f(0.324007273f, 0.0818816423f, 0.f),
    C_Vertex3f(0.324007273f, 0.0818816423f, -1.f),
    C_Normal3f(-0.706715226f, 0.707498193f, 0.f),
    C_Vertex3f(0.334361076f, 0.0912718773f, 0.f),
    C_Vertex3f(0.334361076f, 0.0912718773f, -1.f),
    C_Normal3f(-0.772611499f, 0.634879112f, 0.f),
    C_Vertex3f(0.344151497f, 0.102003574f, 0.f),
    C_Vertex3f(0.344151497f, 0.102003574f, -1.f),
    C_Normal3f(-0.834466457f, 0.551058769f, 0.f),
    C_Vertex3f(0.352718115f, 0.113611341f, 0.f),
    C_Vertex3f(0.352718115f, 0.113611341f, -1.f),
    C_Normal3f(-0.886892378f, 0.4619762f, 0.f),
    C_Vertex3f(0.36006093f, 0.126095176f, 0.f),
    C_Vertex3f(0.36006093f, 0.126095176f, -1.f),
    C_Normal3f(-0.928756475f, 0.370690554f, 0.f),
    C_Vertex3f(0.366179943f, 0.13945508f, 0.f),
    C_Vertex3f(0.366179943f, 0.13945508f, -1.f),
    C_Normal3f(-0.959940314f, 0.280204356f, 0.f),
    C_Vertex3f(0.371075153f, 0.153691053f, 0.f),
    C_Vertex3f(0.371075153f, 0.153691053f, -1.f),
    C_Normal3f(-0.981188774f, 0.193050712f, 0.f),
    C_Vertex3f(0.374746561f, 0.168803096f, 0.f),
    C_Vertex3f(0.374746561f, 0.168803096f, -1.f),
    C_Normal3f(-0.993813336f, 0.111063607f, 0.f),
    C_Vertex3f(0.377194166f, 0.184791207f, 0.f),
    C_Vertex3f(0.377194166f, 0.184791207f, -1.f),
    C_Normal3f(-0.999359131f, 0.035795629f, 0.f),
    C_Vertex3f(0.378417969f, 0.201655388f, 0.f),
    C_Vertex3f(0.378417969f, 0.201655388f, -1.f),
    C_Normal3f(-0.999376833f, -0.035299249f, 0.f),
    C_Vertex3f(0.378417969f, 0.218957901f, 0.f),
    C_Vertex3f(0.378417969f, 0.218957901f, -1.f),
    C_Normal3f(-0.993617296f, -0.112803586f, 0.f),
    C_Vertex3f(0.377242565f, 0.234932899f, 0.f),
    C_Vertex3f(0.377242565f, 0.234932899f, -1.f),
    C_Normal3f(-0.980445921f, -0.19678852f, 0.f),
    C_Vertex3f(0.374891758f, 0.25001812f, 0.f),
    C_Vertex3f(0.374891758f, 0.25001812f, -1.f),
    C_Normal3f(-0.958041847f, -0.286628485f, 0.f),
    C_Vertex3f(0.371365547f, 0.264213562f, 0.f),
    C_Vertex3f(0.371365547f, 0.264213562f, -1.f),
    C_Normal3f(-0.92483741f, -0.38036257f, 0.f),
    C_Vertex3f(0.366663933f, 0.277519226f, 0.f),
    C_Vertex3f(0.366663933f, 0.277519226f, -1.f),
    C_Normal3f(-0.879893661f, -0.475170612f, 0.f),
    C_Vertex3f(0.360786915f, 0.289935112f, 0.f),
    C_Vertex3f(0.360786915f, 0.289935112f, -1.f),
    C_Normal3f(-0.82328552f, -0.56762737f, 0.f),
    C_Vertex3f(0.353734493f, 0.30146122f, 0.f),
    C_Vertex3f(0.353734493f, 0.30146122f, -1.f),
    C_Normal3f(-0.756315172f, -0.654207528f, 0.f),
    C_Vertex3f(0.345506668f, 0.312097549f, 0.f),
    C_Vertex3f(0.345506668f, 0.312097549f, -1.f),
    C_Normal3f(-0.682806849f, -0.730598927f, 0.f),
    C_Vertex3f(0.336103439f, 0.321844101f, 0.f),
    C_Vertex3f(0.336103439f, 0.321844101f, -1.f),
    C_Normal3f(-0.6012519f, -0.79905951f, 0.f),
    C_Vertex3f(0.325952768f, 0.330372334f, 0.f),
    C_Vertex3f(0.325952768f, 0.330372334f, -1.f),
    C_Normal3f(-0.510995567f, -0.859583378f, 0.f),
    C_Vertex3f(0.315054655f, 0.337682247f, 0.f),
    C_Vertex3f(0.315054655f, 0.337682247f, -1.f),
    C_Normal3f(-0.415002555f, -0.909820259f, 0.f),
    C_Vertex3f(0.3034091f, 0.343773842f, 0.f),
    C_Vertex3f(0.3034091f, 0.343773842f, -1.f),
    C_Normal3f(-0.316798955f, -0.948492765f, 0.f),
    C_Vertex3f(0.291016102f, 0.348647118f, 0.f),
    C_Vertex3f(0.291016102f, 0.348647118f, -1.f),
    C_Normal3f(-0.21986334f, -0.975530624f, 0.f),
    C_Vertex3f(0.277875662f, 0.352302074f, 0.f),
    C_Vertex3f(0.277875662f, 0.352302074f, -1.f),
    C_Normal3f(-0.127100512f, -0.991889775f, 0.f),
    C_Vertex3f(0.26398778f, 0.354738712f, 0.f),
    C_Vertex3f(0.26398778f, 0.354738712f, -1.f),
    C_Normal3f(-0.0381734818f, -0.999271154f, 0.f),
    C_Vertex3f(0.249352455f, 0.355957031f, 0.f),
    C_Vertex3f(0.249352455f, 0.355957031f, -1.f),
    C_Normal3f(0.0612595342f, -0.998121858f, 0.f),
    C_Vertex3f(0.232095718f, 0.355957031f, 0.f),
    C_Vertex3f(0.232095718f, 0.355957031f, -1.f),
    C_Normal3f(0.181458637f, -0.983398616f, 0.f),
    C_Vertex3f(0.213940144f, 0.353783607f, 0.f),
    C_Vertex3f(0.213940144f, 0.353783607f, -1.f),
    C_Normal3f(0.309517503f, -0.9508937f, 0.f),
    C_Vertex3f(0.196759701f, 0.34943676f, 0.f),
    C_Vertex3f(0.196759701f, 0.34943676f, -1.f),
    C_Normal3f(0.435636222f, -0.900122821f, 0.f),
    C_Vertex3f(0.18055439f, 0.342916489f, 0.f),
    C_Vertex3f(0.18055439f, 0.342916489f, -1.f),
    C_Normal3f(0.543660343f, -0.839305341f, 0.f),
    C_Vertex3f(0.165324211f, 0.334222794f, 0.f),
    C_Vertex3f(0.165324211f, 0.334222794f, -1.f),
    C_Normal3f(0.639068782f, -0.769149661f, 0.f),
    C_Vertex3f(0.151716709f, 0.32423687f, 0.f),
    C_Vertex3f(0.151716709f, 0.32423687f, -1.f),
    C_Normal3f(0.729705572f, -0.683761537f, 0.f),
    C_Vertex3f(0.139731884f, 0.312958717f, 0.f),
    C_Vertex3f(0.139731884f, 0.312958717f, -1.f),
    C_Normal3f(0.797920883f, -0.602762103f, 0.f),
    C_Vertex3f(0.129369736f, 0.300388336f, 0.f),
    C_Vertex3f(0.129369736f, 0.300388336f, -1.f),
    C_Normal3f(0.845925629f, -0.533300936f, 0.f),
    C_Vertex3f(0.125f, 0.293457031f, 0.f),
    C_Vertex3f(0.125f, 0.293457031f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.131303608f, -0.991342187f, 0.f),
    C_Vertex3f(0.125f, 0.293457031f, 0.f),
    C_Vertex3f(0.125f, 0.293457031f, -1.f),
    C_Vertex3f(0.0512695313f, 0.303222656f, -1.f),
    C_Vertex3f(0.0512695313f, 0.303222656f, 0.f),
    C_Normal3f(-0.982706726f, 0.185168773f, 0.f),
    C_Vertex3f(0.0512695313f, 0.303222656f, 0.f),
    C_Vertex3f(0.0512695313f, 0.303222656f, -1.f),
    C_Vertex3f(0.11328125f, 0.632324219f, -1.f),
    C_Vertex3f(0.11328125f, 0.632324219f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.11328125f, 0.632324219f, 0.f),
    C_Vertex3f(0.11328125f, 0.632324219f, -1.f),
    C_Vertex3f(0.431640625f, 0.632324219f, -1.f),
    C_Vertex3f(0.431640625f, 0.632324219f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.431640625f, 0.632324219f, 0.f),
    C_Vertex3f(0.431640625f, 0.632324219f, -1.f),
    C_Vertex3f(0.431640625f, 0.557128906f, -1.f),
    C_Vertex3f(0.431640625f, 0.557128906f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.431640625f, 0.557128906f, 0.f),
    C_Vertex3f(0.431640625f, 0.557128906f, -1.f),
    C_Vertex3f(0.176139832f, 0.557128906f, -1.f),
    C_Vertex3f(0.176139832f, 0.557128906f, 0.f),
    C_Normal3f(0.980468333f, -0.196676612f, 0.f),
    C_Vertex3f(0.176139832f, 0.557128906f, 0.f),
    C_Vertex3f(0.176139832f, 0.557128906f, -1.f),
    C_Vertex3f(0.141647339f, 0.385177612f, -1.f),
    C_Vertex3f(0.141647339f, 0.385177612f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.571032643f, 0.820927262f, 0.f),
    C_Vertex3f(0.141647339f, 0.385177612f, 0.f),
    C_Vertex3f(0.141647339f, 0.385177612f, -1.f),
    C_Normal3f(-0.534509599f, 0.845162451f, 0.f),
    C_Vertex3f(0.148856163f, 0.390192032f, 0.f),
    C_Vertex3f(0.148856163f, 0.390192032f, -1.f),
    C_Normal3f(-0.485215575f, 0.874394596f, 0.f),
    C_Vertex3f(0.163451433f, 0.398967266f, 0.f),
    C_Vertex3f(0.163451433f, 0.398967266f, -1.f),
    C_Normal3f(-0.420847684f, 0.907131314f, 0.f),
    C_Vertex3f(0.178224325f, 0.406488895f, 0.f),
    C_Vertex3f(0.178224325f, 0.406488895f, -1.f),
    C_Normal3f(-0.351203948f, 0.936298966f, 0.f),
    C_Vertex3f(0.193174839f, 0.41275692f, 0.f),
    C_Vertex3f(0.193174839f, 0.41275692f, -1.f),
    C_Normal3f(-0.277050674f, 0.960855365f, 0.f),
    C_Vertex3f(0.208302975f, 0.417771339f, 0.f),
    C_Vertex3f(0.208302975f, 0.417771339f, -1.f),
    C_Normal3f(-0.199487254f, 0.979900479f, 0.f),
    C_Vertex3f(0.223608732f, 0.421532154f, 0.f),
    C_Vertex3f(0.223608732f, 0.421532154f, -1.f),
    C_Normal3f(-0.119883306f, 0.992787898f, 0.f),
    C_Vertex3f(0.239092112f, 0.424039364f, 0.f),
    C_Vertex3f(0.239092112f, 0.424039364f, -1.f),
    C_Normal3f(-0.0367734171f, 0.999323547f, 0.f),
    C_Vertex3f(0.254753113f, 0.425292969f, 0.f),
    C_Vertex3f(0.254753113f, 0.425292969f, -1.f),
    C_Normal3f(0.0470875688f, 0.998890758f, 0.f),
    C_Vertex3f(0.273159027f, 0.425292969f, 0.f),
    C_Vertex3f(0.273159027f, 0.425292969f, -1.f),
    C_Normal3f(0.13679567f, 0.990599275f, 0.f),
    C_Vertex3f(0.293312907f, 0.423475266f, 0.f),
    C_Vertex3f(0.293312907f, 0.423475266f, -1.f),
    C_Normal3f(0.233483911f, 0.972360671f, 0.f),
    C_Vertex3f(0.312647462f, 0.419839859f, 0.f),
    C_Vertex3f(0.312647462f, 0.419839859f, -1.f),
    C_Normal3f(0.331511319f, 0.943451226f, 0.f),
    C_Vertex3f(0.331162691f, 0.414386749f, 0.f),
    C_Vertex3f(0.331162691f, 0.414386749f, -1.f),
    C_Normal3f(0.427720964f, 0.903910875f, 0.f),
    C_Vertex3f(0.348858595f, 0.407115936f, 0.f),
    C_Vertex3f(0.348858595f, 0.407115936f, -1.f),
    C_Normal3f(0.51896286f, 0.854796767f, 0.f),
    C_Vertex3f(0.365735173f, 0.39802742f, 0.f),
    C_Vertex3f(0.365735173f, 0.39802742f, -1.f),
    C_Normal3f(0.602587819f, 0.798052549f, 0.f),
    C_Vertex3f(0.381792426f, 0.387121201f, 0.f),
    C_Vertex3f(0.381792426f, 0.387121201f, -1.f),
    C_Normal3f(0.676808357f, 0.736159265f, 0.f),
    C_Vertex3f(0.397030354f, 0.374397278f, 0.f),
    C_Vertex3f(0.397030354f, 0.374397278f, -1.f),
    C_Normal3f(0.744214356f, 0.667940915f, 0.f),
    C_Vertex3f(0.411448956f, 0.359855652f, 0.f),
    C_Vertex3f(0.411448956f, 0.359855652f, -1.f),
    C_Normal3f(0.808588862f, 0.588374078f, 0.f),
    C_Vertex3f(0.424065232f, 0.344275236f, 0.f),
    C_Vertex3f(0.424065232f, 0.344275236f, -1.f),
    C_Normal3f(0.865635574f, 0.500674665f, 0.f),
    C_Vertex3f(0.434879184f, 0.327656031f, 0.f),
    C_Vertex3f(0.434879184f, 0.327656031f, -1.f),
    C_Normal3f(0.913220942f, 0.407464713f, 0.f),
    C_Vertex3f(0.44389081f, 0.309998035f, 0.f),
    C_Vertex3f(0.44389081f, 0.309998035f, -1.f),
    C_Normal3f(0.950115919f, 0.311896682f, 0.f),
    C_Vertex3f(0.451100111f, 0.29130125f, 0.f),
    C_Vertex3f(0.451100111f, 0.29130125f, -1.f),
    C_Normal3f(0.976138949f, 0.217146695f, 0.f),
    C_Vertex3f(0.456507087f, 0.271565676f, 0.f),
    C_Vertex3f(0.456507087f, 0.271565676f, -1.f),
    C_Normal3f(0.992036521f, 0.125950456f, 0.f),
    C_Vertex3f(0.460111737f, 0.250791311f, 0.f),
    C_Vertex3f(0.460111737f, 0.250791311f, -1.f),
    C_Normal3f(0.999166608f, 0.0408186205f, 0.f),
    C_Vertex3f(0.461914063f, 0.228978157f, 0.f),
    C_Vertex3f(0.461914063f, 0.228978157f, -1.f),
    C_Normal3f(0.99933207f, -0.0365442298f, 0.f),
    C_Vertex3f(0.461914063f, 0.206673622f, 0.f),
    C_Vertex3f(0.461914063f, 0.206673622f, -1.f),
    C_Normal3f(0.99344337f, -0.114325672f, 0.f),
    C_Vertex3f(0.460329771f, 0.185654402f, 0.f),
    C_Vertex3f(0.460329771f, 0.185654402f, -1.f),
    C_Normal3f(0.980786085f, -0.195086226f, 0.f),
    C_Vertex3f(0.457161188f, 0.165373087f, 0.f),
    C_Vertex3f(0.457161188f, 0.165373087f, -1.f),
    C_Normal3f(0.960637629f, -0.277804732f, 0.f),
    C_Vertex3f(0.452408314f, 0.145829678f, 0.f),
    C_Vertex3f(0.452408314f, 0.145829678f, -1.f),
    C_Normal3f(0.932694376f, -0.360667706f, 0.f),
    C_Vertex3f(0.446071148f, 0.127024174f, 0.f),
    C_Vertex3f(0.446071148f, 0.127024174f, -1.f),
    C_Normal3f(0.897162139f, -0.441701293f, 0.f),
    C_Vertex3f(0.438149691f, 0.108956575f, 0.f),
    C_Vertex3f(0.438149691f, 0.108956575f, -1.f),
    C_Normal3f(0.854780734f, -0.518989265f, 0.f),
    C_Vertex3f(0.428643942f, 0.0916268826f, 0.f),
    C_Vertex3f(0.428643942f, 0.0916268826f, -1.f),
    C_Normal3f(0.806255043f, -0.591567993f, 0.f),
    C_Vertex3f(0.417553902f, 0.0750350952f, 0.f),
    C_Vertex3f(0.417553902f, 0.0750350952f, -1.f),
    C_Normal3f(0.747210979f, -0.664586961f, 0.f),
    C_Vertex3f(0.403513908f, 0.0573768616f, 0.f),
    C_Vertex3f(0.403513908f, 0.0573768616f, -1.f),
    C_Normal3f(0.67011714f, -0.74225533f, 0.f),
    C_Vertex3f(0.386701584f, 0.0403470993f, 0.f),
    C_Vertex3f(0.386701584f, 0.0403470993f, -1.f),
    C_Normal3f(0.577362359f, -0.816488028f, 0.f),
    C_Vertex3f(0.36848259f, 0.0257501602f, 0.f),
    C_Vertex3f(0.36848259f, 0.0257501602f, -1.f),
    C_Normal3f(0.474144131f, -0.880447149f, 0.f),
    C_Vertex3f(0.348856926f, 0.0135860443f, 0.f),
    C_Vertex3f(0.348856926f, 0.0135860443f, -1.f),
    C_Normal3f(0.364756674f, -0.931102872f, 0.f),
    C_Vertex3f(0.327824593f, 0.00385475159f, 0.f),
    C_Vertex3f(0.327824593f, 0.00385475159f, -1.f),
    C_Normal3f(0.254179388f, -0.967157066f, 0.f),
    C_Vertex3f(0.30538559f, -0.00344371796f, 0.f),
    C_Vertex3f(0.30538559f, -0.00344371796f, -1.f),
    C_Normal3f(0.14703536f, -0.989131212f, 0.f),
    C_Vertex3f(0.281539917f, -0.00830936432f, 0.f),
    C_Vertex3f(0.281539917f, -0.00830936432f, -1.f),
    C_Normal3f(0.0490806848f, -0.998794794f, 0.f),
    C_Vertex3f(0.256287575f, -0.0107421875f, 0.f),
    C_Vertex3f(0.256287575f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0338871181f, -0.99942565f, 0.f),
    C_Vertex3f(0.232031822f, -0.0107421875f, 0.f),
    C_Vertex3f(0.232031822f, -0.0107421875f, -1.f),
    C_Normal3f(-0.112035908f, -0.993704081f, 0.f),
    C_Vertex3f(0.21118319f, -0.00921285152f, 0.f),
    C_Vertex3f(0.21118319f, -0.00921285152f, -1.f),
    C_Normal3f(-0.193910003f, -0.981019258f, 0.f),
    C_Vertex3f(0.19133842f, -0.00615417957f, 0.f),
    C_Vertex3f(0.19133842f, -0.00615417957f, -1.f),
    C_Normal3f(-0.280183673f, -0.959946513f, 0.f),
    C_Vertex3f(0.172497511f, -0.00156617165f, 0.f),
    C_Vertex3f(0.172497511f, -0.00156617165f, -1.f),
    C_Normal3f(-0.368984342f, -0.929435551f, 0.f),
    C_Vertex3f(0.154660463f, 0.00455117226f, 0.f),
    C_Vertex3f(0.154660463f, 0.00455117226f, -1.f),
    C_Normal3f(-0.457882464f, -0.889012694f, 0.f),
    C_Vertex3f(0.137827277f, 0.0121978521f, 0.f),
    C_Vertex3f(0.137827277f, 0.0121978521f, -1.f),
    C_Normal3f(-0.544136822f, -0.83899653f, 0.f),
    C_Vertex3f(0.121997952f, 0.021373868f, 0.f),
    C_Vertex3f(0.121997952f, 0.021373868f, -1.f),
    C_Normal3f(-0.625066936f, -0.780571163f, 0.f),
    C_Vertex3f(0.107172489f, 0.0320792198f, 0.f),
    C_Vertex3f(0.107172489f, 0.0320792198f, -1.f),
    C_Normal3f(-0.697521508f, -0.71656394f, 0.f),
    C_Vertex3f(0.0933508873f, 0.0443139076f, 0.f),
    C_Vertex3f(0.0933508873f, 0.0443139076f, -1.f),
    C_Normal3f(-0.763899505f, -0.645335197f, 0.f),
    C_Vertex3f(0.0810111761f, 0.0575453043f, 0.f),
    C_Vertex3f(0.0810111761f, 0.0575453043f, -1.f),
    C_Normal3f(-0.824239552f, -0.566241205f, 0.f),
    C_Vertex3f(0.0701533556f, 0.0717734098f, 0.f),
    C_Vertex3f(0.0701533556f, 0.0717734098f, -1.f),
    C_Normal3f(-0.876516104f, -0.481372625f, 0.f),
    C_Vertex3f(0.0607774258f, 0.0869982243f, 0.f),
    C_Vertex3f(0.0607774258f, 0.0869982243f, -1.f),
    C_Normal3f(-0.91939503f, -0.393335342f, 0.f),
    C_Vertex3f(0.0528833866f, 0.103219748f, 0.f),
    C_Vertex3f(0.0528833866f, 0.103219748f, -1.f),
    C_Normal3f(-0.952394605f, -0.304868162f, 0.f),
    C_Vertex3f(0.0464712381f, 0.12043798f, 0.f),
    C_Vertex3f(0.0464712381f, 0.12043798f, -1.f),
    C_Normal3f(-0.975845754f, -0.218460873f, 0.f),
    C_Vertex3f(0.0415409803f, 0.138652921f, 0.f),
    C_Vertex3f(0.0415409803f, 0.138652921f, -1.f),
    C_Normal3f(-0.988766313f, -0.149469495f, 0.f),
    C_Vertex3f(0.0380926132f, 0.157864571f, 0.f),
    C_Vertex3f(0.0380926132f, 0.157864571f, -1.f),
    C_Normal3f(-0.995298743f, -0.0968525708f, 0.f),
    C_Vertex3f(0.037109375f, 0.16796875f, 0.f),
    C_Vertex3f(0.037109375f, 0.16796875f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431640625f, 0.557128906f, 0.f),
    C_Vertex3f(0.431640625f, 0.632324219f, 0.f),
    C_Vertex3f(0.176139832f, 0.557128906f, 0.f),
    C_Vertex3f(0.11328125f, 0.632324219f, 0.f),
    C_Vertex3f(0.141647339f, 0.385177612f, 0.f),
    C_Vertex3f(0.125f, 0.293457031f, 0.f),
    C_Vertex3f(0.129369736f, 0.300388336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163451433f, 0.398967266f, 0.f),
    C_Vertex3f(0.148856163f, 0.390192032f, 0.f),
    C_Vertex3f(0.151716709f, 0.32423687f, 0.f),
    C_Vertex3f(0.141647339f, 0.385177612f, 0.f),
    C_Vertex3f(0.139731884f, 0.312958717f, 0.f),
    C_Vertex3f(0.129369736f, 0.300388336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.461914063f, 0.206673622f, 0.f),
    C_Vertex3f(0.461914063f, 0.228978157f, 0.f),
    C_Vertex3f(0.460329771f, 0.185654402f, 0.f),
    C_Vertex3f(0.460111737f, 0.250791311f, 0.f),
    C_Vertex3f(0.457161188f, 0.165373087f, 0.f),
    C_Vertex3f(0.456507087f, 0.271565676f, 0.f),
    C_Vertex3f(0.452408314f, 0.145829678f, 0.f),
    C_Vertex3f(0.451100111f, 0.29130125f, 0.f),
    C_Vertex3f(0.446071148f, 0.127024174f, 0.f),
    C_Vertex3f(0.44389081f, 0.309998035f, 0.f),
    C_Vertex3f(0.438149691f, 0.108956575f, 0.f),
    C_Vertex3f(0.434879184f, 0.327656031f, 0.f),
    C_Vertex3f(0.428643942f, 0.0916268826f, 0.f),
    C_Vertex3f(0.424065232f, 0.344275236f, 0.f),
    C_Vertex3f(0.417553902f, 0.0750350952f, 0.f),
    C_Vertex3f(0.411448956f, 0.359855652f, 0.f),
    C_Vertex3f(0.403513908f, 0.0573768616f, 0.f),
    C_Vertex3f(0.397030354f, 0.374397278f, 0.f),
    C_Vertex3f(0.386701584f, 0.0403470993f, 0.f),
    C_Vertex3f(0.381792426f, 0.387121201f, 0.f),
    C_Vertex3f(0.378417969f, 0.218957901f, 0.f),
    C_Vertex3f(0.377242565f, 0.234932899f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.381792426f, 0.387121201f, 0.f),
    C_Vertex3f(0.365735173f, 0.39802742f, 0.f),
    C_Vertex3f(0.366663933f, 0.277519226f, 0.f),
    C_Vertex3f(0.371365547f, 0.264213562f, 0.f),
    C_Vertex3f(0.374891758f, 0.25001812f, 0.f),
    C_Vertex3f(0.377242565f, 0.234932899f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.336103439f, 0.321844101f, 0.f),
    C_Vertex3f(0.345506668f, 0.312097549f, 0.f),
    C_Vertex3f(0.348858595f, 0.407115936f, 0.f),
    C_Vertex3f(0.353734493f, 0.30146122f, 0.f),
    C_Vertex3f(0.365735173f, 0.39802742f, 0.f),
    C_Vertex3f(0.360786915f, 0.289935112f, 0.f),
    C_Vertex3f(0.366663933f, 0.277519226f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.331162691f, 0.414386749f, 0.f),
    C_Vertex3f(0.312647462f, 0.419839859f, 0.f),
    C_Vertex3f(0.315054655f, 0.337682247f, 0.f),
    C_Vertex3f(0.325952768f, 0.330372334f, 0.f),
    C_Vertex3f(0.336103439f, 0.321844101f, 0.f),
    C_Vertex3f(0.348858595f, 0.407115936f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.277875662f, 0.352302074f, 0.f),
    C_Vertex3f(0.291016102f, 0.348647118f, 0.f),
    C_Vertex3f(0.293312907f, 0.423475266f, 0.f),
    C_Vertex3f(0.3034091f, 0.343773842f, 0.f),
    C_Vertex3f(0.312647462f, 0.419839859f, 0.f),
    C_Vertex3f(0.315054655f, 0.337682247f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163451433f, 0.398967266f, 0.f),
    C_Vertex3f(0.165324211f, 0.334222794f, 0.f),
    C_Vertex3f(0.178224325f, 0.406488895f, 0.f),
    C_Vertex3f(0.18055439f, 0.342916489f, 0.f),
    C_Vertex3f(0.193174839f, 0.41275692f, 0.f),
    C_Vertex3f(0.196759701f, 0.34943676f, 0.f),
    C_Vertex3f(0.208302975f, 0.417771339f, 0.f),
    C_Vertex3f(0.213940144f, 0.353783607f, 0.f),
    C_Vertex3f(0.223608732f, 0.421532154f, 0.f),
    C_Vertex3f(0.232095718f, 0.355957031f, 0.f),
    C_Vertex3f(0.239092112f, 0.424039364f, 0.f),
    C_Vertex3f(0.249352455f, 0.355957031f, 0.f),
    C_Vertex3f(0.254753113f, 0.425292969f, 0.f),
    C_Vertex3f(0.26398778f, 0.354738712f, 0.f),
    C_Vertex3f(0.273159027f, 0.425292969f, 0.f),
    C_Vertex3f(0.277875662f, 0.352302074f, 0.f),
    C_Vertex3f(0.293312907f, 0.423475266f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.386701584f, 0.0403470993f, 0.f),
    C_Vertex3f(0.378417969f, 0.218957901f, 0.f),
    C_Vertex3f(0.378417969f, 0.201655388f, 0.f),
    C_Vertex3f(0.377194166f, 0.184791207f, 0.f),
    C_Vertex3f(0.374746561f, 0.168803096f, 0.f),
    C_Vertex3f(0.371075153f, 0.153691053f, 0.f),
    C_Vertex3f(0.36848259f, 0.0257501602f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36848259f, 0.0257501602f, 0.f),
    C_Vertex3f(0.371075153f, 0.153691053f, 0.f),
    C_Vertex3f(0.366179943f, 0.13945508f, 0.f),
    C_Vertex3f(0.36006093f, 0.126095176f, 0.f),
    C_Vertex3f(0.352718115f, 0.113611341f, 0.f),
    C_Vertex3f(0.348856926f, 0.0135860443f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.352718115f, 0.113611341f, 0.f),
    C_Vertex3f(0.344151497f, 0.102003574f, 0.f),
    C_Vertex3f(0.348856926f, 0.0135860443f, 0.f),
    C_Vertex3f(0.334361076f, 0.0912718773f, 0.f),
    C_Vertex3f(0.327824593f, 0.00385475159f, 0.f),
    C_Vertex3f(0.324007273f, 0.0818816423f, 0.f),
    C_Vertex3f(0.313090086f, 0.0738328695f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.30538559f, -0.00344371796f, 0.f),
    C_Vertex3f(0.327824593f, 0.00385475159f, 0.f),
    C_Vertex3f(0.313090086f, 0.0738328695f, 0.f),
    C_Vertex3f(0.301609516f, 0.0671255589f, 0.f),
    C_Vertex3f(0.289565563f, 0.0617597103f, 0.f),
    C_Vertex3f(0.281539917f, -0.00830936432f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.289565563f, 0.0617597103f, 0.f),
    C_Vertex3f(0.276958227f, 0.0577353239f, 0.f),
    C_Vertex3f(0.281539917f, -0.00830936432f, 0.f),
    C_Vertex3f(0.263787508f, 0.0550523996f, 0.f),
    C_Vertex3f(0.256287575f, -0.0107421875f, 0.f),
    C_Vertex3f(0.250053406f, 0.0537109375f, 0.f),
    C_Vertex3f(0.236972809f, 0.0537109375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.232031822f, -0.0107421875f, 0.f),
    C_Vertex3f(0.256287575f, -0.0107421875f, 0.f),
    C_Vertex3f(0.236972809f, 0.0537109375f, 0.f),
    C_Vertex3f(0.225550771f, 0.0546604395f, 0.f),
    C_Vertex3f(0.214570403f, 0.0565594435f, 0.f),
    C_Vertex3f(0.21118319f, -0.00921285152f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.214570403f, 0.0565594435f, 0.f),
    C_Vertex3f(0.204031706f, 0.0594079494f, 0.f),
    C_Vertex3f(0.21118319f, -0.00921285152f, 0.f),
    C_Vertex3f(0.193934679f, 0.0632059574f, 0.f),
    C_Vertex3f(0.19133842f, -0.00615417957f, 0.f),
    C_Vertex3f(0.184279323f, 0.0679534674f, 0.f),
    C_Vertex3f(0.175065637f, 0.0736504793f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172497511f, -0.00156617165f, 0.f),
    C_Vertex3f(0.19133842f, -0.00615417957f, 0.f),
    C_Vertex3f(0.175065637f, 0.0736504793f, 0.f),
    C_Vertex3f(0.166293621f, 0.0802969933f, 0.f),
    C_Vertex3f(0.157963276f, 0.0878930092f, 0.f),
    C_Vertex3f(0.154660463f, 0.00455117226f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157963276f, 0.0878930092f, 0.f),
    C_Vertex3f(0.150387406f, 0.0964246988f, 0.f),
    C_Vertex3f(0.154660463f, 0.00455117226f, 0.f),
    C_Vertex3f(0.143566012f, 0.105892062f, 0.f),
    C_Vertex3f(0.137827277f, 0.0121978521f, 0.f),
    C_Vertex3f(0.137499094f, 0.116295099f, 0.f),
    C_Vertex3f(0.132186651f, 0.12763381f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132186651f, 0.12763381f, 0.f),
    C_Vertex3f(0.127628684f, 0.139908195f, 0.f),
    C_Vertex3f(0.137827277f, 0.0121978521f, 0.f),
    C_Vertex3f(0.123825192f, 0.153118253f, 0.f),
    C_Vertex3f(0.121997952f, 0.021373868f, 0.f),
    C_Vertex3f(0.120776176f, 0.167263985f, 0.f),
    C_Vertex3f(0.119628906f, 0.174804688f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.119628906f, 0.174804688f, 0.f),
    C_Vertex3f(0.037109375f, 0.16796875f, 0.f),
    C_Vertex3f(0.0380926132f, 0.157864571f, 0.f),
    C_Vertex3f(0.0415409803f, 0.138652921f, 0.f),
    C_Vertex3f(0.0464712381f, 0.12043798f, 0.f),
    C_Vertex3f(0.0528833866f, 0.103219748f, 0.f),
    C_Vertex3f(0.0607774258f, 0.0869982243f, 0.f),
    C_Vertex3f(0.0701533556f, 0.0717734098f, 0.f),
    C_Vertex3f(0.0810111761f, 0.0575453043f, 0.f),
    C_Vertex3f(0.0933508873f, 0.0443139076f, 0.f),
    C_Vertex3f(0.107172489f, 0.0320792198f, 0.f),
    C_Vertex3f(0.121997952f, 0.021373868f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.163451433f, 0.398967266f, 0.f),
    C_Vertex3f(0.151716709f, 0.32423687f, 0.f),
    C_Vertex3f(0.165324211f, 0.334222794f, 0.f),
    C_Vertex3f(0.125f, 0.293457031f, 0.f),
    C_Vertex3f(0.11328125f, 0.632324219f, 0.f),
    C_Vertex3f(0.0512695313f, 0.303222656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431640625f, 0.557128906f, -1.f),
    C_Vertex3f(0.431640625f, 0.632324219f, -1.f),
    C_Vertex3f(0.176139832f, 0.557128906f, -1.f),
    C_Vertex3f(0.11328125f, 0.632324219f, -1.f),
    C_Vertex3f(0.141647339f, 0.385177612f, -1.f),
    C_Vertex3f(0.125f, 0.293457031f, -1.f),
    C_Vertex3f(0.129369736f, 0.300388336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163451433f, 0.398967266f, -1.f),
    C_Vertex3f(0.148856163f, 0.390192032f, -1.f),
    C_Vertex3f(0.151716709f, 0.32423687f, -1.f),
    C_Vertex3f(0.141647339f, 0.385177612f, -1.f),
    C_Vertex3f(0.139731884f, 0.312958717f, -1.f),
    C_Vertex3f(0.129369736f, 0.300388336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.461914063f, 0.206673622f, -1.f),
    C_Vertex3f(0.461914063f, 0.228978157f, -1.f),
    C_Vertex3f(0.460329771f, 0.185654402f, -1.f),
    C_Vertex3f(0.460111737f, 0.250791311f, -1.f),
    C_Vertex3f(0.457161188f, 0.165373087f, -1.f),
    C_Vertex3f(0.456507087f, 0.271565676f, -1.f),
    C_Vertex3f(0.452408314f, 0.145829678f, -1.f),
    C_Vertex3f(0.451100111f, 0.29130125f, -1.f),
    C_Vertex3f(0.446071148f, 0.127024174f, -1.f),
    C_Vertex3f(0.44389081f, 0.309998035f, -1.f),
    C_Vertex3f(0.438149691f, 0.108956575f, -1.f),
    C_Vertex3f(0.434879184f, 0.327656031f, -1.f),
    C_Vertex3f(0.428643942f, 0.0916268826f, -1.f),
    C_Vertex3f(0.424065232f, 0.344275236f, -1.f),
    C_Vertex3f(0.417553902f, 0.0750350952f, -1.f),
    C_Vertex3f(0.411448956f, 0.359855652f, -1.f),
    C_Vertex3f(0.403513908f, 0.0573768616f, -1.f),
    C_Vertex3f(0.397030354f, 0.374397278f, -1.f),
    C_Vertex3f(0.386701584f, 0.0403470993f, -1.f),
    C_Vertex3f(0.381792426f, 0.387121201f, -1.f),
    C_Vertex3f(0.378417969f, 0.218957901f, -1.f),
    C_Vertex3f(0.377242565f, 0.234932899f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.381792426f, 0.387121201f, -1.f),
    C_Vertex3f(0.365735173f, 0.39802742f, -1.f),
    C_Vertex3f(0.366663933f, 0.277519226f, -1.f),
    C_Vertex3f(0.371365547f, 0.264213562f, -1.f),
    C_Vertex3f(0.374891758f, 0.25001812f, -1.f),
    C_Vertex3f(0.377242565f, 0.234932899f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.336103439f, 0.321844101f, -1.f),
    C_Vertex3f(0.345506668f, 0.312097549f, -1.f),
    C_Vertex3f(0.348858595f, 0.407115936f, -1.f),
    C_Vertex3f(0.353734493f, 0.30146122f, -1.f),
    C_Vertex3f(0.365735173f, 0.39802742f, -1.f),
    C_Vertex3f(0.360786915f, 0.289935112f, -1.f),
    C_Vertex3f(0.366663933f, 0.277519226f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.331162691f, 0.414386749f, -1.f),
    C_Vertex3f(0.312647462f, 0.419839859f, -1.f),
    C_Vertex3f(0.315054655f, 0.337682247f, -1.f),
    C_Vertex3f(0.325952768f, 0.330372334f, -1.f),
    C_Vertex3f(0.336103439f, 0.321844101f, -1.f),
    C_Vertex3f(0.348858595f, 0.407115936f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.277875662f, 0.352302074f, -1.f),
    C_Vertex3f(0.291016102f, 0.348647118f, -1.f),
    C_Vertex3f(0.293312907f, 0.423475266f, -1.f),
    C_Vertex3f(0.3034091f, 0.343773842f, -1.f),
    C_Vertex3f(0.312647462f, 0.419839859f, -1.f),
    C_Vertex3f(0.315054655f, 0.337682247f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163451433f, 0.398967266f, -1.f),
    C_Vertex3f(0.165324211f, 0.334222794f, -1.f),
    C_Vertex3f(0.178224325f, 0.406488895f, -1.f),
    C_Vertex3f(0.18055439f, 0.342916489f, -1.f),
    C_Vertex3f(0.193174839f, 0.41275692f, -1.f),
    C_Vertex3f(0.196759701f, 0.34943676f, -1.f),
    C_Vertex3f(0.208302975f, 0.417771339f, -1.f),
    C_Vertex3f(0.213940144f, 0.353783607f, -1.f),
    C_Vertex3f(0.223608732f, 0.421532154f, -1.f),
    C_Vertex3f(0.232095718f, 0.355957031f, -1.f),
    C_Vertex3f(0.239092112f, 0.424039364f, -1.f),
    C_Vertex3f(0.249352455f, 0.355957031f, -1.f),
    C_Vertex3f(0.254753113f, 0.425292969f, -1.f),
    C_Vertex3f(0.26398778f, 0.354738712f, -1.f),
    C_Vertex3f(0.273159027f, 0.425292969f, -1.f),
    C_Vertex3f(0.277875662f, 0.352302074f, -1.f),
    C_Vertex3f(0.293312907f, 0.423475266f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.386701584f, 0.0403470993f, -1.f),
    C_Vertex3f(0.378417969f, 0.218957901f, -1.f),
    C_Vertex3f(0.378417969f, 0.201655388f, -1.f),
    C_Vertex3f(0.377194166f, 0.184791207f, -1.f),
    C_Vertex3f(0.374746561f, 0.168803096f, -1.f),
    C_Vertex3f(0.371075153f, 0.153691053f, -1.f),
    C_Vertex3f(0.36848259f, 0.0257501602f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36848259f, 0.0257501602f, -1.f),
    C_Vertex3f(0.371075153f, 0.153691053f, -1.f),
    C_Vertex3f(0.366179943f, 0.13945508f, -1.f),
    C_Vertex3f(0.36006093f, 0.126095176f, -1.f),
    C_Vertex3f(0.352718115f, 0.113611341f, -1.f),
    C_Vertex3f(0.348856926f, 0.0135860443f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.352718115f, 0.113611341f, -1.f),
    C_Vertex3f(0.344151497f, 0.102003574f, -1.f),
    C_Vertex3f(0.348856926f, 0.0135860443f, -1.f),
    C_Vertex3f(0.334361076f, 0.0912718773f, -1.f),
    C_Vertex3f(0.327824593f, 0.00385475159f, -1.f),
    C_Vertex3f(0.324007273f, 0.0818816423f, -1.f),
    C_Vertex3f(0.313090086f, 0.0738328695f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.30538559f, -0.00344371796f, -1.f),
    C_Vertex3f(0.327824593f, 0.00385475159f, -1.f),
    C_Vertex3f(0.313090086f, 0.0738328695f, -1.f),
    C_Vertex3f(0.301609516f, 0.0671255589f, -1.f),
    C_Vertex3f(0.289565563f, 0.0617597103f, -1.f),
    C_Vertex3f(0.281539917f, -0.00830936432f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.289565563f, 0.0617597103f, -1.f),
    C_Vertex3f(0.276958227f, 0.0577353239f, -1.f),
    C_Vertex3f(0.281539917f, -0.00830936432f, -1.f),
    C_Vertex3f(0.263787508f, 0.0550523996f, -1.f),
    C_Vertex3f(0.256287575f, -0.0107421875f, -1.f),
    C_Vertex3f(0.250053406f, 0.0537109375f, -1.f),
    C_Vertex3f(0.236972809f, 0.0537109375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.232031822f, -0.0107421875f, -1.f),
    C_Vertex3f(0.256287575f, -0.0107421875f, -1.f),
    C_Vertex3f(0.236972809f, 0.0537109375f, -1.f),
    C_Vertex3f(0.225550771f, 0.0546604395f, -1.f),
    C_Vertex3f(0.214570403f, 0.0565594435f, -1.f),
    C_Vertex3f(0.21118319f, -0.00921285152f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.214570403f, 0.0565594435f, -1.f),
    C_Vertex3f(0.204031706f, 0.0594079494f, -1.f),
    C_Vertex3f(0.21118319f, -0.00921285152f, -1.f),
    C_Vertex3f(0.193934679f, 0.0632059574f, -1.f),
    C_Vertex3f(0.19133842f, -0.00615417957f, -1.f),
    C_Vertex3f(0.184279323f, 0.0679534674f, -1.f),
    C_Vertex3f(0.175065637f, 0.0736504793f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172497511f, -0.00156617165f, -1.f),
    C_Vertex3f(0.19133842f, -0.00615417957f, -1.f),
    C_Vertex3f(0.175065637f, 0.0736504793f, -1.f),
    C_Vertex3f(0.166293621f, 0.0802969933f, -1.f),
    C_Vertex3f(0.157963276f, 0.0878930092f, -1.f),
    C_Vertex3f(0.154660463f, 0.00455117226f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157963276f, 0.0878930092f, -1.f),
    C_Vertex3f(0.150387406f, 0.0964246988f, -1.f),
    C_Vertex3f(0.154660463f, 0.00455117226f, -1.f),
    C_Vertex3f(0.143566012f, 0.105892062f, -1.f),
    C_Vertex3f(0.137827277f, 0.0121978521f, -1.f),
    C_Vertex3f(0.137499094f, 0.116295099f, -1.f),
    C_Vertex3f(0.132186651f, 0.12763381f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132186651f, 0.12763381f, -1.f),
    C_Vertex3f(0.127628684f, 0.139908195f, -1.f),
    C_Vertex3f(0.137827277f, 0.0121978521f, -1.f),
    C_Vertex3f(0.123825192f, 0.153118253f, -1.f),
    C_Vertex3f(0.121997952f, 0.021373868f, -1.f),
    C_Vertex3f(0.120776176f, 0.167263985f, -1.f),
    C_Vertex3f(0.119628906f, 0.174804688f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.119628906f, 0.174804688f, -1.f),
    C_Vertex3f(0.037109375f, 0.16796875f, -1.f),
    C_Vertex3f(0.0380926132f, 0.157864571f, -1.f),
    C_Vertex3f(0.0415409803f, 0.138652921f, -1.f),
    C_Vertex3f(0.0464712381f, 0.12043798f, -1.f),
    C_Vertex3f(0.0528833866f, 0.103219748f, -1.f),
    C_Vertex3f(0.0607774258f, 0.0869982243f, -1.f),
    C_Vertex3f(0.0701533556f, 0.0717734098f, -1.f),
    C_Vertex3f(0.0810111761f, 0.0575453043f, -1.f),
    C_Vertex3f(0.0933508873f, 0.0443139076f, -1.f),
    C_Vertex3f(0.107172489f, 0.0320792198f, -1.f),
    C_Vertex3f(0.121997952f, 0.021373868f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.163451433f, 0.398967266f, -1.f),
    C_Vertex3f(0.151716709f, 0.32423687f, -1.f),
    C_Vertex3f(0.165324211f, 0.334222794f, -1.f),
    C_Vertex3f(0.125f, 0.293457031f, -1.f),
    C_Vertex3f(0.11328125f, 0.632324219f, -1.f),
    C_Vertex3f(0.0512695313f, 0.303222656f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_54[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.0809831321f, -0.996715486f, 0.f),
    C_Vertex3f(0.445800781f, 0.483398438f, 0.f),
    C_Vertex3f(0.445800781f, 0.483398438f, -1.f),
    C_Vertex3f(0.367675781f, 0.477050781f, -1.f),
    C_Vertex3f(0.367675781f, 0.477050781f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.975490808f, -0.220040321f, 0.f),
    C_Vertex3f(0.367675781f, 0.477050781f, 0.f),
    C_Vertex3f(0.367675781f, 0.477050781f, -1.f),
    C_Normal3f(-0.962286532f, -0.272037834f, 0.f),
    C_Vertex3f(0.365047455f, 0.488702774f, 0.f),
    C_Vertex3f(0.365047455f, 0.488702774f, -1.f),
    C_Normal3f(-0.937081456f, -0.349110663f, 0.f),
    C_Vertex3f(0.35869503f, 0.508818626f, 0.f),
    C_Vertex3f(0.35869503f, 0.508818626f, -1.f),
    C_Normal3f(-0.886681676f, -0.46238035f, 0.f),
    C_Vertex3f(0.351246834f, 0.525746346f, 0.f),
    C_Vertex3f(0.351246834f, 0.525746346f, -1.f),
    C_Normal3f(-0.818156779f, -0.574995279f, 0.f),
    C_Vertex3f(0.342702866f, 0.539485931f, 0.f),
    C_Vertex3f(0.342702866f, 0.539485931f, -1.f),
    C_Normal3f(-0.733194172f, -0.680019438f, 0.f),
    C_Vertex3f(0.337882996f, 0.544761658f, 0.f),
    C_Vertex3f(0.337882996f, 0.544761658f, -1.f),
    C_Normal3f(-0.682577789f, -0.730812907f, 0.f),
    C_Vertex3f(0.333885193f, 0.548993111f, 0.f),
    C_Vertex3f(0.333885193f, 0.548993111f, -1.f),
    C_Normal3f(-0.620348036f, -0.784326613f, 0.f),
    C_Vertex3f(0.325424194f, 0.556398153f, 0.f),
    C_Vertex3f(0.325424194f, 0.556398153f, -1.f),
    C_Normal3f(-0.536202133f, -0.844089627f, 0.f),
    C_Vertex3f(0.316497803f, 0.562745333f, 0.f),
    C_Vertex3f(0.316497803f, 0.562745333f, -1.f),
    C_Normal3f(-0.443345875f, -0.896350682f, 0.f),
    C_Vertex3f(0.307106018f, 0.568034649f, 0.f),
    C_Vertex3f(0.307106018f, 0.568034649f, -1.f),
    C_Normal3f(-0.344492555f, -0.93878907f, 0.f),
    C_Vertex3f(0.29724884f, 0.572266102f, 0.f),
    C_Vertex3f(0.29724884f, 0.572266102f, -1.f),
    C_Normal3f(-0.24304083f, -0.970016122f, 0.f),
    C_Vertex3f(0.28692627f, 0.575439692f, 0.f),
    C_Vertex3f(0.28692627f, 0.575439692f, -1.f),
    C_Normal3f(-0.142513976f, -0.989792824f, 0.f),
    C_Vertex3f(0.276138306f, 0.577555418f, 0.f),
    C_Vertex3f(0.276138306f, 0.577555418f, -1.f),
    C_Normal3f(-0.0398378782f, -0.999206185f, 0.f),
    C_Vertex3f(0.264884949f, 0.578613281f, 0.f),
    C_Vertex3f(0.264884949f, 0.578613281f, -1.f),
    C_Normal3f(0.0792517215f, -0.996854722f, 0.f),
    C_Vertex3f(0.249605179f, 0.578613281f, 0.f),
    C_Vertex3f(0.249605179f, 0.578613281f, -1.f),
    C_Normal3f(0.223530024f, -0.974697053f, 0.f),
    C_Vertex3f(0.231884956f, 0.575989723f, 0.f),
    C_Vertex3f(0.231884956f, 0.575989723f, -1.f),
    C_Normal3f(0.378501058f, -0.925600827f, 0.f),
    C_Vertex3f(0.215285301f, 0.570742607f, 0.f),
    C_Vertex3f(0.215285301f, 0.570742607f, -1.f),
    C_Normal3f(0.529070735f, -0.848577678f, 0.f),
    C_Vertex3f(0.199806213f, 0.562871933f, 0.f),
    C_Vertex3f(0.199806213f, 0.562871933f, -1.f),
    C_Normal3f(0.656264842f, -0.754530609f, 0.f),
    C_Vertex3f(0.183376312f, 0.550848007f, 0.f),
    C_Vertex3f(0.183376312f, 0.550848007f, -1.f),
    C_Normal3f(0.75996846f, -0.649960041f, 0.f),
    C_Vertex3f(0.166817188f, 0.534179211f, 0.f),
    C_Vertex3f(0.166817188f, 0.534179211f, -1.f),
    C_Normal3f(0.842499197f, -0.5386976f, 0.f),
    C_Vertex3f(0.152200222f, 0.514395237f, 0.f),
    C_Vertex3f(0.152200222f, 0.514395237f, -1.f),
    C_Normal3f(0.89846319f, -0.439048886f, 0.f),
    C_Vertex3f(0.139525414f, 0.491496086f, 0.f),
    C_Vertex3f(0.139525414f, 0.491496086f, -1.f),
    C_Normal3f(0.9357723f, -0.352604926f, 0.f),
    C_Vertex3f(0.131475925f, 0.47198534f, 0.f),
    C_Vertex3f(0.131475925f, 0.47198534f, -1.f),
    C_Normal3f(0.959925592f, -0.280255198f, 0.f),
    C_Vertex3f(0.126752973f, 0.457599521f, 0.f),
    C_Vertex3f(0.126752973f, 0.457599521f, -1.f),
    C_Normal3f(0.974896789f, -0.222657278f, 0.f),
    C_Vertex3f(0.122673392f, 0.441835046f, 0.f),
    C_Vertex3f(0.122673392f, 0.441835046f, -1.f),
    C_Normal3f(0.98508811f, -0.172050208f, 0.f),
    C_Vertex3f(0.119237185f, 0.424691916f, 0.f),
    C_Vertex3f(0.119237185f, 0.424691916f, -1.f),
    C_Normal3f(0.991828263f, -0.127580076f, 0.f),
    C_Vertex3f(0.116444349f, 0.40617013f, 0.f),
    C_Vertex3f(0.116444349f, 0.40617013f, -1.f),
    C_Normal3f(0.996083021f, -0.0884233862f, 0.f),
    C_Vertex3f(0.114294887f, 0.386269689f, 0.f),
    C_Vertex3f(0.114294887f, 0.386269689f, -1.f),
    C_Normal3f(0.998549819f, -0.0538357347f, 0.f),
    C_Vertex3f(0.112788796f, 0.364990592f, 0.f),
    C_Vertex3f(0.112788796f, 0.364990592f, -1.f),
    C_Normal3f(0.999607086f, -0.0280311834f, 0.f),
    C_Vertex3f(0.111926079f, 0.34233284f, 0.f),
    C_Vertex3f(0.111926079f, 0.34233284f, -1.f),
    C_Normal3f(0.999958336f, -0.0091251554f, 0.f),
    C_Vertex3f(0.111816406f, 0.330314636f, 0.f),
    C_Vertex3f(0.111816406f, 0.330314636f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.835570157f, 0.5493837f, 0.f),
    C_Vertex3f(0.111816406f, 0.330314636f, 0.f),
    C_Vertex3f(0.111816406f, 0.330314636f, -1.f),
    C_Normal3f(-0.791296482f, 0.611432672f, 0.f),
    C_Vertex3f(0.118927002f, 0.341129303f, 0.f),
    C_Vertex3f(0.118927002f, 0.341129303f, -1.f),
    C_Normal3f(-0.724574983f, 0.68919605f, 0.f),
    C_Vertex3f(0.134733677f, 0.359973431f, 0.f),
    C_Vertex3f(0.134733677f, 0.359973431f, -1.f),
    C_Normal3f(-0.627780378f, 0.778390527f, 0.f),
    C_Vertex3f(0.152125835f, 0.376032352f, 0.f),
    C_Vertex3f(0.152125835f, 0.376032352f, -1.f),
    C_Normal3f(-0.515097558f, 0.857131541f, 0.f),
    C_Vertex3f(0.171103477f, 0.389306068f, 0.f),
    C_Vertex3f(0.171103477f, 0.389306068f, -1.f),
    C_Normal3f(-0.403523296f, 0.914969385f, 0.f),
    C_Vertex3f(0.191666603f, 0.399794579f, 0.f),
    C_Vertex3f(0.191666603f, 0.399794579f, -1.f),
    C_Normal3f(-0.294121206f, 0.955768168f, 0.f),
    C_Vertex3f(0.212722301f, 0.407660961f, 0.f),
    C_Vertex3f(0.212722301f, 0.407660961f, -1.f),
    C_Normal3f(-0.177597836f, 0.984103203f, 0.f),
    C_Vertex3f(0.234270573f, 0.412905216f, 0.f),
    C_Vertex3f(0.234270573f, 0.412905216f, -1.f),
    C_Normal3f(-0.0606525801f, 0.998158991f, 0.f),
    C_Vertex3f(0.256311417f, 0.415527344f, 0.f),
    C_Vertex3f(0.256311417f, 0.415527344f, -1.f),
    C_Normal3f(0.0450858548f, 0.998983085f, 0.f),
    C_Vertex3f(0.277422905f, 0.415527344f, 0.f),
    C_Vertex3f(0.277422905f, 0.415527344f, -1.f),
    C_Normal3f(0.144371927f, 0.98952353f, 0.f),
    C_Vertex3f(0.296380877f, 0.413718939f, 0.f),
    C_Vertex3f(0.296380877f, 0.413718939f, -1.f),
    C_Normal3f(0.245388389f, 0.969424844f, 0.f),
    C_Vertex3f(0.314607263f, 0.410102129f, 0.f),
    C_Vertex3f(0.314607263f, 0.410102129f, -1.f),
    C_Normal3f(0.3466084f, 0.938009977f, 0.f),
    C_Vertex3f(0.33210206f, 0.404676914f, 0.f),
    C_Vertex3f(0.33210206f, 0.404676914f, -1.f),
    C_Normal3f(0.444550753f, 0.895753682f, 0.f),
    C_Vertex3f(0.348865271f, 0.397443295f, 0.f),
    C_Vertex3f(0.348865271f, 0.397443295f, -1.f),
    C_Normal3f(0.535995424f, 0.844220877f, 0.f),
    C_Vertex3f(0.364896894f, 0.38840127f, 0.f),
    C_Vertex3f(0.364896894f, 0.38840127f, -1.f),
    C_Normal3f(0.618490994f, 0.785791874f, 0.f),
    C_Vertex3f(0.380196929f, 0.37755084f, 0.f),
    C_Vertex3f(0.380196929f, 0.37755084f, -1.f),
    C_Normal3f(0.690634489f, 0.723204017f, 0.f),
    C_Vertex3f(0.394765377f, 0.364892006f, 0.f),
    C_Vertex3f(0.394765377f, 0.364892006f, -1.f),
    C_Normal3f(0.75630343f, 0.654221058f, 0.f),
    C_Vertex3f(0.408602238f, 0.350424767f, 0.f),
    C_Vertex3f(0.408602238f, 0.350424767f, -1.f),
    C_Normal3f(0.819121063f, 0.573620737f, 0.f),
    C_Vertex3f(0.420709491f, 0.334899664f, 0.f),
    C_Vertex3f(0.420709491f, 0.334899664f, -1.f),
    C_Normal3f(0.874022186f, 0.485886186f, 0.f),
    C_Vertex3f(0.431087136f, 0.318316698f, 0.f),
    C_Vertex3f(0.431087136f, 0.318316698f, -1.f),
    C_Normal3f(0.919214725f, 0.393756598f, 0.f),
    C_Vertex3f(0.439735174f, 0.300675869f, 0.f),
    C_Vertex3f(0.439735174f, 0.300675869f, -1.f),
    C_Normal3f(0.953842402f, 0.30030781f, 0.f),
    C_Vertex3f(0.446653605f, 0.281977177f, 0.f),
    C_Vertex3f(0.446653605f, 0.281977177f, -1.f),
    C_Normal3f(0.978027701f, 0.208474681f, 0.f),
    C_Vertex3f(0.451842427f, 0.262220621f, 0.f),
    C_Vertex3f(0.451842427f, 0.262220621f, -1.f),
    C_Normal3f(0.992692947f, 0.120667741f, 0.f),
    C_Vertex3f(0.455301642f, 0.241406202f, 0.f),
    C_Vertex3f(0.455301642f, 0.241406202f, -1.f),
    C_Normal3f(0.99936229f, 0.0357089154f, 0.f),
    C_Vertex3f(0.45703125f, 0.21953392f, 0.f),
    C_Vertex3f(0.45703125f, 0.21953392f, -1.f),
    C_Normal3f(0.998293638f, -0.0583938099f, 0.f),
    C_Vertex3f(0.45703125f, 0.193000793f, 0.f),
    C_Vertex3f(0.45703125f, 0.193000793f, -1.f),
    C_Normal3f(0.985700607f, -0.168506458f, 0.f),
    C_Vertex3f(0.453778744f, 0.163929462f, 0.f),
    C_Vertex3f(0.453778744f, 0.163929462f, -1.f),
    C_Normal3f(0.958885789f, -0.283792257f, 0.f),
    C_Vertex3f(0.447273731f, 0.135922909f, 0.f),
    C_Vertex3f(0.447273731f, 0.135922909f, -1.f),
    C_Normal3f(0.918318868f, -0.395841539f, 0.f),
    C_Vertex3f(0.437516212f, 0.108981133f, 0.f),
    C_Vertex3f(0.437516212f, 0.108981133f, -1.f),
    C_Normal3f(0.863543987f, -0.504273474f, 0.f),
    C_Vertex3f(0.424506187f, 0.0831041336f, 0.f),
    C_Vertex3f(0.424506187f, 0.0831041336f, -1.f),
    C_Normal3f(0.788142145f, -0.615493238f, 0.f),
    C_Vertex3f(0.409063339f, 0.060256958f, 0.f),
    C_Vertex3f(0.409063339f, 0.060256958f, -1.f),
    C_Normal3f(0.692020893f, -0.721877515f, 0.f),
    C_Vertex3f(0.391187668f, 0.0404396057f, 0.f),
    C_Vertex3f(0.391187668f, 0.0404396057f, -1.f),
    C_Normal3f(0.578668356f, -0.815563023f, 0.f),
    C_Vertex3f(0.370879173f, 0.0236520767f, 0.f),
    C_Vertex3f(0.370879173f, 0.0236520767f, -1.f),
    C_Normal3f(0.455812365f, -0.890075862f, 0.f),
    C_Vertex3f(0.348137856f, 0.00989437103f, 0.f),
    C_Vertex3f(0.348137856f, 0.00989437103f, -1.f),
    C_Normal3f(0.324801117f, -0.945782244f, 0.f),
    C_Vertex3f(0.323865414f, -0.000423908234f, 0.f),
    C_Vertex3f(0.323865414f, -0.000423908234f, -1.f),
    C_Normal3f(0.190617591f, -0.98166436f, 0.f),
    C_Vertex3f(0.298061848f, -0.00730276108f, 0.f),
    C_Vertex3f(0.298061848f, -0.00730276108f, -1.f),
    C_Normal3f(0.0634856671f, -0.99798274f, 0.f),
    C_Vertex3f(0.270727158f, -0.0107421875f, 0.f),
    C_Vertex3f(0.270727158f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0449524485f, -0.998989165f, 0.f),
    C_Vertex3f(0.243994713f, -0.0107421875f, 0.f),
    C_Vertex3f(0.243994713f, -0.0107421875f, -1.f),
    C_Normal3f(-0.146375075f, -0.989229143f, 0.f),
    C_Vertex3f(0.220529795f, -0.00848340988f, 0.f),
    C_Vertex3f(0.220529795f, -0.00848340988f, -1.f),
    C_Normal3f(-0.251150012f, -0.967948139f, 0.f),
    C_Vertex3f(0.198199034f, -0.00396585464f, 0.f),
    C_Vertex3f(0.198199034f, -0.00396585464f, -1.f),
    C_Normal3f(-0.357846707f, -0.933780313f, 0.f),
    C_Vertex3f(0.17700243f, 0.00281047821f, 0.f),
    C_Vertex3f(0.17700243f, 0.00281047821f, -1.f),
    C_Normal3f(-0.462315828f, -0.886715293f, 0.f),
    C_Vertex3f(0.156939983f, 0.0118455887f, 0.f),
    C_Vertex3f(0.156939983f, 0.0118455887f, -1.f),
    C_Normal3f(-0.560385466f, -0.82823199f, 0.f),
    C_Vertex3f(0.138011694f, 0.0231394768f, 0.f),
    C_Vertex3f(0.138011694f, 0.0231394768f, -1.f),
    C_Normal3f(-0.648649633f, -0.76108712f, 0.f),
    C_Vertex3f(0.120217562f, 0.0366921425f, 0.f),
    C_Vertex3f(0.120217562f, 0.0366921425f, -1.f),
    C_Normal3f(-0.725018561f, -0.688729346f, 0.f),
    C_Vertex3f(0.103557587f, 0.0525035858f, 0.f),
    C_Vertex3f(0.103557587f, 0.0525035858f, -1.f),
    C_Normal3f(-0.801868737f, -0.597500145f, 0.f),
    C_Vertex3f(0.0880317688f, 0.0705738068f, 0.f),
    C_Vertex3f(0.0880317688f, 0.0705738068f, -1.f),
    C_Normal3f(-0.871905208f, -0.489674777f, 0.f),
    C_Vertex3f(0.0744466782f, 0.0915715694f, 0.f),
    C_Vertex3f(0.0744466782f, 0.0915715694f, -1.f),
    C_Normal3f(-0.92184484f, -0.387559265f, 0.f),
    C_Vertex3f(0.0628023148f, 0.115496874f, 0.f),
    C_Vertex3f(0.0628023148f, 0.115496874f, -1.f),
    C_Normal3f(-0.955584586f, -0.294716746f, 0.f),
    C_Vertex3f(0.0530986786f, 0.14234972f, 0.f),
    C_Vertex3f(0.0530986786f, 0.14234972f, -1.f),
    C_Normal3f(-0.977174163f, -0.212439701f, 0.f),
    C_Vertex3f(0.0453357697f, 0.172130108f, 0.f),
    C_Vertex3f(0.0453357697f, 0.172130108f, -1.f),
    C_Normal3f(-0.990070283f, -0.140573651f, 0.f),
    C_Vertex3f(0.039513588f, 0.204838037f, 0.f),
    C_Vertex3f(0.039513588f, 0.204838037f, -1.f),
    C_Normal3f(-0.996935546f, -0.0782272071f, 0.f),
    C_Vertex3f(0.0356321335f, 0.240473509f, 0.f),
    C_Vertex3f(0.0356321335f, 0.240473509f, -1.f),
    C_Normal3f(-0.999723494f, -0.0235142726f, 0.f),
    C_Vertex3f(0.0336914063f, 0.279036522f, 0.f),
    C_Vertex3f(0.0336914063f, 0.279036522f, -1.f),
    C_Normal3f(-0.999697089f, 0.0246117562f, 0.f),
    C_Vertex3f(0.0336914063f, 0.322984695f, 0.f),
    C_Vertex3f(0.0336914063f, 0.322984695f, -1.f),
    C_Normal3f(-0.997026205f, 0.0770637095f, 0.f),
    C_Vertex3f(0.0358378887f, 0.366223812f, 0.f),
    C_Vertex3f(0.0358378887f, 0.366223812f, -1.f),
    C_Normal3f(-0.990420341f, 0.138085827f, 0.f),
    C_Vertex3f(0.0401308537f, 0.406296253f, 0.f),
    C_Vertex3f(0.0401308537f, 0.406296253f, -1.f),
    C_Normal3f(-0.978121221f, 0.208035544f, 0.f),
    C_Vertex3f(0.0465703011f, 0.443202019f, 0.f),
    C_Vertex3f(0.0465703011f, 0.443202019f, -1.f),
    C_Normal3f(-0.957724154f, 0.287687927f, 0.f),
    C_Vertex3f(0.0551562309f, 0.476941109f, 0.f),
    C_Vertex3f(0.0551562309f, 0.476941109f, -1.f),
    C_Normal3f(-0.926144958f, 0.377167821f, 0.f),
    C_Vertex3f(0.0658886433f, 0.507513523f, 0.f),
    C_Vertex3f(0.0658886433f, 0.507513523f, -1.f),
    C_Normal3f(-0.879790843f, 0.47536099f, 0.f),
    C_Vertex3f(0.0787675381f, 0.534919262f, 0.f),
    C_Vertex3f(0.0787675381f, 0.534919262f, -1.f),
    C_Normal3f(-0.816256821f, 0.57768923f, 0.f),
    C_Vertex3f(0.0937929153f, 0.559158325f, 0.f),
    C_Vertex3f(0.0937929153f, 0.559158325f, -1.f),
    C_Normal3f(-0.741472483f, 0.670983195f, 0.f),
    C_Vertex3f(0.109869957f, 0.578866005f, 0.f),
    C_Vertex3f(0.109869957f, 0.578866005f, -1.f),
    C_Normal3f(-0.660738766f, 0.750615895f, 0.f),
    C_Vertex3f(0.126151323f, 0.594916105f, 0.f),
    C_Vertex3f(0.126151323f, 0.594916105f, -1.f),
    C_Normal3f(-0.568903983f, 0.822403967f, 0.f),
    C_Vertex3f(0.143731833f, 0.608673334f, 0.f),
    C_Vertex3f(0.143731833f, 0.608673334f, -1.f),
    C_Normal3f(-0.467142135f, 0.884182274f, 0.f),
    C_Vertex3f(0.162611485f, 0.620137691f, 0.f),
    C_Vertex3f(0.162611485f, 0.620137691f, -1.f),
    C_Normal3f(-0.359530866f, 0.933133185f, 0.f),
    C_Vertex3f(0.182790279f, 0.629309177f, 0.f),
    C_Vertex3f(0.182790279f, 0.629309177f, -1.f),
    C_Normal3f(-0.250774264f, 0.968045533f, 0.f),
    C_Vertex3f(0.204268217f, 0.636187792f, 0.f),
    C_Vertex3f(0.204268217f, 0.636187792f, -1.f),
    C_Normal3f(-0.145254686f, 0.989394248f, 0.f),
    C_Vertex3f(0.227045298f, 0.640773535f, 0.f),
    C_Vertex3f(0.227045298f, 0.640773535f, -1.f),
    C_Normal3f(-0.0495421886f, 0.998771966f, 0.f),
    C_Vertex3f(0.251121521f, 0.643066406f, 0.f),
    C_Vertex3f(0.251121521f, 0.643066406f, -1.f),
    C_Normal3f(0.0329543687f, 0.999456882f, 0.f),
    C_Vertex3f(0.273269653f, 0.643066406f, 0.f),
    C_Vertex3f(0.273269653f, 0.643066406f, -1.f),
    C_Normal3f(0.112047449f, 0.993702948f, 0.f),
    C_Vertex3f(0.29133606f, 0.641740441f, 0.f),
    C_Vertex3f(0.29133606f, 0.641740441f, -1.f),
    C_Normal3f(0.193753928f, 0.981050134f, 0.f),
    C_Vertex3f(0.308547974f, 0.639088511f, 0.f),
    C_Vertex3f(0.308547974f, 0.639088511f, -1.f),
    C_Normal3f(0.27969873f, 0.960087836f, 0.f),
    C_Vertex3f(0.324905396f, 0.635110617f, 0.f),
    C_Vertex3f(0.324905396f, 0.635110617f, -1.f),
    C_Normal3f(0.368015617f, 0.929819643f, 0.f),
    C_Vertex3f(0.340408325f, 0.629806757f, 0.f),
    C_Vertex3f(0.340408325f, 0.629806757f, -1.f),
    C_Normal3f(0.456311136f, 0.889820337f, 0.f),
    C_Vertex3f(0.355056763f, 0.623176932f, 0.f),
    C_Vertex3f(0.355056763f, 0.623176932f, -1.f),
    C_Normal3f(0.541909158f, 0.840437174f, 0.f),
    C_Vertex3f(0.368850708f, 0.615221143f, 0.f),
    C_Vertex3f(0.368850708f, 0.615221143f, -1.f),
    C_Normal3f(0.622211039f, 0.78284961f, 0.f),
    C_Vertex3f(0.381790161f, 0.605939388f, 0.f),
    C_Vertex3f(0.381790161f, 0.605939388f, -1.f),
    C_Normal3f(0.695544779f, 0.718482792f, 0.f),
    C_Vertex3f(0.393875122f, 0.595331669f, 0.f),
    C_Vertex3f(0.393875122f, 0.595331669f, -1.f),
    C_Normal3f(0.763061047f, 0.646326363f, 0.f),
    C_Vertex3f(0.404750347f, 0.58371222f, 0.f),
    C_Vertex3f(0.404750347f, 0.58371222f, -1.f),
    C_Normal3f(0.823195517f, 0.567758143f, 0.f),
    C_Vertex3f(0.414415836f, 0.571081042f, 0.f),
    C_Vertex3f(0.414415836f, 0.571081042f, -1.f),
    C_Normal3f(0.87440753f, 0.48519218f, 0.f),
    C_Vertex3f(0.42287159f, 0.557438135f, 0.f),
    C_Vertex3f(0.42287159f, 0.557438135f, -1.f),
    C_Normal3f(0.915970147f, 0.401246279f, 0.f),
    C_Vertex3f(0.430117607f, 0.542783499f, 0.f),
    C_Vertex3f(0.430117607f, 0.542783499f, -1.f),
    C_Normal3f(0.947965741f, 0.318372548f, 0.f),
    C_Vertex3f(0.436153889f, 0.527117133f, 0.f),
    C_Vertex3f(0.436153889f, 0.527117133f, -1.f),
    C_Normal3f(0.971122622f, 0.238581032f, 0.f),
    C_Vertex3f(0.440980434f, 0.510439038f, 0.f),
    C_Vertex3f(0.440980434f, 0.510439038f, -1.f),
    C_Normal3f(0.984480083f, 0.175496668f, 0.f),
    C_Vertex3f(0.444597244f, 0.492749214f, 0.f),
    C_Vertex3f(0.444597244f, 0.492749214f, -1.f),
    C_Normal3f(0.991818309f, 0.127656788f, 0.f),
    C_Vertex3f(0.445800781f, 0.483398438f, 0.f),
    C_Vertex3f(0.445800781f, 0.483398438f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999371469f, -0.0354490392f, 0.f),
    C_Vertex3f(0.124023438f, 0.215363503f, 0.f),
    C_Vertex3f(0.124023438f, 0.215363503f, -1.f),
    C_Normal3f(0.998379052f, 0.0569146313f, 0.f),
    C_Vertex3f(0.124023438f, 0.197439194f, 0.f),
    C_Vertex3f(0.124023438f, 0.197439194f, -1.f),
    C_Normal3f(0.986759245f, 0.162191764f, 0.f),
    C_Vertex3f(0.126179218f, 0.177547455f, 0.f),
    C_Vertex3f(0.126179218f, 0.177547455f, -1.f),
    C_Normal3f(0.962922275f, 0.269779116f, 0.f),
    C_Vertex3f(0.13049078f, 0.158092499f, 0.f),
    C_Vertex3f(0.13049078f, 0.158092499f, -1.f),
    C_Normal3f(0.928045034f, 0.372467935f, 0.f),
    C_Vertex3f(0.136958122f, 0.139074326f, 0.f),
    C_Vertex3f(0.136958122f, 0.139074326f, -1.f),
    C_Normal3f(0.879080474f, 0.476673335f, 0.f),
    C_Vertex3f(0.145581245f, 0.120492935f, 0.f),
    C_Vertex3f(0.145581245f, 0.120492935f, -1.f),
    C_Normal3f(0.806691349f, 0.59097296f, 0.f),
    C_Vertex3f(0.155909061f, 0.104125023f, 0.f),
    C_Vertex3f(0.155909061f, 0.104125023f, -1.f),
    C_Normal3f(0.71153295f, 0.702652752f, 0.f),
    C_Vertex3f(0.16794157f, 0.0899705887f, 0.f),
    C_Vertex3f(0.16794157f, 0.0899705887f, -1.f),
    C_Normal3f(0.596191645f, 0.80284214f, 0.f),
    C_Vertex3f(0.181678772f, 0.0780296326f, 0.f),
    C_Vertex3f(0.181678772f, 0.0780296326f, -1.f),
    C_Normal3f(0.47764802f, 0.878551245f, 0.f),
    C_Vertex3f(0.197120667f, 0.0683021545f, 0.f),
    C_Vertex3f(0.197120667f, 0.0683021545f, -1.f),
    C_Normal3f(0.35360226f, 0.935395837f, 0.f),
    C_Vertex3f(0.212989807f, 0.061006546f, 0.f),
    C_Vertex3f(0.212989807f, 0.061006546f, -1.f),
    C_Normal3f(0.215741843f, 0.976450384f, 0.f),
    C_Vertex3f(0.229286194f, 0.056142807f, 0.f),
    C_Vertex3f(0.229286194f, 0.056142807f, -1.f),
    C_Normal3f(0.0768462867f, 0.997042954f, 0.f),
    C_Vertex3f(0.246009827f, 0.0537109375f, 0.f),
    C_Vertex3f(0.246009827f, 0.0537109375f, -1.f),
    C_Normal3f(-0.0466981903f, 0.998909056f, 0.f),
    C_Vertex3f(0.260838509f, 0.0537109375f, 0.f),
    C_Vertex3f(0.260838509f, 0.0537109375f, -1.f),
    C_Normal3f(-0.15733555f, 0.987545133f, 0.f),
    C_Vertex3f(0.272896767f, 0.0549678802f, 0.f),
    C_Vertex3f(0.272896767f, 0.0549678802f, -1.f),
    C_Normal3f(-0.266040236f, 0.963961899f, 0.f),
    C_Vertex3f(0.284506798f, 0.0574817657f, 0.f),
    C_Vertex3f(0.284506798f, 0.0574817657f, -1.f),
    C_Normal3f(-0.373161316f, 0.927766562f, 0.f),
    C_Vertex3f(0.295668602f, 0.061252594f, 0.f),
    C_Vertex3f(0.295668602f, 0.061252594f, -1.f),
    C_Normal3f(-0.474624634f, 0.880188227f, 0.f),
    C_Vertex3f(0.306382179f, 0.066280365f, 0.f),
    C_Vertex3f(0.306382179f, 0.066280365f, -1.f),
    C_Normal3f(-0.567076504f, 0.823665082f, 0.f),
    C_Vertex3f(0.31664753f, 0.0725650787f, 0.f),
    C_Vertex3f(0.31664753f, 0.0725650787f, -1.f),
    C_Normal3f(-0.648388445f, 0.761309624f, 0.f),
    C_Vertex3f(0.326464653f, 0.0801067352f, 0.f),
    C_Vertex3f(0.326464653f, 0.0801067352f, -1.f),
    C_Normal3f(-0.717771709f, 0.696278453f, 0.f),
    C_Vertex3f(0.335833549f, 0.0889053345f, 0.f),
    C_Vertex3f(0.335833549f, 0.0889053345f, -1.f),
    C_Normal3f(-0.782391667f, 0.62278676f, 0.f),
    C_Vertex3f(0.344754219f, 0.0989608765f, 0.f),
    C_Vertex3f(0.344754219f, 0.0989608765f, -1.f),
    C_Normal3f(-0.844051719f, 0.536261857f, 0.f),
    C_Vertex3f(0.352559805f, 0.109918118f, 0.f),
    C_Vertex3f(0.352559805f, 0.109918118f, -1.f),
    C_Normal3f(-0.895064235f, 0.445937216f, 0.f),
    C_Vertex3f(0.359250307f, 0.121777058f, 0.f),
    C_Vertex3f(0.359250307f, 0.121777058f, -1.f),
    C_Normal3f(-0.93484205f, 0.355063856f, 0.f),
    C_Vertex3f(0.364825726f, 0.134537697f, 0.f),
    C_Vertex3f(0.364825726f, 0.134537697f, -1.f),
    C_Normal3f(-0.96382612f, 0.266531825f, 0.f),
    C_Vertex3f(0.36928606f, 0.148200035f, 0.f),
    C_Vertex3f(0.36928606f, 0.148200035f, -1.f),
    C_Normal3f(-0.983197868f, 0.182543486f, 0.f),
    C_Vertex3f(0.372631311f, 0.162764072f, 0.f),
    C_Vertex3f(0.372631311f, 0.162764072f, -1.f),
    C_Normal3f(-0.994523644f, 0.104511462f, 0.f),
    C_Vertex3f(0.374861479f, 0.178229809f, 0.f),
    C_Vertex3f(0.374861479f, 0.178229809f, -1.f),
    C_Normal3f(-0.999440134f, 0.0334587023f, 0.f),
    C_Vertex3f(0.375976563f, 0.194597244f, 0.f),
    C_Vertex3f(0.375976563f, 0.194597244f, -1.f),
    C_Normal3f(-0.999432206f, -0.0336936414f, 0.f),
    C_Vertex3f(0.375976563f, 0.211538315f, 0.f),
    C_Vertex3f(0.375976563f, 0.211538315f, -1.f),
    C_Normal3f(-0.994213104f, -0.107426353f, 0.f),
    C_Vertex3f(0.374875069f, 0.227270126f, 0.f),
    C_Vertex3f(0.374875069f, 0.227270126f, -1.f),
    C_Normal3f(-0.982225955f, -0.187702015f, 0.f),
    C_Vertex3f(0.372672081f, 0.242120743f, 0.f),
    C_Vertex3f(0.372672081f, 0.242120743f, -1.f),
    C_Normal3f(-0.961714268f, -0.274053991f, 0.f),
    C_Vertex3f(0.369367599f, 0.256090164f, 0.f),
    C_Vertex3f(0.369367599f, 0.256090164f, -1.f),
    C_Normal3f(-0.931051493f, -0.364887655f, 0.f),
    C_Vertex3f(0.364961624f, 0.269178391f, 0.f),
    C_Vertex3f(0.364961624f, 0.269178391f, -1.f),
    C_Normal3f(-0.889075398f, -0.457760692f, 0.f),
    C_Vertex3f(0.359454155f, 0.281385422f, 0.f),
    C_Vertex3f(0.359454155f, 0.281385422f, -1.f),
    C_Normal3f(-0.83547461f, -0.549529135f, 0.f),
    C_Vertex3f(0.352845192f, 0.292711258f, 0.f),
    C_Vertex3f(0.352845192f, 0.292711258f, -1.f),
    C_Normal3f(-0.771075428f, -0.636743665f, 0.f),
    C_Vertex3f(0.345134735f, 0.303155899f, 0.f),
    C_Vertex3f(0.345134735f, 0.303155899f, -1.f),
    C_Normal3f(-0.70166564f, -0.712506413f, 0.f),
    C_Vertex3f(0.336322784f, 0.312719345f, 0.f),
    C_Vertex3f(0.336322784f, 0.312719345f, -1.f),
    C_Normal3f(-0.625638008f, -0.780113459f, 0.f),
    C_Vertex3f(0.326926231f, 0.32108736f, 0.f),
    C_Vertex3f(0.326926231f, 0.32108736f, -1.f),
    C_Normal3f(-0.539045393f, -0.842276692f, 0.f),
    C_Vertex3f(0.316945076f, 0.328259945f, 0.f),
    C_Vertex3f(0.316945076f, 0.328259945f, -1.f),
    C_Normal3f(-0.443936616f, -0.896058202f, 0.f),
    C_Vertex3f(0.306379318f, 0.334237099f, 0.f),
    C_Vertex3f(0.306379318f, 0.334237099f, -1.f),
    C_Normal3f(-0.343412727f, -0.939184606f, 0.f),
    C_Vertex3f(0.295228958f, 0.339018822f, 0.f),
    C_Vertex3f(0.295228958f, 0.339018822f, -1.f),
    C_Normal3f(-0.241150215f, -0.970487714f, 0.f),
    C_Vertex3f(0.283493996f, 0.342605114f, 0.f),
    C_Vertex3f(0.283493996f, 0.342605114f, -1.f),
    C_Normal3f(-0.14076364f, -0.990043223f, 0.f),
    C_Vertex3f(0.271174431f, 0.344995975f, 0.f),
    C_Vertex3f(0.271174431f, 0.344995975f, -1.f),
    C_Normal3f(-0.0453945212f, -0.998969138f, 0.f),
    C_Vertex3f(0.258270264f, 0.346191406f, 0.f),
    C_Vertex3f(0.258270264f, 0.346191406f, -1.f),
    C_Normal3f(0.0455531366f, -0.998961985f, 0.f),
    C_Vertex3f(0.244867325f, 0.346191406f, 0.f),
    C_Vertex3f(0.244867325f, 0.346191406f, -1.f),
    C_Normal3f(0.141333222f, -0.989962041f, 0.f),
    C_Vertex3f(0.232054949f, 0.344995975f, 0.f),
    C_Vertex3f(0.232054949f, 0.344995975f, -1.f),
    C_Normal3f(0.240622327f, -0.970618844f, 0.f),
    C_Vertex3f(0.219747305f, 0.342605114f, 0.f),
    C_Vertex3f(0.219747305f, 0.342605114f, -1.f),
    C_Normal3f(0.340578765f, -0.940215945f, 0.f),
    C_Vertex3f(0.207944393f, 0.339018822f, 0.f),
    C_Vertex3f(0.207944393f, 0.339018822f, -1.f),
    C_Normal3f(0.437847376f, -0.899049342f, 0.f),
    C_Vertex3f(0.196646214f, 0.334237099f, 0.f),
    C_Vertex3f(0.196646214f, 0.334237099f, -1.f),
    C_Normal3f(0.529228508f, -0.84847939f, 0.f),
    C_Vertex3f(0.185852766f, 0.328259945f, 0.f),
    C_Vertex3f(0.185852766f, 0.328259945f, -1.f),
    C_Normal3f(0.612184167f, -0.790715218f, 0.f),
    C_Vertex3f(0.175564051f, 0.32108736f, 0.f),
    C_Vertex3f(0.175564051f, 0.32108736f, -1.f),
    C_Normal3f(0.685153782f, -0.728398383f, 0.f),
    C_Vertex3f(0.165780067f, 0.312719345f, 0.f),
    C_Vertex3f(0.165780067f, 0.312719345f, -1.f),
    C_Normal3f(0.752369165f, -0.658741653f, 0.f),
    C_Vertex3f(0.156500816f, 0.303155899f, 0.f),
    C_Vertex3f(0.156500816f, 0.303155899f, -1.f),
    C_Normal3f(0.816955507f, -0.576700747f, 0.f),
    C_Vertex3f(0.148381472f, 0.292847872f, 0.f),
    C_Vertex3f(0.148381472f, 0.292847872f, -1.f),
    C_Normal3f(0.873106718f, -0.487529159f, 0.f),
    C_Vertex3f(0.141422033f, 0.281795263f, 0.f),
    C_Vertex3f(0.141422033f, 0.281795263f, -1.f),
    C_Normal3f(0.919033289f, -0.394180059f, 0.f),
    C_Vertex3f(0.135622501f, 0.269998074f, 0.f),
    C_Vertex3f(0.135622501f, 0.269998074f, -1.f),
    C_Normal3f(0.953973293f, -0.299891621f, 0.f),
    C_Vertex3f(0.130982876f, 0.257456303f, 0.f),
    C_Vertex3f(0.130982876f, 0.257456303f, -1.f),
    C_Normal3f(0.978198051f, -0.207674041f, 0.f),
    C_Vertex3f(0.127503157f, 0.24416995f, 0.f),
    C_Vertex3f(0.127503157f, 0.24416995f, -1.f),
    C_Normal3f(0.99278301f, -0.119924754f, 0.f),
    C_Vertex3f(0.125183344f, 0.230139017f, 0.f),
    C_Vertex3f(0.125183344f, 0.230139017f, -1.f),
    C_Normal3f(0.999371469f, -0.0354490392f, 0.f),
    C_Vertex3f(0.124023438f, 0.215363503f, 0.f),
    C_Vertex3f(0.124023438f, 0.215363503f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.380196929f, 0.37755084f, 0.f),
    C_Vertex3f(0.364896894f, 0.38840127f, 0.f),
    C_Vertex3f(0.364961624f, 0.269178391f, 0.f),
    C_Vertex3f(0.369367599f, 0.256090164f, 0.f),
    C_Vertex3f(0.372672081f, 0.242120743f, 0.f),
    C_Vertex3f(0.374875069f, 0.227270126f, 0.f),
    C_Vertex3f(0.375976563f, 0.211538315f, 0.f),
    C_Vertex3f(0.391187668f, 0.0404396057f, 0.f),
    C_Vertex3f(0.394765377f, 0.364892006f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.336322784f, 0.312719345f, 0.f),
    C_Vertex3f(0.345134735f, 0.303155899f, 0.f),
    C_Vertex3f(0.348865271f, 0.397443295f, 0.f),
    C_Vertex3f(0.352845192f, 0.292711258f, 0.f),
    C_Vertex3f(0.364896894f, 0.38840127f, 0.f),
    C_Vertex3f(0.359454155f, 0.281385422f, 0.f),
    C_Vertex3f(0.364961624f, 0.269178391f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33210206f, 0.404676914f, 0.f),
    C_Vertex3f(0.314607263f, 0.410102129f, 0.f),
    C_Vertex3f(0.316945076f, 0.328259945f, 0.f),
    C_Vertex3f(0.326926231f, 0.32108736f, 0.f),
    C_Vertex3f(0.336322784f, 0.312719345f, 0.f),
    C_Vertex3f(0.348865271f, 0.397443295f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.283493996f, 0.342605114f, 0.f),
    C_Vertex3f(0.295228958f, 0.339018822f, 0.f),
    C_Vertex3f(0.296380877f, 0.413718939f, 0.f),
    C_Vertex3f(0.306379318f, 0.334237099f, 0.f),
    C_Vertex3f(0.314607263f, 0.410102129f, 0.f),
    C_Vertex3f(0.316945076f, 0.328259945f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.277422905f, 0.415527344f, 0.f),
    C_Vertex3f(0.256311417f, 0.415527344f, 0.f),
    C_Vertex3f(0.258270264f, 0.346191406f, 0.f),
    C_Vertex3f(0.271174431f, 0.344995975f, 0.f),
    C_Vertex3f(0.283493996f, 0.342605114f, 0.f),
    C_Vertex3f(0.296380877f, 0.413718939f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.219747305f, 0.342605114f, 0.f),
    C_Vertex3f(0.232054949f, 0.344995975f, 0.f),
    C_Vertex3f(0.234270573f, 0.412905216f, 0.f),
    C_Vertex3f(0.244867325f, 0.346191406f, 0.f),
    C_Vertex3f(0.256311417f, 0.415527344f, 0.f),
    C_Vertex3f(0.258270264f, 0.346191406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.212722301f, 0.407660961f, 0.f),
    C_Vertex3f(0.191666603f, 0.399794579f, 0.f),
    C_Vertex3f(0.196646214f, 0.334237099f, 0.f),
    C_Vertex3f(0.207944393f, 0.339018822f, 0.f),
    C_Vertex3f(0.219747305f, 0.342605114f, 0.f),
    C_Vertex3f(0.234270573f, 0.412905216f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.156500816f, 0.303155899f, 0.f),
    C_Vertex3f(0.165780067f, 0.312719345f, 0.f),
    C_Vertex3f(0.171103477f, 0.389306068f, 0.f),
    C_Vertex3f(0.175564051f, 0.32108736f, 0.f),
    C_Vertex3f(0.191666603f, 0.399794579f, 0.f),
    C_Vertex3f(0.185852766f, 0.328259945f, 0.f),
    C_Vertex3f(0.196646214f, 0.334237099f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.152125835f, 0.376032352f, 0.f),
    C_Vertex3f(0.134733677f, 0.359973431f, 0.f),
    C_Vertex3f(0.135622501f, 0.269998074f, 0.f),
    C_Vertex3f(0.141422033f, 0.281795263f, 0.f),
    C_Vertex3f(0.148381472f, 0.292847872f, 0.f),
    C_Vertex3f(0.156500816f, 0.303155899f, 0.f),
    C_Vertex3f(0.171103477f, 0.389306068f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134733677f, 0.359973431f, 0.f),
    C_Vertex3f(0.118927002f, 0.341129303f, 0.f),
    C_Vertex3f(0.124023438f, 0.215363503f, 0.f),
    C_Vertex3f(0.125183344f, 0.230139017f, 0.f),
    C_Vertex3f(0.127503157f, 0.24416995f, 0.f),
    C_Vertex3f(0.130982876f, 0.257456303f, 0.f),
    C_Vertex3f(0.135622501f, 0.269998074f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120217562f, 0.0366921425f, 0.f),
    C_Vertex3f(0.138011694f, 0.0231394768f, 0.f),
    C_Vertex3f(0.124023438f, 0.197439194f, 0.f),
    C_Vertex3f(0.124023438f, 0.215363503f, 0.f),
    C_Vertex3f(0.118927002f, 0.341129303f, 0.f),
    C_Vertex3f(0.111816406f, 0.330314636f, 0.f),
    C_Vertex3f(0.103557587f, 0.0525035858f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.330314636f, 0.f),
    C_Vertex3f(0.109869957f, 0.578866005f, 0.f),
    C_Vertex3f(0.103557587f, 0.0525035858f, 0.f),
    C_Vertex3f(0.0937929153f, 0.559158325f, 0.f),
    C_Vertex3f(0.0880317688f, 0.0705738068f, 0.f),
    C_Vertex3f(0.0787675381f, 0.534919262f, 0.f),
    C_Vertex3f(0.0744466782f, 0.0915715694f, 0.f),
    C_Vertex3f(0.0658886433f, 0.507513523f, 0.f),
    C_Vertex3f(0.0628023148f, 0.115496874f, 0.f),
    C_Vertex3f(0.0551562309f, 0.476941109f, 0.f),
    C_Vertex3f(0.0530986786f, 0.14234972f, 0.f),
    C_Vertex3f(0.0465703011f, 0.443202019f, 0.f),
    C_Vertex3f(0.0453357697f, 0.172130108f, 0.f),
    C_Vertex3f(0.0401308537f, 0.406296253f, 0.f),
    C_Vertex3f(0.039513588f, 0.204838037f, 0.f),
    C_Vertex3f(0.0358378887f, 0.366223812f, 0.f),
    C_Vertex3f(0.0356321335f, 0.240473509f, 0.f),
    C_Vertex3f(0.0336914063f, 0.322984695f, 0.f),
    C_Vertex3f(0.0336914063f, 0.279036522f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446653605f, 0.281977177f, 0.f),
    C_Vertex3f(0.447273731f, 0.135922909f, 0.f),
    C_Vertex3f(0.451842427f, 0.262220621f, 0.f),
    C_Vertex3f(0.453778744f, 0.163929462f, 0.f),
    C_Vertex3f(0.455301642f, 0.241406202f, 0.f),
    C_Vertex3f(0.45703125f, 0.193000793f, 0.f),
    C_Vertex3f(0.45703125f, 0.21953392f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.408602238f, 0.350424767f, 0.f),
    C_Vertex3f(0.409063339f, 0.060256958f, 0.f),
    C_Vertex3f(0.420709491f, 0.334899664f, 0.f),
    C_Vertex3f(0.424506187f, 0.0831041336f, 0.f),
    C_Vertex3f(0.431087136f, 0.318316698f, 0.f),
    C_Vertex3f(0.437516212f, 0.108981133f, 0.f),
    C_Vertex3f(0.439735174f, 0.300675869f, 0.f),
    C_Vertex3f(0.447273731f, 0.135922909f, 0.f),
    C_Vertex3f(0.446653605f, 0.281977177f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.394765377f, 0.364892006f, 0.f),
    C_Vertex3f(0.391187668f, 0.0404396057f, 0.f),
    C_Vertex3f(0.409063339f, 0.060256958f, 0.f),
    C_Vertex3f(0.408602238f, 0.350424767f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.391187668f, 0.0404396057f, 0.f),
    C_Vertex3f(0.375976563f, 0.211538315f, 0.f),
    C_Vertex3f(0.375976563f, 0.194597244f, 0.f),
    C_Vertex3f(0.374861479f, 0.178229809f, 0.f),
    C_Vertex3f(0.372631311f, 0.162764072f, 0.f),
    C_Vertex3f(0.370879173f, 0.0236520767f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.370879173f, 0.0236520767f, 0.f),
    C_Vertex3f(0.372631311f, 0.162764072f, 0.f),
    C_Vertex3f(0.36928606f, 0.148200035f, 0.f),
    C_Vertex3f(0.364825726f, 0.134537697f, 0.f),
    C_Vertex3f(0.359250307f, 0.121777058f, 0.f),
    C_Vertex3f(0.352559805f, 0.109918118f, 0.f),
    C_Vertex3f(0.348137856f, 0.00989437103f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.348137856f, 0.00989437103f, 0.f),
    C_Vertex3f(0.352559805f, 0.109918118f, 0.f),
    C_Vertex3f(0.344754219f, 0.0989608765f, 0.f),
    C_Vertex3f(0.335833549f, 0.0889053345f, 0.f),
    C_Vertex3f(0.326464653f, 0.0801067352f, 0.f),
    C_Vertex3f(0.323865414f, -0.000423908234f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.326464653f, 0.0801067352f, 0.f),
    C_Vertex3f(0.31664753f, 0.0725650787f, 0.f),
    C_Vertex3f(0.323865414f, -0.000423908234f, 0.f),
    C_Vertex3f(0.306382179f, 0.066280365f, 0.f),
    C_Vertex3f(0.298061848f, -0.00730276108f, 0.f),
    C_Vertex3f(0.295668602f, 0.061252594f, 0.f),
    C_Vertex3f(0.284506798f, 0.0574817657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.272896767f, 0.0549678802f, 0.f),
    C_Vertex3f(0.260838509f, 0.0537109375f, 0.f),
    C_Vertex3f(0.270727158f, -0.0107421875f, 0.f),
    C_Vertex3f(0.298061848f, -0.00730276108f, 0.f),
    C_Vertex3f(0.284506798f, 0.0574817657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.197120667f, 0.0683021545f, 0.f),
    C_Vertex3f(0.198199034f, -0.00396585464f, 0.f),
    C_Vertex3f(0.212989807f, 0.061006546f, 0.f),
    C_Vertex3f(0.220529795f, -0.00848340988f, 0.f),
    C_Vertex3f(0.229286194f, 0.056142807f, 0.f),
    C_Vertex3f(0.243994713f, -0.0107421875f, 0.f),
    C_Vertex3f(0.246009827f, 0.0537109375f, 0.f),
    C_Vertex3f(0.270727158f, -0.0107421875f, 0.f),
    C_Vertex3f(0.260838509f, 0.0537109375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.155909061f, 0.104125023f, 0.f),
    C_Vertex3f(0.156939983f, 0.0118455887f, 0.f),
    C_Vertex3f(0.16794157f, 0.0899705887f, 0.f),
    C_Vertex3f(0.17700243f, 0.00281047821f, 0.f),
    C_Vertex3f(0.181678772f, 0.0780296326f, 0.f),
    C_Vertex3f(0.198199034f, -0.00396585464f, 0.f),
    C_Vertex3f(0.197120667f, 0.0683021545f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145581245f, 0.120492935f, 0.f),
    C_Vertex3f(0.136958122f, 0.139074326f, 0.f),
    C_Vertex3f(0.138011694f, 0.0231394768f, 0.f),
    C_Vertex3f(0.156939983f, 0.0118455887f, 0.f),
    C_Vertex3f(0.155909061f, 0.104125023f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138011694f, 0.0231394768f, 0.f),
    C_Vertex3f(0.136958122f, 0.139074326f, 0.f),
    C_Vertex3f(0.13049078f, 0.158092499f, 0.f),
    C_Vertex3f(0.126179218f, 0.177547455f, 0.f),
    C_Vertex3f(0.124023438f, 0.197439194f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.445800781f, 0.483398438f, 0.f),
    C_Vertex3f(0.444597244f, 0.492749214f, 0.f),
    C_Vertex3f(0.440980434f, 0.510439038f, 0.f),
    C_Vertex3f(0.436153889f, 0.527117133f, 0.f),
    C_Vertex3f(0.430117607f, 0.542783499f, 0.f),
    C_Vertex3f(0.42287159f, 0.557438135f, 0.f),
    C_Vertex3f(0.414415836f, 0.571081042f, 0.f),
    C_Vertex3f(0.404750347f, 0.58371222f, 0.f),
    C_Vertex3f(0.393875122f, 0.595331669f, 0.f),
    C_Vertex3f(0.381790161f, 0.605939388f, 0.f),
    C_Vertex3f(0.368850708f, 0.615221143f, 0.f),
    C_Vertex3f(0.367675781f, 0.477050781f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.342702866f, 0.539485931f, 0.f),
    C_Vertex3f(0.351246834f, 0.525746346f, 0.f),
    C_Vertex3f(0.355056763f, 0.623176932f, 0.f),
    C_Vertex3f(0.35869503f, 0.508818626f, 0.f),
    C_Vertex3f(0.368850708f, 0.615221143f, 0.f),
    C_Vertex3f(0.365047455f, 0.488702774f, 0.f),
    C_Vertex3f(0.367675781f, 0.477050781f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.340408325f, 0.629806757f, 0.f),
    C_Vertex3f(0.324905396f, 0.635110617f, 0.f),
    C_Vertex3f(0.325424194f, 0.556398153f, 0.f),
    C_Vertex3f(0.333885193f, 0.548993111f, 0.f),
    C_Vertex3f(0.337882996f, 0.544761658f, 0.f),
    C_Vertex3f(0.342702866f, 0.539485931f, 0.f),
    C_Vertex3f(0.355056763f, 0.623176932f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.29724884f, 0.572266102f, 0.f),
    C_Vertex3f(0.307106018f, 0.568034649f, 0.f),
    C_Vertex3f(0.308547974f, 0.639088511f, 0.f),
    C_Vertex3f(0.316497803f, 0.562745333f, 0.f),
    C_Vertex3f(0.324905396f, 0.635110617f, 0.f),
    C_Vertex3f(0.325424194f, 0.556398153f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.29133606f, 0.641740441f, 0.f),
    C_Vertex3f(0.273269653f, 0.643066406f, 0.f),
    C_Vertex3f(0.276138306f, 0.577555418f, 0.f),
    C_Vertex3f(0.28692627f, 0.575439692f, 0.f),
    C_Vertex3f(0.29724884f, 0.572266102f, 0.f),
    C_Vertex3f(0.308547974f, 0.639088511f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.231884956f, 0.575989723f, 0.f),
    C_Vertex3f(0.249605179f, 0.578613281f, 0.f),
    C_Vertex3f(0.251121521f, 0.643066406f, 0.f),
    C_Vertex3f(0.264884949f, 0.578613281f, 0.f),
    C_Vertex3f(0.273269653f, 0.643066406f, 0.f),
    C_Vertex3f(0.276138306f, 0.577555418f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.183376312f, 0.550848007f, 0.f),
    C_Vertex3f(0.199806213f, 0.562871933f, 0.f),
    C_Vertex3f(0.204268217f, 0.636187792f, 0.f),
    C_Vertex3f(0.215285301f, 0.570742607f, 0.f),
    C_Vertex3f(0.227045298f, 0.640773535f, 0.f),
    C_Vertex3f(0.231884956f, 0.575989723f, 0.f),
    C_Vertex3f(0.251121521f, 0.643066406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.131475925f, 0.47198534f, 0.f),
    C_Vertex3f(0.139525414f, 0.491496086f, 0.f),
    C_Vertex3f(0.143731833f, 0.608673334f, 0.f),
    C_Vertex3f(0.152200222f, 0.514395237f, 0.f),
    C_Vertex3f(0.162611485f, 0.620137691f, 0.f),
    C_Vertex3f(0.166817188f, 0.534179211f, 0.f),
    C_Vertex3f(0.182790279f, 0.629309177f, 0.f),
    C_Vertex3f(0.183376312f, 0.550848007f, 0.f),
    C_Vertex3f(0.204268217f, 0.636187792f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.119237185f, 0.424691916f, 0.f),
    C_Vertex3f(0.122673392f, 0.441835046f, 0.f),
    C_Vertex3f(0.126151323f, 0.594916105f, 0.f),
    C_Vertex3f(0.126752973f, 0.457599521f, 0.f),
    C_Vertex3f(0.143731833f, 0.608673334f, 0.f),
    C_Vertex3f(0.131475925f, 0.47198534f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.126151323f, 0.594916105f, 0.f),
    C_Vertex3f(0.109869957f, 0.578866005f, 0.f),
    C_Vertex3f(0.112788796f, 0.364990592f, 0.f),
    C_Vertex3f(0.114294887f, 0.386269689f, 0.f),
    C_Vertex3f(0.116444349f, 0.40617013f, 0.f),
    C_Vertex3f(0.119237185f, 0.424691916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.111926079f, 0.34233284f, 0.f),
    C_Vertex3f(0.112788796f, 0.364990592f, 0.f),
    C_Vertex3f(0.109869957f, 0.578866005f, 0.f),
    C_Vertex3f(0.111816406f, 0.330314636f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.380196929f, 0.37755084f, -1.f),
    C_Vertex3f(0.364896894f, 0.38840127f, -1.f),
    C_Vertex3f(0.364961624f, 0.269178391f, -1.f),
    C_Vertex3f(0.369367599f, 0.256090164f, -1.f),
    C_Vertex3f(0.372672081f, 0.242120743f, -1.f),
    C_Vertex3f(0.374875069f, 0.227270126f, -1.f),
    C_Vertex3f(0.375976563f, 0.211538315f, -1.f),
    C_Vertex3f(0.391187668f, 0.0404396057f, -1.f),
    C_Vertex3f(0.394765377f, 0.364892006f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.336322784f, 0.312719345f, -1.f),
    C_Vertex3f(0.345134735f, 0.303155899f, -1.f),
    C_Vertex3f(0.348865271f, 0.397443295f, -1.f),
    C_Vertex3f(0.352845192f, 0.292711258f, -1.f),
    C_Vertex3f(0.364896894f, 0.38840127f, -1.f),
    C_Vertex3f(0.359454155f, 0.281385422f, -1.f),
    C_Vertex3f(0.364961624f, 0.269178391f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33210206f, 0.404676914f, -1.f),
    C_Vertex3f(0.314607263f, 0.410102129f, -1.f),
    C_Vertex3f(0.316945076f, 0.328259945f, -1.f),
    C_Vertex3f(0.326926231f, 0.32108736f, -1.f),
    C_Vertex3f(0.336322784f, 0.312719345f, -1.f),
    C_Vertex3f(0.348865271f, 0.397443295f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.283493996f, 0.342605114f, -1.f),
    C_Vertex3f(0.295228958f, 0.339018822f, -1.f),
    C_Vertex3f(0.296380877f, 0.413718939f, -1.f),
    C_Vertex3f(0.306379318f, 0.334237099f, -1.f),
    C_Vertex3f(0.314607263f, 0.410102129f, -1.f),
    C_Vertex3f(0.316945076f, 0.328259945f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.277422905f, 0.415527344f, -1.f),
    C_Vertex3f(0.256311417f, 0.415527344f, -1.f),
    C_Vertex3f(0.258270264f, 0.346191406f, -1.f),
    C_Vertex3f(0.271174431f, 0.344995975f, -1.f),
    C_Vertex3f(0.283493996f, 0.342605114f, -1.f),
    C_Vertex3f(0.296380877f, 0.413718939f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.219747305f, 0.342605114f, -1.f),
    C_Vertex3f(0.232054949f, 0.344995975f, -1.f),
    C_Vertex3f(0.234270573f, 0.412905216f, -1.f),
    C_Vertex3f(0.244867325f, 0.346191406f, -1.f),
    C_Vertex3f(0.256311417f, 0.415527344f, -1.f),
    C_Vertex3f(0.258270264f, 0.346191406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.212722301f, 0.407660961f, -1.f),
    C_Vertex3f(0.191666603f, 0.399794579f, -1.f),
    C_Vertex3f(0.196646214f, 0.334237099f, -1.f),
    C_Vertex3f(0.207944393f, 0.339018822f, -1.f),
    C_Vertex3f(0.219747305f, 0.342605114f, -1.f),
    C_Vertex3f(0.234270573f, 0.412905216f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.156500816f, 0.303155899f, -1.f),
    C_Vertex3f(0.165780067f, 0.312719345f, -1.f),
    C_Vertex3f(0.171103477f, 0.389306068f, -1.f),
    C_Vertex3f(0.175564051f, 0.32108736f, -1.f),
    C_Vertex3f(0.191666603f, 0.399794579f, -1.f),
    C_Vertex3f(0.185852766f, 0.328259945f, -1.f),
    C_Vertex3f(0.196646214f, 0.334237099f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.152125835f, 0.376032352f, -1.f),
    C_Vertex3f(0.134733677f, 0.359973431f, -1.f),
    C_Vertex3f(0.135622501f, 0.269998074f, -1.f),
    C_Vertex3f(0.141422033f, 0.281795263f, -1.f),
    C_Vertex3f(0.148381472f, 0.292847872f, -1.f),
    C_Vertex3f(0.156500816f, 0.303155899f, -1.f),
    C_Vertex3f(0.171103477f, 0.389306068f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134733677f, 0.359973431f, -1.f),
    C_Vertex3f(0.118927002f, 0.341129303f, -1.f),
    C_Vertex3f(0.124023438f, 0.215363503f, -1.f),
    C_Vertex3f(0.125183344f, 0.230139017f, -1.f),
    C_Vertex3f(0.127503157f, 0.24416995f, -1.f),
    C_Vertex3f(0.130982876f, 0.257456303f, -1.f),
    C_Vertex3f(0.135622501f, 0.269998074f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120217562f, 0.0366921425f, -1.f),
    C_Vertex3f(0.138011694f, 0.0231394768f, -1.f),
    C_Vertex3f(0.124023438f, 0.197439194f, -1.f),
    C_Vertex3f(0.124023438f, 0.215363503f, -1.f),
    C_Vertex3f(0.118927002f, 0.341129303f, -1.f),
    C_Vertex3f(0.111816406f, 0.330314636f, -1.f),
    C_Vertex3f(0.103557587f, 0.0525035858f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.330314636f, -1.f),
    C_Vertex3f(0.109869957f, 0.578866005f, -1.f),
    C_Vertex3f(0.103557587f, 0.0525035858f, -1.f),
    C_Vertex3f(0.0937929153f, 0.559158325f, -1.f),
    C_Vertex3f(0.0880317688f, 0.0705738068f, -1.f),
    C_Vertex3f(0.0787675381f, 0.534919262f, -1.f),
    C_Vertex3f(0.0744466782f, 0.0915715694f, -1.f),
    C_Vertex3f(0.0658886433f, 0.507513523f, -1.f),
    C_Vertex3f(0.0628023148f, 0.115496874f, -1.f),
    C_Vertex3f(0.0551562309f, 0.476941109f, -1.f),
    C_Vertex3f(0.0530986786f, 0.14234972f, -1.f),
    C_Vertex3f(0.0465703011f, 0.443202019f, -1.f),
    C_Vertex3f(0.0453357697f, 0.172130108f, -1.f),
    C_Vertex3f(0.0401308537f, 0.406296253f, -1.f),
    C_Vertex3f(0.039513588f, 0.204838037f, -1.f),
    C_Vertex3f(0.0358378887f, 0.366223812f, -1.f),
    C_Vertex3f(0.0356321335f, 0.240473509f, -1.f),
    C_Vertex3f(0.0336914063f, 0.322984695f, -1.f),
    C_Vertex3f(0.0336914063f, 0.279036522f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446653605f, 0.281977177f, -1.f),
    C_Vertex3f(0.447273731f, 0.135922909f, -1.f),
    C_Vertex3f(0.451842427f, 0.262220621f, -1.f),
    C_Vertex3f(0.453778744f, 0.163929462f, -1.f),
    C_Vertex3f(0.455301642f, 0.241406202f, -1.f),
    C_Vertex3f(0.45703125f, 0.193000793f, -1.f),
    C_Vertex3f(0.45703125f, 0.21953392f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.408602238f, 0.350424767f, -1.f),
    C_Vertex3f(0.409063339f, 0.060256958f, -1.f),
    C_Vertex3f(0.420709491f, 0.334899664f, -1.f),
    C_Vertex3f(0.424506187f, 0.0831041336f, -1.f),
    C_Vertex3f(0.431087136f, 0.318316698f, -1.f),
    C_Vertex3f(0.437516212f, 0.108981133f, -1.f),
    C_Vertex3f(0.439735174f, 0.300675869f, -1.f),
    C_Vertex3f(0.447273731f, 0.135922909f, -1.f),
    C_Vertex3f(0.446653605f, 0.281977177f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.394765377f, 0.364892006f, -1.f),
    C_Vertex3f(0.391187668f, 0.0404396057f, -1.f),
    C_Vertex3f(0.409063339f, 0.060256958f, -1.f),
    C_Vertex3f(0.408602238f, 0.350424767f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.391187668f, 0.0404396057f, -1.f),
    C_Vertex3f(0.375976563f, 0.211538315f, -1.f),
    C_Vertex3f(0.375976563f, 0.194597244f, -1.f),
    C_Vertex3f(0.374861479f, 0.178229809f, -1.f),
    C_Vertex3f(0.372631311f, 0.162764072f, -1.f),
    C_Vertex3f(0.370879173f, 0.0236520767f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.370879173f, 0.0236520767f, -1.f),
    C_Vertex3f(0.372631311f, 0.162764072f, -1.f),
    C_Vertex3f(0.36928606f, 0.148200035f, -1.f),
    C_Vertex3f(0.364825726f, 0.134537697f, -1.f),
    C_Vertex3f(0.359250307f, 0.121777058f, -1.f),
    C_Vertex3f(0.352559805f, 0.109918118f, -1.f),
    C_Vertex3f(0.348137856f, 0.00989437103f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.348137856f, 0.00989437103f, -1.f),
    C_Vertex3f(0.352559805f, 0.109918118f, -1.f),
    C_Vertex3f(0.344754219f, 0.0989608765f, -1.f),
    C_Vertex3f(0.335833549f, 0.0889053345f, -1.f),
    C_Vertex3f(0.326464653f, 0.0801067352f, -1.f),
    C_Vertex3f(0.323865414f, -0.000423908234f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.326464653f, 0.0801067352f, -1.f),
    C_Vertex3f(0.31664753f, 0.0725650787f, -1.f),
    C_Vertex3f(0.323865414f, -0.000423908234f, -1.f),
    C_Vertex3f(0.306382179f, 0.066280365f, -1.f),
    C_Vertex3f(0.298061848f, -0.00730276108f, -1.f),
    C_Vertex3f(0.295668602f, 0.061252594f, -1.f),
    C_Vertex3f(0.284506798f, 0.0574817657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.272896767f, 0.0549678802f, -1.f),
    C_Vertex3f(0.260838509f, 0.0537109375f, -1.f),
    C_Vertex3f(0.270727158f, -0.0107421875f, -1.f),
    C_Vertex3f(0.298061848f, -0.00730276108f, -1.f),
    C_Vertex3f(0.284506798f, 0.0574817657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.197120667f, 0.0683021545f, -1.f),
    C_Vertex3f(0.198199034f, -0.00396585464f, -1.f),
    C_Vertex3f(0.212989807f, 0.061006546f, -1.f),
    C_Vertex3f(0.220529795f, -0.00848340988f, -1.f),
    C_Vertex3f(0.229286194f, 0.056142807f, -1.f),
    C_Vertex3f(0.243994713f, -0.0107421875f, -1.f),
    C_Vertex3f(0.246009827f, 0.0537109375f, -1.f),
    C_Vertex3f(0.270727158f, -0.0107421875f, -1.f),
    C_Vertex3f(0.260838509f, 0.0537109375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.155909061f, 0.104125023f, -1.f),
    C_Vertex3f(0.156939983f, 0.0118455887f, -1.f),
    C_Vertex3f(0.16794157f, 0.0899705887f, -1.f),
    C_Vertex3f(0.17700243f, 0.00281047821f, -1.f),
    C_Vertex3f(0.181678772f, 0.0780296326f, -1.f),
    C_Vertex3f(0.198199034f, -0.00396585464f, -1.f),
    C_Vertex3f(0.197120667f, 0.0683021545f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145581245f, 0.120492935f, -1.f),
    C_Vertex3f(0.136958122f, 0.139074326f, -1.f),
    C_Vertex3f(0.138011694f, 0.0231394768f, -1.f),
    C_Vertex3f(0.156939983f, 0.0118455887f, -1.f),
    C_Vertex3f(0.155909061f, 0.104125023f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138011694f, 0.0231394768f, -1.f),
    C_Vertex3f(0.136958122f, 0.139074326f, -1.f),
    C_Vertex3f(0.13049078f, 0.158092499f, -1.f),
    C_Vertex3f(0.126179218f, 0.177547455f, -1.f),
    C_Vertex3f(0.124023438f, 0.197439194f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.445800781f, 0.483398438f, -1.f),
    C_Vertex3f(0.444597244f, 0.492749214f, -1.f),
    C_Vertex3f(0.440980434f, 0.510439038f, -1.f),
    C_Vertex3f(0.436153889f, 0.527117133f, -1.f),
    C_Vertex3f(0.430117607f, 0.542783499f, -1.f),
    C_Vertex3f(0.42287159f, 0.557438135f, -1.f),
    C_Vertex3f(0.414415836f, 0.571081042f, -1.f),
    C_Vertex3f(0.404750347f, 0.58371222f, -1.f),
    C_Vertex3f(0.393875122f, 0.595331669f, -1.f),
    C_Vertex3f(0.381790161f, 0.605939388f, -1.f),
    C_Vertex3f(0.368850708f, 0.615221143f, -1.f),
    C_Vertex3f(0.367675781f, 0.477050781f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.342702866f, 0.539485931f, -1.f),
    C_Vertex3f(0.351246834f, 0.525746346f, -1.f),
    C_Vertex3f(0.355056763f, 0.623176932f, -1.f),
    C_Vertex3f(0.35869503f, 0.508818626f, -1.f),
    C_Vertex3f(0.368850708f, 0.615221143f, -1.f),
    C_Vertex3f(0.365047455f, 0.488702774f, -1.f),
    C_Vertex3f(0.367675781f, 0.477050781f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.340408325f, 0.629806757f, -1.f),
    C_Vertex3f(0.324905396f, 0.635110617f, -1.f),
    C_Vertex3f(0.325424194f, 0.556398153f, -1.f),
    C_Vertex3f(0.333885193f, 0.548993111f, -1.f),
    C_Vertex3f(0.337882996f, 0.544761658f, -1.f),
    C_Vertex3f(0.342702866f, 0.539485931f, -1.f),
    C_Vertex3f(0.355056763f, 0.623176932f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.29724884f, 0.572266102f, -1.f),
    C_Vertex3f(0.307106018f, 0.568034649f, -1.f),
    C_Vertex3f(0.308547974f, 0.639088511f, -1.f),
    C_Vertex3f(0.316497803f, 0.562745333f, -1.f),
    C_Vertex3f(0.324905396f, 0.635110617f, -1.f),
    C_Vertex3f(0.325424194f, 0.556398153f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.29133606f, 0.641740441f, -1.f),
    C_Vertex3f(0.273269653f, 0.643066406f, -1.f),
    C_Vertex3f(0.276138306f, 0.577555418f, -1.f),
    C_Vertex3f(0.28692627f, 0.575439692f, -1.f),
    C_Vertex3f(0.29724884f, 0.572266102f, -1.f),
    C_Vertex3f(0.308547974f, 0.639088511f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.231884956f, 0.575989723f, -1.f),
    C_Vertex3f(0.249605179f, 0.578613281f, -1.f),
    C_Vertex3f(0.251121521f, 0.643066406f, -1.f),
    C_Vertex3f(0.264884949f, 0.578613281f, -1.f),
    C_Vertex3f(0.273269653f, 0.643066406f, -1.f),
    C_Vertex3f(0.276138306f, 0.577555418f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.183376312f, 0.550848007f, -1.f),
    C_Vertex3f(0.199806213f, 0.562871933f, -1.f),
    C_Vertex3f(0.204268217f, 0.636187792f, -1.f),
    C_Vertex3f(0.215285301f, 0.570742607f, -1.f),
    C_Vertex3f(0.227045298f, 0.640773535f, -1.f),
    C_Vertex3f(0.231884956f, 0.575989723f, -1.f),
    C_Vertex3f(0.251121521f, 0.643066406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.131475925f, 0.47198534f, -1.f),
    C_Vertex3f(0.139525414f, 0.491496086f, -1.f),
    C_Vertex3f(0.143731833f, 0.608673334f, -1.f),
    C_Vertex3f(0.152200222f, 0.514395237f, -1.f),
    C_Vertex3f(0.162611485f, 0.620137691f, -1.f),
    C_Vertex3f(0.166817188f, 0.534179211f, -1.f),
    C_Vertex3f(0.182790279f, 0.629309177f, -1.f),
    C_Vertex3f(0.183376312f, 0.550848007f, -1.f),
    C_Vertex3f(0.204268217f, 0.636187792f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.119237185f, 0.424691916f, -1.f),
    C_Vertex3f(0.122673392f, 0.441835046f, -1.f),
    C_Vertex3f(0.126151323f, 0.594916105f, -1.f),
    C_Vertex3f(0.126752973f, 0.457599521f, -1.f),
    C_Vertex3f(0.143731833f, 0.608673334f, -1.f),
    C_Vertex3f(0.131475925f, 0.47198534f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.126151323f, 0.594916105f, -1.f),
    C_Vertex3f(0.109869957f, 0.578866005f, -1.f),
    C_Vertex3f(0.112788796f, 0.364990592f, -1.f),
    C_Vertex3f(0.114294887f, 0.386269689f, -1.f),
    C_Vertex3f(0.116444349f, 0.40617013f, -1.f),
    C_Vertex3f(0.119237185f, 0.424691916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.111926079f, 0.34233284f, -1.f),
    C_Vertex3f(0.112788796f, 0.364990592f, -1.f),
    C_Vertex3f(0.109869957f, 0.578866005f, -1.f),
    C_Vertex3f(0.111816406f, 0.330314636f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_55[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0424804688f, 0.556640625f, 0.f),
    C_Vertex3f(0.0424804688f, 0.556640625f, -1.f),
    C_Vertex3f(0.0424804688f, 0.632324219f, -1.f),
    C_Vertex3f(0.0424804688f, 0.632324219f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0424804688f, 0.632324219f, 0.f),
    C_Vertex3f(0.0424804688f, 0.632324219f, -1.f),
    C_Vertex3f(0.45703125f, 0.632324219f, -1.f),
    C_Vertex3f(0.45703125f, 0.632324219f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.45703125f, 0.632324219f, 0.f),
    C_Vertex3f(0.45703125f, 0.632324219f, -1.f),
    C_Vertex3f(0.45703125f, 0.571075439f, -1.f),
    C_Vertex3f(0.45703125f, 0.571075439f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.45703125f, 0.571075439f, 0.f),
    C_Vertex3f(0.45703125f, 0.571075439f, -1.f),
    C_Normal3f(0.747414827f, -0.664357603f, 0.f),
    C_Vertex3f(0.449388504f, 0.562935829f, 0.f),
    C_Vertex3f(0.449388504f, 0.562935829f, -1.f),
    C_Normal3f(0.768294573f, -0.640096486f, 0.f),
    C_Vertex3f(0.434136987f, 0.545318961f, 0.f),
    C_Vertex3f(0.434136987f, 0.545318961f, -1.f),
    C_Normal3f(0.790561914f, -0.612382054f, 0.f),
    C_Vertex3f(0.418919444f, 0.526364446f, 0.f),
    C_Vertex3f(0.418919444f, 0.526364446f, -1.f),
    C_Normal3f(0.810163617f, -0.586203873f, 0.f),
    C_Vertex3f(0.403735876f, 0.506072283f, 0.f),
    C_Vertex3f(0.403735876f, 0.506072283f, -1.f),
    C_Normal3f(0.827451825f, -0.561536729f, 0.f),
    C_Vertex3f(0.388586283f, 0.484442472f, 0.f),
    C_Vertex3f(0.388586283f, 0.484442472f, -1.f),
    C_Normal3f(0.8427338f, -0.538330317f, 0.f),
    C_Vertex3f(0.373470664f, 0.461475015f, 0.f),
    C_Vertex3f(0.373470664f, 0.461475015f, -1.f),
    C_Normal3f(0.856275916f, -0.516518772f, 0.f),
    C_Vertex3f(0.35838902f, 0.437169909f, 0.f),
    C_Vertex3f(0.35838902f, 0.437169909f, -1.f),
    C_Normal3f(0.869435072f, -0.494047254f, 0.f),
    C_Vertex3f(0.343341351f, 0.411527157f, 0.f),
    C_Vertex3f(0.343341351f, 0.411527157f, -1.f),
    C_Normal3f(0.888613403f, -0.458656967f, 0.f),
    C_Vertex3f(0.320820808f, 0.371056557f, 0.f),
    C_Vertex3f(0.320820808f, 0.371056557f, -1.f),
    C_Normal3f(0.911239624f, -0.411876678f, 0.f),
    C_Vertex3f(0.294206142f, 0.316331387f, 0.f),
    C_Vertex3f(0.294206142f, 0.316331387f, -1.f),
    C_Normal3f(0.933309436f, -0.359073013f, 0.f),
    C_Vertex3f(0.2710042f, 0.260841846f, 0.f),
    C_Vertex3f(0.2710042f, 0.260841846f, -1.f),
    C_Normal3f(0.951583803f, -0.30738923f, 0.f),
    C_Vertex3f(0.251214981f, 0.204587936f, 0.f),
    C_Vertex3f(0.251214981f, 0.204587936f, -1.f),
    C_Normal3f(0.96740371f, -0.253239065f, 0.f),
    C_Vertex3f(0.237131119f, 0.155981064f, 0.f),
    C_Vertex3f(0.237131119f, 0.155981064f, -1.f),
    C_Normal3f(0.98064363f, -0.195800796f, 0.f),
    C_Vertex3f(0.227468491f, 0.113873482f, 0.f),
    C_Vertex3f(0.227468491f, 0.113873482f, -1.f),
    C_Normal3f(0.989811599f, -0.142383605f, 0.f),
    C_Vertex3f(0.219934464f, 0.0698537827f, 0.f),
    C_Vertex3f(0.219934464f, 0.0698537827f, -1.f),
    C_Normal3f(0.994954348f, -0.100328103f, 0.f),
    C_Vertex3f(0.214529037f, 0.0239219666f, 0.f),
    C_Vertex3f(0.214529037f, 0.0239219666f, -1.f),
    C_Normal3f(0.997662842f, -0.0683298036f, 0.f),
    C_Vertex3f(0.212890625f, 0.f, 0.f),
    C_Vertex3f(0.212890625f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.212890625f, 0.f, 0.f),
    C_Vertex3f(0.212890625f, 0.f, -1.f),
    C_Vertex3f(0.131835938f, 0.f, -1.f),
    C_Vertex3f(0.131835938f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999849319f, 0.0173588451f, 0.f),
    C_Vertex3f(0.131835938f, 0.f, 0.f),
    C_Vertex3f(0.131835938f, 0.f, -1.f),
    C_Normal3f(-0.998960197f, 0.0455918498f, 0.f),
    C_Vertex3f(0.131999969f, 0.00944805145f, 0.f),
    C_Vertex3f(0.131999969f, 0.00944805145f, -1.f),
    C_Normal3f(-0.996938705f, 0.0781873763f, 0.f),
    C_Vertex3f(0.133174419f, 0.0293273926f, 0.f),
    C_Vertex3f(0.133174419f, 0.0293273926f, -1.f),
    C_Normal3f(-0.993514061f, 0.113709584f, 0.f),
    C_Vertex3f(0.135195255f, 0.0501899719f, 0.f),
    C_Vertex3f(0.135195255f, 0.0501899719f, -1.f),
    C_Normal3f(-0.989324093f, 0.145732328f, 0.f),
    C_Vertex3f(0.138062477f, 0.0720357895f, 0.f),
    C_Vertex3f(0.138062477f, 0.0720357895f, -1.f),
    C_Normal3f(-0.984628618f, 0.174661011f, 0.f),
    C_Vertex3f(0.141776085f, 0.0948648453f, 0.f),
    C_Vertex3f(0.141776085f, 0.0948648453f, -1.f),
    C_Normal3f(-0.979620337f, 0.200857818f, 0.f),
    C_Vertex3f(0.146336079f, 0.118677139f, 0.f),
    C_Vertex3f(0.146336079f, 0.118677139f, -1.f),
    C_Normal3f(-0.974441171f, 0.224642754f, 0.f),
    C_Vertex3f(0.151742458f, 0.143472672f, 0.f),
    C_Vertex3f(0.151742458f, 0.143472672f, -1.f),
    C_Normal3f(-0.968669295f, 0.248354286f, 0.f),
    C_Vertex3f(0.157995224f, 0.169251442f, 0.f),
    C_Vertex3f(0.157995224f, 0.169251442f, -1.f),
    C_Normal3f(-0.957486272f, 0.288478523f, 0.f),
    C_Vertex3f(0.168643951f, 0.209394455f, 0.f),
    C_Vertex3f(0.168643951f, 0.209394455f, -1.f),
    C_Normal3f(-0.940458834f, 0.339907527f, 0.f),
    C_Vertex3f(0.185927868f, 0.261962414f, 0.f),
    C_Vertex3f(0.185927868f, 0.261962414f, -1.f),
    C_Normal3f(-0.919159234f, 0.393886209f, 0.f),
    C_Vertex3f(0.206297398f, 0.313574314f, 0.f),
    C_Vertex3f(0.206297398f, 0.313574314f, -1.f),
    C_Normal3f(-0.896880388f, 0.44227308f, 0.f),
    C_Vertex3f(0.229752541f, 0.364230156f, 0.f),
    C_Vertex3f(0.229752541f, 0.364230156f, -1.f),
    C_Normal3f(-0.877691567f, 0.479226023f, 0.f),
    C_Vertex3f(0.249658108f, 0.401504993f, 0.f),
    C_Vertex3f(0.249658108f, 0.401504993f, -1.f),
    C_Normal3f(-0.864643693f, 0.502385616f, 0.f),
    C_Vertex3f(0.263140082f, 0.42537868f, 0.f),
    C_Vertex3f(0.263140082f, 0.42537868f, -1.f),
    C_Normal3f(-0.851505995f, 0.524344981f, 0.f),
    C_Vertex3f(0.276833653f, 0.448276162f, 0.f),
    C_Vertex3f(0.276833653f, 0.448276162f, -1.f),
    C_Normal3f(-0.837023914f, 0.547166407f, 0.f),
    C_Vertex3f(0.290738821f, 0.470197439f, 0.f),
    C_Vertex3f(0.290738821f, 0.470197439f, -1.f),
    C_Normal3f(-0.821062744f, 0.570838034f, 0.f),
    C_Vertex3f(0.304855585f, 0.491142511f, 0.f),
    C_Vertex3f(0.304855585f, 0.491142511f, -1.f),
    C_Normal3f(-0.803479016f, 0.59533304f, 0.f),
    C_Vertex3f(0.319183946f, 0.511111379f, 0.f),
    C_Vertex3f(0.319183946f, 0.511111379f, -1.f),
    C_Normal3f(-0.784121931f, 0.620606899f, 0.f),
    C_Vertex3f(0.333723903f, 0.530104041f, 0.f),
    C_Vertex3f(0.333723903f, 0.530104041f, -1.f),
    C_Normal3f(-0.766523719f, 0.642216027f, 0.f),
    C_Vertex3f(0.348475456f, 0.548120499f, 0.f),
    C_Vertex3f(0.348475456f, 0.548120499f, -1.f),
    C_Normal3f(-0.751418889f, 0.659825504f, 0.f),
    C_Vertex3f(0.355957031f, 0.556640625f, 0.f),
    C_Vertex3f(0.355957031f, 0.556640625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.355957031f, 0.556640625f, 0.f),
    C_Vertex3f(0.355957031f, 0.556640625f, -1.f),
    C_Vertex3f(0.0424804688f, 0.556640625f, -1.f),
    C_Vertex3f(0.0424804688f, 0.556640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.45703125f, 0.632324219f, 0.f),
    C_Vertex3f(0.0424804688f, 0.632324219f, 0.f),
    C_Vertex3f(0.355957031f, 0.556640625f, 0.f),
    C_Vertex3f(0.35838902f, 0.437169909f, 0.f),
    C_Vertex3f(0.373470664f, 0.461475015f, 0.f),
    C_Vertex3f(0.388586283f, 0.484442472f, 0.f),
    C_Vertex3f(0.403735876f, 0.506072283f, 0.f),
    C_Vertex3f(0.418919444f, 0.526364446f, 0.f),
    C_Vertex3f(0.434136987f, 0.545318961f, 0.f),
    C_Vertex3f(0.449388504f, 0.562935829f, 0.f),
    C_Vertex3f(0.45703125f, 0.571075439f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.319183946f, 0.511111379f, 0.f),
    C_Vertex3f(0.320820808f, 0.371056557f, 0.f),
    C_Vertex3f(0.333723903f, 0.530104041f, 0.f),
    C_Vertex3f(0.343341351f, 0.411527157f, 0.f),
    C_Vertex3f(0.348475456f, 0.548120499f, 0.f),
    C_Vertex3f(0.35838902f, 0.437169909f, 0.f),
    C_Vertex3f(0.355957031f, 0.556640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304855585f, 0.491142511f, 0.f),
    C_Vertex3f(0.290738821f, 0.470197439f, 0.f),
    C_Vertex3f(0.294206142f, 0.316331387f, 0.f),
    C_Vertex3f(0.320820808f, 0.371056557f, 0.f),
    C_Vertex3f(0.319183946f, 0.511111379f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.219934464f, 0.0698537827f, 0.f),
    C_Vertex3f(0.227468491f, 0.113873482f, 0.f),
    C_Vertex3f(0.229752541f, 0.364230156f, 0.f),
    C_Vertex3f(0.237131119f, 0.155981064f, 0.f),
    C_Vertex3f(0.249658108f, 0.401504993f, 0.f),
    C_Vertex3f(0.251214981f, 0.204587936f, 0.f),
    C_Vertex3f(0.263140082f, 0.42537868f, 0.f),
    C_Vertex3f(0.2710042f, 0.260841846f, 0.f),
    C_Vertex3f(0.276833653f, 0.448276162f, 0.f),
    C_Vertex3f(0.294206142f, 0.316331387f, 0.f),
    C_Vertex3f(0.290738821f, 0.470197439f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.229752541f, 0.364230156f, 0.f),
    C_Vertex3f(0.206297398f, 0.313574314f, 0.f),
    C_Vertex3f(0.212890625f, 0.f, 0.f),
    C_Vertex3f(0.214529037f, 0.0239219666f, 0.f),
    C_Vertex3f(0.219934464f, 0.0698537827f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.212890625f, 0.f, 0.f),
    C_Vertex3f(0.206297398f, 0.313574314f, 0.f),
    C_Vertex3f(0.185927868f, 0.261962414f, 0.f),
    C_Vertex3f(0.168643951f, 0.209394455f, 0.f),
    C_Vertex3f(0.157995224f, 0.169251442f, 0.f),
    C_Vertex3f(0.151742458f, 0.143472672f, 0.f),
    C_Vertex3f(0.146336079f, 0.118677139f, 0.f),
    C_Vertex3f(0.141776085f, 0.0948648453f, 0.f),
    C_Vertex3f(0.138062477f, 0.0720357895f, 0.f),
    C_Vertex3f(0.135195255f, 0.0501899719f, 0.f),
    C_Vertex3f(0.133174419f, 0.0293273926f, 0.f),
    C_Vertex3f(0.131999969f, 0.00944805145f, 0.f),
    C_Vertex3f(0.131835938f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.355957031f, 0.556640625f, 0.f),
    C_Vertex3f(0.0424804688f, 0.632324219f, 0.f),
    C_Vertex3f(0.0424804688f, 0.556640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.45703125f, 0.632324219f, -1.f),
    C_Vertex3f(0.0424804688f, 0.632324219f, -1.f),
    C_Vertex3f(0.355957031f, 0.556640625f, -1.f),
    C_Vertex3f(0.35838902f, 0.437169909f, -1.f),
    C_Vertex3f(0.373470664f, 0.461475015f, -1.f),
    C_Vertex3f(0.388586283f, 0.484442472f, -1.f),
    C_Vertex3f(0.403735876f, 0.506072283f, -1.f),
    C_Vertex3f(0.418919444f, 0.526364446f, -1.f),
    C_Vertex3f(0.434136987f, 0.545318961f, -1.f),
    C_Vertex3f(0.449388504f, 0.562935829f, -1.f),
    C_Vertex3f(0.45703125f, 0.571075439f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.319183946f, 0.511111379f, -1.f),
    C_Vertex3f(0.320820808f, 0.371056557f, -1.f),
    C_Vertex3f(0.333723903f, 0.530104041f, -1.f),
    C_Vertex3f(0.343341351f, 0.411527157f, -1.f),
    C_Vertex3f(0.348475456f, 0.548120499f, -1.f),
    C_Vertex3f(0.35838902f, 0.437169909f, -1.f),
    C_Vertex3f(0.355957031f, 0.556640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304855585f, 0.491142511f, -1.f),
    C_Vertex3f(0.290738821f, 0.470197439f, -1.f),
    C_Vertex3f(0.294206142f, 0.316331387f, -1.f),
    C_Vertex3f(0.320820808f, 0.371056557f, -1.f),
    C_Vertex3f(0.319183946f, 0.511111379f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.219934464f, 0.0698537827f, -1.f),
    C_Vertex3f(0.227468491f, 0.113873482f, -1.f),
    C_Vertex3f(0.229752541f, 0.364230156f, -1.f),
    C_Vertex3f(0.237131119f, 0.155981064f, -1.f),
    C_Vertex3f(0.249658108f, 0.401504993f, -1.f),
    C_Vertex3f(0.251214981f, 0.204587936f, -1.f),
    C_Vertex3f(0.263140082f, 0.42537868f, -1.f),
    C_Vertex3f(0.2710042f, 0.260841846f, -1.f),
    C_Vertex3f(0.276833653f, 0.448276162f, -1.f),
    C_Vertex3f(0.294206142f, 0.316331387f, -1.f),
    C_Vertex3f(0.290738821f, 0.470197439f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.229752541f, 0.364230156f, -1.f),
    C_Vertex3f(0.206297398f, 0.313574314f, -1.f),
    C_Vertex3f(0.212890625f, 0.f, -1.f),
    C_Vertex3f(0.214529037f, 0.0239219666f, -1.f),
    C_Vertex3f(0.219934464f, 0.0698537827f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.212890625f, 0.f, -1.f),
    C_Vertex3f(0.206297398f, 0.313574314f, -1.f),
    C_Vertex3f(0.185927868f, 0.261962414f, -1.f),
    C_Vertex3f(0.168643951f, 0.209394455f, -1.f),
    C_Vertex3f(0.157995224f, 0.169251442f, -1.f),
    C_Vertex3f(0.151742458f, 0.143472672f, -1.f),
    C_Vertex3f(0.146336079f, 0.118677139f, -1.f),
    C_Vertex3f(0.141776085f, 0.0948648453f, -1.f),
    C_Vertex3f(0.138062477f, 0.0720357895f, -1.f),
    C_Vertex3f(0.135195255f, 0.0501899719f, -1.f),
    C_Vertex3f(0.133174419f, 0.0293273926f, -1.f),
    C_Vertex3f(0.131999969f, 0.00944805145f, -1.f),
    C_Vertex3f(0.131835938f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.355957031f, 0.556640625f, -1.f),
    C_Vertex3f(0.0424804688f, 0.632324219f, -1.f),
    C_Vertex3f(0.0424804688f, 0.556640625f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_56[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.34159103f, -0.939848721f, 0.f),
    C_Vertex3f(0.158699036f, 0.347679138f, 0.f),
    C_Vertex3f(0.158699036f, 0.347679138f, -1.f),
    C_Normal3f(-0.413664311f, -0.910429537f, 0.f),
    C_Vertex3f(0.146392822f, 0.352151871f, 0.f),
    C_Vertex3f(0.146392822f, 0.352151871f, -1.f),
    C_Normal3f(-0.511577785f, -0.859237015f, 0.f),
    C_Vertex3f(0.124967098f, 0.363005638f, 0.f),
    C_Vertex3f(0.124967098f, 0.363005638f, -1.f),
    C_Normal3f(-0.635928869f, -0.771747708f, 0.f),
    C_Vertex3f(0.106728077f, 0.375767708f, 0.f),
    C_Vertex3f(0.106728077f, 0.375767708f, -1.f),
    C_Normal3f(-0.757660508f, -0.652648926f, 0.f),
    C_Vertex3f(0.0916757584f, 0.39043808f, 0.f),
    C_Vertex3f(0.0916757584f, 0.39043808f, -1.f),
    C_Normal3f(-0.858710587f, -0.512460887f, 0.f),
    C_Vertex3f(0.0798101425f, 0.407016754f, 0.f),
    C_Vertex3f(0.0798101425f, 0.407016754f, -1.f),
    C_Normal3f(-0.931780219f, -0.363022923f, 0.f),
    C_Vertex3f(0.0709109306f, 0.425232887f, 0.f),
    C_Vertex3f(0.0709109306f, 0.425232887f, -1.f),
    C_Normal3f(-0.977609932f, -0.210425287f, 0.f),
    C_Vertex3f(0.0649781227f, 0.445086479f, 0.f),
    C_Vertex3f(0.0649781227f, 0.445086479f, -1.f),
    C_Normal3f(-0.997489154f, -0.0708198845f, 0.f),
    C_Vertex3f(0.0620117188f, 0.46657753f, 0.f),
    C_Vertex3f(0.0620117188f, 0.46657753f, -1.f),
    C_Normal3f(-0.999105155f, 0.042294994f, 0.f),
    C_Vertex3f(0.0620117188f, 0.486867905f, 0.f),
    C_Vertex3f(0.0620117188f, 0.486867905f, -1.f),
    C_Normal3f(-0.989879251f, 0.141912133f, 0.f),
    C_Vertex3f(0.0635800362f, 0.503624797f, 0.f),
    C_Vertex3f(0.0635800362f, 0.503624797f, -1.f),
    C_Normal3f(-0.970253944f, 0.242089644f, 0.f),
    C_Vertex3f(0.066716671f, 0.519686341f, 0.f),
    C_Vertex3f(0.066716671f, 0.519686341f, -1.f),
    C_Normal3f(-0.939233243f, 0.343279779f, 0.f),
    C_Vertex3f(0.0714216232f, 0.535052538f, 0.f),
    C_Vertex3f(0.0714216232f, 0.535052538f, -1.f),
    C_Normal3f(-0.897021294f, 0.441987455f, 0.f),
    C_Vertex3f(0.0776948929f, 0.549723387f, 0.f),
    C_Vertex3f(0.0776948929f, 0.549723387f, -1.f),
    C_Normal3f(-0.8449651f, 0.534821451f, 0.f),
    C_Vertex3f(0.0855364799f, 0.563698888f, 0.f),
    C_Vertex3f(0.0855364799f, 0.563698888f, -1.f),
    C_Normal3f(-0.785348892f, 0.619053423f, 0.f),
    C_Vertex3f(0.0949463844f, 0.576979041f, 0.f),
    C_Vertex3f(0.0949463844f, 0.576979041f, -1.f),
    C_Normal3f(-0.720953465f, 0.692983389f, 0.f),
    C_Vertex3f(0.105924606f, 0.589563847f, 0.f),
    C_Vertex3f(0.105924606f, 0.589563847f, -1.f),
    C_Normal3f(-0.649044693f, 0.760750294f, 0.f),
    C_Vertex3f(0.118471146f, 0.601453304f, 0.f),
    C_Vertex3f(0.118471146f, 0.601453304f, -1.f),
    C_Normal3f(-0.565163255f, 0.824979126f, 0.f),
    C_Vertex3f(0.13205409f, 0.61185658f, 0.f),
    C_Vertex3f(0.13205409f, 0.61185658f, -1.f),
    C_Normal3f(-0.475136459f, 0.879912078f, 0.f),
    C_Vertex3f(0.146673441f, 0.620773673f, 0.f),
    C_Vertex3f(0.146673441f, 0.620773673f, -1.f),
    C_Normal3f(-0.382115066f, 0.924114764f, 0.f),
    C_Vertex3f(0.162329197f, 0.628204584f, 0.f),
    C_Vertex3f(0.162329197f, 0.628204584f, -1.f),
    C_Normal3f(-0.289313287f, 0.957234442f, 0.f),
    C_Vertex3f(0.179021358f, 0.634149313f, 0.f),
    C_Vertex3f(0.179021358f, 0.634149313f, -1.f),
    C_Normal3f(-0.199528217f, 0.979892075f, 0.f),
    C_Vertex3f(0.196749926f, 0.63860786f, 0.f),
    C_Vertex3f(0.196749926f, 0.63860786f, -1.f),
    C_Normal3f(-0.114842273f, 0.993383646f, 0.f),
    C_Vertex3f(0.215514898f, 0.641580224f, 0.f),
    C_Vertex3f(0.215514898f, 0.641580224f, -1.f),
    C_Normal3f(-0.0364970155f, 0.999333739f, 0.f),
    C_Vertex3f(0.235316277f, 0.643066406f, 0.f),
    C_Vertex3f(0.235316277f, 0.643066406f, -1.f),
    C_Normal3f(0.0372212939f, 0.999306977f, 0.f),
    C_Vertex3f(0.25620842f, 0.643066406f, 0.f),
    C_Vertex3f(0.25620842f, 0.643066406f, -1.f),
    C_Normal3f(0.116669446f, 0.993170798f, 0.f),
    C_Vertex3f(0.276132226f, 0.64154613f, 0.f),
    C_Vertex3f(0.276132226f, 0.64154613f, -1.f),
    C_Normal3f(0.20239684f, 0.979303598f, 0.f),
    C_Vertex3f(0.295033336f, 0.638505578f, 0.f),
    C_Vertex3f(0.295033336f, 0.638505578f, -1.f),
    C_Normal3f(0.292941689f, 0.956130326f, 0.f),
    C_Vertex3f(0.312911749f, 0.63394475f, 0.f),
    C_Vertex3f(0.312911749f, 0.63394475f, -1.f),
    C_Normal3f(0.386109918f, 0.922452748f, 0.f),
    C_Vertex3f(0.329767466f, 0.627863646f, 0.f),
    C_Vertex3f(0.329767466f, 0.627863646f, -1.f),
    C_Normal3f(0.47903949f, 0.877793431f, 0.f),
    C_Vertex3f(0.345600486f, 0.620262265f, 0.f),
    C_Vertex3f(0.345600486f, 0.620262265f, -1.f),
    C_Normal3f(0.568529308f, 0.822663009f, 0.f),
    C_Vertex3f(0.36041081f, 0.611140609f, 0.f),
    C_Vertex3f(0.36041081f, 0.611140609f, -1.f),
    C_Normal3f(0.651529014f, 0.758623719f, 0.f),
    C_Vertex3f(0.374198437f, 0.600498676f, 0.f),
    C_Vertex3f(0.374198437f, 0.600498676f, -1.f),
    C_Normal3f(0.722137749f, 0.691749334f, 0.f),
    C_Vertex3f(0.386963367f, 0.588336468f, 0.f),
    C_Vertex3f(0.386963367f, 0.588336468f, -1.f),
    C_Normal3f(0.785268784f, 0.61915499f, 0.f),
    C_Vertex3f(0.398132682f, 0.575513005f, 0.f),
    C_Vertex3f(0.398132682f, 0.575513005f, -1.f),
    C_Normal3f(0.844095469f, 0.536192894f, 0.f),
    C_Vertex3f(0.40770638f, 0.562028289f, 0.f),
    C_Vertex3f(0.40770638f, 0.562028289f, -1.f),
    C_Normal3f(0.895858824f, 0.44433865f, 0.f),
    C_Vertex3f(0.415684462f, 0.547882318f, 0.f),
    C_Vertex3f(0.415684462f, 0.547882318f, -1.f),
    C_Normal3f(0.938191235f, 0.346117586f, 0.f),
    C_Vertex3f(0.422066927f, 0.533075094f, 0.f),
    C_Vertex3f(0.422066927f, 0.533075094f, -1.f),
    C_Normal3f(0.969572902f, 0.244802848f, 0.f),
    C_Vertex3f(0.426853776f, 0.517606616f, 0.f),
    C_Vertex3f(0.426853776f, 0.517606616f, -1.f),
    C_Normal3f(0.989593446f, 0.143892199f, 0.f),
    C_Vertex3f(0.430045009f, 0.501476884f, 0.f),
    C_Vertex3f(0.430045009f, 0.501476884f, -1.f),
    C_Normal3f(0.999053299f, 0.0435018204f, 0.f),
    C_Vertex3f(0.431640625f, 0.484685898f, 0.f),
    C_Vertex3f(0.431640625f, 0.484685898f, -1.f),
    C_Normal3f(0.997420371f, -0.0717815161f, 0.f),
    C_Vertex3f(0.431640625f, 0.464832306f, 0.f),
    C_Vertex3f(0.431640625f, 0.464832306f, -1.f),
    C_Normal3f(0.976942599f, -0.213502169f, 0.f),
    C_Vertex3f(0.428714275f, 0.444023609f, 0.f),
    C_Vertex3f(0.428714275f, 0.444023609f, -1.f),
    C_Normal3f(0.930865526f, -0.365361661f, 0.f),
    C_Vertex3f(0.422861576f, 0.42466116f, 0.f),
    C_Vertex3f(0.422861576f, 0.42466116f, -1.f),
    C_Normal3f(0.859111667f, -0.51178813f, 0.f),
    C_Vertex3f(0.414082527f, 0.406744957f, 0.f),
    C_Vertex3f(0.414082527f, 0.406744957f, -1.f),
    C_Normal3f(0.761383832f, -0.648301363f, 0.f),
    C_Vertex3f(0.402377129f, 0.390275002f, 0.f),
    C_Vertex3f(0.402377129f, 0.390275002f, -1.f),
    C_Normal3f(0.642958701f, -0.76590085f, 0.f),
    C_Vertex3f(0.387636662f, 0.375686169f, 0.f),
    C_Vertex3f(0.387636662f, 0.375686169f, -1.f),
    C_Normal3f(0.520706773f, -0.853735507f, 0.f),
    C_Vertex3f(0.369861126f, 0.362978458f, 0.f),
    C_Vertex3f(0.369861126f, 0.362978458f, -1.f),
    C_Normal3f(0.423424333f, -0.905931473f, 0.f),
    C_Vertex3f(0.349050522f, 0.352151871f, 0.f),
    C_Vertex3f(0.349050522f, 0.352151871f, -1.f),
    C_Normal3f(0.35123837f, -0.936286151f, 0.f),
    C_Vertex3f(0.337127686f, 0.347679138f, 0.f),
    C_Vertex3f(0.337127686f, 0.347679138f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.309335768f, 0.950952888f, 0.f),
    C_Vertex3f(0.337127686f, 0.347679138f, 0.f),
    C_Vertex3f(0.337127686f, 0.347679138f, -1.f),
    C_Normal3f(0.349827409f, 0.936814249f, 0.f),
    C_Vertex3f(0.344518661f, 0.345274925f, 0.f),
    C_Vertex3f(0.344518661f, 0.345274925f, -1.f),
    C_Normal3f(0.402361244f, 0.915480971f, 0.f),
    C_Vertex3f(0.358417749f, 0.339728951f, 0.f),
    C_Vertex3f(0.358417749f, 0.339728951f, -1.f),
    C_Normal3f(0.467616051f, 0.883931637f, 0.f),
    C_Vertex3f(0.371433973f, 0.33344543f, 0.f),
    C_Vertex3f(0.371433973f, 0.33344543f, -1.f),
    C_Normal3f(0.534275532f, 0.84531039f, 0.f),
    C_Vertex3f(0.383567333f, 0.32642436f, 0.f),
    C_Vertex3f(0.383567333f, 0.32642436f, -1.f),
    C_Normal3f(0.600974143f, 0.799268484f, 0.f),
    C_Vertex3f(0.394817829f, 0.318665743f, 0.f),
    C_Vertex3f(0.394817829f, 0.318665743f, -1.f),
    C_Normal3f(0.666110337f, 0.745853305f, 0.f),
    C_Vertex3f(0.405185461f, 0.310169578f, 0.f),
    C_Vertex3f(0.405185461f, 0.310169578f, -1.f),
    C_Normal3f(0.72798425f, 0.685593843f, 0.f),
    C_Vertex3f(0.414670229f, 0.300935864f, 0.f),
    C_Vertex3f(0.414670229f, 0.300935864f, -1.f),
    C_Normal3f(0.78498292f, 0.619517386f, 0.f),
    C_Vertex3f(0.423272133f, 0.290964603f, 0.f),
    C_Vertex3f(0.423272133f, 0.290964603f, -1.f),
    C_Normal3f(0.834698081f, 0.550707757f, 0.f),
    C_Vertex3f(0.430991173f, 0.280255795f, 0.f),
    C_Vertex3f(0.430991173f, 0.280255795f, -1.f),
    C_Normal3f(0.877709627f, 0.479192853f, 0.f),
    C_Vertex3f(0.437745333f, 0.269027829f, 0.f),
    C_Vertex3f(0.437745333f, 0.269027829f, -1.f),
    C_Normal3f(0.914642811f, 0.404262871f, 0.f),
    C_Vertex3f(0.443534613f, 0.257280707f, 0.f),
    C_Vertex3f(0.443534613f, 0.257280707f, -1.f),
    C_Normal3f(0.944844067f, 0.327520519f, 0.f),
    C_Vertex3f(0.448359013f, 0.245014429f, 0.f),
    C_Vertex3f(0.448359013f, 0.245014429f, -1.f),
    C_Normal3f(0.968086243f, 0.250617206f, 0.f),
    C_Vertex3f(0.452218533f, 0.232228994f, 0.f),
    C_Vertex3f(0.452218533f, 0.232228994f, -1.f),
    C_Normal3f(0.984552622f, 0.175089076f, 0.f),
    C_Vertex3f(0.455113173f, 0.218924403f, 0.f),
    C_Vertex3f(0.455113173f, 0.218924403f, -1.f),
    C_Normal3f(0.994760811f, 0.102229916f, 0.f),
    C_Vertex3f(0.457042933f, 0.205100656f, 0.f),
    C_Vertex3f(0.457042933f, 0.205100656f, -1.f),
    C_Normal3f(0.999547064f, 0.0300965123f, 0.f),
    C_Vertex3f(0.458007813f, 0.190757751f, 0.f),
    C_Vertex3f(0.458007813f, 0.190757751f, -1.f),
    C_Normal3f(0.998826027f, -0.0484423116f, 0.f),
    C_Vertex3f(0.458007813f, 0.173055649f, 0.f),
    C_Vertex3f(0.458007813f, 0.173055649f, -1.f),
    C_Normal3f(0.990209043f, -0.139592096f, 0.f),
    C_Vertex3f(0.456192732f, 0.153332829f, 0.f),
    C_Vertex3f(0.456192732f, 0.153332829f, -1.f),
    C_Normal3f(0.971192896f, -0.23829478f, 0.f),
    C_Vertex3f(0.452562571f, 0.134429336f, 0.f),
    C_Vertex3f(0.452562571f, 0.134429336f, -1.f),
    C_Normal3f(0.941057682f, -0.338246197f, 0.f),
    C_Vertex3f(0.447117329f, 0.116345167f, 0.f),
    C_Vertex3f(0.447117329f, 0.116345167f, -1.f),
    C_Normal3f(0.899905205f, -0.436085641f, 0.f),
    C_Vertex3f(0.439857006f, 0.0990803242f, 0.f),
    C_Vertex3f(0.439857006f, 0.0990803242f, -1.f),
    C_Normal3f(0.848935962f, -0.528495669f, 0.f),
    C_Vertex3f(0.430781603f, 0.0826348066f, 0.f),
    C_Vertex3f(0.430781603f, 0.0826348066f, -1.f),
    C_Normal3f(0.790283918f, -0.612740815f, 0.f),
    C_Vertex3f(0.419891119f, 0.0670086145f, 0.f),
    C_Vertex3f(0.419891119f, 0.0670086145f, -1.f),
    C_Normal3f(0.726616323f, -0.687043369f, 0.f),
    C_Vertex3f(0.407185555f, 0.0522017479f, 0.f),
    C_Vertex3f(0.407185555f, 0.0522017479f, -1.f),
    C_Normal3f(0.655841708f, -0.754898429f, 0.f),
    C_Vertex3f(0.392664909f, 0.0382142067f, 0.f),
    C_Vertex3f(0.392664909f, 0.0382142067f, -1.f),
    C_Normal3f(0.573345304f, -0.819313824f, 0.f),
    C_Vertex3f(0.376997709f, 0.0259751081f, 0.f),
    C_Vertex3f(0.376997709f, 0.0259751081f, -1.f),
    C_Normal3f(0.483987242f, -0.875075042f, 0.f),
    C_Vertex3f(0.360183954f, 0.0154844522f, 0.f),
    C_Vertex3f(0.360183954f, 0.0154844522f, -1.f),
    C_Normal3f(0.390771121f, -0.920487821f, 0.f),
    C_Vertex3f(0.342223644f, 0.006742239f, 0.f),
    C_Vertex3f(0.342223644f, 0.006742239f, -1.f),
    C_Normal3f(0.296926677f, -0.954900265f, 0.f),
    C_Vertex3f(0.323116779f, -0.000251531601f, 0.f),
    C_Vertex3f(0.323116779f, -0.000251531601f, -1.f),
    C_Normal3f(0.205404729f, -0.978677154f, 0.f),
    C_Vertex3f(0.302863359f, -0.00549685955f, 0.f),
    C_Vertex3f(0.302863359f, -0.00549685955f, -1.f),
    C_Normal3f(0.118515901f, -0.992952108f, 0.f),
    C_Vertex3f(0.281463385f, -0.00899374485f, 0.f),
    C_Vertex3f(0.281463385f, -0.00899374485f, -1.f),
    C_Normal3f(0.037785653f, -0.999285877f, 0.f),
    C_Vertex3f(0.258916855f, -0.0107421875f, 0.f),
    C_Vertex3f(0.258916855f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0379322879f, -0.999280334f, 0.f),
    C_Vertex3f(0.23522377f, -0.0107421875f, 0.f),
    C_Vertex3f(0.23522377f, -0.0107421875f, -1.f),
    C_Normal3f(-0.118969977f, -0.992897809f, 0.f),
    C_Vertex3f(0.21267724f, -0.00898694992f, 0.f),
    C_Vertex3f(0.21267724f, -0.00898694992f, -1.f),
    C_Normal3f(-0.206169114f, -0.97851634f, 0.f),
    C_Vertex3f(0.191277266f, -0.00547647476f, 0.f),
    C_Vertex3f(0.191277266f, -0.00547647476f, -1.f),
    C_Normal3f(-0.297978342f, -0.954572618f, 0.f),
    C_Vertex3f(0.171023846f, -0.000210762024f, 0.f),
    C_Vertex3f(0.171023846f, -0.000210762024f, -1.f),
    C_Normal3f(-0.392056733f, -0.919941008f, 0.f),
    C_Vertex3f(0.151916981f, 0.00681018829f, 0.f),
    C_Vertex3f(0.151916981f, 0.00681018829f, -1.f),
    C_Normal3f(-0.485425621f, -0.874278009f, 0.f),
    C_Vertex3f(0.133956671f, 0.0155863762f, 0.f),
    C_Vertex3f(0.133956671f, 0.0155863762f, -1.f),
    C_Normal3f(-0.574838161f, -0.818267167f, 0.f),
    C_Vertex3f(0.117142916f, 0.0261178017f, 0.f),
    C_Vertex3f(0.117142916f, 0.0261178017f, -1.f),
    C_Normal3f(-0.657290518f, -0.753637195f, 0.f),
    C_Vertex3f(0.101475716f, 0.0384044647f, 0.f),
    C_Vertex3f(0.101475716f, 0.0384044647f, -1.f),
    C_Normal3f(-0.728475571f, -0.685071766f, 0.f),
    C_Vertex3f(0.0869550705f, 0.0524463654f, 0.f),
    C_Vertex3f(0.0869550705f, 0.0524463654f, -1.f),
    C_Normal3f(-0.792730033f, -0.609572947f, 0.f),
    C_Vertex3f(0.074249506f, 0.0673556328f, 0.f),
    C_Vertex3f(0.074249506f, 0.0673556328f, -1.f),
    C_Normal3f(-0.85148561f, -0.524377942f, 0.f),
    C_Vertex3f(0.0633590221f, 0.083132267f, 0.f),
    C_Vertex3f(0.0633590221f, 0.083132267f, -1.f),
    C_Normal3f(-0.902128577f, -0.431467295f, 0.f),
    C_Vertex3f(0.0542836189f, 0.099776268f, 0.f),
    C_Vertex3f(0.0542836189f, 0.099776268f, -1.f),
    C_Normal3f(-0.942672729f, -0.333718449f, 0.f),
    C_Vertex3f(0.0470232964f, 0.117287636f, 0.f),
    C_Vertex3f(0.0470232964f, 0.117287636f, -1.f),
    C_Normal3f(-0.972120225f, -0.234482959f, 0.f),
    C_Vertex3f(0.0415780544f, 0.13566637f, 0.f),
    C_Vertex3f(0.0415780544f, 0.13566637f, -1.f),
    C_Normal3f(-0.99056536f, -0.137040824f, 0.f),
    C_Vertex3f(0.0379478931f, 0.154912472f, 0.f),
    C_Vertex3f(0.0379478931f, 0.154912472f, -1.f),
    C_Normal3f(-0.998885691f, -0.0471937433f, 0.f),
    C_Vertex3f(0.0361328125f, 0.17502594f, 0.f),
    C_Vertex3f(0.0361328125f, 0.17502594f, -1.f),
    C_Normal3f(-0.999553561f, 0.0298789237f, 0.f),
    C_Vertex3f(0.0361328125f, 0.193329811f, 0.f),
    C_Vertex3f(0.0361328125f, 0.193329811f, -1.f),
    C_Normal3f(-0.994864404f, 0.101216421f, 0.f),
    C_Vertex3f(0.0371280909f, 0.208321452f, 0.f),
    C_Vertex3f(0.0371280909f, 0.208321452f, -1.f),
    C_Normal3f(-0.984654069f, 0.174517944f, 0.f),
    C_Vertex3f(0.0391186476f, 0.222677827f, 0.f),
    C_Vertex3f(0.0391186476f, 0.222677827f, -1.f),
    C_Normal3f(-0.967847645f, 0.251537174f, 0.f),
    C_Vertex3f(0.0421044827f, 0.236398935f, 0.f),
    C_Vertex3f(0.0421044827f, 0.236398935f, -1.f),
    C_Normal3f(-0.943630457f, 0.331000805f, 0.f),
    C_Vertex3f(0.0460855961f, 0.249484777f, 0.f),
    C_Vertex3f(0.0460855961f, 0.249484777f, -1.f),
    C_Normal3f(-0.911521673f, 0.411252081f, 0.f),
    C_Vertex3f(0.0510619879f, 0.261935353f, 0.f),
    C_Vertex3f(0.0510619879f, 0.261935353f, -1.f),
    C_Normal3f(-0.871515036f, 0.490368694f, 0.f),
    C_Vertex3f(0.057033658f, 0.273750663f, 0.f),
    C_Vertex3f(0.057033658f, 0.273750663f, -1.f),
    C_Normal3f(-0.824159503f, 0.566357851f, 0.f),
    C_Vertex3f(0.0640006065f, 0.284930706f, 0.f),
    C_Vertex3f(0.0640006065f, 0.284930706f, -1.f),
    C_Normal3f(-0.770195901f, 0.637807369f, 0.f),
    C_Vertex3f(0.0719628334f, 0.295475483f, 0.f),
    C_Vertex3f(0.0719628334f, 0.295475483f, -1.f),
    C_Normal3f(-0.709238887f, 0.704968095f, 0.f),
    C_Vertex3f(0.0807694197f, 0.305180192f, 0.f),
    C_Vertex3f(0.0807694197f, 0.305180192f, -1.f),
    C_Normal3f(-0.64244163f, 0.766334653f, 0.f),
    C_Vertex3f(0.0904203653f, 0.314044833f, 0.f),
    C_Vertex3f(0.0904203653f, 0.314044833f, -1.f),
    C_Normal3f(-0.571559072f, 0.820560873f, 0.f),
    C_Vertex3f(0.10091567f, 0.322069407f, 0.f),
    C_Vertex3f(0.10091567f, 0.322069407f, -1.f),
    C_Normal3f(-0.498549283f, 0.866861343f, 0.f),
    C_Vertex3f(0.112255335f, 0.329253912f, 0.f),
    C_Vertex3f(0.112255335f, 0.329253912f, -1.f),
    C_Normal3f(-0.425331026f, 0.90503782f, 0.f),
    C_Vertex3f(0.124439359f, 0.33559835f, 0.f),
    C_Vertex3f(0.124439359f, 0.33559835f, -1.f),
    C_Normal3f(-0.353583843f, 0.935402751f, 0.f),
    C_Vertex3f(0.137467742f, 0.341102719f, 0.f),
    C_Vertex3f(0.137467742f, 0.341102719f, -1.f),
    C_Normal3f(-0.295881957f, 0.955224514f, 0.f),
    C_Vertex3f(0.151340485f, 0.345767021f, 0.f),
    C_Vertex3f(0.151340485f, 0.345767021f, -1.f),
    C_Normal3f(-0.251497567f, 0.967857957f, 0.f),
    C_Vertex3f(0.158699036f, 0.347679138f, 0.f),
    C_Vertex3f(0.158699036f, 0.347679138f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.998984277f, -0.045059707f, 0.f),
    C_Vertex3f(0.143066406f, 0.485905647f, 0.f),
    C_Vertex3f(0.143066406f, 0.485905647f, -1.f),
    C_Normal3f(0.999116063f, 0.0420362838f, 0.f),
    C_Vertex3f(0.143066406f, 0.475216866f, 0.f),
    C_Vertex3f(0.143066406f, 0.475216866f, -1.f),
    C_Normal3f(0.991844535f, 0.127453804f, 0.f),
    C_Vertex3f(0.143971562f, 0.464391947f, 0.f),
    C_Vertex3f(0.143971562f, 0.464391947f, -1.f),
    C_Normal3f(0.975565374f, 0.219709679f, 0.f),
    C_Vertex3f(0.145781875f, 0.454085112f, 0.f),
    C_Vertex3f(0.145781875f, 0.454085112f, -1.f),
    C_Normal3f(0.948937535f, 0.315463781f, 0.f),
    C_Vertex3f(0.148497343f, 0.44429636f, 0.f),
    C_Vertex3f(0.148497343f, 0.44429636f, -1.f),
    C_Normal3f(0.91124022f, 0.411875278f, 0.f),
    C_Vertex3f(0.152117968f, 0.435025692f, 0.f),
    C_Vertex3f(0.152117968f, 0.435025692f, -1.f),
    C_Normal3f(0.862725139f, 0.505672991f, 0.f),
    C_Vertex3f(0.156643748f, 0.426273108f, 0.f),
    C_Vertex3f(0.156643748f, 0.426273108f, -1.f),
    C_Normal3f(0.804725111f, 0.593647599f, 0.f),
    C_Vertex3f(0.162074685f, 0.418038607f, 0.f),
    C_Vertex3f(0.162074685f, 0.418038607f, -1.f),
    C_Normal3f(0.739483893f, 0.673174202f, 0.f),
    C_Vertex3f(0.168410778f, 0.410322189f, 0.f),
    C_Vertex3f(0.168410778f, 0.410322189f, -1.f),
    C_Normal3f(0.667975485f, 0.744183242f, 0.f),
    C_Vertex3f(0.175652027f, 0.403123856f, 0.f),
    C_Vertex3f(0.175652027f, 0.403123856f, -1.f),
    C_Normal3f(0.586707413f, 0.809799016f, 0.f),
    C_Vertex3f(0.18344748f, 0.396825314f, 0.f),
    C_Vertex3f(0.18344748f, 0.396825314f, -1.f),
    C_Normal3f(0.497599483f, 0.867406964f, 0.f),
    C_Vertex3f(0.191797137f, 0.391426563f, 0.f),
    C_Vertex3f(0.191797137f, 0.391426563f, -1.f),
    C_Normal3f(0.403526545f, 0.914967895f, 0.f),
    C_Vertex3f(0.200700998f, 0.386927605f, 0.f),
    C_Vertex3f(0.200700998f, 0.386927605f, -1.f),
    C_Normal3f(0.307790041f, 0.951454282f, 0.f),
    C_Vertex3f(0.210159063f, 0.383328438f, 0.f),
    C_Vertex3f(0.210159063f, 0.383328438f, -1.f),
    C_Normal3f(0.213577211f, 0.976926148f, 0.f),
    C_Vertex3f(0.220171332f, 0.380629063f, 0.f),
    C_Vertex3f(0.220171332f, 0.380629063f, -1.f),
    C_Normal3f(0.12351577f, 0.992342651f, 0.f),
    C_Vertex3f(0.230737805f, 0.378829479f, 0.f),
    C_Vertex3f(0.230737805f, 0.378829479f, -1.f),
    C_Normal3f(0.0397252738f, 0.999210656f, 0.f),
    C_Vertex3f(0.241858482f, 0.377929688f, 0.f),
    C_Vertex3f(0.241858482f, 0.377929688f, -1.f),
    C_Normal3f(-0.0399405695f, 0.999202073f, 0.f),
    C_Vertex3f(0.253370285f, 0.377929688f, 0.f),
    C_Vertex3f(0.253370285f, 0.377929688f, -1.f),
    C_Normal3f(-0.125739619f, 0.992063344f, 0.f),
    C_Vertex3f(0.26419878f, 0.378822684f, 0.f),
    C_Vertex3f(0.26419878f, 0.378822684f, -1.f),
    C_Normal3f(-0.216890603f, 0.976195931f, 0.f),
    C_Vertex3f(0.274507046f, 0.380608678f, 0.f),
    C_Vertex3f(0.274507046f, 0.380608678f, -1.f),
    C_Normal3f(-0.311692953f, 0.950182855f, 0.f),
    C_Vertex3f(0.284295082f, 0.383287668f, 0.f),
    C_Vertex3f(0.284295082f, 0.383287668f, -1.f),
    C_Normal3f(-0.407413632f, 0.913243711f, 0.f),
    C_Vertex3f(0.293562889f, 0.386859655f, 0.f),
    C_Vertex3f(0.293562889f, 0.386859655f, -1.f),
    C_Normal3f(-0.500859618f, 0.865528464f, 0.f),
    C_Vertex3f(0.302310467f, 0.391324639f, 0.f),
    C_Vertex3f(0.302310467f, 0.391324639f, -1.f),
    C_Normal3f(-0.588843465f, 0.808247089f, 0.f),
    C_Vertex3f(0.310537815f, 0.39668262f, 0.f),
    C_Vertex3f(0.310537815f, 0.39668262f, -1.f),
    C_Normal3f(-0.66869837f, 0.74353379f, 0.f),
    C_Vertex3f(0.318244934f, 0.402933598f, 0.f),
    C_Vertex3f(0.318244934f, 0.402933598f, -1.f),
    C_Normal3f(-0.736915588f, 0.675984859f, 0.f),
    C_Vertex3f(0.325431824f, 0.410077572f, 0.f),
    C_Vertex3f(0.325431824f, 0.410077572f, -1.f),
    C_Normal3f(-0.798562229f, 0.60191226f, 0.f),
    C_Vertex3f(0.331720352f, 0.417623639f, 0.f),
    C_Vertex3f(0.331720352f, 0.417623639f, -1.f),
    C_Normal3f(-0.855101407f, 0.51846087f, 0.f),
    C_Vertex3f(0.337110519f, 0.425571799f, 0.f),
    C_Vertex3f(0.337110519f, 0.425571799f, -1.f),
    C_Normal3f(-0.904064357f, 0.427396268f, 0.f),
    C_Vertex3f(0.341602325f, 0.433922052f, 0.f),
    C_Vertex3f(0.341602325f, 0.433922052f, -1.f),
    C_Normal3f(-0.943510234f, 0.331343174f, 0.f),
    C_Vertex3f(0.34519577f, 0.442674398f, 0.f),
    C_Vertex3f(0.34519577f, 0.442674398f, -1.f),
    C_Normal3f(-0.972373486f, 0.233430594f, 0.f),
    C_Vertex3f(0.347890854f, 0.451828837f, 0.f),
    C_Vertex3f(0.347890854f, 0.451828837f, -1.f),
    C_Normal3f(-0.990598202f, 0.136803657f, 0.f),
    C_Vertex3f(0.349687576f, 0.461385369f, 0.f),
    C_Vertex3f(0.349687576f, 0.461385369f, -1.f),
    C_Normal3f(-0.999044776f, 0.0437001996f, 0.f),
    C_Vertex3f(0.350585938f, 0.471343994f, 0.f),
    C_Vertex3f(0.350585938f, 0.471343994f, -1.f),
    C_Normal3f(-0.999025285f, -0.0441413634f, 0.f),
    C_Vertex3f(0.350585938f, 0.481923103f, 0.f),
    C_Vertex3f(0.350585938f, 0.481923103f, -1.f),
    C_Normal3f(-0.990781069f, -0.135472938f, 0.f),
    C_Vertex3f(0.349660397f, 0.492291212f, 0.f),
    C_Vertex3f(0.349660397f, 0.492291212f, -1.f),
    C_Normal3f(-0.972837567f, -0.231488854f, 0.f),
    C_Vertex3f(0.347809315f, 0.502229929f, 0.f),
    C_Vertex3f(0.347809315f, 0.502229929f, -1.f),
    C_Normal3f(-0.944292367f, -0.329107732f, 0.f),
    C_Vertex3f(0.345032692f, 0.511739254f, 0.f),
    C_Vertex3f(0.345032692f, 0.511739254f, -1.f),
    C_Normal3f(-0.905091584f, -0.425216705f, 0.f),
    C_Vertex3f(0.341330528f, 0.520819187f, 0.f),
    C_Vertex3f(0.341330528f, 0.520819187f, -1.f),
    C_Normal3f(-0.856193066f, -0.516655922f, 0.f),
    C_Vertex3f(0.336702824f, 0.529469728f, 0.f),
    C_Vertex3f(0.336702824f, 0.529469728f, -1.f),
    C_Normal3f(-0.799464047f, -0.600713968f, 0.f),
    C_Vertex3f(0.331149578f, 0.537690878f, 0.f),
    C_Vertex3f(0.331149578f, 0.537690878f, -1.f),
    C_Normal3f(-0.737352014f, -0.675508678f, 0.f),
    C_Vertex3f(0.324670792f, 0.545482635f, 0.f),
    C_Vertex3f(0.324670792f, 0.545482635f, -1.f),
    C_Normal3f(-0.670502663f, -0.74190712f, 0.f),
    C_Vertex3f(0.317266464f, 0.552845001f, 0.f),
    C_Vertex3f(0.317266464f, 0.552845001f, -1.f),
    C_Normal3f(-0.593919098f, -0.804524839f, 0.f),
    C_Vertex3f(0.309396267f, 0.559287071f, 0.f),
    C_Vertex3f(0.309396267f, 0.559287071f, -1.f),
    C_Normal3f(-0.508589685f, -0.861009061f, 0.f),
    C_Vertex3f(0.3010602f, 0.564808846f, 0.f),
    C_Vertex3f(0.3010602f, 0.564808846f, -1.f),
    C_Normal3f(-0.416683614f, -0.909051538f, 0.f),
    C_Vertex3f(0.292258263f, 0.569410324f, 0.f),
    C_Vertex3f(0.292258263f, 0.569410324f, -1.f),
    C_Normal3f(-0.321079046f, -0.947052419f, 0.f),
    C_Vertex3f(0.282990456f, 0.573091507f, 0.f),
    C_Vertex3f(0.282990456f, 0.573091507f, -1.f),
    C_Normal3f(-0.224929318f, -0.974375129f, 0.f),
    C_Vertex3f(0.273256779f, 0.575852394f, 0.f),
    C_Vertex3f(0.273256779f, 0.575852394f, -1.f),
    C_Normal3f(-0.13117826f, -0.991358757f, 0.f),
    C_Vertex3f(0.263057232f, 0.577692986f, 0.f),
    C_Vertex3f(0.263057232f, 0.577692986f, -1.f),
    C_Normal3f(-0.0420794412f, -0.999114215f, 0.f),
    C_Vertex3f(0.252391815f, 0.578613281f, 0.f),
    C_Vertex3f(0.252391815f, 0.578613281f, -1.f),
    C_Normal3f(0.0409781486f, -0.999159992f, 0.f),
    C_Vertex3f(0.241206169f, 0.578613281f, 0.f),
    C_Vertex3f(0.241206169f, 0.578613281f, -1.f),
    C_Normal3f(0.127365947f, -0.9918558f, 0.f),
    C_Vertex3f(0.230452418f, 0.57771349f, 0.f),
    C_Vertex3f(0.230452418f, 0.57771349f, -1.f),
    C_Normal3f(0.218955949f, -0.97573477f, 0.f),
    C_Vertex3f(0.220184922f, 0.575913906f, 0.f),
    C_Vertex3f(0.220184922f, 0.575913906f, -1.f),
    C_Normal3f(0.313529462f, -0.949578464f, 0.f),
    C_Vertex3f(0.210403681f, 0.573214531f, 0.f),
    C_Vertex3f(0.210403681f, 0.573214531f, -1.f),
    C_Normal3f(0.408327997f, -0.9128353f, 0.f),
    C_Vertex3f(0.201108694f, 0.569615364f, 0.f),
    C_Vertex3f(0.201108694f, 0.569615364f, -1.f),
    C_Normal3f(0.50026536f, -0.865872085f, 0.f),
    C_Vertex3f(0.192299962f, 0.565116405f, 0.f),
    C_Vertex3f(0.192299962f, 0.565116405f, -1.f),
    C_Normal3f(0.5863837f, -0.8100335f, 0.f),
    C_Vertex3f(0.183977485f, 0.559717655f, 0.f),
    C_Vertex3f(0.183977485f, 0.559717655f, -1.f),
    C_Normal3f(0.664309919f, -0.747457206f, 0.f),
    C_Vertex3f(0.176141262f, 0.553419113f, 0.f),
    C_Vertex3f(0.176141262f, 0.553419113f, -1.f),
    C_Normal3f(0.730723977f, -0.682673156f, 0.f),
    C_Vertex3f(0.168791294f, 0.546220779f, 0.f),
    C_Vertex3f(0.168791294f, 0.546220779f, -1.f),
    C_Normal3f(0.791355908f, -0.611355841f, 0.f),
    C_Vertex3f(0.162360072f, 0.538667917f, 0.f),
    C_Vertex3f(0.162360072f, 0.538667917f, -1.f),
    C_Normal3f(0.847989142f, -0.530013502f, 0.f),
    C_Vertex3f(0.156847596f, 0.530760527f, 0.f),
    C_Vertex3f(0.156847596f, 0.530760527f, -1.f),
    C_Normal3f(0.898025692f, -0.439942896f, 0.f),
    C_Vertex3f(0.152253866f, 0.522498608f, 0.f),
    C_Vertex3f(0.152253866f, 0.522498608f, -1.f),
    C_Normal3f(0.93917793f, -0.343430758f, 0.f),
    C_Vertex3f(0.148578882f, 0.51388216f, 0.f),
    C_Vertex3f(0.148578882f, 0.51388216f, -1.f),
    C_Normal3f(0.969897449f, -0.243513882f, 0.f),
    C_Vertex3f(0.145822644f, 0.504911184f, 0.f),
    C_Vertex3f(0.145822644f, 0.504911184f, -1.f),
    C_Normal3f(0.98964721f, -0.143521503f, 0.f),
    C_Vertex3f(0.143985152f, 0.49558568f, 0.f),
    C_Vertex3f(0.143985152f, 0.49558568f, -1.f),
    C_Normal3f(0.998984277f, -0.045059707f, 0.f),
    C_Vertex3f(0.143066406f, 0.485905647f, 0.f),
    C_Vertex3f(0.143066406f, 0.485905647f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999213278f, -0.0396613367f, 0.f),
    C_Vertex3f(0.1171875f, 0.191793442f, 0.f),
    C_Vertex3f(0.1171875f, 0.191793442f, -1.f),
    C_Normal3f(0.998095453f, 0.0616884455f, 0.f),
    C_Vertex3f(0.1171875f, 0.176486969f, 0.f),
    C_Vertex3f(0.1171875f, 0.176486969f, -1.f),
    C_Normal3f(0.984206617f, 0.1770235f, 0.f),
    C_Vertex3f(0.119157314f, 0.1599226f, 0.f),
    C_Vertex3f(0.119157314f, 0.1599226f, -1.f),
    C_Normal3f(0.956539273f, 0.291603714f, 0.f),
    C_Vertex3f(0.123096943f, 0.143631935f, 0.f),
    C_Vertex3f(0.123096943f, 0.143631935f, -1.f),
    C_Normal3f(0.917282045f, 0.398238093f, 0.f),
    C_Vertex3f(0.129006386f, 0.127614975f, 0.f),
    C_Vertex3f(0.129006386f, 0.127614975f, -1.f),
    C_Normal3f(0.858837605f, 0.51224798f, 0.f),
    C_Vertex3f(0.136885643f, 0.111871719f, 0.f),
    C_Vertex3f(0.136885643f, 0.111871719f, -1.f),
    C_Normal3f(0.770438612f, 0.637514293f, 0.f),
    C_Vertex3f(0.146726608f, 0.097905159f, 0.f),
    C_Vertex3f(0.146726608f, 0.097905159f, -1.f),
    C_Normal3f(0.662344098f, 0.749199748f, 0.f),
    C_Vertex3f(0.158529282f, 0.0857152939f, 0.f),
    C_Vertex3f(0.158529282f, 0.0857152939f, -1.f),
    C_Normal3f(0.542600095f, 0.839991152f, 0.f),
    C_Vertex3f(0.172293663f, 0.075302124f, 0.f),
    C_Vertex3f(0.172293663f, 0.075302124f, -1.f),
    C_Normal3f(0.426683724f, 0.904400945f, 0.f),
    C_Vertex3f(0.188019753f, 0.0666656494f, 0.f),
    C_Vertex3f(0.188019753f, 0.0666656494f, -1.f),
    C_Normal3f(0.309216678f, 0.950991571f, 0.f),
    C_Vertex3f(0.204329014f, 0.0601882935f, 0.f),
    C_Vertex3f(0.204329014f, 0.0601882935f, -1.f),
    C_Normal3f(0.185209617f, 0.982699037f, 0.f),
    C_Vertex3f(0.221221447f, 0.0558700562f, 0.f),
    C_Vertex3f(0.221221447f, 0.0558700562f, -1.f),
    C_Normal3f(0.0642191693f, 0.997935891f, 0.f),
    C_Vertex3f(0.238697052f, 0.0537109375f, 0.f),
    C_Vertex3f(0.238697052f, 0.0537109375f, -1.f),
    C_Normal3f(-0.0383617431f, 0.999263942f, 0.f),
    C_Vertex3f(0.25477314f, 0.0537109375f, 0.f),
    C_Vertex3f(0.25477314f, 0.0537109375f, -1.f),
    C_Normal3f(-0.128235176f, 0.991743803f, 0.f),
    C_Vertex3f(0.268243194f, 0.0548452139f, 0.f),
    C_Vertex3f(0.268243194f, 0.0548452139f, -1.f),
    C_Normal3f(-0.220645785f, 0.975354075f, 0.f),
    C_Vertex3f(0.281089902f, 0.0571137667f, 0.f),
    C_Vertex3f(0.281089902f, 0.0571137667f, -1.f),
    C_Normal3f(-0.316185027f, 0.948697507f, 0.f),
    C_Vertex3f(0.293313265f, 0.0605165958f, 0.f),
    C_Vertex3f(0.293313265f, 0.0605165958f, -1.f),
    C_Normal3f(-0.412007838f, 0.911180317f, 0.f),
    C_Vertex3f(0.304913282f, 0.0650537014f, 0.f),
    C_Vertex3f(0.304913282f, 0.0650537014f, -1.f),
    C_Normal3f(-0.504913568f, 0.863169909f, 0.f),
    C_Vertex3f(0.315889955f, 0.0707250834f, 0.f),
    C_Vertex3f(0.315889955f, 0.0707250834f, -1.f),
    C_Normal3f(-0.591829062f, 0.806063533f, 0.f),
    C_Vertex3f(0.326243281f, 0.0775307417f, 0.f),
    C_Vertex3f(0.326243281f, 0.0775307417f, -1.f),
    C_Normal3f(-0.670295537f, 0.742094278f, 0.f),
    C_Vertex3f(0.335973263f, 0.0854706764f, 0.f),
    C_Vertex3f(0.335973263f, 0.0854706764f, -1.f),
    C_Normal3f(-0.739600837f, 0.673045754f, 0.f),
    C_Vertex3f(0.345079899f, 0.0945448875f, 0.f),
    C_Vertex3f(0.345079899f, 0.0945448875f, -1.f),
    C_Normal3f(-0.80378139f, 0.594924808f, 0.f),
    C_Vertex3f(0.353048205f, 0.104234099f, 0.f),
    C_Vertex3f(0.353048205f, 0.104234099f, -1.f),
    C_Normal3f(-0.861272812f, 0.508142948f, 0.f),
    C_Vertex3f(0.359878182f, 0.114538312f, 0.f),
    C_Vertex3f(0.359878182f, 0.114538312f, -1.f),
    C_Normal3f(-0.909772396f, 0.415107518f, 0.f),
    C_Vertex3f(0.36556983f, 0.125457525f, 0.f),
    C_Vertex3f(0.36556983f, 0.125457525f, -1.f),
    C_Normal3f(-0.947793186f, 0.318885714f, 0.f),
    C_Vertex3f(0.370123148f, 0.136991739f, 0.f),
    C_Vertex3f(0.370123148f, 0.136991739f, -1.f),
    C_Normal3f(-0.974882782f, 0.222718805f, 0.f),
    C_Vertex3f(0.373538136f, 0.149140954f, 0.f),
    C_Vertex3f(0.373538136f, 0.149140954f, -1.f),
    C_Normal3f(-0.991576195f, 0.129524738f, 0.f),
    C_Vertex3f(0.375814795f, 0.161905169f, 0.f),
    C_Vertex3f(0.375814795f, 0.161905169f, -1.f),
    C_Normal3f(-0.999143362f, 0.0413837656f, 0.f),
    C_Vertex3f(0.376953125f, 0.175284386f, 0.f),
    C_Vertex3f(0.376953125f, 0.175284386f, -1.f),
    C_Normal3f(-0.999105573f, -0.0422843359f, 0.f),
    C_Vertex3f(0.376953125f, 0.189388275f, 0.f),
    C_Vertex3f(0.376953125f, 0.189388275f, -1.f),
    C_Normal3f(-0.991359115f, -0.131175563f, 0.f),
    C_Vertex3f(0.375780702f, 0.202986717f, 0.f),
    C_Vertex3f(0.375780702f, 0.202986717f, -1.f),
    C_Normal3f(-0.97429204f, -0.225288779f, 0.f),
    C_Vertex3f(0.373435855f, 0.215970039f, 0.f),
    C_Vertex3f(0.373435855f, 0.215970039f, -1.f),
    C_Normal3f(-0.946704209f, -0.322104186f, 0.f),
    C_Vertex3f(0.369918585f, 0.228338242f, 0.f),
    C_Vertex3f(0.369918585f, 0.228338242f, -1.f),
    C_Normal3f(-0.908159852f, -0.418623537f, 0.f),
    C_Vertex3f(0.365228891f, 0.240091324f, 0.f),
    C_Vertex3f(0.365228891f, 0.240091324f, -1.f),
    C_Normal3f(-0.859234989f, -0.511581123f, 0.f),
    C_Vertex3f(0.359366775f, 0.251229286f, 0.f),
    C_Vertex3f(0.359366775f, 0.251229286f, -1.f),
    C_Normal3f(-0.801530659f, -0.597953737f, 0.f),
    C_Vertex3f(0.352332234f, 0.261752129f, 0.f),
    C_Vertex3f(0.352332234f, 0.261752129f, -1.f),
    C_Normal3f(-0.737420917f, -0.675433457f, 0.f),
    C_Vertex3f(0.344125271f, 0.271659851f, 0.f),
    C_Vertex3f(0.344125271f, 0.271659851f, -1.f),
    C_Normal3f(-0.66892606f, -0.743328929f, 0.f),
    C_Vertex3f(0.334745884f, 0.280952454f, 0.f),
    C_Vertex3f(0.334745884f, 0.280952454f, -1.f),
    C_Normal3f(-0.591817081f, -0.806072295f, 0.f),
    C_Vertex3f(0.324763656f, 0.289083481f, 0.f),
    C_Vertex3f(0.324763656f, 0.289083481f, -1.f),
    C_Normal3f(-0.506127954f, -0.862458408f, 0.f),
    C_Vertex3f(0.314178586f, 0.296052933f, 0.f),
    C_Vertex3f(0.314178586f, 0.296052933f, -1.f),
    C_Normal3f(-0.41410926f, -0.910227239f, 0.f),
    C_Vertex3f(0.302990675f, 0.301860809f, 0.f),
    C_Vertex3f(0.302990675f, 0.301860809f, -1.f),
    C_Normal3f(-0.318681031f, -0.947862089f, 0.f),
    C_Vertex3f(0.291199923f, 0.306507111f, 0.f),
    C_Vertex3f(0.291199923f, 0.306507111f, -1.f),
    C_Normal3f(-0.222986847f, -0.974821389f, 0.f),
    C_Vertex3f(0.278806329f, 0.309991837f, 0.f),
    C_Vertex3f(0.278806329f, 0.309991837f, -1.f),
    C_Normal3f(-0.129915431f, -0.991525114f, 0.f),
    C_Vertex3f(0.265809894f, 0.312314987f, 0.f),
    C_Vertex3f(0.265809894f, 0.312314987f, -1.f),
    C_Normal3f(-0.041990649f, -0.99911803f, 0.f),
    C_Vertex3f(0.252210617f, 0.313476563f, 0.f),
    C_Vertex3f(0.252210617f, 0.313476563f, -1.f),
    C_Normal3f(0.0419951305f, -0.999117851f, 0.f),
    C_Vertex3f(0.238171577f, 0.313476563f, 0.f),
    C_Vertex3f(0.238171577f, 0.313476563f, -1.f),
    C_Normal3f(0.131596208f, -0.991303444f, 0.f),
    C_Vertex3f(0.224898577f, 0.312328577f, 0.f),
    C_Vertex3f(0.224898577f, 0.312328577f, -1.f),
    C_Normal3f(0.226031855f, -0.974119842f, 0.f),
    C_Vertex3f(0.212228537f, 0.310032606f, 0.f),
    C_Vertex3f(0.212228537f, 0.310032606f, -1.f),
    C_Normal3f(0.323174655f, -0.946339369f, 0.f),
    C_Vertex3f(0.200161457f, 0.30658865f, 0.f),
    C_Vertex3f(0.200161457f, 0.30658865f, -1.f),
    C_Normal3f(0.419992745f, -0.907527387f, 0.f),
    C_Vertex3f(0.188697338f, 0.301996708f, 0.f),
    C_Vertex3f(0.188697338f, 0.301996708f, -1.f),
    C_Normal3f(0.51318711f, -0.858276784f, 0.f),
    C_Vertex3f(0.17783618f, 0.296256781f, 0.f),
    C_Vertex3f(0.17783618f, 0.296256781f, -1.f),
    C_Normal3f(0.599710226f, -0.800217211f, 0.f),
    C_Vertex3f(0.167577982f, 0.289368868f, 0.f),
    C_Vertex3f(0.167577982f, 0.289368868f, -1.f),
    C_Normal3f(0.677245438f, -0.735757232f, 0.f),
    C_Vertex3f(0.157922745f, 0.28133297f, 0.f),
    C_Vertex3f(0.157922745f, 0.28133297f, -1.f),
    C_Normal3f(0.744748056f, -0.667345703f, 0.f),
    C_Vertex3f(0.148870468f, 0.272149086f, 0.f),
    C_Vertex3f(0.148870468f, 0.272149086f, -1.f),
    C_Normal3f(0.8068313f, -0.590781927f, 0.f),
    C_Vertex3f(0.140949726f, 0.262391329f, 0.f),
    C_Vertex3f(0.140949726f, 0.262391329f, -1.f),
    C_Normal3f(0.862741828f, -0.505644679f, 0.f),
    C_Vertex3f(0.134160519f, 0.252059698f, 0.f),
    C_Vertex3f(0.134160519f, 0.252059698f, -1.f),
    C_Normal3f(0.910230637f, -0.41410166f, 0.f),
    C_Vertex3f(0.128502846f, 0.241154194f, 0.f),
    C_Vertex3f(0.128502846f, 0.241154194f, -1.f),
    C_Normal3f(0.947754502f, -0.319000423f, 0.f),
    C_Vertex3f(0.123976707f, 0.229674816f, 0.f),
    C_Vertex3f(0.123976707f, 0.229674816f, -1.f),
    C_Normal3f(0.974717259f, -0.223441944f, 0.f),
    C_Vertex3f(0.120582104f, 0.217621565f, 0.f),
    C_Vertex3f(0.120582104f, 0.217621565f, -1.f),
    C_Normal3f(0.991473377f, -0.130309865f, 0.f),
    C_Vertex3f(0.118319035f, 0.20499444f, 0.f),
    C_Vertex3f(0.118319035f, 0.20499444f, -1.f),
    C_Normal3f(0.999213278f, -0.0396613367f, 0.f),
    C_Vertex3f(0.1171875f, 0.191793442f, 0.f),
    C_Vertex3f(0.1171875f, 0.191793442f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158699036f, 0.347679138f, 0.f),
    C_Vertex3f(0.151340485f, 0.345767021f, 0.f),
    C_Vertex3f(0.157922745f, 0.28133297f, 0.f),
    C_Vertex3f(0.167577982f, 0.289368868f, 0.f),
    C_Vertex3f(0.162074685f, 0.418038607f, 0.f),
    C_Vertex3f(0.156643748f, 0.426273108f, 0.f),
    C_Vertex3f(0.152117968f, 0.435025692f, 0.f),
    C_Vertex3f(0.148497343f, 0.44429636f, 0.f),
    C_Vertex3f(0.146392822f, 0.352151871f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.146392822f, 0.352151871f, 0.f),
    C_Vertex3f(0.148497343f, 0.44429636f, 0.f),
    C_Vertex3f(0.145781875f, 0.454085112f, 0.f),
    C_Vertex3f(0.143971562f, 0.464391947f, 0.f),
    C_Vertex3f(0.143066406f, 0.475216866f, 0.f),
    C_Vertex3f(0.13205409f, 0.61185658f, 0.f),
    C_Vertex3f(0.124967098f, 0.363005638f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118471146f, 0.601453304f, 0.f),
    C_Vertex3f(0.105924606f, 0.589563847f, 0.f),
    C_Vertex3f(0.106728077f, 0.375767708f, 0.f),
    C_Vertex3f(0.124967098f, 0.363005638f, 0.f),
    C_Vertex3f(0.13205409f, 0.61185658f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0776948929f, 0.549723387f, 0.f),
    C_Vertex3f(0.0798101425f, 0.407016754f, 0.f),
    C_Vertex3f(0.0855364799f, 0.563698888f, 0.f),
    C_Vertex3f(0.0916757584f, 0.39043808f, 0.f),
    C_Vertex3f(0.0949463844f, 0.576979041f, 0.f),
    C_Vertex3f(0.106728077f, 0.375767708f, 0.f),
    C_Vertex3f(0.105924606f, 0.589563847f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0635800362f, 0.503624797f, 0.f),
    C_Vertex3f(0.0649781227f, 0.445086479f, 0.f),
    C_Vertex3f(0.066716671f, 0.519686341f, 0.f),
    C_Vertex3f(0.0709109306f, 0.425232887f, 0.f),
    C_Vertex3f(0.0714216232f, 0.535052538f, 0.f),
    C_Vertex3f(0.0798101425f, 0.407016754f, 0.f),
    C_Vertex3f(0.0776948929f, 0.549723387f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0620117188f, 0.486867905f, 0.f),
    C_Vertex3f(0.0620117188f, 0.46657753f, 0.f),
    C_Vertex3f(0.0649781227f, 0.445086479f, 0.f),
    C_Vertex3f(0.0635800362f, 0.503624797f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.452218533f, 0.232228994f, 0.f),
    C_Vertex3f(0.452562571f, 0.134429336f, 0.f),
    C_Vertex3f(0.455113173f, 0.218924403f, 0.f),
    C_Vertex3f(0.456192732f, 0.153332829f, 0.f),
    C_Vertex3f(0.457042933f, 0.205100656f, 0.f),
    C_Vertex3f(0.458007813f, 0.173055649f, 0.f),
    C_Vertex3f(0.458007813f, 0.190757751f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.437745333f, 0.269027829f, 0.f),
    C_Vertex3f(0.439857006f, 0.0990803242f, 0.f),
    C_Vertex3f(0.443534613f, 0.257280707f, 0.f),
    C_Vertex3f(0.447117329f, 0.116345167f, 0.f),
    C_Vertex3f(0.448359013f, 0.245014429f, 0.f),
    C_Vertex3f(0.452562571f, 0.134429336f, 0.f),
    C_Vertex3f(0.452218533f, 0.232228994f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.405185461f, 0.310169578f, 0.f),
    C_Vertex3f(0.407185555f, 0.0522017479f, 0.f),
    C_Vertex3f(0.414670229f, 0.300935864f, 0.f),
    C_Vertex3f(0.419891119f, 0.0670086145f, 0.f),
    C_Vertex3f(0.423272133f, 0.290964603f, 0.f),
    C_Vertex3f(0.430781603f, 0.0826348066f, 0.f),
    C_Vertex3f(0.430991173f, 0.280255795f, 0.f),
    C_Vertex3f(0.439857006f, 0.0990803242f, 0.f),
    C_Vertex3f(0.437745333f, 0.269027829f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.376953125f, 0.189388275f, 0.f),
    C_Vertex3f(0.376997709f, 0.0259751081f, 0.f),
    C_Vertex3f(0.383567333f, 0.32642436f, 0.f),
    C_Vertex3f(0.392664909f, 0.0382142067f, 0.f),
    C_Vertex3f(0.394817829f, 0.318665743f, 0.f),
    C_Vertex3f(0.407185555f, 0.0522017479f, 0.f),
    C_Vertex3f(0.405185461f, 0.310169578f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.365228891f, 0.240091324f, 0.f),
    C_Vertex3f(0.369918585f, 0.228338242f, 0.f),
    C_Vertex3f(0.371433973f, 0.33344543f, 0.f),
    C_Vertex3f(0.373435855f, 0.215970039f, 0.f),
    C_Vertex3f(0.383567333f, 0.32642436f, 0.f),
    C_Vertex3f(0.375780702f, 0.202986717f, 0.f),
    C_Vertex3f(0.376953125f, 0.189388275f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.365228891f, 0.240091324f, 0.f),
    C_Vertex3f(0.371433973f, 0.33344543f, 0.f),
    C_Vertex3f(0.359366775f, 0.251229286f, 0.f),
    C_Vertex3f(0.358417749f, 0.339728951f, 0.f),
    C_Vertex3f(0.352332234f, 0.261752129f, 0.f),
    C_Vertex3f(0.344518661f, 0.345274925f, 0.f),
    C_Vertex3f(0.344125271f, 0.271659851f, 0.f),
    C_Vertex3f(0.337127686f, 0.347679138f, 0.f),
    C_Vertex3f(0.334745884f, 0.280952454f, 0.f),
    C_Vertex3f(0.337110519f, 0.425571799f, 0.f),
    C_Vertex3f(0.331720352f, 0.417623639f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.302310467f, 0.391324639f, 0.f),
    C_Vertex3f(0.302990675f, 0.301860809f, 0.f),
    C_Vertex3f(0.310537815f, 0.39668262f, 0.f),
    C_Vertex3f(0.314178586f, 0.296052933f, 0.f),
    C_Vertex3f(0.318244934f, 0.402933598f, 0.f),
    C_Vertex3f(0.324763656f, 0.289083481f, 0.f),
    C_Vertex3f(0.325431824f, 0.410077572f, 0.f),
    C_Vertex3f(0.334745884f, 0.280952454f, 0.f),
    C_Vertex3f(0.331720352f, 0.417623639f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.26419878f, 0.378822684f, 0.f),
    C_Vertex3f(0.265809894f, 0.312314987f, 0.f),
    C_Vertex3f(0.274507046f, 0.380608678f, 0.f),
    C_Vertex3f(0.278806329f, 0.309991837f, 0.f),
    C_Vertex3f(0.284295082f, 0.383287668f, 0.f),
    C_Vertex3f(0.291199923f, 0.306507111f, 0.f),
    C_Vertex3f(0.293562889f, 0.386859655f, 0.f),
    C_Vertex3f(0.302990675f, 0.301860809f, 0.f),
    C_Vertex3f(0.302310467f, 0.391324639f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.210159063f, 0.383328438f, 0.f),
    C_Vertex3f(0.212228537f, 0.310032606f, 0.f),
    C_Vertex3f(0.220171332f, 0.380629063f, 0.f),
    C_Vertex3f(0.224898577f, 0.312328577f, 0.f),
    C_Vertex3f(0.230737805f, 0.378829479f, 0.f),
    C_Vertex3f(0.238171577f, 0.313476563f, 0.f),
    C_Vertex3f(0.241858482f, 0.377929688f, 0.f),
    C_Vertex3f(0.252210617f, 0.313476563f, 0.f),
    C_Vertex3f(0.253370285f, 0.377929688f, 0.f),
    C_Vertex3f(0.265809894f, 0.312314987f, 0.f),
    C_Vertex3f(0.26419878f, 0.378822684f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.175652027f, 0.403123856f, 0.f),
    C_Vertex3f(0.17783618f, 0.296256781f, 0.f),
    C_Vertex3f(0.18344748f, 0.396825314f, 0.f),
    C_Vertex3f(0.188697338f, 0.301996708f, 0.f),
    C_Vertex3f(0.191797137f, 0.391426563f, 0.f),
    C_Vertex3f(0.200161457f, 0.30658865f, 0.f),
    C_Vertex3f(0.200700998f, 0.386927605f, 0.f),
    C_Vertex3f(0.212228537f, 0.310032606f, 0.f),
    C_Vertex3f(0.210159063f, 0.383328438f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.168410778f, 0.410322189f, 0.f),
    C_Vertex3f(0.162074685f, 0.418038607f, 0.f),
    C_Vertex3f(0.167577982f, 0.289368868f, 0.f),
    C_Vertex3f(0.17783618f, 0.296256781f, 0.f),
    C_Vertex3f(0.175652027f, 0.403123856f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.128502846f, 0.241154194f, 0.f),
    C_Vertex3f(0.134160519f, 0.252059698f, 0.f),
    C_Vertex3f(0.137467742f, 0.341102719f, 0.f),
    C_Vertex3f(0.140949726f, 0.262391329f, 0.f),
    C_Vertex3f(0.151340485f, 0.345767021f, 0.f),
    C_Vertex3f(0.148870468f, 0.272149086f, 0.f),
    C_Vertex3f(0.157922745f, 0.28133297f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.124439359f, 0.33559835f, 0.f),
    C_Vertex3f(0.112255335f, 0.329253912f, 0.f),
    C_Vertex3f(0.1171875f, 0.191793442f, 0.f),
    C_Vertex3f(0.118319035f, 0.20499444f, 0.f),
    C_Vertex3f(0.120582104f, 0.217621565f, 0.f),
    C_Vertex3f(0.123976707f, 0.229674816f, 0.f),
    C_Vertex3f(0.128502846f, 0.241154194f, 0.f),
    C_Vertex3f(0.137467742f, 0.341102719f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.123096943f, 0.143631935f, 0.f),
    C_Vertex3f(0.119157314f, 0.1599226f, 0.f),
    C_Vertex3f(0.133956671f, 0.0155863762f, 0.f),
    C_Vertex3f(0.1171875f, 0.176486969f, 0.f),
    C_Vertex3f(0.117142916f, 0.0261178017f, 0.f),
    C_Vertex3f(0.1171875f, 0.191793442f, 0.f),
    C_Vertex3f(0.112255335f, 0.329253912f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.101475716f, 0.0384044647f, 0.f),
    C_Vertex3f(0.117142916f, 0.0261178017f, 0.f),
    C_Vertex3f(0.112255335f, 0.329253912f, 0.f),
    C_Vertex3f(0.10091567f, 0.322069407f, 0.f),
    C_Vertex3f(0.0904203653f, 0.314044833f, 0.f),
    C_Vertex3f(0.0869550705f, 0.0524463654f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0807694197f, 0.305180192f, 0.f),
    C_Vertex3f(0.0719628334f, 0.295475483f, 0.f),
    C_Vertex3f(0.074249506f, 0.0673556328f, 0.f),
    C_Vertex3f(0.0869550705f, 0.0524463654f, 0.f),
    C_Vertex3f(0.0904203653f, 0.314044833f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0460855961f, 0.249484777f, 0.f),
    C_Vertex3f(0.0470232964f, 0.117287636f, 0.f),
    C_Vertex3f(0.0510619879f, 0.261935353f, 0.f),
    C_Vertex3f(0.0542836189f, 0.099776268f, 0.f),
    C_Vertex3f(0.057033658f, 0.273750663f, 0.f),
    C_Vertex3f(0.0633590221f, 0.083132267f, 0.f),
    C_Vertex3f(0.0640006065f, 0.284930706f, 0.f),
    C_Vertex3f(0.074249506f, 0.0673556328f, 0.f),
    C_Vertex3f(0.0719628334f, 0.295475483f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0371280909f, 0.208321452f, 0.f),
    C_Vertex3f(0.0379478931f, 0.154912472f, 0.f),
    C_Vertex3f(0.0391186476f, 0.222677827f, 0.f),
    C_Vertex3f(0.0415780544f, 0.13566637f, 0.f),
    C_Vertex3f(0.0421044827f, 0.236398935f, 0.f),
    C_Vertex3f(0.0470232964f, 0.117287636f, 0.f),
    C_Vertex3f(0.0460855961f, 0.249484777f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0361328125f, 0.193329811f, 0.f),
    C_Vertex3f(0.0361328125f, 0.17502594f, 0.f),
    C_Vertex3f(0.0379478931f, 0.154912472f, 0.f),
    C_Vertex3f(0.0371280909f, 0.208321452f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376997709f, 0.0259751081f, 0.f),
    C_Vertex3f(0.376953125f, 0.189388275f, 0.f),
    C_Vertex3f(0.376953125f, 0.175284386f, 0.f),
    C_Vertex3f(0.375814795f, 0.161905169f, 0.f),
    C_Vertex3f(0.373538136f, 0.149140954f, 0.f),
    C_Vertex3f(0.370123148f, 0.136991739f, 0.f),
    C_Vertex3f(0.36556983f, 0.125457525f, 0.f),
    C_Vertex3f(0.360183954f, 0.0154844522f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360183954f, 0.0154844522f, 0.f),
    C_Vertex3f(0.36556983f, 0.125457525f, 0.f),
    C_Vertex3f(0.359878182f, 0.114538312f, 0.f),
    C_Vertex3f(0.353048205f, 0.104234099f, 0.f),
    C_Vertex3f(0.345079899f, 0.0945448875f, 0.f),
    C_Vertex3f(0.342223644f, 0.006742239f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.345079899f, 0.0945448875f, 0.f),
    C_Vertex3f(0.335973263f, 0.0854706764f, 0.f),
    C_Vertex3f(0.342223644f, 0.006742239f, 0.f),
    C_Vertex3f(0.326243281f, 0.0775307417f, 0.f),
    C_Vertex3f(0.323116779f, -0.000251531601f, 0.f),
    C_Vertex3f(0.315889955f, 0.0707250834f, 0.f),
    C_Vertex3f(0.304913282f, 0.0650537014f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323116779f, -0.000251531601f, 0.f),
    C_Vertex3f(0.304913282f, 0.0650537014f, 0.f),
    C_Vertex3f(0.302863359f, -0.00549685955f, 0.f),
    C_Vertex3f(0.293313265f, 0.0605165958f, 0.f),
    C_Vertex3f(0.281463385f, -0.00899374485f, 0.f),
    C_Vertex3f(0.281089902f, 0.0571137667f, 0.f),
    C_Vertex3f(0.268243194f, 0.0548452139f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.258916855f, -0.0107421875f, 0.f),
    C_Vertex3f(0.281463385f, -0.00899374485f, 0.f),
    C_Vertex3f(0.268243194f, 0.0548452139f, 0.f),
    C_Vertex3f(0.25477314f, 0.0537109375f, 0.f),
    C_Vertex3f(0.238697052f, 0.0537109375f, 0.f),
    C_Vertex3f(0.23522377f, -0.0107421875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.188019753f, 0.0666656494f, 0.f),
    C_Vertex3f(0.191277266f, -0.00547647476f, 0.f),
    C_Vertex3f(0.204329014f, 0.0601882935f, 0.f),
    C_Vertex3f(0.21267724f, -0.00898694992f, 0.f),
    C_Vertex3f(0.221221447f, 0.0558700562f, 0.f),
    C_Vertex3f(0.23522377f, -0.0107421875f, 0.f),
    C_Vertex3f(0.238697052f, 0.0537109375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.146726608f, 0.097905159f, 0.f),
    C_Vertex3f(0.151916981f, 0.00681018829f, 0.f),
    C_Vertex3f(0.158529282f, 0.0857152939f, 0.f),
    C_Vertex3f(0.171023846f, -0.000210762024f, 0.f),
    C_Vertex3f(0.172293663f, 0.075302124f, 0.f),
    C_Vertex3f(0.191277266f, -0.00547647476f, 0.f),
    C_Vertex3f(0.188019753f, 0.0666656494f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136885643f, 0.111871719f, 0.f),
    C_Vertex3f(0.129006386f, 0.127614975f, 0.f),
    C_Vertex3f(0.133956671f, 0.0155863762f, 0.f),
    C_Vertex3f(0.151916981f, 0.00681018829f, 0.f),
    C_Vertex3f(0.146726608f, 0.097905159f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.398132682f, 0.575513005f, 0.f),
    C_Vertex3f(0.386963367f, 0.588336468f, 0.f),
    C_Vertex3f(0.387636662f, 0.375686169f, 0.f),
    C_Vertex3f(0.374198437f, 0.600498676f, 0.f),
    C_Vertex3f(0.369861126f, 0.362978458f, 0.f),
    C_Vertex3f(0.36041081f, 0.611140609f, 0.f),
    C_Vertex3f(0.350585938f, 0.481923103f, 0.f),
    C_Vertex3f(0.349660397f, 0.492291212f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.341330528f, 0.520819187f, 0.f),
    C_Vertex3f(0.345032692f, 0.511739254f, 0.f),
    C_Vertex3f(0.345600486f, 0.620262265f, 0.f),
    C_Vertex3f(0.347809315f, 0.502229929f, 0.f),
    C_Vertex3f(0.36041081f, 0.611140609f, 0.f),
    C_Vertex3f(0.349660397f, 0.492291212f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.317266464f, 0.552845001f, 0.f),
    C_Vertex3f(0.324670792f, 0.545482635f, 0.f),
    C_Vertex3f(0.329767466f, 0.627863646f, 0.f),
    C_Vertex3f(0.331149578f, 0.537690878f, 0.f),
    C_Vertex3f(0.345600486f, 0.620262265f, 0.f),
    C_Vertex3f(0.336702824f, 0.529469728f, 0.f),
    C_Vertex3f(0.341330528f, 0.520819187f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.312911749f, 0.63394475f, 0.f),
    C_Vertex3f(0.295033336f, 0.638505578f, 0.f),
    C_Vertex3f(0.3010602f, 0.564808846f, 0.f),
    C_Vertex3f(0.309396267f, 0.559287071f, 0.f),
    C_Vertex3f(0.317266464f, 0.552845001f, 0.f),
    C_Vertex3f(0.329767466f, 0.627863646f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.263057232f, 0.577692986f, 0.f),
    C_Vertex3f(0.273256779f, 0.575852394f, 0.f),
    C_Vertex3f(0.276132226f, 0.64154613f, 0.f),
    C_Vertex3f(0.282990456f, 0.573091507f, 0.f),
    C_Vertex3f(0.295033336f, 0.638505578f, 0.f),
    C_Vertex3f(0.292258263f, 0.569410324f, 0.f),
    C_Vertex3f(0.3010602f, 0.564808846f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25620842f, 0.643066406f, 0.f),
    C_Vertex3f(0.235316277f, 0.643066406f, 0.f),
    C_Vertex3f(0.241206169f, 0.578613281f, 0.f),
    C_Vertex3f(0.252391815f, 0.578613281f, 0.f),
    C_Vertex3f(0.263057232f, 0.577692986f, 0.f),
    C_Vertex3f(0.276132226f, 0.64154613f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.201108694f, 0.569615364f, 0.f),
    C_Vertex3f(0.210403681f, 0.573214531f, 0.f),
    C_Vertex3f(0.215514898f, 0.641580224f, 0.f),
    C_Vertex3f(0.220184922f, 0.575913906f, 0.f),
    C_Vertex3f(0.235316277f, 0.643066406f, 0.f),
    C_Vertex3f(0.230452418f, 0.57771349f, 0.f),
    C_Vertex3f(0.241206169f, 0.578613281f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.196749926f, 0.63860786f, 0.f),
    C_Vertex3f(0.179021358f, 0.634149313f, 0.f),
    C_Vertex3f(0.183977485f, 0.559717655f, 0.f),
    C_Vertex3f(0.192299962f, 0.565116405f, 0.f),
    C_Vertex3f(0.201108694f, 0.569615364f, 0.f),
    C_Vertex3f(0.215514898f, 0.641580224f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179021358f, 0.634149313f, 0.f),
    C_Vertex3f(0.162329197f, 0.628204584f, 0.f),
    C_Vertex3f(0.162360072f, 0.538667917f, 0.f),
    C_Vertex3f(0.168791294f, 0.546220779f, 0.f),
    C_Vertex3f(0.176141262f, 0.553419113f, 0.f),
    C_Vertex3f(0.183977485f, 0.559717655f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162329197f, 0.628204584f, 0.f),
    C_Vertex3f(0.146673441f, 0.620773673f, 0.f),
    C_Vertex3f(0.148578882f, 0.51388216f, 0.f),
    C_Vertex3f(0.152253866f, 0.522498608f, 0.f),
    C_Vertex3f(0.156847596f, 0.530760527f, 0.f),
    C_Vertex3f(0.162360072f, 0.538667917f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.146673441f, 0.620773673f, 0.f),
    C_Vertex3f(0.13205409f, 0.61185658f, 0.f),
    C_Vertex3f(0.143066406f, 0.485905647f, 0.f),
    C_Vertex3f(0.143985152f, 0.49558568f, 0.f),
    C_Vertex3f(0.145822644f, 0.504911184f, 0.f),
    C_Vertex3f(0.148578882f, 0.51388216f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.422066927f, 0.533075094f, 0.f),
    C_Vertex3f(0.422861576f, 0.42466116f, 0.f),
    C_Vertex3f(0.426853776f, 0.517606616f, 0.f),
    C_Vertex3f(0.428714275f, 0.444023609f, 0.f),
    C_Vertex3f(0.430045009f, 0.501476884f, 0.f),
    C_Vertex3f(0.431640625f, 0.464832306f, 0.f),
    C_Vertex3f(0.431640625f, 0.484685898f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.398132682f, 0.575513005f, 0.f),
    C_Vertex3f(0.402377129f, 0.390275002f, 0.f),
    C_Vertex3f(0.40770638f, 0.562028289f, 0.f),
    C_Vertex3f(0.414082527f, 0.406744957f, 0.f),
    C_Vertex3f(0.415684462f, 0.547882318f, 0.f),
    C_Vertex3f(0.422861576f, 0.42466116f, 0.f),
    C_Vertex3f(0.422066927f, 0.533075094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.350585938f, 0.481923103f, 0.f),
    C_Vertex3f(0.350585938f, 0.471343994f, 0.f),
    C_Vertex3f(0.369861126f, 0.362978458f, 0.f),
    C_Vertex3f(0.349687576f, 0.461385369f, 0.f),
    C_Vertex3f(0.349050522f, 0.352151871f, 0.f),
    C_Vertex3f(0.347890854f, 0.451828837f, 0.f),
    C_Vertex3f(0.34519577f, 0.442674398f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.341602325f, 0.433922052f, 0.f),
    C_Vertex3f(0.337110519f, 0.425571799f, 0.f),
    C_Vertex3f(0.337127686f, 0.347679138f, 0.f),
    C_Vertex3f(0.349050522f, 0.352151871f, 0.f),
    C_Vertex3f(0.34519577f, 0.442674398f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.398132682f, 0.575513005f, 0.f),
    C_Vertex3f(0.387636662f, 0.375686169f, 0.f),
    C_Vertex3f(0.402377129f, 0.390275002f, 0.f),
    C_Vertex3f(0.143066406f, 0.485905647f, 0.f),
    C_Vertex3f(0.13205409f, 0.61185658f, 0.f),
    C_Vertex3f(0.143066406f, 0.475216866f, 0.f),
    C_Vertex3f(0.123096943f, 0.143631935f, 0.f),
    C_Vertex3f(0.133956671f, 0.0155863762f, 0.f),
    C_Vertex3f(0.129006386f, 0.127614975f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158699036f, 0.347679138f, -1.f),
    C_Vertex3f(0.151340485f, 0.345767021f, -1.f),
    C_Vertex3f(0.157922745f, 0.28133297f, -1.f),
    C_Vertex3f(0.167577982f, 0.289368868f, -1.f),
    C_Vertex3f(0.162074685f, 0.418038607f, -1.f),
    C_Vertex3f(0.156643748f, 0.426273108f, -1.f),
    C_Vertex3f(0.152117968f, 0.435025692f, -1.f),
    C_Vertex3f(0.148497343f, 0.44429636f, -1.f),
    C_Vertex3f(0.146392822f, 0.352151871f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.146392822f, 0.352151871f, -1.f),
    C_Vertex3f(0.148497343f, 0.44429636f, -1.f),
    C_Vertex3f(0.145781875f, 0.454085112f, -1.f),
    C_Vertex3f(0.143971562f, 0.464391947f, -1.f),
    C_Vertex3f(0.143066406f, 0.475216866f, -1.f),
    C_Vertex3f(0.13205409f, 0.61185658f, -1.f),
    C_Vertex3f(0.124967098f, 0.363005638f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118471146f, 0.601453304f, -1.f),
    C_Vertex3f(0.105924606f, 0.589563847f, -1.f),
    C_Vertex3f(0.106728077f, 0.375767708f, -1.f),
    C_Vertex3f(0.124967098f, 0.363005638f, -1.f),
    C_Vertex3f(0.13205409f, 0.61185658f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0776948929f, 0.549723387f, -1.f),
    C_Vertex3f(0.0798101425f, 0.407016754f, -1.f),
    C_Vertex3f(0.0855364799f, 0.563698888f, -1.f),
    C_Vertex3f(0.0916757584f, 0.39043808f, -1.f),
    C_Vertex3f(0.0949463844f, 0.576979041f, -1.f),
    C_Vertex3f(0.106728077f, 0.375767708f, -1.f),
    C_Vertex3f(0.105924606f, 0.589563847f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0635800362f, 0.503624797f, -1.f),
    C_Vertex3f(0.0649781227f, 0.445086479f, -1.f),
    C_Vertex3f(0.066716671f, 0.519686341f, -1.f),
    C_Vertex3f(0.0709109306f, 0.425232887f, -1.f),
    C_Vertex3f(0.0714216232f, 0.535052538f, -1.f),
    C_Vertex3f(0.0798101425f, 0.407016754f, -1.f),
    C_Vertex3f(0.0776948929f, 0.549723387f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0620117188f, 0.486867905f, -1.f),
    C_Vertex3f(0.0620117188f, 0.46657753f, -1.f),
    C_Vertex3f(0.0649781227f, 0.445086479f, -1.f),
    C_Vertex3f(0.0635800362f, 0.503624797f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.452218533f, 0.232228994f, -1.f),
    C_Vertex3f(0.452562571f, 0.134429336f, -1.f),
    C_Vertex3f(0.455113173f, 0.218924403f, -1.f),
    C_Vertex3f(0.456192732f, 0.153332829f, -1.f),
    C_Vertex3f(0.457042933f, 0.205100656f, -1.f),
    C_Vertex3f(0.458007813f, 0.173055649f, -1.f),
    C_Vertex3f(0.458007813f, 0.190757751f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.437745333f, 0.269027829f, -1.f),
    C_Vertex3f(0.439857006f, 0.0990803242f, -1.f),
    C_Vertex3f(0.443534613f, 0.257280707f, -1.f),
    C_Vertex3f(0.447117329f, 0.116345167f, -1.f),
    C_Vertex3f(0.448359013f, 0.245014429f, -1.f),
    C_Vertex3f(0.452562571f, 0.134429336f, -1.f),
    C_Vertex3f(0.452218533f, 0.232228994f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.405185461f, 0.310169578f, -1.f),
    C_Vertex3f(0.407185555f, 0.0522017479f, -1.f),
    C_Vertex3f(0.414670229f, 0.300935864f, -1.f),
    C_Vertex3f(0.419891119f, 0.0670086145f, -1.f),
    C_Vertex3f(0.423272133f, 0.290964603f, -1.f),
    C_Vertex3f(0.430781603f, 0.0826348066f, -1.f),
    C_Vertex3f(0.430991173f, 0.280255795f, -1.f),
    C_Vertex3f(0.439857006f, 0.0990803242f, -1.f),
    C_Vertex3f(0.437745333f, 0.269027829f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.376953125f, 0.189388275f, -1.f),
    C_Vertex3f(0.376997709f, 0.0259751081f, -1.f),
    C_Vertex3f(0.383567333f, 0.32642436f, -1.f),
    C_Vertex3f(0.392664909f, 0.0382142067f, -1.f),
    C_Vertex3f(0.394817829f, 0.318665743f, -1.f),
    C_Vertex3f(0.407185555f, 0.0522017479f, -1.f),
    C_Vertex3f(0.405185461f, 0.310169578f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.365228891f, 0.240091324f, -1.f),
    C_Vertex3f(0.369918585f, 0.228338242f, -1.f),
    C_Vertex3f(0.371433973f, 0.33344543f, -1.f),
    C_Vertex3f(0.373435855f, 0.215970039f, -1.f),
    C_Vertex3f(0.383567333f, 0.32642436f, -1.f),
    C_Vertex3f(0.375780702f, 0.202986717f, -1.f),
    C_Vertex3f(0.376953125f, 0.189388275f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.365228891f, 0.240091324f, -1.f),
    C_Vertex3f(0.371433973f, 0.33344543f, -1.f),
    C_Vertex3f(0.359366775f, 0.251229286f, -1.f),
    C_Vertex3f(0.358417749f, 0.339728951f, -1.f),
    C_Vertex3f(0.352332234f, 0.261752129f, -1.f),
    C_Vertex3f(0.344518661f, 0.345274925f, -1.f),
    C_Vertex3f(0.344125271f, 0.271659851f, -1.f),
    C_Vertex3f(0.337127686f, 0.347679138f, -1.f),
    C_Vertex3f(0.334745884f, 0.280952454f, -1.f),
    C_Vertex3f(0.337110519f, 0.425571799f, -1.f),
    C_Vertex3f(0.331720352f, 0.417623639f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.302310467f, 0.391324639f, -1.f),
    C_Vertex3f(0.302990675f, 0.301860809f, -1.f),
    C_Vertex3f(0.310537815f, 0.39668262f, -1.f),
    C_Vertex3f(0.314178586f, 0.296052933f, -1.f),
    C_Vertex3f(0.318244934f, 0.402933598f, -1.f),
    C_Vertex3f(0.324763656f, 0.289083481f, -1.f),
    C_Vertex3f(0.325431824f, 0.410077572f, -1.f),
    C_Vertex3f(0.334745884f, 0.280952454f, -1.f),
    C_Vertex3f(0.331720352f, 0.417623639f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.26419878f, 0.378822684f, -1.f),
    C_Vertex3f(0.265809894f, 0.312314987f, -1.f),
    C_Vertex3f(0.274507046f, 0.380608678f, -1.f),
    C_Vertex3f(0.278806329f, 0.309991837f, -1.f),
    C_Vertex3f(0.284295082f, 0.383287668f, -1.f),
    C_Vertex3f(0.291199923f, 0.306507111f, -1.f),
    C_Vertex3f(0.293562889f, 0.386859655f, -1.f),
    C_Vertex3f(0.302990675f, 0.301860809f, -1.f),
    C_Vertex3f(0.302310467f, 0.391324639f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.210159063f, 0.383328438f, -1.f),
    C_Vertex3f(0.212228537f, 0.310032606f, -1.f),
    C_Vertex3f(0.220171332f, 0.380629063f, -1.f),
    C_Vertex3f(0.224898577f, 0.312328577f, -1.f),
    C_Vertex3f(0.230737805f, 0.378829479f, -1.f),
    C_Vertex3f(0.238171577f, 0.313476563f, -1.f),
    C_Vertex3f(0.241858482f, 0.377929688f, -1.f),
    C_Vertex3f(0.252210617f, 0.313476563f, -1.f),
    C_Vertex3f(0.253370285f, 0.377929688f, -1.f),
    C_Vertex3f(0.265809894f, 0.312314987f, -1.f),
    C_Vertex3f(0.26419878f, 0.378822684f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.175652027f, 0.403123856f, -1.f),
    C_Vertex3f(0.17783618f, 0.296256781f, -1.f),
    C_Vertex3f(0.18344748f, 0.396825314f, -1.f),
    C_Vertex3f(0.188697338f, 0.301996708f, -1.f),
    C_Vertex3f(0.191797137f, 0.391426563f, -1.f),
    C_Vertex3f(0.200161457f, 0.30658865f, -1.f),
    C_Vertex3f(0.200700998f, 0.386927605f, -1.f),
    C_Vertex3f(0.212228537f, 0.310032606f, -1.f),
    C_Vertex3f(0.210159063f, 0.383328438f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.168410778f, 0.410322189f, -1.f),
    C_Vertex3f(0.162074685f, 0.418038607f, -1.f),
    C_Vertex3f(0.167577982f, 0.289368868f, -1.f),
    C_Vertex3f(0.17783618f, 0.296256781f, -1.f),
    C_Vertex3f(0.175652027f, 0.403123856f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.128502846f, 0.241154194f, -1.f),
    C_Vertex3f(0.134160519f, 0.252059698f, -1.f),
    C_Vertex3f(0.137467742f, 0.341102719f, -1.f),
    C_Vertex3f(0.140949726f, 0.262391329f, -1.f),
    C_Vertex3f(0.151340485f, 0.345767021f, -1.f),
    C_Vertex3f(0.148870468f, 0.272149086f, -1.f),
    C_Vertex3f(0.157922745f, 0.28133297f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.124439359f, 0.33559835f, -1.f),
    C_Vertex3f(0.112255335f, 0.329253912f, -1.f),
    C_Vertex3f(0.1171875f, 0.191793442f, -1.f),
    C_Vertex3f(0.118319035f, 0.20499444f, -1.f),
    C_Vertex3f(0.120582104f, 0.217621565f, -1.f),
    C_Vertex3f(0.123976707f, 0.229674816f, -1.f),
    C_Vertex3f(0.128502846f, 0.241154194f, -1.f),
    C_Vertex3f(0.137467742f, 0.341102719f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.123096943f, 0.143631935f, -1.f),
    C_Vertex3f(0.119157314f, 0.1599226f, -1.f),
    C_Vertex3f(0.133956671f, 0.0155863762f, -1.f),
    C_Vertex3f(0.1171875f, 0.176486969f, -1.f),
    C_Vertex3f(0.117142916f, 0.0261178017f, -1.f),
    C_Vertex3f(0.1171875f, 0.191793442f, -1.f),
    C_Vertex3f(0.112255335f, 0.329253912f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.101475716f, 0.0384044647f, -1.f),
    C_Vertex3f(0.117142916f, 0.0261178017f, -1.f),
    C_Vertex3f(0.112255335f, 0.329253912f, -1.f),
    C_Vertex3f(0.10091567f, 0.322069407f, -1.f),
    C_Vertex3f(0.0904203653f, 0.314044833f, -1.f),
    C_Vertex3f(0.0869550705f, 0.0524463654f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0807694197f, 0.305180192f, -1.f),
    C_Vertex3f(0.0719628334f, 0.295475483f, -1.f),
    C_Vertex3f(0.074249506f, 0.0673556328f, -1.f),
    C_Vertex3f(0.0869550705f, 0.0524463654f, -1.f),
    C_Vertex3f(0.0904203653f, 0.314044833f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0460855961f, 0.249484777f, -1.f),
    C_Vertex3f(0.0470232964f, 0.117287636f, -1.f),
    C_Vertex3f(0.0510619879f, 0.261935353f, -1.f),
    C_Vertex3f(0.0542836189f, 0.099776268f, -1.f),
    C_Vertex3f(0.057033658f, 0.273750663f, -1.f),
    C_Vertex3f(0.0633590221f, 0.083132267f, -1.f),
    C_Vertex3f(0.0640006065f, 0.284930706f, -1.f),
    C_Vertex3f(0.074249506f, 0.0673556328f, -1.f),
    C_Vertex3f(0.0719628334f, 0.295475483f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0371280909f, 0.208321452f, -1.f),
    C_Vertex3f(0.0379478931f, 0.154912472f, -1.f),
    C_Vertex3f(0.0391186476f, 0.222677827f, -1.f),
    C_Vertex3f(0.0415780544f, 0.13566637f, -1.f),
    C_Vertex3f(0.0421044827f, 0.236398935f, -1.f),
    C_Vertex3f(0.0470232964f, 0.117287636f, -1.f),
    C_Vertex3f(0.0460855961f, 0.249484777f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0361328125f, 0.193329811f, -1.f),
    C_Vertex3f(0.0361328125f, 0.17502594f, -1.f),
    C_Vertex3f(0.0379478931f, 0.154912472f, -1.f),
    C_Vertex3f(0.0371280909f, 0.208321452f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376997709f, 0.0259751081f, -1.f),
    C_Vertex3f(0.376953125f, 0.189388275f, -1.f),
    C_Vertex3f(0.376953125f, 0.175284386f, -1.f),
    C_Vertex3f(0.375814795f, 0.161905169f, -1.f),
    C_Vertex3f(0.373538136f, 0.149140954f, -1.f),
    C_Vertex3f(0.370123148f, 0.136991739f, -1.f),
    C_Vertex3f(0.36556983f, 0.125457525f, -1.f),
    C_Vertex3f(0.360183954f, 0.0154844522f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360183954f, 0.0154844522f, -1.f),
    C_Vertex3f(0.36556983f, 0.125457525f, -1.f),
    C_Vertex3f(0.359878182f, 0.114538312f, -1.f),
    C_Vertex3f(0.353048205f, 0.104234099f, -1.f),
    C_Vertex3f(0.345079899f, 0.0945448875f, -1.f),
    C_Vertex3f(0.342223644f, 0.006742239f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.345079899f, 0.0945448875f, -1.f),
    C_Vertex3f(0.335973263f, 0.0854706764f, -1.f),
    C_Vertex3f(0.342223644f, 0.006742239f, -1.f),
    C_Vertex3f(0.326243281f, 0.0775307417f, -1.f),
    C_Vertex3f(0.323116779f, -0.000251531601f, -1.f),
    C_Vertex3f(0.315889955f, 0.0707250834f, -1.f),
    C_Vertex3f(0.304913282f, 0.0650537014f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323116779f, -0.000251531601f, -1.f),
    C_Vertex3f(0.304913282f, 0.0650537014f, -1.f),
    C_Vertex3f(0.302863359f, -0.00549685955f, -1.f),
    C_Vertex3f(0.293313265f, 0.0605165958f, -1.f),
    C_Vertex3f(0.281463385f, -0.00899374485f, -1.f),
    C_Vertex3f(0.281089902f, 0.0571137667f, -1.f),
    C_Vertex3f(0.268243194f, 0.0548452139f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.258916855f, -0.0107421875f, -1.f),
    C_Vertex3f(0.281463385f, -0.00899374485f, -1.f),
    C_Vertex3f(0.268243194f, 0.0548452139f, -1.f),
    C_Vertex3f(0.25477314f, 0.0537109375f, -1.f),
    C_Vertex3f(0.238697052f, 0.0537109375f, -1.f),
    C_Vertex3f(0.23522377f, -0.0107421875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.188019753f, 0.0666656494f, -1.f),
    C_Vertex3f(0.191277266f, -0.00547647476f, -1.f),
    C_Vertex3f(0.204329014f, 0.0601882935f, -1.f),
    C_Vertex3f(0.21267724f, -0.00898694992f, -1.f),
    C_Vertex3f(0.221221447f, 0.0558700562f, -1.f),
    C_Vertex3f(0.23522377f, -0.0107421875f, -1.f),
    C_Vertex3f(0.238697052f, 0.0537109375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.146726608f, 0.097905159f, -1.f),
    C_Vertex3f(0.151916981f, 0.00681018829f, -1.f),
    C_Vertex3f(0.158529282f, 0.0857152939f, -1.f),
    C_Vertex3f(0.171023846f, -0.000210762024f, -1.f),
    C_Vertex3f(0.172293663f, 0.075302124f, -1.f),
    C_Vertex3f(0.191277266f, -0.00547647476f, -1.f),
    C_Vertex3f(0.188019753f, 0.0666656494f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136885643f, 0.111871719f, -1.f),
    C_Vertex3f(0.129006386f, 0.127614975f, -1.f),
    C_Vertex3f(0.133956671f, 0.0155863762f, -1.f),
    C_Vertex3f(0.151916981f, 0.00681018829f, -1.f),
    C_Vertex3f(0.146726608f, 0.097905159f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.398132682f, 0.575513005f, -1.f),
    C_Vertex3f(0.386963367f, 0.588336468f, -1.f),
    C_Vertex3f(0.387636662f, 0.375686169f, -1.f),
    C_Vertex3f(0.374198437f, 0.600498676f, -1.f),
    C_Vertex3f(0.369861126f, 0.362978458f, -1.f),
    C_Vertex3f(0.36041081f, 0.611140609f, -1.f),
    C_Vertex3f(0.350585938f, 0.481923103f, -1.f),
    C_Vertex3f(0.349660397f, 0.492291212f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.341330528f, 0.520819187f, -1.f),
    C_Vertex3f(0.345032692f, 0.511739254f, -1.f),
    C_Vertex3f(0.345600486f, 0.620262265f, -1.f),
    C_Vertex3f(0.347809315f, 0.502229929f, -1.f),
    C_Vertex3f(0.36041081f, 0.611140609f, -1.f),
    C_Vertex3f(0.349660397f, 0.492291212f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.317266464f, 0.552845001f, -1.f),
    C_Vertex3f(0.324670792f, 0.545482635f, -1.f),
    C_Vertex3f(0.329767466f, 0.627863646f, -1.f),
    C_Vertex3f(0.331149578f, 0.537690878f, -1.f),
    C_Vertex3f(0.345600486f, 0.620262265f, -1.f),
    C_Vertex3f(0.336702824f, 0.529469728f, -1.f),
    C_Vertex3f(0.341330528f, 0.520819187f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.312911749f, 0.63394475f, -1.f),
    C_Vertex3f(0.295033336f, 0.638505578f, -1.f),
    C_Vertex3f(0.3010602f, 0.564808846f, -1.f),
    C_Vertex3f(0.309396267f, 0.559287071f, -1.f),
    C_Vertex3f(0.317266464f, 0.552845001f, -1.f),
    C_Vertex3f(0.329767466f, 0.627863646f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.263057232f, 0.577692986f, -1.f),
    C_Vertex3f(0.273256779f, 0.575852394f, -1.f),
    C_Vertex3f(0.276132226f, 0.64154613f, -1.f),
    C_Vertex3f(0.282990456f, 0.573091507f, -1.f),
    C_Vertex3f(0.295033336f, 0.638505578f, -1.f),
    C_Vertex3f(0.292258263f, 0.569410324f, -1.f),
    C_Vertex3f(0.3010602f, 0.564808846f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25620842f, 0.643066406f, -1.f),
    C_Vertex3f(0.235316277f, 0.643066406f, -1.f),
    C_Vertex3f(0.241206169f, 0.578613281f, -1.f),
    C_Vertex3f(0.252391815f, 0.578613281f, -1.f),
    C_Vertex3f(0.263057232f, 0.577692986f, -1.f),
    C_Vertex3f(0.276132226f, 0.64154613f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.201108694f, 0.569615364f, -1.f),
    C_Vertex3f(0.210403681f, 0.573214531f, -1.f),
    C_Vertex3f(0.215514898f, 0.641580224f, -1.f),
    C_Vertex3f(0.220184922f, 0.575913906f, -1.f),
    C_Vertex3f(0.235316277f, 0.643066406f, -1.f),
    C_Vertex3f(0.230452418f, 0.57771349f, -1.f),
    C_Vertex3f(0.241206169f, 0.578613281f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.196749926f, 0.63860786f, -1.f),
    C_Vertex3f(0.179021358f, 0.634149313f, -1.f),
    C_Vertex3f(0.183977485f, 0.559717655f, -1.f),
    C_Vertex3f(0.192299962f, 0.565116405f, -1.f),
    C_Vertex3f(0.201108694f, 0.569615364f, -1.f),
    C_Vertex3f(0.215514898f, 0.641580224f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179021358f, 0.634149313f, -1.f),
    C_Vertex3f(0.162329197f, 0.628204584f, -1.f),
    C_Vertex3f(0.162360072f, 0.538667917f, -1.f),
    C_Vertex3f(0.168791294f, 0.546220779f, -1.f),
    C_Vertex3f(0.176141262f, 0.553419113f, -1.f),
    C_Vertex3f(0.183977485f, 0.559717655f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162329197f, 0.628204584f, -1.f),
    C_Vertex3f(0.146673441f, 0.620773673f, -1.f),
    C_Vertex3f(0.148578882f, 0.51388216f, -1.f),
    C_Vertex3f(0.152253866f, 0.522498608f, -1.f),
    C_Vertex3f(0.156847596f, 0.530760527f, -1.f),
    C_Vertex3f(0.162360072f, 0.538667917f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.146673441f, 0.620773673f, -1.f),
    C_Vertex3f(0.13205409f, 0.61185658f, -1.f),
    C_Vertex3f(0.143066406f, 0.485905647f, -1.f),
    C_Vertex3f(0.143985152f, 0.49558568f, -1.f),
    C_Vertex3f(0.145822644f, 0.504911184f, -1.f),
    C_Vertex3f(0.148578882f, 0.51388216f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.422066927f, 0.533075094f, -1.f),
    C_Vertex3f(0.422861576f, 0.42466116f, -1.f),
    C_Vertex3f(0.426853776f, 0.517606616f, -1.f),
    C_Vertex3f(0.428714275f, 0.444023609f, -1.f),
    C_Vertex3f(0.430045009f, 0.501476884f, -1.f),
    C_Vertex3f(0.431640625f, 0.464832306f, -1.f),
    C_Vertex3f(0.431640625f, 0.484685898f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.398132682f, 0.575513005f, -1.f),
    C_Vertex3f(0.402377129f, 0.390275002f, -1.f),
    C_Vertex3f(0.40770638f, 0.562028289f, -1.f),
    C_Vertex3f(0.414082527f, 0.406744957f, -1.f),
    C_Vertex3f(0.415684462f, 0.547882318f, -1.f),
    C_Vertex3f(0.422861576f, 0.42466116f, -1.f),
    C_Vertex3f(0.422066927f, 0.533075094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.350585938f, 0.481923103f, -1.f),
    C_Vertex3f(0.350585938f, 0.471343994f, -1.f),
    C_Vertex3f(0.369861126f, 0.362978458f, -1.f),
    C_Vertex3f(0.349687576f, 0.461385369f, -1.f),
    C_Vertex3f(0.349050522f, 0.352151871f, -1.f),
    C_Vertex3f(0.347890854f, 0.451828837f, -1.f),
    C_Vertex3f(0.34519577f, 0.442674398f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.341602325f, 0.433922052f, -1.f),
    C_Vertex3f(0.337110519f, 0.425571799f, -1.f),
    C_Vertex3f(0.337127686f, 0.347679138f, -1.f),
    C_Vertex3f(0.349050522f, 0.352151871f, -1.f),
    C_Vertex3f(0.34519577f, 0.442674398f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.398132682f, 0.575513005f, -1.f),
    C_Vertex3f(0.387636662f, 0.375686169f, -1.f),
    C_Vertex3f(0.402377129f, 0.390275002f, -1.f),
    C_Vertex3f(0.143066406f, 0.485905647f, -1.f),
    C_Vertex3f(0.13205409f, 0.61185658f, -1.f),
    C_Vertex3f(0.143066406f, 0.475216866f, -1.f),
    C_Vertex3f(0.123096943f, 0.143631935f, -1.f),
    C_Vertex3f(0.133956671f, 0.0155863762f, -1.f),
    C_Vertex3f(0.129006386f, 0.127614975f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_57[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.0899563953f, 0.995945752f, 0.f),
    C_Vertex3f(0.048828125f, 0.1484375f, 0.f),
    C_Vertex3f(0.048828125f, 0.1484375f, -1.f),
    C_Vertex3f(0.124511719f, 0.155273438f, -1.f),
    C_Vertex3f(0.124511719f, 0.155273438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.984207809f, 0.17701681f, 0.f),
    C_Vertex3f(0.124511719f, 0.155273438f, 0.f),
    C_Vertex3f(0.124511719f, 0.155273438f, -1.f),
    C_Normal3f(0.97725147f, 0.212083668f, 0.f),
    C_Vertex3f(0.125712395f, 0.148597717f, 0.f),
    C_Vertex3f(0.125712395f, 0.148597717f, -1.f),
    C_Normal3f(0.965718329f, 0.259591848f, 0.f),
    C_Vertex3f(0.128659129f, 0.136162758f, 0.f),
    C_Vertex3f(0.128659129f, 0.136162758f, -1.f),
    C_Normal3f(0.946659803f, 0.32223472f, 0.f),
    C_Vertex3f(0.132151246f, 0.124644279f, 0.f),
    C_Vertex3f(0.132151246f, 0.124644279f, -1.f),
    C_Normal3f(0.920360506f, 0.391071022f, 0.f),
    C_Vertex3f(0.136188745f, 0.114042282f, 0.f),
    C_Vertex3f(0.136188745f, 0.114042282f, -1.f),
    C_Normal3f(0.88495326f, 0.465679735f, 0.f),
    C_Vertex3f(0.140771627f, 0.104356766f, 0.f),
    C_Vertex3f(0.140771627f, 0.104356766f, -1.f),
    C_Normal3f(0.838492572f, 0.544913113f, 0.f),
    C_Vertex3f(0.145899892f, 0.0955877304f, 0.f),
    C_Vertex3f(0.145899892f, 0.0955877304f, -1.f),
    C_Normal3f(0.779278457f, 0.62667793f, 0.f),
    C_Vertex3f(0.151573539f, 0.0877351761f, 0.f),
    C_Vertex3f(0.151573539f, 0.0877351761f, -1.f),
    C_Normal3f(0.698300898f, 0.715804279f, 0.f),
    C_Vertex3f(0.157792568f, 0.0807991028f, 0.f),
    C_Vertex3f(0.157792568f, 0.0807991028f, -1.f),
    C_Normal3f(0.576654494f, 0.816988111f, 0.f),
    C_Vertex3f(0.167939186f, 0.0717697144f, 0.f),
    C_Vertex3f(0.167939186f, 0.0717697144f, -1.f),
    C_Normal3f(0.417151183f, 0.90883708f, 0.f),
    C_Vertex3f(0.183377743f, 0.0627403259f, 0.f),
    C_Vertex3f(0.183377743f, 0.0627403259f, -1.f),
    C_Normal3f(0.239484251f, 0.970900238f, 0.f),
    C_Vertex3f(0.200726032f, 0.0567207336f, 0.f),
    C_Vertex3f(0.200726032f, 0.0567207336f, -1.f),
    C_Normal3f(0.0771460533f, 0.997019827f, 0.f),
    C_Vertex3f(0.219984055f, 0.0537109375f, 0.f),
    C_Vertex3f(0.219984055f, 0.0537109375f, -1.f),
    C_Normal3f(-0.0565973558f, 0.998397052f, 0.f),
    C_Vertex3f(0.239624023f, 0.0537109375f, 0.f),
    C_Vertex3f(0.239624023f, 0.0537109375f, -1.f),
    C_Normal3f(-0.186049134f, 0.982540429f, 0.f),
    C_Vertex3f(0.256633282f, 0.055788517f, 0.f),
    C_Vertex3f(0.256633282f, 0.055788517f, -1.f),
    C_Normal3f(-0.320425451f, 0.94727385f, 0.f),
    C_Vertex3f(0.272539616f, 0.059943676f, 0.f),
    C_Vertex3f(0.272539616f, 0.059943676f, -1.f),
    C_Normal3f(-0.454476625f, 0.890758693f, 0.f),
    C_Vertex3f(0.287343025f, 0.0661764145f, 0.f),
    C_Vertex3f(0.287343025f, 0.0661764145f, -1.f),
    C_Normal3f(-0.567413867f, 0.823432744f, 0.f),
    C_Vertex3f(0.30104351f, 0.0744867325f, 0.f),
    C_Vertex3f(0.30104351f, 0.0744867325f, -1.f),
    C_Normal3f(-0.660829723f, 0.750535965f, 0.f),
    C_Vertex3f(0.313485622f, 0.0841908455f, 0.f),
    C_Vertex3f(0.313485622f, 0.0841908455f, -1.f),
    C_Normal3f(-0.745200336f, 0.666840672f, 0.f),
    C_Vertex3f(0.324669361f, 0.0952887535f, 0.f),
    C_Vertex3f(0.324669361f, 0.0952887535f, -1.f),
    C_Normal3f(-0.817359567f, 0.576127887f, 0.f),
    C_Vertex3f(0.334594727f, 0.107780457f, 0.f),
    C_Vertex3f(0.334594727f, 0.107780457f, -1.f),
    C_Normal3f(-0.875859439f, 0.482566178f, 0.f),
    C_Vertex3f(0.343261719f, 0.121665955f, 0.f),
    C_Vertex3f(0.343261719f, 0.121665955f, -1.f),
    C_Normal3f(-0.916996717f, 0.398894817f, 0.f),
    C_Vertex3f(0.35123539f, 0.137983322f, 0.f),
    C_Vertex3f(0.35123539f, 0.137983322f, -1.f),
    C_Normal3f(-0.944654047f, 0.328068167f, 0.f),
    C_Vertex3f(0.358515739f, 0.156732559f, 0.f),
    C_Vertex3f(0.358515739f, 0.156732559f, -1.f),
    C_Normal3f(-0.963307202f, 0.268401146f, 0.f),
    C_Vertex3f(0.365102768f, 0.177913666f, 0.f),
    C_Vertex3f(0.365102768f, 0.177913666f, -1.f),
    C_Normal3f(-0.977173686f, 0.212441862f, 0.f),
    C_Vertex3f(0.370996475f, 0.201526642f, 0.f),
    C_Vertex3f(0.370996475f, 0.201526642f, -1.f),
    C_Normal3f(-0.988365114f, 0.152100086f, 0.f),
    C_Vertex3f(0.375416756f, 0.225355148f, 0.f),
    C_Vertex3f(0.375416756f, 0.225355148f, -1.f),
    C_Normal3f(-0.995839059f, 0.0911295936f, 0.f),
    C_Vertex3f(0.378363609f, 0.249399185f, 0.f),
    C_Vertex3f(0.378363609f, 0.249399185f, -1.f),
    C_Normal3f(-0.99929142f, 0.0376396999f, 0.f),
    C_Vertex3f(0.379837036f, 0.273658752f, 0.f),
    C_Vertex3f(0.379837036f, 0.273658752f, -1.f),
    C_Normal3f(-0.999952674f, -0.0097359838f, 0.f),
    C_Vertex3f(0.379837036f, 0.288516998f, 0.f),
    C_Vertex3f(0.379837036f, 0.288516998f, -1.f),
    C_Normal3f(-0.999430895f, -0.0337311551f, 0.f),
    C_Vertex3f(0.379615784f, 0.296382904f, 0.f),
    C_Vertex3f(0.379615784f, 0.296382904f, -1.f),
    C_Normal3f(-0.999111474f, -0.0421443395f, 0.f),
    C_Vertex3f(0.379394531f, 0.301628113f, 0.f),
    C_Vertex3f(0.379394531f, 0.301628113f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.847127616f, -0.531389415f, 0.f),
    C_Vertex3f(0.379394531f, 0.301628113f, 0.f),
    C_Vertex3f(0.379394531f, 0.301628113f, -1.f),
    C_Normal3f(0.799589336f, -0.600547314f, 0.f),
    C_Vertex3f(0.373500824f, 0.292232513f, 0.f),
    C_Vertex3f(0.373500824f, 0.292232513f, -1.f),
    C_Normal3f(0.731778979f, -0.681541979f, 0.f),
    C_Vertex3f(0.359557629f, 0.275216579f, 0.f),
    C_Vertex3f(0.359557629f, 0.275216579f, -1.f),
    C_Normal3f(0.641211569f, -0.767364144f, 0.f),
    C_Vertex3f(0.343458652f, 0.25997591f, 0.f),
    C_Vertex3f(0.343458652f, 0.25997591f, -1.f),
    C_Normal3f(0.545340359f, -0.838214636f, 0.f),
    C_Vertex3f(0.325203896f, 0.246510506f, 0.f),
    C_Vertex3f(0.325203896f, 0.246510506f, -1.f),
    C_Normal3f(0.440721959f, -0.897643685f, 0.f),
    C_Vertex3f(0.304793358f, 0.234820366f, 0.f),
    C_Vertex3f(0.304793358f, 0.234820366f, -1.f),
    C_Normal3f(0.319358647f, -0.947633922f, 0.f),
    C_Vertex3f(0.283536434f, 0.226052761f, 0.f),
    C_Vertex3f(0.283536434f, 0.226052761f, -1.f),
    C_Normal3f(0.191022873f, -0.981585562f, 0.f),
    C_Vertex3f(0.261433125f, 0.220207691f, 0.f),
    C_Vertex3f(0.261433125f, 0.220207691f, -1.f),
    C_Normal3f(0.0651267543f, -0.997877002f, 0.f),
    C_Vertex3f(0.238483429f, 0.217285156f, 0.f),
    C_Vertex3f(0.238483429f, 0.217285156f, -1.f),
    C_Normal3f(-0.0439695865f, -0.999032915f, 0.f),
    C_Vertex3f(0.216653824f, 0.217285156f, 0.f),
    C_Vertex3f(0.216653824f, 0.217285156f, -1.f),
    C_Normal3f(-0.142879069f, -0.989740074f, 0.f),
    C_Vertex3f(0.197554469f, 0.219086528f, 0.f),
    C_Vertex3f(0.197554469f, 0.219086528f, -1.f),
    C_Normal3f(-0.243298545f, -0.96995151f, 0.f),
    C_Vertex3f(0.179218888f, 0.222689271f, 0.f),
    C_Vertex3f(0.179218888f, 0.222689271f, -1.f),
    C_Normal3f(-0.344342738f, -0.938844025f, 0.f),
    C_Vertex3f(0.161647081f, 0.228093386f, 0.f),
    C_Vertex3f(0.161647081f, 0.228093386f, -1.f),
    C_Normal3f(-0.442538708f, -0.896749437f, 0.f),
    C_Vertex3f(0.144839048f, 0.235298872f, 0.f),
    C_Vertex3f(0.144839048f, 0.235298872f, -1.f),
    C_Normal3f(-0.534591436f, -0.845110595f, 0.f),
    C_Vertex3f(0.128794789f, 0.24430573f, 0.f),
    C_Vertex3f(0.128794789f, 0.24430573f, -1.f),
    C_Normal3f(-0.617915332f, -0.78624469f, 0.f),
    C_Vertex3f(0.113514304f, 0.255113959f, 0.f),
    C_Vertex3f(0.113514304f, 0.255113959f, -1.f),
    C_Normal3f(-0.690956056f, -0.722896755f, 0.f),
    C_Vertex3f(0.0989975929f, 0.26772356f, 0.f),
    C_Vertex3f(0.0989975929f, 0.26772356f, -1.f),
    C_Normal3f(-0.758010685f, -0.652242064f, 0.f),
    C_Vertex3f(0.0852446556f, 0.282134533f, 0.f),
    C_Vertex3f(0.0852446556f, 0.282134533f, -1.f),
    C_Normal3f(-0.822027445f, -0.569447815f, 0.f),
    C_Vertex3f(0.0732108355f, 0.297691941f, 0.f),
    C_Vertex3f(0.0732108355f, 0.297691941f, -1.f),
    C_Normal3f(-0.87722677f, -0.480076164f, 0.f),
    C_Vertex3f(0.0628961325f, 0.314395785f, 0.f),
    C_Vertex3f(0.0628961325f, 0.314395785f, -1.f),
    C_Normal3f(-0.922015071f, -0.387154162f, 0.f),
    C_Vertex3f(0.0543005466f, 0.332246065f, 0.f),
    C_Vertex3f(0.0543005466f, 0.332246065f, -1.f),
    C_Normal3f(-0.955842197f, -0.293880224f, 0.f),
    C_Vertex3f(0.047424078f, 0.351242781f, 0.f),
    C_Vertex3f(0.047424078f, 0.351242781f, -1.f),
    C_Normal3f(-0.979151011f, -0.203133494f, 0.f),
    C_Vertex3f(0.0422667265f, 0.371385932f, 0.f),
    C_Vertex3f(0.0422667265f, 0.371385932f, -1.f),
    C_Normal3f(-0.993115783f, -0.117136098f, 0.f),
    C_Vertex3f(0.0388284922f, 0.392675519f, 0.f),
    C_Vertex3f(0.0388284922f, 0.392675519f, -1.f),
    C_Normal3f(-0.999314368f, -0.0370245762f, 0.f),
    C_Vertex3f(0.037109375f, 0.415111542f, 0.f),
    C_Vertex3f(0.037109375f, 0.415111542f, -1.f),
    C_Normal3f(-0.999275863f, 0.0380483679f, 0.f),
    C_Vertex3f(0.037109375f, 0.43907547f, 0.f),
    C_Vertex3f(0.037109375f, 0.43907547f, -1.f),
    C_Normal3f(-0.992962062f, 0.118432775f, 0.f),
    C_Vertex3f(0.038903594f, 0.462233663f, 0.f),
    C_Vertex3f(0.038903594f, 0.462233663f, -1.f),
    C_Normal3f(-0.978687525f, 0.20535484f, 0.f),
    C_Vertex3f(0.0424920321f, 0.48420465f, 0.f),
    C_Vertex3f(0.0424920321f, 0.48420465f, -1.f),
    C_Normal3f(-0.954878569f, 0.296996593f, 0.f),
    C_Vertex3f(0.0478746891f, 0.504988432f, 0.f),
    C_Vertex3f(0.0478746891f, 0.504988432f, -1.f),
    C_Normal3f(-0.920370638f, 0.39104718f, 0.f),
    C_Vertex3f(0.0550515652f, 0.524585009f, 0.f),
    C_Vertex3f(0.0550515652f, 0.524585009f, -1.f),
    C_Normal3f(-0.874770403f, 0.48453778f, 0.f),
    C_Vertex3f(0.0640226603f, 0.54299438f, 0.f),
    C_Vertex3f(0.0640226603f, 0.54299438f, -1.f),
    C_Normal3f(-0.818712711f, 0.574203193f, 0.f),
    C_Vertex3f(0.0747879744f, 0.560216546f, 0.f),
    C_Vertex3f(0.0747879744f, 0.560216546f, -1.f),
    C_Normal3f(-0.753894985f, 0.656994998f, 0.f),
    C_Vertex3f(0.0873475075f, 0.576251507f, 0.f),
    C_Vertex3f(0.0873475075f, 0.576251507f, -1.f),
    C_Normal3f(-0.684931219f, 0.728607714f, 0.f),
    C_Vertex3f(0.10170126f, 0.591099262f, 0.f),
    C_Vertex3f(0.10170126f, 0.591099262f, -1.f),
    C_Normal3f(-0.608935654f, 0.793219626f, 0.f),
    C_Vertex3f(0.116962314f, 0.604091048f, 0.f),
    C_Vertex3f(0.116962314f, 0.604091048f, -1.f),
    C_Normal3f(-0.523312807f, 0.852140665f, 0.f),
    C_Vertex3f(0.13313067f, 0.615226865f, 0.f),
    C_Vertex3f(0.13313067f, 0.615226865f, -1.f),
    C_Normal3f(-0.430125535f, 0.902769089f, 0.f),
    C_Vertex3f(0.150206327f, 0.624506712f, 0.f),
    C_Vertex3f(0.150206327f, 0.624506712f, -1.f),
    C_Normal3f(-0.332313001f, 0.943169117f, 0.f),
    C_Vertex3f(0.168189287f, 0.63193059f, 0.f),
    C_Vertex3f(0.168189287f, 0.63193059f, -1.f),
    C_Normal3f(-0.233248621f, 0.972417116f, 0.f),
    C_Vertex3f(0.187079549f, 0.637498498f, 0.f),
    C_Vertex3f(0.187079549f, 0.637498498f, -1.f),
    C_Normal3f(-0.136190116f, 0.990682721f, 0.f),
    C_Vertex3f(0.206877112f, 0.641210437f, 0.f),
    C_Vertex3f(0.206877112f, 0.641210437f, -1.f),
    C_Normal3f(-0.0471541733f, 0.998887658f, 0.f),
    C_Vertex3f(0.227581978f, 0.643066406f, 0.f),
    C_Vertex3f(0.227581978f, 0.643066406f, -1.f),
    C_Normal3f(0.0309990626f, 0.999519348f, 0.f),
    C_Vertex3f(0.246192932f, 0.643066406f, 0.f),
    C_Vertex3f(0.246192932f, 0.643066406f, -1.f),
    C_Normal3f(0.103779063f, 0.994600356f, 0.f),
    C_Vertex3f(0.261468291f, 0.642015457f, 0.f),
    C_Vertex3f(0.261468291f, 0.642015457f, -1.f),
    C_Normal3f(0.175092891f, 0.984551907f, 0.f),
    C_Vertex3f(0.276409268f, 0.639913559f, 0.f),
    C_Vertex3f(0.276409268f, 0.639913559f, -1.f),
    C_Normal3f(0.246858105f, 0.969051659f, 0.f),
    C_Vertex3f(0.291015863f, 0.636760712f, 0.f),
    C_Vertex3f(0.291015863f, 0.636760712f, -1.f),
    C_Normal3f(0.317896813f, 0.948125303f, 0.f),
    C_Vertex3f(0.305288076f, 0.632556915f, 0.f),
    C_Vertex3f(0.305288076f, 0.632556915f, -1.f),
    C_Normal3f(0.387036175f, 0.922064483f, 0.f),
    C_Vertex3f(0.319225907f, 0.62730217f, 0.f),
    C_Vertex3f(0.319225907f, 0.62730217f, -1.f),
    C_Normal3f(0.453203142f, 0.891407311f, 0.f),
    C_Vertex3f(0.332829356f, 0.620996475f, 0.f),
    C_Vertex3f(0.332829356f, 0.620996475f, -1.f),
    C_Normal3f(0.515504897f, 0.856886566f, 0.f),
    C_Vertex3f(0.346098423f, 0.613639832f, 0.f),
    C_Vertex3f(0.346098423f, 0.613639832f, -1.f),
    C_Normal3f(0.576626301f, 0.817007959f, 0.f),
    C_Vertex3f(0.359033108f, 0.605232239f, 0.f),
    C_Vertex3f(0.359033108f, 0.605232239f, -1.f),
    C_Normal3f(0.637932718f, 0.770092189f, 0.f),
    C_Vertex3f(0.371190071f, 0.595930696f, 0.f),
    C_Vertex3f(0.371190071f, 0.595930696f, -1.f),
    C_Normal3f(0.695644736f, 0.718385994f, 0.f),
    C_Vertex3f(0.382569313f, 0.585735202f, 0.f),
    C_Vertex3f(0.382569313f, 0.585735202f, -1.f),
    C_Normal3f(0.748759329f, 0.662842095f, 0.f),
    C_Vertex3f(0.393170834f, 0.574645758f, 0.f),
    C_Vertex3f(0.393170834f, 0.574645758f, -1.f),
    C_Normal3f(0.796536863f, 0.604589939f, 0.f),
    C_Vertex3f(0.402994633f, 0.562662363f, 0.f),
    C_Vertex3f(0.402994633f, 0.562662363f, -1.f),
    C_Normal3f(0.838546813f, 0.544829428f, 0.f),
    C_Vertex3f(0.41204071f, 0.549785018f, 0.f),
    C_Vertex3f(0.41204071f, 0.549785018f, -1.f),
    C_Normal3f(0.874667168f, 0.484723955f, 0.f),
    C_Vertex3f(0.420309067f, 0.536013722f, 0.f),
    C_Vertex3f(0.420309067f, 0.536013722f, -1.f),
    C_Normal3f(0.905046046f, 0.425313592f, 0.f),
    C_Vertex3f(0.427799702f, 0.521348476f, 0.f),
    C_Vertex3f(0.427799702f, 0.521348476f, -1.f),
    C_Normal3f(0.93383199f, 0.357711941f, 0.f),
    C_Vertex3f(0.434512615f, 0.50578928f, 0.f),
    C_Vertex3f(0.434512615f, 0.50578928f, -1.f),
    C_Normal3f(0.957776368f, 0.28751421f, 0.f),
    C_Vertex3f(0.440386415f, 0.488489985f, 0.f),
    C_Vertex3f(0.440386415f, 0.488489985f, -1.f),
    C_Normal3f(0.974169493f, 0.225818336f, 0.f),
    C_Vertex3f(0.4454211f, 0.469450593f, 0.f),
    C_Vertex3f(0.4454211f, 0.469450593f, -1.f),
    C_Normal3f(0.985128045f, 0.171821535f, 0.f),
    C_Vertex3f(0.449616671f, 0.448671103f, 0.f),
    C_Vertex3f(0.449616671f, 0.448671103f, -1.f),
    C_Normal3f(0.992208898f, 0.124585852f, 0.f),
    C_Vertex3f(0.452973127f, 0.426151514f, 0.f),
    C_Vertex3f(0.452973127f, 0.426151514f, -1.f),
    C_Normal3f(0.996533811f, 0.0831888616f, 0.f),
    C_Vertex3f(0.45549047f, 0.401891828f, 0.f),
    C_Vertex3f(0.45549047f, 0.401891828f, -1.f),
    C_Normal3f(0.998904705f, 0.0467919782f, 0.f),
    C_Vertex3f(0.457168698f, 0.375892043f, 0.f),
    C_Vertex3f(0.457168698f, 0.375892043f, -1.f),
    C_Normal3f(0.999894679f, 0.0145108262f, 0.f),
    C_Vertex3f(0.458007813f, 0.348152161f, 0.f),
    C_Vertex3f(0.458007813f, 0.348152161f, -1.f),
    C_Normal3f(0.999901175f, -0.0140601723f, 0.f),
    C_Vertex3f(0.458007813f, 0.318071365f, 0.f),
    C_Vertex3f(0.458007813f, 0.318071365f, -1.f),
    C_Normal3f(0.99903059f, -0.0440212935f, 0.f),
    C_Vertex3f(0.457175374f, 0.288952589f, 0.f),
    C_Vertex3f(0.457175374f, 0.288952589f, -1.f),
    C_Normal3f(0.996992886f, -0.0774929896f, 0.f),
    C_Vertex3f(0.455510497f, 0.261396646f, 0.f),
    C_Vertex3f(0.455510497f, 0.261396646f, -1.f),
    C_Normal3f(0.993388712f, -0.114798814f, 0.f),
    C_Vertex3f(0.453013182f, 0.235403538f, 0.f),
    C_Vertex3f(0.453013182f, 0.235403538f, -1.f),
    C_Normal3f(0.987685978f, -0.156449214f, 0.f),
    C_Vertex3f(0.449683428f, 0.210973263f, 0.f),
    C_Vertex3f(0.449683428f, 0.210973263f, -1.f),
    C_Normal3f(0.979182124f, -0.202983513f, 0.f),
    C_Vertex3f(0.445521235f, 0.188105822f, 0.f),
    C_Vertex3f(0.445521235f, 0.188105822f, -1.f),
    C_Normal3f(0.966958404f, -0.254934371f, 0.f),
    C_Vertex3f(0.440526605f, 0.166801214f, 0.f),
    C_Vertex3f(0.440526605f, 0.166801214f, -1.f),
    C_Normal3f(0.949831605f, -0.312761664f, 0.f),
    C_Vertex3f(0.434699535f, 0.147059441f, 0.f),
    C_Vertex3f(0.434699535f, 0.147059441f, -1.f),
    C_Normal3f(0.928279042f, -0.371884465f, 0.f),
    C_Vertex3f(0.428040028f, 0.128880501f, 0.f),
    C_Vertex3f(0.428040028f, 0.128880501f, -1.f),
    C_Normal3f(0.902951181f, -0.429743081f, 0.f),
    C_Vertex3f(0.420568585f, 0.111786485f, 0.f),
    C_Vertex3f(0.420568585f, 0.111786485f, -1.f),
    C_Normal3f(0.871831596f, -0.489805758f, 0.f),
    C_Vertex3f(0.412285209f, 0.0957773924f, 0.f),
    C_Vertex3f(0.412285209f, 0.0957773924f, -1.f),
    C_Normal3f(0.834359825f, -0.551220059f, 0.f),
    C_Vertex3f(0.403189898f, 0.0808532238f, 0.f),
    C_Vertex3f(0.403189898f, 0.0808532238f, -1.f),
    C_Normal3f(0.790178895f, -0.612876296f, 0.f),
    C_Vertex3f(0.393282652f, 0.067013979f, 0.f),
    C_Vertex3f(0.393282652f, 0.067013979f, -1.f),
    C_Normal3f(0.739232898f, -0.673449814f, 0.f),
    C_Vertex3f(0.382563472f, 0.0542596579f, 0.f),
    C_Vertex3f(0.382563472f, 0.0542596579f, -1.f),
    C_Normal3f(0.681848466f, -0.731493473f, 0.f),
    C_Vertex3f(0.371032357f, 0.0425902605f, 0.f),
    C_Vertex3f(0.371032357f, 0.0425902605f, -1.f),
    C_Normal3f(0.618770778f, -0.785571575f, 0.f),
    C_Vertex3f(0.358689308f, 0.0320057869f, 0.f),
    C_Vertex3f(0.358689308f, 0.0320057869f, -1.f),
    C_Normal3f(0.552422583f, -0.833564162f, 0.f),
    C_Vertex3f(0.345534325f, 0.022506237f, 0.f),
    C_Vertex3f(0.345534325f, 0.022506237f, -1.f),
    C_Normal3f(0.482676893f, -0.875798404f, 0.f),
    C_Vertex3f(0.331812859f, 0.0141941309f, 0.f),
    C_Vertex3f(0.331812859f, 0.0141941309f, -1.f),
    C_Normal3f(0.409005195f, -0.912532032f, 0.f),
    C_Vertex3f(0.31752491f, 0.0070694685f, 0.f),
    C_Vertex3f(0.31752491f, 0.0070694685f, -1.f),
    C_Normal3f(0.332863837f, -0.942974925f, 0.f),
    C_Vertex3f(0.302670479f, 0.00113224983f, 0.f),
    C_Vertex3f(0.302670479f, 0.00113224983f, -1.f),
    C_Normal3f(0.255839109f, -0.966719389f, 0.f),
    C_Vertex3f(0.287249565f, -0.0036175251f, 0.f),
    C_Vertex3f(0.287249565f, -0.0036175251f, -1.f),
    C_Normal3f(0.179488927f, -0.983760059f, 0.f),
    C_Vertex3f(0.271262169f, -0.0071798563f, 0.f),
    C_Vertex3f(0.271262169f, -0.0071798563f, -1.f),
    C_Normal3f(0.105200998f, -0.994450927f, 0.f),
    C_Vertex3f(0.25470829f, -0.00955474377f, 0.f),
    C_Vertex3f(0.25470829f, -0.00955474377f, -1.f),
    C_Normal3f(0.0335695483f, -0.999436319f, 0.f),
    C_Vertex3f(0.237587929f, -0.0107421875f, 0.f),
    C_Vertex3f(0.237587929f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0360784829f, -0.999348938f, 0.f),
    C_Vertex3f(0.219355583f, -0.0107421875f, 0.f),
    C_Vertex3f(0.219355583f, -0.0107421875f, -1.f),
    C_Normal3f(-0.111246377f, -0.993792892f, 0.f),
    C_Vertex3f(0.201437473f, -0.0094370842f, 0.f),
    C_Vertex3f(0.201437473f, -0.0094370842f, -1.f),
    C_Normal3f(-0.192543298f, -0.981288433f, 0.f),
    C_Vertex3f(0.184379101f, -0.00682687759f, 0.f),
    C_Vertex3f(0.184379101f, -0.00682687759f, -1.f),
    C_Normal3f(-0.278238982f, -0.960511923f, 0.f),
    C_Vertex3f(0.168180466f, -0.00291156769f, 0.f),
    C_Vertex3f(0.168180466f, -0.00291156769f, -1.f),
    C_Normal3f(-0.366508901f, -0.930414557f, 0.f),
    C_Vertex3f(0.152841568f, 0.00230884552f, 0.f),
    C_Vertex3f(0.152841568f, 0.00230884552f, -1.f),
    C_Normal3f(-0.45497632f, -0.890503526f, 0.f),
    C_Vertex3f(0.138362408f, 0.00883436203f, 0.f),
    C_Vertex3f(0.138362408f, 0.00883436203f, -1.f),
    C_Normal3f(-0.540945828f, -0.84105742f, 0.f),
    C_Vertex3f(0.124742985f, 0.0166649818f, 0.f),
    C_Vertex3f(0.124742985f, 0.0166649818f, -1.f),
    C_Normal3f(-0.621763289f, -0.783205152f, 0.f),
    C_Vertex3f(0.111983299f, 0.025800705f, 0.f),
    C_Vertex3f(0.111983299f, 0.025800705f, -1.f),
    C_Normal3f(-0.696364462f, -0.717688382f, 0.f),
    C_Vertex3f(0.100083351f, 0.0362415314f, 0.f),
    C_Vertex3f(0.100083351f, 0.0362415314f, -1.f),
    C_Normal3f(-0.765257955f, -0.643723667f, 0.f),
    C_Vertex3f(0.0893707275f, 0.0477414131f, 0.f),
    C_Vertex3f(0.0893707275f, 0.0477414131f, -1.f),
    C_Normal3f(-0.826001585f, -0.563667774f, 0.f),
    C_Vertex3f(0.0798454285f, 0.0603003502f, 0.f),
    C_Vertex3f(0.0798454285f, 0.0603003502f, -1.f),
    C_Normal3f(-0.877178311f, -0.480164915f, 0.f),
    C_Vertex3f(0.0715074539f, 0.0739183426f, 0.f),
    C_Vertex3f(0.0715074539f, 0.0739183426f, -1.f),
    C_Normal3f(-0.918270767f, -0.395953089f, 0.f),
    C_Vertex3f(0.0643568039f, 0.0885953903f, 0.f),
    C_Vertex3f(0.0643568039f, 0.0885953903f, -1.f),
    C_Normal3f(-0.949593723f, -0.313483f, 0.f),
    C_Vertex3f(0.0583934784f, 0.104331493f, 0.f),
    C_Vertex3f(0.0583934784f, 0.104331493f, -1.f),
    C_Normal3f(-0.972075641f, -0.234668016f, 0.f),
    C_Vertex3f(0.0536174774f, 0.121126652f, 0.f),
    C_Vertex3f(0.0536174774f, 0.121126652f, -1.f),
    C_Normal3f(-0.984969437f, -0.172728643f, 0.f),
    C_Vertex3f(0.050028801f, 0.138980865f, 0.f),
    C_Vertex3f(0.050028801f, 0.138980865f, -1.f),
    C_Normal3f(-0.992035985f, -0.125955358f, 0.f),
    C_Vertex3f(0.048828125f, 0.1484375f, 0.f),
    C_Vertex3f(0.048828125f, 0.1484375f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999446273f, 0.0332718454f, 0.f),
    C_Vertex3f(0.370605469f, 0.422379494f, 0.f),
    C_Vertex3f(0.370605469f, 0.422379494f, -1.f),
    C_Normal3f(-0.99942416f, -0.0339323394f, 0.f),
    C_Vertex3f(0.370605469f, 0.439313889f, 0.f),
    C_Vertex3f(0.370605469f, 0.439313889f, -1.f),
    C_Normal3f(-0.994265616f, -0.106938951f, 0.f),
    C_Vertex3f(0.369485259f, 0.455373526f, 0.f),
    C_Vertex3f(0.369485259f, 0.455373526f, -1.f),
    C_Normal3f(-0.982446074f, -0.186546683f, 0.f),
    C_Vertex3f(0.36724484f, 0.470559359f, 0.f),
    C_Vertex3f(0.36724484f, 0.470559359f, -1.f),
    C_Normal3f(-0.96231848f, -0.271925062f, 0.f),
    C_Vertex3f(0.363884211f, 0.484871387f, 0.f),
    C_Vertex3f(0.363884211f, 0.484871387f, -1.f),
    C_Normal3f(-0.932370663f, -0.361503839f, 0.f),
    C_Vertex3f(0.359403372f, 0.498309612f, 0.f),
    C_Vertex3f(0.359403372f, 0.498309612f, -1.f),
    C_Normal3f(-0.891544878f, -0.452932298f, 0.f),
    C_Vertex3f(0.353802323f, 0.510874033f, 0.f),
    C_Vertex3f(0.353802323f, 0.510874033f, -1.f),
    C_Normal3f(-0.839585543f, -0.543227494f, 0.f),
    C_Vertex3f(0.347081065f, 0.52256465f, 0.f),
    C_Vertex3f(0.347081065f, 0.52256465f, -1.f),
    C_Normal3f(-0.777289271f, -0.629143417f, 0.f),
    C_Vertex3f(0.339239597f, 0.533381462f, 0.f),
    C_Vertex3f(0.339239597f, 0.533381462f, -1.f),
    C_Normal3f(-0.711899638f, -0.702281117f, 0.f),
    C_Vertex3f(0.33027792f, 0.543324471f, 0.f),
    C_Vertex3f(0.33027792f, 0.543324471f, -1.f),
    C_Normal3f(-0.641390502f, -0.767214537f, 0.f),
    C_Vertex3f(0.320848346f, 0.552024603f, 0.f),
    C_Vertex3f(0.320848346f, 0.552024603f, -1.f),
    C_Normal3f(-0.559310377f, -0.828958213f, 0.f),
    C_Vertex3f(0.310950875f, 0.559481859f, 0.f),
    C_Vertex3f(0.310950875f, 0.559481859f, -1.f),
    C_Normal3f(-0.466684163f, -0.884424031f, 0.f),
    C_Vertex3f(0.300585508f, 0.565696239f, 0.f),
    C_Vertex3f(0.300585508f, 0.565696239f, -1.f),
    C_Normal3f(-0.365815043f, -0.930687547f, 0.f),
    C_Vertex3f(0.289752245f, 0.570667744f, 0.f),
    C_Vertex3f(0.289752245f, 0.570667744f, -1.f),
    C_Normal3f(-0.260097116f, -0.96558249f, 0.f),
    C_Vertex3f(0.278451085f, 0.574396372f, 0.f),
    C_Vertex3f(0.278451085f, 0.574396372f, -1.f),
    C_Normal3f(-0.153480321f, -0.98815161f, 0.f),
    C_Vertex3f(0.266682029f, 0.576882124f, 0.f),
    C_Vertex3f(0.266682029f, 0.576882124f, -1.f),
    C_Normal3f(-0.0493383482f, -0.998782158f, 0.f),
    C_Vertex3f(0.254445076f, 0.578125f, 0.f),
    C_Vertex3f(0.254445076f, 0.578125f, -1.f),
    C_Normal3f(0.0521502495f, -0.998639226f, 0.f),
    C_Vertex3f(0.241521835f, 0.578125f, 0.f),
    C_Vertex3f(0.241521835f, 0.578125f, -1.f),
    C_Normal3f(0.158699751f, -0.98732686f, 0.f),
    C_Vertex3f(0.228814125f, 0.576786518f, 0.f),
    C_Vertex3f(0.228814125f, 0.576786518f, -1.f),
    C_Normal3f(0.267427653f, -0.963577926f, 0.f),
    C_Vertex3f(0.216540337f, 0.574109554f, 0.f),
    C_Vertex3f(0.216540337f, 0.574109554f, -1.f),
    C_Normal3f(0.373833179f, -0.927495956f, 0.f),
    C_Vertex3f(0.20470047f, 0.570094109f, 0.f),
    C_Vertex3f(0.20470047f, 0.570094109f, -1.f),
    C_Normal3f(0.47399804f, -0.880525947f, 0.f),
    C_Vertex3f(0.193294525f, 0.564740181f, 0.f),
    C_Vertex3f(0.193294525f, 0.564740181f, -1.f),
    C_Normal3f(0.564836681f, -0.825202644f, 0.f),
    C_Vertex3f(0.182322502f, 0.558047771f, 0.f),
    C_Vertex3f(0.182322502f, 0.558047771f, -1.f),
    C_Normal3f(0.644510388f, -0.764595509f, 0.f),
    C_Vertex3f(0.171784401f, 0.55001688f, 0.f),
    C_Vertex3f(0.171784401f, 0.55001688f, -1.f),
    C_Normal3f(0.712458789f, -0.70171392f, 0.f),
    C_Vertex3f(0.161680222f, 0.540647507f, 0.f),
    C_Vertex3f(0.161680222f, 0.540647507f, -1.f),
    C_Normal3f(0.77461046f, -0.6324386f, 0.f),
    C_Vertex3f(0.152009964f, 0.529939651f, 0.f),
    C_Vertex3f(0.152009964f, 0.529939651f, -1.f),
    C_Normal3f(0.834422529f, -0.551125288f, 0.f),
    C_Vertex3f(0.143548489f, 0.51843977f, 0.f),
    C_Vertex3f(0.143548489f, 0.51843977f, -1.f),
    C_Normal3f(0.885766983f, -0.464130133f, 0.f),
    C_Vertex3f(0.136295795f, 0.506147861f, 0.f),
    C_Vertex3f(0.136295795f, 0.506147861f, -1.f),
    C_Normal3f(0.92734468f, -0.374208391f, 0.f),
    C_Vertex3f(0.130251884f, 0.493063927f, 0.f),
    C_Vertex3f(0.130251884f, 0.493063927f, -1.f),
    C_Normal3f(0.958761394f, -0.284212112f, 0.f),
    C_Vertex3f(0.125416756f, 0.479187965f, 0.f),
    C_Vertex3f(0.125416756f, 0.479187965f, -1.f),
    C_Normal3f(0.980465889f, -0.196689069f, 0.f),
    C_Vertex3f(0.121790409f, 0.464519978f, 0.f),
    C_Vertex3f(0.121790409f, 0.464519978f, -1.f),
    C_Normal3f(0.993525207f, -0.113611899f, 0.f),
    C_Vertex3f(0.119372845f, 0.449059963f, 0.f),
    C_Vertex3f(0.119372845f, 0.449059963f, -1.f),
    C_Normal3f(0.999305665f, -0.0372573622f, 0.f),
    C_Vertex3f(0.118164063f, 0.432807922f, 0.f),
    C_Vertex3f(0.118164063f, 0.432807922f, -1.f),
    C_Normal3f(0.999304354f, 0.0372940972f, 0.f),
    C_Vertex3f(0.118164063f, 0.416638374f, 0.f),
    C_Vertex3f(0.118164063f, 0.416638374f, -1.f),
    C_Normal3f(0.992753327f, 0.120169558f, 0.f),
    C_Vertex3f(0.119311571f, 0.402060151f, 0.f),
    C_Vertex3f(0.119311571f, 0.402060151f, -1.f),
    C_Normal3f(0.978168011f, 0.207815856f, 0.f),
    C_Vertex3f(0.121606588f, 0.388198733f, 0.f),
    C_Vertex3f(0.121606588f, 0.388198733f, -1.f),
    C_Normal3f(0.954041541f, 0.299674273f, 0.f),
    C_Vertex3f(0.125049114f, 0.375054121f, 0.f),
    C_Vertex3f(0.125049114f, 0.375054121f, -1.f),
    C_Normal3f(0.919387937f, 0.393352062f, 0.f),
    C_Vertex3f(0.129639149f, 0.362626314f, 0.f),
    C_Vertex3f(0.129639149f, 0.362626314f, -1.f),
    C_Normal3f(0.874016821f, 0.485895783f, 0.f),
    C_Vertex3f(0.135376692f, 0.350915313f, 0.f),
    C_Vertex3f(0.135376692f, 0.350915313f, -1.f),
    C_Normal3f(0.818734288f, 0.574172556f, 0.f),
    C_Vertex3f(0.142261744f, 0.339921117f, 0.f),
    C_Vertex3f(0.142261744f, 0.339921117f, -1.f),
    C_Normal3f(0.75531745f, 0.655359209f, 0.f),
    C_Vertex3f(0.150294304f, 0.329643726f, 0.f),
    C_Vertex3f(0.150294304f, 0.329643726f, -1.f),
    C_Normal3f(0.687980294f, 0.725729406f, 0.f),
    C_Vertex3f(0.159474373f, 0.320083141f, 0.f),
    C_Vertex3f(0.159474373f, 0.320083141f, -1.f),
    C_Normal3f(0.613289833f, 0.789857984f, 0.f),
    C_Vertex3f(0.169203997f, 0.311717629f, 0.f),
    C_Vertex3f(0.169203997f, 0.311717629f, -1.f),
    C_Normal3f(0.528649509f, 0.848840177f, 0.f),
    C_Vertex3f(0.179483175f, 0.304547191f, 0.f),
    C_Vertex3f(0.179483175f, 0.304547191f, -1.f),
    C_Normal3f(0.435899973f, 0.899995148f, 0.f),
    C_Vertex3f(0.190311909f, 0.298571825f, 0.f),
    C_Vertex3f(0.190311909f, 0.298571825f, -1.f),
    C_Normal3f(0.337836415f, 0.941204846f, 0.f),
    C_Vertex3f(0.201690197f, 0.293791533f, 0.f),
    C_Vertex3f(0.201690197f, 0.293791533f, -1.f),
    C_Normal3f(0.237815097f, 0.971310437f, 0.f),
    C_Vertex3f(0.21361804f, 0.290206313f, 0.f),
    C_Vertex3f(0.21361804f, 0.290206313f, -1.f),
    C_Normal3f(0.139204189f, 0.990263641f, 0.f),
    C_Vertex3f(0.226095438f, 0.287816167f, 0.f),
    C_Vertex3f(0.226095438f, 0.287816167f, -1.f),
    C_Normal3f(0.0448363572f, 0.99899435f, 0.f),
    C_Vertex3f(0.239122391f, 0.286621094f, 0.f),
    C_Vertex3f(0.239122391f, 0.286621094f, -1.f),
    C_Normal3f(-0.0448513888f, 0.998993695f, 0.f),
    C_Vertex3f(0.25272274f, 0.286621094f, 0.f),
    C_Vertex3f(0.25272274f, 0.286621094f, -1.f),
    C_Normal3f(-0.139604107f, 0.990207374f, 0.f),
    C_Vertex3f(0.265740752f, 0.287816167f, 0.f),
    C_Vertex3f(0.265740752f, 0.287816167f, -1.f),
    C_Normal3f(-0.239553437f, 0.970883191f, 0.f),
    C_Vertex3f(0.278152585f, 0.290206313f, 0.f),
    C_Vertex3f(0.278152585f, 0.290206313f, -1.f),
    C_Normal3f(-0.341743469f, 0.939793289f, 0.f),
    C_Vertex3f(0.289958239f, 0.293791533f, 0.f),
    C_Vertex3f(0.289958239f, 0.293791533f, -1.f),
    C_Normal3f(-0.442575127f, 0.896731496f, 0.f),
    C_Vertex3f(0.301157713f, 0.298571825f, 0.f),
    C_Vertex3f(0.301157713f, 0.298571825f, -1.f),
    C_Normal3f(-0.538308084f, 0.842748165f, 0.f),
    C_Vertex3f(0.311751008f, 0.304547191f, 0.f),
    C_Vertex3f(0.311751008f, 0.304547191f, -1.f),
    C_Normal3f(-0.625713885f, 0.780052602f, 0.f),
    C_Vertex3f(0.321738124f, 0.311717629f, 0.f),
    C_Vertex3f(0.321738124f, 0.311717629f, -1.f),
    C_Normal3f(-0.702593148f, 0.711591899f, 0.f),
    C_Vertex3f(0.331119061f, 0.320083141f, 0.f),
    C_Vertex3f(0.331119061f, 0.320083141f, -1.f),
    C_Normal3f(-0.772938728f, 0.634480715f, 0.f),
    C_Vertex3f(0.339893818f, 0.329643726f, 0.f),
    C_Vertex3f(0.339893818f, 0.329643726f, -1.f),
    C_Normal3f(-0.837872505f, 0.545865953f, 0.f),
    C_Vertex3f(0.347571731f, 0.340126157f, 0.f),
    C_Vertex3f(0.347571731f, 0.340126157f, -1.f),
    C_Normal3f(-0.891400456f, 0.453216463f, 0.f),
    C_Vertex3f(0.354152799f, 0.351530433f, 0.f),
    C_Vertex3f(0.354152799f, 0.351530433f, -1.f),
    C_Normal3f(-0.932912111f, 0.360104203f, 0.f),
    C_Vertex3f(0.359637022f, 0.363856554f, 0.f),
    C_Vertex3f(0.359637022f, 0.363856554f, -1.f),
    C_Normal3f(-0.962955475f, 0.269660354f, 0.f),
    C_Vertex3f(0.364024401f, 0.377104521f, 0.f),
    C_Vertex3f(0.364024401f, 0.377104521f, -1.f),
    C_Normal3f(-0.982886195f, 0.184213832f, 0.f),
    C_Vertex3f(0.367314935f, 0.391274333f, 0.f),
    C_Vertex3f(0.367314935f, 0.391274333f, -1.f),
    C_Normal3f(-0.994450986f, 0.105200388f, 0.f),
    C_Vertex3f(0.369508624f, 0.406365991f, 0.f),
    C_Vertex3f(0.369508624f, 0.406365991f, -1.f),
    C_Normal3f(-0.999446273f, 0.0332718454f, 0.f),
    C_Vertex3f(0.370605469f, 0.422379494f, 0.f),
    C_Vertex3f(0.370605469f, 0.422379494f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.371190071f, 0.595930696f, 0.f),
    C_Vertex3f(0.359033108f, 0.605232239f, 0.f),
    C_Vertex3f(0.359403372f, 0.498309612f, 0.f),
    C_Vertex3f(0.363884211f, 0.484871387f, 0.f),
    C_Vertex3f(0.36724484f, 0.470559359f, 0.f),
    C_Vertex3f(0.369485259f, 0.455373526f, 0.f),
    C_Vertex3f(0.370605469f, 0.439313889f, 0.f),
    C_Vertex3f(0.373500824f, 0.292232513f, 0.f),
    C_Vertex3f(0.382569313f, 0.585735202f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320848346f, 0.552024603f, 0.f),
    C_Vertex3f(0.33027792f, 0.543324471f, 0.f),
    C_Vertex3f(0.332829356f, 0.620996475f, 0.f),
    C_Vertex3f(0.339239597f, 0.533381462f, 0.f),
    C_Vertex3f(0.346098423f, 0.613639832f, 0.f),
    C_Vertex3f(0.347081065f, 0.52256465f, 0.f),
    C_Vertex3f(0.359033108f, 0.605232239f, 0.f),
    C_Vertex3f(0.353802323f, 0.510874033f, 0.f),
    C_Vertex3f(0.359403372f, 0.498309612f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278451085f, 0.574396372f, 0.f),
    C_Vertex3f(0.289752245f, 0.570667744f, 0.f),
    C_Vertex3f(0.291015863f, 0.636760712f, 0.f),
    C_Vertex3f(0.300585508f, 0.565696239f, 0.f),
    C_Vertex3f(0.305288076f, 0.632556915f, 0.f),
    C_Vertex3f(0.310950875f, 0.559481859f, 0.f),
    C_Vertex3f(0.319225907f, 0.62730217f, 0.f),
    C_Vertex3f(0.320848346f, 0.552024603f, 0.f),
    C_Vertex3f(0.332829356f, 0.620996475f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.228814125f, 0.576786518f, 0.f),
    C_Vertex3f(0.241521835f, 0.578125f, 0.f),
    C_Vertex3f(0.246192932f, 0.643066406f, 0.f),
    C_Vertex3f(0.254445076f, 0.578125f, 0.f),
    C_Vertex3f(0.261468291f, 0.642015457f, 0.f),
    C_Vertex3f(0.266682029f, 0.576882124f, 0.f),
    C_Vertex3f(0.276409268f, 0.639913559f, 0.f),
    C_Vertex3f(0.278451085f, 0.574396372f, 0.f),
    C_Vertex3f(0.291015863f, 0.636760712f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193294525f, 0.564740181f, 0.f),
    C_Vertex3f(0.20470047f, 0.570094109f, 0.f),
    C_Vertex3f(0.206877112f, 0.641210437f, 0.f),
    C_Vertex3f(0.216540337f, 0.574109554f, 0.f),
    C_Vertex3f(0.227581978f, 0.643066406f, 0.f),
    C_Vertex3f(0.228814125f, 0.576786518f, 0.f),
    C_Vertex3f(0.246192932f, 0.643066406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.187079549f, 0.637498498f, 0.f),
    C_Vertex3f(0.168189287f, 0.63193059f, 0.f),
    C_Vertex3f(0.171784401f, 0.55001688f, 0.f),
    C_Vertex3f(0.182322502f, 0.558047771f, 0.f),
    C_Vertex3f(0.193294525f, 0.564740181f, 0.f),
    C_Vertex3f(0.206877112f, 0.641210437f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.136295795f, 0.506147861f, 0.f),
    C_Vertex3f(0.143548489f, 0.51843977f, 0.f),
    C_Vertex3f(0.150206327f, 0.624506712f, 0.f),
    C_Vertex3f(0.152009964f, 0.529939651f, 0.f),
    C_Vertex3f(0.168189287f, 0.63193059f, 0.f),
    C_Vertex3f(0.161680222f, 0.540647507f, 0.f),
    C_Vertex3f(0.171784401f, 0.55001688f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.13313067f, 0.615226865f, 0.f),
    C_Vertex3f(0.116962314f, 0.604091048f, 0.f),
    C_Vertex3f(0.119372845f, 0.449059963f, 0.f),
    C_Vertex3f(0.121790409f, 0.464519978f, 0.f),
    C_Vertex3f(0.125416756f, 0.479187965f, 0.f),
    C_Vertex3f(0.130251884f, 0.493063927f, 0.f),
    C_Vertex3f(0.136295795f, 0.506147861f, 0.f),
    C_Vertex3f(0.150206327f, 0.624506712f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.037109375f, 0.43907547f, 0.f),
    C_Vertex3f(0.0388284922f, 0.392675519f, 0.f),
    C_Vertex3f(0.038903594f, 0.462233663f, 0.f),
    C_Vertex3f(0.0422667265f, 0.371385932f, 0.f),
    C_Vertex3f(0.0424920321f, 0.48420465f, 0.f),
    C_Vertex3f(0.047424078f, 0.351242781f, 0.f),
    C_Vertex3f(0.0478746891f, 0.504988432f, 0.f),
    C_Vertex3f(0.0543005466f, 0.332246065f, 0.f),
    C_Vertex3f(0.0550515652f, 0.524585009f, 0.f),
    C_Vertex3f(0.0628961325f, 0.314395785f, 0.f),
    C_Vertex3f(0.0640226603f, 0.54299438f, 0.f),
    C_Vertex3f(0.0732108355f, 0.297691941f, 0.f),
    C_Vertex3f(0.0747879744f, 0.560216546f, 0.f),
    C_Vertex3f(0.0852446556f, 0.282134533f, 0.f),
    C_Vertex3f(0.0873475075f, 0.576251507f, 0.f),
    C_Vertex3f(0.0989975929f, 0.26772356f, 0.f),
    C_Vertex3f(0.10170126f, 0.591099262f, 0.f),
    C_Vertex3f(0.113514304f, 0.255113959f, 0.f),
    C_Vertex3f(0.116962314f, 0.604091048f, 0.f),
    C_Vertex3f(0.118164063f, 0.416638374f, 0.f),
    C_Vertex3f(0.118164063f, 0.432807922f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.458007813f, 0.318071365f, 0.f),
    C_Vertex3f(0.458007813f, 0.348152161f, 0.f),
    C_Vertex3f(0.457175374f, 0.288952589f, 0.f),
    C_Vertex3f(0.457168698f, 0.375892043f, 0.f),
    C_Vertex3f(0.455510497f, 0.261396646f, 0.f),
    C_Vertex3f(0.45549047f, 0.401891828f, 0.f),
    C_Vertex3f(0.453013182f, 0.235403538f, 0.f),
    C_Vertex3f(0.452973127f, 0.426151514f, 0.f),
    C_Vertex3f(0.449683428f, 0.210973263f, 0.f),
    C_Vertex3f(0.449616671f, 0.448671103f, 0.f),
    C_Vertex3f(0.445521235f, 0.188105822f, 0.f),
    C_Vertex3f(0.4454211f, 0.469450593f, 0.f),
    C_Vertex3f(0.440526605f, 0.166801214f, 0.f),
    C_Vertex3f(0.440386415f, 0.488489985f, 0.f),
    C_Vertex3f(0.434699535f, 0.147059441f, 0.f),
    C_Vertex3f(0.434512615f, 0.50578928f, 0.f),
    C_Vertex3f(0.428040028f, 0.128880501f, 0.f),
    C_Vertex3f(0.427799702f, 0.521348476f, 0.f),
    C_Vertex3f(0.420568585f, 0.111786485f, 0.f),
    C_Vertex3f(0.420309067f, 0.536013722f, 0.f),
    C_Vertex3f(0.412285209f, 0.0957773924f, 0.f),
    C_Vertex3f(0.41204071f, 0.549785018f, 0.f),
    C_Vertex3f(0.403189898f, 0.0808532238f, 0.f),
    C_Vertex3f(0.402994633f, 0.562662363f, 0.f),
    C_Vertex3f(0.393282652f, 0.067013979f, 0.f),
    C_Vertex3f(0.393170834f, 0.574645758f, 0.f),
    C_Vertex3f(0.382569313f, 0.585735202f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.382563472f, 0.0542596579f, 0.f),
    C_Vertex3f(0.393282652f, 0.067013979f, 0.f),
    C_Vertex3f(0.382569313f, 0.585735202f, 0.f),
    C_Vertex3f(0.379837036f, 0.288516998f, 0.f),
    C_Vertex3f(0.379837036f, 0.273658752f, 0.f),
    C_Vertex3f(0.378363609f, 0.249399185f, 0.f),
    C_Vertex3f(0.375416756f, 0.225355148f, 0.f),
    C_Vertex3f(0.371032357f, 0.0425902605f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.382569313f, 0.585735202f, 0.f),
    C_Vertex3f(0.373500824f, 0.292232513f, 0.f),
    C_Vertex3f(0.379394531f, 0.301628113f, 0.f),
    C_Vertex3f(0.379615784f, 0.296382904f, 0.f),
    C_Vertex3f(0.379837036f, 0.288516998f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.373500824f, 0.292232513f, 0.f),
    C_Vertex3f(0.370605469f, 0.439313889f, 0.f),
    C_Vertex3f(0.370605469f, 0.422379494f, 0.f),
    C_Vertex3f(0.369508624f, 0.406365991f, 0.f),
    C_Vertex3f(0.367314935f, 0.391274333f, 0.f),
    C_Vertex3f(0.364024401f, 0.377104521f, 0.f),
    C_Vertex3f(0.359637022f, 0.363856554f, 0.f),
    C_Vertex3f(0.359557629f, 0.275216579f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.359637022f, 0.363856554f, 0.f),
    C_Vertex3f(0.354152799f, 0.351530433f, 0.f),
    C_Vertex3f(0.359557629f, 0.275216579f, 0.f),
    C_Vertex3f(0.347571731f, 0.340126157f, 0.f),
    C_Vertex3f(0.343458652f, 0.25997591f, 0.f),
    C_Vertex3f(0.339893818f, 0.329643726f, 0.f),
    C_Vertex3f(0.331119061f, 0.320083141f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.325203896f, 0.246510506f, 0.f),
    C_Vertex3f(0.343458652f, 0.25997591f, 0.f),
    C_Vertex3f(0.331119061f, 0.320083141f, 0.f),
    C_Vertex3f(0.321738124f, 0.311717629f, 0.f),
    C_Vertex3f(0.311751008f, 0.304547191f, 0.f),
    C_Vertex3f(0.304793358f, 0.234820366f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.311751008f, 0.304547191f, 0.f),
    C_Vertex3f(0.301157713f, 0.298571825f, 0.f),
    C_Vertex3f(0.304793358f, 0.234820366f, 0.f),
    C_Vertex3f(0.289958239f, 0.293791533f, 0.f),
    C_Vertex3f(0.283536434f, 0.226052761f, 0.f),
    C_Vertex3f(0.278152585f, 0.290206313f, 0.f),
    C_Vertex3f(0.265740752f, 0.287816167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.261433125f, 0.220207691f, 0.f),
    C_Vertex3f(0.283536434f, 0.226052761f, 0.f),
    C_Vertex3f(0.265740752f, 0.287816167f, 0.f),
    C_Vertex3f(0.25272274f, 0.286621094f, 0.f),
    C_Vertex3f(0.239122391f, 0.286621094f, 0.f),
    C_Vertex3f(0.238483429f, 0.217285156f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.226095438f, 0.287816167f, 0.f),
    C_Vertex3f(0.21361804f, 0.290206313f, 0.f),
    C_Vertex3f(0.216653824f, 0.217285156f, 0.f),
    C_Vertex3f(0.238483429f, 0.217285156f, 0.f),
    C_Vertex3f(0.239122391f, 0.286621094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.201690197f, 0.293791533f, 0.f),
    C_Vertex3f(0.190311909f, 0.298571825f, 0.f),
    C_Vertex3f(0.197554469f, 0.219086528f, 0.f),
    C_Vertex3f(0.216653824f, 0.217285156f, 0.f),
    C_Vertex3f(0.21361804f, 0.290206313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.159474373f, 0.320083141f, 0.f),
    C_Vertex3f(0.161647081f, 0.228093386f, 0.f),
    C_Vertex3f(0.169203997f, 0.311717629f, 0.f),
    C_Vertex3f(0.179218888f, 0.222689271f, 0.f),
    C_Vertex3f(0.179483175f, 0.304547191f, 0.f),
    C_Vertex3f(0.197554469f, 0.219086528f, 0.f),
    C_Vertex3f(0.190311909f, 0.298571825f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150294304f, 0.329643726f, 0.f),
    C_Vertex3f(0.142261744f, 0.339921117f, 0.f),
    C_Vertex3f(0.144839048f, 0.235298872f, 0.f),
    C_Vertex3f(0.161647081f, 0.228093386f, 0.f),
    C_Vertex3f(0.159474373f, 0.320083141f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142261744f, 0.339921117f, 0.f),
    C_Vertex3f(0.135376692f, 0.350915313f, 0.f),
    C_Vertex3f(0.144839048f, 0.235298872f, 0.f),
    C_Vertex3f(0.129639149f, 0.362626314f, 0.f),
    C_Vertex3f(0.128794789f, 0.24430573f, 0.f),
    C_Vertex3f(0.125049114f, 0.375054121f, 0.f),
    C_Vertex3f(0.121606588f, 0.388198733f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.128794789f, 0.24430573f, 0.f),
    C_Vertex3f(0.121606588f, 0.388198733f, 0.f),
    C_Vertex3f(0.119311571f, 0.402060151f, 0.f),
    C_Vertex3f(0.118164063f, 0.416638374f, 0.f),
    C_Vertex3f(0.113514304f, 0.255113959f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.375416756f, 0.225355148f, 0.f),
    C_Vertex3f(0.370996475f, 0.201526642f, 0.f),
    C_Vertex3f(0.371032357f, 0.0425902605f, 0.f),
    C_Vertex3f(0.365102768f, 0.177913666f, 0.f),
    C_Vertex3f(0.358689308f, 0.0320057869f, 0.f),
    C_Vertex3f(0.358515739f, 0.156732559f, 0.f),
    C_Vertex3f(0.35123539f, 0.137983322f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.345534325f, 0.022506237f, 0.f),
    C_Vertex3f(0.358689308f, 0.0320057869f, 0.f),
    C_Vertex3f(0.35123539f, 0.137983322f, 0.f),
    C_Vertex3f(0.343261719f, 0.121665955f, 0.f),
    C_Vertex3f(0.334594727f, 0.107780457f, 0.f),
    C_Vertex3f(0.331812859f, 0.0141941309f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.30104351f, 0.0744867325f, 0.f),
    C_Vertex3f(0.302670479f, 0.00113224983f, 0.f),
    C_Vertex3f(0.313485622f, 0.0841908455f, 0.f),
    C_Vertex3f(0.31752491f, 0.0070694685f, 0.f),
    C_Vertex3f(0.324669361f, 0.0952887535f, 0.f),
    C_Vertex3f(0.331812859f, 0.0141941309f, 0.f),
    C_Vertex3f(0.334594727f, 0.107780457f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.201437473f, -0.0094370842f, 0.f),
    C_Vertex3f(0.219355583f, -0.0107421875f, 0.f),
    C_Vertex3f(0.219984055f, 0.0537109375f, 0.f),
    C_Vertex3f(0.237587929f, -0.0107421875f, 0.f),
    C_Vertex3f(0.239624023f, 0.0537109375f, 0.f),
    C_Vertex3f(0.25470829f, -0.00955474377f, 0.f),
    C_Vertex3f(0.256633282f, 0.055788517f, 0.f),
    C_Vertex3f(0.271262169f, -0.0071798563f, 0.f),
    C_Vertex3f(0.272539616f, 0.059943676f, 0.f),
    C_Vertex3f(0.287249565f, -0.0036175251f, 0.f),
    C_Vertex3f(0.287343025f, 0.0661764145f, 0.f),
    C_Vertex3f(0.302670479f, 0.00113224983f, 0.f),
    C_Vertex3f(0.30104351f, 0.0744867325f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.167939186f, 0.0717697144f, 0.f),
    C_Vertex3f(0.168180466f, -0.00291156769f, 0.f),
    C_Vertex3f(0.183377743f, 0.0627403259f, 0.f),
    C_Vertex3f(0.184379101f, -0.00682687759f, 0.f),
    C_Vertex3f(0.200726032f, 0.0567207336f, 0.f),
    C_Vertex3f(0.201437473f, -0.0094370842f, 0.f),
    C_Vertex3f(0.219984055f, 0.0537109375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.157792568f, 0.0807991028f, 0.f),
    C_Vertex3f(0.151573539f, 0.0877351761f, 0.f),
    C_Vertex3f(0.152841568f, 0.00230884552f, 0.f),
    C_Vertex3f(0.168180466f, -0.00291156769f, 0.f),
    C_Vertex3f(0.167939186f, 0.0717697144f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151573539f, 0.0877351761f, 0.f),
    C_Vertex3f(0.145899892f, 0.0955877304f, 0.f),
    C_Vertex3f(0.152841568f, 0.00230884552f, 0.f),
    C_Vertex3f(0.140771627f, 0.104356766f, 0.f),
    C_Vertex3f(0.138362408f, 0.00883436203f, 0.f),
    C_Vertex3f(0.136188745f, 0.114042282f, 0.f),
    C_Vertex3f(0.132151246f, 0.124644279f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.100083351f, 0.0362415314f, 0.f),
    C_Vertex3f(0.111983299f, 0.025800705f, 0.f),
    C_Vertex3f(0.124511719f, 0.155273438f, 0.f),
    C_Vertex3f(0.124742985f, 0.0166649818f, 0.f),
    C_Vertex3f(0.125712395f, 0.148597717f, 0.f),
    C_Vertex3f(0.138362408f, 0.00883436203f, 0.f),
    C_Vertex3f(0.128659129f, 0.136162758f, 0.f),
    C_Vertex3f(0.132151246f, 0.124644279f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.124511719f, 0.155273438f, 0.f),
    C_Vertex3f(0.048828125f, 0.1484375f, 0.f),
    C_Vertex3f(0.050028801f, 0.138980865f, 0.f),
    C_Vertex3f(0.0536174774f, 0.121126652f, 0.f),
    C_Vertex3f(0.0583934784f, 0.104331493f, 0.f),
    C_Vertex3f(0.0643568039f, 0.0885953903f, 0.f),
    C_Vertex3f(0.0715074539f, 0.0739183426f, 0.f),
    C_Vertex3f(0.0798454285f, 0.0603003502f, 0.f),
    C_Vertex3f(0.0893707275f, 0.0477414131f, 0.f),
    C_Vertex3f(0.100083351f, 0.0362415314f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0388284922f, 0.392675519f, 0.f),
    C_Vertex3f(0.037109375f, 0.43907547f, 0.f),
    C_Vertex3f(0.037109375f, 0.415111542f, 0.f),
    C_Vertex3f(0.119372845f, 0.449059963f, 0.f),
    C_Vertex3f(0.116962314f, 0.604091048f, 0.f),
    C_Vertex3f(0.118164063f, 0.432807922f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.371190071f, 0.595930696f, -1.f),
    C_Vertex3f(0.359033108f, 0.605232239f, -1.f),
    C_Vertex3f(0.359403372f, 0.498309612f, -1.f),
    C_Vertex3f(0.363884211f, 0.484871387f, -1.f),
    C_Vertex3f(0.36724484f, 0.470559359f, -1.f),
    C_Vertex3f(0.369485259f, 0.455373526f, -1.f),
    C_Vertex3f(0.370605469f, 0.439313889f, -1.f),
    C_Vertex3f(0.373500824f, 0.292232513f, -1.f),
    C_Vertex3f(0.382569313f, 0.585735202f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320848346f, 0.552024603f, -1.f),
    C_Vertex3f(0.33027792f, 0.543324471f, -1.f),
    C_Vertex3f(0.332829356f, 0.620996475f, -1.f),
    C_Vertex3f(0.339239597f, 0.533381462f, -1.f),
    C_Vertex3f(0.346098423f, 0.613639832f, -1.f),
    C_Vertex3f(0.347081065f, 0.52256465f, -1.f),
    C_Vertex3f(0.359033108f, 0.605232239f, -1.f),
    C_Vertex3f(0.353802323f, 0.510874033f, -1.f),
    C_Vertex3f(0.359403372f, 0.498309612f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278451085f, 0.574396372f, -1.f),
    C_Vertex3f(0.289752245f, 0.570667744f, -1.f),
    C_Vertex3f(0.291015863f, 0.636760712f, -1.f),
    C_Vertex3f(0.300585508f, 0.565696239f, -1.f),
    C_Vertex3f(0.305288076f, 0.632556915f, -1.f),
    C_Vertex3f(0.310950875f, 0.559481859f, -1.f),
    C_Vertex3f(0.319225907f, 0.62730217f, -1.f),
    C_Vertex3f(0.320848346f, 0.552024603f, -1.f),
    C_Vertex3f(0.332829356f, 0.620996475f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.228814125f, 0.576786518f, -1.f),
    C_Vertex3f(0.241521835f, 0.578125f, -1.f),
    C_Vertex3f(0.246192932f, 0.643066406f, -1.f),
    C_Vertex3f(0.254445076f, 0.578125f, -1.f),
    C_Vertex3f(0.261468291f, 0.642015457f, -1.f),
    C_Vertex3f(0.266682029f, 0.576882124f, -1.f),
    C_Vertex3f(0.276409268f, 0.639913559f, -1.f),
    C_Vertex3f(0.278451085f, 0.574396372f, -1.f),
    C_Vertex3f(0.291015863f, 0.636760712f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193294525f, 0.564740181f, -1.f),
    C_Vertex3f(0.20470047f, 0.570094109f, -1.f),
    C_Vertex3f(0.206877112f, 0.641210437f, -1.f),
    C_Vertex3f(0.216540337f, 0.574109554f, -1.f),
    C_Vertex3f(0.227581978f, 0.643066406f, -1.f),
    C_Vertex3f(0.228814125f, 0.576786518f, -1.f),
    C_Vertex3f(0.246192932f, 0.643066406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.187079549f, 0.637498498f, -1.f),
    C_Vertex3f(0.168189287f, 0.63193059f, -1.f),
    C_Vertex3f(0.171784401f, 0.55001688f, -1.f),
    C_Vertex3f(0.182322502f, 0.558047771f, -1.f),
    C_Vertex3f(0.193294525f, 0.564740181f, -1.f),
    C_Vertex3f(0.206877112f, 0.641210437f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.136295795f, 0.506147861f, -1.f),
    C_Vertex3f(0.143548489f, 0.51843977f, -1.f),
    C_Vertex3f(0.150206327f, 0.624506712f, -1.f),
    C_Vertex3f(0.152009964f, 0.529939651f, -1.f),
    C_Vertex3f(0.168189287f, 0.63193059f, -1.f),
    C_Vertex3f(0.161680222f, 0.540647507f, -1.f),
    C_Vertex3f(0.171784401f, 0.55001688f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.13313067f, 0.615226865f, -1.f),
    C_Vertex3f(0.116962314f, 0.604091048f, -1.f),
    C_Vertex3f(0.119372845f, 0.449059963f, -1.f),
    C_Vertex3f(0.121790409f, 0.464519978f, -1.f),
    C_Vertex3f(0.125416756f, 0.479187965f, -1.f),
    C_Vertex3f(0.130251884f, 0.493063927f, -1.f),
    C_Vertex3f(0.136295795f, 0.506147861f, -1.f),
    C_Vertex3f(0.150206327f, 0.624506712f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.037109375f, 0.43907547f, -1.f),
    C_Vertex3f(0.0388284922f, 0.392675519f, -1.f),
    C_Vertex3f(0.038903594f, 0.462233663f, -1.f),
    C_Vertex3f(0.0422667265f, 0.371385932f, -1.f),
    C_Vertex3f(0.0424920321f, 0.48420465f, -1.f),
    C_Vertex3f(0.047424078f, 0.351242781f, -1.f),
    C_Vertex3f(0.0478746891f, 0.504988432f, -1.f),
    C_Vertex3f(0.0543005466f, 0.332246065f, -1.f),
    C_Vertex3f(0.0550515652f, 0.524585009f, -1.f),
    C_Vertex3f(0.0628961325f, 0.314395785f, -1.f),
    C_Vertex3f(0.0640226603f, 0.54299438f, -1.f),
    C_Vertex3f(0.0732108355f, 0.297691941f, -1.f),
    C_Vertex3f(0.0747879744f, 0.560216546f, -1.f),
    C_Vertex3f(0.0852446556f, 0.282134533f, -1.f),
    C_Vertex3f(0.0873475075f, 0.576251507f, -1.f),
    C_Vertex3f(0.0989975929f, 0.26772356f, -1.f),
    C_Vertex3f(0.10170126f, 0.591099262f, -1.f),
    C_Vertex3f(0.113514304f, 0.255113959f, -1.f),
    C_Vertex3f(0.116962314f, 0.604091048f, -1.f),
    C_Vertex3f(0.118164063f, 0.416638374f, -1.f),
    C_Vertex3f(0.118164063f, 0.432807922f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.458007813f, 0.318071365f, -1.f),
    C_Vertex3f(0.458007813f, 0.348152161f, -1.f),
    C_Vertex3f(0.457175374f, 0.288952589f, -1.f),
    C_Vertex3f(0.457168698f, 0.375892043f, -1.f),
    C_Vertex3f(0.455510497f, 0.261396646f, -1.f),
    C_Vertex3f(0.45549047f, 0.401891828f, -1.f),
    C_Vertex3f(0.453013182f, 0.235403538f, -1.f),
    C_Vertex3f(0.452973127f, 0.426151514f, -1.f),
    C_Vertex3f(0.449683428f, 0.210973263f, -1.f),
    C_Vertex3f(0.449616671f, 0.448671103f, -1.f),
    C_Vertex3f(0.445521235f, 0.188105822f, -1.f),
    C_Vertex3f(0.4454211f, 0.469450593f, -1.f),
    C_Vertex3f(0.440526605f, 0.166801214f, -1.f),
    C_Vertex3f(0.440386415f, 0.488489985f, -1.f),
    C_Vertex3f(0.434699535f, 0.147059441f, -1.f),
    C_Vertex3f(0.434512615f, 0.50578928f, -1.f),
    C_Vertex3f(0.428040028f, 0.128880501f, -1.f),
    C_Vertex3f(0.427799702f, 0.521348476f, -1.f),
    C_Vertex3f(0.420568585f, 0.111786485f, -1.f),
    C_Vertex3f(0.420309067f, 0.536013722f, -1.f),
    C_Vertex3f(0.412285209f, 0.0957773924f, -1.f),
    C_Vertex3f(0.41204071f, 0.549785018f, -1.f),
    C_Vertex3f(0.403189898f, 0.0808532238f, -1.f),
    C_Vertex3f(0.402994633f, 0.562662363f, -1.f),
    C_Vertex3f(0.393282652f, 0.067013979f, -1.f),
    C_Vertex3f(0.393170834f, 0.574645758f, -1.f),
    C_Vertex3f(0.382569313f, 0.585735202f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.382563472f, 0.0542596579f, -1.f),
    C_Vertex3f(0.393282652f, 0.067013979f, -1.f),
    C_Vertex3f(0.382569313f, 0.585735202f, -1.f),
    C_Vertex3f(0.379837036f, 0.288516998f, -1.f),
    C_Vertex3f(0.379837036f, 0.273658752f, -1.f),
    C_Vertex3f(0.378363609f, 0.249399185f, -1.f),
    C_Vertex3f(0.375416756f, 0.225355148f, -1.f),
    C_Vertex3f(0.371032357f, 0.0425902605f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.382569313f, 0.585735202f, -1.f),
    C_Vertex3f(0.373500824f, 0.292232513f, -1.f),
    C_Vertex3f(0.379394531f, 0.301628113f, -1.f),
    C_Vertex3f(0.379615784f, 0.296382904f, -1.f),
    C_Vertex3f(0.379837036f, 0.288516998f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.373500824f, 0.292232513f, -1.f),
    C_Vertex3f(0.370605469f, 0.439313889f, -1.f),
    C_Vertex3f(0.370605469f, 0.422379494f, -1.f),
    C_Vertex3f(0.369508624f, 0.406365991f, -1.f),
    C_Vertex3f(0.367314935f, 0.391274333f, -1.f),
    C_Vertex3f(0.364024401f, 0.377104521f, -1.f),
    C_Vertex3f(0.359637022f, 0.363856554f, -1.f),
    C_Vertex3f(0.359557629f, 0.275216579f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.359637022f, 0.363856554f, -1.f),
    C_Vertex3f(0.354152799f, 0.351530433f, -1.f),
    C_Vertex3f(0.359557629f, 0.275216579f, -1.f),
    C_Vertex3f(0.347571731f, 0.340126157f, -1.f),
    C_Vertex3f(0.343458652f, 0.25997591f, -1.f),
    C_Vertex3f(0.339893818f, 0.329643726f, -1.f),
    C_Vertex3f(0.331119061f, 0.320083141f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.325203896f, 0.246510506f, -1.f),
    C_Vertex3f(0.343458652f, 0.25997591f, -1.f),
    C_Vertex3f(0.331119061f, 0.320083141f, -1.f),
    C_Vertex3f(0.321738124f, 0.311717629f, -1.f),
    C_Vertex3f(0.311751008f, 0.304547191f, -1.f),
    C_Vertex3f(0.304793358f, 0.234820366f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.311751008f, 0.304547191f, -1.f),
    C_Vertex3f(0.301157713f, 0.298571825f, -1.f),
    C_Vertex3f(0.304793358f, 0.234820366f, -1.f),
    C_Vertex3f(0.289958239f, 0.293791533f, -1.f),
    C_Vertex3f(0.283536434f, 0.226052761f, -1.f),
    C_Vertex3f(0.278152585f, 0.290206313f, -1.f),
    C_Vertex3f(0.265740752f, 0.287816167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.261433125f, 0.220207691f, -1.f),
    C_Vertex3f(0.283536434f, 0.226052761f, -1.f),
    C_Vertex3f(0.265740752f, 0.287816167f, -1.f),
    C_Vertex3f(0.25272274f, 0.286621094f, -1.f),
    C_Vertex3f(0.239122391f, 0.286621094f, -1.f),
    C_Vertex3f(0.238483429f, 0.217285156f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.226095438f, 0.287816167f, -1.f),
    C_Vertex3f(0.21361804f, 0.290206313f, -1.f),
    C_Vertex3f(0.216653824f, 0.217285156f, -1.f),
    C_Vertex3f(0.238483429f, 0.217285156f, -1.f),
    C_Vertex3f(0.239122391f, 0.286621094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.201690197f, 0.293791533f, -1.f),
    C_Vertex3f(0.190311909f, 0.298571825f, -1.f),
    C_Vertex3f(0.197554469f, 0.219086528f, -1.f),
    C_Vertex3f(0.216653824f, 0.217285156f, -1.f),
    C_Vertex3f(0.21361804f, 0.290206313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.159474373f, 0.320083141f, -1.f),
    C_Vertex3f(0.161647081f, 0.228093386f, -1.f),
    C_Vertex3f(0.169203997f, 0.311717629f, -1.f),
    C_Vertex3f(0.179218888f, 0.222689271f, -1.f),
    C_Vertex3f(0.179483175f, 0.304547191f, -1.f),
    C_Vertex3f(0.197554469f, 0.219086528f, -1.f),
    C_Vertex3f(0.190311909f, 0.298571825f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150294304f, 0.329643726f, -1.f),
    C_Vertex3f(0.142261744f, 0.339921117f, -1.f),
    C_Vertex3f(0.144839048f, 0.235298872f, -1.f),
    C_Vertex3f(0.161647081f, 0.228093386f, -1.f),
    C_Vertex3f(0.159474373f, 0.320083141f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142261744f, 0.339921117f, -1.f),
    C_Vertex3f(0.135376692f, 0.350915313f, -1.f),
    C_Vertex3f(0.144839048f, 0.235298872f, -1.f),
    C_Vertex3f(0.129639149f, 0.362626314f, -1.f),
    C_Vertex3f(0.128794789f, 0.24430573f, -1.f),
    C_Vertex3f(0.125049114f, 0.375054121f, -1.f),
    C_Vertex3f(0.121606588f, 0.388198733f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.128794789f, 0.24430573f, -1.f),
    C_Vertex3f(0.121606588f, 0.388198733f, -1.f),
    C_Vertex3f(0.119311571f, 0.402060151f, -1.f),
    C_Vertex3f(0.118164063f, 0.416638374f, -1.f),
    C_Vertex3f(0.113514304f, 0.255113959f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.375416756f, 0.225355148f, -1.f),
    C_Vertex3f(0.370996475f, 0.201526642f, -1.f),
    C_Vertex3f(0.371032357f, 0.0425902605f, -1.f),
    C_Vertex3f(0.365102768f, 0.177913666f, -1.f),
    C_Vertex3f(0.358689308f, 0.0320057869f, -1.f),
    C_Vertex3f(0.358515739f, 0.156732559f, -1.f),
    C_Vertex3f(0.35123539f, 0.137983322f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.345534325f, 0.022506237f, -1.f),
    C_Vertex3f(0.358689308f, 0.0320057869f, -1.f),
    C_Vertex3f(0.35123539f, 0.137983322f, -1.f),
    C_Vertex3f(0.343261719f, 0.121665955f, -1.f),
    C_Vertex3f(0.334594727f, 0.107780457f, -1.f),
    C_Vertex3f(0.331812859f, 0.0141941309f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.30104351f, 0.0744867325f, -1.f),
    C_Vertex3f(0.302670479f, 0.00113224983f, -1.f),
    C_Vertex3f(0.313485622f, 0.0841908455f, -1.f),
    C_Vertex3f(0.31752491f, 0.0070694685f, -1.f),
    C_Vertex3f(0.324669361f, 0.0952887535f, -1.f),
    C_Vertex3f(0.331812859f, 0.0141941309f, -1.f),
    C_Vertex3f(0.334594727f, 0.107780457f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.201437473f, -0.0094370842f, -1.f),
    C_Vertex3f(0.219355583f, -0.0107421875f, -1.f),
    C_Vertex3f(0.219984055f, 0.0537109375f, -1.f),
    C_Vertex3f(0.237587929f, -0.0107421875f, -1.f),
    C_Vertex3f(0.239624023f, 0.0537109375f, -1.f),
    C_Vertex3f(0.25470829f, -0.00955474377f, -1.f),
    C_Vertex3f(0.256633282f, 0.055788517f, -1.f),
    C_Vertex3f(0.271262169f, -0.0071798563f, -1.f),
    C_Vertex3f(0.272539616f, 0.059943676f, -1.f),
    C_Vertex3f(0.287249565f, -0.0036175251f, -1.f),
    C_Vertex3f(0.287343025f, 0.0661764145f, -1.f),
    C_Vertex3f(0.302670479f, 0.00113224983f, -1.f),
    C_Vertex3f(0.30104351f, 0.0744867325f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.167939186f, 0.0717697144f, -1.f),
    C_Vertex3f(0.168180466f, -0.00291156769f, -1.f),
    C_Vertex3f(0.183377743f, 0.0627403259f, -1.f),
    C_Vertex3f(0.184379101f, -0.00682687759f, -1.f),
    C_Vertex3f(0.200726032f, 0.0567207336f, -1.f),
    C_Vertex3f(0.201437473f, -0.0094370842f, -1.f),
    C_Vertex3f(0.219984055f, 0.0537109375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.157792568f, 0.0807991028f, -1.f),
    C_Vertex3f(0.151573539f, 0.0877351761f, -1.f),
    C_Vertex3f(0.152841568f, 0.00230884552f, -1.f),
    C_Vertex3f(0.168180466f, -0.00291156769f, -1.f),
    C_Vertex3f(0.167939186f, 0.0717697144f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151573539f, 0.0877351761f, -1.f),
    C_Vertex3f(0.145899892f, 0.0955877304f, -1.f),
    C_Vertex3f(0.152841568f, 0.00230884552f, -1.f),
    C_Vertex3f(0.140771627f, 0.104356766f, -1.f),
    C_Vertex3f(0.138362408f, 0.00883436203f, -1.f),
    C_Vertex3f(0.136188745f, 0.114042282f, -1.f),
    C_Vertex3f(0.132151246f, 0.124644279f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.100083351f, 0.0362415314f, -1.f),
    C_Vertex3f(0.111983299f, 0.025800705f, -1.f),
    C_Vertex3f(0.124511719f, 0.155273438f, -1.f),
    C_Vertex3f(0.124742985f, 0.0166649818f, -1.f),
    C_Vertex3f(0.125712395f, 0.148597717f, -1.f),
    C_Vertex3f(0.138362408f, 0.00883436203f, -1.f),
    C_Vertex3f(0.128659129f, 0.136162758f, -1.f),
    C_Vertex3f(0.132151246f, 0.124644279f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.124511719f, 0.155273438f, -1.f),
    C_Vertex3f(0.048828125f, 0.1484375f, -1.f),
    C_Vertex3f(0.050028801f, 0.138980865f, -1.f),
    C_Vertex3f(0.0536174774f, 0.121126652f, -1.f),
    C_Vertex3f(0.0583934784f, 0.104331493f, -1.f),
    C_Vertex3f(0.0643568039f, 0.0885953903f, -1.f),
    C_Vertex3f(0.0715074539f, 0.0739183426f, -1.f),
    C_Vertex3f(0.0798454285f, 0.0603003502f, -1.f),
    C_Vertex3f(0.0893707275f, 0.0477414131f, -1.f),
    C_Vertex3f(0.100083351f, 0.0362415314f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0388284922f, 0.392675519f, -1.f),
    C_Vertex3f(0.037109375f, 0.43907547f, -1.f),
    C_Vertex3f(0.037109375f, 0.415111542f, -1.f),
    C_Vertex3f(0.119372845f, 0.449059963f, -1.f),
    C_Vertex3f(0.116962314f, 0.604091048f, -1.f),
    C_Vertex3f(0.118164063f, 0.432807922f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_58[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0810546875f, 0.375f, 0.f),
    C_Vertex3f(0.0810546875f, 0.375f, -1.f),
    C_Vertex3f(0.0810546875f, 0.464355469f, -1.f),
    C_Vertex3f(0.0810546875f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0810546875f, 0.464355469f, 0.f),
    C_Vertex3f(0.0810546875f, 0.464355469f, -1.f),
    C_Vertex3f(0.170410156f, 0.464355469f, -1.f),
    C_Vertex3f(0.170410156f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.170410156f, 0.464355469f, 0.f),
    C_Vertex3f(0.170410156f, 0.464355469f, -1.f),
    C_Vertex3f(0.170410156f, 0.375f, -1.f),
    C_Vertex3f(0.170410156f, 0.375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.170410156f, 0.375f, 0.f),
    C_Vertex3f(0.170410156f, 0.375f, -1.f),
    C_Vertex3f(0.0810546875f, 0.375f, -1.f),
    C_Vertex3f(0.0810546875f, 0.375f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0810546875f, 0.f, 0.f),
    C_Vertex3f(0.0810546875f, 0.f, -1.f),
    C_Vertex3f(0.0810546875f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0810546875f, 0.0893554688f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0810546875f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0810546875f, 0.0893554688f, -1.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, -1.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, 0.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, -1.f),
    C_Vertex3f(0.170410156f, 0.f, -1.f),
    C_Vertex3f(0.170410156f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.170410156f, 0.f, 0.f),
    C_Vertex3f(0.170410156f, 0.f, -1.f),
    C_Vertex3f(0.0810546875f, 0.f, -1.f),
    C_Vertex3f(0.0810546875f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0810546875f, 0.464355469f, 0.f),
    C_Vertex3f(0.0810546875f, 0.375f, 0.f),
    C_Vertex3f(0.170410156f, 0.375f, 0.f),
    C_Vertex3f(0.170410156f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0810546875f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0810546875f, 0.f, 0.f),
    C_Vertex3f(0.170410156f, 0.f, 0.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0810546875f, 0.464355469f, -1.f),
    C_Vertex3f(0.0810546875f, 0.375f, -1.f),
    C_Vertex3f(0.170410156f, 0.375f, -1.f),
    C_Vertex3f(0.170410156f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0810546875f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0810546875f, 0.f, -1.f),
    C_Vertex3f(0.170410156f, 0.f, -1.f),
    C_Vertex3f(0.170410156f, 0.0893554688f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_59[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0795898438f, 0.375f, 0.f),
    C_Vertex3f(0.0795898438f, 0.375f, -1.f),
    C_Vertex3f(0.0795898438f, 0.464355469f, -1.f),
    C_Vertex3f(0.0795898438f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0795898438f, 0.464355469f, 0.f),
    C_Vertex3f(0.0795898438f, 0.464355469f, -1.f),
    C_Vertex3f(0.168945313f, 0.464355469f, -1.f),
    C_Vertex3f(0.168945313f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.168945313f, 0.464355469f, 0.f),
    C_Vertex3f(0.168945313f, 0.464355469f, -1.f),
    C_Vertex3f(0.168945313f, 0.375f, -1.f),
    C_Vertex3f(0.168945313f, 0.375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.168945313f, 0.375f, 0.f),
    C_Vertex3f(0.168945313f, 0.375f, -1.f),
    C_Vertex3f(0.0795898438f, 0.375f, -1.f),
    C_Vertex3f(0.0795898438f, 0.375f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0795898438f, 0.f, 0.f),
    C_Vertex3f(0.0795898438f, 0.f, -1.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, -1.f),
    C_Vertex3f(0.168945313f, 0.0893554688f, -1.f),
    C_Vertex3f(0.168945313f, 0.0893554688f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.168945313f, 0.0893554688f, 0.f),
    C_Vertex3f(0.168945313f, 0.0893554688f, -1.f),
    C_Normal3f(0.999845922f, -0.017552482f, 0.f),
    C_Vertex3f(0.168945313f, -0.0123672485f, 0.f),
    C_Vertex3f(0.168945313f, -0.0123672485f, -1.f),
    C_Normal3f(0.988292277f, -0.152572513f, 0.f),
    C_Vertex3f(0.16676712f, -0.0347213745f, 0.f),
    C_Vertex3f(0.16676712f, -0.0347213745f, -1.f),
    C_Normal3f(0.960453153f, -0.278441697f, 0.f),
    C_Vertex3f(0.162410736f, -0.0546951294f, 0.f),
    C_Vertex3f(0.162410736f, -0.0546951294f, -1.f),
    C_Normal3f(0.906842053f, -0.421470672f, 0.f),
    C_Vertex3f(0.15587616f, -0.0722885132f, 0.f),
    C_Vertex3f(0.15587616f, -0.0722885132f, -1.f),
    C_Normal3f(0.820653856f, -0.571425498f, 0.f),
    C_Vertex3f(0.147163391f, -0.0875015259f, 0.f),
    C_Vertex3f(0.147163391f, -0.0875015259f, -1.f),
    C_Normal3f(0.709192693f, -0.705014646f, 0.f),
    C_Vertex3f(0.135890961f, -0.100990295f, 0.f),
    C_Vertex3f(0.135890961f, -0.100990295f, -1.f),
    C_Normal3f(0.585075676f, -0.810978711f, 0.f),
    C_Vertex3f(0.122058868f, -0.112754822f, 0.f),
    C_Vertex3f(0.122058868f, -0.112754822f, -1.f),
    C_Normal3f(0.481169224f, -0.876627743f, 0.f),
    C_Vertex3f(0.105667114f, -0.122795105f, 0.f),
    C_Vertex3f(0.105667114f, -0.122795105f, -1.f),
    C_Normal3f(0.401824236f, -0.915716887f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.837327302f, -0.546702027f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, -1.f),
    C_Vertex3f(0.07421875f, -0.0932998657f, -1.f),
    C_Vertex3f(0.07421875f, -0.0932998657f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.402259469f, 0.915525675f, 0.f),
    C_Vertex3f(0.07421875f, -0.0932998657f, 0.f),
    C_Vertex3f(0.07421875f, -0.0932998657f, -1.f),
    C_Normal3f(-0.486692309f, 0.873573482f, 0.f),
    C_Vertex3f(0.0804481506f, -0.0905628204f, 0.f),
    C_Vertex3f(0.0804481506f, -0.0905628204f, -1.f),
    C_Normal3f(-0.595117092f, 0.803638995f, 0.f),
    C_Vertex3f(0.0912671089f, -0.0838017464f, 0.f),
    C_Vertex3f(0.0912671089f, -0.0838017464f, -1.f),
    C_Normal3f(-0.720752895f, 0.693192124f, 0.f),
    C_Vertex3f(0.100446224f, -0.0757536888f, 0.f),
    C_Vertex3f(0.100446224f, -0.0757536888f, -1.f),
    C_Normal3f(-0.829470992f, 0.558549702f, 0.f),
    C_Vertex3f(0.107985497f, -0.0664186478f, 0.f),
    C_Vertex3f(0.107985497f, -0.0664186478f, -1.f),
    C_Normal3f(-0.913641572f, 0.406520873f, 0.f),
    C_Vertex3f(0.113884926f, -0.0557966232f, 0.f),
    C_Vertex3f(0.113884926f, -0.0557966232f, -1.f),
    C_Normal3f(-0.963291168f, 0.268458724f, 0.f),
    C_Vertex3f(0.11847353f, -0.0428471565f, 0.f),
    C_Vertex3f(0.11847353f, -0.0428471565f, -1.f),
    C_Normal3f(-0.987516761f, 0.157514051f, 0.f),
    C_Vertex3f(0.121751308f, -0.0275702477f, 0.f),
    C_Vertex3f(0.121751308f, -0.0275702477f, -1.f),
    C_Normal3f(-0.996553242f, 0.0829556137f, 0.f),
    C_Vertex3f(0.123718262f, -0.00996589661f, 0.f),
    C_Vertex3f(0.123718262f, -0.00996589661f, -1.f),
    C_Normal3f(-0.999458671f, 0.03290084f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, -1.f),
    C_Vertex3f(0.0795898438f, 0.f, -1.f),
    C_Vertex3f(0.0795898438f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0795898438f, 0.464355469f, 0.f),
    C_Vertex3f(0.0795898438f, 0.375f, 0.f),
    C_Vertex3f(0.168945313f, 0.375f, 0.f),
    C_Vertex3f(0.168945313f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.168945313f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, 0.f),
    C_Vertex3f(0.135890961f, -0.100990295f, 0.f),
    C_Vertex3f(0.147163391f, -0.0875015259f, 0.f),
    C_Vertex3f(0.15587616f, -0.0722885132f, 0.f),
    C_Vertex3f(0.162410736f, -0.0546951294f, 0.f),
    C_Vertex3f(0.16676712f, -0.0347213745f, 0.f),
    C_Vertex3f(0.168945313f, -0.0123672485f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123718262f, -0.00996589661f, 0.f),
    C_Vertex3f(0.121751308f, -0.0275702477f, 0.f),
    C_Vertex3f(0.122058868f, -0.112754822f, 0.f),
    C_Vertex3f(0.135890961f, -0.100990295f, 0.f),
    C_Vertex3f(0.124046326f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122058868f, -0.112754822f, 0.f),
    C_Vertex3f(0.121751308f, -0.0275702477f, 0.f),
    C_Vertex3f(0.11847353f, -0.0428471565f, 0.f),
    C_Vertex3f(0.113884926f, -0.0557966232f, 0.f),
    C_Vertex3f(0.107985497f, -0.0664186478f, 0.f),
    C_Vertex3f(0.105667114f, -0.122795105f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.100446224f, -0.0757536888f, 0.f),
    C_Vertex3f(0.0912671089f, -0.0838017464f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, 0.f),
    C_Vertex3f(0.105667114f, -0.122795105f, 0.f),
    C_Vertex3f(0.107985497f, -0.0664186478f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0804481506f, -0.0905628204f, 0.f),
    C_Vertex3f(0.07421875f, -0.0932998657f, 0.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, 0.f),
    C_Vertex3f(0.0912671089f, -0.0838017464f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.124046326f, 0.f, 0.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, 0.f),
    C_Vertex3f(0.0795898438f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0795898438f, 0.464355469f, -1.f),
    C_Vertex3f(0.0795898438f, 0.375f, -1.f),
    C_Vertex3f(0.168945313f, 0.375f, -1.f),
    C_Vertex3f(0.168945313f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.168945313f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, -1.f),
    C_Vertex3f(0.124046326f, 0.f, -1.f),
    C_Vertex3f(0.135890961f, -0.100990295f, -1.f),
    C_Vertex3f(0.147163391f, -0.0875015259f, -1.f),
    C_Vertex3f(0.15587616f, -0.0722885132f, -1.f),
    C_Vertex3f(0.162410736f, -0.0546951294f, -1.f),
    C_Vertex3f(0.16676712f, -0.0347213745f, -1.f),
    C_Vertex3f(0.168945313f, -0.0123672485f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123718262f, -0.00996589661f, -1.f),
    C_Vertex3f(0.121751308f, -0.0275702477f, -1.f),
    C_Vertex3f(0.122058868f, -0.112754822f, -1.f),
    C_Vertex3f(0.135890961f, -0.100990295f, -1.f),
    C_Vertex3f(0.124046326f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122058868f, -0.112754822f, -1.f),
    C_Vertex3f(0.121751308f, -0.0275702477f, -1.f),
    C_Vertex3f(0.11847353f, -0.0428471565f, -1.f),
    C_Vertex3f(0.113884926f, -0.0557966232f, -1.f),
    C_Vertex3f(0.107985497f, -0.0664186478f, -1.f),
    C_Vertex3f(0.105667114f, -0.122795105f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.100446224f, -0.0757536888f, -1.f),
    C_Vertex3f(0.0912671089f, -0.0838017464f, -1.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, -1.f),
    C_Vertex3f(0.105667114f, -0.122795105f, -1.f),
    C_Vertex3f(0.107985497f, -0.0664186478f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0804481506f, -0.0905628204f, -1.f),
    C_Vertex3f(0.07421875f, -0.0932998657f, -1.f),
    C_Vertex3f(0.0961914063f, -0.126953125f, -1.f),
    C_Vertex3f(0.0912671089f, -0.0838017464f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.124046326f, 0.f, -1.f),
    C_Vertex3f(0.0795898438f, 0.0893554688f, -1.f),
    C_Vertex3f(0.0795898438f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_60[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.048828125f, 0.280273438f, 0.f),
    C_Vertex3f(0.048828125f, 0.280273438f, -1.f),
    C_Vertex3f(0.048828125f, 0.354003906f, -1.f),
    C_Vertex3f(0.048828125f, 0.354003906f, 0.f),
    C_Normal3f(-0.389432311f, 0.921055138f, 0.f),
    C_Vertex3f(0.048828125f, 0.354003906f, 0.f),
    C_Vertex3f(0.048828125f, 0.354003906f, -1.f),
    C_Vertex3f(0.47265625f, 0.533203125f, -1.f),
    C_Vertex3f(0.47265625f, 0.533203125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.47265625f, 0.533203125f, 0.f),
    C_Vertex3f(0.47265625f, 0.533203125f, -1.f),
    C_Vertex3f(0.47265625f, 0.454589844f, -1.f),
    C_Vertex3f(0.47265625f, 0.454589844f, 0.f),
    C_Normal3f(0.379836679f, -0.925053537f, 0.f),
    C_Vertex3f(0.47265625f, 0.454589844f, 0.f),
    C_Vertex3f(0.47265625f, 0.454589844f, -1.f),
    C_Vertex3f(0.13671875f, 0.316650391f, -1.f),
    C_Vertex3f(0.13671875f, 0.316650391f, 0.f),
    C_Normal3f(0.382134259f, 0.924106777f, 0.f),
    C_Vertex3f(0.13671875f, 0.316650391f, 0.f),
    C_Vertex3f(0.13671875f, 0.316650391f, -1.f),
    C_Vertex3f(0.47265625f, 0.177734375f, -1.f),
    C_Vertex3f(0.47265625f, 0.177734375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.47265625f, 0.177734375f, 0.f),
    C_Vertex3f(0.47265625f, 0.177734375f, -1.f),
    C_Vertex3f(0.47265625f, 0.0991210938f, -1.f),
    C_Vertex3f(0.47265625f, 0.0991210938f, 0.f),
    C_Normal3f(-0.393024117f, -0.919528127f, 0.f),
    C_Vertex3f(0.47265625f, 0.0991210938f, 0.f),
    C_Vertex3f(0.47265625f, 0.0991210938f, -1.f),
    C_Vertex3f(0.048828125f, 0.280273438f, -1.f),
    C_Vertex3f(0.048828125f, 0.280273438f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.13671875f, 0.316650391f, 0.f),
    C_Vertex3f(0.47265625f, 0.454589844f, 0.f),
    C_Vertex3f(0.47265625f, 0.533203125f, 0.f),
    C_Vertex3f(0.048828125f, 0.354003906f, 0.f),
    C_Vertex3f(0.47265625f, 0.0991210938f, 0.f),
    C_Vertex3f(0.47265625f, 0.177734375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.47265625f, 0.0991210938f, 0.f),
    C_Vertex3f(0.048828125f, 0.354003906f, 0.f),
    C_Vertex3f(0.048828125f, 0.280273438f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.13671875f, 0.316650391f, -1.f),
    C_Vertex3f(0.47265625f, 0.454589844f, -1.f),
    C_Vertex3f(0.47265625f, 0.533203125f, -1.f),
    C_Vertex3f(0.048828125f, 0.354003906f, -1.f),
    C_Vertex3f(0.47265625f, 0.0991210938f, -1.f),
    C_Vertex3f(0.47265625f, 0.177734375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.47265625f, 0.0991210938f, -1.f),
    C_Vertex3f(0.048828125f, 0.354003906f, -1.f),
    C_Vertex3f(0.048828125f, 0.280273438f, -1.f),
    C_End(),
    C_Translatef(0.522460938f, 0.f, 0.f),
};
static const Call char_61[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.47265625f, 0.376953125f, 0.f),
    C_Vertex3f(0.47265625f, 0.376953125f, -1.f),
    C_Vertex3f(0.0498046875f, 0.376953125f, -1.f),
    C_Vertex3f(0.0498046875f, 0.376953125f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0498046875f, 0.376953125f, 0.f),
    C_Vertex3f(0.0498046875f, 0.376953125f, -1.f),
    C_Vertex3f(0.0498046875f, 0.450195313f, -1.f),
    C_Vertex3f(0.0498046875f, 0.450195313f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0498046875f, 0.450195313f, 0.f),
    C_Vertex3f(0.0498046875f, 0.450195313f, -1.f),
    C_Vertex3f(0.47265625f, 0.450195313f, -1.f),
    C_Vertex3f(0.47265625f, 0.450195313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.47265625f, 0.450195313f, 0.f),
    C_Vertex3f(0.47265625f, 0.450195313f, -1.f),
    C_Vertex3f(0.47265625f, 0.376953125f, -1.f),
    C_Vertex3f(0.47265625f, 0.376953125f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.47265625f, 0.182128906f, 0.f),
    C_Vertex3f(0.47265625f, 0.182128906f, -1.f),
    C_Vertex3f(0.0498046875f, 0.182128906f, -1.f),
    C_Vertex3f(0.0498046875f, 0.182128906f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0498046875f, 0.182128906f, 0.f),
    C_Vertex3f(0.0498046875f, 0.182128906f, -1.f),
    C_Vertex3f(0.0498046875f, 0.255371094f, -1.f),
    C_Vertex3f(0.0498046875f, 0.255371094f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0498046875f, 0.255371094f, 0.f),
    C_Vertex3f(0.0498046875f, 0.255371094f, -1.f),
    C_Vertex3f(0.47265625f, 0.255371094f, -1.f),
    C_Vertex3f(0.47265625f, 0.255371094f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.47265625f, 0.255371094f, 0.f),
    C_Vertex3f(0.47265625f, 0.255371094f, -1.f),
    C_Vertex3f(0.47265625f, 0.182128906f, -1.f),
    C_Vertex3f(0.47265625f, 0.182128906f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0498046875f, 0.450195313f, 0.f),
    C_Vertex3f(0.0498046875f, 0.376953125f, 0.f),
    C_Vertex3f(0.47265625f, 0.376953125f, 0.f),
    C_Vertex3f(0.47265625f, 0.450195313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0498046875f, 0.255371094f, 0.f),
    C_Vertex3f(0.0498046875f, 0.182128906f, 0.f),
    C_Vertex3f(0.47265625f, 0.182128906f, 0.f),
    C_Vertex3f(0.47265625f, 0.255371094f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0498046875f, 0.450195313f, -1.f),
    C_Vertex3f(0.0498046875f, 0.376953125f, -1.f),
    C_Vertex3f(0.47265625f, 0.376953125f, -1.f),
    C_Vertex3f(0.47265625f, 0.450195313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0498046875f, 0.255371094f, -1.f),
    C_Vertex3f(0.0498046875f, 0.182128906f, -1.f),
    C_Vertex3f(0.47265625f, 0.182128906f, -1.f),
    C_Vertex3f(0.47265625f, 0.255371094f, -1.f),
    C_End(),
    C_Translatef(0.522460938f, 0.f, 0.f),
};
static const Call char_62[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.393024117f, -0.919528127f, 0.f),
    C_Vertex3f(0.47265625f, 0.280273438f, 0.f),
    C_Vertex3f(0.47265625f, 0.280273438f, -1.f),
    C_Vertex3f(0.048828125f, 0.0991210938f, -1.f),
    C_Vertex3f(0.048828125f, 0.0991210938f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.048828125f, 0.0991210938f, 0.f),
    C_Vertex3f(0.048828125f, 0.0991210938f, -1.f),
    C_Vertex3f(0.048828125f, 0.177734375f, -1.f),
    C_Vertex3f(0.048828125f, 0.177734375f, 0.f),
    C_Normal3f(-0.382609159f, 0.92391032f, 0.f),
    C_Vertex3f(0.048828125f, 0.177734375f, 0.f),
    C_Vertex3f(0.048828125f, 0.177734375f, -1.f),
    C_Vertex3f(0.384277344f, 0.316650391f, -1.f),
    C_Vertex3f(0.384277344f, 0.316650391f, 0.f),
    C_Normal3f(-0.380309671f, -0.924859226f, 0.f),
    C_Vertex3f(0.384277344f, 0.316650391f, 0.f),
    C_Vertex3f(0.384277344f, 0.316650391f, -1.f),
    C_Vertex3f(0.048828125f, 0.454589844f, -1.f),
    C_Vertex3f(0.048828125f, 0.454589844f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.048828125f, 0.454589844f, 0.f),
    C_Vertex3f(0.048828125f, 0.454589844f, -1.f),
    C_Vertex3f(0.048828125f, 0.533203125f, -1.f),
    C_Vertex3f(0.048828125f, 0.533203125f, 0.f),
    C_Normal3f(0.389432311f, 0.921055138f, 0.f),
    C_Vertex3f(0.048828125f, 0.533203125f, 0.f),
    C_Vertex3f(0.048828125f, 0.533203125f, -1.f),
    C_Vertex3f(0.47265625f, 0.354003906f, -1.f),
    C_Vertex3f(0.47265625f, 0.354003906f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.47265625f, 0.354003906f, 0.f),
    C_Vertex3f(0.47265625f, 0.354003906f, -1.f),
    C_Vertex3f(0.47265625f, 0.280273438f, -1.f),
    C_Vertex3f(0.47265625f, 0.280273438f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.384277344f, 0.316650391f, 0.f),
    C_Vertex3f(0.048828125f, 0.177734375f, 0.f),
    C_Vertex3f(0.47265625f, 0.280273438f, 0.f),
    C_Vertex3f(0.47265625f, 0.354003906f, 0.f),
    C_Vertex3f(0.048828125f, 0.533203125f, 0.f),
    C_Vertex3f(0.048828125f, 0.454589844f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.47265625f, 0.280273438f, 0.f),
    C_Vertex3f(0.048828125f, 0.177734375f, 0.f),
    C_Vertex3f(0.048828125f, 0.0991210938f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.384277344f, 0.316650391f, -1.f),
    C_Vertex3f(0.048828125f, 0.177734375f, -1.f),
    C_Vertex3f(0.47265625f, 0.280273438f, -1.f),
    C_Vertex3f(0.47265625f, 0.354003906f, -1.f),
    C_Vertex3f(0.048828125f, 0.533203125f, -1.f),
    C_Vertex3f(0.048828125f, 0.454589844f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.47265625f, 0.280273438f, -1.f),
    C_Vertex3f(0.048828125f, 0.177734375f, -1.f),
    C_Vertex3f(0.048828125f, 0.0991210938f, -1.f),
    C_End(),
    C_Translatef(0.522460938f, 0.f, 0.f),
};
static const Call char_63[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999603152f, -0.028171517f, 0.f),
    C_Vertex3f(0.206054688f, 0.157714844f, 0.f),
    C_Vertex3f(0.206054688f, 0.157714844f, -1.f),
    C_Normal3f(-0.999671996f, -0.025609972f, 0.f),
    C_Vertex3f(0.205944061f, 0.161640167f, 0.f),
    C_Vertex3f(0.205944061f, 0.161640167f, -1.f),
    C_Normal3f(-0.999765217f, -0.0216690525f, 0.f),
    C_Vertex3f(0.205778122f, 0.168510437f, 0.f),
    C_Vertex3f(0.205778122f, 0.168510437f, -1.f),
    C_Normal3f(-0.999881983f, -0.0153637817f, 0.f),
    C_Vertex3f(0.205667496f, 0.17440033f, 0.f),
    C_Vertex3f(0.205667496f, 0.17440033f, -1.f),
    C_Normal3f(-0.99999553f, -0.00300050387f, 0.f),
    C_Vertex3f(0.205612183f, 0.179309845f, 0.f),
    C_Vertex3f(0.205612183f, 0.179309845f, -1.f),
    C_Normal3f(-0.998910069f, 0.0466757566f, 0.f),
    C_Vertex3f(0.205612183f, 0.192834854f, 0.f),
    C_Vertex3f(0.205612183f, 0.192834854f, -1.f),
    C_Normal3f(-0.99305743f, 0.117630437f, 0.f),
    C_Vertex3f(0.207250595f, 0.214373589f, 0.f),
    C_Vertex3f(0.207250595f, 0.214373589f, -1.f),
    C_Normal3f(-0.977915823f, 0.208998993f, 0.f),
    C_Vertex3f(0.21052742f, 0.234330177f, 0.f),
    C_Vertex3f(0.21052742f, 0.234330177f, -1.f),
    C_Normal3f(-0.956230164f, 0.292615771f, 0.f),
    C_Vertex3f(0.215442657f, 0.25270462f, 0.f),
    C_Vertex3f(0.215442657f, 0.25270462f, -1.f),
    C_Normal3f(-0.932974398f, 0.359942704f, 0.f),
    C_Vertex3f(0.218719482f, 0.261100769f, 0.f),
    C_Vertex3f(0.218719482f, 0.261100769f, -1.f),
    C_Normal3f(-0.909795761f, 0.415056139f, 0.f),
    C_Vertex3f(0.221122742f, 0.267427444f, 0.f),
    C_Vertex3f(0.221122742f, 0.267427444f, -1.f),
    C_Normal3f(-0.87583065f, 0.48261863f, 0.f),
    C_Vertex3f(0.227402687f, 0.280134201f, 0.f),
    C_Vertex3f(0.227402687f, 0.280134201f, -1.f),
    C_Normal3f(-0.833085954f, 0.553143442f, 0.f),
    C_Vertex3f(0.235156059f, 0.292894363f, 0.f),
    C_Vertex3f(0.235156059f, 0.292894363f, -1.f),
    C_Normal3f(-0.791139424f, 0.611635864f, 0.f),
    C_Vertex3f(0.244382858f, 0.305707932f, 0.f),
    C_Vertex3f(0.244382858f, 0.305707932f, -1.f),
    C_Normal3f(-0.74516499f, 0.666880071f, 0.f),
    C_Vertex3f(0.253662109f, 0.316831589f, 0.f),
    C_Vertex3f(0.253662109f, 0.316831589f, -1.f),
    C_Normal3f(-0.708310902f, 0.70590055f, 0.f),
    C_Vertex3f(0.264660358f, 0.328365803f, 0.f),
    C_Vertex3f(0.264660358f, 0.328365803f, -1.f),
    C_Normal3f(-0.684892952f, 0.728643656f, 0.f),
    C_Vertex3f(0.27879858f, 0.34205389f, 0.f),
    C_Vertex3f(0.27879858f, 0.34205389f, -1.f),
    C_Normal3f(-0.669042826f, 0.743223846f, 0.f),
    C_Vertex3f(0.296076775f, 0.357895851f, 0.f),
    C_Vertex3f(0.296076775f, 0.357895851f, -1.f),
    C_Normal3f(-0.666873932f, 0.745170534f, 0.f),
    C_Vertex3f(0.3113904f, 0.371392727f, 0.f),
    C_Vertex3f(0.3113904f, 0.371392727f, -1.f),
    C_Normal3f(-0.684084535f, 0.7294029f, 0.f),
    C_Vertex3f(0.320705533f, 0.379936814f, 0.f),
    C_Vertex3f(0.320705533f, 0.379936814f, -1.f),
    C_Normal3f(-0.702139735f, 0.712039113f, 0.f),
    C_Vertex3f(0.329126716f, 0.388027072f, 0.f),
    C_Vertex3f(0.329126716f, 0.388027072f, -1.f),
    C_Normal3f(-0.72298795f, 0.690860569f, 0.f),
    C_Vertex3f(0.336653948f, 0.3956635f, 0.f),
    C_Vertex3f(0.336653948f, 0.3956635f, -1.f),
    C_Normal3f(-0.747223794f, 0.664572418f, 0.f),
    C_Vertex3f(0.34328723f, 0.402846098f, 0.f),
    C_Vertex3f(0.34328723f, 0.402846098f, -1.f),
    C_Normal3f(-0.775553286f, 0.631282151f, 0.f),
    C_Vertex3f(0.349026561f, 0.409574866f, 0.f),
    C_Vertex3f(0.349026561f, 0.409574866f, -1.f),
    C_Normal3f(-0.808746099f, 0.588157952f, 0.f),
    C_Vertex3f(0.353871942f, 0.415849805f, 0.f),
    C_Vertex3f(0.353871942f, 0.415849805f, -1.f),
    C_Normal3f(-0.851630449f, 0.524142802f, 0.f),
    C_Vertex3f(0.357823372f, 0.421670914f, 0.f),
    C_Vertex3f(0.357823372f, 0.421670914f, -1.f),
    C_Normal3f(-0.903002501f, 0.429635316f, 0.f),
    C_Vertex3f(0.362409592f, 0.429721832f, 0.f),
    C_Vertex3f(0.362409592f, 0.429721832f, -1.f),
    C_Normal3f(-0.948754191f, 0.316014826f, 0.f),
    C_Vertex3f(0.366995811f, 0.44094944f, 0.f),
    C_Vertex3f(0.366995811f, 0.44094944f, -1.f),
    C_Normal3f(-0.982132018f, 0.188192725f, 0.f),
    C_Vertex3f(0.370053291f, 0.452670097f, 0.f),
    C_Vertex3f(0.370053291f, 0.452670097f, -1.f),
    C_Normal3f(-0.998029828f, 0.0627413541f, 0.f),
    C_Vertex3f(0.371582031f, 0.464883804f, 0.f),
    C_Vertex3f(0.371582031f, 0.464883804f, -1.f),
    C_Normal3f(-0.998849511f, -0.0479538701f, 0.f),
    C_Vertex3f(0.371582031f, 0.476987839f, 0.f),
    C_Vertex3f(0.371582031f, 0.476987839f, -1.f),
    C_Normal3f(-0.988561869f, -0.150815591f, 0.f),
    C_Vertex3f(0.370465517f, 0.488140106f, 0.f),
    C_Vertex3f(0.370465517f, 0.488140106f, -1.f),
    C_Normal3f(-0.967203021f, -0.254004717f, 0.f),
    C_Vertex3f(0.368232489f, 0.498943329f, 0.f),
    C_Vertex3f(0.368232489f, 0.498943329f, -1.f),
    C_Normal3f(-0.934737384f, -0.355339378f, 0.f),
    C_Vertex3f(0.364882946f, 0.509397507f, 0.f),
    C_Vertex3f(0.364882946f, 0.509397507f, -1.f),
    C_Normal3f(-0.892296255f, -0.451450229f, 0.f),
    C_Vertex3f(0.360416889f, 0.51950264f, 0.f),
    C_Vertex3f(0.360416889f, 0.51950264f, -1.f),
    C_Normal3f(-0.841927052f, -0.539591253f, 0.f),
    C_Vertex3f(0.354834318f, 0.529258728f, 0.f),
    C_Vertex3f(0.354834318f, 0.529258728f, -1.f),
    C_Normal3f(-0.786185205f, -0.61799103f, 0.f),
    C_Vertex3f(0.348135233f, 0.538665771f, 0.f),
    C_Vertex3f(0.348135233f, 0.538665771f, -1.f),
    C_Normal3f(-0.727670848f, -0.685926437f, 0.f),
    C_Vertex3f(0.340319633f, 0.54772377f, 0.f),
    C_Vertex3f(0.340319633f, 0.54772377f, -1.f),
    C_Normal3f(-0.664413273f, -0.747365475f, 0.f),
    C_Vertex3f(0.33138752f, 0.556432724f, 0.f),
    C_Vertex3f(0.33138752f, 0.556432724f, -1.f),
    C_Normal3f(-0.589825988f, -0.807530344f, 0.f),
    C_Vertex3f(0.321951628f, 0.564053059f, 0.f),
    C_Vertex3f(0.321951628f, 0.564053059f, -1.f),
    C_Normal3f(-0.506540477f, -0.862216115f, 0.f),
    C_Vertex3f(0.312011957f, 0.570584774f, 0.f),
    C_Vertex3f(0.312011957f, 0.570584774f, -1.f),
    C_Normal3f(-0.416426837f, -0.909169197f, 0.f),
    C_Vertex3f(0.301568508f, 0.57602787f, 0.f),
    C_Vertex3f(0.301568508f, 0.57602787f, -1.f),
    C_Normal3f(-0.322089702f, -0.946709096f, 0.f),
    C_Vertex3f(0.290621281f, 0.580382347f, 0.f),
    C_Vertex3f(0.290621281f, 0.580382347f, -1.f),
    C_Normal3f(-0.226509124f, -0.974009037f, 0.f),
    C_Vertex3f(0.279170275f, 0.583648205f, 0.f),
    C_Vertex3f(0.279170275f, 0.583648205f, -1.f),
    C_Normal3f(-0.132592306f, -0.991170585f, 0.f),
    C_Vertex3f(0.26721549f, 0.585825443f, 0.f),
    C_Vertex3f(0.26721549f, 0.585825443f, -1.f),
    C_Normal3f(-0.0431529954f, -0.999068379f, 0.f),
    C_Vertex3f(0.254756927f, 0.586914063f, 0.f),
    C_Vertex3f(0.254756927f, 0.586914063f, -1.f),
    C_Normal3f(0.0398654193f, -0.999204993f, 0.f),
    C_Vertex3f(0.242012024f, 0.586914063f, 0.f),
    C_Vertex3f(0.242012024f, 0.586914063f, -1.f),
    C_Normal3f(0.125079542f, -0.992146671f, 0.f),
    C_Vertex3f(0.230002046f, 0.585926414f, 0.f),
    C_Vertex3f(0.230002046f, 0.585926414f, -1.f),
    C_Normal3f(0.214670882f, -0.976686418f, 0.f),
    C_Vertex3f(0.218509555f, 0.583951116f, 0.f),
    C_Vertex3f(0.218509555f, 0.583951116f, -1.f),
    C_Normal3f(0.306995809f, -0.95171082f, 0.f),
    C_Vertex3f(0.207534552f, 0.580988169f, 0.f),
    C_Vertex3f(0.207534552f, 0.580988169f, -1.f),
    C_Normal3f(0.399494916f, -0.916735411f, 0.f),
    C_Vertex3f(0.197077036f, 0.577037573f, 0.f),
    C_Vertex3f(0.197077036f, 0.577037573f, -1.f),
    C_Normal3f(0.489326686f, -0.872100592f, 0.f),
    C_Vertex3f(0.187137008f, 0.572099328f, 0.f),
    C_Vertex3f(0.187137008f, 0.572099328f, -1.f),
    C_Normal3f(0.573765874f, -0.819019377f, 0.f),
    C_Vertex3f(0.177714467f, 0.566173434f, 0.f),
    C_Vertex3f(0.177714467f, 0.566173434f, -1.f),
    C_Normal3f(0.650601208f, -0.75941956f, 0.f),
    C_Vertex3f(0.168809414f, 0.559259892f, 0.f),
    C_Vertex3f(0.168809414f, 0.559259892f, -1.f),
    C_Normal3f(0.725305557f, -0.688427091f, 0.f),
    C_Vertex3f(0.160421848f, 0.5513587f, 0.f),
    C_Vertex3f(0.160421848f, 0.5513587f, -1.f),
    C_Normal3f(0.794816077f, -0.606850445f, 0.f),
    C_Vertex3f(0.152755857f, 0.542346358f, 0.f),
    C_Vertex3f(0.152755857f, 0.542346358f, -1.f),
    C_Normal3f(0.851234734f, -0.524785161f, 0.f),
    C_Vertex3f(0.145811439f, 0.532222867f, 0.f),
    C_Vertex3f(0.145811439f, 0.532222867f, -1.f),
    C_Normal3f(0.895429015f, -0.445204407f, 0.f),
    C_Vertex3f(0.139588594f, 0.520988226f, 0.f),
    C_Vertex3f(0.139588594f, 0.520988226f, -1.f),
    C_Normal3f(0.928974032f, -0.370144963f, 0.f),
    C_Vertex3f(0.134087324f, 0.508642435f, 0.f),
    C_Vertex3f(0.134087324f, 0.508642435f, -1.f),
    C_Normal3f(0.953701735f, -0.3007541f, 0.f),
    C_Vertex3f(0.129307628f, 0.495185494f, 0.f),
    C_Vertex3f(0.129307628f, 0.495185494f, -1.f),
    C_Normal3f(0.971392512f, -0.237479746f, 0.f),
    C_Vertex3f(0.125249505f, 0.480617404f, 0.f),
    C_Vertex3f(0.125249505f, 0.480617404f, -1.f),
    C_Normal3f(0.981897891f, -0.189411268f, 0.f),
    C_Vertex3f(0.121912956f, 0.464938164f, 0.f),
    C_Vertex3f(0.121912956f, 0.464938164f, -1.f),
    C_Normal3f(0.988088369f, -0.153887212f, 0.f),
    C_Vertex3f(0.120605469f, 0.456542969f, 0.f),
    C_Vertex3f(0.120605469f, 0.456542969f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.125505731f, -0.992092907f, 0.f),
    C_Vertex3f(0.120605469f, 0.456542969f, 0.f),
    C_Vertex3f(0.120605469f, 0.456542969f, -1.f),
    C_Vertex3f(0.0395507813f, 0.466796875f, -1.f),
    C_Vertex3f(0.0395507813f, 0.466796875f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.992633283f, 0.121158153f, 0.f),
    C_Vertex3f(0.0395507813f, 0.466796875f, 0.f),
    C_Vertex3f(0.0395507813f, 0.466796875f, -1.f),
    C_Normal3f(-0.98606503f, 0.16636081f, 0.f),
    C_Vertex3f(0.0409154892f, 0.477977753f, 0.f),
    C_Vertex3f(0.0409154892f, 0.477977753f, -1.f),
    C_Normal3f(-0.973989725f, 0.226592034f, 0.f),
    C_Vertex3f(0.0449889898f, 0.49903059f, 0.f),
    C_Vertex3f(0.0449889898f, 0.49903059f, -1.f),
    C_Normal3f(-0.952699065f, 0.303915471f, 0.f),
    C_Vertex3f(0.0504065752f, 0.518774509f, 0.f),
    C_Vertex3f(0.0504065752f, 0.518774509f, -1.f),
    C_Normal3f(-0.922612906f, 0.385727227f, 0.f),
    C_Vertex3f(0.0571682453f, 0.537209511f, 0.f),
    C_Vertex3f(0.0571682453f, 0.537209511f, -1.f),
    C_Normal3f(-0.882510185f, 0.470293373f, 0.f),
    C_Vertex3f(0.0652740002f, 0.554335594f, 0.f),
    C_Vertex3f(0.0652740002f, 0.554335594f, -1.f),
    C_Normal3f(-0.831710637f, 0.555209398f, 0.f),
    C_Vertex3f(0.0747238398f, 0.57015276f, 0.f),
    C_Vertex3f(0.0747238398f, 0.57015276f, -1.f),
    C_Normal3f(-0.770373464f, 0.637592852f, 0.f),
    C_Vertex3f(0.0855177641f, 0.584661007f, 0.f),
    C_Vertex3f(0.0855177641f, 0.584661007f, -1.f),
    C_Normal3f(-0.699673057f, 0.714463174f, 0.f),
    C_Vertex3f(0.0976557732f, 0.597860336f, 0.f),
    C_Vertex3f(0.0976557732f, 0.597860336f, -1.f),
    C_Normal3f(-0.622231483f, 0.782833219f, 0.f),
    C_Vertex3f(0.111137867f, 0.609750748f, 0.f),
    C_Vertex3f(0.111137867f, 0.609750748f, -1.f),
    C_Normal3f(-0.538702428f, 0.842496157f, 0.f),
    C_Vertex3f(0.125704646f, 0.620154858f, 0.f),
    C_Vertex3f(0.125704646f, 0.620154858f, -1.f),
    C_Normal3f(-0.450639337f, 0.892706037f, 0.f),
    C_Vertex3f(0.141356111f, 0.629072666f, 0.f),
    C_Vertex3f(0.141356111f, 0.629072666f, -1.f),
    C_Normal3f(-0.360989749f, 0.932569742f, 0.f),
    C_Vertex3f(0.15809226f, 0.636504173f, 0.f),
    C_Vertex3f(0.15809226f, 0.636504173f, -1.f),
    C_Normal3f(-0.27256155f, 0.962138295f, 0.f),
    C_Vertex3f(0.175913095f, 0.642449379f, 0.f),
    C_Vertex3f(0.175913095f, 0.642449379f, -1.f),
    C_Normal3f(-0.187667623f, 0.982232571f, 0.f),
    C_Vertex3f(0.194818616f, 0.646908283f, 0.f),
    C_Vertex3f(0.194818616f, 0.646908283f, -1.f),
    C_Normal3f(-0.107946888f, 0.994156659f, 0.f),
    C_Vertex3f(0.214808822f, 0.649880886f, 0.f),
    C_Vertex3f(0.214808822f, 0.649880886f, -1.f),
    C_Normal3f(-0.0338443778f, 0.99942708f, 0.f),
    C_Vertex3f(0.235883713f, 0.651367188f, 0.f),
    C_Vertex3f(0.235883713f, 0.651367188f, -1.f),
    C_Normal3f(0.0354284644f, 0.999372184f, 0.f),
    C_Vertex3f(0.258699417f, 0.651367188f, 0.f),
    C_Vertex3f(0.258699417f, 0.651367188f, -1.f),
    C_Normal3f(0.109895512f, 0.993943155f, 0.f),
    C_Vertex3f(0.280983925f, 0.649768353f, 0.f),
    C_Vertex3f(0.280983925f, 0.649768353f, -1.f),
    C_Normal3f(0.191344365f, 0.981522977f, 0.f),
    C_Vertex3f(0.302081108f, 0.646570683f, 0.f),
    C_Vertex3f(0.302081108f, 0.646570683f, -1.f),
    C_Normal3f(0.278259575f, 0.960505962f, 0.f),
    C_Vertex3f(0.321990967f, 0.641774178f, 0.f),
    C_Vertex3f(0.321990967f, 0.641774178f, -1.f),
    C_Normal3f(0.368879229f, 0.929477394f, 0.f),
    C_Vertex3f(0.340713501f, 0.635378838f, 0.f),
    C_Vertex3f(0.340713501f, 0.635378838f, -1.f),
    C_Normal3f(0.460692316f, 0.88755995f, 0.f),
    C_Vertex3f(0.358248711f, 0.627384663f, 0.f),
    C_Vertex3f(0.358248711f, 0.627384663f, -1.f),
    C_Normal3f(0.550644338f, 0.834739923f, 0.f),
    C_Vertex3f(0.374596596f, 0.617791653f, 0.f),
    C_Vertex3f(0.374596596f, 0.617791653f, -1.f),
    C_Normal3f(0.635549128f, 0.772060454f, 0.f),
    C_Vertex3f(0.389757156f, 0.606599808f, 0.f),
    C_Vertex3f(0.389757156f, 0.606599808f, -1.f),
    C_Normal3f(0.707797468f, 0.706415355f, 0.f),
    C_Vertex3f(0.403730392f, 0.593809128f, 0.f),
    C_Vertex3f(0.403730392f, 0.593809128f, -1.f),
    C_Normal3f(0.772137403f, 0.635455549f, 0.f),
    C_Vertex3f(0.415956974f, 0.58034873f, 0.f),
    C_Vertex3f(0.415956974f, 0.58034873f, -1.f),
    C_Normal3f(0.833024919f, 0.553235471f, 0.f),
    C_Vertex3f(0.426436901f, 0.566218615f, 0.f),
    C_Vertex3f(0.426436901f, 0.566218615f, -1.f),
    C_Normal3f(0.88745904f, 0.460886776f, 0.f),
    C_Vertex3f(0.435170174f, 0.551418781f, 0.f),
    C_Vertex3f(0.435170174f, 0.551418781f, -1.f),
    C_Normal3f(0.932658434f, 0.360760838f, 0.f),
    C_Vertex3f(0.442156792f, 0.53594923f, 0.f),
    C_Vertex3f(0.442156792f, 0.53594923f, -1.f),
    C_Normal3f(0.966620386f, 0.256212622f, 0.f),
    C_Vertex3f(0.447396755f, 0.519809961f, 0.f),
    C_Vertex3f(0.447396755f, 0.519809961f, -1.f),
    C_Normal3f(0.988523066f, 0.151069522f, 0.f),
    C_Vertex3f(0.450890064f, 0.503000975f, 0.f),
    C_Vertex3f(0.450890064f, 0.503000975f, -1.f),
    C_Normal3f(0.998890519f, 0.0470920689f, 0.f),
    C_Vertex3f(0.452636719f, 0.48552227f, 0.f),
    C_Vertex3f(0.452636719f, 0.48552227f, -1.f),
    C_Normal3f(0.998091877f, -0.0617452972f, 0.f),
    C_Vertex3f(0.452636719f, 0.46595192f, 0.f),
    C_Vertex3f(0.452636719f, 0.46595192f, -1.f),
    C_Normal3f(0.983028471f, -0.183452785f, 0.f),
    C_Vertex3f(0.450178146f, 0.445780277f, 0.f),
    C_Vertex3f(0.450178146f, 0.445780277f, -1.f),
    C_Normal3f(0.95117116f, -0.308663964f, 0.f),
    C_Vertex3f(0.445261002f, 0.426429272f, 0.f),
    C_Vertex3f(0.445261002f, 0.426429272f, -1.f),
    C_Normal3f(0.907212973f, -0.420671791f, 0.f),
    C_Vertex3f(0.437885284f, 0.407898903f, 0.f),
    C_Vertex3f(0.437885284f, 0.407898903f, -1.f),
    C_Normal3f(0.859156251f, -0.511713445f, 0.f),
    C_Vertex3f(0.430509567f, 0.394616604f, 0.f),
    C_Vertex3f(0.430509567f, 0.394616604f, -1.f),
    C_Normal3f(0.81993717f, -0.572453558f, 0.f),
    C_Vertex3f(0.424417734f, 0.385287166f, 0.f),
    C_Vertex3f(0.424417734f, 0.385287166f, -1.f),
    C_Normal3f(0.787677884f, -0.616087377f, 0.f),
    C_Vertex3f(0.417151213f, 0.375483155f, 0.f),
    C_Vertex3f(0.417151213f, 0.375483155f, -1.f),
    C_Normal3f(0.758723319f, -0.651413083f, 0.f),
    C_Vertex3f(0.408710003f, 0.365204573f, 0.f),
    C_Vertex3f(0.408710003f, 0.365204573f, -1.f),
    C_Normal3f(0.732865512f, -0.68037343f, 0.f),
    C_Vertex3f(0.399094105f, 0.354451418f, 0.f),
    C_Vertex3f(0.399094105f, 0.354451418f, -1.f),
    C_Normal3f(0.709797263f, -0.704405963f, 0.f),
    C_Vertex3f(0.388303518f, 0.343223691f, 0.f),
    C_Vertex3f(0.388303518f, 0.343223691f, -1.f),
    C_Normal3f(0.689192295f, -0.724578619f, 0.f),
    C_Vertex3f(0.376338243f, 0.331521392f, 0.f),
    C_Vertex3f(0.376338243f, 0.331521392f, -1.f),
    C_Normal3f(0.671597004f, -0.74091655f, 0.f),
    C_Vertex3f(0.36319828f, 0.319344521f, 0.f),
    C_Vertex3f(0.36319828f, 0.319344521f, -1.f),
    C_Normal3f(0.666077137f, -0.745882928f, 0.f),
    C_Vertex3f(0.351233482f, 0.308765411f, 0.f),
    C_Vertex3f(0.351233482f, 0.308765411f, -1.f),
    C_Normal3f(0.675726533f, -0.737152398f, 0.f),
    C_Vertex3f(0.342451572f, 0.300817609f, 0.f),
    C_Vertex3f(0.342451572f, 0.300817609f, -1.f),
    C_Normal3f(0.686391294f, -0.727232337f, 0.f),
    C_Vertex3f(0.334502697f, 0.293428779f, 0.f),
    C_Vertex3f(0.334502697f, 0.293428779f, -1.f),
    C_Normal3f(0.699113369f, -0.715010822f, 0.f),
    C_Vertex3f(0.327386856f, 0.286598921f, 0.f),
    C_Vertex3f(0.327386856f, 0.286598921f, -1.f),
    C_Normal3f(0.714527726f, -0.699607193f, 0.f),
    C_Vertex3f(0.32110405f, 0.280328035f, 0.f),
    C_Vertex3f(0.32110405f, 0.280328035f, -1.f),
    C_Normal3f(0.733543277f, -0.679642797f, 0.f),
    C_Vertex3f(0.315654278f, 0.274616122f, 0.f),
    C_Vertex3f(0.315654278f, 0.274616122f, -1.f),
    C_Normal3f(0.757489383f, -0.652847409f, 0.f),
    C_Vertex3f(0.31103754f, 0.269463181f, 0.f),
    C_Vertex3f(0.31103754f, 0.269463181f, -1.f),
    C_Normal3f(0.796097815f, -0.605167985f, 0.f),
    C_Vertex3f(0.307253838f, 0.264869213f, 0.f),
    C_Vertex3f(0.307253838f, 0.264869213f, -1.f),
    C_Normal3f(0.871388972f, -0.490592599f, 0.f),
    C_Vertex3f(0.299877167f, 0.254781723f, 0.f),
    C_Vertex3f(0.299877167f, 0.254781723f, -1.f),
    C_Normal3f(0.934900761f, -0.354909271f, 0.f),
    C_Vertex3f(0.291135788f, 0.236240387f, 0.f),
    C_Vertex3f(0.291135788f, 0.236240387f, -1.f),
    C_Normal3f(0.976498902f, -0.215522289f, 0.f),
    C_Vertex3f(0.286875725f, 0.220533371f, 0.f),
    C_Vertex3f(0.286875725f, 0.220533371f, -1.f),
    C_Normal3f(0.992140055f, -0.12513271f, 0.f),
    C_Vertex3f(0.284636021f, 0.206790924f, 0.f),
    C_Vertex3f(0.284636021f, 0.206790924f, -1.f),
    C_Normal3f(0.997431517f, -0.0716258362f, 0.f),
    C_Vertex3f(0.282996655f, 0.189777374f, 0.f),
    C_Vertex3f(0.282996655f, 0.189777374f, -1.f),
    C_Normal3f(0.999230742f, -0.0392172821f, 0.f),
    C_Vertex3f(0.281957626f, 0.169492722f, 0.f),
    C_Vertex3f(0.281957626f, 0.169492722f, -1.f),
    C_Normal3f(0.99982667f, -0.0186202526f, 0.f),
    C_Vertex3f(0.281738281f, 0.157714844f, 0.f),
    C_Vertex3f(0.281738281f, 0.157714844f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.281738281f, 0.157714844f, 0.f),
    C_Vertex3f(0.281738281f, 0.157714844f, -1.f),
    C_Vertex3f(0.206054688f, 0.157714844f, -1.f),
    C_Vertex3f(0.206054688f, 0.157714844f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.192382813f, 0.f, 0.f),
    C_Vertex3f(0.192382813f, 0.f, -1.f),
    C_Vertex3f(0.192382813f, 0.0893554688f, -1.f),
    C_Vertex3f(0.192382813f, 0.0893554688f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.192382813f, 0.0893554688f, 0.f),
    C_Vertex3f(0.192382813f, 0.0893554688f, -1.f),
    C_Vertex3f(0.281738281f, 0.0893554688f, -1.f),
    C_Vertex3f(0.281738281f, 0.0893554688f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.281738281f, 0.0893554688f, 0.f),
    C_Vertex3f(0.281738281f, 0.0893554688f, -1.f),
    C_Vertex3f(0.281738281f, 0.f, -1.f),
    C_Vertex3f(0.281738281f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.281738281f, 0.f, 0.f),
    C_Vertex3f(0.281738281f, 0.f, -1.f),
    C_Vertex3f(0.192382813f, 0.f, -1.f),
    C_Vertex3f(0.192382813f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.374596596f, 0.617791653f, 0.f),
    C_Vertex3f(0.358248711f, 0.627384663f, 0.f),
    C_Vertex3f(0.360416889f, 0.51950264f, 0.f),
    C_Vertex3f(0.364882946f, 0.509397507f, 0.f),
    C_Vertex3f(0.368232489f, 0.498943329f, 0.f),
    C_Vertex3f(0.370465517f, 0.488140106f, 0.f),
    C_Vertex3f(0.371582031f, 0.476987839f, 0.f),
    C_Vertex3f(0.376338243f, 0.331521392f, 0.f),
    C_Vertex3f(0.389757156f, 0.606599808f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.33138752f, 0.556432724f, 0.f),
    C_Vertex3f(0.340319633f, 0.54772377f, 0.f),
    C_Vertex3f(0.340713501f, 0.635378838f, 0.f),
    C_Vertex3f(0.348135233f, 0.538665771f, 0.f),
    C_Vertex3f(0.358248711f, 0.627384663f, 0.f),
    C_Vertex3f(0.354834318f, 0.529258728f, 0.f),
    C_Vertex3f(0.360416889f, 0.51950264f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.321990967f, 0.641774178f, 0.f),
    C_Vertex3f(0.302081108f, 0.646570683f, 0.f),
    C_Vertex3f(0.312011957f, 0.570584774f, 0.f),
    C_Vertex3f(0.321951628f, 0.564053059f, 0.f),
    C_Vertex3f(0.33138752f, 0.556432724f, 0.f),
    C_Vertex3f(0.340713501f, 0.635378838f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.26721549f, 0.585825443f, 0.f),
    C_Vertex3f(0.279170275f, 0.583648205f, 0.f),
    C_Vertex3f(0.280983925f, 0.649768353f, 0.f),
    C_Vertex3f(0.290621281f, 0.580382347f, 0.f),
    C_Vertex3f(0.302081108f, 0.646570683f, 0.f),
    C_Vertex3f(0.301568508f, 0.57602787f, 0.f),
    C_Vertex3f(0.312011957f, 0.570584774f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.258699417f, 0.651367188f, 0.f),
    C_Vertex3f(0.235883713f, 0.651367188f, 0.f),
    C_Vertex3f(0.242012024f, 0.586914063f, 0.f),
    C_Vertex3f(0.254756927f, 0.586914063f, 0.f),
    C_Vertex3f(0.26721549f, 0.585825443f, 0.f),
    C_Vertex3f(0.280983925f, 0.649768353f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.197077036f, 0.577037573f, 0.f),
    C_Vertex3f(0.207534552f, 0.580988169f, 0.f),
    C_Vertex3f(0.214808822f, 0.649880886f, 0.f),
    C_Vertex3f(0.218509555f, 0.583951116f, 0.f),
    C_Vertex3f(0.235883713f, 0.651367188f, 0.f),
    C_Vertex3f(0.230002046f, 0.585926414f, 0.f),
    C_Vertex3f(0.242012024f, 0.586914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.194818616f, 0.646908283f, 0.f),
    C_Vertex3f(0.175913095f, 0.642449379f, 0.f),
    C_Vertex3f(0.177714467f, 0.566173434f, 0.f),
    C_Vertex3f(0.187137008f, 0.572099328f, 0.f),
    C_Vertex3f(0.197077036f, 0.577037573f, 0.f),
    C_Vertex3f(0.214808822f, 0.649880886f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145811439f, 0.532222867f, 0.f),
    C_Vertex3f(0.152755857f, 0.542346358f, 0.f),
    C_Vertex3f(0.15809226f, 0.636504173f, 0.f),
    C_Vertex3f(0.160421848f, 0.5513587f, 0.f),
    C_Vertex3f(0.175913095f, 0.642449379f, 0.f),
    C_Vertex3f(0.168809414f, 0.559259892f, 0.f),
    C_Vertex3f(0.177714467f, 0.566173434f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.141356111f, 0.629072666f, 0.f),
    C_Vertex3f(0.125704646f, 0.620154858f, 0.f),
    C_Vertex3f(0.129307628f, 0.495185494f, 0.f),
    C_Vertex3f(0.134087324f, 0.508642435f, 0.f),
    C_Vertex3f(0.139588594f, 0.520988226f, 0.f),
    C_Vertex3f(0.145811439f, 0.532222867f, 0.f),
    C_Vertex3f(0.15809226f, 0.636504173f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.125704646f, 0.620154858f, 0.f),
    C_Vertex3f(0.111137867f, 0.609750748f, 0.f),
    C_Vertex3f(0.120605469f, 0.456542969f, 0.f),
    C_Vertex3f(0.121912956f, 0.464938164f, 0.f),
    C_Vertex3f(0.125249505f, 0.480617404f, 0.f),
    C_Vertex3f(0.129307628f, 0.495185494f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120605469f, 0.456542969f, 0.f),
    C_Vertex3f(0.111137867f, 0.609750748f, 0.f),
    C_Vertex3f(0.0976557732f, 0.597860336f, 0.f),
    C_Vertex3f(0.0855177641f, 0.584661007f, 0.f),
    C_Vertex3f(0.0747238398f, 0.57015276f, 0.f),
    C_Vertex3f(0.0652740002f, 0.554335594f, 0.f),
    C_Vertex3f(0.0571682453f, 0.537209511f, 0.f),
    C_Vertex3f(0.0504065752f, 0.518774509f, 0.f),
    C_Vertex3f(0.0449889898f, 0.49903059f, 0.f),
    C_Vertex3f(0.0409154892f, 0.477977753f, 0.f),
    C_Vertex3f(0.0395507813f, 0.466796875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.417151213f, 0.375483155f, 0.f),
    C_Vertex3f(0.424417734f, 0.385287166f, 0.f),
    C_Vertex3f(0.426436901f, 0.566218615f, 0.f),
    C_Vertex3f(0.430509567f, 0.394616604f, 0.f),
    C_Vertex3f(0.435170174f, 0.551418781f, 0.f),
    C_Vertex3f(0.437885284f, 0.407898903f, 0.f),
    C_Vertex3f(0.442156792f, 0.53594923f, 0.f),
    C_Vertex3f(0.445261002f, 0.426429272f, 0.f),
    C_Vertex3f(0.447396755f, 0.519809961f, 0.f),
    C_Vertex3f(0.450178146f, 0.445780277f, 0.f),
    C_Vertex3f(0.450890064f, 0.503000975f, 0.f),
    C_Vertex3f(0.452636719f, 0.46595192f, 0.f),
    C_Vertex3f(0.452636719f, 0.48552227f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.376338243f, 0.331521392f, 0.f),
    C_Vertex3f(0.388303518f, 0.343223691f, 0.f),
    C_Vertex3f(0.389757156f, 0.606599808f, 0.f),
    C_Vertex3f(0.399094105f, 0.354451418f, 0.f),
    C_Vertex3f(0.403730392f, 0.593809128f, 0.f),
    C_Vertex3f(0.408710003f, 0.365204573f, 0.f),
    C_Vertex3f(0.415956974f, 0.58034873f, 0.f),
    C_Vertex3f(0.417151213f, 0.375483155f, 0.f),
    C_Vertex3f(0.426436901f, 0.566218615f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376338243f, 0.331521392f, 0.f),
    C_Vertex3f(0.371582031f, 0.476987839f, 0.f),
    C_Vertex3f(0.371582031f, 0.464883804f, 0.f),
    C_Vertex3f(0.370053291f, 0.452670097f, 0.f),
    C_Vertex3f(0.366995811f, 0.44094944f, 0.f),
    C_Vertex3f(0.36319828f, 0.319344521f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36319828f, 0.319344521f, 0.f),
    C_Vertex3f(0.366995811f, 0.44094944f, 0.f),
    C_Vertex3f(0.362409592f, 0.429721832f, 0.f),
    C_Vertex3f(0.357823372f, 0.421670914f, 0.f),
    C_Vertex3f(0.353871942f, 0.415849805f, 0.f),
    C_Vertex3f(0.351233482f, 0.308765411f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.32110405f, 0.280328035f, 0.f),
    C_Vertex3f(0.327386856f, 0.286598921f, 0.f),
    C_Vertex3f(0.329126716f, 0.388027072f, 0.f),
    C_Vertex3f(0.334502697f, 0.293428779f, 0.f),
    C_Vertex3f(0.336653948f, 0.3956635f, 0.f),
    C_Vertex3f(0.342451572f, 0.300817609f, 0.f),
    C_Vertex3f(0.34328723f, 0.402846098f, 0.f),
    C_Vertex3f(0.351233482f, 0.308765411f, 0.f),
    C_Vertex3f(0.349026561f, 0.409574866f, 0.f),
    C_Vertex3f(0.353871942f, 0.415849805f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.307253838f, 0.264869213f, 0.f),
    C_Vertex3f(0.31103754f, 0.269463181f, 0.f),
    C_Vertex3f(0.3113904f, 0.371392727f, 0.f),
    C_Vertex3f(0.315654278f, 0.274616122f, 0.f),
    C_Vertex3f(0.320705533f, 0.379936814f, 0.f),
    C_Vertex3f(0.32110405f, 0.280328035f, 0.f),
    C_Vertex3f(0.329126716f, 0.388027072f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.286875725f, 0.220533371f, 0.f),
    C_Vertex3f(0.291135788f, 0.236240387f, 0.f),
    C_Vertex3f(0.296076775f, 0.357895851f, 0.f),
    C_Vertex3f(0.299877167f, 0.254781723f, 0.f),
    C_Vertex3f(0.3113904f, 0.371392727f, 0.f),
    C_Vertex3f(0.307253838f, 0.264869213f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.281738281f, 0.157714844f, 0.f),
    C_Vertex3f(0.281957626f, 0.169492722f, 0.f),
    C_Vertex3f(0.27879858f, 0.34205389f, 0.f),
    C_Vertex3f(0.282996655f, 0.189777374f, 0.f),
    C_Vertex3f(0.296076775f, 0.357895851f, 0.f),
    C_Vertex3f(0.284636021f, 0.206790924f, 0.f),
    C_Vertex3f(0.286875725f, 0.220533371f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.281738281f, 0.157714844f, 0.f),
    C_Vertex3f(0.27879858f, 0.34205389f, 0.f),
    C_Vertex3f(0.264660358f, 0.328365803f, 0.f),
    C_Vertex3f(0.253662109f, 0.316831589f, 0.f),
    C_Vertex3f(0.244382858f, 0.305707932f, 0.f),
    C_Vertex3f(0.235156059f, 0.292894363f, 0.f),
    C_Vertex3f(0.227402687f, 0.280134201f, 0.f),
    C_Vertex3f(0.221122742f, 0.267427444f, 0.f),
    C_Vertex3f(0.218719482f, 0.261100769f, 0.f),
    C_Vertex3f(0.215442657f, 0.25270462f, 0.f),
    C_Vertex3f(0.21052742f, 0.234330177f, 0.f),
    C_Vertex3f(0.207250595f, 0.214373589f, 0.f),
    C_Vertex3f(0.206054688f, 0.157714844f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.207250595f, 0.214373589f, 0.f),
    C_Vertex3f(0.205612183f, 0.192834854f, 0.f),
    C_Vertex3f(0.205667496f, 0.17440033f, 0.f),
    C_Vertex3f(0.205778122f, 0.168510437f, 0.f),
    C_Vertex3f(0.205944061f, 0.161640167f, 0.f),
    C_Vertex3f(0.206054688f, 0.157714844f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.192382813f, 0.0893554688f, 0.f),
    C_Vertex3f(0.192382813f, 0.f, 0.f),
    C_Vertex3f(0.281738281f, 0.f, 0.f),
    C_Vertex3f(0.281738281f, 0.0893554688f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.205667496f, 0.17440033f, 0.f),
    C_Vertex3f(0.205612183f, 0.192834854f, 0.f),
    C_Vertex3f(0.205612183f, 0.179309845f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.374596596f, 0.617791653f, -1.f),
    C_Vertex3f(0.358248711f, 0.627384663f, -1.f),
    C_Vertex3f(0.360416889f, 0.51950264f, -1.f),
    C_Vertex3f(0.364882946f, 0.509397507f, -1.f),
    C_Vertex3f(0.368232489f, 0.498943329f, -1.f),
    C_Vertex3f(0.370465517f, 0.488140106f, -1.f),
    C_Vertex3f(0.371582031f, 0.476987839f, -1.f),
    C_Vertex3f(0.376338243f, 0.331521392f, -1.f),
    C_Vertex3f(0.389757156f, 0.606599808f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.33138752f, 0.556432724f, -1.f),
    C_Vertex3f(0.340319633f, 0.54772377f, -1.f),
    C_Vertex3f(0.340713501f, 0.635378838f, -1.f),
    C_Vertex3f(0.348135233f, 0.538665771f, -1.f),
    C_Vertex3f(0.358248711f, 0.627384663f, -1.f),
    C_Vertex3f(0.354834318f, 0.529258728f, -1.f),
    C_Vertex3f(0.360416889f, 0.51950264f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.321990967f, 0.641774178f, -1.f),
    C_Vertex3f(0.302081108f, 0.646570683f, -1.f),
    C_Vertex3f(0.312011957f, 0.570584774f, -1.f),
    C_Vertex3f(0.321951628f, 0.564053059f, -1.f),
    C_Vertex3f(0.33138752f, 0.556432724f, -1.f),
    C_Vertex3f(0.340713501f, 0.635378838f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.26721549f, 0.585825443f, -1.f),
    C_Vertex3f(0.279170275f, 0.583648205f, -1.f),
    C_Vertex3f(0.280983925f, 0.649768353f, -1.f),
    C_Vertex3f(0.290621281f, 0.580382347f, -1.f),
    C_Vertex3f(0.302081108f, 0.646570683f, -1.f),
    C_Vertex3f(0.301568508f, 0.57602787f, -1.f),
    C_Vertex3f(0.312011957f, 0.570584774f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.258699417f, 0.651367188f, -1.f),
    C_Vertex3f(0.235883713f, 0.651367188f, -1.f),
    C_Vertex3f(0.242012024f, 0.586914063f, -1.f),
    C_Vertex3f(0.254756927f, 0.586914063f, -1.f),
    C_Vertex3f(0.26721549f, 0.585825443f, -1.f),
    C_Vertex3f(0.280983925f, 0.649768353f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.197077036f, 0.577037573f, -1.f),
    C_Vertex3f(0.207534552f, 0.580988169f, -1.f),
    C_Vertex3f(0.214808822f, 0.649880886f, -1.f),
    C_Vertex3f(0.218509555f, 0.583951116f, -1.f),
    C_Vertex3f(0.235883713f, 0.651367188f, -1.f),
    C_Vertex3f(0.230002046f, 0.585926414f, -1.f),
    C_Vertex3f(0.242012024f, 0.586914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.194818616f, 0.646908283f, -1.f),
    C_Vertex3f(0.175913095f, 0.642449379f, -1.f),
    C_Vertex3f(0.177714467f, 0.566173434f, -1.f),
    C_Vertex3f(0.187137008f, 0.572099328f, -1.f),
    C_Vertex3f(0.197077036f, 0.577037573f, -1.f),
    C_Vertex3f(0.214808822f, 0.649880886f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145811439f, 0.532222867f, -1.f),
    C_Vertex3f(0.152755857f, 0.542346358f, -1.f),
    C_Vertex3f(0.15809226f, 0.636504173f, -1.f),
    C_Vertex3f(0.160421848f, 0.5513587f, -1.f),
    C_Vertex3f(0.175913095f, 0.642449379f, -1.f),
    C_Vertex3f(0.168809414f, 0.559259892f, -1.f),
    C_Vertex3f(0.177714467f, 0.566173434f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.141356111f, 0.629072666f, -1.f),
    C_Vertex3f(0.125704646f, 0.620154858f, -1.f),
    C_Vertex3f(0.129307628f, 0.495185494f, -1.f),
    C_Vertex3f(0.134087324f, 0.508642435f, -1.f),
    C_Vertex3f(0.139588594f, 0.520988226f, -1.f),
    C_Vertex3f(0.145811439f, 0.532222867f, -1.f),
    C_Vertex3f(0.15809226f, 0.636504173f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.125704646f, 0.620154858f, -1.f),
    C_Vertex3f(0.111137867f, 0.609750748f, -1.f),
    C_Vertex3f(0.120605469f, 0.456542969f, -1.f),
    C_Vertex3f(0.121912956f, 0.464938164f, -1.f),
    C_Vertex3f(0.125249505f, 0.480617404f, -1.f),
    C_Vertex3f(0.129307628f, 0.495185494f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120605469f, 0.456542969f, -1.f),
    C_Vertex3f(0.111137867f, 0.609750748f, -1.f),
    C_Vertex3f(0.0976557732f, 0.597860336f, -1.f),
    C_Vertex3f(0.0855177641f, 0.584661007f, -1.f),
    C_Vertex3f(0.0747238398f, 0.57015276f, -1.f),
    C_Vertex3f(0.0652740002f, 0.554335594f, -1.f),
    C_Vertex3f(0.0571682453f, 0.537209511f, -1.f),
    C_Vertex3f(0.0504065752f, 0.518774509f, -1.f),
    C_Vertex3f(0.0449889898f, 0.49903059f, -1.f),
    C_Vertex3f(0.0409154892f, 0.477977753f, -1.f),
    C_Vertex3f(0.0395507813f, 0.466796875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.417151213f, 0.375483155f, -1.f),
    C_Vertex3f(0.424417734f, 0.385287166f, -1.f),
    C_Vertex3f(0.426436901f, 0.566218615f, -1.f),
    C_Vertex3f(0.430509567f, 0.394616604f, -1.f),
    C_Vertex3f(0.435170174f, 0.551418781f, -1.f),
    C_Vertex3f(0.437885284f, 0.407898903f, -1.f),
    C_Vertex3f(0.442156792f, 0.53594923f, -1.f),
    C_Vertex3f(0.445261002f, 0.426429272f, -1.f),
    C_Vertex3f(0.447396755f, 0.519809961f, -1.f),
    C_Vertex3f(0.450178146f, 0.445780277f, -1.f),
    C_Vertex3f(0.450890064f, 0.503000975f, -1.f),
    C_Vertex3f(0.452636719f, 0.46595192f, -1.f),
    C_Vertex3f(0.452636719f, 0.48552227f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.376338243f, 0.331521392f, -1.f),
    C_Vertex3f(0.388303518f, 0.343223691f, -1.f),
    C_Vertex3f(0.389757156f, 0.606599808f, -1.f),
    C_Vertex3f(0.399094105f, 0.354451418f, -1.f),
    C_Vertex3f(0.403730392f, 0.593809128f, -1.f),
    C_Vertex3f(0.408710003f, 0.365204573f, -1.f),
    C_Vertex3f(0.415956974f, 0.58034873f, -1.f),
    C_Vertex3f(0.417151213f, 0.375483155f, -1.f),
    C_Vertex3f(0.426436901f, 0.566218615f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376338243f, 0.331521392f, -1.f),
    C_Vertex3f(0.371582031f, 0.476987839f, -1.f),
    C_Vertex3f(0.371582031f, 0.464883804f, -1.f),
    C_Vertex3f(0.370053291f, 0.452670097f, -1.f),
    C_Vertex3f(0.366995811f, 0.44094944f, -1.f),
    C_Vertex3f(0.36319828f, 0.319344521f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36319828f, 0.319344521f, -1.f),
    C_Vertex3f(0.366995811f, 0.44094944f, -1.f),
    C_Vertex3f(0.362409592f, 0.429721832f, -1.f),
    C_Vertex3f(0.357823372f, 0.421670914f, -1.f),
    C_Vertex3f(0.353871942f, 0.415849805f, -1.f),
    C_Vertex3f(0.351233482f, 0.308765411f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.32110405f, 0.280328035f, -1.f),
    C_Vertex3f(0.327386856f, 0.286598921f, -1.f),
    C_Vertex3f(0.329126716f, 0.388027072f, -1.f),
    C_Vertex3f(0.334502697f, 0.293428779f, -1.f),
    C_Vertex3f(0.336653948f, 0.3956635f, -1.f),
    C_Vertex3f(0.342451572f, 0.300817609f, -1.f),
    C_Vertex3f(0.34328723f, 0.402846098f, -1.f),
    C_Vertex3f(0.351233482f, 0.308765411f, -1.f),
    C_Vertex3f(0.349026561f, 0.409574866f, -1.f),
    C_Vertex3f(0.353871942f, 0.415849805f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.307253838f, 0.264869213f, -1.f),
    C_Vertex3f(0.31103754f, 0.269463181f, -1.f),
    C_Vertex3f(0.3113904f, 0.371392727f, -1.f),
    C_Vertex3f(0.315654278f, 0.274616122f, -1.f),
    C_Vertex3f(0.320705533f, 0.379936814f, -1.f),
    C_Vertex3f(0.32110405f, 0.280328035f, -1.f),
    C_Vertex3f(0.329126716f, 0.388027072f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.286875725f, 0.220533371f, -1.f),
    C_Vertex3f(0.291135788f, 0.236240387f, -1.f),
    C_Vertex3f(0.296076775f, 0.357895851f, -1.f),
    C_Vertex3f(0.299877167f, 0.254781723f, -1.f),
    C_Vertex3f(0.3113904f, 0.371392727f, -1.f),
    C_Vertex3f(0.307253838f, 0.264869213f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.281738281f, 0.157714844f, -1.f),
    C_Vertex3f(0.281957626f, 0.169492722f, -1.f),
    C_Vertex3f(0.27879858f, 0.34205389f, -1.f),
    C_Vertex3f(0.282996655f, 0.189777374f, -1.f),
    C_Vertex3f(0.296076775f, 0.357895851f, -1.f),
    C_Vertex3f(0.284636021f, 0.206790924f, -1.f),
    C_Vertex3f(0.286875725f, 0.220533371f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.281738281f, 0.157714844f, -1.f),
    C_Vertex3f(0.27879858f, 0.34205389f, -1.f),
    C_Vertex3f(0.264660358f, 0.328365803f, -1.f),
    C_Vertex3f(0.253662109f, 0.316831589f, -1.f),
    C_Vertex3f(0.244382858f, 0.305707932f, -1.f),
    C_Vertex3f(0.235156059f, 0.292894363f, -1.f),
    C_Vertex3f(0.227402687f, 0.280134201f, -1.f),
    C_Vertex3f(0.221122742f, 0.267427444f, -1.f),
    C_Vertex3f(0.218719482f, 0.261100769f, -1.f),
    C_Vertex3f(0.215442657f, 0.25270462f, -1.f),
    C_Vertex3f(0.21052742f, 0.234330177f, -1.f),
    C_Vertex3f(0.207250595f, 0.214373589f, -1.f),
    C_Vertex3f(0.206054688f, 0.157714844f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.207250595f, 0.214373589f, -1.f),
    C_Vertex3f(0.205612183f, 0.192834854f, -1.f),
    C_Vertex3f(0.205667496f, 0.17440033f, -1.f),
    C_Vertex3f(0.205778122f, 0.168510437f, -1.f),
    C_Vertex3f(0.205944061f, 0.161640167f, -1.f),
    C_Vertex3f(0.206054688f, 0.157714844f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.192382813f, 0.0893554688f, -1.f),
    C_Vertex3f(0.192382813f, 0.f, -1.f),
    C_Vertex3f(0.281738281f, 0.f, -1.f),
    C_Vertex3f(0.281738281f, 0.0893554688f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.205667496f, 0.17440033f, -1.f),
    C_Vertex3f(0.205612183f, 0.192834854f, -1.f),
    C_Vertex3f(0.205612183f, 0.179309845f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_64[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.756130219f, -0.65442127f, 0.f),
    C_Vertex3f(0.507324219f, 0.0710906982f, 0.f),
    C_Vertex3f(0.507324219f, 0.0710906982f, -1.f),
    C_Normal3f(0.720308602f, -0.693653762f, 0.f),
    C_Vertex3f(0.500230789f, 0.0628948212f, 0.f),
    C_Vertex3f(0.500230789f, 0.0628948212f, -1.f),
    C_Normal3f(0.670676231f, -0.741750121f, 0.f),
    C_Vertex3f(0.485198975f, 0.0481152534f, 0.f),
    C_Vertex3f(0.485198975f, 0.0481152534f, -1.f),
    C_Normal3f(0.604272902f, -0.796777308f, 0.f),
    C_Vertex3f(0.469322205f, 0.0349478722f, 0.f),
    C_Vertex3f(0.469322205f, 0.0349478722f, -1.f),
    C_Normal3f(0.531205714f, -0.847242832f, 0.f),
    C_Vertex3f(0.452600479f, 0.0233926773f, 0.f),
    C_Vertex3f(0.452600479f, 0.0233926773f, -1.f),
    C_Normal3f(0.443534225f, -0.896257341f, 0.f),
    C_Vertex3f(0.435033798f, 0.0134496689f, 0.f),
    C_Vertex3f(0.435033798f, 0.0134496689f, -1.f),
    C_Normal3f(0.332807451f, -0.942994833f, 0.f),
    C_Vertex3f(0.417439461f, 0.00599241257f, 0.f),
    C_Vertex3f(0.417439461f, 0.00599241257f, -1.f),
    C_Normal3f(0.206850991f, -0.978372514f, 0.f),
    C_Vertex3f(0.399817467f, 0.00102090836f, 0.f),
    C_Vertex3f(0.399817467f, 0.00102090836f, -1.f),
    C_Normal3f(0.068508476f, -0.997650623f, 0.f),
    C_Vertex3f(0.382167816f, -0.00146484375f, 0.f),
    C_Vertex3f(0.382167816f, -0.00146484375f, -1.f),
    C_Normal3f(-0.0751502365f, -0.997172236f, 0.f),
    C_Vertex3f(0.363618851f, -0.00146484375f, 0.f),
    C_Vertex3f(0.363618851f, -0.00146484375f, -1.f),
    C_Normal3f(-0.218718082f, -0.975788057f, 0.f),
    C_Vertex3f(0.344470501f, 0.00137615204f, 0.f),
    C_Vertex3f(0.344470501f, 0.00137615204f, -1.f),
    C_Normal3f(-0.35440284f, -0.935092866f, 0.f),
    C_Vertex3f(0.325594425f, 0.00705814362f, 0.f),
    C_Vertex3f(0.325594425f, 0.00705814362f, -1.f),
    C_Normal3f(-0.474076092f, -0.880483866f, 0.f),
    C_Vertex3f(0.306990623f, 0.015581131f, 0.f),
    C_Vertex3f(0.306990623f, 0.015581131f, -1.f),
    C_Normal3f(-0.59377718f, -0.804629564f, 0.f),
    C_Vertex3f(0.288659096f, 0.0269451141f, 0.f),
    C_Vertex3f(0.288659096f, 0.0269451141f, -1.f),
    C_Normal3f(-0.711537778f, -0.702647805f, 0.f),
    C_Vertex3f(0.272046089f, 0.0413684845f, 0.f),
    C_Vertex3f(0.272046089f, 0.0413684845f, -1.f),
    C_Normal3f(-0.80452913f, -0.593913257f, 0.f),
    C_Vertex3f(0.257151604f, 0.0588512421f, 0.f),
    C_Vertex3f(0.257151604f, 0.0588512421f, -1.f),
    C_Normal3f(-0.8732391f, -0.487291843f, 0.f),
    C_Vertex3f(0.243975639f, 0.0793933868f, 0.f),
    C_Vertex3f(0.243975639f, 0.0793933868f, -1.f),
    C_Normal3f(-0.923723042f, -0.383061111f, 0.f),
    C_Vertex3f(0.232518196f, 0.102994919f, 0.f),
    C_Vertex3f(0.232518196f, 0.102994919f, -1.f),
    C_Normal3f(-0.962264001f, -0.272117496f, 0.f),
    C_Vertex3f(0.223925114f, 0.127743721f, 0.f),
    C_Vertex3f(0.223925114f, 0.127743721f, -1.f),
    C_Normal3f(-0.987080991f, -0.160222173f, 0.f),
    C_Vertex3f(0.218196392f, 0.153639793f, 0.f),
    C_Vertex3f(0.218196392f, 0.153639793f, -1.f),
    C_Normal3f(-0.998351634f, -0.0573928952f, 0.f),
    C_Vertex3f(0.215332031f, 0.180683136f, 0.f),
    C_Vertex3f(0.215332031f, 0.180683136f, -1.f),
    C_Normal3f(-0.999616981f, 0.0276760682f, 0.f),
    C_Vertex3f(0.215332031f, 0.203465462f, 0.f),
    C_Vertex3f(0.215332031f, 0.203465462f, -1.f),
    C_Normal3f(-0.995428562f, 0.0955090448f, 0.f),
    C_Vertex3f(0.216444016f, 0.220846295f, 0.f),
    C_Vertex3f(0.216444016f, 0.220846295f, -1.f),
    C_Normal3f(-0.987463713f, 0.157846361f, 0.f),
    C_Vertex3f(0.218667984f, 0.238233924f, 0.f),
    C_Vertex3f(0.218667984f, 0.238233924f, -1.f),
    C_Normal3f(-0.975880027f, 0.218307301f, 0.f),
    C_Vertex3f(0.222003937f, 0.255628347f, 0.f),
    C_Vertex3f(0.222003937f, 0.255628347f, -1.f),
    C_Normal3f(-0.961067855f, 0.276312649f, 0.f),
    C_Vertex3f(0.226451874f, 0.273029566f, 0.f),
    C_Vertex3f(0.226451874f, 0.273029566f, -1.f),
    C_Normal3f(-0.943487406f, 0.331408381f, 0.f),
    C_Vertex3f(0.232011795f, 0.290437579f, 0.f),
    C_Vertex3f(0.232011795f, 0.290437579f, -1.f),
    C_Normal3f(-0.923634887f, 0.383273602f, 0.f),
    C_Vertex3f(0.238683701f, 0.307852387f, 0.f),
    C_Vertex3f(0.238683701f, 0.307852387f, -1.f),
    C_Normal3f(-0.902009606f, 0.431716114f, 0.f),
    C_Vertex3f(0.24646759f, 0.325273991f, 0.f),
    C_Vertex3f(0.24646759f, 0.325273991f, -1.f),
    C_Normal3f(-0.87887758f, 0.477047443f, 0.f),
    C_Vertex3f(0.255363464f, 0.342702389f, 0.f),
    C_Vertex3f(0.255363464f, 0.342702389f, -1.f),
    C_Normal3f(-0.8521083f, 0.523365438f, 0.f),
    C_Vertex3f(0.264798164f, 0.359044909f, 0.f),
    C_Vertex3f(0.264798164f, 0.359044909f, -1.f),
    C_Normal3f(-0.820713818f, 0.571339548f, 0.f),
    C_Vertex3f(0.27477169f, 0.374301553f, 0.f),
    C_Vertex3f(0.27477169f, 0.374301553f, -1.f),
    C_Normal3f(-0.784239948f, 0.620457768f, 0.f),
    C_Vertex3f(0.285284042f, 0.388472319f, 0.f),
    C_Vertex3f(0.285284042f, 0.388472319f, -1.f),
    C_Normal3f(-0.742327034f, 0.670037806f, 0.f),
    C_Vertex3f(0.29633522f, 0.401557207f, 0.f),
    C_Vertex3f(0.29633522f, 0.401557207f, -1.f),
    C_Normal3f(-0.694770277f, 0.719231665f, 0.f),
    C_Vertex3f(0.307925224f, 0.413556218f, 0.f),
    C_Vertex3f(0.307925224f, 0.413556218f, -1.f),
    C_Normal3f(-0.641582727f, 0.767053902f, 0.f),
    C_Vertex3f(0.320054054f, 0.424469352f, 0.f),
    C_Vertex3f(0.320054054f, 0.424469352f, -1.f),
    C_Normal3f(-0.583047807f, 0.812437773f, 0.f),
    C_Vertex3f(0.33272171f, 0.434296608f, 0.f),
    C_Vertex3f(0.33272171f, 0.434296608f, -1.f),
    C_Normal3f(-0.528643489f, 0.848843992f, 0.f),
    C_Vertex3f(0.345928192f, 0.443037987f, 0.f),
    C_Vertex3f(0.345928192f, 0.443037987f, -1.f),
    C_Normal3f(-0.477652431f, 0.87854892f, 0.f),
    C_Vertex3f(0.359039307f, 0.450686693f, 0.f),
    C_Vertex3f(0.359039307f, 0.450686693f, -1.f),
    C_Normal3f(-0.420467198f, 0.907307744f, 0.f),
    C_Vertex3f(0.372055054f, 0.457242727f, 0.f),
    C_Vertex3f(0.372055054f, 0.457242727f, -1.f),
    C_Normal3f(-0.356827974f, 0.934170067f, 0.f),
    C_Vertex3f(0.384975433f, 0.462706089f, 0.f),
    C_Vertex3f(0.384975433f, 0.462706089f, -1.f),
    C_Normal3f(-0.286739409f, 0.958008647f, 0.f),
    C_Vertex3f(0.397800446f, 0.467076778f, 0.f),
    C_Vertex3f(0.397800446f, 0.467076778f, -1.f),
    C_Normal3f(-0.210569456f, 0.977578878f, 0.f),
    C_Vertex3f(0.41053009f, 0.470354795f, 0.f),
    C_Vertex3f(0.41053009f, 0.470354795f, -1.f),
    C_Normal3f(-0.129128382f, 0.991627812f, 0.f),
    C_Vertex3f(0.423164368f, 0.47254014f, 0.f),
    C_Vertex3f(0.423164368f, 0.47254014f, -1.f),
    C_Normal3f(-0.0386397019f, 0.999253154f, 0.f),
    C_Vertex3f(0.435703278f, 0.473632813f, 0.f),
    C_Vertex3f(0.435703278f, 0.473632813f, -1.f),
    C_Normal3f(0.0723346695f, 0.997380376f, 0.f),
    C_Vertex3f(0.451421738f, 0.473632813f, 0.f),
    C_Vertex3f(0.451421738f, 0.473632813f, -1.f),
    C_Normal3f(0.199256256f, 0.979947388f, 0.f),
    C_Vertex3f(0.469977856f, 0.47114706f, 0.f),
    C_Vertex3f(0.469977856f, 0.47114706f, -1.f),
    C_Normal3f(0.327965677f, 0.944689631f, 0.f),
    C_Vertex3f(0.488096714f, 0.466175556f, 0.f),
    C_Vertex3f(0.488096714f, 0.466175556f, -1.f),
    C_Normal3f(0.44592759f, 0.895069063f, 0.f),
    C_Vertex3f(0.505778313f, 0.4587183f, 0.f),
    C_Vertex3f(0.505778313f, 0.4587183f, -1.f),
    C_Normal3f(0.55967325f, 0.828713357f, 0.f),
    C_Vertex3f(0.523022652f, 0.448775291f, 0.f),
    C_Vertex3f(0.523022652f, 0.448775291f, -1.f),
    C_Normal3f(0.667271316f, 0.744814754f, 0.f),
    C_Vertex3f(0.539066792f, 0.436236858f, 0.f),
    C_Vertex3f(0.539066792f, 0.436236858f, -1.f),
    C_Normal3f(0.755617499f, 0.655013204f, 0.f),
    C_Vertex3f(0.553910732f, 0.421103001f, 0.f),
    C_Vertex3f(0.553910732f, 0.421103001f, -1.f),
    C_Normal3f(0.814521194f, 0.580133796f, 0.f),
    C_Vertex3f(0.567554474f, 0.403373718f, 0.f),
    C_Vertex3f(0.567554474f, 0.403373718f, -1.f),
    C_Normal3f(0.852854252f, 0.522149146f, 0.f),
    C_Vertex3f(0.573776245f, 0.393211365f, 0.f),
    C_Vertex3f(0.573776245f, 0.393211365f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.976062775f, 0.217489257f, 0.f),
    C_Vertex3f(0.573776245f, 0.393211365f, 0.f),
    C_Vertex3f(0.573776245f, 0.393211365f, -1.f),
    C_Vertex3f(0.588867188f, 0.4609375f, -1.f),
    C_Vertex3f(0.588867188f, 0.4609375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.588867188f, 0.4609375f, 0.f),
    C_Vertex3f(0.588867188f, 0.4609375f, -1.f),
    C_Vertex3f(0.666992188f, 0.4609375f, -1.f),
    C_Vertex3f(0.666992188f, 0.4609375f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.977699578f, -0.210008442f, 0.f),
    C_Vertex3f(0.666992188f, 0.4609375f, 0.f),
    C_Vertex3f(0.666992188f, 0.4609375f, -1.f),
    C_Normal3f(0.977730572f, -0.209863976f, 0.f),
    C_Vertex3f(0.602309227f, 0.159804344f, 0.f),
    C_Vertex3f(0.602309227f, 0.159804344f, -1.f),
    C_Normal3f(0.978795111f, -0.204841524f, 0.f),
    C_Vertex3f(0.599436998f, 0.146206141f, 0.f),
    C_Vertex3f(0.599436998f, 0.146206141f, -1.f),
    C_Normal3f(0.979724884f, -0.200347528f, 0.f),
    C_Vertex3f(0.596975088f, 0.134316206f, 0.f),
    C_Vertex3f(0.596975088f, 0.134316206f, -1.f),
    C_Normal3f(0.980964541f, -0.194186971f, 0.f),
    C_Vertex3f(0.594923496f, 0.124134541f, 0.f),
    C_Vertex3f(0.594923496f, 0.124134541f, -1.f),
    C_Normal3f(0.982696414f, -0.185223252f, 0.f),
    C_Vertex3f(0.593282223f, 0.115661144f, 0.f),
    C_Vertex3f(0.593282223f, 0.115661144f, -1.f),
    C_Normal3f(0.9852736f, -0.17098476f, 0.f),
    C_Vertex3f(0.592051268f, 0.108896017f, 0.f),
    C_Vertex3f(0.592051268f, 0.108896017f, -1.f),
    C_Normal3f(0.989446044f, -0.144901723f, 0.f),
    C_Vertex3f(0.591230631f, 0.103839159f, 0.f),
    C_Vertex3f(0.591230631f, 0.103839159f, -1.f),
    C_Normal3f(0.998344183f, -0.0575228594f, 0.f),
    C_Vertex3f(0.590820313f, 0.10049057f, 0.f),
    C_Vertex3f(0.590820313f, 0.10049057f, -1.f),
    C_Normal3f(0.992745161f, 0.120236784f, 0.f),
    C_Vertex3f(0.590820313f, 0.0967178345f, 0.f),
    C_Vertex3f(0.590820313f, 0.0967178345f, -1.f),
    C_Normal3f(0.952898741f, 0.303288758f, 0.f),
    C_Vertex3f(0.591942787f, 0.0912227631f, 0.f),
    C_Vertex3f(0.591942787f, 0.0912227631f, -1.f),
    C_Normal3f(0.866889834f, 0.498499781f, 0.f),
    C_Vertex3f(0.594187737f, 0.0861377716f, 0.f),
    C_Vertex3f(0.594187737f, 0.0861377716f, -1.f),
    C_Normal3f(0.751115561f, 0.660170794f, 0.f),
    C_Vertex3f(0.597555161f, 0.0814628601f, 0.f),
    C_Vertex3f(0.597555161f, 0.0814628601f, -1.f),
    C_Normal3f(0.620026946f, 0.784580529f, 0.f),
    C_Vertex3f(0.602045059f, 0.0771980286f, 0.f),
    C_Vertex3f(0.602045059f, 0.0771980286f, -1.f),
    C_Normal3f(0.457096726f, 0.889416993f, 0.f),
    C_Vertex3f(0.606999397f, 0.0739994049f, 0.f),
    C_Vertex3f(0.606999397f, 0.0739994049f, -1.f),
    C_Normal3f(0.272318214f, 0.962207198f, 0.f),
    C_Vertex3f(0.612418175f, 0.0718669891f, 0.f),
    C_Vertex3f(0.612418175f, 0.0718669891f, -1.f),
    C_Normal3f(0.0715609118f, 0.997436285f, 0.f),
    C_Vertex3f(0.618301392f, 0.0708007813f, 0.f),
    C_Vertex3f(0.618301392f, 0.0708007813f, -1.f),
    C_Normal3f(-0.147234142f, 0.989101648f, 0.f),
    C_Vertex3f(0.627279282f, 0.0708007813f, 0.f),
    C_Vertex3f(0.627279282f, 0.0708007813f, -1.f),
    C_Normal3f(-0.329704612f, 0.944084048f, 0.f),
    C_Vertex3f(0.640692711f, 0.074133873f, 0.f),
    C_Vertex3f(0.640692711f, 0.074133873f, -1.f),
    C_Normal3f(-0.459165573f, 0.888350725f, 0.f),
    C_Vertex3f(0.655911446f, 0.0808000565f, 0.f),
    C_Vertex3f(0.655911446f, 0.0808000565f, -1.f),
    C_Normal3f(-0.541282117f, 0.840841055f, 0.f),
    C_Vertex3f(0.672935486f, 0.0907993317f, 0.f),
    C_Vertex3f(0.672935486f, 0.0907993317f, -1.f),
    C_Normal3f(-0.598323464f, 0.80125463f, 0.f),
    C_Vertex3f(0.688591003f, 0.101837158f, 0.f),
    C_Vertex3f(0.688591003f, 0.101837158f, -1.f),
    C_Normal3f(-0.650683463f, 0.759349048f, 0.f),
    C_Vertex3f(0.700422645f, 0.111324906f, 0.f),
    C_Vertex3f(0.700422645f, 0.111324906f, -1.f),
    C_Normal3f(-0.69875139f, 0.715364635f, 0.f),
    C_Vertex3f(0.711604238f, 0.121557117f, 0.f),
    C_Vertex3f(0.711604238f, 0.121557117f, -1.f),
    C_Normal3f(-0.743538201f, 0.668693483f, 0.f),
    C_Vertex3f(0.722135782f, 0.132533789f, 0.f),
    C_Vertex3f(0.722135782f, 0.132533789f, -1.f),
    C_Normal3f(-0.784596205f, 0.620007038f, 0.f),
    C_Vertex3f(0.732017279f, 0.144254923f, 0.f),
    C_Vertex3f(0.732017279f, 0.144254923f, -1.f),
    C_Normal3f(-0.821631789f, 0.570018709f, 0.f),
    C_Vertex3f(0.741248727f, 0.156720519f, 0.f),
    C_Vertex3f(0.741248727f, 0.156720519f, -1.f),
    C_Normal3f(-0.854508638f, 0.519437075f, 0.f),
    C_Vertex3f(0.749830127f, 0.169930577f, 0.f),
    C_Vertex3f(0.749830127f, 0.169930577f, -1.f),
    C_Normal3f(-0.885886192f, 0.463902652f, 0.f),
    C_Vertex3f(0.757761478f, 0.183885098f, 0.f),
    C_Vertex3f(0.757761478f, 0.183885098f, -1.f),
    C_Normal3f(-0.921718538f, 0.387859404f, 0.f),
    C_Vertex3f(0.768683434f, 0.205933571f, 0.f),
    C_Vertex3f(0.768683434f, 0.205933571f, -1.f),
    C_Normal3f(-0.957180798f, 0.28949064f, 0.f),
    C_Vertex3f(0.779605389f, 0.235795498f, 0.f),
    C_Vertex3f(0.779605389f, 0.235795498f, -1.f),
    C_Normal3f(-0.984404087f, 0.17592223f, 0.f),
    C_Vertex3f(0.786886692f, 0.266121387f, 0.f),
    C_Vertex3f(0.786886692f, 0.266121387f, -1.f),
    C_Normal3f(-0.997858703f, 0.0654068142f, 0.f),
    C_Vertex3f(0.790527344f, 0.29691124f, 0.f),
    C_Vertex3f(0.790527344f, 0.29691124f, -1.f),
    C_Normal3f(-0.9996261f, -0.0273443181f, 0.f),
    C_Vertex3f(0.790527344f, 0.321663857f, 0.f),
    C_Vertex3f(0.790527344f, 0.321663857f, -1.f),
    C_Normal3f(-0.995190084f, -0.0979621187f, 0.f),
    C_Vertex3f(0.789359212f, 0.33961463f, 0.f),
    C_Vertex3f(0.789359212f, 0.33961463f, -1.f),
    C_Normal3f(-0.986360073f, -0.164601639f, 0.f),
    C_Vertex3f(0.787022948f, 0.357264757f, 0.f),
    C_Vertex3f(0.787022948f, 0.357264757f, -1.f),
    C_Normal3f(-0.972890019f, -0.23126848f, 0.f),
    C_Vertex3f(0.783518553f, 0.374614239f, 0.f),
    C_Vertex3f(0.783518553f, 0.374614239f, -1.f),
    C_Normal3f(-0.954868436f, -0.297029376f, 0.f),
    C_Vertex3f(0.778846025f, 0.391663074f, 0.f),
    C_Vertex3f(0.778846025f, 0.391663074f, -1.f),
    C_Normal3f(-0.932572901f, -0.360981584f, 0.f),
    C_Vertex3f(0.773005366f, 0.408411264f, 0.f),
    C_Vertex3f(0.773005366f, 0.408411264f, -1.f),
    C_Normal3f(-0.906449676f, -0.422313869f, 0.f),
    C_Vertex3f(0.765996575f, 0.424858809f, 0.f),
    C_Vertex3f(0.765996575f, 0.424858809f, -1.f),
    C_Normal3f(-0.877074122f, -0.480355114f, 0.f),
    C_Vertex3f(0.757819653f, 0.441005707f, 0.f),
    C_Vertex3f(0.757819653f, 0.441005707f, -1.f),
    C_Normal3f(-0.841507614f, -0.540245235f, 0.f),
    C_Vertex3f(0.748474598f, 0.456851959f, 0.f),
    C_Vertex3f(0.748474598f, 0.456851959f, -1.f),
    C_Normal3f(-0.798080623f, -0.602550566f, 0.f),
    C_Vertex3f(0.737981796f, 0.471905947f, 0.f),
    C_Vertex3f(0.737981796f, 0.471905947f, -1.f),
    C_Normal3f(-0.750372529f, -0.661015272f, 0.f),
    C_Vertex3f(0.726341248f, 0.486167669f, 0.f),
    C_Vertex3f(0.726341248f, 0.486167669f, -1.f),
    C_Normal3f(-0.699338973f, -0.714790285f, 0.f),
    C_Vertex3f(0.713552952f, 0.499637127f, 0.f),
    C_Vertex3f(0.713552952f, 0.499637127f, -1.f),
    C_Normal3f(-0.646047652f, -0.763297081f, 0.f),
    C_Vertex3f(0.699616909f, 0.51231432f, 0.f),
    C_Vertex3f(0.699616909f, 0.51231432f, -1.f),
    C_Normal3f(-0.591581583f, -0.806245148f, 0.f),
    C_Vertex3f(0.684533119f, 0.524199247f, 0.f),
    C_Vertex3f(0.684533119f, 0.524199247f, -1.f),
    C_Normal3f(-0.53695333f, -0.843611956f, 0.f),
    C_Vertex3f(0.668301582f, 0.53529191f, 0.f),
    C_Vertex3f(0.668301582f, 0.53529191f, -1.f),
    C_Normal3f(-0.483042777f, -0.875596821f, 0.f),
    C_Vertex3f(0.650922298f, 0.545592308f, 0.f),
    C_Vertex3f(0.650922298f, 0.545592308f, -1.f),
    C_Normal3f(-0.428991348f, -0.90330857f, 0.f),
    C_Vertex3f(0.632395267f, 0.555100441f, 0.f),
    C_Vertex3f(0.632395267f, 0.555100441f, -1.f),
    C_Normal3f(-0.372367293f, -0.928085506f, 0.f),
    C_Vertex3f(0.613383174f, 0.563420057f, 0.f),
    C_Vertex3f(0.613383174f, 0.563420057f, -1.f),
    C_Normal3f(-0.314363658f, -0.949302554f, 0.f),
    C_Vertex3f(0.593886018f, 0.570551157f, 0.f),
    C_Vertex3f(0.593886018f, 0.570551157f, -1.f),
    C_Normal3f(-0.255656421f, -0.966767669f, 0.f),
    C_Vertex3f(0.573903799f, 0.57649374f, 0.f),
    C_Vertex3f(0.573903799f, 0.57649374f, -1.f),
    C_Normal3f(-0.196928412f, -0.980417967f, 0.f),
    C_Vertex3f(0.553436518f, 0.581247807f, 0.f),
    C_Vertex3f(0.553436518f, 0.581247807f, -1.f),
    C_Normal3f(-0.138831556f, -0.990315974f, 0.f),
    C_Vertex3f(0.532484174f, 0.584813356f, 0.f),
    C_Vertex3f(0.532484174f, 0.584813356f, -1.f),
    C_Normal3f(-0.0819549337f, -0.996636093f, 0.f),
    C_Vertex3f(0.511046767f, 0.58719039f, 0.f),
    C_Vertex3f(0.511046767f, 0.58719039f, -1.f),
    C_Normal3f(-0.0258764774f, -0.999665201f, 0.f),
    C_Vertex3f(0.489124298f, 0.588378906f, 0.f),
    C_Vertex3f(0.489124298f, 0.588378906f, -1.f),
    C_Normal3f(0.0304978415f, -0.999534845f, 0.f),
    C_Vertex3f(0.46513176f, 0.588378906f, 0.f),
    C_Vertex3f(0.46513176f, 0.588378906f, -1.f),
    C_Normal3f(0.0902814046f, -0.995916307f, 0.f),
    C_Vertex3f(0.440107584f, 0.586883307f, 0.f),
    C_Vertex3f(0.440107584f, 0.586883307f, -1.f),
    C_Normal3f(0.152728587f, -0.988268197f, 0.f),
    C_Vertex3f(0.415636778f, 0.583892107f, 0.f),
    C_Vertex3f(0.415636778f, 0.583892107f, -1.f),
    C_Normal3f(0.216186509f, -0.976352036f, 0.f),
    C_Vertex3f(0.391719341f, 0.579405308f, 0.f),
    C_Vertex3f(0.391719341f, 0.579405308f, -1.f),
    C_Normal3f(0.279861212f, -0.96004045f, 0.f),
    C_Vertex3f(0.368355274f, 0.573422909f, 0.f),
    C_Vertex3f(0.368355274f, 0.573422909f, -1.f),
    C_Normal3f(0.342906713f, -0.93936944f, 0.f),
    C_Vertex3f(0.345544577f, 0.56594491f, 0.f),
    C_Vertex3f(0.345544577f, 0.56594491f, -1.f),
    C_Normal3f(0.404477865f, -0.914547861f, 0.f),
    C_Vertex3f(0.323287249f, 0.556971312f, 0.f),
    C_Vertex3f(0.323287249f, 0.556971312f, -1.f),
    C_Normal3f(0.463785976f, -0.885947287f, 0.f),
    C_Vertex3f(0.30158329f, 0.546502113f, 0.f),
    C_Vertex3f(0.30158329f, 0.546502113f, -1.f),
    C_Normal3f(0.523001015f, -0.852331996f, 0.f),
    C_Vertex3f(0.280432701f, 0.534537315f, 0.f),
    C_Vertex3f(0.280432701f, 0.534537315f, -1.f),
    C_Normal3f(0.583377481f, -0.812201202f, 0.f),
    C_Vertex3f(0.260470867f, 0.521275043f, 0.f),
    C_Vertex3f(0.260470867f, 0.521275043f, -1.f),
    C_Normal3f(0.641666353f, -0.766983867f, 0.f),
    C_Vertex3f(0.241697788f, 0.506715298f, 0.f),
    C_Vertex3f(0.241697788f, 0.506715298f, -1.f),
    C_Normal3f(0.696816921f, -0.717249036f, 0.f),
    C_Vertex3f(0.224113464f, 0.490858078f, 0.f),
    C_Vertex3f(0.224113464f, 0.490858078f, -1.f),
    C_Normal3f(0.747911453f, -0.663798571f, 0.f),
    C_Vertex3f(0.207717896f, 0.473703384f, 0.f),
    C_Vertex3f(0.207717896f, 0.473703384f, -1.f),
    C_Normal3f(0.794240177f, -0.607603908f, 0.f),
    C_Vertex3f(0.192511082f, 0.455251217f, 0.f),
    C_Vertex3f(0.192511082f, 0.455251217f, -1.f),
    C_Normal3f(0.835347593f, -0.549722135f, 0.f),
    C_Vertex3f(0.178493023f, 0.435501575f, 0.f),
    C_Vertex3f(0.178493023f, 0.435501575f, -1.f),
    C_Normal3f(0.871042967f, -0.491206646f, 0.f),
    C_Vertex3f(0.165663719f, 0.41445446f, 0.f),
    C_Vertex3f(0.165663719f, 0.41445446f, -1.f),
    C_Normal3f(0.900101483f, -0.43568036f, 0.f),
    C_Vertex3f(0.15402317f, 0.392109871f, 0.f),
    C_Vertex3f(0.15402317f, 0.392109871f, -1.f),
    C_Normal3f(0.924557328f, -0.381043047f, 0.f),
    C_Vertex3f(0.14383769f, 0.369362593f, 0.f),
    C_Vertex3f(0.14383769f, 0.369362593f, -1.f),
    C_Normal3f(0.945986807f, -0.32420519f, 0.f),
    C_Vertex3f(0.135107279f, 0.346212626f, 0.f),
    C_Vertex3f(0.135107279f, 0.346212626f, -1.f),
    C_Normal3f(0.96404475f, -0.265739679f, 0.f),
    C_Vertex3f(0.127831936f, 0.322659969f, 0.f),
    C_Vertex3f(0.127831936f, 0.322659969f, -1.f),
    C_Normal3f(0.978491604f, -0.206286684f, 0.f),
    C_Vertex3f(0.122011662f, 0.298704624f, 0.f),
    C_Vertex3f(0.122011662f, 0.298704624f, -1.f),
    C_Normal3f(0.989207923f, -0.146518901f, 0.f),
    C_Vertex3f(0.117646456f, 0.27434659f, 0.f),
    C_Vertex3f(0.117646456f, 0.27434659f, -1.f),
    C_Normal3f(0.996199191f, -0.0871044472f, 0.f),
    C_Vertex3f(0.114736319f, 0.249585867f, 0.f),
    C_Vertex3f(0.114736319f, 0.249585867f, -1.f),
    C_Normal3f(0.999598503f, -0.0283352435f, 0.f),
    C_Vertex3f(0.11328125f, 0.224422455f, 0.f),
    C_Vertex3f(0.11328125f, 0.224422455f, -1.f),
    C_Normal3f(0.999609113f, 0.027959507f, 0.f),
    C_Vertex3f(0.11328125f, 0.198254585f, 0.f),
    C_Vertex3f(0.11328125f, 0.198254585f, -1.f),
    C_Normal3f(0.996323466f, 0.085670948f, 0.f),
    C_Vertex3f(0.114735961f, 0.172413588f, 0.f),
    C_Vertex3f(0.114735961f, 0.172413588f, -1.f),
    C_Normal3f(0.989116251f, 0.147136375f, 0.f),
    C_Vertex3f(0.117645383f, 0.14750123f, 0.f),
    C_Vertex3f(0.117645383f, 0.14750123f, -1.f),
    C_Normal3f(0.977360845f, 0.211579457f, 0.f),
    C_Vertex3f(0.122009516f, 0.123517513f, 0.f),
    C_Vertex3f(0.122009516f, 0.123517513f, -1.f),
    C_Normal3f(0.960487247f, 0.278323621f, 0.f),
    C_Vertex3f(0.12782836f, 0.100462437f, 0.f),
    C_Vertex3f(0.12782836f, 0.100462437f, -1.f),
    C_Normal3f(0.938059986f, 0.3464728f, 0.f),
    C_Vertex3f(0.135101914f, 0.0783360004f, 0.f),
    C_Vertex3f(0.135101914f, 0.0783360004f, -1.f),
    C_Normal3f(0.909848154f, 0.41494149f, 0.f),
    C_Vertex3f(0.14383018f, 0.0571382046f, 0.f),
    C_Vertex3f(0.14383018f, 0.0571382046f, -1.f),
    C_Normal3f(0.875885606f, 0.482518673f, 0.f),
    C_Vertex3f(0.154013157f, 0.0368690491f, 0.f),
    C_Vertex3f(0.154013157f, 0.0368690491f, -1.f),
    C_Normal3f(0.835863888f, 0.548936725f, 0.f),
    C_Vertex3f(0.165650845f, 0.0175285339f, 0.f),
    C_Vertex3f(0.165650845f, 0.0175285339f, -1.f),
    C_Normal3f(0.78900373f, 0.614388466f, 0.f),
    C_Vertex3f(0.178586364f, -0.000548481941f, 0.f),
    C_Vertex3f(0.178586364f, -0.000548481941f, -1.f),
    C_Normal3f(0.736043334f, 0.676934302f, 0.f),
    C_Vertex3f(0.192819715f, -0.0173619986f, 0.f),
    C_Vertex3f(0.192819715f, -0.0173619986f, -1.f),
    C_Normal3f(0.677857816f, 0.735193014f, 0.f),
    C_Vertex3f(0.208350897f, -0.0329120159f, 0.f),
    C_Vertex3f(0.208350897f, -0.0329120159f, -1.f),
    C_Normal3f(0.615647912f, 0.788021445f, 0.f),
    C_Vertex3f(0.225179911f, -0.047198534f, 0.f),
    C_Vertex3f(0.225179911f, -0.047198534f, -1.f),
    C_Normal3f(0.550819457f, 0.83462435f, 0.f),
    C_Vertex3f(0.243306756f, -0.0602215528f, 0.f),
    C_Vertex3f(0.243306756f, -0.0602215528f, -1.f),
    C_Normal3f(0.484836161f, 0.87460506f, 0.f),
    C_Vertex3f(0.262731433f, -0.0719810724f, 0.f),
    C_Vertex3f(0.262731433f, -0.0719810724f, -1.f),
    C_Normal3f(0.419076949f, 0.907950699f, 0.f),
    C_Vertex3f(0.283453941f, -0.0824770927f, 0.f),
    C_Vertex3f(0.283453941f, -0.0824770927f, -1.f),
    C_Normal3f(0.361593187f, 0.932335973f, 0.f),
    C_Vertex3f(0.305474281f, -0.0917096138f, 0.f),
    C_Vertex3f(0.305474281f, -0.0917096138f, -1.f),
    C_Normal3f(0.31115514f, 0.950359166f, 0.f),
    C_Vertex3f(0.32808876f, -0.0997880697f, 0.f),
    C_Vertex3f(0.32808876f, -0.0997880697f, -1.f),
    C_Normal3f(0.260697335f, 0.965420544f, 0.f),
    C_Vertex3f(0.351297379f, -0.106712461f, 0.f),
    C_Vertex3f(0.351297379f, -0.106712461f, -1.f),
    C_Normal3f(0.210655347f, 0.977560341f, 0.f),
    C_Vertex3f(0.375100136f, -0.112482786f, 0.f),
    C_Vertex3f(0.375100136f, -0.112482786f, -1.f),
    C_Normal3f(0.161426187f, 0.986884832f, 0.f),
    C_Vertex3f(0.399497032f, -0.117099047f, 0.f),
    C_Vertex3f(0.399497032f, -0.117099047f, -1.f),
    C_Normal3f(0.113356315f, 0.993554413f, 0.f),
    C_Vertex3f(0.424488068f, -0.120561242f, 0.f),
    C_Vertex3f(0.424488068f, -0.120561242f, -1.f),
    C_Normal3f(0.0667345002f, 0.997770786f, 0.f),
    C_Vertex3f(0.450073242f, -0.122869372f, 0.f),
    C_Vertex3f(0.450073242f, -0.122869372f, -1.f),
    C_Normal3f(0.0214139894f, 0.999770761f, 0.f),
    C_Vertex3f(0.476252556f, -0.124023438f, 0.f),
    C_Vertex3f(0.476252556f, -0.124023438f, -1.f),
    C_Normal3f(-0.0218358561f, 0.999761581f, 0.f),
    C_Vertex3f(0.503953934f, -0.124023438f, 0.f),
    C_Vertex3f(0.503953934f, -0.124023438f, -1.f),
    C_Normal3f(-0.0670700818f, 0.997748196f, 0.f),
    C_Vertex3f(0.531422377f, -0.12281847f, 0.f),
    C_Vertex3f(0.531422377f, -0.12281847f, -1.f),
    C_Normal3f(-0.116296515f, 0.993214607f, 0.f),
    C_Vertex3f(0.557729959f, -0.120408535f, 0.f),
    C_Vertex3f(0.557729959f, -0.120408535f, -1.f),
    C_Normal3f(-0.169198513f, 0.985581994f, 0.f),
    C_Vertex3f(0.582876682f, -0.116793633f, 0.f),
    C_Vertex3f(0.582876682f, -0.116793633f, -1.f),
    C_Normal3f(-0.225693256f, 0.974198341f, 0.f),
    C_Vertex3f(0.606862545f, -0.111973763f, 0.f),
    C_Vertex3f(0.606862545f, -0.111973763f, -1.f),
    C_Normal3f(-0.28552717f, 0.958370566f, 0.f),
    C_Vertex3f(0.629687548f, -0.105948925f, 0.f),
    C_Vertex3f(0.629687548f, -0.105948925f, -1.f),
    C_Normal3f(-0.348233223f, 0.937407911f, 0.f),
    C_Vertex3f(0.65135169f, -0.09871912f, 0.f),
    C_Vertex3f(0.65135169f, -0.09871912f, -1.f),
    C_Normal3f(-0.413098723f, 0.910686255f, 0.f),
    C_Vertex3f(0.671854973f, -0.0902843475f, 0.f),
    C_Vertex3f(0.671854973f, -0.0902843475f, -1.f),
    C_Normal3f(-0.467207462f, 0.884147644f, 0.f),
    C_Vertex3f(0.691197395f, -0.0806446075f, 0.f),
    C_Vertex3f(0.691197395f, -0.0806446075f, -1.f),
    C_Normal3f(-0.511200547f, 0.859461427f, 0.f),
    C_Vertex3f(0.709324121f, -0.0704846382f, 0.f),
    C_Vertex3f(0.709324121f, -0.0704846382f, -1.f),
    C_Normal3f(-0.557219684f, 0.830365062f, 0.f),
    C_Vertex3f(0.726235151f, -0.0598044395f, 0.f),
    C_Vertex3f(0.726235151f, -0.0598044395f, -1.f),
    C_Normal3f(-0.604884326f, 0.796313345f, 0.f),
    C_Vertex3f(0.741930485f, -0.0486040115f, 0.f),
    C_Vertex3f(0.741930485f, -0.0486040115f, -1.f),
    C_Normal3f(-0.653638363f, 0.756807089f, 0.f),
    C_Vertex3f(0.756410122f, -0.0368833542f, 0.f),
    C_Vertex3f(0.756410122f, -0.0368833542f, -1.f),
    C_Normal3f(-0.702734172f, 0.711452484f, 0.f),
    C_Vertex3f(0.769674063f, -0.0246424675f, 0.f),
    C_Vertex3f(0.769674063f, -0.0246424675f, -1.f),
    C_Normal3f(-0.751236498f, 0.660033107f, 0.f),
    C_Vertex3f(0.781722307f, -0.0118813515f, 0.f),
    C_Vertex3f(0.781722307f, -0.0118813515f, -1.f),
    C_Normal3f(-0.790418446f, 0.612567306f, 0.f),
    C_Vertex3f(0.792554855f, 0.0013999939f, 0.f),
    C_Vertex3f(0.792554855f, 0.0013999939f, -1.f),
    C_Normal3f(-0.820465803f, 0.571695507f, 0.f),
    C_Vertex3f(0.797363281f, 0.00830078125f, 0.f),
    C_Vertex3f(0.797363281f, 0.00830078125f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.797363281f, 0.00830078125f, 0.f),
    C_Vertex3f(0.797363281f, 0.00830078125f, -1.f),
    C_Vertex3f(0.876464844f, 0.00830078125f, -1.f),
    C_Vertex3f(0.876464844f, 0.00830078125f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.899627328f, -0.436658621f, 0.f),
    C_Vertex3f(0.876464844f, 0.00830078125f, 0.f),
    C_Vertex3f(0.876464844f, 0.00830078125f, -1.f),
    C_Normal3f(0.879334152f, -0.476205111f, 0.f),
    C_Vertex3f(0.873679161f, 0.00256156921f, 0.f),
    C_Vertex3f(0.873679161f, 0.00256156921f, -1.f),
    C_Normal3f(0.85322082f, -0.521549821f, 0.f),
    C_Vertex3f(0.867110968f, -0.00897157192f, 0.f),
    C_Vertex3f(0.867110968f, -0.00897157192f, -1.f),
    C_Normal3f(0.821458995f, -0.570267558f, 0.f),
    C_Vertex3f(0.859545946f, -0.0205594301f, 0.f),
    C_Vertex3f(0.859545946f, -0.0205594301f, -1.f),
    C_Normal3f(0.789888263f, -0.613250792f, 0.f),
    C_Vertex3f(0.850984097f, -0.0322020054f, 0.f),
    C_Vertex3f(0.850984097f, -0.0322020054f, -1.f),
    C_Normal3f(0.75902015f, -0.651067197f, 0.f),
    C_Vertex3f(0.841425419f, -0.0438992977f, 0.f),
    C_Vertex3f(0.841425419f, -0.0438992977f, -1.f),
    C_Normal3f(0.729204655f, -0.684295654f, 0.f),
    C_Vertex3f(0.830869913f, -0.0556513071f, 0.f),
    C_Vertex3f(0.830869913f, -0.0556513071f, -1.f),
    C_Normal3f(0.700663686f, -0.713491738f, 0.f),
    C_Vertex3f(0.819317579f, -0.0674580336f, 0.f),
    C_Vertex3f(0.819317579f, -0.0674580336f, -1.f),
    C_Normal3f(0.670886457f, -0.741560042f, 0.f),
    C_Vertex3f(0.806768417f, -0.0793194771f, 0.f),
    C_Vertex3f(0.806768417f, -0.0793194771f, -1.f),
    C_Normal3f(0.617338479f, -0.786697686f, 0.f),
    C_Vertex3f(0.786449432f, -0.097193718f, 0.f),
    C_Vertex3f(0.786449432f, -0.097193718f, -1.f),
    C_Normal3f(0.544763267f, -0.838589907f, 0.f),
    C_Vertex3f(0.756789684f, -0.118538857f, 0.f),
    C_Vertex3f(0.756789684f, -0.118538857f, -1.f),
    C_Normal3f(0.465266168f, -0.885170758f, 0.f),
    C_Vertex3f(0.724562168f, -0.137396812f, 0.f),
    C_Vertex3f(0.724562168f, -0.137396812f, -1.f),
    C_Normal3f(0.392207056f, -0.919876993f, 0.f),
    C_Vertex3f(0.689766884f, -0.153767586f, 0.f),
    C_Vertex3f(0.689766884f, -0.153767586f, -1.f),
    C_Normal3f(0.326920003f, -0.945052028f, 0.f),
    C_Vertex3f(0.661744595f, -0.164180279f, 0.f),
    C_Vertex3f(0.661744595f, -0.164180279f, -1.f),
    C_Normal3f(0.269984424f, -0.962864697f, 0.f),
    C_Vertex3f(0.642107368f, -0.17025435f, 0.f),
    C_Vertex3f(0.642107368f, -0.17025435f, -1.f),
    C_Normal3f(0.220902354f, -0.975295961f, 0.f),
    C_Vertex3f(0.62151444f, -0.175460696f, 0.f),
    C_Vertex3f(0.62151444f, -0.175460696f, -1.f),
    C_Normal3f(0.17455411f, -0.984647572f, 0.f),
    C_Vertex3f(0.599965811f, -0.179799318f, 0.f),
    C_Vertex3f(0.599965811f, -0.179799318f, -1.f),
    C_Normal3f(0.131008446f, -0.991381168f, 0.f),
    C_Vertex3f(0.577461481f, -0.183270216f, 0.f),
    C_Vertex3f(0.577461481f, -0.183270216f, -1.f),
    C_Normal3f(0.0902526826f, -0.99591893f, 0.f),
    C_Vertex3f(0.554001451f, -0.185873389f, 0.f),
    C_Vertex3f(0.554001451f, -0.185873389f, -1.f),
    C_Normal3f(0.0522147082f, -0.998635888f, 0.f),
    C_Vertex3f(0.529585719f, -0.187608838f, 0.f),
    C_Vertex3f(0.529585719f, -0.187608838f, -1.f),
    C_Normal3f(0.0138188228f, -0.999904513f, 0.f),
    C_Vertex3f(0.504214287f, -0.188476563f, 0.f),
    C_Vertex3f(0.504214287f, -0.188476563f, -1.f),
    C_Normal3f(-0.0370273888f, -0.999314249f, 0.f),
    C_Vertex3f(0.466798782f, -0.188476563f, 0.f),
    C_Vertex3f(0.466798782f, -0.188476563f, -1.f),
    C_Normal3f(-0.101720825f, -0.994813025f, 0.f),
    C_Vertex3f(0.42017889f, -0.185362816f, 0.f),
    C_Vertex3f(0.42017889f, -0.185362816f, -1.f),
    C_Normal3f(-0.175006777f, -0.984567225f, 0.f),
    C_Vertex3f(0.375442982f, -0.179135323f, 0.f),
    C_Vertex3f(0.375442982f, -0.179135323f, -1.f),
    C_Normal3f(-0.246104777f, -0.969243288f, 0.f),
    C_Vertex3f(0.332591057f, -0.169794083f, 0.f),
    C_Vertex3f(0.332591057f, -0.169794083f, -1.f),
    C_Normal3f(-0.308354467f, -0.951271534f, 0.f),
    C_Vertex3f(0.301865101f, -0.160452843f, 0.f),
    C_Vertex3f(0.301865101f, -0.160452843f, -1.f),
    C_Normal3f(-0.357474327f, -0.933922946f, 0.f),
    C_Vertex3f(0.282139182f, -0.153440118f, 0.f),
    C_Vertex3f(0.282139182f, -0.153440118f, -1.f),
    C_Normal3f(-0.40321216f, -0.915106595f, 0.f),
    C_Vertex3f(0.263171315f, -0.145642161f, 0.f),
    C_Vertex3f(0.263171315f, -0.145642161f, -1.f),
    C_Normal3f(-0.449632883f, -0.893213451f, 0.f),
    C_Vertex3f(0.2449615f, -0.137058973f, 0.f),
    C_Vertex3f(0.2449615f, -0.137058973f, -1.f),
    C_Normal3f(-0.496337742f, -0.868129551f, 0.f),
    C_Vertex3f(0.227509737f, -0.127690554f, 0.f),
    C_Vertex3f(0.227509737f, -0.127690554f, -1.f),
    C_Normal3f(-0.542877734f, -0.839811802f, 0.f),
    C_Vertex3f(0.210816026f, -0.117536902f, 0.f),
    C_Vertex3f(0.210816026f, -0.117536902f, -1.f),
    C_Normal3f(-0.588769078f, -0.80830127f, 0.f),
    C_Vertex3f(0.194880366f, -0.10659802f, 0.f),
    C_Vertex3f(0.194880366f, -0.10659802f, -1.f),
    C_Normal3f(-0.633513272f, -0.773731709f, 0.f),
    C_Vertex3f(0.179702759f, -0.0948739052f, 0.f),
    C_Vertex3f(0.179702759f, -0.0948739052f, -1.f),
    C_Normal3f(-0.673819363f, -0.738896132f, 0.f),
    C_Vertex3f(0.165283203f, -0.0823645592f, 0.f),
    C_Vertex3f(0.165283203f, -0.0823645592f, -1.f),
    C_Normal3f(-0.710532069f, -0.703664839f, 0.f),
    C_Vertex3f(0.151751041f, -0.0693839788f, 0.f),
    C_Vertex3f(0.151751041f, -0.0693839788f, -1.f),
    C_Normal3f(-0.746475279f, -0.665413201f, 0.f),
    C_Vertex3f(0.139106274f, -0.0559321642f, 0.f),
    C_Vertex3f(0.139106274f, -0.0559321642f, -1.f),
    C_Normal3f(-0.781225443f, -0.624248922f, 0.f),
    C_Vertex3f(0.1273489f, -0.0420091152f, 0.f),
    C_Vertex3f(0.1273489f, -0.0420091152f, -1.f),
    C_Normal3f(-0.814355552f, -0.580366313f, 0.f),
    C_Vertex3f(0.11647892f, -0.0276148319f, 0.f),
    C_Vertex3f(0.11647892f, -0.0276148319f, -1.f),
    C_Normal3f(-0.845455289f, -0.534046233f, 0.f),
    C_Vertex3f(0.106496334f, -0.0127493143f, 0.f),
    C_Vertex3f(0.106496334f, -0.0127493143f, -1.f),
    C_Normal3f(-0.874153018f, -0.485650599f, 0.f),
    C_Vertex3f(0.0974011421f, 0.00258743763f, 0.f),
    C_Vertex3f(0.0974011421f, 0.00258743763f, -1.f),
    C_Normal3f(-0.90146625f, -0.432849288f, 0.f),
    C_Vertex3f(0.0891933441f, 0.0183954239f, 0.f),
    C_Vertex3f(0.0891933441f, 0.0183954239f, -1.f),
    C_Normal3f(-0.924649835f, -0.380818248f, 0.f),
    C_Vertex3f(0.080945015f, 0.0368595123f, 0.f),
    C_Vertex3f(0.080945015f, 0.0368595123f, -1.f),
    C_Normal3f(-0.943672717f, -0.330880374f, 0.f),
    C_Vertex3f(0.0729157925f, 0.0579183102f, 0.f),
    C_Vertex3f(0.0729157925f, 0.0579183102f, -1.f),
    C_Normal3f(-0.96015203f, -0.279478163f, 0.f),
    C_Vertex3f(0.0660336018f, 0.0793869495f, 0.f),
    C_Vertex3f(0.0660336018f, 0.0793869495f, -1.f),
    C_Normal3f(-0.973754823f, -0.227599666f, 0.f),
    C_Vertex3f(0.0602984428f, 0.10126543f, 0.f),
    C_Vertex3f(0.0602984428f, 0.10126543f, -1.f),
    C_Normal3f(-0.984442949f, -0.175704136f, 0.f),
    C_Vertex3f(0.0557103157f, 0.123553753f, 0.f),
    C_Vertex3f(0.0557103157f, 0.123553753f, -1.f),
    C_Normal3f(-0.992252827f, -0.124234959f, 0.f),
    C_Vertex3f(0.0522692204f, 0.146251917f, 0.f),
    C_Vertex3f(0.0522692204f, 0.146251917f, -1.f),
    C_Normal3f(-0.997287631f, -0.073602125f, 0.f),
    C_Vertex3f(0.0499751568f, 0.169359922f, 0.f),
    C_Vertex3f(0.0499751568f, 0.169359922f, -1.f),
    C_Normal3f(-0.999723971f, -0.0234933309f, 0.f),
    C_Vertex3f(0.048828125f, 0.192877769f, 0.f),
    C_Vertex3f(0.048828125f, 0.192877769f, -1.f),
    C_Normal3f(-0.999650598f, 0.0264342558f, 0.f),
    C_Vertex3f(0.048828125f, 0.218170166f, 0.f),
    C_Vertex3f(0.048828125f, 0.218170166f, -1.f),
    C_Normal3f(-0.996955216f, 0.0779764056f, 0.f),
    C_Vertex3f(0.0501935482f, 0.244513273f, 0.f),
    C_Vertex3f(0.0501935482f, 0.244513273f, -1.f),
    C_Normal3f(-0.99140805f, 0.130806059f, 0.f),
    C_Vertex3f(0.0529243946f, 0.270542383f, 0.f),
    C_Vertex3f(0.0529243946f, 0.270542383f, -1.f),
    C_Normal3f(-0.9829638f, 0.183799371f, 0.f),
    C_Vertex3f(0.0570206642f, 0.296257496f, 0.f),
    C_Vertex3f(0.0570206642f, 0.296257496f, -1.f),
    C_Normal3f(-0.971632719f, 0.236494884f, 0.f),
    C_Vertex3f(0.062482357f, 0.321658611f, 0.f),
    C_Vertex3f(0.062482357f, 0.321658611f, -1.f),
    C_Normal3f(-0.95750016f, 0.288432688f, 0.f),
    C_Vertex3f(0.069309473f, 0.346745729f, 0.f),
    C_Vertex3f(0.069309473f, 0.346745729f, -1.f),
    C_Normal3f(-0.940723538f, 0.33917436f, 0.f),
    C_Vertex3f(0.0775020123f, 0.37151885f, 0.f),
    C_Vertex3f(0.0775020123f, 0.37151885f, -1.f),
    C_Normal3f(-0.921198428f, 0.38909325f, 0.f),
    C_Vertex3f(0.0870599747f, 0.395977974f, 0.f),
    C_Vertex3f(0.0870599747f, 0.395977974f, -1.f),
    C_Normal3f(-0.897836387f, 0.440329343f, 0.f),
    C_Vertex3f(0.0991849899f, 0.422854424f, 0.f),
    C_Vertex3f(0.0991849899f, 0.422854424f, -1.f),
    C_Normal3f(-0.868328989f, 0.495988667f, 0.f),
    C_Vertex3f(0.113911152f, 0.45072782f, 0.f),
    C_Vertex3f(0.113911152f, 0.45072782f, -1.f),
    C_Normal3f(-0.8323915f, 0.554188192f, 0.f),
    C_Vertex3f(0.130036831f, 0.476866841f, 0.f),
    C_Vertex3f(0.130036831f, 0.476866841f, -1.f),
    C_Normal3f(-0.790684581f, 0.612223804f, 0.f),
    C_Vertex3f(0.147562027f, 0.501271486f, 0.f),
    C_Vertex3f(0.147562027f, 0.501271486f, -1.f),
    C_Normal3f(-0.74326539f, 0.668996751f, 0.f),
    C_Vertex3f(0.16648674f, 0.523941755f, 0.f),
    C_Vertex3f(0.16648674f, 0.523941755f, -1.f),
    C_Normal3f(-0.69048208f, 0.723349452f, 0.f),
    C_Vertex3f(0.18681097f, 0.544877648f, 0.f),
    C_Vertex3f(0.18681097f, 0.544877648f, -1.f),
    C_Normal3f(-0.63298595f, 0.774163246f, 0.f),
    C_Vertex3f(0.208534718f, 0.564079165f, 0.f),
    C_Vertex3f(0.208534718f, 0.564079165f, -1.f),
    C_Normal3f(-0.571700573f, 0.820462346f, 0.f),
    C_Vertex3f(0.231657982f, 0.581546307f, 0.f),
    C_Vertex3f(0.231657982f, 0.581546307f, -1.f),
    C_Normal3f(-0.505510092f, 0.862820625f, 0.f),
    C_Vertex3f(0.256180763f, 0.597279072f, 0.f),
    C_Vertex3f(0.256180763f, 0.597279072f, -1.f),
    C_Normal3f(-0.434746414f, 0.900552988f, 0.f),
    C_Vertex3f(0.282007694f, 0.611045241f, 0.f),
    C_Vertex3f(0.282007694f, 0.611045241f, -1.f),
    C_Normal3f(-0.362787575f, 0.931871831f, 0.f),
    C_Vertex3f(0.309138775f, 0.622844815f, 0.f),
    C_Vertex3f(0.309138775f, 0.622844815f, -1.f),
    C_Normal3f(-0.291071981f, 0.956701219f, 0.f),
    C_Vertex3f(0.337574005f, 0.632677794f, 0.f),
    C_Vertex3f(0.337574005f, 0.632677794f, -1.f),
    C_Normal3f(-0.220886722f, 0.975299478f, 0.f),
    C_Vertex3f(0.367313385f, 0.640544176f, 0.f),
    C_Vertex3f(0.367313385f, 0.640544176f, -1.f),
    C_Normal3f(-0.153282717f, 0.988182425f, 0.f),
    C_Vertex3f(0.398356915f, 0.646443963f, 0.f),
    C_Vertex3f(0.398356915f, 0.646443963f, -1.f),
    C_Normal3f(-0.0890363529f, 0.996028364f, 0.f),
    C_Vertex3f(0.430704594f, 0.650377154f, 0.f),
    C_Vertex3f(0.430704594f, 0.650377154f, -1.f),
    C_Normal3f(-0.0303933751f, 0.999538004f, 0.f),
    C_Vertex3f(0.464356422f, 0.65234375f, 0.f),
    C_Vertex3f(0.464356422f, 0.65234375f, -1.f),
    C_Normal3f(0.0241343491f, 0.999708712f, 0.f),
    C_Vertex3f(0.495379448f, 0.65234375f, 0.f),
    C_Vertex3f(0.495379448f, 0.65234375f, -1.f),
    C_Normal3f(0.0795537084f, 0.996830523f, 0.f),
    C_Vertex3f(0.521780014f, 0.650957465f, 0.f),
    C_Vertex3f(0.521780014f, 0.650957465f, -1.f),
    C_Normal3f(0.135368988f, 0.990795255f, 0.f),
    C_Vertex3f(0.547491074f, 0.648184896f, 0.f),
    C_Vertex3f(0.547491074f, 0.648184896f, -1.f),
    C_Normal3f(0.192927867f, 0.981212974f, 0.f),
    C_Vertex3f(0.572512627f, 0.644026041f, 0.f),
    C_Vertex3f(0.572512627f, 0.644026041f, -1.f),
    C_Normal3f(0.251693755f, 0.967806876f, 0.f),
    C_Vertex3f(0.596844673f, 0.638480902f, 0.f),
    C_Vertex3f(0.596844673f, 0.638480902f, -1.f),
    C_Normal3f(0.31103304f, 0.950399101f, 0.f),
    C_Vertex3f(0.620487213f, 0.631549478f, 0.f),
    C_Vertex3f(0.620487213f, 0.631549478f, -1.f),
    C_Normal3f(0.370240599f, 0.928935885f, 0.f),
    C_Vertex3f(0.643440247f, 0.623231769f, 0.f),
    C_Vertex3f(0.643440247f, 0.623231769f, -1.f),
    C_Normal3f(0.428576469f, 0.903505564f, 0.f),
    C_Vertex3f(0.665703773f, 0.613527775f, 0.f),
    C_Vertex3f(0.665703773f, 0.613527775f, -1.f),
    C_Normal3f(0.488844037f, 0.872371197f, 0.f),
    C_Vertex3f(0.687277794f, 0.602437496f, 0.f),
    C_Vertex3f(0.687277794f, 0.602437496f, -1.f),
    C_Normal3f(0.551392078f, 0.834246278f, 0.f),
    C_Vertex3f(0.707711697f, 0.589988112f, 0.f),
    C_Vertex3f(0.707711697f, 0.589988112f, -1.f),
    C_Normal3f(0.611968577f, 0.790882111f, 0.f),
    C_Vertex3f(0.727005482f, 0.576179624f, 0.f),
    C_Vertex3f(0.727005482f, 0.576179624f, -1.f),
    C_Normal3f(0.669475675f, 0.742833972f, 0.f),
    C_Vertex3f(0.745159149f, 0.56101203f, 0.f),
    C_Vertex3f(0.745159149f, 0.56101203f, -1.f),
    C_Normal3f(0.722949505f, 0.690900862f, 0.f),
    C_Vertex3f(0.762172699f, 0.544485331f, 0.f),
    C_Vertex3f(0.762172699f, 0.544485331f, -1.f),
    C_Normal3f(0.771639407f, 0.636060297f, 0.f),
    C_Vertex3f(0.778046131f, 0.526599526f, 0.f),
    C_Vertex3f(0.778046131f, 0.526599526f, -1.f),
    C_Normal3f(0.815056741f, 0.579381108f, 0.f),
    C_Vertex3f(0.792779446f, 0.507354617f, 0.f),
    C_Vertex3f(0.792779446f, 0.507354617f, -1.f),
    C_Normal3f(0.852921307f, 0.522039473f, 0.f),
    C_Vertex3f(0.806372643f, 0.486750603f, 0.f),
    C_Vertex3f(0.806372643f, 0.486750603f, -1.f),
    C_Normal3f(0.886058152f, 0.463573962f, 0.f),
    C_Vertex3f(0.817896843f, 0.466317177f, 0.f),
    C_Vertex3f(0.817896843f, 0.466317177f, -1.f),
    C_Normal3f(0.914959371f, 0.403545946f, 0.f),
    C_Vertex3f(0.827167749f, 0.447003603f, 0.f),
    C_Vertex3f(0.827167749f, 0.447003603f, -1.f),
    C_Normal3f(0.939222753f, 0.343308389f, 0.f),
    C_Vertex3f(0.835114241f, 0.427280188f, 0.f),
    C_Vertex3f(0.835114241f, 0.427280188f, -1.f),
    C_Normal3f(0.959645331f, 0.281212986f, 0.f),
    C_Vertex3f(0.841736317f, 0.407146931f, 0.f),
    C_Vertex3f(0.841736317f, 0.407146931f, -1.f),
    C_Normal3f(0.975939512f, 0.218041152f, 0.f),
    C_Vertex3f(0.847033978f, 0.386603832f, 0.f),
    C_Vertex3f(0.847033978f, 0.386603832f, -1.f),
    C_Normal3f(0.987975478f, 0.154610366f, 0.f),
    C_Vertex3f(0.851007223f, 0.365650892f, 0.f),
    C_Vertex3f(0.851007223f, 0.365650892f, -1.f),
    C_Normal3f(0.99578464f, 0.0917227268f, 0.f),
    C_Vertex3f(0.853656054f, 0.344288111f, 0.f),
    C_Vertex3f(0.853656054f, 0.344288111f, -1.f),
    C_Normal3f(0.99963057f, 0.027181292f, 0.f),
    C_Vertex3f(0.854980469f, 0.322515488f, 0.f),
    C_Vertex3f(0.854980469f, 0.322515488f, -1.f),
    C_Normal3f(0.998836219f, -0.0482314564f, 0.f),
    C_Vertex3f(0.854980469f, 0.295580864f, 0.f),
    C_Vertex3f(0.854980469f, 0.295580864f, -1.f),
    C_Normal3f(0.990644097f, -0.136470869f, 0.f),
    C_Vertex3f(0.85219264f, 0.264781713f, 0.f),
    C_Vertex3f(0.85219264f, 0.264781713f, -1.f),
    C_Normal3f(0.973151565f, -0.230165526f, 0.f),
    C_Vertex3f(0.846616983f, 0.234870195f, 0.f),
    C_Vertex3f(0.846616983f, 0.234870195f, -1.f),
    C_Normal3f(0.946366608f, -0.323094964f, 0.f),
    C_Vertex3f(0.838253498f, 0.20584631f, 0.f),
    C_Vertex3f(0.838253498f, 0.20584631f, -1.f),
    C_Normal3f(0.910889149f, -0.412651032f, 0.f),
    C_Vertex3f(0.827102184f, 0.177710056f, 0.f),
    C_Vertex3f(0.827102184f, 0.177710056f, -1.f),
    C_Normal3f(0.868019462f, -0.496530056f, 0.f),
    C_Vertex3f(0.813163042f, 0.150461435f, 0.f),
    C_Vertex3f(0.813163042f, 0.150461435f, -1.f),
    C_Normal3f(0.819545746f, -0.573013663f, 0.f),
    C_Vertex3f(0.796436071f, 0.124100447f, 0.f),
    C_Vertex3f(0.796436071f, 0.124100447f, -1.f),
    C_Normal3f(0.768636346f, -0.639685988f, 0.f),
    C_Vertex3f(0.776921272f, 0.0986270905f, 0.f),
    C_Vertex3f(0.776921272f, 0.0986270905f, -1.f),
    C_Normal3f(0.716952682f, -0.697121918f, 0.f),
    C_Vertex3f(0.755821228f, 0.0752983093f, 0.f),
    C_Vertex3f(0.755821228f, 0.0752983093f, -1.f),
    C_Normal3f(0.656702578f, -0.754149795f, 0.f),
    C_Vertex3f(0.735459089f, 0.0559854507f, 0.f),
    C_Vertex3f(0.735459089f, 0.0559854507f, -1.f),
    C_Normal3f(0.584604025f, -0.811318755f, 0.f),
    C_Vertex3f(0.714632273f, 0.039431572f, 0.f),
    C_Vertex3f(0.714632273f, 0.039431572f, -1.f),
    C_Normal3f(0.499656409f, -0.866223693f, 0.f),
    C_Vertex3f(0.693340778f, 0.025636673f, 0.f),
    C_Vertex3f(0.693340778f, 0.025636673f, -1.f),
    C_Normal3f(0.402092576f, -0.915598989f, 0.f),
    C_Vertex3f(0.671584606f, 0.0146007538f, 0.f),
    C_Vertex3f(0.671584606f, 0.0146007538f, -1.f),
    C_Normal3f(0.293649286f, -0.955913186f, 0.f),
    C_Vertex3f(0.649363756f, 0.00632381439f, 0.f),
    C_Vertex3f(0.649363756f, 0.00632381439f, -1.f),
    C_Normal3f(0.17770423f, -0.98408401f, 0.f),
    C_Vertex3f(0.626678228f, 0.000805854797f, 0.f),
    C_Vertex3f(0.626678228f, 0.000805854797f, -1.f),
    C_Normal3f(0.0647789761f, -0.997899652f, 0.f),
    C_Vertex3f(0.603528023f, -0.001953125f, 0.f),
    C_Vertex3f(0.603528023f, -0.001953125f, -1.f),
    C_Normal3f(-0.0347846746f, -0.999394894f, 0.f),
    C_Vertex3f(0.584177017f, -0.001953125f, 0.f),
    C_Vertex3f(0.584177017f, -0.001953125f, -1.f),
    C_Normal3f(-0.132197276f, -0.991223395f, 0.f),
    C_Vertex3f(0.570538521f, -0.000804901123f, 0.f),
    C_Vertex3f(0.570538521f, -0.000804901123f, -1.f),
    C_Normal3f(-0.242867857f, -0.970059395f, 0.f),
    C_Vertex3f(0.558348656f, 0.00149154663f, 0.f),
    C_Vertex3f(0.558348656f, 0.00149154663f, -1.f),
    C_Normal3f(-0.372350246f, -0.928092241f, 0.f),
    C_Vertex3f(0.547607422f, 0.00493621826f, 0.f),
    C_Vertex3f(0.547607422f, 0.00493621826f, -1.f),
    C_Normal3f(-0.50558567f, -0.862776399f, 0.f),
    C_Vertex3f(0.538314819f, 0.00952911377f, 0.f),
    C_Vertex3f(0.538314819f, 0.00952911377f, -1.f),
    C_Normal3f(-0.631739259f, -0.775180995f, 0.f),
    C_Vertex3f(0.530252457f, 0.0151062012f, 0.f),
    C_Vertex3f(0.530252457f, 0.0151062012f, -1.f),
    C_Normal3f(-0.750184119f, -0.661229074f, 0.f),
    C_Vertex3f(0.523420334f, 0.0216674805f, 0.f),
    C_Vertex3f(0.523420334f, 0.0216674805f, -1.f),
    C_Normal3f(-0.844877541f, -0.534959793f, 0.f),
    C_Vertex3f(0.517818451f, 0.0292129517f, 0.f),
    C_Vertex3f(0.517818451f, 0.0292129517f, -1.f),
    C_Normal3f(-0.921879351f, -0.387476861f, 0.f),
    C_Vertex3f(0.514211655f, 0.0362110138f, 0.f),
    C_Vertex3f(0.514211655f, 0.0362110138f, -1.f),
    C_Normal3f(-0.965568244f, -0.260149986f, 0.f),
    C_Vertex3f(0.511752129f, 0.0436458588f, 0.f),
    C_Vertex3f(0.511752129f, 0.0436458588f, -1.f),
    C_Normal3f(-0.983915329f, -0.178635016f, 0.f),
    C_Vertex3f(0.509675026f, 0.0530490875f, 0.f),
    C_Vertex3f(0.509675026f, 0.0530490875f, -1.f),
    C_Normal3f(-0.99161762f, -0.12920697f, 0.f),
    C_Vertex3f(0.507980347f, 0.0644207001f, 0.f),
    C_Vertex3f(0.507980347f, 0.0644207001f, -1.f),
    C_Normal3f(-0.995196462f, -0.0978975073f, 0.f),
    C_Vertex3f(0.507324219f, 0.0710906982f, 0.f),
    C_Vertex3f(0.507324219f, 0.0710906982f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.998869956f, -0.0475256294f, 0.f),
    C_Vertex3f(0.296386719f, 0.198862076f, 0.f),
    C_Vertex3f(0.296386719f, 0.198862076f, -1.f),
    C_Normal3f(0.999593735f, 0.0285042413f, 0.f),
    C_Vertex3f(0.296386719f, 0.181959152f, 0.f),
    C_Vertex3f(0.296386719f, 0.181959152f, -1.f),
    C_Normal3f(0.9953565f, 0.0962566957f, 0.f),
    C_Vertex3f(0.297272325f, 0.167805433f, 0.f),
    C_Vertex3f(0.297272325f, 0.167805433f, -1.f),
    C_Normal3f(0.985594511f, 0.169125438f, 0.f),
    C_Vertex3f(0.299043536f, 0.154485941f, 0.f),
    C_Vertex3f(0.299043536f, 0.154485941f, -1.f),
    C_Normal3f(0.968562961f, 0.248768613f, 0.f),
    C_Vertex3f(0.301700354f, 0.142000675f, 0.f),
    C_Vertex3f(0.301700354f, 0.142000675f, -1.f),
    C_Normal3f(0.942454398f, 0.33433494f, 0.f),
    C_Vertex3f(0.305242777f, 0.130349636f, 0.f),
    C_Vertex3f(0.305242777f, 0.130349636f, -1.f),
    C_Normal3f(0.905593693f, 0.424146384f, 0.f),
    C_Vertex3f(0.309670806f, 0.119532824f, 0.f),
    C_Vertex3f(0.309670806f, 0.119532824f, -1.f),
    C_Normal3f(0.856814265f, 0.515625119f, 0.f),
    C_Vertex3f(0.314984441f, 0.109550238f, 0.f),
    C_Vertex3f(0.314984441f, 0.109550238f, -1.f),
    C_Normal3f(0.795885623f, 0.605446994f, 0.f),
    C_Vertex3f(0.321183681f, 0.100401878f, 0.f),
    C_Vertex3f(0.321183681f, 0.100401878f, -1.f),
    C_Normal3f(0.733895838f, 0.679261923f, 0.f),
    C_Vertex3f(0.328268528f, 0.0920877457f, 0.f),
    C_Vertex3f(0.328268528f, 0.0920877457f, -1.f),
    C_Normal3f(0.670580268f, 0.741836905f, 0.f),
    C_Vertex3f(0.335612178f, 0.0848128796f, 0.f),
    C_Vertex3f(0.335612178f, 0.0848128796f, -1.f),
    C_Normal3f(0.59446615f, 0.8041206f, 0.f),
    C_Vertex3f(0.343214631f, 0.07857728f, 0.f),
    C_Vertex3f(0.343214631f, 0.07857728f, -1.f),
    C_Normal3f(0.505118191f, 0.863050222f, 0.f),
    C_Vertex3f(0.351075888f, 0.0733809471f, 0.f),
    C_Vertex3f(0.351075888f, 0.0733809471f, -1.f),
    C_Normal3f(0.403451204f, 0.915001154f, 0.f),
    C_Vertex3f(0.359195948f, 0.0692238808f, 0.f),
    C_Vertex3f(0.359195948f, 0.0692238808f, -1.f),
    C_Normal3f(0.292055964f, 0.956401169f, 0.f),
    C_Vertex3f(0.367574811f, 0.066106081f, 0.f),
    C_Vertex3f(0.367574811f, 0.066106081f, -1.f),
    C_Normal3f(0.175067067f, 0.984556437f, 0.f),
    C_Vertex3f(0.376212478f, 0.0640275478f, 0.f),
    C_Vertex3f(0.376212478f, 0.0640275478f, -1.f),
    C_Normal3f(0.0530046299f, 0.998594344f, 0.f),
    C_Vertex3f(0.385108948f, 0.0629882813f, 0.f),
    C_Vertex3f(0.385108948f, 0.0629882813f, -1.f),
    C_Normal3f(-0.0786735713f, 0.996900439f, 0.f),
    C_Vertex3f(0.395792007f, 0.0629882813f, 0.f),
    C_Vertex3f(0.395792007f, 0.0629882813f, -1.f),
    C_Normal3f(-0.211536035f, 0.977370203f, 0.f),
    C_Vertex3f(0.40832901f, 0.0648207664f, 0.f),
    C_Vertex3f(0.40832901f, 0.0648207664f, -1.f),
    C_Normal3f(-0.331771344f, 0.943359792f, 0.f),
    C_Vertex3f(0.421192169f, 0.0684857368f, 0.f),
    C_Vertex3f(0.421192169f, 0.0684857368f, -1.f),
    C_Normal3f(-0.432980657f, 0.901403248f, 0.f),
    C_Vertex3f(0.434381485f, 0.0739831924f, 0.f),
    C_Vertex3f(0.434381485f, 0.0739831924f, -1.f),
    C_Normal3f(-0.523812532f, 0.851833582f, 0.f),
    C_Vertex3f(0.447896957f, 0.0813131332f, 0.f),
    C_Vertex3f(0.447896957f, 0.0813131332f, -1.f),
    C_Normal3f(-0.607645392f, 0.794208527f, 0.f),
    C_Vertex3f(0.461112499f, 0.090420723f, 0.f),
    C_Vertex3f(0.461112499f, 0.090420723f, -1.f),
    C_Normal3f(-0.677981257f, 0.735079169f, 0.f),
    C_Vertex3f(0.474028111f, 0.101305962f, 0.f),
    C_Vertex3f(0.474028111f, 0.101305962f, -1.f),
    C_Normal3f(-0.735981345f, 0.677001834f, 0.f),
    C_Vertex3f(0.486643791f, 0.113968849f, 0.f),
    C_Vertex3f(0.486643791f, 0.113968849f, -1.f),
    C_Normal3f(-0.79504925f, 0.606544852f, 0.f),
    C_Vertex3f(0.498959541f, 0.128409386f, 0.f),
    C_Vertex3f(0.498959541f, 0.128409386f, -1.f),
    C_Normal3f(-0.852494478f, 0.522736132f, 0.f),
    C_Vertex3f(0.510158539f, 0.144791603f, 0.f),
    C_Vertex3f(0.510158539f, 0.144791603f, -1.f),
    C_Normal3f(-0.89671123f, 0.442615986f, 0.f),
    C_Vertex3f(0.520240784f, 0.163115501f, 0.f),
    C_Vertex3f(0.520240784f, 0.163115501f, -1.f),
    C_Normal3f(-0.929791093f, 0.36808756f, 0.f),
    C_Vertex3f(0.529206276f, 0.183381081f, 0.f),
    C_Vertex3f(0.529206276f, 0.183381081f, -1.f),
    C_Normal3f(-0.955411851f, 0.295276314f, 0.f),
    C_Vertex3f(0.537055016f, 0.205588341f, 0.f),
    C_Vertex3f(0.537055016f, 0.205588341f, -1.f),
    C_Normal3f(-0.976546943f, 0.215304747f, 0.f),
    C_Vertex3f(0.54294157f, 0.227823734f, 0.f),
    C_Vertex3f(0.54294157f, 0.227823734f, -1.f),
    C_Normal3f(-0.991384983f, 0.130980074f, 0.f),
    C_Vertex3f(0.54686594f, 0.250087261f, 0.f),
    C_Vertex3f(0.54686594f, 0.250087261f, -1.f),
    C_Normal3f(-0.998850584f, 0.0479315035f, 0.f),
    C_Vertex3f(0.548828125f, 0.272378922f, 0.f),
    C_Vertex3f(0.548828125f, 0.272378922f, -1.f),
    C_Normal3f(-0.999603331f, -0.0281642675f, 0.f),
    C_Vertex3f(0.548828125f, 0.290977478f, 0.f),
    C_Vertex3f(0.548828125f, 0.290977478f, -1.f),
    C_Normal3f(-0.994914234f, -0.100726515f, 0.f),
    C_Vertex3f(0.547908425f, 0.305020809f, 0.f),
    C_Vertex3f(0.547908425f, 0.305020809f, -1.f),
    C_Normal3f(-0.984228492f, -0.176901564f, 0.f),
    C_Vertex3f(0.546069026f, 0.318230152f, 0.f),
    C_Vertex3f(0.546069026f, 0.318230152f, -1.f),
    C_Normal3f(-0.965628147f, -0.259927571f, 0.f),
    C_Vertex3f(0.543309927f, 0.330605507f, 0.f),
    C_Vertex3f(0.543309927f, 0.330605507f, -1.f),
    C_Normal3f(-0.937239826f, -0.348685414f, 0.f),
    C_Vertex3f(0.539631128f, 0.342146873f, 0.f),
    C_Vertex3f(0.539631128f, 0.342146873f, -1.f),
    C_Normal3f(-0.897436202f, -0.441144317f, 0.f),
    C_Vertex3f(0.53503263f, 0.352854252f, 0.f),
    C_Vertex3f(0.53503263f, 0.352854252f, -1.f),
    C_Normal3f(-0.845263362f, -0.534349978f, 0.f),
    C_Vertex3f(0.529514432f, 0.362727642f, 0.f),
    C_Vertex3f(0.529514432f, 0.362727642f, -1.f),
    C_Normal3f(-0.780876756f, -0.624685228f, 0.f),
    C_Vertex3f(0.523076534f, 0.371767044f, 0.f),
    C_Vertex3f(0.523076534f, 0.371767044f, -1.f),
    C_Normal3f(-0.713294387f, -0.700864553f, 0.f),
    C_Vertex3f(0.515718937f, 0.379972458f, 0.f),
    C_Vertex3f(0.515718937f, 0.379972458f, -1.f),
    C_Normal3f(-0.64207226f, -0.76664412f, 0.f),
    C_Vertex3f(0.507959485f, 0.387152195f, 0.f),
    C_Vertex3f(0.507959485f, 0.387152195f, -1.f),
    C_Normal3f(-0.559249043f, -0.828999698f, 0.f),
    C_Vertex3f(0.499798179f, 0.393306255f, 0.f),
    C_Vertex3f(0.499798179f, 0.393306255f, -1.f),
    C_Normal3f(-0.465973079f, -0.884798884f, 0.f),
    C_Vertex3f(0.491235018f, 0.398434639f, 0.f),
    C_Vertex3f(0.491235018f, 0.398434639f, -1.f),
    C_Normal3f(-0.364680827f, -0.931132555f, 0.f),
    C_Vertex3f(0.482270002f, 0.402537346f, 0.f),
    C_Vertex3f(0.482270002f, 0.402537346f, -1.f),
    C_Normal3f(-0.258866906f, -0.965912938f, 0.f),
    C_Vertex3f(0.472903132f, 0.405614376f, 0.f),
    C_Vertex3f(0.472903132f, 0.405614376f, -1.f),
    C_Normal3f(-0.152514488f, -0.988301218f, 0.f),
    C_Vertex3f(0.463134408f, 0.40766573f, 0.f),
    C_Vertex3f(0.463134408f, 0.40766573f, -1.f),
    C_Normal3f(-0.0456761569f, -0.998956263f, 0.f),
    C_Vertex3f(0.452963829f, 0.408691406f, 0.f),
    C_Vertex3f(0.452963829f, 0.408691406f, -1.f),
    C_Normal3f(0.0687286928f, -0.997635365f, 0.f),
    C_Vertex3f(0.440702438f, 0.408691406f, 0.f),
    C_Vertex3f(0.440702438f, 0.408691406f, -1.f),
    C_Normal3f(0.196035236f, -0.98059684f, 0.f),
    C_Vertex3f(0.427160263f, 0.406913757f, 0.f),
    C_Vertex3f(0.427160263f, 0.406913757f, -1.f),
    C_Normal3f(0.325044692f, -0.945698738f, 0.f),
    C_Vertex3f(0.41402626f, 0.403358459f, 0.f),
    C_Vertex3f(0.41402626f, 0.403358459f, -1.f),
    C_Normal3f(0.444975019f, -0.89554292f, 0.f),
    C_Vertex3f(0.40130043f, 0.398025513f, 0.f),
    C_Vertex3f(0.40130043f, 0.398025513f, -1.f),
    C_Normal3f(0.556717873f, -0.83070159f, 0.f),
    C_Vertex3f(0.388982773f, 0.390914917f, 0.f),
    C_Vertex3f(0.388982773f, 0.390914917f, -1.f),
    C_Normal3f(0.65222019f, -0.758029461f, 0.f),
    C_Vertex3f(0.37685585f, 0.381643295f, 0.f),
    C_Vertex3f(0.37685585f, 0.381643295f, -1.f),
    C_Normal3f(0.726351142f, -0.687323809f, 0.f),
    C_Vertex3f(0.364919662f, 0.370210648f, 0.f),
    C_Vertex3f(0.364919662f, 0.370210648f, -1.f),
    C_Normal3f(0.783187389f, -0.62178576f, 0.f),
    C_Vertex3f(0.35317421f, 0.356616974f, 0.f),
    C_Vertex3f(0.35317421f, 0.356616974f, -1.f),
    C_Normal3f(0.834036827f, -0.551708758f, 0.f),
    C_Vertex3f(0.341619492f, 0.340862274f, 0.f),
    C_Vertex3f(0.341619492f, 0.340862274f, -1.f),
    C_Normal3f(0.880773842f, -0.473537147f, 0.f),
    C_Vertex3f(0.33120966f, 0.323412418f, 0.f),
    C_Vertex3f(0.33120966f, 0.323412418f, -1.f),
    C_Normal3f(0.917068839f, -0.398728907f, 0.f),
    C_Vertex3f(0.321944714f, 0.304267406f, 0.f),
    C_Vertex3f(0.321944714f, 0.304267406f, -1.f),
    C_Normal3f(0.944442034f, -0.328678071f, 0.f),
    C_Vertex3f(0.313824654f, 0.283427238f, 0.f),
    C_Vertex3f(0.313824654f, 0.283427238f, -1.f),
    C_Normal3f(0.963826537f, -0.266530514f, 0.f),
    C_Vertex3f(0.30684948f, 0.260891914f, 0.f),
    C_Vertex3f(0.30684948f, 0.260891914f, -1.f),
    C_Normal3f(0.979394197f, -0.201958016f, 0.f),
    C_Vertex3f(0.301618099f, 0.239285946f, 0.f),
    C_Vertex3f(0.301618099f, 0.239285946f, -1.f),
    C_Normal3f(0.991729796f, -0.128342882f, 0.f),
    C_Vertex3f(0.298130512f, 0.218609333f, 0.f),
    C_Vertex3f(0.298130512f, 0.218609333f, -1.f),
    C_Normal3f(0.998869956f, -0.0475256294f, 0.f),
    C_Vertex3f(0.296386719f, 0.198862076f, 0.f),
    C_Vertex3f(0.296386719f, 0.198862076f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.876464844f, 0.00830078125f, 0.f),
    C_Vertex3f(0.797363281f, 0.00830078125f, 0.f),
    C_Vertex3f(0.806768417f, -0.0793194771f, 0.f),
    C_Vertex3f(0.819317579f, -0.0674580336f, 0.f),
    C_Vertex3f(0.830869913f, -0.0556513071f, 0.f),
    C_Vertex3f(0.841425419f, -0.0438992977f, 0.f),
    C_Vertex3f(0.850984097f, -0.0322020054f, 0.f),
    C_Vertex3f(0.859545946f, -0.0205594301f, 0.f),
    C_Vertex3f(0.867110968f, -0.00897157192f, 0.f),
    C_Vertex3f(0.873679161f, 0.00256156921f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.792554855f, 0.0013999939f, 0.f),
    C_Vertex3f(0.781722307f, -0.0118813515f, 0.f),
    C_Vertex3f(0.786449432f, -0.097193718f, 0.f),
    C_Vertex3f(0.806768417f, -0.0793194771f, 0.f),
    C_Vertex3f(0.797363281f, 0.00830078125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.769674063f, -0.0246424675f, 0.f),
    C_Vertex3f(0.756410122f, -0.0368833542f, 0.f),
    C_Vertex3f(0.756789684f, -0.118538857f, 0.f),
    C_Vertex3f(0.786449432f, -0.097193718f, 0.f),
    C_Vertex3f(0.781722307f, -0.0118813515f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.756410122f, -0.0368833542f, 0.f),
    C_Vertex3f(0.741930485f, -0.0486040115f, 0.f),
    C_Vertex3f(0.756789684f, -0.118538857f, 0.f),
    C_Vertex3f(0.726235151f, -0.0598044395f, 0.f),
    C_Vertex3f(0.724562168f, -0.137396812f, 0.f),
    C_Vertex3f(0.709324121f, -0.0704846382f, 0.f),
    C_Vertex3f(0.691197395f, -0.0806446075f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.504214287f, -0.188476563f, 0.f),
    C_Vertex3f(0.529585719f, -0.187608838f, 0.f),
    C_Vertex3f(0.531422377f, -0.12281847f, 0.f),
    C_Vertex3f(0.554001451f, -0.185873389f, 0.f),
    C_Vertex3f(0.557729959f, -0.120408535f, 0.f),
    C_Vertex3f(0.577461481f, -0.183270216f, 0.f),
    C_Vertex3f(0.582876682f, -0.116793633f, 0.f),
    C_Vertex3f(0.599965811f, -0.179799318f, 0.f),
    C_Vertex3f(0.606862545f, -0.111973763f, 0.f),
    C_Vertex3f(0.62151444f, -0.175460696f, 0.f),
    C_Vertex3f(0.629687548f, -0.105948925f, 0.f),
    C_Vertex3f(0.642107368f, -0.17025435f, 0.f),
    C_Vertex3f(0.65135169f, -0.09871912f, 0.f),
    C_Vertex3f(0.661744595f, -0.164180279f, 0.f),
    C_Vertex3f(0.671854973f, -0.0902843475f, 0.f),
    C_Vertex3f(0.689766884f, -0.153767586f, 0.f),
    C_Vertex3f(0.691197395f, -0.0806446075f, 0.f),
    C_Vertex3f(0.724562168f, -0.137396812f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.531422377f, -0.12281847f, 0.f),
    C_Vertex3f(0.503953934f, -0.124023438f, 0.f),
    C_Vertex3f(0.504214287f, -0.188476563f, 0.f),
    C_Vertex3f(0.476252556f, -0.124023438f, 0.f),
    C_Vertex3f(0.466798782f, -0.188476563f, 0.f),
    C_Vertex3f(0.450073242f, -0.122869372f, 0.f),
    C_Vertex3f(0.424488068f, -0.120561242f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466798782f, -0.188476563f, 0.f),
    C_Vertex3f(0.424488068f, -0.120561242f, 0.f),
    C_Vertex3f(0.42017889f, -0.185362816f, 0.f),
    C_Vertex3f(0.399497032f, -0.117099047f, 0.f),
    C_Vertex3f(0.375442982f, -0.179135323f, 0.f),
    C_Vertex3f(0.375100136f, -0.112482786f, 0.f),
    C_Vertex3f(0.351297379f, -0.106712461f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.332591057f, -0.169794083f, 0.f),
    C_Vertex3f(0.375442982f, -0.179135323f, 0.f),
    C_Vertex3f(0.351297379f, -0.106712461f, 0.f),
    C_Vertex3f(0.32808876f, -0.0997880697f, 0.f),
    C_Vertex3f(0.305474281f, -0.0917096138f, 0.f),
    C_Vertex3f(0.301865101f, -0.160452843f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.283453941f, -0.0824770927f, 0.f),
    C_Vertex3f(0.262731433f, -0.0719810724f, 0.f),
    C_Vertex3f(0.263171315f, -0.145642161f, 0.f),
    C_Vertex3f(0.282139182f, -0.153440118f, 0.f),
    C_Vertex3f(0.301865101f, -0.160452843f, 0.f),
    C_Vertex3f(0.305474281f, -0.0917096138f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.263171315f, -0.145642161f, 0.f),
    C_Vertex3f(0.262731433f, -0.0719810724f, 0.f),
    C_Vertex3f(0.2449615f, -0.137058973f, 0.f),
    C_Vertex3f(0.243306756f, -0.0602215528f, 0.f),
    C_Vertex3f(0.227509737f, -0.127690554f, 0.f),
    C_Vertex3f(0.225179911f, -0.047198534f, 0.f),
    C_Vertex3f(0.210816026f, -0.117536902f, 0.f),
    C_Vertex3f(0.208350897f, -0.0329120159f, 0.f),
    C_Vertex3f(0.194880366f, -0.10659802f, 0.f),
    C_Vertex3f(0.192819715f, -0.0173619986f, 0.f),
    C_Vertex3f(0.179702759f, -0.0948739052f, 0.f),
    C_Vertex3f(0.178586364f, -0.000548481941f, 0.f),
    C_Vertex3f(0.165650845f, 0.0175285339f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.179702759f, -0.0948739052f, 0.f),
    C_Vertex3f(0.165650845f, 0.0175285339f, 0.f),
    C_Vertex3f(0.165283203f, -0.0823645592f, 0.f),
    C_Vertex3f(0.154013157f, 0.0368690491f, 0.f),
    C_Vertex3f(0.151751041f, -0.0693839788f, 0.f),
    C_Vertex3f(0.14383018f, 0.0571382046f, 0.f),
    C_Vertex3f(0.139106274f, -0.0559321642f, 0.f),
    C_Vertex3f(0.135101914f, 0.0783360004f, 0.f),
    C_Vertex3f(0.12782836f, 0.100462437f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1273489f, -0.0420091152f, 0.f),
    C_Vertex3f(0.139106274f, -0.0559321642f, 0.f),
    C_Vertex3f(0.12782836f, 0.100462437f, 0.f),
    C_Vertex3f(0.122009516f, 0.123517513f, 0.f),
    C_Vertex3f(0.117645383f, 0.14750123f, 0.f),
    C_Vertex3f(0.11647892f, -0.0276148319f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.117645383f, 0.14750123f, 0.f),
    C_Vertex3f(0.114735961f, 0.172413588f, 0.f),
    C_Vertex3f(0.11647892f, -0.0276148319f, 0.f),
    C_Vertex3f(0.11328125f, 0.198254585f, 0.f),
    C_Vertex3f(0.106496334f, -0.0127493143f, 0.f),
    C_Vertex3f(0.11328125f, 0.224422455f, 0.f),
    C_Vertex3f(0.113911152f, 0.45072782f, 0.f),
    C_Vertex3f(0.114736319f, 0.249585867f, 0.f),
    C_Vertex3f(0.130036831f, 0.476866841f, 0.f),
    C_Vertex3f(0.117646456f, 0.27434659f, 0.f),
    C_Vertex3f(0.122011662f, 0.298704624f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.547908425f, 0.305020809f, 0.f),
    C_Vertex3f(0.548828125f, 0.290977478f, 0.f),
    C_Vertex3f(0.553910732f, 0.421103001f, 0.f),
    C_Vertex3f(0.558348656f, 0.00149154663f, 0.f),
    C_Vertex3f(0.567554474f, 0.403373718f, 0.f),
    C_Vertex3f(0.570538521f, -0.000804901123f, 0.f),
    C_Vertex3f(0.573776245f, 0.393211365f, 0.f),
    C_Vertex3f(0.584177017f, -0.001953125f, 0.f),
    C_Vertex3f(0.588867188f, 0.4609375f, 0.f),
    C_Vertex3f(0.590820313f, 0.0967178345f, 0.f),
    C_Vertex3f(0.590820313f, 0.10049057f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.553910732f, 0.421103001f, 0.f),
    C_Vertex3f(0.539066792f, 0.436236858f, 0.f),
    C_Vertex3f(0.539631128f, 0.342146873f, 0.f),
    C_Vertex3f(0.543309927f, 0.330605507f, 0.f),
    C_Vertex3f(0.546069026f, 0.318230152f, 0.f),
    C_Vertex3f(0.547908425f, 0.305020809f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.539066792f, 0.436236858f, 0.f),
    C_Vertex3f(0.523022652f, 0.448775291f, 0.f),
    C_Vertex3f(0.523076534f, 0.371767044f, 0.f),
    C_Vertex3f(0.529514432f, 0.362727642f, 0.f),
    C_Vertex3f(0.53503263f, 0.352854252f, 0.f),
    C_Vertex3f(0.539631128f, 0.342146873f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.491235018f, 0.398434639f, 0.f),
    C_Vertex3f(0.499798179f, 0.393306255f, 0.f),
    C_Vertex3f(0.505778313f, 0.4587183f, 0.f),
    C_Vertex3f(0.507959485f, 0.387152195f, 0.f),
    C_Vertex3f(0.523022652f, 0.448775291f, 0.f),
    C_Vertex3f(0.515718937f, 0.379972458f, 0.f),
    C_Vertex3f(0.523076534f, 0.371767044f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.488096714f, 0.466175556f, 0.f),
    C_Vertex3f(0.469977856f, 0.47114706f, 0.f),
    C_Vertex3f(0.472903132f, 0.405614376f, 0.f),
    C_Vertex3f(0.482270002f, 0.402537346f, 0.f),
    C_Vertex3f(0.491235018f, 0.398434639f, 0.f),
    C_Vertex3f(0.505778313f, 0.4587183f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.321944714f, 0.304267406f, 0.f),
    C_Vertex3f(0.33120966f, 0.323412418f, 0.f),
    C_Vertex3f(0.33272171f, 0.434296608f, 0.f),
    C_Vertex3f(0.341619492f, 0.340862274f, 0.f),
    C_Vertex3f(0.345928192f, 0.443037987f, 0.f),
    C_Vertex3f(0.35317421f, 0.356616974f, 0.f),
    C_Vertex3f(0.359039307f, 0.450686693f, 0.f),
    C_Vertex3f(0.364919662f, 0.370210648f, 0.f),
    C_Vertex3f(0.372055054f, 0.457242727f, 0.f),
    C_Vertex3f(0.37685585f, 0.381643295f, 0.f),
    C_Vertex3f(0.384975433f, 0.462706089f, 0.f),
    C_Vertex3f(0.388982773f, 0.390914917f, 0.f),
    C_Vertex3f(0.397800446f, 0.467076778f, 0.f),
    C_Vertex3f(0.40130043f, 0.398025513f, 0.f),
    C_Vertex3f(0.41053009f, 0.470354795f, 0.f),
    C_Vertex3f(0.41402626f, 0.403358459f, 0.f),
    C_Vertex3f(0.423164368f, 0.47254014f, 0.f),
    C_Vertex3f(0.427160263f, 0.406913757f, 0.f),
    C_Vertex3f(0.435703278f, 0.473632813f, 0.f),
    C_Vertex3f(0.440702438f, 0.408691406f, 0.f),
    C_Vertex3f(0.451421738f, 0.473632813f, 0.f),
    C_Vertex3f(0.452963829f, 0.408691406f, 0.f),
    C_Vertex3f(0.469977856f, 0.47114706f, 0.f),
    C_Vertex3f(0.463134408f, 0.40766573f, 0.f),
    C_Vertex3f(0.472903132f, 0.405614376f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.301618099f, 0.239285946f, 0.f),
    C_Vertex3f(0.30684948f, 0.260891914f, 0.f),
    C_Vertex3f(0.307925224f, 0.413556218f, 0.f),
    C_Vertex3f(0.313824654f, 0.283427238f, 0.f),
    C_Vertex3f(0.320054054f, 0.424469352f, 0.f),
    C_Vertex3f(0.321944714f, 0.304267406f, 0.f),
    C_Vertex3f(0.33272171f, 0.434296608f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.301700354f, 0.142000675f, 0.f),
    C_Vertex3f(0.299043536f, 0.154485941f, 0.f),
    C_Vertex3f(0.306990623f, 0.015581131f, 0.f),
    C_Vertex3f(0.297272325f, 0.167805433f, 0.f),
    C_Vertex3f(0.288659096f, 0.0269451141f, 0.f),
    C_Vertex3f(0.296386719f, 0.181959152f, 0.f),
    C_Vertex3f(0.29633522f, 0.401557207f, 0.f),
    C_Vertex3f(0.296386719f, 0.198862076f, 0.f),
    C_Vertex3f(0.307925224f, 0.413556218f, 0.f),
    C_Vertex3f(0.298130512f, 0.218609333f, 0.f),
    C_Vertex3f(0.301618099f, 0.239285946f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.841736317f, 0.407146931f, 0.f),
    C_Vertex3f(0.835114241f, 0.427280188f, 0.f),
    C_Vertex3f(0.838253498f, 0.20584631f, 0.f),
    C_Vertex3f(0.827167749f, 0.447003603f, 0.f),
    C_Vertex3f(0.827102184f, 0.177710056f, 0.f),
    C_Vertex3f(0.817896843f, 0.466317177f, 0.f),
    C_Vertex3f(0.813163042f, 0.150461435f, 0.f),
    C_Vertex3f(0.806372643f, 0.486750603f, 0.f),
    C_Vertex3f(0.796436071f, 0.124100447f, 0.f),
    C_Vertex3f(0.792779446f, 0.507354617f, 0.f),
    C_Vertex3f(0.790527344f, 0.321663857f, 0.f),
    C_Vertex3f(0.789359212f, 0.33961463f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.792779446f, 0.507354617f, 0.f),
    C_Vertex3f(0.778046131f, 0.526599526f, 0.f),
    C_Vertex3f(0.778846025f, 0.391663074f, 0.f),
    C_Vertex3f(0.783518553f, 0.374614239f, 0.f),
    C_Vertex3f(0.787022948f, 0.357264757f, 0.f),
    C_Vertex3f(0.789359212f, 0.33961463f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.748474598f, 0.456851959f, 0.f),
    C_Vertex3f(0.757819653f, 0.441005707f, 0.f),
    C_Vertex3f(0.762172699f, 0.544485331f, 0.f),
    C_Vertex3f(0.765996575f, 0.424858809f, 0.f),
    C_Vertex3f(0.778046131f, 0.526599526f, 0.f),
    C_Vertex3f(0.773005366f, 0.408411264f, 0.f),
    C_Vertex3f(0.778846025f, 0.391663074f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.713552952f, 0.499637127f, 0.f),
    C_Vertex3f(0.726341248f, 0.486167669f, 0.f),
    C_Vertex3f(0.727005482f, 0.576179624f, 0.f),
    C_Vertex3f(0.737981796f, 0.471905947f, 0.f),
    C_Vertex3f(0.745159149f, 0.56101203f, 0.f),
    C_Vertex3f(0.748474598f, 0.456851959f, 0.f),
    C_Vertex3f(0.762172699f, 0.544485331f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.668301582f, 0.53529191f, 0.f),
    C_Vertex3f(0.684533119f, 0.524199247f, 0.f),
    C_Vertex3f(0.687277794f, 0.602437496f, 0.f),
    C_Vertex3f(0.699616909f, 0.51231432f, 0.f),
    C_Vertex3f(0.707711697f, 0.589988112f, 0.f),
    C_Vertex3f(0.713552952f, 0.499637127f, 0.f),
    C_Vertex3f(0.727005482f, 0.576179624f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.573903799f, 0.57649374f, 0.f),
    C_Vertex3f(0.593886018f, 0.570551157f, 0.f),
    C_Vertex3f(0.596844673f, 0.638480902f, 0.f),
    C_Vertex3f(0.613383174f, 0.563420057f, 0.f),
    C_Vertex3f(0.620487213f, 0.631549478f, 0.f),
    C_Vertex3f(0.632395267f, 0.555100441f, 0.f),
    C_Vertex3f(0.643440247f, 0.623231769f, 0.f),
    C_Vertex3f(0.650922298f, 0.545592308f, 0.f),
    C_Vertex3f(0.665703773f, 0.613527775f, 0.f),
    C_Vertex3f(0.668301582f, 0.53529191f, 0.f),
    C_Vertex3f(0.687277794f, 0.602437496f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.46513176f, 0.588378906f, 0.f),
    C_Vertex3f(0.489124298f, 0.588378906f, 0.f),
    C_Vertex3f(0.495379448f, 0.65234375f, 0.f),
    C_Vertex3f(0.511046767f, 0.58719039f, 0.f),
    C_Vertex3f(0.521780014f, 0.650957465f, 0.f),
    C_Vertex3f(0.532484174f, 0.584813356f, 0.f),
    C_Vertex3f(0.547491074f, 0.648184896f, 0.f),
    C_Vertex3f(0.553436518f, 0.581247807f, 0.f),
    C_Vertex3f(0.572512627f, 0.644026041f, 0.f),
    C_Vertex3f(0.573903799f, 0.57649374f, 0.f),
    C_Vertex3f(0.596844673f, 0.638480902f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.368355274f, 0.573422909f, 0.f),
    C_Vertex3f(0.391719341f, 0.579405308f, 0.f),
    C_Vertex3f(0.398356915f, 0.646443963f, 0.f),
    C_Vertex3f(0.415636778f, 0.583892107f, 0.f),
    C_Vertex3f(0.430704594f, 0.650377154f, 0.f),
    C_Vertex3f(0.440107584f, 0.586883307f, 0.f),
    C_Vertex3f(0.464356422f, 0.65234375f, 0.f),
    C_Vertex3f(0.46513176f, 0.588378906f, 0.f),
    C_Vertex3f(0.495379448f, 0.65234375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.260470867f, 0.521275043f, 0.f),
    C_Vertex3f(0.280432701f, 0.534537315f, 0.f),
    C_Vertex3f(0.282007694f, 0.611045241f, 0.f),
    C_Vertex3f(0.30158329f, 0.546502113f, 0.f),
    C_Vertex3f(0.309138775f, 0.622844815f, 0.f),
    C_Vertex3f(0.323287249f, 0.556971312f, 0.f),
    C_Vertex3f(0.337574005f, 0.632677794f, 0.f),
    C_Vertex3f(0.345544577f, 0.56594491f, 0.f),
    C_Vertex3f(0.367313385f, 0.640544176f, 0.f),
    C_Vertex3f(0.368355274f, 0.573422909f, 0.f),
    C_Vertex3f(0.398356915f, 0.646443963f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.192511082f, 0.455251217f, 0.f),
    C_Vertex3f(0.207717896f, 0.473703384f, 0.f),
    C_Vertex3f(0.208534718f, 0.564079165f, 0.f),
    C_Vertex3f(0.224113464f, 0.490858078f, 0.f),
    C_Vertex3f(0.231657982f, 0.581546307f, 0.f),
    C_Vertex3f(0.241697788f, 0.506715298f, 0.f),
    C_Vertex3f(0.256180763f, 0.597279072f, 0.f),
    C_Vertex3f(0.260470867f, 0.521275043f, 0.f),
    C_Vertex3f(0.282007694f, 0.611045241f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15402317f, 0.392109871f, 0.f),
    C_Vertex3f(0.165663719f, 0.41445446f, 0.f),
    C_Vertex3f(0.16648674f, 0.523941755f, 0.f),
    C_Vertex3f(0.178493023f, 0.435501575f, 0.f),
    C_Vertex3f(0.18681097f, 0.544877648f, 0.f),
    C_Vertex3f(0.192511082f, 0.455251217f, 0.f),
    C_Vertex3f(0.208534718f, 0.564079165f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147562027f, 0.501271486f, 0.f),
    C_Vertex3f(0.130036831f, 0.476866841f, 0.f),
    C_Vertex3f(0.135107279f, 0.346212626f, 0.f),
    C_Vertex3f(0.14383769f, 0.369362593f, 0.f),
    C_Vertex3f(0.15402317f, 0.392109871f, 0.f),
    C_Vertex3f(0.16648674f, 0.523941755f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130036831f, 0.476866841f, 0.f),
    C_Vertex3f(0.122011662f, 0.298704624f, 0.f),
    C_Vertex3f(0.127831936f, 0.322659969f, 0.f),
    C_Vertex3f(0.135107279f, 0.346212626f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0991849899f, 0.422854424f, 0.f),
    C_Vertex3f(0.0870599747f, 0.395977974f, 0.f),
    C_Vertex3f(0.0891933441f, 0.0183954239f, 0.f),
    C_Vertex3f(0.0974011421f, 0.00258743763f, 0.f),
    C_Vertex3f(0.106496334f, -0.0127493143f, 0.f),
    C_Vertex3f(0.113911152f, 0.45072782f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0891933441f, 0.0183954239f, 0.f),
    C_Vertex3f(0.0870599747f, 0.395977974f, 0.f),
    C_Vertex3f(0.080945015f, 0.0368595123f, 0.f),
    C_Vertex3f(0.0775020123f, 0.37151885f, 0.f),
    C_Vertex3f(0.0729157925f, 0.0579183102f, 0.f),
    C_Vertex3f(0.069309473f, 0.346745729f, 0.f),
    C_Vertex3f(0.0660336018f, 0.0793869495f, 0.f),
    C_Vertex3f(0.062482357f, 0.321658611f, 0.f),
    C_Vertex3f(0.0602984428f, 0.10126543f, 0.f),
    C_Vertex3f(0.0570206642f, 0.296257496f, 0.f),
    C_Vertex3f(0.0557103157f, 0.123553753f, 0.f),
    C_Vertex3f(0.0529243946f, 0.270542383f, 0.f),
    C_Vertex3f(0.0522692204f, 0.146251917f, 0.f),
    C_Vertex3f(0.0501935482f, 0.244513273f, 0.f),
    C_Vertex3f(0.0499751568f, 0.169359922f, 0.f),
    C_Vertex3f(0.048828125f, 0.218170166f, 0.f),
    C_Vertex3f(0.048828125f, 0.192877769f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.853656054f, 0.344288111f, 0.f),
    C_Vertex3f(0.851007223f, 0.365650892f, 0.f),
    C_Vertex3f(0.85219264f, 0.264781713f, 0.f),
    C_Vertex3f(0.854980469f, 0.295580864f, 0.f),
    C_Vertex3f(0.854980469f, 0.322515488f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.847033978f, 0.386603832f, 0.f),
    C_Vertex3f(0.841736317f, 0.407146931f, 0.f),
    C_Vertex3f(0.846616983f, 0.234870195f, 0.f),
    C_Vertex3f(0.85219264f, 0.264781713f, 0.f),
    C_Vertex3f(0.851007223f, 0.365650892f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.796436071f, 0.124100447f, 0.f),
    C_Vertex3f(0.790527344f, 0.321663857f, 0.f),
    C_Vertex3f(0.790527344f, 0.29691124f, 0.f),
    C_Vertex3f(0.786886692f, 0.266121387f, 0.f),
    C_Vertex3f(0.779605389f, 0.235795498f, 0.f),
    C_Vertex3f(0.776921272f, 0.0986270905f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.779605389f, 0.235795498f, 0.f),
    C_Vertex3f(0.768683434f, 0.205933571f, 0.f),
    C_Vertex3f(0.776921272f, 0.0986270905f, 0.f),
    C_Vertex3f(0.757761478f, 0.183885098f, 0.f),
    C_Vertex3f(0.755821228f, 0.0752983093f, 0.f),
    C_Vertex3f(0.749830127f, 0.169930577f, 0.f),
    C_Vertex3f(0.741248727f, 0.156720519f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.735459089f, 0.0559854507f, 0.f),
    C_Vertex3f(0.755821228f, 0.0752983093f, 0.f),
    C_Vertex3f(0.741248727f, 0.156720519f, 0.f),
    C_Vertex3f(0.732017279f, 0.144254923f, 0.f),
    C_Vertex3f(0.722135782f, 0.132533789f, 0.f),
    C_Vertex3f(0.714632273f, 0.039431572f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.722135782f, 0.132533789f, 0.f),
    C_Vertex3f(0.711604238f, 0.121557117f, 0.f),
    C_Vertex3f(0.714632273f, 0.039431572f, 0.f),
    C_Vertex3f(0.700422645f, 0.111324906f, 0.f),
    C_Vertex3f(0.693340778f, 0.025636673f, 0.f),
    C_Vertex3f(0.688591003f, 0.101837158f, 0.f),
    C_Vertex3f(0.672935486f, 0.0907993317f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.693340778f, 0.025636673f, 0.f),
    C_Vertex3f(0.672935486f, 0.0907993317f, 0.f),
    C_Vertex3f(0.671584606f, 0.0146007538f, 0.f),
    C_Vertex3f(0.655911446f, 0.0808000565f, 0.f),
    C_Vertex3f(0.649363756f, 0.00632381439f, 0.f),
    C_Vertex3f(0.640692711f, 0.074133873f, 0.f),
    C_Vertex3f(0.627279282f, 0.0708007813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.626678228f, 0.000805854797f, 0.f),
    C_Vertex3f(0.649363756f, 0.00632381439f, 0.f),
    C_Vertex3f(0.627279282f, 0.0708007813f, 0.f),
    C_Vertex3f(0.618301392f, 0.0708007813f, 0.f),
    C_Vertex3f(0.612418175f, 0.0718669891f, 0.f),
    C_Vertex3f(0.606999397f, 0.0739994049f, 0.f),
    C_Vertex3f(0.603528023f, -0.001953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.603528023f, -0.001953125f, 0.f),
    C_Vertex3f(0.606999397f, 0.0739994049f, 0.f),
    C_Vertex3f(0.602045059f, 0.0771980286f, 0.f),
    C_Vertex3f(0.597555161f, 0.0814628601f, 0.f),
    C_Vertex3f(0.594187737f, 0.0861377716f, 0.f),
    C_Vertex3f(0.591942787f, 0.0912227631f, 0.f),
    C_Vertex3f(0.590820313f, 0.0967178345f, 0.f),
    C_Vertex3f(0.584177017f, -0.001953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.548828125f, 0.272378922f, 0.f),
    C_Vertex3f(0.54686594f, 0.250087261f, 0.f),
    C_Vertex3f(0.547607422f, 0.00493621826f, 0.f),
    C_Vertex3f(0.558348656f, 0.00149154663f, 0.f),
    C_Vertex3f(0.548828125f, 0.290977478f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.514211655f, 0.0362110138f, 0.f),
    C_Vertex3f(0.517818451f, 0.0292129517f, 0.f),
    C_Vertex3f(0.520240784f, 0.163115501f, 0.f),
    C_Vertex3f(0.523420334f, 0.0216674805f, 0.f),
    C_Vertex3f(0.529206276f, 0.183381081f, 0.f),
    C_Vertex3f(0.530252457f, 0.0151062012f, 0.f),
    C_Vertex3f(0.537055016f, 0.205588341f, 0.f),
    C_Vertex3f(0.538314819f, 0.00952911377f, 0.f),
    C_Vertex3f(0.54294157f, 0.227823734f, 0.f),
    C_Vertex3f(0.547607422f, 0.00493621826f, 0.f),
    C_Vertex3f(0.54686594f, 0.250087261f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.507980347f, 0.0644207001f, 0.f),
    C_Vertex3f(0.509675026f, 0.0530490875f, 0.f),
    C_Vertex3f(0.510158539f, 0.144791603f, 0.f),
    C_Vertex3f(0.511752129f, 0.0436458588f, 0.f),
    C_Vertex3f(0.520240784f, 0.163115501f, 0.f),
    C_Vertex3f(0.514211655f, 0.0362110138f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.510158539f, 0.144791603f, 0.f),
    C_Vertex3f(0.498959541f, 0.128409386f, 0.f),
    C_Vertex3f(0.500230789f, 0.0628948212f, 0.f),
    C_Vertex3f(0.507324219f, 0.0710906982f, 0.f),
    C_Vertex3f(0.507980347f, 0.0644207001f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.434381485f, 0.0739831924f, 0.f),
    C_Vertex3f(0.435033798f, 0.0134496689f, 0.f),
    C_Vertex3f(0.447896957f, 0.0813131332f, 0.f),
    C_Vertex3f(0.452600479f, 0.0233926773f, 0.f),
    C_Vertex3f(0.461112499f, 0.090420723f, 0.f),
    C_Vertex3f(0.469322205f, 0.0349478722f, 0.f),
    C_Vertex3f(0.474028111f, 0.101305962f, 0.f),
    C_Vertex3f(0.485198975f, 0.0481152534f, 0.f),
    C_Vertex3f(0.486643791f, 0.113968849f, 0.f),
    C_Vertex3f(0.500230789f, 0.0628948212f, 0.f),
    C_Vertex3f(0.498959541f, 0.128409386f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.395792007f, 0.0629882813f, 0.f),
    C_Vertex3f(0.399817467f, 0.00102090836f, 0.f),
    C_Vertex3f(0.40832901f, 0.0648207664f, 0.f),
    C_Vertex3f(0.417439461f, 0.00599241257f, 0.f),
    C_Vertex3f(0.421192169f, 0.0684857368f, 0.f),
    C_Vertex3f(0.435033798f, 0.0134496689f, 0.f),
    C_Vertex3f(0.434381485f, 0.0739831924f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.385108948f, 0.0629882813f, 0.f),
    C_Vertex3f(0.376212478f, 0.0640275478f, 0.f),
    C_Vertex3f(0.382167816f, -0.00146484375f, 0.f),
    C_Vertex3f(0.399817467f, 0.00102090836f, 0.f),
    C_Vertex3f(0.395792007f, 0.0629882813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.367574811f, 0.066106081f, 0.f),
    C_Vertex3f(0.359195948f, 0.0692238808f, 0.f),
    C_Vertex3f(0.363618851f, -0.00146484375f, 0.f),
    C_Vertex3f(0.382167816f, -0.00146484375f, 0.f),
    C_Vertex3f(0.376212478f, 0.0640275478f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.351075888f, 0.0733809471f, 0.f),
    C_Vertex3f(0.343214631f, 0.07857728f, 0.f),
    C_Vertex3f(0.344470501f, 0.00137615204f, 0.f),
    C_Vertex3f(0.363618851f, -0.00146484375f, 0.f),
    C_Vertex3f(0.359195948f, 0.0692238808f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.343214631f, 0.07857728f, 0.f),
    C_Vertex3f(0.335612178f, 0.0848128796f, 0.f),
    C_Vertex3f(0.344470501f, 0.00137615204f, 0.f),
    C_Vertex3f(0.328268528f, 0.0920877457f, 0.f),
    C_Vertex3f(0.325594425f, 0.00705814362f, 0.f),
    C_Vertex3f(0.321183681f, 0.100401878f, 0.f),
    C_Vertex3f(0.314984441f, 0.109550238f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.309670806f, 0.119532824f, 0.f),
    C_Vertex3f(0.305242777f, 0.130349636f, 0.f),
    C_Vertex3f(0.306990623f, 0.015581131f, 0.f),
    C_Vertex3f(0.325594425f, 0.00705814362f, 0.f),
    C_Vertex3f(0.314984441f, 0.109550238f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.29633522f, 0.401557207f, 0.f),
    C_Vertex3f(0.285284042f, 0.388472319f, 0.f),
    C_Vertex3f(0.288659096f, 0.0269451141f, 0.f),
    C_Vertex3f(0.27477169f, 0.374301553f, 0.f),
    C_Vertex3f(0.272046089f, 0.0413684845f, 0.f),
    C_Vertex3f(0.264798164f, 0.359044909f, 0.f),
    C_Vertex3f(0.257151604f, 0.0588512421f, 0.f),
    C_Vertex3f(0.255363464f, 0.342702389f, 0.f),
    C_Vertex3f(0.24646759f, 0.325273991f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.257151604f, 0.0588512421f, 0.f),
    C_Vertex3f(0.24646759f, 0.325273991f, 0.f),
    C_Vertex3f(0.243975639f, 0.0793933868f, 0.f),
    C_Vertex3f(0.238683701f, 0.307852387f, 0.f),
    C_Vertex3f(0.232518196f, 0.102994919f, 0.f),
    C_Vertex3f(0.232011795f, 0.290437579f, 0.f),
    C_Vertex3f(0.226451874f, 0.273029566f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.223925114f, 0.127743721f, 0.f),
    C_Vertex3f(0.232518196f, 0.102994919f, 0.f),
    C_Vertex3f(0.226451874f, 0.273029566f, 0.f),
    C_Vertex3f(0.222003937f, 0.255628347f, 0.f),
    C_Vertex3f(0.218667984f, 0.238233924f, 0.f),
    C_Vertex3f(0.218196392f, 0.153639793f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.218196392f, 0.153639793f, 0.f),
    C_Vertex3f(0.218667984f, 0.238233924f, 0.f),
    C_Vertex3f(0.216444016f, 0.220846295f, 0.f),
    C_Vertex3f(0.215332031f, 0.203465462f, 0.f),
    C_Vertex3f(0.215332031f, 0.180683136f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.588867188f, 0.4609375f, 0.f),
    C_Vertex3f(0.590820313f, 0.10049057f, 0.f),
    C_Vertex3f(0.591230631f, 0.103839159f, 0.f),
    C_Vertex3f(0.592051268f, 0.108896017f, 0.f),
    C_Vertex3f(0.593282223f, 0.115661144f, 0.f),
    C_Vertex3f(0.594923496f, 0.124134541f, 0.f),
    C_Vertex3f(0.596975088f, 0.134316206f, 0.f),
    C_Vertex3f(0.599436998f, 0.146206141f, 0.f),
    C_Vertex3f(0.602309227f, 0.159804344f, 0.f),
    C_Vertex3f(0.666992188f, 0.4609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.301700354f, 0.142000675f, 0.f),
    C_Vertex3f(0.306990623f, 0.015581131f, 0.f),
    C_Vertex3f(0.305242777f, 0.130349636f, 0.f),
    C_Vertex3f(0.841736317f, 0.407146931f, 0.f),
    C_Vertex3f(0.838253498f, 0.20584631f, 0.f),
    C_Vertex3f(0.846616983f, 0.234870195f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.876464844f, 0.00830078125f, -1.f),
    C_Vertex3f(0.797363281f, 0.00830078125f, -1.f),
    C_Vertex3f(0.806768417f, -0.0793194771f, -1.f),
    C_Vertex3f(0.819317579f, -0.0674580336f, -1.f),
    C_Vertex3f(0.830869913f, -0.0556513071f, -1.f),
    C_Vertex3f(0.841425419f, -0.0438992977f, -1.f),
    C_Vertex3f(0.850984097f, -0.0322020054f, -1.f),
    C_Vertex3f(0.859545946f, -0.0205594301f, -1.f),
    C_Vertex3f(0.867110968f, -0.00897157192f, -1.f),
    C_Vertex3f(0.873679161f, 0.00256156921f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.792554855f, 0.0013999939f, -1.f),
    C_Vertex3f(0.781722307f, -0.0118813515f, -1.f),
    C_Vertex3f(0.786449432f, -0.097193718f, -1.f),
    C_Vertex3f(0.806768417f, -0.0793194771f, -1.f),
    C_Vertex3f(0.797363281f, 0.00830078125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.769674063f, -0.0246424675f, -1.f),
    C_Vertex3f(0.756410122f, -0.0368833542f, -1.f),
    C_Vertex3f(0.756789684f, -0.118538857f, -1.f),
    C_Vertex3f(0.786449432f, -0.097193718f, -1.f),
    C_Vertex3f(0.781722307f, -0.0118813515f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.756410122f, -0.0368833542f, -1.f),
    C_Vertex3f(0.741930485f, -0.0486040115f, -1.f),
    C_Vertex3f(0.756789684f, -0.118538857f, -1.f),
    C_Vertex3f(0.726235151f, -0.0598044395f, -1.f),
    C_Vertex3f(0.724562168f, -0.137396812f, -1.f),
    C_Vertex3f(0.709324121f, -0.0704846382f, -1.f),
    C_Vertex3f(0.691197395f, -0.0806446075f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.504214287f, -0.188476563f, -1.f),
    C_Vertex3f(0.529585719f, -0.187608838f, -1.f),
    C_Vertex3f(0.531422377f, -0.12281847f, -1.f),
    C_Vertex3f(0.554001451f, -0.185873389f, -1.f),
    C_Vertex3f(0.557729959f, -0.120408535f, -1.f),
    C_Vertex3f(0.577461481f, -0.183270216f, -1.f),
    C_Vertex3f(0.582876682f, -0.116793633f, -1.f),
    C_Vertex3f(0.599965811f, -0.179799318f, -1.f),
    C_Vertex3f(0.606862545f, -0.111973763f, -1.f),
    C_Vertex3f(0.62151444f, -0.175460696f, -1.f),
    C_Vertex3f(0.629687548f, -0.105948925f, -1.f),
    C_Vertex3f(0.642107368f, -0.17025435f, -1.f),
    C_Vertex3f(0.65135169f, -0.09871912f, -1.f),
    C_Vertex3f(0.661744595f, -0.164180279f, -1.f),
    C_Vertex3f(0.671854973f, -0.0902843475f, -1.f),
    C_Vertex3f(0.689766884f, -0.153767586f, -1.f),
    C_Vertex3f(0.691197395f, -0.0806446075f, -1.f),
    C_Vertex3f(0.724562168f, -0.137396812f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.531422377f, -0.12281847f, -1.f),
    C_Vertex3f(0.503953934f, -0.124023438f, -1.f),
    C_Vertex3f(0.504214287f, -0.188476563f, -1.f),
    C_Vertex3f(0.476252556f, -0.124023438f, -1.f),
    C_Vertex3f(0.466798782f, -0.188476563f, -1.f),
    C_Vertex3f(0.450073242f, -0.122869372f, -1.f),
    C_Vertex3f(0.424488068f, -0.120561242f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466798782f, -0.188476563f, -1.f),
    C_Vertex3f(0.424488068f, -0.120561242f, -1.f),
    C_Vertex3f(0.42017889f, -0.185362816f, -1.f),
    C_Vertex3f(0.399497032f, -0.117099047f, -1.f),
    C_Vertex3f(0.375442982f, -0.179135323f, -1.f),
    C_Vertex3f(0.375100136f, -0.112482786f, -1.f),
    C_Vertex3f(0.351297379f, -0.106712461f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.332591057f, -0.169794083f, -1.f),
    C_Vertex3f(0.375442982f, -0.179135323f, -1.f),
    C_Vertex3f(0.351297379f, -0.106712461f, -1.f),
    C_Vertex3f(0.32808876f, -0.0997880697f, -1.f),
    C_Vertex3f(0.305474281f, -0.0917096138f, -1.f),
    C_Vertex3f(0.301865101f, -0.160452843f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.283453941f, -0.0824770927f, -1.f),
    C_Vertex3f(0.262731433f, -0.0719810724f, -1.f),
    C_Vertex3f(0.263171315f, -0.145642161f, -1.f),
    C_Vertex3f(0.282139182f, -0.153440118f, -1.f),
    C_Vertex3f(0.301865101f, -0.160452843f, -1.f),
    C_Vertex3f(0.305474281f, -0.0917096138f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.263171315f, -0.145642161f, -1.f),
    C_Vertex3f(0.262731433f, -0.0719810724f, -1.f),
    C_Vertex3f(0.2449615f, -0.137058973f, -1.f),
    C_Vertex3f(0.243306756f, -0.0602215528f, -1.f),
    C_Vertex3f(0.227509737f, -0.127690554f, -1.f),
    C_Vertex3f(0.225179911f, -0.047198534f, -1.f),
    C_Vertex3f(0.210816026f, -0.117536902f, -1.f),
    C_Vertex3f(0.208350897f, -0.0329120159f, -1.f),
    C_Vertex3f(0.194880366f, -0.10659802f, -1.f),
    C_Vertex3f(0.192819715f, -0.0173619986f, -1.f),
    C_Vertex3f(0.179702759f, -0.0948739052f, -1.f),
    C_Vertex3f(0.178586364f, -0.000548481941f, -1.f),
    C_Vertex3f(0.165650845f, 0.0175285339f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.179702759f, -0.0948739052f, -1.f),
    C_Vertex3f(0.165650845f, 0.0175285339f, -1.f),
    C_Vertex3f(0.165283203f, -0.0823645592f, -1.f),
    C_Vertex3f(0.154013157f, 0.0368690491f, -1.f),
    C_Vertex3f(0.151751041f, -0.0693839788f, -1.f),
    C_Vertex3f(0.14383018f, 0.0571382046f, -1.f),
    C_Vertex3f(0.139106274f, -0.0559321642f, -1.f),
    C_Vertex3f(0.135101914f, 0.0783360004f, -1.f),
    C_Vertex3f(0.12782836f, 0.100462437f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1273489f, -0.0420091152f, -1.f),
    C_Vertex3f(0.139106274f, -0.0559321642f, -1.f),
    C_Vertex3f(0.12782836f, 0.100462437f, -1.f),
    C_Vertex3f(0.122009516f, 0.123517513f, -1.f),
    C_Vertex3f(0.117645383f, 0.14750123f, -1.f),
    C_Vertex3f(0.11647892f, -0.0276148319f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.117645383f, 0.14750123f, -1.f),
    C_Vertex3f(0.114735961f, 0.172413588f, -1.f),
    C_Vertex3f(0.11647892f, -0.0276148319f, -1.f),
    C_Vertex3f(0.11328125f, 0.198254585f, -1.f),
    C_Vertex3f(0.106496334f, -0.0127493143f, -1.f),
    C_Vertex3f(0.11328125f, 0.224422455f, -1.f),
    C_Vertex3f(0.113911152f, 0.45072782f, -1.f),
    C_Vertex3f(0.114736319f, 0.249585867f, -1.f),
    C_Vertex3f(0.130036831f, 0.476866841f, -1.f),
    C_Vertex3f(0.117646456f, 0.27434659f, -1.f),
    C_Vertex3f(0.122011662f, 0.298704624f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.547908425f, 0.305020809f, -1.f),
    C_Vertex3f(0.548828125f, 0.290977478f, -1.f),
    C_Vertex3f(0.553910732f, 0.421103001f, -1.f),
    C_Vertex3f(0.558348656f, 0.00149154663f, -1.f),
    C_Vertex3f(0.567554474f, 0.403373718f, -1.f),
    C_Vertex3f(0.570538521f, -0.000804901123f, -1.f),
    C_Vertex3f(0.573776245f, 0.393211365f, -1.f),
    C_Vertex3f(0.584177017f, -0.001953125f, -1.f),
    C_Vertex3f(0.588867188f, 0.4609375f, -1.f),
    C_Vertex3f(0.590820313f, 0.0967178345f, -1.f),
    C_Vertex3f(0.590820313f, 0.10049057f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.553910732f, 0.421103001f, -1.f),
    C_Vertex3f(0.539066792f, 0.436236858f, -1.f),
    C_Vertex3f(0.539631128f, 0.342146873f, -1.f),
    C_Vertex3f(0.543309927f, 0.330605507f, -1.f),
    C_Vertex3f(0.546069026f, 0.318230152f, -1.f),
    C_Vertex3f(0.547908425f, 0.305020809f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.539066792f, 0.436236858f, -1.f),
    C_Vertex3f(0.523022652f, 0.448775291f, -1.f),
    C_Vertex3f(0.523076534f, 0.371767044f, -1.f),
    C_Vertex3f(0.529514432f, 0.362727642f, -1.f),
    C_Vertex3f(0.53503263f, 0.352854252f, -1.f),
    C_Vertex3f(0.539631128f, 0.342146873f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.491235018f, 0.398434639f, -1.f),
    C_Vertex3f(0.499798179f, 0.393306255f, -1.f),
    C_Vertex3f(0.505778313f, 0.4587183f, -1.f),
    C_Vertex3f(0.507959485f, 0.387152195f, -1.f),
    C_Vertex3f(0.523022652f, 0.448775291f, -1.f),
    C_Vertex3f(0.515718937f, 0.379972458f, -1.f),
    C_Vertex3f(0.523076534f, 0.371767044f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.488096714f, 0.466175556f, -1.f),
    C_Vertex3f(0.469977856f, 0.47114706f, -1.f),
    C_Vertex3f(0.472903132f, 0.405614376f, -1.f),
    C_Vertex3f(0.482270002f, 0.402537346f, -1.f),
    C_Vertex3f(0.491235018f, 0.398434639f, -1.f),
    C_Vertex3f(0.505778313f, 0.4587183f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.321944714f, 0.304267406f, -1.f),
    C_Vertex3f(0.33120966f, 0.323412418f, -1.f),
    C_Vertex3f(0.33272171f, 0.434296608f, -1.f),
    C_Vertex3f(0.341619492f, 0.340862274f, -1.f),
    C_Vertex3f(0.345928192f, 0.443037987f, -1.f),
    C_Vertex3f(0.35317421f, 0.356616974f, -1.f),
    C_Vertex3f(0.359039307f, 0.450686693f, -1.f),
    C_Vertex3f(0.364919662f, 0.370210648f, -1.f),
    C_Vertex3f(0.372055054f, 0.457242727f, -1.f),
    C_Vertex3f(0.37685585f, 0.381643295f, -1.f),
    C_Vertex3f(0.384975433f, 0.462706089f, -1.f),
    C_Vertex3f(0.388982773f, 0.390914917f, -1.f),
    C_Vertex3f(0.397800446f, 0.467076778f, -1.f),
    C_Vertex3f(0.40130043f, 0.398025513f, -1.f),
    C_Vertex3f(0.41053009f, 0.470354795f, -1.f),
    C_Vertex3f(0.41402626f, 0.403358459f, -1.f),
    C_Vertex3f(0.423164368f, 0.47254014f, -1.f),
    C_Vertex3f(0.427160263f, 0.406913757f, -1.f),
    C_Vertex3f(0.435703278f, 0.473632813f, -1.f),
    C_Vertex3f(0.440702438f, 0.408691406f, -1.f),
    C_Vertex3f(0.451421738f, 0.473632813f, -1.f),
    C_Vertex3f(0.452963829f, 0.408691406f, -1.f),
    C_Vertex3f(0.469977856f, 0.47114706f, -1.f),
    C_Vertex3f(0.463134408f, 0.40766573f, -1.f),
    C_Vertex3f(0.472903132f, 0.405614376f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.301618099f, 0.239285946f, -1.f),
    C_Vertex3f(0.30684948f, 0.260891914f, -1.f),
    C_Vertex3f(0.307925224f, 0.413556218f, -1.f),
    C_Vertex3f(0.313824654f, 0.283427238f, -1.f),
    C_Vertex3f(0.320054054f, 0.424469352f, -1.f),
    C_Vertex3f(0.321944714f, 0.304267406f, -1.f),
    C_Vertex3f(0.33272171f, 0.434296608f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.301700354f, 0.142000675f, -1.f),
    C_Vertex3f(0.299043536f, 0.154485941f, -1.f),
    C_Vertex3f(0.306990623f, 0.015581131f, -1.f),
    C_Vertex3f(0.297272325f, 0.167805433f, -1.f),
    C_Vertex3f(0.288659096f, 0.0269451141f, -1.f),
    C_Vertex3f(0.296386719f, 0.181959152f, -1.f),
    C_Vertex3f(0.29633522f, 0.401557207f, -1.f),
    C_Vertex3f(0.296386719f, 0.198862076f, -1.f),
    C_Vertex3f(0.307925224f, 0.413556218f, -1.f),
    C_Vertex3f(0.298130512f, 0.218609333f, -1.f),
    C_Vertex3f(0.301618099f, 0.239285946f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.841736317f, 0.407146931f, -1.f),
    C_Vertex3f(0.835114241f, 0.427280188f, -1.f),
    C_Vertex3f(0.838253498f, 0.20584631f, -1.f),
    C_Vertex3f(0.827167749f, 0.447003603f, -1.f),
    C_Vertex3f(0.827102184f, 0.177710056f, -1.f),
    C_Vertex3f(0.817896843f, 0.466317177f, -1.f),
    C_Vertex3f(0.813163042f, 0.150461435f, -1.f),
    C_Vertex3f(0.806372643f, 0.486750603f, -1.f),
    C_Vertex3f(0.796436071f, 0.124100447f, -1.f),
    C_Vertex3f(0.792779446f, 0.507354617f, -1.f),
    C_Vertex3f(0.790527344f, 0.321663857f, -1.f),
    C_Vertex3f(0.789359212f, 0.33961463f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.792779446f, 0.507354617f, -1.f),
    C_Vertex3f(0.778046131f, 0.526599526f, -1.f),
    C_Vertex3f(0.778846025f, 0.391663074f, -1.f),
    C_Vertex3f(0.783518553f, 0.374614239f, -1.f),
    C_Vertex3f(0.787022948f, 0.357264757f, -1.f),
    C_Vertex3f(0.789359212f, 0.33961463f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.748474598f, 0.456851959f, -1.f),
    C_Vertex3f(0.757819653f, 0.441005707f, -1.f),
    C_Vertex3f(0.762172699f, 0.544485331f, -1.f),
    C_Vertex3f(0.765996575f, 0.424858809f, -1.f),
    C_Vertex3f(0.778046131f, 0.526599526f, -1.f),
    C_Vertex3f(0.773005366f, 0.408411264f, -1.f),
    C_Vertex3f(0.778846025f, 0.391663074f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.713552952f, 0.499637127f, -1.f),
    C_Vertex3f(0.726341248f, 0.486167669f, -1.f),
    C_Vertex3f(0.727005482f, 0.576179624f, -1.f),
    C_Vertex3f(0.737981796f, 0.471905947f, -1.f),
    C_Vertex3f(0.745159149f, 0.56101203f, -1.f),
    C_Vertex3f(0.748474598f, 0.456851959f, -1.f),
    C_Vertex3f(0.762172699f, 0.544485331f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.668301582f, 0.53529191f, -1.f),
    C_Vertex3f(0.684533119f, 0.524199247f, -1.f),
    C_Vertex3f(0.687277794f, 0.602437496f, -1.f),
    C_Vertex3f(0.699616909f, 0.51231432f, -1.f),
    C_Vertex3f(0.707711697f, 0.589988112f, -1.f),
    C_Vertex3f(0.713552952f, 0.499637127f, -1.f),
    C_Vertex3f(0.727005482f, 0.576179624f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.573903799f, 0.57649374f, -1.f),
    C_Vertex3f(0.593886018f, 0.570551157f, -1.f),
    C_Vertex3f(0.596844673f, 0.638480902f, -1.f),
    C_Vertex3f(0.613383174f, 0.563420057f, -1.f),
    C_Vertex3f(0.620487213f, 0.631549478f, -1.f),
    C_Vertex3f(0.632395267f, 0.555100441f, -1.f),
    C_Vertex3f(0.643440247f, 0.623231769f, -1.f),
    C_Vertex3f(0.650922298f, 0.545592308f, -1.f),
    C_Vertex3f(0.665703773f, 0.613527775f, -1.f),
    C_Vertex3f(0.668301582f, 0.53529191f, -1.f),
    C_Vertex3f(0.687277794f, 0.602437496f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.46513176f, 0.588378906f, -1.f),
    C_Vertex3f(0.489124298f, 0.588378906f, -1.f),
    C_Vertex3f(0.495379448f, 0.65234375f, -1.f),
    C_Vertex3f(0.511046767f, 0.58719039f, -1.f),
    C_Vertex3f(0.521780014f, 0.650957465f, -1.f),
    C_Vertex3f(0.532484174f, 0.584813356f, -1.f),
    C_Vertex3f(0.547491074f, 0.648184896f, -1.f),
    C_Vertex3f(0.553436518f, 0.581247807f, -1.f),
    C_Vertex3f(0.572512627f, 0.644026041f, -1.f),
    C_Vertex3f(0.573903799f, 0.57649374f, -1.f),
    C_Vertex3f(0.596844673f, 0.638480902f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.368355274f, 0.573422909f, -1.f),
    C_Vertex3f(0.391719341f, 0.579405308f, -1.f),
    C_Vertex3f(0.398356915f, 0.646443963f, -1.f),
    C_Vertex3f(0.415636778f, 0.583892107f, -1.f),
    C_Vertex3f(0.430704594f, 0.650377154f, -1.f),
    C_Vertex3f(0.440107584f, 0.586883307f, -1.f),
    C_Vertex3f(0.464356422f, 0.65234375f, -1.f),
    C_Vertex3f(0.46513176f, 0.588378906f, -1.f),
    C_Vertex3f(0.495379448f, 0.65234375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.260470867f, 0.521275043f, -1.f),
    C_Vertex3f(0.280432701f, 0.534537315f, -1.f),
    C_Vertex3f(0.282007694f, 0.611045241f, -1.f),
    C_Vertex3f(0.30158329f, 0.546502113f, -1.f),
    C_Vertex3f(0.309138775f, 0.622844815f, -1.f),
    C_Vertex3f(0.323287249f, 0.556971312f, -1.f),
    C_Vertex3f(0.337574005f, 0.632677794f, -1.f),
    C_Vertex3f(0.345544577f, 0.56594491f, -1.f),
    C_Vertex3f(0.367313385f, 0.640544176f, -1.f),
    C_Vertex3f(0.368355274f, 0.573422909f, -1.f),
    C_Vertex3f(0.398356915f, 0.646443963f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.192511082f, 0.455251217f, -1.f),
    C_Vertex3f(0.207717896f, 0.473703384f, -1.f),
    C_Vertex3f(0.208534718f, 0.564079165f, -1.f),
    C_Vertex3f(0.224113464f, 0.490858078f, -1.f),
    C_Vertex3f(0.231657982f, 0.581546307f, -1.f),
    C_Vertex3f(0.241697788f, 0.506715298f, -1.f),
    C_Vertex3f(0.256180763f, 0.597279072f, -1.f),
    C_Vertex3f(0.260470867f, 0.521275043f, -1.f),
    C_Vertex3f(0.282007694f, 0.611045241f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15402317f, 0.392109871f, -1.f),
    C_Vertex3f(0.165663719f, 0.41445446f, -1.f),
    C_Vertex3f(0.16648674f, 0.523941755f, -1.f),
    C_Vertex3f(0.178493023f, 0.435501575f, -1.f),
    C_Vertex3f(0.18681097f, 0.544877648f, -1.f),
    C_Vertex3f(0.192511082f, 0.455251217f, -1.f),
    C_Vertex3f(0.208534718f, 0.564079165f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147562027f, 0.501271486f, -1.f),
    C_Vertex3f(0.130036831f, 0.476866841f, -1.f),
    C_Vertex3f(0.135107279f, 0.346212626f, -1.f),
    C_Vertex3f(0.14383769f, 0.369362593f, -1.f),
    C_Vertex3f(0.15402317f, 0.392109871f, -1.f),
    C_Vertex3f(0.16648674f, 0.523941755f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130036831f, 0.476866841f, -1.f),
    C_Vertex3f(0.122011662f, 0.298704624f, -1.f),
    C_Vertex3f(0.127831936f, 0.322659969f, -1.f),
    C_Vertex3f(0.135107279f, 0.346212626f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0991849899f, 0.422854424f, -1.f),
    C_Vertex3f(0.0870599747f, 0.395977974f, -1.f),
    C_Vertex3f(0.0891933441f, 0.0183954239f, -1.f),
    C_Vertex3f(0.0974011421f, 0.00258743763f, -1.f),
    C_Vertex3f(0.106496334f, -0.0127493143f, -1.f),
    C_Vertex3f(0.113911152f, 0.45072782f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0891933441f, 0.0183954239f, -1.f),
    C_Vertex3f(0.0870599747f, 0.395977974f, -1.f),
    C_Vertex3f(0.080945015f, 0.0368595123f, -1.f),
    C_Vertex3f(0.0775020123f, 0.37151885f, -1.f),
    C_Vertex3f(0.0729157925f, 0.0579183102f, -1.f),
    C_Vertex3f(0.069309473f, 0.346745729f, -1.f),
    C_Vertex3f(0.0660336018f, 0.0793869495f, -1.f),
    C_Vertex3f(0.062482357f, 0.321658611f, -1.f),
    C_Vertex3f(0.0602984428f, 0.10126543f, -1.f),
    C_Vertex3f(0.0570206642f, 0.296257496f, -1.f),
    C_Vertex3f(0.0557103157f, 0.123553753f, -1.f),
    C_Vertex3f(0.0529243946f, 0.270542383f, -1.f),
    C_Vertex3f(0.0522692204f, 0.146251917f, -1.f),
    C_Vertex3f(0.0501935482f, 0.244513273f, -1.f),
    C_Vertex3f(0.0499751568f, 0.169359922f, -1.f),
    C_Vertex3f(0.048828125f, 0.218170166f, -1.f),
    C_Vertex3f(0.048828125f, 0.192877769f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.853656054f, 0.344288111f, -1.f),
    C_Vertex3f(0.851007223f, 0.365650892f, -1.f),
    C_Vertex3f(0.85219264f, 0.264781713f, -1.f),
    C_Vertex3f(0.854980469f, 0.295580864f, -1.f),
    C_Vertex3f(0.854980469f, 0.322515488f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.847033978f, 0.386603832f, -1.f),
    C_Vertex3f(0.841736317f, 0.407146931f, -1.f),
    C_Vertex3f(0.846616983f, 0.234870195f, -1.f),
    C_Vertex3f(0.85219264f, 0.264781713f, -1.f),
    C_Vertex3f(0.851007223f, 0.365650892f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.796436071f, 0.124100447f, -1.f),
    C_Vertex3f(0.790527344f, 0.321663857f, -1.f),
    C_Vertex3f(0.790527344f, 0.29691124f, -1.f),
    C_Vertex3f(0.786886692f, 0.266121387f, -1.f),
    C_Vertex3f(0.779605389f, 0.235795498f, -1.f),
    C_Vertex3f(0.776921272f, 0.0986270905f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.779605389f, 0.235795498f, -1.f),
    C_Vertex3f(0.768683434f, 0.205933571f, -1.f),
    C_Vertex3f(0.776921272f, 0.0986270905f, -1.f),
    C_Vertex3f(0.757761478f, 0.183885098f, -1.f),
    C_Vertex3f(0.755821228f, 0.0752983093f, -1.f),
    C_Vertex3f(0.749830127f, 0.169930577f, -1.f),
    C_Vertex3f(0.741248727f, 0.156720519f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.735459089f, 0.0559854507f, -1.f),
    C_Vertex3f(0.755821228f, 0.0752983093f, -1.f),
    C_Vertex3f(0.741248727f, 0.156720519f, -1.f),
    C_Vertex3f(0.732017279f, 0.144254923f, -1.f),
    C_Vertex3f(0.722135782f, 0.132533789f, -1.f),
    C_Vertex3f(0.714632273f, 0.039431572f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.722135782f, 0.132533789f, -1.f),
    C_Vertex3f(0.711604238f, 0.121557117f, -1.f),
    C_Vertex3f(0.714632273f, 0.039431572f, -1.f),
    C_Vertex3f(0.700422645f, 0.111324906f, -1.f),
    C_Vertex3f(0.693340778f, 0.025636673f, -1.f),
    C_Vertex3f(0.688591003f, 0.101837158f, -1.f),
    C_Vertex3f(0.672935486f, 0.0907993317f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.693340778f, 0.025636673f, -1.f),
    C_Vertex3f(0.672935486f, 0.0907993317f, -1.f),
    C_Vertex3f(0.671584606f, 0.0146007538f, -1.f),
    C_Vertex3f(0.655911446f, 0.0808000565f, -1.f),
    C_Vertex3f(0.649363756f, 0.00632381439f, -1.f),
    C_Vertex3f(0.640692711f, 0.074133873f, -1.f),
    C_Vertex3f(0.627279282f, 0.0708007813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.626678228f, 0.000805854797f, -1.f),
    C_Vertex3f(0.649363756f, 0.00632381439f, -1.f),
    C_Vertex3f(0.627279282f, 0.0708007813f, -1.f),
    C_Vertex3f(0.618301392f, 0.0708007813f, -1.f),
    C_Vertex3f(0.612418175f, 0.0718669891f, -1.f),
    C_Vertex3f(0.606999397f, 0.0739994049f, -1.f),
    C_Vertex3f(0.603528023f, -0.001953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.603528023f, -0.001953125f, -1.f),
    C_Vertex3f(0.606999397f, 0.0739994049f, -1.f),
    C_Vertex3f(0.602045059f, 0.0771980286f, -1.f),
    C_Vertex3f(0.597555161f, 0.0814628601f, -1.f),
    C_Vertex3f(0.594187737f, 0.0861377716f, -1.f),
    C_Vertex3f(0.591942787f, 0.0912227631f, -1.f),
    C_Vertex3f(0.590820313f, 0.0967178345f, -1.f),
    C_Vertex3f(0.584177017f, -0.001953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.548828125f, 0.272378922f, -1.f),
    C_Vertex3f(0.54686594f, 0.250087261f, -1.f),
    C_Vertex3f(0.547607422f, 0.00493621826f, -1.f),
    C_Vertex3f(0.558348656f, 0.00149154663f, -1.f),
    C_Vertex3f(0.548828125f, 0.290977478f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.514211655f, 0.0362110138f, -1.f),
    C_Vertex3f(0.517818451f, 0.0292129517f, -1.f),
    C_Vertex3f(0.520240784f, 0.163115501f, -1.f),
    C_Vertex3f(0.523420334f, 0.0216674805f, -1.f),
    C_Vertex3f(0.529206276f, 0.183381081f, -1.f),
    C_Vertex3f(0.530252457f, 0.0151062012f, -1.f),
    C_Vertex3f(0.537055016f, 0.205588341f, -1.f),
    C_Vertex3f(0.538314819f, 0.00952911377f, -1.f),
    C_Vertex3f(0.54294157f, 0.227823734f, -1.f),
    C_Vertex3f(0.547607422f, 0.00493621826f, -1.f),
    C_Vertex3f(0.54686594f, 0.250087261f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.507980347f, 0.0644207001f, -1.f),
    C_Vertex3f(0.509675026f, 0.0530490875f, -1.f),
    C_Vertex3f(0.510158539f, 0.144791603f, -1.f),
    C_Vertex3f(0.511752129f, 0.0436458588f, -1.f),
    C_Vertex3f(0.520240784f, 0.163115501f, -1.f),
    C_Vertex3f(0.514211655f, 0.0362110138f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.510158539f, 0.144791603f, -1.f),
    C_Vertex3f(0.498959541f, 0.128409386f, -1.f),
    C_Vertex3f(0.500230789f, 0.0628948212f, -1.f),
    C_Vertex3f(0.507324219f, 0.0710906982f, -1.f),
    C_Vertex3f(0.507980347f, 0.0644207001f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.434381485f, 0.0739831924f, -1.f),
    C_Vertex3f(0.435033798f, 0.0134496689f, -1.f),
    C_Vertex3f(0.447896957f, 0.0813131332f, -1.f),
    C_Vertex3f(0.452600479f, 0.0233926773f, -1.f),
    C_Vertex3f(0.461112499f, 0.090420723f, -1.f),
    C_Vertex3f(0.469322205f, 0.0349478722f, -1.f),
    C_Vertex3f(0.474028111f, 0.101305962f, -1.f),
    C_Vertex3f(0.485198975f, 0.0481152534f, -1.f),
    C_Vertex3f(0.486643791f, 0.113968849f, -1.f),
    C_Vertex3f(0.500230789f, 0.0628948212f, -1.f),
    C_Vertex3f(0.498959541f, 0.128409386f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.395792007f, 0.0629882813f, -1.f),
    C_Vertex3f(0.399817467f, 0.00102090836f, -1.f),
    C_Vertex3f(0.40832901f, 0.0648207664f, -1.f),
    C_Vertex3f(0.417439461f, 0.00599241257f, -1.f),
    C_Vertex3f(0.421192169f, 0.0684857368f, -1.f),
    C_Vertex3f(0.435033798f, 0.0134496689f, -1.f),
    C_Vertex3f(0.434381485f, 0.0739831924f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.385108948f, 0.0629882813f, -1.f),
    C_Vertex3f(0.376212478f, 0.0640275478f, -1.f),
    C_Vertex3f(0.382167816f, -0.00146484375f, -1.f),
    C_Vertex3f(0.399817467f, 0.00102090836f, -1.f),
    C_Vertex3f(0.395792007f, 0.0629882813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.367574811f, 0.066106081f, -1.f),
    C_Vertex3f(0.359195948f, 0.0692238808f, -1.f),
    C_Vertex3f(0.363618851f, -0.00146484375f, -1.f),
    C_Vertex3f(0.382167816f, -0.00146484375f, -1.f),
    C_Vertex3f(0.376212478f, 0.0640275478f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.351075888f, 0.0733809471f, -1.f),
    C_Vertex3f(0.343214631f, 0.07857728f, -1.f),
    C_Vertex3f(0.344470501f, 0.00137615204f, -1.f),
    C_Vertex3f(0.363618851f, -0.00146484375f, -1.f),
    C_Vertex3f(0.359195948f, 0.0692238808f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.343214631f, 0.07857728f, -1.f),
    C_Vertex3f(0.335612178f, 0.0848128796f, -1.f),
    C_Vertex3f(0.344470501f, 0.00137615204f, -1.f),
    C_Vertex3f(0.328268528f, 0.0920877457f, -1.f),
    C_Vertex3f(0.325594425f, 0.00705814362f, -1.f),
    C_Vertex3f(0.321183681f, 0.100401878f, -1.f),
    C_Vertex3f(0.314984441f, 0.109550238f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.309670806f, 0.119532824f, -1.f),
    C_Vertex3f(0.305242777f, 0.130349636f, -1.f),
    C_Vertex3f(0.306990623f, 0.015581131f, -1.f),
    C_Vertex3f(0.325594425f, 0.00705814362f, -1.f),
    C_Vertex3f(0.314984441f, 0.109550238f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.29633522f, 0.401557207f, -1.f),
    C_Vertex3f(0.285284042f, 0.388472319f, -1.f),
    C_Vertex3f(0.288659096f, 0.0269451141f, -1.f),
    C_Vertex3f(0.27477169f, 0.374301553f, -1.f),
    C_Vertex3f(0.272046089f, 0.0413684845f, -1.f),
    C_Vertex3f(0.264798164f, 0.359044909f, -1.f),
    C_Vertex3f(0.257151604f, 0.0588512421f, -1.f),
    C_Vertex3f(0.255363464f, 0.342702389f, -1.f),
    C_Vertex3f(0.24646759f, 0.325273991f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.257151604f, 0.0588512421f, -1.f),
    C_Vertex3f(0.24646759f, 0.325273991f, -1.f),
    C_Vertex3f(0.243975639f, 0.0793933868f, -1.f),
    C_Vertex3f(0.238683701f, 0.307852387f, -1.f),
    C_Vertex3f(0.232518196f, 0.102994919f, -1.f),
    C_Vertex3f(0.232011795f, 0.290437579f, -1.f),
    C_Vertex3f(0.226451874f, 0.273029566f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.223925114f, 0.127743721f, -1.f),
    C_Vertex3f(0.232518196f, 0.102994919f, -1.f),
    C_Vertex3f(0.226451874f, 0.273029566f, -1.f),
    C_Vertex3f(0.222003937f, 0.255628347f, -1.f),
    C_Vertex3f(0.218667984f, 0.238233924f, -1.f),
    C_Vertex3f(0.218196392f, 0.153639793f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.218196392f, 0.153639793f, -1.f),
    C_Vertex3f(0.218667984f, 0.238233924f, -1.f),
    C_Vertex3f(0.216444016f, 0.220846295f, -1.f),
    C_Vertex3f(0.215332031f, 0.203465462f, -1.f),
    C_Vertex3f(0.215332031f, 0.180683136f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.588867188f, 0.4609375f, -1.f),
    C_Vertex3f(0.590820313f, 0.10049057f, -1.f),
    C_Vertex3f(0.591230631f, 0.103839159f, -1.f),
    C_Vertex3f(0.592051268f, 0.108896017f, -1.f),
    C_Vertex3f(0.593282223f, 0.115661144f, -1.f),
    C_Vertex3f(0.594923496f, 0.124134541f, -1.f),
    C_Vertex3f(0.596975088f, 0.134316206f, -1.f),
    C_Vertex3f(0.599436998f, 0.146206141f, -1.f),
    C_Vertex3f(0.602309227f, 0.159804344f, -1.f),
    C_Vertex3f(0.666992188f, 0.4609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.301700354f, 0.142000675f, -1.f),
    C_Vertex3f(0.306990623f, 0.015581131f, -1.f),
    C_Vertex3f(0.305242777f, 0.130349636f, -1.f),
    C_Vertex3f(0.841736317f, 0.407146931f, -1.f),
    C_Vertex3f(0.838253498f, 0.20584631f, -1.f),
    C_Vertex3f(0.846616983f, 0.234870195f, -1.f),
    C_End(),
    C_Translatef(0.908691406f, 0.f, 0.f),
};
static const Call char_65[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.933440149f, 0.358733237f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Normal3f(0.92576915f, 0.378089279f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Normal3f(-0.933976114f, -0.357335538f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Normal3f(0.940247476f, -0.340491831f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.935599327f, -0.353063881f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Normal3f(-0.935954869f, -0.352120042f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Normal3f(-0.940465152f, -0.339890033f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Normal3f(-0.942786753f, -0.333396405f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Normal3f(-0.945318818f, -0.326147795f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Normal3f(-0.948088706f, -0.318005919f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Normal3f(-0.951127946f, -0.30879733f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Normal3f(-0.954471946f, -0.298300862f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Normal3f(-0.957520068f, -0.288366556f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Normal3f(-0.9601475f, -0.279493541f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.97945708f, -0.201652929f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Normal3f(0.974358559f, -0.225001022f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_Normal3f(0.967271984f, -0.253741771f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Normal3f(0.957781792f, -0.287495911f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Normal3f(0.949109674f, -0.314945877f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_66[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0654296875f, 0.f, 0.f),
    C_Vertex3f(0.0654296875f, 0.f, -1.f),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Normal3f(0.00227682921f, 0.999997437f, 0.f),
    C_Vertex3f(0.314907074f, 0.640625f, 0.f),
    C_Vertex3f(0.314907074f, 0.640625f, -1.f),
    C_Normal3f(0.0535869971f, 0.99856323f, 0.f),
    C_Vertex3f(0.332347989f, 0.640017271f, 0.f),
    C_Vertex3f(0.332347989f, 0.640017271f, -1.f),
    C_Normal3f(0.0940706357f, 0.995565534f, 0.f),
    C_Vertex3f(0.348881125f, 0.638801813f, 0.f),
    C_Vertex3f(0.348881125f, 0.638801813f, -1.f),
    C_Normal3f(0.138842821f, 0.990314364f, 0.f),
    C_Vertex3f(0.364506483f, 0.636978626f, 0.f),
    C_Vertex3f(0.364506483f, 0.636978626f, -1.f),
    C_Normal3f(0.188300446f, 0.982111454f, 0.f),
    C_Vertex3f(0.379224062f, 0.63454771f, 0.f),
    C_Vertex3f(0.379224062f, 0.63454771f, -1.f),
    C_Normal3f(0.24277702f, 0.970082045f, 0.f),
    C_Vertex3f(0.393033862f, 0.631509066f, 0.f),
    C_Vertex3f(0.393033862f, 0.631509066f, -1.f),
    C_Normal3f(0.30247134f, 0.953158498f, 0.f),
    C_Vertex3f(0.405935884f, 0.627862692f, 0.f),
    C_Vertex3f(0.405935884f, 0.627862692f, -1.f),
    C_Normal3f(0.37410599f, 0.927385926f, 0.f),
    C_Vertex3f(0.417930126f, 0.623608589f, 0.f),
    C_Vertex3f(0.417930126f, 0.623608589f, -1.f),
    C_Normal3f(0.474823654f, 0.880081058f, 0.f),
    C_Vertex3f(0.434559822f, 0.616315842f, 0.f),
    C_Vertex3f(0.434559822f, 0.616315842f, -1.f),
    C_Normal3f(0.59248507f, 0.80558151f, 0.f),
    C_Vertex3f(0.454329967f, 0.603970051f, 0.f),
    C_Vertex3f(0.454329967f, 0.603970051f, -1.f),
    C_Normal3f(0.709565461f, 0.704639554f, 0.f),
    C_Vertex3f(0.47169733f, 0.589002132f, 0.f),
    C_Vertex3f(0.47169733f, 0.589002132f, -1.f),
    C_Normal3f(0.808391511f, 0.588645101f, 0.f),
    C_Vertex3f(0.486661911f, 0.571412086f, 0.f),
    C_Vertex3f(0.486661911f, 0.571412086f, -1.f),
    C_Normal3f(0.880777955f, 0.473529458f, 0.f),
    C_Vertex3f(0.499223709f, 0.551199913f, 0.f),
    C_Vertex3f(0.499223709f, 0.551199913f, -1.f),
    C_Normal3f(0.936181247f, 0.351517648f, 0.f),
    C_Vertex3f(0.508645058f, 0.530522823f, 0.f),
    C_Vertex3f(0.508645058f, 0.530522823f, -1.f),
    C_Normal3f(0.976564884f, 0.215223134f, 0.f),
    C_Vertex3f(0.514925957f, 0.509380817f, 0.f),
    C_Vertex3f(0.514925957f, 0.509380817f, -1.f),
    C_Normal3f(0.997333109f, 0.0729850531f, 0.f),
    C_Vertex3f(0.518066406f, 0.487773895f, 0.f),
    C_Vertex3f(0.518066406f, 0.487773895f, -1.f),
    C_Normal3f(0.997725725f, -0.0674047023f, 0.f),
    C_Vertex3f(0.518066406f, 0.466466904f, 0.f),
    C_Vertex3f(0.518066406f, 0.466466904f, -1.f),
    C_Normal3f(0.978093863f, -0.208164185f, 0.f),
    C_Vertex3f(0.51527977f, 0.446526051f, 0.f),
    C_Vertex3f(0.51527977f, 0.446526051f, -1.f),
    C_Normal3f(0.939104795f, -0.34363085f, 0.f),
    C_Vertex3f(0.509706497f, 0.427186489f, 0.f),
    C_Vertex3f(0.509706497f, 0.427186489f, -1.f),
    C_Normal3f(0.883943856f, -0.467593074f, 0.f),
    C_Vertex3f(0.501346588f, 0.408448219f, 0.f),
    C_Vertex3f(0.501346588f, 0.408448219f, -1.f),
    C_Normal3f(0.808460712f, -0.588550031f, 0.f),
    C_Vertex3f(0.490200043f, 0.390311241f, 0.f),
    C_Vertex3f(0.490200043f, 0.390311241f, -1.f),
    C_Normal3f(0.709528685f, -0.704676449f, 0.f),
    C_Vertex3f(0.476212502f, 0.373922825f, 0.f),
    C_Vertex3f(0.476212502f, 0.373922825f, -1.f),
    C_Normal3f(0.602203906f, -0.798342288f, 0.f),
    C_Vertex3f(0.459383965f, 0.35928297f, 0.f),
    C_Vertex3f(0.459383965f, 0.35928297f, -1.f),
    C_Normal3f(0.512611806f, -0.858620465f, 0.f),
    C_Vertex3f(0.439714432f, 0.346391678f, 0.f),
    C_Vertex3f(0.439714432f, 0.346391678f, -1.f),
    C_Normal3f(0.443625778f, -0.896212161f, 0.f),
    C_Vertex3f(0.428459167f, 0.340820313f, 0.f),
    C_Vertex3f(0.428459167f, 0.340820313f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.280859202f, 0.959748983f, 0.f),
    C_Vertex3f(0.428459167f, 0.340820313f, 0.f),
    C_Vertex3f(0.428459167f, 0.340820313f, -1.f),
    C_Normal3f(0.322010159f, 0.946736217f, 0.f),
    C_Vertex3f(0.435739517f, 0.338689804f, 0.f),
    C_Vertex3f(0.435739517f, 0.338689804f, -1.f),
    C_Normal3f(0.375480443f, 0.926830351f, 0.f),
    C_Vertex3f(0.44945848f, 0.333677888f, 0.f),
    C_Vertex3f(0.44945848f, 0.333677888f, -1.f),
    C_Normal3f(0.442026228f, 0.897002101f, 0.f),
    C_Vertex3f(0.462335706f, 0.327915072f, 0.f),
    C_Vertex3f(0.462335706f, 0.327915072f, -1.f),
    C_Normal3f(0.510154545f, 0.860082746f, 0.f),
    C_Vertex3f(0.474371195f, 0.321401358f, 0.f),
    C_Vertex3f(0.474371195f, 0.321401358f, -1.f),
    C_Normal3f(0.578484714f, 0.815693259f, 0.f),
    C_Vertex3f(0.485564947f, 0.314136744f, 0.f),
    C_Vertex3f(0.485564947f, 0.314136744f, -1.f),
    C_Normal3f(0.645390093f, 0.763853192f, 0.f),
    C_Vertex3f(0.495916963f, 0.30612123f, 0.f),
    C_Vertex3f(0.495916963f, 0.30612123f, -1.f),
    C_Normal3f(0.709138215f, 0.705069482f, 0.f),
    C_Vertex3f(0.505427241f, 0.297354817f, 0.f),
    C_Vertex3f(0.505427241f, 0.297354817f, -1.f),
    C_Normal3f(0.768079758f, 0.640354156f, 0.f),
    C_Vertex3f(0.514095783f, 0.287837505f, 0.f),
    C_Vertex3f(0.514095783f, 0.287837505f, -1.f),
    C_Normal3f(0.819691777f, 0.572804809f, 0.f),
    C_Vertex3f(0.521922588f, 0.277569294f, 0.f),
    C_Vertex3f(0.521922588f, 0.277569294f, -1.f),
    C_Normal3f(0.865032375f, 0.501716018f, 0.f),
    C_Vertex3f(0.528771043f, 0.266837001f, 0.f),
    C_Vertex3f(0.528771043f, 0.266837001f, -1.f),
    C_Normal3f(0.90473175f, 0.42598179f, 0.f),
    C_Vertex3f(0.534641147f, 0.255640626f, 0.f),
    C_Vertex3f(0.534641147f, 0.255640626f, -1.f),
    C_Normal3f(0.937801421f, 0.347172141f, 0.f),
    C_Vertex3f(0.5395329f, 0.243980169f, 0.f),
    C_Vertex3f(0.5395329f, 0.243980169f, -1.f),
    C_Normal3f(0.963681757f, 0.267053306f, 0.f),
    C_Vertex3f(0.543446302f, 0.231855631f, 0.f),
    C_Vertex3f(0.543446302f, 0.231855631f, -1.f),
    C_Normal3f(0.982284367f, 0.187396452f, 0.f),
    C_Vertex3f(0.546381354f, 0.219267011f, 0.f),
    C_Vertex3f(0.546381354f, 0.219267011f, -1.f),
    C_Normal3f(0.993953824f, 0.109799117f, 0.f),
    C_Vertex3f(0.548338056f, 0.206214309f, 0.f),
    C_Vertex3f(0.548338056f, 0.206214309f, -1.f),
    C_Normal3f(0.999525905f, 0.0307911802f, 0.f),
    C_Vertex3f(0.549316406f, 0.192697525f, 0.f),
    C_Vertex3f(0.549316406f, 0.192697525f, -1.f),
    C_Normal3f(0.998234808f, -0.0593903735f, 0.f),
    C_Vertex3f(0.549316406f, 0.174455643f, 0.f),
    C_Vertex3f(0.549316406f, 0.174455643f, -1.f),
    C_Normal3f(0.986296535f, -0.164982155f, 0.f),
    C_Vertex3f(0.546939373f, 0.152744293f, 0.f),
    C_Vertex3f(0.546939373f, 0.152744293f, -1.f),
    C_Normal3f(0.960546494f, -0.278119296f, 0.f),
    C_Vertex3f(0.542185307f, 0.131824493f, 0.f),
    C_Vertex3f(0.542185307f, 0.131824493f, -1.f),
    C_Normal3f(0.921448112f, -0.388501436f, 0.f),
    C_Vertex3f(0.535054207f, 0.111696243f, 0.f),
    C_Vertex3f(0.535054207f, 0.111696243f, -1.f),
    C_Normal3f(0.875379264f, -0.483436853f, 0.f),
    C_Vertex3f(0.525546074f, 0.0923595428f, 0.f),
    C_Vertex3f(0.525546074f, 0.0923595428f, -1.f),
    C_Normal3f(0.819832265f, -0.572603762f, 0.f),
    C_Vertex3f(0.514918327f, 0.0752353668f, 0.f),
    C_Vertex3f(0.514918327f, 0.0752353668f, -1.f),
    C_Normal3f(0.746449828f, -0.665441751f, 0.f),
    C_Vertex3f(0.503170967f, 0.0603237152f, 0.f),
    C_Vertex3f(0.503170967f, 0.0603237152f, -1.f),
    C_Normal3f(0.653523266f, -0.75690639f, 0.f),
    C_Vertex3f(0.490303993f, 0.047624588f, 0.f),
    C_Vertex3f(0.490303993f, 0.047624588f, -1.f),
    C_Normal3f(0.55191493f, -0.833900511f, 0.f),
    C_Vertex3f(0.476317406f, 0.0371379852f, 0.f),
    C_Vertex3f(0.476317406f, 0.0371379852f, -1.f),
    C_Normal3f(0.457255363f, -0.889335454f, 0.f),
    C_Vertex3f(0.4605546f, 0.0279350281f, 0.f),
    C_Vertex3f(0.4605546f, 0.0279350281f, -1.f),
    C_Normal3f(0.36732462f, -0.930092812f, 0.f),
    C_Vertex3f(0.443015575f, 0.0200157166f, 0.f),
    C_Vertex3f(0.443015575f, 0.0200157166f, -1.f),
    C_Normal3f(0.284422666f, -0.958698988f, 0.f),
    C_Vertex3f(0.423700333f, 0.0133800507f, 0.f),
    C_Vertex3f(0.423700333f, 0.0133800507f, -1.f),
    C_Normal3f(0.205696955f, -0.978615701f, 0.f),
    C_Vertex3f(0.402608871f, 0.0080280304f, 0.f),
    C_Vertex3f(0.402608871f, 0.0080280304f, -1.f),
    C_Normal3f(0.134434149f, -0.990922451f, 0.f),
    C_Vertex3f(0.379140854f, 0.0040140152f, 0.f),
    C_Vertex3f(0.379140854f, 0.0040140152f, -1.f),
    C_Normal3f(0.0740395114f, -0.997255385f, 0.f),
    C_Vertex3f(0.35329628f, 0.00133800507f, 0.f),
    C_Vertex3f(0.35329628f, 0.00133800507f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.32507515f, 0.f, 0.f),
    C_Vertex3f(0.32507515f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.32507515f, 0.f, 0.f),
    C_Vertex3f(0.32507515f, 0.f, -1.f),
    C_Vertex3f(0.0654296875f, 0.f, -1.f),
    C_Vertex3f(0.0654296875f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.150390625f, 0.371582031f, 0.f),
    C_Vertex3f(0.150390625f, 0.371582031f, -1.f),
    C_Normal3f(-0.00146206818f, 0.999998927f, 0.f),
    C_Vertex3f(0.295811653f, 0.371582031f, 0.f),
    C_Vertex3f(0.295811653f, 0.371582031f, -1.f),
    C_Normal3f(-0.0276122596f, 0.999618709f, 0.f),
    C_Vertex3f(0.308893919f, 0.371813774f, 0.f),
    C_Vertex3f(0.308893919f, 0.371813774f, -1.f),
    C_Normal3f(-0.0499319769f, 0.998752594f, 0.f),
    C_Vertex3f(0.32098031f, 0.37227726f, 0.f),
    C_Vertex3f(0.32098031f, 0.37227726f, -1.f),
    C_Normal3f(-0.0763489679f, 0.99708122f, 0.f),
    C_Vertex3f(0.332070827f, 0.372972488f, 0.f),
    C_Vertex3f(0.332070827f, 0.372972488f, -1.f),
    C_Normal3f(-0.108030796f, 0.994147599f, 0.f),
    C_Vertex3f(0.34216547f, 0.37389946f, 0.f),
    C_Vertex3f(0.34216547f, 0.37389946f, -1.f),
    C_Normal3f(-0.146592394f, 0.989197016f, 0.f),
    C_Vertex3f(0.351264238f, 0.375058174f, 0.f),
    C_Vertex3f(0.351264238f, 0.375058174f, -1.f),
    C_Normal3f(-0.194297239f, 0.980942726f, 0.f),
    C_Vertex3f(0.359367132f, 0.376448631f, 0.f),
    C_Vertex3f(0.359367132f, 0.376448631f, -1.f),
    C_Normal3f(-0.261956841f, 0.965079546f, 0.f),
    C_Vertex3f(0.366474152f, 0.378070831f, 0.f),
    C_Vertex3f(0.366474152f, 0.378070831f, -1.f),
    C_Normal3f(-0.35850963f, 0.933526039f, 0.f),
    C_Vertex3f(0.377605438f, 0.381399155f, 0.f),
    C_Vertex3f(0.377605438f, 0.381399155f, -1.f),
    C_Normal3f(-0.478592545f, 0.878037214f, 0.f),
    C_Vertex3f(0.391765118f, 0.387783527f, 0.f),
    C_Vertex3f(0.391765118f, 0.387783527f, -1.f),
    C_Normal3f(-0.616899133f, 0.78704226f, 0.f),
    C_Vertex3f(0.403933048f, 0.395749569f, 0.f),
    C_Vertex3f(0.403933048f, 0.395749569f, -1.f),
    C_Normal3f(-0.747750759f, 0.66397953f, 0.f),
    C_Vertex3f(0.41410923f, 0.405297279f, 0.f),
    C_Vertex3f(0.41410923f, 0.405297279f, -1.f),
    C_Normal3f(-0.855641603f, 0.517568648f, 0.f),
    C_Vertex3f(0.422293663f, 0.416426659f, 0.f),
    C_Vertex3f(0.422293663f, 0.416426659f, -1.f),
    C_Normal3f(-0.932973266f, 0.359945714f, 0.f),
    C_Vertex3f(0.428431988f, 0.428975582f, 0.f),
    C_Vertex3f(0.428431988f, 0.428975582f, -1.f),
    C_Normal3f(-0.978831053f, 0.204669729f, 0.f),
    C_Vertex3f(0.432524204f, 0.44294405f, 0.f),
    C_Vertex3f(0.432524204f, 0.44294405f, -1.f),
    C_Normal3f(-0.997930765f, 0.06429708f, 0.f),
    C_Vertex3f(0.434570313f, 0.458332062f, 0.f),
    C_Vertex3f(0.434570313f, 0.458332062f, -1.f),
    C_Normal3f(-0.998149037f, -0.0608146004f, 0.f),
    C_Vertex3f(0.434570313f, 0.474700928f, 0.f),
    C_Vertex3f(0.434570313f, 0.474700928f, -1.f),
    C_Normal3f(-0.98103404f, -0.193835288f, 0.f),
    C_Vertex3f(0.43266058f, 0.489676476f, 0.f),
    C_Vertex3f(0.43266058f, 0.489676476f, -1.f),
    C_Normal3f(-0.943116248f, -0.332463175f, 0.f),
    C_Vertex3f(0.428841114f, 0.503697395f, 0.f),
    C_Vertex3f(0.428841114f, 0.503697395f, -1.f),
    C_Normal3f(-0.88322705f, -0.468945712f, 0.f),
    C_Vertex3f(0.423111916f, 0.516763687f, 0.f),
    C_Vertex3f(0.423111916f, 0.516763687f, -1.f),
    C_Normal3f(-0.796916008f, -0.604090095f, 0.f),
    C_Vertex3f(0.415472984f, 0.528875351f, 0.f),
    C_Vertex3f(0.415472984f, 0.528875351f, -1.f),
    C_Normal3f(-0.675509274f, -0.737351418f, 0.f),
    C_Vertex3f(0.406197071f, 0.539077759f, 0.f),
    C_Vertex3f(0.406197071f, 0.539077759f, -1.f),
    C_Normal3f(-0.53033185f, -0.847790122f, 0.f),
    C_Vertex3f(0.395284176f, 0.547370911f, 0.f),
    C_Vertex3f(0.395284176f, 0.547370911f, -1.f),
    C_Normal3f(-0.387234747f, -0.921981156f, 0.f),
    C_Vertex3f(0.382734299f, 0.553754807f, 0.f),
    C_Vertex3f(0.382734299f, 0.553754807f, -1.f),
    C_Normal3f(-0.269298673f, -0.963056684f, 0.f),
    C_Vertex3f(0.372094154f, 0.557110786f, 0.f),
    C_Vertex3f(0.372094154f, 0.557110786f, -1.f),
    C_Normal3f(-0.198051453f, -0.980191648f, 0.f),
    C_Vertex3f(0.363731861f, 0.559068441f, 0.f),
    C_Vertex3f(0.363731861f, 0.559068441f, -1.f),
    C_Normal3f(-0.148182079f, -0.988960087f, 0.f),
    C_Vertex3f(0.354100704f, 0.560746431f, 0.f),
    C_Vertex3f(0.354100704f, 0.560746431f, -1.f),
    C_Normal3f(-0.10846357f, -0.994100451f, 0.f),
    C_Vertex3f(0.343200684f, 0.562144756f, 0.f),
    C_Vertex3f(0.343200684f, 0.562144756f, -1.f),
    C_Normal3f(-0.0762286484f, -0.997090399f, 0.f),
    C_Vertex3f(0.331031799f, 0.563263416f, 0.f),
    C_Vertex3f(0.331031799f, 0.563263416f, -1.f),
    C_Normal3f(-0.0496228114f, -0.998767972f, 0.f),
    C_Vertex3f(0.317594051f, 0.564102411f, 0.f),
    C_Vertex3f(0.317594051f, 0.564102411f, -1.f),
    C_Normal3f(-0.0273345634f, -0.999626338f, 0.f),
    C_Vertex3f(0.30288744f, 0.564661741f, 0.f),
    C_Vertex3f(0.30288744f, 0.564661741f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.286911964f, 0.564941406f, 0.f),
    C_Vertex3f(0.286911964f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.286911964f, 0.564941406f, 0.f),
    C_Vertex3f(0.286911964f, 0.564941406f, -1.f),
    C_Vertex3f(0.150390625f, 0.564941406f, -1.f),
    C_Vertex3f(0.150390625f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.150390625f, 0.564941406f, 0.f),
    C_Vertex3f(0.150390625f, 0.564941406f, -1.f),
    C_Vertex3f(0.150390625f, 0.371582031f, -1.f),
    C_Vertex3f(0.150390625f, 0.371582031f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, 0.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, -1.f),
    C_Normal3f(-0.00203977409f, 0.999997973f, 0.f),
    C_Vertex3f(0.320371628f, 0.0756835938f, 0.f),
    C_Vertex3f(0.320371628f, 0.0756835938f, -1.f),
    C_Normal3f(-0.0358958766f, 0.999355555f, 0.f),
    C_Vertex3f(0.337873459f, 0.0760660172f, 0.f),
    C_Vertex3f(0.337873459f, 0.0760660172f, -1.f),
    C_Normal3f(-0.0738703236f, 0.997267842f, 0.f),
    C_Vertex3f(0.352312088f, 0.076830864f, 0.f),
    C_Vertex3f(0.352312088f, 0.076830864f, -1.f),
    C_Normal3f(-0.139587745f, 0.990209639f, 0.f),
    C_Vertex3f(0.363687515f, 0.0779781342f, 0.f),
    C_Vertex3f(0.363687515f, 0.0779781342f, -1.f),
    C_Normal3f(-0.217983484f, 0.975952446f, 0.f),
    C_Vertex3f(0.375171661f, 0.0800533295f, 0.f),
    C_Vertex3f(0.375171661f, 0.0800533295f, -1.f),
    C_Normal3f(-0.292538524f, 0.956253767f, 0.f),
    C_Vertex3f(0.388625145f, 0.083548069f, 0.f),
    C_Vertex3f(0.388625145f, 0.083548069f, -1.f),
    C_Normal3f(-0.381353974f, 0.924429119f, 0.f),
    C_Vertex3f(0.400876045f, 0.087916851f, 0.f),
    C_Vertex3f(0.400876045f, 0.087916851f, -1.f),
    C_Normal3f(-0.477650583f, 0.878549874f, 0.f),
    C_Vertex3f(0.411924362f, 0.0931596756f, 0.f),
    C_Vertex3f(0.411924362f, 0.0931596756f, -1.f),
    C_Normal3f(-0.586543679f, 0.809917569f, 0.f),
    C_Vertex3f(0.421770096f, 0.0992765427f, 0.f),
    C_Vertex3f(0.421770096f, 0.0992765427f, -1.f),
    C_Normal3f(-0.69318682f, 0.720758021f, 0.f),
    C_Vertex3f(0.430740356f, 0.106786251f, 0.f),
    C_Vertex3f(0.430740356f, 0.106786251f, -1.f),
    C_Normal3f(-0.781746984f, 0.623595715f, 0.f),
    C_Vertex3f(0.438835144f, 0.115688801f, 0.f),
    C_Vertex3f(0.438835144f, 0.115688801f, -1.f),
    C_Normal3f(-0.851056933f, 0.525073469f, 0.f),
    C_Vertex3f(0.446054459f, 0.125984192f, 0.f),
    C_Vertex3f(0.446054459f, 0.125984192f, -1.f),
    C_Normal3f(-0.909426928f, 0.415863574f, 0.f),
    C_Vertex3f(0.4523983f, 0.137672424f, 0.f),
    C_Vertex3f(0.4523983f, 0.137672424f, -1.f),
    C_Normal3f(-0.956751823f, 0.290905446f, 0.f),
    C_Vertex3f(0.457156181f, 0.150261879f, 0.f),
    C_Vertex3f(0.457156181f, 0.150261879f, -1.f),
    C_Normal3f(-0.98575145f, 0.168208569f, 0.f),
    C_Vertex3f(0.460328102f, 0.163752556f, 0.f),
    C_Vertex3f(0.460328102f, 0.163752556f, -1.f),
    C_Normal3f(-0.998693526f, 0.0511008054f, 0.f),
    C_Vertex3f(0.461914063f, 0.178144455f, 0.f),
    C_Vertex3f(0.461914063f, 0.178144455f, -1.f),
    C_Normal3f(-0.997553825f, -0.0699026659f, 0.f),
    C_Vertex3f(0.461914063f, 0.194747925f, 0.f),
    C_Vertex3f(0.461914063f, 0.194747925f, -1.f),
    C_Normal3f(-0.977260888f, -0.212040365f, 0.f),
    C_Vertex3f(0.45957756f, 0.21148777f, 0.f),
    C_Vertex3f(0.45957756f, 0.21148777f, -1.f),
    C_Normal3f(-0.931665301f, -0.363317758f, 0.f),
    C_Vertex3f(0.454904556f, 0.227053642f, 0.f),
    C_Vertex3f(0.454904556f, 0.227053642f, -1.f),
    C_Normal3f(-0.860371709f, -0.509667158f, 0.f),
    C_Vertex3f(0.44789505f, 0.241445541f, 0.f),
    C_Vertex3f(0.44789505f, 0.241445541f, -1.f),
    C_Normal3f(-0.76776731f, -0.640728712f, 0.f),
    C_Vertex3f(0.438549042f, 0.254663467f, 0.f),
    C_Vertex3f(0.438549042f, 0.254663467f, -1.f),
    C_Normal3f(-0.652069747f, -0.758159041f, 0.f),
    C_Vertex3f(0.427474499f, 0.265914917f, 0.f),
    C_Vertex3f(0.427474499f, 0.265914917f, -1.f),
    C_Normal3f(-0.519147336f, -0.854684711f, 0.f),
    C_Vertex3f(0.414671421f, 0.27519989f, 0.f),
    C_Vertex3f(0.414671421f, 0.27519989f, -1.f),
    C_Normal3f(-0.38053301f, -0.924767375f, 0.f),
    C_Vertex3f(0.400139809f, 0.282518387f, 0.f),
    C_Vertex3f(0.400139809f, 0.282518387f, -1.f),
    C_Normal3f(-0.251164079f, -0.967944562f, 0.f),
    C_Vertex3f(0.383879662f, 0.287870407f, 0.f),
    C_Vertex3f(0.383879662f, 0.287870407f, -1.f),
    C_Normal3f(-0.152882516f, -0.988244414f, 0.f),
    C_Vertex3f(0.364044666f, 0.291884422f, 0.f),
    C_Vertex3f(0.364044666f, 0.291884422f, -1.f),
    C_Normal3f(-0.0794003159f, -0.996842742f, 0.f),
    C_Vertex3f(0.340634823f, 0.294560432f, 0.f),
    C_Vertex3f(0.340634823f, 0.294560432f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.313650131f, 0.295898438f, 0.f),
    C_Vertex3f(0.313650131f, 0.295898438f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.313650131f, 0.295898438f, 0.f),
    C_Vertex3f(0.313650131f, 0.295898438f, -1.f),
    C_Vertex3f(0.150390625f, 0.295898438f, -1.f),
    C_Vertex3f(0.150390625f, 0.295898438f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.150390625f, 0.295898438f, 0.f),
    C_Vertex3f(0.150390625f, 0.295898438f, -1.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, -1.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.454904556f, 0.227053642f, 0.f),
    C_Vertex3f(0.45957756f, 0.21148777f, 0.f),
    C_Vertex3f(0.462335706f, 0.327915072f, 0.f),
    C_Vertex3f(0.461914063f, 0.194747925f, 0.f),
    C_Vertex3f(0.476317406f, 0.0371379852f, 0.f),
    C_Vertex3f(0.461914063f, 0.178144455f, 0.f),
    C_Vertex3f(0.4605546f, 0.0279350281f, 0.f),
    C_Vertex3f(0.460328102f, 0.163752556f, 0.f),
    C_Vertex3f(0.457156181f, 0.150261879f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.44945848f, 0.333677888f, 0.f),
    C_Vertex3f(0.435739517f, 0.338689804f, 0.f),
    C_Vertex3f(0.438549042f, 0.254663467f, 0.f),
    C_Vertex3f(0.44789505f, 0.241445541f, 0.f),
    C_Vertex3f(0.454904556f, 0.227053642f, 0.f),
    C_Vertex3f(0.462335706f, 0.327915072f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.428459167f, 0.340820313f, 0.f),
    C_Vertex3f(0.439714432f, 0.346391678f, 0.f),
    C_Vertex3f(0.432524204f, 0.44294405f, 0.f),
    C_Vertex3f(0.428431988f, 0.428975582f, 0.f),
    C_Vertex3f(0.427474499f, 0.265914917f, 0.f),
    C_Vertex3f(0.438549042f, 0.254663467f, 0.f),
    C_Vertex3f(0.435739517f, 0.338689804f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.422293663f, 0.416426659f, 0.f),
    C_Vertex3f(0.41410923f, 0.405297279f, 0.f),
    C_Vertex3f(0.414671421f, 0.27519989f, 0.f),
    C_Vertex3f(0.427474499f, 0.265914917f, 0.f),
    C_Vertex3f(0.428431988f, 0.428975582f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.377605438f, 0.381399155f, 0.f),
    C_Vertex3f(0.383879662f, 0.287870407f, 0.f),
    C_Vertex3f(0.391765118f, 0.387783527f, 0.f),
    C_Vertex3f(0.400139809f, 0.282518387f, 0.f),
    C_Vertex3f(0.403933048f, 0.395749569f, 0.f),
    C_Vertex3f(0.414671421f, 0.27519989f, 0.f),
    C_Vertex3f(0.41410923f, 0.405297279f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.366474152f, 0.378070831f, 0.f),
    C_Vertex3f(0.359367132f, 0.376448631f, 0.f),
    C_Vertex3f(0.364044666f, 0.291884422f, 0.f),
    C_Vertex3f(0.383879662f, 0.287870407f, 0.f),
    C_Vertex3f(0.377605438f, 0.381399155f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.359367132f, 0.376448631f, 0.f),
    C_Vertex3f(0.351264238f, 0.375058174f, 0.f),
    C_Vertex3f(0.364044666f, 0.291884422f, 0.f),
    C_Vertex3f(0.34216547f, 0.37389946f, 0.f),
    C_Vertex3f(0.340634823f, 0.294560432f, 0.f),
    C_Vertex3f(0.332070827f, 0.372972488f, 0.f),
    C_Vertex3f(0.32098031f, 0.37227726f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.313650131f, 0.295898438f, 0.f),
    C_Vertex3f(0.340634823f, 0.294560432f, 0.f),
    C_Vertex3f(0.32098031f, 0.37227726f, 0.f),
    C_Vertex3f(0.308893919f, 0.371813774f, 0.f),
    C_Vertex3f(0.295811653f, 0.371582031f, 0.f),
    C_Vertex3f(0.150390625f, 0.371582031f, 0.f),
    C_Vertex3f(0.150390625f, 0.295898438f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.548338056f, 0.206214309f, 0.f),
    C_Vertex3f(0.546381354f, 0.219267011f, 0.f),
    C_Vertex3f(0.546939373f, 0.152744293f, 0.f),
    C_Vertex3f(0.549316406f, 0.174455643f, 0.f),
    C_Vertex3f(0.549316406f, 0.192697525f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.534641147f, 0.255640626f, 0.f),
    C_Vertex3f(0.535054207f, 0.111696243f, 0.f),
    C_Vertex3f(0.5395329f, 0.243980169f, 0.f),
    C_Vertex3f(0.542185307f, 0.131824493f, 0.f),
    C_Vertex3f(0.543446302f, 0.231855631f, 0.f),
    C_Vertex3f(0.546939373f, 0.152744293f, 0.f),
    C_Vertex3f(0.546381354f, 0.219267011f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.514095783f, 0.287837505f, 0.f),
    C_Vertex3f(0.514918327f, 0.0752353668f, 0.f),
    C_Vertex3f(0.521922588f, 0.277569294f, 0.f),
    C_Vertex3f(0.525546074f, 0.0923595428f, 0.f),
    C_Vertex3f(0.528771043f, 0.266837001f, 0.f),
    C_Vertex3f(0.535054207f, 0.111696243f, 0.f),
    C_Vertex3f(0.534641147f, 0.255640626f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.474371195f, 0.321401358f, 0.f),
    C_Vertex3f(0.476317406f, 0.0371379852f, 0.f),
    C_Vertex3f(0.485564947f, 0.314136744f, 0.f),
    C_Vertex3f(0.490303993f, 0.047624588f, 0.f),
    C_Vertex3f(0.495916963f, 0.30612123f, 0.f),
    C_Vertex3f(0.503170967f, 0.0603237152f, 0.f),
    C_Vertex3f(0.505427241f, 0.297354817f, 0.f),
    C_Vertex3f(0.514918327f, 0.0752353668f, 0.f),
    C_Vertex3f(0.514095783f, 0.287837505f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.457156181f, 0.150261879f, 0.f),
    C_Vertex3f(0.4523983f, 0.137672424f, 0.f),
    C_Vertex3f(0.4605546f, 0.0279350281f, 0.f),
    C_Vertex3f(0.446054459f, 0.125984192f, 0.f),
    C_Vertex3f(0.443015575f, 0.0200157166f, 0.f),
    C_Vertex3f(0.438835144f, 0.115688801f, 0.f),
    C_Vertex3f(0.430740356f, 0.106786251f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.423700333f, 0.0133800507f, 0.f),
    C_Vertex3f(0.443015575f, 0.0200157166f, 0.f),
    C_Vertex3f(0.430740356f, 0.106786251f, 0.f),
    C_Vertex3f(0.421770096f, 0.0992765427f, 0.f),
    C_Vertex3f(0.411924362f, 0.0931596756f, 0.f),
    C_Vertex3f(0.402608871f, 0.0080280304f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.411924362f, 0.0931596756f, 0.f),
    C_Vertex3f(0.400876045f, 0.087916851f, 0.f),
    C_Vertex3f(0.402608871f, 0.0080280304f, 0.f),
    C_Vertex3f(0.388625145f, 0.083548069f, 0.f),
    C_Vertex3f(0.379140854f, 0.0040140152f, 0.f),
    C_Vertex3f(0.375171661f, 0.0800533295f, 0.f),
    C_Vertex3f(0.363687515f, 0.0779781342f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35329628f, 0.00133800507f, 0.f),
    C_Vertex3f(0.379140854f, 0.0040140152f, 0.f),
    C_Vertex3f(0.363687515f, 0.0779781342f, 0.f),
    C_Vertex3f(0.352312088f, 0.076830864f, 0.f),
    C_Vertex3f(0.337873459f, 0.0760660172f, 0.f),
    C_Vertex3f(0.32507515f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_Vertex3f(0.0654296875f, 0.f, 0.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, 0.f),
    C_Vertex3f(0.32507515f, 0.f, 0.f),
    C_Vertex3f(0.320371628f, 0.0756835938f, 0.f),
    C_Vertex3f(0.337873459f, 0.0760660172f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.518066406f, 0.466466904f, 0.f),
    C_Vertex3f(0.518066406f, 0.487773895f, 0.f),
    C_Vertex3f(0.51527977f, 0.446526051f, 0.f),
    C_Vertex3f(0.514925957f, 0.509380817f, 0.f),
    C_Vertex3f(0.509706497f, 0.427186489f, 0.f),
    C_Vertex3f(0.508645058f, 0.530522823f, 0.f),
    C_Vertex3f(0.501346588f, 0.408448219f, 0.f),
    C_Vertex3f(0.499223709f, 0.551199913f, 0.f),
    C_Vertex3f(0.490200043f, 0.390311241f, 0.f),
    C_Vertex3f(0.486661911f, 0.571412086f, 0.f),
    C_Vertex3f(0.476212502f, 0.373922825f, 0.f),
    C_Vertex3f(0.47169733f, 0.589002132f, 0.f),
    C_Vertex3f(0.459383965f, 0.35928297f, 0.f),
    C_Vertex3f(0.454329967f, 0.603970051f, 0.f),
    C_Vertex3f(0.439714432f, 0.346391678f, 0.f),
    C_Vertex3f(0.434570313f, 0.474700928f, 0.f),
    C_Vertex3f(0.434570313f, 0.458332062f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434559822f, 0.616315842f, 0.f),
    C_Vertex3f(0.417930126f, 0.623608589f, 0.f),
    C_Vertex3f(0.423111916f, 0.516763687f, 0.f),
    C_Vertex3f(0.428841114f, 0.503697395f, 0.f),
    C_Vertex3f(0.43266058f, 0.489676476f, 0.f),
    C_Vertex3f(0.434570313f, 0.474700928f, 0.f),
    C_Vertex3f(0.454329967f, 0.603970051f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.354100704f, 0.560746431f, 0.f),
    C_Vertex3f(0.363731861f, 0.559068441f, 0.f),
    C_Vertex3f(0.364506483f, 0.636978626f, 0.f),
    C_Vertex3f(0.372094154f, 0.557110786f, 0.f),
    C_Vertex3f(0.379224062f, 0.63454771f, 0.f),
    C_Vertex3f(0.382734299f, 0.553754807f, 0.f),
    C_Vertex3f(0.393033862f, 0.631509066f, 0.f),
    C_Vertex3f(0.395284176f, 0.547370911f, 0.f),
    C_Vertex3f(0.405935884f, 0.627862692f, 0.f),
    C_Vertex3f(0.406197071f, 0.539077759f, 0.f),
    C_Vertex3f(0.417930126f, 0.623608589f, 0.f),
    C_Vertex3f(0.415472984f, 0.528875351f, 0.f),
    C_Vertex3f(0.423111916f, 0.516763687f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.317594051f, 0.564102411f, 0.f),
    C_Vertex3f(0.331031799f, 0.563263416f, 0.f),
    C_Vertex3f(0.332347989f, 0.640017271f, 0.f),
    C_Vertex3f(0.343200684f, 0.562144756f, 0.f),
    C_Vertex3f(0.348881125f, 0.638801813f, 0.f),
    C_Vertex3f(0.354100704f, 0.560746431f, 0.f),
    C_Vertex3f(0.364506483f, 0.636978626f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.314907074f, 0.640625f, 0.f),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_Vertex3f(0.150390625f, 0.564941406f, 0.f),
    C_Vertex3f(0.286911964f, 0.564941406f, 0.f),
    C_Vertex3f(0.30288744f, 0.564661741f, 0.f),
    C_Vertex3f(0.317594051f, 0.564102411f, 0.f),
    C_Vertex3f(0.332347989f, 0.640017271f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150390625f, 0.564941406f, 0.f),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, 0.f),
    C_Vertex3f(0.150390625f, 0.295898438f, 0.f),
    C_Vertex3f(0.150390625f, 0.371582031f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.432524204f, 0.44294405f, 0.f),
    C_Vertex3f(0.439714432f, 0.346391678f, 0.f),
    C_Vertex3f(0.434570313f, 0.458332062f, 0.f),
    C_Vertex3f(0.462335706f, 0.327915072f, 0.f),
    C_Vertex3f(0.476317406f, 0.0371379852f, 0.f),
    C_Vertex3f(0.474371195f, 0.321401358f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.454904556f, 0.227053642f, -1.f),
    C_Vertex3f(0.45957756f, 0.21148777f, -1.f),
    C_Vertex3f(0.462335706f, 0.327915072f, -1.f),
    C_Vertex3f(0.461914063f, 0.194747925f, -1.f),
    C_Vertex3f(0.476317406f, 0.0371379852f, -1.f),
    C_Vertex3f(0.461914063f, 0.178144455f, -1.f),
    C_Vertex3f(0.4605546f, 0.0279350281f, -1.f),
    C_Vertex3f(0.460328102f, 0.163752556f, -1.f),
    C_Vertex3f(0.457156181f, 0.150261879f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.44945848f, 0.333677888f, -1.f),
    C_Vertex3f(0.435739517f, 0.338689804f, -1.f),
    C_Vertex3f(0.438549042f, 0.254663467f, -1.f),
    C_Vertex3f(0.44789505f, 0.241445541f, -1.f),
    C_Vertex3f(0.454904556f, 0.227053642f, -1.f),
    C_Vertex3f(0.462335706f, 0.327915072f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.428459167f, 0.340820313f, -1.f),
    C_Vertex3f(0.439714432f, 0.346391678f, -1.f),
    C_Vertex3f(0.432524204f, 0.44294405f, -1.f),
    C_Vertex3f(0.428431988f, 0.428975582f, -1.f),
    C_Vertex3f(0.427474499f, 0.265914917f, -1.f),
    C_Vertex3f(0.438549042f, 0.254663467f, -1.f),
    C_Vertex3f(0.435739517f, 0.338689804f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.422293663f, 0.416426659f, -1.f),
    C_Vertex3f(0.41410923f, 0.405297279f, -1.f),
    C_Vertex3f(0.414671421f, 0.27519989f, -1.f),
    C_Vertex3f(0.427474499f, 0.265914917f, -1.f),
    C_Vertex3f(0.428431988f, 0.428975582f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.377605438f, 0.381399155f, -1.f),
    C_Vertex3f(0.383879662f, 0.287870407f, -1.f),
    C_Vertex3f(0.391765118f, 0.387783527f, -1.f),
    C_Vertex3f(0.400139809f, 0.282518387f, -1.f),
    C_Vertex3f(0.403933048f, 0.395749569f, -1.f),
    C_Vertex3f(0.414671421f, 0.27519989f, -1.f),
    C_Vertex3f(0.41410923f, 0.405297279f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.366474152f, 0.378070831f, -1.f),
    C_Vertex3f(0.359367132f, 0.376448631f, -1.f),
    C_Vertex3f(0.364044666f, 0.291884422f, -1.f),
    C_Vertex3f(0.383879662f, 0.287870407f, -1.f),
    C_Vertex3f(0.377605438f, 0.381399155f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.359367132f, 0.376448631f, -1.f),
    C_Vertex3f(0.351264238f, 0.375058174f, -1.f),
    C_Vertex3f(0.364044666f, 0.291884422f, -1.f),
    C_Vertex3f(0.34216547f, 0.37389946f, -1.f),
    C_Vertex3f(0.340634823f, 0.294560432f, -1.f),
    C_Vertex3f(0.332070827f, 0.372972488f, -1.f),
    C_Vertex3f(0.32098031f, 0.37227726f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.313650131f, 0.295898438f, -1.f),
    C_Vertex3f(0.340634823f, 0.294560432f, -1.f),
    C_Vertex3f(0.32098031f, 0.37227726f, -1.f),
    C_Vertex3f(0.308893919f, 0.371813774f, -1.f),
    C_Vertex3f(0.295811653f, 0.371582031f, -1.f),
    C_Vertex3f(0.150390625f, 0.371582031f, -1.f),
    C_Vertex3f(0.150390625f, 0.295898438f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.548338056f, 0.206214309f, -1.f),
    C_Vertex3f(0.546381354f, 0.219267011f, -1.f),
    C_Vertex3f(0.546939373f, 0.152744293f, -1.f),
    C_Vertex3f(0.549316406f, 0.174455643f, -1.f),
    C_Vertex3f(0.549316406f, 0.192697525f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.534641147f, 0.255640626f, -1.f),
    C_Vertex3f(0.535054207f, 0.111696243f, -1.f),
    C_Vertex3f(0.5395329f, 0.243980169f, -1.f),
    C_Vertex3f(0.542185307f, 0.131824493f, -1.f),
    C_Vertex3f(0.543446302f, 0.231855631f, -1.f),
    C_Vertex3f(0.546939373f, 0.152744293f, -1.f),
    C_Vertex3f(0.546381354f, 0.219267011f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.514095783f, 0.287837505f, -1.f),
    C_Vertex3f(0.514918327f, 0.0752353668f, -1.f),
    C_Vertex3f(0.521922588f, 0.277569294f, -1.f),
    C_Vertex3f(0.525546074f, 0.0923595428f, -1.f),
    C_Vertex3f(0.528771043f, 0.266837001f, -1.f),
    C_Vertex3f(0.535054207f, 0.111696243f, -1.f),
    C_Vertex3f(0.534641147f, 0.255640626f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.474371195f, 0.321401358f, -1.f),
    C_Vertex3f(0.476317406f, 0.0371379852f, -1.f),
    C_Vertex3f(0.485564947f, 0.314136744f, -1.f),
    C_Vertex3f(0.490303993f, 0.047624588f, -1.f),
    C_Vertex3f(0.495916963f, 0.30612123f, -1.f),
    C_Vertex3f(0.503170967f, 0.0603237152f, -1.f),
    C_Vertex3f(0.505427241f, 0.297354817f, -1.f),
    C_Vertex3f(0.514918327f, 0.0752353668f, -1.f),
    C_Vertex3f(0.514095783f, 0.287837505f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.457156181f, 0.150261879f, -1.f),
    C_Vertex3f(0.4523983f, 0.137672424f, -1.f),
    C_Vertex3f(0.4605546f, 0.0279350281f, -1.f),
    C_Vertex3f(0.446054459f, 0.125984192f, -1.f),
    C_Vertex3f(0.443015575f, 0.0200157166f, -1.f),
    C_Vertex3f(0.438835144f, 0.115688801f, -1.f),
    C_Vertex3f(0.430740356f, 0.106786251f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.423700333f, 0.0133800507f, -1.f),
    C_Vertex3f(0.443015575f, 0.0200157166f, -1.f),
    C_Vertex3f(0.430740356f, 0.106786251f, -1.f),
    C_Vertex3f(0.421770096f, 0.0992765427f, -1.f),
    C_Vertex3f(0.411924362f, 0.0931596756f, -1.f),
    C_Vertex3f(0.402608871f, 0.0080280304f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.411924362f, 0.0931596756f, -1.f),
    C_Vertex3f(0.400876045f, 0.087916851f, -1.f),
    C_Vertex3f(0.402608871f, 0.0080280304f, -1.f),
    C_Vertex3f(0.388625145f, 0.083548069f, -1.f),
    C_Vertex3f(0.379140854f, 0.0040140152f, -1.f),
    C_Vertex3f(0.375171661f, 0.0800533295f, -1.f),
    C_Vertex3f(0.363687515f, 0.0779781342f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35329628f, 0.00133800507f, -1.f),
    C_Vertex3f(0.379140854f, 0.0040140152f, -1.f),
    C_Vertex3f(0.363687515f, 0.0779781342f, -1.f),
    C_Vertex3f(0.352312088f, 0.076830864f, -1.f),
    C_Vertex3f(0.337873459f, 0.0760660172f, -1.f),
    C_Vertex3f(0.32507515f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Vertex3f(0.0654296875f, 0.f, -1.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, -1.f),
    C_Vertex3f(0.32507515f, 0.f, -1.f),
    C_Vertex3f(0.320371628f, 0.0756835938f, -1.f),
    C_Vertex3f(0.337873459f, 0.0760660172f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.518066406f, 0.466466904f, -1.f),
    C_Vertex3f(0.518066406f, 0.487773895f, -1.f),
    C_Vertex3f(0.51527977f, 0.446526051f, -1.f),
    C_Vertex3f(0.514925957f, 0.509380817f, -1.f),
    C_Vertex3f(0.509706497f, 0.427186489f, -1.f),
    C_Vertex3f(0.508645058f, 0.530522823f, -1.f),
    C_Vertex3f(0.501346588f, 0.408448219f, -1.f),
    C_Vertex3f(0.499223709f, 0.551199913f, -1.f),
    C_Vertex3f(0.490200043f, 0.390311241f, -1.f),
    C_Vertex3f(0.486661911f, 0.571412086f, -1.f),
    C_Vertex3f(0.476212502f, 0.373922825f, -1.f),
    C_Vertex3f(0.47169733f, 0.589002132f, -1.f),
    C_Vertex3f(0.459383965f, 0.35928297f, -1.f),
    C_Vertex3f(0.454329967f, 0.603970051f, -1.f),
    C_Vertex3f(0.439714432f, 0.346391678f, -1.f),
    C_Vertex3f(0.434570313f, 0.474700928f, -1.f),
    C_Vertex3f(0.434570313f, 0.458332062f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434559822f, 0.616315842f, -1.f),
    C_Vertex3f(0.417930126f, 0.623608589f, -1.f),
    C_Vertex3f(0.423111916f, 0.516763687f, -1.f),
    C_Vertex3f(0.428841114f, 0.503697395f, -1.f),
    C_Vertex3f(0.43266058f, 0.489676476f, -1.f),
    C_Vertex3f(0.434570313f, 0.474700928f, -1.f),
    C_Vertex3f(0.454329967f, 0.603970051f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.354100704f, 0.560746431f, -1.f),
    C_Vertex3f(0.363731861f, 0.559068441f, -1.f),
    C_Vertex3f(0.364506483f, 0.636978626f, -1.f),
    C_Vertex3f(0.372094154f, 0.557110786f, -1.f),
    C_Vertex3f(0.379224062f, 0.63454771f, -1.f),
    C_Vertex3f(0.382734299f, 0.553754807f, -1.f),
    C_Vertex3f(0.393033862f, 0.631509066f, -1.f),
    C_Vertex3f(0.395284176f, 0.547370911f, -1.f),
    C_Vertex3f(0.405935884f, 0.627862692f, -1.f),
    C_Vertex3f(0.406197071f, 0.539077759f, -1.f),
    C_Vertex3f(0.417930126f, 0.623608589f, -1.f),
    C_Vertex3f(0.415472984f, 0.528875351f, -1.f),
    C_Vertex3f(0.423111916f, 0.516763687f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.317594051f, 0.564102411f, -1.f),
    C_Vertex3f(0.331031799f, 0.563263416f, -1.f),
    C_Vertex3f(0.332347989f, 0.640017271f, -1.f),
    C_Vertex3f(0.343200684f, 0.562144756f, -1.f),
    C_Vertex3f(0.348881125f, 0.638801813f, -1.f),
    C_Vertex3f(0.354100704f, 0.560746431f, -1.f),
    C_Vertex3f(0.364506483f, 0.636978626f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.314907074f, 0.640625f, -1.f),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Vertex3f(0.150390625f, 0.564941406f, -1.f),
    C_Vertex3f(0.286911964f, 0.564941406f, -1.f),
    C_Vertex3f(0.30288744f, 0.564661741f, -1.f),
    C_Vertex3f(0.317594051f, 0.564102411f, -1.f),
    C_Vertex3f(0.332347989f, 0.640017271f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150390625f, 0.564941406f, -1.f),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, -1.f),
    C_Vertex3f(0.150390625f, 0.295898438f, -1.f),
    C_Vertex3f(0.150390625f, 0.371582031f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.432524204f, 0.44294405f, -1.f),
    C_Vertex3f(0.439714432f, 0.346391678f, -1.f),
    C_Vertex3f(0.434570313f, 0.458332062f, -1.f),
    C_Vertex3f(0.462335706f, 0.327915072f, -1.f),
    C_Vertex3f(0.476317406f, 0.0371379852f, -1.f),
    C_Vertex3f(0.474371195f, 0.321401358f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_67[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.244256333f, 0.969710708f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, -1.f),
    C_Vertex3f(0.610839844f, 0.203514099f, -1.f),
    C_Vertex3f(0.610839844f, 0.203514099f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.96897471f, -0.247160003f, 0.f),
    C_Vertex3f(0.610839844f, 0.203514099f, 0.f),
    C_Vertex3f(0.610839844f, 0.203514099f, -1.f),
    C_Normal3f(0.957571566f, -0.28819555f, 0.f),
    C_Vertex3f(0.607507706f, 0.190450668f, 0.f),
    C_Vertex3f(0.607507706f, 0.190450668f, -1.f),
    C_Normal3f(0.939702749f, -0.341992259f, 0.f),
    C_Vertex3f(0.599511743f, 0.165874839f, 0.f),
    C_Vertex3f(0.599511743f, 0.165874839f, -1.f),
    C_Normal3f(0.912207901f, -0.409727544f, 0.f),
    C_Vertex3f(0.590184093f, 0.142850041f, 0.f),
    C_Vertex3f(0.590184093f, 0.142850041f, -1.f),
    C_Normal3f(0.877267182f, -0.480002373f, 0.f),
    C_Vertex3f(0.579524755f, 0.121376276f, 0.f),
    C_Vertex3f(0.579524755f, 0.121376276f, -1.f),
    C_Normal3f(0.834214509f, -0.55144006f, 0.f),
    C_Vertex3f(0.567533731f, 0.101453543f, 0.f),
    C_Vertex3f(0.567533731f, 0.101453543f, -1.f),
    C_Normal3f(0.78278327f, -0.622294366f, 0.f),
    C_Vertex3f(0.55421102f, 0.0830818415f, 0.f),
    C_Vertex3f(0.55421102f, 0.0830818415f, -1.f),
    C_Normal3f(0.72325623f, -0.690579891f, 0.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, 0.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, -1.f),
    C_Normal3f(0.656544447f, -0.754287302f, 0.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, 0.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, -1.f),
    C_Normal3f(0.585676551f, -0.810544848f, 0.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, 0.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, -1.f),
    C_Normal3f(0.510755539f, -0.859726012f, 0.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, 0.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, -1.f),
    C_Normal3f(0.431376964f, -0.902171791f, 0.f),
    C_Vertex3f(0.468728662f, 0.014980197f, 0.f),
    C_Vertex3f(0.468728662f, 0.014980197f, -1.f),
    C_Normal3f(0.349487215f, -0.936941028f, 0.f),
    C_Vertex3f(0.448522329f, 0.0064060688f, 0.f),
    C_Vertex3f(0.448522329f, 0.0064060688f, -1.f),
    C_Normal3f(0.267140955f, -0.963657379f, 0.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, 0.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, -1.f),
    C_Normal3f(0.186262488f, -0.982499957f, 0.f),
    C_Vertex3f(0.405221105f, -0.00559771061f, 0.f),
    C_Vertex3f(0.405221105f, -0.00559771061f, -1.f),
    C_Normal3f(0.108459026f, -0.994100928f, 0.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, 0.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, -1.f),
    C_Normal3f(0.034610942f, -0.999400854f, 0.f),
    C_Vertex3f(0.358068466f, -0.0107421875f, 0.f),
    C_Vertex3f(0.358068466f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0262763742f, -0.99965477f, 0.f),
    C_Vertex3f(0.33261013f, -0.0107421875f, 0.f),
    C_Vertex3f(0.33261013f, -0.0107421875f, -1.f),
    C_Normal3f(-0.08182282f, -0.996646881f, 0.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, 0.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, -1.f),
    C_Normal3f(-0.142589718f, -0.989781857f, 0.f),
    C_Vertex3f(0.28444469f, -0.00678789616f, 0.f),
    C_Vertex3f(0.28444469f, -0.00678789616f, -1.f),
    C_Normal3f(-0.208271027f, -0.978071153f, 0.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, 0.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, -1.f),
    C_Normal3f(-0.27848947f, -0.960439324f, 0.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, 0.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, -1.f),
    C_Normal3f(-0.352481455f, -0.935818851f, 0.f),
    C_Vertex3f(0.221263289f, 0.00902926922f, 0.f),
    C_Vertex3f(0.221263289f, 0.00902926922f, -1.f),
    C_Normal3f(-0.429035544f, -0.90328753f, 0.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, 0.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, -1.f),
    C_Normal3f(-0.506492198f, -0.862244546f, 0.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, 0.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, -1.f),
    C_Normal3f(-0.577542961f, -0.816360354f, 0.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, 0.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, -1.f),
    C_Normal3f(-0.64043498f, -0.768012404f, 0.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, 0.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, -1.f),
    C_Normal3f(-0.699227035f, -0.714899778f, 0.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, 0.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, -1.f),
    C_Normal3f(-0.752920151f, -0.65811193f, 0.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, 0.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, -1.f),
    C_Normal3f(-0.800829113f, -0.598892927f, 0.f),
    C_Vertex3f(0.113759041f, 0.0907020569f, 0.f),
    C_Vertex3f(0.113759041f, 0.0907020569f, -1.f),
    C_Normal3f(-0.842613876f, -0.53851819f, 0.f),
    C_Vertex3f(0.102382421f, 0.107153654f, 0.f),
    C_Vertex3f(0.102382421f, 0.107153654f, -1.f),
    C_Normal3f(-0.878260493f, -0.478182614f, 0.f),
    C_Vertex3f(0.0919754505f, 0.124786615f, 0.f),
    C_Vertex3f(0.0919754505f, 0.124786615f, -1.f),
    C_Normal3f(-0.90802449f, -0.41891703f, 0.f),
    C_Vertex3f(0.0825381279f, 0.143600941f, 0.f),
    C_Vertex3f(0.0825381279f, 0.143600941f, -1.f),
    C_Normal3f(-0.930585563f, -0.366074562f, 0.f),
    C_Vertex3f(0.0740704536f, 0.16359663f, 0.f),
    C_Vertex3f(0.0740704536f, 0.16359663f, -1.f),
    C_Normal3f(-0.948261619f, -0.317489982f, 0.f),
    C_Vertex3f(0.0666612387f, 0.183961034f, 0.f),
    C_Vertex3f(0.0666612387f, 0.183961034f, -1.f),
    C_Normal3f(-0.963385701f, -0.268119097f, 0.f),
    C_Vertex3f(0.060310483f, 0.204694152f, 0.f),
    C_Vertex3f(0.060310483f, 0.204694152f, -1.f),
    C_Normal3f(-0.975867569f, -0.218363285f, 0.f),
    C_Vertex3f(0.0550181866f, 0.225795984f, 0.f),
    C_Vertex3f(0.0550181866f, 0.225795984f, -1.f),
    C_Normal3f(-0.985680282f, -0.168625012f, 0.f),
    C_Vertex3f(0.0507843494f, 0.247266531f, 0.f),
    C_Vertex3f(0.0507843494f, 0.247266531f, -1.f),
    C_Normal3f(-0.992859125f, -0.119292498f, 0.f),
    C_Vertex3f(0.0476089716f, 0.269105792f, 0.f),
    C_Vertex3f(0.0476089716f, 0.269105792f, -1.f),
    C_Normal3f(-0.99749589f, -0.0707256868f, 0.f),
    C_Vertex3f(0.045492053f, 0.291313767f, 0.f),
    C_Vertex3f(0.045492053f, 0.291313767f, -1.f),
    C_Normal3f(-0.999741733f, -0.0227270462f, 0.f),
    C_Vertex3f(0.0444335938f, 0.313890457f, 0.f),
    C_Vertex3f(0.0444335938f, 0.313890457f, -1.f),
    C_Normal3f(-0.999692678f, 0.0247900393f, 0.f),
    C_Vertex3f(0.0444335938f, 0.337874413f, 0.f),
    C_Vertex3f(0.0444335938f, 0.337874413f, -1.f),
    C_Normal3f(-0.997179449f, 0.0750543475f, 0.f),
    C_Vertex3f(0.0456290245f, 0.362097859f, 0.f),
    C_Vertex3f(0.0456290245f, 0.362097859f, -1.f),
    C_Normal3f(-0.991681218f, 0.128718078f, 0.f),
    C_Vertex3f(0.048019886f, 0.385522246f, 0.f),
    C_Vertex3f(0.048019886f, 0.385522246f, -1.f),
    C_Normal3f(-0.982738376f, 0.185000554f, 0.f),
    C_Vertex3f(0.0516061783f, 0.408147573f, 0.f),
    C_Vertex3f(0.0516061783f, 0.408147573f, -1.f),
    C_Normal3f(-0.969899654f, 0.243504897f, 0.f),
    C_Vertex3f(0.0563879013f, 0.429973841f, 0.f),
    C_Vertex3f(0.0563879013f, 0.429973841f, -1.f),
    C_Normal3f(-0.952771962f, 0.303686678f, 0.f),
    C_Vertex3f(0.0623650551f, 0.451001048f, 0.f),
    C_Vertex3f(0.0623650551f, 0.451001048f, -1.f),
    C_Normal3f(-0.931062996f, 0.364858449f, 0.f),
    C_Vertex3f(0.0695376396f, 0.471229196f, 0.f),
    C_Vertex3f(0.0695376396f, 0.471229196f, -1.f),
    C_Normal3f(-0.904623687f, 0.426211268f, 0.f),
    C_Vertex3f(0.0779056549f, 0.490658283f, 0.f),
    C_Vertex3f(0.0779056549f, 0.490658283f, -1.f),
    C_Normal3f(-0.873540044f, 0.486752242f, 0.f),
    C_Vertex3f(0.087469101f, 0.509288311f, 0.f),
    C_Vertex3f(0.087469101f, 0.509288311f, -1.f),
    C_Normal3f(-0.837032259f, 0.547153473f, 0.f),
    C_Vertex3f(0.098043561f, 0.526798368f, 0.f),
    C_Vertex3f(0.098043561f, 0.526798368f, -1.f),
    C_Normal3f(-0.794706285f, 0.606994212f, 0.f),
    C_Vertex3f(0.109629035f, 0.543188453f, 0.f),
    C_Vertex3f(0.109629035f, 0.543188453f, -1.f),
    C_Normal3f(-0.746745348f, 0.665110052f, 0.f),
    C_Vertex3f(0.122225523f, 0.558458567f, 0.f),
    C_Vertex3f(0.122225523f, 0.558458567f, -1.f),
    C_Normal3f(-0.693637788f, 0.72032398f, 0.f),
    C_Vertex3f(0.135833025f, 0.572608709f, 0.f),
    C_Vertex3f(0.135833025f, 0.572608709f, -1.f),
    C_Normal3f(-0.636166453f, 0.771551907f, 0.f),
    C_Vertex3f(0.150451541f, 0.585638881f, 0.f),
    C_Vertex3f(0.150451541f, 0.585638881f, -1.f),
    C_Normal3f(-0.575355411f, 0.817903519f, 0.f),
    C_Vertex3f(0.166081071f, 0.597549081f, 0.f),
    C_Vertex3f(0.166081071f, 0.597549081f, -1.f),
    C_Normal3f(-0.512380719f, 0.85875833f, 0.f),
    C_Vertex3f(0.182721615f, 0.60833931f, 0.f),
    C_Vertex3f(0.182721615f, 0.60833931f, -1.f),
    C_Normal3f(-0.452358484f, 0.891836226f, 0.f),
    C_Vertex3f(0.200373173f, 0.618009567f, 0.f),
    C_Vertex3f(0.200373173f, 0.618009567f, -1.f),
    C_Normal3f(-0.394201905f, 0.919023931f, 0.f),
    C_Vertex3f(0.218468785f, 0.626471043f, 0.f),
    C_Vertex3f(0.218468785f, 0.626471043f, -1.f),
    C_Normal3f(-0.334004939f, 0.942571342f, 0.f),
    C_Vertex3f(0.237008452f, 0.633723736f, 0.f),
    C_Vertex3f(0.237008452f, 0.633723736f, -1.f),
    C_Normal3f(-0.272504777f, 0.962154448f, 0.f),
    C_Vertex3f(0.255992174f, 0.639767647f, 0.f),
    C_Vertex3f(0.255992174f, 0.639767647f, -1.f),
    C_Normal3f(-0.210483432f, 0.977597356f, 0.f),
    C_Vertex3f(0.27541995f, 0.644602776f, 0.f),
    C_Vertex3f(0.27541995f, 0.644602776f, -1.f),
    C_Normal3f(-0.148719549f, 0.988879383f, 0.f),
    C_Vertex3f(0.295291781f, 0.648229122f, 0.f),
    C_Vertex3f(0.295291781f, 0.648229122f, -1.f),
    C_Normal3f(-0.0879421532f, 0.996125579f, 0.f),
    C_Vertex3f(0.315607667f, 0.650646687f, 0.f),
    C_Vertex3f(0.315607667f, 0.650646687f, -1.f),
    C_Normal3f(-0.027851617f, 0.999612153f, 0.f),
    C_Vertex3f(0.336367607f, 0.651855469f, 0.f),
    C_Vertex3f(0.336367607f, 0.651855469f, -1.f),
    C_Normal3f(0.0334659219f, 0.999439895f, 0.f),
    C_Vertex3f(0.358991623f, 0.651855469f, 0.f),
    C_Vertex3f(0.358991623f, 0.651855469f, -1.f),
    C_Normal3f(0.101028174f, 0.994883597f, 0.f),
    C_Vertex3f(0.382079482f, 0.650324821f, 0.f),
    C_Vertex3f(0.382079482f, 0.650324821f, -1.f),
    C_Normal3f(0.174023315f, 0.984741509f, 0.f),
    C_Vertex3f(0.404211164f, 0.647263527f, 0.f),
    C_Vertex3f(0.404211164f, 0.647263527f, -1.f),
    C_Normal3f(0.250579476f, 0.968096018f, 0.f),
    C_Vertex3f(0.425386667f, 0.642671585f, 0.f),
    C_Vertex3f(0.425386667f, 0.642671585f, -1.f),
    C_Normal3f(0.329433322f, 0.94417882f, 0.f),
    C_Vertex3f(0.445605993f, 0.636548996f, 0.f),
    C_Vertex3f(0.445605993f, 0.636548996f, -1.f),
    C_Normal3f(0.408961713f, 0.912551582f, 0.f),
    C_Vertex3f(0.464869142f, 0.62889576f, 0.f),
    C_Vertex3f(0.464869142f, 0.62889576f, -1.f),
    C_Normal3f(0.487299085f, 0.873235166f, 0.f),
    C_Vertex3f(0.483176112f, 0.619711876f, 0.f),
    C_Vertex3f(0.483176112f, 0.619711876f, -1.f),
    C_Normal3f(0.562525034f, 0.826780319f, 0.f),
    C_Vertex3f(0.500526905f, 0.608997345f, 0.f),
    C_Vertex3f(0.500526905f, 0.608997345f, -1.f),
    C_Normal3f(0.632112801f, 0.774876416f, 0.f),
    C_Vertex3f(0.51692152f, 0.596752167f, 0.f),
    C_Vertex3f(0.51692152f, 0.596752167f, -1.f),
    C_Normal3f(0.696682513f, 0.717379689f, 0.f),
    C_Vertex3f(0.532072902f, 0.583263397f, 0.f),
    C_Vertex3f(0.532072902f, 0.583263397f, -1.f),
    C_Normal3f(0.756186187f, 0.65435642f, 0.f),
    C_Vertex3f(0.54598105f, 0.568531036f, 0.f),
    C_Vertex3f(0.54598105f, 0.568531036f, -1.f),
    C_Normal3f(0.809380293f, 0.587284982f, 0.f),
    C_Vertex3f(0.558645964f, 0.552555084f, 0.f),
    C_Vertex3f(0.558645964f, 0.552555084f, -1.f),
    C_Normal3f(0.855471253f, 0.517850339f, 0.f),
    C_Vertex3f(0.570067644f, 0.535335541f, 0.f),
    C_Vertex3f(0.570067644f, 0.535335541f, -1.f),
    C_Normal3f(0.894157887f, 0.447751939f, 0.f),
    C_Vertex3f(0.580246091f, 0.516872406f, 0.f),
    C_Vertex3f(0.580246091f, 0.516872406f, -1.f),
    C_Normal3f(0.925588667f, 0.378530741f, 0.f),
    C_Vertex3f(0.589181304f, 0.49716568f, 0.f),
    C_Vertex3f(0.589181304f, 0.49716568f, -1.f),
    C_Normal3f(0.946590006f, 0.322439611f, 0.f),
    C_Vertex3f(0.596873283f, 0.476215363f, 0.f),
    C_Vertex3f(0.596873283f, 0.476215363f, -1.f),
    C_Normal3f(0.960284233f, 0.279023796f, 0.f),
    C_Vertex3f(0.600097656f, 0.465118408f, 0.f),
    C_Vertex3f(0.600097656f, 0.465118408f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.229118645f, -0.973398507f, 0.f),
    C_Vertex3f(0.600097656f, 0.465118408f, 0.f),
    C_Vertex3f(0.600097656f, 0.465118408f, -1.f),
    C_Vertex3f(0.516601563f, 0.445465088f, -1.f),
    C_Vertex3f(0.516601563f, 0.445465088f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.952661753f, -0.304032117f, 0.f),
    C_Vertex3f(0.516601563f, 0.445465088f, 0.f),
    C_Vertex3f(0.516601563f, 0.445465088f, -1.f),
    C_Normal3f(-0.941064239f, -0.338227957f, 0.f),
    C_Vertex3f(0.513813972f, 0.454199791f, 0.f),
    C_Vertex3f(0.513813972f, 0.454199791f, -1.f),
    C_Normal3f(-0.923471034f, -0.383668244f, 0.f),
    C_Vertex3f(0.507610321f, 0.470481753f, 0.f),
    C_Vertex3f(0.507610321f, 0.470481753f, -1.f),
    C_Normal3f(-0.896986663f, -0.442057669f, 0.f),
    C_Vertex3f(0.500778198f, 0.485576272f, 0.f),
    C_Vertex3f(0.500778198f, 0.485576272f, -1.f),
    C_Normal3f(-0.86353302f, -0.504292369f, 0.f),
    C_Vertex3f(0.493317604f, 0.499483347f, 0.f),
    C_Vertex3f(0.493317604f, 0.499483347f, -1.f),
    C_Normal3f(-0.821924388f, -0.569596708f, 0.f),
    C_Vertex3f(0.485228539f, 0.512202978f, 0.f),
    C_Vertex3f(0.485228539f, 0.512202978f, -1.f),
    C_Normal3f(-0.771112323f, -0.63669908f, 0.f),
    C_Vertex3f(0.476511002f, 0.523735166f, 0.f),
    C_Vertex3f(0.476511002f, 0.523735166f, -1.f),
    C_Normal3f(-0.710405827f, -0.703792214f, 0.f),
    C_Vertex3f(0.467164993f, 0.534079909f, 0.f),
    C_Vertex3f(0.467164993f, 0.534079909f, -1.f),
    C_Normal3f(-0.63972795f, -0.768601477f, 0.f),
    C_Vertex3f(0.457190514f, 0.543237209f, 0.f),
    C_Vertex3f(0.457190514f, 0.543237209f, -1.f),
    C_Normal3f(-0.563830793f, -0.825890362f, 0.f),
    C_Vertex3f(0.446587563f, 0.551207066f, 0.f),
    C_Vertex3f(0.446587563f, 0.551207066f, -1.f),
    C_Normal3f(-0.486547291f, -0.873654306f, 0.f),
    C_Vertex3f(0.435301542f, 0.55818069f, 0.f),
    C_Vertex3f(0.435301542f, 0.55818069f, -1.f),
    C_Normal3f(-0.406627208f, -0.913594127f, 0.f),
    C_Vertex3f(0.423332453f, 0.564158082f, 0.f),
    C_Vertex3f(0.423332453f, 0.564158082f, -1.f),
    C_Normal3f(-0.326150656f, -0.945317864f, 0.f),
    C_Vertex3f(0.410680294f, 0.569139242f, 0.f),
    C_Vertex3f(0.410680294f, 0.569139242f, -1.f),
    C_Normal3f(-0.247042209f, -0.96900475f, 0.f),
    C_Vertex3f(0.397345066f, 0.57312417f, 0.f),
    C_Vertex3f(0.397345066f, 0.57312417f, -1.f),
    C_Normal3f(-0.170889482f, -0.98529017f, 0.f),
    C_Vertex3f(0.383326769f, 0.576112866f, 0.f),
    C_Vertex3f(0.383326769f, 0.576112866f, -1.f),
    C_Normal3f(-0.0988525376f, -0.995102167f, 0.f),
    C_Vertex3f(0.368625402f, 0.57810533f, 0.f),
    C_Vertex3f(0.368625402f, 0.57810533f, -1.f),
    C_Normal3f(-0.0304940492f, -0.999534905f, 0.f),
    C_Vertex3f(0.353240967f, 0.579101563f, 0.f),
    C_Vertex3f(0.353240967f, 0.579101563f, -1.f),
    C_Normal3f(0.0315818004f, -0.999501169f, 0.f),
    C_Vertex3f(0.335970879f, 0.579101563f, 0.f),
    C_Vertex3f(0.335970879f, 0.579101563f, -1.f),
    C_Normal3f(0.0952159464f, -0.995456576f, 0.f),
    C_Vertex3f(0.318256378f, 0.577996135f, 0.f),
    C_Vertex3f(0.318256378f, 0.577996135f, -1.f),
    C_Normal3f(0.164439261f, -0.986387312f, 0.f),
    C_Vertex3f(0.301300049f, 0.575785279f, 0.f),
    C_Vertex3f(0.301300049f, 0.575785279f, -1.f),
    C_Normal3f(0.237575501f, -0.971369028f, 0.f),
    C_Vertex3f(0.285101891f, 0.572468996f, 0.f),
    C_Vertex3f(0.285101891f, 0.572468996f, -1.f),
    C_Normal3f(0.313623488f, -0.94954741f, 0.f),
    C_Vertex3f(0.269661903f, 0.568047285f, 0.f),
    C_Vertex3f(0.269661903f, 0.568047285f, -1.f),
    C_Normal3f(0.391205609f, -0.920303345f, 0.f),
    C_Vertex3f(0.254980087f, 0.562520146f, 0.f),
    C_Vertex3f(0.254980087f, 0.562520146f, -1.f),
    C_Normal3f(0.468632936f, -0.883393049f, 0.f),
    C_Vertex3f(0.241056442f, 0.55588758f, 0.f),
    C_Vertex3f(0.241056442f, 0.55588758f, -1.f),
    C_Normal3f(0.544044912f, -0.839056134f, 0.f),
    C_Vertex3f(0.227890968f, 0.548149586f, 0.f),
    C_Vertex3f(0.227890968f, 0.548149586f, -1.f),
    C_Normal3f(0.61108011f, -0.791568756f, 0.f),
    C_Vertex3f(0.215483665f, 0.539306164f, 0.f),
    C_Vertex3f(0.215483665f, 0.539306164f, -1.f),
    C_Normal3f(0.671135724f, -0.741334558f, 0.f),
    C_Vertex3f(0.203998804f, 0.529705167f, 0.f),
    C_Vertex3f(0.203998804f, 0.529705167f, -1.f),
    C_Normal3f(0.728354037f, -0.685200989f, 0.f),
    C_Vertex3f(0.193436384f, 0.519346595f, 0.f),
    C_Vertex3f(0.193436384f, 0.519346595f, -1.f),
    C_Normal3f(0.781438589f, -0.623982251f, 0.f),
    C_Vertex3f(0.183796406f, 0.508230448f, 0.f),
    C_Vertex3f(0.183796406f, 0.508230448f, -1.f),
    C_Normal3f(0.829291821f, -0.558815777f, 0.f),
    C_Vertex3f(0.175078869f, 0.496356726f, 0.f),
    C_Vertex3f(0.175078869f, 0.496356726f, -1.f),
    C_Normal3f(0.871125877f, -0.491059691f, 0.f),
    C_Vertex3f(0.167283773f, 0.483725429f, 0.f),
    C_Vertex3f(0.167283773f, 0.483725429f, -1.f),
    C_Normal3f(0.90652287f, -0.422156781f, 0.f),
    C_Vertex3f(0.160411119f, 0.470336556f, 0.f),
    C_Vertex3f(0.160411119f, 0.470336556f, -1.f),
    C_Normal3f(0.937979281f, -0.34669134f, 0.f),
    C_Vertex3f(0.154460907f, 0.456190109f, 0.f),
    C_Vertex3f(0.154460907f, 0.456190109f, -1.f),
    C_Normal3f(0.961297512f, -0.275512487f, 0.f),
    C_Vertex3f(0.14691925f, 0.433834076f, 0.f),
    C_Vertex3f(0.14691925f, 0.433834076f, -1.f),
    C_Normal3f(0.979427636f, -0.201795474f, 0.f),
    C_Vertex3f(0.139377594f, 0.403562546f, 0.f),
    C_Vertex3f(0.139377594f, 0.403562546f, -1.f),
    C_Normal3f(0.992667437f, -0.120877311f, 0.f),
    C_Vertex3f(0.134349823f, 0.37282753f, 0.f),
    C_Vertex3f(0.134349823f, 0.37282753f, -1.f),
    C_Normal3f(0.999036968f, -0.0438771285f, 0.f),
    C_Vertex3f(0.131835938f, 0.341629028f, 0.f),
    C_Vertex3f(0.131835938f, 0.341629028f, -1.f),
    C_Normal3f(0.999868095f, 0.0162386186f, 0.f),
    C_Vertex3f(0.131835938f, 0.315588951f, 0.f),
    C_Vertex3f(0.131835938f, 0.315588951f, -1.f),
    C_Normal3f(0.998357296f, 0.0572959892f, 0.f),
    C_Vertex3f(0.132579923f, 0.295819283f, 0.f),
    C_Vertex3f(0.132579923f, 0.295819283f, -1.f),
    C_Normal3f(0.995140135f, 0.0984695256f, 0.f),
    C_Vertex3f(0.134067893f, 0.276698112f, 0.f),
    C_Vertex3f(0.134067893f, 0.276698112f, -1.f),
    C_Normal3f(0.989862859f, 0.142026201f, 0.f),
    C_Vertex3f(0.136299849f, 0.258225441f, 0.f),
    C_Vertex3f(0.136299849f, 0.258225441f, -1.f),
    C_Normal3f(0.98218751f, 0.187903568f, 0.f),
    C_Vertex3f(0.139275789f, 0.240401268f, 0.f),
    C_Vertex3f(0.139275789f, 0.240401268f, -1.f),
    C_Normal3f(0.971761227f, 0.23596625f, 0.f),
    C_Vertex3f(0.142995715f, 0.223225594f, 0.f),
    C_Vertex3f(0.142995715f, 0.223225594f, -1.f),
    C_Normal3f(0.958231807f, 0.285992563f, 0.f),
    C_Vertex3f(0.147459626f, 0.206698418f, 0.f),
    C_Vertex3f(0.147459626f, 0.206698418f, -1.f),
    C_Normal3f(0.941266835f, 0.337663591f, 0.f),
    C_Vertex3f(0.152667522f, 0.19081974f, 0.f),
    C_Vertex3f(0.152667522f, 0.19081974f, -1.f),
    C_Normal3f(0.918154716f, 0.396222055f, 0.f),
    C_Vertex3f(0.158619404f, 0.175589561f, 0.f),
    C_Vertex3f(0.158619404f, 0.175589561f, -1.f),
    C_Normal3f(0.886648476f, 0.462444037f, 0.f),
    C_Vertex3f(0.165397286f, 0.161321402f, 0.f),
    C_Vertex3f(0.165397286f, 0.161321402f, -1.f),
    C_Normal3f(0.847963274f, 0.530054867f, 0.f),
    C_Vertex3f(0.17300117f, 0.148015261f, 0.f),
    C_Vertex3f(0.17300117f, 0.148015261f, -1.f),
    C_Normal3f(0.801762938f, 0.597642303f, 0.f),
    C_Vertex3f(0.181431055f, 0.135671139f, 0.f),
    C_Vertex3f(0.181431055f, 0.135671139f, -1.f),
    C_Normal3f(0.748123646f, 0.663559377f, 0.f),
    C_Vertex3f(0.190686941f, 0.124289036f, 0.f),
    C_Vertex3f(0.190686941f, 0.124289036f, -1.f),
    C_Normal3f(0.687618852f, 0.726071894f, 0.f),
    C_Vertex3f(0.200768828f, 0.113868952f, 0.f),
    C_Vertex3f(0.200768828f, 0.113868952f, -1.f),
    C_Normal3f(0.621327579f, 0.783550918f, 0.f),
    C_Vertex3f(0.211676717f, 0.104410887f, 0.f),
    C_Vertex3f(0.211676717f, 0.104410887f, -1.f),
    C_Normal3f(0.550750792f, 0.834669828f, 0.f),
    C_Vertex3f(0.223410606f, 0.0959148407f, 0.f),
    C_Vertex3f(0.223410606f, 0.0959148407f, -1.f),
    C_Normal3f(0.486346662f, 0.873765886f, 0.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, 0.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, -1.f),
    C_Normal3f(0.427382827f, 0.904070735f, 0.f),
    C_Vertex3f(0.248789787f, 0.0817885399f, 0.f),
    C_Vertex3f(0.248789787f, 0.0817885399f, -1.f),
    C_Normal3f(0.365080386f, 0.930975974f, 0.f),
    C_Vertex3f(0.261868477f, 0.0761380196f, 0.f),
    C_Vertex3f(0.261868477f, 0.0761380196f, -1.f),
    C_Normal3f(0.300158978f, 0.953889191f, 0.f),
    C_Vertex3f(0.275206566f, 0.0714292526f, 0.f),
    C_Vertex3f(0.275206566f, 0.0714292526f, -1.f),
    C_Normal3f(0.233480901f, 0.972361386f, 0.f),
    C_Vertex3f(0.288804054f, 0.0676622391f, 0.f),
    C_Vertex3f(0.288804054f, 0.0676622391f, -1.f),
    C_Normal3f(0.165996134f, 0.986126363f, 0.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, 0.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, -1.f),
    C_Normal3f(0.098675929f, 0.995119691f, 0.f),
    C_Vertex3f(0.316777229f, 0.0629534721f, 0.f),
    C_Vertex3f(0.316777229f, 0.0629534721f, -1.f),
    C_Normal3f(0.0307678469f, 0.99952662f, 0.f),
    C_Vertex3f(0.331152916f, 0.0620117188f, 0.f),
    C_Vertex3f(0.331152916f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0384552181f, 0.999260366f, 0.f),
    C_Vertex3f(0.347371101f, 0.0620117188f, 0.f),
    C_Vertex3f(0.347371101f, 0.0620117188f, -1.f),
    C_Normal3f(-0.113953315f, 0.993486106f, 0.f),
    C_Vertex3f(0.364489913f, 0.0632946491f, 0.f),
    C_Vertex3f(0.364489913f, 0.0632946491f, -1.f),
    C_Normal3f(-0.19543092f, 0.98071748f, 0.f),
    C_Vertex3f(0.380926251f, 0.0658605099f, 0.f),
    C_Vertex3f(0.380926251f, 0.0658605099f, -1.f),
    C_Normal3f(-0.279707551f, 0.960085154f, 0.f),
    C_Vertex3f(0.396680117f, 0.069709301f, 0.f),
    C_Vertex3f(0.396680117f, 0.069709301f, -1.f),
    C_Normal3f(-0.364904284f, 0.931044996f, 0.f),
    C_Vertex3f(0.411751509f, 0.0748410225f, 0.f),
    C_Vertex3f(0.411751509f, 0.0748410225f, -1.f),
    C_Normal3f(-0.44885537f, 0.893604457f, 0.f),
    C_Vertex3f(0.426140428f, 0.0812556744f, 0.f),
    C_Vertex3f(0.426140428f, 0.0812556744f, -1.f),
    C_Normal3f(-0.529350221f, 0.848403394f, 0.f),
    C_Vertex3f(0.439846873f, 0.0889532566f, 0.f),
    C_Vertex3f(0.439846873f, 0.0889532566f, -1.f),
    C_Normal3f(-0.604410052f, 0.796673357f, 0.f),
    C_Vertex3f(0.452870846f, 0.0979337692f, 0.f),
    C_Vertex3f(0.452870846f, 0.0979337692f, -1.f),
    C_Normal3f(-0.676676273f, 0.73628056f, 0.f),
    C_Vertex3f(0.465212345f, 0.108197212f, 0.f),
    C_Vertex3f(0.465212345f, 0.108197212f, -1.f),
    C_Normal3f(-0.745346248f, 0.666677594f, 0.f),
    C_Vertex3f(0.476557136f, 0.119702578f, 0.f),
    C_Vertex3f(0.476557136f, 0.119702578f, -1.f),
    C_Normal3f(-0.80507195f, 0.59317708f, 0.f),
    C_Vertex3f(0.486905217f, 0.132449865f, 0.f),
    C_Vertex3f(0.486905217f, 0.132449865f, -1.f),
    C_Normal3f(-0.855240166f, 0.518231869f, 0.f),
    C_Vertex3f(0.49625659f, 0.146439075f, 0.f),
    C_Vertex3f(0.49625659f, 0.146439075f, -1.f),
    C_Normal3f(-0.895998299f, 0.444057435f, 0.f),
    C_Vertex3f(0.504611254f, 0.161670208f, 0.f),
    C_Vertex3f(0.504611254f, 0.161670208f, -1.f),
    C_Normal3f(-0.928064346f, 0.372419834f, 0.f),
    C_Vertex3f(0.511969209f, 0.178143263f, 0.f),
    C_Vertex3f(0.511969209f, 0.178143263f, -1.f),
    C_Normal3f(-0.952493608f, 0.304558545f, 0.f),
    C_Vertex3f(0.518330455f, 0.19585824f, 0.f),
    C_Vertex3f(0.518330455f, 0.19585824f, -1.f),
    C_Normal3f(-0.967872739f, 0.251440734f, 0.f),
    C_Vertex3f(0.523694992f, 0.21481514f, 0.f),
    C_Vertex3f(0.523694992f, 0.21481514f, -1.f),
    C_Normal3f(-0.977408946f, 0.21135661f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.600097656f, 0.465118408f, 0.f),
    C_Vertex3f(0.596873283f, 0.476215363f, 0.f),
    C_Vertex3f(0.589181304f, 0.49716568f, 0.f),
    C_Vertex3f(0.580246091f, 0.516872406f, 0.f),
    C_Vertex3f(0.570067644f, 0.535335541f, 0.f),
    C_Vertex3f(0.558645964f, 0.552555084f, 0.f),
    C_Vertex3f(0.54598105f, 0.568531036f, 0.f),
    C_Vertex3f(0.532072902f, 0.583263397f, 0.f),
    C_Vertex3f(0.51692152f, 0.596752167f, 0.f),
    C_Vertex3f(0.516601563f, 0.445465088f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.51692152f, 0.596752167f, 0.f),
    C_Vertex3f(0.500526905f, 0.608997345f, 0.f),
    C_Vertex3f(0.500778198f, 0.485576272f, 0.f),
    C_Vertex3f(0.507610321f, 0.470481753f, 0.f),
    C_Vertex3f(0.513813972f, 0.454199791f, 0.f),
    C_Vertex3f(0.516601563f, 0.445465088f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.467164993f, 0.534079909f, 0.f),
    C_Vertex3f(0.476511002f, 0.523735166f, 0.f),
    C_Vertex3f(0.483176112f, 0.619711876f, 0.f),
    C_Vertex3f(0.485228539f, 0.512202978f, 0.f),
    C_Vertex3f(0.500526905f, 0.608997345f, 0.f),
    C_Vertex3f(0.493317604f, 0.499483347f, 0.f),
    C_Vertex3f(0.500778198f, 0.485576272f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.464869142f, 0.62889576f, 0.f),
    C_Vertex3f(0.445605993f, 0.636548996f, 0.f),
    C_Vertex3f(0.446587563f, 0.551207066f, 0.f),
    C_Vertex3f(0.457190514f, 0.543237209f, 0.f),
    C_Vertex3f(0.467164993f, 0.534079909f, 0.f),
    C_Vertex3f(0.483176112f, 0.619711876f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.410680294f, 0.569139242f, 0.f),
    C_Vertex3f(0.423332453f, 0.564158082f, 0.f),
    C_Vertex3f(0.425386667f, 0.642671585f, 0.f),
    C_Vertex3f(0.435301542f, 0.55818069f, 0.f),
    C_Vertex3f(0.445605993f, 0.636548996f, 0.f),
    C_Vertex3f(0.446587563f, 0.551207066f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.404211164f, 0.647263527f, 0.f),
    C_Vertex3f(0.382079482f, 0.650324821f, 0.f),
    C_Vertex3f(0.383326769f, 0.576112866f, 0.f),
    C_Vertex3f(0.397345066f, 0.57312417f, 0.f),
    C_Vertex3f(0.410680294f, 0.569139242f, 0.f),
    C_Vertex3f(0.425386667f, 0.642671585f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318256378f, 0.577996135f, 0.f),
    C_Vertex3f(0.335970879f, 0.579101563f, 0.f),
    C_Vertex3f(0.336367607f, 0.651855469f, 0.f),
    C_Vertex3f(0.353240967f, 0.579101563f, 0.f),
    C_Vertex3f(0.358991623f, 0.651855469f, 0.f),
    C_Vertex3f(0.368625402f, 0.57810533f, 0.f),
    C_Vertex3f(0.382079482f, 0.650324821f, 0.f),
    C_Vertex3f(0.383326769f, 0.576112866f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241056442f, 0.55588758f, 0.f),
    C_Vertex3f(0.254980087f, 0.562520146f, 0.f),
    C_Vertex3f(0.255992174f, 0.639767647f, 0.f),
    C_Vertex3f(0.269661903f, 0.568047285f, 0.f),
    C_Vertex3f(0.27541995f, 0.644602776f, 0.f),
    C_Vertex3f(0.285101891f, 0.572468996f, 0.f),
    C_Vertex3f(0.295291781f, 0.648229122f, 0.f),
    C_Vertex3f(0.301300049f, 0.575785279f, 0.f),
    C_Vertex3f(0.315607667f, 0.650646687f, 0.f),
    C_Vertex3f(0.318256378f, 0.577996135f, 0.f),
    C_Vertex3f(0.336367607f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.203998804f, 0.529705167f, 0.f),
    C_Vertex3f(0.215483665f, 0.539306164f, 0.f),
    C_Vertex3f(0.218468785f, 0.626471043f, 0.f),
    C_Vertex3f(0.227890968f, 0.548149586f, 0.f),
    C_Vertex3f(0.237008452f, 0.633723736f, 0.f),
    C_Vertex3f(0.241056442f, 0.55588758f, 0.f),
    C_Vertex3f(0.255992174f, 0.639767647f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.200373173f, 0.618009567f, 0.f),
    C_Vertex3f(0.182721615f, 0.60833931f, 0.f),
    C_Vertex3f(0.183796406f, 0.508230448f, 0.f),
    C_Vertex3f(0.193436384f, 0.519346595f, 0.f),
    C_Vertex3f(0.203998804f, 0.529705167f, 0.f),
    C_Vertex3f(0.218468785f, 0.626471043f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154460907f, 0.456190109f, 0.f),
    C_Vertex3f(0.160411119f, 0.470336556f, 0.f),
    C_Vertex3f(0.166081071f, 0.597549081f, 0.f),
    C_Vertex3f(0.167283773f, 0.483725429f, 0.f),
    C_Vertex3f(0.182721615f, 0.60833931f, 0.f),
    C_Vertex3f(0.175078869f, 0.496356726f, 0.f),
    C_Vertex3f(0.183796406f, 0.508230448f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150451541f, 0.585638881f, 0.f),
    C_Vertex3f(0.135833025f, 0.572608709f, 0.f),
    C_Vertex3f(0.139377594f, 0.403562546f, 0.f),
    C_Vertex3f(0.14691925f, 0.433834076f, 0.f),
    C_Vertex3f(0.154460907f, 0.456190109f, 0.f),
    C_Vertex3f(0.166081071f, 0.597549081f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.134067893f, 0.276698112f, 0.f),
    C_Vertex3f(0.132579923f, 0.295819283f, 0.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, 0.f),
    C_Vertex3f(0.131835938f, 0.315588951f, 0.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, 0.f),
    C_Vertex3f(0.131835938f, 0.341629028f, 0.f),
    C_Vertex3f(0.135833025f, 0.572608709f, 0.f),
    C_Vertex3f(0.134349823f, 0.37282753f, 0.f),
    C_Vertex3f(0.139377594f, 0.403562546f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0444335938f, 0.337874413f, 0.f),
    C_Vertex3f(0.045492053f, 0.291313767f, 0.f),
    C_Vertex3f(0.0456290245f, 0.362097859f, 0.f),
    C_Vertex3f(0.0476089716f, 0.269105792f, 0.f),
    C_Vertex3f(0.048019886f, 0.385522246f, 0.f),
    C_Vertex3f(0.0507843494f, 0.247266531f, 0.f),
    C_Vertex3f(0.0516061783f, 0.408147573f, 0.f),
    C_Vertex3f(0.0550181866f, 0.225795984f, 0.f),
    C_Vertex3f(0.0563879013f, 0.429973841f, 0.f),
    C_Vertex3f(0.060310483f, 0.204694152f, 0.f),
    C_Vertex3f(0.0623650551f, 0.451001048f, 0.f),
    C_Vertex3f(0.0666612387f, 0.183961034f, 0.f),
    C_Vertex3f(0.0695376396f, 0.471229196f, 0.f),
    C_Vertex3f(0.0740704536f, 0.16359663f, 0.f),
    C_Vertex3f(0.0779056549f, 0.490658283f, 0.f),
    C_Vertex3f(0.0825381279f, 0.143600941f, 0.f),
    C_Vertex3f(0.087469101f, 0.509288311f, 0.f),
    C_Vertex3f(0.0919754505f, 0.124786615f, 0.f),
    C_Vertex3f(0.098043561f, 0.526798368f, 0.f),
    C_Vertex3f(0.102382421f, 0.107153654f, 0.f),
    C_Vertex3f(0.109629035f, 0.543188453f, 0.f),
    C_Vertex3f(0.113759041f, 0.0907020569f, 0.f),
    C_Vertex3f(0.122225523f, 0.558458567f, 0.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, 0.f),
    C_Vertex3f(0.135833025f, 0.572608709f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.610839844f, 0.203514099f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, 0.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, 0.f),
    C_Vertex3f(0.55421102f, 0.0830818415f, 0.f),
    C_Vertex3f(0.567533731f, 0.101453543f, 0.f),
    C_Vertex3f(0.579524755f, 0.121376276f, 0.f),
    C_Vertex3f(0.590184093f, 0.142850041f, 0.f),
    C_Vertex3f(0.599511743f, 0.165874839f, 0.f),
    C_Vertex3f(0.607507706f, 0.190450668f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.523694992f, 0.21481514f, 0.f),
    C_Vertex3f(0.518330455f, 0.19585824f, 0.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, 0.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.511969209f, 0.178143263f, 0.f),
    C_Vertex3f(0.504611254f, 0.161670208f, 0.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, 0.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, 0.f),
    C_Vertex3f(0.518330455f, 0.19585824f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.49625659f, 0.146439075f, 0.f),
    C_Vertex3f(0.486905217f, 0.132449865f, 0.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, 0.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, 0.f),
    C_Vertex3f(0.504611254f, 0.161670208f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.476557136f, 0.119702578f, 0.f),
    C_Vertex3f(0.465212345f, 0.108197212f, 0.f),
    C_Vertex3f(0.468728662f, 0.014980197f, 0.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, 0.f),
    C_Vertex3f(0.486905217f, 0.132449865f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426140428f, 0.0812556744f, 0.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, 0.f),
    C_Vertex3f(0.439846873f, 0.0889532566f, 0.f),
    C_Vertex3f(0.448522329f, 0.0064060688f, 0.f),
    C_Vertex3f(0.452870846f, 0.0979337692f, 0.f),
    C_Vertex3f(0.468728662f, 0.014980197f, 0.f),
    C_Vertex3f(0.465212345f, 0.108197212f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.380926251f, 0.0658605099f, 0.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, 0.f),
    C_Vertex3f(0.396680117f, 0.069709301f, 0.f),
    C_Vertex3f(0.405221105f, -0.00559771061f, 0.f),
    C_Vertex3f(0.411751509f, 0.0748410225f, 0.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, 0.f),
    C_Vertex3f(0.426140428f, 0.0812556744f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.331152916f, 0.0620117188f, 0.f),
    C_Vertex3f(0.33261013f, -0.0107421875f, 0.f),
    C_Vertex3f(0.347371101f, 0.0620117188f, 0.f),
    C_Vertex3f(0.358068466f, -0.0107421875f, 0.f),
    C_Vertex3f(0.364489913f, 0.0632946491f, 0.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, 0.f),
    C_Vertex3f(0.380926251f, 0.0658605099f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.316777229f, 0.0629534721f, 0.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, 0.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, 0.f),
    C_Vertex3f(0.33261013f, -0.0107421875f, 0.f),
    C_Vertex3f(0.331152916f, 0.0620117188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.261868477f, 0.0761380196f, 0.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, 0.f),
    C_Vertex3f(0.275206566f, 0.0714292526f, 0.f),
    C_Vertex3f(0.28444469f, -0.00678789616f, 0.f),
    C_Vertex3f(0.288804054f, 0.0676622391f, 0.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, 0.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.248789787f, 0.0817885399f, 0.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, 0.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, 0.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, 0.f),
    C_Vertex3f(0.261868477f, 0.0761380196f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200768828f, 0.113868952f, 0.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, 0.f),
    C_Vertex3f(0.211676717f, 0.104410887f, 0.f),
    C_Vertex3f(0.221263289f, 0.00902926922f, 0.f),
    C_Vertex3f(0.223410606f, 0.0959148407f, 0.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, 0.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190686941f, 0.124289036f, 0.f),
    C_Vertex3f(0.181431055f, 0.135671139f, 0.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, 0.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, 0.f),
    C_Vertex3f(0.200768828f, 0.113868952f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.17300117f, 0.148015261f, 0.f),
    C_Vertex3f(0.165397286f, 0.161321402f, 0.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, 0.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, 0.f),
    C_Vertex3f(0.181431055f, 0.135671139f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158619404f, 0.175589561f, 0.f),
    C_Vertex3f(0.152667522f, 0.19081974f, 0.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, 0.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, 0.f),
    C_Vertex3f(0.165397286f, 0.161321402f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152667522f, 0.19081974f, 0.f),
    C_Vertex3f(0.147459626f, 0.206698418f, 0.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, 0.f),
    C_Vertex3f(0.142995715f, 0.223225594f, 0.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, 0.f),
    C_Vertex3f(0.139275789f, 0.240401268f, 0.f),
    C_Vertex3f(0.136299849f, 0.258225441f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.134067893f, 0.276698112f, 0.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, 0.f),
    C_Vertex3f(0.136299849f, 0.258225441f, 0.f),
    C_Vertex3f(0.045492053f, 0.291313767f, 0.f),
    C_Vertex3f(0.0444335938f, 0.337874413f, 0.f),
    C_Vertex3f(0.0444335938f, 0.313890457f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.600097656f, 0.465118408f, -1.f),
    C_Vertex3f(0.596873283f, 0.476215363f, -1.f),
    C_Vertex3f(0.589181304f, 0.49716568f, -1.f),
    C_Vertex3f(0.580246091f, 0.516872406f, -1.f),
    C_Vertex3f(0.570067644f, 0.535335541f, -1.f),
    C_Vertex3f(0.558645964f, 0.552555084f, -1.f),
    C_Vertex3f(0.54598105f, 0.568531036f, -1.f),
    C_Vertex3f(0.532072902f, 0.583263397f, -1.f),
    C_Vertex3f(0.51692152f, 0.596752167f, -1.f),
    C_Vertex3f(0.516601563f, 0.445465088f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.51692152f, 0.596752167f, -1.f),
    C_Vertex3f(0.500526905f, 0.608997345f, -1.f),
    C_Vertex3f(0.500778198f, 0.485576272f, -1.f),
    C_Vertex3f(0.507610321f, 0.470481753f, -1.f),
    C_Vertex3f(0.513813972f, 0.454199791f, -1.f),
    C_Vertex3f(0.516601563f, 0.445465088f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.467164993f, 0.534079909f, -1.f),
    C_Vertex3f(0.476511002f, 0.523735166f, -1.f),
    C_Vertex3f(0.483176112f, 0.619711876f, -1.f),
    C_Vertex3f(0.485228539f, 0.512202978f, -1.f),
    C_Vertex3f(0.500526905f, 0.608997345f, -1.f),
    C_Vertex3f(0.493317604f, 0.499483347f, -1.f),
    C_Vertex3f(0.500778198f, 0.485576272f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.464869142f, 0.62889576f, -1.f),
    C_Vertex3f(0.445605993f, 0.636548996f, -1.f),
    C_Vertex3f(0.446587563f, 0.551207066f, -1.f),
    C_Vertex3f(0.457190514f, 0.543237209f, -1.f),
    C_Vertex3f(0.467164993f, 0.534079909f, -1.f),
    C_Vertex3f(0.483176112f, 0.619711876f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.410680294f, 0.569139242f, -1.f),
    C_Vertex3f(0.423332453f, 0.564158082f, -1.f),
    C_Vertex3f(0.425386667f, 0.642671585f, -1.f),
    C_Vertex3f(0.435301542f, 0.55818069f, -1.f),
    C_Vertex3f(0.445605993f, 0.636548996f, -1.f),
    C_Vertex3f(0.446587563f, 0.551207066f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.404211164f, 0.647263527f, -1.f),
    C_Vertex3f(0.382079482f, 0.650324821f, -1.f),
    C_Vertex3f(0.383326769f, 0.576112866f, -1.f),
    C_Vertex3f(0.397345066f, 0.57312417f, -1.f),
    C_Vertex3f(0.410680294f, 0.569139242f, -1.f),
    C_Vertex3f(0.425386667f, 0.642671585f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318256378f, 0.577996135f, -1.f),
    C_Vertex3f(0.335970879f, 0.579101563f, -1.f),
    C_Vertex3f(0.336367607f, 0.651855469f, -1.f),
    C_Vertex3f(0.353240967f, 0.579101563f, -1.f),
    C_Vertex3f(0.358991623f, 0.651855469f, -1.f),
    C_Vertex3f(0.368625402f, 0.57810533f, -1.f),
    C_Vertex3f(0.382079482f, 0.650324821f, -1.f),
    C_Vertex3f(0.383326769f, 0.576112866f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241056442f, 0.55588758f, -1.f),
    C_Vertex3f(0.254980087f, 0.562520146f, -1.f),
    C_Vertex3f(0.255992174f, 0.639767647f, -1.f),
    C_Vertex3f(0.269661903f, 0.568047285f, -1.f),
    C_Vertex3f(0.27541995f, 0.644602776f, -1.f),
    C_Vertex3f(0.285101891f, 0.572468996f, -1.f),
    C_Vertex3f(0.295291781f, 0.648229122f, -1.f),
    C_Vertex3f(0.301300049f, 0.575785279f, -1.f),
    C_Vertex3f(0.315607667f, 0.650646687f, -1.f),
    C_Vertex3f(0.318256378f, 0.577996135f, -1.f),
    C_Vertex3f(0.336367607f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.203998804f, 0.529705167f, -1.f),
    C_Vertex3f(0.215483665f, 0.539306164f, -1.f),
    C_Vertex3f(0.218468785f, 0.626471043f, -1.f),
    C_Vertex3f(0.227890968f, 0.548149586f, -1.f),
    C_Vertex3f(0.237008452f, 0.633723736f, -1.f),
    C_Vertex3f(0.241056442f, 0.55588758f, -1.f),
    C_Vertex3f(0.255992174f, 0.639767647f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.200373173f, 0.618009567f, -1.f),
    C_Vertex3f(0.182721615f, 0.60833931f, -1.f),
    C_Vertex3f(0.183796406f, 0.508230448f, -1.f),
    C_Vertex3f(0.193436384f, 0.519346595f, -1.f),
    C_Vertex3f(0.203998804f, 0.529705167f, -1.f),
    C_Vertex3f(0.218468785f, 0.626471043f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154460907f, 0.456190109f, -1.f),
    C_Vertex3f(0.160411119f, 0.470336556f, -1.f),
    C_Vertex3f(0.166081071f, 0.597549081f, -1.f),
    C_Vertex3f(0.167283773f, 0.483725429f, -1.f),
    C_Vertex3f(0.182721615f, 0.60833931f, -1.f),
    C_Vertex3f(0.175078869f, 0.496356726f, -1.f),
    C_Vertex3f(0.183796406f, 0.508230448f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150451541f, 0.585638881f, -1.f),
    C_Vertex3f(0.135833025f, 0.572608709f, -1.f),
    C_Vertex3f(0.139377594f, 0.403562546f, -1.f),
    C_Vertex3f(0.14691925f, 0.433834076f, -1.f),
    C_Vertex3f(0.154460907f, 0.456190109f, -1.f),
    C_Vertex3f(0.166081071f, 0.597549081f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.134067893f, 0.276698112f, -1.f),
    C_Vertex3f(0.132579923f, 0.295819283f, -1.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, -1.f),
    C_Vertex3f(0.131835938f, 0.315588951f, -1.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, -1.f),
    C_Vertex3f(0.131835938f, 0.341629028f, -1.f),
    C_Vertex3f(0.135833025f, 0.572608709f, -1.f),
    C_Vertex3f(0.134349823f, 0.37282753f, -1.f),
    C_Vertex3f(0.139377594f, 0.403562546f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0444335938f, 0.337874413f, -1.f),
    C_Vertex3f(0.045492053f, 0.291313767f, -1.f),
    C_Vertex3f(0.0456290245f, 0.362097859f, -1.f),
    C_Vertex3f(0.0476089716f, 0.269105792f, -1.f),
    C_Vertex3f(0.048019886f, 0.385522246f, -1.f),
    C_Vertex3f(0.0507843494f, 0.247266531f, -1.f),
    C_Vertex3f(0.0516061783f, 0.408147573f, -1.f),
    C_Vertex3f(0.0550181866f, 0.225795984f, -1.f),
    C_Vertex3f(0.0563879013f, 0.429973841f, -1.f),
    C_Vertex3f(0.060310483f, 0.204694152f, -1.f),
    C_Vertex3f(0.0623650551f, 0.451001048f, -1.f),
    C_Vertex3f(0.0666612387f, 0.183961034f, -1.f),
    C_Vertex3f(0.0695376396f, 0.471229196f, -1.f),
    C_Vertex3f(0.0740704536f, 0.16359663f, -1.f),
    C_Vertex3f(0.0779056549f, 0.490658283f, -1.f),
    C_Vertex3f(0.0825381279f, 0.143600941f, -1.f),
    C_Vertex3f(0.087469101f, 0.509288311f, -1.f),
    C_Vertex3f(0.0919754505f, 0.124786615f, -1.f),
    C_Vertex3f(0.098043561f, 0.526798368f, -1.f),
    C_Vertex3f(0.102382421f, 0.107153654f, -1.f),
    C_Vertex3f(0.109629035f, 0.543188453f, -1.f),
    C_Vertex3f(0.113759041f, 0.0907020569f, -1.f),
    C_Vertex3f(0.122225523f, 0.558458567f, -1.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, -1.f),
    C_Vertex3f(0.135833025f, 0.572608709f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.610839844f, 0.203514099f, -1.f),
    C_Vertex3f(0.525878906f, 0.224914551f, -1.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, -1.f),
    C_Vertex3f(0.55421102f, 0.0830818415f, -1.f),
    C_Vertex3f(0.567533731f, 0.101453543f, -1.f),
    C_Vertex3f(0.579524755f, 0.121376276f, -1.f),
    C_Vertex3f(0.590184093f, 0.142850041f, -1.f),
    C_Vertex3f(0.599511743f, 0.165874839f, -1.f),
    C_Vertex3f(0.607507706f, 0.190450668f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.523694992f, 0.21481514f, -1.f),
    C_Vertex3f(0.518330455f, 0.19585824f, -1.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, -1.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, -1.f),
    C_Vertex3f(0.525878906f, 0.224914551f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.511969209f, 0.178143263f, -1.f),
    C_Vertex3f(0.504611254f, 0.161670208f, -1.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, -1.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, -1.f),
    C_Vertex3f(0.518330455f, 0.19585824f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.49625659f, 0.146439075f, -1.f),
    C_Vertex3f(0.486905217f, 0.132449865f, -1.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, -1.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, -1.f),
    C_Vertex3f(0.504611254f, 0.161670208f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.476557136f, 0.119702578f, -1.f),
    C_Vertex3f(0.465212345f, 0.108197212f, -1.f),
    C_Vertex3f(0.468728662f, 0.014980197f, -1.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, -1.f),
    C_Vertex3f(0.486905217f, 0.132449865f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426140428f, 0.0812556744f, -1.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, -1.f),
    C_Vertex3f(0.439846873f, 0.0889532566f, -1.f),
    C_Vertex3f(0.448522329f, 0.0064060688f, -1.f),
    C_Vertex3f(0.452870846f, 0.0979337692f, -1.f),
    C_Vertex3f(0.468728662f, 0.014980197f, -1.f),
    C_Vertex3f(0.465212345f, 0.108197212f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.380926251f, 0.0658605099f, -1.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, -1.f),
    C_Vertex3f(0.396680117f, 0.069709301f, -1.f),
    C_Vertex3f(0.405221105f, -0.00559771061f, -1.f),
    C_Vertex3f(0.411751509f, 0.0748410225f, -1.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, -1.f),
    C_Vertex3f(0.426140428f, 0.0812556744f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.331152916f, 0.0620117188f, -1.f),
    C_Vertex3f(0.33261013f, -0.0107421875f, -1.f),
    C_Vertex3f(0.347371101f, 0.0620117188f, -1.f),
    C_Vertex3f(0.358068466f, -0.0107421875f, -1.f),
    C_Vertex3f(0.364489913f, 0.0632946491f, -1.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, -1.f),
    C_Vertex3f(0.380926251f, 0.0658605099f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.316777229f, 0.0629534721f, -1.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, -1.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, -1.f),
    C_Vertex3f(0.33261013f, -0.0107421875f, -1.f),
    C_Vertex3f(0.331152916f, 0.0620117188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.261868477f, 0.0761380196f, -1.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, -1.f),
    C_Vertex3f(0.275206566f, 0.0714292526f, -1.f),
    C_Vertex3f(0.28444469f, -0.00678789616f, -1.f),
    C_Vertex3f(0.288804054f, 0.0676622391f, -1.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, -1.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.248789787f, 0.0817885399f, -1.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, -1.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, -1.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, -1.f),
    C_Vertex3f(0.261868477f, 0.0761380196f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200768828f, 0.113868952f, -1.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, -1.f),
    C_Vertex3f(0.211676717f, 0.104410887f, -1.f),
    C_Vertex3f(0.221263289f, 0.00902926922f, -1.f),
    C_Vertex3f(0.223410606f, 0.0959148407f, -1.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, -1.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190686941f, 0.124289036f, -1.f),
    C_Vertex3f(0.181431055f, 0.135671139f, -1.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, -1.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, -1.f),
    C_Vertex3f(0.200768828f, 0.113868952f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.17300117f, 0.148015261f, -1.f),
    C_Vertex3f(0.165397286f, 0.161321402f, -1.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, -1.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, -1.f),
    C_Vertex3f(0.181431055f, 0.135671139f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158619404f, 0.175589561f, -1.f),
    C_Vertex3f(0.152667522f, 0.19081974f, -1.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, -1.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, -1.f),
    C_Vertex3f(0.165397286f, 0.161321402f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152667522f, 0.19081974f, -1.f),
    C_Vertex3f(0.147459626f, 0.206698418f, -1.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, -1.f),
    C_Vertex3f(0.142995715f, 0.223225594f, -1.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, -1.f),
    C_Vertex3f(0.139275789f, 0.240401268f, -1.f),
    C_Vertex3f(0.136299849f, 0.258225441f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.134067893f, 0.276698112f, -1.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, -1.f),
    C_Vertex3f(0.136299849f, 0.258225441f, -1.f),
    C_Vertex3f(0.045492053f, 0.291313767f, -1.f),
    C_Vertex3f(0.0444335938f, 0.337874413f, -1.f),
    C_Vertex3f(0.0444335938f, 0.313890457f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_68[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Normal3f(0.00115807739f, 0.999999344f, 0.f),
    C_Vertex3f(0.298933029f, 0.640625f, 0.f),
    C_Vertex3f(0.298933029f, 0.640625f, -1.f),
    C_Normal3f(0.0252576191f, 0.999680996f, 0.f),
    C_Vertex3f(0.316514492f, 0.640338182f, 0.f),
    C_Vertex3f(0.316514492f, 0.640338182f, -1.f),
    C_Normal3f(0.0449910089f, 0.998987436f, 0.f),
    C_Vertex3f(0.332989216f, 0.639764547f, 0.f),
    C_Vertex3f(0.332989216f, 0.639764547f, -1.f),
    C_Normal3f(0.0676065236f, 0.997712076f, 0.f),
    C_Vertex3f(0.348357201f, 0.638904095f, 0.f),
    C_Vertex3f(0.348357201f, 0.638904095f, -1.f),
    C_Normal3f(0.093741402f, 0.995596528f, 0.f),
    C_Vertex3f(0.362618446f, 0.637756824f, 0.f),
    C_Vertex3f(0.362618446f, 0.637756824f, -1.f),
    C_Normal3f(0.124217272f, 0.992255032f, 0.f),
    C_Vertex3f(0.375772953f, 0.636322737f, 0.f),
    C_Vertex3f(0.375772953f, 0.636322737f, -1.f),
    C_Normal3f(0.160101086f, 0.987100661f, 0.f),
    C_Vertex3f(0.387820721f, 0.634601831f, 0.f),
    C_Vertex3f(0.387820721f, 0.634601831f, -1.f),
    C_Normal3f(0.208720669f, 0.977975309f, 0.f),
    C_Vertex3f(0.398761749f, 0.632594109f, 0.f),
    C_Vertex3f(0.398761749f, 0.632594109f, -1.f),
    C_Normal3f(0.286513388f, 0.958076298f, 0.f),
    C_Vertex3f(0.417448044f, 0.628278732f, 0.f),
    C_Vertex3f(0.417448044f, 0.628278732f, -1.f),
    C_Normal3f(0.383859992f, 0.923391342f, 0.f),
    C_Vertex3f(0.442964554f, 0.619375229f, 0.f),
    C_Vertex3f(0.442964554f, 0.619375229f, -1.f),
    C_Normal3f(0.493410945f, 0.869796336f, 0.f),
    C_Vertex3f(0.466459274f, 0.607904434f, 0.f),
    C_Vertex3f(0.466459274f, 0.607904434f, -1.f),
    C_Normal3f(0.593687773f, 0.804695487f, 0.f),
    C_Vertex3f(0.487932205f, 0.593866348f, 0.f),
    C_Vertex3f(0.487932205f, 0.593866348f, -1.f),
    C_Normal3f(0.668363988f, 0.743834376f, 0.f),
    C_Vertex3f(0.503995895f, 0.580210686f, 0.f),
    C_Vertex3f(0.503995895f, 0.580210686f, -1.f),
    C_Normal3f(0.717492461f, 0.696566164f, 0.f),
    C_Vertex3f(0.515873075f, 0.568760395f, 0.f),
    C_Vertex3f(0.515873075f, 0.568760395f, -1.f),
    C_Normal3f(0.762034953f, 0.64753598f, 0.f),
    C_Vertex3f(0.526951194f, 0.556565762f, 0.f),
    C_Vertex3f(0.526951194f, 0.556565762f, -1.f),
    C_Normal3f(0.802996516f, 0.595983744f, 0.f),
    C_Vertex3f(0.537230253f, 0.543626785f, 0.f),
    C_Vertex3f(0.537230253f, 0.543626785f, -1.f),
    C_Normal3f(0.839958608f, 0.542650461f, 0.f),
    C_Vertex3f(0.546710253f, 0.529943466f, 0.f),
    C_Vertex3f(0.546710253f, 0.529943466f, -1.f),
    C_Normal3f(0.87266928f, 0.488311708f, 0.f),
    C_Vertex3f(0.555391192f, 0.515515804f, 0.f),
    C_Vertex3f(0.555391192f, 0.515515804f, -1.f),
    C_Normal3f(0.901044369f, 0.433727026f, 0.f),
    C_Vertex3f(0.563273072f, 0.5003438f, 0.f),
    C_Vertex3f(0.563273072f, 0.5003438f, -1.f),
    C_Normal3f(0.925152183f, 0.379596233f, 0.f),
    C_Vertex3f(0.570355892f, 0.484427452f, 0.f),
    C_Vertex3f(0.570355892f, 0.484427452f, -1.f),
    C_Normal3f(0.944622636f, 0.328158587f, 0.f),
    C_Vertex3f(0.576639652f, 0.467766762f, 0.f),
    C_Vertex3f(0.576639652f, 0.467766762f, -1.f),
    C_Normal3f(0.960200369f, 0.279311985f, 0.f),
    C_Vertex3f(0.582137942f, 0.450512171f, 0.f),
    C_Vertex3f(0.582137942f, 0.450512171f, -1.f),
    C_Normal3f(0.972809255f, 0.231607407f, 0.f),
    C_Vertex3f(0.586850762f, 0.432663679f, 0.f),
    C_Vertex3f(0.586850762f, 0.432663679f, -1.f),
    C_Normal3f(0.982672215f, 0.185351655f, 0.f),
    C_Vertex3f(0.590778112f, 0.414221287f, 0.f),
    C_Vertex3f(0.590778112f, 0.414221287f, -1.f),
    C_Normal3f(0.99004072f, 0.140781626f, 0.f),
    C_Vertex3f(0.593919992f, 0.395184994f, 0.f),
    C_Vertex3f(0.593919992f, 0.395184994f, -1.f),
    C_Normal3f(0.995179772f, 0.0980677232f, 0.f),
    C_Vertex3f(0.596276402f, 0.3755548f, 0.f),
    C_Vertex3f(0.596276402f, 0.3755548f, -1.f),
    C_Normal3f(0.998355806f, 0.057320077f, 0.f),
    C_Vertex3f(0.597847342f, 0.355330706f, 0.f),
    C_Vertex3f(0.597847342f, 0.355330706f, -1.f),
    C_Normal3f(0.999875426f, 0.0157807507f, 0.f),
    C_Vertex3f(0.598632813f, 0.334512711f, 0.f),
    C_Vertex3f(0.598632813f, 0.334512711f, -1.f),
    C_Normal3f(0.999434948f, -0.0336130969f, 0.f),
    C_Vertex3f(0.598632813f, 0.305562973f, 0.f),
    C_Vertex3f(0.598632813f, 0.305562973f, -1.f),
    C_Normal3f(0.995444655f, -0.0953409821f, 0.f),
    C_Vertex3f(0.596501827f, 0.271151066f, 0.f),
    C_Vertex3f(0.596501827f, 0.271151066f, -1.f),
    C_Normal3f(0.98582083f, -0.167801693f, 0.f),
    C_Vertex3f(0.592239857f, 0.238814831f, 0.f),
    C_Vertex3f(0.592239857f, 0.238814831f, -1.f),
    C_Normal3f(0.968939006f, -0.247300044f, 0.f),
    C_Vertex3f(0.585846901f, 0.208554268f, 0.f),
    C_Vertex3f(0.585846901f, 0.208554268f, -1.f),
    C_Normal3f(0.947250128f, -0.320495278f, 0.f),
    C_Vertex3f(0.57732296f, 0.180369377f, 0.f),
    C_Vertex3f(0.57732296f, 0.180369377f, -1.f),
    C_Normal3f(0.921705365f, -0.387890875f, 0.f),
    C_Vertex3f(0.567596912f, 0.154614925f, 0.f),
    C_Vertex3f(0.567596912f, 0.154614925f, -1.f),
    C_Normal3f(0.886680424f, -0.462382644f, 0.f),
    C_Vertex3f(0.556668758f, 0.131290913f, 0.f),
    C_Vertex3f(0.556668758f, 0.131290913f, -1.f),
    C_Normal3f(0.839603961f, -0.543198943f, 0.f),
    C_Vertex3f(0.544538498f, 0.110397339f, 0.f),
    C_Vertex3f(0.544538498f, 0.110397339f, -1.f),
    C_Normal3f(0.788238823f, -0.615369439f, 0.f),
    C_Vertex3f(0.531206131f, 0.0919342041f, 0.f),
    C_Vertex3f(0.531206131f, 0.0919342041f, -1.f),
    C_Normal3f(0.735883474f, -0.677108228f, 0.f),
    C_Vertex3f(0.51724577f, 0.0754375458f, 0.f),
    C_Vertex3f(0.51724577f, 0.0754375458f, -1.f),
    C_Normal3f(0.672655046f, -0.739956081f, 0.f),
    C_Vertex3f(0.502657413f, 0.0609073639f, 0.f),
    C_Vertex3f(0.502657413f, 0.0609073639f, -1.f),
    C_Normal3f(0.597774565f, -0.801664293f, 0.f),
    C_Vertex3f(0.487441063f, 0.0483436584f, 0.f),
    C_Vertex3f(0.487441063f, 0.0483436584f, -1.f),
    C_Normal3f(0.512860179f, -0.858472168f, 0.f),
    C_Vertex3f(0.471596718f, 0.0377464294f, 0.f),
    C_Vertex3f(0.471596718f, 0.0377464294f, -1.f),
    C_Normal3f(0.428142518f, -0.9037112f, 0.f),
    C_Vertex3f(0.45411253f, 0.0284328461f, 0.f),
    C_Vertex3f(0.45411253f, 0.0284328461f, -1.f),
    C_Normal3f(0.347374946f, -0.937726259f, 0.f),
    C_Vertex3f(0.434988499f, 0.0204029083f, 0.f),
    C_Vertex3f(0.434988499f, 0.0204029083f, -1.f),
    C_Normal3f(0.272151053f, -0.962254465f, 0.f),
    C_Vertex3f(0.414224625f, 0.0136566162f, 0.f),
    C_Vertex3f(0.414224625f, 0.0136566162f, -1.f),
    C_Normal3f(0.201482207f, -0.979492128f, 0.f),
    C_Vertex3f(0.391820908f, 0.00819396973f, 0.f),
    C_Vertex3f(0.391820908f, 0.00819396973f, -1.f),
    C_Normal3f(0.13582851f, -0.990732372f, 0.f),
    C_Vertex3f(0.367751122f, 0.00409698486f, 0.f),
    C_Vertex3f(0.367751122f, 0.00409698486f, -1.f),
    C_Normal3f(0.0768730193f, -0.997040927f, 0.f),
    C_Vertex3f(0.342015266f, 0.00136566162f, 0.f),
    C_Vertex3f(0.342015266f, 0.00136566162f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.314613342f, 0.f, 0.f),
    C_Vertex3f(0.314613342f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.314613342f, 0.f, 0.f),
    C_Vertex3f(0.314613342f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, 0.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, -1.f),
    C_Normal3f(-0.00815145951f, 0.9999668f, 0.f),
    C_Vertex3f(0.306411743f, 0.0756835938f, 0.f),
    C_Vertex3f(0.306411743f, 0.0756835938f, -1.f),
    C_Normal3f(-0.0829877257f, 0.99655056f, 0.f),
    C_Vertex3f(0.334675789f, 0.0771579742f, 0.f),
    C_Vertex3f(0.334675789f, 0.0771579742f, -1.f),
    C_Normal3f(-0.157277852f, 0.987554371f, 0.f),
    C_Vertex3f(0.359526634f, 0.0801067352f, 0.f),
    C_Vertex3f(0.359526634f, 0.0801067352f, -1.f),
    C_Normal3f(-0.253023356f, 0.967460155f, 0.f),
    C_Vertex3f(0.380964279f, 0.0845298767f, 0.f),
    C_Vertex3f(0.380964279f, 0.0845298767f, -1.f),
    C_Normal3f(-0.35504064f, 0.934850872f, 0.f),
    C_Vertex3f(0.398988724f, 0.0904273987f, 0.f),
    C_Vertex3f(0.398988724f, 0.0904273987f, -1.f),
    C_Normal3f(-0.44996503f, 0.893046141f, 0.f),
    C_Vertex3f(0.415183067f, 0.0975255966f, 0.f),
    C_Vertex3f(0.415183067f, 0.0975255966f, -1.f),
    C_Normal3f(-0.551824272f, 0.833960414f, 0.f),
    C_Vertex3f(0.42954731f, 0.105824471f, 0.f),
    C_Vertex3f(0.42954731f, 0.105824471f, -1.f),
    C_Normal3f(-0.660835385f, 0.750530839f, 0.f),
    C_Vertex3f(0.442081451f, 0.11532402f, 0.f),
    C_Vertex3f(0.442081451f, 0.11532402f, -1.f),
    C_Normal3f(-0.757875502f, 0.652399182f, 0.f),
    C_Vertex3f(0.454971313f, 0.128210068f, 0.f),
    C_Vertex3f(0.454971313f, 0.128210068f, -1.f),
    C_Normal3f(-0.833363414f, 0.552725554f, 0.f),
    C_Vertex3f(0.468380928f, 0.145875454f, 0.f),
    C_Vertex3f(0.468380928f, 0.145875454f, -1.f),
    C_Normal3f(-0.892219663f, 0.451601654f, 0.f),
    C_Vertex3f(0.480124474f, 0.166134357f, 0.f),
    C_Vertex3f(0.480124474f, 0.166134357f, -1.f),
    C_Normal3f(-0.933910728f, 0.357506275f, 0.f),
    C_Vertex3f(0.49020195f, 0.188986778f, 0.f),
    C_Vertex3f(0.49020195f, 0.188986778f, -1.f),
    C_Normal3f(-0.964291275f, 0.264844f, 0.f),
    C_Vertex3f(0.498613358f, 0.214432716f, 0.f),
    C_Vertex3f(0.498613358f, 0.214432716f, -1.f),
    C_Normal3f(-0.984490156f, 0.175439522f, 0.f),
    C_Vertex3f(0.504921913f, 0.242581844f, 0.f),
    C_Vertex3f(0.504921913f, 0.242581844f, -1.f),
    C_Normal3f(-0.99523741f, 0.0974805579f, 0.f),
    C_Vertex3f(0.509127617f, 0.273434162f, 0.f),
    C_Vertex3f(0.509127617f, 0.273434162f, -1.f),
    C_Normal3f(-0.99946475f, 0.0327148512f, 0.f),
    C_Vertex3f(0.511230469f, 0.30698967f, 0.f),
    C_Vertex3f(0.511230469f, 0.30698967f, -1.f),
    C_Normal3f(-0.999820232f, -0.0189639982f, 0.f),
    C_Vertex3f(0.511230469f, 0.337677956f, 0.f),
    C_Vertex3f(0.511230469f, 0.337677956f, -1.f),
    C_Normal3f(-0.997735023f, -0.0672658309f, 0.f),
    C_Vertex3f(0.510199547f, 0.361341953f, 0.f),
    C_Vertex3f(0.510199547f, 0.361341953f, -1.f),
    C_Normal3f(-0.992880464f, -0.119114362f, 0.f),
    C_Vertex3f(0.508137703f, 0.383552074f, 0.f),
    C_Vertex3f(0.508137703f, 0.383552074f, -1.f),
    C_Normal3f(-0.984158039f, -0.177293479f, 0.f),
    C_Vertex3f(0.505044937f, 0.404308319f, 0.f),
    C_Vertex3f(0.505044937f, 0.404308319f, -1.f),
    C_Normal3f(-0.97020036f, -0.242304146f, 0.f),
    C_Vertex3f(0.500921249f, 0.423610687f, 0.f),
    C_Vertex3f(0.500921249f, 0.423610687f, -1.f),
    C_Normal3f(-0.949298799f, -0.314375162f, 0.f),
    C_Vertex3f(0.49576664f, 0.441459179f, 0.f),
    C_Vertex3f(0.49576664f, 0.441459179f, -1.f),
    C_Normal3f(-0.919436395f, -0.393238693f, 0.f),
    C_Vertex3f(0.489581108f, 0.457853794f, 0.f),
    C_Vertex3f(0.489581108f, 0.457853794f, -1.f),
    C_Normal3f(-0.878442347f, -0.477848351f, 0.f),
    C_Vertex3f(0.482364655f, 0.472794533f, 0.f),
    C_Vertex3f(0.482364655f, 0.472794533f, -1.f),
    C_Normal3f(-0.836537242f, -0.547909915f, 0.f),
    C_Vertex3f(0.474117279f, 0.486281395f, 0.f),
    C_Vertex3f(0.474117279f, 0.486281395f, -1.f),
    C_Normal3f(-0.798398674f, -0.602129102f, 0.f),
    C_Vertex3f(0.465426207f, 0.498655796f, 0.f),
    C_Vertex3f(0.465426207f, 0.498655796f, -1.f),
    C_Normal3f(-0.752955735f, -0.65807122f, 0.f),
    C_Vertex3f(0.456291437f, 0.509917736f, 0.f),
    C_Vertex3f(0.456291437f, 0.509917736f, -1.f),
    C_Normal3f(-0.699517012f, -0.714615822f, 0.f),
    C_Vertex3f(0.446712971f, 0.520067215f, 0.f),
    C_Vertex3f(0.446712971f, 0.520067215f, -1.f),
    C_Normal3f(-0.637700915f, -0.770284116f, 0.f),
    C_Vertex3f(0.436690807f, 0.529104233f, 0.f),
    C_Vertex3f(0.436690807f, 0.529104233f, -1.f),
    C_Normal3f(-0.567602038f, -0.823303044f, 0.f),
    C_Vertex3f(0.426224947f, 0.53702879f, 0.f),
    C_Vertex3f(0.426224947f, 0.53702879f, -1.f),
    C_Normal3f(-0.489930987f, -0.871761262f, 0.f),
    C_Vertex3f(0.41531539f, 0.543840885f, 0.f),
    C_Vertex3f(0.41531539f, 0.543840885f, -1.f),
    C_Normal3f(-0.408646643f, -0.912692726f, 0.f),
    C_Vertex3f(0.403962135f, 0.54954052f, 0.f),
    C_Vertex3f(0.403962135f, 0.54954052f, -1.f),
    C_Normal3f(-0.322217703f, -0.946665645f, 0.f),
    C_Vertex3f(0.393803596f, 0.553472519f, 0.f),
    C_Vertex3f(0.393803596f, 0.553472519f, -1.f),
    C_Normal3f(-0.246536016f, -0.969133556f, 0.f),
    C_Vertex3f(0.383986235f, 0.556339741f, 0.f),
    C_Vertex3f(0.383986235f, 0.556339741f, -1.f),
    C_Normal3f(-0.188095346f, -0.982150733f, 0.f),
    C_Vertex3f(0.372871637f, 0.558797359f, 0.f),
    C_Vertex3f(0.372871637f, 0.558797359f, -1.f),
    C_Normal3f(-0.139744624f, -0.990187645f, 0.f),
    C_Vertex3f(0.360459805f, 0.560845375f, 0.f),
    C_Vertex3f(0.360459805f, 0.560845375f, -1.f),
    C_Normal3f(-0.0993556604f, -0.99505198f, 0.f),
    C_Vertex3f(0.346750736f, 0.562483788f, 0.f),
    C_Vertex3f(0.346750736f, 0.562483788f, -1.f),
    C_Normal3f(-0.0652717501f, -0.997867525f, 0.f),
    C_Vertex3f(0.331744432f, 0.563712597f, 0.f),
    C_Vertex3f(0.331744432f, 0.563712597f, -1.f),
    C_Normal3f(-0.0362196714f, -0.999343812f, 0.f),
    C_Vertex3f(0.315440893f, 0.564531803f, 0.f),
    C_Vertex3f(0.315440893f, 0.564531803f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.297840118f, 0.564941406f, 0.f),
    C_Vertex3f(0.297840118f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.297840118f, 0.564941406f, 0.f),
    C_Vertex3f(0.297840118f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, -1.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563273072f, 0.5003438f, 0.f),
    C_Vertex3f(0.555391192f, 0.515515804f, 0.f),
    C_Vertex3f(0.556668758f, 0.131290913f, 0.f),
    C_Vertex3f(0.546710253f, 0.529943466f, 0.f),
    C_Vertex3f(0.544538498f, 0.110397339f, 0.f),
    C_Vertex3f(0.537230253f, 0.543626785f, 0.f),
    C_Vertex3f(0.531206131f, 0.0919342041f, 0.f),
    C_Vertex3f(0.526951194f, 0.556565762f, 0.f),
    C_Vertex3f(0.51724577f, 0.0754375458f, 0.f),
    C_Vertex3f(0.515873075f, 0.568760395f, 0.f),
    C_Vertex3f(0.511230469f, 0.337677956f, 0.f),
    C_Vertex3f(0.510199547f, 0.361341953f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.49576664f, 0.441459179f, 0.f),
    C_Vertex3f(0.500921249f, 0.423610687f, 0.f),
    C_Vertex3f(0.503995895f, 0.580210686f, 0.f),
    C_Vertex3f(0.505044937f, 0.404308319f, 0.f),
    C_Vertex3f(0.515873075f, 0.568760395f, 0.f),
    C_Vertex3f(0.508137703f, 0.383552074f, 0.f),
    C_Vertex3f(0.510199547f, 0.361341953f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.474117279f, 0.486281395f, 0.f),
    C_Vertex3f(0.482364655f, 0.472794533f, 0.f),
    C_Vertex3f(0.487932205f, 0.593866348f, 0.f),
    C_Vertex3f(0.489581108f, 0.457853794f, 0.f),
    C_Vertex3f(0.503995895f, 0.580210686f, 0.f),
    C_Vertex3f(0.49576664f, 0.441459179f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.466459274f, 0.607904434f, 0.f),
    C_Vertex3f(0.442964554f, 0.619375229f, 0.f),
    C_Vertex3f(0.446712971f, 0.520067215f, 0.f),
    C_Vertex3f(0.456291437f, 0.509917736f, 0.f),
    C_Vertex3f(0.465426207f, 0.498655796f, 0.f),
    C_Vertex3f(0.474117279f, 0.486281395f, 0.f),
    C_Vertex3f(0.487932205f, 0.593866348f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.403962135f, 0.54954052f, 0.f),
    C_Vertex3f(0.41531539f, 0.543840885f, 0.f),
    C_Vertex3f(0.417448044f, 0.628278732f, 0.f),
    C_Vertex3f(0.426224947f, 0.53702879f, 0.f),
    C_Vertex3f(0.442964554f, 0.619375229f, 0.f),
    C_Vertex3f(0.436690807f, 0.529104233f, 0.f),
    C_Vertex3f(0.446712971f, 0.520067215f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.297840118f, 0.564941406f, 0.f),
    C_Vertex3f(0.298933029f, 0.640625f, 0.f),
    C_Vertex3f(0.315440893f, 0.564531803f, 0.f),
    C_Vertex3f(0.316514492f, 0.640338182f, 0.f),
    C_Vertex3f(0.331744432f, 0.563712597f, 0.f),
    C_Vertex3f(0.332989216f, 0.639764547f, 0.f),
    C_Vertex3f(0.346750736f, 0.562483788f, 0.f),
    C_Vertex3f(0.348357201f, 0.638904095f, 0.f),
    C_Vertex3f(0.360459805f, 0.560845375f, 0.f),
    C_Vertex3f(0.362618446f, 0.637756824f, 0.f),
    C_Vertex3f(0.372871637f, 0.558797359f, 0.f),
    C_Vertex3f(0.375772953f, 0.636322737f, 0.f),
    C_Vertex3f(0.383986235f, 0.556339741f, 0.f),
    C_Vertex3f(0.387820721f, 0.634601831f, 0.f),
    C_Vertex3f(0.393803596f, 0.553472519f, 0.f),
    C_Vertex3f(0.398761749f, 0.632594109f, 0.f),
    C_Vertex3f(0.403962135f, 0.54954052f, 0.f),
    C_Vertex3f(0.417448044f, 0.628278732f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.153808594f, 0.0756835938f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_Vertex3f(0.314613342f, 0.f, 0.f),
    C_Vertex3f(0.306411743f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.597847342f, 0.355330706f, 0.f),
    C_Vertex3f(0.596276402f, 0.3755548f, 0.f),
    C_Vertex3f(0.596501827f, 0.271151066f, 0.f),
    C_Vertex3f(0.598632813f, 0.305562973f, 0.f),
    C_Vertex3f(0.598632813f, 0.334512711f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.593919992f, 0.395184994f, 0.f),
    C_Vertex3f(0.590778112f, 0.414221287f, 0.f),
    C_Vertex3f(0.592239857f, 0.238814831f, 0.f),
    C_Vertex3f(0.596501827f, 0.271151066f, 0.f),
    C_Vertex3f(0.596276402f, 0.3755548f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.576639652f, 0.467766762f, 0.f),
    C_Vertex3f(0.57732296f, 0.180369377f, 0.f),
    C_Vertex3f(0.582137942f, 0.450512171f, 0.f),
    C_Vertex3f(0.585846901f, 0.208554268f, 0.f),
    C_Vertex3f(0.586850762f, 0.432663679f, 0.f),
    C_Vertex3f(0.592239857f, 0.238814831f, 0.f),
    C_Vertex3f(0.590778112f, 0.414221287f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.570355892f, 0.484427452f, 0.f),
    C_Vertex3f(0.563273072f, 0.5003438f, 0.f),
    C_Vertex3f(0.567596912f, 0.154614925f, 0.f),
    C_Vertex3f(0.57732296f, 0.180369377f, 0.f),
    C_Vertex3f(0.576639652f, 0.467766762f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.51724577f, 0.0754375458f, 0.f),
    C_Vertex3f(0.511230469f, 0.337677956f, 0.f),
    C_Vertex3f(0.511230469f, 0.30698967f, 0.f),
    C_Vertex3f(0.509127617f, 0.273434162f, 0.f),
    C_Vertex3f(0.504921913f, 0.242581844f, 0.f),
    C_Vertex3f(0.502657413f, 0.0609073639f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.504921913f, 0.242581844f, 0.f),
    C_Vertex3f(0.498613358f, 0.214432716f, 0.f),
    C_Vertex3f(0.502657413f, 0.0609073639f, 0.f),
    C_Vertex3f(0.49020195f, 0.188986778f, 0.f),
    C_Vertex3f(0.487441063f, 0.0483436584f, 0.f),
    C_Vertex3f(0.480124474f, 0.166134357f, 0.f),
    C_Vertex3f(0.471596718f, 0.0377464294f, 0.f),
    C_Vertex3f(0.468380928f, 0.145875454f, 0.f),
    C_Vertex3f(0.454971313f, 0.128210068f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.471596718f, 0.0377464294f, 0.f),
    C_Vertex3f(0.454971313f, 0.128210068f, 0.f),
    C_Vertex3f(0.45411253f, 0.0284328461f, 0.f),
    C_Vertex3f(0.442081451f, 0.11532402f, 0.f),
    C_Vertex3f(0.434988499f, 0.0204029083f, 0.f),
    C_Vertex3f(0.42954731f, 0.105824471f, 0.f),
    C_Vertex3f(0.415183067f, 0.0975255966f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.306411743f, 0.0756835938f, 0.f),
    C_Vertex3f(0.314613342f, 0.f, 0.f),
    C_Vertex3f(0.334675789f, 0.0771579742f, 0.f),
    C_Vertex3f(0.342015266f, 0.00136566162f, 0.f),
    C_Vertex3f(0.359526634f, 0.0801067352f, 0.f),
    C_Vertex3f(0.367751122f, 0.00409698486f, 0.f),
    C_Vertex3f(0.380964279f, 0.0845298767f, 0.f),
    C_Vertex3f(0.391820908f, 0.00819396973f, 0.f),
    C_Vertex3f(0.398988724f, 0.0904273987f, 0.f),
    C_Vertex3f(0.414224625f, 0.0136566162f, 0.f),
    C_Vertex3f(0.415183067f, 0.0975255966f, 0.f),
    C_Vertex3f(0.434988499f, 0.0204029083f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.563273072f, 0.5003438f, 0.f),
    C_Vertex3f(0.556668758f, 0.131290913f, 0.f),
    C_Vertex3f(0.567596912f, 0.154614925f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.298933029f, 0.640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563273072f, 0.5003438f, -1.f),
    C_Vertex3f(0.555391192f, 0.515515804f, -1.f),
    C_Vertex3f(0.556668758f, 0.131290913f, -1.f),
    C_Vertex3f(0.546710253f, 0.529943466f, -1.f),
    C_Vertex3f(0.544538498f, 0.110397339f, -1.f),
    C_Vertex3f(0.537230253f, 0.543626785f, -1.f),
    C_Vertex3f(0.531206131f, 0.0919342041f, -1.f),
    C_Vertex3f(0.526951194f, 0.556565762f, -1.f),
    C_Vertex3f(0.51724577f, 0.0754375458f, -1.f),
    C_Vertex3f(0.515873075f, 0.568760395f, -1.f),
    C_Vertex3f(0.511230469f, 0.337677956f, -1.f),
    C_Vertex3f(0.510199547f, 0.361341953f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.49576664f, 0.441459179f, -1.f),
    C_Vertex3f(0.500921249f, 0.423610687f, -1.f),
    C_Vertex3f(0.503995895f, 0.580210686f, -1.f),
    C_Vertex3f(0.505044937f, 0.404308319f, -1.f),
    C_Vertex3f(0.515873075f, 0.568760395f, -1.f),
    C_Vertex3f(0.508137703f, 0.383552074f, -1.f),
    C_Vertex3f(0.510199547f, 0.361341953f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.474117279f, 0.486281395f, -1.f),
    C_Vertex3f(0.482364655f, 0.472794533f, -1.f),
    C_Vertex3f(0.487932205f, 0.593866348f, -1.f),
    C_Vertex3f(0.489581108f, 0.457853794f, -1.f),
    C_Vertex3f(0.503995895f, 0.580210686f, -1.f),
    C_Vertex3f(0.49576664f, 0.441459179f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.466459274f, 0.607904434f, -1.f),
    C_Vertex3f(0.442964554f, 0.619375229f, -1.f),
    C_Vertex3f(0.446712971f, 0.520067215f, -1.f),
    C_Vertex3f(0.456291437f, 0.509917736f, -1.f),
    C_Vertex3f(0.465426207f, 0.498655796f, -1.f),
    C_Vertex3f(0.474117279f, 0.486281395f, -1.f),
    C_Vertex3f(0.487932205f, 0.593866348f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.403962135f, 0.54954052f, -1.f),
    C_Vertex3f(0.41531539f, 0.543840885f, -1.f),
    C_Vertex3f(0.417448044f, 0.628278732f, -1.f),
    C_Vertex3f(0.426224947f, 0.53702879f, -1.f),
    C_Vertex3f(0.442964554f, 0.619375229f, -1.f),
    C_Vertex3f(0.436690807f, 0.529104233f, -1.f),
    C_Vertex3f(0.446712971f, 0.520067215f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.297840118f, 0.564941406f, -1.f),
    C_Vertex3f(0.298933029f, 0.640625f, -1.f),
    C_Vertex3f(0.315440893f, 0.564531803f, -1.f),
    C_Vertex3f(0.316514492f, 0.640338182f, -1.f),
    C_Vertex3f(0.331744432f, 0.563712597f, -1.f),
    C_Vertex3f(0.332989216f, 0.639764547f, -1.f),
    C_Vertex3f(0.346750736f, 0.562483788f, -1.f),
    C_Vertex3f(0.348357201f, 0.638904095f, -1.f),
    C_Vertex3f(0.360459805f, 0.560845375f, -1.f),
    C_Vertex3f(0.362618446f, 0.637756824f, -1.f),
    C_Vertex3f(0.372871637f, 0.558797359f, -1.f),
    C_Vertex3f(0.375772953f, 0.636322737f, -1.f),
    C_Vertex3f(0.383986235f, 0.556339741f, -1.f),
    C_Vertex3f(0.387820721f, 0.634601831f, -1.f),
    C_Vertex3f(0.393803596f, 0.553472519f, -1.f),
    C_Vertex3f(0.398761749f, 0.632594109f, -1.f),
    C_Vertex3f(0.403962135f, 0.54954052f, -1.f),
    C_Vertex3f(0.417448044f, 0.628278732f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.153808594f, 0.0756835938f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.314613342f, 0.f, -1.f),
    C_Vertex3f(0.306411743f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.597847342f, 0.355330706f, -1.f),
    C_Vertex3f(0.596276402f, 0.3755548f, -1.f),
    C_Vertex3f(0.596501827f, 0.271151066f, -1.f),
    C_Vertex3f(0.598632813f, 0.305562973f, -1.f),
    C_Vertex3f(0.598632813f, 0.334512711f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.593919992f, 0.395184994f, -1.f),
    C_Vertex3f(0.590778112f, 0.414221287f, -1.f),
    C_Vertex3f(0.592239857f, 0.238814831f, -1.f),
    C_Vertex3f(0.596501827f, 0.271151066f, -1.f),
    C_Vertex3f(0.596276402f, 0.3755548f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.576639652f, 0.467766762f, -1.f),
    C_Vertex3f(0.57732296f, 0.180369377f, -1.f),
    C_Vertex3f(0.582137942f, 0.450512171f, -1.f),
    C_Vertex3f(0.585846901f, 0.208554268f, -1.f),
    C_Vertex3f(0.586850762f, 0.432663679f, -1.f),
    C_Vertex3f(0.592239857f, 0.238814831f, -1.f),
    C_Vertex3f(0.590778112f, 0.414221287f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.570355892f, 0.484427452f, -1.f),
    C_Vertex3f(0.563273072f, 0.5003438f, -1.f),
    C_Vertex3f(0.567596912f, 0.154614925f, -1.f),
    C_Vertex3f(0.57732296f, 0.180369377f, -1.f),
    C_Vertex3f(0.576639652f, 0.467766762f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.51724577f, 0.0754375458f, -1.f),
    C_Vertex3f(0.511230469f, 0.337677956f, -1.f),
    C_Vertex3f(0.511230469f, 0.30698967f, -1.f),
    C_Vertex3f(0.509127617f, 0.273434162f, -1.f),
    C_Vertex3f(0.504921913f, 0.242581844f, -1.f),
    C_Vertex3f(0.502657413f, 0.0609073639f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.504921913f, 0.242581844f, -1.f),
    C_Vertex3f(0.498613358f, 0.214432716f, -1.f),
    C_Vertex3f(0.502657413f, 0.0609073639f, -1.f),
    C_Vertex3f(0.49020195f, 0.188986778f, -1.f),
    C_Vertex3f(0.487441063f, 0.0483436584f, -1.f),
    C_Vertex3f(0.480124474f, 0.166134357f, -1.f),
    C_Vertex3f(0.471596718f, 0.0377464294f, -1.f),
    C_Vertex3f(0.468380928f, 0.145875454f, -1.f),
    C_Vertex3f(0.454971313f, 0.128210068f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.471596718f, 0.0377464294f, -1.f),
    C_Vertex3f(0.454971313f, 0.128210068f, -1.f),
    C_Vertex3f(0.45411253f, 0.0284328461f, -1.f),
    C_Vertex3f(0.442081451f, 0.11532402f, -1.f),
    C_Vertex3f(0.434988499f, 0.0204029083f, -1.f),
    C_Vertex3f(0.42954731f, 0.105824471f, -1.f),
    C_Vertex3f(0.415183067f, 0.0975255966f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.306411743f, 0.0756835938f, -1.f),
    C_Vertex3f(0.314613342f, 0.f, -1.f),
    C_Vertex3f(0.334675789f, 0.0771579742f, -1.f),
    C_Vertex3f(0.342015266f, 0.00136566162f, -1.f),
    C_Vertex3f(0.359526634f, 0.0801067352f, -1.f),
    C_Vertex3f(0.367751122f, 0.00409698486f, -1.f),
    C_Vertex3f(0.380964279f, 0.0845298767f, -1.f),
    C_Vertex3f(0.391820908f, 0.00819396973f, -1.f),
    C_Vertex3f(0.398988724f, 0.0904273987f, -1.f),
    C_Vertex3f(0.414224625f, 0.0136566162f, -1.f),
    C_Vertex3f(0.415183067f, 0.0975255966f, -1.f),
    C_Vertex3f(0.434988499f, 0.0204029083f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.563273072f, 0.5003438f, -1.f),
    C_Vertex3f(0.556668758f, 0.131290913f, -1.f),
    C_Vertex3f(0.567596912f, 0.154614925f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.298933029f, 0.640625f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_69[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_70[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0732421875f, 0.f, 0.f),
    C_Vertex3f(0.0732421875f, 0.f, -1.f),
    C_Vertex3f(0.0732421875f, 0.640625f, -1.f),
    C_Vertex3f(0.0732421875f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0732421875f, 0.640625f, 0.f),
    C_Vertex3f(0.0732421875f, 0.640625f, -1.f),
    C_Vertex3f(0.505859375f, 0.640625f, -1.f),
    C_Vertex3f(0.505859375f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.505859375f, 0.640625f, 0.f),
    C_Vertex3f(0.505859375f, 0.640625f, -1.f),
    C_Vertex3f(0.505859375f, 0.564941406f, -1.f),
    C_Vertex3f(0.505859375f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.505859375f, 0.564941406f, 0.f),
    C_Vertex3f(0.505859375f, 0.564941406f, -1.f),
    C_Vertex3f(0.158203125f, 0.564941406f, -1.f),
    C_Vertex3f(0.158203125f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.158203125f, 0.564941406f, 0.f),
    C_Vertex3f(0.158203125f, 0.564941406f, -1.f),
    C_Vertex3f(0.158203125f, 0.366699219f, -1.f),
    C_Vertex3f(0.158203125f, 0.366699219f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.158203125f, 0.366699219f, 0.f),
    C_Vertex3f(0.158203125f, 0.366699219f, -1.f),
    C_Vertex3f(0.458984375f, 0.366699219f, -1.f),
    C_Vertex3f(0.458984375f, 0.366699219f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.458984375f, 0.366699219f, 0.f),
    C_Vertex3f(0.458984375f, 0.366699219f, -1.f),
    C_Vertex3f(0.458984375f, 0.291015625f, -1.f),
    C_Vertex3f(0.458984375f, 0.291015625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.458984375f, 0.291015625f, 0.f),
    C_Vertex3f(0.458984375f, 0.291015625f, -1.f),
    C_Vertex3f(0.158203125f, 0.291015625f, -1.f),
    C_Vertex3f(0.158203125f, 0.291015625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.158203125f, 0.291015625f, 0.f),
    C_Vertex3f(0.158203125f, 0.291015625f, -1.f),
    C_Vertex3f(0.158203125f, 0.f, -1.f),
    C_Vertex3f(0.158203125f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.158203125f, 0.f, 0.f),
    C_Vertex3f(0.158203125f, 0.f, -1.f),
    C_Vertex3f(0.0732421875f, 0.f, -1.f),
    C_Vertex3f(0.0732421875f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158203125f, 0.564941406f, 0.f),
    C_Vertex3f(0.505859375f, 0.564941406f, 0.f),
    C_Vertex3f(0.505859375f, 0.640625f, 0.f),
    C_Vertex3f(0.0732421875f, 0.640625f, 0.f),
    C_Vertex3f(0.158203125f, 0.f, 0.f),
    C_Vertex3f(0.158203125f, 0.291015625f, 0.f),
    C_Vertex3f(0.158203125f, 0.366699219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158203125f, 0.366699219f, 0.f),
    C_Vertex3f(0.158203125f, 0.291015625f, 0.f),
    C_Vertex3f(0.458984375f, 0.291015625f, 0.f),
    C_Vertex3f(0.458984375f, 0.366699219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.158203125f, 0.f, 0.f),
    C_Vertex3f(0.0732421875f, 0.640625f, 0.f),
    C_Vertex3f(0.0732421875f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158203125f, 0.564941406f, -1.f),
    C_Vertex3f(0.505859375f, 0.564941406f, -1.f),
    C_Vertex3f(0.505859375f, 0.640625f, -1.f),
    C_Vertex3f(0.0732421875f, 0.640625f, -1.f),
    C_Vertex3f(0.158203125f, 0.f, -1.f),
    C_Vertex3f(0.158203125f, 0.291015625f, -1.f),
    C_Vertex3f(0.158203125f, 0.366699219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158203125f, 0.366699219f, -1.f),
    C_Vertex3f(0.158203125f, 0.291015625f, -1.f),
    C_Vertex3f(0.458984375f, 0.291015625f, -1.f),
    C_Vertex3f(0.458984375f, 0.366699219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.158203125f, 0.f, -1.f),
    C_Vertex3f(0.0732421875f, 0.640625f, -1.f),
    C_Vertex3f(0.0732421875f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.546875f, 0.f, 0.f),
};
static const Call char_71[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.369140625f, 0.250976563f, 0.f),
    C_Vertex3f(0.369140625f, 0.250976563f, -1.f),
    C_Vertex3f(0.369140625f, 0.326660156f, -1.f),
    C_Vertex3f(0.369140625f, 0.326660156f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.369140625f, 0.326660156f, 0.f),
    C_Vertex3f(0.369140625f, 0.326660156f, -1.f),
    C_Vertex3f(0.640136719f, 0.326660156f, -1.f),
    C_Vertex3f(0.640136719f, 0.326660156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.640136719f, 0.326660156f, 0.f),
    C_Vertex3f(0.640136719f, 0.326660156f, -1.f),
    C_Vertex3f(0.640136719f, 0.0892105103f, -1.f),
    C_Vertex3f(0.640136719f, 0.0892105103f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.623085916f, -0.782153368f, 0.f),
    C_Vertex3f(0.640136719f, 0.0892105103f, 0.f),
    C_Vertex3f(0.640136719f, 0.0892105103f, -1.f),
    C_Normal3f(0.586032808f, -0.810287297f, 0.f),
    C_Vertex3f(0.624521255f, 0.0767707825f, 0.f),
    C_Vertex3f(0.624521255f, 0.0767707825f, -1.f),
    C_Normal3f(0.535224736f, -0.844709754f, 0.f),
    C_Vertex3f(0.592799187f, 0.0549740791f, 0.f),
    C_Vertex3f(0.592799187f, 0.0549740791f, -1.f),
    C_Normal3f(0.46764341f, -0.883917272f, 0.f),
    C_Vertex3f(0.560585976f, 0.036260128f, 0.f),
    C_Vertex3f(0.560585976f, 0.036260128f, -1.f),
    C_Normal3f(0.393174857f, -0.919463754f, 0.f),
    C_Vertex3f(0.527881622f, 0.0206289291f, 0.f),
    C_Vertex3f(0.527881622f, 0.0206289291f, -1.f),
    C_Normal3f(0.312180042f, -0.950022995f, 0.f),
    C_Vertex3f(0.494686127f, 0.00808048248f, 0.f),
    C_Vertex3f(0.494686127f, 0.00808048248f, -1.f),
    C_Normal3f(0.225708812f, -0.974194825f, 0.f),
    C_Vertex3f(0.461053848f, -0.00133085251f, 0.f),
    C_Vertex3f(0.461053848f, -0.00133085251f, -1.f),
    C_Normal3f(0.135967165f, -0.990713418f, 0.f),
    C_Vertex3f(0.426984787f, -0.00760507584f, 0.f),
    C_Vertex3f(0.426984787f, -0.00760507584f, -1.f),
    C_Normal3f(0.0491345115f, -0.998792112f, 0.f),
    C_Vertex3f(0.392478943f, -0.0107421875f, 0.f),
    C_Vertex3f(0.392478943f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0241439808f, -0.999708474f, 0.f),
    C_Vertex3f(0.363214493f, -0.0107421875f, 0.f),
    C_Vertex3f(0.363214493f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0829121098f, -0.996556878f, 0.f),
    C_Vertex3f(0.340167284f, -0.00947880745f, 0.f),
    C_Vertex3f(0.340167284f, -0.00947880745f, -1.f),
    C_Normal3f(-0.14061439f, -0.990064383f, 0.f),
    C_Vertex3f(0.31765914f, -0.00695204735f, 0.f),
    C_Vertex3f(0.31765914f, -0.00695204735f, -1.f),
    C_Normal3f(-0.199671775f, -0.979862809f, 0.f),
    C_Vertex3f(0.29569006f, -0.0031619072f, 0.f),
    C_Vertex3f(0.29569006f, -0.0031619072f, -1.f),
    C_Normal3f(-0.25946945f, -0.96575129f, 0.f),
    C_Vertex3f(0.274260044f, 0.00189161301f, 0.f),
    C_Vertex3f(0.274260044f, 0.00189161301f, -1.f),
    C_Normal3f(-0.319319099f, -0.947647274f, 0.f),
    C_Vertex3f(0.253369093f, 0.00820851326f, 0.f),
    C_Vertex3f(0.253369093f, 0.00820851326f, -1.f),
    C_Normal3f(-0.37849313f, -0.925604045f, 0.f),
    C_Vertex3f(0.233017206f, 0.0157887936f, 0.f),
    C_Vertex3f(0.233017206f, 0.0157887936f, -1.f),
    C_Normal3f(-0.436266661f, -0.899817467f, 0.f),
    C_Vertex3f(0.213204384f, 0.0246324539f, 0.f),
    C_Vertex3f(0.213204384f, 0.0246324539f, -1.f),
    C_Normal3f(-0.495938122f, -0.868357837f, 0.f),
    C_Vertex3f(0.193930626f, 0.0347394943f, 0.f),
    C_Vertex3f(0.193930626f, 0.0347394943f, -1.f),
    C_Normal3f(-0.558923006f, -0.82921958f, 0.f),
    C_Vertex3f(0.175837636f, 0.0459734201f, 0.f),
    C_Vertex3f(0.175837636f, 0.0459734201f, -1.f),
    C_Normal3f(-0.620785594f, -0.783980429f, 0.f),
    C_Vertex3f(0.158925414f, 0.0583342314f, 0.f),
    C_Vertex3f(0.158925414f, 0.0583342314f, -1.f),
    C_Normal3f(-0.680231988f, -0.732996821f, 0.f),
    C_Vertex3f(0.14319396f, 0.071821928f, 0.f),
    C_Vertex3f(0.14319396f, 0.071821928f, -1.f),
    C_Normal3f(-0.736015141f, -0.676965177f, 0.f),
    C_Vertex3f(0.128643274f, 0.0864365101f, 0.f),
    C_Vertex3f(0.128643274f, 0.0864365101f, -1.f),
    C_Normal3f(-0.787055671f, -0.616881967f, 0.f),
    C_Vertex3f(0.115273356f, 0.102177978f, 0.f),
    C_Vertex3f(0.115273356f, 0.102177978f, -1.f),
    C_Normal3f(-0.832544446f, -0.553958297f, 0.f),
    C_Vertex3f(0.103084207f, 0.11904633f, 0.f),
    C_Vertex3f(0.103084207f, 0.11904633f, -1.f),
    C_Normal3f(-0.872002065f, -0.489502221f, 0.f),
    C_Vertex3f(0.0920758247f, 0.137041569f, 0.f),
    C_Vertex3f(0.0920758247f, 0.137041569f, -1.f),
    C_Normal3f(-0.903324485f, -0.428957909f, 0.f),
    C_Vertex3f(0.0822482109f, 0.156163692f, 0.f),
    C_Vertex3f(0.0822482109f, 0.156163692f, -1.f),
    C_Normal3f(-0.928474605f, -0.371395946f, 0.f),
    C_Vertex3f(0.0736490488f, 0.175845742f, 0.f),
    C_Vertex3f(0.0736490488f, 0.175845742f, -1.f),
    C_Normal3f(-0.949845433f, -0.312719762f, 0.f),
    C_Vertex3f(0.0662783384f, 0.196087718f, 0.f),
    C_Vertex3f(0.0662783384f, 0.196087718f, -1.f),
    C_Normal3f(-0.967297971f, -0.25364241f, 0.f),
    C_Vertex3f(0.0601360798f, 0.21688962f, 0.f),
    C_Vertex3f(0.0601360798f, 0.21688962f, -1.f),
    C_Normal3f(-0.980830729f, -0.194861934f, 0.f),
    C_Vertex3f(0.0552222729f, 0.238251448f, 0.f),
    C_Vertex3f(0.0552222729f, 0.238251448f, -1.f),
    C_Normal3f(-0.990567744f, -0.137023717f, 0.f),
    C_Vertex3f(0.0515369177f, 0.260173202f, 0.f),
    C_Vertex3f(0.0515369177f, 0.260173202f, -1.f),
    C_Normal3f(-0.996739149f, -0.0806914046f, 0.f),
    C_Vertex3f(0.0490800142f, 0.282654881f, 0.f),
    C_Vertex3f(0.0490800142f, 0.282654881f, -1.f),
    C_Normal3f(-0.99965167f, -0.0263895821f, 0.f),
    C_Vertex3f(0.0478515625f, 0.305696487f, 0.f),
    C_Vertex3f(0.0478515625f, 0.305696487f, -1.f),
    C_Normal3f(-0.999654889f, 0.0262702107f, 0.f),
    C_Vertex3f(0.0478515625f, 0.329189301f, 0.f),
    C_Vertex3f(0.0478515625f, 0.329189301f, -1.f),
    C_Normal3f(-0.9967857f, 0.0801141337f, 0.f),
    C_Vertex3f(0.0490732193f, 0.352183938f, 0.f),
    C_Vertex3f(0.0490732193f, 0.352183938f, -1.f),
    C_Normal3f(-0.990840971f, 0.135033816f, 0.f),
    C_Vertex3f(0.0515165329f, 0.374789119f, 0.f),
    C_Vertex3f(0.0515165329f, 0.374789119f, -1.f),
    C_Normal3f(-0.981665969f, 0.190609276f, 0.f),
    C_Vertex3f(0.0551815033f, 0.397004843f, 0.f),
    C_Vertex3f(0.0551815033f, 0.397004843f, -1.f),
    C_Normal3f(-0.969190896f, 0.246310905f, 0.f),
    C_Vertex3f(0.0600681305f, 0.41883111f, 0.f),
    C_Vertex3f(0.0600681305f, 0.41883111f, -1.f),
    C_Normal3f(-0.953439713f, 0.301583827f, 0.f),
    C_Vertex3f(0.0661764145f, 0.44026792f, 0.f),
    C_Vertex3f(0.0661764145f, 0.44026792f, -1.f),
    C_Normal3f(-0.934533775f, 0.355874479f, 0.f),
    C_Vertex3f(0.0735063553f, 0.461315274f, 0.f),
    C_Vertex3f(0.0735063553f, 0.461315274f, -1.f),
    C_Normal3f(-0.912687659f, 0.408657789f, 0.f),
    C_Vertex3f(0.0820579529f, 0.481973171f, 0.f),
    C_Vertex3f(0.0820579529f, 0.481973171f, -1.f),
    C_Normal3f(-0.885204375f, 0.46520251f, 0.f),
    C_Vertex3f(0.0918312073f, 0.502241611f, 0.f),
    C_Vertex3f(0.0918312073f, 0.502241611f, -1.f),
    C_Normal3f(-0.849341929f, 0.527842939f, 0.f),
    C_Vertex3f(0.102675915f, 0.521205783f, 0.f),
    C_Vertex3f(0.102675915f, 0.521205783f, -1.f),
    C_Normal3f(-0.806868017f, 0.5907318f, 0.f),
    C_Vertex3f(0.114592075f, 0.538865685f, 0.f),
    C_Vertex3f(0.114592075f, 0.538865685f, -1.f),
    C_Normal3f(-0.757747769f, 0.652547598f, 0.f),
    C_Vertex3f(0.127579689f, 0.555221319f, 0.f),
    C_Vertex3f(0.127579689f, 0.555221319f, -1.f),
    C_Normal3f(-0.70232141f, 0.711859941f, 0.f),
    C_Vertex3f(0.141638756f, 0.570272684f, 0.f),
    C_Vertex3f(0.141638756f, 0.570272684f, -1.f),
    C_Normal3f(-0.641332746f, 0.767262816f, 0.f),
    C_Vertex3f(0.156769276f, 0.58401978f, 0.f),
    C_Vertex3f(0.156769276f, 0.58401978f, -1.f),
    C_Normal3f(-0.575895548f, 0.817523301f, 0.f),
    C_Vertex3f(0.172971249f, 0.596462607f, 0.f),
    C_Vertex3f(0.172971249f, 0.596462607f, -1.f),
    C_Normal3f(-0.507395744f, 0.861713052f, 0.f),
    C_Vertex3f(0.190244675f, 0.607601166f, 0.f),
    C_Vertex3f(0.190244675f, 0.607601166f, -1.f),
    C_Normal3f(-0.442338616f, 0.896848142f, 0.f),
    C_Vertex3f(0.208589554f, 0.617435455f, 0.f),
    C_Vertex3f(0.208589554f, 0.617435455f, -1.f),
    C_Normal3f(-0.381018549f, 0.924567401f, 0.f),
    C_Vertex3f(0.227630615f, 0.626040459f, 0.f),
    C_Vertex3f(0.227630615f, 0.626040459f, -1.f),
    C_Normal3f(-0.319027632f, 0.947745383f, 0.f),
    C_Vertex3f(0.247367859f, 0.633416176f, 0.f),
    C_Vertex3f(0.247367859f, 0.633416176f, -1.f),
    C_Normal3f(-0.257230639f, 0.966349959f, 0.f),
    C_Vertex3f(0.267801285f, 0.639562607f, 0.f),
    C_Vertex3f(0.267801285f, 0.639562607f, -1.f),
    C_Normal3f(-0.196421713f, 0.980519414f, 0.f),
    C_Vertex3f(0.288930893f, 0.644479752f, 0.f),
    C_Vertex3f(0.288930893f, 0.644479752f, -1.f),
    C_Normal3f(-0.137283921f, 0.990531683f, 0.f),
    C_Vertex3f(0.310756683f, 0.64816761f, 0.f),
    C_Vertex3f(0.310756683f, 0.64816761f, -1.f),
    C_Normal3f(-0.080365546f, 0.996765554f, 0.f),
    C_Vertex3f(0.333278656f, 0.650626183f, 0.f),
    C_Vertex3f(0.333278656f, 0.650626183f, -1.f),
    C_Normal3f(-0.0233916733f, 0.999726355f, 0.f),
    C_Vertex3f(0.356496811f, 0.651855469f, 0.f),
    C_Vertex3f(0.356496811f, 0.651855469f, -1.f),
    C_Normal3f(0.0451385081f, 0.998980641f, 0.f),
    C_Vertex3f(0.385816574f, 0.651855469f, 0.f),
    C_Vertex3f(0.385816574f, 0.651855469f, -1.f),
    C_Normal3f(0.130076706f, 0.991503894f, 0.f),
    C_Vertex3f(0.418876171f, 0.649036884f, 0.f),
    C_Vertex3f(0.418876171f, 0.649036884f, -1.f),
    C_Normal3f(0.224674076f, 0.974434018f, 0.f),
    C_Vertex3f(0.450270176f, 0.643399715f, 0.f),
    C_Vertex3f(0.450270176f, 0.643399715f, -1.f),
    C_Normal3f(0.32309252f, 0.946367383f, 0.f),
    C_Vertex3f(0.479998589f, 0.634943962f, 0.f),
    C_Vertex3f(0.479998589f, 0.634943962f, -1.f),
    C_Normal3f(0.422699183f, 0.906270027f, 0.f),
    C_Vertex3f(0.508061409f, 0.623669624f, 0.f),
    C_Vertex3f(0.508061409f, 0.623669624f, -1.f),
    C_Normal3f(0.527392685f, 0.849621713f, 0.f),
    C_Vertex3f(0.533093929f, 0.610179424f, 0.f),
    C_Vertex3f(0.533093929f, 0.610179424f, -1.f),
    C_Normal3f(0.634408832f, 0.772997677f, 0.f),
    C_Vertex3f(0.555096149f, 0.594473362f, 0.f),
    C_Vertex3f(0.555096149f, 0.594473362f, -1.f),
    C_Normal3f(0.736908436f, 0.675992548f, 0.f),
    C_Vertex3f(0.574068069f, 0.576551437f, 0.f),
    C_Vertex3f(0.574068069f, 0.576551437f, -1.f),
    C_Normal3f(0.822476745f, 0.568798602f, 0.f),
    C_Vertex3f(0.590009689f, 0.556413651f, 0.f),
    C_Vertex3f(0.590009689f, 0.556413651f, -1.f),
    C_Normal3f(0.884067595f, 0.467359096f, 0.f),
    C_Vertex3f(0.6040411f, 0.533210754f, 0.f),
    C_Vertex3f(0.6040411f, 0.533210754f, -1.f),
    C_Normal3f(0.927947819f, 0.372710288f, 0.f),
    C_Vertex3f(0.6161623f, 0.506942749f, 0.f),
    C_Vertex3f(0.6161623f, 0.506942749f, -1.f),
    C_Normal3f(0.953686178f, 0.300803453f, 0.f),
    C_Vertex3f(0.626373291f, 0.477609634f, 0.f),
    C_Vertex3f(0.626373291f, 0.477609634f, -1.f),
    C_Normal3f(0.968710303f, 0.248194233f, 0.f),
    C_Vertex3f(0.630523682f, 0.461410522f, 0.f),
    C_Vertex3f(0.630523682f, 0.461410522f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.264574856f, -0.964365184f, 0.f),
    C_Vertex3f(0.630523682f, 0.461410522f, 0.f),
    C_Vertex3f(0.630523682f, 0.461410522f, -1.f),
    C_Vertex3f(0.554077148f, 0.440437317f, -1.f),
    C_Vertex3f(0.554077148f, 0.440437317f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.959322631f, -0.282311976f, 0.f),
    C_Vertex3f(0.554077148f, 0.440437317f, 0.f),
    C_Vertex3f(0.554077148f, 0.440437317f, -1.f),
    C_Normal3f(-0.946045637f, -0.324033529f, 0.f),
    C_Vertex3f(0.550474167f, 0.452680588f, 0.f),
    C_Vertex3f(0.550474167f, 0.452680588f, -1.f),
    C_Normal3f(-0.923839152f, -0.38278082f, 0.f),
    C_Vertex3f(0.542394638f, 0.474545479f, 0.f),
    C_Vertex3f(0.542394638f, 0.474545479f, -1.f),
    C_Normal3f(-0.885898411f, -0.463879377f, 0.f),
    C_Vertex3f(0.533441544f, 0.493788719f, 0.f),
    C_Vertex3f(0.533441544f, 0.493788719f, -1.f),
    C_Normal3f(-0.830638766f, -0.55681175f, 0.f),
    C_Vertex3f(0.523614883f, 0.510410309f, 0.f),
    C_Vertex3f(0.523614883f, 0.510410309f, -1.f),
    C_Normal3f(-0.746800542f, -0.665048122f, 0.f),
    C_Vertex3f(0.512914658f, 0.524410248f, 0.f),
    C_Vertex3f(0.512914658f, 0.524410248f, -1.f),
    C_Normal3f(-0.644194484f, -0.764861703f, 0.f),
    C_Vertex3f(0.499921799f, 0.537015915f, 0.f),
    C_Vertex3f(0.499921799f, 0.537015915f, -1.f),
    C_Normal3f(-0.538977444f, -0.842320323f, 0.f),
    C_Vertex3f(0.484636307f, 0.54822731f, 0.f),
    C_Vertex3f(0.484636307f, 0.54822731f, -1.f),
    C_Normal3f(-0.437884986f, -0.899031043f, 0.f),
    C_Vertex3f(0.467058182f, 0.558044434f, 0.f),
    C_Vertex3f(0.467058182f, 0.558044434f, -1.f),
    C_Normal3f(-0.339533776f, -0.940593839f, 0.f),
    C_Vertex3f(0.447187424f, 0.566467285f, 0.f),
    C_Vertex3f(0.447187424f, 0.566467285f, -1.f),
    C_Normal3f(-0.237734407f, -0.971330225f, 0.f),
    C_Vertex3f(0.426224709f, 0.572784424f, 0.f),
    C_Vertex3f(0.426224709f, 0.572784424f, -1.f),
    C_Normal3f(-0.138410524f, -0.990374923f, 0.f),
    C_Vertex3f(0.404170036f, 0.57699585f, 0.f),
    C_Vertex3f(0.404170036f, 0.57699585f, -1.f),
    C_Normal3f(-0.0422570221f, -0.999106824f, 0.f),
    C_Vertex3f(0.381023407f, 0.579101563f, 0.f),
    C_Vertex3f(0.381023407f, 0.579101563f, -1.f),
    C_Normal3f(0.0411277004f, -0.999153912f, 0.f),
    C_Vertex3f(0.354383469f, 0.579101563f, 0.f),
    C_Vertex3f(0.354383469f, 0.579101563f, -1.f),
    C_Normal3f(0.126090482f, -0.9920187f, 0.f),
    C_Vertex3f(0.327307224f, 0.576890469f, 0.f),
    C_Vertex3f(0.327307224f, 0.576890469f, -1.f),
    C_Normal3f(0.223308846f, -0.974747717f, 0.f),
    C_Vertex3f(0.302196026f, 0.572468281f, 0.f),
    C_Vertex3f(0.302196026f, 0.572468281f, -1.f),
    C_Normal3f(0.329650491f, -0.944103003f, 0.f),
    C_Vertex3f(0.279049873f, 0.565834999f, 0.f),
    C_Vertex3f(0.279049873f, 0.565834999f, -1.f),
    C_Normal3f(0.427793115f, -0.903876662f, 0.f),
    C_Vertex3f(0.257868767f, 0.556990623f, 0.f),
    C_Vertex3f(0.257868767f, 0.556990623f, -1.f),
    C_Normal3f(0.516784489f, -0.85611558f, 0.f),
    C_Vertex3f(0.238734722f, 0.54675436f, 0.f),
    C_Vertex3f(0.238734722f, 0.54675436f, -1.f),
    C_Normal3f(0.608704686f, -0.793396831f, 0.f),
    C_Vertex3f(0.221647739f, 0.535126209f, 0.f),
    C_Vertex3f(0.221647739f, 0.535126209f, -1.f),
    C_Normal3f(0.699406087f, -0.714724481f, 0.f),
    C_Vertex3f(0.206607819f, 0.522106171f, 0.f),
    C_Vertex3f(0.206607819f, 0.522106171f, -1.f),
    C_Normal3f(0.769154906f, -0.639062464f, 0.f),
    C_Vertex3f(0.19361496f, 0.507694244f, 0.f),
    C_Vertex3f(0.19361496f, 0.507694244f, -1.f),
    C_Normal3f(0.819520056f, -0.573050499f, 0.f),
    C_Vertex3f(0.182068825f, 0.492571831f, 0.f),
    C_Vertex3f(0.182068825f, 0.492571831f, -1.f),
    C_Normal3f(0.865334332f, -0.501195014f, 0.f),
    C_Vertex3f(0.171969414f, 0.47673893f, 0.f),
    C_Vertex3f(0.171969414f, 0.47673893f, -1.f),
    C_Normal3f(0.90881896f, -0.41719085f, 0.f),
    C_Vertex3f(0.163316727f, 0.460195541f, 0.f),
    C_Vertex3f(0.163316727f, 0.460195541f, -1.f),
    C_Normal3f(0.944719911f, -0.327878535f, 0.f),
    C_Vertex3f(0.153598785f, 0.436719894f, 0.f),
    C_Vertex3f(0.153598785f, 0.436719894f, -1.f),
    C_Normal3f(0.971937418f, -0.235239655f, 0.f),
    C_Vertex3f(0.144426346f, 0.405766487f, 0.f),
    C_Vertex3f(0.144426346f, 0.405766487f, -1.f),
    C_Normal3f(0.990387321f, -0.138322011f, 0.f),
    C_Vertex3f(0.138311386f, 0.373557091f, 0.f),
    C_Vertex3f(0.138311386f, 0.373557091f, -1.f),
    C_Normal3f(0.998767555f, -0.0496318527f, 0.f),
    C_Vertex3f(0.135253906f, 0.340091705f, 0.f),
    C_Vertex3f(0.135253906f, 0.340091705f, -1.f),
    C_Normal3f(0.999820292f, 0.0189560317f, 0.f),
    C_Vertex3f(0.135253906f, 0.312029839f, 0.f),
    C_Vertex3f(0.135253906f, 0.312029839f, -1.f),
    C_Normal3f(0.997643113f, 0.0686158016f, 0.f),
    C_Vertex3f(0.136175156f, 0.291501164f, 0.f),
    C_Vertex3f(0.136175156f, 0.291501164f, -1.f),
    C_Normal3f(0.992895544f, 0.118989676f, 0.f),
    C_Vertex3f(0.138017654f, 0.271846175f, 0.f),
    C_Vertex3f(0.138017654f, 0.271846175f, -1.f),
    C_Normal3f(0.984901726f, 0.173114404f, 0.f),
    C_Vertex3f(0.140781403f, 0.253064871f, 0.f),
    C_Vertex3f(0.140781403f, 0.253064871f, -1.f),
    C_Normal3f(0.972982645f, 0.230878025f, 0.f),
    C_Vertex3f(0.1444664f, 0.235157251f, 0.f),
    C_Vertex3f(0.1444664f, 0.235157251f, -1.f),
    C_Normal3f(0.956423461f, 0.291983366f, 0.f),
    C_Vertex3f(0.149072647f, 0.218123317f, 0.f),
    C_Vertex3f(0.149072647f, 0.218123317f, -1.f),
    C_Normal3f(0.934522271f, 0.355904579f, 0.f),
    C_Vertex3f(0.154600143f, 0.201963067f, 0.f),
    C_Vertex3f(0.154600143f, 0.201963067f, -1.f),
    C_Normal3f(0.906662226f, 0.421857387f, 0.f),
    C_Vertex3f(0.161048889f, 0.186676502f, 0.f),
    C_Vertex3f(0.161048889f, 0.186676502f, -1.f),
    C_Normal3f(0.873084843f, 0.487568319f, 0.f),
    C_Vertex3f(0.168418884f, 0.172263622f, 0.f),
    C_Vertex3f(0.168418884f, 0.172263622f, -1.f),
    C_Normal3f(0.833571136f, 0.552412152f, 0.f),
    C_Vertex3f(0.176628113f, 0.158778906f, 0.f),
    C_Vertex3f(0.176628113f, 0.158778906f, -1.f),
    C_Normal3f(0.787365735f, 0.616486132f, 0.f),
    C_Vertex3f(0.185676575f, 0.146222353f, 0.f),
    C_Vertex3f(0.185676575f, 0.146222353f, -1.f),
    C_Normal3f(0.734744549f, 0.678344011f, 0.f),
    C_Vertex3f(0.19556427f, 0.134593964f, 0.f),
    C_Vertex3f(0.19556427f, 0.134593964f, -1.f),
    C_Normal3f(0.676387787f, 0.736545622f, 0.f),
    C_Vertex3f(0.206291199f, 0.123893738f, 0.f),
    C_Vertex3f(0.206291199f, 0.123893738f, -1.f),
    C_Normal3f(0.61335361f, 0.789808333f, 0.f),
    C_Vertex3f(0.217857361f, 0.114121675f, 0.f),
    C_Vertex3f(0.217857361f, 0.114121675f, -1.f),
    C_Normal3f(0.546985626f, 0.83714205f, 0.f),
    C_Vertex3f(0.230262756f, 0.105277777f, 0.f),
    C_Vertex3f(0.230262756f, 0.105277777f, -1.f),
    C_Normal3f(0.471909761f, 0.881646812f, 0.f),
    C_Vertex3f(0.243507385f, 0.0973620415f, 0.f),
    C_Vertex3f(0.243507385f, 0.0973620415f, -1.f),
    C_Normal3f(0.385551751f, 0.9226861f, 0.f),
    C_Vertex3f(0.264633179f, 0.0868806839f, 0.f),
    C_Vertex3f(0.264633179f, 0.0868806839f, -1.f),
    C_Normal3f(0.28409493f, 0.958796203f, 0.f),
    C_Vertex3f(0.293674469f, 0.0763993263f, 0.f),
    C_Vertex3f(0.293674469f, 0.0763993263f, -1.f),
    C_Normal3f(0.170147717f, 0.985418499f, 0.f),
    C_Vertex3f(0.323589325f, 0.0694117546f, 0.f),
    C_Vertex3f(0.323589325f, 0.0694117546f, -1.f),
    C_Normal3f(0.057769414f, 0.998329937f, 0.f),
    C_Vertex3f(0.354377747f, 0.0659179688f, 0.f),
    C_Vertex3f(0.354377747f, 0.0659179688f, -1.f),
    C_Normal3f(-0.046444267f, 0.998920918f, 0.f),
    C_Vertex3f(0.383966446f, 0.0659179688f, 0.f),
    C_Vertex3f(0.383966446f, 0.0659179688f, -1.f),
    C_Normal3f(-0.144979268f, 0.9894346f, 0.f),
    C_Vertex3f(0.411153793f, 0.0685577393f, 0.f),
    C_Vertex3f(0.411153793f, 0.0685577393f, -1.f),
    C_Normal3f(-0.239988849f, 0.970775723f, 0.f),
    C_Vertex3f(0.438013077f, 0.0738372803f, 0.f),
    C_Vertex3f(0.438013077f, 0.0738372803f, -1.f),
    C_Normal3f(-0.33069098f, 0.943739116f, 0.f),
    C_Vertex3f(0.464544296f, 0.0817565918f, 0.f),
    C_Vertex3f(0.464544296f, 0.0817565918f, -1.f),
    C_Normal3f(-0.399715871f, 0.916639149f, 0.f),
    C_Vertex3f(0.490747452f, 0.0923156738f, 0.f),
    C_Vertex3f(0.490747452f, 0.0923156738f, -1.f),
    C_Normal3f(-0.459193528f, 0.888336241f, 0.f),
    C_Vertex3f(0.513784409f, 0.103228569f, 0.f),
    C_Vertex3f(0.513784409f, 0.103228569f, -1.f),
    C_Normal3f(-0.530459821f, 0.847710133f, 0.f),
    C_Vertex3f(0.533655167f, 0.114495277f, 0.f),
    C_Vertex3f(0.533655167f, 0.114495277f, -1.f),
    C_Normal3f(-0.600051999f, 0.79996103f, 0.f),
    C_Vertex3f(0.550359726f, 0.126115799f, 0.f),
    C_Vertex3f(0.550359726f, 0.126115799f, -1.f),
    C_Normal3f(-0.662514627f, 0.749048889f, 0.f),
    C_Vertex3f(0.557128906f, 0.132102966f, 0.f),
    C_Vertex3f(0.557128906f, 0.132102966f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.557128906f, 0.132102966f, 0.f),
    C_Vertex3f(0.557128906f, 0.132102966f, -1.f),
    C_Vertex3f(0.557128906f, 0.250976563f, -1.f),
    C_Vertex3f(0.557128906f, 0.250976563f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.557128906f, 0.250976563f, 0.f),
    C_Vertex3f(0.557128906f, 0.250976563f, -1.f),
    C_Vertex3f(0.369140625f, 0.250976563f, -1.f),
    C_Vertex3f(0.369140625f, 0.250976563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.630523682f, 0.461410522f, 0.f),
    C_Vertex3f(0.626373291f, 0.477609634f, 0.f),
    C_Vertex3f(0.6161623f, 0.506942749f, 0.f),
    C_Vertex3f(0.6040411f, 0.533210754f, 0.f),
    C_Vertex3f(0.590009689f, 0.556413651f, 0.f),
    C_Vertex3f(0.574068069f, 0.576551437f, 0.f),
    C_Vertex3f(0.555096149f, 0.594473362f, 0.f),
    C_Vertex3f(0.554077148f, 0.440437317f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.555096149f, 0.594473362f, 0.f),
    C_Vertex3f(0.533093929f, 0.610179424f, 0.f),
    C_Vertex3f(0.533441544f, 0.493788719f, 0.f),
    C_Vertex3f(0.542394638f, 0.474545479f, 0.f),
    C_Vertex3f(0.550474167f, 0.452680588f, 0.f),
    C_Vertex3f(0.554077148f, 0.440437317f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.484636307f, 0.54822731f, 0.f),
    C_Vertex3f(0.499921799f, 0.537015915f, 0.f),
    C_Vertex3f(0.508061409f, 0.623669624f, 0.f),
    C_Vertex3f(0.512914658f, 0.524410248f, 0.f),
    C_Vertex3f(0.533093929f, 0.610179424f, 0.f),
    C_Vertex3f(0.523614883f, 0.510410309f, 0.f),
    C_Vertex3f(0.533441544f, 0.493788719f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426224709f, 0.572784424f, 0.f),
    C_Vertex3f(0.447187424f, 0.566467285f, 0.f),
    C_Vertex3f(0.450270176f, 0.643399715f, 0.f),
    C_Vertex3f(0.467058182f, 0.558044434f, 0.f),
    C_Vertex3f(0.479998589f, 0.634943962f, 0.f),
    C_Vertex3f(0.484636307f, 0.54822731f, 0.f),
    C_Vertex3f(0.508061409f, 0.623669624f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.19361496f, 0.507694244f, 0.f),
    C_Vertex3f(0.206607819f, 0.522106171f, 0.f),
    C_Vertex3f(0.208589554f, 0.617435455f, 0.f),
    C_Vertex3f(0.221647739f, 0.535126209f, 0.f),
    C_Vertex3f(0.227630615f, 0.626040459f, 0.f),
    C_Vertex3f(0.238734722f, 0.54675436f, 0.f),
    C_Vertex3f(0.247367859f, 0.633416176f, 0.f),
    C_Vertex3f(0.257868767f, 0.556990623f, 0.f),
    C_Vertex3f(0.267801285f, 0.639562607f, 0.f),
    C_Vertex3f(0.279049873f, 0.565834999f, 0.f),
    C_Vertex3f(0.288930893f, 0.644479752f, 0.f),
    C_Vertex3f(0.302196026f, 0.572468281f, 0.f),
    C_Vertex3f(0.310756683f, 0.64816761f, 0.f),
    C_Vertex3f(0.327307224f, 0.576890469f, 0.f),
    C_Vertex3f(0.333278656f, 0.650626183f, 0.f),
    C_Vertex3f(0.354383469f, 0.579101563f, 0.f),
    C_Vertex3f(0.356496811f, 0.651855469f, 0.f),
    C_Vertex3f(0.381023407f, 0.579101563f, 0.f),
    C_Vertex3f(0.385816574f, 0.651855469f, 0.f),
    C_Vertex3f(0.404170036f, 0.57699585f, 0.f),
    C_Vertex3f(0.418876171f, 0.649036884f, 0.f),
    C_Vertex3f(0.426224709f, 0.572784424f, 0.f),
    C_Vertex3f(0.450270176f, 0.643399715f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163316727f, 0.460195541f, 0.f),
    C_Vertex3f(0.171969414f, 0.47673893f, 0.f),
    C_Vertex3f(0.172971249f, 0.596462607f, 0.f),
    C_Vertex3f(0.182068825f, 0.492571831f, 0.f),
    C_Vertex3f(0.190244675f, 0.607601166f, 0.f),
    C_Vertex3f(0.19361496f, 0.507694244f, 0.f),
    C_Vertex3f(0.208589554f, 0.617435455f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156769276f, 0.58401978f, 0.f),
    C_Vertex3f(0.141638756f, 0.570272684f, 0.f),
    C_Vertex3f(0.144426346f, 0.405766487f, 0.f),
    C_Vertex3f(0.153598785f, 0.436719894f, 0.f),
    C_Vertex3f(0.163316727f, 0.460195541f, 0.f),
    C_Vertex3f(0.172971249f, 0.596462607f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.115273356f, 0.102177978f, 0.f),
    C_Vertex3f(0.128643274f, 0.0864365101f, 0.f),
    C_Vertex3f(0.127579689f, 0.555221319f, 0.f),
    C_Vertex3f(0.135253906f, 0.340091705f, 0.f),
    C_Vertex3f(0.141638756f, 0.570272684f, 0.f),
    C_Vertex3f(0.138311386f, 0.373557091f, 0.f),
    C_Vertex3f(0.144426346f, 0.405766487f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.138017654f, 0.271846175f, 0.f),
    C_Vertex3f(0.136175156f, 0.291501164f, 0.f),
    C_Vertex3f(0.14319396f, 0.071821928f, 0.f),
    C_Vertex3f(0.135253906f, 0.312029839f, 0.f),
    C_Vertex3f(0.128643274f, 0.0864365101f, 0.f),
    C_Vertex3f(0.135253906f, 0.340091705f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0490732193f, 0.352183938f, 0.f),
    C_Vertex3f(0.0490800142f, 0.282654881f, 0.f),
    C_Vertex3f(0.0515165329f, 0.374789119f, 0.f),
    C_Vertex3f(0.0515369177f, 0.260173202f, 0.f),
    C_Vertex3f(0.0551815033f, 0.397004843f, 0.f),
    C_Vertex3f(0.0552222729f, 0.238251448f, 0.f),
    C_Vertex3f(0.0600681305f, 0.41883111f, 0.f),
    C_Vertex3f(0.0601360798f, 0.21688962f, 0.f),
    C_Vertex3f(0.0661764145f, 0.44026792f, 0.f),
    C_Vertex3f(0.0662783384f, 0.196087718f, 0.f),
    C_Vertex3f(0.0735063553f, 0.461315274f, 0.f),
    C_Vertex3f(0.0736490488f, 0.175845742f, 0.f),
    C_Vertex3f(0.0820579529f, 0.481973171f, 0.f),
    C_Vertex3f(0.0822482109f, 0.156163692f, 0.f),
    C_Vertex3f(0.0918312073f, 0.502241611f, 0.f),
    C_Vertex3f(0.0920758247f, 0.137041569f, 0.f),
    C_Vertex3f(0.102675915f, 0.521205783f, 0.f),
    C_Vertex3f(0.103084207f, 0.11904633f, 0.f),
    C_Vertex3f(0.114592075f, 0.538865685f, 0.f),
    C_Vertex3f(0.115273356f, 0.102177978f, 0.f),
    C_Vertex3f(0.127579689f, 0.555221319f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0478515625f, 0.329189301f, 0.f),
    C_Vertex3f(0.0478515625f, 0.305696487f, 0.f),
    C_Vertex3f(0.0490800142f, 0.282654881f, 0.f),
    C_Vertex3f(0.0490732193f, 0.352183938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.640136719f, 0.326660156f, 0.f),
    C_Vertex3f(0.369140625f, 0.326660156f, 0.f),
    C_Vertex3f(0.557128906f, 0.250976563f, 0.f),
    C_Vertex3f(0.560585976f, 0.036260128f, 0.f),
    C_Vertex3f(0.592799187f, 0.0549740791f, 0.f),
    C_Vertex3f(0.624521255f, 0.0767707825f, 0.f),
    C_Vertex3f(0.640136719f, 0.0892105103f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.560585976f, 0.036260128f, 0.f),
    C_Vertex3f(0.557128906f, 0.250976563f, 0.f),
    C_Vertex3f(0.557128906f, 0.132102966f, 0.f),
    C_Vertex3f(0.550359726f, 0.126115799f, 0.f),
    C_Vertex3f(0.533655167f, 0.114495277f, 0.f),
    C_Vertex3f(0.527881622f, 0.0206289291f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.513784409f, 0.103228569f, 0.f),
    C_Vertex3f(0.490747452f, 0.0923156738f, 0.f),
    C_Vertex3f(0.494686127f, 0.00808048248f, 0.f),
    C_Vertex3f(0.527881622f, 0.0206289291f, 0.f),
    C_Vertex3f(0.533655167f, 0.114495277f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.29569006f, -0.0031619072f, 0.f),
    C_Vertex3f(0.31765914f, -0.00695204735f, 0.f),
    C_Vertex3f(0.323589325f, 0.0694117546f, 0.f),
    C_Vertex3f(0.340167284f, -0.00947880745f, 0.f),
    C_Vertex3f(0.354377747f, 0.0659179688f, 0.f),
    C_Vertex3f(0.363214493f, -0.0107421875f, 0.f),
    C_Vertex3f(0.383966446f, 0.0659179688f, 0.f),
    C_Vertex3f(0.392478943f, -0.0107421875f, 0.f),
    C_Vertex3f(0.411153793f, 0.0685577393f, 0.f),
    C_Vertex3f(0.426984787f, -0.00760507584f, 0.f),
    C_Vertex3f(0.438013077f, 0.0738372803f, 0.f),
    C_Vertex3f(0.461053848f, -0.00133085251f, 0.f),
    C_Vertex3f(0.464544296f, 0.0817565918f, 0.f),
    C_Vertex3f(0.494686127f, 0.00808048248f, 0.f),
    C_Vertex3f(0.490747452f, 0.0923156738f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230262756f, 0.105277777f, 0.f),
    C_Vertex3f(0.233017206f, 0.0157887936f, 0.f),
    C_Vertex3f(0.243507385f, 0.0973620415f, 0.f),
    C_Vertex3f(0.253369093f, 0.00820851326f, 0.f),
    C_Vertex3f(0.264633179f, 0.0868806839f, 0.f),
    C_Vertex3f(0.274260044f, 0.00189161301f, 0.f),
    C_Vertex3f(0.293674469f, 0.0763993263f, 0.f),
    C_Vertex3f(0.29569006f, -0.0031619072f, 0.f),
    C_Vertex3f(0.323589325f, 0.0694117546f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.217857361f, 0.114121675f, 0.f),
    C_Vertex3f(0.206291199f, 0.123893738f, 0.f),
    C_Vertex3f(0.213204384f, 0.0246324539f, 0.f),
    C_Vertex3f(0.233017206f, 0.0157887936f, 0.f),
    C_Vertex3f(0.230262756f, 0.105277777f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.19556427f, 0.134593964f, 0.f),
    C_Vertex3f(0.185676575f, 0.146222353f, 0.f),
    C_Vertex3f(0.193930626f, 0.0347394943f, 0.f),
    C_Vertex3f(0.213204384f, 0.0246324539f, 0.f),
    C_Vertex3f(0.206291199f, 0.123893738f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.176628113f, 0.158778906f, 0.f),
    C_Vertex3f(0.168418884f, 0.172263622f, 0.f),
    C_Vertex3f(0.175837636f, 0.0459734201f, 0.f),
    C_Vertex3f(0.193930626f, 0.0347394943f, 0.f),
    C_Vertex3f(0.185676575f, 0.146222353f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161048889f, 0.186676502f, 0.f),
    C_Vertex3f(0.154600143f, 0.201963067f, 0.f),
    C_Vertex3f(0.158925414f, 0.0583342314f, 0.f),
    C_Vertex3f(0.175837636f, 0.0459734201f, 0.f),
    C_Vertex3f(0.168418884f, 0.172263622f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154600143f, 0.201963067f, 0.f),
    C_Vertex3f(0.149072647f, 0.218123317f, 0.f),
    C_Vertex3f(0.158925414f, 0.0583342314f, 0.f),
    C_Vertex3f(0.1444664f, 0.235157251f, 0.f),
    C_Vertex3f(0.14319396f, 0.071821928f, 0.f),
    C_Vertex3f(0.140781403f, 0.253064871f, 0.f),
    C_Vertex3f(0.138017654f, 0.271846175f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.557128906f, 0.250976563f, 0.f),
    C_Vertex3f(0.369140625f, 0.326660156f, 0.f),
    C_Vertex3f(0.369140625f, 0.250976563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.630523682f, 0.461410522f, -1.f),
    C_Vertex3f(0.626373291f, 0.477609634f, -1.f),
    C_Vertex3f(0.6161623f, 0.506942749f, -1.f),
    C_Vertex3f(0.6040411f, 0.533210754f, -1.f),
    C_Vertex3f(0.590009689f, 0.556413651f, -1.f),
    C_Vertex3f(0.574068069f, 0.576551437f, -1.f),
    C_Vertex3f(0.555096149f, 0.594473362f, -1.f),
    C_Vertex3f(0.554077148f, 0.440437317f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.555096149f, 0.594473362f, -1.f),
    C_Vertex3f(0.533093929f, 0.610179424f, -1.f),
    C_Vertex3f(0.533441544f, 0.493788719f, -1.f),
    C_Vertex3f(0.542394638f, 0.474545479f, -1.f),
    C_Vertex3f(0.550474167f, 0.452680588f, -1.f),
    C_Vertex3f(0.554077148f, 0.440437317f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.484636307f, 0.54822731f, -1.f),
    C_Vertex3f(0.499921799f, 0.537015915f, -1.f),
    C_Vertex3f(0.508061409f, 0.623669624f, -1.f),
    C_Vertex3f(0.512914658f, 0.524410248f, -1.f),
    C_Vertex3f(0.533093929f, 0.610179424f, -1.f),
    C_Vertex3f(0.523614883f, 0.510410309f, -1.f),
    C_Vertex3f(0.533441544f, 0.493788719f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426224709f, 0.572784424f, -1.f),
    C_Vertex3f(0.447187424f, 0.566467285f, -1.f),
    C_Vertex3f(0.450270176f, 0.643399715f, -1.f),
    C_Vertex3f(0.467058182f, 0.558044434f, -1.f),
    C_Vertex3f(0.479998589f, 0.634943962f, -1.f),
    C_Vertex3f(0.484636307f, 0.54822731f, -1.f),
    C_Vertex3f(0.508061409f, 0.623669624f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.19361496f, 0.507694244f, -1.f),
    C_Vertex3f(0.206607819f, 0.522106171f, -1.f),
    C_Vertex3f(0.208589554f, 0.617435455f, -1.f),
    C_Vertex3f(0.221647739f, 0.535126209f, -1.f),
    C_Vertex3f(0.227630615f, 0.626040459f, -1.f),
    C_Vertex3f(0.238734722f, 0.54675436f, -1.f),
    C_Vertex3f(0.247367859f, 0.633416176f, -1.f),
    C_Vertex3f(0.257868767f, 0.556990623f, -1.f),
    C_Vertex3f(0.267801285f, 0.639562607f, -1.f),
    C_Vertex3f(0.279049873f, 0.565834999f, -1.f),
    C_Vertex3f(0.288930893f, 0.644479752f, -1.f),
    C_Vertex3f(0.302196026f, 0.572468281f, -1.f),
    C_Vertex3f(0.310756683f, 0.64816761f, -1.f),
    C_Vertex3f(0.327307224f, 0.576890469f, -1.f),
    C_Vertex3f(0.333278656f, 0.650626183f, -1.f),
    C_Vertex3f(0.354383469f, 0.579101563f, -1.f),
    C_Vertex3f(0.356496811f, 0.651855469f, -1.f),
    C_Vertex3f(0.381023407f, 0.579101563f, -1.f),
    C_Vertex3f(0.385816574f, 0.651855469f, -1.f),
    C_Vertex3f(0.404170036f, 0.57699585f, -1.f),
    C_Vertex3f(0.418876171f, 0.649036884f, -1.f),
    C_Vertex3f(0.426224709f, 0.572784424f, -1.f),
    C_Vertex3f(0.450270176f, 0.643399715f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163316727f, 0.460195541f, -1.f),
    C_Vertex3f(0.171969414f, 0.47673893f, -1.f),
    C_Vertex3f(0.172971249f, 0.596462607f, -1.f),
    C_Vertex3f(0.182068825f, 0.492571831f, -1.f),
    C_Vertex3f(0.190244675f, 0.607601166f, -1.f),
    C_Vertex3f(0.19361496f, 0.507694244f, -1.f),
    C_Vertex3f(0.208589554f, 0.617435455f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156769276f, 0.58401978f, -1.f),
    C_Vertex3f(0.141638756f, 0.570272684f, -1.f),
    C_Vertex3f(0.144426346f, 0.405766487f, -1.f),
    C_Vertex3f(0.153598785f, 0.436719894f, -1.f),
    C_Vertex3f(0.163316727f, 0.460195541f, -1.f),
    C_Vertex3f(0.172971249f, 0.596462607f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.115273356f, 0.102177978f, -1.f),
    C_Vertex3f(0.128643274f, 0.0864365101f, -1.f),
    C_Vertex3f(0.127579689f, 0.555221319f, -1.f),
    C_Vertex3f(0.135253906f, 0.340091705f, -1.f),
    C_Vertex3f(0.141638756f, 0.570272684f, -1.f),
    C_Vertex3f(0.138311386f, 0.373557091f, -1.f),
    C_Vertex3f(0.144426346f, 0.405766487f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.138017654f, 0.271846175f, -1.f),
    C_Vertex3f(0.136175156f, 0.291501164f, -1.f),
    C_Vertex3f(0.14319396f, 0.071821928f, -1.f),
    C_Vertex3f(0.135253906f, 0.312029839f, -1.f),
    C_Vertex3f(0.128643274f, 0.0864365101f, -1.f),
    C_Vertex3f(0.135253906f, 0.340091705f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0490732193f, 0.352183938f, -1.f),
    C_Vertex3f(0.0490800142f, 0.282654881f, -1.f),
    C_Vertex3f(0.0515165329f, 0.374789119f, -1.f),
    C_Vertex3f(0.0515369177f, 0.260173202f, -1.f),
    C_Vertex3f(0.0551815033f, 0.397004843f, -1.f),
    C_Vertex3f(0.0552222729f, 0.238251448f, -1.f),
    C_Vertex3f(0.0600681305f, 0.41883111f, -1.f),
    C_Vertex3f(0.0601360798f, 0.21688962f, -1.f),
    C_Vertex3f(0.0661764145f, 0.44026792f, -1.f),
    C_Vertex3f(0.0662783384f, 0.196087718f, -1.f),
    C_Vertex3f(0.0735063553f, 0.461315274f, -1.f),
    C_Vertex3f(0.0736490488f, 0.175845742f, -1.f),
    C_Vertex3f(0.0820579529f, 0.481973171f, -1.f),
    C_Vertex3f(0.0822482109f, 0.156163692f, -1.f),
    C_Vertex3f(0.0918312073f, 0.502241611f, -1.f),
    C_Vertex3f(0.0920758247f, 0.137041569f, -1.f),
    C_Vertex3f(0.102675915f, 0.521205783f, -1.f),
    C_Vertex3f(0.103084207f, 0.11904633f, -1.f),
    C_Vertex3f(0.114592075f, 0.538865685f, -1.f),
    C_Vertex3f(0.115273356f, 0.102177978f, -1.f),
    C_Vertex3f(0.127579689f, 0.555221319f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0478515625f, 0.329189301f, -1.f),
    C_Vertex3f(0.0478515625f, 0.305696487f, -1.f),
    C_Vertex3f(0.0490800142f, 0.282654881f, -1.f),
    C_Vertex3f(0.0490732193f, 0.352183938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.640136719f, 0.326660156f, -1.f),
    C_Vertex3f(0.369140625f, 0.326660156f, -1.f),
    C_Vertex3f(0.557128906f, 0.250976563f, -1.f),
    C_Vertex3f(0.560585976f, 0.036260128f, -1.f),
    C_Vertex3f(0.592799187f, 0.0549740791f, -1.f),
    C_Vertex3f(0.624521255f, 0.0767707825f, -1.f),
    C_Vertex3f(0.640136719f, 0.0892105103f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.560585976f, 0.036260128f, -1.f),
    C_Vertex3f(0.557128906f, 0.250976563f, -1.f),
    C_Vertex3f(0.557128906f, 0.132102966f, -1.f),
    C_Vertex3f(0.550359726f, 0.126115799f, -1.f),
    C_Vertex3f(0.533655167f, 0.114495277f, -1.f),
    C_Vertex3f(0.527881622f, 0.0206289291f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.513784409f, 0.103228569f, -1.f),
    C_Vertex3f(0.490747452f, 0.0923156738f, -1.f),
    C_Vertex3f(0.494686127f, 0.00808048248f, -1.f),
    C_Vertex3f(0.527881622f, 0.0206289291f, -1.f),
    C_Vertex3f(0.533655167f, 0.114495277f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.29569006f, -0.0031619072f, -1.f),
    C_Vertex3f(0.31765914f, -0.00695204735f, -1.f),
    C_Vertex3f(0.323589325f, 0.0694117546f, -1.f),
    C_Vertex3f(0.340167284f, -0.00947880745f, -1.f),
    C_Vertex3f(0.354377747f, 0.0659179688f, -1.f),
    C_Vertex3f(0.363214493f, -0.0107421875f, -1.f),
    C_Vertex3f(0.383966446f, 0.0659179688f, -1.f),
    C_Vertex3f(0.392478943f, -0.0107421875f, -1.f),
    C_Vertex3f(0.411153793f, 0.0685577393f, -1.f),
    C_Vertex3f(0.426984787f, -0.00760507584f, -1.f),
    C_Vertex3f(0.438013077f, 0.0738372803f, -1.f),
    C_Vertex3f(0.461053848f, -0.00133085251f, -1.f),
    C_Vertex3f(0.464544296f, 0.0817565918f, -1.f),
    C_Vertex3f(0.494686127f, 0.00808048248f, -1.f),
    C_Vertex3f(0.490747452f, 0.0923156738f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230262756f, 0.105277777f, -1.f),
    C_Vertex3f(0.233017206f, 0.0157887936f, -1.f),
    C_Vertex3f(0.243507385f, 0.0973620415f, -1.f),
    C_Vertex3f(0.253369093f, 0.00820851326f, -1.f),
    C_Vertex3f(0.264633179f, 0.0868806839f, -1.f),
    C_Vertex3f(0.274260044f, 0.00189161301f, -1.f),
    C_Vertex3f(0.293674469f, 0.0763993263f, -1.f),
    C_Vertex3f(0.29569006f, -0.0031619072f, -1.f),
    C_Vertex3f(0.323589325f, 0.0694117546f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.217857361f, 0.114121675f, -1.f),
    C_Vertex3f(0.206291199f, 0.123893738f, -1.f),
    C_Vertex3f(0.213204384f, 0.0246324539f, -1.f),
    C_Vertex3f(0.233017206f, 0.0157887936f, -1.f),
    C_Vertex3f(0.230262756f, 0.105277777f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.19556427f, 0.134593964f, -1.f),
    C_Vertex3f(0.185676575f, 0.146222353f, -1.f),
    C_Vertex3f(0.193930626f, 0.0347394943f, -1.f),
    C_Vertex3f(0.213204384f, 0.0246324539f, -1.f),
    C_Vertex3f(0.206291199f, 0.123893738f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.176628113f, 0.158778906f, -1.f),
    C_Vertex3f(0.168418884f, 0.172263622f, -1.f),
    C_Vertex3f(0.175837636f, 0.0459734201f, -1.f),
    C_Vertex3f(0.193930626f, 0.0347394943f, -1.f),
    C_Vertex3f(0.185676575f, 0.146222353f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161048889f, 0.186676502f, -1.f),
    C_Vertex3f(0.154600143f, 0.201963067f, -1.f),
    C_Vertex3f(0.158925414f, 0.0583342314f, -1.f),
    C_Vertex3f(0.175837636f, 0.0459734201f, -1.f),
    C_Vertex3f(0.168418884f, 0.172263622f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154600143f, 0.201963067f, -1.f),
    C_Vertex3f(0.149072647f, 0.218123317f, -1.f),
    C_Vertex3f(0.158925414f, 0.0583342314f, -1.f),
    C_Vertex3f(0.1444664f, 0.235157251f, -1.f),
    C_Vertex3f(0.14319396f, 0.071821928f, -1.f),
    C_Vertex3f(0.140781403f, 0.253064871f, -1.f),
    C_Vertex3f(0.138017654f, 0.271846175f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.557128906f, 0.250976563f, -1.f),
    C_Vertex3f(0.369140625f, 0.326660156f, -1.f),
    C_Vertex3f(0.369140625f, 0.250976563f, -1.f),
    C_End(),
    C_Translatef(0.696289063f, 0.f, 0.f),
};
static const Call char_72[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.072265625f, 0.f, 0.f),
    C_Vertex3f(0.072265625f, 0.f, -1.f),
    C_Vertex3f(0.072265625f, 0.640625f, -1.f),
    C_Vertex3f(0.072265625f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.072265625f, 0.640625f, 0.f),
    C_Vertex3f(0.072265625f, 0.640625f, -1.f),
    C_Vertex3f(0.157226563f, 0.640625f, -1.f),
    C_Vertex3f(0.157226563f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.157226563f, 0.640625f, 0.f),
    C_Vertex3f(0.157226563f, 0.640625f, -1.f),
    C_Vertex3f(0.157226563f, 0.377441406f, -1.f),
    C_Vertex3f(0.157226563f, 0.377441406f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.157226563f, 0.377441406f, 0.f),
    C_Vertex3f(0.157226563f, 0.377441406f, -1.f),
    C_Vertex3f(0.489257813f, 0.377441406f, -1.f),
    C_Vertex3f(0.489257813f, 0.377441406f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.377441406f, 0.f),
    C_Vertex3f(0.489257813f, 0.377441406f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.f, -1.f),
    C_Vertex3f(0.57421875f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.57421875f, 0.f, 0.f),
    C_Vertex3f(0.57421875f, 0.f, -1.f),
    C_Vertex3f(0.489257813f, 0.f, -1.f),
    C_Vertex3f(0.489257813f, 0.f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.f, -1.f),
    C_Vertex3f(0.489257813f, 0.301757813f, -1.f),
    C_Vertex3f(0.489257813f, 0.301757813f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.489257813f, 0.301757813f, 0.f),
    C_Vertex3f(0.489257813f, 0.301757813f, -1.f),
    C_Vertex3f(0.157226563f, 0.301757813f, -1.f),
    C_Vertex3f(0.157226563f, 0.301757813f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.157226563f, 0.301757813f, 0.f),
    C_Vertex3f(0.157226563f, 0.301757813f, -1.f),
    C_Vertex3f(0.157226563f, 0.f, -1.f),
    C_Vertex3f(0.157226563f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.157226563f, 0.f, 0.f),
    C_Vertex3f(0.157226563f, 0.f, -1.f),
    C_Vertex3f(0.072265625f, 0.f, -1.f),
    C_Vertex3f(0.072265625f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.57421875f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.377441406f, 0.f),
    C_Vertex3f(0.489257813f, 0.301757813f, 0.f),
    C_Vertex3f(0.157226563f, 0.377441406f, 0.f),
    C_Vertex3f(0.157226563f, 0.301757813f, 0.f),
    C_Vertex3f(0.157226563f, 0.640625f, 0.f),
    C_Vertex3f(0.157226563f, 0.f, 0.f),
    C_Vertex3f(0.072265625f, 0.640625f, 0.f),
    C_Vertex3f(0.072265625f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.57421875f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.301757813f, 0.f),
    C_Vertex3f(0.489257813f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.f, -1.f),
    C_Vertex3f(0.489257813f, 0.377441406f, -1.f),
    C_Vertex3f(0.489257813f, 0.301757813f, -1.f),
    C_Vertex3f(0.157226563f, 0.377441406f, -1.f),
    C_Vertex3f(0.157226563f, 0.301757813f, -1.f),
    C_Vertex3f(0.157226563f, 0.640625f, -1.f),
    C_Vertex3f(0.157226563f, 0.f, -1.f),
    C_Vertex3f(0.072265625f, 0.640625f, -1.f),
    C_Vertex3f(0.072265625f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.57421875f, 0.f, -1.f),
    C_Vertex3f(0.489257813f, 0.301757813f, -1.f),
    C_Vertex3f(0.489257813f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_73[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_74[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.135192409f, 0.990819395f, 0.f),
    C_Vertex3f(0.0258789063f, 0.181434631f, 0.f),
    C_Vertex3f(0.0258789063f, 0.181434631f, -1.f),
    C_Vertex3f(0.102539063f, 0.191894531f, -1.f),
    C_Vertex3f(0.102539063f, 0.191894531f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999123931f, 0.0418475792f, 0.f),
    C_Vertex3f(0.102539063f, 0.191894531f, 0.f),
    C_Vertex3f(0.102539063f, 0.191894531f, -1.f),
    C_Normal3f(0.997560084f, 0.069812797f, 0.f),
    C_Vertex3f(0.102921486f, 0.182764053f, 0.f),
    C_Vertex3f(0.102921486f, 0.182764053f, -1.f),
    C_Normal3f(0.994023681f, 0.109164424f, 0.f),
    C_Vertex3f(0.104355216f, 0.165943384f, 0.f),
    C_Vertex3f(0.104355216f, 0.165943384f, -1.f),
    C_Normal3f(0.986462772f, 0.163985327f, 0.f),
    C_Vertex3f(0.106457829f, 0.150563002f, 0.f),
    C_Vertex3f(0.106457829f, 0.150563002f, -1.f),
    C_Normal3f(0.973493278f, 0.228715658f, 0.f),
    C_Vertex3f(0.109229326f, 0.136622906f, 0.f),
    C_Vertex3f(0.109229326f, 0.136622906f, -1.f),
    C_Normal3f(0.952298999f, 0.305166453f, 0.f),
    C_Vertex3f(0.112669706f, 0.124123096f, 0.f),
    C_Vertex3f(0.112669706f, 0.124123096f, -1.f),
    C_Normal3f(0.918740988f, 0.394860655f, 0.f),
    C_Vertex3f(0.11677897f, 0.113063574f, 0.f),
    C_Vertex3f(0.11677897f, 0.113063574f, -1.f),
    C_Normal3f(0.86709106f, 0.498149633f, 0.f),
    C_Vertex3f(0.121557117f, 0.103444338f, 0.f),
    C_Vertex3f(0.121557117f, 0.103444338f, -1.f),
    C_Normal3f(0.790368974f, 0.612631023f, 0.f),
    C_Vertex3f(0.127004147f, 0.0952653885f, 0.f),
    C_Vertex3f(0.127004147f, 0.0952653885f, -1.f),
    C_Normal3f(0.701978683f, 0.712198079f, 0.f),
    C_Vertex3f(0.13312006f, 0.0885267258f, 0.f),
    C_Vertex3f(0.13312006f, 0.0885267258f, -1.f),
    C_Normal3f(0.616279304f, 0.787527621f, 0.f),
    C_Vertex3f(0.139823079f, 0.0826303959f, 0.f),
    C_Vertex3f(0.139823079f, 0.0826303959f, -1.f),
    C_Normal3f(0.521315753f, 0.853363931f, 0.f),
    C_Vertex3f(0.147113204f, 0.0775763988f, 0.f),
    C_Vertex3f(0.147113204f, 0.0775763988f, -1.f),
    C_Normal3f(0.420829952f, 0.90713954f, 0.f),
    C_Vertex3f(0.154990435f, 0.0733647346f, 0.f),
    C_Vertex3f(0.154990435f, 0.0733647346f, -1.f),
    C_Normal3f(0.319038033f, 0.947741985f, 0.f),
    C_Vertex3f(0.163454771f, 0.0699954033f, 0.f),
    C_Vertex3f(0.163454771f, 0.0699954033f, -1.f),
    C_Normal3f(0.219830945f, 0.975537956f, 0.f),
    C_Vertex3f(0.172506213f, 0.0674684048f, 0.f),
    C_Vertex3f(0.172506213f, 0.0674684048f, -1.f),
    C_Normal3f(0.126196653f, 0.992005229f, 0.f),
    C_Vertex3f(0.182144761f, 0.0657837391f, 0.f),
    C_Vertex3f(0.182144761f, 0.0657837391f, -1.f),
    C_Normal3f(0.0356593132f, 0.999364078f, 0.f),
    C_Vertex3f(0.192370415f, 0.0649414063f, 0.f),
    C_Vertex3f(0.192370415f, 0.0649414063f, -1.f),
    C_Normal3f(-0.0646432042f, 0.997908413f, 0.f),
    C_Vertex3f(0.205751419f, 0.0649414063f, 0.f),
    C_Vertex3f(0.205751419f, 0.0649414063f, -1.f),
    C_Normal3f(-0.188315094f, 0.982108653f, 0.f),
    C_Vertex3f(0.220607281f, 0.0667705536f, 0.f),
    C_Vertex3f(0.220607281f, 0.0667705536f, -1.f),
    C_Normal3f(-0.326994032f, 0.945026398f, 0.f),
    C_Vertex3f(0.234369755f, 0.0704288483f, 0.f),
    C_Vertex3f(0.234369755f, 0.0704288483f, -1.f),
    C_Normal3f(-0.466991752f, 0.884261727f, 0.f),
    C_Vertex3f(0.247038841f, 0.0759162903f, 0.f),
    C_Vertex3f(0.247038841f, 0.0759162903f, -1.f),
    C_Normal3f(-0.59834379f, 0.801239491f, 0.f),
    C_Vertex3f(0.25861454f, 0.0832328796f, 0.f),
    C_Vertex3f(0.25861454f, 0.0832328796f, -1.f),
    C_Normal3f(-0.722839355f, 0.691016018f, 0.f),
    C_Vertex3f(0.268388748f, 0.0918598175f, 0.f),
    C_Vertex3f(0.268388748f, 0.0918598175f, -1.f),
    C_Normal3f(-0.831675708f, 0.555261672f, 0.f),
    C_Vertex3f(0.276361465f, 0.101797104f, 0.f),
    C_Vertex3f(0.276361465f, 0.101797104f, -1.f),
    C_Normal3f(-0.914371073f, 0.404877156f, 0.f),
    C_Vertex3f(0.282532692f, 0.113044739f, 0.f),
    C_Vertex3f(0.282532692f, 0.113044739f, -1.f),
    C_Normal3f(-0.966631651f, 0.256170005f, 0.f),
    C_Vertex3f(0.286902428f, 0.125602722f, 0.f),
    C_Vertex3f(0.286902428f, 0.125602722f, -1.f),
    C_Normal3f(-0.988888085f, 0.148661554f, 0.f),
    C_Vertex3f(0.290179729f, 0.141900063f, 0.f),
    C_Vertex3f(0.290179729f, 0.141900063f, -1.f),
    C_Normal3f(-0.9972139f, 0.0745940581f, 0.f),
    C_Vertex3f(0.292364597f, 0.16193676f, 0.f),
    C_Vertex3f(0.292364597f, 0.16193676f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.293457031f, 0.185712814f, 0.f),
    C_Vertex3f(0.293457031f, 0.185712814f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.293457031f, 0.185712814f, 0.f),
    C_Vertex3f(0.293457031f, 0.185712814f, -1.f),
    C_Vertex3f(0.293457031f, 0.640625f, -1.f),
    C_Vertex3f(0.293457031f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.293457031f, 0.640625f, 0.f),
    C_Vertex3f(0.293457031f, 0.640625f, -1.f),
    C_Vertex3f(0.378417969f, 0.640625f, -1.f),
    C_Vertex3f(0.378417969f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.378417969f, 0.640625f, 0.f),
    C_Vertex3f(0.378417969f, 0.640625f, -1.f),
    C_Normal3f(0.999999166f, -0.00130672508f, 0.f),
    C_Vertex3f(0.378417969f, 0.194145203f, 0.f),
    C_Vertex3f(0.378417969f, 0.194145203f, -1.f),
    C_Normal3f(0.998772502f, -0.0495318472f, 0.f),
    C_Vertex3f(0.377809763f, 0.175182581f, 0.f),
    C_Vertex3f(0.377809763f, 0.175182581f, -1.f),
    C_Normal3f(0.996143281f, -0.0877406225f, 0.f),
    C_Vertex3f(0.376593351f, 0.157353163f, 0.f),
    C_Vertex3f(0.376593351f, 0.157353163f, -1.f),
    C_Normal3f(0.991403282f, -0.130841345f, 0.f),
    C_Vertex3f(0.374768734f, 0.140656948f, 0.f),
    C_Vertex3f(0.374768734f, 0.140656948f, -1.f),
    C_Normal3f(0.983750701f, -0.179539844f, 0.f),
    C_Vertex3f(0.372335911f, 0.125093937f, 0.f),
    C_Vertex3f(0.372335911f, 0.125093937f, -1.f),
    C_Normal3f(0.972100735f, -0.234563723f, 0.f),
    C_Vertex3f(0.369294882f, 0.110664129f, 0.f),
    C_Vertex3f(0.369294882f, 0.110664129f, -1.f),
    C_Normal3f(0.95500797f, -0.296580374f, 0.f),
    C_Vertex3f(0.365645647f, 0.0973675251f, 0.f),
    C_Vertex3f(0.365645647f, 0.0973675251f, -1.f),
    C_Normal3f(0.927667379f, -0.373407453f, 0.f),
    C_Vertex3f(0.361388206f, 0.0852041245f, 0.f),
    C_Vertex3f(0.361388206f, 0.0852041245f, -1.f),
    C_Normal3f(0.87537843f, -0.483438194f, 0.f),
    C_Vertex3f(0.354089737f, 0.0686588287f, 0.f),
    C_Vertex3f(0.354089737f, 0.0686588287f, -1.f),
    C_Normal3f(0.791552842f, -0.611100852f, 0.f),
    C_Vertex3f(0.341515064f, 0.0492191315f, 0.f),
    C_Vertex3f(0.341515064f, 0.0492191315f, -1.f),
    C_Normal3f(0.677437067f, -0.735580802f, 0.f),
    C_Vertex3f(0.326097012f, 0.0324001312f, 0.f),
    C_Vertex3f(0.326097012f, 0.0324001312f, -1.f),
    C_Normal3f(0.547791779f, -0.836614728f, 0.f),
    C_Vertex3f(0.307835579f, 0.018201828f, 0.f),
    C_Vertex3f(0.307835579f, 0.018201828f, -1.f),
    C_Normal3f(0.417937607f, -0.908475757f, 0.f),
    C_Vertex3f(0.286730766f, 0.0066242218f, 0.f),
    C_Vertex3f(0.286730766f, 0.0066242218f, -1.f),
    C_Normal3f(0.29030472f, -0.956934273f, 0.f),
    C_Vertex3f(0.263794422f, -0.00205898285f, 0.f),
    C_Vertex3f(0.263794422f, -0.00205898285f, -1.f),
    C_Normal3f(0.166676909f, -0.986011565f, 0.f),
    C_Vertex3f(0.239026546f, -0.00784778595f, 0.f),
    C_Vertex3f(0.239026546f, -0.00784778595f, -1.f),
    C_Normal3f(0.0562579446f, -0.998416305f, 0.f),
    C_Vertex3f(0.212427139f, -0.0107421875f, 0.f),
    C_Vertex3f(0.212427139f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0338453166f, -0.99942714f, 0.f),
    C_Vertex3f(0.187659264f, -0.0107421875f, 0.f),
    C_Vertex3f(0.187659264f, -0.0107421875f, -1.f),
    C_Normal3f(-0.116980776f, -0.993134201f, 0.f),
    C_Vertex3f(0.167791486f, -0.00923061371f, 0.f),
    C_Vertex3f(0.167791486f, -0.00923061371f, -1.f),
    C_Normal3f(-0.205328166f, -0.978693187f, 0.f),
    C_Vertex3f(0.149160743f, -0.00620746613f, 0.f),
    C_Vertex3f(0.149160743f, -0.00620746613f, -1.f),
    C_Normal3f(-0.300773591f, -0.953695595f, 0.f),
    C_Vertex3f(0.131767035f, -0.00167274475f, 0.f),
    C_Vertex3f(0.131767035f, -0.00167274475f, -1.f),
    C_Normal3f(-0.401023597f, -0.91606766f, 0.f),
    C_Vertex3f(0.115610361f, 0.00437355042f, 0.f),
    C_Vertex3f(0.115610361f, 0.00437355042f, -1.f),
    C_Normal3f(-0.502577543f, -0.864532113f, 0.f),
    C_Vertex3f(0.100690722f, 0.0119314194f, 0.f),
    C_Vertex3f(0.100690722f, 0.0119314194f, -1.f),
    C_Normal3f(-0.601063073f, -0.799201667f, 0.f),
    C_Vertex3f(0.0870081186f, 0.0210008621f, 0.f),
    C_Vertex3f(0.0870081186f, 0.0210008621f, -1.f),
    C_Normal3f(-0.691986382f, -0.721910596f, 0.f),
    C_Vertex3f(0.0745625496f, 0.0315818787f, 0.f),
    C_Vertex3f(0.0745625496f, 0.0315818787f, -1.f),
    C_Normal3f(-0.77440238f, -0.632693529f, 0.f),
    C_Vertex3f(0.0633540154f, 0.043674469f, 0.f),
    C_Vertex3f(0.0633540154f, 0.043674469f, -1.f),
    C_Normal3f(-0.845652819f, -0.533733428f, 0.f),
    C_Vertex3f(0.0536013842f, 0.0572378635f, 0.f),
    C_Vertex3f(0.0536013842f, 0.0572378635f, -1.f),
    C_Normal3f(-0.90153718f, -0.432701707f, 0.f),
    C_Vertex3f(0.045304656f, 0.0722720623f, 0.f),
    C_Vertex3f(0.045304656f, 0.0722720623f, -1.f),
    C_Normal3f(-0.942508519f, -0.334182143f, 0.f),
    C_Vertex3f(0.0384638309f, 0.0887770653f, 0.f),
    C_Vertex3f(0.0384638309f, 0.0887770653f, -1.f),
    C_Normal3f(-0.970396161f, -0.241518661f, 0.f),
    C_Vertex3f(0.0330789089f, 0.106752872f, 0.f),
    C_Vertex3f(0.0330789089f, 0.106752872f, -1.f),
    C_Normal3f(-0.987653911f, -0.156651706f, 0.f),
    C_Vertex3f(0.0291498899f, 0.126199484f, 0.f),
    C_Vertex3f(0.0291498899f, 0.126199484f, -1.f),
    C_Normal3f(-0.9967677f, -0.080337055f, 0.f),
    C_Vertex3f(0.026676774f, 0.147116899f, 0.f),
    C_Vertex3f(0.026676774f, 0.147116899f, -1.f),
    C_Normal3f(-0.999729872f, -0.0232431516f, 0.f),
    C_Vertex3f(0.0256595612f, 0.169505119f, 0.f),
    C_Vertex3f(0.0256595612f, 0.169505119f, -1.f),
    C_Normal3f(-0.999831021f, 0.0183836538f, 0.f),
    C_Vertex3f(0.0258789063f, 0.181434631f, 0.f),
    C_Vertex3f(0.0258789063f, 0.181434631f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378417969f, 0.640625f, 0.f),
    C_Vertex3f(0.293457031f, 0.640625f, 0.f),
    C_Vertex3f(0.307835579f, 0.018201828f, 0.f),
    C_Vertex3f(0.326097012f, 0.0324001312f, 0.f),
    C_Vertex3f(0.341515064f, 0.0492191315f, 0.f),
    C_Vertex3f(0.354089737f, 0.0686588287f, 0.f),
    C_Vertex3f(0.361388206f, 0.0852041245f, 0.f),
    C_Vertex3f(0.365645647f, 0.0973675251f, 0.f),
    C_Vertex3f(0.369294882f, 0.110664129f, 0.f),
    C_Vertex3f(0.372335911f, 0.125093937f, 0.f),
    C_Vertex3f(0.374768734f, 0.140656948f, 0.f),
    C_Vertex3f(0.376593351f, 0.157353163f, 0.f),
    C_Vertex3f(0.377809763f, 0.175182581f, 0.f),
    C_Vertex3f(0.378417969f, 0.194145203f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.307835579f, 0.018201828f, 0.f),
    C_Vertex3f(0.293457031f, 0.640625f, 0.f),
    C_Vertex3f(0.293457031f, 0.185712814f, 0.f),
    C_Vertex3f(0.292364597f, 0.16193676f, 0.f),
    C_Vertex3f(0.290179729f, 0.141900063f, 0.f),
    C_Vertex3f(0.286902428f, 0.125602722f, 0.f),
    C_Vertex3f(0.286730766f, 0.0066242218f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.286730766f, 0.0066242218f, 0.f),
    C_Vertex3f(0.286902428f, 0.125602722f, 0.f),
    C_Vertex3f(0.282532692f, 0.113044739f, 0.f),
    C_Vertex3f(0.276361465f, 0.101797104f, 0.f),
    C_Vertex3f(0.268388748f, 0.0918598175f, 0.f),
    C_Vertex3f(0.263794422f, -0.00205898285f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268388748f, 0.0918598175f, 0.f),
    C_Vertex3f(0.25861454f, 0.0832328796f, 0.f),
    C_Vertex3f(0.263794422f, -0.00205898285f, 0.f),
    C_Vertex3f(0.247038841f, 0.0759162903f, 0.f),
    C_Vertex3f(0.239026546f, -0.00784778595f, 0.f),
    C_Vertex3f(0.234369755f, 0.0704288483f, 0.f),
    C_Vertex3f(0.220607281f, 0.0667705536f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.212427139f, -0.0107421875f, 0.f),
    C_Vertex3f(0.239026546f, -0.00784778595f, 0.f),
    C_Vertex3f(0.220607281f, 0.0667705536f, 0.f),
    C_Vertex3f(0.205751419f, 0.0649414063f, 0.f),
    C_Vertex3f(0.192370415f, 0.0649414063f, 0.f),
    C_Vertex3f(0.187659264f, -0.0107421875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.192370415f, 0.0649414063f, 0.f),
    C_Vertex3f(0.182144761f, 0.0657837391f, 0.f),
    C_Vertex3f(0.187659264f, -0.0107421875f, 0.f),
    C_Vertex3f(0.172506213f, 0.0674684048f, 0.f),
    C_Vertex3f(0.167791486f, -0.00923061371f, 0.f),
    C_Vertex3f(0.163454771f, 0.0699954033f, 0.f),
    C_Vertex3f(0.154990435f, 0.0733647346f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.149160743f, -0.00620746613f, 0.f),
    C_Vertex3f(0.167791486f, -0.00923061371f, 0.f),
    C_Vertex3f(0.154990435f, 0.0733647346f, 0.f),
    C_Vertex3f(0.147113204f, 0.0775763988f, 0.f),
    C_Vertex3f(0.139823079f, 0.0826303959f, 0.f),
    C_Vertex3f(0.13312006f, 0.0885267258f, 0.f),
    C_Vertex3f(0.131767035f, -0.00167274475f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.131767035f, -0.00167274475f, 0.f),
    C_Vertex3f(0.13312006f, 0.0885267258f, 0.f),
    C_Vertex3f(0.127004147f, 0.0952653885f, 0.f),
    C_Vertex3f(0.121557117f, 0.103444338f, 0.f),
    C_Vertex3f(0.11677897f, 0.113063574f, 0.f),
    C_Vertex3f(0.115610361f, 0.00437355042f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115610361f, 0.00437355042f, 0.f),
    C_Vertex3f(0.11677897f, 0.113063574f, 0.f),
    C_Vertex3f(0.112669706f, 0.124123096f, 0.f),
    C_Vertex3f(0.109229326f, 0.136622906f, 0.f),
    C_Vertex3f(0.106457829f, 0.150563002f, 0.f),
    C_Vertex3f(0.104355216f, 0.165943384f, 0.f),
    C_Vertex3f(0.102921486f, 0.182764053f, 0.f),
    C_Vertex3f(0.100690722f, 0.0119314194f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.102539063f, 0.191894531f, 0.f),
    C_Vertex3f(0.0258789063f, 0.181434631f, 0.f),
    C_Vertex3f(0.026676774f, 0.147116899f, 0.f),
    C_Vertex3f(0.0291498899f, 0.126199484f, 0.f),
    C_Vertex3f(0.0330789089f, 0.106752872f, 0.f),
    C_Vertex3f(0.0384638309f, 0.0887770653f, 0.f),
    C_Vertex3f(0.045304656f, 0.0722720623f, 0.f),
    C_Vertex3f(0.0536013842f, 0.0572378635f, 0.f),
    C_Vertex3f(0.0633540154f, 0.043674469f, 0.f),
    C_Vertex3f(0.0745625496f, 0.0315818787f, 0.f),
    C_Vertex3f(0.0870081186f, 0.0210008621f, 0.f),
    C_Vertex3f(0.100690722f, 0.0119314194f, 0.f),
    C_Vertex3f(0.102921486f, 0.182764053f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.026676774f, 0.147116899f, 0.f),
    C_Vertex3f(0.0258789063f, 0.181434631f, 0.f),
    C_Vertex3f(0.0256595612f, 0.169505119f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378417969f, 0.640625f, -1.f),
    C_Vertex3f(0.293457031f, 0.640625f, -1.f),
    C_Vertex3f(0.307835579f, 0.018201828f, -1.f),
    C_Vertex3f(0.326097012f, 0.0324001312f, -1.f),
    C_Vertex3f(0.341515064f, 0.0492191315f, -1.f),
    C_Vertex3f(0.354089737f, 0.0686588287f, -1.f),
    C_Vertex3f(0.361388206f, 0.0852041245f, -1.f),
    C_Vertex3f(0.365645647f, 0.0973675251f, -1.f),
    C_Vertex3f(0.369294882f, 0.110664129f, -1.f),
    C_Vertex3f(0.372335911f, 0.125093937f, -1.f),
    C_Vertex3f(0.374768734f, 0.140656948f, -1.f),
    C_Vertex3f(0.376593351f, 0.157353163f, -1.f),
    C_Vertex3f(0.377809763f, 0.175182581f, -1.f),
    C_Vertex3f(0.378417969f, 0.194145203f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.307835579f, 0.018201828f, -1.f),
    C_Vertex3f(0.293457031f, 0.640625f, -1.f),
    C_Vertex3f(0.293457031f, 0.185712814f, -1.f),
    C_Vertex3f(0.292364597f, 0.16193676f, -1.f),
    C_Vertex3f(0.290179729f, 0.141900063f, -1.f),
    C_Vertex3f(0.286902428f, 0.125602722f, -1.f),
    C_Vertex3f(0.286730766f, 0.0066242218f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.286730766f, 0.0066242218f, -1.f),
    C_Vertex3f(0.286902428f, 0.125602722f, -1.f),
    C_Vertex3f(0.282532692f, 0.113044739f, -1.f),
    C_Vertex3f(0.276361465f, 0.101797104f, -1.f),
    C_Vertex3f(0.268388748f, 0.0918598175f, -1.f),
    C_Vertex3f(0.263794422f, -0.00205898285f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268388748f, 0.0918598175f, -1.f),
    C_Vertex3f(0.25861454f, 0.0832328796f, -1.f),
    C_Vertex3f(0.263794422f, -0.00205898285f, -1.f),
    C_Vertex3f(0.247038841f, 0.0759162903f, -1.f),
    C_Vertex3f(0.239026546f, -0.00784778595f, -1.f),
    C_Vertex3f(0.234369755f, 0.0704288483f, -1.f),
    C_Vertex3f(0.220607281f, 0.0667705536f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.212427139f, -0.0107421875f, -1.f),
    C_Vertex3f(0.239026546f, -0.00784778595f, -1.f),
    C_Vertex3f(0.220607281f, 0.0667705536f, -1.f),
    C_Vertex3f(0.205751419f, 0.0649414063f, -1.f),
    C_Vertex3f(0.192370415f, 0.0649414063f, -1.f),
    C_Vertex3f(0.187659264f, -0.0107421875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.192370415f, 0.0649414063f, -1.f),
    C_Vertex3f(0.182144761f, 0.0657837391f, -1.f),
    C_Vertex3f(0.187659264f, -0.0107421875f, -1.f),
    C_Vertex3f(0.172506213f, 0.0674684048f, -1.f),
    C_Vertex3f(0.167791486f, -0.00923061371f, -1.f),
    C_Vertex3f(0.163454771f, 0.0699954033f, -1.f),
    C_Vertex3f(0.154990435f, 0.0733647346f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.149160743f, -0.00620746613f, -1.f),
    C_Vertex3f(0.167791486f, -0.00923061371f, -1.f),
    C_Vertex3f(0.154990435f, 0.0733647346f, -1.f),
    C_Vertex3f(0.147113204f, 0.0775763988f, -1.f),
    C_Vertex3f(0.139823079f, 0.0826303959f, -1.f),
    C_Vertex3f(0.13312006f, 0.0885267258f, -1.f),
    C_Vertex3f(0.131767035f, -0.00167274475f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.131767035f, -0.00167274475f, -1.f),
    C_Vertex3f(0.13312006f, 0.0885267258f, -1.f),
    C_Vertex3f(0.127004147f, 0.0952653885f, -1.f),
    C_Vertex3f(0.121557117f, 0.103444338f, -1.f),
    C_Vertex3f(0.11677897f, 0.113063574f, -1.f),
    C_Vertex3f(0.115610361f, 0.00437355042f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115610361f, 0.00437355042f, -1.f),
    C_Vertex3f(0.11677897f, 0.113063574f, -1.f),
    C_Vertex3f(0.112669706f, 0.124123096f, -1.f),
    C_Vertex3f(0.109229326f, 0.136622906f, -1.f),
    C_Vertex3f(0.106457829f, 0.150563002f, -1.f),
    C_Vertex3f(0.104355216f, 0.165943384f, -1.f),
    C_Vertex3f(0.102921486f, 0.182764053f, -1.f),
    C_Vertex3f(0.100690722f, 0.0119314194f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.102539063f, 0.191894531f, -1.f),
    C_Vertex3f(0.0258789063f, 0.181434631f, -1.f),
    C_Vertex3f(0.026676774f, 0.147116899f, -1.f),
    C_Vertex3f(0.0291498899f, 0.126199484f, -1.f),
    C_Vertex3f(0.0330789089f, 0.106752872f, -1.f),
    C_Vertex3f(0.0384638309f, 0.0887770653f, -1.f),
    C_Vertex3f(0.045304656f, 0.0722720623f, -1.f),
    C_Vertex3f(0.0536013842f, 0.0572378635f, -1.f),
    C_Vertex3f(0.0633540154f, 0.043674469f, -1.f),
    C_Vertex3f(0.0745625496f, 0.0315818787f, -1.f),
    C_Vertex3f(0.0870081186f, 0.0210008621f, -1.f),
    C_Vertex3f(0.100690722f, 0.0119314194f, -1.f),
    C_Vertex3f(0.102921486f, 0.182764053f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.026676774f, 0.147116899f, -1.f),
    C_Vertex3f(0.0258789063f, 0.181434631f, -1.f),
    C_Vertex3f(0.0256595612f, 0.169505119f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_75[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0654296875f, 0.f, 0.f),
    C_Vertex3f(0.0654296875f, 0.f, -1.f),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Vertex3f(0.150390625f, 0.640625f, -1.f),
    C_Vertex3f(0.150390625f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.150390625f, 0.640625f, 0.f),
    C_Vertex3f(0.150390625f, 0.640625f, -1.f),
    C_Vertex3f(0.150390625f, 0.323074341f, -1.f),
    C_Vertex3f(0.150390625f, 0.323074341f, 0.f),
    C_Normal3f(-0.706453025f, 0.707759976f, 0.f),
    C_Vertex3f(0.150390625f, 0.323074341f, 0.f),
    C_Vertex3f(0.150390625f, 0.323074341f, -1.f),
    C_Vertex3f(0.468528748f, 0.640625f, -1.f),
    C_Vertex3f(0.468528748f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.468528748f, 0.640625f, 0.f),
    C_Vertex3f(0.468528748f, 0.640625f, -1.f),
    C_Vertex3f(0.583496094f, 0.640625f, -1.f),
    C_Vertex3f(0.583496094f, 0.640625f, 0.f),
    C_Normal3f(0.694854915f, -0.719149828f, 0.f),
    C_Vertex3f(0.583496094f, 0.640625f, 0.f),
    C_Vertex3f(0.583496094f, 0.640625f, -1.f),
    C_Vertex3f(0.314750671f, 0.380958557f, -1.f),
    C_Vertex3f(0.314750671f, 0.380958557f, 0.f),
    C_Normal3f(0.805300176f, 0.592867255f, 0.f),
    C_Vertex3f(0.314750671f, 0.380958557f, 0.f),
    C_Vertex3f(0.314750671f, 0.380958557f, -1.f),
    C_Vertex3f(0.595214844f, 0.f, -1.f),
    C_Vertex3f(0.595214844f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.595214844f, 0.f, 0.f),
    C_Vertex3f(0.595214844f, 0.f, -1.f),
    C_Vertex3f(0.483329773f, 0.f, -1.f),
    C_Vertex3f(0.483329773f, 0.f, 0.f),
    C_Normal3f(-0.817530513f, -0.575885355f, 0.f),
    C_Vertex3f(0.483329773f, 0.f, 0.f),
    C_Vertex3f(0.483329773f, 0.f, -1.f),
    C_Vertex3f(0.255378723f, 0.323600769f, -1.f),
    C_Vertex3f(0.255378723f, 0.323600769f, 0.f),
    C_Normal3f(0.694824874f, -0.719179034f, 0.f),
    C_Vertex3f(0.255378723f, 0.323600769f, 0.f),
    C_Vertex3f(0.255378723f, 0.323600769f, -1.f),
    C_Vertex3f(0.150390625f, 0.222167969f, -1.f),
    C_Vertex3f(0.150390625f, 0.222167969f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.150390625f, 0.222167969f, 0.f),
    C_Vertex3f(0.150390625f, 0.222167969f, -1.f),
    C_Vertex3f(0.150390625f, 0.f, -1.f),
    C_Vertex3f(0.150390625f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.150390625f, 0.f, 0.f),
    C_Vertex3f(0.150390625f, 0.f, -1.f),
    C_Vertex3f(0.0654296875f, 0.f, -1.f),
    C_Vertex3f(0.0654296875f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_Vertex3f(0.150390625f, 0.f, 0.f),
    C_Vertex3f(0.150390625f, 0.640625f, 0.f),
    C_Vertex3f(0.150390625f, 0.222167969f, 0.f),
    C_Vertex3f(0.150390625f, 0.323074341f, 0.f),
    C_Vertex3f(0.255378723f, 0.323600769f, 0.f),
    C_Vertex3f(0.468528748f, 0.640625f, 0.f),
    C_Vertex3f(0.314750671f, 0.380958557f, 0.f),
    C_Vertex3f(0.583496094f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.314750671f, 0.380958557f, 0.f),
    C_Vertex3f(0.255378723f, 0.323600769f, 0.f),
    C_Vertex3f(0.483329773f, 0.f, 0.f),
    C_Vertex3f(0.595214844f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.150390625f, 0.f, 0.f),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_Vertex3f(0.0654296875f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Vertex3f(0.150390625f, 0.f, -1.f),
    C_Vertex3f(0.150390625f, 0.640625f, -1.f),
    C_Vertex3f(0.150390625f, 0.222167969f, -1.f),
    C_Vertex3f(0.150390625f, 0.323074341f, -1.f),
    C_Vertex3f(0.255378723f, 0.323600769f, -1.f),
    C_Vertex3f(0.468528748f, 0.640625f, -1.f),
    C_Vertex3f(0.314750671f, 0.380958557f, -1.f),
    C_Vertex3f(0.583496094f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.314750671f, 0.380958557f, -1.f),
    C_Vertex3f(0.255378723f, 0.323600769f, -1.f),
    C_Vertex3f(0.483329773f, 0.f, -1.f),
    C_Vertex3f(0.595214844f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.150390625f, 0.f, -1.f),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Vertex3f(0.0654296875f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_76[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0654296875f, 0.f, 0.f),
    C_Vertex3f(0.0654296875f, 0.f, -1.f),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_Vertex3f(0.150390625f, 0.640625f, -1.f),
    C_Vertex3f(0.150390625f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.150390625f, 0.640625f, 0.f),
    C_Vertex3f(0.150390625f, 0.640625f, -1.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, -1.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, 0.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, -1.f),
    C_Vertex3f(0.465820313f, 0.0756835938f, -1.f),
    C_Vertex3f(0.465820313f, 0.0756835938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.465820313f, 0.0756835938f, 0.f),
    C_Vertex3f(0.465820313f, 0.0756835938f, -1.f),
    C_Vertex3f(0.465820313f, 0.f, -1.f),
    C_Vertex3f(0.465820313f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.465820313f, 0.f, 0.f),
    C_Vertex3f(0.465820313f, 0.f, -1.f),
    C_Vertex3f(0.0654296875f, 0.f, -1.f),
    C_Vertex3f(0.0654296875f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0654296875f, 0.f, 0.f),
    C_Vertex3f(0.465820313f, 0.f, 0.f),
    C_Vertex3f(0.465820313f, 0.0756835938f, 0.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, 0.f),
    C_Vertex3f(0.150390625f, 0.640625f, 0.f),
    C_Vertex3f(0.0654296875f, 0.640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0654296875f, 0.f, -1.f),
    C_Vertex3f(0.465820313f, 0.f, -1.f),
    C_Vertex3f(0.465820313f, 0.0756835938f, -1.f),
    C_Vertex3f(0.150390625f, 0.0756835938f, -1.f),
    C_Vertex3f(0.150390625f, 0.640625f, -1.f),
    C_Vertex3f(0.0654296875f, 0.640625f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_77[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.06640625f, 0.f, 0.f),
    C_Vertex3f(0.06640625f, 0.f, -1.f),
    C_Vertex3f(0.06640625f, 0.640625f, -1.f),
    C_Vertex3f(0.06640625f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.06640625f, 0.640625f, 0.f),
    C_Vertex3f(0.06640625f, 0.640625f, -1.f),
    C_Vertex3f(0.193855286f, 0.640625f, -1.f),
    C_Vertex3f(0.193855286f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.948355973f, 0.317208141f, 0.f),
    C_Vertex3f(0.193855286f, 0.640625f, 0.f),
    C_Vertex3f(0.193855286f, 0.640625f, -1.f),
    C_Normal3f(0.948399723f, 0.31707716f, 0.f),
    C_Vertex3f(0.34823513f, 0.179076195f, 0.f),
    C_Vertex3f(0.34823513f, 0.179076195f, -1.f),
    C_Normal3f(0.950018406f, 0.31219402f, 0.f),
    C_Vertex3f(0.353127003f, 0.164232254f, 0.f),
    C_Vertex3f(0.353127003f, 0.164232254f, -1.f),
    C_Normal3f(0.950602651f, 0.310410231f, 0.f),
    C_Vertex3f(0.357663631f, 0.150384903f, 0.f),
    C_Vertex3f(0.357663631f, 0.150384903f, -1.f),
    C_Normal3f(0.951277971f, 0.308334529f, 0.f),
    C_Vertex3f(0.361845016f, 0.137534142f, 0.f),
    C_Vertex3f(0.361845016f, 0.137534142f, -1.f),
    C_Normal3f(0.952067316f, 0.305888861f, 0.f),
    C_Vertex3f(0.365671158f, 0.12567997f, 0.f),
    C_Vertex3f(0.365671158f, 0.12567997f, -1.f),
    C_Normal3f(0.953001797f, 0.302964658f, 0.f),
    C_Vertex3f(0.369142056f, 0.114822388f, 0.f),
    C_Vertex3f(0.369142056f, 0.114822388f, -1.f),
    C_Normal3f(0.954125643f, 0.299406409f, 0.f),
    C_Vertex3f(0.37225771f, 0.104961395f, 0.f),
    C_Vertex3f(0.37225771f, 0.104961395f, -1.f),
    C_Normal3f(0.955251932f, 0.295793533f, 0.f),
    C_Vertex3f(0.37501812f, 0.0960969925f, 0.f),
    C_Vertex3f(0.37501812f, 0.0960969925f, -1.f),
    C_Normal3f(0.956313729f, 0.292342216f, 0.f),
    C_Vertex3f(0.376220703f, 0.0921630859f, 0.f),
    C_Vertex3f(0.376220703f, 0.0921630859f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.954194307f, 0.29918772f, 0.f),
    C_Vertex3f(0.376220703f, 0.0921630859f, 0.f),
    C_Vertex3f(0.376220703f, 0.0921630859f, -1.f),
    C_Normal3f(-0.952987671f, 0.303009301f, 0.f),
    C_Vertex3f(0.377591133f, 0.0965337753f, 0.f),
    C_Vertex3f(0.377591133f, 0.0965337753f, -1.f),
    C_Normal3f(-0.951691389f, 0.307056338f, 0.f),
    C_Vertex3f(0.38071537f, 0.106299162f, 0.f),
    C_Vertex3f(0.38071537f, 0.106299162f, -1.f),
    C_Normal3f(-0.950380683f, 0.311089456f, 0.f),
    C_Vertex3f(0.384222984f, 0.117088556f, 0.f),
    C_Vertex3f(0.384222984f, 0.117088556f, -1.f),
    C_Normal3f(-0.94927752f, 0.314439625f, 0.f),
    C_Vertex3f(0.388113976f, 0.128901958f, 0.f),
    C_Vertex3f(0.388113976f, 0.128901958f, -1.f),
    C_Normal3f(-0.948336363f, 0.317266673f, 0.f),
    C_Vertex3f(0.392388344f, 0.141739368f, 0.f),
    C_Vertex3f(0.392388344f, 0.141739368f, -1.f),
    C_Normal3f(-0.94752413f, 0.319684178f, 0.f),
    C_Vertex3f(0.397046089f, 0.155600786f, 0.f),
    C_Vertex3f(0.397046089f, 0.155600786f, -1.f),
    C_Normal3f(-0.946816146f, 0.321775138f, 0.f),
    C_Vertex3f(0.402087212f, 0.170486212f, 0.f),
    C_Vertex3f(0.402087212f, 0.170486212f, -1.f),
    C_Normal3f(-0.945276439f, 0.326270342f, 0.f),
    C_Vertex3f(0.407511711f, 0.186395645f, 0.f),
    C_Vertex3f(0.407511711f, 0.186395645f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.945276439f, 0.326270342f, 0.f),
    C_Vertex3f(0.407511711f, 0.186395645f, 0.f),
    C_Vertex3f(0.407511711f, 0.186395645f, -1.f),
    C_Vertex3f(0.564292908f, 0.640625f, -1.f),
    C_Vertex3f(0.564292908f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.564292908f, 0.640625f, 0.f),
    C_Vertex3f(0.564292908f, 0.640625f, -1.f),
    C_Vertex3f(0.678222656f, 0.640625f, -1.f),
    C_Vertex3f(0.678222656f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.678222656f, 0.640625f, 0.f),
    C_Vertex3f(0.678222656f, 0.640625f, -1.f),
    C_Vertex3f(0.678222656f, 0.f, -1.f),
    C_Vertex3f(0.678222656f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.678222656f, 0.f, 0.f),
    C_Vertex3f(0.678222656f, 0.f, -1.f),
    C_Vertex3f(0.596679688f, 0.f, -1.f),
    C_Vertex3f(0.596679688f, 0.f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.596679688f, 0.f, 0.f),
    C_Vertex3f(0.596679688f, 0.f, -1.f),
    C_Vertex3f(0.596679688f, 0.536132813f, -1.f),
    C_Vertex3f(0.596679688f, 0.536132813f, 0.f),
    C_Normal3f(0.944264889f, -0.329186618f, 0.f),
    C_Vertex3f(0.596679688f, 0.536132813f, 0.f),
    C_Vertex3f(0.596679688f, 0.536132813f, -1.f),
    C_Vertex3f(0.40977478f, 0.f, -1.f),
    C_Vertex3f(0.40977478f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.40977478f, 0.f, 0.f),
    C_Vertex3f(0.40977478f, 0.f, -1.f),
    C_Vertex3f(0.33341217f, 0.f, -1.f),
    C_Vertex3f(0.33341217f, 0.f, 0.f),
    C_Normal3f(-0.946760297f, -0.32193929f, 0.f),
    C_Vertex3f(0.33341217f, 0.f, 0.f),
    C_Vertex3f(0.33341217f, 0.f, -1.f),
    C_Vertex3f(0.147949219f, 0.545410156f, -1.f),
    C_Vertex3f(0.147949219f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.147949219f, 0.545410156f, 0.f),
    C_Vertex3f(0.147949219f, 0.545410156f, -1.f),
    C_Vertex3f(0.147949219f, 0.f, -1.f),
    C_Vertex3f(0.147949219f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.147949219f, 0.f, 0.f),
    C_Vertex3f(0.147949219f, 0.f, -1.f),
    C_Vertex3f(0.06640625f, 0.f, -1.f),
    C_Vertex3f(0.06640625f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.678222656f, 0.f, 0.f),
    C_Vertex3f(0.678222656f, 0.640625f, 0.f),
    C_Vertex3f(0.596679688f, 0.536132813f, 0.f),
    C_Vertex3f(0.564292908f, 0.640625f, 0.f),
    C_Vertex3f(0.40977478f, 0.f, 0.f),
    C_Vertex3f(0.407511711f, 0.186395645f, 0.f),
    C_Vertex3f(0.402087212f, 0.170486212f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.40977478f, 0.f, 0.f),
    C_Vertex3f(0.402087212f, 0.170486212f, 0.f),
    C_Vertex3f(0.397046089f, 0.155600786f, 0.f),
    C_Vertex3f(0.392388344f, 0.141739368f, 0.f),
    C_Vertex3f(0.388113976f, 0.128901958f, 0.f),
    C_Vertex3f(0.384222984f, 0.117088556f, 0.f),
    C_Vertex3f(0.38071537f, 0.106299162f, 0.f),
    C_Vertex3f(0.377591133f, 0.0965337753f, 0.f),
    C_Vertex3f(0.376220703f, 0.0921630859f, 0.f),
    C_Vertex3f(0.33341217f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376220703f, 0.0921630859f, 0.f),
    C_Vertex3f(0.37501812f, 0.0960969925f, 0.f),
    C_Vertex3f(0.37225771f, 0.104961395f, 0.f),
    C_Vertex3f(0.369142056f, 0.114822388f, 0.f),
    C_Vertex3f(0.365671158f, 0.12567997f, 0.f),
    C_Vertex3f(0.361845016f, 0.137534142f, 0.f),
    C_Vertex3f(0.357663631f, 0.150384903f, 0.f),
    C_Vertex3f(0.353127003f, 0.164232254f, 0.f),
    C_Vertex3f(0.34823513f, 0.179076195f, 0.f),
    C_Vertex3f(0.33341217f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.06640625f, 0.f, 0.f),
    C_Vertex3f(0.147949219f, 0.f, 0.f),
    C_Vertex3f(0.06640625f, 0.640625f, 0.f),
    C_Vertex3f(0.147949219f, 0.545410156f, 0.f),
    C_Vertex3f(0.193855286f, 0.640625f, 0.f),
    C_Vertex3f(0.33341217f, 0.f, 0.f),
    C_Vertex3f(0.34823513f, 0.179076195f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.678222656f, 0.f, 0.f),
    C_Vertex3f(0.596679688f, 0.536132813f, 0.f),
    C_Vertex3f(0.596679688f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.678222656f, 0.f, -1.f),
    C_Vertex3f(0.678222656f, 0.640625f, -1.f),
    C_Vertex3f(0.596679688f, 0.536132813f, -1.f),
    C_Vertex3f(0.564292908f, 0.640625f, -1.f),
    C_Vertex3f(0.40977478f, 0.f, -1.f),
    C_Vertex3f(0.407511711f, 0.186395645f, -1.f),
    C_Vertex3f(0.402087212f, 0.170486212f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.40977478f, 0.f, -1.f),
    C_Vertex3f(0.402087212f, 0.170486212f, -1.f),
    C_Vertex3f(0.397046089f, 0.155600786f, -1.f),
    C_Vertex3f(0.392388344f, 0.141739368f, -1.f),
    C_Vertex3f(0.388113976f, 0.128901958f, -1.f),
    C_Vertex3f(0.384222984f, 0.117088556f, -1.f),
    C_Vertex3f(0.38071537f, 0.106299162f, -1.f),
    C_Vertex3f(0.377591133f, 0.0965337753f, -1.f),
    C_Vertex3f(0.376220703f, 0.0921630859f, -1.f),
    C_Vertex3f(0.33341217f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376220703f, 0.0921630859f, -1.f),
    C_Vertex3f(0.37501812f, 0.0960969925f, -1.f),
    C_Vertex3f(0.37225771f, 0.104961395f, -1.f),
    C_Vertex3f(0.369142056f, 0.114822388f, -1.f),
    C_Vertex3f(0.365671158f, 0.12567997f, -1.f),
    C_Vertex3f(0.361845016f, 0.137534142f, -1.f),
    C_Vertex3f(0.357663631f, 0.150384903f, -1.f),
    C_Vertex3f(0.353127003f, 0.164232254f, -1.f),
    C_Vertex3f(0.34823513f, 0.179076195f, -1.f),
    C_Vertex3f(0.33341217f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.06640625f, 0.f, -1.f),
    C_Vertex3f(0.147949219f, 0.f, -1.f),
    C_Vertex3f(0.06640625f, 0.640625f, -1.f),
    C_Vertex3f(0.147949219f, 0.545410156f, -1.f),
    C_Vertex3f(0.193855286f, 0.640625f, -1.f),
    C_Vertex3f(0.33341217f, 0.f, -1.f),
    C_Vertex3f(0.34823513f, 0.179076195f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.678222656f, 0.f, -1.f),
    C_Vertex3f(0.596679688f, 0.536132813f, -1.f),
    C_Vertex3f(0.596679688f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.745605469f, 0.f, 0.f),
};
static const Call char_78[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.068359375f, 0.f, 0.f),
    C_Vertex3f(0.068359375f, 0.f, -1.f),
    C_Vertex3f(0.068359375f, 0.640625f, -1.f),
    C_Vertex3f(0.068359375f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.068359375f, 0.640625f, 0.f),
    C_Vertex3f(0.068359375f, 0.640625f, -1.f),
    C_Vertex3f(0.155105591f, 0.640625f, -1.f),
    C_Vertex3f(0.155105591f, 0.640625f, 0.f),
    C_Normal3f(0.830788314f, 0.556588352f, 0.f),
    C_Vertex3f(0.155105591f, 0.640625f, 0.f),
    C_Vertex3f(0.155105591f, 0.640625f, -1.f),
    C_Vertex3f(0.4921875f, 0.137481689f, -1.f),
    C_Vertex3f(0.4921875f, 0.137481689f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.4921875f, 0.137481689f, 0.f),
    C_Vertex3f(0.4921875f, 0.137481689f, -1.f),
    C_Vertex3f(0.4921875f, 0.640625f, -1.f),
    C_Vertex3f(0.4921875f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.4921875f, 0.640625f, 0.f),
    C_Vertex3f(0.4921875f, 0.640625f, -1.f),
    C_Vertex3f(0.573242188f, 0.640625f, -1.f),
    C_Vertex3f(0.573242188f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.573242188f, 0.640625f, 0.f),
    C_Vertex3f(0.573242188f, 0.640625f, -1.f),
    C_Vertex3f(0.573242188f, 0.f, -1.f),
    C_Vertex3f(0.573242188f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.573242188f, 0.f, 0.f),
    C_Vertex3f(0.573242188f, 0.f, -1.f),
    C_Vertex3f(0.486495972f, 0.f, -1.f),
    C_Vertex3f(0.486495972f, 0.f, 0.f),
    C_Normal3f(-0.830800116f, -0.556570947f, 0.f),
    C_Vertex3f(0.486495972f, 0.f, 0.f),
    C_Vertex3f(0.486495972f, 0.f, -1.f),
    C_Vertex3f(0.149414063f, 0.503166199f, -1.f),
    C_Vertex3f(0.149414063f, 0.503166199f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.149414063f, 0.503166199f, 0.f),
    C_Vertex3f(0.149414063f, 0.503166199f, -1.f),
    C_Vertex3f(0.149414063f, 0.f, -1.f),
    C_Vertex3f(0.149414063f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.149414063f, 0.f, 0.f),
    C_Vertex3f(0.149414063f, 0.f, -1.f),
    C_Vertex3f(0.068359375f, 0.f, -1.f),
    C_Vertex3f(0.068359375f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.068359375f, 0.f, 0.f),
    C_Vertex3f(0.149414063f, 0.f, 0.f),
    C_Vertex3f(0.068359375f, 0.640625f, 0.f),
    C_Vertex3f(0.149414063f, 0.503166199f, 0.f),
    C_Vertex3f(0.155105591f, 0.640625f, 0.f),
    C_Vertex3f(0.486495972f, 0.f, 0.f),
    C_Vertex3f(0.4921875f, 0.137481689f, 0.f),
    C_Vertex3f(0.573242188f, 0.f, 0.f),
    C_Vertex3f(0.4921875f, 0.640625f, 0.f),
    C_Vertex3f(0.573242188f, 0.640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.068359375f, 0.f, -1.f),
    C_Vertex3f(0.149414063f, 0.f, -1.f),
    C_Vertex3f(0.068359375f, 0.640625f, -1.f),
    C_Vertex3f(0.149414063f, 0.503166199f, -1.f),
    C_Vertex3f(0.155105591f, 0.640625f, -1.f),
    C_Vertex3f(0.486495972f, 0.f, -1.f),
    C_Vertex3f(0.4921875f, 0.137481689f, -1.f),
    C_Vertex3f(0.573242188f, 0.f, -1.f),
    C_Vertex3f(0.4921875f, 0.640625f, -1.f),
    C_Vertex3f(0.573242188f, 0.640625f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_79[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_Normal3f(-0.999229491f, 0.0392487273f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Normal3f(-0.994037926f, 0.109034933f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Normal3f(-0.981642663f, 0.190729454f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Normal3f(-0.960366726f, 0.278739631f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Normal3f(-0.928485572f, 0.371368498f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Normal3f(-0.884783864f, 0.46600157f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Normal3f(-0.828982711f, 0.559274256f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Normal3f(-0.762047708f, 0.647521019f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Normal3f(-0.689660072f, 0.724133313f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Normal3f(-0.610743105f, 0.791828811f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Normal3f(-0.522311151f, 0.85275501f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Normal3f(-0.426912457f, 0.904293001f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Normal3f(-0.327885866f, 0.944717348f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Normal3f(-0.228800729f, 0.97347337f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Normal3f(-0.132870927f, 0.991133392f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Normal3f(-0.0466583073f, 0.998910844f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Normal3f(0.0266486872f, 0.999644876f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Normal3f(0.0925619975f, 0.995706856f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Normal3f(0.157029793f, 0.98759383f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Normal3f(0.222855464f, 0.974851489f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Normal3f(0.289159536f, 0.957280874f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Normal3f(0.354979903f, 0.934873939f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Normal3f(0.419335544f, 0.907831371f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Normal3f(0.481297553f, 0.876557291f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Normal3f(0.540897012f, 0.841088831f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Normal3f(0.599702179f, 0.800223231f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Normal3f(0.656618118f, 0.754223228f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Normal3f(0.710595846f, 0.703600347f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Normal3f(0.760694981f, 0.649109423f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Normal3f(0.806162775f, 0.591693759f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Normal3f(0.846489131f, 0.532405972f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Normal3f(0.881427884f, 0.47231859f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Normal3f(0.910308123f, 0.41393134f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Normal3f(0.934312224f, 0.356455833f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Normal3f(0.954386652f, 0.298573464f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Normal3f(0.970532179f, 0.240971416f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Normal3f(0.982873797f, 0.184280097f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Normal3f(0.991638541f, 0.129045784f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Normal3f(0.997129619f, 0.0757144392f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Normal3f(0.999698818f, 0.0245408304f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_Normal3f(0.999670804f, -0.0256569553f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Normal3f(0.996906638f, -0.0785944462f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Normal3f(0.990996599f, -0.133886874f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Normal3f(0.981580496f, -0.191049173f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Normal3f(0.968356967f, -0.249568939f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Normal3f(0.951117396f, -0.308829606f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Normal3f(0.929773211f, -0.36813283f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Normal3f(0.90437752f, -0.426733196f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Normal3f(0.874280751f, -0.485420704f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Normal3f(0.838072717f, -0.545558572f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Normal3f(0.795999944f, -0.605296671f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Normal3f(0.74819994f, -0.663473368f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Normal3f(0.695117056f, -0.718896568f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Normal3f(0.637500465f, -0.770449877f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Normal3f(0.576356828f, -0.817198098f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Normal3f(0.512864292f, -0.858469784f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Normal3f(0.45348832f, -0.891262233f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Normal3f(0.396994501f, -0.91782105f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Normal3f(0.33799234f, -0.941148937f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Normal3f(0.277123004f, -0.960834503f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Normal3f(0.215111941f, -0.976589382f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Normal3f(0.152726561f, -0.988268495f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Normal3f(0.0907294229f, -0.995875537f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Normal3f(0.0292172581f, -0.999573112f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0311744045f, -0.999514043f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0938582048f, -0.995585501f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Normal3f(-0.159526795f, -0.987193584f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Normal3f(-0.226795033f, -0.973942518f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Normal3f(-0.294730932f, -0.955580294f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Normal3f(-0.362296075f, -0.932063103f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Normal3f(-0.428416908f, -0.903581202f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Normal3f(-0.492066324f, -0.870557725f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Normal3f(-0.551469922f, -0.834194779f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Normal3f(-0.60842973f, -0.793607712f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Normal3f(-0.663708389f, -0.747991323f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Normal3f(-0.716302752f, -0.697789609f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Normal3f(-0.765292704f, -0.64368248f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Normal3f(-0.809918106f, -0.586542904f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Normal3f(-0.84963721f, -0.527367532f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Normal3f(-0.884154499f, -0.467194706f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Normal3f(-0.911144316f, -0.412087172f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Normal3f(-0.932779849f, -0.360446692f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Normal3f(-0.951756537f, -0.306854159f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Normal3f(-0.967785478f, -0.251776159f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Normal3f(-0.980656683f, -0.195735648f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Normal3f(-0.990251899f, -0.139287636f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Normal3f(-0.996550143f, -0.0829921961f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Normal3f(0.999472976f, 0.0324606337f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Normal3f(0.994043112f, 0.108987279f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Normal3f(0.98171258f, 0.190368786f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Normal3f(0.960640907f, 0.277793169f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Normal3f(0.929206252f, 0.369561613f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Normal3f(0.886297405f, 0.463116556f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Normal3f(0.831705272f, 0.555217385f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Normal3f(0.766394138f, 0.642370582f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Normal3f(0.696807623f, 0.717258096f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Normal3f(0.621382415f, 0.783507407f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Normal3f(0.535601974f, 0.84447062f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Normal3f(0.441416115f, 0.897302508f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Normal3f(0.341795832f, 0.939774215f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Normal3f(0.240293324f, 0.970700264f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Normal3f(0.140439287f, 0.990089297f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Normal3f(0.0449865349f, 0.998987615f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0450702235f, 0.9989838f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Normal3f(-0.139715239f, 0.990191758f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Normal3f(-0.239666194f, 0.970855415f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Normal3f(-0.341789991f, 0.939776421f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Normal3f(-0.442493498f, 0.896771669f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Normal3f(-0.538056731f, 0.842908621f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Normal3f(-0.625280082f, 0.780400395f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Normal3f(-0.701993346f, 0.712183535f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Normal3f(-0.773547888f, 0.633737862f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Normal3f(-0.840062797f, 0.542489231f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Normal3f(-0.894041598f, 0.44798404f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Normal3f(-0.935230792f, 0.354038715f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Normal3f(-0.964579821f, 0.263791323f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Normal3f(-0.983775556f, 0.179403409f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Normal3f(-0.994777799f, 0.102064326f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Normal3f(-0.999368489f, 0.0355346389f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Normal3f(-0.999824762f, -0.0187195782f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Normal3f(-0.997788966f, -0.066463232f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Normal3f(-0.993469477f, -0.114098631f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Normal3f(-0.986415029f, -0.164272755f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Normal3f(-0.976221919f, -0.216773391f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Normal3f(-0.962501347f, -0.2712771f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Normal3f(-0.944906473f, -0.327340305f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Normal3f(-0.923166513f, -0.384400129f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Normal3f(-0.896669626f, -0.442700326f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Normal3f(-0.864540637f, -0.50256294f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Normal3f(-0.826793492f, -0.562505424f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Normal3f(-0.783447027f, -0.62145859f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Normal3f(-0.734794199f, -0.678290129f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Normal3f(-0.681415319f, -0.731896937f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Normal3f(-0.62415117f, -0.781303585f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Normal3f(-0.564038873f, -0.825748205f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Normal3f(-0.503067136f, -0.864247262f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Normal3f(-0.439818412f, -0.898086727f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Normal3f(-0.373481601f, -0.927637637f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Normal3f(-0.305047989f, -0.952336967f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Normal3f(-0.235612959f, -0.971846938f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Normal3f(-0.166289985f, -0.986076891f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Normal3f(-0.098126255f, -0.995173991f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Normal3f(-0.0289524645f, -0.999580801f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Normal3f(0.0471588001f, -0.99888742f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Normal3f(0.134547606f, -0.990907192f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Normal3f(0.228727072f, -0.973490655f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Normal3f(0.323609829f, -0.946190596f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Normal3f(0.416355193f, -0.909202039f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Normal3f(0.504215479f, -0.863577902f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Normal3f(0.584928811f, -0.811084628f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Normal3f(0.656986952f, -0.753901958f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Normal3f(0.740095019f, -0.672502279f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Normal3f(0.826100826f, -0.56352222f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Normal3f(0.891164541f, -0.453680247f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Normal3f(0.937078595f, -0.349118531f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Normal3f(0.967324972f, -0.253539503f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Normal3f(0.985708833f, -0.16845797f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Normal3f(0.995581388f, -0.0939031243f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Translatef(0.696289063f, 0.f, 0.f),
};
static const Call char_80[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Normal3f(0.0026752858f, 0.999996483f, 0.f),
    C_Vertex3f(0.326591492f, 0.640625f, 0.f),
    C_Vertex3f(0.326591492f, 0.640625f, -1.f),
    C_Normal3f(0.0436434634f, 0.99904716f, 0.f),
    C_Vertex3f(0.354740143f, 0.639860153f, 0.f),
    C_Vertex3f(0.354740143f, 0.639860153f, -1.f),
    C_Normal3f(0.0847168267f, 0.996405065f, 0.f),
    C_Vertex3f(0.379116058f, 0.63833046f, 0.f),
    C_Vertex3f(0.379116058f, 0.63833046f, -1.f),
    C_Normal3f(0.130656779f, 0.9914276f, 0.f),
    C_Vertex3f(0.399719238f, 0.636035919f, 0.f),
    C_Vertex3f(0.399719238f, 0.636035919f, -1.f),
    C_Normal3f(0.170454666f, 0.98536551f, 0.f),
    C_Vertex3f(0.40813446f, 0.634506226f, 0.f),
    C_Vertex3f(0.40813446f, 0.634506226f, -1.f),
    C_Normal3f(0.223443985f, 0.974716842f, 0.f),
    C_Vertex3f(0.419940948f, 0.632537842f, 0.f),
    C_Vertex3f(0.419940948f, 0.632537842f, -1.f),
    C_Normal3f(0.303222179f, 0.952919841f, 0.f),
    C_Vertex3f(0.441640377f, 0.626825333f, 0.f),
    C_Vertex3f(0.441640377f, 0.626825333f, -1.f),
    C_Normal3f(0.406447589f, 0.913674116f, 0.f),
    C_Vertex3f(0.461426258f, 0.619337082f, 0.f),
    C_Vertex3f(0.461426258f, 0.619337082f, -1.f),
    C_Normal3f(0.514589965f, 0.857436419f, 0.f),
    C_Vertex3f(0.479298592f, 0.61007309f, 0.f),
    C_Vertex3f(0.479298592f, 0.61007309f, -1.f),
    C_Normal3f(0.624768317f, 0.780810177f, 0.f),
    C_Vertex3f(0.495257378f, 0.599033356f, 0.f),
    C_Vertex3f(0.495257378f, 0.599033356f, -1.f),
    C_Normal3f(0.725484312f, 0.688238621f, 0.f),
    C_Vertex3f(0.509658337f, 0.585780621f, 0.f),
    C_Vertex3f(0.509658337f, 0.585780621f, -1.f),
    C_Normal3f(0.808464885f, 0.588544369f, 0.f),
    C_Vertex3f(0.522501469f, 0.570314884f, 0.f),
    C_Vertex3f(0.522501469f, 0.570314884f, -1.f),
    C_Normal3f(0.872770131f, 0.488131404f, 0.f),
    C_Vertex3f(0.533786774f, 0.552636147f, 0.f),
    C_Vertex3f(0.533786774f, 0.552636147f, -1.f),
    C_Normal3f(0.922784507f, 0.385316521f, 0.f),
    C_Vertex3f(0.543514252f, 0.532744408f, 0.f),
    C_Vertex3f(0.543514252f, 0.532744408f, -1.f),
    C_Normal3f(0.961828709f, 0.273652285f, 0.f),
    C_Vertex3f(0.55080986f, 0.511868f, 0.f),
    C_Vertex3f(0.55080986f, 0.511868f, -1.f),
    C_Normal3f(0.986944973f, 0.161057398f, 0.f),
    C_Vertex3f(0.555673599f, 0.490006924f, 0.f),
    C_Vertex3f(0.555673599f, 0.490006924f, -1.f),
    C_Normal3f(0.998541713f, 0.0539860986f, 0.f),
    C_Vertex3f(0.558105469f, 0.467161179f, 0.f),
    C_Vertex3f(0.558105469f, 0.467161179f, -1.f),
    C_Normal3f(0.999244392f, -0.038867075f, 0.f),
    C_Vertex3f(0.558105469f, 0.445026398f, 0.f),
    C_Vertex3f(0.558105469f, 0.445026398f, -1.f),
    C_Normal3f(0.992083132f, -0.125583053f, 0.f),
    C_Vertex3f(0.55648005f, 0.425372839f, 0.f),
    C_Vertex3f(0.55648005f, 0.425372839f, -1.f),
    C_Normal3f(0.976655483f, -0.214811772f, 0.f),
    C_Vertex3f(0.553229213f, 0.406504869f, 0.f),
    C_Vertex3f(0.553229213f, 0.406504869f, -1.f),
    C_Normal3f(0.951981068f, -0.306156814f, 0.f),
    C_Vertex3f(0.548352957f, 0.388422489f, 0.f),
    C_Vertex3f(0.548352957f, 0.388422489f, -1.f),
    C_Normal3f(0.917767048f, -0.397119135f, 0.f),
    C_Vertex3f(0.541851282f, 0.371125698f, 0.f),
    C_Vertex3f(0.541851282f, 0.371125698f, -1.f),
    C_Normal3f(0.8744995f, -0.485026419f, 0.f),
    C_Vertex3f(0.533724189f, 0.354614496f, 0.f),
    C_Vertex3f(0.533724189f, 0.354614496f, -1.f),
    C_Normal3f(0.823441982f, -0.567400396f, 0.f),
    C_Vertex3f(0.523971677f, 0.338888884f, 0.f),
    C_Vertex3f(0.523971677f, 0.338888884f, -1.f),
    C_Normal3f(0.766457736f, -0.642294884f, 0.f),
    C_Vertex3f(0.512593746f, 0.32394886f, 0.f),
    C_Vertex3f(0.512593746f, 0.32394886f, -1.f),
    C_Normal3f(0.679839313f, -0.733361065f, 0.f),
    C_Vertex3f(0.499590397f, 0.309794426f, 0.f),
    C_Vertex3f(0.499590397f, 0.309794426f, -1.f),
    C_Normal3f(0.561759055f, -0.827300906f, 0.f),
    C_Vertex3f(0.483964801f, 0.297409296f, 0.f),
    C_Vertex3f(0.483964801f, 0.297409296f, -1.f),
    C_Normal3f(0.44544363f, -0.895309985f, 0.f),
    C_Vertex3f(0.465716958f, 0.28679347f, 0.f),
    C_Vertex3f(0.465716958f, 0.28679347f, -1.f),
    C_Normal3f(0.337841511f, -0.941202939f, 0.f),
    C_Vertex3f(0.444846869f, 0.277946949f, 0.f),
    C_Vertex3f(0.444846869f, 0.277946949f, -1.f),
    C_Normal3f(0.242230058f, -0.970218897f, 0.f),
    C_Vertex3f(0.421354532f, 0.270869732f, 0.f),
    C_Vertex3f(0.421354532f, 0.270869732f, -1.f),
    C_Normal3f(0.159224004f, -0.98724246f, 0.f),
    C_Vertex3f(0.395239949f, 0.265561819f, 0.f),
    C_Vertex3f(0.395239949f, 0.265561819f, -1.f),
    C_Normal3f(0.0879815221f, -0.996122122f, 0.f),
    C_Vertex3f(0.366503119f, 0.262023211f, 0.f),
    C_Vertex3f(0.366503119f, 0.262023211f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.335144043f, 0.260253906f, 0.f),
    C_Vertex3f(0.335144043f, 0.260253906f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.335144043f, 0.260253906f, 0.f),
    C_Vertex3f(0.335144043f, 0.260253906f, -1.f),
    C_Vertex3f(0.153808594f, 0.260253906f, -1.f),
    C_Vertex3f(0.153808594f, 0.260253906f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.260253906f, 0.f),
    C_Vertex3f(0.153808594f, 0.260253906f, -1.f),
    C_Vertex3f(0.153808594f, 0.f, -1.f),
    C_Vertex3f(0.153808594f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.153808594f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.153808594f, 0.3359375f, 0.f),
    C_Vertex3f(0.153808594f, 0.3359375f, -1.f),
    C_Normal3f(-0.00490320241f, 0.99998796f, 0.f),
    C_Vertex3f(0.329737663f, 0.3359375f, 0.f),
    C_Vertex3f(0.329737663f, 0.3359375f, -1.f),
    C_Normal3f(-0.0780806988f, 0.99694705f, 0.f),
    C_Vertex3f(0.348792791f, 0.336893559f, 0.f),
    C_Vertex3f(0.348792791f, 0.336893559f, -1.f),
    C_Normal3f(-0.140673816f, 0.990055978f, 0.f),
    C_Vertex3f(0.366358995f, 0.338805676f, 0.f),
    C_Vertex3f(0.366358995f, 0.338805676f, -1.f),
    C_Normal3f(-0.213219538f, 0.97700429f, 0.f),
    C_Vertex3f(0.382436275f, 0.341673851f, 0.f),
    C_Vertex3f(0.382436275f, 0.341673851f, -1.f),
    C_Normal3f(-0.296769321f, 0.9549492f, 0.f),
    C_Vertex3f(0.397024632f, 0.345498085f, 0.f),
    C_Vertex3f(0.397024632f, 0.345498085f, -1.f),
    C_Normal3f(-0.391611129f, 0.920130789f, 0.f),
    C_Vertex3f(0.410124063f, 0.350278378f, 0.f),
    C_Vertex3f(0.410124063f, 0.350278378f, -1.f),
    C_Normal3f(-0.496452749f, 0.868063807f, 0.f),
    C_Vertex3f(0.421734571f, 0.356014729f, 0.f),
    C_Vertex3f(0.421734571f, 0.356014729f, -1.f),
    C_Normal3f(-0.607434034f, 0.794370174f, 0.f),
    C_Vertex3f(0.431856155f, 0.362707138f, 0.f),
    C_Vertex3f(0.431856155f, 0.362707138f, -1.f),
    C_Normal3f(-0.704667389f, 0.709537745f, 0.f),
    C_Vertex3f(0.440488815f, 0.370355606f, 0.f),
    C_Vertex3f(0.440488815f, 0.370355606f, -1.f),
    C_Normal3f(-0.782536387f, 0.622604907f, 0.f),
    C_Vertex3f(0.448042393f, 0.378782272f, 0.f),
    C_Vertex3f(0.448042393f, 0.378782272f, -1.f),
    C_Normal3f(-0.850430369f, 0.526087582f, 0.f),
    C_Vertex3f(0.454516888f, 0.387987137f, 0.f),
    C_Vertex3f(0.454516888f, 0.387987137f, -1.f),
    C_Normal3f(-0.905662835f, 0.423998594f, 0.f),
    C_Vertex3f(0.4599123f, 0.3979702f, 0.f),
    C_Vertex3f(0.4599123f, 0.3979702f, -1.f),
    C_Normal3f(-0.947143197f, 0.320811123f, 0.f),
    C_Vertex3f(0.46422863f, 0.408731461f, 0.f),
    C_Vertex3f(0.46422863f, 0.408731461f, -1.f),
    C_Normal3f(-0.975367844f, 0.220584571f, 0.f),
    C_Vertex3f(0.467465878f, 0.42027092f, 0.f),
    C_Vertex3f(0.467465878f, 0.42027092f, -1.f),
    C_Normal3f(-0.991984189f, 0.126361847f, 0.f),
    C_Vertex3f(0.469624043f, 0.432588577f, 0.f),
    C_Vertex3f(0.469624043f, 0.432588577f, -1.f),
    C_Normal3f(-0.999357402f, 0.0358452089f, 0.f),
    C_Vertex3f(0.470703125f, 0.445684433f, 0.f),
    C_Vertex3f(0.470703125f, 0.445684433f, -1.f),
    C_Normal3f(-0.997469604f, -0.0710953325f, 0.f),
    C_Vertex3f(0.470703125f, 0.462673187f, 0.f),
    C_Vertex3f(0.470703125f, 0.462673187f, -1.f),
    C_Normal3f(-0.978155077f, -0.207876801f, 0.f),
    C_Vertex3f(0.468162537f, 0.481328964f, 0.f),
    C_Vertex3f(0.468162537f, 0.481328964f, -1.f),
    C_Normal3f(-0.933129847f, -0.359539777f, 0.f),
    C_Vertex3f(0.46308136f, 0.498537064f, 0.f),
    C_Vertex3f(0.46308136f, 0.498537064f, -1.f),
    C_Normal3f(-0.860755861f, -0.509018064f, 0.f),
    C_Vertex3f(0.455459595f, 0.514297485f, 0.f),
    C_Vertex3f(0.455459595f, 0.514297485f, -1.f),
    C_Normal3f(-0.766670108f, -0.642041206f, 0.f),
    C_Vertex3f(0.445297241f, 0.528610229f, 0.f),
    C_Vertex3f(0.445297241f, 0.528610229f, -1.f),
    C_Normal3f(-0.646713018f, -0.7627334f, 0.f),
    C_Vertex3f(0.433523178f, 0.540492058f, 0.f),
    C_Vertex3f(0.433523178f, 0.540492058f, -1.f),
    C_Normal3f(-0.50191462f, -0.864917099f, 0.f),
    C_Vertex3f(0.420137405f, 0.54994297f, 0.f),
    C_Vertex3f(0.420137405f, 0.54994297f, -1.f),
    C_Normal3f(-0.371174455f, -0.928563178f, 0.f),
    C_Vertex3f(0.405139923f, 0.556962967f, 0.f),
    C_Vertex3f(0.405139923f, 0.556962967f, -1.f),
    C_Normal3f(-0.26393947f, -0.96453923f, 0.f),
    C_Vertex3f(0.396835327f, 0.559257507f, 0.f),
    C_Vertex3f(0.396835327f, 0.559257507f, -1.f),
    C_Normal3f(-0.208503798f, -0.978021502f, 0.f),
    C_Vertex3f(0.394158363f, 0.559967995f, 0.f),
    C_Vertex3f(0.394158363f, 0.559967995f, -1.f),
    C_Normal3f(-0.161504805f, -0.986871898f, 0.f),
    C_Vertex3f(0.387670517f, 0.561211348f, 0.f),
    C_Vertex3f(0.387670517f, 0.561211348f, -1.f),
    C_Normal3f(-0.118460611f, -0.992958665f, 0.f),
    C_Vertex3f(0.380048752f, 0.562277079f, 0.f),
    C_Vertex3f(0.380048752f, 0.562277079f, -1.f),
    C_Normal3f(-0.0854239091f, -0.996344686f, 0.f),
    C_Vertex3f(0.371293068f, 0.563165188f, 0.f),
    C_Vertex3f(0.371293068f, 0.563165188f, -1.f),
    C_Normal3f(-0.0593484379f, -0.998237312f, 0.f),
    C_Vertex3f(0.361403465f, 0.563875675f, 0.f),
    C_Vertex3f(0.361403465f, 0.563875675f, -1.f),
    C_Normal3f(-0.0382839218f, -0.999266922f, 0.f),
    C_Vertex3f(0.350379944f, 0.564408541f, 0.f),
    C_Vertex3f(0.350379944f, 0.564408541f, -1.f),
    C_Normal3f(-0.0209341403f, -0.999780834f, 0.f),
    C_Vertex3f(0.338222504f, 0.564763784f, 0.f),
    C_Vertex3f(0.338222504f, 0.564763784f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.324931145f, 0.564941406f, 0.f),
    C_Vertex3f(0.324931145f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.324931145f, 0.564941406f, 0.f),
    C_Vertex3f(0.324931145f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.3359375f, -1.f),
    C_Vertex3f(0.153808594f, 0.3359375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.468162537f, 0.481328964f, 0.f),
    C_Vertex3f(0.470703125f, 0.462673187f, 0.f),
    C_Vertex3f(0.479298592f, 0.61007309f, 0.f),
    C_Vertex3f(0.483964801f, 0.297409296f, 0.f),
    C_Vertex3f(0.495257378f, 0.599033356f, 0.f),
    C_Vertex3f(0.499590397f, 0.309794426f, 0.f),
    C_Vertex3f(0.509658337f, 0.585780621f, 0.f),
    C_Vertex3f(0.512593746f, 0.32394886f, 0.f),
    C_Vertex3f(0.522501469f, 0.570314884f, 0.f),
    C_Vertex3f(0.523971677f, 0.338888884f, 0.f),
    C_Vertex3f(0.533786774f, 0.552636147f, 0.f),
    C_Vertex3f(0.533724189f, 0.354614496f, 0.f),
    C_Vertex3f(0.541851282f, 0.371125698f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445297241f, 0.528610229f, 0.f),
    C_Vertex3f(0.455459595f, 0.514297485f, 0.f),
    C_Vertex3f(0.461426258f, 0.619337082f, 0.f),
    C_Vertex3f(0.46308136f, 0.498537064f, 0.f),
    C_Vertex3f(0.479298592f, 0.61007309f, 0.f),
    C_Vertex3f(0.468162537f, 0.481328964f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.441640377f, 0.626825333f, 0.f),
    C_Vertex3f(0.419940948f, 0.632537842f, 0.f),
    C_Vertex3f(0.420137405f, 0.54994297f, 0.f),
    C_Vertex3f(0.433523178f, 0.540492058f, 0.f),
    C_Vertex3f(0.445297241f, 0.528610229f, 0.f),
    C_Vertex3f(0.461426258f, 0.619337082f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.394158363f, 0.559967995f, 0.f),
    C_Vertex3f(0.396835327f, 0.559257507f, 0.f),
    C_Vertex3f(0.399719238f, 0.636035919f, 0.f),
    C_Vertex3f(0.405139923f, 0.556962967f, 0.f),
    C_Vertex3f(0.40813446f, 0.634506226f, 0.f),
    C_Vertex3f(0.420137405f, 0.54994297f, 0.f),
    C_Vertex3f(0.419940948f, 0.632537842f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.361403465f, 0.563875675f, 0.f),
    C_Vertex3f(0.371293068f, 0.563165188f, 0.f),
    C_Vertex3f(0.379116058f, 0.63833046f, 0.f),
    C_Vertex3f(0.380048752f, 0.562277079f, 0.f),
    C_Vertex3f(0.399719238f, 0.636035919f, 0.f),
    C_Vertex3f(0.387670517f, 0.561211348f, 0.f),
    C_Vertex3f(0.394158363f, 0.559967995f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.354740143f, 0.639860153f, 0.f),
    C_Vertex3f(0.326591492f, 0.640625f, 0.f),
    C_Vertex3f(0.338222504f, 0.564763784f, 0.f),
    C_Vertex3f(0.350379944f, 0.564408541f, 0.f),
    C_Vertex3f(0.361403465f, 0.563875675f, 0.f),
    C_Vertex3f(0.379116058f, 0.63833046f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.326591492f, 0.640625f, 0.f),
    C_Vertex3f(0.324931145f, 0.564941406f, 0.f),
    C_Vertex3f(0.338222504f, 0.564763784f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.348792791f, 0.336893559f, 0.f),
    C_Vertex3f(0.329737663f, 0.3359375f, 0.f),
    C_Vertex3f(0.335144043f, 0.260253906f, 0.f),
    C_Vertex3f(0.153808594f, 0.3359375f, 0.f),
    C_Vertex3f(0.153808594f, 0.260253906f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.153808594f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.533786774f, 0.552636147f, 0.f),
    C_Vertex3f(0.541851282f, 0.371125698f, 0.f),
    C_Vertex3f(0.543514252f, 0.532744408f, 0.f),
    C_Vertex3f(0.548352957f, 0.388422489f, 0.f),
    C_Vertex3f(0.55080986f, 0.511868f, 0.f),
    C_Vertex3f(0.553229213f, 0.406504869f, 0.f),
    C_Vertex3f(0.555673599f, 0.490006924f, 0.f),
    C_Vertex3f(0.55648005f, 0.425372839f, 0.f),
    C_Vertex3f(0.558105469f, 0.467161179f, 0.f),
    C_Vertex3f(0.558105469f, 0.445026398f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.483964801f, 0.297409296f, 0.f),
    C_Vertex3f(0.470703125f, 0.462673187f, 0.f),
    C_Vertex3f(0.470703125f, 0.445684433f, 0.f),
    C_Vertex3f(0.469624043f, 0.432588577f, 0.f),
    C_Vertex3f(0.467465878f, 0.42027092f, 0.f),
    C_Vertex3f(0.465716958f, 0.28679347f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.465716958f, 0.28679347f, 0.f),
    C_Vertex3f(0.467465878f, 0.42027092f, 0.f),
    C_Vertex3f(0.46422863f, 0.408731461f, 0.f),
    C_Vertex3f(0.4599123f, 0.3979702f, 0.f),
    C_Vertex3f(0.454516888f, 0.387987137f, 0.f),
    C_Vertex3f(0.448042393f, 0.378782272f, 0.f),
    C_Vertex3f(0.444846869f, 0.277946949f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.444846869f, 0.277946949f, 0.f),
    C_Vertex3f(0.448042393f, 0.378782272f, 0.f),
    C_Vertex3f(0.440488815f, 0.370355606f, 0.f),
    C_Vertex3f(0.431856155f, 0.362707138f, 0.f),
    C_Vertex3f(0.421734571f, 0.356014729f, 0.f),
    C_Vertex3f(0.421354532f, 0.270869732f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.421734571f, 0.356014729f, 0.f),
    C_Vertex3f(0.410124063f, 0.350278378f, 0.f),
    C_Vertex3f(0.421354532f, 0.270869732f, 0.f),
    C_Vertex3f(0.397024632f, 0.345498085f, 0.f),
    C_Vertex3f(0.395239949f, 0.265561819f, 0.f),
    C_Vertex3f(0.382436275f, 0.341673851f, 0.f),
    C_Vertex3f(0.366503119f, 0.262023211f, 0.f),
    C_Vertex3f(0.366358995f, 0.338805676f, 0.f),
    C_Vertex3f(0.348792791f, 0.336893559f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.348792791f, 0.336893559f, 0.f),
    C_Vertex3f(0.335144043f, 0.260253906f, 0.f),
    C_Vertex3f(0.366503119f, 0.262023211f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.468162537f, 0.481328964f, -1.f),
    C_Vertex3f(0.470703125f, 0.462673187f, -1.f),
    C_Vertex3f(0.479298592f, 0.61007309f, -1.f),
    C_Vertex3f(0.483964801f, 0.297409296f, -1.f),
    C_Vertex3f(0.495257378f, 0.599033356f, -1.f),
    C_Vertex3f(0.499590397f, 0.309794426f, -1.f),
    C_Vertex3f(0.509658337f, 0.585780621f, -1.f),
    C_Vertex3f(0.512593746f, 0.32394886f, -1.f),
    C_Vertex3f(0.522501469f, 0.570314884f, -1.f),
    C_Vertex3f(0.523971677f, 0.338888884f, -1.f),
    C_Vertex3f(0.533786774f, 0.552636147f, -1.f),
    C_Vertex3f(0.533724189f, 0.354614496f, -1.f),
    C_Vertex3f(0.541851282f, 0.371125698f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445297241f, 0.528610229f, -1.f),
    C_Vertex3f(0.455459595f, 0.514297485f, -1.f),
    C_Vertex3f(0.461426258f, 0.619337082f, -1.f),
    C_Vertex3f(0.46308136f, 0.498537064f, -1.f),
    C_Vertex3f(0.479298592f, 0.61007309f, -1.f),
    C_Vertex3f(0.468162537f, 0.481328964f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.441640377f, 0.626825333f, -1.f),
    C_Vertex3f(0.419940948f, 0.632537842f, -1.f),
    C_Vertex3f(0.420137405f, 0.54994297f, -1.f),
    C_Vertex3f(0.433523178f, 0.540492058f, -1.f),
    C_Vertex3f(0.445297241f, 0.528610229f, -1.f),
    C_Vertex3f(0.461426258f, 0.619337082f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.394158363f, 0.559967995f, -1.f),
    C_Vertex3f(0.396835327f, 0.559257507f, -1.f),
    C_Vertex3f(0.399719238f, 0.636035919f, -1.f),
    C_Vertex3f(0.405139923f, 0.556962967f, -1.f),
    C_Vertex3f(0.40813446f, 0.634506226f, -1.f),
    C_Vertex3f(0.420137405f, 0.54994297f, -1.f),
    C_Vertex3f(0.419940948f, 0.632537842f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.361403465f, 0.563875675f, -1.f),
    C_Vertex3f(0.371293068f, 0.563165188f, -1.f),
    C_Vertex3f(0.379116058f, 0.63833046f, -1.f),
    C_Vertex3f(0.380048752f, 0.562277079f, -1.f),
    C_Vertex3f(0.399719238f, 0.636035919f, -1.f),
    C_Vertex3f(0.387670517f, 0.561211348f, -1.f),
    C_Vertex3f(0.394158363f, 0.559967995f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.354740143f, 0.639860153f, -1.f),
    C_Vertex3f(0.326591492f, 0.640625f, -1.f),
    C_Vertex3f(0.338222504f, 0.564763784f, -1.f),
    C_Vertex3f(0.350379944f, 0.564408541f, -1.f),
    C_Vertex3f(0.361403465f, 0.563875675f, -1.f),
    C_Vertex3f(0.379116058f, 0.63833046f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.153808594f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.326591492f, 0.640625f, -1.f),
    C_Vertex3f(0.324931145f, 0.564941406f, -1.f),
    C_Vertex3f(0.338222504f, 0.564763784f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.348792791f, 0.336893559f, -1.f),
    C_Vertex3f(0.329737663f, 0.3359375f, -1.f),
    C_Vertex3f(0.335144043f, 0.260253906f, -1.f),
    C_Vertex3f(0.153808594f, 0.3359375f, -1.f),
    C_Vertex3f(0.153808594f, 0.260253906f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.533786774f, 0.552636147f, -1.f),
    C_Vertex3f(0.541851282f, 0.371125698f, -1.f),
    C_Vertex3f(0.543514252f, 0.532744408f, -1.f),
    C_Vertex3f(0.548352957f, 0.388422489f, -1.f),
    C_Vertex3f(0.55080986f, 0.511868f, -1.f),
    C_Vertex3f(0.553229213f, 0.406504869f, -1.f),
    C_Vertex3f(0.555673599f, 0.490006924f, -1.f),
    C_Vertex3f(0.55648005f, 0.425372839f, -1.f),
    C_Vertex3f(0.558105469f, 0.467161179f, -1.f),
    C_Vertex3f(0.558105469f, 0.445026398f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.483964801f, 0.297409296f, -1.f),
    C_Vertex3f(0.470703125f, 0.462673187f, -1.f),
    C_Vertex3f(0.470703125f, 0.445684433f, -1.f),
    C_Vertex3f(0.469624043f, 0.432588577f, -1.f),
    C_Vertex3f(0.467465878f, 0.42027092f, -1.f),
    C_Vertex3f(0.465716958f, 0.28679347f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.465716958f, 0.28679347f, -1.f),
    C_Vertex3f(0.467465878f, 0.42027092f, -1.f),
    C_Vertex3f(0.46422863f, 0.408731461f, -1.f),
    C_Vertex3f(0.4599123f, 0.3979702f, -1.f),
    C_Vertex3f(0.454516888f, 0.387987137f, -1.f),
    C_Vertex3f(0.448042393f, 0.378782272f, -1.f),
    C_Vertex3f(0.444846869f, 0.277946949f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.444846869f, 0.277946949f, -1.f),
    C_Vertex3f(0.448042393f, 0.378782272f, -1.f),
    C_Vertex3f(0.440488815f, 0.370355606f, -1.f),
    C_Vertex3f(0.431856155f, 0.362707138f, -1.f),
    C_Vertex3f(0.421734571f, 0.356014729f, -1.f),
    C_Vertex3f(0.421354532f, 0.270869732f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.421734571f, 0.356014729f, -1.f),
    C_Vertex3f(0.410124063f, 0.350278378f, -1.f),
    C_Vertex3f(0.421354532f, 0.270869732f, -1.f),
    C_Vertex3f(0.397024632f, 0.345498085f, -1.f),
    C_Vertex3f(0.395239949f, 0.265561819f, -1.f),
    C_Vertex3f(0.382436275f, 0.341673851f, -1.f),
    C_Vertex3f(0.366503119f, 0.262023211f, -1.f),
    C_Vertex3f(0.366358995f, 0.338805676f, -1.f),
    C_Vertex3f(0.348792791f, 0.336893559f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.348792791f, 0.336893559f, -1.f),
    C_Vertex3f(0.335144043f, 0.260253906f, -1.f),
    C_Vertex3f(0.366503119f, 0.262023211f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_81[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.568689823f, 0.822552085f, 0.f),
    C_Vertex3f(0.554786682f, 0.0690002441f, 0.f),
    C_Vertex3f(0.554786682f, 0.0690002441f, -1.f),
    C_Normal3f(0.542739391f, 0.839901149f, 0.f),
    C_Vertex3f(0.569532394f, 0.0588054657f, 0.f),
    C_Vertex3f(0.569532394f, 0.0588054657f, -1.f),
    C_Normal3f(0.505729377f, 0.862692177f, 0.f),
    C_Vertex3f(0.597876549f, 0.0411558151f, 0.f),
    C_Vertex3f(0.597876549f, 0.0411558151f, -1.f),
    C_Normal3f(0.453314424f, 0.891350746f, 0.f),
    C_Vertex3f(0.625073433f, 0.0262460709f, 0.f),
    C_Vertex3f(0.625073433f, 0.0262460709f, -1.f),
    C_Normal3f(0.401631474f, 0.915801406f, 0.f),
    C_Vertex3f(0.651123047f, 0.0140762329f, 0.f),
    C_Vertex3f(0.651123047f, 0.0140762329f, -1.f),
    C_Normal3f(0.354135931f, 0.935193956f, 0.f),
    C_Vertex3f(0.663574219f, 0.00936126709f, 0.f),
    C_Vertex3f(0.663574219f, 0.00936126709f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.921422184f, -0.388563007f, 0.f),
    C_Vertex3f(0.663574219f, 0.00936126709f, 0.f),
    C_Vertex3f(0.663574219f, 0.00936126709f, -1.f),
    C_Vertex3f(0.638694763f, -0.0496368408f, -1.f),
    C_Vertex3f(0.638694763f, -0.0496368408f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.339422286f, -0.940634131f, 0.f),
    C_Vertex3f(0.638694763f, -0.0496368408f, 0.f),
    C_Vertex3f(0.638694763f, -0.0496368408f, -1.f),
    C_Normal3f(-0.395755261f, -0.918356061f, 0.f),
    C_Vertex3f(0.621431351f, -0.0434074402f, 0.f),
    C_Vertex3f(0.621431351f, -0.0434074402f, -1.f),
    C_Normal3f(-0.460489333f, -0.887665272f, 0.f),
    C_Vertex3f(0.586958885f, -0.0273418427f, 0.f),
    C_Vertex3f(0.586958885f, -0.0273418427f, -1.f),
    C_Normal3f(-0.529670119f, -0.848203719f, 0.f),
    C_Vertex3f(0.552540779f, -0.00766944885f, 0.f),
    C_Vertex3f(0.552540779f, -0.00766944885f, -1.f),
    C_Normal3f(-0.580435157f, -0.814306557f, 0.f),
    C_Vertex3f(0.518177032f, 0.0156097412f, 0.f),
    C_Vertex3f(0.518177032f, 0.0156097412f, -1.f),
    C_Normal3f(-0.309056163f, -0.951043785f, 0.f),
    C_Vertex3f(0.501022339f, 0.0290527344f, 0.f),
    C_Vertex3f(0.501022339f, 0.0290527344f, -1.f),
    C_Normal3f(0.449553251f, -0.893253565f, 0.f),
    C_Vertex3f(0.492116928f, 0.0240783691f, 0.f),
    C_Vertex3f(0.492116928f, 0.0240783691f, -1.f),
    C_Normal3f(0.400243729f, -0.916408658f, 0.f),
    C_Vertex3f(0.473841429f, 0.01537323f, 0.f),
    C_Vertex3f(0.473841429f, 0.01537323f, -1.f),
    C_Normal3f(0.339142859f, -0.940734923f, 0.f),
    C_Vertex3f(0.455101252f, 0.00791168213f, 0.f),
    C_Vertex3f(0.455101252f, 0.00791168213f, -1.f),
    C_Normal3f(0.276671231f, -0.96096462f, 0.f),
    C_Vertex3f(0.435896397f, 0.00169372559f, 0.f),
    C_Vertex3f(0.435896397f, 0.00169372559f, -1.f),
    C_Normal3f(0.213651687f, -0.976909935f, 0.f),
    C_Vertex3f(0.416226864f, -0.00328063965f, 0.f),
    C_Vertex3f(0.416226864f, -0.00328063965f, -1.f),
    C_Normal3f(0.150903121f, -0.988548517f, 0.f),
    C_Vertex3f(0.396092653f, -0.00701141357f, 0.f),
    C_Vertex3f(0.396092653f, -0.00701141357f, -1.f),
    C_Normal3f(0.0891907439f, -0.996014595f, 0.f),
    C_Vertex3f(0.375493765f, -0.00949859619f, 0.f),
    C_Vertex3f(0.375493765f, -0.00949859619f, -1.f),
    C_Normal3f(0.0291112158f, -0.999576211f, 0.f),
    C_Vertex3f(0.354430199f, -0.0107421875f, 0.f),
    C_Vertex3f(0.354430199f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0305462368f, -0.999533355f, 0.f),
    C_Vertex3f(0.332793236f, -0.0107421875f, 0.f),
    C_Vertex3f(0.332793236f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0932440013f, -0.995643258f, 0.f),
    C_Vertex3f(0.311553001f, -0.00943183899f, 0.f),
    C_Vertex3f(0.311553001f, -0.00943183899f, -1.f),
    C_Normal3f(-0.157931313f, -0.987450004f, 0.f),
    C_Vertex3f(0.290818214f, -0.00681114197f, 0.f),
    C_Vertex3f(0.290818214f, -0.00681114197f, -1.f),
    C_Normal3f(-0.22375834f, -0.974644601f, 0.f),
    C_Vertex3f(0.270588875f, -0.00288009644f, 0.f),
    C_Vertex3f(0.270588875f, -0.00288009644f, -1.f),
    C_Normal3f(-0.289836943f, -0.957076013f, 0.f),
    C_Vertex3f(0.250864983f, 0.00236129761f, 0.f),
    C_Vertex3f(0.250864983f, 0.00236129761f, -1.f),
    C_Normal3f(-0.355215311f, -0.934784532f, 0.f),
    C_Vertex3f(0.231646538f, 0.00891304016f, 0.f),
    C_Vertex3f(0.231646538f, 0.00891304016f, -1.f),
    C_Normal3f(-0.418942183f, -0.908012927f, 0.f),
    C_Vertex3f(0.21293354f, 0.0167751312f, 0.f),
    C_Vertex3f(0.21293354f, 0.0167751312f, -1.f),
    C_Normal3f(-0.480134517f, -0.877194822f, 0.f),
    C_Vertex3f(0.19472599f, 0.0259475708f, 0.f),
    C_Vertex3f(0.19472599f, 0.0259475708f, -1.f),
    C_Normal3f(-0.539449513f, -0.842017949f, 0.f),
    C_Vertex3f(0.177023888f, 0.0364303589f, 0.f),
    C_Vertex3f(0.177023888f, 0.0364303589f, -1.f),
    C_Normal3f(-0.598349154f, -0.801235497f, 0.f),
    C_Vertex3f(0.160339355f, 0.0479778051f, 0.f),
    C_Vertex3f(0.160339355f, 0.0479778051f, -1.f),
    C_Normal3f(-0.655139625f, -0.755507827f, 0.f),
    C_Vertex3f(0.144672394f, 0.0605899096f, 0.f),
    C_Vertex3f(0.144672394f, 0.0605899096f, -1.f),
    C_Normal3f(-0.708812952f, -0.705396533f, 0.f),
    C_Vertex3f(0.130023003f, 0.0742666721f, 0.f),
    C_Vertex3f(0.130023003f, 0.0742666721f, -1.f),
    C_Normal3f(-0.758489311f, -0.651685536f, 0.f),
    C_Vertex3f(0.116391182f, 0.0890080929f, 0.f),
    C_Vertex3f(0.116391182f, 0.0890080929f, -1.f),
    C_Normal3f(-0.803488016f, -0.59532088f, 0.f),
    C_Vertex3f(0.103776932f, 0.104814172f, 0.f),
    C_Vertex3f(0.103776932f, 0.104814172f, -1.f),
    C_Normal3f(-0.843371272f, -0.537331283f, 0.f),
    C_Vertex3f(0.0921802521f, 0.121684909f, 0.f),
    C_Vertex3f(0.0921802521f, 0.121684909f, -1.f),
    C_Normal3f(-0.877955139f, -0.478742927f, 0.f),
    C_Vertex3f(0.0816011429f, 0.139620304f, 0.f),
    C_Vertex3f(0.0816011429f, 0.139620304f, -1.f),
    C_Normal3f(-0.906953335f, -0.421231031f, 0.f),
    C_Vertex3f(0.0720396042f, 0.158620358f, 0.f),
    C_Vertex3f(0.0720396042f, 0.158620358f, -1.f),
    C_Normal3f(-0.931536257f, -0.363648415f, 0.f),
    C_Vertex3f(0.0636732578f, 0.178220868f, 0.f),
    C_Vertex3f(0.0636732578f, 0.178220868f, -1.f),
    C_Normal3f(-0.952246428f, -0.305330813f, 0.f),
    C_Vertex3f(0.0565021038f, 0.198421836f, 0.f),
    C_Vertex3f(0.0565021038f, 0.198421836f, -1.f),
    C_Normal3f(-0.969019949f, -0.246982694f, 0.f),
    C_Vertex3f(0.0505261421f, 0.219223261f, 0.f),
    C_Vertex3f(0.0505261421f, 0.219223261f, -1.f),
    C_Normal3f(-0.98192507f, -0.189270183f, 0.f),
    C_Vertex3f(0.0457453728f, 0.240625143f, 0.f),
    C_Vertex3f(0.0457453728f, 0.240625143f, -1.f),
    C_Normal3f(-0.991144478f, -0.13278833f, 0.f),
    C_Vertex3f(0.0421597958f, 0.262627482f, 0.f),
    C_Vertex3f(0.0421597958f, 0.262627482f, -1.f),
    C_Normal3f(-0.996950388f, -0.0780386552f, 0.f),
    C_Vertex3f(0.0397694111f, 0.285230279f, 0.f),
    C_Vertex3f(0.0397694111f, 0.285230279f, -1.f),
    C_Normal3f(-0.999676168f, -0.0254465323f, 0.f),
    C_Vertex3f(0.0385742188f, 0.308433533f, 0.f),
    C_Vertex3f(0.0385742188f, 0.308433533f, -1.f),
    C_Normal3f(-0.999671876f, 0.0256122909f, 0.f),
    C_Vertex3f(0.0385742188f, 0.332183838f, 0.f),
    C_Vertex3f(0.0385742188f, 0.332183838f, -1.f),
    C_Normal3f(-0.996914327f, 0.0784962326f, 0.f),
    C_Vertex3f(0.039776206f, 0.355348229f, 0.f),
    C_Vertex3f(0.039776206f, 0.355348229f, -1.f),
    C_Normal3f(-0.991094649f, 0.133160025f, 0.f),
    C_Vertex3f(0.0421801805f, 0.377980113f, 0.f),
    C_Vertex3f(0.0421801805f, 0.377980113f, -1.f),
    C_Normal3f(-0.981937349f, 0.189206421f, 0.f),
    C_Vertex3f(0.0457861423f, 0.400079489f, 0.f),
    C_Vertex3f(0.0457861423f, 0.400079489f, -1.f),
    C_Normal3f(-0.969238818f, 0.246122345f, 0.f),
    C_Vertex3f(0.0505940914f, 0.421646357f, 0.f),
    C_Vertex3f(0.0505940914f, 0.421646357f, -1.f),
    C_Normal3f(-0.952887297f, 0.303324312f, 0.f),
    C_Vertex3f(0.0566040277f, 0.442680717f, 0.f),
    C_Vertex3f(0.0566040277f, 0.442680717f, -1.f),
    C_Normal3f(-0.932881474f, 0.360183358f, 0.f),
    C_Vertex3f(0.0638159513f, 0.463182569f, 0.f),
    C_Vertex3f(0.0638159513f, 0.463182569f, -1.f),
    C_Normal3f(-0.909338653f, 0.416056633f, 0.f),
    C_Vertex3f(0.0722298622f, 0.483151913f, 0.f),
    C_Vertex3f(0.0722298622f, 0.483151913f, -1.f),
    C_Normal3f(-0.881034672f, 0.473051757f, 0.f),
    C_Vertex3f(0.0818457603f, 0.502588749f, 0.f),
    C_Vertex3f(0.0818457603f, 0.502588749f, -1.f),
    C_Normal3f(-0.846231759f, 0.532815039f, 0.f),
    C_Vertex3f(0.092479229f, 0.520865321f, 0.f),
    C_Vertex3f(0.092479229f, 0.520865321f, -1.f),
    C_Normal3f(-0.805630803f, 0.592418015f, 0.f),
    C_Vertex3f(0.104130268f, 0.537981629f, 0.f),
    C_Vertex3f(0.104130268f, 0.537981629f, -1.f),
    C_Normal3f(-0.759304404f, 0.650735617f, 0.f),
    C_Vertex3f(0.116798878f, 0.553937674f, 0.f),
    C_Vertex3f(0.116798878f, 0.553937674f, -1.f),
    C_Normal3f(-0.707621336f, 0.706591785f, 0.f),
    C_Vertex3f(0.130485058f, 0.568733454f, 0.f),
    C_Vertex3f(0.130485058f, 0.568733454f, -1.f),
    C_Normal3f(-0.651253343f, 0.758860409f, 0.f),
    C_Vertex3f(0.145188808f, 0.58236897f, 0.f),
    C_Vertex3f(0.145188808f, 0.58236897f, -1.f),
    C_Normal3f(-0.59113878f, 0.806569874f, 0.f),
    C_Vertex3f(0.16091013f, 0.594844222f, 0.f),
    C_Vertex3f(0.16091013f, 0.594844222f, -1.f),
    C_Normal3f(-0.528407216f, 0.848991096f, 0.f),
    C_Vertex3f(0.177649021f, 0.60615921f, 0.f),
    C_Vertex3f(0.177649021f, 0.60615921f, -1.f),
    C_Normal3f(-0.467135668f, 0.884185612f, 0.f),
    C_Vertex3f(0.195405483f, 0.616313934f, 0.f),
    C_Vertex3f(0.195405483f, 0.616313934f, -1.f),
    C_Normal3f(-0.406483948f, 0.913658023f, 0.f),
    C_Vertex3f(0.213687778f, 0.625199318f, 0.f),
    C_Vertex3f(0.213687778f, 0.625199318f, -1.f),
    C_Normal3f(-0.343762279f, 0.939056695f, 0.f),
    C_Vertex3f(0.232495904f, 0.632815361f, 0.f),
    C_Vertex3f(0.232495904f, 0.632815361f, -1.f),
    C_Normal3f(-0.279831916f, 0.960049033f, 0.f),
    C_Vertex3f(0.251829863f, 0.639162064f, 0.f),
    C_Vertex3f(0.251829863f, 0.639162064f, -1.f),
    C_Normal3f(-0.215588257f, 0.976484418f, 0.f),
    C_Vertex3f(0.271689653f, 0.644239426f, 0.f),
    C_Vertex3f(0.271689653f, 0.644239426f, -1.f),
    C_Normal3f(-0.151900679f, 0.988395751f, 0.f),
    C_Vertex3f(0.292075276f, 0.648047447f, 0.f),
    C_Vertex3f(0.292075276f, 0.648047447f, -1.f),
    C_Normal3f(-0.08955919f, 0.995981514f, 0.f),
    C_Vertex3f(0.312986732f, 0.650586128f, 0.f),
    C_Vertex3f(0.312986732f, 0.650586128f, -1.f),
    C_Normal3f(-0.0291610751f, 0.999574721f, 0.f),
    C_Vertex3f(0.334424019f, 0.651855469f, 0.f),
    C_Vertex3f(0.334424019f, 0.651855469f, -1.f),
    C_Normal3f(0.0301099457f, 0.999546647f, 0.f),
    C_Vertex3f(0.356496811f, 0.651855469f, 0.f),
    C_Vertex3f(0.356496811f, 0.651855469f, -1.f),
    C_Normal3f(0.0920579284f, 0.995753646f, 0.f),
    C_Vertex3f(0.37813282f, 0.650538802f, 0.f),
    C_Vertex3f(0.37813282f, 0.650538802f, -1.f),
    C_Normal3f(0.156188414f, 0.987727284f, 0.f),
    C_Vertex3f(0.399222374f, 0.647905469f, 0.f),
    C_Vertex3f(0.399222374f, 0.647905469f, -1.f),
    C_Normal3f(0.221691743f, 0.975116789f, 0.f),
    C_Vertex3f(0.419765472f, 0.643955469f, 0.f),
    C_Vertex3f(0.419765472f, 0.643955469f, -1.f),
    C_Normal3f(0.287703127f, 0.957719564f, 0.f),
    C_Vertex3f(0.439762115f, 0.638688803f, 0.f),
    C_Vertex3f(0.439762115f, 0.638688803f, -1.f),
    C_Normal3f(0.353274047f, 0.935519814f, 0.f),
    C_Vertex3f(0.459212303f, 0.63210547f, 0.f),
    C_Vertex3f(0.459212303f, 0.63210547f, -1.f),
    C_Normal3f(0.417434603f, 0.908707023f, 0.f),
    C_Vertex3f(0.478116035f, 0.62420547f, 0.f),
    C_Vertex3f(0.478116035f, 0.62420547f, -1.f),
    C_Normal3f(0.479262501f, 0.877671599f, 0.f),
    C_Vertex3f(0.496473312f, 0.614988804f, 0.f),
    C_Vertex3f(0.496473312f, 0.614988804f, -1.f),
    C_Normal3f(0.538996637f, 0.842307925f, 0.f),
    C_Vertex3f(0.514284134f, 0.604455471f, 0.f),
    C_Vertex3f(0.514284134f, 0.604455471f, -1.f),
    C_Normal3f(0.598155022f, 0.801380396f, 0.f),
    C_Vertex3f(0.531036496f, 0.592871666f, 0.f),
    C_Vertex3f(0.531036496f, 0.592871666f, -1.f),
    C_Normal3f(0.655408144f, 0.755274892f, 0.f),
    C_Vertex3f(0.546730399f, 0.580237389f, 0.f),
    C_Vertex3f(0.546730399f, 0.580237389f, -1.f),
    C_Normal3f(0.709693372f, 0.704510689f, 0.f),
    C_Vertex3f(0.561365843f, 0.566552639f, 0.f),
    C_Vertex3f(0.561365843f, 0.566552639f, -1.f),
    C_Normal3f(0.760059237f, 0.649853766f, 0.f),
    C_Vertex3f(0.574942827f, 0.551817417f, 0.f),
    C_Vertex3f(0.574942827f, 0.551817417f, -1.f),
    C_Normal3f(0.805746138f, 0.592261076f, 0.f),
    C_Vertex3f(0.587461352f, 0.536031723f, 0.f),
    C_Vertex3f(0.587461352f, 0.536031723f, -1.f),
    C_Normal3f(0.846242011f, 0.532798707f, 0.f),
    C_Vertex3f(0.598921418f, 0.519195557f, 0.f),
    C_Vertex3f(0.598921418f, 0.519195557f, -1.f),
    C_Normal3f(0.881303191f, 0.472551405f, 0.f),
    C_Vertex3f(0.609323025f, 0.501308918f, 0.f),
    C_Vertex3f(0.609323025f, 0.501308918f, -1.f),
    C_Normal3f(0.910126865f, 0.414329708f, 0.f),
    C_Vertex3f(0.618666172f, 0.482371807f, 0.f),
    C_Vertex3f(0.618666172f, 0.482371807f, -1.f),
    C_Normal3f(0.934013546f, 0.357237577f, 0.f),
    C_Vertex3f(0.626841426f, 0.462827563f, 0.f),
    C_Vertex3f(0.626841426f, 0.462827563f, -1.f),
    C_Normal3f(0.95406419f, 0.29960236f, 0.f),
    C_Vertex3f(0.633848786f, 0.442676187f, 0.f),
    C_Vertex3f(0.633848786f, 0.442676187f, -1.f),
    C_Normal3f(0.970250785f, 0.242102087f, 0.f),
    C_Vertex3f(0.639688253f, 0.421917677f, 0.f),
    C_Vertex3f(0.639688253f, 0.421917677f, -1.f),
    C_Normal3f(0.982669055f, 0.185368314f, 0.f),
    C_Vertex3f(0.644359827f, 0.400552034f, 0.f),
    C_Vertex3f(0.644359827f, 0.400552034f, -1.f),
    C_Normal3f(0.99151963f, 0.129957289f, 0.f),
    C_Vertex3f(0.647863507f, 0.378579259f, 0.f),
    C_Vertex3f(0.647863507f, 0.378579259f, -1.f),
    C_Normal3f(0.997082472f, 0.076331459f, 0.f),
    C_Vertex3f(0.650199294f, 0.355999351f, 0.f),
    C_Vertex3f(0.650199294f, 0.355999351f, -1.f),
    C_Normal3f(0.999773324f, 0.0212913789f, 0.f),
    C_Vertex3f(0.651367188f, 0.332812309f, 0.f),
    C_Vertex3f(0.651367188f, 0.332812309f, -1.f),
    C_Normal3f(0.999058068f, -0.0433938876f, 0.f),
    C_Vertex3f(0.651367188f, 0.301158905f, 0.f),
    C_Vertex3f(0.651367188f, 0.301158905f, -1.f),
    C_Normal3f(0.992472947f, -0.122463778f, 0.f),
    C_Vertex3f(0.648362637f, 0.263638496f, 0.f),
    C_Vertex3f(0.648362637f, 0.263638496f, -1.f),
    C_Normal3f(0.977150381f, -0.212548658f, 0.f),
    C_Vertex3f(0.642353535f, 0.228110313f, 0.f),
    C_Vertex3f(0.642353535f, 0.228110313f, -1.f),
    C_Normal3f(0.951544344f, -0.307511359f, 0.f),
    C_Vertex3f(0.633339882f, 0.194574356f, 0.f),
    C_Vertex3f(0.633339882f, 0.194574356f, -1.f),
    C_Normal3f(0.914000094f, -0.405713737f, 0.f),
    C_Vertex3f(0.621321678f, 0.163030624f, 0.f),
    C_Vertex3f(0.621321678f, 0.163030624f, -1.f),
    C_Normal3f(0.863820851f, -0.503799081f, 0.f),
    C_Vertex3f(0.606244564f, 0.133533478f, 0.f),
    C_Vertex3f(0.606244564f, 0.133533478f, -1.f),
    C_Normal3f(0.802252471f, -0.596984863f, 0.f),
    C_Vertex3f(0.58810854f, 0.106082916f, 0.f),
    C_Vertex3f(0.58810854f, 0.106082916f, -1.f),
    C_Normal3f(0.743817985f, -0.668382168f, 0.f),
    C_Vertex3f(0.566913605f, 0.0806789398f, 0.f),
    C_Vertex3f(0.566913605f, 0.0806789398f, -1.f),
    C_Normal3f(0.693669975f, -0.720292985f, 0.f),
    C_Vertex3f(0.554786682f, 0.0690002441f, 0.f),
    C_Vertex3f(0.554786682f, 0.0690002441f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.269351542f, 0.963041902f, 0.f),
    C_Vertex3f(0.368141174f, 0.177352905f, 0.f),
    C_Vertex3f(0.368141174f, 0.177352905f, -1.f),
    C_Normal3f(0.297518522f, 0.954716027f, 0.f),
    C_Vertex3f(0.37732029f, 0.174785614f, 0.f),
    C_Vertex3f(0.37732029f, 0.174785614f, -1.f),
    C_Normal3f(0.334019363f, 0.942566156f, 0.f),
    C_Vertex3f(0.394872665f, 0.16902256f, 0.f),
    C_Vertex3f(0.394872665f, 0.16902256f, -1.f),
    C_Normal3f(0.379590124f, 0.925154805f, 0.f),
    C_Vertex3f(0.411619186f, 0.162631035f, 0.f),
    C_Vertex3f(0.411619186f, 0.162631035f, -1.f),
    C_Normal3f(0.426861942f, 0.904316783f, 0.f),
    C_Vertex3f(0.427559853f, 0.155611038f, 0.f),
    C_Vertex3f(0.427559853f, 0.155611038f, -1.f),
    C_Normal3f(0.475495011f, 0.879718423f, 0.f),
    C_Vertex3f(0.442694664f, 0.14796257f, 0.f),
    C_Vertex3f(0.442694664f, 0.14796257f, -1.f),
    C_Normal3f(0.525041759f, 0.851076424f, 0.f),
    C_Vertex3f(0.457023621f, 0.139685631f, 0.f),
    C_Vertex3f(0.457023621f, 0.139685631f, -1.f),
    C_Normal3f(0.574948192f, 0.8181898f, 0.f),
    C_Vertex3f(0.470546722f, 0.13078022f, 0.f),
    C_Vertex3f(0.470546722f, 0.13078022f, -1.f),
    C_Normal3f(0.61634624f, 0.787475288f, 0.f),
    C_Vertex3f(0.483263969f, 0.121246338f, 0.f),
    C_Vertex3f(0.483263969f, 0.121246338f, -1.f),
    C_Normal3f(-0.21908088f, 0.975706697f, 0.f),
    C_Vertex3f(0.489219666f, 0.116165161f, 0.f),
    C_Vertex3f(0.489219666f, 0.116165161f, -1.f),
    C_Normal3f(-0.733057141f, 0.680167079f, 0.f),
    C_Vertex3f(0.498562813f, 0.124681473f, 0.f),
    C_Vertex3f(0.498562813f, 0.124681473f, -1.f),
    C_Normal3f(-0.800044596f, 0.599940419f, 0.f),
    C_Vertex3f(0.514913321f, 0.143856764f, 0.f),
    C_Vertex3f(0.514913321f, 0.143856764f, -1.f),
    C_Normal3f(-0.867358148f, 0.49768436f, 0.f),
    C_Vertex3f(0.528928041f, 0.165174723f, 0.f),
    C_Vertex3f(0.528928041f, 0.165174723f, -1.f),
    C_Normal3f(-0.919181049f, 0.393835187f, 0.f),
    C_Vertex3f(0.540606976f, 0.188635349f, 0.f),
    C_Vertex3f(0.540606976f, 0.188635349f, -1.f),
    C_Normal3f(-0.95610404f, 0.293027252f, 0.f),
    C_Vertex3f(0.549950123f, 0.214238644f, 0.f),
    C_Vertex3f(0.549950123f, 0.214238644f, -1.f),
    C_Normal3f(-0.980080366f, 0.198601112f, 0.f),
    C_Vertex3f(0.556957483f, 0.241984606f, 0.f),
    C_Vertex3f(0.556957483f, 0.241984606f, -1.f),
    C_Normal3f(-0.993657351f, 0.112450317f, 0.f),
    C_Vertex3f(0.561629057f, 0.271873236f, 0.f),
    C_Vertex3f(0.561629057f, 0.271873236f, -1.f),
    C_Normal3f(-0.999212921f, 0.0396692045f, 0.f),
    C_Vertex3f(0.563964844f, 0.303904533f, 0.f),
    C_Vertex3f(0.563964844f, 0.303904533f, -1.f),
    C_Normal3f(-0.999836028f, -0.0181120541f, 0.f),
    C_Vertex3f(0.563964844f, 0.330708504f, 0.f),
    C_Vertex3f(0.563964844f, 0.330708504f, -1.f),
    C_Normal3f(-0.997765362f, -0.0668153763f, 0.f),
    C_Vertex3f(0.563138366f, 0.349528432f, 0.f),
    C_Vertex3f(0.563138366f, 0.349528432f, -1.f),
    C_Normal3f(-0.993402958f, -0.114675716f, 0.f),
    C_Vertex3f(0.56148541f, 0.367734313f, 0.f),
    C_Vertex3f(0.56148541f, 0.367734313f, -1.f),
    C_Normal3f(-0.986283839f, -0.165058017f, 0.f),
    C_Vertex3f(0.559005976f, 0.385326147f, 0.f),
    C_Vertex3f(0.559005976f, 0.385326147f, -1.f),
    C_Normal3f(-0.976006508f, -0.2177414f, 0.f),
    C_Vertex3f(0.555700064f, 0.402303934f, 0.f),
    C_Vertex3f(0.555700064f, 0.402303934f, -1.f),
    C_Normal3f(-0.962185979f, -0.272393197f, 0.f),
    C_Vertex3f(0.551567674f, 0.418667674f, 0.f),
    C_Vertex3f(0.551567674f, 0.418667674f, -1.f),
    C_Normal3f(-0.944482684f, -0.328561187f, 0.f),
    C_Vertex3f(0.546608806f, 0.434417367f, 0.f),
    C_Vertex3f(0.546608806f, 0.434417367f, -1.f),
    C_Normal3f(-0.922634721f, -0.385675043f, 0.f),
    C_Vertex3f(0.54082346f, 0.449553013f, 0.f),
    C_Vertex3f(0.54082346f, 0.449553013f, -1.f),
    C_Normal3f(-0.896034777f, -0.443984032f, 0.f),
    C_Vertex3f(0.534211636f, 0.464074612f, 0.f),
    C_Vertex3f(0.534211636f, 0.464074612f, -1.f),
    C_Normal3f(-0.863815427f, -0.503808498f, 0.f),
    C_Vertex3f(0.526834965f, 0.477784157f, 0.f),
    C_Vertex3f(0.526834965f, 0.477784157f, -1.f),
    C_Normal3f(-0.826010764f, -0.563654244f, 0.f),
    C_Vertex3f(0.518693447f, 0.490681648f, 0.f),
    C_Vertex3f(0.518693447f, 0.490681648f, -1.f),
    C_Normal3f(-0.782655001f, -0.622455716f, 0.f),
    C_Vertex3f(0.509787083f, 0.502767086f, 0.f),
    C_Vertex3f(0.509787083f, 0.502767086f, -1.f),
    C_Normal3f(-0.734054685f, -0.679090381f, 0.f),
    C_Vertex3f(0.500115871f, 0.51404047f, 0.f),
    C_Vertex3f(0.500115871f, 0.51404047f, -1.f),
    C_Normal3f(-0.680798948f, -0.732470274f, 0.f),
    C_Vertex3f(0.489679813f, 0.524501801f, 0.f),
    C_Vertex3f(0.489679813f, 0.524501801f, -1.f),
    C_Normal3f(-0.623731196f, -0.781638861f, 0.f),
    C_Vertex3f(0.478478909f, 0.534151077f, 0.f),
    C_Vertex3f(0.478478909f, 0.534151077f, -1.f),
    C_Normal3f(-0.563883901f, -0.825854003f, 0.f),
    C_Vertex3f(0.466513157f, 0.5429883f, 0.f),
    C_Vertex3f(0.466513157f, 0.5429883f, -1.f),
    C_Normal3f(-0.503067136f, -0.864247262f, 0.f),
    C_Vertex3f(0.453782558f, 0.55101347f, 0.f),
    C_Vertex3f(0.453782558f, 0.55101347f, -1.f),
    C_Normal3f(-0.439818412f, -0.898086727f, 0.f),
    C_Vertex3f(0.440662742f, 0.558035493f, 0.f),
    C_Vertex3f(0.440662742f, 0.558035493f, -1.f),
    C_Normal3f(-0.373481601f, -0.927637637f, 0.f),
    C_Vertex3f(0.427153707f, 0.56405437f, 0.f),
    C_Vertex3f(0.427153707f, 0.56405437f, -1.f),
    C_Normal3f(-0.305047989f, -0.952336967f, 0.f),
    C_Vertex3f(0.413255453f, 0.569070101f, 0.f),
    C_Vertex3f(0.413255453f, 0.569070101f, -1.f),
    C_Normal3f(-0.235612959f, -0.971846938f, 0.f),
    C_Vertex3f(0.398967981f, 0.573082685f, 0.f),
    C_Vertex3f(0.398967981f, 0.573082685f, -1.f),
    C_Normal3f(-0.166289985f, -0.986076891f, 0.f),
    C_Vertex3f(0.384291291f, 0.576092124f, 0.f),
    C_Vertex3f(0.384291291f, 0.576092124f, -1.f),
    C_Normal3f(-0.098126255f, -0.995173991f, 0.f),
    C_Vertex3f(0.369225383f, 0.578098416f, 0.f),
    C_Vertex3f(0.369225383f, 0.578098416f, -1.f),
    C_Normal3f(-0.0284591708f, -0.999594986f, 0.f),
    C_Vertex3f(0.353770256f, 0.579101563f, 0.f),
    C_Vertex3f(0.353770256f, 0.579101563f, -1.f),
    C_Normal3f(0.0476521067f, -0.998863995f, 0.f),
    C_Vertex3f(0.333991051f, 0.579101563f, 0.f),
    C_Vertex3f(0.333991051f, 0.579101563f, -1.f),
    C_Normal3f(0.135685161f, -0.990751982f, 0.f),
    C_Vertex3f(0.31128788f, 0.577074885f, 0.f),
    C_Vertex3f(0.31128788f, 0.577074885f, -1.f),
    C_Normal3f(0.232585236f, -0.972575903f, 0.f),
    C_Vertex3f(0.289595604f, 0.573021531f, 0.f),
    C_Vertex3f(0.289595604f, 0.573021531f, -1.f),
    C_Normal3f(0.331674427f, -0.943393826f, 0.f),
    C_Vertex3f(0.268914223f, 0.5669415f, 0.f),
    C_Vertex3f(0.268914223f, 0.5669415f, -1.f),
    C_Normal3f(0.429696918f, -0.902973175f, 0.f),
    C_Vertex3f(0.249243736f, 0.558834791f, 0.f),
    C_Vertex3f(0.249243736f, 0.558834791f, -1.f),
    C_Normal3f(0.523244083f, -0.852182925f, 0.f),
    C_Vertex3f(0.230584145f, 0.548701406f, 0.f),
    C_Vertex3f(0.230584145f, 0.548701406f, -1.f),
    C_Normal3f(0.609312177f, -0.792930424f, 0.f),
    C_Vertex3f(0.212935448f, 0.536541343f, 0.f),
    C_Vertex3f(0.212935448f, 0.536541343f, -1.f),
    C_Normal3f(0.685763776f, -0.727824211f, 0.f),
    C_Vertex3f(0.196297646f, 0.522354603f, 0.f),
    C_Vertex3f(0.196297646f, 0.522354603f, -1.f),
    C_Normal3f(0.761523604f, -0.648137212f, 0.f),
    C_Vertex3f(0.180670738f, 0.506141186f, 0.f),
    C_Vertex3f(0.180670738f, 0.506141186f, -1.f),
    C_Normal3f(0.834279776f, -0.551341355f, 0.f),
    C_Vertex3f(0.166997194f, 0.487928271f, 0.f),
    C_Vertex3f(0.166997194f, 0.487928271f, -1.f),
    C_Normal3f(0.892103493f, -0.451831043f, 0.f),
    C_Vertex3f(0.155277014f, 0.467715859f, 0.f),
    C_Vertex3f(0.155277014f, 0.467715859f, -1.f),
    C_Normal3f(0.935188234f, -0.354150981f, 0.f),
    C_Vertex3f(0.145510197f, 0.44550395f, 0.f),
    C_Vertex3f(0.145510197f, 0.44550395f, -1.f),
    C_Normal3f(0.965141773f, -0.261727571f, 0.f),
    C_Vertex3f(0.137696743f, 0.421292543f, 0.f),
    C_Vertex3f(0.137696743f, 0.421292543f, -1.f),
    C_Normal3f(0.984274685f, -0.176644608f, 0.f),
    C_Vertex3f(0.131836653f, 0.395081639f, 0.f),
    C_Vertex3f(0.131836653f, 0.395081639f, -1.f),
    C_Normal3f(0.99500674f, -0.0998082757f, 0.f),
    C_Vertex3f(0.127929926f, 0.366871238f, 0.f),
    C_Vertex3f(0.127929926f, 0.366871238f, -1.f),
    C_Normal3f(0.999503016f, -0.0315267369f, 0.f),
    C_Vertex3f(0.125976563f, 0.336661339f, 0.f),
    C_Vertex3f(0.125976563f, 0.336661339f, -1.f),
    C_Normal3f(0.999500513f, 0.0316030197f, 0.f),
    C_Vertex3f(0.125976563f, 0.304943085f, 0.f),
    C_Vertex3f(0.125976563f, 0.304943085f, -1.f),
    C_Normal3f(0.994866848f, 0.101192571f, 0.f),
    C_Vertex3f(0.127909422f, 0.275531292f, 0.f),
    C_Vertex3f(0.127909422f, 0.275531292f, -1.f),
    C_Normal3f(0.984001815f, 0.178158641f, 0.f),
    C_Vertex3f(0.131775141f, 0.247934818f, 0.f),
    C_Vertex3f(0.131775141f, 0.247934818f, -1.f),
    C_Normal3f(0.964947462f, 0.262443095f, 0.f),
    C_Vertex3f(0.137573719f, 0.222153664f, 0.f),
    C_Vertex3f(0.137573719f, 0.222153664f, -1.f),
    C_Normal3f(0.935646355f, 0.35293895f, 0.f),
    C_Vertex3f(0.145305157f, 0.198187828f, 0.f),
    C_Vertex3f(0.145305157f, 0.198187828f, -1.f),
    C_Normal3f(0.894270599f, 0.447526723f, 0.f),
    C_Vertex3f(0.154969454f, 0.176037312f, 0.f),
    C_Vertex3f(0.154969454f, 0.176037312f, -1.f),
    C_Normal3f(0.839711547f, 0.543032706f, 0.f),
    C_Vertex3f(0.16656661f, 0.155702114f, 0.f),
    C_Vertex3f(0.16656661f, 0.155702114f, -1.f),
    C_Normal3f(0.772096634f, 0.635505199f, 0.f),
    C_Vertex3f(0.180096626f, 0.137182236f, 0.f),
    C_Vertex3f(0.180096626f, 0.137182236f, -1.f),
    C_Normal3f(0.69950527f, 0.714627445f, 0.f),
    C_Vertex3f(0.195559502f, 0.120477676f, 0.f),
    C_Vertex3f(0.195559502f, 0.120477676f, -1.f),
    C_Normal3f(0.622323155f, 0.782760501f, 0.f),
    C_Vertex3f(0.212094784f, 0.105861187f, 0.f),
    C_Vertex3f(0.212094784f, 0.105861187f, -1.f),
    C_Normal3f(0.534827769f, 0.844961047f, 0.f),
    C_Vertex3f(0.229702473f, 0.0933327675f, 0.f),
    C_Vertex3f(0.229702473f, 0.0933327675f, -1.f),
    C_Normal3f(0.439273477f, 0.898353338f, 0.f),
    C_Vertex3f(0.248382568f, 0.0828924179f, 0.f),
    C_Vertex3f(0.248382568f, 0.0828924179f, -1.f),
    C_Normal3f(0.338896424f, 0.940823734f, 0.f),
    C_Vertex3f(0.268135071f, 0.0745401382f, 0.f),
    C_Vertex3f(0.268135071f, 0.0745401382f, -1.f),
    C_Normal3f(0.237391695f, 0.97141403f, 0.f),
    C_Vertex3f(0.28895998f, 0.0682759285f, 0.f),
    C_Vertex3f(0.28895998f, 0.0682759285f, -1.f),
    C_Normal3f(0.138275996f, 0.990393817f, 0.f),
    C_Vertex3f(0.310857296f, 0.0640997887f, 0.f),
    C_Vertex3f(0.310857296f, 0.0640997887f, -1.f),
    C_Normal3f(0.0449997857f, 0.998987019f, 0.f),
    C_Vertex3f(0.333827019f, 0.0620117188f, 0.f),
    C_Vertex3f(0.333827019f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0467745773f, 0.99890542f, 0.f),
    C_Vertex3f(0.357212067f, 0.0620117188f, 0.f),
    C_Vertex3f(0.357212067f, 0.0620117188f, -1.f),
    C_Normal3f(-0.145274758f, 0.989391387f, 0.f),
    C_Vertex3f(0.379284859f, 0.0641403198f, 0.f),
    C_Vertex3f(0.379284859f, 0.0641403198f, -1.f),
    C_Normal3f(-0.244655088f, 0.969610214f, 0.f),
    C_Vertex3f(0.400702477f, 0.068397522f, 0.f),
    C_Vertex3f(0.400702477f, 0.068397522f, -1.f),
    C_Normal3f(-0.326450378f, 0.945214272f, 0.f),
    C_Vertex3f(0.42146492f, 0.0747833252f, 0.f),
    C_Vertex3f(0.42146492f, 0.0747833252f, -1.f),
    C_Normal3f(-0.389930606f, 0.920844197f, 0.f),
    C_Vertex3f(0.431518555f, 0.0790405273f, 0.f),
    C_Vertex3f(0.431518555f, 0.0790405273f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.543731749f, -0.839259088f, 0.f),
    C_Vertex3f(0.431518555f, 0.0790405273f, 0.f),
    C_Vertex3f(0.431518555f, 0.0790405273f, -1.f),
    C_Normal3f(-0.498548836f, -0.866861641f, 0.f),
    C_Vertex3f(0.421573639f, 0.085483551f, 0.f),
    C_Vertex3f(0.421573639f, 0.085483551f, -1.f),
    C_Normal3f(-0.438400388f, -0.898779809f, 0.f),
    C_Vertex3f(0.401138306f, 0.0965127945f, 0.f),
    C_Vertex3f(0.401138306f, 0.0965127945f, -1.f),
    C_Normal3f(-0.361636579f, -0.932319164f, 0.f),
    C_Vertex3f(0.380157471f, 0.105685234f, 0.f),
    C_Vertex3f(0.380157471f, 0.105685234f, -1.f),
    C_Normal3f(-0.294779986f, -0.955565155f, 0.f),
    C_Vertex3f(0.358631134f, 0.11300087f, 0.f),
    C_Vertex3f(0.358631134f, 0.11300087f, -1.f),
    C_Normal3f(-0.240087286f, -0.970751286f, 0.f),
    C_Vertex3f(0.347595215f, 0.115730286f, 0.f),
    C_Vertex3f(0.347595215f, 0.115730286f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.948659778f, 0.316298246f, 0.f),
    C_Vertex3f(0.347595215f, 0.115730286f, 0.f),
    C_Vertex3f(0.347595215f, 0.115730286f, -1.f),
    C_Vertex3f(0.368141174f, 0.177352905f, -1.f),
    C_Vertex3f(0.368141174f, 0.177352905f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.650199294f, 0.355999351f, 0.f),
    C_Vertex3f(0.647863507f, 0.378579259f, 0.f),
    C_Vertex3f(0.648362637f, 0.263638496f, 0.f),
    C_Vertex3f(0.651367188f, 0.301158905f, 0.f),
    C_Vertex3f(0.651367188f, 0.332812309f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.644359827f, 0.400552034f, 0.f),
    C_Vertex3f(0.639688253f, 0.421917677f, 0.f),
    C_Vertex3f(0.642353535f, 0.228110313f, 0.f),
    C_Vertex3f(0.648362637f, 0.263638496f, 0.f),
    C_Vertex3f(0.647863507f, 0.378579259f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.618666172f, 0.482371807f, 0.f),
    C_Vertex3f(0.621321678f, 0.163030624f, 0.f),
    C_Vertex3f(0.626841426f, 0.462827563f, 0.f),
    C_Vertex3f(0.633339882f, 0.194574356f, 0.f),
    C_Vertex3f(0.633848786f, 0.442676187f, 0.f),
    C_Vertex3f(0.642353535f, 0.228110313f, 0.f),
    C_Vertex3f(0.639688253f, 0.421917677f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.587461352f, 0.536031723f, 0.f),
    C_Vertex3f(0.58810854f, 0.106082916f, 0.f),
    C_Vertex3f(0.598921418f, 0.519195557f, 0.f),
    C_Vertex3f(0.606244564f, 0.133533478f, 0.f),
    C_Vertex3f(0.609323025f, 0.501308918f, 0.f),
    C_Vertex3f(0.621321678f, 0.163030624f, 0.f),
    C_Vertex3f(0.618666172f, 0.482371807f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.574942827f, 0.551817417f, 0.f),
    C_Vertex3f(0.561365843f, 0.566552639f, 0.f),
    C_Vertex3f(0.56148541f, 0.367734313f, 0.f),
    C_Vertex3f(0.563138366f, 0.349528432f, 0.f),
    C_Vertex3f(0.563964844f, 0.330708504f, 0.f),
    C_Vertex3f(0.566913605f, 0.0806789398f, 0.f),
    C_Vertex3f(0.58810854f, 0.106082916f, 0.f),
    C_Vertex3f(0.587461352f, 0.536031723f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.561365843f, 0.566552639f, 0.f),
    C_Vertex3f(0.546730399f, 0.580237389f, 0.f),
    C_Vertex3f(0.551567674f, 0.418667674f, 0.f),
    C_Vertex3f(0.555700064f, 0.402303934f, 0.f),
    C_Vertex3f(0.559005976f, 0.385326147f, 0.f),
    C_Vertex3f(0.56148541f, 0.367734313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.546730399f, 0.580237389f, 0.f),
    C_Vertex3f(0.531036496f, 0.592871666f, 0.f),
    C_Vertex3f(0.534211636f, 0.464074612f, 0.f),
    C_Vertex3f(0.54082346f, 0.449553013f, 0.f),
    C_Vertex3f(0.546608806f, 0.434417367f, 0.f),
    C_Vertex3f(0.551567674f, 0.418667674f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.500115871f, 0.51404047f, 0.f),
    C_Vertex3f(0.509787083f, 0.502767086f, 0.f),
    C_Vertex3f(0.514284134f, 0.604455471f, 0.f),
    C_Vertex3f(0.518693447f, 0.490681648f, 0.f),
    C_Vertex3f(0.531036496f, 0.592871666f, 0.f),
    C_Vertex3f(0.526834965f, 0.477784157f, 0.f),
    C_Vertex3f(0.534211636f, 0.464074612f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.496473312f, 0.614988804f, 0.f),
    C_Vertex3f(0.478116035f, 0.62420547f, 0.f),
    C_Vertex3f(0.478478909f, 0.534151077f, 0.f),
    C_Vertex3f(0.489679813f, 0.524501801f, 0.f),
    C_Vertex3f(0.500115871f, 0.51404047f, 0.f),
    C_Vertex3f(0.514284134f, 0.604455471f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.440662742f, 0.558035493f, 0.f),
    C_Vertex3f(0.453782558f, 0.55101347f, 0.f),
    C_Vertex3f(0.459212303f, 0.63210547f, 0.f),
    C_Vertex3f(0.466513157f, 0.5429883f, 0.f),
    C_Vertex3f(0.478116035f, 0.62420547f, 0.f),
    C_Vertex3f(0.478478909f, 0.534151077f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.384291291f, 0.576092124f, 0.f),
    C_Vertex3f(0.398967981f, 0.573082685f, 0.f),
    C_Vertex3f(0.399222374f, 0.647905469f, 0.f),
    C_Vertex3f(0.413255453f, 0.569070101f, 0.f),
    C_Vertex3f(0.419765472f, 0.643955469f, 0.f),
    C_Vertex3f(0.427153707f, 0.56405437f, 0.f),
    C_Vertex3f(0.439762115f, 0.638688803f, 0.f),
    C_Vertex3f(0.440662742f, 0.558035493f, 0.f),
    C_Vertex3f(0.459212303f, 0.63210547f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.196297646f, 0.522354603f, 0.f),
    C_Vertex3f(0.212935448f, 0.536541343f, 0.f),
    C_Vertex3f(0.213687778f, 0.625199318f, 0.f),
    C_Vertex3f(0.230584145f, 0.548701406f, 0.f),
    C_Vertex3f(0.232495904f, 0.632815361f, 0.f),
    C_Vertex3f(0.249243736f, 0.558834791f, 0.f),
    C_Vertex3f(0.251829863f, 0.639162064f, 0.f),
    C_Vertex3f(0.268914223f, 0.5669415f, 0.f),
    C_Vertex3f(0.271689653f, 0.644239426f, 0.f),
    C_Vertex3f(0.289595604f, 0.573021531f, 0.f),
    C_Vertex3f(0.292075276f, 0.648047447f, 0.f),
    C_Vertex3f(0.31128788f, 0.577074885f, 0.f),
    C_Vertex3f(0.312986732f, 0.650586128f, 0.f),
    C_Vertex3f(0.333991051f, 0.579101563f, 0.f),
    C_Vertex3f(0.334424019f, 0.651855469f, 0.f),
    C_Vertex3f(0.353770256f, 0.579101563f, 0.f),
    C_Vertex3f(0.356496811f, 0.651855469f, 0.f),
    C_Vertex3f(0.369225383f, 0.578098416f, 0.f),
    C_Vertex3f(0.37813282f, 0.650538802f, 0.f),
    C_Vertex3f(0.384291291f, 0.576092124f, 0.f),
    C_Vertex3f(0.399222374f, 0.647905469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145510197f, 0.44550395f, 0.f),
    C_Vertex3f(0.155277014f, 0.467715859f, 0.f),
    C_Vertex3f(0.16091013f, 0.594844222f, 0.f),
    C_Vertex3f(0.166997194f, 0.487928271f, 0.f),
    C_Vertex3f(0.177649021f, 0.60615921f, 0.f),
    C_Vertex3f(0.180670738f, 0.506141186f, 0.f),
    C_Vertex3f(0.195405483f, 0.616313934f, 0.f),
    C_Vertex3f(0.196297646f, 0.522354603f, 0.f),
    C_Vertex3f(0.213687778f, 0.625199318f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145188808f, 0.58236897f, 0.f),
    C_Vertex3f(0.130485058f, 0.568733454f, 0.f),
    C_Vertex3f(0.131836653f, 0.395081639f, 0.f),
    C_Vertex3f(0.137696743f, 0.421292543f, 0.f),
    C_Vertex3f(0.145510197f, 0.44550395f, 0.f),
    C_Vertex3f(0.16091013f, 0.594844222f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.131775141f, 0.247934818f, 0.f),
    C_Vertex3f(0.127909422f, 0.275531292f, 0.f),
    C_Vertex3f(0.130023003f, 0.0742666721f, 0.f),
    C_Vertex3f(0.125976563f, 0.304943085f, 0.f),
    C_Vertex3f(0.116798878f, 0.553937674f, 0.f),
    C_Vertex3f(0.125976563f, 0.336661339f, 0.f),
    C_Vertex3f(0.130485058f, 0.568733454f, 0.f),
    C_Vertex3f(0.127929926f, 0.366871238f, 0.f),
    C_Vertex3f(0.131836653f, 0.395081639f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.566913605f, 0.0806789398f, 0.f),
    C_Vertex3f(0.563964844f, 0.330708504f, 0.f),
    C_Vertex3f(0.563964844f, 0.303904533f, 0.f),
    C_Vertex3f(0.561629057f, 0.271873236f, 0.f),
    C_Vertex3f(0.556957483f, 0.241984606f, 0.f),
    C_Vertex3f(0.554786682f, 0.0690002441f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.552540779f, -0.00766944885f, 0.f),
    C_Vertex3f(0.586958885f, -0.0273418427f, 0.f),
    C_Vertex3f(0.554786682f, 0.0690002441f, 0.f),
    C_Vertex3f(0.556957483f, 0.241984606f, 0.f),
    C_Vertex3f(0.549950123f, 0.214238644f, 0.f),
    C_Vertex3f(0.540606976f, 0.188635349f, 0.f),
    C_Vertex3f(0.528928041f, 0.165174723f, 0.f),
    C_Vertex3f(0.518177032f, 0.0156097412f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.489219666f, 0.116165161f, 0.f),
    C_Vertex3f(0.492116928f, 0.0240783691f, 0.f),
    C_Vertex3f(0.498562813f, 0.124681473f, 0.f),
    C_Vertex3f(0.501022339f, 0.0290527344f, 0.f),
    C_Vertex3f(0.514913321f, 0.143856764f, 0.f),
    C_Vertex3f(0.518177032f, 0.0156097412f, 0.f),
    C_Vertex3f(0.528928041f, 0.165174723f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.483263969f, 0.121246338f, 0.f),
    C_Vertex3f(0.470546722f, 0.13078022f, 0.f),
    C_Vertex3f(0.473841429f, 0.01537323f, 0.f),
    C_Vertex3f(0.492116928f, 0.0240783691f, 0.f),
    C_Vertex3f(0.489219666f, 0.116165161f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431518555f, 0.0790405273f, 0.f),
    C_Vertex3f(0.435896397f, 0.00169372559f, 0.f),
    C_Vertex3f(0.442694664f, 0.14796257f, 0.f),
    C_Vertex3f(0.455101252f, 0.00791168213f, 0.f),
    C_Vertex3f(0.457023621f, 0.139685631f, 0.f),
    C_Vertex3f(0.473841429f, 0.01537323f, 0.f),
    C_Vertex3f(0.470546722f, 0.13078022f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.37732029f, 0.174785614f, 0.f),
    C_Vertex3f(0.380157471f, 0.105685234f, 0.f),
    C_Vertex3f(0.394872665f, 0.16902256f, 0.f),
    C_Vertex3f(0.401138306f, 0.0965127945f, 0.f),
    C_Vertex3f(0.411619186f, 0.162631035f, 0.f),
    C_Vertex3f(0.421573639f, 0.085483551f, 0.f),
    C_Vertex3f(0.427559853f, 0.155611038f, 0.f),
    C_Vertex3f(0.431518555f, 0.0790405273f, 0.f),
    C_Vertex3f(0.442694664f, 0.14796257f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.368141174f, 0.177352905f, 0.f),
    C_Vertex3f(0.347595215f, 0.115730286f, 0.f),
    C_Vertex3f(0.358631134f, 0.11300087f, 0.f),
    C_Vertex3f(0.380157471f, 0.105685234f, 0.f),
    C_Vertex3f(0.37732029f, 0.174785614f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.663574219f, 0.00936126709f, 0.f),
    C_Vertex3f(0.651123047f, 0.0140762329f, 0.f),
    C_Vertex3f(0.638694763f, -0.0496368408f, 0.f),
    C_Vertex3f(0.625073433f, 0.0262460709f, 0.f),
    C_Vertex3f(0.621431351f, -0.0434074402f, 0.f),
    C_Vertex3f(0.597876549f, 0.0411558151f, 0.f),
    C_Vertex3f(0.586958885f, -0.0273418427f, 0.f),
    C_Vertex3f(0.569532394f, 0.0588054657f, 0.f),
    C_Vertex3f(0.554786682f, 0.0690002441f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.311553001f, -0.00943183899f, 0.f),
    C_Vertex3f(0.332793236f, -0.0107421875f, 0.f),
    C_Vertex3f(0.333827019f, 0.0620117188f, 0.f),
    C_Vertex3f(0.354430199f, -0.0107421875f, 0.f),
    C_Vertex3f(0.357212067f, 0.0620117188f, 0.f),
    C_Vertex3f(0.375493765f, -0.00949859619f, 0.f),
    C_Vertex3f(0.379284859f, 0.0641403198f, 0.f),
    C_Vertex3f(0.396092653f, -0.00701141357f, 0.f),
    C_Vertex3f(0.400702477f, 0.068397522f, 0.f),
    C_Vertex3f(0.416226864f, -0.00328063965f, 0.f),
    C_Vertex3f(0.42146492f, 0.0747833252f, 0.f),
    C_Vertex3f(0.435896397f, 0.00169372559f, 0.f),
    C_Vertex3f(0.431518555f, 0.0790405273f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.212094784f, 0.105861187f, 0.f),
    C_Vertex3f(0.21293354f, 0.0167751312f, 0.f),
    C_Vertex3f(0.229702473f, 0.0933327675f, 0.f),
    C_Vertex3f(0.231646538f, 0.00891304016f, 0.f),
    C_Vertex3f(0.248382568f, 0.0828924179f, 0.f),
    C_Vertex3f(0.250864983f, 0.00236129761f, 0.f),
    C_Vertex3f(0.268135071f, 0.0745401382f, 0.f),
    C_Vertex3f(0.270588875f, -0.00288009644f, 0.f),
    C_Vertex3f(0.28895998f, 0.0682759285f, 0.f),
    C_Vertex3f(0.290818214f, -0.00681114197f, 0.f),
    C_Vertex3f(0.310857296f, 0.0640997887f, 0.f),
    C_Vertex3f(0.311553001f, -0.00943183899f, 0.f),
    C_Vertex3f(0.333827019f, 0.0620117188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154969454f, 0.176037312f, 0.f),
    C_Vertex3f(0.160339355f, 0.0479778051f, 0.f),
    C_Vertex3f(0.16656661f, 0.155702114f, 0.f),
    C_Vertex3f(0.177023888f, 0.0364303589f, 0.f),
    C_Vertex3f(0.180096626f, 0.137182236f, 0.f),
    C_Vertex3f(0.19472599f, 0.0259475708f, 0.f),
    C_Vertex3f(0.195559502f, 0.120477676f, 0.f),
    C_Vertex3f(0.21293354f, 0.0167751312f, 0.f),
    C_Vertex3f(0.212094784f, 0.105861187f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145305157f, 0.198187828f, 0.f),
    C_Vertex3f(0.137573719f, 0.222153664f, 0.f),
    C_Vertex3f(0.144672394f, 0.0605899096f, 0.f),
    C_Vertex3f(0.160339355f, 0.0479778051f, 0.f),
    C_Vertex3f(0.154969454f, 0.176037312f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.131775141f, 0.247934818f, 0.f),
    C_Vertex3f(0.130023003f, 0.0742666721f, 0.f),
    C_Vertex3f(0.144672394f, 0.0605899096f, 0.f),
    C_Vertex3f(0.137573719f, 0.222153664f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.130023003f, 0.0742666721f, 0.f),
    C_Vertex3f(0.116798878f, 0.553937674f, 0.f),
    C_Vertex3f(0.116391182f, 0.0890080929f, 0.f),
    C_Vertex3f(0.104130268f, 0.537981629f, 0.f),
    C_Vertex3f(0.103776932f, 0.104814172f, 0.f),
    C_Vertex3f(0.092479229f, 0.520865321f, 0.f),
    C_Vertex3f(0.0921802521f, 0.121684909f, 0.f),
    C_Vertex3f(0.0818457603f, 0.502588749f, 0.f),
    C_Vertex3f(0.0816011429f, 0.139620304f, 0.f),
    C_Vertex3f(0.0722298622f, 0.483151913f, 0.f),
    C_Vertex3f(0.0720396042f, 0.158620358f, 0.f),
    C_Vertex3f(0.0638159513f, 0.463182569f, 0.f),
    C_Vertex3f(0.0636732578f, 0.178220868f, 0.f),
    C_Vertex3f(0.0566040277f, 0.442680717f, 0.f),
    C_Vertex3f(0.0565021038f, 0.198421836f, 0.f),
    C_Vertex3f(0.0505940914f, 0.421646357f, 0.f),
    C_Vertex3f(0.0505261421f, 0.219223261f, 0.f),
    C_Vertex3f(0.0457861423f, 0.400079489f, 0.f),
    C_Vertex3f(0.0457453728f, 0.240625143f, 0.f),
    C_Vertex3f(0.0421801805f, 0.377980113f, 0.f),
    C_Vertex3f(0.0421597958f, 0.262627482f, 0.f),
    C_Vertex3f(0.039776206f, 0.355348229f, 0.f),
    C_Vertex3f(0.0397694111f, 0.285230279f, 0.f),
    C_Vertex3f(0.0385742188f, 0.332183838f, 0.f),
    C_Vertex3f(0.0385742188f, 0.308433533f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.650199294f, 0.355999351f, -1.f),
    C_Vertex3f(0.647863507f, 0.378579259f, -1.f),
    C_Vertex3f(0.648362637f, 0.263638496f, -1.f),
    C_Vertex3f(0.651367188f, 0.301158905f, -1.f),
    C_Vertex3f(0.651367188f, 0.332812309f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.644359827f, 0.400552034f, -1.f),
    C_Vertex3f(0.639688253f, 0.421917677f, -1.f),
    C_Vertex3f(0.642353535f, 0.228110313f, -1.f),
    C_Vertex3f(0.648362637f, 0.263638496f, -1.f),
    C_Vertex3f(0.647863507f, 0.378579259f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.618666172f, 0.482371807f, -1.f),
    C_Vertex3f(0.621321678f, 0.163030624f, -1.f),
    C_Vertex3f(0.626841426f, 0.462827563f, -1.f),
    C_Vertex3f(0.633339882f, 0.194574356f, -1.f),
    C_Vertex3f(0.633848786f, 0.442676187f, -1.f),
    C_Vertex3f(0.642353535f, 0.228110313f, -1.f),
    C_Vertex3f(0.639688253f, 0.421917677f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.587461352f, 0.536031723f, -1.f),
    C_Vertex3f(0.58810854f, 0.106082916f, -1.f),
    C_Vertex3f(0.598921418f, 0.519195557f, -1.f),
    C_Vertex3f(0.606244564f, 0.133533478f, -1.f),
    C_Vertex3f(0.609323025f, 0.501308918f, -1.f),
    C_Vertex3f(0.621321678f, 0.163030624f, -1.f),
    C_Vertex3f(0.618666172f, 0.482371807f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.574942827f, 0.551817417f, -1.f),
    C_Vertex3f(0.561365843f, 0.566552639f, -1.f),
    C_Vertex3f(0.56148541f, 0.367734313f, -1.f),
    C_Vertex3f(0.563138366f, 0.349528432f, -1.f),
    C_Vertex3f(0.563964844f, 0.330708504f, -1.f),
    C_Vertex3f(0.566913605f, 0.0806789398f, -1.f),
    C_Vertex3f(0.58810854f, 0.106082916f, -1.f),
    C_Vertex3f(0.587461352f, 0.536031723f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.561365843f, 0.566552639f, -1.f),
    C_Vertex3f(0.546730399f, 0.580237389f, -1.f),
    C_Vertex3f(0.551567674f, 0.418667674f, -1.f),
    C_Vertex3f(0.555700064f, 0.402303934f, -1.f),
    C_Vertex3f(0.559005976f, 0.385326147f, -1.f),
    C_Vertex3f(0.56148541f, 0.367734313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.546730399f, 0.580237389f, -1.f),
    C_Vertex3f(0.531036496f, 0.592871666f, -1.f),
    C_Vertex3f(0.534211636f, 0.464074612f, -1.f),
    C_Vertex3f(0.54082346f, 0.449553013f, -1.f),
    C_Vertex3f(0.546608806f, 0.434417367f, -1.f),
    C_Vertex3f(0.551567674f, 0.418667674f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.500115871f, 0.51404047f, -1.f),
    C_Vertex3f(0.509787083f, 0.502767086f, -1.f),
    C_Vertex3f(0.514284134f, 0.604455471f, -1.f),
    C_Vertex3f(0.518693447f, 0.490681648f, -1.f),
    C_Vertex3f(0.531036496f, 0.592871666f, -1.f),
    C_Vertex3f(0.526834965f, 0.477784157f, -1.f),
    C_Vertex3f(0.534211636f, 0.464074612f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.496473312f, 0.614988804f, -1.f),
    C_Vertex3f(0.478116035f, 0.62420547f, -1.f),
    C_Vertex3f(0.478478909f, 0.534151077f, -1.f),
    C_Vertex3f(0.489679813f, 0.524501801f, -1.f),
    C_Vertex3f(0.500115871f, 0.51404047f, -1.f),
    C_Vertex3f(0.514284134f, 0.604455471f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.440662742f, 0.558035493f, -1.f),
    C_Vertex3f(0.453782558f, 0.55101347f, -1.f),
    C_Vertex3f(0.459212303f, 0.63210547f, -1.f),
    C_Vertex3f(0.466513157f, 0.5429883f, -1.f),
    C_Vertex3f(0.478116035f, 0.62420547f, -1.f),
    C_Vertex3f(0.478478909f, 0.534151077f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.384291291f, 0.576092124f, -1.f),
    C_Vertex3f(0.398967981f, 0.573082685f, -1.f),
    C_Vertex3f(0.399222374f, 0.647905469f, -1.f),
    C_Vertex3f(0.413255453f, 0.569070101f, -1.f),
    C_Vertex3f(0.419765472f, 0.643955469f, -1.f),
    C_Vertex3f(0.427153707f, 0.56405437f, -1.f),
    C_Vertex3f(0.439762115f, 0.638688803f, -1.f),
    C_Vertex3f(0.440662742f, 0.558035493f, -1.f),
    C_Vertex3f(0.459212303f, 0.63210547f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.196297646f, 0.522354603f, -1.f),
    C_Vertex3f(0.212935448f, 0.536541343f, -1.f),
    C_Vertex3f(0.213687778f, 0.625199318f, -1.f),
    C_Vertex3f(0.230584145f, 0.548701406f, -1.f),
    C_Vertex3f(0.232495904f, 0.632815361f, -1.f),
    C_Vertex3f(0.249243736f, 0.558834791f, -1.f),
    C_Vertex3f(0.251829863f, 0.639162064f, -1.f),
    C_Vertex3f(0.268914223f, 0.5669415f, -1.f),
    C_Vertex3f(0.271689653f, 0.644239426f, -1.f),
    C_Vertex3f(0.289595604f, 0.573021531f, -1.f),
    C_Vertex3f(0.292075276f, 0.648047447f, -1.f),
    C_Vertex3f(0.31128788f, 0.577074885f, -1.f),
    C_Vertex3f(0.312986732f, 0.650586128f, -1.f),
    C_Vertex3f(0.333991051f, 0.579101563f, -1.f),
    C_Vertex3f(0.334424019f, 0.651855469f, -1.f),
    C_Vertex3f(0.353770256f, 0.579101563f, -1.f),
    C_Vertex3f(0.356496811f, 0.651855469f, -1.f),
    C_Vertex3f(0.369225383f, 0.578098416f, -1.f),
    C_Vertex3f(0.37813282f, 0.650538802f, -1.f),
    C_Vertex3f(0.384291291f, 0.576092124f, -1.f),
    C_Vertex3f(0.399222374f, 0.647905469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145510197f, 0.44550395f, -1.f),
    C_Vertex3f(0.155277014f, 0.467715859f, -1.f),
    C_Vertex3f(0.16091013f, 0.594844222f, -1.f),
    C_Vertex3f(0.166997194f, 0.487928271f, -1.f),
    C_Vertex3f(0.177649021f, 0.60615921f, -1.f),
    C_Vertex3f(0.180670738f, 0.506141186f, -1.f),
    C_Vertex3f(0.195405483f, 0.616313934f, -1.f),
    C_Vertex3f(0.196297646f, 0.522354603f, -1.f),
    C_Vertex3f(0.213687778f, 0.625199318f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145188808f, 0.58236897f, -1.f),
    C_Vertex3f(0.130485058f, 0.568733454f, -1.f),
    C_Vertex3f(0.131836653f, 0.395081639f, -1.f),
    C_Vertex3f(0.137696743f, 0.421292543f, -1.f),
    C_Vertex3f(0.145510197f, 0.44550395f, -1.f),
    C_Vertex3f(0.16091013f, 0.594844222f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.131775141f, 0.247934818f, -1.f),
    C_Vertex3f(0.127909422f, 0.275531292f, -1.f),
    C_Vertex3f(0.130023003f, 0.0742666721f, -1.f),
    C_Vertex3f(0.125976563f, 0.304943085f, -1.f),
    C_Vertex3f(0.116798878f, 0.553937674f, -1.f),
    C_Vertex3f(0.125976563f, 0.336661339f, -1.f),
    C_Vertex3f(0.130485058f, 0.568733454f, -1.f),
    C_Vertex3f(0.127929926f, 0.366871238f, -1.f),
    C_Vertex3f(0.131836653f, 0.395081639f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.566913605f, 0.0806789398f, -1.f),
    C_Vertex3f(0.563964844f, 0.330708504f, -1.f),
    C_Vertex3f(0.563964844f, 0.303904533f, -1.f),
    C_Vertex3f(0.561629057f, 0.271873236f, -1.f),
    C_Vertex3f(0.556957483f, 0.241984606f, -1.f),
    C_Vertex3f(0.554786682f, 0.0690002441f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.552540779f, -0.00766944885f, -1.f),
    C_Vertex3f(0.586958885f, -0.0273418427f, -1.f),
    C_Vertex3f(0.554786682f, 0.0690002441f, -1.f),
    C_Vertex3f(0.556957483f, 0.241984606f, -1.f),
    C_Vertex3f(0.549950123f, 0.214238644f, -1.f),
    C_Vertex3f(0.540606976f, 0.188635349f, -1.f),
    C_Vertex3f(0.528928041f, 0.165174723f, -1.f),
    C_Vertex3f(0.518177032f, 0.0156097412f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.489219666f, 0.116165161f, -1.f),
    C_Vertex3f(0.492116928f, 0.0240783691f, -1.f),
    C_Vertex3f(0.498562813f, 0.124681473f, -1.f),
    C_Vertex3f(0.501022339f, 0.0290527344f, -1.f),
    C_Vertex3f(0.514913321f, 0.143856764f, -1.f),
    C_Vertex3f(0.518177032f, 0.0156097412f, -1.f),
    C_Vertex3f(0.528928041f, 0.165174723f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.483263969f, 0.121246338f, -1.f),
    C_Vertex3f(0.470546722f, 0.13078022f, -1.f),
    C_Vertex3f(0.473841429f, 0.01537323f, -1.f),
    C_Vertex3f(0.492116928f, 0.0240783691f, -1.f),
    C_Vertex3f(0.489219666f, 0.116165161f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431518555f, 0.0790405273f, -1.f),
    C_Vertex3f(0.435896397f, 0.00169372559f, -1.f),
    C_Vertex3f(0.442694664f, 0.14796257f, -1.f),
    C_Vertex3f(0.455101252f, 0.00791168213f, -1.f),
    C_Vertex3f(0.457023621f, 0.139685631f, -1.f),
    C_Vertex3f(0.473841429f, 0.01537323f, -1.f),
    C_Vertex3f(0.470546722f, 0.13078022f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.37732029f, 0.174785614f, -1.f),
    C_Vertex3f(0.380157471f, 0.105685234f, -1.f),
    C_Vertex3f(0.394872665f, 0.16902256f, -1.f),
    C_Vertex3f(0.401138306f, 0.0965127945f, -1.f),
    C_Vertex3f(0.411619186f, 0.162631035f, -1.f),
    C_Vertex3f(0.421573639f, 0.085483551f, -1.f),
    C_Vertex3f(0.427559853f, 0.155611038f, -1.f),
    C_Vertex3f(0.431518555f, 0.0790405273f, -1.f),
    C_Vertex3f(0.442694664f, 0.14796257f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.368141174f, 0.177352905f, -1.f),
    C_Vertex3f(0.347595215f, 0.115730286f, -1.f),
    C_Vertex3f(0.358631134f, 0.11300087f, -1.f),
    C_Vertex3f(0.380157471f, 0.105685234f, -1.f),
    C_Vertex3f(0.37732029f, 0.174785614f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.663574219f, 0.00936126709f, -1.f),
    C_Vertex3f(0.651123047f, 0.0140762329f, -1.f),
    C_Vertex3f(0.638694763f, -0.0496368408f, -1.f),
    C_Vertex3f(0.625073433f, 0.0262460709f, -1.f),
    C_Vertex3f(0.621431351f, -0.0434074402f, -1.f),
    C_Vertex3f(0.597876549f, 0.0411558151f, -1.f),
    C_Vertex3f(0.586958885f, -0.0273418427f, -1.f),
    C_Vertex3f(0.569532394f, 0.0588054657f, -1.f),
    C_Vertex3f(0.554786682f, 0.0690002441f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.311553001f, -0.00943183899f, -1.f),
    C_Vertex3f(0.332793236f, -0.0107421875f, -1.f),
    C_Vertex3f(0.333827019f, 0.0620117188f, -1.f),
    C_Vertex3f(0.354430199f, -0.0107421875f, -1.f),
    C_Vertex3f(0.357212067f, 0.0620117188f, -1.f),
    C_Vertex3f(0.375493765f, -0.00949859619f, -1.f),
    C_Vertex3f(0.379284859f, 0.0641403198f, -1.f),
    C_Vertex3f(0.396092653f, -0.00701141357f, -1.f),
    C_Vertex3f(0.400702477f, 0.068397522f, -1.f),
    C_Vertex3f(0.416226864f, -0.00328063965f, -1.f),
    C_Vertex3f(0.42146492f, 0.0747833252f, -1.f),
    C_Vertex3f(0.435896397f, 0.00169372559f, -1.f),
    C_Vertex3f(0.431518555f, 0.0790405273f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.212094784f, 0.105861187f, -1.f),
    C_Vertex3f(0.21293354f, 0.0167751312f, -1.f),
    C_Vertex3f(0.229702473f, 0.0933327675f, -1.f),
    C_Vertex3f(0.231646538f, 0.00891304016f, -1.f),
    C_Vertex3f(0.248382568f, 0.0828924179f, -1.f),
    C_Vertex3f(0.250864983f, 0.00236129761f, -1.f),
    C_Vertex3f(0.268135071f, 0.0745401382f, -1.f),
    C_Vertex3f(0.270588875f, -0.00288009644f, -1.f),
    C_Vertex3f(0.28895998f, 0.0682759285f, -1.f),
    C_Vertex3f(0.290818214f, -0.00681114197f, -1.f),
    C_Vertex3f(0.310857296f, 0.0640997887f, -1.f),
    C_Vertex3f(0.311553001f, -0.00943183899f, -1.f),
    C_Vertex3f(0.333827019f, 0.0620117188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154969454f, 0.176037312f, -1.f),
    C_Vertex3f(0.160339355f, 0.0479778051f, -1.f),
    C_Vertex3f(0.16656661f, 0.155702114f, -1.f),
    C_Vertex3f(0.177023888f, 0.0364303589f, -1.f),
    C_Vertex3f(0.180096626f, 0.137182236f, -1.f),
    C_Vertex3f(0.19472599f, 0.0259475708f, -1.f),
    C_Vertex3f(0.195559502f, 0.120477676f, -1.f),
    C_Vertex3f(0.21293354f, 0.0167751312f, -1.f),
    C_Vertex3f(0.212094784f, 0.105861187f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145305157f, 0.198187828f, -1.f),
    C_Vertex3f(0.137573719f, 0.222153664f, -1.f),
    C_Vertex3f(0.144672394f, 0.0605899096f, -1.f),
    C_Vertex3f(0.160339355f, 0.0479778051f, -1.f),
    C_Vertex3f(0.154969454f, 0.176037312f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.131775141f, 0.247934818f, -1.f),
    C_Vertex3f(0.130023003f, 0.0742666721f, -1.f),
    C_Vertex3f(0.144672394f, 0.0605899096f, -1.f),
    C_Vertex3f(0.137573719f, 0.222153664f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.130023003f, 0.0742666721f, -1.f),
    C_Vertex3f(0.116798878f, 0.553937674f, -1.f),
    C_Vertex3f(0.116391182f, 0.0890080929f, -1.f),
    C_Vertex3f(0.104130268f, 0.537981629f, -1.f),
    C_Vertex3f(0.103776932f, 0.104814172f, -1.f),
    C_Vertex3f(0.092479229f, 0.520865321f, -1.f),
    C_Vertex3f(0.0921802521f, 0.121684909f, -1.f),
    C_Vertex3f(0.0818457603f, 0.502588749f, -1.f),
    C_Vertex3f(0.0816011429f, 0.139620304f, -1.f),
    C_Vertex3f(0.0722298622f, 0.483151913f, -1.f),
    C_Vertex3f(0.0720396042f, 0.158620358f, -1.f),
    C_Vertex3f(0.0638159513f, 0.463182569f, -1.f),
    C_Vertex3f(0.0636732578f, 0.178220868f, -1.f),
    C_Vertex3f(0.0566040277f, 0.442680717f, -1.f),
    C_Vertex3f(0.0565021038f, 0.198421836f, -1.f),
    C_Vertex3f(0.0505940914f, 0.421646357f, -1.f),
    C_Vertex3f(0.0505261421f, 0.219223261f, -1.f),
    C_Vertex3f(0.0457861423f, 0.400079489f, -1.f),
    C_Vertex3f(0.0457453728f, 0.240625143f, -1.f),
    C_Vertex3f(0.0421801805f, 0.377980113f, -1.f),
    C_Vertex3f(0.0421597958f, 0.262627482f, -1.f),
    C_Vertex3f(0.039776206f, 0.355348229f, -1.f),
    C_Vertex3f(0.0397694111f, 0.285230279f, -1.f),
    C_Vertex3f(0.0385742188f, 0.332183838f, -1.f),
    C_Vertex3f(0.0385742188f, 0.308433533f, -1.f),
    C_End(),
    C_Translatef(0.696289063f, 0.f, 0.f),
};
static const Call char_82[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Normal3f(0.00171212223f, 0.999998629f, 0.f),
    C_Vertex3f(0.365180016f, 0.640625f, 0.f),
    C_Vertex3f(0.365180016f, 0.640625f, -1.f),
    C_Normal3f(0.0414890237f, 0.999138951f, 0.f),
    C_Vertex3f(0.385302901f, 0.640085697f, 0.f),
    C_Vertex3f(0.385302901f, 0.640085697f, -1.f),
    C_Normal3f(0.0738858804f, 0.99726665f, 0.f),
    C_Vertex3f(0.404142499f, 0.639007092f, 0.f),
    C_Vertex3f(0.404142499f, 0.639007092f, -1.f),
    C_Normal3f(0.110904679f, 0.993831038f, 0.f),
    C_Vertex3f(0.421698809f, 0.637389183f, 0.f),
    C_Vertex3f(0.421698809f, 0.637389183f, -1.f),
    C_Normal3f(0.153417826f, 0.988161385f, 0.f),
    C_Vertex3f(0.43797183f, 0.635231972f, 0.f),
    C_Vertex3f(0.43797183f, 0.635231972f, -1.f),
    C_Normal3f(0.202448204f, 0.979292929f, 0.f),
    C_Vertex3f(0.452961564f, 0.632535458f, 0.f),
    C_Vertex3f(0.452961564f, 0.632535458f, -1.f),
    C_Normal3f(0.259147733f, 0.965837717f, 0.f),
    C_Vertex3f(0.46666801f, 0.629299641f, 0.f),
    C_Vertex3f(0.46666801f, 0.629299641f, -1.f),
    C_Normal3f(0.324711859f, 0.945813f, 0.f),
    C_Vertex3f(0.479091167f, 0.625524521f, 0.f),
    C_Vertex3f(0.479091167f, 0.625524521f, -1.f),
    C_Normal3f(0.399137348f, 0.916891217f, 0.f),
    C_Vertex3f(0.490231037f, 0.621210098f, 0.f),
    C_Vertex3f(0.490231037f, 0.621210098f, -1.f),
    C_Normal3f(0.474592984f, 0.880205274f, 0.f),
    C_Vertex3f(0.500811219f, 0.616069436f, 0.f),
    C_Vertex3f(0.500811219f, 0.616069436f, -1.f),
    C_Normal3f(0.547919035f, 0.836531341f, 0.f),
    C_Vertex3f(0.510831714f, 0.610102534f, 0.f),
    C_Vertex3f(0.510831714f, 0.610102534f, -1.f),
    C_Normal3f(0.617431939f, 0.786624312f, 0.f),
    C_Vertex3f(0.520292521f, 0.603309393f, 0.f),
    C_Vertex3f(0.520292521f, 0.603309393f, -1.f),
    C_Normal3f(0.681681931f, 0.731648624f, 0.f),
    C_Vertex3f(0.52919364f, 0.595690012f, 0.f),
    C_Vertex3f(0.52919364f, 0.595690012f, -1.f),
    C_Normal3f(0.739599764f, 0.673047006f, 0.f),
    C_Vertex3f(0.537535071f, 0.587244391f, 0.f),
    C_Vertex3f(0.537535071f, 0.587244391f, -1.f),
    C_Normal3f(0.790571511f, 0.612369716f, 0.f),
    C_Vertex3f(0.545316815f, 0.577972531f, 0.f),
    C_Vertex3f(0.545316815f, 0.577972531f, -1.f),
    C_Normal3f(0.838434756f, 0.545002043f, 0.f),
    C_Vertex3f(0.552538872f, 0.567874432f, 0.f),
    C_Vertex3f(0.552538872f, 0.567874432f, -1.f),
    C_Normal3f(0.891729116f, 0.452569515f, 0.f),
    C_Vertex3f(0.562532425f, 0.551487923f, 0.f),
    C_Vertex3f(0.562532425f, 0.551487923f, -1.f),
    C_Normal3f(0.942870557f, 0.333159357f, 0.f),
    C_Vertex3f(0.572525978f, 0.528492451f, 0.f),
    C_Vertex3f(0.572525978f, 0.528492451f, -1.f),
    C_Normal3f(0.980169415f, 0.198161289f, 0.f),
    C_Vertex3f(0.579188347f, 0.504350185f, 0.f),
    C_Vertex3f(0.579188347f, 0.504350185f, -1.f),
    C_Normal3f(0.997500598f, 0.070658356f, 0.f),
    C_Vertex3f(0.582519531f, 0.479061127f, 0.f),
    C_Vertex3f(0.582519531f, 0.479061127f, -1.f),
    C_Normal3f(0.999346197f, -0.0361558236f, 0.f),
    C_Vertex3f(0.582519531f, 0.457323074f, 0.f),
    C_Vertex3f(0.582519531f, 0.457323074f, -1.f),
    C_Normal3f(0.99178654f, -0.127904668f, 0.f),
    C_Vertex3f(0.581140757f, 0.440951824f, 0.f),
    C_Vertex3f(0.581140757f, 0.440951824f, -1.f),
    C_Normal3f(0.975756466f, -0.218859062f, 0.f),
    C_Vertex3f(0.578383207f, 0.425249577f, 0.f),
    C_Vertex3f(0.578383207f, 0.425249577f, -1.f),
    C_Normal3f(0.950106025f, -0.31192717f, 0.f),
    C_Vertex3f(0.574246883f, 0.410216331f, 0.f),
    C_Vertex3f(0.574246883f, 0.410216331f, -1.f),
    C_Normal3f(0.914558947f, -0.404452801f, 0.f),
    C_Vertex3f(0.568731785f, 0.395852089f, 0.f),
    C_Vertex3f(0.568731785f, 0.395852089f, -1.f),
    C_Normal3f(0.869682014f, -0.493612349f, 0.f),
    C_Vertex3f(0.561837912f, 0.382156849f, 0.f),
    C_Vertex3f(0.561837912f, 0.382156849f, -1.f),
    C_Normal3f(0.816872001f, -0.576819062f, 0.f),
    C_Vertex3f(0.553565264f, 0.369130611f, 0.f),
    C_Vertex3f(0.553565264f, 0.369130611f, -1.f),
    C_Normal3f(0.758145928f, -0.652084887f, 0.f),
    C_Vertex3f(0.543913841f, 0.356773376f, 0.f),
    C_Vertex3f(0.543913841f, 0.356773376f, -1.f),
    C_Normal3f(0.687572002f, -0.72611618f, 0.f),
    C_Vertex3f(0.532883644f, 0.345085144f, 0.f),
    C_Vertex3f(0.532883644f, 0.345085144f, -1.f),
    C_Normal3f(0.604920149f, -0.796286106f, 0.f),
    C_Vertex3f(0.520351648f, 0.334461927f, 0.f),
    C_Vertex3f(0.520351648f, 0.334461927f, -1.f),
    C_Normal3f(0.521054924f, -0.853523076f, 0.f),
    C_Vertex3f(0.506317854f, 0.324903727f, 0.f),
    C_Vertex3f(0.506317854f, 0.324903727f, -1.f),
    C_Normal3f(0.439138532f, -0.898419321f, 0.f),
    C_Vertex3f(0.490782261f, 0.316410542f, 0.f),
    C_Vertex3f(0.490782261f, 0.316410542f, -1.f),
    C_Normal3f(0.361444056f, -0.932393849f, 0.f),
    C_Vertex3f(0.473744869f, 0.308982372f, 0.f),
    C_Vertex3f(0.473744869f, 0.308982372f, -1.f),
    C_Normal3f(0.289333045f, -0.957228541f, 0.f),
    C_Vertex3f(0.455205679f, 0.302619219f, 0.f),
    C_Vertex3f(0.455205679f, 0.302619219f, -1.f),
    C_Normal3f(0.223412767f, -0.974723876f, 0.f),
    C_Vertex3f(0.43516469f, 0.297321081f, 0.f),
    C_Vertex3f(0.43516469f, 0.297321081f, -1.f),
    C_Normal3f(0.173269987f, -0.984874308f, 0.f),
    C_Vertex3f(0.413621902f, 0.293087959f, 0.f),
    C_Vertex3f(0.413621902f, 0.293087959f, -1.f),
    C_Normal3f(0.136196211f, -0.990681887f, 0.f),
    C_Vertex3f(0.402099609f, 0.291503906f, 0.f),
    C_Vertex3f(0.402099609f, 0.291503906f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.433096468f, 0.901347518f, 0.f),
    C_Vertex3f(0.402099609f, 0.291503906f, 0.f),
    C_Vertex3f(0.402099609f, 0.291503906f, -1.f),
    C_Normal3f(0.461343557f, 0.887221575f, 0.f),
    C_Vertex3f(0.410511017f, 0.287462234f, 0.f),
    C_Vertex3f(0.410511017f, 0.287462234f, -1.f),
    C_Normal3f(0.501844704f, 0.86495769f, 0.f),
    C_Vertex3f(0.425312996f, 0.27943325f, 0.f),
    C_Vertex3f(0.425312996f, 0.27943325f, -1.f),
    C_Normal3f(0.559579551f, 0.828776717f, 0.f),
    C_Vertex3f(0.438094139f, 0.271458626f, 0.f),
    C_Vertex3f(0.438094139f, 0.271458626f, -1.f),
    C_Normal3f(0.616705537f, 0.787193894f, 0.f),
    C_Vertex3f(0.448854446f, 0.263538361f, 0.f),
    C_Vertex3f(0.448854446f, 0.263538361f, -1.f),
    C_Normal3f(0.673848331f, 0.738869667f, 0.f),
    C_Vertex3f(0.453224182f, 0.259605408f, 0.f),
    C_Vertex3f(0.453224182f, 0.259605408f, -1.f),
    C_Normal3f(0.711601377f, 0.702583492f, 0.f),
    C_Vertex3f(0.462509155f, 0.251085281f, 0.f),
    C_Vertex3f(0.462509155f, 0.251085281f, -1.f),
    C_Normal3f(0.750892103f, 0.660424888f, 0.f),
    C_Vertex3f(0.480566978f, 0.231911659f, 0.f),
    C_Vertex3f(0.480566978f, 0.231911659f, -1.f),
    C_Normal3f(0.791269779f, 0.611467183f, 0.f),
    C_Vertex3f(0.498112679f, 0.210604668f, 0.f),
    C_Vertex3f(0.498112679f, 0.210604668f, -1.f),
    C_Normal3f(0.84359175f, 0.53698504f, 0.f),
    C_Vertex3f(0.515146255f, 0.187164307f, 0.f),
    C_Vertex3f(0.515146255f, 0.187164307f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.84359175f, 0.53698504f, 0.f),
    C_Vertex3f(0.515146255f, 0.187164307f, 0.f),
    C_Vertex3f(0.515146255f, 0.187164307f, -1.f),
    C_Vertex3f(0.634284973f, 0.f, -1.f),
    C_Vertex3f(0.634284973f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.634284973f, 0.f, 0.f),
    C_Vertex3f(0.634284973f, 0.f, -1.f),
    C_Vertex3f(0.52835083f, 0.f, -1.f),
    C_Vertex3f(0.52835083f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.843566835f, -0.5370242f, 0.f),
    C_Vertex3f(0.52835083f, 0.f, 0.f),
    C_Vertex3f(0.52835083f, 0.f, -1.f),
    C_Normal3f(-0.841976523f, -0.539514124f, 0.f),
    C_Vertex3f(0.43431282f, 0.147716522f, 0.f),
    C_Vertex3f(0.43431282f, 0.147716522f, -1.f),
    C_Normal3f(-0.827761114f, -0.561080813f, 0.f),
    C_Vertex3f(0.417385101f, 0.173175335f, 0.f),
    C_Vertex3f(0.417385101f, 0.173175335f, -1.f),
    C_Normal3f(-0.815335989f, -0.578988075f, 0.f),
    C_Vertex3f(0.402095795f, 0.19524622f, 0.f),
    C_Vertex3f(0.402095795f, 0.19524622f, -1.f),
    C_Normal3f(-0.797967613f, -0.602700353f, 0.f),
    C_Vertex3f(0.388444901f, 0.213929176f, 0.f),
    C_Vertex3f(0.388444901f, 0.213929176f, -1.f),
    C_Normal3f(-0.770745218f, -0.637143493f, 0.f),
    C_Vertex3f(0.376432419f, 0.229224205f, 0.f),
    C_Vertex3f(0.376432419f, 0.229224205f, -1.f),
    C_Normal3f(-0.731152058f, -0.682214677f, 0.f),
    C_Vertex3f(0.365047455f, 0.2422328f, 0.f),
    C_Vertex3f(0.365047455f, 0.2422328f, -1.f),
    C_Normal3f(-0.675937951f, -0.736958563f, 0.f),
    C_Vertex3f(0.354290009f, 0.25295496f, 0.f),
    C_Vertex3f(0.354290009f, 0.25295496f, -1.f),
    C_Normal3f(-0.586566746f, -0.80990082f, 0.f),
    C_Vertex3f(0.34416008f, 0.261390686f, 0.f),
    C_Vertex3f(0.34416008f, 0.261390686f, -1.f),
    C_Normal3f(-0.468028367f, -0.883713424f, 0.f),
    C_Vertex3f(0.329906464f, 0.270614624f, 0.f),
    C_Vertex3f(0.329906464f, 0.270614624f, -1.f),
    C_Normal3f(-0.352317154f, -0.935880721f, 0.f),
    C_Vertex3f(0.310573578f, 0.279178619f, 0.f),
    C_Vertex3f(0.310573578f, 0.279178619f, -1.f),
    C_Normal3f(-0.230586886f, -0.973051786f, 0.f),
    C_Vertex3f(0.300743103f, 0.281593323f, 0.f),
    C_Vertex3f(0.300743103f, 0.281593323f, -1.f),
    C_Normal3f(-0.145100087f, -0.989417017f, 0.f),
    C_Vertex3f(0.297140121f, 0.282361984f, 0.f),
    C_Vertex3f(0.297140121f, 0.282361984f, -1.f),
    C_Normal3f(-0.0898721814f, -0.995953262f, 0.f),
    C_Vertex3f(0.287639618f, 0.283514977f, 0.f),
    C_Vertex3f(0.287639618f, 0.283514977f, -1.f),
    C_Normal3f(-0.0444993898f, -0.99900943f, 0.f),
    C_Vertex3f(0.275844574f, 0.284283638f, 0.f),
    C_Vertex3f(0.275844574f, 0.284283638f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.26175499f, 0.284667969f, 0.f),
    C_Vertex3f(0.26175499f, 0.284667969f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.26175499f, 0.284667969f, 0.f),
    C_Vertex3f(0.26175499f, 0.284667969f, -1.f),
    C_Vertex3f(0.155273438f, 0.284667969f, -1.f),
    C_Vertex3f(0.155273438f, 0.284667969f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155273438f, 0.284667969f, 0.f),
    C_Vertex3f(0.155273438f, 0.284667969f, -1.f),
    C_Vertex3f(0.155273438f, 0.f, -1.f),
    C_Vertex3f(0.155273438f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.155273438f, 0.f, 0.f),
    C_Vertex3f(0.155273438f, 0.f, -1.f),
    C_Vertex3f(0.0703125f, 0.f, -1.f),
    C_Vertex3f(0.0703125f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.155273438f, 0.357910156f, 0.f),
    C_Vertex3f(0.155273438f, 0.357910156f, -1.f),
    C_Normal3f(-0.00674823625f, 0.999977231f, 0.f),
    C_Vertex3f(0.351966858f, 0.357910156f, 0.f),
    C_Vertex3f(0.351966858f, 0.357910156f, -1.f),
    C_Normal3f(-0.0923255756f, 0.99572891f, 0.f),
    C_Vertex3f(0.377850533f, 0.359412193f, 0.f),
    C_Vertex3f(0.377850533f, 0.359412193f, -1.f),
    C_Normal3f(-0.174971595f, 0.984573483f, 0.f),
    C_Vertex3f(0.400565147f, 0.362416267f, 0.f),
    C_Vertex3f(0.400565147f, 0.362416267f, -1.f),
    C_Normal3f(-0.28091076f, 0.959733844f, 0.f),
    C_Vertex3f(0.420110703f, 0.366922379f, 0.f),
    C_Vertex3f(0.420110703f, 0.366922379f, -1.f),
    C_Normal3f(-0.409481436f, 0.912318408f, 0.f),
    C_Vertex3f(0.436487198f, 0.372930527f, 0.f),
    C_Vertex3f(0.436487198f, 0.372930527f, -1.f),
    C_Normal3f(-0.544242144f, 0.838928282f, 0.f),
    C_Vertex3f(0.450898647f, 0.380741119f, 0.f),
    C_Vertex3f(0.450898647f, 0.380741119f, -1.f),
    C_Normal3f(-0.675923824f, 0.736971498f, 0.f),
    C_Vertex3f(0.463345051f, 0.390354156f, 0.f),
    C_Vertex3f(0.463345051f, 0.390354156f, -1.f),
    C_Normal3f(-0.791863799f, 0.610697746f, 0.f),
    C_Vertex3f(0.473826408f, 0.401769638f, 0.f),
    C_Vertex3f(0.473826408f, 0.401769638f, -1.f),
    C_Normal3f(-0.875549972f, 0.483127594f, 0.f),
    C_Vertex3f(0.48234272f, 0.414987564f, 0.f),
    C_Vertex3f(0.48234272f, 0.414987564f, -1.f),
    C_Normal3f(-0.935373902f, 0.353660226f, 0.f),
    C_Vertex3f(0.488729954f, 0.428778648f, 0.f),
    C_Vertex3f(0.488729954f, 0.428778648f, -1.f),
    C_Normal3f(-0.977056384f, 0.212980807f, 0.f),
    C_Vertex3f(0.49298811f, 0.443142891f, 0.f),
    C_Vertex3f(0.49298811f, 0.443142891f, -1.f),
    C_Normal3f(-0.997196436f, 0.0748291165f, 0.f),
    C_Vertex3f(0.495117188f, 0.458080292f, 0.f),
    C_Vertex3f(0.495117188f, 0.458080292f, -1.f),
    C_Normal3f(-0.9991014f, -0.0423839204f, 0.f),
    C_Vertex3f(0.495117188f, 0.471515656f, 0.f),
    C_Vertex3f(0.495117188f, 0.471515656f, -1.f),
    C_Normal3f(-0.989539802f, -0.144260377f, 0.f),
    C_Vertex3f(0.494086742f, 0.482370615f, 0.f),
    C_Vertex3f(0.494086742f, 0.482370615f, -1.f),
    C_Normal3f(-0.968961f, -0.247213706f, 0.f),
    C_Vertex3f(0.492025852f, 0.492720366f, 0.f),
    C_Vertex3f(0.492025852f, 0.492720366f, -1.f),
    C_Normal3f(-0.936021507f, -0.351943046f, 0.f),
    C_Vertex3f(0.488934517f, 0.502564907f, 0.f),
    C_Vertex3f(0.488934517f, 0.502564907f, -1.f),
    C_Normal3f(-0.890725434f, -0.454541653f, 0.f),
    C_Vertex3f(0.484812737f, 0.51190424f, 0.f),
    C_Vertex3f(0.484812737f, 0.51190424f, -1.f),
    C_Normal3f(-0.83444643f, -0.551089108f, 0.f),
    C_Vertex3f(0.479660511f, 0.520738363f, 0.f),
    C_Vertex3f(0.479660511f, 0.520738363f, -1.f),
    C_Normal3f(-0.76973623f, -0.638362169f, 0.f),
    C_Vertex3f(0.47347784f, 0.529067278f, 0.f),
    C_Vertex3f(0.47347784f, 0.529067278f, -1.f),
    C_Normal3f(-0.699802637f, -0.714336216f, 0.f),
    C_Vertex3f(0.466264725f, 0.536890984f, 0.f),
    C_Vertex3f(0.466264725f, 0.536890984f, -1.f),
    C_Normal3f(-0.613105416f, -0.790001094f, 0.f),
    C_Vertex3f(0.458021164f, 0.54420948f, 0.f),
    C_Vertex3f(0.458021164f, 0.54420948f, -1.f),
    C_Normal3f(-0.509780824f, -0.860304356f, 0.f),
    C_Vertex3f(0.448583364f, 0.550613165f, 0.f),
    C_Vertex3f(0.448583364f, 0.550613165f, -1.f),
    C_Normal3f(-0.408900708f, -0.91257894f, 0.f),
    C_Vertex3f(0.437951326f, 0.556102037f, 0.f),
    C_Vertex3f(0.437951326f, 0.556102037f, -1.f),
    C_Normal3f(-0.314544022f, -0.949242949f, 0.f),
    C_Vertex3f(0.42612505f, 0.560676098f, 0.f),
    C_Vertex3f(0.42612505f, 0.560676098f, -1.f),
    C_Normal3f(-0.228883058f, -0.973453939f, 0.f),
    C_Vertex3f(0.413104534f, 0.564335346f, 0.f),
    C_Vertex3f(0.413104534f, 0.564335346f, -1.f),
    C_Normal3f(-0.152596891f, -0.988288581f, 0.f),
    C_Vertex3f(0.39888978f, 0.567079782f, 0.f),
    C_Vertex3f(0.39888978f, 0.567079782f, -1.f),
    C_Normal3f(-0.0854175612f, -0.996345282f, 0.f),
    C_Vertex3f(0.383480787f, 0.568909407f, 0.f),
    C_Vertex3f(0.383480787f, 0.568909407f, -1.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.366877556f, 0.569824219f, 0.f),
    C_Vertex3f(0.366877556f, 0.569824219f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.366877556f, 0.569824219f, 0.f),
    C_Vertex3f(0.366877556f, 0.569824219f, -1.f),
    C_Vertex3f(0.155273438f, 0.569824219f, -1.f),
    C_Vertex3f(0.155273438f, 0.569824219f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155273438f, 0.569824219f, 0.f),
    C_Vertex3f(0.155273438f, 0.569824219f, -1.f),
    C_Vertex3f(0.155273438f, 0.357910156f, -1.f),
    C_Vertex3f(0.155273438f, 0.357910156f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52919364f, 0.595690012f, 0.f),
    C_Vertex3f(0.520292521f, 0.603309393f, 0.f),
    C_Vertex3f(0.520351648f, 0.334461927f, 0.f),
    C_Vertex3f(0.510831714f, 0.610102534f, 0.f),
    C_Vertex3f(0.506317854f, 0.324903727f, 0.f),
    C_Vertex3f(0.500811219f, 0.616069436f, 0.f),
    C_Vertex3f(0.495117188f, 0.471515656f, 0.f),
    C_Vertex3f(0.494086742f, 0.482370615f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.484812737f, 0.51190424f, 0.f),
    C_Vertex3f(0.488934517f, 0.502564907f, 0.f),
    C_Vertex3f(0.490231037f, 0.621210098f, 0.f),
    C_Vertex3f(0.492025852f, 0.492720366f, 0.f),
    C_Vertex3f(0.500811219f, 0.616069436f, 0.f),
    C_Vertex3f(0.494086742f, 0.482370615f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.458021164f, 0.54420948f, 0.f),
    C_Vertex3f(0.466264725f, 0.536890984f, 0.f),
    C_Vertex3f(0.46666801f, 0.629299641f, 0.f),
    C_Vertex3f(0.47347784f, 0.529067278f, 0.f),
    C_Vertex3f(0.479091167f, 0.625524521f, 0.f),
    C_Vertex3f(0.479660511f, 0.520738363f, 0.f),
    C_Vertex3f(0.490231037f, 0.621210098f, 0.f),
    C_Vertex3f(0.484812737f, 0.51190424f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.42612505f, 0.560676098f, 0.f),
    C_Vertex3f(0.437951326f, 0.556102037f, 0.f),
    C_Vertex3f(0.43797183f, 0.635231972f, 0.f),
    C_Vertex3f(0.448583364f, 0.550613165f, 0.f),
    C_Vertex3f(0.452961564f, 0.632535458f, 0.f),
    C_Vertex3f(0.458021164f, 0.54420948f, 0.f),
    C_Vertex3f(0.46666801f, 0.629299641f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.366877556f, 0.569824219f, 0.f),
    C_Vertex3f(0.383480787f, 0.568909407f, 0.f),
    C_Vertex3f(0.385302901f, 0.640085697f, 0.f),
    C_Vertex3f(0.39888978f, 0.567079782f, 0.f),
    C_Vertex3f(0.404142499f, 0.639007092f, 0.f),
    C_Vertex3f(0.413104534f, 0.564335346f, 0.f),
    C_Vertex3f(0.421698809f, 0.637389183f, 0.f),
    C_Vertex3f(0.42612505f, 0.560676098f, 0.f),
    C_Vertex3f(0.43797183f, 0.635231972f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0703125f, 0.f, 0.f),
    C_Vertex3f(0.155273438f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.155273438f, 0.569824219f, 0.f),
    C_Vertex3f(0.365180016f, 0.640625f, 0.f),
    C_Vertex3f(0.366877556f, 0.569824219f, 0.f),
    C_Vertex3f(0.385302901f, 0.640085697f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.155273438f, 0.569824219f, 0.f),
    C_Vertex3f(0.155273438f, 0.284667969f, 0.f),
    C_Vertex3f(0.155273438f, 0.357910156f, 0.f),
    C_Vertex3f(0.26175499f, 0.284667969f, 0.f),
    C_Vertex3f(0.351966858f, 0.357910156f, 0.f),
    C_Vertex3f(0.275844574f, 0.284283638f, 0.f),
    C_Vertex3f(0.287639618f, 0.283514977f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.574246883f, 0.410216331f, 0.f),
    C_Vertex3f(0.578383207f, 0.425249577f, 0.f),
    C_Vertex3f(0.579188347f, 0.504350185f, 0.f),
    C_Vertex3f(0.581140757f, 0.440951824f, 0.f),
    C_Vertex3f(0.582519531f, 0.479061127f, 0.f),
    C_Vertex3f(0.582519531f, 0.457323074f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.553565264f, 0.369130611f, 0.f),
    C_Vertex3f(0.561837912f, 0.382156849f, 0.f),
    C_Vertex3f(0.562532425f, 0.551487923f, 0.f),
    C_Vertex3f(0.568731785f, 0.395852089f, 0.f),
    C_Vertex3f(0.572525978f, 0.528492451f, 0.f),
    C_Vertex3f(0.574246883f, 0.410216331f, 0.f),
    C_Vertex3f(0.579188347f, 0.504350185f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.562532425f, 0.551487923f, 0.f),
    C_Vertex3f(0.552538872f, 0.567874432f, 0.f),
    C_Vertex3f(0.553565264f, 0.369130611f, 0.f),
    C_Vertex3f(0.545316815f, 0.577972531f, 0.f),
    C_Vertex3f(0.543913841f, 0.356773376f, 0.f),
    C_Vertex3f(0.537535071f, 0.587244391f, 0.f),
    C_Vertex3f(0.532883644f, 0.345085144f, 0.f),
    C_Vertex3f(0.52919364f, 0.595690012f, 0.f),
    C_Vertex3f(0.520351648f, 0.334461927f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.495117188f, 0.471515656f, 0.f),
    C_Vertex3f(0.495117188f, 0.458080292f, 0.f),
    C_Vertex3f(0.506317854f, 0.324903727f, 0.f),
    C_Vertex3f(0.49298811f, 0.443142891f, 0.f),
    C_Vertex3f(0.490782261f, 0.316410542f, 0.f),
    C_Vertex3f(0.488729954f, 0.428778648f, 0.f),
    C_Vertex3f(0.48234272f, 0.414987564f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.450898647f, 0.380741119f, 0.f),
    C_Vertex3f(0.455205679f, 0.302619219f, 0.f),
    C_Vertex3f(0.463345051f, 0.390354156f, 0.f),
    C_Vertex3f(0.473744869f, 0.308982372f, 0.f),
    C_Vertex3f(0.473826408f, 0.401769638f, 0.f),
    C_Vertex3f(0.490782261f, 0.316410542f, 0.f),
    C_Vertex3f(0.48234272f, 0.414987564f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.402099609f, 0.291503906f, 0.f),
    C_Vertex3f(0.413621902f, 0.293087959f, 0.f),
    C_Vertex3f(0.420110703f, 0.366922379f, 0.f),
    C_Vertex3f(0.43516469f, 0.297321081f, 0.f),
    C_Vertex3f(0.436487198f, 0.372930527f, 0.f),
    C_Vertex3f(0.455205679f, 0.302619219f, 0.f),
    C_Vertex3f(0.450898647f, 0.380741119f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.400565147f, 0.362416267f, 0.f),
    C_Vertex3f(0.377850533f, 0.359412193f, 0.f),
    C_Vertex3f(0.388444901f, 0.213929176f, 0.f),
    C_Vertex3f(0.402095795f, 0.19524622f, 0.f),
    C_Vertex3f(0.402099609f, 0.291503906f, 0.f),
    C_Vertex3f(0.420110703f, 0.366922379f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377850533f, 0.359412193f, 0.f),
    C_Vertex3f(0.351966858f, 0.357910156f, 0.f),
    C_Vertex3f(0.354290009f, 0.25295496f, 0.f),
    C_Vertex3f(0.365047455f, 0.2422328f, 0.f),
    C_Vertex3f(0.376432419f, 0.229224205f, 0.f),
    C_Vertex3f(0.388444901f, 0.213929176f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.351966858f, 0.357910156f, 0.f),
    C_Vertex3f(0.287639618f, 0.283514977f, 0.f),
    C_Vertex3f(0.297140121f, 0.282361984f, 0.f),
    C_Vertex3f(0.300743103f, 0.281593323f, 0.f),
    C_Vertex3f(0.310573578f, 0.279178619f, 0.f),
    C_Vertex3f(0.329906464f, 0.270614624f, 0.f),
    C_Vertex3f(0.34416008f, 0.261390686f, 0.f),
    C_Vertex3f(0.354290009f, 0.25295496f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.52835083f, 0.f, 0.f),
    C_Vertex3f(0.634284973f, 0.f, 0.f),
    C_Vertex3f(0.515146255f, 0.187164307f, 0.f),
    C_Vertex3f(0.498112679f, 0.210604668f, 0.f),
    C_Vertex3f(0.480566978f, 0.231911659f, 0.f),
    C_Vertex3f(0.462509155f, 0.251085281f, 0.f),
    C_Vertex3f(0.453224182f, 0.259605408f, 0.f),
    C_Vertex3f(0.448854446f, 0.263538361f, 0.f),
    C_Vertex3f(0.438094139f, 0.271458626f, 0.f),
    C_Vertex3f(0.43431282f, 0.147716522f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.425312996f, 0.27943325f, 0.f),
    C_Vertex3f(0.410511017f, 0.287462234f, 0.f),
    C_Vertex3f(0.417385101f, 0.173175335f, 0.f),
    C_Vertex3f(0.43431282f, 0.147716522f, 0.f),
    C_Vertex3f(0.438094139f, 0.271458626f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402099609f, 0.291503906f, 0.f),
    C_Vertex3f(0.402095795f, 0.19524622f, 0.f),
    C_Vertex3f(0.417385101f, 0.173175335f, 0.f),
    C_Vertex3f(0.410511017f, 0.287462234f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.155273438f, 0.569824219f, 0.f),
    C_Vertex3f(0.155273438f, 0.f, 0.f),
    C_Vertex3f(0.155273438f, 0.284667969f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52919364f, 0.595690012f, -1.f),
    C_Vertex3f(0.520292521f, 0.603309393f, -1.f),
    C_Vertex3f(0.520351648f, 0.334461927f, -1.f),
    C_Vertex3f(0.510831714f, 0.610102534f, -1.f),
    C_Vertex3f(0.506317854f, 0.324903727f, -1.f),
    C_Vertex3f(0.500811219f, 0.616069436f, -1.f),
    C_Vertex3f(0.495117188f, 0.471515656f, -1.f),
    C_Vertex3f(0.494086742f, 0.482370615f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.484812737f, 0.51190424f, -1.f),
    C_Vertex3f(0.488934517f, 0.502564907f, -1.f),
    C_Vertex3f(0.490231037f, 0.621210098f, -1.f),
    C_Vertex3f(0.492025852f, 0.492720366f, -1.f),
    C_Vertex3f(0.500811219f, 0.616069436f, -1.f),
    C_Vertex3f(0.494086742f, 0.482370615f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.458021164f, 0.54420948f, -1.f),
    C_Vertex3f(0.466264725f, 0.536890984f, -1.f),
    C_Vertex3f(0.46666801f, 0.629299641f, -1.f),
    C_Vertex3f(0.47347784f, 0.529067278f, -1.f),
    C_Vertex3f(0.479091167f, 0.625524521f, -1.f),
    C_Vertex3f(0.479660511f, 0.520738363f, -1.f),
    C_Vertex3f(0.490231037f, 0.621210098f, -1.f),
    C_Vertex3f(0.484812737f, 0.51190424f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.42612505f, 0.560676098f, -1.f),
    C_Vertex3f(0.437951326f, 0.556102037f, -1.f),
    C_Vertex3f(0.43797183f, 0.635231972f, -1.f),
    C_Vertex3f(0.448583364f, 0.550613165f, -1.f),
    C_Vertex3f(0.452961564f, 0.632535458f, -1.f),
    C_Vertex3f(0.458021164f, 0.54420948f, -1.f),
    C_Vertex3f(0.46666801f, 0.629299641f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.366877556f, 0.569824219f, -1.f),
    C_Vertex3f(0.383480787f, 0.568909407f, -1.f),
    C_Vertex3f(0.385302901f, 0.640085697f, -1.f),
    C_Vertex3f(0.39888978f, 0.567079782f, -1.f),
    C_Vertex3f(0.404142499f, 0.639007092f, -1.f),
    C_Vertex3f(0.413104534f, 0.564335346f, -1.f),
    C_Vertex3f(0.421698809f, 0.637389183f, -1.f),
    C_Vertex3f(0.42612505f, 0.560676098f, -1.f),
    C_Vertex3f(0.43797183f, 0.635231972f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0703125f, 0.f, -1.f),
    C_Vertex3f(0.155273438f, 0.f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.569824219f, -1.f),
    C_Vertex3f(0.365180016f, 0.640625f, -1.f),
    C_Vertex3f(0.366877556f, 0.569824219f, -1.f),
    C_Vertex3f(0.385302901f, 0.640085697f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.155273438f, 0.569824219f, -1.f),
    C_Vertex3f(0.155273438f, 0.284667969f, -1.f),
    C_Vertex3f(0.155273438f, 0.357910156f, -1.f),
    C_Vertex3f(0.26175499f, 0.284667969f, -1.f),
    C_Vertex3f(0.351966858f, 0.357910156f, -1.f),
    C_Vertex3f(0.275844574f, 0.284283638f, -1.f),
    C_Vertex3f(0.287639618f, 0.283514977f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.574246883f, 0.410216331f, -1.f),
    C_Vertex3f(0.578383207f, 0.425249577f, -1.f),
    C_Vertex3f(0.579188347f, 0.504350185f, -1.f),
    C_Vertex3f(0.581140757f, 0.440951824f, -1.f),
    C_Vertex3f(0.582519531f, 0.479061127f, -1.f),
    C_Vertex3f(0.582519531f, 0.457323074f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.553565264f, 0.369130611f, -1.f),
    C_Vertex3f(0.561837912f, 0.382156849f, -1.f),
    C_Vertex3f(0.562532425f, 0.551487923f, -1.f),
    C_Vertex3f(0.568731785f, 0.395852089f, -1.f),
    C_Vertex3f(0.572525978f, 0.528492451f, -1.f),
    C_Vertex3f(0.574246883f, 0.410216331f, -1.f),
    C_Vertex3f(0.579188347f, 0.504350185f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.562532425f, 0.551487923f, -1.f),
    C_Vertex3f(0.552538872f, 0.567874432f, -1.f),
    C_Vertex3f(0.553565264f, 0.369130611f, -1.f),
    C_Vertex3f(0.545316815f, 0.577972531f, -1.f),
    C_Vertex3f(0.543913841f, 0.356773376f, -1.f),
    C_Vertex3f(0.537535071f, 0.587244391f, -1.f),
    C_Vertex3f(0.532883644f, 0.345085144f, -1.f),
    C_Vertex3f(0.52919364f, 0.595690012f, -1.f),
    C_Vertex3f(0.520351648f, 0.334461927f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.495117188f, 0.471515656f, -1.f),
    C_Vertex3f(0.495117188f, 0.458080292f, -1.f),
    C_Vertex3f(0.506317854f, 0.324903727f, -1.f),
    C_Vertex3f(0.49298811f, 0.443142891f, -1.f),
    C_Vertex3f(0.490782261f, 0.316410542f, -1.f),
    C_Vertex3f(0.488729954f, 0.428778648f, -1.f),
    C_Vertex3f(0.48234272f, 0.414987564f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.450898647f, 0.380741119f, -1.f),
    C_Vertex3f(0.455205679f, 0.302619219f, -1.f),
    C_Vertex3f(0.463345051f, 0.390354156f, -1.f),
    C_Vertex3f(0.473744869f, 0.308982372f, -1.f),
    C_Vertex3f(0.473826408f, 0.401769638f, -1.f),
    C_Vertex3f(0.490782261f, 0.316410542f, -1.f),
    C_Vertex3f(0.48234272f, 0.414987564f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.402099609f, 0.291503906f, -1.f),
    C_Vertex3f(0.413621902f, 0.293087959f, -1.f),
    C_Vertex3f(0.420110703f, 0.366922379f, -1.f),
    C_Vertex3f(0.43516469f, 0.297321081f, -1.f),
    C_Vertex3f(0.436487198f, 0.372930527f, -1.f),
    C_Vertex3f(0.455205679f, 0.302619219f, -1.f),
    C_Vertex3f(0.450898647f, 0.380741119f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.400565147f, 0.362416267f, -1.f),
    C_Vertex3f(0.377850533f, 0.359412193f, -1.f),
    C_Vertex3f(0.388444901f, 0.213929176f, -1.f),
    C_Vertex3f(0.402095795f, 0.19524622f, -1.f),
    C_Vertex3f(0.402099609f, 0.291503906f, -1.f),
    C_Vertex3f(0.420110703f, 0.366922379f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377850533f, 0.359412193f, -1.f),
    C_Vertex3f(0.351966858f, 0.357910156f, -1.f),
    C_Vertex3f(0.354290009f, 0.25295496f, -1.f),
    C_Vertex3f(0.365047455f, 0.2422328f, -1.f),
    C_Vertex3f(0.376432419f, 0.229224205f, -1.f),
    C_Vertex3f(0.388444901f, 0.213929176f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.351966858f, 0.357910156f, -1.f),
    C_Vertex3f(0.287639618f, 0.283514977f, -1.f),
    C_Vertex3f(0.297140121f, 0.282361984f, -1.f),
    C_Vertex3f(0.300743103f, 0.281593323f, -1.f),
    C_Vertex3f(0.310573578f, 0.279178619f, -1.f),
    C_Vertex3f(0.329906464f, 0.270614624f, -1.f),
    C_Vertex3f(0.34416008f, 0.261390686f, -1.f),
    C_Vertex3f(0.354290009f, 0.25295496f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.52835083f, 0.f, -1.f),
    C_Vertex3f(0.634284973f, 0.f, -1.f),
    C_Vertex3f(0.515146255f, 0.187164307f, -1.f),
    C_Vertex3f(0.498112679f, 0.210604668f, -1.f),
    C_Vertex3f(0.480566978f, 0.231911659f, -1.f),
    C_Vertex3f(0.462509155f, 0.251085281f, -1.f),
    C_Vertex3f(0.453224182f, 0.259605408f, -1.f),
    C_Vertex3f(0.448854446f, 0.263538361f, -1.f),
    C_Vertex3f(0.438094139f, 0.271458626f, -1.f),
    C_Vertex3f(0.43431282f, 0.147716522f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.425312996f, 0.27943325f, -1.f),
    C_Vertex3f(0.410511017f, 0.287462234f, -1.f),
    C_Vertex3f(0.417385101f, 0.173175335f, -1.f),
    C_Vertex3f(0.43431282f, 0.147716522f, -1.f),
    C_Vertex3f(0.438094139f, 0.271458626f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402099609f, 0.291503906f, -1.f),
    C_Vertex3f(0.402095795f, 0.19524622f, -1.f),
    C_Vertex3f(0.417385101f, 0.173175335f, -1.f),
    C_Vertex3f(0.410511017f, 0.287462234f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.155273438f, 0.569824219f, -1.f),
    C_Vertex3f(0.155273438f, 0.f, -1.f),
    C_Vertex3f(0.155273438f, 0.284667969f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_83[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.0850564986f, 0.996376157f, 0.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, 0.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, -1.f),
    C_Vertex3f(0.120117188f, 0.212890625f, -1.f),
    C_Vertex3f(0.120117188f, 0.212890625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.993104398f, 0.117233843f, 0.f),
    C_Vertex3f(0.120117188f, 0.212890625f, 0.f),
    C_Vertex3f(0.120117188f, 0.212890625f, -1.f),
    C_Normal3f(0.98398751f, 0.17823717f, 0.f),
    C_Vertex3f(0.121538162f, 0.200853348f, 0.f),
    C_Vertex3f(0.121538162f, 0.200853348f, -1.f),
    C_Normal3f(0.964875579f, 0.262707204f, 0.f),
    C_Vertex3f(0.126266956f, 0.178939819f, 0.f),
    C_Vertex3f(0.126266956f, 0.178939819f, -1.f),
    C_Normal3f(0.926923394f, 0.375250369f, 0.f),
    C_Vertex3f(0.132882595f, 0.159187317f, 0.f),
    C_Vertex3f(0.132882595f, 0.159187317f, -1.f),
    C_Normal3f(0.867992342f, 0.496577561f, 0.f),
    C_Vertex3f(0.141385078f, 0.14159584f, 0.f),
    C_Vertex3f(0.141385078f, 0.14159584f, -1.f),
    C_Normal3f(0.77902025f, 0.626998723f, 0.f),
    C_Vertex3f(0.151774406f, 0.12616539f, 0.f),
    C_Vertex3f(0.151774406f, 0.12616539f, -1.f),
    C_Normal3f(0.668635428f, 0.743590415f, 0.f),
    C_Vertex3f(0.165034294f, 0.112212658f, 0.f),
    C_Vertex3f(0.165034294f, 0.112212658f, -1.f),
    C_Normal3f(0.555541098f, 0.831489027f, 0.f),
    C_Vertex3f(0.181164742f, 0.0997376442f, 0.f),
    C_Vertex3f(0.181164742f, 0.0997376442f, -1.f),
    C_Normal3f(0.448622704f, 0.893721223f, 0.f),
    C_Vertex3f(0.200165749f, 0.0887403488f, 0.f),
    C_Vertex3f(0.200165749f, 0.0887403488f, -1.f),
    C_Normal3f(0.346432209f, 0.938074946f, 0.f),
    C_Vertex3f(0.222037315f, 0.0792207718f, 0.f),
    C_Vertex3f(0.222037315f, 0.0792207718f, -1.f),
    C_Normal3f(0.241358921f, 0.970435977f, 0.f),
    C_Vertex3f(0.245275974f, 0.072081089f, 0.f),
    C_Vertex3f(0.245275974f, 0.072081089f, -1.f),
    C_Normal3f(0.139774442f, 0.990183294f, 0.f),
    C_Vertex3f(0.269881725f, 0.0673213005f, 0.f),
    C_Vertex3f(0.269881725f, 0.0673213005f, -1.f),
    C_Normal3f(0.0459121391f, 0.998945475f, 0.f),
    C_Vertex3f(0.295854568f, 0.0649414063f, 0.f),
    C_Vertex3f(0.295854568f, 0.0649414063f, -1.f),
    C_Normal3f(-0.0370819233f, 0.999312282f, 0.f),
    C_Vertex3f(0.321662903f, 0.0649414063f, 0.f),
    C_Vertex3f(0.321662903f, 0.0649414063f, -1.f),
    C_Normal3f(-0.121412903f, 0.99260205f, 0.f),
    C_Vertex3f(0.344518185f, 0.0667471886f, 0.f),
    C_Vertex3f(0.344518185f, 0.0667471886f, -1.f),
    C_Normal3f(-0.212854326f, 0.977083981f, 0.f),
    C_Vertex3f(0.365952015f, 0.0703587532f, 0.f),
    C_Vertex3f(0.365952015f, 0.0703587532f, -1.f),
    C_Normal3f(-0.311187327f, 0.950348616f, 0.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, 0.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, -1.f),
    C_Normal3f(-0.413075179f, 0.910696864f, 0.f),
    C_Vertex3f(0.404555321f, 0.0829992294f, 0.f),
    C_Vertex3f(0.404555321f, 0.0829992294f, -1.f),
    C_Normal3f(-0.523600936f, 0.85196358f, 0.f),
    C_Vertex3f(0.420767784f, 0.0915622711f, 0.f),
    C_Vertex3f(0.420767784f, 0.0915622711f, -1.f),
    C_Normal3f(-0.641458333f, 0.767157912f, 0.f),
    C_Vertex3f(0.434601784f, 0.101465225f, 0.f),
    C_Vertex3f(0.434601784f, 0.101465225f, -1.f),
    C_Normal3f(-0.757515967f, 0.652816653f, 0.f),
    C_Vertex3f(0.44605732f, 0.112708092f, 0.f),
    C_Vertex3f(0.44605732f, 0.112708092f, -1.f),
    C_Normal3f(-0.850961983f, 0.525227368f, 0.f),
    C_Vertex3f(0.455134392f, 0.125290871f, 0.f),
    C_Vertex3f(0.455134392f, 0.125290871f, -1.f),
    C_Normal3f(-0.921275377f, 0.388911009f, 0.f),
    C_Vertex3f(0.461942196f, 0.138444424f, 0.f),
    C_Vertex3f(0.461942196f, 0.138444424f, -1.f),
    C_Normal3f(-0.971729159f, 0.236098394f, 0.f),
    C_Vertex3f(0.466480732f, 0.152168751f, 0.f),
    C_Vertex3f(0.466480732f, 0.152168751f, -1.f),
    C_Normal3f(-0.997008026f, 0.0772976428f, 0.f),
    C_Vertex3f(0.46875f, 0.166463852f, 0.f),
    C_Vertex3f(0.46875f, 0.166463852f, -1.f),
    C_Normal3f(-0.997219086f, -0.0745259076f, 0.f),
    C_Vertex3f(0.46875f, 0.181438446f, 0.f),
    C_Vertex3f(0.46875f, 0.181438446f, -1.f),
    C_Normal3f(-0.972602963f, -0.232472599f, 0.f),
    C_Vertex3f(0.466575146f, 0.195565224f, 0.f),
    C_Vertex3f(0.466575146f, 0.195565224f, -1.f),
    C_Normal3f(-0.919204891f, -0.393779665f, 0.f),
    C_Vertex3f(0.462225437f, 0.208735466f, 0.f),
    C_Vertex3f(0.462225437f, 0.208735466f, -1.f),
    C_Normal3f(-0.838967443f, -0.544181705f, 0.f),
    C_Vertex3f(0.455700874f, 0.220949173f, 0.f),
    C_Vertex3f(0.455700874f, 0.220949173f, -1.f),
    C_Normal3f(-0.729361534f, -0.684128463f, 0.f),
    C_Vertex3f(0.447001457f, 0.232206345f, 0.f),
    C_Vertex3f(0.447001457f, 0.232206345f, -1.f),
    C_Normal3f(-0.605762899f, -0.795645237f, 0.f),
    C_Vertex3f(0.435428143f, 0.242562294f, 0.f),
    C_Vertex3f(0.435428143f, 0.242562294f, -1.f),
    C_Normal3f(-0.493140727f, -0.869949579f, 0.f),
    C_Vertex3f(0.42098093f, 0.252017021f, 0.f),
    C_Vertex3f(0.42098093f, 0.252017021f, -1.f),
    C_Normal3f(-0.406449646f, -0.913673222f, 0.f),
    C_Vertex3f(0.403659821f, 0.260570526f, 0.f),
    C_Vertex3f(0.403659821f, 0.260570526f, -1.f),
    C_Normal3f(-0.342922598f, -0.939363658f, 0.f),
    C_Vertex3f(0.390324593f, 0.265654564f, 0.f),
    C_Vertex3f(0.390324593f, 0.265654564f, -1.f),
    C_Normal3f(-0.307028621f, -0.951700211f, 0.f),
    C_Vertex3f(0.381887197f, 0.268518806f, 0.f),
    C_Vertex3f(0.381887197f, 0.268518806f, -1.f),
    C_Normal3f(-0.285247117f, -0.958454013f, 0.f),
    C_Vertex3f(0.371487856f, 0.271731496f, 0.f),
    C_Vertex3f(0.371487856f, 0.271731496f, -1.f),
    C_Normal3f(-0.269598514f, -0.962972879f, 0.f),
    C_Vertex3f(0.359126568f, 0.275292635f, 0.f),
    C_Vertex3f(0.359126568f, 0.275292635f, -1.f),
    C_Normal3f(-0.257821202f, -0.966192603f, 0.f),
    C_Vertex3f(0.344803333f, 0.279202223f, 0.f),
    C_Vertex3f(0.344803333f, 0.279202223f, -1.f),
    C_Normal3f(-0.248640731f, -0.968595803f, 0.f),
    C_Vertex3f(0.328518152f, 0.283460259f, 0.f),
    C_Vertex3f(0.328518152f, 0.283460259f, -1.f),
    C_Normal3f(-0.241285399f, -0.970454216f, 0.f),
    C_Vertex3f(0.310271025f, 0.288066745f, 0.f),
    C_Vertex3f(0.310271025f, 0.288066745f, -1.f),
    C_Normal3f(-0.235261127f, -0.971932232f, 0.f),
    C_Vertex3f(0.290061951f, 0.293021679f, 0.f),
    C_Vertex3f(0.290061951f, 0.293021679f, -1.f),
    C_Normal3f(-0.239585206f, -0.970875382f, 0.f),
    C_Vertex3f(0.26789093f, 0.298325062f, 0.f),
    C_Vertex3f(0.26789093f, 0.298325062f, -1.f),
    C_Normal3f(-0.255102754f, -0.966913939f, 0.f),
    C_Vertex3f(0.247384787f, 0.303553224f, 0.f),
    C_Vertex3f(0.247384787f, 0.303553224f, -1.f),
    C_Normal3f(-0.273236841f, -0.961946726f, 0.f),
    C_Vertex3f(0.22854352f, 0.308706164f, 0.f),
    C_Vertex3f(0.22854352f, 0.308706164f, -1.f),
    C_Normal3f(-0.294684052f, -0.955594778f, 0.f),
    C_Vertex3f(0.21136713f, 0.313783884f, 0.f),
    C_Vertex3f(0.21136713f, 0.313783884f, -1.f),
    C_Normal3f(-0.32039851f, -0.947282851f, 0.f),
    C_Vertex3f(0.195855618f, 0.318786383f, 0.f),
    C_Vertex3f(0.195855618f, 0.318786383f, -1.f),
    C_Normal3f(-0.351712555f, -0.936108053f, 0.f),
    C_Vertex3f(0.182008982f, 0.32371366f, 0.f),
    C_Vertex3f(0.182008982f, 0.32371366f, -1.f),
    C_Normal3f(-0.390521169f, -0.920593977f, 0.f),
    C_Vertex3f(0.169827223f, 0.328565717f, 0.f),
    C_Vertex3f(0.169827223f, 0.328565717f, -1.f),
    C_Normal3f(-0.445337921f, -0.895362616f, 0.f),
    C_Vertex3f(0.159310341f, 0.333342552f, 0.f),
    C_Vertex3f(0.159310341f, 0.333342552f, -1.f),
    C_Normal3f(-0.519384444f, -0.854540706f, 0.f),
    C_Vertex3f(0.143409729f, 0.341705322f, 0.f),
    C_Vertex3f(0.143409729f, 0.341705322f, -1.f),
    C_Normal3f(-0.607493222f, -0.794324815f, 0.f),
    C_Vertex3f(0.123383999f, 0.355178356f, 0.f),
    C_Vertex3f(0.123383999f, 0.355178356f, -1.f),
    C_Normal3f(-0.70724684f, -0.706966639f, 0.f),
    C_Vertex3f(0.106281757f, 0.370100498f, 0.f),
    C_Vertex3f(0.106281757f, 0.370100498f, -1.f),
    C_Normal3f(-0.802355707f, -0.596846163f, 0.f),
    C_Vertex3f(0.0921030045f, 0.386471748f, 0.f),
    C_Vertex3f(0.0921030045f, 0.386471748f, -1.f),
    C_Normal3f(-0.881154537f, -0.472828329f, 0.f),
    C_Vertex3f(0.0808477402f, 0.404292107f, 0.f),
    C_Vertex3f(0.0808477402f, 0.404292107f, -1.f),
    C_Normal3f(-0.940210938f, -0.340592563f, 0.f),
    C_Vertex3f(0.072406292f, 0.423178196f, 0.f),
    C_Vertex3f(0.072406292f, 0.423178196f, -1.f),
    C_Normal3f(-0.979426026f, -0.201803774f, 0.f),
    C_Vertex3f(0.0667786598f, 0.443130016f, 0.f),
    C_Vertex3f(0.0667786598f, 0.443130016f, -1.f),
    C_Normal3f(-0.997979105f, -0.0635413378f, 0.f),
    C_Vertex3f(0.0639648438f, 0.464147568f, 0.f),
    C_Vertex3f(0.0639648438f, 0.464147568f, -1.f),
    C_Normal3f(-0.997278392f, 0.0737271234f, 0.f),
    C_Vertex3f(0.0639648438f, 0.487323761f, 0.f),
    C_Vertex3f(0.0639648438f, 0.487323761f, -1.f),
    C_Normal3f(-0.975787222f, 0.218722045f, 0.f),
    C_Vertex3f(0.0674138069f, 0.510800362f, 0.f),
    C_Vertex3f(0.0674138069f, 0.510800362f, -1.f),
    C_Normal3f(-0.932640553f, 0.360806853f, 0.f),
    C_Vertex3f(0.0743117332f, 0.533484459f, 0.f),
    C_Vertex3f(0.0743117332f, 0.533484459f, -1.f),
    C_Normal3f(-0.876705945f, 0.481026679f, 0.f),
    C_Vertex3f(0.0846586227f, 0.555376053f, 0.f),
    C_Vertex3f(0.0846586227f, 0.555376053f, -1.f),
    C_Normal3f(-0.818909228f, 0.573922992f, 0.f),
    C_Vertex3f(0.0950055122f, 0.571200371f, 0.f),
    C_Vertex3f(0.0950055122f, 0.571200371f, -1.f),
    C_Normal3f(-0.764627099f, 0.644472897f, 0.f),
    C_Vertex3f(0.102697134f, 0.581114531f, 0.f),
    C_Vertex3f(0.102697134f, 0.581114531f, -1.f),
    C_Normal3f(-0.710227191f, 0.703972518f, 0.f),
    C_Vertex3f(0.111182451f, 0.590393305f, 0.f),
    C_Vertex3f(0.111182451f, 0.590393305f, -1.f),
    C_Normal3f(-0.652239919f, 0.758012652f, 0.f),
    C_Vertex3f(0.120461464f, 0.599036694f, 0.f),
    C_Vertex3f(0.120461464f, 0.599036694f, -1.f),
    C_Normal3f(-0.591996014f, 0.805940866f, 0.f),
    C_Vertex3f(0.130534172f, 0.607044697f, 0.f),
    C_Vertex3f(0.130534172f, 0.607044697f, -1.f),
    C_Normal3f(-0.530831516f, 0.847477317f, 0.f),
    C_Vertex3f(0.141400576f, 0.614417315f, 0.f),
    C_Vertex3f(0.141400576f, 0.614417315f, -1.f),
    C_Normal3f(-0.469970703f, 0.882681966f, 0.f),
    C_Vertex3f(0.153060675f, 0.621154547f, 0.f),
    C_Vertex3f(0.153060675f, 0.621154547f, -1.f),
    C_Normal3f(-0.404598981f, 0.914494216f, 0.f),
    C_Vertex3f(0.165514469f, 0.627256393f, 0.f),
    C_Vertex3f(0.165514469f, 0.627256393f, -1.f),
    C_Normal3f(-0.324249119f, 0.945971787f, 0.f),
    C_Vertex3f(0.185385704f, 0.635456085f, 0.f),
    C_Vertex3f(0.185385704f, 0.635456085f, -1.f),
    C_Normal3f(-0.23152405f, 0.972829163f, 0.f),
    C_Vertex3f(0.213358402f, 0.643655777f, 0.f),
    C_Vertex3f(0.213358402f, 0.643655777f, -1.f),
    C_Normal3f(-0.134569451f, 0.990904152f, 0.f),
    C_Vertex3f(0.242808819f, 0.649122238f, 0.f),
    C_Vertex3f(0.242808819f, 0.649122238f, -1.f),
    C_Normal3f(-0.0420266986f, 0.99911654f, 0.f),
    C_Vertex3f(0.273736954f, 0.651855469f, 0.f),
    C_Vertex3f(0.273736954f, 0.651855469f, -1.f),
    C_Normal3f(0.0423609316f, 0.999102354f, 0.f),
    C_Vertex3f(0.307786942f, 0.651855469f, 0.f),
    C_Vertex3f(0.307786942f, 0.651855469f, -1.f),
    C_Normal3f(0.131084278f, 0.991371155f, 0.f),
    C_Vertex3f(0.341372967f, 0.64898777f, 0.f),
    C_Vertex3f(0.341372967f, 0.64898777f, -1.f),
    C_Normal3f(0.229363292f, 0.973340929f, 0.f),
    C_Vertex3f(0.372850895f, 0.643252373f, 0.f),
    C_Vertex3f(0.372850895f, 0.643252373f, -1.f),
    C_Normal3f(0.326469094f, 0.945207834f, 0.f),
    C_Vertex3f(0.402220726f, 0.634649277f, 0.f),
    C_Vertex3f(0.402220726f, 0.634649277f, -1.f),
    C_Normal3f(0.411136717f, 0.911573768f, 0.f),
    C_Vertex3f(0.422667027f, 0.626046181f, 0.f),
    C_Vertex3f(0.422667027f, 0.626046181f, -1.f),
    C_Normal3f(0.47657907f, 0.879131556f, 0.f),
    C_Vertex3f(0.435510993f, 0.619634748f, 0.f),
    C_Vertex3f(0.435510993f, 0.619634748f, -1.f),
    C_Normal3f(0.537039518f, 0.84355706f, 0.f),
    C_Vertex3f(0.447568059f, 0.612547278f, 0.f),
    C_Vertex3f(0.447568059f, 0.612547278f, -1.f),
    C_Normal3f(0.597353518f, 0.801978111f, 0.f),
    C_Vertex3f(0.458838224f, 0.604783773f, 0.f),
    C_Vertex3f(0.458838224f, 0.604783773f, -1.f),
    C_Normal3f(0.656340778f, 0.754464507f, 0.f),
    C_Vertex3f(0.469321489f, 0.596344233f, 0.f),
    C_Vertex3f(0.469321489f, 0.596344233f, -1.f),
    C_Normal3f(0.712764323f, 0.701403618f, 0.f),
    C_Vertex3f(0.479017854f, 0.587228656f, 0.f),
    C_Vertex3f(0.479017854f, 0.587228656f, -1.f),
    C_Normal3f(0.765439689f, 0.643507659f, 0.f),
    C_Vertex3f(0.487927318f, 0.577437043f, 0.f),
    C_Vertex3f(0.487927318f, 0.577437043f, -1.f),
    C_Normal3f(0.817846715f, 0.575436234f, 0.f),
    C_Vertex3f(0.496049881f, 0.566969395f, 0.f),
    C_Vertex3f(0.496049881f, 0.566969395f, -1.f),
    C_Normal3f(0.876051962f, 0.482216626f, 0.f),
    C_Vertex3f(0.507053375f, 0.550253868f, 0.f),
    C_Vertex3f(0.507053375f, 0.550253868f, -1.f),
    C_Normal3f(0.931329012f, 0.364178926f, 0.f),
    C_Vertex3f(0.518330574f, 0.526491642f, 0.f),
    C_Vertex3f(0.518330574f, 0.526491642f, -1.f),
    C_Normal3f(0.972792685f, 0.231677249f, 0.f),
    C_Vertex3f(0.526213646f, 0.501254559f, 0.f),
    C_Vertex3f(0.526213646f, 0.501254559f, -1.f),
    C_Normal3f(0.992469192f, 0.122494414f, 0.f),
    C_Vertex3f(0.530702591f, 0.474542618f, 0.f),
    C_Vertex3f(0.530702591f, 0.474542618f, -1.f),
    C_Normal3f(0.999246538f, 0.0388122536f, 0.f),
    C_Vertex3f(0.53125f, 0.460449219f, 0.f),
    C_Vertex3f(0.53125f, 0.460449219f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.0780742019f, -0.996947527f, 0.f),
    C_Vertex3f(0.53125f, 0.460449219f, 0.f),
    C_Vertex3f(0.53125f, 0.460449219f, -1.f),
    C_Vertex3f(0.450195313f, 0.454101563f, -1.f),
    C_Vertex3f(0.450195313f, 0.454101563f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99420023f, -0.1075451f, 0.f),
    C_Vertex3f(0.450195313f, 0.454101563f, 0.f),
    C_Vertex3f(0.450195313f, 0.454101563f, -1.f),
    C_Normal3f(-0.987811804f, -0.155653521f, 0.f),
    C_Vertex3f(0.449372292f, 0.461709976f, 0.f),
    C_Vertex3f(0.449372292f, 0.461709976f, -1.f),
    C_Normal3f(-0.97549212f, -0.22003451f, 0.f),
    C_Vertex3f(0.446745276f, 0.475996256f, 0.f),
    C_Vertex3f(0.446745276f, 0.475996256f, -1.f),
    C_Normal3f(-0.952990592f, -0.302999943f, 0.f),
    C_Vertex3f(0.443137288f, 0.489351988f, 0.f),
    C_Vertex3f(0.443137288f, 0.489351988f, -1.f),
    C_Normal3f(-0.920428216f, -0.390911579f, 0.f),
    C_Vertex3f(0.438548326f, 0.501777172f, 0.f),
    C_Vertex3f(0.438548326f, 0.501777172f, -1.f),
    C_Normal3f(-0.876408041f, -0.48156926f, 0.f),
    C_Vertex3f(0.432978392f, 0.513271809f, 0.f),
    C_Vertex3f(0.432978392f, 0.513271809f, -1.f),
    C_Normal3f(-0.820290387f, -0.571947157f, 0.f),
    C_Vertex3f(0.426427484f, 0.523835897f, 0.f),
    C_Vertex3f(0.426427484f, 0.523835897f, -1.f),
    C_Normal3f(-0.752571285f, -0.658510864f, 0.f),
    C_Vertex3f(0.418895602f, 0.533469439f, 0.f),
    C_Vertex3f(0.418895602f, 0.533469439f, -1.f),
    C_Normal3f(-0.675041139f, -0.737780094f, 0.f),
    C_Vertex3f(0.410382748f, 0.542172432f, 0.f),
    C_Vertex3f(0.410382748f, 0.542172432f, -1.f),
    C_Normal3f(-0.586615622f, -0.809865475f, 0.f),
    C_Vertex3f(0.40088892f, 0.549944878f, 0.f),
    C_Vertex3f(0.40088892f, 0.549944878f, -1.f),
    C_Normal3f(-0.491431266f, -0.870916426f, 0.f),
    C_Vertex3f(0.3902632f, 0.556745768f, 0.f),
    C_Vertex3f(0.3902632f, 0.556745768f, -1.f),
    C_Normal3f(-0.397817284f, -0.917464733f, 0.f),
    C_Vertex3f(0.378505588f, 0.562575102f, 0.f),
    C_Vertex3f(0.378505588f, 0.562575102f, -1.f),
    C_Normal3f(-0.309020817f, -0.951055288f, 0.f),
    C_Vertex3f(0.365616083f, 0.56743288f, 0.f),
    C_Vertex3f(0.365616083f, 0.56743288f, -1.f),
    C_Normal3f(-0.227022663f, -0.97388941f, 0.f),
    C_Vertex3f(0.351594687f, 0.571319103f, 0.f),
    C_Vertex3f(0.351594687f, 0.571319103f, -1.f),
    C_Normal3f(-0.15270482f, -0.988271832f, 0.f),
    C_Vertex3f(0.336441398f, 0.57423377f, 0.f),
    C_Vertex3f(0.336441398f, 0.57423377f, -1.f),
    C_Normal3f(-0.0861612409f, -0.996281266f, 0.f),
    C_Vertex3f(0.320156217f, 0.576176882f, 0.f),
    C_Vertex3f(0.320156217f, 0.576176882f, -1.f),
    C_Normal3f(-0.0267179906f, -0.999642968f, 0.f),
    C_Vertex3f(0.302739143f, 0.577148438f, 0.f),
    C_Vertex3f(0.302739143f, 0.577148438f, -1.f),
    C_Normal3f(0.0238604303f, -0.999715269f, 0.f),
    C_Vertex3f(0.283805847f, 0.577148438f, 0.f),
    C_Vertex3f(0.283805847f, 0.577148438f, -1.f),
    C_Normal3f(0.0759948045f, -0.997108161f, 0.f),
    C_Vertex3f(0.26579845f, 0.576266766f, 0.f),
    C_Vertex3f(0.26579845f, 0.576266766f, -1.f),
    C_Normal3f(0.136121199f, -0.990692139f, 0.f),
    C_Vertex3f(0.249101281f, 0.574503422f, 0.f),
    C_Vertex3f(0.249101281f, 0.574503422f, -1.f),
    C_Normal3f(0.205018878f, -0.978758037f, 0.f),
    C_Vertex3f(0.233714342f, 0.571858406f, 0.f),
    C_Vertex3f(0.233714342f, 0.571858406f, -1.f),
    C_Normal3f(0.283481002f, -0.958977818f, 0.f),
    C_Vertex3f(0.219637632f, 0.568331718f, 0.f),
    C_Vertex3f(0.219637632f, 0.568331718f, -1.f),
    C_Normal3f(0.371697843f, -0.928353667f, 0.f),
    C_Vertex3f(0.206871152f, 0.563923359f, 0.f),
    C_Vertex3f(0.206871152f, 0.563923359f, -1.f),
    C_Normal3f(0.468693584f, -0.883360744f, 0.f),
    C_Vertex3f(0.195414901f, 0.558633327f, 0.f),
    C_Vertex3f(0.195414901f, 0.558633327f, -1.f),
    C_Normal3f(0.571656764f, -0.820492864f, 0.f),
    C_Vertex3f(0.185268879f, 0.552461624f, 0.f),
    C_Vertex3f(0.185268879f, 0.552461624f, -1.f),
    C_Normal3f(0.657800972f, -0.75319171f, 0.f),
    C_Vertex3f(0.176433086f, 0.545408249f, 0.f),
    C_Vertex3f(0.176433086f, 0.545408249f, -1.f),
    C_Normal3f(0.726798058f, -0.686851263f, 0.f),
    C_Vertex3f(0.168701768f, 0.537992716f, 0.f),
    C_Vertex3f(0.168701768f, 0.537992716f, -1.f),
    C_Normal3f(0.794912338f, -0.606724441f, 0.f),
    C_Vertex3f(0.162074924f, 0.530215025f, 0.f),
    C_Vertex3f(0.162074924f, 0.530215025f, -1.f),
    C_Normal3f(0.858511865f, -0.512793601f, 0.f),
    C_Vertex3f(0.156552553f, 0.522075176f, 0.f),
    C_Vertex3f(0.156552553f, 0.522075176f, -1.f),
    C_Normal3f(0.913557231f, -0.406710356f, 0.f),
    C_Vertex3f(0.152134657f, 0.51357317f, 0.f),
    C_Vertex3f(0.152134657f, 0.51357317f, -1.f),
    C_Normal3f(0.956429482f, -0.291963279f, 0.f),
    C_Vertex3f(0.148821235f, 0.504709005f, 0.f),
    C_Vertex3f(0.148821235f, 0.504709005f, -1.f),
    C_Normal3f(0.984844685f, -0.173438251f, 0.f),
    C_Vertex3f(0.146612287f, 0.495482683f, 0.f),
    C_Vertex3f(0.146612287f, 0.495482683f, -1.f),
    C_Normal3f(0.998868942f, -0.0475483239f, 0.f),
    C_Vertex3f(0.145507813f, 0.485894203f, 0.f),
    C_Vertex3f(0.145507813f, 0.485894203f, -1.f),
    C_Normal3f(0.994372308f, 0.105942346f, 0.f),
    C_Vertex3f(0.145507813f, 0.472280502f, 0.f),
    C_Vertex3f(0.145507813f, 0.472280502f, -1.f),
    C_Normal3f(0.954287887f, 0.298888922f, 0.f),
    C_Vertex3f(0.148635864f, 0.456534386f, 0.f),
    C_Vertex3f(0.148635864f, 0.456534386f, -1.f),
    C_Normal3f(0.86449939f, 0.50263375f, 0.f),
    C_Vertex3f(0.154891968f, 0.442318916f, 0.f),
    C_Vertex3f(0.154891968f, 0.442318916f, -1.f),
    C_Normal3f(0.747366309f, 0.664412141f, 0.f),
    C_Vertex3f(0.164276123f, 0.429634094f, 0.f),
    C_Vertex3f(0.164276123f, 0.429634094f, -1.f),
    C_Normal3f(0.614301443f, 0.789071441f, 0.f),
    C_Vertex3f(0.173605919f, 0.421268463f, 0.f),
    C_Vertex3f(0.173605919f, 0.421268463f, -1.f),
    C_Normal3f(0.495683849f, 0.868503153f, 0.f),
    C_Vertex3f(0.182229519f, 0.415657163f, 0.f),
    C_Vertex3f(0.182229519f, 0.415657163f, -1.f),
    C_Normal3f(0.417125732f, 0.908848763f, 0.f),
    C_Vertex3f(0.193329334f, 0.410011649f, 0.f),
    C_Vertex3f(0.193329334f, 0.410011649f, -1.f),
    C_Normal3f(0.358929008f, 0.933364928f, 0.f),
    C_Vertex3f(0.206905365f, 0.404331923f, 0.f),
    C_Vertex3f(0.206905365f, 0.404331923f, -1.f),
    C_Normal3f(0.314626217f, 0.949215591f, 0.f),
    C_Vertex3f(0.222957611f, 0.398617983f, 0.f),
    C_Vertex3f(0.222957611f, 0.398617983f, -1.f),
    C_Normal3f(0.280000448f, 0.959999919f, 0.f),
    C_Vertex3f(0.241486073f, 0.39286983f, 0.f),
    C_Vertex3f(0.241486073f, 0.39286983f, -1.f),
    C_Normal3f(0.252299994f, 0.967649102f, 0.f),
    C_Vertex3f(0.262490749f, 0.387087464f, 0.f),
    C_Vertex3f(0.262490749f, 0.387087464f, -1.f),
    C_Normal3f(0.229690298f, 0.973263741f, 0.f),
    C_Vertex3f(0.285971642f, 0.381270885f, 0.f),
    C_Vertex3f(0.285971642f, 0.381270885f, -1.f),
    C_Normal3f(0.225008681f, 0.974356711f, 0.f),
    C_Vertex3f(0.311928749f, 0.375420094f, 0.f),
    C_Vertex3f(0.311928749f, 0.375420094f, -1.f),
    C_Normal3f(0.236557841f, 0.971617401f, 0.f),
    C_Vertex3f(0.335841417f, 0.369754434f, 0.f),
    C_Vertex3f(0.335841417f, 0.369754434f, -1.f),
    C_Normal3f(0.250241429f, 0.968183458f, 0.f),
    C_Vertex3f(0.357709646f, 0.364273906f, 0.f),
    C_Vertex3f(0.357709646f, 0.364273906f, -1.f),
    C_Normal3f(0.266700119f, 0.963779569f, 0.f),
    C_Vertex3f(0.377533436f, 0.35897851f, 0.f),
    C_Vertex3f(0.377533436f, 0.35897851f, -1.f),
    C_Normal3f(0.286852658f, 0.957974672f, 0.f),
    C_Vertex3f(0.395312786f, 0.353868246f, 0.f),
    C_Vertex3f(0.395312786f, 0.353868246f, -1.f),
    C_Normal3f(0.31205979f, 0.950062454f, 0.f),
    C_Vertex3f(0.411047697f, 0.348943114f, 0.f),
    C_Vertex3f(0.411047697f, 0.348943114f, -1.f),
    C_Normal3f(0.344411969f, 0.938818634f, 0.f),
    C_Vertex3f(0.424738169f, 0.344203115f, 0.f),
    C_Vertex3f(0.424738169f, 0.344203115f, -1.f),
    C_Normal3f(0.391262203f, 0.920279264f, 0.f),
    C_Vertex3f(0.436384201f, 0.339648247f, 0.f),
    C_Vertex3f(0.436384201f, 0.339648247f, -1.f),
    C_Normal3f(0.44322136f, 0.896412134f, 0.f),
    C_Vertex3f(0.448167801f, 0.334241867f, 0.f),
    C_Vertex3f(0.448167801f, 0.334241867f, -1.f),
    C_Normal3f(0.490454525f, 0.871466875f, 0.f),
    C_Vertex3f(0.461219668f, 0.327368617f, 0.f),
    C_Vertex3f(0.461219668f, 0.327368617f, -1.f),
    C_Normal3f(0.541226506f, 0.840876758f, 0.f),
    C_Vertex3f(0.473357797f, 0.320065141f, 0.f),
    C_Vertex3f(0.473357797f, 0.320065141f, -1.f),
    C_Normal3f(0.593918145f, 0.804525495f, 0.f),
    C_Vertex3f(0.484582186f, 0.312331438f, 0.f),
    C_Vertex3f(0.484582186f, 0.312331438f, -1.f),
    C_Normal3f(0.647797883f, 0.761812329f, 0.f),
    C_Vertex3f(0.494892836f, 0.304167509f, 0.f),
    C_Vertex3f(0.494892836f, 0.304167509f, -1.f),
    C_Normal3f(0.701877952f, 0.712297261f, 0.f),
    C_Vertex3f(0.504289746f, 0.295573354f, 0.f),
    C_Vertex3f(0.504289746f, 0.295573354f, -1.f),
    C_Normal3f(0.754930794f, 0.655804515f, 0.f),
    C_Vertex3f(0.512772918f, 0.286548972f, 0.f),
    C_Vertex3f(0.512772918f, 0.286548972f, -1.f),
    C_Normal3f(0.810424566f, 0.585843086f, 0.f),
    C_Vertex3f(0.52034235f, 0.277094364f, 0.f),
    C_Vertex3f(0.52034235f, 0.277094364f, -1.f),
    C_Normal3f(0.875063002f, 0.484009027f, 0.f),
    C_Vertex3f(0.53032589f, 0.262267113f, 0.f),
    C_Vertex3f(0.53032589f, 0.262267113f, -1.f),
    C_Normal3f(0.935477078f, 0.353387594f, 0.f),
    C_Vertex3f(0.540309429f, 0.24099493f, 0.f),
    C_Vertex3f(0.540309429f, 0.24099493f, -1.f),
    C_Normal3f(0.978222191f, 0.207560554f, 0.f),
    C_Vertex3f(0.546965122f, 0.218220234f, 0.f),
    C_Vertex3f(0.546965122f, 0.218220234f, -1.f),
    C_Normal3f(0.997787654f, 0.0664810762f, 0.f),
    C_Vertex3f(0.550292969f, 0.193943024f, 0.f),
    C_Vertex3f(0.550292969f, 0.193943024f, -1.f),
    C_Normal3f(0.997380614f, -0.072332114f, 0.f),
    C_Vertex3f(0.550292969f, 0.168273926f, 0.f),
    C_Vertex3f(0.550292969f, 0.168273926f, -1.f),
    C_Normal3f(0.975688696f, -0.219161078f, 0.f),
    C_Vertex3f(0.546631336f, 0.143453121f, 0.f),
    C_Vertex3f(0.546631336f, 0.143453121f, -1.f),
    C_Normal3f(0.932907403f, -0.360116422f, 0.f),
    C_Vertex3f(0.539308071f, 0.119369984f, 0.f),
    C_Vertex3f(0.539308071f, 0.119369984f, -1.f),
    C_Normal3f(0.8780123f, -0.478638142f, 0.f),
    C_Vertex3f(0.528323174f, 0.0960245132f, 0.f),
    C_Vertex3f(0.528323174f, 0.0960245132f, -1.f),
    C_Normal3f(0.822501123f, -0.568763554f, 0.f),
    C_Vertex3f(0.517338276f, 0.0790686607f, 0.f),
    C_Vertex3f(0.517338276f, 0.0790686607f, -1.f),
    C_Normal3f(0.772335768f, -0.635214388f, 0.f),
    C_Vertex3f(0.509215593f, 0.0683926344f, 0.f),
    C_Vertex3f(0.509215593f, 0.0683926344f, -1.f),
    C_Normal3f(0.722240627f, -0.691641867f, 0.f),
    C_Vertex3f(0.500293493f, 0.0583444834f, 0.f),
    C_Vertex3f(0.500293493f, 0.0583444834f, -1.f),
    C_Normal3f(0.668852508f, -0.74339515f, 0.f),
    C_Vertex3f(0.490571976f, 0.0489242077f, 0.f),
    C_Vertex3f(0.490571976f, 0.0489242077f, -1.f),
    C_Normal3f(0.613247633f, -0.789890707f, 0.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, 0.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, -1.f),
    C_Normal3f(0.556525171f, -0.830830753f, 0.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, 0.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, -1.f),
    C_Normal3f(0.499719143f, -0.866187513f, 0.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, 0.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, -1.f),
    C_Normal3f(0.43820852f, -0.898873389f, 0.f),
    C_Vertex3f(0.44369173f, 0.0175218582f, 0.f),
    C_Vertex3f(0.44369173f, 0.0175218582f, -1.f),
    C_Normal3f(0.354275286f, -0.935141146f, 0.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, 0.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, -1.f),
    C_Normal3f(0.253019869f, -0.96746105f, 0.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, 0.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, -1.f),
    C_Normal3f(0.146797985f, -0.989166439f, 0.f),
    C_Vertex3f(0.363073826f, -0.00760173798f, 0.f),
    C_Vertex3f(0.363073826f, -0.00760173798f, -1.f),
    C_Normal3f(0.0517184138f, -0.998661697f, 0.f),
    C_Vertex3f(0.330471039f, -0.0107421875f, 0.f),
    C_Vertex3f(0.330471039f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0161968842f, -0.99986881f, 0.f),
    C_Vertex3f(0.302433014f, -0.0107421875f, 0.f),
    C_Vertex3f(0.302433014f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0580721684f, -0.998312414f, 0.f),
    C_Vertex3f(0.281562805f, -0.00994992256f, 0.f),
    C_Vertex3f(0.281562805f, -0.00994992256f, -1.f),
    C_Normal3f(-0.100804754f, -0.994906187f, 0.f),
    C_Vertex3f(0.261573792f, -0.00836539268f, 0.f),
    C_Vertex3f(0.261573792f, -0.00836539268f, -1.f),
    C_Normal3f(-0.146933034f, -0.989146531f, 0.f),
    C_Vertex3f(0.242465973f, -0.00598859787f, 0.f),
    C_Vertex3f(0.242465973f, -0.00598859787f, -1.f),
    C_Normal3f(-0.196539491f, -0.98049593f, 0.f),
    C_Vertex3f(0.224239349f, -0.00281953812f, 0.f),
    C_Vertex3f(0.224239349f, -0.00281953812f, -1.f),
    C_Normal3f(-0.249605194f, -0.968347728f, 0.f),
    C_Vertex3f(0.206893921f, 0.00114178658f, 0.f),
    C_Vertex3f(0.206893921f, 0.00114178658f, -1.f),
    C_Normal3f(-0.305969775f, -0.952041268f, 0.f),
    C_Vertex3f(0.190429688f, 0.00589537621f, 0.f),
    C_Vertex3f(0.190429688f, 0.00589537621f, -1.f),
    C_Normal3f(-0.365288645f, -0.930894315f, 0.f),
    C_Vertex3f(0.174846649f, 0.0114412308f, 0.f),
    C_Vertex3f(0.174846649f, 0.0114412308f, -1.f),
    C_Normal3f(-0.42608273f, -0.904684186f, 0.f),
    C_Vertex3f(0.160144806f, 0.0177793503f, 0.f),
    C_Vertex3f(0.160144806f, 0.0177793503f, -1.f),
    C_Normal3f(-0.487082392f, -0.873356044f, 0.f),
    C_Vertex3f(0.146235347f, 0.0249164104f, 0.f),
    C_Vertex3f(0.146235347f, 0.0249164104f, -1.f),
    C_Normal3f(-0.548076808f, -0.836427927f, 0.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, 0.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, -1.f),
    C_Normal3f(-0.607979953f, -0.793952405f, 0.f),
    C_Vertex3f(0.120793581f, 0.0415873528f, 0.f),
    C_Vertex3f(0.120793581f, 0.0415873528f, -1.f),
    C_Normal3f(-0.665657461f, -0.746257424f, 0.f),
    C_Vertex3f(0.109261274f, 0.0511212349f, 0.f),
    C_Vertex3f(0.109261274f, 0.0511212349f, -1.f),
    C_Normal3f(-0.720020831f, -0.693952501f, 0.f),
    C_Vertex3f(0.0985213518f, 0.0614540577f, 0.f),
    C_Vertex3f(0.0985213518f, 0.0614540577f, -1.f),
    C_Normal3f(-0.770123422f, -0.637894928f, 0.f),
    C_Vertex3f(0.0885738134f, 0.0725858212f, 0.f),
    C_Vertex3f(0.0885738134f, 0.0725858212f, -1.f),
    C_Normal3f(-0.815241337f, -0.579121411f, 0.f),
    C_Vertex3f(0.0794186592f, 0.0845165253f, 0.f),
    C_Vertex3f(0.0794186592f, 0.0845165253f, -1.f),
    C_Normal3f(-0.854430854f, -0.519565165f, 0.f),
    C_Vertex3f(0.0710558891f, 0.09724617f, 0.f),
    C_Vertex3f(0.0710558891f, 0.09724617f, -1.f),
    C_Normal3f(-0.889058173f, -0.45779416f, 0.f),
    C_Vertex3f(0.0636838675f, 0.11039257f, 0.f),
    C_Vertex3f(0.0636838675f, 0.11039257f, -1.f),
    C_Normal3f(-0.919535875f, -0.393005997f, 0.f),
    C_Vertex3f(0.0573025942f, 0.123955727f, 0.f),
    C_Vertex3f(0.0573025942f, 0.123955727f, -1.f),
    C_Normal3f(-0.945318818f, -0.326147705f, 0.f),
    C_Vertex3f(0.0519120693f, 0.137935638f, 0.f),
    C_Vertex3f(0.0519120693f, 0.137935638f, -1.f),
    C_Normal3f(-0.966074646f, -0.258263081f, 0.f),
    C_Vertex3f(0.0475122929f, 0.152332306f, 0.f),
    C_Vertex3f(0.0475122929f, 0.152332306f, -1.f),
    C_Normal3f(-0.981704116f, -0.190412983f, 0.f),
    C_Vertex3f(0.0441032648f, 0.167145729f, 0.f),
    C_Vertex3f(0.0441032648f, 0.167145729f, -1.f),
    C_Normal3f(-0.992332458f, -0.123597145f, 0.f),
    C_Vertex3f(0.0416849852f, 0.182375908f, 0.f),
    C_Vertex3f(0.0416849852f, 0.182375908f, -1.f),
    C_Normal3f(-0.997592866f, -0.0693431348f, 0.f),
    C_Vertex3f(0.0402574539f, 0.198022842f, 0.f),
    C_Vertex3f(0.0402574539f, 0.198022842f, -1.f),
    C_Normal3f(-0.999630511f, -0.0271806438f, 0.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, 0.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.53125f, 0.460449219f, 0.f),
    C_Vertex3f(0.530702591f, 0.474542618f, 0.f),
    C_Vertex3f(0.526213646f, 0.501254559f, 0.f),
    C_Vertex3f(0.518330574f, 0.526491642f, 0.f),
    C_Vertex3f(0.507053375f, 0.550253868f, 0.f),
    C_Vertex3f(0.496049881f, 0.566969395f, 0.f),
    C_Vertex3f(0.487927318f, 0.577437043f, 0.f),
    C_Vertex3f(0.479017854f, 0.587228656f, 0.f),
    C_Vertex3f(0.469321489f, 0.596344233f, 0.f),
    C_Vertex3f(0.458838224f, 0.604783773f, 0.f),
    C_Vertex3f(0.450195313f, 0.454101563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.443137288f, 0.489351988f, 0.f),
    C_Vertex3f(0.446745276f, 0.475996256f, 0.f),
    C_Vertex3f(0.447568059f, 0.612547278f, 0.f),
    C_Vertex3f(0.449372292f, 0.461709976f, 0.f),
    C_Vertex3f(0.458838224f, 0.604783773f, 0.f),
    C_Vertex3f(0.450195313f, 0.454101563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426427484f, 0.523835897f, 0.f),
    C_Vertex3f(0.432978392f, 0.513271809f, 0.f),
    C_Vertex3f(0.435510993f, 0.619634748f, 0.f),
    C_Vertex3f(0.438548326f, 0.501777172f, 0.f),
    C_Vertex3f(0.447568059f, 0.612547278f, 0.f),
    C_Vertex3f(0.443137288f, 0.489351988f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.422667027f, 0.626046181f, 0.f),
    C_Vertex3f(0.402220726f, 0.634649277f, 0.f),
    C_Vertex3f(0.410382748f, 0.542172432f, 0.f),
    C_Vertex3f(0.418895602f, 0.533469439f, 0.f),
    C_Vertex3f(0.426427484f, 0.523835897f, 0.f),
    C_Vertex3f(0.435510993f, 0.619634748f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402220726f, 0.634649277f, 0.f),
    C_Vertex3f(0.372850895f, 0.643252373f, 0.f),
    C_Vertex3f(0.378505588f, 0.562575102f, 0.f),
    C_Vertex3f(0.3902632f, 0.556745768f, 0.f),
    C_Vertex3f(0.40088892f, 0.549944878f, 0.f),
    C_Vertex3f(0.410382748f, 0.542172432f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320156217f, 0.576176882f, 0.f),
    C_Vertex3f(0.336441398f, 0.57423377f, 0.f),
    C_Vertex3f(0.341372967f, 0.64898777f, 0.f),
    C_Vertex3f(0.351594687f, 0.571319103f, 0.f),
    C_Vertex3f(0.372850895f, 0.643252373f, 0.f),
    C_Vertex3f(0.365616083f, 0.56743288f, 0.f),
    C_Vertex3f(0.378505588f, 0.562575102f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.307786942f, 0.651855469f, 0.f),
    C_Vertex3f(0.273736954f, 0.651855469f, 0.f),
    C_Vertex3f(0.283805847f, 0.577148438f, 0.f),
    C_Vertex3f(0.302739143f, 0.577148438f, 0.f),
    C_Vertex3f(0.320156217f, 0.576176882f, 0.f),
    C_Vertex3f(0.341372967f, 0.64898777f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.219637632f, 0.568331718f, 0.f),
    C_Vertex3f(0.233714342f, 0.571858406f, 0.f),
    C_Vertex3f(0.242808819f, 0.649122238f, 0.f),
    C_Vertex3f(0.249101281f, 0.574503422f, 0.f),
    C_Vertex3f(0.273736954f, 0.651855469f, 0.f),
    C_Vertex3f(0.26579845f, 0.576266766f, 0.f),
    C_Vertex3f(0.283805847f, 0.577148438f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213358402f, 0.643655777f, 0.f),
    C_Vertex3f(0.185385704f, 0.635456085f, 0.f),
    C_Vertex3f(0.195414901f, 0.558633327f, 0.f),
    C_Vertex3f(0.206871152f, 0.563923359f, 0.f),
    C_Vertex3f(0.219637632f, 0.568331718f, 0.f),
    C_Vertex3f(0.242808819f, 0.649122238f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.185385704f, 0.635456085f, 0.f),
    C_Vertex3f(0.165514469f, 0.627256393f, 0.f),
    C_Vertex3f(0.168701768f, 0.537992716f, 0.f),
    C_Vertex3f(0.176433086f, 0.545408249f, 0.f),
    C_Vertex3f(0.185268879f, 0.552461624f, 0.f),
    C_Vertex3f(0.195414901f, 0.558633327f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.148821235f, 0.504709005f, 0.f),
    C_Vertex3f(0.152134657f, 0.51357317f, 0.f),
    C_Vertex3f(0.153060675f, 0.621154547f, 0.f),
    C_Vertex3f(0.156552553f, 0.522075176f, 0.f),
    C_Vertex3f(0.165514469f, 0.627256393f, 0.f),
    C_Vertex3f(0.162074924f, 0.530215025f, 0.f),
    C_Vertex3f(0.168701768f, 0.537992716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.120461464f, 0.599036694f, 0.f),
    C_Vertex3f(0.123383999f, 0.355178356f, 0.f),
    C_Vertex3f(0.130534172f, 0.607044697f, 0.f),
    C_Vertex3f(0.143409729f, 0.341705322f, 0.f),
    C_Vertex3f(0.141400576f, 0.614417315f, 0.f),
    C_Vertex3f(0.145507813f, 0.485894203f, 0.f),
    C_Vertex3f(0.153060675f, 0.621154547f, 0.f),
    C_Vertex3f(0.146612287f, 0.495482683f, 0.f),
    C_Vertex3f(0.148821235f, 0.504709005f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154891968f, 0.442318916f, 0.f),
    C_Vertex3f(0.148635864f, 0.456534386f, 0.f),
    C_Vertex3f(0.159310341f, 0.333342552f, 0.f),
    C_Vertex3f(0.145507813f, 0.472280502f, 0.f),
    C_Vertex3f(0.143409729f, 0.341705322f, 0.f),
    C_Vertex3f(0.145507813f, 0.485894203f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.111182451f, 0.590393305f, 0.f),
    C_Vertex3f(0.102697134f, 0.581114531f, 0.f),
    C_Vertex3f(0.106281757f, 0.370100498f, 0.f),
    C_Vertex3f(0.123383999f, 0.355178356f, 0.f),
    C_Vertex3f(0.120461464f, 0.599036694f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0639648438f, 0.487323761f, 0.f),
    C_Vertex3f(0.0667786598f, 0.443130016f, 0.f),
    C_Vertex3f(0.0674138069f, 0.510800362f, 0.f),
    C_Vertex3f(0.072406292f, 0.423178196f, 0.f),
    C_Vertex3f(0.0743117332f, 0.533484459f, 0.f),
    C_Vertex3f(0.0808477402f, 0.404292107f, 0.f),
    C_Vertex3f(0.0846586227f, 0.555376053f, 0.f),
    C_Vertex3f(0.0921030045f, 0.386471748f, 0.f),
    C_Vertex3f(0.0950055122f, 0.571200371f, 0.f),
    C_Vertex3f(0.106281757f, 0.370100498f, 0.f),
    C_Vertex3f(0.102697134f, 0.581114531f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466575146f, 0.195565224f, 0.f),
    C_Vertex3f(0.46875f, 0.181438446f, 0.f),
    C_Vertex3f(0.473357797f, 0.320065141f, 0.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, 0.f),
    C_Vertex3f(0.484582186f, 0.312331438f, 0.f),
    C_Vertex3f(0.490571976f, 0.0489242077f, 0.f),
    C_Vertex3f(0.494892836f, 0.304167509f, 0.f),
    C_Vertex3f(0.500293493f, 0.0583444834f, 0.f),
    C_Vertex3f(0.504289746f, 0.295573354f, 0.f),
    C_Vertex3f(0.509215593f, 0.0683926344f, 0.f),
    C_Vertex3f(0.512772918f, 0.286548972f, 0.f),
    C_Vertex3f(0.517338276f, 0.0790686607f, 0.f),
    C_Vertex3f(0.52034235f, 0.277094364f, 0.f),
    C_Vertex3f(0.528323174f, 0.0960245132f, 0.f),
    C_Vertex3f(0.53032589f, 0.262267113f, 0.f),
    C_Vertex3f(0.539308071f, 0.119369984f, 0.f),
    C_Vertex3f(0.540309429f, 0.24099493f, 0.f),
    C_Vertex3f(0.546631336f, 0.143453121f, 0.f),
    C_Vertex3f(0.546965122f, 0.218220234f, 0.f),
    C_Vertex3f(0.550292969f, 0.168273926f, 0.f),
    C_Vertex3f(0.550292969f, 0.193943024f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.381887197f, 0.268518806f, 0.f),
    C_Vertex3f(0.390324593f, 0.265654564f, 0.f),
    C_Vertex3f(0.395312786f, 0.353868246f, 0.f),
    C_Vertex3f(0.403659821f, 0.260570526f, 0.f),
    C_Vertex3f(0.411047697f, 0.348943114f, 0.f),
    C_Vertex3f(0.42098093f, 0.252017021f, 0.f),
    C_Vertex3f(0.424738169f, 0.344203115f, 0.f),
    C_Vertex3f(0.435428143f, 0.242562294f, 0.f),
    C_Vertex3f(0.436384201f, 0.339648247f, 0.f),
    C_Vertex3f(0.447001457f, 0.232206345f, 0.f),
    C_Vertex3f(0.448167801f, 0.334241867f, 0.f),
    C_Vertex3f(0.455700874f, 0.220949173f, 0.f),
    C_Vertex3f(0.461219668f, 0.327368617f, 0.f),
    C_Vertex3f(0.462225437f, 0.208735466f, 0.f),
    C_Vertex3f(0.473357797f, 0.320065141f, 0.f),
    C_Vertex3f(0.466575146f, 0.195565224f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377533436f, 0.35897851f, 0.f),
    C_Vertex3f(0.357709646f, 0.364273906f, 0.f),
    C_Vertex3f(0.359126568f, 0.275292635f, 0.f),
    C_Vertex3f(0.371487856f, 0.271731496f, 0.f),
    C_Vertex3f(0.381887197f, 0.268518806f, 0.f),
    C_Vertex3f(0.395312786f, 0.353868246f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.290061951f, 0.293021679f, 0.f),
    C_Vertex3f(0.310271025f, 0.288066745f, 0.f),
    C_Vertex3f(0.311928749f, 0.375420094f, 0.f),
    C_Vertex3f(0.328518152f, 0.283460259f, 0.f),
    C_Vertex3f(0.335841417f, 0.369754434f, 0.f),
    C_Vertex3f(0.344803333f, 0.279202223f, 0.f),
    C_Vertex3f(0.357709646f, 0.364273906f, 0.f),
    C_Vertex3f(0.359126568f, 0.275292635f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193329334f, 0.410011649f, 0.f),
    C_Vertex3f(0.195855618f, 0.318786383f, 0.f),
    C_Vertex3f(0.206905365f, 0.404331923f, 0.f),
    C_Vertex3f(0.21136713f, 0.313783884f, 0.f),
    C_Vertex3f(0.222957611f, 0.398617983f, 0.f),
    C_Vertex3f(0.22854352f, 0.308706164f, 0.f),
    C_Vertex3f(0.241486073f, 0.39286983f, 0.f),
    C_Vertex3f(0.247384787f, 0.303553224f, 0.f),
    C_Vertex3f(0.262490749f, 0.387087464f, 0.f),
    C_Vertex3f(0.26789093f, 0.298325062f, 0.f),
    C_Vertex3f(0.285971642f, 0.381270885f, 0.f),
    C_Vertex3f(0.290061951f, 0.293021679f, 0.f),
    C_Vertex3f(0.311928749f, 0.375420094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154891968f, 0.442318916f, 0.f),
    C_Vertex3f(0.159310341f, 0.333342552f, 0.f),
    C_Vertex3f(0.164276123f, 0.429634094f, 0.f),
    C_Vertex3f(0.169827223f, 0.328565717f, 0.f),
    C_Vertex3f(0.173605919f, 0.421268463f, 0.f),
    C_Vertex3f(0.182008982f, 0.32371366f, 0.f),
    C_Vertex3f(0.182229519f, 0.415657163f, 0.f),
    C_Vertex3f(0.195855618f, 0.318786383f, 0.f),
    C_Vertex3f(0.193329334f, 0.410011649f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.46875f, 0.166463852f, 0.f),
    C_Vertex3f(0.466480732f, 0.152168751f, 0.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, 0.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, 0.f),
    C_Vertex3f(0.46875f, 0.181438446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.461942196f, 0.138444424f, 0.f),
    C_Vertex3f(0.455134392f, 0.125290871f, 0.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, 0.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, 0.f),
    C_Vertex3f(0.466480732f, 0.152168751f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.420767784f, 0.0915622711f, 0.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, 0.f),
    C_Vertex3f(0.434601784f, 0.101465225f, 0.f),
    C_Vertex3f(0.44369173f, 0.0175218582f, 0.f),
    C_Vertex3f(0.44605732f, 0.112708092f, 0.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, 0.f),
    C_Vertex3f(0.455134392f, 0.125290871f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.404555321f, 0.0829992294f, 0.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, 0.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, 0.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, 0.f),
    C_Vertex3f(0.420767784f, 0.0915622711f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.224239349f, -0.00281953812f, 0.f),
    C_Vertex3f(0.242465973f, -0.00598859787f, 0.f),
    C_Vertex3f(0.245275974f, 0.072081089f, 0.f),
    C_Vertex3f(0.261573792f, -0.00836539268f, 0.f),
    C_Vertex3f(0.269881725f, 0.0673213005f, 0.f),
    C_Vertex3f(0.281562805f, -0.00994992256f, 0.f),
    C_Vertex3f(0.295854568f, 0.0649414063f, 0.f),
    C_Vertex3f(0.302433014f, -0.0107421875f, 0.f),
    C_Vertex3f(0.321662903f, 0.0649414063f, 0.f),
    C_Vertex3f(0.330471039f, -0.0107421875f, 0.f),
    C_Vertex3f(0.344518185f, 0.0667471886f, 0.f),
    C_Vertex3f(0.363073826f, -0.00760173798f, 0.f),
    C_Vertex3f(0.365952015f, 0.0703587532f, 0.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, 0.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132882595f, 0.159187317f, 0.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, 0.f),
    C_Vertex3f(0.141385078f, 0.14159584f, 0.f),
    C_Vertex3f(0.146235347f, 0.0249164104f, 0.f),
    C_Vertex3f(0.151774406f, 0.12616539f, 0.f),
    C_Vertex3f(0.160144806f, 0.0177793503f, 0.f),
    C_Vertex3f(0.165034294f, 0.112212658f, 0.f),
    C_Vertex3f(0.174846649f, 0.0114412308f, 0.f),
    C_Vertex3f(0.181164742f, 0.0997376442f, 0.f),
    C_Vertex3f(0.190429688f, 0.00589537621f, 0.f),
    C_Vertex3f(0.200165749f, 0.0887403488f, 0.f),
    C_Vertex3f(0.206893921f, 0.00114178658f, 0.f),
    C_Vertex3f(0.222037315f, 0.0792207718f, 0.f),
    C_Vertex3f(0.224239349f, -0.00281953812f, 0.f),
    C_Vertex3f(0.245275974f, 0.072081089f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0985213518f, 0.0614540577f, 0.f),
    C_Vertex3f(0.109261274f, 0.0511212349f, 0.f),
    C_Vertex3f(0.120117188f, 0.212890625f, 0.f),
    C_Vertex3f(0.120793581f, 0.0415873528f, 0.f),
    C_Vertex3f(0.121538162f, 0.200853348f, 0.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, 0.f),
    C_Vertex3f(0.126266956f, 0.178939819f, 0.f),
    C_Vertex3f(0.132882595f, 0.159187317f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120117188f, 0.212890625f, 0.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, 0.f),
    C_Vertex3f(0.0402574539f, 0.198022842f, 0.f),
    C_Vertex3f(0.0416849852f, 0.182375908f, 0.f),
    C_Vertex3f(0.0441032648f, 0.167145729f, 0.f),
    C_Vertex3f(0.0475122929f, 0.152332306f, 0.f),
    C_Vertex3f(0.0519120693f, 0.137935638f, 0.f),
    C_Vertex3f(0.0573025942f, 0.123955727f, 0.f),
    C_Vertex3f(0.0636838675f, 0.11039257f, 0.f),
    C_Vertex3f(0.0710558891f, 0.09724617f, 0.f),
    C_Vertex3f(0.0794186592f, 0.0845165253f, 0.f),
    C_Vertex3f(0.0885738134f, 0.0725858212f, 0.f),
    C_Vertex3f(0.0985213518f, 0.0614540577f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0667786598f, 0.443130016f, 0.f),
    C_Vertex3f(0.0639648438f, 0.487323761f, 0.f),
    C_Vertex3f(0.0639648438f, 0.464147568f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.53125f, 0.460449219f, -1.f),
    C_Vertex3f(0.530702591f, 0.474542618f, -1.f),
    C_Vertex3f(0.526213646f, 0.501254559f, -1.f),
    C_Vertex3f(0.518330574f, 0.526491642f, -1.f),
    C_Vertex3f(0.507053375f, 0.550253868f, -1.f),
    C_Vertex3f(0.496049881f, 0.566969395f, -1.f),
    C_Vertex3f(0.487927318f, 0.577437043f, -1.f),
    C_Vertex3f(0.479017854f, 0.587228656f, -1.f),
    C_Vertex3f(0.469321489f, 0.596344233f, -1.f),
    C_Vertex3f(0.458838224f, 0.604783773f, -1.f),
    C_Vertex3f(0.450195313f, 0.454101563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.443137288f, 0.489351988f, -1.f),
    C_Vertex3f(0.446745276f, 0.475996256f, -1.f),
    C_Vertex3f(0.447568059f, 0.612547278f, -1.f),
    C_Vertex3f(0.449372292f, 0.461709976f, -1.f),
    C_Vertex3f(0.458838224f, 0.604783773f, -1.f),
    C_Vertex3f(0.450195313f, 0.454101563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426427484f, 0.523835897f, -1.f),
    C_Vertex3f(0.432978392f, 0.513271809f, -1.f),
    C_Vertex3f(0.435510993f, 0.619634748f, -1.f),
    C_Vertex3f(0.438548326f, 0.501777172f, -1.f),
    C_Vertex3f(0.447568059f, 0.612547278f, -1.f),
    C_Vertex3f(0.443137288f, 0.489351988f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.422667027f, 0.626046181f, -1.f),
    C_Vertex3f(0.402220726f, 0.634649277f, -1.f),
    C_Vertex3f(0.410382748f, 0.542172432f, -1.f),
    C_Vertex3f(0.418895602f, 0.533469439f, -1.f),
    C_Vertex3f(0.426427484f, 0.523835897f, -1.f),
    C_Vertex3f(0.435510993f, 0.619634748f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402220726f, 0.634649277f, -1.f),
    C_Vertex3f(0.372850895f, 0.643252373f, -1.f),
    C_Vertex3f(0.378505588f, 0.562575102f, -1.f),
    C_Vertex3f(0.3902632f, 0.556745768f, -1.f),
    C_Vertex3f(0.40088892f, 0.549944878f, -1.f),
    C_Vertex3f(0.410382748f, 0.542172432f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320156217f, 0.576176882f, -1.f),
    C_Vertex3f(0.336441398f, 0.57423377f, -1.f),
    C_Vertex3f(0.341372967f, 0.64898777f, -1.f),
    C_Vertex3f(0.351594687f, 0.571319103f, -1.f),
    C_Vertex3f(0.372850895f, 0.643252373f, -1.f),
    C_Vertex3f(0.365616083f, 0.56743288f, -1.f),
    C_Vertex3f(0.378505588f, 0.562575102f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.307786942f, 0.651855469f, -1.f),
    C_Vertex3f(0.273736954f, 0.651855469f, -1.f),
    C_Vertex3f(0.283805847f, 0.577148438f, -1.f),
    C_Vertex3f(0.302739143f, 0.577148438f, -1.f),
    C_Vertex3f(0.320156217f, 0.576176882f, -1.f),
    C_Vertex3f(0.341372967f, 0.64898777f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.219637632f, 0.568331718f, -1.f),
    C_Vertex3f(0.233714342f, 0.571858406f, -1.f),
    C_Vertex3f(0.242808819f, 0.649122238f, -1.f),
    C_Vertex3f(0.249101281f, 0.574503422f, -1.f),
    C_Vertex3f(0.273736954f, 0.651855469f, -1.f),
    C_Vertex3f(0.26579845f, 0.576266766f, -1.f),
    C_Vertex3f(0.283805847f, 0.577148438f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213358402f, 0.643655777f, -1.f),
    C_Vertex3f(0.185385704f, 0.635456085f, -1.f),
    C_Vertex3f(0.195414901f, 0.558633327f, -1.f),
    C_Vertex3f(0.206871152f, 0.563923359f, -1.f),
    C_Vertex3f(0.219637632f, 0.568331718f, -1.f),
    C_Vertex3f(0.242808819f, 0.649122238f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.185385704f, 0.635456085f, -1.f),
    C_Vertex3f(0.165514469f, 0.627256393f, -1.f),
    C_Vertex3f(0.168701768f, 0.537992716f, -1.f),
    C_Vertex3f(0.176433086f, 0.545408249f, -1.f),
    C_Vertex3f(0.185268879f, 0.552461624f, -1.f),
    C_Vertex3f(0.195414901f, 0.558633327f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.148821235f, 0.504709005f, -1.f),
    C_Vertex3f(0.152134657f, 0.51357317f, -1.f),
    C_Vertex3f(0.153060675f, 0.621154547f, -1.f),
    C_Vertex3f(0.156552553f, 0.522075176f, -1.f),
    C_Vertex3f(0.165514469f, 0.627256393f, -1.f),
    C_Vertex3f(0.162074924f, 0.530215025f, -1.f),
    C_Vertex3f(0.168701768f, 0.537992716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.120461464f, 0.599036694f, -1.f),
    C_Vertex3f(0.123383999f, 0.355178356f, -1.f),
    C_Vertex3f(0.130534172f, 0.607044697f, -1.f),
    C_Vertex3f(0.143409729f, 0.341705322f, -1.f),
    C_Vertex3f(0.141400576f, 0.614417315f, -1.f),
    C_Vertex3f(0.145507813f, 0.485894203f, -1.f),
    C_Vertex3f(0.153060675f, 0.621154547f, -1.f),
    C_Vertex3f(0.146612287f, 0.495482683f, -1.f),
    C_Vertex3f(0.148821235f, 0.504709005f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154891968f, 0.442318916f, -1.f),
    C_Vertex3f(0.148635864f, 0.456534386f, -1.f),
    C_Vertex3f(0.159310341f, 0.333342552f, -1.f),
    C_Vertex3f(0.145507813f, 0.472280502f, -1.f),
    C_Vertex3f(0.143409729f, 0.341705322f, -1.f),
    C_Vertex3f(0.145507813f, 0.485894203f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.111182451f, 0.590393305f, -1.f),
    C_Vertex3f(0.102697134f, 0.581114531f, -1.f),
    C_Vertex3f(0.106281757f, 0.370100498f, -1.f),
    C_Vertex3f(0.123383999f, 0.355178356f, -1.f),
    C_Vertex3f(0.120461464f, 0.599036694f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0639648438f, 0.487323761f, -1.f),
    C_Vertex3f(0.0667786598f, 0.443130016f, -1.f),
    C_Vertex3f(0.0674138069f, 0.510800362f, -1.f),
    C_Vertex3f(0.072406292f, 0.423178196f, -1.f),
    C_Vertex3f(0.0743117332f, 0.533484459f, -1.f),
    C_Vertex3f(0.0808477402f, 0.404292107f, -1.f),
    C_Vertex3f(0.0846586227f, 0.555376053f, -1.f),
    C_Vertex3f(0.0921030045f, 0.386471748f, -1.f),
    C_Vertex3f(0.0950055122f, 0.571200371f, -1.f),
    C_Vertex3f(0.106281757f, 0.370100498f, -1.f),
    C_Vertex3f(0.102697134f, 0.581114531f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466575146f, 0.195565224f, -1.f),
    C_Vertex3f(0.46875f, 0.181438446f, -1.f),
    C_Vertex3f(0.473357797f, 0.320065141f, -1.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, -1.f),
    C_Vertex3f(0.484582186f, 0.312331438f, -1.f),
    C_Vertex3f(0.490571976f, 0.0489242077f, -1.f),
    C_Vertex3f(0.494892836f, 0.304167509f, -1.f),
    C_Vertex3f(0.500293493f, 0.0583444834f, -1.f),
    C_Vertex3f(0.504289746f, 0.295573354f, -1.f),
    C_Vertex3f(0.509215593f, 0.0683926344f, -1.f),
    C_Vertex3f(0.512772918f, 0.286548972f, -1.f),
    C_Vertex3f(0.517338276f, 0.0790686607f, -1.f),
    C_Vertex3f(0.52034235f, 0.277094364f, -1.f),
    C_Vertex3f(0.528323174f, 0.0960245132f, -1.f),
    C_Vertex3f(0.53032589f, 0.262267113f, -1.f),
    C_Vertex3f(0.539308071f, 0.119369984f, -1.f),
    C_Vertex3f(0.540309429f, 0.24099493f, -1.f),
    C_Vertex3f(0.546631336f, 0.143453121f, -1.f),
    C_Vertex3f(0.546965122f, 0.218220234f, -1.f),
    C_Vertex3f(0.550292969f, 0.168273926f, -1.f),
    C_Vertex3f(0.550292969f, 0.193943024f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.381887197f, 0.268518806f, -1.f),
    C_Vertex3f(0.390324593f, 0.265654564f, -1.f),
    C_Vertex3f(0.395312786f, 0.353868246f, -1.f),
    C_Vertex3f(0.403659821f, 0.260570526f, -1.f),
    C_Vertex3f(0.411047697f, 0.348943114f, -1.f),
    C_Vertex3f(0.42098093f, 0.252017021f, -1.f),
    C_Vertex3f(0.424738169f, 0.344203115f, -1.f),
    C_Vertex3f(0.435428143f, 0.242562294f, -1.f),
    C_Vertex3f(0.436384201f, 0.339648247f, -1.f),
    C_Vertex3f(0.447001457f, 0.232206345f, -1.f),
    C_Vertex3f(0.448167801f, 0.334241867f, -1.f),
    C_Vertex3f(0.455700874f, 0.220949173f, -1.f),
    C_Vertex3f(0.461219668f, 0.327368617f, -1.f),
    C_Vertex3f(0.462225437f, 0.208735466f, -1.f),
    C_Vertex3f(0.473357797f, 0.320065141f, -1.f),
    C_Vertex3f(0.466575146f, 0.195565224f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377533436f, 0.35897851f, -1.f),
    C_Vertex3f(0.357709646f, 0.364273906f, -1.f),
    C_Vertex3f(0.359126568f, 0.275292635f, -1.f),
    C_Vertex3f(0.371487856f, 0.271731496f, -1.f),
    C_Vertex3f(0.381887197f, 0.268518806f, -1.f),
    C_Vertex3f(0.395312786f, 0.353868246f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.290061951f, 0.293021679f, -1.f),
    C_Vertex3f(0.310271025f, 0.288066745f, -1.f),
    C_Vertex3f(0.311928749f, 0.375420094f, -1.f),
    C_Vertex3f(0.328518152f, 0.283460259f, -1.f),
    C_Vertex3f(0.335841417f, 0.369754434f, -1.f),
    C_Vertex3f(0.344803333f, 0.279202223f, -1.f),
    C_Vertex3f(0.357709646f, 0.364273906f, -1.f),
    C_Vertex3f(0.359126568f, 0.275292635f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193329334f, 0.410011649f, -1.f),
    C_Vertex3f(0.195855618f, 0.318786383f, -1.f),
    C_Vertex3f(0.206905365f, 0.404331923f, -1.f),
    C_Vertex3f(0.21136713f, 0.313783884f, -1.f),
    C_Vertex3f(0.222957611f, 0.398617983f, -1.f),
    C_Vertex3f(0.22854352f, 0.308706164f, -1.f),
    C_Vertex3f(0.241486073f, 0.39286983f, -1.f),
    C_Vertex3f(0.247384787f, 0.303553224f, -1.f),
    C_Vertex3f(0.262490749f, 0.387087464f, -1.f),
    C_Vertex3f(0.26789093f, 0.298325062f, -1.f),
    C_Vertex3f(0.285971642f, 0.381270885f, -1.f),
    C_Vertex3f(0.290061951f, 0.293021679f, -1.f),
    C_Vertex3f(0.311928749f, 0.375420094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154891968f, 0.442318916f, -1.f),
    C_Vertex3f(0.159310341f, 0.333342552f, -1.f),
    C_Vertex3f(0.164276123f, 0.429634094f, -1.f),
    C_Vertex3f(0.169827223f, 0.328565717f, -1.f),
    C_Vertex3f(0.173605919f, 0.421268463f, -1.f),
    C_Vertex3f(0.182008982f, 0.32371366f, -1.f),
    C_Vertex3f(0.182229519f, 0.415657163f, -1.f),
    C_Vertex3f(0.195855618f, 0.318786383f, -1.f),
    C_Vertex3f(0.193329334f, 0.410011649f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.46875f, 0.166463852f, -1.f),
    C_Vertex3f(0.466480732f, 0.152168751f, -1.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, -1.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, -1.f),
    C_Vertex3f(0.46875f, 0.181438446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.461942196f, 0.138444424f, -1.f),
    C_Vertex3f(0.455134392f, 0.125290871f, -1.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, -1.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, -1.f),
    C_Vertex3f(0.466480732f, 0.152168751f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.420767784f, 0.0915622711f, -1.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, -1.f),
    C_Vertex3f(0.434601784f, 0.101465225f, -1.f),
    C_Vertex3f(0.44369173f, 0.0175218582f, -1.f),
    C_Vertex3f(0.44605732f, 0.112708092f, -1.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, -1.f),
    C_Vertex3f(0.455134392f, 0.125290871f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.404555321f, 0.0829992294f, -1.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, -1.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, -1.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, -1.f),
    C_Vertex3f(0.420767784f, 0.0915622711f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.224239349f, -0.00281953812f, -1.f),
    C_Vertex3f(0.242465973f, -0.00598859787f, -1.f),
    C_Vertex3f(0.245275974f, 0.072081089f, -1.f),
    C_Vertex3f(0.261573792f, -0.00836539268f, -1.f),
    C_Vertex3f(0.269881725f, 0.0673213005f, -1.f),
    C_Vertex3f(0.281562805f, -0.00994992256f, -1.f),
    C_Vertex3f(0.295854568f, 0.0649414063f, -1.f),
    C_Vertex3f(0.302433014f, -0.0107421875f, -1.f),
    C_Vertex3f(0.321662903f, 0.0649414063f, -1.f),
    C_Vertex3f(0.330471039f, -0.0107421875f, -1.f),
    C_Vertex3f(0.344518185f, 0.0667471886f, -1.f),
    C_Vertex3f(0.363073826f, -0.00760173798f, -1.f),
    C_Vertex3f(0.365952015f, 0.0703587532f, -1.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, -1.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132882595f, 0.159187317f, -1.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, -1.f),
    C_Vertex3f(0.141385078f, 0.14159584f, -1.f),
    C_Vertex3f(0.146235347f, 0.0249164104f, -1.f),
    C_Vertex3f(0.151774406f, 0.12616539f, -1.f),
    C_Vertex3f(0.160144806f, 0.0177793503f, -1.f),
    C_Vertex3f(0.165034294f, 0.112212658f, -1.f),
    C_Vertex3f(0.174846649f, 0.0114412308f, -1.f),
    C_Vertex3f(0.181164742f, 0.0997376442f, -1.f),
    C_Vertex3f(0.190429688f, 0.00589537621f, -1.f),
    C_Vertex3f(0.200165749f, 0.0887403488f, -1.f),
    C_Vertex3f(0.206893921f, 0.00114178658f, -1.f),
    C_Vertex3f(0.222037315f, 0.0792207718f, -1.f),
    C_Vertex3f(0.224239349f, -0.00281953812f, -1.f),
    C_Vertex3f(0.245275974f, 0.072081089f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0985213518f, 0.0614540577f, -1.f),
    C_Vertex3f(0.109261274f, 0.0511212349f, -1.f),
    C_Vertex3f(0.120117188f, 0.212890625f, -1.f),
    C_Vertex3f(0.120793581f, 0.0415873528f, -1.f),
    C_Vertex3f(0.121538162f, 0.200853348f, -1.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, -1.f),
    C_Vertex3f(0.126266956f, 0.178939819f, -1.f),
    C_Vertex3f(0.132882595f, 0.159187317f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120117188f, 0.212890625f, -1.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, -1.f),
    C_Vertex3f(0.0402574539f, 0.198022842f, -1.f),
    C_Vertex3f(0.0416849852f, 0.182375908f, -1.f),
    C_Vertex3f(0.0441032648f, 0.167145729f, -1.f),
    C_Vertex3f(0.0475122929f, 0.152332306f, -1.f),
    C_Vertex3f(0.0519120693f, 0.137935638f, -1.f),
    C_Vertex3f(0.0573025942f, 0.123955727f, -1.f),
    C_Vertex3f(0.0636838675f, 0.11039257f, -1.f),
    C_Vertex3f(0.0710558891f, 0.09724617f, -1.f),
    C_Vertex3f(0.0794186592f, 0.0845165253f, -1.f),
    C_Vertex3f(0.0885738134f, 0.0725858212f, -1.f),
    C_Vertex3f(0.0985213518f, 0.0614540577f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0667786598f, 0.443130016f, -1.f),
    C_Vertex3f(0.0639648438f, 0.487323761f, -1.f),
    C_Vertex3f(0.0639648438f, 0.464147568f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_84[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.231933594f, 0.f, 0.f),
    C_Vertex3f(0.231933594f, 0.f, -1.f),
    C_Vertex3f(0.231933594f, 0.564941406f, -1.f),
    C_Vertex3f(0.231933594f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.231933594f, 0.564941406f, 0.f),
    C_Vertex3f(0.231933594f, 0.564941406f, -1.f),
    C_Vertex3f(0.0209960938f, 0.564941406f, -1.f),
    C_Vertex3f(0.0209960938f, 0.564941406f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0209960938f, 0.564941406f, 0.f),
    C_Vertex3f(0.0209960938f, 0.564941406f, -1.f),
    C_Vertex3f(0.0209960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0209960938f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0209960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0209960938f, 0.640625f, -1.f),
    C_Vertex3f(0.528808594f, 0.640625f, -1.f),
    C_Vertex3f(0.528808594f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.528808594f, 0.640625f, 0.f),
    C_Vertex3f(0.528808594f, 0.640625f, -1.f),
    C_Vertex3f(0.528808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.528808594f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.528808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.528808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.316894531f, 0.564941406f, -1.f),
    C_Vertex3f(0.316894531f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.316894531f, 0.564941406f, 0.f),
    C_Vertex3f(0.316894531f, 0.564941406f, -1.f),
    C_Vertex3f(0.316894531f, 0.f, -1.f),
    C_Vertex3f(0.316894531f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.316894531f, 0.f, 0.f),
    C_Vertex3f(0.316894531f, 0.f, -1.f),
    C_Vertex3f(0.231933594f, 0.f, -1.f),
    C_Vertex3f(0.231933594f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.231933594f, 0.f, 0.f),
    C_Vertex3f(0.316894531f, 0.f, 0.f),
    C_Vertex3f(0.231933594f, 0.564941406f, 0.f),
    C_Vertex3f(0.316894531f, 0.564941406f, 0.f),
    C_Vertex3f(0.528808594f, 0.640625f, 0.f),
    C_Vertex3f(0.528808594f, 0.564941406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0209960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0209960938f, 0.564941406f, 0.f),
    C_Vertex3f(0.231933594f, 0.564941406f, 0.f),
    C_Vertex3f(0.528808594f, 0.640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.231933594f, 0.f, -1.f),
    C_Vertex3f(0.316894531f, 0.f, -1.f),
    C_Vertex3f(0.231933594f, 0.564941406f, -1.f),
    C_Vertex3f(0.316894531f, 0.564941406f, -1.f),
    C_Vertex3f(0.528808594f, 0.640625f, -1.f),
    C_Vertex3f(0.528808594f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0209960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0209960938f, 0.564941406f, -1.f),
    C_Vertex3f(0.231933594f, 0.564941406f, -1.f),
    C_Vertex3f(0.528808594f, 0.640625f, -1.f),
    C_End(),
    C_Translatef(0.546875f, 0.f, 0.f),
};
static const Call char_85[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Normal3f(0.999998629f, -0.00168604392f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, -1.f),
    C_Normal3f(0.998944402f, -0.0459359773f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, -1.f),
    C_Normal3f(0.996721923f, -0.0809027106f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, -1.f),
    C_Normal3f(0.992785931f, -0.119900197f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, -1.f),
    C_Normal3f(0.986551166f, -0.163452342f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, -1.f),
    C_Normal3f(0.977247715f, -0.212101072f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, -1.f),
    C_Normal3f(0.963873327f, -0.266360849f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, -1.f),
    C_Normal3f(0.945149362f, -0.32663849f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, -1.f),
    C_Normal3f(0.917468131f, -0.397809178f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, -1.f),
    C_Normal3f(0.88042134f, -0.474192262f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, -1.f),
    C_Normal3f(0.836648703f, -0.547739744f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, -1.f),
    C_Normal3f(0.787085712f, -0.616843581f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, -1.f),
    C_Normal3f(0.733018816f, -0.680208266f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, -1.f),
    C_Normal3f(0.675931513f, -0.736964524f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, -1.f),
    C_Normal3f(0.617332876f, -0.786702096f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Normal3f(0.558609724f, -0.82943064f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_Normal3f(0.490787864f, -0.871279061f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Normal3f(0.413442761f, -0.91053015f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, -1.f),
    C_Normal3f(0.337979287f, -0.941153526f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_Normal3f(0.265874416f, -0.964007616f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, -1.f),
    C_Normal3f(0.198111668f, -0.980179429f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, -1.f),
    C_Normal3f(0.135230228f, -0.990814209f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, -1.f),
    C_Normal3f(0.0774159282f, -0.996998906f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, -1.f),
    C_Normal3f(0.024781188f, -0.999692857f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0218659379f, -0.999760926f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0691669211f, -0.997605145f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, -1.f),
    C_Normal3f(-0.120602004f, -0.992700875f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, -1.f),
    C_Normal3f(-0.17647697f, -0.984304726f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, -1.f),
    C_Normal3f(-0.236778811f, -0.971563518f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, -1.f),
    C_Normal3f(-0.301268518f, -0.953539252f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Normal3f(-0.369405627f, -0.929268181f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Normal3f(-0.440283626f, -0.897858739f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, -1.f),
    C_Normal3f(-0.510246277f, -0.860028386f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Normal3f(-0.5782184f, -0.815882027f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_Normal3f(-0.645087063f, -0.764109015f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_Normal3f(-0.709094107f, -0.705113769f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, -1.f),
    C_Normal3f(-0.768517613f, -0.639828682f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, -1.f),
    C_Normal3f(-0.821880043f, -0.569660723f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, -1.f),
    C_Normal3f(-0.868125141f, -0.496345371f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, -1.f),
    C_Normal3f(-0.906716228f, -0.421741188f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, -1.f),
    C_Normal3f(-0.936816335f, -0.349821717f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, -1.f),
    C_Normal3f(-0.958463192f, -0.285216063f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, -1.f),
    C_Normal3f(-0.973905027f, -0.226956055f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, -1.f),
    C_Normal3f(-0.984618485f, -0.174717858f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, -1.f),
    C_Normal3f(-0.991773546f, -0.128004253f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, -1.f),
    C_Normal3f(-0.99627322f, -0.0862539262f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, -1.f),
    C_Normal3f(-0.998803377f, -0.048905924f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Normal3f(0.999999225f, 0.00121280376f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_Normal3f(0.99925375f, 0.0386249162f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, -1.f),
    C_Normal3f(0.997593999f, 0.0693271309f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Normal3f(0.99447f, 0.105021358f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, -1.f),
    C_Normal3f(0.989158213f, 0.14685446f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, -1.f),
    C_Normal3f(0.980551183f, 0.196263924f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, -1.f),
    C_Normal3f(0.966938674f, 0.255009204f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, -1.f),
    C_Normal3f(0.942981362f, 0.332845628f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Normal3f(0.892347693f, 0.451348603f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, -1.f),
    C_Normal3f(0.808162808f, 0.588959098f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, -1.f),
    C_Normal3f(0.691974759f, 0.721921682f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_Normal3f(0.560325563f, 0.828272521f, 0.f),
    C_Vertex3f(0.214595795f, 0.0926685333f, 0.f),
    C_Vertex3f(0.214595795f, 0.0926685333f, -1.f),
    C_Normal3f(0.425149351f, 0.905123174f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_Normal3f(0.290447056f, 0.95689106f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, -1.f),
    C_Normal3f(0.164019689f, 0.986457109f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, -1.f),
    C_Normal3f(0.0528943539f, 0.998600125f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, -1.f),
    C_Normal3f(-0.0281555988f, 0.999603629f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, -1.f),
    C_Normal3f(-0.0948538408f, 0.995491266f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, -1.f),
    C_Normal3f(-0.170193285f, 0.98541069f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, -1.f),
    C_Normal3f(-0.256163627f, 0.966633379f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, -1.f),
    C_Normal3f(-0.352678686f, 0.935744524f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, -1.f),
    C_Normal3f(-0.458108366f, 0.888896406f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_Normal3f(-0.568576157f, 0.822630644f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, -1.f),
    C_Normal3f(-0.677751124f, 0.735291302f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, -1.f),
    C_Normal3f(-0.785317659f, 0.619092882f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_Normal3f(-0.870217264f, 0.492668152f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, -1.f),
    C_Normal3f(-0.92558831f, 0.378531873f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, -1.f),
    C_Normal3f(-0.959939063f, 0.280209035f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, -1.f),
    C_Normal3f(-0.980327249f, 0.197379053f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Normal3f(-0.991757751f, 0.12812756f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, -1.f),
    C_Normal3f(-0.997533739f, 0.0701881126f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214595795f, 0.0926685333f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, -1.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, -1.f),
    C_Vertex3f(0.077960968f, 0.169532061f, -1.f),
    C_Vertex3f(0.081785202f, 0.150808215f, -1.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, -1.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, -1.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, -1.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, -1.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, -1.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, -1.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, -1.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, -1.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, -1.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, -1.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, -1.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, -1.f),
    C_Vertex3f(0.549636841f, 0.110161304f, -1.f),
    C_Vertex3f(0.555099487f, 0.124358654f, -1.f),
    C_Vertex3f(0.559879303f, 0.139798522f, -1.f),
    C_Vertex3f(0.563976288f, 0.156480908f, -1.f),
    C_Vertex3f(0.567390442f, 0.174405813f, -1.f),
    C_Vertex3f(0.570121765f, 0.193573236f, -1.f),
    C_Vertex3f(0.572170258f, 0.213983178f, -1.f),
    C_Vertex3f(0.573535919f, 0.235635638f, -1.f),
    C_Vertex3f(0.57421875f, 0.258530617f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_Vertex3f(0.488016248f, 0.228415847f, -1.f),
    C_Vertex3f(0.485533118f, 0.203168988f, -1.f),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Vertex3f(0.476842165f, 0.160003424f, -1.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_Vertex3f(0.470634341f, 0.142084718f, -1.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Vertex3f(0.463184953f, 0.126608729f, -1.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, -1.f),
    C_Vertex3f(0.444561481f, 0.102984905f, -1.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, -1.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, -1.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, -1.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, -1.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, -1.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, -1.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, -1.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, -1.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, -1.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, -1.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, -1.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, -1.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, -1.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, -1.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, -1.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, -1.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, -1.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, -1.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, -1.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214595795f, 0.0926685333f, -1.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Vertex3f(0.174647331f, 0.13835907f, -1.f),
    C_Vertex3f(0.183670044f, 0.016397953f, -1.f),
    C_Vertex3f(0.185180187f, 0.120863914f, -1.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Vertex3f(0.165444732f, 0.164430857f, -1.f),
    C_Vertex3f(0.162538648f, 0.177054167f, -1.f),
    C_Vertex3f(0.160116911f, 0.191049099f, -1.f),
    C_Vertex3f(0.158179522f, 0.206415653f, -1.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Vertex3f(0.155757785f, 0.241263628f, -1.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_86[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.93283987f, -0.360291183f, 0.f),
    C_Vertex3f(0.251373291f, 0.f, 0.f),
    C_Vertex3f(0.251373291f, 0.f, -1.f),
    C_Vertex3f(0.00394439697f, 0.640625f, -1.f),
    C_Vertex3f(0.00394439697f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.00394439697f, 0.640625f, 0.f),
    C_Vertex3f(0.00394439697f, 0.640625f, -1.f),
    C_Vertex3f(0.095703125f, 0.640625f, -1.f),
    C_Vertex3f(0.095703125f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.941609979f, 0.336705565f, 0.f),
    C_Vertex3f(0.095703125f, 0.640625f, 0.f),
    C_Vertex3f(0.095703125f, 0.640625f, -1.f),
    C_Normal3f(0.94188267f, 0.335941851f, 0.f),
    C_Vertex3f(0.267059326f, 0.161420822f, 0.f),
    C_Vertex3f(0.267059326f, 0.161420822f, -1.f),
    C_Normal3f(0.949424803f, 0.313994527f, 0.f),
    C_Vertex3f(0.276281357f, 0.134336472f, 0.f),
    C_Vertex3f(0.276281357f, 0.134336472f, -1.f),
    C_Normal3f(0.955107987f, 0.296258062f, 0.f),
    C_Vertex3f(0.284685135f, 0.108125687f, 0.f),
    C_Vertex3f(0.284685135f, 0.108125687f, -1.f),
    C_Normal3f(0.959920824f, 0.280271351f, 0.f),
    C_Vertex3f(0.29227066f, 0.0827884674f, 0.f),
    C_Vertex3f(0.29227066f, 0.0827884674f, -1.f),
    C_Normal3f(0.963803768f, 0.266612798f, 0.f),
    C_Vertex3f(0.295654297f, 0.0705566406f, 0.f),
    C_Vertex3f(0.295654297f, 0.0705566406f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.962045312f, 0.272889882f, 0.f),
    C_Vertex3f(0.295654297f, 0.0705566406f, 0.f),
    C_Vertex3f(0.295654297f, 0.0705566406f, -1.f),
    C_Normal3f(-0.953998625f, 0.299810946f, 0.f),
    C_Vertex3f(0.303089142f, 0.0967674255f, 0.f),
    C_Vertex3f(0.303089142f, 0.0967674255f, -1.f),
    C_Normal3f(-0.93710202f, 0.349055529f, 0.f),
    C_Vertex3f(0.320365906f, 0.149188995f, 0.f),
    C_Vertex3f(0.320365906f, 0.149188995f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.93710202f, 0.349055529f, 0.f),
    C_Vertex3f(0.320365906f, 0.149188995f, 0.f),
    C_Vertex3f(0.320365906f, 0.149188995f, -1.f),
    C_Vertex3f(0.503417969f, 0.640625f, -1.f),
    C_Vertex3f(0.503417969f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.503417969f, 0.640625f, 0.f),
    C_Vertex3f(0.503417969f, 0.640625f, -1.f),
    C_Vertex3f(0.589942932f, 0.640625f, -1.f),
    C_Vertex3f(0.589942932f, 0.640625f, 0.f),
    C_Normal3f(0.931024611f, -0.364956588f, 0.f),
    C_Vertex3f(0.589942932f, 0.640625f, 0.f),
    C_Vertex3f(0.589942932f, 0.640625f, -1.f),
    C_Vertex3f(0.338821411f, 0.f, -1.f),
    C_Vertex3f(0.338821411f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.338821411f, 0.f, 0.f),
    C_Vertex3f(0.338821411f, 0.f, -1.f),
    C_Vertex3f(0.251373291f, 0.f, -1.f),
    C_Vertex3f(0.251373291f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338821411f, 0.f, 0.f),
    C_Vertex3f(0.589942932f, 0.640625f, 0.f),
    C_Vertex3f(0.503417969f, 0.640625f, 0.f),
    C_Vertex3f(0.320365906f, 0.149188995f, 0.f),
    C_Vertex3f(0.303089142f, 0.0967674255f, 0.f),
    C_Vertex3f(0.295654297f, 0.0705566406f, 0.f),
    C_Vertex3f(0.251373291f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.295654297f, 0.0705566406f, 0.f),
    C_Vertex3f(0.29227066f, 0.0827884674f, 0.f),
    C_Vertex3f(0.284685135f, 0.108125687f, 0.f),
    C_Vertex3f(0.276281357f, 0.134336472f, 0.f),
    C_Vertex3f(0.267059326f, 0.161420822f, 0.f),
    C_Vertex3f(0.251373291f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.095703125f, 0.640625f, 0.f),
    C_Vertex3f(0.00394439697f, 0.640625f, 0.f),
    C_Vertex3f(0.251373291f, 0.f, 0.f),
    C_Vertex3f(0.267059326f, 0.161420822f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338821411f, 0.f, -1.f),
    C_Vertex3f(0.589942932f, 0.640625f, -1.f),
    C_Vertex3f(0.503417969f, 0.640625f, -1.f),
    C_Vertex3f(0.320365906f, 0.149188995f, -1.f),
    C_Vertex3f(0.303089142f, 0.0967674255f, -1.f),
    C_Vertex3f(0.295654297f, 0.0705566406f, -1.f),
    C_Vertex3f(0.251373291f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.295654297f, 0.0705566406f, -1.f),
    C_Vertex3f(0.29227066f, 0.0827884674f, -1.f),
    C_Vertex3f(0.284685135f, 0.108125687f, -1.f),
    C_Vertex3f(0.276281357f, 0.134336472f, -1.f),
    C_Vertex3f(0.267059326f, 0.161420822f, -1.f),
    C_Vertex3f(0.251373291f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.095703125f, 0.640625f, -1.f),
    C_Vertex3f(0.00394439697f, 0.640625f, -1.f),
    C_Vertex3f(0.251373291f, 0.f, -1.f),
    C_Vertex3f(0.267059326f, 0.161420822f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_87[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.966547966f, -0.256486058f, 0.f),
    C_Vertex3f(0.180740356f, 0.f, 0.f),
    C_Vertex3f(0.180740356f, 0.f, -1.f),
    C_Vertex3f(0.0107421875f, 0.640625f, -1.f),
    C_Vertex3f(0.0107421875f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0107421875f, 0.640625f, 0.f),
    C_Vertex3f(0.0107421875f, 0.640625f, -1.f),
    C_Vertex3f(0.0977172852f, 0.640625f, -1.f),
    C_Vertex3f(0.0977172852f, 0.640625f, 0.f),
    C_Normal3f(0.974074125f, 0.226229265f, 0.f),
    C_Vertex3f(0.0977172852f, 0.640625f, 0.f),
    C_Vertex3f(0.0977172852f, 0.640625f, -1.f),
    C_Vertex3f(0.195251465f, 0.220672607f, -1.f),
    C_Vertex3f(0.195251465f, 0.220672607f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.974074125f, 0.226229265f, 0.f),
    C_Vertex3f(0.195251465f, 0.220672607f, 0.f),
    C_Vertex3f(0.195251465f, 0.220672607f, -1.f),
    C_Normal3f(0.976917565f, 0.213616788f, 0.f),
    C_Vertex3f(0.203212738f, 0.187702179f, 0.f),
    C_Vertex3f(0.203212738f, 0.187702179f, -1.f),
    C_Normal3f(0.981388152f, 0.192034781f, 0.f),
    C_Vertex3f(0.21679306f, 0.122158051f, 0.f),
    C_Vertex3f(0.21679306f, 0.122158051f, -1.f),
    C_Normal3f(0.98544544f, 0.16999194f, 0.f),
    C_Vertex3f(0.222412109f, 0.0895843506f, 0.f),
    C_Vertex3f(0.222412109f, 0.0895843506f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.972626746f, 0.232372612f, 0.f),
    C_Vertex3f(0.222412109f, 0.0895843506f, 0.f),
    C_Vertex3f(0.222412109f, 0.0895843506f, -1.f),
    C_Normal3f(-0.972521484f, 0.232813105f, 0.f),
    C_Vertex3f(0.225479126f, 0.102421761f, 0.f),
    C_Vertex3f(0.225479126f, 0.102421761f, -1.f),
    C_Normal3f(-0.972364604f, 0.233467326f, 0.f),
    C_Vertex3f(0.230981112f, 0.125379324f, 0.f),
    C_Vertex3f(0.230981112f, 0.125379324f, -1.f),
    C_Normal3f(-0.972126067f, 0.234458953f, 0.f),
    C_Vertex3f(0.235851049f, 0.145619631f, 0.f),
    C_Vertex3f(0.235851049f, 0.145619631f, -1.f),
    C_Normal3f(-0.971805751f, 0.235782787f, 0.f),
    C_Vertex3f(0.24008894f, 0.163142681f, 0.f),
    C_Vertex3f(0.24008894f, 0.163142681f, -1.f),
    C_Normal3f(-0.971353471f, 0.237639427f, 0.f),
    C_Vertex3f(0.243694782f, 0.177948475f, 0.f),
    C_Vertex3f(0.243694782f, 0.177948475f, -1.f),
    C_Normal3f(-0.970666111f, 0.240431517f, 0.f),
    C_Vertex3f(0.246668577f, 0.190037012f, 0.f),
    C_Vertex3f(0.246668577f, 0.190037012f, -1.f),
    C_Normal3f(-0.969496787f, 0.245103896f, 0.f),
    C_Vertex3f(0.249010324f, 0.199408293f, 0.f),
    C_Vertex3f(0.249010324f, 0.199408293f, -1.f),
    C_Normal3f(-0.962519109f, 0.271213859f, 0.f),
    C_Vertex3f(0.250720024f, 0.206062317f, 0.f),
    C_Vertex3f(0.250720024f, 0.206062317f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.962519109f, 0.271213859f, 0.f),
    C_Vertex3f(0.250720024f, 0.206062317f, 0.f),
    C_Vertex3f(0.250720024f, 0.206062317f, -1.f),
    C_Vertex3f(0.373168945f, 0.640625f, -1.f),
    C_Vertex3f(0.373168945f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.373168945f, 0.640625f, 0.f),
    C_Vertex3f(0.373168945f, 0.640625f, -1.f),
    C_Vertex3f(0.475440979f, 0.640625f, -1.f),
    C_Vertex3f(0.475440979f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.962182641f, 0.272405237f, 0.f),
    C_Vertex3f(0.475440979f, 0.640625f, 0.f),
    C_Vertex3f(0.475440979f, 0.640625f, -1.f),
    C_Normal3f(0.963308394f, 0.268396944f, 0.f),
    C_Vertex3f(0.575777054f, 0.286220551f, 0.f),
    C_Vertex3f(0.575777054f, 0.286220551f, -1.f),
    C_Normal3f(0.97337842f, 0.229204118f, 0.f),
    C_Vertex3f(0.59049511f, 0.227682114f, 0.f),
    C_Vertex3f(0.59049511f, 0.227682114f, -1.f),
    C_Normal3f(0.980202794f, 0.197995976f, 0.f),
    C_Vertex3f(0.602913857f, 0.170976639f, 0.f),
    C_Vertex3f(0.602913857f, 0.170976639f, -1.f),
    C_Normal3f(0.985467553f, 0.169863999f, 0.f),
    C_Vertex3f(0.613033295f, 0.116104126f, 0.f),
    C_Vertex3f(0.613033295f, 0.116104126f, -1.f),
    C_Normal3f(0.989304841f, 0.145862699f, 0.f),
    C_Vertex3f(0.616943359f, 0.0895843506f, 0.f),
    C_Vertex3f(0.616943359f, 0.0895843506f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.979242742f, 0.202691063f, 0.f),
    C_Vertex3f(0.616943359f, 0.0895843506f, 0.f),
    C_Vertex3f(0.616943359f, 0.0895843506f, -1.f),
    C_Normal3f(-0.977631569f, 0.210324541f, 0.f),
    C_Vertex3f(0.620084763f, 0.104761124f, 0.f),
    C_Vertex3f(0.620084763f, 0.104761124f, -1.f),
    C_Normal3f(-0.975755036f, 0.218865186f, 0.f),
    C_Vertex3f(0.627223015f, 0.137366295f, 0.f),
    C_Vertex3f(0.627223015f, 0.137366295f, -1.f),
    C_Normal3f(-0.973688483f, 0.227883443f, 0.f),
    C_Vertex3f(0.635216713f, 0.172223091f, 0.f),
    C_Vertex3f(0.635216713f, 0.172223091f, -1.f),
    C_Normal3f(-0.971524f, 0.236941144f, 0.f),
    C_Vertex3f(0.644065857f, 0.209331512f, 0.f),
    C_Vertex3f(0.644065857f, 0.209331512f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.971524f, 0.236941144f, 0.f),
    C_Vertex3f(0.644065857f, 0.209331512f, 0.f),
    C_Vertex3f(0.644065857f, 0.209331512f, -1.f),
    C_Vertex3f(0.749252319f, 0.640625f, -1.f),
    C_Vertex3f(0.749252319f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.749252319f, 0.640625f, 0.f),
    C_Vertex3f(0.749252319f, 0.640625f, -1.f),
    C_Vertex3f(0.834472656f, 0.640625f, -1.f),
    C_Vertex3f(0.834472656f, 0.640625f, 0.f),
    C_Normal3f(0.964694917f, -0.263370156f, 0.f),
    C_Vertex3f(0.834472656f, 0.640625f, 0.f),
    C_Vertex3f(0.834472656f, 0.640625f, -1.f),
    C_Vertex3f(0.659576416f, 0.f, -1.f),
    C_Vertex3f(0.659576416f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.659576416f, 0.f, 0.f),
    C_Vertex3f(0.659576416f, 0.f, -1.f),
    C_Vertex3f(0.576850891f, 0.f, -1.f),
    C_Vertex3f(0.576850891f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.964246035f, -0.265008509f, 0.f),
    C_Vertex3f(0.576850891f, 0.f, 0.f),
    C_Vertex3f(0.576850891f, 0.f, -1.f),
    C_Normal3f(-0.964201152f, -0.265172005f, 0.f),
    C_Vertex3f(0.440556526f, 0.495913506f, 0.f),
    C_Vertex3f(0.440556526f, 0.495913506f, -1.f),
    C_Normal3f(-0.962778628f, -0.270290941f, 0.f),
    C_Vertex3f(0.436669827f, 0.509717226f, 0.f),
    C_Vertex3f(0.436669827f, 0.509717226f, -1.f),
    C_Normal3f(-0.963271737f, -0.268528312f, 0.f),
    C_Vertex3f(0.433220387f, 0.522044897f, 0.f),
    C_Vertex3f(0.433220387f, 0.522044897f, -1.f),
    C_Normal3f(-0.963904858f, -0.266247183f, 0.f),
    C_Vertex3f(0.430208206f, 0.532896519f, 0.f),
    C_Vertex3f(0.430208206f, 0.532896519f, -1.f),
    C_Normal3f(-0.964746892f, -0.263179362f, 0.f),
    C_Vertex3f(0.427633286f, 0.542272091f, 0.f),
    C_Vertex3f(0.427633286f, 0.542272091f, -1.f),
    C_Normal3f(-0.965921998f, -0.258833349f, 0.f),
    C_Vertex3f(0.425495625f, 0.550171614f, 0.f),
    C_Vertex3f(0.425495625f, 0.550171614f, -1.f),
    C_Normal3f(-0.967674971f, -0.252200544f, 0.f),
    C_Vertex3f(0.423795223f, 0.556595087f, 0.f),
    C_Vertex3f(0.423795223f, 0.556595087f, -1.f),
    C_Normal3f(-0.969961047f, -0.24326019f, 0.f),
    C_Vertex3f(0.422532082f, 0.561542511f, 0.f),
    C_Vertex3f(0.422532082f, 0.561542511f, -1.f),
    C_Normal3f(-0.972846329f, -0.231451914f, 0.f),
    C_Vertex3f(0.422119141f, 0.563278198f, 0.f),
    C_Vertex3f(0.422119141f, 0.563278198f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.972573876f, -0.232594028f, 0.f),
    C_Vertex3f(0.422119141f, 0.563278198f, 0.f),
    C_Vertex3f(0.422119141f, 0.563278198f, -1.f),
    C_Normal3f(0.971627176f, -0.236517429f, 0.f),
    C_Vertex3f(0.419490814f, 0.552288055f, 0.f),
    C_Vertex3f(0.419490814f, 0.552288055f, -1.f),
    C_Normal3f(0.970274508f, -0.242007062f, 0.f),
    C_Vertex3f(0.414484978f, 0.531916618f, 0.f),
    C_Vertex3f(0.414484978f, 0.531916618f, -1.f),
    C_Normal3f(0.968339324f, -0.249637723f, 0.f),
    C_Vertex3f(0.409729958f, 0.51315403f, 0.f),
    C_Vertex3f(0.409729958f, 0.51315403f, -1.f),
    C_Normal3f(0.96379292f, -0.266651839f, 0.f),
    C_Vertex3f(0.405225754f, 0.49600029f, 0.f),
    C_Vertex3f(0.405225754f, 0.49600029f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.96379292f, -0.266651839f, 0.f),
    C_Vertex3f(0.405225754f, 0.49600029f, 0.f),
    C_Vertex3f(0.405225754f, 0.49600029f, -1.f),
    C_Vertex3f(0.267997742f, 0.f, -1.f),
    C_Vertex3f(0.267997742f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.267997742f, 0.f, 0.f),
    C_Vertex3f(0.267997742f, 0.f, -1.f),
    C_Vertex3f(0.180740356f, 0.f, -1.f),
    C_Vertex3f(0.180740356f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.659576416f, 0.f, 0.f),
    C_Vertex3f(0.834472656f, 0.640625f, 0.f),
    C_Vertex3f(0.749252319f, 0.640625f, 0.f),
    C_Vertex3f(0.644065857f, 0.209331512f, 0.f),
    C_Vertex3f(0.635216713f, 0.172223091f, 0.f),
    C_Vertex3f(0.627223015f, 0.137366295f, 0.f),
    C_Vertex3f(0.620084763f, 0.104761124f, 0.f),
    C_Vertex3f(0.616943359f, 0.0895843506f, 0.f),
    C_Vertex3f(0.576850891f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.613033295f, 0.116104126f, 0.f),
    C_Vertex3f(0.602913857f, 0.170976639f, 0.f),
    C_Vertex3f(0.616943359f, 0.0895843506f, 0.f),
    C_Vertex3f(0.59049511f, 0.227682114f, 0.f),
    C_Vertex3f(0.576850891f, 0.f, 0.f),
    C_Vertex3f(0.575777054f, 0.286220551f, 0.f),
    C_Vertex3f(0.475440979f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.475440979f, 0.640625f, 0.f),
    C_Vertex3f(0.373168945f, 0.640625f, 0.f),
    C_Vertex3f(0.422119141f, 0.563278198f, 0.f),
    C_Vertex3f(0.422532082f, 0.561542511f, 0.f),
    C_Vertex3f(0.423795223f, 0.556595087f, 0.f),
    C_Vertex3f(0.425495625f, 0.550171614f, 0.f),
    C_Vertex3f(0.427633286f, 0.542272091f, 0.f),
    C_Vertex3f(0.430208206f, 0.532896519f, 0.f),
    C_Vertex3f(0.433220387f, 0.522044897f, 0.f),
    C_Vertex3f(0.436669827f, 0.509717226f, 0.f),
    C_Vertex3f(0.440556526f, 0.495913506f, 0.f),
    C_Vertex3f(0.576850891f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.422119141f, 0.563278198f, 0.f),
    C_Vertex3f(0.373168945f, 0.640625f, 0.f),
    C_Vertex3f(0.405225754f, 0.49600029f, 0.f),
    C_Vertex3f(0.409729958f, 0.51315403f, 0.f),
    C_Vertex3f(0.414484978f, 0.531916618f, 0.f),
    C_Vertex3f(0.419490814f, 0.552288055f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.267997742f, 0.f, 0.f),
    C_Vertex3f(0.405225754f, 0.49600029f, 0.f),
    C_Vertex3f(0.373168945f, 0.640625f, 0.f),
    C_Vertex3f(0.250720024f, 0.206062317f, 0.f),
    C_Vertex3f(0.249010324f, 0.199408293f, 0.f),
    C_Vertex3f(0.246668577f, 0.190037012f, 0.f),
    C_Vertex3f(0.243694782f, 0.177948475f, 0.f),
    C_Vertex3f(0.24008894f, 0.163142681f, 0.f),
    C_Vertex3f(0.235851049f, 0.145619631f, 0.f),
    C_Vertex3f(0.230981112f, 0.125379324f, 0.f),
    C_Vertex3f(0.225479126f, 0.102421761f, 0.f),
    C_Vertex3f(0.222412109f, 0.0895843506f, 0.f),
    C_Vertex3f(0.180740356f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.21679306f, 0.122158051f, 0.f),
    C_Vertex3f(0.203212738f, 0.187702179f, 0.f),
    C_Vertex3f(0.222412109f, 0.0895843506f, 0.f),
    C_Vertex3f(0.195251465f, 0.220672607f, 0.f),
    C_Vertex3f(0.180740356f, 0.f, 0.f),
    C_Vertex3f(0.0977172852f, 0.640625f, 0.f),
    C_Vertex3f(0.0107421875f, 0.640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.659576416f, 0.f, -1.f),
    C_Vertex3f(0.834472656f, 0.640625f, -1.f),
    C_Vertex3f(0.749252319f, 0.640625f, -1.f),
    C_Vertex3f(0.644065857f, 0.209331512f, -1.f),
    C_Vertex3f(0.635216713f, 0.172223091f, -1.f),
    C_Vertex3f(0.627223015f, 0.137366295f, -1.f),
    C_Vertex3f(0.620084763f, 0.104761124f, -1.f),
    C_Vertex3f(0.616943359f, 0.0895843506f, -1.f),
    C_Vertex3f(0.576850891f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.613033295f, 0.116104126f, -1.f),
    C_Vertex3f(0.602913857f, 0.170976639f, -1.f),
    C_Vertex3f(0.616943359f, 0.0895843506f, -1.f),
    C_Vertex3f(0.59049511f, 0.227682114f, -1.f),
    C_Vertex3f(0.576850891f, 0.f, -1.f),
    C_Vertex3f(0.575777054f, 0.286220551f, -1.f),
    C_Vertex3f(0.475440979f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.475440979f, 0.640625f, -1.f),
    C_Vertex3f(0.373168945f, 0.640625f, -1.f),
    C_Vertex3f(0.422119141f, 0.563278198f, -1.f),
    C_Vertex3f(0.422532082f, 0.561542511f, -1.f),
    C_Vertex3f(0.423795223f, 0.556595087f, -1.f),
    C_Vertex3f(0.425495625f, 0.550171614f, -1.f),
    C_Vertex3f(0.427633286f, 0.542272091f, -1.f),
    C_Vertex3f(0.430208206f, 0.532896519f, -1.f),
    C_Vertex3f(0.433220387f, 0.522044897f, -1.f),
    C_Vertex3f(0.436669827f, 0.509717226f, -1.f),
    C_Vertex3f(0.440556526f, 0.495913506f, -1.f),
    C_Vertex3f(0.576850891f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.422119141f, 0.563278198f, -1.f),
    C_Vertex3f(0.373168945f, 0.640625f, -1.f),
    C_Vertex3f(0.405225754f, 0.49600029f, -1.f),
    C_Vertex3f(0.409729958f, 0.51315403f, -1.f),
    C_Vertex3f(0.414484978f, 0.531916618f, -1.f),
    C_Vertex3f(0.419490814f, 0.552288055f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.267997742f, 0.f, -1.f),
    C_Vertex3f(0.405225754f, 0.49600029f, -1.f),
    C_Vertex3f(0.373168945f, 0.640625f, -1.f),
    C_Vertex3f(0.250720024f, 0.206062317f, -1.f),
    C_Vertex3f(0.249010324f, 0.199408293f, -1.f),
    C_Vertex3f(0.246668577f, 0.190037012f, -1.f),
    C_Vertex3f(0.243694782f, 0.177948475f, -1.f),
    C_Vertex3f(0.24008894f, 0.163142681f, -1.f),
    C_Vertex3f(0.235851049f, 0.145619631f, -1.f),
    C_Vertex3f(0.230981112f, 0.125379324f, -1.f),
    C_Vertex3f(0.225479126f, 0.102421761f, -1.f),
    C_Vertex3f(0.222412109f, 0.0895843506f, -1.f),
    C_Vertex3f(0.180740356f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.21679306f, 0.122158051f, -1.f),
    C_Vertex3f(0.203212738f, 0.187702179f, -1.f),
    C_Vertex3f(0.222412109f, 0.0895843506f, -1.f),
    C_Vertex3f(0.195251465f, 0.220672607f, -1.f),
    C_Vertex3f(0.180740356f, 0.f, -1.f),
    C_Vertex3f(0.0977172852f, 0.640625f, -1.f),
    C_Vertex3f(0.0107421875f, 0.640625f, -1.f),
    C_End(),
    C_Translatef(0.844726563f, 0.f, 0.f),
};
static const Call char_88[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.801677883f, 0.597756326f, 0.f),
    C_Vertex3f(0.00393676758f, 0.f, 0.f),
    C_Vertex3f(0.00393676758f, 0.f, -1.f),
    C_Vertex3f(0.251686096f, 0.332267761f, -1.f),
    C_Vertex3f(0.251686096f, 0.332267761f, 0.f),
    C_Normal3f(-0.815983593f, -0.578074992f, 0.f),
    C_Vertex3f(0.251686096f, 0.332267761f, 0.f),
    C_Vertex3f(0.251686096f, 0.332267761f, -1.f),
    C_Vertex3f(0.0332336426f, 0.640625f, -1.f),
    C_Vertex3f(0.0332336426f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0332336426f, 0.640625f, 0.f),
    C_Vertex3f(0.0332336426f, 0.640625f, -1.f),
    C_Vertex3f(0.133911133f, 0.640625f, -1.f),
    C_Vertex3f(0.133911133f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.813825548f, 0.581109285f, 0.f),
    C_Vertex3f(0.133911133f, 0.640625f, 0.f),
    C_Vertex3f(0.133911133f, 0.640625f, -1.f),
    C_Normal3f(0.814137697f, 0.580671787f, 0.f),
    C_Vertex3f(0.255493164f, 0.470353127f, 0.f),
    C_Vertex3f(0.255493164f, 0.470353127f, -1.f),
    C_Normal3f(0.821179628f, 0.570669889f, 0.f),
    C_Vertex3f(0.263934851f, 0.458323717f, 0.f),
    C_Vertex3f(0.263934851f, 0.458323717f, -1.f),
    C_Normal3f(0.826943874f, 0.56228435f, 0.f),
    C_Vertex3f(0.271708727f, 0.447019339f, 0.f),
    C_Vertex3f(0.271708727f, 0.447019339f, -1.f),
    C_Normal3f(0.833518207f, 0.552492082f, 0.f),
    C_Vertex3f(0.278814793f, 0.436439991f, 0.f),
    C_Vertex3f(0.278814793f, 0.436439991f, -1.f),
    C_Normal3f(0.84107852f, 0.540912867f, 0.f),
    C_Vertex3f(0.285253048f, 0.426585674f, 0.f),
    C_Vertex3f(0.285253048f, 0.426585674f, -1.f),
    C_Normal3f(0.849853098f, 0.52701962f, 0.f),
    C_Vertex3f(0.291023493f, 0.417456388f, 0.f),
    C_Vertex3f(0.291023493f, 0.417456388f, -1.f),
    C_Normal3f(0.860138953f, 0.510060012f, 0.f),
    C_Vertex3f(0.296126127f, 0.409052134f, 0.f),
    C_Vertex3f(0.296126127f, 0.409052134f, -1.f),
    C_Normal3f(0.870142043f, 0.492800921f, 0.f),
    C_Vertex3f(0.300560951f, 0.40137291f, 0.f),
    C_Vertex3f(0.300560951f, 0.40137291f, -1.f),
    C_Normal3f(0.879283607f, 0.47629869f, 0.f),
    C_Vertex3f(0.302444458f, 0.397895813f, 0.f),
    C_Vertex3f(0.302444458f, 0.397895813f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.849212348f, 0.528051555f, 0.f),
    C_Vertex3f(0.302444458f, 0.397895813f, 0.f),
    C_Vertex3f(0.302444458f, 0.397895813f, -1.f),
    C_Normal3f(-0.838466525f, 0.544952989f, 0.f),
    C_Vertex3f(0.307794571f, 0.406499863f, 0.f),
    C_Vertex3f(0.307794571f, 0.406499863f, -1.f),
    C_Normal3f(-0.825458586f, 0.564462662f, 0.f),
    C_Vertex3f(0.319502831f, 0.424141884f, 0.f),
    C_Vertex3f(0.319502831f, 0.424141884f, -1.f),
    C_Normal3f(-0.810494721f, 0.585745811f, 0.f),
    C_Vertex3f(0.332219124f, 0.442217827f, 0.f),
    C_Vertex3f(0.332219124f, 0.442217827f, -1.f),
    C_Normal3f(-0.798570514f, 0.601901352f, 0.f),
    C_Vertex3f(0.345943451f, 0.460727692f, 0.f),
    C_Vertex3f(0.345943451f, 0.460727692f, -1.f),
    C_Normal3f(-0.798925757f, 0.601429641f, 0.f),
    C_Vertex3f(0.353309631f, 0.470199585f, 0.f),
    C_Vertex3f(0.353309631f, 0.470199585f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.798925757f, 0.601429641f, 0.f),
    C_Vertex3f(0.353309631f, 0.470199585f, 0.f),
    C_Vertex3f(0.353309631f, 0.470199585f, -1.f),
    C_Vertex3f(0.48160553f, 0.640625f, -1.f),
    C_Vertex3f(0.48160553f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.48160553f, 0.640625f, 0.f),
    C_Vertex3f(0.48160553f, 0.640625f, -1.f),
    C_Vertex3f(0.573760986f, 0.640625f, -1.f),
    C_Vertex3f(0.573760986f, 0.640625f, 0.f),
    C_Normal3f(0.807480454f, -0.589894295f, 0.f),
    C_Vertex3f(0.573760986f, 0.640625f, 0.f),
    C_Vertex3f(0.573760986f, 0.640625f, -1.f),
    C_Vertex3f(0.351325989f, 0.336143494f, -1.f),
    C_Vertex3f(0.351325989f, 0.336143494f, 0.f),
    C_Normal3f(0.8138358f, 0.581094861f, 0.f),
    C_Vertex3f(0.351325989f, 0.336143494f, 0.f),
    C_Vertex3f(0.351325989f, 0.336143494f, -1.f),
    C_Vertex3f(0.591339111f, 0.f, -1.f),
    C_Vertex3f(0.591339111f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.591339111f, 0.f, 0.f),
    C_Vertex3f(0.591339111f, 0.f, -1.f),
    C_Vertex3f(0.487068176f, 0.f, -1.f),
    C_Vertex3f(0.487068176f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.815992892f, -0.578061879f, 0.f),
    C_Vertex3f(0.487068176f, 0.f, 0.f),
    C_Vertex3f(0.487068176f, 0.f, -1.f),
    C_Normal3f(-0.81770736f, -0.575634122f, 0.f),
    C_Vertex3f(0.311347961f, 0.248046875f, 0.f),
    C_Vertex3f(0.311347961f, 0.248046875f, -1.f),
    C_Normal3f(-0.839068532f, -0.54402566f, 0.f),
    C_Vertex3f(0.298461914f, 0.267921448f, 0.f),
    C_Vertex3f(0.298461914f, 0.267921448f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.842958152f, -0.537979126f, 0.f),
    C_Vertex3f(0.298461914f, 0.267921448f, 0.f),
    C_Vertex3f(0.298461914f, 0.267921448f, -1.f),
    C_Normal3f(0.840375483f, -0.542004645f, 0.f),
    C_Vertex3f(0.293230057f, 0.259723663f, 0.f),
    C_Vertex3f(0.293230057f, 0.259723663f, -1.f),
    C_Normal3f(0.836190939f, -0.548438549f, 0.f),
    C_Vertex3f(0.284202576f, 0.245812416f, 0.f),
    C_Vertex3f(0.284202576f, 0.245812416f, -1.f),
    C_Normal3f(0.828908503f, -0.559384167f, 0.f),
    C_Vertex3f(0.276611328f, 0.23438549f, 0.f),
    C_Vertex3f(0.276611328f, 0.23438549f, -1.f),
    C_Normal3f(0.807474852f, -0.589902043f, 0.f),
    C_Vertex3f(0.270456314f, 0.225442886f, 0.f),
    C_Vertex3f(0.270456314f, 0.225442886f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.807474852f, -0.589902043f, 0.f),
    C_Vertex3f(0.270456314f, 0.225442886f, 0.f),
    C_Vertex3f(0.270456314f, 0.225442886f, -1.f),
    C_Vertex3f(0.105758667f, 0.f, -1.f),
    C_Vertex3f(0.105758667f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.105758667f, 0.f, 0.f),
    C_Vertex3f(0.105758667f, 0.f, -1.f),
    C_Vertex3f(0.00393676758f, 0.f, -1.f),
    C_Vertex3f(0.00393676758f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.251686096f, 0.332267761f, 0.f),
    C_Vertex3f(0.00393676758f, 0.f, 0.f),
    C_Vertex3f(0.105758667f, 0.f, 0.f),
    C_Vertex3f(0.270456314f, 0.225442886f, 0.f),
    C_Vertex3f(0.255493164f, 0.470353127f, 0.f),
    C_Vertex3f(0.133911133f, 0.640625f, 0.f),
    C_Vertex3f(0.0332336426f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.351325989f, 0.336143494f, 0.f),
    C_Vertex3f(0.573760986f, 0.640625f, 0.f),
    C_Vertex3f(0.353309631f, 0.470199585f, 0.f),
    C_Vertex3f(0.345943451f, 0.460727692f, 0.f),
    C_Vertex3f(0.332219124f, 0.442217827f, 0.f),
    C_Vertex3f(0.319502831f, 0.424141884f, 0.f),
    C_Vertex3f(0.311347961f, 0.248046875f, 0.f),
    C_Vertex3f(0.487068176f, 0.f, 0.f),
    C_Vertex3f(0.591339111f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.311347961f, 0.248046875f, 0.f),
    C_Vertex3f(0.319502831f, 0.424141884f, 0.f),
    C_Vertex3f(0.307794571f, 0.406499863f, 0.f),
    C_Vertex3f(0.302444458f, 0.397895813f, 0.f),
    C_Vertex3f(0.300560951f, 0.40137291f, 0.f),
    C_Vertex3f(0.298461914f, 0.267921448f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.296126127f, 0.409052134f, 0.f),
    C_Vertex3f(0.291023493f, 0.417456388f, 0.f),
    C_Vertex3f(0.293230057f, 0.259723663f, 0.f),
    C_Vertex3f(0.298461914f, 0.267921448f, 0.f),
    C_Vertex3f(0.300560951f, 0.40137291f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.263934851f, 0.458323717f, 0.f),
    C_Vertex3f(0.270456314f, 0.225442886f, 0.f),
    C_Vertex3f(0.271708727f, 0.447019339f, 0.f),
    C_Vertex3f(0.276611328f, 0.23438549f, 0.f),
    C_Vertex3f(0.278814793f, 0.436439991f, 0.f),
    C_Vertex3f(0.284202576f, 0.245812416f, 0.f),
    C_Vertex3f(0.285253048f, 0.426585674f, 0.f),
    C_Vertex3f(0.293230057f, 0.259723663f, 0.f),
    C_Vertex3f(0.291023493f, 0.417456388f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.353309631f, 0.470199585f, 0.f),
    C_Vertex3f(0.573760986f, 0.640625f, 0.f),
    C_Vertex3f(0.48160553f, 0.640625f, 0.f),
    C_Vertex3f(0.255493164f, 0.470353127f, 0.f),
    C_Vertex3f(0.270456314f, 0.225442886f, 0.f),
    C_Vertex3f(0.263934851f, 0.458323717f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.251686096f, 0.332267761f, -1.f),
    C_Vertex3f(0.00393676758f, 0.f, -1.f),
    C_Vertex3f(0.105758667f, 0.f, -1.f),
    C_Vertex3f(0.270456314f, 0.225442886f, -1.f),
    C_Vertex3f(0.255493164f, 0.470353127f, -1.f),
    C_Vertex3f(0.133911133f, 0.640625f, -1.f),
    C_Vertex3f(0.0332336426f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.351325989f, 0.336143494f, -1.f),
    C_Vertex3f(0.573760986f, 0.640625f, -1.f),
    C_Vertex3f(0.353309631f, 0.470199585f, -1.f),
    C_Vertex3f(0.345943451f, 0.460727692f, -1.f),
    C_Vertex3f(0.332219124f, 0.442217827f, -1.f),
    C_Vertex3f(0.319502831f, 0.424141884f, -1.f),
    C_Vertex3f(0.311347961f, 0.248046875f, -1.f),
    C_Vertex3f(0.487068176f, 0.f, -1.f),
    C_Vertex3f(0.591339111f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.311347961f, 0.248046875f, -1.f),
    C_Vertex3f(0.319502831f, 0.424141884f, -1.f),
    C_Vertex3f(0.307794571f, 0.406499863f, -1.f),
    C_Vertex3f(0.302444458f, 0.397895813f, -1.f),
    C_Vertex3f(0.300560951f, 0.40137291f, -1.f),
    C_Vertex3f(0.298461914f, 0.267921448f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.296126127f, 0.409052134f, -1.f),
    C_Vertex3f(0.291023493f, 0.417456388f, -1.f),
    C_Vertex3f(0.293230057f, 0.259723663f, -1.f),
    C_Vertex3f(0.298461914f, 0.267921448f, -1.f),
    C_Vertex3f(0.300560951f, 0.40137291f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.263934851f, 0.458323717f, -1.f),
    C_Vertex3f(0.270456314f, 0.225442886f, -1.f),
    C_Vertex3f(0.271708727f, 0.447019339f, -1.f),
    C_Vertex3f(0.276611328f, 0.23438549f, -1.f),
    C_Vertex3f(0.278814793f, 0.436439991f, -1.f),
    C_Vertex3f(0.284202576f, 0.245812416f, -1.f),
    C_Vertex3f(0.285253048f, 0.426585674f, -1.f),
    C_Vertex3f(0.293230057f, 0.259723663f, -1.f),
    C_Vertex3f(0.291023493f, 0.417456388f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.353309631f, 0.470199585f, -1.f),
    C_Vertex3f(0.573760986f, 0.640625f, -1.f),
    C_Vertex3f(0.48160553f, 0.640625f, -1.f),
    C_Vertex3f(0.255493164f, 0.470353127f, -1.f),
    C_Vertex3f(0.270456314f, 0.225442886f, -1.f),
    C_Vertex3f(0.263934851f, 0.458323717f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_89[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.249542236f, 0.f, 0.f),
    C_Vertex3f(0.249542236f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Normal3f(-0.831113875f, -0.556102216f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.00247192383f, 0.640625f, -1.f),
    C_Vertex3f(0.00247192383f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.00247192383f, 0.640625f, 0.f),
    C_Vertex3f(0.00247192383f, 0.640625f, -1.f),
    C_Vertex3f(0.105628967f, 0.640625f, -1.f),
    C_Vertex3f(0.105628967f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.836778283f, 0.547541976f, 0.f),
    C_Vertex3f(0.105628967f, 0.640625f, 0.f),
    C_Vertex3f(0.105628967f, 0.640625f, -1.f),
    C_Normal3f(0.840595007f, 0.541664064f, 0.f),
    C_Vertex3f(0.249668121f, 0.420497894f, 0.f),
    C_Vertex3f(0.249668121f, 0.420497894f, -1.f),
    C_Normal3f(0.862715006f, 0.505690515f, 0.f),
    C_Vertex3f(0.282409668f, 0.366283417f, 0.f),
    C_Vertex3f(0.282409668f, 0.366283417f, -1.f),
    C_Normal3f(0.875902712f, 0.482487768f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.86717093f, 0.498010665f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_Normal3f(-0.86124754f, 0.508185625f, 0.f),
    C_Vertex3f(0.304641724f, 0.351566315f, 0.f),
    C_Vertex3f(0.304641724f, 0.351566315f, -1.f),
    C_Normal3f(-0.854441464f, 0.519547701f, 0.f),
    C_Vertex3f(0.320591927f, 0.37826252f, 0.f),
    C_Vertex3f(0.320591927f, 0.37826252f, -1.f),
    C_Normal3f(-0.847048819f, 0.531515121f, 0.f),
    C_Vertex3f(0.338072777f, 0.406546593f, 0.f),
    C_Vertex3f(0.338072777f, 0.406546593f, -1.f),
    C_Normal3f(-0.835170507f, 0.549991071f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.835170507f, 0.549991071f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, -1.f),
    C_Vertex3f(0.49156189f, 0.640625f, -1.f),
    C_Vertex3f(0.49156189f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.49156189f, 0.640625f, 0.f),
    C_Vertex3f(0.49156189f, 0.640625f, -1.f),
    C_Vertex3f(0.590362549f, 0.640625f, -1.f),
    C_Vertex3f(0.590362549f, 0.640625f, 0.f),
    C_Normal3f(0.821961641f, -0.569542885f, 0.f),
    C_Vertex3f(0.590362549f, 0.640625f, 0.f),
    C_Vertex3f(0.590362549f, 0.640625f, -1.f),
    C_Vertex3f(0.334503174f, 0.271369934f, -1.f),
    C_Vertex3f(0.334503174f, 0.271369934f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.334503174f, 0.271369934f, 0.f),
    C_Vertex3f(0.334503174f, 0.271369934f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_Vertex3f(0.282409668f, 0.366283417f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.249668121f, 0.420497894f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Vertex3f(0.105628967f, 0.640625f, 0.f),
    C_Vertex3f(0.00247192383f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.49156189f, 0.640625f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, 0.f),
    C_Vertex3f(0.590362549f, 0.640625f, 0.f),
    C_Vertex3f(0.338072777f, 0.406546593f, 0.f),
    C_Vertex3f(0.334503174f, 0.271369934f, 0.f),
    C_Vertex3f(0.320591927f, 0.37826252f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.304641724f, 0.351566315f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Vertex3f(0.249542236f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_Vertex3f(0.282409668f, 0.366283417f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.249668121f, 0.420497894f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.105628967f, 0.640625f, -1.f),
    C_Vertex3f(0.00247192383f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.49156189f, 0.640625f, -1.f),
    C_Vertex3f(0.357084274f, 0.436418533f, -1.f),
    C_Vertex3f(0.590362549f, 0.640625f, -1.f),
    C_Vertex3f(0.338072777f, 0.406546593f, -1.f),
    C_Vertex3f(0.334503174f, 0.271369934f, -1.f),
    C_Vertex3f(0.320591927f, 0.37826252f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.304641724f, 0.351566315f, -1.f),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.249542236f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_90[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.017578125f, 0.f, 0.f),
    C_Vertex3f(0.017578125f, 0.f, -1.f),
    C_Vertex3f(0.017578125f, 0.078742981f, -1.f),
    C_Vertex3f(0.017578125f, 0.078742981f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.017578125f, 0.078742981f, 0.f),
    C_Vertex3f(0.017578125f, 0.078742981f, -1.f),
    C_Normal3f(-0.77963239f, 0.626237333f, 0.f),
    C_Vertex3f(0.354995728f, 0.499845505f, 0.f),
    C_Vertex3f(0.354995728f, 0.499845505f, -1.f),
    C_Normal3f(-0.757743835f, 0.652552128f, 0.f),
    C_Vertex3f(0.372172356f, 0.520194054f, 0.f),
    C_Vertex3f(0.372172356f, 0.520194054f, -1.f),
    C_Normal3f(-0.743628263f, 0.668593347f, 0.f),
    C_Vertex3f(0.388871193f, 0.539181709f, 0.f),
    C_Vertex3f(0.388871193f, 0.539181709f, -1.f),
    C_Normal3f(-0.73034364f, 0.683079898f, 0.f),
    C_Vertex3f(0.405092239f, 0.556808472f, 0.f),
    C_Vertex3f(0.405092239f, 0.556808472f, -1.f),
    C_Normal3f(-0.718555987f, 0.695469201f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, -1.f),
    C_Vertex3f(0.0546875f, 0.564941406f, -1.f),
    C_Vertex3f(0.0546875f, 0.564941406f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0546875f, 0.564941406f, 0.f),
    C_Vertex3f(0.0546875f, 0.564941406f, -1.f),
    C_Vertex3f(0.0546875f, 0.640625f, -1.f),
    C_Vertex3f(0.0546875f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0546875f, 0.640625f, 0.f),
    C_Vertex3f(0.0546875f, 0.640625f, -1.f),
    C_Vertex3f(0.514160156f, 0.640625f, -1.f),
    C_Vertex3f(0.514160156f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.514160156f, 0.640625f, 0.f),
    C_Vertex3f(0.514160156f, 0.640625f, -1.f),
    C_Vertex3f(0.514160156f, 0.564941406f, -1.f),
    C_Vertex3f(0.514160156f, 0.564941406f, 0.f),
    C_Normal3f(0.776490152f, -0.630129337f, 0.f),
    C_Vertex3f(0.514160156f, 0.564941406f, 0.f),
    C_Vertex3f(0.514160156f, 0.564941406f, -1.f),
    C_Vertex3f(0.153862f, 0.120956421f, -1.f),
    C_Vertex3f(0.153862f, 0.120956421f, 0.f),
    C_Normal3f(0.759085f, -0.650991499f, 0.f),
    C_Vertex3f(0.153862f, 0.120956421f, 0.f),
    C_Vertex3f(0.153862f, 0.120956421f, -1.f),
    C_Vertex3f(0.115036011f, 0.0756835938f, -1.f),
    C_Vertex3f(0.115036011f, 0.0756835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.115036011f, 0.0756835938f, 0.f),
    C_Vertex3f(0.115036011f, 0.0756835938f, -1.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, -1.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, 0.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, -1.f),
    C_Vertex3f(0.524414063f, 0.f, -1.f),
    C_Vertex3f(0.524414063f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.524414063f, 0.f, 0.f),
    C_Vertex3f(0.524414063f, 0.f, -1.f),
    C_Vertex3f(0.017578125f, 0.f, -1.f),
    C_Vertex3f(0.017578125f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115036011f, 0.0756835938f, 0.f),
    C_Vertex3f(0.153862f, 0.120956421f, 0.f),
    C_Vertex3f(0.354995728f, 0.499845505f, 0.f),
    C_Vertex3f(0.017578125f, 0.078742981f, 0.f),
    C_Vertex3f(0.524414063f, 0.f, 0.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.514160156f, 0.564941406f, 0.f),
    C_Vertex3f(0.514160156f, 0.640625f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, 0.f),
    C_Vertex3f(0.405092239f, 0.556808472f, 0.f),
    C_Vertex3f(0.388871193f, 0.539181709f, 0.f),
    C_Vertex3f(0.372172356f, 0.520194054f, 0.f),
    C_Vertex3f(0.354995728f, 0.499845505f, 0.f),
    C_Vertex3f(0.153862f, 0.120956421f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0546875f, 0.640625f, 0.f),
    C_Vertex3f(0.0546875f, 0.564941406f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, 0.f),
    C_Vertex3f(0.514160156f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.524414063f, 0.f, 0.f),
    C_Vertex3f(0.017578125f, 0.078742981f, 0.f),
    C_Vertex3f(0.017578125f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115036011f, 0.0756835938f, -1.f),
    C_Vertex3f(0.153862f, 0.120956421f, -1.f),
    C_Vertex3f(0.354995728f, 0.499845505f, -1.f),
    C_Vertex3f(0.017578125f, 0.078742981f, -1.f),
    C_Vertex3f(0.524414063f, 0.f, -1.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.514160156f, 0.564941406f, -1.f),
    C_Vertex3f(0.514160156f, 0.640625f, -1.f),
    C_Vertex3f(0.412963867f, 0.564941406f, -1.f),
    C_Vertex3f(0.405092239f, 0.556808472f, -1.f),
    C_Vertex3f(0.388871193f, 0.539181709f, -1.f),
    C_Vertex3f(0.372172356f, 0.520194054f, -1.f),
    C_Vertex3f(0.354995728f, 0.499845505f, -1.f),
    C_Vertex3f(0.153862f, 0.120956421f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0546875f, 0.640625f, -1.f),
    C_Vertex3f(0.0546875f, 0.564941406f, -1.f),
    C_Vertex3f(0.412963867f, 0.564941406f, -1.f),
    C_Vertex3f(0.514160156f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.524414063f, 0.f, -1.f),
    C_Vertex3f(0.017578125f, 0.078742981f, -1.f),
    C_Vertex3f(0.017578125f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.546875f, 0.f, 0.f),
};
static const Call char_91[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.060546875f, -0.177734375f, 0.f),
    C_Vertex3f(0.060546875f, -0.177734375f, -1.f),
    C_Vertex3f(0.060546875f, 0.640625f, -1.f),
    C_Vertex3f(0.060546875f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.060546875f, 0.640625f, 0.f),
    C_Vertex3f(0.060546875f, 0.640625f, -1.f),
    C_Vertex3f(0.233886719f, 0.640625f, -1.f),
    C_Vertex3f(0.233886719f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.233886719f, 0.640625f, 0.f),
    C_Vertex3f(0.233886719f, 0.640625f, -1.f),
    C_Vertex3f(0.233886719f, 0.576171875f, -1.f),
    C_Vertex3f(0.233886719f, 0.576171875f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.233886719f, 0.576171875f, 0.f),
    C_Vertex3f(0.233886719f, 0.576171875f, -1.f),
    C_Vertex3f(0.139160156f, 0.576171875f, -1.f),
    C_Vertex3f(0.139160156f, 0.576171875f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.139160156f, 0.576171875f, 0.f),
    C_Vertex3f(0.139160156f, 0.576171875f, -1.f),
    C_Vertex3f(0.139160156f, -0.11328125f, -1.f),
    C_Vertex3f(0.139160156f, -0.11328125f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.139160156f, -0.11328125f, 0.f),
    C_Vertex3f(0.139160156f, -0.11328125f, -1.f),
    C_Vertex3f(0.233886719f, -0.11328125f, -1.f),
    C_Vertex3f(0.233886719f, -0.11328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.233886719f, -0.11328125f, 0.f),
    C_Vertex3f(0.233886719f, -0.11328125f, -1.f),
    C_Vertex3f(0.233886719f, -0.177734375f, -1.f),
    C_Vertex3f(0.233886719f, -0.177734375f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.233886719f, -0.177734375f, 0.f),
    C_Vertex3f(0.233886719f, -0.177734375f, -1.f),
    C_Vertex3f(0.060546875f, -0.177734375f, -1.f),
    C_Vertex3f(0.060546875f, -0.177734375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.139160156f, -0.11328125f, 0.f),
    C_Vertex3f(0.139160156f, 0.576171875f, 0.f),
    C_Vertex3f(0.060546875f, 0.640625f, 0.f),
    C_Vertex3f(0.060546875f, -0.177734375f, 0.f),
    C_Vertex3f(0.233886719f, -0.177734375f, 0.f),
    C_Vertex3f(0.233886719f, -0.11328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.233886719f, 0.640625f, 0.f),
    C_Vertex3f(0.060546875f, 0.640625f, 0.f),
    C_Vertex3f(0.139160156f, 0.576171875f, 0.f),
    C_Vertex3f(0.233886719f, 0.576171875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.139160156f, -0.11328125f, -1.f),
    C_Vertex3f(0.139160156f, 0.576171875f, -1.f),
    C_Vertex3f(0.060546875f, 0.640625f, -1.f),
    C_Vertex3f(0.060546875f, -0.177734375f, -1.f),
    C_Vertex3f(0.233886719f, -0.177734375f, -1.f),
    C_Vertex3f(0.233886719f, -0.11328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.233886719f, 0.640625f, -1.f),
    C_Vertex3f(0.060546875f, 0.640625f, -1.f),
    C_Vertex3f(0.139160156f, 0.576171875f, -1.f),
    C_Vertex3f(0.233886719f, 0.576171875f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_92[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.963095069f, -0.269161463f, 0.f),
    C_Vertex3f(0.18560791f, -0.0107421875f, 0.f),
    C_Vertex3f(0.18560791f, -0.0107421875f, -1.f),
    C_Vertex3f(0.000564575195f, 0.651367188f, -1.f),
    C_Vertex3f(0.000564575195f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.000564575195f, 0.651367188f, 0.f),
    C_Vertex3f(0.000564575195f, 0.651367188f, -1.f),
    C_Vertex3f(0.0634765625f, 0.651367188f, -1.f),
    C_Vertex3f(0.0634765625f, 0.651367188f, 0.f),
    C_Normal3f(0.963089287f, 0.269182026f, 0.f),
    C_Vertex3f(0.0634765625f, 0.651367188f, 0.f),
    C_Vertex3f(0.0634765625f, 0.651367188f, -1.f),
    C_Vertex3f(0.248535156f, -0.0107421875f, -1.f),
    C_Vertex3f(0.248535156f, -0.0107421875f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.248535156f, -0.0107421875f, 0.f),
    C_Vertex3f(0.248535156f, -0.0107421875f, -1.f),
    C_Vertex3f(0.18560791f, -0.0107421875f, -1.f),
    C_Vertex3f(0.18560791f, -0.0107421875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18560791f, -0.0107421875f, 0.f),
    C_Vertex3f(0.248535156f, -0.0107421875f, 0.f),
    C_Vertex3f(0.0634765625f, 0.651367188f, 0.f),
    C_Vertex3f(0.000564575195f, 0.651367188f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18560791f, -0.0107421875f, -1.f),
    C_Vertex3f(0.248535156f, -0.0107421875f, -1.f),
    C_Vertex3f(0.0634765625f, 0.651367188f, -1.f),
    C_Vertex3f(0.000564575195f, 0.651367188f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_93[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.190429688f, -0.177734375f, 0.f),
    C_Vertex3f(0.190429688f, -0.177734375f, -1.f),
    C_Vertex3f(0.0170898438f, -0.177734375f, -1.f),
    C_Vertex3f(0.0170898438f, -0.177734375f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0170898438f, -0.177734375f, 0.f),
    C_Vertex3f(0.0170898438f, -0.177734375f, -1.f),
    C_Vertex3f(0.0170898438f, -0.11328125f, -1.f),
    C_Vertex3f(0.0170898438f, -0.11328125f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0170898438f, -0.11328125f, 0.f),
    C_Vertex3f(0.0170898438f, -0.11328125f, -1.f),
    C_Vertex3f(0.111816406f, -0.11328125f, -1.f),
    C_Vertex3f(0.111816406f, -0.11328125f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, -0.11328125f, 0.f),
    C_Vertex3f(0.111816406f, -0.11328125f, -1.f),
    C_Vertex3f(0.111816406f, 0.576171875f, -1.f),
    C_Vertex3f(0.111816406f, 0.576171875f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.576171875f, 0.f),
    C_Vertex3f(0.111816406f, 0.576171875f, -1.f),
    C_Vertex3f(0.0170898438f, 0.576171875f, -1.f),
    C_Vertex3f(0.0170898438f, 0.576171875f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0170898438f, 0.576171875f, 0.f),
    C_Vertex3f(0.0170898438f, 0.576171875f, -1.f),
    C_Vertex3f(0.0170898438f, 0.640625f, -1.f),
    C_Vertex3f(0.0170898438f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0170898438f, 0.640625f, 0.f),
    C_Vertex3f(0.0170898438f, 0.640625f, -1.f),
    C_Vertex3f(0.190429688f, 0.640625f, -1.f),
    C_Vertex3f(0.190429688f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.190429688f, 0.640625f, 0.f),
    C_Vertex3f(0.190429688f, 0.640625f, -1.f),
    C_Vertex3f(0.190429688f, -0.177734375f, -1.f),
    C_Vertex3f(0.190429688f, -0.177734375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.111816406f, 0.576171875f, 0.f),
    C_Vertex3f(0.111816406f, -0.11328125f, 0.f),
    C_Vertex3f(0.190429688f, -0.177734375f, 0.f),
    C_Vertex3f(0.190429688f, 0.640625f, 0.f),
    C_Vertex3f(0.0170898438f, 0.640625f, 0.f),
    C_Vertex3f(0.0170898438f, 0.576171875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0170898438f, -0.11328125f, 0.f),
    C_Vertex3f(0.0170898438f, -0.177734375f, 0.f),
    C_Vertex3f(0.190429688f, -0.177734375f, 0.f),
    C_Vertex3f(0.111816406f, -0.11328125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.111816406f, 0.576171875f, -1.f),
    C_Vertex3f(0.111816406f, -0.11328125f, -1.f),
    C_Vertex3f(0.190429688f, -0.177734375f, -1.f),
    C_Vertex3f(0.190429688f, 0.640625f, -1.f),
    C_Vertex3f(0.0170898438f, 0.640625f, -1.f),
    C_Vertex3f(0.0170898438f, 0.576171875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0170898438f, -0.11328125f, -1.f),
    C_Vertex3f(0.0170898438f, -0.177734375f, -1.f),
    C_Vertex3f(0.190429688f, -0.177734375f, -1.f),
    C_Vertex3f(0.111816406f, -0.11328125f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_94[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.104446411f, 0.301269531f, 0.f),
    C_Vertex3f(0.104446411f, 0.301269531f, -1.f),
    C_Vertex3f(0.0233917236f, 0.301269531f, -1.f),
    C_Vertex3f(0.0233917236f, 0.301269531f, 0.f),
    C_Normal3f(-0.915070176f, 0.403294533f, 0.f),
    C_Vertex3f(0.0233917236f, 0.301269531f, 0.f),
    C_Vertex3f(0.0233917236f, 0.301269531f, -1.f),
    C_Vertex3f(0.177688599f, 0.651367188f, -1.f),
    C_Vertex3f(0.177688599f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.177688599f, 0.651367188f, 0.f),
    C_Vertex3f(0.177688599f, 0.651367188f, -1.f),
    C_Vertex3f(0.241165161f, 0.651367188f, -1.f),
    C_Vertex3f(0.241165161f, 0.651367188f, 0.f),
    C_Normal3f(0.914126694f, 0.405428588f, 0.f),
    C_Vertex3f(0.241165161f, 0.651367188f, 0.f),
    C_Vertex3f(0.241165161f, 0.651367188f, -1.f),
    C_Vertex3f(0.396438599f, 0.301269531f, -1.f),
    C_Vertex3f(0.396438599f, 0.301269531f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.396438599f, 0.301269531f, 0.f),
    C_Vertex3f(0.396438599f, 0.301269531f, -1.f),
    C_Vertex3f(0.317337036f, 0.301269531f, -1.f),
    C_Vertex3f(0.317337036f, 0.301269531f, 0.f),
    C_Normal3f(-0.923747301f, -0.383002669f, 0.f),
    C_Vertex3f(0.317337036f, 0.301269531f, 0.f),
    C_Vertex3f(0.317337036f, 0.301269531f, -1.f),
    C_Vertex3f(0.209228516f, 0.562011719f, -1.f),
    C_Vertex3f(0.209228516f, 0.562011719f, 0.f),
    C_Normal3f(0.927880049f, -0.37287876f, 0.f),
    C_Vertex3f(0.209228516f, 0.562011719f, 0.f),
    C_Vertex3f(0.209228516f, 0.562011719f, -1.f),
    C_Vertex3f(0.104446411f, 0.301269531f, -1.f),
    C_Vertex3f(0.104446411f, 0.301269531f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0233917236f, 0.301269531f, 0.f),
    C_Vertex3f(0.104446411f, 0.301269531f, 0.f),
    C_Vertex3f(0.177688599f, 0.651367188f, 0.f),
    C_Vertex3f(0.209228516f, 0.562011719f, 0.f),
    C_Vertex3f(0.241165161f, 0.651367188f, 0.f),
    C_Vertex3f(0.317337036f, 0.301269531f, 0.f),
    C_Vertex3f(0.396438599f, 0.301269531f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0233917236f, 0.301269531f, -1.f),
    C_Vertex3f(0.104446411f, 0.301269531f, -1.f),
    C_Vertex3f(0.177688599f, 0.651367188f, -1.f),
    C_Vertex3f(0.209228516f, 0.562011719f, -1.f),
    C_Vertex3f(0.241165161f, 0.651367188f, -1.f),
    C_Vertex3f(0.317337036f, 0.301269531f, -1.f),
    C_Vertex3f(0.396438599f, 0.301269531f, -1.f),
    C_End(),
    C_Translatef(0.419921875f, 0.f, 0.f),
};
static const Call char_95[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(-0.013671875f, -0.177734375f, 0.f),
    C_Vertex3f(-0.013671875f, -0.177734375f, -1.f),
    C_Vertex3f(-0.013671875f, -0.12109375f, -1.f),
    C_Vertex3f(-0.013671875f, -0.12109375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(-0.013671875f, -0.12109375f, 0.f),
    C_Vertex3f(-0.013671875f, -0.12109375f, -1.f),
    C_Vertex3f(0.5078125f, -0.12109375f, -1.f),
    C_Vertex3f(0.5078125f, -0.12109375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.5078125f, -0.12109375f, 0.f),
    C_Vertex3f(0.5078125f, -0.12109375f, -1.f),
    C_Vertex3f(0.5078125f, -0.177734375f, -1.f),
    C_Vertex3f(0.5078125f, -0.177734375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.5078125f, -0.177734375f, 0.f),
    C_Vertex3f(0.5078125f, -0.177734375f, -1.f),
    C_Vertex3f(-0.013671875f, -0.177734375f, -1.f),
    C_Vertex3f(-0.013671875f, -0.177734375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-0.013671875f, -0.177734375f, 0.f),
    C_Vertex3f(0.5078125f, -0.177734375f, 0.f),
    C_Vertex3f(0.5078125f, -0.12109375f, 0.f),
    C_Vertex3f(-0.013671875f, -0.12109375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-0.013671875f, -0.177734375f, -1.f),
    C_Vertex3f(0.5078125f, -0.177734375f, -1.f),
    C_Vertex3f(0.5078125f, -0.12109375f, -1.f),
    C_Vertex3f(-0.013671875f, -0.12109375f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_96[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.203613281f, 0.521484375f, 0.f),
    C_Vertex3f(0.203613281f, 0.521484375f, -1.f),
    C_Vertex3f(0.140136719f, 0.521484375f, -1.f),
    C_Vertex3f(0.140136719f, 0.521484375f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.140136719f, 0.521484375f, 0.f),
    C_Vertex3f(0.140136719f, 0.521484375f, -1.f),
    C_Vertex3f(0.0390625f, 0.644042969f, -1.f),
    C_Vertex3f(0.0390625f, 0.644042969f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0390625f, 0.644042969f, 0.f),
    C_Vertex3f(0.0390625f, 0.644042969f, -1.f),
    C_Vertex3f(0.14453125f, 0.644042969f, -1.f),
    C_Vertex3f(0.14453125f, 0.644042969f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.14453125f, 0.644042969f, 0.f),
    C_Vertex3f(0.14453125f, 0.644042969f, -1.f),
    C_Vertex3f(0.203613281f, 0.521484375f, -1.f),
    C_Vertex3f(0.203613281f, 0.521484375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.203613281f, 0.521484375f, 0.f),
    C_Vertex3f(0.14453125f, 0.644042969f, 0.f),
    C_Vertex3f(0.0390625f, 0.644042969f, 0.f),
    C_Vertex3f(0.140136719f, 0.521484375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.203613281f, 0.521484375f, -1.f),
    C_Vertex3f(0.14453125f, 0.644042969f, -1.f),
    C_Vertex3f(0.0390625f, 0.644042969f, -1.f),
    C_Vertex3f(0.140136719f, 0.521484375f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_97[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.644008815f, -0.765018106f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Normal3f(0.609832764f, -0.792530119f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Normal3f(0.560195386f, -0.828360498f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Normal3f(0.488698423f, -0.872452796f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Normal3f(0.402021319f, -0.915630281f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Normal3f(0.307879895f, -0.951425254f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Normal3f(0.217794061f, -0.975994766f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Normal3f(0.128345996f, -0.991729438f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Normal3f(0.0436385125f, -0.999047339f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0297879688f, -0.999556243f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0997409225f, -0.995013535f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Normal3f(-0.175596207f, -0.984462261f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Normal3f(-0.2587111f, -0.965954781f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Normal3f(-0.348068923f, -0.937468946f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Normal3f(-0.441677898f, -0.897173643f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Normal3f(-0.536508918f, -0.84389472f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Normal3f(-0.628730416f, -0.777623355f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Normal3f(-0.708205581f, -0.706006348f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Normal3f(-0.777196348f, -0.629258096f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Normal3f(-0.840526164f, -0.541770935f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Normal3f(-0.895173013f, -0.445718706f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Normal3f(-0.93884182f, -0.344348639f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Normal3f(-0.970410466f, -0.241460919f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Normal3f(-0.990046978f, -0.140737399f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Normal3f(-0.999063611f, -0.043263685f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_Normal3f(-0.998176992f, 0.0603547506f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Normal3f(-0.983728349f, 0.179661945f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Normal3f(-0.952656448f, 0.304048836f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Normal3f(-0.90515995f, 0.425071001f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Normal3f(-0.848285139f, 0.529539704f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Normal3f(-0.782067299f, 0.62319386f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Normal3f(-0.701901376f, 0.712274075f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Normal3f(-0.610135138f, 0.792297363f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Normal3f(-0.525150001f, 0.851009607f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Normal3f(-0.451541305f, 0.89225018f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Normal3f(-0.377094716f, 0.926174641f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Normal3f(-0.315602541f, 0.948891461f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Normal3f(-0.262250215f, 0.964999855f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Normal3f(-0.217653468f, 0.976026177f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Normal3f(-0.182780772f, 0.983153641f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Normal3f(-0.152309522f, 0.988332868f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Normal3f(-0.129105836f, 0.991630852f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Normal3f(-0.125778303f, 0.992058396f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Normal3f(-0.140358314f, 0.990100741f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Normal3f(-0.157823548f, 0.987467349f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Normal3f(-0.178095639f, 0.98401314f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Normal3f(-0.20187664f, 0.979411006f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Normal3f(-0.230107605f, 0.973165154f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Normal3f(-0.264068544f, 0.964503944f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Normal3f(-0.297998816f, 0.954566181f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Normal3f(-0.724053323f, 0.689743996f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Normal3f(-0.999676347f, 0.0254404638f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Normal3f(-0.999733686f, 0.0230763834f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Normal3f(-0.999828994f, 0.0184933227f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Normal3f(-0.999987364f, 0.00503883278f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Normal3f(-0.999277949f, -0.0379956067f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Normal3f(-0.995299578f, -0.0968433693f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Normal3f(-0.98471868f, -0.174152732f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Normal3f(-0.964879274f, -0.262693614f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Normal3f(-0.932045639f, -0.362340957f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Normal3f(-0.882021129f, -0.471209705f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Normal3f(-0.811133981f, -0.584860384f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Normal3f(-0.713830113f, -0.700318813f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Normal3f(-0.61406064f, -0.789258838f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Normal3f(-0.521080196f, -0.853507817f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Normal3f(-0.424242347f, -0.905548692f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Normal3f(-0.330907196f, -0.943663299f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Normal3f(-0.243749335f, -0.969838262f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Normal3f(-0.164196789f, -0.986427605f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Normal3f(-0.0926971957f, -0.995694339f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Normal3f(-0.0295525081f, -0.999563217f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Normal3f(0.0223024357f, -0.99975127f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Normal3f(0.0729011372f, -0.997339189f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Normal3f(0.130242616f, -0.991482139f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Normal3f(0.195666403f, -0.980670512f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Normal3f(0.269929171f, -0.962880194f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Normal3f(0.35332042f, -0.93550241f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Normal3f(0.44521296f, -0.895424783f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Normal3f(0.543500543f, -0.839408755f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Normal3f(0.641533256f, -0.767095149f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Normal3f(0.72543782f, -0.688287675f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Normal3f(0.793855846f, -0.608106017f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Normal3f(0.847895145f, -0.530164063f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Normal3f(0.889552832f, -0.45683226f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Normal3f(0.921084702f, -0.389362156f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Normal3f(0.944615781f, -0.328178316f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Normal3f(0.959435046f, -0.281929553f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Normal3f(0.968791068f, -0.247878939f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.131752744f, -0.991282582f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.978250861f, 0.207425416f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Normal3f(-0.964124143f, 0.265451789f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Normal3f(-0.93853426f, 0.345186114f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Normal3f(-0.892804325f, 0.450444669f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Normal3f(-0.82656163f, 0.562846184f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Normal3f(-0.731744051f, 0.68157953f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Normal3f(-0.617528915f, 0.786548138f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Normal3f(-0.501185894f, 0.865339637f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Normal3f(-0.390982777f, 0.920397997f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Normal3f(-0.290884405f, 0.956758201f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Normal3f(-0.197783709f, 0.98024565f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Normal3f(-0.11233186f, 0.993670762f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Normal3f(-0.0353869088f, 0.999373674f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Normal3f(0.0308220927f, 0.999524832f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Normal3f(0.102260895f, 0.994757652f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Normal3f(0.188853115f, 0.982005298f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Normal3f(0.293315113f, 0.956015766f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Normal3f(0.394779444f, 0.918775976f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Normal3f(0.489459455f, 0.872026086f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Normal3f(0.597415328f, 0.801932037f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Normal3f(0.71381259f, 0.700336814f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Normal3f(0.813518882f, 0.581538498f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Normal3f(0.882400215f, 0.470499665f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Normal3f(0.93244344f, 0.361315906f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Normal3f(0.96134758f, 0.275337785f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Normal3f(0.978440464f, 0.206529349f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Normal3f(0.987760723f, 0.155976713f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Normal3f(0.992642164f, 0.121084802f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Normal3f(0.996025085f, 0.0890726f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Normal3f(0.997923374f, 0.0644121915f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Normal3f(0.998993039f, 0.0448651426f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Normal3f(0.999579132f, 0.0290074293f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Normal3f(0.999873698f, 0.015893871f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Normal3f(0.999999583f, 0.000895282777f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Normal3f(0.999999404f, 0.00104705233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Normal3f(0.999950409f, 0.00996590592f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Normal3f(0.999827266f, 0.0185867213f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Normal3f(0.999563813f, 0.0295336749f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Normal3f(0.999036491f, 0.0438889116f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Normal3f(0.997980356f, 0.0635225922f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Normal3f(0.995762765f, 0.0919589549f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Normal3f(0.98978281f, 0.142583564f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Normal3f(0.975923538f, 0.218112797f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Normal3f(0.955474317f, 0.295074373f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Normal3f(0.927006185f, 0.375045985f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Normal3f(0.898835897f, 0.438285291f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Normal3f(0.873886287f, 0.486130446f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.895264685f, -0.445534706f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Normal3f(-0.923641562f, -0.383257538f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Normal3f(-0.95153141f, -0.307551742f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Normal3f(-0.975133479f, -0.221618384f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Normal3f(-0.987897873f, -0.155105904f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Normal3f(-0.994414866f, -0.105542049f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.377323061f, -0.926081717f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Normal3f(0.344282955f, -0.93886584f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Normal3f(0.30806917f, -0.951363921f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Normal3f(0.27083087f, -0.962626934f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Normal3f(0.23911649f, -0.970990896f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Normal3f(0.211867765f, -0.977298319f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Normal3f(0.188257366f, -0.982119739f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Normal3f(0.167636946f, -0.985848784f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Normal3f(0.149529129f, -0.988757372f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Normal3f(0.158242494f, -0.987400293f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Normal3f(0.197629482f, -0.980276704f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Normal3f(0.253822058f, -0.967250884f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Normal3f(0.338508695f, -0.940963209f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Normal3f(0.456908435f, -0.889513731f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Normal3f(0.576411605f, -0.817159474f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Normal3f(0.691899657f, -0.721993685f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Normal3f(0.794279575f, -0.607552409f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Normal3f(0.874713659f, -0.484639972f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Normal3f(0.936157763f, -0.351580352f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Normal3f(0.977780998f, -0.209628955f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Normal3f(0.998205423f, -0.0598839708f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Normal3f(0.99415952f, 0.107920907f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Normal3f(0.951207876f, 0.308550656f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Normal3f(0.858417392f, 0.51295191f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Normal3f(0.730436027f, 0.682981133f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Normal3f(0.568286002f, 0.822831035f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Normal3f(0.378173947f, 0.92573458f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Normal3f(0.204722017f, 0.978820205f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Normal3f(0.0608421154f, 0.998147368f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Normal3f(-0.056054458f, 0.998427689f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Normal3f(-0.176223293f, 0.984350204f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Normal3f(-0.302620173f, 0.953111231f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Normal3f(-0.429032564f, 0.90328896f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Normal3f(-0.549294412f, 0.835628867f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Normal3f(-0.66699791f, 0.74505949f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Normal3f(-0.775314748f, 0.631575108f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Normal3f(-0.863186479f, 0.504885197f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Normal3f(-0.934871018f, 0.354987651f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_Normal3f(-0.976849318f, 0.213928372f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Normal3f(-0.993807793f, 0.111112691f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Normal3f(-0.999024689f, 0.0441556387f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_98[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.131347656f, 0.f, 0.f),
    C_Vertex3f(0.131347656f, 0.f, -1.f),
    C_Vertex3f(0.05859375f, 0.f, -1.f),
    C_Vertex3f(0.05859375f, 0.f, 0.f),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.05859375f, 0.f, 0.f),
    C_Vertex3f(0.05859375f, 0.f, -1.f),
    C_Vertex3f(0.05859375f, 0.640625f, -1.f),
    C_Vertex3f(0.05859375f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.05859375f, 0.640625f, 0.f),
    C_Vertex3f(0.05859375f, 0.640625f, -1.f),
    C_Vertex3f(0.137207031f, 0.640625f, -1.f),
    C_Vertex3f(0.137207031f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137207031f, 0.640625f, 0.f),
    C_Vertex3f(0.137207031f, 0.640625f, -1.f),
    C_Vertex3f(0.137207031f, 0.412330627f, -1.f),
    C_Vertex3f(0.137207031f, 0.412330627f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.780980825f, 0.624555051f, 0.f),
    C_Vertex3f(0.137207031f, 0.412330627f, 0.f),
    C_Vertex3f(0.137207031f, 0.412330627f, -1.f),
    C_Normal3f(-0.738653004f, 0.674085736f, 0.f),
    C_Vertex3f(0.143432617f, 0.420115471f, 0.f),
    C_Vertex3f(0.143432617f, 0.420115471f, -1.f),
    C_Normal3f(-0.677294135f, 0.73571229f, 0.f),
    C_Vertex3f(0.156744003f, 0.433738947f, 0.f),
    C_Vertex3f(0.156744003f, 0.433738947f, -1.f),
    C_Normal3f(-0.591228664f, 0.806504071f, 0.f),
    C_Vertex3f(0.170915604f, 0.445416212f, 0.f),
    C_Vertex3f(0.170915604f, 0.445416212f, -1.f),
    C_Normal3f(-0.492850423f, 0.870114088f, 0.f),
    C_Vertex3f(0.185947418f, 0.455147266f, 0.f),
    C_Vertex3f(0.185947418f, 0.455147266f, -1.f),
    C_Normal3f(-0.385137945f, 0.922859013f, 0.f),
    C_Vertex3f(0.201839447f, 0.46293211f, 0.f),
    C_Vertex3f(0.201839447f, 0.46293211f, -1.f),
    C_Normal3f(-0.272457153f, 0.962167919f, 0.f),
    C_Vertex3f(0.21859169f, 0.468770742f, 0.f),
    C_Vertex3f(0.21859169f, 0.468770742f, -1.f),
    C_Normal3f(-0.15972437f, 0.987161696f, 0.f),
    C_Vertex3f(0.236204147f, 0.472663164f, 0.f),
    C_Vertex3f(0.236204147f, 0.472663164f, -1.f),
    C_Normal3f(-0.0500440225f, 0.998746991f, 0.f),
    C_Vertex3f(0.254676819f, 0.474609375f, 0.f),
    C_Vertex3f(0.254676819f, 0.474609375f, -1.f),
    C_Normal3f(0.0522761419f, 0.998632669f, 0.f),
    C_Vertex3f(0.275045395f, 0.474609375f, 0.f),
    C_Vertex3f(0.275045395f, 0.474609375f, -1.f),
    C_Normal3f(0.155545801f, 0.987828732f, 0.f),
    C_Vertex3f(0.29587698f, 0.472452641f, 0.f),
    C_Vertex3f(0.29587698f, 0.472452641f, -1.f),
    C_Normal3f(0.2606318f, 0.965438247f, 0.f),
    C_Vertex3f(0.316135883f, 0.468139172f, 0.f),
    C_Vertex3f(0.316135883f, 0.468139172f, -1.f),
    C_Normal3f(0.362620354f, 0.93193692f, 0.f),
    C_Vertex3f(0.335822105f, 0.461668968f, 0.f),
    C_Vertex3f(0.335822105f, 0.461668968f, -1.f),
    C_Normal3f(0.461413294f, 0.887185395f, 0.f),
    C_Vertex3f(0.354935646f, 0.45304203f, 0.f),
    C_Vertex3f(0.354935646f, 0.45304203f, -1.f),
    C_Normal3f(0.561526477f, 0.827458799f, 0.f),
    C_Vertex3f(0.372356415f, 0.442667961f, 0.f),
    C_Vertex3f(0.372356415f, 0.442667961f, -1.f),
    C_Normal3f(0.657740474f, 0.753244638f, 0.f),
    C_Vertex3f(0.388084412f, 0.430546761f, 0.f),
    C_Vertex3f(0.388084412f, 0.430546761f, -1.f),
    C_Normal3f(0.745273471f, 0.666758955f, 0.f),
    C_Vertex3f(0.402119637f, 0.416678429f, 0.f),
    C_Vertex3f(0.402119637f, 0.416678429f, -1.f),
    C_Normal3f(0.815043449f, 0.579399765f, 0.f),
    C_Vertex3f(0.41446209f, 0.401062965f, 0.f),
    C_Vertex3f(0.41446209f, 0.401062965f, -1.f),
    C_Normal3f(0.867594361f, 0.497272551f, 0.f),
    C_Vertex3f(0.425466537f, 0.383836269f, 0.f),
    C_Vertex3f(0.425466537f, 0.383836269f, -1.f),
    C_Normal3f(0.909167409f, 0.41643092f, 0.f),
    C_Vertex3f(0.43513298f, 0.364998341f, 0.f),
    C_Vertex3f(0.43513298f, 0.364998341f, -1.f),
    C_Normal3f(0.940778613f, 0.339021772f, 0.f),
    C_Vertex3f(0.443461418f, 0.344549179f, 0.f),
    C_Vertex3f(0.443461418f, 0.344549179f, -1.f),
    C_Normal3f(0.96480757f, 0.262956947f, 0.f),
    C_Vertex3f(0.450451851f, 0.322488785f, 0.f),
    C_Vertex3f(0.450451851f, 0.322488785f, -1.f),
    C_Normal3f(0.982735038f, 0.185018793f, 0.f),
    C_Vertex3f(0.455694675f, 0.299664497f, 0.f),
    C_Vertex3f(0.455694675f, 0.299664497f, -1.f),
    C_Normal3f(0.994073093f, 0.108713463f, 0.f),
    C_Vertex3f(0.459189892f, 0.276076317f, 0.f),
    C_Vertex3f(0.459189892f, 0.276076317f, -1.f),
    C_Normal3f(0.99943167f, 0.0337080322f, 0.f),
    C_Vertex3f(0.4609375f, 0.251724243f, 0.f),
    C_Vertex3f(0.4609375f, 0.251724243f, -1.f),
    C_Normal3f(0.999450088f, -0.0331580974f, 0.f),
    C_Vertex3f(0.4609375f, 0.22426033f, 0.f),
    C_Vertex3f(0.4609375f, 0.22426033f, -1.f),
    C_Normal3f(0.994896233f, -0.100903802f, 0.f),
    C_Vertex3f(0.459093451f, 0.196141005f, 0.f),
    C_Vertex3f(0.459093451f, 0.196141005f, -1.f),
    C_Normal3f(0.984145463f, -0.177362666f, 0.f),
    C_Vertex3f(0.455405354f, 0.169714212f, 0.f),
    C_Vertex3f(0.455405354f, 0.169714212f, -1.f),
    C_Normal3f(0.965384185f, -0.260832042f, 0.f),
    C_Vertex3f(0.449873209f, 0.144979954f, 0.f),
    C_Vertex3f(0.449873209f, 0.144979954f, -1.f),
    C_Normal3f(0.936676621f, -0.350195378f, 0.f),
    C_Vertex3f(0.442497015f, 0.121938229f, 0.f),
    C_Vertex3f(0.442497015f, 0.121938229f, -1.f),
    C_Normal3f(0.896328568f, -0.443390548f, 0.f),
    C_Vertex3f(0.433276772f, 0.100589037f, 0.f),
    C_Vertex3f(0.433276772f, 0.100589037f, -1.f),
    C_Normal3f(0.843335092f, -0.537388086f, 0.f),
    C_Vertex3f(0.422212481f, 0.0809323788f, 0.f),
    C_Vertex3f(0.422212481f, 0.0809323788f, -1.f),
    C_Normal3f(0.777843416f, -0.628458142f, 0.f),
    C_Vertex3f(0.409304142f, 0.0629682541f, 0.f),
    C_Vertex3f(0.409304142f, 0.0629682541f, -1.f),
    C_Normal3f(0.710203826f, -0.703996181f, 0.f),
    C_Vertex3f(0.394551754f, 0.0466966629f, 0.f),
    C_Vertex3f(0.394551754f, 0.0466966629f, -1.f),
    C_Normal3f(0.640320957f, -0.768107474f, 0.f),
    C_Vertex3f(0.37906158f, 0.0324590206f, 0.f),
    C_Vertex3f(0.37906158f, 0.0324590206f, -1.f),
    C_Normal3f(0.558918118f, -0.829222798f, 0.f),
    C_Vertex3f(0.362833619f, 0.0202553272f, 0.f),
    C_Vertex3f(0.362833619f, 0.0202553272f, -1.f),
    C_Normal3f(0.466915488f, -0.884301901f, 0.f),
    C_Vertex3f(0.345867872f, 0.0100855827f, 0.f),
    C_Vertex3f(0.345867872f, 0.0100855827f, -1.f),
    C_Normal3f(0.366496891f, -0.930419266f, 0.f),
    C_Vertex3f(0.328164339f, 0.00194978714f, 0.f),
    C_Vertex3f(0.328164339f, 0.00194978714f, -1.f),
    C_Normal3f(0.260958344f, -0.965350151f, 0.f),
    C_Vertex3f(0.30972302f, -0.00415205956f, 0.f),
    C_Vertex3f(0.30972302f, -0.00415205956f, -1.f),
    C_Normal3f(0.154206589f, -0.988038659f, 0.f),
    C_Vertex3f(0.290543914f, -0.00821995735f, 0.f),
    C_Vertex3f(0.290543914f, -0.00821995735f, -1.f),
    C_Normal3f(0.0501376688f, -0.998742282f, 0.f),
    C_Vertex3f(0.270627022f, -0.0102539063f, 0.f),
    C_Vertex3f(0.270627022f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0532922745f, -0.998578966f, 0.f),
    C_Vertex3f(0.250027657f, -0.0102539063f, 0.f),
    C_Vertex3f(0.250027657f, -0.0102539063f, -1.f),
    C_Normal3f(-0.167324975f, -0.985901773f, 0.f),
    C_Vertex3f(0.230589867f, -0.00811719894f, 0.f),
    C_Vertex3f(0.230589867f, -0.00811719894f, -1.f),
    C_Normal3f(-0.287755549f, -0.957703888f, 0.f),
    C_Vertex3f(0.212258339f, -0.00384378433f, 0.f),
    C_Vertex3f(0.212258339f, -0.00384378433f, -1.f),
    C_Normal3f(-0.40927279f, -0.912412047f, 0.f),
    C_Vertex3f(0.195033073f, 0.00256633759f, 0.f),
    C_Vertex3f(0.195033073f, 0.00256633759f, -1.f),
    C_Normal3f(-0.525531173f, -0.850774407f, 0.f),
    C_Vertex3f(0.17891407f, 0.0111131668f, 0.f),
    C_Vertex3f(0.17891407f, 0.0111131668f, -1.f),
    C_Normal3f(-0.630704403f, -0.776023209f, 0.f),
    C_Vertex3f(0.163901329f, 0.0217967033f, 0.f),
    C_Vertex3f(0.163901329f, 0.0217967033f, -1.f),
    C_Normal3f(-0.720860898f, -0.69307977f, 0.f),
    C_Vertex3f(0.14999485f, 0.0346169472f, 0.f),
    C_Vertex3f(0.14999485f, 0.0346169472f, -1.f),
    C_Normal3f(-0.783396006f, -0.621522903f, 0.f),
    C_Vertex3f(0.137194633f, 0.0495738983f, 0.f),
    C_Vertex3f(0.137194633f, 0.0495738983f, -1.f),
    C_Normal3f(-0.825345695f, -0.564627826f, 0.f),
    C_Vertex3f(0.131347656f, 0.0581207275f, 0.f),
    C_Vertex3f(0.131347656f, 0.0581207275f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.131347656f, 0.0581207275f, 0.f),
    C_Vertex3f(0.131347656f, 0.0581207275f, -1.f),
    C_Vertex3f(0.131347656f, 0.f, -1.f),
    C_Vertex3f(0.131347656f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999591529f, -0.0285795201f, 0.f),
    C_Vertex3f(0.130371094f, 0.246283531f, 0.f),
    C_Vertex3f(0.130371094f, 0.246283531f, -1.f),
    C_Normal3f(0.999846697f, 0.0175086241f, 0.f),
    C_Vertex3f(0.130371094f, 0.2252388f, 0.f),
    C_Vertex3f(0.130371094f, 0.2252388f, -1.f),
    C_Normal3f(0.998389065f, 0.0567390844f, 0.f),
    C_Vertex3f(0.131079912f, 0.205805779f, 0.f),
    C_Vertex3f(0.131079912f, 0.205805779f, -1.f),
    C_Normal3f(0.994771004f, 0.102130502f, 0.f),
    C_Vertex3f(0.132497549f, 0.187821388f, 0.f),
    C_Vertex3f(0.132497549f, 0.187821388f, -1.f),
    C_Normal3f(0.98791337f, 0.155006677f, 0.f),
    C_Vertex3f(0.134624004f, 0.171285629f, 0.f),
    C_Vertex3f(0.134624004f, 0.171285629f, -1.f),
    C_Normal3f(0.976216555f, 0.216797471f, 0.f),
    C_Vertex3f(0.137459278f, 0.156198502f, 0.f),
    C_Vertex3f(0.137459278f, 0.156198502f, -1.f),
    C_Normal3f(0.957330167f, 0.288996488f, 0.f),
    C_Vertex3f(0.14100337f, 0.142560005f, 0.f),
    C_Vertex3f(0.14100337f, 0.142560005f, -1.f),
    C_Normal3f(0.927886724f, 0.372862279f, 0.f),
    C_Vertex3f(0.145256281f, 0.13037014f, 0.f),
    C_Vertex3f(0.145256281f, 0.13037014f, -1.f),
    C_Normal3f(0.879589856f, 0.475732863f, 0.f),
    C_Vertex3f(0.15021801f, 0.119628906f, 0.f),
    C_Vertex3f(0.15021801f, 0.119628906f, -1.f),
    C_Normal3f(0.824065089f, 0.56649524f, 0.f),
    C_Vertex3f(0.157688141f, 0.107384682f, 0.f),
    C_Vertex3f(0.157688141f, 0.107384682f, -1.f),
    C_Normal3f(0.76188612f, 0.647710979f, 0.f),
    C_Vertex3f(0.167775631f, 0.094088316f, 0.f),
    C_Vertex3f(0.167775631f, 0.094088316f, -1.f),
    C_Normal3f(0.678396225f, 0.734696388f, 0.f),
    C_Vertex3f(0.178680897f, 0.082691431f, 0.f),
    C_Vertex3f(0.178680897f, 0.082691431f, -1.f),
    C_Normal3f(0.575961113f, 0.817477167f, 0.f),
    C_Vertex3f(0.190403938f, 0.0731940269f, 0.f),
    C_Vertex3f(0.190403938f, 0.0731940269f, -1.f),
    C_Normal3f(0.456714123f, 0.889613509f, 0.f),
    C_Vertex3f(0.202944756f, 0.0655961037f, 0.f),
    C_Vertex3f(0.202944756f, 0.0655961037f, -1.f),
    C_Normal3f(0.32607016f, 0.94534564f, 0.f),
    C_Vertex3f(0.216303349f, 0.0598976612f, 0.f),
    C_Vertex3f(0.216303349f, 0.0598976612f, -1.f),
    C_Normal3f(0.191725403f, 0.981448591f, 0.f),
    C_Vertex3f(0.230479717f, 0.0560986996f, 0.f),
    C_Vertex3f(0.230479717f, 0.0560986996f, -1.f),
    C_Normal3f(0.0646170303f, 0.997910082f, 0.f),
    C_Vertex3f(0.245473862f, 0.0541992188f, 0.f),
    C_Vertex3f(0.245473862f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0522465669f, 0.998634219f, 0.f),
    C_Vertex3f(0.259814262f, 0.0541992188f, 0.f),
    C_Vertex3f(0.259814262f, 0.0541992188f, -1.f),
    C_Normal3f(-0.169525221f, 0.985525846f, 0.f),
    C_Vertex3f(0.272246838f, 0.0555999279f, 0.f),
    C_Vertex3f(0.272246838f, 0.0555999279f, -1.f),
    C_Normal3f(-0.284983039f, 0.958532512f, 0.f),
    C_Vertex3f(0.284243107f, 0.0584013462f, 0.f),
    C_Vertex3f(0.284243107f, 0.0584013462f, -1.f),
    C_Normal3f(-0.396769017f, 0.917918503f, 0.f),
    C_Vertex3f(0.29580307f, 0.0626034737f, 0.f),
    C_Vertex3f(0.29580307f, 0.0626034737f, -1.f),
    C_Normal3f(-0.500410497f, 0.865788221f, 0.f),
    C_Vertex3f(0.306926727f, 0.0682063103f, 0.f),
    C_Vertex3f(0.306926727f, 0.0682063103f, -1.f),
    C_Normal3f(-0.592688024f, 0.805432081f, 0.f),
    C_Vertex3f(0.317614079f, 0.075209856f, 0.f),
    C_Vertex3f(0.317614079f, 0.075209856f, -1.f),
    C_Normal3f(-0.672017813f, 0.740535021f, 0.f),
    C_Vertex3f(0.327865124f, 0.0836141109f, 0.f),
    C_Vertex3f(0.327865124f, 0.0836141109f, -1.f),
    C_Normal3f(-0.738318443f, 0.674452305f, 0.f),
    C_Vertex3f(0.337679863f, 0.093419075f, 0.f),
    C_Vertex3f(0.337679863f, 0.093419075f, -1.f),
    C_Normal3f(-0.804101646f, 0.594491959f, 0.f),
    C_Vertex3f(0.347058296f, 0.104624748f, 0.f),
    C_Vertex3f(0.347058296f, 0.104624748f, -1.f),
    C_Normal3f(-0.867125094f, 0.498090416f, 0.f),
    C_Vertex3f(0.355264425f, 0.117203712f, 0.f),
    C_Vertex3f(0.355264425f, 0.117203712f, -1.f),
    C_Normal3f(-0.915163815f, 0.403081983f, 0.f),
    C_Vertex3f(0.36229825f, 0.131155968f, 0.f),
    C_Vertex3f(0.36229825f, 0.131155968f, -1.f),
    C_Normal3f(-0.949781299f, 0.312914461f, 0.f),
    C_Vertex3f(0.368159771f, 0.146481514f, 0.f),
    C_Vertex3f(0.368159771f, 0.146481514f, -1.f),
    C_Normal3f(-0.973262668f, 0.229695067f, 0.f),
    C_Vertex3f(0.372848988f, 0.163180351f, 0.f),
    C_Vertex3f(0.372848988f, 0.163180351f, -1.f),
    C_Normal3f(-0.988014519f, 0.154361561f, 0.f),
    C_Vertex3f(0.3763659f, 0.18125248f, 0.f),
    C_Vertex3f(0.3763659f, 0.18125248f, -1.f),
    C_Normal3f(-0.996206999f, 0.0870147422f, 0.f),
    C_Vertex3f(0.378710508f, 0.200697899f, 0.f),
    C_Vertex3f(0.378710508f, 0.200697899f, -1.f),
    C_Normal3f(-0.999633431f, 0.0270738173f, 0.f),
    C_Vertex3f(0.379882813f, 0.221516609f, 0.f),
    C_Vertex3f(0.379882813f, 0.221516609f, -1.f),
    C_Normal3f(-0.999669254f, -0.0257173311f, 0.f),
    C_Vertex3f(0.379882813f, 0.243982315f, 0.f),
    C_Vertex3f(0.379882813f, 0.243982315f, -1.f),
    C_Normal3f(-0.996632993f, -0.0819917098f, 0.f),
    C_Vertex3f(0.378758192f, 0.265232205f, 0.f),
    C_Vertex3f(0.378758192f, 0.265232205f, -1.f),
    C_Normal3f(-0.989245415f, -0.146264777f, 0.f),
    C_Vertex3f(0.376508951f, 0.284992576f, 0.f),
    C_Vertex3f(0.376508951f, 0.284992576f, -1.f),
    C_Normal3f(-0.975695431f, -0.21913071f, 0.f),
    C_Vertex3f(0.37313509f, 0.303263426f, 0.f),
    C_Vertex3f(0.37313509f, 0.303263426f, -1.f),
    C_Normal3f(-0.953640461f, -0.300948143f, 0.f),
    C_Vertex3f(0.368636608f, 0.320044756f, 0.f),
    C_Vertex3f(0.368636608f, 0.320044756f, -1.f),
    C_Normal3f(-0.920264482f, -0.391296893f, 0.f),
    C_Vertex3f(0.363013506f, 0.335336566f, 0.f),
    C_Vertex3f(0.363013506f, 0.335336566f, -1.f),
    C_Normal3f(-0.87256819f, -0.48849225f, 0.f),
    C_Vertex3f(0.356265783f, 0.349138856f, 0.f),
    C_Vertex3f(0.356265783f, 0.349138856f, -1.f),
    C_Normal3f(-0.80801928f, -0.589156032f, 0.f),
    C_Vertex3f(0.34839344f, 0.361451626f, 0.f),
    C_Vertex3f(0.34839344f, 0.361451626f, -1.f),
    C_Normal3f(-0.739661217f, -0.672979414f, 0.f),
    C_Vertex3f(0.339396477f, 0.372274876f, 0.f),
    C_Vertex3f(0.339396477f, 0.372274876f, -1.f),
    C_Normal3f(-0.671572506f, -0.740938842f, 0.f),
    C_Vertex3f(0.329929352f, 0.381745219f, 0.f),
    C_Vertex3f(0.329929352f, 0.381745219f, -1.f),
    C_Normal3f(-0.590395689f, -0.807113886f, 0.f),
    C_Vertex3f(0.319992065f, 0.389862657f, 0.f),
    C_Vertex3f(0.319992065f, 0.389862657f, -1.f),
    C_Normal3f(-0.496546686f, -0.868009984f, 0.f),
    C_Vertex3f(0.309584618f, 0.396627188f, 0.f),
    C_Vertex3f(0.309584618f, 0.396627188f, -1.f),
    C_Normal3f(-0.39200142f, -0.919964612f, 0.f),
    C_Vertex3f(0.298707008f, 0.402038813f, 0.f),
    C_Vertex3f(0.298707008f, 0.402038813f, -1.f),
    C_Normal3f(-0.280300438f, -0.9599123f, 0.f),
    C_Vertex3f(0.287359238f, 0.406097531f, 0.f),
    C_Vertex3f(0.287359238f, 0.406097531f, -1.f),
    C_Normal3f(-0.166032538f, -0.986120224f, 0.f),
    C_Vertex3f(0.275541306f, 0.408803344f, 0.f),
    C_Vertex3f(0.275541306f, 0.408803344f, -1.f),
    C_Normal3f(-0.0538189635f, -0.998550653f, 0.f),
    C_Vertex3f(0.263253212f, 0.41015625f, 0.f),
    C_Vertex3f(0.263253212f, 0.41015625f, -1.f),
    C_Normal3f(0.055396907f, -0.998464465f, 0.f),
    C_Vertex3f(0.250439644f, 0.41015625f, 0.f),
    C_Vertex3f(0.250439644f, 0.41015625f, -1.f),
    C_Normal3f(0.169525221f, -0.985525846f, 0.f),
    C_Vertex3f(0.238007069f, 0.408755541f, 0.f),
    C_Vertex3f(0.238007069f, 0.408755541f, -1.f),
    C_Normal3f(0.284983039f, -0.958532512f, 0.f),
    C_Vertex3f(0.226010799f, 0.405954123f, 0.f),
    C_Vertex3f(0.226010799f, 0.405954123f, -1.f),
    C_Normal3f(0.396769017f, -0.917918503f, 0.f),
    C_Vertex3f(0.214450836f, 0.401751995f, 0.f),
    C_Vertex3f(0.214450836f, 0.401751995f, -1.f),
    C_Normal3f(0.500410497f, -0.865788221f, 0.f),
    C_Vertex3f(0.203327179f, 0.396149158f, 0.f),
    C_Vertex3f(0.203327179f, 0.396149158f, -1.f),
    C_Normal3f(0.592688024f, -0.805432081f, 0.f),
    C_Vertex3f(0.192639828f, 0.389145613f, 0.f),
    C_Vertex3f(0.192639828f, 0.389145613f, -1.f),
    C_Normal3f(0.672017813f, -0.740535021f, 0.f),
    C_Vertex3f(0.182388783f, 0.380741358f, 0.f),
    C_Vertex3f(0.182388783f, 0.380741358f, -1.f),
    C_Normal3f(0.738318443f, -0.674452305f, 0.f),
    C_Vertex3f(0.172574043f, 0.370936394f, 0.f),
    C_Vertex3f(0.172574043f, 0.370936394f, -1.f),
    C_Normal3f(0.802624285f, -0.596484959f, 0.f),
    C_Vertex3f(0.16319561f, 0.359730721f, 0.f),
    C_Vertex3f(0.16319561f, 0.359730721f, -1.f),
    C_Normal3f(0.864084899f, -0.503346086f, 0.f),
    C_Vertex3f(0.154989481f, 0.34727478f, 0.f),
    C_Vertex3f(0.154989481f, 0.34727478f, -1.f),
    C_Normal3f(0.911955595f, -0.41028899f, 0.f),
    C_Vertex3f(0.147955656f, 0.333568573f, 0.f),
    C_Vertex3f(0.147955656f, 0.333568573f, -1.f),
    C_Normal3f(0.947186589f, -0.320683122f, 0.f),
    C_Vertex3f(0.142094135f, 0.318612099f, 0.f),
    C_Vertex3f(0.142094135f, 0.318612099f, -1.f),
    C_Normal3f(0.971550524f, -0.23683247f, 0.f),
    C_Vertex3f(0.137404919f, 0.302405357f, 0.f),
    C_Vertex3f(0.137404919f, 0.302405357f, -1.f),
    C_Normal3f(0.987118244f, -0.159992501f, 0.f),
    C_Vertex3f(0.133888006f, 0.284948349f, 0.f),
    C_Vertex3f(0.133888006f, 0.284948349f, -1.f),
    C_Normal3f(0.99588877f, -0.0905850306f, 0.f),
    C_Vertex3f(0.131543398f, 0.266241074f, 0.f),
    C_Vertex3f(0.131543398f, 0.266241074f, -1.f),
    C_Normal3f(0.999591529f, -0.0285795201f, 0.f),
    C_Vertex3f(0.130371094f, 0.246283531f, 0.f),
    C_Vertex3f(0.130371094f, 0.246283531f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.130371094f, 0.246283531f, 0.f),
    C_Vertex3f(0.131543398f, 0.266241074f, 0.f),
    C_Vertex3f(0.137207031f, 0.640625f, 0.f),
    C_Vertex3f(0.133888006f, 0.284948349f, 0.f),
    C_Vertex3f(0.137207031f, 0.412330627f, 0.f),
    C_Vertex3f(0.137404919f, 0.302405357f, 0.f),
    C_Vertex3f(0.143432617f, 0.420115471f, 0.f),
    C_Vertex3f(0.142094135f, 0.318612099f, 0.f),
    C_Vertex3f(0.147955656f, 0.333568573f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.137459278f, 0.156198502f, 0.f),
    C_Vertex3f(0.134624004f, 0.171285629f, 0.f),
    C_Vertex3f(0.137194633f, 0.0495738983f, 0.f),
    C_Vertex3f(0.132497549f, 0.187821388f, 0.f),
    C_Vertex3f(0.131347656f, 0.0581207275f, 0.f),
    C_Vertex3f(0.131079912f, 0.205805779f, 0.f),
    C_Vertex3f(0.131347656f, 0.f, 0.f),
    C_Vertex3f(0.130371094f, 0.2252388f, 0.f),
    C_Vertex3f(0.05859375f, 0.640625f, 0.f),
    C_Vertex3f(0.130371094f, 0.246283531f, 0.f),
    C_Vertex3f(0.137207031f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.378758192f, 0.265232205f, 0.f),
    C_Vertex3f(0.379882813f, 0.243982315f, 0.f),
    C_Vertex3f(0.388084412f, 0.430546761f, 0.f),
    C_Vertex3f(0.394551754f, 0.0466966629f, 0.f),
    C_Vertex3f(0.402119637f, 0.416678429f, 0.f),
    C_Vertex3f(0.409304142f, 0.0629682541f, 0.f),
    C_Vertex3f(0.41446209f, 0.401062965f, 0.f),
    C_Vertex3f(0.422212481f, 0.0809323788f, 0.f),
    C_Vertex3f(0.425466537f, 0.383836269f, 0.f),
    C_Vertex3f(0.433276772f, 0.100589037f, 0.f),
    C_Vertex3f(0.43513298f, 0.364998341f, 0.f),
    C_Vertex3f(0.442497015f, 0.121938229f, 0.f),
    C_Vertex3f(0.443461418f, 0.344549179f, 0.f),
    C_Vertex3f(0.449873209f, 0.144979954f, 0.f),
    C_Vertex3f(0.450451851f, 0.322488785f, 0.f),
    C_Vertex3f(0.455405354f, 0.169714212f, 0.f),
    C_Vertex3f(0.455694675f, 0.299664497f, 0.f),
    C_Vertex3f(0.459093451f, 0.196141005f, 0.f),
    C_Vertex3f(0.459189892f, 0.276076317f, 0.f),
    C_Vertex3f(0.4609375f, 0.22426033f, 0.f),
    C_Vertex3f(0.4609375f, 0.251724243f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.363013506f, 0.335336566f, 0.f),
    C_Vertex3f(0.368636608f, 0.320044756f, 0.f),
    C_Vertex3f(0.372356415f, 0.442667961f, 0.f),
    C_Vertex3f(0.37313509f, 0.303263426f, 0.f),
    C_Vertex3f(0.388084412f, 0.430546761f, 0.f),
    C_Vertex3f(0.376508951f, 0.284992576f, 0.f),
    C_Vertex3f(0.378758192f, 0.265232205f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339396477f, 0.372274876f, 0.f),
    C_Vertex3f(0.34839344f, 0.361451626f, 0.f),
    C_Vertex3f(0.354935646f, 0.45304203f, 0.f),
    C_Vertex3f(0.356265783f, 0.349138856f, 0.f),
    C_Vertex3f(0.372356415f, 0.442667961f, 0.f),
    C_Vertex3f(0.363013506f, 0.335336566f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.335822105f, 0.461668968f, 0.f),
    C_Vertex3f(0.316135883f, 0.468139172f, 0.f),
    C_Vertex3f(0.319992065f, 0.389862657f, 0.f),
    C_Vertex3f(0.329929352f, 0.381745219f, 0.f),
    C_Vertex3f(0.339396477f, 0.372274876f, 0.f),
    C_Vertex3f(0.354935646f, 0.45304203f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.275541306f, 0.408803344f, 0.f),
    C_Vertex3f(0.287359238f, 0.406097531f, 0.f),
    C_Vertex3f(0.29587698f, 0.472452641f, 0.f),
    C_Vertex3f(0.298707008f, 0.402038813f, 0.f),
    C_Vertex3f(0.316135883f, 0.468139172f, 0.f),
    C_Vertex3f(0.309584618f, 0.396627188f, 0.f),
    C_Vertex3f(0.319992065f, 0.389862657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.238007069f, 0.408755541f, 0.f),
    C_Vertex3f(0.250439644f, 0.41015625f, 0.f),
    C_Vertex3f(0.254676819f, 0.474609375f, 0.f),
    C_Vertex3f(0.263253212f, 0.41015625f, 0.f),
    C_Vertex3f(0.275045395f, 0.474609375f, 0.f),
    C_Vertex3f(0.275541306f, 0.408803344f, 0.f),
    C_Vertex3f(0.29587698f, 0.472452641f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.203327179f, 0.396149158f, 0.f),
    C_Vertex3f(0.214450836f, 0.401751995f, 0.f),
    C_Vertex3f(0.21859169f, 0.468770742f, 0.f),
    C_Vertex3f(0.226010799f, 0.405954123f, 0.f),
    C_Vertex3f(0.236204147f, 0.472663164f, 0.f),
    C_Vertex3f(0.238007069f, 0.408755541f, 0.f),
    C_Vertex3f(0.254676819f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172574043f, 0.370936394f, 0.f),
    C_Vertex3f(0.182388783f, 0.380741358f, 0.f),
    C_Vertex3f(0.185947418f, 0.455147266f, 0.f),
    C_Vertex3f(0.192639828f, 0.389145613f, 0.f),
    C_Vertex3f(0.201839447f, 0.46293211f, 0.f),
    C_Vertex3f(0.203327179f, 0.396149158f, 0.f),
    C_Vertex3f(0.21859169f, 0.468770742f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147955656f, 0.333568573f, 0.f),
    C_Vertex3f(0.154989481f, 0.34727478f, 0.f),
    C_Vertex3f(0.156744003f, 0.433738947f, 0.f),
    C_Vertex3f(0.16319561f, 0.359730721f, 0.f),
    C_Vertex3f(0.170915604f, 0.445416212f, 0.f),
    C_Vertex3f(0.172574043f, 0.370936394f, 0.f),
    C_Vertex3f(0.185947418f, 0.455147266f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.379882813f, 0.221516609f, 0.f),
    C_Vertex3f(0.378710508f, 0.200697899f, 0.f),
    C_Vertex3f(0.37906158f, 0.0324590206f, 0.f),
    C_Vertex3f(0.394551754f, 0.0466966629f, 0.f),
    C_Vertex3f(0.379882813f, 0.243982315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37906158f, 0.0324590206f, 0.f),
    C_Vertex3f(0.378710508f, 0.200697899f, 0.f),
    C_Vertex3f(0.3763659f, 0.18125248f, 0.f),
    C_Vertex3f(0.372848988f, 0.163180351f, 0.f),
    C_Vertex3f(0.368159771f, 0.146481514f, 0.f),
    C_Vertex3f(0.362833619f, 0.0202553272f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.362833619f, 0.0202553272f, 0.f),
    C_Vertex3f(0.368159771f, 0.146481514f, 0.f),
    C_Vertex3f(0.36229825f, 0.131155968f, 0.f),
    C_Vertex3f(0.355264425f, 0.117203712f, 0.f),
    C_Vertex3f(0.347058296f, 0.104624748f, 0.f),
    C_Vertex3f(0.345867872f, 0.0100855827f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.337679863f, 0.093419075f, 0.f),
    C_Vertex3f(0.327865124f, 0.0836141109f, 0.f),
    C_Vertex3f(0.328164339f, 0.00194978714f, 0.f),
    C_Vertex3f(0.345867872f, 0.0100855827f, 0.f),
    C_Vertex3f(0.347058296f, 0.104624748f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.317614079f, 0.075209856f, 0.f),
    C_Vertex3f(0.306926727f, 0.0682063103f, 0.f),
    C_Vertex3f(0.30972302f, -0.00415205956f, 0.f),
    C_Vertex3f(0.328164339f, 0.00194978714f, 0.f),
    C_Vertex3f(0.327865124f, 0.0836141109f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.29580307f, 0.0626034737f, 0.f),
    C_Vertex3f(0.284243107f, 0.0584013462f, 0.f),
    C_Vertex3f(0.290543914f, -0.00821995735f, 0.f),
    C_Vertex3f(0.30972302f, -0.00415205956f, 0.f),
    C_Vertex3f(0.306926727f, 0.0682063103f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230479717f, 0.0560986996f, 0.f),
    C_Vertex3f(0.230589867f, -0.00811719894f, 0.f),
    C_Vertex3f(0.245473862f, 0.0541992188f, 0.f),
    C_Vertex3f(0.250027657f, -0.0102539063f, 0.f),
    C_Vertex3f(0.259814262f, 0.0541992188f, 0.f),
    C_Vertex3f(0.270627022f, -0.0102539063f, 0.f),
    C_Vertex3f(0.272246838f, 0.0555999279f, 0.f),
    C_Vertex3f(0.290543914f, -0.00821995735f, 0.f),
    C_Vertex3f(0.284243107f, 0.0584013462f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.178680897f, 0.082691431f, 0.f),
    C_Vertex3f(0.17891407f, 0.0111131668f, 0.f),
    C_Vertex3f(0.190403938f, 0.0731940269f, 0.f),
    C_Vertex3f(0.195033073f, 0.00256633759f, 0.f),
    C_Vertex3f(0.202944756f, 0.0655961037f, 0.f),
    C_Vertex3f(0.212258339f, -0.00384378433f, 0.f),
    C_Vertex3f(0.216303349f, 0.0598976612f, 0.f),
    C_Vertex3f(0.230589867f, -0.00811719894f, 0.f),
    C_Vertex3f(0.230479717f, 0.0560986996f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.167775631f, 0.094088316f, 0.f),
    C_Vertex3f(0.157688141f, 0.107384682f, 0.f),
    C_Vertex3f(0.163901329f, 0.0217967033f, 0.f),
    C_Vertex3f(0.17891407f, 0.0111131668f, 0.f),
    C_Vertex3f(0.178680897f, 0.082691431f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15021801f, 0.119628906f, 0.f),
    C_Vertex3f(0.145256281f, 0.13037014f, 0.f),
    C_Vertex3f(0.14999485f, 0.0346169472f, 0.f),
    C_Vertex3f(0.163901329f, 0.0217967033f, 0.f),
    C_Vertex3f(0.157688141f, 0.107384682f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14999485f, 0.0346169472f, 0.f),
    C_Vertex3f(0.145256281f, 0.13037014f, 0.f),
    C_Vertex3f(0.14100337f, 0.142560005f, 0.f),
    C_Vertex3f(0.137459278f, 0.156198502f, 0.f),
    C_Vertex3f(0.137194633f, 0.0495738983f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.131347656f, 0.f, 0.f),
    C_Vertex3f(0.05859375f, 0.640625f, 0.f),
    C_Vertex3f(0.05859375f, 0.f, 0.f),
    C_Vertex3f(0.143432617f, 0.420115471f, 0.f),
    C_Vertex3f(0.147955656f, 0.333568573f, 0.f),
    C_Vertex3f(0.156744003f, 0.433738947f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.130371094f, 0.246283531f, -1.f),
    C_Vertex3f(0.131543398f, 0.266241074f, -1.f),
    C_Vertex3f(0.137207031f, 0.640625f, -1.f),
    C_Vertex3f(0.133888006f, 0.284948349f, -1.f),
    C_Vertex3f(0.137207031f, 0.412330627f, -1.f),
    C_Vertex3f(0.137404919f, 0.302405357f, -1.f),
    C_Vertex3f(0.143432617f, 0.420115471f, -1.f),
    C_Vertex3f(0.142094135f, 0.318612099f, -1.f),
    C_Vertex3f(0.147955656f, 0.333568573f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.137459278f, 0.156198502f, -1.f),
    C_Vertex3f(0.134624004f, 0.171285629f, -1.f),
    C_Vertex3f(0.137194633f, 0.0495738983f, -1.f),
    C_Vertex3f(0.132497549f, 0.187821388f, -1.f),
    C_Vertex3f(0.131347656f, 0.0581207275f, -1.f),
    C_Vertex3f(0.131079912f, 0.205805779f, -1.f),
    C_Vertex3f(0.131347656f, 0.f, -1.f),
    C_Vertex3f(0.130371094f, 0.2252388f, -1.f),
    C_Vertex3f(0.05859375f, 0.640625f, -1.f),
    C_Vertex3f(0.130371094f, 0.246283531f, -1.f),
    C_Vertex3f(0.137207031f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.378758192f, 0.265232205f, -1.f),
    C_Vertex3f(0.379882813f, 0.243982315f, -1.f),
    C_Vertex3f(0.388084412f, 0.430546761f, -1.f),
    C_Vertex3f(0.394551754f, 0.0466966629f, -1.f),
    C_Vertex3f(0.402119637f, 0.416678429f, -1.f),
    C_Vertex3f(0.409304142f, 0.0629682541f, -1.f),
    C_Vertex3f(0.41446209f, 0.401062965f, -1.f),
    C_Vertex3f(0.422212481f, 0.0809323788f, -1.f),
    C_Vertex3f(0.425466537f, 0.383836269f, -1.f),
    C_Vertex3f(0.433276772f, 0.100589037f, -1.f),
    C_Vertex3f(0.43513298f, 0.364998341f, -1.f),
    C_Vertex3f(0.442497015f, 0.121938229f, -1.f),
    C_Vertex3f(0.443461418f, 0.344549179f, -1.f),
    C_Vertex3f(0.449873209f, 0.144979954f, -1.f),
    C_Vertex3f(0.450451851f, 0.322488785f, -1.f),
    C_Vertex3f(0.455405354f, 0.169714212f, -1.f),
    C_Vertex3f(0.455694675f, 0.299664497f, -1.f),
    C_Vertex3f(0.459093451f, 0.196141005f, -1.f),
    C_Vertex3f(0.459189892f, 0.276076317f, -1.f),
    C_Vertex3f(0.4609375f, 0.22426033f, -1.f),
    C_Vertex3f(0.4609375f, 0.251724243f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.363013506f, 0.335336566f, -1.f),
    C_Vertex3f(0.368636608f, 0.320044756f, -1.f),
    C_Vertex3f(0.372356415f, 0.442667961f, -1.f),
    C_Vertex3f(0.37313509f, 0.303263426f, -1.f),
    C_Vertex3f(0.388084412f, 0.430546761f, -1.f),
    C_Vertex3f(0.376508951f, 0.284992576f, -1.f),
    C_Vertex3f(0.378758192f, 0.265232205f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339396477f, 0.372274876f, -1.f),
    C_Vertex3f(0.34839344f, 0.361451626f, -1.f),
    C_Vertex3f(0.354935646f, 0.45304203f, -1.f),
    C_Vertex3f(0.356265783f, 0.349138856f, -1.f),
    C_Vertex3f(0.372356415f, 0.442667961f, -1.f),
    C_Vertex3f(0.363013506f, 0.335336566f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.335822105f, 0.461668968f, -1.f),
    C_Vertex3f(0.316135883f, 0.468139172f, -1.f),
    C_Vertex3f(0.319992065f, 0.389862657f, -1.f),
    C_Vertex3f(0.329929352f, 0.381745219f, -1.f),
    C_Vertex3f(0.339396477f, 0.372274876f, -1.f),
    C_Vertex3f(0.354935646f, 0.45304203f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.275541306f, 0.408803344f, -1.f),
    C_Vertex3f(0.287359238f, 0.406097531f, -1.f),
    C_Vertex3f(0.29587698f, 0.472452641f, -1.f),
    C_Vertex3f(0.298707008f, 0.402038813f, -1.f),
    C_Vertex3f(0.316135883f, 0.468139172f, -1.f),
    C_Vertex3f(0.309584618f, 0.396627188f, -1.f),
    C_Vertex3f(0.319992065f, 0.389862657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.238007069f, 0.408755541f, -1.f),
    C_Vertex3f(0.250439644f, 0.41015625f, -1.f),
    C_Vertex3f(0.254676819f, 0.474609375f, -1.f),
    C_Vertex3f(0.263253212f, 0.41015625f, -1.f),
    C_Vertex3f(0.275045395f, 0.474609375f, -1.f),
    C_Vertex3f(0.275541306f, 0.408803344f, -1.f),
    C_Vertex3f(0.29587698f, 0.472452641f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.203327179f, 0.396149158f, -1.f),
    C_Vertex3f(0.214450836f, 0.401751995f, -1.f),
    C_Vertex3f(0.21859169f, 0.468770742f, -1.f),
    C_Vertex3f(0.226010799f, 0.405954123f, -1.f),
    C_Vertex3f(0.236204147f, 0.472663164f, -1.f),
    C_Vertex3f(0.238007069f, 0.408755541f, -1.f),
    C_Vertex3f(0.254676819f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172574043f, 0.370936394f, -1.f),
    C_Vertex3f(0.182388783f, 0.380741358f, -1.f),
    C_Vertex3f(0.185947418f, 0.455147266f, -1.f),
    C_Vertex3f(0.192639828f, 0.389145613f, -1.f),
    C_Vertex3f(0.201839447f, 0.46293211f, -1.f),
    C_Vertex3f(0.203327179f, 0.396149158f, -1.f),
    C_Vertex3f(0.21859169f, 0.468770742f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147955656f, 0.333568573f, -1.f),
    C_Vertex3f(0.154989481f, 0.34727478f, -1.f),
    C_Vertex3f(0.156744003f, 0.433738947f, -1.f),
    C_Vertex3f(0.16319561f, 0.359730721f, -1.f),
    C_Vertex3f(0.170915604f, 0.445416212f, -1.f),
    C_Vertex3f(0.172574043f, 0.370936394f, -1.f),
    C_Vertex3f(0.185947418f, 0.455147266f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.379882813f, 0.221516609f, -1.f),
    C_Vertex3f(0.378710508f, 0.200697899f, -1.f),
    C_Vertex3f(0.37906158f, 0.0324590206f, -1.f),
    C_Vertex3f(0.394551754f, 0.0466966629f, -1.f),
    C_Vertex3f(0.379882813f, 0.243982315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37906158f, 0.0324590206f, -1.f),
    C_Vertex3f(0.378710508f, 0.200697899f, -1.f),
    C_Vertex3f(0.3763659f, 0.18125248f, -1.f),
    C_Vertex3f(0.372848988f, 0.163180351f, -1.f),
    C_Vertex3f(0.368159771f, 0.146481514f, -1.f),
    C_Vertex3f(0.362833619f, 0.0202553272f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.362833619f, 0.0202553272f, -1.f),
    C_Vertex3f(0.368159771f, 0.146481514f, -1.f),
    C_Vertex3f(0.36229825f, 0.131155968f, -1.f),
    C_Vertex3f(0.355264425f, 0.117203712f, -1.f),
    C_Vertex3f(0.347058296f, 0.104624748f, -1.f),
    C_Vertex3f(0.345867872f, 0.0100855827f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.337679863f, 0.093419075f, -1.f),
    C_Vertex3f(0.327865124f, 0.0836141109f, -1.f),
    C_Vertex3f(0.328164339f, 0.00194978714f, -1.f),
    C_Vertex3f(0.345867872f, 0.0100855827f, -1.f),
    C_Vertex3f(0.347058296f, 0.104624748f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.317614079f, 0.075209856f, -1.f),
    C_Vertex3f(0.306926727f, 0.0682063103f, -1.f),
    C_Vertex3f(0.30972302f, -0.00415205956f, -1.f),
    C_Vertex3f(0.328164339f, 0.00194978714f, -1.f),
    C_Vertex3f(0.327865124f, 0.0836141109f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.29580307f, 0.0626034737f, -1.f),
    C_Vertex3f(0.284243107f, 0.0584013462f, -1.f),
    C_Vertex3f(0.290543914f, -0.00821995735f, -1.f),
    C_Vertex3f(0.30972302f, -0.00415205956f, -1.f),
    C_Vertex3f(0.306926727f, 0.0682063103f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230479717f, 0.0560986996f, -1.f),
    C_Vertex3f(0.230589867f, -0.00811719894f, -1.f),
    C_Vertex3f(0.245473862f, 0.0541992188f, -1.f),
    C_Vertex3f(0.250027657f, -0.0102539063f, -1.f),
    C_Vertex3f(0.259814262f, 0.0541992188f, -1.f),
    C_Vertex3f(0.270627022f, -0.0102539063f, -1.f),
    C_Vertex3f(0.272246838f, 0.0555999279f, -1.f),
    C_Vertex3f(0.290543914f, -0.00821995735f, -1.f),
    C_Vertex3f(0.284243107f, 0.0584013462f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.178680897f, 0.082691431f, -1.f),
    C_Vertex3f(0.17891407f, 0.0111131668f, -1.f),
    C_Vertex3f(0.190403938f, 0.0731940269f, -1.f),
    C_Vertex3f(0.195033073f, 0.00256633759f, -1.f),
    C_Vertex3f(0.202944756f, 0.0655961037f, -1.f),
    C_Vertex3f(0.212258339f, -0.00384378433f, -1.f),
    C_Vertex3f(0.216303349f, 0.0598976612f, -1.f),
    C_Vertex3f(0.230589867f, -0.00811719894f, -1.f),
    C_Vertex3f(0.230479717f, 0.0560986996f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.167775631f, 0.094088316f, -1.f),
    C_Vertex3f(0.157688141f, 0.107384682f, -1.f),
    C_Vertex3f(0.163901329f, 0.0217967033f, -1.f),
    C_Vertex3f(0.17891407f, 0.0111131668f, -1.f),
    C_Vertex3f(0.178680897f, 0.082691431f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15021801f, 0.119628906f, -1.f),
    C_Vertex3f(0.145256281f, 0.13037014f, -1.f),
    C_Vertex3f(0.14999485f, 0.0346169472f, -1.f),
    C_Vertex3f(0.163901329f, 0.0217967033f, -1.f),
    C_Vertex3f(0.157688141f, 0.107384682f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14999485f, 0.0346169472f, -1.f),
    C_Vertex3f(0.145256281f, 0.13037014f, -1.f),
    C_Vertex3f(0.14100337f, 0.142560005f, -1.f),
    C_Vertex3f(0.137459278f, 0.156198502f, -1.f),
    C_Vertex3f(0.137194633f, 0.0495738983f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.131347656f, 0.f, -1.f),
    C_Vertex3f(0.05859375f, 0.640625f, -1.f),
    C_Vertex3f(0.05859375f, 0.f, -1.f),
    C_Vertex3f(0.143432617f, 0.420115471f, -1.f),
    C_Vertex3f(0.147955656f, 0.333568573f, -1.f),
    C_Vertex3f(0.156744003f, 0.433738947f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_99[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.131752744f, 0.991282582f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, -1.f),
    C_Vertex3f(0.439453125f, 0.159667969f, -1.f),
    C_Vertex3f(0.439453125f, 0.159667969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.987604678f, -0.156962141f, 0.f),
    C_Vertex3f(0.439453125f, 0.159667969f, 0.f),
    C_Vertex3f(0.439453125f, 0.159667969f, -1.f),
    C_Normal3f(0.97908473f, -0.203453019f, 0.f),
    C_Vertex3f(0.437869072f, 0.149701118f, 0.f),
    C_Vertex3f(0.437869072f, 0.149701118f, -1.f),
    C_Normal3f(0.964377761f, -0.26452899f, 0.f),
    C_Vertex3f(0.433464766f, 0.130849957f, 0.f),
    C_Vertex3f(0.433464766f, 0.130849957f, -1.f),
    C_Normal3f(0.939896464f, -0.341459483f, 0.f),
    C_Vertex3f(0.427824259f, 0.113081336f, 0.f),
    C_Vertex3f(0.427824259f, 0.113081336f, -1.f),
    C_Normal3f(0.907030344f, -0.421065122f, 0.f),
    C_Vertex3f(0.420947552f, 0.0963952541f, 0.f),
    C_Vertex3f(0.420947552f, 0.0963952541f, -1.f),
    C_Normal3f(0.865158498f, -0.501498401f, 0.f),
    C_Vertex3f(0.412834644f, 0.0807917118f, 0.f),
    C_Vertex3f(0.412834644f, 0.0807917118f, -1.f),
    C_Normal3f(0.814237773f, -0.580531597f, 0.f),
    C_Vertex3f(0.403485537f, 0.066270709f, 0.f),
    C_Vertex3f(0.403485537f, 0.066270709f, -1.f),
    C_Normal3f(0.754939914f, -0.655793846f, 0.f),
    C_Vertex3f(0.392900229f, 0.0528322458f, 0.f),
    C_Vertex3f(0.392900229f, 0.0528322458f, -1.f),
    C_Normal3f(0.688658357f, -0.725086033f, 0.f),
    C_Vertex3f(0.38107872f, 0.0404763222f, 0.f),
    C_Vertex3f(0.38107872f, 0.0404763222f, -1.f),
    C_Normal3f(0.618423939f, -0.785844684f, 0.f),
    C_Vertex3f(0.368021011f, 0.0292029381f, 0.f),
    C_Vertex3f(0.368021011f, 0.0292029381f, -1.f),
    C_Normal3f(0.542743206f, -0.839898705f, 0.f),
    C_Vertex3f(0.354218721f, 0.019338727f, 0.f),
    C_Vertex3f(0.354218721f, 0.019338727f, -1.f),
    C_Normal3f(0.461000979f, -0.887399673f, 0.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, 0.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, -1.f),
    C_Normal3f(0.375253379f, -0.926922321f, 0.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, 0.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, -1.f),
    C_Normal3f(0.287862331f, -0.957671821f, 0.f),
    C_Vertex3f(0.308344364f, -0.00179886818f, 0.f),
    C_Vertex3f(0.308344364f, -0.00179886818f, -1.f),
    C_Normal3f(0.201184601f, -0.979553342f, 0.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, 0.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, -1.f),
    C_Normal3f(0.117288537f, -0.993097842f, 0.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, 0.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, -1.f),
    C_Normal3f(0.0354844518f, -0.999370158f, 0.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, 0.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, -1.f),
    C_Normal3f(-0.044093933f, -0.999027431f, 0.f),
    C_Vertex3f(0.234351158f, -0.0102539063f, 0.f),
    C_Vertex3f(0.234351158f, -0.0102539063f, -1.f),
    C_Normal3f(-0.131067663f, -0.99137342f, 0.f),
    C_Vertex3f(0.211698174f, -0.00830876827f, 0.f),
    C_Vertex3f(0.211698174f, -0.00830876827f, -1.f),
    C_Normal3f(-0.226606935f, -0.973986268f, 0.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, 0.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, -1.f),
    C_Normal3f(-0.326136827f, -0.945322514f, 0.f),
    C_Vertex3f(0.169895887f, 0.00141692162f, 0.f),
    C_Vertex3f(0.169895887f, 0.00141692162f, -1.f),
    C_Normal3f(-0.426454365f, -0.904509068f, 0.f),
    C_Vertex3f(0.150746584f, 0.00919747353f, 0.f),
    C_Vertex3f(0.150746584f, 0.00919747353f, -1.f),
    C_Normal3f(-0.523817837f, -0.851830363f, 0.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, 0.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, -1.f),
    C_Normal3f(-0.61455822f, -0.788871467f, 0.f),
    C_Vertex3f(0.115951657f, 0.0305939913f, 0.f),
    C_Vertex3f(0.115951657f, 0.0305939913f, -1.f),
    C_Normal3f(-0.69573462f, -0.718298972f, 0.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, 0.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, -1.f),
    C_Normal3f(-0.770461857f, -0.637486041f, 0.f),
    C_Vertex3f(0.0858283043f, 0.0597710609f, 0.f),
    C_Vertex3f(0.0858283043f, 0.0597710609f, -1.f),
    C_Normal3f(-0.838346004f, -0.545138538f, 0.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, 0.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, -1.f),
    C_Normal3f(-0.893242121f, -0.449575752f, 0.f),
    C_Vertex3f(0.0623019934f, 0.0959511995f, 0.f),
    C_Vertex3f(0.0623019934f, 0.0959511995f, -1.f),
    C_Normal3f(-0.934954882f, -0.354766816f, 0.f),
    C_Vertex3f(0.0532534122f, 0.116570234f, 0.f),
    C_Vertex3f(0.0532534122f, 0.116570234f, -1.f),
    C_Normal3f(-0.96454227f, -0.263928503f, 0.f),
    C_Vertex3f(0.0460145473f, 0.138875246f, 0.f),
    C_Vertex3f(0.0460145473f, 0.138875246f, -1.f),
    C_Normal3f(-0.983806968f, -0.179231405f, 0.f),
    C_Vertex3f(0.0405853987f, 0.162866235f, 0.f),
    C_Vertex3f(0.0405853987f, 0.162866235f, -1.f),
    C_Normal3f(-0.994802058f, -0.101827636f, 0.f),
    C_Vertex3f(0.0369659662f, 0.1885432f, 0.f),
    C_Vertex3f(0.0369659662f, 0.1885432f, -1.f),
    C_Normal3f(-0.999555469f, -0.0298151094f, 0.f),
    C_Vertex3f(0.03515625f, 0.215906143f, 0.f),
    C_Vertex3f(0.03515625f, 0.215906143f, -1.f),
    C_Normal3f(-0.998968303f, 0.0454123914f, 0.f),
    C_Vertex3f(0.03515625f, 0.249214172f, 0.f),
    C_Vertex3f(0.03515625f, 0.249214172f, -1.f),
    C_Normal3f(-0.990711629f, 0.135979757f, 0.f),
    C_Vertex3f(0.0382714272f, 0.284432888f, 0.f),
    C_Vertex3f(0.0382714272f, 0.284432888f, -1.f),
    C_Normal3f(-0.97111553f, 0.238609701f, 0.f),
    C_Vertex3f(0.0445017815f, 0.317303181f, 0.f),
    C_Vertex3f(0.0445017815f, 0.317303181f, -1.f),
    C_Normal3f(-0.940327108f, 0.34027198f, 0.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, 0.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, -1.f),
    C_Normal3f(-0.902472794f, 0.430746824f, 0.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, 0.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, -1.f),
    C_Normal3f(-0.863703251f, 0.504000664f, 0.f),
    C_Vertex3f(0.0702359676f, 0.382161498f, 0.f),
    C_Vertex3f(0.0702359676f, 0.382161498f, -1.f),
    C_Normal3f(-0.821103454f, 0.570779502f, 0.f),
    C_Vertex3f(0.0780918598f, 0.3944875f, 0.f),
    C_Vertex3f(0.0780918598f, 0.3944875f, -1.f),
    C_Normal3f(-0.771526754f, 0.636196911f, 0.f),
    C_Vertex3f(0.0867605209f, 0.405933142f, 0.f),
    C_Vertex3f(0.0867605209f, 0.405933142f, -1.f),
    C_Normal3f(-0.715440392f, 0.698673785f, 0.f),
    C_Vertex3f(0.096241951f, 0.416498423f, 0.f),
    C_Vertex3f(0.096241951f, 0.416498423f, -1.f),
    C_Normal3f(-0.653754652f, 0.756706655f, 0.f),
    C_Vertex3f(0.10653615f, 0.426183343f, 0.f),
    C_Vertex3f(0.10653615f, 0.426183343f, -1.f),
    C_Normal3f(-0.587759495f, 0.809035778f, 0.f),
    C_Vertex3f(0.117643118f, 0.434987903f, 0.f),
    C_Vertex3f(0.117643118f, 0.434987903f, -1.f),
    C_Normal3f(-0.512026966f, 0.85896939f, 0.f),
    C_Vertex3f(0.129562855f, 0.442912102f, 0.f),
    C_Vertex3f(0.129562855f, 0.442912102f, -1.f),
    C_Normal3f(-0.419815004f, 0.907609761f, 0.f),
    C_Vertex3f(0.148661613f, 0.453477859f, 0.f),
    C_Vertex3f(0.148661613f, 0.453477859f, -1.f),
    C_Normal3f(-0.308520645f, 0.951217651f, 0.f),
    C_Vertex3f(0.175247669f, 0.464043617f, 0.f),
    C_Vertex3f(0.175247669f, 0.464043617f, -1.f),
    C_Normal3f(-0.183707342f, 0.982981026f, 0.f),
    C_Vertex3f(0.202954769f, 0.471087456f, 0.f),
    C_Vertex3f(0.202954769f, 0.471087456f, -1.f),
    C_Normal3f(-0.0659859255f, 0.997820497f, 0.f),
    C_Vertex3f(0.231782913f, 0.474609375f, 0.f),
    C_Vertex3f(0.231782913f, 0.474609375f, -1.f),
    C_Normal3f(0.0281452518f, 0.999603868f, 0.f),
    C_Vertex3f(0.256212234f, 0.474609375f, 0.f),
    C_Vertex3f(0.256212234f, 0.474609375f, -1.f),
    C_Normal3f(0.10130471f, 0.994855404f, 0.f),
    C_Vertex3f(0.274260879f, 0.473413348f, 0.f),
    C_Vertex3f(0.274260879f, 0.473413348f, -1.f),
    C_Normal3f(0.175657928f, 0.984451234f, 0.f),
    C_Vertex3f(0.291448712f, 0.471021295f, 0.f),
    C_Vertex3f(0.291448712f, 0.471021295f, -1.f),
    C_Normal3f(0.25465107f, 0.967033029f, 0.f),
    C_Vertex3f(0.307775736f, 0.467433214f, 0.f),
    C_Vertex3f(0.307775736f, 0.467433214f, -1.f),
    C_Normal3f(0.337013245f, 0.941499949f, 0.f),
    C_Vertex3f(0.323241949f, 0.462649107f, 0.f),
    C_Vertex3f(0.323241949f, 0.462649107f, -1.f),
    C_Normal3f(0.420948029f, 0.907084763f, 0.f),
    C_Vertex3f(0.337847352f, 0.456668973f, 0.f),
    C_Vertex3f(0.337847352f, 0.456668973f, -1.f),
    C_Normal3f(0.504237056f, 0.863565207f, 0.f),
    C_Vertex3f(0.351591945f, 0.449492812f, 0.f),
    C_Vertex3f(0.351591945f, 0.449492812f, -1.f),
    C_Normal3f(0.584466457f, 0.811417818f, 0.f),
    C_Vertex3f(0.364475727f, 0.441120625f, 0.f),
    C_Vertex3f(0.364475727f, 0.441120625f, -1.f),
    C_Normal3f(0.659348786f, 0.751837194f, 0.f),
    C_Vertex3f(0.376498699f, 0.43155241f, 0.f),
    C_Vertex3f(0.376498699f, 0.43155241f, -1.f),
    C_Normal3f(0.728860795f, 0.684661984f, 0.f),
    C_Vertex3f(0.387442231f, 0.420979381f, 0.f),
    C_Vertex3f(0.387442231f, 0.420979381f, -1.f),
    C_Normal3f(0.791613638f, 0.611021996f, 0.f),
    C_Vertex3f(0.397306323f, 0.409401536f, 0.f),
    C_Vertex3f(0.397306323f, 0.409401536f, -1.f),
    C_Normal3f(0.846052468f, 0.533099651f, 0.f),
    C_Vertex3f(0.406090975f, 0.396818876f, 0.f),
    C_Vertex3f(0.406090975f, 0.396818876f, -1.f),
    C_Normal3f(0.891355872f, 0.453304201f, 0.f),
    C_Vertex3f(0.413796186f, 0.383231401f, 0.f),
    C_Vertex3f(0.413796186f, 0.383231401f, -1.f),
    C_Normal3f(0.927452505f, 0.373940825f, 0.f),
    C_Vertex3f(0.420421958f, 0.368639112f, 0.f),
    C_Vertex3f(0.420421958f, 0.368639112f, -1.f),
    C_Normal3f(0.954892576f, 0.296951592f, 0.f),
    C_Vertex3f(0.425968289f, 0.353042006f, 0.f),
    C_Vertex3f(0.425968289f, 0.353042006f, -1.f),
    C_Normal3f(0.971834719f, 0.235663593f, 0.f),
    C_Vertex3f(0.430435181f, 0.336440086f, 0.f),
    C_Vertex3f(0.430435181f, 0.336440086f, -1.f),
    C_Normal3f(0.981990516f, 0.188930243f, 0.f),
    C_Vertex3f(0.432128906f, 0.327636719f, 0.f),
    C_Vertex3f(0.432128906f, 0.327636719f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.151110858f, -0.988516808f, 0.f),
    C_Vertex3f(0.432128906f, 0.327636719f, 0.f),
    C_Vertex3f(0.432128906f, 0.327636719f, -1.f),
    C_Vertex3f(0.35546875f, 0.315917969f, -1.f),
    C_Vertex3f(0.35546875f, 0.315917969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.973956823f, -0.226733729f, 0.f),
    C_Vertex3f(0.35546875f, 0.315917969f, 0.f),
    C_Vertex3f(0.35546875f, 0.315917969f, -1.f),
    C_Normal3f(-0.952746391f, -0.303766668f, 0.f),
    C_Vertex3f(0.352739334f, 0.327642441f, 0.f),
    C_Vertex3f(0.352739334f, 0.327642441f, -1.f),
    C_Normal3f(-0.911611378f, -0.41105324f, 0.f),
    C_Vertex3f(0.345180035f, 0.348187923f, 0.f),
    C_Vertex3f(0.345180035f, 0.348187923f, -1.f),
    C_Normal3f(-0.834030807f, -0.551717758f, 0.f),
    C_Vertex3f(0.335520267f, 0.365829945f, 0.f),
    C_Vertex3f(0.335520267f, 0.365829945f, -1.f),
    C_Normal3f(-0.719893813f, -0.694084167f, 0.f),
    C_Vertex3f(0.323760033f, 0.380568504f, 0.f),
    C_Vertex3f(0.323760033f, 0.380568504f, -1.f),
    C_Normal3f(-0.578953743f, -0.815360427f, 0.f),
    C_Vertex3f(0.30989933f, 0.392403603f, 0.f),
    C_Vertex3f(0.30989933f, 0.392403603f, -1.f),
    C_Normal3f(-0.418954581f, -0.908007264f, 0.f),
    C_Vertex3f(0.294591427f, 0.401279926f, 0.f),
    C_Vertex3f(0.294591427f, 0.401279926f, -1.f),
    C_Normal3f(-0.246108323f, -0.969242394f, 0.f),
    C_Vertex3f(0.277836323f, 0.407197475f, 0.f),
    C_Vertex3f(0.277836323f, 0.407197475f, -1.f),
    C_Normal3f(-0.0831752717f, -0.996534884f, 0.f),
    C_Vertex3f(0.259634018f, 0.41015625f, 0.f),
    C_Vertex3f(0.259634018f, 0.41015625f, -1.f),
    C_Normal3f(0.0424002297f, -0.999100745f, 0.f),
    C_Vertex3f(0.242386818f, 0.41015625f, 0.f),
    C_Vertex3f(0.242386818f, 0.41015625f, -1.f),
    C_Normal3f(0.143321022f, -0.989676297f, 0.f),
    C_Vertex3f(0.228237867f, 0.408823848f, 0.f),
    C_Vertex3f(0.228237867f, 0.408823848f, -1.f),
    C_Normal3f(0.246344402f, -0.969182372f, 0.f),
    C_Vertex3f(0.214784861f, 0.406159043f, 0.f),
    C_Vertex3f(0.214784861f, 0.406159043f, -1.f),
    C_Normal3f(0.351784617f, -0.936080992f, 0.f),
    C_Vertex3f(0.202027798f, 0.402161837f, 0.f),
    C_Vertex3f(0.202027798f, 0.402161837f, -1.f),
    C_Normal3f(0.455658734f, -0.8901546f, 0.f),
    C_Vertex3f(0.189966679f, 0.396832228f, 0.f),
    C_Vertex3f(0.189966679f, 0.396832228f, -1.f),
    C_Normal3f(0.553831875f, -0.832628489f, 0.f),
    C_Vertex3f(0.178601503f, 0.390170217f, 0.f),
    C_Vertex3f(0.178601503f, 0.390170217f, -1.f),
    C_Normal3f(0.642789781f, -0.76604265f, 0.f),
    C_Vertex3f(0.167932272f, 0.382175803f, 0.f),
    C_Vertex3f(0.167932272f, 0.382175803f, -1.f),
    C_Normal3f(0.720233381f, -0.693731844f, 0.f),
    C_Vertex3f(0.157958984f, 0.372848988f, 0.f),
    C_Vertex3f(0.157958984f, 0.372848988f, -1.f),
    C_Normal3f(0.79592365f, -0.605397046f, 0.f),
    C_Vertex3f(0.148681641f, 0.36218977f, 0.f),
    C_Vertex3f(0.148681641f, 0.36218977f, -1.f),
    C_Normal3f(0.864854872f, -0.502021909f, 0.f),
    C_Vertex3f(0.140563965f, 0.34997952f, 0.f),
    C_Vertex3f(0.140563965f, 0.34997952f, -1.f),
    C_Normal3f(0.915732861f, -0.40178749f, 0.f),
    C_Vertex3f(0.133605957f, 0.336218238f, 0.f),
    C_Vertex3f(0.133605957f, 0.336218238f, -1.f),
    C_Normal3f(0.951208353f, -0.308549166f, 0.f),
    C_Vertex3f(0.127807617f, 0.320905924f, 0.f),
    C_Vertex3f(0.127807617f, 0.320905924f, -1.f),
    C_Normal3f(0.974532247f, -0.224247366f, 0.f),
    C_Vertex3f(0.123168945f, 0.304042578f, 0.f),
    C_Vertex3f(0.123168945f, 0.304042578f, -1.f),
    C_Normal3f(0.988779366f, -0.14938277f, 0.f),
    C_Vertex3f(0.119689941f, 0.2856282f, 0.f),
    C_Vertex3f(0.119689941f, 0.2856282f, -1.f),
    C_Normal3f(0.996501565f, -0.0835746899f, 0.f),
    C_Vertex3f(0.117370605f, 0.265662789f, 0.f),
    C_Vertex3f(0.117370605f, 0.265662789f, -1.f),
    C_Normal3f(0.999664426f, -0.025906859f, 0.f),
    C_Vertex3f(0.116210938f, 0.244146347f, 0.f),
    C_Vertex3f(0.116210938f, 0.244146347f, -1.f),
    C_Normal3f(0.999690711f, 0.0248676781f, 0.f),
    C_Vertex3f(0.116210938f, 0.220914841f, 0.f),
    C_Vertex3f(0.116210938f, 0.220914841f, -1.f),
    C_Normal3f(0.996813893f, 0.0797625408f, 0.f),
    C_Vertex3f(0.117331028f, 0.199118257f, 0.f),
    C_Vertex3f(0.117331028f, 0.199118257f, -1.f),
    C_Normal3f(0.989740729f, 0.14287433f, 0.f),
    C_Vertex3f(0.119571209f, 0.178920627f, 0.f),
    C_Vertex3f(0.119571209f, 0.178920627f, -1.f),
    C_Normal3f(0.97659272f, 0.215096921f, 0.f),
    C_Vertex3f(0.12293148f, 0.160321951f, 0.f),
    C_Vertex3f(0.12293148f, 0.160321951f, -1.f),
    C_Normal3f(0.95485121f, 0.297084332f, 0.f),
    C_Vertex3f(0.127411842f, 0.143322229f, 0.f),
    C_Vertex3f(0.127411842f, 0.143322229f, -1.f),
    C_Normal3f(0.921351075f, 0.38873148f, 0.f),
    C_Vertex3f(0.133012295f, 0.127921462f, 0.f),
    C_Vertex3f(0.133012295f, 0.127921462f, -1.f),
    C_Normal3f(0.872526586f, 0.488566548f, 0.f),
    C_Vertex3f(0.139732838f, 0.114119649f, 0.f),
    C_Vertex3f(0.139732838f, 0.114119649f, -1.f),
    C_Normal3f(0.805116951f, 0.593116045f, 0.f),
    C_Vertex3f(0.147573471f, 0.10191679f, 0.f),
    C_Vertex3f(0.147573471f, 0.10191679f, -1.f),
    C_Normal3f(0.730181158f, 0.683253646f, 0.f),
    C_Vertex3f(0.156534195f, 0.0913128853f, 0.f),
    C_Vertex3f(0.156534195f, 0.0913128853f, -1.f),
    C_Normal3f(0.653271496f, 0.757123709f, 0.f),
    C_Vertex3f(0.166177988f, 0.0820344687f, 0.f),
    C_Vertex3f(0.166177988f, 0.0820344687f, -1.f),
    C_Normal3f(0.564196467f, 0.825640559f, 0.f),
    C_Vertex3f(0.17650485f, 0.0740815401f, 0.f),
    C_Vertex3f(0.17650485f, 0.0740815401f, -1.f),
    C_Normal3f(0.465150356f, 0.885231733f, 0.f),
    C_Vertex3f(0.187514782f, 0.0674540997f, 0.f),
    C_Vertex3f(0.187514782f, 0.0674540997f, -1.f),
    C_Normal3f(0.359690905f, 0.933071494f, 0.f),
    C_Vertex3f(0.199207783f, 0.0621521473f, 0.f),
    C_Vertex3f(0.199207783f, 0.0621521473f, -1.f),
    C_Normal3f(0.252142906f, 0.967689991f, 0.f),
    C_Vertex3f(0.211583853f, 0.058175683f, 0.f),
    C_Vertex3f(0.211583853f, 0.058175683f, -1.f),
    C_Normal3f(0.146761522f, 0.989171922f, 0.f),
    C_Vertex3f(0.224642992f, 0.0555247068f, 0.f),
    C_Vertex3f(0.224642992f, 0.0555247068f, -1.f),
    C_Normal3f(0.0494966768f, 0.99877435f, 0.f),
    C_Vertex3f(0.238385201f, 0.0541992188f, 0.f),
    C_Vertex3f(0.238385201f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0367671363f, 0.999323905f, 0.f),
    C_Vertex3f(0.251389503f, 0.0541992188f, 0.f),
    C_Vertex3f(0.251389503f, 0.0541992188f, -1.f),
    C_Normal3f(-0.121542171f, 0.992586255f, 0.f),
    C_Vertex3f(0.262494683f, 0.0550862551f, 0.f),
    C_Vertex3f(0.262494683f, 0.0550862551f, -1.f),
    C_Normal3f(-0.208772928f, 0.977964103f, 0.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, 0.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, -1.f),
    C_Normal3f(-0.298960239f, 0.954265594f, 0.f),
    C_Vertex3f(0.283270597f, 0.0595214367f, 0.f),
    C_Vertex3f(0.283270597f, 0.0595214367f, -1.f),
    C_Normal3f(-0.389751405f, 0.920920014f, 0.f),
    C_Vertex3f(0.292941332f, 0.063069582f, 0.f),
    C_Vertex3f(0.292941332f, 0.063069582f, -1.f),
    C_Normal3f(-0.478471577f, 0.878103018f, 0.f),
    C_Vertex3f(0.302133918f, 0.0675047636f, 0.f),
    C_Vertex3f(0.302133918f, 0.0675047636f, -1.f),
    C_Normal3f(-0.562476099f, 0.826813638f, 0.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, 0.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, -1.f),
    C_Normal3f(-0.639527619f, 0.768768132f, 0.f),
    C_Vertex3f(0.319084644f, 0.0790362358f, 0.f),
    C_Vertex3f(0.319084644f, 0.0790362358f, -1.f),
    C_Normal3f(-0.715125501f, 0.698996067f, 0.f),
    C_Vertex3f(0.326842785f, 0.0861325264f, 0.f),
    C_Vertex3f(0.326842785f, 0.0861325264f, -1.f),
    C_Normal3f(-0.786699533f, 0.617336094f, 0.f),
    C_Vertex3f(0.333890796f, 0.094184041f, 0.f),
    C_Vertex3f(0.333890796f, 0.094184041f, -1.f),
    C_Normal3f(-0.845782697f, 0.533527553f, 0.f),
    C_Vertex3f(0.340228677f, 0.10319078f, 0.f),
    C_Vertex3f(0.340228677f, 0.10319078f, -1.f),
    C_Normal3f(-0.892610729f, 0.450828373f, 0.f),
    C_Vertex3f(0.345856428f, 0.113152742f, 0.f),
    C_Vertex3f(0.345856428f, 0.113152742f, -1.f),
    C_Normal3f(-0.928346336f, 0.37171635f, 0.f),
    C_Vertex3f(0.35077405f, 0.124069929f, 0.f),
    C_Vertex3f(0.35077405f, 0.124069929f, -1.f),
    C_Normal3f(-0.954632878f, 0.297785163f, 0.f),
    C_Vertex3f(0.354981542f, 0.13594234f, 0.f),
    C_Vertex3f(0.354981542f, 0.13594234f, -1.f),
    C_Normal3f(-0.973226845f, 0.229846776f, 0.f),
    C_Vertex3f(0.358478904f, 0.148769975f, 0.f),
    C_Vertex3f(0.358478904f, 0.148769975f, -1.f),
    C_Normal3f(-0.984033346f, 0.177983955f, 0.f),
    C_Vertex3f(0.361266136f, 0.162552834f, 0.f),
    C_Vertex3f(0.361266136f, 0.162552834f, -1.f),
    C_Normal3f(-0.990214288f, 0.139555246f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.432128906f, 0.327636719f, 0.f),
    C_Vertex3f(0.430435181f, 0.336440086f, 0.f),
    C_Vertex3f(0.425968289f, 0.353042006f, 0.f),
    C_Vertex3f(0.420421958f, 0.368639112f, 0.f),
    C_Vertex3f(0.413796186f, 0.383231401f, 0.f),
    C_Vertex3f(0.406090975f, 0.396818876f, 0.f),
    C_Vertex3f(0.397306323f, 0.409401536f, 0.f),
    C_Vertex3f(0.387442231f, 0.420979381f, 0.f),
    C_Vertex3f(0.376498699f, 0.43155241f, 0.f),
    C_Vertex3f(0.364475727f, 0.441120625f, 0.f),
    C_Vertex3f(0.35546875f, 0.315917969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323760033f, 0.380568504f, 0.f),
    C_Vertex3f(0.335520267f, 0.365829945f, 0.f),
    C_Vertex3f(0.337847352f, 0.456668973f, 0.f),
    C_Vertex3f(0.345180035f, 0.348187923f, 0.f),
    C_Vertex3f(0.351591945f, 0.449492812f, 0.f),
    C_Vertex3f(0.352739334f, 0.327642441f, 0.f),
    C_Vertex3f(0.364475727f, 0.441120625f, 0.f),
    C_Vertex3f(0.35546875f, 0.315917969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.214784861f, 0.406159043f, 0.f),
    C_Vertex3f(0.228237867f, 0.408823848f, 0.f),
    C_Vertex3f(0.231782913f, 0.474609375f, 0.f),
    C_Vertex3f(0.242386818f, 0.41015625f, 0.f),
    C_Vertex3f(0.256212234f, 0.474609375f, 0.f),
    C_Vertex3f(0.259634018f, 0.41015625f, 0.f),
    C_Vertex3f(0.274260879f, 0.473413348f, 0.f),
    C_Vertex3f(0.277836323f, 0.407197475f, 0.f),
    C_Vertex3f(0.291448712f, 0.471021295f, 0.f),
    C_Vertex3f(0.294591427f, 0.401279926f, 0.f),
    C_Vertex3f(0.307775736f, 0.467433214f, 0.f),
    C_Vertex3f(0.30989933f, 0.392403603f, 0.f),
    C_Vertex3f(0.323241949f, 0.462649107f, 0.f),
    C_Vertex3f(0.323760033f, 0.380568504f, 0.f),
    C_Vertex3f(0.337847352f, 0.456668973f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.202954769f, 0.471087456f, 0.f),
    C_Vertex3f(0.175247669f, 0.464043617f, 0.f),
    C_Vertex3f(0.178601503f, 0.390170217f, 0.f),
    C_Vertex3f(0.189966679f, 0.396832228f, 0.f),
    C_Vertex3f(0.202027798f, 0.402161837f, 0.f),
    C_Vertex3f(0.214784861f, 0.406159043f, 0.f),
    C_Vertex3f(0.231782913f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.175247669f, 0.464043617f, 0.f),
    C_Vertex3f(0.148661613f, 0.453477859f, 0.f),
    C_Vertex3f(0.148681641f, 0.36218977f, 0.f),
    C_Vertex3f(0.157958984f, 0.372848988f, 0.f),
    C_Vertex3f(0.167932272f, 0.382175803f, 0.f),
    C_Vertex3f(0.178601503f, 0.390170217f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.123168945f, 0.304042578f, 0.f),
    C_Vertex3f(0.127807617f, 0.320905924f, 0.f),
    C_Vertex3f(0.129562855f, 0.442912102f, 0.f),
    C_Vertex3f(0.133605957f, 0.336218238f, 0.f),
    C_Vertex3f(0.148661613f, 0.453477859f, 0.f),
    C_Vertex3f(0.140563965f, 0.34997952f, 0.f),
    C_Vertex3f(0.148681641f, 0.36218977f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116210938f, 0.244146347f, 0.f),
    C_Vertex3f(0.117370605f, 0.265662789f, 0.f),
    C_Vertex3f(0.117643118f, 0.434987903f, 0.f),
    C_Vertex3f(0.119689941f, 0.2856282f, 0.f),
    C_Vertex3f(0.129562855f, 0.442912102f, 0.f),
    C_Vertex3f(0.123168945f, 0.304042578f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115951657f, 0.0305939913f, 0.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, 0.f),
    C_Vertex3f(0.117331028f, 0.199118257f, 0.f),
    C_Vertex3f(0.116210938f, 0.220914841f, 0.f),
    C_Vertex3f(0.116210938f, 0.244146347f, 0.f),
    C_Vertex3f(0.10653615f, 0.426183343f, 0.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.10653615f, 0.426183343f, 0.f),
    C_Vertex3f(0.096241951f, 0.416498423f, 0.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, 0.f),
    C_Vertex3f(0.0867605209f, 0.405933142f, 0.f),
    C_Vertex3f(0.0858283043f, 0.0597710609f, 0.f),
    C_Vertex3f(0.0780918598f, 0.3944875f, 0.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, 0.f),
    C_Vertex3f(0.0702359676f, 0.382161498f, 0.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0460145473f, 0.138875246f, 0.f),
    C_Vertex3f(0.0532534122f, 0.116570234f, 0.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, 0.f),
    C_Vertex3f(0.0623019934f, 0.0959511995f, 0.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, 0.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.03515625f, 0.249214172f, 0.f),
    C_Vertex3f(0.0369659662f, 0.1885432f, 0.f),
    C_Vertex3f(0.0382714272f, 0.284432888f, 0.f),
    C_Vertex3f(0.0405853987f, 0.162866235f, 0.f),
    C_Vertex3f(0.0445017815f, 0.317303181f, 0.f),
    C_Vertex3f(0.0460145473f, 0.138875246f, 0.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.439453125f, 0.159667969f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, 0.f),
    C_Vertex3f(0.368021011f, 0.0292029381f, 0.f),
    C_Vertex3f(0.38107872f, 0.0404763222f, 0.f),
    C_Vertex3f(0.392900229f, 0.0528322458f, 0.f),
    C_Vertex3f(0.403485537f, 0.066270709f, 0.f),
    C_Vertex3f(0.412834644f, 0.0807917118f, 0.f),
    C_Vertex3f(0.420947552f, 0.0963952541f, 0.f),
    C_Vertex3f(0.427824259f, 0.113081336f, 0.f),
    C_Vertex3f(0.433464766f, 0.130849957f, 0.f),
    C_Vertex3f(0.437869072f, 0.149701118f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.368021011f, 0.0292029381f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, 0.f),
    C_Vertex3f(0.361266136f, 0.162552834f, 0.f),
    C_Vertex3f(0.358478904f, 0.148769975f, 0.f),
    C_Vertex3f(0.354981542f, 0.13594234f, 0.f),
    C_Vertex3f(0.354218721f, 0.019338727f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.354218721f, 0.019338727f, 0.f),
    C_Vertex3f(0.354981542f, 0.13594234f, 0.f),
    C_Vertex3f(0.35077405f, 0.124069929f, 0.f),
    C_Vertex3f(0.345856428f, 0.113152742f, 0.f),
    C_Vertex3f(0.340228677f, 0.10319078f, 0.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.340228677f, 0.10319078f, 0.f),
    C_Vertex3f(0.333890796f, 0.094184041f, 0.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, 0.f),
    C_Vertex3f(0.326842785f, 0.0861325264f, 0.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, 0.f),
    C_Vertex3f(0.319084644f, 0.0790362358f, 0.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.308344364f, -0.00179886818f, 0.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, 0.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, 0.f),
    C_Vertex3f(0.302133918f, 0.0675047636f, 0.f),
    C_Vertex3f(0.292941332f, 0.063069582f, 0.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.283270597f, 0.0595214367f, 0.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, 0.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, 0.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, 0.f),
    C_Vertex3f(0.292941332f, 0.063069582f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262494683f, 0.0550862551f, 0.f),
    C_Vertex3f(0.251389503f, 0.0541992188f, 0.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, 0.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, 0.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.211583853f, 0.058175683f, 0.f),
    C_Vertex3f(0.211698174f, -0.00830876827f, 0.f),
    C_Vertex3f(0.224642992f, 0.0555247068f, 0.f),
    C_Vertex3f(0.234351158f, -0.0102539063f, 0.f),
    C_Vertex3f(0.238385201f, 0.0541992188f, 0.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, 0.f),
    C_Vertex3f(0.251389503f, 0.0541992188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.199207783f, 0.0621521473f, 0.f),
    C_Vertex3f(0.187514782f, 0.0674540997f, 0.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, 0.f),
    C_Vertex3f(0.211698174f, -0.00830876827f, 0.f),
    C_Vertex3f(0.211583853f, 0.058175683f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.17650485f, 0.0740815401f, 0.f),
    C_Vertex3f(0.166177988f, 0.0820344687f, 0.f),
    C_Vertex3f(0.169895887f, 0.00141692162f, 0.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, 0.f),
    C_Vertex3f(0.187514782f, 0.0674540997f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156534195f, 0.0913128853f, 0.f),
    C_Vertex3f(0.147573471f, 0.10191679f, 0.f),
    C_Vertex3f(0.150746584f, 0.00919747353f, 0.f),
    C_Vertex3f(0.169895887f, 0.00141692162f, 0.f),
    C_Vertex3f(0.166177988f, 0.0820344687f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147573471f, 0.10191679f, 0.f),
    C_Vertex3f(0.139732838f, 0.114119649f, 0.f),
    C_Vertex3f(0.150746584f, 0.00919747353f, 0.f),
    C_Vertex3f(0.133012295f, 0.127921462f, 0.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, 0.f),
    C_Vertex3f(0.127411842f, 0.143322229f, 0.f),
    C_Vertex3f(0.12293148f, 0.160321951f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.119571209f, 0.178920627f, 0.f),
    C_Vertex3f(0.117331028f, 0.199118257f, 0.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, 0.f),
    C_Vertex3f(0.12293148f, 0.160321951f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0369659662f, 0.1885432f, 0.f),
    C_Vertex3f(0.03515625f, 0.249214172f, 0.f),
    C_Vertex3f(0.03515625f, 0.215906143f, 0.f),
    C_Vertex3f(0.10653615f, 0.426183343f, 0.f),
    C_Vertex3f(0.116210938f, 0.244146347f, 0.f),
    C_Vertex3f(0.117643118f, 0.434987903f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.432128906f, 0.327636719f, -1.f),
    C_Vertex3f(0.430435181f, 0.336440086f, -1.f),
    C_Vertex3f(0.425968289f, 0.353042006f, -1.f),
    C_Vertex3f(0.420421958f, 0.368639112f, -1.f),
    C_Vertex3f(0.413796186f, 0.383231401f, -1.f),
    C_Vertex3f(0.406090975f, 0.396818876f, -1.f),
    C_Vertex3f(0.397306323f, 0.409401536f, -1.f),
    C_Vertex3f(0.387442231f, 0.420979381f, -1.f),
    C_Vertex3f(0.376498699f, 0.43155241f, -1.f),
    C_Vertex3f(0.364475727f, 0.441120625f, -1.f),
    C_Vertex3f(0.35546875f, 0.315917969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323760033f, 0.380568504f, -1.f),
    C_Vertex3f(0.335520267f, 0.365829945f, -1.f),
    C_Vertex3f(0.337847352f, 0.456668973f, -1.f),
    C_Vertex3f(0.345180035f, 0.348187923f, -1.f),
    C_Vertex3f(0.351591945f, 0.449492812f, -1.f),
    C_Vertex3f(0.352739334f, 0.327642441f, -1.f),
    C_Vertex3f(0.364475727f, 0.441120625f, -1.f),
    C_Vertex3f(0.35546875f, 0.315917969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.214784861f, 0.406159043f, -1.f),
    C_Vertex3f(0.228237867f, 0.408823848f, -1.f),
    C_Vertex3f(0.231782913f, 0.474609375f, -1.f),
    C_Vertex3f(0.242386818f, 0.41015625f, -1.f),
    C_Vertex3f(0.256212234f, 0.474609375f, -1.f),
    C_Vertex3f(0.259634018f, 0.41015625f, -1.f),
    C_Vertex3f(0.274260879f, 0.473413348f, -1.f),
    C_Vertex3f(0.277836323f, 0.407197475f, -1.f),
    C_Vertex3f(0.291448712f, 0.471021295f, -1.f),
    C_Vertex3f(0.294591427f, 0.401279926f, -1.f),
    C_Vertex3f(0.307775736f, 0.467433214f, -1.f),
    C_Vertex3f(0.30989933f, 0.392403603f, -1.f),
    C_Vertex3f(0.323241949f, 0.462649107f, -1.f),
    C_Vertex3f(0.323760033f, 0.380568504f, -1.f),
    C_Vertex3f(0.337847352f, 0.456668973f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.202954769f, 0.471087456f, -1.f),
    C_Vertex3f(0.175247669f, 0.464043617f, -1.f),
    C_Vertex3f(0.178601503f, 0.390170217f, -1.f),
    C_Vertex3f(0.189966679f, 0.396832228f, -1.f),
    C_Vertex3f(0.202027798f, 0.402161837f, -1.f),
    C_Vertex3f(0.214784861f, 0.406159043f, -1.f),
    C_Vertex3f(0.231782913f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.175247669f, 0.464043617f, -1.f),
    C_Vertex3f(0.148661613f, 0.453477859f, -1.f),
    C_Vertex3f(0.148681641f, 0.36218977f, -1.f),
    C_Vertex3f(0.157958984f, 0.372848988f, -1.f),
    C_Vertex3f(0.167932272f, 0.382175803f, -1.f),
    C_Vertex3f(0.178601503f, 0.390170217f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.123168945f, 0.304042578f, -1.f),
    C_Vertex3f(0.127807617f, 0.320905924f, -1.f),
    C_Vertex3f(0.129562855f, 0.442912102f, -1.f),
    C_Vertex3f(0.133605957f, 0.336218238f, -1.f),
    C_Vertex3f(0.148661613f, 0.453477859f, -1.f),
    C_Vertex3f(0.140563965f, 0.34997952f, -1.f),
    C_Vertex3f(0.148681641f, 0.36218977f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116210938f, 0.244146347f, -1.f),
    C_Vertex3f(0.117370605f, 0.265662789f, -1.f),
    C_Vertex3f(0.117643118f, 0.434987903f, -1.f),
    C_Vertex3f(0.119689941f, 0.2856282f, -1.f),
    C_Vertex3f(0.129562855f, 0.442912102f, -1.f),
    C_Vertex3f(0.123168945f, 0.304042578f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115951657f, 0.0305939913f, -1.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, -1.f),
    C_Vertex3f(0.117331028f, 0.199118257f, -1.f),
    C_Vertex3f(0.116210938f, 0.220914841f, -1.f),
    C_Vertex3f(0.116210938f, 0.244146347f, -1.f),
    C_Vertex3f(0.10653615f, 0.426183343f, -1.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.10653615f, 0.426183343f, -1.f),
    C_Vertex3f(0.096241951f, 0.416498423f, -1.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, -1.f),
    C_Vertex3f(0.0867605209f, 0.405933142f, -1.f),
    C_Vertex3f(0.0858283043f, 0.0597710609f, -1.f),
    C_Vertex3f(0.0780918598f, 0.3944875f, -1.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, -1.f),
    C_Vertex3f(0.0702359676f, 0.382161498f, -1.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0460145473f, 0.138875246f, -1.f),
    C_Vertex3f(0.0532534122f, 0.116570234f, -1.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, -1.f),
    C_Vertex3f(0.0623019934f, 0.0959511995f, -1.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, -1.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.03515625f, 0.249214172f, -1.f),
    C_Vertex3f(0.0369659662f, 0.1885432f, -1.f),
    C_Vertex3f(0.0382714272f, 0.284432888f, -1.f),
    C_Vertex3f(0.0405853987f, 0.162866235f, -1.f),
    C_Vertex3f(0.0445017815f, 0.317303181f, -1.f),
    C_Vertex3f(0.0460145473f, 0.138875246f, -1.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.439453125f, 0.159667969f, -1.f),
    C_Vertex3f(0.362304688f, 0.169921875f, -1.f),
    C_Vertex3f(0.368021011f, 0.0292029381f, -1.f),
    C_Vertex3f(0.38107872f, 0.0404763222f, -1.f),
    C_Vertex3f(0.392900229f, 0.0528322458f, -1.f),
    C_Vertex3f(0.403485537f, 0.066270709f, -1.f),
    C_Vertex3f(0.412834644f, 0.0807917118f, -1.f),
    C_Vertex3f(0.420947552f, 0.0963952541f, -1.f),
    C_Vertex3f(0.427824259f, 0.113081336f, -1.f),
    C_Vertex3f(0.433464766f, 0.130849957f, -1.f),
    C_Vertex3f(0.437869072f, 0.149701118f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.368021011f, 0.0292029381f, -1.f),
    C_Vertex3f(0.362304688f, 0.169921875f, -1.f),
    C_Vertex3f(0.361266136f, 0.162552834f, -1.f),
    C_Vertex3f(0.358478904f, 0.148769975f, -1.f),
    C_Vertex3f(0.354981542f, 0.13594234f, -1.f),
    C_Vertex3f(0.354218721f, 0.019338727f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.354218721f, 0.019338727f, -1.f),
    C_Vertex3f(0.354981542f, 0.13594234f, -1.f),
    C_Vertex3f(0.35077405f, 0.124069929f, -1.f),
    C_Vertex3f(0.345856428f, 0.113152742f, -1.f),
    C_Vertex3f(0.340228677f, 0.10319078f, -1.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.340228677f, 0.10319078f, -1.f),
    C_Vertex3f(0.333890796f, 0.094184041f, -1.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, -1.f),
    C_Vertex3f(0.326842785f, 0.0861325264f, -1.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, -1.f),
    C_Vertex3f(0.319084644f, 0.0790362358f, -1.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.308344364f, -0.00179886818f, -1.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, -1.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, -1.f),
    C_Vertex3f(0.302133918f, 0.0675047636f, -1.f),
    C_Vertex3f(0.292941332f, 0.063069582f, -1.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.283270597f, 0.0595214367f, -1.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, -1.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, -1.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, -1.f),
    C_Vertex3f(0.292941332f, 0.063069582f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262494683f, 0.0550862551f, -1.f),
    C_Vertex3f(0.251389503f, 0.0541992188f, -1.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, -1.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, -1.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.211583853f, 0.058175683f, -1.f),
    C_Vertex3f(0.211698174f, -0.00830876827f, -1.f),
    C_Vertex3f(0.224642992f, 0.0555247068f, -1.f),
    C_Vertex3f(0.234351158f, -0.0102539063f, -1.f),
    C_Vertex3f(0.238385201f, 0.0541992188f, -1.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, -1.f),
    C_Vertex3f(0.251389503f, 0.0541992188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.199207783f, 0.0621521473f, -1.f),
    C_Vertex3f(0.187514782f, 0.0674540997f, -1.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, -1.f),
    C_Vertex3f(0.211698174f, -0.00830876827f, -1.f),
    C_Vertex3f(0.211583853f, 0.058175683f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.17650485f, 0.0740815401f, -1.f),
    C_Vertex3f(0.166177988f, 0.0820344687f, -1.f),
    C_Vertex3f(0.169895887f, 0.00141692162f, -1.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, -1.f),
    C_Vertex3f(0.187514782f, 0.0674540997f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156534195f, 0.0913128853f, -1.f),
    C_Vertex3f(0.147573471f, 0.10191679f, -1.f),
    C_Vertex3f(0.150746584f, 0.00919747353f, -1.f),
    C_Vertex3f(0.169895887f, 0.00141692162f, -1.f),
    C_Vertex3f(0.166177988f, 0.0820344687f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147573471f, 0.10191679f, -1.f),
    C_Vertex3f(0.139732838f, 0.114119649f, -1.f),
    C_Vertex3f(0.150746584f, 0.00919747353f, -1.f),
    C_Vertex3f(0.133012295f, 0.127921462f, -1.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, -1.f),
    C_Vertex3f(0.127411842f, 0.143322229f, -1.f),
    C_Vertex3f(0.12293148f, 0.160321951f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.119571209f, 0.178920627f, -1.f),
    C_Vertex3f(0.117331028f, 0.199118257f, -1.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, -1.f),
    C_Vertex3f(0.12293148f, 0.160321951f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0369659662f, 0.1885432f, -1.f),
    C_Vertex3f(0.03515625f, 0.249214172f, -1.f),
    C_Vertex3f(0.03515625f, 0.215906143f, -1.f),
    C_Vertex3f(0.10653615f, 0.426183343f, -1.f),
    C_Vertex3f(0.116210938f, 0.244146347f, -1.f),
    C_Vertex3f(0.117643118f, 0.434987903f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_100[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.360351563f, 0.f, 0.f),
    C_Vertex3f(0.360351563f, 0.f, -1.f),
    C_Vertex3f(0.360351563f, 0.058555603f, -1.f),
    C_Vertex3f(0.360351563f, 0.058555603f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.841642022f, -0.540035784f, 0.f),
    C_Vertex3f(0.360351563f, 0.058555603f, 0.f),
    C_Vertex3f(0.360351563f, 0.058555603f, -1.f),
    C_Normal3f(0.798145533f, -0.602464736f, 0.f),
    C_Vertex3f(0.354832649f, 0.0499544144f, 0.f),
    C_Vertex3f(0.354832649f, 0.0499544144f, -1.f),
    C_Normal3f(0.732650757f, -0.680604696f, 0.f),
    C_Vertex3f(0.342497349f, 0.0349023342f, 0.f),
    C_Vertex3f(0.342497349f, 0.0349023342f, -1.f),
    C_Normal3f(0.637806177f, -0.770196855f, 0.f),
    C_Vertex3f(0.328864574f, 0.0220005512f, 0.f),
    C_Vertex3f(0.328864574f, 0.0220005512f, -1.f),
    C_Normal3f(0.527623475f, -0.849478304f, 0.f),
    C_Vertex3f(0.313934326f, 0.0112490654f, 0.f),
    C_Vertex3f(0.313934326f, 0.0112490654f, -1.f),
    C_Normal3f(0.407285631f, -0.913300931f, 0.f),
    C_Vertex3f(0.297706604f, 0.00264787674f, 0.f),
    C_Vertex3f(0.297706604f, 0.00264787674f, -1.f),
    C_Normal3f(0.28364563f, -0.958929121f, 0.f),
    C_Vertex3f(0.280181408f, -0.00380301476f, 0.f),
    C_Vertex3f(0.280181408f, -0.00380301476f, -1.f),
    C_Normal3f(0.163423359f, -0.986555994f, 0.f),
    C_Vertex3f(0.261358738f, -0.00810360909f, 0.f),
    C_Vertex3f(0.261358738f, -0.00810360909f, -1.f),
    C_Normal3f(0.0568410531f, -0.998383284f, 0.f),
    C_Vertex3f(0.241238594f, -0.0102539063f, 0.f),
    C_Vertex3f(0.241238594f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0305659585f, -0.999532759f, 0.f),
    C_Vertex3f(0.223589897f, -0.0102539063f, 0.f),
    C_Vertex3f(0.223589897f, -0.0102539063f, -1.f),
    C_Normal3f(-0.105895929f, -0.994377196f, 0.f),
    C_Vertex3f(0.209990263f, -0.00929832458f, 0.f),
    C_Vertex3f(0.209990263f, -0.00929832458f, -1.f),
    C_Normal3f(-0.178357527f, -0.983965695f, 0.f),
    C_Vertex3f(0.196670771f, -0.00738716125f, 0.f),
    C_Vertex3f(0.196670771f, -0.00738716125f, -1.f),
    C_Normal3f(-0.250981718f, -0.967991889f, 0.f),
    C_Vertex3f(0.18363142f, -0.00452041626f, 0.f),
    C_Vertex3f(0.18363142f, -0.00452041626f, -1.f),
    C_Normal3f(-0.322548866f, -0.946552873f, 0.f),
    C_Vertex3f(0.170872211f, -0.0006980896f, 0.f),
    C_Vertex3f(0.170872211f, -0.0006980896f, -1.f),
    C_Normal3f(-0.391874731f, -0.920018554f, 0.f),
    C_Vertex3f(0.158393145f, 0.00407981873f, 0.f),
    C_Vertex3f(0.158393145f, 0.00407981873f, -1.f),
    C_Normal3f(-0.457906723f, -0.889000297f, 0.f),
    C_Vertex3f(0.14619422f, 0.00981330872f, 0.f),
    C_Vertex3f(0.14619422f, 0.00981330872f, -1.f),
    C_Normal3f(-0.519799352f, -0.85428834f, 0.f),
    C_Vertex3f(0.134275436f, 0.0165023804f, 0.f),
    C_Vertex3f(0.134275436f, 0.0165023804f, -1.f),
    C_Normal3f(-0.578574061f, -0.815629959f, 0.f),
    C_Vertex3f(0.122636795f, 0.0241470337f, 0.f),
    C_Vertex3f(0.122636795f, 0.0241470337f, -1.f),
    C_Normal3f(-0.635972321f, -0.771711826f, 0.f),
    C_Vertex3f(0.111653924f, 0.0325491428f, 0.f),
    C_Vertex3f(0.111653924f, 0.0325491428f, -1.f),
    C_Normal3f(-0.690232873f, -0.723587275f, 0.f),
    C_Vertex3f(0.101326823f, 0.0417087078f, 0.f),
    C_Vertex3f(0.101326823f, 0.0417087078f, -1.f),
    C_Normal3f(-0.740524292f, -0.672029614f, 0.f),
    C_Vertex3f(0.0916554928f, 0.0516257286f, 0.f),
    C_Vertex3f(0.0916554928f, 0.0516257286f, -1.f),
    C_Normal3f(-0.786214173f, -0.617954075f, 0.f),
    C_Vertex3f(0.0826399326f, 0.0623002052f, 0.f),
    C_Vertex3f(0.0826399326f, 0.0623002052f, -1.f),
    C_Normal3f(-0.826905191f, -0.562341452f, 0.f),
    C_Vertex3f(0.0742801428f, 0.0737321377f, 0.f),
    C_Vertex3f(0.0742801428f, 0.0737321377f, -1.f),
    C_Normal3f(-0.862439871f, -0.506159604f, 0.f),
    C_Vertex3f(0.0665761232f, 0.085921526f, 0.f),
    C_Vertex3f(0.0665761232f, 0.085921526f, -1.f),
    C_Normal3f(-0.895650148f, -0.444759309f, 0.f),
    C_Vertex3f(0.059527874f, 0.0988683701f, 0.f),
    C_Vertex3f(0.059527874f, 0.0988683701f, -1.f),
    C_Normal3f(-0.93371135f, -0.358026832f, 0.f),
    C_Vertex3f(0.0499391556f, 0.11942482f, 0.f),
    C_Vertex3f(0.0499391556f, 0.11942482f, -1.f),
    C_Normal3f(-0.967314661f, -0.253578871f, 0.f),
    C_Vertex3f(0.0403504372f, 0.148881912f, 0.f),
    C_Vertex3f(0.0403504372f, 0.148881912f, -1.f),
    C_Normal3f(-0.989312887f, -0.145808354f, 0.f),
    C_Vertex3f(0.0339579582f, 0.180387497f, 0.f),
    C_Vertex3f(0.0339579582f, 0.180387497f, -1.f),
    C_Normal3f(-0.998920083f, -0.0464622043f, 0.f),
    C_Vertex3f(0.0307617188f, 0.213941574f, 0.f),
    C_Vertex3f(0.0307617188f, 0.213941574f, -1.f),
    C_Normal3f(-0.999105036f, 0.042297177f, 0.f),
    C_Vertex3f(0.0307617188f, 0.249105453f, 0.f),
    C_Vertex3f(0.0307617188f, 0.249105453f, -1.f),
    C_Normal3f(-0.991115391f, 0.133004889f, 0.f),
    C_Vertex3f(0.0336523056f, 0.282220364f, 0.f),
    C_Vertex3f(0.0336523056f, 0.282220364f, -1.f),
    C_Normal3f(-0.973395526f, 0.229131177f, 0.f),
    C_Vertex3f(0.0394334793f, 0.313724995f, 0.f),
    C_Vertex3f(0.0394334793f, 0.313724995f, -1.f),
    C_Normal3f(-0.946958959f, 0.321354657f, 0.f),
    C_Vertex3f(0.0481052399f, 0.343619347f, 0.f),
    C_Vertex3f(0.0481052399f, 0.343619347f, -1.f),
    C_Normal3f(-0.915484905f, 0.402352303f, 0.f),
    C_Vertex3f(0.0567770004f, 0.364832401f, 0.f),
    C_Vertex3f(0.0567770004f, 0.364832401f, -1.f),
    C_Normal3f(-0.882828295f, 0.469695807f, 0.f),
    C_Vertex3f(0.0632807016f, 0.378148556f, 0.f),
    C_Vertex3f(0.0632807016f, 0.378148556f, -1.f),
    C_Normal3f(-0.846762002f, 0.531971812f, 0.f),
    C_Vertex3f(0.0705069304f, 0.390638828f, 0.f),
    C_Vertex3f(0.0705069304f, 0.390638828f, -1.f),
    C_Normal3f(-0.804390311f, 0.594101191f, 0.f),
    C_Vertex3f(0.0784556866f, 0.402303219f, 0.f),
    C_Vertex3f(0.0784556866f, 0.402303219f, -1.f),
    C_Normal3f(-0.755794704f, 0.65480864f, 0.f),
    C_Vertex3f(0.0871269703f, 0.413141727f, 0.f),
    C_Vertex3f(0.0871269703f, 0.413141727f, -1.f),
    C_Normal3f(-0.701407194f, 0.712760806f, 0.f),
    C_Vertex3f(0.0965207815f, 0.423154354f, 0.f),
    C_Vertex3f(0.0965207815f, 0.423154354f, -1.f),
    C_Normal3f(-0.64201653f, 0.766690731f, 0.f),
    C_Vertex3f(0.10663712f, 0.432341099f, 0.f),
    C_Vertex3f(0.10663712f, 0.432341099f, -1.f),
    C_Normal3f(-0.578719974f, 0.815526307f, 0.f),
    C_Vertex3f(0.117475986f, 0.440701962f, 0.f),
    C_Vertex3f(0.117475986f, 0.440701962f, -1.f),
    C_Normal3f(-0.51583159f, 0.85668999f, 0.f),
    C_Vertex3f(0.12903738f, 0.448236942f, 0.f),
    C_Vertex3f(0.12903738f, 0.448236942f, -1.f),
    C_Normal3f(-0.45227018f, 0.891881049f, 0.f),
    C_Vertex3f(0.140939832f, 0.45483005f, 0.f),
    C_Vertex3f(0.140939832f, 0.45483005f, -1.f),
    C_Normal3f(-0.385108382f, 0.922871411f, 0.f),
    C_Vertex3f(0.153183341f, 0.460481286f, 0.f),
    C_Vertex3f(0.153183341f, 0.460481286f, -1.f),
    C_Normal3f(-0.315338939f, 0.948979139f, 0.f),
    C_Vertex3f(0.165767908f, 0.465190649f, 0.f),
    C_Vertex3f(0.165767908f, 0.465190649f, -1.f),
    C_Normal3f(-0.24410449f, 0.969748914f, 0.f),
    C_Vertex3f(0.178693533f, 0.468958139f, 0.f),
    C_Vertex3f(0.178693533f, 0.468958139f, -1.f),
    C_Normal3f(-0.172605753f, 0.984991014f, 0.f),
    C_Vertex3f(0.191960216f, 0.471783757f, 0.f),
    C_Vertex3f(0.191960216f, 0.471783757f, -1.f),
    C_Normal3f(-0.102004074f, 0.994783938f, 0.f),
    C_Vertex3f(0.205567956f, 0.473667502f, 0.f),
    C_Vertex3f(0.205567956f, 0.473667502f, -1.f),
    C_Normal3f(-0.0298276022f, 0.999554992f, 0.f),
    C_Vertex3f(0.219516754f, 0.474609375f, 0.f),
    C_Vertex3f(0.219516754f, 0.474609375f, -1.f),
    C_Normal3f(0.0588891059f, 0.998264492f, 0.f),
    C_Vertex3f(0.237131119f, 0.474609375f, 0.f),
    C_Vertex3f(0.237131119f, 0.474609375f, -1.f),
    C_Normal3f(0.169708803f, 0.985494256f, 0.f),
    C_Vertex3f(0.256925583f, 0.472402573f, 0.f),
    C_Vertex3f(0.256925583f, 0.472402573f, -1.f),
    C_Normal3f(0.291890353f, 0.956451774f, 0.f),
    C_Vertex3f(0.275575638f, 0.467988968f, 0.f),
    C_Vertex3f(0.275575638f, 0.467988968f, -1.f),
    C_Normal3f(0.414995462f, 0.909823358f, 0.f),
    C_Vertex3f(0.293081284f, 0.461368561f, 0.f),
    C_Vertex3f(0.293081284f, 0.461368561f, -1.f),
    C_Normal3f(0.519946158f, 0.854199111f, 0.f),
    C_Vertex3f(0.30944252f, 0.452541351f, 0.f),
    C_Vertex3f(0.30944252f, 0.452541351f, -1.f),
    C_Normal3f(0.60987556f, 0.792497158f, 0.f),
    C_Vertex3f(0.324276924f, 0.442379951f, 0.f),
    C_Vertex3f(0.324276924f, 0.442379951f, -1.f),
    C_Normal3f(0.696106136f, 0.717938781f, 0.f),
    C_Vertex3f(0.337584496f, 0.430884361f, 0.f),
    C_Vertex3f(0.337584496f, 0.430884361f, -1.f),
    C_Normal3f(0.762266219f, 0.647263646f, 0.f),
    C_Vertex3f(0.349365234f, 0.418054581f, 0.f),
    C_Vertex3f(0.349365234f, 0.418054581f, -1.f),
    C_Normal3f(0.810017169f, 0.586406171f, 0.f),
    C_Vertex3f(0.354492188f, 0.410972595f, 0.f),
    C_Vertex3f(0.354492188f, 0.410972595f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.410972595f, 0.f),
    C_Vertex3f(0.354492188f, 0.410972595f, -1.f),
    C_Vertex3f(0.354492188f, 0.640625f, -1.f),
    C_Vertex3f(0.354492188f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.354492188f, 0.640625f, 0.f),
    C_Vertex3f(0.354492188f, 0.640625f, -1.f),
    C_Vertex3f(0.433105469f, 0.640625f, -1.f),
    C_Vertex3f(0.433105469f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.640625f, 0.f),
    C_Vertex3f(0.433105469f, 0.640625f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.360351563f, 0.f, -1.f),
    C_Vertex3f(0.360351563f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999690771f, -0.0248665623f, 0.f),
    C_Vertex3f(0.111816406f, 0.243244171f, 0.f),
    C_Vertex3f(0.111816406f, 0.243244171f, -1.f),
    C_Normal3f(0.999636889f, 0.0269467719f, 0.f),
    C_Vertex3f(0.111816406f, 0.220562935f, 0.f),
    C_Vertex3f(0.111816406f, 0.220562935f, -1.f),
    C_Normal3f(0.996213257f, 0.0869429857f, 0.f),
    C_Vertex3f(0.112990975f, 0.199671507f, 0.f),
    C_Vertex3f(0.112990975f, 0.199671507f, -1.f),
    C_Normal3f(0.987995982f, 0.154479221f, 0.f),
    C_Vertex3f(0.115340114f, 0.180187464f, 0.f),
    C_Vertex3f(0.115340114f, 0.180187464f, -1.f),
    C_Normal3f(0.973126173f, 0.230272368f, 0.f),
    C_Vertex3f(0.118863821f, 0.162110806f, 0.f),
    C_Vertex3f(0.118863821f, 0.162110806f, -1.f),
    C_Normal3f(0.94932884f, 0.314284563f, 0.f),
    C_Vertex3f(0.123562098f, 0.145441532f, 0.f),
    C_Vertex3f(0.123562098f, 0.145441532f, -1.f),
    C_Normal3f(0.914048672f, 0.40560475f, 0.f),
    C_Vertex3f(0.129434943f, 0.130179644f, 0.f),
    C_Vertex3f(0.129434943f, 0.130179644f, -1.f),
    C_Normal3f(0.864824891f, 0.502073824f, 0.f),
    C_Vertex3f(0.136482358f, 0.11632514f, 0.f),
    C_Vertex3f(0.136482358f, 0.11632514f, -1.f),
    C_Normal3f(0.799944401f, 0.600073993f, 0.f),
    C_Vertex3f(0.144704342f, 0.103878021f, 0.f),
    C_Vertex3f(0.144704342f, 0.103878021f, -1.f),
    C_Normal3f(0.732869506f, 0.680369139f, 0.f),
    C_Vertex3f(0.154100895f, 0.0928382874f, 0.f),
    C_Vertex3f(0.154100895f, 0.0928382874f, -1.f),
    C_Normal3f(0.666542768f, 0.745466828f, 0.f),
    C_Vertex3f(0.163920999f, 0.0831785202f, 0.f),
    C_Vertex3f(0.163920999f, 0.0831785202f, -1.f),
    C_Normal3f(0.587456107f, 0.809256077f, 0.f),
    C_Vertex3f(0.174164653f, 0.0748987198f, 0.f),
    C_Vertex3f(0.174164653f, 0.0748987198f, -1.f),
    C_Normal3f(0.495735109f, 0.868473828f, 0.f),
    C_Vertex3f(0.184831858f, 0.0679988861f, 0.f),
    C_Vertex3f(0.184831858f, 0.0679988861f, -1.f),
    C_Normal3f(0.3929528f, 0.919558644f, 0.f),
    C_Vertex3f(0.195922613f, 0.0624790192f, 0.f),
    C_Vertex3f(0.195922613f, 0.0624790192f, -1.f),
    C_Normal3f(0.282246858f, 0.959341824f, 0.f),
    C_Vertex3f(0.207436919f, 0.058339119f, 0.f),
    C_Vertex3f(0.207436919f, 0.058339119f, -1.f),
    C_Normal3f(0.167951345f, 0.985795259f, 0.f),
    C_Vertex3f(0.219374776f, 0.0555791855f, 0.f),
    C_Vertex3f(0.219374776f, 0.0555791855f, -1.f),
    C_Normal3f(0.0546770953f, 0.998504102f, 0.f),
    C_Vertex3f(0.231736183f, 0.0541992188f, 0.f),
    C_Vertex3f(0.231736183f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0521537066f, 0.998639047f, 0.f),
    C_Vertex3f(0.2445755f, 0.0541992188f, 0.f),
    C_Vertex3f(0.2445755f, 0.0541992188f, -1.f),
    C_Normal3f(-0.160462737f, 0.987041891f, 0.f),
    C_Vertex3f(0.256984234f, 0.0555177927f, 0.f),
    C_Vertex3f(0.256984234f, 0.0555177927f, -1.f),
    C_Normal3f(-0.271590233f, 0.962413013f, 0.f),
    C_Vertex3f(0.268908024f, 0.0581549406f, 0.f),
    C_Vertex3f(0.268908024f, 0.0581549406f, -1.f),
    C_Normal3f(-0.381084204f, 0.924540281f, 0.f),
    C_Vertex3f(0.28034687f, 0.0621106625f, 0.f),
    C_Vertex3f(0.28034687f, 0.0621106625f, -1.f),
    C_Normal3f(-0.484568447f, 0.874753416f, 0.f),
    C_Vertex3f(0.291300774f, 0.0673849583f, 0.f),
    C_Vertex3f(0.291300774f, 0.0673849583f, -1.f),
    C_Normal3f(-0.578463078f, 0.815708458f, 0.f),
    C_Vertex3f(0.301769733f, 0.073977828f, 0.f),
    C_Vertex3f(0.301769733f, 0.073977828f, -1.f),
    C_Normal3f(-0.660548508f, 0.750783443f, 0.f),
    C_Vertex3f(0.31175375f, 0.0818892717f, 0.f),
    C_Vertex3f(0.31175375f, 0.0818892717f, -1.f),
    C_Normal3f(-0.730075538f, 0.683366358f, 0.f),
    C_Vertex3f(0.321252823f, 0.0911192894f, 0.f),
    C_Vertex3f(0.321252823f, 0.0911192894f, -1.f),
    C_Normal3f(-0.799341977f, 0.600876451f, 0.f),
    C_Vertex3f(0.330266953f, 0.101667881f, 0.f),
    C_Vertex3f(0.330266953f, 0.101667881f, -1.f),
    C_Normal3f(-0.865051627f, 0.501682997f, 0.f),
    C_Vertex3f(0.338154316f, 0.113603234f, 0.f),
    C_Vertex3f(0.338154316f, 0.113603234f, -1.f),
    C_Normal3f(-0.914582431f, 0.404399753f, 0.f),
    C_Vertex3f(0.344914913f, 0.126925349f, 0.f),
    C_Vertex3f(0.344914913f, 0.126925349f, -1.f),
    C_Normal3f(-0.949854076f, 0.312693477f, 0.f),
    C_Vertex3f(0.350548744f, 0.141634226f, 0.f),
    C_Vertex3f(0.350548744f, 0.141634226f, -1.f),
    C_Normal3f(-0.973503351f, 0.228672519f, 0.f),
    C_Vertex3f(0.355055809f, 0.157729864f, 0.f),
    C_Vertex3f(0.355055809f, 0.157729864f, -1.f),
    C_Normal3f(-0.988202393f, 0.153153405f, 0.f),
    C_Vertex3f(0.358436108f, 0.175212264f, 0.f),
    C_Vertex3f(0.358436108f, 0.175212264f, -1.f),
    C_Normal3f(-0.996288598f, 0.0860765651f, 0.f),
    C_Vertex3f(0.36068964f, 0.194081426f, 0.f),
    C_Vertex3f(0.36068964f, 0.194081426f, -1.f),
    C_Normal3f(-0.999656737f, 0.0262000374f, 0.f),
    C_Vertex3f(0.361816406f, 0.214337349f, 0.f),
    C_Vertex3f(0.361816406f, 0.214337349f, -1.f),
    C_Normal3f(-0.999674916f, -0.0254983827f, 0.f),
    C_Vertex3f(0.361816406f, 0.237072945f, 0.f),
    C_Vertex3f(0.361816406f, 0.237072945f, -1.f),
    C_Normal3f(-0.996797502f, -0.0799673721f, 0.f),
    C_Vertex3f(0.360669136f, 0.259316564f, 0.f),
    C_Vertex3f(0.360669136f, 0.259316564f, -1.f),
    C_Normal3f(-0.989738226f, -0.142892465f, 0.f),
    C_Vertex3f(0.358374596f, 0.279975295f, 0.f),
    C_Vertex3f(0.358374596f, 0.279975295f, -1.f),
    C_Normal3f(-0.976716876f, -0.21453245f, 0.f),
    C_Vertex3f(0.354932785f, 0.299049139f, 0.f),
    C_Vertex3f(0.354932785f, 0.299049139f, -1.f),
    C_Normal3f(-0.955370724f, -0.295409441f, 0.f),
    C_Vertex3f(0.350343704f, 0.316538095f, 0.f),
    C_Vertex3f(0.350343704f, 0.316538095f, -1.f),
    C_Normal3f(-0.922784507f, -0.385316432f, 0.f),
    C_Vertex3f(0.344607353f, 0.332442164f, 0.f),
    C_Vertex3f(0.344607353f, 0.332442164f, -1.f),
    C_Normal3f(-0.875735641f, -0.482790828f, 0.f),
    C_Vertex3f(0.337723732f, 0.346761346f, 0.f),
    C_Vertex3f(0.337723732f, 0.346761346f, -1.f),
    C_Normal3f(-0.811328232f, -0.584590793f, 0.f),
    C_Vertex3f(0.329692841f, 0.35949564f, 0.f),
    C_Vertex3f(0.329692841f, 0.35949564f, -1.f),
    C_Normal3f(-0.74198103f, -0.670421004f, 0.f),
    C_Vertex3f(0.320514679f, 0.370645046f, 0.f),
    C_Vertex3f(0.320514679f, 0.370645046f, -1.f),
    C_Normal3f(-0.672217548f, -0.740353584f, 0.f),
    C_Vertex3f(0.31080389f, 0.380400777f, 0.f),
    C_Vertex3f(0.31080389f, 0.380400777f, -1.f),
    C_Normal3f(-0.589266121f, -0.807938993f, 0.f),
    C_Vertex3f(0.300560474f, 0.388762832f, 0.f),
    C_Vertex3f(0.300560474f, 0.388762832f, -1.f),
    C_Normal3f(-0.493859082f, -0.869541943f, 0.f),
    C_Vertex3f(0.289784431f, 0.395731211f, 0.f),
    C_Vertex3f(0.289784431f, 0.395731211f, -1.f),
    C_Normal3f(-0.388340235f, -0.92151618f, 0.f),
    C_Vertex3f(0.278475761f, 0.401305914f, 0.f),
    C_Vertex3f(0.278475761f, 0.401305914f, -1.f),
    C_Normal3f(-0.276545793f, -0.961000741f, 0.f),
    C_Vertex3f(0.266634464f, 0.405486941f, 0.f),
    C_Vertex3f(0.266634464f, 0.405486941f, -1.f),
    C_Normal3f(-0.163169146f, -0.986598074f, 0.f),
    C_Vertex3f(0.25426054f, 0.408274293f, 0.f),
    C_Vertex3f(0.25426054f, 0.408274293f, -1.f),
    C_Normal3f(-0.0531537309f, -0.998586357f, 0.f),
    C_Vertex3f(0.241353989f, 0.409667969f, 0.f),
    C_Vertex3f(0.241353989f, 0.409667969f, -1.f),
    C_Normal3f(0.0517335646f, -0.998660862f, 0.f),
    C_Vertex3f(0.228077888f, 0.409667969f, 0.f),
    C_Vertex3f(0.228077888f, 0.409667969f, -1.f),
    C_Normal3f(0.161122307f, -0.986934483f, 0.f),
    C_Vertex3f(0.215504527f, 0.408328891f, 0.f),
    C_Vertex3f(0.215504527f, 0.408328891f, -1.f),
    C_Normal3f(0.273707241f, -0.961813033f, 0.f),
    C_Vertex3f(0.203470826f, 0.405650735f, 0.f),
    C_Vertex3f(0.203470826f, 0.405650735f, -1.f),
    C_Normal3f(0.385317087f, -0.922784209f, 0.f),
    C_Vertex3f(0.191976786f, 0.401633501f, 0.f),
    C_Vertex3f(0.191976786f, 0.401633501f, -1.f),
    C_Normal3f(0.491238832f, -0.871024847f, 0.f),
    C_Vertex3f(0.181022406f, 0.396277189f, 0.f),
    C_Vertex3f(0.181022406f, 0.396277189f, -1.f),
    C_Normal3f(0.587484241f, -0.809235632f, 0.f),
    C_Vertex3f(0.170607686f, 0.3895818f, 0.f),
    C_Vertex3f(0.170607686f, 0.3895818f, -1.f),
    C_Normal3f(0.671488404f, -0.741015077f, 0.f),
    C_Vertex3f(0.160732627f, 0.381547332f, 0.f),
    C_Vertex3f(0.160732627f, 0.381547332f, -1.f),
    C_Normal3f(0.742308736f, -0.670058131f, 0.f),
    C_Vertex3f(0.151397228f, 0.372173786f, 0.f),
    C_Vertex3f(0.151397228f, 0.372173786f, -1.f),
    C_Normal3f(0.81230104f, -0.583238363f, 0.f),
    C_Vertex3f(0.14260149f, 0.361461163f, 0.f),
    C_Vertex3f(0.14260149f, 0.361461163f, -1.f),
    C_Normal3f(0.876758277f, -0.480931282f, 0.f),
    C_Vertex3f(0.134905219f, 0.349204659f, 0.f),
    C_Vertex3f(0.134905219f, 0.349204659f, -1.f),
    C_Normal3f(0.923632443f, -0.383279324f, 0.f),
    C_Vertex3f(0.128308415f, 0.335404277f, 0.f),
    C_Vertex3f(0.128308415f, 0.335404277f, -1.f),
    C_Normal3f(0.955965996f, -0.293477327f, 0.f),
    C_Vertex3f(0.122811079f, 0.320060015f, 0.f),
    C_Vertex3f(0.122811079f, 0.320060015f, -1.f),
    C_Normal3f(0.977073014f, -0.212904409f, 0.f),
    C_Vertex3f(0.11841321f, 0.303171873f, 0.f),
    C_Vertex3f(0.11841321f, 0.303171873f, -1.f),
    C_Normal3f(0.989911556f, -0.14168632f, 0.f),
    C_Vertex3f(0.115114808f, 0.284739852f, 0.f),
    C_Vertex3f(0.115114808f, 0.284739852f, -1.f),
    C_Normal3f(0.996855736f, -0.0792379081f, 0.f),
    C_Vertex3f(0.112915874f, 0.264763951f, 0.f),
    C_Vertex3f(0.112915874f, 0.264763951f, -1.f),
    C_Normal3f(0.999690771f, -0.0248665623f, 0.f),
    C_Vertex3f(0.111816406f, 0.243244171f, 0.f),
    C_Vertex3f(0.111816406f, 0.243244171f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360351563f, 0.058555603f, 0.f),
    C_Vertex3f(0.360351563f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.36068964f, 0.194081426f, 0.f),
    C_Vertex3f(0.358436108f, 0.175212264f, 0.f),
    C_Vertex3f(0.355055809f, 0.157729864f, 0.f),
    C_Vertex3f(0.354832649f, 0.0499544144f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355055809f, 0.157729864f, 0.f),
    C_Vertex3f(0.350548744f, 0.141634226f, 0.f),
    C_Vertex3f(0.354832649f, 0.0499544144f, 0.f),
    C_Vertex3f(0.344914913f, 0.126925349f, 0.f),
    C_Vertex3f(0.342497349f, 0.0349023342f, 0.f),
    C_Vertex3f(0.338154316f, 0.113603234f, 0.f),
    C_Vertex3f(0.330266953f, 0.101667881f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.342497349f, 0.0349023342f, 0.f),
    C_Vertex3f(0.330266953f, 0.101667881f, 0.f),
    C_Vertex3f(0.328864574f, 0.0220005512f, 0.f),
    C_Vertex3f(0.321252823f, 0.0911192894f, 0.f),
    C_Vertex3f(0.313934326f, 0.0112490654f, 0.f),
    C_Vertex3f(0.31175375f, 0.0818892717f, 0.f),
    C_Vertex3f(0.301769733f, 0.073977828f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.297706604f, 0.00264787674f, 0.f),
    C_Vertex3f(0.313934326f, 0.0112490654f, 0.f),
    C_Vertex3f(0.301769733f, 0.073977828f, 0.f),
    C_Vertex3f(0.291300774f, 0.0673849583f, 0.f),
    C_Vertex3f(0.28034687f, 0.0621106625f, 0.f),
    C_Vertex3f(0.280181408f, -0.00380301476f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.268908024f, 0.0581549406f, 0.f),
    C_Vertex3f(0.256984234f, 0.0555177927f, 0.f),
    C_Vertex3f(0.261358738f, -0.00810360909f, 0.f),
    C_Vertex3f(0.280181408f, -0.00380301476f, 0.f),
    C_Vertex3f(0.28034687f, 0.0621106625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.195922613f, 0.0624790192f, 0.f),
    C_Vertex3f(0.196670771f, -0.00738716125f, 0.f),
    C_Vertex3f(0.207436919f, 0.058339119f, 0.f),
    C_Vertex3f(0.209990263f, -0.00929832458f, 0.f),
    C_Vertex3f(0.219374776f, 0.0555791855f, 0.f),
    C_Vertex3f(0.223589897f, -0.0102539063f, 0.f),
    C_Vertex3f(0.231736183f, 0.0541992188f, 0.f),
    C_Vertex3f(0.241238594f, -0.0102539063f, 0.f),
    C_Vertex3f(0.2445755f, 0.0541992188f, 0.f),
    C_Vertex3f(0.261358738f, -0.00810360909f, 0.f),
    C_Vertex3f(0.256984234f, 0.0555177927f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.144704342f, 0.103878021f, 0.f),
    C_Vertex3f(0.14619422f, 0.00981330872f, 0.f),
    C_Vertex3f(0.154100895f, 0.0928382874f, 0.f),
    C_Vertex3f(0.158393145f, 0.00407981873f, 0.f),
    C_Vertex3f(0.163920999f, 0.0831785202f, 0.f),
    C_Vertex3f(0.170872211f, -0.0006980896f, 0.f),
    C_Vertex3f(0.174164653f, 0.0748987198f, 0.f),
    C_Vertex3f(0.18363142f, -0.00452041626f, 0.f),
    C_Vertex3f(0.184831858f, 0.0679988861f, 0.f),
    C_Vertex3f(0.196670771f, -0.00738716125f, 0.f),
    C_Vertex3f(0.195922613f, 0.0624790192f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136482358f, 0.11632514f, 0.f),
    C_Vertex3f(0.129434943f, 0.130179644f, 0.f),
    C_Vertex3f(0.134275436f, 0.0165023804f, 0.f),
    C_Vertex3f(0.14619422f, 0.00981330872f, 0.f),
    C_Vertex3f(0.144704342f, 0.103878021f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123562098f, 0.145441532f, 0.f),
    C_Vertex3f(0.118863821f, 0.162110806f, 0.f),
    C_Vertex3f(0.122636795f, 0.0241470337f, 0.f),
    C_Vertex3f(0.134275436f, 0.0165023804f, 0.f),
    C_Vertex3f(0.129434943f, 0.130179644f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122636795f, 0.0241470337f, 0.f),
    C_Vertex3f(0.118863821f, 0.162110806f, 0.f),
    C_Vertex3f(0.115340114f, 0.180187464f, 0.f),
    C_Vertex3f(0.112990975f, 0.199671507f, 0.f),
    C_Vertex3f(0.111816406f, 0.220562935f, 0.f),
    C_Vertex3f(0.111653924f, 0.0325491428f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.433105469f, 0.640625f, 0.f),
    C_Vertex3f(0.354492188f, 0.640625f, 0.f),
    C_Vertex3f(0.354932785f, 0.299049139f, 0.f),
    C_Vertex3f(0.358374596f, 0.279975295f, 0.f),
    C_Vertex3f(0.360669136f, 0.259316564f, 0.f),
    C_Vertex3f(0.361816406f, 0.237072945f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.337723732f, 0.346761346f, 0.f),
    C_Vertex3f(0.344607353f, 0.332442164f, 0.f),
    C_Vertex3f(0.349365234f, 0.418054581f, 0.f),
    C_Vertex3f(0.350343704f, 0.316538095f, 0.f),
    C_Vertex3f(0.354492188f, 0.410972595f, 0.f),
    C_Vertex3f(0.354932785f, 0.299049139f, 0.f),
    C_Vertex3f(0.354492188f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.31080389f, 0.380400777f, 0.f),
    C_Vertex3f(0.320514679f, 0.370645046f, 0.f),
    C_Vertex3f(0.324276924f, 0.442379951f, 0.f),
    C_Vertex3f(0.329692841f, 0.35949564f, 0.f),
    C_Vertex3f(0.337584496f, 0.430884361f, 0.f),
    C_Vertex3f(0.337723732f, 0.346761346f, 0.f),
    C_Vertex3f(0.349365234f, 0.418054581f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278475761f, 0.401305914f, 0.f),
    C_Vertex3f(0.289784431f, 0.395731211f, 0.f),
    C_Vertex3f(0.293081284f, 0.461368561f, 0.f),
    C_Vertex3f(0.300560474f, 0.388762832f, 0.f),
    C_Vertex3f(0.30944252f, 0.452541351f, 0.f),
    C_Vertex3f(0.31080389f, 0.380400777f, 0.f),
    C_Vertex3f(0.324276924f, 0.442379951f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241353989f, 0.409667969f, 0.f),
    C_Vertex3f(0.25426054f, 0.408274293f, 0.f),
    C_Vertex3f(0.256925583f, 0.472402573f, 0.f),
    C_Vertex3f(0.266634464f, 0.405486941f, 0.f),
    C_Vertex3f(0.275575638f, 0.467988968f, 0.f),
    C_Vertex3f(0.278475761f, 0.401305914f, 0.f),
    C_Vertex3f(0.293081284f, 0.461368561f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.191976786f, 0.401633501f, 0.f),
    C_Vertex3f(0.203470826f, 0.405650735f, 0.f),
    C_Vertex3f(0.205567956f, 0.473667502f, 0.f),
    C_Vertex3f(0.215504527f, 0.408328891f, 0.f),
    C_Vertex3f(0.219516754f, 0.474609375f, 0.f),
    C_Vertex3f(0.228077888f, 0.409667969f, 0.f),
    C_Vertex3f(0.237131119f, 0.474609375f, 0.f),
    C_Vertex3f(0.241353989f, 0.409667969f, 0.f),
    C_Vertex3f(0.256925583f, 0.472402573f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.14260149f, 0.361461163f, 0.f),
    C_Vertex3f(0.151397228f, 0.372173786f, 0.f),
    C_Vertex3f(0.153183341f, 0.460481286f, 0.f),
    C_Vertex3f(0.160732627f, 0.381547332f, 0.f),
    C_Vertex3f(0.165767908f, 0.465190649f, 0.f),
    C_Vertex3f(0.170607686f, 0.3895818f, 0.f),
    C_Vertex3f(0.178693533f, 0.468958139f, 0.f),
    C_Vertex3f(0.181022406f, 0.396277189f, 0.f),
    C_Vertex3f(0.191960216f, 0.471783757f, 0.f),
    C_Vertex3f(0.191976786f, 0.401633501f, 0.f),
    C_Vertex3f(0.205567956f, 0.473667502f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122811079f, 0.320060015f, 0.f),
    C_Vertex3f(0.128308415f, 0.335404277f, 0.f),
    C_Vertex3f(0.12903738f, 0.448236942f, 0.f),
    C_Vertex3f(0.134905219f, 0.349204659f, 0.f),
    C_Vertex3f(0.140939832f, 0.45483005f, 0.f),
    C_Vertex3f(0.14260149f, 0.361461163f, 0.f),
    C_Vertex3f(0.153183341f, 0.460481286f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.112915874f, 0.264763951f, 0.f),
    C_Vertex3f(0.115114808f, 0.284739852f, 0.f),
    C_Vertex3f(0.117475986f, 0.440701962f, 0.f),
    C_Vertex3f(0.11841321f, 0.303171873f, 0.f),
    C_Vertex3f(0.12903738f, 0.448236942f, 0.f),
    C_Vertex3f(0.122811079f, 0.320060015f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.111816406f, 0.243244171f, 0.f),
    C_Vertex3f(0.112915874f, 0.264763951f, 0.f),
    C_Vertex3f(0.117475986f, 0.440701962f, 0.f),
    C_Vertex3f(0.10663712f, 0.432341099f, 0.f),
    C_Vertex3f(0.111653924f, 0.0325491428f, 0.f),
    C_Vertex3f(0.111816406f, 0.220562935f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111653924f, 0.0325491428f, 0.f),
    C_Vertex3f(0.10663712f, 0.432341099f, 0.f),
    C_Vertex3f(0.101326823f, 0.0417087078f, 0.f),
    C_Vertex3f(0.0965207815f, 0.423154354f, 0.f),
    C_Vertex3f(0.0916554928f, 0.0516257286f, 0.f),
    C_Vertex3f(0.0871269703f, 0.413141727f, 0.f),
    C_Vertex3f(0.0826399326f, 0.0623002052f, 0.f),
    C_Vertex3f(0.0784556866f, 0.402303219f, 0.f),
    C_Vertex3f(0.0742801428f, 0.0737321377f, 0.f),
    C_Vertex3f(0.0705069304f, 0.390638828f, 0.f),
    C_Vertex3f(0.0665761232f, 0.085921526f, 0.f),
    C_Vertex3f(0.0632807016f, 0.378148556f, 0.f),
    C_Vertex3f(0.059527874f, 0.0988683701f, 0.f),
    C_Vertex3f(0.0567770004f, 0.364832401f, 0.f),
    C_Vertex3f(0.0499391556f, 0.11942482f, 0.f),
    C_Vertex3f(0.0481052399f, 0.343619347f, 0.f),
    C_Vertex3f(0.0403504372f, 0.148881912f, 0.f),
    C_Vertex3f(0.0394334793f, 0.313724995f, 0.f),
    C_Vertex3f(0.0339579582f, 0.180387497f, 0.f),
    C_Vertex3f(0.0336523056f, 0.282220364f, 0.f),
    C_Vertex3f(0.0307617188f, 0.249105453f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361816406f, 0.214337349f, 0.f),
    C_Vertex3f(0.36068964f, 0.194081426f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.361816406f, 0.237072945f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0339579582f, 0.180387497f, 0.f),
    C_Vertex3f(0.0307617188f, 0.249105453f, 0.f),
    C_Vertex3f(0.0307617188f, 0.213941574f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360351563f, 0.058555603f, -1.f),
    C_Vertex3f(0.360351563f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.36068964f, 0.194081426f, -1.f),
    C_Vertex3f(0.358436108f, 0.175212264f, -1.f),
    C_Vertex3f(0.355055809f, 0.157729864f, -1.f),
    C_Vertex3f(0.354832649f, 0.0499544144f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355055809f, 0.157729864f, -1.f),
    C_Vertex3f(0.350548744f, 0.141634226f, -1.f),
    C_Vertex3f(0.354832649f, 0.0499544144f, -1.f),
    C_Vertex3f(0.344914913f, 0.126925349f, -1.f),
    C_Vertex3f(0.342497349f, 0.0349023342f, -1.f),
    C_Vertex3f(0.338154316f, 0.113603234f, -1.f),
    C_Vertex3f(0.330266953f, 0.101667881f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.342497349f, 0.0349023342f, -1.f),
    C_Vertex3f(0.330266953f, 0.101667881f, -1.f),
    C_Vertex3f(0.328864574f, 0.0220005512f, -1.f),
    C_Vertex3f(0.321252823f, 0.0911192894f, -1.f),
    C_Vertex3f(0.313934326f, 0.0112490654f, -1.f),
    C_Vertex3f(0.31175375f, 0.0818892717f, -1.f),
    C_Vertex3f(0.301769733f, 0.073977828f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.297706604f, 0.00264787674f, -1.f),
    C_Vertex3f(0.313934326f, 0.0112490654f, -1.f),
    C_Vertex3f(0.301769733f, 0.073977828f, -1.f),
    C_Vertex3f(0.291300774f, 0.0673849583f, -1.f),
    C_Vertex3f(0.28034687f, 0.0621106625f, -1.f),
    C_Vertex3f(0.280181408f, -0.00380301476f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.268908024f, 0.0581549406f, -1.f),
    C_Vertex3f(0.256984234f, 0.0555177927f, -1.f),
    C_Vertex3f(0.261358738f, -0.00810360909f, -1.f),
    C_Vertex3f(0.280181408f, -0.00380301476f, -1.f),
    C_Vertex3f(0.28034687f, 0.0621106625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.195922613f, 0.0624790192f, -1.f),
    C_Vertex3f(0.196670771f, -0.00738716125f, -1.f),
    C_Vertex3f(0.207436919f, 0.058339119f, -1.f),
    C_Vertex3f(0.209990263f, -0.00929832458f, -1.f),
    C_Vertex3f(0.219374776f, 0.0555791855f, -1.f),
    C_Vertex3f(0.223589897f, -0.0102539063f, -1.f),
    C_Vertex3f(0.231736183f, 0.0541992188f, -1.f),
    C_Vertex3f(0.241238594f, -0.0102539063f, -1.f),
    C_Vertex3f(0.2445755f, 0.0541992188f, -1.f),
    C_Vertex3f(0.261358738f, -0.00810360909f, -1.f),
    C_Vertex3f(0.256984234f, 0.0555177927f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.144704342f, 0.103878021f, -1.f),
    C_Vertex3f(0.14619422f, 0.00981330872f, -1.f),
    C_Vertex3f(0.154100895f, 0.0928382874f, -1.f),
    C_Vertex3f(0.158393145f, 0.00407981873f, -1.f),
    C_Vertex3f(0.163920999f, 0.0831785202f, -1.f),
    C_Vertex3f(0.170872211f, -0.0006980896f, -1.f),
    C_Vertex3f(0.174164653f, 0.0748987198f, -1.f),
    C_Vertex3f(0.18363142f, -0.00452041626f, -1.f),
    C_Vertex3f(0.184831858f, 0.0679988861f, -1.f),
    C_Vertex3f(0.196670771f, -0.00738716125f, -1.f),
    C_Vertex3f(0.195922613f, 0.0624790192f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136482358f, 0.11632514f, -1.f),
    C_Vertex3f(0.129434943f, 0.130179644f, -1.f),
    C_Vertex3f(0.134275436f, 0.0165023804f, -1.f),
    C_Vertex3f(0.14619422f, 0.00981330872f, -1.f),
    C_Vertex3f(0.144704342f, 0.103878021f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123562098f, 0.145441532f, -1.f),
    C_Vertex3f(0.118863821f, 0.162110806f, -1.f),
    C_Vertex3f(0.122636795f, 0.0241470337f, -1.f),
    C_Vertex3f(0.134275436f, 0.0165023804f, -1.f),
    C_Vertex3f(0.129434943f, 0.130179644f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122636795f, 0.0241470337f, -1.f),
    C_Vertex3f(0.118863821f, 0.162110806f, -1.f),
    C_Vertex3f(0.115340114f, 0.180187464f, -1.f),
    C_Vertex3f(0.112990975f, 0.199671507f, -1.f),
    C_Vertex3f(0.111816406f, 0.220562935f, -1.f),
    C_Vertex3f(0.111653924f, 0.0325491428f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.433105469f, 0.640625f, -1.f),
    C_Vertex3f(0.354492188f, 0.640625f, -1.f),
    C_Vertex3f(0.354932785f, 0.299049139f, -1.f),
    C_Vertex3f(0.358374596f, 0.279975295f, -1.f),
    C_Vertex3f(0.360669136f, 0.259316564f, -1.f),
    C_Vertex3f(0.361816406f, 0.237072945f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.337723732f, 0.346761346f, -1.f),
    C_Vertex3f(0.344607353f, 0.332442164f, -1.f),
    C_Vertex3f(0.349365234f, 0.418054581f, -1.f),
    C_Vertex3f(0.350343704f, 0.316538095f, -1.f),
    C_Vertex3f(0.354492188f, 0.410972595f, -1.f),
    C_Vertex3f(0.354932785f, 0.299049139f, -1.f),
    C_Vertex3f(0.354492188f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.31080389f, 0.380400777f, -1.f),
    C_Vertex3f(0.320514679f, 0.370645046f, -1.f),
    C_Vertex3f(0.324276924f, 0.442379951f, -1.f),
    C_Vertex3f(0.329692841f, 0.35949564f, -1.f),
    C_Vertex3f(0.337584496f, 0.430884361f, -1.f),
    C_Vertex3f(0.337723732f, 0.346761346f, -1.f),
    C_Vertex3f(0.349365234f, 0.418054581f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278475761f, 0.401305914f, -1.f),
    C_Vertex3f(0.289784431f, 0.395731211f, -1.f),
    C_Vertex3f(0.293081284f, 0.461368561f, -1.f),
    C_Vertex3f(0.300560474f, 0.388762832f, -1.f),
    C_Vertex3f(0.30944252f, 0.452541351f, -1.f),
    C_Vertex3f(0.31080389f, 0.380400777f, -1.f),
    C_Vertex3f(0.324276924f, 0.442379951f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241353989f, 0.409667969f, -1.f),
    C_Vertex3f(0.25426054f, 0.408274293f, -1.f),
    C_Vertex3f(0.256925583f, 0.472402573f, -1.f),
    C_Vertex3f(0.266634464f, 0.405486941f, -1.f),
    C_Vertex3f(0.275575638f, 0.467988968f, -1.f),
    C_Vertex3f(0.278475761f, 0.401305914f, -1.f),
    C_Vertex3f(0.293081284f, 0.461368561f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.191976786f, 0.401633501f, -1.f),
    C_Vertex3f(0.203470826f, 0.405650735f, -1.f),
    C_Vertex3f(0.205567956f, 0.473667502f, -1.f),
    C_Vertex3f(0.215504527f, 0.408328891f, -1.f),
    C_Vertex3f(0.219516754f, 0.474609375f, -1.f),
    C_Vertex3f(0.228077888f, 0.409667969f, -1.f),
    C_Vertex3f(0.237131119f, 0.474609375f, -1.f),
    C_Vertex3f(0.241353989f, 0.409667969f, -1.f),
    C_Vertex3f(0.256925583f, 0.472402573f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.14260149f, 0.361461163f, -1.f),
    C_Vertex3f(0.151397228f, 0.372173786f, -1.f),
    C_Vertex3f(0.153183341f, 0.460481286f, -1.f),
    C_Vertex3f(0.160732627f, 0.381547332f, -1.f),
    C_Vertex3f(0.165767908f, 0.465190649f, -1.f),
    C_Vertex3f(0.170607686f, 0.3895818f, -1.f),
    C_Vertex3f(0.178693533f, 0.468958139f, -1.f),
    C_Vertex3f(0.181022406f, 0.396277189f, -1.f),
    C_Vertex3f(0.191960216f, 0.471783757f, -1.f),
    C_Vertex3f(0.191976786f, 0.401633501f, -1.f),
    C_Vertex3f(0.205567956f, 0.473667502f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122811079f, 0.320060015f, -1.f),
    C_Vertex3f(0.128308415f, 0.335404277f, -1.f),
    C_Vertex3f(0.12903738f, 0.448236942f, -1.f),
    C_Vertex3f(0.134905219f, 0.349204659f, -1.f),
    C_Vertex3f(0.140939832f, 0.45483005f, -1.f),
    C_Vertex3f(0.14260149f, 0.361461163f, -1.f),
    C_Vertex3f(0.153183341f, 0.460481286f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.112915874f, 0.264763951f, -1.f),
    C_Vertex3f(0.115114808f, 0.284739852f, -1.f),
    C_Vertex3f(0.117475986f, 0.440701962f, -1.f),
    C_Vertex3f(0.11841321f, 0.303171873f, -1.f),
    C_Vertex3f(0.12903738f, 0.448236942f, -1.f),
    C_Vertex3f(0.122811079f, 0.320060015f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.111816406f, 0.243244171f, -1.f),
    C_Vertex3f(0.112915874f, 0.264763951f, -1.f),
    C_Vertex3f(0.117475986f, 0.440701962f, -1.f),
    C_Vertex3f(0.10663712f, 0.432341099f, -1.f),
    C_Vertex3f(0.111653924f, 0.0325491428f, -1.f),
    C_Vertex3f(0.111816406f, 0.220562935f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111653924f, 0.0325491428f, -1.f),
    C_Vertex3f(0.10663712f, 0.432341099f, -1.f),
    C_Vertex3f(0.101326823f, 0.0417087078f, -1.f),
    C_Vertex3f(0.0965207815f, 0.423154354f, -1.f),
    C_Vertex3f(0.0916554928f, 0.0516257286f, -1.f),
    C_Vertex3f(0.0871269703f, 0.413141727f, -1.f),
    C_Vertex3f(0.0826399326f, 0.0623002052f, -1.f),
    C_Vertex3f(0.0784556866f, 0.402303219f, -1.f),
    C_Vertex3f(0.0742801428f, 0.0737321377f, -1.f),
    C_Vertex3f(0.0705069304f, 0.390638828f, -1.f),
    C_Vertex3f(0.0665761232f, 0.085921526f, -1.f),
    C_Vertex3f(0.0632807016f, 0.378148556f, -1.f),
    C_Vertex3f(0.059527874f, 0.0988683701f, -1.f),
    C_Vertex3f(0.0567770004f, 0.364832401f, -1.f),
    C_Vertex3f(0.0499391556f, 0.11942482f, -1.f),
    C_Vertex3f(0.0481052399f, 0.343619347f, -1.f),
    C_Vertex3f(0.0403504372f, 0.148881912f, -1.f),
    C_Vertex3f(0.0394334793f, 0.313724995f, -1.f),
    C_Vertex3f(0.0339579582f, 0.180387497f, -1.f),
    C_Vertex3f(0.0336523056f, 0.282220364f, -1.f),
    C_Vertex3f(0.0307617188f, 0.249105453f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361816406f, 0.214337349f, -1.f),
    C_Vertex3f(0.36068964f, 0.194081426f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.361816406f, 0.237072945f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0339579582f, 0.180387497f, -1.f),
    C_Vertex3f(0.0307617188f, 0.249105453f, -1.f),
    C_Vertex3f(0.0307617188f, 0.213941574f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_101[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.125192568f, 0.992132485f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.965454936f, -0.260570467f, 0.f),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Normal3f(0.952376783f, -0.304923922f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, -1.f),
    C_Normal3f(0.932039082f, -0.362357676f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, -1.f),
    C_Normal3f(0.901194572f, -0.433414668f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, -1.f),
    C_Normal3f(0.862806261f, -0.505534768f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, -1.f),
    C_Normal3f(0.816684961f, -0.577083826f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, -1.f),
    C_Normal3f(0.763127804f, -0.646247566f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, -1.f),
    C_Normal3f(0.702967107f, -0.711222291f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, -1.f),
    C_Normal3f(0.637524903f, -0.77042979f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, -1.f),
    C_Normal3f(0.565409184f, -0.824810505f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Normal3f(0.48646915f, -0.873697758f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, -1.f),
    C_Normal3f(0.405260772f, -0.9142012f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, -1.f),
    C_Normal3f(0.323988765f, -0.946060896f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Normal3f(0.244618788f, -0.969619274f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, -1.f),
    C_Normal3f(0.168704703f, -0.985666752f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, -1.f),
    C_Normal3f(0.0973226056f, -0.995252907f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, -1.f),
    C_Normal3f(0.029160779f, -0.999574721f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0416765399f, -0.999131143f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Normal3f(-0.124204665f, -0.992256582f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Normal3f(-0.21582f, -0.976433158f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, -1.f),
    C_Normal3f(-0.312524736f, -0.949909627f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Normal3f(-0.411543161f, -0.911390245f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Normal3f(-0.509326518f, -0.860573351f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Normal3f(-0.602056086f, -0.798453748f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Normal3f(-0.686321557f, -0.727298141f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, -1.f),
    C_Normal3f(-0.762221575f, -0.647316158f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, -1.f),
    C_Normal3f(-0.830126584f, -0.557574868f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, -1.f),
    C_Normal3f(-0.886253357f, -0.463200867f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, -1.f),
    C_Normal3f(-0.929815829f, -0.368024975f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, -1.f),
    C_Normal3f(-0.961317182f, -0.275443733f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, -1.f),
    C_Normal3f(-0.982170284f, -0.187993705f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, -1.f),
    C_Normal3f(-0.994233549f, -0.107235931f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Normal3f(-0.999436319f, -0.0335720852f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, -1.f),
    C_Normal3f(-0.999442458f, 0.0333874822f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Normal3f(-0.994488776f, 0.104842685f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, -1.f),
    C_Normal3f(-0.982936144f, 0.183947176f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, -1.f),
    C_Normal3f(-0.962905705f, 0.269838274f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, -1.f),
    C_Normal3f(-0.932518303f, 0.361122698f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, -1.f),
    C_Normal3f(-0.890263319f, 0.455446035f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, -1.f),
    C_Normal3f(-0.835456192f, 0.54955703f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, -1.f),
    C_Normal3f(-0.768646836f, 0.639673412f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, -1.f),
    C_Normal3f(-0.695958257f, 0.718082249f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Normal3f(-0.616909444f, 0.787034094f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Normal3f(-0.527987719f, 0.849251986f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Normal3f(-0.431770295f, 0.901983559f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Normal3f(-0.33168745f, 0.943389297f, 0.f),
    C_Vertex3f(0.172035694f, 0.462487936f, 0.f),
    C_Vertex3f(0.172035694f, 0.462487936f, -1.f),
    C_Normal3f(-0.231436506f, 0.972849965f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_Normal3f(-0.134358883f, 0.990932763f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Normal3f(-0.0433523022f, 0.999059916f, 0.f),
    C_Vertex3f(0.237405777f, 0.474609375f, 0.f),
    C_Vertex3f(0.237405777f, 0.474609375f, -1.f),
    C_Normal3f(0.043157652f, 0.99906826f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_Normal3f(0.135833368f, 0.990731657f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Normal3f(0.233665943f, 0.97231698f, 0.f),
    C_Vertex3f(0.304389715f, 0.468667746f, 0.f),
    C_Vertex3f(0.304389715f, 0.468667746f, -1.f),
    C_Normal3f(0.334365368f, 0.94244355f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_Normal3f(0.434524238f, 0.900660098f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Normal3f(0.530448139f, 0.847717404f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Normal3f(0.618784487f, 0.785560787f, 0.f),
    C_Vertex3f(0.378282785f, 0.433017969f, 0.f),
    C_Vertex3f(0.378282785f, 0.433017969f, -1.f),
    C_Normal3f(0.697084367f, 0.7169891f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_Normal3f(0.769382715f, 0.638788223f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, -1.f),
    C_Normal3f(0.836126029f, 0.548537314f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, -1.f),
    C_Normal3f(0.890802264f, 0.454391062f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, -1.f),
    C_Normal3f(0.932900369f, 0.360134482f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, -1.f),
    C_Normal3f(0.963139415f, 0.269002616f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, -1.f),
    C_Normal3f(0.983052552f, 0.183323756f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, -1.f),
    C_Normal3f(0.994528711f, 0.104463689f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Normal3f(0.999162018f, 0.0409296043f, 0.f),
    C_Vertex3f(0.460449219f, 0.247264862f, 0.f),
    C_Vertex3f(0.460449219f, 0.247264862f, -1.f),
    C_Normal3f(0.999970615f, -0.0076672486f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, -1.f),
    C_Normal3f(0.999690831f, -0.0248668455f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, -1.f),
    C_Normal3f(0.999516904f, -0.0310815405f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.998369277f, 0.0570866764f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Normal3f(0.995165646f, 0.0982105583f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, -1.f),
    C_Normal3f(0.988114417f, 0.153720722f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, -1.f),
    C_Normal3f(0.97403872f, 0.226381436f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, -1.f),
    C_Normal3f(0.95226264f, 0.305279851f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_Normal3f(0.921066403f, 0.389405489f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, -1.f),
    C_Normal3f(0.878937125f, 0.476937532f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, -1.f),
    C_Normal3f(0.824944496f, 0.565213859f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, -1.f),
    C_Normal3f(0.759140372f, 0.650927007f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Normal3f(0.688533545f, 0.725204468f, 0.f),
    C_Vertex3f(0.161923409f, 0.0897507668f, 0.f),
    C_Vertex3f(0.161923409f, 0.0897507668f, -1.f),
    C_Normal3f(0.613223195f, 0.78990972f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_Normal3f(0.528003633f, 0.849242151f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, -1.f),
    C_Normal3f(0.434818447f, 0.900518119f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, -1.f),
    C_Normal3f(0.336550295f, 0.94166553f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_Normal3f(0.236599728f, 0.971607208f, 0.f),
    C_Vertex3f(0.219586134f, 0.0580083132f, 0.f),
    C_Vertex3f(0.219586134f, 0.0580083132f, -1.f),
    C_Normal3f(0.138325498f, 0.990386844f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_Normal3f(0.0396123156f, 0.999215126f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, -1.f),
    C_Normal3f(-0.074584201f, 0.997214675f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, -1.f),
    C_Normal3f(-0.216858819f, 0.976202905f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, -1.f),
    C_Normal3f(-0.374131411f, 0.927375793f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, -1.f),
    C_Normal3f(-0.527194917f, 0.84974438f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, -1.f),
    C_Normal3f(-0.670490623f, 0.741917968f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_Normal3f(-0.787500262f, 0.616314411f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, -1.f),
    C_Normal3f(-0.869012713f, 0.49478972f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, -1.f),
    C_Normal3f(-0.915187478f, 0.40302819f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, -1.f),
    C_Normal3f(-0.9414168f, 0.337245345f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.996035993f, -0.088950254f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Normal3f(-0.992682815f, -0.120750621f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, -1.f),
    C_Normal3f(-0.986461818f, -0.163990855f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, -1.f),
    C_Normal3f(-0.975184321f, -0.221394569f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, -1.f),
    C_Normal3f(-0.958473325f, -0.285182297f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, -1.f),
    C_Normal3f(-0.934688568f, -0.355467528f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_Normal3f(-0.901933372f, -0.431875169f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, -1.f),
    C_Normal3f(-0.85821414f, -0.513291836f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, -1.f),
    C_Normal3f(-0.79895848f, -0.601386249f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_Normal3f(-0.734940171f, -0.678131938f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, -1.f),
    C_Normal3f(-0.663570166f, -0.74811399f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, -1.f),
    C_Normal3f(-0.575440705f, -0.817843437f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, -1.f),
    C_Normal3f(-0.476377249f, -0.87924099f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Normal3f(-0.369776905f, -0.929120481f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, -1.f),
    C_Normal3f(-0.260044605f, -0.965596616f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_Normal3f(-0.151727274f, -0.988422394f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, -1.f),
    C_Normal3f(-0.0498818606f, -0.998755097f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, -1.f),
    C_Normal3f(0.0416725017f, -0.999131382f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Normal3f(0.132727861f, -0.991152465f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, -1.f),
    C_Normal3f(0.226974145f, -0.973900795f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_Normal3f(0.32306844f, -0.946375608f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, -1.f),
    C_Normal3f(0.418056071f, -0.908421159f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, -1.f),
    C_Normal3f(0.50887841f, -0.860838413f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, -1.f),
    C_Normal3f(0.59283942f, -0.80532068f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Normal3f(0.667988896f, -0.744171143f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, -1.f),
    C_Normal3f(0.736174107f, -0.676792264f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_Normal3f(0.80003792f, -0.59994936f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, -1.f),
    C_Normal3f(0.855952084f, -0.517055154f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, -1.f),
    C_Normal3f(0.902495027f, -0.430700183f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, -1.f),
    C_Normal3f(0.939107299f, -0.343623996f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_Normal3f(0.966056883f, -0.258329391f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, -1.f),
    C_Normal3f(0.984245181f, -0.176808938f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, -1.f),
    C_Normal3f(0.993621826f, -0.112763919f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Normal3f(0.997922182f, -0.0644304082f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219586134f, 0.0580083132f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161923409f, 0.0897507668f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378282785f, 0.433017969f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304389715f, 0.468667746f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237405777f, 0.474609375f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172035694f, 0.462487936f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460449219f, 0.247264862f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, -1.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, -1.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, -1.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, -1.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, -1.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, -1.f),
    C_Vertex3f(0.449591041f, 0.113932014f, -1.f),
    C_Vertex3f(0.455421448f, 0.13074398f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, -1.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, -1.f),
    C_Vertex3f(0.361532211f, 0.115768433f, -1.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, -1.f),
    C_Vertex3f(0.372081757f, 0.137390137f, -1.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, -1.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, -1.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, -1.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, -1.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, -1.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, -1.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, -1.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, -1.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, -1.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219586134f, 0.0580083132f, -1.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, -1.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, -1.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, -1.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161923409f, 0.0897507668f, -1.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Vertex3f(0.143544197f, 0.111185431f, -1.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Vertex3f(0.13597393f, 0.123582006f, -1.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Vertex3f(0.129483223f, 0.137098074f, -1.f),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_Vertex3f(0.119740486f, 0.167488694f, -1.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Vertex3f(0.116488457f, 0.184363246f, -1.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Vertex3f(0.114315987f, 0.202357292f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, -1.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, -1.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, -1.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, -1.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, -1.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, -1.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, -1.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, -1.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, -1.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, -1.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, -1.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, -1.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, -1.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, -1.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378282785f, 0.433017969f, -1.f),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_Vertex3f(0.369242191f, 0.322288036f, -1.f),
    C_Vertex3f(0.372357607f, 0.310348272f, -1.f),
    C_Vertex3f(0.374871969f, 0.297490358f, -1.f),
    C_Vertex3f(0.376785278f, 0.283714294f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_Vertex3f(0.356289625f, 0.352598429f, -1.f),
    C_Vertex3f(0.3612082f, 0.343413115f, -1.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Vertex3f(0.322116375f, 0.38877368f, -1.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_Vertex3f(0.332911015f, 0.380220652f, -1.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Vertex3f(0.343008995f, 0.370242119f, -1.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304389715f, 0.468667746f, -1.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_Vertex3f(0.298437119f, 0.401603222f, -1.f),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Vertex3f(0.257693291f, 0.41015625f, -1.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_Vertex3f(0.271971226f, 0.408730745f, -1.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237405777f, 0.474609375f, -1.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_Vertex3f(0.230415225f, 0.409018517f, -1.f),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Vertex3f(0.183755994f, 0.393090248f, -1.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_Vertex3f(0.194611311f, 0.398778915f, -1.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Vertex3f(0.206006289f, 0.403329849f, -1.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172035694f, 0.462487936f, -1.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_Vertex3f(0.163664341f, 0.378299713f, -1.f),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_Vertex3f(0.132908106f, 0.337285995f, -1.f),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Vertex3f(0.139049888f, 0.348690987f, -1.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Vertex3f(0.146223187f, 0.35932827f, -1.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Vertex3f(0.120671868f, 0.298464775f, -1.f),
    C_Vertex3f(0.123719096f, 0.31217289f, -1.f),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460449219f, 0.247264862f, -1.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.460231781f, 0.218906403f, -1.f),
    C_Vertex3f(0.460449219f, 0.22939682f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Vertex3f(0.454919577f, 0.299908876f, -1.f),
    C_Vertex3f(0.449389935f, 0.323812962f, -1.f),
    C_Vertex3f(0.442017078f, 0.346105099f, -1.f),
    C_Vertex3f(0.432801008f, 0.366785288f, -1.f),
    C_Vertex3f(0.421741724f, 0.385853529f, -1.f),
    C_Vertex3f(0.408839226f, 0.403309822f, -1.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_102[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0777893066f, 0.f, 0.f),
    C_Vertex3f(0.0777893066f, 0.f, -1.f),
    C_Vertex3f(0.0777893066f, 0.403320313f, -1.f),
    C_Vertex3f(0.0777893066f, 0.403320313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0777893066f, 0.403320313f, 0.f),
    C_Vertex3f(0.0777893066f, 0.403320313f, -1.f),
    C_Vertex3f(0.00845336914f, 0.403320313f, -1.f),
    C_Vertex3f(0.00845336914f, 0.403320313f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.00845336914f, 0.403320313f, 0.f),
    C_Vertex3f(0.00845336914f, 0.403320313f, -1.f),
    C_Vertex3f(0.00845336914f, 0.464355469f, -1.f),
    C_Vertex3f(0.00845336914f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.00845336914f, 0.464355469f, 0.f),
    C_Vertex3f(0.00845336914f, 0.464355469f, -1.f),
    C_Vertex3f(0.0777893066f, 0.464355469f, -1.f),
    C_Vertex3f(0.0777893066f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0777893066f, 0.464355469f, 0.f),
    C_Vertex3f(0.0777893066f, 0.464355469f, -1.f),
    C_Normal3f(-0.99991858f, 0.0127583751f, 0.f),
    C_Vertex3f(0.0777893066f, 0.525278091f, 0.f),
    C_Vertex3f(0.0777893066f, 0.525278091f, -1.f),
    C_Normal3f(-0.996605694f, 0.0823225901f, 0.f),
    C_Vertex3f(0.0788259506f, 0.545600891f, 0.f),
    C_Vertex3f(0.0788259506f, 0.545600891f, -1.f),
    C_Normal3f(-0.986859202f, 0.161582813f, 0.f),
    C_Vertex3f(0.0808992386f, 0.562927246f, 0.f),
    C_Vertex3f(0.0808992386f, 0.562927246f, -1.f),
    C_Normal3f(-0.960355639f, 0.278777629f, 0.f),
    C_Vertex3f(0.0840091705f, 0.577257156f, 0.f),
    C_Vertex3f(0.0840091705f, 0.577257156f, -1.f),
    C_Normal3f(-0.905084074f, 0.425232708f, 0.f),
    C_Vertex3f(0.0889186859f, 0.590553284f, 0.f),
    C_Vertex3f(0.0889186859f, 0.590553284f, -1.f),
    C_Normal3f(-0.826848328f, 0.562424898f, 0.f),
    C_Vertex3f(0.0967464447f, 0.604367733f, 0.f),
    C_Vertex3f(0.0967464447f, 0.604367733f, -1.f),
    C_Normal3f(-0.725149512f, 0.68859148f, 0.f),
    C_Vertex3f(0.106729507f, 0.616737843f, 0.f),
    C_Vertex3f(0.106729507f, 0.616737843f, -1.f),
    C_Normal3f(-0.611118257f, 0.791539311f, 0.f),
    C_Vertex3f(0.118867874f, 0.627663612f, 0.f),
    C_Vertex3f(0.118867874f, 0.627663612f, -1.f),
    C_Normal3f(-0.466499358f, 0.884521484f, 0.f),
    C_Vertex3f(0.133161545f, 0.637145042f, 0.f),
    C_Vertex3f(0.133161545f, 0.637145042f, -1.f),
    C_Normal3f(-0.303506047f, 0.95282948f, 0.f),
    C_Vertex3f(0.150328636f, 0.644256115f, 0.f),
    C_Vertex3f(0.150328636f, 0.644256115f, -1.f),
    C_Normal3f(-0.163326189f, 0.986572146f, 0.f),
    C_Vertex3f(0.170369148f, 0.64899683f, 0.f),
    C_Vertex3f(0.170369148f, 0.64899683f, -1.f),
    C_Normal3f(-0.0536596328f, 0.998559296f, 0.f),
    C_Vertex3f(0.193283081f, 0.651367188f, 0.f),
    C_Vertex3f(0.193283081f, 0.651367188f, -1.f),
    C_Normal3f(0.025463324f, 0.99967581f, 0.f),
    C_Vertex3f(0.214479446f, 0.651367188f, 0.f),
    C_Vertex3f(0.214479446f, 0.651367188f, -1.f),
    C_Normal3f(0.0821959674f, 0.996616185f, 0.f),
    C_Vertex3f(0.231959343f, 0.650382042f, 0.f),
    C_Vertex3f(0.231959343f, 0.650382042f, -1.f),
    C_Normal3f(0.129950091f, 0.991520464f, 0.f),
    C_Vertex3f(0.250313759f, 0.648411751f, 0.f),
    C_Vertex3f(0.250313759f, 0.648411751f, -1.f),
    C_Normal3f(0.165893629f, 0.986143708f, 0.f),
    C_Vertex3f(0.269542694f, 0.645456314f, 0.f),
    C_Vertex3f(0.269542694f, 0.645456314f, -1.f),
    C_Normal3f(0.192354694f, 0.981325507f, 0.f),
    C_Vertex3f(0.279594421f, 0.643486023f, 0.f),
    C_Vertex3f(0.279594421f, 0.643486023f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.985574841f, -0.169240132f, 0.f),
    C_Vertex3f(0.279594421f, 0.643486023f, 0.f),
    C_Vertex3f(0.279594421f, 0.643486023f, -1.f),
    C_Vertex3f(0.267791748f, 0.574752808f, -1.f),
    C_Vertex3f(0.267791748f, 0.574752808f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.174993649f, -0.98456955f, 0.f),
    C_Vertex3f(0.267791748f, 0.574752808f, 0.f),
    C_Vertex3f(0.267791748f, 0.574752808f, -1.f),
    C_Normal3f(-0.121954165f, -0.99253571f, 0.f),
    C_Vertex3f(0.255558014f, 0.576927185f, 0.f),
    C_Vertex3f(0.255558014f, 0.576927185f, -1.f),
    C_Normal3f(-0.0504519083f, -0.998726487f, 0.f),
    C_Vertex3f(0.232398987f, 0.579101563f, 0.f),
    C_Vertex3f(0.232398987f, 0.579101563f, -1.f),
    C_Normal3f(0.053975258f, -0.998542309f, 0.f),
    C_Vertex3f(0.212514877f, 0.579101563f, 0.f),
    C_Vertex3f(0.212514877f, 0.579101563f, -1.f),
    C_Normal3f(0.19984743f, -0.979827046f, 0.f),
    C_Vertex3f(0.197218895f, 0.577199936f, 0.f),
    C_Vertex3f(0.197218895f, 0.577199936f, -1.f),
    C_Normal3f(0.385948092f, -0.922520459f, 0.f),
    C_Vertex3f(0.184544563f, 0.573396683f, 0.f),
    C_Vertex3f(0.184544563f, 0.573396683f, -1.f),
    C_Normal3f(0.605768561f, -0.795640826f, 0.f),
    C_Vertex3f(0.174491882f, 0.567691803f, 0.f),
    C_Vertex3f(0.174491882f, 0.567691803f, -1.f),
    C_Normal3f(0.818496883f, -0.574510932f, 0.f),
    C_Vertex3f(0.167060852f, 0.560085297f, 0.f),
    C_Vertex3f(0.167060852f, 0.560085297f, -1.f),
    C_Normal3f(0.938087165f, -0.346399099f, 0.f),
    C_Vertex3f(0.161487579f, 0.549164772f, 0.f),
    C_Vertex3f(0.161487579f, 0.549164772f, -1.f),
    C_Normal3f(0.984971285f, -0.172717988f, 0.f),
    C_Vertex3f(0.157772064f, 0.534930229f, 0.f),
    C_Vertex3f(0.157772064f, 0.534930229f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155914307f, 0.517381668f, 0.f),
    C_Vertex3f(0.155914307f, 0.517381668f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155914307f, 0.517381668f, 0.f),
    C_Vertex3f(0.155914307f, 0.517381668f, -1.f),
    C_Vertex3f(0.155914307f, 0.464355469f, -1.f),
    C_Vertex3f(0.155914307f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155914307f, 0.464355469f, 0.f),
    C_Vertex3f(0.155914307f, 0.464355469f, -1.f),
    C_Vertex3f(0.246246338f, 0.464355469f, -1.f),
    C_Vertex3f(0.246246338f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.246246338f, 0.464355469f, 0.f),
    C_Vertex3f(0.246246338f, 0.464355469f, -1.f),
    C_Vertex3f(0.246246338f, 0.403320313f, -1.f),
    C_Vertex3f(0.246246338f, 0.403320313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.246246338f, 0.403320313f, 0.f),
    C_Vertex3f(0.246246338f, 0.403320313f, -1.f),
    C_Vertex3f(0.156402588f, 0.403320313f, -1.f),
    C_Vertex3f(0.156402588f, 0.403320313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.156402588f, 0.403320313f, 0.f),
    C_Vertex3f(0.156402588f, 0.403320313f, -1.f),
    C_Vertex3f(0.156402588f, 0.f, -1.f),
    C_Vertex3f(0.156402588f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.156402588f, 0.f, 0.f),
    C_Vertex3f(0.156402588f, 0.f, -1.f),
    C_Vertex3f(0.0777893066f, 0.f, -1.f),
    C_Vertex3f(0.0777893066f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0788259506f, 0.545600891f, 0.f),
    C_Vertex3f(0.0777893066f, 0.525278091f, 0.f),
    C_Vertex3f(0.156402588f, 0.f, 0.f),
    C_Vertex3f(0.0777893066f, 0.464355469f, 0.f),
    C_Vertex3f(0.0777893066f, 0.403320313f, 0.f),
    C_Vertex3f(0.00845336914f, 0.464355469f, 0.f),
    C_Vertex3f(0.00845336914f, 0.403320313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.246246338f, 0.403320313f, 0.f),
    C_Vertex3f(0.246246338f, 0.464355469f, 0.f),
    C_Vertex3f(0.156402588f, 0.403320313f, 0.f),
    C_Vertex3f(0.155914307f, 0.464355469f, 0.f),
    C_Vertex3f(0.156402588f, 0.f, 0.f),
    C_Vertex3f(0.150328636f, 0.644256115f, 0.f),
    C_Vertex3f(0.133161545f, 0.637145042f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156402588f, 0.f, 0.f),
    C_Vertex3f(0.133161545f, 0.637145042f, 0.f),
    C_Vertex3f(0.118867874f, 0.627663612f, 0.f),
    C_Vertex3f(0.106729507f, 0.616737843f, 0.f),
    C_Vertex3f(0.0967464447f, 0.604367733f, 0.f),
    C_Vertex3f(0.0889186859f, 0.590553284f, 0.f),
    C_Vertex3f(0.0840091705f, 0.577257156f, 0.f),
    C_Vertex3f(0.0808992386f, 0.562927246f, 0.f),
    C_Vertex3f(0.0788259506f, 0.545600891f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.231959343f, 0.650382042f, 0.f),
    C_Vertex3f(0.232398987f, 0.579101563f, 0.f),
    C_Vertex3f(0.250313759f, 0.648411751f, 0.f),
    C_Vertex3f(0.255558014f, 0.576927185f, 0.f),
    C_Vertex3f(0.269542694f, 0.645456314f, 0.f),
    C_Vertex3f(0.267791748f, 0.574752808f, 0.f),
    C_Vertex3f(0.279594421f, 0.643486023f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214479446f, 0.651367188f, 0.f),
    C_Vertex3f(0.193283081f, 0.651367188f, 0.f),
    C_Vertex3f(0.197218895f, 0.577199936f, 0.f),
    C_Vertex3f(0.212514877f, 0.579101563f, 0.f),
    C_Vertex3f(0.232398987f, 0.579101563f, 0.f),
    C_Vertex3f(0.231959343f, 0.650382042f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.161487579f, 0.549164772f, 0.f),
    C_Vertex3f(0.167060852f, 0.560085297f, 0.f),
    C_Vertex3f(0.170369148f, 0.64899683f, 0.f),
    C_Vertex3f(0.174491882f, 0.567691803f, 0.f),
    C_Vertex3f(0.193283081f, 0.651367188f, 0.f),
    C_Vertex3f(0.184544563f, 0.573396683f, 0.f),
    C_Vertex3f(0.197218895f, 0.577199936f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.155914307f, 0.464355469f, 0.f),
    C_Vertex3f(0.155914307f, 0.517381668f, 0.f),
    C_Vertex3f(0.150328636f, 0.644256115f, 0.f),
    C_Vertex3f(0.157772064f, 0.534930229f, 0.f),
    C_Vertex3f(0.170369148f, 0.64899683f, 0.f),
    C_Vertex3f(0.161487579f, 0.549164772f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.156402588f, 0.f, 0.f),
    C_Vertex3f(0.0777893066f, 0.403320313f, 0.f),
    C_Vertex3f(0.0777893066f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0788259506f, 0.545600891f, -1.f),
    C_Vertex3f(0.0777893066f, 0.525278091f, -1.f),
    C_Vertex3f(0.156402588f, 0.f, -1.f),
    C_Vertex3f(0.0777893066f, 0.464355469f, -1.f),
    C_Vertex3f(0.0777893066f, 0.403320313f, -1.f),
    C_Vertex3f(0.00845336914f, 0.464355469f, -1.f),
    C_Vertex3f(0.00845336914f, 0.403320313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.246246338f, 0.403320313f, -1.f),
    C_Vertex3f(0.246246338f, 0.464355469f, -1.f),
    C_Vertex3f(0.156402588f, 0.403320313f, -1.f),
    C_Vertex3f(0.155914307f, 0.464355469f, -1.f),
    C_Vertex3f(0.156402588f, 0.f, -1.f),
    C_Vertex3f(0.150328636f, 0.644256115f, -1.f),
    C_Vertex3f(0.133161545f, 0.637145042f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156402588f, 0.f, -1.f),
    C_Vertex3f(0.133161545f, 0.637145042f, -1.f),
    C_Vertex3f(0.118867874f, 0.627663612f, -1.f),
    C_Vertex3f(0.106729507f, 0.616737843f, -1.f),
    C_Vertex3f(0.0967464447f, 0.604367733f, -1.f),
    C_Vertex3f(0.0889186859f, 0.590553284f, -1.f),
    C_Vertex3f(0.0840091705f, 0.577257156f, -1.f),
    C_Vertex3f(0.0808992386f, 0.562927246f, -1.f),
    C_Vertex3f(0.0788259506f, 0.545600891f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.231959343f, 0.650382042f, -1.f),
    C_Vertex3f(0.232398987f, 0.579101563f, -1.f),
    C_Vertex3f(0.250313759f, 0.648411751f, -1.f),
    C_Vertex3f(0.255558014f, 0.576927185f, -1.f),
    C_Vertex3f(0.269542694f, 0.645456314f, -1.f),
    C_Vertex3f(0.267791748f, 0.574752808f, -1.f),
    C_Vertex3f(0.279594421f, 0.643486023f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214479446f, 0.651367188f, -1.f),
    C_Vertex3f(0.193283081f, 0.651367188f, -1.f),
    C_Vertex3f(0.197218895f, 0.577199936f, -1.f),
    C_Vertex3f(0.212514877f, 0.579101563f, -1.f),
    C_Vertex3f(0.232398987f, 0.579101563f, -1.f),
    C_Vertex3f(0.231959343f, 0.650382042f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.161487579f, 0.549164772f, -1.f),
    C_Vertex3f(0.167060852f, 0.560085297f, -1.f),
    C_Vertex3f(0.170369148f, 0.64899683f, -1.f),
    C_Vertex3f(0.174491882f, 0.567691803f, -1.f),
    C_Vertex3f(0.193283081f, 0.651367188f, -1.f),
    C_Vertex3f(0.184544563f, 0.573396683f, -1.f),
    C_Vertex3f(0.197218895f, 0.577199936f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.155914307f, 0.464355469f, -1.f),
    C_Vertex3f(0.155914307f, 0.517381668f, -1.f),
    C_Vertex3f(0.150328636f, 0.644256115f, -1.f),
    C_Vertex3f(0.157772064f, 0.534930229f, -1.f),
    C_Vertex3f(0.170369148f, 0.64899683f, -1.f),
    C_Vertex3f(0.161487579f, 0.549164772f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.156402588f, 0.f, -1.f),
    C_Vertex3f(0.0777893066f, 0.403320313f, -1.f),
    C_Vertex3f(0.0777893066f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_103[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.144949675f, 0.98943907f, 0.f),
    C_Vertex3f(0.0445785522f, -0.0380859375f, 0.f),
    C_Vertex3f(0.0445785522f, -0.0380859375f, -1.f),
    C_Vertex3f(0.121238708f, -0.0493164063f, -1.f),
    C_Vertex3f(0.121238708f, -0.0493164063f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.990981042f, 0.134002596f, 0.f),
    C_Vertex3f(0.121238708f, -0.0493164063f, 0.f),
    C_Vertex3f(0.121238708f, -0.0493164063f, -1.f),
    C_Normal3f(0.973152995f, 0.230159253f, 0.f),
    C_Vertex3f(0.122440338f, -0.0582027435f, 0.f),
    C_Vertex3f(0.122440338f, -0.0582027435f, -1.f),
    C_Normal3f(0.92991513f, 0.367774218f, 0.f),
    C_Vertex3f(0.126973152f, -0.0735626221f, 0.f),
    C_Vertex3f(0.126973152f, -0.0735626221f, -1.f),
    C_Normal3f(0.835318625f, 0.549766004f, 0.f),
    C_Vertex3f(0.133635521f, -0.0865097046f, 0.f),
    C_Vertex3f(0.133635521f, -0.0865097046f, -1.f),
    C_Normal3f(0.681365311f, 0.731943488f, 0.f),
    C_Vertex3f(0.142427444f, -0.0970439911f, 0.f),
    C_Vertex3f(0.142427444f, -0.0970439911f, -1.f),
    C_Normal3f(0.508564711f, 0.861023724f, 0.f),
    C_Vertex3f(0.15520668f, -0.106590271f, 0.f),
    C_Vertex3f(0.15520668f, -0.106590271f, -1.f),
    C_Normal3f(0.345039219f, 0.938588262f, 0.f),
    C_Vertex3f(0.172520638f, -0.114818573f, 0.f),
    C_Vertex3f(0.172520638f, -0.114818573f, -1.f),
    C_Normal3f(0.189395308f, 0.98190093f, 0.f),
    C_Vertex3f(0.192511559f, -0.120304108f, 0.f),
    C_Vertex3f(0.192511559f, -0.120304108f, -1.f),
    C_Normal3f(0.0649086311f, 0.997891247f, 0.f),
    C_Vertex3f(0.215179443f, -0.123046875f, 0.f),
    C_Vertex3f(0.215179443f, -0.123046875f, -1.f),
    C_Normal3f(-0.0210433956f, 0.999778569f, 0.f),
    C_Vertex3f(0.234678268f, -0.123046875f, 0.f),
    C_Vertex3f(0.234678268f, -0.123046875f, -1.f),
    C_Normal3f(-0.0815888792f, 0.996666074f, 0.f),
    C_Vertex3f(0.247552991f, -0.122365475f, 0.f),
    C_Vertex3f(0.247552991f, -0.122365475f, -1.f),
    C_Normal3f(-0.143987566f, 0.989579499f, 0.f),
    C_Vertex3f(0.259649634f, -0.121002674f, 0.f),
    C_Vertex3f(0.259649634f, -0.121002674f, -1.f),
    C_Normal3f(-0.213190749f, 0.977010548f, 0.f),
    C_Vertex3f(0.270968199f, -0.118958473f, 0.f),
    C_Vertex3f(0.270968199f, -0.118958473f, -1.f),
    C_Normal3f(-0.289152712f, 0.95728302f, 0.f),
    C_Vertex3f(0.281508684f, -0.116232872f, 0.f),
    C_Vertex3f(0.281508684f, -0.116232872f, -1.f),
    C_Normal3f(-0.371250212f, 0.928532898f, 0.f),
    C_Vertex3f(0.291271091f, -0.112825871f, 0.f),
    C_Vertex3f(0.291271091f, -0.112825871f, -1.f),
    C_Normal3f(-0.458056808f, 0.88892287f, 0.f),
    C_Vertex3f(0.300255418f, -0.108737469f, 0.f),
    C_Vertex3f(0.300255418f, -0.108737469f, -1.f),
    C_Normal3f(-0.556115031f, 0.831105351f, 0.f),
    C_Vertex3f(0.308461666f, -0.103967667f, 0.f),
    C_Vertex3f(0.308461666f, -0.103967667f, -1.f),
    C_Normal3f(-0.669381082f, 0.742919326f, 0.f),
    C_Vertex3f(0.31960392f, -0.095790863f, 0.f),
    C_Vertex3f(0.31960392f, -0.095790863f, -1.f),
    C_Normal3f(-0.783216953f, 0.621748626f, 0.f),
    C_Vertex3f(0.332056522f, -0.0827083588f, 0.f),
    C_Vertex3f(0.332056522f, -0.0827083588f, -1.f),
    C_Normal3f(-0.879532576f, 0.475838572f, 0.f),
    C_Vertex3f(0.342105389f, -0.067445755f, 0.f),
    C_Vertex3f(0.342105389f, -0.067445755f, -1.f),
    C_Normal3f(-0.935814023f, 0.352493882f, 0.f),
    C_Vertex3f(0.349750519f, -0.0500030518f, 0.f),
    C_Vertex3f(0.349750519f, -0.0500030518f, -1.f),
    C_Normal3f(-0.966785848f, 0.255587876f, 0.f),
    C_Vertex3f(0.352371216f, -0.0401916504f, 0.f),
    C_Vertex3f(0.352371216f, -0.0401916504f, -1.f),
    C_Normal3f(-0.981225252f, 0.192864701f, 0.f),
    C_Vertex3f(0.353137016f, -0.0371932983f, 0.f),
    C_Vertex3f(0.353137016f, -0.0371932983f, -1.f),
    C_Normal3f(-0.989724576f, 0.142987147f, 0.f),
    C_Vertex3f(0.354463339f, -0.0295476913f, 0.f),
    C_Vertex3f(0.354463339f, -0.0295476913f, -1.f),
    C_Normal3f(-0.99497366f, 0.10013736f, 0.f),
    C_Vertex3f(0.355584383f, -0.0202531815f, 0.f),
    C_Vertex3f(0.355584383f, -0.0202531815f, -1.f),
    C_Normal3f(-0.997621357f, 0.0689328462f, 0.f),
    C_Vertex3f(0.356500149f, -0.00930976868f, 0.f),
    C_Vertex3f(0.356500149f, -0.00930976868f, -1.f),
    C_Normal3f(-0.998975337f, 0.0452586897f, 0.f),
    C_Vertex3f(0.357210636f, 0.003282547f, 0.f),
    C_Vertex3f(0.357210636f, 0.003282547f, -1.f),
    C_Normal3f(-0.999643147f, 0.0267114658f, 0.f),
    C_Vertex3f(0.357715845f, 0.0175237656f, 0.f),
    C_Vertex3f(0.357715845f, 0.0175237656f, -1.f),
    C_Normal3f(-0.999930322f, 0.011802732f, 0.f),
    C_Vertex3f(0.358015776f, 0.033413887f, 0.f),
    C_Vertex3f(0.358015776f, 0.033413887f, -1.f),
    C_Normal3f(-0.999998987f, 0.0014498597f, 0.f),
    C_Vertex3f(0.358110428f, 0.0509529114f, 0.f),
    C_Vertex3f(0.358110428f, 0.0509529114f, -1.f),
    C_Normal3f(-0.99998337f, -0.00576531189f, 0.f),
    C_Vertex3f(0.358055115f, 0.060546875f, 0.f),
    C_Vertex3f(0.358055115f, 0.060546875f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.761257231f, -0.648449957f, 0.f),
    C_Vertex3f(0.358055115f, 0.060546875f, 0.f),
    C_Vertex3f(0.358055115f, 0.060546875f, -1.f),
    C_Normal3f(0.718758523f, -0.695259869f, 0.f),
    C_Vertex3f(0.351608276f, 0.0529785156f, 0.f),
    C_Vertex3f(0.351608276f, 0.0529785156f, -1.f),
    C_Normal3f(0.657810032f, -0.753183842f, 0.f),
    C_Vertex3f(0.337922573f, 0.0397338867f, 0.f),
    C_Vertex3f(0.337922573f, 0.0397338867f, -1.f),
    C_Normal3f(0.573272824f, -0.819364607f, 0.f),
    C_Vertex3f(0.323444843f, 0.0283813477f, 0.f),
    C_Vertex3f(0.323444843f, 0.0283813477f, -1.f),
    C_Normal3f(0.477530628f, -0.878615141f, 0.f),
    C_Vertex3f(0.308175087f, 0.0189208984f, 0.f),
    C_Vertex3f(0.308175087f, 0.0189208984f, -1.f),
    C_Normal3f(0.373294592f, -0.927712917f, 0.f),
    C_Vertex3f(0.292113304f, 0.0113525391f, 0.f),
    C_Vertex3f(0.292113304f, 0.0113525391f, -1.f),
    C_Normal3f(0.264455944f, -0.964397848f, 0.f),
    C_Vertex3f(0.275259495f, 0.00567626953f, 0.f),
    C_Vertex3f(0.275259495f, 0.00567626953f, -1.f),
    C_Normal3f(0.155397445f, -0.987852037f, 0.f),
    C_Vertex3f(0.257613659f, 0.00189208984f, 0.f),
    C_Vertex3f(0.257613659f, 0.00189208984f, -1.f),
    C_Normal3f(0.0472297966f, -0.998884022f, 0.f),
    C_Vertex3f(0.239175797f, 0.f, 0.f),
    C_Vertex3f(0.239175797f, 0.f, -1.f),
    C_Normal3f(-0.0488111973f, -0.998807967f, 0.f),
    C_Vertex3f(0.217597008f, 0.f, 0.f),
    C_Vertex3f(0.217597008f, 0.f, -1.f),
    C_Normal3f(-0.146208391f, -0.989253819f, 0.f),
    C_Vertex3f(0.195021272f, 0.00215780735f, 0.f),
    C_Vertex3f(0.195021272f, 0.00215780735f, -1.f),
    C_Normal3f(-0.253929794f, -0.967222691f, 0.f),
    C_Vertex3f(0.173797488f, 0.00647342205f, 0.f),
    C_Vertex3f(0.173797488f, 0.00647342205f, -1.f),
    C_Normal3f(-0.366118371f, -0.930568278f, 0.f),
    C_Vertex3f(0.153925657f, 0.0129468441f, 0.f),
    C_Vertex3f(0.153925657f, 0.0129468441f, -1.f),
    C_Normal3f(-0.477982938f, -0.878369093f, 0.f),
    C_Vertex3f(0.135405779f, 0.0215780735f, 0.f),
    C_Vertex3f(0.135405779f, 0.0215780735f, -1.f),
    C_Normal3f(-0.584101677f, -0.811680496f, 0.f),
    C_Vertex3f(0.118237853f, 0.0323671103f, 0.f),
    C_Vertex3f(0.118237853f, 0.0323671103f, -1.f),
    C_Normal3f(-0.679596663f, -0.733585954f, 0.f),
    C_Vertex3f(0.10242188f, 0.0453139544f, 0.f),
    C_Vertex3f(0.10242188f, 0.0453139544f, -1.f),
    C_Normal3f(-0.76119554f, -0.648522377f, 0.f),
    C_Vertex3f(0.087957859f, 0.0604186058f, 0.f),
    C_Vertex3f(0.087957859f, 0.0604186058f, -1.f),
    C_Normal3f(-0.821239054f, -0.570584238f, 0.f),
    C_Vertex3f(0.0748457909f, 0.0776810646f, 0.f),
    C_Vertex3f(0.0748457909f, 0.0776810646f, -1.f),
    C_Normal3f(-0.867196918f, -0.497965336f, 0.f),
    C_Vertex3f(0.0633727312f, 0.0958038568f, 0.f),
    C_Vertex3f(0.0633727312f, 0.0958038568f, -1.f),
    C_Normal3f(-0.906986952f, -0.421158671f, 0.f),
    C_Vertex3f(0.0535386801f, 0.114786983f, 0.f),
    C_Vertex3f(0.0535386801f, 0.114786983f, -1.f),
    C_Normal3f(-0.939744115f, -0.341878593f, 0.f),
    C_Vertex3f(0.0453436375f, 0.134630442f, 0.f),
    C_Vertex3f(0.0453436375f, 0.134630442f, -1.f),
    C_Normal3f(-0.965079367f, -0.261957854f, 0.f),
    C_Vertex3f(0.0387876034f, 0.155334234f, 0.f),
    C_Vertex3f(0.0387876034f, 0.155334234f, -1.f),
    C_Normal3f(-0.983085334f, -0.183148101f, 0.f),
    C_Vertex3f(0.0338705778f, 0.17689836f, 0.f),
    C_Vertex3f(0.0338705778f, 0.17689836f, -1.f),
    C_Normal3f(-0.994263887f, -0.106954731f, 0.f),
    C_Vertex3f(0.0305925608f, 0.19932282f, 0.f),
    C_Vertex3f(0.0305925608f, 0.19932282f, -1.f),
    C_Normal3f(-0.999502897f, -0.0315251648f, 0.f),
    C_Vertex3f(0.0289535522f, 0.222607613f, 0.f),
    C_Vertex3f(0.0289535522f, 0.222607613f, -1.f),
    C_Normal3f(-0.998774409f, 0.0494947657f, 0.f),
    C_Vertex3f(0.0289535522f, 0.25128746f, 0.f),
    C_Vertex3f(0.0289535522f, 0.25128746f, -1.f),
    C_Normal3f(-0.989791572f, 0.142522559f, 0.f),
    C_Vertex3f(0.0319571495f, 0.283218384f, 0.f),
    C_Vertex3f(0.0319571495f, 0.283218384f, -1.f),
    C_Normal3f(-0.970084429f, 0.242767766f, 0.f),
    C_Vertex3f(0.037964344f, 0.313865662f, 0.f),
    C_Vertex3f(0.037964344f, 0.313865662f, -1.f),
    C_Normal3f(-0.941664159f, 0.33655417f, 0.f),
    C_Vertex3f(0.0469751358f, 0.343229294f, 0.f),
    C_Vertex3f(0.0469751358f, 0.343229294f, -1.f),
    C_Normal3f(-0.909339249f, 0.416055351f, 0.f),
    C_Vertex3f(0.0559859276f, 0.364289284f, 0.f),
    C_Vertex3f(0.0559859276f, 0.364289284f, -1.f),
    C_Normal3f(-0.877416372f, 0.479729593f, 0.f),
    C_Vertex3f(0.0626690388f, 0.377530217f, 0.f),
    C_Vertex3f(0.0626690388f, 0.377530217f, -1.f),
    C_Normal3f(-0.842592478f, 0.538551569f, 0.f),
    C_Vertex3f(0.0700280666f, 0.389972091f, 0.f),
    C_Vertex3f(0.0700280666f, 0.389972091f, -1.f),
    C_Normal3f(-0.802034736f, 0.597277403f, 0.f),
    C_Vertex3f(0.0780630112f, 0.401614904f, 0.f),
    C_Vertex3f(0.0780630112f, 0.401614904f, -1.f),
    C_Normal3f(-0.75579083f, 0.654813111f, 0.f),
    C_Vertex3f(0.0867738724f, 0.412458658f, 0.f),
    C_Vertex3f(0.0867738724f, 0.412458658f, -1.f),
    C_Normal3f(-0.704196036f, 0.710005581f, 0.f),
    C_Vertex3f(0.0961606503f, 0.422503352f, 0.f),
    C_Vertex3f(0.0961606503f, 0.422503352f, -1.f),
    C_Normal3f(-0.647884309f, 0.761738718f, 0.f),
    C_Vertex3f(0.106223345f, 0.431748986f, 0.f),
    C_Vertex3f(0.106223345f, 0.431748986f, -1.f),
    C_Normal3f(-0.58775878f, 0.809036255f, 0.f),
    C_Vertex3f(0.116961956f, 0.44019556f, 0.f),
    C_Vertex3f(0.116961956f, 0.44019556f, -1.f),
    C_Normal3f(-0.523666203f, 0.851923585f, 0.f),
    C_Vertex3f(0.128376484f, 0.447843075f, 0.f),
    C_Vertex3f(0.128376484f, 0.447843075f, -1.f),
    C_Normal3f(-0.454959899f, 0.89051199f, 0.f),
    C_Vertex3f(0.140289426f, 0.45453465f, 0.f),
    C_Vertex3f(0.140289426f, 0.45453465f, -1.f),
    C_Normal3f(-0.383523136f, 0.923531234f, 0.f),
    C_Vertex3f(0.152700782f, 0.460270286f, 0.f),
    C_Vertex3f(0.152700782f, 0.460270286f, -1.f),
    C_Normal3f(-0.310722768f, 0.950500548f, 0.f),
    C_Vertex3f(0.165610552f, 0.465049982f, 0.f),
    C_Vertex3f(0.165610552f, 0.465049982f, -1.f),
    C_Normal3f(-0.237943888f, 0.971278906f, 0.f),
    C_Vertex3f(0.179018736f, 0.468873739f, 0.f),
    C_Vertex3f(0.179018736f, 0.468873739f, -1.f),
    C_Normal3f(-0.166468948f, 0.986046672f, 0.f),
    C_Vertex3f(0.192925334f, 0.471741557f, 0.f),
    C_Vertex3f(0.192925334f, 0.471741557f, -1.f),
    C_Normal3f(-0.0973844752f, 0.995246887f, 0.f),
    C_Vertex3f(0.207330346f, 0.473653436f, 0.f),
    C_Vertex3f(0.207330346f, 0.473653436f, -1.f),
    C_Normal3f(-0.029068714f, 0.999577343f, 0.f),
    C_Vertex3f(0.222233772f, 0.474609375f, 0.f),
    C_Vertex3f(0.222233772f, 0.474609375f, -1.f),
    C_Normal3f(0.0551050007f, 0.998480558f, 0.f),
    C_Vertex3f(0.24020195f, 0.474609375f, 0.f),
    C_Vertex3f(0.24020195f, 0.474609375f, -1.f),
    C_Normal3f(0.160172939f, 0.987089038f, 0.f),
    C_Vertex3f(0.259835482f, 0.47253418f, 0.f),
    C_Vertex3f(0.259835482f, 0.47253418f, -1.f),
    C_Normal3f(0.27299577f, 0.962015271f, 0.f),
    C_Vertex3f(0.278568029f, 0.468383789f, 0.f),
    C_Vertex3f(0.278568029f, 0.468383789f, -1.f),
    C_Normal3f(0.385568261f, 0.922679365f, 0.f),
    C_Vertex3f(0.296399593f, 0.462158203f, 0.f),
    C_Vertex3f(0.296399593f, 0.462158203f, -1.f),
    C_Normal3f(0.492999107f, 0.870029867f, 0.f),
    C_Vertex3f(0.313330173f, 0.453857422f, 0.f),
    C_Vertex3f(0.313330173f, 0.453857422f, -1.f),
    C_Normal3f(0.590989769f, 0.80667907f, 0.f),
    C_Vertex3f(0.32935977f, 0.443481445f, 0.f),
    C_Vertex3f(0.32935977f, 0.443481445f, -1.f),
    C_Normal3f(0.676646113f, 0.736308336f, 0.f),
    C_Vertex3f(0.344488382f, 0.431030273f, 0.f),
    C_Vertex3f(0.344488382f, 0.431030273f, -1.f),
    C_Normal3f(0.737700224f, 0.67512852f, 0.f),
    C_Vertex3f(0.358716011f, 0.416503906f, 0.f),
    C_Vertex3f(0.358716011f, 0.416503906f, -1.f),
    C_Normal3f(0.779827595f, 0.625994444f, 0.f),
    C_Vertex3f(0.365379333f, 0.408203125f, 0.f),
    C_Vertex3f(0.365379333f, 0.408203125f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.365379333f, 0.408203125f, 0.f),
    C_Vertex3f(0.365379333f, 0.408203125f, -1.f),
    C_Vertex3f(0.365379333f, 0.464355469f, -1.f),
    C_Vertex3f(0.365379333f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.365379333f, 0.464355469f, 0.f),
    C_Vertex3f(0.365379333f, 0.464355469f, -1.f),
    C_Vertex3f(0.43813324f, 0.464355469f, -1.f),
    C_Vertex3f(0.43813324f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.43813324f, 0.464355469f, 0.f),
    C_Vertex3f(0.43813324f, 0.464355469f, -1.f),
    C_Normal3f(0.999998808f, -0.00156923383f, 0.f),
    C_Vertex3f(0.43813324f, 0.0497026443f, 0.f),
    C_Vertex3f(0.43813324f, 0.0497026443f, -1.f),
    C_Normal3f(0.999081135f, -0.0428586937f, 0.f),
    C_Vertex3f(0.437443137f, 0.0245856047f, 0.f),
    C_Vertex3f(0.437443137f, 0.0245856047f, -1.f),
    C_Normal3f(0.996982336f, -0.077628389f, 0.f),
    C_Vertex3f(0.436062932f, 0.00144159794f, 0.f),
    C_Vertex3f(0.436062932f, 0.00144159794f, -1.f),
    C_Normal3f(0.992916167f, -0.118816622f, 0.f),
    C_Vertex3f(0.433992624f, -0.0197293758f, 0.f),
    C_Vertex3f(0.433992624f, -0.0197293758f, -1.f),
    C_Normal3f(0.985770583f, -0.168096289f, 0.f),
    C_Vertex3f(0.431232214f, -0.0389273167f, 0.f),
    C_Vertex3f(0.431232214f, -0.0389273167f, -1.f),
    C_Normal3f(0.973753393f, -0.227605283f, 0.f),
    C_Vertex3f(0.427781701f, -0.0561522245f, 0.f),
    C_Vertex3f(0.427781701f, -0.0561522245f, -1.f),
    C_Normal3f(0.953950346f, -0.299964666f, 0.f),
    C_Vertex3f(0.423641086f, -0.0714040995f, 0.f),
    C_Vertex3f(0.423641086f, -0.0714040995f, -1.f),
    C_Normal3f(0.921633959f, -0.388060391f, 0.f),
    C_Vertex3f(0.418810368f, -0.0846829414f, 0.f),
    C_Vertex3f(0.418810368f, -0.0846829414f, -1.f),
    C_Normal3f(0.88060087f, -0.473858684f, 0.f),
    C_Vertex3f(0.413289547f, -0.0959887505f, 0.f),
    C_Vertex3f(0.413289547f, -0.0959887505f, -1.f),
    C_Normal3f(0.83989501f, -0.542748928f, 0.f),
    C_Vertex3f(0.406962514f, -0.10670054f, 0.f),
    C_Vertex3f(0.406962514f, -0.10670054f, -1.f),
    C_Normal3f(0.793329239f, -0.608792841f, 0.f),
    C_Vertex3f(0.399829268f, -0.116818309f, 0.f),
    C_Vertex3f(0.399829268f, -0.116818309f, -1.f),
    C_Normal3f(0.74175328f, -0.670672894f, 0.f),
    C_Vertex3f(0.391889811f, -0.126342058f, 0.f),
    C_Vertex3f(0.391889811f, -0.126342058f, -1.f),
    C_Normal3f(0.686296046f, -0.72732228f, 0.f),
    C_Vertex3f(0.38314414f, -0.135271788f, 0.f),
    C_Vertex3f(0.38314414f, -0.135271788f, -1.f),
    C_Normal3f(0.628248751f, -0.778012574f, 0.f),
    C_Vertex3f(0.373592257f, -0.143607497f, 0.f),
    C_Vertex3f(0.373592257f, -0.143607497f, -1.f),
    C_Normal3f(0.568931818f, -0.822384596f, 0.f),
    C_Vertex3f(0.363234162f, -0.151349187f, 0.f),
    C_Vertex3f(0.363234162f, -0.151349187f, -1.f),
    C_Normal3f(0.50957644f, -0.860425413f, 0.f),
    C_Vertex3f(0.352069855f, -0.158496857f, 0.f),
    C_Vertex3f(0.352069855f, -0.158496857f, -1.f),
    C_Normal3f(0.446416527f, -0.89482522f, 0.f),
    C_Vertex3f(0.340099335f, -0.165050507f, 0.f),
    C_Vertex3f(0.340099335f, -0.165050507f, -1.f),
    C_Normal3f(0.37890169f, -0.925436974f, 0.f),
    C_Vertex3f(0.327438831f, -0.17078495f, 0.f),
    C_Vertex3f(0.327438831f, -0.17078495f, -1.f),
    C_Normal3f(0.312509567f, -0.949914634f, 0.f),
    C_Vertex3f(0.314088345f, -0.175700188f, 0.f),
    C_Vertex3f(0.314088345f, -0.175700188f, -1.f),
    C_Normal3f(0.248239383f, -0.96869874f, 0.f),
    C_Vertex3f(0.300047874f, -0.179796219f, 0.f),
    C_Vertex3f(0.300047874f, -0.179796219f, -1.f),
    C_Normal3f(0.186842233f, -0.982389927f, 0.f),
    C_Vertex3f(0.285317421f, -0.183073044f, 0.f),
    C_Vertex3f(0.285317421f, -0.183073044f, -1.f),
    C_Normal3f(0.128823057f, -0.991667628f, 0.f),
    C_Vertex3f(0.269896984f, -0.185530663f, 0.f),
    C_Vertex3f(0.269896984f, -0.185530663f, -1.f),
    C_Normal3f(0.0744677484f, -0.997223437f, 0.f),
    C_Vertex3f(0.253786564f, -0.187169075f, 0.f),
    C_Vertex3f(0.253786564f, -0.187169075f, -1.f),
    C_Normal3f(0.0227934364f, -0.999740243f, 0.f),
    C_Vertex3f(0.23698616f, -0.187988281f, 0.f),
    C_Vertex3f(0.23698616f, -0.187988281f, -1.f),
    C_Normal3f(-0.0300000943f, -0.999549866f, 0.f),
    C_Vertex3f(0.217855453f, -0.187988281f, 0.f),
    C_Vertex3f(0.217855453f, -0.187988281f, -1.f),
    C_Normal3f(-0.0904965997f, -0.995896816f, 0.f),
    C_Vertex3f(0.19808197f, -0.186820626f, 0.f),
    C_Vertex3f(0.19808197f, -0.186820626f, -1.f),
    C_Normal3f(-0.157716066f, -0.987484515f, 0.f),
    C_Vertex3f(0.17930603f, -0.184485316f, 0.f),
    C_Vertex3f(0.17930603f, -0.184485316f, -1.f),
    C_Normal3f(-0.23015973f, -0.973152876f, 0.f),
    C_Vertex3f(0.161527634f, -0.180982351f, 0.f),
    C_Vertex3f(0.161527634f, -0.180982351f, -1.f),
    C_Normal3f(-0.307117254f, -0.9516716f, 0.f),
    C_Vertex3f(0.14474678f, -0.176311731f, 0.f),
    C_Vertex3f(0.14474678f, -0.176311731f, -1.f),
    C_Normal3f(-0.38736558f, -0.921926141f, 0.f),
    C_Vertex3f(0.12896347f, -0.170473456f, 0.f),
    C_Vertex3f(0.12896347f, -0.170473456f, -1.f),
    C_Normal3f(-0.469145149f, -0.883121014f, 0.f),
    C_Vertex3f(0.114177704f, -0.163467526f, 0.f),
    C_Vertex3f(0.114177704f, -0.163467526f, -1.f),
    C_Normal3f(-0.550245762f, -0.83500284f, 0.f),
    C_Vertex3f(0.100389481f, -0.155293941f, 0.f),
    C_Vertex3f(0.100389481f, -0.155293941f, -1.f),
    C_Normal3f(-0.638741493f, -0.769421458f, 0.f),
    C_Vertex3f(0.0875988007f, -0.145952702f, 0.f),
    C_Vertex3f(0.0875988007f, -0.145952702f, -1.f),
    C_Normal3f(-0.733093202f, -0.680128157f, 0.f),
    C_Vertex3f(0.0764615536f, -0.135429978f, 0.f),
    C_Vertex3f(0.0764615536f, -0.135429978f, -1.f),
    C_Normal3f(-0.817646861f, -0.575720131f, 0.f),
    C_Vertex3f(0.0669777393f, -0.123725772f, 0.f),
    C_Vertex3f(0.0669777393f, -0.123725772f, -1.f),
    C_Normal3f(-0.887326777f, -0.461141318f, 0.f),
    C_Vertex3f(0.0591473579f, -0.110840082f, 0.f),
    C_Vertex3f(0.0591473579f, -0.110840082f, -1.f),
    C_Normal3f(-0.939379394f, -0.342879444f, 0.f),
    C_Vertex3f(0.0529704094f, -0.0967729092f, 0.f),
    C_Vertex3f(0.0529704094f, -0.0967729092f, -1.f),
    C_Normal3f(-0.973828435f, -0.227284402f, 0.f),
    C_Vertex3f(0.0484468937f, -0.0815242529f, 0.f),
    C_Vertex3f(0.0484468937f, -0.0815242529f, -1.f),
    C_Normal3f(-0.992870867f, -0.119194709f, 0.f),
    C_Vertex3f(0.0455768108f, -0.0650941133f, 0.f),
    C_Vertex3f(0.0455768108f, -0.0650941133f, -1.f),
    C_Normal3f(-0.999317646f, -0.0369361266f, 0.f),
    C_Vertex3f(0.0443601608f, -0.0474824905f, 0.f),
    C_Vertex3f(0.0443601608f, -0.0474824905f, -1.f),
    C_Normal3f(-0.999729931f, 0.0232353769f, 0.f),
    C_Vertex3f(0.0445785522f, -0.0380859375f, 0.f),
    C_Vertex3f(0.0445785522f, -0.0380859375f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999621809f, -0.0275010262f, 0.f),
    C_Vertex3f(0.11000824f, 0.250864029f, 0.f),
    C_Vertex3f(0.11000824f, 0.250864029f, -1.f),
    C_Normal3f(0.999657631f, 0.0261642728f, 0.f),
    C_Vertex3f(0.11000824f, 0.228935242f, 0.f),
    C_Vertex3f(0.11000824f, 0.228935242f, -1.f),
    C_Normal3f(0.996601939f, 0.0823688656f, 0.f),
    C_Vertex3f(0.111140013f, 0.207622409f, 0.f),
    C_Vertex3f(0.111140013f, 0.207622409f, -1.f),
    C_Normal3f(0.98908782f, 0.147326738f, 0.f),
    C_Vertex3f(0.113403559f, 0.187854409f, 0.f),
    C_Vertex3f(0.113403559f, 0.187854409f, -1.f),
    C_Normal3f(0.975191653f, 0.221362099f, 0.f),
    C_Vertex3f(0.116798878f, 0.169631243f, 0.f),
    C_Vertex3f(0.116798878f, 0.169631243f, -1.f),
    C_Normal3f(0.952370584f, 0.304943323f, 0.f),
    C_Vertex3f(0.12132597f, 0.152952909f, 0.f),
    C_Vertex3f(0.12132597f, 0.152952909f, -1.f),
    C_Normal3f(0.91751802f, 0.397694319f, 0.f),
    C_Vertex3f(0.126984835f, 0.137819409f, 0.f),
    C_Vertex3f(0.126984835f, 0.137819409f, -1.f),
    C_Normal3f(0.8672809f, 0.497819036f, 0.f),
    C_Vertex3f(0.133775473f, 0.124230742f, 0.f),
    C_Vertex3f(0.133775473f, 0.124230742f, -1.f),
    C_Normal3f(0.7988168f, 0.601574302f, 0.f),
    C_Vertex3f(0.141697884f, 0.112186909f, 0.f),
    C_Vertex3f(0.141697884f, 0.112186909f, -1.f),
    C_Normal3f(0.725372732f, 0.688356221f, 0.f),
    C_Vertex3f(0.150752068f, 0.101687908f, 0.f),
    C_Vertex3f(0.150752068f, 0.101687908f, -1.f),
    C_Normal3f(0.652262568f, 0.757993162f, 0.f),
    C_Vertex3f(0.160378933f, 0.0925012827f, 0.f),
    C_Vertex3f(0.160378933f, 0.0925012827f, -1.f),
    C_Normal3f(0.567009628f, 0.823711216f, 0.f),
    C_Vertex3f(0.17057848f, 0.0846270323f, 0.f),
    C_Vertex3f(0.17057848f, 0.0846270323f, -1.f),
    C_Normal3f(0.471071541f, 0.88209492f, 0.f),
    C_Vertex3f(0.181350708f, 0.0780651569f, 0.f),
    C_Vertex3f(0.181350708f, 0.0780651569f, -1.f),
    C_Normal3f(0.367306948f, 0.930099726f, 0.f),
    C_Vertex3f(0.192695618f, 0.0728156567f, 0.f),
    C_Vertex3f(0.192695618f, 0.0728156567f, -1.f),
    C_Normal3f(0.259624004f, 0.965709746f, 0.f),
    C_Vertex3f(0.204613209f, 0.0688785315f, 0.f),
    C_Vertex3f(0.204613209f, 0.0688785315f, -1.f),
    C_Normal3f(0.152278572f, 0.988337576f, 0.f),
    C_Vertex3f(0.217103481f, 0.0662537813f, 0.f),
    C_Vertex3f(0.217103481f, 0.0662537813f, -1.f),
    C_Normal3f(0.049195867f, 0.998789132f, 0.f),
    C_Vertex3f(0.230166435f, 0.0649414063f, 0.f),
    C_Vertex3f(0.230166435f, 0.0649414063f, -1.f),
    C_Normal3f(-0.0490914211f, 0.998794317f, 0.f),
    C_Vertex3f(0.243747711f, 0.0649414063f, 0.f),
    C_Vertex3f(0.243747711f, 0.0649414063f, -1.f),
    C_Normal3f(-0.152297512f, 0.988334656f, 0.f),
    C_Vertex3f(0.256729245f, 0.0662469864f, 0.f),
    C_Vertex3f(0.256729245f, 0.0662469864f, -1.f),
    C_Normal3f(-0.259171724f, 0.96583122f, 0.f),
    C_Vertex3f(0.269165397f, 0.0688581467f, 0.f),
    C_Vertex3f(0.269165397f, 0.0688581467f, -1.f),
    C_Normal3f(-0.3660191f, 0.930607319f, 0.f),
    C_Vertex3f(0.281056166f, 0.0727748871f, 0.f),
    C_Vertex3f(0.281056166f, 0.0727748871f, -1.f),
    C_Normal3f(-0.468703538f, 0.883355558f, 0.f),
    C_Vertex3f(0.292401552f, 0.0779972076f, 0.f),
    C_Vertex3f(0.292401552f, 0.0779972076f, -1.f),
    C_Normal3f(-0.563496292f, 0.826118588f, 0.f),
    C_Vertex3f(0.303201556f, 0.0845251083f, 0.f),
    C_Vertex3f(0.303201556f, 0.0845251083f, -1.f),
    C_Normal3f(-0.647720635f, 0.761877894f, 0.f),
    C_Vertex3f(0.313456178f, 0.0923585892f, 0.f),
    C_Vertex3f(0.313456178f, 0.0923585892f, -1.f),
    C_Normal3f(-0.720049858f, 0.6939224f, 0.f),
    C_Vertex3f(0.323165417f, 0.10149765f, 0.f),
    C_Vertex3f(0.323165417f, 0.10149765f, -1.f),
    C_Normal3f(-0.793092728f, 0.609100997f, 0.f),
    C_Vertex3f(0.332329273f, 0.111942291f, 0.f),
    C_Vertex3f(0.332329273f, 0.111942291f, -1.f),
    C_Normal3f(-0.862141669f, 0.506667137f, 0.f),
    C_Vertex3f(0.340347648f, 0.123870134f, 0.f),
    C_Vertex3f(0.340347648f, 0.123870134f, -1.f),
    C_Normal3f(-0.913569748f, 0.406682104f, 0.f),
    C_Vertex3f(0.34722054f, 0.137281179f, 0.f),
    C_Vertex3f(0.34722054f, 0.137281179f, -1.f),
    C_Normal3f(-0.949721873f, 0.313094765f, 0.f),
    C_Vertex3f(0.35294795f, 0.152175426f, 0.f),
    C_Vertex3f(0.35294795f, 0.152175426f, -1.f),
    C_Normal3f(-0.973655283f, 0.228024706f, 0.f),
    C_Vertex3f(0.357529879f, 0.168552876f, 0.f),
    C_Vertex3f(0.357529879f, 0.168552876f, -1.f),
    C_Normal3f(-0.988357186f, 0.152151659f, 0.f),
    C_Vertex3f(0.360966325f, 0.186413527f, 0.f),
    C_Vertex3f(0.360966325f, 0.186413527f, -1.f),
    C_Normal3f(-0.996361017f, 0.085234344f, 0.f),
    C_Vertex3f(0.363257289f, 0.20575738f, 0.f),
    C_Vertex3f(0.363257289f, 0.20575738f, -1.f),
    C_Normal3f(-0.999639392f, 0.0268504545f, 0.f),
    C_Vertex3f(0.364402771f, 0.226584435f, 0.f),
    C_Vertex3f(0.364402771f, 0.226584435f, -1.f),
    C_Normal3f(-0.999602795f, -0.0281850658f, 0.f),
    C_Vertex3f(0.364402771f, 0.248403549f, 0.f),
    C_Vertex3f(0.364402771f, 0.248403549f, -1.f),
    C_Normal3f(-0.995849252f, -0.0910178721f, 0.f),
    C_Vertex3f(0.363223195f, 0.268418908f, 0.f),
    C_Vertex3f(0.363223195f, 0.268418908f, -1.f),
    C_Normal3f(-0.986910403f, -0.161269456f, 0.f),
    C_Vertex3f(0.360864043f, 0.287121654f, 0.f),
    C_Vertex3f(0.360864043f, 0.287121654f, -1.f),
    C_Normal3f(-0.97088784f, -0.239534348f, 0.f),
    C_Vertex3f(0.357325315f, 0.304511786f, 0.f),
    C_Vertex3f(0.357325315f, 0.304511786f, -1.f),
    C_Normal3f(-0.94555372f, -0.325466365f, 0.f),
    C_Vertex3f(0.352607012f, 0.320589304f, 0.f),
    C_Vertex3f(0.352607012f, 0.320589304f, -1.f),
    C_Normal3f(-0.908546388f, -0.417784035f, 0.f),
    C_Vertex3f(0.346709132f, 0.335354209f, 0.f),
    C_Vertex3f(0.346709132f, 0.335354209f, -1.f),
    C_Normal3f(-0.857798815f, -0.513985634f, 0.f),
    C_Vertex3f(0.339631677f, 0.3488065f, 0.f),
    C_Vertex3f(0.339631677f, 0.3488065f, -1.f),
    C_Normal3f(-0.792175114f, -0.610293806f, 0.f),
    C_Vertex3f(0.331374645f, 0.360946178f, 0.f),
    C_Vertex3f(0.331374645f, 0.360946178f, -1.f),
    C_Normal3f(-0.723727584f, -0.690085828f, 0.f),
    C_Vertex3f(0.321938038f, 0.371773243f, 0.f),
    C_Vertex3f(0.321938038f, 0.371773243f, -1.f),
    C_Normal3f(-0.654487073f, -0.756073177f, 0.f),
    C_Vertex3f(0.31201756f, 0.381246924f, 0.f),
    C_Vertex3f(0.31201756f, 0.381246924f, -1.f),
    C_Normal3f(-0.573008001f, -0.819549739f, 0.f),
    C_Vertex3f(0.301613212f, 0.389367223f, 0.f),
    C_Vertex3f(0.301613212f, 0.389367223f, -1.f),
    C_Normal3f(-0.480019867f, -0.877257586f, 0.f),
    C_Vertex3f(0.290724993f, 0.396134138f, 0.f),
    C_Vertex3f(0.290724993f, 0.396134138f, -1.f),
    C_Normal3f(-0.377652496f, -0.925947428f, 0.f),
    C_Vertex3f(0.279352903f, 0.40154767f, 0.f),
    C_Vertex3f(0.279352903f, 0.40154767f, -1.f),
    C_Normal3f(-0.269338131f, -0.963045716f, 0.f),
    C_Vertex3f(0.267496943f, 0.40560782f, 0.f),
    C_Vertex3f(0.267496943f, 0.40560782f, -1.f),
    C_Normal3f(-0.159290373f, -0.987231731f, 0.f),
    C_Vertex3f(0.255157113f, 0.408314586f, 0.f),
    C_Vertex3f(0.255157113f, 0.408314586f, -1.f),
    C_Normal3f(-0.0519378856f, -0.998650312f, 0.f),
    C_Vertex3f(0.242333412f, 0.409667969f, 0.f),
    C_Vertex3f(0.242333412f, 0.409667969f, -1.f),
    C_Normal3f(0.0515967645f, -0.998668015f, 0.f),
    C_Vertex3f(0.22913456f, 0.409667969f, 0.f),
    C_Vertex3f(0.22913456f, 0.409667969f, -1.f),
    C_Normal3f(0.159757718f, -0.987156272f, 0.f),
    C_Vertex3f(0.216535211f, 0.40833509f, 0.f),
    C_Vertex3f(0.216535211f, 0.40833509f, -1.f),
    C_Normal3f(0.2704207f, -0.962742209f, 0.f),
    C_Vertex3f(0.204426646f, 0.405669332f, 0.f),
    C_Vertex3f(0.204426646f, 0.405669332f, -1.f),
    C_Normal3f(0.379520237f, -0.925183356f, 0.f),
    C_Vertex3f(0.192808867f, 0.401670694f, 0.f),
    C_Vertex3f(0.192808867f, 0.401670694f, -1.f),
    C_Normal3f(0.482726425f, -0.875771284f, 0.f),
    C_Vertex3f(0.181681871f, 0.396339178f, 0.f),
    C_Vertex3f(0.181681871f, 0.396339178f, -1.f),
    C_Normal3f(0.576480567f, -0.817110837f, 0.f),
    C_Vertex3f(0.171045661f, 0.389674783f, 0.f),
    C_Vertex3f(0.171045661f, 0.389674783f, -1.f),
    C_Normal3f(0.658555627f, -0.752532065f, 0.f),
    C_Vertex3f(0.160900235f, 0.381677508f, 0.f),
    C_Vertex3f(0.160900235f, 0.381677508f, -1.f),
    C_Normal3f(0.728174925f, -0.685391426f, 0.f),
    C_Vertex3f(0.151245594f, 0.372347355f, 0.f),
    C_Vertex3f(0.151245594f, 0.372347355f, -1.f),
    C_Normal3f(0.796290755f, -0.604914069f, 0.f),
    C_Vertex3f(0.142081738f, 0.361684322f, 0.f),
    C_Vertex3f(0.142081738f, 0.361684322f, -1.f),
    C_Normal3f(0.860942304f, -0.508702517f, 0.f),
    C_Vertex3f(0.134063363f, 0.34972918f, 0.f),
    C_Vertex3f(0.134063363f, 0.34972918f, -1.f),
    C_Normal3f(0.910736263f, -0.412988544f, 0.f),
    C_Vertex3f(0.127190471f, 0.336481929f, 0.f),
    C_Vertex3f(0.127190471f, 0.336481929f, -1.f),
    C_Normal3f(0.946931779f, -0.321434736f, 0.f),
    C_Vertex3f(0.12146306f, 0.321942568f, 0.f),
    C_Vertex3f(0.12146306f, 0.321942568f, -1.f),
    C_Normal3f(0.97165215f, -0.236415073f, 0.f),
    C_Vertex3f(0.116881132f, 0.306111097f, 0.f),
    C_Vertex3f(0.116881132f, 0.306111097f, -1.f),
    C_Normal3f(0.987261772f, -0.159104377f, 0.f),
    C_Vertex3f(0.113444686f, 0.288987517f, 0.f),
    C_Vertex3f(0.113444686f, 0.288987517f, -1.f),
    C_Normal3f(0.995962024f, -0.0897758827f, 0.f),
    C_Vertex3f(0.111153722f, 0.270571828f, 0.f),
    C_Vertex3f(0.111153722f, 0.270571828f, -1.f),
    C_Normal3f(0.999621809f, -0.0275010262f, 0.f),
    C_Vertex3f(0.11000824f, 0.250864029f, 0.f),
    C_Vertex3f(0.11000824f, 0.250864029f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.364402771f, 0.248403549f, 0.f),
    C_Vertex3f(0.364402771f, 0.226584435f, 0.f),
    C_Vertex3f(0.373592257f, -0.143607497f, 0.f),
    C_Vertex3f(0.363257289f, 0.20575738f, 0.f),
    C_Vertex3f(0.363234162f, -0.151349187f, 0.f),
    C_Vertex3f(0.360966325f, 0.186413527f, 0.f),
    C_Vertex3f(0.358110428f, 0.0509529114f, 0.f),
    C_Vertex3f(0.358055115f, 0.060546875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.360966325f, 0.186413527f, 0.f),
    C_Vertex3f(0.357529879f, 0.168552876f, 0.f),
    C_Vertex3f(0.358055115f, 0.060546875f, 0.f),
    C_Vertex3f(0.35294795f, 0.152175426f, 0.f),
    C_Vertex3f(0.351608276f, 0.0529785156f, 0.f),
    C_Vertex3f(0.34722054f, 0.137281179f, 0.f),
    C_Vertex3f(0.340347648f, 0.123870134f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.351608276f, 0.0529785156f, 0.f),
    C_Vertex3f(0.340347648f, 0.123870134f, 0.f),
    C_Vertex3f(0.337922573f, 0.0397338867f, 0.f),
    C_Vertex3f(0.332329273f, 0.111942291f, 0.f),
    C_Vertex3f(0.323444843f, 0.0283813477f, 0.f),
    C_Vertex3f(0.323165417f, 0.10149765f, 0.f),
    C_Vertex3f(0.313456178f, 0.0923585892f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.308175087f, 0.0189208984f, 0.f),
    C_Vertex3f(0.323444843f, 0.0283813477f, 0.f),
    C_Vertex3f(0.313456178f, 0.0923585892f, 0.f),
    C_Vertex3f(0.303201556f, 0.0845251083f, 0.f),
    C_Vertex3f(0.292401552f, 0.0779972076f, 0.f),
    C_Vertex3f(0.292113304f, 0.0113525391f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.256729245f, 0.0662469864f, 0.f),
    C_Vertex3f(0.257613659f, 0.00189208984f, 0.f),
    C_Vertex3f(0.269165397f, 0.0688581467f, 0.f),
    C_Vertex3f(0.275259495f, 0.00567626953f, 0.f),
    C_Vertex3f(0.281056166f, 0.0727748871f, 0.f),
    C_Vertex3f(0.292113304f, 0.0113525391f, 0.f),
    C_Vertex3f(0.292401552f, 0.0779972076f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.217103481f, 0.0662537813f, 0.f),
    C_Vertex3f(0.217597008f, 0.f, 0.f),
    C_Vertex3f(0.230166435f, 0.0649414063f, 0.f),
    C_Vertex3f(0.239175797f, 0.f, 0.f),
    C_Vertex3f(0.243747711f, 0.0649414063f, 0.f),
    C_Vertex3f(0.257613659f, 0.00189208984f, 0.f),
    C_Vertex3f(0.256729245f, 0.0662469864f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.204613209f, 0.0688785315f, 0.f),
    C_Vertex3f(0.192695618f, 0.0728156567f, 0.f),
    C_Vertex3f(0.195021272f, 0.00215780735f, 0.f),
    C_Vertex3f(0.217597008f, 0.f, 0.f),
    C_Vertex3f(0.217103481f, 0.0662537813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.181350708f, 0.0780651569f, 0.f),
    C_Vertex3f(0.17057848f, 0.0846270323f, 0.f),
    C_Vertex3f(0.173797488f, 0.00647342205f, 0.f),
    C_Vertex3f(0.195021272f, 0.00215780735f, 0.f),
    C_Vertex3f(0.192695618f, 0.0728156567f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.160378933f, 0.0925012827f, 0.f),
    C_Vertex3f(0.150752068f, 0.101687908f, 0.f),
    C_Vertex3f(0.153925657f, 0.0129468441f, 0.f),
    C_Vertex3f(0.173797488f, 0.00647342205f, 0.f),
    C_Vertex3f(0.17057848f, 0.0846270323f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.141697884f, 0.112186909f, 0.f),
    C_Vertex3f(0.133775473f, 0.124230742f, 0.f),
    C_Vertex3f(0.135405779f, 0.0215780735f, 0.f),
    C_Vertex3f(0.153925657f, 0.0129468441f, 0.f),
    C_Vertex3f(0.150752068f, 0.101687908f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133775473f, 0.124230742f, 0.f),
    C_Vertex3f(0.126984835f, 0.137819409f, 0.f),
    C_Vertex3f(0.135405779f, 0.0215780735f, 0.f),
    C_Vertex3f(0.12132597f, 0.152952909f, 0.f),
    C_Vertex3f(0.118237853f, 0.0323671103f, 0.f),
    C_Vertex3f(0.116798878f, 0.169631243f, 0.f),
    C_Vertex3f(0.113403559f, 0.187854409f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.113403559f, 0.187854409f, 0.f),
    C_Vertex3f(0.111140013f, 0.207622409f, 0.f),
    C_Vertex3f(0.118237853f, 0.0323671103f, 0.f),
    C_Vertex3f(0.11000824f, 0.228935242f, 0.f),
    C_Vertex3f(0.10242188f, 0.0453139544f, 0.f),
    C_Vertex3f(0.106223345f, 0.431748986f, 0.f),
    C_Vertex3f(0.0961606503f, 0.422503352f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.087957859f, 0.0604186058f, 0.f),
    C_Vertex3f(0.10242188f, 0.0453139544f, 0.f),
    C_Vertex3f(0.0961606503f, 0.422503352f, 0.f),
    C_Vertex3f(0.0867738724f, 0.412458658f, 0.f),
    C_Vertex3f(0.0780630112f, 0.401614904f, 0.f),
    C_Vertex3f(0.0748457909f, 0.0776810646f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0700280666f, 0.389972091f, 0.f),
    C_Vertex3f(0.0626690388f, 0.377530217f, 0.f),
    C_Vertex3f(0.0633727312f, 0.0958038568f, 0.f),
    C_Vertex3f(0.0748457909f, 0.0776810646f, 0.f),
    C_Vertex3f(0.0780630112f, 0.401614904f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0387876034f, 0.155334234f, 0.f),
    C_Vertex3f(0.0453436375f, 0.134630442f, 0.f),
    C_Vertex3f(0.0469751358f, 0.343229294f, 0.f),
    C_Vertex3f(0.0535386801f, 0.114786983f, 0.f),
    C_Vertex3f(0.0559859276f, 0.364289284f, 0.f),
    C_Vertex3f(0.0633727312f, 0.0958038568f, 0.f),
    C_Vertex3f(0.0626690388f, 0.377530217f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0289535522f, 0.25128746f, 0.f),
    C_Vertex3f(0.0305925608f, 0.19932282f, 0.f),
    C_Vertex3f(0.0319571495f, 0.283218384f, 0.f),
    C_Vertex3f(0.0338705778f, 0.17689836f, 0.f),
    C_Vertex3f(0.037964344f, 0.313865662f, 0.f),
    C_Vertex3f(0.0387876034f, 0.155334234f, 0.f),
    C_Vertex3f(0.0469751358f, 0.343229294f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.365379333f, 0.408203125f, 0.f),
    C_Vertex3f(0.358716011f, 0.416503906f, 0.f),
    C_Vertex3f(0.360864043f, 0.287121654f, 0.f),
    C_Vertex3f(0.363223195f, 0.268418908f, 0.f),
    C_Vertex3f(0.364402771f, 0.248403549f, 0.f),
    C_Vertex3f(0.373592257f, -0.143607497f, 0.f),
    C_Vertex3f(0.365379333f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.358716011f, 0.416503906f, 0.f),
    C_Vertex3f(0.344488382f, 0.431030273f, 0.f),
    C_Vertex3f(0.346709132f, 0.335354209f, 0.f),
    C_Vertex3f(0.352607012f, 0.320589304f, 0.f),
    C_Vertex3f(0.357325315f, 0.304511786f, 0.f),
    C_Vertex3f(0.360864043f, 0.287121654f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.301613212f, 0.389367223f, 0.f),
    C_Vertex3f(0.31201756f, 0.381246924f, 0.f),
    C_Vertex3f(0.313330173f, 0.453857422f, 0.f),
    C_Vertex3f(0.321938038f, 0.371773243f, 0.f),
    C_Vertex3f(0.32935977f, 0.443481445f, 0.f),
    C_Vertex3f(0.331374645f, 0.360946178f, 0.f),
    C_Vertex3f(0.344488382f, 0.431030273f, 0.f),
    C_Vertex3f(0.339631677f, 0.3488065f, 0.f),
    C_Vertex3f(0.346709132f, 0.335354209f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.296399593f, 0.462158203f, 0.f),
    C_Vertex3f(0.278568029f, 0.468383789f, 0.f),
    C_Vertex3f(0.279352903f, 0.40154767f, 0.f),
    C_Vertex3f(0.290724993f, 0.396134138f, 0.f),
    C_Vertex3f(0.301613212f, 0.389367223f, 0.f),
    C_Vertex3f(0.313330173f, 0.453857422f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.242333412f, 0.409667969f, 0.f),
    C_Vertex3f(0.255157113f, 0.408314586f, 0.f),
    C_Vertex3f(0.259835482f, 0.47253418f, 0.f),
    C_Vertex3f(0.267496943f, 0.40560782f, 0.f),
    C_Vertex3f(0.278568029f, 0.468383789f, 0.f),
    C_Vertex3f(0.279352903f, 0.40154767f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.181681871f, 0.396339178f, 0.f),
    C_Vertex3f(0.192808867f, 0.401670694f, 0.f),
    C_Vertex3f(0.192925334f, 0.471741557f, 0.f),
    C_Vertex3f(0.204426646f, 0.405669332f, 0.f),
    C_Vertex3f(0.207330346f, 0.473653436f, 0.f),
    C_Vertex3f(0.216535211f, 0.40833509f, 0.f),
    C_Vertex3f(0.222233772f, 0.474609375f, 0.f),
    C_Vertex3f(0.22913456f, 0.409667969f, 0.f),
    C_Vertex3f(0.24020195f, 0.474609375f, 0.f),
    C_Vertex3f(0.242333412f, 0.409667969f, 0.f),
    C_Vertex3f(0.259835482f, 0.47253418f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142081738f, 0.361684322f, 0.f),
    C_Vertex3f(0.151245594f, 0.372347355f, 0.f),
    C_Vertex3f(0.152700782f, 0.460270286f, 0.f),
    C_Vertex3f(0.160900235f, 0.381677508f, 0.f),
    C_Vertex3f(0.165610552f, 0.465049982f, 0.f),
    C_Vertex3f(0.171045661f, 0.389674783f, 0.f),
    C_Vertex3f(0.179018736f, 0.468873739f, 0.f),
    C_Vertex3f(0.181681871f, 0.396339178f, 0.f),
    C_Vertex3f(0.192925334f, 0.471741557f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12146306f, 0.321942568f, 0.f),
    C_Vertex3f(0.127190471f, 0.336481929f, 0.f),
    C_Vertex3f(0.128376484f, 0.447843075f, 0.f),
    C_Vertex3f(0.134063363f, 0.34972918f, 0.f),
    C_Vertex3f(0.140289426f, 0.45453465f, 0.f),
    C_Vertex3f(0.142081738f, 0.361684322f, 0.f),
    C_Vertex3f(0.152700782f, 0.460270286f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116961956f, 0.44019556f, 0.f),
    C_Vertex3f(0.106223345f, 0.431748986f, 0.f),
    C_Vertex3f(0.11000824f, 0.250864029f, 0.f),
    C_Vertex3f(0.111153722f, 0.270571828f, 0.f),
    C_Vertex3f(0.113444686f, 0.288987517f, 0.f),
    C_Vertex3f(0.116881132f, 0.306111097f, 0.f),
    C_Vertex3f(0.12146306f, 0.321942568f, 0.f),
    C_Vertex3f(0.128376484f, 0.447843075f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.43813324f, 0.464355469f, 0.f),
    C_Vertex3f(0.365379333f, 0.464355469f, 0.f),
    C_Vertex3f(0.373592257f, -0.143607497f, 0.f),
    C_Vertex3f(0.38314414f, -0.135271788f, 0.f),
    C_Vertex3f(0.391889811f, -0.126342058f, 0.f),
    C_Vertex3f(0.399829268f, -0.116818309f, 0.f),
    C_Vertex3f(0.406962514f, -0.10670054f, 0.f),
    C_Vertex3f(0.413289547f, -0.0959887505f, 0.f),
    C_Vertex3f(0.418810368f, -0.0846829414f, 0.f),
    C_Vertex3f(0.423641086f, -0.0714040995f, 0.f),
    C_Vertex3f(0.427781701f, -0.0561522245f, 0.f),
    C_Vertex3f(0.431232214f, -0.0389273167f, 0.f),
    C_Vertex3f(0.433992624f, -0.0197293758f, 0.f),
    C_Vertex3f(0.436062932f, 0.00144159794f, 0.f),
    C_Vertex3f(0.437443137f, 0.0245856047f, 0.f),
    C_Vertex3f(0.43813324f, 0.0497026443f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.363234162f, -0.151349187f, 0.f),
    C_Vertex3f(0.358110428f, 0.0509529114f, 0.f),
    C_Vertex3f(0.358015776f, 0.033413887f, 0.f),
    C_Vertex3f(0.357715845f, 0.0175237656f, 0.f),
    C_Vertex3f(0.357210636f, 0.003282547f, 0.f),
    C_Vertex3f(0.356500149f, -0.00930976868f, 0.f),
    C_Vertex3f(0.355584383f, -0.0202531815f, 0.f),
    C_Vertex3f(0.354463339f, -0.0295476913f, 0.f),
    C_Vertex3f(0.353137016f, -0.0371932983f, 0.f),
    C_Vertex3f(0.352371216f, -0.0401916504f, 0.f),
    C_Vertex3f(0.352069855f, -0.158496857f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.352371216f, -0.0401916504f, 0.f),
    C_Vertex3f(0.349750519f, -0.0500030518f, 0.f),
    C_Vertex3f(0.352069855f, -0.158496857f, 0.f),
    C_Vertex3f(0.342105389f, -0.067445755f, 0.f),
    C_Vertex3f(0.340099335f, -0.165050507f, 0.f),
    C_Vertex3f(0.332056522f, -0.0827083588f, 0.f),
    C_Vertex3f(0.327438831f, -0.17078495f, 0.f),
    C_Vertex3f(0.31960392f, -0.095790863f, 0.f),
    C_Vertex3f(0.314088345f, -0.175700188f, 0.f),
    C_Vertex3f(0.308461666f, -0.103967667f, 0.f),
    C_Vertex3f(0.300255418f, -0.108737469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.314088345f, -0.175700188f, 0.f),
    C_Vertex3f(0.300255418f, -0.108737469f, 0.f),
    C_Vertex3f(0.300047874f, -0.179796219f, 0.f),
    C_Vertex3f(0.291271091f, -0.112825871f, 0.f),
    C_Vertex3f(0.285317421f, -0.183073044f, 0.f),
    C_Vertex3f(0.281508684f, -0.116232872f, 0.f),
    C_Vertex3f(0.270968199f, -0.118958473f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.285317421f, -0.183073044f, 0.f),
    C_Vertex3f(0.270968199f, -0.118958473f, 0.f),
    C_Vertex3f(0.269896984f, -0.185530663f, 0.f),
    C_Vertex3f(0.259649634f, -0.121002674f, 0.f),
    C_Vertex3f(0.253786564f, -0.187169075f, 0.f),
    C_Vertex3f(0.247552991f, -0.122365475f, 0.f),
    C_Vertex3f(0.23698616f, -0.187988281f, 0.f),
    C_Vertex3f(0.234678268f, -0.123046875f, 0.f),
    C_Vertex3f(0.217855453f, -0.187988281f, 0.f),
    C_Vertex3f(0.215179443f, -0.123046875f, 0.f),
    C_Vertex3f(0.19808197f, -0.186820626f, 0.f),
    C_Vertex3f(0.192511559f, -0.120304108f, 0.f),
    C_Vertex3f(0.17930603f, -0.184485316f, 0.f),
    C_Vertex3f(0.172520638f, -0.114818573f, 0.f),
    C_Vertex3f(0.161527634f, -0.180982351f, 0.f),
    C_Vertex3f(0.15520668f, -0.106590271f, 0.f),
    C_Vertex3f(0.14474678f, -0.176311731f, 0.f),
    C_Vertex3f(0.142427444f, -0.0970439911f, 0.f),
    C_Vertex3f(0.133635521f, -0.0865097046f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.12896347f, -0.170473456f, 0.f),
    C_Vertex3f(0.14474678f, -0.176311731f, 0.f),
    C_Vertex3f(0.133635521f, -0.0865097046f, 0.f),
    C_Vertex3f(0.126973152f, -0.0735626221f, 0.f),
    C_Vertex3f(0.122440338f, -0.0582027435f, 0.f),
    C_Vertex3f(0.121238708f, -0.0493164063f, 0.f),
    C_Vertex3f(0.114177704f, -0.163467526f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.121238708f, -0.0493164063f, 0.f),
    C_Vertex3f(0.0445785522f, -0.0380859375f, 0.f),
    C_Vertex3f(0.0455768108f, -0.0650941133f, 0.f),
    C_Vertex3f(0.0484468937f, -0.0815242529f, 0.f),
    C_Vertex3f(0.0529704094f, -0.0967729092f, 0.f),
    C_Vertex3f(0.0591473579f, -0.110840082f, 0.f),
    C_Vertex3f(0.0669777393f, -0.123725772f, 0.f),
    C_Vertex3f(0.0764615536f, -0.135429978f, 0.f),
    C_Vertex3f(0.0875988007f, -0.145952702f, 0.f),
    C_Vertex3f(0.100389481f, -0.155293941f, 0.f),
    C_Vertex3f(0.114177704f, -0.163467526f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0455768108f, -0.0650941133f, 0.f),
    C_Vertex3f(0.0445785522f, -0.0380859375f, 0.f),
    C_Vertex3f(0.0443601608f, -0.0474824905f, 0.f),
    C_Vertex3f(0.11000824f, 0.250864029f, 0.f),
    C_Vertex3f(0.106223345f, 0.431748986f, 0.f),
    C_Vertex3f(0.11000824f, 0.228935242f, 0.f),
    C_Vertex3f(0.0305925608f, 0.19932282f, 0.f),
    C_Vertex3f(0.0289535522f, 0.25128746f, 0.f),
    C_Vertex3f(0.0289535522f, 0.222607613f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.364402771f, 0.248403549f, -1.f),
    C_Vertex3f(0.364402771f, 0.226584435f, -1.f),
    C_Vertex3f(0.373592257f, -0.143607497f, -1.f),
    C_Vertex3f(0.363257289f, 0.20575738f, -1.f),
    C_Vertex3f(0.363234162f, -0.151349187f, -1.f),
    C_Vertex3f(0.360966325f, 0.186413527f, -1.f),
    C_Vertex3f(0.358110428f, 0.0509529114f, -1.f),
    C_Vertex3f(0.358055115f, 0.060546875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.360966325f, 0.186413527f, -1.f),
    C_Vertex3f(0.357529879f, 0.168552876f, -1.f),
    C_Vertex3f(0.358055115f, 0.060546875f, -1.f),
    C_Vertex3f(0.35294795f, 0.152175426f, -1.f),
    C_Vertex3f(0.351608276f, 0.0529785156f, -1.f),
    C_Vertex3f(0.34722054f, 0.137281179f, -1.f),
    C_Vertex3f(0.340347648f, 0.123870134f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.351608276f, 0.0529785156f, -1.f),
    C_Vertex3f(0.340347648f, 0.123870134f, -1.f),
    C_Vertex3f(0.337922573f, 0.0397338867f, -1.f),
    C_Vertex3f(0.332329273f, 0.111942291f, -1.f),
    C_Vertex3f(0.323444843f, 0.0283813477f, -1.f),
    C_Vertex3f(0.323165417f, 0.10149765f, -1.f),
    C_Vertex3f(0.313456178f, 0.0923585892f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.308175087f, 0.0189208984f, -1.f),
    C_Vertex3f(0.323444843f, 0.0283813477f, -1.f),
    C_Vertex3f(0.313456178f, 0.0923585892f, -1.f),
    C_Vertex3f(0.303201556f, 0.0845251083f, -1.f),
    C_Vertex3f(0.292401552f, 0.0779972076f, -1.f),
    C_Vertex3f(0.292113304f, 0.0113525391f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.256729245f, 0.0662469864f, -1.f),
    C_Vertex3f(0.257613659f, 0.00189208984f, -1.f),
    C_Vertex3f(0.269165397f, 0.0688581467f, -1.f),
    C_Vertex3f(0.275259495f, 0.00567626953f, -1.f),
    C_Vertex3f(0.281056166f, 0.0727748871f, -1.f),
    C_Vertex3f(0.292113304f, 0.0113525391f, -1.f),
    C_Vertex3f(0.292401552f, 0.0779972076f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.217103481f, 0.0662537813f, -1.f),
    C_Vertex3f(0.217597008f, 0.f, -1.f),
    C_Vertex3f(0.230166435f, 0.0649414063f, -1.f),
    C_Vertex3f(0.239175797f, 0.f, -1.f),
    C_Vertex3f(0.243747711f, 0.0649414063f, -1.f),
    C_Vertex3f(0.257613659f, 0.00189208984f, -1.f),
    C_Vertex3f(0.256729245f, 0.0662469864f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.204613209f, 0.0688785315f, -1.f),
    C_Vertex3f(0.192695618f, 0.0728156567f, -1.f),
    C_Vertex3f(0.195021272f, 0.00215780735f, -1.f),
    C_Vertex3f(0.217597008f, 0.f, -1.f),
    C_Vertex3f(0.217103481f, 0.0662537813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.181350708f, 0.0780651569f, -1.f),
    C_Vertex3f(0.17057848f, 0.0846270323f, -1.f),
    C_Vertex3f(0.173797488f, 0.00647342205f, -1.f),
    C_Vertex3f(0.195021272f, 0.00215780735f, -1.f),
    C_Vertex3f(0.192695618f, 0.0728156567f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.160378933f, 0.0925012827f, -1.f),
    C_Vertex3f(0.150752068f, 0.101687908f, -1.f),
    C_Vertex3f(0.153925657f, 0.0129468441f, -1.f),
    C_Vertex3f(0.173797488f, 0.00647342205f, -1.f),
    C_Vertex3f(0.17057848f, 0.0846270323f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.141697884f, 0.112186909f, -1.f),
    C_Vertex3f(0.133775473f, 0.124230742f, -1.f),
    C_Vertex3f(0.135405779f, 0.0215780735f, -1.f),
    C_Vertex3f(0.153925657f, 0.0129468441f, -1.f),
    C_Vertex3f(0.150752068f, 0.101687908f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133775473f, 0.124230742f, -1.f),
    C_Vertex3f(0.126984835f, 0.137819409f, -1.f),
    C_Vertex3f(0.135405779f, 0.0215780735f, -1.f),
    C_Vertex3f(0.12132597f, 0.152952909f, -1.f),
    C_Vertex3f(0.118237853f, 0.0323671103f, -1.f),
    C_Vertex3f(0.116798878f, 0.169631243f, -1.f),
    C_Vertex3f(0.113403559f, 0.187854409f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.113403559f, 0.187854409f, -1.f),
    C_Vertex3f(0.111140013f, 0.207622409f, -1.f),
    C_Vertex3f(0.118237853f, 0.0323671103f, -1.f),
    C_Vertex3f(0.11000824f, 0.228935242f, -1.f),
    C_Vertex3f(0.10242188f, 0.0453139544f, -1.f),
    C_Vertex3f(0.106223345f, 0.431748986f, -1.f),
    C_Vertex3f(0.0961606503f, 0.422503352f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.087957859f, 0.0604186058f, -1.f),
    C_Vertex3f(0.10242188f, 0.0453139544f, -1.f),
    C_Vertex3f(0.0961606503f, 0.422503352f, -1.f),
    C_Vertex3f(0.0867738724f, 0.412458658f, -1.f),
    C_Vertex3f(0.0780630112f, 0.401614904f, -1.f),
    C_Vertex3f(0.0748457909f, 0.0776810646f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0700280666f, 0.389972091f, -1.f),
    C_Vertex3f(0.0626690388f, 0.377530217f, -1.f),
    C_Vertex3f(0.0633727312f, 0.0958038568f, -1.f),
    C_Vertex3f(0.0748457909f, 0.0776810646f, -1.f),
    C_Vertex3f(0.0780630112f, 0.401614904f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0387876034f, 0.155334234f, -1.f),
    C_Vertex3f(0.0453436375f, 0.134630442f, -1.f),
    C_Vertex3f(0.0469751358f, 0.343229294f, -1.f),
    C_Vertex3f(0.0535386801f, 0.114786983f, -1.f),
    C_Vertex3f(0.0559859276f, 0.364289284f, -1.f),
    C_Vertex3f(0.0633727312f, 0.0958038568f, -1.f),
    C_Vertex3f(0.0626690388f, 0.377530217f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0289535522f, 0.25128746f, -1.f),
    C_Vertex3f(0.0305925608f, 0.19932282f, -1.f),
    C_Vertex3f(0.0319571495f, 0.283218384f, -1.f),
    C_Vertex3f(0.0338705778f, 0.17689836f, -1.f),
    C_Vertex3f(0.037964344f, 0.313865662f, -1.f),
    C_Vertex3f(0.0387876034f, 0.155334234f, -1.f),
    C_Vertex3f(0.0469751358f, 0.343229294f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.365379333f, 0.408203125f, -1.f),
    C_Vertex3f(0.358716011f, 0.416503906f, -1.f),
    C_Vertex3f(0.360864043f, 0.287121654f, -1.f),
    C_Vertex3f(0.363223195f, 0.268418908f, -1.f),
    C_Vertex3f(0.364402771f, 0.248403549f, -1.f),
    C_Vertex3f(0.373592257f, -0.143607497f, -1.f),
    C_Vertex3f(0.365379333f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.358716011f, 0.416503906f, -1.f),
    C_Vertex3f(0.344488382f, 0.431030273f, -1.f),
    C_Vertex3f(0.346709132f, 0.335354209f, -1.f),
    C_Vertex3f(0.352607012f, 0.320589304f, -1.f),
    C_Vertex3f(0.357325315f, 0.304511786f, -1.f),
    C_Vertex3f(0.360864043f, 0.287121654f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.301613212f, 0.389367223f, -1.f),
    C_Vertex3f(0.31201756f, 0.381246924f, -1.f),
    C_Vertex3f(0.313330173f, 0.453857422f, -1.f),
    C_Vertex3f(0.321938038f, 0.371773243f, -1.f),
    C_Vertex3f(0.32935977f, 0.443481445f, -1.f),
    C_Vertex3f(0.331374645f, 0.360946178f, -1.f),
    C_Vertex3f(0.344488382f, 0.431030273f, -1.f),
    C_Vertex3f(0.339631677f, 0.3488065f, -1.f),
    C_Vertex3f(0.346709132f, 0.335354209f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.296399593f, 0.462158203f, -1.f),
    C_Vertex3f(0.278568029f, 0.468383789f, -1.f),
    C_Vertex3f(0.279352903f, 0.40154767f, -1.f),
    C_Vertex3f(0.290724993f, 0.396134138f, -1.f),
    C_Vertex3f(0.301613212f, 0.389367223f, -1.f),
    C_Vertex3f(0.313330173f, 0.453857422f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.242333412f, 0.409667969f, -1.f),
    C_Vertex3f(0.255157113f, 0.408314586f, -1.f),
    C_Vertex3f(0.259835482f, 0.47253418f, -1.f),
    C_Vertex3f(0.267496943f, 0.40560782f, -1.f),
    C_Vertex3f(0.278568029f, 0.468383789f, -1.f),
    C_Vertex3f(0.279352903f, 0.40154767f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.181681871f, 0.396339178f, -1.f),
    C_Vertex3f(0.192808867f, 0.401670694f, -1.f),
    C_Vertex3f(0.192925334f, 0.471741557f, -1.f),
    C_Vertex3f(0.204426646f, 0.405669332f, -1.f),
    C_Vertex3f(0.207330346f, 0.473653436f, -1.f),
    C_Vertex3f(0.216535211f, 0.40833509f, -1.f),
    C_Vertex3f(0.222233772f, 0.474609375f, -1.f),
    C_Vertex3f(0.22913456f, 0.409667969f, -1.f),
    C_Vertex3f(0.24020195f, 0.474609375f, -1.f),
    C_Vertex3f(0.242333412f, 0.409667969f, -1.f),
    C_Vertex3f(0.259835482f, 0.47253418f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142081738f, 0.361684322f, -1.f),
    C_Vertex3f(0.151245594f, 0.372347355f, -1.f),
    C_Vertex3f(0.152700782f, 0.460270286f, -1.f),
    C_Vertex3f(0.160900235f, 0.381677508f, -1.f),
    C_Vertex3f(0.165610552f, 0.465049982f, -1.f),
    C_Vertex3f(0.171045661f, 0.389674783f, -1.f),
    C_Vertex3f(0.179018736f, 0.468873739f, -1.f),
    C_Vertex3f(0.181681871f, 0.396339178f, -1.f),
    C_Vertex3f(0.192925334f, 0.471741557f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12146306f, 0.321942568f, -1.f),
    C_Vertex3f(0.127190471f, 0.336481929f, -1.f),
    C_Vertex3f(0.128376484f, 0.447843075f, -1.f),
    C_Vertex3f(0.134063363f, 0.34972918f, -1.f),
    C_Vertex3f(0.140289426f, 0.45453465f, -1.f),
    C_Vertex3f(0.142081738f, 0.361684322f, -1.f),
    C_Vertex3f(0.152700782f, 0.460270286f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116961956f, 0.44019556f, -1.f),
    C_Vertex3f(0.106223345f, 0.431748986f, -1.f),
    C_Vertex3f(0.11000824f, 0.250864029f, -1.f),
    C_Vertex3f(0.111153722f, 0.270571828f, -1.f),
    C_Vertex3f(0.113444686f, 0.288987517f, -1.f),
    C_Vertex3f(0.116881132f, 0.306111097f, -1.f),
    C_Vertex3f(0.12146306f, 0.321942568f, -1.f),
    C_Vertex3f(0.128376484f, 0.447843075f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.43813324f, 0.464355469f, -1.f),
    C_Vertex3f(0.365379333f, 0.464355469f, -1.f),
    C_Vertex3f(0.373592257f, -0.143607497f, -1.f),
    C_Vertex3f(0.38314414f, -0.135271788f, -1.f),
    C_Vertex3f(0.391889811f, -0.126342058f, -1.f),
    C_Vertex3f(0.399829268f, -0.116818309f, -1.f),
    C_Vertex3f(0.406962514f, -0.10670054f, -1.f),
    C_Vertex3f(0.413289547f, -0.0959887505f, -1.f),
    C_Vertex3f(0.418810368f, -0.0846829414f, -1.f),
    C_Vertex3f(0.423641086f, -0.0714040995f, -1.f),
    C_Vertex3f(0.427781701f, -0.0561522245f, -1.f),
    C_Vertex3f(0.431232214f, -0.0389273167f, -1.f),
    C_Vertex3f(0.433992624f, -0.0197293758f, -1.f),
    C_Vertex3f(0.436062932f, 0.00144159794f, -1.f),
    C_Vertex3f(0.437443137f, 0.0245856047f, -1.f),
    C_Vertex3f(0.43813324f, 0.0497026443f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.363234162f, -0.151349187f, -1.f),
    C_Vertex3f(0.358110428f, 0.0509529114f, -1.f),
    C_Vertex3f(0.358015776f, 0.033413887f, -1.f),
    C_Vertex3f(0.357715845f, 0.0175237656f, -1.f),
    C_Vertex3f(0.357210636f, 0.003282547f, -1.f),
    C_Vertex3f(0.356500149f, -0.00930976868f, -1.f),
    C_Vertex3f(0.355584383f, -0.0202531815f, -1.f),
    C_Vertex3f(0.354463339f, -0.0295476913f, -1.f),
    C_Vertex3f(0.353137016f, -0.0371932983f, -1.f),
    C_Vertex3f(0.352371216f, -0.0401916504f, -1.f),
    C_Vertex3f(0.352069855f, -0.158496857f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.352371216f, -0.0401916504f, -1.f),
    C_Vertex3f(0.349750519f, -0.0500030518f, -1.f),
    C_Vertex3f(0.352069855f, -0.158496857f, -1.f),
    C_Vertex3f(0.342105389f, -0.067445755f, -1.f),
    C_Vertex3f(0.340099335f, -0.165050507f, -1.f),
    C_Vertex3f(0.332056522f, -0.0827083588f, -1.f),
    C_Vertex3f(0.327438831f, -0.17078495f, -1.f),
    C_Vertex3f(0.31960392f, -0.095790863f, -1.f),
    C_Vertex3f(0.314088345f, -0.175700188f, -1.f),
    C_Vertex3f(0.308461666f, -0.103967667f, -1.f),
    C_Vertex3f(0.300255418f, -0.108737469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.314088345f, -0.175700188f, -1.f),
    C_Vertex3f(0.300255418f, -0.108737469f, -1.f),
    C_Vertex3f(0.300047874f, -0.179796219f, -1.f),
    C_Vertex3f(0.291271091f, -0.112825871f, -1.f),
    C_Vertex3f(0.285317421f, -0.183073044f, -1.f),
    C_Vertex3f(0.281508684f, -0.116232872f, -1.f),
    C_Vertex3f(0.270968199f, -0.118958473f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.285317421f, -0.183073044f, -1.f),
    C_Vertex3f(0.270968199f, -0.118958473f, -1.f),
    C_Vertex3f(0.269896984f, -0.185530663f, -1.f),
    C_Vertex3f(0.259649634f, -0.121002674f, -1.f),
    C_Vertex3f(0.253786564f, -0.187169075f, -1.f),
    C_Vertex3f(0.247552991f, -0.122365475f, -1.f),
    C_Vertex3f(0.23698616f, -0.187988281f, -1.f),
    C_Vertex3f(0.234678268f, -0.123046875f, -1.f),
    C_Vertex3f(0.217855453f, -0.187988281f, -1.f),
    C_Vertex3f(0.215179443f, -0.123046875f, -1.f),
    C_Vertex3f(0.19808197f, -0.186820626f, -1.f),
    C_Vertex3f(0.192511559f, -0.120304108f, -1.f),
    C_Vertex3f(0.17930603f, -0.184485316f, -1.f),
    C_Vertex3f(0.172520638f, -0.114818573f, -1.f),
    C_Vertex3f(0.161527634f, -0.180982351f, -1.f),
    C_Vertex3f(0.15520668f, -0.106590271f, -1.f),
    C_Vertex3f(0.14474678f, -0.176311731f, -1.f),
    C_Vertex3f(0.142427444f, -0.0970439911f, -1.f),
    C_Vertex3f(0.133635521f, -0.0865097046f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.12896347f, -0.170473456f, -1.f),
    C_Vertex3f(0.14474678f, -0.176311731f, -1.f),
    C_Vertex3f(0.133635521f, -0.0865097046f, -1.f),
    C_Vertex3f(0.126973152f, -0.0735626221f, -1.f),
    C_Vertex3f(0.122440338f, -0.0582027435f, -1.f),
    C_Vertex3f(0.121238708f, -0.0493164063f, -1.f),
    C_Vertex3f(0.114177704f, -0.163467526f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.121238708f, -0.0493164063f, -1.f),
    C_Vertex3f(0.0445785522f, -0.0380859375f, -1.f),
    C_Vertex3f(0.0455768108f, -0.0650941133f, -1.f),
    C_Vertex3f(0.0484468937f, -0.0815242529f, -1.f),
    C_Vertex3f(0.0529704094f, -0.0967729092f, -1.f),
    C_Vertex3f(0.0591473579f, -0.110840082f, -1.f),
    C_Vertex3f(0.0669777393f, -0.123725772f, -1.f),
    C_Vertex3f(0.0764615536f, -0.135429978f, -1.f),
    C_Vertex3f(0.0875988007f, -0.145952702f, -1.f),
    C_Vertex3f(0.100389481f, -0.155293941f, -1.f),
    C_Vertex3f(0.114177704f, -0.163467526f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0455768108f, -0.0650941133f, -1.f),
    C_Vertex3f(0.0445785522f, -0.0380859375f, -1.f),
    C_Vertex3f(0.0443601608f, -0.0474824905f, -1.f),
    C_Vertex3f(0.11000824f, 0.250864029f, -1.f),
    C_Vertex3f(0.106223345f, 0.431748986f, -1.f),
    C_Vertex3f(0.11000824f, 0.228935242f, -1.f),
    C_Vertex3f(0.0305925608f, 0.19932282f, -1.f),
    C_Vertex3f(0.0289535522f, 0.25128746f, -1.f),
    C_Vertex3f(0.0289535522f, 0.222607613f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_104[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.640625f, -1.f),
    C_Vertex3f(0.0590820313f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.640625f, 0.f),
    C_Vertex3f(0.0590820313f, 0.640625f, -1.f),
    C_Vertex3f(0.137695313f, 0.640625f, -1.f),
    C_Vertex3f(0.137695313f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.640625f, 0.f),
    C_Vertex3f(0.137695313f, 0.640625f, -1.f),
    C_Vertex3f(0.137695313f, 0.410682678f, -1.f),
    C_Vertex3f(0.137695313f, 0.410682678f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.757735252f, 0.652562082f, 0.f),
    C_Vertex3f(0.137695313f, 0.410682678f, 0.f),
    C_Vertex3f(0.137695313f, 0.410682678f, -1.f),
    C_Normal3f(-0.714029312f, 0.7001158f, 0.f),
    C_Vertex3f(0.144577026f, 0.418673515f, 0.f),
    C_Vertex3f(0.144577026f, 0.418673515f, -1.f),
    C_Normal3f(-0.65162605f, 0.758540392f, 0.f),
    C_Vertex3f(0.159241915f, 0.43265748f, 0.f),
    C_Vertex3f(0.159241915f, 0.43265748f, -1.f),
    C_Normal3f(-0.565677524f, 0.824626505f, 0.f),
    C_Vertex3f(0.174808264f, 0.444643736f, 0.f),
    C_Vertex3f(0.174808264f, 0.444643736f, -1.f),
    C_Normal3f(-0.469225168f, 0.883078635f, 0.f),
    C_Vertex3f(0.191276073f, 0.454632282f, 0.f),
    C_Vertex3f(0.191276073f, 0.454632282f, -1.f),
    C_Normal3f(-0.365257144f, 0.930906594f, 0.f),
    C_Vertex3f(0.208645344f, 0.462623119f, 0.f),
    C_Vertex3f(0.208645344f, 0.462623119f, -1.f),
    C_Normal3f(-0.257753402f, 0.966210723f, 0.f),
    C_Vertex3f(0.226916075f, 0.468616247f, 0.f),
    C_Vertex3f(0.226916075f, 0.468616247f, -1.f),
    C_Normal3f(-0.150957346f, 0.988540351f, 0.f),
    C_Vertex3f(0.246088266f, 0.472611666f, 0.f),
    C_Vertex3f(0.246088266f, 0.472611666f, -1.f),
    C_Normal3f(-0.0459275506f, 0.998944759f, 0.f),
    C_Vertex3f(0.266161919f, 0.474609375f, 0.f),
    C_Vertex3f(0.266161919f, 0.474609375f, -1.f),
    C_Normal3f(0.0534397066f, 0.998571098f, 0.f),
    C_Vertex3f(0.289539337f, 0.474609375f, 0.f),
    C_Vertex3f(0.289539337f, 0.474609375f, -1.f),
    C_Normal3f(0.161784887f, 0.986826062f, 0.f),
    C_Vertex3f(0.313626289f, 0.472069263f, 0.f),
    C_Vertex3f(0.313626289f, 0.472069263f, -1.f),
    C_Normal3f(0.282686323f, 0.959212482f, 0.f),
    C_Vertex3f(0.33602047f, 0.46698904f, 0.f),
    C_Vertex3f(0.33602047f, 0.46698904f, -1.f),
    C_Normal3f(0.408667862f, 0.912683129f, 0.f),
    C_Vertex3f(0.356721878f, 0.459368706f, 0.f),
    C_Vertex3f(0.356721878f, 0.459368706f, -1.f),
    C_Normal3f(0.533259153f, 0.845951915f, 0.f),
    C_Vertex3f(0.375730515f, 0.44920826f, 0.f),
    C_Vertex3f(0.375730515f, 0.44920826f, -1.f),
    C_Normal3f(0.65810585f, 0.752925336f, 0.f),
    C_Vertex3f(0.392035007f, 0.437108517f, 0.f),
    C_Vertex3f(0.392035007f, 0.437108517f, -1.f),
    C_Normal3f(0.77475512f, 0.632261515f, 0.f),
    C_Vertex3f(0.405635357f, 0.423069477f, 0.f),
    C_Vertex3f(0.405635357f, 0.423069477f, -1.f),
    C_Normal3f(0.865303636f, 0.501248002f, 0.f),
    C_Vertex3f(0.416531563f, 0.407091141f, 0.f),
    C_Vertex3f(0.416531563f, 0.407091141f, -1.f),
    C_Normal3f(0.923428595f, 0.383770406f, 0.f),
    C_Vertex3f(0.42267561f, 0.393652916f, 0.f),
    C_Vertex3f(0.42267561f, 0.393652916f, -1.f),
    C_Normal3f(0.953053594f, 0.302801609f, 0.f),
    C_Vertex3f(0.426259637f, 0.383683443f, 0.f),
    C_Vertex3f(0.426259637f, 0.383683443f, -1.f),
    C_Normal3f(0.971233666f, 0.238128498f, 0.f),
    C_Vertex3f(0.42933166f, 0.372703314f, 0.f),
    C_Vertex3f(0.42933166f, 0.372703314f, -1.f),
    C_Normal3f(0.98342371f, 0.181322411f, 0.f),
    C_Vertex3f(0.43189168f, 0.360712528f, 0.f),
    C_Vertex3f(0.43189168f, 0.360712528f, -1.f),
    C_Normal3f(0.991313159f, 0.131522685f, 0.f),
    C_Vertex3f(0.433939695f, 0.347711086f, 0.f),
    C_Vertex3f(0.433939695f, 0.347711086f, -1.f),
    C_Normal3f(0.996135533f, 0.0878298357f, 0.f),
    C_Vertex3f(0.435475707f, 0.333698988f, 0.f),
    C_Vertex3f(0.435475707f, 0.333698988f, -1.f),
    C_Normal3f(0.998779178f, 0.0493987761f, 0.f),
    C_Vertex3f(0.436499715f, 0.318676233f, 0.f),
    C_Vertex3f(0.436499715f, 0.318676233f, -1.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.437011719f, 0.302642822f, 0.f),
    C_Vertex3f(0.437011719f, 0.302642822f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.437011719f, 0.302642822f, 0.f),
    C_Vertex3f(0.437011719f, 0.302642822f, -1.f),
    C_Vertex3f(0.437011719f, 0.f, -1.f),
    C_Vertex3f(0.437011719f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.437011719f, 0.f, 0.f),
    C_Vertex3f(0.437011719f, 0.f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_Normal3f(-0.999996781f, -0.00253521232f, 0.f),
    C_Vertex3f(0.358398438f, 0.301394463f, 0.f),
    C_Vertex3f(0.358398438f, 0.301394463f, -1.f),
    C_Normal3f(-0.995928764f, -0.0901440307f, 0.f),
    C_Vertex3f(0.357599497f, 0.315136552f, 0.f),
    C_Vertex3f(0.357599497f, 0.315136552f, -1.f),
    C_Normal3f(-0.986938894f, -0.161095008f, 0.f),
    C_Vertex3f(0.356001616f, 0.327875018f, 0.f),
    C_Vertex3f(0.356001616f, 0.327875018f, -1.f),
    C_Normal3f(-0.970385253f, -0.24156265f, 0.f),
    C_Vertex3f(0.353604794f, 0.339609861f, 0.f),
    C_Vertex3f(0.353604794f, 0.339609861f, -1.f),
    C_Normal3f(-0.94342798f, -0.331577659f, 0.f),
    C_Vertex3f(0.350409031f, 0.350341082f, 0.f),
    C_Vertex3f(0.350409031f, 0.350341082f, -1.f),
    C_Normal3f(-0.902824759f, -0.430008769f, 0.f),
    C_Vertex3f(0.346414328f, 0.360068679f, 0.f),
    C_Vertex3f(0.346414328f, 0.360068679f, -1.f),
    C_Normal3f(-0.845481038f, -0.534005463f, 0.f),
    C_Vertex3f(0.341620684f, 0.368792653f, 0.f),
    C_Vertex3f(0.341620684f, 0.368792653f, -1.f),
    C_Normal3f(-0.769446313f, -0.638711512f, 0.f),
    C_Vertex3f(0.336028099f, 0.376513004f, 0.f),
    C_Vertex3f(0.336028099f, 0.376513004f, -1.f),
    C_Normal3f(-0.683618188f, -0.729839861f, 0.f),
    C_Vertex3f(0.329636574f, 0.383229733f, 0.f),
    C_Vertex3f(0.329636574f, 0.383229733f, -1.f),
    C_Normal3f(-0.594310284f, -0.804235995f, 0.f),
    C_Vertex3f(0.322582722f, 0.38910687f, 0.f),
    C_Vertex3f(0.322582722f, 0.38910687f, -1.f),
    C_Normal3f(-0.497703731f, -0.867347002f, 0.f),
    C_Vertex3f(0.314866543f, 0.394144416f, 0.f),
    C_Vertex3f(0.314866543f, 0.394144416f, -1.f),
    C_Normal3f(-0.397959441f, -0.917403042f, 0.f),
    C_Vertex3f(0.306488037f, 0.398342371f, 0.f),
    C_Vertex3f(0.306488037f, 0.398342371f, -1.f),
    C_Normal3f(-0.299185693f, -0.954194903f, 0.f),
    C_Vertex3f(0.297447205f, 0.401700735f, 0.f),
    C_Vertex3f(0.297447205f, 0.401700735f, -1.f),
    C_Normal3f(-0.204748258f, -0.978814662f, 0.f),
    C_Vertex3f(0.287744045f, 0.404219508f, 0.f),
    C_Vertex3f(0.287744045f, 0.404219508f, -1.f),
    C_Normal3f(-0.116928905f, -0.99314034f, 0.f),
    C_Vertex3f(0.277378559f, 0.40589869f, 0.f),
    C_Vertex3f(0.277378559f, 0.40589869f, -1.f),
    C_Normal3f(-0.0327606685f, -0.999463201f, 0.f),
    C_Vertex3f(0.266350746f, 0.406738281f, 0.f),
    C_Vertex3f(0.266350746f, 0.406738281f, -1.f),
    C_Normal3f(0.0716509297f, -0.997429729f, 0.f),
    C_Vertex3f(0.251764297f, 0.406738281f, 0.f),
    C_Vertex3f(0.251764297f, 0.406738281f, -1.f),
    C_Normal3f(0.199434131f, -0.979911268f, 0.f),
    C_Vertex3f(0.234800816f, 0.404471874f, 0.f),
    C_Vertex3f(0.234800816f, 0.404471874f, -1.f),
    C_Normal3f(0.33042413f, -0.943832576f, 0.f),
    C_Vertex3f(0.218356609f, 0.39993906f, 0.f),
    C_Vertex3f(0.218356609f, 0.39993906f, -1.f),
    C_Normal3f(0.451753765f, -0.892142713f, 0.f),
    C_Vertex3f(0.202431679f, 0.393139839f, 0.f),
    C_Vertex3f(0.202431679f, 0.393139839f, -1.f),
    C_Normal3f(0.56809783f, -0.822961092f, 0.f),
    C_Vertex3f(0.187026024f, 0.384074211f, 0.f),
    C_Vertex3f(0.187026024f, 0.384074211f, -1.f),
    C_Normal3f(0.688725412f, -0.725022256f, 0.f),
    C_Vertex3f(0.173832893f, 0.373397827f, 0.f),
    C_Vertex3f(0.173832893f, 0.373397827f, -1.f),
    C_Normal3f(0.798388541f, -0.602142632f, 0.f),
    C_Vertex3f(0.162852287f, 0.361110687f, 0.f),
    C_Vertex3f(0.162852287f, 0.361110687f, -1.f),
    C_Normal3f(0.886818588f, -0.462117761f, 0.f),
    C_Vertex3f(0.154084206f, 0.347212791f, 0.f),
    C_Vertex3f(0.154084206f, 0.347212791f, -1.f),
    C_Normal3f(0.947420299f, -0.319991916f, 0.f),
    C_Vertex3f(0.147528648f, 0.33170414f, 0.f),
    C_Vertex3f(0.147528648f, 0.33170414f, -1.f),
    C_Normal3f(0.97952038f, -0.201345146f, 0.f),
    C_Vertex3f(0.14261198f, 0.31324625f, 0.f),
    C_Vertex3f(0.14261198f, 0.31324625f, -1.f),
    C_Normal3f(0.994278252f, -0.106821738f, 0.f),
    C_Vertex3f(0.139334202f, 0.291839123f, 0.f),
    C_Vertex3f(0.139334202f, 0.291839123f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.267482758f, 0.f),
    C_Vertex3f(0.137695313f, 0.267482758f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.267482758f, 0.f),
    C_Vertex3f(0.137695313f, 0.267482758f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0590820313f, 0.640625f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.640625f, 0.f),
    C_Vertex3f(0.137695313f, 0.267482758f, 0.f),
    C_Vertex3f(0.137695313f, 0.410682678f, 0.f),
    C_Vertex3f(0.139334202f, 0.291839123f, 0.f),
    C_Vertex3f(0.144577026f, 0.418673515f, 0.f),
    C_Vertex3f(0.14261198f, 0.31324625f, 0.f),
    C_Vertex3f(0.147528648f, 0.33170414f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.353604794f, 0.339609861f, 0.f),
    C_Vertex3f(0.356001616f, 0.327875018f, 0.f),
    C_Vertex3f(0.356721878f, 0.459368706f, 0.f),
    C_Vertex3f(0.357599497f, 0.315136552f, 0.f),
    C_Vertex3f(0.375730515f, 0.44920826f, 0.f),
    C_Vertex3f(0.358398438f, 0.301394463f, 0.f),
    C_Vertex3f(0.437011719f, 0.f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356721878f, 0.459368706f, 0.f),
    C_Vertex3f(0.33602047f, 0.46698904f, 0.f),
    C_Vertex3f(0.336028099f, 0.376513004f, 0.f),
    C_Vertex3f(0.341620684f, 0.368792653f, 0.f),
    C_Vertex3f(0.346414328f, 0.360068679f, 0.f),
    C_Vertex3f(0.350409031f, 0.350341082f, 0.f),
    C_Vertex3f(0.353604794f, 0.339609861f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33602047f, 0.46698904f, 0.f),
    C_Vertex3f(0.313626289f, 0.472069263f, 0.f),
    C_Vertex3f(0.314866543f, 0.394144416f, 0.f),
    C_Vertex3f(0.322582722f, 0.38910687f, 0.f),
    C_Vertex3f(0.329636574f, 0.383229733f, 0.f),
    C_Vertex3f(0.336028099f, 0.376513004f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.277378559f, 0.40589869f, 0.f),
    C_Vertex3f(0.287744045f, 0.404219508f, 0.f),
    C_Vertex3f(0.289539337f, 0.474609375f, 0.f),
    C_Vertex3f(0.297447205f, 0.401700735f, 0.f),
    C_Vertex3f(0.313626289f, 0.472069263f, 0.f),
    C_Vertex3f(0.306488037f, 0.398342371f, 0.f),
    C_Vertex3f(0.314866543f, 0.394144416f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.162852287f, 0.361110687f, 0.f),
    C_Vertex3f(0.173832893f, 0.373397827f, 0.f),
    C_Vertex3f(0.174808264f, 0.444643736f, 0.f),
    C_Vertex3f(0.187026024f, 0.384074211f, 0.f),
    C_Vertex3f(0.191276073f, 0.454632282f, 0.f),
    C_Vertex3f(0.202431679f, 0.393139839f, 0.f),
    C_Vertex3f(0.208645344f, 0.462623119f, 0.f),
    C_Vertex3f(0.218356609f, 0.39993906f, 0.f),
    C_Vertex3f(0.226916075f, 0.468616247f, 0.f),
    C_Vertex3f(0.234800816f, 0.404471874f, 0.f),
    C_Vertex3f(0.246088266f, 0.472611666f, 0.f),
    C_Vertex3f(0.251764297f, 0.406738281f, 0.f),
    C_Vertex3f(0.266161919f, 0.474609375f, 0.f),
    C_Vertex3f(0.266350746f, 0.406738281f, 0.f),
    C_Vertex3f(0.289539337f, 0.474609375f, 0.f),
    C_Vertex3f(0.277378559f, 0.40589869f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.159241915f, 0.43265748f, 0.f),
    C_Vertex3f(0.144577026f, 0.418673515f, 0.f),
    C_Vertex3f(0.147528648f, 0.33170414f, 0.f),
    C_Vertex3f(0.154084206f, 0.347212791f, 0.f),
    C_Vertex3f(0.162852287f, 0.361110687f, 0.f),
    C_Vertex3f(0.174808264f, 0.444643736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.437011719f, 0.f, 0.f),
    C_Vertex3f(0.437011719f, 0.302642822f, 0.f),
    C_Vertex3f(0.436499715f, 0.318676233f, 0.f),
    C_Vertex3f(0.435475707f, 0.333698988f, 0.f),
    C_Vertex3f(0.433939695f, 0.347711086f, 0.f),
    C_Vertex3f(0.43189168f, 0.360712528f, 0.f),
    C_Vertex3f(0.42933166f, 0.372703314f, 0.f),
    C_Vertex3f(0.426259637f, 0.383683443f, 0.f),
    C_Vertex3f(0.42267561f, 0.393652916f, 0.f),
    C_Vertex3f(0.416531563f, 0.407091141f, 0.f),
    C_Vertex3f(0.405635357f, 0.423069477f, 0.f),
    C_Vertex3f(0.392035007f, 0.437108517f, 0.f),
    C_Vertex3f(0.375730515f, 0.44920826f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.640625f, 0.f),
    C_Vertex3f(0.0590820313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0590820313f, 0.640625f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.137695313f, 0.640625f, -1.f),
    C_Vertex3f(0.137695313f, 0.267482758f, -1.f),
    C_Vertex3f(0.137695313f, 0.410682678f, -1.f),
    C_Vertex3f(0.139334202f, 0.291839123f, -1.f),
    C_Vertex3f(0.144577026f, 0.418673515f, -1.f),
    C_Vertex3f(0.14261198f, 0.31324625f, -1.f),
    C_Vertex3f(0.147528648f, 0.33170414f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.353604794f, 0.339609861f, -1.f),
    C_Vertex3f(0.356001616f, 0.327875018f, -1.f),
    C_Vertex3f(0.356721878f, 0.459368706f, -1.f),
    C_Vertex3f(0.357599497f, 0.315136552f, -1.f),
    C_Vertex3f(0.375730515f, 0.44920826f, -1.f),
    C_Vertex3f(0.358398438f, 0.301394463f, -1.f),
    C_Vertex3f(0.437011719f, 0.f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356721878f, 0.459368706f, -1.f),
    C_Vertex3f(0.33602047f, 0.46698904f, -1.f),
    C_Vertex3f(0.336028099f, 0.376513004f, -1.f),
    C_Vertex3f(0.341620684f, 0.368792653f, -1.f),
    C_Vertex3f(0.346414328f, 0.360068679f, -1.f),
    C_Vertex3f(0.350409031f, 0.350341082f, -1.f),
    C_Vertex3f(0.353604794f, 0.339609861f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33602047f, 0.46698904f, -1.f),
    C_Vertex3f(0.313626289f, 0.472069263f, -1.f),
    C_Vertex3f(0.314866543f, 0.394144416f, -1.f),
    C_Vertex3f(0.322582722f, 0.38910687f, -1.f),
    C_Vertex3f(0.329636574f, 0.383229733f, -1.f),
    C_Vertex3f(0.336028099f, 0.376513004f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.277378559f, 0.40589869f, -1.f),
    C_Vertex3f(0.287744045f, 0.404219508f, -1.f),
    C_Vertex3f(0.289539337f, 0.474609375f, -1.f),
    C_Vertex3f(0.297447205f, 0.401700735f, -1.f),
    C_Vertex3f(0.313626289f, 0.472069263f, -1.f),
    C_Vertex3f(0.306488037f, 0.398342371f, -1.f),
    C_Vertex3f(0.314866543f, 0.394144416f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.162852287f, 0.361110687f, -1.f),
    C_Vertex3f(0.173832893f, 0.373397827f, -1.f),
    C_Vertex3f(0.174808264f, 0.444643736f, -1.f),
    C_Vertex3f(0.187026024f, 0.384074211f, -1.f),
    C_Vertex3f(0.191276073f, 0.454632282f, -1.f),
    C_Vertex3f(0.202431679f, 0.393139839f, -1.f),
    C_Vertex3f(0.208645344f, 0.462623119f, -1.f),
    C_Vertex3f(0.218356609f, 0.39993906f, -1.f),
    C_Vertex3f(0.226916075f, 0.468616247f, -1.f),
    C_Vertex3f(0.234800816f, 0.404471874f, -1.f),
    C_Vertex3f(0.246088266f, 0.472611666f, -1.f),
    C_Vertex3f(0.251764297f, 0.406738281f, -1.f),
    C_Vertex3f(0.266161919f, 0.474609375f, -1.f),
    C_Vertex3f(0.266350746f, 0.406738281f, -1.f),
    C_Vertex3f(0.289539337f, 0.474609375f, -1.f),
    C_Vertex3f(0.277378559f, 0.40589869f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.159241915f, 0.43265748f, -1.f),
    C_Vertex3f(0.144577026f, 0.418673515f, -1.f),
    C_Vertex3f(0.147528648f, 0.33170414f, -1.f),
    C_Vertex3f(0.154084206f, 0.347212791f, -1.f),
    C_Vertex3f(0.162852287f, 0.361110687f, -1.f),
    C_Vertex3f(0.174808264f, 0.444643736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.437011719f, 0.f, -1.f),
    C_Vertex3f(0.437011719f, 0.302642822f, -1.f),
    C_Vertex3f(0.436499715f, 0.318676233f, -1.f),
    C_Vertex3f(0.435475707f, 0.333698988f, -1.f),
    C_Vertex3f(0.433939695f, 0.347711086f, -1.f),
    C_Vertex3f(0.43189168f, 0.360712528f, -1.f),
    C_Vertex3f(0.42933166f, 0.372703314f, -1.f),
    C_Vertex3f(0.426259637f, 0.383683443f, -1.f),
    C_Vertex3f(0.42267561f, 0.393652916f, -1.f),
    C_Vertex3f(0.416531563f, 0.407091141f, -1.f),
    C_Vertex3f(0.405635357f, 0.423069477f, -1.f),
    C_Vertex3f(0.392035007f, 0.437108517f, -1.f),
    C_Vertex3f(0.375730515f, 0.44920826f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.640625f, -1.f),
    C_Vertex3f(0.0590820313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_105[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0595703125f, 0.550292969f, 0.f),
    C_Vertex3f(0.0595703125f, 0.550292969f, -1.f),
    C_Vertex3f(0.0595703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0595703125f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0595703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0595703125f, 0.640625f, -1.f),
    C_Vertex3f(0.138183594f, 0.640625f, -1.f),
    C_Vertex3f(0.138183594f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.138183594f, 0.640625f, 0.f),
    C_Vertex3f(0.138183594f, 0.640625f, -1.f),
    C_Vertex3f(0.138183594f, 0.550292969f, -1.f),
    C_Vertex3f(0.138183594f, 0.550292969f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.138183594f, 0.550292969f, 0.f),
    C_Vertex3f(0.138183594f, 0.550292969f, -1.f),
    C_Vertex3f(0.0595703125f, 0.550292969f, -1.f),
    C_Vertex3f(0.0595703125f, 0.550292969f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0595703125f, 0.f, 0.f),
    C_Vertex3f(0.0595703125f, 0.f, -1.f),
    C_Vertex3f(0.0595703125f, 0.464355469f, -1.f),
    C_Vertex3f(0.0595703125f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0595703125f, 0.464355469f, 0.f),
    C_Vertex3f(0.0595703125f, 0.464355469f, -1.f),
    C_Vertex3f(0.138183594f, 0.464355469f, -1.f),
    C_Vertex3f(0.138183594f, 0.464355469f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.138183594f, 0.464355469f, 0.f),
    C_Vertex3f(0.138183594f, 0.464355469f, -1.f),
    C_Vertex3f(0.138183594f, 0.f, -1.f),
    C_Vertex3f(0.138183594f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.138183594f, 0.f, 0.f),
    C_Vertex3f(0.138183594f, 0.f, -1.f),
    C_Vertex3f(0.0595703125f, 0.f, -1.f),
    C_Vertex3f(0.0595703125f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0595703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0595703125f, 0.550292969f, 0.f),
    C_Vertex3f(0.138183594f, 0.550292969f, 0.f),
    C_Vertex3f(0.138183594f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0595703125f, 0.464355469f, 0.f),
    C_Vertex3f(0.0595703125f, 0.f, 0.f),
    C_Vertex3f(0.138183594f, 0.f, 0.f),
    C_Vertex3f(0.138183594f, 0.464355469f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0595703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0595703125f, 0.550292969f, -1.f),
    C_Vertex3f(0.138183594f, 0.550292969f, -1.f),
    C_Vertex3f(0.138183594f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0595703125f, 0.464355469f, -1.f),
    C_Vertex3f(0.0595703125f, 0.f, -1.f),
    C_Vertex3f(0.138183594f, 0.f, -1.f),
    C_Vertex3f(0.138183594f, 0.464355469f, -1.f),
    C_End(),
    C_Translatef(0.198730469f, 0.f, 0.f),
};
static const Call char_106[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0585632324f, 0.549316406f, 0.f),
    C_Vertex3f(0.0585632324f, 0.549316406f, -1.f),
    C_Vertex3f(0.0585632324f, 0.640625f, -1.f),
    C_Vertex3f(0.0585632324f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0585632324f, 0.640625f, 0.f),
    C_Vertex3f(0.0585632324f, 0.640625f, -1.f),
    C_Vertex3f(0.137176514f, 0.640625f, -1.f),
    C_Vertex3f(0.137176514f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137176514f, 0.640625f, 0.f),
    C_Vertex3f(0.137176514f, 0.640625f, -1.f),
    C_Vertex3f(0.137176514f, 0.549316406f, -1.f),
    C_Vertex3f(0.137176514f, 0.549316406f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.137176514f, 0.549316406f, 0.f),
    C_Vertex3f(0.137176514f, 0.549316406f, -1.f),
    C_Vertex3f(0.0585632324f, 0.549316406f, -1.f),
    C_Vertex3f(0.0585632324f, 0.549316406f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.977198541f, 0.212327629f, 0.f),
    C_Vertex3f(-0.0410766602f, -0.179710388f, 0.f),
    C_Vertex3f(-0.0410766602f, -0.179710388f, -1.f),
    C_Vertex3f(-0.0265731812f, -0.112960815f, -1.f),
    C_Vertex3f(-0.0265731812f, -0.112960815f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.2523458f, 0.967637122f, 0.f),
    C_Vertex3f(-0.0265731812f, -0.112960815f, 0.f),
    C_Vertex3f(-0.0265731812f, -0.112960815f, -1.f),
    C_Normal3f(0.227778971f, 0.973712921f, 0.f),
    C_Vertex3f(-0.0206489563f, -0.114505768f, 0.f),
    C_Vertex3f(-0.0206489563f, -0.114505768f, -1.f),
    C_Normal3f(0.190425396f, 0.981701732f, 0.f),
    C_Vertex3f(-0.0100622177f, -0.116823196f, 0.f),
    C_Vertex3f(-0.0100622177f, -0.116823196f, -1.f),
    C_Normal3f(0.132106736f, 0.991235495f, 0.f),
    C_Vertex3f(-0.000737190247f, -0.118368149f, 0.f),
    C_Vertex3f(-0.000737190247f, -0.118368149f, -1.f),
    C_Normal3f(0.0441011451f, 0.999027073f, 0.f),
    C_Vertex3f(0.0073261261f, -0.119140625f, 0.f),
    C_Vertex3f(0.0073261261f, -0.119140625f, -1.f),
    C_Normal3f(-0.0993672684f, 0.995050848f, 0.f),
    C_Vertex3f(0.0167617798f, -0.119140625f, 0.f),
    C_Vertex3f(0.0167617798f, -0.119140625f, -1.f),
    C_Normal3f(-0.292881548f, 0.956148744f, 0.f),
    C_Vertex3f(0.0272951126f, -0.117146492f, 0.f),
    C_Vertex3f(0.0272951126f, -0.117146492f, -1.f),
    C_Normal3f(-0.518164396f, 0.855280995f, 0.f),
    C_Vertex3f(0.0362920761f, -0.113158226f, 0.f),
    C_Vertex3f(0.0362920761f, -0.113158226f, -1.f),
    C_Normal3f(-0.708912671f, 0.705296338f, 0.f),
    C_Vertex3f(0.0437526703f, -0.107175827f, 0.f),
    C_Vertex3f(0.0437526703f, -0.107175827f, -1.f),
    C_Normal3f(-0.852359235f, 0.522956848f, 0.f),
    C_Vertex3f(0.0481958389f, -0.101193428f, 0.f),
    C_Vertex3f(0.0481958389f, -0.101193428f, -1.f),
    C_Normal3f(-0.932847023f, 0.360272586f, 0.f),
    C_Vertex3f(0.0507876873f, -0.0957095623f, 0.f),
    C_Vertex3f(0.0507876873f, -0.0957095623f, -1.f),
    C_Normal3f(-0.966983318f, 0.254839897f, 0.f),
    C_Vertex3f(0.0530092716f, -0.0887300968f, 0.f),
    C_Vertex3f(0.0530092716f, -0.0887300968f, -1.f),
    C_Normal3f(-0.984070122f, 0.177780554f, 0.f),
    C_Vertex3f(0.0548605919f, -0.0802550316f, 0.f),
    C_Vertex3f(0.0548605919f, -0.0802550316f, -1.f),
    C_Normal3f(-0.992770076f, 0.120031625f, 0.f),
    C_Vertex3f(0.0563416481f, -0.0702843666f, 0.f),
    C_Vertex3f(0.0563416481f, -0.0702843666f, -1.f),
    C_Normal3f(-0.997140467f, 0.0755696967f, 0.f),
    C_Vertex3f(0.0574524403f, -0.0588181019f, 0.f),
    C_Vertex3f(0.0574524403f, -0.0588181019f, -1.f),
    C_Normal3f(-0.999180377f, 0.0404780768f, 0.f),
    C_Vertex3f(0.0581929684f, -0.0458562374f, 0.f),
    C_Vertex3f(0.0581929684f, -0.0458562374f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0585632324f, -0.0313987732f, 0.f),
    C_Vertex3f(0.0585632324f, -0.0313987732f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0585632324f, -0.0313987732f, 0.f),
    C_Vertex3f(0.0585632324f, -0.0313987732f, -1.f),
    C_Vertex3f(0.0585632324f, 0.464355469f, -1.f),
    C_Vertex3f(0.0585632324f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0585632324f, 0.464355469f, 0.f),
    C_Vertex3f(0.0585632324f, 0.464355469f, -1.f),
    C_Vertex3f(0.137176514f, 0.464355469f, -1.f),
    C_Vertex3f(0.137176514f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137176514f, 0.464355469f, 0.f),
    C_Vertex3f(0.137176514f, 0.464355469f, -1.f),
    C_Normal3f(0.999999106f, -0.00133918482f, 0.f),
    C_Vertex3f(0.137176514f, -0.0357141495f, 0.f),
    C_Vertex3f(0.137176514f, -0.0357141495f, -1.f),
    C_Normal3f(0.998488784f, -0.0549575053f, 0.f),
    C_Vertex3f(0.136480331f, -0.0554993153f, 0.f),
    C_Vertex3f(0.136480331f, -0.0554993153f, -1.f),
    C_Normal3f(0.995005012f, -0.0998252705f, 0.f),
    C_Vertex3f(0.135087967f, -0.0736596584f, 0.f),
    C_Vertex3f(0.135087967f, -0.0736596584f, -1.f),
    C_Normal3f(0.988203883f, -0.153143689f, 0.f),
    C_Vertex3f(0.13299942f, -0.090195179f, 0.f),
    C_Vertex3f(0.13299942f, -0.090195179f, -1.f),
    C_Normal3f(0.976189017f, -0.216921717f, 0.f),
    C_Vertex3f(0.130214691f, -0.105105877f, 0.f),
    C_Vertex3f(0.130214691f, -0.105105877f, -1.f),
    C_Normal3f(0.955972373f, -0.293456614f, 0.f),
    C_Vertex3f(0.12673378f, -0.118391752f, 0.f),
    C_Vertex3f(0.12673378f, -0.118391752f, -1.f),
    C_Normal3f(0.922928333f, -0.384971827f, 0.f),
    C_Vertex3f(0.122556686f, -0.130052805f, 0.f),
    C_Vertex3f(0.122556686f, -0.130052805f, -1.f),
    C_Normal3f(0.869204581f, -0.494452566f, 0.f),
    C_Vertex3f(0.117683411f, -0.140089035f, 0.f),
    C_Vertex3f(0.117683411f, -0.140089035f, -1.f),
    C_Normal3f(0.796181798f, -0.605057478f, 0.f),
    C_Vertex3f(0.111348152f, -0.149756432f, 0.f),
    C_Vertex3f(0.111348152f, -0.149756432f, -1.f),
    C_Normal3f(0.707634091f, -0.70657903f, 0.f),
    C_Vertex3f(0.10307312f, -0.159314394f, 0.f),
    C_Vertex3f(0.10307312f, -0.159314394f, -1.f),
    C_Normal3f(0.59912318f, -0.800656855f, 0.f),
    C_Vertex3f(0.093624115f, -0.167506933f, 0.f),
    C_Vertex3f(0.093624115f, -0.167506933f, -1.f),
    C_Normal3f(0.480652452f, -0.876911163f, 0.f),
    C_Vertex3f(0.0830011368f, -0.174334049f, 0.f),
    C_Vertex3f(0.0830011368f, -0.174334049f, -1.f),
    C_Normal3f(0.360024184f, -0.932942986f, 0.f),
    C_Vertex3f(0.0712041855f, -0.179795742f, 0.f),
    C_Vertex3f(0.0712041855f, -0.179795742f, -1.f),
    C_Normal3f(0.244156301f, -0.969735801f, 0.f),
    C_Vertex3f(0.0582332611f, -0.183892012f, 0.f),
    C_Vertex3f(0.0582332611f, -0.183892012f, -1.f),
    C_Normal3f(0.137702912f, -0.990473509f, 0.f),
    C_Vertex3f(0.0440883636f, -0.186622858f, 0.f),
    C_Vertex3f(0.0440883636f, -0.186622858f, -1.f),
    C_Normal3f(0.0432540104f, -0.999064207f, 0.f),
    C_Vertex3f(0.0287694931f, -0.187988281f, 0.f),
    C_Vertex3f(0.0287694931f, -0.187988281f, -1.f),
    C_Normal3f(-0.032428138f, -0.999474049f, 0.f),
    C_Vertex3f(0.012550354f, -0.187988281f, 0.f),
    C_Vertex3f(0.012550354f, -0.187988281f, -1.f),
    C_Normal3f(-0.0994069502f, -0.995046794f, 0.f),
    C_Vertex3f(-0.00312232971f, -0.186953545f, 0.f),
    C_Vertex3f(-0.00312232971f, -0.186953545f, -1.f),
    C_Normal3f(-0.167095542f, -0.985940635f, 0.f),
    C_Vertex3f(-0.0185222626f, -0.184884071f, 0.f),
    C_Vertex3f(-0.0185222626f, -0.184884071f, -1.f),
    C_Normal3f(-0.223580092f, -0.97468555f, 0.f),
    C_Vertex3f(-0.0336494446f, -0.181779861f, 0.f),
    C_Vertex3f(-0.0336494446f, -0.181779861f, -1.f),
    C_Normal3f(-0.268409312f, -0.963304937f, 0.f),
    C_Vertex3f(-0.0410766602f, -0.179710388f, 0.f),
    C_Vertex3f(-0.0410766602f, -0.179710388f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0585632324f, 0.640625f, 0.f),
    C_Vertex3f(0.0585632324f, 0.549316406f, 0.f),
    C_Vertex3f(0.137176514f, 0.549316406f, 0.f),
    C_Vertex3f(0.137176514f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.137176514f, 0.464355469f, 0.f),
    C_Vertex3f(0.0585632324f, 0.464355469f, 0.f),
    C_Vertex3f(0.0712041855f, -0.179795742f, 0.f),
    C_Vertex3f(0.0830011368f, -0.174334049f, 0.f),
    C_Vertex3f(0.093624115f, -0.167506933f, 0.f),
    C_Vertex3f(0.10307312f, -0.159314394f, 0.f),
    C_Vertex3f(0.111348152f, -0.149756432f, 0.f),
    C_Vertex3f(0.117683411f, -0.140089035f, 0.f),
    C_Vertex3f(0.122556686f, -0.130052805f, 0.f),
    C_Vertex3f(0.12673378f, -0.118391752f, 0.f),
    C_Vertex3f(0.130214691f, -0.105105877f, 0.f),
    C_Vertex3f(0.13299942f, -0.090195179f, 0.f),
    C_Vertex3f(0.135087967f, -0.0736596584f, 0.f),
    C_Vertex3f(0.136480331f, -0.0554993153f, 0.f),
    C_Vertex3f(0.137176514f, -0.0357141495f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0585632324f, -0.0313987732f, 0.f),
    C_Vertex3f(0.0581929684f, -0.0458562374f, 0.f),
    C_Vertex3f(0.0582332611f, -0.183892012f, 0.f),
    C_Vertex3f(0.0712041855f, -0.179795742f, 0.f),
    C_Vertex3f(0.0585632324f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0582332611f, -0.183892012f, 0.f),
    C_Vertex3f(0.0581929684f, -0.0458562374f, 0.f),
    C_Vertex3f(0.0574524403f, -0.0588181019f, 0.f),
    C_Vertex3f(0.0563416481f, -0.0702843666f, 0.f),
    C_Vertex3f(0.0548605919f, -0.0802550316f, 0.f),
    C_Vertex3f(0.0530092716f, -0.0887300968f, 0.f),
    C_Vertex3f(0.0507876873f, -0.0957095623f, 0.f),
    C_Vertex3f(0.0481958389f, -0.101193428f, 0.f),
    C_Vertex3f(0.0440883636f, -0.186622858f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0481958389f, -0.101193428f, 0.f),
    C_Vertex3f(0.0437526703f, -0.107175827f, 0.f),
    C_Vertex3f(0.0440883636f, -0.186622858f, 0.f),
    C_Vertex3f(0.0362920761f, -0.113158226f, 0.f),
    C_Vertex3f(0.0287694931f, -0.187988281f, 0.f),
    C_Vertex3f(0.0272951126f, -0.117146492f, 0.f),
    C_Vertex3f(0.0167617798f, -0.119140625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.012550354f, -0.187988281f, 0.f),
    C_Vertex3f(0.0287694931f, -0.187988281f, 0.f),
    C_Vertex3f(0.0167617798f, -0.119140625f, 0.f),
    C_Vertex3f(0.0073261261f, -0.119140625f, 0.f),
    C_Vertex3f(-0.000737190247f, -0.118368149f, 0.f),
    C_Vertex3f(-0.00312232971f, -0.186953545f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-0.0100622177f, -0.116823196f, 0.f),
    C_Vertex3f(-0.0206489563f, -0.114505768f, 0.f),
    C_Vertex3f(-0.0185222626f, -0.184884071f, 0.f),
    C_Vertex3f(-0.00312232971f, -0.186953545f, 0.f),
    C_Vertex3f(-0.000737190247f, -0.118368149f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-0.0265731812f, -0.112960815f, 0.f),
    C_Vertex3f(-0.0410766602f, -0.179710388f, 0.f),
    C_Vertex3f(-0.0336494446f, -0.181779861f, 0.f),
    C_Vertex3f(-0.0185222626f, -0.184884071f, 0.f),
    C_Vertex3f(-0.0206489563f, -0.114505768f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0585632324f, 0.640625f, -1.f),
    C_Vertex3f(0.0585632324f, 0.549316406f, -1.f),
    C_Vertex3f(0.137176514f, 0.549316406f, -1.f),
    C_Vertex3f(0.137176514f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.137176514f, 0.464355469f, -1.f),
    C_Vertex3f(0.0585632324f, 0.464355469f, -1.f),
    C_Vertex3f(0.0712041855f, -0.179795742f, -1.f),
    C_Vertex3f(0.0830011368f, -0.174334049f, -1.f),
    C_Vertex3f(0.093624115f, -0.167506933f, -1.f),
    C_Vertex3f(0.10307312f, -0.159314394f, -1.f),
    C_Vertex3f(0.111348152f, -0.149756432f, -1.f),
    C_Vertex3f(0.117683411f, -0.140089035f, -1.f),
    C_Vertex3f(0.122556686f, -0.130052805f, -1.f),
    C_Vertex3f(0.12673378f, -0.118391752f, -1.f),
    C_Vertex3f(0.130214691f, -0.105105877f, -1.f),
    C_Vertex3f(0.13299942f, -0.090195179f, -1.f),
    C_Vertex3f(0.135087967f, -0.0736596584f, -1.f),
    C_Vertex3f(0.136480331f, -0.0554993153f, -1.f),
    C_Vertex3f(0.137176514f, -0.0357141495f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0585632324f, -0.0313987732f, -1.f),
    C_Vertex3f(0.0581929684f, -0.0458562374f, -1.f),
    C_Vertex3f(0.0582332611f, -0.183892012f, -1.f),
    C_Vertex3f(0.0712041855f, -0.179795742f, -1.f),
    C_Vertex3f(0.0585632324f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0582332611f, -0.183892012f, -1.f),
    C_Vertex3f(0.0581929684f, -0.0458562374f, -1.f),
    C_Vertex3f(0.0574524403f, -0.0588181019f, -1.f),
    C_Vertex3f(0.0563416481f, -0.0702843666f, -1.f),
    C_Vertex3f(0.0548605919f, -0.0802550316f, -1.f),
    C_Vertex3f(0.0530092716f, -0.0887300968f, -1.f),
    C_Vertex3f(0.0507876873f, -0.0957095623f, -1.f),
    C_Vertex3f(0.0481958389f, -0.101193428f, -1.f),
    C_Vertex3f(0.0440883636f, -0.186622858f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0481958389f, -0.101193428f, -1.f),
    C_Vertex3f(0.0437526703f, -0.107175827f, -1.f),
    C_Vertex3f(0.0440883636f, -0.186622858f, -1.f),
    C_Vertex3f(0.0362920761f, -0.113158226f, -1.f),
    C_Vertex3f(0.0287694931f, -0.187988281f, -1.f),
    C_Vertex3f(0.0272951126f, -0.117146492f, -1.f),
    C_Vertex3f(0.0167617798f, -0.119140625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.012550354f, -0.187988281f, -1.f),
    C_Vertex3f(0.0287694931f, -0.187988281f, -1.f),
    C_Vertex3f(0.0167617798f, -0.119140625f, -1.f),
    C_Vertex3f(0.0073261261f, -0.119140625f, -1.f),
    C_Vertex3f(-0.000737190247f, -0.118368149f, -1.f),
    C_Vertex3f(-0.00312232971f, -0.186953545f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-0.0100622177f, -0.116823196f, -1.f),
    C_Vertex3f(-0.0206489563f, -0.114505768f, -1.f),
    C_Vertex3f(-0.0185222626f, -0.184884071f, -1.f),
    C_Vertex3f(-0.00312232971f, -0.186953545f, -1.f),
    C_Vertex3f(-0.000737190247f, -0.118368149f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-0.0265731812f, -0.112960815f, -1.f),
    C_Vertex3f(-0.0410766602f, -0.179710388f, -1.f),
    C_Vertex3f(-0.0336494446f, -0.181779861f, -1.f),
    C_Vertex3f(-0.0185222626f, -0.184884071f, -1.f),
    C_Vertex3f(-0.0206489563f, -0.114505768f, -1.f),
    C_End(),
    C_Translatef(0.198730469f, 0.f, 0.f),
};
static const Call char_107[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0594329834f, 0.f, 0.f),
    C_Vertex3f(0.0594329834f, 0.f, -1.f),
    C_Vertex3f(0.0594329834f, 0.640625f, -1.f),
    C_Vertex3f(0.0594329834f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0594329834f, 0.640625f, 0.f),
    C_Vertex3f(0.0594329834f, 0.640625f, -1.f),
    C_Vertex3f(0.138046265f, 0.640625f, -1.f),
    C_Vertex3f(0.138046265f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.138046265f, 0.640625f, 0.f),
    C_Vertex3f(0.138046265f, 0.640625f, -1.f),
    C_Vertex3f(0.138046265f, 0.27557373f, -1.f),
    C_Vertex3f(0.138046265f, 0.27557373f, 0.f),
    C_Normal3f(-0.711593151f, 0.702591836f, 0.f),
    C_Vertex3f(0.138046265f, 0.27557373f, 0.f),
    C_Vertex3f(0.138046265f, 0.27557373f, -1.f),
    C_Vertex3f(0.324440002f, 0.464355469f, -1.f),
    C_Vertex3f(0.324440002f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.324440002f, 0.464355469f, 0.f),
    C_Vertex3f(0.324440002f, 0.464355469f, -1.f),
    C_Vertex3f(0.426269531f, 0.464355469f, -1.f),
    C_Vertex3f(0.426269531f, 0.464355469f, 0.f),
    C_Normal3f(0.697220147f, -0.716857135f, 0.f),
    C_Vertex3f(0.426269531f, 0.464355469f, 0.f),
    C_Vertex3f(0.426269531f, 0.464355469f, -1.f),
    C_Vertex3f(0.248855591f, 0.291801453f, -1.f),
    C_Vertex3f(0.248855591f, 0.291801453f, 0.f),
    C_Normal3f(0.830805063f, 0.556563556f, 0.f),
    C_Vertex3f(0.248855591f, 0.291801453f, 0.f),
    C_Vertex3f(0.248855591f, 0.291801453f, -1.f),
    C_Vertex3f(0.444335938f, 0.f, -1.f),
    C_Vertex3f(0.444335938f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.444335938f, 0.f, 0.f),
    C_Vertex3f(0.444335938f, 0.f, -1.f),
    C_Vertex3f(0.34727478f, 0.f, -1.f),
    C_Vertex3f(0.34727478f, 0.f, 0.f),
    C_Normal3f(-0.838402927f, -0.5450508f, 0.f),
    C_Vertex3f(0.34727478f, 0.f, 0.f),
    C_Vertex3f(0.34727478f, 0.f, -1.f),
    C_Vertex3f(0.192932129f, 0.237411499f, -1.f),
    C_Vertex3f(0.192932129f, 0.237411499f, 0.f),
    C_Normal3f(0.697223485f, -0.716853738f, 0.f),
    C_Vertex3f(0.192932129f, 0.237411499f, 0.f),
    C_Vertex3f(0.192932129f, 0.237411499f, -1.f),
    C_Vertex3f(0.138046265f, 0.184028625f, -1.f),
    C_Vertex3f(0.138046265f, 0.184028625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.138046265f, 0.184028625f, 0.f),
    C_Vertex3f(0.138046265f, 0.184028625f, -1.f),
    C_Vertex3f(0.138046265f, 0.f, -1.f),
    C_Vertex3f(0.138046265f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.138046265f, 0.f, 0.f),
    C_Vertex3f(0.138046265f, 0.f, -1.f),
    C_Vertex3f(0.0594329834f, 0.f, -1.f),
    C_Vertex3f(0.0594329834f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0594329834f, 0.640625f, 0.f),
    C_Vertex3f(0.138046265f, 0.f, 0.f),
    C_Vertex3f(0.138046265f, 0.640625f, 0.f),
    C_Vertex3f(0.138046265f, 0.184028625f, 0.f),
    C_Vertex3f(0.138046265f, 0.27557373f, 0.f),
    C_Vertex3f(0.192932129f, 0.237411499f, 0.f),
    C_Vertex3f(0.324440002f, 0.464355469f, 0.f),
    C_Vertex3f(0.248855591f, 0.291801453f, 0.f),
    C_Vertex3f(0.426269531f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.248855591f, 0.291801453f, 0.f),
    C_Vertex3f(0.192932129f, 0.237411499f, 0.f),
    C_Vertex3f(0.34727478f, 0.f, 0.f),
    C_Vertex3f(0.444335938f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.138046265f, 0.f, 0.f),
    C_Vertex3f(0.0594329834f, 0.640625f, 0.f),
    C_Vertex3f(0.0594329834f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0594329834f, 0.640625f, -1.f),
    C_Vertex3f(0.138046265f, 0.f, -1.f),
    C_Vertex3f(0.138046265f, 0.640625f, -1.f),
    C_Vertex3f(0.138046265f, 0.184028625f, -1.f),
    C_Vertex3f(0.138046265f, 0.27557373f, -1.f),
    C_Vertex3f(0.192932129f, 0.237411499f, -1.f),
    C_Vertex3f(0.324440002f, 0.464355469f, -1.f),
    C_Vertex3f(0.248855591f, 0.291801453f, -1.f),
    C_Vertex3f(0.426269531f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.248855591f, 0.291801453f, -1.f),
    C_Vertex3f(0.192932129f, 0.237411499f, -1.f),
    C_Vertex3f(0.34727478f, 0.f, -1.f),
    C_Vertex3f(0.444335938f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.138046265f, 0.f, -1.f),
    C_Vertex3f(0.0594329834f, 0.640625f, -1.f),
    C_Vertex3f(0.0594329834f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_108[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.f, -1.f),
    C_Vertex3f(0.0571289063f, 0.640625f, -1.f),
    C_Vertex3f(0.0571289063f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.640625f, 0.f),
    C_Vertex3f(0.0571289063f, 0.640625f, -1.f),
    C_Vertex3f(0.135742188f, 0.640625f, -1.f),
    C_Vertex3f(0.135742188f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.135742188f, 0.640625f, 0.f),
    C_Vertex3f(0.135742188f, 0.640625f, -1.f),
    C_Vertex3f(0.135742188f, 0.f, -1.f),
    C_Vertex3f(0.135742188f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.135742188f, 0.f, 0.f),
    C_Vertex3f(0.135742188f, 0.f, -1.f),
    C_Vertex3f(0.0571289063f, 0.f, -1.f),
    C_Vertex3f(0.0571289063f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0571289063f, 0.f, 0.f),
    C_Vertex3f(0.135742188f, 0.f, 0.f),
    C_Vertex3f(0.135742188f, 0.640625f, 0.f),
    C_Vertex3f(0.0571289063f, 0.640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0571289063f, 0.f, -1.f),
    C_Vertex3f(0.135742188f, 0.f, -1.f),
    C_Vertex3f(0.135742188f, 0.640625f, -1.f),
    C_Vertex3f(0.0571289063f, 0.640625f, -1.f),
    C_End(),
    C_Translatef(0.198730469f, 0.f, 0.f),
};
static const Call char_109[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0589981079f, 0.f, 0.f),
    C_Vertex3f(0.0589981079f, 0.f, -1.f),
    C_Vertex3f(0.0589981079f, 0.464355469f, -1.f),
    C_Vertex3f(0.0589981079f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0589981079f, 0.464355469f, 0.f),
    C_Vertex3f(0.0589981079f, 0.464355469f, -1.f),
    C_Vertex3f(0.129310608f, 0.464355469f, -1.f),
    C_Vertex3f(0.129310608f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.129310608f, 0.464355469f, 0.f),
    C_Vertex3f(0.129310608f, 0.464355469f, -1.f),
    C_Vertex3f(0.129310608f, 0.398826599f, -1.f),
    C_Vertex3f(0.129310608f, 0.398826599f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.842350006f, 0.538930953f, 0.f),
    C_Vertex3f(0.129310608f, 0.398826599f, 0.f),
    C_Vertex3f(0.129310608f, 0.398826599f, -1.f),
    C_Normal3f(-0.796281338f, 0.604926527f, 0.f),
    C_Vertex3f(0.134775162f, 0.407367706f, 0.f),
    C_Vertex3f(0.134775162f, 0.407367706f, -1.f),
    C_Normal3f(-0.730337977f, 0.683086097f, 0.f),
    C_Vertex3f(0.147508144f, 0.422780514f, 0.f),
    C_Vertex3f(0.147508144f, 0.422780514f, -1.f),
    C_Normal3f(-0.641445577f, 0.767168522f, 0.f),
    C_Vertex3f(0.162045002f, 0.436523914f, 0.f),
    C_Vertex3f(0.162045002f, 0.436523914f, -1.f),
    C_Normal3f(-0.546065688f, 0.837742388f, 0.f),
    C_Vertex3f(0.178385735f, 0.448597908f, 0.f),
    C_Vertex3f(0.178385735f, 0.448597908f, -1.f),
    C_Normal3f(-0.436343312f, 0.899780273f, 0.f),
    C_Vertex3f(0.196530342f, 0.459002495f, 0.f),
    C_Vertex3f(0.196530342f, 0.459002495f, -1.f),
    C_Normal3f(-0.308780134f, 0.95113343f, 0.f),
    C_Vertex3f(0.215932369f, 0.466805935f, 0.f),
    C_Vertex3f(0.215932369f, 0.466805935f, -1.f),
    C_Normal3f(-0.180278316f, 0.983615637f, 0.f),
    C_Vertex3f(0.236591816f, 0.472008228f, 0.f),
    C_Vertex3f(0.236591816f, 0.472008228f, -1.f),
    C_Normal3f(-0.0559695102f, 0.998432457f, 0.f),
    C_Vertex3f(0.258508682f, 0.474609375f, 0.f),
    C_Vertex3f(0.258508682f, 0.474609375f, -1.f),
    C_Normal3f(0.0558521487f, 0.998439014f, 0.f),
    C_Vertex3f(0.282993317f, 0.474609375f, 0.f),
    C_Vertex3f(0.282993317f, 0.474609375f, -1.f),
    C_Normal3f(0.177497074f, 0.984121263f, 0.f),
    C_Vertex3f(0.306465626f, 0.471926689f, 0.f),
    C_Vertex3f(0.306465626f, 0.471926689f, -1.f),
    C_Normal3f(0.318103194f, 0.948056042f, 0.f),
    C_Vertex3f(0.327615261f, 0.466561317f, 0.f),
    C_Vertex3f(0.327615261f, 0.466561317f, -1.f),
    C_Normal3f(0.469332099f, 0.883021712f, 0.f),
    C_Vertex3f(0.346442223f, 0.45851326f, 0.f),
    C_Vertex3f(0.346442223f, 0.45851326f, -1.f),
    C_Normal3f(0.610943854f, 0.791673899f, 0.f),
    C_Vertex3f(0.36294651f, 0.447782516f, 0.f),
    C_Vertex3f(0.36294651f, 0.447782516f, -1.f),
    C_Normal3f(0.734925449f, 0.678147972f, 0.f),
    C_Vertex3f(0.377019405f, 0.434916496f, 0.f),
    C_Vertex3f(0.377019405f, 0.434916496f, -1.f),
    C_Normal3f(0.838896573f, 0.544290781f, 0.f),
    C_Vertex3f(0.388660908f, 0.419915199f, 0.f),
    C_Vertex3f(0.388660908f, 0.419915199f, -1.f),
    C_Normal3f(0.904810369f, 0.425814927f, 0.f),
    C_Vertex3f(0.397871017f, 0.402778625f, 0.f),
    C_Vertex3f(0.397871017f, 0.402778625f, -1.f),
    C_Normal3f(0.943344831f, 0.331813872f, 0.f),
    C_Vertex3f(0.401260376f, 0.3931427f, 0.f),
    C_Vertex3f(0.401260376f, 0.3931427f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.828080058f, 0.560609758f, 0.f),
    C_Vertex3f(0.401260376f, 0.3931427f, 0.f),
    C_Vertex3f(0.401260376f, 0.3931427f, -1.f),
    C_Normal3f(-0.790163994f, 0.612895429f, 0.f),
    C_Vertex3f(0.408154488f, 0.403326035f, 0.f),
    C_Vertex3f(0.408154488f, 0.403326035f, -1.f),
    C_Normal3f(-0.733208716f, 0.680003643f, 0.f),
    C_Vertex3f(0.422981977f, 0.42114687f, 0.f),
    C_Vertex3f(0.422981977f, 0.42114687f, -1.f),
    C_Normal3f(-0.64963007f, 0.760250509f, 0.f),
    C_Vertex3f(0.438848734f, 0.436421871f, 0.f),
    C_Vertex3f(0.438848734f, 0.436421871f, -1.f),
    C_Normal3f(-0.549348891f, 0.835593164f, 0.f),
    C_Vertex3f(0.455754757f, 0.449151039f, 0.f),
    C_Vertex3f(0.455754757f, 0.449151039f, -1.f),
    C_Normal3f(-0.434603363f, 0.900621951f, 0.f),
    C_Vertex3f(0.473700047f, 0.459334373f, 0.f),
    C_Vertex3f(0.473700047f, 0.459334373f, -1.f),
    C_Normal3f(-0.310219824f, 0.950664878f, 0.f),
    C_Vertex3f(0.492684603f, 0.466971874f, 0.f),
    C_Vertex3f(0.492684603f, 0.466971874f, -1.f),
    C_Normal3f(-0.182755828f, 0.98315835f, 0.f),
    C_Vertex3f(0.512708426f, 0.472063541f, 0.f),
    C_Vertex3f(0.512708426f, 0.472063541f, -1.f),
    C_Normal3f(-0.0623386651f, 0.998054981f, 0.f),
    C_Vertex3f(0.533771515f, 0.474609375f, 0.f),
    C_Vertex3f(0.533771515f, 0.474609375f, -1.f),
    C_Normal3f(0.0331861898f, 0.999449193f, 0.f),
    C_Vertex3f(0.553467751f, 0.474609375f, 0.f),
    C_Vertex3f(0.553467751f, 0.474609375f, -1.f),
    C_Normal3f(0.112776987f, 0.993620276f, 0.f),
    C_Vertex3f(0.56975925f, 0.473414421f, 0.f),
    C_Vertex3f(0.56975925f, 0.473414421f, -1.f),
    C_Normal3f(0.197942361f, 0.980213583f, 0.f),
    C_Vertex3f(0.585052133f, 0.471024513f, 0.f),
    C_Vertex3f(0.585052133f, 0.471024513f, -1.f),
    C_Normal3f(0.290137529f, 0.956984937f, 0.f),
    C_Vertex3f(0.599346399f, 0.467439651f, 0.f),
    C_Vertex3f(0.599346399f, 0.467439651f, -1.f),
    C_Normal3f(0.387405723f, 0.921909332f, 0.f),
    C_Vertex3f(0.61264205f, 0.462659836f, 0.f),
    C_Vertex3f(0.61264205f, 0.462659836f, -1.f),
    C_Normal3f(0.486658543f, 0.873592317f, 0.f),
    C_Vertex3f(0.624939084f, 0.456685066f, 0.f),
    C_Vertex3f(0.624939084f, 0.456685066f, -1.f),
    C_Normal3f(0.58390075f, 0.811825097f, 0.f),
    C_Vertex3f(0.636237502f, 0.449515343f, 0.f),
    C_Vertex3f(0.636237502f, 0.449515343f, -1.f),
    C_Normal3f(0.67483902f, 0.737964928f, 0.f),
    C_Vertex3f(0.646537304f, 0.441150665f, 0.f),
    C_Vertex3f(0.646537304f, 0.441150665f, -1.f),
    C_Normal3f(0.76026088f, 0.649617851f, 0.f),
    C_Vertex3f(0.65583849f, 0.431591034f, 0.f),
    C_Vertex3f(0.65583849f, 0.431591034f, -1.f),
    C_Normal3f(0.835611641f, 0.549320579f, 0.f),
    C_Vertex3f(0.663977027f, 0.420740604f, 0.f),
    C_Vertex3f(0.663977027f, 0.420740604f, -1.f),
    C_Normal3f(0.894392133f, 0.447283626f, 0.f),
    C_Vertex3f(0.670952916f, 0.408599377f, 0.f),
    C_Vertex3f(0.670952916f, 0.408599377f, -1.f),
    C_Normal3f(0.937356591f, 0.348371387f, 0.f),
    C_Vertex3f(0.676766157f, 0.395167351f, 0.f),
    C_Vertex3f(0.676766157f, 0.395167351f, -1.f),
    C_Normal3f(0.966686368f, 0.255963653f, 0.f),
    C_Vertex3f(0.68141675f, 0.380444527f, 0.f),
    C_Vertex3f(0.68141675f, 0.380444527f, -1.f),
    C_Normal3f(0.985117733f, 0.171880648f, 0.f),
    C_Vertex3f(0.684904695f, 0.364430904f, 0.f),
    C_Vertex3f(0.684904695f, 0.364430904f, -1.f),
    C_Normal3f(0.995313346f, 0.0967028588f, 0.f),
    C_Vertex3f(0.687229991f, 0.347126484f, 0.f),
    C_Vertex3f(0.687229991f, 0.347126484f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.688392639f, 0.328531265f, 0.f),
    C_Vertex3f(0.688392639f, 0.328531265f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.688392639f, 0.328531265f, 0.f),
    C_Vertex3f(0.688392639f, 0.328531265f, -1.f),
    C_Vertex3f(0.688392639f, 0.f, -1.f),
    C_Vertex3f(0.688392639f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.688392639f, 0.f, 0.f),
    C_Vertex3f(0.688392639f, 0.f, -1.f),
    C_Vertex3f(0.609779358f, 0.f, -1.f),
    C_Vertex3f(0.609779358f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.609779358f, 0.f, 0.f),
    C_Vertex3f(0.609779358f, 0.f, -1.f),
    C_Normal3f(-0.999995649f, -0.00294936704f, 0.f),
    C_Vertex3f(0.609779358f, 0.304031372f, 0.f),
    C_Vertex3f(0.609779358f, 0.304031372f, -1.f),
    C_Normal3f(-0.997048676f, -0.0767716095f, 0.f),
    C_Vertex3f(0.608822823f, 0.324317455f, 0.f),
    C_Vertex3f(0.608822823f, 0.324317455f, -1.f),
    C_Normal3f(-0.988051176f, -0.154126093f, 0.f),
    C_Vertex3f(0.606909752f, 0.341299534f, 0.f),
    C_Vertex3f(0.606909752f, 0.341299534f, -1.f),
    C_Normal3f(-0.963367045f, -0.268186599f, 0.f),
    C_Vertex3f(0.604040146f, 0.354977608f, 0.f),
    C_Vertex3f(0.604040146f, 0.354977608f, -1.f),
    C_Normal3f(-0.90621841f, -0.422809988f, 0.f),
    C_Vertex3f(0.600214005f, 0.365351677f, 0.f),
    C_Vertex3f(0.600214005f, 0.365351677f, -1.f),
    C_Normal3f(-0.820967793f, -0.570974469f, 0.f),
    C_Vertex3f(0.59483099f, 0.374715805f, 0.f),
    C_Vertex3f(0.59483099f, 0.374715805f, -1.f),
    C_Normal3f(-0.713026941f, -0.701136649f, 0.f),
    C_Vertex3f(0.587891102f, 0.383069992f, 0.f),
    C_Vertex3f(0.587891102f, 0.383069992f, -1.f),
    C_Normal3f(-0.59347564f, -0.804851949f, 0.f),
    C_Vertex3f(0.579394341f, 0.390414238f, 0.f),
    C_Vertex3f(0.579394341f, 0.390414238f, -1.f),
    C_Normal3f(-0.467123091f, -0.884192228f, 0.f),
    C_Vertex3f(0.569340706f, 0.396748543f, 0.f),
    C_Vertex3f(0.569340706f, 0.396748543f, -1.f),
    C_Normal3f(-0.32892856f, -0.944354832f, 0.f),
    C_Vertex3f(0.558412075f, 0.401499271f, 0.f),
    C_Vertex3f(0.558412075f, 0.401499271f, -1.f),
    C_Normal3f(-0.190494508f, -0.981688261f, 0.f),
    C_Vertex3f(0.546608448f, 0.404666424f, 0.f),
    C_Vertex3f(0.546608448f, 0.404666424f, -1.f),
    C_Normal3f(-0.0617978722f, -0.998088717f, 0.f),
    C_Vertex3f(0.533929825f, 0.40625f, 0.f),
    C_Vertex3f(0.533929825f, 0.40625f, -1.f),
    C_Normal3f(0.0412723497f, -0.999147952f, 0.f),
    C_Vertex3f(0.521032333f, 0.40625f, 0.f),
    C_Vertex3f(0.521032333f, 0.40625f, -1.f),
    C_Normal3f(0.131920278f, -0.99126029f, 0.f),
    C_Vertex3f(0.509310246f, 0.405233026f, 0.f),
    C_Vertex3f(0.509310246f, 0.405233026f, -1.f),
    C_Normal3f(0.226303384f, -0.9740569f, 0.f),
    C_Vertex3f(0.498107433f, 0.403199077f, 0.f),
    C_Vertex3f(0.498107433f, 0.403199077f, -1.f),
    C_Normal3f(0.323146343f, -0.946349025f, 0.f),
    C_Vertex3f(0.487423897f, 0.400148153f, 0.f),
    C_Vertex3f(0.487423897f, 0.400148153f, -1.f),
    C_Normal3f(0.419437468f, -0.907784164f, 0.f),
    C_Vertex3f(0.477259636f, 0.396080256f, 0.f),
    C_Vertex3f(0.477259636f, 0.396080256f, -1.f),
    C_Normal3f(0.511946261f, -0.859017491f, 0.f),
    C_Vertex3f(0.467614651f, 0.390995383f, 0.f),
    C_Vertex3f(0.467614651f, 0.390995383f, -1.f),
    C_Normal3f(0.597727895f, -0.801699042f, 0.f),
    C_Vertex3f(0.458488941f, 0.384893537f, 0.f),
    C_Vertex3f(0.458488941f, 0.384893537f, -1.f),
    C_Normal3f(0.674574196f, -0.738207102f, 0.f),
    C_Vertex3f(0.449882507f, 0.377774715f, 0.f),
    C_Vertex3f(0.449882507f, 0.377774715f, -1.f),
    C_Normal3f(0.75563556f, -0.654992282f, 0.f),
    C_Vertex3f(0.441795349f, 0.36963892f, 0.f),
    C_Vertex3f(0.441795349f, 0.36963892f, -1.f),
    C_Normal3f(0.834931493f, -0.550353944f, 0.f),
    C_Vertex3f(0.434719086f, 0.360281348f, 0.f),
    C_Vertex3f(0.434719086f, 0.360281348f, -1.f),
    C_Normal3f(0.895551026f, -0.444958746f, 0.f),
    C_Vertex3f(0.428653717f, 0.349702001f, 0.f),
    C_Vertex3f(0.428653717f, 0.349702001f, -1.f),
    C_Normal3f(0.93892628f, -0.344118327f, 0.f),
    C_Vertex3f(0.423599243f, 0.337900877f, 0.f),
    C_Vertex3f(0.423599243f, 0.337900877f, -1.f),
    C_Normal3f(0.967937469f, -0.251191407f, 0.f),
    C_Vertex3f(0.419555664f, 0.324877977f, 0.f),
    C_Vertex3f(0.419555664f, 0.324877977f, -1.f),
    C_Normal3f(0.985836208f, -0.167711005f, 0.f),
    C_Vertex3f(0.41652298f, 0.310633302f, 0.f),
    C_Vertex3f(0.41652298f, 0.310633302f, -1.f),
    C_Normal3f(0.995581746f, -0.0938987806f, 0.f),
    C_Vertex3f(0.41450119f, 0.29516685f, 0.f),
    C_Vertex3f(0.41450119f, 0.29516685f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.413490295f, 0.278478622f, 0.f),
    C_Vertex3f(0.413490295f, 0.278478622f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.413490295f, 0.278478622f, 0.f),
    C_Vertex3f(0.413490295f, 0.278478622f, -1.f),
    C_Vertex3f(0.413490295f, 0.f, -1.f),
    C_Vertex3f(0.413490295f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.413490295f, 0.f, 0.f),
    C_Vertex3f(0.413490295f, 0.f, -1.f),
    C_Vertex3f(0.334877014f, 0.f, -1.f),
    C_Vertex3f(0.334877014f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.334877014f, 0.f, 0.f),
    C_Vertex3f(0.334877014f, 0.f, -1.f),
    C_Normal3f(-0.999998212f, -0.00187793339f, 0.f),
    C_Vertex3f(0.334877014f, 0.307966232f, 0.f),
    C_Vertex3f(0.334877014f, 0.307966232f, -1.f),
    C_Normal3f(-0.997126997f, -0.0757480785f, 0.f),
    C_Vertex3f(0.334275603f, 0.320250869f, 0.f),
    C_Vertex3f(0.334275603f, 0.320250869f, -1.f),
    C_Normal3f(-0.990880013f, -0.134747297f, 0.f),
    C_Vertex3f(0.333072782f, 0.331716657f, 0.f),
    C_Vertex3f(0.333072782f, 0.331716657f, -1.f),
    C_Normal3f(-0.979509652f, -0.201397225f, 0.f),
    C_Vertex3f(0.331268549f, 0.342363596f, 0.f),
    C_Vertex3f(0.331268549f, 0.342363596f, -1.f),
    C_Normal3f(-0.961110413f, -0.276164383f, 0.f),
    C_Vertex3f(0.328862906f, 0.352191687f, 0.f),
    C_Vertex3f(0.328862906f, 0.352191687f, -1.f),
    C_Normal3f(-0.933340609f, -0.3589921f, 0.f),
    C_Vertex3f(0.325855851f, 0.361200929f, 0.f),
    C_Vertex3f(0.325855851f, 0.361200929f, -1.f),
    C_Normal3f(-0.893567204f, -0.448929489f, 0.f),
    C_Vertex3f(0.322247386f, 0.369391322f, 0.f),
    C_Vertex3f(0.322247386f, 0.369391322f, -1.f),
    C_Normal3f(-0.839257777f, -0.543733835f, 0.f),
    C_Vertex3f(0.31803751f, 0.376762867f, 0.f),
    C_Vertex3f(0.31803751f, 0.376762867f, -1.f),
    C_Normal3f(-0.763634086f, -0.645649254f, 0.f),
    C_Vertex3f(0.313226223f, 0.383315563f, 0.f),
    C_Vertex3f(0.313226223f, 0.383315563f, -1.f),
    C_Normal3f(-0.66623354f, -0.745743275f, 0.f),
    C_Vertex3f(0.307649493f, 0.389049172f, 0.f),
    C_Vertex3f(0.307649493f, 0.389049172f, -1.f),
    C_Normal3f(-0.556556225f, -0.83081001f, 0.f),
    C_Vertex3f(0.301307321f, 0.393963695f, 0.f),
    C_Vertex3f(0.301307321f, 0.393963695f, -1.f),
    C_Normal3f(-0.441524088f, -0.897249401f, 0.f),
    C_Vertex3f(0.294199705f, 0.39805913f, 0.f),
    C_Vertex3f(0.294199705f, 0.39805913f, -1.f),
    C_Normal3f(-0.328033566f, -0.944666028f, 0.f),
    C_Vertex3f(0.286326647f, 0.401335478f, 0.f),
    C_Vertex3f(0.286326647f, 0.401335478f, -1.f),
    C_Normal3f(-0.221357942f, -0.975192726f, 0.f),
    C_Vertex3f(0.277688146f, 0.403792739f, 0.f),
    C_Vertex3f(0.277688146f, 0.403792739f, -1.f),
    C_Normal3f(-0.12456353f, -0.992211699f, 0.f),
    C_Vertex3f(0.268284202f, 0.405430913f, 0.f),
    C_Vertex3f(0.268284202f, 0.405430913f, -1.f),
    C_Normal3f(-0.0341840014f, -0.999415636f, 0.f),
    C_Vertex3f(0.258114815f, 0.40625f, 0.f),
    C_Vertex3f(0.258114815f, 0.40625f, -1.f),
    C_Normal3f(0.0731632113f, -0.997319937f, 0.f),
    C_Vertex3f(0.244337082f, 0.40625f, 0.f),
    C_Vertex3f(0.244337082f, 0.40625f, -1.f),
    C_Normal3f(0.20453082f, -0.97886008f, 0.f),
    C_Vertex3f(0.228344917f, 0.404066086f, 0.f),
    C_Vertex3f(0.228344917f, 0.404066086f, -1.f),
    C_Normal3f(0.341039866f, -0.940048814f, 0.f),
    C_Vertex3f(0.212981224f, 0.399698257f, 0.f),
    C_Vertex3f(0.212981224f, 0.399698257f, -1.f),
    C_Normal3f(0.468321204f, -0.883558214f, 0.f),
    C_Vertex3f(0.198246002f, 0.393146515f, 0.f),
    C_Vertex3f(0.198246002f, 0.393146515f, -1.f),
    C_Normal3f(0.5958336f, -0.803107917f, 0.f),
    C_Vertex3f(0.184139252f, 0.384410858f, 0.f),
    C_Vertex3f(0.184139252f, 0.384410858f, -1.f),
    C_Normal3f(0.724474788f, -0.689301372f, 0.f),
    C_Vertex3f(0.171973705f, 0.373654842f, 0.f),
    C_Vertex3f(0.171973705f, 0.373654842f, -1.f),
    C_Normal3f(0.830303192f, -0.557311833f, 0.f),
    C_Vertex3f(0.161749363f, 0.360878468f, 0.f),
    C_Vertex3f(0.161749363f, 0.360878468f, -1.f),
    C_Normal3f(0.903021336f, -0.429595709f, 0.f),
    C_Vertex3f(0.153466225f, 0.346081734f, 0.f),
    C_Vertex3f(0.153466225f, 0.346081734f, -1.f),
    C_Normal3f(0.945874929f, -0.324531436f, 0.f),
    C_Vertex3f(0.148709774f, 0.333468914f, 0.f),
    C_Vertex3f(0.148709774f, 0.333468914f, -1.f),
    C_Normal3f(0.967144608f, -0.254226953f, 0.f),
    C_Vertex3f(0.145935178f, 0.324131846f, 0.f),
    C_Vertex3f(0.145935178f, 0.324131846f, -1.f),
    C_Normal3f(0.979977727f, -0.199107215f, 0.f),
    C_Vertex3f(0.143556952f, 0.313866258f, 0.f),
    C_Vertex3f(0.143556952f, 0.313866258f, -1.f),
    C_Normal3f(0.988497853f, -0.15123494f, 0.f),
    C_Vertex3f(0.141575098f, 0.302672148f, 0.f),
    C_Vertex3f(0.141575098f, 0.302672148f, -1.f),
    C_Normal3f(0.993980825f, -0.109554268f, 0.f),
    C_Vertex3f(0.139989614f, 0.290549517f, 0.f),
    C_Vertex3f(0.139989614f, 0.290549517f, -1.f),
    C_Normal3f(0.997323036f, -0.0731220394f, 0.f),
    C_Vertex3f(0.138800502f, 0.277498364f, 0.f),
    C_Vertex3f(0.138800502f, 0.277498364f, -1.f),
    C_Normal3f(0.999153912f, -0.0411282182f, 0.f),
    C_Vertex3f(0.13800776f, 0.263518691f, 0.f),
    C_Vertex3f(0.13800776f, 0.263518691f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137611389f, 0.248610497f, 0.f),
    C_Vertex3f(0.137611389f, 0.248610497f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137611389f, 0.248610497f, 0.f),
    C_Vertex3f(0.137611389f, 0.248610497f, -1.f),
    C_Vertex3f(0.137611389f, 0.f, -1.f),
    C_Vertex3f(0.137611389f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.137611389f, 0.f, 0.f),
    C_Vertex3f(0.137611389f, 0.f, -1.f),
    C_Vertex3f(0.0589981079f, 0.f, -1.f),
    C_Vertex3f(0.0589981079f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.346442223f, 0.45851326f, 0.f),
    C_Vertex3f(0.327615261f, 0.466561317f, 0.f),
    C_Vertex3f(0.328862906f, 0.352191687f, 0.f),
    C_Vertex3f(0.331268549f, 0.342363596f, 0.f),
    C_Vertex3f(0.333072782f, 0.331716657f, 0.f),
    C_Vertex3f(0.334275603f, 0.320250869f, 0.f),
    C_Vertex3f(0.334877014f, 0.307966232f, 0.f),
    C_Vertex3f(0.413490295f, 0.f, 0.f),
    C_Vertex3f(0.36294651f, 0.447782516f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.327615261f, 0.466561317f, 0.f),
    C_Vertex3f(0.306465626f, 0.471926689f, 0.f),
    C_Vertex3f(0.307649493f, 0.389049172f, 0.f),
    C_Vertex3f(0.313226223f, 0.383315563f, 0.f),
    C_Vertex3f(0.31803751f, 0.376762867f, 0.f),
    C_Vertex3f(0.322247386f, 0.369391322f, 0.f),
    C_Vertex3f(0.325855851f, 0.361200929f, 0.f),
    C_Vertex3f(0.328862906f, 0.352191687f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.306465626f, 0.471926689f, 0.f),
    C_Vertex3f(0.282993317f, 0.474609375f, 0.f),
    C_Vertex3f(0.286326647f, 0.401335478f, 0.f),
    C_Vertex3f(0.294199705f, 0.39805913f, 0.f),
    C_Vertex3f(0.301307321f, 0.393963695f, 0.f),
    C_Vertex3f(0.307649493f, 0.389049172f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.244337082f, 0.40625f, 0.f),
    C_Vertex3f(0.258114815f, 0.40625f, 0.f),
    C_Vertex3f(0.258508682f, 0.474609375f, 0.f),
    C_Vertex3f(0.268284202f, 0.405430913f, 0.f),
    C_Vertex3f(0.282993317f, 0.474609375f, 0.f),
    C_Vertex3f(0.277688146f, 0.403792739f, 0.f),
    C_Vertex3f(0.286326647f, 0.401335478f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.198246002f, 0.393146515f, 0.f),
    C_Vertex3f(0.212981224f, 0.399698257f, 0.f),
    C_Vertex3f(0.215932369f, 0.466805935f, 0.f),
    C_Vertex3f(0.228344917f, 0.404066086f, 0.f),
    C_Vertex3f(0.236591816f, 0.472008228f, 0.f),
    C_Vertex3f(0.244337082f, 0.40625f, 0.f),
    C_Vertex3f(0.258508682f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153466225f, 0.346081734f, 0.f),
    C_Vertex3f(0.161749363f, 0.360878468f, 0.f),
    C_Vertex3f(0.162045002f, 0.436523914f, 0.f),
    C_Vertex3f(0.171973705f, 0.373654842f, 0.f),
    C_Vertex3f(0.178385735f, 0.448597908f, 0.f),
    C_Vertex3f(0.184139252f, 0.384410858f, 0.f),
    C_Vertex3f(0.196530342f, 0.459002495f, 0.f),
    C_Vertex3f(0.198246002f, 0.393146515f, 0.f),
    C_Vertex3f(0.215932369f, 0.466805935f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.143556952f, 0.313866258f, 0.f),
    C_Vertex3f(0.145935178f, 0.324131846f, 0.f),
    C_Vertex3f(0.147508144f, 0.422780514f, 0.f),
    C_Vertex3f(0.148709774f, 0.333468914f, 0.f),
    C_Vertex3f(0.162045002f, 0.436523914f, 0.f),
    C_Vertex3f(0.153466225f, 0.346081734f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147508144f, 0.422780514f, 0.f),
    C_Vertex3f(0.134775162f, 0.407367706f, 0.f),
    C_Vertex3f(0.138800502f, 0.277498364f, 0.f),
    C_Vertex3f(0.139989614f, 0.290549517f, 0.f),
    C_Vertex3f(0.141575098f, 0.302672148f, 0.f),
    C_Vertex3f(0.143556952f, 0.313866258f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134775162f, 0.407367706f, 0.f),
    C_Vertex3f(0.129310608f, 0.398826599f, 0.f),
    C_Vertex3f(0.137611389f, 0.f, 0.f),
    C_Vertex3f(0.137611389f, 0.248610497f, 0.f),
    C_Vertex3f(0.13800776f, 0.263518691f, 0.f),
    C_Vertex3f(0.138800502f, 0.277498364f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0589981079f, 0.464355469f, 0.f),
    C_Vertex3f(0.0589981079f, 0.f, 0.f),
    C_Vertex3f(0.137611389f, 0.f, 0.f),
    C_Vertex3f(0.129310608f, 0.398826599f, 0.f),
    C_Vertex3f(0.129310608f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.61264205f, 0.462659836f, 0.f),
    C_Vertex3f(0.599346399f, 0.467439651f, 0.f),
    C_Vertex3f(0.600214005f, 0.365351677f, 0.f),
    C_Vertex3f(0.604040146f, 0.354977608f, 0.f),
    C_Vertex3f(0.606909752f, 0.341299534f, 0.f),
    C_Vertex3f(0.608822823f, 0.324317455f, 0.f),
    C_Vertex3f(0.609779358f, 0.304031372f, 0.f),
    C_Vertex3f(0.688392639f, 0.f, 0.f),
    C_Vertex3f(0.624939084f, 0.456685066f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.558412075f, 0.401499271f, 0.f),
    C_Vertex3f(0.569340706f, 0.396748543f, 0.f),
    C_Vertex3f(0.56975925f, 0.473414421f, 0.f),
    C_Vertex3f(0.579394341f, 0.390414238f, 0.f),
    C_Vertex3f(0.585052133f, 0.471024513f, 0.f),
    C_Vertex3f(0.587891102f, 0.383069992f, 0.f),
    C_Vertex3f(0.599346399f, 0.467439651f, 0.f),
    C_Vertex3f(0.59483099f, 0.374715805f, 0.f),
    C_Vertex3f(0.600214005f, 0.365351677f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.553467751f, 0.474609375f, 0.f),
    C_Vertex3f(0.533771515f, 0.474609375f, 0.f),
    C_Vertex3f(0.533929825f, 0.40625f, 0.f),
    C_Vertex3f(0.546608448f, 0.404666424f, 0.f),
    C_Vertex3f(0.558412075f, 0.401499271f, 0.f),
    C_Vertex3f(0.56975925f, 0.473414421f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.498107433f, 0.403199077f, 0.f),
    C_Vertex3f(0.509310246f, 0.405233026f, 0.f),
    C_Vertex3f(0.512708426f, 0.472063541f, 0.f),
    C_Vertex3f(0.521032333f, 0.40625f, 0.f),
    C_Vertex3f(0.533771515f, 0.474609375f, 0.f),
    C_Vertex3f(0.533929825f, 0.40625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.492684603f, 0.466971874f, 0.f),
    C_Vertex3f(0.473700047f, 0.459334373f, 0.f),
    C_Vertex3f(0.477259636f, 0.396080256f, 0.f),
    C_Vertex3f(0.487423897f, 0.400148153f, 0.f),
    C_Vertex3f(0.498107433f, 0.403199077f, 0.f),
    C_Vertex3f(0.512708426f, 0.472063541f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.441795349f, 0.36963892f, 0.f),
    C_Vertex3f(0.449882507f, 0.377774715f, 0.f),
    C_Vertex3f(0.455754757f, 0.449151039f, 0.f),
    C_Vertex3f(0.458488941f, 0.384893537f, 0.f),
    C_Vertex3f(0.473700047f, 0.459334373f, 0.f),
    C_Vertex3f(0.467614651f, 0.390995383f, 0.f),
    C_Vertex3f(0.477259636f, 0.396080256f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.438848734f, 0.436421871f, 0.f),
    C_Vertex3f(0.422981977f, 0.42114687f, 0.f),
    C_Vertex3f(0.423599243f, 0.337900877f, 0.f),
    C_Vertex3f(0.428653717f, 0.349702001f, 0.f),
    C_Vertex3f(0.434719086f, 0.360281348f, 0.f),
    C_Vertex3f(0.441795349f, 0.36963892f, 0.f),
    C_Vertex3f(0.455754757f, 0.449151039f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.422981977f, 0.42114687f, 0.f),
    C_Vertex3f(0.408154488f, 0.403326035f, 0.f),
    C_Vertex3f(0.41450119f, 0.29516685f, 0.f),
    C_Vertex3f(0.41652298f, 0.310633302f, 0.f),
    C_Vertex3f(0.419555664f, 0.324877977f, 0.f),
    C_Vertex3f(0.423599243f, 0.337900877f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.408154488f, 0.403326035f, 0.f),
    C_Vertex3f(0.401260376f, 0.3931427f, 0.f),
    C_Vertex3f(0.413490295f, 0.f, 0.f),
    C_Vertex3f(0.413490295f, 0.278478622f, 0.f),
    C_Vertex3f(0.41450119f, 0.29516685f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.413490295f, 0.f, 0.f),
    C_Vertex3f(0.401260376f, 0.3931427f, 0.f),
    C_Vertex3f(0.397871017f, 0.402778625f, 0.f),
    C_Vertex3f(0.388660908f, 0.419915199f, 0.f),
    C_Vertex3f(0.377019405f, 0.434916496f, 0.f),
    C_Vertex3f(0.36294651f, 0.447782516f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.688392639f, 0.f, 0.f),
    C_Vertex3f(0.688392639f, 0.328531265f, 0.f),
    C_Vertex3f(0.687229991f, 0.347126484f, 0.f),
    C_Vertex3f(0.684904695f, 0.364430904f, 0.f),
    C_Vertex3f(0.68141675f, 0.380444527f, 0.f),
    C_Vertex3f(0.676766157f, 0.395167351f, 0.f),
    C_Vertex3f(0.670952916f, 0.408599377f, 0.f),
    C_Vertex3f(0.663977027f, 0.420740604f, 0.f),
    C_Vertex3f(0.65583849f, 0.431591034f, 0.f),
    C_Vertex3f(0.646537304f, 0.441150665f, 0.f),
    C_Vertex3f(0.636237502f, 0.449515343f, 0.f),
    C_Vertex3f(0.624939084f, 0.456685066f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.688392639f, 0.f, 0.f),
    C_Vertex3f(0.609779358f, 0.304031372f, 0.f),
    C_Vertex3f(0.609779358f, 0.f, 0.f),
    C_Vertex3f(0.413490295f, 0.f, 0.f),
    C_Vertex3f(0.334877014f, 0.307966232f, 0.f),
    C_Vertex3f(0.334877014f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.346442223f, 0.45851326f, -1.f),
    C_Vertex3f(0.327615261f, 0.466561317f, -1.f),
    C_Vertex3f(0.328862906f, 0.352191687f, -1.f),
    C_Vertex3f(0.331268549f, 0.342363596f, -1.f),
    C_Vertex3f(0.333072782f, 0.331716657f, -1.f),
    C_Vertex3f(0.334275603f, 0.320250869f, -1.f),
    C_Vertex3f(0.334877014f, 0.307966232f, -1.f),
    C_Vertex3f(0.413490295f, 0.f, -1.f),
    C_Vertex3f(0.36294651f, 0.447782516f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.327615261f, 0.466561317f, -1.f),
    C_Vertex3f(0.306465626f, 0.471926689f, -1.f),
    C_Vertex3f(0.307649493f, 0.389049172f, -1.f),
    C_Vertex3f(0.313226223f, 0.383315563f, -1.f),
    C_Vertex3f(0.31803751f, 0.376762867f, -1.f),
    C_Vertex3f(0.322247386f, 0.369391322f, -1.f),
    C_Vertex3f(0.325855851f, 0.361200929f, -1.f),
    C_Vertex3f(0.328862906f, 0.352191687f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.306465626f, 0.471926689f, -1.f),
    C_Vertex3f(0.282993317f, 0.474609375f, -1.f),
    C_Vertex3f(0.286326647f, 0.401335478f, -1.f),
    C_Vertex3f(0.294199705f, 0.39805913f, -1.f),
    C_Vertex3f(0.301307321f, 0.393963695f, -1.f),
    C_Vertex3f(0.307649493f, 0.389049172f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.244337082f, 0.40625f, -1.f),
    C_Vertex3f(0.258114815f, 0.40625f, -1.f),
    C_Vertex3f(0.258508682f, 0.474609375f, -1.f),
    C_Vertex3f(0.268284202f, 0.405430913f, -1.f),
    C_Vertex3f(0.282993317f, 0.474609375f, -1.f),
    C_Vertex3f(0.277688146f, 0.403792739f, -1.f),
    C_Vertex3f(0.286326647f, 0.401335478f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.198246002f, 0.393146515f, -1.f),
    C_Vertex3f(0.212981224f, 0.399698257f, -1.f),
    C_Vertex3f(0.215932369f, 0.466805935f, -1.f),
    C_Vertex3f(0.228344917f, 0.404066086f, -1.f),
    C_Vertex3f(0.236591816f, 0.472008228f, -1.f),
    C_Vertex3f(0.244337082f, 0.40625f, -1.f),
    C_Vertex3f(0.258508682f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153466225f, 0.346081734f, -1.f),
    C_Vertex3f(0.161749363f, 0.360878468f, -1.f),
    C_Vertex3f(0.162045002f, 0.436523914f, -1.f),
    C_Vertex3f(0.171973705f, 0.373654842f, -1.f),
    C_Vertex3f(0.178385735f, 0.448597908f, -1.f),
    C_Vertex3f(0.184139252f, 0.384410858f, -1.f),
    C_Vertex3f(0.196530342f, 0.459002495f, -1.f),
    C_Vertex3f(0.198246002f, 0.393146515f, -1.f),
    C_Vertex3f(0.215932369f, 0.466805935f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.143556952f, 0.313866258f, -1.f),
    C_Vertex3f(0.145935178f, 0.324131846f, -1.f),
    C_Vertex3f(0.147508144f, 0.422780514f, -1.f),
    C_Vertex3f(0.148709774f, 0.333468914f, -1.f),
    C_Vertex3f(0.162045002f, 0.436523914f, -1.f),
    C_Vertex3f(0.153466225f, 0.346081734f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147508144f, 0.422780514f, -1.f),
    C_Vertex3f(0.134775162f, 0.407367706f, -1.f),
    C_Vertex3f(0.138800502f, 0.277498364f, -1.f),
    C_Vertex3f(0.139989614f, 0.290549517f, -1.f),
    C_Vertex3f(0.141575098f, 0.302672148f, -1.f),
    C_Vertex3f(0.143556952f, 0.313866258f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134775162f, 0.407367706f, -1.f),
    C_Vertex3f(0.129310608f, 0.398826599f, -1.f),
    C_Vertex3f(0.137611389f, 0.f, -1.f),
    C_Vertex3f(0.137611389f, 0.248610497f, -1.f),
    C_Vertex3f(0.13800776f, 0.263518691f, -1.f),
    C_Vertex3f(0.138800502f, 0.277498364f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0589981079f, 0.464355469f, -1.f),
    C_Vertex3f(0.0589981079f, 0.f, -1.f),
    C_Vertex3f(0.137611389f, 0.f, -1.f),
    C_Vertex3f(0.129310608f, 0.398826599f, -1.f),
    C_Vertex3f(0.129310608f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.61264205f, 0.462659836f, -1.f),
    C_Vertex3f(0.599346399f, 0.467439651f, -1.f),
    C_Vertex3f(0.600214005f, 0.365351677f, -1.f),
    C_Vertex3f(0.604040146f, 0.354977608f, -1.f),
    C_Vertex3f(0.606909752f, 0.341299534f, -1.f),
    C_Vertex3f(0.608822823f, 0.324317455f, -1.f),
    C_Vertex3f(0.609779358f, 0.304031372f, -1.f),
    C_Vertex3f(0.688392639f, 0.f, -1.f),
    C_Vertex3f(0.624939084f, 0.456685066f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.558412075f, 0.401499271f, -1.f),
    C_Vertex3f(0.569340706f, 0.396748543f, -1.f),
    C_Vertex3f(0.56975925f, 0.473414421f, -1.f),
    C_Vertex3f(0.579394341f, 0.390414238f, -1.f),
    C_Vertex3f(0.585052133f, 0.471024513f, -1.f),
    C_Vertex3f(0.587891102f, 0.383069992f, -1.f),
    C_Vertex3f(0.599346399f, 0.467439651f, -1.f),
    C_Vertex3f(0.59483099f, 0.374715805f, -1.f),
    C_Vertex3f(0.600214005f, 0.365351677f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.553467751f, 0.474609375f, -1.f),
    C_Vertex3f(0.533771515f, 0.474609375f, -1.f),
    C_Vertex3f(0.533929825f, 0.40625f, -1.f),
    C_Vertex3f(0.546608448f, 0.404666424f, -1.f),
    C_Vertex3f(0.558412075f, 0.401499271f, -1.f),
    C_Vertex3f(0.56975925f, 0.473414421f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.498107433f, 0.403199077f, -1.f),
    C_Vertex3f(0.509310246f, 0.405233026f, -1.f),
    C_Vertex3f(0.512708426f, 0.472063541f, -1.f),
    C_Vertex3f(0.521032333f, 0.40625f, -1.f),
    C_Vertex3f(0.533771515f, 0.474609375f, -1.f),
    C_Vertex3f(0.533929825f, 0.40625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.492684603f, 0.466971874f, -1.f),
    C_Vertex3f(0.473700047f, 0.459334373f, -1.f),
    C_Vertex3f(0.477259636f, 0.396080256f, -1.f),
    C_Vertex3f(0.487423897f, 0.400148153f, -1.f),
    C_Vertex3f(0.498107433f, 0.403199077f, -1.f),
    C_Vertex3f(0.512708426f, 0.472063541f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.441795349f, 0.36963892f, -1.f),
    C_Vertex3f(0.449882507f, 0.377774715f, -1.f),
    C_Vertex3f(0.455754757f, 0.449151039f, -1.f),
    C_Vertex3f(0.458488941f, 0.384893537f, -1.f),
    C_Vertex3f(0.473700047f, 0.459334373f, -1.f),
    C_Vertex3f(0.467614651f, 0.390995383f, -1.f),
    C_Vertex3f(0.477259636f, 0.396080256f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.438848734f, 0.436421871f, -1.f),
    C_Vertex3f(0.422981977f, 0.42114687f, -1.f),
    C_Vertex3f(0.423599243f, 0.337900877f, -1.f),
    C_Vertex3f(0.428653717f, 0.349702001f, -1.f),
    C_Vertex3f(0.434719086f, 0.360281348f, -1.f),
    C_Vertex3f(0.441795349f, 0.36963892f, -1.f),
    C_Vertex3f(0.455754757f, 0.449151039f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.422981977f, 0.42114687f, -1.f),
    C_Vertex3f(0.408154488f, 0.403326035f, -1.f),
    C_Vertex3f(0.41450119f, 0.29516685f, -1.f),
    C_Vertex3f(0.41652298f, 0.310633302f, -1.f),
    C_Vertex3f(0.419555664f, 0.324877977f, -1.f),
    C_Vertex3f(0.423599243f, 0.337900877f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.408154488f, 0.403326035f, -1.f),
    C_Vertex3f(0.401260376f, 0.3931427f, -1.f),
    C_Vertex3f(0.413490295f, 0.f, -1.f),
    C_Vertex3f(0.413490295f, 0.278478622f, -1.f),
    C_Vertex3f(0.41450119f, 0.29516685f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.413490295f, 0.f, -1.f),
    C_Vertex3f(0.401260376f, 0.3931427f, -1.f),
    C_Vertex3f(0.397871017f, 0.402778625f, -1.f),
    C_Vertex3f(0.388660908f, 0.419915199f, -1.f),
    C_Vertex3f(0.377019405f, 0.434916496f, -1.f),
    C_Vertex3f(0.36294651f, 0.447782516f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.688392639f, 0.f, -1.f),
    C_Vertex3f(0.688392639f, 0.328531265f, -1.f),
    C_Vertex3f(0.687229991f, 0.347126484f, -1.f),
    C_Vertex3f(0.684904695f, 0.364430904f, -1.f),
    C_Vertex3f(0.68141675f, 0.380444527f, -1.f),
    C_Vertex3f(0.676766157f, 0.395167351f, -1.f),
    C_Vertex3f(0.670952916f, 0.408599377f, -1.f),
    C_Vertex3f(0.663977027f, 0.420740604f, -1.f),
    C_Vertex3f(0.65583849f, 0.431591034f, -1.f),
    C_Vertex3f(0.646537304f, 0.441150665f, -1.f),
    C_Vertex3f(0.636237502f, 0.449515343f, -1.f),
    C_Vertex3f(0.624939084f, 0.456685066f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.688392639f, 0.f, -1.f),
    C_Vertex3f(0.609779358f, 0.304031372f, -1.f),
    C_Vertex3f(0.609779358f, 0.f, -1.f),
    C_Vertex3f(0.413490295f, 0.f, -1.f),
    C_Vertex3f(0.334877014f, 0.307966232f, -1.f),
    C_Vertex3f(0.334877014f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.745605469f, 0.f, 0.f),
};
static const Call char_110[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.129882813f, 0.464355469f, -1.f),
    C_Vertex3f(0.129882813f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.129882813f, 0.464355469f, 0.f),
    C_Vertex3f(0.129882813f, 0.464355469f, -1.f),
    C_Vertex3f(0.129882813f, 0.398361206f, -1.f),
    C_Vertex3f(0.129882813f, 0.398361206f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.830692351f, 0.556731761f, 0.f),
    C_Vertex3f(0.129882813f, 0.398361206f, 0.f),
    C_Vertex3f(0.129882813f, 0.398361206f, -1.f),
    C_Normal3f(-0.786487639f, 0.617606044f, 0.f),
    C_Vertex3f(0.136270523f, 0.407892227f, 0.f),
    C_Vertex3f(0.136270523f, 0.407892227f, -1.f),
    C_Normal3f(-0.720558703f, 0.693393946f, 0.f),
    C_Vertex3f(0.150465012f, 0.424571514f, 0.f),
    C_Vertex3f(0.150465012f, 0.424571514f, -1.f),
    C_Normal3f(-0.626053929f, 0.779779792f, 0.f),
    C_Vertex3f(0.166078568f, 0.438868046f, 0.f),
    C_Vertex3f(0.166078568f, 0.438868046f, -1.f),
    C_Normal3f(-0.517227948f, 0.855847657f, 0.f),
    C_Vertex3f(0.183111191f, 0.450781822f, 0.f),
    C_Vertex3f(0.183111191f, 0.450781822f, -1.f),
    C_Normal3f(-0.39907527f, 0.916918218f, 0.f),
    C_Vertex3f(0.201562881f, 0.460312843f, 0.f),
    C_Vertex3f(0.201562881f, 0.460312843f, -1.f),
    C_Normal3f(-0.278033763f, 0.960571289f, 0.f),
    C_Vertex3f(0.22143364f, 0.467461109f, 0.f),
    C_Vertex3f(0.22143364f, 0.467461109f, -1.f),
    C_Normal3f(-0.16036272f, 0.987058163f, 0.f),
    C_Vertex3f(0.242723465f, 0.47222662f, 0.f),
    C_Vertex3f(0.242723465f, 0.47222662f, -1.f),
    C_Normal3f(-0.0525781661f, 0.998616874f, 0.f),
    C_Vertex3f(0.265432358f, 0.474609375f, 0.f),
    C_Vertex3f(0.265432358f, 0.474609375f, -1.f),
    C_Normal3f(0.0441252515f, 0.99902606f, 0.f),
    C_Vertex3f(0.287979126f, 0.474609375f, 0.f),
    C_Vertex3f(0.287979126f, 0.474609375f, -1.f),
    C_Normal3f(0.142065495f, 0.989857197f, 0.f),
    C_Vertex3f(0.308097839f, 0.472724915f, 0.f),
    C_Vertex3f(0.308097839f, 0.472724915f, -1.f),
    C_Normal3f(0.242489442f, 0.970154107f, 0.f),
    C_Vertex3f(0.32736969f, 0.468955994f, 0.f),
    C_Vertex3f(0.32736969f, 0.468955994f, -1.f),
    C_Normal3f(0.344026625f, 0.938959897f, 0.f),
    C_Vertex3f(0.345794678f, 0.463302612f, 0.f),
    C_Vertex3f(0.345794678f, 0.463302612f, -1.f),
    C_Normal3f(0.442368001f, 0.896833658f, 0.f),
    C_Vertex3f(0.363372803f, 0.455764771f, 0.f),
    C_Vertex3f(0.363372803f, 0.455764771f, -1.f),
    C_Normal3f(0.546218216f, 0.837642908f, 0.f),
    C_Vertex3f(0.378740311f, 0.447052002f, 0.f),
    C_Vertex3f(0.378740311f, 0.447052002f, -1.f),
    C_Normal3f(0.656018972f, 0.754744351f, 0.f),
    C_Vertex3f(0.391897202f, 0.437164307f, 0.f),
    C_Vertex3f(0.391897202f, 0.437164307f, -1.f),
    C_Normal3f(0.763929188f, 0.645300031f, 0.f),
    C_Vertex3f(0.402843475f, 0.426101685f, 0.f),
    C_Vertex3f(0.402843475f, 0.426101685f, -1.f),
    C_Normal3f(0.845810473f, 0.533483624f, 0.f),
    C_Vertex3f(0.411579132f, 0.413864136f, 0.f),
    C_Vertex3f(0.411579132f, 0.413864136f, -1.f),
    C_Normal3f(0.899753392f, 0.436398685f, 0.f),
    C_Vertex3f(0.41900444f, 0.400479317f, 0.f),
    C_Vertex3f(0.41900444f, 0.400479317f, -1.f),
    C_Normal3f(0.940455556f, 0.339916557f, 0.f),
    C_Vertex3f(0.4251194f, 0.385947227f, 0.f),
    C_Vertex3f(0.4251194f, 0.385947227f, -1.f),
    C_Normal3f(0.966733932f, 0.255783916f, 0.f),
    C_Vertex3f(0.429924011f, 0.370267868f, 0.f),
    C_Vertex3f(0.429924011f, 0.370267868f, -1.f),
    C_Normal3f(0.983459711f, 0.181127235f, 0.f),
    C_Vertex3f(0.432216644f, 0.35912323f, 0.f),
    C_Vertex3f(0.432216644f, 0.35912323f, -1.f),
    C_Normal3f(0.992074847f, 0.125648394f, 0.f),
    C_Vertex3f(0.433171272f, 0.352636337f, 0.f),
    C_Vertex3f(0.433171272f, 0.352636337f, -1.f),
    C_Normal3f(0.995658576f, 0.0930804089f, 0.f),
    C_Vertex3f(0.433989525f, 0.345125198f, 0.f),
    C_Vertex3f(0.433989525f, 0.345125198f, -1.f),
    C_Normal3f(0.997707486f, 0.06767416f, 0.f),
    C_Vertex3f(0.434671402f, 0.336589813f, 0.f),
    C_Vertex3f(0.434671402f, 0.336589813f, -1.f),
    C_Normal3f(0.998878896f, 0.0473382138f, 0.f),
    C_Vertex3f(0.435216904f, 0.327030182f, 0.f),
    C_Vertex3f(0.435216904f, 0.327030182f, -1.f),
    C_Normal3f(0.999528348f, 0.0307117645f, 0.f),
    C_Vertex3f(0.43562603f, 0.316446304f, 0.f),
    C_Vertex3f(0.43562603f, 0.316446304f, -1.f),
    C_Normal3f(0.999857664f, 0.0168754011f, 0.f),
    C_Vertex3f(0.435898781f, 0.304838181f, 0.f),
    C_Vertex3f(0.435898781f, 0.304838181f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.436035156f, 0.292205811f, 0.f),
    C_Vertex3f(0.436035156f, 0.292205811f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.436035156f, 0.292205811f, 0.f),
    C_Vertex3f(0.436035156f, 0.292205811f, -1.f),
    C_Vertex3f(0.436035156f, 0.f, -1.f),
    C_Vertex3f(0.436035156f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.436035156f, 0.f, 0.f),
    C_Vertex3f(0.436035156f, 0.f, -1.f),
    C_Vertex3f(0.357421875f, 0.f, -1.f),
    C_Vertex3f(0.357421875f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.357421875f, 0.f, 0.f),
    C_Vertex3f(0.357421875f, 0.f, -1.f),
    C_Normal3f(-0.999993324f, -0.00363833085f, 0.f),
    C_Vertex3f(0.357421875f, 0.294200897f, 0.f),
    C_Vertex3f(0.357421875f, 0.294200897f, -1.f),
    C_Normal3f(-0.996123493f, -0.0879653245f, 0.f),
    C_Vertex3f(0.356275082f, 0.31519556f, 0.f),
    C_Vertex3f(0.356275082f, 0.31519556f, -1.f),
    C_Normal3f(-0.98514843f, -0.171704516f, 0.f),
    C_Vertex3f(0.353981495f, 0.333159924f, 0.f),
    C_Vertex3f(0.353981495f, 0.333159924f, -1.f),
    C_Normal3f(-0.958059728f, -0.286568522f, 0.f),
    C_Vertex3f(0.350541115f, 0.348093987f, 0.f),
    C_Vertex3f(0.350541115f, 0.348093987f, -1.f),
    C_Normal3f(-0.900001645f, -0.435886383f, 0.f),
    C_Vertex3f(0.345953941f, 0.359997749f, 0.f),
    C_Vertex3f(0.345953941f, 0.359997749f, -1.f),
    C_Normal3f(-0.812479973f, -0.582989037f, 0.f),
    C_Vertex3f(0.339592457f, 0.370700359f, 0.f),
    C_Vertex3f(0.339592457f, 0.370700359f, -1.f),
    C_Normal3f(-0.70227474f, -0.711906016f, 0.f),
    C_Vertex3f(0.331456661f, 0.380201817f, 0.f),
    C_Vertex3f(0.331456661f, 0.380201817f, -1.f),
    C_Normal3f(-0.580608785f, -0.814182699f, 0.f),
    C_Vertex3f(0.321546555f, 0.388502121f, 0.f),
    C_Vertex3f(0.321546555f, 0.388502121f, -1.f),
    C_Normal3f(-0.454031795f, -0.890985548f, 0.f),
    C_Vertex3f(0.309862137f, 0.395601273f, 0.f),
    C_Vertex3f(0.309862137f, 0.395601273f, -1.f),
    C_Normal3f(-0.318599582f, -0.947889447f, 0.f),
    C_Vertex3f(0.297166824f, 0.400925636f, 0.f),
    C_Vertex3f(0.297166824f, 0.400925636f, -1.f),
    C_Normal3f(-0.184121281f, -0.98290354f, 0.f),
    C_Vertex3f(0.283460617f, 0.404475212f, 0.f),
    C_Vertex3f(0.283460617f, 0.404475212f, -1.f),
    C_Normal3f(-0.0613803118f, -0.998114467f, 0.f),
    C_Vertex3f(0.268743515f, 0.40625f, 0.f),
    C_Vertex3f(0.268743515f, 0.40625f, -1.f),
    C_Normal3f(0.0379030257f, -0.999281406f, 0.f),
    C_Vertex3f(0.254600525f, 0.40625f, 0.f),
    C_Vertex3f(0.254600525f, 0.40625f, -1.f),
    C_Normal3f(0.124490522f, -0.992220819f, 0.f),
    C_Vertex3f(0.24247241f, 0.40525353f, 0.f),
    C_Vertex3f(0.24247241f, 0.40525353f, -1.f),
    C_Normal3f(0.212007388f, -0.9772681f, 0.f),
    C_Vertex3f(0.230774164f, 0.403260589f, 0.f),
    C_Vertex3f(0.230774164f, 0.403260589f, -1.f),
    C_Normal3f(0.300902724f, -0.953654885f, 0.f),
    C_Vertex3f(0.219505787f, 0.400271177f, 0.f),
    C_Vertex3f(0.219505787f, 0.400271177f, -1.f),
    C_Normal3f(0.388869584f, -0.921292841f, 0.f),
    C_Vertex3f(0.208667278f, 0.396285295f, 0.f),
    C_Vertex3f(0.208667278f, 0.396285295f, -1.f),
    C_Normal3f(0.473541707f, -0.880771399f, 0.f),
    C_Vertex3f(0.198258638f, 0.391302943f, 0.f),
    C_Vertex3f(0.198258638f, 0.391302943f, -1.f),
    C_Normal3f(0.552798092f, -0.833315134f, 0.f),
    C_Vertex3f(0.188279867f, 0.385324121f, 0.f),
    C_Vertex3f(0.188279867f, 0.385324121f, -1.f),
    C_Normal3f(0.625019014f, -0.780609548f, 0.f),
    C_Vertex3f(0.178730965f, 0.378348827f, 0.f),
    C_Vertex3f(0.178730965f, 0.378348827f, -1.f),
    C_Normal3f(0.719850719f, -0.694128811f, 0.f),
    C_Vertex3f(0.169611931f, 0.370377064f, 0.f),
    C_Vertex3f(0.169611931f, 0.370377064f, -1.f),
    C_Normal3f(0.820990741f, -0.570941508f, 0.f),
    C_Vertex3f(0.161632776f, 0.360617042f, 0.f),
    C_Vertex3f(0.161632776f, 0.360617042f, -1.f),
    C_Normal3f(0.893041074f, -0.449975073f, 0.f),
    C_Vertex3f(0.154793501f, 0.349068761f, 0.f),
    C_Vertex3f(0.154793501f, 0.349068761f, -1.f),
    C_Normal3f(0.940752149f, -0.339094907f, 0.f),
    C_Vertex3f(0.149094105f, 0.335732222f, 0.f),
    C_Vertex3f(0.149094105f, 0.335732222f, -1.f),
    C_Normal3f(0.970358133f, -0.241671532f, 0.f),
    C_Vertex3f(0.144534588f, 0.320607424f, 0.f),
    C_Vertex3f(0.144534588f, 0.320607424f, -1.f),
    C_Normal3f(0.987435877f, -0.158020154f, 0.f),
    C_Vertex3f(0.14111495f, 0.303694367f, 0.f),
    C_Vertex3f(0.14111495f, 0.303694367f, -1.f),
    C_Normal3f(0.996214747f, -0.0869256482f, 0.f),
    C_Vertex3f(0.138835192f, 0.284993052f, 0.f),
    C_Vertex3f(0.138835192f, 0.284993052f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.264503479f, 0.f),
    C_Vertex3f(0.137695313f, 0.264503479f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.264503479f, 0.f),
    C_Vertex3f(0.137695313f, 0.264503479f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.363372803f, 0.455764771f, 0.f),
    C_Vertex3f(0.345794678f, 0.463302612f, 0.f),
    C_Vertex3f(0.345953941f, 0.359997749f, 0.f),
    C_Vertex3f(0.350541115f, 0.348093987f, 0.f),
    C_Vertex3f(0.353981495f, 0.333159924f, 0.f),
    C_Vertex3f(0.356275082f, 0.31519556f, 0.f),
    C_Vertex3f(0.357421875f, 0.294200897f, 0.f),
    C_Vertex3f(0.436035156f, 0.f, 0.f),
    C_Vertex3f(0.378740311f, 0.447052002f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.309862137f, 0.395601273f, 0.f),
    C_Vertex3f(0.321546555f, 0.388502121f, 0.f),
    C_Vertex3f(0.32736969f, 0.468955994f, 0.f),
    C_Vertex3f(0.331456661f, 0.380201817f, 0.f),
    C_Vertex3f(0.345794678f, 0.463302612f, 0.f),
    C_Vertex3f(0.339592457f, 0.370700359f, 0.f),
    C_Vertex3f(0.345953941f, 0.359997749f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268743515f, 0.40625f, 0.f),
    C_Vertex3f(0.283460617f, 0.404475212f, 0.f),
    C_Vertex3f(0.287979126f, 0.474609375f, 0.f),
    C_Vertex3f(0.297166824f, 0.400925636f, 0.f),
    C_Vertex3f(0.308097839f, 0.472724915f, 0.f),
    C_Vertex3f(0.309862137f, 0.395601273f, 0.f),
    C_Vertex3f(0.32736969f, 0.468955994f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230774164f, 0.403260589f, 0.f),
    C_Vertex3f(0.24247241f, 0.40525353f, 0.f),
    C_Vertex3f(0.242723465f, 0.47222662f, 0.f),
    C_Vertex3f(0.254600525f, 0.40625f, 0.f),
    C_Vertex3f(0.265432358f, 0.474609375f, 0.f),
    C_Vertex3f(0.268743515f, 0.40625f, 0.f),
    C_Vertex3f(0.287979126f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.22143364f, 0.467461109f, 0.f),
    C_Vertex3f(0.201562881f, 0.460312843f, 0.f),
    C_Vertex3f(0.208667278f, 0.396285295f, 0.f),
    C_Vertex3f(0.219505787f, 0.400271177f, 0.f),
    C_Vertex3f(0.230774164f, 0.403260589f, 0.f),
    C_Vertex3f(0.242723465f, 0.47222662f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.169611931f, 0.370377064f, 0.f),
    C_Vertex3f(0.178730965f, 0.378348827f, 0.f),
    C_Vertex3f(0.183111191f, 0.450781822f, 0.f),
    C_Vertex3f(0.188279867f, 0.385324121f, 0.f),
    C_Vertex3f(0.201562881f, 0.460312843f, 0.f),
    C_Vertex3f(0.198258638f, 0.391302943f, 0.f),
    C_Vertex3f(0.208667278f, 0.396285295f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.166078568f, 0.438868046f, 0.f),
    C_Vertex3f(0.150465012f, 0.424571514f, 0.f),
    C_Vertex3f(0.154793501f, 0.349068761f, 0.f),
    C_Vertex3f(0.161632776f, 0.360617042f, 0.f),
    C_Vertex3f(0.169611931f, 0.370377064f, 0.f),
    C_Vertex3f(0.183111191f, 0.450781822f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150465012f, 0.424571514f, 0.f),
    C_Vertex3f(0.136270523f, 0.407892227f, 0.f),
    C_Vertex3f(0.138835192f, 0.284993052f, 0.f),
    C_Vertex3f(0.14111495f, 0.303694367f, 0.f),
    C_Vertex3f(0.144534588f, 0.320607424f, 0.f),
    C_Vertex3f(0.149094105f, 0.335732222f, 0.f),
    C_Vertex3f(0.154793501f, 0.349068761f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129882813f, 0.464355469f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Vertex3f(0.129882813f, 0.398361206f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Vertex3f(0.136270523f, 0.407892227f, 0.f),
    C_Vertex3f(0.137695313f, 0.264503479f, 0.f),
    C_Vertex3f(0.138835192f, 0.284993052f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.436035156f, 0.f, 0.f),
    C_Vertex3f(0.436035156f, 0.292205811f, 0.f),
    C_Vertex3f(0.435898781f, 0.304838181f, 0.f),
    C_Vertex3f(0.43562603f, 0.316446304f, 0.f),
    C_Vertex3f(0.435216904f, 0.327030182f, 0.f),
    C_Vertex3f(0.434671402f, 0.336589813f, 0.f),
    C_Vertex3f(0.433989525f, 0.345125198f, 0.f),
    C_Vertex3f(0.433171272f, 0.352636337f, 0.f),
    C_Vertex3f(0.432216644f, 0.35912323f, 0.f),
    C_Vertex3f(0.429924011f, 0.370267868f, 0.f),
    C_Vertex3f(0.4251194f, 0.385947227f, 0.f),
    C_Vertex3f(0.41900444f, 0.400479317f, 0.f),
    C_Vertex3f(0.411579132f, 0.413864136f, 0.f),
    C_Vertex3f(0.402843475f, 0.426101685f, 0.f),
    C_Vertex3f(0.391897202f, 0.437164307f, 0.f),
    C_Vertex3f(0.378740311f, 0.447052002f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.436035156f, 0.f, 0.f),
    C_Vertex3f(0.357421875f, 0.294200897f, 0.f),
    C_Vertex3f(0.357421875f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Vertex3f(0.0590820313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.363372803f, 0.455764771f, -1.f),
    C_Vertex3f(0.345794678f, 0.463302612f, -1.f),
    C_Vertex3f(0.345953941f, 0.359997749f, -1.f),
    C_Vertex3f(0.350541115f, 0.348093987f, -1.f),
    C_Vertex3f(0.353981495f, 0.333159924f, -1.f),
    C_Vertex3f(0.356275082f, 0.31519556f, -1.f),
    C_Vertex3f(0.357421875f, 0.294200897f, -1.f),
    C_Vertex3f(0.436035156f, 0.f, -1.f),
    C_Vertex3f(0.378740311f, 0.447052002f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.309862137f, 0.395601273f, -1.f),
    C_Vertex3f(0.321546555f, 0.388502121f, -1.f),
    C_Vertex3f(0.32736969f, 0.468955994f, -1.f),
    C_Vertex3f(0.331456661f, 0.380201817f, -1.f),
    C_Vertex3f(0.345794678f, 0.463302612f, -1.f),
    C_Vertex3f(0.339592457f, 0.370700359f, -1.f),
    C_Vertex3f(0.345953941f, 0.359997749f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268743515f, 0.40625f, -1.f),
    C_Vertex3f(0.283460617f, 0.404475212f, -1.f),
    C_Vertex3f(0.287979126f, 0.474609375f, -1.f),
    C_Vertex3f(0.297166824f, 0.400925636f, -1.f),
    C_Vertex3f(0.308097839f, 0.472724915f, -1.f),
    C_Vertex3f(0.309862137f, 0.395601273f, -1.f),
    C_Vertex3f(0.32736969f, 0.468955994f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230774164f, 0.403260589f, -1.f),
    C_Vertex3f(0.24247241f, 0.40525353f, -1.f),
    C_Vertex3f(0.242723465f, 0.47222662f, -1.f),
    C_Vertex3f(0.254600525f, 0.40625f, -1.f),
    C_Vertex3f(0.265432358f, 0.474609375f, -1.f),
    C_Vertex3f(0.268743515f, 0.40625f, -1.f),
    C_Vertex3f(0.287979126f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.22143364f, 0.467461109f, -1.f),
    C_Vertex3f(0.201562881f, 0.460312843f, -1.f),
    C_Vertex3f(0.208667278f, 0.396285295f, -1.f),
    C_Vertex3f(0.219505787f, 0.400271177f, -1.f),
    C_Vertex3f(0.230774164f, 0.403260589f, -1.f),
    C_Vertex3f(0.242723465f, 0.47222662f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.169611931f, 0.370377064f, -1.f),
    C_Vertex3f(0.178730965f, 0.378348827f, -1.f),
    C_Vertex3f(0.183111191f, 0.450781822f, -1.f),
    C_Vertex3f(0.188279867f, 0.385324121f, -1.f),
    C_Vertex3f(0.201562881f, 0.460312843f, -1.f),
    C_Vertex3f(0.198258638f, 0.391302943f, -1.f),
    C_Vertex3f(0.208667278f, 0.396285295f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.166078568f, 0.438868046f, -1.f),
    C_Vertex3f(0.150465012f, 0.424571514f, -1.f),
    C_Vertex3f(0.154793501f, 0.349068761f, -1.f),
    C_Vertex3f(0.161632776f, 0.360617042f, -1.f),
    C_Vertex3f(0.169611931f, 0.370377064f, -1.f),
    C_Vertex3f(0.183111191f, 0.450781822f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150465012f, 0.424571514f, -1.f),
    C_Vertex3f(0.136270523f, 0.407892227f, -1.f),
    C_Vertex3f(0.138835192f, 0.284993052f, -1.f),
    C_Vertex3f(0.14111495f, 0.303694367f, -1.f),
    C_Vertex3f(0.144534588f, 0.320607424f, -1.f),
    C_Vertex3f(0.149094105f, 0.335732222f, -1.f),
    C_Vertex3f(0.154793501f, 0.349068761f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129882813f, 0.464355469f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.129882813f, 0.398361206f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.136270523f, 0.407892227f, -1.f),
    C_Vertex3f(0.137695313f, 0.264503479f, -1.f),
    C_Vertex3f(0.138835192f, 0.284993052f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.436035156f, 0.f, -1.f),
    C_Vertex3f(0.436035156f, 0.292205811f, -1.f),
    C_Vertex3f(0.435898781f, 0.304838181f, -1.f),
    C_Vertex3f(0.43562603f, 0.316446304f, -1.f),
    C_Vertex3f(0.435216904f, 0.327030182f, -1.f),
    C_Vertex3f(0.434671402f, 0.336589813f, -1.f),
    C_Vertex3f(0.433989525f, 0.345125198f, -1.f),
    C_Vertex3f(0.433171272f, 0.352636337f, -1.f),
    C_Vertex3f(0.432216644f, 0.35912323f, -1.f),
    C_Vertex3f(0.429924011f, 0.370267868f, -1.f),
    C_Vertex3f(0.4251194f, 0.385947227f, -1.f),
    C_Vertex3f(0.41900444f, 0.400479317f, -1.f),
    C_Vertex3f(0.411579132f, 0.413864136f, -1.f),
    C_Vertex3f(0.402843475f, 0.426101685f, -1.f),
    C_Vertex3f(0.391897202f, 0.437164307f, -1.f),
    C_Vertex3f(0.378740311f, 0.447052002f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.436035156f, 0.f, -1.f),
    C_Vertex3f(0.357421875f, 0.294200897f, -1.f),
    C_Vertex3f(0.357421875f, 0.f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.0590820313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_111[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_Normal3f(-0.999325812f, 0.0367154181f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Normal3f(-0.993402183f, 0.114683464f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Normal3f(-0.979182839f, 0.20298022f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Normal3f(-0.953921318f, 0.300056934f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Normal3f(-0.914828122f, 0.403843552f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Normal3f(-0.859796643f, 0.510636568f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Normal3f(-0.788298845f, 0.615292549f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Normal3f(-0.703678906f, 0.710518122f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Normal3f(-0.620272636f, 0.784386277f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Normal3f(-0.543085396f, 0.839677453f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Normal3f(-0.461994827f, 0.886882663f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Normal3f(-0.376675963f, 0.92634511f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Normal3f(-0.289430827f, 0.957198918f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Normal3f(-0.202602044f, 0.97926122f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Normal3f(-0.118287019f, 0.992979467f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Normal3f(-0.0370853581f, 0.999312103f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Normal3f(0.0429813564f, 0.99907589f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Normal3f(0.130778968f, 0.991411507f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Normal3f(0.22524552f, 0.974301994f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Normal3f(0.322965264f, 0.946410835f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Normal3f(0.420876533f, 0.907118022f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Normal3f(0.515538931f, 0.856866181f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Normal3f(0.603679597f, 0.797226965f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Normal3f(0.682734966f, 0.730666041f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Normal3f(0.755707204f, 0.65490967f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Normal3f(0.823695898f, 0.56703198f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Normal3f(0.880775928f, 0.473533183f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Normal3f(0.925762534f, 0.378105432f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Normal3f(0.958753943f, 0.284237236f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Normal3f(0.980860114f, 0.194713563f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Normal3f(0.993775964f, 0.111397475f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Normal3f(0.999312639f, 0.0370702483f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Normal3f(0.999835014f, -0.0181650277f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_Normal3f(0.99813658f, -0.0610187873f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Normal3f(0.994201541f, -0.107533179f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Normal3f(0.987230599f, -0.159297675f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Normal3f(0.976227641f, -0.216747552f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Normal3f(0.95995146f, -0.280166566f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Normal3f(0.936915934f, -0.349554807f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Normal3f(0.90544951f, -0.424453944f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Normal3f(0.869160771f, -0.494529665f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Normal3f(0.831279218f, -0.555855095f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Normal3f(0.787997842f, -0.615678072f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Normal3f(0.739750624f, -0.672881126f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Normal3f(0.687237859f, -0.726432562f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Normal3f(0.631377041f, -0.775475979f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Normal3f(0.573225498f, -0.819397628f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Normal3f(0.507929087f, -0.861398876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Normal3f(0.419512093f, -0.907749832f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Normal3f(0.30787161f, -0.951427937f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Normal3f(0.183039665f, -0.9831056f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Normal3f(0.0626576319f, -0.998035133f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Normal3f(-0.038599968f, -0.999254763f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Normal3f(-0.128493756f, -0.991710246f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Normal3f(-0.222021356f, -0.975041866f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Normal3f(-0.319475144f, -0.947594643f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Normal3f(-0.417872697f, -0.908505559f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Normal3f(-0.513709545f, -0.857964098f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Normal3f(-0.60350287f, -0.797360897f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Normal3f(-0.684390008f, -0.729116023f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Normal3f(-0.760332525f, -0.649533927f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Normal3f(-0.830618083f, -0.556842566f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Normal3f(-0.887843728f, -0.460145056f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Normal3f(-0.931549191f, -0.363615453f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Normal3f(-0.962654054f, -0.270734549f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Normal3f(-0.982942224f, -0.183914542f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Normal3f(-0.994526625f, -0.104483135f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Normal3f(0.999606252f, 0.0280574113f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Normal3f(0.995970607f, 0.0896810889f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Normal3f(0.987246752f, 0.159197494f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Normal3f(0.971514463f, 0.236980483f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Normal3f(0.946460783f, 0.322818667f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Normal3f(0.909567118f, 0.415557086f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Normal3f(0.858534515f, 0.51275593f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Normal3f(0.791965246f, 0.610566258f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Normal3f(0.720430911f, 0.693526745f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Normal3f(0.646434367f, 0.762969673f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Normal3f(0.5606305f, 0.828066111f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Normal3f(0.464660376f, 0.885488927f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Normal3f(0.361490935f, 0.932375669f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Normal3f(0.25501439f, 0.966937244f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Normal3f(0.149346635f, 0.988784969f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Normal3f(0.0481923707f, 0.998838007f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0485911854f, 0.998818815f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Normal3f(-0.151025206f, 0.988529861f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Normal3f(-0.257574052f, 0.966258526f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Normal3f(-0.36460194f, 0.93116349f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Normal3f(-0.467931896f, 0.883764625f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Normal3f(-0.563691318f, 0.825985491f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Normal3f(-0.649005592f, 0.760783732f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Normal3f(-0.72236073f, 0.69151628f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Normal3f(-0.794082761f, 0.607809663f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Normal3f(-0.861015081f, 0.508579373f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_Normal3f(-0.911770999f, 0.410699069f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Normal3f(-0.948089004f, 0.318005204f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Normal3f(-0.972527146f, 0.232789084f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Normal3f(-0.987754703f, 0.156014875f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Normal3f(-0.99614507f, 0.0877208859f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Normal3f(-0.999613106f, 0.0278154202f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Normal3f(-0.999587059f, -0.0287333243f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Normal3f(-0.995663822f, -0.0930251554f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Normal3f(-0.986366749f, -0.164561793f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Normal3f(-0.969791949f, -0.243933469f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Normal3f(-0.943758905f, -0.330634564f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Normal3f(-0.906032026f, -0.423209161f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Normal3f(-0.854762733f, -0.519018829f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Normal3f(-0.789101183f, -0.614263177f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Normal3f(-0.719142795f, -0.694862306f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Normal3f(-0.64632684f, -0.763060749f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Normal3f(-0.561695635f, -0.827343881f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Normal3f(-0.466670901f, -0.884431064f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Normal3f(-0.364006966f, -0.931396246f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Normal3f(-0.257463753f, -0.966287971f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Normal3f(-0.151149184f, -0.988511026f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Normal3f(-0.0485963188f, -0.998818517f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Normal3f(0.0480433144f, -0.99884522f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Normal3f(0.148618564f, -0.988894641f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Normal3f(0.253825158f, -0.967250109f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Normal3f(0.359922737f, -0.932982147f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Normal3f(0.462841064f, -0.886441231f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Normal3f(0.558709502f, -0.829363465f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Normal3f(0.644552469f, -0.764560103f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Normal3f(0.718696654f, -0.695323706f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Normal3f(0.790661871f, -0.612253129f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Normal3f(0.857762754f, -0.514045715f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Normal3f(0.909156442f, -0.416454464f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Normal3f(0.946268499f, -0.32338205f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Normal3f(0.971438348f, -0.237291992f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Normal3f(0.987223685f, -0.159340307f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Normal3f(0.995966375f, -0.0897267833f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_112[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, 0.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.130859375f, 0.464355469f, -1.f),
    C_Vertex3f(0.130859375f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.130859375f, 0.464355469f, 0.f),
    C_Vertex3f(0.130859375f, 0.464355469f, -1.f),
    C_Vertex3f(0.130859375f, 0.403907776f, -1.f),
    C_Vertex3f(0.130859375f, 0.403907776f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.812902927f, 0.58239913f, 0.f),
    C_Vertex3f(0.130859375f, 0.403907776f, 0.f),
    C_Vertex3f(0.130859375f, 0.403907776f, -1.f),
    C_Normal3f(-0.775052965f, 0.631896317f, 0.f),
    C_Vertex3f(0.137191772f, 0.412746429f, 0.f),
    C_Vertex3f(0.137191772f, 0.412746429f, -1.f),
    C_Normal3f(-0.718732178f, 0.695287108f, 0.f),
    C_Vertex3f(0.150675774f, 0.428213596f, 0.f),
    C_Vertex3f(0.150675774f, 0.428213596f, -1.f),
    C_Normal3f(-0.636815846f, 0.771015823f, 0.f),
    C_Vertex3f(0.164978981f, 0.441470623f, 0.f),
    C_Vertex3f(0.164978981f, 0.441470623f, -1.f),
    C_Normal3f(-0.53910464f, 0.842238665f, 0.f),
    C_Vertex3f(0.180101395f, 0.452517509f, 0.f),
    C_Vertex3f(0.180101395f, 0.452517509f, -1.f),
    C_Normal3f(-0.418337375f, 0.908291698f, 0.f),
    C_Vertex3f(0.196043015f, 0.461354256f, 0.f),
    C_Vertex3f(0.196043015f, 0.461354256f, -1.f),
    C_Normal3f(-0.286336184f, 0.958129168f, 0.f),
    C_Vertex3f(0.213677406f, 0.467981815f, 0.f),
    C_Vertex3f(0.213677406f, 0.467981815f, -1.f),
    C_Normal3f(-0.16209133f, 0.986775756f, 0.f),
    C_Vertex3f(0.23300457f, 0.472400188f, 0.f),
    C_Vertex3f(0.23300457f, 0.472400188f, -1.f),
    C_Normal3f(-0.055420991f, 0.998463035f, 0.f),
    C_Vertex3f(0.254024506f, 0.474609375f, 0.f),
    C_Vertex3f(0.254024506f, 0.474609375f, -1.f),
    C_Normal3f(0.0287769381f, 0.999585807f, 0.f),
    C_Vertex3f(0.272805214f, 0.474609375f, 0.f),
    C_Vertex3f(0.272805214f, 0.474609375f, -1.f),
    C_Normal3f(0.100475393f, 0.994939566f, 0.f),
    C_Vertex3f(0.287217259f, 0.473653793f, 0.f),
    C_Vertex3f(0.287217259f, 0.473653793f, -1.f),
    C_Normal3f(0.171092555f, 0.985254884f, 0.f),
    C_Vertex3f(0.301192641f, 0.47174263f, 0.f),
    C_Vertex3f(0.301192641f, 0.47174263f, -1.f),
    C_Normal3f(0.243525013f, 0.969894588f, 0.f),
    C_Vertex3f(0.314731359f, 0.468875885f, 0.f),
    C_Vertex3f(0.314731359f, 0.468875885f, -1.f),
    C_Normal3f(0.316595078f, 0.948560834f, 0.f),
    C_Vertex3f(0.327833414f, 0.465053558f, 0.f),
    C_Vertex3f(0.327833414f, 0.465053558f, -1.f),
    C_Normal3f(0.38898927f, 0.921242297f, 0.f),
    C_Vertex3f(0.340498805f, 0.46027565f, 0.f),
    C_Vertex3f(0.340498805f, 0.46027565f, -1.f),
    C_Normal3f(0.45936501f, 0.888247609f, 0.f),
    C_Vertex3f(0.352727532f, 0.45454216f, 0.f),
    C_Vertex3f(0.352727532f, 0.45454216f, -1.f),
    C_Normal3f(0.526469886f, 0.850193679f, 0.f),
    C_Vertex3f(0.364519596f, 0.447853088f, 0.f),
    C_Vertex3f(0.364519596f, 0.447853088f, -1.f),
    C_Normal3f(0.589690566f, 0.807629287f, 0.f),
    C_Vertex3f(0.375874996f, 0.440208435f, 0.f),
    C_Vertex3f(0.375874996f, 0.440208435f, -1.f),
    C_Normal3f(0.650091171f, 0.759856284f, 0.f),
    C_Vertex3f(0.386534333f, 0.431779027f, 0.f),
    C_Vertex3f(0.386534333f, 0.431779027f, -1.f),
    C_Normal3f(0.706786573f, 0.707426846f, 0.f),
    C_Vertex3f(0.396497607f, 0.422564864f, 0.f),
    C_Vertex3f(0.396497607f, 0.422564864f, -1.f),
    C_Normal3f(0.758810639f, 0.651311338f, 0.f),
    C_Vertex3f(0.405764818f, 0.412565947f, 0.f),
    C_Vertex3f(0.405764818f, 0.412565947f, -1.f),
    C_Normal3f(0.805464149f, 0.592644572f, 0.f),
    C_Vertex3f(0.414335966f, 0.401782274f, 0.f),
    C_Vertex3f(0.414335966f, 0.401782274f, -1.f),
    C_Normal3f(0.846354783f, 0.532619476f, 0.f),
    C_Vertex3f(0.422211051f, 0.390213847f, 0.f),
    C_Vertex3f(0.422211051f, 0.390213847f, -1.f),
    C_Normal3f(0.881393254f, 0.472383261f, 0.f),
    C_Vertex3f(0.429390073f, 0.377860665f, 0.f),
    C_Vertex3f(0.429390073f, 0.377860665f, -1.f),
    C_Normal3f(0.913388312f, 0.407089353f, 0.f),
    C_Vertex3f(0.435873032f, 0.364722729f, 0.f),
    C_Vertex3f(0.435873032f, 0.364722729f, -1.f),
    C_Normal3f(0.944813907f, 0.327607483f, 0.f),
    C_Vertex3f(0.444553375f, 0.343838692f, 0.f),
    C_Vertex3f(0.444553375f, 0.343838692f, -1.f),
    C_Normal3f(0.971876144f, 0.235492513f, 0.f),
    C_Vertex3f(0.453233719f, 0.314654827f, 0.f),
    C_Vertex3f(0.453233719f, 0.314654827f, -1.f),
    C_Normal3f(0.990457535f, 0.1378178f, 0.f),
    C_Vertex3f(0.459020615f, 0.284132481f, 0.f),
    C_Vertex3f(0.459020615f, 0.284132481f, -1.f),
    C_Normal3f(0.999047935f, 0.0436266102f, 0.f),
    C_Vertex3f(0.461914063f, 0.252271652f, 0.f),
    C_Vertex3f(0.461914063f, 0.252271652f, -1.f),
    C_Normal3f(0.998904943f, -0.0467839614f, 0.f),
    C_Vertex3f(0.461914063f, 0.21787262f, 0.f),
    C_Vertex3f(0.461914063f, 0.21787262f, -1.f),
    C_Normal3f(0.989578068f, -0.143997103f, 0.f),
    C_Vertex3f(0.458718777f, 0.184047699f, 0.f),
    C_Vertex3f(0.458718777f, 0.184047699f, -1.f),
    C_Normal3f(0.968683243f, -0.248299569f, 0.f),
    C_Vertex3f(0.452328205f, 0.151996613f, 0.f),
    C_Vertex3f(0.452328205f, 0.151996613f, -1.f),
    C_Normal3f(0.937510371f, -0.347957522f, 0.f),
    C_Vertex3f(0.442742348f, 0.12171936f, 0.f),
    C_Vertex3f(0.442742348f, 0.12171936f, -1.f),
    C_Normal3f(0.901544869f, -0.432685554f, 0.f),
    C_Vertex3f(0.43315649f, 0.100341797f, 0.f),
    C_Vertex3f(0.43315649f, 0.100341797f, -1.f),
    C_Normal3f(0.866914868f, -0.498456359f, 0.f),
    C_Vertex3f(0.426042199f, 0.086922884f, 0.f),
    C_Vertex3f(0.426042199f, 0.086922884f, -1.f),
    C_Normal3f(0.829312563f, -0.558784962f, 0.f),
    C_Vertex3f(0.418204188f, 0.0743367672f, 0.f),
    C_Vertex3f(0.418204188f, 0.0743367672f, -1.f),
    C_Normal3f(0.785779059f, -0.618507385f, 0.f),
    C_Vertex3f(0.409642458f, 0.0625834465f, 0.f),
    C_Vertex3f(0.409642458f, 0.0625834465f, -1.f),
    C_Normal3f(0.736506641f, -0.676430166f, 0.f),
    C_Vertex3f(0.400357008f, 0.0516629219f, 0.f),
    C_Vertex3f(0.400357008f, 0.0516629219f, -1.f),
    C_Normal3f(0.682005465f, -0.731347024f, 0.f),
    C_Vertex3f(0.390347838f, 0.0415751934f, 0.f),
    C_Vertex3f(0.390347838f, 0.0415751934f, -1.f),
    C_Normal3f(0.623091877f, -0.782148659f, 0.f),
    C_Vertex3f(0.379614949f, 0.032320261f, 0.f),
    C_Vertex3f(0.379614949f, 0.032320261f, -1.f),
    C_Normal3f(0.554469109f, -0.832204282f, 0.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, 0.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, -1.f),
    C_Normal3f(0.465751261f, -0.884915709f, 0.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, 0.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, -1.f),
    C_Normal3f(0.351073146f, -0.936347961f, 0.f),
    C_Vertex3f(0.324899673f, 0.00113010406f, 0.f),
    C_Vertex3f(0.324899673f, 0.00113010406f, -1.f),
    C_Normal3f(0.214413568f, -0.976742983f, 0.f),
    C_Vertex3f(0.299283981f, -0.00645923615f, 0.f),
    C_Vertex3f(0.299283981f, -0.00645923615f, -1.f),
    C_Normal3f(0.0764189214f, -0.997075856f, 0.f),
    C_Vertex3f(0.273040771f, -0.0102539063f, 0.f),
    C_Vertex3f(0.273040771f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0492831543f, -0.99878484f, 0.f),
    C_Vertex3f(0.249773026f, -0.0102539063f, 0.f),
    C_Vertex3f(0.249773026f, -0.0102539063f, -1.f),
    C_Normal3f(-0.168528154f, -0.985696793f, 0.f),
    C_Vertex3f(0.231119633f, -0.00818538666f, 0.f),
    C_Vertex3f(0.231119633f, -0.00818538666f, -1.f),
    C_Normal3f(-0.288905561f, -0.957357645f, 0.f),
    C_Vertex3f(0.213477612f, -0.00404834747f, 0.f),
    C_Vertex3f(0.213477612f, -0.00404834747f, -1.f),
    C_Normal3f(-0.409592241f, -0.912268758f, 0.f),
    C_Vertex3f(0.196846962f, 0.0021572113f, 0.f),
    C_Vertex3f(0.196846962f, 0.0021572113f, -1.f),
    C_Normal3f(-0.508741796f, -0.860919118f, 0.f),
    C_Vertex3f(0.181227684f, 0.0104312897f, 0.f),
    C_Vertex3f(0.181227684f, 0.0104312897f, -1.f),
    C_Normal3f(-0.590768754f, -0.806840956f, 0.f),
    C_Vertex3f(0.167000294f, 0.0197944641f, 0.f),
    C_Vertex3f(0.167000294f, 0.0197944641f, -1.f),
    C_Normal3f(-0.671363354f, -0.741128445f, 0.f),
    C_Vertex3f(0.154164791f, 0.0302467346f, 0.f),
    C_Vertex3f(0.154164791f, 0.0302467346f, -1.f),
    C_Normal3f(-0.735081077f, -0.677979231f, 0.f),
    C_Vertex3f(0.142721176f, 0.0417881012f, 0.f),
    C_Vertex3f(0.142721176f, 0.0417881012f, -1.f),
    C_Normal3f(-0.782457054f, -0.622704446f, 0.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, 0.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, 0.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, -1.f),
    C_Vertex3f(0.137695313f, -0.177734375f, -1.f),
    C_Vertex3f(0.137695313f, -0.177734375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.137695313f, -0.177734375f, 0.f),
    C_Vertex3f(0.137695313f, -0.177734375f, -1.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, -1.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999616206f, -0.0277042985f, 0.f),
    C_Vertex3f(0.130371094f, 0.240693092f, 0.f),
    C_Vertex3f(0.130371094f, 0.240693092f, -1.f),
    C_Normal3f(0.999660492f, 0.0260581132f, 0.f),
    C_Vertex3f(0.130371094f, 0.218290329f, 0.f),
    C_Vertex3f(0.130371094f, 0.218290329f, -1.f),
    C_Normal3f(0.996502697f, 0.0835615695f, 0.f),
    C_Vertex3f(0.131500959f, 0.197348356f, 0.f),
    C_Vertex3f(0.131500959f, 0.197348356f, -1.f),
    C_Normal3f(0.988841295f, 0.148972824f, 0.f),
    C_Vertex3f(0.133760691f, 0.177868128f, 0.f),
    C_Vertex3f(0.133760691f, 0.177868128f, -1.f),
    C_Normal3f(0.974820793f, 0.222989604f, 0.f),
    C_Vertex3f(0.137150288f, 0.159849644f, 0.f),
    C_Vertex3f(0.137150288f, 0.159849644f, -1.f),
    C_Normal3f(0.952072144f, 0.305873334f, 0.f),
    C_Vertex3f(0.14166975f, 0.143292904f, 0.f),
    C_Vertex3f(0.14166975f, 0.143292904f, -1.f),
    C_Normal3f(0.917793036f, 0.397059083f, 0.f),
    C_Vertex3f(0.147319078f, 0.128197908f, 0.f),
    C_Vertex3f(0.147319078f, 0.128197908f, -1.f),
    C_Normal3f(0.86907357f, 0.494682908f, 0.f),
    C_Vertex3f(0.154098272f, 0.114564657f, 0.f),
    C_Vertex3f(0.154098272f, 0.114564657f, -1.f),
    C_Normal3f(0.803575754f, 0.595202565f, 0.f),
    C_Vertex3f(0.162007332f, 0.10239315f, 0.f),
    C_Vertex3f(0.162007332f, 0.10239315f, -1.f),
    C_Normal3f(0.734469414f, 0.678641856f, 0.f),
    C_Vertex3f(0.171046257f, 0.0916833878f, 0.f),
    C_Vertex3f(0.171046257f, 0.0916833878f, -1.f),
    C_Normal3f(0.665713906f, 0.746206999f, 0.f),
    C_Vertex3f(0.180561781f, 0.0823123455f, 0.f),
    C_Vertex3f(0.180561781f, 0.0823123455f, -1.f),
    C_Normal3f(0.58415103f, 0.811645031f, 0.f),
    C_Vertex3f(0.190553904f, 0.0742800236f, 0.f),
    C_Vertex3f(0.190553904f, 0.0742800236f, -1.f),
    C_Normal3f(0.490356565f, 0.871521831f, 0.f),
    C_Vertex3f(0.201022625f, 0.067586422f, 0.f),
    C_Vertex3f(0.201022625f, 0.067586422f, -1.f),
    C_Normal3f(0.386419147f, 0.922323167f, 0.f),
    C_Vertex3f(0.211967945f, 0.0622315407f, 0.f),
    C_Vertex3f(0.211967945f, 0.0622315407f, -1.f),
    C_Normal3f(0.275887758f, 0.961189985f, 0.f),
    C_Vertex3f(0.223389864f, 0.0582153797f, 0.f),
    C_Vertex3f(0.223389864f, 0.0582153797f, -1.f),
    C_Normal3f(0.163234457f, 0.986587346f, 0.f),
    C_Vertex3f(0.235288382f, 0.0555379391f, 0.f),
    C_Vertex3f(0.235288382f, 0.0555379391f, -1.f),
    C_Normal3f(0.0527659878f, 0.99860692f, 0.f),
    C_Vertex3f(0.247663498f, 0.0541992188f, 0.f),
    C_Vertex3f(0.247663498f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0541663952f, 0.998531938f, 0.f),
    C_Vertex3f(0.260623932f, 0.0541992188f, 0.f),
    C_Vertex3f(0.260623932f, 0.0541992188f, -1.f),
    C_Normal3f(-0.165874287f, 0.986146927f, 0.f),
    C_Vertex3f(0.2732234f, 0.055585742f, 0.f),
    C_Vertex3f(0.2732234f, 0.055585742f, -1.f),
    C_Normal3f(-0.279773116f, 0.96006614f, 0.f),
    C_Vertex3f(0.285353184f, 0.0583587885f, 0.f),
    C_Vertex3f(0.285353184f, 0.0583587885f, -1.f),
    C_Normal3f(-0.390942574f, 0.920415103f, 0.f),
    C_Vertex3f(0.297013283f, 0.0625183582f, 0.f),
    C_Vertex3f(0.297013283f, 0.0625183582f, -1.f),
    C_Normal3f(-0.494884431f, 0.868958771f, 0.f),
    C_Vertex3f(0.308203697f, 0.0680644512f, 0.f),
    C_Vertex3f(0.308203697f, 0.0680644512f, -1.f),
    C_Normal3f(-0.588160515f, 0.808744252f, 0.f),
    C_Vertex3f(0.318924427f, 0.0749970675f, 0.f),
    C_Vertex3f(0.318924427f, 0.0749970675f, -1.f),
    C_Normal3f(-0.668875277f, 0.743374646f, 0.f),
    C_Vertex3f(0.329175472f, 0.0833162069f, 0.f),
    C_Vertex3f(0.329175472f, 0.0833162069f, -1.f),
    C_Normal3f(-0.736653864f, 0.676269948f, 0.f),
    C_Vertex3f(0.338956833f, 0.0930218697f, 0.f),
    C_Vertex3f(0.338956833f, 0.0930218697f, -1.f),
    C_Normal3f(-0.805206656f, 0.592994332f, 0.f),
    C_Vertex3f(0.348268509f, 0.104114056f, 0.f),
    C_Vertex3f(0.348268509f, 0.104114056f, -1.f),
    C_Normal3f(-0.870426655f, 0.492298156f, 0.f),
    C_Vertex3f(0.356416225f, 0.116729379f, 0.f),
    C_Vertex3f(0.356416225f, 0.116729379f, -1.f),
    C_Normal3f(-0.918786287f, 0.394755304f, 0.f),
    C_Vertex3f(0.363399982f, 0.130867839f, 0.f),
    C_Vertex3f(0.363399982f, 0.130867839f, -1.f),
    C_Normal3f(-0.952719927f, 0.303849667f, 0.f),
    C_Vertex3f(0.36921978f, 0.146529436f, 0.f),
    C_Vertex3f(0.36921978f, 0.146529436f, -1.f),
    C_Normal3f(-0.975189686f, 0.221370652f, 0.f),
    C_Vertex3f(0.373875618f, 0.16371417f, 0.f),
    C_Vertex3f(0.373875618f, 0.16371417f, -1.f),
    C_Normal3f(-0.989014626f, 0.14781794f, 0.f),
    C_Vertex3f(0.377367496f, 0.182422042f, 0.f),
    C_Vertex3f(0.377367496f, 0.182422042f, -1.f),
    C_Normal3f(-0.996559322f, 0.0828832015f, 0.f),
    C_Vertex3f(0.379695415f, 0.20265305f, 0.f),
    C_Vertex3f(0.379695415f, 0.20265305f, -1.f),
    C_Normal3f(-0.99965781f, 0.0261562765f, 0.f),
    C_Vertex3f(0.380859375f, 0.224407196f, 0.f),
    C_Vertex3f(0.380859375f, 0.224407196f, -1.f),
    C_Normal3f(-0.999659002f, -0.0261108223f, 0.f),
    C_Vertex3f(0.380859375f, 0.247138023f, 0.f),
    C_Vertex3f(0.380859375f, 0.247138023f, -1.f),
    C_Normal3f(-0.996418953f, -0.0845540985f, 0.f),
    C_Vertex3f(0.379722595f, 0.267929077f, 0.f),
    C_Vertex3f(0.379722595f, 0.267929077f, -1.f),
    C_Normal3f(-0.988649189f, -0.150242209f, 0.f),
    C_Vertex3f(0.377449036f, 0.287326813f, 0.f),
    C_Vertex3f(0.377449036f, 0.287326813f, -1.f),
    C_Normal3f(-0.974580407f, -0.224037766f, 0.f),
    C_Vertex3f(0.374038696f, 0.30533123f, 0.f),
    C_Vertex3f(0.374038696f, 0.30533123f, -1.f),
    C_Normal3f(-0.952025831f, -0.306017488f, 0.f),
    C_Vertex3f(0.369491577f, 0.321942329f, 0.f),
    C_Vertex3f(0.369491577f, 0.321942329f, -1.f),
    C_Normal3f(-0.918480754f, -0.395465791f, 0.f),
    C_Vertex3f(0.363807678f, 0.33716011f, 0.f),
    C_Vertex3f(0.363807678f, 0.33716011f, -1.f),
    C_Normal3f(-0.871442795f, -0.490497172f, 0.f),
    C_Vertex3f(0.356987f, 0.350984573f, 0.f),
    C_Vertex3f(0.356987f, 0.350984573f, -1.f),
    C_Normal3f(-0.809008956f, -0.58779633f, 0.f),
    C_Vertex3f(0.349029541f, 0.363415718f, 0.f),
    C_Vertex3f(0.349029541f, 0.363415718f, -1.f),
    C_Normal3f(-0.743215382f, -0.669052243f, 0.f),
    C_Vertex3f(0.339935303f, 0.374453545f, 0.f),
    C_Vertex3f(0.339935303f, 0.374453545f, -1.f),
    C_Normal3f(-0.676697135f, -0.736261547f, 0.f),
    C_Vertex3f(0.330398798f, 0.384111643f, 0.f),
    C_Vertex3f(0.330398798f, 0.384111643f, -1.f),
    C_Normal3f(-0.596903384f, -0.802313089f, 0.f),
    C_Vertex3f(0.320420027f, 0.392390013f, 0.f),
    C_Vertex3f(0.320420027f, 0.392390013f, -1.f),
    C_Normal3f(-0.503922164f, -0.863749087f, 0.f),
    C_Vertex3f(0.309998989f, 0.399288654f, 0.f),
    C_Vertex3f(0.309998989f, 0.399288654f, -1.f),
    C_Normal3f(-0.399402738f, -0.916775644f, 0.f),
    C_Vertex3f(0.299135685f, 0.404807568f, 0.f),
    C_Vertex3f(0.299135685f, 0.404807568f, -1.f),
    C_Normal3f(-0.286685169f, -0.9580248f, 0.f),
    C_Vertex3f(0.287830114f, 0.408946753f, 0.f),
    C_Vertex3f(0.287830114f, 0.408946753f, -1.f),
    C_Normal3f(-0.17038475f, -0.98537761f, 0.f),
    C_Vertex3f(0.276082277f, 0.411706209f, 0.f),
    C_Vertex3f(0.276082277f, 0.411706209f, -1.f),
    C_Normal3f(-0.0556195974f, -0.998452008f, 0.f),
    C_Vertex3f(0.263892174f, 0.413085938f, 0.f),
    C_Vertex3f(0.263892174f, 0.413085938f, -1.f),
    C_Normal3f(0.059252087f, -0.998243034f, 0.f),
    C_Vertex3f(0.251314163f, 0.413085938f, 0.f),
    C_Vertex3f(0.251314163f, 0.413085938f, -1.f),
    C_Normal3f(0.180802196f, -0.983519554f, 0.f),
    C_Vertex3f(0.23915112f, 0.411617398f, 0.f),
    C_Vertex3f(0.23915112f, 0.411617398f, -1.f),
    C_Normal3f(0.301221013f, -0.953554392f, 0.f),
    C_Vertex3f(0.227348685f, 0.40868032f, 0.f),
    C_Vertex3f(0.227348685f, 0.40868032f, -1.f),
    C_Normal3f(0.415209264f, -0.909725904f, 0.f),
    C_Vertex3f(0.215906858f, 0.404274702f, 0.f),
    C_Vertex3f(0.215906858f, 0.404274702f, -1.f),
    C_Normal3f(0.518405676f, -0.855134904f, 0.f),
    C_Vertex3f(0.20482564f, 0.398400545f, 0.f),
    C_Vertex3f(0.20482564f, 0.398400545f, -1.f),
    C_Normal3f(0.608244836f, -0.793749452f, 0.f),
    C_Vertex3f(0.194105029f, 0.391057849f, 0.f),
    C_Vertex3f(0.194105029f, 0.391057849f, -1.f),
    C_Normal3f(0.68401885f, -0.729464352f, 0.f),
    C_Vertex3f(0.183745027f, 0.382246614f, 0.f),
    C_Vertex3f(0.183745027f, 0.382246614f, -1.f),
    C_Normal3f(0.746438384f, -0.665454566f, 0.f),
    C_Vertex3f(0.173745632f, 0.371966839f, 0.f),
    C_Vertex3f(0.173745632f, 0.371966839f, -1.f),
    C_Normal3f(0.808491766f, -0.588507473f, 0.f),
    C_Vertex3f(0.164106846f, 0.360218525f, 0.f),
    C_Vertex3f(0.164106846f, 0.360218525f, -1.f),
    C_Normal3f(0.868875623f, -0.495030463f, 0.f),
    C_Vertex3f(0.155672908f, 0.347138524f, 0.f),
    C_Vertex3f(0.155672908f, 0.347138524f, -1.f),
    C_Normal3f(0.915481687f, -0.402359694f, 0.f),
    C_Vertex3f(0.148443818f, 0.332726836f, 0.f),
    C_Vertex3f(0.148443818f, 0.332726836f, -1.f),
    C_Normal3f(0.949511886f, -0.313730955f, 0.f),
    C_Vertex3f(0.142419577f, 0.316983461f, 0.f),
    C_Vertex3f(0.142419577f, 0.316983461f, -1.f),
    C_Normal3f(0.972893298f, -0.231254324f, 0.f),
    C_Vertex3f(0.137600183f, 0.2999084f, 0.f),
    C_Vertex3f(0.137600183f, 0.2999084f, -1.f),
    C_Normal3f(0.987757683f, -0.155995861f, 0.f),
    C_Vertex3f(0.133985639f, 0.281501651f, 0.f),
    C_Vertex3f(0.133985639f, 0.281501651f, -1.f),
    C_Normal3f(0.996100307f, -0.088227801f, 0.f),
    C_Vertex3f(0.131575942f, 0.261763215f, 0.f),
    C_Vertex3f(0.131575942f, 0.261763215f, -1.f),
    C_Normal3f(0.999616206f, -0.0277042985f, 0.f),
    C_Vertex3f(0.130371094f, 0.240693092f, 0.f),
    C_Vertex3f(0.130371094f, 0.240693092f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133760691f, 0.177868128f, 0.f),
    C_Vertex3f(0.131500959f, 0.197348356f, 0.f),
    C_Vertex3f(0.137695313f, -0.177734375f, 0.f),
    C_Vertex3f(0.130371094f, 0.218290329f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Vertex3f(0.130371094f, 0.240693092f, 0.f),
    C_Vertex3f(0.130859375f, 0.464355469f, 0.f),
    C_Vertex3f(0.130859375f, 0.403907776f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379722595f, 0.267929077f, 0.f),
    C_Vertex3f(0.380859375f, 0.247138023f, 0.f),
    C_Vertex3f(0.386534333f, 0.431779027f, 0.f),
    C_Vertex3f(0.390347838f, 0.0415751934f, 0.f),
    C_Vertex3f(0.396497607f, 0.422564864f, 0.f),
    C_Vertex3f(0.400357008f, 0.0516629219f, 0.f),
    C_Vertex3f(0.405764818f, 0.412565947f, 0.f),
    C_Vertex3f(0.409642458f, 0.0625834465f, 0.f),
    C_Vertex3f(0.414335966f, 0.401782274f, 0.f),
    C_Vertex3f(0.418204188f, 0.0743367672f, 0.f),
    C_Vertex3f(0.422211051f, 0.390213847f, 0.f),
    C_Vertex3f(0.426042199f, 0.086922884f, 0.f),
    C_Vertex3f(0.429390073f, 0.377860665f, 0.f),
    C_Vertex3f(0.43315649f, 0.100341797f, 0.f),
    C_Vertex3f(0.435873032f, 0.364722729f, 0.f),
    C_Vertex3f(0.442742348f, 0.12171936f, 0.f),
    C_Vertex3f(0.444553375f, 0.343838692f, 0.f),
    C_Vertex3f(0.452328205f, 0.151996613f, 0.f),
    C_Vertex3f(0.453233719f, 0.314654827f, 0.f),
    C_Vertex3f(0.458718777f, 0.184047699f, 0.f),
    C_Vertex3f(0.459020615f, 0.284132481f, 0.f),
    C_Vertex3f(0.461914063f, 0.21787262f, 0.f),
    C_Vertex3f(0.461914063f, 0.252271652f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369491577f, 0.321942329f, 0.f),
    C_Vertex3f(0.374038696f, 0.30533123f, 0.f),
    C_Vertex3f(0.375874996f, 0.440208435f, 0.f),
    C_Vertex3f(0.377449036f, 0.287326813f, 0.f),
    C_Vertex3f(0.386534333f, 0.431779027f, 0.f),
    C_Vertex3f(0.379722595f, 0.267929077f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.364519596f, 0.447853088f, 0.f),
    C_Vertex3f(0.352727532f, 0.45454216f, 0.f),
    C_Vertex3f(0.356987f, 0.350984573f, 0.f),
    C_Vertex3f(0.363807678f, 0.33716011f, 0.f),
    C_Vertex3f(0.369491577f, 0.321942329f, 0.f),
    C_Vertex3f(0.375874996f, 0.440208435f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.330398798f, 0.384111643f, 0.f),
    C_Vertex3f(0.339935303f, 0.374453545f, 0.f),
    C_Vertex3f(0.340498805f, 0.46027565f, 0.f),
    C_Vertex3f(0.349029541f, 0.363415718f, 0.f),
    C_Vertex3f(0.352727532f, 0.45454216f, 0.f),
    C_Vertex3f(0.356987f, 0.350984573f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.287830114f, 0.408946753f, 0.f),
    C_Vertex3f(0.299135685f, 0.404807568f, 0.f),
    C_Vertex3f(0.301192641f, 0.47174263f, 0.f),
    C_Vertex3f(0.309998989f, 0.399288654f, 0.f),
    C_Vertex3f(0.314731359f, 0.468875885f, 0.f),
    C_Vertex3f(0.320420027f, 0.392390013f, 0.f),
    C_Vertex3f(0.327833414f, 0.465053558f, 0.f),
    C_Vertex3f(0.330398798f, 0.384111643f, 0.f),
    C_Vertex3f(0.340498805f, 0.46027565f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.23915112f, 0.411617398f, 0.f),
    C_Vertex3f(0.251314163f, 0.413085938f, 0.f),
    C_Vertex3f(0.254024506f, 0.474609375f, 0.f),
    C_Vertex3f(0.263892174f, 0.413085938f, 0.f),
    C_Vertex3f(0.272805214f, 0.474609375f, 0.f),
    C_Vertex3f(0.276082277f, 0.411706209f, 0.f),
    C_Vertex3f(0.287217259f, 0.473653793f, 0.f),
    C_Vertex3f(0.287830114f, 0.408946753f, 0.f),
    C_Vertex3f(0.301192641f, 0.47174263f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.23300457f, 0.472400188f, 0.f),
    C_Vertex3f(0.213677406f, 0.467981815f, 0.f),
    C_Vertex3f(0.215906858f, 0.404274702f, 0.f),
    C_Vertex3f(0.227348685f, 0.40868032f, 0.f),
    C_Vertex3f(0.23915112f, 0.411617398f, 0.f),
    C_Vertex3f(0.254024506f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.183745027f, 0.382246614f, 0.f),
    C_Vertex3f(0.194105029f, 0.391057849f, 0.f),
    C_Vertex3f(0.196043015f, 0.461354256f, 0.f),
    C_Vertex3f(0.20482564f, 0.398400545f, 0.f),
    C_Vertex3f(0.213677406f, 0.467981815f, 0.f),
    C_Vertex3f(0.215906858f, 0.404274702f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.155672908f, 0.347138524f, 0.f),
    C_Vertex3f(0.164106846f, 0.360218525f, 0.f),
    C_Vertex3f(0.164978981f, 0.441470623f, 0.f),
    C_Vertex3f(0.173745632f, 0.371966839f, 0.f),
    C_Vertex3f(0.180101395f, 0.452517509f, 0.f),
    C_Vertex3f(0.183745027f, 0.382246614f, 0.f),
    C_Vertex3f(0.196043015f, 0.461354256f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150675774f, 0.428213596f, 0.f),
    C_Vertex3f(0.137191772f, 0.412746429f, 0.f),
    C_Vertex3f(0.137600183f, 0.2999084f, 0.f),
    C_Vertex3f(0.142419577f, 0.316983461f, 0.f),
    C_Vertex3f(0.148443818f, 0.332726836f, 0.f),
    C_Vertex3f(0.155672908f, 0.347138524f, 0.f),
    C_Vertex3f(0.164978981f, 0.441470623f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.137191772f, 0.412746429f, 0.f),
    C_Vertex3f(0.130859375f, 0.403907776f, 0.f),
    C_Vertex3f(0.131575942f, 0.261763215f, 0.f),
    C_Vertex3f(0.133985639f, 0.281501651f, 0.f),
    C_Vertex3f(0.137600183f, 0.2999084f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.380859375f, 0.247138023f, 0.f),
    C_Vertex3f(0.380859375f, 0.224407196f, 0.f),
    C_Vertex3f(0.390347838f, 0.0415751934f, 0.f),
    C_Vertex3f(0.379695415f, 0.20265305f, 0.f),
    C_Vertex3f(0.379614949f, 0.032320261f, 0.f),
    C_Vertex3f(0.377367496f, 0.182422042f, 0.f),
    C_Vertex3f(0.373875618f, 0.16371417f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36921978f, 0.146529436f, 0.f),
    C_Vertex3f(0.363399982f, 0.130867839f, 0.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, 0.f),
    C_Vertex3f(0.379614949f, 0.032320261f, 0.f),
    C_Vertex3f(0.373875618f, 0.16371417f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356416225f, 0.116729379f, 0.f),
    C_Vertex3f(0.348268509f, 0.104114056f, 0.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, 0.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, 0.f),
    C_Vertex3f(0.363399982f, 0.130867839f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.348268509f, 0.104114056f, 0.f),
    C_Vertex3f(0.338956833f, 0.0930218697f, 0.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, 0.f),
    C_Vertex3f(0.329175472f, 0.0833162069f, 0.f),
    C_Vertex3f(0.324899673f, 0.00113010406f, 0.f),
    C_Vertex3f(0.318924427f, 0.0749970675f, 0.f),
    C_Vertex3f(0.308203697f, 0.0680644512f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.299283981f, -0.00645923615f, 0.f),
    C_Vertex3f(0.324899673f, 0.00113010406f, 0.f),
    C_Vertex3f(0.308203697f, 0.0680644512f, 0.f),
    C_Vertex3f(0.297013283f, 0.0625183582f, 0.f),
    C_Vertex3f(0.285353184f, 0.0583587885f, 0.f),
    C_Vertex3f(0.2732234f, 0.055585742f, 0.f),
    C_Vertex3f(0.273040771f, -0.0102539063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.260623932f, 0.0541992188f, 0.f),
    C_Vertex3f(0.247663498f, 0.0541992188f, 0.f),
    C_Vertex3f(0.249773026f, -0.0102539063f, 0.f),
    C_Vertex3f(0.273040771f, -0.0102539063f, 0.f),
    C_Vertex3f(0.2732234f, 0.055585742f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.211967945f, 0.0622315407f, 0.f),
    C_Vertex3f(0.213477612f, -0.00404834747f, 0.f),
    C_Vertex3f(0.223389864f, 0.0582153797f, 0.f),
    C_Vertex3f(0.231119633f, -0.00818538666f, 0.f),
    C_Vertex3f(0.235288382f, 0.0555379391f, 0.f),
    C_Vertex3f(0.249773026f, -0.0102539063f, 0.f),
    C_Vertex3f(0.247663498f, 0.0541992188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.180561781f, 0.0823123455f, 0.f),
    C_Vertex3f(0.181227684f, 0.0104312897f, 0.f),
    C_Vertex3f(0.190553904f, 0.0742800236f, 0.f),
    C_Vertex3f(0.196846962f, 0.0021572113f, 0.f),
    C_Vertex3f(0.201022625f, 0.067586422f, 0.f),
    C_Vertex3f(0.213477612f, -0.00404834747f, 0.f),
    C_Vertex3f(0.211967945f, 0.0622315407f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154098272f, 0.114564657f, 0.f),
    C_Vertex3f(0.154164791f, 0.0302467346f, 0.f),
    C_Vertex3f(0.162007332f, 0.10239315f, 0.f),
    C_Vertex3f(0.167000294f, 0.0197944641f, 0.f),
    C_Vertex3f(0.171046257f, 0.0916833878f, 0.f),
    C_Vertex3f(0.181227684f, 0.0104312897f, 0.f),
    C_Vertex3f(0.180561781f, 0.0823123455f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133760691f, 0.177868128f, 0.f),
    C_Vertex3f(0.137695313f, -0.177734375f, 0.f),
    C_Vertex3f(0.137150288f, 0.159849644f, 0.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, 0.f),
    C_Vertex3f(0.14166975f, 0.143292904f, 0.f),
    C_Vertex3f(0.142721176f, 0.0417881012f, 0.f),
    C_Vertex3f(0.147319078f, 0.128197908f, 0.f),
    C_Vertex3f(0.154164791f, 0.0302467346f, 0.f),
    C_Vertex3f(0.154098272f, 0.114564657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.137695313f, -0.177734375f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, 0.f),
    C_Vertex3f(0.131575942f, 0.261763215f, 0.f),
    C_Vertex3f(0.130859375f, 0.403907776f, 0.f),
    C_Vertex3f(0.130371094f, 0.240693092f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133760691f, 0.177868128f, -1.f),
    C_Vertex3f(0.131500959f, 0.197348356f, -1.f),
    C_Vertex3f(0.137695313f, -0.177734375f, -1.f),
    C_Vertex3f(0.130371094f, 0.218290329f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.130371094f, 0.240693092f, -1.f),
    C_Vertex3f(0.130859375f, 0.464355469f, -1.f),
    C_Vertex3f(0.130859375f, 0.403907776f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379722595f, 0.267929077f, -1.f),
    C_Vertex3f(0.380859375f, 0.247138023f, -1.f),
    C_Vertex3f(0.386534333f, 0.431779027f, -1.f),
    C_Vertex3f(0.390347838f, 0.0415751934f, -1.f),
    C_Vertex3f(0.396497607f, 0.422564864f, -1.f),
    C_Vertex3f(0.400357008f, 0.0516629219f, -1.f),
    C_Vertex3f(0.405764818f, 0.412565947f, -1.f),
    C_Vertex3f(0.409642458f, 0.0625834465f, -1.f),
    C_Vertex3f(0.414335966f, 0.401782274f, -1.f),
    C_Vertex3f(0.418204188f, 0.0743367672f, -1.f),
    C_Vertex3f(0.422211051f, 0.390213847f, -1.f),
    C_Vertex3f(0.426042199f, 0.086922884f, -1.f),
    C_Vertex3f(0.429390073f, 0.377860665f, -1.f),
    C_Vertex3f(0.43315649f, 0.100341797f, -1.f),
    C_Vertex3f(0.435873032f, 0.364722729f, -1.f),
    C_Vertex3f(0.442742348f, 0.12171936f, -1.f),
    C_Vertex3f(0.444553375f, 0.343838692f, -1.f),
    C_Vertex3f(0.452328205f, 0.151996613f, -1.f),
    C_Vertex3f(0.453233719f, 0.314654827f, -1.f),
    C_Vertex3f(0.458718777f, 0.184047699f, -1.f),
    C_Vertex3f(0.459020615f, 0.284132481f, -1.f),
    C_Vertex3f(0.461914063f, 0.21787262f, -1.f),
    C_Vertex3f(0.461914063f, 0.252271652f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369491577f, 0.321942329f, -1.f),
    C_Vertex3f(0.374038696f, 0.30533123f, -1.f),
    C_Vertex3f(0.375874996f, 0.440208435f, -1.f),
    C_Vertex3f(0.377449036f, 0.287326813f, -1.f),
    C_Vertex3f(0.386534333f, 0.431779027f, -1.f),
    C_Vertex3f(0.379722595f, 0.267929077f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.364519596f, 0.447853088f, -1.f),
    C_Vertex3f(0.352727532f, 0.45454216f, -1.f),
    C_Vertex3f(0.356987f, 0.350984573f, -1.f),
    C_Vertex3f(0.363807678f, 0.33716011f, -1.f),
    C_Vertex3f(0.369491577f, 0.321942329f, -1.f),
    C_Vertex3f(0.375874996f, 0.440208435f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.330398798f, 0.384111643f, -1.f),
    C_Vertex3f(0.339935303f, 0.374453545f, -1.f),
    C_Vertex3f(0.340498805f, 0.46027565f, -1.f),
    C_Vertex3f(0.349029541f, 0.363415718f, -1.f),
    C_Vertex3f(0.352727532f, 0.45454216f, -1.f),
    C_Vertex3f(0.356987f, 0.350984573f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.287830114f, 0.408946753f, -1.f),
    C_Vertex3f(0.299135685f, 0.404807568f, -1.f),
    C_Vertex3f(0.301192641f, 0.47174263f, -1.f),
    C_Vertex3f(0.309998989f, 0.399288654f, -1.f),
    C_Vertex3f(0.314731359f, 0.468875885f, -1.f),
    C_Vertex3f(0.320420027f, 0.392390013f, -1.f),
    C_Vertex3f(0.327833414f, 0.465053558f, -1.f),
    C_Vertex3f(0.330398798f, 0.384111643f, -1.f),
    C_Vertex3f(0.340498805f, 0.46027565f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.23915112f, 0.411617398f, -1.f),
    C_Vertex3f(0.251314163f, 0.413085938f, -1.f),
    C_Vertex3f(0.254024506f, 0.474609375f, -1.f),
    C_Vertex3f(0.263892174f, 0.413085938f, -1.f),
    C_Vertex3f(0.272805214f, 0.474609375f, -1.f),
    C_Vertex3f(0.276082277f, 0.411706209f, -1.f),
    C_Vertex3f(0.287217259f, 0.473653793f, -1.f),
    C_Vertex3f(0.287830114f, 0.408946753f, -1.f),
    C_Vertex3f(0.301192641f, 0.47174263f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.23300457f, 0.472400188f, -1.f),
    C_Vertex3f(0.213677406f, 0.467981815f, -1.f),
    C_Vertex3f(0.215906858f, 0.404274702f, -1.f),
    C_Vertex3f(0.227348685f, 0.40868032f, -1.f),
    C_Vertex3f(0.23915112f, 0.411617398f, -1.f),
    C_Vertex3f(0.254024506f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.183745027f, 0.382246614f, -1.f),
    C_Vertex3f(0.194105029f, 0.391057849f, -1.f),
    C_Vertex3f(0.196043015f, 0.461354256f, -1.f),
    C_Vertex3f(0.20482564f, 0.398400545f, -1.f),
    C_Vertex3f(0.213677406f, 0.467981815f, -1.f),
    C_Vertex3f(0.215906858f, 0.404274702f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.155672908f, 0.347138524f, -1.f),
    C_Vertex3f(0.164106846f, 0.360218525f, -1.f),
    C_Vertex3f(0.164978981f, 0.441470623f, -1.f),
    C_Vertex3f(0.173745632f, 0.371966839f, -1.f),
    C_Vertex3f(0.180101395f, 0.452517509f, -1.f),
    C_Vertex3f(0.183745027f, 0.382246614f, -1.f),
    C_Vertex3f(0.196043015f, 0.461354256f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150675774f, 0.428213596f, -1.f),
    C_Vertex3f(0.137191772f, 0.412746429f, -1.f),
    C_Vertex3f(0.137600183f, 0.2999084f, -1.f),
    C_Vertex3f(0.142419577f, 0.316983461f, -1.f),
    C_Vertex3f(0.148443818f, 0.332726836f, -1.f),
    C_Vertex3f(0.155672908f, 0.347138524f, -1.f),
    C_Vertex3f(0.164978981f, 0.441470623f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.137191772f, 0.412746429f, -1.f),
    C_Vertex3f(0.130859375f, 0.403907776f, -1.f),
    C_Vertex3f(0.131575942f, 0.261763215f, -1.f),
    C_Vertex3f(0.133985639f, 0.281501651f, -1.f),
    C_Vertex3f(0.137600183f, 0.2999084f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.380859375f, 0.247138023f, -1.f),
    C_Vertex3f(0.380859375f, 0.224407196f, -1.f),
    C_Vertex3f(0.390347838f, 0.0415751934f, -1.f),
    C_Vertex3f(0.379695415f, 0.20265305f, -1.f),
    C_Vertex3f(0.379614949f, 0.032320261f, -1.f),
    C_Vertex3f(0.377367496f, 0.182422042f, -1.f),
    C_Vertex3f(0.373875618f, 0.16371417f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36921978f, 0.146529436f, -1.f),
    C_Vertex3f(0.363399982f, 0.130867839f, -1.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, -1.f),
    C_Vertex3f(0.379614949f, 0.032320261f, -1.f),
    C_Vertex3f(0.373875618f, 0.16371417f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356416225f, 0.116729379f, -1.f),
    C_Vertex3f(0.348268509f, 0.104114056f, -1.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, -1.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, -1.f),
    C_Vertex3f(0.363399982f, 0.130867839f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.348268509f, 0.104114056f, -1.f),
    C_Vertex3f(0.338956833f, 0.0930218697f, -1.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, -1.f),
    C_Vertex3f(0.329175472f, 0.0833162069f, -1.f),
    C_Vertex3f(0.324899673f, 0.00113010406f, -1.f),
    C_Vertex3f(0.318924427f, 0.0749970675f, -1.f),
    C_Vertex3f(0.308203697f, 0.0680644512f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.299283981f, -0.00645923615f, -1.f),
    C_Vertex3f(0.324899673f, 0.00113010406f, -1.f),
    C_Vertex3f(0.308203697f, 0.0680644512f, -1.f),
    C_Vertex3f(0.297013283f, 0.0625183582f, -1.f),
    C_Vertex3f(0.285353184f, 0.0583587885f, -1.f),
    C_Vertex3f(0.2732234f, 0.055585742f, -1.f),
    C_Vertex3f(0.273040771f, -0.0102539063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.260623932f, 0.0541992188f, -1.f),
    C_Vertex3f(0.247663498f, 0.0541992188f, -1.f),
    C_Vertex3f(0.249773026f, -0.0102539063f, -1.f),
    C_Vertex3f(0.273040771f, -0.0102539063f, -1.f),
    C_Vertex3f(0.2732234f, 0.055585742f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.211967945f, 0.0622315407f, -1.f),
    C_Vertex3f(0.213477612f, -0.00404834747f, -1.f),
    C_Vertex3f(0.223389864f, 0.0582153797f, -1.f),
    C_Vertex3f(0.231119633f, -0.00818538666f, -1.f),
    C_Vertex3f(0.235288382f, 0.0555379391f, -1.f),
    C_Vertex3f(0.249773026f, -0.0102539063f, -1.f),
    C_Vertex3f(0.247663498f, 0.0541992188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.180561781f, 0.0823123455f, -1.f),
    C_Vertex3f(0.181227684f, 0.0104312897f, -1.f),
    C_Vertex3f(0.190553904f, 0.0742800236f, -1.f),
    C_Vertex3f(0.196846962f, 0.0021572113f, -1.f),
    C_Vertex3f(0.201022625f, 0.067586422f, -1.f),
    C_Vertex3f(0.213477612f, -0.00404834747f, -1.f),
    C_Vertex3f(0.211967945f, 0.0622315407f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154098272f, 0.114564657f, -1.f),
    C_Vertex3f(0.154164791f, 0.0302467346f, -1.f),
    C_Vertex3f(0.162007332f, 0.10239315f, -1.f),
    C_Vertex3f(0.167000294f, 0.0197944641f, -1.f),
    C_Vertex3f(0.171046257f, 0.0916833878f, -1.f),
    C_Vertex3f(0.181227684f, 0.0104312897f, -1.f),
    C_Vertex3f(0.180561781f, 0.0823123455f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133760691f, 0.177868128f, -1.f),
    C_Vertex3f(0.137695313f, -0.177734375f, -1.f),
    C_Vertex3f(0.137150288f, 0.159849644f, -1.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, -1.f),
    C_Vertex3f(0.14166975f, 0.143292904f, -1.f),
    C_Vertex3f(0.142721176f, 0.0417881012f, -1.f),
    C_Vertex3f(0.147319078f, 0.128197908f, -1.f),
    C_Vertex3f(0.154164791f, 0.0302467346f, -1.f),
    C_Vertex3f(0.154098272f, 0.114564657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.137695313f, -0.177734375f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, -1.f),
    C_Vertex3f(0.131575942f, 0.261763215f, -1.f),
    C_Vertex3f(0.130859375f, 0.403907776f, -1.f),
    C_Vertex3f(0.130371094f, 0.240693092f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_113[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354858398f, -0.177734375f, 0.f),
    C_Vertex3f(0.354858398f, -0.177734375f, -1.f),
    C_Vertex3f(0.354858398f, 0.049407959f, -1.f),
    C_Vertex3f(0.354858398f, 0.049407959f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.813815057f, -0.581123888f, 0.f),
    C_Vertex3f(0.354858398f, 0.049407959f, 0.f),
    C_Vertex3f(0.354858398f, 0.049407959f, -1.f),
    C_Normal3f(0.759154022f, -0.650911033f, 0.f),
    C_Vertex3f(0.350271225f, 0.0429840088f, 0.f),
    C_Vertex3f(0.350271225f, 0.0429840088f, -1.f),
    C_Normal3f(0.68529743f, -0.728263319f, 0.f),
    C_Vertex3f(0.339268208f, 0.0312252045f, 0.f),
    C_Vertex3f(0.339268208f, 0.0312252045f, -1.f),
    C_Normal3f(0.593066633f, -0.80515343f, 0.f),
    C_Vertex3f(0.32643652f, 0.0205554962f, 0.f),
    C_Vertex3f(0.32643652f, 0.0205554962f, -1.f),
    C_Normal3f(0.501832008f, -0.864965081f, 0.f),
    C_Vertex3f(0.311776161f, 0.010974884f, 0.f),
    C_Vertex3f(0.311776161f, 0.010974884f, -1.f),
    C_Normal3f(0.405510545f, -0.914090335f, 0.f),
    C_Vertex3f(0.295287132f, 0.00248336792f, 0.f),
    C_Vertex3f(0.295287132f, 0.00248336792f, -1.f),
    C_Normal3f(0.293781638f, -0.955872536f, 0.f),
    C_Vertex3f(0.278278828f, -0.00388526917f, 0.f),
    C_Vertex3f(0.278278828f, -0.00388526917f, -1.f),
    C_Normal3f(0.176221251f, -0.984350562f, 0.f),
    C_Vertex3f(0.260751247f, -0.00813102722f, 0.f),
    C_Vertex3f(0.260751247f, -0.00813102722f, -1.f),
    C_Normal3f(0.0562957004f, -0.998414099f, 0.f),
    C_Vertex3f(0.242704391f, -0.0102539063f, 0.f),
    C_Vertex3f(0.242704391f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0520790406f, -0.998642921f, 0.f),
    C_Vertex3f(0.223101616f, -0.0102539063f, 0.f),
    C_Vertex3f(0.223101616f, -0.0102539063f, -1.f),
    C_Normal3f(-0.156098396f, -0.98774147f, 0.f),
    C_Vertex3f(0.203178883f, -0.00819265842f, 0.f),
    C_Vertex3f(0.203178883f, -0.00819265842f, -1.f),
    C_Normal3f(-0.263728827f, -0.964596868f, 0.f),
    C_Vertex3f(0.183972836f, -0.00407016277f, 0.f),
    C_Vertex3f(0.183972836f, -0.00407016277f, -1.f),
    C_Normal3f(-0.369709551f, -0.929147363f, 0.f),
    C_Vertex3f(0.165483475f, 0.0021135807f, 0.f),
    C_Vertex3f(0.165483475f, 0.0021135807f, -1.f),
    C_Normal3f(-0.470113546f, -0.88260597f, 0.f),
    C_Vertex3f(0.1477108f, 0.010358572f, 0.f),
    C_Vertex3f(0.1477108f, 0.010358572f, -1.f),
    C_Normal3f(-0.561715722f, -0.827330291f, 0.f),
    C_Vertex3f(0.130654812f, 0.0206648111f, 0.f),
    C_Vertex3f(0.130654812f, 0.0206648111f, -1.f),
    C_Normal3f(-0.642463803f, -0.766316056f, 0.f),
    C_Vertex3f(0.11431551f, 0.0330322981f, 0.f),
    C_Vertex3f(0.11431551f, 0.0330322981f, -1.f),
    C_Normal3f(-0.711582363f, -0.702602684f, 0.f),
    C_Vertex3f(0.098692894f, 0.0474610329f, 0.f),
    C_Vertex3f(0.098692894f, 0.0474610329f, -1.f),
    C_Normal3f(-0.777102709f, -0.629373729f, 0.f),
    C_Vertex3f(0.0837869644f, 0.0639510155f, 0.f),
    C_Vertex3f(0.0837869644f, 0.0639510155f, -1.f),
    C_Normal3f(-0.840442777f, -0.541900218f, 0.f),
    C_Vertex3f(0.070744276f, 0.0819698572f, 0.f),
    C_Vertex3f(0.070744276f, 0.0819698572f, -1.f),
    C_Normal3f(-0.89280808f, -0.450437367f, 0.f),
    C_Vertex3f(0.0595648289f, 0.101517558f, 0.f),
    C_Vertex3f(0.0595648289f, 0.101517558f, -1.f),
    C_Normal3f(-0.933571637f, -0.358390868f, 0.f),
    C_Vertex3f(0.0502486229f, 0.122594118f, 0.f),
    C_Vertex3f(0.0502486229f, 0.122594118f, -1.f),
    C_Normal3f(-0.963201344f, -0.268780828f, 0.f),
    C_Vertex3f(0.0427956581f, 0.145199537f, 0.f),
    C_Vertex3f(0.0427956581f, 0.145199537f, -1.f),
    C_Normal3f(-0.982946455f, -0.183891699f, 0.f),
    C_Vertex3f(0.0372059345f, 0.169333816f, 0.f),
    C_Vertex3f(0.0372059345f, 0.169333816f, -1.f),
    C_Normal3f(-0.994454324f, -0.105169065f, 0.f),
    C_Vertex3f(0.0334794521f, 0.194996953f, 0.f),
    C_Vertex3f(0.0334794521f, 0.194996953f, -1.f),
    C_Normal3f(-0.999502122f, -0.0315519422f, 0.f),
    C_Vertex3f(0.0316162109f, 0.22218895f, 0.f),
    C_Vertex3f(0.0316162109f, 0.22218895f, -1.f),
    C_Normal3f(-0.99891448f, 0.0465811603f, 0.f),
    C_Vertex3f(0.0316162109f, 0.254020691f, 0.f),
    C_Vertex3f(0.0316162109f, 0.254020691f, -1.f),
    C_Normal3f(-0.990210652f, 0.13958098f, 0.f),
    C_Vertex3f(0.0346460342f, 0.287162304f, 0.f),
    C_Vertex3f(0.0346460342f, 0.287162304f, -1.f),
    C_Normal3f(-0.970407963f, 0.24147141f, 0.f),
    C_Vertex3f(0.0407056808f, 0.318502903f, 0.f),
    C_Vertex3f(0.0407056808f, 0.318502903f, -1.f),
    C_Normal3f(-0.940557837f, 0.339633495f, 0.f),
    C_Vertex3f(0.0497951508f, 0.348042488f, 0.f),
    C_Vertex3f(0.0497951508f, 0.348042488f, -1.f),
    C_Normal3f(-0.905745983f, 0.423820972f, 0.f),
    C_Vertex3f(0.0588846207f, 0.368846416f, 0.f),
    C_Vertex3f(0.0588846207f, 0.368846416f, -1.f),
    C_Normal3f(-0.871862352f, 0.489751101f, 0.f),
    C_Vertex3f(0.0656267405f, 0.381876111f, 0.f),
    C_Vertex3f(0.0656267405f, 0.381876111f, -1.f),
    C_Normal3f(-0.834693313f, 0.550715029f, 0.f),
    C_Vertex3f(0.0730513334f, 0.394066215f, 0.f),
    C_Vertex3f(0.0730513334f, 0.394066215f, -1.f),
    C_Normal3f(-0.791220546f, 0.61153084f, 0.f),
    C_Vertex3f(0.0811583996f, 0.405416727f, 0.f),
    C_Vertex3f(0.0811583996f, 0.405416727f, -1.f),
    C_Normal3f(-0.741509855f, 0.670942008f, 0.f),
    C_Vertex3f(0.0899479389f, 0.415927649f, 0.f),
    C_Vertex3f(0.0899479389f, 0.415927649f, -1.f),
    C_Normal3f(-0.685975313f, 0.727624834f, 0.f),
    C_Vertex3f(0.0994199514f, 0.425598979f, 0.f),
    C_Vertex3f(0.0994199514f, 0.425598979f, -1.f),
    C_Normal3f(-0.625391364f, 0.780311227f, 0.f),
    C_Vertex3f(0.109574437f, 0.434430718f, 0.f),
    C_Vertex3f(0.109574437f, 0.434430718f, -1.f),
    C_Normal3f(-0.554226935f, 0.832365572f, 0.f),
    C_Vertex3f(0.120411396f, 0.442422867f, 0.f),
    C_Vertex3f(0.120411396f, 0.442422867f, -1.f),
    C_Normal3f(-0.459839851f, 0.888001859f, 0.f),
    C_Vertex3f(0.137690544f, 0.453151703f, 0.f),
    C_Vertex3f(0.137690544f, 0.453151703f, -1.f),
    C_Normal3f(-0.340146154f, 0.940372646f, 0.f),
    C_Vertex3f(0.161848545f, 0.463880539f, 0.f),
    C_Vertex3f(0.161848545f, 0.463880539f, -1.f),
    C_Normal3f(-0.203292131f, 0.979118168f, 0.f),
    C_Vertex3f(0.187125683f, 0.471033096f, 0.f),
    C_Vertex3f(0.187125683f, 0.471033096f, -1.f),
    C_Normal3f(-0.0700763613f, 0.997541666f, 0.f),
    C_Vertex3f(0.213521957f, 0.474609375f, 0.f),
    C_Vertex3f(0.213521957f, 0.474609375f, -1.f),
    C_Normal3f(0.0504245572f, 0.998727798f, 0.f),
    C_Vertex3f(0.238034248f, 0.474609375f, 0.f),
    C_Vertex3f(0.238034248f, 0.474609375f, -1.f),
    C_Normal3f(0.169254988f, 0.985572338f, 0.f),
    C_Vertex3f(0.258397102f, 0.472343683f, 0.f),
    C_Vertex3f(0.258397102f, 0.472343683f, -1.f),
    C_Normal3f(0.290699273f, 0.956814468f, 0.f),
    C_Vertex3f(0.27761364f, 0.4678123f, 0.f),
    C_Vertex3f(0.27761364f, 0.4678123f, -1.f),
    C_Normal3f(0.412797958f, 0.91082263f, 0.f),
    C_Vertex3f(0.295683861f, 0.461015224f, 0.f),
    C_Vertex3f(0.295683861f, 0.461015224f, -1.f),
    C_Normal3f(0.529117823f, 0.848548412f, 0.f),
    C_Vertex3f(0.312607765f, 0.451952457f, 0.f),
    C_Vertex3f(0.312607765f, 0.451952457f, -1.f),
    C_Normal3f(0.633887112f, 0.773425579f, 0.f),
    C_Vertex3f(0.328385353f, 0.440623999f, 0.f),
    C_Vertex3f(0.328385353f, 0.440623999f, -1.f),
    C_Normal3f(0.723342657f, 0.690489233f, 0.f),
    C_Vertex3f(0.343016624f, 0.427029848f, 0.f),
    C_Vertex3f(0.343016624f, 0.427029848f, -1.f),
    C_Normal3f(0.785211623f, 0.619227469f, 0.f),
    C_Vertex3f(0.356501579f, 0.411170006f, 0.f),
    C_Vertex3f(0.356501579f, 0.411170006f, -1.f),
    C_Normal3f(0.826644897f, 0.562723994f, 0.f),
    C_Vertex3f(0.362670898f, 0.402107239f, 0.f),
    C_Vertex3f(0.362670898f, 0.402107239f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.362670898f, 0.402107239f, 0.f),
    C_Vertex3f(0.362670898f, 0.402107239f, -1.f),
    C_Vertex3f(0.362670898f, 0.464355469f, -1.f),
    C_Vertex3f(0.362670898f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.362670898f, 0.464355469f, 0.f),
    C_Vertex3f(0.362670898f, 0.464355469f, -1.f),
    C_Vertex3f(0.43347168f, 0.464355469f, -1.f),
    C_Vertex3f(0.43347168f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.43347168f, 0.464355469f, 0.f),
    C_Vertex3f(0.43347168f, 0.464355469f, -1.f),
    C_Vertex3f(0.43347168f, -0.177734375f, -1.f),
    C_Vertex3f(0.43347168f, -0.177734375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.43347168f, -0.177734375f, 0.f),
    C_Vertex3f(0.43347168f, -0.177734375f, -1.f),
    C_Vertex3f(0.354858398f, -0.177734375f, -1.f),
    C_Vertex3f(0.354858398f, -0.177734375f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999684989f, -0.0250965934f, 0.f),
    C_Vertex3f(0.112670898f, 0.244926453f, 0.f),
    C_Vertex3f(0.112670898f, 0.244926453f, -1.f),
    C_Normal3f(0.999640703f, 0.0268048979f, 0.f),
    C_Vertex3f(0.112670898f, 0.222198486f, 0.f),
    C_Vertex3f(0.112670898f, 0.222198486f, -1.f),
    C_Normal3f(0.996267676f, 0.0863180757f, 0.f),
    C_Vertex3f(0.113843441f, 0.201198578f, 0.f),
    C_Vertex3f(0.113843441f, 0.201198578f, -1.f),
    C_Normal3f(0.988184929f, 0.1532664f, 0.f),
    C_Vertex3f(0.116188526f, 0.181598663f, 0.f),
    C_Vertex3f(0.116188526f, 0.181598663f, -1.f),
    C_Normal3f(0.973587215f, 0.228315488f, 0.f),
    C_Vertex3f(0.119706154f, 0.163398743f, 0.f),
    C_Vertex3f(0.119706154f, 0.163398743f, -1.f),
    C_Normal3f(0.950268328f, 0.311432123f, 0.f),
    C_Vertex3f(0.124396324f, 0.146598816f, 0.f),
    C_Vertex3f(0.124396324f, 0.146598816f, -1.f),
    C_Normal3f(0.915750504f, 0.401747346f, 0.f),
    C_Vertex3f(0.130259037f, 0.131198883f, 0.f),
    C_Vertex3f(0.130259037f, 0.131198883f, -1.f),
    C_Normal3f(0.867636621f, 0.49719885f, 0.f),
    C_Vertex3f(0.137294292f, 0.117198944f, 0.f),
    C_Vertex3f(0.137294292f, 0.117198944f, -1.f),
    C_Normal3f(0.804226041f, 0.594323575f, 0.f),
    C_Vertex3f(0.14550209f, 0.104598999f, 0.f),
    C_Vertex3f(0.14550209f, 0.104598999f, -1.f),
    C_Normal3f(0.737600327f, 0.675237596f, 0.f),
    C_Vertex3f(0.154882431f, 0.0933990479f, 0.f),
    C_Vertex3f(0.154882431f, 0.0933990479f, -1.f),
    C_Normal3f(0.670250714f, 0.74213475f, 0.f),
    C_Vertex3f(0.164726496f, 0.0835990906f, 0.f),
    C_Vertex3f(0.164726496f, 0.0835990906f, -1.f),
    C_Normal3f(0.589912415f, 0.807467282f, 0.f),
    C_Vertex3f(0.175034285f, 0.0751991272f, 0.f),
    C_Vertex3f(0.175034285f, 0.0751991272f, -1.f),
    C_Normal3f(0.496871829f, 0.867823958f, 0.f),
    C_Vertex3f(0.185805798f, 0.0681991577f, 0.f),
    C_Vertex3f(0.185805798f, 0.0681991577f, -1.f),
    C_Normal3f(0.392937213f, 0.91956526f, 0.f),
    C_Vertex3f(0.197041035f, 0.0625991821f, 0.f),
    C_Vertex3f(0.197041035f, 0.0625991821f, -1.f),
    C_Normal3f(0.281494051f, 0.959563017f, 0.f),
    C_Vertex3f(0.208739996f, 0.0583992004f, 0.f),
    C_Vertex3f(0.208739996f, 0.0583992004f, -1.f),
    C_Normal3f(0.167047918f, 0.985948741f, 0.f),
    C_Vertex3f(0.220902681f, 0.0555992126f, 0.f),
    C_Vertex3f(0.220902681f, 0.0555992126f, -1.f),
    C_Normal3f(0.0549419522f, 0.998489499f, 0.f),
    C_Vertex3f(0.233529091f, 0.0541992188f, 0.f),
    C_Vertex3f(0.233529091f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0533560067f, 0.998575509f, 0.f),
    C_Vertex3f(0.24634552f, 0.0541992188f, 0.f),
    C_Vertex3f(0.24634552f, 0.0541992188f, -1.f),
    C_Normal3f(-0.165700138f, 0.986176193f, 0.f),
    C_Vertex3f(0.258452058f, 0.0555309057f, 0.f),
    C_Vertex3f(0.258452058f, 0.0555309057f, -1.f),
    C_Normal3f(-0.279147983f, 0.960248113f, 0.f),
    C_Vertex3f(0.270122409f, 0.0581942797f, 0.f),
    C_Vertex3f(0.270122409f, 0.0581942797f, -1.f),
    C_Normal3f(-0.389658362f, 0.920959473f, 0.f),
    C_Vertex3f(0.281356573f, 0.0621893406f, 0.f),
    C_Vertex3f(0.281356573f, 0.0621893406f, -1.f),
    C_Normal3f(-0.492845923f, 0.870116532f, 0.f),
    C_Vertex3f(0.292154551f, 0.0675160885f, 0.f),
    C_Vertex3f(0.292154551f, 0.0675160885f, -1.f),
    C_Normal3f(-0.585399926f, 0.810744643f, 0.f),
    C_Vertex3f(0.302516341f, 0.0741745234f, 0.f),
    C_Vertex3f(0.302516341f, 0.0741745234f, -1.f),
    C_Normal3f(-0.665525615f, 0.746375084f, 0.f),
    C_Vertex3f(0.312441945f, 0.0821646452f, 0.f),
    C_Vertex3f(0.312441945f, 0.0821646452f, -1.f),
    C_Normal3f(-0.7329005f, 0.680335879f, 0.f),
    C_Vertex3f(0.321931362f, 0.091486454f, 0.f),
    C_Vertex3f(0.321931362f, 0.091486454f, -1.f),
    C_Normal3f(-0.800764024f, 0.598979831f, 0.f),
    C_Vertex3f(0.330984592f, 0.10213995f, 0.f),
    C_Vertex3f(0.330984592f, 0.10213995f, -1.f),
    C_Normal3f(-0.865895987f, 0.500224054f, 0.f),
    C_Vertex3f(0.338906169f, 0.11417973f, 0.f),
    C_Vertex3f(0.338906169f, 0.11417973f, -1.f),
    C_Normal3f(-0.915036619f, 0.403370768f, 0.f),
    C_Vertex3f(0.345696092f, 0.127605796f, 0.f),
    C_Vertex3f(0.345696092f, 0.127605796f, -1.f),
    C_Normal3f(-0.95007205f, 0.312030524f, 0.f),
    C_Vertex3f(0.351354361f, 0.142418146f, 0.f),
    C_Vertex3f(0.351354361f, 0.142418146f, -1.f),
    C_Normal3f(-0.973593593f, 0.228288472f, 0.f),
    C_Vertex3f(0.355880976f, 0.158616781f, 0.f),
    C_Vertex3f(0.355880976f, 0.158616781f, -1.f),
    C_Normal3f(-0.988232076f, 0.152961582f, 0.f),
    C_Vertex3f(0.359275937f, 0.176201701f, 0.f),
    C_Vertex3f(0.359275937f, 0.176201701f, -1.f),
    C_Normal3f(-0.996294916f, 0.0860029235f, 0.f),
    C_Vertex3f(0.361539245f, 0.195172906f, 0.f),
    C_Vertex3f(0.361539245f, 0.195172906f, -1.f),
    C_Normal3f(-0.999650836f, 0.0264240205f, 0.f),
    C_Vertex3f(0.362670898f, 0.215530396f, 0.f),
    C_Vertex3f(0.362670898f, 0.215530396f, -1.f),
    C_Normal3f(-0.999635696f, -0.0269913767f, 0.f),
    C_Vertex3f(0.362670898f, 0.237984657f, 0.f),
    C_Vertex3f(0.362670898f, 0.237984657f, -1.f),
    C_Normal3f(-0.996393263f, -0.0848552287f, 0.f),
    C_Vertex3f(0.361477852f, 0.259715319f, 0.f),
    C_Vertex3f(0.361477852f, 0.259715319f, -1.f),
    C_Normal3f(-0.988595366f, -0.150596097f, 0.f),
    C_Vertex3f(0.359091759f, 0.280011892f, 0.f),
    C_Vertex3f(0.359091759f, 0.280011892f, -1.f),
    C_Normal3f(-0.974529266f, -0.224260494f, 0.f),
    C_Vertex3f(0.355512619f, 0.298874378f, 0.f),
    C_Vertex3f(0.355512619f, 0.298874378f, -1.f),
    C_Normal3f(-0.952075005f, -0.305864751f, 0.f),
    C_Vertex3f(0.350740433f, 0.316302776f, 0.f),
    C_Vertex3f(0.350740433f, 0.316302776f, -1.f),
    C_Normal3f(-0.918832123f, -0.394648641f, 0.f),
    C_Vertex3f(0.3447752f, 0.332297087f, 0.f),
    C_Vertex3f(0.3447752f, 0.332297087f, -1.f),
    C_Normal3f(-0.87243551f, -0.488729179f, 0.f),
    C_Vertex3f(0.33761692f, 0.346857309f, 0.f),
    C_Vertex3f(0.33761692f, 0.346857309f, -1.f),
    C_Normal3f(-0.811121821f, -0.584877312f, 0.f),
    C_Vertex3f(0.329265594f, 0.359983444f, 0.f),
    C_Vertex3f(0.329265594f, 0.359983444f, -1.f),
    C_Normal3f(-0.747107983f, -0.664702654f, 0.f),
    C_Vertex3f(0.319721222f, 0.371675491f, 0.f),
    C_Vertex3f(0.319721222f, 0.371675491f, -1.f),
    C_Normal3f(-0.682686269f, -0.730711579f, 0.f),
    C_Vertex3f(0.309761047f, 0.381906033f, 0.f),
    C_Vertex3f(0.309761047f, 0.381906033f, -1.f),
    C_Normal3f(-0.604822636f, -0.796360254f, 0.f),
    C_Vertex3f(0.299385071f, 0.390675068f, 0.f),
    C_Vertex3f(0.299385071f, 0.390675068f, -1.f),
    C_Normal3f(-0.513173759f, -0.858284771f, 0.f),
    C_Vertex3f(0.288593292f, 0.397982597f, 0.f),
    C_Vertex3f(0.288593292f, 0.397982597f, -1.f),
    C_Normal3f(-0.408921808f, -0.912569404f, 0.f),
    C_Vertex3f(0.277385712f, 0.403828621f, 0.f),
    C_Vertex3f(0.277385712f, 0.403828621f, -1.f),
    C_Normal3f(-0.295072019f, -0.955474973f, 0.f),
    C_Vertex3f(0.265762329f, 0.408213139f, 0.f),
    C_Vertex3f(0.265762329f, 0.408213139f, -1.f),
    C_Normal3f(-0.17620182f, -0.984354019f, 0.f),
    C_Vertex3f(0.253723145f, 0.41113615f, 0.f),
    C_Vertex3f(0.253723145f, 0.41113615f, -1.f),
    C_Normal3f(-0.0577359013f, -0.998331904f, 0.f),
    C_Vertex3f(0.241268158f, 0.412597656f, 0.f),
    C_Vertex3f(0.241268158f, 0.412597656f, -1.f),
    C_Normal3f(0.0540757291f, -0.998536825f, 0.f),
    C_Vertex3f(0.228451729f, 0.412597656f, 0.f),
    C_Vertex3f(0.228451729f, 0.412597656f, -1.f),
    C_Normal3f(0.166988358f, -0.985958815f, 0.f),
    C_Vertex3f(0.216173768f, 0.41123867f, 0.f),
    C_Vertex3f(0.216173768f, 0.41123867f, -1.f),
    C_Normal3f(0.282156944f, -0.959368289f, 0.f),
    C_Vertex3f(0.204379916f, 0.408520699f, 0.f),
    C_Vertex3f(0.204379916f, 0.408520699f, -1.f),
    C_Normal3f(0.394841701f, -0.918749094f, 0.f),
    C_Vertex3f(0.193070173f, 0.404443741f, 0.f),
    C_Vertex3f(0.193070173f, 0.404443741f, -1.f),
    C_Normal3f(0.500306964f, -0.865848005f, 0.f),
    C_Vertex3f(0.182244539f, 0.399007797f, 0.f),
    C_Vertex3f(0.182244539f, 0.399007797f, -1.f),
    C_Normal3f(0.594884932f, -0.803810954f, 0.f),
    C_Vertex3f(0.171903014f, 0.392212868f, 0.f),
    C_Vertex3f(0.171903014f, 0.392212868f, -1.f),
    C_Normal3f(0.676528513f, -0.736416399f, 0.f),
    C_Vertex3f(0.162045598f, 0.384058952f, 0.f),
    C_Vertex3f(0.162045598f, 0.384058952f, -1.f),
    C_Normal3f(0.744814754f, -0.667271256f, 0.f),
    C_Vertex3f(0.152672291f, 0.374546051f, 0.f),
    C_Vertex3f(0.152672291f, 0.374546051f, -1.f),
    C_Normal3f(0.812937677f, -0.582350731f, 0.f),
    C_Vertex3f(0.143783092f, 0.363674164f, 0.f),
    C_Vertex3f(0.143783092f, 0.363674164f, -1.f),
    C_Normal3f(0.876596987f, -0.481225252f, 0.f),
    C_Vertex3f(0.136005044f, 0.351279259f, 0.f),
    C_Vertex3f(0.136005044f, 0.351279259f, -1.f),
    C_Normal3f(0.923190892f, -0.384341747f, 0.f),
    C_Vertex3f(0.129338145f, 0.337361336f, 0.f),
    C_Vertex3f(0.129338145f, 0.337361336f, -1.f),
    C_Normal3f(0.955533326f, -0.29488337f, 0.f),
    C_Vertex3f(0.123782396f, 0.321920395f, 0.f),
    C_Vertex3f(0.123782396f, 0.321920395f, -1.f),
    C_Normal3f(0.976766706f, -0.214305729f, 0.f),
    C_Vertex3f(0.119337797f, 0.304956436f, 0.f),
    C_Vertex3f(0.119337797f, 0.304956436f, -1.f),
    C_Normal3f(0.989746273f, -0.14283675f, 0.f),
    C_Vertex3f(0.116004348f, 0.28646946f, 0.f),
    C_Vertex3f(0.116004348f, 0.28646946f, -1.f),
    C_Normal3f(0.996796131f, -0.0799838454f, 0.f),
    C_Vertex3f(0.113782048f, 0.266459465f, 0.f),
    C_Vertex3f(0.113782048f, 0.266459465f, -1.f),
    C_Normal3f(0.999684989f, -0.0250965934f, 0.f),
    C_Vertex3f(0.112670898f, 0.244926453f, 0.f),
    C_Vertex3f(0.112670898f, 0.244926453f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.361539245f, 0.195172906f, 0.f),
    C_Vertex3f(0.359275937f, 0.176201701f, 0.f),
    C_Vertex3f(0.43347168f, -0.177734375f, 0.f),
    C_Vertex3f(0.355880976f, 0.158616781f, 0.f),
    C_Vertex3f(0.354858398f, 0.049407959f, 0.f),
    C_Vertex3f(0.351354361f, 0.142418146f, 0.f),
    C_Vertex3f(0.350271225f, 0.0429840088f, 0.f),
    C_Vertex3f(0.345696092f, 0.127605796f, 0.f),
    C_Vertex3f(0.339268208f, 0.0312252045f, 0.f),
    C_Vertex3f(0.338906169f, 0.11417973f, 0.f),
    C_Vertex3f(0.330984592f, 0.10213995f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.32643652f, 0.0205554962f, 0.f),
    C_Vertex3f(0.339268208f, 0.0312252045f, 0.f),
    C_Vertex3f(0.330984592f, 0.10213995f, 0.f),
    C_Vertex3f(0.321931362f, 0.091486454f, 0.f),
    C_Vertex3f(0.312441945f, 0.0821646452f, 0.f),
    C_Vertex3f(0.311776161f, 0.010974884f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.302516341f, 0.0741745234f, 0.f),
    C_Vertex3f(0.292154551f, 0.0675160885f, 0.f),
    C_Vertex3f(0.295287132f, 0.00248336792f, 0.f),
    C_Vertex3f(0.311776161f, 0.010974884f, 0.f),
    C_Vertex3f(0.312441945f, 0.0821646452f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.258452058f, 0.0555309057f, 0.f),
    C_Vertex3f(0.260751247f, -0.00813102722f, 0.f),
    C_Vertex3f(0.270122409f, 0.0581942797f, 0.f),
    C_Vertex3f(0.278278828f, -0.00388526917f, 0.f),
    C_Vertex3f(0.281356573f, 0.0621893406f, 0.f),
    C_Vertex3f(0.295287132f, 0.00248336792f, 0.f),
    C_Vertex3f(0.292154551f, 0.0675160885f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220902681f, 0.0555992126f, 0.f),
    C_Vertex3f(0.223101616f, -0.0102539063f, 0.f),
    C_Vertex3f(0.233529091f, 0.0541992188f, 0.f),
    C_Vertex3f(0.242704391f, -0.0102539063f, 0.f),
    C_Vertex3f(0.24634552f, 0.0541992188f, 0.f),
    C_Vertex3f(0.260751247f, -0.00813102722f, 0.f),
    C_Vertex3f(0.258452058f, 0.0555309057f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.208739996f, 0.0583992004f, 0.f),
    C_Vertex3f(0.197041035f, 0.0625991821f, 0.f),
    C_Vertex3f(0.203178883f, -0.00819265842f, 0.f),
    C_Vertex3f(0.223101616f, -0.0102539063f, 0.f),
    C_Vertex3f(0.220902681f, 0.0555992126f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164726496f, 0.0835990906f, 0.f),
    C_Vertex3f(0.165483475f, 0.0021135807f, 0.f),
    C_Vertex3f(0.175034285f, 0.0751991272f, 0.f),
    C_Vertex3f(0.183972836f, -0.00407016277f, 0.f),
    C_Vertex3f(0.185805798f, 0.0681991577f, 0.f),
    C_Vertex3f(0.203178883f, -0.00819265842f, 0.f),
    C_Vertex3f(0.197041035f, 0.0625991821f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.154882431f, 0.0933990479f, 0.f),
    C_Vertex3f(0.14550209f, 0.104598999f, 0.f),
    C_Vertex3f(0.1477108f, 0.010358572f, 0.f),
    C_Vertex3f(0.165483475f, 0.0021135807f, 0.f),
    C_Vertex3f(0.164726496f, 0.0835990906f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.137294292f, 0.117198944f, 0.f),
    C_Vertex3f(0.130259037f, 0.131198883f, 0.f),
    C_Vertex3f(0.130654812f, 0.0206648111f, 0.f),
    C_Vertex3f(0.1477108f, 0.010358572f, 0.f),
    C_Vertex3f(0.14550209f, 0.104598999f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130654812f, 0.0206648111f, 0.f),
    C_Vertex3f(0.130259037f, 0.131198883f, 0.f),
    C_Vertex3f(0.124396324f, 0.146598816f, 0.f),
    C_Vertex3f(0.119706154f, 0.163398743f, 0.f),
    C_Vertex3f(0.116188526f, 0.181598663f, 0.f),
    C_Vertex3f(0.11431551f, 0.0330322981f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116188526f, 0.181598663f, 0.f),
    C_Vertex3f(0.113843441f, 0.201198578f, 0.f),
    C_Vertex3f(0.11431551f, 0.0330322981f, 0.f),
    C_Vertex3f(0.112670898f, 0.222198486f, 0.f),
    C_Vertex3f(0.109574437f, 0.434430718f, 0.f),
    C_Vertex3f(0.112670898f, 0.244926453f, 0.f),
    C_Vertex3f(0.120411396f, 0.442422867f, 0.f),
    C_Vertex3f(0.113782048f, 0.266459465f, 0.f),
    C_Vertex3f(0.116004348f, 0.28646946f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0811583996f, 0.405416727f, 0.f),
    C_Vertex3f(0.0837869644f, 0.0639510155f, 0.f),
    C_Vertex3f(0.0899479389f, 0.415927649f, 0.f),
    C_Vertex3f(0.098692894f, 0.0474610329f, 0.f),
    C_Vertex3f(0.0994199514f, 0.425598979f, 0.f),
    C_Vertex3f(0.11431551f, 0.0330322981f, 0.f),
    C_Vertex3f(0.109574437f, 0.434430718f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0316162109f, 0.254020691f, 0.f),
    C_Vertex3f(0.0334794521f, 0.194996953f, 0.f),
    C_Vertex3f(0.0346460342f, 0.287162304f, 0.f),
    C_Vertex3f(0.0372059345f, 0.169333816f, 0.f),
    C_Vertex3f(0.0407056808f, 0.318502903f, 0.f),
    C_Vertex3f(0.0427956581f, 0.145199537f, 0.f),
    C_Vertex3f(0.0497951508f, 0.348042488f, 0.f),
    C_Vertex3f(0.0502486229f, 0.122594118f, 0.f),
    C_Vertex3f(0.0588846207f, 0.368846416f, 0.f),
    C_Vertex3f(0.0595648289f, 0.101517558f, 0.f),
    C_Vertex3f(0.0656267405f, 0.381876111f, 0.f),
    C_Vertex3f(0.070744276f, 0.0819698572f, 0.f),
    C_Vertex3f(0.0730513334f, 0.394066215f, 0.f),
    C_Vertex3f(0.0837869644f, 0.0639510155f, 0.f),
    C_Vertex3f(0.0811583996f, 0.405416727f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.362670898f, 0.402107239f, 0.f),
    C_Vertex3f(0.356501579f, 0.411170006f, 0.f),
    C_Vertex3f(0.359091759f, 0.280011892f, 0.f),
    C_Vertex3f(0.361477852f, 0.259715319f, 0.f),
    C_Vertex3f(0.362670898f, 0.237984657f, 0.f),
    C_Vertex3f(0.43347168f, -0.177734375f, 0.f),
    C_Vertex3f(0.362670898f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356501579f, 0.411170006f, 0.f),
    C_Vertex3f(0.343016624f, 0.427029848f, 0.f),
    C_Vertex3f(0.3447752f, 0.332297087f, 0.f),
    C_Vertex3f(0.350740433f, 0.316302776f, 0.f),
    C_Vertex3f(0.355512619f, 0.298874378f, 0.f),
    C_Vertex3f(0.359091759f, 0.280011892f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.299385071f, 0.390675068f, 0.f),
    C_Vertex3f(0.309761047f, 0.381906033f, 0.f),
    C_Vertex3f(0.312607765f, 0.451952457f, 0.f),
    C_Vertex3f(0.319721222f, 0.371675491f, 0.f),
    C_Vertex3f(0.328385353f, 0.440623999f, 0.f),
    C_Vertex3f(0.329265594f, 0.359983444f, 0.f),
    C_Vertex3f(0.343016624f, 0.427029848f, 0.f),
    C_Vertex3f(0.33761692f, 0.346857309f, 0.f),
    C_Vertex3f(0.3447752f, 0.332297087f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.265762329f, 0.408213139f, 0.f),
    C_Vertex3f(0.277385712f, 0.403828621f, 0.f),
    C_Vertex3f(0.27761364f, 0.4678123f, 0.f),
    C_Vertex3f(0.288593292f, 0.397982597f, 0.f),
    C_Vertex3f(0.295683861f, 0.461015224f, 0.f),
    C_Vertex3f(0.299385071f, 0.390675068f, 0.f),
    C_Vertex3f(0.312607765f, 0.451952457f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.258397102f, 0.472343683f, 0.f),
    C_Vertex3f(0.238034248f, 0.474609375f, 0.f),
    C_Vertex3f(0.241268158f, 0.412597656f, 0.f),
    C_Vertex3f(0.253723145f, 0.41113615f, 0.f),
    C_Vertex3f(0.265762329f, 0.408213139f, 0.f),
    C_Vertex3f(0.27761364f, 0.4678123f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193070173f, 0.404443741f, 0.f),
    C_Vertex3f(0.204379916f, 0.408520699f, 0.f),
    C_Vertex3f(0.213521957f, 0.474609375f, 0.f),
    C_Vertex3f(0.216173768f, 0.41123867f, 0.f),
    C_Vertex3f(0.238034248f, 0.474609375f, 0.f),
    C_Vertex3f(0.228451729f, 0.412597656f, 0.f),
    C_Vertex3f(0.241268158f, 0.412597656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.187125683f, 0.471033096f, 0.f),
    C_Vertex3f(0.161848545f, 0.463880539f, 0.f),
    C_Vertex3f(0.162045598f, 0.384058952f, 0.f),
    C_Vertex3f(0.171903014f, 0.392212868f, 0.f),
    C_Vertex3f(0.182244539f, 0.399007797f, 0.f),
    C_Vertex3f(0.193070173f, 0.404443741f, 0.f),
    C_Vertex3f(0.213521957f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129338145f, 0.337361336f, 0.f),
    C_Vertex3f(0.136005044f, 0.351279259f, 0.f),
    C_Vertex3f(0.137690544f, 0.453151703f, 0.f),
    C_Vertex3f(0.143783092f, 0.363674164f, 0.f),
    C_Vertex3f(0.161848545f, 0.463880539f, 0.f),
    C_Vertex3f(0.152672291f, 0.374546051f, 0.f),
    C_Vertex3f(0.162045598f, 0.384058952f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116004348f, 0.28646946f, 0.f),
    C_Vertex3f(0.119337797f, 0.304956436f, 0.f),
    C_Vertex3f(0.120411396f, 0.442422867f, 0.f),
    C_Vertex3f(0.123782396f, 0.321920395f, 0.f),
    C_Vertex3f(0.137690544f, 0.453151703f, 0.f),
    C_Vertex3f(0.129338145f, 0.337361336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.362670898f, 0.215530396f, 0.f),
    C_Vertex3f(0.361539245f, 0.195172906f, 0.f),
    C_Vertex3f(0.43347168f, -0.177734375f, 0.f),
    C_Vertex3f(0.362670898f, 0.237984657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.43347168f, -0.177734375f, 0.f),
    C_Vertex3f(0.354858398f, 0.049407959f, 0.f),
    C_Vertex3f(0.354858398f, -0.177734375f, 0.f),
    C_Vertex3f(0.362670898f, 0.464355469f, 0.f),
    C_Vertex3f(0.43347168f, -0.177734375f, 0.f),
    C_Vertex3f(0.43347168f, 0.464355469f, 0.f),
    C_Vertex3f(0.0334794521f, 0.194996953f, 0.f),
    C_Vertex3f(0.0316162109f, 0.254020691f, 0.f),
    C_Vertex3f(0.0316162109f, 0.22218895f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.361539245f, 0.195172906f, -1.f),
    C_Vertex3f(0.359275937f, 0.176201701f, -1.f),
    C_Vertex3f(0.43347168f, -0.177734375f, -1.f),
    C_Vertex3f(0.355880976f, 0.158616781f, -1.f),
    C_Vertex3f(0.354858398f, 0.049407959f, -1.f),
    C_Vertex3f(0.351354361f, 0.142418146f, -1.f),
    C_Vertex3f(0.350271225f, 0.0429840088f, -1.f),
    C_Vertex3f(0.345696092f, 0.127605796f, -1.f),
    C_Vertex3f(0.339268208f, 0.0312252045f, -1.f),
    C_Vertex3f(0.338906169f, 0.11417973f, -1.f),
    C_Vertex3f(0.330984592f, 0.10213995f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.32643652f, 0.0205554962f, -1.f),
    C_Vertex3f(0.339268208f, 0.0312252045f, -1.f),
    C_Vertex3f(0.330984592f, 0.10213995f, -1.f),
    C_Vertex3f(0.321931362f, 0.091486454f, -1.f),
    C_Vertex3f(0.312441945f, 0.0821646452f, -1.f),
    C_Vertex3f(0.311776161f, 0.010974884f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.302516341f, 0.0741745234f, -1.f),
    C_Vertex3f(0.292154551f, 0.0675160885f, -1.f),
    C_Vertex3f(0.295287132f, 0.00248336792f, -1.f),
    C_Vertex3f(0.311776161f, 0.010974884f, -1.f),
    C_Vertex3f(0.312441945f, 0.0821646452f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.258452058f, 0.0555309057f, -1.f),
    C_Vertex3f(0.260751247f, -0.00813102722f, -1.f),
    C_Vertex3f(0.270122409f, 0.0581942797f, -1.f),
    C_Vertex3f(0.278278828f, -0.00388526917f, -1.f),
    C_Vertex3f(0.281356573f, 0.0621893406f, -1.f),
    C_Vertex3f(0.295287132f, 0.00248336792f, -1.f),
    C_Vertex3f(0.292154551f, 0.0675160885f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220902681f, 0.0555992126f, -1.f),
    C_Vertex3f(0.223101616f, -0.0102539063f, -1.f),
    C_Vertex3f(0.233529091f, 0.0541992188f, -1.f),
    C_Vertex3f(0.242704391f, -0.0102539063f, -1.f),
    C_Vertex3f(0.24634552f, 0.0541992188f, -1.f),
    C_Vertex3f(0.260751247f, -0.00813102722f, -1.f),
    C_Vertex3f(0.258452058f, 0.0555309057f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.208739996f, 0.0583992004f, -1.f),
    C_Vertex3f(0.197041035f, 0.0625991821f, -1.f),
    C_Vertex3f(0.203178883f, -0.00819265842f, -1.f),
    C_Vertex3f(0.223101616f, -0.0102539063f, -1.f),
    C_Vertex3f(0.220902681f, 0.0555992126f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164726496f, 0.0835990906f, -1.f),
    C_Vertex3f(0.165483475f, 0.0021135807f, -1.f),
    C_Vertex3f(0.175034285f, 0.0751991272f, -1.f),
    C_Vertex3f(0.183972836f, -0.00407016277f, -1.f),
    C_Vertex3f(0.185805798f, 0.0681991577f, -1.f),
    C_Vertex3f(0.203178883f, -0.00819265842f, -1.f),
    C_Vertex3f(0.197041035f, 0.0625991821f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.154882431f, 0.0933990479f, -1.f),
    C_Vertex3f(0.14550209f, 0.104598999f, -1.f),
    C_Vertex3f(0.1477108f, 0.010358572f, -1.f),
    C_Vertex3f(0.165483475f, 0.0021135807f, -1.f),
    C_Vertex3f(0.164726496f, 0.0835990906f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.137294292f, 0.117198944f, -1.f),
    C_Vertex3f(0.130259037f, 0.131198883f, -1.f),
    C_Vertex3f(0.130654812f, 0.0206648111f, -1.f),
    C_Vertex3f(0.1477108f, 0.010358572f, -1.f),
    C_Vertex3f(0.14550209f, 0.104598999f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130654812f, 0.0206648111f, -1.f),
    C_Vertex3f(0.130259037f, 0.131198883f, -1.f),
    C_Vertex3f(0.124396324f, 0.146598816f, -1.f),
    C_Vertex3f(0.119706154f, 0.163398743f, -1.f),
    C_Vertex3f(0.116188526f, 0.181598663f, -1.f),
    C_Vertex3f(0.11431551f, 0.0330322981f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116188526f, 0.181598663f, -1.f),
    C_Vertex3f(0.113843441f, 0.201198578f, -1.f),
    C_Vertex3f(0.11431551f, 0.0330322981f, -1.f),
    C_Vertex3f(0.112670898f, 0.222198486f, -1.f),
    C_Vertex3f(0.109574437f, 0.434430718f, -1.f),
    C_Vertex3f(0.112670898f, 0.244926453f, -1.f),
    C_Vertex3f(0.120411396f, 0.442422867f, -1.f),
    C_Vertex3f(0.113782048f, 0.266459465f, -1.f),
    C_Vertex3f(0.116004348f, 0.28646946f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0811583996f, 0.405416727f, -1.f),
    C_Vertex3f(0.0837869644f, 0.0639510155f, -1.f),
    C_Vertex3f(0.0899479389f, 0.415927649f, -1.f),
    C_Vertex3f(0.098692894f, 0.0474610329f, -1.f),
    C_Vertex3f(0.0994199514f, 0.425598979f, -1.f),
    C_Vertex3f(0.11431551f, 0.0330322981f, -1.f),
    C_Vertex3f(0.109574437f, 0.434430718f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0316162109f, 0.254020691f, -1.f),
    C_Vertex3f(0.0334794521f, 0.194996953f, -1.f),
    C_Vertex3f(0.0346460342f, 0.287162304f, -1.f),
    C_Vertex3f(0.0372059345f, 0.169333816f, -1.f),
    C_Vertex3f(0.0407056808f, 0.318502903f, -1.f),
    C_Vertex3f(0.0427956581f, 0.145199537f, -1.f),
    C_Vertex3f(0.0497951508f, 0.348042488f, -1.f),
    C_Vertex3f(0.0502486229f, 0.122594118f, -1.f),
    C_Vertex3f(0.0588846207f, 0.368846416f, -1.f),
    C_Vertex3f(0.0595648289f, 0.101517558f, -1.f),
    C_Vertex3f(0.0656267405f, 0.381876111f, -1.f),
    C_Vertex3f(0.070744276f, 0.0819698572f, -1.f),
    C_Vertex3f(0.0730513334f, 0.394066215f, -1.f),
    C_Vertex3f(0.0837869644f, 0.0639510155f, -1.f),
    C_Vertex3f(0.0811583996f, 0.405416727f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.362670898f, 0.402107239f, -1.f),
    C_Vertex3f(0.356501579f, 0.411170006f, -1.f),
    C_Vertex3f(0.359091759f, 0.280011892f, -1.f),
    C_Vertex3f(0.361477852f, 0.259715319f, -1.f),
    C_Vertex3f(0.362670898f, 0.237984657f, -1.f),
    C_Vertex3f(0.43347168f, -0.177734375f, -1.f),
    C_Vertex3f(0.362670898f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356501579f, 0.411170006f, -1.f),
    C_Vertex3f(0.343016624f, 0.427029848f, -1.f),
    C_Vertex3f(0.3447752f, 0.332297087f, -1.f),
    C_Vertex3f(0.350740433f, 0.316302776f, -1.f),
    C_Vertex3f(0.355512619f, 0.298874378f, -1.f),
    C_Vertex3f(0.359091759f, 0.280011892f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.299385071f, 0.390675068f, -1.f),
    C_Vertex3f(0.309761047f, 0.381906033f, -1.f),
    C_Vertex3f(0.312607765f, 0.451952457f, -1.f),
    C_Vertex3f(0.319721222f, 0.371675491f, -1.f),
    C_Vertex3f(0.328385353f, 0.440623999f, -1.f),
    C_Vertex3f(0.329265594f, 0.359983444f, -1.f),
    C_Vertex3f(0.343016624f, 0.427029848f, -1.f),
    C_Vertex3f(0.33761692f, 0.346857309f, -1.f),
    C_Vertex3f(0.3447752f, 0.332297087f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.265762329f, 0.408213139f, -1.f),
    C_Vertex3f(0.277385712f, 0.403828621f, -1.f),
    C_Vertex3f(0.27761364f, 0.4678123f, -1.f),
    C_Vertex3f(0.288593292f, 0.397982597f, -1.f),
    C_Vertex3f(0.295683861f, 0.461015224f, -1.f),
    C_Vertex3f(0.299385071f, 0.390675068f, -1.f),
    C_Vertex3f(0.312607765f, 0.451952457f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.258397102f, 0.472343683f, -1.f),
    C_Vertex3f(0.238034248f, 0.474609375f, -1.f),
    C_Vertex3f(0.241268158f, 0.412597656f, -1.f),
    C_Vertex3f(0.253723145f, 0.41113615f, -1.f),
    C_Vertex3f(0.265762329f, 0.408213139f, -1.f),
    C_Vertex3f(0.27761364f, 0.4678123f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193070173f, 0.404443741f, -1.f),
    C_Vertex3f(0.204379916f, 0.408520699f, -1.f),
    C_Vertex3f(0.213521957f, 0.474609375f, -1.f),
    C_Vertex3f(0.216173768f, 0.41123867f, -1.f),
    C_Vertex3f(0.238034248f, 0.474609375f, -1.f),
    C_Vertex3f(0.228451729f, 0.412597656f, -1.f),
    C_Vertex3f(0.241268158f, 0.412597656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.187125683f, 0.471033096f, -1.f),
    C_Vertex3f(0.161848545f, 0.463880539f, -1.f),
    C_Vertex3f(0.162045598f, 0.384058952f, -1.f),
    C_Vertex3f(0.171903014f, 0.392212868f, -1.f),
    C_Vertex3f(0.182244539f, 0.399007797f, -1.f),
    C_Vertex3f(0.193070173f, 0.404443741f, -1.f),
    C_Vertex3f(0.213521957f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129338145f, 0.337361336f, -1.f),
    C_Vertex3f(0.136005044f, 0.351279259f, -1.f),
    C_Vertex3f(0.137690544f, 0.453151703f, -1.f),
    C_Vertex3f(0.143783092f, 0.363674164f, -1.f),
    C_Vertex3f(0.161848545f, 0.463880539f, -1.f),
    C_Vertex3f(0.152672291f, 0.374546051f, -1.f),
    C_Vertex3f(0.162045598f, 0.384058952f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116004348f, 0.28646946f, -1.f),
    C_Vertex3f(0.119337797f, 0.304956436f, -1.f),
    C_Vertex3f(0.120411396f, 0.442422867f, -1.f),
    C_Vertex3f(0.123782396f, 0.321920395f, -1.f),
    C_Vertex3f(0.137690544f, 0.453151703f, -1.f),
    C_Vertex3f(0.129338145f, 0.337361336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.362670898f, 0.215530396f, -1.f),
    C_Vertex3f(0.361539245f, 0.195172906f, -1.f),
    C_Vertex3f(0.43347168f, -0.177734375f, -1.f),
    C_Vertex3f(0.362670898f, 0.237984657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.43347168f, -0.177734375f, -1.f),
    C_Vertex3f(0.354858398f, 0.049407959f, -1.f),
    C_Vertex3f(0.354858398f, -0.177734375f, -1.f),
    C_Vertex3f(0.362670898f, 0.464355469f, -1.f),
    C_Vertex3f(0.43347168f, -0.177734375f, -1.f),
    C_Vertex3f(0.43347168f, 0.464355469f, -1.f),
    C_Vertex3f(0.0334794521f, 0.194996953f, -1.f),
    C_Vertex3f(0.0316162109f, 0.254020691f, -1.f),
    C_Vertex3f(0.0316162109f, 0.22218895f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_114[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0581054688f, 0.f, 0.f),
    C_Vertex3f(0.0581054688f, 0.f, -1.f),
    C_Vertex3f(0.0581054688f, 0.464355469f, -1.f),
    C_Vertex3f(0.0581054688f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0581054688f, 0.464355469f, 0.f),
    C_Vertex3f(0.0581054688f, 0.464355469f, -1.f),
    C_Vertex3f(0.12890625f, 0.464355469f, -1.f),
    C_Vertex3f(0.12890625f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.12890625f, 0.464355469f, 0.f),
    C_Vertex3f(0.12890625f, 0.464355469f, -1.f),
    C_Vertex3f(0.12890625f, 0.393989563f, -1.f),
    C_Vertex3f(0.12890625f, 0.393989563f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.876370728f, 0.481637299f, 0.f),
    C_Vertex3f(0.12890625f, 0.393989563f, 0.f),
    C_Vertex3f(0.12890625f, 0.393989563f, -1.f),
    C_Normal3f(-0.866819322f, 0.498622358f, 0.f),
    C_Vertex3f(0.132288933f, 0.400144577f, 0.f),
    C_Vertex3f(0.132288933f, 0.400144577f, -1.f),
    C_Normal3f(-0.852947414f, 0.521996677f, 0.f),
    C_Vertex3f(0.138924837f, 0.411406159f, 0.f),
    C_Vertex3f(0.138924837f, 0.411406159f, -1.f),
    C_Normal3f(-0.832746387f, 0.553654671f, 0.f),
    C_Vertex3f(0.14543128f, 0.421619296f, 0.f),
    C_Vertex3f(0.14543128f, 0.421619296f, -1.f),
    C_Normal3f(-0.807477355f, 0.589898586f, 0.f),
    C_Vertex3f(0.151808262f, 0.430783987f, 0.f),
    C_Vertex3f(0.151808262f, 0.430783987f, -1.f),
    C_Normal3f(-0.775400102f, 0.631470203f, 0.f),
    C_Vertex3f(0.158055782f, 0.438900232f, 0.f),
    C_Vertex3f(0.158055782f, 0.438900232f, -1.f),
    C_Normal3f(-0.734070361f, 0.679073453f, 0.f),
    C_Vertex3f(0.164173841f, 0.445968032f, 0.f),
    C_Vertex3f(0.164173841f, 0.445968032f, -1.f),
    C_Normal3f(-0.680078626f, 0.733139157f, 0.f),
    C_Vertex3f(0.170162439f, 0.451987386f, 0.f),
    C_Vertex3f(0.170162439f, 0.451987386f, -1.f),
    C_Normal3f(-0.600508511f, 0.799618423f, 0.f),
    C_Vertex3f(0.176021576f, 0.456958294f, 0.f),
    C_Vertex3f(0.176021576f, 0.456958294f, -1.f),
    C_Normal3f(-0.495537668f, 0.868586481f, 0.f),
    C_Vertex3f(0.184616089f, 0.462841988f, 0.f),
    C_Vertex3f(0.184616089f, 0.462841988f, -1.f),
    C_Normal3f(-0.369830549f, 0.929099202f, 0.f),
    C_Vertex3f(0.196647644f, 0.468725681f, 0.f),
    C_Vertex3f(0.196647644f, 0.468725681f, -1.f),
    C_Normal3f(-0.222507983f, 0.974930823f, 0.f),
    C_Vertex3f(0.209251404f, 0.472648144f, 0.f),
    C_Vertex3f(0.209251404f, 0.472648144f, -1.f),
    C_Normal3f(-0.0652749613f, 0.997867286f, 0.f),
    C_Vertex3f(0.222427368f, 0.474609375f, 0.f),
    C_Vertex3f(0.222427368f, 0.474609375f, -1.f),
    C_Normal3f(0.0849306211f, 0.996386826f, 0.f),
    C_Vertex3f(0.239233017f, 0.474609375f, 0.f),
    C_Vertex3f(0.239233017f, 0.474609375f, -1.f),
    C_Normal3f(0.228030235f, 0.973654032f, 0.f),
    C_Vertex3f(0.259259224f, 0.471469879f, 0.f),
    C_Vertex3f(0.259259224f, 0.471469879f, -1.f),
    C_Normal3f(0.361072868f, 0.932537556f, 0.f),
    C_Vertex3f(0.279448509f, 0.465190887f, 0.f),
    C_Vertex3f(0.279448509f, 0.465190887f, -1.f),
    C_Normal3f(0.45631662f, 0.889817417f, 0.f),
    C_Vertex3f(0.299800873f, 0.4557724f, 0.f),
    C_Vertex3f(0.299800873f, 0.4557724f, -1.f),
    C_Normal3f(0.522078514f, 0.852897406f, 0.f),
    C_Vertex3f(0.310058594f, 0.449493408f, 0.f),
    C_Vertex3f(0.310058594f, 0.449493408f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.935289741f, -0.353882819f, 0.f),
    C_Vertex3f(0.310058594f, 0.449493408f, 0.f),
    C_Vertex3f(0.310058594f, 0.449493408f, -1.f),
    C_Vertex3f(0.282516479f, 0.376701355f, -1.f),
    C_Vertex3f(0.282516479f, 0.376701355f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.506248713f, -0.862387478f, 0.f),
    C_Vertex3f(0.282516479f, 0.376701355f, 0.f),
    C_Vertex3f(0.282516479f, 0.376701355f, -1.f),
    C_Normal3f(-0.43944633f, -0.898268878f, 0.f),
    C_Vertex3f(0.275339127f, 0.380914688f, 0.f),
    C_Vertex3f(0.275339127f, 0.380914688f, -1.f),
    C_Normal3f(-0.34447366f, -0.938795984f, 0.f),
    C_Vertex3f(0.260985374f, 0.387234688f, 0.f),
    C_Vertex3f(0.260985374f, 0.387234688f, -1.f),
    C_Normal3f(-0.215023324f, -0.976608872f, 0.f),
    C_Vertex3f(0.246632576f, 0.391448021f, 0.f),
    C_Vertex3f(0.246632576f, 0.391448021f, -1.f),
    C_Normal3f(-0.0751765817f, -0.997170269f, 0.f),
    C_Vertex3f(0.232280731f, 0.393554688f, 0.f),
    C_Vertex3f(0.232280731f, 0.393554688f, -1.f),
    C_Normal3f(0.075073868f, -0.997177958f, 0.f),
    C_Vertex3f(0.218688965f, 0.393554688f, 0.f),
    C_Vertex3f(0.218688965f, 0.393554688f, -1.f),
    C_Normal3f(0.238447502f, -0.971155405f, 0.f),
    C_Vertex3f(0.206509113f, 0.391614437f, 0.f),
    C_Vertex3f(0.206509113f, 0.391614437f, -1.f),
    C_Normal3f(0.397400796f, -0.917645216f, 0.f),
    C_Vertex3f(0.194982052f, 0.387733936f, 0.f),
    C_Vertex3f(0.194982052f, 0.387733936f, -1.f),
    C_Normal3f(0.541327059f, -0.840812087f, 0.f),
    C_Vertex3f(0.18410778f, 0.381913185f, 0.f),
    C_Vertex3f(0.18410778f, 0.381913185f, -1.f),
    C_Normal3f(0.667747676f, -0.744387627f, 0.f),
    C_Vertex3f(0.173886299f, 0.374152184f, 0.f),
    C_Vertex3f(0.173886299f, 0.374152184f, -1.f),
    C_Normal3f(0.780359268f, -0.625331342f, 0.f),
    C_Vertex3f(0.165128231f, 0.364887714f, 0.f),
    C_Vertex3f(0.165128231f, 0.364887714f, -1.f),
    C_Normal3f(0.868883133f, -0.495017201f, 0.f),
    C_Vertex3f(0.157833576f, 0.354119778f, 0.f),
    C_Vertex3f(0.157833576f, 0.354119778f, -1.f),
    C_Normal3f(0.93455416f, -0.355820984f, 0.f),
    C_Vertex3f(0.152002335f, 0.341848373f, 0.f),
    C_Vertex3f(0.152002335f, 0.341848373f, -1.f),
    C_Normal3f(0.966984987f, -0.254833549f, 0.f),
    C_Vertex3f(0.146543503f, 0.324466705f, 0.f),
    C_Vertex3f(0.146543503f, 0.324466705f, -1.f),
    C_Normal3f(0.983799279f, -0.179273158f, 0.f),
    C_Vertex3f(0.141631126f, 0.302494049f, 0.f),
    C_Vertex3f(0.141631126f, 0.302494049f, -1.f),
    C_Normal3f(0.994558811f, -0.104176275f, 0.f),
    C_Vertex3f(0.138356209f, 0.279537201f, 0.f),
    C_Vertex3f(0.138356209f, 0.279537201f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.13671875f, 0.255596161f, 0.f),
    C_Vertex3f(0.13671875f, 0.255596161f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.13671875f, 0.255596161f, 0.f),
    C_Vertex3f(0.13671875f, 0.255596161f, -1.f),
    C_Vertex3f(0.13671875f, 0.f, -1.f),
    C_Vertex3f(0.13671875f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.13671875f, 0.f, 0.f),
    C_Vertex3f(0.13671875f, 0.f, -1.f),
    C_Vertex3f(0.0581054688f, 0.f, -1.f),
    C_Vertex3f(0.0581054688f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.260985374f, 0.387234688f, 0.f),
    C_Vertex3f(0.275339127f, 0.380914688f, 0.f),
    C_Vertex3f(0.279448509f, 0.465190887f, 0.f),
    C_Vertex3f(0.282516479f, 0.376701355f, 0.f),
    C_Vertex3f(0.299800873f, 0.4557724f, 0.f),
    C_Vertex3f(0.310058594f, 0.449493408f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164173841f, 0.445968032f, 0.f),
    C_Vertex3f(0.165128231f, 0.364887714f, 0.f),
    C_Vertex3f(0.170162439f, 0.451987386f, 0.f),
    C_Vertex3f(0.173886299f, 0.374152184f, 0.f),
    C_Vertex3f(0.176021576f, 0.456958294f, 0.f),
    C_Vertex3f(0.18410778f, 0.381913185f, 0.f),
    C_Vertex3f(0.184616089f, 0.462841988f, 0.f),
    C_Vertex3f(0.194982052f, 0.387733936f, 0.f),
    C_Vertex3f(0.196647644f, 0.468725681f, 0.f),
    C_Vertex3f(0.206509113f, 0.391614437f, 0.f),
    C_Vertex3f(0.209251404f, 0.472648144f, 0.f),
    C_Vertex3f(0.218688965f, 0.393554688f, 0.f),
    C_Vertex3f(0.222427368f, 0.474609375f, 0.f),
    C_Vertex3f(0.232280731f, 0.393554688f, 0.f),
    C_Vertex3f(0.239233017f, 0.474609375f, 0.f),
    C_Vertex3f(0.246632576f, 0.391448021f, 0.f),
    C_Vertex3f(0.259259224f, 0.471469879f, 0.f),
    C_Vertex3f(0.260985374f, 0.387234688f, 0.f),
    C_Vertex3f(0.279448509f, 0.465190887f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158055782f, 0.438900232f, 0.f),
    C_Vertex3f(0.151808262f, 0.430783987f, 0.f),
    C_Vertex3f(0.152002335f, 0.341848373f, 0.f),
    C_Vertex3f(0.157833576f, 0.354119778f, 0.f),
    C_Vertex3f(0.165128231f, 0.364887714f, 0.f),
    C_Vertex3f(0.164173841f, 0.445968032f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13671875f, 0.255596161f, 0.f),
    C_Vertex3f(0.138356209f, 0.279537201f, 0.f),
    C_Vertex3f(0.138924837f, 0.411406159f, 0.f),
    C_Vertex3f(0.141631126f, 0.302494049f, 0.f),
    C_Vertex3f(0.14543128f, 0.421619296f, 0.f),
    C_Vertex3f(0.146543503f, 0.324466705f, 0.f),
    C_Vertex3f(0.151808262f, 0.430783987f, 0.f),
    C_Vertex3f(0.152002335f, 0.341848373f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12890625f, 0.464355469f, 0.f),
    C_Vertex3f(0.0581054688f, 0.464355469f, 0.f),
    C_Vertex3f(0.12890625f, 0.393989563f, 0.f),
    C_Vertex3f(0.13671875f, 0.f, 0.f),
    C_Vertex3f(0.132288933f, 0.400144577f, 0.f),
    C_Vertex3f(0.13671875f, 0.255596161f, 0.f),
    C_Vertex3f(0.138924837f, 0.411406159f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.13671875f, 0.f, 0.f),
    C_Vertex3f(0.0581054688f, 0.464355469f, 0.f),
    C_Vertex3f(0.0581054688f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.260985374f, 0.387234688f, -1.f),
    C_Vertex3f(0.275339127f, 0.380914688f, -1.f),
    C_Vertex3f(0.279448509f, 0.465190887f, -1.f),
    C_Vertex3f(0.282516479f, 0.376701355f, -1.f),
    C_Vertex3f(0.299800873f, 0.4557724f, -1.f),
    C_Vertex3f(0.310058594f, 0.449493408f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164173841f, 0.445968032f, -1.f),
    C_Vertex3f(0.165128231f, 0.364887714f, -1.f),
    C_Vertex3f(0.170162439f, 0.451987386f, -1.f),
    C_Vertex3f(0.173886299f, 0.374152184f, -1.f),
    C_Vertex3f(0.176021576f, 0.456958294f, -1.f),
    C_Vertex3f(0.18410778f, 0.381913185f, -1.f),
    C_Vertex3f(0.184616089f, 0.462841988f, -1.f),
    C_Vertex3f(0.194982052f, 0.387733936f, -1.f),
    C_Vertex3f(0.196647644f, 0.468725681f, -1.f),
    C_Vertex3f(0.206509113f, 0.391614437f, -1.f),
    C_Vertex3f(0.209251404f, 0.472648144f, -1.f),
    C_Vertex3f(0.218688965f, 0.393554688f, -1.f),
    C_Vertex3f(0.222427368f, 0.474609375f, -1.f),
    C_Vertex3f(0.232280731f, 0.393554688f, -1.f),
    C_Vertex3f(0.239233017f, 0.474609375f, -1.f),
    C_Vertex3f(0.246632576f, 0.391448021f, -1.f),
    C_Vertex3f(0.259259224f, 0.471469879f, -1.f),
    C_Vertex3f(0.260985374f, 0.387234688f, -1.f),
    C_Vertex3f(0.279448509f, 0.465190887f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158055782f, 0.438900232f, -1.f),
    C_Vertex3f(0.151808262f, 0.430783987f, -1.f),
    C_Vertex3f(0.152002335f, 0.341848373f, -1.f),
    C_Vertex3f(0.157833576f, 0.354119778f, -1.f),
    C_Vertex3f(0.165128231f, 0.364887714f, -1.f),
    C_Vertex3f(0.164173841f, 0.445968032f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13671875f, 0.255596161f, -1.f),
    C_Vertex3f(0.138356209f, 0.279537201f, -1.f),
    C_Vertex3f(0.138924837f, 0.411406159f, -1.f),
    C_Vertex3f(0.141631126f, 0.302494049f, -1.f),
    C_Vertex3f(0.14543128f, 0.421619296f, -1.f),
    C_Vertex3f(0.146543503f, 0.324466705f, -1.f),
    C_Vertex3f(0.151808262f, 0.430783987f, -1.f),
    C_Vertex3f(0.152002335f, 0.341848373f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12890625f, 0.464355469f, -1.f),
    C_Vertex3f(0.0581054688f, 0.464355469f, -1.f),
    C_Vertex3f(0.12890625f, 0.393989563f, -1.f),
    C_Vertex3f(0.13671875f, 0.f, -1.f),
    C_Vertex3f(0.132288933f, 0.400144577f, -1.f),
    C_Vertex3f(0.13671875f, 0.255596161f, -1.f),
    C_Vertex3f(0.138924837f, 0.411406159f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.13671875f, 0.f, -1.f),
    C_Vertex3f(0.0581054688f, 0.464355469f, -1.f),
    C_Vertex3f(0.0581054688f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_115[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.155324444f, 0.987863481f, 0.f),
    C_Vertex3f(0.02734375f, 0.138671875f, 0.f),
    C_Vertex3f(0.02734375f, 0.138671875f, -1.f),
    C_Vertex3f(0.104980469f, 0.150878906f, -1.f),
    C_Vertex3f(0.104980469f, 0.150878906f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.990272939f, 0.139138594f, 0.f),
    C_Vertex3f(0.104980469f, 0.150878906f, 0.f),
    C_Vertex3f(0.104980469f, 0.150878906f, -1.f),
    C_Normal3f(0.982487202f, 0.186330065f, 0.f),
    C_Vertex3f(0.105802536f, 0.145028114f, 0.f),
    C_Vertex3f(0.105802536f, 0.145028114f, -1.f),
    C_Normal3f(0.968530297f, 0.248895541f, 0.f),
    C_Vertex3f(0.108179688f, 0.134009957f, 0.f),
    C_Vertex3f(0.108179688f, 0.134009957f, -1.f),
    C_Normal3f(0.944471955f, 0.328592092f, 0.f),
    C_Vertex3f(0.111289859f, 0.123675227f, 0.f),
    C_Vertex3f(0.111289859f, 0.123675227f, -1.f),
    C_Normal3f(0.911173463f, 0.412022918f, 0.f),
    C_Vertex3f(0.115133047f, 0.114023924f, 0.f),
    C_Vertex3f(0.115133047f, 0.114023924f, -1.f),
    C_Normal3f(0.867658496f, 0.497160673f, 0.f),
    C_Vertex3f(0.119709253f, 0.105056047f, 0.f),
    C_Vertex3f(0.119709253f, 0.105056047f, -1.f),
    C_Normal3f(0.813619614f, 0.581397593f, 0.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, 0.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, -1.f),
    C_Normal3f(0.749659598f, 0.661823511f, 0.f),
    C_Vertex3f(0.131060719f, 0.0891705751f, 0.f),
    C_Vertex3f(0.131060719f, 0.0891705751f, -1.f),
    C_Normal3f(0.677360773f, 0.735650957f, 0.f),
    C_Vertex3f(0.137835979f, 0.0822529793f, 0.f),
    C_Vertex3f(0.137835979f, 0.0822529793f, -1.f),
    C_Normal3f(0.59564364f, 0.803248763f, 0.f),
    C_Vertex3f(0.145344257f, 0.0760188103f, 0.f),
    C_Vertex3f(0.145344257f, 0.0760188103f, -1.f),
    C_Normal3f(0.506266713f, 0.862376928f, 0.f),
    C_Vertex3f(0.153599143f, 0.0705639124f, 0.f),
    C_Vertex3f(0.153599143f, 0.0705639124f, -1.f),
    C_Normal3f(0.415789396f, 0.909460902f, 0.f),
    C_Vertex3f(0.162600636f, 0.0658882856f, 0.f),
    C_Vertex3f(0.162600636f, 0.0658882856f, -1.f),
    C_Normal3f(0.327373713f, 0.94489491f, 0.f),
    C_Vertex3f(0.172348738f, 0.06199193f, 0.f),
    C_Vertex3f(0.172348738f, 0.06199193f, -1.f),
    C_Normal3f(0.243412912f, 0.969922721f, 0.f),
    C_Vertex3f(0.182843447f, 0.0588748455f, 0.f),
    C_Vertex3f(0.182843447f, 0.0588748455f, -1.f),
    C_Normal3f(0.165423974f, 0.986222446f, 0.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, 0.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, -1.f),
    C_Normal3f(0.0941423625f, 0.995558739f, 0.f),
    C_Vertex3f(0.206072688f, 0.0549784899f, 0.f),
    C_Vertex3f(0.206072688f, 0.0549784899f, -1.f),
    C_Normal3f(0.0296508372f, 0.999560237f, 0.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, 0.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0263045318f, 0.999653935f, 0.f),
    C_Vertex3f(0.23234272f, 0.0541992188f, 0.f),
    C_Vertex3f(0.23234272f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0841049999f, 0.996456981f, 0.f),
    C_Vertex3f(0.245055914f, 0.0548899174f, 0.f),
    C_Vertex3f(0.245055914f, 0.0548899174f, -1.f),
    C_Normal3f(-0.149783954f, 0.988718808f, 0.f),
    C_Vertex3f(0.256892443f, 0.0562713146f, 0.f),
    C_Vertex3f(0.256892443f, 0.0562713146f, -1.f),
    C_Normal3f(-0.223927125f, 0.974605858f, 0.f),
    C_Vertex3f(0.267852306f, 0.0583434105f, 0.f),
    C_Vertex3f(0.267852306f, 0.0583434105f, -1.f),
    C_Normal3f(-0.306725383f, 0.951798081f, 0.f),
    C_Vertex3f(0.277935505f, 0.061106205f, 0.f),
    C_Vertex3f(0.277935505f, 0.061106205f, -1.f),
    C_Normal3f(-0.397544771f, 0.91758281f, 0.f),
    C_Vertex3f(0.287142038f, 0.0645596981f, 0.f),
    C_Vertex3f(0.287142038f, 0.0645596981f, -1.f),
    C_Normal3f(-0.49448508f, 0.869186103f, 0.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, 0.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, -1.f),
    C_Normal3f(-0.603962064f, 0.797013104f, 0.f),
    C_Vertex3f(0.30292511f, 0.0735387802f, 0.f),
    C_Vertex3f(0.30292511f, 0.0735387802f, -1.f),
    C_Normal3f(-0.717075408f, 0.696995676f, 0.f),
    C_Vertex3f(0.312789917f, 0.0818271637f, 0.f),
    C_Vertex3f(0.312789917f, 0.0818271637f, -1.f),
    C_Normal3f(-0.835284114f, 0.549818575f, 0.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, 0.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, -1.f),
    C_Normal3f(-0.938837171f, 0.344361275f, 0.f),
    C_Vertex3f(0.329231262f, 0.106804848f, 0.f),
    C_Vertex3f(0.329231262f, 0.106804848f, -1.f),
    C_Normal3f(-0.993195236f, 0.116461053f, 0.f),
    C_Vertex3f(0.332519531f, 0.120731354f, 0.f),
    C_Vertex3f(0.332519531f, 0.120731354f, -1.f),
    C_Normal3f(-0.993828118f, -0.110930823f, 0.f),
    C_Vertex3f(0.332519531f, 0.134847641f, 0.f),
    C_Vertex3f(0.332519531f, 0.134847641f, -1.f),
    C_Normal3f(-0.932035148f, -0.362367868f, 0.f),
    C_Vertex3f(0.329612732f, 0.146773338f, 0.f),
    C_Vertex3f(0.329612732f, 0.146773338f, -1.f),
    C_Normal3f(-0.80304426f, -0.595919311f, 0.f),
    C_Vertex3f(0.323799133f, 0.157277107f, 0.f),
    C_Vertex3f(0.323799133f, 0.157277107f, -1.f),
    C_Normal3f(-0.651524901f, -0.758627295f, 0.f),
    C_Vertex3f(0.315078735f, 0.166358948f, 0.f),
    C_Vertex3f(0.315078735f, 0.166358948f, -1.f),
    C_Normal3f(-0.509991109f, -0.860179663f, 0.f),
    C_Vertex3f(0.307235718f, 0.171502113f, 0.f),
    C_Vertex3f(0.307235718f, 0.171502113f, -1.f),
    C_Normal3f(-0.41662693f, -0.909077585f, 0.f),
    C_Vertex3f(0.301674128f, 0.174306393f, 0.f),
    C_Vertex3f(0.301674128f, 0.174306393f, -1.f),
    C_Normal3f(-0.365715176f, -0.930726826f, 0.f),
    C_Vertex3f(0.294609785f, 0.177288532f, 0.f),
    C_Vertex3f(0.294609785f, 0.177288532f, -1.f),
    C_Normal3f(-0.329218626f, -0.944253683f, 0.f),
    C_Vertex3f(0.28604269f, 0.180448532f, 0.f),
    C_Vertex3f(0.28604269f, 0.180448532f, -1.f),
    C_Normal3f(-0.301897258f, -0.953340411f, 0.f),
    C_Vertex3f(0.275972843f, 0.183786392f, 0.f),
    C_Vertex3f(0.275972843f, 0.183786392f, -1.f),
    C_Normal3f(-0.280728996f, -0.95978713f, 0.f),
    C_Vertex3f(0.264400244f, 0.187302113f, 0.f),
    C_Vertex3f(0.264400244f, 0.187302113f, -1.f),
    C_Normal3f(-0.263869554f, -0.964558363f, 0.f),
    C_Vertex3f(0.251324892f, 0.190995693f, 0.f),
    C_Vertex3f(0.251324892f, 0.190995693f, -1.f),
    C_Normal3f(-0.249712259f, -0.968320072f, 0.f),
    C_Vertex3f(0.236746788f, 0.194867134f, 0.f),
    C_Vertex3f(0.236746788f, 0.194867134f, -1.f),
    C_Normal3f(-0.249843717f, -0.968286216f, 0.f),
    C_Vertex3f(0.217877388f, 0.199621201f, 0.f),
    C_Vertex3f(0.217877388f, 0.199621201f, -1.f),
    C_Normal3f(-0.261011988f, -0.965335548f, 0.f),
    C_Vertex3f(0.197880268f, 0.204895735f, 0.f),
    C_Vertex3f(0.197880268f, 0.204895735f, -1.f),
    C_Normal3f(-0.274538636f, -0.961576045f, 0.f),
    C_Vertex3f(0.179543972f, 0.209985971f, 0.f),
    C_Vertex3f(0.179543972f, 0.209985971f, -1.f),
    C_Normal3f(-0.290685207f, -0.956818759f, 0.f),
    C_Vertex3f(0.1628685f, 0.214891911f, 0.f),
    C_Vertex3f(0.1628685f, 0.214891911f, -1.f),
    C_Normal3f(-0.310274899f, -0.950646877f, 0.f),
    C_Vertex3f(0.147853851f, 0.219613552f, 0.f),
    C_Vertex3f(0.147853851f, 0.219613552f, -1.f),
    C_Normal3f(-0.334503651f, -0.942394435f, 0.f),
    C_Vertex3f(0.134500027f, 0.224150896f, 0.f),
    C_Vertex3f(0.134500027f, 0.224150896f, -1.f),
    C_Normal3f(-0.365166187f, -0.930942416f, 0.f),
    C_Vertex3f(0.122807026f, 0.228503942f, 0.f),
    C_Vertex3f(0.122807026f, 0.228503942f, -1.f),
    C_Normal3f(-0.40970242f, -0.912219226f, 0.f),
    C_Vertex3f(0.112774849f, 0.232672691f, 0.f),
    C_Vertex3f(0.112774849f, 0.232672691f, -1.f),
    C_Normal3f(-0.493446022f, -0.869776368f, 0.f),
    C_Vertex3f(0.100217819f, 0.238649368f, 0.f),
    C_Vertex3f(0.100217819f, 0.238649368f, -1.f),
    C_Normal3f(-0.599326909f, -0.800504327f, 0.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, 0.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, -1.f),
    C_Normal3f(-0.710367441f, -0.703831077f, 0.f),
    C_Vertex3f(0.0728063583f, 0.259171963f, 0.f),
    C_Vertex3f(0.0728063583f, 0.259171963f, -1.f),
    C_Normal3f(-0.809581339f, -0.587007701f, 0.f),
    C_Vertex3f(0.0621376038f, 0.271725655f, 0.f),
    C_Vertex3f(0.0621376038f, 0.271725655f, -1.f),
    C_Normal3f(-0.885973155f, -0.463736624f, 0.f),
    C_Vertex3f(0.0534934998f, 0.28580761f, 0.f),
    C_Vertex3f(0.0534934998f, 0.28580761f, -1.f),
    C_Normal3f(-0.942161143f, -0.335160226f, 0.f),
    C_Vertex3f(0.0470104218f, 0.300626278f, 0.f),
    C_Vertex3f(0.0470104218f, 0.300626278f, -1.f),
    C_Normal3f(-0.979902864f, -0.199475348f, 0.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, 0.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, -1.f),
    C_Normal3f(-0.997804403f, -0.0662289932f, 0.f),
    C_Vertex3f(0.0405273438f, 0.332473755f, 0.f),
    C_Vertex3f(0.0405273438f, 0.332473755f, -1.f),
    C_Normal3f(-0.998378634f, 0.0569228902f, 0.f),
    C_Vertex3f(0.0405273438f, 0.348739624f, 0.f),
    C_Vertex3f(0.0405273438f, 0.348739624f, -1.f),
    C_Normal3f(-0.983803272f, 0.179251954f, 0.f),
    C_Vertex3f(0.0423059464f, 0.363668919f, 0.f),
    C_Vertex3f(0.0423059464f, 0.363668919f, -1.f),
    C_Normal3f(-0.953510284f, 0.30136025f, 0.f),
    C_Vertex3f(0.0458631516f, 0.378024578f, 0.f),
    C_Vertex3f(0.0458631516f, 0.378024578f, -1.f),
    C_Normal3f(-0.908047616f, 0.418866903f, 0.f),
    C_Vertex3f(0.0511989594f, 0.391806602f, 0.f),
    C_Vertex3f(0.0511989594f, 0.391806602f, -1.f),
    C_Normal3f(-0.852469385f, 0.5227772f, 0.f),
    C_Vertex3f(0.0583133698f, 0.405014992f, 0.f),
    C_Vertex3f(0.0583133698f, 0.405014992f, -1.f),
    C_Normal3f(-0.786895573f, 0.617086232f, 0.f),
    C_Vertex3f(0.0667133331f, 0.417105198f, 0.f),
    C_Vertex3f(0.0667133331f, 0.417105198f, -1.f),
    C_Normal3f(-0.709986389f, 0.704215467f, 0.f),
    C_Vertex3f(0.0763988495f, 0.428077221f, 0.f),
    C_Vertex3f(0.0763988495f, 0.428077221f, -1.f),
    C_Normal3f(-0.63944602f, 0.768836021f, 0.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, 0.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, -1.f),
    C_Normal3f(-0.585769415f, 0.810477734f, 0.f),
    C_Vertex3f(0.09349823f, 0.442298889f, 0.f),
    C_Vertex3f(0.09349823f, 0.442298889f, -1.f),
    C_Normal3f(-0.528840721f, 0.848721147f, 0.f),
    C_Vertex3f(0.0980949402f, 0.445682526f, 0.f),
    C_Vertex3f(0.0980949402f, 0.445682526f, -1.f),
    C_Normal3f(-0.455799788f, 0.8900823f, 0.f),
    C_Vertex3f(0.108957291f, 0.451931477f, 0.f),
    C_Vertex3f(0.108957291f, 0.451931477f, -1.f),
    C_Normal3f(-0.378850102f, 0.925458074f, 0.f),
    C_Vertex3f(0.121488571f, 0.457662106f, 0.f),
    C_Vertex3f(0.121488571f, 0.457662106f, -1.f),
    C_Normal3f(-0.312904745f, 0.949784517f, 0.f),
    C_Vertex3f(0.135688782f, 0.462874413f, 0.f),
    C_Vertex3f(0.135688782f, 0.462874413f, -1.f),
    C_Normal3f(-0.246383533f, 0.969172418f, 0.f),
    C_Vertex3f(0.151557922f, 0.467568398f, 0.f),
    C_Vertex3f(0.151557922f, 0.467568398f, -1.f),
    C_Normal3f(-0.172719449f, 0.984971046f, 0.f),
    C_Vertex3f(0.168001175f, 0.471088886f, 0.f),
    C_Vertex3f(0.168001175f, 0.471088886f, -1.f),
    C_Normal3f(-0.101200007f, 0.994866133f, 0.f),
    C_Vertex3f(0.185018539f, 0.473435879f, 0.f),
    C_Vertex3f(0.185018539f, 0.473435879f, -1.f),
    C_Normal3f(-0.0290665496f, 0.999577463f, 0.f),
    C_Vertex3f(0.202610016f, 0.474609375f, 0.f),
    C_Vertex3f(0.202610016f, 0.474609375f, -1.f),
    C_Normal3f(0.0405974723f, 0.999175668f, 0.f),
    C_Vertex3f(0.225374222f, 0.474609375f, 0.f),
    C_Vertex3f(0.225374222f, 0.474609375f, -1.f),
    C_Normal3f(0.117976762f, 0.993016362f, 0.f),
    C_Vertex3f(0.251067162f, 0.472640514f, 0.f),
    C_Vertex3f(0.251067162f, 0.472640514f, -1.f),
    C_Normal3f(0.207643896f, 0.978204429f, 0.f),
    C_Vertex3f(0.275090218f, 0.468702793f, 0.f),
    C_Vertex3f(0.275090218f, 0.468702793f, -1.f),
    C_Normal3f(0.304983616f, 0.95235759f, 0.f),
    C_Vertex3f(0.29744339f, 0.462796211f, 0.f),
    C_Vertex3f(0.29744339f, 0.462796211f, -1.f),
    C_Normal3f(0.405418575f, 0.914131165f, 0.f),
    C_Vertex3f(0.318126678f, 0.454920769f, 0.f),
    C_Vertex3f(0.318126678f, 0.454920769f, -1.f),
    C_Normal3f(0.51426959f, 0.857628584f, 0.f),
    C_Vertex3f(0.336102009f, 0.445651054f, 0.f),
    C_Vertex3f(0.336102009f, 0.445651054f, -1.f),
    C_Normal3f(0.632485092f, 0.774572551f, 0.f),
    C_Vertex3f(0.351369381f, 0.434987068f, 0.f),
    C_Vertex3f(0.351369381f, 0.434987068f, -1.f),
    C_Normal3f(0.751277328f, 0.659986556f, 0.f),
    C_Vertex3f(0.363928795f, 0.42292881f, 0.f),
    C_Vertex3f(0.363928795f, 0.42292881f, -1.f),
    C_Normal3f(0.848835111f, 0.528657675f, 0.f),
    C_Vertex3f(0.373780251f, 0.40947628f, 0.f),
    C_Vertex3f(0.373780251f, 0.40947628f, -1.f),
    C_Normal3f(0.91259867f, 0.408856481f, 0.f),
    C_Vertex3f(0.382098675f, 0.393754482f, 0.f),
    C_Vertex3f(0.382098675f, 0.393754482f, -1.f),
    C_Normal3f(0.95388037f, 0.300187111f, 0.f),
    C_Vertex3f(0.388884068f, 0.375763416f, 0.f),
    C_Vertex3f(0.388884068f, 0.375763416f, -1.f),
    C_Normal3f(0.97548455f, 0.220067948f, 0.f),
    C_Vertex3f(0.394136429f, 0.355503082f, 0.f),
    C_Vertex3f(0.394136429f, 0.355503082f, -1.f),
    C_Normal3f(0.986645639f, 0.162881732f, 0.f),
    C_Vertex3f(0.395996094f, 0.344238281f, 0.f),
    C_Vertex3f(0.395996094f, 0.344238281f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.131752744f, -0.991282582f, 0.f),
    C_Vertex3f(0.395996094f, 0.344238281f, 0.f),
    C_Vertex3f(0.395996094f, 0.344238281f, -1.f),
    C_Vertex3f(0.318847656f, 0.333984375f, -1.f),
    C_Vertex3f(0.318847656f, 0.333984375f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.989477277f, -0.144688696f, 0.f),
    C_Vertex3f(0.318847656f, 0.333984375f, 0.f),
    C_Vertex3f(0.318847656f, 0.333984375f, -1.f),
    C_Normal3f(-0.968403697f, -0.249387681f, 0.f),
    C_Vertex3f(0.3175354f, 0.34295845f, 0.f),
    C_Vertex3f(0.3175354f, 0.34295845f, -1.f),
    C_Normal3f(-0.920641422f, -0.390409231f, 0.f),
    C_Vertex3f(0.312421799f, 0.358936787f, 0.f),
    C_Vertex3f(0.312421799f, 0.358936787f, -1.f),
    C_Normal3f(-0.827186882f, -0.561926782f, 0.f),
    C_Vertex3f(0.304819107f, 0.372945309f, 0.f),
    C_Vertex3f(0.304819107f, 0.372945309f, -1.f),
    C_Normal3f(-0.6981287f, -0.715972304f, 0.f),
    C_Vertex3f(0.294727325f, 0.384984016f, 0.f),
    C_Vertex3f(0.294727325f, 0.384984016f, -1.f),
    C_Normal3f(-0.535922647f, -0.84426707f, 0.f),
    C_Vertex3f(0.282146454f, 0.39505291f, 0.f),
    C_Vertex3f(0.282146454f, 0.39505291f, -1.f),
    C_Normal3f(-0.356808603f, -0.934177518f, 0.f),
    C_Vertex3f(0.266968727f, 0.40260458f, 0.f),
    C_Vertex3f(0.266968727f, 0.40260458f, -1.f),
    C_Normal3f(-0.194198593f, -0.980962276f, 0.f),
    C_Vertex3f(0.249194145f, 0.407639027f, 0.f),
    C_Vertex3f(0.249194145f, 0.407639027f, -1.f),
    C_Normal3f(-0.0650108978f, -0.997884631f, 0.f),
    C_Vertex3f(0.228822708f, 0.41015625f, 0.f),
    C_Vertex3f(0.228822708f, 0.41015625f, -1.f),
    C_Normal3f(0.0181715135f, -0.999834895f, 0.f),
    C_Vertex3f(0.21055603f, 0.41015625f, 0.f),
    C_Vertex3f(0.21055603f, 0.41015625f, -1.f),
    C_Normal3f(0.069313921f, -0.997594953f, 0.f),
    C_Vertex3f(0.197961926f, 0.40959537f, 0.f),
    C_Vertex3f(0.197961926f, 0.40959537f, -1.f),
    C_Normal3f(0.124911025f, -0.99216789f, 0.f),
    C_Vertex3f(0.186338782f, 0.408473611f, 0.f),
    C_Vertex3f(0.186338782f, 0.408473611f, -1.f),
    C_Normal3f(0.189587131f, -0.981863916f, 0.f),
    C_Vertex3f(0.175686598f, 0.406790972f, 0.f),
    C_Vertex3f(0.175686598f, 0.406790972f, -1.f),
    C_Normal3f(0.264681518f, -0.964335859f, 0.f),
    C_Vertex3f(0.166005373f, 0.404547453f, 0.f),
    C_Vertex3f(0.166005373f, 0.404547453f, -1.f),
    C_Normal3f(0.351177722f, -0.936308861f, 0.f),
    C_Vertex3f(0.157295108f, 0.401743054f, 0.f),
    C_Vertex3f(0.157295108f, 0.401743054f, -1.f),
    C_Normal3f(0.449065775f, -0.893498778f, 0.f),
    C_Vertex3f(0.149555802f, 0.398377776f, 0.f),
    C_Vertex3f(0.149555802f, 0.398377776f, -1.f),
    C_Normal3f(0.567429543f, -0.823421896f, 0.f),
    C_Vertex3f(0.142787457f, 0.394451618f, 0.f),
    C_Vertex3f(0.142787457f, 0.394451618f, -1.f),
    C_Normal3f(0.687615454f, -0.726075113f, 0.f),
    C_Vertex3f(0.134091377f, 0.387721062f, 0.f),
    C_Vertex3f(0.134091377f, 0.387721062f, -1.f),
    C_Normal3f(0.813137829f, -0.582071185f, 0.f),
    C_Vertex3f(0.125395298f, 0.377980709f, 0.f),
    C_Vertex3f(0.125395298f, 0.377980709f, -1.f),
    C_Normal3f(0.928707838f, -0.370812148f, 0.f),
    C_Vertex3f(0.119597912f, 0.367474079f, 0.f),
    C_Vertex3f(0.119597912f, 0.367474079f, -1.f),
    C_Normal3f(0.990716219f, -0.135946453f, 0.f),
    C_Vertex3f(0.116699219f, 0.356201172f, 0.f),
    C_Vertex3f(0.116699219f, 0.356201172f, -1.f),
    C_Normal3f(0.99754113f, 0.0700827762f, 0.f),
    C_Vertex3f(0.116699219f, 0.346349716f, 0.f),
    C_Vertex3f(0.116699219f, 0.346349716f, -1.f),
    C_Normal3f(0.969055176f, 0.246844038f, 0.f),
    C_Vertex3f(0.117902756f, 0.33907032f, 0.f),
    C_Vertex3f(0.117902756f, 0.33907032f, -1.f),
    C_Normal3f(0.912301242f, 0.409519881f, 0.f),
    C_Vertex3f(0.12030983f, 0.332175255f, 0.f),
    C_Vertex3f(0.12030983f, 0.332175255f, -1.f),
    C_Normal3f(0.846645236f, 0.532157719f, 0.f),
    C_Vertex3f(0.123920441f, 0.32566452f, 0.f),
    C_Vertex3f(0.123920441f, 0.32566452f, -1.f),
    C_Normal3f(0.791732073f, 0.610868514f, 0.f),
    C_Vertex3f(0.126327515f, 0.322601318f, 0.f),
    C_Vertex3f(0.126327515f, 0.322601318f, -1.f),
    C_Normal3f(0.722636878f, 0.691227913f, 0.f),
    C_Vertex3f(0.128732681f, 0.31942749f, 0.f),
    C_Vertex3f(0.128732681f, 0.31942749f, -1.f),
    C_Normal3f(0.627186358f, 0.778869152f, 0.f),
    C_Vertex3f(0.134911537f, 0.313627243f, 0.f),
    C_Vertex3f(0.134911537f, 0.313627243f, -1.f),
    C_Normal3f(0.517561138f, 0.855646193f, 0.f),
    C_Vertex3f(0.142458916f, 0.308374405f, 0.f),
    C_Vertex3f(0.142458916f, 0.308374405f, -1.f),
    C_Normal3f(0.434632897f, 0.900607765f, 0.f),
    C_Vertex3f(0.151374817f, 0.303668976f, 0.f),
    C_Vertex3f(0.151374817f, 0.303668976f, -1.f),
    C_Normal3f(0.368849158f, 0.929489195f, 0.f),
    C_Vertex3f(0.156517029f, 0.301589966f, 0.f),
    C_Vertex3f(0.156517029f, 0.301589966f, -1.f),
    C_Normal3f(0.322466284f, 0.946580887f, 0.f),
    C_Vertex3f(0.157993317f, 0.301042557f, 0.f),
    C_Vertex3f(0.157993317f, 0.301042557f, -1.f),
    C_Normal3f(0.303810388f, 0.952732503f, 0.f),
    C_Vertex3f(0.162381887f, 0.299592018f, 0.f),
    C_Vertex3f(0.162381887f, 0.299592018f, -1.f),
    C_Normal3f(0.290211618f, 0.956962526f, 0.f),
    C_Vertex3f(0.168206453f, 0.297785759f, 0.f),
    C_Vertex3f(0.168206453f, 0.297785759f, -1.f),
    C_Normal3f(0.281414062f, 0.959586382f, 0.f),
    C_Vertex3f(0.175467014f, 0.295623779f, 0.f),
    C_Vertex3f(0.175467014f, 0.295623779f, -1.f),
    C_Normal3f(0.275258064f, 0.961370349f, 0.f),
    C_Vertex3f(0.18416357f, 0.293106079f, 0.f),
    C_Vertex3f(0.18416357f, 0.293106079f, -1.f),
    C_Normal3f(0.270709753f, 0.962661028f, 0.f),
    C_Vertex3f(0.194296122f, 0.290232658f, 0.f),
    C_Vertex3f(0.194296122f, 0.290232658f, -1.f),
    C_Normal3f(0.267212242f, 0.96363765f, 0.f),
    C_Vertex3f(0.205864668f, 0.287003517f, 0.f),
    C_Vertex3f(0.205864668f, 0.287003517f, -1.f),
    C_Normal3f(0.262356311f, 0.964971066f, 0.f),
    C_Vertex3f(0.218869209f, 0.283418655f, 0.f),
    C_Vertex3f(0.218869209f, 0.283418655f, -1.f),
    C_Normal3f(0.262702644f, 0.964876831f, 0.f),
    C_Vertex3f(0.236536026f, 0.278664589f, 0.f),
    C_Vertex3f(0.236536026f, 0.278664589f, -1.f),
    C_Normal3f(0.269685537f, 0.962948442f, 0.f),
    C_Vertex3f(0.255851984f, 0.273349524f, 0.f),
    C_Vertex3f(0.255851984f, 0.273349524f, -1.f),
    C_Normal3f(0.279609233f, 0.960113943f, 0.f),
    C_Vertex3f(0.273590803f, 0.268286943f, 0.f),
    C_Vertex3f(0.273590803f, 0.268286943f, -1.f),
    C_Normal3f(0.291450858f, 0.956585765f, 0.f),
    C_Vertex3f(0.289752483f, 0.263476849f, 0.f),
    C_Vertex3f(0.289752483f, 0.263476849f, -1.f),
    C_Normal3f(0.30581823f, 0.952089965f, 0.f),
    C_Vertex3f(0.304337025f, 0.258919239f, 0.f),
    C_Vertex3f(0.304337025f, 0.258919239f, -1.f),
    C_Normal3f(0.323601365f, 0.946193576f, 0.f),
    C_Vertex3f(0.317344427f, 0.254614115f, 0.f),
    C_Vertex3f(0.317344427f, 0.254614115f, -1.f),
    C_Normal3f(0.346153408f, 0.938177943f, 0.f),
    C_Vertex3f(0.328774691f, 0.250561476f, 0.f),
    C_Vertex3f(0.328774691f, 0.250561476f, -1.f),
    C_Normal3f(0.379073739f, 0.925366402f, 0.f),
    C_Vertex3f(0.338627815f, 0.246761322f, 0.f),
    C_Vertex3f(0.338627815f, 0.246761322f, -1.f),
    C_Normal3f(0.458562732f, 0.88866204f, 0.f),
    C_Vertex3f(0.351041794f, 0.241439819f, 0.f),
    C_Vertex3f(0.351041794f, 0.241439819f, -1.f),
    C_Normal3f(0.565282106f, 0.824897587f, 0.f),
    C_Vertex3f(0.365812302f, 0.232733727f, 0.f),
    C_Vertex3f(0.365812302f, 0.232733727f, -1.f),
    C_Normal3f(0.676785588f, 0.736180186f, 0.f),
    C_Vertex3f(0.378801346f, 0.222416878f, 0.f),
    C_Vertex3f(0.378801346f, 0.222416878f, -1.f),
    C_Normal3f(0.776970387f, 0.629537165f, 0.f),
    C_Vertex3f(0.390008926f, 0.210489273f, 0.f),
    C_Vertex3f(0.390008926f, 0.210489273f, -1.f),
    C_Normal3f(0.867101312f, 0.498131841f, 0.f),
    C_Vertex3f(0.399435043f, 0.196950912f, 0.f),
    C_Vertex3f(0.399435043f, 0.196950912f, -1.f),
    C_Normal3f(0.93836987f, 0.345632762f, 0.f),
    C_Vertex3f(0.406504631f, 0.181775093f, 0.f),
    C_Vertex3f(0.406504631f, 0.181775093f, -1.f),
    C_Normal3f(0.980490804f, 0.196564838f, 0.f),
    C_Vertex3f(0.41121769f, 0.164961815f, 0.f),
    C_Vertex3f(0.41121769f, 0.164961815f, -1.f),
    C_Normal3f(0.998114407f, 0.0613807701f, 0.f),
    C_Vertex3f(0.413574219f, 0.146511078f, 0.f),
    C_Vertex3f(0.413574219f, 0.146511078f, -1.f),
    C_Normal3f(0.997288287f, -0.0735932663f, 0.f),
    C_Vertex3f(0.413574219f, 0.126642227f, 0.f),
    C_Vertex3f(0.413574219f, 0.126642227f, -1.f),
    C_Normal3f(0.974677503f, -0.223615587f, 0.f),
    C_Vertex3f(0.410700321f, 0.10756588f, 0.f),
    C_Vertex3f(0.410700321f, 0.10756588f, -1.f),
    C_Normal3f(0.930260181f, -0.366900384f, 0.f),
    C_Vertex3f(0.404952526f, 0.0890626907f, 0.f),
    C_Vertex3f(0.404952526f, 0.0890626907f, -1.f),
    C_Normal3f(0.868737638f, -0.495272696f, 0.f),
    C_Vertex3f(0.396330833f, 0.0711326599f, 0.f),
    C_Vertex3f(0.396330833f, 0.0711326599f, -1.f),
    C_Normal3f(0.788599133f, -0.614907563f, 0.f),
    C_Vertex3f(0.384835243f, 0.0537757874f, 0.f),
    C_Vertex3f(0.384835243f, 0.0537757874f, -1.f),
    C_Normal3f(0.6853109f, -0.728250563f, 0.f),
    C_Vertex3f(0.370794773f, 0.0383834839f, 0.f),
    C_Vertex3f(0.370794773f, 0.0383834839f, -1.f),
    C_Normal3f(0.571765184f, -0.820417285f, 0.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, 0.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, -1.f),
    C_Normal3f(0.456936926f, -0.889499009f, 0.f),
    C_Vertex3f(0.335079193f, 0.0134925842f, 0.f),
    C_Vertex3f(0.335079193f, 0.0134925842f, -1.f),
    C_Normal3f(0.348042637f, -0.937478721f, 0.f),
    C_Vertex3f(0.313404083f, 0.00399398804f, 0.f),
    C_Vertex3f(0.313404083f, 0.00399398804f, -1.f),
    C_Normal3f(0.24191834f, -0.970296621f, 0.f),
    C_Vertex3f(0.290305138f, -0.00312995911f, 0.f),
    C_Vertex3f(0.290305138f, -0.00312995911f, -1.f),
    C_Normal3f(0.139768258f, -0.990184247f, 0.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, 0.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, -1.f),
    C_Normal3f(0.046545554f, -0.998916209f, 0.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, 0.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0253265686f, -0.999679208f, 0.f),
    C_Vertex3f(0.214819908f, -0.0102539063f, 0.f),
    C_Vertex3f(0.214819908f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0849944577f, -0.996381402f, 0.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, 0.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, -1.f),
    C_Normal3f(-0.150375873f, -0.988628924f, 0.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, 0.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, -1.f),
    C_Normal3f(-0.223169982f, -0.974779606f, 0.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, 0.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, -1.f),
    C_Normal3f(-0.303275019f, -0.952903092f, 0.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, 0.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, -1.f),
    C_Normal3f(-0.389873773f, -0.920868278f, 0.f),
    C_Vertex3f(0.121732354f, 0.00735211372f, 0.f),
    C_Vertex3f(0.121732354f, 0.00735211372f, -1.f),
    C_Normal3f(-0.481160849f, -0.876632392f, 0.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, 0.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, -1.f),
    C_Normal3f(-0.574190378f, -0.818721831f, 0.f),
    C_Vertex3f(0.0939326286f, 0.0226106644f, 0.f),
    C_Vertex3f(0.0939326286f, 0.0226106644f, -1.f),
    C_Normal3f(-0.659739971f, -0.75149405f, 0.f),
    C_Vertex3f(0.082054615f, 0.0320005417f, 0.f),
    C_Vertex3f(0.082054615f, 0.0320005417f, -1.f),
    C_Normal3f(-0.733788431f, -0.679378092f, 0.f),
    C_Vertex3f(0.0712507963f, 0.0425231457f, 0.f),
    C_Vertex3f(0.0712507963f, 0.0425231457f, -1.f),
    C_Normal3f(-0.799175143f, -0.601098239f, 0.f),
    C_Vertex3f(0.0615211725f, 0.0541784763f, 0.f),
    C_Vertex3f(0.0615211725f, 0.0541784763f, -1.f),
    C_Normal3f(-0.854495287f, -0.519459128f, 0.f),
    C_Vertex3f(0.0528657436f, 0.0669665337f, 0.f),
    C_Vertex3f(0.0528657436f, 0.0669665337f, -1.f),
    C_Normal3f(-0.899324596f, -0.437281758f, 0.f),
    C_Vertex3f(0.0452845097f, 0.0808873177f, 0.f),
    C_Vertex3f(0.0452845097f, 0.0808873177f, -1.f),
    C_Normal3f(-0.934099019f, -0.357014179f, 0.f),
    C_Vertex3f(0.0387774706f, 0.0959408283f, 0.f),
    C_Vertex3f(0.0387774706f, 0.0959408283f, -1.f),
    C_Normal3f(-0.959852457f, -0.280505359f, 0.f),
    C_Vertex3f(0.0333446264f, 0.112127066f, 0.f),
    C_Vertex3f(0.0333446264f, 0.112127066f, -1.f),
    C_Normal3f(-0.975386679f, -0.220501676f, 0.f),
    C_Vertex3f(0.0289859772f, 0.12944603f, 0.f),
    C_Vertex3f(0.0289859772f, 0.12944603f, -1.f),
    C_Normal3f(-0.98452431f, -0.175248176f, 0.f),
    C_Vertex3f(0.02734375f, 0.138671875f, 0.f),
    C_Vertex3f(0.02734375f, 0.138671875f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.395996094f, 0.344238281f, 0.f),
    C_Vertex3f(0.394136429f, 0.355503082f, 0.f),
    C_Vertex3f(0.388884068f, 0.375763416f, 0.f),
    C_Vertex3f(0.382098675f, 0.393754482f, 0.f),
    C_Vertex3f(0.373780251f, 0.40947628f, 0.f),
    C_Vertex3f(0.363928795f, 0.42292881f, 0.f),
    C_Vertex3f(0.351369381f, 0.434987068f, 0.f),
    C_Vertex3f(0.336102009f, 0.445651054f, 0.f),
    C_Vertex3f(0.318847656f, 0.333984375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.318126678f, 0.454920769f, 0.f),
    C_Vertex3f(0.29744339f, 0.462796211f, 0.f),
    C_Vertex3f(0.304819107f, 0.372945309f, 0.f),
    C_Vertex3f(0.312421799f, 0.358936787f, 0.f),
    C_Vertex3f(0.3175354f, 0.34295845f, 0.f),
    C_Vertex3f(0.318847656f, 0.333984375f, 0.f),
    C_Vertex3f(0.336102009f, 0.445651054f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.228822708f, 0.41015625f, 0.f),
    C_Vertex3f(0.249194145f, 0.407639027f, 0.f),
    C_Vertex3f(0.251067162f, 0.472640514f, 0.f),
    C_Vertex3f(0.266968727f, 0.40260458f, 0.f),
    C_Vertex3f(0.275090218f, 0.468702793f, 0.f),
    C_Vertex3f(0.282146454f, 0.39505291f, 0.f),
    C_Vertex3f(0.29744339f, 0.462796211f, 0.f),
    C_Vertex3f(0.294727325f, 0.384984016f, 0.f),
    C_Vertex3f(0.304819107f, 0.372945309f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186338782f, 0.408473611f, 0.f),
    C_Vertex3f(0.197961926f, 0.40959537f, 0.f),
    C_Vertex3f(0.202610016f, 0.474609375f, 0.f),
    C_Vertex3f(0.21055603f, 0.41015625f, 0.f),
    C_Vertex3f(0.225374222f, 0.474609375f, 0.f),
    C_Vertex3f(0.228822708f, 0.41015625f, 0.f),
    C_Vertex3f(0.251067162f, 0.472640514f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157295108f, 0.401743054f, 0.f),
    C_Vertex3f(0.166005373f, 0.404547453f, 0.f),
    C_Vertex3f(0.168001175f, 0.471088886f, 0.f),
    C_Vertex3f(0.175686598f, 0.406790972f, 0.f),
    C_Vertex3f(0.185018539f, 0.473435879f, 0.f),
    C_Vertex3f(0.186338782f, 0.408473611f, 0.f),
    C_Vertex3f(0.202610016f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.151557922f, 0.467568398f, 0.f),
    C_Vertex3f(0.135688782f, 0.462874413f, 0.f),
    C_Vertex3f(0.142787457f, 0.394451618f, 0.f),
    C_Vertex3f(0.149555802f, 0.398377776f, 0.f),
    C_Vertex3f(0.157295108f, 0.401743054f, 0.f),
    C_Vertex3f(0.168001175f, 0.471088886f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116699219f, 0.356201172f, 0.f),
    C_Vertex3f(0.119597912f, 0.367474079f, 0.f),
    C_Vertex3f(0.121488571f, 0.457662106f, 0.f),
    C_Vertex3f(0.125395298f, 0.377980709f, 0.f),
    C_Vertex3f(0.135688782f, 0.462874413f, 0.f),
    C_Vertex3f(0.134091377f, 0.387721062f, 0.f),
    C_Vertex3f(0.142787457f, 0.394451618f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12030983f, 0.332175255f, 0.f),
    C_Vertex3f(0.117902756f, 0.33907032f, 0.f),
    C_Vertex3f(0.122807026f, 0.228503942f, 0.f),
    C_Vertex3f(0.116699219f, 0.346349716f, 0.f),
    C_Vertex3f(0.112774849f, 0.232672691f, 0.f),
    C_Vertex3f(0.116699219f, 0.356201172f, 0.f),
    C_Vertex3f(0.108957291f, 0.451931477f, 0.f),
    C_Vertex3f(0.121488571f, 0.457662106f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.100217819f, 0.238649368f, 0.f),
    C_Vertex3f(0.112774849f, 0.232672691f, 0.f),
    C_Vertex3f(0.108957291f, 0.451931477f, 0.f),
    C_Vertex3f(0.0980949402f, 0.445682526f, 0.f),
    C_Vertex3f(0.09349823f, 0.442298889f, 0.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, 0.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0423059464f, 0.363668919f, 0.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, 0.f),
    C_Vertex3f(0.0458631516f, 0.378024578f, 0.f),
    C_Vertex3f(0.0470104218f, 0.300626278f, 0.f),
    C_Vertex3f(0.0511989594f, 0.391806602f, 0.f),
    C_Vertex3f(0.0534934998f, 0.28580761f, 0.f),
    C_Vertex3f(0.0583133698f, 0.405014992f, 0.f),
    C_Vertex3f(0.0621376038f, 0.271725655f, 0.f),
    C_Vertex3f(0.0667133331f, 0.417105198f, 0.f),
    C_Vertex3f(0.0728063583f, 0.259171963f, 0.f),
    C_Vertex3f(0.0763988495f, 0.428077221f, 0.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, 0.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0405273438f, 0.348739624f, 0.f),
    C_Vertex3f(0.0405273438f, 0.332473755f, 0.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, 0.f),
    C_Vertex3f(0.0423059464f, 0.363668919f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.351041794f, 0.241439819f, 0.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, 0.f),
    C_Vertex3f(0.365812302f, 0.232733727f, 0.f),
    C_Vertex3f(0.370794773f, 0.0383834839f, 0.f),
    C_Vertex3f(0.378801346f, 0.222416878f, 0.f),
    C_Vertex3f(0.384835243f, 0.0537757874f, 0.f),
    C_Vertex3f(0.390008926f, 0.210489273f, 0.f),
    C_Vertex3f(0.396330833f, 0.0711326599f, 0.f),
    C_Vertex3f(0.399435043f, 0.196950912f, 0.f),
    C_Vertex3f(0.404952526f, 0.0890626907f, 0.f),
    C_Vertex3f(0.406504631f, 0.181775093f, 0.f),
    C_Vertex3f(0.410700321f, 0.10756588f, 0.f),
    C_Vertex3f(0.41121769f, 0.164961815f, 0.f),
    C_Vertex3f(0.413574219f, 0.126642227f, 0.f),
    C_Vertex3f(0.413574219f, 0.146511078f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338627815f, 0.246761322f, 0.f),
    C_Vertex3f(0.328774691f, 0.250561476f, 0.f),
    C_Vertex3f(0.329612732f, 0.146773338f, 0.f),
    C_Vertex3f(0.332519531f, 0.134847641f, 0.f),
    C_Vertex3f(0.335079193f, 0.0134925842f, 0.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, 0.f),
    C_Vertex3f(0.351041794f, 0.241439819f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.307235718f, 0.171502113f, 0.f),
    C_Vertex3f(0.315078735f, 0.166358948f, 0.f),
    C_Vertex3f(0.317344427f, 0.254614115f, 0.f),
    C_Vertex3f(0.323799133f, 0.157277107f, 0.f),
    C_Vertex3f(0.328774691f, 0.250561476f, 0.f),
    C_Vertex3f(0.329612732f, 0.146773338f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304337025f, 0.258919239f, 0.f),
    C_Vertex3f(0.289752483f, 0.263476849f, 0.f),
    C_Vertex3f(0.294609785f, 0.177288532f, 0.f),
    C_Vertex3f(0.301674128f, 0.174306393f, 0.f),
    C_Vertex3f(0.307235718f, 0.171502113f, 0.f),
    C_Vertex3f(0.317344427f, 0.254614115f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.236746788f, 0.194867134f, 0.f),
    C_Vertex3f(0.251324892f, 0.190995693f, 0.f),
    C_Vertex3f(0.255851984f, 0.273349524f, 0.f),
    C_Vertex3f(0.264400244f, 0.187302113f, 0.f),
    C_Vertex3f(0.273590803f, 0.268286943f, 0.f),
    C_Vertex3f(0.275972843f, 0.183786392f, 0.f),
    C_Vertex3f(0.289752483f, 0.263476849f, 0.f),
    C_Vertex3f(0.28604269f, 0.180448532f, 0.f),
    C_Vertex3f(0.294609785f, 0.177288532f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.255851984f, 0.273349524f, 0.f),
    C_Vertex3f(0.236536026f, 0.278664589f, 0.f),
    C_Vertex3f(0.236746788f, 0.194867134f, 0.f),
    C_Vertex3f(0.218869209f, 0.283418655f, 0.f),
    C_Vertex3f(0.217877388f, 0.199621201f, 0.f),
    C_Vertex3f(0.205864668f, 0.287003517f, 0.f),
    C_Vertex3f(0.197880268f, 0.204895735f, 0.f),
    C_Vertex3f(0.194296122f, 0.290232658f, 0.f),
    C_Vertex3f(0.18416357f, 0.293106079f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179543972f, 0.209985971f, 0.f),
    C_Vertex3f(0.197880268f, 0.204895735f, 0.f),
    C_Vertex3f(0.18416357f, 0.293106079f, 0.f),
    C_Vertex3f(0.175467014f, 0.295623779f, 0.f),
    C_Vertex3f(0.168206453f, 0.297785759f, 0.f),
    C_Vertex3f(0.1628685f, 0.214891911f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1628685f, 0.214891911f, 0.f),
    C_Vertex3f(0.168206453f, 0.297785759f, 0.f),
    C_Vertex3f(0.162381887f, 0.299592018f, 0.f),
    C_Vertex3f(0.157993317f, 0.301042557f, 0.f),
    C_Vertex3f(0.156517029f, 0.301589966f, 0.f),
    C_Vertex3f(0.151374817f, 0.303668976f, 0.f),
    C_Vertex3f(0.147853851f, 0.219613552f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151374817f, 0.303668976f, 0.f),
    C_Vertex3f(0.142458916f, 0.308374405f, 0.f),
    C_Vertex3f(0.147853851f, 0.219613552f, 0.f),
    C_Vertex3f(0.134911537f, 0.313627243f, 0.f),
    C_Vertex3f(0.134500027f, 0.224150896f, 0.f),
    C_Vertex3f(0.128732681f, 0.31942749f, 0.f),
    C_Vertex3f(0.126327515f, 0.322601318f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123920441f, 0.32566452f, 0.f),
    C_Vertex3f(0.12030983f, 0.332175255f, 0.f),
    C_Vertex3f(0.122807026f, 0.228503942f, 0.f),
    C_Vertex3f(0.134500027f, 0.224150896f, 0.f),
    C_Vertex3f(0.126327515f, 0.322601318f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.335079193f, 0.0134925842f, 0.f),
    C_Vertex3f(0.332519531f, 0.134847641f, 0.f),
    C_Vertex3f(0.332519531f, 0.120731354f, 0.f),
    C_Vertex3f(0.329231262f, 0.106804848f, 0.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, 0.f),
    C_Vertex3f(0.313404083f, 0.00399398804f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.313404083f, 0.00399398804f, 0.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, 0.f),
    C_Vertex3f(0.312789917f, 0.0818271637f, 0.f),
    C_Vertex3f(0.30292511f, 0.0735387802f, 0.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, 0.f),
    C_Vertex3f(0.290305138f, -0.00312995911f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.290305138f, -0.00312995911f, 0.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, 0.f),
    C_Vertex3f(0.287142038f, 0.0645596981f, 0.f),
    C_Vertex3f(0.277935505f, 0.061106205f, 0.f),
    C_Vertex3f(0.267852306f, 0.0583434105f, 0.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.267852306f, 0.0583434105f, 0.f),
    C_Vertex3f(0.256892443f, 0.0562713146f, 0.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, 0.f),
    C_Vertex3f(0.245055914f, 0.0548899174f, 0.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, 0.f),
    C_Vertex3f(0.23234272f, 0.0541992188f, 0.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214819908f, -0.0102539063f, 0.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, 0.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, 0.f),
    C_Vertex3f(0.206072688f, 0.0549784899f, 0.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, 0.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.182843447f, 0.0588748455f, 0.f),
    C_Vertex3f(0.172348738f, 0.06199193f, 0.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, 0.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, 0.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162600636f, 0.0658882856f, 0.f),
    C_Vertex3f(0.153599143f, 0.0705639124f, 0.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, 0.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, 0.f),
    C_Vertex3f(0.172348738f, 0.06199193f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153599143f, 0.0705639124f, 0.f),
    C_Vertex3f(0.145344257f, 0.0760188103f, 0.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, 0.f),
    C_Vertex3f(0.137835979f, 0.0822529793f, 0.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, 0.f),
    C_Vertex3f(0.131060719f, 0.0891705751f, 0.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.121732354f, 0.00735211372f, 0.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, 0.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, 0.f),
    C_Vertex3f(0.119709253f, 0.105056047f, 0.f),
    C_Vertex3f(0.115133047f, 0.114023924f, 0.f),
    C_Vertex3f(0.111289859f, 0.123675227f, 0.f),
    C_Vertex3f(0.108179688f, 0.134009957f, 0.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.082054615f, 0.0320005417f, 0.f),
    C_Vertex3f(0.0939326286f, 0.0226106644f, 0.f),
    C_Vertex3f(0.104980469f, 0.150878906f, 0.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, 0.f),
    C_Vertex3f(0.105802536f, 0.145028114f, 0.f),
    C_Vertex3f(0.108179688f, 0.134009957f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.104980469f, 0.150878906f, 0.f),
    C_Vertex3f(0.02734375f, 0.138671875f, 0.f),
    C_Vertex3f(0.0289859772f, 0.12944603f, 0.f),
    C_Vertex3f(0.0333446264f, 0.112127066f, 0.f),
    C_Vertex3f(0.0387774706f, 0.0959408283f, 0.f),
    C_Vertex3f(0.0452845097f, 0.0808873177f, 0.f),
    C_Vertex3f(0.0528657436f, 0.0669665337f, 0.f),
    C_Vertex3f(0.0615211725f, 0.0541784763f, 0.f),
    C_Vertex3f(0.0712507963f, 0.0425231457f, 0.f),
    C_Vertex3f(0.082054615f, 0.0320005417f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.395996094f, 0.344238281f, -1.f),
    C_Vertex3f(0.394136429f, 0.355503082f, -1.f),
    C_Vertex3f(0.388884068f, 0.375763416f, -1.f),
    C_Vertex3f(0.382098675f, 0.393754482f, -1.f),
    C_Vertex3f(0.373780251f, 0.40947628f, -1.f),
    C_Vertex3f(0.363928795f, 0.42292881f, -1.f),
    C_Vertex3f(0.351369381f, 0.434987068f, -1.f),
    C_Vertex3f(0.336102009f, 0.445651054f, -1.f),
    C_Vertex3f(0.318847656f, 0.333984375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.318126678f, 0.454920769f, -1.f),
    C_Vertex3f(0.29744339f, 0.462796211f, -1.f),
    C_Vertex3f(0.304819107f, 0.372945309f, -1.f),
    C_Vertex3f(0.312421799f, 0.358936787f, -1.f),
    C_Vertex3f(0.3175354f, 0.34295845f, -1.f),
    C_Vertex3f(0.318847656f, 0.333984375f, -1.f),
    C_Vertex3f(0.336102009f, 0.445651054f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.228822708f, 0.41015625f, -1.f),
    C_Vertex3f(0.249194145f, 0.407639027f, -1.f),
    C_Vertex3f(0.251067162f, 0.472640514f, -1.f),
    C_Vertex3f(0.266968727f, 0.40260458f, -1.f),
    C_Vertex3f(0.275090218f, 0.468702793f, -1.f),
    C_Vertex3f(0.282146454f, 0.39505291f, -1.f),
    C_Vertex3f(0.29744339f, 0.462796211f, -1.f),
    C_Vertex3f(0.294727325f, 0.384984016f, -1.f),
    C_Vertex3f(0.304819107f, 0.372945309f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186338782f, 0.408473611f, -1.f),
    C_Vertex3f(0.197961926f, 0.40959537f, -1.f),
    C_Vertex3f(0.202610016f, 0.474609375f, -1.f),
    C_Vertex3f(0.21055603f, 0.41015625f, -1.f),
    C_Vertex3f(0.225374222f, 0.474609375f, -1.f),
    C_Vertex3f(0.228822708f, 0.41015625f, -1.f),
    C_Vertex3f(0.251067162f, 0.472640514f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157295108f, 0.401743054f, -1.f),
    C_Vertex3f(0.166005373f, 0.404547453f, -1.f),
    C_Vertex3f(0.168001175f, 0.471088886f, -1.f),
    C_Vertex3f(0.175686598f, 0.406790972f, -1.f),
    C_Vertex3f(0.185018539f, 0.473435879f, -1.f),
    C_Vertex3f(0.186338782f, 0.408473611f, -1.f),
    C_Vertex3f(0.202610016f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.151557922f, 0.467568398f, -1.f),
    C_Vertex3f(0.135688782f, 0.462874413f, -1.f),
    C_Vertex3f(0.142787457f, 0.394451618f, -1.f),
    C_Vertex3f(0.149555802f, 0.398377776f, -1.f),
    C_Vertex3f(0.157295108f, 0.401743054f, -1.f),
    C_Vertex3f(0.168001175f, 0.471088886f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116699219f, 0.356201172f, -1.f),
    C_Vertex3f(0.119597912f, 0.367474079f, -1.f),
    C_Vertex3f(0.121488571f, 0.457662106f, -1.f),
    C_Vertex3f(0.125395298f, 0.377980709f, -1.f),
    C_Vertex3f(0.135688782f, 0.462874413f, -1.f),
    C_Vertex3f(0.134091377f, 0.387721062f, -1.f),
    C_Vertex3f(0.142787457f, 0.394451618f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12030983f, 0.332175255f, -1.f),
    C_Vertex3f(0.117902756f, 0.33907032f, -1.f),
    C_Vertex3f(0.122807026f, 0.228503942f, -1.f),
    C_Vertex3f(0.116699219f, 0.346349716f, -1.f),
    C_Vertex3f(0.112774849f, 0.232672691f, -1.f),
    C_Vertex3f(0.116699219f, 0.356201172f, -1.f),
    C_Vertex3f(0.108957291f, 0.451931477f, -1.f),
    C_Vertex3f(0.121488571f, 0.457662106f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.100217819f, 0.238649368f, -1.f),
    C_Vertex3f(0.112774849f, 0.232672691f, -1.f),
    C_Vertex3f(0.108957291f, 0.451931477f, -1.f),
    C_Vertex3f(0.0980949402f, 0.445682526f, -1.f),
    C_Vertex3f(0.09349823f, 0.442298889f, -1.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, -1.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0423059464f, 0.363668919f, -1.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, -1.f),
    C_Vertex3f(0.0458631516f, 0.378024578f, -1.f),
    C_Vertex3f(0.0470104218f, 0.300626278f, -1.f),
    C_Vertex3f(0.0511989594f, 0.391806602f, -1.f),
    C_Vertex3f(0.0534934998f, 0.28580761f, -1.f),
    C_Vertex3f(0.0583133698f, 0.405014992f, -1.f),
    C_Vertex3f(0.0621376038f, 0.271725655f, -1.f),
    C_Vertex3f(0.0667133331f, 0.417105198f, -1.f),
    C_Vertex3f(0.0728063583f, 0.259171963f, -1.f),
    C_Vertex3f(0.0763988495f, 0.428077221f, -1.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, -1.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0405273438f, 0.348739624f, -1.f),
    C_Vertex3f(0.0405273438f, 0.332473755f, -1.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, -1.f),
    C_Vertex3f(0.0423059464f, 0.363668919f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.351041794f, 0.241439819f, -1.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, -1.f),
    C_Vertex3f(0.365812302f, 0.232733727f, -1.f),
    C_Vertex3f(0.370794773f, 0.0383834839f, -1.f),
    C_Vertex3f(0.378801346f, 0.222416878f, -1.f),
    C_Vertex3f(0.384835243f, 0.0537757874f, -1.f),
    C_Vertex3f(0.390008926f, 0.210489273f, -1.f),
    C_Vertex3f(0.396330833f, 0.0711326599f, -1.f),
    C_Vertex3f(0.399435043f, 0.196950912f, -1.f),
    C_Vertex3f(0.404952526f, 0.0890626907f, -1.f),
    C_Vertex3f(0.406504631f, 0.181775093f, -1.f),
    C_Vertex3f(0.410700321f, 0.10756588f, -1.f),
    C_Vertex3f(0.41121769f, 0.164961815f, -1.f),
    C_Vertex3f(0.413574219f, 0.126642227f, -1.f),
    C_Vertex3f(0.413574219f, 0.146511078f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338627815f, 0.246761322f, -1.f),
    C_Vertex3f(0.328774691f, 0.250561476f, -1.f),
    C_Vertex3f(0.329612732f, 0.146773338f, -1.f),
    C_Vertex3f(0.332519531f, 0.134847641f, -1.f),
    C_Vertex3f(0.335079193f, 0.0134925842f, -1.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, -1.f),
    C_Vertex3f(0.351041794f, 0.241439819f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.307235718f, 0.171502113f, -1.f),
    C_Vertex3f(0.315078735f, 0.166358948f, -1.f),
    C_Vertex3f(0.317344427f, 0.254614115f, -1.f),
    C_Vertex3f(0.323799133f, 0.157277107f, -1.f),
    C_Vertex3f(0.328774691f, 0.250561476f, -1.f),
    C_Vertex3f(0.329612732f, 0.146773338f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304337025f, 0.258919239f, -1.f),
    C_Vertex3f(0.289752483f, 0.263476849f, -1.f),
    C_Vertex3f(0.294609785f, 0.177288532f, -1.f),
    C_Vertex3f(0.301674128f, 0.174306393f, -1.f),
    C_Vertex3f(0.307235718f, 0.171502113f, -1.f),
    C_Vertex3f(0.317344427f, 0.254614115f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.236746788f, 0.194867134f, -1.f),
    C_Vertex3f(0.251324892f, 0.190995693f, -1.f),
    C_Vertex3f(0.255851984f, 0.273349524f, -1.f),
    C_Vertex3f(0.264400244f, 0.187302113f, -1.f),
    C_Vertex3f(0.273590803f, 0.268286943f, -1.f),
    C_Vertex3f(0.275972843f, 0.183786392f, -1.f),
    C_Vertex3f(0.289752483f, 0.263476849f, -1.f),
    C_Vertex3f(0.28604269f, 0.180448532f, -1.f),
    C_Vertex3f(0.294609785f, 0.177288532f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.255851984f, 0.273349524f, -1.f),
    C_Vertex3f(0.236536026f, 0.278664589f, -1.f),
    C_Vertex3f(0.236746788f, 0.194867134f, -1.f),
    C_Vertex3f(0.218869209f, 0.283418655f, -1.f),
    C_Vertex3f(0.217877388f, 0.199621201f, -1.f),
    C_Vertex3f(0.205864668f, 0.287003517f, -1.f),
    C_Vertex3f(0.197880268f, 0.204895735f, -1.f),
    C_Vertex3f(0.194296122f, 0.290232658f, -1.f),
    C_Vertex3f(0.18416357f, 0.293106079f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179543972f, 0.209985971f, -1.f),
    C_Vertex3f(0.197880268f, 0.204895735f, -1.f),
    C_Vertex3f(0.18416357f, 0.293106079f, -1.f),
    C_Vertex3f(0.175467014f, 0.295623779f, -1.f),
    C_Vertex3f(0.168206453f, 0.297785759f, -1.f),
    C_Vertex3f(0.1628685f, 0.214891911f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1628685f, 0.214891911f, -1.f),
    C_Vertex3f(0.168206453f, 0.297785759f, -1.f),
    C_Vertex3f(0.162381887f, 0.299592018f, -1.f),
    C_Vertex3f(0.157993317f, 0.301042557f, -1.f),
    C_Vertex3f(0.156517029f, 0.301589966f, -1.f),
    C_Vertex3f(0.151374817f, 0.303668976f, -1.f),
    C_Vertex3f(0.147853851f, 0.219613552f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151374817f, 0.303668976f, -1.f),
    C_Vertex3f(0.142458916f, 0.308374405f, -1.f),
    C_Vertex3f(0.147853851f, 0.219613552f, -1.f),
    C_Vertex3f(0.134911537f, 0.313627243f, -1.f),
    C_Vertex3f(0.134500027f, 0.224150896f, -1.f),
    C_Vertex3f(0.128732681f, 0.31942749f, -1.f),
    C_Vertex3f(0.126327515f, 0.322601318f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123920441f, 0.32566452f, -1.f),
    C_Vertex3f(0.12030983f, 0.332175255f, -1.f),
    C_Vertex3f(0.122807026f, 0.228503942f, -1.f),
    C_Vertex3f(0.134500027f, 0.224150896f, -1.f),
    C_Vertex3f(0.126327515f, 0.322601318f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.335079193f, 0.0134925842f, -1.f),
    C_Vertex3f(0.332519531f, 0.134847641f, -1.f),
    C_Vertex3f(0.332519531f, 0.120731354f, -1.f),
    C_Vertex3f(0.329231262f, 0.106804848f, -1.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, -1.f),
    C_Vertex3f(0.313404083f, 0.00399398804f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.313404083f, 0.00399398804f, -1.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, -1.f),
    C_Vertex3f(0.312789917f, 0.0818271637f, -1.f),
    C_Vertex3f(0.30292511f, 0.0735387802f, -1.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, -1.f),
    C_Vertex3f(0.290305138f, -0.00312995911f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.290305138f, -0.00312995911f, -1.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, -1.f),
    C_Vertex3f(0.287142038f, 0.0645596981f, -1.f),
    C_Vertex3f(0.277935505f, 0.061106205f, -1.f),
    C_Vertex3f(0.267852306f, 0.0583434105f, -1.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.267852306f, 0.0583434105f, -1.f),
    C_Vertex3f(0.256892443f, 0.0562713146f, -1.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, -1.f),
    C_Vertex3f(0.245055914f, 0.0548899174f, -1.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, -1.f),
    C_Vertex3f(0.23234272f, 0.0541992188f, -1.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214819908f, -0.0102539063f, -1.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, -1.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, -1.f),
    C_Vertex3f(0.206072688f, 0.0549784899f, -1.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, -1.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.182843447f, 0.0588748455f, -1.f),
    C_Vertex3f(0.172348738f, 0.06199193f, -1.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, -1.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, -1.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162600636f, 0.0658882856f, -1.f),
    C_Vertex3f(0.153599143f, 0.0705639124f, -1.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, -1.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, -1.f),
    C_Vertex3f(0.172348738f, 0.06199193f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153599143f, 0.0705639124f, -1.f),
    C_Vertex3f(0.145344257f, 0.0760188103f, -1.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, -1.f),
    C_Vertex3f(0.137835979f, 0.0822529793f, -1.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, -1.f),
    C_Vertex3f(0.131060719f, 0.0891705751f, -1.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.121732354f, 0.00735211372f, -1.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, -1.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, -1.f),
    C_Vertex3f(0.119709253f, 0.105056047f, -1.f),
    C_Vertex3f(0.115133047f, 0.114023924f, -1.f),
    C_Vertex3f(0.111289859f, 0.123675227f, -1.f),
    C_Vertex3f(0.108179688f, 0.134009957f, -1.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.082054615f, 0.0320005417f, -1.f),
    C_Vertex3f(0.0939326286f, 0.0226106644f, -1.f),
    C_Vertex3f(0.104980469f, 0.150878906f, -1.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, -1.f),
    C_Vertex3f(0.105802536f, 0.145028114f, -1.f),
    C_Vertex3f(0.108179688f, 0.134009957f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.104980469f, 0.150878906f, -1.f),
    C_Vertex3f(0.02734375f, 0.138671875f, -1.f),
    C_Vertex3f(0.0289859772f, 0.12944603f, -1.f),
    C_Vertex3f(0.0333446264f, 0.112127066f, -1.f),
    C_Vertex3f(0.0387774706f, 0.0959408283f, -1.f),
    C_Vertex3f(0.0452845097f, 0.0808873177f, -1.f),
    C_Vertex3f(0.0528657436f, 0.0669665337f, -1.f),
    C_Vertex3f(0.0615211725f, 0.0541784763f, -1.f),
    C_Vertex3f(0.0712507963f, 0.0425231457f, -1.f),
    C_Vertex3f(0.082054615f, 0.0320005417f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_116[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.987190545f, 0.159545943f, 0.f),
    C_Vertex3f(0.23046875f, 0.0701141357f, 0.f),
    C_Vertex3f(0.23046875f, 0.0701141357f, -1.f),
    C_Vertex3f(0.241699219f, 0.000625610352f, -1.f),
    C_Vertex3f(0.241699219f, 0.000625610352f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.205760881f, -0.97860235f, 0.f),
    C_Vertex3f(0.241699219f, 0.000625610352f, 0.f),
    C_Vertex3f(0.241699219f, 0.000625610352f, -1.f),
    C_Normal3f(0.178665251f, -0.983909905f, 0.f),
    C_Vertex3f(0.233407974f, -0.0011177063f, 0.f),
    C_Vertex3f(0.233407974f, -0.0011177063f, -1.f),
    C_Normal3f(0.14124395f, -0.989974856f, 0.f),
    C_Vertex3f(0.217698097f, -0.00373268127f, 0.f),
    C_Vertex3f(0.217698097f, -0.00373268127f, -1.f),
    C_Normal3f(0.0904197916f, -0.99590379f, 0.f),
    C_Vertex3f(0.202860832f, -0.00547599792f, 0.f),
    C_Vertex3f(0.202860832f, -0.00547599792f, -1.f),
    C_Normal3f(0.0279239472f, -0.999610066f, 0.f),
    C_Vertex3f(0.188896179f, -0.00634765625f, 0.f),
    C_Vertex3f(0.188896179f, -0.00634765625f, -1.f),
    C_Normal3f(-0.0467723384f, -0.998905659f, 0.f),
    C_Vertex3f(0.171657562f, -0.00634765625f, 0.f),
    C_Vertex3f(0.171657562f, -0.00634765625f, -1.f),
    C_Normal3f(-0.141612709f, -0.989922047f, 0.f),
    C_Vertex3f(0.152672768f, -0.00465154648f, 0.f),
    C_Vertex3f(0.152672768f, -0.00465154648f, -1.f),
    C_Normal3f(-0.26571691f, -0.964051127f, 0.f),
    C_Vertex3f(0.136088371f, -0.00125932693f, 0.f),
    C_Vertex3f(0.136088371f, -0.00125932693f, -1.f),
    C_Normal3f(-0.415813833f, -0.909449756f, 0.f),
    C_Vertex3f(0.121904373f, 0.00382900238f, 0.f),
    C_Vertex3f(0.121904373f, 0.00382900238f, -1.f),
    C_Normal3f(-0.557014048f, -0.830503106f, 0.f),
    C_Vertex3f(0.110120773f, 0.0106134415f, 0.f),
    C_Vertex3f(0.110120773f, 0.0106134415f, -1.f),
    C_Normal3f(-0.674988091f, -0.737828553f, 0.f),
    C_Vertex3f(0.100082874f, 0.0184645653f, 0.f),
    C_Vertex3f(0.100082874f, 0.0184645653f, -1.f),
    C_Normal3f(-0.786583424f, -0.617484152f, 0.f),
    C_Vertex3f(0.0917906761f, 0.0273823738f, 0.f),
    C_Vertex3f(0.0917906761f, 0.0273823738f, -1.f),
    C_Normal3f(-0.874248743f, -0.485478282f, 0.f),
    C_Vertex3f(0.0852441788f, 0.0373668671f, 0.f),
    C_Vertex3f(0.0852441788f, 0.0373668671f, -1.f),
    C_Normal3f(-0.937352538f, -0.348382026f, 0.f),
    C_Vertex3f(0.0816435814f, 0.0456552505f, 0.f),
    C_Vertex3f(0.0816435814f, 0.0456552505f, -1.f),
    C_Normal3f(-0.970227718f, -0.24219425f, 0.f),
    C_Vertex3f(0.0795432329f, 0.0527057648f, 0.f),
    C_Vertex3f(0.0795432329f, 0.0527057648f, -1.f),
    C_Normal3f(-0.984740019f, -0.17403166f, 0.f),
    C_Vertex3f(0.0777429342f, 0.0612812042f, 0.f),
    C_Vertex3f(0.0777429342f, 0.0612812042f, -1.f),
    C_Normal3f(-0.992363334f, -0.123348445f, 0.f),
    C_Vertex3f(0.0762426853f, 0.0713815689f, 0.f),
    C_Vertex3f(0.0762426853f, 0.0713815689f, -1.f),
    C_Normal3f(-0.996425748f, -0.0844721943f, 0.f),
    C_Vertex3f(0.0750424862f, 0.0830068588f, 0.f),
    C_Vertex3f(0.0750424862f, 0.0830068588f, -1.f),
    C_Normal3f(-0.99854964f, -0.0538384169f, 0.f),
    C_Vertex3f(0.0741423368f, 0.096157074f, 0.f),
    C_Vertex3f(0.0741423368f, 0.096157074f, -1.f),
    C_Normal3f(-0.999575257f, -0.0291420575f, 0.f),
    C_Vertex3f(0.0735422373f, 0.110832214f, 0.f),
    C_Vertex3f(0.0735422373f, 0.110832214f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0732421875f, 0.12703228f, 0.f),
    C_Vertex3f(0.0732421875f, 0.12703228f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0732421875f, 0.12703228f, 0.f),
    C_Vertex3f(0.0732421875f, 0.12703228f, -1.f),
    C_Vertex3f(0.0732421875f, 0.403320313f, -1.f),
    C_Vertex3f(0.0732421875f, 0.403320313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0732421875f, 0.403320313f, 0.f),
    C_Vertex3f(0.0732421875f, 0.403320313f, -1.f),
    C_Vertex3f(0.0151367188f, 0.403320313f, -1.f),
    C_Vertex3f(0.0151367188f, 0.403320313f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0151367188f, 0.403320313f, 0.f),
    C_Vertex3f(0.0151367188f, 0.403320313f, -1.f),
    C_Vertex3f(0.0151367188f, 0.464355469f, -1.f),
    C_Vertex3f(0.0151367188f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0151367188f, 0.464355469f, 0.f),
    C_Vertex3f(0.0151367188f, 0.464355469f, -1.f),
    C_Vertex3f(0.0732421875f, 0.464355469f, -1.f),
    C_Vertex3f(0.0732421875f, 0.464355469f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0732421875f, 0.464355469f, 0.f),
    C_Vertex3f(0.0732421875f, 0.464355469f, -1.f),
    C_Vertex3f(0.0732421875f, 0.579292297f, -1.f),
    C_Vertex3f(0.0732421875f, 0.579292297f, 0.f),
    C_Normal3f(-0.51471597f, 0.857360721f, 0.f),
    C_Vertex3f(0.0732421875f, 0.579292297f, 0.f),
    C_Vertex3f(0.0732421875f, 0.579292297f, -1.f),
    C_Vertex3f(0.151855469f, 0.626487732f, -1.f),
    C_Vertex3f(0.151855469f, 0.626487732f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.151855469f, 0.626487732f, 0.f),
    C_Vertex3f(0.151855469f, 0.626487732f, -1.f),
    C_Vertex3f(0.151855469f, 0.464355469f, -1.f),
    C_Vertex3f(0.151855469f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.151855469f, 0.464355469f, 0.f),
    C_Vertex3f(0.151855469f, 0.464355469f, -1.f),
    C_Vertex3f(0.23046875f, 0.464355469f, -1.f),
    C_Vertex3f(0.23046875f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.23046875f, 0.464355469f, 0.f),
    C_Vertex3f(0.23046875f, 0.464355469f, -1.f),
    C_Vertex3f(0.23046875f, 0.403320313f, -1.f),
    C_Vertex3f(0.23046875f, 0.403320313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.23046875f, 0.403320313f, 0.f),
    C_Vertex3f(0.23046875f, 0.403320313f, -1.f),
    C_Vertex3f(0.151855469f, 0.403320313f, -1.f),
    C_Vertex3f(0.151855469f, 0.403320313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.151855469f, 0.403320313f, 0.f),
    C_Vertex3f(0.151855469f, 0.403320313f, -1.f),
    C_Normal3f(0.99999845f, 0.00175298483f, 0.f),
    C_Vertex3f(0.151855469f, 0.122854233f, 0.f),
    C_Vertex3f(0.151855469f, 0.122854233f, -1.f),
    C_Normal3f(0.998041451f, 0.062555857f, 0.f),
    C_Vertex3f(0.152371407f, 0.10900116f, 0.f),
    C_Vertex3f(0.152371407f, 0.10900116f, -1.f),
    C_Normal3f(0.990589738f, 0.136865124f, 0.f),
    C_Vertex3f(0.153403282f, 0.09815979f, 0.f),
    C_Vertex3f(0.153403282f, 0.09815979f, -1.f),
    C_Normal3f(0.955589056f, 0.2947025f, 0.f),
    C_Vertex3f(0.154951096f, 0.0903301239f, 0.f),
    C_Vertex3f(0.154951096f, 0.0903301239f, -1.f),
    C_Normal3f(0.834623158f, 0.550821364f, 0.f),
    C_Vertex3f(0.158046722f, 0.0831031799f, 0.f),
    C_Vertex3f(0.158046722f, 0.0831031799f, -1.f),
    C_Normal3f(0.65593344f, 0.754818678f, 0.f),
    C_Vertex3f(0.164779663f, 0.0754375458f, 0.f),
    C_Vertex3f(0.164779663f, 0.0754375458f, -1.f),
    C_Normal3f(0.451076657f, 0.892485201f, 0.f),
    C_Vertex3f(0.171783447f, 0.0711660385f, 0.f),
    C_Vertex3f(0.171783447f, 0.0711660385f, -1.f),
    C_Normal3f(0.276930124f, 0.960889995f, 0.f),
    C_Vertex3f(0.177456856f, 0.0690302849f, 0.f),
    C_Vertex3f(0.177456856f, 0.0690302849f, -1.f),
    C_Normal3f(0.147116452f, 0.989119232f, 0.f),
    C_Vertex3f(0.184134483f, 0.0676064491f, 0.f),
    C_Vertex3f(0.184134483f, 0.0676064491f, -1.f),
    C_Normal3f(0.0465329997f, 0.998916745f, 0.f),
    C_Vertex3f(0.19181633f, 0.0668945313f, 0.f),
    C_Vertex3f(0.19181633f, 0.0668945313f, -1.f),
    C_Normal3f(-0.0265599135f, 0.999647141f, 0.f),
    C_Vertex3f(0.199417114f, 0.0668945313f, 0.f),
    C_Vertex3f(0.199417114f, 0.0668945313f, -1.f),
    C_Normal3f(-0.0746711716f, 0.997208118f, 0.f),
    C_Vertex3f(0.206963539f, 0.0672969818f, 0.f),
    C_Vertex3f(0.206963539f, 0.0672969818f, -1.f),
    C_Normal3f(-0.109979577f, 0.993933916f, 0.f),
    C_Vertex3f(0.215540886f, 0.0681018829f, 0.f),
    C_Vertex3f(0.215540886f, 0.0681018829f, -1.f),
    C_Normal3f(-0.133590192f, 0.991036654f, 0.f),
    C_Vertex3f(0.225149155f, 0.0693092346f, 0.f),
    C_Vertex3f(0.225149155f, 0.0693092346f, -1.f),
    C_Normal3f(-0.149605855f, 0.988745749f, 0.f),
    C_Vertex3f(0.23046875f, 0.0701141357f, 0.f),
    C_Vertex3f(0.23046875f, 0.0701141357f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151855469f, 0.626487732f, 0.f),
    C_Vertex3f(0.0732421875f, 0.579292297f, 0.f),
    C_Vertex3f(0.0735422373f, 0.110832214f, 0.f),
    C_Vertex3f(0.0732421875f, 0.464355469f, 0.f),
    C_Vertex3f(0.0732421875f, 0.403320313f, 0.f),
    C_Vertex3f(0.0151367188f, 0.464355469f, 0.f),
    C_Vertex3f(0.0151367188f, 0.403320313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.151855469f, 0.626487732f, 0.f),
    C_Vertex3f(0.0735422373f, 0.110832214f, 0.f),
    C_Vertex3f(0.0741423368f, 0.096157074f, 0.f),
    C_Vertex3f(0.0750424862f, 0.0830068588f, 0.f),
    C_Vertex3f(0.0762426853f, 0.0713815689f, 0.f),
    C_Vertex3f(0.0777429342f, 0.0612812042f, 0.f),
    C_Vertex3f(0.0795432329f, 0.0527057648f, 0.f),
    C_Vertex3f(0.0816435814f, 0.0456552505f, 0.f),
    C_Vertex3f(0.0852441788f, 0.0373668671f, 0.f),
    C_Vertex3f(0.0917906761f, 0.0273823738f, 0.f),
    C_Vertex3f(0.100082874f, 0.0184645653f, 0.f),
    C_Vertex3f(0.110120773f, 0.0106134415f, 0.f),
    C_Vertex3f(0.121904373f, 0.00382900238f, 0.f),
    C_Vertex3f(0.136088371f, -0.00125932693f, 0.f),
    C_Vertex3f(0.151855469f, 0.122854233f, 0.f),
    C_Vertex3f(0.151855469f, 0.403320313f, 0.f),
    C_Vertex3f(0.151855469f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.151855469f, 0.464355469f, 0.f),
    C_Vertex3f(0.151855469f, 0.403320313f, 0.f),
    C_Vertex3f(0.23046875f, 0.403320313f, 0.f),
    C_Vertex3f(0.23046875f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241699219f, 0.000625610352f, 0.f),
    C_Vertex3f(0.23046875f, 0.0701141357f, 0.f),
    C_Vertex3f(0.233407974f, -0.0011177063f, 0.f),
    C_Vertex3f(0.225149155f, 0.0693092346f, 0.f),
    C_Vertex3f(0.217698097f, -0.00373268127f, 0.f),
    C_Vertex3f(0.215540886f, 0.0681018829f, 0.f),
    C_Vertex3f(0.206963539f, 0.0672969818f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.202860832f, -0.00547599792f, 0.f),
    C_Vertex3f(0.217698097f, -0.00373268127f, 0.f),
    C_Vertex3f(0.206963539f, 0.0672969818f, 0.f),
    C_Vertex3f(0.199417114f, 0.0668945313f, 0.f),
    C_Vertex3f(0.19181633f, 0.0668945313f, 0.f),
    C_Vertex3f(0.188896179f, -0.00634765625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.188896179f, -0.00634765625f, 0.f),
    C_Vertex3f(0.19181633f, 0.0668945313f, 0.f),
    C_Vertex3f(0.184134483f, 0.0676064491f, 0.f),
    C_Vertex3f(0.177456856f, 0.0690302849f, 0.f),
    C_Vertex3f(0.171783447f, 0.0711660385f, 0.f),
    C_Vertex3f(0.171657562f, -0.00634765625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.171657562f, -0.00634765625f, 0.f),
    C_Vertex3f(0.171783447f, 0.0711660385f, 0.f),
    C_Vertex3f(0.164779663f, 0.0754375458f, 0.f),
    C_Vertex3f(0.158046722f, 0.0831031799f, 0.f),
    C_Vertex3f(0.154951096f, 0.0903301239f, 0.f),
    C_Vertex3f(0.153403282f, 0.09815979f, 0.f),
    C_Vertex3f(0.152672768f, -0.00465154648f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.152672768f, -0.00465154648f, 0.f),
    C_Vertex3f(0.153403282f, 0.09815979f, 0.f),
    C_Vertex3f(0.152371407f, 0.10900116f, 0.f),
    C_Vertex3f(0.151855469f, 0.122854233f, 0.f),
    C_Vertex3f(0.136088371f, -0.00125932693f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0735422373f, 0.110832214f, 0.f),
    C_Vertex3f(0.0732421875f, 0.403320313f, 0.f),
    C_Vertex3f(0.0732421875f, 0.12703228f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151855469f, 0.626487732f, -1.f),
    C_Vertex3f(0.0732421875f, 0.579292297f, -1.f),
    C_Vertex3f(0.0735422373f, 0.110832214f, -1.f),
    C_Vertex3f(0.0732421875f, 0.464355469f, -1.f),
    C_Vertex3f(0.0732421875f, 0.403320313f, -1.f),
    C_Vertex3f(0.0151367188f, 0.464355469f, -1.f),
    C_Vertex3f(0.0151367188f, 0.403320313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.151855469f, 0.626487732f, -1.f),
    C_Vertex3f(0.0735422373f, 0.110832214f, -1.f),
    C_Vertex3f(0.0741423368f, 0.096157074f, -1.f),
    C_Vertex3f(0.0750424862f, 0.0830068588f, -1.f),
    C_Vertex3f(0.0762426853f, 0.0713815689f, -1.f),
    C_Vertex3f(0.0777429342f, 0.0612812042f, -1.f),
    C_Vertex3f(0.0795432329f, 0.0527057648f, -1.f),
    C_Vertex3f(0.0816435814f, 0.0456552505f, -1.f),
    C_Vertex3f(0.0852441788f, 0.0373668671f, -1.f),
    C_Vertex3f(0.0917906761f, 0.0273823738f, -1.f),
    C_Vertex3f(0.100082874f, 0.0184645653f, -1.f),
    C_Vertex3f(0.110120773f, 0.0106134415f, -1.f),
    C_Vertex3f(0.121904373f, 0.00382900238f, -1.f),
    C_Vertex3f(0.136088371f, -0.00125932693f, -1.f),
    C_Vertex3f(0.151855469f, 0.122854233f, -1.f),
    C_Vertex3f(0.151855469f, 0.403320313f, -1.f),
    C_Vertex3f(0.151855469f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.151855469f, 0.464355469f, -1.f),
    C_Vertex3f(0.151855469f, 0.403320313f, -1.f),
    C_Vertex3f(0.23046875f, 0.403320313f, -1.f),
    C_Vertex3f(0.23046875f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241699219f, 0.000625610352f, -1.f),
    C_Vertex3f(0.23046875f, 0.0701141357f, -1.f),
    C_Vertex3f(0.233407974f, -0.0011177063f, -1.f),
    C_Vertex3f(0.225149155f, 0.0693092346f, -1.f),
    C_Vertex3f(0.217698097f, -0.00373268127f, -1.f),
    C_Vertex3f(0.215540886f, 0.0681018829f, -1.f),
    C_Vertex3f(0.206963539f, 0.0672969818f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.202860832f, -0.00547599792f, -1.f),
    C_Vertex3f(0.217698097f, -0.00373268127f, -1.f),
    C_Vertex3f(0.206963539f, 0.0672969818f, -1.f),
    C_Vertex3f(0.199417114f, 0.0668945313f, -1.f),
    C_Vertex3f(0.19181633f, 0.0668945313f, -1.f),
    C_Vertex3f(0.188896179f, -0.00634765625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.188896179f, -0.00634765625f, -1.f),
    C_Vertex3f(0.19181633f, 0.0668945313f, -1.f),
    C_Vertex3f(0.184134483f, 0.0676064491f, -1.f),
    C_Vertex3f(0.177456856f, 0.0690302849f, -1.f),
    C_Vertex3f(0.171783447f, 0.0711660385f, -1.f),
    C_Vertex3f(0.171657562f, -0.00634765625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.171657562f, -0.00634765625f, -1.f),
    C_Vertex3f(0.171783447f, 0.0711660385f, -1.f),
    C_Vertex3f(0.164779663f, 0.0754375458f, -1.f),
    C_Vertex3f(0.158046722f, 0.0831031799f, -1.f),
    C_Vertex3f(0.154951096f, 0.0903301239f, -1.f),
    C_Vertex3f(0.153403282f, 0.09815979f, -1.f),
    C_Vertex3f(0.152672768f, -0.00465154648f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.152672768f, -0.00465154648f, -1.f),
    C_Vertex3f(0.153403282f, 0.09815979f, -1.f),
    C_Vertex3f(0.152371407f, 0.10900116f, -1.f),
    C_Vertex3f(0.151855469f, 0.122854233f, -1.f),
    C_Vertex3f(0.136088371f, -0.00125932693f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0735422373f, 0.110832214f, -1.f),
    C_Vertex3f(0.0732421875f, 0.403320313f, -1.f),
    C_Vertex3f(0.0732421875f, 0.12703228f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_117[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.824172258f, -0.566339135f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Normal3f(0.782976985f, -0.622050703f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Normal3f(0.721546888f, -0.692365587f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_Normal3f(0.632779717f, -0.774331868f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, -1.f),
    C_Normal3f(0.528753817f, -0.848775268f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Normal3f(0.413056374f, -0.910705447f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, -1.f),
    C_Normal3f(0.291300356f, -0.95663166f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, -1.f),
    C_Normal3f(0.169845596f, -0.985470653f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, -1.f),
    C_Normal3f(0.0559700541f, -0.998432457f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0470732637f, -0.998891473f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, -1.f),
    C_Normal3f(-0.149569377f, -0.988751233f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, -1.f),
    C_Normal3f(-0.252776206f, -0.967524767f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Normal3f(-0.354831725f, -0.934930205f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Normal3f(-0.449254483f, -0.893403828f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Normal3f(-0.547086418f, -0.837076187f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, -1.f),
    C_Normal3f(-0.652715385f, -0.757603228f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, -1.f),
    C_Normal3f(-0.759380996f, -0.65064621f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, -1.f),
    C_Normal3f(-0.844308019f, -0.535858274f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, -1.f),
    C_Normal3f(-0.901026189f, -0.433764547f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, -1.f),
    C_Normal3f(-0.942368746f, -0.3345761f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, -1.f),
    C_Normal3f(-0.969121754f, -0.246582538f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, -1.f),
    C_Normal3f(-0.987194002f, -0.1595245f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, -1.f),
    C_Normal3f(-0.995961547f, -0.0897801891f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, -1.f),
    C_Normal3f(-0.99899143f, -0.0449032933f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Normal3f(0.999999881f, 0.000538040244f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_Normal3f(0.999861121f, 0.0166672915f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, -1.f),
    C_Normal3f(0.999531269f, 0.0306145567f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, -1.f),
    C_Normal3f(0.998860717f, 0.0477204137f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, -1.f),
    C_Normal3f(0.997604668f, 0.0691730455f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Normal3f(0.995301366f, 0.0968251824f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, -1.f),
    C_Normal3f(0.991019964f, 0.133714229f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, -1.f),
    C_Normal3f(0.984530509f, 0.175213054f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, -1.f),
    C_Normal3f(0.97341013f, 0.229069084f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, -1.f),
    C_Normal3f(0.942804158f, 0.333346993f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Normal3f(0.885687292f, 0.464282304f, 0.f),
    C_Vertex3f(0.14818573f, 0.102423668f, 0.f),
    C_Vertex3f(0.14818573f, 0.102423668f, -1.f),
    C_Normal3f(0.786936522f, 0.617033899f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_Normal3f(0.661581755f, 0.749873102f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, -1.f),
    C_Normal3f(0.519911349f, 0.854220212f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, -1.f),
    C_Normal3f(0.364257187f, 0.931298375f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, -1.f),
    C_Normal3f(0.208589152f, 0.978003383f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_Normal3f(0.065011777f, 0.997884512f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, -1.f),
    C_Normal3f(-0.0649489835f, 0.997888565f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, -1.f),
    C_Normal3f(-0.197600305f, 0.980282605f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, -1.f),
    C_Normal3f(-0.327999711f, 0.94467783f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, -1.f),
    C_Normal3f(-0.449308246f, 0.893376768f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_Normal3f(-0.566825271f, 0.823837996f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, -1.f),
    C_Normal3f(-0.689566791f, 0.724222064f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, -1.f),
    C_Normal3f(-0.800973952f, 0.598699152f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, -1.f),
    C_Normal3f(-0.890218019f, 0.455534756f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, -1.f),
    C_Normal3f(-0.951235354f, 0.308466077f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Normal3f(-0.981952488f, 0.189127818f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, -1.f),
    C_Normal3f(-0.995145023f, 0.0984196588f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14818573f, 0.102423668f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, -1.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, -1.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, -1.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, -1.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, -1.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, -1.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, -1.f),
    C_Vertex3f(0.101396084f, 0.028175354f, -1.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, -1.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_Vertex3f(0.352881908f, 0.174730778f, -1.f),
    C_Vertex3f(0.34966135f, 0.152068615f, -1.f),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Vertex3f(0.338389397f, 0.117308617f, -1.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_Vertex3f(0.329683781f, 0.103327751f, -1.f),
    C_Vertex3f(0.325308323f, 0.0266819f, -1.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, -1.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, -1.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, -1.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, -1.f),
    C_Vertex3f(0.270124912f, -0.002866745f, -1.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, -1.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, -1.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, -1.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, -1.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, -1.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, -1.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, -1.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, -1.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, -1.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14818573f, 0.102423668f, -1.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Vertex3f(0.140548706f, 0.124023438f, -1.f),
    C_Vertex3f(0.139947891f, 0.126699448f, -1.f),
    C_Vertex3f(0.138896465f, 0.133307457f, -1.f),
    C_Vertex3f(0.137995243f, 0.141171455f, -1.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Vertex3f(0.13664341f, 0.160667419f, -1.f),
    C_Vertex3f(0.136192799f, 0.172299385f, -1.f),
    C_Vertex3f(0.135892391f, 0.18518734f, -1.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_118[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.935192704f, -0.35413909f, 0.f),
    C_Vertex3f(0.187072754f, 0.f, 0.f),
    C_Vertex3f(0.187072754f, 0.f, -1.f),
    C_Vertex3f(0.0112304688f, 0.464355469f, -1.f),
    C_Vertex3f(0.0112304688f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0112304688f, 0.464355469f, 0.f),
    C_Vertex3f(0.0112304688f, 0.464355469f, -1.f),
    C_Vertex3f(0.09425354f, 0.464355469f, -1.f),
    C_Vertex3f(0.09425354f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.941578746f, 0.336792767f, 0.f),
    C_Vertex3f(0.09425354f, 0.464355469f, 0.f),
    C_Vertex3f(0.09425354f, 0.464355469f, -1.f),
    C_Normal3f(0.943195939f, 0.332237124f, 0.f),
    C_Vertex3f(0.201778412f, 0.16374588f, 0.f),
    C_Vertex3f(0.201778412f, 0.16374588f, -1.f),
    C_Normal3f(0.95672363f, 0.290998071f, 0.f),
    C_Vertex3f(0.216621399f, 0.116962433f, 0.f),
    C_Vertex3f(0.216621399f, 0.116962433f, -1.f),
    C_Normal3f(0.963241875f, 0.268635601f, 0.f),
    C_Vertex3f(0.223388672f, 0.0926971436f, 0.f),
    C_Vertex3f(0.223388672f, 0.0926971436f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.96159935f, 0.274457306f, 0.f),
    C_Vertex3f(0.223388672f, 0.0926971436f, 0.f),
    C_Vertex3f(0.223388672f, 0.0926971436f, -1.f),
    C_Normal3f(-0.957130671f, 0.289656729f, 0.f),
    C_Vertex3f(0.226009369f, 0.10187912f, 0.f),
    C_Vertex3f(0.226009369f, 0.10187912f, -1.f),
    C_Normal3f(-0.952061296f, 0.305907041f, 0.f),
    C_Vertex3f(0.232287407f, 0.122101784f, 0.f),
    C_Vertex3f(0.232287407f, 0.122101784f, -1.f),
    C_Normal3f(-0.946655393f, 0.322247714f, 0.f),
    C_Vertex3f(0.239602089f, 0.144183159f, 0.f),
    C_Vertex3f(0.239602089f, 0.144183159f, -1.f),
    C_Normal3f(-0.939785123f, 0.34176591f, 0.f),
    C_Vertex3f(0.247953415f, 0.168123245f, 0.f),
    C_Vertex3f(0.247953415f, 0.168123245f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.939785123f, 0.34176591f, 0.f),
    C_Vertex3f(0.247953415f, 0.168123245f, 0.f),
    C_Vertex3f(0.247953415f, 0.168123245f, -1.f),
    C_Vertex3f(0.355682373f, 0.464355469f, -1.f),
    C_Vertex3f(0.355682373f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.355682373f, 0.464355469f, 0.f),
    C_Vertex3f(0.355682373f, 0.464355469f, -1.f),
    C_Vertex3f(0.436523438f, 0.464355469f, -1.f),
    C_Vertex3f(0.436523438f, 0.464355469f, 0.f),
    C_Normal3f(0.935370743f, -0.35366863f, 0.f),
    C_Vertex3f(0.436523438f, 0.464355469f, 0.f),
    C_Vertex3f(0.436523438f, 0.464355469f, -1.f),
    C_Vertex3f(0.260948181f, 0.f, -1.f),
    C_Vertex3f(0.260948181f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.260948181f, 0.f, 0.f),
    C_Vertex3f(0.260948181f, 0.f, -1.f),
    C_Vertex3f(0.187072754f, 0.f, -1.f),
    C_Vertex3f(0.187072754f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.260948181f, 0.f, 0.f),
    C_Vertex3f(0.436523438f, 0.464355469f, 0.f),
    C_Vertex3f(0.355682373f, 0.464355469f, 0.f),
    C_Vertex3f(0.247953415f, 0.168123245f, 0.f),
    C_Vertex3f(0.239602089f, 0.144183159f, 0.f),
    C_Vertex3f(0.232287407f, 0.122101784f, 0.f),
    C_Vertex3f(0.226009369f, 0.10187912f, 0.f),
    C_Vertex3f(0.223388672f, 0.0926971436f, 0.f),
    C_Vertex3f(0.187072754f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.201778412f, 0.16374588f, 0.f),
    C_Vertex3f(0.09425354f, 0.464355469f, 0.f),
    C_Vertex3f(0.187072754f, 0.f, 0.f),
    C_Vertex3f(0.223388672f, 0.0926971436f, 0.f),
    C_Vertex3f(0.216621399f, 0.116962433f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.187072754f, 0.f, 0.f),
    C_Vertex3f(0.09425354f, 0.464355469f, 0.f),
    C_Vertex3f(0.0112304688f, 0.464355469f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.260948181f, 0.f, -1.f),
    C_Vertex3f(0.436523438f, 0.464355469f, -1.f),
    C_Vertex3f(0.355682373f, 0.464355469f, -1.f),
    C_Vertex3f(0.247953415f, 0.168123245f, -1.f),
    C_Vertex3f(0.239602089f, 0.144183159f, -1.f),
    C_Vertex3f(0.232287407f, 0.122101784f, -1.f),
    C_Vertex3f(0.226009369f, 0.10187912f, -1.f),
    C_Vertex3f(0.223388672f, 0.0926971436f, -1.f),
    C_Vertex3f(0.187072754f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.201778412f, 0.16374588f, -1.f),
    C_Vertex3f(0.09425354f, 0.464355469f, -1.f),
    C_Vertex3f(0.187072754f, 0.f, -1.f),
    C_Vertex3f(0.223388672f, 0.0926971436f, -1.f),
    C_Vertex3f(0.216621399f, 0.116962433f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.187072754f, 0.f, -1.f),
    C_Vertex3f(0.09425354f, 0.464355469f, -1.f),
    C_Vertex3f(0.0112304688f, 0.464355469f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_119[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.956629276f, -0.291307956f, 0.f),
    C_Vertex3f(0.143600464f, 0.f, 0.f),
    C_Vertex3f(0.143600464f, 0.f, -1.f),
    C_Vertex3f(0.00219726563f, 0.464355469f, -1.f),
    C_Vertex3f(0.00219726563f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.00219726563f, 0.464355469f, 0.f),
    C_Vertex3f(0.00219726563f, 0.464355469f, -1.f),
    C_Vertex3f(0.0834884644f, 0.464355469f, -1.f),
    C_Vertex3f(0.0834884644f, 0.464355469f, 0.f),
    C_Normal3f(0.963955045f, 0.26606521f, 0.f),
    C_Vertex3f(0.0834884644f, 0.464355469f, 0.f),
    C_Vertex3f(0.0834884644f, 0.464355469f, -1.f),
    C_Vertex3f(0.157516479f, 0.196151733f, -1.f),
    C_Vertex3f(0.157516479f, 0.196151733f, 0.f),
    C_Normal3f(0.964997351f, 0.262259692f, 0.f),
    C_Vertex3f(0.157516479f, 0.196151733f, 0.f),
    C_Vertex3f(0.157516479f, 0.196151733f, -1.f),
    C_Vertex3f(0.184562683f, 0.0966339111f, -1.f),
    C_Vertex3f(0.184562683f, 0.0966339111f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.974391222f, 0.224859506f, 0.f),
    C_Vertex3f(0.184562683f, 0.0966339111f, 0.f),
    C_Vertex3f(0.184562683f, 0.0966339111f, -1.f),
    C_Normal3f(-0.971657813f, 0.236391574f, 0.f),
    C_Vertex3f(0.18477726f, 0.0975637436f, 0.f),
    C_Vertex3f(0.18477726f, 0.0975637436f, -1.f),
    C_Normal3f(-0.970498979f, 0.241105184f, 0.f),
    C_Vertex3f(0.185856581f, 0.101952314f, 0.f),
    C_Vertex3f(0.185856581f, 0.101952314f, -1.f),
    C_Normal3f(-0.969888389f, 0.243549585f, 0.f),
    C_Vertex3f(0.187586069f, 0.108869791f, 0.f),
    C_Vertex3f(0.187586069f, 0.108869791f, -1.f),
    C_Normal3f(-0.96956408f, 0.24483788f, 0.f),
    C_Vertex3f(0.189965725f, 0.118316174f, 0.f),
    C_Vertex3f(0.189965725f, 0.118316174f, -1.f),
    C_Normal3f(-0.969362736f, 0.245633349f, 0.f),
    C_Vertex3f(0.192995548f, 0.130291462f, 0.f),
    C_Vertex3f(0.192995548f, 0.130291462f, -1.f),
    C_Normal3f(-0.969225824f, 0.246173397f, 0.f),
    C_Vertex3f(0.196675539f, 0.144795656f, 0.f),
    C_Vertex3f(0.196675539f, 0.144795656f, -1.f),
    C_Normal3f(-0.969126523f, 0.246564046f, 0.f),
    C_Vertex3f(0.201005697f, 0.161828756f, 0.f),
    C_Vertex3f(0.201005697f, 0.161828756f, -1.f),
    C_Normal3f(-0.969062448f, 0.246815503f, 0.f),
    C_Vertex3f(0.205986023f, 0.181390762f, 0.f),
    C_Vertex3f(0.205986023f, 0.181390762f, -1.f),
    C_Normal3f(-0.965142787f, 0.261723995f, 0.f),
    C_Vertex3f(0.20880127f, 0.192436218f, 0.f),
    C_Vertex3f(0.20880127f, 0.192436218f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.965142787f, 0.261723995f, 0.f),
    C_Vertex3f(0.20880127f, 0.192436218f, 0.f),
    C_Vertex3f(0.20880127f, 0.192436218f, -1.f),
    C_Vertex3f(0.282539368f, 0.464355469f, -1.f),
    C_Vertex3f(0.282539368f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.282539368f, 0.464355469f, 0.f),
    C_Vertex3f(0.282539368f, 0.464355469f, -1.f),
    C_Vertex3f(0.363632202f, 0.464355469f, -1.f),
    C_Vertex3f(0.363632202f, 0.464355469f, 0.f),
    C_Normal3f(0.968348026f, 0.249603629f, 0.f),
    C_Vertex3f(0.363632202f, 0.464355469f, 0.f),
    C_Vertex3f(0.363632202f, 0.464355469f, -1.f),
    C_Vertex3f(0.433113098f, 0.194801331f, -1.f),
    C_Vertex3f(0.433113098f, 0.194801331f, 0.f),
    C_Normal3f(0.968681872f, 0.248305053f, 0.f),
    C_Vertex3f(0.433113098f, 0.194801331f, 0.f),
    C_Vertex3f(0.433113098f, 0.194801331f, -1.f),
    C_Vertex3f(0.455810547f, 0.106254578f, -1.f),
    C_Vertex3f(0.455810547f, 0.106254578f, 0.f),
    C_Normal3f(-0.958571851f, 0.284850925f, 0.f),
    C_Vertex3f(0.455810547f, 0.106254578f, 0.f),
    C_Vertex3f(0.455810547f, 0.106254578f, -1.f),
    C_Vertex3f(0.48249054f, 0.196037292f, -1.f),
    C_Vertex3f(0.48249054f, 0.196037292f, 0.f),
    C_Normal3f(-0.958369792f, 0.285529822f, 0.f),
    C_Vertex3f(0.48249054f, 0.196037292f, 0.f),
    C_Vertex3f(0.48249054f, 0.196037292f, -1.f),
    C_Vertex3f(0.562431335f, 0.464355469f, -1.f),
    C_Vertex3f(0.562431335f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.562431335f, 0.464355469f, 0.f),
    C_Vertex3f(0.562431335f, 0.464355469f, -1.f),
    C_Vertex3f(0.638916016f, 0.464355469f, -1.f),
    C_Vertex3f(0.638916016f, 0.464355469f, 0.f),
    C_Normal3f(0.954153538f, -0.29931739f, 0.f),
    C_Vertex3f(0.638916016f, 0.464355469f, 0.f),
    C_Vertex3f(0.638916016f, 0.464355469f, -1.f),
    C_Vertex3f(0.493247986f, 0.f, -1.f),
    C_Vertex3f(0.493247986f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.493247986f, 0.f, 0.f),
    C_Vertex3f(0.493247986f, 0.f, -1.f),
    C_Vertex3f(0.411895752f, 0.f, -1.f),
    C_Vertex3f(0.411895752f, 0.f, 0.f),
    C_Normal3f(-0.966750205f, -0.255722553f, 0.f),
    C_Vertex3f(0.411895752f, 0.f, 0.f),
    C_Vertex3f(0.411895752f, 0.f, -1.f),
    C_Vertex3f(0.338394165f, 0.277870178f, -1.f),
    C_Vertex3f(0.338394165f, 0.277870178f, 0.f),
    C_Normal3f(-0.974359691f, -0.224996343f, 0.f),
    C_Vertex3f(0.338394165f, 0.277870178f, 0.f),
    C_Vertex3f(0.338394165f, 0.277870178f, -1.f),
    C_Vertex3f(0.320068359f, 0.35723114f, -1.f),
    C_Vertex3f(0.320068359f, 0.35723114f, 0.f),
    C_Normal3f(0.966875255f, -0.255249441f, 0.f),
    C_Vertex3f(0.320068359f, 0.35723114f, 0.f),
    C_Vertex3f(0.320068359f, 0.35723114f, -1.f),
    C_Vertex3f(0.225761414f, 0.f, -1.f),
    C_Vertex3f(0.225761414f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.225761414f, 0.f, 0.f),
    C_Vertex3f(0.225761414f, 0.f, -1.f),
    C_Vertex3f(0.143600464f, 0.f, -1.f),
    C_Vertex3f(0.143600464f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493247986f, 0.f, 0.f),
    C_Vertex3f(0.638916016f, 0.464355469f, 0.f),
    C_Vertex3f(0.562431335f, 0.464355469f, 0.f),
    C_Vertex3f(0.48249054f, 0.196037292f, 0.f),
    C_Vertex3f(0.455810547f, 0.106254578f, 0.f),
    C_Vertex3f(0.411895752f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320068359f, 0.35723114f, 0.f),
    C_Vertex3f(0.338394165f, 0.277870178f, 0.f),
    C_Vertex3f(0.363632202f, 0.464355469f, 0.f),
    C_Vertex3f(0.411895752f, 0.f, 0.f),
    C_Vertex3f(0.433113098f, 0.194801331f, 0.f),
    C_Vertex3f(0.455810547f, 0.106254578f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.282539368f, 0.464355469f, 0.f),
    C_Vertex3f(0.20880127f, 0.192436218f, 0.f),
    C_Vertex3f(0.225761414f, 0.f, 0.f),
    C_Vertex3f(0.320068359f, 0.35723114f, 0.f),
    C_Vertex3f(0.363632202f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.225761414f, 0.f, 0.f),
    C_Vertex3f(0.20880127f, 0.192436218f, 0.f),
    C_Vertex3f(0.205986023f, 0.181390762f, 0.f),
    C_Vertex3f(0.201005697f, 0.161828756f, 0.f),
    C_Vertex3f(0.196675539f, 0.144795656f, 0.f),
    C_Vertex3f(0.192995548f, 0.130291462f, 0.f),
    C_Vertex3f(0.189965725f, 0.118316174f, 0.f),
    C_Vertex3f(0.187586069f, 0.108869791f, 0.f),
    C_Vertex3f(0.185856581f, 0.101952314f, 0.f),
    C_Vertex3f(0.18477726f, 0.0975637436f, 0.f),
    C_Vertex3f(0.184562683f, 0.0966339111f, 0.f),
    C_Vertex3f(0.143600464f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.143600464f, 0.f, 0.f),
    C_Vertex3f(0.184562683f, 0.0966339111f, 0.f),
    C_Vertex3f(0.157516479f, 0.196151733f, 0.f),
    C_Vertex3f(0.0834884644f, 0.464355469f, 0.f),
    C_Vertex3f(0.00219726563f, 0.464355469f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493247986f, 0.f, -1.f),
    C_Vertex3f(0.638916016f, 0.464355469f, -1.f),
    C_Vertex3f(0.562431335f, 0.464355469f, -1.f),
    C_Vertex3f(0.48249054f, 0.196037292f, -1.f),
    C_Vertex3f(0.455810547f, 0.106254578f, -1.f),
    C_Vertex3f(0.411895752f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320068359f, 0.35723114f, -1.f),
    C_Vertex3f(0.338394165f, 0.277870178f, -1.f),
    C_Vertex3f(0.363632202f, 0.464355469f, -1.f),
    C_Vertex3f(0.411895752f, 0.f, -1.f),
    C_Vertex3f(0.433113098f, 0.194801331f, -1.f),
    C_Vertex3f(0.455810547f, 0.106254578f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.282539368f, 0.464355469f, -1.f),
    C_Vertex3f(0.20880127f, 0.192436218f, -1.f),
    C_Vertex3f(0.225761414f, 0.f, -1.f),
    C_Vertex3f(0.320068359f, 0.35723114f, -1.f),
    C_Vertex3f(0.363632202f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.225761414f, 0.f, -1.f),
    C_Vertex3f(0.20880127f, 0.192436218f, -1.f),
    C_Vertex3f(0.205986023f, 0.181390762f, -1.f),
    C_Vertex3f(0.201005697f, 0.161828756f, -1.f),
    C_Vertex3f(0.196675539f, 0.144795656f, -1.f),
    C_Vertex3f(0.192995548f, 0.130291462f, -1.f),
    C_Vertex3f(0.189965725f, 0.118316174f, -1.f),
    C_Vertex3f(0.187586069f, 0.108869791f, -1.f),
    C_Vertex3f(0.185856581f, 0.101952314f, -1.f),
    C_Vertex3f(0.18477726f, 0.0975637436f, -1.f),
    C_Vertex3f(0.184562683f, 0.0966339111f, -1.f),
    C_Vertex3f(0.143600464f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.143600464f, 0.f, -1.f),
    C_Vertex3f(0.184562683f, 0.0966339111f, -1.f),
    C_Vertex3f(0.157516479f, 0.196151733f, -1.f),
    C_Vertex3f(0.0834884644f, 0.464355469f, -1.f),
    C_Vertex3f(0.00219726563f, 0.464355469f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_120[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.816278398f, 0.577658772f, 0.f),
    C_Vertex3f(0.00634765625f, 0.f, 0.f),
    C_Vertex3f(0.00634765625f, 0.f, -1.f),
    C_Vertex3f(0.176193237f, 0.240005493f, -1.f),
    C_Vertex3f(0.176193237f, 0.240005493f, 0.f),
    C_Normal3f(-0.819051325f, -0.573720276f, 0.f),
    C_Vertex3f(0.176193237f, 0.240005493f, 0.f),
    C_Vertex3f(0.176193237f, 0.240005493f, -1.f),
    C_Vertex3f(0.0190429688f, 0.464355469f, -1.f),
    C_Vertex3f(0.0190429688f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0190429688f, 0.464355469f, 0.f),
    C_Vertex3f(0.0190429688f, 0.464355469f, -1.f),
    C_Vertex3f(0.116821289f, 0.464355469f, -1.f),
    C_Vertex3f(0.116821289f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.832796276f, 0.553579569f, 0.f),
    C_Vertex3f(0.116821289f, 0.464355469f, 0.f),
    C_Vertex3f(0.116821289f, 0.464355469f, -1.f),
    C_Normal3f(0.833608389f, 0.552355826f, 0.f),
    C_Vertex3f(0.194263458f, 0.347852707f, 0.f),
    C_Vertex3f(0.194263458f, 0.347852707f, -1.f),
    C_Normal3f(0.843319356f, 0.537412643f, 0.f),
    C_Vertex3f(0.203393936f, 0.333701134f, 0.f),
    C_Vertex3f(0.203393936f, 0.333701134f, -1.f),
    C_Normal3f(0.850319564f, 0.526266754f, 0.f),
    C_Vertex3f(0.211484909f, 0.320828438f, 0.f),
    C_Vertex3f(0.211484909f, 0.320828438f, -1.f),
    C_Normal3f(0.857342184f, 0.514746964f, 0.f),
    C_Vertex3f(0.218536377f, 0.309234619f, 0.f),
    C_Vertex3f(0.218536377f, 0.309234619f, -1.f),
    C_Normal3f(0.86396426f, 0.503553152f, 0.f),
    C_Vertex3f(0.221542358f, 0.304077148f, 0.f),
    C_Vertex3f(0.221542358f, 0.304077148f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.816228569f, 0.577729046f, 0.f),
    C_Vertex3f(0.221542358f, 0.304077148f, 0.f),
    C_Vertex3f(0.221542358f, 0.304077148f, -1.f),
    C_Normal3f(-0.81627357f, 0.577665567f, 0.f),
    C_Vertex3f(0.231300354f, 0.317863464f, 0.f),
    C_Vertex3f(0.231300354f, 0.317863464f, -1.f),
    C_Normal3f(-0.816299498f, 0.577628911f, 0.f),
    C_Vertex3f(0.249008179f, 0.342887878f, 0.f),
    C_Vertex3f(0.249008179f, 0.342887878f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.816299498f, 0.577628911f, 0.f),
    C_Vertex3f(0.249008179f, 0.342887878f, 0.f),
    C_Vertex3f(0.249008179f, 0.342887878f, -1.f),
    C_Vertex3f(0.334960938f, 0.464355469f, -1.f),
    C_Vertex3f(0.334960938f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.334960938f, 0.464355469f, 0.f),
    C_Vertex3f(0.334960938f, 0.464355469f, -1.f),
    C_Vertex3f(0.429634094f, 0.464355469f, -1.f),
    C_Vertex3f(0.429634094f, 0.464355469f, 0.f),
    C_Normal3f(0.816791892f, -0.57693249f, 0.f),
    C_Vertex3f(0.429634094f, 0.464355469f, 0.f),
    C_Vertex3f(0.429634094f, 0.464355469f, -1.f),
    C_Vertex3f(0.272293091f, 0.241600037f, -1.f),
    C_Vertex3f(0.272293091f, 0.241600037f, 0.f),
    C_Normal3f(0.820020676f, 0.572333872f, 0.f),
    C_Vertex3f(0.272293091f, 0.241600037f, 0.f),
    C_Vertex3f(0.272293091f, 0.241600037f, -1.f),
    C_Vertex3f(0.440917969f, 0.f, -1.f),
    C_Vertex3f(0.440917969f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.440917969f, 0.f, 0.f),
    C_Vertex3f(0.440917969f, 0.f, -1.f),
    C_Vertex3f(0.344314575f, 0.f, -1.f),
    C_Vertex3f(0.344314575f, 0.f, 0.f),
    C_Normal3f(-0.831812441f, -0.555056751f, 0.f),
    C_Vertex3f(0.344314575f, 0.f, 0.f),
    C_Vertex3f(0.344314575f, 0.f, -1.f),
    C_Vertex3f(0.248405457f, 0.143730164f, -1.f),
    C_Vertex3f(0.248405457f, 0.143730164f, 0.f),
    C_Normal3f(-0.844951868f, -0.534842432f, 0.f),
    C_Vertex3f(0.248405457f, 0.143730164f, 0.f),
    C_Vertex3f(0.248405457f, 0.143730164f, -1.f),
    C_Vertex3f(0.22618103f, 0.178840637f, -1.f),
    C_Vertex3f(0.22618103f, 0.178840637f, 0.f),
    C_Normal3f(0.820622563f, -0.571470618f, 0.f),
    C_Vertex3f(0.22618103f, 0.178840637f, 0.f),
    C_Vertex3f(0.22618103f, 0.178840637f, -1.f),
    C_Vertex3f(0.101638794f, 0.f, -1.f),
    C_Vertex3f(0.101638794f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.101638794f, 0.f, 0.f),
    C_Vertex3f(0.101638794f, 0.f, -1.f),
    C_Vertex3f(0.00634765625f, 0.f, -1.f),
    C_Vertex3f(0.00634765625f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.176193237f, 0.240005493f, 0.f),
    C_Vertex3f(0.00634765625f, 0.f, 0.f),
    C_Vertex3f(0.101638794f, 0.f, 0.f),
    C_Vertex3f(0.22618103f, 0.178840637f, 0.f),
    C_Vertex3f(0.194263458f, 0.347852707f, 0.f),
    C_Vertex3f(0.116821289f, 0.464355469f, 0.f),
    C_Vertex3f(0.0190429688f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.272293091f, 0.241600037f, 0.f),
    C_Vertex3f(0.429634094f, 0.464355469f, 0.f),
    C_Vertex3f(0.334960938f, 0.464355469f, 0.f),
    C_Vertex3f(0.249008179f, 0.342887878f, 0.f),
    C_Vertex3f(0.248405457f, 0.143730164f, 0.f),
    C_Vertex3f(0.344314575f, 0.f, 0.f),
    C_Vertex3f(0.440917969f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.231300354f, 0.317863464f, 0.f),
    C_Vertex3f(0.221542358f, 0.304077148f, 0.f),
    C_Vertex3f(0.22618103f, 0.178840637f, 0.f),
    C_Vertex3f(0.248405457f, 0.143730164f, 0.f),
    C_Vertex3f(0.249008179f, 0.342887878f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.22618103f, 0.178840637f, 0.f),
    C_Vertex3f(0.221542358f, 0.304077148f, 0.f),
    C_Vertex3f(0.218536377f, 0.309234619f, 0.f),
    C_Vertex3f(0.211484909f, 0.320828438f, 0.f),
    C_Vertex3f(0.203393936f, 0.333701134f, 0.f),
    C_Vertex3f(0.194263458f, 0.347852707f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.176193237f, 0.240005493f, -1.f),
    C_Vertex3f(0.00634765625f, 0.f, -1.f),
    C_Vertex3f(0.101638794f, 0.f, -1.f),
    C_Vertex3f(0.22618103f, 0.178840637f, -1.f),
    C_Vertex3f(0.194263458f, 0.347852707f, -1.f),
    C_Vertex3f(0.116821289f, 0.464355469f, -1.f),
    C_Vertex3f(0.0190429688f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.272293091f, 0.241600037f, -1.f),
    C_Vertex3f(0.429634094f, 0.464355469f, -1.f),
    C_Vertex3f(0.334960938f, 0.464355469f, -1.f),
    C_Vertex3f(0.249008179f, 0.342887878f, -1.f),
    C_Vertex3f(0.248405457f, 0.143730164f, -1.f),
    C_Vertex3f(0.344314575f, 0.f, -1.f),
    C_Vertex3f(0.440917969f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.231300354f, 0.317863464f, -1.f),
    C_Vertex3f(0.221542358f, 0.304077148f, -1.f),
    C_Vertex3f(0.22618103f, 0.178840637f, -1.f),
    C_Vertex3f(0.248405457f, 0.143730164f, -1.f),
    C_Vertex3f(0.249008179f, 0.342887878f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.22618103f, 0.178840637f, -1.f),
    C_Vertex3f(0.221542358f, 0.304077148f, -1.f),
    C_Vertex3f(0.218536377f, 0.309234619f, -1.f),
    C_Vertex3f(0.211484909f, 0.320828438f, -1.f),
    C_Vertex3f(0.203393936f, 0.333701134f, -1.f),
    C_Vertex3f(0.194263458f, 0.347852707f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_121[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.993022859f, -0.117922105f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_Vertex3f(0.046875f, -0.1043396f, -1.f),
    C_Vertex3f(0.046875f, -0.1043396f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.262003899f, 0.965066791f, 0.f),
    C_Vertex3f(0.046875f, -0.1043396f, 0.f),
    C_Vertex3f(0.046875f, -0.1043396f, -1.f),
    C_Normal3f(0.229912221f, 0.973211408f, 0.f),
    C_Vertex3f(0.0533103943f, -0.106086731f, 0.f),
    C_Vertex3f(0.0533103943f, -0.106086731f, -1.f),
    C_Normal3f(0.184329838f, 0.982864439f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, -1.f),
    C_Normal3f(0.120143346f, 0.992756546f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, -1.f),
    C_Normal3f(0.0401061773f, 0.999195457f, 0.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, 0.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, -1.f),
    C_Normal3f(-0.0471678302f, 0.998887002f, 0.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, 0.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, -1.f),
    C_Normal3f(-0.145652682f, 0.989335775f, 0.f),
    C_Vertex3f(0.110143661f, -0.110236168f, 0.f),
    C_Vertex3f(0.110143661f, -0.110236168f, -1.f),
    C_Normal3f(-0.267909348f, 0.963444173f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, -1.f),
    C_Normal3f(-0.409735411f, 0.912204444f, 0.f),
    C_Vertex3f(0.129777908f, -0.104776382f, 0.f),
    C_Vertex3f(0.129777908f, -0.104776382f, -1.f),
    C_Normal3f(-0.53943646f, 0.842026293f, 0.f),
    C_Vertex3f(0.13763237f, -0.100408554f, 0.f),
    C_Vertex3f(0.13763237f, -0.100408554f, -1.f),
    C_Normal3f(-0.641497552f, 0.76712507f, 0.f),
    C_Vertex3f(0.144777298f, -0.09516716f, 0.f),
    C_Vertex3f(0.144777298f, -0.09516716f, -1.f),
    C_Normal3f(-0.732968271f, 0.680262864f, 0.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, 0.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, -1.f),
    C_Normal3f(-0.804138839f, 0.594441473f, 0.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, 0.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, -1.f),
    C_Normal3f(-0.86107707f, 0.508474469f, 0.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, 0.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, -1.f),
    C_Normal3f(-0.895077646f, 0.445910394f, 0.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, 0.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, -1.f),
    C_Normal3f(-0.910975873f, 0.412459642f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, -1.f),
    C_Normal3f(-0.921322405f, 0.388799489f, 0.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, 0.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, -1.f),
    C_Normal3f(-0.928546607f, 0.37121591f, 0.f),
    C_Vertex3f(0.170659065f, -0.053943634f, 0.f),
    C_Vertex3f(0.170659065f, -0.053943634f, -1.f),
    C_Normal3f(-0.93385601f, 0.357649088f, 0.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, 0.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, -1.f),
    C_Normal3f(-0.937913001f, 0.346870661f, 0.f),
    C_Vertex3f(0.177517891f, -0.036034584f, 0.f),
    C_Vertex3f(0.177517891f, -0.036034584f, -1.f),
    C_Normal3f(-0.940624177f, 0.339449972f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, -1.f),
    C_Normal3f(-0.944727242f, 0.327857316f, 0.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, 0.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, -1.f),
    C_Normal3f(-0.942189276f, 0.335081249f, 0.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, 0.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, -1.f),
    C_Normal3f(-0.938051581f, 0.346495599f, 0.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, 0.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, -1.f),
    C_Normal3f(-0.935004234f, -0.354636669f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.935004234f, -0.354636669f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, -1.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, 0.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.94061166f, 0.339484304f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Normal3f(0.942783356f, 0.333405942f, 0.f),
    C_Vertex3f(0.20488739f, 0.170864105f, 0.f),
    C_Vertex3f(0.20488739f, 0.170864105f, -1.f),
    C_Normal3f(0.95868063f, 0.284484535f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, -1.f),
    C_Normal3f(0.966813982f, 0.255481392f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.968233764f, 0.250046819f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_Normal3f(-0.959497869f, 0.281715602f, 0.f),
    C_Vertex3f(0.235969543f, 0.115627289f, 0.f),
    C_Vertex3f(0.235969543f, 0.115627289f, -1.f),
    C_Normal3f(-0.938995123f, 0.343930542f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.938995123f, 0.343930542f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_Vertex3f(0.360839844f, 0.464355469f, -1.f),
    C_Vertex3f(0.360839844f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.360839844f, 0.464355469f, 0.f),
    C_Vertex3f(0.360839844f, 0.464355469f, -1.f),
    C_Vertex3f(0.439498901f, 0.464355469f, -1.f),
    C_Vertex3f(0.439498901f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.936978996f, -0.349385917f, 0.f),
    C_Vertex3f(0.439498901f, 0.464355469f, 0.f),
    C_Vertex3f(0.439498901f, 0.464355469f, -1.f),
    C_Normal3f(0.936883092f, -0.349642813f, 0.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, 0.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, -1.f),
    C_Normal3f(0.932654917f, -0.360769868f, 0.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, 0.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, -1.f),
    C_Normal3f(0.928989351f, -0.370106369f, 0.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, 0.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, -1.f),
    C_Normal3f(0.92445457f, -0.381292284f, 0.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, 0.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, -1.f),
    C_Normal3f(0.918711185f, -0.394930184f, 0.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, 0.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, -1.f),
    C_Normal3f(0.911223829f, -0.411911488f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, -1.f),
    C_Normal3f(0.901101232f, -0.433608681f, 0.f),
    C_Vertex3f(0.225468159f, -0.100803614f, 0.f),
    C_Vertex3f(0.225468159f, -0.100803614f, -1.f),
    C_Normal3f(0.88511163f, -0.465378761f, 0.f),
    C_Vertex3f(0.221137047f, -0.109474182f, 0.f),
    C_Vertex3f(0.221137047f, -0.109474182f, -1.f),
    C_Normal3f(0.852197766f, -0.523219883f, 0.f),
    C_Vertex3f(0.213924408f, -0.122758865f, 0.f),
    C_Vertex3f(0.213924408f, -0.122758865f, -1.f),
    C_Normal3f(0.799718082f, -0.600375772f, 0.f),
    C_Vertex3f(0.20266819f, -0.139555454f, 0.f),
    C_Vertex3f(0.20266819f, -0.139555454f, -1.f),
    C_Normal3f(0.721279263f, -0.692644298f, 0.f),
    C_Vertex3f(0.190646172f, -0.153766155f, 0.f),
    C_Vertex3f(0.190646172f, -0.153766155f, -1.f),
    C_Normal3f(0.617209971f, -0.786798477f, 0.f),
    C_Vertex3f(0.177858353f, -0.165390968f, 0.f),
    C_Vertex3f(0.177858353f, -0.165390968f, -1.f),
    C_Normal3f(0.486348569f, -0.873764873f, 0.f),
    C_Vertex3f(0.164304733f, -0.174429893f, 0.f),
    C_Vertex3f(0.164304733f, -0.174429893f, -1.f),
    C_Normal3f(0.342036247f, -0.939686716f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, -1.f),
    C_Normal3f(0.197416425f, -0.980319679f, 0.f),
    C_Vertex3f(0.133263588f, -0.18572855f, 0.f),
    C_Vertex3f(0.133263588f, -0.18572855f, -1.f),
    C_Normal3f(0.0692126006f, -0.997601986f, 0.f),
    C_Vertex3f(0.115776062f, -0.187988281f, 0.f),
    C_Vertex3f(0.115776062f, -0.187988281f, -1.f),
    C_Normal3f(-0.0443920419f, -0.999014199f, 0.f),
    C_Vertex3f(0.100692749f, -0.187988281f, 0.f),
    C_Vertex3f(0.100692749f, -0.187988281f, -1.f),
    C_Normal3f(-0.144747168f, -0.989468634f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, -1.f),
    C_Normal3f(-0.225563779f, -0.974228382f, 0.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, 0.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, -1.f),
    C_Normal3f(-0.284065366f, -0.958804965f, 0.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, 0.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, -1.f),
    C_Normal3f(-0.325648993f, -0.945490777f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.240839005f, -0.0658898354f, 0.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, 0.f),
    C_Vertex3f(0.360839844f, 0.464355469f, 0.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, 0.f),
    C_Vertex3f(0.439498901f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235969543f, 0.115627289f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, 0.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, 0.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213924408f, -0.122758865f, 0.f),
    C_Vertex3f(0.221137047f, -0.109474182f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, 0.f),
    C_Vertex3f(0.225468159f, -0.100803614f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.20488739f, 0.170864105f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.190646172f, -0.153766155f, 0.f),
    C_Vertex3f(0.20266819f, -0.139555454f, 0.f),
    C_Vertex3f(0.213924408f, -0.122758865f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190646172f, -0.153766155f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, 0.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, 0.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, 0.f),
    C_Vertex3f(0.177858353f, -0.165390968f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.177858353f, -0.165390968f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, 0.f),
    C_Vertex3f(0.177517891f, -0.036034584f, 0.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, 0.f),
    C_Vertex3f(0.170659065f, -0.053943634f, 0.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, 0.f),
    C_Vertex3f(0.164304733f, -0.174429893f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.164304733f, -0.174429893f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, 0.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, 0.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, 0.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, 0.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151212692f, -0.0890522003f, 0.f),
    C_Vertex3f(0.144777298f, -0.09516716f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, 0.f),
    C_Vertex3f(0.13763237f, -0.100408554f, 0.f),
    C_Vertex3f(0.133263588f, -0.18572855f, 0.f),
    C_Vertex3f(0.129777908f, -0.104776382f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133263588f, -0.18572855f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, 0.f),
    C_Vertex3f(0.115776062f, -0.187988281f, 0.f),
    C_Vertex3f(0.110143661f, -0.110236168f, 0.f),
    C_Vertex3f(0.100692749f, -0.187988281f, 0.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, 0.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, 0.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0533103943f, -0.106086731f, 0.f),
    C_Vertex3f(0.046875f, -0.1043396f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.240839005f, -0.0658898354f, -1.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, -1.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, -1.f),
    C_Vertex3f(0.360839844f, 0.464355469f, -1.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, -1.f),
    C_Vertex3f(0.439498901f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235969543f, 0.115627289f, -1.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, -1.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, -1.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, -1.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213924408f, -0.122758865f, -1.f),
    C_Vertex3f(0.221137047f, -0.109474182f, -1.f),
    C_Vertex3f(0.22170639f, 0.117263794f, -1.f),
    C_Vertex3f(0.225468159f, -0.100803614f, -1.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.20488739f, 0.170864105f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.190646172f, -0.153766155f, -1.f),
    C_Vertex3f(0.20266819f, -0.139555454f, -1.f),
    C_Vertex3f(0.213924408f, -0.122758865f, -1.f),
    C_Vertex3f(0.22170639f, 0.117263794f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190646172f, -0.153766155f, -1.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, -1.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, -1.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, -1.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, -1.f),
    C_Vertex3f(0.177858353f, -0.165390968f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.177858353f, -0.165390968f, -1.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, -1.f),
    C_Vertex3f(0.177517891f, -0.036034584f, -1.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, -1.f),
    C_Vertex3f(0.170659065f, -0.053943634f, -1.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, -1.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, -1.f),
    C_Vertex3f(0.164304733f, -0.174429893f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.164304733f, -0.174429893f, -1.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, -1.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, -1.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, -1.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, -1.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, -1.f),
    C_Vertex3f(0.149439812f, -0.181209087f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151212692f, -0.0890522003f, -1.f),
    C_Vertex3f(0.144777298f, -0.09516716f, -1.f),
    C_Vertex3f(0.149439812f, -0.181209087f, -1.f),
    C_Vertex3f(0.13763237f, -0.100408554f, -1.f),
    C_Vertex3f(0.133263588f, -0.18572855f, -1.f),
    C_Vertex3f(0.129777908f, -0.104776382f, -1.f),
    C_Vertex3f(0.120615005f, -0.108052254f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133263588f, -0.18572855f, -1.f),
    C_Vertex3f(0.120615005f, -0.108052254f, -1.f),
    C_Vertex3f(0.115776062f, -0.187988281f, -1.f),
    C_Vertex3f(0.110143661f, -0.110236168f, -1.f),
    C_Vertex3f(0.100692749f, -0.187988281f, -1.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, -1.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, -1.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, -1.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, -1.f),
    C_Vertex3f(0.065363884f, -0.108707428f, -1.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, -1.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, -1.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0533103943f, -0.106086731f, -1.f),
    C_Vertex3f(0.046875f, -0.1043396f, -1.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_Vertex3f(0.065363884f, -0.108707428f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_122[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0174789429f, 0.f, 0.f),
    C_Vertex3f(0.0174789429f, 0.f, -1.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, -1.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, 0.f),
    C_Normal3f(-0.754504204f, 0.65629518f, 0.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, 0.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, -1.f),
    C_Vertex3f(0.312271118f, 0.402832031f, -1.f),
    C_Vertex3f(0.312271118f, 0.402832031f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.0486485697f, -0.998815954f, 0.f),
    C_Vertex3f(0.312271118f, 0.402832031f, 0.f),
    C_Vertex3f(0.312271118f, 0.402832031f, -1.f),
    C_Normal3f(0.0422043577f, -0.99910903f, 0.f),
    C_Vertex3f(0.299739838f, 0.40222168f, 0.f),
    C_Vertex3f(0.299739838f, 0.40222168f, -1.f),
    C_Normal3f(0.0333627351f, -0.999443233f, 0.f),
    C_Vertex3f(0.276148796f, 0.401306152f, 0.f),
    C_Vertex3f(0.276148796f, 0.401306152f, -1.f),
    C_Normal3f(0.0214021672f, -0.999770999f, 0.f),
    C_Vertex3f(0.254029274f, 0.400695801f, 0.f),
    C_Vertex3f(0.254029274f, 0.400695801f, -1.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.233381271f, 0.400390625f, 0.f),
    C_Vertex3f(0.233381271f, 0.400390625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.233381271f, 0.400390625f, 0.f),
    C_Vertex3f(0.233381271f, 0.400390625f, -1.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, -1.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, 0.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, -1.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, -1.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, 0.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, -1.f),
    C_Vertex3f(0.413963318f, 0.464355469f, -1.f),
    C_Vertex3f(0.413963318f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.413963318f, 0.464355469f, 0.f),
    C_Vertex3f(0.413963318f, 0.464355469f, -1.f),
    C_Vertex3f(0.413963318f, 0.412063599f, -1.f),
    C_Vertex3f(0.413963318f, 0.412063599f, 0.f),
    C_Normal3f(0.760531485f, -0.649300992f, 0.f),
    C_Vertex3f(0.413963318f, 0.412063599f, 0.f),
    C_Vertex3f(0.413963318f, 0.412063599f, -1.f),
    C_Vertex3f(0.162734985f, 0.117797852f, -1.f),
    C_Vertex3f(0.162734985f, 0.117797852f, 0.f),
    C_Normal3f(0.744730413f, -0.667365432f, 0.f),
    C_Vertex3f(0.162734985f, 0.117797852f, 0.f),
    C_Vertex3f(0.162734985f, 0.117797852f, -1.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, -1.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0737020075f, 0.9972803f, 0.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, 0.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, -1.f),
    C_Normal3f(-0.0513249896f, 0.998682022f, 0.f),
    C_Vertex3f(0.140922546f, 0.0659179688f, 0.f),
    C_Vertex3f(0.140922546f, 0.0659179688f, -1.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, 0.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, 0.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, -1.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, -1.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, 0.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, -1.f),
    C_Vertex3f(0.428123474f, 0.f, -1.f),
    C_Vertex3f(0.428123474f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.428123474f, 0.f, 0.f),
    C_Vertex3f(0.428123474f, 0.f, -1.f),
    C_Vertex3f(0.0174789429f, 0.f, -1.f),
    C_Vertex3f(0.0174789429f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.428123474f, 0.f, 0.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, 0.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, 0.f),
    C_Vertex3f(0.140922546f, 0.0659179688f, 0.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, 0.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, 0.f),
    C_Vertex3f(0.0174789429f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.413963318f, 0.464355469f, 0.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, 0.f),
    C_Vertex3f(0.233381271f, 0.400390625f, 0.f),
    C_Vertex3f(0.254029274f, 0.400695801f, 0.f),
    C_Vertex3f(0.276148796f, 0.401306152f, 0.f),
    C_Vertex3f(0.299739838f, 0.40222168f, 0.f),
    C_Vertex3f(0.312271118f, 0.402832031f, 0.f),
    C_Vertex3f(0.413963318f, 0.412063599f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.312271118f, 0.402832031f, 0.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, 0.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, 0.f),
    C_Vertex3f(0.162734985f, 0.117797852f, 0.f),
    C_Vertex3f(0.413963318f, 0.412063599f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.233381271f, 0.400390625f, 0.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, 0.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.428123474f, 0.f, -1.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, -1.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, -1.f),
    C_Vertex3f(0.140922546f, 0.0659179688f, -1.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, -1.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, -1.f),
    C_Vertex3f(0.0174789429f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.413963318f, 0.464355469f, -1.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, -1.f),
    C_Vertex3f(0.233381271f, 0.400390625f, -1.f),
    C_Vertex3f(0.254029274f, 0.400695801f, -1.f),
    C_Vertex3f(0.276148796f, 0.401306152f, -1.f),
    C_Vertex3f(0.299739838f, 0.40222168f, -1.f),
    C_Vertex3f(0.312271118f, 0.402832031f, -1.f),
    C_Vertex3f(0.413963318f, 0.412063599f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.312271118f, 0.402832031f, -1.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, -1.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, -1.f),
    C_Vertex3f(0.162734985f, 0.117797852f, -1.f),
    C_Vertex3f(0.413963318f, 0.412063599f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.233381271f, 0.400390625f, -1.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, -1.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_123[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0257483516f, 0.999668479f, 0.f),
    C_Vertex3f(0.0249023438f, 0.267578125f, 0.f),
    C_Vertex3f(0.0249023438f, 0.267578125f, -1.f),
    C_Normal3f(-0.112888664f, 0.99360764f, 0.f),
    C_Vertex3f(0.0333442688f, 0.267795563f, 0.f),
    C_Vertex3f(0.0333442688f, 0.267795563f, -1.f),
    C_Normal3f(-0.234421864f, 0.972135007f, 0.f),
    C_Vertex3f(0.0486655235f, 0.270277977f, 0.f),
    C_Vertex3f(0.0486655235f, 0.270277977f, -1.f),
    C_Normal3f(-0.393440217f, 0.919350207f, 0.f),
    C_Vertex3f(0.0624241829f, 0.27480793f, 0.f),
    C_Vertex3f(0.0624241829f, 0.27480793f, -1.f),
    C_Normal3f(-0.55426687f, 0.832339048f, 0.f),
    C_Vertex3f(0.0746202469f, 0.281385422f, 0.f),
    C_Vertex3f(0.0746202469f, 0.281385422f, -1.f),
    C_Normal3f(-0.695468605f, 0.718556523f, 0.f),
    C_Vertex3f(0.0852537155f, 0.290010452f, 0.f),
    C_Vertex3f(0.0852537155f, 0.290010452f, -1.f),
    C_Normal3f(-0.811074972f, 0.584942341f, 0.f),
    C_Vertex3f(0.0941057205f, 0.300244808f, 0.f),
    C_Vertex3f(0.0941057205f, 0.300244808f, -1.f),
    C_Normal3f(-0.898491502f, 0.438990772f, 0.f),
    C_Vertex3f(0.101176262f, 0.31208849f, 0.f),
    C_Vertex3f(0.101176262f, 0.31208849f, -1.f),
    C_Normal3f(-0.952425659f, 0.304771066f, 0.f),
    C_Vertex3f(0.10646534f, 0.325541496f, 0.f),
    C_Vertex3f(0.10646534f, 0.325541496f, -1.f),
    C_Normal3f(-0.979199469f, 0.202899784f, 0.f),
    C_Vertex3f(0.10909605f, 0.336838245f, 0.f),
    C_Vertex3f(0.10909605f, 0.336838245f, -1.f),
    C_Normal3f(-0.988720834f, 0.149769858f, 0.f),
    C_Vertex3f(0.110643744f, 0.345706582f, 0.f),
    C_Vertex3f(0.110643744f, 0.345706582f, -1.f),
    C_Normal3f(-0.993576229f, 0.113165207f, 0.f),
    C_Vertex3f(0.111985326f, 0.355912089f, 0.f),
    C_Vertex3f(0.111985326f, 0.355912089f, -1.f),
    C_Normal3f(-0.996445f, 0.0842453241f, 0.f),
    C_Vertex3f(0.113120794f, 0.367454767f, 0.f),
    C_Vertex3f(0.113120794f, 0.367454767f, -1.f),
    C_Normal3f(-0.998145342f, 0.060875427f, 0.f),
    C_Vertex3f(0.11405015f, 0.380334616f, 0.f),
    C_Vertex3f(0.11405015f, 0.380334616f, -1.f),
    C_Normal3f(-0.999133289f, 0.0416273847f, 0.f),
    C_Vertex3f(0.114773393f, 0.394551635f, 0.f),
    C_Vertex3f(0.114773393f, 0.394551635f, -1.f),
    C_Normal3f(-0.99967438f, 0.0255158953f, 0.f),
    C_Vertex3f(0.115290523f, 0.410105824f, 0.f),
    C_Vertex3f(0.115290523f, 0.410105824f, -1.f),
    C_Normal3f(-0.999929845f, 0.0118420655f, 0.f),
    C_Vertex3f(0.11560154f, 0.426997185f, 0.f),
    C_Vertex3f(0.11560154f, 0.426997185f, -1.f),
    C_Normal3f(-0.999971807f, 0.00750718312f, 0.f),
    C_Vertex3f(0.115706444f, 0.445225716f, 0.f),
    C_Vertex3f(0.115706444f, 0.445225716f, -1.f),
    C_Normal3f(-0.999933541f, 0.0115258982f, 0.f),
    C_Vertex3f(0.115863562f, 0.461899042f, 0.f),
    C_Vertex3f(0.115863562f, 0.461899042f, -1.f),
    C_Normal3f(-0.999865294f, 0.016415488f, 0.f),
    C_Vertex3f(0.116072893f, 0.477017164f, 0.f),
    C_Vertex3f(0.116072893f, 0.477017164f, -1.f),
    C_Normal3f(-0.999746978f, 0.0224929322f, 0.f),
    C_Vertex3f(0.116334438f, 0.490580082f, 0.f),
    C_Vertex3f(0.116334438f, 0.490580082f, -1.f),
    C_Normal3f(-0.999542356f, 0.0302499365f, 0.f),
    C_Vertex3f(0.116648197f, 0.502587795f, 0.f),
    C_Vertex3f(0.116648197f, 0.502587795f, -1.f),
    C_Normal3f(-0.99917984f, 0.0404921509f, 0.f),
    C_Vertex3f(0.11701417f, 0.513040304f, 0.f),
    C_Vertex3f(0.11701417f, 0.513040304f, -1.f),
    C_Normal3f(-0.998506308f, 0.0546361655f, 0.f),
    C_Vertex3f(0.117432356f, 0.521937609f, 0.f),
    C_Vertex3f(0.117432356f, 0.521937609f, -1.f),
    C_Normal3f(-0.997454584f, 0.0713048428f, 0.f),
    C_Vertex3f(0.117902756f, 0.529279709f, 0.f),
    C_Vertex3f(0.117902756f, 0.529279709f, -1.f),
    C_Normal3f(-0.994727314f, 0.102555692f, 0.f),
    C_Vertex3f(0.118164063f, 0.532173157f, 0.f),
    C_Vertex3f(0.118164063f, 0.532173157f, -1.f),
    C_Normal3f(-0.989296973f, 0.145916119f, 0.f),
    C_Vertex3f(0.119146347f, 0.541341782f, 0.f),
    C_Vertex3f(0.119146347f, 0.541341782f, -1.f),
    C_Normal3f(-0.97930634f, 0.202383906f, 0.f),
    C_Vertex3f(0.121955395f, 0.557878017f, 0.f),
    C_Vertex3f(0.121955395f, 0.557878017f, -1.f),
    C_Normal3f(-0.959239006f, 0.282595873f, 0.f),
    C_Vertex3f(0.125608921f, 0.572613239f, 0.f),
    C_Vertex3f(0.125608921f, 0.572613239f, -1.f),
    C_Normal3f(-0.925617933f, 0.378459334f, 0.f),
    C_Vertex3f(0.130106926f, 0.585547447f, 0.f),
    C_Vertex3f(0.130106926f, 0.585547447f, -1.f),
    C_Normal3f(-0.881768882f, 0.471681744f, 0.f),
    C_Vertex3f(0.135449409f, 0.596680641f, 0.f),
    C_Vertex3f(0.135449409f, 0.596680641f, -1.f),
    C_Normal3f(-0.832850754f, 0.553497732f, 0.f),
    C_Vertex3f(0.141419411f, 0.606695175f, 0.f),
    C_Vertex3f(0.141419411f, 0.606695175f, -1.f),
    C_Normal3f(-0.769846618f, 0.638228953f, 0.f),
    C_Vertex3f(0.14801693f, 0.615591049f, 0.f),
    C_Vertex3f(0.14801693f, 0.615591049f, -1.f),
    C_Normal3f(-0.691565871f, 0.722313285f, 0.f),
    C_Vertex3f(0.155241966f, 0.623368263f, 0.f),
    C_Vertex3f(0.155241966f, 0.623368263f, -1.f),
    C_Normal3f(-0.597816706f, 0.801632881f, 0.f),
    C_Vertex3f(0.163094521f, 0.630026817f, 0.f),
    C_Vertex3f(0.163094521f, 0.630026817f, -1.f),
    C_Normal3f(-0.49952206f, 0.866301119f, 0.f),
    C_Vertex3f(0.172038555f, 0.635894299f, 0.f),
    C_Vertex3f(0.172038555f, 0.635894299f, -1.f),
    C_Normal3f(-0.404556721f, 0.914512932f, 0.f),
    C_Vertex3f(0.18207407f, 0.640970707f, 0.f),
    C_Vertex3f(0.18207407f, 0.640970707f, -1.f),
    C_Normal3f(-0.330339968f, 0.943862081f, 0.f),
    C_Vertex3f(0.193201065f, 0.645256042f, 0.f),
    C_Vertex3f(0.193201065f, 0.645256042f, -1.f),
    C_Normal3f(-0.266802222f, 0.963751256f, 0.f),
    C_Vertex3f(0.199310303f, 0.647003174f, 0.f),
    C_Vertex3f(0.199310303f, 0.647003174f, -1.f),
    C_Normal3f(-0.17835857f, 0.983965516f, 0.f),
    C_Vertex3f(0.203453064f, 0.648094177f, 0.f),
    C_Vertex3f(0.203453064f, 0.648094177f, -1.f),
    C_Normal3f(-0.110968083f, 0.993823946f, 0.f),
    C_Vertex3f(0.214357376f, 0.649730682f, 0.f),
    C_Vertex3f(0.214357376f, 0.649730682f, -1.f),
    C_Normal3f(-0.0550824516f, 0.99848181f, 0.f),
    C_Vertex3f(0.227880478f, 0.650821686f, 0.f),
    C_Vertex3f(0.227880478f, 0.650821686f, -1.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.244022369f, 0.651367188f, 0.f),
    C_Vertex3f(0.244022369f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.244022369f, 0.651367188f, 0.f),
    C_Vertex3f(0.244022369f, 0.651367188f, -1.f),
    C_Vertex3f(0.277832031f, 0.651367188f, -1.f),
    C_Vertex3f(0.277832031f, 0.651367188f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.277832031f, 0.651367188f, 0.f),
    C_Vertex3f(0.277832031f, 0.651367188f, -1.f),
    C_Vertex3f(0.277832031f, 0.582519531f, -1.f),
    C_Vertex3f(0.277832031f, 0.582519531f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.277832031f, 0.582519531f, 0.f),
    C_Vertex3f(0.277832031f, 0.582519531f, -1.f),
    C_Normal3f(0.0173150823f, -0.999850094f, 0.f),
    C_Vertex3f(0.258675575f, 0.582519531f, 0.f),
    C_Vertex3f(0.258675575f, 0.582519531f, -1.f),
    C_Normal3f(0.0773033872f, -0.997007608f, 0.f),
    C_Vertex3f(0.248307943f, 0.582008243f, 0.f),
    C_Vertex3f(0.248307943f, 0.582008243f, -1.f),
    C_Normal3f(0.141555354f, -0.989930332f, 0.f),
    C_Vertex3f(0.238892794f, 0.580985665f, 0.f),
    C_Vertex3f(0.238892794f, 0.580985665f, -1.f),
    C_Normal3f(0.218647376f, -0.975803912f, 0.f),
    C_Vertex3f(0.230430126f, 0.579451799f, 0.f),
    C_Vertex3f(0.230430126f, 0.579451799f, -1.f),
    C_Normal3f(0.310890257f, -0.950445831f, 0.f),
    C_Vertex3f(0.222919941f, 0.577406645f, 0.f),
    C_Vertex3f(0.222919941f, 0.577406645f, -1.f),
    C_Normal3f(0.419705123f, -0.907660484f, 0.f),
    C_Vertex3f(0.216362238f, 0.574850202f, 0.f),
    C_Vertex3f(0.216362238f, 0.574850202f, -1.f),
    C_Normal3f(0.543784559f, -0.839224815f, 0.f),
    C_Vertex3f(0.210757017f, 0.57178247f, 0.f),
    C_Vertex3f(0.210757017f, 0.57178247f, -1.f),
    C_Normal3f(0.67631948f, -0.736608446f, 0.f),
    C_Vertex3f(0.206104279f, 0.568203449f, 0.f),
    C_Vertex3f(0.206104279f, 0.568203449f, -1.f),
    C_Normal3f(0.805840194f, -0.592133105f, 0.f),
    C_Vertex3f(0.202404022f, 0.56411314f, 0.f),
    C_Vertex3f(0.202404022f, 0.56411314f, -1.f),
    C_Normal3f(0.893511891f, -0.449039429f, 0.f),
    C_Vertex3f(0.199166298f, 0.558761477f, 0.f),
    C_Vertex3f(0.199166298f, 0.558761477f, -1.f),
    C_Normal3f(0.943507254f, -0.331352085f, 0.f),
    C_Vertex3f(0.196391106f, 0.552148461f, 0.f),
    C_Vertex3f(0.196391106f, 0.552148461f, -1.f),
    C_Normal3f(0.971336186f, -0.237709865f, 0.f),
    C_Vertex3f(0.194078445f, 0.544274092f, 0.f),
    C_Vertex3f(0.194078445f, 0.544274092f, -1.f),
    C_Normal3f(0.986538768f, -0.163527027f, 0.f),
    C_Vertex3f(0.192228317f, 0.535138369f, 0.f),
    C_Vertex3f(0.192228317f, 0.535138369f, -1.f),
    C_Normal3f(0.994547486f, -0.104284637f, 0.f),
    C_Vertex3f(0.190840721f, 0.524741292f, 0.f),
    C_Vertex3f(0.190840721f, 0.524741292f, -1.f),
    C_Normal3f(0.998410165f, -0.0563665889f, 0.f),
    C_Vertex3f(0.189915657f, 0.513082862f, 0.f),
    C_Vertex3f(0.189915657f, 0.513082862f, -1.f),
    C_Normal3f(0.999909103f, -0.0134835914f, 0.f),
    C_Vertex3f(0.189453125f, 0.500163078f, 0.f),
    C_Vertex3f(0.189453125f, 0.500163078f, -1.f),
    C_Normal3f(0.999994874f, -0.00319881481f, 0.f),
    C_Vertex3f(0.189453125f, 0.478782654f, 0.f),
    C_Vertex3f(0.189453125f, 0.478782654f, -1.f),
    C_Normal3f(0.999957502f, -0.00921898615f, 0.f),
    C_Vertex3f(0.189301729f, 0.452834606f, 0.f),
    C_Vertex3f(0.189301729f, 0.452834606f, -1.f),
    C_Normal3f(0.999852061f, -0.0172010921f, 0.f),
    C_Vertex3f(0.188998938f, 0.429518223f, 0.f),
    C_Vertex3f(0.188998938f, 0.429518223f, -1.f),
    C_Normal3f(0.999625981f, -0.027347317f, 0.f),
    C_Vertex3f(0.18854475f, 0.408833504f, 0.f),
    C_Vertex3f(0.18854475f, 0.408833504f, -1.f),
    C_Normal3f(0.999172628f, -0.0406708717f, 0.f),
    C_Vertex3f(0.187939167f, 0.390780449f, 0.f),
    C_Vertex3f(0.187939167f, 0.390780449f, -1.f),
    C_Normal3f(0.998262167f, -0.0589292385f, 0.f),
    C_Vertex3f(0.187182188f, 0.375359058f, 0.f),
    C_Vertex3f(0.187182188f, 0.375359058f, -1.f),
    C_Normal3f(0.996342242f, -0.0854525194f, 0.f),
    C_Vertex3f(0.186273813f, 0.362569332f, 0.f),
    C_Vertex3f(0.186273813f, 0.362569332f, -1.f),
    C_Normal3f(0.992920637f, -0.118779764f, 0.f),
    C_Vertex3f(0.185214043f, 0.35241127f, 0.f),
    C_Vertex3f(0.185214043f, 0.35241127f, -1.f),
    C_Normal3f(0.98621875f, -0.165446371f, 0.f),
    C_Vertex3f(0.184608459f, 0.348648071f, 0.f),
    C_Vertex3f(0.184608459f, 0.348648071f, -1.f),
    C_Normal3f(0.975097239f, -0.221777678f, 0.f),
    C_Vertex3f(0.182624817f, 0.336977005f, 0.f),
    C_Vertex3f(0.182624817f, 0.336977005f, -1.f),
    C_Normal3f(0.954020798f, -0.299740314f, 0.f),
    C_Vertex3f(0.177252769f, 0.316307068f, 0.f),
    C_Vertex3f(0.177252769f, 0.316307068f, -1.f),
    C_Normal3f(0.912301242f, -0.409519911f, 0.f),
    C_Vertex3f(0.17047596f, 0.298309326f, 0.f),
    C_Vertex3f(0.17047596f, 0.298309326f, -1.f),
    C_Normal3f(0.84416467f, -0.536083996f, 0.f),
    C_Vertex3f(0.162294388f, 0.28298378f, 0.f),
    C_Vertex3f(0.162294388f, 0.28298378f, -1.f),
    C_Normal3f(0.746893048f, -0.664944172f, 0.f),
    C_Vertex3f(0.152708054f, 0.270330429f, 0.f),
    C_Vertex3f(0.152708054f, 0.270330429f, -1.f),
    C_Normal3f(0.646538198f, -0.762881517f, 0.f),
    C_Vertex3f(0.140395164f, 0.258385658f, 0.f),
    C_Vertex3f(0.140395164f, 0.258385658f, -1.f),
    C_Normal3f(0.552826047f, -0.833296657f, 0.f),
    C_Vertex3f(0.125355721f, 0.247149467f, 0.f),
    C_Vertex3f(0.125355721f, 0.247149467f, -1.f),
    C_Normal3f(0.482647717f, -0.875814617f, 0.f),
    C_Vertex3f(0.107589722f, 0.236621857f, 0.f),
    C_Vertex3f(0.107589722f, 0.236621857f, -1.f),
    C_Normal3f(0.432108819f, -0.901821494f, 0.f),
    C_Vertex3f(0.0973434448f, 0.231712341f, 0.f),
    C_Vertex3f(0.0973434448f, 0.231712341f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.382570148f, 0.923926532f, 0.f),
    C_Vertex3f(0.0973434448f, 0.231712341f, 0.f),
    C_Vertex3f(0.0973434448f, 0.231712341f, -1.f),
    C_Normal3f(0.426906526f, 0.904295802f, 0.f),
    C_Vertex3f(0.103403091f, 0.229203224f, 0.f),
    C_Vertex3f(0.103403091f, 0.229203224f, -1.f),
    C_Normal3f(0.483897686f, 0.875124574f, 0.f),
    C_Vertex3f(0.114689231f, 0.223523617f, 0.f),
    C_Vertex3f(0.114689231f, 0.223523617f, -1.f),
    C_Normal3f(0.553592622f, 0.832787633f, 0.f),
    C_Vertex3f(0.125142217f, 0.217182636f, 0.f),
    C_Vertex3f(0.125142217f, 0.217182636f, -1.f),
    C_Normal3f(0.623161912f, 0.78209281f, 0.f),
    C_Vertex3f(0.134762049f, 0.210180283f, 0.f),
    C_Vertex3f(0.134762049f, 0.210180283f, -1.f),
    C_Normal3f(0.690691471f, 0.723149717f, 0.f),
    C_Vertex3f(0.143548727f, 0.202516556f, 0.f),
    C_Vertex3f(0.143548727f, 0.202516556f, -1.f),
    C_Normal3f(0.754166305f, 0.656683505f, 0.f),
    C_Vertex3f(0.151502252f, 0.194191456f, 0.f),
    C_Vertex3f(0.151502252f, 0.194191456f, -1.f),
    C_Normal3f(0.811722696f, 0.584042966f, 0.f),
    C_Vertex3f(0.158622622f, 0.185204983f, 0.f),
    C_Vertex3f(0.158622622f, 0.185204983f, -1.f),
    C_Normal3f(0.861899674f, 0.507078826f, 0.f),
    C_Vertex3f(0.16490984f, 0.175557137f, 0.f),
    C_Vertex3f(0.16490984f, 0.175557137f, -1.f),
    C_Normal3f(0.908708632f, 0.417431027f, 0.f),
    C_Vertex3f(0.170363903f, 0.165247917f, 0.f),
    C_Vertex3f(0.170363903f, 0.165247917f, -1.f),
    C_Normal3f(0.944767892f, 0.327740371f, 0.f),
    C_Vertex3f(0.175136209f, 0.153295279f, 0.f),
    C_Vertex3f(0.175136209f, 0.153295279f, -1.f),
    C_Normal3f(0.967805266f, 0.251699805f, 0.f),
    C_Vertex3f(0.179226756f, 0.139699221f, 0.f),
    C_Vertex3f(0.179226756f, 0.139699221f, -1.f),
    C_Normal3f(0.982241273f, 0.187621787f, 0.f),
    C_Vertex3f(0.182635546f, 0.124459743f, 0.f),
    C_Vertex3f(0.182635546f, 0.124459743f, -1.f),
    C_Normal3f(0.991039693f, 0.133568183f, 0.f),
    C_Vertex3f(0.185362577f, 0.107576847f, 0.f),
    C_Vertex3f(0.185362577f, 0.107576847f, -1.f),
    C_Normal3f(0.996142387f, 0.0877515897f, 0.f),
    C_Vertex3f(0.187407851f, 0.0890505314f, 0.f),
    C_Vertex3f(0.187407851f, 0.0890505314f, -1.f),
    C_Normal3f(0.998815417f, 0.0486591309f, 0.f),
    C_Vertex3f(0.188771367f, 0.0688807964f, 0.f),
    C_Vertex3f(0.188771367f, 0.0688807964f, -1.f),
    C_Normal3f(0.99988097f, 0.0154299783f, 0.f),
    C_Vertex3f(0.189453125f, 0.0470676422f, 0.f),
    C_Vertex3f(0.189453125f, 0.0470676422f, -1.f),
    C_Normal3f(0.999999344f, 0.00110343483f, 0.f),
    C_Vertex3f(0.189453125f, 0.0247020721f, 0.f),
    C_Vertex3f(0.189453125f, 0.0247020721f, -1.f),
    C_Normal3f(0.999992788f, 0.00380572095f, 0.f),
    C_Vertex3f(0.189498901f, 0.00558233261f, 0.f),
    C_Vertex3f(0.189498901f, 0.00558233261f, -1.f),
    C_Normal3f(0.999974072f, 0.00720301876f, 0.f),
    C_Vertex3f(0.189590454f, -0.0113825798f, 0.f),
    C_Vertex3f(0.189590454f, -0.0113825798f, -1.f),
    C_Normal3f(0.999931931f, 0.0116660753f, 0.f),
    C_Vertex3f(0.189727783f, -0.0261926651f, 0.f),
    C_Vertex3f(0.189727783f, -0.0261926651f, -1.f),
    C_Normal3f(0.999841809f, 0.0177892409f, 0.f),
    C_Vertex3f(0.189910889f, -0.0388479233f, 0.f),
    C_Vertex3f(0.189910889f, -0.0388479233f, -1.f),
    C_Normal3f(0.999643266f, 0.0267090872f, 0.f),
    C_Vertex3f(0.190139771f, -0.0493483543f, 0.f),
    C_Vertex3f(0.190139771f, -0.0493483543f, -1.f),
    C_Normal3f(0.999163032f, 0.0409039035f, 0.f),
    C_Vertex3f(0.190414429f, -0.0576939583f, 0.f),
    C_Vertex3f(0.190414429f, -0.0576939583f, -1.f),
    C_Normal3f(0.998123944f, 0.061226774f, 0.f),
    C_Vertex3f(0.190734863f, -0.0638847351f, 0.f),
    C_Vertex3f(0.190734863f, -0.0638847351f, -1.f),
    C_Normal3f(0.993787825f, 0.111291379f, 0.f),
    C_Vertex3f(0.190917969f, -0.06590271f, 0.f),
    C_Vertex3f(0.190917969f, -0.06590271f, -1.f),
    C_Normal3f(0.981004477f, 0.193985522f, 0.f),
    C_Vertex3f(0.191791534f, -0.0733203888f, 0.f),
    C_Vertex3f(0.191791534f, -0.0733203888f, -1.f),
    C_Normal3f(0.950498998f, 0.310727537f, 0.f),
    C_Vertex3f(0.194876194f, -0.085919857f, 0.f),
    C_Vertex3f(0.194876194f, -0.085919857f, -1.f),
    C_Normal3f(0.875981629f, 0.482344478f, 0.f),
    C_Vertex3f(0.199298382f, -0.0962834358f, 0.f),
    C_Vertex3f(0.199298382f, -0.0962834358f, -1.f),
    C_Normal3f(0.737003267f, 0.675889194f, 0.f),
    C_Vertex3f(0.205058098f, -0.104411125f, 0.f),
    C_Vertex3f(0.205058098f, -0.104411125f, -1.f),
    C_Normal3f(0.505345166f, 0.862917304f, 0.f),
    C_Vertex3f(0.212155342f, -0.110302925f, 0.f),
    C_Vertex3f(0.212155342f, -0.110302925f, -1.f),
    C_Normal3f(0.288635224f, 0.957439184f, 0.f),
    C_Vertex3f(0.222664356f, -0.114721775f, 0.f),
    C_Vertex3f(0.222664356f, -0.114721775f, -1.f),
    C_Normal3f(0.139995709f, 0.990152061f, 0.f),
    C_Vertex3f(0.23658514f, -0.117667675f, 0.f),
    C_Vertex3f(0.23658514f, -0.117667675f, -1.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.253917694f, -0.119140625f, 0.f),
    C_Vertex3f(0.253917694f, -0.119140625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.253917694f, -0.119140625f, 0.f),
    C_Vertex3f(0.253917694f, -0.119140625f, -1.f),
    C_Vertex3f(0.277832031f, -0.119140625f, -1.f),
    C_Vertex3f(0.277832031f, -0.119140625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.277832031f, -0.119140625f, 0.f),
    C_Vertex3f(0.277832031f, -0.119140625f, -1.f),
    C_Vertex3f(0.277832031f, -0.187988281f, -1.f),
    C_Vertex3f(0.277832031f, -0.187988281f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.277832031f, -0.187988281f, 0.f),
    C_Vertex3f(0.277832031f, -0.187988281f, -1.f),
    C_Normal3f(-0.0162848905f, -0.99986732f, 0.f),
    C_Vertex3f(0.242658615f, -0.187988281f, 0.f),
    C_Vertex3f(0.242658615f, -0.187988281f, -1.f),
    C_Normal3f(-0.0769083798f, -0.997038186f, 0.f),
    C_Vertex3f(0.224196434f, -0.187114716f, 0.f),
    C_Vertex3f(0.224196434f, -0.187114716f, -1.f),
    C_Normal3f(-0.153726667f, -0.988113403f, 0.f),
    C_Vertex3f(0.208683968f, -0.185367584f, 0.f),
    C_Vertex3f(0.208683968f, -0.185367584f, -1.f),
    C_Normal3f(-0.271724135f, -0.962375164f, 0.f),
    C_Vertex3f(0.196121216f, -0.182746887f, 0.f),
    C_Vertex3f(0.196121216f, -0.182746887f, -1.f),
    C_Normal3f(-0.403362662f, -0.915040255f, 0.f),
    C_Vertex3f(0.184324265f, -0.178489685f, 0.f),
    C_Vertex3f(0.184324265f, -0.178489685f, -1.f),
    C_Normal3f(-0.516935825f, -0.856024086f, 0.f),
    C_Vertex3f(0.171544552f, -0.171913147f, 0.f),
    C_Vertex3f(0.171544552f, -0.171913147f, -1.f),
    C_Normal3f(-0.63023138f, -0.776407421f, 0.f),
    C_Vertex3f(0.159965992f, -0.163780212f, 0.f),
    C_Vertex3f(0.159965992f, -0.163780212f, -1.f),
    C_Normal3f(-0.730806708f, -0.682584524f, 0.f),
    C_Vertex3f(0.149588585f, -0.154090881f, 0.f),
    C_Vertex3f(0.149588585f, -0.154090881f, -1.f),
    C_Normal3f(-0.82008481f, -0.572242022f, 0.f),
    C_Vertex3f(0.140412331f, -0.142845154f, 0.f),
    C_Vertex3f(0.140412331f, -0.142845154f, -1.f),
    C_Normal3f(-0.89413327f, -0.447801113f, 0.f),
    C_Vertex3f(0.132847309f, -0.13009882f, 0.f),
    C_Vertex3f(0.132847309f, -0.13009882f, -1.f),
    C_Normal3f(-0.945824802f, -0.324677587f, 0.f),
    C_Vertex3f(0.12689352f, -0.115851879f, 0.f),
    C_Vertex3f(0.12689352f, -0.115851879f, -1.f),
    C_Normal3f(-0.97606504f, -0.217478469f, 0.f),
    C_Vertex3f(0.122550964f, -0.100104332f, 0.f),
    C_Vertex3f(0.122550964f, -0.100104332f, -1.f),
    C_Normal3f(-0.990013421f, -0.140973315f, 0.f),
    C_Vertex3f(0.120502472f, -0.0871682167f, 0.f),
    C_Vertex3f(0.120502472f, -0.0871682167f, -1.f),
    C_Normal3f(-0.994324744f, -0.106387101f, 0.f),
    C_Vertex3f(0.11928463f, -0.0771658421f, 0.f),
    C_Vertex3f(0.11928463f, -0.0771658421f, -1.f),
    C_Normal3f(-0.996611774f, -0.0822495818f, 0.f),
    C_Vertex3f(0.118214607f, -0.0657851696f, 0.f),
    C_Vertex3f(0.118214607f, -0.0657851696f, -1.f),
    C_Normal3f(-0.998016417f, -0.0629538298f, 0.f),
    C_Vertex3f(0.117292404f, -0.0530261993f, 0.f),
    C_Vertex3f(0.117292404f, -0.0530261993f, -1.f),
    C_Normal3f(-0.99888587f, -0.0471923426f, 0.f),
    C_Vertex3f(0.116518021f, -0.0388889313f, 0.f),
    C_Vertex3f(0.116518021f, -0.0388889313f, -1.f),
    C_Normal3f(-0.999418974f, -0.0340847187f, 0.f),
    C_Vertex3f(0.115891457f, -0.0233733654f, 0.f),
    C_Vertex3f(0.115891457f, -0.0233733654f, -1.f),
    C_Normal3f(-0.999734998f, -0.0230180956f, 0.f),
    C_Vertex3f(0.115412712f, -0.00647950172f, 0.f),
    C_Vertex3f(0.115412712f, -0.00647950172f, -1.f),
    C_Normal3f(-0.999908149f, -0.0135534741f, 0.f),
    C_Vertex3f(0.115081787f, 0.0117926598f, 0.f),
    C_Vertex3f(0.115081787f, 0.0117926598f, -1.f),
    C_Normal3f(-0.999890089f, -0.014822931f, 0.f),
    C_Vertex3f(0.114898682f, 0.031443119f, 0.f),
    C_Vertex3f(0.114898682f, 0.031443119f, -1.f),
    C_Normal3f(-0.999628246f, -0.0272670761f, 0.f),
    C_Vertex3f(0.114521623f, 0.0495789051f, 0.f),
    C_Vertex3f(0.114521623f, 0.0495789051f, -1.f),
    C_Normal3f(-0.999114633f, -0.0420705527f, 0.f),
    C_Vertex3f(0.11395061f, 0.0662000179f, 0.f),
    C_Vertex3f(0.11395061f, 0.0662000179f, -1.f),
    C_Normal3f(-0.998200774f, -0.0599613748f, 0.f),
    C_Vertex3f(0.113185644f, 0.0813064575f, 0.f),
    C_Vertex3f(0.113185644f, 0.0813064575f, -1.f),
    C_Normal3f(-0.996632874f, -0.0819935873f, 0.f),
    C_Vertex3f(0.112226725f, 0.0948982239f, 0.f),
    C_Vertex3f(0.112226725f, 0.0948982239f, -1.f),
    C_Normal3f(-0.993959665f, -0.109746195f, 0.f),
    C_Vertex3f(0.111073852f, 0.106975317f, 0.f),
    C_Vertex3f(0.111073852f, 0.106975317f, -1.f),
    C_Normal3f(-0.989331782f, -0.145679593f, 0.f),
    C_Vertex3f(0.109727025f, 0.117537737f, 0.f),
    C_Vertex3f(0.109727025f, 0.117537737f, -1.f),
    C_Normal3f(-0.979894936f, -0.19951424f, 0.f),
    C_Vertex3f(0.108186245f, 0.126585484f, 0.f),
    C_Vertex3f(0.108186245f, 0.126585484f, -1.f),
    C_Normal3f(-0.95349586f, -0.301406205f, 0.f),
    C_Vertex3f(0.105584145f, 0.137885094f, 0.f),
    C_Vertex3f(0.105584145f, 0.137885094f, -1.f),
    C_Normal3f(-0.900936961f, -0.433949947f, 0.f),
    C_Vertex3f(0.100352287f, 0.151368141f, 0.f),
    C_Vertex3f(0.100352287f, 0.151368141f, -1.f),
    C_Normal3f(-0.815782189f, -0.578359306f, 0.f),
    C_Vertex3f(0.0933580399f, 0.163268089f, 0.f),
    C_Vertex3f(0.0933580399f, 0.163268089f, -1.f),
    C_Normal3f(-0.702942252f, -0.711246908f, 0.f),
    C_Vertex3f(0.0846014023f, 0.173584938f, 0.f),
    C_Vertex3f(0.0846014023f, 0.173584938f, -1.f),
    C_Normal3f(-0.563210666f, -0.826313317f, 0.f),
    C_Vertex3f(0.0740823746f, 0.182318687f, 0.f),
    C_Vertex3f(0.0740823746f, 0.182318687f, -1.f),
    C_Normal3f(-0.401142776f, -0.916015506f, 0.f),
    C_Vertex3f(0.0620179176f, 0.188977718f, 0.f),
    C_Vertex3f(0.0620179176f, 0.188977718f, -1.f),
    C_Normal3f(-0.239440426f, -0.970911026f, 0.f),
    C_Vertex3f(0.0484080315f, 0.193562031f, 0.f),
    C_Vertex3f(0.0484080315f, 0.193562031f, -1.f),
    C_Normal3f(-0.115242854f, -0.993337393f, 0.f),
    C_Vertex3f(0.0332527161f, 0.196071625f, 0.f),
    C_Vertex3f(0.0332527161f, 0.196071625f, -1.f),
    C_Normal3f(-0.0260304641f, -0.999661148f, 0.f),
    C_Vertex3f(0.0249023438f, 0.196289063f, 0.f),
    C_Vertex3f(0.0249023438f, 0.196289063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0249023438f, 0.196289063f, 0.f),
    C_Vertex3f(0.0249023438f, 0.196289063f, -1.f),
    C_Vertex3f(0.0249023438f, 0.267578125f, -1.f),
    C_Vertex3f(0.0249023438f, 0.267578125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.222664356f, -0.114721775f, 0.f),
    C_Vertex3f(0.224196434f, -0.187114716f, 0.f),
    C_Vertex3f(0.23658514f, -0.117667675f, 0.f),
    C_Vertex3f(0.242658615f, -0.187988281f, 0.f),
    C_Vertex3f(0.253917694f, -0.119140625f, 0.f),
    C_Vertex3f(0.277832031f, -0.187988281f, 0.f),
    C_Vertex3f(0.277832031f, -0.119140625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.212155342f, -0.110302925f, 0.f),
    C_Vertex3f(0.205058098f, -0.104411125f, 0.f),
    C_Vertex3f(0.208683968f, -0.185367584f, 0.f),
    C_Vertex3f(0.224196434f, -0.187114716f, 0.f),
    C_Vertex3f(0.222664356f, -0.114721775f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.199298382f, -0.0962834358f, 0.f),
    C_Vertex3f(0.194876194f, -0.085919857f, 0.f),
    C_Vertex3f(0.196121216f, -0.182746887f, 0.f),
    C_Vertex3f(0.208683968f, -0.185367584f, 0.f),
    C_Vertex3f(0.205058098f, -0.104411125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.196121216f, -0.182746887f, 0.f),
    C_Vertex3f(0.194876194f, -0.085919857f, 0.f),
    C_Vertex3f(0.191791534f, -0.0733203888f, 0.f),
    C_Vertex3f(0.190917969f, -0.06590271f, 0.f),
    C_Vertex3f(0.190734863f, -0.0638847351f, 0.f),
    C_Vertex3f(0.190414429f, -0.0576939583f, 0.f),
    C_Vertex3f(0.184324265f, -0.178489685f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.188771367f, 0.0688807964f, 0.f),
    C_Vertex3f(0.187407851f, 0.0890505314f, 0.f),
    C_Vertex3f(0.189727783f, -0.0261926651f, 0.f),
    C_Vertex3f(0.189590454f, -0.0113825798f, 0.f),
    C_Vertex3f(0.189498901f, 0.00558233261f, 0.f),
    C_Vertex3f(0.189453125f, 0.0247020721f, 0.f),
    C_Vertex3f(0.189453125f, 0.0470676422f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.189727783f, -0.0261926651f, 0.f),
    C_Vertex3f(0.187407851f, 0.0890505314f, 0.f),
    C_Vertex3f(0.189910889f, -0.0388479233f, 0.f),
    C_Vertex3f(0.185362577f, 0.107576847f, 0.f),
    C_Vertex3f(0.190139771f, -0.0493483543f, 0.f),
    C_Vertex3f(0.184324265f, -0.178489685f, 0.f),
    C_Vertex3f(0.190414429f, -0.0576939583f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.184324265f, -0.178489685f, 0.f),
    C_Vertex3f(0.185362577f, 0.107576847f, 0.f),
    C_Vertex3f(0.182635546f, 0.124459743f, 0.f),
    C_Vertex3f(0.179226756f, 0.139699221f, 0.f),
    C_Vertex3f(0.175136209f, 0.153295279f, 0.f),
    C_Vertex3f(0.171544552f, -0.171913147f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.175136209f, 0.153295279f, 0.f),
    C_Vertex3f(0.170363903f, 0.165247917f, 0.f),
    C_Vertex3f(0.171544552f, -0.171913147f, 0.f),
    C_Vertex3f(0.16490984f, 0.175557137f, 0.f),
    C_Vertex3f(0.159965992f, -0.163780212f, 0.f),
    C_Vertex3f(0.158622622f, 0.185204983f, 0.f),
    C_Vertex3f(0.151502252f, 0.194191456f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12689352f, -0.115851879f, 0.f),
    C_Vertex3f(0.132847309f, -0.13009882f, 0.f),
    C_Vertex3f(0.134762049f, 0.210180283f, 0.f),
    C_Vertex3f(0.140412331f, -0.142845154f, 0.f),
    C_Vertex3f(0.143548727f, 0.202516556f, 0.f),
    C_Vertex3f(0.149588585f, -0.154090881f, 0.f),
    C_Vertex3f(0.151502252f, 0.194191456f, 0.f),
    C_Vertex3f(0.159965992f, -0.163780212f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.125142217f, 0.217182636f, 0.f),
    C_Vertex3f(0.114689231f, 0.223523617f, 0.f),
    C_Vertex3f(0.114898682f, 0.031443119f, 0.f),
    C_Vertex3f(0.115081787f, 0.0117926598f, 0.f),
    C_Vertex3f(0.115412712f, -0.00647950172f, 0.f),
    C_Vertex3f(0.115891457f, -0.0233733654f, 0.f),
    C_Vertex3f(0.116518021f, -0.0388889313f, 0.f),
    C_Vertex3f(0.117292404f, -0.0530261993f, 0.f),
    C_Vertex3f(0.118214607f, -0.0657851696f, 0.f),
    C_Vertex3f(0.11928463f, -0.0771658421f, 0.f),
    C_Vertex3f(0.120502472f, -0.0871682167f, 0.f),
    C_Vertex3f(0.122550964f, -0.100104332f, 0.f),
    C_Vertex3f(0.12689352f, -0.115851879f, 0.f),
    C_Vertex3f(0.134762049f, 0.210180283f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.114689231f, 0.223523617f, 0.f),
    C_Vertex3f(0.103403091f, 0.229203224f, 0.f),
    C_Vertex3f(0.105584145f, 0.137885094f, 0.f),
    C_Vertex3f(0.108186245f, 0.126585484f, 0.f),
    C_Vertex3f(0.109727025f, 0.117537737f, 0.f),
    C_Vertex3f(0.111073852f, 0.106975317f, 0.f),
    C_Vertex3f(0.112226725f, 0.0948982239f, 0.f),
    C_Vertex3f(0.113185644f, 0.0813064575f, 0.f),
    C_Vertex3f(0.11395061f, 0.0662000179f, 0.f),
    C_Vertex3f(0.114521623f, 0.0495789051f, 0.f),
    C_Vertex3f(0.114898682f, 0.031443119f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.100352287f, 0.151368141f, 0.f),
    C_Vertex3f(0.105584145f, 0.137885094f, 0.f),
    C_Vertex3f(0.103403091f, 0.229203224f, 0.f),
    C_Vertex3f(0.0973434448f, 0.231712341f, 0.f),
    C_Vertex3f(0.0941057205f, 0.300244808f, 0.f),
    C_Vertex3f(0.0933580399f, 0.163268089f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0249023438f, 0.267578125f, 0.f),
    C_Vertex3f(0.0332527161f, 0.196071625f, 0.f),
    C_Vertex3f(0.0333442688f, 0.267795563f, 0.f),
    C_Vertex3f(0.0484080315f, 0.193562031f, 0.f),
    C_Vertex3f(0.0486655235f, 0.270277977f, 0.f),
    C_Vertex3f(0.0620179176f, 0.188977718f, 0.f),
    C_Vertex3f(0.0624241829f, 0.27480793f, 0.f),
    C_Vertex3f(0.0740823746f, 0.182318687f, 0.f),
    C_Vertex3f(0.0746202469f, 0.281385422f, 0.f),
    C_Vertex3f(0.0846014023f, 0.173584938f, 0.f),
    C_Vertex3f(0.0852537155f, 0.290010452f, 0.f),
    C_Vertex3f(0.0933580399f, 0.163268089f, 0.f),
    C_Vertex3f(0.0941057205f, 0.300244808f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230430126f, 0.579451799f, 0.f),
    C_Vertex3f(0.238892794f, 0.580985665f, 0.f),
    C_Vertex3f(0.244022369f, 0.651367188f, 0.f),
    C_Vertex3f(0.248307943f, 0.582008243f, 0.f),
    C_Vertex3f(0.277832031f, 0.651367188f, 0.f),
    C_Vertex3f(0.258675575f, 0.582519531f, 0.f),
    C_Vertex3f(0.277832031f, 0.582519531f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.227880478f, 0.650821686f, 0.f),
    C_Vertex3f(0.214357376f, 0.649730682f, 0.f),
    C_Vertex3f(0.216362238f, 0.574850202f, 0.f),
    C_Vertex3f(0.222919941f, 0.577406645f, 0.f),
    C_Vertex3f(0.230430126f, 0.579451799f, 0.f),
    C_Vertex3f(0.244022369f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.196391106f, 0.552148461f, 0.f),
    C_Vertex3f(0.199166298f, 0.558761477f, 0.f),
    C_Vertex3f(0.199310303f, 0.647003174f, 0.f),
    C_Vertex3f(0.202404022f, 0.56411314f, 0.f),
    C_Vertex3f(0.203453064f, 0.648094177f, 0.f),
    C_Vertex3f(0.206104279f, 0.568203449f, 0.f),
    C_Vertex3f(0.214357376f, 0.649730682f, 0.f),
    C_Vertex3f(0.210757017f, 0.57178247f, 0.f),
    C_Vertex3f(0.216362238f, 0.574850202f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.190840721f, 0.524741292f, 0.f),
    C_Vertex3f(0.192228317f, 0.535138369f, 0.f),
    C_Vertex3f(0.193201065f, 0.645256042f, 0.f),
    C_Vertex3f(0.194078445f, 0.544274092f, 0.f),
    C_Vertex3f(0.199310303f, 0.647003174f, 0.f),
    C_Vertex3f(0.196391106f, 0.552148461f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.193201065f, 0.645256042f, 0.f),
    C_Vertex3f(0.18207407f, 0.640970707f, 0.f),
    C_Vertex3f(0.182624817f, 0.336977005f, 0.f),
    C_Vertex3f(0.184608459f, 0.348648071f, 0.f),
    C_Vertex3f(0.185214043f, 0.35241127f, 0.f),
    C_Vertex3f(0.189453125f, 0.500163078f, 0.f),
    C_Vertex3f(0.189915657f, 0.513082862f, 0.f),
    C_Vertex3f(0.190840721f, 0.524741292f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.189453125f, 0.500163078f, 0.f),
    C_Vertex3f(0.185214043f, 0.35241127f, 0.f),
    C_Vertex3f(0.186273813f, 0.362569332f, 0.f),
    C_Vertex3f(0.187182188f, 0.375359058f, 0.f),
    C_Vertex3f(0.187939167f, 0.390780449f, 0.f),
    C_Vertex3f(0.18854475f, 0.408833504f, 0.f),
    C_Vertex3f(0.188998938f, 0.429518223f, 0.f),
    C_Vertex3f(0.189301729f, 0.452834606f, 0.f),
    C_Vertex3f(0.189453125f, 0.478782654f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.182624817f, 0.336977005f, 0.f),
    C_Vertex3f(0.18207407f, 0.640970707f, 0.f),
    C_Vertex3f(0.177252769f, 0.316307068f, 0.f),
    C_Vertex3f(0.172038555f, 0.635894299f, 0.f),
    C_Vertex3f(0.17047596f, 0.298309326f, 0.f),
    C_Vertex3f(0.163094521f, 0.630026817f, 0.f),
    C_Vertex3f(0.162294388f, 0.28298378f, 0.f),
    C_Vertex3f(0.155241966f, 0.623368263f, 0.f),
    C_Vertex3f(0.152708054f, 0.270330429f, 0.f),
    C_Vertex3f(0.14801693f, 0.615591049f, 0.f),
    C_Vertex3f(0.141419411f, 0.606695175f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.140395164f, 0.258385658f, 0.f),
    C_Vertex3f(0.152708054f, 0.270330429f, 0.f),
    C_Vertex3f(0.141419411f, 0.606695175f, 0.f),
    C_Vertex3f(0.135449409f, 0.596680641f, 0.f),
    C_Vertex3f(0.130106926f, 0.585547447f, 0.f),
    C_Vertex3f(0.125608921f, 0.572613239f, 0.f),
    C_Vertex3f(0.125355721f, 0.247149467f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.125355721f, 0.247149467f, 0.f),
    C_Vertex3f(0.125608921f, 0.572613239f, 0.f),
    C_Vertex3f(0.121955395f, 0.557878017f, 0.f),
    C_Vertex3f(0.119146347f, 0.541341782f, 0.f),
    C_Vertex3f(0.118164063f, 0.532173157f, 0.f),
    C_Vertex3f(0.117902756f, 0.529279709f, 0.f),
    C_Vertex3f(0.117432356f, 0.521937609f, 0.f),
    C_Vertex3f(0.11701417f, 0.513040304f, 0.f),
    C_Vertex3f(0.116648197f, 0.502587795f, 0.f),
    C_Vertex3f(0.116334438f, 0.490580082f, 0.f),
    C_Vertex3f(0.116072893f, 0.477017164f, 0.f),
    C_Vertex3f(0.115863562f, 0.461899042f, 0.f),
    C_Vertex3f(0.115706444f, 0.445225716f, 0.f),
    C_Vertex3f(0.11560154f, 0.426997185f, 0.f),
    C_Vertex3f(0.115290523f, 0.410105824f, 0.f),
    C_Vertex3f(0.114773393f, 0.394551635f, 0.f),
    C_Vertex3f(0.11405015f, 0.380334616f, 0.f),
    C_Vertex3f(0.113120794f, 0.367454767f, 0.f),
    C_Vertex3f(0.111985326f, 0.355912089f, 0.f),
    C_Vertex3f(0.110643744f, 0.345706582f, 0.f),
    C_Vertex3f(0.10909605f, 0.336838245f, 0.f),
    C_Vertex3f(0.107589722f, 0.236621857f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0941057205f, 0.300244808f, 0.f),
    C_Vertex3f(0.0973434448f, 0.231712341f, 0.f),
    C_Vertex3f(0.101176262f, 0.31208849f, 0.f),
    C_Vertex3f(0.107589722f, 0.236621857f, 0.f),
    C_Vertex3f(0.10646534f, 0.325541496f, 0.f),
    C_Vertex3f(0.10909605f, 0.336838245f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0332527161f, 0.196071625f, 0.f),
    C_Vertex3f(0.0249023438f, 0.267578125f, 0.f),
    C_Vertex3f(0.0249023438f, 0.196289063f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.222664356f, -0.114721775f, -1.f),
    C_Vertex3f(0.224196434f, -0.187114716f, -1.f),
    C_Vertex3f(0.23658514f, -0.117667675f, -1.f),
    C_Vertex3f(0.242658615f, -0.187988281f, -1.f),
    C_Vertex3f(0.253917694f, -0.119140625f, -1.f),
    C_Vertex3f(0.277832031f, -0.187988281f, -1.f),
    C_Vertex3f(0.277832031f, -0.119140625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.212155342f, -0.110302925f, -1.f),
    C_Vertex3f(0.205058098f, -0.104411125f, -1.f),
    C_Vertex3f(0.208683968f, -0.185367584f, -1.f),
    C_Vertex3f(0.224196434f, -0.187114716f, -1.f),
    C_Vertex3f(0.222664356f, -0.114721775f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.199298382f, -0.0962834358f, -1.f),
    C_Vertex3f(0.194876194f, -0.085919857f, -1.f),
    C_Vertex3f(0.196121216f, -0.182746887f, -1.f),
    C_Vertex3f(0.208683968f, -0.185367584f, -1.f),
    C_Vertex3f(0.205058098f, -0.104411125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.196121216f, -0.182746887f, -1.f),
    C_Vertex3f(0.194876194f, -0.085919857f, -1.f),
    C_Vertex3f(0.191791534f, -0.0733203888f, -1.f),
    C_Vertex3f(0.190917969f, -0.06590271f, -1.f),
    C_Vertex3f(0.190734863f, -0.0638847351f, -1.f),
    C_Vertex3f(0.190414429f, -0.0576939583f, -1.f),
    C_Vertex3f(0.184324265f, -0.178489685f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.188771367f, 0.0688807964f, -1.f),
    C_Vertex3f(0.187407851f, 0.0890505314f, -1.f),
    C_Vertex3f(0.189727783f, -0.0261926651f, -1.f),
    C_Vertex3f(0.189590454f, -0.0113825798f, -1.f),
    C_Vertex3f(0.189498901f, 0.00558233261f, -1.f),
    C_Vertex3f(0.189453125f, 0.0247020721f, -1.f),
    C_Vertex3f(0.189453125f, 0.0470676422f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.189727783f, -0.0261926651f, -1.f),
    C_Vertex3f(0.187407851f, 0.0890505314f, -1.f),
    C_Vertex3f(0.189910889f, -0.0388479233f, -1.f),
    C_Vertex3f(0.185362577f, 0.107576847f, -1.f),
    C_Vertex3f(0.190139771f, -0.0493483543f, -1.f),
    C_Vertex3f(0.184324265f, -0.178489685f, -1.f),
    C_Vertex3f(0.190414429f, -0.0576939583f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.184324265f, -0.178489685f, -1.f),
    C_Vertex3f(0.185362577f, 0.107576847f, -1.f),
    C_Vertex3f(0.182635546f, 0.124459743f, -1.f),
    C_Vertex3f(0.179226756f, 0.139699221f, -1.f),
    C_Vertex3f(0.175136209f, 0.153295279f, -1.f),
    C_Vertex3f(0.171544552f, -0.171913147f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.175136209f, 0.153295279f, -1.f),
    C_Vertex3f(0.170363903f, 0.165247917f, -1.f),
    C_Vertex3f(0.171544552f, -0.171913147f, -1.f),
    C_Vertex3f(0.16490984f, 0.175557137f, -1.f),
    C_Vertex3f(0.159965992f, -0.163780212f, -1.f),
    C_Vertex3f(0.158622622f, 0.185204983f, -1.f),
    C_Vertex3f(0.151502252f, 0.194191456f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12689352f, -0.115851879f, -1.f),
    C_Vertex3f(0.132847309f, -0.13009882f, -1.f),
    C_Vertex3f(0.134762049f, 0.210180283f, -1.f),
    C_Vertex3f(0.140412331f, -0.142845154f, -1.f),
    C_Vertex3f(0.143548727f, 0.202516556f, -1.f),
    C_Vertex3f(0.149588585f, -0.154090881f, -1.f),
    C_Vertex3f(0.151502252f, 0.194191456f, -1.f),
    C_Vertex3f(0.159965992f, -0.163780212f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.125142217f, 0.217182636f, -1.f),
    C_Vertex3f(0.114689231f, 0.223523617f, -1.f),
    C_Vertex3f(0.114898682f, 0.031443119f, -1.f),
    C_Vertex3f(0.115081787f, 0.0117926598f, -1.f),
    C_Vertex3f(0.115412712f, -0.00647950172f, -1.f),
    C_Vertex3f(0.115891457f, -0.0233733654f, -1.f),
    C_Vertex3f(0.116518021f, -0.0388889313f, -1.f),
    C_Vertex3f(0.117292404f, -0.0530261993f, -1.f),
    C_Vertex3f(0.118214607f, -0.0657851696f, -1.f),
    C_Vertex3f(0.11928463f, -0.0771658421f, -1.f),
    C_Vertex3f(0.120502472f, -0.0871682167f, -1.f),
    C_Vertex3f(0.122550964f, -0.100104332f, -1.f),
    C_Vertex3f(0.12689352f, -0.115851879f, -1.f),
    C_Vertex3f(0.134762049f, 0.210180283f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.114689231f, 0.223523617f, -1.f),
    C_Vertex3f(0.103403091f, 0.229203224f, -1.f),
    C_Vertex3f(0.105584145f, 0.137885094f, -1.f),
    C_Vertex3f(0.108186245f, 0.126585484f, -1.f),
    C_Vertex3f(0.109727025f, 0.117537737f, -1.f),
    C_Vertex3f(0.111073852f, 0.106975317f, -1.f),
    C_Vertex3f(0.112226725f, 0.0948982239f, -1.f),
    C_Vertex3f(0.113185644f, 0.0813064575f, -1.f),
    C_Vertex3f(0.11395061f, 0.0662000179f, -1.f),
    C_Vertex3f(0.114521623f, 0.0495789051f, -1.f),
    C_Vertex3f(0.114898682f, 0.031443119f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.100352287f, 0.151368141f, -1.f),
    C_Vertex3f(0.105584145f, 0.137885094f, -1.f),
    C_Vertex3f(0.103403091f, 0.229203224f, -1.f),
    C_Vertex3f(0.0973434448f, 0.231712341f, -1.f),
    C_Vertex3f(0.0941057205f, 0.300244808f, -1.f),
    C_Vertex3f(0.0933580399f, 0.163268089f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0249023438f, 0.267578125f, -1.f),
    C_Vertex3f(0.0332527161f, 0.196071625f, -1.f),
    C_Vertex3f(0.0333442688f, 0.267795563f, -1.f),
    C_Vertex3f(0.0484080315f, 0.193562031f, -1.f),
    C_Vertex3f(0.0486655235f, 0.270277977f, -1.f),
    C_Vertex3f(0.0620179176f, 0.188977718f, -1.f),
    C_Vertex3f(0.0624241829f, 0.27480793f, -1.f),
    C_Vertex3f(0.0740823746f, 0.182318687f, -1.f),
    C_Vertex3f(0.0746202469f, 0.281385422f, -1.f),
    C_Vertex3f(0.0846014023f, 0.173584938f, -1.f),
    C_Vertex3f(0.0852537155f, 0.290010452f, -1.f),
    C_Vertex3f(0.0933580399f, 0.163268089f, -1.f),
    C_Vertex3f(0.0941057205f, 0.300244808f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230430126f, 0.579451799f, -1.f),
    C_Vertex3f(0.238892794f, 0.580985665f, -1.f),
    C_Vertex3f(0.244022369f, 0.651367188f, -1.f),
    C_Vertex3f(0.248307943f, 0.582008243f, -1.f),
    C_Vertex3f(0.277832031f, 0.651367188f, -1.f),
    C_Vertex3f(0.258675575f, 0.582519531f, -1.f),
    C_Vertex3f(0.277832031f, 0.582519531f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.227880478f, 0.650821686f, -1.f),
    C_Vertex3f(0.214357376f, 0.649730682f, -1.f),
    C_Vertex3f(0.216362238f, 0.574850202f, -1.f),
    C_Vertex3f(0.222919941f, 0.577406645f, -1.f),
    C_Vertex3f(0.230430126f, 0.579451799f, -1.f),
    C_Vertex3f(0.244022369f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.196391106f, 0.552148461f, -1.f),
    C_Vertex3f(0.199166298f, 0.558761477f, -1.f),
    C_Vertex3f(0.199310303f, 0.647003174f, -1.f),
    C_Vertex3f(0.202404022f, 0.56411314f, -1.f),
    C_Vertex3f(0.203453064f, 0.648094177f, -1.f),
    C_Vertex3f(0.206104279f, 0.568203449f, -1.f),
    C_Vertex3f(0.214357376f, 0.649730682f, -1.f),
    C_Vertex3f(0.210757017f, 0.57178247f, -1.f),
    C_Vertex3f(0.216362238f, 0.574850202f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.190840721f, 0.524741292f, -1.f),
    C_Vertex3f(0.192228317f, 0.535138369f, -1.f),
    C_Vertex3f(0.193201065f, 0.645256042f, -1.f),
    C_Vertex3f(0.194078445f, 0.544274092f, -1.f),
    C_Vertex3f(0.199310303f, 0.647003174f, -1.f),
    C_Vertex3f(0.196391106f, 0.552148461f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.193201065f, 0.645256042f, -1.f),
    C_Vertex3f(0.18207407f, 0.640970707f, -1.f),
    C_Vertex3f(0.182624817f, 0.336977005f, -1.f),
    C_Vertex3f(0.184608459f, 0.348648071f, -1.f),
    C_Vertex3f(0.185214043f, 0.35241127f, -1.f),
    C_Vertex3f(0.189453125f, 0.500163078f, -1.f),
    C_Vertex3f(0.189915657f, 0.513082862f, -1.f),
    C_Vertex3f(0.190840721f, 0.524741292f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.189453125f, 0.500163078f, -1.f),
    C_Vertex3f(0.185214043f, 0.35241127f, -1.f),
    C_Vertex3f(0.186273813f, 0.362569332f, -1.f),
    C_Vertex3f(0.187182188f, 0.375359058f, -1.f),
    C_Vertex3f(0.187939167f, 0.390780449f, -1.f),
    C_Vertex3f(0.18854475f, 0.408833504f, -1.f),
    C_Vertex3f(0.188998938f, 0.429518223f, -1.f),
    C_Vertex3f(0.189301729f, 0.452834606f, -1.f),
    C_Vertex3f(0.189453125f, 0.478782654f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.182624817f, 0.336977005f, -1.f),
    C_Vertex3f(0.18207407f, 0.640970707f, -1.f),
    C_Vertex3f(0.177252769f, 0.316307068f, -1.f),
    C_Vertex3f(0.172038555f, 0.635894299f, -1.f),
    C_Vertex3f(0.17047596f, 0.298309326f, -1.f),
    C_Vertex3f(0.163094521f, 0.630026817f, -1.f),
    C_Vertex3f(0.162294388f, 0.28298378f, -1.f),
    C_Vertex3f(0.155241966f, 0.623368263f, -1.f),
    C_Vertex3f(0.152708054f, 0.270330429f, -1.f),
    C_Vertex3f(0.14801693f, 0.615591049f, -1.f),
    C_Vertex3f(0.141419411f, 0.606695175f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.140395164f, 0.258385658f, -1.f),
    C_Vertex3f(0.152708054f, 0.270330429f, -1.f),
    C_Vertex3f(0.141419411f, 0.606695175f, -1.f),
    C_Vertex3f(0.135449409f, 0.596680641f, -1.f),
    C_Vertex3f(0.130106926f, 0.585547447f, -1.f),
    C_Vertex3f(0.125608921f, 0.572613239f, -1.f),
    C_Vertex3f(0.125355721f, 0.247149467f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.125355721f, 0.247149467f, -1.f),
    C_Vertex3f(0.125608921f, 0.572613239f, -1.f),
    C_Vertex3f(0.121955395f, 0.557878017f, -1.f),
    C_Vertex3f(0.119146347f, 0.541341782f, -1.f),
    C_Vertex3f(0.118164063f, 0.532173157f, -1.f),
    C_Vertex3f(0.117902756f, 0.529279709f, -1.f),
    C_Vertex3f(0.117432356f, 0.521937609f, -1.f),
    C_Vertex3f(0.11701417f, 0.513040304f, -1.f),
    C_Vertex3f(0.116648197f, 0.502587795f, -1.f),
    C_Vertex3f(0.116334438f, 0.490580082f, -1.f),
    C_Vertex3f(0.116072893f, 0.477017164f, -1.f),
    C_Vertex3f(0.115863562f, 0.461899042f, -1.f),
    C_Vertex3f(0.115706444f, 0.445225716f, -1.f),
    C_Vertex3f(0.11560154f, 0.426997185f, -1.f),
    C_Vertex3f(0.115290523f, 0.410105824f, -1.f),
    C_Vertex3f(0.114773393f, 0.394551635f, -1.f),
    C_Vertex3f(0.11405015f, 0.380334616f, -1.f),
    C_Vertex3f(0.113120794f, 0.367454767f, -1.f),
    C_Vertex3f(0.111985326f, 0.355912089f, -1.f),
    C_Vertex3f(0.110643744f, 0.345706582f, -1.f),
    C_Vertex3f(0.10909605f, 0.336838245f, -1.f),
    C_Vertex3f(0.107589722f, 0.236621857f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0941057205f, 0.300244808f, -1.f),
    C_Vertex3f(0.0973434448f, 0.231712341f, -1.f),
    C_Vertex3f(0.101176262f, 0.31208849f, -1.f),
    C_Vertex3f(0.107589722f, 0.236621857f, -1.f),
    C_Vertex3f(0.10646534f, 0.325541496f, -1.f),
    C_Vertex3f(0.10909605f, 0.336838245f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0332527161f, 0.196071625f, -1.f),
    C_Vertex3f(0.0249023438f, 0.267578125f, -1.f),
    C_Vertex3f(0.0249023438f, 0.196289063f, -1.f),
    C_End(),
    C_Translatef(0.298828125f, 0.f, 0.f),
};
static const Call char_124[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.08203125f, -0.188476563f, 0.f),
    C_Vertex3f(0.08203125f, -0.188476563f, -1.f),
    C_Vertex3f(0.08203125f, 0.651367188f, -1.f),
    C_Vertex3f(0.08203125f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.08203125f, 0.651367188f, 0.f),
    C_Vertex3f(0.08203125f, 0.651367188f, -1.f),
    C_Vertex3f(0.150878906f, 0.651367188f, -1.f),
    C_Vertex3f(0.150878906f, 0.651367188f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.150878906f, 0.651367188f, 0.f),
    C_Vertex3f(0.150878906f, 0.651367188f, -1.f),
    C_Vertex3f(0.150878906f, -0.188476563f, -1.f),
    C_Vertex3f(0.150878906f, -0.188476563f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.150878906f, -0.188476563f, 0.f),
    C_Vertex3f(0.150878906f, -0.188476563f, -1.f),
    C_Vertex3f(0.08203125f, -0.188476563f, -1.f),
    C_Vertex3f(0.08203125f, -0.188476563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.08203125f, -0.188476563f, 0.f),
    C_Vertex3f(0.150878906f, -0.188476563f, 0.f),
    C_Vertex3f(0.150878906f, 0.651367188f, 0.f),
    C_Vertex3f(0.08203125f, 0.651367188f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.08203125f, -0.188476563f, -1.f),
    C_Vertex3f(0.150878906f, -0.188476563f, -1.f),
    C_Vertex3f(0.150878906f, 0.651367188f, -1.f),
    C_Vertex3f(0.08203125f, 0.651367188f, -1.f),
    C_End(),
    C_Translatef(0.232421875f, 0.f, 0.f),
};
static const Call char_125[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.2734375f, 0.267578125f, 0.f),
    C_Vertex3f(0.2734375f, 0.267578125f, -1.f),
    C_Vertex3f(0.2734375f, 0.196289063f, -1.f),
    C_Vertex3f(0.2734375f, 0.196289063f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.0257483516f, -0.999668479f, 0.f),
    C_Vertex3f(0.2734375f, 0.196289063f, 0.f),
    C_Vertex3f(0.2734375f, 0.196289063f, -1.f),
    C_Normal3f(0.114010431f, -0.99347955f, 0.f),
    C_Vertex3f(0.264995575f, 0.196071625f, 0.f),
    C_Vertex3f(0.264995575f, 0.196071625f, -1.f),
    C_Normal3f(0.236995429f, -0.971510768f, 0.f),
    C_Vertex3f(0.24967432f, 0.193562031f, 0.f),
    C_Vertex3f(0.24967432f, 0.193562031f, -1.f),
    C_Normal3f(0.397497207f, -0.917603374f, 0.f),
    C_Vertex3f(0.235915661f, 0.188977718f, 0.f),
    C_Vertex3f(0.235915661f, 0.188977718f, -1.f),
    C_Normal3f(0.559044778f, -0.829137444f, 0.f),
    C_Vertex3f(0.223719597f, 0.182318687f, 0.f),
    C_Vertex3f(0.223719597f, 0.182318687f, -1.f),
    C_Normal3f(0.698572159f, -0.715539634f, 0.f),
    C_Vertex3f(0.213086128f, 0.173584938f, 0.f),
    C_Vertex3f(0.213086128f, 0.173584938f, -1.f),
    C_Normal3f(0.811792314f, -0.583946228f, 0.f),
    C_Vertex3f(0.204234123f, 0.163295269f, 0.f),
    C_Vertex3f(0.204234123f, 0.163295269f, -1.f),
    C_Normal3f(0.898151219f, -0.439686477f, 0.f),
    C_Vertex3f(0.197163582f, 0.15144968f, 0.f),
    C_Vertex3f(0.197163582f, 0.15144968f, -1.f),
    C_Normal3f(0.951956987f, -0.306231797f, 0.f),
    C_Vertex3f(0.191874504f, 0.138048172f, 0.f),
    C_Vertex3f(0.191874504f, 0.138048172f, -1.f),
    C_Normal3f(0.978948951f, -0.204105169f, 0.f),
    C_Vertex3f(0.189243793f, 0.126830101f, 0.f),
    C_Vertex3f(0.189243793f, 0.126830101f, -1.f),
    C_Normal3f(0.988622367f, -0.150418654f, 0.f),
    C_Vertex3f(0.187696099f, 0.118007302f, 0.f),
    C_Vertex3f(0.187696099f, 0.118007302f, -1.f),
    C_Normal3f(0.993534744f, -0.113527633f, 0.f),
    C_Vertex3f(0.186354518f, 0.107840419f, 0.f),
    C_Vertex3f(0.186354518f, 0.107840419f, -1.f),
    C_Normal3f(0.996428668f, -0.0844393075f, 0.f),
    C_Vertex3f(0.185219049f, 0.0963294506f, 0.f),
    C_Vertex3f(0.185219049f, 0.0963294506f, -1.f),
    C_Normal3f(0.998139501f, -0.0609710999f, 0.f),
    C_Vertex3f(0.184289694f, 0.0834743977f, 0.f),
    C_Vertex3f(0.184289694f, 0.0834743977f, -1.f),
    C_Normal3f(0.99913156f, -0.0416676886f, 0.f),
    C_Vertex3f(0.183566451f, 0.06927526f, 0.f),
    C_Vertex3f(0.183566451f, 0.06927526f, -1.f),
    C_Normal3f(0.999674141f, -0.0255277064f, 0.f),
    C_Vertex3f(0.183049321f, 0.0537320375f, 0.f),
    C_Vertex3f(0.183049321f, 0.0537320375f, -1.f),
    C_Normal3f(0.999929905f, -0.0118424669f, 0.f),
    C_Vertex3f(0.182738304f, 0.0368447304f, 0.f),
    C_Vertex3f(0.182738304f, 0.0368447304f, -1.f),
    C_Normal3f(0.999971807f, -0.00750608044f, 0.f),
    C_Vertex3f(0.1826334f, 0.0186133385f, 0.f),
    C_Vertex3f(0.1826334f, 0.0186133385f, -1.f),
    C_Normal3f(0.9999336f, -0.0115244724f, 0.f),
    C_Vertex3f(0.182476282f, 0.001937747f, 0.f),
    C_Vertex3f(0.182476282f, 0.001937747f, -1.f),
    C_Normal3f(0.999865353f, -0.0164139196f, 0.f),
    C_Vertex3f(0.182266951f, -0.013182044f, 0.f),
    C_Vertex3f(0.182266951f, -0.013182044f, -1.f),
    C_Normal3f(0.999746978f, -0.0224915687f, 0.f),
    C_Vertex3f(0.182005405f, -0.0267460346f, 0.f),
    C_Vertex3f(0.182005405f, -0.0267460346f, -1.f),
    C_Normal3f(0.999542356f, -0.0302494541f, 0.f),
    C_Vertex3f(0.181691647f, -0.0387542248f, 0.f),
    C_Vertex3f(0.181691647f, -0.0387542248f, -1.f),
    C_Normal3f(0.99917978f, -0.0404938944f, 0.f),
    C_Vertex3f(0.181325674f, -0.0492066145f, 0.f),
    C_Vertex3f(0.181325674f, -0.0492066145f, -1.f),
    C_Normal3f(0.99850589f, -0.0546429604f, 0.f),
    C_Vertex3f(0.180907488f, -0.0581032038f, 0.f),
    C_Vertex3f(0.180907488f, -0.0581032038f, -1.f),
    C_Normal3f(0.997453392f, -0.0713205412f, 0.f),
    C_Vertex3f(0.180437088f, -0.0654439926f, 0.f),
    C_Vertex3f(0.180437088f, -0.0654439926f, -1.f),
    C_Normal3f(0.994805336f, -0.101795487f, 0.f),
    C_Vertex3f(0.180175781f, -0.0683364868f, 0.f),
    C_Vertex3f(0.180175781f, -0.0683364868f, -1.f),
    C_Normal3f(0.9894467f, -0.144896835f, 0.f),
    C_Vertex3f(0.179191589f, -0.0776157379f, 0.f),
    C_Vertex3f(0.179191589f, -0.0776157379f, -1.f),
    C_Normal3f(0.979438901f, -0.201741293f, 0.f),
    C_Vertex3f(0.176374912f, -0.0942912102f, 0.f),
    C_Vertex3f(0.176374912f, -0.0942912102f, -1.f),
    C_Normal3f(0.959077895f, -0.283142179f, 0.f),
    C_Vertex3f(0.172709942f, -0.109083652f, 0.f),
    C_Vertex3f(0.172709942f, -0.109083652f, -1.f),
    C_Normal3f(0.924420834f, -0.381374002f, 0.f),
    C_Vertex3f(0.168196678f, -0.121993065f, 0.f),
    C_Vertex3f(0.168196678f, -0.121993065f, -1.f),
    C_Normal3f(0.879334927f, -0.476203918f, 0.f),
    C_Vertex3f(0.162835121f, -0.133019447f, 0.f),
    C_Vertex3f(0.162835121f, -0.133019447f, -1.f),
    C_Normal3f(0.830216527f, -0.557441056f, 0.f),
    C_Vertex3f(0.156845093f, -0.14295435f, 0.f),
    C_Vertex3f(0.156845093f, -0.14295435f, -1.f),
    C_Normal3f(0.767401934f, -0.641166389f, 0.f),
    C_Vertex3f(0.150226593f, -0.151797771f, 0.f),
    C_Vertex3f(0.150226593f, -0.151797771f, -1.f),
    C_Normal3f(0.689908743f, -0.723896325f, 0.f),
    C_Vertex3f(0.142979622f, -0.159549713f, 0.f),
    C_Vertex3f(0.142979622f, -0.159549713f, -1.f),
    C_Normal3f(0.596778929f, -0.802405775f, 0.f),
    C_Vertex3f(0.135104179f, -0.166210175f, 0.f),
    C_Vertex3f(0.135104179f, -0.166210175f, -1.f),
    C_Normal3f(0.498434007f, -0.866927683f, 0.f),
    C_Vertex3f(0.126134396f, -0.172078133f, 0.f),
    C_Vertex3f(0.126134396f, -0.172078133f, -1.f),
    C_Normal3f(0.403477728f, -0.914989471f, 0.f),
    C_Vertex3f(0.116070271f, -0.177153587f, 0.f),
    C_Vertex3f(0.116070271f, -0.177153587f, -1.f),
    C_Normal3f(0.319494188f, -0.947588265f, 0.f),
    C_Vertex3f(0.104911804f, -0.181436539f, 0.f),
    C_Vertex3f(0.104911804f, -0.181436539f, -1.f),
    C_Normal3f(0.218360171f, -0.975868285f, 0.f),
    C_Vertex3f(0.0946273804f, -0.184383392f, 0.f),
    C_Vertex3f(0.0946273804f, -0.184383392f, -1.f),
    C_Normal3f(0.121676326f, -0.992569804f, 0.f),
    C_Vertex3f(0.0836868286f, -0.186185837f, 0.f),
    C_Vertex3f(0.0836868286f, -0.186185837f, -1.f),
    C_Normal3f(0.0604660846f, -0.998170257f, 0.f),
    C_Vertex3f(0.0701217651f, -0.187387466f, 0.f),
    C_Vertex3f(0.0701217651f, -0.187387466f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0539321899f, -0.187988281f, 0.f),
    C_Vertex3f(0.0539321899f, -0.187988281f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0539321899f, -0.187988281f, 0.f),
    C_Vertex3f(0.0539321899f, -0.187988281f, -1.f),
    C_Vertex3f(0.0200195313f, -0.187988281f, -1.f),
    C_Vertex3f(0.0200195313f, -0.187988281f, 0.f),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0200195313f, -0.187988281f, 0.f),
    C_Vertex3f(0.0200195313f, -0.187988281f, -1.f),
    C_Vertex3f(0.0200195313f, -0.119140625f, -1.f),
    C_Vertex3f(0.0200195313f, -0.119140625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0200195313f, -0.119140625f, 0.f),
    C_Vertex3f(0.0200195313f, -0.119140625f, -1.f),
    C_Normal3f(-0.0172241461f, 0.999851644f, 0.f),
    C_Vertex3f(0.0392827988f, -0.119140625f, 0.f),
    C_Vertex3f(0.0392827988f, -0.119140625f, -1.f),
    C_Normal3f(-0.0769076198f, 0.997038305f, 0.f),
    C_Vertex3f(0.049706459f, -0.118629217f, 0.f),
    C_Vertex3f(0.049706459f, -0.118629217f, -1.f),
    C_Normal3f(-0.140838534f, 0.990032554f, 0.f),
    C_Vertex3f(0.0591726303f, -0.117606401f, 0.f),
    C_Vertex3f(0.0591726303f, -0.117606401f, -1.f),
    C_Normal3f(-0.217567116f, 0.97604531f, 0.f),
    C_Vertex3f(0.0676813126f, -0.116072178f, 0.f),
    C_Vertex3f(0.0676813126f, -0.116072178f, -1.f),
    C_Normal3f(-0.309425265f, 0.950923741f, 0.f),
    C_Vertex3f(0.0752325058f, -0.114026546f, 0.f),
    C_Vertex3f(0.0752325058f, -0.114026546f, -1.f),
    C_Normal3f(-0.417888641f, 0.908498228f, 0.f),
    C_Vertex3f(0.08182621f, -0.111469507f, 0.f),
    C_Vertex3f(0.08182621f, -0.111469507f, -1.f),
    C_Normal3f(-0.541752934f, 0.840537846f, 0.f),
    C_Vertex3f(0.0874624252f, -0.10840106f, 0.f),
    C_Vertex3f(0.0874624252f, -0.10840106f, -1.f),
    C_Normal3f(-0.674345374f, 0.738416076f, 0.f),
    C_Vertex3f(0.0921411514f, -0.104821205f, 0.f),
    C_Vertex3f(0.0921411514f, -0.104821205f, -1.f),
    C_Normal3f(-0.804707468f, 0.59367156f, 0.f),
    C_Vertex3f(0.0958623886f, -0.100729942f, 0.f),
    C_Vertex3f(0.0958623886f, -0.100729942f, -1.f),
    C_Normal3f(-0.89313966f, 0.44977963f, 0.f),
    C_Vertex3f(0.0991184711f, -0.0953636169f, 0.f),
    C_Vertex3f(0.0991184711f, -0.0953636169f, -1.f),
    C_Normal3f(-0.943423927f, 0.331588954f, 0.f),
    C_Vertex3f(0.101909399f, -0.088722229f, 0.f),
    C_Vertex3f(0.101909399f, -0.088722229f, -1.f),
    C_Normal3f(-0.971340775f, 0.237691239f, 0.f),
    C_Vertex3f(0.104235172f, -0.0808057785f, 0.f),
    C_Vertex3f(0.104235172f, -0.0808057785f, -1.f),
    C_Normal3f(-0.986558378f, 0.163409352f, 0.f),
    C_Vertex3f(0.106095791f, -0.0716142654f, 0.f),
    C_Vertex3f(0.106095791f, -0.0716142654f, -1.f),
    C_Normal3f(-0.994560957f, 0.104156204f, 0.f),
    C_Vertex3f(0.107491255f, -0.0611476898f, 0.f),
    C_Vertex3f(0.107491255f, -0.0611476898f, -1.f),
    C_Normal3f(-0.998415351f, 0.0562740751f, 0.f),
    C_Vertex3f(0.108421564f, -0.0494060516f, 0.f),
    C_Vertex3f(0.108421564f, -0.0494060516f, -1.f),
    C_Normal3f(-0.999905348f, 0.0137610063f, 0.f),
    C_Vertex3f(0.108886719f, -0.0363893509f, 0.f),
    C_Vertex3f(0.108886719f, -0.0363893509f, -1.f),
    C_Normal3f(-0.999996364f, 0.002679314f, 0.f),
    C_Vertex3f(0.108886719f, -0.0156068802f, 0.f),
    C_Vertex3f(0.108886719f, -0.0156068802f, -1.f),
    C_Normal3f(-0.999969661f, 0.0077815894f, 0.f),
    C_Vertex3f(0.109008789f, 0.00917077065f, 0.f),
    C_Vertex3f(0.109008789f, 0.00917077065f, -1.f),
    C_Normal3f(-0.999894798f, 0.0145069147f, 0.f),
    C_Vertex3f(0.10925293f, 0.0314528942f, 0.f),
    C_Vertex3f(0.10925293f, 0.0314528942f, -1.f),
    C_Normal3f(-0.999734521f, 0.0230398905f, 0.f),
    C_Vertex3f(0.109619141f, 0.0512394905f, 0.f),
    C_Vertex3f(0.109619141f, 0.0512394905f, -1.f),
    C_Normal3f(-0.999414384f, 0.0342195518f, 0.f),
    C_Vertex3f(0.110107422f, 0.0685305595f, 0.f),
    C_Vertex3f(0.110107422f, 0.0685305595f, -1.f),
    C_Normal3f(-0.99877429f, 0.0494962186f, 0.f),
    C_Vertex3f(0.110717773f, 0.0833261013f, 0.f),
    C_Vertex3f(0.110717773f, 0.0833261013f, -1.f),
    C_Normal3f(-0.997432947f, 0.0716071501f, 0.f),
    C_Vertex3f(0.111450195f, 0.0956261158f, 0.f),
    C_Vertex3f(0.111450195f, 0.0956261158f, -1.f),
    C_Normal3f(-0.995060086f, 0.0992750898f, 0.f),
    C_Vertex3f(0.112304688f, 0.105430603f, 0.f),
    C_Vertex3f(0.112304688f, 0.105430603f, -1.f),
    C_Normal3f(-0.990011752f, 0.14098534f, 0.f),
    C_Vertex3f(0.112792969f, 0.109085083f, 0.f),
    C_Vertex3f(0.112792969f, 0.109085083f, -1.f),
    C_Normal3f(-0.979194641f, 0.202923477f, 0.f),
    C_Vertex3f(0.114549637f, 0.121194839f, 0.f),
    C_Vertex3f(0.114549637f, 0.121194839f, -1.f),
    C_Normal3f(-0.958168447f, 0.286204994f, 0.f),
    C_Vertex3f(0.119818687f, 0.142987251f, 0.f),
    C_Vertex3f(0.119818687f, 0.142987251f, -1.f),
    C_Normal3f(-0.916875005f, 0.399174333f, 0.f),
    C_Vertex3f(0.126843452f, 0.162352562f, 0.f),
    C_Vertex3f(0.126843452f, 0.162352562f, -1.f),
    C_Normal3f(-0.852102518f, 0.523374975f, 0.f),
    C_Vertex3f(0.135623932f, 0.179290771f, 0.f),
    C_Vertex3f(0.135623932f, 0.179290771f, -1.f),
    C_Normal3f(-0.762389362f, 0.647118568f, 0.f),
    C_Vertex3f(0.146160126f, 0.19380188f, 0.f),
    C_Vertex3f(0.146160126f, 0.19380188f, -1.f),
    C_Normal3f(-0.658375978f, 0.752689183f, 0.f),
    C_Vertex3f(0.158891678f, 0.206703186f, 0.f),
    C_Vertex3f(0.158891678f, 0.206703186f, -1.f),
    C_Normal3f(-0.547578037f, 0.83675462f, 0.f),
    C_Vertex3f(0.173818588f, 0.21799469f, 0.f),
    C_Vertex3f(0.173818588f, 0.21799469f, -1.f),
    C_Normal3f(-0.455889344f, 0.890036464f, 0.f),
    C_Vertex3f(0.190940857f, 0.227676392f, 0.f),
    C_Vertex3f(0.190940857f, 0.227676392f, -1.f),
    C_Normal3f(-0.385546684f, 0.922688305f, 0.f),
    C_Vertex3f(0.20059967f, 0.231712341f, 0.f),
    C_Vertex3f(0.20059967f, 0.231712341f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.427531123f, -0.90400064f, 0.f),
    C_Vertex3f(0.20059967f, 0.231712341f, 0.f),
    C_Vertex3f(0.20059967f, 0.231712341f, -1.f),
    C_Normal3f(-0.464289844f, -0.885683298f, 0.f),
    C_Vertex3f(0.194257736f, 0.234711647f, 0.f),
    C_Vertex3f(0.194257736f, 0.234711647f, -1.f),
    C_Normal3f(-0.512167394f, -0.858885705f, 0.f),
    C_Vertex3f(0.182511449f, 0.241194487f, 0.f),
    C_Vertex3f(0.182511449f, 0.241194487f, -1.f),
    C_Normal3f(-0.571930051f, -0.820302308f, 0.f),
    C_Vertex3f(0.171702743f, 0.248161554f, 0.f),
    C_Vertex3f(0.171702743f, 0.248161554f, -1.f),
    C_Normal3f(-0.633266926f, -0.77393347f, 0.f),
    C_Vertex3f(0.161831617f, 0.25561285f, 0.f),
    C_Vertex3f(0.161831617f, 0.25561285f, -1.f),
    C_Normal3f(-0.69480437f, -0.719198823f, 0.f),
    C_Vertex3f(0.152898073f, 0.263548374f, 0.f),
    C_Vertex3f(0.152898073f, 0.263548374f, -1.f),
    C_Normal3f(-0.754816294f, -0.655936301f, 0.f),
    C_Vertex3f(0.14490211f, 0.271968126f, 0.f),
    C_Vertex3f(0.14490211f, 0.271968126f, -1.f),
    C_Normal3f(-0.811349034f, -0.584561944f, 0.f),
    C_Vertex3f(0.137843728f, 0.280872107f, 0.f),
    C_Vertex3f(0.137843728f, 0.280872107f, -1.f),
    C_Normal3f(-0.862438083f, -0.506162584f, 0.f),
    C_Vertex3f(0.131722927f, 0.290260315f, 0.f),
    C_Vertex3f(0.131722927f, 0.290260315f, -1.f),
    C_Normal3f(-0.910814166f, -0.412816465f, 0.f),
    C_Vertex3f(0.126539707f, 0.300132751f, 0.f),
    C_Vertex3f(0.126539707f, 0.300132751f, -1.f),
    C_Normal3f(-0.947028279f, -0.321150184f, 0.f),
    C_Vertex3f(0.12200439f, 0.311702728f, 0.f),
    C_Vertex3f(0.12200439f, 0.311702728f, -1.f),
    C_Normal3f(-0.969584167f, -0.244758263f, 0.f),
    C_Vertex3f(0.118116975f, 0.324970245f, 0.f),
    C_Vertex3f(0.118116975f, 0.324970245f, -1.f),
    C_Normal3f(-0.983425617f, -0.181312054f, 0.f),
    C_Vertex3f(0.114877462f, 0.339935303f, 0.f),
    C_Vertex3f(0.114877462f, 0.339935303f, -1.f),
    C_Normal3f(-0.991719365f, -0.128424063f, 0.f),
    C_Vertex3f(0.112285852f, 0.3565979f, 0.f),
    C_Vertex3f(0.112285852f, 0.3565979f, -1.f),
    C_Normal3f(-0.996463656f, -0.0840249807f, 0.f),
    C_Vertex3f(0.110342145f, 0.374958038f, 0.f),
    C_Vertex3f(0.110342145f, 0.374958038f, -1.f),
    C_Normal3f(-0.998921216f, -0.0464356877f, 0.f),
    C_Vertex3f(0.10904634f, 0.395015717f, 0.f),
    C_Vertex3f(0.10904634f, 0.395015717f, -1.f),
    C_Normal3f(-0.999892175f, -0.0146844303f, 0.f),
    C_Vertex3f(0.108398438f, 0.416770935f, 0.f),
    C_Vertex3f(0.108398438f, 0.416770935f, -1.f),
    C_Normal3f(-0.999998927f, -0.0014710176f, 0.f),
    C_Vertex3f(0.108398438f, 0.43913269f, 0.f),
    C_Vertex3f(0.108398438f, 0.43913269f, -1.f),
    C_Normal3f(-0.999987125f, -0.00506940903f, 0.f),
    C_Vertex3f(0.108337402f, 0.458262682f, 0.f),
    C_Vertex3f(0.108337402f, 0.458262682f, -1.f),
    C_Normal3f(-0.999953985f, -0.00958491769f, 0.f),
    C_Vertex3f(0.108215332f, 0.475251913f, 0.f),
    C_Vertex3f(0.108215332f, 0.475251913f, -1.f),
    C_Normal3f(-0.999879777f, -0.0155026829f, 0.f),
    C_Vertex3f(0.108032227f, 0.490100384f, 0.f),
    C_Vertex3f(0.108032227f, 0.490100384f, -1.f),
    C_Normal3f(-0.999721587f, -0.0235949084f, 0.f),
    C_Vertex3f(0.107788086f, 0.502808094f, 0.f),
    C_Vertex3f(0.107788086f, 0.502808094f, -1.f),
    C_Normal3f(-0.999375761f, -0.0353268422f, 0.f),
    C_Vertex3f(0.10748291f, 0.513375044f, 0.f),
    C_Vertex3f(0.10748291f, 0.513375044f, -1.f),
    C_Normal3f(-0.998548746f, -0.0538557693f, 0.f),
    C_Vertex3f(0.107116699f, 0.521801233f, 0.f),
    C_Vertex3f(0.107116699f, 0.521801233f, -1.f),
    C_Normal3f(-0.996789098f, -0.0800729692f, 0.f),
    C_Vertex3f(0.106689453f, 0.528086662f, 0.f),
    C_Vertex3f(0.106689453f, 0.528086662f, -1.f),
    C_Normal3f(-0.994287312f, -0.106736541f, 0.f),
    C_Vertex3f(0.106445313f, 0.530158997f, 0.f),
    C_Vertex3f(0.106445313f, 0.530158997f, -1.f),
    C_Normal3f(-0.98301959f, -0.183500469f, 0.f),
    C_Vertex3f(0.105682373f, 0.537467957f, 0.f),
    C_Vertex3f(0.105682373f, 0.537467957f, -1.f),
    C_Normal3f(-0.952512503f, -0.304499447f, 0.f),
    C_Vertex3f(0.102764606f, 0.54987669f, 0.f),
    C_Vertex3f(0.102764606f, 0.54987669f, -1.f),
    C_Normal3f(-0.876078665f, -0.482168227f, 0.f),
    C_Vertex3f(0.0984549522f, 0.560076237f, 0.f),
    C_Vertex3f(0.0984549522f, 0.560076237f, -1.f),
    C_Normal3f(-0.732604623f, -0.680654407f, 0.f),
    C_Vertex3f(0.0927534103f, 0.568066597f, 0.f),
    C_Vertex3f(0.0927534103f, 0.568066597f, -1.f),
    C_Normal3f(-0.498431861f, -0.866928935f, 0.f),
    C_Vertex3f(0.0856599808f, 0.573847771f, 0.f),
    C_Vertex3f(0.0856599808f, 0.573847771f, -1.f),
    C_Normal3f(-0.283798218f, -0.95888406f, 0.f),
    C_Vertex3f(0.0751566887f, 0.578183651f, 0.f),
    C_Vertex3f(0.0751566887f, 0.578183651f, -1.f),
    C_Normal3f(-0.137491316f, -0.990502954f, 0.f),
    C_Vertex3f(0.0612435341f, 0.581074238f, 0.f),
    C_Vertex3f(0.0612435341f, 0.581074238f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.043920517f, 0.582519531f, 0.f),
    C_Vertex3f(0.043920517f, 0.582519531f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.043920517f, 0.582519531f, 0.f),
    C_Vertex3f(0.043920517f, 0.582519531f, -1.f),
    C_Vertex3f(0.0200195313f, 0.582519531f, -1.f),
    C_Vertex3f(0.0200195313f, 0.582519531f, 0.f),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0200195313f, 0.582519531f, 0.f),
    C_Vertex3f(0.0200195313f, 0.582519531f, -1.f),
    C_Vertex3f(0.0200195313f, 0.651367188f, -1.f),
    C_Vertex3f(0.0200195313f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0200195313f, 0.651367188f, 0.f),
    C_Vertex3f(0.0200195313f, 0.651367188f, -1.f),
    C_Normal3f(0.0162848905f, 0.99986732f, 0.f),
    C_Vertex3f(0.0551929474f, 0.651367188f, 0.f),
    C_Vertex3f(0.0551929474f, 0.651367188f, -1.f),
    C_Normal3f(0.0769083798f, 0.997038186f, 0.f),
    C_Vertex3f(0.0736551285f, 0.650493622f, 0.f),
    C_Vertex3f(0.0736551285f, 0.650493622f, -1.f),
    C_Normal3f(0.153726667f, 0.988113403f, 0.f),
    C_Vertex3f(0.0891675949f, 0.64874649f, 0.f),
    C_Vertex3f(0.0891675949f, 0.64874649f, -1.f),
    C_Normal3f(0.24387598f, 0.969806373f, 0.f),
    C_Vertex3f(0.101730347f, 0.646125793f, 0.f),
    C_Vertex3f(0.101730347f, 0.646125793f, -1.f),
    C_Normal3f(0.331742853f, 0.943369925f, 0.f),
    C_Vertex3f(0.106536865f, 0.644378662f, 0.f),
    C_Vertex3f(0.106536865f, 0.644378662f, -1.f),
    C_Normal3f(0.408231348f, 0.912878454f, 0.f),
    C_Vertex3f(0.113527298f, 0.64197731f, 0.f),
    C_Vertex3f(0.113527298f, 0.64197731f, -1.f),
    C_Normal3f(0.511936426f, 0.859023392f, 0.f),
    C_Vertex3f(0.126307011f, 0.635537624f, 0.f),
    C_Vertex3f(0.126307011f, 0.635537624f, -1.f),
    C_Normal3f(0.629549563f, 0.776960254f, 0.f),
    C_Vertex3f(0.137885571f, 0.627460957f, 0.f),
    C_Vertex3f(0.137885571f, 0.627460957f, -1.f),
    C_Normal3f(0.732898057f, 0.680338442f, 0.f),
    C_Vertex3f(0.148262978f, 0.617747307f, 0.f),
    C_Vertex3f(0.148262978f, 0.617747307f, -1.f),
    C_Normal3f(0.822123885f, 0.569308639f, 0.f),
    C_Vertex3f(0.157439232f, 0.606396675f, 0.f),
    C_Vertex3f(0.157439232f, 0.606396675f, -1.f),
    C_Normal3f(0.894999206f, 0.44606784f, 0.f),
    C_Vertex3f(0.165004253f, 0.593571663f, 0.f),
    C_Vertex3f(0.165004253f, 0.593571663f, -1.f),
    C_Normal3f(0.946085393f, 0.323917389f, 0.f),
    C_Vertex3f(0.170958042f, 0.57927227f, 0.f),
    C_Vertex3f(0.170958042f, 0.57927227f, -1.f),
    C_Normal3f(0.976107299f, 0.217289209f, 0.f),
    C_Vertex3f(0.175300598f, 0.563498497f, 0.f),
    C_Vertex3f(0.175300598f, 0.563498497f, -1.f),
    C_Normal3f(0.990013003f, 0.14097546f, 0.f),
    C_Vertex3f(0.177349091f, 0.550562382f, 0.f),
    C_Vertex3f(0.177349091f, 0.550562382f, -1.f),
    C_Normal3f(0.994324267f, 0.106392391f, 0.f),
    C_Vertex3f(0.178566933f, 0.540560365f, 0.f),
    C_Vertex3f(0.178566933f, 0.540560365f, -1.f),
    C_Normal3f(0.996611238f, 0.0822556317f, 0.f),
    C_Vertex3f(0.179636955f, 0.529180408f, 0.f),
    C_Vertex3f(0.179636955f, 0.529180408f, -1.f),
    C_Normal3f(0.998016059f, 0.062959671f, 0.f),
    C_Vertex3f(0.180559158f, 0.51642251f, 0.f),
    C_Vertex3f(0.180559158f, 0.51642251f, -1.f),
    C_Normal3f(0.998885572f, 0.04719745f, 0.f),
    C_Vertex3f(0.181333542f, 0.502286673f, 0.f),
    C_Vertex3f(0.181333542f, 0.502286673f, -1.f),
    C_Normal3f(0.999418855f, 0.0340888537f, 0.f),
    C_Vertex3f(0.181960106f, 0.486772895f, 0.f),
    C_Vertex3f(0.181960106f, 0.486772895f, -1.f),
    C_Normal3f(0.999734998f, 0.0230211392f, 0.f),
    C_Vertex3f(0.18243885f, 0.469881177f, 0.f),
    C_Vertex3f(0.18243885f, 0.469881177f, -1.f),
    C_Normal3f(0.999908149f, 0.0135553917f, 0.f),
    C_Vertex3f(0.182769775f, 0.451611519f, 0.f),
    C_Vertex3f(0.182769775f, 0.451611519f, -1.f),
    C_Normal3f(0.999889553f, 0.0148593429f, 0.f),
    C_Vertex3f(0.182952881f, 0.431963921f, 0.f),
    C_Vertex3f(0.182952881f, 0.431963921f, -1.f),
    C_Normal3f(0.999624848f, 0.0273901317f, 0.f),
    C_Vertex3f(0.183331132f, 0.413837671f, 0.f),
    C_Vertex3f(0.183331132f, 0.413837671f, -1.f),
    C_Normal3f(0.9991045f, 0.0423099659f, 0.f),
    C_Vertex3f(0.183904529f, 0.397232771f, 0.f),
    C_Vertex3f(0.183904529f, 0.397232771f, -1.f),
    C_Normal3f(0.998176634f, 0.0603604876f, 0.f),
    C_Vertex3f(0.184673071f, 0.38214922f, 0.f),
    C_Vertex3f(0.184673071f, 0.38214922f, -1.f),
    C_Normal3f(0.996581376f, 0.0826176852f, 0.f),
    C_Vertex3f(0.185636759f, 0.368587017f, 0.f),
    C_Vertex3f(0.185636759f, 0.368587017f, -1.f),
    C_Normal3f(0.993854225f, 0.110697247f, 0.f),
    C_Vertex3f(0.186795592f, 0.356546164f, 0.f),
    C_Vertex3f(0.186795592f, 0.356546164f, -1.f),
    C_Normal3f(0.989118099f, 0.147123814f, 0.f),
    C_Vertex3f(0.188149571f, 0.346026659f, 0.f),
    C_Vertex3f(0.188149571f, 0.346026659f, -1.f),
    C_Normal3f(0.979419172f, 0.20183675f, 0.f),
    C_Vertex3f(0.189698696f, 0.337028503f, 0.f),
    C_Vertex3f(0.189698696f, 0.337028503f, -1.f),
    C_Normal3f(0.952419221f, 0.304791123f, 0.f),
    C_Vertex3f(0.192315102f, 0.325813293f, 0.f),
    C_Vertex3f(0.192315102f, 0.325813293f, -1.f),
    C_Normal3f(0.899055541f, 0.43783465f, 0.f),
    C_Vertex3f(0.197575569f, 0.312414646f, 0.f),
    C_Vertex3f(0.197575569f, 0.312414646f, -1.f),
    C_Normal3f(0.813248992f, 0.581915855f, 0.f),
    C_Vertex3f(0.204607964f, 0.300570965f, 0.f),
    C_Vertex3f(0.204607964f, 0.300570965f, -1.f),
    C_Normal3f(0.700482607f, 0.713669479f, 0.f),
    C_Vertex3f(0.213412285f, 0.290282249f, 0.f),
    C_Vertex3f(0.213412285f, 0.290282249f, -1.f),
    C_Normal3f(0.561121762f, 0.827733219f, 0.f),
    C_Vertex3f(0.223988533f, 0.2815485f, 0.f),
    C_Vertex3f(0.223988533f, 0.2815485f, -1.f),
    C_Normal3f(0.399312466f, 0.916814923f, 0.f),
    C_Vertex3f(0.236118793f, 0.274889469f, 0.f),
    C_Vertex3f(0.236118793f, 0.274889469f, -1.f),
    C_Normal3f(0.23821184f, 0.971213281f, 0.f),
    C_Vertex3f(0.249803066f, 0.270305157f, 0.f),
    C_Vertex3f(0.249803066f, 0.270305157f, -1.f),
    C_Normal3f(0.114623353f, 0.993409038f, 0.f),
    C_Vertex3f(0.265041351f, 0.267795563f, 0.f),
    C_Vertex3f(0.265041351f, 0.267795563f, -1.f),
    C_Normal3f(0.0258886386f, 0.999664783f, 0.f),
    C_Vertex3f(0.2734375f, 0.267578125f, 0.f),
    C_Vertex3f(0.2734375f, 0.267578125f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.191874504f, 0.138048172f, 0.f),
    C_Vertex3f(0.197163582f, 0.15144968f, 0.f),
    C_Vertex3f(0.20059967f, 0.231712341f, 0.f),
    C_Vertex3f(0.204234123f, 0.163295269f, 0.f),
    C_Vertex3f(0.204607964f, 0.300570965f, 0.f),
    C_Vertex3f(0.213086128f, 0.173584938f, 0.f),
    C_Vertex3f(0.213412285f, 0.290282249f, 0.f),
    C_Vertex3f(0.223719597f, 0.182318687f, 0.f),
    C_Vertex3f(0.223988533f, 0.2815485f, 0.f),
    C_Vertex3f(0.235915661f, 0.188977718f, 0.f),
    C_Vertex3f(0.236118793f, 0.274889469f, 0.f),
    C_Vertex3f(0.24967432f, 0.193562031f, 0.f),
    C_Vertex3f(0.249803066f, 0.270305157f, 0.f),
    C_Vertex3f(0.264995575f, 0.196071625f, 0.f),
    C_Vertex3f(0.265041351f, 0.267795563f, 0.f),
    C_Vertex3f(0.2734375f, 0.196289063f, 0.f),
    C_Vertex3f(0.2734375f, 0.267578125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.197575569f, 0.312414646f, 0.f),
    C_Vertex3f(0.192315102f, 0.325813293f, 0.f),
    C_Vertex3f(0.194257736f, 0.234711647f, 0.f),
    C_Vertex3f(0.20059967f, 0.231712341f, 0.f),
    C_Vertex3f(0.204607964f, 0.300570965f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.194257736f, 0.234711647f, 0.f),
    C_Vertex3f(0.192315102f, 0.325813293f, 0.f),
    C_Vertex3f(0.189698696f, 0.337028503f, 0.f),
    C_Vertex3f(0.188149571f, 0.346026659f, 0.f),
    C_Vertex3f(0.186795592f, 0.356546164f, 0.f),
    C_Vertex3f(0.185636759f, 0.368587017f, 0.f),
    C_Vertex3f(0.184673071f, 0.38214922f, 0.f),
    C_Vertex3f(0.182511449f, 0.241194487f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.182511449f, 0.241194487f, 0.f),
    C_Vertex3f(0.184673071f, 0.38214922f, 0.f),
    C_Vertex3f(0.183904529f, 0.397232771f, 0.f),
    C_Vertex3f(0.183331132f, 0.413837671f, 0.f),
    C_Vertex3f(0.182952881f, 0.431963921f, 0.f),
    C_Vertex3f(0.182769775f, 0.451611519f, 0.f),
    C_Vertex3f(0.18243885f, 0.469881177f, 0.f),
    C_Vertex3f(0.181960106f, 0.486772895f, 0.f),
    C_Vertex3f(0.181333542f, 0.502286673f, 0.f),
    C_Vertex3f(0.180559158f, 0.51642251f, 0.f),
    C_Vertex3f(0.179636955f, 0.529180408f, 0.f),
    C_Vertex3f(0.178566933f, 0.540560365f, 0.f),
    C_Vertex3f(0.177349091f, 0.550562382f, 0.f),
    C_Vertex3f(0.175300598f, 0.563498497f, 0.f),
    C_Vertex3f(0.171702743f, 0.248161554f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.131722927f, 0.290260315f, 0.f),
    C_Vertex3f(0.137843728f, 0.280872107f, 0.f),
    C_Vertex3f(0.137885571f, 0.627460957f, 0.f),
    C_Vertex3f(0.14490211f, 0.271968126f, 0.f),
    C_Vertex3f(0.148262978f, 0.617747307f, 0.f),
    C_Vertex3f(0.152898073f, 0.263548374f, 0.f),
    C_Vertex3f(0.157439232f, 0.606396675f, 0.f),
    C_Vertex3f(0.161831617f, 0.25561285f, 0.f),
    C_Vertex3f(0.165004253f, 0.593571663f, 0.f),
    C_Vertex3f(0.171702743f, 0.248161554f, 0.f),
    C_Vertex3f(0.170958042f, 0.57927227f, 0.f),
    C_Vertex3f(0.175300598f, 0.563498497f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.118116975f, 0.324970245f, 0.f),
    C_Vertex3f(0.12200439f, 0.311702728f, 0.f),
    C_Vertex3f(0.126307011f, 0.635537624f, 0.f),
    C_Vertex3f(0.126539707f, 0.300132751f, 0.f),
    C_Vertex3f(0.137885571f, 0.627460957f, 0.f),
    C_Vertex3f(0.131722927f, 0.290260315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110342145f, 0.374958038f, 0.f),
    C_Vertex3f(0.112285852f, 0.3565979f, 0.f),
    C_Vertex3f(0.113527298f, 0.64197731f, 0.f),
    C_Vertex3f(0.114877462f, 0.339935303f, 0.f),
    C_Vertex3f(0.126307011f, 0.635537624f, 0.f),
    C_Vertex3f(0.118116975f, 0.324970245f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.113527298f, 0.64197731f, 0.f),
    C_Vertex3f(0.106536865f, 0.644378662f, 0.f),
    C_Vertex3f(0.106689453f, 0.528086662f, 0.f),
    C_Vertex3f(0.107116699f, 0.521801233f, 0.f),
    C_Vertex3f(0.10748291f, 0.513375044f, 0.f),
    C_Vertex3f(0.107788086f, 0.502808094f, 0.f),
    C_Vertex3f(0.108032227f, 0.490100384f, 0.f),
    C_Vertex3f(0.108215332f, 0.475251913f, 0.f),
    C_Vertex3f(0.108337402f, 0.458262682f, 0.f),
    C_Vertex3f(0.108398438f, 0.43913269f, 0.f),
    C_Vertex3f(0.10904634f, 0.395015717f, 0.f),
    C_Vertex3f(0.110342145f, 0.374958038f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.106536865f, 0.644378662f, 0.f),
    C_Vertex3f(0.101730347f, 0.646125793f, 0.f),
    C_Vertex3f(0.102764606f, 0.54987669f, 0.f),
    C_Vertex3f(0.105682373f, 0.537467957f, 0.f),
    C_Vertex3f(0.106445313f, 0.530158997f, 0.f),
    C_Vertex3f(0.106689453f, 0.528086662f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0751566887f, 0.578183651f, 0.f),
    C_Vertex3f(0.0856599808f, 0.573847771f, 0.f),
    C_Vertex3f(0.0891675949f, 0.64874649f, 0.f),
    C_Vertex3f(0.0927534103f, 0.568066597f, 0.f),
    C_Vertex3f(0.101730347f, 0.646125793f, 0.f),
    C_Vertex3f(0.0984549522f, 0.560076237f, 0.f),
    C_Vertex3f(0.102764606f, 0.54987669f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0200195313f, 0.651367188f, 0.f),
    C_Vertex3f(0.043920517f, 0.582519531f, 0.f),
    C_Vertex3f(0.0551929474f, 0.651367188f, 0.f),
    C_Vertex3f(0.0612435341f, 0.581074238f, 0.f),
    C_Vertex3f(0.0736551285f, 0.650493622f, 0.f),
    C_Vertex3f(0.0751566887f, 0.578183651f, 0.f),
    C_Vertex3f(0.0891675949f, 0.64874649f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190940857f, 0.227676392f, 0.f),
    C_Vertex3f(0.173818588f, 0.21799469f, 0.f),
    C_Vertex3f(0.183566451f, 0.06927526f, 0.f),
    C_Vertex3f(0.184289694f, 0.0834743977f, 0.f),
    C_Vertex3f(0.185219049f, 0.0963294506f, 0.f),
    C_Vertex3f(0.186354518f, 0.107840419f, 0.f),
    C_Vertex3f(0.187696099f, 0.118007302f, 0.f),
    C_Vertex3f(0.189243793f, 0.126830101f, 0.f),
    C_Vertex3f(0.191874504f, 0.138048172f, 0.f),
    C_Vertex3f(0.20059967f, 0.231712341f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.182738304f, 0.0368447304f, 0.f),
    C_Vertex3f(0.183049321f, 0.0537320375f, 0.f),
    C_Vertex3f(0.180907488f, -0.0581032038f, 0.f),
    C_Vertex3f(0.181325674f, -0.0492066145f, 0.f),
    C_Vertex3f(0.181691647f, -0.0387542248f, 0.f),
    C_Vertex3f(0.182005405f, -0.0267460346f, 0.f),
    C_Vertex3f(0.182266951f, -0.013182044f, 0.f),
    C_Vertex3f(0.182476282f, 0.001937747f, 0.f),
    C_Vertex3f(0.1826334f, 0.0186133385f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.183049321f, 0.0537320375f, 0.f),
    C_Vertex3f(0.183566451f, 0.06927526f, 0.f),
    C_Vertex3f(0.176374912f, -0.0942912102f, 0.f),
    C_Vertex3f(0.179191589f, -0.0776157379f, 0.f),
    C_Vertex3f(0.180175781f, -0.0683364868f, 0.f),
    C_Vertex3f(0.180437088f, -0.0654439926f, 0.f),
    C_Vertex3f(0.180907488f, -0.0581032038f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.173818588f, 0.21799469f, 0.f),
    C_Vertex3f(0.158891678f, 0.206703186f, 0.f),
    C_Vertex3f(0.162835121f, -0.133019447f, 0.f),
    C_Vertex3f(0.168196678f, -0.121993065f, 0.f),
    C_Vertex3f(0.172709942f, -0.109083652f, 0.f),
    C_Vertex3f(0.176374912f, -0.0942912102f, 0.f),
    C_Vertex3f(0.183566451f, 0.06927526f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.114549637f, 0.121194839f, 0.f),
    C_Vertex3f(0.116070271f, -0.177153587f, 0.f),
    C_Vertex3f(0.119818687f, 0.142987251f, 0.f),
    C_Vertex3f(0.126134396f, -0.172078133f, 0.f),
    C_Vertex3f(0.126843452f, 0.162352562f, 0.f),
    C_Vertex3f(0.135104179f, -0.166210175f, 0.f),
    C_Vertex3f(0.135623932f, 0.179290771f, 0.f),
    C_Vertex3f(0.142979622f, -0.159549713f, 0.f),
    C_Vertex3f(0.146160126f, 0.19380188f, 0.f),
    C_Vertex3f(0.150226593f, -0.151797771f, 0.f),
    C_Vertex3f(0.158891678f, 0.206703186f, 0.f),
    C_Vertex3f(0.156845093f, -0.14295435f, 0.f),
    C_Vertex3f(0.162835121f, -0.133019447f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116070271f, -0.177153587f, 0.f),
    C_Vertex3f(0.114549637f, 0.121194839f, 0.f),
    C_Vertex3f(0.112792969f, 0.109085083f, 0.f),
    C_Vertex3f(0.112304688f, 0.105430603f, 0.f),
    C_Vertex3f(0.111450195f, 0.0956261158f, 0.f),
    C_Vertex3f(0.110717773f, 0.0833261013f, 0.f),
    C_Vertex3f(0.110107422f, 0.0685305595f, 0.f),
    C_Vertex3f(0.109619141f, 0.0512394905f, 0.f),
    C_Vertex3f(0.10925293f, 0.0314528942f, 0.f),
    C_Vertex3f(0.109008789f, 0.00917077065f, 0.f),
    C_Vertex3f(0.108886719f, -0.0156068802f, 0.f),
    C_Vertex3f(0.108886719f, -0.0363893509f, 0.f),
    C_Vertex3f(0.108421564f, -0.0494060516f, 0.f),
    C_Vertex3f(0.107491255f, -0.0611476898f, 0.f),
    C_Vertex3f(0.106095791f, -0.0716142654f, 0.f),
    C_Vertex3f(0.104911804f, -0.181436539f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.104911804f, -0.181436539f, 0.f),
    C_Vertex3f(0.106095791f, -0.0716142654f, 0.f),
    C_Vertex3f(0.104235172f, -0.0808057785f, 0.f),
    C_Vertex3f(0.101909399f, -0.088722229f, 0.f),
    C_Vertex3f(0.0991184711f, -0.0953636169f, 0.f),
    C_Vertex3f(0.0958623886f, -0.100729942f, 0.f),
    C_Vertex3f(0.0946273804f, -0.184383392f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0958623886f, -0.100729942f, 0.f),
    C_Vertex3f(0.0921411514f, -0.104821205f, 0.f),
    C_Vertex3f(0.0946273804f, -0.184383392f, 0.f),
    C_Vertex3f(0.0874624252f, -0.10840106f, 0.f),
    C_Vertex3f(0.0836868286f, -0.186185837f, 0.f),
    C_Vertex3f(0.08182621f, -0.111469507f, 0.f),
    C_Vertex3f(0.0752325058f, -0.114026546f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0701217651f, -0.187387466f, 0.f),
    C_Vertex3f(0.0836868286f, -0.186185837f, 0.f),
    C_Vertex3f(0.0752325058f, -0.114026546f, 0.f),
    C_Vertex3f(0.0676813126f, -0.116072178f, 0.f),
    C_Vertex3f(0.0591726303f, -0.117606401f, 0.f),
    C_Vertex3f(0.0539321899f, -0.187988281f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0539321899f, -0.187988281f, 0.f),
    C_Vertex3f(0.0591726303f, -0.117606401f, 0.f),
    C_Vertex3f(0.049706459f, -0.118629217f, 0.f),
    C_Vertex3f(0.0392827988f, -0.119140625f, 0.f),
    C_Vertex3f(0.0200195313f, -0.119140625f, 0.f),
    C_Vertex3f(0.0200195313f, -0.187988281f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.10904634f, 0.395015717f, 0.f),
    C_Vertex3f(0.108398438f, 0.43913269f, 0.f),
    C_Vertex3f(0.108398438f, 0.416770935f, 0.f),
    C_Vertex3f(0.043920517f, 0.582519531f, 0.f),
    C_Vertex3f(0.0200195313f, 0.651367188f, 0.f),
    C_Vertex3f(0.0200195313f, 0.582519531f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.191874504f, 0.138048172f, -1.f),
    C_Vertex3f(0.197163582f, 0.15144968f, -1.f),
    C_Vertex3f(0.20059967f, 0.231712341f, -1.f),
    C_Vertex3f(0.204234123f, 0.163295269f, -1.f),
    C_Vertex3f(0.204607964f, 0.300570965f, -1.f),
    C_Vertex3f(0.213086128f, 0.173584938f, -1.f),
    C_Vertex3f(0.213412285f, 0.290282249f, -1.f),
    C_Vertex3f(0.223719597f, 0.182318687f, -1.f),
    C_Vertex3f(0.223988533f, 0.2815485f, -1.f),
    C_Vertex3f(0.235915661f, 0.188977718f, -1.f),
    C_Vertex3f(0.236118793f, 0.274889469f, -1.f),
    C_Vertex3f(0.24967432f, 0.193562031f, -1.f),
    C_Vertex3f(0.249803066f, 0.270305157f, -1.f),
    C_Vertex3f(0.264995575f, 0.196071625f, -1.f),
    C_Vertex3f(0.265041351f, 0.267795563f, -1.f),
    C_Vertex3f(0.2734375f, 0.196289063f, -1.f),
    C_Vertex3f(0.2734375f, 0.267578125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.197575569f, 0.312414646f, -1.f),
    C_Vertex3f(0.192315102f, 0.325813293f, -1.f),
    C_Vertex3f(0.194257736f, 0.234711647f, -1.f),
    C_Vertex3f(0.20059967f, 0.231712341f, -1.f),
    C_Vertex3f(0.204607964f, 0.300570965f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.194257736f, 0.234711647f, -1.f),
    C_Vertex3f(0.192315102f, 0.325813293f, -1.f),
    C_Vertex3f(0.189698696f, 0.337028503f, -1.f),
    C_Vertex3f(0.188149571f, 0.346026659f, -1.f),
    C_Vertex3f(0.186795592f, 0.356546164f, -1.f),
    C_Vertex3f(0.185636759f, 0.368587017f, -1.f),
    C_Vertex3f(0.184673071f, 0.38214922f, -1.f),
    C_Vertex3f(0.182511449f, 0.241194487f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.182511449f, 0.241194487f, -1.f),
    C_Vertex3f(0.184673071f, 0.38214922f, -1.f),
    C_Vertex3f(0.183904529f, 0.397232771f, -1.f),
    C_Vertex3f(0.183331132f, 0.413837671f, -1.f),
    C_Vertex3f(0.182952881f, 0.431963921f, -1.f),
    C_Vertex3f(0.182769775f, 0.451611519f, -1.f),
    C_Vertex3f(0.18243885f, 0.469881177f, -1.f),
    C_Vertex3f(0.181960106f, 0.486772895f, -1.f),
    C_Vertex3f(0.181333542f, 0.502286673f, -1.f),
    C_Vertex3f(0.180559158f, 0.51642251f, -1.f),
    C_Vertex3f(0.179636955f, 0.529180408f, -1.f),
    C_Vertex3f(0.178566933f, 0.540560365f, -1.f),
    C_Vertex3f(0.177349091f, 0.550562382f, -1.f),
    C_Vertex3f(0.175300598f, 0.563498497f, -1.f),
    C_Vertex3f(0.171702743f, 0.248161554f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.131722927f, 0.290260315f, -1.f),
    C_Vertex3f(0.137843728f, 0.280872107f, -1.f),
    C_Vertex3f(0.137885571f, 0.627460957f, -1.f),
    C_Vertex3f(0.14490211f, 0.271968126f, -1.f),
    C_Vertex3f(0.148262978f, 0.617747307f, -1.f),
    C_Vertex3f(0.152898073f, 0.263548374f, -1.f),
    C_Vertex3f(0.157439232f, 0.606396675f, -1.f),
    C_Vertex3f(0.161831617f, 0.25561285f, -1.f),
    C_Vertex3f(0.165004253f, 0.593571663f, -1.f),
    C_Vertex3f(0.171702743f, 0.248161554f, -1.f),
    C_Vertex3f(0.170958042f, 0.57927227f, -1.f),
    C_Vertex3f(0.175300598f, 0.563498497f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.118116975f, 0.324970245f, -1.f),
    C_Vertex3f(0.12200439f, 0.311702728f, -1.f),
    C_Vertex3f(0.126307011f, 0.635537624f, -1.f),
    C_Vertex3f(0.126539707f, 0.300132751f, -1.f),
    C_Vertex3f(0.137885571f, 0.627460957f, -1.f),
    C_Vertex3f(0.131722927f, 0.290260315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110342145f, 0.374958038f, -1.f),
    C_Vertex3f(0.112285852f, 0.3565979f, -1.f),
    C_Vertex3f(0.113527298f, 0.64197731f, -1.f),
    C_Vertex3f(0.114877462f, 0.339935303f, -1.f),
    C_Vertex3f(0.126307011f, 0.635537624f, -1.f),
    C_Vertex3f(0.118116975f, 0.324970245f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.113527298f, 0.64197731f, -1.f),
    C_Vertex3f(0.106536865f, 0.644378662f, -1.f),
    C_Vertex3f(0.106689453f, 0.528086662f, -1.f),
    C_Vertex3f(0.107116699f, 0.521801233f, -1.f),
    C_Vertex3f(0.10748291f, 0.513375044f, -1.f),
    C_Vertex3f(0.107788086f, 0.502808094f, -1.f),
    C_Vertex3f(0.108032227f, 0.490100384f, -1.f),
    C_Vertex3f(0.108215332f, 0.475251913f, -1.f),
    C_Vertex3f(0.108337402f, 0.458262682f, -1.f),
    C_Vertex3f(0.108398438f, 0.43913269f, -1.f),
    C_Vertex3f(0.10904634f, 0.395015717f, -1.f),
    C_Vertex3f(0.110342145f, 0.374958038f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.106536865f, 0.644378662f, -1.f),
    C_Vertex3f(0.101730347f, 0.646125793f, -1.f),
    C_Vertex3f(0.102764606f, 0.54987669f, -1.f),
    C_Vertex3f(0.105682373f, 0.537467957f, -1.f),
    C_Vertex3f(0.106445313f, 0.530158997f, -1.f),
    C_Vertex3f(0.106689453f, 0.528086662f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0751566887f, 0.578183651f, -1.f),
    C_Vertex3f(0.0856599808f, 0.573847771f, -1.f),
    C_Vertex3f(0.0891675949f, 0.64874649f, -1.f),
    C_Vertex3f(0.0927534103f, 0.568066597f, -1.f),
    C_Vertex3f(0.101730347f, 0.646125793f, -1.f),
    C_Vertex3f(0.0984549522f, 0.560076237f, -1.f),
    C_Vertex3f(0.102764606f, 0.54987669f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0200195313f, 0.651367188f, -1.f),
    C_Vertex3f(0.043920517f, 0.582519531f, -1.f),
    C_Vertex3f(0.0551929474f, 0.651367188f, -1.f),
    C_Vertex3f(0.0612435341f, 0.581074238f, -1.f),
    C_Vertex3f(0.0736551285f, 0.650493622f, -1.f),
    C_Vertex3f(0.0751566887f, 0.578183651f, -1.f),
    C_Vertex3f(0.0891675949f, 0.64874649f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190940857f, 0.227676392f, -1.f),
    C_Vertex3f(0.173818588f, 0.21799469f, -1.f),
    C_Vertex3f(0.183566451f, 0.06927526f, -1.f),
    C_Vertex3f(0.184289694f, 0.0834743977f, -1.f),
    C_Vertex3f(0.185219049f, 0.0963294506f, -1.f),
    C_Vertex3f(0.186354518f, 0.107840419f, -1.f),
    C_Vertex3f(0.187696099f, 0.118007302f, -1.f),
    C_Vertex3f(0.189243793f, 0.126830101f, -1.f),
    C_Vertex3f(0.191874504f, 0.138048172f, -1.f),
    C_Vertex3f(0.20059967f, 0.231712341f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.182738304f, 0.0368447304f, -1.f),
    C_Vertex3f(0.183049321f, 0.0537320375f, -1.f),
    C_Vertex3f(0.180907488f, -0.0581032038f, -1.f),
    C_Vertex3f(0.181325674f, -0.0492066145f, -1.f),
    C_Vertex3f(0.181691647f, -0.0387542248f, -1.f),
    C_Vertex3f(0.182005405f, -0.0267460346f, -1.f),
    C_Vertex3f(0.182266951f, -0.013182044f, -1.f),
    C_Vertex3f(0.182476282f, 0.001937747f, -1.f),
    C_Vertex3f(0.1826334f, 0.0186133385f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.183049321f, 0.0537320375f, -1.f),
    C_Vertex3f(0.183566451f, 0.06927526f, -1.f),
    C_Vertex3f(0.176374912f, -0.0942912102f, -1.f),
    C_Vertex3f(0.179191589f, -0.0776157379f, -1.f),
    C_Vertex3f(0.180175781f, -0.0683364868f, -1.f),
    C_Vertex3f(0.180437088f, -0.0654439926f, -1.f),
    C_Vertex3f(0.180907488f, -0.0581032038f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.173818588f, 0.21799469f, -1.f),
    C_Vertex3f(0.158891678f, 0.206703186f, -1.f),
    C_Vertex3f(0.162835121f, -0.133019447f, -1.f),
    C_Vertex3f(0.168196678f, -0.121993065f, -1.f),
    C_Vertex3f(0.172709942f, -0.109083652f, -1.f),
    C_Vertex3f(0.176374912f, -0.0942912102f, -1.f),
    C_Vertex3f(0.183566451f, 0.06927526f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.114549637f, 0.121194839f, -1.f),
    C_Vertex3f(0.116070271f, -0.177153587f, -1.f),
    C_Vertex3f(0.119818687f, 0.142987251f, -1.f),
    C_Vertex3f(0.126134396f, -0.172078133f, -1.f),
    C_Vertex3f(0.126843452f, 0.162352562f, -1.f),
    C_Vertex3f(0.135104179f, -0.166210175f, -1.f),
    C_Vertex3f(0.135623932f, 0.179290771f, -1.f),
    C_Vertex3f(0.142979622f, -0.159549713f, -1.f),
    C_Vertex3f(0.146160126f, 0.19380188f, -1.f),
    C_Vertex3f(0.150226593f, -0.151797771f, -1.f),
    C_Vertex3f(0.158891678f, 0.206703186f, -1.f),
    C_Vertex3f(0.156845093f, -0.14295435f, -1.f),
    C_Vertex3f(0.162835121f, -0.133019447f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116070271f, -0.177153587f, -1.f),
    C_Vertex3f(0.114549637f, 0.121194839f, -1.f),
    C_Vertex3f(0.112792969f, 0.109085083f, -1.f),
    C_Vertex3f(0.112304688f, 0.105430603f, -1.f),
    C_Vertex3f(0.111450195f, 0.0956261158f, -1.f),
    C_Vertex3f(0.110717773f, 0.0833261013f, -1.f),
    C_Vertex3f(0.110107422f, 0.0685305595f, -1.f),
    C_Vertex3f(0.109619141f, 0.0512394905f, -1.f),
    C_Vertex3f(0.10925293f, 0.0314528942f, -1.f),
    C_Vertex3f(0.109008789f, 0.00917077065f, -1.f),
    C_Vertex3f(0.108886719f, -0.0156068802f, -1.f),
    C_Vertex3f(0.108886719f, -0.0363893509f, -1.f),
    C_Vertex3f(0.108421564f, -0.0494060516f, -1.f),
    C_Vertex3f(0.107491255f, -0.0611476898f, -1.f),
    C_Vertex3f(0.106095791f, -0.0716142654f, -1.f),
    C_Vertex3f(0.104911804f, -0.181436539f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.104911804f, -0.181436539f, -1.f),
    C_Vertex3f(0.106095791f, -0.0716142654f, -1.f),
    C_Vertex3f(0.104235172f, -0.0808057785f, -1.f),
    C_Vertex3f(0.101909399f, -0.088722229f, -1.f),
    C_Vertex3f(0.0991184711f, -0.0953636169f, -1.f),
    C_Vertex3f(0.0958623886f, -0.100729942f, -1.f),
    C_Vertex3f(0.0946273804f, -0.184383392f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0958623886f, -0.100729942f, -1.f),
    C_Vertex3f(0.0921411514f, -0.104821205f, -1.f),
    C_Vertex3f(0.0946273804f, -0.184383392f, -1.f),
    C_Vertex3f(0.0874624252f, -0.10840106f, -1.f),
    C_Vertex3f(0.0836868286f, -0.186185837f, -1.f),
    C_Vertex3f(0.08182621f, -0.111469507f, -1.f),
    C_Vertex3f(0.0752325058f, -0.114026546f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0701217651f, -0.187387466f, -1.f),
    C_Vertex3f(0.0836868286f, -0.186185837f, -1.f),
    C_Vertex3f(0.0752325058f, -0.114026546f, -1.f),
    C_Vertex3f(0.0676813126f, -0.116072178f, -1.f),
    C_Vertex3f(0.0591726303f, -0.117606401f, -1.f),
    C_Vertex3f(0.0539321899f, -0.187988281f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0539321899f, -0.187988281f, -1.f),
    C_Vertex3f(0.0591726303f, -0.117606401f, -1.f),
    C_Vertex3f(0.049706459f, -0.118629217f, -1.f),
    C_Vertex3f(0.0392827988f, -0.119140625f, -1.f),
    C_Vertex3f(0.0200195313f, -0.119140625f, -1.f),
    C_Vertex3f(0.0200195313f, -0.187988281f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.10904634f, 0.395015717f, -1.f),
    C_Vertex3f(0.108398438f, 0.43913269f, -1.f),
    C_Vertex3f(0.108398438f, 0.416770935f, -1.f),
    C_Vertex3f(0.043920517f, 0.582519531f, -1.f),
    C_Vertex3f(0.0200195313f, 0.651367188f, -1.f),
    C_Vertex3f(0.0200195313f, 0.582519531f, -1.f),
    C_End(),
    C_Translatef(0.298828125f, 0.f, 0.f),
};
static const Call char_126[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0375976563f, 0.243164063f, 0.f),
    C_Vertex3f(0.0375976563f, 0.243164063f, -1.f),
    C_Vertex3f(0.0375976563f, 0.332519531f, -1.f),
    C_Vertex3f(0.0375976563f, 0.332519531f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0375976563f, 0.332519531f, 0.f),
    C_Vertex3f(0.0375976563f, 0.332519531f, -1.f),
    C_Normal3f(-0.704141676f, 0.710059524f, 0.f),
    C_Vertex3f(0.043390274f, 0.339111328f, 0.f),
    C_Vertex3f(0.043390274f, 0.339111328f, -1.f),
    C_Normal3f(-0.637717545f, 0.770270169f, 0.f),
    C_Vertex3f(0.0558774471f, 0.350646973f, 0.f),
    C_Vertex3f(0.0558774471f, 0.350646973f, -1.f),
    C_Normal3f(-0.547860622f, 0.836569726f, 0.f),
    C_Vertex3f(0.0692665577f, 0.360534668f, 0.f),
    C_Vertex3f(0.0692665577f, 0.360534668f, -1.f),
    C_Normal3f(-0.449382305f, 0.893339455f, 0.f),
    C_Vertex3f(0.0835576057f, 0.368774414f, 0.f),
    C_Vertex3f(0.0835576057f, 0.368774414f, -1.f),
    C_Normal3f(-0.345930338f, 0.938260198f, 0.f),
    C_Vertex3f(0.0987505913f, 0.375366211f, 0.f),
    C_Vertex3f(0.0987505913f, 0.375366211f, -1.f),
    C_Normal3f(-0.241619259f, 0.970371187f, 0.f),
    C_Vertex3f(0.114845514f, 0.380310059f, 0.f),
    C_Vertex3f(0.114845514f, 0.380310059f, -1.f),
    C_Normal3f(-0.140274361f, 0.990112662f, 0.f),
    C_Vertex3f(0.131842375f, 0.383605957f, 0.f),
    C_Vertex3f(0.131842375f, 0.383605957f, -1.f),
    C_Normal3f(-0.0486167856f, 0.998817563f, 0.f),
    C_Vertex3f(0.149741173f, 0.385253906f, 0.f),
    C_Vertex3f(0.149741173f, 0.385253906f, -1.f),
    C_Normal3f(0.0325165726f, 0.999471247f, 0.f),
    C_Vertex3f(0.165699005f, 0.385253906f, 0.f),
    C_Vertex3f(0.165699005f, 0.385253906f, -1.f),
    C_Normal3f(0.104858935f, 0.994487107f, 0.f),
    C_Vertex3f(0.179142475f, 0.384297371f, 0.f),
    C_Vertex3f(0.179142475f, 0.384297371f, -1.f),
    C_Normal3f(0.169119194f, 0.985595644f, 0.f),
    C_Vertex3f(0.192914486f, 0.3823843f, 0.f),
    C_Vertex3f(0.192914486f, 0.3823843f, -1.f),
    C_Normal3f(0.228486076f, 0.97354722f, 0.f),
    C_Vertex3f(0.207015038f, 0.379514694f, 0.f),
    C_Vertex3f(0.207015038f, 0.379514694f, -1.f),
    C_Normal3f(0.286308944f, 0.958137274f, 0.f),
    C_Vertex3f(0.22144413f, 0.375688553f, 0.f),
    C_Vertex3f(0.22144413f, 0.375688553f, -1.f),
    C_Normal3f(0.330246598f, 0.943894684f, 0.f),
    C_Vertex3f(0.238933086f, 0.369976997f, 0.f),
    C_Vertex3f(0.238933086f, 0.369976997f, -1.f),
    C_Normal3f(0.360739201f, 0.932666659f, 0.f),
    C_Vertex3f(0.259481907f, 0.362380028f, 0.f),
    C_Vertex3f(0.259481907f, 0.362380028f, -1.f),
    C_Normal3f(0.381089211f, 0.924538255f, 0.f),
    C_Vertex3f(0.283090591f, 0.352897644f, 0.f),
    C_Vertex3f(0.283090591f, 0.352897644f, -1.f),
    C_Normal3f(0.382386804f, 0.92400229f, 0.f),
    C_Vertex3f(0.303966522f, 0.344043732f, 0.f),
    C_Vertex3f(0.303966522f, 0.344043732f, -1.f),
    C_Normal3f(0.358124495f, 0.933673918f, 0.f),
    C_Vertex3f(0.317164421f, 0.338796616f, 0.f),
    C_Vertex3f(0.317164421f, 0.338796616f, -1.f),
    C_Normal3f(0.328591377f, 0.944472194f, 0.f),
    C_Vertex3f(0.328476906f, 0.33464241f, 0.f),
    C_Vertex3f(0.328476906f, 0.33464241f, -1.f),
    C_Normal3f(0.278498769f, 0.960436583f, 0.f),
    C_Vertex3f(0.337903976f, 0.331581116f, 0.f),
    C_Vertex3f(0.337903976f, 0.331581116f, -1.f),
    C_Normal3f(0.182532594f, 0.983199835f, 0.f),
    C_Vertex3f(0.349216461f, 0.32862854f, 0.f),
    C_Vertex3f(0.349216461f, 0.32862854f, -1.f),
    C_Normal3f(0.0655784607f, 0.997847438f, 0.f),
    C_Vertex3f(0.3644104f, 0.326660156f, 0.f),
    C_Vertex3f(0.3644104f, 0.326660156f, -1.f),
    C_Normal3f(-0.0729649737f, 0.99733448f, 0.f),
    C_Vertex3f(0.379167557f, 0.326660156f, 0.f),
    C_Vertex3f(0.379167557f, 0.326660156f, -1.f),
    C_Normal3f(-0.214771897f, 0.976664245f, 0.f),
    C_Vertex3f(0.393622875f, 0.32879734f, 0.f),
    C_Vertex3f(0.393622875f, 0.32879734f, -1.f),
    C_Normal3f(-0.339081109f, 0.940757096f, 0.f),
    C_Vertex3f(0.408323765f, 0.333071709f, 0.f),
    C_Vertex3f(0.408323765f, 0.333071709f, -1.f),
    C_Normal3f(-0.444621146f, 0.895718753f, 0.f),
    C_Vertex3f(0.423270226f, 0.339483261f, 0.f),
    C_Vertex3f(0.423270226f, 0.339483261f, -1.f),
    C_Normal3f(-0.524436355f, 0.851449668f, 0.f),
    C_Vertex3f(0.438462257f, 0.348031998f, 0.f),
    C_Vertex3f(0.438462257f, 0.348031998f, -1.f),
    C_Normal3f(-0.5907076f, 0.806885839f, 0.f),
    C_Vertex3f(0.452807903f, 0.357676506f, 0.f),
    C_Vertex3f(0.452807903f, 0.357676506f, -1.f),
    C_Normal3f(-0.653460622f, 0.756960452f, 0.f),
    C_Vertex3f(0.466307163f, 0.368416786f, 0.f),
    C_Vertex3f(0.466307163f, 0.368416786f, -1.f),
    C_Normal3f(-0.702214301f, 0.71196574f, 0.f),
    C_Vertex3f(0.478960037f, 0.380252838f, 0.f),
    C_Vertex3f(0.478960037f, 0.380252838f, -1.f),
    C_Normal3f(-0.738509059f, 0.674243569f, 0.f),
    C_Vertex3f(0.484863281f, 0.38671875f, 0.f),
    C_Vertex3f(0.484863281f, 0.38671875f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.484863281f, 0.38671875f, 0.f),
    C_Vertex3f(0.484863281f, 0.38671875f, -1.f),
    C_Vertex3f(0.484863281f, 0.293945313f, -1.f),
    C_Vertex3f(0.484863281f, 0.293945313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.683649898f, -0.729810178f, 0.f),
    C_Vertex3f(0.484863281f, 0.293945313f, 0.f),
    C_Vertex3f(0.484863281f, 0.293945313f, -1.f),
    C_Normal3f(0.646575868f, -0.762849748f, 0.f),
    C_Vertex3f(0.477867126f, 0.287391663f, 0.f),
    C_Vertex3f(0.477867126f, 0.287391663f, -1.f),
    C_Normal3f(0.593324244f, -0.804963529f, 0.f),
    C_Vertex3f(0.463793278f, 0.276086807f, 0.f),
    C_Vertex3f(0.463793278f, 0.276086807f, -1.f),
    C_Normal3f(0.518189669f, -0.855265737f, 0.f),
    C_Vertex3f(0.44963789f, 0.266584396f, 0.f),
    C_Vertex3f(0.44963789f, 0.266584396f, -1.f),
    C_Normal3f(0.429925382f, -0.902864397f, 0.f),
    C_Vertex3f(0.435400963f, 0.25888443f, 0.f),
    C_Vertex3f(0.435400963f, 0.25888443f, -1.f),
    C_Normal3f(0.329841971f, -0.944036186f, 0.f),
    C_Vertex3f(0.421082497f, 0.252986908f, 0.f),
    C_Vertex3f(0.421082497f, 0.252986908f, -1.f),
    C_Normal3f(0.228960082f, -0.97343576f, 0.f),
    C_Vertex3f(0.405862331f, 0.248563766f, 0.f),
    C_Vertex3f(0.405862331f, 0.248563766f, -1.f),
    C_Normal3f(0.132273927f, -0.991213143f, 0.f),
    C_Vertex3f(0.389740467f, 0.245615005f, 0.f),
    C_Vertex3f(0.389740467f, 0.245615005f, -1.f),
    C_Normal3f(0.037680082f, -0.99928987f, 0.f),
    C_Vertex3f(0.372716904f, 0.244140625f, 0.f),
    C_Vertex3f(0.372716904f, 0.244140625f, -1.f),
    C_Normal3f(-0.0648508519f, -0.997894943f, 0.f),
    C_Vertex3f(0.350639343f, 0.244140625f, 0.f),
    C_Vertex3f(0.350639343f, 0.244140625f, -1.f),
    C_Normal3f(-0.170031086f, -0.985438704f, 0.f),
    C_Vertex3f(0.325611115f, 0.24720192f, 0.f),
    C_Vertex3f(0.325611115f, 0.24720192f, -1.f),
    C_Normal3f(-0.261700034f, -0.965149283f, 0.f),
    C_Vertex3f(0.31071949f, 0.251028538f, 0.f),
    C_Vertex3f(0.31071949f, 0.251028538f, -1.f),
    C_Normal3f(-0.305256814f, -0.952270091f, 0.f),
    C_Vertex3f(0.303888321f, 0.253092051f, 0.f),
    C_Vertex3f(0.303888321f, 0.253092051f, -1.f),
    C_Normal3f(-0.331669092f, -0.943395853f, 0.f),
    C_Vertex3f(0.296182632f, 0.255688429f, 0.f),
    C_Vertex3f(0.296182632f, 0.255688429f, -1.f),
    C_Normal3f(-0.352408409f, -0.935846329f, 0.f),
    C_Vertex3f(0.287602425f, 0.258817673f, 0.f),
    C_Vertex3f(0.287602425f, 0.258817673f, -1.f),
    C_Normal3f(-0.369101226f, -0.929389179f, 0.f),
    C_Vertex3f(0.278147697f, 0.262479782f, 0.f),
    C_Vertex3f(0.278147697f, 0.262479782f, -1.f),
    C_Normal3f(-0.382813394f, -0.923825681f, 0.f),
    C_Vertex3f(0.267818451f, 0.266674757f, 0.f),
    C_Vertex3f(0.267818451f, 0.266674757f, -1.f),
    C_Normal3f(-0.394270241f, -0.918994546f, 0.f),
    C_Vertex3f(0.256614685f, 0.271402597f, 0.f),
    C_Vertex3f(0.256614685f, 0.271402597f, -1.f),
    C_Normal3f(-0.404871434f, -0.914373577f, 0.f),
    C_Vertex3f(0.2445364f, 0.276663303f, 0.f),
    C_Vertex3f(0.2445364f, 0.276663303f, -1.f),
    C_Normal3f(-0.373435289f, -0.927656293f, 0.f),
    C_Vertex3f(0.225107193f, 0.285353661f, 0.f),
    C_Vertex3f(0.225107193f, 0.285353661f, -1.f),
    C_Normal3f(-0.304339081f, -0.952563822f, 0.f),
    C_Vertex3f(0.201360703f, 0.294044018f, 0.f),
    C_Vertex3f(0.201360703f, 0.294044018f, -1.f),
    C_Normal3f(-0.207277626f, -0.978282154f, 0.f),
    C_Vertex3f(0.179773331f, 0.299837589f, 0.f),
    C_Vertex3f(0.179773331f, 0.299837589f, -1.f),
    C_Normal3f(-0.0823454633f, -0.996603787f, 0.f),
    C_Vertex3f(0.160345078f, 0.302734375f, 0.f),
    C_Vertex3f(0.160345078f, 0.302734375f, -1.f),
    C_Normal3f(0.0514798947f, -0.998674095f, 0.f),
    C_Vertex3f(0.144714355f, 0.302734375f, 0.f),
    C_Vertex3f(0.144714355f, 0.302734375f, -1.f),
    C_Normal3f(0.166504756f, -0.986040652f, 0.f),
    C_Vertex3f(0.131132603f, 0.301228523f, 0.f),
    C_Vertex3f(0.131132603f, 0.301228523f, -1.f),
    C_Normal3f(0.278832793f, -0.960339725f, 0.f),
    C_Vertex3f(0.117961407f, 0.29821682f, 0.f),
    C_Vertex3f(0.117961407f, 0.29821682f, -1.f),
    C_Normal3f(0.38705951f, -0.922054708f, 0.f),
    C_Vertex3f(0.105200768f, 0.293699265f, 0.f),
    C_Vertex3f(0.105200768f, 0.293699265f, -1.f),
    C_Normal3f(0.496189952f, -0.868213952f, 0.f),
    C_Vertex3f(0.0928506851f, 0.287675858f, 0.f),
    C_Vertex3f(0.0928506851f, 0.287675858f, -1.f),
    C_Normal3f(0.580653489f, -0.81415087f, 0.f),
    C_Vertex3f(0.0789971352f, 0.278723717f, 0.f),
    C_Vertex3f(0.0789971352f, 0.278723717f, -1.f),
    C_Normal3f(0.637960851f, -0.770068705f, 0.f),
    C_Vertex3f(0.0636401176f, 0.266842842f, 0.f),
    C_Vertex3f(0.0636401176f, 0.266842842f, -1.f),
    C_Normal3f(0.672732115f, -0.739886105f, 0.f),
    C_Vertex3f(0.0467796326f, 0.252033234f, 0.f),
    C_Vertex3f(0.0467796326f, 0.252033234f, -1.f),
    C_Normal3f(0.694749117f, -0.71925211f, 0.f),
    C_Vertex3f(0.0375976563f, 0.243164063f, 0.f),
    C_Vertex3f(0.0375976563f, 0.243164063f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.349216461f, 0.32862854f, 0.f),
    C_Vertex3f(0.350639343f, 0.244140625f, 0.f),
    C_Vertex3f(0.3644104f, 0.326660156f, 0.f),
    C_Vertex3f(0.372716904f, 0.244140625f, 0.f),
    C_Vertex3f(0.379167557f, 0.326660156f, 0.f),
    C_Vertex3f(0.389740467f, 0.245615005f, 0.f),
    C_Vertex3f(0.393622875f, 0.32879734f, 0.f),
    C_Vertex3f(0.405862331f, 0.248563766f, 0.f),
    C_Vertex3f(0.408323765f, 0.333071709f, 0.f),
    C_Vertex3f(0.421082497f, 0.252986908f, 0.f),
    C_Vertex3f(0.423270226f, 0.339483261f, 0.f),
    C_Vertex3f(0.435400963f, 0.25888443f, 0.f),
    C_Vertex3f(0.438462257f, 0.348031998f, 0.f),
    C_Vertex3f(0.44963789f, 0.266584396f, 0.f),
    C_Vertex3f(0.452807903f, 0.357676506f, 0.f),
    C_Vertex3f(0.463793278f, 0.276086807f, 0.f),
    C_Vertex3f(0.466307163f, 0.368416786f, 0.f),
    C_Vertex3f(0.477867126f, 0.287391663f, 0.f),
    C_Vertex3f(0.478960037f, 0.380252838f, 0.f),
    C_Vertex3f(0.484863281f, 0.293945313f, 0.f),
    C_Vertex3f(0.484863281f, 0.38671875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.296182632f, 0.255688429f, 0.f),
    C_Vertex3f(0.303888321f, 0.253092051f, 0.f),
    C_Vertex3f(0.303966522f, 0.344043732f, 0.f),
    C_Vertex3f(0.31071949f, 0.251028538f, 0.f),
    C_Vertex3f(0.317164421f, 0.338796616f, 0.f),
    C_Vertex3f(0.325611115f, 0.24720192f, 0.f),
    C_Vertex3f(0.328476906f, 0.33464241f, 0.f),
    C_Vertex3f(0.350639343f, 0.244140625f, 0.f),
    C_Vertex3f(0.337903976f, 0.331581116f, 0.f),
    C_Vertex3f(0.349216461f, 0.32862854f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.267818451f, 0.266674757f, 0.f),
    C_Vertex3f(0.278147697f, 0.262479782f, 0.f),
    C_Vertex3f(0.283090591f, 0.352897644f, 0.f),
    C_Vertex3f(0.287602425f, 0.258817673f, 0.f),
    C_Vertex3f(0.303966522f, 0.344043732f, 0.f),
    C_Vertex3f(0.296182632f, 0.255688429f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259481907f, 0.362380028f, 0.f),
    C_Vertex3f(0.238933086f, 0.369976997f, 0.f),
    C_Vertex3f(0.2445364f, 0.276663303f, 0.f),
    C_Vertex3f(0.256614685f, 0.271402597f, 0.f),
    C_Vertex3f(0.267818451f, 0.266674757f, 0.f),
    C_Vertex3f(0.283090591f, 0.352897644f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.225107193f, 0.285353661f, 0.f),
    C_Vertex3f(0.2445364f, 0.276663303f, 0.f),
    C_Vertex3f(0.238933086f, 0.369976997f, 0.f),
    C_Vertex3f(0.22144413f, 0.375688553f, 0.f),
    C_Vertex3f(0.207015038f, 0.379514694f, 0.f),
    C_Vertex3f(0.201360703f, 0.294044018f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.192914486f, 0.3823843f, 0.f),
    C_Vertex3f(0.179142475f, 0.384297371f, 0.f),
    C_Vertex3f(0.179773331f, 0.299837589f, 0.f),
    C_Vertex3f(0.201360703f, 0.294044018f, 0.f),
    C_Vertex3f(0.207015038f, 0.379514694f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.117961407f, 0.29821682f, 0.f),
    C_Vertex3f(0.131132603f, 0.301228523f, 0.f),
    C_Vertex3f(0.131842375f, 0.383605957f, 0.f),
    C_Vertex3f(0.144714355f, 0.302734375f, 0.f),
    C_Vertex3f(0.149741173f, 0.385253906f, 0.f),
    C_Vertex3f(0.160345078f, 0.302734375f, 0.f),
    C_Vertex3f(0.165699005f, 0.385253906f, 0.f),
    C_Vertex3f(0.179773331f, 0.299837589f, 0.f),
    C_Vertex3f(0.179142475f, 0.384297371f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.043390274f, 0.339111328f, 0.f),
    C_Vertex3f(0.0467796326f, 0.252033234f, 0.f),
    C_Vertex3f(0.0558774471f, 0.350646973f, 0.f),
    C_Vertex3f(0.0636401176f, 0.266842842f, 0.f),
    C_Vertex3f(0.0692665577f, 0.360534668f, 0.f),
    C_Vertex3f(0.0789971352f, 0.278723717f, 0.f),
    C_Vertex3f(0.0835576057f, 0.368774414f, 0.f),
    C_Vertex3f(0.0928506851f, 0.287675858f, 0.f),
    C_Vertex3f(0.0987505913f, 0.375366211f, 0.f),
    C_Vertex3f(0.105200768f, 0.293699265f, 0.f),
    C_Vertex3f(0.114845514f, 0.380310059f, 0.f),
    C_Vertex3f(0.117961407f, 0.29821682f, 0.f),
    C_Vertex3f(0.131842375f, 0.383605957f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0375976563f, 0.332519531f, 0.f),
    C_Vertex3f(0.0375976563f, 0.243164063f, 0.f),
    C_Vertex3f(0.0467796326f, 0.252033234f, 0.f),
    C_Vertex3f(0.043390274f, 0.339111328f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.349216461f, 0.32862854f, -1.f),
    C_Vertex3f(0.350639343f, 0.244140625f, -1.f),
    C_Vertex3f(0.3644104f, 0.326660156f, -1.f),
    C_Vertex3f(0.372716904f, 0.244140625f, -1.f),
    C_Vertex3f(0.379167557f, 0.326660156f, -1.f),
    C_Vertex3f(0.389740467f, 0.245615005f, -1.f),
    C_Vertex3f(0.393622875f, 0.32879734f, -1.f),
    C_Vertex3f(0.405862331f, 0.248563766f, -1.f),
    C_Vertex3f(0.408323765f, 0.333071709f, -1.f),
    C_Vertex3f(0.421082497f, 0.252986908f, -1.f),
    C_Vertex3f(0.423270226f, 0.339483261f, -1.f),
    C_Vertex3f(0.435400963f, 0.25888443f, -1.f),
    C_Vertex3f(0.438462257f, 0.348031998f, -1.f),
    C_Vertex3f(0.44963789f, 0.266584396f, -1.f),
    C_Vertex3f(0.452807903f, 0.357676506f, -1.f),
    C_Vertex3f(0.463793278f, 0.276086807f, -1.f),
    C_Vertex3f(0.466307163f, 0.368416786f, -1.f),
    C_Vertex3f(0.477867126f, 0.287391663f, -1.f),
    C_Vertex3f(0.478960037f, 0.380252838f, -1.f),
    C_Vertex3f(0.484863281f, 0.293945313f, -1.f),
    C_Vertex3f(0.484863281f, 0.38671875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.296182632f, 0.255688429f, -1.f),
    C_Vertex3f(0.303888321f, 0.253092051f, -1.f),
    C_Vertex3f(0.303966522f, 0.344043732f, -1.f),
    C_Vertex3f(0.31071949f, 0.251028538f, -1.f),
    C_Vertex3f(0.317164421f, 0.338796616f, -1.f),
    C_Vertex3f(0.325611115f, 0.24720192f, -1.f),
    C_Vertex3f(0.328476906f, 0.33464241f, -1.f),
    C_Vertex3f(0.350639343f, 0.244140625f, -1.f),
    C_Vertex3f(0.337903976f, 0.331581116f, -1.f),
    C_Vertex3f(0.349216461f, 0.32862854f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.267818451f, 0.266674757f, -1.f),
    C_Vertex3f(0.278147697f, 0.262479782f, -1.f),
    C_Vertex3f(0.283090591f, 0.352897644f, -1.f),
    C_Vertex3f(0.287602425f, 0.258817673f, -1.f),
    C_Vertex3f(0.303966522f, 0.344043732f, -1.f),
    C_Vertex3f(0.296182632f, 0.255688429f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259481907f, 0.362380028f, -1.f),
    C_Vertex3f(0.238933086f, 0.369976997f, -1.f),
    C_Vertex3f(0.2445364f, 0.276663303f, -1.f),
    C_Vertex3f(0.256614685f, 0.271402597f, -1.f),
    C_Vertex3f(0.267818451f, 0.266674757f, -1.f),
    C_Vertex3f(0.283090591f, 0.352897644f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.225107193f, 0.285353661f, -1.f),
    C_Vertex3f(0.2445364f, 0.276663303f, -1.f),
    C_Vertex3f(0.238933086f, 0.369976997f, -1.f),
    C_Vertex3f(0.22144413f, 0.375688553f, -1.f),
    C_Vertex3f(0.207015038f, 0.379514694f, -1.f),
    C_Vertex3f(0.201360703f, 0.294044018f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.192914486f, 0.3823843f, -1.f),
    C_Vertex3f(0.179142475f, 0.384297371f, -1.f),
    C_Vertex3f(0.179773331f, 0.299837589f, -1.f),
    C_Vertex3f(0.201360703f, 0.294044018f, -1.f),
    C_Vertex3f(0.207015038f, 0.379514694f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.117961407f, 0.29821682f, -1.f),
    C_Vertex3f(0.131132603f, 0.301228523f, -1.f),
    C_Vertex3f(0.131842375f, 0.383605957f, -1.f),
    C_Vertex3f(0.144714355f, 0.302734375f, -1.f),
    C_Vertex3f(0.149741173f, 0.385253906f, -1.f),
    C_Vertex3f(0.160345078f, 0.302734375f, -1.f),
    C_Vertex3f(0.165699005f, 0.385253906f, -1.f),
    C_Vertex3f(0.179773331f, 0.299837589f, -1.f),
    C_Vertex3f(0.179142475f, 0.384297371f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.043390274f, 0.339111328f, -1.f),
    C_Vertex3f(0.0467796326f, 0.252033234f, -1.f),
    C_Vertex3f(0.0558774471f, 0.350646973f, -1.f),
    C_Vertex3f(0.0636401176f, 0.266842842f, -1.f),
    C_Vertex3f(0.0692665577f, 0.360534668f, -1.f),
    C_Vertex3f(0.0789971352f, 0.278723717f, -1.f),
    C_Vertex3f(0.0835576057f, 0.368774414f, -1.f),
    C_Vertex3f(0.0928506851f, 0.287675858f, -1.f),
    C_Vertex3f(0.0987505913f, 0.375366211f, -1.f),
    C_Vertex3f(0.105200768f, 0.293699265f, -1.f),
    C_Vertex3f(0.114845514f, 0.380310059f, -1.f),
    C_Vertex3f(0.117961407f, 0.29821682f, -1.f),
    C_Vertex3f(0.131842375f, 0.383605957f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0375976563f, 0.332519531f, -1.f),
    C_Vertex3f(0.0375976563f, 0.243164063f, -1.f),
    C_Vertex3f(0.0467796326f, 0.252033234f, -1.f),
    C_Vertex3f(0.043390274f, 0.339111328f, -1.f),
    C_End(),
    C_Translatef(0.522460938f, 0.f, 0.f),
};
static const Call char_127[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_128[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0401875973f, -0.999192059f, 0.f),
    C_Vertex3f(0.353671074f, 0.579101563f, 0.f),
    C_Vertex3f(0.353671074f, 0.579101563f, -1.f),
    C_Normal3f(0.0312495157f, -0.999511659f, 0.f),
    C_Vertex3f(0.336018562f, 0.579101563f, 0.f),
    C_Vertex3f(0.336018562f, 0.579101563f, -1.f),
    C_Normal3f(0.095438011f, -0.995435357f, 0.f),
    C_Vertex3f(0.318394184f, 0.577998638f, 0.f),
    C_Vertex3f(0.318394184f, 0.577998638f, -1.f),
    C_Normal3f(0.164650634f, -0.986351967f, 0.f),
    C_Vertex3f(0.301507473f, 0.575792789f, 0.f),
    C_Vertex3f(0.301507473f, 0.575792789f, -1.f),
    C_Normal3f(0.23761867f, -0.971358538f, 0.f),
    C_Vertex3f(0.285358429f, 0.572484016f, 0.f),
    C_Vertex3f(0.285358429f, 0.572484016f, -1.f),
    C_Normal3f(0.313327342f, -0.949645162f, 0.f),
    C_Vertex3f(0.269947052f, 0.568072319f, 0.f),
    C_Vertex3f(0.269947052f, 0.568072319f, -1.f),
    C_Normal3f(0.390405238f, -0.92064321f, 0.f),
    C_Vertex3f(0.255273342f, 0.562557697f, 0.f),
    C_Vertex3f(0.255273342f, 0.562557697f, -1.f),
    C_Normal3f(0.467193693f, -0.884155035f, 0.f),
    C_Vertex3f(0.241337299f, 0.555940151f, 0.f),
    C_Vertex3f(0.241337299f, 0.555940151f, -1.f),
    C_Normal3f(0.542791247f, -0.839867651f, 0.f),
    C_Vertex3f(0.228138924f, 0.548219681f, 0.f),
    C_Vertex3f(0.228138924f, 0.548219681f, -1.f),
    C_Normal3f(0.625021338f, -0.780607641f, 0.f),
    C_Vertex3f(0.214475632f, 0.538579941f, 0.f),
    C_Vertex3f(0.214475632f, 0.538579941f, -1.f),
    C_Normal3f(0.70513171f, -0.709076345f, 0.f),
    C_Vertex3f(0.200320244f, 0.525945663f, 0.f),
    C_Vertex3f(0.200320244f, 0.525945663f, -1.f),
    C_Normal3f(0.772067189f, -0.635540903f, 0.f),
    C_Vertex3f(0.186875343f, 0.511133194f, 0.f),
    C_Vertex3f(0.186875343f, 0.511133194f, -1.f),
    C_Normal3f(0.82634151f, -0.563169241f, 0.f),
    C_Vertex3f(0.17414093f, 0.494142532f, 0.f),
    C_Vertex3f(0.17414093f, 0.494142532f, -1.f),
    C_Normal3f(0.871409774f, -0.490555882f, 0.f),
    C_Vertex3f(0.162008286f, 0.474645615f, 0.f),
    C_Vertex3f(0.162008286f, 0.474645615f, -1.f),
    C_Normal3f(0.914056182f, -0.405587584f, 0.f),
    C_Vertex3f(0.152280807f, 0.455310822f, 0.f),
    C_Vertex3f(0.152280807f, 0.455310822f, -1.f),
    C_Normal3f(0.952352703f, -0.304998964f, 0.f),
    C_Vertex3f(0.145067215f, 0.436466217f, 0.f),
    C_Vertex3f(0.145067215f, 0.436466217f, -1.f),
    C_Normal3f(0.978200912f, -0.207660854f, 0.f),
    C_Vertex3f(0.140367508f, 0.418111801f, 0.f),
    C_Vertex3f(0.140367508f, 0.418111801f, -1.f),
    C_Normal3f(0.992597342f, -0.121451691f, 0.f),
    C_Vertex3f(0.139274597f, 0.409179688f, 0.f),
    C_Vertex3f(0.139274597f, 0.409179688f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.139274597f, 0.409179688f, 0.f),
    C_Vertex3f(0.139274597f, 0.409179688f, -1.f),
    C_Vertex3f(0.4375f, 0.409179688f, -1.f),
    C_Vertex3f(0.4375f, 0.409179688f, 0.f),
    C_Normal3f(0.980624497f, -0.195896968f, 0.f),
    C_Vertex3f(0.4375f, 0.409179688f, 0.f),
    C_Vertex3f(0.4375f, 0.409179688f, -1.f),
    C_Vertex3f(0.425697327f, 0.350097656f, -1.f),
    C_Vertex3f(0.425697327f, 0.350097656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.425697327f, 0.350097656f, 0.f),
    C_Vertex3f(0.425697327f, 0.350097656f, -1.f),
    C_Vertex3f(0.132270813f, 0.350097656f, -1.f),
    C_Vertex3f(0.132270813f, 0.350097656f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.998875678f, -0.0474070869f, 0.f),
    C_Vertex3f(0.132270813f, 0.350097656f, 0.f),
    C_Vertex3f(0.132270813f, 0.350097656f, -1.f),
    C_Normal3f(0.999858618f, -0.0168142617f, 0.f),
    C_Vertex3f(0.131835938f, 0.340934753f, 0.f),
    C_Vertex3f(0.131835938f, 0.340934753f, -1.f),
    C_Normal3f(0.99999845f, 0.00180470839f, 0.f),
    C_Vertex3f(0.131835938f, 0.324237823f, 0.f),
    C_Vertex3f(0.131835938f, 0.324237823f, -1.f),
    C_Normal3f(0.999975801f, 0.0069499216f, 0.f),
    C_Vertex3f(0.131890297f, 0.310813904f, 0.f),
    C_Vertex3f(0.131890297f, 0.310813904f, -1.f),
    C_Normal3f(0.999867439f, 0.0162761156f, 0.f),
    C_Vertex3f(0.131999016f, 0.300773621f, 0.f),
    C_Vertex3f(0.131999016f, 0.300773621f, -1.f),
    C_Normal3f(0.999463379f, 0.0327561013f, 0.f),
    C_Vertex3f(0.132162094f, 0.294116974f, 0.f),
    C_Vertex3f(0.132162094f, 0.294116974f, -1.f),
    C_Normal3f(0.997800589f, 0.0662874505f, 0.f),
    C_Vertex3f(0.132270813f, 0.292480469f, 0.f),
    C_Vertex3f(0.132270813f, 0.292480469f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.132270813f, 0.292480469f, 0.f),
    C_Vertex3f(0.132270813f, 0.292480469f, -1.f),
    C_Vertex3f(0.4140625f, 0.292480469f, -1.f),
    C_Vertex3f(0.4140625f, 0.292480469f, 0.f),
    C_Normal3f(0.979215384f, -0.202823028f, 0.f),
    C_Vertex3f(0.4140625f, 0.292480469f, 0.f),
    C_Vertex3f(0.4140625f, 0.292480469f, -1.f),
    C_Vertex3f(0.401824951f, 0.233398438f, -1.f),
    C_Vertex3f(0.401824951f, 0.233398438f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.401824951f, 0.233398438f, 0.f),
    C_Vertex3f(0.401824951f, 0.233398438f, -1.f),
    C_Vertex3f(0.141464233f, 0.233398438f, -1.f),
    C_Vertex3f(0.141464233f, 0.233398438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.983427048f, 0.181304291f, 0.f),
    C_Vertex3f(0.141464233f, 0.233398438f, 0.f),
    C_Vertex3f(0.141464233f, 0.233398438f, -1.f),
    C_Normal3f(0.972587645f, 0.23253645f, 0.f),
    C_Vertex3f(0.143759727f, 0.220947266f, 0.f),
    C_Vertex3f(0.143759727f, 0.220947266f, -1.f),
    C_Normal3f(0.953562379f, 0.301195592f, 0.f),
    C_Vertex3f(0.149962902f, 0.197852612f, 0.f),
    C_Vertex3f(0.149962902f, 0.197852612f, -1.f),
    C_Normal3f(0.921002507f, 0.389556706f, 0.f),
    C_Vertex3f(0.157778263f, 0.176565647f, 0.f),
    C_Vertex3f(0.157778263f, 0.176565647f, -1.f),
    C_Normal3f(0.875874162f, 0.482539624f, 0.f),
    C_Vertex3f(0.167205811f, 0.157086372f, 0.f),
    C_Vertex3f(0.167205811f, 0.157086372f, -1.f),
    C_Normal3f(0.816743433f, 0.577000916f, 0.f),
    C_Vertex3f(0.178245544f, 0.139414787f, 0.f),
    C_Vertex3f(0.178245544f, 0.139414787f, -1.f),
    C_Normal3f(0.743443191f, 0.668799102f, 0.f),
    C_Vertex3f(0.190897465f, 0.123550892f, 0.f),
    C_Vertex3f(0.190897465f, 0.123550892f, -1.f),
    C_Normal3f(0.657539606f, 0.753420115f, 0.f),
    C_Vertex3f(0.205161572f, 0.109494686f, 0.f),
    C_Vertex3f(0.205161572f, 0.109494686f, -1.f),
    C_Normal3f(0.557465553f, 0.830200136f, 0.f),
    C_Vertex3f(0.221037865f, 0.09724617f, 0.f),
    C_Vertex3f(0.221037865f, 0.09724617f, -1.f),
    C_Normal3f(0.450723588f, 0.892663598f, 0.f),
    C_Vertex3f(0.242351532f, 0.0845222473f, 0.f),
    C_Vertex3f(0.242351532f, 0.0845222473f, -1.f),
    C_Normal3f(0.331469357f, 0.943465948f, 0.f),
    C_Vertex3f(0.268528938f, 0.073266983f, 0.f),
    C_Vertex3f(0.268528938f, 0.073266983f, -1.f),
    C_Normal3f(0.198853344f, 0.980029225f, 0.f),
    C_Vertex3f(0.295744896f, 0.0657634735f, 0.f),
    C_Vertex3f(0.295744896f, 0.0657634735f, -1.f),
    C_Normal3f(0.0704516023f, 0.997515202f, 0.f),
    C_Vertex3f(0.323999405f, 0.0620117188f, 0.f),
    C_Vertex3f(0.323999405f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0300309658f, 0.999548972f, 0.f),
    C_Vertex3f(0.348865509f, 0.0620117188f, 0.f),
    C_Vertex3f(0.348865509f, 0.0620117188f, -1.f),
    C_Normal3f(-0.105559111f, 0.994413078f, 0.f),
    C_Vertex3f(0.368184328f, 0.0633392334f, 0.f),
    C_Vertex3f(0.368184328f, 0.0633392334f, -1.f),
    C_Normal3f(-0.18491289f, 0.982754886f, 0.f),
    C_Vertex3f(0.386382818f, 0.0659942627f, 0.f),
    C_Vertex3f(0.386382818f, 0.0659942627f, -1.f),
    C_Normal3f(-0.270778805f, 0.962641656f, 0.f),
    C_Vertex3f(0.403460979f, 0.0699768066f, 0.f),
    C_Vertex3f(0.403460979f, 0.0699768066f, -1.f),
    C_Normal3f(-0.361701071f, 0.932294071f, 0.f),
    C_Vertex3f(0.419418812f, 0.0752868652f, 0.f),
    C_Vertex3f(0.419418812f, 0.0752868652f, -1.f),
    C_Normal3f(-0.455309987f, 0.890332878f, 0.f),
    C_Vertex3f(0.434256315f, 0.0819244385f, 0.f),
    C_Vertex3f(0.434256315f, 0.0819244385f, -1.f),
    C_Normal3f(-0.548414648f, 0.836206555f, 0.f),
    C_Vertex3f(0.44797349f, 0.0898895264f, 0.f),
    C_Vertex3f(0.44797349f, 0.0898895264f, -1.f),
    C_Normal3f(-0.622991681f, 0.78222847f, 0.f),
    C_Vertex3f(0.460570335f, 0.0991821289f, 0.f),
    C_Vertex3f(0.460570335f, 0.0991821289f, -1.f),
    C_Normal3f(-0.67919153f, 0.733961105f, 0.f),
    C_Vertex3f(0.466308594f, 0.104492188f, 0.f),
    C_Vertex3f(0.466308594f, 0.104492188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.466308594f, 0.104492188f, 0.f),
    C_Vertex3f(0.466308594f, 0.104492188f, -1.f),
    C_Vertex3f(0.466308594f, 0.0145721436f, -1.f),
    C_Vertex3f(0.466308594f, 0.0145721436f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.420310944f, -0.907380164f, 0.f),
    C_Vertex3f(0.466308594f, 0.0145721436f, 0.f),
    C_Vertex3f(0.466308594f, 0.0145721436f, -1.f),
    C_Normal3f(0.349543601f, -0.936920106f, 0.f),
    C_Vertex3f(0.452646255f, 0.00824356079f, 0.f),
    C_Vertex3f(0.452646255f, 0.00824356079f, -1.f),
    C_Normal3f(0.259369016f, -0.965778291f, 0.f),
    C_Vertex3f(0.423900604f, -0.00124931335f, 0.f),
    C_Vertex3f(0.423900604f, -0.00124931335f, -1.f),
    C_Normal3f(0.151935592f, -0.988390386f, 0.f),
    C_Vertex3f(0.393733978f, -0.00757789612f, 0.f),
    C_Vertex3f(0.393733978f, -0.00757789612f, -1.f),
    C_Normal3f(0.0522809587f, -0.998632491f, 0.f),
    C_Vertex3f(0.362146378f, -0.0107421875f, 0.f),
    C_Vertex3f(0.362146378f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0243614074f, -0.999703228f, 0.f),
    C_Vertex3f(0.333292007f, -0.0107421875f, 0.f),
    C_Vertex3f(0.333292007f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0825223923f, -0.996589303f, 0.f),
    C_Vertex3f(0.309592545f, -0.0094615221f, 0.f),
    C_Vertex3f(0.309592545f, -0.0094615221f, -1.f),
    C_Normal3f(-0.142752275f, -0.989758372f, 0.f),
    C_Vertex3f(0.286893785f, -0.00690019131f, 0.f),
    C_Vertex3f(0.286893785f, -0.00690019131f, -1.f),
    C_Normal3f(-0.206879005f, -0.978366554f, 0.f),
    C_Vertex3f(0.265195727f, -0.00305819511f, 0.f),
    C_Vertex3f(0.265195727f, -0.00305819511f, -1.f),
    C_Normal3f(-0.274387538f, -0.961619198f, 0.f),
    C_Vertex3f(0.244498372f, 0.00206446648f, 0.f),
    C_Vertex3f(0.244498372f, 0.00206446648f, -1.f),
    C_Normal3f(-0.344470859f, -0.938796997f, 0.f),
    C_Vertex3f(0.224801719f, 0.00846779346f, 0.f),
    C_Vertex3f(0.224801719f, 0.00846779346f, -1.f),
    C_Normal3f(-0.416021705f, -0.909354687f, 0.f),
    C_Vertex3f(0.206105769f, 0.0161517859f, 0.f),
    C_Vertex3f(0.206105769f, 0.0161517859f, -1.f),
    C_Normal3f(-0.487669587f, -0.873028278f, 0.f),
    C_Vertex3f(0.188410521f, 0.0251164436f, 0.f),
    C_Vertex3f(0.188410521f, 0.0251164436f, -1.f),
    C_Normal3f(-0.557872534f, -0.82992667f, 0.f),
    C_Vertex3f(0.171715975f, 0.0353617668f, 0.f),
    C_Vertex3f(0.171715975f, 0.0353617668f, -1.f),
    C_Normal3f(-0.625056088f, -0.780579805f, 0.f),
    C_Vertex3f(0.156022131f, 0.0468877554f, 0.f),
    C_Vertex3f(0.156022131f, 0.0468877554f, -1.f),
    C_Normal3f(-0.687776625f, -0.725922406f, 0.f),
    C_Vertex3f(0.14132899f, 0.0596944094f, 0.f),
    C_Vertex3f(0.14132899f, 0.0596944094f, -1.f),
    C_Normal3f(-0.744871438f, -0.667208076f, 0.f),
    C_Vertex3f(0.127636552f, 0.0737817287f, 0.f),
    C_Vertex3f(0.127636552f, 0.0737817287f, -1.f),
    C_Normal3f(-0.795561731f, -0.605872512f, 0.f),
    C_Vertex3f(0.114944816f, 0.0891497135f, 0.f),
    C_Vertex3f(0.114944816f, 0.0891497135f, -1.f),
    C_Normal3f(-0.839488328f, -0.543377697f, 0.f),
    C_Vertex3f(0.103253782f, 0.105798364f, 0.f),
    C_Vertex3f(0.103253782f, 0.105798364f, -1.f),
    C_Normal3f(-0.876680493f, -0.481073052f, 0.f),
    C_Vertex3f(0.0925634503f, 0.123727679f, 0.f),
    C_Vertex3f(0.0925634503f, 0.123727679f, -1.f),
    C_Normal3f(-0.906912267f, -0.421319425f, 0.f),
    C_Vertex3f(0.0828738213f, 0.14293766f, 0.f),
    C_Vertex3f(0.0828738213f, 0.14293766f, -1.f),
    C_Normal3f(-0.931832552f, -0.362888604f, 0.f),
    C_Vertex3f(0.0750312805f, 0.161466599f, 0.f),
    C_Vertex3f(0.0750312805f, 0.161466599f, -1.f),
    C_Normal3f(-0.950686812f, -0.310152471f, 0.f),
    C_Vertex3f(0.0685272217f, 0.179777145f, 0.f),
    C_Vertex3f(0.0685272217f, 0.179777145f, -1.f),
    C_Normal3f(-0.964071035f, -0.26564458f, 0.f),
    C_Vertex3f(0.0625152588f, 0.199831009f, 0.f),
    C_Vertex3f(0.0625152588f, 0.199831009f, -1.f),
    C_Normal3f(-0.972534478f, -0.232758418f, 0.f),
    C_Vertex3f(0.0569953918f, 0.221628189f, 0.f),
    C_Vertex3f(0.0569953918f, 0.221628189f, -1.f),
    C_Normal3f(-0.977943718f, -0.208868876f, 0.f),
    C_Vertex3f(0.0544815063f, 0.233398438f, 0.f),
    C_Vertex3f(0.0544815063f, 0.233398438f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0544815063f, 0.233398438f, 0.f),
    C_Vertex3f(0.0544815063f, 0.233398438f, -1.f),
    C_Vertex3f(-0.0122070313f, 0.233398438f, -1.f),
    C_Vertex3f(-0.0122070313f, 0.233398438f, 0.f),
    C_Normal3f(-0.979340851f, 0.202216685f, 0.f),
    C_Vertex3f(-0.0122070313f, 0.233398438f, 0.f),
    C_Vertex3f(-0.0122070313f, 0.233398438f, -1.f),
    C_Vertex3f(-7.62939453e-006f, 0.292480469f, -1.f),
    C_Vertex3f(-7.62939453e-006f, 0.292480469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(-7.62939453e-006f, 0.292480469f, 0.f),
    C_Vertex3f(-7.62939453e-006f, 0.292480469f, -1.f),
    C_Vertex3f(0.0457458496f, 0.292480469f, -1.f),
    C_Vertex3f(0.0457458496f, 0.292480469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.997448206f, -0.0713945478f, 0.f),
    C_Vertex3f(0.0457458496f, 0.292480469f, 0.f),
    C_Vertex3f(0.0457458496f, 0.292480469f, -1.f),
    C_Normal3f(-0.998979628f, -0.0451634899f, 0.f),
    C_Vertex3f(0.0450897217f, 0.301647186f, 0.f),
    C_Vertex3f(0.0450897217f, 0.301647186f, -1.f),
    C_Normal3f(-0.999860525f, -0.0166983213f, 0.f),
    C_Vertex3f(0.0444335938f, 0.3215065f, 0.f),
    C_Vertex3f(0.0444335938f, 0.3215065f, -1.f),
    C_Normal3f(-0.999884367f, 0.0152083812f, 0.f),
    C_Vertex3f(0.0444335938f, 0.340934753f, 0.f),
    C_Vertex3f(0.0444335938f, 0.340934753f, -1.f),
    C_Normal3f(-0.998875678f, 0.0474070869f, 0.f),
    C_Vertex3f(0.0448684692f, 0.350097656f, 0.f),
    C_Vertex3f(0.0448684692f, 0.350097656f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0448684692f, 0.350097656f, 0.f),
    C_Vertex3f(0.0448684692f, 0.350097656f, -1.f),
    C_Vertex3f(-0.0122070313f, 0.350097656f, -1.f),
    C_Vertex3f(-0.0122070313f, 0.350097656f, 0.f),
    C_Normal3f(-0.979340851f, 0.202216685f, 0.f),
    C_Vertex3f(-0.0122070313f, 0.350097656f, 0.f),
    C_Vertex3f(-0.0122070313f, 0.350097656f, -1.f),
    C_Vertex3f(-762939453e-006f, 0.409179688f, -1.f),
    C_Vertex3f(-762939453e-006f, 0.409179688f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(-762939453e-006f, 0.409179688f, 0.f),
    C_Vertex3f(-762939453e-006f, 0.409179688f, -1.f),
    C_Vertex3f(0.0505523682f, 0.409179688f, -1.f),
    C_Vertex3f(0.0505523682f, 0.409179688f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.982332587f, 0.187143579f, 0.f),
    C_Vertex3f(0.0505523682f, 0.409179688f, 0.f),
    C_Vertex3f(0.0505523682f, 0.409179688f, -1.f),
    C_Normal3f(-0.9693681f, 0.245612279f, 0.f),
    C_Vertex3f(0.053940773f, 0.426965714f, 0.f),
    C_Vertex3f(0.053940773f, 0.426965714f, -1.f),
    C_Normal3f(-0.946410537f, 0.322966129f, 0.f),
    C_Vertex3f(0.0634360313f, 0.460028172f, 0.f),
    C_Vertex3f(0.0634360313f, 0.460028172f, -1.f),
    C_Normal3f(-0.907360554f, 0.420353413f, 0.f),
    C_Vertex3f(0.0756497383f, 0.490581036f, 0.f),
    C_Vertex3f(0.0756497383f, 0.490581036f, -1.f),
    C_Normal3f(-0.85440594f, 0.519605994f, 0.f),
    C_Vertex3f(0.0905818939f, 0.518624306f, 0.f),
    C_Vertex3f(0.0905818939f, 0.518624306f, -1.f),
    C_Normal3f(-0.787363887f, 0.616488576f, 0.f),
    C_Vertex3f(0.108232498f, 0.544157982f, 0.f),
    C_Vertex3f(0.108232498f, 0.544157982f, -1.f),
    C_Normal3f(-0.707773089f, 0.706439793f, 0.f),
    C_Vertex3f(0.128601551f, 0.567182064f, 0.f),
    C_Vertex3f(0.128601551f, 0.567182064f, -1.f),
    C_Normal3f(-0.618845105f, 0.785512984f, 0.f),
    C_Vertex3f(0.151689053f, 0.587696552f, 0.f),
    C_Vertex3f(0.151689053f, 0.587696552f, -1.f),
    C_Normal3f(-0.530143976f, 0.847907662f, 0.f),
    C_Vertex3f(0.177495003f, 0.605701447f, 0.f),
    C_Vertex3f(0.177495003f, 0.605701447f, -1.f),
    C_Normal3f(-0.454058617f, 0.890971839f, 0.f),
    C_Vertex3f(0.200555801f, 0.618249893f, 0.f),
    C_Vertex3f(0.200555801f, 0.618249893f, -1.f),
    C_Normal3f(-0.392641574f, 0.919691563f, 0.f),
    C_Vertex3f(0.218603611f, 0.626651287f, 0.f),
    C_Vertex3f(0.218603611f, 0.626651287f, -1.f),
    C_Normal3f(-0.33249408f, 0.94310534f, 0.f),
    C_Vertex3f(0.237102032f, 0.633852482f, 0.f),
    C_Vertex3f(0.237102032f, 0.633852482f, -1.f),
    C_Normal3f(-0.271122992f, 0.96254468f, 0.f),
    C_Vertex3f(0.256051064f, 0.639853477f, 0.f),
    C_Vertex3f(0.256051064f, 0.639853477f, -1.f),
    C_Normal3f(-0.209307596f, 0.977849841f, 0.f),
    C_Vertex3f(0.275450706f, 0.644654274f, 0.f),
    C_Vertex3f(0.275450706f, 0.644654274f, -1.f),
    C_Normal3f(-0.147818342f, 0.989014566f, 0.f),
    C_Vertex3f(0.295300961f, 0.648254871f, 0.f),
    C_Vertex3f(0.295300961f, 0.648254871f, -1.f),
    C_Normal3f(-0.0873720646f, 0.996175766f, 0.f),
    C_Vertex3f(0.315601826f, 0.65065527f, 0.f),
    C_Vertex3f(0.315601826f, 0.65065527f, -1.f),
    C_Normal3f(-0.0288966987f, 0.99958241f, 0.f),
    C_Vertex3f(0.336353302f, 0.651855469f, 0.f),
    C_Vertex3f(0.336353302f, 0.651855469f, -1.f),
    C_Normal3f(0.0274486877f, 0.999623239f, 0.f),
    C_Vertex3f(0.357118607f, 0.651855469f, 0.f),
    C_Vertex3f(0.357118607f, 0.651855469f, -1.f),
    C_Normal3f(0.0864661038f, 0.996254742f, 0.f),
    C_Vertex3f(0.376641035f, 0.650749207f, 0.f),
    C_Vertex3f(0.376641035f, 0.650749207f, -1.f),
    C_Normal3f(0.149325833f, 0.988788009f, 0.f),
    C_Vertex3f(0.39535737f, 0.648536682f, 0.f),
    C_Vertex3f(0.39535737f, 0.648536682f, -1.f),
    C_Normal3f(0.215943396f, 0.976405859f, 0.f),
    C_Vertex3f(0.413267612f, 0.645217896f, 0.f),
    C_Vertex3f(0.413267612f, 0.645217896f, -1.f),
    C_Normal3f(0.285654932f, 0.958332598f, 0.f),
    C_Vertex3f(0.430371761f, 0.640792847f, 0.f),
    C_Vertex3f(0.430371761f, 0.640792847f, -1.f),
    C_Normal3f(0.357493132f, 0.933915794f, 0.f),
    C_Vertex3f(0.446669817f, 0.635261536f, 0.f),
    C_Vertex3f(0.446669817f, 0.635261536f, -1.f),
    C_Normal3f(0.430202067f, 0.902732551f, 0.f),
    C_Vertex3f(0.462161779f, 0.628623962f, 0.f),
    C_Vertex3f(0.462161779f, 0.628623962f, -1.f),
    C_Normal3f(0.490395814f, 0.871499836f, 0.f),
    C_Vertex3f(0.476847649f, 0.620880127f, 0.f),
    C_Vertex3f(0.476847649f, 0.620880127f, -1.f),
    C_Normal3f(0.537632465f, 0.843179226f, 0.f),
    C_Vertex3f(0.483787537f, 0.616455078f, 0.f),
    C_Vertex3f(0.483787537f, 0.616455078f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.977659881f, -0.210193202f, 0.f),
    C_Vertex3f(0.483787537f, 0.616455078f, 0.f),
    C_Vertex3f(0.483787537f, 0.616455078f, -1.f),
    C_Vertex3f(0.466308594f, 0.53515625f, -1.f),
    C_Vertex3f(0.466308594f, 0.53515625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.635885239f, -0.77178365f, 0.f),
    C_Vertex3f(0.466308594f, 0.53515625f, 0.f),
    C_Vertex3f(0.466308594f, 0.53515625f, -1.f),
    C_Normal3f(-0.594114721f, -0.804380298f, 0.f),
    C_Vertex3f(0.459641457f, 0.540649414f, 0.f),
    C_Vertex3f(0.459641457f, 0.540649414f, -1.f),
    C_Normal3f(-0.537319362f, -0.843378961f, 0.f),
    C_Vertex3f(0.445856094f, 0.550262451f, 0.f),
    C_Vertex3f(0.445856094f, 0.550262451f, -1.f),
    C_Normal3f(-0.462937534f, -0.886390924f, 0.f),
    C_Vertex3f(0.431619644f, 0.558502197f, 0.f),
    C_Vertex3f(0.431619644f, 0.558502197f, -1.f),
    C_Normal3f(-0.382821351f, -0.923822403f, 0.f),
    C_Vertex3f(0.416932106f, 0.565368652f, 0.f),
    C_Vertex3f(0.416932106f, 0.565368652f, -1.f),
    C_Normal3f(-0.298570096f, -0.954387665f, 0.f),
    C_Vertex3f(0.40179348f, 0.570861816f, 0.f),
    C_Vertex3f(0.40179348f, 0.570861816f, -1.f),
    C_Normal3f(-0.212142169f, -0.977238834f, 0.f),
    C_Vertex3f(0.386203766f, 0.574981689f, 0.f),
    C_Vertex3f(0.386203766f, 0.574981689f, -1.f),
    C_Normal3f(-0.125634536f, -0.992076576f, 0.f),
    C_Vertex3f(0.370162964f, 0.577728271f, 0.f),
    C_Vertex3f(0.370162964f, 0.577728271f, -1.f),
    C_Normal3f(-0.0401875973f, -0.999192059f, 0.f),
    C_Vertex3f(0.353671074f, 0.579101563f, 0.f),
    C_Vertex3f(0.353671074f, 0.579101563f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0505523682f, 0.409179688f, 0.f),
    C_Vertex3f(-7.62939453e-006f, 0.409179688f, 0.f),
    C_Vertex3f(0.0448684692f, 0.350097656f, 0.f),
    C_Vertex3f(0.0450897217f, 0.301647186f, 0.f),
    C_Vertex3f(0.0457458496f, 0.292480469f, 0.f),
    C_Vertex3f(0.0544815063f, 0.233398438f, 0.f),
    C_Vertex3f(0.053940773f, 0.426965714f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.401824951f, 0.233398438f, 0.f),
    C_Vertex3f(0.4140625f, 0.292480469f, 0.f),
    C_Vertex3f(0.141464233f, 0.233398438f, 0.f),
    C_Vertex3f(0.132270813f, 0.292480469f, 0.f),
    C_Vertex3f(0.14132899f, 0.0596944094f, 0.f),
    C_Vertex3f(0.132162094f, 0.294116974f, 0.f),
    C_Vertex3f(0.127636552f, 0.0737817287f, 0.f),
    C_Vertex3f(0.131999016f, 0.300773621f, 0.f),
    C_Vertex3f(0.128601551f, 0.567182064f, 0.f),
    C_Vertex3f(0.131890297f, 0.310813904f, 0.f),
    C_Vertex3f(0.131835938f, 0.324237823f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0925634503f, 0.123727679f, 0.f),
    C_Vertex3f(0.103253782f, 0.105798364f, 0.f),
    C_Vertex3f(0.108232498f, 0.544157982f, 0.f),
    C_Vertex3f(0.114944816f, 0.0891497135f, 0.f),
    C_Vertex3f(0.128601551f, 0.567182064f, 0.f),
    C_Vertex3f(0.127636552f, 0.0737817287f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0685272217f, 0.179777145f, 0.f),
    C_Vertex3f(0.0750312805f, 0.161466599f, 0.f),
    C_Vertex3f(0.0756497383f, 0.490581036f, 0.f),
    C_Vertex3f(0.0828738213f, 0.14293766f, 0.f),
    C_Vertex3f(0.0905818939f, 0.518624306f, 0.f),
    C_Vertex3f(0.0925634503f, 0.123727679f, 0.f),
    C_Vertex3f(0.108232498f, 0.544157982f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0634360313f, 0.460028172f, 0.f),
    C_Vertex3f(0.053940773f, 0.426965714f, 0.f),
    C_Vertex3f(0.0544815063f, 0.233398438f, 0.f),
    C_Vertex3f(0.0569953918f, 0.221628189f, 0.f),
    C_Vertex3f(0.0625152588f, 0.199831009f, 0.f),
    C_Vertex3f(0.0685272217f, 0.179777145f, 0.f),
    C_Vertex3f(0.0756497383f, 0.490581036f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-7.62939453e-006f, 0.292480469f, 0.f),
    C_Vertex3f(-0.0122070313f, 0.233398438f, 0.f),
    C_Vertex3f(0.0544815063f, 0.233398438f, 0.f),
    C_Vertex3f(0.0457458496f, 0.292480469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460570335f, 0.0991821289f, 0.f),
    C_Vertex3f(0.44797349f, 0.0898895264f, 0.f),
    C_Vertex3f(0.452646255f, 0.00824356079f, 0.f),
    C_Vertex3f(0.466308594f, 0.0145721436f, 0.f),
    C_Vertex3f(0.466308594f, 0.104492188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434256315f, 0.0819244385f, 0.f),
    C_Vertex3f(0.419418812f, 0.0752868652f, 0.f),
    C_Vertex3f(0.423900604f, -0.00124931335f, 0.f),
    C_Vertex3f(0.452646255f, 0.00824356079f, 0.f),
    C_Vertex3f(0.44797349f, 0.0898895264f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.403460979f, 0.0699768066f, 0.f),
    C_Vertex3f(0.386382818f, 0.0659942627f, 0.f),
    C_Vertex3f(0.393733978f, -0.00757789612f, 0.f),
    C_Vertex3f(0.423900604f, -0.00124931335f, 0.f),
    C_Vertex3f(0.419418812f, 0.0752868652f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.244498372f, 0.00206446648f, 0.f),
    C_Vertex3f(0.265195727f, -0.00305819511f, 0.f),
    C_Vertex3f(0.268528938f, 0.073266983f, 0.f),
    C_Vertex3f(0.286893785f, -0.00690019131f, 0.f),
    C_Vertex3f(0.295744896f, 0.0657634735f, 0.f),
    C_Vertex3f(0.309592545f, -0.0094615221f, 0.f),
    C_Vertex3f(0.323999405f, 0.0620117188f, 0.f),
    C_Vertex3f(0.333292007f, -0.0107421875f, 0.f),
    C_Vertex3f(0.348865509f, 0.0620117188f, 0.f),
    C_Vertex3f(0.362146378f, -0.0107421875f, 0.f),
    C_Vertex3f(0.368184328f, 0.0633392334f, 0.f),
    C_Vertex3f(0.393733978f, -0.00757789612f, 0.f),
    C_Vertex3f(0.386382818f, 0.0659942627f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.205161572f, 0.109494686f, 0.f),
    C_Vertex3f(0.206105769f, 0.0161517859f, 0.f),
    C_Vertex3f(0.221037865f, 0.09724617f, 0.f),
    C_Vertex3f(0.224801719f, 0.00846779346f, 0.f),
    C_Vertex3f(0.242351532f, 0.0845222473f, 0.f),
    C_Vertex3f(0.244498372f, 0.00206446648f, 0.f),
    C_Vertex3f(0.268528938f, 0.073266983f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.167205811f, 0.157086372f, 0.f),
    C_Vertex3f(0.171715975f, 0.0353617668f, 0.f),
    C_Vertex3f(0.178245544f, 0.139414787f, 0.f),
    C_Vertex3f(0.188410521f, 0.0251164436f, 0.f),
    C_Vertex3f(0.190897465f, 0.123550892f, 0.f),
    C_Vertex3f(0.206105769f, 0.0161517859f, 0.f),
    C_Vertex3f(0.205161572f, 0.109494686f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.157778263f, 0.176565647f, 0.f),
    C_Vertex3f(0.149962902f, 0.197852612f, 0.f),
    C_Vertex3f(0.156022131f, 0.0468877554f, 0.f),
    C_Vertex3f(0.171715975f, 0.0353617668f, 0.f),
    C_Vertex3f(0.167205811f, 0.157086372f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156022131f, 0.0468877554f, 0.f),
    C_Vertex3f(0.149962902f, 0.197852612f, 0.f),
    C_Vertex3f(0.143759727f, 0.220947266f, 0.f),
    C_Vertex3f(0.141464233f, 0.233398438f, 0.f),
    C_Vertex3f(0.14132899f, 0.0596944094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431619644f, 0.558502197f, 0.f),
    C_Vertex3f(0.445856094f, 0.550262451f, 0.f),
    C_Vertex3f(0.446669817f, 0.635261536f, 0.f),
    C_Vertex3f(0.459641457f, 0.540649414f, 0.f),
    C_Vertex3f(0.462161779f, 0.628623962f, 0.f),
    C_Vertex3f(0.466308594f, 0.53515625f, 0.f),
    C_Vertex3f(0.476847649f, 0.620880127f, 0.f),
    C_Vertex3f(0.483787537f, 0.616455078f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318394184f, 0.577998638f, 0.f),
    C_Vertex3f(0.336018562f, 0.579101563f, 0.f),
    C_Vertex3f(0.336353302f, 0.651855469f, 0.f),
    C_Vertex3f(0.353671074f, 0.579101563f, 0.f),
    C_Vertex3f(0.357118607f, 0.651855469f, 0.f),
    C_Vertex3f(0.370162964f, 0.577728271f, 0.f),
    C_Vertex3f(0.376641035f, 0.650749207f, 0.f),
    C_Vertex3f(0.386203766f, 0.574981689f, 0.f),
    C_Vertex3f(0.39535737f, 0.648536682f, 0.f),
    C_Vertex3f(0.40179348f, 0.570861816f, 0.f),
    C_Vertex3f(0.413267612f, 0.645217896f, 0.f),
    C_Vertex3f(0.416932106f, 0.565368652f, 0.f),
    C_Vertex3f(0.430371761f, 0.640792847f, 0.f),
    C_Vertex3f(0.431619644f, 0.558502197f, 0.f),
    C_Vertex3f(0.446669817f, 0.635261536f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241337299f, 0.555940151f, 0.f),
    C_Vertex3f(0.255273342f, 0.562557697f, 0.f),
    C_Vertex3f(0.256051064f, 0.639853477f, 0.f),
    C_Vertex3f(0.269947052f, 0.568072319f, 0.f),
    C_Vertex3f(0.275450706f, 0.644654274f, 0.f),
    C_Vertex3f(0.285358429f, 0.572484016f, 0.f),
    C_Vertex3f(0.295300961f, 0.648254871f, 0.f),
    C_Vertex3f(0.301507473f, 0.575792789f, 0.f),
    C_Vertex3f(0.315601826f, 0.65065527f, 0.f),
    C_Vertex3f(0.318394184f, 0.577998638f, 0.f),
    C_Vertex3f(0.336353302f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186875343f, 0.511133194f, 0.f),
    C_Vertex3f(0.200320244f, 0.525945663f, 0.f),
    C_Vertex3f(0.200555801f, 0.618249893f, 0.f),
    C_Vertex3f(0.214475632f, 0.538579941f, 0.f),
    C_Vertex3f(0.218603611f, 0.626651287f, 0.f),
    C_Vertex3f(0.228138924f, 0.548219681f, 0.f),
    C_Vertex3f(0.237102032f, 0.633852482f, 0.f),
    C_Vertex3f(0.241337299f, 0.555940151f, 0.f),
    C_Vertex3f(0.256051064f, 0.639853477f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.177495003f, 0.605701447f, 0.f),
    C_Vertex3f(0.151689053f, 0.587696552f, 0.f),
    C_Vertex3f(0.152280807f, 0.455310822f, 0.f),
    C_Vertex3f(0.162008286f, 0.474645615f, 0.f),
    C_Vertex3f(0.17414093f, 0.494142532f, 0.f),
    C_Vertex3f(0.186875343f, 0.511133194f, 0.f),
    C_Vertex3f(0.200555801f, 0.618249893f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.151689053f, 0.587696552f, 0.f),
    C_Vertex3f(0.128601551f, 0.567182064f, 0.f),
    C_Vertex3f(0.132270813f, 0.350097656f, 0.f),
    C_Vertex3f(0.139274597f, 0.409179688f, 0.f),
    C_Vertex3f(0.140367508f, 0.418111801f, 0.f),
    C_Vertex3f(0.145067215f, 0.436466217f, 0.f),
    C_Vertex3f(0.152280807f, 0.455310822f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.131835938f, 0.340934753f, 0.f),
    C_Vertex3f(0.132270813f, 0.350097656f, 0.f),
    C_Vertex3f(0.128601551f, 0.567182064f, 0.f),
    C_Vertex3f(0.131835938f, 0.324237823f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.139274597f, 0.409179688f, 0.f),
    C_Vertex3f(0.132270813f, 0.350097656f, 0.f),
    C_Vertex3f(0.425697327f, 0.350097656f, 0.f),
    C_Vertex3f(0.4375f, 0.409179688f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0444335938f, 0.340934753f, 0.f),
    C_Vertex3f(0.0444335938f, 0.3215065f, 0.f),
    C_Vertex3f(0.0450897217f, 0.301647186f, 0.f),
    C_Vertex3f(0.0448684692f, 0.350097656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0448684692f, 0.350097656f, 0.f),
    C_Vertex3f(-7.62939453e-006f, 0.409179688f, 0.f),
    C_Vertex3f(-0.0122070313f, 0.350097656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0505523682f, 0.409179688f, -1.f),
    C_Vertex3f(-7.62939453e-006f, 0.409179688f, -1.f),
    C_Vertex3f(0.0448684692f, 0.350097656f, -1.f),
    C_Vertex3f(0.0450897217f, 0.301647186f, -1.f),
    C_Vertex3f(0.0457458496f, 0.292480469f, -1.f),
    C_Vertex3f(0.0544815063f, 0.233398438f, -1.f),
    C_Vertex3f(0.053940773f, 0.426965714f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.401824951f, 0.233398438f, -1.f),
    C_Vertex3f(0.4140625f, 0.292480469f, -1.f),
    C_Vertex3f(0.141464233f, 0.233398438f, -1.f),
    C_Vertex3f(0.132270813f, 0.292480469f, -1.f),
    C_Vertex3f(0.14132899f, 0.0596944094f, -1.f),
    C_Vertex3f(0.132162094f, 0.294116974f, -1.f),
    C_Vertex3f(0.127636552f, 0.0737817287f, -1.f),
    C_Vertex3f(0.131999016f, 0.300773621f, -1.f),
    C_Vertex3f(0.128601551f, 0.567182064f, -1.f),
    C_Vertex3f(0.131890297f, 0.310813904f, -1.f),
    C_Vertex3f(0.131835938f, 0.324237823f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0925634503f, 0.123727679f, -1.f),
    C_Vertex3f(0.103253782f, 0.105798364f, -1.f),
    C_Vertex3f(0.108232498f, 0.544157982f, -1.f),
    C_Vertex3f(0.114944816f, 0.0891497135f, -1.f),
    C_Vertex3f(0.128601551f, 0.567182064f, -1.f),
    C_Vertex3f(0.127636552f, 0.0737817287f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0685272217f, 0.179777145f, -1.f),
    C_Vertex3f(0.0750312805f, 0.161466599f, -1.f),
    C_Vertex3f(0.0756497383f, 0.490581036f, -1.f),
    C_Vertex3f(0.0828738213f, 0.14293766f, -1.f),
    C_Vertex3f(0.0905818939f, 0.518624306f, -1.f),
    C_Vertex3f(0.0925634503f, 0.123727679f, -1.f),
    C_Vertex3f(0.108232498f, 0.544157982f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0634360313f, 0.460028172f, -1.f),
    C_Vertex3f(0.053940773f, 0.426965714f, -1.f),
    C_Vertex3f(0.0544815063f, 0.233398438f, -1.f),
    C_Vertex3f(0.0569953918f, 0.221628189f, -1.f),
    C_Vertex3f(0.0625152588f, 0.199831009f, -1.f),
    C_Vertex3f(0.0685272217f, 0.179777145f, -1.f),
    C_Vertex3f(0.0756497383f, 0.490581036f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-7.62939453e-006f, 0.292480469f, -1.f),
    C_Vertex3f(-0.0122070313f, 0.233398438f, -1.f),
    C_Vertex3f(0.0544815063f, 0.233398438f, -1.f),
    C_Vertex3f(0.0457458496f, 0.292480469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460570335f, 0.0991821289f, -1.f),
    C_Vertex3f(0.44797349f, 0.0898895264f, -1.f),
    C_Vertex3f(0.452646255f, 0.00824356079f, -1.f),
    C_Vertex3f(0.466308594f, 0.0145721436f, -1.f),
    C_Vertex3f(0.466308594f, 0.104492188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434256315f, 0.0819244385f, -1.f),
    C_Vertex3f(0.419418812f, 0.0752868652f, -1.f),
    C_Vertex3f(0.423900604f, -0.00124931335f, -1.f),
    C_Vertex3f(0.452646255f, 0.00824356079f, -1.f),
    C_Vertex3f(0.44797349f, 0.0898895264f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.403460979f, 0.0699768066f, -1.f),
    C_Vertex3f(0.386382818f, 0.0659942627f, -1.f),
    C_Vertex3f(0.393733978f, -0.00757789612f, -1.f),
    C_Vertex3f(0.423900604f, -0.00124931335f, -1.f),
    C_Vertex3f(0.419418812f, 0.0752868652f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.244498372f, 0.00206446648f, -1.f),
    C_Vertex3f(0.265195727f, -0.00305819511f, -1.f),
    C_Vertex3f(0.268528938f, 0.073266983f, -1.f),
    C_Vertex3f(0.286893785f, -0.00690019131f, -1.f),
    C_Vertex3f(0.295744896f, 0.0657634735f, -1.f),
    C_Vertex3f(0.309592545f, -0.0094615221f, -1.f),
    C_Vertex3f(0.323999405f, 0.0620117188f, -1.f),
    C_Vertex3f(0.333292007f, -0.0107421875f, -1.f),
    C_Vertex3f(0.348865509f, 0.0620117188f, -1.f),
    C_Vertex3f(0.362146378f, -0.0107421875f, -1.f),
    C_Vertex3f(0.368184328f, 0.0633392334f, -1.f),
    C_Vertex3f(0.393733978f, -0.00757789612f, -1.f),
    C_Vertex3f(0.386382818f, 0.0659942627f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.205161572f, 0.109494686f, -1.f),
    C_Vertex3f(0.206105769f, 0.0161517859f, -1.f),
    C_Vertex3f(0.221037865f, 0.09724617f, -1.f),
    C_Vertex3f(0.224801719f, 0.00846779346f, -1.f),
    C_Vertex3f(0.242351532f, 0.0845222473f, -1.f),
    C_Vertex3f(0.244498372f, 0.00206446648f, -1.f),
    C_Vertex3f(0.268528938f, 0.073266983f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.167205811f, 0.157086372f, -1.f),
    C_Vertex3f(0.171715975f, 0.0353617668f, -1.f),
    C_Vertex3f(0.178245544f, 0.139414787f, -1.f),
    C_Vertex3f(0.188410521f, 0.0251164436f, -1.f),
    C_Vertex3f(0.190897465f, 0.123550892f, -1.f),
    C_Vertex3f(0.206105769f, 0.0161517859f, -1.f),
    C_Vertex3f(0.205161572f, 0.109494686f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.157778263f, 0.176565647f, -1.f),
    C_Vertex3f(0.149962902f, 0.197852612f, -1.f),
    C_Vertex3f(0.156022131f, 0.0468877554f, -1.f),
    C_Vertex3f(0.171715975f, 0.0353617668f, -1.f),
    C_Vertex3f(0.167205811f, 0.157086372f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156022131f, 0.0468877554f, -1.f),
    C_Vertex3f(0.149962902f, 0.197852612f, -1.f),
    C_Vertex3f(0.143759727f, 0.220947266f, -1.f),
    C_Vertex3f(0.141464233f, 0.233398438f, -1.f),
    C_Vertex3f(0.14132899f, 0.0596944094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431619644f, 0.558502197f, -1.f),
    C_Vertex3f(0.445856094f, 0.550262451f, -1.f),
    C_Vertex3f(0.446669817f, 0.635261536f, -1.f),
    C_Vertex3f(0.459641457f, 0.540649414f, -1.f),
    C_Vertex3f(0.462161779f, 0.628623962f, -1.f),
    C_Vertex3f(0.466308594f, 0.53515625f, -1.f),
    C_Vertex3f(0.476847649f, 0.620880127f, -1.f),
    C_Vertex3f(0.483787537f, 0.616455078f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318394184f, 0.577998638f, -1.f),
    C_Vertex3f(0.336018562f, 0.579101563f, -1.f),
    C_Vertex3f(0.336353302f, 0.651855469f, -1.f),
    C_Vertex3f(0.353671074f, 0.579101563f, -1.f),
    C_Vertex3f(0.357118607f, 0.651855469f, -1.f),
    C_Vertex3f(0.370162964f, 0.577728271f, -1.f),
    C_Vertex3f(0.376641035f, 0.650749207f, -1.f),
    C_Vertex3f(0.386203766f, 0.574981689f, -1.f),
    C_Vertex3f(0.39535737f, 0.648536682f, -1.f),
    C_Vertex3f(0.40179348f, 0.570861816f, -1.f),
    C_Vertex3f(0.413267612f, 0.645217896f, -1.f),
    C_Vertex3f(0.416932106f, 0.565368652f, -1.f),
    C_Vertex3f(0.430371761f, 0.640792847f, -1.f),
    C_Vertex3f(0.431619644f, 0.558502197f, -1.f),
    C_Vertex3f(0.446669817f, 0.635261536f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241337299f, 0.555940151f, -1.f),
    C_Vertex3f(0.255273342f, 0.562557697f, -1.f),
    C_Vertex3f(0.256051064f, 0.639853477f, -1.f),
    C_Vertex3f(0.269947052f, 0.568072319f, -1.f),
    C_Vertex3f(0.275450706f, 0.644654274f, -1.f),
    C_Vertex3f(0.285358429f, 0.572484016f, -1.f),
    C_Vertex3f(0.295300961f, 0.648254871f, -1.f),
    C_Vertex3f(0.301507473f, 0.575792789f, -1.f),
    C_Vertex3f(0.315601826f, 0.65065527f, -1.f),
    C_Vertex3f(0.318394184f, 0.577998638f, -1.f),
    C_Vertex3f(0.336353302f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186875343f, 0.511133194f, -1.f),
    C_Vertex3f(0.200320244f, 0.525945663f, -1.f),
    C_Vertex3f(0.200555801f, 0.618249893f, -1.f),
    C_Vertex3f(0.214475632f, 0.538579941f, -1.f),
    C_Vertex3f(0.218603611f, 0.626651287f, -1.f),
    C_Vertex3f(0.228138924f, 0.548219681f, -1.f),
    C_Vertex3f(0.237102032f, 0.633852482f, -1.f),
    C_Vertex3f(0.241337299f, 0.555940151f, -1.f),
    C_Vertex3f(0.256051064f, 0.639853477f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.177495003f, 0.605701447f, -1.f),
    C_Vertex3f(0.151689053f, 0.587696552f, -1.f),
    C_Vertex3f(0.152280807f, 0.455310822f, -1.f),
    C_Vertex3f(0.162008286f, 0.474645615f, -1.f),
    C_Vertex3f(0.17414093f, 0.494142532f, -1.f),
    C_Vertex3f(0.186875343f, 0.511133194f, -1.f),
    C_Vertex3f(0.200555801f, 0.618249893f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.151689053f, 0.587696552f, -1.f),
    C_Vertex3f(0.128601551f, 0.567182064f, -1.f),
    C_Vertex3f(0.132270813f, 0.350097656f, -1.f),
    C_Vertex3f(0.139274597f, 0.409179688f, -1.f),
    C_Vertex3f(0.140367508f, 0.418111801f, -1.f),
    C_Vertex3f(0.145067215f, 0.436466217f, -1.f),
    C_Vertex3f(0.152280807f, 0.455310822f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.131835938f, 0.340934753f, -1.f),
    C_Vertex3f(0.132270813f, 0.350097656f, -1.f),
    C_Vertex3f(0.128601551f, 0.567182064f, -1.f),
    C_Vertex3f(0.131835938f, 0.324237823f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.139274597f, 0.409179688f, -1.f),
    C_Vertex3f(0.132270813f, 0.350097656f, -1.f),
    C_Vertex3f(0.425697327f, 0.350097656f, -1.f),
    C_Vertex3f(0.4375f, 0.409179688f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0444335938f, 0.340934753f, -1.f),
    C_Vertex3f(0.0444335938f, 0.3215065f, -1.f),
    C_Vertex3f(0.0450897217f, 0.301647186f, -1.f),
    C_Vertex3f(0.0448684692f, 0.350097656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0448684692f, 0.350097656f, -1.f),
    C_Vertex3f(-7.62939453e-006f, 0.409179688f, -1.f),
    C_Vertex3f(-0.0122070313f, 0.350097656f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_129[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_130[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0541992188f, 0.f, 0.f),
    C_Vertex3f(0.0541992188f, 0.f, -1.f),
    C_Vertex3f(0.0541992188f, 0.0913085938f, -1.f),
    C_Vertex3f(0.0541992188f, 0.0913085938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0541992188f, 0.0913085938f, 0.f),
    C_Vertex3f(0.0541992188f, 0.0913085938f, -1.f),
    C_Vertex3f(0.138671875f, 0.0913085938f, -1.f),
    C_Vertex3f(0.138671875f, 0.0913085938f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.138671875f, 0.0913085938f, 0.f),
    C_Vertex3f(0.138671875f, 0.0913085938f, -1.f),
    C_Normal3f(0.999989629f, -0.00455305772f, 0.f),
    C_Vertex3f(0.138671875f, 0.0117206573f, 0.f),
    C_Vertex3f(0.138671875f, 0.0117206573f, -1.f),
    C_Normal3f(0.998822749f, -0.0485085919f, 0.f),
    C_Vertex3f(0.13824749f, -0.0018992424f, 0.f),
    C_Vertex3f(0.13824749f, -0.0018992424f, -1.f),
    C_Normal3f(0.996167064f, -0.0874710083f, 0.f),
    C_Vertex3f(0.13739872f, -0.0144944191f, 0.f),
    C_Vertex3f(0.13739872f, -0.0144944191f, -1.f),
    C_Normal3f(0.99109906f, -0.133126646f, 0.f),
    C_Vertex3f(0.136125565f, -0.0260648727f, 0.f),
    C_Vertex3f(0.136125565f, -0.0260648727f, -1.f),
    C_Normal3f(0.982363462f, -0.186981231f, 0.f),
    C_Vertex3f(0.134428024f, -0.0366106033f, 0.f),
    C_Vertex3f(0.134428024f, -0.0366106033f, -1.f),
    C_Normal3f(0.968034923f, -0.250815451f, 0.f),
    C_Vertex3f(0.132306099f, -0.0461316109f, 0.f),
    C_Vertex3f(0.132306099f, -0.0461316109f, -1.f),
    C_Normal3f(0.945174932f, -0.326564759f, 0.f),
    C_Vertex3f(0.129759789f, -0.0546278954f, 0.f),
    C_Vertex3f(0.129759789f, -0.0546278954f, -1.f),
    C_Normal3f(0.901648641f, -0.432469338f, 0.f),
    C_Vertex3f(0.126789093f, -0.0620994568f, 0.f),
    C_Vertex3f(0.126789093f, -0.0620994568f, -1.f),
    C_Normal3f(0.831324697f, -0.555787146f, 0.f),
    C_Vertex3f(0.120384216f, -0.074174881f, 0.f),
    C_Vertex3f(0.120384216f, -0.074174881f, -1.f),
    C_Normal3f(0.733749926f, -0.679419637f, 0.f),
    C_Vertex3f(0.108341217f, -0.0896930695f, 0.f),
    C_Vertex3f(0.108341217f, -0.0896930695f, -1.f),
    C_Normal3f(0.607928216f, -0.793991923f, 0.f),
    C_Vertex3f(0.0936698914f, -0.103025436f, 0.f),
    C_Vertex3f(0.0936698914f, -0.103025436f, -1.f),
    C_Normal3f(0.497282386f, -0.867588699f, 0.f),
    C_Vertex3f(0.0763702393f, -0.114171982f, 0.f),
    C_Vertex3f(0.0763702393f, -0.114171982f, -1.f),
    C_Normal3f(0.410103351f, -0.912039101f, 0.f),
    C_Vertex3f(0.06640625f, -0.118652344f, 0.f),
    C_Vertex3f(0.06640625f, -0.118652344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.852290988f, -0.52306819f, 0.f),
    C_Vertex3f(0.06640625f, -0.118652344f, 0.f),
    C_Vertex3f(0.06640625f, -0.118652344f, -1.f),
    C_Vertex3f(0.0473632813f, -0.0876235962f, -1.f),
    C_Vertex3f(0.0473632813f, -0.0876235962f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.378667861f, 0.925532579f, 0.f),
    C_Vertex3f(0.0473632813f, -0.0876235962f, 0.f),
    C_Vertex3f(0.0473632813f, -0.0876235962f, -1.f),
    C_Normal3f(-0.470709801f, 0.882288098f, 0.f),
    C_Vertex3f(0.0532512665f, -0.0852146149f, 0.f),
    C_Vertex3f(0.0532512665f, -0.0852146149f, -1.f),
    C_Normal3f(-0.58732599f, 0.80935055f, 0.f),
    C_Vertex3f(0.0635280609f, -0.0789995193f, 0.f),
    C_Vertex3f(0.0635280609f, -0.0789995193f, -1.f),
    C_Normal3f(-0.719236493f, 0.694765329f, 0.f),
    C_Vertex3f(0.0723056793f, -0.071387291f, 0.f),
    C_Vertex3f(0.0723056793f, -0.071387291f, -1.f),
    C_Normal3f(-0.829798639f, 0.558062911f, 0.f),
    C_Vertex3f(0.0795841217f, -0.0623779297f, 0.f),
    C_Vertex3f(0.0795841217f, -0.0623779297f, -1.f),
    C_Normal3f(-0.911361814f, 0.411606163f, 0.f),
    C_Vertex3f(0.0853633881f, -0.0519714355f, 0.f),
    C_Vertex3f(0.0853633881f, -0.0519714355f, -1.f),
    C_Normal3f(-0.96053046f, 0.278174967f, 0.f),
    C_Vertex3f(0.0898618698f, -0.0396213531f, 0.f),
    C_Vertex3f(0.0898618698f, -0.0396213531f, -1.f),
    C_Normal3f(-0.98604548f, 0.166476414f, 0.f),
    C_Vertex3f(0.093079567f, -0.0253276825f, 0.f),
    C_Vertex3f(0.093079567f, -0.0253276825f, -1.f),
    C_Normal3f(-0.996025264f, 0.0890714675f, 0.f),
    C_Vertex3f(0.0950164795f, -0.00909042358f, 0.f),
    C_Vertex3f(0.0950164795f, -0.00909042358f, -1.f),
    C_Normal3f(-0.999349475f, 0.0360654853f, 0.f),
    C_Vertex3f(0.0953445435f, 0.f, 0.f),
    C_Vertex3f(0.0953445435f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0953445435f, 0.f, 0.f),
    C_Vertex3f(0.0953445435f, 0.f, -1.f),
    C_Vertex3f(0.0541992188f, 0.f, -1.f),
    C_Vertex3f(0.0541992188f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138671875f, 0.0913085938f, 0.f),
    C_Vertex3f(0.0541992188f, 0.0913085938f, 0.f),
    C_Vertex3f(0.0953445435f, 0.f, 0.f),
    C_Vertex3f(0.108341217f, -0.0896930695f, 0.f),
    C_Vertex3f(0.120384216f, -0.074174881f, 0.f),
    C_Vertex3f(0.126789093f, -0.0620994568f, 0.f),
    C_Vertex3f(0.129759789f, -0.0546278954f, 0.f),
    C_Vertex3f(0.132306099f, -0.0461316109f, 0.f),
    C_Vertex3f(0.134428024f, -0.0366106033f, 0.f),
    C_Vertex3f(0.136125565f, -0.0260648727f, 0.f),
    C_Vertex3f(0.13739872f, -0.0144944191f, 0.f),
    C_Vertex3f(0.13824749f, -0.0018992424f, 0.f),
    C_Vertex3f(0.138671875f, 0.0117206573f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0950164795f, -0.00909042358f, 0.f),
    C_Vertex3f(0.093079567f, -0.0253276825f, 0.f),
    C_Vertex3f(0.0936698914f, -0.103025436f, 0.f),
    C_Vertex3f(0.108341217f, -0.0896930695f, 0.f),
    C_Vertex3f(0.0953445435f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0936698914f, -0.103025436f, 0.f),
    C_Vertex3f(0.093079567f, -0.0253276825f, 0.f),
    C_Vertex3f(0.0898618698f, -0.0396213531f, 0.f),
    C_Vertex3f(0.0853633881f, -0.0519714355f, 0.f),
    C_Vertex3f(0.0795841217f, -0.0623779297f, 0.f),
    C_Vertex3f(0.0763702393f, -0.114171982f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0723056793f, -0.071387291f, 0.f),
    C_Vertex3f(0.0635280609f, -0.0789995193f, 0.f),
    C_Vertex3f(0.06640625f, -0.118652344f, 0.f),
    C_Vertex3f(0.0763702393f, -0.114171982f, 0.f),
    C_Vertex3f(0.0795841217f, -0.0623779297f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0532512665f, -0.0852146149f, 0.f),
    C_Vertex3f(0.0473632813f, -0.0876235962f, 0.f),
    C_Vertex3f(0.06640625f, -0.118652344f, 0.f),
    C_Vertex3f(0.0635280609f, -0.0789995193f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0953445435f, 0.f, 0.f),
    C_Vertex3f(0.0541992188f, 0.0913085938f, 0.f),
    C_Vertex3f(0.0541992188f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138671875f, 0.0913085938f, -1.f),
    C_Vertex3f(0.0541992188f, 0.0913085938f, -1.f),
    C_Vertex3f(0.0953445435f, 0.f, -1.f),
    C_Vertex3f(0.108341217f, -0.0896930695f, -1.f),
    C_Vertex3f(0.120384216f, -0.074174881f, -1.f),
    C_Vertex3f(0.126789093f, -0.0620994568f, -1.f),
    C_Vertex3f(0.129759789f, -0.0546278954f, -1.f),
    C_Vertex3f(0.132306099f, -0.0461316109f, -1.f),
    C_Vertex3f(0.134428024f, -0.0366106033f, -1.f),
    C_Vertex3f(0.136125565f, -0.0260648727f, -1.f),
    C_Vertex3f(0.13739872f, -0.0144944191f, -1.f),
    C_Vertex3f(0.13824749f, -0.0018992424f, -1.f),
    C_Vertex3f(0.138671875f, 0.0117206573f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0950164795f, -0.00909042358f, -1.f),
    C_Vertex3f(0.093079567f, -0.0253276825f, -1.f),
    C_Vertex3f(0.0936698914f, -0.103025436f, -1.f),
    C_Vertex3f(0.108341217f, -0.0896930695f, -1.f),
    C_Vertex3f(0.0953445435f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0936698914f, -0.103025436f, -1.f),
    C_Vertex3f(0.093079567f, -0.0253276825f, -1.f),
    C_Vertex3f(0.0898618698f, -0.0396213531f, -1.f),
    C_Vertex3f(0.0853633881f, -0.0519714355f, -1.f),
    C_Vertex3f(0.0795841217f, -0.0623779297f, -1.f),
    C_Vertex3f(0.0763702393f, -0.114171982f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0723056793f, -0.071387291f, -1.f),
    C_Vertex3f(0.0635280609f, -0.0789995193f, -1.f),
    C_Vertex3f(0.06640625f, -0.118652344f, -1.f),
    C_Vertex3f(0.0763702393f, -0.114171982f, -1.f),
    C_Vertex3f(0.0795841217f, -0.0623779297f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0532512665f, -0.0852146149f, -1.f),
    C_Vertex3f(0.0473632813f, -0.0876235962f, -1.f),
    C_Vertex3f(0.06640625f, -0.118652344f, -1.f),
    C_Vertex3f(0.0635280609f, -0.0789995193f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0953445435f, 0.f, -1.f),
    C_Vertex3f(0.0541992188f, 0.0913085938f, -1.f),
    C_Vertex3f(0.0541992188f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.198730469f, 0.f, 0.f),
};
static const Call char_131[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.977637827f, 0.210295573f, 0.f),
    C_Vertex3f(0.0201263428f, -0.176651001f, 0.f),
    C_Vertex3f(0.0201263428f, -0.176651001f, -1.f),
    C_Vertex3f(0.0347061157f, -0.10887146f, -1.f),
    C_Vertex3f(0.0347061157f, -0.10887146f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.214906573f, 0.976634681f, 0.f),
    C_Vertex3f(0.0347061157f, -0.10887146f, 0.f),
    C_Vertex3f(0.0347061157f, -0.10887146f, -1.f),
    C_Normal3f(0.195992649f, 0.980605364f, 0.f),
    C_Vertex3f(0.0458183289f, -0.111316681f, 0.f),
    C_Vertex3f(0.0458183289f, -0.111316681f, -1.f),
    C_Normal3f(0.166534618f, 0.986035645f, 0.f),
    C_Vertex3f(0.0652914047f, -0.114984512f, 0.f),
    C_Vertex3f(0.0652914047f, -0.114984512f, -1.f),
    C_Normal3f(0.118659787f, 0.992934942f, 0.f),
    C_Vertex3f(0.0820131302f, -0.117429733f, 0.f),
    C_Vertex3f(0.0820131302f, -0.117429733f, -1.f),
    C_Normal3f(0.047850661f, 0.998854518f, 0.f),
    C_Vertex3f(0.0959835052f, -0.118652344f, 0.f),
    C_Vertex3f(0.0959835052f, -0.118652344f, -1.f),
    C_Normal3f(-0.0692039728f, 0.997602582f, 0.f),
    C_Vertex3f(0.107534409f, -0.118652344f, 0.f),
    C_Vertex3f(0.107534409f, -0.118652344f, -1.f),
    C_Normal3f(-0.227614522f, 0.973751366f, 0.f),
    C_Vertex3f(0.118041515f, -0.117122173f, 0.f),
    C_Vertex3f(0.118041515f, -0.117122173f, -1.f),
    C_Normal3f(-0.412678719f, 0.910876751f, 0.f),
    C_Vertex3f(0.127172947f, -0.114061832f, 0.f),
    C_Vertex3f(0.127172947f, -0.114061832f, -1.f),
    C_Normal3f(-0.603937447f, 0.79703182f, 0.f),
    C_Vertex3f(0.134928703f, -0.109471321f, 0.f),
    C_Vertex3f(0.134928703f, -0.109471321f, -1.f),
    C_Normal3f(-0.798138738f, 0.602473676f, 0.f),
    C_Vertex3f(0.141308784f, -0.103350639f, 0.f),
    C_Vertex3f(0.141308784f, -0.103350639f, -1.f),
    C_Normal3f(-0.91059643f, 0.413296521f, 0.f),
    C_Vertex3f(0.146974087f, -0.0935139656f, 0.f),
    C_Vertex3f(0.146974087f, -0.0935139656f, -1.f),
    C_Normal3f(-0.958339274f, 0.285632163f, 0.f),
    C_Vertex3f(0.15192461f, -0.0799612999f, 0.f),
    C_Vertex3f(0.15192461f, -0.0799612999f, -1.f),
    C_Normal3f(-0.985764682f, 0.168131202f, 0.f),
    C_Vertex3f(0.156160355f, -0.0626926422f, 0.f),
    C_Vertex3f(0.156160355f, -0.0626926422f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.985764682f, 0.168131202f, 0.f),
    C_Vertex3f(0.156160355f, -0.0626926422f, 0.f),
    C_Vertex3f(0.156160355f, -0.0626926422f, -1.f),
    C_Vertex3f(0.235626221f, 0.40322113f, -1.f),
    C_Vertex3f(0.235626221f, 0.40322113f, 0.f),
    C_Normal3f(-0.00112495606f, -0.999999404f, 0.f),
    C_Vertex3f(0.235626221f, 0.40322113f, 0.f),
    C_Vertex3f(0.235626221f, 0.40322113f, -1.f),
    C_Vertex3f(0.147460938f, 0.403320313f, -1.f),
    C_Vertex3f(0.147460938f, 0.403320313f, 0.f),
    C_Normal3f(-0.985631049f, 0.168912515f, 0.f),
    C_Vertex3f(0.147460938f, 0.403320313f, 0.f),
    C_Vertex3f(0.147460938f, 0.403320313f, -1.f),
    C_Vertex3f(0.157920837f, 0.464355469f, -1.f),
    C_Vertex3f(0.157920837f, 0.464355469f, 0.f),
    C_Normal3f(0.00103896041f, 0.999999464f, 0.f),
    C_Vertex3f(0.157920837f, 0.464355469f, 0.f),
    C_Vertex3f(0.157920837f, 0.464355469f, -1.f),
    C_Vertex3f(0.246040344f, 0.464263916f, -1.f),
    C_Vertex3f(0.246040344f, 0.464263916f, 0.f),
    C_Normal3f(-0.985771477f, 0.168090984f, 0.f),
    C_Vertex3f(0.246040344f, 0.464263916f, 0.f),
    C_Vertex3f(0.246040344f, 0.464263916f, -1.f),
    C_Vertex3f(0.255950928f, 0.522384644f, -1.f),
    C_Vertex3f(0.255950928f, 0.522384644f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.985771477f, 0.168090984f, 0.f),
    C_Vertex3f(0.255950928f, 0.522384644f, 0.f),
    C_Vertex3f(0.255950928f, 0.522384644f, -1.f),
    C_Normal3f(-0.981618404f, 0.190854162f, 0.f),
    C_Vertex3f(0.257151604f, 0.528923988f, 0.f),
    C_Vertex3f(0.257151604f, 0.528923988f, -1.f),
    C_Normal3f(-0.978702903f, 0.205281645f, 0.f),
    C_Vertex3f(0.259566545f, 0.540980816f, 0.f),
    C_Vertex3f(0.259566545f, 0.540980816f, -1.f),
    C_Normal3f(-0.974255085f, 0.225448638f, 0.f),
    C_Vertex3f(0.261995077f, 0.552015781f, 0.f),
    C_Vertex3f(0.261995077f, 0.552015781f, -1.f),
    C_Normal3f(-0.968357563f, 0.249566808f, 0.f),
    C_Vertex3f(0.264437199f, 0.562028885f, 0.f),
    C_Vertex3f(0.264437199f, 0.562028885f, -1.f),
    C_Normal3f(-0.960331619f, 0.27886039f, 0.f),
    C_Vertex3f(0.26689291f, 0.571020126f, 0.f),
    C_Vertex3f(0.26689291f, 0.571020126f, -1.f),
    C_Normal3f(-0.949066401f, 0.315076202f, 0.f),
    C_Vertex3f(0.269362211f, 0.578989506f, 0.f),
    C_Vertex3f(0.269362211f, 0.578989506f, -1.f),
    C_Normal3f(-0.93266052f, 0.360755175f, 0.f),
    C_Vertex3f(0.271845102f, 0.585937023f, 0.f),
    C_Vertex3f(0.271845102f, 0.585937023f, -1.f),
    C_Normal3f(-0.904462218f, 0.426553816f, 0.f),
    C_Vertex3f(0.274341583f, 0.591862679f, 0.f),
    C_Vertex3f(0.274341583f, 0.591862679f, -1.f),
    C_Normal3f(-0.855574369f, 0.51767993f, 0.f),
    C_Vertex3f(0.278982162f, 0.601070404f, 0.f),
    C_Vertex3f(0.278982162f, 0.601070404f, -1.f),
    C_Normal3f(-0.784430742f, 0.62021637f, 0.f),
    C_Vertex3f(0.287198544f, 0.613111496f, 0.f),
    C_Vertex3f(0.287198544f, 0.613111496f, -1.f),
    C_Normal3f(-0.687651217f, 0.726041138f, 0.f),
    C_Vertex3f(0.296860218f, 0.623682022f, 0.f),
    C_Vertex3f(0.296860218f, 0.623682022f, -1.f),
    C_Normal3f(-0.577344596f, 0.816500604f, 0.f),
    C_Vertex3f(0.307967186f, 0.632781982f, 0.f),
    C_Vertex3f(0.307967186f, 0.632781982f, -1.f),
    C_Normal3f(-0.442698717f, 0.896670461f, 0.f),
    C_Vertex3f(0.320519447f, 0.640411377f, 0.f),
    C_Vertex3f(0.320519447f, 0.640411377f, -1.f),
    C_Normal3f(-0.294736117f, 0.955578685f, 0.f),
    C_Vertex3f(0.335010052f, 0.646133423f, 0.f),
    C_Vertex3f(0.335010052f, 0.646133423f, -1.f),
    C_Normal3f(-0.162265033f, 0.986747205f, 0.f),
    C_Vertex3f(0.351438999f, 0.64994812f, 0.f),
    C_Vertex3f(0.351438999f, 0.64994812f, -1.f),
    C_Normal3f(-0.0511317179f, 0.998691976f, 0.f),
    C_Vertex3f(0.36980629f, 0.651855469f, 0.f),
    C_Vertex3f(0.36980629f, 0.651855469f, -1.f),
    C_Normal3f(0.0525118373f, 0.998620272f, 0.f),
    C_Vertex3f(0.388692856f, 0.651855469f, 0.f),
    C_Vertex3f(0.388692856f, 0.651855469f, -1.f),
    C_Normal3f(0.139907777f, 0.990164518f, 0.f),
    C_Vertex3f(0.409161568f, 0.649785995f, 0.f),
    C_Vertex3f(0.409161568f, 0.649785995f, -1.f),
    C_Normal3f(0.202882618f, 0.979203105f, 0.f),
    C_Vertex3f(0.432631493f, 0.645647049f, 0.f),
    C_Vertex3f(0.432631493f, 0.645647049f, -1.f),
    C_Normal3f(0.243544266f, 0.96988982f, 0.f),
    C_Vertex3f(0.459102631f, 0.639438629f, 0.f),
    C_Vertex3f(0.459102631f, 0.639438629f, -1.f),
    C_Normal3f(0.270406365f, 0.962746322f, 0.f),
    C_Vertex3f(0.473838806f, 0.635299683f, 0.f),
    C_Vertex3f(0.473838806f, 0.635299683f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.974451303f, -0.22459887f, 0.f),
    C_Vertex3f(0.473838806f, 0.635299683f, 0.f),
    C_Vertex3f(0.473838806f, 0.635299683f, -1.f),
    C_Vertex3f(0.458473206f, 0.568634033f, -1.f),
    C_Vertex3f(0.458473206f, 0.568634033f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.22704652f, -0.973883986f, 0.f),
    C_Vertex3f(0.458473206f, 0.568634033f, 0.f),
    C_Vertex3f(0.458473206f, 0.568634033f, -1.f),
    C_Normal3f(-0.207510337f, -0.978232801f, 0.f),
    C_Vertex3f(0.447248459f, 0.571250916f, 0.f),
    C_Vertex3f(0.447248459f, 0.571250916f, -1.f),
    C_Normal3f(-0.176901728f, -0.984228611f, 0.f),
    C_Vertex3f(0.427632332f, 0.575176239f, 0.f),
    C_Vertex3f(0.427632332f, 0.575176239f, -1.f),
    C_Normal3f(-0.12669763f, -0.991941452f, 0.f),
    C_Vertex3f(0.410849571f, 0.577793121f, 0.f),
    C_Vertex3f(0.410849571f, 0.577793121f, -1.f),
    C_Normal3f(-0.0510226414f, -0.998697579f, 0.f),
    C_Vertex3f(0.396900177f, 0.579101563f, 0.f),
    C_Vertex3f(0.396900177f, 0.579101563f, -1.f),
    C_Normal3f(0.0670348704f, -0.99775064f, 0.f),
    C_Vertex3f(0.385238647f, 0.579101563f, 0.f),
    C_Vertex3f(0.385238647f, 0.579101563f, -1.f),
    C_Normal3f(0.221401215f, -0.975182772f, 0.f),
    C_Vertex3f(0.374557972f, 0.577600479f, 0.f),
    C_Vertex3f(0.374557972f, 0.577600479f, -1.f),
    C_Normal3f(0.40825516f, -0.912867844f, 0.f),
    C_Vertex3f(0.365403652f, 0.574598312f, 0.f),
    C_Vertex3f(0.365403652f, 0.574598312f, -1.f),
    C_Normal3f(0.593795121f, -0.804616272f, 0.f),
    C_Vertex3f(0.357775688f, 0.570095062f, 0.f),
    C_Vertex3f(0.357775688f, 0.570095062f, -1.f),
    C_Normal3f(0.752048552f, -0.659107804f, 0.f),
    C_Vertex3f(0.353199482f, 0.565591812f, 0.f),
    C_Vertex3f(0.353199482f, 0.565591812f, -1.f),
    C_Normal3f(0.84803921f, -0.529933631f, 0.f),
    C_Vertex3f(0.350271106f, 0.561532259f, 0.f),
    C_Vertex3f(0.350271106f, 0.561532259f, -1.f),
    C_Normal3f(0.899347246f, -0.437235177f, 0.f),
    C_Vertex3f(0.347465158f, 0.556415319f, 0.f),
    C_Vertex3f(0.347465158f, 0.556415319f, -1.f),
    C_Normal3f(0.931271672f, -0.364325315f, 0.f),
    C_Vertex3f(0.344781637f, 0.550240993f, 0.f),
    C_Vertex3f(0.344781637f, 0.550240993f, -1.f),
    C_Normal3f(0.951833189f, -0.306616336f, 0.f),
    C_Vertex3f(0.342220545f, 0.543009281f, 0.f),
    C_Vertex3f(0.342220545f, 0.543009281f, -1.f),
    C_Normal3f(0.965521395f, -0.260323584f, 0.f),
    C_Vertex3f(0.33978188f, 0.534720182f, 0.f),
    C_Vertex3f(0.33978188f, 0.534720182f, -1.f),
    C_Normal3f(0.974904895f, -0.222622052f, 0.f),
    C_Vertex3f(0.337465644f, 0.525373697f, 0.f),
    C_Vertex3f(0.337465644f, 0.525373697f, -1.f),
    C_Normal3f(0.984859109f, -0.173356578f, 0.f),
    C_Vertex3f(0.335271835f, 0.514969826f, 0.f),
    C_Vertex3f(0.335271835f, 0.514969826f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.984859109f, -0.173356578f, 0.f),
    C_Vertex3f(0.335271835f, 0.514969826f, 0.f),
    C_Vertex3f(0.335271835f, 0.514969826f, -1.f),
    C_Vertex3f(0.32636261f, 0.464355469f, -1.f),
    C_Vertex3f(0.32636261f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.32636261f, 0.464355469f, 0.f),
    C_Vertex3f(0.32636261f, 0.464355469f, -1.f),
    C_Vertex3f(0.414550781f, 0.464355469f, -1.f),
    C_Vertex3f(0.414550781f, 0.464355469f, 0.f),
    C_Normal3f(0.984439731f, -0.175722495f, 0.f),
    C_Vertex3f(0.414550781f, 0.464355469f, 0.f),
    C_Vertex3f(0.414550781f, 0.464355469f, -1.f),
    C_Vertex3f(0.403656006f, 0.403320313f, -1.f),
    C_Vertex3f(0.403656006f, 0.403320313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.403656006f, 0.403320313f, 0.f),
    C_Vertex3f(0.403656006f, 0.403320313f, -1.f),
    C_Vertex3f(0.315643311f, 0.403320313f, -1.f),
    C_Vertex3f(0.315643311f, 0.403320313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.984913886f, -0.173045263f, 0.f),
    C_Vertex3f(0.315643311f, 0.403320313f, 0.f),
    C_Vertex3f(0.315643311f, 0.403320313f, -1.f),
    C_Normal3f(0.984677255f, -0.174386486f, 0.f),
    C_Vertex3f(0.230344772f, -0.0821695328f, 0.f),
    C_Vertex3f(0.230344772f, -0.0821695328f, -1.f),
    C_Normal3f(0.970288813f, -0.24194926f, 0.f),
    C_Vertex3f(0.227024555f, -0.0970674753f, 0.f),
    C_Vertex3f(0.227024555f, -0.0970674753f, -1.f),
    C_Normal3f(0.955168426f, -0.296063095f, 0.f),
    C_Vertex3f(0.22322607f, -0.110717654f, 0.f),
    C_Vertex3f(0.22322607f, -0.110717654f, -1.f),
    C_Normal3f(0.933725357f, -0.357990026f, 0.f),
    C_Vertex3f(0.218949318f, -0.12312007f, 0.f),
    C_Vertex3f(0.218949318f, -0.12312007f, -1.f),
    C_Normal3f(0.903542101f, -0.428499371f, 0.f),
    C_Vertex3f(0.214194298f, -0.134274721f, 0.f),
    C_Vertex3f(0.214194298f, -0.134274721f, -1.f),
    C_Normal3f(0.86145407f, -0.507835507f, 0.f),
    C_Vertex3f(0.20896101f, -0.144181609f, 0.f),
    C_Vertex3f(0.20896101f, -0.144181609f, -1.f),
    C_Normal3f(0.803621411f, -0.595140874f, 0.f),
    C_Vertex3f(0.203249454f, -0.152840734f, 0.f),
    C_Vertex3f(0.203249454f, -0.152840734f, -1.f),
    C_Normal3f(0.726021349f, -0.687672138f, 0.f),
    C_Vertex3f(0.197059631f, -0.160252094f, 0.f),
    C_Vertex3f(0.197059631f, -0.160252094f, -1.f),
    C_Normal3f(0.635703862f, -0.771933079f, 0.f),
    C_Vertex3f(0.190391541f, -0.166415691f, 0.f),
    C_Vertex3f(0.190391541f, -0.166415691f, -1.f),
    C_Normal3f(0.544526279f, -0.838743746f, 0.f),
    C_Vertex3f(0.183026314f, -0.171808839f, 0.f),
    C_Vertex3f(0.183026314f, -0.171808839f, -1.f),
    C_Normal3f(0.4499304f, -0.893063724f, 0.f),
    C_Vertex3f(0.174963951f, -0.176431537f, 0.f),
    C_Vertex3f(0.174963951f, -0.176431537f, -1.f),
    C_Normal3f(0.355751902f, -0.934580386f, 0.f),
    C_Vertex3f(0.166204453f, -0.180283785f, 0.f),
    C_Vertex3f(0.166204453f, -0.180283785f, -1.f),
    C_Normal3f(0.265169561f, -0.964201748f, 0.f),
    C_Vertex3f(0.156747818f, -0.183365583f, 0.f),
    C_Vertex3f(0.156747818f, -0.183365583f, -1.f),
    C_Normal3f(0.18039088f, -0.983595073f, 0.f),
    C_Vertex3f(0.146594048f, -0.185676932f, 0.f),
    C_Vertex3f(0.146594048f, -0.185676932f, -1.f),
    C_Normal3f(0.10264501f, -0.994718015f, 0.f),
    C_Vertex3f(0.135743141f, -0.187217832f, 0.f),
    C_Vertex3f(0.135743141f, -0.187217832f, -1.f),
    C_Normal3f(0.0275592301f, -0.99962014f, 0.f),
    C_Vertex3f(0.124195099f, -0.187988281f, 0.f),
    C_Vertex3f(0.124195099f, -0.187988281f, -1.f),
    C_Normal3f(-0.0363920927f, -0.999337614f, 0.f),
    C_Vertex3f(0.107797623f, -0.187988281f, 0.f),
    C_Vertex3f(0.107797623f, -0.187988281f, -1.f),
    C_Normal3f(-0.0900802612f, -0.995934486f, 0.f),
    C_Vertex3f(0.0852794647f, -0.186571121f, 0.f),
    C_Vertex3f(0.0852794647f, -0.186571121f, -1.f),
    C_Normal3f(-0.137770504f, -0.990464211f, 0.f),
    C_Vertex3f(0.060792923f, -0.183736801f, 0.f),
    C_Vertex3f(0.060792923f, -0.183736801f, -1.f),
    C_Normal3f(-0.171655104f, -0.985157073f, 0.f),
    C_Vertex3f(0.0343379974f, -0.179485321f, 0.f),
    C_Vertex3f(0.0343379974f, -0.179485321f, -1.f),
    C_Normal3f(-0.195584565f, -0.980686784f, 0.f),
    C_Vertex3f(0.0201263428f, -0.176651001f, 0.f),
    C_Vertex3f(0.0201263428f, -0.176651001f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235626221f, 0.40322113f, 0.f),
    C_Vertex3f(0.156160355f, -0.0626926422f, 0.f),
    C_Vertex3f(0.156747818f, -0.183365583f, 0.f),
    C_Vertex3f(0.166204453f, -0.180283785f, 0.f),
    C_Vertex3f(0.174963951f, -0.176431537f, 0.f),
    C_Vertex3f(0.183026314f, -0.171808839f, 0.f),
    C_Vertex3f(0.190391541f, -0.166415691f, 0.f),
    C_Vertex3f(0.197059631f, -0.160252094f, 0.f),
    C_Vertex3f(0.203249454f, -0.152840734f, 0.f),
    C_Vertex3f(0.20896101f, -0.144181609f, 0.f),
    C_Vertex3f(0.214194298f, -0.134274721f, 0.f),
    C_Vertex3f(0.218949318f, -0.12312007f, 0.f),
    C_Vertex3f(0.22322607f, -0.110717654f, 0.f),
    C_Vertex3f(0.227024555f, -0.0970674753f, 0.f),
    C_Vertex3f(0.230344772f, -0.0821695328f, 0.f),
    C_Vertex3f(0.315643311f, 0.403320313f, 0.f),
    C_Vertex3f(0.246040344f, 0.464263916f, 0.f),
    C_Vertex3f(0.157920837f, 0.464355469f, 0.f),
    C_Vertex3f(0.147460938f, 0.403320313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.32636261f, 0.464355469f, 0.f),
    C_Vertex3f(0.335271835f, 0.514969826f, 0.f),
    C_Vertex3f(0.335010052f, 0.646133423f, 0.f),
    C_Vertex3f(0.320519447f, 0.640411377f, 0.f),
    C_Vertex3f(0.315643311f, 0.403320313f, 0.f),
    C_Vertex3f(0.403656006f, 0.403320313f, 0.f),
    C_Vertex3f(0.414550781f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.315643311f, 0.403320313f, 0.f),
    C_Vertex3f(0.320519447f, 0.640411377f, 0.f),
    C_Vertex3f(0.307967186f, 0.632781982f, 0.f),
    C_Vertex3f(0.296860218f, 0.623682022f, 0.f),
    C_Vertex3f(0.287198544f, 0.613111496f, 0.f),
    C_Vertex3f(0.278982162f, 0.601070404f, 0.f),
    C_Vertex3f(0.274341583f, 0.591862679f, 0.f),
    C_Vertex3f(0.271845102f, 0.585937023f, 0.f),
    C_Vertex3f(0.269362211f, 0.578989506f, 0.f),
    C_Vertex3f(0.26689291f, 0.571020126f, 0.f),
    C_Vertex3f(0.264437199f, 0.562028885f, 0.f),
    C_Vertex3f(0.261995077f, 0.552015781f, 0.f),
    C_Vertex3f(0.259566545f, 0.540980816f, 0.f),
    C_Vertex3f(0.257151604f, 0.528923988f, 0.f),
    C_Vertex3f(0.255950928f, 0.522384644f, 0.f),
    C_Vertex3f(0.246040344f, 0.464263916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.156160355f, -0.0626926422f, 0.f),
    C_Vertex3f(0.15192461f, -0.0799612999f, 0.f),
    C_Vertex3f(0.156747818f, -0.183365583f, 0.f),
    C_Vertex3f(0.146974087f, -0.0935139656f, 0.f),
    C_Vertex3f(0.146594048f, -0.185676932f, 0.f),
    C_Vertex3f(0.141308784f, -0.103350639f, 0.f),
    C_Vertex3f(0.135743141f, -0.187217832f, 0.f),
    C_Vertex3f(0.134928703f, -0.109471321f, 0.f),
    C_Vertex3f(0.127172947f, -0.114061832f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.135743141f, -0.187217832f, 0.f),
    C_Vertex3f(0.127172947f, -0.114061832f, 0.f),
    C_Vertex3f(0.124195099f, -0.187988281f, 0.f),
    C_Vertex3f(0.118041515f, -0.117122173f, 0.f),
    C_Vertex3f(0.107797623f, -0.187988281f, 0.f),
    C_Vertex3f(0.107534409f, -0.118652344f, 0.f),
    C_Vertex3f(0.0959835052f, -0.118652344f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0852794647f, -0.186571121f, 0.f),
    C_Vertex3f(0.107797623f, -0.187988281f, 0.f),
    C_Vertex3f(0.0959835052f, -0.118652344f, 0.f),
    C_Vertex3f(0.0820131302f, -0.117429733f, 0.f),
    C_Vertex3f(0.0652914047f, -0.114984512f, 0.f),
    C_Vertex3f(0.060792923f, -0.183736801f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0201263428f, -0.176651001f, 0.f),
    C_Vertex3f(0.0343379974f, -0.179485321f, 0.f),
    C_Vertex3f(0.0347061157f, -0.10887146f, 0.f),
    C_Vertex3f(0.060792923f, -0.183736801f, 0.f),
    C_Vertex3f(0.0458183289f, -0.111316681f, 0.f),
    C_Vertex3f(0.0652914047f, -0.114984512f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.410849571f, 0.577793121f, 0.f),
    C_Vertex3f(0.427632332f, 0.575176239f, 0.f),
    C_Vertex3f(0.432631493f, 0.645647049f, 0.f),
    C_Vertex3f(0.447248459f, 0.571250916f, 0.f),
    C_Vertex3f(0.459102631f, 0.639438629f, 0.f),
    C_Vertex3f(0.458473206f, 0.568634033f, 0.f),
    C_Vertex3f(0.473838806f, 0.635299683f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.374557972f, 0.577600479f, 0.f),
    C_Vertex3f(0.385238647f, 0.579101563f, 0.f),
    C_Vertex3f(0.388692856f, 0.651855469f, 0.f),
    C_Vertex3f(0.396900177f, 0.579101563f, 0.f),
    C_Vertex3f(0.409161568f, 0.649785995f, 0.f),
    C_Vertex3f(0.410849571f, 0.577793121f, 0.f),
    C_Vertex3f(0.432631493f, 0.645647049f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36980629f, 0.651855469f, 0.f),
    C_Vertex3f(0.351438999f, 0.64994812f, 0.f),
    C_Vertex3f(0.353199482f, 0.565591812f, 0.f),
    C_Vertex3f(0.357775688f, 0.570095062f, 0.f),
    C_Vertex3f(0.365403652f, 0.574598312f, 0.f),
    C_Vertex3f(0.374557972f, 0.577600479f, 0.f),
    C_Vertex3f(0.388692856f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.351438999f, 0.64994812f, 0.f),
    C_Vertex3f(0.335010052f, 0.646133423f, 0.f),
    C_Vertex3f(0.335271835f, 0.514969826f, 0.f),
    C_Vertex3f(0.337465644f, 0.525373697f, 0.f),
    C_Vertex3f(0.33978188f, 0.534720182f, 0.f),
    C_Vertex3f(0.342220545f, 0.543009281f, 0.f),
    C_Vertex3f(0.344781637f, 0.550240993f, 0.f),
    C_Vertex3f(0.347465158f, 0.556415319f, 0.f),
    C_Vertex3f(0.350271106f, 0.561532259f, 0.f),
    C_Vertex3f(0.353199482f, 0.565591812f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235626221f, 0.40322113f, -1.f),
    C_Vertex3f(0.156160355f, -0.0626926422f, -1.f),
    C_Vertex3f(0.156747818f, -0.183365583f, -1.f),
    C_Vertex3f(0.166204453f, -0.180283785f, -1.f),
    C_Vertex3f(0.174963951f, -0.176431537f, -1.f),
    C_Vertex3f(0.183026314f, -0.171808839f, -1.f),
    C_Vertex3f(0.190391541f, -0.166415691f, -1.f),
    C_Vertex3f(0.197059631f, -0.160252094f, -1.f),
    C_Vertex3f(0.203249454f, -0.152840734f, -1.f),
    C_Vertex3f(0.20896101f, -0.144181609f, -1.f),
    C_Vertex3f(0.214194298f, -0.134274721f, -1.f),
    C_Vertex3f(0.218949318f, -0.12312007f, -1.f),
    C_Vertex3f(0.22322607f, -0.110717654f, -1.f),
    C_Vertex3f(0.227024555f, -0.0970674753f, -1.f),
    C_Vertex3f(0.230344772f, -0.0821695328f, -1.f),
    C_Vertex3f(0.315643311f, 0.403320313f, -1.f),
    C_Vertex3f(0.246040344f, 0.464263916f, -1.f),
    C_Vertex3f(0.157920837f, 0.464355469f, -1.f),
    C_Vertex3f(0.147460938f, 0.403320313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.32636261f, 0.464355469f, -1.f),
    C_Vertex3f(0.335271835f, 0.514969826f, -1.f),
    C_Vertex3f(0.335010052f, 0.646133423f, -1.f),
    C_Vertex3f(0.320519447f, 0.640411377f, -1.f),
    C_Vertex3f(0.315643311f, 0.403320313f, -1.f),
    C_Vertex3f(0.403656006f, 0.403320313f, -1.f),
    C_Vertex3f(0.414550781f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.315643311f, 0.403320313f, -1.f),
    C_Vertex3f(0.320519447f, 0.640411377f, -1.f),
    C_Vertex3f(0.307967186f, 0.632781982f, -1.f),
    C_Vertex3f(0.296860218f, 0.623682022f, -1.f),
    C_Vertex3f(0.287198544f, 0.613111496f, -1.f),
    C_Vertex3f(0.278982162f, 0.601070404f, -1.f),
    C_Vertex3f(0.274341583f, 0.591862679f, -1.f),
    C_Vertex3f(0.271845102f, 0.585937023f, -1.f),
    C_Vertex3f(0.269362211f, 0.578989506f, -1.f),
    C_Vertex3f(0.26689291f, 0.571020126f, -1.f),
    C_Vertex3f(0.264437199f, 0.562028885f, -1.f),
    C_Vertex3f(0.261995077f, 0.552015781f, -1.f),
    C_Vertex3f(0.259566545f, 0.540980816f, -1.f),
    C_Vertex3f(0.257151604f, 0.528923988f, -1.f),
    C_Vertex3f(0.255950928f, 0.522384644f, -1.f),
    C_Vertex3f(0.246040344f, 0.464263916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.156160355f, -0.0626926422f, -1.f),
    C_Vertex3f(0.15192461f, -0.0799612999f, -1.f),
    C_Vertex3f(0.156747818f, -0.183365583f, -1.f),
    C_Vertex3f(0.146974087f, -0.0935139656f, -1.f),
    C_Vertex3f(0.146594048f, -0.185676932f, -1.f),
    C_Vertex3f(0.141308784f, -0.103350639f, -1.f),
    C_Vertex3f(0.135743141f, -0.187217832f, -1.f),
    C_Vertex3f(0.134928703f, -0.109471321f, -1.f),
    C_Vertex3f(0.127172947f, -0.114061832f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.135743141f, -0.187217832f, -1.f),
    C_Vertex3f(0.127172947f, -0.114061832f, -1.f),
    C_Vertex3f(0.124195099f, -0.187988281f, -1.f),
    C_Vertex3f(0.118041515f, -0.117122173f, -1.f),
    C_Vertex3f(0.107797623f, -0.187988281f, -1.f),
    C_Vertex3f(0.107534409f, -0.118652344f, -1.f),
    C_Vertex3f(0.0959835052f, -0.118652344f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0852794647f, -0.186571121f, -1.f),
    C_Vertex3f(0.107797623f, -0.187988281f, -1.f),
    C_Vertex3f(0.0959835052f, -0.118652344f, -1.f),
    C_Vertex3f(0.0820131302f, -0.117429733f, -1.f),
    C_Vertex3f(0.0652914047f, -0.114984512f, -1.f),
    C_Vertex3f(0.060792923f, -0.183736801f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0201263428f, -0.176651001f, -1.f),
    C_Vertex3f(0.0343379974f, -0.179485321f, -1.f),
    C_Vertex3f(0.0347061157f, -0.10887146f, -1.f),
    C_Vertex3f(0.060792923f, -0.183736801f, -1.f),
    C_Vertex3f(0.0458183289f, -0.111316681f, -1.f),
    C_Vertex3f(0.0652914047f, -0.114984512f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.410849571f, 0.577793121f, -1.f),
    C_Vertex3f(0.427632332f, 0.575176239f, -1.f),
    C_Vertex3f(0.432631493f, 0.645647049f, -1.f),
    C_Vertex3f(0.447248459f, 0.571250916f, -1.f),
    C_Vertex3f(0.459102631f, 0.639438629f, -1.f),
    C_Vertex3f(0.458473206f, 0.568634033f, -1.f),
    C_Vertex3f(0.473838806f, 0.635299683f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.374557972f, 0.577600479f, -1.f),
    C_Vertex3f(0.385238647f, 0.579101563f, -1.f),
    C_Vertex3f(0.388692856f, 0.651855469f, -1.f),
    C_Vertex3f(0.396900177f, 0.579101563f, -1.f),
    C_Vertex3f(0.409161568f, 0.649785995f, -1.f),
    C_Vertex3f(0.410849571f, 0.577793121f, -1.f),
    C_Vertex3f(0.432631493f, 0.645647049f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36980629f, 0.651855469f, -1.f),
    C_Vertex3f(0.351438999f, 0.64994812f, -1.f),
    C_Vertex3f(0.353199482f, 0.565591812f, -1.f),
    C_Vertex3f(0.357775688f, 0.570095062f, -1.f),
    C_Vertex3f(0.365403652f, 0.574598312f, -1.f),
    C_Vertex3f(0.374557972f, 0.577600479f, -1.f),
    C_Vertex3f(0.388692856f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.351438999f, 0.64994812f, -1.f),
    C_Vertex3f(0.335010052f, 0.646133423f, -1.f),
    C_Vertex3f(0.335271835f, 0.514969826f, -1.f),
    C_Vertex3f(0.337465644f, 0.525373697f, -1.f),
    C_Vertex3f(0.33978188f, 0.534720182f, -1.f),
    C_Vertex3f(0.342220545f, 0.543009281f, -1.f),
    C_Vertex3f(0.344781637f, 0.550240993f, -1.f),
    C_Vertex3f(0.347465158f, 0.556415319f, -1.f),
    C_Vertex3f(0.350271106f, 0.561532259f, -1.f),
    C_Vertex3f(0.353199482f, 0.565591812f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_132[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0380859375f, 0.f, 0.f),
    C_Vertex3f(0.0380859375f, 0.f, -1.f),
    C_Vertex3f(0.0380859375f, 0.0913085938f, -1.f),
    C_Vertex3f(0.0380859375f, 0.0913085938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0380859375f, 0.0913085938f, 0.f),
    C_Vertex3f(0.0380859375f, 0.0913085938f, -1.f),
    C_Vertex3f(0.122558594f, 0.0913085938f, -1.f),
    C_Vertex3f(0.122558594f, 0.0913085938f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.122558594f, 0.0913085938f, 0.f),
    C_Vertex3f(0.122558594f, 0.0913085938f, -1.f),
    C_Normal3f(0.999989629f, -0.00455305772f, 0.f),
    C_Vertex3f(0.122558594f, 0.0117206573f, 0.f),
    C_Vertex3f(0.122558594f, 0.0117206573f, -1.f),
    C_Normal3f(0.998822749f, -0.0485085919f, 0.f),
    C_Vertex3f(0.122134209f, -0.0018992424f, 0.f),
    C_Vertex3f(0.122134209f, -0.0018992424f, -1.f),
    C_Normal3f(0.996167064f, -0.0874710083f, 0.f),
    C_Vertex3f(0.121285439f, -0.0144944191f, 0.f),
    C_Vertex3f(0.121285439f, -0.0144944191f, -1.f),
    C_Normal3f(0.99109906f, -0.133126646f, 0.f),
    C_Vertex3f(0.120012283f, -0.0260648727f, 0.f),
    C_Vertex3f(0.120012283f, -0.0260648727f, -1.f),
    C_Normal3f(0.982363462f, -0.186981231f, 0.f),
    C_Vertex3f(0.118314743f, -0.0366106033f, 0.f),
    C_Vertex3f(0.118314743f, -0.0366106033f, -1.f),
    C_Normal3f(0.968034923f, -0.250815451f, 0.f),
    C_Vertex3f(0.116192818f, -0.0461316109f, 0.f),
    C_Vertex3f(0.116192818f, -0.0461316109f, -1.f),
    C_Normal3f(0.945174932f, -0.326564759f, 0.f),
    C_Vertex3f(0.113646507f, -0.0546278954f, 0.f),
    C_Vertex3f(0.113646507f, -0.0546278954f, -1.f),
    C_Normal3f(0.901648641f, -0.432469338f, 0.f),
    C_Vertex3f(0.110675812f, -0.0620994568f, 0.f),
    C_Vertex3f(0.110675812f, -0.0620994568f, -1.f),
    C_Normal3f(0.831324697f, -0.555787146f, 0.f),
    C_Vertex3f(0.104270935f, -0.074174881f, 0.f),
    C_Vertex3f(0.104270935f, -0.074174881f, -1.f),
    C_Normal3f(0.733749926f, -0.679419637f, 0.f),
    C_Vertex3f(0.0922279358f, -0.0896930695f, 0.f),
    C_Vertex3f(0.0922279358f, -0.0896930695f, -1.f),
    C_Normal3f(0.607928216f, -0.793991923f, 0.f),
    C_Vertex3f(0.0775566101f, -0.103025436f, 0.f),
    C_Vertex3f(0.0775566101f, -0.103025436f, -1.f),
    C_Normal3f(0.497282386f, -0.867588699f, 0.f),
    C_Vertex3f(0.060256958f, -0.114171982f, 0.f),
    C_Vertex3f(0.060256958f, -0.114171982f, -1.f),
    C_Normal3f(0.410103351f, -0.912039101f, 0.f),
    C_Vertex3f(0.0502929688f, -0.118652344f, 0.f),
    C_Vertex3f(0.0502929688f, -0.118652344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.852290988f, -0.52306819f, 0.f),
    C_Vertex3f(0.0502929688f, -0.118652344f, 0.f),
    C_Vertex3f(0.0502929688f, -0.118652344f, -1.f),
    C_Vertex3f(0.03125f, -0.0876235962f, -1.f),
    C_Vertex3f(0.03125f, -0.0876235962f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.378667861f, 0.925532579f, 0.f),
    C_Vertex3f(0.03125f, -0.0876235962f, 0.f),
    C_Vertex3f(0.03125f, -0.0876235962f, -1.f),
    C_Normal3f(-0.470709801f, 0.882288098f, 0.f),
    C_Vertex3f(0.0371379852f, -0.0852146149f, 0.f),
    C_Vertex3f(0.0371379852f, -0.0852146149f, -1.f),
    C_Normal3f(-0.58732599f, 0.80935055f, 0.f),
    C_Vertex3f(0.0474147797f, -0.0789995193f, 0.f),
    C_Vertex3f(0.0474147797f, -0.0789995193f, -1.f),
    C_Normal3f(-0.719236493f, 0.694765329f, 0.f),
    C_Vertex3f(0.0561923981f, -0.071387291f, 0.f),
    C_Vertex3f(0.0561923981f, -0.071387291f, -1.f),
    C_Normal3f(-0.829798639f, 0.558062911f, 0.f),
    C_Vertex3f(0.0634708405f, -0.0623779297f, 0.f),
    C_Vertex3f(0.0634708405f, -0.0623779297f, -1.f),
    C_Normal3f(-0.911361814f, 0.411606163f, 0.f),
    C_Vertex3f(0.0692501068f, -0.0519714355f, 0.f),
    C_Vertex3f(0.0692501068f, -0.0519714355f, -1.f),
    C_Normal3f(-0.96053046f, 0.278174967f, 0.f),
    C_Vertex3f(0.0737485886f, -0.0396213531f, 0.f),
    C_Vertex3f(0.0737485886f, -0.0396213531f, -1.f),
    C_Normal3f(-0.98604548f, 0.166476414f, 0.f),
    C_Vertex3f(0.0769662857f, -0.0253276825f, 0.f),
    C_Vertex3f(0.0769662857f, -0.0253276825f, -1.f),
    C_Normal3f(-0.996025264f, 0.0890714675f, 0.f),
    C_Vertex3f(0.0789031982f, -0.00909042358f, 0.f),
    C_Vertex3f(0.0789031982f, -0.00909042358f, -1.f),
    C_Normal3f(-0.999349475f, 0.0360654853f, 0.f),
    C_Vertex3f(0.0792312622f, 0.f, 0.f),
    C_Vertex3f(0.0792312622f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0792312622f, 0.f, 0.f),
    C_Vertex3f(0.0792312622f, 0.f, -1.f),
    C_Vertex3f(0.0380859375f, 0.f, -1.f),
    C_Vertex3f(0.0380859375f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.173828125f, 0.f, 0.f),
    C_Vertex3f(0.173828125f, 0.f, -1.f),
    C_Vertex3f(0.173828125f, 0.0913085938f, -1.f),
    C_Vertex3f(0.173828125f, 0.0913085938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.173828125f, 0.0913085938f, 0.f),
    C_Vertex3f(0.173828125f, 0.0913085938f, -1.f),
    C_Vertex3f(0.258300781f, 0.0913085938f, -1.f),
    C_Vertex3f(0.258300781f, 0.0913085938f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.258300781f, 0.0913085938f, 0.f),
    C_Vertex3f(0.258300781f, 0.0913085938f, -1.f),
    C_Normal3f(0.999989629f, -0.00455305772f, 0.f),
    C_Vertex3f(0.258300781f, 0.0117206573f, 0.f),
    C_Vertex3f(0.258300781f, 0.0117206573f, -1.f),
    C_Normal3f(0.998822749f, -0.0485085919f, 0.f),
    C_Vertex3f(0.257876396f, -0.0018992424f, 0.f),
    C_Vertex3f(0.257876396f, -0.0018992424f, -1.f),
    C_Normal3f(0.996167064f, -0.0874710083f, 0.f),
    C_Vertex3f(0.257027626f, -0.0144944191f, 0.f),
    C_Vertex3f(0.257027626f, -0.0144944191f, -1.f),
    C_Normal3f(0.99109906f, -0.133126646f, 0.f),
    C_Vertex3f(0.255754471f, -0.0260648727f, 0.f),
    C_Vertex3f(0.255754471f, -0.0260648727f, -1.f),
    C_Normal3f(0.982363462f, -0.186981231f, 0.f),
    C_Vertex3f(0.254056931f, -0.0366106033f, 0.f),
    C_Vertex3f(0.254056931f, -0.0366106033f, -1.f),
    C_Normal3f(0.968034923f, -0.250815451f, 0.f),
    C_Vertex3f(0.251935005f, -0.0461316109f, 0.f),
    C_Vertex3f(0.251935005f, -0.0461316109f, -1.f),
    C_Normal3f(0.945174932f, -0.326564759f, 0.f),
    C_Vertex3f(0.249388695f, -0.0546278954f, 0.f),
    C_Vertex3f(0.249388695f, -0.0546278954f, -1.f),
    C_Normal3f(0.901648641f, -0.432469338f, 0.f),
    C_Vertex3f(0.246417999f, -0.0620994568f, 0.f),
    C_Vertex3f(0.246417999f, -0.0620994568f, -1.f),
    C_Normal3f(0.831324697f, -0.555787146f, 0.f),
    C_Vertex3f(0.240013123f, -0.074174881f, 0.f),
    C_Vertex3f(0.240013123f, -0.074174881f, -1.f),
    C_Normal3f(0.733749926f, -0.679419637f, 0.f),
    C_Vertex3f(0.227970123f, -0.0896930695f, 0.f),
    C_Vertex3f(0.227970123f, -0.0896930695f, -1.f),
    C_Normal3f(0.607928216f, -0.793991923f, 0.f),
    C_Vertex3f(0.213298798f, -0.103025436f, 0.f),
    C_Vertex3f(0.213298798f, -0.103025436f, -1.f),
    C_Normal3f(0.497282386f, -0.867588699f, 0.f),
    C_Vertex3f(0.195999146f, -0.114171982f, 0.f),
    C_Vertex3f(0.195999146f, -0.114171982f, -1.f),
    C_Normal3f(0.410103351f, -0.912039101f, 0.f),
    C_Vertex3f(0.186035156f, -0.118652344f, 0.f),
    C_Vertex3f(0.186035156f, -0.118652344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.852290988f, -0.52306819f, 0.f),
    C_Vertex3f(0.186035156f, -0.118652344f, 0.f),
    C_Vertex3f(0.186035156f, -0.118652344f, -1.f),
    C_Vertex3f(0.166992188f, -0.0876235962f, -1.f),
    C_Vertex3f(0.166992188f, -0.0876235962f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.378667861f, 0.925532579f, 0.f),
    C_Vertex3f(0.166992188f, -0.0876235962f, 0.f),
    C_Vertex3f(0.166992188f, -0.0876235962f, -1.f),
    C_Normal3f(-0.470709801f, 0.882288098f, 0.f),
    C_Vertex3f(0.172880173f, -0.0852146149f, 0.f),
    C_Vertex3f(0.172880173f, -0.0852146149f, -1.f),
    C_Normal3f(-0.58732599f, 0.80935055f, 0.f),
    C_Vertex3f(0.183156967f, -0.0789995193f, 0.f),
    C_Vertex3f(0.183156967f, -0.0789995193f, -1.f),
    C_Normal3f(-0.719236493f, 0.694765329f, 0.f),
    C_Vertex3f(0.191934586f, -0.071387291f, 0.f),
    C_Vertex3f(0.191934586f, -0.071387291f, -1.f),
    C_Normal3f(-0.829798639f, 0.558062911f, 0.f),
    C_Vertex3f(0.199213028f, -0.0623779297f, 0.f),
    C_Vertex3f(0.199213028f, -0.0623779297f, -1.f),
    C_Normal3f(-0.911361814f, 0.411606163f, 0.f),
    C_Vertex3f(0.204992294f, -0.0519714355f, 0.f),
    C_Vertex3f(0.204992294f, -0.0519714355f, -1.f),
    C_Normal3f(-0.96053046f, 0.278174967f, 0.f),
    C_Vertex3f(0.209490776f, -0.0396213531f, 0.f),
    C_Vertex3f(0.209490776f, -0.0396213531f, -1.f),
    C_Normal3f(-0.98604548f, 0.166476414f, 0.f),
    C_Vertex3f(0.212708473f, -0.0253276825f, 0.f),
    C_Vertex3f(0.212708473f, -0.0253276825f, -1.f),
    C_Normal3f(-0.996025264f, 0.0890714675f, 0.f),
    C_Vertex3f(0.214645386f, -0.00909042358f, 0.f),
    C_Vertex3f(0.214645386f, -0.00909042358f, -1.f),
    C_Normal3f(-0.999349475f, 0.0360654853f, 0.f),
    C_Vertex3f(0.21497345f, 0.f, 0.f),
    C_Vertex3f(0.21497345f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.21497345f, 0.f, 0.f),
    C_Vertex3f(0.21497345f, 0.f, -1.f),
    C_Vertex3f(0.173828125f, 0.f, -1.f),
    C_Vertex3f(0.173828125f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122558594f, 0.0913085938f, 0.f),
    C_Vertex3f(0.0380859375f, 0.0913085938f, 0.f),
    C_Vertex3f(0.0792312622f, 0.f, 0.f),
    C_Vertex3f(0.0922279358f, -0.0896930695f, 0.f),
    C_Vertex3f(0.104270935f, -0.074174881f, 0.f),
    C_Vertex3f(0.110675812f, -0.0620994568f, 0.f),
    C_Vertex3f(0.113646507f, -0.0546278954f, 0.f),
    C_Vertex3f(0.116192818f, -0.0461316109f, 0.f),
    C_Vertex3f(0.118314743f, -0.0366106033f, 0.f),
    C_Vertex3f(0.120012283f, -0.0260648727f, 0.f),
    C_Vertex3f(0.121285439f, -0.0144944191f, 0.f),
    C_Vertex3f(0.122134209f, -0.0018992424f, 0.f),
    C_Vertex3f(0.122558594f, 0.0117206573f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0789031982f, -0.00909042358f, 0.f),
    C_Vertex3f(0.0769662857f, -0.0253276825f, 0.f),
    C_Vertex3f(0.0775566101f, -0.103025436f, 0.f),
    C_Vertex3f(0.0922279358f, -0.0896930695f, 0.f),
    C_Vertex3f(0.0792312622f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0775566101f, -0.103025436f, 0.f),
    C_Vertex3f(0.0769662857f, -0.0253276825f, 0.f),
    C_Vertex3f(0.0737485886f, -0.0396213531f, 0.f),
    C_Vertex3f(0.0692501068f, -0.0519714355f, 0.f),
    C_Vertex3f(0.0634708405f, -0.0623779297f, 0.f),
    C_Vertex3f(0.060256958f, -0.114171982f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0561923981f, -0.071387291f, 0.f),
    C_Vertex3f(0.0474147797f, -0.0789995193f, 0.f),
    C_Vertex3f(0.0502929688f, -0.118652344f, 0.f),
    C_Vertex3f(0.060256958f, -0.114171982f, 0.f),
    C_Vertex3f(0.0634708405f, -0.0623779297f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0371379852f, -0.0852146149f, 0.f),
    C_Vertex3f(0.03125f, -0.0876235962f, 0.f),
    C_Vertex3f(0.0502929688f, -0.118652344f, 0.f),
    C_Vertex3f(0.0474147797f, -0.0789995193f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.258300781f, 0.0913085938f, 0.f),
    C_Vertex3f(0.173828125f, 0.0913085938f, 0.f),
    C_Vertex3f(0.21497345f, 0.f, 0.f),
    C_Vertex3f(0.227970123f, -0.0896930695f, 0.f),
    C_Vertex3f(0.240013123f, -0.074174881f, 0.f),
    C_Vertex3f(0.246417999f, -0.0620994568f, 0.f),
    C_Vertex3f(0.249388695f, -0.0546278954f, 0.f),
    C_Vertex3f(0.251935005f, -0.0461316109f, 0.f),
    C_Vertex3f(0.254056931f, -0.0366106033f, 0.f),
    C_Vertex3f(0.255754471f, -0.0260648727f, 0.f),
    C_Vertex3f(0.257027626f, -0.0144944191f, 0.f),
    C_Vertex3f(0.257876396f, -0.0018992424f, 0.f),
    C_Vertex3f(0.258300781f, 0.0117206573f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214645386f, -0.00909042358f, 0.f),
    C_Vertex3f(0.212708473f, -0.0253276825f, 0.f),
    C_Vertex3f(0.213298798f, -0.103025436f, 0.f),
    C_Vertex3f(0.227970123f, -0.0896930695f, 0.f),
    C_Vertex3f(0.21497345f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213298798f, -0.103025436f, 0.f),
    C_Vertex3f(0.212708473f, -0.0253276825f, 0.f),
    C_Vertex3f(0.209490776f, -0.0396213531f, 0.f),
    C_Vertex3f(0.204992294f, -0.0519714355f, 0.f),
    C_Vertex3f(0.199213028f, -0.0623779297f, 0.f),
    C_Vertex3f(0.195999146f, -0.114171982f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191934586f, -0.071387291f, 0.f),
    C_Vertex3f(0.183156967f, -0.0789995193f, 0.f),
    C_Vertex3f(0.186035156f, -0.118652344f, 0.f),
    C_Vertex3f(0.195999146f, -0.114171982f, 0.f),
    C_Vertex3f(0.199213028f, -0.0623779297f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172880173f, -0.0852146149f, 0.f),
    C_Vertex3f(0.166992188f, -0.0876235962f, 0.f),
    C_Vertex3f(0.186035156f, -0.118652344f, 0.f),
    C_Vertex3f(0.183156967f, -0.0789995193f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.21497345f, 0.f, 0.f),
    C_Vertex3f(0.173828125f, 0.0913085938f, 0.f),
    C_Vertex3f(0.173828125f, 0.f, 0.f),
    C_Vertex3f(0.0792312622f, 0.f, 0.f),
    C_Vertex3f(0.0380859375f, 0.0913085938f, 0.f),
    C_Vertex3f(0.0380859375f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122558594f, 0.0913085938f, -1.f),
    C_Vertex3f(0.0380859375f, 0.0913085938f, -1.f),
    C_Vertex3f(0.0792312622f, 0.f, -1.f),
    C_Vertex3f(0.0922279358f, -0.0896930695f, -1.f),
    C_Vertex3f(0.104270935f, -0.074174881f, -1.f),
    C_Vertex3f(0.110675812f, -0.0620994568f, -1.f),
    C_Vertex3f(0.113646507f, -0.0546278954f, -1.f),
    C_Vertex3f(0.116192818f, -0.0461316109f, -1.f),
    C_Vertex3f(0.118314743f, -0.0366106033f, -1.f),
    C_Vertex3f(0.120012283f, -0.0260648727f, -1.f),
    C_Vertex3f(0.121285439f, -0.0144944191f, -1.f),
    C_Vertex3f(0.122134209f, -0.0018992424f, -1.f),
    C_Vertex3f(0.122558594f, 0.0117206573f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0789031982f, -0.00909042358f, -1.f),
    C_Vertex3f(0.0769662857f, -0.0253276825f, -1.f),
    C_Vertex3f(0.0775566101f, -0.103025436f, -1.f),
    C_Vertex3f(0.0922279358f, -0.0896930695f, -1.f),
    C_Vertex3f(0.0792312622f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0775566101f, -0.103025436f, -1.f),
    C_Vertex3f(0.0769662857f, -0.0253276825f, -1.f),
    C_Vertex3f(0.0737485886f, -0.0396213531f, -1.f),
    C_Vertex3f(0.0692501068f, -0.0519714355f, -1.f),
    C_Vertex3f(0.0634708405f, -0.0623779297f, -1.f),
    C_Vertex3f(0.060256958f, -0.114171982f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0561923981f, -0.071387291f, -1.f),
    C_Vertex3f(0.0474147797f, -0.0789995193f, -1.f),
    C_Vertex3f(0.0502929688f, -0.118652344f, -1.f),
    C_Vertex3f(0.060256958f, -0.114171982f, -1.f),
    C_Vertex3f(0.0634708405f, -0.0623779297f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0371379852f, -0.0852146149f, -1.f),
    C_Vertex3f(0.03125f, -0.0876235962f, -1.f),
    C_Vertex3f(0.0502929688f, -0.118652344f, -1.f),
    C_Vertex3f(0.0474147797f, -0.0789995193f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.258300781f, 0.0913085938f, -1.f),
    C_Vertex3f(0.173828125f, 0.0913085938f, -1.f),
    C_Vertex3f(0.21497345f, 0.f, -1.f),
    C_Vertex3f(0.227970123f, -0.0896930695f, -1.f),
    C_Vertex3f(0.240013123f, -0.074174881f, -1.f),
    C_Vertex3f(0.246417999f, -0.0620994568f, -1.f),
    C_Vertex3f(0.249388695f, -0.0546278954f, -1.f),
    C_Vertex3f(0.251935005f, -0.0461316109f, -1.f),
    C_Vertex3f(0.254056931f, -0.0366106033f, -1.f),
    C_Vertex3f(0.255754471f, -0.0260648727f, -1.f),
    C_Vertex3f(0.257027626f, -0.0144944191f, -1.f),
    C_Vertex3f(0.257876396f, -0.0018992424f, -1.f),
    C_Vertex3f(0.258300781f, 0.0117206573f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214645386f, -0.00909042358f, -1.f),
    C_Vertex3f(0.212708473f, -0.0253276825f, -1.f),
    C_Vertex3f(0.213298798f, -0.103025436f, -1.f),
    C_Vertex3f(0.227970123f, -0.0896930695f, -1.f),
    C_Vertex3f(0.21497345f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213298798f, -0.103025436f, -1.f),
    C_Vertex3f(0.212708473f, -0.0253276825f, -1.f),
    C_Vertex3f(0.209490776f, -0.0396213531f, -1.f),
    C_Vertex3f(0.204992294f, -0.0519714355f, -1.f),
    C_Vertex3f(0.199213028f, -0.0623779297f, -1.f),
    C_Vertex3f(0.195999146f, -0.114171982f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191934586f, -0.071387291f, -1.f),
    C_Vertex3f(0.183156967f, -0.0789995193f, -1.f),
    C_Vertex3f(0.186035156f, -0.118652344f, -1.f),
    C_Vertex3f(0.195999146f, -0.114171982f, -1.f),
    C_Vertex3f(0.199213028f, -0.0623779297f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172880173f, -0.0852146149f, -1.f),
    C_Vertex3f(0.166992188f, -0.0876235962f, -1.f),
    C_Vertex3f(0.186035156f, -0.118652344f, -1.f),
    C_Vertex3f(0.183156967f, -0.0789995193f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.21497345f, 0.f, -1.f),
    C_Vertex3f(0.173828125f, 0.0913085938f, -1.f),
    C_Vertex3f(0.173828125f, 0.f, -1.f),
    C_Vertex3f(0.0792312622f, 0.f, -1.f),
    C_Vertex3f(0.0380859375f, 0.0913085938f, -1.f),
    C_Vertex3f(0.0380859375f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_133[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.104492188f, 0.f, 0.f),
    C_Vertex3f(0.104492188f, 0.f, -1.f),
    C_Vertex3f(0.104492188f, 0.0893554688f, -1.f),
    C_Vertex3f(0.104492188f, 0.0893554688f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.104492188f, 0.0893554688f, 0.f),
    C_Vertex3f(0.104492188f, 0.0893554688f, -1.f),
    C_Vertex3f(0.193847656f, 0.0893554688f, -1.f),
    C_Vertex3f(0.193847656f, 0.0893554688f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.193847656f, 0.0893554688f, 0.f),
    C_Vertex3f(0.193847656f, 0.0893554688f, -1.f),
    C_Vertex3f(0.193847656f, 0.f, -1.f),
    C_Vertex3f(0.193847656f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.193847656f, 0.f, 0.f),
    C_Vertex3f(0.193847656f, 0.f, -1.f),
    C_Vertex3f(0.104492188f, 0.f, -1.f),
    C_Vertex3f(0.104492188f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.402832031f, 0.f, 0.f),
    C_Vertex3f(0.402832031f, 0.f, -1.f),
    C_Vertex3f(0.402832031f, 0.0893554688f, -1.f),
    C_Vertex3f(0.402832031f, 0.0893554688f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.402832031f, 0.0893554688f, 0.f),
    C_Vertex3f(0.402832031f, 0.0893554688f, -1.f),
    C_Vertex3f(0.4921875f, 0.0893554688f, -1.f),
    C_Vertex3f(0.4921875f, 0.0893554688f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.4921875f, 0.0893554688f, 0.f),
    C_Vertex3f(0.4921875f, 0.0893554688f, -1.f),
    C_Vertex3f(0.4921875f, 0.f, -1.f),
    C_Vertex3f(0.4921875f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.4921875f, 0.f, 0.f),
    C_Vertex3f(0.4921875f, 0.f, -1.f),
    C_Vertex3f(0.402832031f, 0.f, -1.f),
    C_Vertex3f(0.402832031f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.700683594f, 0.f, 0.f),
    C_Vertex3f(0.700683594f, 0.f, -1.f),
    C_Vertex3f(0.700683594f, 0.0893554688f, -1.f),
    C_Vertex3f(0.700683594f, 0.0893554688f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.700683594f, 0.0893554688f, 0.f),
    C_Vertex3f(0.700683594f, 0.0893554688f, -1.f),
    C_Vertex3f(0.790039063f, 0.0893554688f, -1.f),
    C_Vertex3f(0.790039063f, 0.0893554688f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.790039063f, 0.0893554688f, 0.f),
    C_Vertex3f(0.790039063f, 0.0893554688f, -1.f),
    C_Vertex3f(0.790039063f, 0.f, -1.f),
    C_Vertex3f(0.790039063f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.790039063f, 0.f, 0.f),
    C_Vertex3f(0.790039063f, 0.f, -1.f),
    C_Vertex3f(0.700683594f, 0.f, -1.f),
    C_Vertex3f(0.700683594f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.104492188f, 0.0893554688f, 0.f),
    C_Vertex3f(0.104492188f, 0.f, 0.f),
    C_Vertex3f(0.193847656f, 0.f, 0.f),
    C_Vertex3f(0.193847656f, 0.0893554688f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.700683594f, 0.0893554688f, 0.f),
    C_Vertex3f(0.700683594f, 0.f, 0.f),
    C_Vertex3f(0.790039063f, 0.f, 0.f),
    C_Vertex3f(0.790039063f, 0.0893554688f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402832031f, 0.0893554688f, 0.f),
    C_Vertex3f(0.402832031f, 0.f, 0.f),
    C_Vertex3f(0.4921875f, 0.f, 0.f),
    C_Vertex3f(0.4921875f, 0.0893554688f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.104492188f, 0.0893554688f, -1.f),
    C_Vertex3f(0.104492188f, 0.f, -1.f),
    C_Vertex3f(0.193847656f, 0.f, -1.f),
    C_Vertex3f(0.193847656f, 0.0893554688f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.700683594f, 0.0893554688f, -1.f),
    C_Vertex3f(0.700683594f, 0.f, -1.f),
    C_Vertex3f(0.790039063f, 0.f, -1.f),
    C_Vertex3f(0.790039063f, 0.0893554688f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402832031f, 0.0893554688f, -1.f),
    C_Vertex3f(0.402832031f, 0.f, -1.f),
    C_Vertex3f(0.4921875f, 0.f, -1.f),
    C_Vertex3f(0.4921875f, 0.0893554688f, -1.f),
    C_End(),
    C_Translatef(0.895019531f, 0.f, 0.f),
};
static const Call char_134[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.206054688f, -0.151367188f, 0.f),
    C_Vertex3f(0.206054688f, -0.151367188f, -1.f),
    C_Vertex3f(0.206054688f, 0.37890625f, -1.f),
    C_Vertex3f(0.206054688f, 0.37890625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.206054688f, 0.37890625f, 0.f),
    C_Vertex3f(0.206054688f, 0.37890625f, -1.f),
    C_Vertex3f(0.0317382813f, 0.37890625f, -1.f),
    C_Vertex3f(0.0317382813f, 0.37890625f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0317382813f, 0.37890625f, 0.f),
    C_Vertex3f(0.0317382813f, 0.37890625f, -1.f),
    C_Vertex3f(0.0317382813f, 0.448730469f, -1.f),
    C_Vertex3f(0.0317382813f, 0.448730469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0317382813f, 0.448730469f, 0.f),
    C_Vertex3f(0.0317382813f, 0.448730469f, -1.f),
    C_Vertex3f(0.206054688f, 0.448730469f, -1.f),
    C_Vertex3f(0.206054688f, 0.448730469f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.206054688f, 0.448730469f, 0.f),
    C_Vertex3f(0.206054688f, 0.448730469f, -1.f),
    C_Vertex3f(0.206054688f, 0.625976563f, -1.f),
    C_Vertex3f(0.206054688f, 0.625976563f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.206054688f, 0.625976563f, 0.f),
    C_Vertex3f(0.206054688f, 0.625976563f, -1.f),
    C_Vertex3f(0.284667969f, 0.625976563f, -1.f),
    C_Vertex3f(0.284667969f, 0.625976563f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.284667969f, 0.625976563f, 0.f),
    C_Vertex3f(0.284667969f, 0.625976563f, -1.f),
    C_Vertex3f(0.284667969f, 0.448730469f, -1.f),
    C_Vertex3f(0.284667969f, 0.448730469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.284667969f, 0.448730469f, 0.f),
    C_Vertex3f(0.284667969f, 0.448730469f, -1.f),
    C_Vertex3f(0.460449219f, 0.448730469f, -1.f),
    C_Vertex3f(0.460449219f, 0.448730469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.460449219f, 0.448730469f, 0.f),
    C_Vertex3f(0.460449219f, 0.448730469f, -1.f),
    C_Vertex3f(0.460449219f, 0.37890625f, -1.f),
    C_Vertex3f(0.460449219f, 0.37890625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.460449219f, 0.37890625f, 0.f),
    C_Vertex3f(0.460449219f, 0.37890625f, -1.f),
    C_Vertex3f(0.284667969f, 0.37890625f, -1.f),
    C_Vertex3f(0.284667969f, 0.37890625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.284667969f, 0.37890625f, 0.f),
    C_Vertex3f(0.284667969f, 0.37890625f, -1.f),
    C_Vertex3f(0.284667969f, -0.151367188f, -1.f),
    C_Vertex3f(0.284667969f, -0.151367188f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.284667969f, -0.151367188f, 0.f),
    C_Vertex3f(0.284667969f, -0.151367188f, -1.f),
    C_Vertex3f(0.206054688f, -0.151367188f, -1.f),
    C_Vertex3f(0.206054688f, -0.151367188f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.284667969f, 0.625976563f, 0.f),
    C_Vertex3f(0.206054688f, 0.625976563f, 0.f),
    C_Vertex3f(0.284667969f, -0.151367188f, 0.f),
    C_Vertex3f(0.206054688f, 0.448730469f, 0.f),
    C_Vertex3f(0.206054688f, 0.37890625f, 0.f),
    C_Vertex3f(0.0317382813f, 0.448730469f, 0.f),
    C_Vertex3f(0.0317382813f, 0.37890625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.284667969f, 0.37890625f, 0.f),
    C_Vertex3f(0.460449219f, 0.37890625f, 0.f),
    C_Vertex3f(0.284667969f, 0.448730469f, 0.f),
    C_Vertex3f(0.284667969f, 0.625976563f, 0.f),
    C_Vertex3f(0.284667969f, -0.151367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.284667969f, 0.448730469f, 0.f),
    C_Vertex3f(0.460449219f, 0.37890625f, 0.f),
    C_Vertex3f(0.460449219f, 0.448730469f, 0.f),
    C_Vertex3f(0.284667969f, -0.151367188f, 0.f),
    C_Vertex3f(0.206054688f, 0.37890625f, 0.f),
    C_Vertex3f(0.206054688f, -0.151367188f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.284667969f, 0.625976563f, -1.f),
    C_Vertex3f(0.206054688f, 0.625976563f, -1.f),
    C_Vertex3f(0.284667969f, -0.151367188f, -1.f),
    C_Vertex3f(0.206054688f, 0.448730469f, -1.f),
    C_Vertex3f(0.206054688f, 0.37890625f, -1.f),
    C_Vertex3f(0.0317382813f, 0.448730469f, -1.f),
    C_Vertex3f(0.0317382813f, 0.37890625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.284667969f, 0.37890625f, -1.f),
    C_Vertex3f(0.460449219f, 0.37890625f, -1.f),
    C_Vertex3f(0.284667969f, 0.448730469f, -1.f),
    C_Vertex3f(0.284667969f, 0.625976563f, -1.f),
    C_Vertex3f(0.284667969f, -0.151367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.284667969f, 0.448730469f, -1.f),
    C_Vertex3f(0.460449219f, 0.37890625f, -1.f),
    C_Vertex3f(0.460449219f, 0.448730469f, -1.f),
    C_Vertex3f(0.284667969f, -0.151367188f, -1.f),
    C_Vertex3f(0.206054688f, 0.37890625f, -1.f),
    C_Vertex3f(0.206054688f, -0.151367188f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_135[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.207519531f, -0.151367188f, 0.f),
    C_Vertex3f(0.207519531f, -0.151367188f, -1.f),
    C_Vertex3f(0.207519531f, 0.0102539063f, -1.f),
    C_Vertex3f(0.207519531f, 0.0102539063f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.207519531f, 0.0102539063f, 0.f),
    C_Vertex3f(0.207519531f, 0.0102539063f, -1.f),
    C_Vertex3f(0.0317382813f, 0.0102539063f, -1.f),
    C_Vertex3f(0.0317382813f, 0.0102539063f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0317382813f, 0.0102539063f, 0.f),
    C_Vertex3f(0.0317382813f, 0.0102539063f, -1.f),
    C_Vertex3f(0.0317382813f, 0.0805664063f, -1.f),
    C_Vertex3f(0.0317382813f, 0.0805664063f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0317382813f, 0.0805664063f, 0.f),
    C_Vertex3f(0.0317382813f, 0.0805664063f, -1.f),
    C_Vertex3f(0.207519531f, 0.0805664063f, -1.f),
    C_Vertex3f(0.207519531f, 0.0805664063f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.207519531f, 0.0805664063f, 0.f),
    C_Vertex3f(0.207519531f, 0.0805664063f, -1.f),
    C_Vertex3f(0.207519531f, 0.397460938f, -1.f),
    C_Vertex3f(0.207519531f, 0.397460938f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.207519531f, 0.397460938f, 0.f),
    C_Vertex3f(0.207519531f, 0.397460938f, -1.f),
    C_Vertex3f(0.0317382813f, 0.397460938f, -1.f),
    C_Vertex3f(0.0317382813f, 0.397460938f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0317382813f, 0.397460938f, 0.f),
    C_Vertex3f(0.0317382813f, 0.397460938f, -1.f),
    C_Vertex3f(0.0317382813f, 0.467773438f, -1.f),
    C_Vertex3f(0.0317382813f, 0.467773438f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0317382813f, 0.467773438f, 0.f),
    C_Vertex3f(0.0317382813f, 0.467773438f, -1.f),
    C_Vertex3f(0.207519531f, 0.467773438f, -1.f),
    C_Vertex3f(0.207519531f, 0.467773438f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.207519531f, 0.467773438f, 0.f),
    C_Vertex3f(0.207519531f, 0.467773438f, -1.f),
    C_Vertex3f(0.207519531f, 0.631835938f, -1.f),
    C_Vertex3f(0.207519531f, 0.631835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.207519531f, 0.631835938f, 0.f),
    C_Vertex3f(0.207519531f, 0.631835938f, -1.f),
    C_Vertex3f(0.286132813f, 0.631835938f, -1.f),
    C_Vertex3f(0.286132813f, 0.631835938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.286132813f, 0.631835938f, 0.f),
    C_Vertex3f(0.286132813f, 0.631835938f, -1.f),
    C_Vertex3f(0.286132813f, 0.467773438f, -1.f),
    C_Vertex3f(0.286132813f, 0.467773438f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.286132813f, 0.467773438f, 0.f),
    C_Vertex3f(0.286132813f, 0.467773438f, -1.f),
    C_Vertex3f(0.461914063f, 0.467773438f, -1.f),
    C_Vertex3f(0.461914063f, 0.467773438f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.461914063f, 0.467773438f, 0.f),
    C_Vertex3f(0.461914063f, 0.467773438f, -1.f),
    C_Vertex3f(0.461914063f, 0.397460938f, -1.f),
    C_Vertex3f(0.461914063f, 0.397460938f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.461914063f, 0.397460938f, 0.f),
    C_Vertex3f(0.461914063f, 0.397460938f, -1.f),
    C_Vertex3f(0.286132813f, 0.397460938f, -1.f),
    C_Vertex3f(0.286132813f, 0.397460938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.286132813f, 0.397460938f, 0.f),
    C_Vertex3f(0.286132813f, 0.397460938f, -1.f),
    C_Vertex3f(0.286132813f, 0.0805664063f, -1.f),
    C_Vertex3f(0.286132813f, 0.0805664063f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.286132813f, 0.0805664063f, 0.f),
    C_Vertex3f(0.286132813f, 0.0805664063f, -1.f),
    C_Vertex3f(0.461914063f, 0.0805664063f, -1.f),
    C_Vertex3f(0.461914063f, 0.0805664063f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.461914063f, 0.0805664063f, 0.f),
    C_Vertex3f(0.461914063f, 0.0805664063f, -1.f),
    C_Vertex3f(0.461914063f, 0.0102539063f, -1.f),
    C_Vertex3f(0.461914063f, 0.0102539063f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.461914063f, 0.0102539063f, 0.f),
    C_Vertex3f(0.461914063f, 0.0102539063f, -1.f),
    C_Vertex3f(0.286132813f, 0.0102539063f, -1.f),
    C_Vertex3f(0.286132813f, 0.0102539063f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.286132813f, 0.0102539063f, 0.f),
    C_Vertex3f(0.286132813f, 0.0102539063f, -1.f),
    C_Vertex3f(0.286132813f, -0.151367188f, -1.f),
    C_Vertex3f(0.286132813f, -0.151367188f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.286132813f, -0.151367188f, 0.f),
    C_Vertex3f(0.286132813f, -0.151367188f, -1.f),
    C_Vertex3f(0.207519531f, -0.151367188f, -1.f),
    C_Vertex3f(0.207519531f, -0.151367188f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0317382813f, 0.0805664063f, 0.f),
    C_Vertex3f(0.207519531f, 0.0102539063f, 0.f),
    C_Vertex3f(0.207519531f, 0.0805664063f, 0.f),
    C_Vertex3f(0.286132813f, -0.151367188f, 0.f),
    C_Vertex3f(0.207519531f, 0.397460938f, 0.f),
    C_Vertex3f(0.207519531f, 0.467773438f, 0.f),
    C_Vertex3f(0.0317382813f, 0.467773438f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.286132813f, 0.631835938f, 0.f),
    C_Vertex3f(0.207519531f, 0.631835938f, 0.f),
    C_Vertex3f(0.286132813f, -0.151367188f, 0.f),
    C_Vertex3f(0.286132813f, 0.0102539063f, 0.f),
    C_Vertex3f(0.286132813f, 0.0805664063f, 0.f),
    C_Vertex3f(0.286132813f, 0.397460938f, 0.f),
    C_Vertex3f(0.286132813f, 0.467773438f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.286132813f, 0.467773438f, 0.f),
    C_Vertex3f(0.286132813f, 0.397460938f, 0.f),
    C_Vertex3f(0.461914063f, 0.397460938f, 0.f),
    C_Vertex3f(0.461914063f, 0.467773438f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.286132813f, 0.0805664063f, 0.f),
    C_Vertex3f(0.286132813f, 0.0102539063f, 0.f),
    C_Vertex3f(0.461914063f, 0.0102539063f, 0.f),
    C_Vertex3f(0.461914063f, 0.0805664063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.286132813f, -0.151367188f, 0.f),
    C_Vertex3f(0.207519531f, 0.0102539063f, 0.f),
    C_Vertex3f(0.207519531f, -0.151367188f, 0.f),
    C_Vertex3f(0.207519531f, 0.467773438f, 0.f),
    C_Vertex3f(0.286132813f, -0.151367188f, 0.f),
    C_Vertex3f(0.207519531f, 0.631835938f, 0.f),
    C_Vertex3f(0.207519531f, 0.397460938f, 0.f),
    C_Vertex3f(0.0317382813f, 0.467773438f, 0.f),
    C_Vertex3f(0.0317382813f, 0.397460938f, 0.f),
    C_Vertex3f(0.207519531f, 0.0102539063f, 0.f),
    C_Vertex3f(0.0317382813f, 0.0805664063f, 0.f),
    C_Vertex3f(0.0317382813f, 0.0102539063f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0317382813f, 0.0805664063f, -1.f),
    C_Vertex3f(0.207519531f, 0.0102539063f, -1.f),
    C_Vertex3f(0.207519531f, 0.0805664063f, -1.f),
    C_Vertex3f(0.286132813f, -0.151367188f, -1.f),
    C_Vertex3f(0.207519531f, 0.397460938f, -1.f),
    C_Vertex3f(0.207519531f, 0.467773438f, -1.f),
    C_Vertex3f(0.0317382813f, 0.467773438f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.286132813f, 0.631835938f, -1.f),
    C_Vertex3f(0.207519531f, 0.631835938f, -1.f),
    C_Vertex3f(0.286132813f, -0.151367188f, -1.f),
    C_Vertex3f(0.286132813f, 0.0102539063f, -1.f),
    C_Vertex3f(0.286132813f, 0.0805664063f, -1.f),
    C_Vertex3f(0.286132813f, 0.397460938f, -1.f),
    C_Vertex3f(0.286132813f, 0.467773438f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.286132813f, 0.467773438f, -1.f),
    C_Vertex3f(0.286132813f, 0.397460938f, -1.f),
    C_Vertex3f(0.461914063f, 0.397460938f, -1.f),
    C_Vertex3f(0.461914063f, 0.467773438f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.286132813f, 0.0805664063f, -1.f),
    C_Vertex3f(0.286132813f, 0.0102539063f, -1.f),
    C_Vertex3f(0.461914063f, 0.0102539063f, -1.f),
    C_Vertex3f(0.461914063f, 0.0805664063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.286132813f, -0.151367188f, -1.f),
    C_Vertex3f(0.207519531f, 0.0102539063f, -1.f),
    C_Vertex3f(0.207519531f, -0.151367188f, -1.f),
    C_Vertex3f(0.207519531f, 0.467773438f, -1.f),
    C_Vertex3f(0.286132813f, -0.151367188f, -1.f),
    C_Vertex3f(0.207519531f, 0.631835938f, -1.f),
    C_Vertex3f(0.207519531f, 0.397460938f, -1.f),
    C_Vertex3f(0.0317382813f, 0.467773438f, -1.f),
    C_Vertex3f(0.0317382813f, 0.397460938f, -1.f),
    C_Vertex3f(0.207519531f, 0.0102539063f, -1.f),
    C_Vertex3f(0.0317382813f, 0.0805664063f, -1.f),
    C_Vertex3f(0.0317382813f, 0.0102539063f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_136[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.149414063f, 0.596191406f, 0.f),
    C_Vertex3f(0.149414063f, 0.596191406f, -1.f),
    C_Vertex3f(0.100341797f, 0.521484375f, -1.f),
    C_Vertex3f(0.100341797f, 0.521484375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.100341797f, 0.521484375f, 0.f),
    C_Vertex3f(0.100341797f, 0.521484375f, -1.f),
    C_Vertex3f(0.0109863281f, 0.521484375f, -1.f),
    C_Vertex3f(0.0109863281f, 0.521484375f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(0.0109863281f, 0.521484375f, 0.f),
    C_Vertex3f(0.0109863281f, 0.521484375f, -1.f),
    C_Vertex3f(0.104492188f, 0.644042969f, -1.f),
    C_Vertex3f(0.104492188f, 0.644042969f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.104492188f, 0.644042969f, 0.f),
    C_Vertex3f(0.104492188f, 0.644042969f, -1.f),
    C_Vertex3f(0.188476563f, 0.644042969f, -1.f),
    C_Vertex3f(0.188476563f, 0.644042969f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.188476563f, 0.644042969f, 0.f),
    C_Vertex3f(0.188476563f, 0.644042969f, -1.f),
    C_Vertex3f(0.286865234f, 0.521484375f, -1.f),
    C_Vertex3f(0.286865234f, 0.521484375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.286865234f, 0.521484375f, 0.f),
    C_Vertex3f(0.286865234f, 0.521484375f, -1.f),
    C_Vertex3f(0.197509766f, 0.521484375f, -1.f),
    C_Vertex3f(0.197509766f, 0.521484375f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.197509766f, 0.521484375f, 0.f),
    C_Vertex3f(0.197509766f, 0.521484375f, -1.f),
    C_Vertex3f(0.149414063f, 0.596191406f, -1.f),
    C_Vertex3f(0.149414063f, 0.596191406f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.149414063f, 0.596191406f, 0.f),
    C_Vertex3f(0.197509766f, 0.521484375f, 0.f),
    C_Vertex3f(0.286865234f, 0.521484375f, 0.f),
    C_Vertex3f(0.188476563f, 0.644042969f, 0.f),
    C_Vertex3f(0.104492188f, 0.644042969f, 0.f),
    C_Vertex3f(0.0109863281f, 0.521484375f, 0.f),
    C_Vertex3f(0.100341797f, 0.521484375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.149414063f, 0.596191406f, -1.f),
    C_Vertex3f(0.197509766f, 0.521484375f, -1.f),
    C_Vertex3f(0.286865234f, 0.521484375f, -1.f),
    C_Vertex3f(0.188476563f, 0.644042969f, -1.f),
    C_Vertex3f(0.104492188f, 0.644042969f, -1.f),
    C_Vertex3f(0.0109863281f, 0.521484375f, -1.f),
    C_Vertex3f(0.100341797f, 0.521484375f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_137[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.931913316f, 0.362680972f, 0.f),
    C_Vertex3f(0.140045166f, -0.0234375f, 0.f),
    C_Vertex3f(0.140045166f, -0.0234375f, -1.f),
    C_Vertex3f(0.402854919f, 0.651855469f, -1.f),
    C_Vertex3f(0.402854919f, 0.651855469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.402854919f, 0.651855469f, 0.f),
    C_Vertex3f(0.402854919f, 0.651855469f, -1.f),
    C_Vertex3f(0.460548401f, 0.651855469f, -1.f),
    C_Vertex3f(0.460548401f, 0.651855469f, 0.f),
    C_Normal3f(0.931913316f, -0.362680972f, 0.f),
    C_Vertex3f(0.460548401f, 0.651855469f, 0.f),
    C_Vertex3f(0.460548401f, 0.651855469f, -1.f),
    C_Vertex3f(0.197738647f, -0.0234375f, -1.f),
    C_Vertex3f(0.197738647f, -0.0234375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.197738647f, -0.0234375f, 0.f),
    C_Vertex3f(0.197738647f, -0.0234375f, -1.f),
    C_Vertex3f(0.140045166f, -0.0234375f, -1.f),
    C_Vertex3f(0.140045166f, -0.0234375f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999676049f, -0.0254525878f, 0.f),
    C_Vertex3f(0.0163116455f, 0.469232559f, 0.f),
    C_Vertex3f(0.0163116455f, 0.469232559f, -1.f),
    C_Normal3f(-0.999679029f, 0.0253341328f, 0.f),
    C_Vertex3f(0.0163116455f, 0.491300583f, 0.f),
    C_Vertex3f(0.0163116455f, 0.491300583f, -1.f),
    C_Normal3f(-0.996619225f, 0.0821585581f, 0.f),
    C_Vertex3f(0.0173842907f, 0.51155889f, 0.f),
    C_Vertex3f(0.0173842907f, 0.51155889f, -1.f),
    C_Normal3f(-0.989126623f, 0.147065863f, 0.f),
    C_Vertex3f(0.0195295811f, 0.530335546f, 0.f),
    C_Vertex3f(0.0195295811f, 0.530335546f, -1.f),
    C_Normal3f(-0.975235343f, 0.221169874f, 0.f),
    C_Vertex3f(0.0227475166f, 0.547630548f, 0.f),
    C_Vertex3f(0.0227475166f, 0.547630548f, -1.f),
    C_Normal3f(-0.952356279f, 0.30498752f, 0.f),
    C_Vertex3f(0.0270380974f, 0.563443899f, 0.f),
    C_Vertex3f(0.0270380974f, 0.563443899f, -1.f),
    C_Normal3f(-0.917304993f, 0.398185283f, 0.f),
    C_Vertex3f(0.0324013233f, 0.577775598f, 0.f),
    C_Vertex3f(0.0324013233f, 0.577775598f, -1.f),
    C_Normal3f(-0.866615653f, 0.498976171f, 0.f),
    C_Vertex3f(0.0388371944f, 0.590625644f, 0.f),
    C_Vertex3f(0.0388371944f, 0.590625644f, -1.f),
    C_Normal3f(-0.797321379f, 0.603555024f, 0.f),
    C_Vertex3f(0.0463457108f, 0.601994038f, 0.f),
    C_Vertex3f(0.0463457108f, 0.601994038f, -1.f),
    C_Normal3f(-0.720296085f, 0.693666816f, 0.f),
    C_Vertex3f(0.0549268723f, 0.611880779f, 0.f),
    C_Vertex3f(0.0549268723f, 0.611880779f, -1.f),
    C_Normal3f(-0.641103804f, 0.767454147f, 0.f),
    C_Vertex3f(0.0641980171f, 0.620531678f, 0.f),
    C_Vertex3f(0.0641980171f, 0.620531678f, -1.f),
    C_Normal3f(-0.550565481f, 0.834792018f, 0.f),
    C_Vertex3f(0.0741591454f, 0.627946734f, 0.f),
    C_Vertex3f(0.0741591454f, 0.627946734f, -1.f),
    C_Normal3f(-0.451314569f, 0.89236486f, 0.f),
    C_Vertex3f(0.084810257f, 0.634125948f, 0.f),
    C_Vertex3f(0.084810257f, 0.634125948f, -1.f),
    C_Normal3f(-0.347121656f, 0.937820137f, 0.f),
    C_Vertex3f(0.0961513519f, 0.639069319f, 0.f),
    C_Vertex3f(0.0961513519f, 0.639069319f, -1.f),
    C_Normal3f(-0.242210165f, 0.970223784f, 0.f),
    C_Vertex3f(0.10818243f, 0.642776847f, 0.f),
    C_Vertex3f(0.10818243f, 0.642776847f, -1.f),
    C_Normal3f(-0.140469655f, 0.990085006f, 0.f),
    C_Vertex3f(0.120903492f, 0.645248532f, 0.f),
    C_Vertex3f(0.120903492f, 0.645248532f, -1.f),
    C_Normal3f(-0.0449649841f, 0.998988628f, 0.f),
    C_Vertex3f(0.134314537f, 0.646484375f, 0.f),
    C_Vertex3f(0.134314537f, 0.646484375f, -1.f),
    C_Normal3f(0.0453374274f, 0.99897176f, 0.f),
    C_Vertex3f(0.148360252f, 0.646484375f, 0.f),
    C_Vertex3f(0.148360252f, 0.646484375f, -1.f),
    C_Normal3f(0.141855732f, 0.989887357f, 0.f),
    C_Vertex3f(0.161695004f, 0.645241737f, 0.f),
    C_Vertex3f(0.161695004f, 0.645241737f, -1.f),
    C_Normal3f(0.243922889f, 0.969794691f, 0.f),
    C_Vertex3f(0.174374104f, 0.642756462f, 0.f),
    C_Vertex3f(0.174374104f, 0.642756462f, -1.f),
    C_Normal3f(0.348548681f, 0.937290668f, 0.f),
    C_Vertex3f(0.186397552f, 0.639028549f, 0.f),
    C_Vertex3f(0.186397552f, 0.639028549f, -1.f),
    C_Normal3f(0.451854408f, 0.892091691f, 0.f),
    C_Vertex3f(0.19776535f, 0.634057999f, 0.f),
    C_Vertex3f(0.19776535f, 0.634057999f, -1.f),
    C_Normal3f(0.549768984f, 0.835316718f, 0.f),
    C_Vertex3f(0.208477497f, 0.62784481f, 0.f),
    C_Vertex3f(0.208477497f, 0.62784481f, -1.f),
    C_Normal3f(0.638779402f, 0.769389927f, 0.f),
    C_Vertex3f(0.218533993f, 0.620388985f, 0.f),
    C_Vertex3f(0.218533993f, 0.620388985f, -1.f),
    C_Normal3f(0.716529012f, 0.697557211f, 0.f),
    C_Vertex3f(0.227934837f, 0.611690521f, 0.f),
    C_Vertex3f(0.227934837f, 0.611690521f, -1.f),
    C_Normal3f(0.792242885f, 0.610205889f, 0.f),
    C_Vertex3f(0.236680031f, 0.60174942f, 0.f),
    C_Vertex3f(0.236680031f, 0.60174942f, -1.f),
    C_Normal3f(0.861415744f, 0.507900596f, 0.f),
    C_Vertex3f(0.244332075f, 0.590401649f, 0.f),
    C_Vertex3f(0.244332075f, 0.590401649f, -1.f),
    C_Normal3f(0.913027048f, 0.407898873f, 0.f),
    C_Vertex3f(0.25089097f, 0.577647209f, 0.f),
    C_Vertex3f(0.25089097f, 0.577647209f, -1.f),
    C_Normal3f(0.949363947f, 0.314178169f, 0.f),
    C_Vertex3f(0.256356716f, 0.563486099f, 0.f),
    C_Vertex3f(0.256356716f, 0.563486099f, -1.f),
    C_Normal3f(0.973450243f, 0.228899017f, 0.f),
    C_Vertex3f(0.260729313f, 0.54791832f, 0.f),
    C_Vertex3f(0.260729313f, 0.54791832f, -1.f),
    C_Normal3f(0.988260448f, 0.152778774f, 0.f),
    C_Vertex3f(0.26400876f, 0.530943871f, 0.f),
    C_Vertex3f(0.26400876f, 0.530943871f, -1.f),
    C_Normal3f(0.996329188f, 0.0856039524f, 0.f),
    C_Vertex3f(0.266195059f, 0.512562752f, 0.f),
    C_Vertex3f(0.266195059f, 0.512562752f, -1.f),
    C_Normal3f(0.999646306f, 0.0265937429f, 0.f),
    C_Vertex3f(0.267288208f, 0.492774963f, 0.f),
    C_Vertex3f(0.267288208f, 0.492774963f, -1.f),
    C_Normal3f(0.999733269f, -0.0230971649f, 0.f),
    C_Vertex3f(0.267288208f, 0.471471786f, 0.f),
    C_Vertex3f(0.267288208f, 0.471471786f, -1.f),
    C_Normal3f(0.997286737f, -0.0736151561f, 0.f),
    C_Vertex3f(0.266331792f, 0.45137763f, 0.f),
    C_Vertex3f(0.266331792f, 0.45137763f, -1.f),
    C_Normal3f(0.991403639f, -0.130839244f, 0.f),
    C_Vertex3f(0.26441896f, 0.432601213f, 0.f),
    C_Vertex3f(0.26441896f, 0.432601213f, -1.f),
    C_Normal3f(0.980720758f, -0.195414141f, 0.f),
    C_Vertex3f(0.261549711f, 0.415142536f, 0.f),
    C_Vertex3f(0.261549711f, 0.415142536f, -1.f),
    C_Normal3f(0.963464975f, -0.267834395f, 0.f),
    C_Vertex3f(0.257724047f, 0.399001598f, 0.f),
    C_Vertex3f(0.257724047f, 0.399001598f, -1.f),
    C_Normal3f(0.937441409f, -0.348143041f, 0.f),
    C_Vertex3f(0.252941966f, 0.3841784f, 0.f),
    C_Vertex3f(0.252941966f, 0.3841784f, -1.f),
    C_Normal3f(0.900142372f, -0.435595721f, 0.f),
    C_Vertex3f(0.247203469f, 0.370672941f, 0.f),
    C_Vertex3f(0.247203469f, 0.370672941f, -1.f),
    C_Normal3f(0.849069357f, -0.52828145f, 0.f),
    C_Vertex3f(0.240508556f, 0.358485222f, 0.f),
    C_Vertex3f(0.240508556f, 0.358485222f, -1.f),
    C_Normal3f(0.780189455f, -0.625543356f, 0.f),
    C_Vertex3f(0.232857227f, 0.347615242f, 0.f),
    C_Vertex3f(0.232857227f, 0.347615242f, -1.f),
    C_Normal3f(0.692160487f, -0.721743584f, 0.f),
    C_Vertex3f(0.224167228f, 0.33810401f, 0.f),
    C_Vertex3f(0.224167228f, 0.33810401f, -1.f),
    C_Normal3f(0.589203358f, -0.80798471f, 0.f),
    C_Vertex3f(0.214438558f, 0.329951525f, 0.f),
    C_Vertex3f(0.214438558f, 0.329951525f, -1.f),
    C_Normal3f(0.476328164f, -0.879267573f, 0.f),
    C_Vertex3f(0.203671217f, 0.323157787f, 0.f),
    C_Vertex3f(0.203671217f, 0.323157787f, -1.f),
    C_Normal3f(0.359974474f, -0.932962179f, 0.f),
    C_Vertex3f(0.191865206f, 0.317722797f, 0.f),
    C_Vertex3f(0.191865206f, 0.317722797f, -1.f),
    C_Normal3f(0.246346816f, -0.969181716f, 0.f),
    C_Vertex3f(0.179020524f, 0.313646555f, 0.f),
    C_Vertex3f(0.179020524f, 0.313646555f, -1.f),
    C_Normal3f(0.140113831f, -0.990135312f, 0.f),
    C_Vertex3f(0.165137172f, 0.31092906f, 0.f),
    C_Vertex3f(0.165137172f, 0.31092906f, -1.f),
    C_Normal3f(0.0453998931f, -0.998968959f, 0.f),
    C_Vertex3f(0.150215149f, 0.309570313f, 0.f),
    C_Vertex3f(0.150215149f, 0.309570313f, -1.f),
    C_Normal3f(-0.0448126942f, -0.998995423f, 0.f),
    C_Vertex3f(0.135239601f, 0.309570313f, 0.f),
    C_Vertex3f(0.135239601f, 0.309570313f, -1.f),
    C_Normal3f(-0.14490734f, -0.989445269f, 0.f),
    C_Vertex3f(0.12190485f, 0.310840249f, 0.f),
    C_Vertex3f(0.12190485f, 0.310840249f, -1.f),
    C_Normal3f(-0.248952717f, -0.968515575f, 0.f),
    C_Vertex3f(0.10922575f, 0.313380122f, 0.f),
    C_Vertex3f(0.10922575f, 0.313380122f, -1.f),
    C_Normal3f(-0.3552486f, -0.934771955f, 0.f),
    C_Vertex3f(0.097202301f, 0.317189932f, 0.f),
    C_Vertex3f(0.097202301f, 0.317189932f, -1.f),
    C_Normal3f(-0.459701121f, -0.888073683f, 0.f),
    C_Vertex3f(0.0858345032f, 0.322269678f, 0.f),
    C_Vertex3f(0.0858345032f, 0.322269678f, -1.f),
    C_Normal3f(-0.558112562f, -0.829765201f, 0.f),
    C_Vertex3f(0.0751223564f, 0.328619361f, 0.f),
    C_Vertex3f(0.0751223564f, 0.328619361f, -1.f),
    C_Normal3f(-0.646975517f, -0.762510717f, 0.f),
    C_Vertex3f(0.0650658607f, 0.33623898f, 0.f),
    C_Vertex3f(0.0650658607f, 0.33623898f, -1.f),
    C_Normal3f(-0.724060357f, -0.689736605f, 0.f),
    C_Vertex3f(0.0556650162f, 0.345128536f, 0.f),
    C_Vertex3f(0.0556650162f, 0.345128536f, -1.f),
    C_Normal3f(-0.799809873f, -0.600253403f, 0.f),
    C_Vertex3f(0.0469198227f, 0.355288029f, 0.f),
    C_Vertex3f(0.0469198227f, 0.355288029f, -1.f),
    C_Normal3f(-0.868573964f, -0.495559514f, 0.f),
    C_Vertex3f(0.0392677784f, 0.366977096f, 0.f),
    C_Vertex3f(0.0392677784f, 0.366977096f, -1.f),
    C_Normal3f(-0.918682396f, -0.39499706f, 0.f),
    C_Vertex3f(0.0327088833f, 0.380195737f, 0.f),
    C_Vertex3f(0.0327088833f, 0.380195737f, -1.f),
    C_Normal3f(-0.953223348f, -0.302267015f, 0.f),
    C_Vertex3f(0.0272431374f, 0.394943953f, 0.f),
    C_Vertex3f(0.0272431374f, 0.394943953f, -1.f),
    C_Normal3f(-0.975714564f, -0.219045833f, 0.f),
    C_Vertex3f(0.0228705406f, 0.411221743f, 0.f),
    C_Vertex3f(0.0228705406f, 0.411221743f, -1.f),
    C_Normal3f(-0.989346325f, -0.14558129f, 0.f),
    C_Vertex3f(0.0195910931f, 0.429029107f, 0.f),
    C_Vertex3f(0.0195910931f, 0.429029107f, -1.f),
    C_Normal3f(-0.996689558f, -0.081301257f, 0.f),
    C_Vertex3f(0.0174047947f, 0.448366046f, 0.f),
    C_Vertex3f(0.0174047947f, 0.448366046f, -1.f),
    C_Normal3f(-0.999676049f, -0.0254525878f, 0.f),
    C_Vertex3f(0.0163116455f, 0.469232559f, 0.f),
    C_Vertex3f(0.0163116455f, 0.469232559f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999818802f, -0.0190345086f, 0.f),
    C_Vertex3f(0.0807647705f, 0.485423088f, 0.f),
    C_Vertex3f(0.0807647705f, 0.485423088f, -1.f),
    C_Normal3f(0.999856651f, 0.016930189f, 0.f),
    C_Vertex3f(0.0807647705f, 0.468659401f, 0.f),
    C_Vertex3f(0.0807647705f, 0.468659401f, -1.f),
    C_Normal3f(0.998447061f, 0.0557081886f, 0.f),
    C_Vertex3f(0.0813068151f, 0.453411222f, 0.f),
    C_Vertex3f(0.0813068151f, 0.453411222f, -1.f),
    C_Normal3f(0.994788229f, 0.101962976f, 0.f),
    C_Vertex3f(0.0823909044f, 0.439514518f, 0.f),
    C_Vertex3f(0.0823909044f, 0.439514518f, -1.f),
    C_Normal3f(0.987466037f, 0.157831326f, 0.f),
    C_Vertex3f(0.0840170383f, 0.42696929f, 0.f),
    C_Vertex3f(0.0840170383f, 0.42696929f, -1.f),
    C_Normal3f(0.974147856f, 0.225911677f, 0.f),
    C_Vertex3f(0.0861852169f, 0.415775537f, 0.f),
    C_Vertex3f(0.0861852169f, 0.415775537f, -1.f),
    C_Normal3f(0.950969338f, 0.309285015f, 0.f),
    C_Vertex3f(0.0888954401f, 0.405933261f, 0.f),
    C_Vertex3f(0.0888954401f, 0.405933261f, -1.f),
    C_Normal3f(0.911637008f, 0.410996258f, 0.f),
    C_Vertex3f(0.0921477079f, 0.39744246f, 0.f),
    C_Vertex3f(0.0921477079f, 0.39744246f, -1.f),
    C_Normal3f(0.838077068f, 0.545551777f, 0.f),
    C_Vertex3f(0.0959420204f, 0.390303135f, 0.f),
    C_Vertex3f(0.0959420204f, 0.390303135f, -1.f),
    C_Normal3f(0.723304152f, 0.690529644f, 0.f),
    C_Vertex3f(0.102446556f, 0.381621361f, 0.f),
    C_Vertex3f(0.102446556f, 0.381621361f, -1.f),
    C_Normal3f(0.55755049f, 0.830143034f, 0.f),
    C_Vertex3f(0.112518787f, 0.372939587f, 0.f),
    C_Vertex3f(0.112518787f, 0.372939587f, -1.f),
    C_Normal3f(0.335919112f, 0.941890895f, 0.f),
    C_Vertex3f(0.123990536f, 0.367151737f, 0.f),
    C_Vertex3f(0.123990536f, 0.367151737f, -1.f),
    C_Normal3f(0.108631797f, 0.994082093f, 0.f),
    C_Vertex3f(0.136861801f, 0.364257813f, 0.f),
    C_Vertex3f(0.136861801f, 0.364257813f, -1.f),
    C_Normal3f(-0.0975214392f, 0.995233357f, 0.f),
    C_Vertex3f(0.150472641f, 0.364257813f, 0.f),
    C_Vertex3f(0.150472641f, 0.364257813f, -1.f),
    C_Normal3f(-0.324939996f, 0.94573468f, 0.f),
    C_Vertex3f(0.161942482f, 0.366715431f, 0.f),
    C_Vertex3f(0.161942482f, 0.366715431f, -1.f),
    C_Normal3f(-0.553361595f, 0.832941234f, 0.f),
    C_Vertex3f(0.171931267f, 0.371630669f, 0.f),
    C_Vertex3f(0.171931267f, 0.371630669f, -1.f),
    C_Normal3f(-0.734458268f, 0.678653836f, 0.f),
    C_Vertex3f(0.180438995f, 0.379003525f, 0.f),
    C_Vertex3f(0.180438995f, 0.379003525f, -1.f),
    C_Normal3f(-0.848488271f, 0.529214263f, 0.f),
    C_Vertex3f(0.186312675f, 0.387194633f, 0.f),
    C_Vertex3f(0.186312675f, 0.387194633f, -1.f),
    C_Normal3f(-0.911447227f, 0.411417156f, 0.f),
    C_Vertex3f(0.190443277f, 0.395043373f, 0.f),
    C_Vertex3f(0.190443277f, 0.395043373f, -1.f),
    C_Normal3f(-0.949242175f, 0.314546227f, 0.f),
    C_Vertex3f(0.193983793f, 0.40418911f, 0.f),
    C_Vertex3f(0.193983793f, 0.40418911f, -1.f),
    C_Normal3f(-0.972516656f, 0.232833281f, 0.f),
    C_Vertex3f(0.196934223f, 0.414631844f, 0.f),
    C_Vertex3f(0.196934223f, 0.414631844f, -1.f),
    C_Normal3f(-0.986386299f, 0.164445177f, 0.f),
    C_Vertex3f(0.199294567f, 0.426371574f, 0.f),
    C_Vertex3f(0.199294567f, 0.426371574f, -1.f),
    C_Normal3f(-0.994240165f, 0.107175283f, 0.f),
    C_Vertex3f(0.201064825f, 0.439408302f, 0.f),
    C_Vertex3f(0.201064825f, 0.439408302f, -1.f),
    C_Normal3f(-0.998259366f, 0.0589757748f, 0.f),
    C_Vertex3f(0.202244997f, 0.453742027f, 0.f),
    C_Vertex3f(0.202244997f, 0.453742027f, -1.f),
    C_Normal3f(-0.99983412f, 0.0182126518f, 0.f),
    C_Vertex3f(0.202835083f, 0.469372749f, 0.f),
    C_Vertex3f(0.202835083f, 0.469372749f, -1.f),
    C_Normal3f(-0.999823809f, -0.018774502f, 0.f),
    C_Vertex3f(0.202835083f, 0.486136436f, 0.f),
    C_Vertex3f(0.202835083f, 0.486136436f, -1.f),
    C_Normal3f(-0.998127759f, -0.0611631051f, 0.f),
    C_Vertex3f(0.202231407f, 0.501521111f, 0.f),
    C_Vertex3f(0.202231407f, 0.501521111f, -1.f),
    C_Normal3f(-0.993865371f, -0.110596441f, 0.f),
    C_Vertex3f(0.201024055f, 0.515690804f, 0.f),
    C_Vertex3f(0.201024055f, 0.515690804f, -1.f),
    C_Normal3f(-0.985672772f, -0.16866903f, 0.f),
    C_Vertex3f(0.199213028f, 0.528645515f, 0.f),
    C_Vertex3f(0.199213028f, 0.528645515f, -1.f),
    C_Normal3f(-0.971493006f, -0.237068266f, 0.f),
    C_Vertex3f(0.196798325f, 0.540385246f, 0.f),
    C_Vertex3f(0.196798325f, 0.540385246f, -1.f),
    C_Normal3f(-0.948267639f, -0.317472219f, 0.f),
    C_Vertex3f(0.193779945f, 0.550909996f, 0.f),
    C_Vertex3f(0.193779945f, 0.550909996f, -1.f),
    C_Normal3f(-0.911612749f, -0.411049932f, 0.f),
    C_Vertex3f(0.19015789f, 0.560219765f, 0.f),
    C_Vertex3f(0.19015789f, 0.560219765f, -1.f),
    C_Normal3f(-0.866646886f, -0.49892205f, 0.f),
    C_Vertex3f(0.185932159f, 0.568314552f, 0.f),
    C_Vertex3f(0.185932159f, 0.568314552f, -1.f),
    C_Normal3f(-0.812650859f, -0.582750857f, 0.f),
    C_Vertex3f(0.183517456f, 0.571754456f, 0.f),
    C_Vertex3f(0.183517456f, 0.571754456f, -1.f),
    C_Normal3f(-0.712831557f, -0.701335132f, 0.f),
    C_Vertex3f(0.179784775f, 0.576887131f, 0.f),
    C_Vertex3f(0.179784775f, 0.576887131f, -1.f),
    C_Normal3f(-0.555433512f, -0.831560969f, 0.f),
    C_Vertex3f(0.170892715f, 0.584586143f, 0.f),
    C_Vertex3f(0.170892715f, 0.584586143f, -1.f),
    C_Normal3f(-0.329455912f, -0.944170952f, 0.f),
    C_Vertex3f(0.160573959f, 0.589718819f, 0.f),
    C_Vertex3f(0.160573959f, 0.589718819f, -1.f),
    C_Normal3f(-0.101558663f, -0.994829535f, 0.f),
    C_Vertex3f(0.148828506f, 0.592285156f, 0.f),
    C_Vertex3f(0.148828506f, 0.592285156f, -1.f),
    C_Normal3f(0.0983575284f, -0.995151162f, 0.f),
    C_Vertex3f(0.135435104f, 0.592285156f, 0.f),
    C_Vertex3f(0.135435104f, 0.592285156f, -1.f),
    C_Normal3f(0.318208337f, -0.948020816f, 0.f),
    C_Vertex3f(0.123413086f, 0.589773178f, 0.f),
    C_Vertex3f(0.123413086f, 0.589773178f, -1.f),
    C_Normal3f(0.546115994f, -0.837709546f, 0.f),
    C_Vertex3f(0.112983704f, 0.584749222f, 0.f),
    C_Vertex3f(0.112983704f, 0.584749222f, -1.f),
    C_Normal3f(0.731153846f, -0.68221271f, 0.f),
    C_Vertex3f(0.104146957f, 0.577213287f, 0.f),
    C_Vertex3f(0.104146957f, 0.577213287f, -1.f),
    C_Normal3f(0.847083867f, -0.531459212f, 0.f),
    C_Vertex3f(0.0980548859f, 0.568749428f, 0.f),
    C_Vertex3f(0.0980548859f, 0.568749428f, -1.f),
    C_Normal3f(0.909181118f, -0.41640088f, 0.f),
    C_Vertex3f(0.093732357f, 0.560613632f, 0.f),
    C_Vertex3f(0.093732357f, 0.560613632f, -1.f),
    C_Normal3f(0.94707346f, -0.321017027f, 0.f),
    C_Vertex3f(0.0900273323f, 0.551221848f, 0.f),
    C_Vertex3f(0.0900273323f, 0.551221848f, -1.f),
    C_Normal3f(0.970951021f, -0.239278421f, 0.f),
    C_Vertex3f(0.0869398117f, 0.540574074f, 0.f),
    C_Vertex3f(0.0869398117f, 0.540574074f, -1.f),
    C_Normal3f(0.985451996f, -0.169954002f, 0.f),
    C_Vertex3f(0.0844697952f, 0.528670311f, 0.f),
    C_Vertex3f(0.0844697952f, 0.528670311f, -1.f),
    C_Normal3f(0.99379009f, -0.111270808f, 0.f),
    C_Vertex3f(0.0826172829f, 0.515510559f, 0.f),
    C_Vertex3f(0.0826172829f, 0.515510559f, -1.f),
    C_Normal3f(0.998109937f, -0.0614545159f, 0.f),
    C_Vertex3f(0.0813822746f, 0.501094818f, 0.f),
    C_Vertex3f(0.0813822746f, 0.501094818f, -1.f),
    C_Normal3f(0.999818802f, -0.0190345086f, 0.f),
    C_Vertex3f(0.0807647705f, 0.485423088f, 0.f),
    C_Vertex3f(0.0807647705f, 0.485423088f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99967587f, -0.0254604295f, 0.f),
    C_Vertex3f(0.334182739f, 0.135761261f, 0.f),
    C_Vertex3f(0.334182739f, 0.135761261f, -1.f),
    C_Normal3f(-0.999679029f, 0.0253369864f, 0.f),
    C_Vertex3f(0.334182739f, 0.157826424f, 0.f),
    C_Vertex3f(0.334182739f, 0.157826424f, -1.f),
    C_Normal3f(-0.996618569f, 0.0821665376f, 0.f),
    C_Vertex3f(0.335255384f, 0.178082824f, 0.f),
    C_Vertex3f(0.335255384f, 0.178082824f, -1.f),
    C_Normal3f(-0.989124417f, 0.147081092f, 0.f),
    C_Vertex3f(0.337400675f, 0.196857572f, 0.f),
    C_Vertex3f(0.337400675f, 0.196857572f, -1.f),
    C_Normal3f(-0.975229859f, 0.221194118f, 0.f),
    C_Vertex3f(0.34061861f, 0.214150667f, 0.f),
    C_Vertex3f(0.34061861f, 0.214150667f, -1.f),
    C_Normal3f(-0.952345073f, 0.305022508f, 0.f),
    C_Vertex3f(0.344909191f, 0.229962111f, 0.f),
    C_Vertex3f(0.344909191f, 0.229962111f, -1.f),
    C_Normal3f(-0.917284548f, 0.398232311f, 0.f),
    C_Vertex3f(0.350272417f, 0.244291902f, 0.f),
    C_Vertex3f(0.350272417f, 0.244291902f, -1.f),
    C_Normal3f(-0.866581678f, 0.499035209f, 0.f),
    C_Vertex3f(0.356708288f, 0.25714004f, 0.f),
    C_Vertex3f(0.356708288f, 0.25714004f, -1.f),
    C_Normal3f(-0.797269225f, 0.603623867f, 0.f),
    C_Vertex3f(0.364216805f, 0.268506527f, 0.f),
    C_Vertex3f(0.364216805f, 0.268506527f, -1.f),
    C_Normal3f(-0.720497668f, 0.693457305f, 0.f),
    C_Vertex3f(0.372797966f, 0.278391361f, 0.f),
    C_Vertex3f(0.372797966f, 0.278391361f, -1.f),
    C_Normal3f(-0.64184618f, 0.766833425f, 0.f),
    C_Vertex3f(0.382055283f, 0.287040591f, 0.f),
    C_Vertex3f(0.382055283f, 0.287040591f, -1.f),
    C_Normal3f(-0.55178076f, 0.833989203f, 0.f),
    C_Vertex3f(0.391988754f, 0.294454217f, 0.f),
    C_Vertex3f(0.391988754f, 0.294454217f, -1.f),
    C_Normal3f(-0.452831835f, 0.89159596f, 0.f),
    C_Vertex3f(0.402598381f, 0.300632238f, 0.f),
    C_Vertex3f(0.402598381f, 0.300632238f, -1.f),
    C_Normal3f(-0.348695308f, 0.937236071f, 0.f),
    C_Vertex3f(0.413884163f, 0.305574656f, 0.f),
    C_Vertex3f(0.413884163f, 0.305574656f, -1.f),
    C_Normal3f(-0.243574992f, 0.969882011f, 0.f),
    C_Vertex3f(0.4258461f, 0.309281468f, 0.f),
    C_Vertex3f(0.4258461f, 0.309281468f, -1.f),
    C_Normal3f(-0.141396523f, 0.989952981f, 0.f),
    C_Vertex3f(0.438484192f, 0.311752677f, 0.f),
    C_Vertex3f(0.438484192f, 0.311752677f, -1.f),
    C_Normal3f(-0.0452046841f, 0.99897778f, 0.f),
    C_Vertex3f(0.451798439f, 0.312988281f, 0.f),
    C_Vertex3f(0.451798439f, 0.312988281f, -1.f),
    C_Normal3f(0.0454090424f, 0.998968422f, 0.f),
    C_Vertex3f(0.465789795f, 0.312988281f, 0.f),
    C_Vertex3f(0.465789795f, 0.312988281f, -1.f),
    C_Normal3f(0.141717017f, 0.989907146f, 0.f),
    C_Vertex3f(0.47913301f, 0.311745763f, 0.f),
    C_Vertex3f(0.47913301f, 0.311745763f, -1.f),
    C_Normal3f(0.243561462f, 0.969885468f, 0.f),
    C_Vertex3f(0.49182713f, 0.309260726f, 0.f),
    C_Vertex3f(0.49182713f, 0.309260726f, -1.f),
    C_Normal3f(0.347869754f, 0.937542856f, 0.f),
    C_Vertex3f(0.503872156f, 0.305533171f, 0.f),
    C_Vertex3f(0.503872156f, 0.305533171f, -1.f),
    C_Normal3f(0.450798839f, 0.89262563f, 0.f),
    C_Vertex3f(0.515268087f, 0.300563097f, 0.f),
    C_Vertex3f(0.515268087f, 0.300563097f, -1.f),
    C_Normal3f(0.548332453f, 0.836260438f, 0.f),
    C_Vertex3f(0.526014924f, 0.294350505f, 0.f),
    C_Vertex3f(0.526014924f, 0.294350505f, -1.f),
    C_Normal3f(0.637015522f, 0.770850897f, 0.f),
    C_Vertex3f(0.536112666f, 0.286895394f, 0.f),
    C_Vertex3f(0.536112666f, 0.286895394f, -1.f),
    C_Normal3f(0.714535177f, 0.699599624f, 0.f),
    C_Vertex3f(0.545561314f, 0.278197765f, 0.f),
    C_Vertex3f(0.545561314f, 0.278197765f, -1.f),
    C_Normal3f(0.790378392f, 0.612619042f, 0.f),
    C_Vertex3f(0.554360867f, 0.268257618f, 0.f),
    C_Vertex3f(0.554360867f, 0.268257618f, -1.f),
    C_Normal3f(0.860005796f, 0.510284245f, 0.f),
    C_Vertex3f(0.562060475f, 0.256911159f, 0.f),
    C_Vertex3f(0.562060475f, 0.256911159f, -1.f),
    C_Normal3f(0.91206032f, 0.410056055f, 0.f),
    C_Vertex3f(0.56866014f, 0.244158387f, 0.f),
    C_Vertex3f(0.56866014f, 0.244158387f, -1.f),
    C_Normal3f(0.948766232f, 0.315978914f, 0.f),
    C_Vertex3f(0.574159861f, 0.229999304f, 0.f),
    C_Vertex3f(0.574159861f, 0.229999304f, -1.f),
    C_Normal3f(0.973124206f, 0.23028098f, 0.f),
    C_Vertex3f(0.578559637f, 0.214433908f, 0.f),
    C_Vertex3f(0.578559637f, 0.214433908f, -1.f),
    C_Normal3f(0.988112688f, 0.153730854f, 0.f),
    C_Vertex3f(0.581859469f, 0.197462201f, 0.f),
    C_Vertex3f(0.581859469f, 0.197462201f, -1.f),
    C_Normal3f(0.996282458f, 0.0861468166f, 0.f),
    C_Vertex3f(0.584059358f, 0.179084182f, 0.f),
    C_Vertex3f(0.584059358f, 0.179084182f, -1.f),
    C_Normal3f(0.999641776f, 0.0267636627f, 0.f),
    C_Vertex3f(0.585159302f, 0.15929985f, 0.f),
    C_Vertex3f(0.585159302f, 0.15929985f, -1.f),
    C_Normal3f(0.99973309f, -0.0231050774f, 0.f),
    C_Vertex3f(0.585159302f, 0.138000488f, 0.f),
    C_Vertex3f(0.585159302f, 0.138000488f, -1.f),
    C_Normal3f(0.997282863f, -0.0736666098f, 0.f),
    C_Vertex3f(0.584202886f, 0.117916703f, 0.f),
    C_Vertex3f(0.584202886f, 0.117916703f, -1.f),
    C_Normal3f(0.991384625f, -0.130982816f, 0.f),
    C_Vertex3f(0.582290053f, 0.0991572142f, 0.f),
    C_Vertex3f(0.582290053f, 0.0991572142f, -1.f),
    C_Normal3f(0.980661035f, -0.195714027f, 0.f),
    C_Vertex3f(0.579420805f, 0.0817220211f, 0.f),
    C_Vertex3f(0.579420805f, 0.0817220211f, -1.f),
    C_Normal3f(0.963315785f, -0.268370479f, 0.f),
    C_Vertex3f(0.57559514f, 0.065611124f, 0.f),
    C_Vertex3f(0.57559514f, 0.065611124f, -1.f),
    C_Normal3f(0.937120378f, -0.349006355f, 0.f),
    C_Vertex3f(0.57081306f, 0.050824523f, 0.f),
    C_Vertex3f(0.57081306f, 0.050824523f, -1.f),
    C_Normal3f(0.899522364f, -0.436874717f, 0.f),
    C_Vertex3f(0.565074563f, 0.0373622179f, 0.f),
    C_Vertex3f(0.565074563f, 0.0373622179f, -1.f),
    C_Normal3f(0.847974837f, -0.530036569f, 0.f),
    C_Vertex3f(0.55837965f, 0.0252242088f, 0.f),
    C_Vertex3f(0.55837965f, 0.0252242088f, -1.f),
    C_Normal3f(0.778601587f, -0.627518535f, 0.f),
    C_Vertex3f(0.550728321f, 0.0144104958f, 0.f),
    C_Vertex3f(0.550728321f, 0.0144104958f, -1.f),
    C_Normal3f(0.690287232f, -0.723535538f, 0.f),
    C_Vertex3f(0.542038321f, 0.00494849682f, 0.f),
    C_Vertex3f(0.542038321f, 0.00494849682f, -1.f),
    C_Normal3f(0.5872069f, -0.809436858f, 0.f),
    C_Vertex3f(0.532309651f, -0.00316178799f, 0.f),
    C_Vertex3f(0.532309651f, -0.00316178799f, -1.f),
    C_Normal3f(0.47441861f, -0.880299389f, 0.f),
    C_Vertex3f(0.521542311f, -0.00992035866f, 0.f),
    C_Vertex3f(0.521542311f, -0.00992035866f, -1.f),
    C_Normal3f(0.358350933f, -0.933586955f, 0.f),
    C_Vertex3f(0.5097363f, -0.0153272152f, 0.f),
    C_Vertex3f(0.5097363f, -0.0153272152f, -1.f),
    C_Normal3f(0.245148465f, -0.969485581f, 0.f),
    C_Vertex3f(0.496891618f, -0.0193823576f, 0.f),
    C_Vertex3f(0.496891618f, -0.0193823576f, -1.f),
    C_Normal3f(0.139402702f, -0.990235746f, 0.f),
    C_Vertex3f(0.483008265f, -0.0220857859f, 0.f),
    C_Vertex3f(0.483008265f, -0.0220857859f, -1.f),
    C_Normal3f(0.0451653637f, -0.998979449f, 0.f),
    C_Vertex3f(0.468086243f, -0.0234375f, 0.f),
    C_Vertex3f(0.468086243f, -0.0234375f, -1.f),
    C_Normal3f(-0.0445650034f, -0.999006569f, 0.f),
    C_Vertex3f(0.453110695f, -0.0234375f, 0.f),
    C_Vertex3f(0.453110695f, -0.0234375f, -1.f),
    C_Normal3f(-0.144121513f, -0.989559948f, 0.f),
    C_Vertex3f(0.439775944f, -0.0221745968f, 0.f),
    C_Vertex3f(0.439775944f, -0.0221745968f, -1.f),
    C_Normal3f(-0.24765873f, -0.968847275f, 0.f),
    C_Vertex3f(0.427096844f, -0.0196487904f, 0.f),
    C_Vertex3f(0.427096844f, -0.0196487904f, -1.f),
    C_Normal3f(-0.353527576f, -0.935424089f, 0.f),
    C_Vertex3f(0.415073395f, -0.0158600807f, 0.f),
    C_Vertex3f(0.415073395f, -0.0158600807f, -1.f),
    C_Normal3f(-0.457689643f, -0.889112055f, 0.f),
    C_Vertex3f(0.403705597f, -0.0108084679f, 0.f),
    C_Vertex3f(0.403705597f, -0.0108084679f, -1.f),
    C_Normal3f(-0.555978894f, -0.831196427f, 0.f),
    C_Vertex3f(0.39299345f, -0.0044939518f, 0.f),
    C_Vertex3f(0.39299345f, -0.0044939518f, -1.f),
    C_Normal3f(-0.644885004f, -0.764279723f, 0.f),
    C_Vertex3f(0.382936954f, 0.00308346748f, 0.f),
    C_Vertex3f(0.382936954f, 0.00308346748f, -1.f),
    C_Normal3f(-0.722144306f, -0.691742539f, 0.f),
    C_Vertex3f(0.37353611f, 0.01192379f, 0.f),
    C_Vertex3f(0.37353611f, 0.01192379f, -1.f),
    C_Normal3f(-0.798405468f, -0.602120042f, 0.f),
    C_Vertex3f(0.364790916f, 0.0220270157f, 0.f),
    C_Vertex3f(0.364790916f, 0.0220270157f, -1.f),
    C_Normal3f(-0.867775321f, -0.496956676f, 0.f),
    C_Vertex3f(0.357138872f, 0.0336663723f, 0.f),
    C_Vertex3f(0.357138872f, 0.0336663723f, -1.f),
    C_Normal3f(-0.918272197f, -0.395949781f, 0.f),
    C_Vertex3f(0.350579977f, 0.0468418598f, 0.f),
    C_Vertex3f(0.350579977f, 0.0468418598f, -1.f),
    C_Normal3f(-0.953035653f, -0.302857965f, 0.f),
    C_Vertex3f(0.345114231f, 0.0615534782f, 0.f),
    C_Vertex3f(0.345114231f, 0.0615534782f, -1.f),
    C_Normal3f(-0.975640953f, -0.219373792f, 0.f),
    C_Vertex3f(0.340741634f, 0.0778012276f, 0.f),
    C_Vertex3f(0.340741634f, 0.0778012276f, -1.f),
    C_Normal3f(-0.989323437f, -0.145736486f, 0.f),
    C_Vertex3f(0.337462187f, 0.0955851078f, 0.f),
    C_Vertex3f(0.337462187f, 0.0955851078f, -1.f),
    C_Normal3f(-0.996685147f, -0.081356138f, 0.f),
    C_Vertex3f(0.335275888f, 0.114905119f, 0.f),
    C_Vertex3f(0.335275888f, 0.114905119f, -1.f),
    C_Normal3f(-0.99967587f, -0.0254604295f, 0.f),
    C_Vertex3f(0.334182739f, 0.135761261f, 0.f),
    C_Vertex3f(0.334182739f, 0.135761261f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999819696f, -0.0189867951f, 0.f),
    C_Vertex3f(0.398635864f, 0.151981354f, 0.f),
    C_Vertex3f(0.398635864f, 0.151981354f, -1.f),
    C_Normal3f(0.999855936f, 0.0169770233f, 0.f),
    C_Vertex3f(0.398635864f, 0.135217667f, 0.f),
    C_Vertex3f(0.398635864f, 0.135217667f, -1.f),
    C_Normal3f(0.998430371f, 0.0560075864f, 0.f),
    C_Vertex3f(0.399177909f, 0.120057821f, 0.f),
    C_Vertex3f(0.399177909f, 0.120057821f, -1.f),
    C_Normal3f(0.994742692f, 0.102405682f, 0.f),
    C_Vertex3f(0.400261998f, 0.106229067f, 0.f),
    C_Vertex3f(0.400261998f, 0.106229067f, -1.f),
    C_Normal3f(0.987388313f, 0.158317372f, 0.f),
    C_Vertex3f(0.401888132f, 0.0937314034f, 0.f),
    C_Vertex3f(0.401888132f, 0.0937314034f, -1.f),
    C_Normal3f(0.974067271f, 0.226258412f, 0.f),
    C_Vertex3f(0.404056311f, 0.0825648308f, 0.f),
    C_Vertex3f(0.404056311f, 0.0825648308f, -1.f),
    C_Normal3f(0.951003075f, 0.309181392f, 0.f),
    C_Vertex3f(0.406766534f, 0.0727293491f, 0.f),
    C_Vertex3f(0.406766534f, 0.0727293491f, -1.f),
    C_Normal3f(0.912103832f, 0.409959167f, 0.f),
    C_Vertex3f(0.410018802f, 0.0642249584f, 0.f),
    C_Vertex3f(0.410018802f, 0.0642249584f, -1.f),
    C_Normal3f(0.8398844f, 0.542765379f, 0.f),
    C_Vertex3f(0.413813114f, 0.0570516586f, 0.f),
    C_Vertex3f(0.413813114f, 0.0570516586f, -1.f),
    C_Normal3f(0.726519585f, 0.68714571f, 0.f),
    C_Vertex3f(0.42031765f, 0.0482883453f, 0.f),
    C_Vertex3f(0.42031765f, 0.0482883453f, -1.f),
    C_Normal3f(0.561143398f, 0.827718616f, 0.f),
    C_Vertex3f(0.430389881f, 0.039525032f, 0.f),
    C_Vertex3f(0.430389881f, 0.039525032f, -1.f),
    C_Normal3f(0.338713616f, 0.940889537f, 0.f),
    C_Vertex3f(0.441861629f, 0.0336828232f, 0.f),
    C_Vertex3f(0.441861629f, 0.0336828232f, -1.f),
    C_Normal3f(0.109639853f, 0.993971348f, 0.f),
    C_Vertex3f(0.454732895f, 0.0307617188f, 0.f),
    C_Vertex3f(0.454732895f, 0.0307617188f, -1.f),
    C_Normal3f(-0.0975214392f, 0.995233357f, 0.f),
    C_Vertex3f(0.468343735f, 0.0307617188f, 0.f),
    C_Vertex3f(0.468343735f, 0.0307617188f, -1.f),
    C_Normal3f(-0.324939996f, 0.94573468f, 0.f),
    C_Vertex3f(0.479813576f, 0.0332193375f, 0.f),
    C_Vertex3f(0.479813576f, 0.0332193375f, -1.f),
    C_Normal3f(-0.553361595f, 0.832941234f, 0.f),
    C_Vertex3f(0.489802361f, 0.0381345749f, 0.f),
    C_Vertex3f(0.489802361f, 0.0381345749f, -1.f),
    C_Normal3f(-0.735636413f, 0.677376628f, 0.f),
    C_Vertex3f(0.498310089f, 0.045507431f, 0.f),
    C_Vertex3f(0.498310089f, 0.045507431f, -1.f),
    C_Normal3f(-0.850483358f, 0.52600199f, 0.f),
    C_Vertex3f(0.504183769f, 0.0537528992f, 0.f),
    C_Vertex3f(0.504183769f, 0.0537528992f, -1.f),
    C_Normal3f(-0.912668705f, 0.408700287f, 0.f),
    C_Vertex3f(0.508314371f, 0.0616831779f, 0.f),
    C_Vertex3f(0.508314371f, 0.0616831779f, -1.f),
    C_Normal3f(-0.949630976f, 0.313370615f, 0.f),
    C_Vertex3f(0.511854887f, 0.0708832741f, 0.f),
    C_Vertex3f(0.511854887f, 0.0708832741f, -1.f),
    C_Normal3f(-0.972581208f, 0.232563779f, 0.f),
    C_Vertex3f(0.514805317f, 0.0813531876f, 0.f),
    C_Vertex3f(0.514805317f, 0.0813531876f, -1.f),
    C_Normal3f(-0.986356914f, 0.164620861f, 0.f),
    C_Vertex3f(0.517165661f, 0.0930929184f, 0.f),
    C_Vertex3f(0.517165661f, 0.0930929184f, -1.f),
    C_Normal3f(-0.994205952f, 0.107491814f, 0.f),
    C_Vertex3f(0.518935919f, 0.106102467f, 0.f),
    C_Vertex3f(0.518935919f, 0.106102467f, -1.f),
    C_Normal3f(-0.99824357f, 0.0592435338f, 0.f),
    C_Vertex3f(0.520116091f, 0.120381832f, 0.f),
    C_Vertex3f(0.520116091f, 0.120381832f, -1.f),
    C_Normal3f(-0.999833286f, 0.0182585958f, 0.f),
    C_Vertex3f(0.520706177f, 0.135931015f, 0.f),
    C_Vertex3f(0.520706177f, 0.135931015f, -1.f),
    C_Normal3f(-0.999816477f, -0.0191558395f, 0.f),
    C_Vertex3f(0.520706177f, 0.152694702f, 0.f),
    C_Vertex3f(0.520706177f, 0.152694702f, -1.f),
    C_Normal3f(-0.998059034f, -0.0622734986f, 0.f),
    C_Vertex3f(0.520088673f, 0.168160915f, 0.f),
    C_Vertex3f(0.520088673f, 0.168160915f, -1.f),
    C_Normal3f(-0.993621886f, -0.112763159f, 0.f),
    C_Vertex3f(0.518853664f, 0.182384968f, 0.f),
    C_Vertex3f(0.518853664f, 0.182384968f, -1.f),
    C_Normal3f(-0.985055923f, -0.172235161f, 0.f),
    C_Vertex3f(0.517001152f, 0.195366859f, 0.f),
    C_Vertex3f(0.517001152f, 0.195366859f, -1.f),
    C_Normal3f(-0.970160842f, -0.242462203f, 0.f),
    C_Vertex3f(0.514531136f, 0.20710659f, 0.f),
    C_Vertex3f(0.514531136f, 0.20710659f, -1.f),
    C_Normal3f(-0.945650458f, -0.325185061f, 0.f),
    C_Vertex3f(0.511443615f, 0.21760416f, 0.f),
    C_Vertex3f(0.511443615f, 0.21760416f, -1.f),
    C_Normal3f(-0.906809092f, -0.421541631f, 0.f),
    C_Vertex3f(0.50773859f, 0.22685957f, 0.f),
    C_Vertex3f(0.50773859f, 0.22685957f, -1.f),
    C_Normal3f(-0.843541205f, -0.537064373f, 0.f),
    C_Vertex3f(0.503416061f, 0.234872818f, 0.f),
    C_Vertex3f(0.503416061f, 0.234872818f, -1.f),
    C_Normal3f(-0.733231068f, -0.679979682f, 0.f),
    C_Vertex3f(0.497213364f, 0.243391037f, 0.f),
    C_Vertex3f(0.497213364f, 0.243391037f, -1.f),
    C_Normal3f(-0.558766544f, -0.82932502f, 0.f),
    C_Vertex3f(0.488376617f, 0.25109005f, 0.f),
    C_Vertex3f(0.488376617f, 0.25109005f, -1.f),
    C_Normal3f(-0.333176285f, -0.942864537f, 0.f),
    C_Vertex3f(0.478168488f, 0.256222725f, 0.f),
    C_Vertex3f(0.478168488f, 0.256222725f, -1.f),
    C_Normal3f(-0.102676541f, -0.994714737f, 0.f),
    C_Vertex3f(0.466588974f, 0.258789063f, 0.f),
    C_Vertex3f(0.466588974f, 0.258789063f, -1.f),
    C_Normal3f(0.098783344f, -0.995109022f, 0.f),
    C_Vertex3f(0.453306198f, 0.258789063f, 0.f),
    C_Vertex3f(0.453306198f, 0.258789063f, -1.f),
    C_Normal3f(0.318208337f, -0.948020816f, 0.f),
    C_Vertex3f(0.44128418f, 0.256277084f, 0.f),
    C_Vertex3f(0.44128418f, 0.256277084f, -1.f),
    C_Normal3f(0.546115994f, -0.837709546f, 0.f),
    C_Vertex3f(0.430854797f, 0.251253128f, 0.f),
    C_Vertex3f(0.430854797f, 0.251253128f, -1.f),
    C_Normal3f(0.729994476f, -0.683452964f, 0.f),
    C_Vertex3f(0.422018051f, 0.243717194f, 0.f),
    C_Vertex3f(0.422018051f, 0.243717194f, -1.f),
    C_Normal3f(0.845107675f, -0.534596026f, 0.f),
    C_Vertex3f(0.41592598f, 0.235307693f, 0.f),
    C_Vertex3f(0.41592598f, 0.235307693f, -1.f),
    C_Normal3f(0.907946944f, -0.419084996f, 0.f),
    C_Vertex3f(0.411603451f, 0.227253437f, 0.f),
    C_Vertex3f(0.411603451f, 0.227253437f, -1.f),
    C_Normal3f(0.946674109f, -0.322192639f, 0.f),
    C_Vertex3f(0.407898426f, 0.217916012f, 0.f),
    C_Vertex3f(0.407898426f, 0.217916012f, -1.f),
    C_Normal3f(0.970883965f, -0.23955062f, 0.f),
    C_Vertex3f(0.404810905f, 0.207295418f, 0.f),
    C_Vertex3f(0.404810905f, 0.207295418f, -1.f),
    C_Normal3f(0.985482812f, -0.169775203f, 0.f),
    C_Vertex3f(0.402340889f, 0.195391655f, 0.f),
    C_Vertex3f(0.402340889f, 0.195391655f, -1.f),
    C_Normal3f(0.993826389f, -0.110946804f, 0.f),
    C_Vertex3f(0.400488377f, 0.182204723f, 0.f),
    C_Vertex3f(0.400488377f, 0.182204723f, -1.f),
    C_Normal3f(0.998126805f, -0.0611792244f, 0.f),
    C_Vertex3f(0.399253368f, 0.167734623f, 0.f),
    C_Vertex3f(0.399253368f, 0.167734623f, -1.f),
    C_Normal3f(0.999819696f, -0.0189867951f, 0.f),
    C_Vertex3f(0.398635864f, 0.151981354f, 0.f),
    C_Vertex3f(0.398635864f, 0.151981354f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99967587f, -0.0254604295f, 0.f),
    C_Vertex3f(0.629104614f, 0.135761261f, 0.f),
    C_Vertex3f(0.629104614f, 0.135761261f, -1.f),
    C_Normal3f(-0.999679029f, 0.0253369864f, 0.f),
    C_Vertex3f(0.629104614f, 0.157826424f, 0.f),
    C_Vertex3f(0.629104614f, 0.157826424f, -1.f),
    C_Normal3f(-0.996618569f, 0.0821665376f, 0.f),
    C_Vertex3f(0.630177259f, 0.178082824f, 0.f),
    C_Vertex3f(0.630177259f, 0.178082824f, -1.f),
    C_Normal3f(-0.989124417f, 0.147081092f, 0.f),
    C_Vertex3f(0.63232255f, 0.196857572f, 0.f),
    C_Vertex3f(0.63232255f, 0.196857572f, -1.f),
    C_Normal3f(-0.975229859f, 0.221194118f, 0.f),
    C_Vertex3f(0.635540485f, 0.214150667f, 0.f),
    C_Vertex3f(0.635540485f, 0.214150667f, -1.f),
    C_Normal3f(-0.952345073f, 0.305022508f, 0.f),
    C_Vertex3f(0.639831066f, 0.229962111f, 0.f),
    C_Vertex3f(0.639831066f, 0.229962111f, -1.f),
    C_Normal3f(-0.917284548f, 0.398232311f, 0.f),
    C_Vertex3f(0.645194292f, 0.244291902f, 0.f),
    C_Vertex3f(0.645194292f, 0.244291902f, -1.f),
    C_Normal3f(-0.866581678f, 0.499035209f, 0.f),
    C_Vertex3f(0.651630163f, 0.25714004f, 0.f),
    C_Vertex3f(0.651630163f, 0.25714004f, -1.f),
    C_Normal3f(-0.797269225f, 0.603623867f, 0.f),
    C_Vertex3f(0.65913868f, 0.268506527f, 0.f),
    C_Vertex3f(0.65913868f, 0.268506527f, -1.f),
    C_Normal3f(-0.720229149f, 0.693736196f, 0.f),
    C_Vertex3f(0.667719841f, 0.278391361f, 0.f),
    C_Vertex3f(0.667719841f, 0.278391361f, -1.f),
    C_Normal3f(-0.641030967f, 0.767514944f, 0.f),
    C_Vertex3f(0.676990986f, 0.287040591f, 0.f),
    C_Vertex3f(0.676990986f, 0.287040591f, -1.f),
    C_Normal3f(-0.550491393f, 0.834840775f, 0.f),
    C_Vertex3f(0.686952114f, 0.294454217f, 0.f),
    C_Vertex3f(0.686952114f, 0.294454217f, -1.f),
    C_Normal3f(-0.451245219f, 0.892399907f, 0.f),
    C_Vertex3f(0.697603226f, 0.300632238f, 0.f),
    C_Vertex3f(0.697603226f, 0.300632238f, -1.f),
    C_Normal3f(-0.347062737f, 0.937841892f, 0.f),
    C_Vertex3f(0.708944321f, 0.305574656f, 0.f),
    C_Vertex3f(0.708944321f, 0.305574656f, -1.f),
    C_Normal3f(-0.242166191f, 0.970234811f, 0.f),
    C_Vertex3f(0.720975399f, 0.309281468f, 0.f),
    C_Vertex3f(0.720975399f, 0.309281468f, -1.f),
    C_Normal3f(-0.140443072f, 0.990088701f, 0.f),
    C_Vertex3f(0.733696461f, 0.311752677f, 0.f),
    C_Vertex3f(0.733696461f, 0.311752677f, -1.f),
    C_Normal3f(-0.0449563228f, 0.998988926f, 0.f),
    C_Vertex3f(0.747107506f, 0.312988281f, 0.f),
    C_Vertex3f(0.747107506f, 0.312988281f, -1.f),
    C_Normal3f(0.0453330874f, 0.998971939f, 0.f),
    C_Vertex3f(0.761153221f, 0.312988281f, 0.f),
    C_Vertex3f(0.761153221f, 0.312988281f, -1.f),
    C_Normal3f(0.141842395f, 0.989889264f, 0.f),
    C_Vertex3f(0.774487972f, 0.311745763f, 0.f),
    C_Vertex3f(0.774487972f, 0.311745763f, -1.f),
    C_Normal3f(0.243900865f, 0.969800174f, 0.f),
    C_Vertex3f(0.787167072f, 0.309260726f, 0.f),
    C_Vertex3f(0.787167072f, 0.309260726f, -1.f),
    C_Normal3f(0.348519325f, 0.937301636f, 0.f),
    C_Vertex3f(0.799190521f, 0.305533171f, 0.f),
    C_Vertex3f(0.799190521f, 0.305533171f, -1.f),
    C_Normal3f(0.451819897f, 0.892109156f, 0.f),
    C_Vertex3f(0.810558319f, 0.300563097f, 0.f),
    C_Vertex3f(0.810558319f, 0.300563097f, -1.f),
    C_Normal3f(0.549732208f, 0.835340977f, 0.f),
    C_Vertex3f(0.821270466f, 0.294350505f, 0.f),
    C_Vertex3f(0.821270466f, 0.294350505f, -1.f),
    C_Normal3f(0.638743103f, 0.769420087f, 0.f),
    C_Vertex3f(0.831326962f, 0.286895394f, 0.f),
    C_Vertex3f(0.831326962f, 0.286895394f, -1.f),
    C_Normal3f(0.716495633f, 0.697591603f, 0.f),
    C_Vertex3f(0.840727806f, 0.278197765f, 0.f),
    C_Vertex3f(0.840727806f, 0.278197765f, -1.f),
    C_Normal3f(0.792211473f, 0.610246599f, 0.f),
    C_Vertex3f(0.849473f, 0.268257618f, 0.f),
    C_Vertex3f(0.849473f, 0.268257618f, -1.f),
    C_Normal3f(0.861388206f, 0.507947147f, 0.f),
    C_Vertex3f(0.857125044f, 0.256911159f, 0.f),
    C_Vertex3f(0.857125044f, 0.256911159f, -1.f),
    C_Normal3f(0.913006186f, 0.407945573f, 0.f),
    C_Vertex3f(0.863683939f, 0.244158387f, 0.f),
    C_Vertex3f(0.863683939f, 0.244158387f, -1.f),
    C_Normal3f(0.949350119f, 0.31422022f, 0.f),
    C_Vertex3f(0.869149685f, 0.229999304f, 0.f),
    C_Vertex3f(0.869149685f, 0.229999304f, -1.f),
    C_Normal3f(0.973442078f, 0.22893317f, 0.f),
    C_Vertex3f(0.873522282f, 0.214433908f, 0.f),
    C_Vertex3f(0.873522282f, 0.214433908f, -1.f),
    C_Normal3f(0.988256574f, 0.152803421f, 0.f),
    C_Vertex3f(0.876801729f, 0.197462201f, 0.f),
    C_Vertex3f(0.876801729f, 0.197462201f, -1.f),
    C_Normal3f(0.996327996f, 0.0856185555f, 0.f),
    C_Vertex3f(0.878988028f, 0.179084182f, 0.f),
    C_Vertex3f(0.878988028f, 0.179084182f, -1.f),
    C_Normal3f(0.999646187f, 0.0265984461f, 0.f),
    C_Vertex3f(0.880081177f, 0.15929985f, 0.f),
    C_Vertex3f(0.880081177f, 0.15929985f, -1.f),
    C_Normal3f(0.999736845f, -0.0229410138f, 0.f),
    C_Vertex3f(0.880081177f, 0.138000488f, 0.f),
    C_Vertex3f(0.880081177f, 0.138000488f, -1.f),
    C_Normal3f(0.997321248f, -0.0731460527f, 0.f),
    C_Vertex3f(0.879131556f, 0.117916703f, 0.f),
    C_Vertex3f(0.879131556f, 0.117916703f, -1.f),
    C_Normal3f(0.991505086f, -0.130068034f, 0.f),
    C_Vertex3f(0.877232313f, 0.0991572142f, 0.f),
    C_Vertex3f(0.877232313f, 0.0991572142f, -1.f),
    C_Normal3f(0.980926991f, -0.19437626f, 0.f),
    C_Vertex3f(0.87438345f, 0.0817220211f, 0.f),
    C_Vertex3f(0.87438345f, 0.0817220211f, -1.f),
    C_Normal3f(0.963807285f, -0.266599804f, 0.f),
    C_Vertex3f(0.870584965f, 0.065611124f, 0.f),
    C_Vertex3f(0.870584965f, 0.065611124f, -1.f),
    C_Normal3f(0.937929451f, -0.346825987f, 0.f),
    C_Vertex3f(0.865836859f, 0.050824523f, 0.f),
    C_Vertex3f(0.865836859f, 0.050824523f, -1.f),
    C_Normal3f(0.900740266f, -0.434358209f, 0.f),
    C_Vertex3f(0.860139132f, 0.0373622179f, 0.f),
    C_Vertex3f(0.860139132f, 0.0373622179f, -1.f),
    C_Normal3f(0.849666357f, -0.527320683f, 0.f),
    C_Vertex3f(0.853491783f, 0.0252242088f, 0.f),
    C_Vertex3f(0.853491783f, 0.0252242088f, -1.f),
    C_Normal3f(0.780259728f, -0.625455558f, 0.f),
    C_Vertex3f(0.845894814f, 0.0144104958f, 0.f),
    C_Vertex3f(0.845894814f, 0.0144104958f, -1.f),
    C_Normal3f(0.691221058f, -0.722643375f, 0.f),
    C_Vertex3f(0.837238789f, 0.00494849682f, 0.f),
    C_Vertex3f(0.837238789f, 0.00494849682f, -1.f),
    C_Normal3f(0.587334454f, -0.809344292f, 0.f),
    C_Vertex3f(0.827523708f, -0.00316178799f, 0.f),
    C_Vertex3f(0.827523708f, -0.00316178799f, -1.f),
    C_Normal3f(0.473865807f, -0.880597055f, 0.f),
    C_Vertex3f(0.816749573f, -0.00992035866f, 0.f),
    C_Vertex3f(0.816749573f, -0.00992035866f, -1.f),
    C_Normal3f(0.357406199f, -0.933948994f, 0.f),
    C_Vertex3f(0.804916382f, -0.0153272152f, 0.f),
    C_Vertex3f(0.804916382f, -0.0153272152f, -1.f),
    C_Normal3f(0.244156569f, -0.969735861f, 0.f),
    C_Vertex3f(0.792024136f, -0.0193823576f, 0.f),
    C_Vertex3f(0.792024136f, -0.0193823576f, -1.f),
    C_Normal3f(0.138664961f, -0.990339398f, 0.f),
    C_Vertex3f(0.778072834f, -0.0220857859f, 0.f),
    C_Vertex3f(0.778072834f, -0.0220857859f, -1.f),
    C_Normal3f(0.0449512601f, -0.998989105f, 0.f),
    C_Vertex3f(0.763062477f, -0.0234375f, 0.f),
    C_Vertex3f(0.763062477f, -0.0234375f, -1.f),
    C_Normal3f(-0.044479765f, -0.999010324f, 0.f),
    C_Vertex3f(0.74803257f, -0.0234375f, 0.f),
    C_Vertex3f(0.74803257f, -0.0234375f, -1.f),
    C_Normal3f(-0.144121513f, -0.989559948f, 0.f),
    C_Vertex3f(0.734697819f, -0.0221745968f, 0.f),
    C_Vertex3f(0.734697819f, -0.0221745968f, -1.f),
    C_Normal3f(-0.24765873f, -0.968847275f, 0.f),
    C_Vertex3f(0.722018719f, -0.0196487904f, 0.f),
    C_Vertex3f(0.722018719f, -0.0196487904f, -1.f),
    C_Normal3f(-0.353527576f, -0.935424089f, 0.f),
    C_Vertex3f(0.70999527f, -0.0158600807f, 0.f),
    C_Vertex3f(0.70999527f, -0.0158600807f, -1.f),
    C_Normal3f(-0.457689643f, -0.889112055f, 0.f),
    C_Vertex3f(0.698627472f, -0.0108084679f, 0.f),
    C_Vertex3f(0.698627472f, -0.0108084679f, -1.f),
    C_Normal3f(-0.555978894f, -0.831196427f, 0.f),
    C_Vertex3f(0.687915325f, -0.0044939518f, 0.f),
    C_Vertex3f(0.687915325f, -0.0044939518f, -1.f),
    C_Normal3f(-0.644885004f, -0.764279723f, 0.f),
    C_Vertex3f(0.677858829f, 0.00308346748f, 0.f),
    C_Vertex3f(0.677858829f, 0.00308346748f, -1.f),
    C_Normal3f(-0.722144306f, -0.691742539f, 0.f),
    C_Vertex3f(0.668457985f, 0.01192379f, 0.f),
    C_Vertex3f(0.668457985f, 0.01192379f, -1.f),
    C_Normal3f(-0.798405468f, -0.602120042f, 0.f),
    C_Vertex3f(0.659712791f, 0.0220270157f, 0.f),
    C_Vertex3f(0.659712791f, 0.0220270157f, -1.f),
    C_Normal3f(-0.867775321f, -0.496956676f, 0.f),
    C_Vertex3f(0.652060747f, 0.0336663723f, 0.f),
    C_Vertex3f(0.652060747f, 0.0336663723f, -1.f),
    C_Normal3f(-0.918272197f, -0.395949781f, 0.f),
    C_Vertex3f(0.645501852f, 0.0468418598f, 0.f),
    C_Vertex3f(0.645501852f, 0.0468418598f, -1.f),
    C_Normal3f(-0.953035653f, -0.302857965f, 0.f),
    C_Vertex3f(0.640036106f, 0.0615534782f, 0.f),
    C_Vertex3f(0.640036106f, 0.0615534782f, -1.f),
    C_Normal3f(-0.975640953f, -0.219373792f, 0.f),
    C_Vertex3f(0.635663509f, 0.0778012276f, 0.f),
    C_Vertex3f(0.635663509f, 0.0778012276f, -1.f),
    C_Normal3f(-0.989323437f, -0.145736486f, 0.f),
    C_Vertex3f(0.632384062f, 0.0955851078f, 0.f),
    C_Vertex3f(0.632384062f, 0.0955851078f, -1.f),
    C_Normal3f(-0.996685147f, -0.081356138f, 0.f),
    C_Vertex3f(0.630197763f, 0.114905119f, 0.f),
    C_Vertex3f(0.630197763f, 0.114905119f, -1.f),
    C_Normal3f(-0.99967587f, -0.0254604295f, 0.f),
    C_Vertex3f(0.629104614f, 0.135761261f, 0.f),
    C_Vertex3f(0.629104614f, 0.135761261f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999819696f, -0.0189867951f, 0.f),
    C_Vertex3f(0.693557739f, 0.151981354f, 0.f),
    C_Vertex3f(0.693557739f, 0.151981354f, -1.f),
    C_Normal3f(0.999855936f, 0.0169770233f, 0.f),
    C_Vertex3f(0.693557739f, 0.135217667f, 0.f),
    C_Vertex3f(0.693557739f, 0.135217667f, -1.f),
    C_Normal3f(0.998430371f, 0.0560075864f, 0.f),
    C_Vertex3f(0.694099784f, 0.120057821f, 0.f),
    C_Vertex3f(0.694099784f, 0.120057821f, -1.f),
    C_Normal3f(0.994742692f, 0.102405682f, 0.f),
    C_Vertex3f(0.695183873f, 0.106229067f, 0.f),
    C_Vertex3f(0.695183873f, 0.106229067f, -1.f),
    C_Normal3f(0.987388313f, 0.158317372f, 0.f),
    C_Vertex3f(0.696810007f, 0.0937314034f, 0.f),
    C_Vertex3f(0.696810007f, 0.0937314034f, -1.f),
    C_Normal3f(0.974067271f, 0.226258412f, 0.f),
    C_Vertex3f(0.698978186f, 0.0825648308f, 0.f),
    C_Vertex3f(0.698978186f, 0.0825648308f, -1.f),
    C_Normal3f(0.951003075f, 0.309181392f, 0.f),
    C_Vertex3f(0.701688409f, 0.0727293491f, 0.f),
    C_Vertex3f(0.701688409f, 0.0727293491f, -1.f),
    C_Normal3f(0.912103832f, 0.409959167f, 0.f),
    C_Vertex3f(0.704940677f, 0.0642249584f, 0.f),
    C_Vertex3f(0.704940677f, 0.0642249584f, -1.f),
    C_Normal3f(0.8398844f, 0.542765379f, 0.f),
    C_Vertex3f(0.708734989f, 0.0570516586f, 0.f),
    C_Vertex3f(0.708734989f, 0.0570516586f, -1.f),
    C_Normal3f(0.726519585f, 0.68714571f, 0.f),
    C_Vertex3f(0.715239525f, 0.0482883453f, 0.f),
    C_Vertex3f(0.715239525f, 0.0482883453f, -1.f),
    C_Normal3f(0.561143398f, 0.827718616f, 0.f),
    C_Vertex3f(0.725311756f, 0.039525032f, 0.f),
    C_Vertex3f(0.725311756f, 0.039525032f, -1.f),
    C_Normal3f(0.338713616f, 0.940889537f, 0.f),
    C_Vertex3f(0.736783504f, 0.0336828232f, 0.f),
    C_Vertex3f(0.736783504f, 0.0336828232f, -1.f),
    C_Normal3f(0.109639853f, 0.993971348f, 0.f),
    C_Vertex3f(0.74965477f, 0.0307617188f, 0.f),
    C_Vertex3f(0.74965477f, 0.0307617188f, -1.f),
    C_Normal3f(-0.0975214392f, 0.995233357f, 0.f),
    C_Vertex3f(0.76326561f, 0.0307617188f, 0.f),
    C_Vertex3f(0.76326561f, 0.0307617188f, -1.f),
    C_Normal3f(-0.324939996f, 0.94573468f, 0.f),
    C_Vertex3f(0.774735451f, 0.0332193375f, 0.f),
    C_Vertex3f(0.774735451f, 0.0332193375f, -1.f),
    C_Normal3f(-0.553361595f, 0.832941234f, 0.f),
    C_Vertex3f(0.784724236f, 0.0381345749f, 0.f),
    C_Vertex3f(0.784724236f, 0.0381345749f, -1.f),
    C_Normal3f(-0.735636413f, 0.677376628f, 0.f),
    C_Vertex3f(0.793231964f, 0.045507431f, 0.f),
    C_Vertex3f(0.793231964f, 0.045507431f, -1.f),
    C_Normal3f(-0.850483358f, 0.52600199f, 0.f),
    C_Vertex3f(0.799105644f, 0.0537528992f, 0.f),
    C_Vertex3f(0.799105644f, 0.0537528992f, -1.f),
    C_Normal3f(-0.912668705f, 0.408700287f, 0.f),
    C_Vertex3f(0.803236246f, 0.0616831779f, 0.f),
    C_Vertex3f(0.803236246f, 0.0616831779f, -1.f),
    C_Normal3f(-0.949630976f, 0.313370615f, 0.f),
    C_Vertex3f(0.806776762f, 0.0708832741f, 0.f),
    C_Vertex3f(0.806776762f, 0.0708832741f, -1.f),
    C_Normal3f(-0.972581208f, 0.232563779f, 0.f),
    C_Vertex3f(0.809727192f, 0.0813531876f, 0.f),
    C_Vertex3f(0.809727192f, 0.0813531876f, -1.f),
    C_Normal3f(-0.986356914f, 0.164620861f, 0.f),
    C_Vertex3f(0.812087536f, 0.0930929184f, 0.f),
    C_Vertex3f(0.812087536f, 0.0930929184f, -1.f),
    C_Normal3f(-0.994205952f, 0.107491814f, 0.f),
    C_Vertex3f(0.813857794f, 0.106102467f, 0.f),
    C_Vertex3f(0.813857794f, 0.106102467f, -1.f),
    C_Normal3f(-0.99824357f, 0.0592435338f, 0.f),
    C_Vertex3f(0.815037966f, 0.120381832f, 0.f),
    C_Vertex3f(0.815037966f, 0.120381832f, -1.f),
    C_Normal3f(-0.999833286f, 0.0182585958f, 0.f),
    C_Vertex3f(0.815628052f, 0.135931015f, 0.f),
    C_Vertex3f(0.815628052f, 0.135931015f, -1.f),
    C_Normal3f(-0.999824703f, -0.0187270213f, 0.f),
    C_Vertex3f(0.815628052f, 0.152694702f, 0.f),
    C_Vertex3f(0.815628052f, 0.152694702f, -1.f),
    C_Normal3f(-0.998144805f, -0.0608841889f, 0.f),
    C_Vertex3f(0.815024376f, 0.168160915f, 0.f),
    C_Vertex3f(0.815024376f, 0.168160915f, -1.f),
    C_Normal3f(-0.993901789f, -0.11026901f, 0.f),
    C_Vertex3f(0.813817024f, 0.182384968f, 0.f),
    C_Vertex3f(0.813817024f, 0.182384968f, -1.f),
    C_Normal3f(-0.985703588f, -0.16848886f, 0.f),
    C_Vertex3f(0.812005997f, 0.195366859f, 0.f),
    C_Vertex3f(0.812005997f, 0.195366859f, -1.f),
    C_Normal3f(-0.971426189f, -0.237341687f, 0.f),
    C_Vertex3f(0.809591293f, 0.20710659f, 0.f),
    C_Vertex3f(0.809591293f, 0.20710659f, -1.f),
    C_Normal3f(-0.947872519f, -0.318649918f, 0.f),
    C_Vertex3f(0.806572914f, 0.21760416f, 0.f),
    C_Vertex3f(0.806572914f, 0.21760416f, -1.f),
    C_Normal3f(-0.910398364f, -0.413732886f, 0.f),
    C_Vertex3f(0.802950859f, 0.22685957f, 0.f),
    C_Vertex3f(0.802950859f, 0.22685957f, -1.f),
    C_Normal3f(-0.864071131f, -0.503369749f, 0.f),
    C_Vertex3f(0.798725128f, 0.234872818f, 0.f),
    C_Vertex3f(0.798725128f, 0.234872818f, -1.f),
    C_Normal3f(-0.81088984f, -0.58519882f, 0.f),
    C_Vertex3f(0.796310425f, 0.238258362f, 0.f),
    C_Vertex3f(0.796310425f, 0.238258362f, -1.f),
    C_Normal3f(-0.712831557f, -0.701335132f, 0.f),
    C_Vertex3f(0.792577744f, 0.243391037f, 0.f),
    C_Vertex3f(0.792577744f, 0.243391037f, -1.f),
    C_Normal3f(-0.555433512f, -0.831560969f, 0.f),
    C_Vertex3f(0.783685684f, 0.25109005f, 0.f),
    C_Vertex3f(0.783685684f, 0.25109005f, -1.f),
    C_Normal3f(-0.329455912f, -0.944170952f, 0.f),
    C_Vertex3f(0.773366928f, 0.256222725f, 0.f),
    C_Vertex3f(0.773366928f, 0.256222725f, -1.f),
    C_Normal3f(-0.101558663f, -0.994829535f, 0.f),
    C_Vertex3f(0.761621475f, 0.258789063f, 0.f),
    C_Vertex3f(0.761621475f, 0.258789063f, -1.f),
    C_Normal3f(0.0983575284f, -0.995151162f, 0.f),
    C_Vertex3f(0.748228073f, 0.258789063f, 0.f),
    C_Vertex3f(0.748228073f, 0.258789063f, -1.f),
    C_Normal3f(0.318208337f, -0.948020816f, 0.f),
    C_Vertex3f(0.736206055f, 0.256277084f, 0.f),
    C_Vertex3f(0.736206055f, 0.256277084f, -1.f),
    C_Normal3f(0.546115994f, -0.837709546f, 0.f),
    C_Vertex3f(0.725776672f, 0.251253128f, 0.f),
    C_Vertex3f(0.725776672f, 0.251253128f, -1.f),
    C_Normal3f(0.729994476f, -0.683452964f, 0.f),
    C_Vertex3f(0.716939926f, 0.243717194f, 0.f),
    C_Vertex3f(0.716939926f, 0.243717194f, -1.f),
    C_Normal3f(0.845107675f, -0.534596026f, 0.f),
    C_Vertex3f(0.710847855f, 0.235307693f, 0.f),
    C_Vertex3f(0.710847855f, 0.235307693f, -1.f),
    C_Normal3f(0.907946944f, -0.419084996f, 0.f),
    C_Vertex3f(0.706525326f, 0.227253437f, 0.f),
    C_Vertex3f(0.706525326f, 0.227253437f, -1.f),
    C_Normal3f(0.946674109f, -0.322192639f, 0.f),
    C_Vertex3f(0.702820301f, 0.217916012f, 0.f),
    C_Vertex3f(0.702820301f, 0.217916012f, -1.f),
    C_Normal3f(0.970883965f, -0.23955062f, 0.f),
    C_Vertex3f(0.69973278f, 0.207295418f, 0.f),
    C_Vertex3f(0.69973278f, 0.207295418f, -1.f),
    C_Normal3f(0.985482812f, -0.169775203f, 0.f),
    C_Vertex3f(0.697262764f, 0.195391655f, 0.f),
    C_Vertex3f(0.697262764f, 0.195391655f, -1.f),
    C_Normal3f(0.993826389f, -0.110946804f, 0.f),
    C_Vertex3f(0.695410252f, 0.182204723f, 0.f),
    C_Vertex3f(0.695410252f, 0.182204723f, -1.f),
    C_Normal3f(0.998126805f, -0.0611792244f, 0.f),
    C_Vertex3f(0.694175243f, 0.167734623f, 0.f),
    C_Vertex3f(0.694175243f, 0.167734623f, -1.f),
    C_Normal3f(0.999819696f, -0.0189867951f, 0.f),
    C_Vertex3f(0.693557739f, 0.151981354f, 0.f),
    C_Vertex3f(0.693557739f, 0.151981354f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402854919f, 0.651855469f, 0.f),
    C_Vertex3f(0.140045166f, -0.0234375f, 0.f),
    C_Vertex3f(0.197738647f, -0.0234375f, 0.f),
    C_Vertex3f(0.460548401f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.267288208f, 0.471471786f, 0.f),
    C_Vertex3f(0.267288208f, 0.492774963f, 0.f),
    C_Vertex3f(0.266331792f, 0.45137763f, 0.f),
    C_Vertex3f(0.266195059f, 0.512562752f, 0.f),
    C_Vertex3f(0.26441896f, 0.432601213f, 0.f),
    C_Vertex3f(0.26400876f, 0.530943871f, 0.f),
    C_Vertex3f(0.261549711f, 0.415142536f, 0.f),
    C_Vertex3f(0.260729313f, 0.54791832f, 0.f),
    C_Vertex3f(0.257724047f, 0.399001598f, 0.f),
    C_Vertex3f(0.256356716f, 0.563486099f, 0.f),
    C_Vertex3f(0.252941966f, 0.3841784f, 0.f),
    C_Vertex3f(0.25089097f, 0.577647209f, 0.f),
    C_Vertex3f(0.247203469f, 0.370672941f, 0.f),
    C_Vertex3f(0.244332075f, 0.590401649f, 0.f),
    C_Vertex3f(0.240508556f, 0.358485222f, 0.f),
    C_Vertex3f(0.236680031f, 0.60174942f, 0.f),
    C_Vertex3f(0.232857227f, 0.347615242f, 0.f),
    C_Vertex3f(0.227934837f, 0.611690521f, 0.f),
    C_Vertex3f(0.224167228f, 0.33810401f, 0.f),
    C_Vertex3f(0.218533993f, 0.620388985f, 0.f),
    C_Vertex3f(0.214438558f, 0.329951525f, 0.f),
    C_Vertex3f(0.208477497f, 0.62784481f, 0.f),
    C_Vertex3f(0.203671217f, 0.323157787f, 0.f),
    C_Vertex3f(0.202835083f, 0.486136436f, 0.f),
    C_Vertex3f(0.202835083f, 0.469372749f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.208477497f, 0.62784481f, 0.f),
    C_Vertex3f(0.19776535f, 0.634057999f, 0.f),
    C_Vertex3f(0.199213028f, 0.528645515f, 0.f),
    C_Vertex3f(0.201024055f, 0.515690804f, 0.f),
    C_Vertex3f(0.202231407f, 0.501521111f, 0.f),
    C_Vertex3f(0.202835083f, 0.486136436f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.19776535f, 0.634057999f, 0.f),
    C_Vertex3f(0.186397552f, 0.639028549f, 0.f),
    C_Vertex3f(0.19015789f, 0.560219765f, 0.f),
    C_Vertex3f(0.193779945f, 0.550909996f, 0.f),
    C_Vertex3f(0.196798325f, 0.540385246f, 0.f),
    C_Vertex3f(0.199213028f, 0.528645515f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.186397552f, 0.639028549f, 0.f),
    C_Vertex3f(0.174374104f, 0.642756462f, 0.f),
    C_Vertex3f(0.179784775f, 0.576887131f, 0.f),
    C_Vertex3f(0.183517456f, 0.571754456f, 0.f),
    C_Vertex3f(0.185932159f, 0.568314552f, 0.f),
    C_Vertex3f(0.19015789f, 0.560219765f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.148828506f, 0.592285156f, 0.f),
    C_Vertex3f(0.160573959f, 0.589718819f, 0.f),
    C_Vertex3f(0.161695004f, 0.645241737f, 0.f),
    C_Vertex3f(0.170892715f, 0.584586143f, 0.f),
    C_Vertex3f(0.174374104f, 0.642756462f, 0.f),
    C_Vertex3f(0.179784775f, 0.576887131f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0980548859f, 0.568749428f, 0.f),
    C_Vertex3f(0.104146957f, 0.577213287f, 0.f),
    C_Vertex3f(0.10818243f, 0.642776847f, 0.f),
    C_Vertex3f(0.112983704f, 0.584749222f, 0.f),
    C_Vertex3f(0.120903492f, 0.645248532f, 0.f),
    C_Vertex3f(0.123413086f, 0.589773178f, 0.f),
    C_Vertex3f(0.134314537f, 0.646484375f, 0.f),
    C_Vertex3f(0.135435104f, 0.592285156f, 0.f),
    C_Vertex3f(0.148360252f, 0.646484375f, 0.f),
    C_Vertex3f(0.148828506f, 0.592285156f, 0.f),
    C_Vertex3f(0.161695004f, 0.645241737f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0961513519f, 0.639069319f, 0.f),
    C_Vertex3f(0.084810257f, 0.634125948f, 0.f),
    C_Vertex3f(0.0869398117f, 0.540574074f, 0.f),
    C_Vertex3f(0.0900273323f, 0.551221848f, 0.f),
    C_Vertex3f(0.093732357f, 0.560613632f, 0.f),
    C_Vertex3f(0.0980548859f, 0.568749428f, 0.f),
    C_Vertex3f(0.10818243f, 0.642776847f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.084810257f, 0.634125948f, 0.f),
    C_Vertex3f(0.0741591454f, 0.627946734f, 0.f),
    C_Vertex3f(0.0751223564f, 0.328619361f, 0.f),
    C_Vertex3f(0.0807647705f, 0.485423088f, 0.f),
    C_Vertex3f(0.0813822746f, 0.501094818f, 0.f),
    C_Vertex3f(0.0826172829f, 0.515510559f, 0.f),
    C_Vertex3f(0.0844697952f, 0.528670311f, 0.f),
    C_Vertex3f(0.0869398117f, 0.540574074f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0823909044f, 0.439514518f, 0.f),
    C_Vertex3f(0.0813068151f, 0.453411222f, 0.f),
    C_Vertex3f(0.0858345032f, 0.322269678f, 0.f),
    C_Vertex3f(0.0807647705f, 0.468659401f, 0.f),
    C_Vertex3f(0.0751223564f, 0.328619361f, 0.f),
    C_Vertex3f(0.0807647705f, 0.485423088f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0751223564f, 0.328619361f, 0.f),
    C_Vertex3f(0.0741591454f, 0.627946734f, 0.f),
    C_Vertex3f(0.0650658607f, 0.33623898f, 0.f),
    C_Vertex3f(0.0641980171f, 0.620531678f, 0.f),
    C_Vertex3f(0.0556650162f, 0.345128536f, 0.f),
    C_Vertex3f(0.0549268723f, 0.611880779f, 0.f),
    C_Vertex3f(0.0469198227f, 0.355288029f, 0.f),
    C_Vertex3f(0.0463457108f, 0.601994038f, 0.f),
    C_Vertex3f(0.0392677784f, 0.366977096f, 0.f),
    C_Vertex3f(0.0388371944f, 0.590625644f, 0.f),
    C_Vertex3f(0.0327088833f, 0.380195737f, 0.f),
    C_Vertex3f(0.0324013233f, 0.577775598f, 0.f),
    C_Vertex3f(0.0272431374f, 0.394943953f, 0.f),
    C_Vertex3f(0.0270380974f, 0.563443899f, 0.f),
    C_Vertex3f(0.0228705406f, 0.411221743f, 0.f),
    C_Vertex3f(0.0227475166f, 0.547630548f, 0.f),
    C_Vertex3f(0.0195910931f, 0.429029107f, 0.f),
    C_Vertex3f(0.0195295811f, 0.530335546f, 0.f),
    C_Vertex3f(0.0174047947f, 0.448366046f, 0.f),
    C_Vertex3f(0.0173842907f, 0.51155889f, 0.f),
    C_Vertex3f(0.0163116455f, 0.491300583f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.203671217f, 0.323157787f, 0.f),
    C_Vertex3f(0.202835083f, 0.469372749f, 0.f),
    C_Vertex3f(0.202244997f, 0.453742027f, 0.f),
    C_Vertex3f(0.201064825f, 0.439408302f, 0.f),
    C_Vertex3f(0.199294567f, 0.426371574f, 0.f),
    C_Vertex3f(0.196934223f, 0.414631844f, 0.f),
    C_Vertex3f(0.193983793f, 0.40418911f, 0.f),
    C_Vertex3f(0.191865206f, 0.317722797f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191865206f, 0.317722797f, 0.f),
    C_Vertex3f(0.193983793f, 0.40418911f, 0.f),
    C_Vertex3f(0.190443277f, 0.395043373f, 0.f),
    C_Vertex3f(0.186312675f, 0.387194633f, 0.f),
    C_Vertex3f(0.180438995f, 0.379003525f, 0.f),
    C_Vertex3f(0.179020524f, 0.313646555f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.171931267f, 0.371630669f, 0.f),
    C_Vertex3f(0.161942482f, 0.366715431f, 0.f),
    C_Vertex3f(0.165137172f, 0.31092906f, 0.f),
    C_Vertex3f(0.179020524f, 0.313646555f, 0.f),
    C_Vertex3f(0.180438995f, 0.379003525f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0959420204f, 0.390303135f, 0.f),
    C_Vertex3f(0.097202301f, 0.317189932f, 0.f),
    C_Vertex3f(0.102446556f, 0.381621361f, 0.f),
    C_Vertex3f(0.10922575f, 0.313380122f, 0.f),
    C_Vertex3f(0.112518787f, 0.372939587f, 0.f),
    C_Vertex3f(0.12190485f, 0.310840249f, 0.f),
    C_Vertex3f(0.123990536f, 0.367151737f, 0.f),
    C_Vertex3f(0.135239601f, 0.309570313f, 0.f),
    C_Vertex3f(0.136861801f, 0.364257813f, 0.f),
    C_Vertex3f(0.150215149f, 0.309570313f, 0.f),
    C_Vertex3f(0.150472641f, 0.364257813f, 0.f),
    C_Vertex3f(0.165137172f, 0.31092906f, 0.f),
    C_Vertex3f(0.161942482f, 0.366715431f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.097202301f, 0.317189932f, 0.f),
    C_Vertex3f(0.0959420204f, 0.390303135f, 0.f),
    C_Vertex3f(0.0921477079f, 0.39744246f, 0.f),
    C_Vertex3f(0.0888954401f, 0.405933261f, 0.f),
    C_Vertex3f(0.0861852169f, 0.415775537f, 0.f),
    C_Vertex3f(0.0858345032f, 0.322269678f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0840170383f, 0.42696929f, 0.f),
    C_Vertex3f(0.0823909044f, 0.439514518f, 0.f),
    C_Vertex3f(0.0858345032f, 0.322269678f, 0.f),
    C_Vertex3f(0.0861852169f, 0.415775537f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.880081177f, 0.138000488f, 0.f),
    C_Vertex3f(0.880081177f, 0.15929985f, 0.f),
    C_Vertex3f(0.879131556f, 0.117916703f, 0.f),
    C_Vertex3f(0.878988028f, 0.179084182f, 0.f),
    C_Vertex3f(0.877232313f, 0.0991572142f, 0.f),
    C_Vertex3f(0.876801729f, 0.197462201f, 0.f),
    C_Vertex3f(0.87438345f, 0.0817220211f, 0.f),
    C_Vertex3f(0.873522282f, 0.214433908f, 0.f),
    C_Vertex3f(0.870584965f, 0.065611124f, 0.f),
    C_Vertex3f(0.869149685f, 0.229999304f, 0.f),
    C_Vertex3f(0.865836859f, 0.050824523f, 0.f),
    C_Vertex3f(0.863683939f, 0.244158387f, 0.f),
    C_Vertex3f(0.860139132f, 0.0373622179f, 0.f),
    C_Vertex3f(0.857125044f, 0.256911159f, 0.f),
    C_Vertex3f(0.853491783f, 0.0252242088f, 0.f),
    C_Vertex3f(0.849473f, 0.268257618f, 0.f),
    C_Vertex3f(0.845894814f, 0.0144104958f, 0.f),
    C_Vertex3f(0.840727806f, 0.278197765f, 0.f),
    C_Vertex3f(0.837238789f, 0.00494849682f, 0.f),
    C_Vertex3f(0.831326962f, 0.286895394f, 0.f),
    C_Vertex3f(0.827523708f, -0.00316178799f, 0.f),
    C_Vertex3f(0.821270466f, 0.294350505f, 0.f),
    C_Vertex3f(0.816749573f, -0.00992035866f, 0.f),
    C_Vertex3f(0.815628052f, 0.152694702f, 0.f),
    C_Vertex3f(0.815628052f, 0.135931015f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.821270466f, 0.294350505f, 0.f),
    C_Vertex3f(0.810558319f, 0.300563097f, 0.f),
    C_Vertex3f(0.812005997f, 0.195366859f, 0.f),
    C_Vertex3f(0.813817024f, 0.182384968f, 0.f),
    C_Vertex3f(0.815024376f, 0.168160915f, 0.f),
    C_Vertex3f(0.815628052f, 0.152694702f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.810558319f, 0.300563097f, 0.f),
    C_Vertex3f(0.799190521f, 0.305533171f, 0.f),
    C_Vertex3f(0.802950859f, 0.22685957f, 0.f),
    C_Vertex3f(0.806572914f, 0.21760416f, 0.f),
    C_Vertex3f(0.809591293f, 0.20710659f, 0.f),
    C_Vertex3f(0.812005997f, 0.195366859f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.799190521f, 0.305533171f, 0.f),
    C_Vertex3f(0.787167072f, 0.309260726f, 0.f),
    C_Vertex3f(0.792577744f, 0.243391037f, 0.f),
    C_Vertex3f(0.796310425f, 0.238258362f, 0.f),
    C_Vertex3f(0.798725128f, 0.234872818f, 0.f),
    C_Vertex3f(0.802950859f, 0.22685957f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.761621475f, 0.258789063f, 0.f),
    C_Vertex3f(0.773366928f, 0.256222725f, 0.f),
    C_Vertex3f(0.774487972f, 0.311745763f, 0.f),
    C_Vertex3f(0.783685684f, 0.25109005f, 0.f),
    C_Vertex3f(0.787167072f, 0.309260726f, 0.f),
    C_Vertex3f(0.792577744f, 0.243391037f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.710847855f, 0.235307693f, 0.f),
    C_Vertex3f(0.716939926f, 0.243717194f, 0.f),
    C_Vertex3f(0.720975399f, 0.309281468f, 0.f),
    C_Vertex3f(0.725776672f, 0.251253128f, 0.f),
    C_Vertex3f(0.733696461f, 0.311752677f, 0.f),
    C_Vertex3f(0.736206055f, 0.256277084f, 0.f),
    C_Vertex3f(0.747107506f, 0.312988281f, 0.f),
    C_Vertex3f(0.748228073f, 0.258789063f, 0.f),
    C_Vertex3f(0.761153221f, 0.312988281f, 0.f),
    C_Vertex3f(0.761621475f, 0.258789063f, 0.f),
    C_Vertex3f(0.774487972f, 0.311745763f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.708944321f, 0.305574656f, 0.f),
    C_Vertex3f(0.697603226f, 0.300632238f, 0.f),
    C_Vertex3f(0.69973278f, 0.207295418f, 0.f),
    C_Vertex3f(0.702820301f, 0.217916012f, 0.f),
    C_Vertex3f(0.706525326f, 0.227253437f, 0.f),
    C_Vertex3f(0.710847855f, 0.235307693f, 0.f),
    C_Vertex3f(0.720975399f, 0.309281468f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.697603226f, 0.300632238f, 0.f),
    C_Vertex3f(0.686952114f, 0.294454217f, 0.f),
    C_Vertex3f(0.687915325f, -0.0044939518f, 0.f),
    C_Vertex3f(0.693557739f, 0.151981354f, 0.f),
    C_Vertex3f(0.694175243f, 0.167734623f, 0.f),
    C_Vertex3f(0.695410252f, 0.182204723f, 0.f),
    C_Vertex3f(0.697262764f, 0.195391655f, 0.f),
    C_Vertex3f(0.69973278f, 0.207295418f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.695183873f, 0.106229067f, 0.f),
    C_Vertex3f(0.694099784f, 0.120057821f, 0.f),
    C_Vertex3f(0.698627472f, -0.0108084679f, 0.f),
    C_Vertex3f(0.693557739f, 0.135217667f, 0.f),
    C_Vertex3f(0.687915325f, -0.0044939518f, 0.f),
    C_Vertex3f(0.693557739f, 0.151981354f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.687915325f, -0.0044939518f, 0.f),
    C_Vertex3f(0.686952114f, 0.294454217f, 0.f),
    C_Vertex3f(0.677858829f, 0.00308346748f, 0.f),
    C_Vertex3f(0.676990986f, 0.287040591f, 0.f),
    C_Vertex3f(0.668457985f, 0.01192379f, 0.f),
    C_Vertex3f(0.667719841f, 0.278391361f, 0.f),
    C_Vertex3f(0.659712791f, 0.0220270157f, 0.f),
    C_Vertex3f(0.65913868f, 0.268506527f, 0.f),
    C_Vertex3f(0.652060747f, 0.0336663723f, 0.f),
    C_Vertex3f(0.651630163f, 0.25714004f, 0.f),
    C_Vertex3f(0.645501852f, 0.0468418598f, 0.f),
    C_Vertex3f(0.645194292f, 0.244291902f, 0.f),
    C_Vertex3f(0.640036106f, 0.0615534782f, 0.f),
    C_Vertex3f(0.639831066f, 0.229962111f, 0.f),
    C_Vertex3f(0.635663509f, 0.0778012276f, 0.f),
    C_Vertex3f(0.635540485f, 0.214150667f, 0.f),
    C_Vertex3f(0.632384062f, 0.0955851078f, 0.f),
    C_Vertex3f(0.63232255f, 0.196857572f, 0.f),
    C_Vertex3f(0.630197763f, 0.114905119f, 0.f),
    C_Vertex3f(0.630177259f, 0.178082824f, 0.f),
    C_Vertex3f(0.629104614f, 0.157826424f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.816749573f, -0.00992035866f, 0.f),
    C_Vertex3f(0.815628052f, 0.135931015f, 0.f),
    C_Vertex3f(0.815037966f, 0.120381832f, 0.f),
    C_Vertex3f(0.813857794f, 0.106102467f, 0.f),
    C_Vertex3f(0.812087536f, 0.0930929184f, 0.f),
    C_Vertex3f(0.809727192f, 0.0813531876f, 0.f),
    C_Vertex3f(0.806776762f, 0.0708832741f, 0.f),
    C_Vertex3f(0.804916382f, -0.0153272152f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.804916382f, -0.0153272152f, 0.f),
    C_Vertex3f(0.806776762f, 0.0708832741f, 0.f),
    C_Vertex3f(0.803236246f, 0.0616831779f, 0.f),
    C_Vertex3f(0.799105644f, 0.0537528992f, 0.f),
    C_Vertex3f(0.793231964f, 0.045507431f, 0.f),
    C_Vertex3f(0.792024136f, -0.0193823576f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.784724236f, 0.0381345749f, 0.f),
    C_Vertex3f(0.774735451f, 0.0332193375f, 0.f),
    C_Vertex3f(0.778072834f, -0.0220857859f, 0.f),
    C_Vertex3f(0.792024136f, -0.0193823576f, 0.f),
    C_Vertex3f(0.793231964f, 0.045507431f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.708734989f, 0.0570516586f, 0.f),
    C_Vertex3f(0.70999527f, -0.0158600807f, 0.f),
    C_Vertex3f(0.715239525f, 0.0482883453f, 0.f),
    C_Vertex3f(0.722018719f, -0.0196487904f, 0.f),
    C_Vertex3f(0.725311756f, 0.039525032f, 0.f),
    C_Vertex3f(0.734697819f, -0.0221745968f, 0.f),
    C_Vertex3f(0.736783504f, 0.0336828232f, 0.f),
    C_Vertex3f(0.74803257f, -0.0234375f, 0.f),
    C_Vertex3f(0.74965477f, 0.0307617188f, 0.f),
    C_Vertex3f(0.763062477f, -0.0234375f, 0.f),
    C_Vertex3f(0.76326561f, 0.0307617188f, 0.f),
    C_Vertex3f(0.778072834f, -0.0220857859f, 0.f),
    C_Vertex3f(0.774735451f, 0.0332193375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.70999527f, -0.0158600807f, 0.f),
    C_Vertex3f(0.708734989f, 0.0570516586f, 0.f),
    C_Vertex3f(0.704940677f, 0.0642249584f, 0.f),
    C_Vertex3f(0.701688409f, 0.0727293491f, 0.f),
    C_Vertex3f(0.698978186f, 0.0825648308f, 0.f),
    C_Vertex3f(0.698627472f, -0.0108084679f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.696810007f, 0.0937314034f, 0.f),
    C_Vertex3f(0.695183873f, 0.106229067f, 0.f),
    C_Vertex3f(0.698627472f, -0.0108084679f, 0.f),
    C_Vertex3f(0.698978186f, 0.0825648308f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.585159302f, 0.138000488f, 0.f),
    C_Vertex3f(0.585159302f, 0.15929985f, 0.f),
    C_Vertex3f(0.584202886f, 0.117916703f, 0.f),
    C_Vertex3f(0.584059358f, 0.179084182f, 0.f),
    C_Vertex3f(0.582290053f, 0.0991572142f, 0.f),
    C_Vertex3f(0.581859469f, 0.197462201f, 0.f),
    C_Vertex3f(0.579420805f, 0.0817220211f, 0.f),
    C_Vertex3f(0.578559637f, 0.214433908f, 0.f),
    C_Vertex3f(0.57559514f, 0.065611124f, 0.f),
    C_Vertex3f(0.574159861f, 0.229999304f, 0.f),
    C_Vertex3f(0.57081306f, 0.050824523f, 0.f),
    C_Vertex3f(0.56866014f, 0.244158387f, 0.f),
    C_Vertex3f(0.565074563f, 0.0373622179f, 0.f),
    C_Vertex3f(0.562060475f, 0.256911159f, 0.f),
    C_Vertex3f(0.55837965f, 0.0252242088f, 0.f),
    C_Vertex3f(0.554360867f, 0.268257618f, 0.f),
    C_Vertex3f(0.550728321f, 0.0144104958f, 0.f),
    C_Vertex3f(0.545561314f, 0.278197765f, 0.f),
    C_Vertex3f(0.542038321f, 0.00494849682f, 0.f),
    C_Vertex3f(0.536112666f, 0.286895394f, 0.f),
    C_Vertex3f(0.532309651f, -0.00316178799f, 0.f),
    C_Vertex3f(0.526014924f, 0.294350505f, 0.f),
    C_Vertex3f(0.521542311f, -0.00992035866f, 0.f),
    C_Vertex3f(0.520706177f, 0.152694702f, 0.f),
    C_Vertex3f(0.520706177f, 0.135931015f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.526014924f, 0.294350505f, 0.f),
    C_Vertex3f(0.515268087f, 0.300563097f, 0.f),
    C_Vertex3f(0.517001152f, 0.195366859f, 0.f),
    C_Vertex3f(0.518853664f, 0.182384968f, 0.f),
    C_Vertex3f(0.520088673f, 0.168160915f, 0.f),
    C_Vertex3f(0.520706177f, 0.152694702f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.515268087f, 0.300563097f, 0.f),
    C_Vertex3f(0.503872156f, 0.305533171f, 0.f),
    C_Vertex3f(0.50773859f, 0.22685957f, 0.f),
    C_Vertex3f(0.511443615f, 0.21760416f, 0.f),
    C_Vertex3f(0.514531136f, 0.20710659f, 0.f),
    C_Vertex3f(0.517001152f, 0.195366859f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466588974f, 0.258789063f, 0.f),
    C_Vertex3f(0.478168488f, 0.256222725f, 0.f),
    C_Vertex3f(0.47913301f, 0.311745763f, 0.f),
    C_Vertex3f(0.488376617f, 0.25109005f, 0.f),
    C_Vertex3f(0.49182713f, 0.309260726f, 0.f),
    C_Vertex3f(0.497213364f, 0.243391037f, 0.f),
    C_Vertex3f(0.503872156f, 0.305533171f, 0.f),
    C_Vertex3f(0.503416061f, 0.234872818f, 0.f),
    C_Vertex3f(0.50773859f, 0.22685957f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.41592598f, 0.235307693f, 0.f),
    C_Vertex3f(0.422018051f, 0.243717194f, 0.f),
    C_Vertex3f(0.4258461f, 0.309281468f, 0.f),
    C_Vertex3f(0.430854797f, 0.251253128f, 0.f),
    C_Vertex3f(0.438484192f, 0.311752677f, 0.f),
    C_Vertex3f(0.44128418f, 0.256277084f, 0.f),
    C_Vertex3f(0.451798439f, 0.312988281f, 0.f),
    C_Vertex3f(0.453306198f, 0.258789063f, 0.f),
    C_Vertex3f(0.465789795f, 0.312988281f, 0.f),
    C_Vertex3f(0.466588974f, 0.258789063f, 0.f),
    C_Vertex3f(0.47913301f, 0.311745763f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.413884163f, 0.305574656f, 0.f),
    C_Vertex3f(0.402598381f, 0.300632238f, 0.f),
    C_Vertex3f(0.404810905f, 0.207295418f, 0.f),
    C_Vertex3f(0.407898426f, 0.217916012f, 0.f),
    C_Vertex3f(0.411603451f, 0.227253437f, 0.f),
    C_Vertex3f(0.41592598f, 0.235307693f, 0.f),
    C_Vertex3f(0.4258461f, 0.309281468f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402598381f, 0.300632238f, 0.f),
    C_Vertex3f(0.391988754f, 0.294454217f, 0.f),
    C_Vertex3f(0.39299345f, -0.0044939518f, 0.f),
    C_Vertex3f(0.398635864f, 0.151981354f, 0.f),
    C_Vertex3f(0.399253368f, 0.167734623f, 0.f),
    C_Vertex3f(0.400488377f, 0.182204723f, 0.f),
    C_Vertex3f(0.402340889f, 0.195391655f, 0.f),
    C_Vertex3f(0.404810905f, 0.207295418f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.400261998f, 0.106229067f, 0.f),
    C_Vertex3f(0.399177909f, 0.120057821f, 0.f),
    C_Vertex3f(0.403705597f, -0.0108084679f, 0.f),
    C_Vertex3f(0.398635864f, 0.135217667f, 0.f),
    C_Vertex3f(0.39299345f, -0.0044939518f, 0.f),
    C_Vertex3f(0.398635864f, 0.151981354f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.39299345f, -0.0044939518f, 0.f),
    C_Vertex3f(0.391988754f, 0.294454217f, 0.f),
    C_Vertex3f(0.382936954f, 0.00308346748f, 0.f),
    C_Vertex3f(0.382055283f, 0.287040591f, 0.f),
    C_Vertex3f(0.37353611f, 0.01192379f, 0.f),
    C_Vertex3f(0.372797966f, 0.278391361f, 0.f),
    C_Vertex3f(0.364790916f, 0.0220270157f, 0.f),
    C_Vertex3f(0.364216805f, 0.268506527f, 0.f),
    C_Vertex3f(0.357138872f, 0.0336663723f, 0.f),
    C_Vertex3f(0.356708288f, 0.25714004f, 0.f),
    C_Vertex3f(0.350579977f, 0.0468418598f, 0.f),
    C_Vertex3f(0.350272417f, 0.244291902f, 0.f),
    C_Vertex3f(0.345114231f, 0.0615534782f, 0.f),
    C_Vertex3f(0.344909191f, 0.229962111f, 0.f),
    C_Vertex3f(0.340741634f, 0.0778012276f, 0.f),
    C_Vertex3f(0.34061861f, 0.214150667f, 0.f),
    C_Vertex3f(0.337462187f, 0.0955851078f, 0.f),
    C_Vertex3f(0.337400675f, 0.196857572f, 0.f),
    C_Vertex3f(0.335275888f, 0.114905119f, 0.f),
    C_Vertex3f(0.335255384f, 0.178082824f, 0.f),
    C_Vertex3f(0.334182739f, 0.157826424f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.521542311f, -0.00992035866f, 0.f),
    C_Vertex3f(0.520706177f, 0.135931015f, 0.f),
    C_Vertex3f(0.520116091f, 0.120381832f, 0.f),
    C_Vertex3f(0.518935919f, 0.106102467f, 0.f),
    C_Vertex3f(0.517165661f, 0.0930929184f, 0.f),
    C_Vertex3f(0.514805317f, 0.0813531876f, 0.f),
    C_Vertex3f(0.511854887f, 0.0708832741f, 0.f),
    C_Vertex3f(0.5097363f, -0.0153272152f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.5097363f, -0.0153272152f, 0.f),
    C_Vertex3f(0.511854887f, 0.0708832741f, 0.f),
    C_Vertex3f(0.508314371f, 0.0616831779f, 0.f),
    C_Vertex3f(0.504183769f, 0.0537528992f, 0.f),
    C_Vertex3f(0.498310089f, 0.045507431f, 0.f),
    C_Vertex3f(0.496891618f, -0.0193823576f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.489802361f, 0.0381345749f, 0.f),
    C_Vertex3f(0.479813576f, 0.0332193375f, 0.f),
    C_Vertex3f(0.483008265f, -0.0220857859f, 0.f),
    C_Vertex3f(0.496891618f, -0.0193823576f, 0.f),
    C_Vertex3f(0.498310089f, 0.045507431f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.413813114f, 0.0570516586f, 0.f),
    C_Vertex3f(0.415073395f, -0.0158600807f, 0.f),
    C_Vertex3f(0.42031765f, 0.0482883453f, 0.f),
    C_Vertex3f(0.427096844f, -0.0196487904f, 0.f),
    C_Vertex3f(0.430389881f, 0.039525032f, 0.f),
    C_Vertex3f(0.439775944f, -0.0221745968f, 0.f),
    C_Vertex3f(0.441861629f, 0.0336828232f, 0.f),
    C_Vertex3f(0.453110695f, -0.0234375f, 0.f),
    C_Vertex3f(0.454732895f, 0.0307617188f, 0.f),
    C_Vertex3f(0.468086243f, -0.0234375f, 0.f),
    C_Vertex3f(0.468343735f, 0.0307617188f, 0.f),
    C_Vertex3f(0.483008265f, -0.0220857859f, 0.f),
    C_Vertex3f(0.479813576f, 0.0332193375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.415073395f, -0.0158600807f, 0.f),
    C_Vertex3f(0.413813114f, 0.0570516586f, 0.f),
    C_Vertex3f(0.410018802f, 0.0642249584f, 0.f),
    C_Vertex3f(0.406766534f, 0.0727293491f, 0.f),
    C_Vertex3f(0.404056311f, 0.0825648308f, 0.f),
    C_Vertex3f(0.403705597f, -0.0108084679f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.401888132f, 0.0937314034f, 0.f),
    C_Vertex3f(0.400261998f, 0.106229067f, 0.f),
    C_Vertex3f(0.403705597f, -0.0108084679f, 0.f),
    C_Vertex3f(0.404056311f, 0.0825648308f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.335275888f, 0.114905119f, 0.f),
    C_Vertex3f(0.334182739f, 0.157826424f, 0.f),
    C_Vertex3f(0.334182739f, 0.135761261f, 0.f),
    C_Vertex3f(0.630197763f, 0.114905119f, 0.f),
    C_Vertex3f(0.629104614f, 0.157826424f, 0.f),
    C_Vertex3f(0.629104614f, 0.135761261f, 0.f),
    C_Vertex3f(0.0174047947f, 0.448366046f, 0.f),
    C_Vertex3f(0.0163116455f, 0.491300583f, 0.f),
    C_Vertex3f(0.0163116455f, 0.469232559f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402854919f, 0.651855469f, -1.f),
    C_Vertex3f(0.140045166f, -0.0234375f, -1.f),
    C_Vertex3f(0.197738647f, -0.0234375f, -1.f),
    C_Vertex3f(0.460548401f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.267288208f, 0.471471786f, -1.f),
    C_Vertex3f(0.267288208f, 0.492774963f, -1.f),
    C_Vertex3f(0.266331792f, 0.45137763f, -1.f),
    C_Vertex3f(0.266195059f, 0.512562752f, -1.f),
    C_Vertex3f(0.26441896f, 0.432601213f, -1.f),
    C_Vertex3f(0.26400876f, 0.530943871f, -1.f),
    C_Vertex3f(0.261549711f, 0.415142536f, -1.f),
    C_Vertex3f(0.260729313f, 0.54791832f, -1.f),
    C_Vertex3f(0.257724047f, 0.399001598f, -1.f),
    C_Vertex3f(0.256356716f, 0.563486099f, -1.f),
    C_Vertex3f(0.252941966f, 0.3841784f, -1.f),
    C_Vertex3f(0.25089097f, 0.577647209f, -1.f),
    C_Vertex3f(0.247203469f, 0.370672941f, -1.f),
    C_Vertex3f(0.244332075f, 0.590401649f, -1.f),
    C_Vertex3f(0.240508556f, 0.358485222f, -1.f),
    C_Vertex3f(0.236680031f, 0.60174942f, -1.f),
    C_Vertex3f(0.232857227f, 0.347615242f, -1.f),
    C_Vertex3f(0.227934837f, 0.611690521f, -1.f),
    C_Vertex3f(0.224167228f, 0.33810401f, -1.f),
    C_Vertex3f(0.218533993f, 0.620388985f, -1.f),
    C_Vertex3f(0.214438558f, 0.329951525f, -1.f),
    C_Vertex3f(0.208477497f, 0.62784481f, -1.f),
    C_Vertex3f(0.203671217f, 0.323157787f, -1.f),
    C_Vertex3f(0.202835083f, 0.486136436f, -1.f),
    C_Vertex3f(0.202835083f, 0.469372749f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.208477497f, 0.62784481f, -1.f),
    C_Vertex3f(0.19776535f, 0.634057999f, -1.f),
    C_Vertex3f(0.199213028f, 0.528645515f, -1.f),
    C_Vertex3f(0.201024055f, 0.515690804f, -1.f),
    C_Vertex3f(0.202231407f, 0.501521111f, -1.f),
    C_Vertex3f(0.202835083f, 0.486136436f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.19776535f, 0.634057999f, -1.f),
    C_Vertex3f(0.186397552f, 0.639028549f, -1.f),
    C_Vertex3f(0.19015789f, 0.560219765f, -1.f),
    C_Vertex3f(0.193779945f, 0.550909996f, -1.f),
    C_Vertex3f(0.196798325f, 0.540385246f, -1.f),
    C_Vertex3f(0.199213028f, 0.528645515f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.186397552f, 0.639028549f, -1.f),
    C_Vertex3f(0.174374104f, 0.642756462f, -1.f),
    C_Vertex3f(0.179784775f, 0.576887131f, -1.f),
    C_Vertex3f(0.183517456f, 0.571754456f, -1.f),
    C_Vertex3f(0.185932159f, 0.568314552f, -1.f),
    C_Vertex3f(0.19015789f, 0.560219765f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.148828506f, 0.592285156f, -1.f),
    C_Vertex3f(0.160573959f, 0.589718819f, -1.f),
    C_Vertex3f(0.161695004f, 0.645241737f, -1.f),
    C_Vertex3f(0.170892715f, 0.584586143f, -1.f),
    C_Vertex3f(0.174374104f, 0.642756462f, -1.f),
    C_Vertex3f(0.179784775f, 0.576887131f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0980548859f, 0.568749428f, -1.f),
    C_Vertex3f(0.104146957f, 0.577213287f, -1.f),
    C_Vertex3f(0.10818243f, 0.642776847f, -1.f),
    C_Vertex3f(0.112983704f, 0.584749222f, -1.f),
    C_Vertex3f(0.120903492f, 0.645248532f, -1.f),
    C_Vertex3f(0.123413086f, 0.589773178f, -1.f),
    C_Vertex3f(0.134314537f, 0.646484375f, -1.f),
    C_Vertex3f(0.135435104f, 0.592285156f, -1.f),
    C_Vertex3f(0.148360252f, 0.646484375f, -1.f),
    C_Vertex3f(0.148828506f, 0.592285156f, -1.f),
    C_Vertex3f(0.161695004f, 0.645241737f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0961513519f, 0.639069319f, -1.f),
    C_Vertex3f(0.084810257f, 0.634125948f, -1.f),
    C_Vertex3f(0.0869398117f, 0.540574074f, -1.f),
    C_Vertex3f(0.0900273323f, 0.551221848f, -1.f),
    C_Vertex3f(0.093732357f, 0.560613632f, -1.f),
    C_Vertex3f(0.0980548859f, 0.568749428f, -1.f),
    C_Vertex3f(0.10818243f, 0.642776847f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.084810257f, 0.634125948f, -1.f),
    C_Vertex3f(0.0741591454f, 0.627946734f, -1.f),
    C_Vertex3f(0.0751223564f, 0.328619361f, -1.f),
    C_Vertex3f(0.0807647705f, 0.485423088f, -1.f),
    C_Vertex3f(0.0813822746f, 0.501094818f, -1.f),
    C_Vertex3f(0.0826172829f, 0.515510559f, -1.f),
    C_Vertex3f(0.0844697952f, 0.528670311f, -1.f),
    C_Vertex3f(0.0869398117f, 0.540574074f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0823909044f, 0.439514518f, -1.f),
    C_Vertex3f(0.0813068151f, 0.453411222f, -1.f),
    C_Vertex3f(0.0858345032f, 0.322269678f, -1.f),
    C_Vertex3f(0.0807647705f, 0.468659401f, -1.f),
    C_Vertex3f(0.0751223564f, 0.328619361f, -1.f),
    C_Vertex3f(0.0807647705f, 0.485423088f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0751223564f, 0.328619361f, -1.f),
    C_Vertex3f(0.0741591454f, 0.627946734f, -1.f),
    C_Vertex3f(0.0650658607f, 0.33623898f, -1.f),
    C_Vertex3f(0.0641980171f, 0.620531678f, -1.f),
    C_Vertex3f(0.0556650162f, 0.345128536f, -1.f),
    C_Vertex3f(0.0549268723f, 0.611880779f, -1.f),
    C_Vertex3f(0.0469198227f, 0.355288029f, -1.f),
    C_Vertex3f(0.0463457108f, 0.601994038f, -1.f),
    C_Vertex3f(0.0392677784f, 0.366977096f, -1.f),
    C_Vertex3f(0.0388371944f, 0.590625644f, -1.f),
    C_Vertex3f(0.0327088833f, 0.380195737f, -1.f),
    C_Vertex3f(0.0324013233f, 0.577775598f, -1.f),
    C_Vertex3f(0.0272431374f, 0.394943953f, -1.f),
    C_Vertex3f(0.0270380974f, 0.563443899f, -1.f),
    C_Vertex3f(0.0228705406f, 0.411221743f, -1.f),
    C_Vertex3f(0.0227475166f, 0.547630548f, -1.f),
    C_Vertex3f(0.0195910931f, 0.429029107f, -1.f),
    C_Vertex3f(0.0195295811f, 0.530335546f, -1.f),
    C_Vertex3f(0.0174047947f, 0.448366046f, -1.f),
    C_Vertex3f(0.0173842907f, 0.51155889f, -1.f),
    C_Vertex3f(0.0163116455f, 0.491300583f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.203671217f, 0.323157787f, -1.f),
    C_Vertex3f(0.202835083f, 0.469372749f, -1.f),
    C_Vertex3f(0.202244997f, 0.453742027f, -1.f),
    C_Vertex3f(0.201064825f, 0.439408302f, -1.f),
    C_Vertex3f(0.199294567f, 0.426371574f, -1.f),
    C_Vertex3f(0.196934223f, 0.414631844f, -1.f),
    C_Vertex3f(0.193983793f, 0.40418911f, -1.f),
    C_Vertex3f(0.191865206f, 0.317722797f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191865206f, 0.317722797f, -1.f),
    C_Vertex3f(0.193983793f, 0.40418911f, -1.f),
    C_Vertex3f(0.190443277f, 0.395043373f, -1.f),
    C_Vertex3f(0.186312675f, 0.387194633f, -1.f),
    C_Vertex3f(0.180438995f, 0.379003525f, -1.f),
    C_Vertex3f(0.179020524f, 0.313646555f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.171931267f, 0.371630669f, -1.f),
    C_Vertex3f(0.161942482f, 0.366715431f, -1.f),
    C_Vertex3f(0.165137172f, 0.31092906f, -1.f),
    C_Vertex3f(0.179020524f, 0.313646555f, -1.f),
    C_Vertex3f(0.180438995f, 0.379003525f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0959420204f, 0.390303135f, -1.f),
    C_Vertex3f(0.097202301f, 0.317189932f, -1.f),
    C_Vertex3f(0.102446556f, 0.381621361f, -1.f),
    C_Vertex3f(0.10922575f, 0.313380122f, -1.f),
    C_Vertex3f(0.112518787f, 0.372939587f, -1.f),
    C_Vertex3f(0.12190485f, 0.310840249f, -1.f),
    C_Vertex3f(0.123990536f, 0.367151737f, -1.f),
    C_Vertex3f(0.135239601f, 0.309570313f, -1.f),
    C_Vertex3f(0.136861801f, 0.364257813f, -1.f),
    C_Vertex3f(0.150215149f, 0.309570313f, -1.f),
    C_Vertex3f(0.150472641f, 0.364257813f, -1.f),
    C_Vertex3f(0.165137172f, 0.31092906f, -1.f),
    C_Vertex3f(0.161942482f, 0.366715431f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.097202301f, 0.317189932f, -1.f),
    C_Vertex3f(0.0959420204f, 0.390303135f, -1.f),
    C_Vertex3f(0.0921477079f, 0.39744246f, -1.f),
    C_Vertex3f(0.0888954401f, 0.405933261f, -1.f),
    C_Vertex3f(0.0861852169f, 0.415775537f, -1.f),
    C_Vertex3f(0.0858345032f, 0.322269678f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0840170383f, 0.42696929f, -1.f),
    C_Vertex3f(0.0823909044f, 0.439514518f, -1.f),
    C_Vertex3f(0.0858345032f, 0.322269678f, -1.f),
    C_Vertex3f(0.0861852169f, 0.415775537f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.880081177f, 0.138000488f, -1.f),
    C_Vertex3f(0.880081177f, 0.15929985f, -1.f),
    C_Vertex3f(0.879131556f, 0.117916703f, -1.f),
    C_Vertex3f(0.878988028f, 0.179084182f, -1.f),
    C_Vertex3f(0.877232313f, 0.0991572142f, -1.f),
    C_Vertex3f(0.876801729f, 0.197462201f, -1.f),
    C_Vertex3f(0.87438345f, 0.0817220211f, -1.f),
    C_Vertex3f(0.873522282f, 0.214433908f, -1.f),
    C_Vertex3f(0.870584965f, 0.065611124f, -1.f),
    C_Vertex3f(0.869149685f, 0.229999304f, -1.f),
    C_Vertex3f(0.865836859f, 0.050824523f, -1.f),
    C_Vertex3f(0.863683939f, 0.244158387f, -1.f),
    C_Vertex3f(0.860139132f, 0.0373622179f, -1.f),
    C_Vertex3f(0.857125044f, 0.256911159f, -1.f),
    C_Vertex3f(0.853491783f, 0.0252242088f, -1.f),
    C_Vertex3f(0.849473f, 0.268257618f, -1.f),
    C_Vertex3f(0.845894814f, 0.0144104958f, -1.f),
    C_Vertex3f(0.840727806f, 0.278197765f, -1.f),
    C_Vertex3f(0.837238789f, 0.00494849682f, -1.f),
    C_Vertex3f(0.831326962f, 0.286895394f, -1.f),
    C_Vertex3f(0.827523708f, -0.00316178799f, -1.f),
    C_Vertex3f(0.821270466f, 0.294350505f, -1.f),
    C_Vertex3f(0.816749573f, -0.00992035866f, -1.f),
    C_Vertex3f(0.815628052f, 0.152694702f, -1.f),
    C_Vertex3f(0.815628052f, 0.135931015f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.821270466f, 0.294350505f, -1.f),
    C_Vertex3f(0.810558319f, 0.300563097f, -1.f),
    C_Vertex3f(0.812005997f, 0.195366859f, -1.f),
    C_Vertex3f(0.813817024f, 0.182384968f, -1.f),
    C_Vertex3f(0.815024376f, 0.168160915f, -1.f),
    C_Vertex3f(0.815628052f, 0.152694702f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.810558319f, 0.300563097f, -1.f),
    C_Vertex3f(0.799190521f, 0.305533171f, -1.f),
    C_Vertex3f(0.802950859f, 0.22685957f, -1.f),
    C_Vertex3f(0.806572914f, 0.21760416f, -1.f),
    C_Vertex3f(0.809591293f, 0.20710659f, -1.f),
    C_Vertex3f(0.812005997f, 0.195366859f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.799190521f, 0.305533171f, -1.f),
    C_Vertex3f(0.787167072f, 0.309260726f, -1.f),
    C_Vertex3f(0.792577744f, 0.243391037f, -1.f),
    C_Vertex3f(0.796310425f, 0.238258362f, -1.f),
    C_Vertex3f(0.798725128f, 0.234872818f, -1.f),
    C_Vertex3f(0.802950859f, 0.22685957f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.761621475f, 0.258789063f, -1.f),
    C_Vertex3f(0.773366928f, 0.256222725f, -1.f),
    C_Vertex3f(0.774487972f, 0.311745763f, -1.f),
    C_Vertex3f(0.783685684f, 0.25109005f, -1.f),
    C_Vertex3f(0.787167072f, 0.309260726f, -1.f),
    C_Vertex3f(0.792577744f, 0.243391037f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.710847855f, 0.235307693f, -1.f),
    C_Vertex3f(0.716939926f, 0.243717194f, -1.f),
    C_Vertex3f(0.720975399f, 0.309281468f, -1.f),
    C_Vertex3f(0.725776672f, 0.251253128f, -1.f),
    C_Vertex3f(0.733696461f, 0.311752677f, -1.f),
    C_Vertex3f(0.736206055f, 0.256277084f, -1.f),
    C_Vertex3f(0.747107506f, 0.312988281f, -1.f),
    C_Vertex3f(0.748228073f, 0.258789063f, -1.f),
    C_Vertex3f(0.761153221f, 0.312988281f, -1.f),
    C_Vertex3f(0.761621475f, 0.258789063f, -1.f),
    C_Vertex3f(0.774487972f, 0.311745763f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.708944321f, 0.305574656f, -1.f),
    C_Vertex3f(0.697603226f, 0.300632238f, -1.f),
    C_Vertex3f(0.69973278f, 0.207295418f, -1.f),
    C_Vertex3f(0.702820301f, 0.217916012f, -1.f),
    C_Vertex3f(0.706525326f, 0.227253437f, -1.f),
    C_Vertex3f(0.710847855f, 0.235307693f, -1.f),
    C_Vertex3f(0.720975399f, 0.309281468f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.697603226f, 0.300632238f, -1.f),
    C_Vertex3f(0.686952114f, 0.294454217f, -1.f),
    C_Vertex3f(0.687915325f, -0.0044939518f, -1.f),
    C_Vertex3f(0.693557739f, 0.151981354f, -1.f),
    C_Vertex3f(0.694175243f, 0.167734623f, -1.f),
    C_Vertex3f(0.695410252f, 0.182204723f, -1.f),
    C_Vertex3f(0.697262764f, 0.195391655f, -1.f),
    C_Vertex3f(0.69973278f, 0.207295418f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.695183873f, 0.106229067f, -1.f),
    C_Vertex3f(0.694099784f, 0.120057821f, -1.f),
    C_Vertex3f(0.698627472f, -0.0108084679f, -1.f),
    C_Vertex3f(0.693557739f, 0.135217667f, -1.f),
    C_Vertex3f(0.687915325f, -0.0044939518f, -1.f),
    C_Vertex3f(0.693557739f, 0.151981354f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.687915325f, -0.0044939518f, -1.f),
    C_Vertex3f(0.686952114f, 0.294454217f, -1.f),
    C_Vertex3f(0.677858829f, 0.00308346748f, -1.f),
    C_Vertex3f(0.676990986f, 0.287040591f, -1.f),
    C_Vertex3f(0.668457985f, 0.01192379f, -1.f),
    C_Vertex3f(0.667719841f, 0.278391361f, -1.f),
    C_Vertex3f(0.659712791f, 0.0220270157f, -1.f),
    C_Vertex3f(0.65913868f, 0.268506527f, -1.f),
    C_Vertex3f(0.652060747f, 0.0336663723f, -1.f),
    C_Vertex3f(0.651630163f, 0.25714004f, -1.f),
    C_Vertex3f(0.645501852f, 0.0468418598f, -1.f),
    C_Vertex3f(0.645194292f, 0.244291902f, -1.f),
    C_Vertex3f(0.640036106f, 0.0615534782f, -1.f),
    C_Vertex3f(0.639831066f, 0.229962111f, -1.f),
    C_Vertex3f(0.635663509f, 0.0778012276f, -1.f),
    C_Vertex3f(0.635540485f, 0.214150667f, -1.f),
    C_Vertex3f(0.632384062f, 0.0955851078f, -1.f),
    C_Vertex3f(0.63232255f, 0.196857572f, -1.f),
    C_Vertex3f(0.630197763f, 0.114905119f, -1.f),
    C_Vertex3f(0.630177259f, 0.178082824f, -1.f),
    C_Vertex3f(0.629104614f, 0.157826424f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.816749573f, -0.00992035866f, -1.f),
    C_Vertex3f(0.815628052f, 0.135931015f, -1.f),
    C_Vertex3f(0.815037966f, 0.120381832f, -1.f),
    C_Vertex3f(0.813857794f, 0.106102467f, -1.f),
    C_Vertex3f(0.812087536f, 0.0930929184f, -1.f),
    C_Vertex3f(0.809727192f, 0.0813531876f, -1.f),
    C_Vertex3f(0.806776762f, 0.0708832741f, -1.f),
    C_Vertex3f(0.804916382f, -0.0153272152f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.804916382f, -0.0153272152f, -1.f),
    C_Vertex3f(0.806776762f, 0.0708832741f, -1.f),
    C_Vertex3f(0.803236246f, 0.0616831779f, -1.f),
    C_Vertex3f(0.799105644f, 0.0537528992f, -1.f),
    C_Vertex3f(0.793231964f, 0.045507431f, -1.f),
    C_Vertex3f(0.792024136f, -0.0193823576f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.784724236f, 0.0381345749f, -1.f),
    C_Vertex3f(0.774735451f, 0.0332193375f, -1.f),
    C_Vertex3f(0.778072834f, -0.0220857859f, -1.f),
    C_Vertex3f(0.792024136f, -0.0193823576f, -1.f),
    C_Vertex3f(0.793231964f, 0.045507431f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.708734989f, 0.0570516586f, -1.f),
    C_Vertex3f(0.70999527f, -0.0158600807f, -1.f),
    C_Vertex3f(0.715239525f, 0.0482883453f, -1.f),
    C_Vertex3f(0.722018719f, -0.0196487904f, -1.f),
    C_Vertex3f(0.725311756f, 0.039525032f, -1.f),
    C_Vertex3f(0.734697819f, -0.0221745968f, -1.f),
    C_Vertex3f(0.736783504f, 0.0336828232f, -1.f),
    C_Vertex3f(0.74803257f, -0.0234375f, -1.f),
    C_Vertex3f(0.74965477f, 0.0307617188f, -1.f),
    C_Vertex3f(0.763062477f, -0.0234375f, -1.f),
    C_Vertex3f(0.76326561f, 0.0307617188f, -1.f),
    C_Vertex3f(0.778072834f, -0.0220857859f, -1.f),
    C_Vertex3f(0.774735451f, 0.0332193375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.70999527f, -0.0158600807f, -1.f),
    C_Vertex3f(0.708734989f, 0.0570516586f, -1.f),
    C_Vertex3f(0.704940677f, 0.0642249584f, -1.f),
    C_Vertex3f(0.701688409f, 0.0727293491f, -1.f),
    C_Vertex3f(0.698978186f, 0.0825648308f, -1.f),
    C_Vertex3f(0.698627472f, -0.0108084679f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.696810007f, 0.0937314034f, -1.f),
    C_Vertex3f(0.695183873f, 0.106229067f, -1.f),
    C_Vertex3f(0.698627472f, -0.0108084679f, -1.f),
    C_Vertex3f(0.698978186f, 0.0825648308f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.585159302f, 0.138000488f, -1.f),
    C_Vertex3f(0.585159302f, 0.15929985f, -1.f),
    C_Vertex3f(0.584202886f, 0.117916703f, -1.f),
    C_Vertex3f(0.584059358f, 0.179084182f, -1.f),
    C_Vertex3f(0.582290053f, 0.0991572142f, -1.f),
    C_Vertex3f(0.581859469f, 0.197462201f, -1.f),
    C_Vertex3f(0.579420805f, 0.0817220211f, -1.f),
    C_Vertex3f(0.578559637f, 0.214433908f, -1.f),
    C_Vertex3f(0.57559514f, 0.065611124f, -1.f),
    C_Vertex3f(0.574159861f, 0.229999304f, -1.f),
    C_Vertex3f(0.57081306f, 0.050824523f, -1.f),
    C_Vertex3f(0.56866014f, 0.244158387f, -1.f),
    C_Vertex3f(0.565074563f, 0.0373622179f, -1.f),
    C_Vertex3f(0.562060475f, 0.256911159f, -1.f),
    C_Vertex3f(0.55837965f, 0.0252242088f, -1.f),
    C_Vertex3f(0.554360867f, 0.268257618f, -1.f),
    C_Vertex3f(0.550728321f, 0.0144104958f, -1.f),
    C_Vertex3f(0.545561314f, 0.278197765f, -1.f),
    C_Vertex3f(0.542038321f, 0.00494849682f, -1.f),
    C_Vertex3f(0.536112666f, 0.286895394f, -1.f),
    C_Vertex3f(0.532309651f, -0.00316178799f, -1.f),
    C_Vertex3f(0.526014924f, 0.294350505f, -1.f),
    C_Vertex3f(0.521542311f, -0.00992035866f, -1.f),
    C_Vertex3f(0.520706177f, 0.152694702f, -1.f),
    C_Vertex3f(0.520706177f, 0.135931015f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.526014924f, 0.294350505f, -1.f),
    C_Vertex3f(0.515268087f, 0.300563097f, -1.f),
    C_Vertex3f(0.517001152f, 0.195366859f, -1.f),
    C_Vertex3f(0.518853664f, 0.182384968f, -1.f),
    C_Vertex3f(0.520088673f, 0.168160915f, -1.f),
    C_Vertex3f(0.520706177f, 0.152694702f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.515268087f, 0.300563097f, -1.f),
    C_Vertex3f(0.503872156f, 0.305533171f, -1.f),
    C_Vertex3f(0.50773859f, 0.22685957f, -1.f),
    C_Vertex3f(0.511443615f, 0.21760416f, -1.f),
    C_Vertex3f(0.514531136f, 0.20710659f, -1.f),
    C_Vertex3f(0.517001152f, 0.195366859f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466588974f, 0.258789063f, -1.f),
    C_Vertex3f(0.478168488f, 0.256222725f, -1.f),
    C_Vertex3f(0.47913301f, 0.311745763f, -1.f),
    C_Vertex3f(0.488376617f, 0.25109005f, -1.f),
    C_Vertex3f(0.49182713f, 0.309260726f, -1.f),
    C_Vertex3f(0.497213364f, 0.243391037f, -1.f),
    C_Vertex3f(0.503872156f, 0.305533171f, -1.f),
    C_Vertex3f(0.503416061f, 0.234872818f, -1.f),
    C_Vertex3f(0.50773859f, 0.22685957f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.41592598f, 0.235307693f, -1.f),
    C_Vertex3f(0.422018051f, 0.243717194f, -1.f),
    C_Vertex3f(0.4258461f, 0.309281468f, -1.f),
    C_Vertex3f(0.430854797f, 0.251253128f, -1.f),
    C_Vertex3f(0.438484192f, 0.311752677f, -1.f),
    C_Vertex3f(0.44128418f, 0.256277084f, -1.f),
    C_Vertex3f(0.451798439f, 0.312988281f, -1.f),
    C_Vertex3f(0.453306198f, 0.258789063f, -1.f),
    C_Vertex3f(0.465789795f, 0.312988281f, -1.f),
    C_Vertex3f(0.466588974f, 0.258789063f, -1.f),
    C_Vertex3f(0.47913301f, 0.311745763f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.413884163f, 0.305574656f, -1.f),
    C_Vertex3f(0.402598381f, 0.300632238f, -1.f),
    C_Vertex3f(0.404810905f, 0.207295418f, -1.f),
    C_Vertex3f(0.407898426f, 0.217916012f, -1.f),
    C_Vertex3f(0.411603451f, 0.227253437f, -1.f),
    C_Vertex3f(0.41592598f, 0.235307693f, -1.f),
    C_Vertex3f(0.4258461f, 0.309281468f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402598381f, 0.300632238f, -1.f),
    C_Vertex3f(0.391988754f, 0.294454217f, -1.f),
    C_Vertex3f(0.39299345f, -0.0044939518f, -1.f),
    C_Vertex3f(0.398635864f, 0.151981354f, -1.f),
    C_Vertex3f(0.399253368f, 0.167734623f, -1.f),
    C_Vertex3f(0.400488377f, 0.182204723f, -1.f),
    C_Vertex3f(0.402340889f, 0.195391655f, -1.f),
    C_Vertex3f(0.404810905f, 0.207295418f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.400261998f, 0.106229067f, -1.f),
    C_Vertex3f(0.399177909f, 0.120057821f, -1.f),
    C_Vertex3f(0.403705597f, -0.0108084679f, -1.f),
    C_Vertex3f(0.398635864f, 0.135217667f, -1.f),
    C_Vertex3f(0.39299345f, -0.0044939518f, -1.f),
    C_Vertex3f(0.398635864f, 0.151981354f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.39299345f, -0.0044939518f, -1.f),
    C_Vertex3f(0.391988754f, 0.294454217f, -1.f),
    C_Vertex3f(0.382936954f, 0.00308346748f, -1.f),
    C_Vertex3f(0.382055283f, 0.287040591f, -1.f),
    C_Vertex3f(0.37353611f, 0.01192379f, -1.f),
    C_Vertex3f(0.372797966f, 0.278391361f, -1.f),
    C_Vertex3f(0.364790916f, 0.0220270157f, -1.f),
    C_Vertex3f(0.364216805f, 0.268506527f, -1.f),
    C_Vertex3f(0.357138872f, 0.0336663723f, -1.f),
    C_Vertex3f(0.356708288f, 0.25714004f, -1.f),
    C_Vertex3f(0.350579977f, 0.0468418598f, -1.f),
    C_Vertex3f(0.350272417f, 0.244291902f, -1.f),
    C_Vertex3f(0.345114231f, 0.0615534782f, -1.f),
    C_Vertex3f(0.344909191f, 0.229962111f, -1.f),
    C_Vertex3f(0.340741634f, 0.0778012276f, -1.f),
    C_Vertex3f(0.34061861f, 0.214150667f, -1.f),
    C_Vertex3f(0.337462187f, 0.0955851078f, -1.f),
    C_Vertex3f(0.337400675f, 0.196857572f, -1.f),
    C_Vertex3f(0.335275888f, 0.114905119f, -1.f),
    C_Vertex3f(0.335255384f, 0.178082824f, -1.f),
    C_Vertex3f(0.334182739f, 0.157826424f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.521542311f, -0.00992035866f, -1.f),
    C_Vertex3f(0.520706177f, 0.135931015f, -1.f),
    C_Vertex3f(0.520116091f, 0.120381832f, -1.f),
    C_Vertex3f(0.518935919f, 0.106102467f, -1.f),
    C_Vertex3f(0.517165661f, 0.0930929184f, -1.f),
    C_Vertex3f(0.514805317f, 0.0813531876f, -1.f),
    C_Vertex3f(0.511854887f, 0.0708832741f, -1.f),
    C_Vertex3f(0.5097363f, -0.0153272152f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.5097363f, -0.0153272152f, -1.f),
    C_Vertex3f(0.511854887f, 0.0708832741f, -1.f),
    C_Vertex3f(0.508314371f, 0.0616831779f, -1.f),
    C_Vertex3f(0.504183769f, 0.0537528992f, -1.f),
    C_Vertex3f(0.498310089f, 0.045507431f, -1.f),
    C_Vertex3f(0.496891618f, -0.0193823576f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.489802361f, 0.0381345749f, -1.f),
    C_Vertex3f(0.479813576f, 0.0332193375f, -1.f),
    C_Vertex3f(0.483008265f, -0.0220857859f, -1.f),
    C_Vertex3f(0.496891618f, -0.0193823576f, -1.f),
    C_Vertex3f(0.498310089f, 0.045507431f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.413813114f, 0.0570516586f, -1.f),
    C_Vertex3f(0.415073395f, -0.0158600807f, -1.f),
    C_Vertex3f(0.42031765f, 0.0482883453f, -1.f),
    C_Vertex3f(0.427096844f, -0.0196487904f, -1.f),
    C_Vertex3f(0.430389881f, 0.039525032f, -1.f),
    C_Vertex3f(0.439775944f, -0.0221745968f, -1.f),
    C_Vertex3f(0.441861629f, 0.0336828232f, -1.f),
    C_Vertex3f(0.453110695f, -0.0234375f, -1.f),
    C_Vertex3f(0.454732895f, 0.0307617188f, -1.f),
    C_Vertex3f(0.468086243f, -0.0234375f, -1.f),
    C_Vertex3f(0.468343735f, 0.0307617188f, -1.f),
    C_Vertex3f(0.483008265f, -0.0220857859f, -1.f),
    C_Vertex3f(0.479813576f, 0.0332193375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.415073395f, -0.0158600807f, -1.f),
    C_Vertex3f(0.413813114f, 0.0570516586f, -1.f),
    C_Vertex3f(0.410018802f, 0.0642249584f, -1.f),
    C_Vertex3f(0.406766534f, 0.0727293491f, -1.f),
    C_Vertex3f(0.404056311f, 0.0825648308f, -1.f),
    C_Vertex3f(0.403705597f, -0.0108084679f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.401888132f, 0.0937314034f, -1.f),
    C_Vertex3f(0.400261998f, 0.106229067f, -1.f),
    C_Vertex3f(0.403705597f, -0.0108084679f, -1.f),
    C_Vertex3f(0.404056311f, 0.0825648308f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.335275888f, 0.114905119f, -1.f),
    C_Vertex3f(0.334182739f, 0.157826424f, -1.f),
    C_Vertex3f(0.334182739f, 0.135761261f, -1.f),
    C_Vertex3f(0.630197763f, 0.114905119f, -1.f),
    C_Vertex3f(0.629104614f, 0.157826424f, -1.f),
    C_Vertex3f(0.629104614f, 0.135761261f, -1.f),
    C_Vertex3f(0.0174047947f, 0.448366046f, -1.f),
    C_Vertex3f(0.0163116455f, 0.491300583f, -1.f),
    C_Vertex3f(0.0163116455f, 0.469232559f, -1.f),
    C_End(),
    C_Translatef(0.895019531f, 0.f, 0.f),
};
static const Call char_138[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.0850564986f, 0.996376157f, 0.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, 0.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, -1.f),
    C_Vertex3f(0.120117188f, 0.212890625f, -1.f),
    C_Vertex3f(0.120117188f, 0.212890625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.993104398f, 0.117233843f, 0.f),
    C_Vertex3f(0.120117188f, 0.212890625f, 0.f),
    C_Vertex3f(0.120117188f, 0.212890625f, -1.f),
    C_Normal3f(0.98398751f, 0.17823717f, 0.f),
    C_Vertex3f(0.121538162f, 0.200853348f, 0.f),
    C_Vertex3f(0.121538162f, 0.200853348f, -1.f),
    C_Normal3f(0.964875579f, 0.262707204f, 0.f),
    C_Vertex3f(0.126266956f, 0.178939819f, 0.f),
    C_Vertex3f(0.126266956f, 0.178939819f, -1.f),
    C_Normal3f(0.926923394f, 0.375250369f, 0.f),
    C_Vertex3f(0.132882595f, 0.159187317f, 0.f),
    C_Vertex3f(0.132882595f, 0.159187317f, -1.f),
    C_Normal3f(0.867992342f, 0.496577561f, 0.f),
    C_Vertex3f(0.141385078f, 0.14159584f, 0.f),
    C_Vertex3f(0.141385078f, 0.14159584f, -1.f),
    C_Normal3f(0.77902025f, 0.626998723f, 0.f),
    C_Vertex3f(0.151774406f, 0.12616539f, 0.f),
    C_Vertex3f(0.151774406f, 0.12616539f, -1.f),
    C_Normal3f(0.668635428f, 0.743590415f, 0.f),
    C_Vertex3f(0.165034294f, 0.112212658f, 0.f),
    C_Vertex3f(0.165034294f, 0.112212658f, -1.f),
    C_Normal3f(0.555541098f, 0.831489027f, 0.f),
    C_Vertex3f(0.181164742f, 0.0997376442f, 0.f),
    C_Vertex3f(0.181164742f, 0.0997376442f, -1.f),
    C_Normal3f(0.448622704f, 0.893721223f, 0.f),
    C_Vertex3f(0.200165749f, 0.0887403488f, 0.f),
    C_Vertex3f(0.200165749f, 0.0887403488f, -1.f),
    C_Normal3f(0.346432209f, 0.938074946f, 0.f),
    C_Vertex3f(0.222037315f, 0.0792207718f, 0.f),
    C_Vertex3f(0.222037315f, 0.0792207718f, -1.f),
    C_Normal3f(0.241358921f, 0.970435977f, 0.f),
    C_Vertex3f(0.245275974f, 0.072081089f, 0.f),
    C_Vertex3f(0.245275974f, 0.072081089f, -1.f),
    C_Normal3f(0.139774442f, 0.990183294f, 0.f),
    C_Vertex3f(0.269881725f, 0.0673213005f, 0.f),
    C_Vertex3f(0.269881725f, 0.0673213005f, -1.f),
    C_Normal3f(0.0459121391f, 0.998945475f, 0.f),
    C_Vertex3f(0.295854568f, 0.0649414063f, 0.f),
    C_Vertex3f(0.295854568f, 0.0649414063f, -1.f),
    C_Normal3f(-0.0370819233f, 0.999312282f, 0.f),
    C_Vertex3f(0.321662903f, 0.0649414063f, 0.f),
    C_Vertex3f(0.321662903f, 0.0649414063f, -1.f),
    C_Normal3f(-0.121412903f, 0.99260205f, 0.f),
    C_Vertex3f(0.344518185f, 0.0667471886f, 0.f),
    C_Vertex3f(0.344518185f, 0.0667471886f, -1.f),
    C_Normal3f(-0.212854326f, 0.977083981f, 0.f),
    C_Vertex3f(0.365952015f, 0.0703587532f, 0.f),
    C_Vertex3f(0.365952015f, 0.0703587532f, -1.f),
    C_Normal3f(-0.311187327f, 0.950348616f, 0.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, 0.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, -1.f),
    C_Normal3f(-0.413075179f, 0.910696864f, 0.f),
    C_Vertex3f(0.404555321f, 0.0829992294f, 0.f),
    C_Vertex3f(0.404555321f, 0.0829992294f, -1.f),
    C_Normal3f(-0.523600936f, 0.85196358f, 0.f),
    C_Vertex3f(0.420767784f, 0.0915622711f, 0.f),
    C_Vertex3f(0.420767784f, 0.0915622711f, -1.f),
    C_Normal3f(-0.641458333f, 0.767157912f, 0.f),
    C_Vertex3f(0.434601784f, 0.101465225f, 0.f),
    C_Vertex3f(0.434601784f, 0.101465225f, -1.f),
    C_Normal3f(-0.757515967f, 0.652816653f, 0.f),
    C_Vertex3f(0.44605732f, 0.112708092f, 0.f),
    C_Vertex3f(0.44605732f, 0.112708092f, -1.f),
    C_Normal3f(-0.850961983f, 0.525227368f, 0.f),
    C_Vertex3f(0.455134392f, 0.125290871f, 0.f),
    C_Vertex3f(0.455134392f, 0.125290871f, -1.f),
    C_Normal3f(-0.921275377f, 0.388911009f, 0.f),
    C_Vertex3f(0.461942196f, 0.138444424f, 0.f),
    C_Vertex3f(0.461942196f, 0.138444424f, -1.f),
    C_Normal3f(-0.971729159f, 0.236098394f, 0.f),
    C_Vertex3f(0.466480732f, 0.152168751f, 0.f),
    C_Vertex3f(0.466480732f, 0.152168751f, -1.f),
    C_Normal3f(-0.997008026f, 0.0772976428f, 0.f),
    C_Vertex3f(0.46875f, 0.166463852f, 0.f),
    C_Vertex3f(0.46875f, 0.166463852f, -1.f),
    C_Normal3f(-0.997219086f, -0.0745259076f, 0.f),
    C_Vertex3f(0.46875f, 0.181438446f, 0.f),
    C_Vertex3f(0.46875f, 0.181438446f, -1.f),
    C_Normal3f(-0.972602963f, -0.232472599f, 0.f),
    C_Vertex3f(0.466575146f, 0.195565224f, 0.f),
    C_Vertex3f(0.466575146f, 0.195565224f, -1.f),
    C_Normal3f(-0.919204891f, -0.393779665f, 0.f),
    C_Vertex3f(0.462225437f, 0.208735466f, 0.f),
    C_Vertex3f(0.462225437f, 0.208735466f, -1.f),
    C_Normal3f(-0.838967443f, -0.544181705f, 0.f),
    C_Vertex3f(0.455700874f, 0.220949173f, 0.f),
    C_Vertex3f(0.455700874f, 0.220949173f, -1.f),
    C_Normal3f(-0.729361534f, -0.684128463f, 0.f),
    C_Vertex3f(0.447001457f, 0.232206345f, 0.f),
    C_Vertex3f(0.447001457f, 0.232206345f, -1.f),
    C_Normal3f(-0.605762899f, -0.795645237f, 0.f),
    C_Vertex3f(0.435428143f, 0.242562294f, 0.f),
    C_Vertex3f(0.435428143f, 0.242562294f, -1.f),
    C_Normal3f(-0.493140727f, -0.869949579f, 0.f),
    C_Vertex3f(0.42098093f, 0.252017021f, 0.f),
    C_Vertex3f(0.42098093f, 0.252017021f, -1.f),
    C_Normal3f(-0.406449646f, -0.913673222f, 0.f),
    C_Vertex3f(0.403659821f, 0.260570526f, 0.f),
    C_Vertex3f(0.403659821f, 0.260570526f, -1.f),
    C_Normal3f(-0.342922598f, -0.939363658f, 0.f),
    C_Vertex3f(0.390324593f, 0.265654564f, 0.f),
    C_Vertex3f(0.390324593f, 0.265654564f, -1.f),
    C_Normal3f(-0.307028621f, -0.951700211f, 0.f),
    C_Vertex3f(0.381887197f, 0.268518806f, 0.f),
    C_Vertex3f(0.381887197f, 0.268518806f, -1.f),
    C_Normal3f(-0.285247117f, -0.958454013f, 0.f),
    C_Vertex3f(0.371487856f, 0.271731496f, 0.f),
    C_Vertex3f(0.371487856f, 0.271731496f, -1.f),
    C_Normal3f(-0.269598514f, -0.962972879f, 0.f),
    C_Vertex3f(0.359126568f, 0.275292635f, 0.f),
    C_Vertex3f(0.359126568f, 0.275292635f, -1.f),
    C_Normal3f(-0.257821202f, -0.966192603f, 0.f),
    C_Vertex3f(0.344803333f, 0.279202223f, 0.f),
    C_Vertex3f(0.344803333f, 0.279202223f, -1.f),
    C_Normal3f(-0.248640731f, -0.968595803f, 0.f),
    C_Vertex3f(0.328518152f, 0.283460259f, 0.f),
    C_Vertex3f(0.328518152f, 0.283460259f, -1.f),
    C_Normal3f(-0.241285399f, -0.970454216f, 0.f),
    C_Vertex3f(0.310271025f, 0.288066745f, 0.f),
    C_Vertex3f(0.310271025f, 0.288066745f, -1.f),
    C_Normal3f(-0.235261127f, -0.971932232f, 0.f),
    C_Vertex3f(0.290061951f, 0.293021679f, 0.f),
    C_Vertex3f(0.290061951f, 0.293021679f, -1.f),
    C_Normal3f(-0.239585206f, -0.970875382f, 0.f),
    C_Vertex3f(0.26789093f, 0.298325062f, 0.f),
    C_Vertex3f(0.26789093f, 0.298325062f, -1.f),
    C_Normal3f(-0.255102754f, -0.966913939f, 0.f),
    C_Vertex3f(0.247384787f, 0.303553224f, 0.f),
    C_Vertex3f(0.247384787f, 0.303553224f, -1.f),
    C_Normal3f(-0.273236841f, -0.961946726f, 0.f),
    C_Vertex3f(0.22854352f, 0.308706164f, 0.f),
    C_Vertex3f(0.22854352f, 0.308706164f, -1.f),
    C_Normal3f(-0.294684052f, -0.955594778f, 0.f),
    C_Vertex3f(0.21136713f, 0.313783884f, 0.f),
    C_Vertex3f(0.21136713f, 0.313783884f, -1.f),
    C_Normal3f(-0.32039851f, -0.947282851f, 0.f),
    C_Vertex3f(0.195855618f, 0.318786383f, 0.f),
    C_Vertex3f(0.195855618f, 0.318786383f, -1.f),
    C_Normal3f(-0.351712555f, -0.936108053f, 0.f),
    C_Vertex3f(0.182008982f, 0.32371366f, 0.f),
    C_Vertex3f(0.182008982f, 0.32371366f, -1.f),
    C_Normal3f(-0.390521169f, -0.920593977f, 0.f),
    C_Vertex3f(0.169827223f, 0.328565717f, 0.f),
    C_Vertex3f(0.169827223f, 0.328565717f, -1.f),
    C_Normal3f(-0.445337921f, -0.895362616f, 0.f),
    C_Vertex3f(0.159310341f, 0.333342552f, 0.f),
    C_Vertex3f(0.159310341f, 0.333342552f, -1.f),
    C_Normal3f(-0.519384444f, -0.854540706f, 0.f),
    C_Vertex3f(0.143409729f, 0.341705322f, 0.f),
    C_Vertex3f(0.143409729f, 0.341705322f, -1.f),
    C_Normal3f(-0.607493222f, -0.794324815f, 0.f),
    C_Vertex3f(0.123383999f, 0.355178356f, 0.f),
    C_Vertex3f(0.123383999f, 0.355178356f, -1.f),
    C_Normal3f(-0.70724684f, -0.706966639f, 0.f),
    C_Vertex3f(0.106281757f, 0.370100498f, 0.f),
    C_Vertex3f(0.106281757f, 0.370100498f, -1.f),
    C_Normal3f(-0.802355707f, -0.596846163f, 0.f),
    C_Vertex3f(0.0921030045f, 0.386471748f, 0.f),
    C_Vertex3f(0.0921030045f, 0.386471748f, -1.f),
    C_Normal3f(-0.881154537f, -0.472828329f, 0.f),
    C_Vertex3f(0.0808477402f, 0.404292107f, 0.f),
    C_Vertex3f(0.0808477402f, 0.404292107f, -1.f),
    C_Normal3f(-0.940210938f, -0.340592563f, 0.f),
    C_Vertex3f(0.072406292f, 0.423178196f, 0.f),
    C_Vertex3f(0.072406292f, 0.423178196f, -1.f),
    C_Normal3f(-0.979426026f, -0.201803774f, 0.f),
    C_Vertex3f(0.0667786598f, 0.443130016f, 0.f),
    C_Vertex3f(0.0667786598f, 0.443130016f, -1.f),
    C_Normal3f(-0.997979105f, -0.0635413378f, 0.f),
    C_Vertex3f(0.0639648438f, 0.464147568f, 0.f),
    C_Vertex3f(0.0639648438f, 0.464147568f, -1.f),
    C_Normal3f(-0.997278392f, 0.0737271234f, 0.f),
    C_Vertex3f(0.0639648438f, 0.487323761f, 0.f),
    C_Vertex3f(0.0639648438f, 0.487323761f, -1.f),
    C_Normal3f(-0.975787222f, 0.218722045f, 0.f),
    C_Vertex3f(0.0674138069f, 0.510800362f, 0.f),
    C_Vertex3f(0.0674138069f, 0.510800362f, -1.f),
    C_Normal3f(-0.932640553f, 0.360806853f, 0.f),
    C_Vertex3f(0.0743117332f, 0.533484459f, 0.f),
    C_Vertex3f(0.0743117332f, 0.533484459f, -1.f),
    C_Normal3f(-0.876705945f, 0.481026679f, 0.f),
    C_Vertex3f(0.0846586227f, 0.555376053f, 0.f),
    C_Vertex3f(0.0846586227f, 0.555376053f, -1.f),
    C_Normal3f(-0.818909228f, 0.573922992f, 0.f),
    C_Vertex3f(0.0950055122f, 0.571200371f, 0.f),
    C_Vertex3f(0.0950055122f, 0.571200371f, -1.f),
    C_Normal3f(-0.764627099f, 0.644472897f, 0.f),
    C_Vertex3f(0.102697134f, 0.581114531f, 0.f),
    C_Vertex3f(0.102697134f, 0.581114531f, -1.f),
    C_Normal3f(-0.710227191f, 0.703972518f, 0.f),
    C_Vertex3f(0.111182451f, 0.590393305f, 0.f),
    C_Vertex3f(0.111182451f, 0.590393305f, -1.f),
    C_Normal3f(-0.652239919f, 0.758012652f, 0.f),
    C_Vertex3f(0.120461464f, 0.599036694f, 0.f),
    C_Vertex3f(0.120461464f, 0.599036694f, -1.f),
    C_Normal3f(-0.591996014f, 0.805940866f, 0.f),
    C_Vertex3f(0.130534172f, 0.607044697f, 0.f),
    C_Vertex3f(0.130534172f, 0.607044697f, -1.f),
    C_Normal3f(-0.530831516f, 0.847477317f, 0.f),
    C_Vertex3f(0.141400576f, 0.614417315f, 0.f),
    C_Vertex3f(0.141400576f, 0.614417315f, -1.f),
    C_Normal3f(-0.469970703f, 0.882681966f, 0.f),
    C_Vertex3f(0.153060675f, 0.621154547f, 0.f),
    C_Vertex3f(0.153060675f, 0.621154547f, -1.f),
    C_Normal3f(-0.404598981f, 0.914494216f, 0.f),
    C_Vertex3f(0.165514469f, 0.627256393f, 0.f),
    C_Vertex3f(0.165514469f, 0.627256393f, -1.f),
    C_Normal3f(-0.324249119f, 0.945971787f, 0.f),
    C_Vertex3f(0.185385704f, 0.635456085f, 0.f),
    C_Vertex3f(0.185385704f, 0.635456085f, -1.f),
    C_Normal3f(-0.23152405f, 0.972829163f, 0.f),
    C_Vertex3f(0.213358402f, 0.643655777f, 0.f),
    C_Vertex3f(0.213358402f, 0.643655777f, -1.f),
    C_Normal3f(-0.134569451f, 0.990904152f, 0.f),
    C_Vertex3f(0.242808819f, 0.649122238f, 0.f),
    C_Vertex3f(0.242808819f, 0.649122238f, -1.f),
    C_Normal3f(-0.0420266986f, 0.99911654f, 0.f),
    C_Vertex3f(0.273736954f, 0.651855469f, 0.f),
    C_Vertex3f(0.273736954f, 0.651855469f, -1.f),
    C_Normal3f(0.0423609316f, 0.999102354f, 0.f),
    C_Vertex3f(0.307786942f, 0.651855469f, 0.f),
    C_Vertex3f(0.307786942f, 0.651855469f, -1.f),
    C_Normal3f(0.131084278f, 0.991371155f, 0.f),
    C_Vertex3f(0.341372967f, 0.64898777f, 0.f),
    C_Vertex3f(0.341372967f, 0.64898777f, -1.f),
    C_Normal3f(0.229363292f, 0.973340929f, 0.f),
    C_Vertex3f(0.372850895f, 0.643252373f, 0.f),
    C_Vertex3f(0.372850895f, 0.643252373f, -1.f),
    C_Normal3f(0.326469094f, 0.945207834f, 0.f),
    C_Vertex3f(0.402220726f, 0.634649277f, 0.f),
    C_Vertex3f(0.402220726f, 0.634649277f, -1.f),
    C_Normal3f(0.411136717f, 0.911573768f, 0.f),
    C_Vertex3f(0.422667027f, 0.626046181f, 0.f),
    C_Vertex3f(0.422667027f, 0.626046181f, -1.f),
    C_Normal3f(0.47657907f, 0.879131556f, 0.f),
    C_Vertex3f(0.435510993f, 0.619634748f, 0.f),
    C_Vertex3f(0.435510993f, 0.619634748f, -1.f),
    C_Normal3f(0.537039518f, 0.84355706f, 0.f),
    C_Vertex3f(0.447568059f, 0.612547278f, 0.f),
    C_Vertex3f(0.447568059f, 0.612547278f, -1.f),
    C_Normal3f(0.597353518f, 0.801978111f, 0.f),
    C_Vertex3f(0.458838224f, 0.604783773f, 0.f),
    C_Vertex3f(0.458838224f, 0.604783773f, -1.f),
    C_Normal3f(0.656340778f, 0.754464507f, 0.f),
    C_Vertex3f(0.469321489f, 0.596344233f, 0.f),
    C_Vertex3f(0.469321489f, 0.596344233f, -1.f),
    C_Normal3f(0.712764323f, 0.701403618f, 0.f),
    C_Vertex3f(0.479017854f, 0.587228656f, 0.f),
    C_Vertex3f(0.479017854f, 0.587228656f, -1.f),
    C_Normal3f(0.765439689f, 0.643507659f, 0.f),
    C_Vertex3f(0.487927318f, 0.577437043f, 0.f),
    C_Vertex3f(0.487927318f, 0.577437043f, -1.f),
    C_Normal3f(0.817846715f, 0.575436234f, 0.f),
    C_Vertex3f(0.496049881f, 0.566969395f, 0.f),
    C_Vertex3f(0.496049881f, 0.566969395f, -1.f),
    C_Normal3f(0.876051962f, 0.482216626f, 0.f),
    C_Vertex3f(0.507053375f, 0.550253868f, 0.f),
    C_Vertex3f(0.507053375f, 0.550253868f, -1.f),
    C_Normal3f(0.931329012f, 0.364178926f, 0.f),
    C_Vertex3f(0.518330574f, 0.526491642f, 0.f),
    C_Vertex3f(0.518330574f, 0.526491642f, -1.f),
    C_Normal3f(0.972792685f, 0.231677249f, 0.f),
    C_Vertex3f(0.526213646f, 0.501254559f, 0.f),
    C_Vertex3f(0.526213646f, 0.501254559f, -1.f),
    C_Normal3f(0.992469192f, 0.122494414f, 0.f),
    C_Vertex3f(0.530702591f, 0.474542618f, 0.f),
    C_Vertex3f(0.530702591f, 0.474542618f, -1.f),
    C_Normal3f(0.999246538f, 0.0388122536f, 0.f),
    C_Vertex3f(0.53125f, 0.460449219f, 0.f),
    C_Vertex3f(0.53125f, 0.460449219f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.0780742019f, -0.996947527f, 0.f),
    C_Vertex3f(0.53125f, 0.460449219f, 0.f),
    C_Vertex3f(0.53125f, 0.460449219f, -1.f),
    C_Vertex3f(0.450195313f, 0.454101563f, -1.f),
    C_Vertex3f(0.450195313f, 0.454101563f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99420023f, -0.1075451f, 0.f),
    C_Vertex3f(0.450195313f, 0.454101563f, 0.f),
    C_Vertex3f(0.450195313f, 0.454101563f, -1.f),
    C_Normal3f(-0.987811804f, -0.155653521f, 0.f),
    C_Vertex3f(0.449372292f, 0.461709976f, 0.f),
    C_Vertex3f(0.449372292f, 0.461709976f, -1.f),
    C_Normal3f(-0.97549212f, -0.22003451f, 0.f),
    C_Vertex3f(0.446745276f, 0.475996256f, 0.f),
    C_Vertex3f(0.446745276f, 0.475996256f, -1.f),
    C_Normal3f(-0.952990592f, -0.302999943f, 0.f),
    C_Vertex3f(0.443137288f, 0.489351988f, 0.f),
    C_Vertex3f(0.443137288f, 0.489351988f, -1.f),
    C_Normal3f(-0.920428216f, -0.390911579f, 0.f),
    C_Vertex3f(0.438548326f, 0.501777172f, 0.f),
    C_Vertex3f(0.438548326f, 0.501777172f, -1.f),
    C_Normal3f(-0.876408041f, -0.48156926f, 0.f),
    C_Vertex3f(0.432978392f, 0.513271809f, 0.f),
    C_Vertex3f(0.432978392f, 0.513271809f, -1.f),
    C_Normal3f(-0.820290387f, -0.571947157f, 0.f),
    C_Vertex3f(0.426427484f, 0.523835897f, 0.f),
    C_Vertex3f(0.426427484f, 0.523835897f, -1.f),
    C_Normal3f(-0.752571285f, -0.658510864f, 0.f),
    C_Vertex3f(0.418895602f, 0.533469439f, 0.f),
    C_Vertex3f(0.418895602f, 0.533469439f, -1.f),
    C_Normal3f(-0.675041139f, -0.737780094f, 0.f),
    C_Vertex3f(0.410382748f, 0.542172432f, 0.f),
    C_Vertex3f(0.410382748f, 0.542172432f, -1.f),
    C_Normal3f(-0.586615622f, -0.809865475f, 0.f),
    C_Vertex3f(0.40088892f, 0.549944878f, 0.f),
    C_Vertex3f(0.40088892f, 0.549944878f, -1.f),
    C_Normal3f(-0.491431266f, -0.870916426f, 0.f),
    C_Vertex3f(0.3902632f, 0.556745768f, 0.f),
    C_Vertex3f(0.3902632f, 0.556745768f, -1.f),
    C_Normal3f(-0.397817284f, -0.917464733f, 0.f),
    C_Vertex3f(0.378505588f, 0.562575102f, 0.f),
    C_Vertex3f(0.378505588f, 0.562575102f, -1.f),
    C_Normal3f(-0.309020817f, -0.951055288f, 0.f),
    C_Vertex3f(0.365616083f, 0.56743288f, 0.f),
    C_Vertex3f(0.365616083f, 0.56743288f, -1.f),
    C_Normal3f(-0.227022663f, -0.97388941f, 0.f),
    C_Vertex3f(0.351594687f, 0.571319103f, 0.f),
    C_Vertex3f(0.351594687f, 0.571319103f, -1.f),
    C_Normal3f(-0.15270482f, -0.988271832f, 0.f),
    C_Vertex3f(0.336441398f, 0.57423377f, 0.f),
    C_Vertex3f(0.336441398f, 0.57423377f, -1.f),
    C_Normal3f(-0.0861612409f, -0.996281266f, 0.f),
    C_Vertex3f(0.320156217f, 0.576176882f, 0.f),
    C_Vertex3f(0.320156217f, 0.576176882f, -1.f),
    C_Normal3f(-0.0267179906f, -0.999642968f, 0.f),
    C_Vertex3f(0.302739143f, 0.577148438f, 0.f),
    C_Vertex3f(0.302739143f, 0.577148438f, -1.f),
    C_Normal3f(0.0238604303f, -0.999715269f, 0.f),
    C_Vertex3f(0.283805847f, 0.577148438f, 0.f),
    C_Vertex3f(0.283805847f, 0.577148438f, -1.f),
    C_Normal3f(0.0759948045f, -0.997108161f, 0.f),
    C_Vertex3f(0.26579845f, 0.576266766f, 0.f),
    C_Vertex3f(0.26579845f, 0.576266766f, -1.f),
    C_Normal3f(0.136121199f, -0.990692139f, 0.f),
    C_Vertex3f(0.249101281f, 0.574503422f, 0.f),
    C_Vertex3f(0.249101281f, 0.574503422f, -1.f),
    C_Normal3f(0.205018878f, -0.978758037f, 0.f),
    C_Vertex3f(0.233714342f, 0.571858406f, 0.f),
    C_Vertex3f(0.233714342f, 0.571858406f, -1.f),
    C_Normal3f(0.283481002f, -0.958977818f, 0.f),
    C_Vertex3f(0.219637632f, 0.568331718f, 0.f),
    C_Vertex3f(0.219637632f, 0.568331718f, -1.f),
    C_Normal3f(0.371697843f, -0.928353667f, 0.f),
    C_Vertex3f(0.206871152f, 0.563923359f, 0.f),
    C_Vertex3f(0.206871152f, 0.563923359f, -1.f),
    C_Normal3f(0.468693584f, -0.883360744f, 0.f),
    C_Vertex3f(0.195414901f, 0.558633327f, 0.f),
    C_Vertex3f(0.195414901f, 0.558633327f, -1.f),
    C_Normal3f(0.571656764f, -0.820492864f, 0.f),
    C_Vertex3f(0.185268879f, 0.552461624f, 0.f),
    C_Vertex3f(0.185268879f, 0.552461624f, -1.f),
    C_Normal3f(0.657800972f, -0.75319171f, 0.f),
    C_Vertex3f(0.176433086f, 0.545408249f, 0.f),
    C_Vertex3f(0.176433086f, 0.545408249f, -1.f),
    C_Normal3f(0.726798058f, -0.686851263f, 0.f),
    C_Vertex3f(0.168701768f, 0.537992716f, 0.f),
    C_Vertex3f(0.168701768f, 0.537992716f, -1.f),
    C_Normal3f(0.794912338f, -0.606724441f, 0.f),
    C_Vertex3f(0.162074924f, 0.530215025f, 0.f),
    C_Vertex3f(0.162074924f, 0.530215025f, -1.f),
    C_Normal3f(0.858511865f, -0.512793601f, 0.f),
    C_Vertex3f(0.156552553f, 0.522075176f, 0.f),
    C_Vertex3f(0.156552553f, 0.522075176f, -1.f),
    C_Normal3f(0.913557231f, -0.406710356f, 0.f),
    C_Vertex3f(0.152134657f, 0.51357317f, 0.f),
    C_Vertex3f(0.152134657f, 0.51357317f, -1.f),
    C_Normal3f(0.956429482f, -0.291963279f, 0.f),
    C_Vertex3f(0.148821235f, 0.504709005f, 0.f),
    C_Vertex3f(0.148821235f, 0.504709005f, -1.f),
    C_Normal3f(0.984844685f, -0.173438251f, 0.f),
    C_Vertex3f(0.146612287f, 0.495482683f, 0.f),
    C_Vertex3f(0.146612287f, 0.495482683f, -1.f),
    C_Normal3f(0.998868942f, -0.0475483239f, 0.f),
    C_Vertex3f(0.145507813f, 0.485894203f, 0.f),
    C_Vertex3f(0.145507813f, 0.485894203f, -1.f),
    C_Normal3f(0.994372308f, 0.105942346f, 0.f),
    C_Vertex3f(0.145507813f, 0.472280502f, 0.f),
    C_Vertex3f(0.145507813f, 0.472280502f, -1.f),
    C_Normal3f(0.954287887f, 0.298888922f, 0.f),
    C_Vertex3f(0.148635864f, 0.456534386f, 0.f),
    C_Vertex3f(0.148635864f, 0.456534386f, -1.f),
    C_Normal3f(0.86449939f, 0.50263375f, 0.f),
    C_Vertex3f(0.154891968f, 0.442318916f, 0.f),
    C_Vertex3f(0.154891968f, 0.442318916f, -1.f),
    C_Normal3f(0.747366309f, 0.664412141f, 0.f),
    C_Vertex3f(0.164276123f, 0.429634094f, 0.f),
    C_Vertex3f(0.164276123f, 0.429634094f, -1.f),
    C_Normal3f(0.614301443f, 0.789071441f, 0.f),
    C_Vertex3f(0.173605919f, 0.421268463f, 0.f),
    C_Vertex3f(0.173605919f, 0.421268463f, -1.f),
    C_Normal3f(0.495683849f, 0.868503153f, 0.f),
    C_Vertex3f(0.182229519f, 0.415657163f, 0.f),
    C_Vertex3f(0.182229519f, 0.415657163f, -1.f),
    C_Normal3f(0.417125732f, 0.908848763f, 0.f),
    C_Vertex3f(0.193329334f, 0.410011649f, 0.f),
    C_Vertex3f(0.193329334f, 0.410011649f, -1.f),
    C_Normal3f(0.358929008f, 0.933364928f, 0.f),
    C_Vertex3f(0.206905365f, 0.404331923f, 0.f),
    C_Vertex3f(0.206905365f, 0.404331923f, -1.f),
    C_Normal3f(0.314626217f, 0.949215591f, 0.f),
    C_Vertex3f(0.222957611f, 0.398617983f, 0.f),
    C_Vertex3f(0.222957611f, 0.398617983f, -1.f),
    C_Normal3f(0.280000448f, 0.959999919f, 0.f),
    C_Vertex3f(0.241486073f, 0.39286983f, 0.f),
    C_Vertex3f(0.241486073f, 0.39286983f, -1.f),
    C_Normal3f(0.252299994f, 0.967649102f, 0.f),
    C_Vertex3f(0.262490749f, 0.387087464f, 0.f),
    C_Vertex3f(0.262490749f, 0.387087464f, -1.f),
    C_Normal3f(0.229690298f, 0.973263741f, 0.f),
    C_Vertex3f(0.285971642f, 0.381270885f, 0.f),
    C_Vertex3f(0.285971642f, 0.381270885f, -1.f),
    C_Normal3f(0.225008681f, 0.974356711f, 0.f),
    C_Vertex3f(0.311928749f, 0.375420094f, 0.f),
    C_Vertex3f(0.311928749f, 0.375420094f, -1.f),
    C_Normal3f(0.236557841f, 0.971617401f, 0.f),
    C_Vertex3f(0.335841417f, 0.369754434f, 0.f),
    C_Vertex3f(0.335841417f, 0.369754434f, -1.f),
    C_Normal3f(0.250241429f, 0.968183458f, 0.f),
    C_Vertex3f(0.357709646f, 0.364273906f, 0.f),
    C_Vertex3f(0.357709646f, 0.364273906f, -1.f),
    C_Normal3f(0.266700119f, 0.963779569f, 0.f),
    C_Vertex3f(0.377533436f, 0.35897851f, 0.f),
    C_Vertex3f(0.377533436f, 0.35897851f, -1.f),
    C_Normal3f(0.286852658f, 0.957974672f, 0.f),
    C_Vertex3f(0.395312786f, 0.353868246f, 0.f),
    C_Vertex3f(0.395312786f, 0.353868246f, -1.f),
    C_Normal3f(0.31205979f, 0.950062454f, 0.f),
    C_Vertex3f(0.411047697f, 0.348943114f, 0.f),
    C_Vertex3f(0.411047697f, 0.348943114f, -1.f),
    C_Normal3f(0.344411969f, 0.938818634f, 0.f),
    C_Vertex3f(0.424738169f, 0.344203115f, 0.f),
    C_Vertex3f(0.424738169f, 0.344203115f, -1.f),
    C_Normal3f(0.391262203f, 0.920279264f, 0.f),
    C_Vertex3f(0.436384201f, 0.339648247f, 0.f),
    C_Vertex3f(0.436384201f, 0.339648247f, -1.f),
    C_Normal3f(0.44322136f, 0.896412134f, 0.f),
    C_Vertex3f(0.448167801f, 0.334241867f, 0.f),
    C_Vertex3f(0.448167801f, 0.334241867f, -1.f),
    C_Normal3f(0.490454525f, 0.871466875f, 0.f),
    C_Vertex3f(0.461219668f, 0.327368617f, 0.f),
    C_Vertex3f(0.461219668f, 0.327368617f, -1.f),
    C_Normal3f(0.541226506f, 0.840876758f, 0.f),
    C_Vertex3f(0.473357797f, 0.320065141f, 0.f),
    C_Vertex3f(0.473357797f, 0.320065141f, -1.f),
    C_Normal3f(0.593918145f, 0.804525495f, 0.f),
    C_Vertex3f(0.484582186f, 0.312331438f, 0.f),
    C_Vertex3f(0.484582186f, 0.312331438f, -1.f),
    C_Normal3f(0.647797883f, 0.761812329f, 0.f),
    C_Vertex3f(0.494892836f, 0.304167509f, 0.f),
    C_Vertex3f(0.494892836f, 0.304167509f, -1.f),
    C_Normal3f(0.701877952f, 0.712297261f, 0.f),
    C_Vertex3f(0.504289746f, 0.295573354f, 0.f),
    C_Vertex3f(0.504289746f, 0.295573354f, -1.f),
    C_Normal3f(0.754930794f, 0.655804515f, 0.f),
    C_Vertex3f(0.512772918f, 0.286548972f, 0.f),
    C_Vertex3f(0.512772918f, 0.286548972f, -1.f),
    C_Normal3f(0.810424566f, 0.585843086f, 0.f),
    C_Vertex3f(0.52034235f, 0.277094364f, 0.f),
    C_Vertex3f(0.52034235f, 0.277094364f, -1.f),
    C_Normal3f(0.875063002f, 0.484009027f, 0.f),
    C_Vertex3f(0.53032589f, 0.262267113f, 0.f),
    C_Vertex3f(0.53032589f, 0.262267113f, -1.f),
    C_Normal3f(0.935477078f, 0.353387594f, 0.f),
    C_Vertex3f(0.540309429f, 0.24099493f, 0.f),
    C_Vertex3f(0.540309429f, 0.24099493f, -1.f),
    C_Normal3f(0.978222191f, 0.207560554f, 0.f),
    C_Vertex3f(0.546965122f, 0.218220234f, 0.f),
    C_Vertex3f(0.546965122f, 0.218220234f, -1.f),
    C_Normal3f(0.997787654f, 0.0664810762f, 0.f),
    C_Vertex3f(0.550292969f, 0.193943024f, 0.f),
    C_Vertex3f(0.550292969f, 0.193943024f, -1.f),
    C_Normal3f(0.997380614f, -0.072332114f, 0.f),
    C_Vertex3f(0.550292969f, 0.168273926f, 0.f),
    C_Vertex3f(0.550292969f, 0.168273926f, -1.f),
    C_Normal3f(0.975688696f, -0.219161078f, 0.f),
    C_Vertex3f(0.546631336f, 0.143453121f, 0.f),
    C_Vertex3f(0.546631336f, 0.143453121f, -1.f),
    C_Normal3f(0.932907403f, -0.360116422f, 0.f),
    C_Vertex3f(0.539308071f, 0.119369984f, 0.f),
    C_Vertex3f(0.539308071f, 0.119369984f, -1.f),
    C_Normal3f(0.8780123f, -0.478638142f, 0.f),
    C_Vertex3f(0.528323174f, 0.0960245132f, 0.f),
    C_Vertex3f(0.528323174f, 0.0960245132f, -1.f),
    C_Normal3f(0.822501123f, -0.568763554f, 0.f),
    C_Vertex3f(0.517338276f, 0.0790686607f, 0.f),
    C_Vertex3f(0.517338276f, 0.0790686607f, -1.f),
    C_Normal3f(0.772335768f, -0.635214388f, 0.f),
    C_Vertex3f(0.509215593f, 0.0683926344f, 0.f),
    C_Vertex3f(0.509215593f, 0.0683926344f, -1.f),
    C_Normal3f(0.722240627f, -0.691641867f, 0.f),
    C_Vertex3f(0.500293493f, 0.0583444834f, 0.f),
    C_Vertex3f(0.500293493f, 0.0583444834f, -1.f),
    C_Normal3f(0.668852508f, -0.74339515f, 0.f),
    C_Vertex3f(0.490571976f, 0.0489242077f, 0.f),
    C_Vertex3f(0.490571976f, 0.0489242077f, -1.f),
    C_Normal3f(0.613247633f, -0.789890707f, 0.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, 0.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, -1.f),
    C_Normal3f(0.556525171f, -0.830830753f, 0.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, 0.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, -1.f),
    C_Normal3f(0.499719143f, -0.866187513f, 0.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, 0.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, -1.f),
    C_Normal3f(0.43820852f, -0.898873389f, 0.f),
    C_Vertex3f(0.44369173f, 0.0175218582f, 0.f),
    C_Vertex3f(0.44369173f, 0.0175218582f, -1.f),
    C_Normal3f(0.354275286f, -0.935141146f, 0.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, 0.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, -1.f),
    C_Normal3f(0.253019869f, -0.96746105f, 0.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, 0.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, -1.f),
    C_Normal3f(0.146797985f, -0.989166439f, 0.f),
    C_Vertex3f(0.363073826f, -0.00760173798f, 0.f),
    C_Vertex3f(0.363073826f, -0.00760173798f, -1.f),
    C_Normal3f(0.0517184138f, -0.998661697f, 0.f),
    C_Vertex3f(0.330471039f, -0.0107421875f, 0.f),
    C_Vertex3f(0.330471039f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0161968842f, -0.99986881f, 0.f),
    C_Vertex3f(0.302433014f, -0.0107421875f, 0.f),
    C_Vertex3f(0.302433014f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0580721684f, -0.998312414f, 0.f),
    C_Vertex3f(0.281562805f, -0.00994992256f, 0.f),
    C_Vertex3f(0.281562805f, -0.00994992256f, -1.f),
    C_Normal3f(-0.100804754f, -0.994906187f, 0.f),
    C_Vertex3f(0.261573792f, -0.00836539268f, 0.f),
    C_Vertex3f(0.261573792f, -0.00836539268f, -1.f),
    C_Normal3f(-0.146933034f, -0.989146531f, 0.f),
    C_Vertex3f(0.242465973f, -0.00598859787f, 0.f),
    C_Vertex3f(0.242465973f, -0.00598859787f, -1.f),
    C_Normal3f(-0.196539491f, -0.98049593f, 0.f),
    C_Vertex3f(0.224239349f, -0.00281953812f, 0.f),
    C_Vertex3f(0.224239349f, -0.00281953812f, -1.f),
    C_Normal3f(-0.249605194f, -0.968347728f, 0.f),
    C_Vertex3f(0.206893921f, 0.00114178658f, 0.f),
    C_Vertex3f(0.206893921f, 0.00114178658f, -1.f),
    C_Normal3f(-0.305969775f, -0.952041268f, 0.f),
    C_Vertex3f(0.190429688f, 0.00589537621f, 0.f),
    C_Vertex3f(0.190429688f, 0.00589537621f, -1.f),
    C_Normal3f(-0.365288645f, -0.930894315f, 0.f),
    C_Vertex3f(0.174846649f, 0.0114412308f, 0.f),
    C_Vertex3f(0.174846649f, 0.0114412308f, -1.f),
    C_Normal3f(-0.42608273f, -0.904684186f, 0.f),
    C_Vertex3f(0.160144806f, 0.0177793503f, 0.f),
    C_Vertex3f(0.160144806f, 0.0177793503f, -1.f),
    C_Normal3f(-0.487082392f, -0.873356044f, 0.f),
    C_Vertex3f(0.146235347f, 0.0249164104f, 0.f),
    C_Vertex3f(0.146235347f, 0.0249164104f, -1.f),
    C_Normal3f(-0.548076808f, -0.836427927f, 0.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, 0.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, -1.f),
    C_Normal3f(-0.607979953f, -0.793952405f, 0.f),
    C_Vertex3f(0.120793581f, 0.0415873528f, 0.f),
    C_Vertex3f(0.120793581f, 0.0415873528f, -1.f),
    C_Normal3f(-0.665657461f, -0.746257424f, 0.f),
    C_Vertex3f(0.109261274f, 0.0511212349f, 0.f),
    C_Vertex3f(0.109261274f, 0.0511212349f, -1.f),
    C_Normal3f(-0.720020831f, -0.693952501f, 0.f),
    C_Vertex3f(0.0985213518f, 0.0614540577f, 0.f),
    C_Vertex3f(0.0985213518f, 0.0614540577f, -1.f),
    C_Normal3f(-0.770123422f, -0.637894928f, 0.f),
    C_Vertex3f(0.0885738134f, 0.0725858212f, 0.f),
    C_Vertex3f(0.0885738134f, 0.0725858212f, -1.f),
    C_Normal3f(-0.815241337f, -0.579121411f, 0.f),
    C_Vertex3f(0.0794186592f, 0.0845165253f, 0.f),
    C_Vertex3f(0.0794186592f, 0.0845165253f, -1.f),
    C_Normal3f(-0.854430854f, -0.519565165f, 0.f),
    C_Vertex3f(0.0710558891f, 0.09724617f, 0.f),
    C_Vertex3f(0.0710558891f, 0.09724617f, -1.f),
    C_Normal3f(-0.889058173f, -0.45779416f, 0.f),
    C_Vertex3f(0.0636838675f, 0.11039257f, 0.f),
    C_Vertex3f(0.0636838675f, 0.11039257f, -1.f),
    C_Normal3f(-0.919535875f, -0.393005997f, 0.f),
    C_Vertex3f(0.0573025942f, 0.123955727f, 0.f),
    C_Vertex3f(0.0573025942f, 0.123955727f, -1.f),
    C_Normal3f(-0.945318818f, -0.326147705f, 0.f),
    C_Vertex3f(0.0519120693f, 0.137935638f, 0.f),
    C_Vertex3f(0.0519120693f, 0.137935638f, -1.f),
    C_Normal3f(-0.966074646f, -0.258263081f, 0.f),
    C_Vertex3f(0.0475122929f, 0.152332306f, 0.f),
    C_Vertex3f(0.0475122929f, 0.152332306f, -1.f),
    C_Normal3f(-0.981704116f, -0.190412983f, 0.f),
    C_Vertex3f(0.0441032648f, 0.167145729f, 0.f),
    C_Vertex3f(0.0441032648f, 0.167145729f, -1.f),
    C_Normal3f(-0.992332458f, -0.123597145f, 0.f),
    C_Vertex3f(0.0416849852f, 0.182375908f, 0.f),
    C_Vertex3f(0.0416849852f, 0.182375908f, -1.f),
    C_Normal3f(-0.997592866f, -0.0693431348f, 0.f),
    C_Vertex3f(0.0402574539f, 0.198022842f, 0.f),
    C_Vertex3f(0.0402574539f, 0.198022842f, -1.f),
    C_Normal3f(-0.999630511f, -0.0271806438f, 0.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, 0.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.840821147f, 0.541312993f, 0.f),
    C_Vertex3f(0.297851563f, 0.728027344f, 0.f),
    C_Vertex3f(0.297851563f, 0.728027344f, -1.f),
    C_Vertex3f(0.345947266f, 0.802734375f, -1.f),
    C_Vertex3f(0.345947266f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.345947266f, 0.802734375f, 0.f),
    C_Vertex3f(0.345947266f, 0.802734375f, -1.f),
    C_Vertex3f(0.435302734f, 0.802734375f, -1.f),
    C_Vertex3f(0.435302734f, 0.802734375f, 0.f),
    C_Normal3f(0.779806852f, -0.626020253f, 0.f),
    C_Vertex3f(0.435302734f, 0.802734375f, 0.f),
    C_Vertex3f(0.435302734f, 0.802734375f, -1.f),
    C_Vertex3f(0.336914063f, 0.680175781f, -1.f),
    C_Vertex3f(0.336914063f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.336914063f, 0.680175781f, 0.f),
    C_Vertex3f(0.336914063f, 0.680175781f, -1.f),
    C_Vertex3f(0.252929688f, 0.680175781f, -1.f),
    C_Vertex3f(0.252929688f, 0.680175781f, 0.f),
    C_Normal3f(-0.795031667f, -0.606567979f, 0.f),
    C_Vertex3f(0.252929688f, 0.680175781f, 0.f),
    C_Vertex3f(0.252929688f, 0.680175781f, -1.f),
    C_Vertex3f(0.159423828f, 0.802734375f, -1.f),
    C_Vertex3f(0.159423828f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.159423828f, 0.802734375f, 0.f),
    C_Vertex3f(0.159423828f, 0.802734375f, -1.f),
    C_Vertex3f(0.248779297f, 0.802734375f, -1.f),
    C_Vertex3f(0.248779297f, 0.802734375f, 0.f),
    C_Normal3f(0.835812926f, 0.54901439f, 0.f),
    C_Vertex3f(0.248779297f, 0.802734375f, 0.f),
    C_Vertex3f(0.248779297f, 0.802734375f, -1.f),
    C_Vertex3f(0.297851563f, 0.728027344f, -1.f),
    C_Vertex3f(0.297851563f, 0.728027344f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.53125f, 0.460449219f, 0.f),
    C_Vertex3f(0.530702591f, 0.474542618f, 0.f),
    C_Vertex3f(0.526213646f, 0.501254559f, 0.f),
    C_Vertex3f(0.518330574f, 0.526491642f, 0.f),
    C_Vertex3f(0.507053375f, 0.550253868f, 0.f),
    C_Vertex3f(0.496049881f, 0.566969395f, 0.f),
    C_Vertex3f(0.487927318f, 0.577437043f, 0.f),
    C_Vertex3f(0.479017854f, 0.587228656f, 0.f),
    C_Vertex3f(0.469321489f, 0.596344233f, 0.f),
    C_Vertex3f(0.458838224f, 0.604783773f, 0.f),
    C_Vertex3f(0.450195313f, 0.454101563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.443137288f, 0.489351988f, 0.f),
    C_Vertex3f(0.446745276f, 0.475996256f, 0.f),
    C_Vertex3f(0.447568059f, 0.612547278f, 0.f),
    C_Vertex3f(0.449372292f, 0.461709976f, 0.f),
    C_Vertex3f(0.458838224f, 0.604783773f, 0.f),
    C_Vertex3f(0.450195313f, 0.454101563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426427484f, 0.523835897f, 0.f),
    C_Vertex3f(0.432978392f, 0.513271809f, 0.f),
    C_Vertex3f(0.435510993f, 0.619634748f, 0.f),
    C_Vertex3f(0.438548326f, 0.501777172f, 0.f),
    C_Vertex3f(0.447568059f, 0.612547278f, 0.f),
    C_Vertex3f(0.443137288f, 0.489351988f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.422667027f, 0.626046181f, 0.f),
    C_Vertex3f(0.402220726f, 0.634649277f, 0.f),
    C_Vertex3f(0.410382748f, 0.542172432f, 0.f),
    C_Vertex3f(0.418895602f, 0.533469439f, 0.f),
    C_Vertex3f(0.426427484f, 0.523835897f, 0.f),
    C_Vertex3f(0.435510993f, 0.619634748f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402220726f, 0.634649277f, 0.f),
    C_Vertex3f(0.372850895f, 0.643252373f, 0.f),
    C_Vertex3f(0.378505588f, 0.562575102f, 0.f),
    C_Vertex3f(0.3902632f, 0.556745768f, 0.f),
    C_Vertex3f(0.40088892f, 0.549944878f, 0.f),
    C_Vertex3f(0.410382748f, 0.542172432f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320156217f, 0.576176882f, 0.f),
    C_Vertex3f(0.336441398f, 0.57423377f, 0.f),
    C_Vertex3f(0.341372967f, 0.64898777f, 0.f),
    C_Vertex3f(0.351594687f, 0.571319103f, 0.f),
    C_Vertex3f(0.372850895f, 0.643252373f, 0.f),
    C_Vertex3f(0.365616083f, 0.56743288f, 0.f),
    C_Vertex3f(0.378505588f, 0.562575102f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.307786942f, 0.651855469f, 0.f),
    C_Vertex3f(0.273736954f, 0.651855469f, 0.f),
    C_Vertex3f(0.283805847f, 0.577148438f, 0.f),
    C_Vertex3f(0.302739143f, 0.577148438f, 0.f),
    C_Vertex3f(0.320156217f, 0.576176882f, 0.f),
    C_Vertex3f(0.341372967f, 0.64898777f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.219637632f, 0.568331718f, 0.f),
    C_Vertex3f(0.233714342f, 0.571858406f, 0.f),
    C_Vertex3f(0.242808819f, 0.649122238f, 0.f),
    C_Vertex3f(0.249101281f, 0.574503422f, 0.f),
    C_Vertex3f(0.273736954f, 0.651855469f, 0.f),
    C_Vertex3f(0.26579845f, 0.576266766f, 0.f),
    C_Vertex3f(0.283805847f, 0.577148438f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213358402f, 0.643655777f, 0.f),
    C_Vertex3f(0.185385704f, 0.635456085f, 0.f),
    C_Vertex3f(0.195414901f, 0.558633327f, 0.f),
    C_Vertex3f(0.206871152f, 0.563923359f, 0.f),
    C_Vertex3f(0.219637632f, 0.568331718f, 0.f),
    C_Vertex3f(0.242808819f, 0.649122238f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.185385704f, 0.635456085f, 0.f),
    C_Vertex3f(0.165514469f, 0.627256393f, 0.f),
    C_Vertex3f(0.168701768f, 0.537992716f, 0.f),
    C_Vertex3f(0.176433086f, 0.545408249f, 0.f),
    C_Vertex3f(0.185268879f, 0.552461624f, 0.f),
    C_Vertex3f(0.195414901f, 0.558633327f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.148821235f, 0.504709005f, 0.f),
    C_Vertex3f(0.152134657f, 0.51357317f, 0.f),
    C_Vertex3f(0.153060675f, 0.621154547f, 0.f),
    C_Vertex3f(0.156552553f, 0.522075176f, 0.f),
    C_Vertex3f(0.165514469f, 0.627256393f, 0.f),
    C_Vertex3f(0.162074924f, 0.530215025f, 0.f),
    C_Vertex3f(0.168701768f, 0.537992716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.120461464f, 0.599036694f, 0.f),
    C_Vertex3f(0.123383999f, 0.355178356f, 0.f),
    C_Vertex3f(0.130534172f, 0.607044697f, 0.f),
    C_Vertex3f(0.143409729f, 0.341705322f, 0.f),
    C_Vertex3f(0.141400576f, 0.614417315f, 0.f),
    C_Vertex3f(0.145507813f, 0.485894203f, 0.f),
    C_Vertex3f(0.153060675f, 0.621154547f, 0.f),
    C_Vertex3f(0.146612287f, 0.495482683f, 0.f),
    C_Vertex3f(0.148821235f, 0.504709005f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154891968f, 0.442318916f, 0.f),
    C_Vertex3f(0.148635864f, 0.456534386f, 0.f),
    C_Vertex3f(0.159310341f, 0.333342552f, 0.f),
    C_Vertex3f(0.145507813f, 0.472280502f, 0.f),
    C_Vertex3f(0.143409729f, 0.341705322f, 0.f),
    C_Vertex3f(0.145507813f, 0.485894203f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.111182451f, 0.590393305f, 0.f),
    C_Vertex3f(0.102697134f, 0.581114531f, 0.f),
    C_Vertex3f(0.106281757f, 0.370100498f, 0.f),
    C_Vertex3f(0.123383999f, 0.355178356f, 0.f),
    C_Vertex3f(0.120461464f, 0.599036694f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0639648438f, 0.487323761f, 0.f),
    C_Vertex3f(0.0667786598f, 0.443130016f, 0.f),
    C_Vertex3f(0.0674138069f, 0.510800362f, 0.f),
    C_Vertex3f(0.072406292f, 0.423178196f, 0.f),
    C_Vertex3f(0.0743117332f, 0.533484459f, 0.f),
    C_Vertex3f(0.0808477402f, 0.404292107f, 0.f),
    C_Vertex3f(0.0846586227f, 0.555376053f, 0.f),
    C_Vertex3f(0.0921030045f, 0.386471748f, 0.f),
    C_Vertex3f(0.0950055122f, 0.571200371f, 0.f),
    C_Vertex3f(0.106281757f, 0.370100498f, 0.f),
    C_Vertex3f(0.102697134f, 0.581114531f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466575146f, 0.195565224f, 0.f),
    C_Vertex3f(0.46875f, 0.181438446f, 0.f),
    C_Vertex3f(0.473357797f, 0.320065141f, 0.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, 0.f),
    C_Vertex3f(0.484582186f, 0.312331438f, 0.f),
    C_Vertex3f(0.490571976f, 0.0489242077f, 0.f),
    C_Vertex3f(0.494892836f, 0.304167509f, 0.f),
    C_Vertex3f(0.500293493f, 0.0583444834f, 0.f),
    C_Vertex3f(0.504289746f, 0.295573354f, 0.f),
    C_Vertex3f(0.509215593f, 0.0683926344f, 0.f),
    C_Vertex3f(0.512772918f, 0.286548972f, 0.f),
    C_Vertex3f(0.517338276f, 0.0790686607f, 0.f),
    C_Vertex3f(0.52034235f, 0.277094364f, 0.f),
    C_Vertex3f(0.528323174f, 0.0960245132f, 0.f),
    C_Vertex3f(0.53032589f, 0.262267113f, 0.f),
    C_Vertex3f(0.539308071f, 0.119369984f, 0.f),
    C_Vertex3f(0.540309429f, 0.24099493f, 0.f),
    C_Vertex3f(0.546631336f, 0.143453121f, 0.f),
    C_Vertex3f(0.546965122f, 0.218220234f, 0.f),
    C_Vertex3f(0.550292969f, 0.168273926f, 0.f),
    C_Vertex3f(0.550292969f, 0.193943024f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.381887197f, 0.268518806f, 0.f),
    C_Vertex3f(0.390324593f, 0.265654564f, 0.f),
    C_Vertex3f(0.395312786f, 0.353868246f, 0.f),
    C_Vertex3f(0.403659821f, 0.260570526f, 0.f),
    C_Vertex3f(0.411047697f, 0.348943114f, 0.f),
    C_Vertex3f(0.42098093f, 0.252017021f, 0.f),
    C_Vertex3f(0.424738169f, 0.344203115f, 0.f),
    C_Vertex3f(0.435428143f, 0.242562294f, 0.f),
    C_Vertex3f(0.436384201f, 0.339648247f, 0.f),
    C_Vertex3f(0.447001457f, 0.232206345f, 0.f),
    C_Vertex3f(0.448167801f, 0.334241867f, 0.f),
    C_Vertex3f(0.455700874f, 0.220949173f, 0.f),
    C_Vertex3f(0.461219668f, 0.327368617f, 0.f),
    C_Vertex3f(0.462225437f, 0.208735466f, 0.f),
    C_Vertex3f(0.473357797f, 0.320065141f, 0.f),
    C_Vertex3f(0.466575146f, 0.195565224f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377533436f, 0.35897851f, 0.f),
    C_Vertex3f(0.357709646f, 0.364273906f, 0.f),
    C_Vertex3f(0.359126568f, 0.275292635f, 0.f),
    C_Vertex3f(0.371487856f, 0.271731496f, 0.f),
    C_Vertex3f(0.381887197f, 0.268518806f, 0.f),
    C_Vertex3f(0.395312786f, 0.353868246f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.290061951f, 0.293021679f, 0.f),
    C_Vertex3f(0.310271025f, 0.288066745f, 0.f),
    C_Vertex3f(0.311928749f, 0.375420094f, 0.f),
    C_Vertex3f(0.328518152f, 0.283460259f, 0.f),
    C_Vertex3f(0.335841417f, 0.369754434f, 0.f),
    C_Vertex3f(0.344803333f, 0.279202223f, 0.f),
    C_Vertex3f(0.357709646f, 0.364273906f, 0.f),
    C_Vertex3f(0.359126568f, 0.275292635f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193329334f, 0.410011649f, 0.f),
    C_Vertex3f(0.195855618f, 0.318786383f, 0.f),
    C_Vertex3f(0.206905365f, 0.404331923f, 0.f),
    C_Vertex3f(0.21136713f, 0.313783884f, 0.f),
    C_Vertex3f(0.222957611f, 0.398617983f, 0.f),
    C_Vertex3f(0.22854352f, 0.308706164f, 0.f),
    C_Vertex3f(0.241486073f, 0.39286983f, 0.f),
    C_Vertex3f(0.247384787f, 0.303553224f, 0.f),
    C_Vertex3f(0.262490749f, 0.387087464f, 0.f),
    C_Vertex3f(0.26789093f, 0.298325062f, 0.f),
    C_Vertex3f(0.285971642f, 0.381270885f, 0.f),
    C_Vertex3f(0.290061951f, 0.293021679f, 0.f),
    C_Vertex3f(0.311928749f, 0.375420094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154891968f, 0.442318916f, 0.f),
    C_Vertex3f(0.159310341f, 0.333342552f, 0.f),
    C_Vertex3f(0.164276123f, 0.429634094f, 0.f),
    C_Vertex3f(0.169827223f, 0.328565717f, 0.f),
    C_Vertex3f(0.173605919f, 0.421268463f, 0.f),
    C_Vertex3f(0.182008982f, 0.32371366f, 0.f),
    C_Vertex3f(0.182229519f, 0.415657163f, 0.f),
    C_Vertex3f(0.195855618f, 0.318786383f, 0.f),
    C_Vertex3f(0.193329334f, 0.410011649f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.46875f, 0.166463852f, 0.f),
    C_Vertex3f(0.466480732f, 0.152168751f, 0.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, 0.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, 0.f),
    C_Vertex3f(0.46875f, 0.181438446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.461942196f, 0.138444424f, 0.f),
    C_Vertex3f(0.455134392f, 0.125290871f, 0.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, 0.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, 0.f),
    C_Vertex3f(0.466480732f, 0.152168751f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.420767784f, 0.0915622711f, 0.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, 0.f),
    C_Vertex3f(0.434601784f, 0.101465225f, 0.f),
    C_Vertex3f(0.44369173f, 0.0175218582f, 0.f),
    C_Vertex3f(0.44605732f, 0.112708092f, 0.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, 0.f),
    C_Vertex3f(0.455134392f, 0.125290871f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.404555321f, 0.0829992294f, 0.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, 0.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, 0.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, 0.f),
    C_Vertex3f(0.420767784f, 0.0915622711f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.224239349f, -0.00281953812f, 0.f),
    C_Vertex3f(0.242465973f, -0.00598859787f, 0.f),
    C_Vertex3f(0.245275974f, 0.072081089f, 0.f),
    C_Vertex3f(0.261573792f, -0.00836539268f, 0.f),
    C_Vertex3f(0.269881725f, 0.0673213005f, 0.f),
    C_Vertex3f(0.281562805f, -0.00994992256f, 0.f),
    C_Vertex3f(0.295854568f, 0.0649414063f, 0.f),
    C_Vertex3f(0.302433014f, -0.0107421875f, 0.f),
    C_Vertex3f(0.321662903f, 0.0649414063f, 0.f),
    C_Vertex3f(0.330471039f, -0.0107421875f, 0.f),
    C_Vertex3f(0.344518185f, 0.0667471886f, 0.f),
    C_Vertex3f(0.363073826f, -0.00760173798f, 0.f),
    C_Vertex3f(0.365952015f, 0.0703587532f, 0.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, 0.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132882595f, 0.159187317f, 0.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, 0.f),
    C_Vertex3f(0.141385078f, 0.14159584f, 0.f),
    C_Vertex3f(0.146235347f, 0.0249164104f, 0.f),
    C_Vertex3f(0.151774406f, 0.12616539f, 0.f),
    C_Vertex3f(0.160144806f, 0.0177793503f, 0.f),
    C_Vertex3f(0.165034294f, 0.112212658f, 0.f),
    C_Vertex3f(0.174846649f, 0.0114412308f, 0.f),
    C_Vertex3f(0.181164742f, 0.0997376442f, 0.f),
    C_Vertex3f(0.190429688f, 0.00589537621f, 0.f),
    C_Vertex3f(0.200165749f, 0.0887403488f, 0.f),
    C_Vertex3f(0.206893921f, 0.00114178658f, 0.f),
    C_Vertex3f(0.222037315f, 0.0792207718f, 0.f),
    C_Vertex3f(0.224239349f, -0.00281953812f, 0.f),
    C_Vertex3f(0.245275974f, 0.072081089f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0985213518f, 0.0614540577f, 0.f),
    C_Vertex3f(0.109261274f, 0.0511212349f, 0.f),
    C_Vertex3f(0.120117188f, 0.212890625f, 0.f),
    C_Vertex3f(0.120793581f, 0.0415873528f, 0.f),
    C_Vertex3f(0.121538162f, 0.200853348f, 0.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, 0.f),
    C_Vertex3f(0.126266956f, 0.178939819f, 0.f),
    C_Vertex3f(0.132882595f, 0.159187317f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120117188f, 0.212890625f, 0.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, 0.f),
    C_Vertex3f(0.0402574539f, 0.198022842f, 0.f),
    C_Vertex3f(0.0416849852f, 0.182375908f, 0.f),
    C_Vertex3f(0.0441032648f, 0.167145729f, 0.f),
    C_Vertex3f(0.0475122929f, 0.152332306f, 0.f),
    C_Vertex3f(0.0519120693f, 0.137935638f, 0.f),
    C_Vertex3f(0.0573025942f, 0.123955727f, 0.f),
    C_Vertex3f(0.0636838675f, 0.11039257f, 0.f),
    C_Vertex3f(0.0710558891f, 0.09724617f, 0.f),
    C_Vertex3f(0.0794186592f, 0.0845165253f, 0.f),
    C_Vertex3f(0.0885738134f, 0.0725858212f, 0.f),
    C_Vertex3f(0.0985213518f, 0.0614540577f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.435302734f, 0.802734375f, 0.f),
    C_Vertex3f(0.345947266f, 0.802734375f, 0.f),
    C_Vertex3f(0.336914063f, 0.680175781f, 0.f),
    C_Vertex3f(0.297851563f, 0.728027344f, 0.f),
    C_Vertex3f(0.252929688f, 0.680175781f, 0.f),
    C_Vertex3f(0.248779297f, 0.802734375f, 0.f),
    C_Vertex3f(0.159423828f, 0.802734375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0667786598f, 0.443130016f, 0.f),
    C_Vertex3f(0.0639648438f, 0.487323761f, 0.f),
    C_Vertex3f(0.0639648438f, 0.464147568f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.53125f, 0.460449219f, -1.f),
    C_Vertex3f(0.530702591f, 0.474542618f, -1.f),
    C_Vertex3f(0.526213646f, 0.501254559f, -1.f),
    C_Vertex3f(0.518330574f, 0.526491642f, -1.f),
    C_Vertex3f(0.507053375f, 0.550253868f, -1.f),
    C_Vertex3f(0.496049881f, 0.566969395f, -1.f),
    C_Vertex3f(0.487927318f, 0.577437043f, -1.f),
    C_Vertex3f(0.479017854f, 0.587228656f, -1.f),
    C_Vertex3f(0.469321489f, 0.596344233f, -1.f),
    C_Vertex3f(0.458838224f, 0.604783773f, -1.f),
    C_Vertex3f(0.450195313f, 0.454101563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.443137288f, 0.489351988f, -1.f),
    C_Vertex3f(0.446745276f, 0.475996256f, -1.f),
    C_Vertex3f(0.447568059f, 0.612547278f, -1.f),
    C_Vertex3f(0.449372292f, 0.461709976f, -1.f),
    C_Vertex3f(0.458838224f, 0.604783773f, -1.f),
    C_Vertex3f(0.450195313f, 0.454101563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426427484f, 0.523835897f, -1.f),
    C_Vertex3f(0.432978392f, 0.513271809f, -1.f),
    C_Vertex3f(0.435510993f, 0.619634748f, -1.f),
    C_Vertex3f(0.438548326f, 0.501777172f, -1.f),
    C_Vertex3f(0.447568059f, 0.612547278f, -1.f),
    C_Vertex3f(0.443137288f, 0.489351988f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.422667027f, 0.626046181f, -1.f),
    C_Vertex3f(0.402220726f, 0.634649277f, -1.f),
    C_Vertex3f(0.410382748f, 0.542172432f, -1.f),
    C_Vertex3f(0.418895602f, 0.533469439f, -1.f),
    C_Vertex3f(0.426427484f, 0.523835897f, -1.f),
    C_Vertex3f(0.435510993f, 0.619634748f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.402220726f, 0.634649277f, -1.f),
    C_Vertex3f(0.372850895f, 0.643252373f, -1.f),
    C_Vertex3f(0.378505588f, 0.562575102f, -1.f),
    C_Vertex3f(0.3902632f, 0.556745768f, -1.f),
    C_Vertex3f(0.40088892f, 0.549944878f, -1.f),
    C_Vertex3f(0.410382748f, 0.542172432f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.320156217f, 0.576176882f, -1.f),
    C_Vertex3f(0.336441398f, 0.57423377f, -1.f),
    C_Vertex3f(0.341372967f, 0.64898777f, -1.f),
    C_Vertex3f(0.351594687f, 0.571319103f, -1.f),
    C_Vertex3f(0.372850895f, 0.643252373f, -1.f),
    C_Vertex3f(0.365616083f, 0.56743288f, -1.f),
    C_Vertex3f(0.378505588f, 0.562575102f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.307786942f, 0.651855469f, -1.f),
    C_Vertex3f(0.273736954f, 0.651855469f, -1.f),
    C_Vertex3f(0.283805847f, 0.577148438f, -1.f),
    C_Vertex3f(0.302739143f, 0.577148438f, -1.f),
    C_Vertex3f(0.320156217f, 0.576176882f, -1.f),
    C_Vertex3f(0.341372967f, 0.64898777f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.219637632f, 0.568331718f, -1.f),
    C_Vertex3f(0.233714342f, 0.571858406f, -1.f),
    C_Vertex3f(0.242808819f, 0.649122238f, -1.f),
    C_Vertex3f(0.249101281f, 0.574503422f, -1.f),
    C_Vertex3f(0.273736954f, 0.651855469f, -1.f),
    C_Vertex3f(0.26579845f, 0.576266766f, -1.f),
    C_Vertex3f(0.283805847f, 0.577148438f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213358402f, 0.643655777f, -1.f),
    C_Vertex3f(0.185385704f, 0.635456085f, -1.f),
    C_Vertex3f(0.195414901f, 0.558633327f, -1.f),
    C_Vertex3f(0.206871152f, 0.563923359f, -1.f),
    C_Vertex3f(0.219637632f, 0.568331718f, -1.f),
    C_Vertex3f(0.242808819f, 0.649122238f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.185385704f, 0.635456085f, -1.f),
    C_Vertex3f(0.165514469f, 0.627256393f, -1.f),
    C_Vertex3f(0.168701768f, 0.537992716f, -1.f),
    C_Vertex3f(0.176433086f, 0.545408249f, -1.f),
    C_Vertex3f(0.185268879f, 0.552461624f, -1.f),
    C_Vertex3f(0.195414901f, 0.558633327f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.148821235f, 0.504709005f, -1.f),
    C_Vertex3f(0.152134657f, 0.51357317f, -1.f),
    C_Vertex3f(0.153060675f, 0.621154547f, -1.f),
    C_Vertex3f(0.156552553f, 0.522075176f, -1.f),
    C_Vertex3f(0.165514469f, 0.627256393f, -1.f),
    C_Vertex3f(0.162074924f, 0.530215025f, -1.f),
    C_Vertex3f(0.168701768f, 0.537992716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.120461464f, 0.599036694f, -1.f),
    C_Vertex3f(0.123383999f, 0.355178356f, -1.f),
    C_Vertex3f(0.130534172f, 0.607044697f, -1.f),
    C_Vertex3f(0.143409729f, 0.341705322f, -1.f),
    C_Vertex3f(0.141400576f, 0.614417315f, -1.f),
    C_Vertex3f(0.145507813f, 0.485894203f, -1.f),
    C_Vertex3f(0.153060675f, 0.621154547f, -1.f),
    C_Vertex3f(0.146612287f, 0.495482683f, -1.f),
    C_Vertex3f(0.148821235f, 0.504709005f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154891968f, 0.442318916f, -1.f),
    C_Vertex3f(0.148635864f, 0.456534386f, -1.f),
    C_Vertex3f(0.159310341f, 0.333342552f, -1.f),
    C_Vertex3f(0.145507813f, 0.472280502f, -1.f),
    C_Vertex3f(0.143409729f, 0.341705322f, -1.f),
    C_Vertex3f(0.145507813f, 0.485894203f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.111182451f, 0.590393305f, -1.f),
    C_Vertex3f(0.102697134f, 0.581114531f, -1.f),
    C_Vertex3f(0.106281757f, 0.370100498f, -1.f),
    C_Vertex3f(0.123383999f, 0.355178356f, -1.f),
    C_Vertex3f(0.120461464f, 0.599036694f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0639648438f, 0.487323761f, -1.f),
    C_Vertex3f(0.0667786598f, 0.443130016f, -1.f),
    C_Vertex3f(0.0674138069f, 0.510800362f, -1.f),
    C_Vertex3f(0.072406292f, 0.423178196f, -1.f),
    C_Vertex3f(0.0743117332f, 0.533484459f, -1.f),
    C_Vertex3f(0.0808477402f, 0.404292107f, -1.f),
    C_Vertex3f(0.0846586227f, 0.555376053f, -1.f),
    C_Vertex3f(0.0921030045f, 0.386471748f, -1.f),
    C_Vertex3f(0.0950055122f, 0.571200371f, -1.f),
    C_Vertex3f(0.106281757f, 0.370100498f, -1.f),
    C_Vertex3f(0.102697134f, 0.581114531f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466575146f, 0.195565224f, -1.f),
    C_Vertex3f(0.46875f, 0.181438446f, -1.f),
    C_Vertex3f(0.473357797f, 0.320065141f, -1.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, -1.f),
    C_Vertex3f(0.484582186f, 0.312331438f, -1.f),
    C_Vertex3f(0.490571976f, 0.0489242077f, -1.f),
    C_Vertex3f(0.494892836f, 0.304167509f, -1.f),
    C_Vertex3f(0.500293493f, 0.0583444834f, -1.f),
    C_Vertex3f(0.504289746f, 0.295573354f, -1.f),
    C_Vertex3f(0.509215593f, 0.0683926344f, -1.f),
    C_Vertex3f(0.512772918f, 0.286548972f, -1.f),
    C_Vertex3f(0.517338276f, 0.0790686607f, -1.f),
    C_Vertex3f(0.52034235f, 0.277094364f, -1.f),
    C_Vertex3f(0.528323174f, 0.0960245132f, -1.f),
    C_Vertex3f(0.53032589f, 0.262267113f, -1.f),
    C_Vertex3f(0.539308071f, 0.119369984f, -1.f),
    C_Vertex3f(0.540309429f, 0.24099493f, -1.f),
    C_Vertex3f(0.546631336f, 0.143453121f, -1.f),
    C_Vertex3f(0.546965122f, 0.218220234f, -1.f),
    C_Vertex3f(0.550292969f, 0.168273926f, -1.f),
    C_Vertex3f(0.550292969f, 0.193943024f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.381887197f, 0.268518806f, -1.f),
    C_Vertex3f(0.390324593f, 0.265654564f, -1.f),
    C_Vertex3f(0.395312786f, 0.353868246f, -1.f),
    C_Vertex3f(0.403659821f, 0.260570526f, -1.f),
    C_Vertex3f(0.411047697f, 0.348943114f, -1.f),
    C_Vertex3f(0.42098093f, 0.252017021f, -1.f),
    C_Vertex3f(0.424738169f, 0.344203115f, -1.f),
    C_Vertex3f(0.435428143f, 0.242562294f, -1.f),
    C_Vertex3f(0.436384201f, 0.339648247f, -1.f),
    C_Vertex3f(0.447001457f, 0.232206345f, -1.f),
    C_Vertex3f(0.448167801f, 0.334241867f, -1.f),
    C_Vertex3f(0.455700874f, 0.220949173f, -1.f),
    C_Vertex3f(0.461219668f, 0.327368617f, -1.f),
    C_Vertex3f(0.462225437f, 0.208735466f, -1.f),
    C_Vertex3f(0.473357797f, 0.320065141f, -1.f),
    C_Vertex3f(0.466575146f, 0.195565224f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377533436f, 0.35897851f, -1.f),
    C_Vertex3f(0.357709646f, 0.364273906f, -1.f),
    C_Vertex3f(0.359126568f, 0.275292635f, -1.f),
    C_Vertex3f(0.371487856f, 0.271731496f, -1.f),
    C_Vertex3f(0.381887197f, 0.268518806f, -1.f),
    C_Vertex3f(0.395312786f, 0.353868246f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.290061951f, 0.293021679f, -1.f),
    C_Vertex3f(0.310271025f, 0.288066745f, -1.f),
    C_Vertex3f(0.311928749f, 0.375420094f, -1.f),
    C_Vertex3f(0.328518152f, 0.283460259f, -1.f),
    C_Vertex3f(0.335841417f, 0.369754434f, -1.f),
    C_Vertex3f(0.344803333f, 0.279202223f, -1.f),
    C_Vertex3f(0.357709646f, 0.364273906f, -1.f),
    C_Vertex3f(0.359126568f, 0.275292635f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193329334f, 0.410011649f, -1.f),
    C_Vertex3f(0.195855618f, 0.318786383f, -1.f),
    C_Vertex3f(0.206905365f, 0.404331923f, -1.f),
    C_Vertex3f(0.21136713f, 0.313783884f, -1.f),
    C_Vertex3f(0.222957611f, 0.398617983f, -1.f),
    C_Vertex3f(0.22854352f, 0.308706164f, -1.f),
    C_Vertex3f(0.241486073f, 0.39286983f, -1.f),
    C_Vertex3f(0.247384787f, 0.303553224f, -1.f),
    C_Vertex3f(0.262490749f, 0.387087464f, -1.f),
    C_Vertex3f(0.26789093f, 0.298325062f, -1.f),
    C_Vertex3f(0.285971642f, 0.381270885f, -1.f),
    C_Vertex3f(0.290061951f, 0.293021679f, -1.f),
    C_Vertex3f(0.311928749f, 0.375420094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154891968f, 0.442318916f, -1.f),
    C_Vertex3f(0.159310341f, 0.333342552f, -1.f),
    C_Vertex3f(0.164276123f, 0.429634094f, -1.f),
    C_Vertex3f(0.169827223f, 0.328565717f, -1.f),
    C_Vertex3f(0.173605919f, 0.421268463f, -1.f),
    C_Vertex3f(0.182008982f, 0.32371366f, -1.f),
    C_Vertex3f(0.182229519f, 0.415657163f, -1.f),
    C_Vertex3f(0.195855618f, 0.318786383f, -1.f),
    C_Vertex3f(0.193329334f, 0.410011649f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.46875f, 0.166463852f, -1.f),
    C_Vertex3f(0.466480732f, 0.152168751f, -1.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, -1.f),
    C_Vertex3f(0.480051041f, 0.0401318073f, -1.f),
    C_Vertex3f(0.46875f, 0.181438446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.461942196f, 0.138444424f, -1.f),
    C_Vertex3f(0.455134392f, 0.125290871f, -1.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, -1.f),
    C_Vertex3f(0.468730688f, 0.0319672823f, -1.f),
    C_Vertex3f(0.466480732f, 0.152168751f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.420767784f, 0.0915622711f, -1.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, -1.f),
    C_Vertex3f(0.434601784f, 0.101465225f, -1.f),
    C_Vertex3f(0.44369173f, 0.0175218582f, -1.f),
    C_Vertex3f(0.44605732f, 0.112708092f, -1.f),
    C_Vertex3f(0.456610918f, 0.0244306326f, -1.f),
    C_Vertex3f(0.455134392f, 0.125290871f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.404555321f, 0.0829992294f, -1.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, -1.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, -1.f),
    C_Vertex3f(0.423113823f, 0.00810050964f, -1.f),
    C_Vertex3f(0.420767784f, 0.0915622711f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.224239349f, -0.00281953812f, -1.f),
    C_Vertex3f(0.242465973f, -0.00598859787f, -1.f),
    C_Vertex3f(0.245275974f, 0.072081089f, -1.f),
    C_Vertex3f(0.261573792f, -0.00836539268f, -1.f),
    C_Vertex3f(0.269881725f, 0.0673213005f, -1.f),
    C_Vertex3f(0.281562805f, -0.00994992256f, -1.f),
    C_Vertex3f(0.295854568f, 0.0649414063f, -1.f),
    C_Vertex3f(0.302433014f, -0.0107421875f, -1.f),
    C_Vertex3f(0.321662903f, 0.0649414063f, -1.f),
    C_Vertex3f(0.330471039f, -0.0107421875f, -1.f),
    C_Vertex3f(0.344518185f, 0.0667471886f, -1.f),
    C_Vertex3f(0.363073826f, -0.00760173798f, -1.f),
    C_Vertex3f(0.365952015f, 0.0703587532f, -1.f),
    C_Vertex3f(0.393954754f, -0.00132083893f, -1.f),
    C_Vertex3f(0.385964394f, 0.0757761002f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132882595f, 0.159187317f, -1.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, -1.f),
    C_Vertex3f(0.141385078f, 0.14159584f, -1.f),
    C_Vertex3f(0.146235347f, 0.0249164104f, -1.f),
    C_Vertex3f(0.151774406f, 0.12616539f, -1.f),
    C_Vertex3f(0.160144806f, 0.0177793503f, -1.f),
    C_Vertex3f(0.165034294f, 0.112212658f, -1.f),
    C_Vertex3f(0.174846649f, 0.0114412308f, -1.f),
    C_Vertex3f(0.181164742f, 0.0997376442f, -1.f),
    C_Vertex3f(0.190429688f, 0.00589537621f, -1.f),
    C_Vertex3f(0.200165749f, 0.0887403488f, -1.f),
    C_Vertex3f(0.206893921f, 0.00114178658f, -1.f),
    C_Vertex3f(0.222037315f, 0.0792207718f, -1.f),
    C_Vertex3f(0.224239349f, -0.00281953812f, -1.f),
    C_Vertex3f(0.245275974f, 0.072081089f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0985213518f, 0.0614540577f, -1.f),
    C_Vertex3f(0.109261274f, 0.0511212349f, -1.f),
    C_Vertex3f(0.120117188f, 0.212890625f, -1.f),
    C_Vertex3f(0.120793581f, 0.0415873528f, -1.f),
    C_Vertex3f(0.121538162f, 0.200853348f, -1.f),
    C_Vertex3f(0.133118272f, 0.0328524113f, -1.f),
    C_Vertex3f(0.126266956f, 0.178939819f, -1.f),
    C_Vertex3f(0.132882595f, 0.159187317f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120117188f, 0.212890625f, -1.f),
    C_Vertex3f(0.0400390625f, 0.206054688f, -1.f),
    C_Vertex3f(0.0402574539f, 0.198022842f, -1.f),
    C_Vertex3f(0.0416849852f, 0.182375908f, -1.f),
    C_Vertex3f(0.0441032648f, 0.167145729f, -1.f),
    C_Vertex3f(0.0475122929f, 0.152332306f, -1.f),
    C_Vertex3f(0.0519120693f, 0.137935638f, -1.f),
    C_Vertex3f(0.0573025942f, 0.123955727f, -1.f),
    C_Vertex3f(0.0636838675f, 0.11039257f, -1.f),
    C_Vertex3f(0.0710558891f, 0.09724617f, -1.f),
    C_Vertex3f(0.0794186592f, 0.0845165253f, -1.f),
    C_Vertex3f(0.0885738134f, 0.0725858212f, -1.f),
    C_Vertex3f(0.0985213518f, 0.0614540577f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.435302734f, 0.802734375f, -1.f),
    C_Vertex3f(0.345947266f, 0.802734375f, -1.f),
    C_Vertex3f(0.336914063f, 0.680175781f, -1.f),
    C_Vertex3f(0.297851563f, 0.728027344f, -1.f),
    C_Vertex3f(0.252929688f, 0.680175781f, -1.f),
    C_Vertex3f(0.248779297f, 0.802734375f, -1.f),
    C_Vertex3f(0.159423828f, 0.802734375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0667786598f, 0.443130016f, -1.f),
    C_Vertex3f(0.0639648438f, 0.487323761f, -1.f),
    C_Vertex3f(0.0639648438f, 0.464147568f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_139[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.864222527f, 0.503109872f, 0.f),
    C_Vertex3f(0.126953125f, 0.230033875f, 0.f),
    C_Vertex3f(0.126953125f, 0.230033875f, -1.f),
    C_Vertex3f(0.242675781f, 0.03125f, -1.f),
    C_Vertex3f(0.242675781f, 0.03125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.242675781f, 0.03125f, 0.f),
    C_Vertex3f(0.242675781f, 0.03125f, -1.f),
    C_Vertex3f(0.177734375f, 0.03125f, -1.f),
    C_Vertex3f(0.177734375f, 0.03125f, 0.f),
    C_Normal3f(-0.822045684f, -0.569421649f, 0.f),
    C_Vertex3f(0.177734375f, 0.03125f, 0.f),
    C_Vertex3f(0.177734375f, 0.03125f, -1.f),
    C_Vertex3f(0.0400390625f, 0.230033875f, -1.f),
    C_Vertex3f(0.0400390625f, 0.230033875f, 0.f),
    C_Normal3f(-0.823206663f, 0.567741752f, 0.f),
    C_Vertex3f(0.0400390625f, 0.230033875f, 0.f),
    C_Vertex3f(0.0400390625f, 0.230033875f, -1.f),
    C_Vertex3f(0.177734375f, 0.4296875f, -1.f),
    C_Vertex3f(0.177734375f, 0.4296875f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.177734375f, 0.4296875f, 0.f),
    C_Vertex3f(0.177734375f, 0.4296875f, -1.f),
    C_Vertex3f(0.242675781f, 0.4296875f, -1.f),
    C_Vertex3f(0.242675781f, 0.4296875f, 0.f),
    C_Normal3f(0.86517489f, -0.501470149f, 0.f),
    C_Vertex3f(0.242675781f, 0.4296875f, 0.f),
    C_Vertex3f(0.242675781f, 0.4296875f, -1.f),
    C_Vertex3f(0.126953125f, 0.230033875f, -1.f),
    C_Vertex3f(0.126953125f, 0.230033875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.126953125f, 0.230033875f, 0.f),
    C_Vertex3f(0.242675781f, 0.4296875f, 0.f),
    C_Vertex3f(0.177734375f, 0.4296875f, 0.f),
    C_Vertex3f(0.0400390625f, 0.230033875f, 0.f),
    C_Vertex3f(0.177734375f, 0.03125f, 0.f),
    C_Vertex3f(0.242675781f, 0.03125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.126953125f, 0.230033875f, -1.f),
    C_Vertex3f(0.242675781f, 0.4296875f, -1.f),
    C_Vertex3f(0.177734375f, 0.4296875f, -1.f),
    C_Vertex3f(0.0400390625f, 0.230033875f, -1.f),
    C_Vertex3f(0.177734375f, 0.03125f, -1.f),
    C_Vertex3f(0.242675781f, 0.03125f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_140[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.866699219f, 0.0756835938f, 0.f),
    C_Vertex3f(0.866699219f, 0.0756835938f, -1.f),
    C_Vertex3f(0.866699219f, 0.f, -1.f),
    C_Vertex3f(0.866699219f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.866699219f, 0.f, 0.f),
    C_Vertex3f(0.866699219f, 0.f, -1.f),
    C_Vertex3f(0.489746094f, 0.f, -1.f),
    C_Vertex3f(0.489746094f, 0.f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489746094f, 0.f, 0.f),
    C_Vertex3f(0.489746094f, 0.f, -1.f),
    C_Vertex3f(0.489746094f, 0.0927047729f, -1.f),
    C_Vertex3f(0.489746094f, 0.0927047729f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.868629158f, -0.495462656f, 0.f),
    C_Vertex3f(0.489746094f, 0.0927047729f, 0.f),
    C_Vertex3f(0.489746094f, 0.0927047729f, -1.f),
    C_Normal3f(0.829929769f, -0.557867885f, 0.f),
    C_Vertex3f(0.482370377f, 0.0797739029f, 0.f),
    C_Vertex3f(0.482370377f, 0.0797739029f, -1.f),
    C_Normal3f(0.769939601f, -0.638116777f, 0.f),
    C_Vertex3f(0.465843201f, 0.0571448803f, 0.f),
    C_Vertex3f(0.465843201f, 0.0571448803f, -1.f),
    C_Normal3f(0.679628193f, -0.733556807f, 0.f),
    C_Vertex3f(0.447540283f, 0.0377485752f, 0.f),
    C_Vertex3f(0.447540283f, 0.0377485752f, -1.f),
    C_Normal3f(0.570055783f, -0.82160604f, 0.f),
    C_Vertex3f(0.427461624f, 0.0215849876f, 0.f),
    C_Vertex3f(0.427461624f, 0.0215849876f, -1.f),
    C_Normal3f(0.445429385f, -0.895317137f, 0.f),
    C_Vertex3f(0.405607224f, 0.00865411758f, 0.f),
    C_Vertex3f(0.405607224f, 0.00865411758f, -1.f),
    C_Normal3f(0.31305778f, -0.949734032f, 0.f),
    C_Vertex3f(0.381977081f, -0.00104403496f, 0.f),
    C_Vertex3f(0.381977081f, -0.00104403496f, -1.f),
    C_Normal3f(0.181361005f, -0.983416557f, 0.f),
    C_Vertex3f(0.356571198f, -0.00750946999f, 0.f),
    C_Vertex3f(0.356571198f, -0.00750946999f, -1.f),
    C_Normal3f(0.0555485263f, -0.998456001f, 0.f),
    C_Vertex3f(0.329389572f, -0.0107421875f, 0.f),
    C_Vertex3f(0.329389572f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0511894412f, -0.998688996f, 0.f),
    C_Vertex3f(0.298464775f, -0.0107421875f, 0.f),
    C_Vertex3f(0.298464775f, -0.0107421875f, -1.f),
    C_Normal3f(-0.156996086f, -0.987599254f, 0.f),
    C_Vertex3f(0.267566919f, -0.00757336617f, 0.f),
    C_Vertex3f(0.267566919f, -0.00757336617f, -1.f),
    C_Normal3f(-0.273089767f, -0.961988568f, 0.f),
    C_Vertex3f(0.238663435f, -0.0012357235f, 0.f),
    C_Vertex3f(0.238663435f, -0.0012357235f, -1.f),
    C_Normal3f(-0.39349252f, -0.919327796f, 0.f),
    C_Vertex3f(0.211754322f, 0.00827074051f, 0.f),
    C_Vertex3f(0.211754322f, 0.00827074051f, -1.f),
    C_Normal3f(-0.512095094f, -0.8589288f, 0.f),
    C_Vertex3f(0.186839581f, 0.0209460258f, 0.f),
    C_Vertex3f(0.186839581f, 0.0209460258f, -1.f),
    C_Normal3f(-0.622296572f, -0.782781541f, 0.f),
    C_Vertex3f(0.16391921f, 0.0367901325f, 0.f),
    C_Vertex3f(0.16391921f, 0.0367901325f, -1.f),
    C_Normal3f(-0.718673587f, -0.695347667f, 0.f),
    C_Vertex3f(0.142993212f, 0.0558030605f, 0.f),
    C_Vertex3f(0.142993212f, 0.0558030605f, -1.f),
    C_Normal3f(-0.797603071f, -0.603182733f, 0.f),
    C_Vertex3f(0.124061584f, 0.0779848099f, 0.f),
    C_Vertex3f(0.124061584f, 0.0779848099f, -1.f),
    C_Normal3f(-0.856155276f, -0.516718745f, 0.f),
    C_Vertex3f(0.10816288f, 0.101860046f, 0.f),
    C_Vertex3f(0.10816288f, 0.101860046f, -1.f),
    C_Normal3f(-0.899227798f, -0.437480718f, 0.f),
    C_Vertex3f(0.0951602459f, 0.125871658f, 0.f),
    C_Vertex3f(0.0951602459f, 0.125871658f, -1.f),
    C_Normal3f(-0.932737887f, -0.360555381f, 0.f),
    C_Vertex3f(0.084015131f, 0.15149498f, 0.f),
    C_Vertex3f(0.084015131f, 0.15149498f, -1.f),
    C_Normal3f(-0.958327413f, -0.285672218f, 0.f),
    C_Vertex3f(0.0747275352f, 0.178730011f, 0.f),
    C_Vertex3f(0.0747275352f, 0.178730011f, -1.f),
    C_Normal3f(-0.976798058f, -0.214162499f, 0.f),
    C_Vertex3f(0.0672974586f, 0.207576752f, 0.f),
    C_Vertex3f(0.0672974586f, 0.207576752f, -1.f),
    C_Normal3f(-0.98914808f, -0.146921664f, 0.f),
    C_Vertex3f(0.0617249012f, 0.238035202f, 0.f),
    C_Vertex3f(0.0617249012f, 0.238035202f, -1.f),
    C_Normal3f(-0.996427774f, -0.0844483599f, 0.f),
    C_Vertex3f(0.0580098629f, 0.270105362f, 0.f),
    C_Vertex3f(0.0580098629f, 0.270105362f, -1.f),
    C_Normal3f(-0.999637008f, -0.0269419793f, 0.f),
    C_Vertex3f(0.0561523438f, 0.303787231f, 0.f),
    C_Vertex3f(0.0561523438f, 0.303787231f, -1.f),
    C_Normal3f(-0.999605f, 0.0281032939f, 0.f),
    C_Vertex3f(0.0561523438f, 0.339025497f, 0.f),
    C_Vertex3f(0.0561523438f, 0.339025497f, -1.f),
    C_Normal3f(-0.996143818f, 0.0877356455f, 0.f),
    C_Vertex3f(0.0580918789f, 0.372774482f, 0.f),
    C_Vertex3f(0.0580918789f, 0.372774482f, -1.f),
    C_Normal3f(-0.988429964f, 0.151678324f, 0.f),
    C_Vertex3f(0.0619709492f, 0.405089498f, 0.f),
    C_Vertex3f(0.0619709492f, 0.405089498f, -1.f),
    C_Normal3f(-0.975599706f, 0.219557121f, 0.f),
    C_Vertex3f(0.0677895546f, 0.435970545f, 0.f),
    C_Vertex3f(0.0677895546f, 0.435970545f, -1.f),
    C_Normal3f(-0.956822515f, 0.290672868f, 0.f),
    C_Vertex3f(0.0755476952f, 0.465417624f, 0.f),
    C_Vertex3f(0.0755476952f, 0.465417624f, -1.f),
    C_Normal3f(-0.931400716f, 0.363995403f, 0.f),
    C_Vertex3f(0.0852453709f, 0.493430734f, 0.f),
    C_Vertex3f(0.0852453709f, 0.493430734f, -1.f),
    C_Normal3f(-0.898887634f, 0.438179106f, 0.f),
    C_Vertex3f(0.0968825817f, 0.520009875f, 0.f),
    C_Vertex3f(0.0968825817f, 0.520009875f, -1.f),
    C_Normal3f(-0.859201312f, 0.511637747f, 0.f),
    C_Vertex3f(0.110459328f, 0.545155048f, 0.f),
    C_Vertex3f(0.110459328f, 0.545155048f, -1.f),
    C_Normal3f(-0.800377607f, 0.599496126f, 0.f),
    C_Vertex3f(0.125975609f, 0.568866253f, 0.f),
    C_Vertex3f(0.125975609f, 0.568866253f, -1.f),
    C_Normal3f(-0.713515759f, 0.700639188f, 0.f),
    C_Vertex3f(0.143759489f, 0.589613557f, 0.f),
    C_Vertex3f(0.143759489f, 0.589613557f, -1.f),
    C_Normal3f(-0.60983026f, 0.792532086f, 0.f),
    C_Vertex3f(0.163810968f, 0.60739696f, 0.f),
    C_Vertex3f(0.163810968f, 0.60739696f, -1.f),
    C_Normal3f(-0.494347095f, 0.869264603f, 0.f),
    C_Vertex3f(0.186130047f, 0.622216463f, 0.f),
    C_Vertex3f(0.186130047f, 0.622216463f, -1.f),
    C_Normal3f(-0.37404561f, 0.927410305f, 0.f),
    C_Vertex3f(0.210716724f, 0.634072065f, 0.f),
    C_Vertex3f(0.210716724f, 0.634072065f, -1.f),
    C_Normal3f(-0.25592947f, 0.966695487f, 0.f),
    C_Vertex3f(0.237571001f, 0.642963767f, 0.f),
    C_Vertex3f(0.237571001f, 0.642963767f, -1.f),
    C_Normal3f(-0.145381525f, 0.989375591f, 0.f),
    C_Vertex3f(0.266692877f, 0.648891568f, 0.f),
    C_Vertex3f(0.266692877f, 0.648891568f, -1.f),
    C_Normal3f(-0.047261782f, 0.998882532f, 0.f),
    C_Vertex3f(0.298082352f, 0.651855469f, 0.f),
    C_Vertex3f(0.298082352f, 0.651855469f, -1.f),
    C_Normal3f(0.0523123406f, 0.998630762f, 0.f),
    C_Vertex3f(0.329335213f, 0.651855469f, 0.f),
    C_Vertex3f(0.329335213f, 0.651855469f, -1.f),
    C_Normal3f(0.172221094f, 0.985058308f, 0.f),
    C_Vertex3f(0.356435299f, 0.648798704f, 0.f),
    C_Vertex3f(0.356435299f, 0.648798704f, -1.f),
    C_Normal3f(0.298018217f, 0.95456022f, 0.f),
    C_Vertex3f(0.381786823f, 0.642685175f, 0.f),
    C_Vertex3f(0.381786823f, 0.642685175f, -1.f),
    C_Normal3f(0.425889134f, 0.904775321f, 0.f),
    C_Vertex3f(0.405389786f, 0.633514881f, 0.f),
    C_Vertex3f(0.405389786f, 0.633514881f, -1.f),
    C_Normal3f(0.548300862f, 0.8362813f, 0.f),
    C_Vertex3f(0.427244186f, 0.621287823f, 0.f),
    C_Vertex3f(0.427244186f, 0.621287823f, -1.f),
    C_Normal3f(0.658162892f, 0.752875507f, 0.f),
    C_Vertex3f(0.447350025f, 0.606004f, 0.f),
    C_Vertex3f(0.447350025f, 0.606004f, -1.f),
    C_Normal3f(0.750745296f, 0.660591722f, 0.f),
    C_Vertex3f(0.465707302f, 0.587663412f, 0.f),
    C_Vertex3f(0.465707302f, 0.587663412f, -1.f),
    C_Normal3f(0.813489079f, 0.581580281f, 0.f),
    C_Vertex3f(0.482316017f, 0.56626606f, 0.f),
    C_Vertex3f(0.482316017f, 0.56626606f, -1.f),
    C_Normal3f(0.854585767f, 0.519310355f, 0.f),
    C_Vertex3f(0.489746094f, 0.554039001f, 0.f),
    C_Vertex3f(0.489746094f, 0.554039001f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489746094f, 0.554039001f, 0.f),
    C_Vertex3f(0.489746094f, 0.554039001f, -1.f),
    C_Vertex3f(0.489746094f, 0.640625f, -1.f),
    C_Vertex3f(0.489746094f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.489746094f, 0.640625f, 0.f),
    C_Vertex3f(0.489746094f, 0.640625f, -1.f),
    C_Vertex3f(0.853027344f, 0.640625f, -1.f),
    C_Vertex3f(0.853027344f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.853027344f, 0.640625f, 0.f),
    C_Vertex3f(0.853027344f, 0.640625f, -1.f),
    C_Vertex3f(0.853027344f, 0.564941406f, -1.f),
    C_Vertex3f(0.853027344f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.853027344f, 0.564941406f, 0.f),
    C_Vertex3f(0.853027344f, 0.564941406f, -1.f),
    C_Vertex3f(0.568847656f, 0.564941406f, -1.f),
    C_Vertex3f(0.568847656f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.568847656f, 0.564941406f, 0.f),
    C_Vertex3f(0.568847656f, 0.564941406f, -1.f),
    C_Vertex3f(0.568847656f, 0.369628906f, -1.f),
    C_Vertex3f(0.568847656f, 0.369628906f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.568847656f, 0.369628906f, 0.f),
    C_Vertex3f(0.568847656f, 0.369628906f, -1.f),
    C_Vertex3f(0.830566406f, 0.369628906f, -1.f),
    C_Vertex3f(0.830566406f, 0.369628906f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.830566406f, 0.369628906f, 0.f),
    C_Vertex3f(0.830566406f, 0.369628906f, -1.f),
    C_Vertex3f(0.830566406f, 0.293945313f, -1.f),
    C_Vertex3f(0.830566406f, 0.293945313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.830566406f, 0.293945313f, 0.f),
    C_Vertex3f(0.830566406f, 0.293945313f, -1.f),
    C_Vertex3f(0.568847656f, 0.293945313f, -1.f),
    C_Vertex3f(0.568847656f, 0.293945313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.568847656f, 0.293945313f, 0.f),
    C_Vertex3f(0.568847656f, 0.293945313f, -1.f),
    C_Vertex3f(0.568847656f, 0.0756835938f, -1.f),
    C_Vertex3f(0.568847656f, 0.0756835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.568847656f, 0.0756835938f, 0.f),
    C_Vertex3f(0.568847656f, 0.0756835938f, -1.f),
    C_Vertex3f(0.866699219f, 0.0756835938f, -1.f),
    C_Vertex3f(0.866699219f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0528392456f, -0.998602986f, 0.f),
    C_Vertex3f(0.323115349f, 0.583007813f, 0.f),
    C_Vertex3f(0.323115349f, 0.583007813f, -1.f),
    C_Normal3f(0.0851590186f, -0.996367395f, 0.f),
    C_Vertex3f(0.303394318f, 0.583007813f, 0.f),
    C_Vertex3f(0.303394318f, 0.583007813f, -1.f),
    C_Normal3f(0.238635436f, -0.971109152f, 0.f),
    C_Vertex3f(0.281596184f, 0.57945919f, 0.f),
    C_Vertex3f(0.281596184f, 0.57945919f, -1.f),
    C_Normal3f(0.383145303f, -0.923688054f, 0.f),
    C_Vertex3f(0.260071754f, 0.572361946f, 0.f),
    C_Vertex3f(0.260071754f, 0.572361946f, -1.f),
    C_Normal3f(0.498932868f, -0.866640568f, 0.f),
    C_Vertex3f(0.23882103f, 0.56171608f, 0.f),
    C_Vertex3f(0.23882103f, 0.56171608f, -1.f),
    C_Normal3f(0.589432776f, -0.80781734f, 0.f),
    C_Vertex3f(0.223088264f, 0.551070213f, 0.f),
    C_Vertex3f(0.223088264f, 0.551070213f, -1.f),
    C_Normal3f(0.66487807f, -0.746951938f, 0.f),
    C_Vertex3f(0.213241696f, 0.543051839f, 0.f),
    C_Vertex3f(0.213241696f, 0.543051839f, -1.f),
    C_Normal3f(0.726788878f, -0.686861038f, 0.f),
    C_Vertex3f(0.20403707f, 0.534112334f, 0.f),
    C_Vertex3f(0.20403707f, 0.534112334f, -1.f),
    C_Normal3f(0.781431556f, -0.623990893f, 0.f),
    C_Vertex3f(0.195474386f, 0.524251699f, 0.f),
    C_Vertex3f(0.195474386f, 0.524251699f, -1.f),
    C_Normal3f(0.828465879f, -0.560039639f, 0.f),
    C_Vertex3f(0.187553644f, 0.513469934f, 0.f),
    C_Vertex3f(0.187553644f, 0.513469934f, -1.f),
    C_Normal3f(0.868019223f, -0.496530533f, 0.f),
    C_Vertex3f(0.180274844f, 0.501767039f, 0.f),
    C_Vertex3f(0.180274844f, 0.501767039f, -1.f),
    C_Normal3f(0.900573552f, -0.434703767f, 0.f),
    C_Vertex3f(0.173637986f, 0.489143014f, 0.f),
    C_Vertex3f(0.173637986f, 0.489143014f, -1.f),
    C_Normal3f(0.926832557f, -0.37547493f, 0.f),
    C_Vertex3f(0.16764307f, 0.475597858f, 0.f),
    C_Vertex3f(0.16764307f, 0.475597858f, -1.f),
    C_Normal3f(0.948692262f, -0.316200703f, 0.f),
    C_Vertex3f(0.162290096f, 0.461131573f, 0.f),
    C_Vertex3f(0.162290096f, 0.461131573f, -1.f),
    C_Normal3f(0.965948403f, -0.258734852f, 0.f),
    C_Vertex3f(0.157606244f, 0.445484519f, 0.f),
    C_Vertex3f(0.157606244f, 0.445484519f, -1.f),
    C_Normal3f(0.97840029f, -0.206719294f, 0.f),
    C_Vertex3f(0.153591514f, 0.428656697f, 0.f),
    C_Vertex3f(0.153591514f, 0.428656697f, -1.f),
    C_Normal3f(0.98714745f, -0.15981248f, 0.f),
    C_Vertex3f(0.150245905f, 0.410648108f, 0.f),
    C_Vertex3f(0.150245905f, 0.410648108f, -1.f),
    C_Normal3f(0.993063569f, -0.117578961f, 0.f),
    C_Vertex3f(0.147569418f, 0.39145875f, 0.f),
    C_Vertex3f(0.147569418f, 0.39145875f, -1.f),
    C_Normal3f(0.996830523f, -0.0795544833f, 0.f),
    C_Vertex3f(0.145562053f, 0.371088624f, 0.f),
    C_Vertex3f(0.145562053f, 0.371088624f, -1.f),
    C_Normal3f(0.998974144f, -0.0452843308f, 0.f),
    C_Vertex3f(0.144223809f, 0.34953773f, 0.f),
    C_Vertex3f(0.144223809f, 0.34953773f, -1.f),
    C_Normal3f(0.999914646f, -0.0130604235f, 0.f),
    C_Vertex3f(0.143554688f, 0.326806068f, 0.f),
    C_Vertex3f(0.143554688f, 0.326806068f, -1.f),
    C_Normal3f(0.999647617f, 0.0265483372f, 0.f),
    C_Vertex3f(0.143554688f, 0.298309326f, 0.f),
    C_Vertex3f(0.143554688f, 0.298309326f, -1.f),
    C_Normal3f(0.996863961f, 0.0791335255f, 0.f),
    C_Vertex3f(0.145132065f, 0.267411709f, 0.f),
    C_Vertex3f(0.145132065f, 0.267411709f, -1.f),
    C_Normal3f(0.989962697f, 0.141329244f, 0.f),
    C_Vertex3f(0.148286819f, 0.238697529f, 0.f),
    C_Vertex3f(0.148286819f, 0.238697529f, -1.f),
    C_Normal3f(0.977251291f, 0.212084755f, 0.f),
    C_Vertex3f(0.153018951f, 0.212166786f, 0.f),
    C_Vertex3f(0.153018951f, 0.212166786f, -1.f),
    C_Normal3f(0.956439734f, 0.291929811f, 0.f),
    C_Vertex3f(0.159328461f, 0.187819481f, 0.f),
    C_Vertex3f(0.159328461f, 0.187819481f, -1.f),
    C_Normal3f(0.924696505f, 0.380705059f, 0.f),
    C_Vertex3f(0.167215347f, 0.165655613f, 0.f),
    C_Vertex3f(0.167215347f, 0.165655613f, -1.f),
    C_Normal3f(0.878871381f, 0.477058798f, 0.f),
    C_Vertex3f(0.176679611f, 0.145675182f, 0.f),
    C_Vertex3f(0.176679611f, 0.145675182f, -1.f),
    C_Normal3f(0.816083729f, 0.577933609f, 0.f),
    C_Vertex3f(0.187721252f, 0.127878189f, 0.f),
    C_Vertex3f(0.187721252f, 0.127878189f, -1.f),
    C_Normal3f(0.749653459f, 0.661830604f, 0.f),
    C_Vertex3f(0.200340271f, 0.112264633f, 0.f),
    C_Vertex3f(0.200340271f, 0.112264633f, -1.f),
    C_Normal3f(0.683971703f, 0.729508519f, 0.f),
    C_Vertex3f(0.213567019f, 0.0986027718f, 0.f),
    C_Vertex3f(0.213567019f, 0.0986027718f, -1.f),
    C_Normal3f(0.60469687f, 0.796455741f, 0.f),
    C_Vertex3f(0.227401495f, 0.0868926048f, 0.f),
    C_Vertex3f(0.227401495f, 0.0868926048f, -1.f),
    C_Normal3f(0.511707664f, 0.859159648f, 0.f),
    C_Vertex3f(0.2418437f, 0.0771341324f, 0.f),
    C_Vertex3f(0.2418437f, 0.0771341324f, -1.f),
    C_Normal3f(0.406487256f, 0.913656414f, 0.f),
    C_Vertex3f(0.256893635f, 0.0693273544f, 0.f),
    C_Vertex3f(0.256893635f, 0.0693273544f, -1.f),
    C_Normal3f(0.292333752f, 0.956316352f, 0.f),
    C_Vertex3f(0.272551298f, 0.063472271f, 0.f),
    C_Vertex3f(0.272551298f, 0.063472271f, -1.f),
    C_Normal3f(0.17399025f, 0.98474741f, 0.f),
    C_Vertex3f(0.28881669f, 0.059568882f, 0.f),
    C_Vertex3f(0.28881669f, 0.059568882f, -1.f),
    C_Normal3f(0.0566305444f, 0.998395264f, 0.f),
    C_Vertex3f(0.305689812f, 0.0576171875f, 0.f),
    C_Vertex3f(0.305689812f, 0.0576171875f, -1.f),
    C_Normal3f(-0.0562977642f, 0.99841398f, 0.f),
    C_Vertex3f(0.323225021f, 0.0576171875f, 0.f),
    C_Vertex3f(0.323225021f, 0.0576171875f, -1.f),
    C_Normal3f(-0.17270723f, 0.984973192f, 0.f),
    C_Vertex3f(0.340179563f, 0.0595619678f, 0.f),
    C_Vertex3f(0.340179563f, 0.0595619678f, -1.f),
    C_Normal3f(-0.290711612f, 0.956810772f, 0.f),
    C_Vertex3f(0.356499076f, 0.0634515285f, 0.f),
    C_Vertex3f(0.356499076f, 0.0634515285f, -1.f),
    C_Normal3f(-0.404993713f, 0.914319456f, 0.f),
    C_Vertex3f(0.372183561f, 0.0692858696f, 0.f),
    C_Vertex3f(0.372183561f, 0.0692858696f, -1.f),
    C_Normal3f(-0.510729432f, 0.859741449f, 0.f),
    C_Vertex3f(0.387233019f, 0.077064991f, 0.f),
    C_Vertex3f(0.387233019f, 0.077064991f, -1.f),
    C_Normal3f(-0.604461133f, 0.796634674f, 0.f),
    C_Vertex3f(0.401647449f, 0.0867888927f, 0.f),
    C_Vertex3f(0.401647449f, 0.0867888927f, -1.f),
    C_Normal3f(-0.684533238f, 0.728981614f, 0.f),
    C_Vertex3f(0.41542685f, 0.0984575748f, 0.f),
    C_Vertex3f(0.41542685f, 0.0984575748f, -1.f),
    C_Normal3f(-0.750935316f, 0.660375655f, 0.f),
    C_Vertex3f(0.428571224f, 0.112071037f, 0.f),
    C_Vertex3f(0.428571224f, 0.112071037f, -1.f),
    C_Normal3f(-0.820471585f, 0.571687222f, 0.f),
    C_Vertex3f(0.44108057f, 0.12762928f, 0.f),
    C_Vertex3f(0.44108057f, 0.12762928f, -1.f),
    C_Normal3f(-0.885565042f, 0.464515358f, 0.f),
    C_Vertex3f(0.452026248f, 0.145733118f, 0.f),
    C_Vertex3f(0.452026248f, 0.145733118f, -1.f),
    C_Normal3f(-0.930926681f, 0.365206122f, 0.f),
    C_Vertex3f(0.461408257f, 0.166382551f, 0.f),
    C_Vertex3f(0.461408257f, 0.166382551f, -1.f),
    C_Normal3f(-0.961048305f, 0.276380211f, 0.f),
    C_Vertex3f(0.469226599f, 0.189577579f, 0.f),
    C_Vertex3f(0.469226599f, 0.189577579f, -1.f),
    C_Normal3f(-0.980088115f, 0.198562935f, 0.f),
    C_Vertex3f(0.475481272f, 0.215318203f, 0.f),
    C_Vertex3f(0.475481272f, 0.215318203f, -1.f),
    C_Normal3f(-0.991368055f, 0.13110812f, 0.f),
    C_Vertex3f(0.480172276f, 0.243604422f, 0.f),
    C_Vertex3f(0.480172276f, 0.243604422f, -1.f),
    C_Normal3f(-0.997341812f, 0.0728639141f, 0.f),
    C_Vertex3f(0.483299613f, 0.274436235f, 0.f),
    C_Vertex3f(0.483299613f, 0.274436235f, -1.f),
    C_Normal3f(-0.999736071f, 0.0229743663f, 0.f),
    C_Vertex3f(0.484863281f, 0.307813644f, 0.f),
    C_Vertex3f(0.484863281f, 0.307813644f, -1.f),
    C_Normal3f(-0.999713063f, -0.0239544157f, 0.f),
    C_Vertex3f(0.484863281f, 0.342479706f, 0.f),
    C_Vertex3f(0.484863281f, 0.342479706f, -1.f),
    C_Normal3f(-0.996912837f, -0.0785162598f, 0.f),
    C_Vertex3f(0.483285904f, 0.373643875f, 0.f),
    C_Vertex3f(0.483285904f, 0.373643875f, -1.f),
    C_Normal3f(-0.990086198f, -0.14046064f, 0.f),
    C_Vertex3f(0.480131149f, 0.402563095f, 0.f),
    C_Vertex3f(0.480131149f, 0.402563095f, -1.f),
    C_Normal3f(-0.977444887f, -0.211190552f, 0.f),
    C_Vertex3f(0.475399017f, 0.429237366f, 0.f),
    C_Vertex3f(0.475399017f, 0.429237366f, -1.f),
    C_Normal3f(-0.956618905f, -0.291342288f, 0.f),
    C_Vertex3f(0.469089508f, 0.453666687f, 0.f),
    C_Vertex3f(0.469089508f, 0.453666687f, -1.f),
    C_Normal3f(-0.924631238f, -0.380863488f, 0.f),
    C_Vertex3f(0.461202621f, 0.475851059f, 0.f),
    C_Vertex3f(0.461202621f, 0.475851059f, -1.f),
    C_Normal3f(-0.878108084f, -0.478462338f, 0.f),
    C_Vertex3f(0.451738358f, 0.495790482f, 0.f),
    C_Vertex3f(0.451738358f, 0.495790482f, -1.f),
    C_Normal3f(-0.813891649f, -0.5810166f, 0.f),
    C_Vertex3f(0.440696716f, 0.513484955f, 0.f),
    C_Vertex3f(0.440696716f, 0.513484955f, -1.f),
    C_Normal3f(-0.74634999f, -0.665553749f, 0.f),
    C_Vertex3f(0.428077698f, 0.528934479f, 0.f),
    C_Vertex3f(0.428077698f, 0.528934479f, -1.f),
    C_Normal3f(-0.680679917f, -0.73258096f, 0.f),
    C_Vertex3f(0.414864779f, 0.542452812f, 0.f),
    C_Vertex3f(0.414864779f, 0.542452812f, -1.f),
    C_Normal3f(-0.6015836f, -0.798809826f, 0.f),
    C_Vertex3f(0.401057959f, 0.554039955f, 0.f),
    C_Vertex3f(0.401057959f, 0.554039955f, -1.f),
    C_Normal3f(-0.508962333f, -0.860788703f, 0.f),
    C_Vertex3f(0.386657238f, 0.563695908f, 0.f),
    C_Vertex3f(0.386657238f, 0.563695908f, -1.f),
    C_Normal3f(-0.404285252f, -0.914632976f, 0.f),
    C_Vertex3f(0.371662617f, 0.57142067f, 0.f),
    C_Vertex3f(0.371662617f, 0.57142067f, -1.f),
    C_Normal3f(-0.290790468f, -0.956786752f, 0.f),
    C_Vertex3f(0.356074095f, 0.577214241f, 0.f),
    C_Vertex3f(0.356074095f, 0.577214241f, -1.f),
    C_Normal3f(-0.173128083f, -0.984899282f, 0.f),
    C_Vertex3f(0.339891672f, 0.581076622f, 0.f),
    C_Vertex3f(0.339891672f, 0.581076622f, -1.f),
    C_Normal3f(-0.0528392456f, -0.998602986f, 0.f),
    C_Vertex3f(0.323115349f, 0.583007813f, 0.f),
    C_Vertex3f(0.323115349f, 0.583007813f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.489746094f, 0.0927047729f, 0.f),
    C_Vertex3f(0.489746094f, 0.f, 0.f),
    C_Vertex3f(0.866699219f, 0.f, 0.f),
    C_Vertex3f(0.568847656f, 0.0756835938f, 0.f),
    C_Vertex3f(0.489746094f, 0.554039001f, 0.f),
    C_Vertex3f(0.484863281f, 0.342479706f, 0.f),
    C_Vertex3f(0.484863281f, 0.307813644f, 0.f),
    C_Vertex3f(0.483299613f, 0.274436235f, 0.f),
    C_Vertex3f(0.482370377f, 0.0797739029f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.475399017f, 0.429237366f, 0.f),
    C_Vertex3f(0.480131149f, 0.402563095f, 0.f),
    C_Vertex3f(0.482316017f, 0.56626606f, 0.f),
    C_Vertex3f(0.483285904f, 0.373643875f, 0.f),
    C_Vertex3f(0.489746094f, 0.554039001f, 0.f),
    C_Vertex3f(0.484863281f, 0.342479706f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.451738358f, 0.495790482f, 0.f),
    C_Vertex3f(0.461202621f, 0.475851059f, 0.f),
    C_Vertex3f(0.465707302f, 0.587663412f, 0.f),
    C_Vertex3f(0.469089508f, 0.453666687f, 0.f),
    C_Vertex3f(0.482316017f, 0.56626606f, 0.f),
    C_Vertex3f(0.475399017f, 0.429237366f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.447350025f, 0.606004f, 0.f),
    C_Vertex3f(0.427244186f, 0.621287823f, 0.f),
    C_Vertex3f(0.428077698f, 0.528934479f, 0.f),
    C_Vertex3f(0.440696716f, 0.513484955f, 0.f),
    C_Vertex3f(0.451738358f, 0.495790482f, 0.f),
    C_Vertex3f(0.465707302f, 0.587663412f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.386657238f, 0.563695908f, 0.f),
    C_Vertex3f(0.401057959f, 0.554039955f, 0.f),
    C_Vertex3f(0.405389786f, 0.633514881f, 0.f),
    C_Vertex3f(0.414864779f, 0.542452812f, 0.f),
    C_Vertex3f(0.427244186f, 0.621287823f, 0.f),
    C_Vertex3f(0.428077698f, 0.528934479f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339891672f, 0.581076622f, 0.f),
    C_Vertex3f(0.356074095f, 0.577214241f, 0.f),
    C_Vertex3f(0.356435299f, 0.648798704f, 0.f),
    C_Vertex3f(0.371662617f, 0.57142067f, 0.f),
    C_Vertex3f(0.381786823f, 0.642685175f, 0.f),
    C_Vertex3f(0.386657238f, 0.563695908f, 0.f),
    C_Vertex3f(0.405389786f, 0.633514881f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329335213f, 0.651855469f, 0.f),
    C_Vertex3f(0.298082352f, 0.651855469f, 0.f),
    C_Vertex3f(0.303394318f, 0.583007813f, 0.f),
    C_Vertex3f(0.323115349f, 0.583007813f, 0.f),
    C_Vertex3f(0.339891672f, 0.581076622f, 0.f),
    C_Vertex3f(0.356435299f, 0.648798704f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.23882103f, 0.56171608f, 0.f),
    C_Vertex3f(0.260071754f, 0.572361946f, 0.f),
    C_Vertex3f(0.266692877f, 0.648891568f, 0.f),
    C_Vertex3f(0.281596184f, 0.57945919f, 0.f),
    C_Vertex3f(0.298082352f, 0.651855469f, 0.f),
    C_Vertex3f(0.303394318f, 0.583007813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237571001f, 0.642963767f, 0.f),
    C_Vertex3f(0.210716724f, 0.634072065f, 0.f),
    C_Vertex3f(0.213241696f, 0.543051839f, 0.f),
    C_Vertex3f(0.223088264f, 0.551070213f, 0.f),
    C_Vertex3f(0.23882103f, 0.56171608f, 0.f),
    C_Vertex3f(0.266692877f, 0.648891568f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.210716724f, 0.634072065f, 0.f),
    C_Vertex3f(0.186130047f, 0.622216463f, 0.f),
    C_Vertex3f(0.187553644f, 0.513469934f, 0.f),
    C_Vertex3f(0.195474386f, 0.524251699f, 0.f),
    C_Vertex3f(0.20403707f, 0.534112334f, 0.f),
    C_Vertex3f(0.213241696f, 0.543051839f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.186130047f, 0.622216463f, 0.f),
    C_Vertex3f(0.163810968f, 0.60739696f, 0.f),
    C_Vertex3f(0.16764307f, 0.475597858f, 0.f),
    C_Vertex3f(0.173637986f, 0.489143014f, 0.f),
    C_Vertex3f(0.180274844f, 0.501767039f, 0.f),
    C_Vertex3f(0.187553644f, 0.513469934f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.163810968f, 0.60739696f, 0.f),
    C_Vertex3f(0.143759489f, 0.589613557f, 0.f),
    C_Vertex3f(0.145562053f, 0.371088624f, 0.f),
    C_Vertex3f(0.147569418f, 0.39145875f, 0.f),
    C_Vertex3f(0.150245905f, 0.410648108f, 0.f),
    C_Vertex3f(0.153591514f, 0.428656697f, 0.f),
    C_Vertex3f(0.157606244f, 0.445484519f, 0.f),
    C_Vertex3f(0.162290096f, 0.461131573f, 0.f),
    C_Vertex3f(0.16764307f, 0.475597858f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145132065f, 0.267411709f, 0.f),
    C_Vertex3f(0.143554688f, 0.298309326f, 0.f),
    C_Vertex3f(0.142993212f, 0.0558030605f, 0.f),
    C_Vertex3f(0.143554688f, 0.326806068f, 0.f),
    C_Vertex3f(0.143759489f, 0.589613557f, 0.f),
    C_Vertex3f(0.144223809f, 0.34953773f, 0.f),
    C_Vertex3f(0.145562053f, 0.371088624f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0561523438f, 0.339025497f, 0.f),
    C_Vertex3f(0.0580098629f, 0.270105362f, 0.f),
    C_Vertex3f(0.0580918789f, 0.372774482f, 0.f),
    C_Vertex3f(0.0617249012f, 0.238035202f, 0.f),
    C_Vertex3f(0.0619709492f, 0.405089498f, 0.f),
    C_Vertex3f(0.0672974586f, 0.207576752f, 0.f),
    C_Vertex3f(0.0677895546f, 0.435970545f, 0.f),
    C_Vertex3f(0.0747275352f, 0.178730011f, 0.f),
    C_Vertex3f(0.0755476952f, 0.465417624f, 0.f),
    C_Vertex3f(0.084015131f, 0.15149498f, 0.f),
    C_Vertex3f(0.0852453709f, 0.493430734f, 0.f),
    C_Vertex3f(0.0951602459f, 0.125871658f, 0.f),
    C_Vertex3f(0.0968825817f, 0.520009875f, 0.f),
    C_Vertex3f(0.10816288f, 0.101860046f, 0.f),
    C_Vertex3f(0.110459328f, 0.545155048f, 0.f),
    C_Vertex3f(0.124061584f, 0.0779848099f, 0.f),
    C_Vertex3f(0.125975609f, 0.568866253f, 0.f),
    C_Vertex3f(0.142993212f, 0.0558030605f, 0.f),
    C_Vertex3f(0.143759489f, 0.589613557f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.830566406f, 0.293945313f, 0.f),
    C_Vertex3f(0.568847656f, 0.369628906f, 0.f),
    C_Vertex3f(0.568847656f, 0.293945313f, 0.f),
    C_Vertex3f(0.568847656f, 0.564941406f, 0.f),
    C_Vertex3f(0.568847656f, 0.0756835938f, 0.f),
    C_Vertex3f(0.489746094f, 0.640625f, 0.f),
    C_Vertex3f(0.489746094f, 0.554039001f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.853027344f, 0.640625f, 0.f),
    C_Vertex3f(0.489746094f, 0.640625f, 0.f),
    C_Vertex3f(0.568847656f, 0.564941406f, 0.f),
    C_Vertex3f(0.853027344f, 0.564941406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.482370377f, 0.0797739029f, 0.f),
    C_Vertex3f(0.483299613f, 0.274436235f, 0.f),
    C_Vertex3f(0.480172276f, 0.243604422f, 0.f),
    C_Vertex3f(0.475481272f, 0.215318203f, 0.f),
    C_Vertex3f(0.469226599f, 0.189577579f, 0.f),
    C_Vertex3f(0.465843201f, 0.0571448803f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.469226599f, 0.189577579f, 0.f),
    C_Vertex3f(0.461408257f, 0.166382551f, 0.f),
    C_Vertex3f(0.465843201f, 0.0571448803f, 0.f),
    C_Vertex3f(0.452026248f, 0.145733118f, 0.f),
    C_Vertex3f(0.447540283f, 0.0377485752f, 0.f),
    C_Vertex3f(0.44108057f, 0.12762928f, 0.f),
    C_Vertex3f(0.428571224f, 0.112071037f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.447540283f, 0.0377485752f, 0.f),
    C_Vertex3f(0.428571224f, 0.112071037f, 0.f),
    C_Vertex3f(0.427461624f, 0.0215849876f, 0.f),
    C_Vertex3f(0.41542685f, 0.0984575748f, 0.f),
    C_Vertex3f(0.405607224f, 0.00865411758f, 0.f),
    C_Vertex3f(0.401647449f, 0.0867888927f, 0.f),
    C_Vertex3f(0.387233019f, 0.077064991f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.405607224f, 0.00865411758f, 0.f),
    C_Vertex3f(0.387233019f, 0.077064991f, 0.f),
    C_Vertex3f(0.381977081f, -0.00104403496f, 0.f),
    C_Vertex3f(0.372183561f, 0.0692858696f, 0.f),
    C_Vertex3f(0.356571198f, -0.00750946999f, 0.f),
    C_Vertex3f(0.356499076f, 0.0634515285f, 0.f),
    C_Vertex3f(0.340179563f, 0.0595619678f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329389572f, -0.0107421875f, 0.f),
    C_Vertex3f(0.356571198f, -0.00750946999f, 0.f),
    C_Vertex3f(0.340179563f, 0.0595619678f, 0.f),
    C_Vertex3f(0.323225021f, 0.0576171875f, 0.f),
    C_Vertex3f(0.305689812f, 0.0576171875f, 0.f),
    C_Vertex3f(0.298464775f, -0.0107421875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.305689812f, 0.0576171875f, 0.f),
    C_Vertex3f(0.28881669f, 0.059568882f, 0.f),
    C_Vertex3f(0.298464775f, -0.0107421875f, 0.f),
    C_Vertex3f(0.272551298f, 0.063472271f, 0.f),
    C_Vertex3f(0.267566919f, -0.00757336617f, 0.f),
    C_Vertex3f(0.256893635f, 0.0693273544f, 0.f),
    C_Vertex3f(0.2418437f, 0.0771341324f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.238663435f, -0.0012357235f, 0.f),
    C_Vertex3f(0.267566919f, -0.00757336617f, 0.f),
    C_Vertex3f(0.2418437f, 0.0771341324f, 0.f),
    C_Vertex3f(0.227401495f, 0.0868926048f, 0.f),
    C_Vertex3f(0.213567019f, 0.0986027718f, 0.f),
    C_Vertex3f(0.211754322f, 0.00827074051f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213567019f, 0.0986027718f, 0.f),
    C_Vertex3f(0.200340271f, 0.112264633f, 0.f),
    C_Vertex3f(0.211754322f, 0.00827074051f, 0.f),
    C_Vertex3f(0.187721252f, 0.127878189f, 0.f),
    C_Vertex3f(0.186839581f, 0.0209460258f, 0.f),
    C_Vertex3f(0.176679611f, 0.145675182f, 0.f),
    C_Vertex3f(0.167215347f, 0.165655613f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.16391921f, 0.0367901325f, 0.f),
    C_Vertex3f(0.186839581f, 0.0209460258f, 0.f),
    C_Vertex3f(0.167215347f, 0.165655613f, 0.f),
    C_Vertex3f(0.159328461f, 0.187819481f, 0.f),
    C_Vertex3f(0.153018951f, 0.212166786f, 0.f),
    C_Vertex3f(0.148286819f, 0.238697529f, 0.f),
    C_Vertex3f(0.145132065f, 0.267411709f, 0.f),
    C_Vertex3f(0.142993212f, 0.0558030605f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.568847656f, 0.369628906f, 0.f),
    C_Vertex3f(0.830566406f, 0.293945313f, 0.f),
    C_Vertex3f(0.830566406f, 0.369628906f, 0.f),
    C_Vertex3f(0.568847656f, 0.0756835938f, 0.f),
    C_Vertex3f(0.866699219f, 0.f, 0.f),
    C_Vertex3f(0.866699219f, 0.0756835938f, 0.f),
    C_Vertex3f(0.0580098629f, 0.270105362f, 0.f),
    C_Vertex3f(0.0561523438f, 0.339025497f, 0.f),
    C_Vertex3f(0.0561523438f, 0.303787231f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.489746094f, 0.0927047729f, -1.f),
    C_Vertex3f(0.489746094f, 0.f, -1.f),
    C_Vertex3f(0.866699219f, 0.f, -1.f),
    C_Vertex3f(0.568847656f, 0.0756835938f, -1.f),
    C_Vertex3f(0.489746094f, 0.554039001f, -1.f),
    C_Vertex3f(0.484863281f, 0.342479706f, -1.f),
    C_Vertex3f(0.484863281f, 0.307813644f, -1.f),
    C_Vertex3f(0.483299613f, 0.274436235f, -1.f),
    C_Vertex3f(0.482370377f, 0.0797739029f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.475399017f, 0.429237366f, -1.f),
    C_Vertex3f(0.480131149f, 0.402563095f, -1.f),
    C_Vertex3f(0.482316017f, 0.56626606f, -1.f),
    C_Vertex3f(0.483285904f, 0.373643875f, -1.f),
    C_Vertex3f(0.489746094f, 0.554039001f, -1.f),
    C_Vertex3f(0.484863281f, 0.342479706f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.451738358f, 0.495790482f, -1.f),
    C_Vertex3f(0.461202621f, 0.475851059f, -1.f),
    C_Vertex3f(0.465707302f, 0.587663412f, -1.f),
    C_Vertex3f(0.469089508f, 0.453666687f, -1.f),
    C_Vertex3f(0.482316017f, 0.56626606f, -1.f),
    C_Vertex3f(0.475399017f, 0.429237366f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.447350025f, 0.606004f, -1.f),
    C_Vertex3f(0.427244186f, 0.621287823f, -1.f),
    C_Vertex3f(0.428077698f, 0.528934479f, -1.f),
    C_Vertex3f(0.440696716f, 0.513484955f, -1.f),
    C_Vertex3f(0.451738358f, 0.495790482f, -1.f),
    C_Vertex3f(0.465707302f, 0.587663412f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.386657238f, 0.563695908f, -1.f),
    C_Vertex3f(0.401057959f, 0.554039955f, -1.f),
    C_Vertex3f(0.405389786f, 0.633514881f, -1.f),
    C_Vertex3f(0.414864779f, 0.542452812f, -1.f),
    C_Vertex3f(0.427244186f, 0.621287823f, -1.f),
    C_Vertex3f(0.428077698f, 0.528934479f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339891672f, 0.581076622f, -1.f),
    C_Vertex3f(0.356074095f, 0.577214241f, -1.f),
    C_Vertex3f(0.356435299f, 0.648798704f, -1.f),
    C_Vertex3f(0.371662617f, 0.57142067f, -1.f),
    C_Vertex3f(0.381786823f, 0.642685175f, -1.f),
    C_Vertex3f(0.386657238f, 0.563695908f, -1.f),
    C_Vertex3f(0.405389786f, 0.633514881f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329335213f, 0.651855469f, -1.f),
    C_Vertex3f(0.298082352f, 0.651855469f, -1.f),
    C_Vertex3f(0.303394318f, 0.583007813f, -1.f),
    C_Vertex3f(0.323115349f, 0.583007813f, -1.f),
    C_Vertex3f(0.339891672f, 0.581076622f, -1.f),
    C_Vertex3f(0.356435299f, 0.648798704f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.23882103f, 0.56171608f, -1.f),
    C_Vertex3f(0.260071754f, 0.572361946f, -1.f),
    C_Vertex3f(0.266692877f, 0.648891568f, -1.f),
    C_Vertex3f(0.281596184f, 0.57945919f, -1.f),
    C_Vertex3f(0.298082352f, 0.651855469f, -1.f),
    C_Vertex3f(0.303394318f, 0.583007813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237571001f, 0.642963767f, -1.f),
    C_Vertex3f(0.210716724f, 0.634072065f, -1.f),
    C_Vertex3f(0.213241696f, 0.543051839f, -1.f),
    C_Vertex3f(0.223088264f, 0.551070213f, -1.f),
    C_Vertex3f(0.23882103f, 0.56171608f, -1.f),
    C_Vertex3f(0.266692877f, 0.648891568f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.210716724f, 0.634072065f, -1.f),
    C_Vertex3f(0.186130047f, 0.622216463f, -1.f),
    C_Vertex3f(0.187553644f, 0.513469934f, -1.f),
    C_Vertex3f(0.195474386f, 0.524251699f, -1.f),
    C_Vertex3f(0.20403707f, 0.534112334f, -1.f),
    C_Vertex3f(0.213241696f, 0.543051839f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.186130047f, 0.622216463f, -1.f),
    C_Vertex3f(0.163810968f, 0.60739696f, -1.f),
    C_Vertex3f(0.16764307f, 0.475597858f, -1.f),
    C_Vertex3f(0.173637986f, 0.489143014f, -1.f),
    C_Vertex3f(0.180274844f, 0.501767039f, -1.f),
    C_Vertex3f(0.187553644f, 0.513469934f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.163810968f, 0.60739696f, -1.f),
    C_Vertex3f(0.143759489f, 0.589613557f, -1.f),
    C_Vertex3f(0.145562053f, 0.371088624f, -1.f),
    C_Vertex3f(0.147569418f, 0.39145875f, -1.f),
    C_Vertex3f(0.150245905f, 0.410648108f, -1.f),
    C_Vertex3f(0.153591514f, 0.428656697f, -1.f),
    C_Vertex3f(0.157606244f, 0.445484519f, -1.f),
    C_Vertex3f(0.162290096f, 0.461131573f, -1.f),
    C_Vertex3f(0.16764307f, 0.475597858f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145132065f, 0.267411709f, -1.f),
    C_Vertex3f(0.143554688f, 0.298309326f, -1.f),
    C_Vertex3f(0.142993212f, 0.0558030605f, -1.f),
    C_Vertex3f(0.143554688f, 0.326806068f, -1.f),
    C_Vertex3f(0.143759489f, 0.589613557f, -1.f),
    C_Vertex3f(0.144223809f, 0.34953773f, -1.f),
    C_Vertex3f(0.145562053f, 0.371088624f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0561523438f, 0.339025497f, -1.f),
    C_Vertex3f(0.0580098629f, 0.270105362f, -1.f),
    C_Vertex3f(0.0580918789f, 0.372774482f, -1.f),
    C_Vertex3f(0.0617249012f, 0.238035202f, -1.f),
    C_Vertex3f(0.0619709492f, 0.405089498f, -1.f),
    C_Vertex3f(0.0672974586f, 0.207576752f, -1.f),
    C_Vertex3f(0.0677895546f, 0.435970545f, -1.f),
    C_Vertex3f(0.0747275352f, 0.178730011f, -1.f),
    C_Vertex3f(0.0755476952f, 0.465417624f, -1.f),
    C_Vertex3f(0.084015131f, 0.15149498f, -1.f),
    C_Vertex3f(0.0852453709f, 0.493430734f, -1.f),
    C_Vertex3f(0.0951602459f, 0.125871658f, -1.f),
    C_Vertex3f(0.0968825817f, 0.520009875f, -1.f),
    C_Vertex3f(0.10816288f, 0.101860046f, -1.f),
    C_Vertex3f(0.110459328f, 0.545155048f, -1.f),
    C_Vertex3f(0.124061584f, 0.0779848099f, -1.f),
    C_Vertex3f(0.125975609f, 0.568866253f, -1.f),
    C_Vertex3f(0.142993212f, 0.0558030605f, -1.f),
    C_Vertex3f(0.143759489f, 0.589613557f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.830566406f, 0.293945313f, -1.f),
    C_Vertex3f(0.568847656f, 0.369628906f, -1.f),
    C_Vertex3f(0.568847656f, 0.293945313f, -1.f),
    C_Vertex3f(0.568847656f, 0.564941406f, -1.f),
    C_Vertex3f(0.568847656f, 0.0756835938f, -1.f),
    C_Vertex3f(0.489746094f, 0.640625f, -1.f),
    C_Vertex3f(0.489746094f, 0.554039001f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.853027344f, 0.640625f, -1.f),
    C_Vertex3f(0.489746094f, 0.640625f, -1.f),
    C_Vertex3f(0.568847656f, 0.564941406f, -1.f),
    C_Vertex3f(0.853027344f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.482370377f, 0.0797739029f, -1.f),
    C_Vertex3f(0.483299613f, 0.274436235f, -1.f),
    C_Vertex3f(0.480172276f, 0.243604422f, -1.f),
    C_Vertex3f(0.475481272f, 0.215318203f, -1.f),
    C_Vertex3f(0.469226599f, 0.189577579f, -1.f),
    C_Vertex3f(0.465843201f, 0.0571448803f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.469226599f, 0.189577579f, -1.f),
    C_Vertex3f(0.461408257f, 0.166382551f, -1.f),
    C_Vertex3f(0.465843201f, 0.0571448803f, -1.f),
    C_Vertex3f(0.452026248f, 0.145733118f, -1.f),
    C_Vertex3f(0.447540283f, 0.0377485752f, -1.f),
    C_Vertex3f(0.44108057f, 0.12762928f, -1.f),
    C_Vertex3f(0.428571224f, 0.112071037f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.447540283f, 0.0377485752f, -1.f),
    C_Vertex3f(0.428571224f, 0.112071037f, -1.f),
    C_Vertex3f(0.427461624f, 0.0215849876f, -1.f),
    C_Vertex3f(0.41542685f, 0.0984575748f, -1.f),
    C_Vertex3f(0.405607224f, 0.00865411758f, -1.f),
    C_Vertex3f(0.401647449f, 0.0867888927f, -1.f),
    C_Vertex3f(0.387233019f, 0.077064991f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.405607224f, 0.00865411758f, -1.f),
    C_Vertex3f(0.387233019f, 0.077064991f, -1.f),
    C_Vertex3f(0.381977081f, -0.00104403496f, -1.f),
    C_Vertex3f(0.372183561f, 0.0692858696f, -1.f),
    C_Vertex3f(0.356571198f, -0.00750946999f, -1.f),
    C_Vertex3f(0.356499076f, 0.0634515285f, -1.f),
    C_Vertex3f(0.340179563f, 0.0595619678f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329389572f, -0.0107421875f, -1.f),
    C_Vertex3f(0.356571198f, -0.00750946999f, -1.f),
    C_Vertex3f(0.340179563f, 0.0595619678f, -1.f),
    C_Vertex3f(0.323225021f, 0.0576171875f, -1.f),
    C_Vertex3f(0.305689812f, 0.0576171875f, -1.f),
    C_Vertex3f(0.298464775f, -0.0107421875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.305689812f, 0.0576171875f, -1.f),
    C_Vertex3f(0.28881669f, 0.059568882f, -1.f),
    C_Vertex3f(0.298464775f, -0.0107421875f, -1.f),
    C_Vertex3f(0.272551298f, 0.063472271f, -1.f),
    C_Vertex3f(0.267566919f, -0.00757336617f, -1.f),
    C_Vertex3f(0.256893635f, 0.0693273544f, -1.f),
    C_Vertex3f(0.2418437f, 0.0771341324f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.238663435f, -0.0012357235f, -1.f),
    C_Vertex3f(0.267566919f, -0.00757336617f, -1.f),
    C_Vertex3f(0.2418437f, 0.0771341324f, -1.f),
    C_Vertex3f(0.227401495f, 0.0868926048f, -1.f),
    C_Vertex3f(0.213567019f, 0.0986027718f, -1.f),
    C_Vertex3f(0.211754322f, 0.00827074051f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213567019f, 0.0986027718f, -1.f),
    C_Vertex3f(0.200340271f, 0.112264633f, -1.f),
    C_Vertex3f(0.211754322f, 0.00827074051f, -1.f),
    C_Vertex3f(0.187721252f, 0.127878189f, -1.f),
    C_Vertex3f(0.186839581f, 0.0209460258f, -1.f),
    C_Vertex3f(0.176679611f, 0.145675182f, -1.f),
    C_Vertex3f(0.167215347f, 0.165655613f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.16391921f, 0.0367901325f, -1.f),
    C_Vertex3f(0.186839581f, 0.0209460258f, -1.f),
    C_Vertex3f(0.167215347f, 0.165655613f, -1.f),
    C_Vertex3f(0.159328461f, 0.187819481f, -1.f),
    C_Vertex3f(0.153018951f, 0.212166786f, -1.f),
    C_Vertex3f(0.148286819f, 0.238697529f, -1.f),
    C_Vertex3f(0.145132065f, 0.267411709f, -1.f),
    C_Vertex3f(0.142993212f, 0.0558030605f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.568847656f, 0.369628906f, -1.f),
    C_Vertex3f(0.830566406f, 0.293945313f, -1.f),
    C_Vertex3f(0.830566406f, 0.369628906f, -1.f),
    C_Vertex3f(0.568847656f, 0.0756835938f, -1.f),
    C_Vertex3f(0.866699219f, 0.f, -1.f),
    C_Vertex3f(0.866699219f, 0.0756835938f, -1.f),
    C_Vertex3f(0.0580098629f, 0.270105362f, -1.f),
    C_Vertex3f(0.0561523438f, 0.339025497f, -1.f),
    C_Vertex3f(0.0561523438f, 0.303787231f, -1.f),
    C_End(),
    C_Translatef(0.895019531f, 0.f, 0.f),
};
static const Call char_141[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_142[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.017578125f, 0.f, 0.f),
    C_Vertex3f(0.017578125f, 0.f, -1.f),
    C_Vertex3f(0.017578125f, 0.078742981f, -1.f),
    C_Vertex3f(0.017578125f, 0.078742981f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.017578125f, 0.078742981f, 0.f),
    C_Vertex3f(0.017578125f, 0.078742981f, -1.f),
    C_Normal3f(-0.77963239f, 0.626237333f, 0.f),
    C_Vertex3f(0.354995728f, 0.499845505f, 0.f),
    C_Vertex3f(0.354995728f, 0.499845505f, -1.f),
    C_Normal3f(-0.757743835f, 0.652552128f, 0.f),
    C_Vertex3f(0.372172356f, 0.520194054f, 0.f),
    C_Vertex3f(0.372172356f, 0.520194054f, -1.f),
    C_Normal3f(-0.743628263f, 0.668593347f, 0.f),
    C_Vertex3f(0.388871193f, 0.539181709f, 0.f),
    C_Vertex3f(0.388871193f, 0.539181709f, -1.f),
    C_Normal3f(-0.73034364f, 0.683079898f, 0.f),
    C_Vertex3f(0.405092239f, 0.556808472f, 0.f),
    C_Vertex3f(0.405092239f, 0.556808472f, -1.f),
    C_Normal3f(-0.718555987f, 0.695469201f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, -1.f),
    C_Vertex3f(0.0546875f, 0.564941406f, -1.f),
    C_Vertex3f(0.0546875f, 0.564941406f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0546875f, 0.564941406f, 0.f),
    C_Vertex3f(0.0546875f, 0.564941406f, -1.f),
    C_Vertex3f(0.0546875f, 0.640625f, -1.f),
    C_Vertex3f(0.0546875f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0546875f, 0.640625f, 0.f),
    C_Vertex3f(0.0546875f, 0.640625f, -1.f),
    C_Vertex3f(0.514160156f, 0.640625f, -1.f),
    C_Vertex3f(0.514160156f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.514160156f, 0.640625f, 0.f),
    C_Vertex3f(0.514160156f, 0.640625f, -1.f),
    C_Vertex3f(0.514160156f, 0.564941406f, -1.f),
    C_Vertex3f(0.514160156f, 0.564941406f, 0.f),
    C_Normal3f(0.776490152f, -0.630129337f, 0.f),
    C_Vertex3f(0.514160156f, 0.564941406f, 0.f),
    C_Vertex3f(0.514160156f, 0.564941406f, -1.f),
    C_Vertex3f(0.153862f, 0.120956421f, -1.f),
    C_Vertex3f(0.153862f, 0.120956421f, 0.f),
    C_Normal3f(0.759085f, -0.650991499f, 0.f),
    C_Vertex3f(0.153862f, 0.120956421f, 0.f),
    C_Vertex3f(0.153862f, 0.120956421f, -1.f),
    C_Vertex3f(0.115036011f, 0.0756835938f, -1.f),
    C_Vertex3f(0.115036011f, 0.0756835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.115036011f, 0.0756835938f, 0.f),
    C_Vertex3f(0.115036011f, 0.0756835938f, -1.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, -1.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, 0.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, -1.f),
    C_Vertex3f(0.524414063f, 0.f, -1.f),
    C_Vertex3f(0.524414063f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.524414063f, 0.f, 0.f),
    C_Vertex3f(0.524414063f, 0.f, -1.f),
    C_Vertex3f(0.017578125f, 0.f, -1.f),
    C_Vertex3f(0.017578125f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.840821147f, 0.541312993f, 0.f),
    C_Vertex3f(0.276855469f, 0.728027344f, 0.f),
    C_Vertex3f(0.276855469f, 0.728027344f, -1.f),
    C_Vertex3f(0.324951172f, 0.802734375f, -1.f),
    C_Vertex3f(0.324951172f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.324951172f, 0.802734375f, 0.f),
    C_Vertex3f(0.324951172f, 0.802734375f, -1.f),
    C_Vertex3f(0.414306641f, 0.802734375f, -1.f),
    C_Vertex3f(0.414306641f, 0.802734375f, 0.f),
    C_Normal3f(0.779806852f, -0.626020253f, 0.f),
    C_Vertex3f(0.414306641f, 0.802734375f, 0.f),
    C_Vertex3f(0.414306641f, 0.802734375f, -1.f),
    C_Vertex3f(0.315917969f, 0.680175781f, -1.f),
    C_Vertex3f(0.315917969f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.315917969f, 0.680175781f, 0.f),
    C_Vertex3f(0.315917969f, 0.680175781f, -1.f),
    C_Vertex3f(0.231933594f, 0.680175781f, -1.f),
    C_Vertex3f(0.231933594f, 0.680175781f, 0.f),
    C_Normal3f(-0.795031667f, -0.606567979f, 0.f),
    C_Vertex3f(0.231933594f, 0.680175781f, 0.f),
    C_Vertex3f(0.231933594f, 0.680175781f, -1.f),
    C_Vertex3f(0.138427734f, 0.802734375f, -1.f),
    C_Vertex3f(0.138427734f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.138427734f, 0.802734375f, 0.f),
    C_Vertex3f(0.138427734f, 0.802734375f, -1.f),
    C_Vertex3f(0.227783203f, 0.802734375f, -1.f),
    C_Vertex3f(0.227783203f, 0.802734375f, 0.f),
    C_Normal3f(0.835812926f, 0.54901439f, 0.f),
    C_Vertex3f(0.227783203f, 0.802734375f, 0.f),
    C_Vertex3f(0.227783203f, 0.802734375f, -1.f),
    C_Vertex3f(0.276855469f, 0.728027344f, -1.f),
    C_Vertex3f(0.276855469f, 0.728027344f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115036011f, 0.0756835938f, 0.f),
    C_Vertex3f(0.153862f, 0.120956421f, 0.f),
    C_Vertex3f(0.354995728f, 0.499845505f, 0.f),
    C_Vertex3f(0.017578125f, 0.078742981f, 0.f),
    C_Vertex3f(0.524414063f, 0.f, 0.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.514160156f, 0.564941406f, 0.f),
    C_Vertex3f(0.514160156f, 0.640625f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, 0.f),
    C_Vertex3f(0.405092239f, 0.556808472f, 0.f),
    C_Vertex3f(0.388871193f, 0.539181709f, 0.f),
    C_Vertex3f(0.372172356f, 0.520194054f, 0.f),
    C_Vertex3f(0.354995728f, 0.499845505f, 0.f),
    C_Vertex3f(0.153862f, 0.120956421f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0546875f, 0.640625f, 0.f),
    C_Vertex3f(0.0546875f, 0.564941406f, 0.f),
    C_Vertex3f(0.412963867f, 0.564941406f, 0.f),
    C_Vertex3f(0.514160156f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.414306641f, 0.802734375f, 0.f),
    C_Vertex3f(0.324951172f, 0.802734375f, 0.f),
    C_Vertex3f(0.315917969f, 0.680175781f, 0.f),
    C_Vertex3f(0.276855469f, 0.728027344f, 0.f),
    C_Vertex3f(0.231933594f, 0.680175781f, 0.f),
    C_Vertex3f(0.227783203f, 0.802734375f, 0.f),
    C_Vertex3f(0.138427734f, 0.802734375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.524414063f, 0.f, 0.f),
    C_Vertex3f(0.017578125f, 0.078742981f, 0.f),
    C_Vertex3f(0.017578125f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115036011f, 0.0756835938f, -1.f),
    C_Vertex3f(0.153862f, 0.120956421f, -1.f),
    C_Vertex3f(0.354995728f, 0.499845505f, -1.f),
    C_Vertex3f(0.017578125f, 0.078742981f, -1.f),
    C_Vertex3f(0.524414063f, 0.f, -1.f),
    C_Vertex3f(0.524414063f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.514160156f, 0.564941406f, -1.f),
    C_Vertex3f(0.514160156f, 0.640625f, -1.f),
    C_Vertex3f(0.412963867f, 0.564941406f, -1.f),
    C_Vertex3f(0.405092239f, 0.556808472f, -1.f),
    C_Vertex3f(0.388871193f, 0.539181709f, -1.f),
    C_Vertex3f(0.372172356f, 0.520194054f, -1.f),
    C_Vertex3f(0.354995728f, 0.499845505f, -1.f),
    C_Vertex3f(0.153862f, 0.120956421f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0546875f, 0.640625f, -1.f),
    C_Vertex3f(0.0546875f, 0.564941406f, -1.f),
    C_Vertex3f(0.412963867f, 0.564941406f, -1.f),
    C_Vertex3f(0.514160156f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.414306641f, 0.802734375f, -1.f),
    C_Vertex3f(0.324951172f, 0.802734375f, -1.f),
    C_Vertex3f(0.315917969f, 0.680175781f, -1.f),
    C_Vertex3f(0.276855469f, 0.728027344f, -1.f),
    C_Vertex3f(0.231933594f, 0.680175781f, -1.f),
    C_Vertex3f(0.227783203f, 0.802734375f, -1.f),
    C_Vertex3f(0.138427734f, 0.802734375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.524414063f, 0.f, -1.f),
    C_Vertex3f(0.017578125f, 0.078742981f, -1.f),
    C_Vertex3f(0.017578125f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.546875f, 0.f, 0.f),
};
static const Call char_143[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_144[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_145[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.140625f, 0.532714844f, 0.f),
    C_Vertex3f(0.140625f, 0.532714844f, -1.f),
    C_Vertex3f(0.140625f, 0.44140625f, -1.f),
    C_Vertex3f(0.140625f, 0.44140625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.140625f, 0.44140625f, 0.f),
    C_Vertex3f(0.140625f, 0.44140625f, -1.f),
    C_Vertex3f(0.0561523438f, 0.44140625f, -1.f),
    C_Vertex3f(0.0561523438f, 0.44140625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0561523438f, 0.44140625f, 0.f),
    C_Vertex3f(0.0561523438f, 0.44140625f, -1.f),
    C_Normal3f(-0.999989033f, 0.00469817035f, 0.f),
    C_Vertex3f(0.0561523438f, 0.520994186f, 0.f),
    C_Vertex3f(0.0561523438f, 0.520994186f, -1.f),
    C_Normal3f(-0.99875015f, 0.0499805883f, 0.f),
    C_Vertex3f(0.0565903187f, 0.534627676f, 0.f),
    C_Vertex3f(0.0565903187f, 0.534627676f, -1.f),
    C_Normal3f(-0.995941818f, 0.0899985805f, 0.f),
    C_Vertex3f(0.0574662685f, 0.547250032f, 0.f),
    C_Vertex3f(0.0574662685f, 0.547250032f, -1.f),
    C_Normal3f(-0.990607917f, 0.136733472f, 0.f),
    C_Vertex3f(0.0587801933f, 0.558861256f, 0.f),
    C_Vertex3f(0.0587801933f, 0.558861256f, -1.f),
    C_Normal3f(-0.981468201f, 0.191625014f, 0.f),
    C_Vertex3f(0.060532093f, 0.569461346f, 0.f),
    C_Vertex3f(0.060532093f, 0.569461346f, -1.f),
    C_Normal3f(-0.966588378f, 0.256333619f, 0.f),
    C_Vertex3f(0.0627219677f, 0.579050303f, 0.f),
    C_Vertex3f(0.0627219677f, 0.579050303f, -1.f),
    C_Normal3f(-0.943070769f, 0.3325921f, 0.f),
    C_Vertex3f(0.0653498173f, 0.587628126f, 0.f),
    C_Vertex3f(0.0653498173f, 0.587628126f, -1.f),
    C_Normal3f(-0.901267648f, 0.433262706f, 0.f),
    C_Vertex3f(0.0684156418f, 0.595194817f, 0.f),
    C_Vertex3f(0.0684156418f, 0.595194817f, -1.f),
    C_Normal3f(-0.832335711f, 0.554271877f, 0.f),
    C_Vertex3f(0.0747661591f, 0.607215881f, 0.f),
    C_Vertex3f(0.0747661591f, 0.607215881f, -1.f),
    C_Normal3f(-0.734007537f, 0.679141283f, 0.f),
    C_Vertex3f(0.08664608f, 0.622570992f, 0.f),
    C_Vertex3f(0.08664608f, 0.622570992f, -1.f),
    C_Normal3f(-0.607326269f, 0.794452429f, 0.f),
    C_Vertex3f(0.101208687f, 0.63579464f, 0.f),
    C_Vertex3f(0.101208687f, 0.63579464f, -1.f),
    C_Normal3f(-0.496725231f, 0.867907882f, 0.f),
    C_Vertex3f(0.118453979f, 0.646886826f, 0.f),
    C_Vertex3f(0.118453979f, 0.646886826f, -1.f),
    C_Normal3f(-0.410103351f, 0.912039101f, 0.f),
    C_Vertex3f(0.128417969f, 0.651367188f, 0.f),
    C_Vertex3f(0.128417969f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.848972201f, 0.528437555f, 0.f),
    C_Vertex3f(0.128417969f, 0.651367188f, 0.f),
    C_Vertex3f(0.128417969f, 0.651367188f, -1.f),
    C_Vertex3f(0.147460938f, 0.620773315f, -1.f),
    C_Vertex3f(0.147460938f, 0.620773315f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.387362093f, -0.921927631f, 0.f),
    C_Vertex3f(0.147460938f, 0.620773315f, 0.f),
    C_Vertex3f(0.147460938f, 0.620773315f, -1.f),
    C_Normal3f(0.472210735f, -0.881485641f, 0.f),
    C_Vertex3f(0.141464233f, 0.618253708f, 0.f),
    C_Vertex3f(0.141464233f, 0.618253708f, -1.f),
    C_Normal3f(0.582000375f, -0.813188553f, 0.f),
    C_Vertex3f(0.13105154f, 0.611982822f, 0.f),
    C_Vertex3f(0.13105154f, 0.611982822f, -1.f),
    C_Normal3f(0.710457981f, -0.703739643f, 0.f),
    C_Vertex3f(0.122219563f, 0.604480267f, 0.f),
    C_Vertex3f(0.122219563f, 0.604480267f, -1.f),
    C_Normal3f(0.822694838f, -0.568483233f, 0.f),
    C_Vertex3f(0.1149683f, 0.59574604f, 0.f),
    C_Vertex3f(0.1149683f, 0.59574604f, -1.f),
    C_Normal3f(0.910453439f, -0.413611412f, 0.f),
    C_Vertex3f(0.109297752f, 0.585780144f, 0.f),
    C_Vertex3f(0.109297752f, 0.585780144f, -1.f),
    C_Normal3f(0.962118983f, -0.272629827f, 0.f),
    C_Vertex3f(0.10488081f, 0.573541164f, 0.f),
    C_Vertex3f(0.10488081f, 0.573541164f, -1.f),
    C_Normal3f(0.987116218f, -0.160005018f, 0.f),
    C_Vertex3f(0.101717472f, 0.559029102f, 0.f),
    C_Vertex3f(0.101717472f, 0.559029102f, -1.f),
    C_Normal3f(0.996403515f, -0.0847353712f, 0.f),
    C_Vertex3f(0.0998077393f, 0.542243958f, 0.f),
    C_Vertex3f(0.0998077393f, 0.542243958f, -1.f),
    C_Normal3f(0.999407887f, -0.0344071575f, 0.f),
    C_Vertex3f(0.0994796753f, 0.532714844f, 0.f),
    C_Vertex3f(0.0994796753f, 0.532714844f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0994796753f, 0.532714844f, 0.f),
    C_Vertex3f(0.0994796753f, 0.532714844f, -1.f),
    C_Vertex3f(0.140625f, 0.532714844f, -1.f),
    C_Vertex3f(0.140625f, 0.532714844f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.140625f, 0.44140625f, 0.f),
    C_Vertex3f(0.140625f, 0.532714844f, 0.f),
    C_Vertex3f(0.0994796753f, 0.532714844f, 0.f),
    C_Vertex3f(0.0684156418f, 0.595194817f, 0.f),
    C_Vertex3f(0.0653498173f, 0.587628126f, 0.f),
    C_Vertex3f(0.0627219677f, 0.579050303f, 0.f),
    C_Vertex3f(0.060532093f, 0.569461346f, 0.f),
    C_Vertex3f(0.0587801933f, 0.558861256f, 0.f),
    C_Vertex3f(0.0574662685f, 0.547250032f, 0.f),
    C_Vertex3f(0.0565903187f, 0.534627676f, 0.f),
    C_Vertex3f(0.0561523438f, 0.520994186f, 0.f),
    C_Vertex3f(0.0561523438f, 0.44140625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0994796753f, 0.532714844f, 0.f),
    C_Vertex3f(0.0998077393f, 0.542243958f, 0.f),
    C_Vertex3f(0.101208687f, 0.63579464f, 0.f),
    C_Vertex3f(0.08664608f, 0.622570992f, 0.f),
    C_Vertex3f(0.0747661591f, 0.607215881f, 0.f),
    C_Vertex3f(0.0684156418f, 0.595194817f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.128417969f, 0.651367188f, 0.f),
    C_Vertex3f(0.118453979f, 0.646886826f, 0.f),
    C_Vertex3f(0.122219563f, 0.604480267f, 0.f),
    C_Vertex3f(0.13105154f, 0.611982822f, 0.f),
    C_Vertex3f(0.141464233f, 0.618253708f, 0.f),
    C_Vertex3f(0.147460938f, 0.620773315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118453979f, 0.646886826f, 0.f),
    C_Vertex3f(0.101208687f, 0.63579464f, 0.f),
    C_Vertex3f(0.101717472f, 0.559029102f, 0.f),
    C_Vertex3f(0.10488081f, 0.573541164f, 0.f),
    C_Vertex3f(0.109297752f, 0.585780144f, 0.f),
    C_Vertex3f(0.1149683f, 0.59574604f, 0.f),
    C_Vertex3f(0.122219563f, 0.604480267f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.101208687f, 0.63579464f, 0.f),
    C_Vertex3f(0.0998077393f, 0.542243958f, 0.f),
    C_Vertex3f(0.101717472f, 0.559029102f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.140625f, 0.44140625f, -1.f),
    C_Vertex3f(0.140625f, 0.532714844f, -1.f),
    C_Vertex3f(0.0994796753f, 0.532714844f, -1.f),
    C_Vertex3f(0.0684156418f, 0.595194817f, -1.f),
    C_Vertex3f(0.0653498173f, 0.587628126f, -1.f),
    C_Vertex3f(0.0627219677f, 0.579050303f, -1.f),
    C_Vertex3f(0.060532093f, 0.569461346f, -1.f),
    C_Vertex3f(0.0587801933f, 0.558861256f, -1.f),
    C_Vertex3f(0.0574662685f, 0.547250032f, -1.f),
    C_Vertex3f(0.0565903187f, 0.534627676f, -1.f),
    C_Vertex3f(0.0561523438f, 0.520994186f, -1.f),
    C_Vertex3f(0.0561523438f, 0.44140625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0994796753f, 0.532714844f, -1.f),
    C_Vertex3f(0.0998077393f, 0.542243958f, -1.f),
    C_Vertex3f(0.101208687f, 0.63579464f, -1.f),
    C_Vertex3f(0.08664608f, 0.622570992f, -1.f),
    C_Vertex3f(0.0747661591f, 0.607215881f, -1.f),
    C_Vertex3f(0.0684156418f, 0.595194817f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.128417969f, 0.651367188f, -1.f),
    C_Vertex3f(0.118453979f, 0.646886826f, -1.f),
    C_Vertex3f(0.122219563f, 0.604480267f, -1.f),
    C_Vertex3f(0.13105154f, 0.611982822f, -1.f),
    C_Vertex3f(0.141464233f, 0.618253708f, -1.f),
    C_Vertex3f(0.147460938f, 0.620773315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118453979f, 0.646886826f, -1.f),
    C_Vertex3f(0.101208687f, 0.63579464f, -1.f),
    C_Vertex3f(0.101717472f, 0.559029102f, -1.f),
    C_Vertex3f(0.10488081f, 0.573541164f, -1.f),
    C_Vertex3f(0.109297752f, 0.585780144f, -1.f),
    C_Vertex3f(0.1149683f, 0.59574604f, -1.f),
    C_Vertex3f(0.122219563f, 0.604480267f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.101208687f, 0.63579464f, -1.f),
    C_Vertex3f(0.0998077393f, 0.542243958f, -1.f),
    C_Vertex3f(0.101717472f, 0.559029102f, -1.f),
    C_End(),
    C_Translatef(0.198730469f, 0.f, 0.f),
};
static const Call char_146[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0541992188f, 0.556152344f, 0.f),
    C_Vertex3f(0.0541992188f, 0.556152344f, -1.f),
    C_Vertex3f(0.0541992188f, 0.647460938f, -1.f),
    C_Vertex3f(0.0541992188f, 0.647460938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0541992188f, 0.647460938f, 0.f),
    C_Vertex3f(0.0541992188f, 0.647460938f, -1.f),
    C_Vertex3f(0.138671875f, 0.647460938f, -1.f),
    C_Vertex3f(0.138671875f, 0.647460938f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.138671875f, 0.647460938f, 0.f),
    C_Vertex3f(0.138671875f, 0.647460938f, -1.f),
    C_Normal3f(0.999989629f, -0.00455305772f, 0.f),
    C_Vertex3f(0.138671875f, 0.567873001f, 0.f),
    C_Vertex3f(0.138671875f, 0.567873001f, -1.f),
    C_Normal3f(0.998822749f, -0.0485085919f, 0.f),
    C_Vertex3f(0.13824749f, 0.554253101f, 0.f),
    C_Vertex3f(0.13824749f, 0.554253101f, -1.f),
    C_Normal3f(0.996167064f, -0.0874710083f, 0.f),
    C_Vertex3f(0.13739872f, 0.541657925f, 0.f),
    C_Vertex3f(0.13739872f, 0.541657925f, -1.f),
    C_Normal3f(0.99109906f, -0.133126646f, 0.f),
    C_Vertex3f(0.136125565f, 0.530087471f, 0.f),
    C_Vertex3f(0.136125565f, 0.530087471f, -1.f),
    C_Normal3f(0.982363462f, -0.186981231f, 0.f),
    C_Vertex3f(0.134428024f, 0.51954174f, 0.f),
    C_Vertex3f(0.134428024f, 0.51954174f, -1.f),
    C_Normal3f(0.968034923f, -0.250815451f, 0.f),
    C_Vertex3f(0.132306099f, 0.510020733f, 0.f),
    C_Vertex3f(0.132306099f, 0.510020733f, -1.f),
    C_Normal3f(0.945174932f, -0.326564759f, 0.f),
    C_Vertex3f(0.129759789f, 0.501524448f, 0.f),
    C_Vertex3f(0.129759789f, 0.501524448f, -1.f),
    C_Normal3f(0.901648641f, -0.432469338f, 0.f),
    C_Vertex3f(0.126789093f, 0.494052887f, 0.f),
    C_Vertex3f(0.126789093f, 0.494052887f, -1.f),
    C_Normal3f(0.831324697f, -0.555787146f, 0.f),
    C_Vertex3f(0.120384216f, 0.481977463f, 0.f),
    C_Vertex3f(0.120384216f, 0.481977463f, -1.f),
    C_Normal3f(0.733749926f, -0.679419637f, 0.f),
    C_Vertex3f(0.108341217f, 0.466459274f, 0.f),
    C_Vertex3f(0.108341217f, 0.466459274f, -1.f),
    C_Normal3f(0.607928216f, -0.793991923f, 0.f),
    C_Vertex3f(0.0936698914f, 0.453126907f, 0.f),
    C_Vertex3f(0.0936698914f, 0.453126907f, -1.f),
    C_Normal3f(0.497282386f, -0.867588699f, 0.f),
    C_Vertex3f(0.0763702393f, 0.441980362f, 0.f),
    C_Vertex3f(0.0763702393f, 0.441980362f, -1.f),
    C_Normal3f(0.410103351f, -0.912039101f, 0.f),
    C_Vertex3f(0.06640625f, 0.4375f, 0.f),
    C_Vertex3f(0.06640625f, 0.4375f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.852290988f, -0.52306819f, 0.f),
    C_Vertex3f(0.06640625f, 0.4375f, 0.f),
    C_Vertex3f(0.06640625f, 0.4375f, -1.f),
    C_Vertex3f(0.0473632813f, 0.468528748f, -1.f),
    C_Vertex3f(0.0473632813f, 0.468528748f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.378667861f, 0.925532579f, 0.f),
    C_Vertex3f(0.0473632813f, 0.468528748f, 0.f),
    C_Vertex3f(0.0473632813f, 0.468528748f, -1.f),
    C_Normal3f(-0.470709801f, 0.882288098f, 0.f),
    C_Vertex3f(0.0532512665f, 0.470937729f, 0.f),
    C_Vertex3f(0.0532512665f, 0.470937729f, -1.f),
    C_Normal3f(-0.58732599f, 0.80935055f, 0.f),
    C_Vertex3f(0.0635280609f, 0.477152824f, 0.f),
    C_Vertex3f(0.0635280609f, 0.477152824f, -1.f),
    C_Normal3f(-0.719236493f, 0.694765329f, 0.f),
    C_Vertex3f(0.0723056793f, 0.484765053f, 0.f),
    C_Vertex3f(0.0723056793f, 0.484765053f, -1.f),
    C_Normal3f(-0.829798639f, 0.558062911f, 0.f),
    C_Vertex3f(0.0795841217f, 0.493774414f, 0.f),
    C_Vertex3f(0.0795841217f, 0.493774414f, -1.f),
    C_Normal3f(-0.911361814f, 0.411606163f, 0.f),
    C_Vertex3f(0.0853633881f, 0.504180908f, 0.f),
    C_Vertex3f(0.0853633881f, 0.504180908f, -1.f),
    C_Normal3f(-0.96053046f, 0.278174967f, 0.f),
    C_Vertex3f(0.0898618698f, 0.516530991f, 0.f),
    C_Vertex3f(0.0898618698f, 0.516530991f, -1.f),
    C_Normal3f(-0.98604548f, 0.166476414f, 0.f),
    C_Vertex3f(0.093079567f, 0.530824661f, 0.f),
    C_Vertex3f(0.093079567f, 0.530824661f, -1.f),
    C_Normal3f(-0.996025264f, 0.0890714675f, 0.f),
    C_Vertex3f(0.0950164795f, 0.54706192f, 0.f),
    C_Vertex3f(0.0950164795f, 0.54706192f, -1.f),
    C_Normal3f(-0.999349475f, 0.0360654853f, 0.f),
    C_Vertex3f(0.0953445435f, 0.556152344f, 0.f),
    C_Vertex3f(0.0953445435f, 0.556152344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0953445435f, 0.556152344f, 0.f),
    C_Vertex3f(0.0953445435f, 0.556152344f, -1.f),
    C_Vertex3f(0.0541992188f, 0.556152344f, -1.f),
    C_Vertex3f(0.0541992188f, 0.556152344f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138671875f, 0.647460938f, 0.f),
    C_Vertex3f(0.0541992188f, 0.647460938f, 0.f),
    C_Vertex3f(0.0953445435f, 0.556152344f, 0.f),
    C_Vertex3f(0.108341217f, 0.466459274f, 0.f),
    C_Vertex3f(0.120384216f, 0.481977463f, 0.f),
    C_Vertex3f(0.126789093f, 0.494052887f, 0.f),
    C_Vertex3f(0.129759789f, 0.501524448f, 0.f),
    C_Vertex3f(0.132306099f, 0.510020733f, 0.f),
    C_Vertex3f(0.134428024f, 0.51954174f, 0.f),
    C_Vertex3f(0.136125565f, 0.530087471f, 0.f),
    C_Vertex3f(0.13739872f, 0.541657925f, 0.f),
    C_Vertex3f(0.13824749f, 0.554253101f, 0.f),
    C_Vertex3f(0.138671875f, 0.567873001f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0950164795f, 0.54706192f, 0.f),
    C_Vertex3f(0.093079567f, 0.530824661f, 0.f),
    C_Vertex3f(0.0936698914f, 0.453126907f, 0.f),
    C_Vertex3f(0.108341217f, 0.466459274f, 0.f),
    C_Vertex3f(0.0953445435f, 0.556152344f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0936698914f, 0.453126907f, 0.f),
    C_Vertex3f(0.093079567f, 0.530824661f, 0.f),
    C_Vertex3f(0.0898618698f, 0.516530991f, 0.f),
    C_Vertex3f(0.0853633881f, 0.504180908f, 0.f),
    C_Vertex3f(0.0795841217f, 0.493774414f, 0.f),
    C_Vertex3f(0.0763702393f, 0.441980362f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0723056793f, 0.484765053f, 0.f),
    C_Vertex3f(0.0635280609f, 0.477152824f, 0.f),
    C_Vertex3f(0.06640625f, 0.4375f, 0.f),
    C_Vertex3f(0.0763702393f, 0.441980362f, 0.f),
    C_Vertex3f(0.0795841217f, 0.493774414f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0532512665f, 0.470937729f, 0.f),
    C_Vertex3f(0.0473632813f, 0.468528748f, 0.f),
    C_Vertex3f(0.06640625f, 0.4375f, 0.f),
    C_Vertex3f(0.0635280609f, 0.477152824f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0953445435f, 0.556152344f, 0.f),
    C_Vertex3f(0.0541992188f, 0.647460938f, 0.f),
    C_Vertex3f(0.0541992188f, 0.556152344f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138671875f, 0.647460938f, -1.f),
    C_Vertex3f(0.0541992188f, 0.647460938f, -1.f),
    C_Vertex3f(0.0953445435f, 0.556152344f, -1.f),
    C_Vertex3f(0.108341217f, 0.466459274f, -1.f),
    C_Vertex3f(0.120384216f, 0.481977463f, -1.f),
    C_Vertex3f(0.126789093f, 0.494052887f, -1.f),
    C_Vertex3f(0.129759789f, 0.501524448f, -1.f),
    C_Vertex3f(0.132306099f, 0.510020733f, -1.f),
    C_Vertex3f(0.134428024f, 0.51954174f, -1.f),
    C_Vertex3f(0.136125565f, 0.530087471f, -1.f),
    C_Vertex3f(0.13739872f, 0.541657925f, -1.f),
    C_Vertex3f(0.13824749f, 0.554253101f, -1.f),
    C_Vertex3f(0.138671875f, 0.567873001f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0950164795f, 0.54706192f, -1.f),
    C_Vertex3f(0.093079567f, 0.530824661f, -1.f),
    C_Vertex3f(0.0936698914f, 0.453126907f, -1.f),
    C_Vertex3f(0.108341217f, 0.466459274f, -1.f),
    C_Vertex3f(0.0953445435f, 0.556152344f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0936698914f, 0.453126907f, -1.f),
    C_Vertex3f(0.093079567f, 0.530824661f, -1.f),
    C_Vertex3f(0.0898618698f, 0.516530991f, -1.f),
    C_Vertex3f(0.0853633881f, 0.504180908f, -1.f),
    C_Vertex3f(0.0795841217f, 0.493774414f, -1.f),
    C_Vertex3f(0.0763702393f, 0.441980362f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0723056793f, 0.484765053f, -1.f),
    C_Vertex3f(0.0635280609f, 0.477152824f, -1.f),
    C_Vertex3f(0.06640625f, 0.4375f, -1.f),
    C_Vertex3f(0.0763702393f, 0.441980362f, -1.f),
    C_Vertex3f(0.0795841217f, 0.493774414f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0532512665f, 0.470937729f, -1.f),
    C_Vertex3f(0.0473632813f, 0.468528748f, -1.f),
    C_Vertex3f(0.06640625f, 0.4375f, -1.f),
    C_Vertex3f(0.0635280609f, 0.477152824f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0953445435f, 0.556152344f, -1.f),
    C_Vertex3f(0.0541992188f, 0.647460938f, -1.f),
    C_Vertex3f(0.0541992188f, 0.556152344f, -1.f),
    C_End(),
    C_Translatef(0.198730469f, 0.f, 0.f),
};
static const Call char_147[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.120605469f, 0.532714844f, 0.f),
    C_Vertex3f(0.120605469f, 0.532714844f, -1.f),
    C_Vertex3f(0.120605469f, 0.44140625f, -1.f),
    C_Vertex3f(0.120605469f, 0.44140625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.120605469f, 0.44140625f, 0.f),
    C_Vertex3f(0.120605469f, 0.44140625f, -1.f),
    C_Vertex3f(0.0361328125f, 0.44140625f, -1.f),
    C_Vertex3f(0.0361328125f, 0.44140625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0361328125f, 0.44140625f, 0.f),
    C_Vertex3f(0.0361328125f, 0.44140625f, -1.f),
    C_Normal3f(-0.999989033f, 0.00469817035f, 0.f),
    C_Vertex3f(0.0361328125f, 0.520994186f, 0.f),
    C_Vertex3f(0.0361328125f, 0.520994186f, -1.f),
    C_Normal3f(-0.99875015f, 0.0499805883f, 0.f),
    C_Vertex3f(0.0365707874f, 0.534627676f, 0.f),
    C_Vertex3f(0.0365707874f, 0.534627676f, -1.f),
    C_Normal3f(-0.995941818f, 0.0899985805f, 0.f),
    C_Vertex3f(0.0374467373f, 0.547250032f, 0.f),
    C_Vertex3f(0.0374467373f, 0.547250032f, -1.f),
    C_Normal3f(-0.990607917f, 0.136733472f, 0.f),
    C_Vertex3f(0.0387606621f, 0.558861256f, 0.f),
    C_Vertex3f(0.0387606621f, 0.558861256f, -1.f),
    C_Normal3f(-0.981468201f, 0.191625014f, 0.f),
    C_Vertex3f(0.0405125618f, 0.569461346f, 0.f),
    C_Vertex3f(0.0405125618f, 0.569461346f, -1.f),
    C_Normal3f(-0.966588378f, 0.256333619f, 0.f),
    C_Vertex3f(0.0427024364f, 0.579050303f, 0.f),
    C_Vertex3f(0.0427024364f, 0.579050303f, -1.f),
    C_Normal3f(-0.943070769f, 0.3325921f, 0.f),
    C_Vertex3f(0.045330286f, 0.587628126f, 0.f),
    C_Vertex3f(0.045330286f, 0.587628126f, -1.f),
    C_Normal3f(-0.901267648f, 0.433262706f, 0.f),
    C_Vertex3f(0.0483961105f, 0.595194817f, 0.f),
    C_Vertex3f(0.0483961105f, 0.595194817f, -1.f),
    C_Normal3f(-0.832335711f, 0.554271877f, 0.f),
    C_Vertex3f(0.0547466278f, 0.607215881f, 0.f),
    C_Vertex3f(0.0547466278f, 0.607215881f, -1.f),
    C_Normal3f(-0.734007537f, 0.679141283f, 0.f),
    C_Vertex3f(0.0666265488f, 0.622570992f, 0.f),
    C_Vertex3f(0.0666265488f, 0.622570992f, -1.f),
    C_Normal3f(-0.607326269f, 0.794452429f, 0.f),
    C_Vertex3f(0.0811891556f, 0.63579464f, 0.f),
    C_Vertex3f(0.0811891556f, 0.63579464f, -1.f),
    C_Normal3f(-0.496725231f, 0.867907882f, 0.f),
    C_Vertex3f(0.0984344482f, 0.646886826f, 0.f),
    C_Vertex3f(0.0984344482f, 0.646886826f, -1.f),
    C_Normal3f(-0.410103351f, 0.912039101f, 0.f),
    C_Vertex3f(0.108398438f, 0.651367188f, 0.f),
    C_Vertex3f(0.108398438f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.848972201f, 0.528437555f, 0.f),
    C_Vertex3f(0.108398438f, 0.651367188f, 0.f),
    C_Vertex3f(0.108398438f, 0.651367188f, -1.f),
    C_Vertex3f(0.127441406f, 0.620773315f, -1.f),
    C_Vertex3f(0.127441406f, 0.620773315f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.387362093f, -0.921927631f, 0.f),
    C_Vertex3f(0.127441406f, 0.620773315f, 0.f),
    C_Vertex3f(0.127441406f, 0.620773315f, -1.f),
    C_Normal3f(0.472210735f, -0.881485641f, 0.f),
    C_Vertex3f(0.121444702f, 0.618253708f, 0.f),
    C_Vertex3f(0.121444702f, 0.618253708f, -1.f),
    C_Normal3f(0.582000375f, -0.813188553f, 0.f),
    C_Vertex3f(0.111032009f, 0.611982822f, 0.f),
    C_Vertex3f(0.111032009f, 0.611982822f, -1.f),
    C_Normal3f(0.710457981f, -0.703739643f, 0.f),
    C_Vertex3f(0.102200031f, 0.604480267f, 0.f),
    C_Vertex3f(0.102200031f, 0.604480267f, -1.f),
    C_Normal3f(0.822694838f, -0.568483233f, 0.f),
    C_Vertex3f(0.0949487686f, 0.59574604f, 0.f),
    C_Vertex3f(0.0949487686f, 0.59574604f, -1.f),
    C_Normal3f(0.910453439f, -0.413611412f, 0.f),
    C_Vertex3f(0.0892782211f, 0.585780144f, 0.f),
    C_Vertex3f(0.0892782211f, 0.585780144f, -1.f),
    C_Normal3f(0.962118983f, -0.272629827f, 0.f),
    C_Vertex3f(0.0848612785f, 0.573541164f, 0.f),
    C_Vertex3f(0.0848612785f, 0.573541164f, -1.f),
    C_Normal3f(0.987116218f, -0.160005018f, 0.f),
    C_Vertex3f(0.0816979408f, 0.559029102f, 0.f),
    C_Vertex3f(0.0816979408f, 0.559029102f, -1.f),
    C_Normal3f(0.996403515f, -0.0847353712f, 0.f),
    C_Vertex3f(0.079788208f, 0.542243958f, 0.f),
    C_Vertex3f(0.079788208f, 0.542243958f, -1.f),
    C_Normal3f(0.999407887f, -0.0344071575f, 0.f),
    C_Vertex3f(0.079460144f, 0.532714844f, 0.f),
    C_Vertex3f(0.079460144f, 0.532714844f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.079460144f, 0.532714844f, 0.f),
    C_Vertex3f(0.079460144f, 0.532714844f, -1.f),
    C_Vertex3f(0.120605469f, 0.532714844f, -1.f),
    C_Vertex3f(0.120605469f, 0.532714844f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.256347656f, 0.532714844f, 0.f),
    C_Vertex3f(0.256347656f, 0.532714844f, -1.f),
    C_Vertex3f(0.256347656f, 0.44140625f, -1.f),
    C_Vertex3f(0.256347656f, 0.44140625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.256347656f, 0.44140625f, 0.f),
    C_Vertex3f(0.256347656f, 0.44140625f, -1.f),
    C_Vertex3f(0.171875f, 0.44140625f, -1.f),
    C_Vertex3f(0.171875f, 0.44140625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.171875f, 0.44140625f, 0.f),
    C_Vertex3f(0.171875f, 0.44140625f, -1.f),
    C_Normal3f(-0.999989033f, 0.00469817035f, 0.f),
    C_Vertex3f(0.171875f, 0.520994186f, 0.f),
    C_Vertex3f(0.171875f, 0.520994186f, -1.f),
    C_Normal3f(-0.99875015f, 0.0499805883f, 0.f),
    C_Vertex3f(0.172312975f, 0.534627676f, 0.f),
    C_Vertex3f(0.172312975f, 0.534627676f, -1.f),
    C_Normal3f(-0.995941818f, 0.0899985805f, 0.f),
    C_Vertex3f(0.173188925f, 0.547250032f, 0.f),
    C_Vertex3f(0.173188925f, 0.547250032f, -1.f),
    C_Normal3f(-0.990607917f, 0.136733472f, 0.f),
    C_Vertex3f(0.17450285f, 0.558861256f, 0.f),
    C_Vertex3f(0.17450285f, 0.558861256f, -1.f),
    C_Normal3f(-0.981468201f, 0.191625014f, 0.f),
    C_Vertex3f(0.176254749f, 0.569461346f, 0.f),
    C_Vertex3f(0.176254749f, 0.569461346f, -1.f),
    C_Normal3f(-0.966588378f, 0.256333619f, 0.f),
    C_Vertex3f(0.178444624f, 0.579050303f, 0.f),
    C_Vertex3f(0.178444624f, 0.579050303f, -1.f),
    C_Normal3f(-0.943070769f, 0.3325921f, 0.f),
    C_Vertex3f(0.181072474f, 0.587628126f, 0.f),
    C_Vertex3f(0.181072474f, 0.587628126f, -1.f),
    C_Normal3f(-0.901267648f, 0.433262706f, 0.f),
    C_Vertex3f(0.184138298f, 0.595194817f, 0.f),
    C_Vertex3f(0.184138298f, 0.595194817f, -1.f),
    C_Normal3f(-0.832335711f, 0.554271877f, 0.f),
    C_Vertex3f(0.190488815f, 0.607215881f, 0.f),
    C_Vertex3f(0.190488815f, 0.607215881f, -1.f),
    C_Normal3f(-0.734007537f, 0.679141283f, 0.f),
    C_Vertex3f(0.202368736f, 0.622570992f, 0.f),
    C_Vertex3f(0.202368736f, 0.622570992f, -1.f),
    C_Normal3f(-0.607326269f, 0.794452429f, 0.f),
    C_Vertex3f(0.216931343f, 0.63579464f, 0.f),
    C_Vertex3f(0.216931343f, 0.63579464f, -1.f),
    C_Normal3f(-0.496725231f, 0.867907882f, 0.f),
    C_Vertex3f(0.234176636f, 0.646886826f, 0.f),
    C_Vertex3f(0.234176636f, 0.646886826f, -1.f),
    C_Normal3f(-0.410103351f, 0.912039101f, 0.f),
    C_Vertex3f(0.244140625f, 0.651367188f, 0.f),
    C_Vertex3f(0.244140625f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.848972201f, 0.528437555f, 0.f),
    C_Vertex3f(0.244140625f, 0.651367188f, 0.f),
    C_Vertex3f(0.244140625f, 0.651367188f, -1.f),
    C_Vertex3f(0.263183594f, 0.620773315f, -1.f),
    C_Vertex3f(0.263183594f, 0.620773315f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.387362093f, -0.921927631f, 0.f),
    C_Vertex3f(0.263183594f, 0.620773315f, 0.f),
    C_Vertex3f(0.263183594f, 0.620773315f, -1.f),
    C_Normal3f(0.472210735f, -0.881485641f, 0.f),
    C_Vertex3f(0.25718689f, 0.618253708f, 0.f),
    C_Vertex3f(0.25718689f, 0.618253708f, -1.f),
    C_Normal3f(0.582000375f, -0.813188553f, 0.f),
    C_Vertex3f(0.246774197f, 0.611982822f, 0.f),
    C_Vertex3f(0.246774197f, 0.611982822f, -1.f),
    C_Normal3f(0.710457981f, -0.703739643f, 0.f),
    C_Vertex3f(0.237942219f, 0.604480267f, 0.f),
    C_Vertex3f(0.237942219f, 0.604480267f, -1.f),
    C_Normal3f(0.822694838f, -0.568483233f, 0.f),
    C_Vertex3f(0.230690956f, 0.59574604f, 0.f),
    C_Vertex3f(0.230690956f, 0.59574604f, -1.f),
    C_Normal3f(0.910453439f, -0.413611412f, 0.f),
    C_Vertex3f(0.225020409f, 0.585780144f, 0.f),
    C_Vertex3f(0.225020409f, 0.585780144f, -1.f),
    C_Normal3f(0.962118983f, -0.272629827f, 0.f),
    C_Vertex3f(0.220603466f, 0.573541164f, 0.f),
    C_Vertex3f(0.220603466f, 0.573541164f, -1.f),
    C_Normal3f(0.987116218f, -0.160005018f, 0.f),
    C_Vertex3f(0.217440128f, 0.559029102f, 0.f),
    C_Vertex3f(0.217440128f, 0.559029102f, -1.f),
    C_Normal3f(0.996403515f, -0.0847353712f, 0.f),
    C_Vertex3f(0.215530396f, 0.542243958f, 0.f),
    C_Vertex3f(0.215530396f, 0.542243958f, -1.f),
    C_Normal3f(0.999407887f, -0.0344071575f, 0.f),
    C_Vertex3f(0.215202332f, 0.532714844f, 0.f),
    C_Vertex3f(0.215202332f, 0.532714844f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.215202332f, 0.532714844f, 0.f),
    C_Vertex3f(0.215202332f, 0.532714844f, -1.f),
    C_Vertex3f(0.256347656f, 0.532714844f, -1.f),
    C_Vertex3f(0.256347656f, 0.532714844f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120605469f, 0.44140625f, 0.f),
    C_Vertex3f(0.120605469f, 0.532714844f, 0.f),
    C_Vertex3f(0.079460144f, 0.532714844f, 0.f),
    C_Vertex3f(0.0483961105f, 0.595194817f, 0.f),
    C_Vertex3f(0.045330286f, 0.587628126f, 0.f),
    C_Vertex3f(0.0427024364f, 0.579050303f, 0.f),
    C_Vertex3f(0.0405125618f, 0.569461346f, 0.f),
    C_Vertex3f(0.0387606621f, 0.558861256f, 0.f),
    C_Vertex3f(0.0374467373f, 0.547250032f, 0.f),
    C_Vertex3f(0.0365707874f, 0.534627676f, 0.f),
    C_Vertex3f(0.0361328125f, 0.520994186f, 0.f),
    C_Vertex3f(0.0361328125f, 0.44140625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.079460144f, 0.532714844f, 0.f),
    C_Vertex3f(0.079788208f, 0.542243958f, 0.f),
    C_Vertex3f(0.0811891556f, 0.63579464f, 0.f),
    C_Vertex3f(0.0666265488f, 0.622570992f, 0.f),
    C_Vertex3f(0.0547466278f, 0.607215881f, 0.f),
    C_Vertex3f(0.0483961105f, 0.595194817f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.108398438f, 0.651367188f, 0.f),
    C_Vertex3f(0.0984344482f, 0.646886826f, 0.f),
    C_Vertex3f(0.102200031f, 0.604480267f, 0.f),
    C_Vertex3f(0.111032009f, 0.611982822f, 0.f),
    C_Vertex3f(0.121444702f, 0.618253708f, 0.f),
    C_Vertex3f(0.127441406f, 0.620773315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0984344482f, 0.646886826f, 0.f),
    C_Vertex3f(0.0811891556f, 0.63579464f, 0.f),
    C_Vertex3f(0.0816979408f, 0.559029102f, 0.f),
    C_Vertex3f(0.0848612785f, 0.573541164f, 0.f),
    C_Vertex3f(0.0892782211f, 0.585780144f, 0.f),
    C_Vertex3f(0.0949487686f, 0.59574604f, 0.f),
    C_Vertex3f(0.102200031f, 0.604480267f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.256347656f, 0.44140625f, 0.f),
    C_Vertex3f(0.256347656f, 0.532714844f, 0.f),
    C_Vertex3f(0.215202332f, 0.532714844f, 0.f),
    C_Vertex3f(0.184138298f, 0.595194817f, 0.f),
    C_Vertex3f(0.181072474f, 0.587628126f, 0.f),
    C_Vertex3f(0.178444624f, 0.579050303f, 0.f),
    C_Vertex3f(0.176254749f, 0.569461346f, 0.f),
    C_Vertex3f(0.17450285f, 0.558861256f, 0.f),
    C_Vertex3f(0.173188925f, 0.547250032f, 0.f),
    C_Vertex3f(0.172312975f, 0.534627676f, 0.f),
    C_Vertex3f(0.171875f, 0.520994186f, 0.f),
    C_Vertex3f(0.171875f, 0.44140625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.215202332f, 0.532714844f, 0.f),
    C_Vertex3f(0.215530396f, 0.542243958f, 0.f),
    C_Vertex3f(0.216931343f, 0.63579464f, 0.f),
    C_Vertex3f(0.202368736f, 0.622570992f, 0.f),
    C_Vertex3f(0.190488815f, 0.607215881f, 0.f),
    C_Vertex3f(0.184138298f, 0.595194817f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.244140625f, 0.651367188f, 0.f),
    C_Vertex3f(0.234176636f, 0.646886826f, 0.f),
    C_Vertex3f(0.237942219f, 0.604480267f, 0.f),
    C_Vertex3f(0.246774197f, 0.611982822f, 0.f),
    C_Vertex3f(0.25718689f, 0.618253708f, 0.f),
    C_Vertex3f(0.263183594f, 0.620773315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.234176636f, 0.646886826f, 0.f),
    C_Vertex3f(0.216931343f, 0.63579464f, 0.f),
    C_Vertex3f(0.217440128f, 0.559029102f, 0.f),
    C_Vertex3f(0.220603466f, 0.573541164f, 0.f),
    C_Vertex3f(0.225020409f, 0.585780144f, 0.f),
    C_Vertex3f(0.230690956f, 0.59574604f, 0.f),
    C_Vertex3f(0.237942219f, 0.604480267f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.216931343f, 0.63579464f, 0.f),
    C_Vertex3f(0.215530396f, 0.542243958f, 0.f),
    C_Vertex3f(0.217440128f, 0.559029102f, 0.f),
    C_Vertex3f(0.0811891556f, 0.63579464f, 0.f),
    C_Vertex3f(0.079788208f, 0.542243958f, 0.f),
    C_Vertex3f(0.0816979408f, 0.559029102f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120605469f, 0.44140625f, -1.f),
    C_Vertex3f(0.120605469f, 0.532714844f, -1.f),
    C_Vertex3f(0.079460144f, 0.532714844f, -1.f),
    C_Vertex3f(0.0483961105f, 0.595194817f, -1.f),
    C_Vertex3f(0.045330286f, 0.587628126f, -1.f),
    C_Vertex3f(0.0427024364f, 0.579050303f, -1.f),
    C_Vertex3f(0.0405125618f, 0.569461346f, -1.f),
    C_Vertex3f(0.0387606621f, 0.558861256f, -1.f),
    C_Vertex3f(0.0374467373f, 0.547250032f, -1.f),
    C_Vertex3f(0.0365707874f, 0.534627676f, -1.f),
    C_Vertex3f(0.0361328125f, 0.520994186f, -1.f),
    C_Vertex3f(0.0361328125f, 0.44140625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.079460144f, 0.532714844f, -1.f),
    C_Vertex3f(0.079788208f, 0.542243958f, -1.f),
    C_Vertex3f(0.0811891556f, 0.63579464f, -1.f),
    C_Vertex3f(0.0666265488f, 0.622570992f, -1.f),
    C_Vertex3f(0.0547466278f, 0.607215881f, -1.f),
    C_Vertex3f(0.0483961105f, 0.595194817f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.108398438f, 0.651367188f, -1.f),
    C_Vertex3f(0.0984344482f, 0.646886826f, -1.f),
    C_Vertex3f(0.102200031f, 0.604480267f, -1.f),
    C_Vertex3f(0.111032009f, 0.611982822f, -1.f),
    C_Vertex3f(0.121444702f, 0.618253708f, -1.f),
    C_Vertex3f(0.127441406f, 0.620773315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0984344482f, 0.646886826f, -1.f),
    C_Vertex3f(0.0811891556f, 0.63579464f, -1.f),
    C_Vertex3f(0.0816979408f, 0.559029102f, -1.f),
    C_Vertex3f(0.0848612785f, 0.573541164f, -1.f),
    C_Vertex3f(0.0892782211f, 0.585780144f, -1.f),
    C_Vertex3f(0.0949487686f, 0.59574604f, -1.f),
    C_Vertex3f(0.102200031f, 0.604480267f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.256347656f, 0.44140625f, -1.f),
    C_Vertex3f(0.256347656f, 0.532714844f, -1.f),
    C_Vertex3f(0.215202332f, 0.532714844f, -1.f),
    C_Vertex3f(0.184138298f, 0.595194817f, -1.f),
    C_Vertex3f(0.181072474f, 0.587628126f, -1.f),
    C_Vertex3f(0.178444624f, 0.579050303f, -1.f),
    C_Vertex3f(0.176254749f, 0.569461346f, -1.f),
    C_Vertex3f(0.17450285f, 0.558861256f, -1.f),
    C_Vertex3f(0.173188925f, 0.547250032f, -1.f),
    C_Vertex3f(0.172312975f, 0.534627676f, -1.f),
    C_Vertex3f(0.171875f, 0.520994186f, -1.f),
    C_Vertex3f(0.171875f, 0.44140625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.215202332f, 0.532714844f, -1.f),
    C_Vertex3f(0.215530396f, 0.542243958f, -1.f),
    C_Vertex3f(0.216931343f, 0.63579464f, -1.f),
    C_Vertex3f(0.202368736f, 0.622570992f, -1.f),
    C_Vertex3f(0.190488815f, 0.607215881f, -1.f),
    C_Vertex3f(0.184138298f, 0.595194817f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.244140625f, 0.651367188f, -1.f),
    C_Vertex3f(0.234176636f, 0.646886826f, -1.f),
    C_Vertex3f(0.237942219f, 0.604480267f, -1.f),
    C_Vertex3f(0.246774197f, 0.611982822f, -1.f),
    C_Vertex3f(0.25718689f, 0.618253708f, -1.f),
    C_Vertex3f(0.263183594f, 0.620773315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.234176636f, 0.646886826f, -1.f),
    C_Vertex3f(0.216931343f, 0.63579464f, -1.f),
    C_Vertex3f(0.217440128f, 0.559029102f, -1.f),
    C_Vertex3f(0.220603466f, 0.573541164f, -1.f),
    C_Vertex3f(0.225020409f, 0.585780144f, -1.f),
    C_Vertex3f(0.230690956f, 0.59574604f, -1.f),
    C_Vertex3f(0.237942219f, 0.604480267f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.216931343f, 0.63579464f, -1.f),
    C_Vertex3f(0.215530396f, 0.542243958f, -1.f),
    C_Vertex3f(0.217440128f, 0.559029102f, -1.f),
    C_Vertex3f(0.0811891556f, 0.63579464f, -1.f),
    C_Vertex3f(0.079788208f, 0.542243958f, -1.f),
    C_Vertex3f(0.0816979408f, 0.559029102f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_148[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0380859375f, 0.556152344f, 0.f),
    C_Vertex3f(0.0380859375f, 0.556152344f, -1.f),
    C_Vertex3f(0.0380859375f, 0.647460938f, -1.f),
    C_Vertex3f(0.0380859375f, 0.647460938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0380859375f, 0.647460938f, 0.f),
    C_Vertex3f(0.0380859375f, 0.647460938f, -1.f),
    C_Vertex3f(0.122558594f, 0.647460938f, -1.f),
    C_Vertex3f(0.122558594f, 0.647460938f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.122558594f, 0.647460938f, 0.f),
    C_Vertex3f(0.122558594f, 0.647460938f, -1.f),
    C_Normal3f(0.999989629f, -0.00455305772f, 0.f),
    C_Vertex3f(0.122558594f, 0.567873001f, 0.f),
    C_Vertex3f(0.122558594f, 0.567873001f, -1.f),
    C_Normal3f(0.998822749f, -0.0485085919f, 0.f),
    C_Vertex3f(0.122134209f, 0.554253101f, 0.f),
    C_Vertex3f(0.122134209f, 0.554253101f, -1.f),
    C_Normal3f(0.996167064f, -0.0874710083f, 0.f),
    C_Vertex3f(0.121285439f, 0.541657925f, 0.f),
    C_Vertex3f(0.121285439f, 0.541657925f, -1.f),
    C_Normal3f(0.99109906f, -0.133126646f, 0.f),
    C_Vertex3f(0.120012283f, 0.530087471f, 0.f),
    C_Vertex3f(0.120012283f, 0.530087471f, -1.f),
    C_Normal3f(0.982363462f, -0.186981231f, 0.f),
    C_Vertex3f(0.118314743f, 0.51954174f, 0.f),
    C_Vertex3f(0.118314743f, 0.51954174f, -1.f),
    C_Normal3f(0.968034923f, -0.250815451f, 0.f),
    C_Vertex3f(0.116192818f, 0.510020733f, 0.f),
    C_Vertex3f(0.116192818f, 0.510020733f, -1.f),
    C_Normal3f(0.945174932f, -0.326564759f, 0.f),
    C_Vertex3f(0.113646507f, 0.501524448f, 0.f),
    C_Vertex3f(0.113646507f, 0.501524448f, -1.f),
    C_Normal3f(0.901648641f, -0.432469338f, 0.f),
    C_Vertex3f(0.110675812f, 0.494052887f, 0.f),
    C_Vertex3f(0.110675812f, 0.494052887f, -1.f),
    C_Normal3f(0.831324697f, -0.555787146f, 0.f),
    C_Vertex3f(0.104270935f, 0.481977463f, 0.f),
    C_Vertex3f(0.104270935f, 0.481977463f, -1.f),
    C_Normal3f(0.733749926f, -0.679419637f, 0.f),
    C_Vertex3f(0.0922279358f, 0.466459274f, 0.f),
    C_Vertex3f(0.0922279358f, 0.466459274f, -1.f),
    C_Normal3f(0.607928216f, -0.793991923f, 0.f),
    C_Vertex3f(0.0775566101f, 0.453126907f, 0.f),
    C_Vertex3f(0.0775566101f, 0.453126907f, -1.f),
    C_Normal3f(0.497282386f, -0.867588699f, 0.f),
    C_Vertex3f(0.060256958f, 0.441980362f, 0.f),
    C_Vertex3f(0.060256958f, 0.441980362f, -1.f),
    C_Normal3f(0.410103351f, -0.912039101f, 0.f),
    C_Vertex3f(0.0502929688f, 0.4375f, 0.f),
    C_Vertex3f(0.0502929688f, 0.4375f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.852290988f, -0.52306819f, 0.f),
    C_Vertex3f(0.0502929688f, 0.4375f, 0.f),
    C_Vertex3f(0.0502929688f, 0.4375f, -1.f),
    C_Vertex3f(0.03125f, 0.468528748f, -1.f),
    C_Vertex3f(0.03125f, 0.468528748f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.378667861f, 0.925532579f, 0.f),
    C_Vertex3f(0.03125f, 0.468528748f, 0.f),
    C_Vertex3f(0.03125f, 0.468528748f, -1.f),
    C_Normal3f(-0.470709801f, 0.882288098f, 0.f),
    C_Vertex3f(0.0371379852f, 0.470937729f, 0.f),
    C_Vertex3f(0.0371379852f, 0.470937729f, -1.f),
    C_Normal3f(-0.58732599f, 0.80935055f, 0.f),
    C_Vertex3f(0.0474147797f, 0.477152824f, 0.f),
    C_Vertex3f(0.0474147797f, 0.477152824f, -1.f),
    C_Normal3f(-0.719236493f, 0.694765329f, 0.f),
    C_Vertex3f(0.0561923981f, 0.484765053f, 0.f),
    C_Vertex3f(0.0561923981f, 0.484765053f, -1.f),
    C_Normal3f(-0.829798639f, 0.558062911f, 0.f),
    C_Vertex3f(0.0634708405f, 0.493774414f, 0.f),
    C_Vertex3f(0.0634708405f, 0.493774414f, -1.f),
    C_Normal3f(-0.911361814f, 0.411606163f, 0.f),
    C_Vertex3f(0.0692501068f, 0.504180908f, 0.f),
    C_Vertex3f(0.0692501068f, 0.504180908f, -1.f),
    C_Normal3f(-0.96053046f, 0.278174967f, 0.f),
    C_Vertex3f(0.0737485886f, 0.516530991f, 0.f),
    C_Vertex3f(0.0737485886f, 0.516530991f, -1.f),
    C_Normal3f(-0.98604548f, 0.166476414f, 0.f),
    C_Vertex3f(0.0769662857f, 0.530824661f, 0.f),
    C_Vertex3f(0.0769662857f, 0.530824661f, -1.f),
    C_Normal3f(-0.996025264f, 0.0890714675f, 0.f),
    C_Vertex3f(0.0789031982f, 0.54706192f, 0.f),
    C_Vertex3f(0.0789031982f, 0.54706192f, -1.f),
    C_Normal3f(-0.999349475f, 0.0360654853f, 0.f),
    C_Vertex3f(0.0792312622f, 0.556152344f, 0.f),
    C_Vertex3f(0.0792312622f, 0.556152344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0792312622f, 0.556152344f, 0.f),
    C_Vertex3f(0.0792312622f, 0.556152344f, -1.f),
    C_Vertex3f(0.0380859375f, 0.556152344f, -1.f),
    C_Vertex3f(0.0380859375f, 0.556152344f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.173828125f, 0.556152344f, 0.f),
    C_Vertex3f(0.173828125f, 0.556152344f, -1.f),
    C_Vertex3f(0.173828125f, 0.647460938f, -1.f),
    C_Vertex3f(0.173828125f, 0.647460938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.173828125f, 0.647460938f, 0.f),
    C_Vertex3f(0.173828125f, 0.647460938f, -1.f),
    C_Vertex3f(0.258300781f, 0.647460938f, -1.f),
    C_Vertex3f(0.258300781f, 0.647460938f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.258300781f, 0.647460938f, 0.f),
    C_Vertex3f(0.258300781f, 0.647460938f, -1.f),
    C_Normal3f(0.999989629f, -0.00455305772f, 0.f),
    C_Vertex3f(0.258300781f, 0.567873001f, 0.f),
    C_Vertex3f(0.258300781f, 0.567873001f, -1.f),
    C_Normal3f(0.998822749f, -0.0485085919f, 0.f),
    C_Vertex3f(0.257876396f, 0.554253101f, 0.f),
    C_Vertex3f(0.257876396f, 0.554253101f, -1.f),
    C_Normal3f(0.996167064f, -0.0874710083f, 0.f),
    C_Vertex3f(0.257027626f, 0.541657925f, 0.f),
    C_Vertex3f(0.257027626f, 0.541657925f, -1.f),
    C_Normal3f(0.99109906f, -0.133126646f, 0.f),
    C_Vertex3f(0.255754471f, 0.530087471f, 0.f),
    C_Vertex3f(0.255754471f, 0.530087471f, -1.f),
    C_Normal3f(0.982363462f, -0.186981231f, 0.f),
    C_Vertex3f(0.254056931f, 0.51954174f, 0.f),
    C_Vertex3f(0.254056931f, 0.51954174f, -1.f),
    C_Normal3f(0.968034923f, -0.250815451f, 0.f),
    C_Vertex3f(0.251935005f, 0.510020733f, 0.f),
    C_Vertex3f(0.251935005f, 0.510020733f, -1.f),
    C_Normal3f(0.945174932f, -0.326564759f, 0.f),
    C_Vertex3f(0.249388695f, 0.501524448f, 0.f),
    C_Vertex3f(0.249388695f, 0.501524448f, -1.f),
    C_Normal3f(0.901648641f, -0.432469338f, 0.f),
    C_Vertex3f(0.246417999f, 0.494052887f, 0.f),
    C_Vertex3f(0.246417999f, 0.494052887f, -1.f),
    C_Normal3f(0.831324697f, -0.555787146f, 0.f),
    C_Vertex3f(0.240013123f, 0.481977463f, 0.f),
    C_Vertex3f(0.240013123f, 0.481977463f, -1.f),
    C_Normal3f(0.733749926f, -0.679419637f, 0.f),
    C_Vertex3f(0.227970123f, 0.466459274f, 0.f),
    C_Vertex3f(0.227970123f, 0.466459274f, -1.f),
    C_Normal3f(0.607928216f, -0.793991923f, 0.f),
    C_Vertex3f(0.213298798f, 0.453126907f, 0.f),
    C_Vertex3f(0.213298798f, 0.453126907f, -1.f),
    C_Normal3f(0.497282386f, -0.867588699f, 0.f),
    C_Vertex3f(0.195999146f, 0.441980362f, 0.f),
    C_Vertex3f(0.195999146f, 0.441980362f, -1.f),
    C_Normal3f(0.410103351f, -0.912039101f, 0.f),
    C_Vertex3f(0.186035156f, 0.4375f, 0.f),
    C_Vertex3f(0.186035156f, 0.4375f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.852290988f, -0.52306819f, 0.f),
    C_Vertex3f(0.186035156f, 0.4375f, 0.f),
    C_Vertex3f(0.186035156f, 0.4375f, -1.f),
    C_Vertex3f(0.166992188f, 0.468528748f, -1.f),
    C_Vertex3f(0.166992188f, 0.468528748f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.378667861f, 0.925532579f, 0.f),
    C_Vertex3f(0.166992188f, 0.468528748f, 0.f),
    C_Vertex3f(0.166992188f, 0.468528748f, -1.f),
    C_Normal3f(-0.470709801f, 0.882288098f, 0.f),
    C_Vertex3f(0.172880173f, 0.470937729f, 0.f),
    C_Vertex3f(0.172880173f, 0.470937729f, -1.f),
    C_Normal3f(-0.58732599f, 0.80935055f, 0.f),
    C_Vertex3f(0.183156967f, 0.477152824f, 0.f),
    C_Vertex3f(0.183156967f, 0.477152824f, -1.f),
    C_Normal3f(-0.719236493f, 0.694765329f, 0.f),
    C_Vertex3f(0.191934586f, 0.484765053f, 0.f),
    C_Vertex3f(0.191934586f, 0.484765053f, -1.f),
    C_Normal3f(-0.829798639f, 0.558062911f, 0.f),
    C_Vertex3f(0.199213028f, 0.493774414f, 0.f),
    C_Vertex3f(0.199213028f, 0.493774414f, -1.f),
    C_Normal3f(-0.911361814f, 0.411606163f, 0.f),
    C_Vertex3f(0.204992294f, 0.504180908f, 0.f),
    C_Vertex3f(0.204992294f, 0.504180908f, -1.f),
    C_Normal3f(-0.96053046f, 0.278174967f, 0.f),
    C_Vertex3f(0.209490776f, 0.516530991f, 0.f),
    C_Vertex3f(0.209490776f, 0.516530991f, -1.f),
    C_Normal3f(-0.98604548f, 0.166476414f, 0.f),
    C_Vertex3f(0.212708473f, 0.530824661f, 0.f),
    C_Vertex3f(0.212708473f, 0.530824661f, -1.f),
    C_Normal3f(-0.996025264f, 0.0890714675f, 0.f),
    C_Vertex3f(0.214645386f, 0.54706192f, 0.f),
    C_Vertex3f(0.214645386f, 0.54706192f, -1.f),
    C_Normal3f(-0.999349475f, 0.0360654853f, 0.f),
    C_Vertex3f(0.21497345f, 0.556152344f, 0.f),
    C_Vertex3f(0.21497345f, 0.556152344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.21497345f, 0.556152344f, 0.f),
    C_Vertex3f(0.21497345f, 0.556152344f, -1.f),
    C_Vertex3f(0.173828125f, 0.556152344f, -1.f),
    C_Vertex3f(0.173828125f, 0.556152344f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122558594f, 0.647460938f, 0.f),
    C_Vertex3f(0.0380859375f, 0.647460938f, 0.f),
    C_Vertex3f(0.0792312622f, 0.556152344f, 0.f),
    C_Vertex3f(0.0922279358f, 0.466459274f, 0.f),
    C_Vertex3f(0.104270935f, 0.481977463f, 0.f),
    C_Vertex3f(0.110675812f, 0.494052887f, 0.f),
    C_Vertex3f(0.113646507f, 0.501524448f, 0.f),
    C_Vertex3f(0.116192818f, 0.510020733f, 0.f),
    C_Vertex3f(0.118314743f, 0.51954174f, 0.f),
    C_Vertex3f(0.120012283f, 0.530087471f, 0.f),
    C_Vertex3f(0.121285439f, 0.541657925f, 0.f),
    C_Vertex3f(0.122134209f, 0.554253101f, 0.f),
    C_Vertex3f(0.122558594f, 0.567873001f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0789031982f, 0.54706192f, 0.f),
    C_Vertex3f(0.0769662857f, 0.530824661f, 0.f),
    C_Vertex3f(0.0775566101f, 0.453126907f, 0.f),
    C_Vertex3f(0.0922279358f, 0.466459274f, 0.f),
    C_Vertex3f(0.0792312622f, 0.556152344f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0775566101f, 0.453126907f, 0.f),
    C_Vertex3f(0.0769662857f, 0.530824661f, 0.f),
    C_Vertex3f(0.0737485886f, 0.516530991f, 0.f),
    C_Vertex3f(0.0692501068f, 0.504180908f, 0.f),
    C_Vertex3f(0.0634708405f, 0.493774414f, 0.f),
    C_Vertex3f(0.060256958f, 0.441980362f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0561923981f, 0.484765053f, 0.f),
    C_Vertex3f(0.0474147797f, 0.477152824f, 0.f),
    C_Vertex3f(0.0502929688f, 0.4375f, 0.f),
    C_Vertex3f(0.060256958f, 0.441980362f, 0.f),
    C_Vertex3f(0.0634708405f, 0.493774414f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0371379852f, 0.470937729f, 0.f),
    C_Vertex3f(0.03125f, 0.468528748f, 0.f),
    C_Vertex3f(0.0502929688f, 0.4375f, 0.f),
    C_Vertex3f(0.0474147797f, 0.477152824f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.258300781f, 0.647460938f, 0.f),
    C_Vertex3f(0.173828125f, 0.647460938f, 0.f),
    C_Vertex3f(0.21497345f, 0.556152344f, 0.f),
    C_Vertex3f(0.227970123f, 0.466459274f, 0.f),
    C_Vertex3f(0.240013123f, 0.481977463f, 0.f),
    C_Vertex3f(0.246417999f, 0.494052887f, 0.f),
    C_Vertex3f(0.249388695f, 0.501524448f, 0.f),
    C_Vertex3f(0.251935005f, 0.510020733f, 0.f),
    C_Vertex3f(0.254056931f, 0.51954174f, 0.f),
    C_Vertex3f(0.255754471f, 0.530087471f, 0.f),
    C_Vertex3f(0.257027626f, 0.541657925f, 0.f),
    C_Vertex3f(0.257876396f, 0.554253101f, 0.f),
    C_Vertex3f(0.258300781f, 0.567873001f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214645386f, 0.54706192f, 0.f),
    C_Vertex3f(0.212708473f, 0.530824661f, 0.f),
    C_Vertex3f(0.213298798f, 0.453126907f, 0.f),
    C_Vertex3f(0.227970123f, 0.466459274f, 0.f),
    C_Vertex3f(0.21497345f, 0.556152344f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213298798f, 0.453126907f, 0.f),
    C_Vertex3f(0.212708473f, 0.530824661f, 0.f),
    C_Vertex3f(0.209490776f, 0.516530991f, 0.f),
    C_Vertex3f(0.204992294f, 0.504180908f, 0.f),
    C_Vertex3f(0.199213028f, 0.493774414f, 0.f),
    C_Vertex3f(0.195999146f, 0.441980362f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191934586f, 0.484765053f, 0.f),
    C_Vertex3f(0.183156967f, 0.477152824f, 0.f),
    C_Vertex3f(0.186035156f, 0.4375f, 0.f),
    C_Vertex3f(0.195999146f, 0.441980362f, 0.f),
    C_Vertex3f(0.199213028f, 0.493774414f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172880173f, 0.470937729f, 0.f),
    C_Vertex3f(0.166992188f, 0.468528748f, 0.f),
    C_Vertex3f(0.186035156f, 0.4375f, 0.f),
    C_Vertex3f(0.183156967f, 0.477152824f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.21497345f, 0.556152344f, 0.f),
    C_Vertex3f(0.173828125f, 0.647460938f, 0.f),
    C_Vertex3f(0.173828125f, 0.556152344f, 0.f),
    C_Vertex3f(0.0792312622f, 0.556152344f, 0.f),
    C_Vertex3f(0.0380859375f, 0.647460938f, 0.f),
    C_Vertex3f(0.0380859375f, 0.556152344f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122558594f, 0.647460938f, -1.f),
    C_Vertex3f(0.0380859375f, 0.647460938f, -1.f),
    C_Vertex3f(0.0792312622f, 0.556152344f, -1.f),
    C_Vertex3f(0.0922279358f, 0.466459274f, -1.f),
    C_Vertex3f(0.104270935f, 0.481977463f, -1.f),
    C_Vertex3f(0.110675812f, 0.494052887f, -1.f),
    C_Vertex3f(0.113646507f, 0.501524448f, -1.f),
    C_Vertex3f(0.116192818f, 0.510020733f, -1.f),
    C_Vertex3f(0.118314743f, 0.51954174f, -1.f),
    C_Vertex3f(0.120012283f, 0.530087471f, -1.f),
    C_Vertex3f(0.121285439f, 0.541657925f, -1.f),
    C_Vertex3f(0.122134209f, 0.554253101f, -1.f),
    C_Vertex3f(0.122558594f, 0.567873001f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0789031982f, 0.54706192f, -1.f),
    C_Vertex3f(0.0769662857f, 0.530824661f, -1.f),
    C_Vertex3f(0.0775566101f, 0.453126907f, -1.f),
    C_Vertex3f(0.0922279358f, 0.466459274f, -1.f),
    C_Vertex3f(0.0792312622f, 0.556152344f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0775566101f, 0.453126907f, -1.f),
    C_Vertex3f(0.0769662857f, 0.530824661f, -1.f),
    C_Vertex3f(0.0737485886f, 0.516530991f, -1.f),
    C_Vertex3f(0.0692501068f, 0.504180908f, -1.f),
    C_Vertex3f(0.0634708405f, 0.493774414f, -1.f),
    C_Vertex3f(0.060256958f, 0.441980362f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0561923981f, 0.484765053f, -1.f),
    C_Vertex3f(0.0474147797f, 0.477152824f, -1.f),
    C_Vertex3f(0.0502929688f, 0.4375f, -1.f),
    C_Vertex3f(0.060256958f, 0.441980362f, -1.f),
    C_Vertex3f(0.0634708405f, 0.493774414f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0371379852f, 0.470937729f, -1.f),
    C_Vertex3f(0.03125f, 0.468528748f, -1.f),
    C_Vertex3f(0.0502929688f, 0.4375f, -1.f),
    C_Vertex3f(0.0474147797f, 0.477152824f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.258300781f, 0.647460938f, -1.f),
    C_Vertex3f(0.173828125f, 0.647460938f, -1.f),
    C_Vertex3f(0.21497345f, 0.556152344f, -1.f),
    C_Vertex3f(0.227970123f, 0.466459274f, -1.f),
    C_Vertex3f(0.240013123f, 0.481977463f, -1.f),
    C_Vertex3f(0.246417999f, 0.494052887f, -1.f),
    C_Vertex3f(0.249388695f, 0.501524448f, -1.f),
    C_Vertex3f(0.251935005f, 0.510020733f, -1.f),
    C_Vertex3f(0.254056931f, 0.51954174f, -1.f),
    C_Vertex3f(0.255754471f, 0.530087471f, -1.f),
    C_Vertex3f(0.257027626f, 0.541657925f, -1.f),
    C_Vertex3f(0.257876396f, 0.554253101f, -1.f),
    C_Vertex3f(0.258300781f, 0.567873001f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214645386f, 0.54706192f, -1.f),
    C_Vertex3f(0.212708473f, 0.530824661f, -1.f),
    C_Vertex3f(0.213298798f, 0.453126907f, -1.f),
    C_Vertex3f(0.227970123f, 0.466459274f, -1.f),
    C_Vertex3f(0.21497345f, 0.556152344f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213298798f, 0.453126907f, -1.f),
    C_Vertex3f(0.212708473f, 0.530824661f, -1.f),
    C_Vertex3f(0.209490776f, 0.516530991f, -1.f),
    C_Vertex3f(0.204992294f, 0.504180908f, -1.f),
    C_Vertex3f(0.199213028f, 0.493774414f, -1.f),
    C_Vertex3f(0.195999146f, 0.441980362f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191934586f, 0.484765053f, -1.f),
    C_Vertex3f(0.183156967f, 0.477152824f, -1.f),
    C_Vertex3f(0.186035156f, 0.4375f, -1.f),
    C_Vertex3f(0.195999146f, 0.441980362f, -1.f),
    C_Vertex3f(0.199213028f, 0.493774414f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172880173f, 0.470937729f, -1.f),
    C_Vertex3f(0.166992188f, 0.468528748f, -1.f),
    C_Vertex3f(0.186035156f, 0.4375f, -1.f),
    C_Vertex3f(0.183156967f, 0.477152824f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.21497345f, 0.556152344f, -1.f),
    C_Vertex3f(0.173828125f, 0.647460938f, -1.f),
    C_Vertex3f(0.173828125f, 0.556152344f, -1.f),
    C_Vertex3f(0.0792312622f, 0.556152344f, -1.f),
    C_Vertex3f(0.0380859375f, 0.647460938f, -1.f),
    C_Vertex3f(0.0380859375f, 0.556152344f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_149[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.998981714f, -0.0451169387f, 0.f),
    C_Vertex3f(0.0478515625f, 0.307958603f, 0.f),
    C_Vertex3f(0.0478515625f, 0.307958603f, -1.f),
    C_Normal3f(-0.998980939f, 0.0451338552f, 0.f),
    C_Vertex3f(0.0478515625f, 0.319436073f, 0.f),
    C_Vertex3f(0.0478515625f, 0.319436073f, -1.f),
    C_Normal3f(-0.990227342f, 0.139463037f, 0.f),
    C_Vertex3f(0.0488694906f, 0.330489159f, 0.f),
    C_Vertex3f(0.0488694906f, 0.330489159f, -1.f),
    C_Normal3f(-0.971437454f, 0.237295866f, 0.f),
    C_Vertex3f(0.0509053469f, 0.341118813f, 0.f),
    C_Vertex3f(0.0509053469f, 0.341118813f, -1.f),
    C_Normal3f(-0.941942215f, 0.335774779f, 0.f),
    C_Vertex3f(0.0539591312f, 0.351325035f, 0.f),
    C_Vertex3f(0.0539591312f, 0.351325035f, -1.f),
    C_Normal3f(-0.90201664f, 0.431701362f, 0.f),
    C_Vertex3f(0.0580308437f, 0.361107826f, 0.f),
    C_Vertex3f(0.0580308437f, 0.361107826f, -1.f),
    C_Normal3f(-0.852934062f, 0.522018611f, 0.f),
    C_Vertex3f(0.0631204844f, 0.370467186f, 0.f),
    C_Vertex3f(0.0631204844f, 0.370467186f, -1.f),
    C_Normal3f(-0.796772003f, 0.604280055f, 0.f),
    C_Vertex3f(0.0692280531f, 0.379403114f, 0.f),
    C_Vertex3f(0.0692280531f, 0.379403114f, -1.f),
    C_Normal3f(-0.736030757f, 0.67694813f, 0.f),
    C_Vertex3f(0.07635355f, 0.387915611f, 0.f),
    C_Vertex3f(0.07635355f, 0.387915611f, -1.f),
    C_Normal3f(-0.672536016f, 0.740064442f, 0.f),
    C_Vertex3f(0.0844969749f, 0.396004677f, 0.f),
    C_Vertex3f(0.0844969749f, 0.396004677f, -1.f),
    C_Normal3f(-0.600664794f, 0.799500942f, 0.f),
    C_Vertex3f(0.0930434465f, 0.403082609f, 0.f),
    C_Vertex3f(0.0930434465f, 0.403082609f, -1.f),
    C_Normal3f(-0.519335866f, 0.85457027f, 0.f),
    C_Vertex3f(0.101992965f, 0.409149408f, 0.f),
    C_Vertex3f(0.101992965f, 0.409149408f, -1.f),
    C_Normal3f(-0.429974675f, 0.902840912f, 0.f),
    C_Vertex3f(0.11134553f, 0.414205074f, 0.f),
    C_Vertex3f(0.11134553f, 0.414205074f, -1.f),
    C_Normal3f(-0.334896386f, 0.942254961f, 0.f),
    C_Vertex3f(0.121101141f, 0.418249607f, 0.f),
    C_Vertex3f(0.121101141f, 0.418249607f, -1.f),
    C_Normal3f(-0.237041101f, 0.971499622f, 0.f),
    C_Vertex3f(0.131259799f, 0.421283007f, 0.f),
    C_Vertex3f(0.131259799f, 0.421283007f, -1.f),
    C_Normal3f(-0.139536381f, 0.99021697f, 0.f),
    C_Vertex3f(0.141821504f, 0.423305273f, 0.f),
    C_Vertex3f(0.141821504f, 0.423305273f, -1.f),
    C_Normal3f(-0.0451183543f, 0.998981655f, 0.f),
    C_Vertex3f(0.152786255f, 0.424316406f, 0.f),
    C_Vertex3f(0.152786255f, 0.424316406f, -1.f),
    C_Normal3f(0.0449245088f, 0.998990357f, 0.f),
    C_Vertex3f(0.164209366f, 0.424316406f, 0.f),
    C_Vertex3f(0.164209366f, 0.424316406f, -1.f),
    C_Normal3f(0.138402969f, 0.990375936f, 0.f),
    C_Vertex3f(0.175270915f, 0.423305273f, 0.f),
    C_Vertex3f(0.175270915f, 0.423305273f, -1.f),
    C_Normal3f(0.235409692f, 0.971896231f, 0.f),
    C_Vertex3f(0.185915589f, 0.421283007f, 0.f),
    C_Vertex3f(0.185915589f, 0.421283007f, -1.f),
    C_Normal3f(0.333047777f, 0.942909956f, 0.f),
    C_Vertex3f(0.196143389f, 0.418249607f, 0.f),
    C_Vertex3f(0.196143389f, 0.418249607f, -1.f),
    C_Normal3f(0.428205699f, 0.903681278f, 0.f),
    C_Vertex3f(0.205954313f, 0.414205074f, 0.f),
    C_Vertex3f(0.205954313f, 0.414205074f, -1.f),
    C_Normal3f(0.517906308f, 0.855437398f, 0.f),
    C_Vertex3f(0.215348363f, 0.409149408f, 0.f),
    C_Vertex3f(0.215348363f, 0.409149408f, -1.f),
    C_Normal3f(0.599755406f, 0.800183356f, 0.f),
    C_Vertex3f(0.224325538f, 0.403082609f, 0.f),
    C_Vertex3f(0.224325538f, 0.403082609f, -1.f),
    C_Normal3f(0.672230899f, 0.740341663f, 0.f),
    C_Vertex3f(0.232885838f, 0.396004677f, 0.f),
    C_Vertex3f(0.232885838f, 0.396004677f, -1.f),
    C_Normal3f(0.736030757f, 0.67694813f, 0.f),
    C_Vertex3f(0.241029263f, 0.387915611f, 0.f),
    C_Vertex3f(0.241029263f, 0.387915611f, -1.f),
    C_Normal3f(0.796772003f, 0.604280055f, 0.f),
    C_Vertex3f(0.248154759f, 0.379403114f, 0.f),
    C_Vertex3f(0.248154759f, 0.379403114f, -1.f),
    C_Normal3f(0.852934062f, 0.522018611f, 0.f),
    C_Vertex3f(0.254262328f, 0.370467186f, 0.f),
    C_Vertex3f(0.254262328f, 0.370467186f, -1.f),
    C_Normal3f(0.90201664f, 0.431701362f, 0.f),
    C_Vertex3f(0.259351969f, 0.361107826f, 0.f),
    C_Vertex3f(0.259351969f, 0.361107826f, -1.f),
    C_Normal3f(0.941942215f, 0.335774779f, 0.f),
    C_Vertex3f(0.263423681f, 0.351325035f, 0.f),
    C_Vertex3f(0.263423681f, 0.351325035f, -1.f),
    C_Normal3f(0.971437454f, 0.237295866f, 0.f),
    C_Vertex3f(0.266477466f, 0.341118813f, 0.f),
    C_Vertex3f(0.266477466f, 0.341118813f, -1.f),
    C_Normal3f(0.990227342f, 0.139463037f, 0.f),
    C_Vertex3f(0.268513322f, 0.330489159f, 0.f),
    C_Vertex3f(0.268513322f, 0.330489159f, -1.f),
    C_Normal3f(0.998980939f, 0.0451338552f, 0.f),
    C_Vertex3f(0.26953125f, 0.319436073f, 0.f),
    C_Vertex3f(0.26953125f, 0.319436073f, -1.f),
    C_Normal3f(0.998981714f, -0.0451169387f, 0.f),
    C_Vertex3f(0.26953125f, 0.307958603f, 0.f),
    C_Vertex3f(0.26953125f, 0.307958603f, -1.f),
    C_Normal3f(0.990248144f, -0.139315084f, 0.f),
    C_Vertex3f(0.268513322f, 0.296897054f, 0.f),
    C_Vertex3f(0.268513322f, 0.296897054f, -1.f),
    C_Normal3f(0.971533298f, -0.236903176f, 0.f),
    C_Vertex3f(0.266477466f, 0.286252379f, 0.f),
    C_Vertex3f(0.266477466f, 0.286252379f, -1.f),
    C_Normal3f(0.942205548f, -0.335035443f, 0.f),
    C_Vertex3f(0.263423681f, 0.27602458f, 0.f),
    C_Vertex3f(0.263423681f, 0.27602458f, -1.f),
    C_Normal3f(0.902566075f, -0.430551291f, 0.f),
    C_Vertex3f(0.259351969f, 0.266213655f, 0.f),
    C_Vertex3f(0.259351969f, 0.266213655f, -1.f),
    C_Normal3f(0.853894413f, -0.520446241f, 0.f),
    C_Vertex3f(0.254262328f, 0.256819606f, 0.f),
    C_Vertex3f(0.254262328f, 0.256819606f, -1.f),
    C_Normal3f(0.798249662f, -0.602326751f, 0.f),
    C_Vertex3f(0.248154759f, 0.247842431f, 0.f),
    C_Vertex3f(0.248154759f, 0.247842431f, -1.f),
    C_Normal3f(0.738096058f, -0.674695671f, 0.f),
    C_Vertex3f(0.241029263f, 0.239282131f, 0.f),
    C_Vertex3f(0.241029263f, 0.239282131f, -1.f),
    C_Normal3f(0.674695671f, -0.738096058f, 0.f),
    C_Vertex3f(0.232885838f, 0.231138706f, 0.f),
    C_Vertex3f(0.232885838f, 0.231138706f, -1.f),
    C_Normal3f(0.602326751f, -0.798249662f, 0.f),
    C_Vertex3f(0.224325538f, 0.224013209f, 0.f),
    C_Vertex3f(0.224325538f, 0.224013209f, -1.f),
    C_Normal3f(0.520446241f, -0.853894413f, 0.f),
    C_Vertex3f(0.215348363f, 0.217905641f, 0.f),
    C_Vertex3f(0.215348363f, 0.217905641f, -1.f),
    C_Normal3f(0.430551291f, -0.902566075f, 0.f),
    C_Vertex3f(0.205954313f, 0.212816f, 0.f),
    C_Vertex3f(0.205954313f, 0.212816f, -1.f),
    C_Normal3f(0.335035443f, -0.942205548f, 0.f),
    C_Vertex3f(0.196143389f, 0.208744287f, 0.f),
    C_Vertex3f(0.196143389f, 0.208744287f, -1.f),
    C_Normal3f(0.236903176f, -0.971533298f, 0.f),
    C_Vertex3f(0.185915589f, 0.205690503f, 0.f),
    C_Vertex3f(0.185915589f, 0.205690503f, -1.f),
    C_Normal3f(0.139315084f, -0.990248144f, 0.f),
    C_Vertex3f(0.175270915f, 0.203654647f, 0.f),
    C_Vertex3f(0.175270915f, 0.203654647f, -1.f),
    C_Normal3f(0.0452257954f, -0.998976827f, 0.f),
    C_Vertex3f(0.164209366f, 0.202636719f, 0.f),
    C_Vertex3f(0.164209366f, 0.202636719f, -1.f),
    C_Normal3f(-0.0454209335f, -0.998968005f, 0.f),
    C_Vertex3f(0.152786255f, 0.202636719f, 0.f),
    C_Vertex3f(0.152786255f, 0.202636719f, -1.f),
    C_Normal3f(-0.140455633f, -0.990086973f, 0.f),
    C_Vertex3f(0.141821504f, 0.203654647f, 0.f),
    C_Vertex3f(0.141821504f, 0.203654647f, -1.f),
    C_Normal3f(-0.238543689f, -0.971131742f, 0.f),
    C_Vertex3f(0.131259799f, 0.205690503f, 0.f),
    C_Vertex3f(0.131259799f, 0.205690503f, -1.f),
    C_Normal3f(-0.336892247f, -0.941543162f, 0.f),
    C_Vertex3f(0.121101141f, 0.208744287f, 0.f),
    C_Vertex3f(0.121101141f, 0.208744287f, -1.f),
    C_Normal3f(-0.432325572f, -0.901717603f, 0.f),
    C_Vertex3f(0.11134553f, 0.212816f, 0.f),
    C_Vertex3f(0.11134553f, 0.212816f, -1.f),
    C_Normal3f(-0.521877646f, -0.85302043f, 0.f),
    C_Vertex3f(0.101992965f, 0.217905641f, 0.f),
    C_Vertex3f(0.101992965f, 0.217905641f, -1.f),
    C_Normal3f(-0.603235602f, -0.797563016f, 0.f),
    C_Vertex3f(0.0930434465f, 0.224013209f, 0.f),
    C_Vertex3f(0.0930434465f, 0.224013209f, -1.f),
    C_Normal3f(-0.675000072f, -0.737817705f, 0.f),
    C_Vertex3f(0.0844969749f, 0.231138706f, 0.f),
    C_Vertex3f(0.0844969749f, 0.231138706f, -1.f),
    C_Normal3f(-0.738096058f, -0.674695671f, 0.f),
    C_Vertex3f(0.07635355f, 0.239282131f, 0.f),
    C_Vertex3f(0.07635355f, 0.239282131f, -1.f),
    C_Normal3f(-0.798249662f, -0.602326751f, 0.f),
    C_Vertex3f(0.0692280531f, 0.247842431f, 0.f),
    C_Vertex3f(0.0692280531f, 0.247842431f, -1.f),
    C_Normal3f(-0.853894413f, -0.520446241f, 0.f),
    C_Vertex3f(0.0631204844f, 0.256819606f, 0.f),
    C_Vertex3f(0.0631204844f, 0.256819606f, -1.f),
    C_Normal3f(-0.902566075f, -0.430551291f, 0.f),
    C_Vertex3f(0.0580308437f, 0.266213655f, 0.f),
    C_Vertex3f(0.0580308437f, 0.266213655f, -1.f),
    C_Normal3f(-0.942205548f, -0.335035443f, 0.f),
    C_Vertex3f(0.0539591312f, 0.27602458f, 0.f),
    C_Vertex3f(0.0539591312f, 0.27602458f, -1.f),
    C_Normal3f(-0.971533298f, -0.236903176f, 0.f),
    C_Vertex3f(0.0509053469f, 0.286252379f, 0.f),
    C_Vertex3f(0.0509053469f, 0.286252379f, -1.f),
    C_Normal3f(-0.990248144f, -0.139315084f, 0.f),
    C_Vertex3f(0.0488694906f, 0.296897054f, 0.f),
    C_Vertex3f(0.0488694906f, 0.296897054f, -1.f),
    C_Normal3f(-0.998981714f, -0.0451169387f, 0.f),
    C_Vertex3f(0.0478515625f, 0.307958603f, 0.f),
    C_Vertex3f(0.0478515625f, 0.307958603f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0478515625f, 0.307958603f, 0.f),
    C_Vertex3f(0.0488694906f, 0.296897054f, 0.f),
    C_Vertex3f(0.0509053469f, 0.286252379f, 0.f),
    C_Vertex3f(0.0539591312f, 0.27602458f, 0.f),
    C_Vertex3f(0.0580308437f, 0.266213655f, 0.f),
    C_Vertex3f(0.0631204844f, 0.256819606f, 0.f),
    C_Vertex3f(0.0692280531f, 0.247842431f, 0.f),
    C_Vertex3f(0.07635355f, 0.239282131f, 0.f),
    C_Vertex3f(0.0844969749f, 0.231138706f, 0.f),
    C_Vertex3f(0.0930434465f, 0.224013209f, 0.f),
    C_Vertex3f(0.101992965f, 0.217905641f, 0.f),
    C_Vertex3f(0.11134553f, 0.212816f, 0.f),
    C_Vertex3f(0.121101141f, 0.208744287f, 0.f),
    C_Vertex3f(0.131259799f, 0.205690503f, 0.f),
    C_Vertex3f(0.141821504f, 0.203654647f, 0.f),
    C_Vertex3f(0.152786255f, 0.202636719f, 0.f),
    C_Vertex3f(0.164209366f, 0.202636719f, 0.f),
    C_Vertex3f(0.175270915f, 0.203654647f, 0.f),
    C_Vertex3f(0.185915589f, 0.205690503f, 0.f),
    C_Vertex3f(0.196143389f, 0.208744287f, 0.f),
    C_Vertex3f(0.205954313f, 0.212816f, 0.f),
    C_Vertex3f(0.215348363f, 0.217905641f, 0.f),
    C_Vertex3f(0.224325538f, 0.224013209f, 0.f),
    C_Vertex3f(0.232885838f, 0.231138706f, 0.f),
    C_Vertex3f(0.241029263f, 0.239282131f, 0.f),
    C_Vertex3f(0.248154759f, 0.247842431f, 0.f),
    C_Vertex3f(0.254262328f, 0.256819606f, 0.f),
    C_Vertex3f(0.259351969f, 0.266213655f, 0.f),
    C_Vertex3f(0.263423681f, 0.27602458f, 0.f),
    C_Vertex3f(0.266477466f, 0.286252379f, 0.f),
    C_Vertex3f(0.268513322f, 0.296897054f, 0.f),
    C_Vertex3f(0.26953125f, 0.307958603f, 0.f),
    C_Vertex3f(0.26953125f, 0.319436073f, 0.f),
    C_Vertex3f(0.268513322f, 0.330489159f, 0.f),
    C_Vertex3f(0.266477466f, 0.341118813f, 0.f),
    C_Vertex3f(0.263423681f, 0.351325035f, 0.f),
    C_Vertex3f(0.259351969f, 0.361107826f, 0.f),
    C_Vertex3f(0.254262328f, 0.370467186f, 0.f),
    C_Vertex3f(0.248154759f, 0.379403114f, 0.f),
    C_Vertex3f(0.241029263f, 0.387915611f, 0.f),
    C_Vertex3f(0.232885838f, 0.396004677f, 0.f),
    C_Vertex3f(0.224325538f, 0.403082609f, 0.f),
    C_Vertex3f(0.215348363f, 0.409149408f, 0.f),
    C_Vertex3f(0.205954313f, 0.414205074f, 0.f),
    C_Vertex3f(0.196143389f, 0.418249607f, 0.f),
    C_Vertex3f(0.185915589f, 0.421283007f, 0.f),
    C_Vertex3f(0.175270915f, 0.423305273f, 0.f),
    C_Vertex3f(0.164209366f, 0.424316406f, 0.f),
    C_Vertex3f(0.152786255f, 0.424316406f, 0.f),
    C_Vertex3f(0.141821504f, 0.423305273f, 0.f),
    C_Vertex3f(0.131259799f, 0.421283007f, 0.f),
    C_Vertex3f(0.121101141f, 0.418249607f, 0.f),
    C_Vertex3f(0.11134553f, 0.414205074f, 0.f),
    C_Vertex3f(0.101992965f, 0.409149408f, 0.f),
    C_Vertex3f(0.0930434465f, 0.403082609f, 0.f),
    C_Vertex3f(0.0844969749f, 0.396004677f, 0.f),
    C_Vertex3f(0.07635355f, 0.387915611f, 0.f),
    C_Vertex3f(0.0692280531f, 0.379403114f, 0.f),
    C_Vertex3f(0.0631204844f, 0.370467186f, 0.f),
    C_Vertex3f(0.0580308437f, 0.361107826f, 0.f),
    C_Vertex3f(0.0539591312f, 0.351325035f, 0.f),
    C_Vertex3f(0.0509053469f, 0.341118813f, 0.f),
    C_Vertex3f(0.0488694906f, 0.330489159f, 0.f),
    C_Vertex3f(0.0478515625f, 0.319436073f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0478515625f, 0.307958603f, -1.f),
    C_Vertex3f(0.0488694906f, 0.296897054f, -1.f),
    C_Vertex3f(0.0509053469f, 0.286252379f, -1.f),
    C_Vertex3f(0.0539591312f, 0.27602458f, -1.f),
    C_Vertex3f(0.0580308437f, 0.266213655f, -1.f),
    C_Vertex3f(0.0631204844f, 0.256819606f, -1.f),
    C_Vertex3f(0.0692280531f, 0.247842431f, -1.f),
    C_Vertex3f(0.07635355f, 0.239282131f, -1.f),
    C_Vertex3f(0.0844969749f, 0.231138706f, -1.f),
    C_Vertex3f(0.0930434465f, 0.224013209f, -1.f),
    C_Vertex3f(0.101992965f, 0.217905641f, -1.f),
    C_Vertex3f(0.11134553f, 0.212816f, -1.f),
    C_Vertex3f(0.121101141f, 0.208744287f, -1.f),
    C_Vertex3f(0.131259799f, 0.205690503f, -1.f),
    C_Vertex3f(0.141821504f, 0.203654647f, -1.f),
    C_Vertex3f(0.152786255f, 0.202636719f, -1.f),
    C_Vertex3f(0.164209366f, 0.202636719f, -1.f),
    C_Vertex3f(0.175270915f, 0.203654647f, -1.f),
    C_Vertex3f(0.185915589f, 0.205690503f, -1.f),
    C_Vertex3f(0.196143389f, 0.208744287f, -1.f),
    C_Vertex3f(0.205954313f, 0.212816f, -1.f),
    C_Vertex3f(0.215348363f, 0.217905641f, -1.f),
    C_Vertex3f(0.224325538f, 0.224013209f, -1.f),
    C_Vertex3f(0.232885838f, 0.231138706f, -1.f),
    C_Vertex3f(0.241029263f, 0.239282131f, -1.f),
    C_Vertex3f(0.248154759f, 0.247842431f, -1.f),
    C_Vertex3f(0.254262328f, 0.256819606f, -1.f),
    C_Vertex3f(0.259351969f, 0.266213655f, -1.f),
    C_Vertex3f(0.263423681f, 0.27602458f, -1.f),
    C_Vertex3f(0.266477466f, 0.286252379f, -1.f),
    C_Vertex3f(0.268513322f, 0.296897054f, -1.f),
    C_Vertex3f(0.26953125f, 0.307958603f, -1.f),
    C_Vertex3f(0.26953125f, 0.319436073f, -1.f),
    C_Vertex3f(0.268513322f, 0.330489159f, -1.f),
    C_Vertex3f(0.266477466f, 0.341118813f, -1.f),
    C_Vertex3f(0.263423681f, 0.351325035f, -1.f),
    C_Vertex3f(0.259351969f, 0.361107826f, -1.f),
    C_Vertex3f(0.254262328f, 0.370467186f, -1.f),
    C_Vertex3f(0.248154759f, 0.379403114f, -1.f),
    C_Vertex3f(0.241029263f, 0.387915611f, -1.f),
    C_Vertex3f(0.232885838f, 0.396004677f, -1.f),
    C_Vertex3f(0.224325538f, 0.403082609f, -1.f),
    C_Vertex3f(0.215348363f, 0.409149408f, -1.f),
    C_Vertex3f(0.205954313f, 0.414205074f, -1.f),
    C_Vertex3f(0.196143389f, 0.418249607f, -1.f),
    C_Vertex3f(0.185915589f, 0.421283007f, -1.f),
    C_Vertex3f(0.175270915f, 0.423305273f, -1.f),
    C_Vertex3f(0.164209366f, 0.424316406f, -1.f),
    C_Vertex3f(0.152786255f, 0.424316406f, -1.f),
    C_Vertex3f(0.141821504f, 0.423305273f, -1.f),
    C_Vertex3f(0.131259799f, 0.421283007f, -1.f),
    C_Vertex3f(0.121101141f, 0.418249607f, -1.f),
    C_Vertex3f(0.11134553f, 0.414205074f, -1.f),
    C_Vertex3f(0.101992965f, 0.409149408f, -1.f),
    C_Vertex3f(0.0930434465f, 0.403082609f, -1.f),
    C_Vertex3f(0.0844969749f, 0.396004677f, -1.f),
    C_Vertex3f(0.07635355f, 0.387915611f, -1.f),
    C_Vertex3f(0.0692280531f, 0.379403114f, -1.f),
    C_Vertex3f(0.0631204844f, 0.370467186f, -1.f),
    C_Vertex3f(0.0580308437f, 0.361107826f, -1.f),
    C_Vertex3f(0.0539591312f, 0.351325035f, -1.f),
    C_Vertex3f(0.0509053469f, 0.341118813f, -1.f),
    C_Vertex3f(0.0488694906f, 0.330489159f, -1.f),
    C_Vertex3f(0.0478515625f, 0.319436073f, -1.f),
    C_End(),
    C_Translatef(0.313476563f, 0.f, 0.f),
};
static const Call char_150[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(-0.001953125f, 0.200195313f, 0.f),
    C_Vertex3f(-0.001953125f, 0.200195313f, -1.f),
    C_Vertex3f(-0.001953125f, 0.263671875f, -1.f),
    C_Vertex3f(-0.001953125f, 0.263671875f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(-0.001953125f, 0.263671875f, 0.f),
    C_Vertex3f(-0.001953125f, 0.263671875f, -1.f),
    C_Vertex3f(0.495605469f, 0.263671875f, -1.f),
    C_Vertex3f(0.495605469f, 0.263671875f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.495605469f, 0.263671875f, 0.f),
    C_Vertex3f(0.495605469f, 0.263671875f, -1.f),
    C_Vertex3f(0.495605469f, 0.200195313f, -1.f),
    C_Vertex3f(0.495605469f, 0.200195313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.495605469f, 0.200195313f, 0.f),
    C_Vertex3f(0.495605469f, 0.200195313f, -1.f),
    C_Vertex3f(-0.001953125f, 0.200195313f, -1.f),
    C_Vertex3f(-0.001953125f, 0.200195313f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-0.001953125f, 0.200195313f, 0.f),
    C_Vertex3f(0.495605469f, 0.200195313f, 0.f),
    C_Vertex3f(0.495605469f, 0.263671875f, 0.f),
    C_Vertex3f(-0.001953125f, 0.263671875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-0.001953125f, 0.200195313f, -1.f),
    C_Vertex3f(0.495605469f, 0.200195313f, -1.f),
    C_Vertex3f(0.495605469f, 0.263671875f, -1.f),
    C_Vertex3f(-0.001953125f, 0.263671875f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_151[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.f, 0.200195313f, 0.f),
    C_Vertex3f(0.f, 0.200195313f, -1.f),
    C_Vertex3f(0.f, 0.263671875f, -1.f),
    C_Vertex3f(0.f, 0.263671875f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.f, 0.263671875f, 0.f),
    C_Vertex3f(0.f, 0.263671875f, -1.f),
    C_Vertex3f(0.895019531f, 0.263671875f, -1.f),
    C_Vertex3f(0.895019531f, 0.263671875f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.895019531f, 0.263671875f, 0.f),
    C_Vertex3f(0.895019531f, 0.263671875f, -1.f),
    C_Vertex3f(0.895019531f, 0.200195313f, -1.f),
    C_Vertex3f(0.895019531f, 0.200195313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.895019531f, 0.200195313f, 0.f),
    C_Vertex3f(0.895019531f, 0.200195313f, -1.f),
    C_Vertex3f(0.f, 0.200195313f, -1.f),
    C_Vertex3f(0.f, 0.200195313f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.f, 0.200195313f, 0.f),
    C_Vertex3f(0.895019531f, 0.200195313f, 0.f),
    C_Vertex3f(0.895019531f, 0.263671875f, 0.f),
    C_Vertex3f(0.f, 0.263671875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.f, 0.200195313f, -1.f),
    C_Vertex3f(0.895019531f, 0.200195313f, -1.f),
    C_Vertex3f(0.895019531f, 0.263671875f, -1.f),
    C_Vertex3f(0.f, 0.263671875f, -1.f),
    C_End(),
    C_Translatef(0.895019531f, 0.f, 0.f),
};
static const Call char_152[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999953806f, -0.00961818546f, 0.f),
    C_Vertex3f(0.00305938721f, 0.532714844f, 0.f),
    C_Vertex3f(0.00305938721f, 0.532714844f, -1.f),
    C_Normal3f(-0.999230683f, 0.0392165966f, 0.f),
    C_Vertex3f(0.00300502777f, 0.538366318f, 0.f),
    C_Vertex3f(0.00300502777f, 0.538366318f, -1.f),
    C_Normal3f(-0.994498312f, 0.104752719f, 0.f),
    C_Vertex3f(0.00370240211f, 0.54909873f, 0.f),
    C_Vertex3f(0.00370240211f, 0.54909873f, -1.f),
    C_Normal3f(-0.981874108f, 0.189534381f, 0.f),
    C_Vertex3f(0.00520586967f, 0.559260607f, 0.f),
    C_Vertex3f(0.00520586967f, 0.559260607f, -1.f),
    C_Normal3f(-0.960046053f, 0.279841781f, 0.f),
    C_Vertex3f(0.00751543045f, 0.568851948f, 0.f),
    C_Vertex3f(0.00751543045f, 0.568851948f, -1.f),
    C_Normal3f(-0.927577436f, 0.373630971f, 0.f),
    C_Vertex3f(0.0106310844f, 0.577872753f, 0.f),
    C_Vertex3f(0.0106310844f, 0.577872753f, -1.f),
    C_Normal3f(-0.883691847f, 0.468069077f, 0.f),
    C_Vertex3f(0.0145528316f, 0.586323023f, 0.f),
    C_Vertex3f(0.0145528316f, 0.586323023f, -1.f),
    C_Normal3f(-0.828614712f, 0.559819281f, 0.f),
    C_Vertex3f(0.0192806721f, 0.594202757f, 0.f),
    C_Vertex3f(0.0192806721f, 0.594202757f, -1.f),
    C_Normal3f(-0.759081304f, 0.650995851f, 0.f),
    C_Vertex3f(0.0248146057f, 0.601511955f, 0.f),
    C_Vertex3f(0.0248146057f, 0.601511955f, -1.f),
    C_Normal3f(-0.650275886f, 0.759698212f, 0.f),
    C_Vertex3f(0.0342178345f, 0.611619949f, 0.f),
    C_Vertex3f(0.0342178345f, 0.611619949f, -1.f),
    C_Normal3f(-0.487445921f, 0.87315321f, 0.f),
    C_Vertex3f(0.0484323502f, 0.621727943f, 0.f),
    C_Vertex3f(0.0484323502f, 0.621727943f, -1.f),
    C_Normal3f(-0.287504673f, 0.957779288f, 0.f),
    C_Vertex3f(0.0643949509f, 0.628466606f, 0.f),
    C_Vertex3f(0.0643949509f, 0.628466606f, -1.f),
    C_Normal3f(-0.0979905054f, 0.995187461f, 0.f),
    C_Vertex3f(0.0821056366f, 0.631835938f, 0.f),
    C_Vertex3f(0.0821056366f, 0.631835938f, -1.f),
    C_Normal3f(0.0889112428f, 0.99603951f, 0.f),
    C_Vertex3f(0.098613739f, 0.631835938f, 0.f),
    C_Vertex3f(0.098613739f, 0.631835938f, -1.f),
    C_Normal3f(0.244741365f, 0.969588399f, 0.f),
    C_Vertex3f(0.114637375f, 0.628931999f, 0.f),
    C_Vertex3f(0.114637375f, 0.628931999f, -1.f),
    C_Normal3f(0.345433027f, 0.938443422f, 0.f),
    C_Vertex3f(0.133127213f, 0.623124123f, 0.f),
    C_Vertex3f(0.133127213f, 0.623124123f, -1.f),
    C_Normal3f(0.413061887f, 0.910703003f, 0.f),
    C_Vertex3f(0.154083252f, 0.614412308f, 0.f),
    C_Vertex3f(0.154083252f, 0.614412308f, -1.f),
    C_Normal3f(0.42238301f, 0.906417429f, 0.f),
    C_Vertex3f(0.172250748f, 0.605379105f, 0.f),
    C_Vertex3f(0.172250748f, 0.605379105f, -1.f),
    C_Normal3f(0.345709145f, 0.938341737f, 0.f),
    C_Vertex3f(0.183850288f, 0.600541115f, 0.f),
    C_Vertex3f(0.183850288f, 0.600541115f, -1.f),
    C_Normal3f(0.243631572f, 0.969867826f, 0.f),
    C_Vertex3f(0.19413662f, 0.597315788f, 0.f),
    C_Vertex3f(0.19413662f, 0.597315788f, -1.f),
    C_Normal3f(0.0984259844f, 0.995144367f, 0.f),
    C_Vertex3f(0.203109741f, 0.595703125f, 0.f),
    C_Vertex3f(0.203109741f, 0.595703125f, -1.f),
    C_Normal3f(-0.070669286f, 0.997499824f, 0.f),
    C_Vertex3f(0.210441589f, 0.595703125f, 0.f),
    C_Vertex3f(0.210441589f, 0.595703125f, -1.f),
    C_Normal3f(-0.24143213f, 0.970417738f, 0.f),
    C_Vertex3f(0.216624737f, 0.596660614f, 0.f),
    C_Vertex3f(0.216624737f, 0.596660614f, -1.f),
    C_Normal3f(-0.435192645f, 0.900337279f, 0.f),
    C_Vertex3f(0.221987247f, 0.598575592f, 0.f),
    C_Vertex3f(0.221987247f, 0.598575592f, -1.f),
    C_Normal3f(-0.62994957f, 0.776636064f, 0.f),
    C_Vertex3f(0.226529121f, 0.601448059f, 0.f),
    C_Vertex3f(0.226529121f, 0.601448059f, -1.f),
    C_Normal3f(-0.810694158f, 0.585469842f, 0.f),
    C_Vertex3f(0.230250359f, 0.605278015f, 0.f),
    C_Vertex3f(0.230250359f, 0.605278015f, -1.f),
    C_Normal3f(-0.9168607f, 0.399207145f, 0.f),
    C_Vertex3f(0.233424664f, 0.610996246f, 0.f),
    C_Vertex3f(0.233424664f, 0.610996246f, -1.f),
    C_Normal3f(-0.964133918f, 0.265416414f, 0.f),
    C_Vertex3f(0.236052036f, 0.618602753f, 0.f),
    C_Vertex3f(0.236052036f, 0.618602753f, -1.f),
    C_Normal3f(-0.982875049f, 0.18427363f, 0.f),
    C_Vertex3f(0.238132477f, 0.628097534f, 0.f),
    C_Vertex3f(0.238132477f, 0.628097534f, -1.f),
    C_Normal3f(-0.991047204f, 0.133512393f, 0.f),
    C_Vertex3f(0.238899231f, 0.633789063f, 0.f),
    C_Vertex3f(0.238899231f, 0.633789063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.238899231f, 0.633789063f, 0.f),
    C_Vertex3f(0.238899231f, 0.633789063f, -1.f),
    C_Vertex3f(0.295539856f, 0.633789063f, -1.f),
    C_Vertex3f(0.295539856f, 0.633789063f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999653041f, -0.0263388976f, 0.f),
    C_Vertex3f(0.295539856f, 0.633789063f, 0.f),
    C_Vertex3f(0.295539856f, 0.633789063f, -1.f),
    C_Normal3f(0.99777931f, -0.066607669f, 0.f),
    C_Vertex3f(0.295375824f, 0.627563477f, 0.f),
    C_Vertex3f(0.295375824f, 0.627563477f, -1.f),
    C_Normal3f(0.992581308f, -0.121582195f, 0.f),
    C_Vertex3f(0.294345021f, 0.615890503f, 0.f),
    C_Vertex3f(0.294345021f, 0.615890503f, -1.f),
    C_Normal3f(0.980870843f, -0.194659367f, 0.f),
    C_Vertex3f(0.29261148f, 0.604995728f, 0.f),
    C_Vertex3f(0.29261148f, 0.604995728f, -1.f),
    C_Normal3f(0.961305082f, -0.275485754f, 0.f),
    C_Vertex3f(0.2901752f, 0.59487915f, 0.f),
    C_Vertex3f(0.2901752f, 0.59487915f, -1.f),
    C_Normal3f(0.931648552f, -0.363360554f, 0.f),
    C_Vertex3f(0.28703618f, 0.585540771f, 0.f),
    C_Vertex3f(0.28703618f, 0.585540771f, -1.f),
    C_Normal3f(0.889692426f, -0.456560522f, 0.f),
    C_Vertex3f(0.283194423f, 0.576980591f, 0.f),
    C_Vertex3f(0.283194423f, 0.576980591f, -1.f),
    C_Normal3f(0.833749712f, -0.552142739f, 0.f),
    C_Vertex3f(0.278649926f, 0.569198608f, 0.f),
    C_Vertex3f(0.278649926f, 0.569198608f, -1.f),
    C_Normal3f(0.755483389f, -0.655167818f, 0.f),
    C_Vertex3f(0.273402691f, 0.562194824f, 0.f),
    C_Vertex3f(0.273402691f, 0.562194824f, -1.f),
    C_Normal3f(0.639779389f, -0.768558562f, 0.f),
    C_Vertex3f(0.26447773f, 0.552856445f, 0.f),
    C_Vertex3f(0.26447773f, 0.552856445f, -1.f),
    C_Normal3f(0.477566123f, -0.878595829f, 0.f),
    C_Vertex3f(0.250966549f, 0.543518066f, 0.f),
    C_Vertex3f(0.250966549f, 0.543518066f, -1.f),
    C_Normal3f(0.281305313f, -0.95961827f, 0.f),
    C_Vertex3f(0.235844135f, 0.53729248f, 0.f),
    C_Vertex3f(0.235844135f, 0.53729248f, -1.f),
    C_Normal3f(0.0945234597f, -0.995522618f, 0.f),
    C_Vertex3f(0.219110489f, 0.534179688f, 0.f),
    C_Vertex3f(0.219110489f, 0.534179688f, -1.f),
    C_Normal3f(-0.0947503f, -0.995501101f, 0.f),
    C_Vertex3f(0.20306015f, 0.534179688f, 0.f),
    C_Vertex3f(0.20306015f, 0.534179688f, -1.f),
    C_Normal3f(-0.258822501f, -0.965924919f, 0.f),
    C_Vertex3f(0.187107086f, 0.537225723f, 0.f),
    C_Vertex3f(0.187107086f, 0.537225723f, -1.f),
    C_Normal3f(-0.367869198f, -0.92987752f, 0.f),
    C_Vertex3f(0.168956757f, 0.543317795f, 0.f),
    C_Vertex3f(0.168956757f, 0.543317795f, -1.f),
    C_Normal3f(-0.441156626f, -0.897430122f, 0.f),
    C_Vertex3f(0.148609161f, 0.552455902f, 0.f),
    C_Vertex3f(0.148609161f, 0.552455902f, -1.f),
    C_Normal3f(-0.452606946f, -0.891710103f, 0.f),
    C_Vertex3f(0.130004883f, 0.562465668f, 0.f),
    C_Vertex3f(0.130004883f, 0.562465668f, -1.f),
    C_Normal3f(-0.385828793f, -0.922570407f, 0.f),
    C_Vertex3f(0.117310524f, 0.568342209f, 0.f),
    C_Vertex3f(0.117310524f, 0.568342209f, -1.f),
    C_Normal3f(-0.288805217f, -0.957387865f, 0.f),
    C_Vertex3f(0.106585503f, 0.572259903f, 0.f),
    C_Vertex3f(0.106585503f, 0.572259903f, -1.f),
    C_Normal3f(-0.122500464f, -0.992468476f, 0.f),
    C_Vertex3f(0.0978298187f, 0.57421875f, 0.f),
    C_Vertex3f(0.0978298187f, 0.57421875f, -1.f),
    C_Normal3f(0.0942201167f, -0.995551467f, 0.f),
    C_Vertex3f(0.0907154083f, 0.57421875f, 0.f),
    C_Vertex3f(0.0907154083f, 0.57421875f, -1.f),
    C_Normal3f(0.294618219f, -0.955615103f, 0.f),
    C_Vertex3f(0.083984375f, 0.572908401f, 0.f),
    C_Vertex3f(0.083984375f, 0.572908401f, -1.f),
    C_Normal3f(0.500668526f, -0.865639091f, 0.f),
    C_Vertex3f(0.0779647827f, 0.570287704f, 0.f),
    C_Vertex3f(0.0779647827f, 0.570287704f, -1.f),
    C_Normal3f(0.652620137f, -0.757685304f, 0.f),
    C_Vertex3f(0.0726566315f, 0.566356659f, 0.f),
    C_Vertex3f(0.0726566315f, 0.566356659f, -1.f),
    C_Normal3f(0.744249165f, -0.667902052f, 0.f),
    C_Vertex3f(0.0703582764f, 0.563735962f, 0.f),
    C_Vertex3f(0.0703582764f, 0.563735962f, -1.f),
    C_Normal3f(0.837037444f, -0.547145605f, 0.f),
    C_Vertex3f(0.0679512024f, 0.561113358f, 0.f),
    C_Vertex3f(0.0679512024f, 0.561113358f, -1.f),
    C_Normal3f(0.925027549f, -0.379899919f, 0.f),
    C_Vertex3f(0.0643949509f, 0.554613113f, 0.f),
    C_Vertex3f(0.0643949509f, 0.554613113f, -1.f),
    C_Normal3f(0.980741322f, -0.195311442f, 0.f),
    C_Vertex3f(0.0620965958f, 0.546857834f, 0.f),
    C_Vertex3f(0.0620965958f, 0.546857834f, -1.f),
    C_Normal3f(0.997837007f, -0.065737471f, 0.f),
    C_Vertex3f(0.0610561371f, 0.537847519f, 0.f),
    C_Vertex3f(0.0610561371f, 0.537847519f, -1.f),
    C_Normal3f(0.999775767f, 0.0211769659f, 0.f),
    C_Vertex3f(0.061164856f, 0.532714844f, 0.f),
    C_Vertex3f(0.061164856f, 0.532714844f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.061164856f, 0.532714844f, 0.f),
    C_Vertex3f(0.061164856f, 0.532714844f, -1.f),
    C_Vertex3f(0.00305938721f, 0.532714844f, -1.f),
    C_Vertex3f(0.00305938721f, 0.532714844f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.061164856f, 0.532714844f, 0.f),
    C_Vertex3f(0.0610561371f, 0.537847519f, 0.f),
    C_Vertex3f(0.0484323502f, 0.621727943f, 0.f),
    C_Vertex3f(0.0342178345f, 0.611619949f, 0.f),
    C_Vertex3f(0.0248146057f, 0.601511955f, 0.f),
    C_Vertex3f(0.0192806721f, 0.594202757f, 0.f),
    C_Vertex3f(0.0145528316f, 0.586323023f, 0.f),
    C_Vertex3f(0.0106310844f, 0.577872753f, 0.f),
    C_Vertex3f(0.00751543045f, 0.568851948f, 0.f),
    C_Vertex3f(0.00520586967f, 0.559260607f, 0.f),
    C_Vertex3f(0.00370240211f, 0.54909873f, 0.f),
    C_Vertex3f(0.00305938721f, 0.532714844f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.295539856f, 0.633789063f, 0.f),
    C_Vertex3f(0.238899231f, 0.633789063f, 0.f),
    C_Vertex3f(0.250966549f, 0.543518066f, 0.f),
    C_Vertex3f(0.26447773f, 0.552856445f, 0.f),
    C_Vertex3f(0.273402691f, 0.562194824f, 0.f),
    C_Vertex3f(0.278649926f, 0.569198608f, 0.f),
    C_Vertex3f(0.283194423f, 0.576980591f, 0.f),
    C_Vertex3f(0.28703618f, 0.585540771f, 0.f),
    C_Vertex3f(0.2901752f, 0.59487915f, 0.f),
    C_Vertex3f(0.29261148f, 0.604995728f, 0.f),
    C_Vertex3f(0.294345021f, 0.615890503f, 0.f),
    C_Vertex3f(0.295375824f, 0.627563477f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.238899231f, 0.633789063f, 0.f),
    C_Vertex3f(0.238132477f, 0.628097534f, 0.f),
    C_Vertex3f(0.250966549f, 0.543518066f, 0.f),
    C_Vertex3f(0.236052036f, 0.618602753f, 0.f),
    C_Vertex3f(0.235844135f, 0.53729248f, 0.f),
    C_Vertex3f(0.233424664f, 0.610996246f, 0.f),
    C_Vertex3f(0.230250359f, 0.605278015f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230250359f, 0.605278015f, 0.f),
    C_Vertex3f(0.226529121f, 0.601448059f, 0.f),
    C_Vertex3f(0.235844135f, 0.53729248f, 0.f),
    C_Vertex3f(0.221987247f, 0.598575592f, 0.f),
    C_Vertex3f(0.219110489f, 0.534179688f, 0.f),
    C_Vertex3f(0.216624737f, 0.596660614f, 0.f),
    C_Vertex3f(0.210441589f, 0.595703125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.183850288f, 0.600541115f, 0.f),
    C_Vertex3f(0.187107086f, 0.537225723f, 0.f),
    C_Vertex3f(0.19413662f, 0.597315788f, 0.f),
    C_Vertex3f(0.20306015f, 0.534179688f, 0.f),
    C_Vertex3f(0.203109741f, 0.595703125f, 0.f),
    C_Vertex3f(0.219110489f, 0.534179688f, 0.f),
    C_Vertex3f(0.210441589f, 0.595703125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.117310524f, 0.568342209f, 0.f),
    C_Vertex3f(0.130004883f, 0.562465668f, 0.f),
    C_Vertex3f(0.133127213f, 0.623124123f, 0.f),
    C_Vertex3f(0.148609161f, 0.552455902f, 0.f),
    C_Vertex3f(0.154083252f, 0.614412308f, 0.f),
    C_Vertex3f(0.168956757f, 0.543317795f, 0.f),
    C_Vertex3f(0.172250748f, 0.605379105f, 0.f),
    C_Vertex3f(0.187107086f, 0.537225723f, 0.f),
    C_Vertex3f(0.183850288f, 0.600541115f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0907154083f, 0.57421875f, 0.f),
    C_Vertex3f(0.0978298187f, 0.57421875f, 0.f),
    C_Vertex3f(0.098613739f, 0.631835938f, 0.f),
    C_Vertex3f(0.106585503f, 0.572259903f, 0.f),
    C_Vertex3f(0.114637375f, 0.628931999f, 0.f),
    C_Vertex3f(0.117310524f, 0.568342209f, 0.f),
    C_Vertex3f(0.133127213f, 0.623124123f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0726566315f, 0.566356659f, 0.f),
    C_Vertex3f(0.0779647827f, 0.570287704f, 0.f),
    C_Vertex3f(0.0821056366f, 0.631835938f, 0.f),
    C_Vertex3f(0.083984375f, 0.572908401f, 0.f),
    C_Vertex3f(0.098613739f, 0.631835938f, 0.f),
    C_Vertex3f(0.0907154083f, 0.57421875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0620965958f, 0.546857834f, 0.f),
    C_Vertex3f(0.0643949509f, 0.554613113f, 0.f),
    C_Vertex3f(0.0643949509f, 0.628466606f, 0.f),
    C_Vertex3f(0.0679512024f, 0.561113358f, 0.f),
    C_Vertex3f(0.0821056366f, 0.631835938f, 0.f),
    C_Vertex3f(0.0703582764f, 0.563735962f, 0.f),
    C_Vertex3f(0.0726566315f, 0.566356659f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0643949509f, 0.628466606f, 0.f),
    C_Vertex3f(0.0484323502f, 0.621727943f, 0.f),
    C_Vertex3f(0.0610561371f, 0.537847519f, 0.f),
    C_Vertex3f(0.0620965958f, 0.546857834f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.00305938721f, 0.532714844f, 0.f),
    C_Vertex3f(0.00370240211f, 0.54909873f, 0.f),
    C_Vertex3f(0.00300502777f, 0.538366318f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.061164856f, 0.532714844f, -1.f),
    C_Vertex3f(0.0610561371f, 0.537847519f, -1.f),
    C_Vertex3f(0.0484323502f, 0.621727943f, -1.f),
    C_Vertex3f(0.0342178345f, 0.611619949f, -1.f),
    C_Vertex3f(0.0248146057f, 0.601511955f, -1.f),
    C_Vertex3f(0.0192806721f, 0.594202757f, -1.f),
    C_Vertex3f(0.0145528316f, 0.586323023f, -1.f),
    C_Vertex3f(0.0106310844f, 0.577872753f, -1.f),
    C_Vertex3f(0.00751543045f, 0.568851948f, -1.f),
    C_Vertex3f(0.00520586967f, 0.559260607f, -1.f),
    C_Vertex3f(0.00370240211f, 0.54909873f, -1.f),
    C_Vertex3f(0.00305938721f, 0.532714844f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.295539856f, 0.633789063f, -1.f),
    C_Vertex3f(0.238899231f, 0.633789063f, -1.f),
    C_Vertex3f(0.250966549f, 0.543518066f, -1.f),
    C_Vertex3f(0.26447773f, 0.552856445f, -1.f),
    C_Vertex3f(0.273402691f, 0.562194824f, -1.f),
    C_Vertex3f(0.278649926f, 0.569198608f, -1.f),
    C_Vertex3f(0.283194423f, 0.576980591f, -1.f),
    C_Vertex3f(0.28703618f, 0.585540771f, -1.f),
    C_Vertex3f(0.2901752f, 0.59487915f, -1.f),
    C_Vertex3f(0.29261148f, 0.604995728f, -1.f),
    C_Vertex3f(0.294345021f, 0.615890503f, -1.f),
    C_Vertex3f(0.295375824f, 0.627563477f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.238899231f, 0.633789063f, -1.f),
    C_Vertex3f(0.238132477f, 0.628097534f, -1.f),
    C_Vertex3f(0.250966549f, 0.543518066f, -1.f),
    C_Vertex3f(0.236052036f, 0.618602753f, -1.f),
    C_Vertex3f(0.235844135f, 0.53729248f, -1.f),
    C_Vertex3f(0.233424664f, 0.610996246f, -1.f),
    C_Vertex3f(0.230250359f, 0.605278015f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230250359f, 0.605278015f, -1.f),
    C_Vertex3f(0.226529121f, 0.601448059f, -1.f),
    C_Vertex3f(0.235844135f, 0.53729248f, -1.f),
    C_Vertex3f(0.221987247f, 0.598575592f, -1.f),
    C_Vertex3f(0.219110489f, 0.534179688f, -1.f),
    C_Vertex3f(0.216624737f, 0.596660614f, -1.f),
    C_Vertex3f(0.210441589f, 0.595703125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.183850288f, 0.600541115f, -1.f),
    C_Vertex3f(0.187107086f, 0.537225723f, -1.f),
    C_Vertex3f(0.19413662f, 0.597315788f, -1.f),
    C_Vertex3f(0.20306015f, 0.534179688f, -1.f),
    C_Vertex3f(0.203109741f, 0.595703125f, -1.f),
    C_Vertex3f(0.219110489f, 0.534179688f, -1.f),
    C_Vertex3f(0.210441589f, 0.595703125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.117310524f, 0.568342209f, -1.f),
    C_Vertex3f(0.130004883f, 0.562465668f, -1.f),
    C_Vertex3f(0.133127213f, 0.623124123f, -1.f),
    C_Vertex3f(0.148609161f, 0.552455902f, -1.f),
    C_Vertex3f(0.154083252f, 0.614412308f, -1.f),
    C_Vertex3f(0.168956757f, 0.543317795f, -1.f),
    C_Vertex3f(0.172250748f, 0.605379105f, -1.f),
    C_Vertex3f(0.187107086f, 0.537225723f, -1.f),
    C_Vertex3f(0.183850288f, 0.600541115f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0907154083f, 0.57421875f, -1.f),
    C_Vertex3f(0.0978298187f, 0.57421875f, -1.f),
    C_Vertex3f(0.098613739f, 0.631835938f, -1.f),
    C_Vertex3f(0.106585503f, 0.572259903f, -1.f),
    C_Vertex3f(0.114637375f, 0.628931999f, -1.f),
    C_Vertex3f(0.117310524f, 0.568342209f, -1.f),
    C_Vertex3f(0.133127213f, 0.623124123f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0726566315f, 0.566356659f, -1.f),
    C_Vertex3f(0.0779647827f, 0.570287704f, -1.f),
    C_Vertex3f(0.0821056366f, 0.631835938f, -1.f),
    C_Vertex3f(0.083984375f, 0.572908401f, -1.f),
    C_Vertex3f(0.098613739f, 0.631835938f, -1.f),
    C_Vertex3f(0.0907154083f, 0.57421875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0620965958f, 0.546857834f, -1.f),
    C_Vertex3f(0.0643949509f, 0.554613113f, -1.f),
    C_Vertex3f(0.0643949509f, 0.628466606f, -1.f),
    C_Vertex3f(0.0679512024f, 0.561113358f, -1.f),
    C_Vertex3f(0.0821056366f, 0.631835938f, -1.f),
    C_Vertex3f(0.0703582764f, 0.563735962f, -1.f),
    C_Vertex3f(0.0726566315f, 0.566356659f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0643949509f, 0.628466606f, -1.f),
    C_Vertex3f(0.0484323502f, 0.621727943f, -1.f),
    C_Vertex3f(0.0610561371f, 0.537847519f, -1.f),
    C_Vertex3f(0.0620965958f, 0.546857834f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.00305938721f, 0.532714844f, -1.f),
    C_Vertex3f(0.00370240211f, 0.54909873f, -1.f),
    C_Vertex3f(0.00300502777f, 0.538366318f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_153[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.213378906f, 0.284667969f, 0.f),
    C_Vertex3f(0.213378906f, 0.284667969f, -1.f),
    C_Vertex3f(0.213378906f, 0.587890625f, -1.f),
    C_Vertex3f(0.213378906f, 0.587890625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.213378906f, 0.587890625f, 0.f),
    C_Vertex3f(0.213378906f, 0.587890625f, -1.f),
    C_Vertex3f(0.0981445313f, 0.587890625f, -1.f),
    C_Vertex3f(0.0981445313f, 0.587890625f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0981445313f, 0.587890625f, 0.f),
    C_Vertex3f(0.0981445313f, 0.587890625f, -1.f),
    C_Vertex3f(0.0981445313f, 0.640625f, -1.f),
    C_Vertex3f(0.0981445313f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0981445313f, 0.640625f, 0.f),
    C_Vertex3f(0.0981445313f, 0.640625f, -1.f),
    C_Vertex3f(0.389160156f, 0.640625f, -1.f),
    C_Vertex3f(0.389160156f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.389160156f, 0.640625f, 0.f),
    C_Vertex3f(0.389160156f, 0.640625f, -1.f),
    C_Vertex3f(0.389160156f, 0.587890625f, -1.f),
    C_Vertex3f(0.389160156f, 0.587890625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.389160156f, 0.587890625f, 0.f),
    C_Vertex3f(0.389160156f, 0.587890625f, -1.f),
    C_Vertex3f(0.272949219f, 0.587890625f, -1.f),
    C_Vertex3f(0.272949219f, 0.587890625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.272949219f, 0.587890625f, 0.f),
    C_Vertex3f(0.272949219f, 0.587890625f, -1.f),
    C_Vertex3f(0.272949219f, 0.284667969f, -1.f),
    C_Vertex3f(0.272949219f, 0.284667969f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.272949219f, 0.284667969f, 0.f),
    C_Vertex3f(0.272949219f, 0.284667969f, -1.f),
    C_Vertex3f(0.213378906f, 0.284667969f, -1.f),
    C_Vertex3f(0.213378906f, 0.284667969f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.428710938f, 0.284667969f, 0.f),
    C_Vertex3f(0.428710938f, 0.284667969f, -1.f),
    C_Vertex3f(0.428710938f, 0.640625f, -1.f),
    C_Vertex3f(0.428710938f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.428710938f, 0.640625f, 0.f),
    C_Vertex3f(0.428710938f, 0.640625f, -1.f),
    C_Vertex3f(0.516136169f, 0.640625f, -1.f),
    C_Vertex3f(0.516136169f, 0.640625f, 0.f),
    C_Normal3f(0.953329504f, 0.301932037f, 0.f),
    C_Vertex3f(0.516136169f, 0.640625f, 0.f),
    C_Vertex3f(0.516136169f, 0.640625f, -1.f),
    C_Vertex3f(0.606178284f, 0.356323242f, -1.f),
    C_Vertex3f(0.606178284f, 0.356323242f, 0.f),
    C_Normal3f(-0.956259906f, 0.292518437f, 0.f),
    C_Vertex3f(0.606178284f, 0.356323242f, 0.f),
    C_Vertex3f(0.606178284f, 0.356323242f, -1.f),
    C_Vertex3f(0.693145752f, 0.640625f, -1.f),
    C_Vertex3f(0.693145752f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.693145752f, 0.640625f, 0.f),
    C_Vertex3f(0.693145752f, 0.640625f, -1.f),
    C_Vertex3f(0.778808594f, 0.640625f, -1.f),
    C_Vertex3f(0.778808594f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.778808594f, 0.640625f, 0.f),
    C_Vertex3f(0.778808594f, 0.640625f, -1.f),
    C_Vertex3f(0.778808594f, 0.284667969f, -1.f),
    C_Vertex3f(0.778808594f, 0.284667969f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.778808594f, 0.284667969f, 0.f),
    C_Vertex3f(0.778808594f, 0.284667969f, -1.f),
    C_Vertex3f(0.724609375f, 0.284667969f, -1.f),
    C_Vertex3f(0.724609375f, 0.284667969f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.724609375f, 0.284667969f, 0.f),
    C_Vertex3f(0.724609375f, 0.284667969f, -1.f),
    C_Vertex3f(0.724609375f, 0.583518982f, -1.f),
    C_Vertex3f(0.724609375f, 0.583518982f, 0.f),
    C_Normal3f(0.955920339f, -0.293626249f, 0.f),
    C_Vertex3f(0.724609375f, 0.583518982f, 0.f),
    C_Vertex3f(0.724609375f, 0.583518982f, -1.f),
    C_Vertex3f(0.6328125f, 0.284667969f, -1.f),
    C_Vertex3f(0.6328125f, 0.284667969f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.6328125f, 0.284667969f, 0.f),
    C_Vertex3f(0.6328125f, 0.284667969f, -1.f),
    C_Vertex3f(0.579101563f, 0.284667969f, -1.f),
    C_Vertex3f(0.579101563f, 0.284667969f, 0.f),
    C_Normal3f(-0.953629017f, -0.300984353f, 0.f),
    C_Vertex3f(0.579101563f, 0.284667969f, 0.f),
    C_Vertex3f(0.579101563f, 0.284667969f, -1.f),
    C_Vertex3f(0.483398438f, 0.587890625f, -1.f),
    C_Vertex3f(0.483398438f, 0.587890625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.483398438f, 0.587890625f, 0.f),
    C_Vertex3f(0.483398438f, 0.587890625f, -1.f),
    C_Vertex3f(0.483398438f, 0.284667969f, -1.f),
    C_Vertex3f(0.483398438f, 0.284667969f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.483398438f, 0.284667969f, 0.f),
    C_Vertex3f(0.483398438f, 0.284667969f, -1.f),
    C_Vertex3f(0.428710938f, 0.284667969f, -1.f),
    C_Vertex3f(0.428710938f, 0.284667969f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213378906f, 0.284667969f, 0.f),
    C_Vertex3f(0.272949219f, 0.284667969f, 0.f),
    C_Vertex3f(0.213378906f, 0.587890625f, 0.f),
    C_Vertex3f(0.272949219f, 0.587890625f, 0.f),
    C_Vertex3f(0.389160156f, 0.640625f, 0.f),
    C_Vertex3f(0.389160156f, 0.587890625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0981445313f, 0.640625f, 0.f),
    C_Vertex3f(0.0981445313f, 0.587890625f, 0.f),
    C_Vertex3f(0.213378906f, 0.587890625f, 0.f),
    C_Vertex3f(0.389160156f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.428710938f, 0.284667969f, 0.f),
    C_Vertex3f(0.483398438f, 0.284667969f, 0.f),
    C_Vertex3f(0.428710938f, 0.640625f, 0.f),
    C_Vertex3f(0.483398438f, 0.587890625f, 0.f),
    C_Vertex3f(0.516136169f, 0.640625f, 0.f),
    C_Vertex3f(0.579101563f, 0.284667969f, 0.f),
    C_Vertex3f(0.606178284f, 0.356323242f, 0.f),
    C_Vertex3f(0.6328125f, 0.284667969f, 0.f),
    C_Vertex3f(0.693145752f, 0.640625f, 0.f),
    C_Vertex3f(0.724609375f, 0.583518982f, 0.f),
    C_Vertex3f(0.778808594f, 0.640625f, 0.f),
    C_Vertex3f(0.778808594f, 0.284667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.778808594f, 0.284667969f, 0.f),
    C_Vertex3f(0.724609375f, 0.583518982f, 0.f),
    C_Vertex3f(0.724609375f, 0.284667969f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213378906f, 0.284667969f, -1.f),
    C_Vertex3f(0.272949219f, 0.284667969f, -1.f),
    C_Vertex3f(0.213378906f, 0.587890625f, -1.f),
    C_Vertex3f(0.272949219f, 0.587890625f, -1.f),
    C_Vertex3f(0.389160156f, 0.640625f, -1.f),
    C_Vertex3f(0.389160156f, 0.587890625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0981445313f, 0.640625f, -1.f),
    C_Vertex3f(0.0981445313f, 0.587890625f, -1.f),
    C_Vertex3f(0.213378906f, 0.587890625f, -1.f),
    C_Vertex3f(0.389160156f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.428710938f, 0.284667969f, -1.f),
    C_Vertex3f(0.483398438f, 0.284667969f, -1.f),
    C_Vertex3f(0.428710938f, 0.640625f, -1.f),
    C_Vertex3f(0.483398438f, 0.587890625f, -1.f),
    C_Vertex3f(0.516136169f, 0.640625f, -1.f),
    C_Vertex3f(0.579101563f, 0.284667969f, -1.f),
    C_Vertex3f(0.606178284f, 0.356323242f, -1.f),
    C_Vertex3f(0.6328125f, 0.284667969f, -1.f),
    C_Vertex3f(0.693145752f, 0.640625f, -1.f),
    C_Vertex3f(0.724609375f, 0.583518982f, -1.f),
    C_Vertex3f(0.778808594f, 0.640625f, -1.f),
    C_Vertex3f(0.778808594f, 0.284667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.778808594f, 0.284667969f, -1.f),
    C_Vertex3f(0.724609375f, 0.583518982f, -1.f),
    C_Vertex3f(0.724609375f, 0.284667969f, -1.f),
    C_End(),
    C_Translatef(0.895019531f, 0.f, 0.f),
};
static const Call char_154[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.155324444f, 0.987863481f, 0.f),
    C_Vertex3f(0.02734375f, 0.138671875f, 0.f),
    C_Vertex3f(0.02734375f, 0.138671875f, -1.f),
    C_Vertex3f(0.104980469f, 0.150878906f, -1.f),
    C_Vertex3f(0.104980469f, 0.150878906f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.990272939f, 0.139138594f, 0.f),
    C_Vertex3f(0.104980469f, 0.150878906f, 0.f),
    C_Vertex3f(0.104980469f, 0.150878906f, -1.f),
    C_Normal3f(0.982487202f, 0.186330065f, 0.f),
    C_Vertex3f(0.105802536f, 0.145028114f, 0.f),
    C_Vertex3f(0.105802536f, 0.145028114f, -1.f),
    C_Normal3f(0.968530297f, 0.248895541f, 0.f),
    C_Vertex3f(0.108179688f, 0.134009957f, 0.f),
    C_Vertex3f(0.108179688f, 0.134009957f, -1.f),
    C_Normal3f(0.944471955f, 0.328592092f, 0.f),
    C_Vertex3f(0.111289859f, 0.123675227f, 0.f),
    C_Vertex3f(0.111289859f, 0.123675227f, -1.f),
    C_Normal3f(0.911173463f, 0.412022918f, 0.f),
    C_Vertex3f(0.115133047f, 0.114023924f, 0.f),
    C_Vertex3f(0.115133047f, 0.114023924f, -1.f),
    C_Normal3f(0.867658496f, 0.497160673f, 0.f),
    C_Vertex3f(0.119709253f, 0.105056047f, 0.f),
    C_Vertex3f(0.119709253f, 0.105056047f, -1.f),
    C_Normal3f(0.813619614f, 0.581397593f, 0.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, 0.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, -1.f),
    C_Normal3f(0.749659598f, 0.661823511f, 0.f),
    C_Vertex3f(0.131060719f, 0.0891705751f, 0.f),
    C_Vertex3f(0.131060719f, 0.0891705751f, -1.f),
    C_Normal3f(0.677360773f, 0.735650957f, 0.f),
    C_Vertex3f(0.137835979f, 0.0822529793f, 0.f),
    C_Vertex3f(0.137835979f, 0.0822529793f, -1.f),
    C_Normal3f(0.59564364f, 0.803248763f, 0.f),
    C_Vertex3f(0.145344257f, 0.0760188103f, 0.f),
    C_Vertex3f(0.145344257f, 0.0760188103f, -1.f),
    C_Normal3f(0.506266713f, 0.862376928f, 0.f),
    C_Vertex3f(0.153599143f, 0.0705639124f, 0.f),
    C_Vertex3f(0.153599143f, 0.0705639124f, -1.f),
    C_Normal3f(0.415789396f, 0.909460902f, 0.f),
    C_Vertex3f(0.162600636f, 0.0658882856f, 0.f),
    C_Vertex3f(0.162600636f, 0.0658882856f, -1.f),
    C_Normal3f(0.327373713f, 0.94489491f, 0.f),
    C_Vertex3f(0.172348738f, 0.06199193f, 0.f),
    C_Vertex3f(0.172348738f, 0.06199193f, -1.f),
    C_Normal3f(0.243412912f, 0.969922721f, 0.f),
    C_Vertex3f(0.182843447f, 0.0588748455f, 0.f),
    C_Vertex3f(0.182843447f, 0.0588748455f, -1.f),
    C_Normal3f(0.165423974f, 0.986222446f, 0.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, 0.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, -1.f),
    C_Normal3f(0.0941423625f, 0.995558739f, 0.f),
    C_Vertex3f(0.206072688f, 0.0549784899f, 0.f),
    C_Vertex3f(0.206072688f, 0.0549784899f, -1.f),
    C_Normal3f(0.0296508372f, 0.999560237f, 0.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, 0.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0263045318f, 0.999653935f, 0.f),
    C_Vertex3f(0.23234272f, 0.0541992188f, 0.f),
    C_Vertex3f(0.23234272f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0841049999f, 0.996456981f, 0.f),
    C_Vertex3f(0.245055914f, 0.0548899174f, 0.f),
    C_Vertex3f(0.245055914f, 0.0548899174f, -1.f),
    C_Normal3f(-0.149783954f, 0.988718808f, 0.f),
    C_Vertex3f(0.256892443f, 0.0562713146f, 0.f),
    C_Vertex3f(0.256892443f, 0.0562713146f, -1.f),
    C_Normal3f(-0.223927125f, 0.974605858f, 0.f),
    C_Vertex3f(0.267852306f, 0.0583434105f, 0.f),
    C_Vertex3f(0.267852306f, 0.0583434105f, -1.f),
    C_Normal3f(-0.306725383f, 0.951798081f, 0.f),
    C_Vertex3f(0.277935505f, 0.061106205f, 0.f),
    C_Vertex3f(0.277935505f, 0.061106205f, -1.f),
    C_Normal3f(-0.397544771f, 0.91758281f, 0.f),
    C_Vertex3f(0.287142038f, 0.0645596981f, 0.f),
    C_Vertex3f(0.287142038f, 0.0645596981f, -1.f),
    C_Normal3f(-0.49448508f, 0.869186103f, 0.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, 0.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, -1.f),
    C_Normal3f(-0.603962064f, 0.797013104f, 0.f),
    C_Vertex3f(0.30292511f, 0.0735387802f, 0.f),
    C_Vertex3f(0.30292511f, 0.0735387802f, -1.f),
    C_Normal3f(-0.717075408f, 0.696995676f, 0.f),
    C_Vertex3f(0.312789917f, 0.0818271637f, 0.f),
    C_Vertex3f(0.312789917f, 0.0818271637f, -1.f),
    C_Normal3f(-0.835284114f, 0.549818575f, 0.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, 0.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, -1.f),
    C_Normal3f(-0.938837171f, 0.344361275f, 0.f),
    C_Vertex3f(0.329231262f, 0.106804848f, 0.f),
    C_Vertex3f(0.329231262f, 0.106804848f, -1.f),
    C_Normal3f(-0.993195236f, 0.116461053f, 0.f),
    C_Vertex3f(0.332519531f, 0.120731354f, 0.f),
    C_Vertex3f(0.332519531f, 0.120731354f, -1.f),
    C_Normal3f(-0.993828118f, -0.110930823f, 0.f),
    C_Vertex3f(0.332519531f, 0.134847641f, 0.f),
    C_Vertex3f(0.332519531f, 0.134847641f, -1.f),
    C_Normal3f(-0.932035148f, -0.362367868f, 0.f),
    C_Vertex3f(0.329612732f, 0.146773338f, 0.f),
    C_Vertex3f(0.329612732f, 0.146773338f, -1.f),
    C_Normal3f(-0.80304426f, -0.595919311f, 0.f),
    C_Vertex3f(0.323799133f, 0.157277107f, 0.f),
    C_Vertex3f(0.323799133f, 0.157277107f, -1.f),
    C_Normal3f(-0.651524901f, -0.758627295f, 0.f),
    C_Vertex3f(0.315078735f, 0.166358948f, 0.f),
    C_Vertex3f(0.315078735f, 0.166358948f, -1.f),
    C_Normal3f(-0.509991109f, -0.860179663f, 0.f),
    C_Vertex3f(0.307235718f, 0.171502113f, 0.f),
    C_Vertex3f(0.307235718f, 0.171502113f, -1.f),
    C_Normal3f(-0.41662693f, -0.909077585f, 0.f),
    C_Vertex3f(0.301674128f, 0.174306393f, 0.f),
    C_Vertex3f(0.301674128f, 0.174306393f, -1.f),
    C_Normal3f(-0.365715176f, -0.930726826f, 0.f),
    C_Vertex3f(0.294609785f, 0.177288532f, 0.f),
    C_Vertex3f(0.294609785f, 0.177288532f, -1.f),
    C_Normal3f(-0.329218626f, -0.944253683f, 0.f),
    C_Vertex3f(0.28604269f, 0.180448532f, 0.f),
    C_Vertex3f(0.28604269f, 0.180448532f, -1.f),
    C_Normal3f(-0.301897258f, -0.953340411f, 0.f),
    C_Vertex3f(0.275972843f, 0.183786392f, 0.f),
    C_Vertex3f(0.275972843f, 0.183786392f, -1.f),
    C_Normal3f(-0.280728996f, -0.95978713f, 0.f),
    C_Vertex3f(0.264400244f, 0.187302113f, 0.f),
    C_Vertex3f(0.264400244f, 0.187302113f, -1.f),
    C_Normal3f(-0.263869554f, -0.964558363f, 0.f),
    C_Vertex3f(0.251324892f, 0.190995693f, 0.f),
    C_Vertex3f(0.251324892f, 0.190995693f, -1.f),
    C_Normal3f(-0.249712259f, -0.968320072f, 0.f),
    C_Vertex3f(0.236746788f, 0.194867134f, 0.f),
    C_Vertex3f(0.236746788f, 0.194867134f, -1.f),
    C_Normal3f(-0.249843717f, -0.968286216f, 0.f),
    C_Vertex3f(0.217877388f, 0.199621201f, 0.f),
    C_Vertex3f(0.217877388f, 0.199621201f, -1.f),
    C_Normal3f(-0.261011988f, -0.965335548f, 0.f),
    C_Vertex3f(0.197880268f, 0.204895735f, 0.f),
    C_Vertex3f(0.197880268f, 0.204895735f, -1.f),
    C_Normal3f(-0.274538636f, -0.961576045f, 0.f),
    C_Vertex3f(0.179543972f, 0.209985971f, 0.f),
    C_Vertex3f(0.179543972f, 0.209985971f, -1.f),
    C_Normal3f(-0.290685207f, -0.956818759f, 0.f),
    C_Vertex3f(0.1628685f, 0.214891911f, 0.f),
    C_Vertex3f(0.1628685f, 0.214891911f, -1.f),
    C_Normal3f(-0.310274899f, -0.950646877f, 0.f),
    C_Vertex3f(0.147853851f, 0.219613552f, 0.f),
    C_Vertex3f(0.147853851f, 0.219613552f, -1.f),
    C_Normal3f(-0.334503651f, -0.942394435f, 0.f),
    C_Vertex3f(0.134500027f, 0.224150896f, 0.f),
    C_Vertex3f(0.134500027f, 0.224150896f, -1.f),
    C_Normal3f(-0.365166187f, -0.930942416f, 0.f),
    C_Vertex3f(0.122807026f, 0.228503942f, 0.f),
    C_Vertex3f(0.122807026f, 0.228503942f, -1.f),
    C_Normal3f(-0.40970242f, -0.912219226f, 0.f),
    C_Vertex3f(0.112774849f, 0.232672691f, 0.f),
    C_Vertex3f(0.112774849f, 0.232672691f, -1.f),
    C_Normal3f(-0.493446022f, -0.869776368f, 0.f),
    C_Vertex3f(0.100217819f, 0.238649368f, 0.f),
    C_Vertex3f(0.100217819f, 0.238649368f, -1.f),
    C_Normal3f(-0.599326909f, -0.800504327f, 0.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, 0.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, -1.f),
    C_Normal3f(-0.710367441f, -0.703831077f, 0.f),
    C_Vertex3f(0.0728063583f, 0.259171963f, 0.f),
    C_Vertex3f(0.0728063583f, 0.259171963f, -1.f),
    C_Normal3f(-0.809581339f, -0.587007701f, 0.f),
    C_Vertex3f(0.0621376038f, 0.271725655f, 0.f),
    C_Vertex3f(0.0621376038f, 0.271725655f, -1.f),
    C_Normal3f(-0.885973155f, -0.463736624f, 0.f),
    C_Vertex3f(0.0534934998f, 0.28580761f, 0.f),
    C_Vertex3f(0.0534934998f, 0.28580761f, -1.f),
    C_Normal3f(-0.942161143f, -0.335160226f, 0.f),
    C_Vertex3f(0.0470104218f, 0.300626278f, 0.f),
    C_Vertex3f(0.0470104218f, 0.300626278f, -1.f),
    C_Normal3f(-0.979902864f, -0.199475348f, 0.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, 0.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, -1.f),
    C_Normal3f(-0.997804403f, -0.0662289932f, 0.f),
    C_Vertex3f(0.0405273438f, 0.332473755f, 0.f),
    C_Vertex3f(0.0405273438f, 0.332473755f, -1.f),
    C_Normal3f(-0.998378634f, 0.0569228902f, 0.f),
    C_Vertex3f(0.0405273438f, 0.348739624f, 0.f),
    C_Vertex3f(0.0405273438f, 0.348739624f, -1.f),
    C_Normal3f(-0.983803272f, 0.179251954f, 0.f),
    C_Vertex3f(0.0423059464f, 0.363668919f, 0.f),
    C_Vertex3f(0.0423059464f, 0.363668919f, -1.f),
    C_Normal3f(-0.953510284f, 0.30136025f, 0.f),
    C_Vertex3f(0.0458631516f, 0.378024578f, 0.f),
    C_Vertex3f(0.0458631516f, 0.378024578f, -1.f),
    C_Normal3f(-0.908047616f, 0.418866903f, 0.f),
    C_Vertex3f(0.0511989594f, 0.391806602f, 0.f),
    C_Vertex3f(0.0511989594f, 0.391806602f, -1.f),
    C_Normal3f(-0.852469385f, 0.5227772f, 0.f),
    C_Vertex3f(0.0583133698f, 0.405014992f, 0.f),
    C_Vertex3f(0.0583133698f, 0.405014992f, -1.f),
    C_Normal3f(-0.786895573f, 0.617086232f, 0.f),
    C_Vertex3f(0.0667133331f, 0.417105198f, 0.f),
    C_Vertex3f(0.0667133331f, 0.417105198f, -1.f),
    C_Normal3f(-0.709986389f, 0.704215467f, 0.f),
    C_Vertex3f(0.0763988495f, 0.428077221f, 0.f),
    C_Vertex3f(0.0763988495f, 0.428077221f, -1.f),
    C_Normal3f(-0.63944602f, 0.768836021f, 0.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, 0.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, -1.f),
    C_Normal3f(-0.585769415f, 0.810477734f, 0.f),
    C_Vertex3f(0.09349823f, 0.442298889f, 0.f),
    C_Vertex3f(0.09349823f, 0.442298889f, -1.f),
    C_Normal3f(-0.528840721f, 0.848721147f, 0.f),
    C_Vertex3f(0.0980949402f, 0.445682526f, 0.f),
    C_Vertex3f(0.0980949402f, 0.445682526f, -1.f),
    C_Normal3f(-0.455799788f, 0.8900823f, 0.f),
    C_Vertex3f(0.108957291f, 0.451931477f, 0.f),
    C_Vertex3f(0.108957291f, 0.451931477f, -1.f),
    C_Normal3f(-0.378850102f, 0.925458074f, 0.f),
    C_Vertex3f(0.121488571f, 0.457662106f, 0.f),
    C_Vertex3f(0.121488571f, 0.457662106f, -1.f),
    C_Normal3f(-0.312904745f, 0.949784517f, 0.f),
    C_Vertex3f(0.135688782f, 0.462874413f, 0.f),
    C_Vertex3f(0.135688782f, 0.462874413f, -1.f),
    C_Normal3f(-0.246383533f, 0.969172418f, 0.f),
    C_Vertex3f(0.151557922f, 0.467568398f, 0.f),
    C_Vertex3f(0.151557922f, 0.467568398f, -1.f),
    C_Normal3f(-0.172719449f, 0.984971046f, 0.f),
    C_Vertex3f(0.168001175f, 0.471088886f, 0.f),
    C_Vertex3f(0.168001175f, 0.471088886f, -1.f),
    C_Normal3f(-0.101200007f, 0.994866133f, 0.f),
    C_Vertex3f(0.185018539f, 0.473435879f, 0.f),
    C_Vertex3f(0.185018539f, 0.473435879f, -1.f),
    C_Normal3f(-0.0290665496f, 0.999577463f, 0.f),
    C_Vertex3f(0.202610016f, 0.474609375f, 0.f),
    C_Vertex3f(0.202610016f, 0.474609375f, -1.f),
    C_Normal3f(0.0405974723f, 0.999175668f, 0.f),
    C_Vertex3f(0.225374222f, 0.474609375f, 0.f),
    C_Vertex3f(0.225374222f, 0.474609375f, -1.f),
    C_Normal3f(0.117976762f, 0.993016362f, 0.f),
    C_Vertex3f(0.251067162f, 0.472640514f, 0.f),
    C_Vertex3f(0.251067162f, 0.472640514f, -1.f),
    C_Normal3f(0.207643896f, 0.978204429f, 0.f),
    C_Vertex3f(0.275090218f, 0.468702793f, 0.f),
    C_Vertex3f(0.275090218f, 0.468702793f, -1.f),
    C_Normal3f(0.304983616f, 0.95235759f, 0.f),
    C_Vertex3f(0.29744339f, 0.462796211f, 0.f),
    C_Vertex3f(0.29744339f, 0.462796211f, -1.f),
    C_Normal3f(0.405418575f, 0.914131165f, 0.f),
    C_Vertex3f(0.318126678f, 0.454920769f, 0.f),
    C_Vertex3f(0.318126678f, 0.454920769f, -1.f),
    C_Normal3f(0.51426959f, 0.857628584f, 0.f),
    C_Vertex3f(0.336102009f, 0.445651054f, 0.f),
    C_Vertex3f(0.336102009f, 0.445651054f, -1.f),
    C_Normal3f(0.632485092f, 0.774572551f, 0.f),
    C_Vertex3f(0.351369381f, 0.434987068f, 0.f),
    C_Vertex3f(0.351369381f, 0.434987068f, -1.f),
    C_Normal3f(0.751277328f, 0.659986556f, 0.f),
    C_Vertex3f(0.363928795f, 0.42292881f, 0.f),
    C_Vertex3f(0.363928795f, 0.42292881f, -1.f),
    C_Normal3f(0.848835111f, 0.528657675f, 0.f),
    C_Vertex3f(0.373780251f, 0.40947628f, 0.f),
    C_Vertex3f(0.373780251f, 0.40947628f, -1.f),
    C_Normal3f(0.91259867f, 0.408856481f, 0.f),
    C_Vertex3f(0.382098675f, 0.393754482f, 0.f),
    C_Vertex3f(0.382098675f, 0.393754482f, -1.f),
    C_Normal3f(0.95388037f, 0.300187111f, 0.f),
    C_Vertex3f(0.388884068f, 0.375763416f, 0.f),
    C_Vertex3f(0.388884068f, 0.375763416f, -1.f),
    C_Normal3f(0.97548455f, 0.220067948f, 0.f),
    C_Vertex3f(0.394136429f, 0.355503082f, 0.f),
    C_Vertex3f(0.394136429f, 0.355503082f, -1.f),
    C_Normal3f(0.986645639f, 0.162881732f, 0.f),
    C_Vertex3f(0.395996094f, 0.344238281f, 0.f),
    C_Vertex3f(0.395996094f, 0.344238281f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.131752744f, -0.991282582f, 0.f),
    C_Vertex3f(0.395996094f, 0.344238281f, 0.f),
    C_Vertex3f(0.395996094f, 0.344238281f, -1.f),
    C_Vertex3f(0.318847656f, 0.333984375f, -1.f),
    C_Vertex3f(0.318847656f, 0.333984375f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.989477277f, -0.144688696f, 0.f),
    C_Vertex3f(0.318847656f, 0.333984375f, 0.f),
    C_Vertex3f(0.318847656f, 0.333984375f, -1.f),
    C_Normal3f(-0.968403697f, -0.249387681f, 0.f),
    C_Vertex3f(0.3175354f, 0.34295845f, 0.f),
    C_Vertex3f(0.3175354f, 0.34295845f, -1.f),
    C_Normal3f(-0.920641422f, -0.390409231f, 0.f),
    C_Vertex3f(0.312421799f, 0.358936787f, 0.f),
    C_Vertex3f(0.312421799f, 0.358936787f, -1.f),
    C_Normal3f(-0.827186882f, -0.561926782f, 0.f),
    C_Vertex3f(0.304819107f, 0.372945309f, 0.f),
    C_Vertex3f(0.304819107f, 0.372945309f, -1.f),
    C_Normal3f(-0.6981287f, -0.715972304f, 0.f),
    C_Vertex3f(0.294727325f, 0.384984016f, 0.f),
    C_Vertex3f(0.294727325f, 0.384984016f, -1.f),
    C_Normal3f(-0.535922647f, -0.84426707f, 0.f),
    C_Vertex3f(0.282146454f, 0.39505291f, 0.f),
    C_Vertex3f(0.282146454f, 0.39505291f, -1.f),
    C_Normal3f(-0.356808603f, -0.934177518f, 0.f),
    C_Vertex3f(0.266968727f, 0.40260458f, 0.f),
    C_Vertex3f(0.266968727f, 0.40260458f, -1.f),
    C_Normal3f(-0.194198593f, -0.980962276f, 0.f),
    C_Vertex3f(0.249194145f, 0.407639027f, 0.f),
    C_Vertex3f(0.249194145f, 0.407639027f, -1.f),
    C_Normal3f(-0.0650108978f, -0.997884631f, 0.f),
    C_Vertex3f(0.228822708f, 0.41015625f, 0.f),
    C_Vertex3f(0.228822708f, 0.41015625f, -1.f),
    C_Normal3f(0.0181715135f, -0.999834895f, 0.f),
    C_Vertex3f(0.21055603f, 0.41015625f, 0.f),
    C_Vertex3f(0.21055603f, 0.41015625f, -1.f),
    C_Normal3f(0.069313921f, -0.997594953f, 0.f),
    C_Vertex3f(0.197961926f, 0.40959537f, 0.f),
    C_Vertex3f(0.197961926f, 0.40959537f, -1.f),
    C_Normal3f(0.124911025f, -0.99216789f, 0.f),
    C_Vertex3f(0.186338782f, 0.408473611f, 0.f),
    C_Vertex3f(0.186338782f, 0.408473611f, -1.f),
    C_Normal3f(0.189587131f, -0.981863916f, 0.f),
    C_Vertex3f(0.175686598f, 0.406790972f, 0.f),
    C_Vertex3f(0.175686598f, 0.406790972f, -1.f),
    C_Normal3f(0.264681518f, -0.964335859f, 0.f),
    C_Vertex3f(0.166005373f, 0.404547453f, 0.f),
    C_Vertex3f(0.166005373f, 0.404547453f, -1.f),
    C_Normal3f(0.351177722f, -0.936308861f, 0.f),
    C_Vertex3f(0.157295108f, 0.401743054f, 0.f),
    C_Vertex3f(0.157295108f, 0.401743054f, -1.f),
    C_Normal3f(0.449065775f, -0.893498778f, 0.f),
    C_Vertex3f(0.149555802f, 0.398377776f, 0.f),
    C_Vertex3f(0.149555802f, 0.398377776f, -1.f),
    C_Normal3f(0.567429543f, -0.823421896f, 0.f),
    C_Vertex3f(0.142787457f, 0.394451618f, 0.f),
    C_Vertex3f(0.142787457f, 0.394451618f, -1.f),
    C_Normal3f(0.687615454f, -0.726075113f, 0.f),
    C_Vertex3f(0.134091377f, 0.387721062f, 0.f),
    C_Vertex3f(0.134091377f, 0.387721062f, -1.f),
    C_Normal3f(0.813137829f, -0.582071185f, 0.f),
    C_Vertex3f(0.125395298f, 0.377980709f, 0.f),
    C_Vertex3f(0.125395298f, 0.377980709f, -1.f),
    C_Normal3f(0.928707838f, -0.370812148f, 0.f),
    C_Vertex3f(0.119597912f, 0.367474079f, 0.f),
    C_Vertex3f(0.119597912f, 0.367474079f, -1.f),
    C_Normal3f(0.990716219f, -0.135946453f, 0.f),
    C_Vertex3f(0.116699219f, 0.356201172f, 0.f),
    C_Vertex3f(0.116699219f, 0.356201172f, -1.f),
    C_Normal3f(0.99754113f, 0.0700827762f, 0.f),
    C_Vertex3f(0.116699219f, 0.346349716f, 0.f),
    C_Vertex3f(0.116699219f, 0.346349716f, -1.f),
    C_Normal3f(0.969055176f, 0.246844038f, 0.f),
    C_Vertex3f(0.117902756f, 0.33907032f, 0.f),
    C_Vertex3f(0.117902756f, 0.33907032f, -1.f),
    C_Normal3f(0.912301242f, 0.409519881f, 0.f),
    C_Vertex3f(0.12030983f, 0.332175255f, 0.f),
    C_Vertex3f(0.12030983f, 0.332175255f, -1.f),
    C_Normal3f(0.846645236f, 0.532157719f, 0.f),
    C_Vertex3f(0.123920441f, 0.32566452f, 0.f),
    C_Vertex3f(0.123920441f, 0.32566452f, -1.f),
    C_Normal3f(0.791732073f, 0.610868514f, 0.f),
    C_Vertex3f(0.126327515f, 0.322601318f, 0.f),
    C_Vertex3f(0.126327515f, 0.322601318f, -1.f),
    C_Normal3f(0.722636878f, 0.691227913f, 0.f),
    C_Vertex3f(0.128732681f, 0.31942749f, 0.f),
    C_Vertex3f(0.128732681f, 0.31942749f, -1.f),
    C_Normal3f(0.627186358f, 0.778869152f, 0.f),
    C_Vertex3f(0.134911537f, 0.313627243f, 0.f),
    C_Vertex3f(0.134911537f, 0.313627243f, -1.f),
    C_Normal3f(0.517561138f, 0.855646193f, 0.f),
    C_Vertex3f(0.142458916f, 0.308374405f, 0.f),
    C_Vertex3f(0.142458916f, 0.308374405f, -1.f),
    C_Normal3f(0.434632897f, 0.900607765f, 0.f),
    C_Vertex3f(0.151374817f, 0.303668976f, 0.f),
    C_Vertex3f(0.151374817f, 0.303668976f, -1.f),
    C_Normal3f(0.368849158f, 0.929489195f, 0.f),
    C_Vertex3f(0.156517029f, 0.301589966f, 0.f),
    C_Vertex3f(0.156517029f, 0.301589966f, -1.f),
    C_Normal3f(0.322466284f, 0.946580887f, 0.f),
    C_Vertex3f(0.157993317f, 0.301042557f, 0.f),
    C_Vertex3f(0.157993317f, 0.301042557f, -1.f),
    C_Normal3f(0.303810388f, 0.952732503f, 0.f),
    C_Vertex3f(0.162381887f, 0.299592018f, 0.f),
    C_Vertex3f(0.162381887f, 0.299592018f, -1.f),
    C_Normal3f(0.290211618f, 0.956962526f, 0.f),
    C_Vertex3f(0.168206453f, 0.297785759f, 0.f),
    C_Vertex3f(0.168206453f, 0.297785759f, -1.f),
    C_Normal3f(0.281414062f, 0.959586382f, 0.f),
    C_Vertex3f(0.175467014f, 0.295623779f, 0.f),
    C_Vertex3f(0.175467014f, 0.295623779f, -1.f),
    C_Normal3f(0.275258064f, 0.961370349f, 0.f),
    C_Vertex3f(0.18416357f, 0.293106079f, 0.f),
    C_Vertex3f(0.18416357f, 0.293106079f, -1.f),
    C_Normal3f(0.270709753f, 0.962661028f, 0.f),
    C_Vertex3f(0.194296122f, 0.290232658f, 0.f),
    C_Vertex3f(0.194296122f, 0.290232658f, -1.f),
    C_Normal3f(0.267212242f, 0.96363765f, 0.f),
    C_Vertex3f(0.205864668f, 0.287003517f, 0.f),
    C_Vertex3f(0.205864668f, 0.287003517f, -1.f),
    C_Normal3f(0.262356311f, 0.964971066f, 0.f),
    C_Vertex3f(0.218869209f, 0.283418655f, 0.f),
    C_Vertex3f(0.218869209f, 0.283418655f, -1.f),
    C_Normal3f(0.262702644f, 0.964876831f, 0.f),
    C_Vertex3f(0.236536026f, 0.278664589f, 0.f),
    C_Vertex3f(0.236536026f, 0.278664589f, -1.f),
    C_Normal3f(0.269685537f, 0.962948442f, 0.f),
    C_Vertex3f(0.255851984f, 0.273349524f, 0.f),
    C_Vertex3f(0.255851984f, 0.273349524f, -1.f),
    C_Normal3f(0.279609233f, 0.960113943f, 0.f),
    C_Vertex3f(0.273590803f, 0.268286943f, 0.f),
    C_Vertex3f(0.273590803f, 0.268286943f, -1.f),
    C_Normal3f(0.291450858f, 0.956585765f, 0.f),
    C_Vertex3f(0.289752483f, 0.263476849f, 0.f),
    C_Vertex3f(0.289752483f, 0.263476849f, -1.f),
    C_Normal3f(0.30581823f, 0.952089965f, 0.f),
    C_Vertex3f(0.304337025f, 0.258919239f, 0.f),
    C_Vertex3f(0.304337025f, 0.258919239f, -1.f),
    C_Normal3f(0.323601365f, 0.946193576f, 0.f),
    C_Vertex3f(0.317344427f, 0.254614115f, 0.f),
    C_Vertex3f(0.317344427f, 0.254614115f, -1.f),
    C_Normal3f(0.346153408f, 0.938177943f, 0.f),
    C_Vertex3f(0.328774691f, 0.250561476f, 0.f),
    C_Vertex3f(0.328774691f, 0.250561476f, -1.f),
    C_Normal3f(0.379073739f, 0.925366402f, 0.f),
    C_Vertex3f(0.338627815f, 0.246761322f, 0.f),
    C_Vertex3f(0.338627815f, 0.246761322f, -1.f),
    C_Normal3f(0.458562732f, 0.88866204f, 0.f),
    C_Vertex3f(0.351041794f, 0.241439819f, 0.f),
    C_Vertex3f(0.351041794f, 0.241439819f, -1.f),
    C_Normal3f(0.565282106f, 0.824897587f, 0.f),
    C_Vertex3f(0.365812302f, 0.232733727f, 0.f),
    C_Vertex3f(0.365812302f, 0.232733727f, -1.f),
    C_Normal3f(0.676785588f, 0.736180186f, 0.f),
    C_Vertex3f(0.378801346f, 0.222416878f, 0.f),
    C_Vertex3f(0.378801346f, 0.222416878f, -1.f),
    C_Normal3f(0.776970387f, 0.629537165f, 0.f),
    C_Vertex3f(0.390008926f, 0.210489273f, 0.f),
    C_Vertex3f(0.390008926f, 0.210489273f, -1.f),
    C_Normal3f(0.867101312f, 0.498131841f, 0.f),
    C_Vertex3f(0.399435043f, 0.196950912f, 0.f),
    C_Vertex3f(0.399435043f, 0.196950912f, -1.f),
    C_Normal3f(0.93836987f, 0.345632762f, 0.f),
    C_Vertex3f(0.406504631f, 0.181775093f, 0.f),
    C_Vertex3f(0.406504631f, 0.181775093f, -1.f),
    C_Normal3f(0.980490804f, 0.196564838f, 0.f),
    C_Vertex3f(0.41121769f, 0.164961815f, 0.f),
    C_Vertex3f(0.41121769f, 0.164961815f, -1.f),
    C_Normal3f(0.998114407f, 0.0613807701f, 0.f),
    C_Vertex3f(0.413574219f, 0.146511078f, 0.f),
    C_Vertex3f(0.413574219f, 0.146511078f, -1.f),
    C_Normal3f(0.997288287f, -0.0735932663f, 0.f),
    C_Vertex3f(0.413574219f, 0.126642227f, 0.f),
    C_Vertex3f(0.413574219f, 0.126642227f, -1.f),
    C_Normal3f(0.974677503f, -0.223615587f, 0.f),
    C_Vertex3f(0.410700321f, 0.10756588f, 0.f),
    C_Vertex3f(0.410700321f, 0.10756588f, -1.f),
    C_Normal3f(0.930260181f, -0.366900384f, 0.f),
    C_Vertex3f(0.404952526f, 0.0890626907f, 0.f),
    C_Vertex3f(0.404952526f, 0.0890626907f, -1.f),
    C_Normal3f(0.868737638f, -0.495272696f, 0.f),
    C_Vertex3f(0.396330833f, 0.0711326599f, 0.f),
    C_Vertex3f(0.396330833f, 0.0711326599f, -1.f),
    C_Normal3f(0.788599133f, -0.614907563f, 0.f),
    C_Vertex3f(0.384835243f, 0.0537757874f, 0.f),
    C_Vertex3f(0.384835243f, 0.0537757874f, -1.f),
    C_Normal3f(0.6853109f, -0.728250563f, 0.f),
    C_Vertex3f(0.370794773f, 0.0383834839f, 0.f),
    C_Vertex3f(0.370794773f, 0.0383834839f, -1.f),
    C_Normal3f(0.571765184f, -0.820417285f, 0.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, 0.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, -1.f),
    C_Normal3f(0.456936926f, -0.889499009f, 0.f),
    C_Vertex3f(0.335079193f, 0.0134925842f, 0.f),
    C_Vertex3f(0.335079193f, 0.0134925842f, -1.f),
    C_Normal3f(0.348042637f, -0.937478721f, 0.f),
    C_Vertex3f(0.313404083f, 0.00399398804f, 0.f),
    C_Vertex3f(0.313404083f, 0.00399398804f, -1.f),
    C_Normal3f(0.24191834f, -0.970296621f, 0.f),
    C_Vertex3f(0.290305138f, -0.00312995911f, 0.f),
    C_Vertex3f(0.290305138f, -0.00312995911f, -1.f),
    C_Normal3f(0.139768258f, -0.990184247f, 0.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, 0.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, -1.f),
    C_Normal3f(0.046545554f, -0.998916209f, 0.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, 0.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0253265686f, -0.999679208f, 0.f),
    C_Vertex3f(0.214819908f, -0.0102539063f, 0.f),
    C_Vertex3f(0.214819908f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0849944577f, -0.996381402f, 0.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, 0.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, -1.f),
    C_Normal3f(-0.150375873f, -0.988628924f, 0.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, 0.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, -1.f),
    C_Normal3f(-0.223169982f, -0.974779606f, 0.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, 0.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, -1.f),
    C_Normal3f(-0.303275019f, -0.952903092f, 0.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, 0.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, -1.f),
    C_Normal3f(-0.389873773f, -0.920868278f, 0.f),
    C_Vertex3f(0.121732354f, 0.00735211372f, 0.f),
    C_Vertex3f(0.121732354f, 0.00735211372f, -1.f),
    C_Normal3f(-0.481160849f, -0.876632392f, 0.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, 0.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, -1.f),
    C_Normal3f(-0.574190378f, -0.818721831f, 0.f),
    C_Vertex3f(0.0939326286f, 0.0226106644f, 0.f),
    C_Vertex3f(0.0939326286f, 0.0226106644f, -1.f),
    C_Normal3f(-0.659739971f, -0.75149405f, 0.f),
    C_Vertex3f(0.082054615f, 0.0320005417f, 0.f),
    C_Vertex3f(0.082054615f, 0.0320005417f, -1.f),
    C_Normal3f(-0.733788431f, -0.679378092f, 0.f),
    C_Vertex3f(0.0712507963f, 0.0425231457f, 0.f),
    C_Vertex3f(0.0712507963f, 0.0425231457f, -1.f),
    C_Normal3f(-0.799175143f, -0.601098239f, 0.f),
    C_Vertex3f(0.0615211725f, 0.0541784763f, 0.f),
    C_Vertex3f(0.0615211725f, 0.0541784763f, -1.f),
    C_Normal3f(-0.854495287f, -0.519459128f, 0.f),
    C_Vertex3f(0.0528657436f, 0.0669665337f, 0.f),
    C_Vertex3f(0.0528657436f, 0.0669665337f, -1.f),
    C_Normal3f(-0.899324596f, -0.437281758f, 0.f),
    C_Vertex3f(0.0452845097f, 0.0808873177f, 0.f),
    C_Vertex3f(0.0452845097f, 0.0808873177f, -1.f),
    C_Normal3f(-0.934099019f, -0.357014179f, 0.f),
    C_Vertex3f(0.0387774706f, 0.0959408283f, 0.f),
    C_Vertex3f(0.0387774706f, 0.0959408283f, -1.f),
    C_Normal3f(-0.959852457f, -0.280505359f, 0.f),
    C_Vertex3f(0.0333446264f, 0.112127066f, 0.f),
    C_Vertex3f(0.0333446264f, 0.112127066f, -1.f),
    C_Normal3f(-0.975386679f, -0.220501676f, 0.f),
    C_Vertex3f(0.0289859772f, 0.12944603f, 0.f),
    C_Vertex3f(0.0289859772f, 0.12944603f, -1.f),
    C_Normal3f(-0.98452431f, -0.175248176f, 0.f),
    C_Vertex3f(0.02734375f, 0.138671875f, 0.f),
    C_Vertex3f(0.02734375f, 0.138671875f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.840821147f, 0.541312993f, 0.f),
    C_Vertex3f(0.219726563f, 0.569824219f, 0.f),
    C_Vertex3f(0.219726563f, 0.569824219f, -1.f),
    C_Vertex3f(0.267822266f, 0.64453125f, -1.f),
    C_Vertex3f(0.267822266f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.267822266f, 0.64453125f, 0.f),
    C_Vertex3f(0.267822266f, 0.64453125f, -1.f),
    C_Vertex3f(0.357177734f, 0.64453125f, -1.f),
    C_Vertex3f(0.357177734f, 0.64453125f, 0.f),
    C_Normal3f(0.779806852f, -0.626020253f, 0.f),
    C_Vertex3f(0.357177734f, 0.64453125f, 0.f),
    C_Vertex3f(0.357177734f, 0.64453125f, -1.f),
    C_Vertex3f(0.258789063f, 0.521972656f, -1.f),
    C_Vertex3f(0.258789063f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.258789063f, 0.521972656f, 0.f),
    C_Vertex3f(0.258789063f, 0.521972656f, -1.f),
    C_Vertex3f(0.174804688f, 0.521972656f, -1.f),
    C_Vertex3f(0.174804688f, 0.521972656f, 0.f),
    C_Normal3f(-0.795031667f, -0.606567979f, 0.f),
    C_Vertex3f(0.174804688f, 0.521972656f, 0.f),
    C_Vertex3f(0.174804688f, 0.521972656f, -1.f),
    C_Vertex3f(0.0812988281f, 0.64453125f, -1.f),
    C_Vertex3f(0.0812988281f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0812988281f, 0.64453125f, 0.f),
    C_Vertex3f(0.0812988281f, 0.64453125f, -1.f),
    C_Vertex3f(0.170654297f, 0.64453125f, -1.f),
    C_Vertex3f(0.170654297f, 0.64453125f, 0.f),
    C_Normal3f(0.835812926f, 0.54901439f, 0.f),
    C_Vertex3f(0.170654297f, 0.64453125f, 0.f),
    C_Vertex3f(0.170654297f, 0.64453125f, -1.f),
    C_Vertex3f(0.219726563f, 0.569824219f, -1.f),
    C_Vertex3f(0.219726563f, 0.569824219f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.395996094f, 0.344238281f, 0.f),
    C_Vertex3f(0.394136429f, 0.355503082f, 0.f),
    C_Vertex3f(0.388884068f, 0.375763416f, 0.f),
    C_Vertex3f(0.382098675f, 0.393754482f, 0.f),
    C_Vertex3f(0.373780251f, 0.40947628f, 0.f),
    C_Vertex3f(0.363928795f, 0.42292881f, 0.f),
    C_Vertex3f(0.351369381f, 0.434987068f, 0.f),
    C_Vertex3f(0.336102009f, 0.445651054f, 0.f),
    C_Vertex3f(0.318847656f, 0.333984375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.318126678f, 0.454920769f, 0.f),
    C_Vertex3f(0.29744339f, 0.462796211f, 0.f),
    C_Vertex3f(0.304819107f, 0.372945309f, 0.f),
    C_Vertex3f(0.312421799f, 0.358936787f, 0.f),
    C_Vertex3f(0.3175354f, 0.34295845f, 0.f),
    C_Vertex3f(0.318847656f, 0.333984375f, 0.f),
    C_Vertex3f(0.336102009f, 0.445651054f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.228822708f, 0.41015625f, 0.f),
    C_Vertex3f(0.249194145f, 0.407639027f, 0.f),
    C_Vertex3f(0.251067162f, 0.472640514f, 0.f),
    C_Vertex3f(0.266968727f, 0.40260458f, 0.f),
    C_Vertex3f(0.275090218f, 0.468702793f, 0.f),
    C_Vertex3f(0.282146454f, 0.39505291f, 0.f),
    C_Vertex3f(0.29744339f, 0.462796211f, 0.f),
    C_Vertex3f(0.294727325f, 0.384984016f, 0.f),
    C_Vertex3f(0.304819107f, 0.372945309f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186338782f, 0.408473611f, 0.f),
    C_Vertex3f(0.197961926f, 0.40959537f, 0.f),
    C_Vertex3f(0.202610016f, 0.474609375f, 0.f),
    C_Vertex3f(0.21055603f, 0.41015625f, 0.f),
    C_Vertex3f(0.225374222f, 0.474609375f, 0.f),
    C_Vertex3f(0.228822708f, 0.41015625f, 0.f),
    C_Vertex3f(0.251067162f, 0.472640514f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157295108f, 0.401743054f, 0.f),
    C_Vertex3f(0.166005373f, 0.404547453f, 0.f),
    C_Vertex3f(0.168001175f, 0.471088886f, 0.f),
    C_Vertex3f(0.175686598f, 0.406790972f, 0.f),
    C_Vertex3f(0.185018539f, 0.473435879f, 0.f),
    C_Vertex3f(0.186338782f, 0.408473611f, 0.f),
    C_Vertex3f(0.202610016f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.151557922f, 0.467568398f, 0.f),
    C_Vertex3f(0.135688782f, 0.462874413f, 0.f),
    C_Vertex3f(0.142787457f, 0.394451618f, 0.f),
    C_Vertex3f(0.149555802f, 0.398377776f, 0.f),
    C_Vertex3f(0.157295108f, 0.401743054f, 0.f),
    C_Vertex3f(0.168001175f, 0.471088886f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116699219f, 0.356201172f, 0.f),
    C_Vertex3f(0.119597912f, 0.367474079f, 0.f),
    C_Vertex3f(0.121488571f, 0.457662106f, 0.f),
    C_Vertex3f(0.125395298f, 0.377980709f, 0.f),
    C_Vertex3f(0.135688782f, 0.462874413f, 0.f),
    C_Vertex3f(0.134091377f, 0.387721062f, 0.f),
    C_Vertex3f(0.142787457f, 0.394451618f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12030983f, 0.332175255f, 0.f),
    C_Vertex3f(0.117902756f, 0.33907032f, 0.f),
    C_Vertex3f(0.122807026f, 0.228503942f, 0.f),
    C_Vertex3f(0.116699219f, 0.346349716f, 0.f),
    C_Vertex3f(0.112774849f, 0.232672691f, 0.f),
    C_Vertex3f(0.116699219f, 0.356201172f, 0.f),
    C_Vertex3f(0.108957291f, 0.451931477f, 0.f),
    C_Vertex3f(0.121488571f, 0.457662106f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.100217819f, 0.238649368f, 0.f),
    C_Vertex3f(0.112774849f, 0.232672691f, 0.f),
    C_Vertex3f(0.108957291f, 0.451931477f, 0.f),
    C_Vertex3f(0.0980949402f, 0.445682526f, 0.f),
    C_Vertex3f(0.09349823f, 0.442298889f, 0.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, 0.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0423059464f, 0.363668919f, 0.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, 0.f),
    C_Vertex3f(0.0458631516f, 0.378024578f, 0.f),
    C_Vertex3f(0.0470104218f, 0.300626278f, 0.f),
    C_Vertex3f(0.0511989594f, 0.391806602f, 0.f),
    C_Vertex3f(0.0534934998f, 0.28580761f, 0.f),
    C_Vertex3f(0.0583133698f, 0.405014992f, 0.f),
    C_Vertex3f(0.0621376038f, 0.271725655f, 0.f),
    C_Vertex3f(0.0667133331f, 0.417105198f, 0.f),
    C_Vertex3f(0.0728063583f, 0.259171963f, 0.f),
    C_Vertex3f(0.0763988495f, 0.428077221f, 0.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, 0.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0405273438f, 0.348739624f, 0.f),
    C_Vertex3f(0.0405273438f, 0.332473755f, 0.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, 0.f),
    C_Vertex3f(0.0423059464f, 0.363668919f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.351041794f, 0.241439819f, 0.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, 0.f),
    C_Vertex3f(0.365812302f, 0.232733727f, 0.f),
    C_Vertex3f(0.370794773f, 0.0383834839f, 0.f),
    C_Vertex3f(0.378801346f, 0.222416878f, 0.f),
    C_Vertex3f(0.384835243f, 0.0537757874f, 0.f),
    C_Vertex3f(0.390008926f, 0.210489273f, 0.f),
    C_Vertex3f(0.396330833f, 0.0711326599f, 0.f),
    C_Vertex3f(0.399435043f, 0.196950912f, 0.f),
    C_Vertex3f(0.404952526f, 0.0890626907f, 0.f),
    C_Vertex3f(0.406504631f, 0.181775093f, 0.f),
    C_Vertex3f(0.410700321f, 0.10756588f, 0.f),
    C_Vertex3f(0.41121769f, 0.164961815f, 0.f),
    C_Vertex3f(0.413574219f, 0.126642227f, 0.f),
    C_Vertex3f(0.413574219f, 0.146511078f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338627815f, 0.246761322f, 0.f),
    C_Vertex3f(0.328774691f, 0.250561476f, 0.f),
    C_Vertex3f(0.329612732f, 0.146773338f, 0.f),
    C_Vertex3f(0.332519531f, 0.134847641f, 0.f),
    C_Vertex3f(0.335079193f, 0.0134925842f, 0.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, 0.f),
    C_Vertex3f(0.351041794f, 0.241439819f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.307235718f, 0.171502113f, 0.f),
    C_Vertex3f(0.315078735f, 0.166358948f, 0.f),
    C_Vertex3f(0.317344427f, 0.254614115f, 0.f),
    C_Vertex3f(0.323799133f, 0.157277107f, 0.f),
    C_Vertex3f(0.328774691f, 0.250561476f, 0.f),
    C_Vertex3f(0.329612732f, 0.146773338f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304337025f, 0.258919239f, 0.f),
    C_Vertex3f(0.289752483f, 0.263476849f, 0.f),
    C_Vertex3f(0.294609785f, 0.177288532f, 0.f),
    C_Vertex3f(0.301674128f, 0.174306393f, 0.f),
    C_Vertex3f(0.307235718f, 0.171502113f, 0.f),
    C_Vertex3f(0.317344427f, 0.254614115f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.236746788f, 0.194867134f, 0.f),
    C_Vertex3f(0.251324892f, 0.190995693f, 0.f),
    C_Vertex3f(0.255851984f, 0.273349524f, 0.f),
    C_Vertex3f(0.264400244f, 0.187302113f, 0.f),
    C_Vertex3f(0.273590803f, 0.268286943f, 0.f),
    C_Vertex3f(0.275972843f, 0.183786392f, 0.f),
    C_Vertex3f(0.289752483f, 0.263476849f, 0.f),
    C_Vertex3f(0.28604269f, 0.180448532f, 0.f),
    C_Vertex3f(0.294609785f, 0.177288532f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.255851984f, 0.273349524f, 0.f),
    C_Vertex3f(0.236536026f, 0.278664589f, 0.f),
    C_Vertex3f(0.236746788f, 0.194867134f, 0.f),
    C_Vertex3f(0.218869209f, 0.283418655f, 0.f),
    C_Vertex3f(0.217877388f, 0.199621201f, 0.f),
    C_Vertex3f(0.205864668f, 0.287003517f, 0.f),
    C_Vertex3f(0.197880268f, 0.204895735f, 0.f),
    C_Vertex3f(0.194296122f, 0.290232658f, 0.f),
    C_Vertex3f(0.18416357f, 0.293106079f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179543972f, 0.209985971f, 0.f),
    C_Vertex3f(0.197880268f, 0.204895735f, 0.f),
    C_Vertex3f(0.18416357f, 0.293106079f, 0.f),
    C_Vertex3f(0.175467014f, 0.295623779f, 0.f),
    C_Vertex3f(0.168206453f, 0.297785759f, 0.f),
    C_Vertex3f(0.1628685f, 0.214891911f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1628685f, 0.214891911f, 0.f),
    C_Vertex3f(0.168206453f, 0.297785759f, 0.f),
    C_Vertex3f(0.162381887f, 0.299592018f, 0.f),
    C_Vertex3f(0.157993317f, 0.301042557f, 0.f),
    C_Vertex3f(0.156517029f, 0.301589966f, 0.f),
    C_Vertex3f(0.151374817f, 0.303668976f, 0.f),
    C_Vertex3f(0.147853851f, 0.219613552f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151374817f, 0.303668976f, 0.f),
    C_Vertex3f(0.142458916f, 0.308374405f, 0.f),
    C_Vertex3f(0.147853851f, 0.219613552f, 0.f),
    C_Vertex3f(0.134911537f, 0.313627243f, 0.f),
    C_Vertex3f(0.134500027f, 0.224150896f, 0.f),
    C_Vertex3f(0.128732681f, 0.31942749f, 0.f),
    C_Vertex3f(0.126327515f, 0.322601318f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123920441f, 0.32566452f, 0.f),
    C_Vertex3f(0.12030983f, 0.332175255f, 0.f),
    C_Vertex3f(0.122807026f, 0.228503942f, 0.f),
    C_Vertex3f(0.134500027f, 0.224150896f, 0.f),
    C_Vertex3f(0.126327515f, 0.322601318f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.335079193f, 0.0134925842f, 0.f),
    C_Vertex3f(0.332519531f, 0.134847641f, 0.f),
    C_Vertex3f(0.332519531f, 0.120731354f, 0.f),
    C_Vertex3f(0.329231262f, 0.106804848f, 0.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, 0.f),
    C_Vertex3f(0.313404083f, 0.00399398804f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.313404083f, 0.00399398804f, 0.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, 0.f),
    C_Vertex3f(0.312789917f, 0.0818271637f, 0.f),
    C_Vertex3f(0.30292511f, 0.0735387802f, 0.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, 0.f),
    C_Vertex3f(0.290305138f, -0.00312995911f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.290305138f, -0.00312995911f, 0.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, 0.f),
    C_Vertex3f(0.287142038f, 0.0645596981f, 0.f),
    C_Vertex3f(0.277935505f, 0.061106205f, 0.f),
    C_Vertex3f(0.267852306f, 0.0583434105f, 0.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.267852306f, 0.0583434105f, 0.f),
    C_Vertex3f(0.256892443f, 0.0562713146f, 0.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, 0.f),
    C_Vertex3f(0.245055914f, 0.0548899174f, 0.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, 0.f),
    C_Vertex3f(0.23234272f, 0.0541992188f, 0.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214819908f, -0.0102539063f, 0.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, 0.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, 0.f),
    C_Vertex3f(0.206072688f, 0.0549784899f, 0.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, 0.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.182843447f, 0.0588748455f, 0.f),
    C_Vertex3f(0.172348738f, 0.06199193f, 0.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, 0.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, 0.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162600636f, 0.0658882856f, 0.f),
    C_Vertex3f(0.153599143f, 0.0705639124f, 0.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, 0.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, 0.f),
    C_Vertex3f(0.172348738f, 0.06199193f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153599143f, 0.0705639124f, 0.f),
    C_Vertex3f(0.145344257f, 0.0760188103f, 0.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, 0.f),
    C_Vertex3f(0.137835979f, 0.0822529793f, 0.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, 0.f),
    C_Vertex3f(0.131060719f, 0.0891705751f, 0.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.121732354f, 0.00735211372f, 0.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, 0.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, 0.f),
    C_Vertex3f(0.119709253f, 0.105056047f, 0.f),
    C_Vertex3f(0.115133047f, 0.114023924f, 0.f),
    C_Vertex3f(0.111289859f, 0.123675227f, 0.f),
    C_Vertex3f(0.108179688f, 0.134009957f, 0.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.082054615f, 0.0320005417f, 0.f),
    C_Vertex3f(0.0939326286f, 0.0226106644f, 0.f),
    C_Vertex3f(0.104980469f, 0.150878906f, 0.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, 0.f),
    C_Vertex3f(0.105802536f, 0.145028114f, 0.f),
    C_Vertex3f(0.108179688f, 0.134009957f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.104980469f, 0.150878906f, 0.f),
    C_Vertex3f(0.02734375f, 0.138671875f, 0.f),
    C_Vertex3f(0.0289859772f, 0.12944603f, 0.f),
    C_Vertex3f(0.0333446264f, 0.112127066f, 0.f),
    C_Vertex3f(0.0387774706f, 0.0959408283f, 0.f),
    C_Vertex3f(0.0452845097f, 0.0808873177f, 0.f),
    C_Vertex3f(0.0528657436f, 0.0669665337f, 0.f),
    C_Vertex3f(0.0615211725f, 0.0541784763f, 0.f),
    C_Vertex3f(0.0712507963f, 0.0425231457f, 0.f),
    C_Vertex3f(0.082054615f, 0.0320005417f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.357177734f, 0.64453125f, 0.f),
    C_Vertex3f(0.267822266f, 0.64453125f, 0.f),
    C_Vertex3f(0.258789063f, 0.521972656f, 0.f),
    C_Vertex3f(0.219726563f, 0.569824219f, 0.f),
    C_Vertex3f(0.174804688f, 0.521972656f, 0.f),
    C_Vertex3f(0.170654297f, 0.64453125f, 0.f),
    C_Vertex3f(0.0812988281f, 0.64453125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.395996094f, 0.344238281f, -1.f),
    C_Vertex3f(0.394136429f, 0.355503082f, -1.f),
    C_Vertex3f(0.388884068f, 0.375763416f, -1.f),
    C_Vertex3f(0.382098675f, 0.393754482f, -1.f),
    C_Vertex3f(0.373780251f, 0.40947628f, -1.f),
    C_Vertex3f(0.363928795f, 0.42292881f, -1.f),
    C_Vertex3f(0.351369381f, 0.434987068f, -1.f),
    C_Vertex3f(0.336102009f, 0.445651054f, -1.f),
    C_Vertex3f(0.318847656f, 0.333984375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.318126678f, 0.454920769f, -1.f),
    C_Vertex3f(0.29744339f, 0.462796211f, -1.f),
    C_Vertex3f(0.304819107f, 0.372945309f, -1.f),
    C_Vertex3f(0.312421799f, 0.358936787f, -1.f),
    C_Vertex3f(0.3175354f, 0.34295845f, -1.f),
    C_Vertex3f(0.318847656f, 0.333984375f, -1.f),
    C_Vertex3f(0.336102009f, 0.445651054f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.228822708f, 0.41015625f, -1.f),
    C_Vertex3f(0.249194145f, 0.407639027f, -1.f),
    C_Vertex3f(0.251067162f, 0.472640514f, -1.f),
    C_Vertex3f(0.266968727f, 0.40260458f, -1.f),
    C_Vertex3f(0.275090218f, 0.468702793f, -1.f),
    C_Vertex3f(0.282146454f, 0.39505291f, -1.f),
    C_Vertex3f(0.29744339f, 0.462796211f, -1.f),
    C_Vertex3f(0.294727325f, 0.384984016f, -1.f),
    C_Vertex3f(0.304819107f, 0.372945309f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186338782f, 0.408473611f, -1.f),
    C_Vertex3f(0.197961926f, 0.40959537f, -1.f),
    C_Vertex3f(0.202610016f, 0.474609375f, -1.f),
    C_Vertex3f(0.21055603f, 0.41015625f, -1.f),
    C_Vertex3f(0.225374222f, 0.474609375f, -1.f),
    C_Vertex3f(0.228822708f, 0.41015625f, -1.f),
    C_Vertex3f(0.251067162f, 0.472640514f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157295108f, 0.401743054f, -1.f),
    C_Vertex3f(0.166005373f, 0.404547453f, -1.f),
    C_Vertex3f(0.168001175f, 0.471088886f, -1.f),
    C_Vertex3f(0.175686598f, 0.406790972f, -1.f),
    C_Vertex3f(0.185018539f, 0.473435879f, -1.f),
    C_Vertex3f(0.186338782f, 0.408473611f, -1.f),
    C_Vertex3f(0.202610016f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.151557922f, 0.467568398f, -1.f),
    C_Vertex3f(0.135688782f, 0.462874413f, -1.f),
    C_Vertex3f(0.142787457f, 0.394451618f, -1.f),
    C_Vertex3f(0.149555802f, 0.398377776f, -1.f),
    C_Vertex3f(0.157295108f, 0.401743054f, -1.f),
    C_Vertex3f(0.168001175f, 0.471088886f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116699219f, 0.356201172f, -1.f),
    C_Vertex3f(0.119597912f, 0.367474079f, -1.f),
    C_Vertex3f(0.121488571f, 0.457662106f, -1.f),
    C_Vertex3f(0.125395298f, 0.377980709f, -1.f),
    C_Vertex3f(0.135688782f, 0.462874413f, -1.f),
    C_Vertex3f(0.134091377f, 0.387721062f, -1.f),
    C_Vertex3f(0.142787457f, 0.394451618f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12030983f, 0.332175255f, -1.f),
    C_Vertex3f(0.117902756f, 0.33907032f, -1.f),
    C_Vertex3f(0.122807026f, 0.228503942f, -1.f),
    C_Vertex3f(0.116699219f, 0.346349716f, -1.f),
    C_Vertex3f(0.112774849f, 0.232672691f, -1.f),
    C_Vertex3f(0.116699219f, 0.356201172f, -1.f),
    C_Vertex3f(0.108957291f, 0.451931477f, -1.f),
    C_Vertex3f(0.121488571f, 0.457662106f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.100217819f, 0.238649368f, -1.f),
    C_Vertex3f(0.112774849f, 0.232672691f, -1.f),
    C_Vertex3f(0.108957291f, 0.451931477f, -1.f),
    C_Vertex3f(0.0980949402f, 0.445682526f, -1.f),
    C_Vertex3f(0.09349823f, 0.442298889f, -1.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, -1.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0423059464f, 0.363668919f, -1.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, -1.f),
    C_Vertex3f(0.0458631516f, 0.378024578f, -1.f),
    C_Vertex3f(0.0470104218f, 0.300626278f, -1.f),
    C_Vertex3f(0.0511989594f, 0.391806602f, -1.f),
    C_Vertex3f(0.0534934998f, 0.28580761f, -1.f),
    C_Vertex3f(0.0583133698f, 0.405014992f, -1.f),
    C_Vertex3f(0.0621376038f, 0.271725655f, -1.f),
    C_Vertex3f(0.0667133331f, 0.417105198f, -1.f),
    C_Vertex3f(0.0728063583f, 0.259171963f, -1.f),
    C_Vertex3f(0.0763988495f, 0.428077221f, -1.f),
    C_Vertex3f(0.0854997635f, 0.248146534f, -1.f),
    C_Vertex3f(0.0873699188f, 0.437931061f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0405273438f, 0.348739624f, -1.f),
    C_Vertex3f(0.0405273438f, 0.332473755f, -1.f),
    C_Vertex3f(0.0426883698f, 0.31618166f, -1.f),
    C_Vertex3f(0.0423059464f, 0.363668919f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.351041794f, 0.241439819f, -1.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, -1.f),
    C_Vertex3f(0.365812302f, 0.232733727f, -1.f),
    C_Vertex3f(0.370794773f, 0.0383834839f, -1.f),
    C_Vertex3f(0.378801346f, 0.222416878f, -1.f),
    C_Vertex3f(0.384835243f, 0.0537757874f, -1.f),
    C_Vertex3f(0.390008926f, 0.210489273f, -1.f),
    C_Vertex3f(0.396330833f, 0.0711326599f, -1.f),
    C_Vertex3f(0.399435043f, 0.196950912f, -1.f),
    C_Vertex3f(0.404952526f, 0.0890626907f, -1.f),
    C_Vertex3f(0.406504631f, 0.181775093f, -1.f),
    C_Vertex3f(0.410700321f, 0.10756588f, -1.f),
    C_Vertex3f(0.41121769f, 0.164961815f, -1.f),
    C_Vertex3f(0.413574219f, 0.126642227f, -1.f),
    C_Vertex3f(0.413574219f, 0.146511078f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338627815f, 0.246761322f, -1.f),
    C_Vertex3f(0.328774691f, 0.250561476f, -1.f),
    C_Vertex3f(0.329612732f, 0.146773338f, -1.f),
    C_Vertex3f(0.332519531f, 0.134847641f, -1.f),
    C_Vertex3f(0.335079193f, 0.0134925842f, -1.f),
    C_Vertex3f(0.354209423f, 0.0249557495f, -1.f),
    C_Vertex3f(0.351041794f, 0.241439819f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.307235718f, 0.171502113f, -1.f),
    C_Vertex3f(0.315078735f, 0.166358948f, -1.f),
    C_Vertex3f(0.317344427f, 0.254614115f, -1.f),
    C_Vertex3f(0.323799133f, 0.157277107f, -1.f),
    C_Vertex3f(0.328774691f, 0.250561476f, -1.f),
    C_Vertex3f(0.329612732f, 0.146773338f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304337025f, 0.258919239f, -1.f),
    C_Vertex3f(0.289752483f, 0.263476849f, -1.f),
    C_Vertex3f(0.294609785f, 0.177288532f, -1.f),
    C_Vertex3f(0.301674128f, 0.174306393f, -1.f),
    C_Vertex3f(0.307235718f, 0.171502113f, -1.f),
    C_Vertex3f(0.317344427f, 0.254614115f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.236746788f, 0.194867134f, -1.f),
    C_Vertex3f(0.251324892f, 0.190995693f, -1.f),
    C_Vertex3f(0.255851984f, 0.273349524f, -1.f),
    C_Vertex3f(0.264400244f, 0.187302113f, -1.f),
    C_Vertex3f(0.273590803f, 0.268286943f, -1.f),
    C_Vertex3f(0.275972843f, 0.183786392f, -1.f),
    C_Vertex3f(0.289752483f, 0.263476849f, -1.f),
    C_Vertex3f(0.28604269f, 0.180448532f, -1.f),
    C_Vertex3f(0.294609785f, 0.177288532f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.255851984f, 0.273349524f, -1.f),
    C_Vertex3f(0.236536026f, 0.278664589f, -1.f),
    C_Vertex3f(0.236746788f, 0.194867134f, -1.f),
    C_Vertex3f(0.218869209f, 0.283418655f, -1.f),
    C_Vertex3f(0.217877388f, 0.199621201f, -1.f),
    C_Vertex3f(0.205864668f, 0.287003517f, -1.f),
    C_Vertex3f(0.197880268f, 0.204895735f, -1.f),
    C_Vertex3f(0.194296122f, 0.290232658f, -1.f),
    C_Vertex3f(0.18416357f, 0.293106079f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179543972f, 0.209985971f, -1.f),
    C_Vertex3f(0.197880268f, 0.204895735f, -1.f),
    C_Vertex3f(0.18416357f, 0.293106079f, -1.f),
    C_Vertex3f(0.175467014f, 0.295623779f, -1.f),
    C_Vertex3f(0.168206453f, 0.297785759f, -1.f),
    C_Vertex3f(0.1628685f, 0.214891911f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1628685f, 0.214891911f, -1.f),
    C_Vertex3f(0.168206453f, 0.297785759f, -1.f),
    C_Vertex3f(0.162381887f, 0.299592018f, -1.f),
    C_Vertex3f(0.157993317f, 0.301042557f, -1.f),
    C_Vertex3f(0.156517029f, 0.301589966f, -1.f),
    C_Vertex3f(0.151374817f, 0.303668976f, -1.f),
    C_Vertex3f(0.147853851f, 0.219613552f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151374817f, 0.303668976f, -1.f),
    C_Vertex3f(0.142458916f, 0.308374405f, -1.f),
    C_Vertex3f(0.147853851f, 0.219613552f, -1.f),
    C_Vertex3f(0.134911537f, 0.313627243f, -1.f),
    C_Vertex3f(0.134500027f, 0.224150896f, -1.f),
    C_Vertex3f(0.128732681f, 0.31942749f, -1.f),
    C_Vertex3f(0.126327515f, 0.322601318f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123920441f, 0.32566452f, -1.f),
    C_Vertex3f(0.12030983f, 0.332175255f, -1.f),
    C_Vertex3f(0.122807026f, 0.228503942f, -1.f),
    C_Vertex3f(0.134500027f, 0.224150896f, -1.f),
    C_Vertex3f(0.126327515f, 0.322601318f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.335079193f, 0.0134925842f, -1.f),
    C_Vertex3f(0.332519531f, 0.134847641f, -1.f),
    C_Vertex3f(0.332519531f, 0.120731354f, -1.f),
    C_Vertex3f(0.329231262f, 0.106804848f, -1.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, -1.f),
    C_Vertex3f(0.313404083f, 0.00399398804f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.313404083f, 0.00399398804f, -1.f),
    C_Vertex3f(0.322654724f, 0.0938367844f, -1.f),
    C_Vertex3f(0.312789917f, 0.0818271637f, -1.f),
    C_Vertex3f(0.30292511f, 0.0735387802f, -1.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, -1.f),
    C_Vertex3f(0.290305138f, -0.00312995911f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.290305138f, -0.00312995911f, -1.f),
    C_Vertex3f(0.295471907f, 0.0687038898f, -1.f),
    C_Vertex3f(0.287142038f, 0.0645596981f, -1.f),
    C_Vertex3f(0.277935505f, 0.061106205f, -1.f),
    C_Vertex3f(0.267852306f, 0.0583434105f, -1.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.267852306f, 0.0583434105f, -1.f),
    C_Vertex3f(0.256892443f, 0.0562713146f, -1.f),
    C_Vertex3f(0.265782356f, -0.0078792572f, -1.f),
    C_Vertex3f(0.245055914f, 0.0548899174f, -1.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, -1.f),
    C_Vertex3f(0.23234272f, 0.0541992188f, -1.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214819908f, -0.0102539063f, -1.f),
    C_Vertex3f(0.239835739f, -0.0102539063f, -1.f),
    C_Vertex3f(0.21880722f, 0.0541992188f, -1.f),
    C_Vertex3f(0.206072688f, 0.0549784899f, -1.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, -1.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.182843447f, 0.0588748455f, -1.f),
    C_Vertex3f(0.172348738f, 0.06199193f, -1.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, -1.f),
    C_Vertex3f(0.193506598f, -0.00908017159f, -1.f),
    C_Vertex3f(0.194084764f, 0.0565370321f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162600636f, 0.0658882856f, -1.f),
    C_Vertex3f(0.153599143f, 0.0705639124f, -1.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, -1.f),
    C_Vertex3f(0.173541188f, -0.00673270226f, -1.f),
    C_Vertex3f(0.172348738f, 0.06199193f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153599143f, 0.0705639124f, -1.f),
    C_Vertex3f(0.145344257f, 0.0760188103f, -1.f),
    C_Vertex3f(0.154923677f, -0.00321149826f, -1.f),
    C_Vertex3f(0.137835979f, 0.0822529793f, -1.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, -1.f),
    C_Vertex3f(0.131060719f, 0.0891705751f, -1.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.121732354f, 0.00735211372f, -1.f),
    C_Vertex3f(0.137654066f, 0.0014834404f, -1.f),
    C_Vertex3f(0.125018477f, 0.0967715979f, -1.f),
    C_Vertex3f(0.119709253f, 0.105056047f, -1.f),
    C_Vertex3f(0.115133047f, 0.114023924f, -1.f),
    C_Vertex3f(0.111289859f, 0.123675227f, -1.f),
    C_Vertex3f(0.108179688f, 0.134009957f, -1.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.082054615f, 0.0320005417f, -1.f),
    C_Vertex3f(0.0939326286f, 0.0226106644f, -1.f),
    C_Vertex3f(0.104980469f, 0.150878906f, -1.f),
    C_Vertex3f(0.107158542f, 0.0143945217f, -1.f),
    C_Vertex3f(0.105802536f, 0.145028114f, -1.f),
    C_Vertex3f(0.108179688f, 0.134009957f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.104980469f, 0.150878906f, -1.f),
    C_Vertex3f(0.02734375f, 0.138671875f, -1.f),
    C_Vertex3f(0.0289859772f, 0.12944603f, -1.f),
    C_Vertex3f(0.0333446264f, 0.112127066f, -1.f),
    C_Vertex3f(0.0387774706f, 0.0959408283f, -1.f),
    C_Vertex3f(0.0452845097f, 0.0808873177f, -1.f),
    C_Vertex3f(0.0528657436f, 0.0669665337f, -1.f),
    C_Vertex3f(0.0615211725f, 0.0541784763f, -1.f),
    C_Vertex3f(0.0712507963f, 0.0425231457f, -1.f),
    C_Vertex3f(0.082054615f, 0.0320005417f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.357177734f, 0.64453125f, -1.f),
    C_Vertex3f(0.267822266f, 0.64453125f, -1.f),
    C_Vertex3f(0.258789063f, 0.521972656f, -1.f),
    C_Vertex3f(0.219726563f, 0.569824219f, -1.f),
    C_Vertex3f(0.174804688f, 0.521972656f, -1.f),
    C_Vertex3f(0.170654297f, 0.64453125f, -1.f),
    C_Vertex3f(0.0812988281f, 0.64453125f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_155[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.863734305f, -0.503947437f, 0.f),
    C_Vertex3f(0.155761719f, 0.23134613f, 0.f),
    C_Vertex3f(0.155761719f, 0.23134613f, -1.f),
    C_Vertex3f(0.0400390625f, 0.4296875f, -1.f),
    C_Vertex3f(0.0400390625f, 0.4296875f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0400390625f, 0.4296875f, 0.f),
    C_Vertex3f(0.0400390625f, 0.4296875f, -1.f),
    C_Vertex3f(0.104980469f, 0.4296875f, -1.f),
    C_Vertex3f(0.104980469f, 0.4296875f, 0.f),
    C_Normal3f(0.827476859f, 0.561499834f, 0.f),
    C_Vertex3f(0.104980469f, 0.4296875f, 0.f),
    C_Vertex3f(0.104980469f, 0.4296875f, -1.f),
    C_Vertex3f(0.237792969f, 0.233963013f, -1.f),
    C_Vertex3f(0.237792969f, 0.233963013f, 0.f),
    C_Normal3f(0.836460114f, -0.548027754f, 0.f),
    C_Vertex3f(0.237792969f, 0.233963013f, 0.f),
    C_Vertex3f(0.237792969f, 0.233963013f, -1.f),
    C_Vertex3f(0.104980469f, 0.03125f, -1.f),
    C_Vertex3f(0.104980469f, 0.03125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.104980469f, 0.03125f, 0.f),
    C_Vertex3f(0.104980469f, 0.03125f, -1.f),
    C_Vertex3f(0.0400390625f, 0.03125f, -1.f),
    C_Vertex3f(0.0400390625f, 0.03125f, 0.f),
    C_Normal3f(-0.865655899f, 0.500639379f, 0.f),
    C_Vertex3f(0.0400390625f, 0.03125f, 0.f),
    C_Vertex3f(0.0400390625f, 0.03125f, -1.f),
    C_Vertex3f(0.155761719f, 0.23134613f, -1.f),
    C_Vertex3f(0.155761719f, 0.23134613f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.23134613f, 0.f),
    C_Vertex3f(0.0400390625f, 0.03125f, 0.f),
    C_Vertex3f(0.104980469f, 0.03125f, 0.f),
    C_Vertex3f(0.237792969f, 0.233963013f, 0.f),
    C_Vertex3f(0.104980469f, 0.4296875f, 0.f),
    C_Vertex3f(0.0400390625f, 0.4296875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.23134613f, -1.f),
    C_Vertex3f(0.0400390625f, 0.03125f, -1.f),
    C_Vertex3f(0.104980469f, 0.03125f, -1.f),
    C_Vertex3f(0.237792969f, 0.233963013f, -1.f),
    C_Vertex3f(0.104980469f, 0.4296875f, -1.f),
    C_Vertex3f(0.0400390625f, 0.4296875f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_156[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.792629182f, -0.609704018f, 0.f),
    C_Vertex3f(0.427177429f, 0.0765380859f, 0.f),
    C_Vertex3f(0.427177429f, 0.0765380859f, -1.f),
    C_Normal3f(0.750469565f, -0.660905123f, 0.f),
    C_Vertex3f(0.418874741f, 0.0657444f, 0.f),
    C_Vertex3f(0.418874741f, 0.0657444f, -1.f),
    C_Normal3f(0.689071834f, -0.724693f, 0.f),
    C_Vertex3f(0.401012897f, 0.0468277931f, 0.f),
    C_Vertex3f(0.401012897f, 0.0468277931f, -1.f),
    C_Normal3f(0.602597654f, -0.798045158f, 0.f),
    C_Vertex3f(0.381894588f, 0.0305819511f, 0.f),
    C_Vertex3f(0.381894588f, 0.0305819511f, -1.f),
    C_Normal3f(0.503500819f, -0.863994777f, 0.f),
    C_Vertex3f(0.361519814f, 0.0170068741f, 0.f),
    C_Vertex3f(0.361519814f, 0.0170068741f, -1.f),
    C_Normal3f(0.393978715f, -0.919119596f, 0.f),
    C_Vertex3f(0.339888573f, 0.00610256195f, 0.f),
    C_Vertex3f(0.339888573f, 0.00610256195f, -1.f),
    C_Normal3f(0.278373241f, -0.960472941f, 0.f),
    C_Vertex3f(0.31700182f, -0.00207567215f, 0.f),
    C_Vertex3f(0.31700182f, -0.00207567215f, -1.f),
    C_Normal3f(0.162878811f, -0.986646116f, 0.f),
    C_Vertex3f(0.292859554f, -0.00752782822f, 0.f),
    C_Vertex3f(0.292859554f, -0.00752782822f, -1.f),
    C_Normal3f(0.0533612631f, -0.99857527f, 0.f),
    C_Vertex3f(0.267461777f, -0.0102539063f, 0.f),
    C_Vertex3f(0.267461777f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0398047529f, -0.999207497f, 0.f),
    C_Vertex3f(0.241845131f, -0.0102539063f, 0.f),
    C_Vertex3f(0.241845131f, -0.0102539063f, -1.f),
    C_Normal3f(-0.127445742f, -0.991845608f, 0.f),
    C_Vertex3f(0.218459964f, -0.00830185413f, 0.f),
    C_Vertex3f(0.218459964f, -0.00830185413f, -1.f),
    C_Normal3f(-0.22044614f, -0.975399137f, 0.f),
    C_Vertex3f(0.196269631f, -0.0043977499f, 0.f),
    C_Vertex3f(0.196269631f, -0.0043977499f, -1.f),
    C_Normal3f(-0.317600638f, -0.948224604f, 0.f),
    C_Vertex3f(0.175274134f, 0.00145840645f, 0.f),
    C_Vertex3f(0.175274134f, 0.00145840645f, -1.f),
    C_Normal3f(-0.415979445f, -0.909374058f, 0.f),
    C_Vertex3f(0.155473471f, 0.00926661491f, 0.f),
    C_Vertex3f(0.155473471f, 0.00926661491f, -1.f),
    C_Normal3f(-0.512082279f, -0.858936429f, 0.f),
    C_Vertex3f(0.136867642f, 0.0190268755f, 0.f),
    C_Vertex3f(0.136867642f, 0.0190268755f, -1.f),
    C_Normal3f(-0.60237205f, -0.798215449f, 0.f),
    C_Vertex3f(0.119456649f, 0.0307391882f, 0.f),
    C_Vertex3f(0.119456649f, 0.0307391882f, -1.f),
    C_Normal3f(-0.683886945f, -0.729587972f, 0.f),
    C_Vertex3f(0.10324049f, 0.044403553f, 0.f),
    C_Vertex3f(0.10324049f, 0.044403553f, -1.f),
    C_Normal3f(-0.758434057f, -0.65174973f, 0.f),
    C_Vertex3f(0.0882191658f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0882191658f, 0.0600199699f, -1.f),
    C_Normal3f(-0.826661348f, -0.562699795f, 0.f),
    C_Vertex3f(0.0750755072f, 0.0771788359f, 0.f),
    C_Vertex3f(0.0750755072f, 0.0771788359f, -1.f),
    C_Normal3f(-0.883446872f, -0.468531489f, 0.f),
    C_Vertex3f(0.063809514f, 0.0958801508f, 0.f),
    C_Vertex3f(0.063809514f, 0.0958801508f, -1.f),
    C_Normal3f(-0.927812994f, -0.373045683f, 0.f),
    C_Vertex3f(0.0544211864f, 0.116123915f, 0.f),
    C_Vertex3f(0.0544211864f, 0.116123915f, -1.f),
    C_Normal3f(-0.960084915f, -0.279708833f, 0.f),
    C_Vertex3f(0.0469105244f, 0.137910128f, 0.f),
    C_Vertex3f(0.0469105244f, 0.137910128f, -1.f),
    C_Normal3f(-0.981553733f, -0.191186532f, 0.f),
    C_Vertex3f(0.0412775278f, 0.16123879f, 0.f),
    C_Vertex3f(0.0412775278f, 0.16123879f, -1.f),
    C_Normal3f(-0.994021893f, -0.109181203f, 0.f),
    C_Vertex3f(0.0375221968f, 0.1861099f, 0.f),
    C_Vertex3f(0.0375221968f, 0.1861099f, -1.f),
    C_Normal3f(-0.999491334f, -0.0318913162f, 0.f),
    C_Vertex3f(0.0356445313f, 0.21252346f, 0.f),
    C_Vertex3f(0.0356445313f, 0.21252346f, -1.f),
    C_Normal3f(-0.99887836f, 0.0473506227f, 0.f),
    C_Vertex3f(0.0356445313f, 0.24495697f, 0.f),
    C_Vertex3f(0.0356445313f, 0.24495697f, -1.f),
    C_Normal3f(-0.990211189f, 0.139577165f, 0.f),
    C_Vertex3f(0.038839817f, 0.279929161f, 0.f),
    C_Vertex3f(0.038839817f, 0.279929161f, -1.f),
    C_Normal3f(-0.970340014f, 0.241744429f, 0.f),
    C_Vertex3f(0.0452303886f, 0.312962532f, 0.f),
    C_Vertex3f(0.0452303886f, 0.312962532f, -1.f),
    C_Normal3f(-0.940288305f, 0.340379149f, 0.f),
    C_Vertex3f(0.054816246f, 0.344057083f, 0.f),
    C_Vertex3f(0.054816246f, 0.344057083f, -1.f),
    C_Normal3f(-0.903909206f, 0.427724332f, 0.f),
    C_Vertex3f(0.0644021034f, 0.365923882f, 0.f),
    C_Vertex3f(0.0644021034f, 0.365923882f, -1.f),
    C_Normal3f(-0.865416706f, 0.501052916f, 0.f),
    C_Vertex3f(0.0716257095f, 0.379580498f, 0.f),
    C_Vertex3f(0.0716257095f, 0.379580498f, -1.f),
    C_Normal3f(-0.822967768f, 0.568088174f, 0.f),
    C_Vertex3f(0.0796823502f, 0.392315865f, 0.f),
    C_Vertex3f(0.0796823502f, 0.392315865f, -1.f),
    C_Normal3f(-0.773390412f, 0.633930027f, 0.f),
    C_Vertex3f(0.0885720253f, 0.404129982f, 0.f),
    C_Vertex3f(0.0885720253f, 0.404129982f, -1.f),
    C_Normal3f(-0.717105925f, 0.696964145f, 0.f),
    C_Vertex3f(0.098294735f, 0.41502285f, 0.f),
    C_Vertex3f(0.098294735f, 0.41502285f, -1.f),
    C_Normal3f(-0.654997408f, 0.755631149f, 0.f),
    C_Vertex3f(0.108850479f, 0.424994469f, 0.f),
    C_Vertex3f(0.108850479f, 0.424994469f, -1.f),
    C_Normal3f(-0.58835423f, 0.808603346f, 0.f),
    C_Vertex3f(0.120239258f, 0.434044838f, 0.f),
    C_Vertex3f(0.120239258f, 0.434044838f, -1.f),
    C_Normal3f(-0.511681736f, 0.859175026f, 0.f),
    C_Vertex3f(0.132461071f, 0.442173958f, 0.f),
    C_Vertex3f(0.132461071f, 0.442173958f, -1.f),
    C_Normal3f(-0.416944027f, 0.90893209f, 0.f),
    C_Vertex3f(0.152043343f, 0.452985764f, 0.f),
    C_Vertex3f(0.152043343f, 0.452985764f, -1.f),
    C_Normal3f(-0.303555429f, 0.952813804f, 0.f),
    C_Vertex3f(0.179600239f, 0.463797569f, 0.f),
    C_Vertex3f(0.179600239f, 0.463797569f, -1.f),
    C_Normal3f(-0.178913206f, 0.983864844f, 0.f),
    C_Vertex3f(0.208604336f, 0.47100544f, 0.f),
    C_Vertex3f(0.208604336f, 0.47100544f, -1.f),
    C_Normal3f(-0.0585227571f, 0.998286128f, 0.f),
    C_Vertex3f(0.239055634f, 0.474609375f, 0.f),
    C_Vertex3f(0.239055634f, 0.474609375f, -1.f),
    C_Normal3f(0.0528280027f, 0.998603642f, 0.f),
    C_Vertex3f(0.270080566f, 0.474609375f, 0.f),
    C_Vertex3f(0.270080566f, 0.474609375f, -1.f),
    C_Normal3f(0.170032248f, 0.985438466f, 0.f),
    C_Vertex3f(0.298293114f, 0.471475601f, 0.f),
    C_Vertex3f(0.298293114f, 0.471475601f, -1.f),
    C_Normal3f(0.295757681f, 0.955263078f, 0.f),
    C_Vertex3f(0.324566841f, 0.465208054f, 0.f),
    C_Vertex3f(0.324566841f, 0.465208054f, -1.f),
    C_Normal3f(0.424930304f, 0.905226052f, 0.f),
    C_Vertex3f(0.348901749f, 0.455806732f, 0.f),
    C_Vertex3f(0.348901749f, 0.455806732f, -1.f),
    C_Normal3f(0.537142038f, 0.843491733f, 0.f),
    C_Vertex3f(0.371297836f, 0.443271637f, 0.f),
    C_Vertex3f(0.371297836f, 0.443271637f, -1.f),
    C_Normal3f(0.637220502f, 0.77068156f, 0.f),
    C_Vertex3f(0.390880585f, 0.429074287f, 0.f),
    C_Vertex3f(0.390880585f, 0.429074287f, -1.f),
    C_Normal3f(0.735771298f, 0.677229941f, 0.f),
    C_Vertex3f(0.407649994f, 0.413214684f, 0.f),
    C_Vertex3f(0.407649994f, 0.413214684f, -1.f),
    C_Normal3f(0.811457455f, 0.584411502f, 0.f),
    C_Vertex3f(0.421606064f, 0.395692825f, 0.f),
    C_Vertex3f(0.421606064f, 0.395692825f, -1.f),
    C_Normal3f(0.864719152f, 0.502255857f, 0.f),
    C_Vertex3f(0.427177429f, 0.386100769f, 0.f),
    C_Vertex3f(0.427177429f, 0.386100769f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.823033094f, 0.567993462f, 0.f),
    C_Vertex3f(0.427177429f, 0.386100769f, 0.f),
    C_Vertex3f(0.427177429f, 0.386100769f, -1.f),
    C_Normal3f(-0.776700437f, 0.629870176f, 0.f),
    C_Vertex3f(0.434173584f, 0.396238327f, 0.f),
    C_Vertex3f(0.434173584f, 0.396238327f, -1.f),
    C_Normal3f(-0.710887611f, 0.703305662f, 0.f),
    C_Vertex3f(0.45016098f, 0.414442062f, 0.f),
    C_Vertex3f(0.45016098f, 0.414442062f, -1.f),
    C_Normal3f(-0.622496188f, 0.782622814f, 0.f),
    C_Vertex3f(0.468143463f, 0.430574417f, 0.f),
    C_Vertex3f(0.468143463f, 0.430574417f, -1.f),
    C_Normal3f(-0.527545512f, 0.849526703f, 0.f),
    C_Vertex3f(0.488121033f, 0.444635391f, 0.f),
    C_Vertex3f(0.488121033f, 0.444635391f, -1.f),
    C_Normal3f(-0.420824677f, 0.907141984f, 0.f),
    C_Vertex3f(0.510093689f, 0.456624985f, 0.f),
    C_Vertex3f(0.510093689f, 0.456624985f, -1.f),
    C_Normal3f(-0.299196303f, 0.954191566f, 0.f),
    C_Vertex3f(0.533349991f, 0.46561718f, 0.f),
    C_Vertex3f(0.533349991f, 0.46561718f, -1.f),
    C_Normal3f(-0.17576611f, 0.984431863f, 0.f),
    C_Vertex3f(0.557889938f, 0.471611977f, 0.f),
    C_Vertex3f(0.557889938f, 0.471611977f, -1.f),
    C_Normal3f(-0.0582138859f, 0.998304248f, 0.f),
    C_Vertex3f(0.583713531f, 0.474609375f, 0.f),
    C_Vertex3f(0.583713531f, 0.474609375f, -1.f),
    C_Normal3f(0.0424444377f, 0.999098897f, 0.f),
    C_Vertex3f(0.60929203f, 0.474609375f, 0.f),
    C_Vertex3f(0.60929203f, 0.474609375f, -1.f),
    C_Normal3f(0.137057498f, 0.990563095f, 0.f),
    C_Vertex3f(0.632193923f, 0.472549796f, 0.f),
    C_Vertex3f(0.632193923f, 0.472549796f, -1.f),
    C_Normal3f(0.236223042f, 0.971698821f, 0.f),
    C_Vertex3f(0.653947949f, 0.468430638f, 0.f),
    C_Vertex3f(0.653947949f, 0.468430638f, -1.f),
    C_Normal3f(0.338591576f, 0.940933466f, 0.f),
    C_Vertex3f(0.67455411f, 0.462251902f, 0.f),
    C_Vertex3f(0.67455411f, 0.462251902f, -1.f),
    C_Normal3f(0.440581292f, 0.897712767f, 0.f),
    C_Vertex3f(0.694012403f, 0.454013586f, 0.f),
    C_Vertex3f(0.694012403f, 0.454013586f, -1.f),
    C_Normal3f(0.538255453f, 0.842781723f, 0.f),
    C_Vertex3f(0.712322831f, 0.443715692f, 0.f),
    C_Vertex3f(0.712322831f, 0.443715692f, -1.f),
    C_Normal3f(0.628017843f, 0.778199017f, 0.f),
    C_Vertex3f(0.729485393f, 0.431358218f, 0.f),
    C_Vertex3f(0.729485393f, 0.431358218f, -1.f),
    C_Normal3f(0.707245946f, 0.706967592f, 0.f),
    C_Vertex3f(0.745500088f, 0.416941166f, 0.f),
    C_Vertex3f(0.745500088f, 0.416941166f, -1.f),
    C_Normal3f(0.782516479f, 0.622629821f, 0.f),
    C_Vertex3f(0.760366917f, 0.400464535f, 0.f),
    C_Vertex3f(0.760366917f, 0.400464535f, -1.f),
    C_Normal3f(0.851457655f, 0.52442354f, 0.f),
    C_Vertex3f(0.773347974f, 0.381942153f, 0.f),
    C_Vertex3f(0.773347974f, 0.381942153f, -1.f),
    C_Normal3f(0.904949069f, 0.425519943f, 0.f),
    C_Vertex3f(0.784443259f, 0.361374021f, 0.f),
    C_Vertex3f(0.784443259f, 0.361374021f, -1.f),
    C_Normal3f(0.94393605f, 0.330128521f, 0.f),
    C_Vertex3f(0.793652773f, 0.338760138f, 0.f),
    C_Vertex3f(0.793652773f, 0.338760138f, -1.f),
    C_Normal3f(0.970494986f, 0.241120875f, 0.f),
    C_Vertex3f(0.800976515f, 0.314100504f, 0.f),
    C_Vertex3f(0.800976515f, 0.314100504f, -1.f),
    C_Normal3f(0.987113297f, 0.160022929f, 0.f),
    C_Vertex3f(0.806414485f, 0.28739512f, 0.f),
    C_Vertex3f(0.806414485f, 0.28739512f, -1.f),
    C_Normal3f(0.996181905f, 0.0873026475f, 0.f),
    C_Vertex3f(0.809966683f, 0.258643985f, 0.f),
    C_Vertex3f(0.809966683f, 0.258643985f, -1.f),
    C_Normal3f(0.999457002f, 0.0329514593f, 0.f),
    C_Vertex3f(0.81163311f, 0.227847099f, 0.f),
    C_Vertex3f(0.81163311f, 0.227847099f, -1.f),
    C_Normal3f(0.999977648f, -0.00667851977f, 0.f),
    C_Vertex3f(0.811523438f, 0.211425781f, 0.f),
    C_Vertex3f(0.811523438f, 0.211425781f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.811523438f, 0.211425781f, 0.f),
    C_Vertex3f(0.811523438f, 0.211425781f, -1.f),
    C_Vertex3f(0.46875f, 0.211425781f, -1.f),
    C_Vertex3f(0.46875f, 0.211425781f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.99983716f, 0.0180472229f, 0.f),
    C_Vertex3f(0.46875f, 0.211425781f, 0.f),
    C_Vertex3f(0.46875f, 0.211425781f, -1.f),
    C_Normal3f(0.997984111f, 0.0634645522f, 0.f),
    C_Vertex3f(0.468914032f, 0.202338219f, 0.f),
    C_Vertex3f(0.468914032f, 0.202338219f, -1.f),
    C_Normal3f(0.992232621f, 0.124396764f, 0.f),
    C_Vertex3f(0.470423222f, 0.185114264f, 0.f),
    C_Vertex3f(0.470423222f, 0.185114264f, -1.f),
    C_Normal3f(0.979106545f, 0.203348026f, 0.f),
    C_Vertex3f(0.473113537f, 0.168841481f, 0.f),
    C_Vertex3f(0.473113537f, 0.168841481f, -1.f),
    C_Normal3f(0.957680643f, 0.287832886f, 0.f),
    C_Vertex3f(0.476984978f, 0.153519869f, 0.f),
    C_Vertex3f(0.476984978f, 0.153519869f, -1.f),
    C_Normal3f(0.926505089f, 0.376282275f, 0.f),
    C_Vertex3f(0.482037544f, 0.139149427f, 0.f),
    C_Vertex3f(0.482037544f, 0.139149427f, -1.f),
    C_Normal3f(0.884585917f, 0.466377228f, 0.f),
    C_Vertex3f(0.488271236f, 0.125730157f, 0.f),
    C_Vertex3f(0.488271236f, 0.125730157f, -1.f),
    C_Normal3f(0.831718147f, 0.555198073f, 0.f),
    C_Vertex3f(0.495686054f, 0.113262057f, 0.f),
    C_Vertex3f(0.495686054f, 0.113262057f, -1.f),
    C_Normal3f(0.768721402f, 0.639583766f, 0.f),
    C_Vertex3f(0.504281998f, 0.101745129f, 0.f),
    C_Vertex3f(0.504281998f, 0.101745129f, -1.f),
    C_Normal3f(0.700869381f, 0.713289678f, 0.f),
    C_Vertex3f(0.514059067f, 0.0911793709f, 0.f),
    C_Vertex3f(0.514059067f, 0.0911793709f, -1.f),
    C_Normal3f(0.626194954f, 0.779666543f, 0.f),
    C_Vertex3f(0.524443865f, 0.0819343328f, 0.f),
    C_Vertex3f(0.524443865f, 0.0819343328f, -1.f),
    C_Normal3f(0.54086262f, 0.841111004f, 0.f),
    C_Vertex3f(0.535436392f, 0.0740100145f, 0.f),
    C_Vertex3f(0.535436392f, 0.0740100145f, -1.f),
    C_Normal3f(0.446683109f, 0.894692183f, 0.f),
    C_Vertex3f(0.547036648f, 0.0674064159f, 0.f),
    C_Vertex3f(0.547036648f, 0.0674064159f, -1.f),
    C_Normal3f(0.346556097f, 0.93802923f, 0.f),
    C_Vertex3f(0.559244633f, 0.0621235371f, 0.f),
    C_Vertex3f(0.559244633f, 0.0621235371f, -1.f),
    C_Normal3f(0.244059205f, 0.969760299f, 0.f),
    C_Vertex3f(0.572060347f, 0.0581613779f, 0.f),
    C_Vertex3f(0.572060347f, 0.0581613779f, -1.f),
    C_Normal3f(0.142836913f, 0.989746273f, 0.f),
    C_Vertex3f(0.585483789f, 0.0555199385f, 0.f),
    C_Vertex3f(0.585483789f, 0.0555199385f, -1.f),
    C_Normal3f(0.0410258323f, 0.999158025f, 0.f),
    C_Vertex3f(0.599514961f, 0.0541992188f, 0.f),
    C_Vertex3f(0.599514961f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0792489648f, 0.996854782f, 0.f),
    C_Vertex3f(0.617649078f, 0.0541992188f, 0.f),
    C_Vertex3f(0.617649078f, 0.0541992188f, -1.f),
    C_Normal3f(-0.229205564f, 0.973378062f, 0.f),
    C_Vertex3f(0.637776375f, 0.057240963f, 0.f),
    C_Vertex3f(0.637776375f, 0.057240963f, -1.f),
    C_Normal3f(-0.391481131f, 0.920186222f, 0.f),
    C_Vertex3f(0.656401634f, 0.0633244514f, 0.f),
    C_Vertex3f(0.656401634f, 0.0633244514f, -1.f),
    C_Normal3f(-0.545138538f, 0.838345945f, 0.f),
    C_Vertex3f(0.673524857f, 0.0724496841f, 0.f),
    C_Vertex3f(0.673524857f, 0.0724496841f, -1.f),
    C_Normal3f(-0.685320377f, 0.728241682f, 0.f),
    C_Vertex3f(0.689146042f, 0.0846166611f, 0.f),
    C_Vertex3f(0.689146042f, 0.0846166611f, -1.f),
    C_Normal3f(-0.80359453f, 0.595177114f, 0.f),
    C_Vertex3f(0.702610016f, 0.0998206139f, 0.f),
    C_Vertex3f(0.702610016f, 0.0998206139f, -1.f),
    C_Normal3f(-0.888075054f, 0.459698349f, 0.f),
    C_Vertex3f(0.713916779f, 0.118061543f, 0.f),
    C_Vertex3f(0.713916779f, 0.118061543f, -1.f),
    C_Normal3f(-0.935337067f, 0.353757739f, 0.f),
    C_Vertex3f(0.72306633f, 0.139339447f, 0.f),
    C_Vertex3f(0.72306633f, 0.139339447f, -1.f),
    C_Normal3f(-0.961050332f, 0.276373595f, 0.f),
    C_Vertex3f(0.7265625f, 0.151496887f, 0.f),
    C_Vertex3f(0.7265625f, 0.151496887f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.143021405f, 0.989719629f, 0.f),
    C_Vertex3f(0.7265625f, 0.151496887f, 0.f),
    C_Vertex3f(0.7265625f, 0.151496887f, -1.f),
    C_Vertex3f(0.805175781f, 0.140136719f, -1.f),
    C_Vertex3f(0.805175781f, 0.140136719f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.96770525f, -0.252084404f, 0.f),
    C_Vertex3f(0.805175781f, 0.140136719f, 0.f),
    C_Vertex3f(0.805175781f, 0.140136719f, -1.f),
    C_Normal3f(0.955178082f, -0.296031833f, 0.f),
    C_Vertex3f(0.802826881f, 0.131119728f, 0.f),
    C_Vertex3f(0.802826881f, 0.131119728f, -1.f),
    C_Normal3f(0.93555671f, -0.353176445f, 0.f),
    C_Vertex3f(0.79711175f, 0.114117265f, 0.f),
    C_Vertex3f(0.79711175f, 0.114117265f, -1.f),
    C_Normal3f(0.905537486f, -0.424266279f, 0.f),
    C_Vertex3f(0.790379286f, 0.0981463194f, 0.f),
    C_Vertex3f(0.790379286f, 0.0981463194f, -1.f),
    C_Normal3f(0.867818713f, -0.496881068f, 0.f),
    C_Vertex3f(0.78262949f, 0.083206892f, 0.f),
    C_Vertex3f(0.78262949f, 0.083206892f, -1.f),
    C_Normal3f(0.822068453f, -0.569388568f, 0.f),
    C_Vertex3f(0.773862362f, 0.0692989826f, 0.f),
    C_Vertex3f(0.773862362f, 0.0692989826f, -1.f),
    C_Normal3f(0.768447459f, -0.639912784f, 0.f),
    C_Vertex3f(0.764077902f, 0.0564225912f, 0.f),
    C_Vertex3f(0.764077902f, 0.0564225912f, -1.f),
    C_Normal3f(0.707689226f, -0.706523895f, 0.f),
    C_Vertex3f(0.75327611f, 0.0445777178f, 0.f),
    C_Vertex3f(0.75327611f, 0.0445777178f, -1.f),
    C_Normal3f(0.64108181f, -0.767472506f, 0.f),
    C_Vertex3f(0.741456985f, 0.0337643623f, 0.f),
    C_Vertex3f(0.741456985f, 0.0337643623f, -1.f),
    C_Normal3f(0.568913996f, -0.822396934f, 0.f),
    C_Vertex3f(0.728620529f, 0.0239825249f, 0.f),
    C_Vertex3f(0.728620529f, 0.0239825249f, -1.f),
    C_Normal3f(0.491130799f, -0.871085763f, 0.f),
    C_Vertex3f(0.714944124f, 0.0154234171f, 0.f),
    C_Vertex3f(0.714944124f, 0.0154234171f, -1.f),
    C_Normal3f(0.410550773f, -0.911837757f, 0.f),
    C_Vertex3f(0.700427771f, 0.00808703899f, 0.f),
    C_Vertex3f(0.700427771f, 0.00808703899f, -1.f),
    C_Normal3f(0.329315126f, -0.944220066f, 0.f),
    C_Vertex3f(0.685071468f, 0.00197339058f, 0.f),
    C_Vertex3f(0.685071468f, 0.00197339058f, -1.f),
    C_Normal3f(0.249413311f, -0.968397141f, 0.f),
    C_Vertex3f(0.668875217f, -0.00291752815f, 0.f),
    C_Vertex3f(0.668875217f, -0.00291752815f, -1.f),
    C_Normal3f(0.172489494f, -0.985011339f, 0.f),
    C_Vertex3f(0.651839018f, -0.0065857172f, 0.f),
    C_Vertex3f(0.651839018f, -0.0065857172f, -1.f),
    C_Normal3f(0.099745065f, -0.995012999f, 0.f),
    C_Vertex3f(0.63396287f, -0.00903117657f, 0.f),
    C_Vertex3f(0.63396287f, -0.00903117657f, -1.f),
    C_Normal3f(0.0283364989f, -0.999598503f, 0.f),
    C_Vertex3f(0.615246773f, -0.0102539063f, 0.f),
    C_Vertex3f(0.615246773f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0547810048f, -0.998498321f, 0.f),
    C_Vertex3f(0.590829849f, -0.0102539063f, 0.f),
    C_Vertex3f(0.590829849f, -0.0102539063f, -1.f),
    C_Normal3f(-0.157117173f, -0.987579942f, 0.f),
    C_Vertex3f(0.563081264f, -0.00739192963f, 0.f),
    C_Vertex3f(0.563081264f, -0.00739192963f, -1.f),
    C_Normal3f(-0.270598024f, -0.96269238f, 0.f),
    C_Vertex3f(0.536861897f, -0.00166797638f, 0.f),
    C_Vertex3f(0.536861897f, -0.00166797638f, -1.f),
    C_Normal3f(-0.386189759f, -0.922419369f, 0.f),
    C_Vertex3f(0.512171745f, 0.00691795349f, 0.f),
    C_Vertex3f(0.512171745f, 0.00691795349f, -1.f),
    C_Normal3f(-0.497074097f, -0.867708147f, 0.f),
    C_Vertex3f(0.489010811f, 0.01836586f, 0.f),
    C_Vertex3f(0.489010811f, 0.01836586f, -1.f),
    C_Normal3f(-0.604578614f, -0.796545506f, 0.f),
    C_Vertex3f(0.468253613f, 0.0320768356f, 0.f),
    C_Vertex3f(0.468253613f, 0.0320768356f, -1.f),
    C_Normal3f(-0.706156313f, -0.708055913f, 0.f),
    C_Vertex3f(0.44990015f, 0.0480508804f, 0.f),
    C_Vertex3f(0.44990015f, 0.0480508804f, -1.f),
    C_Normal3f(-0.781765521f, -0.623572528f, 0.f),
    C_Vertex3f(0.433950424f, 0.0662879944f, 0.f),
    C_Vertex3f(0.433950424f, 0.0662879944f, -1.f),
    C_Normal3f(-0.834312439f, -0.551292062f, 0.f),
    C_Vertex3f(0.427177429f, 0.0765380859f, 0.f),
    C_Vertex3f(0.427177429f, 0.0765380859f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.998995781f, -0.0448056757f, 0.f),
    C_Vertex3f(0.116699219f, 0.24562645f, 0.f),
    C_Vertex3f(0.116699219f, 0.24562645f, -1.f),
    C_Normal3f(0.999732256f, 0.0231370199f, 0.f),
    C_Vertex3f(0.116699219f, 0.222887993f, 0.f),
    C_Vertex3f(0.116699219f, 0.222887993f, -1.f),
    C_Normal3f(0.996932983f, 0.0782596841f, 0.f),
    C_Vertex3f(0.117667913f, 0.203769922f, 0.f),
    C_Vertex3f(0.117667913f, 0.203769922f, -1.f),
    C_Normal3f(0.990337014f, 0.138681725f, 0.f),
    C_Vertex3f(0.119605303f, 0.185868025f, 0.f),
    C_Vertex3f(0.119605303f, 0.185868025f, -1.f),
    C_Normal3f(0.97848022f, 0.206340641f, 0.f),
    C_Vertex3f(0.122511387f, 0.169182301f, 0.f),
    C_Vertex3f(0.122511387f, 0.169182301f, -1.f),
    C_Normal3f(0.95957315f, 0.281459481f, 0.f),
    C_Vertex3f(0.126386166f, 0.153712749f, 0.f),
    C_Vertex3f(0.126386166f, 0.153712749f, -1.f),
    C_Normal3f(0.931512237f, 0.363710105f, 0.f),
    C_Vertex3f(0.131229639f, 0.139459372f, 0.f),
    C_Vertex3f(0.131229639f, 0.139459372f, -1.f),
    C_Normal3f(0.892061949f, 0.451913148f, 0.f),
    C_Vertex3f(0.137041807f, 0.126422167f, 0.f),
    C_Vertex3f(0.137041807f, 0.126422167f, -1.f),
    C_Normal3f(0.83734715f, 0.54667145f, 0.f),
    C_Vertex3f(0.14382267f, 0.114601135f, 0.f),
    C_Vertex3f(0.14382267f, 0.114601135f, -1.f),
    C_Normal3f(0.776299417f, 0.630364418f, 0.f),
    C_Vertex3f(0.15271759f, 0.10241127f, 0.f),
    C_Vertex3f(0.15271759f, 0.10241127f, -1.f),
    C_Normal3f(0.707939506f, 0.706273079f, 0.f),
    C_Vertex3f(0.163508177f, 0.0903582573f, 0.f),
    C_Vertex3f(0.163508177f, 0.0903582573f, -1.f),
    C_Normal3f(0.622183681f, 0.782871246f, 0.f),
    C_Vertex3f(0.175049067f, 0.0800271034f, 0.f),
    C_Vertex3f(0.175049067f, 0.0800271034f, -1.f),
    C_Normal3f(0.521833479f, 0.853047371f, 0.f),
    C_Vertex3f(0.18734026f, 0.0714178085f, 0.f),
    C_Vertex3f(0.18734026f, 0.0714178085f, -1.f),
    C_Normal3f(0.409743488f, 0.912200809f, 0.f),
    C_Vertex3f(0.200381756f, 0.0645303726f, 0.f),
    C_Vertex3f(0.200381756f, 0.0645303726f, -1.f),
    C_Normal3f(0.29072684f, 0.956806064f, 0.f),
    C_Vertex3f(0.214173555f, 0.0593647957f, 0.f),
    C_Vertex3f(0.214173555f, 0.0593647957f, -1.f),
    C_Normal3f(0.170602739f, 0.98533988f, 0.f),
    C_Vertex3f(0.228715658f, 0.0559210777f, 0.f),
    C_Vertex3f(0.228715658f, 0.0559210777f, -1.f),
    C_Normal3f(0.0566352904f, 0.998395026f, 0.f),
    C_Vertex3f(0.244008064f, 0.0541992188f, 0.f),
    C_Vertex3f(0.244008064f, 0.0541992188f, -1.f),
    C_Normal3f(-0.050518211f, 0.998723149f, 0.f),
    C_Vertex3f(0.259069443f, 0.0541992188f, 0.f),
    C_Vertex3f(0.259069443f, 0.0541992188f, -1.f),
    C_Normal3f(-0.160962418f, 0.98696059f, 0.f),
    C_Vertex3f(0.27264452f, 0.0556477308f, 0.f),
    C_Vertex3f(0.27264452f, 0.0556477308f, -1.f),
    C_Normal3f(-0.271881163f, 0.962330878f, 0.f),
    C_Vertex3f(0.285714626f, 0.058544755f, 0.f),
    C_Vertex3f(0.285714626f, 0.058544755f, -1.f),
    C_Normal3f(-0.380741924f, 0.924681365f, 0.f),
    C_Vertex3f(0.298279762f, 0.0628902912f, 0.f),
    C_Vertex3f(0.298279762f, 0.0628902912f, -1.f),
    C_Normal3f(-0.483288914f, 0.875460923f, 0.f),
    C_Vertex3f(0.310339928f, 0.0686843395f, 0.f),
    C_Vertex3f(0.310339928f, 0.0686843395f, -1.f),
    C_Normal3f(-0.576124787f, 0.817361772f, 0.f),
    C_Vertex3f(0.321895123f, 0.0759268999f, 0.f),
    C_Vertex3f(0.321895123f, 0.0759268999f, -1.f),
    C_Normal3f(-0.657211125f, 0.753706515f, 0.f),
    C_Vertex3f(0.332945347f, 0.0846179724f, 0.f),
    C_Vertex3f(0.332945347f, 0.0846179724f, -1.f),
    C_Normal3f(-0.725929022f, 0.687769592f, 0.f),
    C_Vertex3f(0.343490601f, 0.0947575569f, 0.f),
    C_Vertex3f(0.343490601f, 0.0947575569f, -1.f),
    C_Normal3f(-0.792309344f, 0.610119581f, 0.f),
    C_Vertex3f(0.353530884f, 0.106345654f, 0.f),
    C_Vertex3f(0.353530884f, 0.106345654f, -1.f),
    C_Normal3f(-0.855779588f, 0.51734072f, 0.f),
    C_Vertex3f(0.362316132f, 0.11920464f, 0.f),
    C_Vertex3f(0.362316132f, 0.11920464f, -1.f),
    C_Normal3f(-0.905895829f, 0.423500568f, 0.f),
    C_Vertex3f(0.369846344f, 0.133334517f, 0.f),
    C_Vertex3f(0.369846344f, 0.133334517f, -1.f),
    C_Normal3f(-0.943215013f, 0.332183003f, 0.f),
    C_Vertex3f(0.376121521f, 0.148735285f, 0.f),
    C_Vertex3f(0.376121521f, 0.148735285f, -1.f),
    C_Normal3f(-0.96926862f, 0.246004924f, 0.f),
    C_Vertex3f(0.381141663f, 0.165406942f, 0.f),
    C_Vertex3f(0.381141663f, 0.165406942f, -1.f),
    C_Normal3f(-0.986036479f, 0.166529119f, 0.f),
    C_Vertex3f(0.384906769f, 0.18334949f, 0.f),
    C_Vertex3f(0.384906769f, 0.18334949f, -1.f),
    C_Normal3f(-0.995532393f, 0.0944205523f, 0.f),
    C_Vertex3f(0.38741684f, 0.202562928f, 0.f),
    C_Vertex3f(0.38741684f, 0.202562928f, -1.f),
    C_Normal3f(-0.999557793f, 0.0297374632f, 0.f),
    C_Vertex3f(0.388671875f, 0.223047256f, 0.f),
    C_Vertex3f(0.388671875f, 0.223047256f, -1.f),
    C_Normal3f(-0.999569058f, -0.0293542426f, 0.f),
    C_Vertex3f(0.388671875f, 0.244748116f, 0.f),
    C_Vertex3f(0.388671875f, 0.244748116f, -1.f),
    C_Normal3f(-0.99560523f, -0.0936488137f, 0.f),
    C_Vertex3f(0.387437224f, 0.265089512f, 0.f),
    C_Vertex3f(0.387437224f, 0.265089512f, -1.f),
    C_Normal3f(-0.986196041f, -0.16558212f, 0.f),
    C_Vertex3f(0.384967923f, 0.284125805f, 0.f),
    C_Vertex3f(0.384967923f, 0.284125805f, -1.f),
    C_Normal3f(-0.969449162f, -0.245292306f, 0.f),
    C_Vertex3f(0.381263971f, 0.301856995f, 0.f),
    C_Vertex3f(0.381263971f, 0.301856995f, -1.f),
    C_Normal3f(-0.943200409f, -0.332224399f, 0.f),
    C_Vertex3f(0.376325369f, 0.318283081f, 0.f),
    C_Vertex3f(0.376325369f, 0.318283081f, -1.f),
    C_Normal3f(-0.905254364f, -0.424869895f, 0.f),
    C_Vertex3f(0.370152116f, 0.333404064f, 0.f),
    C_Vertex3f(0.370152116f, 0.333404064f, -1.f),
    C_Normal3f(-0.853830516f, -0.520551145f, 0.f),
    C_Vertex3f(0.362744212f, 0.347219944f, 0.f),
    C_Vertex3f(0.362744212f, 0.347219944f, -1.f),
    C_Normal3f(-0.788164914f, -0.615464151f, 0.f),
    C_Vertex3f(0.354101658f, 0.359730721f, 0.f),
    C_Vertex3f(0.354101658f, 0.359730721f, -1.f),
    C_Normal3f(-0.718813062f, -0.695203364f, 0.f),
    C_Vertex3f(0.344224453f, 0.370936394f, 0.f),
    C_Vertex3f(0.344224453f, 0.370936394f, -1.f),
    C_Normal3f(-0.647119105f, -0.762388945f, 0.f),
    C_Vertex3f(0.333781123f, 0.380741358f, 0.f),
    C_Vertex3f(0.333781123f, 0.380741358f, -1.f),
    C_Normal3f(-0.563559353f, -0.826075554f, 0.f),
    C_Vertex3f(0.322771668f, 0.389145613f, 0.f),
    C_Vertex3f(0.322771668f, 0.389145613f, -1.f),
    C_Normal3f(-0.469346106f, -0.883014202f, 0.f),
    C_Vertex3f(0.311196089f, 0.396149158f, 0.f),
    C_Vertex3f(0.311196089f, 0.396149158f, -1.f),
    C_Normal3f(-0.367034972f, -0.930207193f, 0.f),
    C_Vertex3f(0.299054384f, 0.401751995f, 0.f),
    C_Vertex3f(0.299054384f, 0.401751995f, -1.f),
    C_Normal3f(-0.260266274f, -0.965536892f, 0.f),
    C_Vertex3f(0.286346555f, 0.405954123f, 0.f),
    C_Vertex3f(0.286346555f, 0.405954123f, -1.f),
    C_Normal3f(-0.153151453f, -0.988202691f, 0.f),
    C_Vertex3f(0.2730726f, 0.408755541f, 0.f),
    C_Vertex3f(0.2730726f, 0.408755541f, -1.f),
    C_Normal3f(-0.0458196253f, -0.998949766f, 0.f),
    C_Vertex3f(0.259232521f, 0.41015625f, 0.f),
    C_Vertex3f(0.259232521f, 0.41015625f, -1.f),
    C_Normal3f(0.076468505f, -0.997071981f, 0.f),
    C_Vertex3f(0.242534637f, 0.41015625f, 0.f),
    C_Vertex3f(0.242534637f, 0.41015625f, -1.f),
    C_Normal3f(0.219655499f, -0.975577533f, 0.f),
    C_Vertex3f(0.224308968f, 0.407477856f, 0.f),
    C_Vertex3f(0.224308968f, 0.407477856f, -1.f),
    C_Normal3f(0.364993691f, -0.931010067f, 0.f),
    C_Vertex3f(0.206847191f, 0.402121067f, 0.f),
    C_Vertex3f(0.206847191f, 0.402121067f, -1.f),
    C_Normal3f(0.498180181f, -0.867073536f, 0.f),
    C_Vertex3f(0.190149307f, 0.394085884f, 0.f),
    C_Vertex3f(0.190149307f, 0.394085884f, -1.f),
    C_Normal3f(0.624822259f, -0.780767083f, 0.f),
    C_Vertex3f(0.174215317f, 0.383372307f, 0.f),
    C_Vertex3f(0.174215317f, 0.383372307f, -1.f),
    C_Normal3f(0.74330622f, -0.668951273f, 0.f),
    C_Vertex3f(0.160163879f, 0.370089531f, 0.f),
    C_Vertex3f(0.160163879f, 0.370089531f, -1.f),
    C_Normal3f(0.836456716f, -0.548032999f, 0.f),
    C_Vertex3f(0.147994995f, 0.354237556f, 0.f),
    C_Vertex3f(0.147994995f, 0.354237556f, -1.f),
    C_Normal3f(0.903547227f, -0.428488553f, 0.f),
    C_Vertex3f(0.137708664f, 0.335816383f, 0.f),
    C_Vertex3f(0.137708664f, 0.335816383f, -1.f),
    C_Normal3f(0.946233749f, -0.323483825f, 0.f),
    C_Vertex3f(0.129304886f, 0.314826012f, 0.f),
    C_Vertex3f(0.129304886f, 0.314826012f, -1.f),
    C_Normal3f(0.973924458f, -0.226872459f, 0.f),
    C_Vertex3f(0.123002052f, 0.292797565f, 0.f),
    C_Vertex3f(0.123002052f, 0.292797565f, -1.f),
    C_Normal3f(0.991191089f, -0.13243936f, 0.f),
    C_Vertex3f(0.118800163f, 0.269731045f, 0.f),
    C_Vertex3f(0.118800163f, 0.269731045f, -1.f),
    C_Normal3f(0.998995781f, -0.0448056757f, 0.f),
    C_Vertex3f(0.116699219f, 0.24562645f, 0.f),
    C_Vertex3f(0.116699219f, 0.24562645f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.471679688f, 0.275878906f, 0.f),
    C_Vertex3f(0.471679688f, 0.275878906f, -1.f),
    C_Vertex3f(0.728515625f, 0.275878906f, -1.f),
    C_Vertex3f(0.728515625f, 0.275878906f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.996856868f, -0.0792234614f, 0.f),
    C_Vertex3f(0.728515625f, 0.275878906f, 0.f),
    C_Vertex3f(0.728515625f, 0.275878906f, -1.f),
    C_Normal3f(-0.992682695f, -0.120751664f, 0.f),
    C_Vertex3f(0.727859497f, 0.284134865f, 0.f),
    C_Vertex3f(0.727859497f, 0.284134865f, -1.f),
    C_Normal3f(-0.984250665f, -0.176778391f, 0.f),
    C_Vertex3f(0.725624204f, 0.299648881f, 0.f),
    C_Vertex3f(0.725624204f, 0.299648881f, -1.f),
    C_Normal3f(-0.968235314f, -0.250041008f, 0.f),
    C_Vertex3f(0.722465873f, 0.314164996f, 0.f),
    C_Vertex3f(0.722465873f, 0.314164996f, -1.f),
    C_Normal3f(-0.94417125f, -0.329455167f, 0.f),
    C_Vertex3f(0.718384504f, 0.32768321f, 0.f),
    C_Vertex3f(0.718384504f, 0.32768321f, -1.f),
    C_Normal3f(-0.910318851f, -0.413907677f, 0.f),
    C_Vertex3f(0.713380098f, 0.340203524f, 0.f),
    C_Vertex3f(0.713380098f, 0.340203524f, -1.f),
    C_Normal3f(-0.865189373f, -0.501445234f, 0.f),
    C_Vertex3f(0.707452655f, 0.351725936f, 0.f),
    C_Vertex3f(0.707452655f, 0.351725936f, -1.f),
    C_Normal3f(-0.807942569f, -0.589261353f, 0.f),
    C_Vertex3f(0.700602174f, 0.362250447f, 0.f),
    C_Vertex3f(0.700602174f, 0.362250447f, -1.f),
    C_Normal3f(-0.7387954f, -0.673929691f, 0.f),
    C_Vertex3f(0.692828655f, 0.371777058f, 0.f),
    C_Vertex3f(0.692828655f, 0.371777058f, -1.f),
    C_Normal3f(-0.665826857f, -0.746106327f, 0.f),
    C_Vertex3f(0.684132099f, 0.380305767f, 0.f),
    C_Vertex3f(0.684132099f, 0.380305767f, -1.f),
    C_Normal3f(-0.589877009f, -0.807493091f, 0.f),
    C_Vertex3f(0.674909234f, 0.387768388f, 0.f),
    C_Vertex3f(0.674909234f, 0.387768388f, -1.f),
    C_Normal3f(-0.505346894f, -0.862916291f, 0.f),
    C_Vertex3f(0.66516006f, 0.39416492f, 0.f),
    C_Vertex3f(0.66516006f, 0.39416492f, -1.f),
    C_Normal3f(-0.414311707f, -0.910135031f, 0.f),
    C_Vertex3f(0.654884577f, 0.399495363f, 0.f),
    C_Vertex3f(0.654884577f, 0.399495363f, -1.f),
    C_Normal3f(-0.319539607f, -0.947572887f, 0.f),
    C_Vertex3f(0.644082785f, 0.403759718f, 0.f),
    C_Vertex3f(0.644082785f, 0.403759718f, -1.f),
    C_Normal3f(-0.224086434f, -0.974569261f, 0.f),
    C_Vertex3f(0.632754683f, 0.406957984f, 0.f),
    C_Vertex3f(0.632754683f, 0.406957984f, -1.f),
    C_Normal3f(-0.130833834f, -0.991404295f, 0.f),
    C_Vertex3f(0.620900273f, 0.409090161f, 0.f),
    C_Vertex3f(0.620900273f, 0.409090161f, -1.f),
    C_Normal3f(-0.0419393256f, -0.999120176f, 0.f),
    C_Vertex3f(0.608519554f, 0.41015625f, 0.f),
    C_Vertex3f(0.608519554f, 0.41015625f, -1.f),
    C_Normal3f(0.0436647646f, -0.999046266f, 0.f),
    C_Vertex3f(0.595502853f, 0.41015625f, 0.f),
    C_Vertex3f(0.595502853f, 0.41015625f, -1.f),
    C_Normal3f(0.13459f, -0.990901351f, 0.f),
    C_Vertex3f(0.582875609f, 0.409035444f, 0.f),
    C_Vertex3f(0.582875609f, 0.409035444f, -1.f),
    C_Normal3f(0.229587525f, -0.973288f, 0.f),
    C_Vertex3f(0.570747495f, 0.406793833f, 0.f),
    C_Vertex3f(0.570747495f, 0.406793833f, -1.f),
    C_Normal3f(0.325907648f, -0.945401609f, 0.f),
    C_Vertex3f(0.559118509f, 0.403431416f, 0.f),
    C_Vertex3f(0.559118509f, 0.403431416f, -1.f),
    C_Normal3f(0.420566708f, -0.90726167f, 0.f),
    C_Vertex3f(0.547988653f, 0.398948193f, 0.f),
    C_Vertex3f(0.547988653f, 0.398948193f, -1.f),
    C_Normal3f(0.510576487f, -0.859832346f, 0.f),
    C_Vertex3f(0.537357926f, 0.393344164f, 0.f),
    C_Vertex3f(0.537357926f, 0.393344164f, -1.f),
    C_Normal3f(0.593396425f, -0.804910302f, 0.f),
    C_Vertex3f(0.527226329f, 0.386619329f, 0.f),
    C_Vertex3f(0.527226329f, 0.386619329f, -1.f),
    C_Normal3f(0.667271435f, -0.744814575f, 0.f),
    C_Vertex3f(0.517593861f, 0.378773689f, 0.f),
    C_Vertex3f(0.517593861f, 0.378773689f, -1.f),
    C_Normal3f(0.735732853f, -0.677271843f, 0.f),
    C_Vertex3f(0.508460522f, 0.369807243f, 0.f),
    C_Vertex3f(0.508460522f, 0.369807243f, -1.f),
    C_Normal3f(0.800612688f, -0.599182129f, 0.f),
    C_Vertex3f(0.500318408f, 0.360007048f, 0.f),
    C_Vertex3f(0.500318408f, 0.360007048f, -1.f),
    C_Normal3f(0.856644213f, -0.515907705f, 0.f),
    C_Vertex3f(0.49316752f, 0.349373102f, 0.f),
    C_Vertex3f(0.49316752f, 0.349373102f, -1.f),
    C_Normal3f(0.902720928f, -0.430226505f, 0.f),
    C_Vertex3f(0.487007856f, 0.337905407f, 0.f),
    C_Vertex3f(0.487007856f, 0.337905407f, -1.f),
    C_Normal3f(0.938650846f, -0.344868779f, 0.f),
    C_Vertex3f(0.481839418f, 0.325603962f, 0.f),
    C_Vertex3f(0.481839418f, 0.325603962f, -1.f),
    C_Normal3f(0.965023935f, -0.262161613f, 0.f),
    C_Vertex3f(0.477662206f, 0.312468767f, 0.f),
    C_Vertex3f(0.477662206f, 0.312468767f, -1.f),
    C_Normal3f(0.982958257f, -0.183828682f, 0.f),
    C_Vertex3f(0.474476218f, 0.298499823f, 0.f),
    C_Vertex3f(0.474476218f, 0.298499823f, -1.f),
    C_Normal3f(0.992444873f, -0.122691877f, 0.f),
    C_Vertex3f(0.472281456f, 0.283697128f, 0.f),
    C_Vertex3f(0.472281456f, 0.283697128f, -1.f),
    C_Normal3f(0.997050941f, -0.0767429993f, 0.f),
    C_Vertex3f(0.471679688f, 0.275878906f, 0.f),
    C_Vertex3f(0.471679688f, 0.275878906f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.725624204f, 0.299648881f, 0.f),
    C_Vertex3f(0.727859497f, 0.284134865f, 0.f),
    C_Vertex3f(0.729485393f, 0.431358218f, 0.f),
    C_Vertex3f(0.728515625f, 0.275878906f, 0.f),
    C_Vertex3f(0.811523438f, 0.211425781f, 0.f),
    C_Vertex3f(0.471679688f, 0.275878906f, 0.f),
    C_Vertex3f(0.46875f, 0.211425781f, 0.f),
    C_Vertex3f(0.468143463f, 0.430574417f, 0.f),
    C_Vertex3f(0.468253613f, 0.0320768356f, 0.f),
    C_Vertex3f(0.45016098f, 0.414442062f, 0.f),
    C_Vertex3f(0.44990015f, 0.0480508804f, 0.f),
    C_Vertex3f(0.434173584f, 0.396238327f, 0.f),
    C_Vertex3f(0.433950424f, 0.0662879944f, 0.f),
    C_Vertex3f(0.427177429f, 0.386100769f, 0.f),
    C_Vertex3f(0.427177429f, 0.0765380859f, 0.f),
    C_Vertex3f(0.421606064f, 0.395692825f, 0.f),
    C_Vertex3f(0.418874741f, 0.0657444f, 0.f),
    C_Vertex3f(0.407649994f, 0.413214684f, 0.f),
    C_Vertex3f(0.401012897f, 0.0468277931f, 0.f),
    C_Vertex3f(0.390880585f, 0.429074287f, 0.f),
    C_Vertex3f(0.388671875f, 0.244748116f, 0.f),
    C_Vertex3f(0.387437224f, 0.265089512f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.390880585f, 0.429074287f, 0.f),
    C_Vertex3f(0.371297836f, 0.443271637f, 0.f),
    C_Vertex3f(0.376325369f, 0.318283081f, 0.f),
    C_Vertex3f(0.381263971f, 0.301856995f, 0.f),
    C_Vertex3f(0.384967923f, 0.284125805f, 0.f),
    C_Vertex3f(0.387437224f, 0.265089512f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.371297836f, 0.443271637f, 0.f),
    C_Vertex3f(0.348901749f, 0.455806732f, 0.f),
    C_Vertex3f(0.354101658f, 0.359730721f, 0.f),
    C_Vertex3f(0.362744212f, 0.347219944f, 0.f),
    C_Vertex3f(0.370152116f, 0.333404064f, 0.f),
    C_Vertex3f(0.376325369f, 0.318283081f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.311196089f, 0.396149158f, 0.f),
    C_Vertex3f(0.322771668f, 0.389145613f, 0.f),
    C_Vertex3f(0.324566841f, 0.465208054f, 0.f),
    C_Vertex3f(0.333781123f, 0.380741358f, 0.f),
    C_Vertex3f(0.348901749f, 0.455806732f, 0.f),
    C_Vertex3f(0.344224453f, 0.370936394f, 0.f),
    C_Vertex3f(0.354101658f, 0.359730721f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.2730726f, 0.408755541f, 0.f),
    C_Vertex3f(0.286346555f, 0.405954123f, 0.f),
    C_Vertex3f(0.298293114f, 0.471475601f, 0.f),
    C_Vertex3f(0.299054384f, 0.401751995f, 0.f),
    C_Vertex3f(0.324566841f, 0.465208054f, 0.f),
    C_Vertex3f(0.311196089f, 0.396149158f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.270080566f, 0.474609375f, 0.f),
    C_Vertex3f(0.239055634f, 0.474609375f, 0.f),
    C_Vertex3f(0.242534637f, 0.41015625f, 0.f),
    C_Vertex3f(0.259232521f, 0.41015625f, 0.f),
    C_Vertex3f(0.2730726f, 0.408755541f, 0.f),
    C_Vertex3f(0.298293114f, 0.471475601f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.190149307f, 0.394085884f, 0.f),
    C_Vertex3f(0.206847191f, 0.402121067f, 0.f),
    C_Vertex3f(0.208604336f, 0.47100544f, 0.f),
    C_Vertex3f(0.224308968f, 0.407477856f, 0.f),
    C_Vertex3f(0.239055634f, 0.474609375f, 0.f),
    C_Vertex3f(0.242534637f, 0.41015625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179600239f, 0.463797569f, 0.f),
    C_Vertex3f(0.152043343f, 0.452985764f, 0.f),
    C_Vertex3f(0.160163879f, 0.370089531f, 0.f),
    C_Vertex3f(0.174215317f, 0.383372307f, 0.f),
    C_Vertex3f(0.190149307f, 0.394085884f, 0.f),
    C_Vertex3f(0.208604336f, 0.47100544f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.123002052f, 0.292797565f, 0.f),
    C_Vertex3f(0.129304886f, 0.314826012f, 0.f),
    C_Vertex3f(0.132461071f, 0.442173958f, 0.f),
    C_Vertex3f(0.137708664f, 0.335816383f, 0.f),
    C_Vertex3f(0.152043343f, 0.452985764f, 0.f),
    C_Vertex3f(0.147994995f, 0.354237556f, 0.f),
    C_Vertex3f(0.160163879f, 0.370089531f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120239258f, 0.434044838f, 0.f),
    C_Vertex3f(0.108850479f, 0.424994469f, 0.f),
    C_Vertex3f(0.116699219f, 0.24562645f, 0.f),
    C_Vertex3f(0.118800163f, 0.269731045f, 0.f),
    C_Vertex3f(0.123002052f, 0.292797565f, 0.f),
    C_Vertex3f(0.132461071f, 0.442173958f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.108850479f, 0.424994469f, 0.f),
    C_Vertex3f(0.098294735f, 0.41502285f, 0.f),
    C_Vertex3f(0.10324049f, 0.044403553f, 0.f),
    C_Vertex3f(0.119456649f, 0.0307391882f, 0.f),
    C_Vertex3f(0.116699219f, 0.222887993f, 0.f),
    C_Vertex3f(0.116699219f, 0.24562645f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.729485393f, 0.431358218f, 0.f),
    C_Vertex3f(0.712322831f, 0.443715692f, 0.f),
    C_Vertex3f(0.713380098f, 0.340203524f, 0.f),
    C_Vertex3f(0.718384504f, 0.32768321f, 0.f),
    C_Vertex3f(0.722465873f, 0.314164996f, 0.f),
    C_Vertex3f(0.725624204f, 0.299648881f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.684132099f, 0.380305767f, 0.f),
    C_Vertex3f(0.692828655f, 0.371777058f, 0.f),
    C_Vertex3f(0.694012403f, 0.454013586f, 0.f),
    C_Vertex3f(0.700602174f, 0.362250447f, 0.f),
    C_Vertex3f(0.712322831f, 0.443715692f, 0.f),
    C_Vertex3f(0.707452655f, 0.351725936f, 0.f),
    C_Vertex3f(0.713380098f, 0.340203524f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.654884577f, 0.399495363f, 0.f),
    C_Vertex3f(0.66516006f, 0.39416492f, 0.f),
    C_Vertex3f(0.67455411f, 0.462251902f, 0.f),
    C_Vertex3f(0.674909234f, 0.387768388f, 0.f),
    C_Vertex3f(0.694012403f, 0.454013586f, 0.f),
    C_Vertex3f(0.684132099f, 0.380305767f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.653947949f, 0.468430638f, 0.f),
    C_Vertex3f(0.632193923f, 0.472549796f, 0.f),
    C_Vertex3f(0.632754683f, 0.406957984f, 0.f),
    C_Vertex3f(0.644082785f, 0.403759718f, 0.f),
    C_Vertex3f(0.654884577f, 0.399495363f, 0.f),
    C_Vertex3f(0.67455411f, 0.462251902f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.595502853f, 0.41015625f, 0.f),
    C_Vertex3f(0.608519554f, 0.41015625f, 0.f),
    C_Vertex3f(0.60929203f, 0.474609375f, 0.f),
    C_Vertex3f(0.620900273f, 0.409090161f, 0.f),
    C_Vertex3f(0.632193923f, 0.472549796f, 0.f),
    C_Vertex3f(0.632754683f, 0.406957984f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.583713531f, 0.474609375f, 0.f),
    C_Vertex3f(0.557889938f, 0.471611977f, 0.f),
    C_Vertex3f(0.559118509f, 0.403431416f, 0.f),
    C_Vertex3f(0.570747495f, 0.406793833f, 0.f),
    C_Vertex3f(0.582875609f, 0.409035444f, 0.f),
    C_Vertex3f(0.595502853f, 0.41015625f, 0.f),
    C_Vertex3f(0.60929203f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.517593861f, 0.378773689f, 0.f),
    C_Vertex3f(0.527226329f, 0.386619329f, 0.f),
    C_Vertex3f(0.533349991f, 0.46561718f, 0.f),
    C_Vertex3f(0.537357926f, 0.393344164f, 0.f),
    C_Vertex3f(0.557889938f, 0.471611977f, 0.f),
    C_Vertex3f(0.547988653f, 0.398948193f, 0.f),
    C_Vertex3f(0.559118509f, 0.403431416f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.510093689f, 0.456624985f, 0.f),
    C_Vertex3f(0.488121033f, 0.444635391f, 0.f),
    C_Vertex3f(0.49316752f, 0.349373102f, 0.f),
    C_Vertex3f(0.500318408f, 0.360007048f, 0.f),
    C_Vertex3f(0.508460522f, 0.369807243f, 0.f),
    C_Vertex3f(0.517593861f, 0.378773689f, 0.f),
    C_Vertex3f(0.533349991f, 0.46561718f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.488121033f, 0.444635391f, 0.f),
    C_Vertex3f(0.468143463f, 0.430574417f, 0.f),
    C_Vertex3f(0.472281456f, 0.283697128f, 0.f),
    C_Vertex3f(0.474476218f, 0.298499823f, 0.f),
    C_Vertex3f(0.477662206f, 0.312468767f, 0.f),
    C_Vertex3f(0.481839418f, 0.325603962f, 0.f),
    C_Vertex3f(0.487007856f, 0.337905407f, 0.f),
    C_Vertex3f(0.49316752f, 0.349373102f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.401012897f, 0.0468277931f, 0.f),
    C_Vertex3f(0.388671875f, 0.244748116f, 0.f),
    C_Vertex3f(0.388671875f, 0.223047256f, 0.f),
    C_Vertex3f(0.38741684f, 0.202562928f, 0.f),
    C_Vertex3f(0.384906769f, 0.18334949f, 0.f),
    C_Vertex3f(0.381894588f, 0.0305819511f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.381894588f, 0.0305819511f, 0.f),
    C_Vertex3f(0.384906769f, 0.18334949f, 0.f),
    C_Vertex3f(0.381141663f, 0.165406942f, 0.f),
    C_Vertex3f(0.376121521f, 0.148735285f, 0.f),
    C_Vertex3f(0.369846344f, 0.133334517f, 0.f),
    C_Vertex3f(0.362316132f, 0.11920464f, 0.f),
    C_Vertex3f(0.361519814f, 0.0170068741f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362316132f, 0.11920464f, 0.f),
    C_Vertex3f(0.353530884f, 0.106345654f, 0.f),
    C_Vertex3f(0.361519814f, 0.0170068741f, 0.f),
    C_Vertex3f(0.343490601f, 0.0947575569f, 0.f),
    C_Vertex3f(0.339888573f, 0.00610256195f, 0.f),
    C_Vertex3f(0.332945347f, 0.0846179724f, 0.f),
    C_Vertex3f(0.321895123f, 0.0759268999f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.31700182f, -0.00207567215f, 0.f),
    C_Vertex3f(0.339888573f, 0.00610256195f, 0.f),
    C_Vertex3f(0.321895123f, 0.0759268999f, 0.f),
    C_Vertex3f(0.310339928f, 0.0686843395f, 0.f),
    C_Vertex3f(0.298279762f, 0.0628902912f, 0.f),
    C_Vertex3f(0.292859554f, -0.00752782822f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.298279762f, 0.0628902912f, 0.f),
    C_Vertex3f(0.285714626f, 0.058544755f, 0.f),
    C_Vertex3f(0.292859554f, -0.00752782822f, 0.f),
    C_Vertex3f(0.27264452f, 0.0556477308f, 0.f),
    C_Vertex3f(0.267461777f, -0.0102539063f, 0.f),
    C_Vertex3f(0.259069443f, 0.0541992188f, 0.f),
    C_Vertex3f(0.244008064f, 0.0541992188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.267461777f, -0.0102539063f, 0.f),
    C_Vertex3f(0.244008064f, 0.0541992188f, 0.f),
    C_Vertex3f(0.241845131f, -0.0102539063f, 0.f),
    C_Vertex3f(0.228715658f, 0.0559210777f, 0.f),
    C_Vertex3f(0.218459964f, -0.00830185413f, 0.f),
    C_Vertex3f(0.214173555f, 0.0593647957f, 0.f),
    C_Vertex3f(0.200381756f, 0.0645303726f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.218459964f, -0.00830185413f, 0.f),
    C_Vertex3f(0.200381756f, 0.0645303726f, 0.f),
    C_Vertex3f(0.196269631f, -0.0043977499f, 0.f),
    C_Vertex3f(0.18734026f, 0.0714178085f, 0.f),
    C_Vertex3f(0.175274134f, 0.00145840645f, 0.f),
    C_Vertex3f(0.175049067f, 0.0800271034f, 0.f),
    C_Vertex3f(0.163508177f, 0.0903582573f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155473471f, 0.00926661491f, 0.f),
    C_Vertex3f(0.175274134f, 0.00145840645f, 0.f),
    C_Vertex3f(0.163508177f, 0.0903582573f, 0.f),
    C_Vertex3f(0.15271759f, 0.10241127f, 0.f),
    C_Vertex3f(0.14382267f, 0.114601135f, 0.f),
    C_Vertex3f(0.137041807f, 0.126422167f, 0.f),
    C_Vertex3f(0.136867642f, 0.0190268755f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136867642f, 0.0190268755f, 0.f),
    C_Vertex3f(0.137041807f, 0.126422167f, 0.f),
    C_Vertex3f(0.131229639f, 0.139459372f, 0.f),
    C_Vertex3f(0.126386166f, 0.153712749f, 0.f),
    C_Vertex3f(0.122511387f, 0.169182301f, 0.f),
    C_Vertex3f(0.119605303f, 0.185868025f, 0.f),
    C_Vertex3f(0.119456649f, 0.0307391882f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.117667913f, 0.203769922f, 0.f),
    C_Vertex3f(0.116699219f, 0.222887993f, 0.f),
    C_Vertex3f(0.119456649f, 0.0307391882f, 0.f),
    C_Vertex3f(0.119605303f, 0.185868025f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0716257095f, 0.379580498f, 0.f),
    C_Vertex3f(0.0750755072f, 0.0771788359f, 0.f),
    C_Vertex3f(0.0796823502f, 0.392315865f, 0.f),
    C_Vertex3f(0.0882191658f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0885720253f, 0.404129982f, 0.f),
    C_Vertex3f(0.10324049f, 0.044403553f, 0.f),
    C_Vertex3f(0.098294735f, 0.41502285f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0469105244f, 0.137910128f, 0.f),
    C_Vertex3f(0.0544211864f, 0.116123915f, 0.f),
    C_Vertex3f(0.054816246f, 0.344057083f, 0.f),
    C_Vertex3f(0.063809514f, 0.0958801508f, 0.f),
    C_Vertex3f(0.0644021034f, 0.365923882f, 0.f),
    C_Vertex3f(0.0750755072f, 0.0771788359f, 0.f),
    C_Vertex3f(0.0716257095f, 0.379580498f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0356445313f, 0.24495697f, 0.f),
    C_Vertex3f(0.0375221968f, 0.1861099f, 0.f),
    C_Vertex3f(0.038839817f, 0.279929161f, 0.f),
    C_Vertex3f(0.0412775278f, 0.16123879f, 0.f),
    C_Vertex3f(0.0452303886f, 0.312962532f, 0.f),
    C_Vertex3f(0.0469105244f, 0.137910128f, 0.f),
    C_Vertex3f(0.054816246f, 0.344057083f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.811523438f, 0.211425781f, 0.f),
    C_Vertex3f(0.81163311f, 0.227847099f, 0.f),
    C_Vertex3f(0.809966683f, 0.258643985f, 0.f),
    C_Vertex3f(0.806414485f, 0.28739512f, 0.f),
    C_Vertex3f(0.800976515f, 0.314100504f, 0.f),
    C_Vertex3f(0.793652773f, 0.338760138f, 0.f),
    C_Vertex3f(0.784443259f, 0.361374021f, 0.f),
    C_Vertex3f(0.773347974f, 0.381942153f, 0.f),
    C_Vertex3f(0.760366917f, 0.400464535f, 0.f),
    C_Vertex3f(0.745500088f, 0.416941166f, 0.f),
    C_Vertex3f(0.729485393f, 0.431358218f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.805175781f, 0.140136719f, 0.f),
    C_Vertex3f(0.7265625f, 0.151496887f, 0.f),
    C_Vertex3f(0.728620529f, 0.0239825249f, 0.f),
    C_Vertex3f(0.741456985f, 0.0337643623f, 0.f),
    C_Vertex3f(0.75327611f, 0.0445777178f, 0.f),
    C_Vertex3f(0.764077902f, 0.0564225912f, 0.f),
    C_Vertex3f(0.773862362f, 0.0692989826f, 0.f),
    C_Vertex3f(0.78262949f, 0.083206892f, 0.f),
    C_Vertex3f(0.790379286f, 0.0981463194f, 0.f),
    C_Vertex3f(0.79711175f, 0.114117265f, 0.f),
    C_Vertex3f(0.802826881f, 0.131119728f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.72306633f, 0.139339447f, 0.f),
    C_Vertex3f(0.713916779f, 0.118061543f, 0.f),
    C_Vertex3f(0.714944124f, 0.0154234171f, 0.f),
    C_Vertex3f(0.728620529f, 0.0239825249f, 0.f),
    C_Vertex3f(0.7265625f, 0.151496887f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.585483789f, 0.0555199385f, 0.f),
    C_Vertex3f(0.590829849f, -0.0102539063f, 0.f),
    C_Vertex3f(0.599514961f, 0.0541992188f, 0.f),
    C_Vertex3f(0.615246773f, -0.0102539063f, 0.f),
    C_Vertex3f(0.617649078f, 0.0541992188f, 0.f),
    C_Vertex3f(0.63396287f, -0.00903117657f, 0.f),
    C_Vertex3f(0.637776375f, 0.057240963f, 0.f),
    C_Vertex3f(0.651839018f, -0.0065857172f, 0.f),
    C_Vertex3f(0.656401634f, 0.0633244514f, 0.f),
    C_Vertex3f(0.668875217f, -0.00291752815f, 0.f),
    C_Vertex3f(0.673524857f, 0.0724496841f, 0.f),
    C_Vertex3f(0.685071468f, 0.00197339058f, 0.f),
    C_Vertex3f(0.689146042f, 0.0846166611f, 0.f),
    C_Vertex3f(0.700427771f, 0.00808703899f, 0.f),
    C_Vertex3f(0.702610016f, 0.0998206139f, 0.f),
    C_Vertex3f(0.714944124f, 0.0154234171f, 0.f),
    C_Vertex3f(0.713916779f, 0.118061543f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.572060347f, 0.0581613779f, 0.f),
    C_Vertex3f(0.559244633f, 0.0621235371f, 0.f),
    C_Vertex3f(0.563081264f, -0.00739192963f, 0.f),
    C_Vertex3f(0.590829849f, -0.0102539063f, 0.f),
    C_Vertex3f(0.585483789f, 0.0555199385f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.547036648f, 0.0674064159f, 0.f),
    C_Vertex3f(0.535436392f, 0.0740100145f, 0.f),
    C_Vertex3f(0.536861897f, -0.00166797638f, 0.f),
    C_Vertex3f(0.563081264f, -0.00739192963f, 0.f),
    C_Vertex3f(0.559244633f, 0.0621235371f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.535436392f, 0.0740100145f, 0.f),
    C_Vertex3f(0.524443865f, 0.0819343328f, 0.f),
    C_Vertex3f(0.536861897f, -0.00166797638f, 0.f),
    C_Vertex3f(0.514059067f, 0.0911793709f, 0.f),
    C_Vertex3f(0.512171745f, 0.00691795349f, 0.f),
    C_Vertex3f(0.504281998f, 0.101745129f, 0.f),
    C_Vertex3f(0.495686054f, 0.113262057f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.489010811f, 0.01836586f, 0.f),
    C_Vertex3f(0.512171745f, 0.00691795349f, 0.f),
    C_Vertex3f(0.495686054f, 0.113262057f, 0.f),
    C_Vertex3f(0.488271236f, 0.125730157f, 0.f),
    C_Vertex3f(0.482037544f, 0.139149427f, 0.f),
    C_Vertex3f(0.476984978f, 0.153519869f, 0.f),
    C_Vertex3f(0.473113537f, 0.168841481f, 0.f),
    C_Vertex3f(0.470423222f, 0.185114264f, 0.f),
    C_Vertex3f(0.468253613f, 0.0320768356f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.468914032f, 0.202338219f, 0.f),
    C_Vertex3f(0.46875f, 0.211425781f, 0.f),
    C_Vertex3f(0.468253613f, 0.0320768356f, 0.f),
    C_Vertex3f(0.470423222f, 0.185114264f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0375221968f, 0.1861099f, 0.f),
    C_Vertex3f(0.0356445313f, 0.24495697f, 0.f),
    C_Vertex3f(0.0356445313f, 0.21252346f, 0.f),
    C_Vertex3f(0.468143463f, 0.430574417f, 0.f),
    C_Vertex3f(0.471679688f, 0.275878906f, 0.f),
    C_Vertex3f(0.472281456f, 0.283697128f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.725624204f, 0.299648881f, -1.f),
    C_Vertex3f(0.727859497f, 0.284134865f, -1.f),
    C_Vertex3f(0.729485393f, 0.431358218f, -1.f),
    C_Vertex3f(0.728515625f, 0.275878906f, -1.f),
    C_Vertex3f(0.811523438f, 0.211425781f, -1.f),
    C_Vertex3f(0.471679688f, 0.275878906f, -1.f),
    C_Vertex3f(0.46875f, 0.211425781f, -1.f),
    C_Vertex3f(0.468143463f, 0.430574417f, -1.f),
    C_Vertex3f(0.468253613f, 0.0320768356f, -1.f),
    C_Vertex3f(0.45016098f, 0.414442062f, -1.f),
    C_Vertex3f(0.44990015f, 0.0480508804f, -1.f),
    C_Vertex3f(0.434173584f, 0.396238327f, -1.f),
    C_Vertex3f(0.433950424f, 0.0662879944f, -1.f),
    C_Vertex3f(0.427177429f, 0.386100769f, -1.f),
    C_Vertex3f(0.427177429f, 0.0765380859f, -1.f),
    C_Vertex3f(0.421606064f, 0.395692825f, -1.f),
    C_Vertex3f(0.418874741f, 0.0657444f, -1.f),
    C_Vertex3f(0.407649994f, 0.413214684f, -1.f),
    C_Vertex3f(0.401012897f, 0.0468277931f, -1.f),
    C_Vertex3f(0.390880585f, 0.429074287f, -1.f),
    C_Vertex3f(0.388671875f, 0.244748116f, -1.f),
    C_Vertex3f(0.387437224f, 0.265089512f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.390880585f, 0.429074287f, -1.f),
    C_Vertex3f(0.371297836f, 0.443271637f, -1.f),
    C_Vertex3f(0.376325369f, 0.318283081f, -1.f),
    C_Vertex3f(0.381263971f, 0.301856995f, -1.f),
    C_Vertex3f(0.384967923f, 0.284125805f, -1.f),
    C_Vertex3f(0.387437224f, 0.265089512f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.371297836f, 0.443271637f, -1.f),
    C_Vertex3f(0.348901749f, 0.455806732f, -1.f),
    C_Vertex3f(0.354101658f, 0.359730721f, -1.f),
    C_Vertex3f(0.362744212f, 0.347219944f, -1.f),
    C_Vertex3f(0.370152116f, 0.333404064f, -1.f),
    C_Vertex3f(0.376325369f, 0.318283081f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.311196089f, 0.396149158f, -1.f),
    C_Vertex3f(0.322771668f, 0.389145613f, -1.f),
    C_Vertex3f(0.324566841f, 0.465208054f, -1.f),
    C_Vertex3f(0.333781123f, 0.380741358f, -1.f),
    C_Vertex3f(0.348901749f, 0.455806732f, -1.f),
    C_Vertex3f(0.344224453f, 0.370936394f, -1.f),
    C_Vertex3f(0.354101658f, 0.359730721f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.2730726f, 0.408755541f, -1.f),
    C_Vertex3f(0.286346555f, 0.405954123f, -1.f),
    C_Vertex3f(0.298293114f, 0.471475601f, -1.f),
    C_Vertex3f(0.299054384f, 0.401751995f, -1.f),
    C_Vertex3f(0.324566841f, 0.465208054f, -1.f),
    C_Vertex3f(0.311196089f, 0.396149158f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.270080566f, 0.474609375f, -1.f),
    C_Vertex3f(0.239055634f, 0.474609375f, -1.f),
    C_Vertex3f(0.242534637f, 0.41015625f, -1.f),
    C_Vertex3f(0.259232521f, 0.41015625f, -1.f),
    C_Vertex3f(0.2730726f, 0.408755541f, -1.f),
    C_Vertex3f(0.298293114f, 0.471475601f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.190149307f, 0.394085884f, -1.f),
    C_Vertex3f(0.206847191f, 0.402121067f, -1.f),
    C_Vertex3f(0.208604336f, 0.47100544f, -1.f),
    C_Vertex3f(0.224308968f, 0.407477856f, -1.f),
    C_Vertex3f(0.239055634f, 0.474609375f, -1.f),
    C_Vertex3f(0.242534637f, 0.41015625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179600239f, 0.463797569f, -1.f),
    C_Vertex3f(0.152043343f, 0.452985764f, -1.f),
    C_Vertex3f(0.160163879f, 0.370089531f, -1.f),
    C_Vertex3f(0.174215317f, 0.383372307f, -1.f),
    C_Vertex3f(0.190149307f, 0.394085884f, -1.f),
    C_Vertex3f(0.208604336f, 0.47100544f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.123002052f, 0.292797565f, -1.f),
    C_Vertex3f(0.129304886f, 0.314826012f, -1.f),
    C_Vertex3f(0.132461071f, 0.442173958f, -1.f),
    C_Vertex3f(0.137708664f, 0.335816383f, -1.f),
    C_Vertex3f(0.152043343f, 0.452985764f, -1.f),
    C_Vertex3f(0.147994995f, 0.354237556f, -1.f),
    C_Vertex3f(0.160163879f, 0.370089531f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.120239258f, 0.434044838f, -1.f),
    C_Vertex3f(0.108850479f, 0.424994469f, -1.f),
    C_Vertex3f(0.116699219f, 0.24562645f, -1.f),
    C_Vertex3f(0.118800163f, 0.269731045f, -1.f),
    C_Vertex3f(0.123002052f, 0.292797565f, -1.f),
    C_Vertex3f(0.132461071f, 0.442173958f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.108850479f, 0.424994469f, -1.f),
    C_Vertex3f(0.098294735f, 0.41502285f, -1.f),
    C_Vertex3f(0.10324049f, 0.044403553f, -1.f),
    C_Vertex3f(0.119456649f, 0.0307391882f, -1.f),
    C_Vertex3f(0.116699219f, 0.222887993f, -1.f),
    C_Vertex3f(0.116699219f, 0.24562645f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.729485393f, 0.431358218f, -1.f),
    C_Vertex3f(0.712322831f, 0.443715692f, -1.f),
    C_Vertex3f(0.713380098f, 0.340203524f, -1.f),
    C_Vertex3f(0.718384504f, 0.32768321f, -1.f),
    C_Vertex3f(0.722465873f, 0.314164996f, -1.f),
    C_Vertex3f(0.725624204f, 0.299648881f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.684132099f, 0.380305767f, -1.f),
    C_Vertex3f(0.692828655f, 0.371777058f, -1.f),
    C_Vertex3f(0.694012403f, 0.454013586f, -1.f),
    C_Vertex3f(0.700602174f, 0.362250447f, -1.f),
    C_Vertex3f(0.712322831f, 0.443715692f, -1.f),
    C_Vertex3f(0.707452655f, 0.351725936f, -1.f),
    C_Vertex3f(0.713380098f, 0.340203524f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.654884577f, 0.399495363f, -1.f),
    C_Vertex3f(0.66516006f, 0.39416492f, -1.f),
    C_Vertex3f(0.67455411f, 0.462251902f, -1.f),
    C_Vertex3f(0.674909234f, 0.387768388f, -1.f),
    C_Vertex3f(0.694012403f, 0.454013586f, -1.f),
    C_Vertex3f(0.684132099f, 0.380305767f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.653947949f, 0.468430638f, -1.f),
    C_Vertex3f(0.632193923f, 0.472549796f, -1.f),
    C_Vertex3f(0.632754683f, 0.406957984f, -1.f),
    C_Vertex3f(0.644082785f, 0.403759718f, -1.f),
    C_Vertex3f(0.654884577f, 0.399495363f, -1.f),
    C_Vertex3f(0.67455411f, 0.462251902f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.595502853f, 0.41015625f, -1.f),
    C_Vertex3f(0.608519554f, 0.41015625f, -1.f),
    C_Vertex3f(0.60929203f, 0.474609375f, -1.f),
    C_Vertex3f(0.620900273f, 0.409090161f, -1.f),
    C_Vertex3f(0.632193923f, 0.472549796f, -1.f),
    C_Vertex3f(0.632754683f, 0.406957984f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.583713531f, 0.474609375f, -1.f),
    C_Vertex3f(0.557889938f, 0.471611977f, -1.f),
    C_Vertex3f(0.559118509f, 0.403431416f, -1.f),
    C_Vertex3f(0.570747495f, 0.406793833f, -1.f),
    C_Vertex3f(0.582875609f, 0.409035444f, -1.f),
    C_Vertex3f(0.595502853f, 0.41015625f, -1.f),
    C_Vertex3f(0.60929203f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.517593861f, 0.378773689f, -1.f),
    C_Vertex3f(0.527226329f, 0.386619329f, -1.f),
    C_Vertex3f(0.533349991f, 0.46561718f, -1.f),
    C_Vertex3f(0.537357926f, 0.393344164f, -1.f),
    C_Vertex3f(0.557889938f, 0.471611977f, -1.f),
    C_Vertex3f(0.547988653f, 0.398948193f, -1.f),
    C_Vertex3f(0.559118509f, 0.403431416f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.510093689f, 0.456624985f, -1.f),
    C_Vertex3f(0.488121033f, 0.444635391f, -1.f),
    C_Vertex3f(0.49316752f, 0.349373102f, -1.f),
    C_Vertex3f(0.500318408f, 0.360007048f, -1.f),
    C_Vertex3f(0.508460522f, 0.369807243f, -1.f),
    C_Vertex3f(0.517593861f, 0.378773689f, -1.f),
    C_Vertex3f(0.533349991f, 0.46561718f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.488121033f, 0.444635391f, -1.f),
    C_Vertex3f(0.468143463f, 0.430574417f, -1.f),
    C_Vertex3f(0.472281456f, 0.283697128f, -1.f),
    C_Vertex3f(0.474476218f, 0.298499823f, -1.f),
    C_Vertex3f(0.477662206f, 0.312468767f, -1.f),
    C_Vertex3f(0.481839418f, 0.325603962f, -1.f),
    C_Vertex3f(0.487007856f, 0.337905407f, -1.f),
    C_Vertex3f(0.49316752f, 0.349373102f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.401012897f, 0.0468277931f, -1.f),
    C_Vertex3f(0.388671875f, 0.244748116f, -1.f),
    C_Vertex3f(0.388671875f, 0.223047256f, -1.f),
    C_Vertex3f(0.38741684f, 0.202562928f, -1.f),
    C_Vertex3f(0.384906769f, 0.18334949f, -1.f),
    C_Vertex3f(0.381894588f, 0.0305819511f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.381894588f, 0.0305819511f, -1.f),
    C_Vertex3f(0.384906769f, 0.18334949f, -1.f),
    C_Vertex3f(0.381141663f, 0.165406942f, -1.f),
    C_Vertex3f(0.376121521f, 0.148735285f, -1.f),
    C_Vertex3f(0.369846344f, 0.133334517f, -1.f),
    C_Vertex3f(0.362316132f, 0.11920464f, -1.f),
    C_Vertex3f(0.361519814f, 0.0170068741f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362316132f, 0.11920464f, -1.f),
    C_Vertex3f(0.353530884f, 0.106345654f, -1.f),
    C_Vertex3f(0.361519814f, 0.0170068741f, -1.f),
    C_Vertex3f(0.343490601f, 0.0947575569f, -1.f),
    C_Vertex3f(0.339888573f, 0.00610256195f, -1.f),
    C_Vertex3f(0.332945347f, 0.0846179724f, -1.f),
    C_Vertex3f(0.321895123f, 0.0759268999f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.31700182f, -0.00207567215f, -1.f),
    C_Vertex3f(0.339888573f, 0.00610256195f, -1.f),
    C_Vertex3f(0.321895123f, 0.0759268999f, -1.f),
    C_Vertex3f(0.310339928f, 0.0686843395f, -1.f),
    C_Vertex3f(0.298279762f, 0.0628902912f, -1.f),
    C_Vertex3f(0.292859554f, -0.00752782822f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.298279762f, 0.0628902912f, -1.f),
    C_Vertex3f(0.285714626f, 0.058544755f, -1.f),
    C_Vertex3f(0.292859554f, -0.00752782822f, -1.f),
    C_Vertex3f(0.27264452f, 0.0556477308f, -1.f),
    C_Vertex3f(0.267461777f, -0.0102539063f, -1.f),
    C_Vertex3f(0.259069443f, 0.0541992188f, -1.f),
    C_Vertex3f(0.244008064f, 0.0541992188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.267461777f, -0.0102539063f, -1.f),
    C_Vertex3f(0.244008064f, 0.0541992188f, -1.f),
    C_Vertex3f(0.241845131f, -0.0102539063f, -1.f),
    C_Vertex3f(0.228715658f, 0.0559210777f, -1.f),
    C_Vertex3f(0.218459964f, -0.00830185413f, -1.f),
    C_Vertex3f(0.214173555f, 0.0593647957f, -1.f),
    C_Vertex3f(0.200381756f, 0.0645303726f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.218459964f, -0.00830185413f, -1.f),
    C_Vertex3f(0.200381756f, 0.0645303726f, -1.f),
    C_Vertex3f(0.196269631f, -0.0043977499f, -1.f),
    C_Vertex3f(0.18734026f, 0.0714178085f, -1.f),
    C_Vertex3f(0.175274134f, 0.00145840645f, -1.f),
    C_Vertex3f(0.175049067f, 0.0800271034f, -1.f),
    C_Vertex3f(0.163508177f, 0.0903582573f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155473471f, 0.00926661491f, -1.f),
    C_Vertex3f(0.175274134f, 0.00145840645f, -1.f),
    C_Vertex3f(0.163508177f, 0.0903582573f, -1.f),
    C_Vertex3f(0.15271759f, 0.10241127f, -1.f),
    C_Vertex3f(0.14382267f, 0.114601135f, -1.f),
    C_Vertex3f(0.137041807f, 0.126422167f, -1.f),
    C_Vertex3f(0.136867642f, 0.0190268755f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136867642f, 0.0190268755f, -1.f),
    C_Vertex3f(0.137041807f, 0.126422167f, -1.f),
    C_Vertex3f(0.131229639f, 0.139459372f, -1.f),
    C_Vertex3f(0.126386166f, 0.153712749f, -1.f),
    C_Vertex3f(0.122511387f, 0.169182301f, -1.f),
    C_Vertex3f(0.119605303f, 0.185868025f, -1.f),
    C_Vertex3f(0.119456649f, 0.0307391882f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.117667913f, 0.203769922f, -1.f),
    C_Vertex3f(0.116699219f, 0.222887993f, -1.f),
    C_Vertex3f(0.119456649f, 0.0307391882f, -1.f),
    C_Vertex3f(0.119605303f, 0.185868025f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0716257095f, 0.379580498f, -1.f),
    C_Vertex3f(0.0750755072f, 0.0771788359f, -1.f),
    C_Vertex3f(0.0796823502f, 0.392315865f, -1.f),
    C_Vertex3f(0.0882191658f, 0.0600199699f, -1.f),
    C_Vertex3f(0.0885720253f, 0.404129982f, -1.f),
    C_Vertex3f(0.10324049f, 0.044403553f, -1.f),
    C_Vertex3f(0.098294735f, 0.41502285f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0469105244f, 0.137910128f, -1.f),
    C_Vertex3f(0.0544211864f, 0.116123915f, -1.f),
    C_Vertex3f(0.054816246f, 0.344057083f, -1.f),
    C_Vertex3f(0.063809514f, 0.0958801508f, -1.f),
    C_Vertex3f(0.0644021034f, 0.365923882f, -1.f),
    C_Vertex3f(0.0750755072f, 0.0771788359f, -1.f),
    C_Vertex3f(0.0716257095f, 0.379580498f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0356445313f, 0.24495697f, -1.f),
    C_Vertex3f(0.0375221968f, 0.1861099f, -1.f),
    C_Vertex3f(0.038839817f, 0.279929161f, -1.f),
    C_Vertex3f(0.0412775278f, 0.16123879f, -1.f),
    C_Vertex3f(0.0452303886f, 0.312962532f, -1.f),
    C_Vertex3f(0.0469105244f, 0.137910128f, -1.f),
    C_Vertex3f(0.054816246f, 0.344057083f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.811523438f, 0.211425781f, -1.f),
    C_Vertex3f(0.81163311f, 0.227847099f, -1.f),
    C_Vertex3f(0.809966683f, 0.258643985f, -1.f),
    C_Vertex3f(0.806414485f, 0.28739512f, -1.f),
    C_Vertex3f(0.800976515f, 0.314100504f, -1.f),
    C_Vertex3f(0.793652773f, 0.338760138f, -1.f),
    C_Vertex3f(0.784443259f, 0.361374021f, -1.f),
    C_Vertex3f(0.773347974f, 0.381942153f, -1.f),
    C_Vertex3f(0.760366917f, 0.400464535f, -1.f),
    C_Vertex3f(0.745500088f, 0.416941166f, -1.f),
    C_Vertex3f(0.729485393f, 0.431358218f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.805175781f, 0.140136719f, -1.f),
    C_Vertex3f(0.7265625f, 0.151496887f, -1.f),
    C_Vertex3f(0.728620529f, 0.0239825249f, -1.f),
    C_Vertex3f(0.741456985f, 0.0337643623f, -1.f),
    C_Vertex3f(0.75327611f, 0.0445777178f, -1.f),
    C_Vertex3f(0.764077902f, 0.0564225912f, -1.f),
    C_Vertex3f(0.773862362f, 0.0692989826f, -1.f),
    C_Vertex3f(0.78262949f, 0.083206892f, -1.f),
    C_Vertex3f(0.790379286f, 0.0981463194f, -1.f),
    C_Vertex3f(0.79711175f, 0.114117265f, -1.f),
    C_Vertex3f(0.802826881f, 0.131119728f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.72306633f, 0.139339447f, -1.f),
    C_Vertex3f(0.713916779f, 0.118061543f, -1.f),
    C_Vertex3f(0.714944124f, 0.0154234171f, -1.f),
    C_Vertex3f(0.728620529f, 0.0239825249f, -1.f),
    C_Vertex3f(0.7265625f, 0.151496887f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.585483789f, 0.0555199385f, -1.f),
    C_Vertex3f(0.590829849f, -0.0102539063f, -1.f),
    C_Vertex3f(0.599514961f, 0.0541992188f, -1.f),
    C_Vertex3f(0.615246773f, -0.0102539063f, -1.f),
    C_Vertex3f(0.617649078f, 0.0541992188f, -1.f),
    C_Vertex3f(0.63396287f, -0.00903117657f, -1.f),
    C_Vertex3f(0.637776375f, 0.057240963f, -1.f),
    C_Vertex3f(0.651839018f, -0.0065857172f, -1.f),
    C_Vertex3f(0.656401634f, 0.0633244514f, -1.f),
    C_Vertex3f(0.668875217f, -0.00291752815f, -1.f),
    C_Vertex3f(0.673524857f, 0.0724496841f, -1.f),
    C_Vertex3f(0.685071468f, 0.00197339058f, -1.f),
    C_Vertex3f(0.689146042f, 0.0846166611f, -1.f),
    C_Vertex3f(0.700427771f, 0.00808703899f, -1.f),
    C_Vertex3f(0.702610016f, 0.0998206139f, -1.f),
    C_Vertex3f(0.714944124f, 0.0154234171f, -1.f),
    C_Vertex3f(0.713916779f, 0.118061543f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.572060347f, 0.0581613779f, -1.f),
    C_Vertex3f(0.559244633f, 0.0621235371f, -1.f),
    C_Vertex3f(0.563081264f, -0.00739192963f, -1.f),
    C_Vertex3f(0.590829849f, -0.0102539063f, -1.f),
    C_Vertex3f(0.585483789f, 0.0555199385f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.547036648f, 0.0674064159f, -1.f),
    C_Vertex3f(0.535436392f, 0.0740100145f, -1.f),
    C_Vertex3f(0.536861897f, -0.00166797638f, -1.f),
    C_Vertex3f(0.563081264f, -0.00739192963f, -1.f),
    C_Vertex3f(0.559244633f, 0.0621235371f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.535436392f, 0.0740100145f, -1.f),
    C_Vertex3f(0.524443865f, 0.0819343328f, -1.f),
    C_Vertex3f(0.536861897f, -0.00166797638f, -1.f),
    C_Vertex3f(0.514059067f, 0.0911793709f, -1.f),
    C_Vertex3f(0.512171745f, 0.00691795349f, -1.f),
    C_Vertex3f(0.504281998f, 0.101745129f, -1.f),
    C_Vertex3f(0.495686054f, 0.113262057f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.489010811f, 0.01836586f, -1.f),
    C_Vertex3f(0.512171745f, 0.00691795349f, -1.f),
    C_Vertex3f(0.495686054f, 0.113262057f, -1.f),
    C_Vertex3f(0.488271236f, 0.125730157f, -1.f),
    C_Vertex3f(0.482037544f, 0.139149427f, -1.f),
    C_Vertex3f(0.476984978f, 0.153519869f, -1.f),
    C_Vertex3f(0.473113537f, 0.168841481f, -1.f),
    C_Vertex3f(0.470423222f, 0.185114264f, -1.f),
    C_Vertex3f(0.468253613f, 0.0320768356f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.468914032f, 0.202338219f, -1.f),
    C_Vertex3f(0.46875f, 0.211425781f, -1.f),
    C_Vertex3f(0.468253613f, 0.0320768356f, -1.f),
    C_Vertex3f(0.470423222f, 0.185114264f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0375221968f, 0.1861099f, -1.f),
    C_Vertex3f(0.0356445313f, 0.24495697f, -1.f),
    C_Vertex3f(0.0356445313f, 0.21252346f, -1.f),
    C_Vertex3f(0.468143463f, 0.430574417f, -1.f),
    C_Vertex3f(0.471679688f, 0.275878906f, -1.f),
    C_Vertex3f(0.472281456f, 0.283697128f, -1.f),
    C_End(),
    C_Translatef(0.844726563f, 0.f, 0.f),
};
static const Call char_157[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.559570313f, 0.f),
    C_Vertex3f(0.125976563f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.559570313f, 0.f),
    C_Vertex3f(0.545410156f, 0.545410156f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, 0.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, 0.f),
    C_Vertex3f(0.111816406f, 0.f, 0.f),
    C_Vertex3f(0.559570313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.559570313f, -1.f),
    C_Vertex3f(0.125976563f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.559570313f, -1.f),
    C_Vertex3f(0.545410156f, 0.545410156f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_Vertex3f(0.545410156f, 0.0141601563f, -1.f),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.125976563f, 0.0141601563f, -1.f),
    C_Vertex3f(0.111816406f, 0.f, -1.f),
    C_Vertex3f(0.559570313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.671386719f, 0.f, 0.f),
};
static const Call char_158[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0174789429f, 0.f, 0.f),
    C_Vertex3f(0.0174789429f, 0.f, -1.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, -1.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, 0.f),
    C_Normal3f(-0.754504204f, 0.65629518f, 0.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, 0.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, -1.f),
    C_Vertex3f(0.312271118f, 0.402832031f, -1.f),
    C_Vertex3f(0.312271118f, 0.402832031f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.0486485697f, -0.998815954f, 0.f),
    C_Vertex3f(0.312271118f, 0.402832031f, 0.f),
    C_Vertex3f(0.312271118f, 0.402832031f, -1.f),
    C_Normal3f(0.0422043577f, -0.99910903f, 0.f),
    C_Vertex3f(0.299739838f, 0.40222168f, 0.f),
    C_Vertex3f(0.299739838f, 0.40222168f, -1.f),
    C_Normal3f(0.0333627351f, -0.999443233f, 0.f),
    C_Vertex3f(0.276148796f, 0.401306152f, 0.f),
    C_Vertex3f(0.276148796f, 0.401306152f, -1.f),
    C_Normal3f(0.0214021672f, -0.999770999f, 0.f),
    C_Vertex3f(0.254029274f, 0.400695801f, 0.f),
    C_Vertex3f(0.254029274f, 0.400695801f, -1.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.233381271f, 0.400390625f, 0.f),
    C_Vertex3f(0.233381271f, 0.400390625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.233381271f, 0.400390625f, 0.f),
    C_Vertex3f(0.233381271f, 0.400390625f, -1.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, -1.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, 0.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, -1.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, -1.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, 0.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, -1.f),
    C_Vertex3f(0.413963318f, 0.464355469f, -1.f),
    C_Vertex3f(0.413963318f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.413963318f, 0.464355469f, 0.f),
    C_Vertex3f(0.413963318f, 0.464355469f, -1.f),
    C_Vertex3f(0.413963318f, 0.412063599f, -1.f),
    C_Vertex3f(0.413963318f, 0.412063599f, 0.f),
    C_Normal3f(0.760531485f, -0.649300992f, 0.f),
    C_Vertex3f(0.413963318f, 0.412063599f, 0.f),
    C_Vertex3f(0.413963318f, 0.412063599f, -1.f),
    C_Vertex3f(0.162734985f, 0.117797852f, -1.f),
    C_Vertex3f(0.162734985f, 0.117797852f, 0.f),
    C_Normal3f(0.744730413f, -0.667365432f, 0.f),
    C_Vertex3f(0.162734985f, 0.117797852f, 0.f),
    C_Vertex3f(0.162734985f, 0.117797852f, -1.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, -1.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0737020075f, 0.9972803f, 0.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, 0.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, -1.f),
    C_Normal3f(-0.0513249896f, 0.998682022f, 0.f),
    C_Vertex3f(0.140922546f, 0.0659179688f, 0.f),
    C_Vertex3f(0.140922546f, 0.0659179688f, -1.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, 0.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, 0.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, -1.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, -1.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, 0.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, -1.f),
    C_Vertex3f(0.428123474f, 0.f, -1.f),
    C_Vertex3f(0.428123474f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.428123474f, 0.f, 0.f),
    C_Vertex3f(0.428123474f, 0.f, -1.f),
    C_Vertex3f(0.0174789429f, 0.f, -1.f),
    C_Vertex3f(0.0174789429f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.840821147f, 0.541312993f, 0.f),
    C_Vertex3f(0.235839844f, 0.569824219f, 0.f),
    C_Vertex3f(0.235839844f, 0.569824219f, -1.f),
    C_Vertex3f(0.283935547f, 0.64453125f, -1.f),
    C_Vertex3f(0.283935547f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.283935547f, 0.64453125f, 0.f),
    C_Vertex3f(0.283935547f, 0.64453125f, -1.f),
    C_Vertex3f(0.373291016f, 0.64453125f, -1.f),
    C_Vertex3f(0.373291016f, 0.64453125f, 0.f),
    C_Normal3f(0.779806852f, -0.626020253f, 0.f),
    C_Vertex3f(0.373291016f, 0.64453125f, 0.f),
    C_Vertex3f(0.373291016f, 0.64453125f, -1.f),
    C_Vertex3f(0.274902344f, 0.521972656f, -1.f),
    C_Vertex3f(0.274902344f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.274902344f, 0.521972656f, 0.f),
    C_Vertex3f(0.274902344f, 0.521972656f, -1.f),
    C_Vertex3f(0.190917969f, 0.521972656f, -1.f),
    C_Vertex3f(0.190917969f, 0.521972656f, 0.f),
    C_Normal3f(-0.795031667f, -0.606567979f, 0.f),
    C_Vertex3f(0.190917969f, 0.521972656f, 0.f),
    C_Vertex3f(0.190917969f, 0.521972656f, -1.f),
    C_Vertex3f(0.0974121094f, 0.64453125f, -1.f),
    C_Vertex3f(0.0974121094f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0974121094f, 0.64453125f, 0.f),
    C_Vertex3f(0.0974121094f, 0.64453125f, -1.f),
    C_Vertex3f(0.186767578f, 0.64453125f, -1.f),
    C_Vertex3f(0.186767578f, 0.64453125f, 0.f),
    C_Normal3f(0.835812926f, 0.54901439f, 0.f),
    C_Vertex3f(0.186767578f, 0.64453125f, 0.f),
    C_Vertex3f(0.186767578f, 0.64453125f, -1.f),
    C_Vertex3f(0.235839844f, 0.569824219f, -1.f),
    C_Vertex3f(0.235839844f, 0.569824219f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.428123474f, 0.f, 0.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, 0.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, 0.f),
    C_Vertex3f(0.140922546f, 0.0659179688f, 0.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, 0.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, 0.f),
    C_Vertex3f(0.0174789429f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.413963318f, 0.464355469f, 0.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, 0.f),
    C_Vertex3f(0.233381271f, 0.400390625f, 0.f),
    C_Vertex3f(0.254029274f, 0.400695801f, 0.f),
    C_Vertex3f(0.276148796f, 0.401306152f, 0.f),
    C_Vertex3f(0.299739838f, 0.40222168f, 0.f),
    C_Vertex3f(0.312271118f, 0.402832031f, 0.f),
    C_Vertex3f(0.413963318f, 0.412063599f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.312271118f, 0.402832031f, 0.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, 0.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, 0.f),
    C_Vertex3f(0.162734985f, 0.117797852f, 0.f),
    C_Vertex3f(0.413963318f, 0.412063599f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.373291016f, 0.64453125f, 0.f),
    C_Vertex3f(0.283935547f, 0.64453125f, 0.f),
    C_Vertex3f(0.274902344f, 0.521972656f, 0.f),
    C_Vertex3f(0.235839844f, 0.569824219f, 0.f),
    C_Vertex3f(0.190917969f, 0.521972656f, 0.f),
    C_Vertex3f(0.186767578f, 0.64453125f, 0.f),
    C_Vertex3f(0.0974121094f, 0.64453125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.233381271f, 0.400390625f, 0.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, 0.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.428123474f, 0.f, -1.f),
    C_Vertex3f(0.428123474f, 0.0678710938f, -1.f),
    C_Vertex3f(0.190502167f, 0.0678710938f, -1.f),
    C_Vertex3f(0.140922546f, 0.0659179688f, -1.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, -1.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, -1.f),
    C_Vertex3f(0.0174789429f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.413963318f, 0.464355469f, -1.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, -1.f),
    C_Vertex3f(0.233381271f, 0.400390625f, -1.f),
    C_Vertex3f(0.254029274f, 0.400695801f, -1.f),
    C_Vertex3f(0.276148796f, 0.401306152f, -1.f),
    C_Vertex3f(0.299739838f, 0.40222168f, -1.f),
    C_Vertex3f(0.312271118f, 0.402832031f, -1.f),
    C_Vertex3f(0.413963318f, 0.412063599f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.312271118f, 0.402832031f, -1.f),
    C_Vertex3f(0.0174789429f, 0.0639266968f, -1.f),
    C_Vertex3f(0.114494324f, 0.0639648438f, -1.f),
    C_Vertex3f(0.162734985f, 0.117797852f, -1.f),
    C_Vertex3f(0.413963318f, 0.412063599f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.373291016f, 0.64453125f, -1.f),
    C_Vertex3f(0.283935547f, 0.64453125f, -1.f),
    C_Vertex3f(0.274902344f, 0.521972656f, -1.f),
    C_Vertex3f(0.235839844f, 0.569824219f, -1.f),
    C_Vertex3f(0.190917969f, 0.521972656f, -1.f),
    C_Vertex3f(0.186767578f, 0.64453125f, -1.f),
    C_Vertex3f(0.0974121094f, 0.64453125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.233381271f, 0.400390625f, -1.f),
    C_Vertex3f(0.0350570679f, 0.464355469f, -1.f),
    C_Vertex3f(0.0350570679f, 0.400390625f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_159[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.249542236f, 0.f, 0.f),
    C_Vertex3f(0.249542236f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Normal3f(-0.831113875f, -0.556102216f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.00247192383f, 0.640625f, -1.f),
    C_Vertex3f(0.00247192383f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.00247192383f, 0.640625f, 0.f),
    C_Vertex3f(0.00247192383f, 0.640625f, -1.f),
    C_Vertex3f(0.105628967f, 0.640625f, -1.f),
    C_Vertex3f(0.105628967f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.836778283f, 0.547541976f, 0.f),
    C_Vertex3f(0.105628967f, 0.640625f, 0.f),
    C_Vertex3f(0.105628967f, 0.640625f, -1.f),
    C_Normal3f(0.840595007f, 0.541664064f, 0.f),
    C_Vertex3f(0.249668121f, 0.420497894f, 0.f),
    C_Vertex3f(0.249668121f, 0.420497894f, -1.f),
    C_Normal3f(0.862715006f, 0.505690515f, 0.f),
    C_Vertex3f(0.282409668f, 0.366283417f, 0.f),
    C_Vertex3f(0.282409668f, 0.366283417f, -1.f),
    C_Normal3f(0.875902712f, 0.482487768f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.86717093f, 0.498010665f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_Normal3f(-0.86124754f, 0.508185625f, 0.f),
    C_Vertex3f(0.304641724f, 0.351566315f, 0.f),
    C_Vertex3f(0.304641724f, 0.351566315f, -1.f),
    C_Normal3f(-0.854441464f, 0.519547701f, 0.f),
    C_Vertex3f(0.320591927f, 0.37826252f, 0.f),
    C_Vertex3f(0.320591927f, 0.37826252f, -1.f),
    C_Normal3f(-0.847048819f, 0.531515121f, 0.f),
    C_Vertex3f(0.338072777f, 0.406546593f, 0.f),
    C_Vertex3f(0.338072777f, 0.406546593f, -1.f),
    C_Normal3f(-0.835170507f, 0.549991071f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.835170507f, 0.549991071f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, -1.f),
    C_Vertex3f(0.49156189f, 0.640625f, -1.f),
    C_Vertex3f(0.49156189f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.49156189f, 0.640625f, 0.f),
    C_Vertex3f(0.49156189f, 0.640625f, -1.f),
    C_Vertex3f(0.590362549f, 0.640625f, -1.f),
    C_Vertex3f(0.590362549f, 0.640625f, 0.f),
    C_Normal3f(0.821961641f, -0.569542885f, 0.f),
    C_Vertex3f(0.590362549f, 0.640625f, 0.f),
    C_Vertex3f(0.590362549f, 0.640625f, -1.f),
    C_Vertex3f(0.334503174f, 0.271369934f, -1.f),
    C_Vertex3f(0.334503174f, 0.271369934f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.334503174f, 0.271369934f, 0.f),
    C_Vertex3f(0.334503174f, 0.271369934f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.171386719f, 0.680175781f, 0.f),
    C_Vertex3f(0.171386719f, 0.680175781f, -1.f),
    C_Vertex3f(0.171386719f, 0.76953125f, -1.f),
    C_Vertex3f(0.171386719f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.171386719f, 0.76953125f, 0.f),
    C_Vertex3f(0.171386719f, 0.76953125f, -1.f),
    C_Vertex3f(0.253417969f, 0.76953125f, -1.f),
    C_Vertex3f(0.253417969f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.253417969f, 0.76953125f, 0.f),
    C_Vertex3f(0.253417969f, 0.76953125f, -1.f),
    C_Vertex3f(0.253417969f, 0.680175781f, -1.f),
    C_Vertex3f(0.253417969f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.253417969f, 0.680175781f, 0.f),
    C_Vertex3f(0.253417969f, 0.680175781f, -1.f),
    C_Vertex3f(0.171386719f, 0.680175781f, -1.f),
    C_Vertex3f(0.171386719f, 0.680175781f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.334472656f, 0.680175781f, 0.f),
    C_Vertex3f(0.334472656f, 0.680175781f, -1.f),
    C_Vertex3f(0.334472656f, 0.76953125f, -1.f),
    C_Vertex3f(0.334472656f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.334472656f, 0.76953125f, 0.f),
    C_Vertex3f(0.334472656f, 0.76953125f, -1.f),
    C_Vertex3f(0.416503906f, 0.76953125f, -1.f),
    C_Vertex3f(0.416503906f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.416503906f, 0.76953125f, 0.f),
    C_Vertex3f(0.416503906f, 0.76953125f, -1.f),
    C_Vertex3f(0.416503906f, 0.680175781f, -1.f),
    C_Vertex3f(0.416503906f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.416503906f, 0.680175781f, 0.f),
    C_Vertex3f(0.416503906f, 0.680175781f, -1.f),
    C_Vertex3f(0.334472656f, 0.680175781f, -1.f),
    C_Vertex3f(0.334472656f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_Vertex3f(0.282409668f, 0.366283417f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.249668121f, 0.420497894f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Vertex3f(0.105628967f, 0.640625f, 0.f),
    C_Vertex3f(0.00247192383f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.49156189f, 0.640625f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, 0.f),
    C_Vertex3f(0.590362549f, 0.640625f, 0.f),
    C_Vertex3f(0.338072777f, 0.406546593f, 0.f),
    C_Vertex3f(0.334503174f, 0.271369934f, 0.f),
    C_Vertex3f(0.320591927f, 0.37826252f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.304641724f, 0.351566315f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.171386719f, 0.76953125f, 0.f),
    C_Vertex3f(0.171386719f, 0.680175781f, 0.f),
    C_Vertex3f(0.253417969f, 0.680175781f, 0.f),
    C_Vertex3f(0.253417969f, 0.76953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.334472656f, 0.76953125f, 0.f),
    C_Vertex3f(0.334472656f, 0.680175781f, 0.f),
    C_Vertex3f(0.416503906f, 0.680175781f, 0.f),
    C_Vertex3f(0.416503906f, 0.76953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Vertex3f(0.249542236f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_Vertex3f(0.282409668f, 0.366283417f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.249668121f, 0.420497894f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.105628967f, 0.640625f, -1.f),
    C_Vertex3f(0.00247192383f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.49156189f, 0.640625f, -1.f),
    C_Vertex3f(0.357084274f, 0.436418533f, -1.f),
    C_Vertex3f(0.590362549f, 0.640625f, -1.f),
    C_Vertex3f(0.338072777f, 0.406546593f, -1.f),
    C_Vertex3f(0.334503174f, 0.271369934f, -1.f),
    C_Vertex3f(0.320591927f, 0.37826252f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.304641724f, 0.351566315f, -1.f),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.171386719f, 0.76953125f, -1.f),
    C_Vertex3f(0.171386719f, 0.680175781f, -1.f),
    C_Vertex3f(0.253417969f, 0.680175781f, -1.f),
    C_Vertex3f(0.253417969f, 0.76953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.334472656f, 0.76953125f, -1.f),
    C_Vertex3f(0.334472656f, 0.680175781f, -1.f),
    C_Vertex3f(0.416503906f, 0.680175781f, -1.f),
    C_Vertex3f(0.416503906f, 0.76953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.249542236f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_160[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_161[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.195838928f, 0.464355469f, 0.f),
    C_Vertex3f(0.195838928f, 0.464355469f, -1.f),
    C_Vertex3f(0.195838928f, 0.375f, -1.f),
    C_Vertex3f(0.195838928f, 0.375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.195838928f, 0.375f, 0.f),
    C_Vertex3f(0.195838928f, 0.375f, -1.f),
    C_Vertex3f(0.105018616f, 0.375f, -1.f),
    C_Vertex3f(0.105018616f, 0.375f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.105018616f, 0.375f, 0.f),
    C_Vertex3f(0.105018616f, 0.375f, -1.f),
    C_Vertex3f(0.105018616f, 0.464355469f, -1.f),
    C_Vertex3f(0.105018616f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.105018616f, 0.464355469f, 0.f),
    C_Vertex3f(0.105018616f, 0.464355469f, -1.f),
    C_Vertex3f(0.195838928f, 0.464355469f, -1.f),
    C_Vertex3f(0.195838928f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.997528553f, 0.0702630058f, 0.f),
    C_Vertex3f(0.175331116f, 0.305175781f, 0.f),
    C_Vertex3f(0.175331116f, 0.305175781f, -1.f),
    C_Vertex3f(0.199256897f, -0.0345001221f, -1.f),
    C_Vertex3f(0.199256897f, -0.0345001221f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.199256897f, -0.0345001221f, 0.f),
    C_Vertex3f(0.199256897f, -0.0345001221f, -1.f),
    C_Vertex3f(0.199256897f, -0.176757813f, -1.f),
    C_Vertex3f(0.199256897f, -0.176757813f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.199256897f, -0.176757813f, 0.f),
    C_Vertex3f(0.199256897f, -0.176757813f, -1.f),
    C_Vertex3f(0.101600647f, -0.176757813f, -1.f),
    C_Vertex3f(0.101600647f, -0.176757813f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.101600647f, -0.176757813f, 0.f),
    C_Vertex3f(0.101600647f, -0.176757813f, -1.f),
    C_Vertex3f(0.101600647f, -0.0345001221f, -1.f),
    C_Vertex3f(0.101600647f, -0.0345001221f, 0.f),
    C_Normal3f(-0.997725487f, 0.067408435f, 0.f),
    C_Vertex3f(0.101600647f, -0.0345001221f, 0.f),
    C_Vertex3f(0.101600647f, -0.0345001221f, -1.f),
    C_Vertex3f(0.124549866f, 0.305175781f, -1.f),
    C_Vertex3f(0.124549866f, 0.305175781f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.124549866f, 0.305175781f, 0.f),
    C_Vertex3f(0.124549866f, 0.305175781f, -1.f),
    C_Vertex3f(0.175331116f, 0.305175781f, -1.f),
    C_Vertex3f(0.175331116f, 0.305175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.105018616f, 0.464355469f, 0.f),
    C_Vertex3f(0.105018616f, 0.375f, 0.f),
    C_Vertex3f(0.195838928f, 0.375f, 0.f),
    C_Vertex3f(0.195838928f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.199256897f, -0.176757813f, 0.f),
    C_Vertex3f(0.199256897f, -0.0345001221f, 0.f),
    C_Vertex3f(0.175331116f, 0.305175781f, 0.f),
    C_Vertex3f(0.124549866f, 0.305175781f, 0.f),
    C_Vertex3f(0.101600647f, -0.0345001221f, 0.f),
    C_Vertex3f(0.101600647f, -0.176757813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.105018616f, 0.464355469f, -1.f),
    C_Vertex3f(0.105018616f, 0.375f, -1.f),
    C_Vertex3f(0.195838928f, 0.375f, -1.f),
    C_Vertex3f(0.195838928f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.199256897f, -0.176757813f, -1.f),
    C_Vertex3f(0.199256897f, -0.0345001221f, -1.f),
    C_Vertex3f(0.175331116f, 0.305175781f, -1.f),
    C_Vertex3f(0.124549866f, 0.305175781f, -1.f),
    C_Vertex3f(0.101600647f, -0.0345001221f, -1.f),
    C_Vertex3f(0.101600647f, -0.176757813f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_162[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.960344732f, -0.278815418f, 0.f),
    C_Vertex3f(0.324981689f, 0.390205383f, 0.f),
    C_Vertex3f(0.324981689f, 0.390205383f, -1.f),
    C_Vertex3f(0.228012085f, 0.0562057495f, -1.f),
    C_Vertex3f(0.228012085f, 0.0562057495f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.264800698f, 0.964303136f, 0.f),
    C_Vertex3f(0.228012085f, 0.0562057495f, 0.f),
    C_Vertex3f(0.228012085f, 0.0562057495f, -1.f),
    C_Normal3f(0.189307347f, 0.981917858f, 0.f),
    C_Vertex3f(0.235221863f, 0.0542259216f, 0.f),
    C_Vertex3f(0.235221863f, 0.0542259216f, -1.f),
    C_Normal3f(0.0785451531f, 0.996910512f, 0.f),
    C_Vertex3f(0.248550415f, 0.0522460938f, 0.f),
    C_Vertex3f(0.248550415f, 0.0522460938f, -1.f),
    C_Normal3f(-0.0424260795f, 0.999099612f, 0.f),
    C_Vertex3f(0.260350227f, 0.0522460938f, 0.f),
    C_Vertex3f(0.260350227f, 0.0522460938f, -1.f),
    C_Normal3f(-0.133377761f, 0.991065323f, 0.f),
    C_Vertex3f(0.271323085f, 0.0532131195f, 0.f),
    C_Vertex3f(0.271323085f, 0.0532131195f, -1.f),
    C_Normal3f(-0.226647884f, 0.973976791f, 0.f),
    C_Vertex3f(0.281906724f, 0.055147171f, 0.f),
    C_Vertex3f(0.281906724f, 0.055147171f, -1.f),
    C_Normal3f(-0.320599467f, 0.947214901f, 0.f),
    C_Vertex3f(0.292101145f, 0.0580482483f, 0.f),
    C_Vertex3f(0.292101145f, 0.0580482483f, -1.f),
    C_Normal3f(-0.412479967f, 0.910966754f, 0.f),
    C_Vertex3f(0.301906347f, 0.0619163513f, 0.f),
    C_Vertex3f(0.301906347f, 0.0619163513f, -1.f),
    C_Normal3f(-0.499625027f, 0.866241753f, 0.f),
    C_Vertex3f(0.311322331f, 0.0667514801f, 0.f),
    C_Vertex3f(0.311322331f, 0.0667514801f, -1.f),
    C_Normal3f(-0.579831481f, 0.814736485f, 0.f),
    C_Vertex3f(0.320349097f, 0.0725536346f, 0.f),
    C_Vertex3f(0.320349097f, 0.0725536346f, -1.f),
    C_Normal3f(-0.651613891f, 0.758550763f, 0.f),
    C_Vertex3f(0.328986645f, 0.0793228149f, 0.f),
    C_Vertex3f(0.328986645f, 0.0793228149f, -1.f),
    C_Normal3f(-0.719192803f, 0.694810629f, 0.f),
    C_Vertex3f(0.337234974f, 0.087059021f, 0.f),
    C_Vertex3f(0.337234974f, 0.087059021f, -1.f),
    C_Normal3f(-0.783884883f, 0.620906234f, 0.f),
    C_Vertex3f(0.344684482f, 0.0955715179f, 0.f),
    C_Vertex3f(0.344684482f, 0.0955715179f, -1.f),
    C_Normal3f(-0.83997649f, 0.542622805f, 0.f),
    C_Vertex3f(0.351335168f, 0.104860306f, 0.f),
    C_Vertex3f(0.351335168f, 0.104860306f, -1.f),
    C_Normal3f(-0.886633635f, 0.462472528f, 0.f),
    C_Vertex3f(0.357187033f, 0.114925385f, 0.f),
    C_Vertex3f(0.357187033f, 0.114925385f, -1.f),
    C_Normal3f(-0.923815131f, 0.382838994f, 0.f),
    C_Vertex3f(0.362240076f, 0.125766754f, 0.f),
    C_Vertex3f(0.362240076f, 0.125766754f, -1.f),
    C_Normal3f(-0.95212549f, 0.305707425f, 0.f),
    C_Vertex3f(0.366494298f, 0.137384415f, 0.f),
    C_Vertex3f(0.366494298f, 0.137384415f, -1.f),
    C_Normal3f(-0.972589195f, 0.232530132f, 0.f),
    C_Vertex3f(0.369949698f, 0.149778366f, 0.f),
    C_Vertex3f(0.369949698f, 0.149778366f, -1.f),
    C_Normal3f(-0.984525442f, 0.175241292f, 0.f),
    C_Vertex3f(0.372606277f, 0.162948608f, 0.f),
    C_Vertex3f(0.372606277f, 0.162948608f, -1.f),
    C_Normal3f(-0.991244495f, 0.132039398f, 0.f),
    C_Vertex3f(0.373535156f, 0.169921875f, 0.f),
    C_Vertex3f(0.373535156f, 0.169921875f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.111794762f, 0.99373126f, 0.f),
    C_Vertex3f(0.373535156f, 0.169921875f, 0.f),
    C_Vertex3f(0.373535156f, 0.169921875f, -1.f),
    C_Vertex3f(0.451660156f, 0.161132813f, -1.f),
    C_Vertex3f(0.451660156f, 0.161132813f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.984436214f, -0.175742358f, 0.f),
    C_Vertex3f(0.451660156f, 0.161132813f, 0.f),
    C_Vertex3f(0.451660156f, 0.161132813f, -1.f),
    C_Normal3f(0.975103259f, -0.22175166f, 0.f),
    C_Vertex3f(0.449857712f, 0.151036263f, 0.f),
    C_Vertex3f(0.449857712f, 0.151036263f, -1.f),
    C_Normal3f(0.959435403f, -0.281928301f, 0.f),
    C_Vertex3f(0.44501698f, 0.131920934f, 0.f),
    C_Vertex3f(0.44501698f, 0.131920934f, -1.f),
    C_Normal3f(0.933983684f, -0.35731566f, 0.f),
    C_Vertex3f(0.438940406f, 0.113883376f, 0.f),
    C_Vertex3f(0.438940406f, 0.113883376f, -1.f),
    C_Normal3f(0.900487006f, -0.43488276f, 0.f),
    C_Vertex3f(0.431627989f, 0.0969235897f, 0.f),
    C_Vertex3f(0.431627989f, 0.0969235897f, -1.f),
    C_Normal3f(0.858471274f, -0.512861669f, 0.f),
    C_Vertex3f(0.423079729f, 0.0810415745f, 0.f),
    C_Vertex3f(0.423079729f, 0.0810415745f, -1.f),
    C_Normal3f(0.808002055f, -0.589179516f, 0.f),
    C_Vertex3f(0.413295627f, 0.0662373304f, 0.f),
    C_Vertex3f(0.413295627f, 0.0662373304f, -1.f),
    C_Normal3f(0.749788225f, -0.661677897f, 0.f),
    C_Vertex3f(0.402275681f, 0.0525108576f, 0.f),
    C_Vertex3f(0.402275681f, 0.0525108576f, -1.f),
    C_Normal3f(0.68516618f, -0.72838676f, 0.f),
    C_Vertex3f(0.390019894f, 0.0398621559f, 0.f),
    C_Vertex3f(0.390019894f, 0.0398621559f, -1.f),
    C_Normal3f(0.618116856f, -0.786086202f, 0.f),
    C_Vertex3f(0.376528263f, 0.0282912254f, 0.f),
    C_Vertex3f(0.376528263f, 0.0282912254f, -1.f),
    C_Normal3f(0.546588242f, -0.83740145f, 0.f),
    C_Vertex3f(0.362428784f, 0.0181666613f, 0.f),
    C_Vertex3f(0.362428784f, 0.0181666613f, -1.f),
    C_Normal3f(0.468250036f, -0.883595943f, 0.f),
    C_Vertex3f(0.347721457f, 0.0094884634f, 0.f),
    C_Vertex3f(0.347721457f, 0.0094884634f, -1.f),
    C_Normal3f(0.384650171f, -0.923062444f, 0.f),
    C_Vertex3f(0.332406282f, 0.00225663185f, 0.f),
    C_Vertex3f(0.332406282f, 0.00225663185f, -1.f),
    C_Normal3f(0.297811955f, -0.954624593f, 0.f),
    C_Vertex3f(0.316483259f, -0.00352883339f, 0.f),
    C_Vertex3f(0.316483259f, -0.00352883339f, -1.f),
    C_Normal3f(0.209998801f, -0.977701724f, 0.f),
    C_Vertex3f(0.299952388f, -0.00786793232f, 0.f),
    C_Vertex3f(0.299952388f, -0.00786793232f, -1.f),
    C_Normal3f(0.123430789f, -0.992353201f, 0.f),
    C_Vertex3f(0.282813668f, -0.0107606649f, 0.f),
    C_Vertex3f(0.282813668f, -0.0107606649f, -1.f),
    C_Normal3f(0.0384103879f, -0.999262094f, 0.f),
    C_Vertex3f(0.265067101f, -0.0122070313f, 0.f),
    C_Vertex3f(0.265067101f, -0.0122070313f, -1.f),
    C_Normal3f(-0.0694938451f, -0.997582376f, 0.f),
    C_Vertex3f(0.245185852f, -0.0122070313f, 0.f),
    C_Vertex3f(0.245185852f, -0.0122070313f, -1.f),
    C_Normal3f(-0.169657335f, -0.985503078f, 0.f),
    C_Vertex3f(0.222682953f, -0.00925445557f, 0.f),
    C_Vertex3f(0.222682953f, -0.00925445557f, -1.f),
    C_Normal3f(-0.242757007f, -0.970087111f, 0.f),
    C_Vertex3f(0.210884094f, -0.00630187988f, 0.f),
    C_Vertex3f(0.210884094f, -0.00630187988f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.958800197f, -0.284081101f, 0.f),
    C_Vertex3f(0.210884094f, -0.00630187988f, 0.f),
    C_Vertex3f(0.210884094f, -0.00630187988f, -1.f),
    C_Vertex3f(0.159790039f, -0.178749084f, -1.f),
    C_Vertex3f(0.159790039f, -0.178749084f, 0.f),
    C_Normal3f(-0.274839878f, -0.961490035f, 0.f),
    C_Vertex3f(0.159790039f, -0.178749084f, 0.f),
    C_Vertex3f(0.159790039f, -0.178749084f, -1.f),
    C_Vertex3f(0.110839844f, -0.164756775f, -1.f),
    C_Vertex3f(0.110839844f, -0.164756775f, 0.f),
    C_Normal3f(-0.960156202f, 0.279463887f, 0.f),
    C_Vertex3f(0.110839844f, -0.164756775f, 0.f),
    C_Vertex3f(0.110839844f, -0.164756775f, -1.f),
    C_Vertex3f(0.161392212f, 0.0089263916f, -1.f),
    C_Vertex3f(0.161392212f, 0.0089263916f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.429875284f, -0.902888238f, 0.f),
    C_Vertex3f(0.161392212f, 0.0089263916f, 0.f),
    C_Vertex3f(0.161392212f, 0.0089263916f, -1.f),
    C_Normal3f(-0.479055941f, -0.877784431f, 0.f),
    C_Vertex3f(0.15508461f, 0.011929512f, 0.f),
    C_Vertex3f(0.15508461f, 0.011929512f, -1.f),
    C_Normal3f(-0.538582802f, -0.84257257f, 0.f),
    C_Vertex3f(0.143045425f, 0.0189392567f, 0.f),
    C_Vertex3f(0.143045425f, 0.0189392567f, -1.f),
    C_Normal3f(-0.606070042f, -0.795411289f, 0.f),
    C_Vertex3f(0.13158226f, 0.0269525051f, 0.f),
    C_Vertex3f(0.13158226f, 0.0269525051f, -1.f),
    C_Normal3f(-0.668160856f, -0.744016886f, 0.f),
    C_Vertex3f(0.120695114f, 0.0359692574f, 0.f),
    C_Vertex3f(0.120695114f, 0.0359692574f, -1.f),
    C_Normal3f(-0.724063814f, -0.689733028f, 0.f),
    C_Vertex3f(0.110383987f, 0.0459895134f, 0.f),
    C_Vertex3f(0.110383987f, 0.0459895134f, -1.f),
    C_Normal3f(-0.773393571f, -0.633926272f, 0.f),
    C_Vertex3f(0.10064888f, 0.0570132732f, 0.f),
    C_Vertex3f(0.10064888f, 0.0570132732f, -1.f),
    C_Normal3f(-0.81613636f, -0.577859402f, 0.f),
    C_Vertex3f(0.0914897919f, 0.0690405369f, 0.f),
    C_Vertex3f(0.0914897919f, 0.0690405369f, -1.f),
    C_Normal3f(-0.852575123f, -0.522604883f, 0.f),
    C_Vertex3f(0.082906723f, 0.0820713043f, 0.f),
    C_Vertex3f(0.082906723f, 0.0820713043f, -1.f),
    C_Normal3f(-0.887721717f, -0.460380316f, 0.f),
    C_Vertex3f(0.0748996735f, 0.0961055756f, 0.f),
    C_Vertex3f(0.0748996735f, 0.0961055756f, -1.f),
    C_Normal3f(-0.920765162f, -0.390117288f, 0.f),
    C_Vertex3f(0.0678935051f, 0.111020327f, 0.f),
    C_Vertex3f(0.0678935051f, 0.111020327f, -1.f),
    C_Normal3f(-0.947042823f, -0.321107149f, 0.f),
    C_Vertex3f(0.0618882179f, 0.126815557f, 0.f),
    C_Vertex3f(0.0618882179f, 0.126815557f, -1.f),
    C_Normal3f(-0.967077792f, -0.254481047f, 0.f),
    C_Vertex3f(0.056883812f, 0.143491268f, 0.f),
    C_Vertex3f(0.056883812f, 0.143491268f, -1.f),
    C_Normal3f(-0.981576741f, -0.191068217f, 0.f),
    C_Vertex3f(0.0528802872f, 0.161047459f, 0.f),
    C_Vertex3f(0.0528802872f, 0.161047459f, -1.f),
    C_Normal3f(-0.991328835f, -0.131404236f, 0.f),
    C_Vertex3f(0.0498776436f, 0.179484129f, 0.f),
    C_Vertex3f(0.0498776436f, 0.179484129f, -1.f),
    C_Normal3f(-0.997125328f, -0.0757694244f, 0.f),
    C_Vertex3f(0.0478758812f, 0.198801279f, 0.f),
    C_Vertex3f(0.0478758812f, 0.198801279f, -1.f),
    C_Normal3f(-0.999694109f, -0.0247320998f, 0.f),
    C_Vertex3f(0.046875f, 0.218998909f, 0.f),
    C_Vertex3f(0.046875f, 0.218998909f, -1.f),
    C_Normal3f(-0.999791026f, 0.020442931f, 0.f),
    C_Vertex3f(0.046875f, 0.239257813f, 0.f),
    C_Vertex3f(0.046875f, 0.239257813f, -1.f),
    C_Normal3f(-0.997893929f, 0.0648673847f, 0.f),
    C_Vertex3f(0.0476726294f, 0.258008122f, 0.f),
    C_Vertex3f(0.0476726294f, 0.258008122f, -1.f),
    C_Normal3f(-0.993725061f, 0.111850023f, 0.f),
    C_Vertex3f(0.0492678881f, 0.276069045f, 0.f),
    C_Vertex3f(0.0492678881f, 0.276069045f, -1.f),
    C_Normal3f(-0.986823738f, 0.161798567f, 0.f),
    C_Vertex3f(0.0516607761f, 0.29344058f, 0.f),
    C_Vertex3f(0.0516607761f, 0.29344058f, -1.f),
    C_Normal3f(-0.976705968f, 0.214581907f, 0.f),
    C_Vertex3f(0.0548512936f, 0.310122728f, 0.f),
    C_Vertex3f(0.0548512936f, 0.310122728f, -1.f),
    C_Normal3f(-0.962876081f, 0.269943893f, 0.f),
    C_Vertex3f(0.0588394403f, 0.326115489f, 0.f),
    C_Vertex3f(0.0588394403f, 0.326115489f, -1.f),
    C_Normal3f(-0.944857419f, 0.327482045f, 0.f),
    C_Vertex3f(0.0636252165f, 0.341418862f, 0.f),
    C_Vertex3f(0.0636252165f, 0.341418862f, -1.f),
    C_Normal3f(-0.922233045f, 0.386634529f, 0.f),
    C_Vertex3f(0.069208622f, 0.356032848f, 0.f),
    C_Vertex3f(0.069208622f, 0.356032848f, -1.f),
    C_Normal3f(-0.893607676f, 0.448848903f, 0.f),
    C_Vertex3f(0.0755896568f, 0.369957447f, 0.f),
    C_Vertex3f(0.0755896568f, 0.369957447f, -1.f),
    C_Normal3f(-0.857710838f, 0.51413244f, 0.f),
    C_Vertex3f(0.0827615261f, 0.383015156f, 0.f),
    C_Vertex3f(0.0827615261f, 0.383015156f, -1.f),
    C_Normal3f(-0.815043032f, 0.57940048f, 0.f),
    C_Vertex3f(0.0907242298f, 0.395205975f, 0.f),
    C_Vertex3f(0.0907242298f, 0.395205975f, -1.f),
    C_Normal3f(-0.765672028f, 0.643231213f, 0.f),
    C_Vertex3f(0.0994777679f, 0.406529903f, 0.f),
    C_Vertex3f(0.0994777679f, 0.406529903f, -1.f),
    C_Normal3f(-0.710068882f, 0.704132259f, 0.f),
    C_Vertex3f(0.109022141f, 0.416986942f, 0.f),
    C_Vertex3f(0.109022141f, 0.416986942f, -1.f),
    C_Normal3f(-0.649115384f, 0.760690033f, 0.f),
    C_Vertex3f(0.119357347f, 0.426577091f, 0.f),
    C_Vertex3f(0.119357347f, 0.426577091f, -1.f),
    C_Normal3f(-0.584043205f, 0.811722577f, 0.f),
    C_Vertex3f(0.130483389f, 0.43530035f, 0.f),
    C_Vertex3f(0.130483389f, 0.43530035f, -1.f),
    C_Normal3f(-0.50945276f, 0.860498667f, 0.f),
    C_Vertex3f(0.142400265f, 0.443156719f, 0.f),
    C_Vertex3f(0.142400265f, 0.443156719f, -1.f),
    C_Normal3f(-0.424512535f, 0.905422091f, 0.f),
    C_Vertex3f(0.16146183f, 0.453640938f, 0.f),
    C_Vertex3f(0.16146183f, 0.453640938f, -1.f),
    C_Normal3f(-0.320926785f, 0.947104037f, 0.f),
    C_Vertex3f(0.187122822f, 0.464125156f, 0.f),
    C_Vertex3f(0.187122822f, 0.464125156f, -1.f),
    C_Normal3f(-0.197428316f, 0.980317295f, 0.f),
    C_Vertex3f(0.213029385f, 0.471114635f, 0.f),
    C_Vertex3f(0.213029385f, 0.471114635f, -1.f),
    C_Normal3f(-0.0804885551f, 0.99675554f, 0.f),
    C_Vertex3f(0.239181519f, 0.474609375f, 0.f),
    C_Vertex3f(0.239181519f, 0.474609375f, -1.f),
    C_Normal3f(0.0138141168f, 0.999904633f, 0.f),
    C_Vertex3f(0.256307602f, 0.474609375f, 0.f),
    C_Vertex3f(0.256307602f, 0.474609375f, -1.f),
    C_Normal3f(0.0540688522f, 0.998537183f, 0.f),
    C_Vertex3f(0.265688896f, 0.474243164f, 0.f),
    C_Vertex3f(0.265688896f, 0.474243164f, -1.f),
    C_Normal3f(0.0782007724f, 0.996937692f, 0.f),
    C_Vertex3f(0.276597023f, 0.473510742f, 0.f),
    C_Vertex3f(0.276597023f, 0.473510742f, -1.f),
    C_Normal3f(0.093890585f, 0.995582521f, 0.f),
    C_Vertex3f(0.289031982f, 0.472412109f, 0.f),
    C_Vertex3f(0.289031982f, 0.472412109f, -1.f),
    C_Normal3f(0.104345292f, 0.994541109f, 0.f),
    C_Vertex3f(0.296012878f, 0.471679688f, 0.f),
    C_Vertex3f(0.296012878f, 0.471679688f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.960178375f, 0.279387653f, 0.f),
    C_Vertex3f(0.296012878f, 0.471679688f, 0.f),
    C_Vertex3f(0.296012878f, 0.471679688f, -1.f),
    C_Vertex3f(0.345207214f, 0.64074707f, -1.f),
    C_Vertex3f(0.345207214f, 0.64074707f, 0.f),
    C_Normal3f(0.280331135f, 0.959903419f, 0.f),
    C_Vertex3f(0.345207214f, 0.64074707f, 0.f),
    C_Vertex3f(0.345207214f, 0.64074707f, -1.f),
    C_Vertex3f(0.393615723f, 0.626609802f, -1.f),
    C_Vertex3f(0.393615723f, 0.626609802f, 0.f),
    C_Normal3f(0.960350335f, -0.278796017f, 0.f),
    C_Vertex3f(0.393615723f, 0.626609802f, 0.f),
    C_Vertex3f(0.393615723f, 0.626609802f, -1.f),
    C_Vertex3f(0.344558716f, 0.457626343f, -1.f),
    C_Vertex3f(0.344558716f, 0.457626343f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.399606317f, 0.916686833f, 0.f),
    C_Vertex3f(0.344558716f, 0.457626343f, 0.f),
    C_Vertex3f(0.344558716f, 0.457626343f, -1.f),
    C_Normal3f(0.465898007f, 0.884838521f, 0.f),
    C_Vertex3f(0.355379105f, 0.45290947f, 0.f),
    C_Vertex3f(0.355379105f, 0.45290947f, -1.f),
    C_Normal3f(0.552982867f, 0.833192587f, 0.f),
    C_Vertex3f(0.374506474f, 0.441857815f, 0.f),
    C_Vertex3f(0.374506474f, 0.441857815f, -1.f),
    C_Normal3f(0.659637988f, 0.751583397f, 0.f),
    C_Vertex3f(0.391120434f, 0.429188251f, 0.f),
    C_Vertex3f(0.391120434f, 0.429188251f, -1.f),
    C_Normal3f(0.76164335f, 0.647996485f, 0.f),
    C_Vertex3f(0.405220985f, 0.41490078f, 0.f),
    C_Vertex3f(0.405220985f, 0.41490078f, -1.f),
    C_Normal3f(0.844700813f, 0.535238862f, 0.f),
    C_Vertex3f(0.416808128f, 0.398995399f, 0.f),
    C_Vertex3f(0.416808128f, 0.398995399f, -1.f),
    C_Normal3f(0.903612554f, 0.428350836f, 0.f),
    C_Vertex3f(0.426645756f, 0.381088734f, 0.f),
    C_Vertex3f(0.426645756f, 0.381088734f, -1.f),
    C_Normal3f(0.945324242f, 0.32613185f, 0.f),
    C_Vertex3f(0.434733868f, 0.361180782f, 0.f),
    C_Vertex3f(0.434733868f, 0.361180782f, -1.f),
    C_Normal3f(0.969007611f, 0.247030839f, 0.f),
    C_Vertex3f(0.441072464f, 0.339271545f, 0.f),
    C_Vertex3f(0.441072464f, 0.339271545f, -1.f),
    C_Normal3f(0.982075632f, 0.188487083f, 0.f),
    C_Vertex3f(0.443367004f, 0.327316284f, 0.f),
    C_Vertex3f(0.443367004f, 0.327316284f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.152331665f, -0.98832947f, 0.f),
    C_Vertex3f(0.443367004f, 0.327316284f, 0.f),
    C_Vertex3f(0.443367004f, 0.327316284f, -1.f),
    C_Vertex3f(0.366989136f, 0.315544128f, -1.f),
    C_Vertex3f(0.366989136f, 0.315544128f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.973997891f, -0.226557076f, 0.f),
    C_Vertex3f(0.366989136f, 0.315544128f, 0.f),
    C_Vertex3f(0.366989136f, 0.315544128f, -1.f),
    C_Normal3f(-0.951376498f, -0.308030486f, 0.f),
    C_Vertex3f(0.364145279f, 0.327770233f, 0.f),
    C_Vertex3f(0.364145279f, 0.327770233f, -1.f),
    C_Normal3f(-0.907687128f, -0.419647664f, 0.f),
    C_Vertex3f(0.356050491f, 0.349328995f, 0.f),
    C_Vertex3f(0.356050491f, 0.349328995f, -1.f),
    C_Normal3f(-0.826998949f, -0.562203526f, 0.f),
    C_Vertex3f(0.34554863f, 0.367994308f, 0.f),
    C_Vertex3f(0.34554863f, 0.367994308f, -1.f),
    C_Normal3f(-0.733741999f, -0.679428101f, 0.f),
    C_Vertex3f(0.332639694f, 0.383766174f, 0.f),
    C_Vertex3f(0.332639694f, 0.383766174f, -1.f),
    C_Normal3f(-0.643571973f, -0.765385509f, 0.f),
    C_Vertex3f(0.324981689f, 0.390205383f, 0.f),
    C_Vertex3f(0.324981689f, 0.390205383f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.104855351f, -0.994487464f, 0.f),
    C_Vertex3f(0.278121948f, 0.41015625f, 0.f),
    C_Vertex3f(0.278121948f, 0.41015625f, -1.f),
    C_Normal3f(-0.0820201859f, -0.996630669f, 0.f),
    C_Vertex3f(0.271175385f, 0.410888672f, 0.f),
    C_Vertex3f(0.271175385f, 0.410888672f, -1.f),
    C_Normal3f(-0.0309454221f, -0.999521017f, 0.f),
    C_Vertex3f(0.260322571f, 0.411621094f, 0.f),
    C_Vertex3f(0.260322571f, 0.411621094f, -1.f),
    C_Normal3f(0.0822884887f, -0.996608555f, 0.f),
    C_Vertex3f(0.247518539f, 0.411621094f, 0.f),
    C_Vertex3f(0.247518539f, 0.411621094f, -1.f),
    C_Normal3f(0.214432254f, -0.97673893f, 0.f),
    C_Vertex3f(0.23029232f, 0.409141541f, 0.f),
    C_Vertex3f(0.23029232f, 0.409141541f, -1.f),
    C_Normal3f(0.354091704f, -0.935210705f, 0.f),
    C_Vertex3f(0.213635445f, 0.404182434f, 0.f),
    C_Vertex3f(0.213635445f, 0.404182434f, -1.f),
    C_Normal3f(0.481359214f, -0.876523376f, 0.f),
    C_Vertex3f(0.197547913f, 0.396743774f, 0.f),
    C_Vertex3f(0.197547913f, 0.396743774f, -1.f),
    C_Normal3f(0.612945557f, -0.790125251f, 0.f),
    C_Vertex3f(0.182029724f, 0.386825562f, 0.f),
    C_Vertex3f(0.182029724f, 0.386825562f, -1.f),
    C_Normal3f(0.743737757f, -0.668471396f, 0.f),
    C_Vertex3f(0.168464661f, 0.374182224f, 0.f),
    C_Vertex3f(0.168464661f, 0.374182224f, -1.f),
    C_Normal3f(0.843927324f, -0.536457539f, 0.f),
    C_Vertex3f(0.156852722f, 0.358813763f, 0.f),
    C_Vertex3f(0.156852722f, 0.358813763f, -1.f),
    C_Normal3f(0.913199484f, -0.407512695f, 0.f),
    C_Vertex3f(0.147193909f, 0.340720177f, 0.f),
    C_Vertex3f(0.147193909f, 0.340720177f, -1.f),
    C_Normal3f(0.954768598f, -0.29734987f, 0.f),
    C_Vertex3f(0.13948822f, 0.319901466f, 0.f),
    C_Vertex3f(0.13948822f, 0.319901466f, -1.f),
    C_Normal3f(0.979318619f, -0.202324077f, 0.f),
    C_Vertex3f(0.133708954f, 0.297420979f, 0.f),
    C_Vertex3f(0.133708954f, 0.297420979f, -1.f),
    C_Normal3f(0.993372142f, -0.11494267f, 0.f),
    C_Vertex3f(0.12985611f, 0.273278713f, 0.f),
    C_Vertex3f(0.12985611f, 0.273278713f, -1.f),
    C_Normal3f(0.999362469f, -0.0357019305f, 0.f),
    C_Vertex3f(0.127929688f, 0.24747467f, 0.f),
    C_Vertex3f(0.127929688f, 0.24747467f, -1.f),
    C_Normal3f(0.999494791f, 0.0317816176f, 0.f),
    C_Vertex3f(0.127929688f, 0.21935463f, 0.f),
    C_Vertex3f(0.127929688f, 0.21935463f, -1.f),
    C_Normal3f(0.994783163f, 0.102013014f, 0.f),
    C_Vertex3f(0.129667759f, 0.19281435f, 0.f),
    C_Vertex3f(0.129667759f, 0.19281435f, -1.f),
    C_Normal3f(0.982893825f, 0.184173226f, 0.f),
    C_Vertex3f(0.133143902f, 0.168508053f, 0.f),
    C_Vertex3f(0.133143902f, 0.168508053f, -1.f),
    C_Normal3f(0.960352957f, 0.278786778f, 0.f),
    C_Vertex3f(0.138358116f, 0.146435738f, 0.f),
    C_Vertex3f(0.138358116f, 0.146435738f, -1.f),
    C_Normal3f(0.923893452f, 0.382649869f, 0.f),
    C_Vertex3f(0.145310402f, 0.126597404f, 0.f),
    C_Vertex3f(0.145310402f, 0.126597404f, -1.f),
    C_Normal3f(0.867939591f, 0.49666965f, 0.f),
    C_Vertex3f(0.153729916f, 0.109321117f, 0.f),
    C_Vertex3f(0.153729916f, 0.109321117f, -1.f),
    C_Normal3f(0.784451365f, 0.620190382f, 0.f),
    C_Vertex3f(0.163616657f, 0.0946068764f, 0.f),
    C_Vertex3f(0.163616657f, 0.0946068764f, -1.f),
    C_Normal3f(0.69026649f, 0.723555267f, 0.f),
    C_Vertex3f(0.174970627f, 0.0824546814f, 0.f),
    C_Vertex3f(0.174970627f, 0.0824546814f, -1.f),
    C_Normal3f(0.598970294f, 0.800771356f, 0.f),
    C_Vertex3f(0.181381226f, 0.0776596069f, 0.f),
    C_Vertex3f(0.181381226f, 0.0776596069f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.960184097f, 0.279367954f, 0.f),
    C_Vertex3f(0.181381226f, 0.0776596069f, 0.f),
    C_Vertex3f(0.181381226f, 0.0776596069f, -1.f),
    C_Vertex3f(0.278121948f, 0.41015625f, -1.f),
    C_Vertex3f(0.278121948f, 0.41015625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.159790039f, -0.178749084f, 0.f),
    C_Vertex3f(0.210884094f, -0.00630187988f, 0.f),
    C_Vertex3f(0.161392212f, 0.0089263916f, 0.f),
    C_Vertex3f(0.163616657f, 0.0946068764f, 0.f),
    C_Vertex3f(0.15508461f, 0.011929512f, 0.f),
    C_Vertex3f(0.153729916f, 0.109321117f, 0.f),
    C_Vertex3f(0.145310402f, 0.126597404f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.143045425f, 0.0189392567f, 0.f),
    C_Vertex3f(0.15508461f, 0.011929512f, 0.f),
    C_Vertex3f(0.145310402f, 0.126597404f, 0.f),
    C_Vertex3f(0.138358116f, 0.146435738f, 0.f),
    C_Vertex3f(0.133143902f, 0.168508053f, 0.f),
    C_Vertex3f(0.13158226f, 0.0269525051f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133143902f, 0.168508053f, 0.f),
    C_Vertex3f(0.129667759f, 0.19281435f, 0.f),
    C_Vertex3f(0.13158226f, 0.0269525051f, 0.f),
    C_Vertex3f(0.127929688f, 0.21935463f, 0.f),
    C_Vertex3f(0.120695114f, 0.0359692574f, 0.f),
    C_Vertex3f(0.127929688f, 0.24747467f, 0.f),
    C_Vertex3f(0.130483389f, 0.43530035f, 0.f),
    C_Vertex3f(0.12985611f, 0.273278713f, 0.f),
    C_Vertex3f(0.133708954f, 0.297420979f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.345207214f, 0.64074707f, 0.f),
    C_Vertex3f(0.296012878f, 0.471679688f, 0.f),
    C_Vertex3f(0.324981689f, 0.390205383f, 0.f),
    C_Vertex3f(0.289031982f, 0.472412109f, 0.f),
    C_Vertex3f(0.278121948f, 0.41015625f, 0.f),
    C_Vertex3f(0.276597023f, 0.473510742f, 0.f),
    C_Vertex3f(0.271175385f, 0.410888672f, 0.f),
    C_Vertex3f(0.265688896f, 0.474243164f, 0.f),
    C_Vertex3f(0.260322571f, 0.411621094f, 0.f),
    C_Vertex3f(0.256307602f, 0.474609375f, 0.f),
    C_Vertex3f(0.247518539f, 0.411621094f, 0.f),
    C_Vertex3f(0.239181519f, 0.474609375f, 0.f),
    C_Vertex3f(0.23029232f, 0.409141541f, 0.f),
    C_Vertex3f(0.213635445f, 0.404182434f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168464661f, 0.374182224f, 0.f),
    C_Vertex3f(0.182029724f, 0.386825562f, 0.f),
    C_Vertex3f(0.187122822f, 0.464125156f, 0.f),
    C_Vertex3f(0.197547913f, 0.396743774f, 0.f),
    C_Vertex3f(0.213029385f, 0.471114635f, 0.f),
    C_Vertex3f(0.213635445f, 0.404182434f, 0.f),
    C_Vertex3f(0.239181519f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.16146183f, 0.453640938f, 0.f),
    C_Vertex3f(0.142400265f, 0.443156719f, 0.f),
    C_Vertex3f(0.147193909f, 0.340720177f, 0.f),
    C_Vertex3f(0.156852722f, 0.358813763f, 0.f),
    C_Vertex3f(0.168464661f, 0.374182224f, 0.f),
    C_Vertex3f(0.187122822f, 0.464125156f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.142400265f, 0.443156719f, 0.f),
    C_Vertex3f(0.130483389f, 0.43530035f, 0.f),
    C_Vertex3f(0.133708954f, 0.297420979f, 0.f),
    C_Vertex3f(0.13948822f, 0.319901466f, 0.f),
    C_Vertex3f(0.147193909f, 0.340720177f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0827615261f, 0.383015156f, 0.f),
    C_Vertex3f(0.082906723f, 0.0820713043f, 0.f),
    C_Vertex3f(0.0907242298f, 0.395205975f, 0.f),
    C_Vertex3f(0.0914897919f, 0.0690405369f, 0.f),
    C_Vertex3f(0.0994777679f, 0.406529903f, 0.f),
    C_Vertex3f(0.10064888f, 0.0570132732f, 0.f),
    C_Vertex3f(0.109022141f, 0.416986942f, 0.f),
    C_Vertex3f(0.110383987f, 0.0459895134f, 0.f),
    C_Vertex3f(0.119357347f, 0.426577091f, 0.f),
    C_Vertex3f(0.120695114f, 0.0359692574f, 0.f),
    C_Vertex3f(0.130483389f, 0.43530035f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0476726294f, 0.258008122f, 0.f),
    C_Vertex3f(0.0478758812f, 0.198801279f, 0.f),
    C_Vertex3f(0.0492678881f, 0.276069045f, 0.f),
    C_Vertex3f(0.0498776436f, 0.179484129f, 0.f),
    C_Vertex3f(0.0516607761f, 0.29344058f, 0.f),
    C_Vertex3f(0.0528802872f, 0.161047459f, 0.f),
    C_Vertex3f(0.0548512936f, 0.310122728f, 0.f),
    C_Vertex3f(0.056883812f, 0.143491268f, 0.f),
    C_Vertex3f(0.0588394403f, 0.326115489f, 0.f),
    C_Vertex3f(0.0618882179f, 0.126815557f, 0.f),
    C_Vertex3f(0.0636252165f, 0.341418862f, 0.f),
    C_Vertex3f(0.0678935051f, 0.111020327f, 0.f),
    C_Vertex3f(0.069208622f, 0.356032848f, 0.f),
    C_Vertex3f(0.0748996735f, 0.0961055756f, 0.f),
    C_Vertex3f(0.0755896568f, 0.369957447f, 0.f),
    C_Vertex3f(0.082906723f, 0.0820713043f, 0.f),
    C_Vertex3f(0.0827615261f, 0.383015156f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.046875f, 0.239257813f, 0.f),
    C_Vertex3f(0.046875f, 0.218998909f, 0.f),
    C_Vertex3f(0.0478758812f, 0.198801279f, 0.f),
    C_Vertex3f(0.0476726294f, 0.258008122f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.344558716f, 0.457626343f, 0.f),
    C_Vertex3f(0.393615723f, 0.626609802f, 0.f),
    C_Vertex3f(0.345207214f, 0.64074707f, 0.f),
    C_Vertex3f(0.332639694f, 0.383766174f, 0.f),
    C_Vertex3f(0.34554863f, 0.367994308f, 0.f),
    C_Vertex3f(0.355379105f, 0.45290947f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.228012085f, 0.0562057495f, 0.f),
    C_Vertex3f(0.324981689f, 0.390205383f, 0.f),
    C_Vertex3f(0.278121948f, 0.41015625f, 0.f),
    C_Vertex3f(0.210884094f, -0.00630187988f, 0.f),
    C_Vertex3f(0.222682953f, -0.00925445557f, 0.f),
    C_Vertex3f(0.245185852f, -0.0122070313f, 0.f),
    C_Vertex3f(0.235221863f, 0.0542259216f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.210884094f, -0.00630187988f, 0.f),
    C_Vertex3f(0.278121948f, 0.41015625f, 0.f),
    C_Vertex3f(0.181381226f, 0.0776596069f, 0.f),
    C_Vertex3f(0.174970627f, 0.0824546814f, 0.f),
    C_Vertex3f(0.163616657f, 0.0946068764f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.443367004f, 0.327316284f, 0.f),
    C_Vertex3f(0.441072464f, 0.339271545f, 0.f),
    C_Vertex3f(0.434733868f, 0.361180782f, 0.f),
    C_Vertex3f(0.426645756f, 0.381088734f, 0.f),
    C_Vertex3f(0.416808128f, 0.398995399f, 0.f),
    C_Vertex3f(0.405220985f, 0.41490078f, 0.f),
    C_Vertex3f(0.391120434f, 0.429188251f, 0.f),
    C_Vertex3f(0.374506474f, 0.441857815f, 0.f),
    C_Vertex3f(0.366989136f, 0.315544128f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.374506474f, 0.441857815f, 0.f),
    C_Vertex3f(0.355379105f, 0.45290947f, 0.f),
    C_Vertex3f(0.356050491f, 0.349328995f, 0.f),
    C_Vertex3f(0.364145279f, 0.327770233f, 0.f),
    C_Vertex3f(0.366989136f, 0.315544128f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.451660156f, 0.161132813f, 0.f),
    C_Vertex3f(0.373535156f, 0.169921875f, 0.f),
    C_Vertex3f(0.376528263f, 0.0282912254f, 0.f),
    C_Vertex3f(0.390019894f, 0.0398621559f, 0.f),
    C_Vertex3f(0.402275681f, 0.0525108576f, 0.f),
    C_Vertex3f(0.413295627f, 0.0662373304f, 0.f),
    C_Vertex3f(0.423079729f, 0.0810415745f, 0.f),
    C_Vertex3f(0.431627989f, 0.0969235897f, 0.f),
    C_Vertex3f(0.438940406f, 0.113883376f, 0.f),
    C_Vertex3f(0.44501698f, 0.131920934f, 0.f),
    C_Vertex3f(0.449857712f, 0.151036263f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376528263f, 0.0282912254f, 0.f),
    C_Vertex3f(0.373535156f, 0.169921875f, 0.f),
    C_Vertex3f(0.372606277f, 0.162948608f, 0.f),
    C_Vertex3f(0.369949698f, 0.149778366f, 0.f),
    C_Vertex3f(0.366494298f, 0.137384415f, 0.f),
    C_Vertex3f(0.362428784f, 0.0181666613f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.362428784f, 0.0181666613f, 0.f),
    C_Vertex3f(0.366494298f, 0.137384415f, 0.f),
    C_Vertex3f(0.362240076f, 0.125766754f, 0.f),
    C_Vertex3f(0.357187033f, 0.114925385f, 0.f),
    C_Vertex3f(0.351335168f, 0.104860306f, 0.f),
    C_Vertex3f(0.347721457f, 0.0094884634f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.351335168f, 0.104860306f, 0.f),
    C_Vertex3f(0.344684482f, 0.0955715179f, 0.f),
    C_Vertex3f(0.347721457f, 0.0094884634f, 0.f),
    C_Vertex3f(0.337234974f, 0.087059021f, 0.f),
    C_Vertex3f(0.332406282f, 0.00225663185f, 0.f),
    C_Vertex3f(0.328986645f, 0.0793228149f, 0.f),
    C_Vertex3f(0.320349097f, 0.0725536346f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.316483259f, -0.00352883339f, 0.f),
    C_Vertex3f(0.332406282f, 0.00225663185f, 0.f),
    C_Vertex3f(0.320349097f, 0.0725536346f, 0.f),
    C_Vertex3f(0.311322331f, 0.0667514801f, 0.f),
    C_Vertex3f(0.301906347f, 0.0619163513f, 0.f),
    C_Vertex3f(0.299952388f, -0.00786793232f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.292101145f, 0.0580482483f, 0.f),
    C_Vertex3f(0.281906724f, 0.055147171f, 0.f),
    C_Vertex3f(0.282813668f, -0.0107606649f, 0.f),
    C_Vertex3f(0.299952388f, -0.00786793232f, 0.f),
    C_Vertex3f(0.301906347f, 0.0619163513f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.271323085f, 0.0532131195f, 0.f),
    C_Vertex3f(0.260350227f, 0.0522460938f, 0.f),
    C_Vertex3f(0.265067101f, -0.0122070313f, 0.f),
    C_Vertex3f(0.282813668f, -0.0107606649f, 0.f),
    C_Vertex3f(0.281906724f, 0.055147171f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.248550415f, 0.0522460938f, 0.f),
    C_Vertex3f(0.235221863f, 0.0542259216f, 0.f),
    C_Vertex3f(0.245185852f, -0.0122070313f, 0.f),
    C_Vertex3f(0.265067101f, -0.0122070313f, 0.f),
    C_Vertex3f(0.260350227f, 0.0522460938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.355379105f, 0.45290947f, 0.f),
    C_Vertex3f(0.34554863f, 0.367994308f, 0.f),
    C_Vertex3f(0.356050491f, 0.349328995f, 0.f),
    C_Vertex3f(0.159790039f, -0.178749084f, 0.f),
    C_Vertex3f(0.161392212f, 0.0089263916f, 0.f),
    C_Vertex3f(0.110839844f, -0.164756775f, 0.f),
    C_Vertex3f(0.345207214f, 0.64074707f, 0.f),
    C_Vertex3f(0.324981689f, 0.390205383f, 0.f),
    C_Vertex3f(0.332639694f, 0.383766174f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.159790039f, -0.178749084f, -1.f),
    C_Vertex3f(0.210884094f, -0.00630187988f, -1.f),
    C_Vertex3f(0.161392212f, 0.0089263916f, -1.f),
    C_Vertex3f(0.163616657f, 0.0946068764f, -1.f),
    C_Vertex3f(0.15508461f, 0.011929512f, -1.f),
    C_Vertex3f(0.153729916f, 0.109321117f, -1.f),
    C_Vertex3f(0.145310402f, 0.126597404f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.143045425f, 0.0189392567f, -1.f),
    C_Vertex3f(0.15508461f, 0.011929512f, -1.f),
    C_Vertex3f(0.145310402f, 0.126597404f, -1.f),
    C_Vertex3f(0.138358116f, 0.146435738f, -1.f),
    C_Vertex3f(0.133143902f, 0.168508053f, -1.f),
    C_Vertex3f(0.13158226f, 0.0269525051f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133143902f, 0.168508053f, -1.f),
    C_Vertex3f(0.129667759f, 0.19281435f, -1.f),
    C_Vertex3f(0.13158226f, 0.0269525051f, -1.f),
    C_Vertex3f(0.127929688f, 0.21935463f, -1.f),
    C_Vertex3f(0.120695114f, 0.0359692574f, -1.f),
    C_Vertex3f(0.127929688f, 0.24747467f, -1.f),
    C_Vertex3f(0.130483389f, 0.43530035f, -1.f),
    C_Vertex3f(0.12985611f, 0.273278713f, -1.f),
    C_Vertex3f(0.133708954f, 0.297420979f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.345207214f, 0.64074707f, -1.f),
    C_Vertex3f(0.296012878f, 0.471679688f, -1.f),
    C_Vertex3f(0.324981689f, 0.390205383f, -1.f),
    C_Vertex3f(0.289031982f, 0.472412109f, -1.f),
    C_Vertex3f(0.278121948f, 0.41015625f, -1.f),
    C_Vertex3f(0.276597023f, 0.473510742f, -1.f),
    C_Vertex3f(0.271175385f, 0.410888672f, -1.f),
    C_Vertex3f(0.265688896f, 0.474243164f, -1.f),
    C_Vertex3f(0.260322571f, 0.411621094f, -1.f),
    C_Vertex3f(0.256307602f, 0.474609375f, -1.f),
    C_Vertex3f(0.247518539f, 0.411621094f, -1.f),
    C_Vertex3f(0.239181519f, 0.474609375f, -1.f),
    C_Vertex3f(0.23029232f, 0.409141541f, -1.f),
    C_Vertex3f(0.213635445f, 0.404182434f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168464661f, 0.374182224f, -1.f),
    C_Vertex3f(0.182029724f, 0.386825562f, -1.f),
    C_Vertex3f(0.187122822f, 0.464125156f, -1.f),
    C_Vertex3f(0.197547913f, 0.396743774f, -1.f),
    C_Vertex3f(0.213029385f, 0.471114635f, -1.f),
    C_Vertex3f(0.213635445f, 0.404182434f, -1.f),
    C_Vertex3f(0.239181519f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.16146183f, 0.453640938f, -1.f),
    C_Vertex3f(0.142400265f, 0.443156719f, -1.f),
    C_Vertex3f(0.147193909f, 0.340720177f, -1.f),
    C_Vertex3f(0.156852722f, 0.358813763f, -1.f),
    C_Vertex3f(0.168464661f, 0.374182224f, -1.f),
    C_Vertex3f(0.187122822f, 0.464125156f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.142400265f, 0.443156719f, -1.f),
    C_Vertex3f(0.130483389f, 0.43530035f, -1.f),
    C_Vertex3f(0.133708954f, 0.297420979f, -1.f),
    C_Vertex3f(0.13948822f, 0.319901466f, -1.f),
    C_Vertex3f(0.147193909f, 0.340720177f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0827615261f, 0.383015156f, -1.f),
    C_Vertex3f(0.082906723f, 0.0820713043f, -1.f),
    C_Vertex3f(0.0907242298f, 0.395205975f, -1.f),
    C_Vertex3f(0.0914897919f, 0.0690405369f, -1.f),
    C_Vertex3f(0.0994777679f, 0.406529903f, -1.f),
    C_Vertex3f(0.10064888f, 0.0570132732f, -1.f),
    C_Vertex3f(0.109022141f, 0.416986942f, -1.f),
    C_Vertex3f(0.110383987f, 0.0459895134f, -1.f),
    C_Vertex3f(0.119357347f, 0.426577091f, -1.f),
    C_Vertex3f(0.120695114f, 0.0359692574f, -1.f),
    C_Vertex3f(0.130483389f, 0.43530035f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0476726294f, 0.258008122f, -1.f),
    C_Vertex3f(0.0478758812f, 0.198801279f, -1.f),
    C_Vertex3f(0.0492678881f, 0.276069045f, -1.f),
    C_Vertex3f(0.0498776436f, 0.179484129f, -1.f),
    C_Vertex3f(0.0516607761f, 0.29344058f, -1.f),
    C_Vertex3f(0.0528802872f, 0.161047459f, -1.f),
    C_Vertex3f(0.0548512936f, 0.310122728f, -1.f),
    C_Vertex3f(0.056883812f, 0.143491268f, -1.f),
    C_Vertex3f(0.0588394403f, 0.326115489f, -1.f),
    C_Vertex3f(0.0618882179f, 0.126815557f, -1.f),
    C_Vertex3f(0.0636252165f, 0.341418862f, -1.f),
    C_Vertex3f(0.0678935051f, 0.111020327f, -1.f),
    C_Vertex3f(0.069208622f, 0.356032848f, -1.f),
    C_Vertex3f(0.0748996735f, 0.0961055756f, -1.f),
    C_Vertex3f(0.0755896568f, 0.369957447f, -1.f),
    C_Vertex3f(0.082906723f, 0.0820713043f, -1.f),
    C_Vertex3f(0.0827615261f, 0.383015156f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.046875f, 0.239257813f, -1.f),
    C_Vertex3f(0.046875f, 0.218998909f, -1.f),
    C_Vertex3f(0.0478758812f, 0.198801279f, -1.f),
    C_Vertex3f(0.0476726294f, 0.258008122f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.344558716f, 0.457626343f, -1.f),
    C_Vertex3f(0.393615723f, 0.626609802f, -1.f),
    C_Vertex3f(0.345207214f, 0.64074707f, -1.f),
    C_Vertex3f(0.332639694f, 0.383766174f, -1.f),
    C_Vertex3f(0.34554863f, 0.367994308f, -1.f),
    C_Vertex3f(0.355379105f, 0.45290947f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.228012085f, 0.0562057495f, -1.f),
    C_Vertex3f(0.324981689f, 0.390205383f, -1.f),
    C_Vertex3f(0.278121948f, 0.41015625f, -1.f),
    C_Vertex3f(0.210884094f, -0.00630187988f, -1.f),
    C_Vertex3f(0.222682953f, -0.00925445557f, -1.f),
    C_Vertex3f(0.245185852f, -0.0122070313f, -1.f),
    C_Vertex3f(0.235221863f, 0.0542259216f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.210884094f, -0.00630187988f, -1.f),
    C_Vertex3f(0.278121948f, 0.41015625f, -1.f),
    C_Vertex3f(0.181381226f, 0.0776596069f, -1.f),
    C_Vertex3f(0.174970627f, 0.0824546814f, -1.f),
    C_Vertex3f(0.163616657f, 0.0946068764f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.443367004f, 0.327316284f, -1.f),
    C_Vertex3f(0.441072464f, 0.339271545f, -1.f),
    C_Vertex3f(0.434733868f, 0.361180782f, -1.f),
    C_Vertex3f(0.426645756f, 0.381088734f, -1.f),
    C_Vertex3f(0.416808128f, 0.398995399f, -1.f),
    C_Vertex3f(0.405220985f, 0.41490078f, -1.f),
    C_Vertex3f(0.391120434f, 0.429188251f, -1.f),
    C_Vertex3f(0.374506474f, 0.441857815f, -1.f),
    C_Vertex3f(0.366989136f, 0.315544128f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.374506474f, 0.441857815f, -1.f),
    C_Vertex3f(0.355379105f, 0.45290947f, -1.f),
    C_Vertex3f(0.356050491f, 0.349328995f, -1.f),
    C_Vertex3f(0.364145279f, 0.327770233f, -1.f),
    C_Vertex3f(0.366989136f, 0.315544128f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.451660156f, 0.161132813f, -1.f),
    C_Vertex3f(0.373535156f, 0.169921875f, -1.f),
    C_Vertex3f(0.376528263f, 0.0282912254f, -1.f),
    C_Vertex3f(0.390019894f, 0.0398621559f, -1.f),
    C_Vertex3f(0.402275681f, 0.0525108576f, -1.f),
    C_Vertex3f(0.413295627f, 0.0662373304f, -1.f),
    C_Vertex3f(0.423079729f, 0.0810415745f, -1.f),
    C_Vertex3f(0.431627989f, 0.0969235897f, -1.f),
    C_Vertex3f(0.438940406f, 0.113883376f, -1.f),
    C_Vertex3f(0.44501698f, 0.131920934f, -1.f),
    C_Vertex3f(0.449857712f, 0.151036263f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376528263f, 0.0282912254f, -1.f),
    C_Vertex3f(0.373535156f, 0.169921875f, -1.f),
    C_Vertex3f(0.372606277f, 0.162948608f, -1.f),
    C_Vertex3f(0.369949698f, 0.149778366f, -1.f),
    C_Vertex3f(0.366494298f, 0.137384415f, -1.f),
    C_Vertex3f(0.362428784f, 0.0181666613f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.362428784f, 0.0181666613f, -1.f),
    C_Vertex3f(0.366494298f, 0.137384415f, -1.f),
    C_Vertex3f(0.362240076f, 0.125766754f, -1.f),
    C_Vertex3f(0.357187033f, 0.114925385f, -1.f),
    C_Vertex3f(0.351335168f, 0.104860306f, -1.f),
    C_Vertex3f(0.347721457f, 0.0094884634f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.351335168f, 0.104860306f, -1.f),
    C_Vertex3f(0.344684482f, 0.0955715179f, -1.f),
    C_Vertex3f(0.347721457f, 0.0094884634f, -1.f),
    C_Vertex3f(0.337234974f, 0.087059021f, -1.f),
    C_Vertex3f(0.332406282f, 0.00225663185f, -1.f),
    C_Vertex3f(0.328986645f, 0.0793228149f, -1.f),
    C_Vertex3f(0.320349097f, 0.0725536346f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.316483259f, -0.00352883339f, -1.f),
    C_Vertex3f(0.332406282f, 0.00225663185f, -1.f),
    C_Vertex3f(0.320349097f, 0.0725536346f, -1.f),
    C_Vertex3f(0.311322331f, 0.0667514801f, -1.f),
    C_Vertex3f(0.301906347f, 0.0619163513f, -1.f),
    C_Vertex3f(0.299952388f, -0.00786793232f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.292101145f, 0.0580482483f, -1.f),
    C_Vertex3f(0.281906724f, 0.055147171f, -1.f),
    C_Vertex3f(0.282813668f, -0.0107606649f, -1.f),
    C_Vertex3f(0.299952388f, -0.00786793232f, -1.f),
    C_Vertex3f(0.301906347f, 0.0619163513f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.271323085f, 0.0532131195f, -1.f),
    C_Vertex3f(0.260350227f, 0.0522460938f, -1.f),
    C_Vertex3f(0.265067101f, -0.0122070313f, -1.f),
    C_Vertex3f(0.282813668f, -0.0107606649f, -1.f),
    C_Vertex3f(0.281906724f, 0.055147171f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.248550415f, 0.0522460938f, -1.f),
    C_Vertex3f(0.235221863f, 0.0542259216f, -1.f),
    C_Vertex3f(0.245185852f, -0.0122070313f, -1.f),
    C_Vertex3f(0.265067101f, -0.0122070313f, -1.f),
    C_Vertex3f(0.260350227f, 0.0522460938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.355379105f, 0.45290947f, -1.f),
    C_Vertex3f(0.34554863f, 0.367994308f, -1.f),
    C_Vertex3f(0.356050491f, 0.349328995f, -1.f),
    C_Vertex3f(0.159790039f, -0.178749084f, -1.f),
    C_Vertex3f(0.161392212f, 0.0089263916f, -1.f),
    C_Vertex3f(0.110839844f, -0.164756775f, -1.f),
    C_Vertex3f(0.345207214f, 0.64074707f, -1.f),
    C_Vertex3f(0.324981689f, 0.390205383f, -1.f),
    C_Vertex3f(0.332639694f, 0.383766174f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_163[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.172920227f, 0.353515625f, 0.f),
    C_Vertex3f(0.172920227f, 0.353515625f, -1.f),
    C_Vertex3f(0.310546875f, 0.353515625f, -1.f),
    C_Vertex3f(0.310546875f, 0.353515625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.310546875f, 0.353515625f, 0.f),
    C_Vertex3f(0.310546875f, 0.353515625f, -1.f),
    C_Vertex3f(0.310546875f, 0.2890625f, -1.f),
    C_Vertex3f(0.310546875f, 0.2890625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.310546875f, 0.2890625f, 0.f),
    C_Vertex3f(0.310546875f, 0.2890625f, -1.f),
    C_Vertex3f(0.186500549f, 0.2890625f, -1.f),
    C_Vertex3f(0.186500549f, 0.2890625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.917745829f, 0.39716807f, 0.f),
    C_Vertex3f(0.186500549f, 0.2890625f, 0.f),
    C_Vertex3f(0.186500549f, 0.2890625f, -1.f),
    C_Normal3f(0.940689087f, 0.339269996f, 0.f),
    C_Vertex3f(0.188581467f, 0.284254074f, 0.f),
    C_Vertex3f(0.188581467f, 0.284254074f, -1.f),
    C_Normal3f(0.965303063f, 0.261132061f, 0.f),
    C_Vertex3f(0.191702843f, 0.274638176f, 0.f),
    C_Vertex3f(0.191702843f, 0.274638176f, -1.f),
    C_Normal3f(0.98707968f, 0.160229713f, 0.f),
    C_Vertex3f(0.19378376f, 0.265023232f, 0.f),
    C_Vertex3f(0.19378376f, 0.265023232f, -1.f),
    C_Normal3f(0.999055922f, 0.043444138f, 0.f),
    C_Vertex3f(0.194824219f, 0.255409241f, 0.f),
    C_Vertex3f(0.194824219f, 0.255409241f, -1.f),
    C_Normal3f(0.997767925f, -0.0667772517f, 0.f),
    C_Vertex3f(0.194824219f, 0.241096497f, 0.f),
    C_Vertex3f(0.194824219f, 0.241096497f, -1.f),
    C_Normal3f(0.985616863f, -0.168995395f, 0.f),
    C_Vertex3f(0.192557335f, 0.221538067f, 0.f),
    C_Vertex3f(0.192557335f, 0.221538067f, -1.f),
    C_Normal3f(0.963434696f, -0.267943263f, 0.f),
    C_Vertex3f(0.188023567f, 0.201433659f, 0.f),
    C_Vertex3f(0.188023567f, 0.201433659f, -1.f),
    C_Normal3f(0.935032427f, -0.354562044f, 0.f),
    C_Vertex3f(0.181222916f, 0.180783272f, 0.f),
    C_Vertex3f(0.181222916f, 0.180783272f, -1.f),
    C_Normal3f(0.896502137f, -0.443039387f, 0.f),
    C_Vertex3f(0.17215538f, 0.159586906f, 0.f),
    C_Vertex3f(0.17215538f, 0.159586906f, -1.f),
    C_Normal3f(0.848013639f, -0.529974461f, 0.f),
    C_Vertex3f(0.160164833f, 0.138171673f, 0.f),
    C_Vertex3f(0.160164833f, 0.138171673f, -1.f),
    C_Normal3f(0.798809707f, -0.601583838f, 0.f),
    C_Vertex3f(0.145251274f, 0.116537571f, 0.f),
    C_Vertex3f(0.145251274f, 0.116537571f, -1.f),
    C_Normal3f(0.75896138f, -0.651135683f, 0.f),
    C_Vertex3f(0.127414703f, 0.0946846008f, 0.f),
    C_Vertex3f(0.127414703f, 0.0946846008f, -1.f),
    C_Normal3f(0.728429079f, -0.685121179f, 0.f),
    C_Vertex3f(0.117034912f, 0.0836486816f, 0.f),
    C_Vertex3f(0.117034912f, 0.0836486816f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.268192708f, 0.963365257f, 0.f),
    C_Vertex3f(0.117034912f, 0.0836486816f, 0.f),
    C_Vertex3f(0.117034912f, 0.0836486816f, -1.f),
    C_Normal3f(-0.232510209f, 0.972593904f, 0.f),
    C_Vertex3f(0.125667572f, 0.0860519409f, 0.f),
    C_Vertex3f(0.125667572f, 0.0860519409f, -1.f),
    C_Normal3f(-0.183240011f, 0.983068168f, 0.f),
    C_Vertex3f(0.142167091f, 0.0896568298f, 0.f),
    C_Vertex3f(0.142167091f, 0.0896568298f, -1.f),
    C_Normal3f(-0.116615705f, 0.993177056f, 0.f),
    C_Vertex3f(0.15790081f, 0.0920600891f, 0.f),
    C_Vertex3f(0.15790081f, 0.0920600891f, -1.f),
    C_Normal3f(-0.0385683142f, 0.999256015f, 0.f),
    C_Vertex3f(0.172868729f, 0.0932617188f, 0.f),
    C_Vertex3f(0.172868729f, 0.0932617188f, -1.f),
    C_Normal3f(0.0378093272f, 0.999284923f, 0.f),
    C_Vertex3f(0.189033508f, 0.0932617188f, 0.f),
    C_Vertex3f(0.189033508f, 0.0932617188f, -1.f),
    C_Normal3f(0.100824237f, 0.99490428f, 0.f),
    C_Vertex3f(0.208307266f, 0.0919208527f, 0.f),
    C_Vertex3f(0.208307266f, 0.0919208527f, -1.f),
    C_Normal3f(0.157680809f, 0.987490118f, 0.f),
    C_Vertex3f(0.228727341f, 0.0892391205f, 0.f),
    C_Vertex3f(0.228727341f, 0.0892391205f, -1.f),
    C_Normal3f(0.205912679f, 0.978570402f, 0.f),
    C_Vertex3f(0.250293732f, 0.0852165222f, 0.f),
    C_Vertex3f(0.250293732f, 0.0852165222f, -1.f),
    C_Normal3f(0.223828942f, 0.974628508f, 0.f),
    C_Vertex3f(0.271039963f, 0.0803356171f, 0.f),
    C_Vertex3f(0.271039963f, 0.0803356171f, -1.f),
    C_Normal3f(0.211586803f, 0.977359116f, 0.f),
    C_Vertex3f(0.288304806f, 0.0764870644f, 0.f),
    C_Vertex3f(0.288304806f, 0.0764870644f, -1.f),
    C_Normal3f(0.197710708f, 0.980260372f, 0.f),
    C_Vertex3f(0.304054737f, 0.0731883049f, 0.f),
    C_Vertex3f(0.304054737f, 0.0731883049f, -1.f),
    C_Normal3f(0.180552587f, 0.983565331f, 0.f),
    C_Vertex3f(0.318289757f, 0.0704393387f, 0.f),
    C_Vertex3f(0.318289757f, 0.0704393387f, -1.f),
    C_Normal3f(0.158815458f, 0.987308323f, 0.f),
    C_Vertex3f(0.331009865f, 0.0682401657f, 0.f),
    C_Vertex3f(0.331009865f, 0.0682401657f, -1.f),
    C_Normal3f(0.130434021f, 0.991456985f, 0.f),
    C_Vertex3f(0.342215061f, 0.066590786f, 0.f),
    C_Vertex3f(0.342215061f, 0.066590786f, -1.f),
    C_Normal3f(0.0919303f, 0.995765388f, 0.f),
    C_Vertex3f(0.351905346f, 0.0654911995f, 0.f),
    C_Vertex3f(0.351905346f, 0.0654911995f, -1.f),
    C_Normal3f(0.0280609149f, 0.999606192f, 0.f),
    C_Vertex3f(0.360080719f, 0.0649414063f, 0.f),
    C_Vertex3f(0.360080719f, 0.0649414063f, -1.f),
    C_Normal3f(-0.0992936864f, 0.995058239f, 0.f),
    C_Vertex3f(0.371490479f, 0.0649414063f, 0.f),
    C_Vertex3f(0.371490479f, 0.0649414063f, -1.f),
    C_Normal3f(-0.221367851f, 0.975190341f, 0.f),
    C_Vertex3f(0.390051842f, 0.0679321289f, 0.f),
    C_Vertex3f(0.390051842f, 0.0679321289f, -1.f),
    C_Normal3f(-0.319631994f, 0.947541714f, 0.f),
    C_Vertex3f(0.411015511f, 0.0739135742f, 0.f),
    C_Vertex3f(0.411015511f, 0.0739135742f, -1.f),
    C_Normal3f(-0.381340533f, 0.924434602f, 0.f),
    C_Vertex3f(0.434381485f, 0.0828857422f, 0.f),
    C_Vertex3f(0.434381485f, 0.0828857422f, -1.f),
    C_Normal3f(-0.421083689f, 0.907021761f, 0.f),
    C_Vertex3f(0.447265625f, 0.0888671875f, 0.f),
    C_Vertex3f(0.447265625f, 0.0888671875f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.943460166f, 0.331485987f, 0.f),
    C_Vertex3f(0.447265625f, 0.0888671875f, 0.f),
    C_Vertex3f(0.447265625f, 0.0888671875f, -1.f),
    C_Vertex3f(0.47265625f, 0.0166015625f, -1.f),
    C_Vertex3f(0.47265625f, 0.0166015625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.388064831f, -0.921632111f, 0.f),
    C_Vertex3f(0.47265625f, 0.0166015625f, 0.f),
    C_Vertex3f(0.47265625f, 0.0166015625f, -1.f),
    C_Normal3f(0.373590946f, -0.927593589f, 0.f),
    C_Vertex3f(0.462604523f, 0.0123691559f, 0.f),
    C_Vertex3f(0.462604523f, 0.0123691559f, -1.f),
    C_Normal3f(0.351291865f, -0.936266005f, 0.f),
    C_Vertex3f(0.444768429f, 0.00536966324f, 0.f),
    C_Vertex3f(0.444768429f, 0.00536966324f, -1.f),
    C_Normal3f(0.315625876f, -0.948883712f, 0.f),
    C_Vertex3f(0.429199696f, -0.00016450882f, 0.f),
    C_Vertex3f(0.429199696f, -0.00016450882f, -1.f),
    C_Normal3f(0.258070827f, -0.966125965f, 0.f),
    C_Vertex3f(0.415898323f, -0.00423336029f, 0.f),
    C_Vertex3f(0.415898323f, -0.00423336029f, -1.f),
    C_Normal3f(0.166635543f, -0.986018479f, 0.f),
    C_Vertex3f(0.399347305f, -0.00813865662f, 0.f),
    C_Vertex3f(0.399347305f, -0.00813865662f, -1.f),
    C_Normal3f(0.0617725365f, -0.998090327f, 0.f),
    C_Vertex3f(0.377384186f, -0.0107421875f, 0.f),
    C_Vertex3f(0.377384186f, -0.0107421875f, -1.f),
    C_Normal3f(-0.045425538f, -0.998967648f, 0.f),
    C_Vertex3f(0.357280731f, -0.0107421875f, 0.f),
    C_Vertex3f(0.357280731f, -0.0107421875f, -1.f),
    C_Normal3f(-0.127738908f, -0.991807818f, 0.f),
    C_Vertex3f(0.338710785f, -0.00898361206f, 0.f),
    C_Vertex3f(0.338710785f, -0.00898361206f, -1.f),
    C_Normal3f(-0.194589794f, -0.980884671f, 0.f),
    C_Vertex3f(0.327839851f, -0.00695037842f, 0.f),
    C_Vertex3f(0.327839851f, -0.00695037842f, -1.f),
    C_Normal3f(-0.236619696f, -0.971602321f, 0.f),
    C_Vertex3f(0.323858142f, -0.00603711605f, 0.f),
    C_Vertex3f(0.323858142f, -0.00603711605f, -1.f),
    C_Normal3f(-0.255208522f, -0.966885984f, 0.f),
    C_Vertex3f(0.318845391f, -0.00475990772f, 0.f),
    C_Vertex3f(0.318845391f, -0.00475990772f, -1.f),
    C_Normal3f(-0.267793328f, -0.96347636f, 0.f),
    C_Vertex3f(0.3128016f, -0.00311875343f, 0.f),
    C_Vertex3f(0.3128016f, -0.00311875343f, -1.f),
    C_Normal3f(-0.276874095f, -0.960906267f, 0.f),
    C_Vertex3f(0.305726767f, -0.00111365318f, 0.f),
    C_Vertex3f(0.305726767f, -0.00111365318f, -1.f),
    C_Normal3f(-0.283733696f, -0.958903074f, 0.f),
    C_Vertex3f(0.297620893f, 0.00125539303f, 0.f),
    C_Vertex3f(0.297620893f, 0.00125539303f, -1.f),
    C_Normal3f(-0.289097518f, -0.95729965f, 0.f),
    C_Vertex3f(0.288483977f, 0.0039883852f, 0.f),
    C_Vertex3f(0.288483977f, 0.0039883852f, -1.f),
    C_Normal3f(-0.293406278f, -0.955987871f, 0.f),
    C_Vertex3f(0.278316021f, 0.00708532333f, 0.f),
    C_Vertex3f(0.278316021f, 0.00708532333f, -1.f),
    C_Normal3f(-0.293450236f, -0.95597434f, 0.f),
    C_Vertex3f(0.267117023f, 0.0105462074f, 0.f),
    C_Vertex3f(0.267117023f, 0.0105462074f, -1.f),
    C_Normal3f(-0.28927058f, -0.957247376f, 0.f),
    C_Vertex3f(0.256908298f, 0.0136567354f, 0.f),
    C_Vertex3f(0.256908298f, 0.0136567354f, -1.f),
    C_Normal3f(-0.284119338f, -0.958788872f, 0.f),
    C_Vertex3f(0.247689843f, 0.0164169073f, 0.f),
    C_Vertex3f(0.247689843f, 0.0164169073f, -1.f),
    C_Normal3f(-0.277613759f, -0.960692763f, 0.f),
    C_Vertex3f(0.23946166f, 0.0188267231f, 0.f),
    C_Vertex3f(0.23946166f, 0.0188267231f, -1.f),
    C_Normal3f(-0.269140422f, -0.96310097f, 0.f),
    C_Vertex3f(0.232223749f, 0.0208861828f, 0.f),
    C_Vertex3f(0.232223749f, 0.0208861828f, -1.f),
    C_Normal3f(-0.257650793f, -0.966238081f, 0.f),
    C_Vertex3f(0.22597611f, 0.0225952864f, 0.f),
    C_Vertex3f(0.22597611f, 0.0225952864f, -1.f),
    C_Normal3f(-0.241194978f, -0.970476687f, 0.f),
    C_Vertex3f(0.220718741f, 0.0239540339f, 0.f),
    C_Vertex3f(0.220718741f, 0.0239540339f, -1.f),
    C_Normal3f(-0.220874771f, -0.97530216f, 0.f),
    C_Vertex3f(0.216451645f, 0.0249624252f, 0.f),
    C_Vertex3f(0.216451645f, 0.0249624252f, -1.f),
    C_Normal3f(-0.172628999f, -0.984986842f, 0.f),
    C_Vertex3f(0.214813232f, 0.0252914429f, 0.f),
    C_Vertex3f(0.214813232f, 0.0252914429f, -1.f),
    C_Normal3f(-0.113400906f, -0.993549347f, 0.f),
    C_Vertex3f(0.204540253f, 0.0270500183f, 0.f),
    C_Vertex3f(0.204540253f, 0.0270500183f, -1.f),
    C_Normal3f(-0.0454210714f, -0.998968005f, 0.f),
    C_Vertex3f(0.183998108f, 0.0288085938f, 0.f),
    C_Vertex3f(0.183998108f, 0.0288085938f, -1.f),
    C_Normal3f(0.0400438458f, -0.999197841f, 0.f),
    C_Vertex3f(0.165863037f, 0.0288085938f, 0.f),
    C_Vertex3f(0.165863037f, 0.0288085938f, -1.f),
    C_Normal3f(0.124226958f, -0.99225384f, 0.f),
    C_Vertex3f(0.14961195f, 0.0274305344f, 0.f),
    C_Vertex3f(0.14961195f, 0.0274305344f, -1.f),
    C_Normal3f(0.198282138f, -0.980144978f, 0.f),
    C_Vertex3f(0.132841587f, 0.0246744156f, 0.f),
    C_Vertex3f(0.132841587f, 0.0246744156f, -1.f),
    C_Normal3f(0.265011251f, -0.964245319f, 0.f),
    C_Vertex3f(0.115551949f, 0.0205402374f, 0.f),
    C_Vertex3f(0.115551949f, 0.0205402374f, -1.f),
    C_Normal3f(0.326193333f, -0.945303142f, 0.f),
    C_Vertex3f(0.0977430344f, 0.0150279999f, 0.f),
    C_Vertex3f(0.0977430344f, 0.0150279999f, -1.f),
    C_Normal3f(0.388204753f, -0.921573162f, 0.f),
    C_Vertex3f(0.0807261467f, 0.00852298737f, 0.f),
    C_Vertex3f(0.0807261467f, 0.00852298737f, -1.f),
    C_Normal3f(0.450807869f, -0.8926211f, 0.f),
    C_Vertex3f(0.0645012856f, 0.00102519989f, 0.f),
    C_Vertex3f(0.0645012856f, 0.00102519989f, -1.f),
    C_Normal3f(0.5027228f, -0.864447713f, 0.f),
    C_Vertex3f(0.0490684509f, -0.00746536255f, 0.f),
    C_Vertex3f(0.0490684509f, -0.00746536255f, -1.f),
    C_Normal3f(0.54365021f, -0.839311898f, 0.f),
    C_Vertex3f(0.0417480469f, -0.0122070313f, 0.f),
    C_Vertex3f(0.0417480469f, -0.0122070313f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.929506958f, -0.368804365f, 0.f),
    C_Vertex3f(0.0417480469f, -0.0122070313f, 0.f),
    C_Vertex3f(0.0417480469f, -0.0122070313f, -1.f),
    C_Vertex3f(0.01171875f, 0.0634765625f, -1.f),
    C_Vertex3f(0.01171875f, 0.0634765625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.48318252f, 0.875519693f, 0.f),
    C_Vertex3f(0.01171875f, 0.0634765625f, 0.f),
    C_Vertex3f(0.01171875f, 0.0634765625f, -1.f),
    C_Normal3f(-0.553335428f, 0.832958519f, 0.f),
    C_Vertex3f(0.0222148895f, 0.0692691803f, 0.f),
    C_Vertex3f(0.0222148895f, 0.0692691803f, -1.f),
    C_Normal3f(-0.635362923f, 0.772213697f, 0.f),
    C_Vertex3f(0.041621685f, 0.0833411217f, 0.f),
    C_Vertex3f(0.041621685f, 0.0833411217f, -1.f),
    C_Normal3f(-0.722630322f, 0.691234767f, 0.f),
    C_Vertex3f(0.059442997f, 0.0998997688f, 0.f),
    C_Vertex3f(0.059442997f, 0.0998997688f, -1.f),
    C_Normal3f(-0.790981352f, 0.611840248f, 0.f),
    C_Vertex3f(0.0756788254f, 0.118945122f, 0.f),
    C_Vertex3f(0.0756788254f, 0.118945122f, -1.f),
    C_Normal3f(-0.8451702f, 0.53449738f, 0.f),
    C_Vertex3f(0.086666584f, 0.135094166f, 0.f),
    C_Vertex3f(0.086666584f, 0.135094166f, -1.f),
    C_Normal3f(-0.890636861f, 0.454715312f, 0.f),
    C_Vertex3f(0.0930761099f, 0.146454453f, 0.f),
    C_Vertex3f(0.0930761099f, 0.146454453f, -1.f),
    C_Normal3f(-0.924910545f, 0.380184859f, 0.f),
    C_Vertex3f(0.0985699892f, 0.158408999f, 0.f),
    C_Vertex3f(0.0985699892f, 0.158408999f, -1.f),
    C_Normal3f(-0.952215016f, 0.305428624f, 0.f),
    C_Vertex3f(0.103148222f, 0.170957804f, 0.f),
    C_Vertex3f(0.103148222f, 0.170957804f, -1.f),
    C_Normal3f(-0.972729266f, 0.231943622f, 0.f),
    C_Vertex3f(0.106810808f, 0.184100866f, 0.f),
    C_Vertex3f(0.106810808f, 0.184100866f, -1.f),
    C_Normal3f(-0.986957788f, 0.160979673f, 0.f),
    C_Vertex3f(0.109557748f, 0.197838187f, 0.f),
    C_Vertex3f(0.109557748f, 0.197838187f, -1.f),
    C_Normal3f(-0.995621383f, 0.0934775546f, 0.f),
    C_Vertex3f(0.111389041f, 0.212169766f, 0.f),
    C_Vertex3f(0.111389041f, 0.212169766f, -1.f),
    C_Normal3f(-0.999677002f, 0.0254135262f, 0.f),
    C_Vertex3f(0.112304688f, 0.227095604f, 0.f),
    C_Vertex3f(0.112304688f, 0.227095604f, -1.f),
    C_Normal3f(-0.997011244f, -0.0772561878f, 0.f),
    C_Vertex3f(0.112304688f, 0.248188019f, 0.f),
    C_Vertex3f(0.112304688f, 0.248188019f, -1.f),
    C_Normal3f(-0.983710885f, -0.179757908f, 0.f),
    C_Vertex3f(0.108570099f, 0.275291443f, 0.f),
    C_Vertex3f(0.108570099f, 0.275291443f, -1.f),
    C_Normal3f(-0.965139151f, -0.261737198f, 0.f),
    C_Vertex3f(0.10483551f, 0.2890625f, 0.f),
    C_Vertex3f(0.10483551f, 0.2890625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.10483551f, 0.2890625f, 0.f),
    C_Vertex3f(0.10483551f, 0.2890625f, -1.f),
    C_Vertex3f(0.0190429688f, 0.2890625f, -1.f),
    C_Vertex3f(0.0190429688f, 0.2890625f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0190429688f, 0.2890625f, 0.f),
    C_Vertex3f(0.0190429688f, 0.2890625f, -1.f),
    C_Vertex3f(0.0190429688f, 0.353515625f, -1.f),
    C_Vertex3f(0.0190429688f, 0.353515625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0190429688f, 0.353515625f, 0.f),
    C_Vertex3f(0.0190429688f, 0.353515625f, -1.f),
    C_Vertex3f(0.0863800049f, 0.353515625f, -1.f),
    C_Vertex3f(0.0863800049f, 0.353515625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.958739221f, -0.284286886f, 0.f),
    C_Vertex3f(0.0863800049f, 0.353515625f, 0.f),
    C_Vertex3f(0.0863800049f, 0.353515625f, -1.f),
    C_Normal3f(-0.96135956f, -0.275296032f, 0.f),
    C_Vertex3f(0.0827541351f, 0.365743637f, 0.f),
    C_Vertex3f(0.0827541351f, 0.365743637f, -1.f),
    C_Normal3f(-0.965323448f, -0.261056751f, 0.f),
    C_Vertex3f(0.0768208504f, 0.386897087f, 0.f),
    C_Vertex3f(0.0768208504f, 0.386897087f, -1.f),
    C_Normal3f(-0.971458375f, -0.237209931f, 0.f),
    C_Vertex3f(0.0722060204f, 0.404747963f, 0.f),
    C_Vertex3f(0.0722060204f, 0.404747963f, -1.f),
    C_Normal3f(-0.980701387f, -0.195512146f, 0.f),
    C_Vertex3f(0.0689096451f, 0.419296265f, 0.f),
    C_Vertex3f(0.0689096451f, 0.419296265f, -1.f),
    C_Normal3f(-0.991982639f, -0.1263742f, 0.f),
    C_Vertex3f(0.0659427643f, 0.436164856f, 0.f),
    C_Vertex3f(0.0659427643f, 0.436164856f, -1.f),
    C_Normal3f(-0.998991251f, -0.0449057817f, 0.f),
    C_Vertex3f(0.0639648438f, 0.458110809f, 0.f),
    C_Vertex3f(0.0639648438f, 0.458110809f, -1.f),
    C_Normal3f(-0.998834133f, 0.0482745543f, 0.f),
    C_Vertex3f(0.0639648438f, 0.480166435f, 0.f),
    C_Vertex3f(0.0639648438f, 0.480166435f, -1.f),
    C_Normal3f(-0.988739133f, 0.149649113f, 0.f),
    C_Vertex3f(0.0660684109f, 0.501635075f, 0.f),
    C_Vertex3f(0.0660684109f, 0.501635075f, -1.f),
    C_Normal3f(-0.965855956f, 0.259079516f, 0.f),
    C_Vertex3f(0.0702755451f, 0.521861553f, 0.f),
    C_Vertex3f(0.0702755451f, 0.521861553f, -1.f),
    C_Normal3f(-0.928176761f, 0.372139573f, 0.f),
    C_Vertex3f(0.0765862465f, 0.540845871f, 0.f),
    C_Vertex3f(0.0765862465f, 0.540845871f, -1.f),
    C_Normal3f(-0.87514621f, 0.483858556f, 0.f),
    C_Vertex3f(0.085000515f, 0.558588028f, 0.f),
    C_Vertex3f(0.085000515f, 0.558588028f, -1.f),
    C_Normal3f(-0.808219612f, 0.588881075f, 0.f),
    C_Vertex3f(0.0955183506f, 0.575088024f, 0.f),
    C_Vertex3f(0.0955183506f, 0.575088024f, -1.f),
    C_Normal3f(-0.730750978f, 0.682644069f, 0.f),
    C_Vertex3f(0.108139753f, 0.59034586f, 0.f),
    C_Vertex3f(0.108139753f, 0.59034586f, -1.f),
    C_Normal3f(-0.648415804f, 0.761286378f, 0.f),
    C_Vertex3f(0.122864723f, 0.604361534f, 0.f),
    C_Vertex3f(0.122864723f, 0.604361534f, -1.f),
    C_Normal3f(-0.568449914f, 0.822717905f, 0.f),
    C_Vertex3f(0.138054848f, 0.615825653f, 0.f),
    C_Vertex3f(0.138054848f, 0.615825653f, -1.f),
    C_Normal3f(-0.491564304f, 0.870841265f, 0.f),
    C_Vertex3f(0.15231657f, 0.624711037f, 0.f),
    C_Vertex3f(0.15231657f, 0.624711037f, -1.f),
    C_Normal3f(-0.414524883f, 0.910037994f, 0.f),
    C_Vertex3f(0.167288303f, 0.63232708f, 0.f),
    C_Vertex3f(0.167288303f, 0.63232708f, -1.f),
    C_Normal3f(-0.335535139f, 0.942027688f, 0.f),
    C_Vertex3f(0.182970047f, 0.638673782f, 0.f),
    C_Vertex3f(0.182970047f, 0.638673782f, -1.f),
    C_Normal3f(-0.256417096f, 0.966566205f, 0.f),
    C_Vertex3f(0.199361801f, 0.643751144f, 0.f),
    C_Vertex3f(0.199361801f, 0.643751144f, -1.f),
    C_Normal3f(-0.17885232f, 0.983875871f, 0.f),
    C_Vertex3f(0.216463566f, 0.647559166f, 0.f),
    C_Vertex3f(0.216463566f, 0.647559166f, -1.f),
    C_Normal3f(-0.104236349f, 0.994552553f, 0.f),
    C_Vertex3f(0.234275341f, 0.650097847f, 0.f),
    C_Vertex3f(0.234275341f, 0.650097847f, -1.f),
    C_Normal3f(-0.0333611481f, 0.999443352f, 0.f),
    C_Vertex3f(0.252797127f, 0.651367188f, 0.f),
    C_Vertex3f(0.252797127f, 0.651367188f, -1.f),
    C_Normal3f(0.0353273973f, 0.99937582f, 0.f),
    C_Vertex3f(0.272302628f, 0.651367188f, 0.f),
    C_Vertex3f(0.272302628f, 0.651367188f, -1.f),
    C_Normal3f(0.10957706f, 0.993978322f, 0.f),
    C_Vertex3f(0.291214466f, 0.650009155f, 0.f),
    C_Vertex3f(0.291214466f, 0.650009155f, -1.f),
    C_Normal3f(0.189298511f, 0.981919527f, 0.f),
    C_Vertex3f(0.309258938f, 0.647293091f, 0.f),
    C_Vertex3f(0.309258938f, 0.647293091f, -1.f),
    C_Normal3f(0.273088843f, 0.961988926f, 0.f),
    C_Vertex3f(0.326436043f, 0.643218994f, 0.f),
    C_Vertex3f(0.326436043f, 0.643218994f, -1.f),
    C_Normal3f(0.359233558f, 0.933247626f, 0.f),
    C_Vertex3f(0.342745781f, 0.637786865f, 0.f),
    C_Vertex3f(0.342745781f, 0.637786865f, -1.f),
    C_Normal3f(0.445534676f, 0.895264685f, 0.f),
    C_Vertex3f(0.358188152f, 0.630996704f, 0.f),
    C_Vertex3f(0.358188152f, 0.630996704f, -1.f),
    C_Normal3f(0.52951932f, 0.848297834f, 0.f),
    C_Vertex3f(0.372763157f, 0.622848511f, 0.f),
    C_Vertex3f(0.372763157f, 0.622848511f, -1.f),
    C_Normal3f(0.608753085f, 0.793359756f, 0.f),
    C_Vertex3f(0.386470795f, 0.613342285f, 0.f),
    C_Vertex3f(0.386470795f, 0.613342285f, -1.f),
    C_Normal3f(0.682516873f, 0.730869889f, 0.f),
    C_Vertex3f(0.399311066f, 0.602478027f, 0.f),
    C_Vertex3f(0.399311066f, 0.602478027f, -1.f),
    C_Normal3f(0.751450837f, 0.659789026f, 0.f),
    C_Vertex3f(0.410915136f, 0.590515137f, 0.f),
    C_Vertex3f(0.410915136f, 0.590515137f, -1.f),
    C_Normal3f(0.812949002f, 0.582334816f, 0.f),
    C_Vertex3f(0.421283007f, 0.577453613f, 0.f),
    C_Vertex3f(0.421283007f, 0.577453613f, -1.f),
    C_Normal3f(0.865488529f, 0.5009287f, 0.f),
    C_Vertex3f(0.430414677f, 0.563293457f, 0.f),
    C_Vertex3f(0.430414677f, 0.563293457f, -1.f),
    C_Normal3f(0.908367395f, 0.418173134f, 0.f),
    C_Vertex3f(0.438310146f, 0.548034668f, 0.f),
    C_Vertex3f(0.438310146f, 0.548034668f, -1.f),
    C_Normal3f(0.94168812f, 0.336486965f, 0.f),
    C_Vertex3f(0.444969416f, 0.531677246f, 0.f),
    C_Vertex3f(0.444969416f, 0.531677246f, -1.f),
    C_Normal3f(0.966187716f, 0.257839829f, 0.f),
    C_Vertex3f(0.450392485f, 0.514221191f, 0.f),
    C_Vertex3f(0.450392485f, 0.514221191f, -1.f),
    C_Normal3f(0.980674088f, 0.19564876f, 0.f),
    C_Vertex3f(0.454579353f, 0.495666504f, 0.f),
    C_Vertex3f(0.454579353f, 0.495666504f, -1.f),
    C_Normal3f(0.988916695f, 0.148471862f, 0.f),
    C_Vertex3f(0.456054688f, 0.485839844f, 0.f),
    C_Vertex3f(0.456054688f, 0.485839844f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.148340449f, -0.988936365f, 0.f),
    C_Vertex3f(0.456054688f, 0.485839844f, 0.f),
    C_Vertex3f(0.456054688f, 0.485839844f, -1.f),
    C_Vertex3f(0.377929688f, 0.474121094f, -1.f),
    C_Vertex3f(0.377929688f, 0.474121094f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99226898f, -0.124105781f, 0.f),
    C_Vertex3f(0.377929688f, 0.474121094f, 0.f),
    C_Vertex3f(0.377929688f, 0.474121094f, -1.f),
    C_Normal3f(-0.98544544f, -0.169992119f, 0.f),
    C_Vertex3f(0.377109528f, 0.480678558f, 0.f),
    C_Vertex3f(0.377109528f, 0.480678558f, -1.f),
    C_Normal3f(-0.973062515f, -0.230541334f, 0.f),
    C_Vertex3f(0.37465632f, 0.493096828f, 0.f),
    C_Vertex3f(0.37465632f, 0.493096828f, -1.f),
    C_Normal3f(-0.951617181f, -0.307286203f, 0.f),
    C_Vertex3f(0.371390224f, 0.504818439f, 0.f),
    C_Vertex3f(0.371390224f, 0.504818439f, -1.f),
    C_Normal3f(-0.921943665f, -0.387324005f, 0.f),
    C_Vertex3f(0.367311239f, 0.515843391f, 0.f),
    C_Vertex3f(0.367311239f, 0.515843391f, -1.f),
    C_Normal3f(-0.883245468f, -0.468910962f, 0.f),
    C_Vertex3f(0.362419367f, 0.526171684f, 0.f),
    C_Vertex3f(0.362419367f, 0.526171684f, -1.f),
    C_Normal3f(-0.835261405f, -0.549853086f, 0.f),
    C_Vertex3f(0.356714606f, 0.535803318f, 0.f),
    C_Vertex3f(0.356714606f, 0.535803318f, -1.f),
    C_Normal3f(-0.778440595f, -0.627718389f, 0.f),
    C_Vertex3f(0.350196958f, 0.544738293f, 0.f),
    C_Vertex3f(0.350196958f, 0.544738293f, -1.f),
    C_Normal3f(-0.713995874f, -0.700149834f, 0.f),
    C_Vertex3f(0.342866421f, 0.552976608f, 0.f),
    C_Vertex3f(0.342866421f, 0.552976608f, -1.f),
    C_Normal3f(-0.646431863f, -0.76297164f, 0.f),
    C_Vertex3f(0.334722996f, 0.560518265f, 0.f),
    C_Vertex3f(0.334722996f, 0.560518265f, -1.f),
    C_Normal3f(-0.57371366f, -0.819055974f, 0.f),
    C_Vertex3f(0.326176524f, 0.567117214f, 0.f),
    C_Vertex3f(0.326176524f, 0.567117214f, -1.f),
    C_Normal3f(-0.492962122f, -0.870050788f, 0.f),
    C_Vertex3f(0.317227006f, 0.572773457f, 0.f),
    C_Vertex3f(0.317227006f, 0.572773457f, -1.f),
    C_Normal3f(-0.405812532f, -0.913956285f, 0.f),
    C_Vertex3f(0.307874441f, 0.577486992f, 0.f),
    C_Vertex3f(0.307874441f, 0.577486992f, -1.f),
    C_Normal3f(-0.314548165f, -0.949241579f, 0.f),
    C_Vertex3f(0.29811883f, 0.58125782f, 0.f),
    C_Vertex3f(0.29811883f, 0.58125782f, -1.f),
    C_Normal3f(-0.221816286f, -0.975088418f, 0.f),
    C_Vertex3f(0.287960172f, 0.584085941f, 0.f),
    C_Vertex3f(0.287960172f, 0.584085941f, -1.f),
    C_Normal3f(-0.130259499f, -0.991479933f, 0.f),
    C_Vertex3f(0.277398467f, 0.585971355f, 0.f),
    C_Vertex3f(0.277398467f, 0.585971355f, -1.f),
    C_Normal3f(-0.0414658822f, -0.999140024f, 0.f),
    C_Vertex3f(0.266433716f, 0.586914063f, 0.f),
    C_Vertex3f(0.266433716f, 0.586914063f, -1.f),
    C_Normal3f(0.0417496786f, -0.999128103f, 0.f),
    C_Vertex3f(0.254683495f, 0.586914063f, 0.f),
    C_Vertex3f(0.254683495f, 0.586914063f, -1.f),
    C_Normal3f(0.127843261f, -0.991794407f, 0.f),
    C_Vertex3f(0.243063211f, 0.5859375f, 0.f),
    C_Vertex3f(0.243063211f, 0.5859375f, -1.f),
    C_Normal3f(0.219492033f, -0.97561425f, 0.f),
    C_Vertex3f(0.23195529f, 0.583984375f, 0.f),
    C_Vertex3f(0.23195529f, 0.583984375f, -1.f),
    C_Normal3f(0.313872099f, -0.949465275f, 0.f),
    C_Vertex3f(0.22135973f, 0.581054688f, 0.f),
    C_Vertex3f(0.22135973f, 0.581054688f, -1.f),
    C_Normal3f(0.408229351f, -0.912879407f, 0.f),
    C_Vertex3f(0.211276531f, 0.577148438f, 0.f),
    C_Vertex3f(0.211276531f, 0.577148438f, -1.f),
    C_Normal3f(0.499530822f, -0.866296172f, 0.f),
    C_Vertex3f(0.201705694f, 0.572265625f, 0.f),
    C_Vertex3f(0.201705694f, 0.572265625f, -1.f),
    C_Normal3f(0.584911644f, -0.811096966f, 0.f),
    C_Vertex3f(0.192647219f, 0.56640625f, 0.f),
    C_Vertex3f(0.192647219f, 0.56640625f, -1.f),
    C_Normal3f(0.662108064f, -0.749408484f, 0.f),
    C_Vertex3f(0.184101105f, 0.559570313f, 0.f),
    C_Vertex3f(0.184101105f, 0.559570313f, -1.f),
    C_Normal3f(0.728694975f, -0.684838474f, 0.f),
    C_Vertex3f(0.176067352f, 0.551757813f, 0.f),
    C_Vertex3f(0.176067352f, 0.551757813f, -1.f),
    C_Normal3f(0.790218055f, -0.612825811f, 0.f),
    C_Vertex3f(0.169037819f, 0.543542385f, 0.f),
    C_Vertex3f(0.169037819f, 0.543542385f, -1.f),
    C_Normal3f(0.847530782f, -0.530746281f, 0.f),
    C_Vertex3f(0.163012505f, 0.53492403f, 0.f),
    C_Vertex3f(0.163012505f, 0.53492403f, -1.f),
    C_Normal3f(0.897996426f, -0.44000268f, 0.f),
    C_Vertex3f(0.157991409f, 0.525902748f, 0.f),
    C_Vertex3f(0.157991409f, 0.525902748f, -1.f),
    C_Normal3f(0.939337492f, -0.342994392f, 0.f),
    C_Vertex3f(0.153974533f, 0.516478539f, 0.f),
    C_Vertex3f(0.153974533f, 0.516478539f, -1.f),
    C_Normal3f(0.970065117f, -0.242844924f, 0.f),
    C_Vertex3f(0.150961876f, 0.506651402f, 0.f),
    C_Vertex3f(0.150961876f, 0.506651402f, -1.f),
    C_Normal3f(0.989734352f, -0.142919153f, 0.f),
    C_Vertex3f(0.148953438f, 0.496421337f, 0.f),
    C_Vertex3f(0.148953438f, 0.496421337f, -1.f),
    C_Normal3f(0.998979867f, -0.0451576896f, 0.f),
    C_Vertex3f(0.147949219f, 0.485788345f, 0.f),
    C_Vertex3f(0.147949219f, 0.485788345f, -1.f),
    C_Normal3f(0.999498904f, 0.0316538289f, 0.f),
    C_Vertex3f(0.147949219f, 0.474205971f, 0.f),
    C_Vertex3f(0.147949219f, 0.474205971f, -1.f),
    C_Normal3f(0.996274352f, 0.0862403661f, 0.f),
    C_Vertex3f(0.148729563f, 0.461148262f, 0.f),
    C_Vertex3f(0.148729563f, 0.461148262f, -1.f),
    C_Normal3f(0.991010427f, 0.133784443f, 0.f),
    C_Vertex3f(0.150290251f, 0.447161674f, 0.f),
    C_Vertex3f(0.150290251f, 0.447161674f, -1.f),
    C_Normal3f(0.984595597f, 0.174847066f, 0.f),
    C_Vertex3f(0.152631283f, 0.432246208f, 0.f),
    C_Vertex3f(0.152631283f, 0.432246208f, -1.f),
    C_Normal3f(0.977595508f, 0.210492298f, 0.f),
    C_Vertex3f(0.155752659f, 0.416401863f, 0.f),
    C_Vertex3f(0.155752659f, 0.416401863f, -1.f),
    C_Normal3f(0.970374167f, 0.241607055f, 0.f),
    C_Vertex3f(0.159654379f, 0.399628639f, 0.f),
    C_Vertex3f(0.159654379f, 0.399628639f, -1.f),
    C_Normal3f(0.963162005f, 0.26892212f, 0.f),
    C_Vertex3f(0.164336443f, 0.381926537f, 0.f),
    C_Vertex3f(0.164336443f, 0.381926537f, -1.f),
    C_Normal3f(0.95726341f, 0.28921783f, 0.f),
    C_Vertex3f(0.169798851f, 0.363295555f, 0.f),
    C_Vertex3f(0.169798851f, 0.363295555f, -1.f),
    C_Normal3f(0.952655792f, 0.304050952f, 0.f),
    C_Vertex3f(0.172920227f, 0.353515625f, 0.f),
    C_Vertex3f(0.172920227f, 0.353515625f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0768208504f, 0.386897087f, 0.f),
    C_Vertex3f(0.0827541351f, 0.365743637f, 0.f),
    C_Vertex3f(0.085000515f, 0.558588028f, 0.f),
    C_Vertex3f(0.0863800049f, 0.353515625f, 0.f),
    C_Vertex3f(0.0955183506f, 0.575088024f, 0.f),
    C_Vertex3f(0.10483551f, 0.2890625f, 0.f),
    C_Vertex3f(0.108139753f, 0.59034586f, 0.f),
    C_Vertex3f(0.108570099f, 0.275291443f, 0.f),
    C_Vertex3f(0.122864723f, 0.604361534f, 0.f),
    C_Vertex3f(0.112304688f, 0.248188019f, 0.f),
    C_Vertex3f(0.115551949f, 0.0205402374f, 0.f),
    C_Vertex3f(0.112304688f, 0.227095604f, 0.f),
    C_Vertex3f(0.111389041f, 0.212169766f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0639648438f, 0.480166435f, 0.f),
    C_Vertex3f(0.0659427643f, 0.436164856f, 0.f),
    C_Vertex3f(0.0660684109f, 0.501635075f, 0.f),
    C_Vertex3f(0.0689096451f, 0.419296265f, 0.f),
    C_Vertex3f(0.0702755451f, 0.521861553f, 0.f),
    C_Vertex3f(0.0722060204f, 0.404747963f, 0.f),
    C_Vertex3f(0.0765862465f, 0.540845871f, 0.f),
    C_Vertex3f(0.0768208504f, 0.386897087f, 0.f),
    C_Vertex3f(0.085000515f, 0.558588028f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.456054688f, 0.485839844f, 0.f),
    C_Vertex3f(0.454579353f, 0.495666504f, 0.f),
    C_Vertex3f(0.450392485f, 0.514221191f, 0.f),
    C_Vertex3f(0.444969416f, 0.531677246f, 0.f),
    C_Vertex3f(0.438310146f, 0.548034668f, 0.f),
    C_Vertex3f(0.430414677f, 0.563293457f, 0.f),
    C_Vertex3f(0.421283007f, 0.577453613f, 0.f),
    C_Vertex3f(0.410915136f, 0.590515137f, 0.f),
    C_Vertex3f(0.399311066f, 0.602478027f, 0.f),
    C_Vertex3f(0.386470795f, 0.613342285f, 0.f),
    C_Vertex3f(0.377929688f, 0.474121094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.367311239f, 0.515843391f, 0.f),
    C_Vertex3f(0.371390224f, 0.504818439f, 0.f),
    C_Vertex3f(0.372763157f, 0.622848511f, 0.f),
    C_Vertex3f(0.37465632f, 0.493096828f, 0.f),
    C_Vertex3f(0.386470795f, 0.613342285f, 0.f),
    C_Vertex3f(0.377109528f, 0.480678558f, 0.f),
    C_Vertex3f(0.377929688f, 0.474121094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.350196958f, 0.544738293f, 0.f),
    C_Vertex3f(0.356714606f, 0.535803318f, 0.f),
    C_Vertex3f(0.358188152f, 0.630996704f, 0.f),
    C_Vertex3f(0.362419367f, 0.526171684f, 0.f),
    C_Vertex3f(0.372763157f, 0.622848511f, 0.f),
    C_Vertex3f(0.367311239f, 0.515843391f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.317227006f, 0.572773457f, 0.f),
    C_Vertex3f(0.326176524f, 0.567117214f, 0.f),
    C_Vertex3f(0.326436043f, 0.643218994f, 0.f),
    C_Vertex3f(0.334722996f, 0.560518265f, 0.f),
    C_Vertex3f(0.342745781f, 0.637786865f, 0.f),
    C_Vertex3f(0.342866421f, 0.552976608f, 0.f),
    C_Vertex3f(0.358188152f, 0.630996704f, 0.f),
    C_Vertex3f(0.350196958f, 0.544738293f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.309258938f, 0.647293091f, 0.f),
    C_Vertex3f(0.291214466f, 0.650009155f, 0.f),
    C_Vertex3f(0.29811883f, 0.58125782f, 0.f),
    C_Vertex3f(0.307874441f, 0.577486992f, 0.f),
    C_Vertex3f(0.317227006f, 0.572773457f, 0.f),
    C_Vertex3f(0.326436043f, 0.643218994f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.254683495f, 0.586914063f, 0.f),
    C_Vertex3f(0.266433716f, 0.586914063f, 0.f),
    C_Vertex3f(0.272302628f, 0.651367188f, 0.f),
    C_Vertex3f(0.277398467f, 0.585971355f, 0.f),
    C_Vertex3f(0.291214466f, 0.650009155f, 0.f),
    C_Vertex3f(0.287960172f, 0.584085941f, 0.f),
    C_Vertex3f(0.29811883f, 0.58125782f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.22135973f, 0.581054688f, 0.f),
    C_Vertex3f(0.23195529f, 0.583984375f, 0.f),
    C_Vertex3f(0.234275341f, 0.650097847f, 0.f),
    C_Vertex3f(0.243063211f, 0.5859375f, 0.f),
    C_Vertex3f(0.252797127f, 0.651367188f, 0.f),
    C_Vertex3f(0.254683495f, 0.586914063f, 0.f),
    C_Vertex3f(0.272302628f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.216463566f, 0.647559166f, 0.f),
    C_Vertex3f(0.199361801f, 0.643751144f, 0.f),
    C_Vertex3f(0.201705694f, 0.572265625f, 0.f),
    C_Vertex3f(0.211276531f, 0.577148438f, 0.f),
    C_Vertex3f(0.22135973f, 0.581054688f, 0.f),
    C_Vertex3f(0.234275341f, 0.650097847f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.169037819f, 0.543542385f, 0.f),
    C_Vertex3f(0.176067352f, 0.551757813f, 0.f),
    C_Vertex3f(0.182970047f, 0.638673782f, 0.f),
    C_Vertex3f(0.184101105f, 0.559570313f, 0.f),
    C_Vertex3f(0.199361801f, 0.643751144f, 0.f),
    C_Vertex3f(0.192647219f, 0.56640625f, 0.f),
    C_Vertex3f(0.201705694f, 0.572265625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.167288303f, 0.63232708f, 0.f),
    C_Vertex3f(0.15231657f, 0.624711037f, 0.f),
    C_Vertex3f(0.153974533f, 0.516478539f, 0.f),
    C_Vertex3f(0.157991409f, 0.525902748f, 0.f),
    C_Vertex3f(0.163012505f, 0.53492403f, 0.f),
    C_Vertex3f(0.169037819f, 0.543542385f, 0.f),
    C_Vertex3f(0.182970047f, 0.638673782f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15231657f, 0.624711037f, 0.f),
    C_Vertex3f(0.138054848f, 0.615825653f, 0.f),
    C_Vertex3f(0.147949219f, 0.485788345f, 0.f),
    C_Vertex3f(0.148953438f, 0.496421337f, 0.f),
    C_Vertex3f(0.150961876f, 0.506651402f, 0.f),
    C_Vertex3f(0.153974533f, 0.516478539f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147949219f, 0.474205971f, 0.f),
    C_Vertex3f(0.147949219f, 0.485788345f, 0.f),
    C_Vertex3f(0.145251274f, 0.116537571f, 0.f),
    C_Vertex3f(0.138054848f, 0.615825653f, 0.f),
    C_Vertex3f(0.127414703f, 0.0946846008f, 0.f),
    C_Vertex3f(0.122864723f, 0.604361534f, 0.f),
    C_Vertex3f(0.117034912f, 0.0836486816f, 0.f),
    C_Vertex3f(0.115551949f, 0.0205402374f, 0.f),
    C_Vertex3f(0.132841587f, 0.0246744156f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0190429688f, 0.353515625f, 0.f),
    C_Vertex3f(0.0190429688f, 0.2890625f, 0.f),
    C_Vertex3f(0.10483551f, 0.2890625f, 0.f),
    C_Vertex3f(0.0863800049f, 0.353515625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.186500549f, 0.2890625f, 0.f),
    C_Vertex3f(0.310546875f, 0.2890625f, 0.f),
    C_Vertex3f(0.310546875f, 0.353515625f, 0.f),
    C_Vertex3f(0.172920227f, 0.353515625f, 0.f),
    C_Vertex3f(0.181222916f, 0.180783272f, 0.f),
    C_Vertex3f(0.188023567f, 0.201433659f, 0.f),
    C_Vertex3f(0.188581467f, 0.284254074f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.17215538f, 0.159586906f, 0.f),
    C_Vertex3f(0.181222916f, 0.180783272f, 0.f),
    C_Vertex3f(0.172920227f, 0.353515625f, 0.f),
    C_Vertex3f(0.169798851f, 0.363295555f, 0.f),
    C_Vertex3f(0.164336443f, 0.381926537f, 0.f),
    C_Vertex3f(0.160164833f, 0.138171673f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.160164833f, 0.138171673f, 0.f),
    C_Vertex3f(0.164336443f, 0.381926537f, 0.f),
    C_Vertex3f(0.159654379f, 0.399628639f, 0.f),
    C_Vertex3f(0.155752659f, 0.416401863f, 0.f),
    C_Vertex3f(0.152631283f, 0.432246208f, 0.f),
    C_Vertex3f(0.150290251f, 0.447161674f, 0.f),
    C_Vertex3f(0.148729563f, 0.461148262f, 0.f),
    C_Vertex3f(0.147949219f, 0.474205971f, 0.f),
    C_Vertex3f(0.145251274f, 0.116537571f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.19378376f, 0.265023232f, 0.f),
    C_Vertex3f(0.191702843f, 0.274638176f, 0.f),
    C_Vertex3f(0.192557335f, 0.221538067f, 0.f),
    C_Vertex3f(0.194824219f, 0.241096497f, 0.f),
    C_Vertex3f(0.194824219f, 0.255409241f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.188581467f, 0.284254074f, 0.f),
    C_Vertex3f(0.188023567f, 0.201433659f, 0.f),
    C_Vertex3f(0.192557335f, 0.221538067f, 0.f),
    C_Vertex3f(0.191702843f, 0.274638176f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.415898323f, -0.00423336029f, 0.f),
    C_Vertex3f(0.429199696f, -0.00016450882f, 0.f),
    C_Vertex3f(0.434381485f, 0.0828857422f, 0.f),
    C_Vertex3f(0.444768429f, 0.00536966324f, 0.f),
    C_Vertex3f(0.447265625f, 0.0888671875f, 0.f),
    C_Vertex3f(0.462604523f, 0.0123691559f, 0.f),
    C_Vertex3f(0.47265625f, 0.0166015625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.371490479f, 0.0649414063f, 0.f),
    C_Vertex3f(0.377384186f, -0.0107421875f, 0.f),
    C_Vertex3f(0.390051842f, 0.0679321289f, 0.f),
    C_Vertex3f(0.399347305f, -0.00813865662f, 0.f),
    C_Vertex3f(0.411015511f, 0.0739135742f, 0.f),
    C_Vertex3f(0.415898323f, -0.00423336029f, 0.f),
    C_Vertex3f(0.434381485f, 0.0828857422f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360080719f, 0.0649414063f, 0.f),
    C_Vertex3f(0.351905346f, 0.0654911995f, 0.f),
    C_Vertex3f(0.357280731f, -0.0107421875f, 0.f),
    C_Vertex3f(0.377384186f, -0.0107421875f, 0.f),
    C_Vertex3f(0.371490479f, 0.0649414063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323858142f, -0.00603711605f, 0.f),
    C_Vertex3f(0.327839851f, -0.00695037842f, 0.f),
    C_Vertex3f(0.331009865f, 0.0682401657f, 0.f),
    C_Vertex3f(0.338710785f, -0.00898361206f, 0.f),
    C_Vertex3f(0.342215061f, 0.066590786f, 0.f),
    C_Vertex3f(0.357280731f, -0.0107421875f, 0.f),
    C_Vertex3f(0.351905346f, 0.0654911995f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.305726767f, -0.00111365318f, 0.f),
    C_Vertex3f(0.3128016f, -0.00311875343f, 0.f),
    C_Vertex3f(0.318289757f, 0.0704393387f, 0.f),
    C_Vertex3f(0.318845391f, -0.00475990772f, 0.f),
    C_Vertex3f(0.331009865f, 0.0682401657f, 0.f),
    C_Vertex3f(0.323858142f, -0.00603711605f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304054737f, 0.0731883049f, 0.f),
    C_Vertex3f(0.288304806f, 0.0764870644f, 0.f),
    C_Vertex3f(0.288483977f, 0.0039883852f, 0.f),
    C_Vertex3f(0.297620893f, 0.00125539303f, 0.f),
    C_Vertex3f(0.305726767f, -0.00111365318f, 0.f),
    C_Vertex3f(0.318289757f, 0.0704393387f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.256908298f, 0.0136567354f, 0.f),
    C_Vertex3f(0.267117023f, 0.0105462074f, 0.f),
    C_Vertex3f(0.271039963f, 0.0803356171f, 0.f),
    C_Vertex3f(0.278316021f, 0.00708532333f, 0.f),
    C_Vertex3f(0.288304806f, 0.0764870644f, 0.f),
    C_Vertex3f(0.288483977f, 0.0039883852f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.250293732f, 0.0852165222f, 0.f),
    C_Vertex3f(0.228727341f, 0.0892391205f, 0.f),
    C_Vertex3f(0.232223749f, 0.0208861828f, 0.f),
    C_Vertex3f(0.23946166f, 0.0188267231f, 0.f),
    C_Vertex3f(0.247689843f, 0.0164169073f, 0.f),
    C_Vertex3f(0.256908298f, 0.0136567354f, 0.f),
    C_Vertex3f(0.271039963f, 0.0803356171f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.228727341f, 0.0892391205f, 0.f),
    C_Vertex3f(0.208307266f, 0.0919208527f, 0.f),
    C_Vertex3f(0.214813232f, 0.0252914429f, 0.f),
    C_Vertex3f(0.216451645f, 0.0249624252f, 0.f),
    C_Vertex3f(0.220718741f, 0.0239540339f, 0.f),
    C_Vertex3f(0.22597611f, 0.0225952864f, 0.f),
    C_Vertex3f(0.232223749f, 0.0208861828f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.214813232f, 0.0252914429f, 0.f),
    C_Vertex3f(0.208307266f, 0.0919208527f, 0.f),
    C_Vertex3f(0.204540253f, 0.0270500183f, 0.f),
    C_Vertex3f(0.189033508f, 0.0932617188f, 0.f),
    C_Vertex3f(0.183998108f, 0.0288085938f, 0.f),
    C_Vertex3f(0.172868729f, 0.0932617188f, 0.f),
    C_Vertex3f(0.165863037f, 0.0288085938f, 0.f),
    C_Vertex3f(0.15790081f, 0.0920600891f, 0.f),
    C_Vertex3f(0.14961195f, 0.0274305344f, 0.f),
    C_Vertex3f(0.142167091f, 0.0896568298f, 0.f),
    C_Vertex3f(0.132841587f, 0.0246744156f, 0.f),
    C_Vertex3f(0.125667572f, 0.0860519409f, 0.f),
    C_Vertex3f(0.117034912f, 0.0836486816f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115551949f, 0.0205402374f, 0.f),
    C_Vertex3f(0.111389041f, 0.212169766f, 0.f),
    C_Vertex3f(0.109557748f, 0.197838187f, 0.f),
    C_Vertex3f(0.106810808f, 0.184100866f, 0.f),
    C_Vertex3f(0.103148222f, 0.170957804f, 0.f),
    C_Vertex3f(0.0985699892f, 0.158408999f, 0.f),
    C_Vertex3f(0.0977430344f, 0.0150279999f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0417480469f, -0.0122070313f, 0.f),
    C_Vertex3f(0.0490684509f, -0.00746536255f, 0.f),
    C_Vertex3f(0.059442997f, 0.0998997688f, 0.f),
    C_Vertex3f(0.0645012856f, 0.00102519989f, 0.f),
    C_Vertex3f(0.0756788254f, 0.118945122f, 0.f),
    C_Vertex3f(0.0807261467f, 0.00852298737f, 0.f),
    C_Vertex3f(0.086666584f, 0.135094166f, 0.f),
    C_Vertex3f(0.0977430344f, 0.0150279999f, 0.f),
    C_Vertex3f(0.0930761099f, 0.146454453f, 0.f),
    C_Vertex3f(0.0985699892f, 0.158408999f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0417480469f, -0.0122070313f, 0.f),
    C_Vertex3f(0.059442997f, 0.0998997688f, 0.f),
    C_Vertex3f(0.041621685f, 0.0833411217f, 0.f),
    C_Vertex3f(0.0222148895f, 0.0692691803f, 0.f),
    C_Vertex3f(0.01171875f, 0.0634765625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0659427643f, 0.436164856f, 0.f),
    C_Vertex3f(0.0639648438f, 0.480166435f, 0.f),
    C_Vertex3f(0.0639648438f, 0.458110809f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0768208504f, 0.386897087f, -1.f),
    C_Vertex3f(0.0827541351f, 0.365743637f, -1.f),
    C_Vertex3f(0.085000515f, 0.558588028f, -1.f),
    C_Vertex3f(0.0863800049f, 0.353515625f, -1.f),
    C_Vertex3f(0.0955183506f, 0.575088024f, -1.f),
    C_Vertex3f(0.10483551f, 0.2890625f, -1.f),
    C_Vertex3f(0.108139753f, 0.59034586f, -1.f),
    C_Vertex3f(0.108570099f, 0.275291443f, -1.f),
    C_Vertex3f(0.122864723f, 0.604361534f, -1.f),
    C_Vertex3f(0.112304688f, 0.248188019f, -1.f),
    C_Vertex3f(0.115551949f, 0.0205402374f, -1.f),
    C_Vertex3f(0.112304688f, 0.227095604f, -1.f),
    C_Vertex3f(0.111389041f, 0.212169766f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0639648438f, 0.480166435f, -1.f),
    C_Vertex3f(0.0659427643f, 0.436164856f, -1.f),
    C_Vertex3f(0.0660684109f, 0.501635075f, -1.f),
    C_Vertex3f(0.0689096451f, 0.419296265f, -1.f),
    C_Vertex3f(0.0702755451f, 0.521861553f, -1.f),
    C_Vertex3f(0.0722060204f, 0.404747963f, -1.f),
    C_Vertex3f(0.0765862465f, 0.540845871f, -1.f),
    C_Vertex3f(0.0768208504f, 0.386897087f, -1.f),
    C_Vertex3f(0.085000515f, 0.558588028f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.456054688f, 0.485839844f, -1.f),
    C_Vertex3f(0.454579353f, 0.495666504f, -1.f),
    C_Vertex3f(0.450392485f, 0.514221191f, -1.f),
    C_Vertex3f(0.444969416f, 0.531677246f, -1.f),
    C_Vertex3f(0.438310146f, 0.548034668f, -1.f),
    C_Vertex3f(0.430414677f, 0.563293457f, -1.f),
    C_Vertex3f(0.421283007f, 0.577453613f, -1.f),
    C_Vertex3f(0.410915136f, 0.590515137f, -1.f),
    C_Vertex3f(0.399311066f, 0.602478027f, -1.f),
    C_Vertex3f(0.386470795f, 0.613342285f, -1.f),
    C_Vertex3f(0.377929688f, 0.474121094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.367311239f, 0.515843391f, -1.f),
    C_Vertex3f(0.371390224f, 0.504818439f, -1.f),
    C_Vertex3f(0.372763157f, 0.622848511f, -1.f),
    C_Vertex3f(0.37465632f, 0.493096828f, -1.f),
    C_Vertex3f(0.386470795f, 0.613342285f, -1.f),
    C_Vertex3f(0.377109528f, 0.480678558f, -1.f),
    C_Vertex3f(0.377929688f, 0.474121094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.350196958f, 0.544738293f, -1.f),
    C_Vertex3f(0.356714606f, 0.535803318f, -1.f),
    C_Vertex3f(0.358188152f, 0.630996704f, -1.f),
    C_Vertex3f(0.362419367f, 0.526171684f, -1.f),
    C_Vertex3f(0.372763157f, 0.622848511f, -1.f),
    C_Vertex3f(0.367311239f, 0.515843391f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.317227006f, 0.572773457f, -1.f),
    C_Vertex3f(0.326176524f, 0.567117214f, -1.f),
    C_Vertex3f(0.326436043f, 0.643218994f, -1.f),
    C_Vertex3f(0.334722996f, 0.560518265f, -1.f),
    C_Vertex3f(0.342745781f, 0.637786865f, -1.f),
    C_Vertex3f(0.342866421f, 0.552976608f, -1.f),
    C_Vertex3f(0.358188152f, 0.630996704f, -1.f),
    C_Vertex3f(0.350196958f, 0.544738293f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.309258938f, 0.647293091f, -1.f),
    C_Vertex3f(0.291214466f, 0.650009155f, -1.f),
    C_Vertex3f(0.29811883f, 0.58125782f, -1.f),
    C_Vertex3f(0.307874441f, 0.577486992f, -1.f),
    C_Vertex3f(0.317227006f, 0.572773457f, -1.f),
    C_Vertex3f(0.326436043f, 0.643218994f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.254683495f, 0.586914063f, -1.f),
    C_Vertex3f(0.266433716f, 0.586914063f, -1.f),
    C_Vertex3f(0.272302628f, 0.651367188f, -1.f),
    C_Vertex3f(0.277398467f, 0.585971355f, -1.f),
    C_Vertex3f(0.291214466f, 0.650009155f, -1.f),
    C_Vertex3f(0.287960172f, 0.584085941f, -1.f),
    C_Vertex3f(0.29811883f, 0.58125782f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.22135973f, 0.581054688f, -1.f),
    C_Vertex3f(0.23195529f, 0.583984375f, -1.f),
    C_Vertex3f(0.234275341f, 0.650097847f, -1.f),
    C_Vertex3f(0.243063211f, 0.5859375f, -1.f),
    C_Vertex3f(0.252797127f, 0.651367188f, -1.f),
    C_Vertex3f(0.254683495f, 0.586914063f, -1.f),
    C_Vertex3f(0.272302628f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.216463566f, 0.647559166f, -1.f),
    C_Vertex3f(0.199361801f, 0.643751144f, -1.f),
    C_Vertex3f(0.201705694f, 0.572265625f, -1.f),
    C_Vertex3f(0.211276531f, 0.577148438f, -1.f),
    C_Vertex3f(0.22135973f, 0.581054688f, -1.f),
    C_Vertex3f(0.234275341f, 0.650097847f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.169037819f, 0.543542385f, -1.f),
    C_Vertex3f(0.176067352f, 0.551757813f, -1.f),
    C_Vertex3f(0.182970047f, 0.638673782f, -1.f),
    C_Vertex3f(0.184101105f, 0.559570313f, -1.f),
    C_Vertex3f(0.199361801f, 0.643751144f, -1.f),
    C_Vertex3f(0.192647219f, 0.56640625f, -1.f),
    C_Vertex3f(0.201705694f, 0.572265625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.167288303f, 0.63232708f, -1.f),
    C_Vertex3f(0.15231657f, 0.624711037f, -1.f),
    C_Vertex3f(0.153974533f, 0.516478539f, -1.f),
    C_Vertex3f(0.157991409f, 0.525902748f, -1.f),
    C_Vertex3f(0.163012505f, 0.53492403f, -1.f),
    C_Vertex3f(0.169037819f, 0.543542385f, -1.f),
    C_Vertex3f(0.182970047f, 0.638673782f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15231657f, 0.624711037f, -1.f),
    C_Vertex3f(0.138054848f, 0.615825653f, -1.f),
    C_Vertex3f(0.147949219f, 0.485788345f, -1.f),
    C_Vertex3f(0.148953438f, 0.496421337f, -1.f),
    C_Vertex3f(0.150961876f, 0.506651402f, -1.f),
    C_Vertex3f(0.153974533f, 0.516478539f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147949219f, 0.474205971f, -1.f),
    C_Vertex3f(0.147949219f, 0.485788345f, -1.f),
    C_Vertex3f(0.145251274f, 0.116537571f, -1.f),
    C_Vertex3f(0.138054848f, 0.615825653f, -1.f),
    C_Vertex3f(0.127414703f, 0.0946846008f, -1.f),
    C_Vertex3f(0.122864723f, 0.604361534f, -1.f),
    C_Vertex3f(0.117034912f, 0.0836486816f, -1.f),
    C_Vertex3f(0.115551949f, 0.0205402374f, -1.f),
    C_Vertex3f(0.132841587f, 0.0246744156f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0190429688f, 0.353515625f, -1.f),
    C_Vertex3f(0.0190429688f, 0.2890625f, -1.f),
    C_Vertex3f(0.10483551f, 0.2890625f, -1.f),
    C_Vertex3f(0.0863800049f, 0.353515625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.186500549f, 0.2890625f, -1.f),
    C_Vertex3f(0.310546875f, 0.2890625f, -1.f),
    C_Vertex3f(0.310546875f, 0.353515625f, -1.f),
    C_Vertex3f(0.172920227f, 0.353515625f, -1.f),
    C_Vertex3f(0.181222916f, 0.180783272f, -1.f),
    C_Vertex3f(0.188023567f, 0.201433659f, -1.f),
    C_Vertex3f(0.188581467f, 0.284254074f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.17215538f, 0.159586906f, -1.f),
    C_Vertex3f(0.181222916f, 0.180783272f, -1.f),
    C_Vertex3f(0.172920227f, 0.353515625f, -1.f),
    C_Vertex3f(0.169798851f, 0.363295555f, -1.f),
    C_Vertex3f(0.164336443f, 0.381926537f, -1.f),
    C_Vertex3f(0.160164833f, 0.138171673f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.160164833f, 0.138171673f, -1.f),
    C_Vertex3f(0.164336443f, 0.381926537f, -1.f),
    C_Vertex3f(0.159654379f, 0.399628639f, -1.f),
    C_Vertex3f(0.155752659f, 0.416401863f, -1.f),
    C_Vertex3f(0.152631283f, 0.432246208f, -1.f),
    C_Vertex3f(0.150290251f, 0.447161674f, -1.f),
    C_Vertex3f(0.148729563f, 0.461148262f, -1.f),
    C_Vertex3f(0.147949219f, 0.474205971f, -1.f),
    C_Vertex3f(0.145251274f, 0.116537571f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.19378376f, 0.265023232f, -1.f),
    C_Vertex3f(0.191702843f, 0.274638176f, -1.f),
    C_Vertex3f(0.192557335f, 0.221538067f, -1.f),
    C_Vertex3f(0.194824219f, 0.241096497f, -1.f),
    C_Vertex3f(0.194824219f, 0.255409241f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.188581467f, 0.284254074f, -1.f),
    C_Vertex3f(0.188023567f, 0.201433659f, -1.f),
    C_Vertex3f(0.192557335f, 0.221538067f, -1.f),
    C_Vertex3f(0.191702843f, 0.274638176f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.415898323f, -0.00423336029f, -1.f),
    C_Vertex3f(0.429199696f, -0.00016450882f, -1.f),
    C_Vertex3f(0.434381485f, 0.0828857422f, -1.f),
    C_Vertex3f(0.444768429f, 0.00536966324f, -1.f),
    C_Vertex3f(0.447265625f, 0.0888671875f, -1.f),
    C_Vertex3f(0.462604523f, 0.0123691559f, -1.f),
    C_Vertex3f(0.47265625f, 0.0166015625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.371490479f, 0.0649414063f, -1.f),
    C_Vertex3f(0.377384186f, -0.0107421875f, -1.f),
    C_Vertex3f(0.390051842f, 0.0679321289f, -1.f),
    C_Vertex3f(0.399347305f, -0.00813865662f, -1.f),
    C_Vertex3f(0.411015511f, 0.0739135742f, -1.f),
    C_Vertex3f(0.415898323f, -0.00423336029f, -1.f),
    C_Vertex3f(0.434381485f, 0.0828857422f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360080719f, 0.0649414063f, -1.f),
    C_Vertex3f(0.351905346f, 0.0654911995f, -1.f),
    C_Vertex3f(0.357280731f, -0.0107421875f, -1.f),
    C_Vertex3f(0.377384186f, -0.0107421875f, -1.f),
    C_Vertex3f(0.371490479f, 0.0649414063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323858142f, -0.00603711605f, -1.f),
    C_Vertex3f(0.327839851f, -0.00695037842f, -1.f),
    C_Vertex3f(0.331009865f, 0.0682401657f, -1.f),
    C_Vertex3f(0.338710785f, -0.00898361206f, -1.f),
    C_Vertex3f(0.342215061f, 0.066590786f, -1.f),
    C_Vertex3f(0.357280731f, -0.0107421875f, -1.f),
    C_Vertex3f(0.351905346f, 0.0654911995f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.305726767f, -0.00111365318f, -1.f),
    C_Vertex3f(0.3128016f, -0.00311875343f, -1.f),
    C_Vertex3f(0.318289757f, 0.0704393387f, -1.f),
    C_Vertex3f(0.318845391f, -0.00475990772f, -1.f),
    C_Vertex3f(0.331009865f, 0.0682401657f, -1.f),
    C_Vertex3f(0.323858142f, -0.00603711605f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304054737f, 0.0731883049f, -1.f),
    C_Vertex3f(0.288304806f, 0.0764870644f, -1.f),
    C_Vertex3f(0.288483977f, 0.0039883852f, -1.f),
    C_Vertex3f(0.297620893f, 0.00125539303f, -1.f),
    C_Vertex3f(0.305726767f, -0.00111365318f, -1.f),
    C_Vertex3f(0.318289757f, 0.0704393387f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.256908298f, 0.0136567354f, -1.f),
    C_Vertex3f(0.267117023f, 0.0105462074f, -1.f),
    C_Vertex3f(0.271039963f, 0.0803356171f, -1.f),
    C_Vertex3f(0.278316021f, 0.00708532333f, -1.f),
    C_Vertex3f(0.288304806f, 0.0764870644f, -1.f),
    C_Vertex3f(0.288483977f, 0.0039883852f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.250293732f, 0.0852165222f, -1.f),
    C_Vertex3f(0.228727341f, 0.0892391205f, -1.f),
    C_Vertex3f(0.232223749f, 0.0208861828f, -1.f),
    C_Vertex3f(0.23946166f, 0.0188267231f, -1.f),
    C_Vertex3f(0.247689843f, 0.0164169073f, -1.f),
    C_Vertex3f(0.256908298f, 0.0136567354f, -1.f),
    C_Vertex3f(0.271039963f, 0.0803356171f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.228727341f, 0.0892391205f, -1.f),
    C_Vertex3f(0.208307266f, 0.0919208527f, -1.f),
    C_Vertex3f(0.214813232f, 0.0252914429f, -1.f),
    C_Vertex3f(0.216451645f, 0.0249624252f, -1.f),
    C_Vertex3f(0.220718741f, 0.0239540339f, -1.f),
    C_Vertex3f(0.22597611f, 0.0225952864f, -1.f),
    C_Vertex3f(0.232223749f, 0.0208861828f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.214813232f, 0.0252914429f, -1.f),
    C_Vertex3f(0.208307266f, 0.0919208527f, -1.f),
    C_Vertex3f(0.204540253f, 0.0270500183f, -1.f),
    C_Vertex3f(0.189033508f, 0.0932617188f, -1.f),
    C_Vertex3f(0.183998108f, 0.0288085938f, -1.f),
    C_Vertex3f(0.172868729f, 0.0932617188f, -1.f),
    C_Vertex3f(0.165863037f, 0.0288085938f, -1.f),
    C_Vertex3f(0.15790081f, 0.0920600891f, -1.f),
    C_Vertex3f(0.14961195f, 0.0274305344f, -1.f),
    C_Vertex3f(0.142167091f, 0.0896568298f, -1.f),
    C_Vertex3f(0.132841587f, 0.0246744156f, -1.f),
    C_Vertex3f(0.125667572f, 0.0860519409f, -1.f),
    C_Vertex3f(0.117034912f, 0.0836486816f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115551949f, 0.0205402374f, -1.f),
    C_Vertex3f(0.111389041f, 0.212169766f, -1.f),
    C_Vertex3f(0.109557748f, 0.197838187f, -1.f),
    C_Vertex3f(0.106810808f, 0.184100866f, -1.f),
    C_Vertex3f(0.103148222f, 0.170957804f, -1.f),
    C_Vertex3f(0.0985699892f, 0.158408999f, -1.f),
    C_Vertex3f(0.0977430344f, 0.0150279999f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0417480469f, -0.0122070313f, -1.f),
    C_Vertex3f(0.0490684509f, -0.00746536255f, -1.f),
    C_Vertex3f(0.059442997f, 0.0998997688f, -1.f),
    C_Vertex3f(0.0645012856f, 0.00102519989f, -1.f),
    C_Vertex3f(0.0756788254f, 0.118945122f, -1.f),
    C_Vertex3f(0.0807261467f, 0.00852298737f, -1.f),
    C_Vertex3f(0.086666584f, 0.135094166f, -1.f),
    C_Vertex3f(0.0977430344f, 0.0150279999f, -1.f),
    C_Vertex3f(0.0930761099f, 0.146454453f, -1.f),
    C_Vertex3f(0.0985699892f, 0.158408999f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0417480469f, -0.0122070313f, -1.f),
    C_Vertex3f(0.059442997f, 0.0998997688f, -1.f),
    C_Vertex3f(0.041621685f, 0.0833411217f, -1.f),
    C_Vertex3f(0.0222148895f, 0.0692691803f, -1.f),
    C_Vertex3f(0.01171875f, 0.0634765625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0659427643f, 0.436164856f, -1.f),
    C_Vertex3f(0.0639648438f, 0.480166435f, -1.f),
    C_Vertex3f(0.0639648438f, 0.458110809f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_164[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.697962463f, -0.716134369f, 0.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, 0.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, -1.f),
    C_Vertex3f(0.0322265625f, 0.479415894f, -1.f),
    C_Vertex3f(0.0322265625f, 0.479415894f, 0.f),
    C_Normal3f(-0.71627146f, 0.697821736f, 0.f),
    C_Vertex3f(0.0322265625f, 0.479415894f, 0.f),
    C_Vertex3f(0.0322265625f, 0.479415894f, -1.f),
    C_Vertex3f(0.0827255249f, 0.53125f, -1.f),
    C_Vertex3f(0.0827255249f, 0.53125f, 0.f),
    C_Normal3f(0.70634985f, 0.707862973f, 0.f),
    C_Vertex3f(0.0827255249f, 0.53125f, 0.f),
    C_Vertex3f(0.0827255249f, 0.53125f, -1.f),
    C_Vertex3f(0.1434021f, 0.470703125f, -1.f),
    C_Vertex3f(0.1434021f, 0.470703125f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.559735656f, 0.828671157f, 0.f),
    C_Vertex3f(0.1434021f, 0.470703125f, 0.f),
    C_Vertex3f(0.1434021f, 0.470703125f, -1.f),
    C_Normal3f(-0.519028664f, 0.854756892f, 0.f),
    C_Vertex3f(0.149185181f, 0.474609375f, 0.f),
    C_Vertex3f(0.149185181f, 0.474609375f, -1.f),
    C_Normal3f(-0.465210527f, 0.885200024f, 0.f),
    C_Vertex3f(0.161092758f, 0.481445313f, 0.f),
    C_Vertex3f(0.161092758f, 0.481445313f, -1.f),
    C_Normal3f(-0.396936834f, 0.917845905f, 0.f),
    C_Vertex3f(0.173341751f, 0.487304688f, 0.f),
    C_Vertex3f(0.173341751f, 0.487304688f, -1.f),
    C_Normal3f(-0.325603038f, 0.945506573f, 0.f),
    C_Vertex3f(0.185932159f, 0.4921875f, 0.f),
    C_Vertex3f(0.185932159f, 0.4921875f, -1.f),
    C_Normal3f(-0.252416164f, 0.967618704f, 0.f),
    C_Vertex3f(0.198863983f, 0.49609375f, 0.f),
    C_Vertex3f(0.198863983f, 0.49609375f, -1.f),
    C_Normal3f(-0.178676605f, 0.983907819f, 0.f),
    C_Vertex3f(0.212137222f, 0.499023438f, 0.f),
    C_Vertex3f(0.212137222f, 0.499023438f, -1.f),
    C_Normal3f(-0.105665922f, 0.994401693f, 0.f),
    C_Vertex3f(0.225751877f, 0.500976563f, 0.f),
    C_Vertex3f(0.225751877f, 0.500976563f, -1.f),
    C_Normal3f(-0.0344785117f, 0.999405444f, 0.f),
    C_Vertex3f(0.239707947f, 0.501953125f, 0.f),
    C_Vertex3f(0.239707947f, 0.501953125f, -1.f),
    C_Normal3f(0.0343816169f, 0.999408782f, 0.f),
    C_Vertex3f(0.254058838f, 0.501953125f, 0.f),
    C_Vertex3f(0.254058838f, 0.501953125f, -1.f),
    C_Normal3f(0.10516502f, 0.994454741f, 0.f),
    C_Vertex3f(0.268094778f, 0.500976563f, 0.f),
    C_Vertex3f(0.268094778f, 0.500976563f, -1.f),
    C_Normal3f(0.178170338f, 0.98399961f, 0.f),
    C_Vertex3f(0.281762362f, 0.499023438f, 0.f),
    C_Vertex3f(0.281762362f, 0.499023438f, -1.f),
    C_Normal3f(0.25219059f, 0.967677593f, 0.f),
    C_Vertex3f(0.295061588f, 0.49609375f, 0.f),
    C_Vertex3f(0.295061588f, 0.49609375f, -1.f),
    C_Normal3f(0.325932384f, 0.945393145f, 0.f),
    C_Vertex3f(0.307992458f, 0.4921875f, 0.f),
    C_Vertex3f(0.307992458f, 0.4921875f, -1.f),
    C_Normal3f(0.398053437f, 0.917362273f, 0.f),
    C_Vertex3f(0.320554972f, 0.487304688f, 0.f),
    C_Vertex3f(0.320554972f, 0.487304688f, -1.f),
    C_Normal3f(0.46728f, 0.884109378f, 0.f),
    C_Vertex3f(0.332749128f, 0.481445313f, 0.f),
    C_Vertex3f(0.332749128f, 0.481445313f, -1.f),
    C_Normal3f(0.521960199f, 0.852969766f, 0.f),
    C_Vertex3f(0.344574928f, 0.474609375f, 0.f),
    C_Vertex3f(0.344574928f, 0.474609375f, -1.f),
    C_Normal3f(0.563366711f, 0.826206982f, 0.f),
    C_Vertex3f(0.35030365f, 0.470703125f, 0.f),
    C_Vertex3f(0.35030365f, 0.470703125f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.70634985f, 0.707862973f, 0.f),
    C_Vertex3f(0.35030365f, 0.470703125f, 0.f),
    C_Vertex3f(0.35030365f, 0.470703125f, -1.f),
    C_Vertex3f(0.410980225f, 0.53125f, -1.f),
    C_Vertex3f(0.410980225f, 0.53125f, 0.f),
    C_Normal3f(0.713273823f, 0.700885475f, 0.f),
    C_Vertex3f(0.410980225f, 0.53125f, 0.f),
    C_Vertex3f(0.410980225f, 0.53125f, -1.f),
    C_Vertex3f(0.461914063f, 0.479415894f, -1.f),
    C_Vertex3f(0.461914063f, 0.479415894f, 0.f),
    C_Normal3f(0.697962463f, -0.716134369f, 0.f),
    C_Vertex3f(0.461914063f, 0.479415894f, 0.f),
    C_Vertex3f(0.461914063f, 0.479415894f, -1.f),
    C_Vertex3f(0.400878906f, 0.419929504f, -1.f),
    C_Vertex3f(0.400878906f, 0.419929504f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.838659227f, 0.544656456f, 0.f),
    C_Vertex3f(0.400878906f, 0.419929504f, 0.f),
    C_Vertex3f(0.400878906f, 0.419929504f, -1.f),
    C_Normal3f(0.861783087f, 0.507276893f, 0.f),
    C_Vertex3f(0.404785156f, 0.41391468f, 0.f),
    C_Vertex3f(0.404785156f, 0.41391468f, -1.f),
    C_Normal3f(0.889200091f, 0.457518458f, 0.f),
    C_Vertex3f(0.411621094f, 0.401680231f, 0.f),
    C_Vertex3f(0.411621094f, 0.401680231f, -1.f),
    C_Normal3f(0.919248462f, 0.393677801f, 0.f),
    C_Vertex3f(0.417480469f, 0.38924098f, 0.f),
    C_Vertex3f(0.417480469f, 0.38924098f, -1.f),
    C_Normal3f(0.945391178f, 0.325937808f, 0.f),
    C_Vertex3f(0.422363281f, 0.376596928f, 0.f),
    C_Vertex3f(0.422363281f, 0.376596928f, -1.f),
    C_Normal3f(0.966895282f, 0.255173564f, 0.f),
    C_Vertex3f(0.426269531f, 0.363748074f, 0.f),
    C_Vertex3f(0.426269531f, 0.363748074f, -1.f),
    C_Normal3f(0.983213663f, 0.182457715f, 0.f),
    C_Vertex3f(0.429199219f, 0.350694418f, 0.f),
    C_Vertex3f(0.429199219f, 0.350694418f, -1.f),
    C_Normal3f(0.994043589f, 0.108983904f, 0.f),
    C_Vertex3f(0.431152344f, 0.337435961f, 0.f),
    C_Vertex3f(0.431152344f, 0.337435961f, -1.f),
    C_Normal3f(0.999355435f, 0.0358975343f, 0.f),
    C_Vertex3f(0.432128906f, 0.323972702f, 0.f),
    C_Vertex3f(0.432128906f, 0.323972702f, -1.f),
    C_Normal3f(0.999360025f, -0.035770338f, 0.f),
    C_Vertex3f(0.432128906f, 0.310249329f, 0.f),
    C_Vertex3f(0.432128906f, 0.310249329f, -1.f),
    C_Normal3f(0.994122207f, -0.108264193f, 0.f),
    C_Vertex3f(0.431152344f, 0.296689272f, 0.f),
    C_Vertex3f(0.431152344f, 0.296689272f, -1.f),
    C_Normal3f(0.983401358f, -0.181443617f, 0.f),
    C_Vertex3f(0.429199219f, 0.283347845f, 0.f),
    C_Vertex3f(0.429199219f, 0.283347845f, -1.f),
    C_Normal3f(0.967195749f, -0.254032314f, 0.f),
    C_Vertex3f(0.426269531f, 0.270225048f, 0.f),
    C_Vertex3f(0.426269531f, 0.270225048f, -1.f),
    C_Normal3f(0.945770681f, -0.324835241f, 0.f),
    C_Vertex3f(0.422363281f, 0.257320881f, 0.f),
    C_Vertex3f(0.422363281f, 0.257320881f, -1.f),
    C_Normal3f(0.919639826f, -0.39276275f, 0.f),
    C_Vertex3f(0.417480469f, 0.244635344f, 0.f),
    C_Vertex3f(0.417480469f, 0.244635344f, -1.f),
    C_Normal3f(0.889512479f, -0.456910968f, 0.f),
    C_Vertex3f(0.411621094f, 0.232168436f, 0.f),
    C_Vertex3f(0.411621094f, 0.232168436f, -1.f),
    C_Normal3f(0.861951053f, -0.506991625f, 0.f),
    C_Vertex3f(0.404785156f, 0.219920158f, 0.f),
    C_Vertex3f(0.404785156f, 0.219920158f, -1.f),
    C_Normal3f(0.838659227f, -0.544656456f, 0.f),
    C_Vertex3f(0.400878906f, 0.213905334f, 0.f),
    C_Vertex3f(0.400878906f, 0.213905334f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.697962463f, 0.716134369f, 0.f),
    C_Vertex3f(0.400878906f, 0.213905334f, 0.f),
    C_Vertex3f(0.400878906f, 0.213905334f, -1.f),
    C_Vertex3f(0.461914063f, 0.154418945f, -1.f),
    C_Vertex3f(0.461914063f, 0.154418945f, 0.f),
    C_Normal3f(0.713583052f, -0.700570643f, 0.f),
    C_Vertex3f(0.461914063f, 0.154418945f, 0.f),
    C_Vertex3f(0.461914063f, 0.154418945f, -1.f),
    C_Vertex3f(0.410980225f, 0.102539063f, -1.f),
    C_Vertex3f(0.410980225f, 0.102539063f, 0.f),
    C_Normal3f(-0.709186971f, -0.705020487f, 0.f),
    C_Vertex3f(0.410980225f, 0.102539063f, 0.f),
    C_Vertex3f(0.410980225f, 0.102539063f, -1.f),
    C_Vertex3f(0.35030365f, 0.163574219f, -1.f),
    C_Vertex3f(0.35030365f, 0.163574219f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.56933099f, -0.822108448f, 0.f),
    C_Vertex3f(0.35030365f, 0.163574219f, 0.f),
    C_Vertex3f(0.35030365f, 0.163574219f, -1.f),
    C_Normal3f(0.527856052f, -0.849333763f, 0.f),
    C_Vertex3f(0.344574928f, 0.159606934f, 0.f),
    C_Vertex3f(0.344574928f, 0.159606934f, -1.f),
    C_Normal3f(0.47295782f, -0.881085098f, 0.f),
    C_Vertex3f(0.332749128f, 0.152664185f, 0.f),
    C_Vertex3f(0.332749128f, 0.152664185f, -1.f),
    C_Normal3f(0.403268069f, -0.915081859f, 0.f),
    C_Vertex3f(0.320554972f, 0.146713257f, 0.f),
    C_Vertex3f(0.320554972f, 0.146713257f, -1.f),
    C_Normal3f(0.330472738f, -0.943815649f, 0.f),
    C_Vertex3f(0.307992458f, 0.14175415f, 0.f),
    C_Vertex3f(0.307992458f, 0.14175415f, -1.f),
    C_Normal3f(0.255874932f, -0.966709912f, 0.f),
    C_Vertex3f(0.295061588f, 0.137786865f, 0.f),
    C_Vertex3f(0.295061588f, 0.137786865f, -1.f),
    C_Normal3f(0.180863857f, -0.98350811f, 0.f),
    C_Vertex3f(0.281762362f, 0.134811401f, 0.f),
    C_Vertex3f(0.281762362f, 0.134811401f, -1.f),
    C_Normal3f(0.106789626f, -0.99428153f, 0.f),
    C_Vertex3f(0.268094778f, 0.132827759f, 0.f),
    C_Vertex3f(0.268094778f, 0.132827759f, -1.f),
    C_Normal3f(0.0349181779f, -0.999390185f, 0.f),
    C_Vertex3f(0.254058838f, 0.131835938f, 0.f),
    C_Vertex3f(0.254058838f, 0.131835938f, -1.f),
    C_Normal3f(-0.0350165814f, -0.999386668f, 0.f),
    C_Vertex3f(0.239707947f, 0.131835938f, 0.f),
    C_Vertex3f(0.239707947f, 0.131835938f, -1.f),
    C_Normal3f(-0.107298091f, -0.994226873f, 0.f),
    C_Vertex3f(0.225751877f, 0.132827759f, 0.f),
    C_Vertex3f(0.225751877f, 0.132827759f, -1.f),
    C_Normal3f(-0.181377262f, -0.983413577f, 0.f),
    C_Vertex3f(0.212137222f, 0.134811401f, 0.f),
    C_Vertex3f(0.212137222f, 0.134811401f, -1.f),
    C_Normal3f(-0.256103337f, -0.966649354f, 0.f),
    C_Vertex3f(0.198863983f, 0.137786865f, 0.f),
    C_Vertex3f(0.198863983f, 0.137786865f, -1.f),
    C_Normal3f(-0.330139875f, -0.943931997f, 0.f),
    C_Vertex3f(0.185932159f, 0.14175415f, 0.f),
    C_Vertex3f(0.185932159f, 0.14175415f, -1.f),
    C_Normal3f(-0.402142465f, -0.915577114f, 0.f),
    C_Vertex3f(0.173341751f, 0.146713257f, 0.f),
    C_Vertex3f(0.173341751f, 0.146713257f, -1.f),
    C_Normal3f(-0.470877469f, -0.882198691f, 0.f),
    C_Vertex3f(0.161092758f, 0.152664185f, 0.f),
    C_Vertex3f(0.161092758f, 0.152664185f, -1.f),
    C_Normal3f(-0.524916351f, -0.851153731f, 0.f),
    C_Vertex3f(0.149185181f, 0.159606934f, 0.f),
    C_Vertex3f(0.149185181f, 0.159606934f, -1.f),
    C_Normal3f(-0.565697491f, -0.824612856f, 0.f),
    C_Vertex3f(0.1434021f, 0.163574219f, 0.f),
    C_Vertex3f(0.1434021f, 0.163574219f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.709186971f, -0.705020487f, 0.f),
    C_Vertex3f(0.1434021f, 0.163574219f, 0.f),
    C_Vertex3f(0.1434021f, 0.163574219f, -1.f),
    C_Vertex3f(0.0827255249f, 0.102539063f, -1.f),
    C_Vertex3f(0.0827255249f, 0.102539063f, 0.f),
    C_Normal3f(-0.716579258f, -0.697505593f, 0.f),
    C_Vertex3f(0.0827255249f, 0.102539063f, 0.f),
    C_Vertex3f(0.0827255249f, 0.102539063f, -1.f),
    C_Vertex3f(0.0322265625f, 0.154418945f, -1.f),
    C_Vertex3f(0.0322265625f, 0.154418945f, 0.f),
    C_Normal3f(-0.697962463f, 0.716134369f, 0.f),
    C_Vertex3f(0.0322265625f, 0.154418945f, 0.f),
    C_Vertex3f(0.0322265625f, 0.154418945f, -1.f),
    C_Vertex3f(0.0932617188f, 0.213905334f, -1.f),
    C_Vertex3f(0.0932617188f, 0.213905334f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.838659227f, -0.544656456f, 0.f),
    C_Vertex3f(0.0932617188f, 0.213905334f, 0.f),
    C_Vertex3f(0.0932617188f, 0.213905334f, -1.f),
    C_Normal3f(-0.861951053f, -0.506991625f, 0.f),
    C_Vertex3f(0.0893554688f, 0.219920158f, 0.f),
    C_Vertex3f(0.0893554688f, 0.219920158f, -1.f),
    C_Normal3f(-0.889512479f, -0.456910968f, 0.f),
    C_Vertex3f(0.0825195313f, 0.232168436f, 0.f),
    C_Vertex3f(0.0825195313f, 0.232168436f, -1.f),
    C_Normal3f(-0.919639826f, -0.39276275f, 0.f),
    C_Vertex3f(0.0766601563f, 0.244635344f, 0.f),
    C_Vertex3f(0.0766601563f, 0.244635344f, -1.f),
    C_Normal3f(-0.945770681f, -0.324835241f, 0.f),
    C_Vertex3f(0.0717773438f, 0.257320881f, 0.f),
    C_Vertex3f(0.0717773438f, 0.257320881f, -1.f),
    C_Normal3f(-0.967195749f, -0.254032314f, 0.f),
    C_Vertex3f(0.0678710938f, 0.270225048f, 0.f),
    C_Vertex3f(0.0678710938f, 0.270225048f, -1.f),
    C_Normal3f(-0.983401358f, -0.181443617f, 0.f),
    C_Vertex3f(0.0649414063f, 0.283347845f, 0.f),
    C_Vertex3f(0.0649414063f, 0.283347845f, -1.f),
    C_Normal3f(-0.994122207f, -0.108264193f, 0.f),
    C_Vertex3f(0.0629882813f, 0.296689272f, 0.f),
    C_Vertex3f(0.0629882813f, 0.296689272f, -1.f),
    C_Normal3f(-0.999360025f, -0.035770338f, 0.f),
    C_Vertex3f(0.0620117188f, 0.310249329f, 0.f),
    C_Vertex3f(0.0620117188f, 0.310249329f, -1.f),
    C_Normal3f(-0.999355435f, 0.0358975343f, 0.f),
    C_Vertex3f(0.0620117188f, 0.323972702f, 0.f),
    C_Vertex3f(0.0620117188f, 0.323972702f, -1.f),
    C_Normal3f(-0.994043589f, 0.108983904f, 0.f),
    C_Vertex3f(0.0629882813f, 0.337435961f, 0.f),
    C_Vertex3f(0.0629882813f, 0.337435961f, -1.f),
    C_Normal3f(-0.983213663f, 0.182457715f, 0.f),
    C_Vertex3f(0.0649414063f, 0.350694418f, 0.f),
    C_Vertex3f(0.0649414063f, 0.350694418f, -1.f),
    C_Normal3f(-0.966895282f, 0.255173564f, 0.f),
    C_Vertex3f(0.0678710938f, 0.363748074f, 0.f),
    C_Vertex3f(0.0678710938f, 0.363748074f, -1.f),
    C_Normal3f(-0.945391178f, 0.325937808f, 0.f),
    C_Vertex3f(0.0717773438f, 0.376596928f, 0.f),
    C_Vertex3f(0.0717773438f, 0.376596928f, -1.f),
    C_Normal3f(-0.919248462f, 0.393677801f, 0.f),
    C_Vertex3f(0.0766601563f, 0.38924098f, 0.f),
    C_Vertex3f(0.0766601563f, 0.38924098f, -1.f),
    C_Normal3f(-0.889200091f, 0.457518458f, 0.f),
    C_Vertex3f(0.0825195313f, 0.401680231f, 0.f),
    C_Vertex3f(0.0825195313f, 0.401680231f, -1.f),
    C_Normal3f(-0.861783087f, 0.507276893f, 0.f),
    C_Vertex3f(0.0893554688f, 0.41391468f, 0.f),
    C_Vertex3f(0.0893554688f, 0.41391468f, -1.f),
    C_Normal3f(-0.838659227f, 0.544656456f, 0.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, 0.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.998984873f, -0.0450478606f, 0.f),
    C_Vertex3f(0.133300781f, 0.322954178f, 0.f),
    C_Vertex3f(0.133300781f, 0.322954178f, -1.f),
    C_Normal3f(0.998993337f, 0.0448593311f, 0.f),
    C_Vertex3f(0.133300781f, 0.311222076f, 0.f),
    C_Vertex3f(0.133300781f, 0.311222076f, -1.f),
    C_Normal3f(0.990387499f, 0.138320491f, 0.f),
    C_Vertex3f(0.134337306f, 0.299871325f, 0.f),
    C_Vertex3f(0.134337306f, 0.299871325f, -1.f),
    C_Normal3f(0.971883357f, 0.235462829f, 0.f),
    C_Vertex3f(0.136410356f, 0.288957238f, 0.f),
    C_Vertex3f(0.136410356f, 0.288957238f, -1.f),
    C_Normal3f(0.94278878f, 0.333390504f, 0.f),
    C_Vertex3f(0.13951993f, 0.278479815f, 0.f),
    C_Vertex3f(0.13951993f, 0.278479815f, -1.f),
    C_Normal3f(0.903323531f, 0.428959876f, 0.f),
    C_Vertex3f(0.143666029f, 0.268439054f, 0.f),
    C_Vertex3f(0.143666029f, 0.268439054f, -1.f),
    C_Normal3f(0.854691029f, 0.519136846f, 0.f),
    C_Vertex3f(0.148848653f, 0.258834958f, 0.f),
    C_Vertex3f(0.148848653f, 0.258834958f, -1.f),
    C_Normal3f(0.798902333f, 0.601460814f, 0.f),
    C_Vertex3f(0.155067801f, 0.249667525f, 0.f),
    C_Vertex3f(0.155067801f, 0.249667525f, -1.f),
    C_Normal3f(0.738411605f, 0.674350262f, 0.f),
    C_Vertex3f(0.162323475f, 0.240936756f, 0.f),
    C_Vertex3f(0.162323475f, 0.240936756f, -1.f),
    C_Normal3f(0.674815953f, 0.737986028f, 0.f),
    C_Vertex3f(0.170615673f, 0.232642651f, 0.f),
    C_Vertex3f(0.170615673f, 0.232642651f, -1.f),
    C_Normal3f(0.602521777f, 0.798102438f, 0.f),
    C_Vertex3f(0.179330707f, 0.225385308f, 0.f),
    C_Vertex3f(0.179330707f, 0.225385308f, -1.f),
    C_Normal3f(0.520701468f, 0.853738785f, 0.f),
    C_Vertex3f(0.188468575f, 0.219164729f, 0.f),
    C_Vertex3f(0.188468575f, 0.219164729f, -1.f),
    C_Normal3f(0.430839926f, 0.902428389f, 0.f),
    C_Vertex3f(0.19802928f, 0.213980913f, 0.f),
    C_Vertex3f(0.19802928f, 0.213980913f, -1.f),
    C_Normal3f(0.335321158f, 0.942103803f, 0.f),
    C_Vertex3f(0.208012819f, 0.20983386f, 0.f),
    C_Vertex3f(0.208012819f, 0.20983386f, -1.f),
    C_Normal3f(0.237146378f, 0.971473932f, 0.f),
    C_Vertex3f(0.218419194f, 0.206723571f, 0.f),
    C_Vertex3f(0.218419194f, 0.206723571f, -1.f),
    C_Normal3f(0.139479667f, 0.990224957f, 0.f),
    C_Vertex3f(0.229248405f, 0.204650044f, 0.f),
    C_Vertex3f(0.229248405f, 0.204650044f, -1.f),
    C_Normal3f(0.0451741628f, 0.998979092f, 0.f),
    C_Vertex3f(0.24050045f, 0.203613281f, 0.f),
    C_Vertex3f(0.24050045f, 0.203613281f, -1.f),
    C_Normal3f(-0.0451741628f, 0.998979092f, 0.f),
    C_Vertex3f(0.252175331f, 0.203613281f, 0.f),
    C_Vertex3f(0.252175331f, 0.203613281f, -1.f),
    C_Normal3f(-0.139479667f, 0.990224957f, 0.f),
    C_Vertex3f(0.263427377f, 0.204650044f, 0.f),
    C_Vertex3f(0.263427377f, 0.204650044f, -1.f),
    C_Normal3f(-0.237146378f, 0.971473932f, 0.f),
    C_Vertex3f(0.274256587f, 0.206723571f, 0.f),
    C_Vertex3f(0.274256587f, 0.206723571f, -1.f),
    C_Normal3f(-0.335321158f, 0.942103803f, 0.f),
    C_Vertex3f(0.284662962f, 0.20983386f, 0.f),
    C_Vertex3f(0.284662962f, 0.20983386f, -1.f),
    C_Normal3f(-0.430839926f, 0.902428389f, 0.f),
    C_Vertex3f(0.294646502f, 0.213980913f, 0.f),
    C_Vertex3f(0.294646502f, 0.213980913f, -1.f),
    C_Normal3f(-0.520701468f, 0.853738785f, 0.f),
    C_Vertex3f(0.304207206f, 0.219164729f, 0.f),
    C_Vertex3f(0.304207206f, 0.219164729f, -1.f),
    C_Normal3f(-0.602521777f, 0.798102438f, 0.f),
    C_Vertex3f(0.313345075f, 0.225385308f, 0.f),
    C_Vertex3f(0.313345075f, 0.225385308f, -1.f),
    C_Normal3f(-0.674815953f, 0.737986028f, 0.f),
    C_Vertex3f(0.322060108f, 0.232642651f, 0.f),
    C_Vertex3f(0.322060108f, 0.232642651f, -1.f),
    C_Normal3f(-0.738411605f, 0.674350262f, 0.f),
    C_Vertex3f(0.330352306f, 0.240936756f, 0.f),
    C_Vertex3f(0.330352306f, 0.240936756f, -1.f),
    C_Normal3f(-0.798902333f, 0.601460814f, 0.f),
    C_Vertex3f(0.33760798f, 0.249667525f, 0.f),
    C_Vertex3f(0.33760798f, 0.249667525f, -1.f),
    C_Normal3f(-0.854691029f, 0.519136846f, 0.f),
    C_Vertex3f(0.343827128f, 0.258834958f, 0.f),
    C_Vertex3f(0.343827128f, 0.258834958f, -1.f),
    C_Normal3f(-0.903323531f, 0.428959876f, 0.f),
    C_Vertex3f(0.349009752f, 0.268439054f, 0.f),
    C_Vertex3f(0.349009752f, 0.268439054f, -1.f),
    C_Normal3f(-0.94278878f, 0.333390504f, 0.f),
    C_Vertex3f(0.353155851f, 0.278479815f, 0.f),
    C_Vertex3f(0.353155851f, 0.278479815f, -1.f),
    C_Normal3f(-0.971883357f, 0.235462829f, 0.f),
    C_Vertex3f(0.356265426f, 0.288957238f, 0.f),
    C_Vertex3f(0.356265426f, 0.288957238f, -1.f),
    C_Normal3f(-0.990387499f, 0.138320491f, 0.f),
    C_Vertex3f(0.358338475f, 0.299871325f, 0.f),
    C_Vertex3f(0.358338475f, 0.299871325f, -1.f),
    C_Normal3f(-0.998993337f, 0.0448593311f, 0.f),
    C_Vertex3f(0.359375f, 0.311222076f, 0.f),
    C_Vertex3f(0.359375f, 0.311222076f, -1.f),
    C_Normal3f(-0.998998046f, -0.044753138f, 0.f),
    C_Vertex3f(0.359375f, 0.322954178f, 0.f),
    C_Vertex3f(0.359375f, 0.322954178f, -1.f),
    C_Normal3f(-0.990358412f, -0.138528198f, 0.f),
    C_Vertex3f(0.35834527f, 0.334208131f, 0.f),
    C_Vertex3f(0.35834527f, 0.334208131f, -1.f),
    C_Normal3f(-0.971853256f, -0.235587239f, 0.f),
    C_Vertex3f(0.35628581f, 0.345039248f, 0.f),
    C_Vertex3f(0.35628581f, 0.345039248f, -1.f),
    C_Normal3f(-0.942840457f, -0.333244413f, 0.f),
    C_Vertex3f(0.353196621f, 0.355447531f, 0.f),
    C_Vertex3f(0.353196621f, 0.355447531f, -1.f),
    C_Normal3f(-0.903595388f, -0.428386897f, 0.f),
    C_Vertex3f(0.349077702f, 0.365432978f, 0.f),
    C_Vertex3f(0.349077702f, 0.365432978f, -1.f),
    C_Normal3f(-0.855354905f, -0.518042386f, 0.f),
    C_Vertex3f(0.343929052f, 0.374995589f, 0.f),
    C_Vertex3f(0.343929052f, 0.374995589f, -1.f),
    C_Normal3f(-0.800130069f, -0.599826455f, 0.f),
    C_Vertex3f(0.337750673f, 0.384135365f, 0.f),
    C_Vertex3f(0.337750673f, 0.384135365f, -1.f),
    C_Normal3f(-0.74034369f, -0.672228515f, 0.f),
    C_Vertex3f(0.330542564f, 0.392852306f, 0.f),
    C_Vertex3f(0.330542564f, 0.392852306f, -1.f),
    C_Normal3f(-0.67672801f, -0.736233175f, 0.f),
    C_Vertex3f(0.322304726f, 0.401146412f, 0.f),
    C_Vertex3f(0.322304726f, 0.401146412f, -1.f),
    C_Normal3f(-0.603545547f, -0.797328591f, 0.f),
    C_Vertex3f(0.313623667f, 0.408403754f, 0.f),
    C_Vertex3f(0.313623667f, 0.408403754f, -1.f),
    C_Normal3f(-0.520839453f, -0.853654683f, 0.f),
    C_Vertex3f(0.304499388f, 0.414624333f, 0.f),
    C_Vertex3f(0.304499388f, 0.414624333f, -1.f),
    C_Normal3f(-0.430230767f, -0.902718961f, 0.f),
    C_Vertex3f(0.294931889f, 0.419808149f, 0.f),
    C_Vertex3f(0.294931889f, 0.419808149f, -1.f),
    C_Normal3f(-0.334233522f, -0.942490339f, 0.f),
    C_Vertex3f(0.284921169f, 0.423955202f, 0.f),
    C_Vertex3f(0.284921169f, 0.423955202f, -1.f),
    C_Normal3f(-0.235934973f, -0.971768856f, 0.f),
    C_Vertex3f(0.27446723f, 0.427065492f, 0.f),
    C_Vertex3f(0.27446723f, 0.427065492f, -1.f),
    C_Normal3f(-0.138518289f, -0.990359843f, 0.f),
    C_Vertex3f(0.26357007f, 0.429139018f, 0.f),
    C_Vertex3f(0.26357007f, 0.429139018f, -1.f),
    C_Normal3f(-0.0448953137f, -0.998991728f, 0.f),
    C_Vertex3f(0.252229691f, 0.430175781f, 0.f),
    C_Vertex3f(0.252229691f, 0.430175781f, -1.f),
    C_Normal3f(0.0450675264f, -0.99898392f, 0.f),
    C_Vertex3f(0.24050045f, 0.430175781f, 0.f),
    C_Vertex3f(0.24050045f, 0.430175781f, -1.f),
    C_Normal3f(0.139479667f, -0.990224957f, 0.f),
    C_Vertex3f(0.229248405f, 0.429139018f, 0.f),
    C_Vertex3f(0.229248405f, 0.429139018f, -1.f),
    C_Normal3f(0.237146378f, -0.971473932f, 0.f),
    C_Vertex3f(0.218419194f, 0.427065492f, 0.f),
    C_Vertex3f(0.218419194f, 0.427065492f, -1.f),
    C_Normal3f(0.335321158f, -0.942103803f, 0.f),
    C_Vertex3f(0.208012819f, 0.423955202f, 0.f),
    C_Vertex3f(0.208012819f, 0.423955202f, -1.f),
    C_Normal3f(0.430839926f, -0.902428389f, 0.f),
    C_Vertex3f(0.19802928f, 0.419808149f, 0.f),
    C_Vertex3f(0.19802928f, 0.419808149f, -1.f),
    C_Normal3f(0.520701468f, -0.853738785f, 0.f),
    C_Vertex3f(0.188468575f, 0.414624333f, 0.f),
    C_Vertex3f(0.188468575f, 0.414624333f, -1.f),
    C_Normal3f(0.602521777f, -0.798102438f, 0.f),
    C_Vertex3f(0.179330707f, 0.408403754f, 0.f),
    C_Vertex3f(0.179330707f, 0.408403754f, -1.f),
    C_Normal3f(0.674815953f, -0.737986028f, 0.f),
    C_Vertex3f(0.170615673f, 0.401146412f, 0.f),
    C_Vertex3f(0.170615673f, 0.401146412f, -1.f),
    C_Normal3f(0.738138676f, -0.674649f, 0.f),
    C_Vertex3f(0.162323475f, 0.392852306f, 0.f),
    C_Vertex3f(0.162323475f, 0.392852306f, -1.f),
    C_Normal3f(0.798231006f, -0.602351546f, 0.f),
    C_Vertex3f(0.155067801f, 0.384135365f, 0.f),
    C_Vertex3f(0.155067801f, 0.384135365f, -1.f),
    C_Normal3f(0.853839219f, -0.52053678f, 0.f),
    C_Vertex3f(0.148848653f, 0.374995589f, 0.f),
    C_Vertex3f(0.148848653f, 0.374995589f, -1.f),
    C_Normal3f(0.902499616f, -0.430690795f, 0.f),
    C_Vertex3f(0.143666029f, 0.365432978f, 0.f),
    C_Vertex3f(0.143666029f, 0.365432978f, -1.f),
    C_Normal3f(0.94214797f, -0.335197061f, 0.f),
    C_Vertex3f(0.13951993f, 0.355447531f, 0.f),
    C_Vertex3f(0.13951993f, 0.355447531f, -1.f),
    C_Normal3f(0.971496284f, -0.237054706f, 0.f),
    C_Vertex3f(0.136410356f, 0.345039248f, 0.f),
    C_Vertex3f(0.136410356f, 0.345039248f, -1.f),
    C_Normal3f(0.990232646f, -0.139424592f, 0.f),
    C_Vertex3f(0.134337306f, 0.334208131f, 0.f),
    C_Vertex3f(0.134337306f, 0.334208131f, -1.f),
    C_Normal3f(0.998984873f, -0.0450478606f, 0.f),
    C_Vertex3f(0.133300781f, 0.322954178f, 0.f),
    C_Vertex3f(0.133300781f, 0.322954178f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.134337306f, 0.299871325f, 0.f),
    C_Vertex3f(0.133300781f, 0.311222076f, 0.f),
    C_Vertex3f(0.1434021f, 0.163574219f, 0.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, 0.f),
    C_Vertex3f(0.0932617188f, 0.213905334f, 0.f),
    C_Vertex3f(0.0893554688f, 0.41391468f, 0.f),
    C_Vertex3f(0.0893554688f, 0.219920158f, 0.f),
    C_Vertex3f(0.0825195313f, 0.401680231f, 0.f),
    C_Vertex3f(0.0825195313f, 0.232168436f, 0.f),
    C_Vertex3f(0.0766601563f, 0.38924098f, 0.f),
    C_Vertex3f(0.0766601563f, 0.244635344f, 0.f),
    C_Vertex3f(0.0717773438f, 0.376596928f, 0.f),
    C_Vertex3f(0.0717773438f, 0.257320881f, 0.f),
    C_Vertex3f(0.0678710938f, 0.363748074f, 0.f),
    C_Vertex3f(0.0678710938f, 0.270225048f, 0.f),
    C_Vertex3f(0.0649414063f, 0.350694418f, 0.f),
    C_Vertex3f(0.0649414063f, 0.283347845f, 0.f),
    C_Vertex3f(0.0629882813f, 0.337435961f, 0.f),
    C_Vertex3f(0.0629882813f, 0.296689272f, 0.f),
    C_Vertex3f(0.0620117188f, 0.323972702f, 0.f),
    C_Vertex3f(0.0620117188f, 0.310249329f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.410980225f, 0.53125f, 0.f),
    C_Vertex3f(0.35030365f, 0.470703125f, 0.f),
    C_Vertex3f(0.353196621f, 0.355447531f, 0.f),
    C_Vertex3f(0.35628581f, 0.345039248f, 0.f),
    C_Vertex3f(0.35834527f, 0.334208131f, 0.f),
    C_Vertex3f(0.359375f, 0.322954178f, 0.f),
    C_Vertex3f(0.400878906f, 0.419929504f, 0.f),
    C_Vertex3f(0.461914063f, 0.479415894f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.359375f, 0.311222076f, 0.f),
    C_Vertex3f(0.410980225f, 0.102539063f, 0.f),
    C_Vertex3f(0.359375f, 0.322954178f, 0.f),
    C_Vertex3f(0.400878906f, 0.213905334f, 0.f),
    C_Vertex3f(0.400878906f, 0.419929504f, 0.f),
    C_Vertex3f(0.404785156f, 0.219920158f, 0.f),
    C_Vertex3f(0.404785156f, 0.41391468f, 0.f),
    C_Vertex3f(0.411621094f, 0.232168436f, 0.f),
    C_Vertex3f(0.411621094f, 0.401680231f, 0.f),
    C_Vertex3f(0.417480469f, 0.244635344f, 0.f),
    C_Vertex3f(0.417480469f, 0.38924098f, 0.f),
    C_Vertex3f(0.422363281f, 0.257320881f, 0.f),
    C_Vertex3f(0.422363281f, 0.376596928f, 0.f),
    C_Vertex3f(0.426269531f, 0.270225048f, 0.f),
    C_Vertex3f(0.426269531f, 0.363748074f, 0.f),
    C_Vertex3f(0.429199219f, 0.283347845f, 0.f),
    C_Vertex3f(0.429199219f, 0.350694418f, 0.f),
    C_Vertex3f(0.431152344f, 0.296689272f, 0.f),
    C_Vertex3f(0.431152344f, 0.337435961f, 0.f),
    C_Vertex3f(0.432128906f, 0.310249329f, 0.f),
    C_Vertex3f(0.432128906f, 0.323972702f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.337750673f, 0.384135365f, 0.f),
    C_Vertex3f(0.343929052f, 0.374995589f, 0.f),
    C_Vertex3f(0.344574928f, 0.474609375f, 0.f),
    C_Vertex3f(0.349077702f, 0.365432978f, 0.f),
    C_Vertex3f(0.35030365f, 0.470703125f, 0.f),
    C_Vertex3f(0.353196621f, 0.355447531f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.332749128f, 0.481445313f, 0.f),
    C_Vertex3f(0.320554972f, 0.487304688f, 0.f),
    C_Vertex3f(0.322304726f, 0.401146412f, 0.f),
    C_Vertex3f(0.330542564f, 0.392852306f, 0.f),
    C_Vertex3f(0.337750673f, 0.384135365f, 0.f),
    C_Vertex3f(0.344574928f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.284921169f, 0.423955202f, 0.f),
    C_Vertex3f(0.294931889f, 0.419808149f, 0.f),
    C_Vertex3f(0.295061588f, 0.49609375f, 0.f),
    C_Vertex3f(0.304499388f, 0.414624333f, 0.f),
    C_Vertex3f(0.307992458f, 0.4921875f, 0.f),
    C_Vertex3f(0.313623667f, 0.408403754f, 0.f),
    C_Vertex3f(0.320554972f, 0.487304688f, 0.f),
    C_Vertex3f(0.322304726f, 0.401146412f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.24050045f, 0.430175781f, 0.f),
    C_Vertex3f(0.252229691f, 0.430175781f, 0.f),
    C_Vertex3f(0.254058838f, 0.501953125f, 0.f),
    C_Vertex3f(0.26357007f, 0.429139018f, 0.f),
    C_Vertex3f(0.268094778f, 0.500976563f, 0.f),
    C_Vertex3f(0.27446723f, 0.427065492f, 0.f),
    C_Vertex3f(0.281762362f, 0.499023438f, 0.f),
    C_Vertex3f(0.284921169f, 0.423955202f, 0.f),
    C_Vertex3f(0.295061588f, 0.49609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.188468575f, 0.414624333f, 0.f),
    C_Vertex3f(0.19802928f, 0.419808149f, 0.f),
    C_Vertex3f(0.198863983f, 0.49609375f, 0.f),
    C_Vertex3f(0.208012819f, 0.423955202f, 0.f),
    C_Vertex3f(0.212137222f, 0.499023438f, 0.f),
    C_Vertex3f(0.218419194f, 0.427065492f, 0.f),
    C_Vertex3f(0.225751877f, 0.500976563f, 0.f),
    C_Vertex3f(0.229248405f, 0.429139018f, 0.f),
    C_Vertex3f(0.239707947f, 0.501953125f, 0.f),
    C_Vertex3f(0.24050045f, 0.430175781f, 0.f),
    C_Vertex3f(0.254058838f, 0.501953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.162323475f, 0.392852306f, 0.f),
    C_Vertex3f(0.170615673f, 0.401146412f, 0.f),
    C_Vertex3f(0.173341751f, 0.487304688f, 0.f),
    C_Vertex3f(0.179330707f, 0.408403754f, 0.f),
    C_Vertex3f(0.185932159f, 0.4921875f, 0.f),
    C_Vertex3f(0.188468575f, 0.414624333f, 0.f),
    C_Vertex3f(0.198863983f, 0.49609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.143666029f, 0.365432978f, 0.f),
    C_Vertex3f(0.148848653f, 0.374995589f, 0.f),
    C_Vertex3f(0.149185181f, 0.474609375f, 0.f),
    C_Vertex3f(0.155067801f, 0.384135365f, 0.f),
    C_Vertex3f(0.161092758f, 0.481445313f, 0.f),
    C_Vertex3f(0.162323475f, 0.392852306f, 0.f),
    C_Vertex3f(0.173341751f, 0.487304688f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1434021f, 0.470703125f, 0.f),
    C_Vertex3f(0.0827255249f, 0.53125f, 0.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, 0.f),
    C_Vertex3f(0.133300781f, 0.322954178f, 0.f),
    C_Vertex3f(0.134337306f, 0.334208131f, 0.f),
    C_Vertex3f(0.136410356f, 0.345039248f, 0.f),
    C_Vertex3f(0.13951993f, 0.355447531f, 0.f),
    C_Vertex3f(0.143666029f, 0.365432978f, 0.f),
    C_Vertex3f(0.149185181f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.410980225f, 0.102539063f, 0.f),
    C_Vertex3f(0.359375f, 0.311222076f, 0.f),
    C_Vertex3f(0.358338475f, 0.299871325f, 0.f),
    C_Vertex3f(0.356265426f, 0.288957238f, 0.f),
    C_Vertex3f(0.353155851f, 0.278479815f, 0.f),
    C_Vertex3f(0.35030365f, 0.163574219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349009752f, 0.268439054f, 0.f),
    C_Vertex3f(0.343827128f, 0.258834958f, 0.f),
    C_Vertex3f(0.344574928f, 0.159606934f, 0.f),
    C_Vertex3f(0.35030365f, 0.163574219f, 0.f),
    C_Vertex3f(0.353155851f, 0.278479815f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33760798f, 0.249667525f, 0.f),
    C_Vertex3f(0.330352306f, 0.240936756f, 0.f),
    C_Vertex3f(0.332749128f, 0.152664185f, 0.f),
    C_Vertex3f(0.344574928f, 0.159606934f, 0.f),
    C_Vertex3f(0.343827128f, 0.258834958f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.294646502f, 0.213980913f, 0.f),
    C_Vertex3f(0.295061588f, 0.137786865f, 0.f),
    C_Vertex3f(0.304207206f, 0.219164729f, 0.f),
    C_Vertex3f(0.307992458f, 0.14175415f, 0.f),
    C_Vertex3f(0.313345075f, 0.225385308f, 0.f),
    C_Vertex3f(0.320554972f, 0.146713257f, 0.f),
    C_Vertex3f(0.322060108f, 0.232642651f, 0.f),
    C_Vertex3f(0.332749128f, 0.152664185f, 0.f),
    C_Vertex3f(0.330352306f, 0.240936756f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.252175331f, 0.203613281f, 0.f),
    C_Vertex3f(0.254058838f, 0.131835938f, 0.f),
    C_Vertex3f(0.263427377f, 0.204650044f, 0.f),
    C_Vertex3f(0.268094778f, 0.132827759f, 0.f),
    C_Vertex3f(0.274256587f, 0.206723571f, 0.f),
    C_Vertex3f(0.281762362f, 0.134811401f, 0.f),
    C_Vertex3f(0.284662962f, 0.20983386f, 0.f),
    C_Vertex3f(0.295061588f, 0.137786865f, 0.f),
    C_Vertex3f(0.294646502f, 0.213980913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.19802928f, 0.213980913f, 0.f),
    C_Vertex3f(0.198863983f, 0.137786865f, 0.f),
    C_Vertex3f(0.208012819f, 0.20983386f, 0.f),
    C_Vertex3f(0.212137222f, 0.134811401f, 0.f),
    C_Vertex3f(0.218419194f, 0.206723571f, 0.f),
    C_Vertex3f(0.225751877f, 0.132827759f, 0.f),
    C_Vertex3f(0.229248405f, 0.204650044f, 0.f),
    C_Vertex3f(0.239707947f, 0.131835938f, 0.f),
    C_Vertex3f(0.24050045f, 0.203613281f, 0.f),
    C_Vertex3f(0.254058838f, 0.131835938f, 0.f),
    C_Vertex3f(0.252175331f, 0.203613281f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.170615673f, 0.232642651f, 0.f),
    C_Vertex3f(0.173341751f, 0.146713257f, 0.f),
    C_Vertex3f(0.179330707f, 0.225385308f, 0.f),
    C_Vertex3f(0.185932159f, 0.14175415f, 0.f),
    C_Vertex3f(0.188468575f, 0.219164729f, 0.f),
    C_Vertex3f(0.198863983f, 0.137786865f, 0.f),
    C_Vertex3f(0.19802928f, 0.213980913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.148848653f, 0.258834958f, 0.f),
    C_Vertex3f(0.149185181f, 0.159606934f, 0.f),
    C_Vertex3f(0.155067801f, 0.249667525f, 0.f),
    C_Vertex3f(0.161092758f, 0.152664185f, 0.f),
    C_Vertex3f(0.162323475f, 0.240936756f, 0.f),
    C_Vertex3f(0.173341751f, 0.146713257f, 0.f),
    C_Vertex3f(0.170615673f, 0.232642651f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.143666029f, 0.268439054f, 0.f),
    C_Vertex3f(0.13951993f, 0.278479815f, 0.f),
    C_Vertex3f(0.1434021f, 0.163574219f, 0.f),
    C_Vertex3f(0.149185181f, 0.159606934f, 0.f),
    C_Vertex3f(0.148848653f, 0.258834958f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136410356f, 0.288957238f, 0.f),
    C_Vertex3f(0.134337306f, 0.299871325f, 0.f),
    C_Vertex3f(0.1434021f, 0.163574219f, 0.f),
    C_Vertex3f(0.13951993f, 0.278479815f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0932617188f, 0.213905334f, 0.f),
    C_Vertex3f(0.0322265625f, 0.154418945f, 0.f),
    C_Vertex3f(0.0827255249f, 0.102539063f, 0.f),
    C_Vertex3f(0.1434021f, 0.163574219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.400878906f, 0.213905334f, 0.f),
    C_Vertex3f(0.410980225f, 0.102539063f, 0.f),
    C_Vertex3f(0.461914063f, 0.154418945f, 0.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, 0.f),
    C_Vertex3f(0.0827255249f, 0.53125f, 0.f),
    C_Vertex3f(0.0322265625f, 0.479415894f, 0.f),
    C_Vertex3f(0.133300781f, 0.322954178f, 0.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, 0.f),
    C_Vertex3f(0.133300781f, 0.311222076f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.134337306f, 0.299871325f, -1.f),
    C_Vertex3f(0.133300781f, 0.311222076f, -1.f),
    C_Vertex3f(0.1434021f, 0.163574219f, -1.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, -1.f),
    C_Vertex3f(0.0932617188f, 0.213905334f, -1.f),
    C_Vertex3f(0.0893554688f, 0.41391468f, -1.f),
    C_Vertex3f(0.0893554688f, 0.219920158f, -1.f),
    C_Vertex3f(0.0825195313f, 0.401680231f, -1.f),
    C_Vertex3f(0.0825195313f, 0.232168436f, -1.f),
    C_Vertex3f(0.0766601563f, 0.38924098f, -1.f),
    C_Vertex3f(0.0766601563f, 0.244635344f, -1.f),
    C_Vertex3f(0.0717773438f, 0.376596928f, -1.f),
    C_Vertex3f(0.0717773438f, 0.257320881f, -1.f),
    C_Vertex3f(0.0678710938f, 0.363748074f, -1.f),
    C_Vertex3f(0.0678710938f, 0.270225048f, -1.f),
    C_Vertex3f(0.0649414063f, 0.350694418f, -1.f),
    C_Vertex3f(0.0649414063f, 0.283347845f, -1.f),
    C_Vertex3f(0.0629882813f, 0.337435961f, -1.f),
    C_Vertex3f(0.0629882813f, 0.296689272f, -1.f),
    C_Vertex3f(0.0620117188f, 0.323972702f, -1.f),
    C_Vertex3f(0.0620117188f, 0.310249329f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.410980225f, 0.53125f, -1.f),
    C_Vertex3f(0.35030365f, 0.470703125f, -1.f),
    C_Vertex3f(0.353196621f, 0.355447531f, -1.f),
    C_Vertex3f(0.35628581f, 0.345039248f, -1.f),
    C_Vertex3f(0.35834527f, 0.334208131f, -1.f),
    C_Vertex3f(0.359375f, 0.322954178f, -1.f),
    C_Vertex3f(0.400878906f, 0.419929504f, -1.f),
    C_Vertex3f(0.461914063f, 0.479415894f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.359375f, 0.311222076f, -1.f),
    C_Vertex3f(0.410980225f, 0.102539063f, -1.f),
    C_Vertex3f(0.359375f, 0.322954178f, -1.f),
    C_Vertex3f(0.400878906f, 0.213905334f, -1.f),
    C_Vertex3f(0.400878906f, 0.419929504f, -1.f),
    C_Vertex3f(0.404785156f, 0.219920158f, -1.f),
    C_Vertex3f(0.404785156f, 0.41391468f, -1.f),
    C_Vertex3f(0.411621094f, 0.232168436f, -1.f),
    C_Vertex3f(0.411621094f, 0.401680231f, -1.f),
    C_Vertex3f(0.417480469f, 0.244635344f, -1.f),
    C_Vertex3f(0.417480469f, 0.38924098f, -1.f),
    C_Vertex3f(0.422363281f, 0.257320881f, -1.f),
    C_Vertex3f(0.422363281f, 0.376596928f, -1.f),
    C_Vertex3f(0.426269531f, 0.270225048f, -1.f),
    C_Vertex3f(0.426269531f, 0.363748074f, -1.f),
    C_Vertex3f(0.429199219f, 0.283347845f, -1.f),
    C_Vertex3f(0.429199219f, 0.350694418f, -1.f),
    C_Vertex3f(0.431152344f, 0.296689272f, -1.f),
    C_Vertex3f(0.431152344f, 0.337435961f, -1.f),
    C_Vertex3f(0.432128906f, 0.310249329f, -1.f),
    C_Vertex3f(0.432128906f, 0.323972702f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.337750673f, 0.384135365f, -1.f),
    C_Vertex3f(0.343929052f, 0.374995589f, -1.f),
    C_Vertex3f(0.344574928f, 0.474609375f, -1.f),
    C_Vertex3f(0.349077702f, 0.365432978f, -1.f),
    C_Vertex3f(0.35030365f, 0.470703125f, -1.f),
    C_Vertex3f(0.353196621f, 0.355447531f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.332749128f, 0.481445313f, -1.f),
    C_Vertex3f(0.320554972f, 0.487304688f, -1.f),
    C_Vertex3f(0.322304726f, 0.401146412f, -1.f),
    C_Vertex3f(0.330542564f, 0.392852306f, -1.f),
    C_Vertex3f(0.337750673f, 0.384135365f, -1.f),
    C_Vertex3f(0.344574928f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.284921169f, 0.423955202f, -1.f),
    C_Vertex3f(0.294931889f, 0.419808149f, -1.f),
    C_Vertex3f(0.295061588f, 0.49609375f, -1.f),
    C_Vertex3f(0.304499388f, 0.414624333f, -1.f),
    C_Vertex3f(0.307992458f, 0.4921875f, -1.f),
    C_Vertex3f(0.313623667f, 0.408403754f, -1.f),
    C_Vertex3f(0.320554972f, 0.487304688f, -1.f),
    C_Vertex3f(0.322304726f, 0.401146412f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.24050045f, 0.430175781f, -1.f),
    C_Vertex3f(0.252229691f, 0.430175781f, -1.f),
    C_Vertex3f(0.254058838f, 0.501953125f, -1.f),
    C_Vertex3f(0.26357007f, 0.429139018f, -1.f),
    C_Vertex3f(0.268094778f, 0.500976563f, -1.f),
    C_Vertex3f(0.27446723f, 0.427065492f, -1.f),
    C_Vertex3f(0.281762362f, 0.499023438f, -1.f),
    C_Vertex3f(0.284921169f, 0.423955202f, -1.f),
    C_Vertex3f(0.295061588f, 0.49609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.188468575f, 0.414624333f, -1.f),
    C_Vertex3f(0.19802928f, 0.419808149f, -1.f),
    C_Vertex3f(0.198863983f, 0.49609375f, -1.f),
    C_Vertex3f(0.208012819f, 0.423955202f, -1.f),
    C_Vertex3f(0.212137222f, 0.499023438f, -1.f),
    C_Vertex3f(0.218419194f, 0.427065492f, -1.f),
    C_Vertex3f(0.225751877f, 0.500976563f, -1.f),
    C_Vertex3f(0.229248405f, 0.429139018f, -1.f),
    C_Vertex3f(0.239707947f, 0.501953125f, -1.f),
    C_Vertex3f(0.24050045f, 0.430175781f, -1.f),
    C_Vertex3f(0.254058838f, 0.501953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.162323475f, 0.392852306f, -1.f),
    C_Vertex3f(0.170615673f, 0.401146412f, -1.f),
    C_Vertex3f(0.173341751f, 0.487304688f, -1.f),
    C_Vertex3f(0.179330707f, 0.408403754f, -1.f),
    C_Vertex3f(0.185932159f, 0.4921875f, -1.f),
    C_Vertex3f(0.188468575f, 0.414624333f, -1.f),
    C_Vertex3f(0.198863983f, 0.49609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.143666029f, 0.365432978f, -1.f),
    C_Vertex3f(0.148848653f, 0.374995589f, -1.f),
    C_Vertex3f(0.149185181f, 0.474609375f, -1.f),
    C_Vertex3f(0.155067801f, 0.384135365f, -1.f),
    C_Vertex3f(0.161092758f, 0.481445313f, -1.f),
    C_Vertex3f(0.162323475f, 0.392852306f, -1.f),
    C_Vertex3f(0.173341751f, 0.487304688f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1434021f, 0.470703125f, -1.f),
    C_Vertex3f(0.0827255249f, 0.53125f, -1.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, -1.f),
    C_Vertex3f(0.133300781f, 0.322954178f, -1.f),
    C_Vertex3f(0.134337306f, 0.334208131f, -1.f),
    C_Vertex3f(0.136410356f, 0.345039248f, -1.f),
    C_Vertex3f(0.13951993f, 0.355447531f, -1.f),
    C_Vertex3f(0.143666029f, 0.365432978f, -1.f),
    C_Vertex3f(0.149185181f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.410980225f, 0.102539063f, -1.f),
    C_Vertex3f(0.359375f, 0.311222076f, -1.f),
    C_Vertex3f(0.358338475f, 0.299871325f, -1.f),
    C_Vertex3f(0.356265426f, 0.288957238f, -1.f),
    C_Vertex3f(0.353155851f, 0.278479815f, -1.f),
    C_Vertex3f(0.35030365f, 0.163574219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349009752f, 0.268439054f, -1.f),
    C_Vertex3f(0.343827128f, 0.258834958f, -1.f),
    C_Vertex3f(0.344574928f, 0.159606934f, -1.f),
    C_Vertex3f(0.35030365f, 0.163574219f, -1.f),
    C_Vertex3f(0.353155851f, 0.278479815f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33760798f, 0.249667525f, -1.f),
    C_Vertex3f(0.330352306f, 0.240936756f, -1.f),
    C_Vertex3f(0.332749128f, 0.152664185f, -1.f),
    C_Vertex3f(0.344574928f, 0.159606934f, -1.f),
    C_Vertex3f(0.343827128f, 0.258834958f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.294646502f, 0.213980913f, -1.f),
    C_Vertex3f(0.295061588f, 0.137786865f, -1.f),
    C_Vertex3f(0.304207206f, 0.219164729f, -1.f),
    C_Vertex3f(0.307992458f, 0.14175415f, -1.f),
    C_Vertex3f(0.313345075f, 0.225385308f, -1.f),
    C_Vertex3f(0.320554972f, 0.146713257f, -1.f),
    C_Vertex3f(0.322060108f, 0.232642651f, -1.f),
    C_Vertex3f(0.332749128f, 0.152664185f, -1.f),
    C_Vertex3f(0.330352306f, 0.240936756f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.252175331f, 0.203613281f, -1.f),
    C_Vertex3f(0.254058838f, 0.131835938f, -1.f),
    C_Vertex3f(0.263427377f, 0.204650044f, -1.f),
    C_Vertex3f(0.268094778f, 0.132827759f, -1.f),
    C_Vertex3f(0.274256587f, 0.206723571f, -1.f),
    C_Vertex3f(0.281762362f, 0.134811401f, -1.f),
    C_Vertex3f(0.284662962f, 0.20983386f, -1.f),
    C_Vertex3f(0.295061588f, 0.137786865f, -1.f),
    C_Vertex3f(0.294646502f, 0.213980913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.19802928f, 0.213980913f, -1.f),
    C_Vertex3f(0.198863983f, 0.137786865f, -1.f),
    C_Vertex3f(0.208012819f, 0.20983386f, -1.f),
    C_Vertex3f(0.212137222f, 0.134811401f, -1.f),
    C_Vertex3f(0.218419194f, 0.206723571f, -1.f),
    C_Vertex3f(0.225751877f, 0.132827759f, -1.f),
    C_Vertex3f(0.229248405f, 0.204650044f, -1.f),
    C_Vertex3f(0.239707947f, 0.131835938f, -1.f),
    C_Vertex3f(0.24050045f, 0.203613281f, -1.f),
    C_Vertex3f(0.254058838f, 0.131835938f, -1.f),
    C_Vertex3f(0.252175331f, 0.203613281f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.170615673f, 0.232642651f, -1.f),
    C_Vertex3f(0.173341751f, 0.146713257f, -1.f),
    C_Vertex3f(0.179330707f, 0.225385308f, -1.f),
    C_Vertex3f(0.185932159f, 0.14175415f, -1.f),
    C_Vertex3f(0.188468575f, 0.219164729f, -1.f),
    C_Vertex3f(0.198863983f, 0.137786865f, -1.f),
    C_Vertex3f(0.19802928f, 0.213980913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.148848653f, 0.258834958f, -1.f),
    C_Vertex3f(0.149185181f, 0.159606934f, -1.f),
    C_Vertex3f(0.155067801f, 0.249667525f, -1.f),
    C_Vertex3f(0.161092758f, 0.152664185f, -1.f),
    C_Vertex3f(0.162323475f, 0.240936756f, -1.f),
    C_Vertex3f(0.173341751f, 0.146713257f, -1.f),
    C_Vertex3f(0.170615673f, 0.232642651f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.143666029f, 0.268439054f, -1.f),
    C_Vertex3f(0.13951993f, 0.278479815f, -1.f),
    C_Vertex3f(0.1434021f, 0.163574219f, -1.f),
    C_Vertex3f(0.149185181f, 0.159606934f, -1.f),
    C_Vertex3f(0.148848653f, 0.258834958f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136410356f, 0.288957238f, -1.f),
    C_Vertex3f(0.134337306f, 0.299871325f, -1.f),
    C_Vertex3f(0.1434021f, 0.163574219f, -1.f),
    C_Vertex3f(0.13951993f, 0.278479815f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0932617188f, 0.213905334f, -1.f),
    C_Vertex3f(0.0322265625f, 0.154418945f, -1.f),
    C_Vertex3f(0.0827255249f, 0.102539063f, -1.f),
    C_Vertex3f(0.1434021f, 0.163574219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.400878906f, 0.213905334f, -1.f),
    C_Vertex3f(0.410980225f, 0.102539063f, -1.f),
    C_Vertex3f(0.461914063f, 0.154418945f, -1.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, -1.f),
    C_Vertex3f(0.0827255249f, 0.53125f, -1.f),
    C_Vertex3f(0.0322265625f, 0.479415894f, -1.f),
    C_Vertex3f(0.133300781f, 0.322954178f, -1.f),
    C_Vertex3f(0.0932617188f, 0.419929504f, -1.f),
    C_Vertex3f(0.133300781f, 0.311222076f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_165[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.208496094f, 0.f, 0.f),
    C_Vertex3f(0.208496094f, 0.f, -1.f),
    C_Vertex3f(0.208496094f, 0.142578125f, -1.f),
    C_Vertex3f(0.208496094f, 0.142578125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.208496094f, 0.142578125f, 0.f),
    C_Vertex3f(0.208496094f, 0.142578125f, -1.f),
    C_Vertex3f(0.0268554688f, 0.142578125f, -1.f),
    C_Vertex3f(0.0268554688f, 0.142578125f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0268554688f, 0.142578125f, 0.f),
    C_Vertex3f(0.0268554688f, 0.142578125f, -1.f),
    C_Vertex3f(0.0268554688f, 0.203125f, -1.f),
    C_Vertex3f(0.0268554688f, 0.203125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0268554688f, 0.203125f, 0.f),
    C_Vertex3f(0.0268554688f, 0.203125f, -1.f),
    C_Vertex3f(0.208496094f, 0.203125f, -1.f),
    C_Vertex3f(0.208496094f, 0.203125f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.208496094f, 0.203125f, 0.f),
    C_Vertex3f(0.208496094f, 0.203125f, -1.f),
    C_Vertex3f(0.208496094f, 0.265625f, -1.f),
    C_Vertex3f(0.208496094f, 0.265625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.208496094f, 0.265625f, 0.f),
    C_Vertex3f(0.208496094f, 0.265625f, -1.f),
    C_Vertex3f(0.0268554688f, 0.265625f, -1.f),
    C_Vertex3f(0.0268554688f, 0.265625f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0268554688f, 0.265625f, 0.f),
    C_Vertex3f(0.0268554688f, 0.265625f, -1.f),
    C_Vertex3f(0.0268554688f, 0.330078125f, -1.f),
    C_Vertex3f(0.0268554688f, 0.330078125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0268554688f, 0.330078125f, 0.f),
    C_Vertex3f(0.0268554688f, 0.330078125f, -1.f),
    C_Vertex3f(0.175758362f, 0.330078125f, -1.f),
    C_Vertex3f(0.175758362f, 0.330078125f, 0.f),
    C_Normal3f(-0.868522644f, -0.495649397f, 0.f),
    C_Vertex3f(0.175758362f, 0.330078125f, 0.f),
    C_Vertex3f(0.175758362f, 0.330078125f, -1.f),
    C_Vertex3f(-0.00146484375f, 0.640625f, -1.f),
    C_Vertex3f(-0.00146484375f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(-0.00146484375f, 0.640625f, 0.f),
    C_Vertex3f(-0.00146484375f, 0.640625f, -1.f),
    C_Vertex3f(0.0859375f, 0.640625f, -1.f),
    C_Vertex3f(0.0859375f, 0.640625f, 0.f),
    C_Normal3f(0.871901393f, 0.489681482f, 0.f),
    C_Vertex3f(0.0859375f, 0.640625f, 0.f),
    C_Vertex3f(0.0859375f, 0.640625f, -1.f),
    C_Vertex3f(0.212623596f, 0.415054321f, -1.f),
    C_Vertex3f(0.212623596f, 0.415054321f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.871901393f, 0.489681482f, 0.f),
    C_Vertex3f(0.212623596f, 0.415054321f, 0.f),
    C_Vertex3f(0.212623596f, 0.415054321f, -1.f),
    C_Normal3f(0.882852077f, 0.469651312f, 0.f),
    C_Vertex3f(0.217905045f, 0.405399323f, 0.f),
    C_Vertex3f(0.217905045f, 0.405399323f, -1.f),
    C_Normal3f(0.890426338f, 0.455127358f, 0.f),
    C_Vertex3f(0.227365494f, 0.387342453f, 0.f),
    C_Vertex3f(0.227365494f, 0.387342453f, -1.f),
    C_Normal3f(0.900614679f, 0.434618384f, 0.f),
    C_Vertex3f(0.235723495f, 0.370538712f, 0.f),
    C_Vertex3f(0.235723495f, 0.370538712f, -1.f),
    C_Normal3f(0.910151601f, 0.414275378f, 0.f),
    C_Vertex3f(0.24297905f, 0.354988098f, 0.f),
    C_Vertex3f(0.24297905f, 0.354988098f, -1.f),
    C_Normal3f(0.918548226f, 0.395309031f, 0.f),
    C_Vertex3f(0.246055603f, 0.347839355f, 0.f),
    C_Vertex3f(0.246055603f, 0.347839355f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.909088731f, 0.416602701f, 0.f),
    C_Vertex3f(0.246055603f, 0.347839355f, 0.f),
    C_Vertex3f(0.246055603f, 0.347839355f, -1.f),
    C_Normal3f(-0.901670575f, 0.432423502f, 0.f),
    C_Vertex3f(0.248683929f, 0.353574753f, 0.f),
    C_Vertex3f(0.248683929f, 0.353574753f, -1.f),
    C_Normal3f(-0.894984782f, 0.446096689f, 0.f),
    C_Vertex3f(0.255821228f, 0.368202209f, 0.f),
    C_Vertex3f(0.255821228f, 0.368202209f, -1.f),
    C_Normal3f(-0.889279664f, 0.457363844f, 0.f),
    C_Vertex3f(0.264839172f, 0.385986328f, 0.f),
    C_Vertex3f(0.264839172f, 0.385986328f, -1.f),
    C_Normal3f(-0.880797982f, 0.473492265f, 0.f),
    C_Vertex3f(0.275737762f, 0.406927109f, 0.f),
    C_Vertex3f(0.275737762f, 0.406927109f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.880797982f, 0.473492265f, 0.f),
    C_Vertex3f(0.275737762f, 0.406927109f, 0.f),
    C_Vertex3f(0.275737762f, 0.406927109f, -1.f),
    C_Vertex3f(0.401367188f, 0.640625f, -1.f),
    C_Vertex3f(0.401367188f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.401367188f, 0.640625f, 0.f),
    C_Vertex3f(0.401367188f, 0.640625f, -1.f),
    C_Vertex3f(0.495117188f, 0.640625f, -1.f),
    C_Vertex3f(0.495117188f, 0.640625f, 0.f),
    C_Normal3f(0.869128704f, -0.494585991f, 0.f),
    C_Vertex3f(0.495117188f, 0.640625f, 0.f),
    C_Vertex3f(0.495117188f, 0.640625f, -1.f),
    C_Vertex3f(0.318397522f, 0.330078125f, -1.f),
    C_Vertex3f(0.318397522f, 0.330078125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.318397522f, 0.330078125f, 0.f),
    C_Vertex3f(0.318397522f, 0.330078125f, -1.f),
    C_Vertex3f(0.467285156f, 0.330078125f, -1.f),
    C_Vertex3f(0.467285156f, 0.330078125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.467285156f, 0.330078125f, 0.f),
    C_Vertex3f(0.467285156f, 0.330078125f, -1.f),
    C_Vertex3f(0.467285156f, 0.265625f, -1.f),
    C_Vertex3f(0.467285156f, 0.265625f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.467285156f, 0.265625f, 0.f),
    C_Vertex3f(0.467285156f, 0.265625f, -1.f),
    C_Vertex3f(0.287109375f, 0.265625f, -1.f),
    C_Vertex3f(0.287109375f, 0.265625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.287109375f, 0.265625f, 0.f),
    C_Vertex3f(0.287109375f, 0.265625f, -1.f),
    C_Vertex3f(0.287109375f, 0.203125f, -1.f),
    C_Vertex3f(0.287109375f, 0.203125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.287109375f, 0.203125f, 0.f),
    C_Vertex3f(0.287109375f, 0.203125f, -1.f),
    C_Vertex3f(0.467285156f, 0.203125f, -1.f),
    C_Vertex3f(0.467285156f, 0.203125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.467285156f, 0.203125f, 0.f),
    C_Vertex3f(0.467285156f, 0.203125f, -1.f),
    C_Vertex3f(0.467285156f, 0.142578125f, -1.f),
    C_Vertex3f(0.467285156f, 0.142578125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.467285156f, 0.142578125f, 0.f),
    C_Vertex3f(0.467285156f, 0.142578125f, -1.f),
    C_Vertex3f(0.287109375f, 0.142578125f, -1.f),
    C_Vertex3f(0.287109375f, 0.142578125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.287109375f, 0.142578125f, 0.f),
    C_Vertex3f(0.287109375f, 0.142578125f, -1.f),
    C_Vertex3f(0.287109375f, 0.f, -1.f),
    C_Vertex3f(0.287109375f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.287109375f, 0.f, 0.f),
    C_Vertex3f(0.287109375f, 0.f, -1.f),
    C_Vertex3f(0.208496094f, 0.f, -1.f),
    C_Vertex3f(0.208496094f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0268554688f, 0.203125f, 0.f),
    C_Vertex3f(0.208496094f, 0.142578125f, 0.f),
    C_Vertex3f(0.208496094f, 0.203125f, 0.f),
    C_Vertex3f(0.287109375f, 0.f, 0.f),
    C_Vertex3f(0.208496094f, 0.265625f, 0.f),
    C_Vertex3f(0.212623596f, 0.415054321f, 0.f),
    C_Vertex3f(0.175758362f, 0.330078125f, 0.f),
    C_Vertex3f(0.0859375f, 0.640625f, 0.f),
    C_Vertex3f(-0.00146484375f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.264839172f, 0.385986328f, 0.f),
    C_Vertex3f(0.287109375f, 0.f, 0.f),
    C_Vertex3f(0.275737762f, 0.406927109f, 0.f),
    C_Vertex3f(0.287109375f, 0.265625f, 0.f),
    C_Vertex3f(0.401367188f, 0.640625f, 0.f),
    C_Vertex3f(0.318397522f, 0.330078125f, 0.f),
    C_Vertex3f(0.495117188f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.287109375f, 0.142578125f, 0.f),
    C_Vertex3f(0.467285156f, 0.142578125f, 0.f),
    C_Vertex3f(0.287109375f, 0.203125f, 0.f),
    C_Vertex3f(0.287109375f, 0.265625f, 0.f),
    C_Vertex3f(0.287109375f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.287109375f, 0.f, 0.f),
    C_Vertex3f(0.264839172f, 0.385986328f, 0.f),
    C_Vertex3f(0.255821228f, 0.368202209f, 0.f),
    C_Vertex3f(0.248683929f, 0.353574753f, 0.f),
    C_Vertex3f(0.246055603f, 0.347839355f, 0.f),
    C_Vertex3f(0.24297905f, 0.354988098f, 0.f),
    C_Vertex3f(0.235723495f, 0.370538712f, 0.f),
    C_Vertex3f(0.227365494f, 0.387342453f, 0.f),
    C_Vertex3f(0.217905045f, 0.405399323f, 0.f),
    C_Vertex3f(0.212623596f, 0.415054321f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.318397522f, 0.330078125f, 0.f),
    C_Vertex3f(0.287109375f, 0.265625f, 0.f),
    C_Vertex3f(0.467285156f, 0.265625f, 0.f),
    C_Vertex3f(0.467285156f, 0.330078125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0268554688f, 0.330078125f, 0.f),
    C_Vertex3f(0.0268554688f, 0.265625f, 0.f),
    C_Vertex3f(0.208496094f, 0.265625f, 0.f),
    C_Vertex3f(0.175758362f, 0.330078125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.287109375f, 0.203125f, 0.f),
    C_Vertex3f(0.467285156f, 0.142578125f, 0.f),
    C_Vertex3f(0.467285156f, 0.203125f, 0.f),
    C_Vertex3f(0.287109375f, 0.f, 0.f),
    C_Vertex3f(0.208496094f, 0.142578125f, 0.f),
    C_Vertex3f(0.208496094f, 0.f, 0.f),
    C_Vertex3f(0.208496094f, 0.142578125f, 0.f),
    C_Vertex3f(0.0268554688f, 0.203125f, 0.f),
    C_Vertex3f(0.0268554688f, 0.142578125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0268554688f, 0.203125f, -1.f),
    C_Vertex3f(0.208496094f, 0.142578125f, -1.f),
    C_Vertex3f(0.208496094f, 0.203125f, -1.f),
    C_Vertex3f(0.287109375f, 0.f, -1.f),
    C_Vertex3f(0.208496094f, 0.265625f, -1.f),
    C_Vertex3f(0.212623596f, 0.415054321f, -1.f),
    C_Vertex3f(0.175758362f, 0.330078125f, -1.f),
    C_Vertex3f(0.0859375f, 0.640625f, -1.f),
    C_Vertex3f(-0.00146484375f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.264839172f, 0.385986328f, -1.f),
    C_Vertex3f(0.287109375f, 0.f, -1.f),
    C_Vertex3f(0.275737762f, 0.406927109f, -1.f),
    C_Vertex3f(0.287109375f, 0.265625f, -1.f),
    C_Vertex3f(0.401367188f, 0.640625f, -1.f),
    C_Vertex3f(0.318397522f, 0.330078125f, -1.f),
    C_Vertex3f(0.495117188f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.287109375f, 0.142578125f, -1.f),
    C_Vertex3f(0.467285156f, 0.142578125f, -1.f),
    C_Vertex3f(0.287109375f, 0.203125f, -1.f),
    C_Vertex3f(0.287109375f, 0.265625f, -1.f),
    C_Vertex3f(0.287109375f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.287109375f, 0.f, -1.f),
    C_Vertex3f(0.264839172f, 0.385986328f, -1.f),
    C_Vertex3f(0.255821228f, 0.368202209f, -1.f),
    C_Vertex3f(0.248683929f, 0.353574753f, -1.f),
    C_Vertex3f(0.246055603f, 0.347839355f, -1.f),
    C_Vertex3f(0.24297905f, 0.354988098f, -1.f),
    C_Vertex3f(0.235723495f, 0.370538712f, -1.f),
    C_Vertex3f(0.227365494f, 0.387342453f, -1.f),
    C_Vertex3f(0.217905045f, 0.405399323f, -1.f),
    C_Vertex3f(0.212623596f, 0.415054321f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.318397522f, 0.330078125f, -1.f),
    C_Vertex3f(0.287109375f, 0.265625f, -1.f),
    C_Vertex3f(0.467285156f, 0.265625f, -1.f),
    C_Vertex3f(0.467285156f, 0.330078125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0268554688f, 0.330078125f, -1.f),
    C_Vertex3f(0.0268554688f, 0.265625f, -1.f),
    C_Vertex3f(0.208496094f, 0.265625f, -1.f),
    C_Vertex3f(0.175758362f, 0.330078125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.287109375f, 0.203125f, -1.f),
    C_Vertex3f(0.467285156f, 0.142578125f, -1.f),
    C_Vertex3f(0.467285156f, 0.203125f, -1.f),
    C_Vertex3f(0.287109375f, 0.f, -1.f),
    C_Vertex3f(0.208496094f, 0.142578125f, -1.f),
    C_Vertex3f(0.208496094f, 0.f, -1.f),
    C_Vertex3f(0.208496094f, 0.142578125f, -1.f),
    C_Vertex3f(0.0268554688f, 0.203125f, -1.f),
    C_Vertex3f(0.0268554688f, 0.142578125f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_166[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.150878906f, 0.651367188f, 0.f),
    C_Vertex3f(0.150878906f, 0.651367188f, -1.f),
    C_Vertex3f(0.150878906f, 0.306152344f, -1.f),
    C_Vertex3f(0.150878906f, 0.306152344f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.150878906f, 0.306152344f, 0.f),
    C_Vertex3f(0.150878906f, 0.306152344f, -1.f),
    C_Vertex3f(0.08203125f, 0.306152344f, -1.f),
    C_Vertex3f(0.08203125f, 0.306152344f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.08203125f, 0.306152344f, 0.f),
    C_Vertex3f(0.08203125f, 0.306152344f, -1.f),
    C_Vertex3f(0.08203125f, 0.651367188f, -1.f),
    C_Vertex3f(0.08203125f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.08203125f, 0.651367188f, 0.f),
    C_Vertex3f(0.08203125f, 0.651367188f, -1.f),
    C_Vertex3f(0.150878906f, 0.651367188f, -1.f),
    C_Vertex3f(0.150878906f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.150878906f, 0.157226563f, 0.f),
    C_Vertex3f(0.150878906f, 0.157226563f, -1.f),
    C_Vertex3f(0.150878906f, -0.188476563f, -1.f),
    C_Vertex3f(0.150878906f, -0.188476563f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.150878906f, -0.188476563f, 0.f),
    C_Vertex3f(0.150878906f, -0.188476563f, -1.f),
    C_Vertex3f(0.08203125f, -0.188476563f, -1.f),
    C_Vertex3f(0.08203125f, -0.188476563f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.08203125f, -0.188476563f, 0.f),
    C_Vertex3f(0.08203125f, -0.188476563f, -1.f),
    C_Vertex3f(0.08203125f, 0.157226563f, -1.f),
    C_Vertex3f(0.08203125f, 0.157226563f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.08203125f, 0.157226563f, 0.f),
    C_Vertex3f(0.08203125f, 0.157226563f, -1.f),
    C_Vertex3f(0.150878906f, 0.157226563f, -1.f),
    C_Vertex3f(0.150878906f, 0.157226563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.08203125f, 0.651367188f, 0.f),
    C_Vertex3f(0.08203125f, 0.306152344f, 0.f),
    C_Vertex3f(0.150878906f, 0.306152344f, 0.f),
    C_Vertex3f(0.150878906f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.08203125f, 0.157226563f, 0.f),
    C_Vertex3f(0.08203125f, -0.188476563f, 0.f),
    C_Vertex3f(0.150878906f, -0.188476563f, 0.f),
    C_Vertex3f(0.150878906f, 0.157226563f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.08203125f, 0.651367188f, -1.f),
    C_Vertex3f(0.08203125f, 0.306152344f, -1.f),
    C_Vertex3f(0.150878906f, 0.306152344f, -1.f),
    C_Vertex3f(0.150878906f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.08203125f, 0.157226563f, -1.f),
    C_Vertex3f(0.08203125f, -0.188476563f, -1.f),
    C_Vertex3f(0.150878906f, -0.188476563f, -1.f),
    C_Vertex3f(0.150878906f, 0.157226563f, -1.f),
    C_End(),
    C_Translatef(0.232421875f, 0.f, 0.f),
};
static const Call char_167[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.140565678f, 0.990071356f, 0.f),
    C_Vertex3f(0.0625f, -0.0302734375f, 0.f),
    C_Vertex3f(0.0625f, -0.0302734375f, -1.f),
    C_Vertex3f(0.141601563f, -0.0190429688f, -1.f),
    C_Vertex3f(0.141601563f, -0.0190429688f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.977565169f, 0.210633412f, 0.f),
    C_Vertex3f(0.141601563f, -0.0190429688f, 0.f),
    C_Vertex3f(0.141601563f, -0.0190429688f, -1.f),
    C_Normal3f(0.970369279f, 0.24162668f, 0.f),
    C_Vertex3f(0.143129349f, -0.0261335373f, 0.f),
    C_Vertex3f(0.143129349f, -0.0261335373f, -1.f),
    C_Normal3f(0.958822608f, 0.284005672f, 0.f),
    C_Vertex3f(0.146635175f, -0.039257884f, 0.f),
    C_Vertex3f(0.146635175f, -0.039257884f, -1.f),
    C_Normal3f(0.940181494f, 0.340674043f, 0.f),
    C_Vertex3f(0.150591254f, -0.0513254404f, 0.f),
    C_Vertex3f(0.150591254f, -0.0513254404f, -1.f),
    C_Normal3f(0.914695859f, 0.404143035f, 0.f),
    C_Vertex3f(0.154997587f, -0.0623362064f, 0.f),
    C_Vertex3f(0.154997587f, -0.0623362064f, -1.f),
    C_Normal3f(0.880221426f, 0.474563122f, 0.f),
    C_Vertex3f(0.159854174f, -0.0722901821f, 0.f),
    C_Vertex3f(0.159854174f, -0.0722901821f, -1.f),
    C_Normal3f(0.834216654f, 0.551436782f, 0.f),
    C_Vertex3f(0.165161014f, -0.0811873674f, 0.f),
    C_Vertex3f(0.165161014f, -0.0811873674f, -1.f),
    C_Normal3f(0.77397418f, 0.633217156f, 0.f),
    C_Vertex3f(0.170918107f, -0.0890277624f, 0.f),
    C_Vertex3f(0.170918107f, -0.0890277624f, -1.f),
    C_Normal3f(0.688508451f, 0.72522819f, 0.f),
    C_Vertex3f(0.177125454f, -0.095811367f, 0.f),
    C_Vertex3f(0.177125454f, -0.095811367f, -1.f),
    C_Normal3f(0.555717766f, 0.831371069f, 0.f),
    C_Vertex3f(0.187111855f, -0.104401588f, 0.f),
    C_Vertex3f(0.187111855f, -0.104401588f, -1.f),
    C_Normal3f(0.389710724f, 0.920937359f, 0.f),
    C_Vertex3f(0.20282793f, -0.11299181f, 0.f),
    C_Vertex3f(0.20282793f, -0.11299181f, -1.f),
    C_Normal3f(0.217043981f, 0.976161778f, 0.f),
    C_Vertex3f(0.220944881f, -0.118718624f, 0.f),
    C_Vertex3f(0.220944881f, -0.118718624f, -1.f),
    C_Normal3f(0.0662743449f, 0.997801363f, 0.f),
    C_Vertex3f(0.241462708f, -0.121582031f, 0.f),
    C_Vertex3f(0.241462708f, -0.121582031f, -1.f),
    C_Normal3f(-0.0667072982f, 0.997772574f, 0.f),
    C_Vertex3f(0.264055252f, -0.121582031f, 0.f),
    C_Vertex3f(0.264055252f, -0.121582031f, -1.f),
    C_Normal3f(-0.222745165f, 0.974876702f, 0.f),
    C_Vertex3f(0.283892632f, -0.118745327f, 0.f),
    C_Vertex3f(0.283892632f, -0.118745327f, -1.f),
    C_Normal3f(-0.401148498f, 0.916013002f, 0.f),
    C_Vertex3f(0.301301003f, -0.113071918f, 0.f),
    C_Vertex3f(0.301301003f, -0.113071918f, -1.f),
    C_Normal3f(-0.584993958f, 0.81103766f, 0.f),
    C_Vertex3f(0.316280365f, -0.104561806f, 0.f),
    C_Vertex3f(0.316280365f, -0.104561806f, -1.f),
    C_Normal3f(-0.733107388f, 0.680112898f, 0.f),
    C_Vertex3f(0.328830719f, -0.0932149887f, 0.f),
    C_Vertex3f(0.328830719f, -0.0932149887f, -1.f),
    C_Normal3f(-0.852965236f, 0.521967828f, 0.f),
    C_Vertex3f(0.338243484f, -0.0808873177f, 0.f),
    C_Vertex3f(0.338243484f, -0.0808873177f, -1.f),
    C_Normal3f(-0.946464181f, 0.322808772f, 0.f),
    C_Vertex3f(0.344518661f, -0.0675787926f, 0.f),
    C_Vertex3f(0.344518661f, -0.0675787926f, -1.f),
    C_Normal3f(-0.993549526f, 0.113398828f, 0.f),
    C_Vertex3f(0.34765625f, -0.0532894135f, 0.f),
    C_Vertex3f(0.34765625f, -0.0532894135f, -1.f),
    C_Normal3f(-0.996656537f, -0.0817055479f, 0.f),
    C_Vertex3f(0.34765625f, -0.0400886536f, 0.f),
    C_Vertex3f(0.34765625f, -0.0400886536f, -1.f),
    C_Normal3f(-0.962836027f, -0.270086735f, 0.f),
    C_Vertex3f(0.345697403f, -0.0293951035f, 0.f),
    C_Vertex3f(0.345697403f, -0.0293951035f, -1.f),
    C_Normal3f(-0.89873153f, -0.438499361f, 0.f),
    C_Vertex3f(0.341779709f, -0.0191392899f, 0.f),
    C_Vertex3f(0.341779709f, -0.0191392899f, -1.f),
    C_Normal3f(-0.816155255f, -0.577832758f, 0.f),
    C_Vertex3f(0.335903168f, -0.00932121277f, 0.f),
    C_Vertex3f(0.335903168f, -0.00932121277f, -1.f),
    C_Normal3f(-0.732208908f, -0.681080103f, 0.f),
    C_Vertex3f(0.328611374f, -0.000539779663f, 0.f),
    C_Vertex3f(0.328611374f, -0.000539779663f, -1.f),
    C_Normal3f(-0.671801627f, -0.74073112f, 0.f),
    C_Vertex3f(0.318918228f, 0.00893878937f, 0.f),
    C_Vertex3f(0.318918228f, 0.00893878937f, -1.f),
    C_Normal3f(-0.629191458f, -0.777250409f, 0.f),
    C_Vertex3f(0.306280136f, 0.0197134018f, 0.f),
    C_Vertex3f(0.306280136f, 0.0197134018f, -1.f),
    C_Normal3f(-0.597803116f, -0.801642954f, 0.f),
    C_Vertex3f(0.290697098f, 0.0317840576f, 0.f),
    C_Vertex3f(0.290697098f, 0.0317840576f, -1.f),
    C_Normal3f(-0.573857963f, -0.818954766f, 0.f),
    C_Vertex3f(0.272169113f, 0.0451507568f, 0.f),
    C_Vertex3f(0.272169113f, 0.0451507568f, -1.f),
    C_Normal3f(-0.555047214f, -0.831818879f, 0.f),
    C_Vertex3f(0.250696182f, 0.0598134995f, 0.f),
    C_Vertex3f(0.250696182f, 0.0598134995f, -1.f),
    C_Normal3f(-0.539906442f, -0.841725051f, 0.f),
    C_Vertex3f(0.226278305f, 0.0757722855f, 0.f),
    C_Vertex3f(0.226278305f, 0.0757722855f, -1.f),
    C_Normal3f(-0.528552353f, -0.848900676f, 0.f),
    C_Vertex3f(0.198915482f, 0.0930271149f, 0.f),
    C_Vertex3f(0.198915482f, 0.0930271149f, -1.f),
    C_Normal3f(-0.526703596f, -0.850048959f, 0.f),
    C_Vertex3f(0.175694466f, 0.10726738f, 0.f),
    C_Vertex3f(0.175694466f, 0.10726738f, -1.f),
    C_Normal3f(-0.537538052f, -0.843239427f, 0.f),
    C_Vertex3f(0.160776973f, 0.11665833f, 0.f),
    C_Vertex3f(0.160776973f, 0.11665833f, -1.f),
    C_Normal3f(-0.548256457f, -0.836310267f, 0.f),
    C_Vertex3f(0.147076249f, 0.125510573f, 0.f),
    C_Vertex3f(0.147076249f, 0.125510573f, -1.f),
    C_Normal3f(-0.560821474f, -0.827936769f, 0.f),
    C_Vertex3f(0.134592295f, 0.13382411f, 0.f),
    C_Vertex3f(0.134592295f, 0.13382411f, -1.f),
    C_Normal3f(-0.575741589f, -0.817631602f, 0.f),
    C_Vertex3f(0.123325109f, 0.14159894f, 0.f),
    C_Vertex3f(0.123325109f, 0.14159894f, -1.f),
    C_Normal3f(-0.593722761f, -0.804669738f, 0.f),
    C_Vertex3f(0.113274693f, 0.148835063f, 0.f),
    C_Vertex3f(0.113274693f, 0.148835063f, -1.f),
    C_Normal3f(-0.615767598f, -0.787927866f, 0.f),
    C_Vertex3f(0.104441047f, 0.155532479f, 0.f),
    C_Vertex3f(0.104441047f, 0.155532479f, -1.f),
    C_Normal3f(-0.646470189f, -0.762939274f, 0.f),
    C_Vertex3f(0.0968241692f, 0.161691189f, 0.f),
    C_Vertex3f(0.0968241692f, 0.161691189f, -1.f),
    C_Normal3f(-0.697850525f, -0.716243386f, 0.f),
    C_Vertex3f(0.0872240067f, 0.170121193f, 0.f),
    C_Vertex3f(0.0872240067f, 0.170121193f, -1.f),
    C_Normal3f(-0.756197751f, -0.654343188f, 0.f),
    C_Vertex3f(0.0756001472f, 0.182370186f, 0.f),
    C_Vertex3f(0.0756001472f, 0.182370186f, -1.f),
    C_Normal3f(-0.812922955f, -0.582371294f, 0.f),
    C_Vertex3f(0.0651526451f, 0.195628166f, 0.f),
    C_Vertex3f(0.0651526451f, 0.195628166f, -1.f),
    C_Normal3f(-0.862090528f, -0.506754398f, 0.f),
    C_Vertex3f(0.0558815002f, 0.209895134f, 0.f),
    C_Vertex3f(0.0558815002f, 0.209895134f, -1.f),
    C_Normal3f(-0.908922374f, -0.416965455f, 0.f),
    C_Vertex3f(0.0477867126f, 0.225171089f, 0.f),
    C_Vertex3f(0.0477867126f, 0.225171089f, -1.f),
    C_Normal3f(-0.952183723f, -0.30552581f, 0.f),
    C_Vertex3f(0.0417156219f, 0.240774632f, 0.f),
    C_Vertex3f(0.0417156219f, 0.240774632f, -1.f),
    C_Normal3f(-0.982675374f, -0.185335189f, 0.f),
    C_Vertex3f(0.0376682281f, 0.256705761f, 0.f),
    C_Vertex3f(0.0376682281f, 0.256705761f, -1.f),
    C_Normal3f(-0.998335302f, -0.05767598f, 0.f),
    C_Vertex3f(0.0356445313f, 0.272964478f, 0.f),
    C_Vertex3f(0.0356445313f, 0.272964478f, -1.f),
    C_Normal3f(-0.996427476f, 0.0844521523f, 0.f),
    C_Vertex3f(0.0356445313f, 0.291734695f, 0.f),
    C_Vertex3f(0.0356445313f, 0.291734695f, -1.f),
    C_Normal3f(-0.969489813f, 0.245131344f, 0.f),
    C_Vertex3f(0.0389280319f, 0.311705589f, 0.f),
    C_Vertex3f(0.0389280319f, 0.311705589f, -1.f),
    C_Normal3f(-0.914026201f, 0.405655235f, 0.f),
    C_Vertex3f(0.0454950333f, 0.330693245f, 0.f),
    C_Vertex3f(0.0454950333f, 0.330693245f, -1.f),
    C_Normal3f(-0.836059332f, 0.548639059f, 0.f),
    C_Vertex3f(0.0553455353f, 0.348697662f, 0.f),
    C_Vertex3f(0.0553455353f, 0.348697662f, -1.f),
    C_Normal3f(-0.738605142f, 0.674138308f, 0.f),
    C_Vertex3f(0.068479538f, 0.365718842f, 0.f),
    C_Vertex3f(0.068479538f, 0.365718842f, -1.f),
    C_Normal3f(-0.616465867f, 0.787381589f, 0.f),
    C_Vertex3f(0.0840768814f, 0.380176544f, 0.f),
    C_Vertex3f(0.0840768814f, 0.380176544f, -1.f),
    C_Normal3f(-0.481977791f, 0.87618351f, 0.f),
    C_Vertex3f(0.102137566f, 0.39207077f, 0.f),
    C_Vertex3f(0.102137566f, 0.39207077f, -1.f),
    C_Normal3f(-0.369069874f, 0.929401636f, 0.f),
    C_Vertex3f(0.122661591f, 0.40140152f, 0.f),
    C_Vertex3f(0.122661591f, 0.40140152f, -1.f),
    C_Normal3f(-0.282407641f, 0.959294498f, 0.f),
    C_Vertex3f(0.134155273f, 0.404785156f, 0.f),
    C_Vertex3f(0.134155273f, 0.404785156f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.69096899f, -0.722884417f, 0.f),
    C_Vertex3f(0.134155273f, 0.404785156f, 0.f),
    C_Vertex3f(0.134155273f, 0.404785156f, -1.f),
    C_Normal3f(-0.718517959f, -0.695508361f, 0.f),
    C_Vertex3f(0.126379013f, 0.412218094f, 0.f),
    C_Vertex3f(0.126379013f, 0.412218094f, -1.f),
    C_Normal3f(-0.756612718f, -0.653863311f, 0.f),
    C_Vertex3f(0.113126278f, 0.426509857f, 0.f),
    C_Vertex3f(0.113126278f, 0.426509857f, -1.f),
    C_Normal3f(-0.807723582f, -0.589561343f, 0.f),
    C_Vertex3f(0.102173328f, 0.440227509f, 0.f),
    C_Vertex3f(0.102173328f, 0.440227509f, -1.f),
    C_Normal3f(-0.863671958f, -0.504054248f, 0.f),
    C_Vertex3f(0.0935201645f, 0.453371048f, 0.f),
    C_Vertex3f(0.0935201645f, 0.453371048f, -1.f),
    C_Normal3f(-0.916936457f, -0.399033368f, 0.f),
    C_Vertex3f(0.0871667862f, 0.465940475f, 0.f),
    C_Vertex3f(0.0871667862f, 0.465940475f, -1.f),
    C_Normal3f(-0.957516015f, -0.288380265f, 0.f),
    C_Vertex3f(0.0824017525f, 0.478919983f, 0.f),
    C_Vertex3f(0.0824017525f, 0.478919983f, -1.f),
    C_Normal3f(-0.984987974f, -0.172623366f, 0.f),
    C_Vertex3f(0.0792250633f, 0.49230957f, 0.f),
    C_Vertex3f(0.0792250633f, 0.49230957f, -1.f),
    C_Normal3f(-0.998383343f, -0.0568384007f, 0.f),
    C_Vertex3f(0.0776367188f, 0.506109238f, 0.f),
    C_Vertex3f(0.0776367188f, 0.506109238f, -1.f),
    C_Normal3f(-0.998783112f, 0.0493178889f, 0.f),
    C_Vertex3f(0.0776367188f, 0.520209312f, 0.f),
    C_Vertex3f(0.0776367188f, 0.520209312f, -1.f),
    C_Normal3f(-0.988360465f, 0.152130112f, 0.f),
    C_Vertex3f(0.0790008307f, 0.533735156f, 0.f),
    C_Vertex3f(0.0790008307f, 0.533735156f, -1.f),
    C_Normal3f(-0.966437042f, 0.256903827f, 0.f),
    C_Vertex3f(0.0817290545f, 0.546796441f, 0.f),
    C_Vertex3f(0.0817290545f, 0.546796441f, -1.f),
    C_Normal3f(-0.932868659f, 0.360216677f, 0.f),
    C_Vertex3f(0.0858213902f, 0.559393167f, 0.f),
    C_Vertex3f(0.0858213902f, 0.559393167f, -1.f),
    C_Normal3f(-0.888723373f, 0.45844391f, 0.f),
    C_Vertex3f(0.0912778378f, 0.571525335f, 0.f),
    C_Vertex3f(0.0912778378f, 0.571525335f, -1.f),
    C_Normal3f(-0.836108983f, 0.548563421f, 0.f),
    C_Vertex3f(0.0980983973f, 0.583192945f, 0.f),
    C_Vertex3f(0.0980983973f, 0.583192945f, -1.f),
    C_Normal3f(-0.777742922f, 0.628582478f, 0.f),
    C_Vertex3f(0.106283069f, 0.594395995f, 0.f),
    C_Vertex3f(0.106283069f, 0.594395995f, -1.f),
    C_Normal3f(-0.716433525f, 0.69765538f, 0.f),
    C_Vertex3f(0.115831852f, 0.605134487f, 0.f),
    C_Vertex3f(0.115831852f, 0.605134487f, -1.f),
    C_Normal3f(-0.646827817f, 0.762636006f, 0.f),
    C_Vertex3f(0.126744747f, 0.615408421f, 0.f),
    C_Vertex3f(0.126744747f, 0.615408421f, -1.f),
    C_Normal3f(-0.563337207f, 0.826227129f, 0.f),
    C_Vertex3f(0.13854444f, 0.624398112f, 0.f),
    C_Vertex3f(0.13854444f, 0.624398112f, -1.f),
    C_Normal3f(-0.473756015f, 0.880656183f, 0.f),
    C_Vertex3f(0.151230931f, 0.632103562f, 0.f),
    C_Vertex3f(0.151230931f, 0.632103562f, -1.f),
    C_Normal3f(-0.381173313f, 0.924503624f, 0.f),
    C_Vertex3f(0.16480422f, 0.638524771f, 0.f),
    C_Vertex3f(0.16480422f, 0.638524771f, -1.f),
    C_Normal3f(-0.288750261f, 0.957404435f, 0.f),
    C_Vertex3f(0.179264307f, 0.643661737f, 0.f),
    C_Vertex3f(0.179264307f, 0.643661737f, -1.f),
    C_Normal3f(-0.199250817f, 0.97994858f, 0.f),
    C_Vertex3f(0.194611192f, 0.647514462f, 0.f),
    C_Vertex3f(0.194611192f, 0.647514462f, -1.f),
    C_Normal3f(-0.1147466f, 0.993394792f, 0.f),
    C_Vertex3f(0.210844874f, 0.650082946f, 0.f),
    C_Vertex3f(0.210844874f, 0.650082946f, -1.f),
    C_Normal3f(-0.0353297256f, 0.999375641f, 0.f),
    C_Vertex3f(0.227965355f, 0.651367188f, 0.f),
    C_Vertex3f(0.227965355f, 0.651367188f, -1.f),
    C_Normal3f(0.0316018462f, 0.999500513f, 0.f),
    C_Vertex3f(0.247172356f, 0.651367188f, 0.f),
    C_Vertex3f(0.247172356f, 0.651367188f, -1.f),
    C_Normal3f(0.0970199183f, 0.995282471f, 0.f),
    C_Vertex3f(0.266460538f, 0.650150061f, 0.f),
    C_Vertex3f(0.266460538f, 0.650150061f, -1.f),
    C_Normal3f(0.170262635f, 0.985398769f, 0.f),
    C_Vertex3f(0.284630179f, 0.647715807f, 0.f),
    C_Vertex3f(0.284630179f, 0.647715807f, -1.f),
    C_Normal3f(0.250097513f, 0.968220711f, 0.f),
    C_Vertex3f(0.30168128f, 0.644064426f, 0.f),
    C_Vertex3f(0.30168128f, 0.644064426f, -1.f),
    C_Normal3f(0.335608631f, 0.942001462f, 0.f),
    C_Vertex3f(0.31761384f, 0.639195919f, 0.f),
    C_Vertex3f(0.31761384f, 0.639195919f, -1.f),
    C_Normal3f(0.425073266f, 0.905158877f, 0.f),
    C_Vertex3f(0.332427859f, 0.633110285f, 0.f),
    C_Vertex3f(0.332427859f, 0.633110285f, -1.f),
    C_Normal3f(0.515914381f, 0.85664016f, 0.f),
    C_Vertex3f(0.346123338f, 0.625807524f, 0.f),
    C_Vertex3f(0.346123338f, 0.625807524f, -1.f),
    C_Normal3f(0.604874253f, 0.796321034f, 0.f),
    C_Vertex3f(0.358700275f, 0.617287636f, 0.f),
    C_Vertex3f(0.358700275f, 0.617287636f, -1.f),
    C_Normal3f(0.686739326f, 0.726903737f, 0.f),
    C_Vertex3f(0.370158672f, 0.607550621f, 0.f),
    C_Vertex3f(0.370158672f, 0.607550621f, -1.f),
    C_Normal3f(0.759682f, 0.650294721f, 0.f),
    C_Vertex3f(0.380471349f, 0.596719503f, 0.f),
    C_Vertex3f(0.380471349f, 0.596719503f, -1.f),
    C_Normal3f(0.823443234f, 0.567398608f, 0.f),
    C_Vertex3f(0.389638305f, 0.584794283f, 0.f),
    C_Vertex3f(0.389638305f, 0.584794283f, -1.f),
    C_Normal3f(0.87657398f, 0.481267273f, 0.f),
    C_Vertex3f(0.39765954f, 0.57177496f, 0.f),
    C_Vertex3f(0.39765954f, 0.57177496f, -1.f),
    C_Normal3f(0.91869992f, 0.394956321f, 0.f),
    C_Vertex3f(0.404535055f, 0.557661533f, 0.f),
    C_Vertex3f(0.404535055f, 0.557661533f, -1.f),
    C_Normal3f(0.950381219f, 0.311087519f, 0.f),
    C_Vertex3f(0.41026485f, 0.542454004f, 0.f),
    C_Vertex3f(0.41026485f, 0.542454004f, -1.f),
    C_Normal3f(0.972811043f, 0.231599882f, 0.f),
    C_Vertex3f(0.414848924f, 0.526152372f, 0.f),
    C_Vertex3f(0.414848924f, 0.526152372f, -1.f),
    C_Normal3f(0.985512912f, 0.169599876f, 0.f),
    C_Vertex3f(0.418287277f, 0.508756638f, 0.f),
    C_Vertex3f(0.418287277f, 0.508756638f, -1.f),
    C_Normal3f(0.992400289f, 0.123051897f, 0.f),
    C_Vertex3f(0.419433594f, 0.499511719f, 0.f),
    C_Vertex3f(0.419433594f, 0.499511719f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.101273037f, -0.994858682f, 0.f),
    C_Vertex3f(0.419433594f, 0.499511719f, 0.f),
    C_Vertex3f(0.419433594f, 0.499511719f, -1.f),
    C_Vertex3f(0.337890625f, 0.491210938f, -1.f),
    C_Vertex3f(0.337890625f, 0.491210938f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.985702336f, -0.168496132f, 0.f),
    C_Vertex3f(0.337890625f, 0.491210938f, 0.f),
    C_Vertex3f(0.337890625f, 0.491210938f, -1.f),
    C_Normal3f(-0.98009795f, -0.198514491f, 0.f),
    C_Vertex3f(0.336746216f, 0.497905731f, 0.f),
    C_Vertex3f(0.336746216f, 0.497905731f, -1.f),
    C_Normal3f(-0.970796347f, -0.239905074f, 0.f),
    C_Vertex3f(0.334028244f, 0.510280132f, 0.f),
    C_Vertex3f(0.334028244f, 0.510280132f, -1.f),
    C_Normal3f(-0.95521462f, -0.295913994f, 0.f),
    C_Vertex3f(0.330881119f, 0.521639347f, 0.f),
    C_Vertex3f(0.330881119f, 0.521639347f, -1.f),
    C_Normal3f(-0.933103502f, -0.359607965f, 0.f),
    C_Vertex3f(0.32730484f, 0.531983376f, 0.f),
    C_Vertex3f(0.32730484f, 0.531983376f, -1.f),
    C_Normal3f(-0.902088404f, -0.431551248f, 0.f),
    C_Vertex3f(0.323299408f, 0.541312218f, 0.f),
    C_Vertex3f(0.323299408f, 0.541312218f, -1.f),
    C_Normal3f(-0.85916239f, -0.511702955f, 0.f),
    C_Vertex3f(0.318864822f, 0.549625874f, 0.f),
    C_Vertex3f(0.318864822f, 0.549625874f, -1.f),
    C_Normal3f(-0.800839365f, -0.598879278f, 0.f),
    C_Vertex3f(0.314001083f, 0.556924343f, 0.f),
    C_Vertex3f(0.314001083f, 0.556924343f, -1.f),
    C_Normal3f(-0.714861453f, -0.699266136f, 0.f),
    C_Vertex3f(0.308708191f, 0.563207626f, 0.f),
    C_Vertex3f(0.308708191f, 0.563207626f, -1.f),
    C_Normal3f(-0.583147228f, -0.812366486f, 0.f),
    C_Vertex3f(0.300125122f, 0.571109772f, 0.f),
    C_Vertex3f(0.300125122f, 0.571109772f, -1.f),
    C_Normal3f(-0.415206879f, -0.909726977f, 0.f),
    C_Vertex3f(0.286691666f, 0.579011917f, 0.f),
    C_Vertex3f(0.286691666f, 0.579011917f, -1.f),
    C_Normal3f(-0.233981669f, -0.972240984f, 0.f),
    C_Vertex3f(0.271268845f, 0.584280014f, 0.f),
    C_Vertex3f(0.271268845f, 0.584280014f, -1.f),
    C_Normal3f(-0.0707446486f, -0.997494459f, 0.f),
    C_Vertex3f(0.253856659f, 0.586914063f, 0.f),
    C_Vertex3f(0.253856659f, 0.586914063f, -1.f),
    C_Normal3f(0.0714551359f, -0.997443855f, 0.f),
    C_Vertex3f(0.234128952f, 0.586914063f, 0.f),
    C_Vertex3f(0.234128952f, 0.586914063f, -1.f),
    C_Normal3f(0.230714098f, -0.973021626f, 0.f),
    C_Vertex3f(0.216009617f, 0.584202766f, 0.f),
    C_Vertex3f(0.216009617f, 0.584202766f, -1.f),
    C_Normal3f(0.406884611f, -0.913479567f, 0.f),
    C_Vertex3f(0.19982481f, 0.578780174f, 0.f),
    C_Vertex3f(0.19982481f, 0.578780174f, -1.f),
    C_Normal3f(0.581305861f, -0.813685119f, 0.f),
    C_Vertex3f(0.185574532f, 0.570646286f, 0.f),
    C_Vertex3f(0.185574532f, 0.570646286f, -1.f),
    C_Normal3f(0.718515575f, -0.695510924f, 0.f),
    C_Vertex3f(0.173258781f, 0.559801102f, 0.f),
    C_Vertex3f(0.173258781f, 0.559801102f, -1.f),
    C_Normal3f(0.835585654f, -0.549360216f, 0.f),
    C_Vertex3f(0.164021969f, 0.548380852f, 0.f),
    C_Vertex3f(0.164021969f, 0.548380852f, -1.f),
    C_Normal3f(0.936019659f, -0.351947606f, 0.f),
    C_Vertex3f(0.157864094f, 0.536385536f, 0.f),
    C_Vertex3f(0.157864094f, 0.536385536f, -1.f),
    C_Normal3f(0.99240458f, -0.123017013f, 0.f),
    C_Vertex3f(0.154785156f, 0.523815155f, 0.f),
    C_Vertex3f(0.154785156f, 0.523815155f, -1.f),
    C_Normal3f(0.996461093f, 0.0840546638f, 0.f),
    C_Vertex3f(0.154785156f, 0.511547089f, 0.f),
    C_Vertex3f(0.154785156f, 0.511547089f, -1.f),
    C_Normal3f(0.964526176f, 0.263987303f, 0.f),
    C_Vertex3f(0.156744003f, 0.500593185f, 0.f),
    C_Vertex3f(0.156744003f, 0.500593185f, -1.f),
    C_Normal3f(0.903100073f, 0.429430395f, 0.f),
    C_Vertex3f(0.160661697f, 0.490076065f, 0.f),
    C_Vertex3f(0.160661697f, 0.490076065f, -1.f),
    C_Normal3f(0.823284805f, 0.567628562f, 0.f),
    C_Vertex3f(0.166538239f, 0.479995728f, 0.f),
    C_Vertex3f(0.166538239f, 0.479995728f, -1.f),
    C_Normal3f(0.737232864f, 0.675638676f, 0.f),
    C_Vertex3f(0.173503876f, 0.471449852f, 0.f),
    C_Vertex3f(0.173503876f, 0.471449852f, -1.f),
    C_Normal3f(0.673301756f, 0.739367723f, 0.f),
    C_Vertex3f(0.18224287f, 0.462859392f, 0.f),
    C_Vertex3f(0.18224287f, 0.462859392f, -1.f),
    C_Normal3f(0.62993443f, 0.776648283f, 0.f),
    C_Vertex3f(0.193624973f, 0.453126669f, 0.f),
    C_Vertex3f(0.193624973f, 0.453126669f, -1.f),
    C_Normal3f(0.597859859f, 0.801600575f, 0.f),
    C_Vertex3f(0.207650185f, 0.442251682f, 0.f),
    C_Vertex3f(0.207650185f, 0.442251682f, -1.f),
    C_Normal3f(0.573323846f, 0.819328785f, 0.f),
    C_Vertex3f(0.224318504f, 0.430234432f, 0.f),
    C_Vertex3f(0.224318504f, 0.430234432f, -1.f),
    C_Normal3f(0.554010689f, 0.832509577f, 0.f),
    C_Vertex3f(0.243629932f, 0.417074919f, 0.f),
    C_Vertex3f(0.243629932f, 0.417074919f, -1.f),
    C_Normal3f(0.538442612f, 0.842662275f, 0.f),
    C_Vertex3f(0.265584469f, 0.402773142f, 0.f),
    C_Vertex3f(0.265584469f, 0.402773142f, -1.f),
    C_Normal3f(0.526642621f, 0.850086749f, 0.f),
    C_Vertex3f(0.290182114f, 0.387329102f, 0.f),
    C_Vertex3f(0.290182114f, 0.387329102f, -1.f),
    C_Normal3f(0.523997307f, 0.851719975f, 0.f),
    C_Vertex3f(0.312355995f, 0.373797417f, 0.f),
    C_Vertex3f(0.312355995f, 0.373797417f, -1.f),
    C_Normal3f(0.531705916f, 0.846929014f, 0.f),
    C_Vertex3f(0.328075409f, 0.364016294f, 0.f),
    C_Vertex3f(0.328075409f, 0.364016294f, -1.f),
    C_Normal3f(0.539442301f, 0.842022538f, 0.f),
    C_Vertex3f(0.342407227f, 0.354931116f, 0.f),
    C_Vertex3f(0.342407227f, 0.354931116f, -1.f),
    C_Normal3f(0.548748136f, 0.835987747f, 0.f),
    C_Vertex3f(0.355351448f, 0.346541882f, 0.f),
    C_Vertex3f(0.355351448f, 0.346541882f, -1.f),
    C_Normal3f(0.560149014f, 0.82839185f, 0.f),
    C_Vertex3f(0.366908073f, 0.338848591f, 0.f),
    C_Vertex3f(0.366908073f, 0.338848591f, -1.f),
    C_Normal3f(0.574430943f, 0.818552971f, 0.f),
    C_Vertex3f(0.377077103f, 0.331851244f, 0.f),
    C_Vertex3f(0.377077103f, 0.331851244f, -1.f),
    C_Normal3f(0.592819571f, 0.805335283f, 0.f),
    C_Vertex3f(0.385858536f, 0.325549841f, 0.f),
    C_Vertex3f(0.385858536f, 0.325549841f, -1.f),
    C_Normal3f(0.612693965f, 0.790320277f, 0.f),
    C_Vertex3f(0.393252373f, 0.319944382f, 0.f),
    C_Vertex3f(0.393252373f, 0.319944382f, -1.f),
    C_Normal3f(0.641754508f, 0.766910195f, 0.f),
    C_Vertex3f(0.396255493f, 0.317489624f, 0.f),
    C_Vertex3f(0.396255493f, 0.317489624f, -1.f),
    C_Normal3f(0.683399558f, 0.730044603f, 0.f),
    C_Vertex3f(0.404010773f, 0.310941696f, 0.f),
    C_Vertex3f(0.404010773f, 0.310941696f, -1.f),
    C_Normal3f(0.733505309f, 0.679683566f, 0.f),
    C_Vertex3f(0.417472839f, 0.297627926f, 0.f),
    C_Vertex3f(0.417472839f, 0.297627926f, -1.f),
    C_Normal3f(0.795531273f, 0.605912507f, 0.f),
    C_Vertex3f(0.428886414f, 0.284096241f, 0.f),
    C_Vertex3f(0.428886414f, 0.284096241f, -1.f),
    C_Normal3f(0.856791139f, 0.515663624f, 0.f),
    C_Vertex3f(0.438251495f, 0.270346642f, 0.f),
    C_Vertex3f(0.438251495f, 0.270346642f, -1.f),
    C_Normal3f(0.912362218f, 0.409383744f, 0.f),
    C_Vertex3f(0.445568085f, 0.256379128f, 0.f),
    C_Vertex3f(0.445568085f, 0.256379128f, -1.f),
    C_Normal3f(0.955813468f, 0.293973833f, 0.f),
    C_Vertex3f(0.451055527f, 0.241811275f, 0.f),
    C_Vertex3f(0.451055527f, 0.241811275f, -1.f),
    C_Normal3f(0.984630585f, 0.174650177f, 0.f),
    C_Vertex3f(0.454713821f, 0.226643085f, 0.f),
    C_Vertex3f(0.454713821f, 0.226643085f, -1.f),
    C_Normal3f(0.998160839f, 0.0606200621f, 0.f),
    C_Vertex3f(0.456542969f, 0.210874557f, 0.f),
    C_Vertex3f(0.456542969f, 0.210874557f, -1.f),
    C_Normal3f(0.999277949f, -0.0379932523f, 0.f),
    C_Vertex3f(0.456542969f, 0.19652462f, 0.f),
    C_Vertex3f(0.456542969f, 0.19652462f, -1.f),
    C_Normal3f(0.991863728f, -0.127304047f, 0.f),
    C_Vertex3f(0.455545664f, 0.184643984f, 0.f),
    C_Vertex3f(0.455545664f, 0.184643984f, -1.f),
    C_Normal3f(0.976127207f, -0.217199847f, 0.f),
    C_Vertex3f(0.453551054f, 0.17321372f, 0.f),
    C_Vertex3f(0.453551054f, 0.17321372f, -1.f),
    C_Normal3f(0.951156497f, -0.308709145f, 0.f),
    C_Vertex3f(0.450559139f, 0.162233829f, 0.f),
    C_Vertex3f(0.450559139f, 0.162233829f, -1.f),
    C_Normal3f(0.9168185f, -0.399304301f, 0.f),
    C_Vertex3f(0.44656992f, 0.151704311f, 0.f),
    C_Vertex3f(0.44656992f, 0.151704311f, -1.f),
    C_Normal3f(0.873752058f, -0.486371487f, 0.f),
    C_Vertex3f(0.441583395f, 0.141625166f, 0.f),
    C_Vertex3f(0.441583395f, 0.141625166f, -1.f),
    C_Normal3f(0.823325336f, -0.567569792f, 0.f),
    C_Vertex3f(0.435599566f, 0.131996393f, 0.f),
    C_Vertex3f(0.435599566f, 0.131996393f, -1.f),
    C_Normal3f(0.766377687f, -0.642390192f, 0.f),
    C_Vertex3f(0.428618431f, 0.122817993f, 0.f),
    C_Vertex3f(0.428618431f, 0.122817993f, -1.f),
    C_Normal3f(0.669998944f, -0.742362142f, 0.f),
    C_Vertex3f(0.420040131f, 0.113433838f, 0.f),
    C_Vertex3f(0.420040131f, 0.113433838f, -1.f),
    C_Normal3f(0.561475575f, -0.82749337f, 0.f),
    C_Vertex3f(0.407035828f, 0.103339195f, 0.f),
    C_Vertex3f(0.407035828f, 0.103339195f, -1.f),
    C_Normal3f(0.477132767f, -0.878831148f, 0.f),
    C_Vertex3f(0.390205383f, 0.0931901932f, 0.f),
    C_Vertex3f(0.390205383f, 0.0931901932f, -1.f),
    C_Normal3f(0.422429472f, -0.906395912f, 0.f),
    C_Vertex3f(0.369548798f, 0.0829868317f, 0.f),
    C_Vertex3f(0.369548798f, 0.0829868317f, -1.f),
    C_Normal3f(0.386430979f, -0.922318339f, 0.f),
    C_Vertex3f(0.357307434f, 0.0778579712f, 0.f),
    C_Vertex3f(0.357307434f, 0.0778579712f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.606095552f, 0.795391917f, 0.f),
    C_Vertex3f(0.357307434f, 0.0778579712f, 0.f),
    C_Vertex3f(0.357307434f, 0.0778579712f, -1.f),
    C_Normal3f(0.648845375f, 0.760920227f, 0.f),
    C_Vertex3f(0.366043091f, 0.0712013245f, 0.f),
    C_Vertex3f(0.366043091f, 0.0712013245f, -1.f),
    C_Normal3f(0.704495907f, 0.709707916f, 0.f),
    C_Vertex3f(0.38130331f, 0.0573964119f, 0.f),
    C_Vertex3f(0.38130331f, 0.0573964119f, -1.f),
    C_Normal3f(0.772776186f, 0.634678602f, 0.f),
    C_Vertex3f(0.394352436f, 0.0430998802f, 0.f),
    C_Vertex3f(0.394352436f, 0.0430998802f, -1.f),
    C_Normal3f(0.839452326f, 0.543433428f, 0.f),
    C_Vertex3f(0.405190468f, 0.0283117294f, 0.f),
    C_Vertex3f(0.405190468f, 0.0283117294f, -1.f),
    C_Normal3f(0.899630785f, 0.436651468f, 0.f),
    C_Vertex3f(0.413817406f, 0.0130319595f, 0.f),
    C_Vertex3f(0.413817406f, 0.0130319595f, -1.f),
    C_Normal3f(0.948221564f, 0.31760934f, 0.f),
    C_Vertex3f(0.420287609f, -0.00279283524f, 0.f),
    C_Vertex3f(0.420287609f, -0.00279283524f, -1.f),
    C_Normal3f(0.981625438f, 0.190818101f, 0.f),
    C_Vertex3f(0.424601078f, -0.0191626549f, 0.f),
    C_Vertex3f(0.424601078f, -0.0191626549f, -1.f),
    C_Normal3f(0.998133361f, 0.0610727295f, 0.f),
    C_Vertex3f(0.426757813f, -0.0360774994f, 0.f),
    C_Vertex3f(0.426757813f, -0.0360774994f, -1.f),
    C_Normal3f(0.997292519f, -0.0735371709f, 0.f),
    C_Vertex3f(0.426757813f, -0.0544109344f, 0.f),
    C_Vertex3f(0.426757813f, -0.0544109344f, -1.f),
    C_Normal3f(0.976193905f, -0.216899708f, 0.f),
    C_Vertex3f(0.42402792f, -0.0730996132f, 0.f),
    C_Vertex3f(0.42402792f, -0.0730996132f, -1.f),
    C_Normal3f(0.934463143f, -0.356059998f, 0.f),
    C_Vertex3f(0.418568134f, -0.0912699699f, 0.f),
    C_Vertex3f(0.418568134f, -0.0912699699f, -1.f),
    C_Normal3f(0.876459539f, -0.481475472f, 0.f),
    C_Vertex3f(0.410378456f, -0.108922005f, 0.f),
    C_Vertex3f(0.410378456f, -0.108922005f, -1.f),
    C_Normal3f(0.799349904f, -0.600865781f, 0.f),
    C_Vertex3f(0.399458885f, -0.126055717f, 0.f),
    C_Vertex3f(0.399458885f, -0.126055717f, -1.f),
    C_Normal3f(0.696733117f, -0.717330515f, 0.f),
    C_Vertex3f(0.386137486f, -0.141170502f, 0.f),
    C_Vertex3f(0.386137486f, -0.141170502f, -1.f),
    C_Normal3f(0.581165552f, -0.813785374f, 0.f),
    C_Vertex3f(0.370414257f, -0.154266357f, 0.f),
    C_Vertex3f(0.370414257f, -0.154266357f, -1.f),
    C_Normal3f(0.462001324f, -0.886879265f, 0.f),
    C_Vertex3f(0.3522892f, -0.165343285f, 0.f),
    C_Vertex3f(0.3522892f, -0.165343285f, -1.f),
    C_Normal3f(0.354041904f, -0.935229599f, 0.f),
    C_Vertex3f(0.331762314f, -0.174401283f, 0.f),
    C_Vertex3f(0.331762314f, -0.174401283f, -1.f),
    C_Normal3f(0.251832545f, -0.967770815f, 0.f),
    C_Vertex3f(0.310416222f, -0.181194782f, 0.f),
    C_Vertex3f(0.310416222f, -0.181194782f, -1.f),
    C_Normal3f(0.148790866f, -0.988868654f, 0.f),
    C_Vertex3f(0.288250923f, -0.185723782f, 0.f),
    C_Vertex3f(0.288250923f, -0.185723782f, -1.f),
    C_Normal3f(0.0499114096f, -0.998753726f, 0.f),
    C_Vertex3f(0.265266418f, -0.187988281f, 0.f),
    C_Vertex3f(0.265266418f, -0.187988281f, -1.f),
    C_Normal3f(-0.030949574f, -0.999521017f, 0.f),
    C_Vertex3f(0.242937088f, -0.187988281f, 0.f),
    C_Vertex3f(0.242937088f, -0.187988281f, -1.f),
    C_Normal3f(-0.101137578f, -0.994872451f, 0.f),
    C_Vertex3f(0.223160267f, -0.186684489f, 0.f),
    C_Vertex3f(0.223160267f, -0.186684489f, -1.f),
    C_Normal3f(-0.176667184f, -0.984270632f, 0.f),
    C_Vertex3f(0.204461575f, -0.184076905f, 0.f),
    C_Vertex3f(0.204461575f, -0.184076905f, -1.f),
    C_Normal3f(-0.25809592f, -0.966119289f, 0.f),
    C_Vertex3f(0.186841011f, -0.180165529f, 0.f),
    C_Vertex3f(0.186841011f, -0.180165529f, -1.f),
    C_Normal3f(-0.34421131f, -0.938892245f, 0.f),
    C_Vertex3f(0.170298576f, -0.174950361f, 0.f),
    C_Vertex3f(0.170298576f, -0.174950361f, -1.f),
    C_Normal3f(-0.433061898f, -0.901364267f, 0.f),
    C_Vertex3f(0.15483427f, -0.168431401f, 0.f),
    C_Vertex3f(0.15483427f, -0.168431401f, -1.f),
    C_Normal3f(-0.522011518f, -0.852938354f, 0.f),
    C_Vertex3f(0.140448093f, -0.160608649f, 0.f),
    C_Vertex3f(0.140448093f, -0.160608649f, -1.f),
    C_Normal3f(-0.607996225f, -0.793939948f, 0.f),
    C_Vertex3f(0.127140045f, -0.151482105f, 0.f),
    C_Vertex3f(0.127140045f, -0.151482105f, -1.f),
    C_Normal3f(-0.685583413f, -0.727994144f, 0.f),
    C_Vertex3f(0.114910126f, -0.141051769f, 0.f),
    C_Vertex3f(0.114910126f, -0.141051769f, -1.f),
    C_Normal3f(-0.754905343f, -0.655833662f, 0.f),
    C_Vertex3f(0.103908777f, -0.12960422f, 0.f),
    C_Vertex3f(0.103908777f, -0.12960422f, -1.f),
    C_Normal3f(-0.816943526f, -0.576717675f, 0.f),
    C_Vertex3f(0.0941359997f, -0.117139459f, 0.f),
    C_Vertex3f(0.0941359997f, -0.117139459f, -1.f),
    C_Normal3f(-0.869970858f, -0.493103176f, 0.f),
    C_Vertex3f(0.0855917931f, -0.103657484f, 0.f),
    C_Vertex3f(0.0855917931f, -0.103657484f, -1.f),
    C_Normal3f(-0.913105249f, -0.407724023f, 0.f),
    C_Vertex3f(0.0782761574f, -0.0891582966f, 0.f),
    C_Vertex3f(0.0782761574f, -0.0891582966f, -1.f),
    C_Normal3f(-0.946334958f, -0.32318753f, 0.f),
    C_Vertex3f(0.0721890926f, -0.0736418962f, 0.f),
    C_Vertex3f(0.0721890926f, -0.0736418962f, -1.f),
    C_Normal3f(-0.970361173f, -0.241659537f, 0.f),
    C_Vertex3f(0.0673305988f, -0.057108283f, 0.f),
    C_Vertex3f(0.0673305988f, -0.057108283f, -1.f),
    C_Normal3f(-0.984181225f, -0.177164599f, 0.f),
    C_Vertex3f(0.063700676f, -0.039557457f, 0.f),
    C_Vertex3f(0.063700676f, -0.039557457f, -1.f),
    C_Normal3f(-0.991740704f, -0.128259018f, 0.f),
    C_Vertex3f(0.0625f, -0.0302734375f, 0.f),
    C_Vertex3f(0.0625f, -0.0302734375f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.178699166f, 0.983903706f, 0.f),
    C_Vertex3f(0.318054199f, 0.106666565f, 0.f),
    C_Vertex3f(0.318054199f, 0.106666565f, -1.f),
    C_Normal3f(-0.509701848f, 0.860351086f, 0.f),
    C_Vertex3f(0.326112747f, 0.110813141f, 0.f),
    C_Vertex3f(0.326112747f, 0.110813141f, -1.f),
    C_Normal3f(-0.580230594f, 0.814452231f, 0.f),
    C_Vertex3f(0.340202808f, 0.11978817f, 0.f),
    C_Vertex3f(0.340202808f, 0.11978817f, -1.f),
    C_Normal3f(-0.670938075f, 0.741513431f, 0.f),
    C_Vertex3f(0.352265835f, 0.129445076f, 0.f),
    C_Vertex3f(0.352265835f, 0.129445076f, -1.f),
    C_Normal3f(-0.763887346f, 0.645349562f, 0.f),
    C_Vertex3f(0.362301826f, 0.139783859f, 0.f),
    C_Vertex3f(0.362301826f, 0.139783859f, -1.f),
    C_Normal3f(-0.842332304f, 0.53895849f, 0.f),
    C_Vertex3f(0.370310783f, 0.15080452f, 0.f),
    C_Vertex3f(0.370310783f, 0.15080452f, -1.f),
    C_Normal3f(-0.907526255f, 0.419995457f, 0.f),
    C_Vertex3f(0.376317501f, 0.161688805f, 0.f),
    C_Vertex3f(0.376317501f, 0.161688805f, -1.f),
    C_Normal3f(-0.962658286f, 0.270719439f, 0.f),
    C_Vertex3f(0.38032198f, 0.172436714f, 0.f),
    C_Vertex3f(0.38032198f, 0.172436714f, -1.f),
    C_Normal3f(-0.995909572f, 0.0903553069f, 0.f),
    C_Vertex3f(0.382324219f, 0.183048248f, 0.f),
    C_Vertex3f(0.382324219f, 0.183048248f, -1.f),
    C_Normal3f(-0.993166327f, -0.116707407f, 0.f),
    C_Vertex3f(0.382324219f, 0.194505692f, 0.f),
    C_Vertex3f(0.382324219f, 0.194505692f, -1.f),
    C_Normal3f(-0.949572384f, -0.313547969f, 0.f),
    C_Vertex3f(0.379477501f, 0.207273483f, 0.f),
    C_Vertex3f(0.379477501f, 0.207273483f, -1.f),
    C_Normal3f(-0.881112337f, -0.472907126f, 0.f),
    C_Vertex3f(0.373784065f, 0.220369339f, 0.f),
    C_Vertex3f(0.373784065f, 0.220369339f, -1.f),
    C_Normal3f(-0.818783522f, -0.574102402f, 0.f),
    C_Vertex3f(0.365243912f, 0.233793259f, 0.f),
    C_Vertex3f(0.365243912f, 0.233793259f, -1.f),
    C_Normal3f(-0.768174708f, -0.640240312f, 0.f),
    C_Vertex3f(0.359550476f, 0.24066925f, 0.f),
    C_Vertex3f(0.359550476f, 0.24066925f, -1.f),
    C_Normal3f(-0.726073742f, -0.687616885f, 0.f),
    C_Vertex3f(0.356648445f, 0.244106293f, 0.f),
    C_Vertex3f(0.356648445f, 0.244106293f, -1.f),
    C_Normal3f(-0.68593514f, -0.727662623f, 0.f),
    C_Vertex3f(0.349229813f, 0.251567125f, 0.f),
    C_Vertex3f(0.349229813f, 0.251567125f, -1.f),
    C_Normal3f(-0.646592021f, -0.762836039f, 0.f),
    C_Vertex3f(0.340196609f, 0.259614706f, 0.f),
    C_Vertex3f(0.340196609f, 0.259614706f, -1.f),
    C_Normal3f(-0.614721715f, -0.788744152f, 0.f),
    C_Vertex3f(0.329548836f, 0.268249035f, 0.f),
    C_Vertex3f(0.329548836f, 0.268249035f, -1.f),
    C_Normal3f(-0.588546515f, -0.808463335f, 0.f),
    C_Vertex3f(0.317286491f, 0.277470112f, 0.f),
    C_Vertex3f(0.317286491f, 0.277470112f, -1.f),
    C_Normal3f(-0.566750228f, -0.823889673f, 0.f),
    C_Vertex3f(0.303409576f, 0.287277937f, 0.f),
    C_Vertex3f(0.303409576f, 0.287277937f, -1.f),
    C_Normal3f(-0.548365176f, -0.83623898f, 0.f),
    C_Vertex3f(0.287918091f, 0.29767251f, 0.f),
    C_Vertex3f(0.287918091f, 0.29767251f, -1.f),
    C_Normal3f(-0.532622814f, -0.846352696f, 0.f),
    C_Vertex3f(0.270812035f, 0.308653831f, 0.f),
    C_Vertex3f(0.270812035f, 0.308653831f, -1.f),
    C_Normal3f(-0.527770162f, -0.849387228f, 0.f),
    C_Vertex3f(0.253995895f, 0.319020271f, 0.f),
    C_Vertex3f(0.253995895f, 0.319020271f, -1.f),
    C_Normal3f(-0.535601377f, -0.844470918f, 0.f),
    C_Vertex3f(0.240036726f, 0.327776194f, 0.f),
    C_Vertex3f(0.240036726f, 0.327776194f, -1.f),
    C_Normal3f(-0.544884861f, -0.838510871f, 0.f),
    C_Vertex3f(0.227030039f, 0.336123228f, 0.f),
    C_Vertex3f(0.227030039f, 0.336123228f, -1.f),
    C_Normal3f(-0.55544883f, -0.831550717f, 0.f),
    C_Vertex3f(0.214975834f, 0.344061375f, 0.f),
    C_Vertex3f(0.214975834f, 0.344061375f, -1.f),
    C_Normal3f(-0.567569971f, -0.823325157f, 0.f),
    C_Vertex3f(0.203874111f, 0.351590633f, 0.f),
    C_Vertex3f(0.203874111f, 0.351590633f, -1.f),
    C_Normal3f(-0.581607997f, -0.813469172f, 0.f),
    C_Vertex3f(0.193724871f, 0.358711004f, 0.f),
    C_Vertex3f(0.193724871f, 0.358711004f, -1.f),
    C_Normal3f(-0.598036587f, -0.80146879f, 0.f),
    C_Vertex3f(0.184528112f, 0.365422487f, 0.f),
    C_Vertex3f(0.184528112f, 0.365422487f, -1.f),
    C_Normal3f(-0.614004135f, -0.789302766f, 0.f),
    C_Vertex3f(0.176283836f, 0.371725082f, 0.f),
    C_Vertex3f(0.176283836f, 0.371725082f, -1.f),
    C_Normal3f(0.163630009f, -0.98652184f, 0.f),
    C_Vertex3f(0.172637939f, 0.374671936f, 0.f),
    C_Vertex3f(0.172637939f, 0.374671936f, -1.f),
    C_Normal3f(0.531089783f, -0.84731555f, 0.f),
    C_Vertex3f(0.163784027f, 0.369651794f, 0.f),
    C_Vertex3f(0.163784027f, 0.369651794f, -1.f),
    C_Normal3f(0.585351467f, -0.810779572f, 0.f),
    C_Vertex3f(0.148610115f, 0.359611511f, 0.f),
    C_Vertex3f(0.148610115f, 0.359611511f, -1.f),
    C_Normal3f(0.661816001f, -0.749666274f, 0.f),
    C_Vertex3f(0.135970116f, 0.349571228f, 0.f),
    C_Vertex3f(0.135970116f, 0.349571228f, -1.f),
    C_Normal3f(0.750577033f, -0.660782993f, 0.f),
    C_Vertex3f(0.125864029f, 0.339530945f, 0.f),
    C_Vertex3f(0.125864029f, 0.339530945f, -1.f),
    C_Normal3f(0.84132123f, -0.54053539f, 0.f),
    C_Vertex3f(0.118291855f, 0.329490662f, 0.f),
    C_Vertex3f(0.118291855f, 0.329490662f, -1.f),
    C_Normal3f(0.916695118f, -0.399587125f, 0.f),
    C_Vertex3f(0.112612724f, 0.31890583f, 0.f),
    C_Vertex3f(0.112612724f, 0.31890583f, -1.f),
    C_Normal3f(0.970359325f, -0.241666585f, 0.f),
    C_Vertex3f(0.108826637f, 0.307776451f, 0.f),
    C_Vertex3f(0.108826637f, 0.307776451f, -1.f),
    C_Normal3f(0.997066319f, -0.0765431672f, 0.f),
    C_Vertex3f(0.106933594f, 0.296102524f, 0.f),
    C_Vertex3f(0.106933594f, 0.296102524f, -1.f),
    C_Normal3f(0.995462537f, 0.095153451f, 0.f),
    C_Vertex3f(0.106933594f, 0.283117294f, 0.f),
    C_Vertex3f(0.106933594f, 0.283117294f, -1.f),
    C_Normal3f(0.961561799f, 0.274588674f, 0.f),
    C_Vertex3f(0.10945797f, 0.269693375f, 0.f),
    C_Vertex3f(0.10945797f, 0.269693375f, -1.f),
    C_Normal3f(0.898693562f, 0.438577056f, 0.f),
    C_Vertex3f(0.114506721f, 0.256597519f, 0.f),
    C_Vertex3f(0.114506721f, 0.256597519f, -1.f),
    C_Normal3f(0.82483983f, 0.565366447f, 0.f),
    C_Vertex3f(0.122079849f, 0.243829727f, 0.f),
    C_Vertex3f(0.122079849f, 0.243829727f, -1.f),
    C_Normal3f(0.749734581f, 0.661738634f, 0.f),
    C_Vertex3f(0.129652977f, 0.234499931f, 0.f),
    C_Vertex3f(0.129652977f, 0.234499931f, -1.f),
    C_Normal3f(0.687598467f, 0.726091206f, 0.f),
    C_Vertex3f(0.13626194f, 0.227761745f, 0.f),
    C_Vertex3f(0.13626194f, 0.227761745f, -1.f),
    C_Normal3f(0.643105447f, 0.765777588f, 0.f),
    C_Vertex3f(0.144431114f, 0.220505238f, 0.f),
    C_Vertex3f(0.144431114f, 0.220505238f, -1.f),
    C_Normal3f(0.607322335f, 0.794455528f, 0.f),
    C_Vertex3f(0.1541605f, 0.212730408f, 0.f),
    C_Vertex3f(0.1541605f, 0.212730408f, -1.f),
    C_Normal3f(0.578149736f, 0.815930724f, 0.f),
    C_Vertex3f(0.165450096f, 0.204437256f, 0.f),
    C_Vertex3f(0.165450096f, 0.204437256f, -1.f),
    C_Normal3f(0.554024577f, 0.832500339f, 0.f),
    C_Vertex3f(0.178299904f, 0.195625782f, 0.f),
    C_Vertex3f(0.178299904f, 0.195625782f, -1.f),
    C_Normal3f(0.533801734f, 0.845609665f, 0.f),
    C_Vertex3f(0.192709923f, 0.186295986f, 0.f),
    C_Vertex3f(0.192709923f, 0.186295986f, -1.f),
    C_Normal3f(0.515401602f, 0.856948853f, 0.f),
    C_Vertex3f(0.208680153f, 0.176447868f, 0.f),
    C_Vertex3f(0.208680153f, 0.176447868f, -1.f),
    C_Normal3f(0.516813219f, 0.856098175f, 0.f),
    C_Vertex3f(0.232036591f, 0.162643433f, 0.f),
    C_Vertex3f(0.232036591f, 0.162643433f, -1.f),
    C_Normal3f(0.531695008f, 0.846935928f, 0.f),
    C_Vertex3f(0.259203911f, 0.145947456f, 0.f),
    C_Vertex3f(0.259203911f, 0.145947456f, -1.f),
    C_Normal3f(0.549414933f, 0.835549653f, 0.f),
    C_Vertex3f(0.284356117f, 0.129797935f, 0.f),
    C_Vertex3f(0.284356117f, 0.129797935f, -1.f),
    C_Normal3f(0.565929592f, 0.824453592f, 0.f),
    C_Vertex3f(0.30749321f, 0.11419487f, 0.f),
    C_Vertex3f(0.30749321f, 0.11419487f, -1.f),
    C_Normal3f(0.178699166f, 0.983903706f, 0.f),
    C_Vertex3f(0.318054199f, 0.106666565f, 0.f),
    C_Vertex3f(0.318054199f, 0.106666565f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.352265835f, 0.129445076f, 0.f),
    C_Vertex3f(0.340202808f, 0.11978817f, 0.f),
    C_Vertex3f(0.341779709f, -0.0191392899f, 0.f),
    C_Vertex3f(0.345697403f, -0.0293951035f, 0.f),
    C_Vertex3f(0.34765625f, -0.0400886536f, 0.f),
    C_Vertex3f(0.3522892f, -0.165343285f, 0.f),
    C_Vertex3f(0.357307434f, 0.0778579712f, 0.f),
    C_Vertex3f(0.362301826f, 0.139783859f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318054199f, 0.106666565f, 0.f),
    C_Vertex3f(0.318918228f, 0.00893878937f, 0.f),
    C_Vertex3f(0.326112747f, 0.110813141f, 0.f),
    C_Vertex3f(0.328611374f, -0.000539779663f, 0.f),
    C_Vertex3f(0.340202808f, 0.11978817f, 0.f),
    C_Vertex3f(0.335903168f, -0.00932121277f, 0.f),
    C_Vertex3f(0.341779709f, -0.0191392899f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.30749321f, 0.11419487f, 0.f),
    C_Vertex3f(0.284356117f, 0.129797935f, 0.f),
    C_Vertex3f(0.290697098f, 0.0317840576f, 0.f),
    C_Vertex3f(0.306280136f, 0.0197134018f, 0.f),
    C_Vertex3f(0.318918228f, 0.00893878937f, 0.f),
    C_Vertex3f(0.318054199f, 0.106666565f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.290697098f, 0.0317840576f, 0.f),
    C_Vertex3f(0.284356117f, 0.129797935f, 0.f),
    C_Vertex3f(0.272169113f, 0.0451507568f, 0.f),
    C_Vertex3f(0.259203911f, 0.145947456f, 0.f),
    C_Vertex3f(0.250696182f, 0.0598134995f, 0.f),
    C_Vertex3f(0.232036591f, 0.162643433f, 0.f),
    C_Vertex3f(0.226278305f, 0.0757722855f, 0.f),
    C_Vertex3f(0.208680153f, 0.176447868f, 0.f),
    C_Vertex3f(0.198915482f, 0.0930271149f, 0.f),
    C_Vertex3f(0.192709923f, 0.186295986f, 0.f),
    C_Vertex3f(0.178299904f, 0.195625782f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.198915482f, 0.0930271149f, 0.f),
    C_Vertex3f(0.178299904f, 0.195625782f, 0.f),
    C_Vertex3f(0.175694466f, 0.10726738f, 0.f),
    C_Vertex3f(0.165450096f, 0.204437256f, 0.f),
    C_Vertex3f(0.160776973f, 0.11665833f, 0.f),
    C_Vertex3f(0.1541605f, 0.212730408f, 0.f),
    C_Vertex3f(0.147076249f, 0.125510573f, 0.f),
    C_Vertex3f(0.144431114f, 0.220505238f, 0.f),
    C_Vertex3f(0.13626194f, 0.227761745f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147076249f, 0.125510573f, 0.f),
    C_Vertex3f(0.13626194f, 0.227761745f, 0.f),
    C_Vertex3f(0.134592295f, 0.13382411f, 0.f),
    C_Vertex3f(0.129652977f, 0.234499931f, 0.f),
    C_Vertex3f(0.123325109f, 0.14159894f, 0.f),
    C_Vertex3f(0.122079849f, 0.243829727f, 0.f),
    C_Vertex3f(0.114506721f, 0.256597519f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.113274693f, 0.148835063f, 0.f),
    C_Vertex3f(0.123325109f, 0.14159894f, 0.f),
    C_Vertex3f(0.114506721f, 0.256597519f, 0.f),
    C_Vertex3f(0.10945797f, 0.269693375f, 0.f),
    C_Vertex3f(0.106933594f, 0.283117294f, 0.f),
    C_Vertex3f(0.104441047f, 0.155532479f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.359550476f, 0.24066925f, 0.f),
    C_Vertex3f(0.365243912f, 0.233793259f, 0.f),
    C_Vertex3f(0.366908073f, 0.338848591f, 0.f),
    C_Vertex3f(0.373784065f, 0.220369339f, 0.f),
    C_Vertex3f(0.377077103f, 0.331851244f, 0.f),
    C_Vertex3f(0.379477501f, 0.207273483f, 0.f),
    C_Vertex3f(0.385858536f, 0.325549841f, 0.f),
    C_Vertex3f(0.382324219f, 0.194505692f, 0.f),
    C_Vertex3f(0.390205383f, 0.0931901932f, 0.f),
    C_Vertex3f(0.382324219f, 0.183048248f, 0.f),
    C_Vertex3f(0.38032198f, 0.172436714f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329548836f, 0.268249035f, 0.f),
    C_Vertex3f(0.340196609f, 0.259614706f, 0.f),
    C_Vertex3f(0.342407227f, 0.354931116f, 0.f),
    C_Vertex3f(0.349229813f, 0.251567125f, 0.f),
    C_Vertex3f(0.355351448f, 0.346541882f, 0.f),
    C_Vertex3f(0.356648445f, 0.244106293f, 0.f),
    C_Vertex3f(0.366908073f, 0.338848591f, 0.f),
    C_Vertex3f(0.359550476f, 0.24066925f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.270812035f, 0.308653831f, 0.f),
    C_Vertex3f(0.287918091f, 0.29767251f, 0.f),
    C_Vertex3f(0.290182114f, 0.387329102f, 0.f),
    C_Vertex3f(0.303409576f, 0.287277937f, 0.f),
    C_Vertex3f(0.312355995f, 0.373797417f, 0.f),
    C_Vertex3f(0.317286491f, 0.277470112f, 0.f),
    C_Vertex3f(0.328075409f, 0.364016294f, 0.f),
    C_Vertex3f(0.329548836f, 0.268249035f, 0.f),
    C_Vertex3f(0.342407227f, 0.354931116f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227030039f, 0.336123228f, 0.f),
    C_Vertex3f(0.240036726f, 0.327776194f, 0.f),
    C_Vertex3f(0.243629932f, 0.417074919f, 0.f),
    C_Vertex3f(0.253995895f, 0.319020271f, 0.f),
    C_Vertex3f(0.265584469f, 0.402773142f, 0.f),
    C_Vertex3f(0.270812035f, 0.308653831f, 0.f),
    C_Vertex3f(0.290182114f, 0.387329102f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193724871f, 0.358711004f, 0.f),
    C_Vertex3f(0.203874111f, 0.351590633f, 0.f),
    C_Vertex3f(0.207650185f, 0.442251682f, 0.f),
    C_Vertex3f(0.214975834f, 0.344061375f, 0.f),
    C_Vertex3f(0.224318504f, 0.430234432f, 0.f),
    C_Vertex3f(0.227030039f, 0.336123228f, 0.f),
    C_Vertex3f(0.243629932f, 0.417074919f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160661697f, 0.490076065f, 0.f),
    C_Vertex3f(0.163784027f, 0.369651794f, 0.f),
    C_Vertex3f(0.166538239f, 0.479995728f, 0.f),
    C_Vertex3f(0.172637939f, 0.374671936f, 0.f),
    C_Vertex3f(0.173503876f, 0.471449852f, 0.f),
    C_Vertex3f(0.176283836f, 0.371725082f, 0.f),
    C_Vertex3f(0.18224287f, 0.462859392f, 0.f),
    C_Vertex3f(0.184528112f, 0.365422487f, 0.f),
    C_Vertex3f(0.193624973f, 0.453126669f, 0.f),
    C_Vertex3f(0.193724871f, 0.358711004f, 0.f),
    C_Vertex3f(0.207650185f, 0.442251682f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160661697f, 0.490076065f, 0.f),
    C_Vertex3f(0.156744003f, 0.500593185f, 0.f),
    C_Vertex3f(0.163784027f, 0.369651794f, 0.f),
    C_Vertex3f(0.154785156f, 0.511547089f, 0.f),
    C_Vertex3f(0.148610115f, 0.359611511f, 0.f),
    C_Vertex3f(0.151230931f, 0.632103562f, 0.f),
    C_Vertex3f(0.13854444f, 0.624398112f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0790008307f, 0.533735156f, 0.f),
    C_Vertex3f(0.0792250633f, 0.49230957f, 0.f),
    C_Vertex3f(0.0817290545f, 0.546796441f, 0.f),
    C_Vertex3f(0.0824017525f, 0.478919983f, 0.f),
    C_Vertex3f(0.0858213902f, 0.559393167f, 0.f),
    C_Vertex3f(0.0871667862f, 0.465940475f, 0.f),
    C_Vertex3f(0.0912778378f, 0.571525335f, 0.f),
    C_Vertex3f(0.0935201645f, 0.453371048f, 0.f),
    C_Vertex3f(0.0980983973f, 0.583192945f, 0.f),
    C_Vertex3f(0.102173328f, 0.440227509f, 0.f),
    C_Vertex3f(0.106283069f, 0.594395995f, 0.f),
    C_Vertex3f(0.113126278f, 0.426509857f, 0.f),
    C_Vertex3f(0.115831852f, 0.605134487f, 0.f),
    C_Vertex3f(0.126379013f, 0.412218094f, 0.f),
    C_Vertex3f(0.126744747f, 0.615408421f, 0.f),
    C_Vertex3f(0.134155273f, 0.404785156f, 0.f),
    C_Vertex3f(0.13854444f, 0.624398112f, 0.f),
    C_Vertex3f(0.135970116f, 0.349571228f, 0.f),
    C_Vertex3f(0.148610115f, 0.359611511f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0776367188f, 0.520209312f, 0.f),
    C_Vertex3f(0.0776367188f, 0.506109238f, 0.f),
    C_Vertex3f(0.0792250633f, 0.49230957f, 0.f),
    C_Vertex3f(0.0790008307f, 0.533735156f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.44656992f, 0.151704311f, 0.f),
    C_Vertex3f(0.450559139f, 0.162233829f, 0.f),
    C_Vertex3f(0.451055527f, 0.241811275f, 0.f),
    C_Vertex3f(0.453551054f, 0.17321372f, 0.f),
    C_Vertex3f(0.454713821f, 0.226643085f, 0.f),
    C_Vertex3f(0.455545664f, 0.184643984f, 0.f),
    C_Vertex3f(0.456542969f, 0.210874557f, 0.f),
    C_Vertex3f(0.456542969f, 0.19652462f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.420040131f, 0.113433838f, 0.f),
    C_Vertex3f(0.428618431f, 0.122817993f, 0.f),
    C_Vertex3f(0.428886414f, 0.284096241f, 0.f),
    C_Vertex3f(0.435599566f, 0.131996393f, 0.f),
    C_Vertex3f(0.438251495f, 0.270346642f, 0.f),
    C_Vertex3f(0.441583395f, 0.141625166f, 0.f),
    C_Vertex3f(0.445568085f, 0.256379128f, 0.f),
    C_Vertex3f(0.44656992f, 0.151704311f, 0.f),
    C_Vertex3f(0.451055527f, 0.241811275f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.417472839f, 0.297627926f, 0.f),
    C_Vertex3f(0.404010773f, 0.310941696f, 0.f),
    C_Vertex3f(0.407035828f, 0.103339195f, 0.f),
    C_Vertex3f(0.420040131f, 0.113433838f, 0.f),
    C_Vertex3f(0.428886414f, 0.284096241f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.385858536f, 0.325549841f, 0.f),
    C_Vertex3f(0.390205383f, 0.0931901932f, 0.f),
    C_Vertex3f(0.393252373f, 0.319944382f, 0.f),
    C_Vertex3f(0.407035828f, 0.103339195f, 0.f),
    C_Vertex3f(0.396255493f, 0.317489624f, 0.f),
    C_Vertex3f(0.404010773f, 0.310941696f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.38032198f, 0.172436714f, 0.f),
    C_Vertex3f(0.376317501f, 0.161688805f, 0.f),
    C_Vertex3f(0.390205383f, 0.0931901932f, 0.f),
    C_Vertex3f(0.370310783f, 0.15080452f, 0.f),
    C_Vertex3f(0.369548798f, 0.0829868317f, 0.f),
    C_Vertex3f(0.362301826f, 0.139783859f, 0.f),
    C_Vertex3f(0.357307434f, 0.0778579712f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.3522892f, -0.165343285f, 0.f),
    C_Vertex3f(0.34765625f, -0.0400886536f, 0.f),
    C_Vertex3f(0.34765625f, -0.0532894135f, 0.f),
    C_Vertex3f(0.344518661f, -0.0675787926f, 0.f),
    C_Vertex3f(0.338243484f, -0.0808873177f, 0.f),
    C_Vertex3f(0.331762314f, -0.174401283f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338243484f, -0.0808873177f, 0.f),
    C_Vertex3f(0.328830719f, -0.0932149887f, 0.f),
    C_Vertex3f(0.331762314f, -0.174401283f, 0.f),
    C_Vertex3f(0.316280365f, -0.104561806f, 0.f),
    C_Vertex3f(0.310416222f, -0.181194782f, 0.f),
    C_Vertex3f(0.301301003f, -0.113071918f, 0.f),
    C_Vertex3f(0.288250923f, -0.185723782f, 0.f),
    C_Vertex3f(0.283892632f, -0.118745327f, 0.f),
    C_Vertex3f(0.265266418f, -0.187988281f, 0.f),
    C_Vertex3f(0.264055252f, -0.121582031f, 0.f),
    C_Vertex3f(0.242937088f, -0.187988281f, 0.f),
    C_Vertex3f(0.241462708f, -0.121582031f, 0.f),
    C_Vertex3f(0.223160267f, -0.186684489f, 0.f),
    C_Vertex3f(0.220944881f, -0.118718624f, 0.f),
    C_Vertex3f(0.204461575f, -0.184076905f, 0.f),
    C_Vertex3f(0.20282793f, -0.11299181f, 0.f),
    C_Vertex3f(0.187111855f, -0.104401588f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.186841011f, -0.180165529f, 0.f),
    C_Vertex3f(0.204461575f, -0.184076905f, 0.f),
    C_Vertex3f(0.187111855f, -0.104401588f, 0.f),
    C_Vertex3f(0.177125454f, -0.095811367f, 0.f),
    C_Vertex3f(0.170918107f, -0.0890277624f, 0.f),
    C_Vertex3f(0.170298576f, -0.174950361f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.170298576f, -0.174950361f, 0.f),
    C_Vertex3f(0.170918107f, -0.0890277624f, 0.f),
    C_Vertex3f(0.165161014f, -0.0811873674f, 0.f),
    C_Vertex3f(0.159854174f, -0.0722901821f, 0.f),
    C_Vertex3f(0.154997587f, -0.0623362064f, 0.f),
    C_Vertex3f(0.15483427f, -0.168431401f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15483427f, -0.168431401f, 0.f),
    C_Vertex3f(0.154997587f, -0.0623362064f, 0.f),
    C_Vertex3f(0.150591254f, -0.0513254404f, 0.f),
    C_Vertex3f(0.146635175f, -0.039257884f, 0.f),
    C_Vertex3f(0.143129349f, -0.0261335373f, 0.f),
    C_Vertex3f(0.141601563f, -0.0190429688f, 0.f),
    C_Vertex3f(0.140448093f, -0.160608649f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.141601563f, -0.0190429688f, 0.f),
    C_Vertex3f(0.0625f, -0.0302734375f, 0.f),
    C_Vertex3f(0.063700676f, -0.039557457f, 0.f),
    C_Vertex3f(0.0673305988f, -0.057108283f, 0.f),
    C_Vertex3f(0.0721890926f, -0.0736418962f, 0.f),
    C_Vertex3f(0.0782761574f, -0.0891582966f, 0.f),
    C_Vertex3f(0.0855917931f, -0.103657484f, 0.f),
    C_Vertex3f(0.0941359997f, -0.117139459f, 0.f),
    C_Vertex3f(0.103908777f, -0.12960422f, 0.f),
    C_Vertex3f(0.114910126f, -0.141051769f, 0.f),
    C_Vertex3f(0.127140045f, -0.151482105f, 0.f),
    C_Vertex3f(0.140448093f, -0.160608649f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.366043091f, 0.0712013245f, 0.f),
    C_Vertex3f(0.370414257f, -0.154266357f, 0.f),
    C_Vertex3f(0.38130331f, 0.0573964119f, 0.f),
    C_Vertex3f(0.386137486f, -0.141170502f, 0.f),
    C_Vertex3f(0.394352436f, 0.0430998802f, 0.f),
    C_Vertex3f(0.399458885f, -0.126055717f, 0.f),
    C_Vertex3f(0.405190468f, 0.0283117294f, 0.f),
    C_Vertex3f(0.410378456f, -0.108922005f, 0.f),
    C_Vertex3f(0.413817406f, 0.0130319595f, 0.f),
    C_Vertex3f(0.418568134f, -0.0912699699f, 0.f),
    C_Vertex3f(0.420287609f, -0.00279283524f, 0.f),
    C_Vertex3f(0.42402792f, -0.0730996132f, 0.f),
    C_Vertex3f(0.424601078f, -0.0191626549f, 0.f),
    C_Vertex3f(0.426757813f, -0.0544109344f, 0.f),
    C_Vertex3f(0.426757813f, -0.0360774994f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.357307434f, 0.0778579712f, 0.f),
    C_Vertex3f(0.3522892f, -0.165343285f, 0.f),
    C_Vertex3f(0.370414257f, -0.154266357f, 0.f),
    C_Vertex3f(0.366043091f, 0.0712013245f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.419433594f, 0.499511719f, 0.f),
    C_Vertex3f(0.418287277f, 0.508756638f, 0.f),
    C_Vertex3f(0.414848924f, 0.526152372f, 0.f),
    C_Vertex3f(0.41026485f, 0.542454004f, 0.f),
    C_Vertex3f(0.404535055f, 0.557661533f, 0.f),
    C_Vertex3f(0.39765954f, 0.57177496f, 0.f),
    C_Vertex3f(0.389638305f, 0.584794283f, 0.f),
    C_Vertex3f(0.380471349f, 0.596719503f, 0.f),
    C_Vertex3f(0.370158672f, 0.607550621f, 0.f),
    C_Vertex3f(0.358700275f, 0.617287636f, 0.f),
    C_Vertex3f(0.346123338f, 0.625807524f, 0.f),
    C_Vertex3f(0.337890625f, 0.491210938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.32730484f, 0.531983376f, 0.f),
    C_Vertex3f(0.330881119f, 0.521639347f, 0.f),
    C_Vertex3f(0.332427859f, 0.633110285f, 0.f),
    C_Vertex3f(0.334028244f, 0.510280132f, 0.f),
    C_Vertex3f(0.346123338f, 0.625807524f, 0.f),
    C_Vertex3f(0.336746216f, 0.497905731f, 0.f),
    C_Vertex3f(0.337890625f, 0.491210938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308708191f, 0.563207626f, 0.f),
    C_Vertex3f(0.314001083f, 0.556924343f, 0.f),
    C_Vertex3f(0.31761384f, 0.639195919f, 0.f),
    C_Vertex3f(0.318864822f, 0.549625874f, 0.f),
    C_Vertex3f(0.332427859f, 0.633110285f, 0.f),
    C_Vertex3f(0.323299408f, 0.541312218f, 0.f),
    C_Vertex3f(0.32730484f, 0.531983376f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.30168128f, 0.644064426f, 0.f),
    C_Vertex3f(0.284630179f, 0.647715807f, 0.f),
    C_Vertex3f(0.286691666f, 0.579011917f, 0.f),
    C_Vertex3f(0.300125122f, 0.571109772f, 0.f),
    C_Vertex3f(0.308708191f, 0.563207626f, 0.f),
    C_Vertex3f(0.31761384f, 0.639195919f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157864094f, 0.536385536f, 0.f),
    C_Vertex3f(0.164021969f, 0.548380852f, 0.f),
    C_Vertex3f(0.16480422f, 0.638524771f, 0.f),
    C_Vertex3f(0.173258781f, 0.559801102f, 0.f),
    C_Vertex3f(0.179264307f, 0.643661737f, 0.f),
    C_Vertex3f(0.185574532f, 0.570646286f, 0.f),
    C_Vertex3f(0.194611192f, 0.647514462f, 0.f),
    C_Vertex3f(0.19982481f, 0.578780174f, 0.f),
    C_Vertex3f(0.210844874f, 0.650082946f, 0.f),
    C_Vertex3f(0.216009617f, 0.584202766f, 0.f),
    C_Vertex3f(0.227965355f, 0.651367188f, 0.f),
    C_Vertex3f(0.234128952f, 0.586914063f, 0.f),
    C_Vertex3f(0.247172356f, 0.651367188f, 0.f),
    C_Vertex3f(0.253856659f, 0.586914063f, 0.f),
    C_Vertex3f(0.266460538f, 0.650150061f, 0.f),
    C_Vertex3f(0.271268845f, 0.584280014f, 0.f),
    C_Vertex3f(0.284630179f, 0.647715807f, 0.f),
    C_Vertex3f(0.286691666f, 0.579011917f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.154785156f, 0.523815155f, 0.f),
    C_Vertex3f(0.157864094f, 0.536385536f, 0.f),
    C_Vertex3f(0.16480422f, 0.638524771f, 0.f),
    C_Vertex3f(0.151230931f, 0.632103562f, 0.f),
    C_Vertex3f(0.154785156f, 0.511547089f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.112612724f, 0.31890583f, 0.f),
    C_Vertex3f(0.118291855f, 0.329490662f, 0.f),
    C_Vertex3f(0.122661591f, 0.40140152f, 0.f),
    C_Vertex3f(0.125864029f, 0.339530945f, 0.f),
    C_Vertex3f(0.134155273f, 0.404785156f, 0.f),
    C_Vertex3f(0.135970116f, 0.349571228f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0968241692f, 0.161691189f, 0.f),
    C_Vertex3f(0.104441047f, 0.155532479f, 0.f),
    C_Vertex3f(0.102137566f, 0.39207077f, 0.f),
    C_Vertex3f(0.106933594f, 0.296102524f, 0.f),
    C_Vertex3f(0.122661591f, 0.40140152f, 0.f),
    C_Vertex3f(0.108826637f, 0.307776451f, 0.f),
    C_Vertex3f(0.112612724f, 0.31890583f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0558815002f, 0.209895134f, 0.f),
    C_Vertex3f(0.0651526451f, 0.195628166f, 0.f),
    C_Vertex3f(0.068479538f, 0.365718842f, 0.f),
    C_Vertex3f(0.0756001472f, 0.182370186f, 0.f),
    C_Vertex3f(0.0840768814f, 0.380176544f, 0.f),
    C_Vertex3f(0.0872240067f, 0.170121193f, 0.f),
    C_Vertex3f(0.102137566f, 0.39207077f, 0.f),
    C_Vertex3f(0.0968241692f, 0.161691189f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0356445313f, 0.291734695f, 0.f),
    C_Vertex3f(0.0376682281f, 0.256705761f, 0.f),
    C_Vertex3f(0.0389280319f, 0.311705589f, 0.f),
    C_Vertex3f(0.0417156219f, 0.240774632f, 0.f),
    C_Vertex3f(0.0454950333f, 0.330693245f, 0.f),
    C_Vertex3f(0.0477867126f, 0.225171089f, 0.f),
    C_Vertex3f(0.0553455353f, 0.348697662f, 0.f),
    C_Vertex3f(0.0558815002f, 0.209895134f, 0.f),
    C_Vertex3f(0.068479538f, 0.365718842f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0376682281f, 0.256705761f, 0.f),
    C_Vertex3f(0.0356445313f, 0.291734695f, 0.f),
    C_Vertex3f(0.0356445313f, 0.272964478f, 0.f),
    C_Vertex3f(0.106933594f, 0.296102524f, 0.f),
    C_Vertex3f(0.104441047f, 0.155532479f, 0.f),
    C_Vertex3f(0.106933594f, 0.283117294f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.352265835f, 0.129445076f, -1.f),
    C_Vertex3f(0.340202808f, 0.11978817f, -1.f),
    C_Vertex3f(0.341779709f, -0.0191392899f, -1.f),
    C_Vertex3f(0.345697403f, -0.0293951035f, -1.f),
    C_Vertex3f(0.34765625f, -0.0400886536f, -1.f),
    C_Vertex3f(0.3522892f, -0.165343285f, -1.f),
    C_Vertex3f(0.357307434f, 0.0778579712f, -1.f),
    C_Vertex3f(0.362301826f, 0.139783859f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318054199f, 0.106666565f, -1.f),
    C_Vertex3f(0.318918228f, 0.00893878937f, -1.f),
    C_Vertex3f(0.326112747f, 0.110813141f, -1.f),
    C_Vertex3f(0.328611374f, -0.000539779663f, -1.f),
    C_Vertex3f(0.340202808f, 0.11978817f, -1.f),
    C_Vertex3f(0.335903168f, -0.00932121277f, -1.f),
    C_Vertex3f(0.341779709f, -0.0191392899f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.30749321f, 0.11419487f, -1.f),
    C_Vertex3f(0.284356117f, 0.129797935f, -1.f),
    C_Vertex3f(0.290697098f, 0.0317840576f, -1.f),
    C_Vertex3f(0.306280136f, 0.0197134018f, -1.f),
    C_Vertex3f(0.318918228f, 0.00893878937f, -1.f),
    C_Vertex3f(0.318054199f, 0.106666565f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.290697098f, 0.0317840576f, -1.f),
    C_Vertex3f(0.284356117f, 0.129797935f, -1.f),
    C_Vertex3f(0.272169113f, 0.0451507568f, -1.f),
    C_Vertex3f(0.259203911f, 0.145947456f, -1.f),
    C_Vertex3f(0.250696182f, 0.0598134995f, -1.f),
    C_Vertex3f(0.232036591f, 0.162643433f, -1.f),
    C_Vertex3f(0.226278305f, 0.0757722855f, -1.f),
    C_Vertex3f(0.208680153f, 0.176447868f, -1.f),
    C_Vertex3f(0.198915482f, 0.0930271149f, -1.f),
    C_Vertex3f(0.192709923f, 0.186295986f, -1.f),
    C_Vertex3f(0.178299904f, 0.195625782f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.198915482f, 0.0930271149f, -1.f),
    C_Vertex3f(0.178299904f, 0.195625782f, -1.f),
    C_Vertex3f(0.175694466f, 0.10726738f, -1.f),
    C_Vertex3f(0.165450096f, 0.204437256f, -1.f),
    C_Vertex3f(0.160776973f, 0.11665833f, -1.f),
    C_Vertex3f(0.1541605f, 0.212730408f, -1.f),
    C_Vertex3f(0.147076249f, 0.125510573f, -1.f),
    C_Vertex3f(0.144431114f, 0.220505238f, -1.f),
    C_Vertex3f(0.13626194f, 0.227761745f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147076249f, 0.125510573f, -1.f),
    C_Vertex3f(0.13626194f, 0.227761745f, -1.f),
    C_Vertex3f(0.134592295f, 0.13382411f, -1.f),
    C_Vertex3f(0.129652977f, 0.234499931f, -1.f),
    C_Vertex3f(0.123325109f, 0.14159894f, -1.f),
    C_Vertex3f(0.122079849f, 0.243829727f, -1.f),
    C_Vertex3f(0.114506721f, 0.256597519f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.113274693f, 0.148835063f, -1.f),
    C_Vertex3f(0.123325109f, 0.14159894f, -1.f),
    C_Vertex3f(0.114506721f, 0.256597519f, -1.f),
    C_Vertex3f(0.10945797f, 0.269693375f, -1.f),
    C_Vertex3f(0.106933594f, 0.283117294f, -1.f),
    C_Vertex3f(0.104441047f, 0.155532479f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.359550476f, 0.24066925f, -1.f),
    C_Vertex3f(0.365243912f, 0.233793259f, -1.f),
    C_Vertex3f(0.366908073f, 0.338848591f, -1.f),
    C_Vertex3f(0.373784065f, 0.220369339f, -1.f),
    C_Vertex3f(0.377077103f, 0.331851244f, -1.f),
    C_Vertex3f(0.379477501f, 0.207273483f, -1.f),
    C_Vertex3f(0.385858536f, 0.325549841f, -1.f),
    C_Vertex3f(0.382324219f, 0.194505692f, -1.f),
    C_Vertex3f(0.390205383f, 0.0931901932f, -1.f),
    C_Vertex3f(0.382324219f, 0.183048248f, -1.f),
    C_Vertex3f(0.38032198f, 0.172436714f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329548836f, 0.268249035f, -1.f),
    C_Vertex3f(0.340196609f, 0.259614706f, -1.f),
    C_Vertex3f(0.342407227f, 0.354931116f, -1.f),
    C_Vertex3f(0.349229813f, 0.251567125f, -1.f),
    C_Vertex3f(0.355351448f, 0.346541882f, -1.f),
    C_Vertex3f(0.356648445f, 0.244106293f, -1.f),
    C_Vertex3f(0.366908073f, 0.338848591f, -1.f),
    C_Vertex3f(0.359550476f, 0.24066925f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.270812035f, 0.308653831f, -1.f),
    C_Vertex3f(0.287918091f, 0.29767251f, -1.f),
    C_Vertex3f(0.290182114f, 0.387329102f, -1.f),
    C_Vertex3f(0.303409576f, 0.287277937f, -1.f),
    C_Vertex3f(0.312355995f, 0.373797417f, -1.f),
    C_Vertex3f(0.317286491f, 0.277470112f, -1.f),
    C_Vertex3f(0.328075409f, 0.364016294f, -1.f),
    C_Vertex3f(0.329548836f, 0.268249035f, -1.f),
    C_Vertex3f(0.342407227f, 0.354931116f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227030039f, 0.336123228f, -1.f),
    C_Vertex3f(0.240036726f, 0.327776194f, -1.f),
    C_Vertex3f(0.243629932f, 0.417074919f, -1.f),
    C_Vertex3f(0.253995895f, 0.319020271f, -1.f),
    C_Vertex3f(0.265584469f, 0.402773142f, -1.f),
    C_Vertex3f(0.270812035f, 0.308653831f, -1.f),
    C_Vertex3f(0.290182114f, 0.387329102f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193724871f, 0.358711004f, -1.f),
    C_Vertex3f(0.203874111f, 0.351590633f, -1.f),
    C_Vertex3f(0.207650185f, 0.442251682f, -1.f),
    C_Vertex3f(0.214975834f, 0.344061375f, -1.f),
    C_Vertex3f(0.224318504f, 0.430234432f, -1.f),
    C_Vertex3f(0.227030039f, 0.336123228f, -1.f),
    C_Vertex3f(0.243629932f, 0.417074919f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160661697f, 0.490076065f, -1.f),
    C_Vertex3f(0.163784027f, 0.369651794f, -1.f),
    C_Vertex3f(0.166538239f, 0.479995728f, -1.f),
    C_Vertex3f(0.172637939f, 0.374671936f, -1.f),
    C_Vertex3f(0.173503876f, 0.471449852f, -1.f),
    C_Vertex3f(0.176283836f, 0.371725082f, -1.f),
    C_Vertex3f(0.18224287f, 0.462859392f, -1.f),
    C_Vertex3f(0.184528112f, 0.365422487f, -1.f),
    C_Vertex3f(0.193624973f, 0.453126669f, -1.f),
    C_Vertex3f(0.193724871f, 0.358711004f, -1.f),
    C_Vertex3f(0.207650185f, 0.442251682f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160661697f, 0.490076065f, -1.f),
    C_Vertex3f(0.156744003f, 0.500593185f, -1.f),
    C_Vertex3f(0.163784027f, 0.369651794f, -1.f),
    C_Vertex3f(0.154785156f, 0.511547089f, -1.f),
    C_Vertex3f(0.148610115f, 0.359611511f, -1.f),
    C_Vertex3f(0.151230931f, 0.632103562f, -1.f),
    C_Vertex3f(0.13854444f, 0.624398112f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0790008307f, 0.533735156f, -1.f),
    C_Vertex3f(0.0792250633f, 0.49230957f, -1.f),
    C_Vertex3f(0.0817290545f, 0.546796441f, -1.f),
    C_Vertex3f(0.0824017525f, 0.478919983f, -1.f),
    C_Vertex3f(0.0858213902f, 0.559393167f, -1.f),
    C_Vertex3f(0.0871667862f, 0.465940475f, -1.f),
    C_Vertex3f(0.0912778378f, 0.571525335f, -1.f),
    C_Vertex3f(0.0935201645f, 0.453371048f, -1.f),
    C_Vertex3f(0.0980983973f, 0.583192945f, -1.f),
    C_Vertex3f(0.102173328f, 0.440227509f, -1.f),
    C_Vertex3f(0.106283069f, 0.594395995f, -1.f),
    C_Vertex3f(0.113126278f, 0.426509857f, -1.f),
    C_Vertex3f(0.115831852f, 0.605134487f, -1.f),
    C_Vertex3f(0.126379013f, 0.412218094f, -1.f),
    C_Vertex3f(0.126744747f, 0.615408421f, -1.f),
    C_Vertex3f(0.134155273f, 0.404785156f, -1.f),
    C_Vertex3f(0.13854444f, 0.624398112f, -1.f),
    C_Vertex3f(0.135970116f, 0.349571228f, -1.f),
    C_Vertex3f(0.148610115f, 0.359611511f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0776367188f, 0.520209312f, -1.f),
    C_Vertex3f(0.0776367188f, 0.506109238f, -1.f),
    C_Vertex3f(0.0792250633f, 0.49230957f, -1.f),
    C_Vertex3f(0.0790008307f, 0.533735156f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.44656992f, 0.151704311f, -1.f),
    C_Vertex3f(0.450559139f, 0.162233829f, -1.f),
    C_Vertex3f(0.451055527f, 0.241811275f, -1.f),
    C_Vertex3f(0.453551054f, 0.17321372f, -1.f),
    C_Vertex3f(0.454713821f, 0.226643085f, -1.f),
    C_Vertex3f(0.455545664f, 0.184643984f, -1.f),
    C_Vertex3f(0.456542969f, 0.210874557f, -1.f),
    C_Vertex3f(0.456542969f, 0.19652462f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.420040131f, 0.113433838f, -1.f),
    C_Vertex3f(0.428618431f, 0.122817993f, -1.f),
    C_Vertex3f(0.428886414f, 0.284096241f, -1.f),
    C_Vertex3f(0.435599566f, 0.131996393f, -1.f),
    C_Vertex3f(0.438251495f, 0.270346642f, -1.f),
    C_Vertex3f(0.441583395f, 0.141625166f, -1.f),
    C_Vertex3f(0.445568085f, 0.256379128f, -1.f),
    C_Vertex3f(0.44656992f, 0.151704311f, -1.f),
    C_Vertex3f(0.451055527f, 0.241811275f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.417472839f, 0.297627926f, -1.f),
    C_Vertex3f(0.404010773f, 0.310941696f, -1.f),
    C_Vertex3f(0.407035828f, 0.103339195f, -1.f),
    C_Vertex3f(0.420040131f, 0.113433838f, -1.f),
    C_Vertex3f(0.428886414f, 0.284096241f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.385858536f, 0.325549841f, -1.f),
    C_Vertex3f(0.390205383f, 0.0931901932f, -1.f),
    C_Vertex3f(0.393252373f, 0.319944382f, -1.f),
    C_Vertex3f(0.407035828f, 0.103339195f, -1.f),
    C_Vertex3f(0.396255493f, 0.317489624f, -1.f),
    C_Vertex3f(0.404010773f, 0.310941696f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.38032198f, 0.172436714f, -1.f),
    C_Vertex3f(0.376317501f, 0.161688805f, -1.f),
    C_Vertex3f(0.390205383f, 0.0931901932f, -1.f),
    C_Vertex3f(0.370310783f, 0.15080452f, -1.f),
    C_Vertex3f(0.369548798f, 0.0829868317f, -1.f),
    C_Vertex3f(0.362301826f, 0.139783859f, -1.f),
    C_Vertex3f(0.357307434f, 0.0778579712f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.3522892f, -0.165343285f, -1.f),
    C_Vertex3f(0.34765625f, -0.0400886536f, -1.f),
    C_Vertex3f(0.34765625f, -0.0532894135f, -1.f),
    C_Vertex3f(0.344518661f, -0.0675787926f, -1.f),
    C_Vertex3f(0.338243484f, -0.0808873177f, -1.f),
    C_Vertex3f(0.331762314f, -0.174401283f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338243484f, -0.0808873177f, -1.f),
    C_Vertex3f(0.328830719f, -0.0932149887f, -1.f),
    C_Vertex3f(0.331762314f, -0.174401283f, -1.f),
    C_Vertex3f(0.316280365f, -0.104561806f, -1.f),
    C_Vertex3f(0.310416222f, -0.181194782f, -1.f),
    C_Vertex3f(0.301301003f, -0.113071918f, -1.f),
    C_Vertex3f(0.288250923f, -0.185723782f, -1.f),
    C_Vertex3f(0.283892632f, -0.118745327f, -1.f),
    C_Vertex3f(0.265266418f, -0.187988281f, -1.f),
    C_Vertex3f(0.264055252f, -0.121582031f, -1.f),
    C_Vertex3f(0.242937088f, -0.187988281f, -1.f),
    C_Vertex3f(0.241462708f, -0.121582031f, -1.f),
    C_Vertex3f(0.223160267f, -0.186684489f, -1.f),
    C_Vertex3f(0.220944881f, -0.118718624f, -1.f),
    C_Vertex3f(0.204461575f, -0.184076905f, -1.f),
    C_Vertex3f(0.20282793f, -0.11299181f, -1.f),
    C_Vertex3f(0.187111855f, -0.104401588f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.186841011f, -0.180165529f, -1.f),
    C_Vertex3f(0.204461575f, -0.184076905f, -1.f),
    C_Vertex3f(0.187111855f, -0.104401588f, -1.f),
    C_Vertex3f(0.177125454f, -0.095811367f, -1.f),
    C_Vertex3f(0.170918107f, -0.0890277624f, -1.f),
    C_Vertex3f(0.170298576f, -0.174950361f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.170298576f, -0.174950361f, -1.f),
    C_Vertex3f(0.170918107f, -0.0890277624f, -1.f),
    C_Vertex3f(0.165161014f, -0.0811873674f, -1.f),
    C_Vertex3f(0.159854174f, -0.0722901821f, -1.f),
    C_Vertex3f(0.154997587f, -0.0623362064f, -1.f),
    C_Vertex3f(0.15483427f, -0.168431401f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15483427f, -0.168431401f, -1.f),
    C_Vertex3f(0.154997587f, -0.0623362064f, -1.f),
    C_Vertex3f(0.150591254f, -0.0513254404f, -1.f),
    C_Vertex3f(0.146635175f, -0.039257884f, -1.f),
    C_Vertex3f(0.143129349f, -0.0261335373f, -1.f),
    C_Vertex3f(0.141601563f, -0.0190429688f, -1.f),
    C_Vertex3f(0.140448093f, -0.160608649f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.141601563f, -0.0190429688f, -1.f),
    C_Vertex3f(0.0625f, -0.0302734375f, -1.f),
    C_Vertex3f(0.063700676f, -0.039557457f, -1.f),
    C_Vertex3f(0.0673305988f, -0.057108283f, -1.f),
    C_Vertex3f(0.0721890926f, -0.0736418962f, -1.f),
    C_Vertex3f(0.0782761574f, -0.0891582966f, -1.f),
    C_Vertex3f(0.0855917931f, -0.103657484f, -1.f),
    C_Vertex3f(0.0941359997f, -0.117139459f, -1.f),
    C_Vertex3f(0.103908777f, -0.12960422f, -1.f),
    C_Vertex3f(0.114910126f, -0.141051769f, -1.f),
    C_Vertex3f(0.127140045f, -0.151482105f, -1.f),
    C_Vertex3f(0.140448093f, -0.160608649f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.366043091f, 0.0712013245f, -1.f),
    C_Vertex3f(0.370414257f, -0.154266357f, -1.f),
    C_Vertex3f(0.38130331f, 0.0573964119f, -1.f),
    C_Vertex3f(0.386137486f, -0.141170502f, -1.f),
    C_Vertex3f(0.394352436f, 0.0430998802f, -1.f),
    C_Vertex3f(0.399458885f, -0.126055717f, -1.f),
    C_Vertex3f(0.405190468f, 0.0283117294f, -1.f),
    C_Vertex3f(0.410378456f, -0.108922005f, -1.f),
    C_Vertex3f(0.413817406f, 0.0130319595f, -1.f),
    C_Vertex3f(0.418568134f, -0.0912699699f, -1.f),
    C_Vertex3f(0.420287609f, -0.00279283524f, -1.f),
    C_Vertex3f(0.42402792f, -0.0730996132f, -1.f),
    C_Vertex3f(0.424601078f, -0.0191626549f, -1.f),
    C_Vertex3f(0.426757813f, -0.0544109344f, -1.f),
    C_Vertex3f(0.426757813f, -0.0360774994f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.357307434f, 0.0778579712f, -1.f),
    C_Vertex3f(0.3522892f, -0.165343285f, -1.f),
    C_Vertex3f(0.370414257f, -0.154266357f, -1.f),
    C_Vertex3f(0.366043091f, 0.0712013245f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.419433594f, 0.499511719f, -1.f),
    C_Vertex3f(0.418287277f, 0.508756638f, -1.f),
    C_Vertex3f(0.414848924f, 0.526152372f, -1.f),
    C_Vertex3f(0.41026485f, 0.542454004f, -1.f),
    C_Vertex3f(0.404535055f, 0.557661533f, -1.f),
    C_Vertex3f(0.39765954f, 0.57177496f, -1.f),
    C_Vertex3f(0.389638305f, 0.584794283f, -1.f),
    C_Vertex3f(0.380471349f, 0.596719503f, -1.f),
    C_Vertex3f(0.370158672f, 0.607550621f, -1.f),
    C_Vertex3f(0.358700275f, 0.617287636f, -1.f),
    C_Vertex3f(0.346123338f, 0.625807524f, -1.f),
    C_Vertex3f(0.337890625f, 0.491210938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.32730484f, 0.531983376f, -1.f),
    C_Vertex3f(0.330881119f, 0.521639347f, -1.f),
    C_Vertex3f(0.332427859f, 0.633110285f, -1.f),
    C_Vertex3f(0.334028244f, 0.510280132f, -1.f),
    C_Vertex3f(0.346123338f, 0.625807524f, -1.f),
    C_Vertex3f(0.336746216f, 0.497905731f, -1.f),
    C_Vertex3f(0.337890625f, 0.491210938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308708191f, 0.563207626f, -1.f),
    C_Vertex3f(0.314001083f, 0.556924343f, -1.f),
    C_Vertex3f(0.31761384f, 0.639195919f, -1.f),
    C_Vertex3f(0.318864822f, 0.549625874f, -1.f),
    C_Vertex3f(0.332427859f, 0.633110285f, -1.f),
    C_Vertex3f(0.323299408f, 0.541312218f, -1.f),
    C_Vertex3f(0.32730484f, 0.531983376f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.30168128f, 0.644064426f, -1.f),
    C_Vertex3f(0.284630179f, 0.647715807f, -1.f),
    C_Vertex3f(0.286691666f, 0.579011917f, -1.f),
    C_Vertex3f(0.300125122f, 0.571109772f, -1.f),
    C_Vertex3f(0.308708191f, 0.563207626f, -1.f),
    C_Vertex3f(0.31761384f, 0.639195919f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157864094f, 0.536385536f, -1.f),
    C_Vertex3f(0.164021969f, 0.548380852f, -1.f),
    C_Vertex3f(0.16480422f, 0.638524771f, -1.f),
    C_Vertex3f(0.173258781f, 0.559801102f, -1.f),
    C_Vertex3f(0.179264307f, 0.643661737f, -1.f),
    C_Vertex3f(0.185574532f, 0.570646286f, -1.f),
    C_Vertex3f(0.194611192f, 0.647514462f, -1.f),
    C_Vertex3f(0.19982481f, 0.578780174f, -1.f),
    C_Vertex3f(0.210844874f, 0.650082946f, -1.f),
    C_Vertex3f(0.216009617f, 0.584202766f, -1.f),
    C_Vertex3f(0.227965355f, 0.651367188f, -1.f),
    C_Vertex3f(0.234128952f, 0.586914063f, -1.f),
    C_Vertex3f(0.247172356f, 0.651367188f, -1.f),
    C_Vertex3f(0.253856659f, 0.586914063f, -1.f),
    C_Vertex3f(0.266460538f, 0.650150061f, -1.f),
    C_Vertex3f(0.271268845f, 0.584280014f, -1.f),
    C_Vertex3f(0.284630179f, 0.647715807f, -1.f),
    C_Vertex3f(0.286691666f, 0.579011917f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.154785156f, 0.523815155f, -1.f),
    C_Vertex3f(0.157864094f, 0.536385536f, -1.f),
    C_Vertex3f(0.16480422f, 0.638524771f, -1.f),
    C_Vertex3f(0.151230931f, 0.632103562f, -1.f),
    C_Vertex3f(0.154785156f, 0.511547089f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.112612724f, 0.31890583f, -1.f),
    C_Vertex3f(0.118291855f, 0.329490662f, -1.f),
    C_Vertex3f(0.122661591f, 0.40140152f, -1.f),
    C_Vertex3f(0.125864029f, 0.339530945f, -1.f),
    C_Vertex3f(0.134155273f, 0.404785156f, -1.f),
    C_Vertex3f(0.135970116f, 0.349571228f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0968241692f, 0.161691189f, -1.f),
    C_Vertex3f(0.104441047f, 0.155532479f, -1.f),
    C_Vertex3f(0.102137566f, 0.39207077f, -1.f),
    C_Vertex3f(0.106933594f, 0.296102524f, -1.f),
    C_Vertex3f(0.122661591f, 0.40140152f, -1.f),
    C_Vertex3f(0.108826637f, 0.307776451f, -1.f),
    C_Vertex3f(0.112612724f, 0.31890583f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0558815002f, 0.209895134f, -1.f),
    C_Vertex3f(0.0651526451f, 0.195628166f, -1.f),
    C_Vertex3f(0.068479538f, 0.365718842f, -1.f),
    C_Vertex3f(0.0756001472f, 0.182370186f, -1.f),
    C_Vertex3f(0.0840768814f, 0.380176544f, -1.f),
    C_Vertex3f(0.0872240067f, 0.170121193f, -1.f),
    C_Vertex3f(0.102137566f, 0.39207077f, -1.f),
    C_Vertex3f(0.0968241692f, 0.161691189f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0356445313f, 0.291734695f, -1.f),
    C_Vertex3f(0.0376682281f, 0.256705761f, -1.f),
    C_Vertex3f(0.0389280319f, 0.311705589f, -1.f),
    C_Vertex3f(0.0417156219f, 0.240774632f, -1.f),
    C_Vertex3f(0.0454950333f, 0.330693245f, -1.f),
    C_Vertex3f(0.0477867126f, 0.225171089f, -1.f),
    C_Vertex3f(0.0553455353f, 0.348697662f, -1.f),
    C_Vertex3f(0.0558815002f, 0.209895134f, -1.f),
    C_Vertex3f(0.068479538f, 0.365718842f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0376682281f, 0.256705761f, -1.f),
    C_Vertex3f(0.0356445313f, 0.291734695f, -1.f),
    C_Vertex3f(0.0356445313f, 0.272964478f, -1.f),
    C_Vertex3f(0.106933594f, 0.296102524f, -1.f),
    C_Vertex3f(0.104441047f, 0.155532479f, -1.f),
    C_Vertex3f(0.106933594f, 0.283117294f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_168[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0263671875f, 0.555175781f, 0.f),
    C_Vertex3f(0.0263671875f, 0.555175781f, -1.f),
    C_Vertex3f(0.0263671875f, 0.64453125f, -1.f),
    C_Vertex3f(0.0263671875f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0263671875f, 0.64453125f, 0.f),
    C_Vertex3f(0.0263671875f, 0.64453125f, -1.f),
    C_Vertex3f(0.108398438f, 0.64453125f, -1.f),
    C_Vertex3f(0.108398438f, 0.64453125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.108398438f, 0.64453125f, 0.f),
    C_Vertex3f(0.108398438f, 0.64453125f, -1.f),
    C_Vertex3f(0.108398438f, 0.555175781f, -1.f),
    C_Vertex3f(0.108398438f, 0.555175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.108398438f, 0.555175781f, 0.f),
    C_Vertex3f(0.108398438f, 0.555175781f, -1.f),
    C_Vertex3f(0.0263671875f, 0.555175781f, -1.f),
    C_Vertex3f(0.0263671875f, 0.555175781f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.189453125f, 0.555175781f, 0.f),
    C_Vertex3f(0.189453125f, 0.555175781f, -1.f),
    C_Vertex3f(0.189453125f, 0.64453125f, -1.f),
    C_Vertex3f(0.189453125f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.189453125f, 0.64453125f, 0.f),
    C_Vertex3f(0.189453125f, 0.64453125f, -1.f),
    C_Vertex3f(0.271484375f, 0.64453125f, -1.f),
    C_Vertex3f(0.271484375f, 0.64453125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.271484375f, 0.64453125f, 0.f),
    C_Vertex3f(0.271484375f, 0.64453125f, -1.f),
    C_Vertex3f(0.271484375f, 0.555175781f, -1.f),
    C_Vertex3f(0.271484375f, 0.555175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.271484375f, 0.555175781f, 0.f),
    C_Vertex3f(0.271484375f, 0.555175781f, -1.f),
    C_Vertex3f(0.189453125f, 0.555175781f, -1.f),
    C_Vertex3f(0.189453125f, 0.555175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0263671875f, 0.64453125f, 0.f),
    C_Vertex3f(0.0263671875f, 0.555175781f, 0.f),
    C_Vertex3f(0.108398438f, 0.555175781f, 0.f),
    C_Vertex3f(0.108398438f, 0.64453125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.189453125f, 0.64453125f, 0.f),
    C_Vertex3f(0.189453125f, 0.555175781f, 0.f),
    C_Vertex3f(0.271484375f, 0.555175781f, 0.f),
    C_Vertex3f(0.271484375f, 0.64453125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0263671875f, 0.64453125f, -1.f),
    C_Vertex3f(0.0263671875f, 0.555175781f, -1.f),
    C_Vertex3f(0.108398438f, 0.555175781f, -1.f),
    C_Vertex3f(0.108398438f, 0.64453125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.189453125f, 0.64453125f, -1.f),
    C_Vertex3f(0.189453125f, 0.555175781f, -1.f),
    C_Vertex3f(0.271484375f, 0.555175781f, -1.f),
    C_Vertex3f(0.271484375f, 0.64453125f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_169[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0321503282f, 0.999482989f, 0.f),
    C_Vertex3f(0.321137428f, 0.651855469f, 0.f),
    C_Vertex3f(0.321137428f, 0.651855469f, -1.f),
    C_Normal3f(0.0321515314f, 0.999483049f, 0.f),
    C_Vertex3f(0.341901779f, 0.651855469f, 0.f),
    C_Vertex3f(0.341901779f, 0.651855469f, -1.f),
    C_Normal3f(0.0966276675f, 0.995320618f, 0.f),
    C_Vertex3f(0.36254251f, 0.650523543f, 0.f),
    C_Vertex3f(0.36254251f, 0.650523543f, -1.f),
    C_Normal3f(0.16065906f, 0.987010002f, 0.f),
    C_Vertex3f(0.383060575f, 0.647859693f, 0.f),
    C_Vertex3f(0.383060575f, 0.647859693f, -1.f),
    C_Normal3f(0.223460346f, 0.974712968f, 0.f),
    C_Vertex3f(0.403455973f, 0.643863916f, 0.f),
    C_Vertex3f(0.403455973f, 0.643863916f, -1.f),
    C_Normal3f(0.284310788f, 0.958732128f, 0.f),
    C_Vertex3f(0.423728704f, 0.638536215f, 0.f),
    C_Vertex3f(0.423728704f, 0.638536215f, -1.f),
    C_Normal3f(0.34259367f, 0.939483702f, 0.f),
    C_Vertex3f(0.44387877f, 0.631876588f, 0.f),
    C_Vertex3f(0.44387877f, 0.631876588f, -1.f),
    C_Normal3f(0.397822261f, 0.917462528f, 0.f),
    C_Vertex3f(0.463906169f, 0.623885036f, 0.f),
    C_Vertex3f(0.463906169f, 0.623885036f, -1.f),
    C_Normal3f(0.449651271f, 0.893204153f, 0.f),
    C_Vertex3f(0.483810902f, 0.614561558f, 0.f),
    C_Vertex3f(0.483810902f, 0.614561558f, -1.f),
    C_Normal3f(0.50415504f, 0.863613188f, 0.f),
    C_Vertex3f(0.503592968f, 0.603906155f, 0.f),
    C_Vertex3f(0.503592968f, 0.603906155f, -1.f),
    C_Normal3f(0.563552618f, 0.826080143f, 0.f),
    C_Vertex3f(0.522282004f, 0.592103124f, 0.f),
    C_Vertex3f(0.522282004f, 0.592103124f, -1.f),
    C_Normal3f(0.621464193f, 0.783442616f, 0.f),
    C_Vertex3f(0.539878011f, 0.579152465f, 0.f),
    C_Vertex3f(0.539878011f, 0.579152465f, -1.f),
    C_Normal3f(0.676860511f, 0.736111283f, 0.f),
    C_Vertex3f(0.556380987f, 0.565054178f, 0.f),
    C_Vertex3f(0.556380987f, 0.565054178f, -1.f),
    C_Normal3f(0.728791237f, 0.684736073f, 0.f),
    C_Vertex3f(0.571790934f, 0.549808264f, 0.f),
    C_Vertex3f(0.571790934f, 0.549808264f, -1.f),
    C_Normal3f(0.776462734f, 0.630163252f, 0.f),
    C_Vertex3f(0.58610785f, 0.533414721f, 0.f),
    C_Vertex3f(0.58610785f, 0.533414721f, -1.f),
    C_Normal3f(0.819298148f, 0.573367715f, 0.f),
    C_Vertex3f(0.599331737f, 0.515873551f, 0.f),
    C_Vertex3f(0.599331737f, 0.515873551f, -1.f),
    C_Normal3f(0.856966972f, 0.515371323f, 0.f),
    C_Vertex3f(0.611462593f, 0.497184753f, 0.f),
    C_Vertex3f(0.611462593f, 0.497184753f, -1.f),
    C_Normal3f(0.887617767f, 0.460580915f, 0.f),
    C_Vertex3f(0.62250042f, 0.477348328f, 0.f),
    C_Vertex3f(0.62250042f, 0.477348328f, -1.f),
    C_Normal3f(0.913628042f, 0.406551063f, 0.f),
    C_Vertex3f(0.632158518f, 0.457300186f, 0.f),
    C_Vertex3f(0.632158518f, 0.457300186f, -1.f),
    C_Normal3f(0.93706286f, 0.349160671f, 0.f),
    C_Vertex3f(0.640436888f, 0.437040329f, 0.f),
    C_Vertex3f(0.640436888f, 0.437040329f, -1.f),
    C_Normal3f(0.957369864f, 0.288864851f, 0.f),
    C_Vertex3f(0.647335529f, 0.416568756f, 0.f),
    C_Vertex3f(0.647335529f, 0.416568756f, -1.f),
    C_Normal3f(0.974066257f, 0.226262957f, 0.f),
    C_Vertex3f(0.652854443f, 0.395885468f, 0.f),
    C_Vertex3f(0.652854443f, 0.395885468f, -1.f),
    C_Normal3f(0.986778498f, 0.162075043f, 0.f),
    C_Vertex3f(0.656993628f, 0.374990463f, 0.f),
    C_Vertex3f(0.656993628f, 0.374990463f, -1.f),
    C_Normal3f(0.995274305f, 0.0971033424f, 0.f),
    C_Vertex3f(0.659753084f, 0.353883743f, 0.f),
    C_Vertex3f(0.659753084f, 0.353883743f, -1.f),
    C_Normal3f(0.999479294f, 0.032265909f, 0.f),
    C_Vertex3f(0.661132813f, 0.332565308f, 0.f),
    C_Vertex3f(0.661132813f, 0.332565308f, -1.f),
    C_Normal3f(0.999489665f, -0.0319455191f, 0.f),
    C_Vertex3f(0.661132813f, 0.311144829f, 0.f),
    C_Vertex3f(0.661132813f, 0.311144829f, -1.f),
    C_Normal3f(0.995321691f, -0.096616663f, 0.f),
    C_Vertex3f(0.659773588f, 0.290038824f, 0.f),
    C_Vertex3f(0.659773588f, 0.290038824f, -1.f),
    C_Normal3f(0.986915231f, -0.161239952f, 0.f),
    C_Vertex3f(0.65705514f, 0.269137621f, 0.f),
    C_Vertex3f(0.65705514f, 0.269137621f, -1.f),
    C_Normal3f(0.974341154f, -0.225076169f, 0.f),
    C_Vertex3f(0.652977467f, 0.248441219f, 0.f),
    C_Vertex3f(0.652977467f, 0.248441219f, -1.f),
    C_Normal3f(0.957829475f, -0.287337095f, 0.f),
    C_Vertex3f(0.647540569f, 0.227949619f, 0.f),
    C_Vertex3f(0.647540569f, 0.227949619f, -1.f),
    C_Normal3f(0.937748313f, -0.347315639f, 0.f),
    C_Vertex3f(0.640744448f, 0.207662821f, 0.f),
    C_Vertex3f(0.640744448f, 0.207662821f, -1.f),
    C_Normal3f(0.91457212f, -0.404422909f, 0.f),
    C_Vertex3f(0.632589102f, 0.187580824f, 0.f),
    C_Vertex3f(0.632589102f, 0.187580824f, -1.f),
    C_Normal3f(0.888843417f, -0.458211154f, 0.f),
    C_Vertex3f(0.623074532f, 0.167703629f, 0.f),
    C_Vertex3f(0.623074532f, 0.167703629f, -1.f),
    C_Normal3f(0.858544528f, -0.512739122f, 0.f),
    C_Vertex3f(0.612200737f, 0.148031235f, 0.f),
    C_Vertex3f(0.612200737f, 0.148031235f, -1.f),
    C_Normal3f(0.821421087f, -0.570322216f, 0.f),
    C_Vertex3f(0.600233912f, 0.129458666f, 0.f),
    C_Vertex3f(0.600233912f, 0.129458666f, -1.f),
    C_Normal3f(0.779336512f, -0.62660557f, 0.f),
    C_Vertex3f(0.587174058f, 0.111985922f, 0.f),
    C_Vertex3f(0.587174058f, 0.111985922f, -1.f),
    C_Normal3f(0.732640922f, -0.680615366f, 0.f),
    C_Vertex3f(0.573021173f, 0.0956130028f, 0.f),
    C_Vertex3f(0.573021173f, 0.0956130028f, -1.f),
    C_Normal3f(0.681911707f, -0.731434464f, 0.f),
    C_Vertex3f(0.557775259f, 0.0803399086f, 0.f),
    C_Vertex3f(0.557775259f, 0.0803399086f, -1.f),
    C_Normal3f(0.627921402f, -0.778276742f, 0.f),
    C_Vertex3f(0.541436315f, 0.0661666393f, 0.f),
    C_Vertex3f(0.541436315f, 0.0661666393f, -1.f),
    C_Normal3f(0.571579754f, -0.820546567f, 0.f),
    C_Vertex3f(0.52400434f, 0.053093195f, 0.f),
    C_Vertex3f(0.52400434f, 0.053093195f, -1.f),
    C_Normal3f(0.513860404f, -0.857873738f, 0.f),
    C_Vertex3f(0.505479336f, 0.0411195755f, 0.f),
    C_Vertex3f(0.505479336f, 0.0411195755f, -1.f),
    C_Normal3f(0.459020972f, -0.888425529f, 0.f),
    C_Vertex3f(0.485861301f, 0.0302457809f, 0.f),
    C_Vertex3f(0.485861301f, 0.0302457809f, -1.f),
    C_Normal3f(0.404825926f, -0.914393783f, 0.f),
    C_Vertex3f(0.466018081f, 0.0207312107f, 0.f),
    C_Vertex3f(0.466018081f, 0.0207312107f, -1.f),
    C_Normal3f(0.347367018f, -0.93772918f, 0.f),
    C_Vertex3f(0.445949674f, 0.0125758648f, 0.f),
    C_Vertex3f(0.445949674f, 0.0125758648f, -1.f),
    C_Normal3f(0.28711763f, -0.957895339f, 0.f),
    C_Vertex3f(0.42565608f, 0.00577974319f, 0.f),
    C_Vertex3f(0.42565608f, 0.00577974319f, -1.f),
    C_Normal3f(0.224689066f, -0.974430501f, 0.f),
    C_Vertex3f(0.4051373f, 0.000342845917f, 0.f),
    C_Vertex3f(0.4051373f, 0.000342845917f, -1.f),
    C_Normal3f(0.160805002f, -0.98698616f, 0.f),
    C_Vertex3f(0.384393334f, -0.00373482704f, 0.f),
    C_Vertex3f(0.384393334f, -0.00373482704f, -1.f),
    C_Normal3f(0.0962618664f, -0.995356023f, 0.f),
    C_Vertex3f(0.363424182f, -0.00645327568f, 0.f),
    C_Vertex3f(0.363424182f, -0.00645327568f, -1.f),
    C_Normal3f(0.0318793692f, -0.999491751f, 0.f),
    C_Vertex3f(0.342229843f, -0.0078125f, 0.f),
    C_Vertex3f(0.342229843f, -0.0078125f, -1.f),
    C_Normal3f(-0.0318782106f, -0.999491811f, 0.f),
    C_Vertex3f(0.320809364f, -0.0078125f, 0.f),
    C_Vertex3f(0.320809364f, -0.0078125f, -1.f),
    C_Normal3f(-0.0962556675f, -0.995356679f, 0.f),
    C_Vertex3f(0.299613476f, -0.00645327568f, 0.f),
    C_Vertex3f(0.299613476f, -0.00645327568f, -1.f),
    C_Normal3f(-0.160797387f, -0.986987412f, 0.f),
    C_Vertex3f(0.278643131f, -0.00373482704f, 0.f),
    C_Vertex3f(0.278643131f, -0.00373482704f, -1.f),
    C_Normal3f(-0.224682286f, -0.974432111f, 0.f),
    C_Vertex3f(0.257898331f, 0.000342845917f, 0.f),
    C_Vertex3f(0.257898331f, 0.000342845917f, -1.f),
    C_Normal3f(-0.287113786f, -0.957896471f, 0.f),
    C_Vertex3f(0.237379074f, 0.00577974319f, 0.f),
    C_Vertex3f(0.237379074f, 0.00577974319f, -1.f),
    C_Normal3f(-0.347367913f, -0.937728882f, 0.f),
    C_Vertex3f(0.217085361f, 0.0125758648f, 0.f),
    C_Vertex3f(0.217085361f, 0.0125758648f, -1.f),
    C_Normal3f(-0.404833019f, -0.914390683f, 0.f),
    C_Vertex3f(0.197017193f, 0.0207312107f, 0.f),
    C_Vertex3f(0.197017193f, 0.0207312107f, -1.f),
    C_Normal3f(-0.459035158f, -0.888418078f, 0.f),
    C_Vertex3f(0.177174568f, 0.0302457809f, 0.f),
    C_Vertex3f(0.177174568f, 0.0302457809f, -1.f),
    C_Normal3f(-0.513809621f, -0.857904255f, 0.f),
    C_Vertex3f(0.157557487f, 0.0411195755f, 0.f),
    C_Vertex3f(0.157557487f, 0.0411195755f, -1.f),
    C_Normal3f(-0.571374416f, -0.820689559f, 0.f),
    C_Vertex3f(0.139026403f, 0.053093195f, 0.f),
    C_Vertex3f(0.139026403f, 0.053093195f, -1.f),
    C_Normal3f(-0.627547026f, -0.778578699f, 0.f),
    C_Vertex3f(0.121581316f, 0.0661666393f, 0.f),
    C_Vertex3f(0.121581316f, 0.0661666393f, -1.f),
    C_Normal3f(-0.681365311f, -0.731943488f, 0.f),
    C_Vertex3f(0.105222225f, 0.0803399086f, 0.f),
    C_Vertex3f(0.105222225f, 0.0803399086f, -1.f),
    C_Normal3f(-0.731932461f, -0.681377172f, 0.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, 0.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, -1.f),
    C_Normal3f(-0.778488278f, -0.627659202f, 0.f),
    C_Vertex3f(0.0757620335f, 0.111985922f, 0.f),
    C_Vertex3f(0.0757620335f, 0.111985922f, -1.f),
    C_Normal3f(-0.820465326f, -0.571696401f, 0.f),
    C_Vertex3f(0.0626609325f, 0.129458666f, 0.f),
    C_Vertex3f(0.0626609325f, 0.129458666f, -1.f),
    C_Normal3f(-0.857520282f, -0.514450133f, 0.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, 0.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, -1.f),
    C_Normal3f(-0.88789314f, -0.460049629f, 0.f),
    C_Vertex3f(0.0397167206f, 0.167703629f, 0.f),
    C_Vertex3f(0.0397167206f, 0.167703629f, -1.f),
    C_Normal3f(-0.913810194f, -0.40614152f, 0.f),
    C_Vertex3f(0.0301537514f, 0.187580824f, 0.f),
    C_Vertex3f(0.0301537514f, 0.187580824f, -1.f),
    C_Normal3f(-0.937171936f, -0.348867804f, 0.f),
    C_Vertex3f(0.0219569206f, 0.207662821f, 0.f),
    C_Vertex3f(0.0219569206f, 0.207662821f, -1.f),
    C_Normal3f(-0.957426429f, -0.288677186f, 0.f),
    C_Vertex3f(0.0151262283f, 0.227949619f, 0.f),
    C_Vertex3f(0.0151262283f, 0.227949619f, -1.f),
    C_Normal3f(-0.974089563f, -0.226162672f, 0.f),
    C_Vertex3f(0.0096616745f, 0.248441219f, 0.f),
    C_Vertex3f(0.0096616745f, 0.248441219f, -1.f),
    C_Normal3f(-0.986784399f, -0.162038669f, 0.f),
    C_Vertex3f(0.00556325912f, 0.269137621f, 0.f),
    C_Vertex3f(0.00556325912f, 0.269137621f, -1.f),
    C_Normal3f(-0.995274305f, -0.0971035138f, 0.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, 0.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, -1.f),
    C_Normal3f(-0.99948436f, -0.0321078524f, 0.f),
    C_Vertex3f(0.00146484375f, 0.311144829f, 0.f),
    C_Vertex3f(0.00146484375f, 0.311144829f, -1.f),
    C_Normal3f(-0.999474049f, 0.0324274302f, 0.f),
    C_Vertex3f(0.00146484375f, 0.332565308f, 0.f),
    C_Vertex3f(0.00146484375f, 0.332565308f, -1.f),
    C_Normal3f(-0.995227098f, 0.0975853205f, 0.f),
    C_Vertex3f(0.00285148621f, 0.353883743f, 0.f),
    C_Vertex3f(0.00285148621f, 0.353883743f, -1.f),
    C_Normal3f(-0.986648202f, 0.162865728f, 0.f),
    C_Vertex3f(0.00562477112f, 0.374990463f, 0.f),
    C_Vertex3f(0.00562477112f, 0.374990463f, -1.f),
    C_Normal3f(-0.973815799f, 0.227338344f, 0.f),
    C_Vertex3f(0.00978469849f, 0.395885468f, 0.f),
    C_Vertex3f(0.00978469849f, 0.395885468f, -1.f),
    C_Normal3f(-0.956968844f, 0.290190816f, 0.f),
    C_Vertex3f(0.0153312683f, 0.416568756f, 0.f),
    C_Vertex3f(0.0153312683f, 0.416568756f, -1.f),
    C_Normal3f(-0.936489463f, 0.35069567f, 0.f),
    C_Vertex3f(0.0222644806f, 0.437040329f, 0.f),
    C_Vertex3f(0.0222644806f, 0.437040329f, -1.f),
    C_Normal3f(-0.912870347f, 0.408249557f, 0.f),
    C_Vertex3f(0.0305843353f, 0.457300186f, 0.f),
    C_Vertex3f(0.0305843353f, 0.457300186f, -1.f),
    C_Normal3f(-0.886673331f, 0.462396473f, 0.f),
    C_Vertex3f(0.0402908325f, 0.477348328f, 0.f),
    C_Vertex3f(0.0402908325f, 0.477348328f, -1.f),
    C_Normal3f(-0.855948746f, 0.517060578f, 0.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, 0.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, -1.f),
    C_Normal3f(-0.818347096f, 0.574724317f, 0.f),
    C_Vertex3f(0.0635631084f, 0.515873551f, 0.f),
    C_Vertex3f(0.0635631084f, 0.515873551f, -1.f),
    C_Normal3f(-0.775618136f, 0.631202459f, 0.f),
    C_Vertex3f(0.0768282413f, 0.533414721f, 0.f),
    C_Vertex3f(0.0768282413f, 0.533414721f, -1.f),
    C_Normal3f(-0.728085756f, 0.685486019f, 0.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, 0.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, -1.f),
    C_Normal3f(-0.676316857f, 0.73661083f, 0.f),
    C_Vertex3f(0.106616497f, 0.565054178f, 0.f),
    C_Vertex3f(0.106616497f, 0.565054178f, -1.f),
    C_Normal3f(-0.62109226f, 0.783737421f, 0.f),
    C_Vertex3f(0.12313962f, 0.579152465f, 0.f),
    C_Vertex3f(0.12313962f, 0.579152465f, -1.f),
    C_Normal3f(-0.563349247f, 0.826218843f, 0.f),
    C_Vertex3f(0.140748739f, 0.592103124f, 0.f),
    C_Vertex3f(0.140748739f, 0.592103124f, -1.f),
    C_Normal3f(-0.504104972f, 0.863642454f, 0.f),
    C_Vertex3f(0.159443855f, 0.603906155f, 0.f),
    C_Vertex3f(0.159443855f, 0.603906155f, -1.f),
    C_Normal3f(-0.449665278f, 0.893197119f, 0.f),
    C_Vertex3f(0.179224968f, 0.614561558f, 0.f),
    C_Vertex3f(0.179224968f, 0.614561558f, -1.f),
    C_Normal3f(-0.397829235f, 0.917459428f, 0.f),
    C_Vertex3f(0.199129105f, 0.623885036f, 0.f),
    C_Vertex3f(0.199129105f, 0.623885036f, -1.f),
    C_Normal3f(-0.342594564f, 0.939483404f, 0.f),
    C_Vertex3f(0.219156265f, 0.631876588f, 0.f),
    C_Vertex3f(0.219156265f, 0.631876588f, -1.f),
    C_Normal3f(-0.284306943f, 0.95873332f, 0.f),
    C_Vertex3f(0.23930645f, 0.638536215f, 0.f),
    C_Vertex3f(0.23930645f, 0.638536215f, -1.f),
    C_Normal3f(-0.223453522f, 0.974714637f, 0.f),
    C_Vertex3f(0.259579659f, 0.643863916f, 0.f),
    C_Vertex3f(0.259579659f, 0.643863916f, -1.f),
    C_Normal3f(-0.160651311f, 0.987011254f, 0.f),
    C_Vertex3f(0.279975891f, 0.647859693f, 0.f),
    C_Vertex3f(0.279975891f, 0.647859693f, -1.f),
    C_Normal3f(-0.0966212898f, 0.995321214f, 0.f),
    C_Vertex3f(0.300495148f, 0.650523543f, 0.f),
    C_Vertex3f(0.300495148f, 0.650523543f, -1.f),
    C_Normal3f(-0.0321503282f, 0.999482989f, 0.f),
    C_Vertex3f(0.321137428f, 0.651855469f, 0.f),
    C_Vertex3f(0.321137428f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0322615579f, -0.999479473f, 0.f),
    C_Vertex3f(0.340151787f, 0.597167969f, 0.f),
    C_Vertex3f(0.340151787f, 0.597167969f, -1.f),
    C_Normal3f(0.0321776681f, -0.999482155f, 0.f),
    C_Vertex3f(0.322833061f, 0.597167969f, 0.f),
    C_Vertex3f(0.322833061f, 0.597167969f, -1.f),
    C_Normal3f(0.0966247916f, -0.995320916f, 0.f),
    C_Vertex3f(0.305575013f, 0.596054792f, 0.f),
    C_Vertex3f(0.305575013f, 0.596054792f, -1.f),
    C_Normal3f(0.160777718f, -0.986990571f, 0.f),
    C_Vertex3f(0.288432956f, 0.59382844f, 0.f),
    C_Vertex3f(0.288432956f, 0.59382844f, -1.f),
    C_Normal3f(0.223790944f, -0.97463721f, 0.f),
    C_Vertex3f(0.271406889f, 0.590488911f, 0.f),
    C_Vertex3f(0.271406889f, 0.590488911f, -1.f),
    C_Normal3f(0.284929276f, -0.958548546f, 0.f),
    C_Vertex3f(0.254496813f, 0.586036205f, 0.f),
    C_Vertex3f(0.254496813f, 0.586036205f, -1.f),
    C_Normal3f(0.343557268f, -0.939131677f, 0.f),
    C_Vertex3f(0.237702727f, 0.580470324f, 0.f),
    C_Vertex3f(0.237702727f, 0.580470324f, -1.f),
    C_Normal3f(0.399167597f, -0.916878045f, 0.f),
    C_Vertex3f(0.221024632f, 0.573791265f, 0.f),
    C_Vertex3f(0.221024632f, 0.573791265f, -1.f),
    C_Normal3f(0.45139423f, -0.892324567f, 0.f),
    C_Vertex3f(0.204462528f, 0.565999031f, 0.f),
    C_Vertex3f(0.204462528f, 0.565999031f, -1.f),
    C_Normal3f(0.505917668f, -0.862581789f, 0.f),
    C_Vertex3f(0.188016415f, 0.55709362f, 0.f),
    C_Vertex3f(0.188016415f, 0.55709362f, -1.f),
    C_Normal3f(0.56488502f, -0.825169563f, 0.f),
    C_Vertex3f(0.172465086f, 0.547232032f, 0.f),
    C_Vertex3f(0.172465086f, 0.547232032f, -1.f),
    C_Normal3f(0.622291386f, -0.782785594f, 0.f),
    C_Vertex3f(0.157808542f, 0.536414266f, 0.f),
    C_Vertex3f(0.157808542f, 0.536414266f, -1.f),
    C_Normal3f(0.677140772f, -0.735853434f, 0.f),
    C_Vertex3f(0.144046783f, 0.524640322f, 0.f),
    C_Vertex3f(0.144046783f, 0.524640322f, -1.f),
    C_Normal3f(0.728521347f, -0.685023189f, 0.f),
    C_Vertex3f(0.13117981f, 0.5119102f, 0.f),
    C_Vertex3f(0.13117981f, 0.5119102f, -1.f),
    C_Normal3f(0.775678813f, -0.631127894f, 0.f),
    C_Vertex3f(0.119207621f, 0.498223901f, 0.f),
    C_Vertex3f(0.119207621f, 0.498223901f, -1.f),
    C_Normal3f(0.818070829f, -0.575117528f, 0.f),
    C_Vertex3f(0.108130217f, 0.483581424f, 0.f),
    C_Vertex3f(0.108130217f, 0.483581424f, -1.f),
    C_Normal3f(0.855392396f, -0.517980456f, 0.f),
    C_Vertex3f(0.0979475975f, 0.467982769f, 0.f),
    C_Vertex3f(0.0979475975f, 0.467982769f, -1.f),
    C_Normal3f(0.886027932f, -0.463631988f, 0.f),
    C_Vertex3f(0.0886597633f, 0.451427937f, 0.f),
    C_Vertex3f(0.0886597633f, 0.451427937f, -1.f),
    C_Normal3f(0.912299216f, -0.409524322f, 0.f),
    C_Vertex3f(0.0805329084f, 0.434702277f, 0.f),
    C_Vertex3f(0.0805329084f, 0.434702277f, -1.f),
    C_Normal3f(0.936017632f, -0.351953179f, 0.f),
    C_Vertex3f(0.0735670328f, 0.417805791f, 0.f),
    C_Vertex3f(0.0735670328f, 0.417805791f, -1.f),
    C_Normal3f(0.956611574f, -0.291366369f, 0.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, 0.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, -1.f),
    C_Normal3f(0.973576069f, -0.228362828f, 0.f),
    C_Vertex3f(0.0631182194f, 0.383500338f, 0.f),
    C_Vertex3f(0.0631182194f, 0.383500338f, -1.f),
    C_Normal3f(0.986515164f, -0.163670227f, 0.f),
    C_Vertex3f(0.0596352816f, 0.366091371f, 0.f),
    C_Vertex3f(0.0596352816f, 0.366091371f, -1.f),
    C_Normal3f(0.995175898f, -0.0981062874f, 0.f),
    C_Vertex3f(0.057313323f, 0.348511577f, 0.f),
    C_Vertex3f(0.057313323f, 0.348511577f, -1.f),
    C_Normal3f(0.999469161f, -0.0325791314f, 0.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, 0.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, -1.f),
    C_Normal3f(0.999484181f, 0.0321161337f, 0.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, 0.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, -1.f),
    C_Normal3f(0.995274365f, 0.0971023366f, 0.f),
    C_Vertex3f(0.057292819f, 0.295268297f, 0.f),
    C_Vertex3f(0.057292819f, 0.295268297f, -1.f),
    C_Normal3f(0.986764789f, 0.162157908f, 0.f),
    C_Vertex3f(0.0595737696f, 0.277826071f, 0.f),
    C_Vertex3f(0.0595737696f, 0.277826071f, -1.f),
    C_Normal3f(0.974012792f, 0.226493135f, 0.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, 0.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, -1.f),
    C_Normal3f(0.957239568f, 0.289296657f, 0.f),
    C_Vertex3f(0.0675570965f, 0.243494511f, 0.f),
    C_Vertex3f(0.0675570965f, 0.243494511f, -1.f),
    C_Normal3f(0.93681109f, 0.349835664f, 0.f),
    C_Vertex3f(0.0732594728f, 0.226605177f, 0.f),
    C_Vertex3f(0.0732594728f, 0.226605177f, -1.f),
    C_Normal3f(0.913206995f, 0.407496005f, 0.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, 0.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, -1.f),
    C_Normal3f(0.886980236f, 0.461807281f, 0.f),
    C_Vertex3f(0.0880856514f, 0.193379402f, 0.f),
    C_Vertex3f(0.0880856514f, 0.193379402f, -1.f),
    C_Normal3f(0.856382728f, 0.516341746f, 0.f),
    C_Vertex3f(0.0972094536f, 0.177042961f, 0.f),
    C_Vertex3f(0.0972094536f, 0.177042961f, -1.f),
    C_Normal3f(0.819295764f, 0.573371112f, 0.f),
    C_Vertex3f(0.107241631f, 0.1616081f, 0.f),
    C_Vertex3f(0.107241631f, 0.1616081f, -1.f),
    C_Normal3f(0.777389884f, 0.629019022f, 0.f),
    C_Vertex3f(0.118182182f, 0.147074819f, 0.f),
    C_Vertex3f(0.118182182f, 0.147074819f, -1.f),
    C_Normal3f(0.73102957f, 0.682345867f, 0.f),
    C_Vertex3f(0.130031109f, 0.133443117f, 0.f),
    C_Vertex3f(0.130031109f, 0.133443117f, -1.f),
    C_Normal3f(0.68079406f, 0.732474983f, 0.f),
    C_Vertex3f(0.14278841f, 0.120712996f, 0.f),
    C_Vertex3f(0.14278841f, 0.120712996f, -1.f),
    C_Normal3f(0.62744385f, 0.778661847f, 0.f),
    C_Vertex3f(0.156454086f, 0.108884454f, 0.f),
    C_Vertex3f(0.156454086f, 0.108884454f, -1.f),
    C_Normal3f(0.571864069f, 0.820348322f, 0.f),
    C_Vertex3f(0.171028137f, 0.0979574919f, 0.f),
    C_Vertex3f(0.171028137f, 0.0979574919f, -1.f),
    C_Normal3f(0.514994919f, 0.857193172f, 0.f),
    C_Vertex3f(0.186510563f, 0.0879321098f, 0.f),
    C_Vertex3f(0.186510563f, 0.0879321098f, -1.f),
    C_Normal3f(0.460680038f, 0.887566268f, 0.f),
    C_Vertex3f(0.202901363f, 0.0788083076f, 0.f),
    C_Vertex3f(0.202901363f, 0.0788083076f, -1.f),
    C_Normal3f(0.406587005f, 0.913612008f, 0.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, 0.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, -1.f),
    C_Normal3f(0.349142969f, 0.937069476f, 0.f),
    C_Vertex3f(0.236216187f, 0.0639821291f, 0.f),
    C_Vertex3f(0.236216187f, 0.0639821291f, -1.f),
    C_Normal3f(0.288807094f, 0.957387328f, 0.f),
    C_Vertex3f(0.253140211f, 0.0582797527f, 0.f),
    C_Vertex3f(0.253140211f, 0.0582797527f, -1.f),
    C_Normal3f(0.226182401f, 0.974084914f, 0.f),
    C_Vertex3f(0.270241976f, 0.0537178516f, 0.f),
    C_Vertex3f(0.270241976f, 0.0537178516f, -1.f),
    C_Normal3f(0.161991552f, 0.986792147f, 0.f),
    C_Vertex3f(0.287521482f, 0.0502964258f, 0.f),
    C_Vertex3f(0.287521482f, 0.0502964258f, -1.f),
    C_Normal3f(0.0970379636f, 0.995280623f, 0.f),
    C_Vertex3f(0.304978728f, 0.0480154753f, 0.f),
    C_Vertex3f(0.304978728f, 0.0480154753f, -1.f),
    C_Normal3f(0.0321576744f, 0.99948281f, 0.f),
    C_Vertex3f(0.322613716f, 0.046875f, 0.f),
    C_Vertex3f(0.322613716f, 0.046875f, -1.f),
    C_Normal3f(-0.0321591832f, 0.999482751f, 0.f),
    C_Vertex3f(0.340425491f, 0.046875f, 0.f),
    C_Vertex3f(0.340425491f, 0.046875f, -1.f),
    C_Normal3f(-0.0970464498f, 0.995279849f, 0.f),
    C_Vertex3f(0.35805881f, 0.0480154753f, 0.f),
    C_Vertex3f(0.35805881f, 0.0480154753f, -1.f),
    C_Normal3f(-0.162003458f, 0.98679018f, 0.f),
    C_Vertex3f(0.375514627f, 0.0502964258f, 0.f),
    C_Vertex3f(0.375514627f, 0.0502964258f, -1.f),
    C_Normal3f(-0.226195812f, 0.974081814f, 0.f),
    C_Vertex3f(0.39279294f, 0.0537178516f, 0.f),
    C_Vertex3f(0.39279294f, 0.0537178516f, -1.f),
    C_Normal3f(-0.288820088f, 0.957383394f, 0.f),
    C_Vertex3f(0.409893751f, 0.0582797527f, 0.f),
    C_Vertex3f(0.409893751f, 0.0582797527f, -1.f),
    C_Normal3f(-0.349153817f, 0.937065423f, 0.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, 0.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, -1.f),
    C_Normal3f(-0.406594306f, 0.913608789f, 0.f),
    C_Vertex3f(0.443562865f, 0.0708249807f, 0.f),
    C_Vertex3f(0.443562865f, 0.0708249807f, -1.f),
    C_Normal3f(-0.46068269f, 0.887564957f, 0.f),
    C_Vertex3f(0.460131168f, 0.0788083076f, 0.f),
    C_Vertex3f(0.460131168f, 0.0788083076f, -1.f),
    C_Normal3f(-0.515075624f, 0.857144773f, 0.f),
    C_Vertex3f(0.476521969f, 0.0879321098f, 0.f),
    C_Vertex3f(0.476521969f, 0.0879321098f, -1.f),
    C_Normal3f(-0.572125196f, 0.82016629f, 0.f),
    C_Vertex3f(0.4919976f, 0.0979574919f, 0.f),
    C_Vertex3f(0.4919976f, 0.0979574919f, -1.f),
    C_Normal3f(-0.627901733f, 0.778292656f, 0.f),
    C_Vertex3f(0.506558061f, 0.108884454f, 0.f),
    C_Vertex3f(0.506558061f, 0.108884454f, -1.f),
    C_Normal3f(-0.681451917f, 0.731862903f, 0.f),
    C_Vertex3f(0.520203352f, 0.120712996f, 0.f),
    C_Vertex3f(0.520203352f, 0.120712996f, -1.f),
    C_Normal3f(-0.731875896f, 0.681438029f, 0.f),
    C_Vertex3f(0.532933474f, 0.133443117f, 0.f),
    C_Vertex3f(0.532933474f, 0.133443117f, -1.f),
    C_Normal3f(-0.77839905f, 0.627769887f, 0.f),
    C_Vertex3f(0.544748425f, 0.147074819f, 0.f),
    C_Vertex3f(0.544748425f, 0.147074819f, -1.f),
    C_Normal3f(-0.82043016f, 0.571746707f, 0.f),
    C_Vertex3f(0.555648208f, 0.1616081f, 0.f),
    C_Vertex3f(0.555648208f, 0.1616081f, -1.f),
    C_Normal3f(-0.857596815f, 0.514322579f, 0.f),
    C_Vertex3f(0.56563282f, 0.177042961f, 0.f),
    C_Vertex3f(0.56563282f, 0.177042961f, -1.f),
    C_Normal3f(-0.888106108f, 0.459638536f, 0.f),
    C_Vertex3f(0.574702263f, 0.193379402f, 0.f),
    C_Vertex3f(0.574702263f, 0.193379402f, -1.f),
    C_Normal3f(-0.914109051f, 0.405468255f, 0.f),
    C_Vertex3f(0.582638025f, 0.209900141f, 0.f),
    C_Vertex3f(0.582638025f, 0.209900141f, -1.f),
    C_Normal3f(-0.937492907f, 0.348004431f, 0.f),
    C_Vertex3f(0.589440107f, 0.226605177f, 0.f),
    C_Vertex3f(0.589440107f, 0.226605177f, -1.f),
    C_Normal3f(-0.95771575f, 0.287716091f, 0.f),
    C_Vertex3f(0.595108509f, 0.243494511f, 0.f),
    C_Vertex3f(0.595108509f, 0.243494511f, -1.f),
    C_Normal3f(-0.974309683f, 0.225212321f, 0.f),
    C_Vertex3f(0.59964323f, 0.260568142f, 0.f),
    C_Vertex3f(0.59964323f, 0.260568142f, -1.f),
    C_Normal3f(-0.986919045f, 0.161216959f, 0.f),
    C_Vertex3f(0.603044271f, 0.277826071f, 0.f),
    C_Vertex3f(0.603044271f, 0.277826071f, -1.f),
    C_Normal3f(-0.995330095f, 0.0965292007f, 0.f),
    C_Vertex3f(0.605311632f, 0.295268297f, 0.f),
    C_Vertex3f(0.605311632f, 0.295268297f, -1.f),
    C_Normal3f(-0.999490261f, 0.0319249816f, 0.f),
    C_Vertex3f(0.606445313f, 0.312894821f, 0.f),
    C_Vertex3f(0.606445313f, 0.312894821f, -1.f),
    C_Normal3f(-0.9994753f, -0.0323886536f, 0.f),
    C_Vertex3f(0.606445313f, 0.330760956f, 0.f),
    C_Vertex3f(0.606445313f, 0.330760956f, -1.f),
    C_Normal3f(-0.995231748f, -0.0975375697f, 0.f),
    C_Vertex3f(0.605291128f, 0.348511577f, 0.f),
    C_Vertex3f(0.605291128f, 0.348511577f, -1.f),
    C_Normal3f(-0.986669362f, -0.162737727f, 0.f),
    C_Vertex3f(0.602982759f, 0.366091371f, 0.f),
    C_Vertex3f(0.602982759f, 0.366091371f, -1.f),
    C_Normal3f(-0.973872483f, -0.22709541f, 0.f),
    C_Vertex3f(0.599520206f, 0.383500338f, 0.f),
    C_Vertex3f(0.599520206f, 0.383500338f, -1.f),
    C_Normal3f(-0.957085848f, -0.289804697f, 0.f),
    C_Vertex3f(0.594903469f, 0.400738478f, 0.f),
    C_Vertex3f(0.594903469f, 0.400738478f, -1.f),
    C_Normal3f(-0.93669498f, -0.350146502f, 0.f),
    C_Vertex3f(0.589132547f, 0.417805791f, 0.f),
    C_Vertex3f(0.589132547f, 0.417805791f, -1.f),
    C_Normal3f(-0.913193345f, -0.407526493f, 0.f),
    C_Vertex3f(0.582207441f, 0.434702277f, 0.f),
    C_Vertex3f(0.582207441f, 0.434702277f, -1.f),
    C_Normal3f(-0.887141407f, -0.461497724f, 0.f),
    C_Vertex3f(0.574128151f, 0.451427937f, 0.f),
    C_Vertex3f(0.574128151f, 0.451427937f, -1.f),
    C_Normal3f(-0.856513202f, -0.516125083f, 0.f),
    C_Vertex3f(0.564894676f, 0.467982769f, 0.f),
    C_Vertex3f(0.564894676f, 0.467982769f, -1.f),
    C_Normal3f(-0.818935633f, -0.573885381f, 0.f),
    C_Vertex3f(0.554752827f, 0.483581424f, 0.f),
    C_Vertex3f(0.554752827f, 0.483581424f, -1.f),
    C_Normal3f(-0.776225388f, -0.630455494f, 0.f),
    C_Vertex3f(0.543702602f, 0.498223901f, 0.f),
    C_Vertex3f(0.543702602f, 0.498223901f, -1.f),
    C_Normal3f(-0.728708327f, -0.684824109f, 0.f),
    C_Vertex3f(0.531744003f, 0.5119102f, 0.f),
    C_Vertex3f(0.531744003f, 0.5119102f, -1.f),
    C_Normal3f(-0.676953673f, -0.736025572f, 0.f),
    C_Vertex3f(0.518877029f, 0.524640322f, 0.f),
    C_Vertex3f(0.518877029f, 0.524640322f, -1.f),
    C_Normal3f(-0.621744752f, -0.783219934f, 0.f),
    C_Vertex3f(0.505101681f, 0.536414266f, 0.f),
    C_Vertex3f(0.505101681f, 0.536414266f, -1.f),
    C_Normal3f(-0.564020872f, -0.825760543f, 0.f),
    C_Vertex3f(0.490417957f, 0.547232032f, 0.f),
    C_Vertex3f(0.490417957f, 0.547232032f, -1.f),
    C_Normal3f(-0.504800618f, -0.86323607f, 0.f),
    C_Vertex3f(0.474825859f, 0.55709362f, 0.f),
    C_Vertex3f(0.474825859f, 0.55709362f, -1.f),
    C_Normal3f(-0.450436771f, -0.892808318f, 0.f),
    C_Vertex3f(0.458325386f, 0.565999031f, 0.f),
    C_Vertex3f(0.458325386f, 0.565999031f, -1.f),
    C_Normal3f(-0.398695141f, -0.917083561f, 0.f),
    C_Vertex3f(0.441729546f, 0.573791265f, 0.f),
    C_Vertex3f(0.441729546f, 0.573791265f, -1.f),
    C_Normal3f(-0.343506575f, -0.939150333f, 0.f),
    C_Vertex3f(0.425038338f, 0.580470324f, 0.f),
    C_Vertex3f(0.425038338f, 0.580470324f, -1.f),
    C_Normal3f(-0.285206378f, -0.958466113f, 0.f),
    C_Vertex3f(0.408251762f, 0.586036205f, 0.f),
    C_Vertex3f(0.408251762f, 0.586036205f, -1.f),
    C_Normal3f(-0.224273607f, -0.974526227f, 0.f),
    C_Vertex3f(0.39136982f, 0.590488911f, 0.f),
    C_Vertex3f(0.39136982f, 0.590488911f, -1.f),
    C_Normal3f(-0.161321044f, -0.986901939f, 0.f),
    C_Vertex3f(0.374392509f, 0.59382844f, 0.f),
    C_Vertex3f(0.374392509f, 0.59382844f, -1.f),
    C_Normal3f(-0.0970703289f, -0.995277524f, 0.f),
    C_Vertex3f(0.357319832f, 0.596054792f, 0.f),
    C_Vertex3f(0.357319832f, 0.596054792f, -1.f),
    C_Normal3f(-0.0322615579f, -0.999479473f, 0.f),
    C_Vertex3f(0.340151787f, 0.597167969f, 0.f),
    C_Vertex3f(0.340151787f, 0.597167969f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.279329538f, 0.960195303f, 0.f),
    C_Vertex3f(0.4375f, 0.268066406f, 0.f),
    C_Vertex3f(0.4375f, 0.268066406f, -1.f),
    C_Vertex3f(0.491210938f, 0.252441406f, -1.f),
    C_Vertex3f(0.491210938f, 0.252441406f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.972285986f, -0.233794481f, 0.f),
    C_Vertex3f(0.491210938f, 0.252441406f, 0.f),
    C_Vertex3f(0.491210938f, 0.252441406f, -1.f),
    C_Normal3f(0.959003329f, -0.283394843f, 0.f),
    C_Vertex3f(0.489570618f, 0.245619774f, 0.f),
    C_Vertex3f(0.489570618f, 0.245619774f, -1.f),
    C_Normal3f(0.937873065f, -0.346978337f, 0.f),
    C_Vertex3f(0.48536694f, 0.23266542f, 0.f),
    C_Vertex3f(0.48536694f, 0.23266542f, -1.f),
    C_Normal3f(0.905451f, -0.424450666f, 0.f),
    C_Vertex3f(0.480240226f, 0.220399976f, 0.f),
    C_Vertex3f(0.480240226f, 0.220399976f, -1.f),
    C_Normal3f(0.865142584f, -0.501525879f, 0.f),
    C_Vertex3f(0.474190474f, 0.208823442f, 0.f),
    C_Vertex3f(0.474190474f, 0.208823442f, -1.f),
    C_Normal3f(0.817257226f, -0.576272964f, 0.f),
    C_Vertex3f(0.467217684f, 0.19793582f, 0.f),
    C_Vertex3f(0.467217684f, 0.19793582f, -1.f),
    C_Normal3f(0.762655854f, -0.646804571f, 0.f),
    C_Vertex3f(0.459321856f, 0.187737107f, 0.f),
    C_Vertex3f(0.459321856f, 0.187737107f, -1.f),
    C_Normal3f(0.70268321f, -0.71150279f, 0.f),
    C_Vertex3f(0.450502992f, 0.178227305f, 0.f),
    C_Vertex3f(0.450502992f, 0.178227305f, -1.f),
    C_Normal3f(0.639008462f, -0.769199729f, 0.f),
    C_Vertex3f(0.440761089f, 0.169406414f, 0.f),
    C_Vertex3f(0.440761089f, 0.169406414f, -1.f),
    C_Normal3f(0.571433961f, -0.820648015f, 0.f),
    C_Vertex3f(0.430096149f, 0.161274433f, 0.f),
    C_Vertex3f(0.430096149f, 0.161274433f, -1.f),
    C_Normal3f(0.497654825f, -0.867375135f, 0.f),
    C_Vertex3f(0.418863893f, 0.15415895f, 0.f),
    C_Vertex3f(0.418863893f, 0.15415895f, -1.f),
    C_Normal3f(0.419916183f, -0.907562852f, 0.f),
    C_Vertex3f(0.407064319f, 0.148059964f, 0.f),
    C_Vertex3f(0.407064319f, 0.148059964f, -1.f),
    C_Normal3f(0.340038031f, -0.940411687f, 0.f),
    C_Vertex3f(0.394697428f, 0.142977476f, 0.f),
    C_Vertex3f(0.394697428f, 0.142977476f, -1.f),
    C_Normal3f(0.25991118f, -0.965632558f, 0.f),
    C_Vertex3f(0.38176322f, 0.138911486f, 0.f),
    C_Vertex3f(0.38176322f, 0.138911486f, -1.f),
    C_Normal3f(0.181291327f, -0.983429492f, 0.f),
    C_Vertex3f(0.368261695f, 0.135861993f, 0.f),
    C_Vertex3f(0.368261695f, 0.135861993f, -1.f),
    C_Normal3f(0.105641134f, -0.994404256f, 0.f),
    C_Vertex3f(0.354192853f, 0.133828998f, 0.f),
    C_Vertex3f(0.354192853f, 0.133828998f, -1.f),
    C_Normal3f(0.0318846256f, -0.999491572f, 0.f),
    C_Vertex3f(0.339556694f, 0.1328125f, 0.f),
    C_Vertex3f(0.339556694f, 0.1328125f, -1.f),
    C_Normal3f(-0.0437674113f, -0.999041855f, 0.f),
    C_Vertex3f(0.322328568f, 0.1328125f, 0.f),
    C_Vertex3f(0.322328568f, 0.1328125f, -1.f),
    C_Normal3f(-0.129522562f, -0.991576493f, 0.f),
    C_Vertex3f(0.303978443f, 0.134371161f, 0.f),
    C_Vertex3f(0.303978443f, 0.134371161f, -1.f),
    C_Normal3f(-0.223468989f, -0.974711061f, 0.f),
    C_Vertex3f(0.286530972f, 0.137488484f, 0.f),
    C_Vertex3f(0.286530972f, 0.137488484f, -1.f),
    C_Normal3f(-0.321033686f, -0.947067797f, 0.f),
    C_Vertex3f(0.269986153f, 0.142164469f, 0.f),
    C_Vertex3f(0.269986153f, 0.142164469f, -1.f),
    C_Normal3f(-0.419197083f, -0.907895207f, 0.f),
    C_Vertex3f(0.254343987f, 0.148399115f, 0.f),
    C_Vertex3f(0.254343987f, 0.148399115f, -1.f),
    C_Normal3f(-0.51448369f, -0.857500196f, 0.f),
    C_Vertex3f(0.239604473f, 0.156192422f, 0.f),
    C_Vertex3f(0.239604473f, 0.156192422f, -1.f),
    C_Normal3f(-0.60350728f, -0.7973575f, 0.f),
    C_Vertex3f(0.225767612f, 0.165544391f, 0.f),
    C_Vertex3f(0.225767612f, 0.165544391f, -1.f),
    C_Normal3f(-0.683544576f, -0.729908884f, 0.f),
    C_Vertex3f(0.212833405f, 0.176455021f, 0.f),
    C_Vertex3f(0.212833405f, 0.176455021f, -1.f),
    C_Normal3f(-0.757078111f, -0.653324425f, 0.f),
    C_Vertex3f(0.200801849f, 0.188924313f, 0.f),
    C_Vertex3f(0.200801849f, 0.188924313f, -1.f),
    C_Normal3f(-0.82505995f, -0.565045118f, 0.f),
    C_Vertex3f(0.190274239f, 0.202596784f, 0.f),
    C_Vertex3f(0.190274239f, 0.202596784f, -1.f),
    C_Normal3f(-0.881946504f, -0.471349657f, 0.f),
    C_Vertex3f(0.181250572f, 0.217472434f, 0.f),
    C_Vertex3f(0.181250572f, 0.217472434f, -1.f),
    C_Normal3f(-0.92663449f, -0.375963598f, 0.f),
    C_Vertex3f(0.17373085f, 0.233551264f, 0.f),
    C_Vertex3f(0.17373085f, 0.233551264f, -1.f),
    C_Normal3f(-0.959308445f, -0.282359987f, 0.f),
    C_Vertex3f(0.167715073f, 0.250833273f, 0.f),
    C_Vertex3f(0.167715073f, 0.250833273f, -1.f),
    C_Normal3f(-0.981144965f, -0.193273515f, 0.f),
    C_Vertex3f(0.163203239f, 0.269318461f, 0.f),
    C_Vertex3f(0.163203239f, 0.269318461f, -1.f),
    C_Normal3f(-0.993875802f, -0.110502988f, 0.f),
    C_Vertex3f(0.160195351f, 0.289006829f, 0.f),
    C_Vertex3f(0.160195351f, 0.289006829f, -1.f),
    C_Normal3f(-0.999474525f, -0.0324125849f, 0.f),
    C_Vertex3f(0.158691406f, 0.309898376f, 0.f),
    C_Vertex3f(0.158691406f, 0.309898376f, -1.f),
    C_Normal3f(-0.998642802f, 0.0520825051f, 0.f),
    C_Vertex3f(0.158691406f, 0.335382462f, 0.f),
    C_Vertex3f(0.158691406f, 0.335382462f, -1.f),
    C_Normal3f(-0.987806499f, 0.155686751f, 0.f),
    C_Vertex3f(0.161427975f, 0.362370014f, 0.f),
    C_Vertex3f(0.161427975f, 0.362370014f, -1.f),
    C_Normal3f(-0.962163746f, 0.272471726f, 0.f),
    C_Vertex3f(0.166901112f, 0.387471676f, 0.f),
    C_Vertex3f(0.166901112f, 0.387471676f, -1.f),
    C_Normal3f(-0.918659031f, 0.395051211f, 0.f),
    C_Vertex3f(0.175110817f, 0.410687447f, 0.f),
    C_Vertex3f(0.175110817f, 0.410687447f, -1.f),
    C_Normal3f(-0.855625093f, 0.517596185f, 0.f),
    C_Vertex3f(0.186057091f, 0.432017326f, 0.f),
    C_Vertex3f(0.186057091f, 0.432017326f, -1.f),
    C_Normal3f(-0.769492567f, 0.638655841f, 0.f),
    C_Vertex3f(0.199328899f, 0.450721741f, 0.f),
    C_Vertex3f(0.199328899f, 0.450721741f, -1.f),
    C_Normal3f(-0.661064744f, 0.750328898f, 0.f),
    C_Vertex3f(0.214926243f, 0.46680069f, 0.f),
    C_Vertex3f(0.214926243f, 0.46680069f, -1.f),
    C_Normal3f(-0.536728859f, 0.843754828f, 0.f),
    C_Vertex3f(0.232849121f, 0.480254173f, 0.f),
    C_Vertex3f(0.232849121f, 0.480254173f, -1.f),
    C_Normal3f(-0.411678016f, 0.911329329f, 0.f),
    C_Vertex3f(0.253097534f, 0.491082191f, 0.f),
    C_Vertex3f(0.253097534f, 0.491082191f, -1.f),
    C_Normal3f(-0.2889238f, 0.957352042f, 0.f),
    C_Vertex3f(0.274796486f, 0.499203205f, 0.f),
    C_Vertex3f(0.274796486f, 0.499203205f, -1.f),
    C_Normal3f(-0.167667642f, 0.985843599f, 0.f),
    C_Vertex3f(0.297945976f, 0.504617214f, 0.f),
    C_Vertex3f(0.297945976f, 0.504617214f, -1.f),
    C_Normal3f(-0.0601753481f, 0.99818784f, 0.f),
    C_Vertex3f(0.322546005f, 0.507324219f, 0.f),
    C_Vertex3f(0.322546005f, 0.507324219f, -1.f),
    C_Normal3f(0.0254070852f, 0.999677122f, 0.f),
    C_Vertex3f(0.342849731f, 0.507324219f, 0.f),
    C_Vertex3f(0.342849731f, 0.507324219f, -1.f),
    C_Normal3f(0.0953223109f, 0.995446444f, 0.f),
    C_Vertex3f(0.356791019f, 0.506453872f, 0.f),
    C_Vertex3f(0.356791019f, 0.506453872f, -1.f),
    C_Normal3f(0.164857402f, 0.986317456f, 0.f),
    C_Vertex3f(0.370116711f, 0.504713178f, 0.f),
    C_Vertex3f(0.370116711f, 0.504713178f, -1.f),
    C_Normal3f(0.238527447f, 0.971135795f, 0.f),
    C_Vertex3f(0.382826805f, 0.502102137f, 0.f),
    C_Vertex3f(0.382826805f, 0.502102137f, -1.f),
    C_Normal3f(0.315333575f, 0.948980927f, 0.f),
    C_Vertex3f(0.394921303f, 0.498620749f, 0.f),
    C_Vertex3f(0.394921303f, 0.498620749f, -1.f),
    C_Normal3f(0.393870145f, 0.919166028f, 0.f),
    C_Vertex3f(0.406400204f, 0.494269013f, 0.f),
    C_Vertex3f(0.406400204f, 0.494269013f, -1.f),
    C_Normal3f(0.472384959f, 0.88139236f, 0.f),
    C_Vertex3f(0.417263508f, 0.489046931f, 0.f),
    C_Vertex3f(0.417263508f, 0.489046931f, -1.f),
    C_Normal3f(0.548923135f, 0.835872829f, 0.f),
    C_Vertex3f(0.427511215f, 0.482954502f, 0.f),
    C_Vertex3f(0.427511215f, 0.482954502f, -1.f),
    C_Normal3f(0.618222773f, 0.786002874f, 0.f),
    C_Vertex3f(0.437143326f, 0.475991726f, 0.f),
    C_Vertex3f(0.437143326f, 0.475991726f, -1.f),
    C_Normal3f(0.681456804f, 0.731858313f, 0.f),
    C_Vertex3f(0.446009159f, 0.468405128f, 0.f),
    C_Vertex3f(0.446009159f, 0.468405128f, -1.f),
    C_Normal3f(0.741287529f, 0.671187639f, 0.f),
    C_Vertex3f(0.454108715f, 0.460194707f, 0.f),
    C_Vertex3f(0.454108715f, 0.460194707f, -1.f),
    C_Normal3f(0.796201468f, 0.60503155f, 0.f),
    C_Vertex3f(0.461441994f, 0.451360464f, 0.f),
    C_Vertex3f(0.461441994f, 0.451360464f, -1.f),
    C_Normal3f(0.844967127f, 0.534818172f, 0.f),
    C_Vertex3f(0.468008995f, 0.441902399f, 0.f),
    C_Vertex3f(0.468008995f, 0.441902399f, -1.f),
    C_Normal3f(0.886769056f, 0.462212831f, 0.f),
    C_Vertex3f(0.473809719f, 0.431820512f, 0.f),
    C_Vertex3f(0.473809719f, 0.431820512f, -1.f),
    C_Normal3f(0.921266913f, 0.388931185f, 0.f),
    C_Vertex3f(0.478844166f, 0.421114802f, 0.f),
    C_Vertex3f(0.478844166f, 0.421114802f, -1.f),
    C_Normal3f(0.944504201f, 0.328499436f, 0.f),
    C_Vertex3f(0.483112335f, 0.409785271f, 0.f),
    C_Vertex3f(0.483112335f, 0.409785271f, -1.f),
    C_Normal3f(0.959664822f, 0.281146437f, 0.f),
    C_Vertex3f(0.484863281f, 0.403808594f, 0.f),
    C_Vertex3f(0.484863281f, 0.403808594f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.236119837f, -0.971723974f, 0.f),
    C_Vertex3f(0.484863281f, 0.403808594f, 0.f),
    C_Vertex3f(0.484863281f, 0.403808594f, -1.f),
    C_Vertex3f(0.432617188f, 0.391113281f, -1.f),
    C_Vertex3f(0.432617188f, 0.391113281f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.92763567f, -0.37348628f, 0.f),
    C_Vertex3f(0.432617188f, 0.391113281f, 0.f),
    C_Vertex3f(0.432617188f, 0.391113281f, -1.f),
    C_Normal3f(-0.892387152f, -0.451270819f, 0.f),
    C_Vertex3f(0.429336548f, 0.399261475f, 0.f),
    C_Vertex3f(0.429336548f, 0.399261475f, -1.f),
    C_Normal3f(-0.833908617f, -0.551902592f, 0.f),
    C_Vertex3f(0.421216488f, 0.413658142f, 0.f),
    C_Vertex3f(0.421216488f, 0.413658142f, -1.f),
    C_Normal3f(-0.741188705f, -0.671296716f, 0.f),
    C_Vertex3f(0.411537647f, 0.42615509f, 0.f),
    C_Vertex3f(0.411537647f, 0.42615509f, -1.f),
    C_Normal3f(-0.626029909f, -0.779798985f, 0.f),
    C_Vertex3f(0.400300026f, 0.436752319f, 0.f),
    C_Vertex3f(0.400300026f, 0.436752319f, -1.f),
    C_Normal3f(-0.495049536f, -0.868864715f, 0.f),
    C_Vertex3f(0.387503624f, 0.445449829f, 0.f),
    C_Vertex3f(0.387503624f, 0.445449829f, -1.f),
    C_Normal3f(-0.351503551f, -0.936186552f, 0.f),
    C_Vertex3f(0.373586178f, 0.451972961f, 0.f),
    C_Vertex3f(0.373586178f, 0.451972961f, -1.f),
    C_Normal3f(-0.204662144f, -0.978832603f, 0.f),
    C_Vertex3f(0.358547688f, 0.456321716f, 0.f),
    C_Vertex3f(0.358547688f, 0.456321716f, -1.f),
    C_Normal3f(-0.0697714165f, -0.997562945f, 0.f),
    C_Vertex3f(0.342388153f, 0.458496094f, 0.f),
    C_Vertex3f(0.342388153f, 0.458496094f, -1.f),
    C_Normal3f(0.0415755846f, -0.999135315f, 0.f),
    C_Vertex3f(0.327459335f, 0.458496094f, 0.f),
    C_Vertex3f(0.327459335f, 0.458496094f, -1.f),
    C_Normal3f(0.141769439f, -0.989899695f, 0.f),
    C_Vertex3f(0.315436006f, 0.457374573f, 0.f),
    C_Vertex3f(0.315436006f, 0.457374573f, -1.f),
    C_Normal3f(0.242993817f, -0.970027804f, 0.f),
    C_Vertex3f(0.303966403f, 0.455131531f, 0.f),
    C_Vertex3f(0.303966403f, 0.455131531f, -1.f),
    C_Normal3f(0.346146911f, -0.938180268f, 0.f),
    C_Vertex3f(0.293050528f, 0.451766968f, 0.f),
    C_Vertex3f(0.293050528f, 0.451766968f, -1.f),
    C_Normal3f(0.447511673f, -0.894278109f, 0.f),
    C_Vertex3f(0.282688379f, 0.447280884f, 0.f),
    C_Vertex3f(0.282688379f, 0.447280884f, -1.f),
    C_Normal3f(0.543306351f, -0.839534521f, 0.f),
    C_Vertex3f(0.272879958f, 0.441673279f, 0.f),
    C_Vertex3f(0.272879958f, 0.441673279f, -1.f),
    C_Normal3f(0.630352974f, -0.776308715f, 0.f),
    C_Vertex3f(0.263625264f, 0.434944153f, 0.f),
    C_Vertex3f(0.263625264f, 0.434944153f, -1.f),
    C_Normal3f(0.706573367f, -0.707639813f, 0.f),
    C_Vertex3f(0.254924297f, 0.427093506f, 0.f),
    C_Vertex3f(0.254924297f, 0.427093506f, -1.f),
    C_Normal3f(0.777976573f, -0.628293276f, 0.f),
    C_Vertex3f(0.246777058f, 0.418121338f, 0.f),
    C_Vertex3f(0.246777058f, 0.418121338f, -1.f),
    C_Normal3f(0.844279528f, -0.535903215f, 0.f),
    C_Vertex3f(0.239648223f, 0.408178091f, 0.f),
    C_Vertex3f(0.239648223f, 0.408178091f, -1.f),
    C_Normal3f(0.897514045f, -0.440985918f, 0.f),
    C_Vertex3f(0.233537793f, 0.397263765f, 0.f),
    C_Vertex3f(0.233537793f, 0.397263765f, -1.f),
    C_Normal3f(0.937724292f, -0.34738043f, 0.f),
    C_Vertex3f(0.228445768f, 0.385378361f, 0.f),
    C_Vertex3f(0.228445768f, 0.385378361f, -1.f),
    C_Normal3f(0.966116905f, -0.258105069f, 0.f),
    C_Vertex3f(0.224372149f, 0.372521877f, 0.f),
    C_Vertex3f(0.224372149f, 0.372521877f, -1.f),
    C_Normal3f(0.984545231f, -0.175130755f, 0.f),
    C_Vertex3f(0.221316934f, 0.358694315f, 0.f),
    C_Vertex3f(0.221316934f, 0.358694315f, -1.f),
    C_Normal3f(0.995042324f, -0.0994514599f, 0.f),
    C_Vertex3f(0.219280124f, 0.343895674f, 0.f),
    C_Vertex3f(0.219280124f, 0.343895674f, -1.f),
    C_Normal3f(0.999509692f, -0.0313100405f, 0.f),
    C_Vertex3f(0.218261719f, 0.328125954f, 0.f),
    C_Vertex3f(0.218261719f, 0.328125954f, -1.f),
    C_Normal3f(0.999559283f, 0.0296840295f, 0.f),
    C_Vertex3f(0.218261719f, 0.311385155f, 0.f),
    C_Vertex3f(0.218261719f, 0.311385155f, -1.f),
    C_Normal3f(0.995512307f, 0.094632566f, 0.f),
    C_Vertex3f(0.219225764f, 0.295663357f, 0.f),
    C_Vertex3f(0.219225764f, 0.295663357f, -1.f),
    C_Normal3f(0.985887349f, 0.167410076f, 0.f),
    C_Vertex3f(0.221153855f, 0.28096056f, 0.f),
    C_Vertex3f(0.221153855f, 0.28096056f, -1.f),
    C_Normal3f(0.968732238f, 0.24810867f, 0.f),
    C_Vertex3f(0.224045992f, 0.267276764f, 0.f),
    C_Vertex3f(0.224045992f, 0.267276764f, -1.f),
    C_Normal3f(0.94181478f, 0.336132169f, 0.f),
    C_Vertex3f(0.227902174f, 0.254611969f, 0.f),
    C_Vertex3f(0.227902174f, 0.254611969f, -1.f),
    C_Normal3f(0.902883947f, 0.429884225f, 0.f),
    C_Vertex3f(0.232722402f, 0.242966175f, 0.f),
    C_Vertex3f(0.232722402f, 0.242966175f, -1.f),
    C_Normal3f(0.850144386f, 0.526549637f, 0.f),
    C_Vertex3f(0.238506675f, 0.232339382f, 0.f),
    C_Vertex3f(0.238506675f, 0.232339382f, -1.f),
    C_Normal3f(0.78289175f, 0.622157991f, 0.f),
    C_Vertex3f(0.245254993f, 0.22273159f, 0.f),
    C_Vertex3f(0.245254993f, 0.22273159f, -1.f),
    C_Normal3f(0.709563315f, 0.7046417f, 0.f),
    C_Vertex3f(0.252967358f, 0.214142799f, 0.f),
    C_Vertex3f(0.252967358f, 0.214142799f, -1.f),
    C_Normal3f(0.631703734f, 0.775209963f, 0.f),
    C_Vertex3f(0.261247277f, 0.206627607f, 0.f),
    C_Vertex3f(0.261247277f, 0.206627607f, -1.f),
    C_Normal3f(0.543013752f, 0.839723766f, 0.f),
    C_Vertex3f(0.270094752f, 0.200186014f, 0.f),
    C_Vertex3f(0.270094752f, 0.200186014f, -1.f),
    C_Normal3f(0.445868701f, 0.895098448f, 0.f),
    C_Vertex3f(0.279509783f, 0.19481802f, 0.f),
    C_Vertex3f(0.279509783f, 0.19481802f, -1.f),
    C_Normal3f(0.343711346f, 0.939075351f, 0.f),
    C_Vertex3f(0.289492369f, 0.190523624f, 0.f),
    C_Vertex3f(0.289492369f, 0.190523624f, -1.f),
    C_Normal3f(0.240470603f, 0.970656395f, 0.f),
    C_Vertex3f(0.30004251f, 0.187302828f, 0.f),
    C_Vertex3f(0.30004251f, 0.187302828f, -1.f),
    C_Normal3f(0.139856562f, 0.99017185f, 0.f),
    C_Vertex3f(0.311160207f, 0.18515563f, 0.f),
    C_Vertex3f(0.311160207f, 0.18515563f, -1.f),
    C_Normal3f(0.0387859084f, 0.999247491f, 0.f),
    C_Vertex3f(0.322845459f, 0.184082031f, 0.f),
    C_Vertex3f(0.322845459f, 0.184082031f, -1.f),
    C_Normal3f(-0.0817969814f, 0.996649027f, 0.f),
    C_Vertex3f(0.338819504f, 0.184082031f, 0.f),
    C_Vertex3f(0.338819504f, 0.184082031f, -1.f),
    C_Normal3f(-0.230782405f, 0.973005354f, 0.f),
    C_Vertex3f(0.357310295f, 0.186910629f, 0.f),
    C_Vertex3f(0.357310295f, 0.186910629f, -1.f),
    C_Normal3f(-0.390241712f, 0.920712531f, 0.f),
    C_Vertex3f(0.374596596f, 0.192567825f, 0.f),
    C_Vertex3f(0.374596596f, 0.192567825f, -1.f),
    C_Normal3f(-0.538789988f, 0.842440069f, 0.f),
    C_Vertex3f(0.390678406f, 0.201053619f, 0.f),
    C_Vertex3f(0.390678406f, 0.201053619f, -1.f),
    C_Normal3f(-0.67085892f, 0.741584957f, 0.f),
    C_Vertex3f(0.405555725f, 0.212368011f, 0.f),
    C_Vertex3f(0.405555725f, 0.212368011f, -1.f),
    C_Normal3f(-0.790529013f, 0.612424612f, 0.f),
    C_Vertex3f(0.417917252f, 0.225694656f, 0.f),
    C_Vertex3f(0.417917252f, 0.225694656f, -1.f),
    C_Normal3f(-0.885248482f, 0.465118319f, 0.f),
    C_Vertex3f(0.427762985f, 0.241033554f, 0.f),
    C_Vertex3f(0.427762985f, 0.241033554f, -1.f),
    C_Normal3f(-0.940829873f, 0.338879317f, 0.f),
    C_Vertex3f(0.435092926f, 0.258384705f, 0.f),
    C_Vertex3f(0.435092926f, 0.258384705f, -1.f),
    C_Normal3f(-0.970456541f, 0.241275832f, 0.f),
    C_Vertex3f(0.4375f, 0.268066406f, 0.f),
    C_Vertex3f(0.4375f, 0.268066406f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.661132813f, 0.311144829f, 0.f),
    C_Vertex3f(0.661132813f, 0.332565308f, 0.f),
    C_Vertex3f(0.659773588f, 0.290038824f, 0.f),
    C_Vertex3f(0.659753084f, 0.353883743f, 0.f),
    C_Vertex3f(0.65705514f, 0.269137621f, 0.f),
    C_Vertex3f(0.656993628f, 0.374990463f, 0.f),
    C_Vertex3f(0.652977467f, 0.248441219f, 0.f),
    C_Vertex3f(0.652854443f, 0.395885468f, 0.f),
    C_Vertex3f(0.647540569f, 0.227949619f, 0.f),
    C_Vertex3f(0.647335529f, 0.416568756f, 0.f),
    C_Vertex3f(0.640744448f, 0.207662821f, 0.f),
    C_Vertex3f(0.640436888f, 0.437040329f, 0.f),
    C_Vertex3f(0.632589102f, 0.187580824f, 0.f),
    C_Vertex3f(0.632158518f, 0.457300186f, 0.f),
    C_Vertex3f(0.623074532f, 0.167703629f, 0.f),
    C_Vertex3f(0.62250042f, 0.477348328f, 0.f),
    C_Vertex3f(0.612200737f, 0.148031235f, 0.f),
    C_Vertex3f(0.611462593f, 0.497184753f, 0.f),
    C_Vertex3f(0.606445313f, 0.330760956f, 0.f),
    C_Vertex3f(0.605291128f, 0.348511577f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.589132547f, 0.417805791f, 0.f),
    C_Vertex3f(0.594903469f, 0.400738478f, 0.f),
    C_Vertex3f(0.599331737f, 0.515873551f, 0.f),
    C_Vertex3f(0.599520206f, 0.383500338f, 0.f),
    C_Vertex3f(0.611462593f, 0.497184753f, 0.f),
    C_Vertex3f(0.602982759f, 0.366091371f, 0.f),
    C_Vertex3f(0.605291128f, 0.348511577f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.58610785f, 0.533414721f, 0.f),
    C_Vertex3f(0.571790934f, 0.549808264f, 0.f),
    C_Vertex3f(0.574128151f, 0.451427937f, 0.f),
    C_Vertex3f(0.582207441f, 0.434702277f, 0.f),
    C_Vertex3f(0.589132547f, 0.417805791f, 0.f),
    C_Vertex3f(0.599331737f, 0.515873551f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.543702602f, 0.498223901f, 0.f),
    C_Vertex3f(0.554752827f, 0.483581424f, 0.f),
    C_Vertex3f(0.556380987f, 0.565054178f, 0.f),
    C_Vertex3f(0.564894676f, 0.467982769f, 0.f),
    C_Vertex3f(0.571790934f, 0.549808264f, 0.f),
    C_Vertex3f(0.574128151f, 0.451427937f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.505101681f, 0.536414266f, 0.f),
    C_Vertex3f(0.518877029f, 0.524640322f, 0.f),
    C_Vertex3f(0.522282004f, 0.592103124f, 0.f),
    C_Vertex3f(0.531744003f, 0.5119102f, 0.f),
    C_Vertex3f(0.539878011f, 0.579152465f, 0.f),
    C_Vertex3f(0.543702602f, 0.498223901f, 0.f),
    C_Vertex3f(0.556380987f, 0.565054178f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.425038338f, 0.580470324f, 0.f),
    C_Vertex3f(0.441729546f, 0.573791265f, 0.f),
    C_Vertex3f(0.44387877f, 0.631876588f, 0.f),
    C_Vertex3f(0.458325386f, 0.565999031f, 0.f),
    C_Vertex3f(0.463906169f, 0.623885036f, 0.f),
    C_Vertex3f(0.474825859f, 0.55709362f, 0.f),
    C_Vertex3f(0.483810902f, 0.614561558f, 0.f),
    C_Vertex3f(0.490417957f, 0.547232032f, 0.f),
    C_Vertex3f(0.503592968f, 0.603906155f, 0.f),
    C_Vertex3f(0.505101681f, 0.536414266f, 0.f),
    C_Vertex3f(0.522282004f, 0.592103124f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.322833061f, 0.597167969f, 0.f),
    C_Vertex3f(0.340151787f, 0.597167969f, 0.f),
    C_Vertex3f(0.341901779f, 0.651855469f, 0.f),
    C_Vertex3f(0.357319832f, 0.596054792f, 0.f),
    C_Vertex3f(0.36254251f, 0.650523543f, 0.f),
    C_Vertex3f(0.374392509f, 0.59382844f, 0.f),
    C_Vertex3f(0.383060575f, 0.647859693f, 0.f),
    C_Vertex3f(0.39136982f, 0.590488911f, 0.f),
    C_Vertex3f(0.403455973f, 0.643863916f, 0.f),
    C_Vertex3f(0.408251762f, 0.586036205f, 0.f),
    C_Vertex3f(0.423728704f, 0.638536215f, 0.f),
    C_Vertex3f(0.425038338f, 0.580470324f, 0.f),
    C_Vertex3f(0.44387877f, 0.631876588f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.221024632f, 0.573791265f, 0.f),
    C_Vertex3f(0.237702727f, 0.580470324f, 0.f),
    C_Vertex3f(0.23930645f, 0.638536215f, 0.f),
    C_Vertex3f(0.254496813f, 0.586036205f, 0.f),
    C_Vertex3f(0.259579659f, 0.643863916f, 0.f),
    C_Vertex3f(0.271406889f, 0.590488911f, 0.f),
    C_Vertex3f(0.279975891f, 0.647859693f, 0.f),
    C_Vertex3f(0.288432956f, 0.59382844f, 0.f),
    C_Vertex3f(0.300495148f, 0.650523543f, 0.f),
    C_Vertex3f(0.305575013f, 0.596054792f, 0.f),
    C_Vertex3f(0.321137428f, 0.651855469f, 0.f),
    C_Vertex3f(0.322833061f, 0.597167969f, 0.f),
    C_Vertex3f(0.341901779f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.144046783f, 0.524640322f, 0.f),
    C_Vertex3f(0.157808542f, 0.536414266f, 0.f),
    C_Vertex3f(0.159443855f, 0.603906155f, 0.f),
    C_Vertex3f(0.172465086f, 0.547232032f, 0.f),
    C_Vertex3f(0.179224968f, 0.614561558f, 0.f),
    C_Vertex3f(0.188016415f, 0.55709362f, 0.f),
    C_Vertex3f(0.199129105f, 0.623885036f, 0.f),
    C_Vertex3f(0.204462528f, 0.565999031f, 0.f),
    C_Vertex3f(0.219156265f, 0.631876588f, 0.f),
    C_Vertex3f(0.221024632f, 0.573791265f, 0.f),
    C_Vertex3f(0.23930645f, 0.638536215f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.108130217f, 0.483581424f, 0.f),
    C_Vertex3f(0.119207621f, 0.498223901f, 0.f),
    C_Vertex3f(0.12313962f, 0.579152465f, 0.f),
    C_Vertex3f(0.13117981f, 0.5119102f, 0.f),
    C_Vertex3f(0.140748739f, 0.592103124f, 0.f),
    C_Vertex3f(0.144046783f, 0.524640322f, 0.f),
    C_Vertex3f(0.159443855f, 0.603906155f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0805329084f, 0.434702277f, 0.f),
    C_Vertex3f(0.0886597633f, 0.451427937f, 0.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, 0.f),
    C_Vertex3f(0.0979475975f, 0.467982769f, 0.f),
    C_Vertex3f(0.106616497f, 0.565054178f, 0.f),
    C_Vertex3f(0.108130217f, 0.483581424f, 0.f),
    C_Vertex3f(0.12313962f, 0.579152465f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0768282413f, 0.533414721f, 0.f),
    C_Vertex3f(0.0635631084f, 0.515873551f, 0.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, 0.f),
    C_Vertex3f(0.0735670328f, 0.417805791f, 0.f),
    C_Vertex3f(0.0805329084f, 0.434702277f, 0.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0635631084f, 0.515873551f, 0.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, 0.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, 0.f),
    C_Vertex3f(0.057313323f, 0.348511577f, 0.f),
    C_Vertex3f(0.0596352816f, 0.366091371f, 0.f),
    C_Vertex3f(0.0631182194f, 0.383500338f, 0.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.00146484375f, 0.332565308f, 0.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, 0.f),
    C_Vertex3f(0.00285148621f, 0.353883743f, 0.f),
    C_Vertex3f(0.00556325912f, 0.269137621f, 0.f),
    C_Vertex3f(0.00562477112f, 0.374990463f, 0.f),
    C_Vertex3f(0.0096616745f, 0.248441219f, 0.f),
    C_Vertex3f(0.00978469849f, 0.395885468f, 0.f),
    C_Vertex3f(0.0151262283f, 0.227949619f, 0.f),
    C_Vertex3f(0.0153312683f, 0.416568756f, 0.f),
    C_Vertex3f(0.0219569206f, 0.207662821f, 0.f),
    C_Vertex3f(0.0222644806f, 0.437040329f, 0.f),
    C_Vertex3f(0.0301537514f, 0.187580824f, 0.f),
    C_Vertex3f(0.0305843353f, 0.457300186f, 0.f),
    C_Vertex3f(0.0397167206f, 0.167703629f, 0.f),
    C_Vertex3f(0.0402908325f, 0.477348328f, 0.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, 0.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, 0.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, 0.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.612200737f, 0.148031235f, 0.f),
    C_Vertex3f(0.606445313f, 0.330760956f, 0.f),
    C_Vertex3f(0.606445313f, 0.312894821f, 0.f),
    C_Vertex3f(0.605311632f, 0.295268297f, 0.f),
    C_Vertex3f(0.603044271f, 0.277826071f, 0.f),
    C_Vertex3f(0.600233912f, 0.129458666f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.600233912f, 0.129458666f, 0.f),
    C_Vertex3f(0.603044271f, 0.277826071f, 0.f),
    C_Vertex3f(0.59964323f, 0.260568142f, 0.f),
    C_Vertex3f(0.595108509f, 0.243494511f, 0.f),
    C_Vertex3f(0.589440107f, 0.226605177f, 0.f),
    C_Vertex3f(0.587174058f, 0.111985922f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.589440107f, 0.226605177f, 0.f),
    C_Vertex3f(0.582638025f, 0.209900141f, 0.f),
    C_Vertex3f(0.587174058f, 0.111985922f, 0.f),
    C_Vertex3f(0.574702263f, 0.193379402f, 0.f),
    C_Vertex3f(0.573021173f, 0.0956130028f, 0.f),
    C_Vertex3f(0.56563282f, 0.177042961f, 0.f),
    C_Vertex3f(0.557775259f, 0.0803399086f, 0.f),
    C_Vertex3f(0.555648208f, 0.1616081f, 0.f),
    C_Vertex3f(0.544748425f, 0.147074819f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.557775259f, 0.0803399086f, 0.f),
    C_Vertex3f(0.544748425f, 0.147074819f, 0.f),
    C_Vertex3f(0.541436315f, 0.0661666393f, 0.f),
    C_Vertex3f(0.532933474f, 0.133443117f, 0.f),
    C_Vertex3f(0.52400434f, 0.053093195f, 0.f),
    C_Vertex3f(0.520203352f, 0.120712996f, 0.f),
    C_Vertex3f(0.506558061f, 0.108884454f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52400434f, 0.053093195f, 0.f),
    C_Vertex3f(0.506558061f, 0.108884454f, 0.f),
    C_Vertex3f(0.505479336f, 0.0411195755f, 0.f),
    C_Vertex3f(0.4919976f, 0.0979574919f, 0.f),
    C_Vertex3f(0.485861301f, 0.0302457809f, 0.f),
    C_Vertex3f(0.476521969f, 0.0879321098f, 0.f),
    C_Vertex3f(0.466018081f, 0.0207312107f, 0.f),
    C_Vertex3f(0.460131168f, 0.0788083076f, 0.f),
    C_Vertex3f(0.445949674f, 0.0125758648f, 0.f),
    C_Vertex3f(0.443562865f, 0.0708249807f, 0.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445949674f, 0.0125758648f, 0.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, 0.f),
    C_Vertex3f(0.42565608f, 0.00577974319f, 0.f),
    C_Vertex3f(0.409893751f, 0.0582797527f, 0.f),
    C_Vertex3f(0.4051373f, 0.000342845917f, 0.f),
    C_Vertex3f(0.39279294f, 0.0537178516f, 0.f),
    C_Vertex3f(0.384393334f, -0.00373482704f, 0.f),
    C_Vertex3f(0.375514627f, 0.0502964258f, 0.f),
    C_Vertex3f(0.363424182f, -0.00645327568f, 0.f),
    C_Vertex3f(0.35805881f, 0.0480154753f, 0.f),
    C_Vertex3f(0.342229843f, -0.0078125f, 0.f),
    C_Vertex3f(0.340425491f, 0.046875f, 0.f),
    C_Vertex3f(0.322613716f, 0.046875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.342229843f, -0.0078125f, 0.f),
    C_Vertex3f(0.322613716f, 0.046875f, 0.f),
    C_Vertex3f(0.320809364f, -0.0078125f, 0.f),
    C_Vertex3f(0.304978728f, 0.0480154753f, 0.f),
    C_Vertex3f(0.299613476f, -0.00645327568f, 0.f),
    C_Vertex3f(0.287521482f, 0.0502964258f, 0.f),
    C_Vertex3f(0.278643131f, -0.00373482704f, 0.f),
    C_Vertex3f(0.270241976f, 0.0537178516f, 0.f),
    C_Vertex3f(0.257898331f, 0.000342845917f, 0.f),
    C_Vertex3f(0.253140211f, 0.0582797527f, 0.f),
    C_Vertex3f(0.237379074f, 0.00577974319f, 0.f),
    C_Vertex3f(0.236216187f, 0.0639821291f, 0.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.237379074f, 0.00577974319f, 0.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, 0.f),
    C_Vertex3f(0.217085361f, 0.0125758648f, 0.f),
    C_Vertex3f(0.202901363f, 0.0788083076f, 0.f),
    C_Vertex3f(0.197017193f, 0.0207312107f, 0.f),
    C_Vertex3f(0.186510563f, 0.0879321098f, 0.f),
    C_Vertex3f(0.177174568f, 0.0302457809f, 0.f),
    C_Vertex3f(0.171028137f, 0.0979574919f, 0.f),
    C_Vertex3f(0.157557487f, 0.0411195755f, 0.f),
    C_Vertex3f(0.156454086f, 0.108884454f, 0.f),
    C_Vertex3f(0.14278841f, 0.120712996f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157557487f, 0.0411195755f, 0.f),
    C_Vertex3f(0.14278841f, 0.120712996f, 0.f),
    C_Vertex3f(0.139026403f, 0.053093195f, 0.f),
    C_Vertex3f(0.130031109f, 0.133443117f, 0.f),
    C_Vertex3f(0.121581316f, 0.0661666393f, 0.f),
    C_Vertex3f(0.118182182f, 0.147074819f, 0.f),
    C_Vertex3f(0.107241631f, 0.1616081f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.121581316f, 0.0661666393f, 0.f),
    C_Vertex3f(0.107241631f, 0.1616081f, 0.f),
    C_Vertex3f(0.105222225f, 0.0803399086f, 0.f),
    C_Vertex3f(0.0972094536f, 0.177042961f, 0.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, 0.f),
    C_Vertex3f(0.0880856514f, 0.193379402f, 0.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0757620335f, 0.111985922f, 0.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, 0.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, 0.f),
    C_Vertex3f(0.0732594728f, 0.226605177f, 0.f),
    C_Vertex3f(0.0675570965f, 0.243494511f, 0.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, 0.f),
    C_Vertex3f(0.0626609325f, 0.129458666f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0626609325f, 0.129458666f, 0.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, 0.f),
    C_Vertex3f(0.0595737696f, 0.277826071f, 0.f),
    C_Vertex3f(0.057292819f, 0.295268297f, 0.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, 0.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.491210938f, 0.252441406f, 0.f),
    C_Vertex3f(0.4375f, 0.268066406f, 0.f),
    C_Vertex3f(0.440761089f, 0.169406414f, 0.f),
    C_Vertex3f(0.450502992f, 0.178227305f, 0.f),
    C_Vertex3f(0.459321856f, 0.187737107f, 0.f),
    C_Vertex3f(0.467217684f, 0.19793582f, 0.f),
    C_Vertex3f(0.474190474f, 0.208823442f, 0.f),
    C_Vertex3f(0.480240226f, 0.220399976f, 0.f),
    C_Vertex3f(0.48536694f, 0.23266542f, 0.f),
    C_Vertex3f(0.489570618f, 0.245619774f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339556694f, 0.1328125f, 0.f),
    C_Vertex3f(0.354192853f, 0.133828998f, 0.f),
    C_Vertex3f(0.357310295f, 0.186910629f, 0.f),
    C_Vertex3f(0.368261695f, 0.135861993f, 0.f),
    C_Vertex3f(0.374596596f, 0.192567825f, 0.f),
    C_Vertex3f(0.38176322f, 0.138911486f, 0.f),
    C_Vertex3f(0.390678406f, 0.201053619f, 0.f),
    C_Vertex3f(0.394697428f, 0.142977476f, 0.f),
    C_Vertex3f(0.405555725f, 0.212368011f, 0.f),
    C_Vertex3f(0.407064319f, 0.148059964f, 0.f),
    C_Vertex3f(0.417917252f, 0.225694656f, 0.f),
    C_Vertex3f(0.418863893f, 0.15415895f, 0.f),
    C_Vertex3f(0.427762985f, 0.241033554f, 0.f),
    C_Vertex3f(0.430096149f, 0.161274433f, 0.f),
    C_Vertex3f(0.435092926f, 0.258384705f, 0.f),
    C_Vertex3f(0.440761089f, 0.169406414f, 0.f),
    C_Vertex3f(0.4375f, 0.268066406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.357310295f, 0.186910629f, 0.f),
    C_Vertex3f(0.338819504f, 0.184082031f, 0.f),
    C_Vertex3f(0.339556694f, 0.1328125f, 0.f),
    C_Vertex3f(0.322845459f, 0.184082031f, 0.f),
    C_Vertex3f(0.322328568f, 0.1328125f, 0.f),
    C_Vertex3f(0.311160207f, 0.18515563f, 0.f),
    C_Vertex3f(0.303978443f, 0.134371161f, 0.f),
    C_Vertex3f(0.30004251f, 0.187302828f, 0.f),
    C_Vertex3f(0.289492369f, 0.190523624f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.286530972f, 0.137488484f, 0.f),
    C_Vertex3f(0.303978443f, 0.134371161f, 0.f),
    C_Vertex3f(0.289492369f, 0.190523624f, 0.f),
    C_Vertex3f(0.279509783f, 0.19481802f, 0.f),
    C_Vertex3f(0.270094752f, 0.200186014f, 0.f),
    C_Vertex3f(0.269986153f, 0.142164469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.261247277f, 0.206627607f, 0.f),
    C_Vertex3f(0.252967358f, 0.214142799f, 0.f),
    C_Vertex3f(0.254343987f, 0.148399115f, 0.f),
    C_Vertex3f(0.269986153f, 0.142164469f, 0.f),
    C_Vertex3f(0.270094752f, 0.200186014f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245254993f, 0.22273159f, 0.f),
    C_Vertex3f(0.238506675f, 0.232339382f, 0.f),
    C_Vertex3f(0.239604473f, 0.156192422f, 0.f),
    C_Vertex3f(0.254343987f, 0.148399115f, 0.f),
    C_Vertex3f(0.252967358f, 0.214142799f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.238506675f, 0.232339382f, 0.f),
    C_Vertex3f(0.232722402f, 0.242966175f, 0.f),
    C_Vertex3f(0.239604473f, 0.156192422f, 0.f),
    C_Vertex3f(0.227902174f, 0.254611969f, 0.f),
    C_Vertex3f(0.225767612f, 0.165544391f, 0.f),
    C_Vertex3f(0.224045992f, 0.267276764f, 0.f),
    C_Vertex3f(0.221153855f, 0.28096056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.167715073f, 0.250833273f, 0.f),
    C_Vertex3f(0.17373085f, 0.233551264f, 0.f),
    C_Vertex3f(0.175110817f, 0.410687447f, 0.f),
    C_Vertex3f(0.181250572f, 0.217472434f, 0.f),
    C_Vertex3f(0.186057091f, 0.432017326f, 0.f),
    C_Vertex3f(0.190274239f, 0.202596784f, 0.f),
    C_Vertex3f(0.199328899f, 0.450721741f, 0.f),
    C_Vertex3f(0.200801849f, 0.188924313f, 0.f),
    C_Vertex3f(0.214926243f, 0.46680069f, 0.f),
    C_Vertex3f(0.212833405f, 0.176455021f, 0.f),
    C_Vertex3f(0.218261719f, 0.311385155f, 0.f),
    C_Vertex3f(0.225767612f, 0.165544391f, 0.f),
    C_Vertex3f(0.219225764f, 0.295663357f, 0.f),
    C_Vertex3f(0.221153855f, 0.28096056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.158691406f, 0.335382462f, 0.f),
    C_Vertex3f(0.160195351f, 0.289006829f, 0.f),
    C_Vertex3f(0.161427975f, 0.362370014f, 0.f),
    C_Vertex3f(0.163203239f, 0.269318461f, 0.f),
    C_Vertex3f(0.166901112f, 0.387471676f, 0.f),
    C_Vertex3f(0.167715073f, 0.250833273f, 0.f),
    C_Vertex3f(0.175110817f, 0.410687447f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.484863281f, 0.403808594f, 0.f),
    C_Vertex3f(0.483112335f, 0.409785271f, 0.f),
    C_Vertex3f(0.478844166f, 0.421114802f, 0.f),
    C_Vertex3f(0.473809719f, 0.431820512f, 0.f),
    C_Vertex3f(0.468008995f, 0.441902399f, 0.f),
    C_Vertex3f(0.461441994f, 0.451360464f, 0.f),
    C_Vertex3f(0.454108715f, 0.460194707f, 0.f),
    C_Vertex3f(0.446009159f, 0.468405128f, 0.f),
    C_Vertex3f(0.437143326f, 0.475991726f, 0.f),
    C_Vertex3f(0.432617188f, 0.391113281f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.432617188f, 0.391113281f, 0.f),
    C_Vertex3f(0.437143326f, 0.475991726f, 0.f),
    C_Vertex3f(0.429336548f, 0.399261475f, 0.f),
    C_Vertex3f(0.427511215f, 0.482954502f, 0.f),
    C_Vertex3f(0.421216488f, 0.413658142f, 0.f),
    C_Vertex3f(0.417263508f, 0.489046931f, 0.f),
    C_Vertex3f(0.411537647f, 0.42615509f, 0.f),
    C_Vertex3f(0.406400204f, 0.494269013f, 0.f),
    C_Vertex3f(0.400300026f, 0.436752319f, 0.f),
    C_Vertex3f(0.394921303f, 0.498620749f, 0.f),
    C_Vertex3f(0.387503624f, 0.445449829f, 0.f),
    C_Vertex3f(0.382826805f, 0.502102137f, 0.f),
    C_Vertex3f(0.373586178f, 0.451972961f, 0.f),
    C_Vertex3f(0.370116711f, 0.504713178f, 0.f),
    C_Vertex3f(0.358547688f, 0.456321716f, 0.f),
    C_Vertex3f(0.356791019f, 0.506453872f, 0.f),
    C_Vertex3f(0.342849731f, 0.507324219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.303966403f, 0.455131531f, 0.f),
    C_Vertex3f(0.315436006f, 0.457374573f, 0.f),
    C_Vertex3f(0.322546005f, 0.507324219f, 0.f),
    C_Vertex3f(0.327459335f, 0.458496094f, 0.f),
    C_Vertex3f(0.342849731f, 0.507324219f, 0.f),
    C_Vertex3f(0.342388153f, 0.458496094f, 0.f),
    C_Vertex3f(0.358547688f, 0.456321716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.297945976f, 0.504617214f, 0.f),
    C_Vertex3f(0.274796486f, 0.499203205f, 0.f),
    C_Vertex3f(0.282688379f, 0.447280884f, 0.f),
    C_Vertex3f(0.293050528f, 0.451766968f, 0.f),
    C_Vertex3f(0.303966403f, 0.455131531f, 0.f),
    C_Vertex3f(0.322546005f, 0.507324219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.274796486f, 0.499203205f, 0.f),
    C_Vertex3f(0.253097534f, 0.491082191f, 0.f),
    C_Vertex3f(0.254924297f, 0.427093506f, 0.f),
    C_Vertex3f(0.263625264f, 0.434944153f, 0.f),
    C_Vertex3f(0.272879958f, 0.441673279f, 0.f),
    C_Vertex3f(0.282688379f, 0.447280884f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.253097534f, 0.491082191f, 0.f),
    C_Vertex3f(0.232849121f, 0.480254173f, 0.f),
    C_Vertex3f(0.233537793f, 0.397263765f, 0.f),
    C_Vertex3f(0.239648223f, 0.408178091f, 0.f),
    C_Vertex3f(0.246777058f, 0.418121338f, 0.f),
    C_Vertex3f(0.254924297f, 0.427093506f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.232849121f, 0.480254173f, 0.f),
    C_Vertex3f(0.214926243f, 0.46680069f, 0.f),
    C_Vertex3f(0.219280124f, 0.343895674f, 0.f),
    C_Vertex3f(0.221316934f, 0.358694315f, 0.f),
    C_Vertex3f(0.224372149f, 0.372521877f, 0.f),
    C_Vertex3f(0.228445768f, 0.385378361f, 0.f),
    C_Vertex3f(0.233537793f, 0.397263765f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.218261719f, 0.328125954f, 0.f),
    C_Vertex3f(0.219280124f, 0.343895674f, 0.f),
    C_Vertex3f(0.214926243f, 0.46680069f, 0.f),
    C_Vertex3f(0.218261719f, 0.311385155f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.160195351f, 0.289006829f, 0.f),
    C_Vertex3f(0.158691406f, 0.335382462f, 0.f),
    C_Vertex3f(0.158691406f, 0.309898376f, 0.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, 0.f),
    C_Vertex3f(0.00146484375f, 0.332565308f, 0.f),
    C_Vertex3f(0.00146484375f, 0.311144829f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.661132813f, 0.311144829f, -1.f),
    C_Vertex3f(0.661132813f, 0.332565308f, -1.f),
    C_Vertex3f(0.659773588f, 0.290038824f, -1.f),
    C_Vertex3f(0.659753084f, 0.353883743f, -1.f),
    C_Vertex3f(0.65705514f, 0.269137621f, -1.f),
    C_Vertex3f(0.656993628f, 0.374990463f, -1.f),
    C_Vertex3f(0.652977467f, 0.248441219f, -1.f),
    C_Vertex3f(0.652854443f, 0.395885468f, -1.f),
    C_Vertex3f(0.647540569f, 0.227949619f, -1.f),
    C_Vertex3f(0.647335529f, 0.416568756f, -1.f),
    C_Vertex3f(0.640744448f, 0.207662821f, -1.f),
    C_Vertex3f(0.640436888f, 0.437040329f, -1.f),
    C_Vertex3f(0.632589102f, 0.187580824f, -1.f),
    C_Vertex3f(0.632158518f, 0.457300186f, -1.f),
    C_Vertex3f(0.623074532f, 0.167703629f, -1.f),
    C_Vertex3f(0.62250042f, 0.477348328f, -1.f),
    C_Vertex3f(0.612200737f, 0.148031235f, -1.f),
    C_Vertex3f(0.611462593f, 0.497184753f, -1.f),
    C_Vertex3f(0.606445313f, 0.330760956f, -1.f),
    C_Vertex3f(0.605291128f, 0.348511577f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.589132547f, 0.417805791f, -1.f),
    C_Vertex3f(0.594903469f, 0.400738478f, -1.f),
    C_Vertex3f(0.599331737f, 0.515873551f, -1.f),
    C_Vertex3f(0.599520206f, 0.383500338f, -1.f),
    C_Vertex3f(0.611462593f, 0.497184753f, -1.f),
    C_Vertex3f(0.602982759f, 0.366091371f, -1.f),
    C_Vertex3f(0.605291128f, 0.348511577f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.58610785f, 0.533414721f, -1.f),
    C_Vertex3f(0.571790934f, 0.549808264f, -1.f),
    C_Vertex3f(0.574128151f, 0.451427937f, -1.f),
    C_Vertex3f(0.582207441f, 0.434702277f, -1.f),
    C_Vertex3f(0.589132547f, 0.417805791f, -1.f),
    C_Vertex3f(0.599331737f, 0.515873551f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.543702602f, 0.498223901f, -1.f),
    C_Vertex3f(0.554752827f, 0.483581424f, -1.f),
    C_Vertex3f(0.556380987f, 0.565054178f, -1.f),
    C_Vertex3f(0.564894676f, 0.467982769f, -1.f),
    C_Vertex3f(0.571790934f, 0.549808264f, -1.f),
    C_Vertex3f(0.574128151f, 0.451427937f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.505101681f, 0.536414266f, -1.f),
    C_Vertex3f(0.518877029f, 0.524640322f, -1.f),
    C_Vertex3f(0.522282004f, 0.592103124f, -1.f),
    C_Vertex3f(0.531744003f, 0.5119102f, -1.f),
    C_Vertex3f(0.539878011f, 0.579152465f, -1.f),
    C_Vertex3f(0.543702602f, 0.498223901f, -1.f),
    C_Vertex3f(0.556380987f, 0.565054178f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.425038338f, 0.580470324f, -1.f),
    C_Vertex3f(0.441729546f, 0.573791265f, -1.f),
    C_Vertex3f(0.44387877f, 0.631876588f, -1.f),
    C_Vertex3f(0.458325386f, 0.565999031f, -1.f),
    C_Vertex3f(0.463906169f, 0.623885036f, -1.f),
    C_Vertex3f(0.474825859f, 0.55709362f, -1.f),
    C_Vertex3f(0.483810902f, 0.614561558f, -1.f),
    C_Vertex3f(0.490417957f, 0.547232032f, -1.f),
    C_Vertex3f(0.503592968f, 0.603906155f, -1.f),
    C_Vertex3f(0.505101681f, 0.536414266f, -1.f),
    C_Vertex3f(0.522282004f, 0.592103124f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.322833061f, 0.597167969f, -1.f),
    C_Vertex3f(0.340151787f, 0.597167969f, -1.f),
    C_Vertex3f(0.341901779f, 0.651855469f, -1.f),
    C_Vertex3f(0.357319832f, 0.596054792f, -1.f),
    C_Vertex3f(0.36254251f, 0.650523543f, -1.f),
    C_Vertex3f(0.374392509f, 0.59382844f, -1.f),
    C_Vertex3f(0.383060575f, 0.647859693f, -1.f),
    C_Vertex3f(0.39136982f, 0.590488911f, -1.f),
    C_Vertex3f(0.403455973f, 0.643863916f, -1.f),
    C_Vertex3f(0.408251762f, 0.586036205f, -1.f),
    C_Vertex3f(0.423728704f, 0.638536215f, -1.f),
    C_Vertex3f(0.425038338f, 0.580470324f, -1.f),
    C_Vertex3f(0.44387877f, 0.631876588f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.221024632f, 0.573791265f, -1.f),
    C_Vertex3f(0.237702727f, 0.580470324f, -1.f),
    C_Vertex3f(0.23930645f, 0.638536215f, -1.f),
    C_Vertex3f(0.254496813f, 0.586036205f, -1.f),
    C_Vertex3f(0.259579659f, 0.643863916f, -1.f),
    C_Vertex3f(0.271406889f, 0.590488911f, -1.f),
    C_Vertex3f(0.279975891f, 0.647859693f, -1.f),
    C_Vertex3f(0.288432956f, 0.59382844f, -1.f),
    C_Vertex3f(0.300495148f, 0.650523543f, -1.f),
    C_Vertex3f(0.305575013f, 0.596054792f, -1.f),
    C_Vertex3f(0.321137428f, 0.651855469f, -1.f),
    C_Vertex3f(0.322833061f, 0.597167969f, -1.f),
    C_Vertex3f(0.341901779f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.144046783f, 0.524640322f, -1.f),
    C_Vertex3f(0.157808542f, 0.536414266f, -1.f),
    C_Vertex3f(0.159443855f, 0.603906155f, -1.f),
    C_Vertex3f(0.172465086f, 0.547232032f, -1.f),
    C_Vertex3f(0.179224968f, 0.614561558f, -1.f),
    C_Vertex3f(0.188016415f, 0.55709362f, -1.f),
    C_Vertex3f(0.199129105f, 0.623885036f, -1.f),
    C_Vertex3f(0.204462528f, 0.565999031f, -1.f),
    C_Vertex3f(0.219156265f, 0.631876588f, -1.f),
    C_Vertex3f(0.221024632f, 0.573791265f, -1.f),
    C_Vertex3f(0.23930645f, 0.638536215f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.108130217f, 0.483581424f, -1.f),
    C_Vertex3f(0.119207621f, 0.498223901f, -1.f),
    C_Vertex3f(0.12313962f, 0.579152465f, -1.f),
    C_Vertex3f(0.13117981f, 0.5119102f, -1.f),
    C_Vertex3f(0.140748739f, 0.592103124f, -1.f),
    C_Vertex3f(0.144046783f, 0.524640322f, -1.f),
    C_Vertex3f(0.159443855f, 0.603906155f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0805329084f, 0.434702277f, -1.f),
    C_Vertex3f(0.0886597633f, 0.451427937f, -1.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, -1.f),
    C_Vertex3f(0.0979475975f, 0.467982769f, -1.f),
    C_Vertex3f(0.106616497f, 0.565054178f, -1.f),
    C_Vertex3f(0.108130217f, 0.483581424f, -1.f),
    C_Vertex3f(0.12313962f, 0.579152465f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0768282413f, 0.533414721f, -1.f),
    C_Vertex3f(0.0635631084f, 0.515873551f, -1.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, -1.f),
    C_Vertex3f(0.0735670328f, 0.417805791f, -1.f),
    C_Vertex3f(0.0805329084f, 0.434702277f, -1.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0635631084f, 0.515873551f, -1.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, -1.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, -1.f),
    C_Vertex3f(0.057313323f, 0.348511577f, -1.f),
    C_Vertex3f(0.0596352816f, 0.366091371f, -1.f),
    C_Vertex3f(0.0631182194f, 0.383500338f, -1.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.00146484375f, 0.332565308f, -1.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, -1.f),
    C_Vertex3f(0.00285148621f, 0.353883743f, -1.f),
    C_Vertex3f(0.00556325912f, 0.269137621f, -1.f),
    C_Vertex3f(0.00562477112f, 0.374990463f, -1.f),
    C_Vertex3f(0.0096616745f, 0.248441219f, -1.f),
    C_Vertex3f(0.00978469849f, 0.395885468f, -1.f),
    C_Vertex3f(0.0151262283f, 0.227949619f, -1.f),
    C_Vertex3f(0.0153312683f, 0.416568756f, -1.f),
    C_Vertex3f(0.0219569206f, 0.207662821f, -1.f),
    C_Vertex3f(0.0222644806f, 0.437040329f, -1.f),
    C_Vertex3f(0.0301537514f, 0.187580824f, -1.f),
    C_Vertex3f(0.0305843353f, 0.457300186f, -1.f),
    C_Vertex3f(0.0397167206f, 0.167703629f, -1.f),
    C_Vertex3f(0.0402908325f, 0.477348328f, -1.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, -1.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, -1.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, -1.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.612200737f, 0.148031235f, -1.f),
    C_Vertex3f(0.606445313f, 0.330760956f, -1.f),
    C_Vertex3f(0.606445313f, 0.312894821f, -1.f),
    C_Vertex3f(0.605311632f, 0.295268297f, -1.f),
    C_Vertex3f(0.603044271f, 0.277826071f, -1.f),
    C_Vertex3f(0.600233912f, 0.129458666f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.600233912f, 0.129458666f, -1.f),
    C_Vertex3f(0.603044271f, 0.277826071f, -1.f),
    C_Vertex3f(0.59964323f, 0.260568142f, -1.f),
    C_Vertex3f(0.595108509f, 0.243494511f, -1.f),
    C_Vertex3f(0.589440107f, 0.226605177f, -1.f),
    C_Vertex3f(0.587174058f, 0.111985922f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.589440107f, 0.226605177f, -1.f),
    C_Vertex3f(0.582638025f, 0.209900141f, -1.f),
    C_Vertex3f(0.587174058f, 0.111985922f, -1.f),
    C_Vertex3f(0.574702263f, 0.193379402f, -1.f),
    C_Vertex3f(0.573021173f, 0.0956130028f, -1.f),
    C_Vertex3f(0.56563282f, 0.177042961f, -1.f),
    C_Vertex3f(0.557775259f, 0.0803399086f, -1.f),
    C_Vertex3f(0.555648208f, 0.1616081f, -1.f),
    C_Vertex3f(0.544748425f, 0.147074819f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.557775259f, 0.0803399086f, -1.f),
    C_Vertex3f(0.544748425f, 0.147074819f, -1.f),
    C_Vertex3f(0.541436315f, 0.0661666393f, -1.f),
    C_Vertex3f(0.532933474f, 0.133443117f, -1.f),
    C_Vertex3f(0.52400434f, 0.053093195f, -1.f),
    C_Vertex3f(0.520203352f, 0.120712996f, -1.f),
    C_Vertex3f(0.506558061f, 0.108884454f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52400434f, 0.053093195f, -1.f),
    C_Vertex3f(0.506558061f, 0.108884454f, -1.f),
    C_Vertex3f(0.505479336f, 0.0411195755f, -1.f),
    C_Vertex3f(0.4919976f, 0.0979574919f, -1.f),
    C_Vertex3f(0.485861301f, 0.0302457809f, -1.f),
    C_Vertex3f(0.476521969f, 0.0879321098f, -1.f),
    C_Vertex3f(0.466018081f, 0.0207312107f, -1.f),
    C_Vertex3f(0.460131168f, 0.0788083076f, -1.f),
    C_Vertex3f(0.445949674f, 0.0125758648f, -1.f),
    C_Vertex3f(0.443562865f, 0.0708249807f, -1.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445949674f, 0.0125758648f, -1.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, -1.f),
    C_Vertex3f(0.42565608f, 0.00577974319f, -1.f),
    C_Vertex3f(0.409893751f, 0.0582797527f, -1.f),
    C_Vertex3f(0.4051373f, 0.000342845917f, -1.f),
    C_Vertex3f(0.39279294f, 0.0537178516f, -1.f),
    C_Vertex3f(0.384393334f, -0.00373482704f, -1.f),
    C_Vertex3f(0.375514627f, 0.0502964258f, -1.f),
    C_Vertex3f(0.363424182f, -0.00645327568f, -1.f),
    C_Vertex3f(0.35805881f, 0.0480154753f, -1.f),
    C_Vertex3f(0.342229843f, -0.0078125f, -1.f),
    C_Vertex3f(0.340425491f, 0.046875f, -1.f),
    C_Vertex3f(0.322613716f, 0.046875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.342229843f, -0.0078125f, -1.f),
    C_Vertex3f(0.322613716f, 0.046875f, -1.f),
    C_Vertex3f(0.320809364f, -0.0078125f, -1.f),
    C_Vertex3f(0.304978728f, 0.0480154753f, -1.f),
    C_Vertex3f(0.299613476f, -0.00645327568f, -1.f),
    C_Vertex3f(0.287521482f, 0.0502964258f, -1.f),
    C_Vertex3f(0.278643131f, -0.00373482704f, -1.f),
    C_Vertex3f(0.270241976f, 0.0537178516f, -1.f),
    C_Vertex3f(0.257898331f, 0.000342845917f, -1.f),
    C_Vertex3f(0.253140211f, 0.0582797527f, -1.f),
    C_Vertex3f(0.237379074f, 0.00577974319f, -1.f),
    C_Vertex3f(0.236216187f, 0.0639821291f, -1.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.237379074f, 0.00577974319f, -1.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, -1.f),
    C_Vertex3f(0.217085361f, 0.0125758648f, -1.f),
    C_Vertex3f(0.202901363f, 0.0788083076f, -1.f),
    C_Vertex3f(0.197017193f, 0.0207312107f, -1.f),
    C_Vertex3f(0.186510563f, 0.0879321098f, -1.f),
    C_Vertex3f(0.177174568f, 0.0302457809f, -1.f),
    C_Vertex3f(0.171028137f, 0.0979574919f, -1.f),
    C_Vertex3f(0.157557487f, 0.0411195755f, -1.f),
    C_Vertex3f(0.156454086f, 0.108884454f, -1.f),
    C_Vertex3f(0.14278841f, 0.120712996f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157557487f, 0.0411195755f, -1.f),
    C_Vertex3f(0.14278841f, 0.120712996f, -1.f),
    C_Vertex3f(0.139026403f, 0.053093195f, -1.f),
    C_Vertex3f(0.130031109f, 0.133443117f, -1.f),
    C_Vertex3f(0.121581316f, 0.0661666393f, -1.f),
    C_Vertex3f(0.118182182f, 0.147074819f, -1.f),
    C_Vertex3f(0.107241631f, 0.1616081f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.121581316f, 0.0661666393f, -1.f),
    C_Vertex3f(0.107241631f, 0.1616081f, -1.f),
    C_Vertex3f(0.105222225f, 0.0803399086f, -1.f),
    C_Vertex3f(0.0972094536f, 0.177042961f, -1.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, -1.f),
    C_Vertex3f(0.0880856514f, 0.193379402f, -1.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0757620335f, 0.111985922f, -1.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, -1.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, -1.f),
    C_Vertex3f(0.0732594728f, 0.226605177f, -1.f),
    C_Vertex3f(0.0675570965f, 0.243494511f, -1.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, -1.f),
    C_Vertex3f(0.0626609325f, 0.129458666f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0626609325f, 0.129458666f, -1.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, -1.f),
    C_Vertex3f(0.0595737696f, 0.277826071f, -1.f),
    C_Vertex3f(0.057292819f, 0.295268297f, -1.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, -1.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.491210938f, 0.252441406f, -1.f),
    C_Vertex3f(0.4375f, 0.268066406f, -1.f),
    C_Vertex3f(0.440761089f, 0.169406414f, -1.f),
    C_Vertex3f(0.450502992f, 0.178227305f, -1.f),
    C_Vertex3f(0.459321856f, 0.187737107f, -1.f),
    C_Vertex3f(0.467217684f, 0.19793582f, -1.f),
    C_Vertex3f(0.474190474f, 0.208823442f, -1.f),
    C_Vertex3f(0.480240226f, 0.220399976f, -1.f),
    C_Vertex3f(0.48536694f, 0.23266542f, -1.f),
    C_Vertex3f(0.489570618f, 0.245619774f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339556694f, 0.1328125f, -1.f),
    C_Vertex3f(0.354192853f, 0.133828998f, -1.f),
    C_Vertex3f(0.357310295f, 0.186910629f, -1.f),
    C_Vertex3f(0.368261695f, 0.135861993f, -1.f),
    C_Vertex3f(0.374596596f, 0.192567825f, -1.f),
    C_Vertex3f(0.38176322f, 0.138911486f, -1.f),
    C_Vertex3f(0.390678406f, 0.201053619f, -1.f),
    C_Vertex3f(0.394697428f, 0.142977476f, -1.f),
    C_Vertex3f(0.405555725f, 0.212368011f, -1.f),
    C_Vertex3f(0.407064319f, 0.148059964f, -1.f),
    C_Vertex3f(0.417917252f, 0.225694656f, -1.f),
    C_Vertex3f(0.418863893f, 0.15415895f, -1.f),
    C_Vertex3f(0.427762985f, 0.241033554f, -1.f),
    C_Vertex3f(0.430096149f, 0.161274433f, -1.f),
    C_Vertex3f(0.435092926f, 0.258384705f, -1.f),
    C_Vertex3f(0.440761089f, 0.169406414f, -1.f),
    C_Vertex3f(0.4375f, 0.268066406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.357310295f, 0.186910629f, -1.f),
    C_Vertex3f(0.338819504f, 0.184082031f, -1.f),
    C_Vertex3f(0.339556694f, 0.1328125f, -1.f),
    C_Vertex3f(0.322845459f, 0.184082031f, -1.f),
    C_Vertex3f(0.322328568f, 0.1328125f, -1.f),
    C_Vertex3f(0.311160207f, 0.18515563f, -1.f),
    C_Vertex3f(0.303978443f, 0.134371161f, -1.f),
    C_Vertex3f(0.30004251f, 0.187302828f, -1.f),
    C_Vertex3f(0.289492369f, 0.190523624f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.286530972f, 0.137488484f, -1.f),
    C_Vertex3f(0.303978443f, 0.134371161f, -1.f),
    C_Vertex3f(0.289492369f, 0.190523624f, -1.f),
    C_Vertex3f(0.279509783f, 0.19481802f, -1.f),
    C_Vertex3f(0.270094752f, 0.200186014f, -1.f),
    C_Vertex3f(0.269986153f, 0.142164469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.261247277f, 0.206627607f, -1.f),
    C_Vertex3f(0.252967358f, 0.214142799f, -1.f),
    C_Vertex3f(0.254343987f, 0.148399115f, -1.f),
    C_Vertex3f(0.269986153f, 0.142164469f, -1.f),
    C_Vertex3f(0.270094752f, 0.200186014f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245254993f, 0.22273159f, -1.f),
    C_Vertex3f(0.238506675f, 0.232339382f, -1.f),
    C_Vertex3f(0.239604473f, 0.156192422f, -1.f),
    C_Vertex3f(0.254343987f, 0.148399115f, -1.f),
    C_Vertex3f(0.252967358f, 0.214142799f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.238506675f, 0.232339382f, -1.f),
    C_Vertex3f(0.232722402f, 0.242966175f, -1.f),
    C_Vertex3f(0.239604473f, 0.156192422f, -1.f),
    C_Vertex3f(0.227902174f, 0.254611969f, -1.f),
    C_Vertex3f(0.225767612f, 0.165544391f, -1.f),
    C_Vertex3f(0.224045992f, 0.267276764f, -1.f),
    C_Vertex3f(0.221153855f, 0.28096056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.167715073f, 0.250833273f, -1.f),
    C_Vertex3f(0.17373085f, 0.233551264f, -1.f),
    C_Vertex3f(0.175110817f, 0.410687447f, -1.f),
    C_Vertex3f(0.181250572f, 0.217472434f, -1.f),
    C_Vertex3f(0.186057091f, 0.432017326f, -1.f),
    C_Vertex3f(0.190274239f, 0.202596784f, -1.f),
    C_Vertex3f(0.199328899f, 0.450721741f, -1.f),
    C_Vertex3f(0.200801849f, 0.188924313f, -1.f),
    C_Vertex3f(0.214926243f, 0.46680069f, -1.f),
    C_Vertex3f(0.212833405f, 0.176455021f, -1.f),
    C_Vertex3f(0.218261719f, 0.311385155f, -1.f),
    C_Vertex3f(0.225767612f, 0.165544391f, -1.f),
    C_Vertex3f(0.219225764f, 0.295663357f, -1.f),
    C_Vertex3f(0.221153855f, 0.28096056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.158691406f, 0.335382462f, -1.f),
    C_Vertex3f(0.160195351f, 0.289006829f, -1.f),
    C_Vertex3f(0.161427975f, 0.362370014f, -1.f),
    C_Vertex3f(0.163203239f, 0.269318461f, -1.f),
    C_Vertex3f(0.166901112f, 0.387471676f, -1.f),
    C_Vertex3f(0.167715073f, 0.250833273f, -1.f),
    C_Vertex3f(0.175110817f, 0.410687447f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.484863281f, 0.403808594f, -1.f),
    C_Vertex3f(0.483112335f, 0.409785271f, -1.f),
    C_Vertex3f(0.478844166f, 0.421114802f, -1.f),
    C_Vertex3f(0.473809719f, 0.431820512f, -1.f),
    C_Vertex3f(0.468008995f, 0.441902399f, -1.f),
    C_Vertex3f(0.461441994f, 0.451360464f, -1.f),
    C_Vertex3f(0.454108715f, 0.460194707f, -1.f),
    C_Vertex3f(0.446009159f, 0.468405128f, -1.f),
    C_Vertex3f(0.437143326f, 0.475991726f, -1.f),
    C_Vertex3f(0.432617188f, 0.391113281f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.432617188f, 0.391113281f, -1.f),
    C_Vertex3f(0.437143326f, 0.475991726f, -1.f),
    C_Vertex3f(0.429336548f, 0.399261475f, -1.f),
    C_Vertex3f(0.427511215f, 0.482954502f, -1.f),
    C_Vertex3f(0.421216488f, 0.413658142f, -1.f),
    C_Vertex3f(0.417263508f, 0.489046931f, -1.f),
    C_Vertex3f(0.411537647f, 0.42615509f, -1.f),
    C_Vertex3f(0.406400204f, 0.494269013f, -1.f),
    C_Vertex3f(0.400300026f, 0.436752319f, -1.f),
    C_Vertex3f(0.394921303f, 0.498620749f, -1.f),
    C_Vertex3f(0.387503624f, 0.445449829f, -1.f),
    C_Vertex3f(0.382826805f, 0.502102137f, -1.f),
    C_Vertex3f(0.373586178f, 0.451972961f, -1.f),
    C_Vertex3f(0.370116711f, 0.504713178f, -1.f),
    C_Vertex3f(0.358547688f, 0.456321716f, -1.f),
    C_Vertex3f(0.356791019f, 0.506453872f, -1.f),
    C_Vertex3f(0.342849731f, 0.507324219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.303966403f, 0.455131531f, -1.f),
    C_Vertex3f(0.315436006f, 0.457374573f, -1.f),
    C_Vertex3f(0.322546005f, 0.507324219f, -1.f),
    C_Vertex3f(0.327459335f, 0.458496094f, -1.f),
    C_Vertex3f(0.342849731f, 0.507324219f, -1.f),
    C_Vertex3f(0.342388153f, 0.458496094f, -1.f),
    C_Vertex3f(0.358547688f, 0.456321716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.297945976f, 0.504617214f, -1.f),
    C_Vertex3f(0.274796486f, 0.499203205f, -1.f),
    C_Vertex3f(0.282688379f, 0.447280884f, -1.f),
    C_Vertex3f(0.293050528f, 0.451766968f, -1.f),
    C_Vertex3f(0.303966403f, 0.455131531f, -1.f),
    C_Vertex3f(0.322546005f, 0.507324219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.274796486f, 0.499203205f, -1.f),
    C_Vertex3f(0.253097534f, 0.491082191f, -1.f),
    C_Vertex3f(0.254924297f, 0.427093506f, -1.f),
    C_Vertex3f(0.263625264f, 0.434944153f, -1.f),
    C_Vertex3f(0.272879958f, 0.441673279f, -1.f),
    C_Vertex3f(0.282688379f, 0.447280884f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.253097534f, 0.491082191f, -1.f),
    C_Vertex3f(0.232849121f, 0.480254173f, -1.f),
    C_Vertex3f(0.233537793f, 0.397263765f, -1.f),
    C_Vertex3f(0.239648223f, 0.408178091f, -1.f),
    C_Vertex3f(0.246777058f, 0.418121338f, -1.f),
    C_Vertex3f(0.254924297f, 0.427093506f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.232849121f, 0.480254173f, -1.f),
    C_Vertex3f(0.214926243f, 0.46680069f, -1.f),
    C_Vertex3f(0.219280124f, 0.343895674f, -1.f),
    C_Vertex3f(0.221316934f, 0.358694315f, -1.f),
    C_Vertex3f(0.224372149f, 0.372521877f, -1.f),
    C_Vertex3f(0.228445768f, 0.385378361f, -1.f),
    C_Vertex3f(0.233537793f, 0.397263765f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.218261719f, 0.328125954f, -1.f),
    C_Vertex3f(0.219280124f, 0.343895674f, -1.f),
    C_Vertex3f(0.214926243f, 0.46680069f, -1.f),
    C_Vertex3f(0.218261719f, 0.311385155f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.160195351f, 0.289006829f, -1.f),
    C_Vertex3f(0.158691406f, 0.335382462f, -1.f),
    C_Vertex3f(0.158691406f, 0.309898376f, -1.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, -1.f),
    C_Vertex3f(0.00146484375f, 0.332565308f, -1.f),
    C_Vertex3f(0.00146484375f, 0.311144829f, -1.f),
    C_End(),
    C_Translatef(0.659667969f, 0.f, 0.f),
};
static const Call char_170[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.590193629f, -0.807261765f, 0.f),
    C_Vertex3f(0.238990784f, 0.364578247f, 0.f),
    C_Vertex3f(0.238990784f, 0.364578247f, -1.f),
    C_Normal3f(0.553578854f, -0.832796693f, 0.f),
    C_Vertex3f(0.232340813f, 0.359716415f, 0.f),
    C_Vertex3f(0.232340813f, 0.359716415f, -1.f),
    C_Normal3f(0.503816664f, -0.863810539f, 0.f),
    C_Vertex3f(0.218877077f, 0.35120821f, 0.f),
    C_Vertex3f(0.218877077f, 0.35120821f, -1.f),
    C_Normal3f(0.438290864f, -0.898833156f, 0.f),
    C_Vertex3f(0.205249548f, 0.343915462f, 0.f),
    C_Vertex3f(0.205249548f, 0.343915462f, -1.f),
    C_Normal3f(0.366777211f, -0.930308878f, 0.f),
    C_Vertex3f(0.191458225f, 0.337838173f, 0.f),
    C_Vertex3f(0.191458225f, 0.337838173f, -1.f),
    C_Normal3f(0.290036291f, -0.957015693f, 0.f),
    C_Vertex3f(0.177503109f, 0.332976341f, 0.f),
    C_Vertex3f(0.177503109f, 0.332976341f, -1.f),
    C_Normal3f(0.209240392f, -0.977864325f, 0.f),
    C_Vertex3f(0.163384199f, 0.329329967f, 0.f),
    C_Vertex3f(0.163384199f, 0.329329967f, -1.f),
    C_Normal3f(0.125912085f, -0.992041409f, 0.f),
    C_Vertex3f(0.149101496f, 0.326899052f, 0.f),
    C_Vertex3f(0.149101496f, 0.326899052f, -1.f),
    C_Normal3f(0.043505732f, -0.999053061f, 0.f),
    C_Vertex3f(0.134654999f, 0.325683594f, 0.f),
    C_Vertex3f(0.134654999f, 0.325683594f, -1.f),
    C_Normal3f(-0.0333679989f, -0.999443114f, 0.f),
    C_Vertex3f(0.121190071f, 0.325683594f, 0.f),
    C_Vertex3f(0.121190071f, 0.325683594f, -1.f),
    C_Normal3f(-0.110182509f, -0.993911326f, 0.f),
    C_Vertex3f(0.109511018f, 0.326523066f, 0.f),
    C_Vertex3f(0.109511018f, 0.326523066f, -1.f),
    C_Normal3f(-0.19215478f, -0.981364608f, 0.f),
    C_Vertex3f(0.098472476f, 0.328202009f, 0.f),
    C_Vertex3f(0.098472476f, 0.328202009f, -1.f),
    C_Normal3f(-0.27989468f, -0.960030675f, 0.f),
    C_Vertex3f(0.0880744457f, 0.330720425f, 0.f),
    C_Vertex3f(0.0880744457f, 0.330720425f, -1.f),
    C_Normal3f(-0.371621162f, -0.928384483f, 0.f),
    C_Vertex3f(0.078316927f, 0.334078312f, 0.f),
    C_Vertex3f(0.078316927f, 0.334078312f, -1.f),
    C_Normal3f(-0.464739263f, -0.885447621f, 0.f),
    C_Vertex3f(0.0691999197f, 0.338275671f, 0.f),
    C_Vertex3f(0.0691999197f, 0.338275671f, -1.f),
    C_Normal3f(-0.556045055f, -0.831152081f, 0.f),
    C_Vertex3f(0.060723424f, 0.343312502f, 0.f),
    C_Vertex3f(0.060723424f, 0.343312502f, -1.f),
    C_Normal3f(-0.642164588f, -0.766566873f, 0.f),
    C_Vertex3f(0.0528874397f, 0.349188805f, 0.f),
    C_Vertex3f(0.0528874397f, 0.349188805f, -1.f),
    C_Normal3f(-0.71763283f, -0.696421742f, 0.f),
    C_Vertex3f(0.045691967f, 0.355904579f, 0.f),
    C_Vertex3f(0.045691967f, 0.355904579f, -1.f),
    C_Normal3f(-0.785562932f, -0.618781745f, 0.f),
    C_Vertex3f(0.0393959284f, 0.36309123f, 0.f),
    C_Vertex3f(0.0393959284f, 0.36309123f, -1.f),
    C_Normal3f(-0.847332597f, -0.531062603f, 0.f),
    C_Vertex3f(0.0339993238f, 0.370748758f, 0.f),
    C_Vertex3f(0.0339993238f, 0.370748758f, -1.f),
    C_Normal3f(-0.90014106f, -0.435598671f, 0.f),
    C_Vertex3f(0.0295021534f, 0.378877163f, 0.f),
    C_Vertex3f(0.0295021534f, 0.378877163f, -1.f),
    C_Normal3f(-0.941986084f, -0.335651904f, 0.f),
    C_Vertex3f(0.025904417f, 0.387476444f, 0.f),
    C_Vertex3f(0.025904417f, 0.387476444f, -1.f),
    C_Normal3f(-0.972024858f, -0.234878108f, 0.f),
    C_Vertex3f(0.0232061148f, 0.396546602f, 0.f),
    C_Vertex3f(0.0232061148f, 0.396546602f, -1.f),
    C_Normal3f(-0.990611851f, -0.136704221f, 0.f),
    C_Vertex3f(0.0214072466f, 0.406087637f, 0.f),
    C_Vertex3f(0.0214072466f, 0.406087637f, -1.f),
    C_Normal3f(-0.999038339f, -0.0438439772f, 0.f),
    C_Vertex3f(0.0205078125f, 0.416099548f, 0.f),
    C_Vertex3f(0.0205078125f, 0.416099548f, -1.f),
    C_Normal3f(-0.999079347f, 0.0429007858f, 0.f),
    C_Vertex3f(0.0205078125f, 0.426582336f, 0.f),
    C_Vertex3f(0.0205078125f, 0.426582336f, -1.f),
    C_Normal3f(-0.990651369f, 0.136417806f, 0.f),
    C_Vertex3f(0.0213799477f, 0.43640995f, 0.f),
    C_Vertex3f(0.0213799477f, 0.43640995f, -1.f),
    C_Normal3f(-0.970934689f, 0.239344388f, 0.f),
    C_Vertex3f(0.023124218f, 0.44558239f, 0.f),
    C_Vertex3f(0.023124218f, 0.44558239f, -1.f),
    C_Normal3f(-0.932766974f, 0.360479802f, 0.f),
    C_Vertex3f(0.0257406235f, 0.454099655f, 0.f),
    C_Vertex3f(0.0257406235f, 0.454099655f, -1.f),
    C_Normal3f(-0.855646253f, 0.517561138f, 0.f),
    C_Vertex3f(0.0309734344f, 0.465892792f, 0.f),
    C_Vertex3f(0.0309734344f, 0.465892792f, -1.f),
    C_Normal3f(-0.746169984f, 0.66575551f, 0.f),
    C_Vertex3f(0.0413284302f, 0.479869843f, 0.f),
    C_Vertex3f(0.0413284302f, 0.479869843f, -1.f),
    C_Normal3f(-0.628554642f, 0.777765453f, 0.f),
    C_Vertex3f(0.0516281128f, 0.489042282f, 0.f),
    C_Vertex3f(0.0516281128f, 0.489042282f, -1.f),
    C_Normal3f(-0.528438866f, 0.848971307f, 0.f),
    C_Vertex3f(0.0595026016f, 0.494557381f, 0.f),
    C_Vertex3f(0.0595026016f, 0.494557381f, -1.f),
    C_Normal3f(-0.441240817f, 0.897388756f, 0.f),
    C_Vertex3f(0.0683851242f, 0.499472618f, 0.f),
    C_Vertex3f(0.0683851242f, 0.499472618f, -1.f),
    C_Normal3f(-0.373328924f, 0.927699029f, 0.f),
    C_Vertex3f(0.0782756805f, 0.503787994f, 0.f),
    C_Vertex3f(0.0782756805f, 0.503787994f, -1.f),
    C_Normal3f(-0.313410968f, 0.949617565f, 0.f),
    C_Vertex3f(0.0837249756f, 0.505645752f, 0.f),
    C_Vertex3f(0.0837249756f, 0.505645752f, -1.f),
    C_Normal3f(-0.262680978f, 0.964882731f, 0.f),
    C_Vertex3f(0.0875396729f, 0.506845474f, 0.f),
    C_Vertex3f(0.0875396729f, 0.506845474f, -1.f),
    C_Normal3f(-0.225941598f, 0.974140882f, 0.f),
    C_Vertex3f(0.0967493057f, 0.509191513f, 0.f),
    C_Vertex3f(0.0967493057f, 0.509191513f, -1.f),
    C_Normal3f(-0.192034453f, 0.981388211f, 0.f),
    C_Vertex3f(0.107539177f, 0.511484146f, 0.f),
    C_Vertex3f(0.107539177f, 0.511484146f, -1.f),
    C_Normal3f(-0.148969412f, 0.988841891f, 0.f),
    C_Vertex3f(0.119909286f, 0.513723373f, 0.f),
    C_Vertex3f(0.119909286f, 0.513723373f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.148969412f, 0.988841891f, 0.f),
    C_Vertex3f(0.119909286f, 0.513723373f, 0.f),
    C_Vertex3f(0.119909286f, 0.513723373f, -1.f),
    C_Vertex3f(0.19102478f, 0.524436951f, -1.f),
    C_Vertex3f(0.19102478f, 0.524436951f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.148969412f, 0.988841891f, 0.f),
    C_Vertex3f(0.19102478f, 0.524436951f, 0.f),
    C_Vertex3f(0.19102478f, 0.524436951f, -1.f),
    C_Normal3f(-0.194715708f, 0.980859697f, 0.f),
    C_Vertex3f(0.198877335f, 0.525968552f, 0.f),
    C_Vertex3f(0.198877335f, 0.525968552f, -1.f),
    C_Normal3f(-0.200164378f, 0.979762375f, 0.f),
    C_Vertex3f(0.211965561f, 0.528594017f, 0.f),
    C_Vertex3f(0.211965561f, 0.528594017f, -1.f),
    C_Normal3f(-0.210078061f, 0.977684617f, 0.f),
    C_Vertex3f(0.222436905f, 0.530781746f, 0.f),
    C_Vertex3f(0.222436905f, 0.530781746f, -1.f),
    C_Normal3f(-0.223906621f, 0.974610627f, 0.f),
    C_Vertex3f(0.230291367f, 0.532531738f, 0.f),
    C_Vertex3f(0.230291367f, 0.532531738f, -1.f),
    C_Normal3f(-0.957268f, 0.289202183f, 0.f),
    C_Vertex3f(0.232910156f, 0.533187866f, 0.f),
    C_Vertex3f(0.232910156f, 0.533187866f, -1.f),
    C_Normal3f(-0.99885422f, -0.0478551649f, 0.f),
    C_Vertex3f(0.232801437f, 0.540840149f, 0.f),
    C_Vertex3f(0.232801437f, 0.540840149f, -1.f),
    C_Normal3f(-0.994654715f, -0.103256904f, 0.f),
    C_Vertex3f(0.231930733f, 0.553630829f, 0.f),
    C_Vertex3f(0.231930733f, 0.553630829f, -1.f),
    C_Normal3f(-0.979594529f, -0.200983644f, 0.f),
    C_Vertex3f(0.230406761f, 0.563907623f, 0.f),
    C_Vertex3f(0.230406761f, 0.563907623f, -1.f),
    C_Normal3f(-0.933269739f, -0.359176129f, 0.f),
    C_Vertex3f(0.228229523f, 0.571670532f, 0.f),
    C_Vertex3f(0.228229523f, 0.571670532f, -1.f),
    C_Normal3f(-0.82737118f, -0.561655462f, 0.f),
    C_Vertex3f(0.225399017f, 0.576919556f, 0.f),
    C_Vertex3f(0.225399017f, 0.576919556f, -1.f),
    C_Normal3f(-0.710780799f, -0.703413546f, 0.f),
    C_Vertex3f(0.221344471f, 0.581812859f, 0.f),
    C_Vertex3f(0.221344471f, 0.581812859f, -1.f),
    C_Normal3f(-0.594902277f, -0.80379802f, 0.f),
    C_Vertex3f(0.216065884f, 0.586350441f, 0.f),
    C_Vertex3f(0.216065884f, 0.586350441f, -1.f),
    C_Normal3f(-0.490448415f, -0.871470213f, 0.f),
    C_Vertex3f(0.209563255f, 0.590532303f, 0.f),
    C_Vertex3f(0.209563255f, 0.590532303f, -1.f),
    C_Normal3f(-0.371518433f, -0.92842567f, 0.f),
    C_Vertex3f(0.201836586f, 0.594358444f, 0.f),
    C_Vertex3f(0.201836586f, 0.594358444f, -1.f),
    C_Normal3f(-0.240632802f, -0.970616281f, 0.f),
    C_Vertex3f(0.192830563f, 0.59722805f, 0.f),
    C_Vertex3f(0.192830563f, 0.59722805f, -1.f),
    C_Normal3f(-0.130213261f, -0.991485953f, 0.f),
    C_Vertex3f(0.182545185f, 0.599141121f, 0.f),
    C_Vertex3f(0.182545185f, 0.599141121f, -1.f),
    C_Normal3f(-0.0359568596f, -0.999353349f, 0.f),
    C_Vertex3f(0.170980453f, 0.600097656f, 0.f),
    C_Vertex3f(0.170980453f, 0.600097656f, -1.f),
    C_Normal3f(0.0585353076f, -0.998285294f, 0.f),
    C_Vertex3f(0.155960083f, 0.600097656f, 0.f),
    C_Vertex3f(0.155960083f, 0.600097656f, -1.f),
    C_Normal3f(0.186331689f, -0.982486904f, 0.f),
    C_Vertex3f(0.140940189f, 0.59833622f, 0.f),
    C_Vertex3f(0.140940189f, 0.59833622f, -1.f),
    C_Normal3f(0.35081327f, -0.936445415f, 0.f),
    C_Vertex3f(0.128097057f, 0.594813347f, 0.f),
    C_Vertex3f(0.128097057f, 0.594813347f, -1.f),
    C_Normal3f(0.541238904f, -0.84086895f, 0.f),
    C_Vertex3f(0.117430687f, 0.589529037f, 0.f),
    C_Vertex3f(0.117430687f, 0.589529037f, -1.f),
    C_Normal3f(0.712460637f, -0.701712072f, 0.f),
    C_Vertex3f(0.108941078f, 0.582483292f, 0.f),
    C_Vertex3f(0.108941078f, 0.582483292f, -1.f),
    C_Normal3f(0.838104069f, -0.545510411f, 0.f),
    C_Vertex3f(0.102083683f, 0.573946953f, 0.f),
    C_Vertex3f(0.102083683f, 0.573946953f, -1.f),
    C_Normal3f(0.925479054f, -0.378798723f, 0.f),
    C_Vertex3f(0.0968585014f, 0.563920021f, 0.f),
    C_Vertex3f(0.0968585014f, 0.563920021f, -1.f),
    C_Normal3f(0.969276488f, -0.245973721f, 0.f),
    C_Vertex3f(0.0932655334f, 0.552402496f, 0.f),
    C_Vertex3f(0.0932655334f, 0.552402496f, -1.f),
    C_Normal3f(0.988829851f, -0.149049416f, 0.f),
    C_Vertex3f(0.0922851563f, 0.545898438f, 0.f),
    C_Vertex3f(0.0922851563f, 0.545898438f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.236956179f, -0.971520305f, 0.f),
    C_Vertex3f(0.0922851563f, 0.545898438f, 0.f),
    C_Vertex3f(0.0922851563f, 0.545898438f, -1.f),
    C_Vertex3f(0.0322265625f, 0.560546875f, -1.f),
    C_Vertex3f(0.0322265625f, 0.560546875f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.991027951f, 0.133654535f, 0.f),
    C_Vertex3f(0.0322265625f, 0.560546875f, 0.f),
    C_Vertex3f(0.0322265625f, 0.560546875f, -1.f),
    C_Normal3f(-0.965536296f, 0.26026845f, 0.f),
    C_Vertex3f(0.033536911f, 0.570262909f, 0.f),
    C_Vertex3f(0.033536911f, 0.570262909f, -1.f),
    C_Normal3f(-0.907857716f, 0.419278383f, 0.f),
    C_Vertex3f(0.0396494865f, 0.588084221f, 0.f),
    C_Vertex3f(0.0396494865f, 0.588084221f, -1.f),
    C_Normal3f(-0.805075586f, 0.593172252f, 0.f),
    C_Vertex3f(0.0492539406f, 0.604294777f, 0.f),
    C_Vertex3f(0.0492539406f, 0.604294777f, -1.f),
    C_Normal3f(-0.689954817f, 0.723852456f, 0.f),
    C_Vertex3f(0.0623502731f, 0.618894577f, 0.f),
    C_Vertex3f(0.0623502731f, 0.618894577f, -1.f),
    C_Normal3f(-0.580743492f, 0.814086676f, 0.f),
    C_Vertex3f(0.0747914314f, 0.62863636f, 0.f),
    C_Vertex3f(0.0747914314f, 0.62863636f, -1.f),
    C_Normal3f(-0.480759472f, 0.876852512f, 0.f),
    C_Vertex3f(0.0839723349f, 0.634319067f, 0.f),
    C_Vertex3f(0.0839723349f, 0.634319067f, -1.f),
    C_Normal3f(-0.390974939f, 0.920401216f, 0.f),
    C_Vertex3f(0.0940400362f, 0.639189959f, 0.f),
    C_Vertex3f(0.0940400362f, 0.639189959f, -1.f),
    C_Normal3f(-0.305218399f, 0.952282429f, 0.f),
    C_Vertex3f(0.104994535f, 0.643249035f, 0.f),
    C_Vertex3f(0.104994535f, 0.643249035f, -1.f),
    C_Normal3f(-0.225343153f, 0.974279463f, 0.f),
    C_Vertex3f(0.116835833f, 0.646496296f, 0.f),
    C_Vertex3f(0.116835833f, 0.646496296f, -1.f),
    C_Normal3f(-0.152288407f, 0.988336086f, 0.f),
    C_Vertex3f(0.129563928f, 0.648931742f, 0.f),
    C_Vertex3f(0.129563928f, 0.648931742f, -1.f),
    C_Normal3f(-0.0862964094f, 0.996269464f, 0.f),
    C_Vertex3f(0.143178821f, 0.650555372f, 0.f),
    C_Vertex3f(0.143178821f, 0.650555372f, -1.f),
    C_Normal3f(-0.0260574389f, 0.999660552f, 0.f),
    C_Vertex3f(0.157680511f, 0.651367188f, 0.f),
    C_Vertex3f(0.157680511f, 0.651367188f, -1.f),
    C_Normal3f(0.0256917812f, 0.999669969f, 0.f),
    C_Vertex3f(0.174323082f, 0.651367188f, 0.f),
    C_Vertex3f(0.174323082f, 0.651367188f, -1.f),
    C_Normal3f(0.0801442042f, 0.996783257f, 0.f),
    C_Vertex3f(0.190910339f, 0.650513172f, 0.f),
    C_Vertex3f(0.190910339f, 0.650513172f, -1.f),
    C_Normal3f(0.144472092f, 0.989508808f, 0.f),
    C_Vertex3f(0.206188202f, 0.648805141f, 0.f),
    C_Vertex3f(0.206188202f, 0.648805141f, -1.f),
    C_Normal3f(0.219055668f, 0.975712419f, 0.f),
    C_Vertex3f(0.22015667f, 0.646243095f, 0.f),
    C_Vertex3f(0.22015667f, 0.646243095f, -1.f),
    C_Normal3f(0.304895848f, 0.952385724f, 0.f),
    C_Vertex3f(0.232815742f, 0.642827034f, 0.f),
    C_Vertex3f(0.232815742f, 0.642827034f, -1.f),
    C_Normal3f(0.402114034f, 0.915589631f, 0.f),
    C_Vertex3f(0.244165421f, 0.638556957f, 0.f),
    C_Vertex3f(0.244165421f, 0.638556957f, -1.f),
    C_Normal3f(0.509073973f, 0.860722721f, 0.f),
    C_Vertex3f(0.254205704f, 0.633432865f, 0.f),
    C_Vertex3f(0.254205704f, 0.633432865f, -1.f),
    C_Normal3f(0.602580786f, 0.798057795f, 0.f),
    C_Vertex3f(0.262936592f, 0.627454758f, 0.f),
    C_Vertex3f(0.262936592f, 0.627454758f, -1.f),
    C_Normal3f(0.682924449f, 0.730488956f, 0.f),
    C_Vertex3f(0.266647339f, 0.624038696f, 0.f),
    C_Vertex3f(0.266647339f, 0.624038696f, -1.f),
    C_Normal3f(0.744483113f, 0.667641222f, 0.f),
    C_Vertex3f(0.270303726f, 0.620567322f, 0.f),
    C_Vertex3f(0.270303726f, 0.620567322f, -1.f),
    C_Normal3f(0.807716966f, 0.589570463f, 0.f),
    C_Vertex3f(0.276702404f, 0.612826347f, 0.f),
    C_Vertex3f(0.276702404f, 0.612826347f, -1.f),
    C_Normal3f(0.87158674f, 0.490241379f, 0.f),
    C_Vertex3f(0.282186985f, 0.604287148f, 0.f),
    C_Vertex3f(0.282186985f, 0.604287148f, -1.f),
    C_Normal3f(0.92116636f, 0.389168978f, 0.f),
    C_Vertex3f(0.286757469f, 0.594949722f, 0.f),
    C_Vertex3f(0.286757469f, 0.594949722f, -1.f),
    C_Normal3f(0.956848085f, 0.290588528f, 0.f),
    C_Vertex3f(0.290413857f, 0.584814072f, 0.f),
    C_Vertex3f(0.290413857f, 0.584814072f, -1.f),
    C_Normal3f(0.980269372f, 0.197666556f, 0.f),
    C_Vertex3f(0.293156147f, 0.573880196f, 0.f),
    C_Vertex3f(0.293156147f, 0.573880196f, -1.f),
    C_Normal3f(0.993673086f, 0.11231111f, 0.f),
    C_Vertex3f(0.294984341f, 0.562148094f, 0.f),
    C_Vertex3f(0.294984341f, 0.562148094f, -1.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.295898438f, 0.549617767f, 0.f),
    C_Vertex3f(0.295898438f, 0.549617767f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.295898438f, 0.549617767f, 0.f),
    C_Vertex3f(0.295898438f, 0.549617767f, -1.f),
    C_Vertex3f(0.295898438f, 0.435440063f, -1.f),
    C_Vertex3f(0.295898438f, 0.435440063f, 0.f),
    C_Normal3f(0.999868929f, -0.0161899496f, 0.f),
    C_Vertex3f(0.295898438f, 0.435440063f, 0.f),
    C_Vertex3f(0.295898438f, 0.435440063f, -1.f),
    C_Vertex3f(0.295455933f, 0.408111572f, -1.f),
    C_Vertex3f(0.295455933f, 0.408111572f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999868929f, -0.0161899496f, 0.f),
    C_Vertex3f(0.295455933f, 0.408111572f, 0.f),
    C_Vertex3f(0.295455933f, 0.408111572f, -1.f),
    C_Normal3f(0.99965179f, 0.0263865888f, 0.f),
    C_Vertex3f(0.295455933f, 0.400608063f, 0.f),
    C_Vertex3f(0.295455933f, 0.400608063f, -1.f),
    C_Normal3f(0.997971892f, 0.0636556074f, 0.f),
    C_Vertex3f(0.296019077f, 0.386776924f, 0.f),
    C_Vertex3f(0.296019077f, 0.386776924f, -1.f),
    C_Normal3f(0.993263006f, 0.115881339f, 0.f),
    C_Vertex3f(0.297145367f, 0.374121666f, 0.f),
    C_Vertex3f(0.297145367f, 0.374121666f, -1.f),
    C_Normal3f(0.984016657f, 0.178075954f, 0.f),
    C_Vertex3f(0.298834801f, 0.362642288f, 0.f),
    C_Vertex3f(0.298834801f, 0.362642288f, -1.f),
    C_Normal3f(0.967625499f, 0.252389967f, 0.f),
    C_Vertex3f(0.301087379f, 0.352338791f, 0.f),
    C_Vertex3f(0.301087379f, 0.352338791f, -1.f),
    C_Normal3f(0.940077424f, 0.340961128f, 0.f),
    C_Vertex3f(0.303903103f, 0.343211174f, 0.f),
    C_Vertex3f(0.303903103f, 0.343211174f, -1.f),
    C_Normal3f(0.895468533f, 0.445124924f, 0.f),
    C_Vertex3f(0.307281971f, 0.335259438f, 0.f),
    C_Vertex3f(0.307281971f, 0.335259438f, -1.f),
    C_Normal3f(0.839631855f, 0.543155909f, 0.f),
    C_Vertex3f(0.311223984f, 0.328483582f, 0.f),
    C_Vertex3f(0.311223984f, 0.328483582f, -1.f),
    C_Normal3f(0.779157102f, 0.62682873f, 0.f),
    C_Vertex3f(0.313476563f, 0.325683594f, 0.f),
    C_Vertex3f(0.313476563f, 0.325683594f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.313476563f, 0.325683594f, 0.f),
    C_Vertex3f(0.313476563f, 0.325683594f, -1.f),
    C_Vertex3f(0.249023438f, 0.325683594f, -1.f),
    C_Vertex3f(0.249023438f, 0.325683594f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.886633098f, -0.462473392f, 0.f),
    C_Vertex3f(0.249023438f, 0.325683594f, 0.f),
    C_Vertex3f(0.249023438f, 0.325683594f, -1.f),
    C_Normal3f(-0.919548094f, -0.392977536f, 0.f),
    C_Vertex3f(0.246744156f, 0.330053329f, 0.f),
    C_Vertex3f(0.246744156f, 0.330053329f, -1.f),
    C_Normal3f(-0.952775836f, -0.303674519f, 0.f),
    C_Vertex3f(0.243210793f, 0.339284897f, 0.f),
    C_Vertex3f(0.243210793f, 0.339284897f, -1.f),
    C_Normal3f(-0.980549991f, -0.196269453f, 0.f),
    C_Vertex3f(0.240702629f, 0.34900856f, 0.f),
    C_Vertex3f(0.240702629f, 0.34900856f, -1.f),
    C_Normal3f(-0.994010091f, -0.109288752f, 0.f),
    C_Vertex3f(0.239219666f, 0.359224319f, 0.f),
    C_Vertex3f(0.239219666f, 0.359224319f, -1.f),
    C_Normal3f(-0.999087453f, -0.0427112542f, 0.f),
    C_Vertex3f(0.238990784f, 0.364578247f, 0.f),
    C_Vertex3f(0.238990784f, 0.364578247f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.255875021f, -0.966709793f, 0.f),
    C_Vertex3f(0.232421875f, 0.482009888f, 0.f),
    C_Vertex3f(0.232421875f, 0.482009888f, -1.f),
    C_Normal3f(0.237309963f, -0.971433997f, 0.f),
    C_Vertex3f(0.229532242f, 0.481245041f, 0.f),
    C_Vertex3f(0.229532242f, 0.481245041f, -1.f),
    C_Normal3f(0.219267994f, -0.975664675f, 0.f),
    C_Vertex3f(0.222581148f, 0.479605913f, 0.f),
    C_Vertex3f(0.222581148f, 0.479605913f, -1.f),
    C_Normal3f(0.202762067f, -0.97922802f, 0.f),
    C_Vertex3f(0.214458227f, 0.477857351f, 0.f),
    C_Vertex3f(0.214458227f, 0.477857351f, -1.f),
    C_Normal3f(0.19005321f, -0.981773794f, 0.f),
    C_Vertex3f(0.205163479f, 0.475999355f, 0.f),
    C_Vertex3f(0.205163479f, 0.475999355f, -1.f),
    C_Normal3f(0.179971159f, -0.983671904f, 0.f),
    C_Vertex3f(0.194696903f, 0.474031925f, 0.f),
    C_Vertex3f(0.194696903f, 0.474031925f, -1.f),
    C_Normal3f(0.171780229f, -0.985135257f, 0.f),
    C_Vertex3f(0.1830585f, 0.471955061f, 0.f),
    C_Vertex3f(0.1830585f, 0.471955061f, -1.f),
    C_Normal3f(0.164995313f, -0.986294389f, 0.f),
    C_Vertex3f(0.17024827f, 0.469768763f, 0.f),
    C_Vertex3f(0.17024827f, 0.469768763f, -1.f),
    C_Normal3f(0.158308029f, -0.987389803f, 0.f),
    C_Vertex3f(0.156266212f, 0.46747303f, 0.f),
    C_Vertex3f(0.156266212f, 0.46747303f, -1.f),
    C_Normal3f(0.187159553f, -0.982329547f, 0.f),
    C_Vertex3f(0.138875961f, 0.464738846f, 0.f),
    C_Vertex3f(0.138875961f, 0.464738846f, -1.f),
    C_Normal3f(0.264171153f, -0.964475751f, 0.f),
    C_Vertex3f(0.122684479f, 0.461074829f, 0.f),
    C_Vertex3f(0.122684479f, 0.461074829f, -1.f),
    C_Normal3f(0.382170767f, -0.924091637f, 0.f),
    C_Vertex3f(0.109928131f, 0.456809998f, 0.f),
    C_Vertex3f(0.109928131f, 0.456809998f, -1.f),
    C_Normal3f(0.563218713f, -0.826307893f, 0.f),
    C_Vertex3f(0.100606918f, 0.451944351f, 0.f),
    C_Vertex3f(0.100606918f, 0.451944351f, -1.f),
    C_Normal3f(0.747848451f, -0.663869619f, 0.f),
    C_Vertex3f(0.0946102142f, 0.446369171f, 0.f),
    C_Vertex3f(0.0946102142f, 0.446369171f, -1.f),
    C_Normal3f(0.867864788f, -0.496800452f, 0.f),
    C_Vertex3f(0.0900297165f, 0.440029144f, 0.f),
    C_Vertex3f(0.0900297165f, 0.440029144f, -1.f),
    C_Normal3f(0.954426885f, -0.298444808f, 0.f),
    C_Vertex3f(0.0869760513f, 0.43303299f, 0.f),
    C_Vertex3f(0.0869760513f, 0.43303299f, -1.f),
    C_Normal3f(0.995635569f, -0.0933262631f, 0.f),
    C_Vertex3f(0.0854492188f, 0.425380707f, 0.f),
    C_Vertex3f(0.0854492188f, 0.425380707f, -1.f),
    C_Normal3f(0.994065225f, 0.108785763f, 0.f),
    C_Vertex3f(0.0854492188f, 0.416744232f, 0.f),
    C_Vertex3f(0.0854492188f, 0.416744232f, -1.f),
    C_Normal3f(0.944290757f, 0.32911247f, 0.f),
    C_Vertex3f(0.0873036385f, 0.408435345f, 0.f),
    C_Vertex3f(0.0873036385f, 0.408435345f, -1.f),
    C_Normal3f(0.844995141f, 0.534773946f, 0.f),
    C_Vertex3f(0.0910124779f, 0.400782108f, 0.f),
    C_Vertex3f(0.0910124779f, 0.400782108f, -1.f),
    C_Normal3f(0.716673672f, 0.697408676f, 0.f),
    C_Vertex3f(0.096575737f, 0.393784523f, 0.f),
    C_Vertex3f(0.096575737f, 0.393784523f, -1.f),
    C_Normal3f(0.561333358f, 0.82758981f, 0.f),
    C_Vertex3f(0.103993416f, 0.387442589f, 0.f),
    C_Vertex3f(0.103993416f, 0.387442589f, -1.f),
    C_Normal3f(0.377978534f, 0.925814331f, 0.f),
    C_Vertex3f(0.112938404f, 0.382686138f, 0.f),
    C_Vertex3f(0.112938404f, 0.382686138f, -1.f),
    C_Normal3f(0.207074389f, 0.978325248f, 0.f),
    C_Vertex3f(0.123410702f, 0.379515171f, 0.f),
    C_Vertex3f(0.123410702f, 0.379515171f, -1.f),
    C_Normal3f(0.0592189804f, 0.998245001f, 0.f),
    C_Vertex3f(0.135410309f, 0.377929688f, 0.f),
    C_Vertex3f(0.135410309f, 0.377929688f, -1.f),
    C_Normal3f(-0.0717982575f, 0.997419238f, 0.f),
    C_Vertex3f(0.150136948f, 0.377929688f, 0.f),
    C_Vertex3f(0.150136948f, 0.377929688f, -1.f),
    C_Normal3f(-0.217196196f, 0.976127923f, 0.f),
    C_Vertex3f(0.165027142f, 0.380061626f, 0.f),
    C_Vertex3f(0.165027142f, 0.380061626f, -1.f),
    C_Normal3f(-0.371118754f, 0.92858547f, 0.f),
    C_Vertex3f(0.178881168f, 0.384325504f, 0.f),
    C_Vertex3f(0.178881168f, 0.384325504f, -1.f),
    C_Normal3f(-0.518676639f, 0.854970396f, 0.f),
    C_Vertex3f(0.191699028f, 0.390721321f, 0.f),
    C_Vertex3f(0.191699028f, 0.390721321f, -1.f),
    C_Normal3f(-0.641018987f, 0.767524958f, 0.f),
    C_Vertex3f(0.203480721f, 0.399249077f, 0.f),
    C_Vertex3f(0.203480721f, 0.399249077f, -1.f),
    C_Normal3f(-0.749500036f, 0.662004471f, 0.f),
    C_Vertex3f(0.213298321f, 0.408760548f, 0.f),
    C_Vertex3f(0.213298321f, 0.408760548f, -1.f),
    C_Normal3f(-0.84784013f, 0.53025192f, 0.f),
    C_Vertex3f(0.221151829f, 0.419255733f, 0.f),
    C_Vertex3f(0.221151829f, 0.419255733f, -1.f),
    C_Normal3f(-0.914177597f, 0.405313879f, 0.f),
    C_Vertex3f(0.227041245f, 0.430734634f, 0.f),
    C_Vertex3f(0.227041245f, 0.430734634f, -1.f),
    C_Normal3f(-0.951889455f, 0.306441963f, 0.f),
    C_Vertex3f(0.229003906f, 0.436965942f, 0.f),
    C_Vertex3f(0.229003906f, 0.436965942f, -1.f),
    C_Normal3f(-0.982035875f, 0.188694656f, 0.f),
    C_Vertex3f(0.229751587f, 0.439153671f, 0.f),
    C_Vertex3f(0.229751587f, 0.439153671f, -1.f),
    C_Normal3f(-0.994390488f, 0.105771415f, 0.f),
    C_Vertex3f(0.230926514f, 0.446971893f, 0.f),
    C_Vertex3f(0.230926514f, 0.446971893f, -1.f),
    C_Normal3f(-0.998573065f, 0.0534019805f, 0.f),
    C_Vertex3f(0.231781006f, 0.45823288f, 0.f),
    C_Vertex3f(0.231781006f, 0.45823288f, -1.f),
    C_Normal3f(-0.999636948f, 0.0269435272f, 0.f),
    C_Vertex3f(0.232315063f, 0.47293663f, 0.f),
    C_Vertex3f(0.232315063f, 0.47293663f, -1.f),
    C_Normal3f(-0.999930739f, 0.0117713101f, 0.f),
    C_Vertex3f(0.232421875f, 0.482009888f, 0.f),
    C_Vertex3f(0.232421875f, 0.482009888f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.222581148f, 0.479605913f, 0.f),
    C_Vertex3f(0.229532242f, 0.481245041f, 0.f),
    C_Vertex3f(0.230291367f, 0.532531738f, 0.f),
    C_Vertex3f(0.232421875f, 0.482009888f, 0.f),
    C_Vertex3f(0.232910156f, 0.533187866f, 0.f),
    C_Vertex3f(0.238990784f, 0.364578247f, 0.f),
    C_Vertex3f(0.244165421f, 0.638556957f, 0.f),
    C_Vertex3f(0.239219666f, 0.359224319f, 0.f),
    C_Vertex3f(0.240702629f, 0.34900856f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.17024827f, 0.469768763f, 0.f),
    C_Vertex3f(0.1830585f, 0.471955061f, 0.f),
    C_Vertex3f(0.19102478f, 0.524436951f, 0.f),
    C_Vertex3f(0.194696903f, 0.474031925f, 0.f),
    C_Vertex3f(0.198877335f, 0.525968552f, 0.f),
    C_Vertex3f(0.205163479f, 0.475999355f, 0.f),
    C_Vertex3f(0.211965561f, 0.528594017f, 0.f),
    C_Vertex3f(0.214458227f, 0.477857351f, 0.f),
    C_Vertex3f(0.222436905f, 0.530781746f, 0.f),
    C_Vertex3f(0.222581148f, 0.479605913f, 0.f),
    C_Vertex3f(0.230291367f, 0.532531738f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.19102478f, 0.524436951f, 0.f),
    C_Vertex3f(0.119909286f, 0.513723373f, 0.f),
    C_Vertex3f(0.122684479f, 0.461074829f, 0.f),
    C_Vertex3f(0.138875961f, 0.464738846f, 0.f),
    C_Vertex3f(0.156266212f, 0.46747303f, 0.f),
    C_Vertex3f(0.17024827f, 0.469768763f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0900297165f, 0.440029144f, 0.f),
    C_Vertex3f(0.0946102142f, 0.446369171f, 0.f),
    C_Vertex3f(0.0967493057f, 0.509191513f, 0.f),
    C_Vertex3f(0.100606918f, 0.451944351f, 0.f),
    C_Vertex3f(0.107539177f, 0.511484146f, 0.f),
    C_Vertex3f(0.109928131f, 0.456809998f, 0.f),
    C_Vertex3f(0.119909286f, 0.513723373f, 0.f),
    C_Vertex3f(0.122684479f, 0.461074829f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0875396729f, 0.506845474f, 0.f),
    C_Vertex3f(0.0837249756f, 0.505645752f, 0.f),
    C_Vertex3f(0.0854492188f, 0.425380707f, 0.f),
    C_Vertex3f(0.0869760513f, 0.43303299f, 0.f),
    C_Vertex3f(0.0900297165f, 0.440029144f, 0.f),
    C_Vertex3f(0.0967493057f, 0.509191513f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0339993238f, 0.370748758f, 0.f),
    C_Vertex3f(0.0393959284f, 0.36309123f, 0.f),
    C_Vertex3f(0.0413284302f, 0.479869843f, 0.f),
    C_Vertex3f(0.045691967f, 0.355904579f, 0.f),
    C_Vertex3f(0.0516281128f, 0.489042282f, 0.f),
    C_Vertex3f(0.0528874397f, 0.349188805f, 0.f),
    C_Vertex3f(0.0595026016f, 0.494557381f, 0.f),
    C_Vertex3f(0.060723424f, 0.343312502f, 0.f),
    C_Vertex3f(0.0683851242f, 0.499472618f, 0.f),
    C_Vertex3f(0.0691999197f, 0.338275671f, 0.f),
    C_Vertex3f(0.0782756805f, 0.503787994f, 0.f),
    C_Vertex3f(0.078316927f, 0.334078312f, 0.f),
    C_Vertex3f(0.0837249756f, 0.505645752f, 0.f),
    C_Vertex3f(0.0854492188f, 0.416744232f, 0.f),
    C_Vertex3f(0.0854492188f, 0.425380707f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.295455933f, 0.400608063f, 0.f),
    C_Vertex3f(0.295455933f, 0.408111572f, 0.f),
    C_Vertex3f(0.294984341f, 0.562148094f, 0.f),
    C_Vertex3f(0.293156147f, 0.573880196f, 0.f),
    C_Vertex3f(0.290413857f, 0.584814072f, 0.f),
    C_Vertex3f(0.286757469f, 0.594949722f, 0.f),
    C_Vertex3f(0.296019077f, 0.386776924f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.296019077f, 0.386776924f, 0.f),
    C_Vertex3f(0.286757469f, 0.594949722f, 0.f),
    C_Vertex3f(0.282186985f, 0.604287148f, 0.f),
    C_Vertex3f(0.276702404f, 0.612826347f, 0.f),
    C_Vertex3f(0.270303726f, 0.620567322f, 0.f),
    C_Vertex3f(0.297145367f, 0.374121666f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.297145367f, 0.374121666f, 0.f),
    C_Vertex3f(0.270303726f, 0.620567322f, 0.f),
    C_Vertex3f(0.266647339f, 0.624038696f, 0.f),
    C_Vertex3f(0.262936592f, 0.627454758f, 0.f),
    C_Vertex3f(0.254205704f, 0.633432865f, 0.f),
    C_Vertex3f(0.298834801f, 0.362642288f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.249023438f, 0.325683594f, 0.f),
    C_Vertex3f(0.313476563f, 0.325683594f, 0.f),
    C_Vertex3f(0.311223984f, 0.328483582f, 0.f),
    C_Vertex3f(0.307281971f, 0.335259438f, 0.f),
    C_Vertex3f(0.303903103f, 0.343211174f, 0.f),
    C_Vertex3f(0.301087379f, 0.352338791f, 0.f),
    C_Vertex3f(0.298834801f, 0.362642288f, 0.f),
    C_Vertex3f(0.254205704f, 0.633432865f, 0.f),
    C_Vertex3f(0.246744156f, 0.330053329f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.244165421f, 0.638556957f, 0.f),
    C_Vertex3f(0.240702629f, 0.34900856f, 0.f),
    C_Vertex3f(0.243210793f, 0.339284897f, 0.f),
    C_Vertex3f(0.246744156f, 0.330053329f, 0.f),
    C_Vertex3f(0.254205704f, 0.633432865f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.232815742f, 0.642827034f, 0.f),
    C_Vertex3f(0.22015667f, 0.646243095f, 0.f),
    C_Vertex3f(0.221344471f, 0.581812859f, 0.f),
    C_Vertex3f(0.225399017f, 0.576919556f, 0.f),
    C_Vertex3f(0.228229523f, 0.571670532f, 0.f),
    C_Vertex3f(0.230406761f, 0.563907623f, 0.f),
    C_Vertex3f(0.231930733f, 0.553630829f, 0.f),
    C_Vertex3f(0.232801437f, 0.540840149f, 0.f),
    C_Vertex3f(0.232910156f, 0.533187866f, 0.f),
    C_Vertex3f(0.244165421f, 0.638556957f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.192830563f, 0.59722805f, 0.f),
    C_Vertex3f(0.201836586f, 0.594358444f, 0.f),
    C_Vertex3f(0.206188202f, 0.648805141f, 0.f),
    C_Vertex3f(0.209563255f, 0.590532303f, 0.f),
    C_Vertex3f(0.22015667f, 0.646243095f, 0.f),
    C_Vertex3f(0.216065884f, 0.586350441f, 0.f),
    C_Vertex3f(0.221344471f, 0.581812859f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.117430687f, 0.589529037f, 0.f),
    C_Vertex3f(0.128097057f, 0.594813347f, 0.f),
    C_Vertex3f(0.129563928f, 0.648931742f, 0.f),
    C_Vertex3f(0.140940189f, 0.59833622f, 0.f),
    C_Vertex3f(0.143178821f, 0.650555372f, 0.f),
    C_Vertex3f(0.155960083f, 0.600097656f, 0.f),
    C_Vertex3f(0.157680511f, 0.651367188f, 0.f),
    C_Vertex3f(0.170980453f, 0.600097656f, 0.f),
    C_Vertex3f(0.174323082f, 0.651367188f, 0.f),
    C_Vertex3f(0.182545185f, 0.599141121f, 0.f),
    C_Vertex3f(0.190910339f, 0.650513172f, 0.f),
    C_Vertex3f(0.192830563f, 0.59722805f, 0.f),
    C_Vertex3f(0.206188202f, 0.648805141f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0968585014f, 0.563920021f, 0.f),
    C_Vertex3f(0.102083683f, 0.573946953f, 0.f),
    C_Vertex3f(0.104994535f, 0.643249035f, 0.f),
    C_Vertex3f(0.108941078f, 0.582483292f, 0.f),
    C_Vertex3f(0.116835833f, 0.646496296f, 0.f),
    C_Vertex3f(0.117430687f, 0.589529037f, 0.f),
    C_Vertex3f(0.129563928f, 0.648931742f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0940400362f, 0.639189959f, 0.f),
    C_Vertex3f(0.0839723349f, 0.634319067f, 0.f),
    C_Vertex3f(0.0922851563f, 0.545898438f, 0.f),
    C_Vertex3f(0.0932655334f, 0.552402496f, 0.f),
    C_Vertex3f(0.0968585014f, 0.563920021f, 0.f),
    C_Vertex3f(0.104994535f, 0.643249035f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0922851563f, 0.545898438f, 0.f),
    C_Vertex3f(0.0839723349f, 0.634319067f, 0.f),
    C_Vertex3f(0.0747914314f, 0.62863636f, 0.f),
    C_Vertex3f(0.0623502731f, 0.618894577f, 0.f),
    C_Vertex3f(0.0492539406f, 0.604294777f, 0.f),
    C_Vertex3f(0.0396494865f, 0.588084221f, 0.f),
    C_Vertex3f(0.033536911f, 0.570262909f, 0.f),
    C_Vertex3f(0.0322265625f, 0.560546875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.295898438f, 0.549617767f, 0.f),
    C_Vertex3f(0.294984341f, 0.562148094f, 0.f),
    C_Vertex3f(0.295455933f, 0.408111572f, 0.f),
    C_Vertex3f(0.295898438f, 0.435440063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.232340813f, 0.359716415f, 0.f),
    C_Vertex3f(0.238990784f, 0.364578247f, 0.f),
    C_Vertex3f(0.232421875f, 0.482009888f, 0.f),
    C_Vertex3f(0.232315063f, 0.47293663f, 0.f),
    C_Vertex3f(0.231781006f, 0.45823288f, 0.f),
    C_Vertex3f(0.230926514f, 0.446971893f, 0.f),
    C_Vertex3f(0.229751587f, 0.439153671f, 0.f),
    C_Vertex3f(0.229003906f, 0.436965942f, 0.f),
    C_Vertex3f(0.227041245f, 0.430734634f, 0.f),
    C_Vertex3f(0.221151829f, 0.419255733f, 0.f),
    C_Vertex3f(0.218877077f, 0.35120821f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213298321f, 0.408760548f, 0.f),
    C_Vertex3f(0.203480721f, 0.399249077f, 0.f),
    C_Vertex3f(0.205249548f, 0.343915462f, 0.f),
    C_Vertex3f(0.218877077f, 0.35120821f, 0.f),
    C_Vertex3f(0.221151829f, 0.419255733f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.096575737f, 0.393784523f, 0.f),
    C_Vertex3f(0.098472476f, 0.328202009f, 0.f),
    C_Vertex3f(0.103993416f, 0.387442589f, 0.f),
    C_Vertex3f(0.109511018f, 0.326523066f, 0.f),
    C_Vertex3f(0.112938404f, 0.382686138f, 0.f),
    C_Vertex3f(0.121190071f, 0.325683594f, 0.f),
    C_Vertex3f(0.123410702f, 0.379515171f, 0.f),
    C_Vertex3f(0.134654999f, 0.325683594f, 0.f),
    C_Vertex3f(0.135410309f, 0.377929688f, 0.f),
    C_Vertex3f(0.149101496f, 0.326899052f, 0.f),
    C_Vertex3f(0.150136948f, 0.377929688f, 0.f),
    C_Vertex3f(0.163384199f, 0.329329967f, 0.f),
    C_Vertex3f(0.165027142f, 0.380061626f, 0.f),
    C_Vertex3f(0.177503109f, 0.332976341f, 0.f),
    C_Vertex3f(0.178881168f, 0.384325504f, 0.f),
    C_Vertex3f(0.191458225f, 0.337838173f, 0.f),
    C_Vertex3f(0.191699028f, 0.390721321f, 0.f),
    C_Vertex3f(0.205249548f, 0.343915462f, 0.f),
    C_Vertex3f(0.203480721f, 0.399249077f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0910124779f, 0.400782108f, 0.f),
    C_Vertex3f(0.0873036385f, 0.408435345f, 0.f),
    C_Vertex3f(0.0880744457f, 0.330720425f, 0.f),
    C_Vertex3f(0.098472476f, 0.328202009f, 0.f),
    C_Vertex3f(0.096575737f, 0.393784523f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0854492188f, 0.416744232f, 0.f),
    C_Vertex3f(0.078316927f, 0.334078312f, 0.f),
    C_Vertex3f(0.0880744457f, 0.330720425f, 0.f),
    C_Vertex3f(0.0873036385f, 0.408435345f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0309734344f, 0.465892792f, 0.f),
    C_Vertex3f(0.0257406235f, 0.454099655f, 0.f),
    C_Vertex3f(0.025904417f, 0.387476444f, 0.f),
    C_Vertex3f(0.0295021534f, 0.378877163f, 0.f),
    C_Vertex3f(0.0339993238f, 0.370748758f, 0.f),
    C_Vertex3f(0.0413284302f, 0.479869843f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.025904417f, 0.387476444f, 0.f),
    C_Vertex3f(0.0257406235f, 0.454099655f, 0.f),
    C_Vertex3f(0.0232061148f, 0.396546602f, 0.f),
    C_Vertex3f(0.023124218f, 0.44558239f, 0.f),
    C_Vertex3f(0.0214072466f, 0.406087637f, 0.f),
    C_Vertex3f(0.0213799477f, 0.43640995f, 0.f),
    C_Vertex3f(0.0205078125f, 0.426582336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0214072466f, 0.406087637f, 0.f),
    C_Vertex3f(0.0205078125f, 0.426582336f, 0.f),
    C_Vertex3f(0.0205078125f, 0.416099548f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.222581148f, 0.479605913f, -1.f),
    C_Vertex3f(0.229532242f, 0.481245041f, -1.f),
    C_Vertex3f(0.230291367f, 0.532531738f, -1.f),
    C_Vertex3f(0.232421875f, 0.482009888f, -1.f),
    C_Vertex3f(0.232910156f, 0.533187866f, -1.f),
    C_Vertex3f(0.238990784f, 0.364578247f, -1.f),
    C_Vertex3f(0.244165421f, 0.638556957f, -1.f),
    C_Vertex3f(0.239219666f, 0.359224319f, -1.f),
    C_Vertex3f(0.240702629f, 0.34900856f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.17024827f, 0.469768763f, -1.f),
    C_Vertex3f(0.1830585f, 0.471955061f, -1.f),
    C_Vertex3f(0.19102478f, 0.524436951f, -1.f),
    C_Vertex3f(0.194696903f, 0.474031925f, -1.f),
    C_Vertex3f(0.198877335f, 0.525968552f, -1.f),
    C_Vertex3f(0.205163479f, 0.475999355f, -1.f),
    C_Vertex3f(0.211965561f, 0.528594017f, -1.f),
    C_Vertex3f(0.214458227f, 0.477857351f, -1.f),
    C_Vertex3f(0.222436905f, 0.530781746f, -1.f),
    C_Vertex3f(0.222581148f, 0.479605913f, -1.f),
    C_Vertex3f(0.230291367f, 0.532531738f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.19102478f, 0.524436951f, -1.f),
    C_Vertex3f(0.119909286f, 0.513723373f, -1.f),
    C_Vertex3f(0.122684479f, 0.461074829f, -1.f),
    C_Vertex3f(0.138875961f, 0.464738846f, -1.f),
    C_Vertex3f(0.156266212f, 0.46747303f, -1.f),
    C_Vertex3f(0.17024827f, 0.469768763f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0900297165f, 0.440029144f, -1.f),
    C_Vertex3f(0.0946102142f, 0.446369171f, -1.f),
    C_Vertex3f(0.0967493057f, 0.509191513f, -1.f),
    C_Vertex3f(0.100606918f, 0.451944351f, -1.f),
    C_Vertex3f(0.107539177f, 0.511484146f, -1.f),
    C_Vertex3f(0.109928131f, 0.456809998f, -1.f),
    C_Vertex3f(0.119909286f, 0.513723373f, -1.f),
    C_Vertex3f(0.122684479f, 0.461074829f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0875396729f, 0.506845474f, -1.f),
    C_Vertex3f(0.0837249756f, 0.505645752f, -1.f),
    C_Vertex3f(0.0854492188f, 0.425380707f, -1.f),
    C_Vertex3f(0.0869760513f, 0.43303299f, -1.f),
    C_Vertex3f(0.0900297165f, 0.440029144f, -1.f),
    C_Vertex3f(0.0967493057f, 0.509191513f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0339993238f, 0.370748758f, -1.f),
    C_Vertex3f(0.0393959284f, 0.36309123f, -1.f),
    C_Vertex3f(0.0413284302f, 0.479869843f, -1.f),
    C_Vertex3f(0.045691967f, 0.355904579f, -1.f),
    C_Vertex3f(0.0516281128f, 0.489042282f, -1.f),
    C_Vertex3f(0.0528874397f, 0.349188805f, -1.f),
    C_Vertex3f(0.0595026016f, 0.494557381f, -1.f),
    C_Vertex3f(0.060723424f, 0.343312502f, -1.f),
    C_Vertex3f(0.0683851242f, 0.499472618f, -1.f),
    C_Vertex3f(0.0691999197f, 0.338275671f, -1.f),
    C_Vertex3f(0.0782756805f, 0.503787994f, -1.f),
    C_Vertex3f(0.078316927f, 0.334078312f, -1.f),
    C_Vertex3f(0.0837249756f, 0.505645752f, -1.f),
    C_Vertex3f(0.0854492188f, 0.416744232f, -1.f),
    C_Vertex3f(0.0854492188f, 0.425380707f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.295455933f, 0.400608063f, -1.f),
    C_Vertex3f(0.295455933f, 0.408111572f, -1.f),
    C_Vertex3f(0.294984341f, 0.562148094f, -1.f),
    C_Vertex3f(0.293156147f, 0.573880196f, -1.f),
    C_Vertex3f(0.290413857f, 0.584814072f, -1.f),
    C_Vertex3f(0.286757469f, 0.594949722f, -1.f),
    C_Vertex3f(0.296019077f, 0.386776924f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.296019077f, 0.386776924f, -1.f),
    C_Vertex3f(0.286757469f, 0.594949722f, -1.f),
    C_Vertex3f(0.282186985f, 0.604287148f, -1.f),
    C_Vertex3f(0.276702404f, 0.612826347f, -1.f),
    C_Vertex3f(0.270303726f, 0.620567322f, -1.f),
    C_Vertex3f(0.297145367f, 0.374121666f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.297145367f, 0.374121666f, -1.f),
    C_Vertex3f(0.270303726f, 0.620567322f, -1.f),
    C_Vertex3f(0.266647339f, 0.624038696f, -1.f),
    C_Vertex3f(0.262936592f, 0.627454758f, -1.f),
    C_Vertex3f(0.254205704f, 0.633432865f, -1.f),
    C_Vertex3f(0.298834801f, 0.362642288f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.249023438f, 0.325683594f, -1.f),
    C_Vertex3f(0.313476563f, 0.325683594f, -1.f),
    C_Vertex3f(0.311223984f, 0.328483582f, -1.f),
    C_Vertex3f(0.307281971f, 0.335259438f, -1.f),
    C_Vertex3f(0.303903103f, 0.343211174f, -1.f),
    C_Vertex3f(0.301087379f, 0.352338791f, -1.f),
    C_Vertex3f(0.298834801f, 0.362642288f, -1.f),
    C_Vertex3f(0.254205704f, 0.633432865f, -1.f),
    C_Vertex3f(0.246744156f, 0.330053329f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.244165421f, 0.638556957f, -1.f),
    C_Vertex3f(0.240702629f, 0.34900856f, -1.f),
    C_Vertex3f(0.243210793f, 0.339284897f, -1.f),
    C_Vertex3f(0.246744156f, 0.330053329f, -1.f),
    C_Vertex3f(0.254205704f, 0.633432865f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.232815742f, 0.642827034f, -1.f),
    C_Vertex3f(0.22015667f, 0.646243095f, -1.f),
    C_Vertex3f(0.221344471f, 0.581812859f, -1.f),
    C_Vertex3f(0.225399017f, 0.576919556f, -1.f),
    C_Vertex3f(0.228229523f, 0.571670532f, -1.f),
    C_Vertex3f(0.230406761f, 0.563907623f, -1.f),
    C_Vertex3f(0.231930733f, 0.553630829f, -1.f),
    C_Vertex3f(0.232801437f, 0.540840149f, -1.f),
    C_Vertex3f(0.232910156f, 0.533187866f, -1.f),
    C_Vertex3f(0.244165421f, 0.638556957f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.192830563f, 0.59722805f, -1.f),
    C_Vertex3f(0.201836586f, 0.594358444f, -1.f),
    C_Vertex3f(0.206188202f, 0.648805141f, -1.f),
    C_Vertex3f(0.209563255f, 0.590532303f, -1.f),
    C_Vertex3f(0.22015667f, 0.646243095f, -1.f),
    C_Vertex3f(0.216065884f, 0.586350441f, -1.f),
    C_Vertex3f(0.221344471f, 0.581812859f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.117430687f, 0.589529037f, -1.f),
    C_Vertex3f(0.128097057f, 0.594813347f, -1.f),
    C_Vertex3f(0.129563928f, 0.648931742f, -1.f),
    C_Vertex3f(0.140940189f, 0.59833622f, -1.f),
    C_Vertex3f(0.143178821f, 0.650555372f, -1.f),
    C_Vertex3f(0.155960083f, 0.600097656f, -1.f),
    C_Vertex3f(0.157680511f, 0.651367188f, -1.f),
    C_Vertex3f(0.170980453f, 0.600097656f, -1.f),
    C_Vertex3f(0.174323082f, 0.651367188f, -1.f),
    C_Vertex3f(0.182545185f, 0.599141121f, -1.f),
    C_Vertex3f(0.190910339f, 0.650513172f, -1.f),
    C_Vertex3f(0.192830563f, 0.59722805f, -1.f),
    C_Vertex3f(0.206188202f, 0.648805141f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0968585014f, 0.563920021f, -1.f),
    C_Vertex3f(0.102083683f, 0.573946953f, -1.f),
    C_Vertex3f(0.104994535f, 0.643249035f, -1.f),
    C_Vertex3f(0.108941078f, 0.582483292f, -1.f),
    C_Vertex3f(0.116835833f, 0.646496296f, -1.f),
    C_Vertex3f(0.117430687f, 0.589529037f, -1.f),
    C_Vertex3f(0.129563928f, 0.648931742f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0940400362f, 0.639189959f, -1.f),
    C_Vertex3f(0.0839723349f, 0.634319067f, -1.f),
    C_Vertex3f(0.0922851563f, 0.545898438f, -1.f),
    C_Vertex3f(0.0932655334f, 0.552402496f, -1.f),
    C_Vertex3f(0.0968585014f, 0.563920021f, -1.f),
    C_Vertex3f(0.104994535f, 0.643249035f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0922851563f, 0.545898438f, -1.f),
    C_Vertex3f(0.0839723349f, 0.634319067f, -1.f),
    C_Vertex3f(0.0747914314f, 0.62863636f, -1.f),
    C_Vertex3f(0.0623502731f, 0.618894577f, -1.f),
    C_Vertex3f(0.0492539406f, 0.604294777f, -1.f),
    C_Vertex3f(0.0396494865f, 0.588084221f, -1.f),
    C_Vertex3f(0.033536911f, 0.570262909f, -1.f),
    C_Vertex3f(0.0322265625f, 0.560546875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.295898438f, 0.549617767f, -1.f),
    C_Vertex3f(0.294984341f, 0.562148094f, -1.f),
    C_Vertex3f(0.295455933f, 0.408111572f, -1.f),
    C_Vertex3f(0.295898438f, 0.435440063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.232340813f, 0.359716415f, -1.f),
    C_Vertex3f(0.238990784f, 0.364578247f, -1.f),
    C_Vertex3f(0.232421875f, 0.482009888f, -1.f),
    C_Vertex3f(0.232315063f, 0.47293663f, -1.f),
    C_Vertex3f(0.231781006f, 0.45823288f, -1.f),
    C_Vertex3f(0.230926514f, 0.446971893f, -1.f),
    C_Vertex3f(0.229751587f, 0.439153671f, -1.f),
    C_Vertex3f(0.229003906f, 0.436965942f, -1.f),
    C_Vertex3f(0.227041245f, 0.430734634f, -1.f),
    C_Vertex3f(0.221151829f, 0.419255733f, -1.f),
    C_Vertex3f(0.218877077f, 0.35120821f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213298321f, 0.408760548f, -1.f),
    C_Vertex3f(0.203480721f, 0.399249077f, -1.f),
    C_Vertex3f(0.205249548f, 0.343915462f, -1.f),
    C_Vertex3f(0.218877077f, 0.35120821f, -1.f),
    C_Vertex3f(0.221151829f, 0.419255733f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.096575737f, 0.393784523f, -1.f),
    C_Vertex3f(0.098472476f, 0.328202009f, -1.f),
    C_Vertex3f(0.103993416f, 0.387442589f, -1.f),
    C_Vertex3f(0.109511018f, 0.326523066f, -1.f),
    C_Vertex3f(0.112938404f, 0.382686138f, -1.f),
    C_Vertex3f(0.121190071f, 0.325683594f, -1.f),
    C_Vertex3f(0.123410702f, 0.379515171f, -1.f),
    C_Vertex3f(0.134654999f, 0.325683594f, -1.f),
    C_Vertex3f(0.135410309f, 0.377929688f, -1.f),
    C_Vertex3f(0.149101496f, 0.326899052f, -1.f),
    C_Vertex3f(0.150136948f, 0.377929688f, -1.f),
    C_Vertex3f(0.163384199f, 0.329329967f, -1.f),
    C_Vertex3f(0.165027142f, 0.380061626f, -1.f),
    C_Vertex3f(0.177503109f, 0.332976341f, -1.f),
    C_Vertex3f(0.178881168f, 0.384325504f, -1.f),
    C_Vertex3f(0.191458225f, 0.337838173f, -1.f),
    C_Vertex3f(0.191699028f, 0.390721321f, -1.f),
    C_Vertex3f(0.205249548f, 0.343915462f, -1.f),
    C_Vertex3f(0.203480721f, 0.399249077f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0910124779f, 0.400782108f, -1.f),
    C_Vertex3f(0.0873036385f, 0.408435345f, -1.f),
    C_Vertex3f(0.0880744457f, 0.330720425f, -1.f),
    C_Vertex3f(0.098472476f, 0.328202009f, -1.f),
    C_Vertex3f(0.096575737f, 0.393784523f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0854492188f, 0.416744232f, -1.f),
    C_Vertex3f(0.078316927f, 0.334078312f, -1.f),
    C_Vertex3f(0.0880744457f, 0.330720425f, -1.f),
    C_Vertex3f(0.0873036385f, 0.408435345f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0309734344f, 0.465892792f, -1.f),
    C_Vertex3f(0.0257406235f, 0.454099655f, -1.f),
    C_Vertex3f(0.025904417f, 0.387476444f, -1.f),
    C_Vertex3f(0.0295021534f, 0.378877163f, -1.f),
    C_Vertex3f(0.0339993238f, 0.370748758f, -1.f),
    C_Vertex3f(0.0413284302f, 0.479869843f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.025904417f, 0.387476444f, -1.f),
    C_Vertex3f(0.0257406235f, 0.454099655f, -1.f),
    C_Vertex3f(0.0232061148f, 0.396546602f, -1.f),
    C_Vertex3f(0.023124218f, 0.44558239f, -1.f),
    C_Vertex3f(0.0214072466f, 0.406087637f, -1.f),
    C_Vertex3f(0.0213799477f, 0.43640995f, -1.f),
    C_Vertex3f(0.0205078125f, 0.426582336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0214072466f, 0.406087637f, -1.f),
    C_Vertex3f(0.0205078125f, 0.426582336f, -1.f),
    C_Vertex3f(0.0205078125f, 0.416099548f, -1.f),
    C_End(),
    C_Translatef(0.331054688f, 0.f, 0.f),
};
static const Call char_171[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.870205104f, 0.492689639f, 0.f),
    C_Vertex3f(0.1484375f, 0.23046875f, 0.f),
    C_Vertex3f(0.1484375f, 0.23046875f, -1.f),
    C_Vertex3f(0.261230469f, 0.03125f, -1.f),
    C_Vertex3f(0.261230469f, 0.03125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.261230469f, 0.03125f, 0.f),
    C_Vertex3f(0.261230469f, 0.03125f, -1.f),
    C_Vertex3f(0.197265625f, 0.03125f, -1.f),
    C_Vertex3f(0.197265625f, 0.03125f, 0.f),
    C_Normal3f(-0.820741117f, -0.571300209f, 0.f),
    C_Vertex3f(0.197265625f, 0.03125f, 0.f),
    C_Vertex3f(0.197265625f, 0.03125f, -1.f),
    C_Vertex3f(0.05859375f, 0.23046875f, -1.f),
    C_Vertex3f(0.05859375f, 0.23046875f, 0.f),
    C_Normal3f(-0.81885463f, 0.574001014f, 0.f),
    C_Vertex3f(0.05859375f, 0.23046875f, 0.f),
    C_Vertex3f(0.05859375f, 0.23046875f, -1.f),
    C_Vertex3f(0.198242188f, 0.4296875f, -1.f),
    C_Vertex3f(0.198242188f, 0.4296875f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.198242188f, 0.4296875f, 0.f),
    C_Vertex3f(0.198242188f, 0.4296875f, -1.f),
    C_Vertex3f(0.262695313f, 0.4296875f, -1.f),
    C_Vertex3f(0.262695313f, 0.4296875f, 0.f),
    C_Normal3f(0.867457092f, -0.497512132f, 0.f),
    C_Vertex3f(0.262695313f, 0.4296875f, 0.f),
    C_Vertex3f(0.262695313f, 0.4296875f, -1.f),
    C_Vertex3f(0.1484375f, 0.23046875f, -1.f),
    C_Vertex3f(0.1484375f, 0.23046875f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.865619838f, 0.500701666f, 0.f),
    C_Vertex3f(0.317382813f, 0.23046875f, 0.f),
    C_Vertex3f(0.317382813f, 0.23046875f, -1.f),
    C_Vertex3f(0.432617188f, 0.03125f, -1.f),
    C_Vertex3f(0.432617188f, 0.03125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.432617188f, 0.03125f, 0.f),
    C_Vertex3f(0.432617188f, 0.03125f, -1.f),
    C_Vertex3f(0.366210938f, 0.03125f, -1.f),
    C_Vertex3f(0.366210938f, 0.03125f, 0.f),
    C_Normal3f(-0.824513376f, -0.565842509f, 0.f),
    C_Vertex3f(0.366210938f, 0.03125f, 0.f),
    C_Vertex3f(0.366210938f, 0.03125f, -1.f),
    C_Vertex3f(0.229492188f, 0.23046875f, -1.f),
    C_Vertex3f(0.229492188f, 0.23046875f, 0.f),
    C_Normal3f(-0.824513376f, 0.565842509f, 0.f),
    C_Vertex3f(0.229492188f, 0.23046875f, 0.f),
    C_Vertex3f(0.229492188f, 0.23046875f, -1.f),
    C_Vertex3f(0.366210938f, 0.4296875f, -1.f),
    C_Vertex3f(0.366210938f, 0.4296875f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.366210938f, 0.4296875f, 0.f),
    C_Vertex3f(0.366210938f, 0.4296875f, -1.f),
    C_Vertex3f(0.432617188f, 0.4296875f, -1.f),
    C_Vertex3f(0.432617188f, 0.4296875f, 0.f),
    C_Normal3f(0.865619838f, -0.500701666f, 0.f),
    C_Vertex3f(0.432617188f, 0.4296875f, 0.f),
    C_Vertex3f(0.432617188f, 0.4296875f, -1.f),
    C_Vertex3f(0.317382813f, 0.23046875f, -1.f),
    C_Vertex3f(0.317382813f, 0.23046875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1484375f, 0.23046875f, 0.f),
    C_Vertex3f(0.262695313f, 0.4296875f, 0.f),
    C_Vertex3f(0.198242188f, 0.4296875f, 0.f),
    C_Vertex3f(0.05859375f, 0.23046875f, 0.f),
    C_Vertex3f(0.197265625f, 0.03125f, 0.f),
    C_Vertex3f(0.261230469f, 0.03125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.317382813f, 0.23046875f, 0.f),
    C_Vertex3f(0.432617188f, 0.4296875f, 0.f),
    C_Vertex3f(0.366210938f, 0.4296875f, 0.f),
    C_Vertex3f(0.229492188f, 0.23046875f, 0.f),
    C_Vertex3f(0.366210938f, 0.03125f, 0.f),
    C_Vertex3f(0.432617188f, 0.03125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1484375f, 0.23046875f, -1.f),
    C_Vertex3f(0.262695313f, 0.4296875f, -1.f),
    C_Vertex3f(0.198242188f, 0.4296875f, -1.f),
    C_Vertex3f(0.05859375f, 0.23046875f, -1.f),
    C_Vertex3f(0.197265625f, 0.03125f, -1.f),
    C_Vertex3f(0.261230469f, 0.03125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.317382813f, 0.23046875f, -1.f),
    C_Vertex3f(0.432617188f, 0.4296875f, -1.f),
    C_Vertex3f(0.366210938f, 0.4296875f, -1.f),
    C_Vertex3f(0.229492188f, 0.23046875f, -1.f),
    C_Vertex3f(0.366210938f, 0.03125f, -1.f),
    C_Vertex3f(0.432617188f, 0.03125f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_172[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.47265625f, 0.185058594f, 0.f),
    C_Vertex3f(0.47265625f, 0.185058594f, -1.f),
    C_Vertex3f(0.3984375f, 0.185058594f, -1.f),
    C_Vertex3f(0.3984375f, 0.185058594f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.3984375f, 0.185058594f, 0.f),
    C_Vertex3f(0.3984375f, 0.185058594f, -1.f),
    C_Vertex3f(0.3984375f, 0.376464844f, -1.f),
    C_Vertex3f(0.3984375f, 0.376464844f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.3984375f, 0.376464844f, 0.f),
    C_Vertex3f(0.3984375f, 0.376464844f, -1.f),
    C_Vertex3f(0.0498046875f, 0.376464844f, -1.f),
    C_Vertex3f(0.0498046875f, 0.376464844f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0498046875f, 0.376464844f, 0.f),
    C_Vertex3f(0.0498046875f, 0.376464844f, -1.f),
    C_Vertex3f(0.0498046875f, 0.449707031f, -1.f),
    C_Vertex3f(0.0498046875f, 0.449707031f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0498046875f, 0.449707031f, 0.f),
    C_Vertex3f(0.0498046875f, 0.449707031f, -1.f),
    C_Vertex3f(0.47265625f, 0.449707031f, -1.f),
    C_Vertex3f(0.47265625f, 0.449707031f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.47265625f, 0.449707031f, 0.f),
    C_Vertex3f(0.47265625f, 0.449707031f, -1.f),
    C_Vertex3f(0.47265625f, 0.185058594f, -1.f),
    C_Vertex3f(0.47265625f, 0.185058594f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.3984375f, 0.376464844f, 0.f),
    C_Vertex3f(0.3984375f, 0.185058594f, 0.f),
    C_Vertex3f(0.47265625f, 0.185058594f, 0.f),
    C_Vertex3f(0.47265625f, 0.449707031f, 0.f),
    C_Vertex3f(0.0498046875f, 0.449707031f, 0.f),
    C_Vertex3f(0.0498046875f, 0.376464844f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.3984375f, 0.376464844f, -1.f),
    C_Vertex3f(0.3984375f, 0.185058594f, -1.f),
    C_Vertex3f(0.47265625f, 0.185058594f, -1.f),
    C_Vertex3f(0.47265625f, 0.449707031f, -1.f),
    C_Vertex3f(0.0498046875f, 0.449707031f, -1.f),
    C_Vertex3f(0.0498046875f, 0.376464844f, -1.f),
    C_End(),
    C_Translatef(0.522460938f, 0.f, 0.f),
};
static const Call char_173[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0278320313f, 0.192382813f, 0.f),
    C_Vertex3f(0.0278320313f, 0.192382813f, -1.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, -1.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, 0.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, -1.f),
    C_Vertex3f(0.270019531f, 0.271484375f, -1.f),
    C_Vertex3f(0.270019531f, 0.271484375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.270019531f, 0.271484375f, 0.f),
    C_Vertex3f(0.270019531f, 0.271484375f, -1.f),
    C_Vertex3f(0.270019531f, 0.192382813f, -1.f),
    C_Vertex3f(0.270019531f, 0.192382813f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.270019531f, 0.192382813f, 0.f),
    C_Vertex3f(0.270019531f, 0.192382813f, -1.f),
    C_Vertex3f(0.0278320313f, 0.192382813f, -1.f),
    C_Vertex3f(0.0278320313f, 0.192382813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0278320313f, 0.192382813f, 0.f),
    C_Vertex3f(0.270019531f, 0.192382813f, 0.f),
    C_Vertex3f(0.270019531f, 0.271484375f, 0.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0278320313f, 0.192382813f, -1.f),
    C_Vertex3f(0.270019531f, 0.192382813f, -1.f),
    C_Vertex3f(0.270019531f, 0.271484375f, -1.f),
    C_Vertex3f(0.0278320313f, 0.271484375f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_174[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0321503282f, 0.999482989f, 0.f),
    C_Vertex3f(0.321137428f, 0.651855469f, 0.f),
    C_Vertex3f(0.321137428f, 0.651855469f, -1.f),
    C_Normal3f(0.0321515314f, 0.999483049f, 0.f),
    C_Vertex3f(0.341901779f, 0.651855469f, 0.f),
    C_Vertex3f(0.341901779f, 0.651855469f, -1.f),
    C_Normal3f(0.0966276675f, 0.995320618f, 0.f),
    C_Vertex3f(0.36254251f, 0.650523543f, 0.f),
    C_Vertex3f(0.36254251f, 0.650523543f, -1.f),
    C_Normal3f(0.16065906f, 0.987010002f, 0.f),
    C_Vertex3f(0.383060575f, 0.647859693f, 0.f),
    C_Vertex3f(0.383060575f, 0.647859693f, -1.f),
    C_Normal3f(0.223460346f, 0.974712968f, 0.f),
    C_Vertex3f(0.403455973f, 0.643863916f, 0.f),
    C_Vertex3f(0.403455973f, 0.643863916f, -1.f),
    C_Normal3f(0.284310788f, 0.958732128f, 0.f),
    C_Vertex3f(0.423728704f, 0.638536215f, 0.f),
    C_Vertex3f(0.423728704f, 0.638536215f, -1.f),
    C_Normal3f(0.34259367f, 0.939483702f, 0.f),
    C_Vertex3f(0.44387877f, 0.631876588f, 0.f),
    C_Vertex3f(0.44387877f, 0.631876588f, -1.f),
    C_Normal3f(0.397822261f, 0.917462528f, 0.f),
    C_Vertex3f(0.463906169f, 0.623885036f, 0.f),
    C_Vertex3f(0.463906169f, 0.623885036f, -1.f),
    C_Normal3f(0.449651271f, 0.893204153f, 0.f),
    C_Vertex3f(0.483810902f, 0.614561558f, 0.f),
    C_Vertex3f(0.483810902f, 0.614561558f, -1.f),
    C_Normal3f(0.50415504f, 0.863613188f, 0.f),
    C_Vertex3f(0.503592968f, 0.603906155f, 0.f),
    C_Vertex3f(0.503592968f, 0.603906155f, -1.f),
    C_Normal3f(0.563552618f, 0.826080143f, 0.f),
    C_Vertex3f(0.522282004f, 0.592103124f, 0.f),
    C_Vertex3f(0.522282004f, 0.592103124f, -1.f),
    C_Normal3f(0.621464193f, 0.783442616f, 0.f),
    C_Vertex3f(0.539878011f, 0.579152465f, 0.f),
    C_Vertex3f(0.539878011f, 0.579152465f, -1.f),
    C_Normal3f(0.676860511f, 0.736111283f, 0.f),
    C_Vertex3f(0.556380987f, 0.565054178f, 0.f),
    C_Vertex3f(0.556380987f, 0.565054178f, -1.f),
    C_Normal3f(0.728791237f, 0.684736073f, 0.f),
    C_Vertex3f(0.571790934f, 0.549808264f, 0.f),
    C_Vertex3f(0.571790934f, 0.549808264f, -1.f),
    C_Normal3f(0.776462734f, 0.630163252f, 0.f),
    C_Vertex3f(0.58610785f, 0.533414721f, 0.f),
    C_Vertex3f(0.58610785f, 0.533414721f, -1.f),
    C_Normal3f(0.819298148f, 0.573367715f, 0.f),
    C_Vertex3f(0.599331737f, 0.515873551f, 0.f),
    C_Vertex3f(0.599331737f, 0.515873551f, -1.f),
    C_Normal3f(0.856966972f, 0.515371323f, 0.f),
    C_Vertex3f(0.611462593f, 0.497184753f, 0.f),
    C_Vertex3f(0.611462593f, 0.497184753f, -1.f),
    C_Normal3f(0.887617767f, 0.460580915f, 0.f),
    C_Vertex3f(0.62250042f, 0.477348328f, 0.f),
    C_Vertex3f(0.62250042f, 0.477348328f, -1.f),
    C_Normal3f(0.913628042f, 0.406551063f, 0.f),
    C_Vertex3f(0.632158518f, 0.457300186f, 0.f),
    C_Vertex3f(0.632158518f, 0.457300186f, -1.f),
    C_Normal3f(0.93706286f, 0.349160671f, 0.f),
    C_Vertex3f(0.640436888f, 0.437040329f, 0.f),
    C_Vertex3f(0.640436888f, 0.437040329f, -1.f),
    C_Normal3f(0.957369864f, 0.288864851f, 0.f),
    C_Vertex3f(0.647335529f, 0.416568756f, 0.f),
    C_Vertex3f(0.647335529f, 0.416568756f, -1.f),
    C_Normal3f(0.974066257f, 0.226262957f, 0.f),
    C_Vertex3f(0.652854443f, 0.395885468f, 0.f),
    C_Vertex3f(0.652854443f, 0.395885468f, -1.f),
    C_Normal3f(0.986778498f, 0.162075043f, 0.f),
    C_Vertex3f(0.656993628f, 0.374990463f, 0.f),
    C_Vertex3f(0.656993628f, 0.374990463f, -1.f),
    C_Normal3f(0.995274305f, 0.0971033424f, 0.f),
    C_Vertex3f(0.659753084f, 0.353883743f, 0.f),
    C_Vertex3f(0.659753084f, 0.353883743f, -1.f),
    C_Normal3f(0.999479294f, 0.032265909f, 0.f),
    C_Vertex3f(0.661132813f, 0.332565308f, 0.f),
    C_Vertex3f(0.661132813f, 0.332565308f, -1.f),
    C_Normal3f(0.999489665f, -0.0319455191f, 0.f),
    C_Vertex3f(0.661132813f, 0.311144829f, 0.f),
    C_Vertex3f(0.661132813f, 0.311144829f, -1.f),
    C_Normal3f(0.995321691f, -0.096616663f, 0.f),
    C_Vertex3f(0.659773588f, 0.290038824f, 0.f),
    C_Vertex3f(0.659773588f, 0.290038824f, -1.f),
    C_Normal3f(0.986915231f, -0.161239952f, 0.f),
    C_Vertex3f(0.65705514f, 0.269137621f, 0.f),
    C_Vertex3f(0.65705514f, 0.269137621f, -1.f),
    C_Normal3f(0.974341154f, -0.225076169f, 0.f),
    C_Vertex3f(0.652977467f, 0.248441219f, 0.f),
    C_Vertex3f(0.652977467f, 0.248441219f, -1.f),
    C_Normal3f(0.957829475f, -0.287337095f, 0.f),
    C_Vertex3f(0.647540569f, 0.227949619f, 0.f),
    C_Vertex3f(0.647540569f, 0.227949619f, -1.f),
    C_Normal3f(0.937748313f, -0.347315639f, 0.f),
    C_Vertex3f(0.640744448f, 0.207662821f, 0.f),
    C_Vertex3f(0.640744448f, 0.207662821f, -1.f),
    C_Normal3f(0.91457212f, -0.404422909f, 0.f),
    C_Vertex3f(0.632589102f, 0.187580824f, 0.f),
    C_Vertex3f(0.632589102f, 0.187580824f, -1.f),
    C_Normal3f(0.888843417f, -0.458211154f, 0.f),
    C_Vertex3f(0.623074532f, 0.167703629f, 0.f),
    C_Vertex3f(0.623074532f, 0.167703629f, -1.f),
    C_Normal3f(0.858544528f, -0.512739122f, 0.f),
    C_Vertex3f(0.612200737f, 0.148031235f, 0.f),
    C_Vertex3f(0.612200737f, 0.148031235f, -1.f),
    C_Normal3f(0.821421087f, -0.570322216f, 0.f),
    C_Vertex3f(0.600233912f, 0.129458666f, 0.f),
    C_Vertex3f(0.600233912f, 0.129458666f, -1.f),
    C_Normal3f(0.779336512f, -0.62660557f, 0.f),
    C_Vertex3f(0.587174058f, 0.111985922f, 0.f),
    C_Vertex3f(0.587174058f, 0.111985922f, -1.f),
    C_Normal3f(0.732640922f, -0.680615366f, 0.f),
    C_Vertex3f(0.573021173f, 0.0956130028f, 0.f),
    C_Vertex3f(0.573021173f, 0.0956130028f, -1.f),
    C_Normal3f(0.681911707f, -0.731434464f, 0.f),
    C_Vertex3f(0.557775259f, 0.0803399086f, 0.f),
    C_Vertex3f(0.557775259f, 0.0803399086f, -1.f),
    C_Normal3f(0.627921402f, -0.778276742f, 0.f),
    C_Vertex3f(0.541436315f, 0.0661666393f, 0.f),
    C_Vertex3f(0.541436315f, 0.0661666393f, -1.f),
    C_Normal3f(0.571579754f, -0.820546567f, 0.f),
    C_Vertex3f(0.52400434f, 0.053093195f, 0.f),
    C_Vertex3f(0.52400434f, 0.053093195f, -1.f),
    C_Normal3f(0.513860404f, -0.857873738f, 0.f),
    C_Vertex3f(0.505479336f, 0.0411195755f, 0.f),
    C_Vertex3f(0.505479336f, 0.0411195755f, -1.f),
    C_Normal3f(0.459020972f, -0.888425529f, 0.f),
    C_Vertex3f(0.485861301f, 0.0302457809f, 0.f),
    C_Vertex3f(0.485861301f, 0.0302457809f, -1.f),
    C_Normal3f(0.404825926f, -0.914393783f, 0.f),
    C_Vertex3f(0.466018081f, 0.0207312107f, 0.f),
    C_Vertex3f(0.466018081f, 0.0207312107f, -1.f),
    C_Normal3f(0.347367018f, -0.93772918f, 0.f),
    C_Vertex3f(0.445949674f, 0.0125758648f, 0.f),
    C_Vertex3f(0.445949674f, 0.0125758648f, -1.f),
    C_Normal3f(0.28711763f, -0.957895339f, 0.f),
    C_Vertex3f(0.42565608f, 0.00577974319f, 0.f),
    C_Vertex3f(0.42565608f, 0.00577974319f, -1.f),
    C_Normal3f(0.224689066f, -0.974430501f, 0.f),
    C_Vertex3f(0.4051373f, 0.000342845917f, 0.f),
    C_Vertex3f(0.4051373f, 0.000342845917f, -1.f),
    C_Normal3f(0.160805002f, -0.98698616f, 0.f),
    C_Vertex3f(0.384393334f, -0.00373482704f, 0.f),
    C_Vertex3f(0.384393334f, -0.00373482704f, -1.f),
    C_Normal3f(0.0962618664f, -0.995356023f, 0.f),
    C_Vertex3f(0.363424182f, -0.00645327568f, 0.f),
    C_Vertex3f(0.363424182f, -0.00645327568f, -1.f),
    C_Normal3f(0.0318793692f, -0.999491751f, 0.f),
    C_Vertex3f(0.342229843f, -0.0078125f, 0.f),
    C_Vertex3f(0.342229843f, -0.0078125f, -1.f),
    C_Normal3f(-0.0318782106f, -0.999491811f, 0.f),
    C_Vertex3f(0.320809364f, -0.0078125f, 0.f),
    C_Vertex3f(0.320809364f, -0.0078125f, -1.f),
    C_Normal3f(-0.0962556675f, -0.995356679f, 0.f),
    C_Vertex3f(0.299613476f, -0.00645327568f, 0.f),
    C_Vertex3f(0.299613476f, -0.00645327568f, -1.f),
    C_Normal3f(-0.160797387f, -0.986987412f, 0.f),
    C_Vertex3f(0.278643131f, -0.00373482704f, 0.f),
    C_Vertex3f(0.278643131f, -0.00373482704f, -1.f),
    C_Normal3f(-0.224682286f, -0.974432111f, 0.f),
    C_Vertex3f(0.257898331f, 0.000342845917f, 0.f),
    C_Vertex3f(0.257898331f, 0.000342845917f, -1.f),
    C_Normal3f(-0.287113786f, -0.957896471f, 0.f),
    C_Vertex3f(0.237379074f, 0.00577974319f, 0.f),
    C_Vertex3f(0.237379074f, 0.00577974319f, -1.f),
    C_Normal3f(-0.347367913f, -0.937728882f, 0.f),
    C_Vertex3f(0.217085361f, 0.0125758648f, 0.f),
    C_Vertex3f(0.217085361f, 0.0125758648f, -1.f),
    C_Normal3f(-0.404833019f, -0.914390683f, 0.f),
    C_Vertex3f(0.197017193f, 0.0207312107f, 0.f),
    C_Vertex3f(0.197017193f, 0.0207312107f, -1.f),
    C_Normal3f(-0.459035158f, -0.888418078f, 0.f),
    C_Vertex3f(0.177174568f, 0.0302457809f, 0.f),
    C_Vertex3f(0.177174568f, 0.0302457809f, -1.f),
    C_Normal3f(-0.513809621f, -0.857904255f, 0.f),
    C_Vertex3f(0.157557487f, 0.0411195755f, 0.f),
    C_Vertex3f(0.157557487f, 0.0411195755f, -1.f),
    C_Normal3f(-0.571374416f, -0.820689559f, 0.f),
    C_Vertex3f(0.139026403f, 0.053093195f, 0.f),
    C_Vertex3f(0.139026403f, 0.053093195f, -1.f),
    C_Normal3f(-0.627547026f, -0.778578699f, 0.f),
    C_Vertex3f(0.121581316f, 0.0661666393f, 0.f),
    C_Vertex3f(0.121581316f, 0.0661666393f, -1.f),
    C_Normal3f(-0.681365311f, -0.731943488f, 0.f),
    C_Vertex3f(0.105222225f, 0.0803399086f, 0.f),
    C_Vertex3f(0.105222225f, 0.0803399086f, -1.f),
    C_Normal3f(-0.731932461f, -0.681377172f, 0.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, 0.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, -1.f),
    C_Normal3f(-0.778488278f, -0.627659202f, 0.f),
    C_Vertex3f(0.0757620335f, 0.111985922f, 0.f),
    C_Vertex3f(0.0757620335f, 0.111985922f, -1.f),
    C_Normal3f(-0.820465326f, -0.571696401f, 0.f),
    C_Vertex3f(0.0626609325f, 0.129458666f, 0.f),
    C_Vertex3f(0.0626609325f, 0.129458666f, -1.f),
    C_Normal3f(-0.857520282f, -0.514450133f, 0.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, 0.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, -1.f),
    C_Normal3f(-0.88789314f, -0.460049629f, 0.f),
    C_Vertex3f(0.0397167206f, 0.167703629f, 0.f),
    C_Vertex3f(0.0397167206f, 0.167703629f, -1.f),
    C_Normal3f(-0.913810194f, -0.40614152f, 0.f),
    C_Vertex3f(0.0301537514f, 0.187580824f, 0.f),
    C_Vertex3f(0.0301537514f, 0.187580824f, -1.f),
    C_Normal3f(-0.937171936f, -0.348867804f, 0.f),
    C_Vertex3f(0.0219569206f, 0.207662821f, 0.f),
    C_Vertex3f(0.0219569206f, 0.207662821f, -1.f),
    C_Normal3f(-0.957426429f, -0.288677186f, 0.f),
    C_Vertex3f(0.0151262283f, 0.227949619f, 0.f),
    C_Vertex3f(0.0151262283f, 0.227949619f, -1.f),
    C_Normal3f(-0.974089563f, -0.226162672f, 0.f),
    C_Vertex3f(0.0096616745f, 0.248441219f, 0.f),
    C_Vertex3f(0.0096616745f, 0.248441219f, -1.f),
    C_Normal3f(-0.986784399f, -0.162038669f, 0.f),
    C_Vertex3f(0.00556325912f, 0.269137621f, 0.f),
    C_Vertex3f(0.00556325912f, 0.269137621f, -1.f),
    C_Normal3f(-0.995274305f, -0.0971035138f, 0.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, 0.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, -1.f),
    C_Normal3f(-0.99948436f, -0.0321078524f, 0.f),
    C_Vertex3f(0.00146484375f, 0.311144829f, 0.f),
    C_Vertex3f(0.00146484375f, 0.311144829f, -1.f),
    C_Normal3f(-0.999474049f, 0.0324274302f, 0.f),
    C_Vertex3f(0.00146484375f, 0.332565308f, 0.f),
    C_Vertex3f(0.00146484375f, 0.332565308f, -1.f),
    C_Normal3f(-0.995227098f, 0.0975853205f, 0.f),
    C_Vertex3f(0.00285148621f, 0.353883743f, 0.f),
    C_Vertex3f(0.00285148621f, 0.353883743f, -1.f),
    C_Normal3f(-0.986648202f, 0.162865728f, 0.f),
    C_Vertex3f(0.00562477112f, 0.374990463f, 0.f),
    C_Vertex3f(0.00562477112f, 0.374990463f, -1.f),
    C_Normal3f(-0.973815799f, 0.227338344f, 0.f),
    C_Vertex3f(0.00978469849f, 0.395885468f, 0.f),
    C_Vertex3f(0.00978469849f, 0.395885468f, -1.f),
    C_Normal3f(-0.956968844f, 0.290190816f, 0.f),
    C_Vertex3f(0.0153312683f, 0.416568756f, 0.f),
    C_Vertex3f(0.0153312683f, 0.416568756f, -1.f),
    C_Normal3f(-0.936489463f, 0.35069567f, 0.f),
    C_Vertex3f(0.0222644806f, 0.437040329f, 0.f),
    C_Vertex3f(0.0222644806f, 0.437040329f, -1.f),
    C_Normal3f(-0.912870347f, 0.408249557f, 0.f),
    C_Vertex3f(0.0305843353f, 0.457300186f, 0.f),
    C_Vertex3f(0.0305843353f, 0.457300186f, -1.f),
    C_Normal3f(-0.886673331f, 0.462396473f, 0.f),
    C_Vertex3f(0.0402908325f, 0.477348328f, 0.f),
    C_Vertex3f(0.0402908325f, 0.477348328f, -1.f),
    C_Normal3f(-0.855948746f, 0.517060578f, 0.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, 0.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, -1.f),
    C_Normal3f(-0.818347096f, 0.574724317f, 0.f),
    C_Vertex3f(0.0635631084f, 0.515873551f, 0.f),
    C_Vertex3f(0.0635631084f, 0.515873551f, -1.f),
    C_Normal3f(-0.775618136f, 0.631202459f, 0.f),
    C_Vertex3f(0.0768282413f, 0.533414721f, 0.f),
    C_Vertex3f(0.0768282413f, 0.533414721f, -1.f),
    C_Normal3f(-0.728085756f, 0.685486019f, 0.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, 0.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, -1.f),
    C_Normal3f(-0.676316857f, 0.73661083f, 0.f),
    C_Vertex3f(0.106616497f, 0.565054178f, 0.f),
    C_Vertex3f(0.106616497f, 0.565054178f, -1.f),
    C_Normal3f(-0.62109226f, 0.783737421f, 0.f),
    C_Vertex3f(0.12313962f, 0.579152465f, 0.f),
    C_Vertex3f(0.12313962f, 0.579152465f, -1.f),
    C_Normal3f(-0.563349247f, 0.826218843f, 0.f),
    C_Vertex3f(0.140748739f, 0.592103124f, 0.f),
    C_Vertex3f(0.140748739f, 0.592103124f, -1.f),
    C_Normal3f(-0.504104972f, 0.863642454f, 0.f),
    C_Vertex3f(0.159443855f, 0.603906155f, 0.f),
    C_Vertex3f(0.159443855f, 0.603906155f, -1.f),
    C_Normal3f(-0.449665278f, 0.893197119f, 0.f),
    C_Vertex3f(0.179224968f, 0.614561558f, 0.f),
    C_Vertex3f(0.179224968f, 0.614561558f, -1.f),
    C_Normal3f(-0.397829235f, 0.917459428f, 0.f),
    C_Vertex3f(0.199129105f, 0.623885036f, 0.f),
    C_Vertex3f(0.199129105f, 0.623885036f, -1.f),
    C_Normal3f(-0.342594564f, 0.939483404f, 0.f),
    C_Vertex3f(0.219156265f, 0.631876588f, 0.f),
    C_Vertex3f(0.219156265f, 0.631876588f, -1.f),
    C_Normal3f(-0.284306943f, 0.95873332f, 0.f),
    C_Vertex3f(0.23930645f, 0.638536215f, 0.f),
    C_Vertex3f(0.23930645f, 0.638536215f, -1.f),
    C_Normal3f(-0.223453522f, 0.974714637f, 0.f),
    C_Vertex3f(0.259579659f, 0.643863916f, 0.f),
    C_Vertex3f(0.259579659f, 0.643863916f, -1.f),
    C_Normal3f(-0.160651311f, 0.987011254f, 0.f),
    C_Vertex3f(0.279975891f, 0.647859693f, 0.f),
    C_Vertex3f(0.279975891f, 0.647859693f, -1.f),
    C_Normal3f(-0.0966212898f, 0.995321214f, 0.f),
    C_Vertex3f(0.300495148f, 0.650523543f, 0.f),
    C_Vertex3f(0.300495148f, 0.650523543f, -1.f),
    C_Normal3f(-0.0321503282f, 0.999482989f, 0.f),
    C_Vertex3f(0.321137428f, 0.651855469f, 0.f),
    C_Vertex3f(0.321137428f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0322615579f, -0.999479473f, 0.f),
    C_Vertex3f(0.340151787f, 0.597167969f, 0.f),
    C_Vertex3f(0.340151787f, 0.597167969f, -1.f),
    C_Normal3f(0.0321776681f, -0.999482155f, 0.f),
    C_Vertex3f(0.322833061f, 0.597167969f, 0.f),
    C_Vertex3f(0.322833061f, 0.597167969f, -1.f),
    C_Normal3f(0.0966247916f, -0.995320916f, 0.f),
    C_Vertex3f(0.305575013f, 0.596054792f, 0.f),
    C_Vertex3f(0.305575013f, 0.596054792f, -1.f),
    C_Normal3f(0.160777718f, -0.986990571f, 0.f),
    C_Vertex3f(0.288432956f, 0.59382844f, 0.f),
    C_Vertex3f(0.288432956f, 0.59382844f, -1.f),
    C_Normal3f(0.223790944f, -0.97463721f, 0.f),
    C_Vertex3f(0.271406889f, 0.590488911f, 0.f),
    C_Vertex3f(0.271406889f, 0.590488911f, -1.f),
    C_Normal3f(0.284929276f, -0.958548546f, 0.f),
    C_Vertex3f(0.254496813f, 0.586036205f, 0.f),
    C_Vertex3f(0.254496813f, 0.586036205f, -1.f),
    C_Normal3f(0.343557268f, -0.939131677f, 0.f),
    C_Vertex3f(0.237702727f, 0.580470324f, 0.f),
    C_Vertex3f(0.237702727f, 0.580470324f, -1.f),
    C_Normal3f(0.399167597f, -0.916878045f, 0.f),
    C_Vertex3f(0.221024632f, 0.573791265f, 0.f),
    C_Vertex3f(0.221024632f, 0.573791265f, -1.f),
    C_Normal3f(0.45139423f, -0.892324567f, 0.f),
    C_Vertex3f(0.204462528f, 0.565999031f, 0.f),
    C_Vertex3f(0.204462528f, 0.565999031f, -1.f),
    C_Normal3f(0.505917668f, -0.862581789f, 0.f),
    C_Vertex3f(0.188016415f, 0.55709362f, 0.f),
    C_Vertex3f(0.188016415f, 0.55709362f, -1.f),
    C_Normal3f(0.56488502f, -0.825169563f, 0.f),
    C_Vertex3f(0.172465086f, 0.547232032f, 0.f),
    C_Vertex3f(0.172465086f, 0.547232032f, -1.f),
    C_Normal3f(0.622291386f, -0.782785594f, 0.f),
    C_Vertex3f(0.157808542f, 0.536414266f, 0.f),
    C_Vertex3f(0.157808542f, 0.536414266f, -1.f),
    C_Normal3f(0.677140772f, -0.735853434f, 0.f),
    C_Vertex3f(0.144046783f, 0.524640322f, 0.f),
    C_Vertex3f(0.144046783f, 0.524640322f, -1.f),
    C_Normal3f(0.728521347f, -0.685023189f, 0.f),
    C_Vertex3f(0.13117981f, 0.5119102f, 0.f),
    C_Vertex3f(0.13117981f, 0.5119102f, -1.f),
    C_Normal3f(0.775678813f, -0.631127894f, 0.f),
    C_Vertex3f(0.119207621f, 0.498223901f, 0.f),
    C_Vertex3f(0.119207621f, 0.498223901f, -1.f),
    C_Normal3f(0.818070829f, -0.575117528f, 0.f),
    C_Vertex3f(0.108130217f, 0.483581424f, 0.f),
    C_Vertex3f(0.108130217f, 0.483581424f, -1.f),
    C_Normal3f(0.855392396f, -0.517980456f, 0.f),
    C_Vertex3f(0.0979475975f, 0.467982769f, 0.f),
    C_Vertex3f(0.0979475975f, 0.467982769f, -1.f),
    C_Normal3f(0.886027932f, -0.463631988f, 0.f),
    C_Vertex3f(0.0886597633f, 0.451427937f, 0.f),
    C_Vertex3f(0.0886597633f, 0.451427937f, -1.f),
    C_Normal3f(0.912299216f, -0.409524322f, 0.f),
    C_Vertex3f(0.0805329084f, 0.434702277f, 0.f),
    C_Vertex3f(0.0805329084f, 0.434702277f, -1.f),
    C_Normal3f(0.936017632f, -0.351953179f, 0.f),
    C_Vertex3f(0.0735670328f, 0.417805791f, 0.f),
    C_Vertex3f(0.0735670328f, 0.417805791f, -1.f),
    C_Normal3f(0.956611574f, -0.291366369f, 0.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, 0.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, -1.f),
    C_Normal3f(0.973576069f, -0.228362828f, 0.f),
    C_Vertex3f(0.0631182194f, 0.383500338f, 0.f),
    C_Vertex3f(0.0631182194f, 0.383500338f, -1.f),
    C_Normal3f(0.986515164f, -0.163670227f, 0.f),
    C_Vertex3f(0.0596352816f, 0.366091371f, 0.f),
    C_Vertex3f(0.0596352816f, 0.366091371f, -1.f),
    C_Normal3f(0.995175898f, -0.0981062874f, 0.f),
    C_Vertex3f(0.057313323f, 0.348511577f, 0.f),
    C_Vertex3f(0.057313323f, 0.348511577f, -1.f),
    C_Normal3f(0.999469161f, -0.0325791314f, 0.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, 0.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, -1.f),
    C_Normal3f(0.999484181f, 0.0321161337f, 0.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, 0.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, -1.f),
    C_Normal3f(0.995274365f, 0.0971023366f, 0.f),
    C_Vertex3f(0.057292819f, 0.295268297f, 0.f),
    C_Vertex3f(0.057292819f, 0.295268297f, -1.f),
    C_Normal3f(0.986764789f, 0.162157908f, 0.f),
    C_Vertex3f(0.0595737696f, 0.277826071f, 0.f),
    C_Vertex3f(0.0595737696f, 0.277826071f, -1.f),
    C_Normal3f(0.974012792f, 0.226493135f, 0.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, 0.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, -1.f),
    C_Normal3f(0.957239568f, 0.289296657f, 0.f),
    C_Vertex3f(0.0675570965f, 0.243494511f, 0.f),
    C_Vertex3f(0.0675570965f, 0.243494511f, -1.f),
    C_Normal3f(0.93681109f, 0.349835664f, 0.f),
    C_Vertex3f(0.0732594728f, 0.226605177f, 0.f),
    C_Vertex3f(0.0732594728f, 0.226605177f, -1.f),
    C_Normal3f(0.913206995f, 0.407496005f, 0.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, 0.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, -1.f),
    C_Normal3f(0.886980236f, 0.461807281f, 0.f),
    C_Vertex3f(0.0880856514f, 0.193379402f, 0.f),
    C_Vertex3f(0.0880856514f, 0.193379402f, -1.f),
    C_Normal3f(0.856382728f, 0.516341746f, 0.f),
    C_Vertex3f(0.0972094536f, 0.177042961f, 0.f),
    C_Vertex3f(0.0972094536f, 0.177042961f, -1.f),
    C_Normal3f(0.819295764f, 0.573371112f, 0.f),
    C_Vertex3f(0.107241631f, 0.1616081f, 0.f),
    C_Vertex3f(0.107241631f, 0.1616081f, -1.f),
    C_Normal3f(0.777389884f, 0.629019022f, 0.f),
    C_Vertex3f(0.118182182f, 0.147074819f, 0.f),
    C_Vertex3f(0.118182182f, 0.147074819f, -1.f),
    C_Normal3f(0.73102957f, 0.682345867f, 0.f),
    C_Vertex3f(0.130031109f, 0.133443117f, 0.f),
    C_Vertex3f(0.130031109f, 0.133443117f, -1.f),
    C_Normal3f(0.68079406f, 0.732474983f, 0.f),
    C_Vertex3f(0.14278841f, 0.120712996f, 0.f),
    C_Vertex3f(0.14278841f, 0.120712996f, -1.f),
    C_Normal3f(0.62744385f, 0.778661847f, 0.f),
    C_Vertex3f(0.156454086f, 0.108884454f, 0.f),
    C_Vertex3f(0.156454086f, 0.108884454f, -1.f),
    C_Normal3f(0.571864069f, 0.820348322f, 0.f),
    C_Vertex3f(0.171028137f, 0.0979574919f, 0.f),
    C_Vertex3f(0.171028137f, 0.0979574919f, -1.f),
    C_Normal3f(0.514994919f, 0.857193172f, 0.f),
    C_Vertex3f(0.186510563f, 0.0879321098f, 0.f),
    C_Vertex3f(0.186510563f, 0.0879321098f, -1.f),
    C_Normal3f(0.460680038f, 0.887566268f, 0.f),
    C_Vertex3f(0.202901363f, 0.0788083076f, 0.f),
    C_Vertex3f(0.202901363f, 0.0788083076f, -1.f),
    C_Normal3f(0.406587005f, 0.913612008f, 0.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, 0.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, -1.f),
    C_Normal3f(0.349142969f, 0.937069476f, 0.f),
    C_Vertex3f(0.236216187f, 0.0639821291f, 0.f),
    C_Vertex3f(0.236216187f, 0.0639821291f, -1.f),
    C_Normal3f(0.288807094f, 0.957387328f, 0.f),
    C_Vertex3f(0.253140211f, 0.0582797527f, 0.f),
    C_Vertex3f(0.253140211f, 0.0582797527f, -1.f),
    C_Normal3f(0.226182401f, 0.974084914f, 0.f),
    C_Vertex3f(0.270241976f, 0.0537178516f, 0.f),
    C_Vertex3f(0.270241976f, 0.0537178516f, -1.f),
    C_Normal3f(0.161991552f, 0.986792147f, 0.f),
    C_Vertex3f(0.287521482f, 0.0502964258f, 0.f),
    C_Vertex3f(0.287521482f, 0.0502964258f, -1.f),
    C_Normal3f(0.0970379636f, 0.995280623f, 0.f),
    C_Vertex3f(0.304978728f, 0.0480154753f, 0.f),
    C_Vertex3f(0.304978728f, 0.0480154753f, -1.f),
    C_Normal3f(0.0321576744f, 0.99948281f, 0.f),
    C_Vertex3f(0.322613716f, 0.046875f, 0.f),
    C_Vertex3f(0.322613716f, 0.046875f, -1.f),
    C_Normal3f(-0.0321591832f, 0.999482751f, 0.f),
    C_Vertex3f(0.340425491f, 0.046875f, 0.f),
    C_Vertex3f(0.340425491f, 0.046875f, -1.f),
    C_Normal3f(-0.0970464498f, 0.995279849f, 0.f),
    C_Vertex3f(0.35805881f, 0.0480154753f, 0.f),
    C_Vertex3f(0.35805881f, 0.0480154753f, -1.f),
    C_Normal3f(-0.162003458f, 0.98679018f, 0.f),
    C_Vertex3f(0.375514627f, 0.0502964258f, 0.f),
    C_Vertex3f(0.375514627f, 0.0502964258f, -1.f),
    C_Normal3f(-0.226195812f, 0.974081814f, 0.f),
    C_Vertex3f(0.39279294f, 0.0537178516f, 0.f),
    C_Vertex3f(0.39279294f, 0.0537178516f, -1.f),
    C_Normal3f(-0.288820088f, 0.957383394f, 0.f),
    C_Vertex3f(0.409893751f, 0.0582797527f, 0.f),
    C_Vertex3f(0.409893751f, 0.0582797527f, -1.f),
    C_Normal3f(-0.349153817f, 0.937065423f, 0.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, 0.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, -1.f),
    C_Normal3f(-0.406594306f, 0.913608789f, 0.f),
    C_Vertex3f(0.443562865f, 0.0708249807f, 0.f),
    C_Vertex3f(0.443562865f, 0.0708249807f, -1.f),
    C_Normal3f(-0.46068269f, 0.887564957f, 0.f),
    C_Vertex3f(0.460131168f, 0.0788083076f, 0.f),
    C_Vertex3f(0.460131168f, 0.0788083076f, -1.f),
    C_Normal3f(-0.515075624f, 0.857144773f, 0.f),
    C_Vertex3f(0.476521969f, 0.0879321098f, 0.f),
    C_Vertex3f(0.476521969f, 0.0879321098f, -1.f),
    C_Normal3f(-0.572125196f, 0.82016629f, 0.f),
    C_Vertex3f(0.4919976f, 0.0979574919f, 0.f),
    C_Vertex3f(0.4919976f, 0.0979574919f, -1.f),
    C_Normal3f(-0.627901733f, 0.778292656f, 0.f),
    C_Vertex3f(0.506558061f, 0.108884454f, 0.f),
    C_Vertex3f(0.506558061f, 0.108884454f, -1.f),
    C_Normal3f(-0.681451917f, 0.731862903f, 0.f),
    C_Vertex3f(0.520203352f, 0.120712996f, 0.f),
    C_Vertex3f(0.520203352f, 0.120712996f, -1.f),
    C_Normal3f(-0.731875896f, 0.681438029f, 0.f),
    C_Vertex3f(0.532933474f, 0.133443117f, 0.f),
    C_Vertex3f(0.532933474f, 0.133443117f, -1.f),
    C_Normal3f(-0.77839905f, 0.627769887f, 0.f),
    C_Vertex3f(0.544748425f, 0.147074819f, 0.f),
    C_Vertex3f(0.544748425f, 0.147074819f, -1.f),
    C_Normal3f(-0.82043016f, 0.571746707f, 0.f),
    C_Vertex3f(0.555648208f, 0.1616081f, 0.f),
    C_Vertex3f(0.555648208f, 0.1616081f, -1.f),
    C_Normal3f(-0.857596815f, 0.514322579f, 0.f),
    C_Vertex3f(0.56563282f, 0.177042961f, 0.f),
    C_Vertex3f(0.56563282f, 0.177042961f, -1.f),
    C_Normal3f(-0.888106108f, 0.459638536f, 0.f),
    C_Vertex3f(0.574702263f, 0.193379402f, 0.f),
    C_Vertex3f(0.574702263f, 0.193379402f, -1.f),
    C_Normal3f(-0.914109051f, 0.405468255f, 0.f),
    C_Vertex3f(0.582638025f, 0.209900141f, 0.f),
    C_Vertex3f(0.582638025f, 0.209900141f, -1.f),
    C_Normal3f(-0.937492907f, 0.348004431f, 0.f),
    C_Vertex3f(0.589440107f, 0.226605177f, 0.f),
    C_Vertex3f(0.589440107f, 0.226605177f, -1.f),
    C_Normal3f(-0.95771575f, 0.287716091f, 0.f),
    C_Vertex3f(0.595108509f, 0.243494511f, 0.f),
    C_Vertex3f(0.595108509f, 0.243494511f, -1.f),
    C_Normal3f(-0.974309683f, 0.225212321f, 0.f),
    C_Vertex3f(0.59964323f, 0.260568142f, 0.f),
    C_Vertex3f(0.59964323f, 0.260568142f, -1.f),
    C_Normal3f(-0.986919045f, 0.161216959f, 0.f),
    C_Vertex3f(0.603044271f, 0.277826071f, 0.f),
    C_Vertex3f(0.603044271f, 0.277826071f, -1.f),
    C_Normal3f(-0.995330095f, 0.0965292007f, 0.f),
    C_Vertex3f(0.605311632f, 0.295268297f, 0.f),
    C_Vertex3f(0.605311632f, 0.295268297f, -1.f),
    C_Normal3f(-0.999490261f, 0.0319249816f, 0.f),
    C_Vertex3f(0.606445313f, 0.312894821f, 0.f),
    C_Vertex3f(0.606445313f, 0.312894821f, -1.f),
    C_Normal3f(-0.9994753f, -0.0323886536f, 0.f),
    C_Vertex3f(0.606445313f, 0.330760956f, 0.f),
    C_Vertex3f(0.606445313f, 0.330760956f, -1.f),
    C_Normal3f(-0.995231748f, -0.0975375697f, 0.f),
    C_Vertex3f(0.605291128f, 0.348511577f, 0.f),
    C_Vertex3f(0.605291128f, 0.348511577f, -1.f),
    C_Normal3f(-0.986669362f, -0.162737727f, 0.f),
    C_Vertex3f(0.602982759f, 0.366091371f, 0.f),
    C_Vertex3f(0.602982759f, 0.366091371f, -1.f),
    C_Normal3f(-0.973872483f, -0.22709541f, 0.f),
    C_Vertex3f(0.599520206f, 0.383500338f, 0.f),
    C_Vertex3f(0.599520206f, 0.383500338f, -1.f),
    C_Normal3f(-0.957085848f, -0.289804697f, 0.f),
    C_Vertex3f(0.594903469f, 0.400738478f, 0.f),
    C_Vertex3f(0.594903469f, 0.400738478f, -1.f),
    C_Normal3f(-0.93669498f, -0.350146502f, 0.f),
    C_Vertex3f(0.589132547f, 0.417805791f, 0.f),
    C_Vertex3f(0.589132547f, 0.417805791f, -1.f),
    C_Normal3f(-0.913193345f, -0.407526493f, 0.f),
    C_Vertex3f(0.582207441f, 0.434702277f, 0.f),
    C_Vertex3f(0.582207441f, 0.434702277f, -1.f),
    C_Normal3f(-0.887141407f, -0.461497724f, 0.f),
    C_Vertex3f(0.574128151f, 0.451427937f, 0.f),
    C_Vertex3f(0.574128151f, 0.451427937f, -1.f),
    C_Normal3f(-0.856513202f, -0.516125083f, 0.f),
    C_Vertex3f(0.564894676f, 0.467982769f, 0.f),
    C_Vertex3f(0.564894676f, 0.467982769f, -1.f),
    C_Normal3f(-0.818935633f, -0.573885381f, 0.f),
    C_Vertex3f(0.554752827f, 0.483581424f, 0.f),
    C_Vertex3f(0.554752827f, 0.483581424f, -1.f),
    C_Normal3f(-0.776225388f, -0.630455494f, 0.f),
    C_Vertex3f(0.543702602f, 0.498223901f, 0.f),
    C_Vertex3f(0.543702602f, 0.498223901f, -1.f),
    C_Normal3f(-0.728708327f, -0.684824109f, 0.f),
    C_Vertex3f(0.531744003f, 0.5119102f, 0.f),
    C_Vertex3f(0.531744003f, 0.5119102f, -1.f),
    C_Normal3f(-0.676953673f, -0.736025572f, 0.f),
    C_Vertex3f(0.518877029f, 0.524640322f, 0.f),
    C_Vertex3f(0.518877029f, 0.524640322f, -1.f),
    C_Normal3f(-0.621744752f, -0.783219934f, 0.f),
    C_Vertex3f(0.505101681f, 0.536414266f, 0.f),
    C_Vertex3f(0.505101681f, 0.536414266f, -1.f),
    C_Normal3f(-0.564020872f, -0.825760543f, 0.f),
    C_Vertex3f(0.490417957f, 0.547232032f, 0.f),
    C_Vertex3f(0.490417957f, 0.547232032f, -1.f),
    C_Normal3f(-0.504800618f, -0.86323607f, 0.f),
    C_Vertex3f(0.474825859f, 0.55709362f, 0.f),
    C_Vertex3f(0.474825859f, 0.55709362f, -1.f),
    C_Normal3f(-0.450436771f, -0.892808318f, 0.f),
    C_Vertex3f(0.458325386f, 0.565999031f, 0.f),
    C_Vertex3f(0.458325386f, 0.565999031f, -1.f),
    C_Normal3f(-0.398695141f, -0.917083561f, 0.f),
    C_Vertex3f(0.441729546f, 0.573791265f, 0.f),
    C_Vertex3f(0.441729546f, 0.573791265f, -1.f),
    C_Normal3f(-0.343506575f, -0.939150333f, 0.f),
    C_Vertex3f(0.425038338f, 0.580470324f, 0.f),
    C_Vertex3f(0.425038338f, 0.580470324f, -1.f),
    C_Normal3f(-0.285206378f, -0.958466113f, 0.f),
    C_Vertex3f(0.408251762f, 0.586036205f, 0.f),
    C_Vertex3f(0.408251762f, 0.586036205f, -1.f),
    C_Normal3f(-0.224273607f, -0.974526227f, 0.f),
    C_Vertex3f(0.39136982f, 0.590488911f, 0.f),
    C_Vertex3f(0.39136982f, 0.590488911f, -1.f),
    C_Normal3f(-0.161321044f, -0.986901939f, 0.f),
    C_Vertex3f(0.374392509f, 0.59382844f, 0.f),
    C_Vertex3f(0.374392509f, 0.59382844f, -1.f),
    C_Normal3f(-0.0970703289f, -0.995277524f, 0.f),
    C_Vertex3f(0.357319832f, 0.596054792f, 0.f),
    C_Vertex3f(0.357319832f, 0.596054792f, -1.f),
    C_Normal3f(-0.0322615579f, -0.999479473f, 0.f),
    C_Vertex3f(0.340151787f, 0.597167969f, 0.f),
    C_Vertex3f(0.340151787f, 0.597167969f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.186523438f, 0.139648438f, 0.f),
    C_Vertex3f(0.186523438f, 0.139648438f, -1.f),
    C_Vertex3f(0.186523438f, 0.494628906f, -1.f),
    C_Vertex3f(0.186523438f, 0.494628906f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.186523438f, 0.494628906f, 0.f),
    C_Vertex3f(0.186523438f, 0.494628906f, -1.f),
    C_Normal3f(0.00212429347f, 0.999997795f, 0.f),
    C_Vertex3f(0.316658974f, 0.494628906f, 0.f),
    C_Vertex3f(0.316658974f, 0.494628906f, -1.f),
    C_Normal3f(0.0328073986f, 0.999461651f, 0.f),
    C_Vertex3f(0.331249237f, 0.494321465f, 0.f),
    C_Vertex3f(0.331249237f, 0.494321465f, -1.f),
    C_Normal3f(0.0591715015f, 0.998247802f, 0.f),
    C_Vertex3f(0.34475708f, 0.493706584f, 0.f),
    C_Vertex3f(0.34475708f, 0.493706584f, -1.f),
    C_Normal3f(0.0901749954f, 0.995925963f, 0.f),
    C_Vertex3f(0.357182503f, 0.492784262f, 0.f),
    C_Vertex3f(0.357182503f, 0.492784262f, -1.f),
    C_Normal3f(0.127041295f, 0.991897464f, 0.f),
    C_Vertex3f(0.368525505f, 0.491554499f, 0.f),
    C_Vertex3f(0.368525505f, 0.491554499f, -1.f),
    C_Normal3f(0.17140007f, 0.985201478f, 0.f),
    C_Vertex3f(0.378786087f, 0.490017295f, 0.f),
    C_Vertex3f(0.378786087f, 0.490017295f, -1.f),
    C_Normal3f(0.225418791f, 0.974262059f, 0.f),
    C_Vertex3f(0.387964249f, 0.48817265f, 0.f),
    C_Vertex3f(0.387964249f, 0.48817265f, -1.f),
    C_Normal3f(0.299393624f, 0.954129636f, 0.f),
    C_Vertex3f(0.39605999f, 0.486020565f, 0.f),
    C_Vertex3f(0.39605999f, 0.486020565f, -1.f),
    C_Normal3f(0.411459327f, 0.911428154f, 0.f),
    C_Vertex3f(0.406579971f, 0.482331276f, 0.f),
    C_Vertex3f(0.406579971f, 0.482331276f, -1.f),
    C_Normal3f(0.541783631f, 0.840517998f, 0.f),
    C_Vertex3f(0.419182301f, 0.475582123f, 0.f),
    C_Vertex3f(0.419182301f, 0.475582123f, -1.f),
    C_Normal3f(0.67189306f, 0.74064815f, 0.f),
    C_Vertex3f(0.430360317f, 0.467002869f, 0.f),
    C_Vertex3f(0.430360317f, 0.467002869f, -1.f),
    C_Normal3f(0.781473577f, 0.623938322f, 0.f),
    C_Vertex3f(0.440114021f, 0.456593513f, 0.f),
    C_Vertex3f(0.440114021f, 0.456593513f, -1.f),
    C_Normal3f(0.862663686f, 0.505778134f, 0.f),
    C_Vertex3f(0.448443413f, 0.444354057f, 0.f),
    C_Vertex3f(0.448443413f, 0.444354057f, -1.f),
    C_Normal3f(0.926459968f, 0.376393199f, 0.f),
    C_Vertex3f(0.454690456f, 0.431731701f, 0.f),
    C_Vertex3f(0.454690456f, 0.431731701f, -1.f),
    C_Normal3f(0.973113537f, 0.230325744f, 0.f),
    C_Vertex3f(0.458855152f, 0.418726444f, 0.f),
    C_Vertex3f(0.458855152f, 0.418726444f, -1.f),
    C_Normal3f(0.9975999f, 0.0692423135f, 0.f),
    C_Vertex3f(0.4609375f, 0.405338287f, 0.f),
    C_Vertex3f(0.4609375f, 0.405338287f, -1.f),
    C_Normal3f(0.994998515f, -0.0998896509f, 0.f),
    C_Vertex3f(0.4609375f, 0.388725281f, 0.f),
    C_Vertex3f(0.4609375f, 0.388725281f, -1.f),
    C_Normal3f(0.958256841f, -0.285908818f, 0.f),
    C_Vertex3f(0.45744276f, 0.370527267f, 0.f),
    C_Vertex3f(0.45744276f, 0.370527267f, -1.f),
    C_Normal3f(0.881525934f, -0.472135574f, 0.f),
    C_Vertex3f(0.450453281f, 0.353586197f, 0.f),
    C_Vertex3f(0.450453281f, 0.353586197f, -1.f),
    C_Normal3f(0.78425169f, -0.620442688f, 0.f),
    C_Vertex3f(0.439969063f, 0.337902069f, 0.f),
    C_Vertex3f(0.439969063f, 0.337902069f, -1.f),
    C_Normal3f(0.690829337f, -0.723017871f, 0.f),
    C_Vertex3f(0.429484844f, 0.32708168f, 0.f),
    C_Vertex3f(0.429484844f, 0.32708168f, -1.f),
    C_Normal3f(0.609032094f, -0.793145597f, 0.f),
    C_Vertex3f(0.421924233f, 0.320660591f, 0.f),
    C_Vertex3f(0.421924233f, 0.320660591f, -1.f),
    C_Normal3f(0.527930737f, -0.84928745f, 0.f),
    C_Vertex3f(0.413792491f, 0.315032005f, 0.f),
    C_Vertex3f(0.413792491f, 0.315032005f, -1.f),
    C_Normal3f(0.442868143f, -0.896586835f, 0.f),
    C_Vertex3f(0.405089617f, 0.310195923f, 0.f),
    C_Vertex3f(0.405089617f, 0.310195923f, -1.f),
    C_Normal3f(0.356472015f, -0.934305966f, 0.f),
    C_Vertex3f(0.395815611f, 0.306152344f, 0.f),
    C_Vertex3f(0.395815611f, 0.306152344f, -1.f),
    C_Normal3f(0.271235377f, -0.96251297f, 0.f),
    C_Vertex3f(0.385970473f, 0.302901268f, 0.f),
    C_Vertex3f(0.385970473f, 0.302901268f, -1.f),
    C_Normal3f(0.189225674f, -0.981933534f, 0.f),
    C_Vertex3f(0.375554204f, 0.300442696f, 0.f),
    C_Vertex3f(0.375554204f, 0.300442696f, -1.f),
    C_Normal3f(0.12444368f, -0.99222672f, 0.f),
    C_Vertex3f(0.364566803f, 0.298776627f, 0.f),
    C_Vertex3f(0.364566803f, 0.298776627f, -1.f),
    C_Normal3f(0.0753626302f, -0.997156143f, 0.f),
    C_Vertex3f(0.358787537f, 0.298339844f, 0.f),
    C_Vertex3f(0.358787537f, 0.298339844f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.385597736f, 0.922666967f, 0.f),
    C_Vertex3f(0.358787537f, 0.298339844f, 0.f),
    C_Vertex3f(0.358787537f, 0.298339844f, -1.f),
    C_Normal3f(0.430131227f, 0.902766287f, 0.f),
    C_Vertex3f(0.363492966f, 0.296373367f, 0.f),
    C_Vertex3f(0.363492966f, 0.296373367f, -1.f),
    C_Normal3f(0.490175068f, 0.871623933f, 0.f),
    C_Vertex3f(0.371973991f, 0.292057037f, 0.f),
    C_Vertex3f(0.371973991f, 0.292057037f, -1.f),
    C_Normal3f(0.568072021f, 0.822978973f, 0.f),
    C_Vertex3f(0.379525185f, 0.28735733f, 0.f),
    C_Vertex3f(0.379525185f, 0.28735733f, -1.f),
    C_Normal3f(0.636670649f, 0.771135867f, 0.f),
    C_Vertex3f(0.386146545f, 0.282274246f, 0.f),
    C_Vertex3f(0.386146545f, 0.282274246f, -1.f),
    C_Normal3f(0.696954668f, 0.717115164f, 0.f),
    C_Vertex3f(0.38899231f, 0.279541016f, 0.f),
    C_Vertex3f(0.38899231f, 0.279541016f, -1.f),
    C_Normal3f(0.74456495f, 0.667549908f, 0.f),
    C_Vertex3f(0.394355774f, 0.274295807f, 0.f),
    C_Vertex3f(0.394355774f, 0.274295807f, -1.f),
    C_Normal3f(0.783868015f, 0.620927393f, 0.f),
    C_Vertex3f(0.406285286f, 0.260252953f, 0.f),
    C_Vertex3f(0.406285286f, 0.260252953f, -1.f),
    C_Normal3f(0.815788031f, 0.578351021f, 0.f),
    C_Vertex3f(0.419417381f, 0.242657661f, 0.f),
    C_Vertex3f(0.419417381f, 0.242657661f, -1.f),
    C_Normal3f(0.848241389f, 0.529609799f, 0.f),
    C_Vertex3f(0.43375206f, 0.221509933f, 0.f),
    C_Vertex3f(0.43375206f, 0.221509933f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.848241389f, 0.529609799f, 0.f),
    C_Vertex3f(0.43375206f, 0.221509933f, 0.f),
    C_Vertex3f(0.43375206f, 0.221509933f, -1.f),
    C_Vertex3f(0.484863281f, 0.139648438f, -1.f),
    C_Vertex3f(0.484863281f, 0.139648438f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.484863281f, 0.139648438f, 0.f),
    C_Vertex3f(0.484863281f, 0.139648438f, -1.f),
    C_Vertex3f(0.414680481f, 0.139648438f, -1.f),
    C_Vertex3f(0.414680481f, 0.139648438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.871156931f, -0.491004676f, 0.f),
    C_Vertex3f(0.414680481f, 0.139648438f, 0.f),
    C_Vertex3f(0.414680481f, 0.139648438f, -1.f),
    C_Normal3f(-0.869124293f, -0.49459374f, 0.f),
    C_Vertex3f(0.378504753f, 0.203832626f, 0.f),
    C_Vertex3f(0.378504753f, 0.203832626f, -1.f),
    C_Normal3f(-0.854305625f, -0.51977098f, 0.f),
    C_Vertex3f(0.36966157f, 0.218757868f, 0.f),
    C_Vertex3f(0.36966157f, 0.218757868f, -1.f),
    C_Normal3f(-0.839956462f, -0.542653859f, 0.f),
    C_Vertex3f(0.361277342f, 0.232147932f, 0.f),
    C_Vertex3f(0.361277342f, 0.232147932f, -1.f),
    C_Normal3f(-0.82149899f, -0.57020992f, 0.f),
    C_Vertex3f(0.35335207f, 0.244002819f, 0.f),
    C_Vertex3f(0.35335207f, 0.244002819f, -1.f),
    C_Normal3f(-0.797076762f, -0.603877962f, 0.f),
    C_Vertex3f(0.345885754f, 0.254322529f, 0.f),
    C_Vertex3f(0.345885754f, 0.254322529f, -1.f),
    C_Normal3f(-0.763653278f, -0.645626605f, 0.f),
    C_Vertex3f(0.338878393f, 0.263107061f, 0.f),
    C_Vertex3f(0.338878393f, 0.263107061f, -1.f),
    C_Normal3f(-0.716044128f, -0.698055148f, 0.f),
    C_Vertex3f(0.332329988f, 0.270356417f, 0.f),
    C_Vertex3f(0.332329988f, 0.270356417f, -1.f),
    C_Normal3f(-0.659081757f, -0.752071261f, 0.f),
    C_Vertex3f(0.32624054f, 0.276070595f, 0.f),
    C_Vertex3f(0.32624054f, 0.276070595f, -1.f),
    C_Normal3f(-0.607362747f, -0.794424653f, 0.f),
    C_Vertex3f(0.323425293f, 0.278160095f, 0.f),
    C_Vertex3f(0.323425293f, 0.278160095f, -1.f),
    C_Normal3f(-0.491694152f, -0.870768011f, 0.f),
    C_Vertex3f(0.319480896f, 0.281238556f, 0.f),
    C_Vertex3f(0.319480896f, 0.281238556f, -1.f),
    C_Normal3f(-0.341990799f, -0.939703286f, 0.f),
    C_Vertex3f(0.309783936f, 0.285862923f, 0.f),
    C_Vertex3f(0.309783936f, 0.285862923f, -1.f),
    C_Normal3f(-0.184197322f, -0.982889295f, 0.f),
    C_Vertex3f(0.298278809f, 0.288954735f, 0.f),
    C_Vertex3f(0.298278809f, 0.288954735f, -1.f),
    C_Normal3f(-0.000323176326f, -0.999999881f, 0.f),
    C_Vertex3f(0.284965515f, 0.290513992f, 0.f),
    C_Vertex3f(0.284965515f, 0.290513992f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.000323176326f, -0.999999881f, 0.f),
    C_Vertex3f(0.284965515f, 0.290513992f, 0.f),
    C_Vertex3f(0.284965515f, 0.290513992f, -1.f),
    C_Vertex3f(0.243652344f, 0.290527344f, -1.f),
    C_Vertex3f(0.243652344f, 0.290527344f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.243652344f, 0.290527344f, 0.f),
    C_Vertex3f(0.243652344f, 0.290527344f, -1.f),
    C_Vertex3f(0.243652344f, 0.139648438f, -1.f),
    C_Vertex3f(0.243652344f, 0.139648438f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.243652344f, 0.139648438f, 0.f),
    C_Vertex3f(0.243652344f, 0.139648438f, -1.f),
    C_Vertex3f(0.186523438f, 0.139648438f, -1.f),
    C_Vertex3f(0.186523438f, 0.139648438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.243652344f, 0.339355469f, 0.f),
    C_Vertex3f(0.243652344f, 0.339355469f, -1.f),
    C_Normal3f(-0.00530415867f, 0.999985933f, 0.f),
    C_Vertex3f(0.319932938f, 0.339355469f, 0.f),
    C_Vertex3f(0.319932938f, 0.339355469f, -1.f),
    C_Normal3f(-0.0631108657f, 0.998006523f, 0.f),
    C_Vertex3f(0.331482172f, 0.339821339f, 0.f),
    C_Vertex3f(0.331482172f, 0.339821339f, -1.f),
    C_Normal3f(-0.115080558f, 0.993356049f, 0.f),
    C_Vertex3f(0.342034101f, 0.340753078f, 0.f),
    C_Vertex3f(0.342034101f, 0.340753078f, -1.f),
    C_Normal3f(-0.177202433f, 0.984174371f, 0.f),
    C_Vertex3f(0.351588726f, 0.342150688f, 0.f),
    C_Vertex3f(0.351588726f, 0.342150688f, -1.f),
    C_Normal3f(-0.251764476f, 0.967788577f, 0.f),
    C_Vertex3f(0.360146046f, 0.344014168f, 0.f),
    C_Vertex3f(0.360146046f, 0.344014168f, -1.f),
    C_Normal3f(-0.341098219f, 0.940027595f, 0.f),
    C_Vertex3f(0.36770606f, 0.346343517f, 0.f),
    C_Vertex3f(0.36770606f, 0.346343517f, -1.f),
    C_Normal3f(-0.446756274f, 0.894655704f, 0.f),
    C_Vertex3f(0.37426877f, 0.349138737f, 0.f),
    C_Vertex3f(0.37426877f, 0.349138737f, -1.f),
    C_Normal3f(-0.580449581f, 0.814296067f, 0.f),
    C_Vertex3f(0.379834175f, 0.352399826f, 0.f),
    C_Vertex3f(0.379834175f, 0.352399826f, -1.f),
    C_Normal3f(-0.720768154f, 0.69317615f, 0.f),
    C_Vertex3f(0.386686325f, 0.357990265f, 0.f),
    C_Vertex3f(0.386686325f, 0.357990265f, -1.f),
    C_Normal3f(-0.851232171f, 0.524789214f, 0.f),
    C_Vertex3f(0.393538475f, 0.366649628f, 0.f),
    C_Vertex3f(0.393538475f, 0.366649628f, -1.f),
    C_Normal3f(-0.950388014f, 0.311066777f, 0.f),
    C_Vertex3f(0.398106575f, 0.376514435f, 0.f),
    C_Vertex3f(0.398106575f, 0.376514435f, -1.f),
    C_Normal3f(-0.994221628f, 0.107346535f, 0.f),
    C_Vertex3f(0.400390625f, 0.387584686f, 0.f),
    C_Vertex3f(0.400390625f, 0.387584686f, -1.f),
    C_Normal3f(-0.99804306f, -0.0625300854f, 0.f),
    C_Vertex3f(0.400390625f, 0.397668839f, 0.f),
    C_Vertex3f(0.400390625f, 0.397668839f, -1.f),
    C_Normal3f(-0.975215077f, -0.221259221f, 0.f),
    C_Vertex3f(0.399290085f, 0.405150414f, 0.f),
    C_Vertex3f(0.399290085f, 0.405150414f, -1.f),
    C_Normal3f(-0.928227484f, -0.372013092f, 0.f),
    C_Vertex3f(0.397089005f, 0.412220955f, 0.f),
    C_Vertex3f(0.397089005f, 0.412220955f, -1.f),
    C_Normal3f(-0.858694136f, -0.512488484f, 0.f),
    C_Vertex3f(0.393787384f, 0.418880463f, 0.f),
    C_Vertex3f(0.393787384f, 0.418880463f, -1.f),
    C_Normal3f(-0.771254838f, -0.636526465f, 0.f),
    C_Vertex3f(0.389385223f, 0.425128937f, 0.f),
    C_Vertex3f(0.389385223f, 0.425128937f, -1.f),
    C_Normal3f(-0.664735317f, -0.747078896f, 0.f),
    C_Vertex3f(0.384129524f, 0.430582523f, 0.f),
    C_Vertex3f(0.384129524f, 0.430582523f, -1.f),
    C_Normal3f(-0.546145618f, -0.837690175f, 0.f),
    C_Vertex3f(0.378020287f, 0.435241222f, 0.f),
    C_Vertex3f(0.378020287f, 0.435241222f, -1.f),
    C_Normal3f(-0.42468679f, -0.905340314f, 0.f),
    C_Vertex3f(0.37105751f, 0.439105034f, 0.f),
    C_Vertex3f(0.37105751f, 0.439105034f, -1.f),
    C_Normal3f(-0.272512227f, -0.962152362f, 0.f),
    C_Vertex3f(0.363241196f, 0.442173958f, 0.f),
    C_Vertex3f(0.363241196f, 0.442173958f, -1.f),
    C_Normal3f(-0.148079425f, -0.988975465f, 0.f),
    C_Vertex3f(0.352095604f, 0.444475651f, 0.f),
    C_Vertex3f(0.352095604f, 0.444475651f, -1.f),
    C_Normal3f(-0.0711255819f, -0.997467399f, 0.f),
    C_Vertex3f(0.337620735f, 0.446010113f, 0.f),
    C_Vertex3f(0.337620735f, 0.446010113f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.319816589f, 0.446777344f, 0.f),
    C_Vertex3f(0.319816589f, 0.446777344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.319816589f, 0.446777344f, 0.f),
    C_Vertex3f(0.319816589f, 0.446777344f, -1.f),
    C_Vertex3f(0.243652344f, 0.446777344f, -1.f),
    C_Vertex3f(0.243652344f, 0.446777344f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.243652344f, 0.446777344f, 0.f),
    C_Vertex3f(0.243652344f, 0.446777344f, -1.f),
    C_Vertex3f(0.243652344f, 0.339355469f, -1.f),
    C_Vertex3f(0.243652344f, 0.339355469f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.661132813f, 0.311144829f, 0.f),
    C_Vertex3f(0.661132813f, 0.332565308f, 0.f),
    C_Vertex3f(0.659773588f, 0.290038824f, 0.f),
    C_Vertex3f(0.659753084f, 0.353883743f, 0.f),
    C_Vertex3f(0.65705514f, 0.269137621f, 0.f),
    C_Vertex3f(0.656993628f, 0.374990463f, 0.f),
    C_Vertex3f(0.652977467f, 0.248441219f, 0.f),
    C_Vertex3f(0.652854443f, 0.395885468f, 0.f),
    C_Vertex3f(0.647540569f, 0.227949619f, 0.f),
    C_Vertex3f(0.647335529f, 0.416568756f, 0.f),
    C_Vertex3f(0.640744448f, 0.207662821f, 0.f),
    C_Vertex3f(0.640436888f, 0.437040329f, 0.f),
    C_Vertex3f(0.632589102f, 0.187580824f, 0.f),
    C_Vertex3f(0.632158518f, 0.457300186f, 0.f),
    C_Vertex3f(0.623074532f, 0.167703629f, 0.f),
    C_Vertex3f(0.62250042f, 0.477348328f, 0.f),
    C_Vertex3f(0.612200737f, 0.148031235f, 0.f),
    C_Vertex3f(0.611462593f, 0.497184753f, 0.f),
    C_Vertex3f(0.606445313f, 0.330760956f, 0.f),
    C_Vertex3f(0.605291128f, 0.348511577f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.589132547f, 0.417805791f, 0.f),
    C_Vertex3f(0.594903469f, 0.400738478f, 0.f),
    C_Vertex3f(0.599331737f, 0.515873551f, 0.f),
    C_Vertex3f(0.599520206f, 0.383500338f, 0.f),
    C_Vertex3f(0.611462593f, 0.497184753f, 0.f),
    C_Vertex3f(0.602982759f, 0.366091371f, 0.f),
    C_Vertex3f(0.605291128f, 0.348511577f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.58610785f, 0.533414721f, 0.f),
    C_Vertex3f(0.571790934f, 0.549808264f, 0.f),
    C_Vertex3f(0.574128151f, 0.451427937f, 0.f),
    C_Vertex3f(0.582207441f, 0.434702277f, 0.f),
    C_Vertex3f(0.589132547f, 0.417805791f, 0.f),
    C_Vertex3f(0.599331737f, 0.515873551f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.543702602f, 0.498223901f, 0.f),
    C_Vertex3f(0.554752827f, 0.483581424f, 0.f),
    C_Vertex3f(0.556380987f, 0.565054178f, 0.f),
    C_Vertex3f(0.564894676f, 0.467982769f, 0.f),
    C_Vertex3f(0.571790934f, 0.549808264f, 0.f),
    C_Vertex3f(0.574128151f, 0.451427937f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.505101681f, 0.536414266f, 0.f),
    C_Vertex3f(0.518877029f, 0.524640322f, 0.f),
    C_Vertex3f(0.522282004f, 0.592103124f, 0.f),
    C_Vertex3f(0.531744003f, 0.5119102f, 0.f),
    C_Vertex3f(0.539878011f, 0.579152465f, 0.f),
    C_Vertex3f(0.543702602f, 0.498223901f, 0.f),
    C_Vertex3f(0.556380987f, 0.565054178f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.425038338f, 0.580470324f, 0.f),
    C_Vertex3f(0.441729546f, 0.573791265f, 0.f),
    C_Vertex3f(0.44387877f, 0.631876588f, 0.f),
    C_Vertex3f(0.458325386f, 0.565999031f, 0.f),
    C_Vertex3f(0.463906169f, 0.623885036f, 0.f),
    C_Vertex3f(0.474825859f, 0.55709362f, 0.f),
    C_Vertex3f(0.483810902f, 0.614561558f, 0.f),
    C_Vertex3f(0.490417957f, 0.547232032f, 0.f),
    C_Vertex3f(0.503592968f, 0.603906155f, 0.f),
    C_Vertex3f(0.505101681f, 0.536414266f, 0.f),
    C_Vertex3f(0.522282004f, 0.592103124f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.322833061f, 0.597167969f, 0.f),
    C_Vertex3f(0.340151787f, 0.597167969f, 0.f),
    C_Vertex3f(0.341901779f, 0.651855469f, 0.f),
    C_Vertex3f(0.357319832f, 0.596054792f, 0.f),
    C_Vertex3f(0.36254251f, 0.650523543f, 0.f),
    C_Vertex3f(0.374392509f, 0.59382844f, 0.f),
    C_Vertex3f(0.383060575f, 0.647859693f, 0.f),
    C_Vertex3f(0.39136982f, 0.590488911f, 0.f),
    C_Vertex3f(0.403455973f, 0.643863916f, 0.f),
    C_Vertex3f(0.408251762f, 0.586036205f, 0.f),
    C_Vertex3f(0.423728704f, 0.638536215f, 0.f),
    C_Vertex3f(0.425038338f, 0.580470324f, 0.f),
    C_Vertex3f(0.44387877f, 0.631876588f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.221024632f, 0.573791265f, 0.f),
    C_Vertex3f(0.237702727f, 0.580470324f, 0.f),
    C_Vertex3f(0.23930645f, 0.638536215f, 0.f),
    C_Vertex3f(0.254496813f, 0.586036205f, 0.f),
    C_Vertex3f(0.259579659f, 0.643863916f, 0.f),
    C_Vertex3f(0.271406889f, 0.590488911f, 0.f),
    C_Vertex3f(0.279975891f, 0.647859693f, 0.f),
    C_Vertex3f(0.288432956f, 0.59382844f, 0.f),
    C_Vertex3f(0.300495148f, 0.650523543f, 0.f),
    C_Vertex3f(0.305575013f, 0.596054792f, 0.f),
    C_Vertex3f(0.321137428f, 0.651855469f, 0.f),
    C_Vertex3f(0.322833061f, 0.597167969f, 0.f),
    C_Vertex3f(0.341901779f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.144046783f, 0.524640322f, 0.f),
    C_Vertex3f(0.157808542f, 0.536414266f, 0.f),
    C_Vertex3f(0.159443855f, 0.603906155f, 0.f),
    C_Vertex3f(0.172465086f, 0.547232032f, 0.f),
    C_Vertex3f(0.179224968f, 0.614561558f, 0.f),
    C_Vertex3f(0.188016415f, 0.55709362f, 0.f),
    C_Vertex3f(0.199129105f, 0.623885036f, 0.f),
    C_Vertex3f(0.204462528f, 0.565999031f, 0.f),
    C_Vertex3f(0.219156265f, 0.631876588f, 0.f),
    C_Vertex3f(0.221024632f, 0.573791265f, 0.f),
    C_Vertex3f(0.23930645f, 0.638536215f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.108130217f, 0.483581424f, 0.f),
    C_Vertex3f(0.119207621f, 0.498223901f, 0.f),
    C_Vertex3f(0.12313962f, 0.579152465f, 0.f),
    C_Vertex3f(0.13117981f, 0.5119102f, 0.f),
    C_Vertex3f(0.140748739f, 0.592103124f, 0.f),
    C_Vertex3f(0.144046783f, 0.524640322f, 0.f),
    C_Vertex3f(0.159443855f, 0.603906155f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0805329084f, 0.434702277f, 0.f),
    C_Vertex3f(0.0886597633f, 0.451427937f, 0.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, 0.f),
    C_Vertex3f(0.0979475975f, 0.467982769f, 0.f),
    C_Vertex3f(0.106616497f, 0.565054178f, 0.f),
    C_Vertex3f(0.108130217f, 0.483581424f, 0.f),
    C_Vertex3f(0.12313962f, 0.579152465f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0768282413f, 0.533414721f, 0.f),
    C_Vertex3f(0.0635631084f, 0.515873551f, 0.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, 0.f),
    C_Vertex3f(0.0735670328f, 0.417805791f, 0.f),
    C_Vertex3f(0.0805329084f, 0.434702277f, 0.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0635631084f, 0.515873551f, 0.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, 0.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, 0.f),
    C_Vertex3f(0.057313323f, 0.348511577f, 0.f),
    C_Vertex3f(0.0596352816f, 0.366091371f, 0.f),
    C_Vertex3f(0.0631182194f, 0.383500338f, 0.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.00146484375f, 0.332565308f, 0.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, 0.f),
    C_Vertex3f(0.00285148621f, 0.353883743f, 0.f),
    C_Vertex3f(0.00556325912f, 0.269137621f, 0.f),
    C_Vertex3f(0.00562477112f, 0.374990463f, 0.f),
    C_Vertex3f(0.0096616745f, 0.248441219f, 0.f),
    C_Vertex3f(0.00978469849f, 0.395885468f, 0.f),
    C_Vertex3f(0.0151262283f, 0.227949619f, 0.f),
    C_Vertex3f(0.0153312683f, 0.416568756f, 0.f),
    C_Vertex3f(0.0219569206f, 0.207662821f, 0.f),
    C_Vertex3f(0.0222644806f, 0.437040329f, 0.f),
    C_Vertex3f(0.0301537514f, 0.187580824f, 0.f),
    C_Vertex3f(0.0305843353f, 0.457300186f, 0.f),
    C_Vertex3f(0.0397167206f, 0.167703629f, 0.f),
    C_Vertex3f(0.0402908325f, 0.477348328f, 0.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, 0.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, 0.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, 0.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.612200737f, 0.148031235f, 0.f),
    C_Vertex3f(0.606445313f, 0.330760956f, 0.f),
    C_Vertex3f(0.606445313f, 0.312894821f, 0.f),
    C_Vertex3f(0.605311632f, 0.295268297f, 0.f),
    C_Vertex3f(0.603044271f, 0.277826071f, 0.f),
    C_Vertex3f(0.600233912f, 0.129458666f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.600233912f, 0.129458666f, 0.f),
    C_Vertex3f(0.603044271f, 0.277826071f, 0.f),
    C_Vertex3f(0.59964323f, 0.260568142f, 0.f),
    C_Vertex3f(0.595108509f, 0.243494511f, 0.f),
    C_Vertex3f(0.589440107f, 0.226605177f, 0.f),
    C_Vertex3f(0.587174058f, 0.111985922f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.589440107f, 0.226605177f, 0.f),
    C_Vertex3f(0.582638025f, 0.209900141f, 0.f),
    C_Vertex3f(0.587174058f, 0.111985922f, 0.f),
    C_Vertex3f(0.574702263f, 0.193379402f, 0.f),
    C_Vertex3f(0.573021173f, 0.0956130028f, 0.f),
    C_Vertex3f(0.56563282f, 0.177042961f, 0.f),
    C_Vertex3f(0.557775259f, 0.0803399086f, 0.f),
    C_Vertex3f(0.555648208f, 0.1616081f, 0.f),
    C_Vertex3f(0.544748425f, 0.147074819f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.557775259f, 0.0803399086f, 0.f),
    C_Vertex3f(0.544748425f, 0.147074819f, 0.f),
    C_Vertex3f(0.541436315f, 0.0661666393f, 0.f),
    C_Vertex3f(0.532933474f, 0.133443117f, 0.f),
    C_Vertex3f(0.52400434f, 0.053093195f, 0.f),
    C_Vertex3f(0.520203352f, 0.120712996f, 0.f),
    C_Vertex3f(0.506558061f, 0.108884454f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52400434f, 0.053093195f, 0.f),
    C_Vertex3f(0.506558061f, 0.108884454f, 0.f),
    C_Vertex3f(0.505479336f, 0.0411195755f, 0.f),
    C_Vertex3f(0.4919976f, 0.0979574919f, 0.f),
    C_Vertex3f(0.485861301f, 0.0302457809f, 0.f),
    C_Vertex3f(0.476521969f, 0.0879321098f, 0.f),
    C_Vertex3f(0.466018081f, 0.0207312107f, 0.f),
    C_Vertex3f(0.460131168f, 0.0788083076f, 0.f),
    C_Vertex3f(0.445949674f, 0.0125758648f, 0.f),
    C_Vertex3f(0.443562865f, 0.0708249807f, 0.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445949674f, 0.0125758648f, 0.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, 0.f),
    C_Vertex3f(0.42565608f, 0.00577974319f, 0.f),
    C_Vertex3f(0.409893751f, 0.0582797527f, 0.f),
    C_Vertex3f(0.4051373f, 0.000342845917f, 0.f),
    C_Vertex3f(0.39279294f, 0.0537178516f, 0.f),
    C_Vertex3f(0.384393334f, -0.00373482704f, 0.f),
    C_Vertex3f(0.375514627f, 0.0502964258f, 0.f),
    C_Vertex3f(0.363424182f, -0.00645327568f, 0.f),
    C_Vertex3f(0.35805881f, 0.0480154753f, 0.f),
    C_Vertex3f(0.342229843f, -0.0078125f, 0.f),
    C_Vertex3f(0.340425491f, 0.046875f, 0.f),
    C_Vertex3f(0.322613716f, 0.046875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.342229843f, -0.0078125f, 0.f),
    C_Vertex3f(0.322613716f, 0.046875f, 0.f),
    C_Vertex3f(0.320809364f, -0.0078125f, 0.f),
    C_Vertex3f(0.304978728f, 0.0480154753f, 0.f),
    C_Vertex3f(0.299613476f, -0.00645327568f, 0.f),
    C_Vertex3f(0.287521482f, 0.0502964258f, 0.f),
    C_Vertex3f(0.278643131f, -0.00373482704f, 0.f),
    C_Vertex3f(0.270241976f, 0.0537178516f, 0.f),
    C_Vertex3f(0.257898331f, 0.000342845917f, 0.f),
    C_Vertex3f(0.253140211f, 0.0582797527f, 0.f),
    C_Vertex3f(0.237379074f, 0.00577974319f, 0.f),
    C_Vertex3f(0.236216187f, 0.0639821291f, 0.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.237379074f, 0.00577974319f, 0.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, 0.f),
    C_Vertex3f(0.217085361f, 0.0125758648f, 0.f),
    C_Vertex3f(0.202901363f, 0.0788083076f, 0.f),
    C_Vertex3f(0.197017193f, 0.0207312107f, 0.f),
    C_Vertex3f(0.186510563f, 0.0879321098f, 0.f),
    C_Vertex3f(0.177174568f, 0.0302457809f, 0.f),
    C_Vertex3f(0.171028137f, 0.0979574919f, 0.f),
    C_Vertex3f(0.157557487f, 0.0411195755f, 0.f),
    C_Vertex3f(0.156454086f, 0.108884454f, 0.f),
    C_Vertex3f(0.14278841f, 0.120712996f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157557487f, 0.0411195755f, 0.f),
    C_Vertex3f(0.14278841f, 0.120712996f, 0.f),
    C_Vertex3f(0.139026403f, 0.053093195f, 0.f),
    C_Vertex3f(0.130031109f, 0.133443117f, 0.f),
    C_Vertex3f(0.121581316f, 0.0661666393f, 0.f),
    C_Vertex3f(0.118182182f, 0.147074819f, 0.f),
    C_Vertex3f(0.107241631f, 0.1616081f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.121581316f, 0.0661666393f, 0.f),
    C_Vertex3f(0.107241631f, 0.1616081f, 0.f),
    C_Vertex3f(0.105222225f, 0.0803399086f, 0.f),
    C_Vertex3f(0.0972094536f, 0.177042961f, 0.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, 0.f),
    C_Vertex3f(0.0880856514f, 0.193379402f, 0.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0757620335f, 0.111985922f, 0.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, 0.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, 0.f),
    C_Vertex3f(0.0732594728f, 0.226605177f, 0.f),
    C_Vertex3f(0.0675570965f, 0.243494511f, 0.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, 0.f),
    C_Vertex3f(0.0626609325f, 0.129458666f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0626609325f, 0.129458666f, 0.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, 0.f),
    C_Vertex3f(0.0595737696f, 0.277826071f, 0.f),
    C_Vertex3f(0.057292819f, 0.295268297f, 0.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, 0.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.448443413f, 0.444354057f, 0.f),
    C_Vertex3f(0.450453281f, 0.353586197f, 0.f),
    C_Vertex3f(0.454690456f, 0.431731701f, 0.f),
    C_Vertex3f(0.45744276f, 0.370527267f, 0.f),
    C_Vertex3f(0.458855152f, 0.418726444f, 0.f),
    C_Vertex3f(0.4609375f, 0.388725281f, 0.f),
    C_Vertex3f(0.4609375f, 0.405338287f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.421924233f, 0.320660591f, 0.f),
    C_Vertex3f(0.429484844f, 0.32708168f, 0.f),
    C_Vertex3f(0.430360317f, 0.467002869f, 0.f),
    C_Vertex3f(0.439969063f, 0.337902069f, 0.f),
    C_Vertex3f(0.440114021f, 0.456593513f, 0.f),
    C_Vertex3f(0.450453281f, 0.353586197f, 0.f),
    C_Vertex3f(0.448443413f, 0.444354057f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.400390625f, 0.397668839f, 0.f),
    C_Vertex3f(0.405089617f, 0.310195923f, 0.f),
    C_Vertex3f(0.406579971f, 0.482331276f, 0.f),
    C_Vertex3f(0.413792491f, 0.315032005f, 0.f),
    C_Vertex3f(0.419182301f, 0.475582123f, 0.f),
    C_Vertex3f(0.421924233f, 0.320660591f, 0.f),
    C_Vertex3f(0.430360317f, 0.467002869f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389385223f, 0.425128937f, 0.f),
    C_Vertex3f(0.393787384f, 0.418880463f, 0.f),
    C_Vertex3f(0.39605999f, 0.486020565f, 0.f),
    C_Vertex3f(0.397089005f, 0.412220955f, 0.f),
    C_Vertex3f(0.406579971f, 0.482331276f, 0.f),
    C_Vertex3f(0.399290085f, 0.405150414f, 0.f),
    C_Vertex3f(0.400390625f, 0.397668839f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.37105751f, 0.439105034f, 0.f),
    C_Vertex3f(0.378020287f, 0.435241222f, 0.f),
    C_Vertex3f(0.378786087f, 0.490017295f, 0.f),
    C_Vertex3f(0.384129524f, 0.430582523f, 0.f),
    C_Vertex3f(0.387964249f, 0.48817265f, 0.f),
    C_Vertex3f(0.389385223f, 0.425128937f, 0.f),
    C_Vertex3f(0.39605999f, 0.486020565f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186523438f, 0.139648438f, 0.f),
    C_Vertex3f(0.243652344f, 0.139648438f, 0.f),
    C_Vertex3f(0.186523438f, 0.494628906f, 0.f),
    C_Vertex3f(0.243652344f, 0.446777344f, 0.f),
    C_Vertex3f(0.316658974f, 0.494628906f, 0.f),
    C_Vertex3f(0.319816589f, 0.446777344f, 0.f),
    C_Vertex3f(0.331249237f, 0.494321465f, 0.f),
    C_Vertex3f(0.337620735f, 0.446010113f, 0.f),
    C_Vertex3f(0.34475708f, 0.493706584f, 0.f),
    C_Vertex3f(0.352095604f, 0.444475651f, 0.f),
    C_Vertex3f(0.357182503f, 0.492784262f, 0.f),
    C_Vertex3f(0.363241196f, 0.442173958f, 0.f),
    C_Vertex3f(0.368525505f, 0.491554499f, 0.f),
    C_Vertex3f(0.37105751f, 0.439105034f, 0.f),
    C_Vertex3f(0.378786087f, 0.490017295f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243652344f, 0.446777344f, 0.f),
    C_Vertex3f(0.243652344f, 0.290527344f, 0.f),
    C_Vertex3f(0.243652344f, 0.339355469f, 0.f),
    C_Vertex3f(0.284965515f, 0.290513992f, 0.f),
    C_Vertex3f(0.319932938f, 0.339355469f, 0.f),
    C_Vertex3f(0.298278809f, 0.288954735f, 0.f),
    C_Vertex3f(0.309783936f, 0.285862923f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.400390625f, 0.397668839f, 0.f),
    C_Vertex3f(0.400390625f, 0.387584686f, 0.f),
    C_Vertex3f(0.405089617f, 0.310195923f, 0.f),
    C_Vertex3f(0.398106575f, 0.376514435f, 0.f),
    C_Vertex3f(0.395815611f, 0.306152344f, 0.f),
    C_Vertex3f(0.393538475f, 0.366649628f, 0.f),
    C_Vertex3f(0.386686325f, 0.357990265f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.395815611f, 0.306152344f, 0.f),
    C_Vertex3f(0.386686325f, 0.357990265f, 0.f),
    C_Vertex3f(0.385970473f, 0.302901268f, 0.f),
    C_Vertex3f(0.379834175f, 0.352399826f, 0.f),
    C_Vertex3f(0.375554204f, 0.300442696f, 0.f),
    C_Vertex3f(0.37426877f, 0.349138737f, 0.f),
    C_Vertex3f(0.36770606f, 0.346343517f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.375554204f, 0.300442696f, 0.f),
    C_Vertex3f(0.36770606f, 0.346343517f, 0.f),
    C_Vertex3f(0.364566803f, 0.298776627f, 0.f),
    C_Vertex3f(0.360146046f, 0.344014168f, 0.f),
    C_Vertex3f(0.358787537f, 0.298339844f, 0.f),
    C_Vertex3f(0.35335207f, 0.244002819f, 0.f),
    C_Vertex3f(0.361277342f, 0.232147932f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.332329988f, 0.270356417f, 0.f),
    C_Vertex3f(0.338878393f, 0.263107061f, 0.f),
    C_Vertex3f(0.342034101f, 0.340753078f, 0.f),
    C_Vertex3f(0.345885754f, 0.254322529f, 0.f),
    C_Vertex3f(0.351588726f, 0.342150688f, 0.f),
    C_Vertex3f(0.35335207f, 0.244002819f, 0.f),
    C_Vertex3f(0.360146046f, 0.344014168f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.331482172f, 0.339821339f, 0.f),
    C_Vertex3f(0.319932938f, 0.339355469f, 0.f),
    C_Vertex3f(0.323425293f, 0.278160095f, 0.f),
    C_Vertex3f(0.32624054f, 0.276070595f, 0.f),
    C_Vertex3f(0.332329988f, 0.270356417f, 0.f),
    C_Vertex3f(0.342034101f, 0.340753078f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.319932938f, 0.339355469f, 0.f),
    C_Vertex3f(0.309783936f, 0.285862923f, 0.f),
    C_Vertex3f(0.319480896f, 0.281238556f, 0.f),
    C_Vertex3f(0.323425293f, 0.278160095f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.419417381f, 0.242657661f, 0.f),
    C_Vertex3f(0.406285286f, 0.260252953f, 0.f),
    C_Vertex3f(0.414680481f, 0.139648438f, 0.f),
    C_Vertex3f(0.484863281f, 0.139648438f, 0.f),
    C_Vertex3f(0.43375206f, 0.221509933f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.414680481f, 0.139648438f, 0.f),
    C_Vertex3f(0.406285286f, 0.260252953f, 0.f),
    C_Vertex3f(0.394355774f, 0.274295807f, 0.f),
    C_Vertex3f(0.38899231f, 0.279541016f, 0.f),
    C_Vertex3f(0.386146545f, 0.282274246f, 0.f),
    C_Vertex3f(0.379525185f, 0.28735733f, 0.f),
    C_Vertex3f(0.378504753f, 0.203832626f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.358787537f, 0.298339844f, 0.f),
    C_Vertex3f(0.361277342f, 0.232147932f, 0.f),
    C_Vertex3f(0.363492966f, 0.296373367f, 0.f),
    C_Vertex3f(0.36966157f, 0.218757868f, 0.f),
    C_Vertex3f(0.371973991f, 0.292057037f, 0.f),
    C_Vertex3f(0.378504753f, 0.203832626f, 0.f),
    C_Vertex3f(0.379525185f, 0.28735733f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.243652344f, 0.446777344f, 0.f),
    C_Vertex3f(0.243652344f, 0.139648438f, 0.f),
    C_Vertex3f(0.243652344f, 0.290527344f, 0.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, 0.f),
    C_Vertex3f(0.00146484375f, 0.332565308f, 0.f),
    C_Vertex3f(0.00146484375f, 0.311144829f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.661132813f, 0.311144829f, -1.f),
    C_Vertex3f(0.661132813f, 0.332565308f, -1.f),
    C_Vertex3f(0.659773588f, 0.290038824f, -1.f),
    C_Vertex3f(0.659753084f, 0.353883743f, -1.f),
    C_Vertex3f(0.65705514f, 0.269137621f, -1.f),
    C_Vertex3f(0.656993628f, 0.374990463f, -1.f),
    C_Vertex3f(0.652977467f, 0.248441219f, -1.f),
    C_Vertex3f(0.652854443f, 0.395885468f, -1.f),
    C_Vertex3f(0.647540569f, 0.227949619f, -1.f),
    C_Vertex3f(0.647335529f, 0.416568756f, -1.f),
    C_Vertex3f(0.640744448f, 0.207662821f, -1.f),
    C_Vertex3f(0.640436888f, 0.437040329f, -1.f),
    C_Vertex3f(0.632589102f, 0.187580824f, -1.f),
    C_Vertex3f(0.632158518f, 0.457300186f, -1.f),
    C_Vertex3f(0.623074532f, 0.167703629f, -1.f),
    C_Vertex3f(0.62250042f, 0.477348328f, -1.f),
    C_Vertex3f(0.612200737f, 0.148031235f, -1.f),
    C_Vertex3f(0.611462593f, 0.497184753f, -1.f),
    C_Vertex3f(0.606445313f, 0.330760956f, -1.f),
    C_Vertex3f(0.605291128f, 0.348511577f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.589132547f, 0.417805791f, -1.f),
    C_Vertex3f(0.594903469f, 0.400738478f, -1.f),
    C_Vertex3f(0.599331737f, 0.515873551f, -1.f),
    C_Vertex3f(0.599520206f, 0.383500338f, -1.f),
    C_Vertex3f(0.611462593f, 0.497184753f, -1.f),
    C_Vertex3f(0.602982759f, 0.366091371f, -1.f),
    C_Vertex3f(0.605291128f, 0.348511577f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.58610785f, 0.533414721f, -1.f),
    C_Vertex3f(0.571790934f, 0.549808264f, -1.f),
    C_Vertex3f(0.574128151f, 0.451427937f, -1.f),
    C_Vertex3f(0.582207441f, 0.434702277f, -1.f),
    C_Vertex3f(0.589132547f, 0.417805791f, -1.f),
    C_Vertex3f(0.599331737f, 0.515873551f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.543702602f, 0.498223901f, -1.f),
    C_Vertex3f(0.554752827f, 0.483581424f, -1.f),
    C_Vertex3f(0.556380987f, 0.565054178f, -1.f),
    C_Vertex3f(0.564894676f, 0.467982769f, -1.f),
    C_Vertex3f(0.571790934f, 0.549808264f, -1.f),
    C_Vertex3f(0.574128151f, 0.451427937f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.505101681f, 0.536414266f, -1.f),
    C_Vertex3f(0.518877029f, 0.524640322f, -1.f),
    C_Vertex3f(0.522282004f, 0.592103124f, -1.f),
    C_Vertex3f(0.531744003f, 0.5119102f, -1.f),
    C_Vertex3f(0.539878011f, 0.579152465f, -1.f),
    C_Vertex3f(0.543702602f, 0.498223901f, -1.f),
    C_Vertex3f(0.556380987f, 0.565054178f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.425038338f, 0.580470324f, -1.f),
    C_Vertex3f(0.441729546f, 0.573791265f, -1.f),
    C_Vertex3f(0.44387877f, 0.631876588f, -1.f),
    C_Vertex3f(0.458325386f, 0.565999031f, -1.f),
    C_Vertex3f(0.463906169f, 0.623885036f, -1.f),
    C_Vertex3f(0.474825859f, 0.55709362f, -1.f),
    C_Vertex3f(0.483810902f, 0.614561558f, -1.f),
    C_Vertex3f(0.490417957f, 0.547232032f, -1.f),
    C_Vertex3f(0.503592968f, 0.603906155f, -1.f),
    C_Vertex3f(0.505101681f, 0.536414266f, -1.f),
    C_Vertex3f(0.522282004f, 0.592103124f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.322833061f, 0.597167969f, -1.f),
    C_Vertex3f(0.340151787f, 0.597167969f, -1.f),
    C_Vertex3f(0.341901779f, 0.651855469f, -1.f),
    C_Vertex3f(0.357319832f, 0.596054792f, -1.f),
    C_Vertex3f(0.36254251f, 0.650523543f, -1.f),
    C_Vertex3f(0.374392509f, 0.59382844f, -1.f),
    C_Vertex3f(0.383060575f, 0.647859693f, -1.f),
    C_Vertex3f(0.39136982f, 0.590488911f, -1.f),
    C_Vertex3f(0.403455973f, 0.643863916f, -1.f),
    C_Vertex3f(0.408251762f, 0.586036205f, -1.f),
    C_Vertex3f(0.423728704f, 0.638536215f, -1.f),
    C_Vertex3f(0.425038338f, 0.580470324f, -1.f),
    C_Vertex3f(0.44387877f, 0.631876588f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.221024632f, 0.573791265f, -1.f),
    C_Vertex3f(0.237702727f, 0.580470324f, -1.f),
    C_Vertex3f(0.23930645f, 0.638536215f, -1.f),
    C_Vertex3f(0.254496813f, 0.586036205f, -1.f),
    C_Vertex3f(0.259579659f, 0.643863916f, -1.f),
    C_Vertex3f(0.271406889f, 0.590488911f, -1.f),
    C_Vertex3f(0.279975891f, 0.647859693f, -1.f),
    C_Vertex3f(0.288432956f, 0.59382844f, -1.f),
    C_Vertex3f(0.300495148f, 0.650523543f, -1.f),
    C_Vertex3f(0.305575013f, 0.596054792f, -1.f),
    C_Vertex3f(0.321137428f, 0.651855469f, -1.f),
    C_Vertex3f(0.322833061f, 0.597167969f, -1.f),
    C_Vertex3f(0.341901779f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.144046783f, 0.524640322f, -1.f),
    C_Vertex3f(0.157808542f, 0.536414266f, -1.f),
    C_Vertex3f(0.159443855f, 0.603906155f, -1.f),
    C_Vertex3f(0.172465086f, 0.547232032f, -1.f),
    C_Vertex3f(0.179224968f, 0.614561558f, -1.f),
    C_Vertex3f(0.188016415f, 0.55709362f, -1.f),
    C_Vertex3f(0.199129105f, 0.623885036f, -1.f),
    C_Vertex3f(0.204462528f, 0.565999031f, -1.f),
    C_Vertex3f(0.219156265f, 0.631876588f, -1.f),
    C_Vertex3f(0.221024632f, 0.573791265f, -1.f),
    C_Vertex3f(0.23930645f, 0.638536215f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.108130217f, 0.483581424f, -1.f),
    C_Vertex3f(0.119207621f, 0.498223901f, -1.f),
    C_Vertex3f(0.12313962f, 0.579152465f, -1.f),
    C_Vertex3f(0.13117981f, 0.5119102f, -1.f),
    C_Vertex3f(0.140748739f, 0.592103124f, -1.f),
    C_Vertex3f(0.144046783f, 0.524640322f, -1.f),
    C_Vertex3f(0.159443855f, 0.603906155f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0805329084f, 0.434702277f, -1.f),
    C_Vertex3f(0.0886597633f, 0.451427937f, -1.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, -1.f),
    C_Vertex3f(0.0979475975f, 0.467982769f, -1.f),
    C_Vertex3f(0.106616497f, 0.565054178f, -1.f),
    C_Vertex3f(0.108130217f, 0.483581424f, -1.f),
    C_Vertex3f(0.12313962f, 0.579152465f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0768282413f, 0.533414721f, -1.f),
    C_Vertex3f(0.0635631084f, 0.515873551f, -1.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, -1.f),
    C_Vertex3f(0.0735670328f, 0.417805791f, -1.f),
    C_Vertex3f(0.0805329084f, 0.434702277f, -1.f),
    C_Vertex3f(0.0911793709f, 0.549808264f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0635631084f, 0.515873551f, -1.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, -1.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, -1.f),
    C_Vertex3f(0.057313323f, 0.348511577f, -1.f),
    C_Vertex3f(0.0596352816f, 0.366091371f, -1.f),
    C_Vertex3f(0.0631182194f, 0.383500338f, -1.f),
    C_Vertex3f(0.0677621365f, 0.400738478f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.00146484375f, 0.332565308f, -1.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, -1.f),
    C_Vertex3f(0.00285148621f, 0.353883743f, -1.f),
    C_Vertex3f(0.00556325912f, 0.269137621f, -1.f),
    C_Vertex3f(0.00562477112f, 0.374990463f, -1.f),
    C_Vertex3f(0.0096616745f, 0.248441219f, -1.f),
    C_Vertex3f(0.00978469849f, 0.395885468f, -1.f),
    C_Vertex3f(0.0151262283f, 0.227949619f, -1.f),
    C_Vertex3f(0.0153312683f, 0.416568756f, -1.f),
    C_Vertex3f(0.0219569206f, 0.207662821f, -1.f),
    C_Vertex3f(0.0222644806f, 0.437040329f, -1.f),
    C_Vertex3f(0.0301537514f, 0.187580824f, -1.f),
    C_Vertex3f(0.0305843353f, 0.457300186f, -1.f),
    C_Vertex3f(0.0397167206f, 0.167703629f, -1.f),
    C_Vertex3f(0.0402908325f, 0.477348328f, -1.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, -1.f),
    C_Vertex3f(0.0513839722f, 0.497184753f, -1.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, -1.f),
    C_Vertex3f(0.0561523438f, 0.330760956f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.612200737f, 0.148031235f, -1.f),
    C_Vertex3f(0.606445313f, 0.330760956f, -1.f),
    C_Vertex3f(0.606445313f, 0.312894821f, -1.f),
    C_Vertex3f(0.605311632f, 0.295268297f, -1.f),
    C_Vertex3f(0.603044271f, 0.277826071f, -1.f),
    C_Vertex3f(0.600233912f, 0.129458666f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.600233912f, 0.129458666f, -1.f),
    C_Vertex3f(0.603044271f, 0.277826071f, -1.f),
    C_Vertex3f(0.59964323f, 0.260568142f, -1.f),
    C_Vertex3f(0.595108509f, 0.243494511f, -1.f),
    C_Vertex3f(0.589440107f, 0.226605177f, -1.f),
    C_Vertex3f(0.587174058f, 0.111985922f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.589440107f, 0.226605177f, -1.f),
    C_Vertex3f(0.582638025f, 0.209900141f, -1.f),
    C_Vertex3f(0.587174058f, 0.111985922f, -1.f),
    C_Vertex3f(0.574702263f, 0.193379402f, -1.f),
    C_Vertex3f(0.573021173f, 0.0956130028f, -1.f),
    C_Vertex3f(0.56563282f, 0.177042961f, -1.f),
    C_Vertex3f(0.557775259f, 0.0803399086f, -1.f),
    C_Vertex3f(0.555648208f, 0.1616081f, -1.f),
    C_Vertex3f(0.544748425f, 0.147074819f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.557775259f, 0.0803399086f, -1.f),
    C_Vertex3f(0.544748425f, 0.147074819f, -1.f),
    C_Vertex3f(0.541436315f, 0.0661666393f, -1.f),
    C_Vertex3f(0.532933474f, 0.133443117f, -1.f),
    C_Vertex3f(0.52400434f, 0.053093195f, -1.f),
    C_Vertex3f(0.520203352f, 0.120712996f, -1.f),
    C_Vertex3f(0.506558061f, 0.108884454f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52400434f, 0.053093195f, -1.f),
    C_Vertex3f(0.506558061f, 0.108884454f, -1.f),
    C_Vertex3f(0.505479336f, 0.0411195755f, -1.f),
    C_Vertex3f(0.4919976f, 0.0979574919f, -1.f),
    C_Vertex3f(0.485861301f, 0.0302457809f, -1.f),
    C_Vertex3f(0.476521969f, 0.0879321098f, -1.f),
    C_Vertex3f(0.466018081f, 0.0207312107f, -1.f),
    C_Vertex3f(0.460131168f, 0.0788083076f, -1.f),
    C_Vertex3f(0.445949674f, 0.0125758648f, -1.f),
    C_Vertex3f(0.443562865f, 0.0708249807f, -1.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445949674f, 0.0125758648f, -1.f),
    C_Vertex3f(0.42681706f, 0.0639821291f, -1.f),
    C_Vertex3f(0.42565608f, 0.00577974319f, -1.f),
    C_Vertex3f(0.409893751f, 0.0582797527f, -1.f),
    C_Vertex3f(0.4051373f, 0.000342845917f, -1.f),
    C_Vertex3f(0.39279294f, 0.0537178516f, -1.f),
    C_Vertex3f(0.384393334f, -0.00373482704f, -1.f),
    C_Vertex3f(0.375514627f, 0.0502964258f, -1.f),
    C_Vertex3f(0.363424182f, -0.00645327568f, -1.f),
    C_Vertex3f(0.35805881f, 0.0480154753f, -1.f),
    C_Vertex3f(0.342229843f, -0.0078125f, -1.f),
    C_Vertex3f(0.340425491f, 0.046875f, -1.f),
    C_Vertex3f(0.322613716f, 0.046875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.342229843f, -0.0078125f, -1.f),
    C_Vertex3f(0.322613716f, 0.046875f, -1.f),
    C_Vertex3f(0.320809364f, -0.0078125f, -1.f),
    C_Vertex3f(0.304978728f, 0.0480154753f, -1.f),
    C_Vertex3f(0.299613476f, -0.00645327568f, -1.f),
    C_Vertex3f(0.287521482f, 0.0502964258f, -1.f),
    C_Vertex3f(0.278643131f, -0.00373482704f, -1.f),
    C_Vertex3f(0.270241976f, 0.0537178516f, -1.f),
    C_Vertex3f(0.257898331f, 0.000342845917f, -1.f),
    C_Vertex3f(0.253140211f, 0.0582797527f, -1.f),
    C_Vertex3f(0.237379074f, 0.00577974319f, -1.f),
    C_Vertex3f(0.236216187f, 0.0639821291f, -1.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.237379074f, 0.00577974319f, -1.f),
    C_Vertex3f(0.219469905f, 0.0708249807f, -1.f),
    C_Vertex3f(0.217085361f, 0.0125758648f, -1.f),
    C_Vertex3f(0.202901363f, 0.0788083076f, -1.f),
    C_Vertex3f(0.197017193f, 0.0207312107f, -1.f),
    C_Vertex3f(0.186510563f, 0.0879321098f, -1.f),
    C_Vertex3f(0.177174568f, 0.0302457809f, -1.f),
    C_Vertex3f(0.171028137f, 0.0979574919f, -1.f),
    C_Vertex3f(0.157557487f, 0.0411195755f, -1.f),
    C_Vertex3f(0.156454086f, 0.108884454f, -1.f),
    C_Vertex3f(0.14278841f, 0.120712996f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.157557487f, 0.0411195755f, -1.f),
    C_Vertex3f(0.14278841f, 0.120712996f, -1.f),
    C_Vertex3f(0.139026403f, 0.053093195f, -1.f),
    C_Vertex3f(0.130031109f, 0.133443117f, -1.f),
    C_Vertex3f(0.121581316f, 0.0661666393f, -1.f),
    C_Vertex3f(0.118182182f, 0.147074819f, -1.f),
    C_Vertex3f(0.107241631f, 0.1616081f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.121581316f, 0.0661666393f, -1.f),
    C_Vertex3f(0.107241631f, 0.1616081f, -1.f),
    C_Vertex3f(0.105222225f, 0.0803399086f, -1.f),
    C_Vertex3f(0.0972094536f, 0.177042961f, -1.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, -1.f),
    C_Vertex3f(0.0880856514f, 0.193379402f, -1.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0757620335f, 0.111985922f, -1.f),
    C_Vertex3f(0.089949131f, 0.0956130028f, -1.f),
    C_Vertex3f(0.0801023245f, 0.209900141f, -1.f),
    C_Vertex3f(0.0732594728f, 0.226605177f, -1.f),
    C_Vertex3f(0.0675570965f, 0.243494511f, -1.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, -1.f),
    C_Vertex3f(0.0626609325f, 0.129458666f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0626609325f, 0.129458666f, -1.f),
    C_Vertex3f(0.0629951954f, 0.260568142f, -1.f),
    C_Vertex3f(0.0595737696f, 0.277826071f, -1.f),
    C_Vertex3f(0.057292819f, 0.295268297f, -1.f),
    C_Vertex3f(0.0561523438f, 0.312894821f, -1.f),
    C_Vertex3f(0.0506458282f, 0.148031235f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.448443413f, 0.444354057f, -1.f),
    C_Vertex3f(0.450453281f, 0.353586197f, -1.f),
    C_Vertex3f(0.454690456f, 0.431731701f, -1.f),
    C_Vertex3f(0.45744276f, 0.370527267f, -1.f),
    C_Vertex3f(0.458855152f, 0.418726444f, -1.f),
    C_Vertex3f(0.4609375f, 0.388725281f, -1.f),
    C_Vertex3f(0.4609375f, 0.405338287f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.421924233f, 0.320660591f, -1.f),
    C_Vertex3f(0.429484844f, 0.32708168f, -1.f),
    C_Vertex3f(0.430360317f, 0.467002869f, -1.f),
    C_Vertex3f(0.439969063f, 0.337902069f, -1.f),
    C_Vertex3f(0.440114021f, 0.456593513f, -1.f),
    C_Vertex3f(0.450453281f, 0.353586197f, -1.f),
    C_Vertex3f(0.448443413f, 0.444354057f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.400390625f, 0.397668839f, -1.f),
    C_Vertex3f(0.405089617f, 0.310195923f, -1.f),
    C_Vertex3f(0.406579971f, 0.482331276f, -1.f),
    C_Vertex3f(0.413792491f, 0.315032005f, -1.f),
    C_Vertex3f(0.419182301f, 0.475582123f, -1.f),
    C_Vertex3f(0.421924233f, 0.320660591f, -1.f),
    C_Vertex3f(0.430360317f, 0.467002869f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389385223f, 0.425128937f, -1.f),
    C_Vertex3f(0.393787384f, 0.418880463f, -1.f),
    C_Vertex3f(0.39605999f, 0.486020565f, -1.f),
    C_Vertex3f(0.397089005f, 0.412220955f, -1.f),
    C_Vertex3f(0.406579971f, 0.482331276f, -1.f),
    C_Vertex3f(0.399290085f, 0.405150414f, -1.f),
    C_Vertex3f(0.400390625f, 0.397668839f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.37105751f, 0.439105034f, -1.f),
    C_Vertex3f(0.378020287f, 0.435241222f, -1.f),
    C_Vertex3f(0.378786087f, 0.490017295f, -1.f),
    C_Vertex3f(0.384129524f, 0.430582523f, -1.f),
    C_Vertex3f(0.387964249f, 0.48817265f, -1.f),
    C_Vertex3f(0.389385223f, 0.425128937f, -1.f),
    C_Vertex3f(0.39605999f, 0.486020565f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186523438f, 0.139648438f, -1.f),
    C_Vertex3f(0.243652344f, 0.139648438f, -1.f),
    C_Vertex3f(0.186523438f, 0.494628906f, -1.f),
    C_Vertex3f(0.243652344f, 0.446777344f, -1.f),
    C_Vertex3f(0.316658974f, 0.494628906f, -1.f),
    C_Vertex3f(0.319816589f, 0.446777344f, -1.f),
    C_Vertex3f(0.331249237f, 0.494321465f, -1.f),
    C_Vertex3f(0.337620735f, 0.446010113f, -1.f),
    C_Vertex3f(0.34475708f, 0.493706584f, -1.f),
    C_Vertex3f(0.352095604f, 0.444475651f, -1.f),
    C_Vertex3f(0.357182503f, 0.492784262f, -1.f),
    C_Vertex3f(0.363241196f, 0.442173958f, -1.f),
    C_Vertex3f(0.368525505f, 0.491554499f, -1.f),
    C_Vertex3f(0.37105751f, 0.439105034f, -1.f),
    C_Vertex3f(0.378786087f, 0.490017295f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243652344f, 0.446777344f, -1.f),
    C_Vertex3f(0.243652344f, 0.290527344f, -1.f),
    C_Vertex3f(0.243652344f, 0.339355469f, -1.f),
    C_Vertex3f(0.284965515f, 0.290513992f, -1.f),
    C_Vertex3f(0.319932938f, 0.339355469f, -1.f),
    C_Vertex3f(0.298278809f, 0.288954735f, -1.f),
    C_Vertex3f(0.309783936f, 0.285862923f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.400390625f, 0.397668839f, -1.f),
    C_Vertex3f(0.400390625f, 0.387584686f, -1.f),
    C_Vertex3f(0.405089617f, 0.310195923f, -1.f),
    C_Vertex3f(0.398106575f, 0.376514435f, -1.f),
    C_Vertex3f(0.395815611f, 0.306152344f, -1.f),
    C_Vertex3f(0.393538475f, 0.366649628f, -1.f),
    C_Vertex3f(0.386686325f, 0.357990265f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.395815611f, 0.306152344f, -1.f),
    C_Vertex3f(0.386686325f, 0.357990265f, -1.f),
    C_Vertex3f(0.385970473f, 0.302901268f, -1.f),
    C_Vertex3f(0.379834175f, 0.352399826f, -1.f),
    C_Vertex3f(0.375554204f, 0.300442696f, -1.f),
    C_Vertex3f(0.37426877f, 0.349138737f, -1.f),
    C_Vertex3f(0.36770606f, 0.346343517f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.375554204f, 0.300442696f, -1.f),
    C_Vertex3f(0.36770606f, 0.346343517f, -1.f),
    C_Vertex3f(0.364566803f, 0.298776627f, -1.f),
    C_Vertex3f(0.360146046f, 0.344014168f, -1.f),
    C_Vertex3f(0.358787537f, 0.298339844f, -1.f),
    C_Vertex3f(0.35335207f, 0.244002819f, -1.f),
    C_Vertex3f(0.361277342f, 0.232147932f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.332329988f, 0.270356417f, -1.f),
    C_Vertex3f(0.338878393f, 0.263107061f, -1.f),
    C_Vertex3f(0.342034101f, 0.340753078f, -1.f),
    C_Vertex3f(0.345885754f, 0.254322529f, -1.f),
    C_Vertex3f(0.351588726f, 0.342150688f, -1.f),
    C_Vertex3f(0.35335207f, 0.244002819f, -1.f),
    C_Vertex3f(0.360146046f, 0.344014168f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.331482172f, 0.339821339f, -1.f),
    C_Vertex3f(0.319932938f, 0.339355469f, -1.f),
    C_Vertex3f(0.323425293f, 0.278160095f, -1.f),
    C_Vertex3f(0.32624054f, 0.276070595f, -1.f),
    C_Vertex3f(0.332329988f, 0.270356417f, -1.f),
    C_Vertex3f(0.342034101f, 0.340753078f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.319932938f, 0.339355469f, -1.f),
    C_Vertex3f(0.309783936f, 0.285862923f, -1.f),
    C_Vertex3f(0.319480896f, 0.281238556f, -1.f),
    C_Vertex3f(0.323425293f, 0.278160095f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.419417381f, 0.242657661f, -1.f),
    C_Vertex3f(0.406285286f, 0.260252953f, -1.f),
    C_Vertex3f(0.414680481f, 0.139648438f, -1.f),
    C_Vertex3f(0.484863281f, 0.139648438f, -1.f),
    C_Vertex3f(0.43375206f, 0.221509933f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.414680481f, 0.139648438f, -1.f),
    C_Vertex3f(0.406285286f, 0.260252953f, -1.f),
    C_Vertex3f(0.394355774f, 0.274295807f, -1.f),
    C_Vertex3f(0.38899231f, 0.279541016f, -1.f),
    C_Vertex3f(0.386146545f, 0.282274246f, -1.f),
    C_Vertex3f(0.379525185f, 0.28735733f, -1.f),
    C_Vertex3f(0.378504753f, 0.203832626f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.358787537f, 0.298339844f, -1.f),
    C_Vertex3f(0.361277342f, 0.232147932f, -1.f),
    C_Vertex3f(0.363492966f, 0.296373367f, -1.f),
    C_Vertex3f(0.36966157f, 0.218757868f, -1.f),
    C_Vertex3f(0.371973991f, 0.292057037f, -1.f),
    C_Vertex3f(0.378504753f, 0.203832626f, -1.f),
    C_Vertex3f(0.379525185f, 0.28735733f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.243652344f, 0.446777344f, -1.f),
    C_Vertex3f(0.243652344f, 0.139648438f, -1.f),
    C_Vertex3f(0.243652344f, 0.290527344f, -1.f),
    C_Vertex3f(0.00283098221f, 0.290038824f, -1.f),
    C_Vertex3f(0.00146484375f, 0.332565308f, -1.f),
    C_Vertex3f(0.00146484375f, 0.311144829f, -1.f),
    C_End(),
    C_Translatef(0.659667969f, 0.f, 0.f),
};
static const Call char_175[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(-0.013671875f, 0.684570313f, 0.f),
    C_Vertex3f(-0.013671875f, 0.684570313f, -1.f),
    C_Vertex3f(-0.013671875f, 0.740722656f, -1.f),
    C_Vertex3f(-0.013671875f, 0.740722656f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(-0.013671875f, 0.740722656f, 0.f),
    C_Vertex3f(-0.013671875f, 0.740722656f, -1.f),
    C_Vertex3f(0.5078125f, 0.740722656f, -1.f),
    C_Vertex3f(0.5078125f, 0.740722656f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.5078125f, 0.740722656f, 0.f),
    C_Vertex3f(0.5078125f, 0.740722656f, -1.f),
    C_Vertex3f(0.5078125f, 0.684570313f, -1.f),
    C_Vertex3f(0.5078125f, 0.684570313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.5078125f, 0.684570313f, 0.f),
    C_Vertex3f(0.5078125f, 0.684570313f, -1.f),
    C_Vertex3f(-0.013671875f, 0.684570313f, -1.f),
    C_Vertex3f(-0.013671875f, 0.684570313f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-0.013671875f, 0.684570313f, 0.f),
    C_Vertex3f(0.5078125f, 0.684570313f, 0.f),
    C_Vertex3f(0.5078125f, 0.740722656f, 0.f),
    C_Vertex3f(-0.013671875f, 0.740722656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(-0.013671875f, 0.684570313f, -1.f),
    C_Vertex3f(0.5078125f, 0.684570313f, -1.f),
    C_Vertex3f(0.5078125f, 0.740722656f, -1.f),
    C_Vertex3f(-0.013671875f, 0.740722656f, -1.f),
    C_End(),
    C_Translatef(0.494140625f, 0.f, 0.f),
};
static const Call char_176[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.998995721f, -0.0448058881f, 0.f),
    C_Vertex3f(0.0556640625f, 0.523909569f, 0.f),
    C_Vertex3f(0.0556640625f, 0.523909569f, -1.f),
    C_Normal3f(-0.998987794f, 0.0449808836f, 0.f),
    C_Vertex3f(0.0556640625f, 0.536536217f, 0.f),
    C_Vertex3f(0.0556640625f, 0.536536217f, -1.f),
    C_Normal3f(-0.990249097f, 0.139308304f, 0.f),
    C_Vertex3f(0.0567777157f, 0.548642874f, 0.f),
    C_Vertex3f(0.0567777157f, 0.548642874f, -1.f),
    C_Normal3f(-0.971494913f, 0.237060159f, 0.f),
    C_Vertex3f(0.059005022f, 0.560284853f, 0.f),
    C_Vertex3f(0.059005022f, 0.560284853f, -1.f),
    C_Normal3f(-0.942044973f, 0.33548674f, 0.f),
    C_Vertex3f(0.0623459816f, 0.571462154f, 0.f),
    C_Vertex3f(0.0623459816f, 0.571462154f, -1.f),
    C_Normal3f(-0.902163327f, 0.431394458f, 0.f),
    C_Vertex3f(0.0668005943f, 0.582174778f, 0.f),
    C_Vertex3f(0.0668005943f, 0.582174778f, -1.f),
    C_Normal3f(-0.853113413f, 0.521725595f, 0.f),
    C_Vertex3f(0.0723688602f, 0.592422724f, 0.f),
    C_Vertex3f(0.0723688602f, 0.592422724f, -1.f),
    C_Normal3f(-0.796963871f, 0.604026973f, 0.f),
    C_Vertex3f(0.0790507793f, 0.602205992f, 0.f),
    C_Vertex3f(0.0790507793f, 0.602205992f, -1.f),
    C_Normal3f(-0.736211658f, 0.676751435f, 0.f),
    C_Vertex3f(0.0868463516f, 0.611524582f, 0.f),
    C_Vertex3f(0.0868463516f, 0.611524582f, -1.f),
    C_Normal3f(-0.672648728f, 0.739961863f, 0.f),
    C_Vertex3f(0.0957555771f, 0.620378494f, 0.f),
    C_Vertex3f(0.0957555771f, 0.620378494f, -1.f),
    C_Normal3f(-0.60066551f, 0.799500465f, 0.f),
    C_Vertex3f(0.105108738f, 0.628125668f, 0.f),
    C_Vertex3f(0.105108738f, 0.628125668f, -1.f),
    C_Normal3f(-0.51923126f, 0.854633749f, 0.f),
    C_Vertex3f(0.114905834f, 0.634766102f, 0.f),
    C_Vertex3f(0.114905834f, 0.634766102f, -1.f),
    C_Normal3f(-0.429789007f, 0.902929306f, 0.f),
    C_Vertex3f(0.125146866f, 0.640299797f, 0.f),
    C_Vertex3f(0.125146866f, 0.640299797f, -1.f),
    C_Normal3f(-0.33466953f, 0.942335606f, 0.f),
    C_Vertex3f(0.135831833f, 0.644726753f, 0.f),
    C_Vertex3f(0.135831833f, 0.644726753f, -1.f),
    C_Normal3f(-0.236822292f, 0.971552968f, 0.f),
    C_Vertex3f(0.146960735f, 0.64804697f, 0.f),
    C_Vertex3f(0.146960735f, 0.64804697f, -1.f),
    C_Normal3f(-0.139375359f, 0.990239561f, 0.f),
    C_Vertex3f(0.158533573f, 0.650260448f, 0.f),
    C_Vertex3f(0.158533573f, 0.650260448f, -1.f),
    C_Normal3f(-0.0449656211f, 0.998988509f, 0.f),
    C_Vertex3f(0.170550346f, 0.651367188f, 0.f),
    C_Vertex3f(0.170550346f, 0.651367188f, -1.f),
    C_Normal3f(0.0446274802f, 0.999003708f, 0.f),
    C_Vertex3f(0.183121681f, 0.651367188f, 0.f),
    C_Vertex3f(0.183121681f, 0.651367188f, -1.f),
    C_Normal3f(0.137439534f, 0.990510166f, 0.f),
    C_Vertex3f(0.195325136f, 0.650260448f, 0.f),
    C_Vertex3f(0.195325136f, 0.650260448f, -1.f),
    C_Normal3f(0.234195665f, 0.972189486f, 0.f),
    C_Vertex3f(0.207050085f, 0.64804697f, 0.f),
    C_Vertex3f(0.207050085f, 0.64804697f, -1.f),
    C_Normal3f(0.331958562f, 0.943293929f, 0.f),
    C_Vertex3f(0.218296528f, 0.644726753f, 0.f),
    C_Vertex3f(0.218296528f, 0.644726753f, -1.f),
    C_Normal3f(0.427599251f, 0.903968453f, 0.f),
    C_Vertex3f(0.229064465f, 0.640299797f, 0.f),
    C_Vertex3f(0.229064465f, 0.640299797f, -1.f),
    C_Normal3f(0.518055737f, 0.855346918f, 0.f),
    C_Vertex3f(0.239353895f, 0.634766102f, 0.f),
    C_Vertex3f(0.239353895f, 0.634766102f, -1.f),
    C_Normal3f(0.60080415f, 0.799396276f, 0.f),
    C_Vertex3f(0.24916482f, 0.628125668f, 0.f),
    C_Vertex3f(0.24916482f, 0.628125668f, -1.f),
    C_Normal3f(0.67418468f, 0.738562882f, 0.f),
    C_Vertex3f(0.258497238f, 0.620378494f, 0.f),
    C_Vertex3f(0.258497238f, 0.620378494f, -1.f),
    C_Normal3f(0.738307476f, 0.674464405f, 0.f),
    C_Vertex3f(0.267351151f, 0.611524582f, 0.f),
    C_Vertex3f(0.267351151f, 0.611524582f, -1.f),
    C_Normal3f(0.798769653f, 0.601637006f, 0.f),
    C_Vertex3f(0.275098324f, 0.602205992f, 0.f),
    C_Vertex3f(0.275098324f, 0.602205992f, -1.f),
    C_Normal3f(0.854554176f, 0.519362092f, 0.f),
    C_Vertex3f(0.281738758f, 0.592422724f, 0.f),
    C_Vertex3f(0.281738758f, 0.592422724f, -1.f),
    C_Normal3f(0.903204262f, 0.429210812f, 0.f),
    C_Vertex3f(0.287272453f, 0.582174778f, 0.f),
    C_Vertex3f(0.287272453f, 0.582174778f, -1.f),
    C_Normal3f(0.942701757f, 0.333636314f, 0.f),
    C_Vertex3f(0.291699409f, 0.571462154f, 0.f),
    C_Vertex3f(0.291699409f, 0.571462154f, -1.f),
    C_Normal3f(0.97183305f, 0.235670343f, 0.f),
    C_Vertex3f(0.295019627f, 0.560284853f, 0.f),
    C_Vertex3f(0.295019627f, 0.560284853f, -1.f),
    C_Normal3f(0.990368068f, 0.13846004f, 0.f),
    C_Vertex3f(0.297233105f, 0.548642874f, 0.f),
    C_Vertex3f(0.297233105f, 0.548642874f, -1.f),
    C_Normal3f(0.9989959f, 0.0448021777f, 0.f),
    C_Vertex3f(0.298339844f, 0.536536217f, 0.f),
    C_Vertex3f(0.298339844f, 0.536536217f, -1.f),
    C_Normal3f(0.99898386f, -0.045068901f, 0.f),
    C_Vertex3f(0.298339844f, 0.523964882f, 0.f),
    C_Vertex3f(0.298339844f, 0.523964882f, -1.f),
    C_Normal3f(0.990265787f, -0.139189079f, 0.f),
    C_Vertex3f(0.297226191f, 0.511851311f, 0.f),
    C_Vertex3f(0.297226191f, 0.511851311f, -1.f),
    C_Normal3f(0.971577525f, -0.23672168f, 0.f),
    C_Vertex3f(0.294998884f, 0.500195503f, 0.f),
    C_Vertex3f(0.294998884f, 0.500195503f, -1.f),
    C_Normal3f(0.942278624f, -0.334829658f, 0.f),
    C_Vertex3f(0.291657925f, 0.488997459f, 0.f),
    C_Vertex3f(0.291657925f, 0.488997459f, -1.f),
    C_Normal3f(0.902660072f, -0.430354357f, 0.f),
    C_Vertex3f(0.287203312f, 0.478257179f, 0.f),
    C_Vertex3f(0.287203312f, 0.478257179f, -1.f),
    C_Normal3f(0.853991389f, -0.520287097f, 0.f),
    C_Vertex3f(0.281635046f, 0.467974663f, 0.f),
    C_Vertex3f(0.281635046f, 0.467974663f, -1.f),
    C_Normal3f(0.798326015f, -0.602225542f, 0.f),
    C_Vertex3f(0.274953127f, 0.45814991f, 0.f),
    C_Vertex3f(0.274953127f, 0.45814991f, -1.f),
    C_Normal3f(0.738127053f, -0.674661756f, 0.f),
    C_Vertex3f(0.267157555f, 0.448782921f, 0.f),
    C_Vertex3f(0.267157555f, 0.448782921f, -1.f),
    C_Normal3f(0.674661756f, -0.738127053f, 0.f),
    C_Vertex3f(0.258248329f, 0.439873695f, 0.f),
    C_Vertex3f(0.258248329f, 0.439873695f, -1.f),
    C_Normal3f(0.602225542f, -0.798326015f, 0.f),
    C_Vertex3f(0.24888134f, 0.432078123f, 0.f),
    C_Vertex3f(0.24888134f, 0.432078123f, -1.f),
    C_Normal3f(0.520287097f, -0.853991389f, 0.f),
    C_Vertex3f(0.239056587f, 0.425396204f, 0.f),
    C_Vertex3f(0.239056587f, 0.425396204f, -1.f),
    C_Normal3f(0.430354357f, -0.902660072f, 0.f),
    C_Vertex3f(0.228774071f, 0.419827938f, 0.f),
    C_Vertex3f(0.228774071f, 0.419827938f, -1.f),
    C_Normal3f(0.334829658f, -0.942278624f, 0.f),
    C_Vertex3f(0.218033791f, 0.415373325f, 0.f),
    C_Vertex3f(0.218033791f, 0.415373325f, -1.f),
    C_Normal3f(0.23672168f, -0.971577525f, 0.f),
    C_Vertex3f(0.206835747f, 0.412032366f, 0.f),
    C_Vertex3f(0.206835747f, 0.412032366f, -1.f),
    C_Normal3f(0.139189079f, -0.990265787f, 0.f),
    C_Vertex3f(0.195179939f, 0.409805059f, 0.f),
    C_Vertex3f(0.195179939f, 0.409805059f, -1.f),
    C_Normal3f(0.0451699123f, -0.99897933f, 0.f),
    C_Vertex3f(0.183066368f, 0.408691406f, 0.f),
    C_Vertex3f(0.183066368f, 0.408691406f, -1.f),
    C_Normal3f(-0.0450668037f, -0.998984039f, 0.f),
    C_Vertex3f(0.170550346f, 0.408691406f, 0.f),
    C_Vertex3f(0.170550346f, 0.408691406f, -1.f),
    C_Normal3f(-0.139375359f, -0.990239561f, 0.f),
    C_Vertex3f(0.158533573f, 0.409798145f, 0.f),
    C_Vertex3f(0.158533573f, 0.409798145f, -1.f),
    C_Normal3f(-0.236822292f, -0.971552968f, 0.f),
    C_Vertex3f(0.146960735f, 0.412011623f, 0.f),
    C_Vertex3f(0.146960735f, 0.412011623f, -1.f),
    C_Normal3f(-0.33466953f, -0.942335606f, 0.f),
    C_Vertex3f(0.135831833f, 0.415331841f, 0.f),
    C_Vertex3f(0.135831833f, 0.415331841f, -1.f),
    C_Normal3f(-0.429789007f, -0.902929306f, 0.f),
    C_Vertex3f(0.125146866f, 0.419758797f, 0.f),
    C_Vertex3f(0.125146866f, 0.419758797f, -1.f),
    C_Normal3f(-0.51923126f, -0.854633749f, 0.f),
    C_Vertex3f(0.114905834f, 0.425292492f, 0.f),
    C_Vertex3f(0.114905834f, 0.425292492f, -1.f),
    C_Normal3f(-0.60066551f, -0.799500465f, 0.f),
    C_Vertex3f(0.105108738f, 0.431932926f, 0.f),
    C_Vertex3f(0.105108738f, 0.431932926f, -1.f),
    C_Normal3f(-0.672648728f, -0.739961863f, 0.f),
    C_Vertex3f(0.0957555771f, 0.439680099f, 0.f),
    C_Vertex3f(0.0957555771f, 0.439680099f, -1.f),
    C_Normal3f(-0.736468077f, -0.676472366f, 0.f),
    C_Vertex3f(0.0868463516f, 0.448534012f, 0.f),
    C_Vertex3f(0.0868463516f, 0.448534012f, -1.f),
    C_Normal3f(-0.797594011f, -0.603194594f, 0.f),
    C_Vertex3f(0.0790507793f, 0.45786643f, 0.f),
    C_Vertex3f(0.0790507793f, 0.45786643f, -1.f),
    C_Normal3f(-0.853911817f, -0.520417571f, 0.f),
    C_Vertex3f(0.0723688602f, 0.467677355f, 0.f),
    C_Vertex3f(0.0723688602f, 0.467677355f, -1.f),
    C_Normal3f(-0.902934372f, -0.429778397f, 0.f),
    C_Vertex3f(0.0668005943f, 0.477966785f, 0.f),
    C_Vertex3f(0.0668005943f, 0.477966785f, -1.f),
    C_Normal3f(-0.942643225f, -0.333801985f, 0.f),
    C_Vertex3f(0.0623459816f, 0.488734722f, 0.f),
    C_Vertex3f(0.0623459816f, 0.488734722f, -1.f),
    C_Normal3f(-0.971855462f, -0.235577792f, 0.f),
    C_Vertex3f(0.059005022f, 0.499981165f, 0.f),
    C_Vertex3f(0.059005022f, 0.499981165f, -1.f),
    C_Normal3f(-0.990392983f, -0.138281807f, 0.f),
    C_Vertex3f(0.0567777157f, 0.511706114f, 0.f),
    C_Vertex3f(0.0567777157f, 0.511706114f, -1.f),
    C_Normal3f(-0.998995721f, -0.0448058881f, 0.f),
    C_Vertex3f(0.0556640625f, 0.523909569f, 0.f),
    C_Vertex3f(0.0556640625f, 0.523909569f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.995793521f, -0.091626212f, 0.f),
    C_Vertex3f(0.103515625f, 0.537639618f, 0.f),
    C_Vertex3f(0.103515625f, 0.537639618f, -1.f),
    C_Normal3f(0.995793521f, 0.091626212f, 0.f),
    C_Vertex3f(0.103515625f, 0.522418976f, 0.f),
    C_Vertex3f(0.103515625f, 0.522418976f, -1.f),
    C_Normal3f(0.958154142f, 0.286252469f, 0.f),
    C_Vertex3f(0.106214046f, 0.508313179f, 0.f),
    C_Vertex3f(0.106214046f, 0.508313179f, -1.f),
    C_Normal3f(0.878959417f, 0.476896554f, 0.f),
    C_Vertex3f(0.111610889f, 0.495322227f, 0.f),
    C_Vertex3f(0.111610889f, 0.495322227f, -1.f),
    C_Normal3f(0.767812848f, 0.640674233f, 0.f),
    C_Vertex3f(0.119706154f, 0.483446121f, 0.f),
    C_Vertex3f(0.119706154f, 0.483446121f, -1.f),
    C_Normal3f(0.638400316f, 0.769704461f, 0.f),
    C_Vertex3f(0.13049984f, 0.47268486f, 0.f),
    C_Vertex3f(0.13049984f, 0.47268486f, -1.f),
    C_Normal3f(0.474682689f, 0.880156934f, 0.f),
    C_Vertex3f(0.142411709f, 0.464613914f, 0.f),
    C_Vertex3f(0.142411709f, 0.464613914f, -1.f),
    C_Normal3f(0.284675807f, 0.958623827f, 0.f),
    C_Vertex3f(0.155441761f, 0.459233284f, 0.f),
    C_Vertex3f(0.155441761f, 0.459233284f, -1.f),
    C_Normal3f(0.0914225429f, 0.995812237f, 0.f),
    C_Vertex3f(0.169589996f, 0.456542969f, 0.f),
    C_Vertex3f(0.169589996f, 0.456542969f, -1.f),
    C_Normal3f(-0.0919388011f, 0.995764673f, 0.f),
    C_Vertex3f(0.184745789f, 0.456542969f, 0.f),
    C_Vertex3f(0.184745789f, 0.456542969f, -1.f),
    C_Normal3f(-0.28736189f, 0.957822025f, 0.f),
    C_Vertex3f(0.198728085f, 0.459233284f, 0.f),
    C_Vertex3f(0.198728085f, 0.459233284f, -1.f),
    C_Normal3f(-0.477140039f, 0.878827333f, 0.f),
    C_Vertex3f(0.211647511f, 0.464613914f, 0.f),
    C_Vertex3f(0.211647511f, 0.464613914f, -1.f),
    C_Normal3f(-0.639322639f, 0.768938661f, 0.f),
    C_Vertex3f(0.223504066f, 0.47268486f, 0.f),
    C_Vertex3f(0.223504066f, 0.47268486f, -1.f),
    C_Normal3f(-0.767812848f, 0.640674233f, 0.f),
    C_Vertex3f(0.234297752f, 0.483446121f, 0.f),
    C_Vertex3f(0.234297752f, 0.483446121f, -1.f),
    C_Normal3f(-0.878959417f, 0.476896554f, 0.f),
    C_Vertex3f(0.242393017f, 0.495322227f, 0.f),
    C_Vertex3f(0.242393017f, 0.495322227f, -1.f),
    C_Normal3f(-0.958154142f, 0.286252469f, 0.f),
    C_Vertex3f(0.24778986f, 0.508313179f, 0.f),
    C_Vertex3f(0.24778986f, 0.508313179f, -1.f),
    C_Normal3f(-0.995793521f, 0.091626212f, 0.f),
    C_Vertex3f(0.250488281f, 0.522418976f, 0.f),
    C_Vertex3f(0.250488281f, 0.522418976f, -1.f),
    C_Normal3f(-0.995793521f, -0.091626212f, 0.f),
    C_Vertex3f(0.250488281f, 0.537639618f, 0.f),
    C_Vertex3f(0.250488281f, 0.537639618f, -1.f),
    C_Normal3f(-0.958154142f, -0.286252469f, 0.f),
    C_Vertex3f(0.24778986f, 0.551745415f, 0.f),
    C_Vertex3f(0.24778986f, 0.551745415f, -1.f),
    C_Normal3f(-0.878959417f, -0.476896554f, 0.f),
    C_Vertex3f(0.242393017f, 0.564736366f, 0.f),
    C_Vertex3f(0.242393017f, 0.564736366f, -1.f),
    C_Normal3f(-0.767812848f, -0.640674233f, 0.f),
    C_Vertex3f(0.234297752f, 0.576612473f, 0.f),
    C_Vertex3f(0.234297752f, 0.576612473f, -1.f),
    C_Normal3f(-0.639322639f, -0.768938661f, 0.f),
    C_Vertex3f(0.223504066f, 0.587373734f, 0.f),
    C_Vertex3f(0.223504066f, 0.587373734f, -1.f),
    C_Normal3f(-0.477140039f, -0.878827333f, 0.f),
    C_Vertex3f(0.211647511f, 0.595444679f, 0.f),
    C_Vertex3f(0.211647511f, 0.595444679f, -1.f),
    C_Normal3f(-0.28736189f, -0.957822025f, 0.f),
    C_Vertex3f(0.198728085f, 0.60082531f, 0.f),
    C_Vertex3f(0.198728085f, 0.60082531f, -1.f),
    C_Normal3f(-0.0919388011f, -0.995764673f, 0.f),
    C_Vertex3f(0.184745789f, 0.603515625f, 0.f),
    C_Vertex3f(0.184745789f, 0.603515625f, -1.f),
    C_Normal3f(0.0914225429f, -0.995812237f, 0.f),
    C_Vertex3f(0.169589996f, 0.603515625f, 0.f),
    C_Vertex3f(0.169589996f, 0.603515625f, -1.f),
    C_Normal3f(0.284675807f, -0.958623827f, 0.f),
    C_Vertex3f(0.155441761f, 0.60082531f, 0.f),
    C_Vertex3f(0.155441761f, 0.60082531f, -1.f),
    C_Normal3f(0.474682689f, -0.880156934f, 0.f),
    C_Vertex3f(0.142411709f, 0.595444679f, 0.f),
    C_Vertex3f(0.142411709f, 0.595444679f, -1.f),
    C_Normal3f(0.638400316f, -0.769704461f, 0.f),
    C_Vertex3f(0.13049984f, 0.587373734f, 0.f),
    C_Vertex3f(0.13049984f, 0.587373734f, -1.f),
    C_Normal3f(0.767812848f, -0.640674233f, 0.f),
    C_Vertex3f(0.119706154f, 0.576612473f, 0.f),
    C_Vertex3f(0.119706154f, 0.576612473f, -1.f),
    C_Normal3f(0.878959417f, -0.476896554f, 0.f),
    C_Vertex3f(0.111610889f, 0.564736366f, 0.f),
    C_Vertex3f(0.111610889f, 0.564736366f, -1.f),
    C_Normal3f(0.958154142f, -0.286252469f, 0.f),
    C_Vertex3f(0.106214046f, 0.551745415f, 0.f),
    C_Vertex3f(0.106214046f, 0.551745415f, -1.f),
    C_Normal3f(0.995793521f, -0.091626212f, 0.f),
    C_Vertex3f(0.103515625f, 0.537639618f, 0.f),
    C_Vertex3f(0.103515625f, 0.537639618f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.250488281f, 0.537639618f, 0.f),
    C_Vertex3f(0.258248329f, 0.439873695f, 0.f),
    C_Vertex3f(0.258497238f, 0.620378494f, 0.f),
    C_Vertex3f(0.267157555f, 0.448782921f, 0.f),
    C_Vertex3f(0.267351151f, 0.611524582f, 0.f),
    C_Vertex3f(0.274953127f, 0.45814991f, 0.f),
    C_Vertex3f(0.275098324f, 0.602205992f, 0.f),
    C_Vertex3f(0.281635046f, 0.467974663f, 0.f),
    C_Vertex3f(0.281738758f, 0.592422724f, 0.f),
    C_Vertex3f(0.287203312f, 0.478257179f, 0.f),
    C_Vertex3f(0.287272453f, 0.582174778f, 0.f),
    C_Vertex3f(0.291657925f, 0.488997459f, 0.f),
    C_Vertex3f(0.291699409f, 0.571462154f, 0.f),
    C_Vertex3f(0.294998884f, 0.500195503f, 0.f),
    C_Vertex3f(0.295019627f, 0.560284853f, 0.f),
    C_Vertex3f(0.297226191f, 0.511851311f, 0.f),
    C_Vertex3f(0.297233105f, 0.548642874f, 0.f),
    C_Vertex3f(0.298339844f, 0.523964882f, 0.f),
    C_Vertex3f(0.298339844f, 0.536536217f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.24916482f, 0.628125668f, 0.f),
    C_Vertex3f(0.239353895f, 0.634766102f, 0.f),
    C_Vertex3f(0.242393017f, 0.564736366f, 0.f),
    C_Vertex3f(0.24778986f, 0.551745415f, 0.f),
    C_Vertex3f(0.250488281f, 0.537639618f, 0.f),
    C_Vertex3f(0.258497238f, 0.620378494f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.242393017f, 0.564736366f, 0.f),
    C_Vertex3f(0.239353895f, 0.634766102f, 0.f),
    C_Vertex3f(0.234297752f, 0.576612473f, 0.f),
    C_Vertex3f(0.229064465f, 0.640299797f, 0.f),
    C_Vertex3f(0.223504066f, 0.587373734f, 0.f),
    C_Vertex3f(0.218296528f, 0.644726753f, 0.f),
    C_Vertex3f(0.211647511f, 0.595444679f, 0.f),
    C_Vertex3f(0.207050085f, 0.64804697f, 0.f),
    C_Vertex3f(0.198728085f, 0.60082531f, 0.f),
    C_Vertex3f(0.195325136f, 0.650260448f, 0.f),
    C_Vertex3f(0.184745789f, 0.603515625f, 0.f),
    C_Vertex3f(0.183121681f, 0.651367188f, 0.f),
    C_Vertex3f(0.170550346f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.106214046f, 0.551745415f, 0.f),
    C_Vertex3f(0.111610889f, 0.564736366f, 0.f),
    C_Vertex3f(0.114905834f, 0.634766102f, 0.f),
    C_Vertex3f(0.119706154f, 0.576612473f, 0.f),
    C_Vertex3f(0.125146866f, 0.640299797f, 0.f),
    C_Vertex3f(0.13049984f, 0.587373734f, 0.f),
    C_Vertex3f(0.135831833f, 0.644726753f, 0.f),
    C_Vertex3f(0.142411709f, 0.595444679f, 0.f),
    C_Vertex3f(0.146960735f, 0.64804697f, 0.f),
    C_Vertex3f(0.155441761f, 0.60082531f, 0.f),
    C_Vertex3f(0.158533573f, 0.650260448f, 0.f),
    C_Vertex3f(0.169589996f, 0.603515625f, 0.f),
    C_Vertex3f(0.170550346f, 0.651367188f, 0.f),
    C_Vertex3f(0.184745789f, 0.603515625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.119706154f, 0.483446121f, 0.f),
    C_Vertex3f(0.111610889f, 0.495322227f, 0.f),
    C_Vertex3f(0.114905834f, 0.425292492f, 0.f),
    C_Vertex3f(0.106214046f, 0.508313179f, 0.f),
    C_Vertex3f(0.105108738f, 0.431932926f, 0.f),
    C_Vertex3f(0.103515625f, 0.522418976f, 0.f),
    C_Vertex3f(0.0957555771f, 0.620378494f, 0.f),
    C_Vertex3f(0.103515625f, 0.537639618f, 0.f),
    C_Vertex3f(0.105108738f, 0.628125668f, 0.f),
    C_Vertex3f(0.106214046f, 0.551745415f, 0.f),
    C_Vertex3f(0.114905834f, 0.634766102f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.250488281f, 0.522418976f, 0.f),
    C_Vertex3f(0.24778986f, 0.508313179f, 0.f),
    C_Vertex3f(0.24888134f, 0.432078123f, 0.f),
    C_Vertex3f(0.258248329f, 0.439873695f, 0.f),
    C_Vertex3f(0.250488281f, 0.537639618f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.170550346f, 0.408691406f, 0.f),
    C_Vertex3f(0.183066368f, 0.408691406f, 0.f),
    C_Vertex3f(0.184745789f, 0.456542969f, 0.f),
    C_Vertex3f(0.195179939f, 0.409805059f, 0.f),
    C_Vertex3f(0.198728085f, 0.459233284f, 0.f),
    C_Vertex3f(0.206835747f, 0.412032366f, 0.f),
    C_Vertex3f(0.211647511f, 0.464613914f, 0.f),
    C_Vertex3f(0.218033791f, 0.415373325f, 0.f),
    C_Vertex3f(0.223504066f, 0.47268486f, 0.f),
    C_Vertex3f(0.228774071f, 0.419827938f, 0.f),
    C_Vertex3f(0.234297752f, 0.483446121f, 0.f),
    C_Vertex3f(0.239056587f, 0.425396204f, 0.f),
    C_Vertex3f(0.242393017f, 0.495322227f, 0.f),
    C_Vertex3f(0.24888134f, 0.432078123f, 0.f),
    C_Vertex3f(0.24778986f, 0.508313179f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.119706154f, 0.483446121f, 0.f),
    C_Vertex3f(0.125146866f, 0.419758797f, 0.f),
    C_Vertex3f(0.13049984f, 0.47268486f, 0.f),
    C_Vertex3f(0.135831833f, 0.415331841f, 0.f),
    C_Vertex3f(0.142411709f, 0.464613914f, 0.f),
    C_Vertex3f(0.146960735f, 0.412011623f, 0.f),
    C_Vertex3f(0.155441761f, 0.459233284f, 0.f),
    C_Vertex3f(0.158533573f, 0.409798145f, 0.f),
    C_Vertex3f(0.169589996f, 0.456542969f, 0.f),
    C_Vertex3f(0.170550346f, 0.408691406f, 0.f),
    C_Vertex3f(0.184745789f, 0.456542969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.105108738f, 0.431932926f, 0.f),
    C_Vertex3f(0.0957555771f, 0.620378494f, 0.f),
    C_Vertex3f(0.0957555771f, 0.439680099f, 0.f),
    C_Vertex3f(0.0868463516f, 0.611524582f, 0.f),
    C_Vertex3f(0.0868463516f, 0.448534012f, 0.f),
    C_Vertex3f(0.0790507793f, 0.602205992f, 0.f),
    C_Vertex3f(0.0790507793f, 0.45786643f, 0.f),
    C_Vertex3f(0.0723688602f, 0.592422724f, 0.f),
    C_Vertex3f(0.0723688602f, 0.467677355f, 0.f),
    C_Vertex3f(0.0668005943f, 0.582174778f, 0.f),
    C_Vertex3f(0.0668005943f, 0.477966785f, 0.f),
    C_Vertex3f(0.0623459816f, 0.571462154f, 0.f),
    C_Vertex3f(0.0623459816f, 0.488734722f, 0.f),
    C_Vertex3f(0.059005022f, 0.560284853f, 0.f),
    C_Vertex3f(0.059005022f, 0.499981165f, 0.f),
    C_Vertex3f(0.0567777157f, 0.548642874f, 0.f),
    C_Vertex3f(0.0567777157f, 0.511706114f, 0.f),
    C_Vertex3f(0.0556640625f, 0.536536217f, 0.f),
    C_Vertex3f(0.0556640625f, 0.523909569f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.119706154f, 0.483446121f, 0.f),
    C_Vertex3f(0.114905834f, 0.425292492f, 0.f),
    C_Vertex3f(0.125146866f, 0.419758797f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.250488281f, 0.537639618f, -1.f),
    C_Vertex3f(0.258248329f, 0.439873695f, -1.f),
    C_Vertex3f(0.258497238f, 0.620378494f, -1.f),
    C_Vertex3f(0.267157555f, 0.448782921f, -1.f),
    C_Vertex3f(0.267351151f, 0.611524582f, -1.f),
    C_Vertex3f(0.274953127f, 0.45814991f, -1.f),
    C_Vertex3f(0.275098324f, 0.602205992f, -1.f),
    C_Vertex3f(0.281635046f, 0.467974663f, -1.f),
    C_Vertex3f(0.281738758f, 0.592422724f, -1.f),
    C_Vertex3f(0.287203312f, 0.478257179f, -1.f),
    C_Vertex3f(0.287272453f, 0.582174778f, -1.f),
    C_Vertex3f(0.291657925f, 0.488997459f, -1.f),
    C_Vertex3f(0.291699409f, 0.571462154f, -1.f),
    C_Vertex3f(0.294998884f, 0.500195503f, -1.f),
    C_Vertex3f(0.295019627f, 0.560284853f, -1.f),
    C_Vertex3f(0.297226191f, 0.511851311f, -1.f),
    C_Vertex3f(0.297233105f, 0.548642874f, -1.f),
    C_Vertex3f(0.298339844f, 0.523964882f, -1.f),
    C_Vertex3f(0.298339844f, 0.536536217f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.24916482f, 0.628125668f, -1.f),
    C_Vertex3f(0.239353895f, 0.634766102f, -1.f),
    C_Vertex3f(0.242393017f, 0.564736366f, -1.f),
    C_Vertex3f(0.24778986f, 0.551745415f, -1.f),
    C_Vertex3f(0.250488281f, 0.537639618f, -1.f),
    C_Vertex3f(0.258497238f, 0.620378494f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.242393017f, 0.564736366f, -1.f),
    C_Vertex3f(0.239353895f, 0.634766102f, -1.f),
    C_Vertex3f(0.234297752f, 0.576612473f, -1.f),
    C_Vertex3f(0.229064465f, 0.640299797f, -1.f),
    C_Vertex3f(0.223504066f, 0.587373734f, -1.f),
    C_Vertex3f(0.218296528f, 0.644726753f, -1.f),
    C_Vertex3f(0.211647511f, 0.595444679f, -1.f),
    C_Vertex3f(0.207050085f, 0.64804697f, -1.f),
    C_Vertex3f(0.198728085f, 0.60082531f, -1.f),
    C_Vertex3f(0.195325136f, 0.650260448f, -1.f),
    C_Vertex3f(0.184745789f, 0.603515625f, -1.f),
    C_Vertex3f(0.183121681f, 0.651367188f, -1.f),
    C_Vertex3f(0.170550346f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.106214046f, 0.551745415f, -1.f),
    C_Vertex3f(0.111610889f, 0.564736366f, -1.f),
    C_Vertex3f(0.114905834f, 0.634766102f, -1.f),
    C_Vertex3f(0.119706154f, 0.576612473f, -1.f),
    C_Vertex3f(0.125146866f, 0.640299797f, -1.f),
    C_Vertex3f(0.13049984f, 0.587373734f, -1.f),
    C_Vertex3f(0.135831833f, 0.644726753f, -1.f),
    C_Vertex3f(0.142411709f, 0.595444679f, -1.f),
    C_Vertex3f(0.146960735f, 0.64804697f, -1.f),
    C_Vertex3f(0.155441761f, 0.60082531f, -1.f),
    C_Vertex3f(0.158533573f, 0.650260448f, -1.f),
    C_Vertex3f(0.169589996f, 0.603515625f, -1.f),
    C_Vertex3f(0.170550346f, 0.651367188f, -1.f),
    C_Vertex3f(0.184745789f, 0.603515625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.119706154f, 0.483446121f, -1.f),
    C_Vertex3f(0.111610889f, 0.495322227f, -1.f),
    C_Vertex3f(0.114905834f, 0.425292492f, -1.f),
    C_Vertex3f(0.106214046f, 0.508313179f, -1.f),
    C_Vertex3f(0.105108738f, 0.431932926f, -1.f),
    C_Vertex3f(0.103515625f, 0.522418976f, -1.f),
    C_Vertex3f(0.0957555771f, 0.620378494f, -1.f),
    C_Vertex3f(0.103515625f, 0.537639618f, -1.f),
    C_Vertex3f(0.105108738f, 0.628125668f, -1.f),
    C_Vertex3f(0.106214046f, 0.551745415f, -1.f),
    C_Vertex3f(0.114905834f, 0.634766102f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.250488281f, 0.522418976f, -1.f),
    C_Vertex3f(0.24778986f, 0.508313179f, -1.f),
    C_Vertex3f(0.24888134f, 0.432078123f, -1.f),
    C_Vertex3f(0.258248329f, 0.439873695f, -1.f),
    C_Vertex3f(0.250488281f, 0.537639618f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.170550346f, 0.408691406f, -1.f),
    C_Vertex3f(0.183066368f, 0.408691406f, -1.f),
    C_Vertex3f(0.184745789f, 0.456542969f, -1.f),
    C_Vertex3f(0.195179939f, 0.409805059f, -1.f),
    C_Vertex3f(0.198728085f, 0.459233284f, -1.f),
    C_Vertex3f(0.206835747f, 0.412032366f, -1.f),
    C_Vertex3f(0.211647511f, 0.464613914f, -1.f),
    C_Vertex3f(0.218033791f, 0.415373325f, -1.f),
    C_Vertex3f(0.223504066f, 0.47268486f, -1.f),
    C_Vertex3f(0.228774071f, 0.419827938f, -1.f),
    C_Vertex3f(0.234297752f, 0.483446121f, -1.f),
    C_Vertex3f(0.239056587f, 0.425396204f, -1.f),
    C_Vertex3f(0.242393017f, 0.495322227f, -1.f),
    C_Vertex3f(0.24888134f, 0.432078123f, -1.f),
    C_Vertex3f(0.24778986f, 0.508313179f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.119706154f, 0.483446121f, -1.f),
    C_Vertex3f(0.125146866f, 0.419758797f, -1.f),
    C_Vertex3f(0.13049984f, 0.47268486f, -1.f),
    C_Vertex3f(0.135831833f, 0.415331841f, -1.f),
    C_Vertex3f(0.142411709f, 0.464613914f, -1.f),
    C_Vertex3f(0.146960735f, 0.412011623f, -1.f),
    C_Vertex3f(0.155441761f, 0.459233284f, -1.f),
    C_Vertex3f(0.158533573f, 0.409798145f, -1.f),
    C_Vertex3f(0.169589996f, 0.456542969f, -1.f),
    C_Vertex3f(0.170550346f, 0.408691406f, -1.f),
    C_Vertex3f(0.184745789f, 0.456542969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.105108738f, 0.431932926f, -1.f),
    C_Vertex3f(0.0957555771f, 0.620378494f, -1.f),
    C_Vertex3f(0.0957555771f, 0.439680099f, -1.f),
    C_Vertex3f(0.0868463516f, 0.611524582f, -1.f),
    C_Vertex3f(0.0868463516f, 0.448534012f, -1.f),
    C_Vertex3f(0.0790507793f, 0.602205992f, -1.f),
    C_Vertex3f(0.0790507793f, 0.45786643f, -1.f),
    C_Vertex3f(0.0723688602f, 0.592422724f, -1.f),
    C_Vertex3f(0.0723688602f, 0.467677355f, -1.f),
    C_Vertex3f(0.0668005943f, 0.582174778f, -1.f),
    C_Vertex3f(0.0668005943f, 0.477966785f, -1.f),
    C_Vertex3f(0.0623459816f, 0.571462154f, -1.f),
    C_Vertex3f(0.0623459816f, 0.488734722f, -1.f),
    C_Vertex3f(0.059005022f, 0.560284853f, -1.f),
    C_Vertex3f(0.059005022f, 0.499981165f, -1.f),
    C_Vertex3f(0.0567777157f, 0.548642874f, -1.f),
    C_Vertex3f(0.0567777157f, 0.511706114f, -1.f),
    C_Vertex3f(0.0556640625f, 0.536536217f, -1.f),
    C_Vertex3f(0.0556640625f, 0.523909569f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.119706154f, 0.483446121f, -1.f),
    C_Vertex3f(0.114905834f, 0.425292492f, -1.f),
    C_Vertex3f(0.125146866f, 0.419758797f, -1.f),
    C_End(),
    C_Translatef(0.357910156f, 0.f, 0.f),
};
static const Call char_177[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.208496094f, 0.114257813f, 0.f),
    C_Vertex3f(0.208496094f, 0.114257813f, -1.f),
    C_Vertex3f(0.208496094f, 0.290039063f, -1.f),
    C_Vertex3f(0.208496094f, 0.290039063f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.208496094f, 0.290039063f, 0.f),
    C_Vertex3f(0.208496094f, 0.290039063f, -1.f),
    C_Vertex3f(0.0341796875f, 0.290039063f, -1.f),
    C_Vertex3f(0.0341796875f, 0.290039063f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0341796875f, 0.290039063f, 0.f),
    C_Vertex3f(0.0341796875f, 0.290039063f, -1.f),
    C_Vertex3f(0.0341796875f, 0.362792969f, -1.f),
    C_Vertex3f(0.0341796875f, 0.362792969f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0341796875f, 0.362792969f, 0.f),
    C_Vertex3f(0.0341796875f, 0.362792969f, -1.f),
    C_Vertex3f(0.208496094f, 0.362792969f, -1.f),
    C_Vertex3f(0.208496094f, 0.362792969f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.208496094f, 0.362792969f, 0.f),
    C_Vertex3f(0.208496094f, 0.362792969f, -1.f),
    C_Vertex3f(0.208496094f, 0.537109375f, -1.f),
    C_Vertex3f(0.208496094f, 0.537109375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.208496094f, 0.537109375f, 0.f),
    C_Vertex3f(0.208496094f, 0.537109375f, -1.f),
    C_Vertex3f(0.282714844f, 0.537109375f, -1.f),
    C_Vertex3f(0.282714844f, 0.537109375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.282714844f, 0.537109375f, 0.f),
    C_Vertex3f(0.282714844f, 0.537109375f, -1.f),
    C_Vertex3f(0.282714844f, 0.362792969f, -1.f),
    C_Vertex3f(0.282714844f, 0.362792969f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.282714844f, 0.362792969f, 0.f),
    C_Vertex3f(0.282714844f, 0.362792969f, -1.f),
    C_Vertex3f(0.45703125f, 0.362792969f, -1.f),
    C_Vertex3f(0.45703125f, 0.362792969f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.45703125f, 0.362792969f, 0.f),
    C_Vertex3f(0.45703125f, 0.362792969f, -1.f),
    C_Vertex3f(0.45703125f, 0.290039063f, -1.f),
    C_Vertex3f(0.45703125f, 0.290039063f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.45703125f, 0.290039063f, 0.f),
    C_Vertex3f(0.45703125f, 0.290039063f, -1.f),
    C_Vertex3f(0.282714844f, 0.290039063f, -1.f),
    C_Vertex3f(0.282714844f, 0.290039063f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.282714844f, 0.290039063f, 0.f),
    C_Vertex3f(0.282714844f, 0.290039063f, -1.f),
    C_Vertex3f(0.282714844f, 0.114257813f, -1.f),
    C_Vertex3f(0.282714844f, 0.114257813f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.282714844f, 0.114257813f, 0.f),
    C_Vertex3f(0.282714844f, 0.114257813f, -1.f),
    C_Vertex3f(0.208496094f, 0.114257813f, -1.f),
    C_Vertex3f(0.208496094f, 0.114257813f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.45703125f, 0.f, 0.f),
    C_Vertex3f(0.45703125f, 0.f, -1.f),
    C_Vertex3f(0.0341796875f, 0.f, -1.f),
    C_Vertex3f(0.0341796875f, 0.f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0341796875f, 0.f, 0.f),
    C_Vertex3f(0.0341796875f, 0.f, -1.f),
    C_Vertex3f(0.0341796875f, 0.0732421875f, -1.f),
    C_Vertex3f(0.0341796875f, 0.0732421875f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0341796875f, 0.0732421875f, 0.f),
    C_Vertex3f(0.0341796875f, 0.0732421875f, -1.f),
    C_Vertex3f(0.45703125f, 0.0732421875f, -1.f),
    C_Vertex3f(0.45703125f, 0.0732421875f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.45703125f, 0.0732421875f, 0.f),
    C_Vertex3f(0.45703125f, 0.0732421875f, -1.f),
    C_Vertex3f(0.45703125f, 0.f, -1.f),
    C_Vertex3f(0.45703125f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.282714844f, 0.537109375f, 0.f),
    C_Vertex3f(0.208496094f, 0.537109375f, 0.f),
    C_Vertex3f(0.282714844f, 0.114257813f, 0.f),
    C_Vertex3f(0.208496094f, 0.362792969f, 0.f),
    C_Vertex3f(0.208496094f, 0.290039063f, 0.f),
    C_Vertex3f(0.0341796875f, 0.362792969f, 0.f),
    C_Vertex3f(0.0341796875f, 0.290039063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.282714844f, 0.290039063f, 0.f),
    C_Vertex3f(0.45703125f, 0.290039063f, 0.f),
    C_Vertex3f(0.282714844f, 0.362792969f, 0.f),
    C_Vertex3f(0.282714844f, 0.537109375f, 0.f),
    C_Vertex3f(0.282714844f, 0.114257813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0341796875f, 0.0732421875f, 0.f),
    C_Vertex3f(0.0341796875f, 0.f, 0.f),
    C_Vertex3f(0.45703125f, 0.f, 0.f),
    C_Vertex3f(0.45703125f, 0.0732421875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.282714844f, 0.362792969f, 0.f),
    C_Vertex3f(0.45703125f, 0.290039063f, 0.f),
    C_Vertex3f(0.45703125f, 0.362792969f, 0.f),
    C_Vertex3f(0.282714844f, 0.114257813f, 0.f),
    C_Vertex3f(0.208496094f, 0.290039063f, 0.f),
    C_Vertex3f(0.208496094f, 0.114257813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.282714844f, 0.537109375f, -1.f),
    C_Vertex3f(0.208496094f, 0.537109375f, -1.f),
    C_Vertex3f(0.282714844f, 0.114257813f, -1.f),
    C_Vertex3f(0.208496094f, 0.362792969f, -1.f),
    C_Vertex3f(0.208496094f, 0.290039063f, -1.f),
    C_Vertex3f(0.0341796875f, 0.362792969f, -1.f),
    C_Vertex3f(0.0341796875f, 0.290039063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.282714844f, 0.290039063f, -1.f),
    C_Vertex3f(0.45703125f, 0.290039063f, -1.f),
    C_Vertex3f(0.282714844f, 0.362792969f, -1.f),
    C_Vertex3f(0.282714844f, 0.537109375f, -1.f),
    C_Vertex3f(0.282714844f, 0.114257813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0341796875f, 0.0732421875f, -1.f),
    C_Vertex3f(0.0341796875f, 0.f, -1.f),
    C_Vertex3f(0.45703125f, 0.f, -1.f),
    C_Vertex3f(0.45703125f, 0.0732421875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.282714844f, 0.362792969f, -1.f),
    C_Vertex3f(0.45703125f, 0.290039063f, -1.f),
    C_Vertex3f(0.45703125f, 0.362792969f, -1.f),
    C_Vertex3f(0.282714844f, 0.114257813f, -1.f),
    C_Vertex3f(0.208496094f, 0.290039063f, -1.f),
    C_Vertex3f(0.208496094f, 0.114257813f, -1.f),
    C_End(),
    C_Translatef(0.491210938f, 0.f, 0.f),
};
static const Call char_178[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.994465232f, 0.105066352f, 0.f),
    C_Vertex3f(0.0107421875f, 0.3203125f, 0.f),
    C_Vertex3f(0.0107421875f, 0.3203125f, -1.f),
    C_Normal3f(-0.978970349f, 0.204002276f, 0.f),
    C_Vertex3f(0.0113983154f, 0.326522827f, 0.f),
    C_Vertex3f(0.0113983154f, 0.326522827f, -1.f),
    C_Normal3f(-0.947692752f, 0.319184124f, 0.f),
    C_Vertex3f(0.0146245956f, 0.338943481f, 0.f),
    C_Vertex3f(0.0146245956f, 0.338943481f, -1.f),
    C_Normal3f(-0.897670329f, 0.440667629f, 0.f),
    C_Vertex3f(0.0197649002f, 0.351364136f, 0.f),
    C_Vertex3f(0.0197649002f, 0.351364136f, -1.f),
    C_Normal3f(-0.850162208f, 0.526520908f, 0.f),
    C_Vertex3f(0.0268192291f, 0.36378479f, 0.f),
    C_Vertex3f(0.0268192291f, 0.36378479f, -1.f),
    C_Normal3f(-0.808503449f, 0.58849144f, 0.f),
    C_Vertex3f(0.0313034058f, 0.369995117f, 0.f),
    C_Vertex3f(0.0313034058f, 0.369995117f, -1.f),
    C_Normal3f(-0.780852914f, 0.624714911f, 0.f),
    C_Vertex3f(0.0347499847f, 0.374680519f, 0.f),
    C_Vertex3f(0.0347499847f, 0.374680519f, -1.f),
    C_Normal3f(-0.751654983f, 0.659556508f, 0.f),
    C_Vertex3f(0.0427504778f, 0.384303212f, 0.f),
    C_Vertex3f(0.0427504778f, 0.384303212f, -1.f),
    C_Normal3f(-0.719192505f, 0.694810867f, 0.f),
    C_Vertex3f(0.0518583059f, 0.394177794f, 0.f),
    C_Vertex3f(0.0518583059f, 0.394177794f, -1.f),
    C_Normal3f(-0.68952769f, 0.724259257f, 0.f),
    C_Vertex3f(0.0620734692f, 0.404304266f, 0.f),
    C_Vertex3f(0.0620734692f, 0.404304266f, -1.f),
    C_Normal3f(-0.662519574f, 0.749044478f, 0.f),
    C_Vertex3f(0.0733959675f, 0.414682627f, 0.f),
    C_Vertex3f(0.0733959675f, 0.414682627f, -1.f),
    C_Normal3f(-0.637962878f, 0.770067096f, 0.f),
    C_Vertex3f(0.0858258009f, 0.425312877f, 0.f),
    C_Vertex3f(0.0858258009f, 0.425312877f, -1.f),
    C_Normal3f(-0.615630746f, 0.788034737f, 0.f),
    C_Vertex3f(0.0993629694f, 0.436195016f, 0.f),
    C_Vertex3f(0.0993629694f, 0.436195016f, -1.f),
    C_Normal3f(-0.595298111f, 0.803504884f, 0.f),
    C_Vertex3f(0.114007473f, 0.447329044f, 0.f),
    C_Vertex3f(0.114007473f, 0.447329044f, -1.f),
    C_Normal3f(-0.588331699f, 0.808619618f, 0.f),
    C_Vertex3f(0.129759312f, 0.458714962f, 0.f),
    C_Vertex3f(0.129759312f, 0.458714962f, -1.f),
    C_Normal3f(-0.594203293f, 0.804314971f, 0.f),
    C_Vertex3f(0.143911242f, 0.469086289f, 0.f),
    C_Vertex3f(0.143911242f, 0.469086289f, -1.f),
    C_Normal3f(-0.601538539f, 0.798843801f, 0.f),
    C_Vertex3f(0.156463265f, 0.478443027f, 0.f),
    C_Vertex3f(0.156463265f, 0.478443027f, -1.f),
    C_Normal3f(-0.610959113f, 0.791662157f, 0.f),
    C_Vertex3f(0.16741538f, 0.486785173f, 0.f),
    C_Vertex3f(0.16741538f, 0.486785173f, -1.f),
    C_Normal3f(-0.623493552f, 0.781828523f, 0.f),
    C_Vertex3f(0.176767588f, 0.49411273f, 0.f),
    C_Vertex3f(0.176767588f, 0.49411273f, -1.f),
    C_Normal3f(-0.640969217f, 0.767566562f, 0.f),
    C_Vertex3f(0.184519887f, 0.500425696f, 0.f),
    C_Vertex3f(0.184519887f, 0.500425696f, -1.f),
    C_Normal3f(-0.666953921f, 0.745098889f, 0.f),
    C_Vertex3f(0.190672278f, 0.505724072f, 0.f),
    C_Vertex3f(0.190672278f, 0.505724072f, -1.f),
    C_Normal3f(-0.717711151f, 0.696341038f, 0.f),
    C_Vertex3f(0.195224762f, 0.510007858f, 0.f),
    C_Vertex3f(0.195224762f, 0.510007858f, -1.f),
    C_Normal3f(-0.794008613f, 0.60790658f, 0.f),
    C_Vertex3f(0.200748444f, 0.516109467f, 0.f),
    C_Vertex3f(0.200748444f, 0.516109467f, -1.f),
    C_Normal3f(-0.873941123f, 0.48603183f, 0.f),
    C_Vertex3f(0.206819534f, 0.525152206f, 0.f),
    C_Vertex3f(0.206819534f, 0.525152206f, -1.f),
    C_Normal3f(-0.949701548f, 0.313156515f, 0.f),
    C_Vertex3f(0.210866928f, 0.534303665f, 0.f),
    C_Vertex3f(0.210866928f, 0.534303665f, -1.f),
    C_Normal3f(-0.994476676f, 0.1049577f, 0.f),
    C_Vertex3f(0.212890625f, 0.543563843f, 0.f),
    C_Vertex3f(0.212890625f, 0.543563843f, -1.f),
    C_Normal3f(-0.995184839f, -0.098015435f, 0.f),
    C_Vertex3f(0.212890625f, 0.553478241f, 0.f),
    C_Vertex3f(0.212890625f, 0.553478241f, -1.f),
    C_Normal3f(-0.953319848f, -0.301962286f, 0.f),
    C_Vertex3f(0.21097517f, 0.563012123f, 0.f),
    C_Vertex3f(0.21097517f, 0.563012123f, -1.f),
    C_Normal3f(-0.862047732f, -0.506827176f, 0.f),
    C_Vertex3f(0.20714426f, 0.571619987f, 0.f),
    C_Vertex3f(0.20714426f, 0.571619987f, -1.f),
    C_Normal3f(-0.732748687f, -0.680499434f, 0.f),
    C_Vertex3f(0.201397896f, 0.579301834f, 0.f),
    C_Vertex3f(0.201397896f, 0.579301834f, -1.f),
    C_Normal3f(-0.568057418f, -0.822988987f, 0.f),
    C_Vertex3f(0.193736076f, 0.586057663f, 0.f),
    C_Vertex3f(0.193736076f, 0.586057663f, -1.f),
    C_Normal3f(-0.377140611f, -0.926155925f, 0.f),
    C_Vertex3f(0.184269428f, 0.591124535f, 0.f),
    C_Vertex3f(0.184269428f, 0.591124535f, -1.f),
    C_Normal3f(-0.203739092f, -0.979025245f, 0.f),
    C_Vertex3f(0.172997952f, 0.594502449f, 0.f),
    C_Vertex3f(0.172997952f, 0.594502449f, -1.f),
    C_Normal3f(-0.0607727244f, -0.99815172f, 0.f),
    C_Vertex3f(0.159921646f, 0.596191406f, 0.f),
    C_Vertex3f(0.159921646f, 0.596191406f, -1.f),
    C_Normal3f(0.0469901972f, -0.998895288f, 0.f),
    C_Vertex3f(0.14525795f, 0.596191406f, 0.f),
    C_Vertex3f(0.14525795f, 0.596191406f, -1.f),
    C_Normal3f(0.162369981f, -0.98672998f, 0.f),
    C_Vertex3f(0.132644653f, 0.594908237f, 0.f),
    C_Vertex3f(0.132644653f, 0.594908237f, -1.f),
    C_Normal3f(0.30927524f, -0.950972497f, 0.f),
    C_Vertex3f(0.121864319f, 0.5923419f, 0.f),
    C_Vertex3f(0.121864319f, 0.5923419f, -1.f),
    C_Normal3f(0.488089651f, -0.872793555f, 0.f),
    C_Vertex3f(0.112916946f, 0.588492393f, 0.f),
    C_Vertex3f(0.112916946f, 0.588492393f, -1.f),
    C_Normal3f(0.680839121f, -0.732433021f, 0.f),
    C_Vertex3f(0.105802536f, 0.583359718f, 0.f),
    C_Vertex3f(0.105802536f, 0.583359718f, -1.f),
    C_Normal3f(0.81558001f, -0.578644395f, 0.f),
    C_Vertex3f(0.0993175507f, 0.575850964f, 0.f),
    C_Vertex3f(0.0993175507f, 0.575850964f, -1.f),
    C_Normal3f(0.894284606f, -0.44749853f, 0.f),
    C_Vertex3f(0.0934619904f, 0.565966129f, 0.f),
    C_Vertex3f(0.0934619904f, 0.565966129f, -1.f),
    C_Normal3f(0.933442175f, -0.358727694f, 0.f),
    C_Vertex3f(0.0882358551f, 0.553705215f, 0.f),
    C_Vertex3f(0.0882358551f, 0.553705215f, -1.f),
    C_Normal3f(0.954058647f, -0.299619675f, 0.f),
    C_Vertex3f(0.0859375f, 0.546386719f, 0.f),
    C_Vertex3f(0.0859375f, 0.546386719f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.103150524f, -0.994665742f, 0.f),
    C_Vertex3f(0.0859375f, 0.546386719f, 0.f),
    C_Vertex3f(0.0859375f, 0.546386719f, -1.f),
    C_Vertex3f(0.0200195313f, 0.553222656f, -1.f),
    C_Vertex3f(0.0200195313f, 0.553222656f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.968093276f, 0.250590175f, 0.f),
    C_Vertex3f(0.0200195313f, 0.553222656f, 0.f),
    C_Vertex3f(0.0200195313f, 0.553222656f, -1.f),
    C_Normal3f(-0.956814468f, 0.290699214f, 0.f),
    C_Vertex3f(0.0216026306f, 0.55933857f, 0.f),
    C_Vertex3f(0.0216026306f, 0.55933857f, -1.f),
    C_Normal3f(-0.939018965f, 0.343865395f, 0.f),
    C_Vertex3f(0.0253491402f, 0.570764661f, 0.f),
    C_Vertex3f(0.0253491402f, 0.570764661f, -1.f),
    C_Normal3f(-0.911282718f, 0.411781281f, 0.f),
    C_Vertex3f(0.0296759605f, 0.581385016f, 0.f),
    C_Vertex3f(0.0296759605f, 0.581385016f, -1.f),
    C_Normal3f(-0.875396848f, 0.483405024f, 0.f),
    C_Vertex3f(0.0345830917f, 0.591199636f, 0.f),
    C_Vertex3f(0.0345830917f, 0.591199636f, -1.f),
    C_Normal3f(-0.830253601f, 0.557385743f, 0.f),
    C_Vertex3f(0.0400705338f, 0.600208521f, 0.f),
    C_Vertex3f(0.0400705338f, 0.600208521f, -1.f),
    C_Normal3f(-0.775131643f, 0.631799757f, 0.f),
    C_Vertex3f(0.0461382866f, 0.60841167f, 0.f),
    C_Vertex3f(0.0461382866f, 0.60841167f, -1.f),
    C_Normal3f(-0.709959626f, 0.704242349f, 0.f),
    C_Vertex3f(0.0527863503f, 0.615809083f, 0.f),
    C_Vertex3f(0.0527863503f, 0.615809083f, -1.f),
    C_Normal3f(-0.635528505f, 0.772077382f, 0.f),
    C_Vertex3f(0.0600147247f, 0.622400761f, 0.f),
    C_Vertex3f(0.0600147247f, 0.622400761f, -1.f),
    C_Normal3f(-0.550064921f, 0.83512187f, 0.f),
    C_Vertex3f(0.06782341f, 0.628186703f, 0.f),
    C_Vertex3f(0.06782341f, 0.628186703f, -1.f),
    C_Normal3f(-0.459427536f, 0.888215303f, 0.f),
    C_Vertex3f(0.0764853954f, 0.633249402f, 0.f),
    C_Vertex3f(0.0764853954f, 0.633249402f, -1.f),
    C_Normal3f(-0.371476054f, 0.928442478f, 0.f),
    C_Vertex3f(0.0860006809f, 0.637588859f, 0.f),
    C_Vertex3f(0.0860006809f, 0.637588859f, -1.f),
    C_Normal3f(-0.288651258f, 0.957434237f, 0.f),
    C_Vertex3f(0.0963692665f, 0.641205072f, 0.f),
    C_Vertex3f(0.0963692665f, 0.641205072f, -1.f),
    C_Normal3f(-0.212354288f, 0.9771927f, 0.f),
    C_Vertex3f(0.107591152f, 0.644098043f, 0.f),
    C_Vertex3f(0.107591152f, 0.644098043f, -1.f),
    C_Normal3f(-0.143138036f, 0.989702702f, 0.f),
    C_Vertex3f(0.119666338f, 0.646267772f, 0.f),
    C_Vertex3f(0.119666338f, 0.646267772f, -1.f),
    C_Normal3f(-0.0809652954f, 0.996716976f, 0.f),
    C_Vertex3f(0.132594824f, 0.647714257f, 0.f),
    C_Vertex3f(0.132594824f, 0.647714257f, -1.f),
    C_Normal3f(-0.0246388894f, 0.999696434f, 0.f),
    C_Vertex3f(0.14637661f, 0.6484375f, 0.f),
    C_Vertex3f(0.14637661f, 0.6484375f, -1.f),
    C_Normal3f(0.0260243267f, 0.999661326f, 0.f),
    C_Vertex3f(0.161939621f, 0.6484375f, 0.f),
    C_Vertex3f(0.161939621f, 0.6484375f, -1.f),
    C_Normal3f(0.081172131f, 0.996700108f, 0.f),
    C_Vertex3f(0.177331567f, 0.647631645f, 0.f),
    C_Vertex3f(0.177331567f, 0.647631645f, -1.f),
    C_Normal3f(0.145038739f, 0.989425957f, 0.f),
    C_Vertex3f(0.191624522f, 0.646019936f, 0.f),
    C_Vertex3f(0.191624522f, 0.646019936f, -1.f),
    C_Normal3f(0.217710793f, 0.976013362f, 0.f),
    C_Vertex3f(0.204818487f, 0.643602371f, 0.f),
    C_Vertex3f(0.204818487f, 0.643602371f, -1.f),
    C_Normal3f(0.299658716f, 0.954046547f, 0.f),
    C_Vertex3f(0.216913462f, 0.640378952f, 0.f),
    C_Vertex3f(0.216913462f, 0.640378952f, -1.f),
    C_Normal3f(0.390576601f, 0.920570552f, 0.f),
    C_Vertex3f(0.227909446f, 0.636349678f, 0.f),
    C_Vertex3f(0.227909446f, 0.636349678f, -1.f),
    C_Normal3f(0.488848925f, 0.872368395f, 0.f),
    C_Vertex3f(0.237806439f, 0.631514549f, 0.f),
    C_Vertex3f(0.237806439f, 0.631514549f, -1.f),
    C_Normal3f(0.591045976f, 0.806637824f, 0.f),
    C_Vertex3f(0.246604443f, 0.625873566f, 0.f),
    C_Vertex3f(0.246604443f, 0.625873566f, -1.f),
    C_Normal3f(0.676174939f, 0.736741066f, 0.f),
    C_Vertex3f(0.254303455f, 0.619426727f, 0.f),
    C_Vertex3f(0.254303455f, 0.619426727f, -1.f),
    C_Normal3f(0.744687021f, 0.667413831f, 0.f),
    C_Vertex3f(0.261040092f, 0.612624645f, 0.f),
    C_Vertex3f(0.261040092f, 0.612624645f, -1.f),
    C_Normal3f(0.810909569f, 0.585171521f, 0.f),
    C_Vertex3f(0.266814351f, 0.605467319f, 0.f),
    C_Vertex3f(0.266814351f, 0.605467319f, -1.f),
    C_Normal3f(0.871335208f, 0.490688264f, 0.f),
    C_Vertex3f(0.271626234f, 0.59795475f, 0.f),
    C_Vertex3f(0.271626234f, 0.59795475f, -1.f),
    C_Normal3f(0.922422051f, 0.386183083f, 0.f),
    C_Vertex3f(0.27547574f, 0.590086937f, 0.f),
    C_Vertex3f(0.27547574f, 0.590086937f, -1.f),
    C_Normal3f(0.961350024f, 0.275329173f, 0.f),
    C_Vertex3f(0.27836287f, 0.58186388f, 0.f),
    C_Vertex3f(0.27836287f, 0.58186388f, -1.f),
    C_Normal3f(0.986680329f, 0.162671283f, 0.f),
    C_Vertex3f(0.280287623f, 0.57328558f, 0.f),
    C_Vertex3f(0.280287623f, 0.57328558f, -1.f),
    C_Normal3f(0.999106944f, 0.0422531627f, 0.f),
    C_Vertex3f(0.28125f, 0.564352036f, 0.f),
    C_Vertex3f(0.28125f, 0.564352036f, -1.f),
    C_Normal3f(0.994880974f, -0.101054065f, 0.f),
    C_Vertex3f(0.28125f, 0.55052948f, 0.f),
    C_Vertex3f(0.28125f, 0.55052948f, -1.f),
    C_Normal3f(0.96444875f, -0.264270067f, 0.f),
    C_Vertex3f(0.278031349f, 0.532664299f, 0.f),
    C_Vertex3f(0.278031349f, 0.532664299f, -1.f),
    C_Normal3f(0.905099988f, -0.425198704f, 0.f),
    C_Vertex3f(0.271594048f, 0.51529026f, 0.f),
    C_Vertex3f(0.271594048f, 0.51529026f, -1.f),
    C_Normal3f(0.841618061f, -0.540073156f, 0.f),
    C_Vertex3f(0.261938095f, 0.498407364f, 0.f),
    C_Vertex3f(0.261938095f, 0.498407364f, -1.f),
    C_Normal3f(0.784726143f, -0.619842589f, 0.f),
    C_Vertex3f(0.255500793f, 0.490211487f, 0.f),
    C_Vertex3f(0.255500793f, 0.490211487f, -1.f),
    C_Normal3f(0.745443285f, -0.666569054f, 0.f),
    C_Vertex3f(0.253046989f, 0.487151146f, 0.f),
    C_Vertex3f(0.253046989f, 0.487151146f, -1.f),
    C_Normal3f(0.710426807f, -0.703770995f, 0.f),
    C_Vertex3f(0.246570826f, 0.480224848f, 0.f),
    C_Vertex3f(0.246570826f, 0.480224848f, -1.f),
    C_Normal3f(0.677599728f, -0.735430896f, 0.f),
    C_Vertex3f(0.238526106f, 0.472492933f, 0.f),
    C_Vertex3f(0.238526106f, 0.472492933f, -1.f),
    C_Normal3f(0.651973963f, -0.758241296f, 0.f),
    C_Vertex3f(0.22891283f, 0.463955402f, 0.f),
    C_Vertex3f(0.22891283f, 0.463955402f, -1.f),
    C_Normal3f(0.631509066f, -0.775368512f, 0.f),
    C_Vertex3f(0.217730999f, 0.454612255f, 0.f),
    C_Vertex3f(0.217730999f, 0.454612255f, -1.f),
    C_Normal3f(0.614833713f, -0.788656771f, 0.f),
    C_Vertex3f(0.204980612f, 0.444463491f, 0.f),
    C_Vertex3f(0.204980612f, 0.444463491f, -1.f),
    C_Normal3f(0.601007879f, -0.799243152f, 0.f),
    C_Vertex3f(0.190661669f, 0.433509111f, 0.f),
    C_Vertex3f(0.190661669f, 0.433509111f, -1.f),
    C_Normal3f(0.591180742f, -0.806539059f, 0.f),
    C_Vertex3f(0.17477417f, 0.421749115f, 0.f),
    C_Vertex3f(0.17477417f, 0.421749115f, -1.f),
    C_Normal3f(0.588156223f, -0.808747292f, 0.f),
    C_Vertex3f(0.166046143f, 0.415466309f, 0.f),
    C_Vertex3f(0.166046143f, 0.415466309f, -1.f),
    C_Normal3f(0.603640079f, -0.797256947f, 0.f),
    C_Vertex3f(0.156963348f, 0.40879631f, 0.f),
    C_Vertex3f(0.156963348f, 0.40879631f, -1.f),
    C_Normal3f(0.621596992f, -0.783337235f, 0.f),
    C_Vertex3f(0.141424179f, 0.396823883f, 0.f),
    C_Vertex3f(0.141424179f, 0.396823883f, -1.f),
    C_Normal3f(0.649983168f, -0.759948611f, 0.f),
    C_Vertex3f(0.128511429f, 0.386219025f, 0.f),
    C_Vertex3f(0.128511429f, 0.386219025f, -1.f),
    C_Normal3f(0.682233751f, -0.731134117f, 0.f),
    C_Vertex3f(0.118225098f, 0.376981735f, 0.f),
    C_Vertex3f(0.118225098f, 0.376981735f, -1.f),
    C_Normal3f(0.716594577f, -0.697490036f, 0.f),
    C_Vertex3f(0.114395142f, 0.373046875f, 0.f),
    C_Vertex3f(0.114395142f, 0.373046875f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.114395142f, 0.373046875f, 0.f),
    C_Vertex3f(0.114395142f, 0.373046875f, -1.f),
    C_Vertex3f(0.282714844f, 0.373046875f, -1.f),
    C_Vertex3f(0.282714844f, 0.373046875f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.282714844f, 0.373046875f, 0.f),
    C_Vertex3f(0.282714844f, 0.373046875f, -1.f),
    C_Vertex3f(0.282714844f, 0.3203125f, -1.f),
    C_Vertex3f(0.282714844f, 0.3203125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.282714844f, 0.3203125f, 0.f),
    C_Vertex3f(0.282714844f, 0.3203125f, -1.f),
    C_Vertex3f(0.0107421875f, 0.3203125f, -1.f),
    C_Vertex3f(0.0107421875f, 0.3203125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.114395142f, 0.373046875f, 0.f),
    C_Vertex3f(0.118225098f, 0.376981735f, 0.f),
    C_Vertex3f(0.129759312f, 0.458714962f, 0.f),
    C_Vertex3f(0.114007473f, 0.447329044f, 0.f),
    C_Vertex3f(0.0993629694f, 0.436195016f, 0.f),
    C_Vertex3f(0.0858258009f, 0.425312877f, 0.f),
    C_Vertex3f(0.0733959675f, 0.414682627f, 0.f),
    C_Vertex3f(0.0620734692f, 0.404304266f, 0.f),
    C_Vertex3f(0.0518583059f, 0.394177794f, 0.f),
    C_Vertex3f(0.0427504778f, 0.384303212f, 0.f),
    C_Vertex3f(0.282714844f, 0.3203125f, 0.f),
    C_Vertex3f(0.282714844f, 0.373046875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.282714844f, 0.3203125f, 0.f),
    C_Vertex3f(0.0427504778f, 0.384303212f, 0.f),
    C_Vertex3f(0.0347499847f, 0.374680519f, 0.f),
    C_Vertex3f(0.0313034058f, 0.369995117f, 0.f),
    C_Vertex3f(0.0268192291f, 0.36378479f, 0.f),
    C_Vertex3f(0.0197649002f, 0.351364136f, 0.f),
    C_Vertex3f(0.0146245956f, 0.338943481f, 0.f),
    C_Vertex3f(0.0113983154f, 0.326522827f, 0.f),
    C_Vertex3f(0.0107421875f, 0.3203125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.21097517f, 0.563012123f, 0.f),
    C_Vertex3f(0.212890625f, 0.553478241f, 0.f),
    C_Vertex3f(0.216913462f, 0.640378952f, 0.f),
    C_Vertex3f(0.217730999f, 0.454612255f, 0.f),
    C_Vertex3f(0.227909446f, 0.636349678f, 0.f),
    C_Vertex3f(0.22891283f, 0.463955402f, 0.f),
    C_Vertex3f(0.237806439f, 0.631514549f, 0.f),
    C_Vertex3f(0.238526106f, 0.472492933f, 0.f),
    C_Vertex3f(0.246604443f, 0.625873566f, 0.f),
    C_Vertex3f(0.246570826f, 0.480224848f, 0.f),
    C_Vertex3f(0.253046989f, 0.487151146f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193736076f, 0.586057663f, 0.f),
    C_Vertex3f(0.201397896f, 0.579301834f, 0.f),
    C_Vertex3f(0.204818487f, 0.643602371f, 0.f),
    C_Vertex3f(0.20714426f, 0.571619987f, 0.f),
    C_Vertex3f(0.216913462f, 0.640378952f, 0.f),
    C_Vertex3f(0.21097517f, 0.563012123f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132644653f, 0.594908237f, 0.f),
    C_Vertex3f(0.14525795f, 0.596191406f, 0.f),
    C_Vertex3f(0.14637661f, 0.6484375f, 0.f),
    C_Vertex3f(0.159921646f, 0.596191406f, 0.f),
    C_Vertex3f(0.161939621f, 0.6484375f, 0.f),
    C_Vertex3f(0.172997952f, 0.594502449f, 0.f),
    C_Vertex3f(0.177331567f, 0.647631645f, 0.f),
    C_Vertex3f(0.184269428f, 0.591124535f, 0.f),
    C_Vertex3f(0.191624522f, 0.646019936f, 0.f),
    C_Vertex3f(0.193736076f, 0.586057663f, 0.f),
    C_Vertex3f(0.204818487f, 0.643602371f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0993175507f, 0.575850964f, 0.f),
    C_Vertex3f(0.105802536f, 0.583359718f, 0.f),
    C_Vertex3f(0.107591152f, 0.644098043f, 0.f),
    C_Vertex3f(0.112916946f, 0.588492393f, 0.f),
    C_Vertex3f(0.119666338f, 0.646267772f, 0.f),
    C_Vertex3f(0.121864319f, 0.5923419f, 0.f),
    C_Vertex3f(0.132594824f, 0.647714257f, 0.f),
    C_Vertex3f(0.132644653f, 0.594908237f, 0.f),
    C_Vertex3f(0.14637661f, 0.6484375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0963692665f, 0.641205072f, 0.f),
    C_Vertex3f(0.0860006809f, 0.637588859f, 0.f),
    C_Vertex3f(0.0882358551f, 0.553705215f, 0.f),
    C_Vertex3f(0.0934619904f, 0.565966129f, 0.f),
    C_Vertex3f(0.0993175507f, 0.575850964f, 0.f),
    C_Vertex3f(0.107591152f, 0.644098043f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0859375f, 0.546386719f, 0.f),
    C_Vertex3f(0.0882358551f, 0.553705215f, 0.f),
    C_Vertex3f(0.0860006809f, 0.637588859f, 0.f),
    C_Vertex3f(0.0764853954f, 0.633249402f, 0.f),
    C_Vertex3f(0.06782341f, 0.628186703f, 0.f),
    C_Vertex3f(0.0600147247f, 0.622400761f, 0.f),
    C_Vertex3f(0.0527863503f, 0.615809083f, 0.f),
    C_Vertex3f(0.0461382866f, 0.60841167f, 0.f),
    C_Vertex3f(0.0400705338f, 0.600208521f, 0.f),
    C_Vertex3f(0.0345830917f, 0.591199636f, 0.f),
    C_Vertex3f(0.0296759605f, 0.581385016f, 0.f),
    C_Vertex3f(0.0253491402f, 0.570764661f, 0.f),
    C_Vertex3f(0.0216026306f, 0.55933857f, 0.f),
    C_Vertex3f(0.0200195313f, 0.553222656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.28125f, 0.564352036f, 0.f),
    C_Vertex3f(0.280287623f, 0.57328558f, 0.f),
    C_Vertex3f(0.28125f, 0.55052948f, 0.f),
    C_Vertex3f(0.27836287f, 0.58186388f, 0.f),
    C_Vertex3f(0.278031349f, 0.532664299f, 0.f),
    C_Vertex3f(0.27547574f, 0.590086937f, 0.f),
    C_Vertex3f(0.271626234f, 0.59795475f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.246604443f, 0.625873566f, 0.f),
    C_Vertex3f(0.253046989f, 0.487151146f, 0.f),
    C_Vertex3f(0.254303455f, 0.619426727f, 0.f),
    C_Vertex3f(0.255500793f, 0.490211487f, 0.f),
    C_Vertex3f(0.261040092f, 0.612624645f, 0.f),
    C_Vertex3f(0.261938095f, 0.498407364f, 0.f),
    C_Vertex3f(0.266814351f, 0.605467319f, 0.f),
    C_Vertex3f(0.271594048f, 0.51529026f, 0.f),
    C_Vertex3f(0.271626234f, 0.59795475f, 0.f),
    C_Vertex3f(0.278031349f, 0.532664299f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.217730999f, 0.454612255f, 0.f),
    C_Vertex3f(0.212890625f, 0.553478241f, 0.f),
    C_Vertex3f(0.212890625f, 0.543563843f, 0.f),
    C_Vertex3f(0.210866928f, 0.534303665f, 0.f),
    C_Vertex3f(0.206819534f, 0.525152206f, 0.f),
    C_Vertex3f(0.204980612f, 0.444463491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.204980612f, 0.444463491f, 0.f),
    C_Vertex3f(0.206819534f, 0.525152206f, 0.f),
    C_Vertex3f(0.200748444f, 0.516109467f, 0.f),
    C_Vertex3f(0.195224762f, 0.510007858f, 0.f),
    C_Vertex3f(0.190672278f, 0.505724072f, 0.f),
    C_Vertex3f(0.190661669f, 0.433509111f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.156963348f, 0.40879631f, 0.f),
    C_Vertex3f(0.166046143f, 0.415466309f, 0.f),
    C_Vertex3f(0.16741538f, 0.486785173f, 0.f),
    C_Vertex3f(0.17477417f, 0.421749115f, 0.f),
    C_Vertex3f(0.176767588f, 0.49411273f, 0.f),
    C_Vertex3f(0.190661669f, 0.433509111f, 0.f),
    C_Vertex3f(0.184519887f, 0.500425696f, 0.f),
    C_Vertex3f(0.190672278f, 0.505724072f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.118225098f, 0.376981735f, 0.f),
    C_Vertex3f(0.128511429f, 0.386219025f, 0.f),
    C_Vertex3f(0.129759312f, 0.458714962f, 0.f),
    C_Vertex3f(0.141424179f, 0.396823883f, 0.f),
    C_Vertex3f(0.143911242f, 0.469086289f, 0.f),
    C_Vertex3f(0.156963348f, 0.40879631f, 0.f),
    C_Vertex3f(0.156463265f, 0.478443027f, 0.f),
    C_Vertex3f(0.16741538f, 0.486785173f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.114395142f, 0.373046875f, -1.f),
    C_Vertex3f(0.118225098f, 0.376981735f, -1.f),
    C_Vertex3f(0.129759312f, 0.458714962f, -1.f),
    C_Vertex3f(0.114007473f, 0.447329044f, -1.f),
    C_Vertex3f(0.0993629694f, 0.436195016f, -1.f),
    C_Vertex3f(0.0858258009f, 0.425312877f, -1.f),
    C_Vertex3f(0.0733959675f, 0.414682627f, -1.f),
    C_Vertex3f(0.0620734692f, 0.404304266f, -1.f),
    C_Vertex3f(0.0518583059f, 0.394177794f, -1.f),
    C_Vertex3f(0.0427504778f, 0.384303212f, -1.f),
    C_Vertex3f(0.282714844f, 0.3203125f, -1.f),
    C_Vertex3f(0.282714844f, 0.373046875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.282714844f, 0.3203125f, -1.f),
    C_Vertex3f(0.0427504778f, 0.384303212f, -1.f),
    C_Vertex3f(0.0347499847f, 0.374680519f, -1.f),
    C_Vertex3f(0.0313034058f, 0.369995117f, -1.f),
    C_Vertex3f(0.0268192291f, 0.36378479f, -1.f),
    C_Vertex3f(0.0197649002f, 0.351364136f, -1.f),
    C_Vertex3f(0.0146245956f, 0.338943481f, -1.f),
    C_Vertex3f(0.0113983154f, 0.326522827f, -1.f),
    C_Vertex3f(0.0107421875f, 0.3203125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.21097517f, 0.563012123f, -1.f),
    C_Vertex3f(0.212890625f, 0.553478241f, -1.f),
    C_Vertex3f(0.216913462f, 0.640378952f, -1.f),
    C_Vertex3f(0.217730999f, 0.454612255f, -1.f),
    C_Vertex3f(0.227909446f, 0.636349678f, -1.f),
    C_Vertex3f(0.22891283f, 0.463955402f, -1.f),
    C_Vertex3f(0.237806439f, 0.631514549f, -1.f),
    C_Vertex3f(0.238526106f, 0.472492933f, -1.f),
    C_Vertex3f(0.246604443f, 0.625873566f, -1.f),
    C_Vertex3f(0.246570826f, 0.480224848f, -1.f),
    C_Vertex3f(0.253046989f, 0.487151146f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.193736076f, 0.586057663f, -1.f),
    C_Vertex3f(0.201397896f, 0.579301834f, -1.f),
    C_Vertex3f(0.204818487f, 0.643602371f, -1.f),
    C_Vertex3f(0.20714426f, 0.571619987f, -1.f),
    C_Vertex3f(0.216913462f, 0.640378952f, -1.f),
    C_Vertex3f(0.21097517f, 0.563012123f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132644653f, 0.594908237f, -1.f),
    C_Vertex3f(0.14525795f, 0.596191406f, -1.f),
    C_Vertex3f(0.14637661f, 0.6484375f, -1.f),
    C_Vertex3f(0.159921646f, 0.596191406f, -1.f),
    C_Vertex3f(0.161939621f, 0.6484375f, -1.f),
    C_Vertex3f(0.172997952f, 0.594502449f, -1.f),
    C_Vertex3f(0.177331567f, 0.647631645f, -1.f),
    C_Vertex3f(0.184269428f, 0.591124535f, -1.f),
    C_Vertex3f(0.191624522f, 0.646019936f, -1.f),
    C_Vertex3f(0.193736076f, 0.586057663f, -1.f),
    C_Vertex3f(0.204818487f, 0.643602371f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0993175507f, 0.575850964f, -1.f),
    C_Vertex3f(0.105802536f, 0.583359718f, -1.f),
    C_Vertex3f(0.107591152f, 0.644098043f, -1.f),
    C_Vertex3f(0.112916946f, 0.588492393f, -1.f),
    C_Vertex3f(0.119666338f, 0.646267772f, -1.f),
    C_Vertex3f(0.121864319f, 0.5923419f, -1.f),
    C_Vertex3f(0.132594824f, 0.647714257f, -1.f),
    C_Vertex3f(0.132644653f, 0.594908237f, -1.f),
    C_Vertex3f(0.14637661f, 0.6484375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0963692665f, 0.641205072f, -1.f),
    C_Vertex3f(0.0860006809f, 0.637588859f, -1.f),
    C_Vertex3f(0.0882358551f, 0.553705215f, -1.f),
    C_Vertex3f(0.0934619904f, 0.565966129f, -1.f),
    C_Vertex3f(0.0993175507f, 0.575850964f, -1.f),
    C_Vertex3f(0.107591152f, 0.644098043f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0859375f, 0.546386719f, -1.f),
    C_Vertex3f(0.0882358551f, 0.553705215f, -1.f),
    C_Vertex3f(0.0860006809f, 0.637588859f, -1.f),
    C_Vertex3f(0.0764853954f, 0.633249402f, -1.f),
    C_Vertex3f(0.06782341f, 0.628186703f, -1.f),
    C_Vertex3f(0.0600147247f, 0.622400761f, -1.f),
    C_Vertex3f(0.0527863503f, 0.615809083f, -1.f),
    C_Vertex3f(0.0461382866f, 0.60841167f, -1.f),
    C_Vertex3f(0.0400705338f, 0.600208521f, -1.f),
    C_Vertex3f(0.0345830917f, 0.591199636f, -1.f),
    C_Vertex3f(0.0296759605f, 0.581385016f, -1.f),
    C_Vertex3f(0.0253491402f, 0.570764661f, -1.f),
    C_Vertex3f(0.0216026306f, 0.55933857f, -1.f),
    C_Vertex3f(0.0200195313f, 0.553222656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.28125f, 0.564352036f, -1.f),
    C_Vertex3f(0.280287623f, 0.57328558f, -1.f),
    C_Vertex3f(0.28125f, 0.55052948f, -1.f),
    C_Vertex3f(0.27836287f, 0.58186388f, -1.f),
    C_Vertex3f(0.278031349f, 0.532664299f, -1.f),
    C_Vertex3f(0.27547574f, 0.590086937f, -1.f),
    C_Vertex3f(0.271626234f, 0.59795475f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.246604443f, 0.625873566f, -1.f),
    C_Vertex3f(0.253046989f, 0.487151146f, -1.f),
    C_Vertex3f(0.254303455f, 0.619426727f, -1.f),
    C_Vertex3f(0.255500793f, 0.490211487f, -1.f),
    C_Vertex3f(0.261040092f, 0.612624645f, -1.f),
    C_Vertex3f(0.261938095f, 0.498407364f, -1.f),
    C_Vertex3f(0.266814351f, 0.605467319f, -1.f),
    C_Vertex3f(0.271594048f, 0.51529026f, -1.f),
    C_Vertex3f(0.271626234f, 0.59795475f, -1.f),
    C_Vertex3f(0.278031349f, 0.532664299f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.217730999f, 0.454612255f, -1.f),
    C_Vertex3f(0.212890625f, 0.553478241f, -1.f),
    C_Vertex3f(0.212890625f, 0.543563843f, -1.f),
    C_Vertex3f(0.210866928f, 0.534303665f, -1.f),
    C_Vertex3f(0.206819534f, 0.525152206f, -1.f),
    C_Vertex3f(0.204980612f, 0.444463491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.204980612f, 0.444463491f, -1.f),
    C_Vertex3f(0.206819534f, 0.525152206f, -1.f),
    C_Vertex3f(0.200748444f, 0.516109467f, -1.f),
    C_Vertex3f(0.195224762f, 0.510007858f, -1.f),
    C_Vertex3f(0.190672278f, 0.505724072f, -1.f),
    C_Vertex3f(0.190661669f, 0.433509111f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.156963348f, 0.40879631f, -1.f),
    C_Vertex3f(0.166046143f, 0.415466309f, -1.f),
    C_Vertex3f(0.16741538f, 0.486785173f, -1.f),
    C_Vertex3f(0.17477417f, 0.421749115f, -1.f),
    C_Vertex3f(0.176767588f, 0.49411273f, -1.f),
    C_Vertex3f(0.190661669f, 0.433509111f, -1.f),
    C_Vertex3f(0.184519887f, 0.500425696f, -1.f),
    C_Vertex3f(0.190672278f, 0.505724072f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.118225098f, 0.376981735f, -1.f),
    C_Vertex3f(0.128511429f, 0.386219025f, -1.f),
    C_Vertex3f(0.129759312f, 0.458714962f, -1.f),
    C_Vertex3f(0.141424179f, 0.396823883f, -1.f),
    C_Vertex3f(0.143911242f, 0.469086289f, -1.f),
    C_Vertex3f(0.156963348f, 0.40879631f, -1.f),
    C_Vertex3f(0.156463265f, 0.478443027f, -1.f),
    C_Vertex3f(0.16741538f, 0.486785173f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_179[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.0987515822f, 0.995112062f, 0.f),
    C_Vertex3f(0.0146484375f, 0.40625f, 0.f),
    C_Vertex3f(0.0146484375f, 0.40625f, -1.f),
    C_Vertex3f(0.0786132813f, 0.412597656f, -1.f),
    C_Vertex3f(0.0786132813f, 0.412597656f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.948600233f, 0.316476732f, 0.f),
    C_Vertex3f(0.0786132813f, 0.412597656f, 0.f),
    C_Vertex3f(0.0786132813f, 0.412597656f, -1.f),
    C_Normal3f(0.925142825f, 0.379619241f, 0.f),
    C_Vertex3f(0.0807952881f, 0.406057358f, 0.f),
    C_Vertex3f(0.0807952881f, 0.406057358f, -1.f),
    C_Normal3f(0.88133961f, 0.472483307f, 0.f),
    C_Vertex3f(0.085814476f, 0.395048141f, 0.f),
    C_Vertex3f(0.085814476f, 0.395048141f, -1.f),
    C_Normal3f(0.796340883f, 0.604848087f, 0.f),
    C_Vertex3f(0.0914888382f, 0.386110306f, 0.f),
    C_Vertex3f(0.0914888382f, 0.386110306f, -1.f),
    C_Normal3f(0.653571963f, 0.756864369f, 0.f),
    C_Vertex3f(0.0978183746f, 0.379243851f, 0.f),
    C_Vertex3f(0.0978183746f, 0.379243851f, -1.f),
    C_Normal3f(0.497298121f, 0.867579758f, 0.f),
    C_Vertex3f(0.106002808f, 0.373577118f, 0.f),
    C_Vertex3f(0.106002808f, 0.373577118f, -1.f),
    C_Normal3f(0.35698247f, 0.934111059f, 0.f),
    C_Vertex3f(0.116259575f, 0.368673325f, 0.f),
    C_Vertex3f(0.116259575f, 0.368673325f, -1.f),
    C_Normal3f(0.207388401f, 0.97825861f, 0.f),
    C_Vertex3f(0.127388954f, 0.365404129f, 0.f),
    C_Vertex3f(0.127388954f, 0.365404129f, -1.f),
    C_Normal3f(0.0622990876f, 0.998057544f, 0.f),
    C_Vertex3f(0.139390945f, 0.363769531f, 0.f),
    C_Vertex3f(0.139390945f, 0.363769531f, -1.f),
    C_Normal3f(-0.0755289048f, 0.997143626f, 0.f),
    C_Vertex3f(0.153575897f, 0.363769531f, 0.f),
    C_Vertex3f(0.153575897f, 0.363769531f, -1.f),
    C_Normal3f(-0.235443294f, 0.971888006f, 0.f),
    C_Vertex3f(0.167543411f, 0.365901947f, 0.f),
    C_Vertex3f(0.167543411f, 0.365901947f, -1.f),
    C_Normal3f(-0.415341794f, 0.909665406f, 0.f),
    C_Vertex3f(0.179983139f, 0.370166779f, 0.f),
    C_Vertex3f(0.179983139f, 0.370166779f, -1.f),
    C_Normal3f(-0.592475772f, 0.805588365f, 0.f),
    C_Vertex3f(0.190895081f, 0.376564026f, 0.f),
    C_Vertex3f(0.190895081f, 0.376564026f, -1.f),
    C_Normal3f(-0.737504065f, 0.675342739f, 0.f),
    C_Vertex3f(0.200279236f, 0.385093689f, 0.f),
    C_Vertex3f(0.200279236f, 0.385093689f, -1.f),
    C_Normal3f(-0.859020531f, 0.511941075f, 0.f),
    C_Vertex3f(0.207317352f, 0.394497871f, 0.f),
    C_Vertex3f(0.207317352f, 0.394497871f, -1.f),
    C_Normal3f(-0.950081706f, 0.312001199f, 0.f),
    C_Vertex3f(0.21200943f, 0.404776573f, 0.f),
    C_Vertex3f(0.21200943f, 0.404776573f, -1.f),
    C_Normal3f(-0.994671762f, 0.103092164f, 0.f),
    C_Vertex3f(0.214355469f, 0.415929794f, 0.f),
    C_Vertex3f(0.214355469f, 0.415929794f, -1.f),
    C_Normal3f(-0.99584192f, -0.0910978988f, 0.f),
    C_Vertex3f(0.214355469f, 0.427412033f, 0.f),
    C_Vertex3f(0.214355469f, 0.427412033f, -1.f),
    C_Normal3f(-0.953854024f, -0.300270617f, 0.f),
    C_Vertex3f(0.212404728f, 0.437254429f, 0.f),
    C_Vertex3f(0.212404728f, 0.437254429f, -1.f),
    C_Normal3f(-0.859505892f, -0.511125743f, 0.f),
    C_Vertex3f(0.208503246f, 0.446002483f, 0.f),
    C_Vertex3f(0.208503246f, 0.446002483f, -1.f),
    C_Normal3f(-0.721117973f, -0.692812264f, 0.f),
    C_Vertex3f(0.202651024f, 0.453656197f, 0.f),
    C_Vertex3f(0.202651024f, 0.453656197f, -1.f),
    C_Normal3f(-0.529917777f, -0.848048925f, 0.f),
    C_Vertex3f(0.194848061f, 0.460215569f, 0.f),
    C_Vertex3f(0.194848061f, 0.460215569f, -1.f),
    C_Normal3f(-0.324515879f, -0.945880234f, 0.f),
    C_Vertex3f(0.184280872f, 0.465135098f, 0.f),
    C_Vertex3f(0.184280872f, 0.465135098f, -1.f),
    C_Normal3f(-0.164888799f, -0.986312151f, 0.f),
    C_Vertex3f(0.170949459f, 0.468414783f, 0.f),
    C_Vertex3f(0.170949459f, 0.468414783f, -1.f),
    C_Normal3f(-0.0582021512f, -0.998304844f, 0.f),
    C_Vertex3f(0.154853821f, 0.470054626f, 0.f),
    C_Vertex3f(0.154853821f, 0.470054626f, -1.f),
    C_Normal3f(0.0226627011f, -0.999743164f, 0.f),
    C_Vertex3f(0.142822266f, 0.470054626f, 0.f),
    C_Vertex3f(0.142822266f, 0.470054626f, -1.f),
    C_Normal3f(0.0741111487f, -0.997250021f, 0.f),
    C_Vertex3f(0.135669708f, 0.469619751f, 0.f),
    C_Vertex3f(0.135669708f, 0.469619751f, -1.f),
    C_Normal3f(0.0856996402f, -0.996320963f, 0.f),
    C_Vertex3f(0.131118774f, 0.469184875f, 0.f),
    C_Vertex3f(0.131118774f, 0.469184875f, -1.f),
    C_Normal3f(0.0556593053f, -0.998449802f, 0.f),
    C_Vertex3f(0.12808609f, 0.468967438f, 0.f),
    C_Vertex3f(0.12808609f, 0.468967438f, -1.f),
    C_Normal3f(0.0334124193f, -0.999441683f, 0.f),
    C_Vertex3f(0.123317719f, 0.46875f, 0.f),
    C_Vertex3f(0.123317719f, 0.46875f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.121582031f, 0.46875f, 0.f),
    C_Vertex3f(0.121582031f, 0.46875f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.980853558f, 0.194746494f, 0.f),
    C_Vertex3f(0.121582031f, 0.46875f, 0.f),
    C_Vertex3f(0.121582031f, 0.46875f, -1.f),
    C_Vertex3f(0.13117981f, 0.517089844f, -1.f),
    C_Vertex3f(0.13117981f, 0.517089844f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.012307425f, 0.999924302f, 0.f),
    C_Vertex3f(0.13117981f, 0.517089844f, 0.f),
    C_Vertex3f(0.13117981f, 0.517089844f, -1.f),
    C_Normal3f(-0.0561989658f, 0.998419583f, 0.f),
    C_Vertex3f(0.140012741f, 0.516981125f, 0.f),
    C_Vertex3f(0.140012741f, 0.516981125f, -1.f),
    C_Normal3f(-0.161512494f, 0.986870706f, 0.f),
    C_Vertex3f(0.155306339f, 0.518447876f, 0.f),
    C_Vertex3f(0.155306339f, 0.518447876f, -1.f),
    C_Normal3f(-0.322124451f, 0.946697354f, 0.f),
    C_Vertex3f(0.168227673f, 0.521598816f, 0.f),
    C_Vertex3f(0.168227673f, 0.521598816f, -1.f),
    C_Normal3f(-0.518483162f, 0.855087817f, 0.f),
    C_Vertex3f(0.178776741f, 0.526433945f, 0.f),
    C_Vertex3f(0.178776741f, 0.526433945f, -1.f),
    C_Normal3f(-0.69272387f, 0.72120285f, 0.f),
    C_Vertex3f(0.186953545f, 0.532953262f, 0.f),
    C_Vertex3f(0.186953545f, 0.532953262f, -1.f),
    C_Normal3f(-0.82908088f, 0.559128761f, 0.f),
    C_Vertex3f(0.193086147f, 0.540178299f, 0.f),
    C_Vertex3f(0.193086147f, 0.540178299f, -1.f),
    C_Normal3f(-0.937811553f, 0.347144663f, 0.f),
    C_Vertex3f(0.197174549f, 0.548109055f, 0.f),
    C_Vertex3f(0.197174549f, 0.548109055f, -1.f),
    C_Normal3f(-0.993113458f, 0.117156781f, 0.f),
    C_Vertex3f(0.19921875f, 0.556745529f, 0.f),
    C_Vertex3f(0.19921875f, 0.556745529f, -1.f),
    C_Normal3f(-0.994821727f, -0.101634219f, 0.f),
    C_Vertex3f(0.19921875f, 0.565437317f, 0.f),
    C_Vertex3f(0.19921875f, 0.565437317f, -1.f),
    C_Normal3f(-0.945004106f, -0.327058583f, 0.f),
    C_Vertex3f(0.197570324f, 0.572880745f, 0.f),
    C_Vertex3f(0.197570324f, 0.572880745f, -1.f),
    C_Normal3f(-0.846279621f, -0.532738864f, 0.f),
    C_Vertex3f(0.194273472f, 0.579726219f, 0.f),
    C_Vertex3f(0.194273472f, 0.579726219f, -1.f),
    C_Normal3f(-0.717827618f, -0.696220875f, 0.f),
    C_Vertex3f(0.189328194f, 0.58597374f, 0.f),
    C_Vertex3f(0.189328194f, 0.58597374f, -1.f),
    C_Normal3f(-0.555961907f, -0.831207812f, 0.f),
    C_Vertex3f(0.182734489f, 0.591623306f, 0.f),
    C_Vertex3f(0.182734489f, 0.591623306f, -1.f),
    C_Normal3f(-0.365763187f, -0.930707991f, 0.f),
    C_Vertex3f(0.174546719f, 0.595860481f, 0.f),
    C_Vertex3f(0.174546719f, 0.595860481f, -1.f),
    C_Normal3f(-0.196366847f, -0.980530441f, 0.f),
    C_Vertex3f(0.164764881f, 0.598685265f, 0.f),
    C_Vertex3f(0.164764881f, 0.598685265f, -1.f),
    C_Normal3f(-0.0587092414f, -0.998275161f, 0.f),
    C_Vertex3f(0.153388977f, 0.600097656f, 0.f),
    C_Vertex3f(0.153388977f, 0.600097656f, -1.f),
    C_Normal3f(0.0512446761f, -0.998686135f, 0.f),
    C_Vertex3f(0.140748978f, 0.600097656f, 0.f),
    C_Vertex3f(0.140748978f, 0.600097656f, -1.f),
    C_Normal3f(0.173851714f, -0.984771848f, 0.f),
    C_Vertex3f(0.129785061f, 0.59888649f, 0.f),
    C_Vertex3f(0.129785061f, 0.59888649f, -1.f),
    C_Normal3f(0.320640087f, -0.947201133f, 0.f),
    C_Vertex3f(0.120167255f, 0.596464157f, 0.f),
    C_Vertex3f(0.120167255f, 0.596464157f, -1.f),
    C_Normal3f(0.487189293f, -0.87329638f, 0.f),
    C_Vertex3f(0.111895561f, 0.592830658f, 0.f),
    C_Vertex3f(0.111895561f, 0.592830658f, -1.f),
    C_Normal3f(0.657247066f, -0.753675222f, 0.f),
    C_Vertex3f(0.104969978f, 0.587985992f, 0.f),
    C_Vertex3f(0.104969978f, 0.587985992f, -1.f),
    C_Normal3f(0.782326162f, -0.622868896f, 0.f),
    C_Vertex3f(0.0985121727f, 0.581159592f, 0.f),
    C_Vertex3f(0.0985121727f, 0.581159592f, -1.f),
    C_Normal3f(0.862240911f, -0.506498337f, 0.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, 0.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, -1.f),
    C_Normal3f(0.905491531f, -0.424364328f, 0.f),
    C_Vertex3f(0.0869998932f, 0.561561584f, 0.f),
    C_Vertex3f(0.0869998932f, 0.561561584f, -1.f),
    C_Normal3f(0.929830372f, -0.367988437f, 0.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, 0.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.161898106f, -0.986807525f, 0.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, 0.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, -1.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, -1.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.932396173f, 0.361437947f, 0.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, 0.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, -1.f),
    C_Normal3f(-0.905484498f, 0.424379319f, 0.f),
    C_Vertex3f(0.0264606476f, 0.577007294f, 0.f),
    C_Vertex3f(0.0264606476f, 0.577007294f, -1.f),
    C_Normal3f(-0.857638955f, 0.514252305f, 0.f),
    C_Vertex3f(0.0366129875f, 0.59666729f, 0.f),
    C_Vertex3f(0.0366129875f, 0.59666729f, -1.f),
    C_Normal3f(-0.770776927f, 0.637105107f, 0.f),
    C_Vertex3f(0.0479416847f, 0.612832069f, 0.f),
    C_Vertex3f(0.0479416847f, 0.612832069f, -1.f),
    C_Normal3f(-0.640562534f, 0.767905951f, 0.f),
    C_Vertex3f(0.0604467392f, 0.625501633f, 0.f),
    C_Vertex3f(0.0604467392f, 0.625501633f, -1.f),
    C_Normal3f(-0.465689421f, 0.884948194f, 0.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, 0.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, -1.f),
    C_Normal3f(-0.297496587f, 0.954722881f, 0.f),
    C_Vertex3f(0.0909562111f, 0.64155674f, 0.f),
    C_Vertex3f(0.0909562111f, 0.64155674f, -1.f),
    C_Normal3f(-0.157664075f, 0.9874928f, 0.f),
    C_Vertex3f(0.11093092f, 0.646143913f, 0.f),
    C_Vertex3f(0.11093092f, 0.646143913f, -1.f),
    C_Normal3f(-0.0518968254f, 0.998652458f, 0.f),
    C_Vertex3f(0.134052277f, 0.6484375f, 0.f),
    C_Vertex3f(0.134052277f, 0.6484375f, -1.f),
    C_Normal3f(0.0203911792f, 0.999792099f, 0.f),
    C_Vertex3f(0.15506649f, 0.6484375f, 0.f),
    C_Vertex3f(0.15506649f, 0.6484375f, -1.f),
    C_Normal3f(0.076960817f, 0.997034073f, 0.f),
    C_Vertex3f(0.169752955f, 0.64770937f, 0.f),
    C_Vertex3f(0.169752955f, 0.64770937f, -1.f),
    C_Normal3f(0.137886435f, 0.990448117f, 0.f),
    C_Vertex3f(0.183365464f, 0.646253109f, 0.f),
    C_Vertex3f(0.183365464f, 0.646253109f, -1.f),
    C_Normal3f(0.207712278f, 0.978190005f, 0.f),
    C_Vertex3f(0.195904016f, 0.644068718f, 0.f),
    C_Vertex3f(0.195904016f, 0.644068718f, -1.f),
    C_Normal3f(0.287211269f, 0.957867205f, 0.f),
    C_Vertex3f(0.207368612f, 0.641156197f, 0.f),
    C_Vertex3f(0.207368612f, 0.641156197f, -1.f),
    C_Normal3f(0.376511544f, 0.926411867f, 0.f),
    C_Vertex3f(0.217759252f, 0.637515545f, 0.f),
    C_Vertex3f(0.217759252f, 0.637515545f, -1.f),
    C_Normal3f(0.474512786f, 0.880248606f, 0.f),
    C_Vertex3f(0.227075934f, 0.633146763f, 0.f),
    C_Vertex3f(0.227075934f, 0.633146763f, -1.f),
    C_Normal3f(0.58870846f, 0.808345437f, 0.f),
    C_Vertex3f(0.235318661f, 0.62804985f, 0.f),
    C_Vertex3f(0.235318661f, 0.62804985f, -1.f),
    C_Normal3f(0.712377429f, 0.701796532f, 0.f),
    C_Vertex3f(0.246071815f, 0.619312286f, 0.f),
    C_Vertex3f(0.246071815f, 0.619312286f, -1.f),
    C_Normal3f(0.838954687f, 0.544201314f, 0.f),
    C_Vertex3f(0.25682497f, 0.606219292f, 0.f),
    C_Vertex3f(0.25682497f, 0.606219292f, -1.f),
    C_Normal3f(0.943152666f, 0.33235985f, 0.f),
    C_Vertex3f(0.26399374f, 0.591683388f, 0.f),
    C_Vertex3f(0.26399374f, 0.591683388f, -1.f),
    C_Normal3f(0.993242919f, 0.116053827f, 0.f),
    C_Vertex3f(0.267578125f, 0.575704575f, 0.f),
    C_Vertex3f(0.267578125f, 0.575704575f, -1.f),
    C_Normal3f(0.997228563f, -0.0743987635f, 0.f),
    C_Vertex3f(0.267578125f, 0.561006546f, 0.f),
    C_Vertex3f(0.267578125f, 0.561006546f, -1.f),
    C_Normal3f(0.966676414f, -0.256001443f, 0.f),
    C_Vertex3f(0.265636921f, 0.549685001f, 0.f),
    C_Vertex3f(0.265636921f, 0.549685001f, -1.f),
    C_Normal3f(0.90528971f, -0.424794763f, 0.f),
    C_Vertex3f(0.261754513f, 0.539016247f, 0.f),
    C_Vertex3f(0.261754513f, 0.539016247f, -1.f),
    C_Normal3f(0.818774521f, -0.574115276f, 0.f),
    C_Vertex3f(0.255930901f, 0.529000282f, 0.f),
    C_Vertex3f(0.255930901f, 0.529000282f, -1.f),
    C_Normal3f(0.710521519f, -0.703675508f, 0.f),
    C_Vertex3f(0.248166084f, 0.519637108f, 0.f),
    C_Vertex3f(0.248166084f, 0.519637108f, -1.f),
    C_Normal3f(0.581858337f, -0.813290119f, 0.f),
    C_Vertex3f(0.238678455f, 0.51157999f, 0.f),
    C_Vertex3f(0.238678455f, 0.51157999f, -1.f),
    C_Normal3f(0.450883567f, -0.892582715f, 0.f),
    C_Vertex3f(0.227468014f, 0.50482893f, 0.f),
    C_Vertex3f(0.227468014f, 0.50482893f, -1.f),
    C_Normal3f(0.347733289f, -0.93759346f, 0.f),
    C_Vertex3f(0.21453476f, 0.499383926f, 0.f),
    C_Vertex3f(0.21453476f, 0.499383926f, -1.f),
    C_Normal3f(0.271775454f, -0.96236068f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.240741223f, 0.97058934f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, -1.f),
    C_Normal3f(0.313972771f, 0.949431956f, 0.f),
    C_Vertex3f(0.216918945f, 0.494905472f, 0.f),
    C_Vertex3f(0.216918945f, 0.494905472f, -1.f),
    C_Normal3f(0.417560071f, 0.908649325f, 0.f),
    C_Vertex3f(0.233778477f, 0.488527298f, 0.f),
    C_Vertex3f(0.233778477f, 0.488527298f, -1.f),
    C_Normal3f(0.556630671f, 0.830760002f, 0.f),
    C_Vertex3f(0.248073101f, 0.480588913f, 0.f),
    C_Vertex3f(0.248073101f, 0.480588913f, -1.f),
    C_Normal3f(0.701332808f, 0.71283406f, 0.f),
    C_Vertex3f(0.259802818f, 0.471090317f, 0.f),
    C_Vertex3f(0.259802818f, 0.471090317f, -1.f),
    C_Normal3f(0.821135819f, 0.570732832f, 0.f),
    C_Vertex3f(0.268967628f, 0.460031509f, 0.f),
    C_Vertex3f(0.268967628f, 0.460031509f, -1.f),
    C_Normal3f(0.909034133f, 0.416721582f, 0.f),
    C_Vertex3f(0.275841236f, 0.448015213f, 0.f),
    C_Vertex3f(0.275841236f, 0.448015213f, -1.f),
    C_Normal3f(0.968881249f, 0.247526288f, 0.f),
    C_Vertex3f(0.280423641f, 0.435041428f, 0.f),
    C_Vertex3f(0.280423641f, 0.435041428f, -1.f),
    C_Normal3f(0.996273696f, 0.0862477496f, 0.f),
    C_Vertex3f(0.282714844f, 0.421110153f, 0.f),
    C_Vertex3f(0.282714844f, 0.421110153f, -1.f),
    C_Normal3f(0.998839438f, -0.04816401f, 0.f),
    C_Vertex3f(0.282714844f, 0.408575058f, 0.f),
    C_Vertex3f(0.282714844f, 0.408575058f, -1.f),
    C_Normal3f(0.986319363f, -0.164845675f, 0.f),
    C_Vertex3f(0.281635284f, 0.398721933f, 0.f),
    C_Vertex3f(0.281635284f, 0.398721933f, -1.f),
    C_Normal3f(0.960858941f, -0.277037889f, 0.f),
    C_Vertex3f(0.279476166f, 0.389197111f, 0.f),
    C_Vertex3f(0.279476166f, 0.389197111f, -1.f),
    C_Normal3f(0.922533095f, -0.385917842f, 0.f),
    C_Vertex3f(0.276237488f, 0.380000591f, 0.f),
    C_Vertex3f(0.276237488f, 0.380000591f, -1.f),
    C_Normal3f(0.873200536f, -0.487361073f, 0.f),
    C_Vertex3f(0.27191925f, 0.371132374f, 0.f),
    C_Vertex3f(0.27191925f, 0.371132374f, -1.f),
    C_Normal3f(0.815806508f, -0.578324974f, 0.f),
    C_Vertex3f(0.266521454f, 0.362592459f, 0.f),
    C_Vertex3f(0.266521454f, 0.362592459f, -1.f),
    C_Normal3f(0.753707826f, -0.657209754f, 0.f),
    C_Vertex3f(0.260044098f, 0.354380846f, 0.f),
    C_Vertex3f(0.260044098f, 0.354380846f, -1.f),
    C_Normal3f(0.690031052f, -0.723779738f, 0.f),
    C_Vertex3f(0.252487183f, 0.346497536f, 0.f),
    C_Vertex3f(0.252487183f, 0.346497536f, -1.f),
    C_Normal3f(0.614450276f, -0.788955569f, 0.f),
    C_Vertex3f(0.243850708f, 0.338942528f, 0.f),
    C_Vertex3f(0.243850708f, 0.338942528f, -1.f),
    C_Normal3f(0.52276057f, -0.852479517f, 0.f),
    C_Vertex3f(0.234298468f, 0.332331896f, 0.f),
    C_Vertex3f(0.234298468f, 0.332331896f, -1.f),
    C_Normal3f(0.429344833f, -0.903140724f, 0.f),
    C_Vertex3f(0.223830462f, 0.32666564f, 0.f),
    C_Vertex3f(0.223830462f, 0.32666564f, -1.f),
    C_Normal3f(0.337783158f, -0.941223919f, 0.f),
    C_Vertex3f(0.21244669f, 0.32194376f, 0.f),
    C_Vertex3f(0.21244669f, 0.32194376f, -1.f),
    C_Normal3f(0.250808328f, -0.968036711f, 0.f),
    C_Vertex3f(0.200147152f, 0.318166256f, 0.f),
    C_Vertex3f(0.200147152f, 0.318166256f, -1.f),
    C_Normal3f(0.170151472f, -0.985417962f, 0.f),
    C_Vertex3f(0.186931849f, 0.315333128f, 0.f),
    C_Vertex3f(0.186931849f, 0.315333128f, -1.f),
    C_Normal3f(0.0966438875f, -0.995319009f, 0.f),
    C_Vertex3f(0.172800779f, 0.313444376f, 0.f),
    C_Vertex3f(0.172800779f, 0.313444376f, -1.f),
    C_Normal3f(0.0307664964f, -0.99952662f, 0.f),
    C_Vertex3f(0.157753944f, 0.3125f, 0.f),
    C_Vertex3f(0.157753944f, 0.3125f, -1.f),
    C_Normal3f(-0.0252892748f, -0.999680161f, 0.f),
    C_Vertex3f(0.142120361f, 0.3125f, 0.f),
    C_Vertex3f(0.142120361f, 0.3125f, -1.f),
    C_Normal3f(-0.0814212263f, -0.996679723f, 0.f),
    C_Vertex3f(0.127717614f, 0.31325984f, 0.f),
    C_Vertex3f(0.127717614f, 0.31325984f, -1.f),
    C_Normal3f(-0.144044951f, -0.989571154f, 0.f),
    C_Vertex3f(0.114216685f, 0.31477952f, 0.f),
    C_Vertex3f(0.114216685f, 0.31477952f, -1.f),
    C_Normal3f(-0.213851959f, -0.976866007f, 0.f),
    C_Vertex3f(0.101617575f, 0.31705904f, 0.f),
    C_Vertex3f(0.101617575f, 0.31705904f, -1.f),
    C_Normal3f(-0.290886581f, -0.956757545f, 0.f),
    C_Vertex3f(0.0899202824f, 0.3200984f, 0.f),
    C_Vertex3f(0.0899202824f, 0.3200984f, -1.f),
    C_Normal3f(-0.374579161f, -0.927194834f, 0.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, 0.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, -1.f),
    C_Normal3f(-0.463473737f, -0.886110604f, 0.f),
    C_Vertex3f(0.0692311525f, 0.32845664f, 0.f),
    C_Vertex3f(0.0692311525f, 0.32845664f, -1.f),
    C_Normal3f(-0.564197719f, -0.825639844f, 0.f),
    C_Vertex3f(0.060239315f, 0.33377552f, 0.f),
    C_Vertex3f(0.060239315f, 0.33377552f, -1.f),
    C_Normal3f(-0.684651256f, -0.728870809f, 0.f),
    C_Vertex3f(0.0481042862f, 0.3428936f, 0.f),
    C_Vertex3f(0.0481042862f, 0.3428936f, -1.f),
    C_Normal3f(-0.805000246f, -0.593274474f, 0.f),
    C_Vertex3f(0.0348210335f, 0.357651711f, 0.f),
    C_Vertex3f(0.0348210335f, 0.357651711f, -1.f),
    C_Normal3f(-0.901867867f, -0.432012022f, 0.f),
    C_Vertex3f(0.0244345665f, 0.37501049f, 0.f),
    C_Vertex3f(0.0244345665f, 0.37501049f, -1.f),
    C_Normal3f(-0.954272866f, -0.298936754f, 0.f),
    C_Vertex3f(0.0169448853f, 0.39496994f, 0.f),
    C_Vertex3f(0.0169448853f, 0.39496994f, -1.f),
    C_Normal3f(-0.979899347f, -0.199492529f, 0.f),
    C_Vertex3f(0.0146484375f, 0.40625f, 0.f),
    C_Vertex3f(0.0146484375f, 0.40625f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.248166084f, 0.519637108f, 0.f),
    C_Vertex3f(0.255930901f, 0.529000282f, 0.f),
    C_Vertex3f(0.25682497f, 0.606219292f, 0.f),
    C_Vertex3f(0.261754513f, 0.539016247f, 0.f),
    C_Vertex3f(0.26399374f, 0.591683388f, 0.f),
    C_Vertex3f(0.265636921f, 0.549685001f, 0.f),
    C_Vertex3f(0.267578125f, 0.575704575f, 0.f),
    C_Vertex3f(0.267578125f, 0.561006546f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227075934f, 0.633146763f, 0.f),
    C_Vertex3f(0.227468014f, 0.50482893f, 0.f),
    C_Vertex3f(0.235318661f, 0.62804985f, 0.f),
    C_Vertex3f(0.238678455f, 0.51157999f, 0.f),
    C_Vertex3f(0.246071815f, 0.619312286f, 0.f),
    C_Vertex3f(0.248166084f, 0.519637108f, 0.f),
    C_Vertex3f(0.25682497f, 0.606219292f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.202651024f, 0.453656197f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, 0.f),
    C_Vertex3f(0.207368612f, 0.641156197f, 0.f),
    C_Vertex3f(0.21453476f, 0.499383926f, 0.f),
    C_Vertex3f(0.217759252f, 0.637515545f, 0.f),
    C_Vertex3f(0.227468014f, 0.50482893f, 0.f),
    C_Vertex3f(0.227075934f, 0.633146763f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.189328194f, 0.58597374f, 0.f),
    C_Vertex3f(0.194273472f, 0.579726219f, 0.f),
    C_Vertex3f(0.195904016f, 0.644068718f, 0.f),
    C_Vertex3f(0.197570324f, 0.572880745f, 0.f),
    C_Vertex3f(0.207368612f, 0.641156197f, 0.f),
    C_Vertex3f(0.19921875f, 0.565437317f, 0.f),
    C_Vertex3f(0.202651024f, 0.453656197f, 0.f),
    C_Vertex3f(0.19921875f, 0.556745529f, 0.f),
    C_Vertex3f(0.197174549f, 0.548109055f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.183365464f, 0.646253109f, 0.f),
    C_Vertex3f(0.169752955f, 0.64770937f, 0.f),
    C_Vertex3f(0.174546719f, 0.595860481f, 0.f),
    C_Vertex3f(0.182734489f, 0.591623306f, 0.f),
    C_Vertex3f(0.189328194f, 0.58597374f, 0.f),
    C_Vertex3f(0.195904016f, 0.644068718f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.140748978f, 0.600097656f, 0.f),
    C_Vertex3f(0.153388977f, 0.600097656f, 0.f),
    C_Vertex3f(0.15506649f, 0.6484375f, 0.f),
    C_Vertex3f(0.164764881f, 0.598685265f, 0.f),
    C_Vertex3f(0.169752955f, 0.64770937f, 0.f),
    C_Vertex3f(0.174546719f, 0.595860481f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052277f, 0.6484375f, 0.f),
    C_Vertex3f(0.11093092f, 0.646143913f, 0.f),
    C_Vertex3f(0.111895561f, 0.592830658f, 0.f),
    C_Vertex3f(0.120167255f, 0.596464157f, 0.f),
    C_Vertex3f(0.129785061f, 0.59888649f, 0.f),
    C_Vertex3f(0.140748978f, 0.600097656f, 0.f),
    C_Vertex3f(0.15506649f, 0.6484375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.11093092f, 0.646143913f, 0.f),
    C_Vertex3f(0.0909562111f, 0.64155674f, 0.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, 0.f),
    C_Vertex3f(0.0985121727f, 0.581159592f, 0.f),
    C_Vertex3f(0.104969978f, 0.587985992f, 0.f),
    C_Vertex3f(0.111895561f, 0.592830658f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0909562111f, 0.64155674f, 0.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, 0.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, 0.f),
    C_Vertex3f(0.0869998932f, 0.561561584f, 0.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0844726563f, 0.555175781f, 0.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, 0.f),
    C_Vertex3f(0.0604467392f, 0.625501633f, 0.f),
    C_Vertex3f(0.0479416847f, 0.612832069f, 0.f),
    C_Vertex3f(0.0366129875f, 0.59666729f, 0.f),
    C_Vertex3f(0.0264606476f, 0.577007294f, 0.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.212404728f, 0.437254429f, 0.f),
    C_Vertex3f(0.214355469f, 0.427412033f, 0.f),
    C_Vertex3f(0.216918945f, 0.494905472f, 0.f),
    C_Vertex3f(0.223830462f, 0.32666564f, 0.f),
    C_Vertex3f(0.233778477f, 0.488527298f, 0.f),
    C_Vertex3f(0.234298468f, 0.332331896f, 0.f),
    C_Vertex3f(0.248073101f, 0.480588913f, 0.f),
    C_Vertex3f(0.243850708f, 0.338942528f, 0.f),
    C_Vertex3f(0.252487183f, 0.346497536f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.208503246f, 0.446002483f, 0.f),
    C_Vertex3f(0.212404728f, 0.437254429f, 0.f),
    C_Vertex3f(0.216918945f, 0.494905472f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, 0.f),
    C_Vertex3f(0.202651024f, 0.453656197f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276237488f, 0.380000591f, 0.f),
    C_Vertex3f(0.279476166f, 0.389197111f, 0.f),
    C_Vertex3f(0.280423641f, 0.435041428f, 0.f),
    C_Vertex3f(0.281635284f, 0.398721933f, 0.f),
    C_Vertex3f(0.282714844f, 0.421110153f, 0.f),
    C_Vertex3f(0.282714844f, 0.408575058f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.260044098f, 0.354380846f, 0.f),
    C_Vertex3f(0.266521454f, 0.362592459f, 0.f),
    C_Vertex3f(0.268967628f, 0.460031509f, 0.f),
    C_Vertex3f(0.27191925f, 0.371132374f, 0.f),
    C_Vertex3f(0.275841236f, 0.448015213f, 0.f),
    C_Vertex3f(0.276237488f, 0.380000591f, 0.f),
    C_Vertex3f(0.280423641f, 0.435041428f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259802818f, 0.471090317f, 0.f),
    C_Vertex3f(0.248073101f, 0.480588913f, 0.f),
    C_Vertex3f(0.252487183f, 0.346497536f, 0.f),
    C_Vertex3f(0.260044098f, 0.354380846f, 0.f),
    C_Vertex3f(0.268967628f, 0.460031509f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214355469f, 0.415929794f, 0.f),
    C_Vertex3f(0.21200943f, 0.404776573f, 0.f),
    C_Vertex3f(0.21244669f, 0.32194376f, 0.f),
    C_Vertex3f(0.223830462f, 0.32666564f, 0.f),
    C_Vertex3f(0.214355469f, 0.427412033f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.21200943f, 0.404776573f, 0.f),
    C_Vertex3f(0.207317352f, 0.394497871f, 0.f),
    C_Vertex3f(0.21244669f, 0.32194376f, 0.f),
    C_Vertex3f(0.200279236f, 0.385093689f, 0.f),
    C_Vertex3f(0.200147152f, 0.318166256f, 0.f),
    C_Vertex3f(0.190895081f, 0.376564026f, 0.f),
    C_Vertex3f(0.186931849f, 0.315333128f, 0.f),
    C_Vertex3f(0.179983139f, 0.370166779f, 0.f),
    C_Vertex3f(0.172800779f, 0.313444376f, 0.f),
    C_Vertex3f(0.167543411f, 0.365901947f, 0.f),
    C_Vertex3f(0.157753944f, 0.3125f, 0.f),
    C_Vertex3f(0.153575897f, 0.363769531f, 0.f),
    C_Vertex3f(0.142120361f, 0.3125f, 0.f),
    C_Vertex3f(0.139390945f, 0.363769531f, 0.f),
    C_Vertex3f(0.127717614f, 0.31325984f, 0.f),
    C_Vertex3f(0.127388954f, 0.365404129f, 0.f),
    C_Vertex3f(0.116259575f, 0.368673325f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127717614f, 0.31325984f, 0.f),
    C_Vertex3f(0.116259575f, 0.368673325f, 0.f),
    C_Vertex3f(0.114216685f, 0.31477952f, 0.f),
    C_Vertex3f(0.106002808f, 0.373577118f, 0.f),
    C_Vertex3f(0.101617575f, 0.31705904f, 0.f),
    C_Vertex3f(0.0978183746f, 0.379243851f, 0.f),
    C_Vertex3f(0.0914888382f, 0.386110306f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0899202824f, 0.3200984f, 0.f),
    C_Vertex3f(0.101617575f, 0.31705904f, 0.f),
    C_Vertex3f(0.0914888382f, 0.386110306f, 0.f),
    C_Vertex3f(0.085814476f, 0.395048141f, 0.f),
    C_Vertex3f(0.0807952881f, 0.406057358f, 0.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0786132813f, 0.412597656f, 0.f),
    C_Vertex3f(0.0146484375f, 0.40625f, 0.f),
    C_Vertex3f(0.0169448853f, 0.39496994f, 0.f),
    C_Vertex3f(0.0244345665f, 0.37501049f, 0.f),
    C_Vertex3f(0.0348210335f, 0.357651711f, 0.f),
    C_Vertex3f(0.0481042862f, 0.3428936f, 0.f),
    C_Vertex3f(0.060239315f, 0.33377552f, 0.f),
    C_Vertex3f(0.0692311525f, 0.32845664f, 0.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, 0.f),
    C_Vertex3f(0.0807952881f, 0.406057358f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.194848061f, 0.460215569f, 0.f),
    C_Vertex3f(0.202651024f, 0.453656197f, 0.f),
    C_Vertex3f(0.197174549f, 0.548109055f, 0.f),
    C_Vertex3f(0.193086147f, 0.540178299f, 0.f),
    C_Vertex3f(0.186953545f, 0.532953262f, 0.f),
    C_Vertex3f(0.184280872f, 0.465135098f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.178776741f, 0.526433945f, 0.f),
    C_Vertex3f(0.168227673f, 0.521598816f, 0.f),
    C_Vertex3f(0.170949459f, 0.468414783f, 0.f),
    C_Vertex3f(0.184280872f, 0.465135098f, 0.f),
    C_Vertex3f(0.186953545f, 0.532953262f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155306339f, 0.518447876f, 0.f),
    C_Vertex3f(0.140012741f, 0.516981125f, 0.f),
    C_Vertex3f(0.142822266f, 0.470054626f, 0.f),
    C_Vertex3f(0.154853821f, 0.470054626f, 0.f),
    C_Vertex3f(0.170949459f, 0.468414783f, 0.f),
    C_Vertex3f(0.168227673f, 0.521598816f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12808609f, 0.468967438f, 0.f),
    C_Vertex3f(0.131118774f, 0.469184875f, 0.f),
    C_Vertex3f(0.13117981f, 0.517089844f, 0.f),
    C_Vertex3f(0.135669708f, 0.469619751f, 0.f),
    C_Vertex3f(0.140012741f, 0.516981125f, 0.f),
    C_Vertex3f(0.142822266f, 0.470054626f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.13117981f, 0.517089844f, 0.f),
    C_Vertex3f(0.121582031f, 0.46875f, 0.f),
    C_Vertex3f(0.123317719f, 0.46875f, 0.f),
    C_Vertex3f(0.12808609f, 0.468967438f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.248166084f, 0.519637108f, -1.f),
    C_Vertex3f(0.255930901f, 0.529000282f, -1.f),
    C_Vertex3f(0.25682497f, 0.606219292f, -1.f),
    C_Vertex3f(0.261754513f, 0.539016247f, -1.f),
    C_Vertex3f(0.26399374f, 0.591683388f, -1.f),
    C_Vertex3f(0.265636921f, 0.549685001f, -1.f),
    C_Vertex3f(0.267578125f, 0.575704575f, -1.f),
    C_Vertex3f(0.267578125f, 0.561006546f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227075934f, 0.633146763f, -1.f),
    C_Vertex3f(0.227468014f, 0.50482893f, -1.f),
    C_Vertex3f(0.235318661f, 0.62804985f, -1.f),
    C_Vertex3f(0.238678455f, 0.51157999f, -1.f),
    C_Vertex3f(0.246071815f, 0.619312286f, -1.f),
    C_Vertex3f(0.248166084f, 0.519637108f, -1.f),
    C_Vertex3f(0.25682497f, 0.606219292f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.202651024f, 0.453656197f, -1.f),
    C_Vertex3f(0.207206726f, 0.497314453f, -1.f),
    C_Vertex3f(0.207368612f, 0.641156197f, -1.f),
    C_Vertex3f(0.21453476f, 0.499383926f, -1.f),
    C_Vertex3f(0.217759252f, 0.637515545f, -1.f),
    C_Vertex3f(0.227468014f, 0.50482893f, -1.f),
    C_Vertex3f(0.227075934f, 0.633146763f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.189328194f, 0.58597374f, -1.f),
    C_Vertex3f(0.194273472f, 0.579726219f, -1.f),
    C_Vertex3f(0.195904016f, 0.644068718f, -1.f),
    C_Vertex3f(0.197570324f, 0.572880745f, -1.f),
    C_Vertex3f(0.207368612f, 0.641156197f, -1.f),
    C_Vertex3f(0.19921875f, 0.565437317f, -1.f),
    C_Vertex3f(0.202651024f, 0.453656197f, -1.f),
    C_Vertex3f(0.19921875f, 0.556745529f, -1.f),
    C_Vertex3f(0.197174549f, 0.548109055f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.183365464f, 0.646253109f, -1.f),
    C_Vertex3f(0.169752955f, 0.64770937f, -1.f),
    C_Vertex3f(0.174546719f, 0.595860481f, -1.f),
    C_Vertex3f(0.182734489f, 0.591623306f, -1.f),
    C_Vertex3f(0.189328194f, 0.58597374f, -1.f),
    C_Vertex3f(0.195904016f, 0.644068718f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.140748978f, 0.600097656f, -1.f),
    C_Vertex3f(0.153388977f, 0.600097656f, -1.f),
    C_Vertex3f(0.15506649f, 0.6484375f, -1.f),
    C_Vertex3f(0.164764881f, 0.598685265f, -1.f),
    C_Vertex3f(0.169752955f, 0.64770937f, -1.f),
    C_Vertex3f(0.174546719f, 0.595860481f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052277f, 0.6484375f, -1.f),
    C_Vertex3f(0.11093092f, 0.646143913f, -1.f),
    C_Vertex3f(0.111895561f, 0.592830658f, -1.f),
    C_Vertex3f(0.120167255f, 0.596464157f, -1.f),
    C_Vertex3f(0.129785061f, 0.59888649f, -1.f),
    C_Vertex3f(0.140748978f, 0.600097656f, -1.f),
    C_Vertex3f(0.15506649f, 0.6484375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.11093092f, 0.646143913f, -1.f),
    C_Vertex3f(0.0909562111f, 0.64155674f, -1.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, -1.f),
    C_Vertex3f(0.0985121727f, 0.581159592f, -1.f),
    C_Vertex3f(0.104969978f, 0.587985992f, -1.f),
    C_Vertex3f(0.111895561f, 0.592830658f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0909562111f, 0.64155674f, -1.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, -1.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, -1.f),
    C_Vertex3f(0.0869998932f, 0.561561584f, -1.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0844726563f, 0.555175781f, -1.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, -1.f),
    C_Vertex3f(0.0604467392f, 0.625501633f, -1.f),
    C_Vertex3f(0.0479416847f, 0.612832069f, -1.f),
    C_Vertex3f(0.0366129875f, 0.59666729f, -1.f),
    C_Vertex3f(0.0264606476f, 0.577007294f, -1.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.212404728f, 0.437254429f, -1.f),
    C_Vertex3f(0.214355469f, 0.427412033f, -1.f),
    C_Vertex3f(0.216918945f, 0.494905472f, -1.f),
    C_Vertex3f(0.223830462f, 0.32666564f, -1.f),
    C_Vertex3f(0.233778477f, 0.488527298f, -1.f),
    C_Vertex3f(0.234298468f, 0.332331896f, -1.f),
    C_Vertex3f(0.248073101f, 0.480588913f, -1.f),
    C_Vertex3f(0.243850708f, 0.338942528f, -1.f),
    C_Vertex3f(0.252487183f, 0.346497536f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.208503246f, 0.446002483f, -1.f),
    C_Vertex3f(0.212404728f, 0.437254429f, -1.f),
    C_Vertex3f(0.216918945f, 0.494905472f, -1.f),
    C_Vertex3f(0.207206726f, 0.497314453f, -1.f),
    C_Vertex3f(0.202651024f, 0.453656197f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276237488f, 0.380000591f, -1.f),
    C_Vertex3f(0.279476166f, 0.389197111f, -1.f),
    C_Vertex3f(0.280423641f, 0.435041428f, -1.f),
    C_Vertex3f(0.281635284f, 0.398721933f, -1.f),
    C_Vertex3f(0.282714844f, 0.421110153f, -1.f),
    C_Vertex3f(0.282714844f, 0.408575058f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.260044098f, 0.354380846f, -1.f),
    C_Vertex3f(0.266521454f, 0.362592459f, -1.f),
    C_Vertex3f(0.268967628f, 0.460031509f, -1.f),
    C_Vertex3f(0.27191925f, 0.371132374f, -1.f),
    C_Vertex3f(0.275841236f, 0.448015213f, -1.f),
    C_Vertex3f(0.276237488f, 0.380000591f, -1.f),
    C_Vertex3f(0.280423641f, 0.435041428f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259802818f, 0.471090317f, -1.f),
    C_Vertex3f(0.248073101f, 0.480588913f, -1.f),
    C_Vertex3f(0.252487183f, 0.346497536f, -1.f),
    C_Vertex3f(0.260044098f, 0.354380846f, -1.f),
    C_Vertex3f(0.268967628f, 0.460031509f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214355469f, 0.415929794f, -1.f),
    C_Vertex3f(0.21200943f, 0.404776573f, -1.f),
    C_Vertex3f(0.21244669f, 0.32194376f, -1.f),
    C_Vertex3f(0.223830462f, 0.32666564f, -1.f),
    C_Vertex3f(0.214355469f, 0.427412033f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.21200943f, 0.404776573f, -1.f),
    C_Vertex3f(0.207317352f, 0.394497871f, -1.f),
    C_Vertex3f(0.21244669f, 0.32194376f, -1.f),
    C_Vertex3f(0.200279236f, 0.385093689f, -1.f),
    C_Vertex3f(0.200147152f, 0.318166256f, -1.f),
    C_Vertex3f(0.190895081f, 0.376564026f, -1.f),
    C_Vertex3f(0.186931849f, 0.315333128f, -1.f),
    C_Vertex3f(0.179983139f, 0.370166779f, -1.f),
    C_Vertex3f(0.172800779f, 0.313444376f, -1.f),
    C_Vertex3f(0.167543411f, 0.365901947f, -1.f),
    C_Vertex3f(0.157753944f, 0.3125f, -1.f),
    C_Vertex3f(0.153575897f, 0.363769531f, -1.f),
    C_Vertex3f(0.142120361f, 0.3125f, -1.f),
    C_Vertex3f(0.139390945f, 0.363769531f, -1.f),
    C_Vertex3f(0.127717614f, 0.31325984f, -1.f),
    C_Vertex3f(0.127388954f, 0.365404129f, -1.f),
    C_Vertex3f(0.116259575f, 0.368673325f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127717614f, 0.31325984f, -1.f),
    C_Vertex3f(0.116259575f, 0.368673325f, -1.f),
    C_Vertex3f(0.114216685f, 0.31477952f, -1.f),
    C_Vertex3f(0.106002808f, 0.373577118f, -1.f),
    C_Vertex3f(0.101617575f, 0.31705904f, -1.f),
    C_Vertex3f(0.0978183746f, 0.379243851f, -1.f),
    C_Vertex3f(0.0914888382f, 0.386110306f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0899202824f, 0.3200984f, -1.f),
    C_Vertex3f(0.101617575f, 0.31705904f, -1.f),
    C_Vertex3f(0.0914888382f, 0.386110306f, -1.f),
    C_Vertex3f(0.085814476f, 0.395048141f, -1.f),
    C_Vertex3f(0.0807952881f, 0.406057358f, -1.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0786132813f, 0.412597656f, -1.f),
    C_Vertex3f(0.0146484375f, 0.40625f, -1.f),
    C_Vertex3f(0.0169448853f, 0.39496994f, -1.f),
    C_Vertex3f(0.0244345665f, 0.37501049f, -1.f),
    C_Vertex3f(0.0348210335f, 0.357651711f, -1.f),
    C_Vertex3f(0.0481042862f, 0.3428936f, -1.f),
    C_Vertex3f(0.060239315f, 0.33377552f, -1.f),
    C_Vertex3f(0.0692311525f, 0.32845664f, -1.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, -1.f),
    C_Vertex3f(0.0807952881f, 0.406057358f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.194848061f, 0.460215569f, -1.f),
    C_Vertex3f(0.202651024f, 0.453656197f, -1.f),
    C_Vertex3f(0.197174549f, 0.548109055f, -1.f),
    C_Vertex3f(0.193086147f, 0.540178299f, -1.f),
    C_Vertex3f(0.186953545f, 0.532953262f, -1.f),
    C_Vertex3f(0.184280872f, 0.465135098f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.178776741f, 0.526433945f, -1.f),
    C_Vertex3f(0.168227673f, 0.521598816f, -1.f),
    C_Vertex3f(0.170949459f, 0.468414783f, -1.f),
    C_Vertex3f(0.184280872f, 0.465135098f, -1.f),
    C_Vertex3f(0.186953545f, 0.532953262f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155306339f, 0.518447876f, -1.f),
    C_Vertex3f(0.140012741f, 0.516981125f, -1.f),
    C_Vertex3f(0.142822266f, 0.470054626f, -1.f),
    C_Vertex3f(0.154853821f, 0.470054626f, -1.f),
    C_Vertex3f(0.170949459f, 0.468414783f, -1.f),
    C_Vertex3f(0.168227673f, 0.521598816f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.12808609f, 0.468967438f, -1.f),
    C_Vertex3f(0.131118774f, 0.469184875f, -1.f),
    C_Vertex3f(0.13117981f, 0.517089844f, -1.f),
    C_Vertex3f(0.135669708f, 0.469619751f, -1.f),
    C_Vertex3f(0.140012741f, 0.516981125f, -1.f),
    C_Vertex3f(0.142822266f, 0.470054626f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.13117981f, 0.517089844f, -1.f),
    C_Vertex3f(0.121582031f, 0.46875f, -1.f),
    C_Vertex3f(0.123317719f, 0.46875f, -1.f),
    C_Vertex3f(0.12808609f, 0.468967438f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_180[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.0971679688f, 0.521484375f, 0.f),
    C_Vertex3f(0.0971679688f, 0.521484375f, -1.f),
    C_Vertex3f(0.154785156f, 0.644042969f, -1.f),
    C_Vertex3f(0.154785156f, 0.644042969f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.154785156f, 0.644042969f, 0.f),
    C_Vertex3f(0.154785156f, 0.644042969f, -1.f),
    C_Vertex3f(0.258300781f, 0.644042969f, -1.f),
    C_Vertex3f(0.258300781f, 0.644042969f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.258300781f, 0.644042969f, 0.f),
    C_Vertex3f(0.258300781f, 0.644042969f, -1.f),
    C_Vertex3f(0.162109375f, 0.521484375f, -1.f),
    C_Vertex3f(0.162109375f, 0.521484375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.162109375f, 0.521484375f, 0.f),
    C_Vertex3f(0.162109375f, 0.521484375f, -1.f),
    C_Vertex3f(0.0971679688f, 0.521484375f, -1.f),
    C_Vertex3f(0.0971679688f, 0.521484375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0971679688f, 0.521484375f, 0.f),
    C_Vertex3f(0.162109375f, 0.521484375f, 0.f),
    C_Vertex3f(0.258300781f, 0.644042969f, 0.f),
    C_Vertex3f(0.154785156f, 0.644042969f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0971679688f, 0.521484375f, -1.f),
    C_Vertex3f(0.162109375f, 0.521484375f, -1.f),
    C_Vertex3f(0.258300781f, 0.644042969f, -1.f),
    C_Vertex3f(0.154785156f, 0.644042969f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_181[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.444824219f, 0.464355469f, 0.f),
    C_Vertex3f(0.444824219f, 0.464355469f, -1.f),
    C_Vertex3f(0.444824219f, 0.f, -1.f),
    C_Vertex3f(0.444824219f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.444824219f, 0.f, 0.f),
    C_Vertex3f(0.444824219f, 0.f, -1.f),
    C_Vertex3f(0.374511719f, 0.f, -1.f),
    C_Vertex3f(0.374511719f, 0.f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.374511719f, 0.f, 0.f),
    C_Vertex3f(0.374511719f, 0.f, -1.f),
    C_Vertex3f(0.374511719f, 0.0552825928f, -1.f),
    C_Vertex3f(0.374511719f, 0.0552825928f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.838297367f, -0.545213282f, 0.f),
    C_Vertex3f(0.374511719f, 0.0552825928f, 0.f),
    C_Vertex3f(0.374511719f, 0.0552825928f, -1.f),
    C_Normal3f(0.81012404f, -0.586258411f, 0.f),
    C_Vertex3f(0.368824005f, 0.0465373993f, 0.f),
    C_Vertex3f(0.368824005f, 0.0465373993f, -1.f),
    C_Normal3f(0.764454961f, -0.644677103f, 0.f),
    C_Vertex3f(0.357503891f, 0.0317802429f, 0.f),
    C_Vertex3f(0.357503891f, 0.0317802429f, -1.f),
    C_Normal3f(0.688139141f, -0.725578785f, 0.f),
    C_Vertex3f(0.34623909f, 0.0197563171f, 0.f),
    C_Vertex3f(0.34623909f, 0.0197563171f, -1.f),
    C_Normal3f(0.572151542f, -0.820147872f, 0.f),
    C_Vertex3f(0.335029602f, 0.0104656219f, 0.f),
    C_Vertex3f(0.335029602f, 0.0104656219f, -1.f),
    C_Normal3f(0.437112212f, -0.899407029f, 0.f),
    C_Vertex3f(0.321796417f, 0.00270462036f, 0.f),
    C_Vertex3f(0.321796417f, 0.00270462036f, -1.f),
    C_Normal3f(0.309743494f, -0.950820208f, 0.f),
    C_Vertex3f(0.305226326f, -0.00401878357f, 0.f),
    C_Vertex3f(0.305226326f, -0.00401878357f, -1.f),
    C_Normal3f(0.179189622f, -0.983814597f, 0.f),
    C_Vertex3f(0.287398338f, -0.00850105286f, 0.f),
    C_Vertex3f(0.287398338f, -0.00850105286f, -1.f),
    C_Normal3f(0.0583624169f, -0.998295367f, 0.f),
    C_Vertex3f(0.268312454f, -0.0107421875f, 0.f),
    C_Vertex3f(0.268312454f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0495052002f, -0.998773932f, 0.f),
    C_Vertex3f(0.249063492f, -0.0107421875f, 0.f),
    C_Vertex3f(0.249063492f, -0.0107421875f, -1.f),
    C_Normal3f(-0.160788015f, -0.986988902f, 0.f),
    C_Vertex3f(0.231947899f, -0.00893974304f, 0.f),
    C_Vertex3f(0.231947899f, -0.00893974304f, -1.f),
    C_Normal3f(-0.27820316f, -0.960522175f, 0.f),
    C_Vertex3f(0.215870857f, -0.00533485413f, 0.f),
    C_Vertex3f(0.215870857f, -0.00533485413f, -1.f),
    C_Normal3f(-0.395309776f, -0.918547869f, 0.f),
    C_Vertex3f(0.200832367f, 7.2479248e-005f, 0.f),
    C_Vertex3f(0.200832367f, 7.2479248e-005f, -1.f),
    C_Normal3f(-0.527815521f, -0.849359155f, 0.f),
    C_Vertex3f(0.188583374f, 0.00640869141f, 0.f),
    C_Vertex3f(0.188583374f, 0.00640869141f, -1.f),
    C_Normal3f(-0.641999781f, -0.766704798f, 0.f),
    C_Vertex3f(0.177537918f, 0.0145483017f, 0.f),
    C_Vertex3f(0.177537918f, 0.0145483017f, -1.f),
    C_Normal3f(-0.715547144f, -0.69856441f, 0.f),
    C_Vertex3f(0.165945053f, 0.0253648758f, 0.f),
    C_Vertex3f(0.165945053f, 0.0253648758f, -1.f),
    C_Normal3f(-0.759465039f, -0.6505481f, 0.f),
    C_Vertex3f(0.153804779f, 0.0388584137f, 0.f),
    C_Vertex3f(0.153804779f, 0.0388584137f, -1.f),
    C_Normal3f(-0.786736727f, -0.617288589f, 0.f),
    C_Vertex3f(0.147460938f, 0.0469436646f, 0.f),
    C_Vertex3f(0.147460938f, 0.0469436646f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.147460938f, 0.0469436646f, 0.f),
    C_Vertex3f(0.147460938f, 0.0469436646f, -1.f),
    C_Vertex3f(0.147460938f, -0.177734375f, -1.f),
    C_Vertex3f(0.147460938f, -0.177734375f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.147460938f, -0.177734375f, 0.f),
    C_Vertex3f(0.147460938f, -0.177734375f, -1.f),
    C_Vertex3f(0.0698242188f, -0.177734375f, -1.f),
    C_Vertex3f(0.0698242188f, -0.177734375f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0698242188f, -0.177734375f, 0.f),
    C_Vertex3f(0.0698242188f, -0.177734375f, -1.f),
    C_Vertex3f(0.0698242188f, 0.464355469f, -1.f),
    C_Vertex3f(0.0698242188f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0698242188f, 0.464355469f, 0.f),
    C_Vertex3f(0.0698242188f, 0.464355469f, -1.f),
    C_Vertex3f(0.147460938f, 0.464355469f, -1.f),
    C_Vertex3f(0.147460938f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.147460938f, 0.464355469f, 0.f),
    C_Vertex3f(0.147460938f, 0.464355469f, -1.f),
    C_Normal3f(0.999998868f, 0.00151724683f, 0.f),
    C_Vertex3f(0.147460938f, 0.25431633f, 0.f),
    C_Vertex3f(0.147460938f, 0.25431633f, -1.f),
    C_Normal3f(0.999736369f, 0.0229594503f, 0.f),
    C_Vertex3f(0.147816658f, 0.229904413f, 0.f),
    C_Vertex3f(0.147816658f, 0.229904413f, -1.f),
    C_Normal3f(0.999094009f, 0.042556148f, 0.f),
    C_Vertex3f(0.148528099f, 0.20784831f, 0.f),
    C_Vertex3f(0.148528099f, 0.20784831f, -1.f),
    C_Normal3f(0.997748554f, 0.0670658201f, 0.f),
    C_Vertex3f(0.149595261f, 0.188148022f, 0.f),
    C_Vertex3f(0.149595261f, 0.188148022f, -1.f),
    C_Normal3f(0.995133698f, 0.098533757f, 0.f),
    C_Vertex3f(0.151018143f, 0.170803547f, 0.f),
    C_Vertex3f(0.151018143f, 0.170803547f, -1.f),
    C_Normal3f(0.99011445f, 0.140261874f, 0.f),
    C_Vertex3f(0.152796745f, 0.155814886f, 0.f),
    C_Vertex3f(0.152796745f, 0.155814886f, -1.f),
    C_Normal3f(0.980230331f, 0.197859809f, 0.f),
    C_Vertex3f(0.154931068f, 0.143182039f, 0.f),
    C_Vertex3f(0.154931068f, 0.143182039f, -1.f),
    C_Normal3f(0.956497073f, 0.291741848f, 0.f),
    C_Vertex3f(0.157421112f, 0.132905006f, 0.f),
    C_Vertex3f(0.157421112f, 0.132905006f, -1.f),
    C_Normal3f(0.911756694f, 0.410730958f, 0.f),
    C_Vertex3f(0.161689758f, 0.121023178f, 0.f),
    C_Vertex3f(0.161689758f, 0.121023178f, -1.f),
    C_Normal3f(0.847431242f, 0.530905068f, 0.f),
    C_Vertex3f(0.169159889f, 0.106846809f, 0.f),
    C_Vertex3f(0.169159889f, 0.106846809f, -1.f),
    C_Normal3f(0.755121052f, 0.655585289f, 0.f),
    C_Vertex3f(0.178408623f, 0.0943365097f, 0.f),
    C_Vertex3f(0.178408623f, 0.0943365097f, -1.f),
    C_Normal3f(0.643238068f, 0.765666246f, 0.f),
    C_Vertex3f(0.189435959f, 0.0834922791f, 0.f),
    C_Vertex3f(0.189435959f, 0.0834922791f, -1.f),
    C_Normal3f(0.517338812f, 0.855780721f, 0.f),
    C_Vertex3f(0.202241898f, 0.0743141174f, 0.f),
    C_Vertex3f(0.202241898f, 0.0743141174f, -1.f),
    C_Normal3f(0.373606384f, 0.92758733f, 0.f),
    C_Vertex3f(0.216005325f, 0.0674304962f, 0.f),
    C_Vertex3f(0.216005325f, 0.0674304962f, -1.f),
    C_Normal3f(0.220848724f, 0.975308061f, 0.f),
    C_Vertex3f(0.230726242f, 0.0628414154f, 0.f),
    C_Vertex3f(0.230726242f, 0.0628414154f, -1.f),
    C_Normal3f(0.0698846206f, 0.997555077f, 0.f),
    C_Vertex3f(0.246404648f, 0.060546875f, 0.f),
    C_Vertex3f(0.246404648f, 0.060546875f, -1.f),
    C_Normal3f(-0.0712370053f, 0.997459412f, 0.f),
    C_Vertex3f(0.263479233f, 0.060546875f, 0.f),
    C_Vertex3f(0.263479233f, 0.060546875f, -1.f),
    C_Normal3f(-0.218316615f, 0.97587806f, 0.f),
    C_Vertex3f(0.280061722f, 0.0629506111f, 0.f),
    C_Vertex3f(0.280061722f, 0.0629506111f, -1.f),
    C_Normal3f(-0.367947489f, 0.929846585f, 0.f),
    C_Vertex3f(0.295713425f, 0.0677580833f, 0.f),
    C_Vertex3f(0.295713425f, 0.0677580833f, -1.f),
    C_Normal3f(-0.50825268f, 0.861207962f, 0.f),
    C_Vertex3f(0.310434341f, 0.0749692917f, 0.f),
    C_Vertex3f(0.310434341f, 0.0749692917f, -1.f),
    C_Normal3f(-0.627999067f, 0.778214097f, 0.f),
    C_Vertex3f(0.324224472f, 0.0845842361f, 0.f),
    C_Vertex3f(0.324224472f, 0.0845842361f, -1.f),
    C_Normal3f(-0.7363922f, 0.676554978f, 0.f),
    C_Vertex3f(0.335989952f, 0.0955920219f, 0.f),
    C_Vertex3f(0.335989952f, 0.0955920219f, -1.f),
    C_Normal3f(-0.832134187f, 0.554574311f, 0.f),
    C_Vertex3f(0.345730782f, 0.107992649f, 0.f),
    C_Vertex3f(0.345730782f, 0.107992649f, -1.f),
    C_Normal3f(-0.90296036f, 0.429723859f, 0.f),
    C_Vertex3f(0.35344696f, 0.121786118f, 0.f),
    C_Vertex3f(0.35344696f, 0.121786118f, -1.f),
    C_Normal3f(-0.953435183f, 0.301597714f, 0.f),
    C_Vertex3f(0.357715607f, 0.13317585f, 0.f),
    C_Vertex3f(0.357715607f, 0.13317585f, -1.f),
    C_Normal3f(-0.979233265f, 0.202736944f, 0.f),
    C_Vertex3f(0.36020565f, 0.143152237f, 0.f),
    C_Vertex3f(0.36020565f, 0.143152237f, -1.f),
    C_Normal3f(-0.989738107f, 0.142893672f, 0.f),
    C_Vertex3f(0.362339973f, 0.155511856f, 0.f),
    C_Vertex3f(0.362339973f, 0.155511856f, -1.f),
    C_Normal3f(-0.994992018f, 0.0999547616f, 0.f),
    C_Vertex3f(0.364118576f, 0.170254707f, 0.f),
    C_Vertex3f(0.364118576f, 0.170254707f, -1.f),
    C_Normal3f(-0.997698188f, 0.0678117871f, 0.f),
    C_Vertex3f(0.365541458f, 0.187380791f, 0.f),
    C_Vertex3f(0.365541458f, 0.187380791f, -1.f),
    C_Normal3f(-0.999078512f, 0.0429198928f, 0.f),
    C_Vertex3f(0.36660862f, 0.206890106f, 0.f),
    C_Vertex3f(0.36660862f, 0.206890106f, -1.f),
    C_Normal3f(-0.999732971f, 0.0231084097f, 0.f),
    C_Vertex3f(0.367320061f, 0.228782654f, 0.f),
    C_Vertex3f(0.367320061f, 0.228782654f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.367675781f, 0.253058434f, 0.f),
    C_Vertex3f(0.367675781f, 0.253058434f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.367675781f, 0.253058434f, 0.f),
    C_Vertex3f(0.367675781f, 0.253058434f, -1.f),
    C_Vertex3f(0.367675781f, 0.464355469f, -1.f),
    C_Vertex3f(0.367675781f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.367675781f, 0.464355469f, 0.f),
    C_Vertex3f(0.367675781f, 0.464355469f, -1.f),
    C_Vertex3f(0.444824219f, 0.464355469f, -1.f),
    C_Vertex3f(0.444824219f, 0.464355469f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147816658f, 0.229904413f, 0.f),
    C_Vertex3f(0.147460938f, 0.25431633f, 0.f),
    C_Vertex3f(0.147460938f, 0.0469436646f, 0.f),
    C_Vertex3f(0.147460938f, 0.464355469f, 0.f),
    C_Vertex3f(0.147460938f, -0.177734375f, 0.f),
    C_Vertex3f(0.0698242188f, 0.464355469f, 0.f),
    C_Vertex3f(0.0698242188f, -0.177734375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.374511719f, 0.f, 0.f),
    C_Vertex3f(0.444824219f, 0.f, 0.f),
    C_Vertex3f(0.374511719f, 0.0552825928f, 0.f),
    C_Vertex3f(0.444824219f, 0.464355469f, 0.f),
    C_Vertex3f(0.368824005f, 0.0465373993f, 0.f),
    C_Vertex3f(0.367675781f, 0.464355469f, 0.f),
    C_Vertex3f(0.367675781f, 0.253058434f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.368824005f, 0.0465373993f, 0.f),
    C_Vertex3f(0.367675781f, 0.253058434f, 0.f),
    C_Vertex3f(0.367320061f, 0.228782654f, 0.f),
    C_Vertex3f(0.36660862f, 0.206890106f, 0.f),
    C_Vertex3f(0.365541458f, 0.187380791f, 0.f),
    C_Vertex3f(0.364118576f, 0.170254707f, 0.f),
    C_Vertex3f(0.362339973f, 0.155511856f, 0.f),
    C_Vertex3f(0.36020565f, 0.143152237f, 0.f),
    C_Vertex3f(0.357715607f, 0.13317585f, 0.f),
    C_Vertex3f(0.357503891f, 0.0317802429f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35344696f, 0.121786118f, 0.f),
    C_Vertex3f(0.345730782f, 0.107992649f, 0.f),
    C_Vertex3f(0.34623909f, 0.0197563171f, 0.f),
    C_Vertex3f(0.357503891f, 0.0317802429f, 0.f),
    C_Vertex3f(0.357715607f, 0.13317585f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230726242f, 0.0628414154f, 0.f),
    C_Vertex3f(0.231947899f, -0.00893974304f, 0.f),
    C_Vertex3f(0.246404648f, 0.060546875f, 0.f),
    C_Vertex3f(0.249063492f, -0.0107421875f, 0.f),
    C_Vertex3f(0.263479233f, 0.060546875f, 0.f),
    C_Vertex3f(0.268312454f, -0.0107421875f, 0.f),
    C_Vertex3f(0.280061722f, 0.0629506111f, 0.f),
    C_Vertex3f(0.287398338f, -0.00850105286f, 0.f),
    C_Vertex3f(0.295713425f, 0.0677580833f, 0.f),
    C_Vertex3f(0.305226326f, -0.00401878357f, 0.f),
    C_Vertex3f(0.310434341f, 0.0749692917f, 0.f),
    C_Vertex3f(0.321796417f, 0.00270462036f, 0.f),
    C_Vertex3f(0.324224472f, 0.0845842361f, 0.f),
    C_Vertex3f(0.335029602f, 0.0104656219f, 0.f),
    C_Vertex3f(0.335989952f, 0.0955920219f, 0.f),
    C_Vertex3f(0.34623909f, 0.0197563171f, 0.f),
    C_Vertex3f(0.345730782f, 0.107992649f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.161689758f, 0.121023178f, 0.f),
    C_Vertex3f(0.165945053f, 0.0253648758f, 0.f),
    C_Vertex3f(0.169159889f, 0.106846809f, 0.f),
    C_Vertex3f(0.177537918f, 0.0145483017f, 0.f),
    C_Vertex3f(0.178408623f, 0.0943365097f, 0.f),
    C_Vertex3f(0.188583374f, 0.00640869141f, 0.f),
    C_Vertex3f(0.189435959f, 0.0834922791f, 0.f),
    C_Vertex3f(0.200832367f, 7.2479248e-005f, 0.f),
    C_Vertex3f(0.202241898f, 0.0743141174f, 0.f),
    C_Vertex3f(0.215870857f, -0.00533485413f, 0.f),
    C_Vertex3f(0.216005325f, 0.0674304962f, 0.f),
    C_Vertex3f(0.231947899f, -0.00893974304f, 0.f),
    C_Vertex3f(0.230726242f, 0.0628414154f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.161689758f, 0.121023178f, 0.f),
    C_Vertex3f(0.157421112f, 0.132905006f, 0.f),
    C_Vertex3f(0.165945053f, 0.0253648758f, 0.f),
    C_Vertex3f(0.154931068f, 0.143182039f, 0.f),
    C_Vertex3f(0.153804779f, 0.0388584137f, 0.f),
    C_Vertex3f(0.152796745f, 0.155814886f, 0.f),
    C_Vertex3f(0.151018143f, 0.170803547f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.153804779f, 0.0388584137f, 0.f),
    C_Vertex3f(0.151018143f, 0.170803547f, 0.f),
    C_Vertex3f(0.149595261f, 0.188148022f, 0.f),
    C_Vertex3f(0.148528099f, 0.20784831f, 0.f),
    C_Vertex3f(0.147816658f, 0.229904413f, 0.f),
    C_Vertex3f(0.147460938f, 0.0469436646f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147816658f, 0.229904413f, -1.f),
    C_Vertex3f(0.147460938f, 0.25431633f, -1.f),
    C_Vertex3f(0.147460938f, 0.0469436646f, -1.f),
    C_Vertex3f(0.147460938f, 0.464355469f, -1.f),
    C_Vertex3f(0.147460938f, -0.177734375f, -1.f),
    C_Vertex3f(0.0698242188f, 0.464355469f, -1.f),
    C_Vertex3f(0.0698242188f, -0.177734375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.374511719f, 0.f, -1.f),
    C_Vertex3f(0.444824219f, 0.f, -1.f),
    C_Vertex3f(0.374511719f, 0.0552825928f, -1.f),
    C_Vertex3f(0.444824219f, 0.464355469f, -1.f),
    C_Vertex3f(0.368824005f, 0.0465373993f, -1.f),
    C_Vertex3f(0.367675781f, 0.464355469f, -1.f),
    C_Vertex3f(0.367675781f, 0.253058434f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.368824005f, 0.0465373993f, -1.f),
    C_Vertex3f(0.367675781f, 0.253058434f, -1.f),
    C_Vertex3f(0.367320061f, 0.228782654f, -1.f),
    C_Vertex3f(0.36660862f, 0.206890106f, -1.f),
    C_Vertex3f(0.365541458f, 0.187380791f, -1.f),
    C_Vertex3f(0.364118576f, 0.170254707f, -1.f),
    C_Vertex3f(0.362339973f, 0.155511856f, -1.f),
    C_Vertex3f(0.36020565f, 0.143152237f, -1.f),
    C_Vertex3f(0.357715607f, 0.13317585f, -1.f),
    C_Vertex3f(0.357503891f, 0.0317802429f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35344696f, 0.121786118f, -1.f),
    C_Vertex3f(0.345730782f, 0.107992649f, -1.f),
    C_Vertex3f(0.34623909f, 0.0197563171f, -1.f),
    C_Vertex3f(0.357503891f, 0.0317802429f, -1.f),
    C_Vertex3f(0.357715607f, 0.13317585f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230726242f, 0.0628414154f, -1.f),
    C_Vertex3f(0.231947899f, -0.00893974304f, -1.f),
    C_Vertex3f(0.246404648f, 0.060546875f, -1.f),
    C_Vertex3f(0.249063492f, -0.0107421875f, -1.f),
    C_Vertex3f(0.263479233f, 0.060546875f, -1.f),
    C_Vertex3f(0.268312454f, -0.0107421875f, -1.f),
    C_Vertex3f(0.280061722f, 0.0629506111f, -1.f),
    C_Vertex3f(0.287398338f, -0.00850105286f, -1.f),
    C_Vertex3f(0.295713425f, 0.0677580833f, -1.f),
    C_Vertex3f(0.305226326f, -0.00401878357f, -1.f),
    C_Vertex3f(0.310434341f, 0.0749692917f, -1.f),
    C_Vertex3f(0.321796417f, 0.00270462036f, -1.f),
    C_Vertex3f(0.324224472f, 0.0845842361f, -1.f),
    C_Vertex3f(0.335029602f, 0.0104656219f, -1.f),
    C_Vertex3f(0.335989952f, 0.0955920219f, -1.f),
    C_Vertex3f(0.34623909f, 0.0197563171f, -1.f),
    C_Vertex3f(0.345730782f, 0.107992649f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.161689758f, 0.121023178f, -1.f),
    C_Vertex3f(0.165945053f, 0.0253648758f, -1.f),
    C_Vertex3f(0.169159889f, 0.106846809f, -1.f),
    C_Vertex3f(0.177537918f, 0.0145483017f, -1.f),
    C_Vertex3f(0.178408623f, 0.0943365097f, -1.f),
    C_Vertex3f(0.188583374f, 0.00640869141f, -1.f),
    C_Vertex3f(0.189435959f, 0.0834922791f, -1.f),
    C_Vertex3f(0.200832367f, 7.2479248e-005f, -1.f),
    C_Vertex3f(0.202241898f, 0.0743141174f, -1.f),
    C_Vertex3f(0.215870857f, -0.00533485413f, -1.f),
    C_Vertex3f(0.216005325f, 0.0674304962f, -1.f),
    C_Vertex3f(0.231947899f, -0.00893974304f, -1.f),
    C_Vertex3f(0.230726242f, 0.0628414154f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.161689758f, 0.121023178f, -1.f),
    C_Vertex3f(0.157421112f, 0.132905006f, -1.f),
    C_Vertex3f(0.165945053f, 0.0253648758f, -1.f),
    C_Vertex3f(0.154931068f, 0.143182039f, -1.f),
    C_Vertex3f(0.153804779f, 0.0388584137f, -1.f),
    C_Vertex3f(0.152796745f, 0.155814886f, -1.f),
    C_Vertex3f(0.151018143f, 0.170803547f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.153804779f, 0.0388584137f, -1.f),
    C_Vertex3f(0.151018143f, 0.170803547f, -1.f),
    C_Vertex3f(0.149595261f, 0.188148022f, -1.f),
    C_Vertex3f(0.148528099f, 0.20784831f, -1.f),
    C_Vertex3f(0.147816658f, 0.229904413f, -1.f),
    C_Vertex3f(0.147460938f, 0.0469436646f, -1.f),
    C_End(),
    C_Translatef(0.515625f, 0.f, 0.f),
};
static const Call char_182[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.176757813f, -0.177734375f, 0.f),
    C_Vertex3f(0.176757813f, -0.177734375f, -1.f),
    C_Vertex3f(0.176757813f, 0.278808594f, -1.f),
    C_Vertex3f(0.176757813f, 0.278808594f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0533554479f, -0.998575628f, 0.f),
    C_Vertex3f(0.176757813f, 0.278808594f, 0.f),
    C_Vertex3f(0.176757813f, 0.278808594f, -1.f),
    C_Normal3f(-0.101648144f, -0.994820416f, 0.f),
    C_Vertex3f(0.166530609f, 0.279355049f, 0.f),
    C_Vertex3f(0.166530609f, 0.279355049f, -1.f),
    C_Normal3f(-0.16598092f, -0.986128926f, 0.f),
    C_Vertex3f(0.147149563f, 0.281833887f, 0.f),
    C_Vertex3f(0.147149563f, 0.281833887f, -1.f),
    C_Normal3f(-0.248429596f, -0.968649983f, 0.f),
    C_Vertex3f(0.128841877f, 0.285698652f, 0.f),
    C_Vertex3f(0.128841877f, 0.285698652f, -1.f),
    C_Normal3f(-0.335346013f, -0.942095041f, 0.f),
    C_Vertex3f(0.111607552f, 0.290949345f, 0.f),
    C_Vertex3f(0.111607552f, 0.290949345f, -1.f),
    C_Normal3f(-0.424704641f, -0.905332029f, 0.f),
    C_Vertex3f(0.0954465866f, 0.297585964f, 0.f),
    C_Vertex3f(0.0954465866f, 0.297585964f, -1.f),
    C_Normal3f(-0.513843417f, -0.85788399f, 0.f),
    C_Vertex3f(0.0803589821f, 0.305608511f, 0.f),
    C_Vertex3f(0.0803589821f, 0.305608511f, -1.f),
    C_Normal3f(-0.599744022f, -0.800191879f, 0.f),
    C_Vertex3f(0.066344738f, 0.315016985f, 0.f),
    C_Vertex3f(0.066344738f, 0.315016985f, -1.f),
    C_Normal3f(-0.679471433f, -0.733701944f, 0.f),
    C_Vertex3f(0.0534038544f, 0.325811386f, 0.f),
    C_Vertex3f(0.0534038544f, 0.325811386f, -1.f),
    C_Normal3f(-0.749605f, -0.6618855f, 0.f),
    C_Vertex3f(0.0415363312f, 0.337991714f, 0.f),
    C_Vertex3f(0.0415363312f, 0.337991714f, -1.f),
    C_Normal3f(-0.81252563f, -0.582925618f, 0.f),
    C_Vertex3f(0.0311522484f, 0.351011992f, 0.f),
    C_Vertex3f(0.0311522484f, 0.351011992f, -1.f),
    C_Normal3f(-0.868274391f, -0.496084303f, 0.f),
    C_Vertex3f(0.022251606f, 0.364872217f, 0.f),
    C_Vertex3f(0.022251606f, 0.364872217f, -1.f),
    C_Normal3f(-0.914809823f, -0.403884679f, 0.f),
    C_Vertex3f(0.014834404f, 0.379572392f, 0.f),
    C_Vertex3f(0.014834404f, 0.379572392f, -1.f),
    C_Normal3f(-0.950946271f, -0.309356034f, 0.f),
    C_Vertex3f(0.0089006424f, 0.395112514f, 0.f),
    C_Vertex3f(0.0089006424f, 0.395112514f, -1.f),
    C_Normal3f(-0.976490617f, -0.215559781f, 0.f),
    C_Vertex3f(0.0044503212f, 0.411492586f, 0.f),
    C_Vertex3f(0.0044503212f, 0.411492586f, -1.f),
    C_Normal3f(-0.99213773f, -0.125151157f, 0.f),
    C_Vertex3f(0.0014834404f, 0.428712606f, 0.f),
    C_Vertex3f(0.0014834404f, 0.428712606f, -1.f),
    C_Normal3f(-0.999240994f, -0.0389551297f, 0.f),
    C_Vertex3f(0.f, 0.446772575f, 0.f),
    C_Vertex3f(0.f, 0.446772575f, -1.f),
    C_Normal3f(-0.999153078f, 0.0411480442f, 0.f),
    C_Vertex3f(0.f, 0.46676445f, 0.f),
    C_Vertex3f(0.f, 0.46676445f, -1.f),
    C_Normal3f(-0.992045939f, 0.125875965f, 0.f),
    C_Vertex3f(0.00164747238f, 0.486776352f, 0.f),
    C_Vertex3f(0.00164747238f, 0.486776352f, -1.f),
    C_Normal3f(-0.97586143f, 0.218390584f, 0.f),
    C_Vertex3f(0.00494241714f, 0.505716324f, 0.f),
    C_Vertex3f(0.00494241714f, 0.505716324f, -1.f),
    C_Normal3f(-0.948867202f, 0.315675616f, 0.f),
    C_Vertex3f(0.00988483429f, 0.523584366f, 0.f),
    C_Vertex3f(0.00988483429f, 0.523584366f, -1.f),
    C_Normal3f(-0.909888327f, 0.414853066f, 0.f),
    C_Vertex3f(0.0164747238f, 0.540380478f, 0.f),
    C_Vertex3f(0.0164747238f, 0.540380478f, -1.f),
    C_Normal3f(-0.858783185f, 0.512339175f, 0.f),
    C_Vertex3f(0.0247120857f, 0.55610466f, 0.f),
    C_Vertex3f(0.0247120857f, 0.55610466f, -1.f),
    C_Normal3f(-0.796700478f, 0.604374409f, 0.f),
    C_Vertex3f(0.03459692f, 0.570756912f, 0.f),
    C_Vertex3f(0.03459692f, 0.570756912f, -1.f),
    C_Normal3f(-0.726001859f, 0.687692761f, 0.f),
    C_Vertex3f(0.0461292267f, 0.584337234f, 0.f),
    C_Vertex3f(0.0461292267f, 0.584337234f, -1.f),
    C_Normal3f(-0.64368552f, 0.765290141f, 0.f),
    C_Vertex3f(0.0593090057f, 0.596845627f, 0.f),
    C_Vertex3f(0.0593090057f, 0.596845627f, -1.f),
    C_Normal3f(-0.549155295f, 0.83572042f, 0.f),
    C_Vertex3f(0.0740132332f, 0.60779047f, 0.f),
    C_Vertex3f(0.0740132332f, 0.60779047f, -1.f),
    C_Normal3f(-0.451580197f, 0.892230511f, 0.f),
    C_Vertex3f(0.090241909f, 0.617171764f, 0.f),
    C_Vertex3f(0.090241909f, 0.617171764f, -1.f),
    C_Normal3f(-0.355223864f, 0.934781253f, 0.f),
    C_Vertex3f(0.107995033f, 0.62498951f, 0.f),
    C_Vertex3f(0.107995033f, 0.62498951f, -1.f),
    C_Normal3f(-0.263432056f, 0.964677989f, 0.f),
    C_Vertex3f(0.127272606f, 0.631243706f, 0.f),
    C_Vertex3f(0.127272606f, 0.631243706f, -1.f),
    C_Normal3f(-0.178359836f, 0.983965337f, 0.f),
    C_Vertex3f(0.148074627f, 0.635934353f, 0.f),
    C_Vertex3f(0.148074627f, 0.635934353f, -1.f),
    C_Normal3f(-0.101058841f, 0.994880438f, 0.f),
    C_Vertex3f(0.170401096f, 0.639061451f, 0.f),
    C_Vertex3f(0.170401096f, 0.639061451f, -1.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.194252014f, 0.640625f, 0.f),
    C_Vertex3f(0.194252014f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.194252014f, 0.640625f, 0.f),
    C_Vertex3f(0.194252014f, 0.640625f, -1.f),
    C_Vertex3f(0.483886719f, 0.640625f, -1.f),
    C_Vertex3f(0.483886719f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.483886719f, 0.640625f, 0.f),
    C_Vertex3f(0.483886719f, 0.640625f, -1.f),
    C_Vertex3f(0.483886719f, 0.564941406f, -1.f),
    C_Vertex3f(0.483886719f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.483886719f, 0.564941406f, 0.f),
    C_Vertex3f(0.483886719f, 0.564941406f, -1.f),
    C_Vertex3f(0.420898438f, 0.564941406f, -1.f),
    C_Vertex3f(0.420898438f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.420898438f, 0.564941406f, 0.f),
    C_Vertex3f(0.420898438f, 0.564941406f, -1.f),
    C_Vertex3f(0.420898438f, -0.177734375f, -1.f),
    C_Vertex3f(0.420898438f, -0.177734375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.420898438f, -0.177734375f, 0.f),
    C_Vertex3f(0.420898438f, -0.177734375f, -1.f),
    C_Vertex3f(0.346679688f, -0.177734375f, -1.f),
    C_Vertex3f(0.346679688f, -0.177734375f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.346679688f, -0.177734375f, 0.f),
    C_Vertex3f(0.346679688f, -0.177734375f, -1.f),
    C_Vertex3f(0.346679688f, 0.564941406f, -1.f),
    C_Vertex3f(0.346679688f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.346679688f, 0.564941406f, 0.f),
    C_Vertex3f(0.346679688f, 0.564941406f, -1.f),
    C_Vertex3f(0.249023438f, 0.564941406f, -1.f),
    C_Vertex3f(0.249023438f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.249023438f, 0.564941406f, 0.f),
    C_Vertex3f(0.249023438f, 0.564941406f, -1.f),
    C_Vertex3f(0.249023438f, -0.177734375f, -1.f),
    C_Vertex3f(0.249023438f, -0.177734375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.249023438f, -0.177734375f, 0.f),
    C_Vertex3f(0.249023438f, -0.177734375f, -1.f),
    C_Vertex3f(0.176757813f, -0.177734375f, -1.f),
    C_Vertex3f(0.176757813f, -0.177734375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.128841877f, 0.285698652f, 0.f),
    C_Vertex3f(0.147149563f, 0.281833887f, 0.f),
    C_Vertex3f(0.148074627f, 0.635934353f, 0.f),
    C_Vertex3f(0.166530609f, 0.279355049f, 0.f),
    C_Vertex3f(0.170401096f, 0.639061451f, 0.f),
    C_Vertex3f(0.176757813f, 0.278808594f, 0.f),
    C_Vertex3f(0.194252014f, 0.640625f, 0.f),
    C_Vertex3f(0.249023438f, -0.177734375f, 0.f),
    C_Vertex3f(0.249023438f, 0.564941406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.f, 0.46676445f, 0.f),
    C_Vertex3f(0.0014834404f, 0.428712606f, 0.f),
    C_Vertex3f(0.00164747238f, 0.486776352f, 0.f),
    C_Vertex3f(0.0044503212f, 0.411492586f, 0.f),
    C_Vertex3f(0.00494241714f, 0.505716324f, 0.f),
    C_Vertex3f(0.0089006424f, 0.395112514f, 0.f),
    C_Vertex3f(0.00988483429f, 0.523584366f, 0.f),
    C_Vertex3f(0.014834404f, 0.379572392f, 0.f),
    C_Vertex3f(0.0164747238f, 0.540380478f, 0.f),
    C_Vertex3f(0.022251606f, 0.364872217f, 0.f),
    C_Vertex3f(0.0247120857f, 0.55610466f, 0.f),
    C_Vertex3f(0.0311522484f, 0.351011992f, 0.f),
    C_Vertex3f(0.03459692f, 0.570756912f, 0.f),
    C_Vertex3f(0.0415363312f, 0.337991714f, 0.f),
    C_Vertex3f(0.0461292267f, 0.584337234f, 0.f),
    C_Vertex3f(0.0534038544f, 0.325811386f, 0.f),
    C_Vertex3f(0.0593090057f, 0.596845627f, 0.f),
    C_Vertex3f(0.066344738f, 0.315016985f, 0.f),
    C_Vertex3f(0.0740132332f, 0.60779047f, 0.f),
    C_Vertex3f(0.0803589821f, 0.305608511f, 0.f),
    C_Vertex3f(0.090241909f, 0.617171764f, 0.f),
    C_Vertex3f(0.0954465866f, 0.297585964f, 0.f),
    C_Vertex3f(0.107995033f, 0.62498951f, 0.f),
    C_Vertex3f(0.111607552f, 0.290949345f, 0.f),
    C_Vertex3f(0.127272606f, 0.631243706f, 0.f),
    C_Vertex3f(0.128841877f, 0.285698652f, 0.f),
    C_Vertex3f(0.148074627f, 0.635934353f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.483886719f, 0.640625f, 0.f),
    C_Vertex3f(0.194252014f, 0.640625f, 0.f),
    C_Vertex3f(0.249023438f, 0.564941406f, 0.f),
    C_Vertex3f(0.346679688f, 0.564941406f, 0.f),
    C_Vertex3f(0.420898438f, 0.564941406f, 0.f),
    C_Vertex3f(0.483886719f, 0.564941406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.346679688f, 0.564941406f, 0.f),
    C_Vertex3f(0.346679688f, -0.177734375f, 0.f),
    C_Vertex3f(0.420898438f, -0.177734375f, 0.f),
    C_Vertex3f(0.420898438f, 0.564941406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.249023438f, -0.177734375f, 0.f),
    C_Vertex3f(0.176757813f, 0.278808594f, 0.f),
    C_Vertex3f(0.176757813f, -0.177734375f, 0.f),
    C_Vertex3f(0.0014834404f, 0.428712606f, 0.f),
    C_Vertex3f(0.f, 0.46676445f, 0.f),
    C_Vertex3f(0.f, 0.446772575f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.128841877f, 0.285698652f, -1.f),
    C_Vertex3f(0.147149563f, 0.281833887f, -1.f),
    C_Vertex3f(0.148074627f, 0.635934353f, -1.f),
    C_Vertex3f(0.166530609f, 0.279355049f, -1.f),
    C_Vertex3f(0.170401096f, 0.639061451f, -1.f),
    C_Vertex3f(0.176757813f, 0.278808594f, -1.f),
    C_Vertex3f(0.194252014f, 0.640625f, -1.f),
    C_Vertex3f(0.249023438f, -0.177734375f, -1.f),
    C_Vertex3f(0.249023438f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.f, 0.46676445f, -1.f),
    C_Vertex3f(0.0014834404f, 0.428712606f, -1.f),
    C_Vertex3f(0.00164747238f, 0.486776352f, -1.f),
    C_Vertex3f(0.0044503212f, 0.411492586f, -1.f),
    C_Vertex3f(0.00494241714f, 0.505716324f, -1.f),
    C_Vertex3f(0.0089006424f, 0.395112514f, -1.f),
    C_Vertex3f(0.00988483429f, 0.523584366f, -1.f),
    C_Vertex3f(0.014834404f, 0.379572392f, -1.f),
    C_Vertex3f(0.0164747238f, 0.540380478f, -1.f),
    C_Vertex3f(0.022251606f, 0.364872217f, -1.f),
    C_Vertex3f(0.0247120857f, 0.55610466f, -1.f),
    C_Vertex3f(0.0311522484f, 0.351011992f, -1.f),
    C_Vertex3f(0.03459692f, 0.570756912f, -1.f),
    C_Vertex3f(0.0415363312f, 0.337991714f, -1.f),
    C_Vertex3f(0.0461292267f, 0.584337234f, -1.f),
    C_Vertex3f(0.0534038544f, 0.325811386f, -1.f),
    C_Vertex3f(0.0593090057f, 0.596845627f, -1.f),
    C_Vertex3f(0.066344738f, 0.315016985f, -1.f),
    C_Vertex3f(0.0740132332f, 0.60779047f, -1.f),
    C_Vertex3f(0.0803589821f, 0.305608511f, -1.f),
    C_Vertex3f(0.090241909f, 0.617171764f, -1.f),
    C_Vertex3f(0.0954465866f, 0.297585964f, -1.f),
    C_Vertex3f(0.107995033f, 0.62498951f, -1.f),
    C_Vertex3f(0.111607552f, 0.290949345f, -1.f),
    C_Vertex3f(0.127272606f, 0.631243706f, -1.f),
    C_Vertex3f(0.128841877f, 0.285698652f, -1.f),
    C_Vertex3f(0.148074627f, 0.635934353f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.483886719f, 0.640625f, -1.f),
    C_Vertex3f(0.194252014f, 0.640625f, -1.f),
    C_Vertex3f(0.249023438f, 0.564941406f, -1.f),
    C_Vertex3f(0.346679688f, 0.564941406f, -1.f),
    C_Vertex3f(0.420898438f, 0.564941406f, -1.f),
    C_Vertex3f(0.483886719f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.346679688f, 0.564941406f, -1.f),
    C_Vertex3f(0.346679688f, -0.177734375f, -1.f),
    C_Vertex3f(0.420898438f, -0.177734375f, -1.f),
    C_Vertex3f(0.420898438f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.249023438f, -0.177734375f, -1.f),
    C_Vertex3f(0.176757813f, 0.278808594f, -1.f),
    C_Vertex3f(0.176757813f, -0.177734375f, -1.f),
    C_Vertex3f(0.0014834404f, 0.428712606f, -1.f),
    C_Vertex3f(0.f, 0.46676445f, -1.f),
    C_Vertex3f(0.f, 0.446772575f, -1.f),
    C_End(),
    C_Translatef(0.480957031f, 0.f, 0.f),
};
static const Call char_183[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.104003906f, 0.278320313f, 0.f),
    C_Vertex3f(0.104003906f, 0.278320313f, -1.f),
    C_Vertex3f(0.104003906f, 0.367675781f, -1.f),
    C_Vertex3f(0.104003906f, 0.367675781f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.104003906f, 0.367675781f, 0.f),
    C_Vertex3f(0.104003906f, 0.367675781f, -1.f),
    C_Vertex3f(0.193359375f, 0.367675781f, -1.f),
    C_Vertex3f(0.193359375f, 0.367675781f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.193359375f, 0.367675781f, 0.f),
    C_Vertex3f(0.193359375f, 0.367675781f, -1.f),
    C_Vertex3f(0.193359375f, 0.278320313f, -1.f),
    C_Vertex3f(0.193359375f, 0.278320313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.193359375f, 0.278320313f, 0.f),
    C_Vertex3f(0.193359375f, 0.278320313f, -1.f),
    C_Vertex3f(0.104003906f, 0.278320313f, -1.f),
    C_Vertex3f(0.104003906f, 0.278320313f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.104003906f, 0.278320313f, 0.f),
    C_Vertex3f(0.193359375f, 0.278320313f, 0.f),
    C_Vertex3f(0.193359375f, 0.367675781f, 0.f),
    C_Vertex3f(0.104003906f, 0.367675781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.104003906f, 0.278320313f, -1.f),
    C_Vertex3f(0.193359375f, 0.278320313f, -1.f),
    C_Vertex3f(0.193359375f, 0.367675781f, -1.f),
    C_Vertex3f(0.104003906f, 0.367675781f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_184[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.959265828f, 0.282505304f, 0.f),
    C_Vertex3f(0.0944671631f, -0.0668945313f, 0.f),
    C_Vertex3f(0.0944671631f, -0.0668945313f, -1.f),
    C_Vertex3f(0.1171875f, 0.0102539063f, -1.f),
    C_Vertex3f(0.1171875f, 0.0102539063f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.1171875f, 0.0102539063f, 0.f),
    C_Vertex3f(0.1171875f, 0.0102539063f, -1.f),
    C_Vertex3f(0.17578125f, 0.0102539063f, -1.f),
    C_Vertex3f(0.17578125f, 0.0102539063f, 0.f),
    C_Normal3f(0.955515504f, -0.294941097f, 0.f),
    C_Vertex3f(0.17578125f, 0.0102539063f, 0.f),
    C_Vertex3f(0.17578125f, 0.0102539063f, -1.f),
    C_Vertex3f(0.161361694f, -0.0364608765f, -1.f),
    C_Vertex3f(0.161361694f, -0.0364608765f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.116983637f, 0.993133843f, 0.f),
    C_Vertex3f(0.161361694f, -0.0364608765f, 0.f),
    C_Vertex3f(0.161361694f, -0.0364608765f, -1.f),
    C_Normal3f(0.194189295f, 0.980964124f, 0.f),
    C_Vertex3f(0.170623779f, -0.0375518799f, 0.f),
    C_Vertex3f(0.170623779f, -0.0375518799f, -1.f),
    C_Normal3f(0.305636048f, 0.952148438f, 0.f),
    C_Vertex3f(0.18685627f, -0.0415077209f, 0.f),
    C_Vertex3f(0.18685627f, -0.0415077209f, -1.f),
    C_Normal3f(0.45934099f, 0.888260067f, 0.f),
    C_Vertex3f(0.200797081f, -0.0472373962f, 0.f),
    C_Vertex3f(0.200797081f, -0.0472373962f, -1.f),
    C_Normal3f(0.624141216f, 0.781311631f, 0.f),
    C_Vertex3f(0.212446213f, -0.0547409058f, 0.f),
    C_Vertex3f(0.212446213f, -0.0547409058f, -1.f),
    C_Normal3f(0.758728862f, 0.651406646f, 0.f),
    C_Vertex3f(0.221803665f, -0.0640182495f, 0.f),
    C_Vertex3f(0.221803665f, -0.0640182495f, -1.f),
    C_Normal3f(0.864426374f, 0.502759397f, 0.f),
    C_Vertex3f(0.228821754f, -0.0738143921f, 0.f),
    C_Vertex3f(0.228821754f, -0.0738143921f, -1.f),
    C_Normal3f(0.949106157f, 0.314956367f, 0.f),
    C_Vertex3f(0.233500481f, -0.0841293335f, 0.f),
    C_Vertex3f(0.233500481f, -0.0841293335f, -1.f),
    C_Normal3f(0.993615568f, 0.112818882f, 0.f),
    C_Vertex3f(0.235839844f, -0.0949630737f, 0.f),
    C_Vertex3f(0.235839844f, -0.0949630737f, -1.f),
    C_Normal3f(0.998467624f, -0.0553379245f, 0.f),
    C_Vertex3f(0.235839844f, -0.104732513f, 0.f),
    C_Vertex3f(0.235839844f, -0.104732513f, -1.f),
    C_Normal3f(0.98275584f, -0.184907913f, 0.f),
    C_Vertex3f(0.234857917f, -0.112680078f, 0.f),
    C_Vertex3f(0.234857917f, -0.112680078f, -1.f),
    C_Normal3f(0.951465666f, -0.307754815f, 0.f),
    C_Vertex3f(0.232894063f, -0.120388865f, 0.f),
    C_Vertex3f(0.232894063f, -0.120388865f, -1.f),
    C_Normal3f(0.905878186f, -0.423538417f, 0.f),
    C_Vertex3f(0.229948282f, -0.127858877f, 0.f),
    C_Vertex3f(0.229948282f, -0.127858877f, -1.f),
    C_Normal3f(0.84940356f, -0.527743995f, 0.f),
    C_Vertex3f(0.226020575f, -0.135090113f, 0.f),
    C_Vertex3f(0.226020575f, -0.135090113f, -1.f),
    C_Normal3f(0.786300123f, -0.61784476f, 0.f),
    C_Vertex3f(0.22111094f, -0.142082572f, 0.f),
    C_Vertex3f(0.22111094f, -0.142082572f, -1.f),
    C_Normal3f(0.720648766f, -0.693300307f, 0.f),
    C_Vertex3f(0.215219378f, -0.148836255f, 0.f),
    C_Vertex3f(0.215219378f, -0.148836255f, -1.f),
    C_Normal3f(0.655690968f, -0.75502938f, 0.f),
    C_Vertex3f(0.20834589f, -0.155351162f, 0.f),
    C_Vertex3f(0.20834589f, -0.155351162f, -1.f),
    C_Normal3f(0.575867295f, -0.817543089f, 0.f),
    C_Vertex3f(0.200490475f, -0.161627293f, 0.f),
    C_Vertex3f(0.200490475f, -0.161627293f, -1.f),
    C_Normal3f(0.478857249f, -0.877892852f, 0.f),
    C_Vertex3f(0.191639543f, -0.167118907f, 0.f),
    C_Vertex3f(0.191639543f, -0.167118907f, -1.f),
    C_Normal3f(0.384976745f, -0.922926307f, 0.f),
    C_Vertex3f(0.181793094f, -0.171826005f, 0.f),
    C_Vertex3f(0.181793094f, -0.171826005f, -1.f),
    C_Normal3f(0.297251523f, -0.954799294f, 0.f),
    C_Vertex3f(0.170951128f, -0.175748587f, 0.f),
    C_Vertex3f(0.170951128f, -0.175748587f, -1.f),
    C_Normal3f(0.217280537f, -0.976109207f, 0.f),
    C_Vertex3f(0.159113646f, -0.178886652f, 0.f),
    C_Vertex3f(0.159113646f, -0.178886652f, -1.f),
    C_Normal3f(0.145558298f, -0.989349663f, 0.f),
    C_Vertex3f(0.146280646f, -0.181240201f, 0.f),
    C_Vertex3f(0.146280646f, -0.181240201f, -1.f),
    C_Normal3f(0.0818652809f, -0.996643364f, 0.f),
    C_Vertex3f(0.13245213f, -0.182809234f, 0.f),
    C_Vertex3f(0.13245213f, -0.182809234f, -1.f),
    C_Normal3f(0.0247552618f, -0.999693513f, 0.f),
    C_Vertex3f(0.117628098f, -0.18359375f, 0.f),
    C_Vertex3f(0.117628098f, -0.18359375f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.10077095f, -0.18359375f, 0.f),
    C_Vertex3f(0.10077095f, -0.18359375f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0839681625f, -0.18359375f, 0.f),
    C_Vertex3f(0.0839681625f, -0.18359375f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0682573318f, -0.18359375f, 0.f),
    C_Vertex3f(0.0682573318f, -0.18359375f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0536384583f, -0.18359375f, 0.f),
    C_Vertex3f(0.0536384583f, -0.18359375f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.046875f, -0.18359375f, 0.f),
    C_Vertex3f(0.046875f, -0.18359375f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.996240556f, 0.0866296142f, 0.f),
    C_Vertex3f(0.046875f, -0.18359375f, 0.f),
    C_Vertex3f(0.046875f, -0.18359375f, -1.f),
    C_Vertex3f(0.0517578125f, -0.127441406f, -1.f),
    C_Vertex3f(0.0517578125f, -0.127441406f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.0523584522f, 0.998628378f, 0.f),
    C_Vertex3f(0.0517578125f, -0.127441406f, 0.f),
    C_Vertex3f(0.0517578125f, -0.127441406f, -1.f),
    C_Normal3f(0.0478803068f, 0.998853087f, 0.f),
    C_Vertex3f(0.0587425232f, -0.127807617f, 0.f),
    C_Vertex3f(0.0587425232f, -0.127807617f, -1.f),
    C_Normal3f(0.0408847593f, 0.999163866f, 0.f),
    C_Vertex3f(0.070857048f, -0.128356934f, 0.f),
    C_Vertex3f(0.070857048f, -0.128356934f, -1.f),
    C_Normal3f(0.0294185635f, 0.999567151f, 0.f),
    C_Vertex3f(0.0811166763f, -0.128723145f, 0.f),
    C_Vertex3f(0.0811166763f, -0.128723145f, -1.f),
    C_Normal3f(0.00834572129f, 0.999965191f, 0.f),
    C_Vertex3f(0.0895214081f, -0.12890625f, 0.f),
    C_Vertex3f(0.0895214081f, -0.12890625f, -1.f),
    C_Normal3f(-0.04573109f, 0.998953879f, 0.f),
    C_Vertex3f(0.103055954f, -0.12890625f, 0.f),
    C_Vertex3f(0.103055954f, -0.12890625f, -1.f),
    C_Normal3f(-0.132396847f, 0.991196811f, 0.f),
    C_Vertex3f(0.120519638f, -0.127487183f, 0.f),
    C_Vertex3f(0.120519638f, -0.127487183f, -1.f),
    C_Normal3f(-0.265544683f, 0.964098513f, 0.f),
    C_Vertex3f(0.13492775f, -0.124649048f, 0.f),
    C_Vertex3f(0.13492775f, -0.124649048f, -1.f),
    C_Normal3f(-0.445162714f, 0.895449579f, 0.f),
    C_Vertex3f(0.146280289f, -0.120391846f, 0.f),
    C_Vertex3f(0.146280289f, -0.120391846f, -1.f),
    C_Normal3f(-0.633790612f, 0.773504674f, 0.f),
    C_Vertex3f(0.153593063f, -0.115369797f, 0.f),
    C_Vertex3f(0.153593063f, -0.115369797f, -1.f),
    C_Normal3f(-0.790464282f, 0.612508297f, 0.f),
    C_Vertex3f(0.1583395f, -0.110510826f, 0.f),
    C_Vertex3f(0.1583395f, -0.110510826f, -1.f),
    C_Normal3f(-0.92062366f, 0.390451133f, 0.f),
    C_Vertex3f(0.161503792f, -0.105160713f, 0.f),
    C_Vertex3f(0.161503792f, -0.105160713f, -1.f),
    C_Normal3f(-0.992604256f, 0.121395536f, 0.f),
    C_Vertex3f(0.163085938f, -0.099319458f, 0.f),
    C_Vertex3f(0.163085938f, -0.099319458f, -1.f),
    C_Normal3f(-0.984436631f, -0.175739869f, 0.f),
    C_Vertex3f(0.163085938f, -0.0922241211f, 0.f),
    C_Vertex3f(0.163085938f, -0.0922241211f, -1.f),
    C_Normal3f(-0.878779709f, -0.477227718f, 0.f),
    C_Vertex3f(0.160572052f, -0.0852375031f, 0.f),
    C_Vertex3f(0.160572052f, -0.0852375031f, -1.f),
    C_Normal3f(-0.710703135f, -0.703492045f, 0.f),
    C_Vertex3f(0.156801224f, -0.0806512833f, 0.f),
    C_Vertex3f(0.156801224f, -0.0806512833f, -1.f),
    C_Normal3f(-0.553116202f, -0.833104074f, 0.f),
    C_Vertex3f(0.153222084f, -0.0778121948f, 0.f),
    C_Vertex3f(0.153222084f, -0.0778121948f, -1.f),
    C_Normal3f(-0.4364748f, -0.899716496f, 0.f),
    C_Vertex3f(0.14857769f, -0.0751914978f, 0.f),
    C_Vertex3f(0.14857769f, -0.0751914978f, -1.f),
    C_Normal3f(-0.344821334f, -0.938668311f, 0.f),
    C_Vertex3f(0.142868042f, -0.0727891922f, 0.f),
    C_Vertex3f(0.142868042f, -0.0727891922f, -1.f),
    C_Normal3f(-0.238420397f, -0.971162021f, 0.f),
    C_Vertex3f(0.13609314f, -0.070605278f, 0.f),
    C_Vertex3f(0.13609314f, -0.070605278f, -1.f),
    C_Normal3f(-0.141307235f, -0.989965796f, 0.f),
    C_Vertex3f(0.12707901f, -0.0689129829f, 0.f),
    C_Vertex3f(0.12707901f, -0.0689129829f, -1.f),
    C_Normal3f(-0.0769447833f, -0.997035325f, 0.f),
    C_Vertex3f(0.115825653f, -0.067712307f, 0.f),
    C_Vertex3f(0.115825653f, -0.067712307f, -1.f),
    C_Normal3f(-0.0382600538f, -0.999267817f, 0.f),
    C_Vertex3f(0.102333069f, -0.0670032501f, 0.f),
    C_Vertex3f(0.102333069f, -0.0670032501f, -1.f),
    C_Normal3f(-0.0138202142f, -0.999904573f, 0.f),
    C_Vertex3f(0.0944671631f, -0.0668945313f, 0.f),
    C_Vertex3f(0.0944671631f, -0.0668945313f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.156801224f, -0.0806512833f, 0.f),
    C_Vertex3f(0.160572052f, -0.0852375031f, 0.f),
    C_Vertex3f(0.161361694f, -0.0364608765f, 0.f),
    C_Vertex3f(0.163085938f, -0.0922241211f, 0.f),
    C_Vertex3f(0.170623779f, -0.0375518799f, 0.f),
    C_Vertex3f(0.170951128f, -0.175748587f, 0.f),
    C_Vertex3f(0.18685627f, -0.0415077209f, 0.f),
    C_Vertex3f(0.181793094f, -0.171826005f, 0.f),
    C_Vertex3f(0.191639543f, -0.167118907f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13609314f, -0.070605278f, 0.f),
    C_Vertex3f(0.142868042f, -0.0727891922f, 0.f),
    C_Vertex3f(0.17578125f, 0.0102539063f, 0.f),
    C_Vertex3f(0.14857769f, -0.0751914978f, 0.f),
    C_Vertex3f(0.161361694f, -0.0364608765f, 0.f),
    C_Vertex3f(0.153222084f, -0.0778121948f, 0.f),
    C_Vertex3f(0.156801224f, -0.0806512833f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.102333069f, -0.0670032501f, 0.f),
    C_Vertex3f(0.115825653f, -0.067712307f, 0.f),
    C_Vertex3f(0.1171875f, 0.0102539063f, 0.f),
    C_Vertex3f(0.12707901f, -0.0689129829f, 0.f),
    C_Vertex3f(0.17578125f, 0.0102539063f, 0.f),
    C_Vertex3f(0.13609314f, -0.070605278f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.229948282f, -0.127858877f, 0.f),
    C_Vertex3f(0.232894063f, -0.120388865f, 0.f),
    C_Vertex3f(0.233500481f, -0.0841293335f, 0.f),
    C_Vertex3f(0.234857917f, -0.112680078f, 0.f),
    C_Vertex3f(0.235839844f, -0.0949630737f, 0.f),
    C_Vertex3f(0.235839844f, -0.104732513f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215219378f, -0.148836255f, 0.f),
    C_Vertex3f(0.22111094f, -0.142082572f, 0.f),
    C_Vertex3f(0.221803665f, -0.0640182495f, 0.f),
    C_Vertex3f(0.226020575f, -0.135090113f, 0.f),
    C_Vertex3f(0.228821754f, -0.0738143921f, 0.f),
    C_Vertex3f(0.229948282f, -0.127858877f, 0.f),
    C_Vertex3f(0.233500481f, -0.0841293335f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.191639543f, -0.167118907f, 0.f),
    C_Vertex3f(0.200490475f, -0.161627293f, 0.f),
    C_Vertex3f(0.200797081f, -0.0472373962f, 0.f),
    C_Vertex3f(0.20834589f, -0.155351162f, 0.f),
    C_Vertex3f(0.212446213f, -0.0547409058f, 0.f),
    C_Vertex3f(0.215219378f, -0.148836255f, 0.f),
    C_Vertex3f(0.221803665f, -0.0640182495f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163085938f, -0.0922241211f, 0.f),
    C_Vertex3f(0.163085938f, -0.099319458f, 0.f),
    C_Vertex3f(0.170951128f, -0.175748587f, 0.f),
    C_Vertex3f(0.161503792f, -0.105160713f, 0.f),
    C_Vertex3f(0.159113646f, -0.178886652f, 0.f),
    C_Vertex3f(0.1583395f, -0.110510826f, 0.f),
    C_Vertex3f(0.153593063f, -0.115369797f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.146280646f, -0.181240201f, 0.f),
    C_Vertex3f(0.159113646f, -0.178886652f, 0.f),
    C_Vertex3f(0.153593063f, -0.115369797f, 0.f),
    C_Vertex3f(0.146280289f, -0.120391846f, 0.f),
    C_Vertex3f(0.13492775f, -0.124649048f, 0.f),
    C_Vertex3f(0.13245213f, -0.182809234f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0811166763f, -0.128723145f, 0.f),
    C_Vertex3f(0.0839681625f, -0.18359375f, 0.f),
    C_Vertex3f(0.0895214081f, -0.12890625f, 0.f),
    C_Vertex3f(0.10077095f, -0.18359375f, 0.f),
    C_Vertex3f(0.103055954f, -0.12890625f, 0.f),
    C_Vertex3f(0.117628098f, -0.18359375f, 0.f),
    C_Vertex3f(0.120519638f, -0.127487183f, 0.f),
    C_Vertex3f(0.13245213f, -0.182809234f, 0.f),
    C_Vertex3f(0.13492775f, -0.124649048f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0517578125f, -0.127441406f, 0.f),
    C_Vertex3f(0.0536384583f, -0.18359375f, 0.f),
    C_Vertex3f(0.0587425232f, -0.127807617f, 0.f),
    C_Vertex3f(0.0682573318f, -0.18359375f, 0.f),
    C_Vertex3f(0.070857048f, -0.128356934f, 0.f),
    C_Vertex3f(0.0839681625f, -0.18359375f, 0.f),
    C_Vertex3f(0.0811166763f, -0.128723145f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0536384583f, -0.18359375f, 0.f),
    C_Vertex3f(0.0517578125f, -0.127441406f, 0.f),
    C_Vertex3f(0.046875f, -0.18359375f, 0.f),
    C_Vertex3f(0.18685627f, -0.0415077209f, 0.f),
    C_Vertex3f(0.191639543f, -0.167118907f, 0.f),
    C_Vertex3f(0.200797081f, -0.0472373962f, 0.f),
    C_Vertex3f(0.102333069f, -0.0670032501f, 0.f),
    C_Vertex3f(0.1171875f, 0.0102539063f, 0.f),
    C_Vertex3f(0.0944671631f, -0.0668945313f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.156801224f, -0.0806512833f, -1.f),
    C_Vertex3f(0.160572052f, -0.0852375031f, -1.f),
    C_Vertex3f(0.161361694f, -0.0364608765f, -1.f),
    C_Vertex3f(0.163085938f, -0.0922241211f, -1.f),
    C_Vertex3f(0.170623779f, -0.0375518799f, -1.f),
    C_Vertex3f(0.170951128f, -0.175748587f, -1.f),
    C_Vertex3f(0.18685627f, -0.0415077209f, -1.f),
    C_Vertex3f(0.181793094f, -0.171826005f, -1.f),
    C_Vertex3f(0.191639543f, -0.167118907f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13609314f, -0.070605278f, -1.f),
    C_Vertex3f(0.142868042f, -0.0727891922f, -1.f),
    C_Vertex3f(0.17578125f, 0.0102539063f, -1.f),
    C_Vertex3f(0.14857769f, -0.0751914978f, -1.f),
    C_Vertex3f(0.161361694f, -0.0364608765f, -1.f),
    C_Vertex3f(0.153222084f, -0.0778121948f, -1.f),
    C_Vertex3f(0.156801224f, -0.0806512833f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.102333069f, -0.0670032501f, -1.f),
    C_Vertex3f(0.115825653f, -0.067712307f, -1.f),
    C_Vertex3f(0.1171875f, 0.0102539063f, -1.f),
    C_Vertex3f(0.12707901f, -0.0689129829f, -1.f),
    C_Vertex3f(0.17578125f, 0.0102539063f, -1.f),
    C_Vertex3f(0.13609314f, -0.070605278f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.229948282f, -0.127858877f, -1.f),
    C_Vertex3f(0.232894063f, -0.120388865f, -1.f),
    C_Vertex3f(0.233500481f, -0.0841293335f, -1.f),
    C_Vertex3f(0.234857917f, -0.112680078f, -1.f),
    C_Vertex3f(0.235839844f, -0.0949630737f, -1.f),
    C_Vertex3f(0.235839844f, -0.104732513f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215219378f, -0.148836255f, -1.f),
    C_Vertex3f(0.22111094f, -0.142082572f, -1.f),
    C_Vertex3f(0.221803665f, -0.0640182495f, -1.f),
    C_Vertex3f(0.226020575f, -0.135090113f, -1.f),
    C_Vertex3f(0.228821754f, -0.0738143921f, -1.f),
    C_Vertex3f(0.229948282f, -0.127858877f, -1.f),
    C_Vertex3f(0.233500481f, -0.0841293335f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.191639543f, -0.167118907f, -1.f),
    C_Vertex3f(0.200490475f, -0.161627293f, -1.f),
    C_Vertex3f(0.200797081f, -0.0472373962f, -1.f),
    C_Vertex3f(0.20834589f, -0.155351162f, -1.f),
    C_Vertex3f(0.212446213f, -0.0547409058f, -1.f),
    C_Vertex3f(0.215219378f, -0.148836255f, -1.f),
    C_Vertex3f(0.221803665f, -0.0640182495f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163085938f, -0.0922241211f, -1.f),
    C_Vertex3f(0.163085938f, -0.099319458f, -1.f),
    C_Vertex3f(0.170951128f, -0.175748587f, -1.f),
    C_Vertex3f(0.161503792f, -0.105160713f, -1.f),
    C_Vertex3f(0.159113646f, -0.178886652f, -1.f),
    C_Vertex3f(0.1583395f, -0.110510826f, -1.f),
    C_Vertex3f(0.153593063f, -0.115369797f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.146280646f, -0.181240201f, -1.f),
    C_Vertex3f(0.159113646f, -0.178886652f, -1.f),
    C_Vertex3f(0.153593063f, -0.115369797f, -1.f),
    C_Vertex3f(0.146280289f, -0.120391846f, -1.f),
    C_Vertex3f(0.13492775f, -0.124649048f, -1.f),
    C_Vertex3f(0.13245213f, -0.182809234f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0811166763f, -0.128723145f, -1.f),
    C_Vertex3f(0.0839681625f, -0.18359375f, -1.f),
    C_Vertex3f(0.0895214081f, -0.12890625f, -1.f),
    C_Vertex3f(0.10077095f, -0.18359375f, -1.f),
    C_Vertex3f(0.103055954f, -0.12890625f, -1.f),
    C_Vertex3f(0.117628098f, -0.18359375f, -1.f),
    C_Vertex3f(0.120519638f, -0.127487183f, -1.f),
    C_Vertex3f(0.13245213f, -0.182809234f, -1.f),
    C_Vertex3f(0.13492775f, -0.124649048f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0517578125f, -0.127441406f, -1.f),
    C_Vertex3f(0.0536384583f, -0.18359375f, -1.f),
    C_Vertex3f(0.0587425232f, -0.127807617f, -1.f),
    C_Vertex3f(0.0682573318f, -0.18359375f, -1.f),
    C_Vertex3f(0.070857048f, -0.128356934f, -1.f),
    C_Vertex3f(0.0839681625f, -0.18359375f, -1.f),
    C_Vertex3f(0.0811166763f, -0.128723145f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0536384583f, -0.18359375f, -1.f),
    C_Vertex3f(0.0517578125f, -0.127441406f, -1.f),
    C_Vertex3f(0.046875f, -0.18359375f, -1.f),
    C_Vertex3f(0.18685627f, -0.0415077209f, -1.f),
    C_Vertex3f(0.191639543f, -0.167118907f, -1.f),
    C_Vertex3f(0.200797081f, -0.0472373962f, -1.f),
    C_Vertex3f(0.102333069f, -0.0670032501f, -1.f),
    C_Vertex3f(0.1171875f, 0.0102539063f, -1.f),
    C_Vertex3f(0.0944671631f, -0.0668945313f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_185[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.14453125f, 0.3203125f, 0.f),
    C_Vertex3f(0.14453125f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.62246567f, -0.782647192f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Normal3f(0.569342554f, -0.822100401f, 0.f),
    C_Vertex3f(0.133413315f, 0.553688049f, 0.f),
    C_Vertex3f(0.133413315f, 0.553688049f, -1.f),
    C_Normal3f(0.497072011f, -0.867709279f, 0.f),
    C_Vertex3f(0.110088348f, 0.538677216f, 0.f),
    C_Vertex3f(0.110088348f, 0.538677216f, -1.f),
    C_Normal3f(0.403288066f, -0.915073037f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, -1.f),
    C_Normal3f(0.321139812f, -0.947031856f, 0.f),
    C_Vertex3f(0.0601711273f, 0.516677856f, 0.f),
    C_Vertex3f(0.0601711273f, 0.516677856f, -1.f),
    C_Normal3f(0.254172325f, -0.967158973f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, -1.f),
    C_Vertex3f(0.046875f, 0.566894531f, -1.f),
    C_Vertex3f(0.046875f, 0.566894531f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.306869417f, 0.951751649f, 0.f),
    C_Vertex3f(0.046875f, 0.566894531f, 0.f),
    C_Vertex3f(0.046875f, 0.566894531f, -1.f),
    C_Normal3f(-0.368349254f, 0.92968744f, 0.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, 0.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, -1.f),
    C_Normal3f(-0.433635086f, 0.901088595f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, -1.f),
    C_Normal3f(-0.497827768f, 0.867275894f, 0.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, 0.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, -1.f),
    C_Normal3f(-0.549703896f, 0.835359514f, 0.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, 0.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, -1.f),
    C_Normal3f(-0.597880721f, 0.801585138f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, -1.f),
    C_Normal3f(-0.651854753f, 0.758343875f, 0.f),
    C_Vertex3f(0.130358219f, 0.61649847f, 0.f),
    C_Vertex3f(0.130358219f, 0.61649847f, -1.f),
    C_Normal3f(-0.707733631f, 0.706479251f, 0.f),
    C_Vertex3f(0.143771648f, 0.6289258f, 0.f),
    C_Vertex3f(0.143771648f, 0.6289258f, -1.f),
    C_Normal3f(-0.754794896f, 0.655960798f, 0.f),
    C_Vertex3f(0.155603409f, 0.641788483f, 0.f),
    C_Vertex3f(0.155603409f, 0.641788483f, -1.f),
    C_Normal3f(-0.792024016f, 0.610490084f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.6484375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.208007813f, 0.6484375f, 0.f),
    C_Vertex3f(0.208007813f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.3203125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.208007813f, 0.6484375f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, 0.f),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Vertex3f(0.155603409f, 0.641788483f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_Vertex3f(0.143771648f, 0.6289258f, 0.f),
    C_Vertex3f(0.133413315f, 0.553688049f, 0.f),
    C_Vertex3f(0.130358219f, 0.61649847f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133413315f, 0.553688049f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, 0.f),
    C_Vertex3f(0.110088348f, 0.538677216f, 0.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, 0.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0601711273f, 0.516677856f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, 0.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, 0.f),
    C_Vertex3f(0.046875f, 0.566894531f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_Vertex3f(0.14453125f, 0.3203125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.208007813f, 0.6484375f, -1.f),
    C_Vertex3f(0.160728455f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.155603409f, 0.641788483f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Vertex3f(0.143771648f, 0.6289258f, -1.f),
    C_Vertex3f(0.133413315f, 0.553688049f, -1.f),
    C_Vertex3f(0.130358219f, 0.61649847f, -1.f),
    C_Vertex3f(0.115363121f, 0.604506493f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133413315f, 0.553688049f, -1.f),
    C_Vertex3f(0.115363121f, 0.604506493f, -1.f),
    C_Vertex3f(0.110088348f, 0.538677216f, -1.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, -1.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, -1.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, -1.f),
    C_Vertex3f(0.067923069f, 0.575233936f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0601711273f, 0.516677856f, -1.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, -1.f),
    C_Vertex3f(0.067923069f, 0.575233936f, -1.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, -1.f),
    C_Vertex3f(0.046875f, 0.566894531f, -1.f),
    C_Vertex3f(0.046875f, 0.513183594f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Vertex3f(0.14453125f, 0.3203125f, -1.f),
    C_End(),
    C_Translatef(0.297851563f, 0.f, 0.f),
};
static const Call char_186[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0440788455f, 0.999028087f, 0.f),
    C_Vertex3f(0.154922485f, 0.651367188f, 0.f),
    C_Vertex3f(0.154922485f, 0.651367188f, -1.f),
    C_Normal3f(0.0440788455f, 0.999028087f, 0.f),
    C_Vertex3f(0.170761108f, 0.651367188f, 0.f),
    C_Vertex3f(0.170761108f, 0.651367188f, -1.f),
    C_Normal3f(0.137602419f, 0.990487516f, 0.f),
    C_Vertex3f(0.185869217f, 0.650001764f, 0.f),
    C_Vertex3f(0.185869217f, 0.650001764f, -1.f),
    C_Normal3f(0.236688837f, 0.971585572f, 0.f),
    C_Vertex3f(0.200246811f, 0.647270918f, 0.f),
    C_Vertex3f(0.200246811f, 0.647270918f, -1.f),
    C_Normal3f(0.338564277f, 0.940943301f, 0.f),
    C_Vertex3f(0.21389389f, 0.643174648f, 0.f),
    C_Vertex3f(0.21389389f, 0.643174648f, -1.f),
    C_Normal3f(0.439684182f, 0.898152411f, 0.f),
    C_Vertex3f(0.226810455f, 0.637712955f, 0.f),
    C_Vertex3f(0.226810455f, 0.637712955f, -1.f),
    C_Normal3f(0.53624028f, 0.844065368f, 0.f),
    C_Vertex3f(0.238996506f, 0.630885839f, 0.f),
    C_Vertex3f(0.238996506f, 0.630885839f, -1.f),
    C_Normal3f(0.624825001f, 0.780764759f, 0.f),
    C_Vertex3f(0.250452042f, 0.6226933f, 0.f),
    C_Vertex3f(0.250452042f, 0.6226933f, -1.f),
    C_Normal3f(0.703007281f, 0.711182654f, 0.f),
    C_Vertex3f(0.261177063f, 0.613135338f, 0.f),
    C_Vertex3f(0.261177063f, 0.613135338f, -1.f),
    C_Normal3f(0.773690104f, 0.633564353f, 0.f),
    C_Vertex3f(0.27117157f, 0.602211952f, 0.f),
    C_Vertex3f(0.27117157f, 0.602211952f, -1.f),
    C_Normal3f(0.838177145f, 0.545397997f, 0.f),
    C_Vertex3f(0.279916763f, 0.590250969f, 0.f),
    C_Vertex3f(0.279916763f, 0.590250969f, -1.f),
    C_Normal3f(0.891438961f, 0.453140795f, 0.f),
    C_Vertex3f(0.287412643f, 0.577252388f, 0.f),
    C_Vertex3f(0.287412643f, 0.577252388f, -1.f),
    C_Normal3f(0.932832658f, 0.360309839f, 0.f),
    C_Vertex3f(0.29365921f, 0.563216209f, 0.f),
    C_Vertex3f(0.29365921f, 0.563216209f, -1.f),
    C_Normal3f(0.962856829f, 0.27001214f, 0.f),
    C_Vertex3f(0.298656464f, 0.548142433f, 0.f),
    C_Vertex3f(0.298656464f, 0.548142433f, -1.f),
    C_Normal3f(0.982817233f, 0.184581175f, 0.f),
    C_Vertex3f(0.302404404f, 0.532031059f, 0.f),
    C_Vertex3f(0.302404404f, 0.532031059f, -1.f),
    C_Normal3f(0.994421899f, 0.105475493f, 0.f),
    C_Vertex3f(0.30490303f, 0.514882088f, 0.f),
    C_Vertex3f(0.30490303f, 0.514882088f, -1.f),
    C_Normal3f(0.999441266f, 0.0334252752f, 0.f),
    C_Vertex3f(0.306152344f, 0.496695518f, 0.f),
    C_Vertex3f(0.306152344f, 0.496695518f, -1.f),
    C_Normal3f(0.999432206f, -0.033694081f, 0.f),
    C_Vertex3f(0.306152344f, 0.477526665f, 0.f),
    C_Vertex3f(0.306152344f, 0.477526665f, -1.f),
    C_Normal3f(0.99430424f, -0.106578842f, 0.f),
    C_Vertex3f(0.304896235f, 0.459436893f, 0.f),
    C_Vertex3f(0.304896235f, 0.459436893f, -1.f),
    C_Normal3f(0.982478917f, -0.186373562f, 0.f),
    C_Vertex3f(0.302384019f, 0.442370892f, 0.f),
    C_Vertex3f(0.302384019f, 0.442370892f, -1.f),
    C_Normal3f(0.962189913f, -0.272379696f, 0.f),
    C_Vertex3f(0.298615694f, 0.426328659f, 0.f),
    C_Vertex3f(0.298615694f, 0.426328659f, -1.f),
    C_Normal3f(0.931764901f, -0.363062114f, 0.f),
    C_Vertex3f(0.293591261f, 0.411310196f, 0.f),
    C_Vertex3f(0.293591261f, 0.411310196f, -1.f),
    C_Normal3f(0.889969289f, -0.456020534f, 0.f),
    C_Vertex3f(0.287310719f, 0.397315502f, 0.f),
    C_Vertex3f(0.287310719f, 0.397315502f, -1.f),
    C_Normal3f(0.836403131f, -0.548114777f, 0.f),
    C_Vertex3f(0.27977407f, 0.384344578f, 0.f),
    C_Vertex3f(0.27977407f, 0.384344578f, -1.f),
    C_Normal3f(0.771813333f, -0.635849237f, 0.f),
    C_Vertex3f(0.270981312f, 0.372397423f, 0.f),
    C_Vertex3f(0.270981312f, 0.372397423f, -1.f),
    C_Normal3f(0.701725781f, -0.712447107f, 0.f),
    C_Vertex3f(0.260932446f, 0.361474037f, 0.f),
    C_Vertex3f(0.260932446f, 0.361474037f, -1.f),
    C_Normal3f(0.624708354f, -0.780858219f, 0.f),
    C_Vertex3f(0.250187039f, 0.351916075f, 0.f),
    C_Vertex3f(0.250187039f, 0.351916075f, -1.f),
    C_Normal3f(0.537229955f, -0.843435824f, 0.f),
    C_Vertex3f(0.238745093f, 0.343723536f, 0.f),
    C_Vertex3f(0.238745093f, 0.343723536f, -1.f),
    C_Normal3f(0.441515058f, -0.897253871f, 0.f),
    C_Vertex3f(0.226606607f, 0.33689642f, 0.f),
    C_Vertex3f(0.226606607f, 0.33689642f, -1.f),
    C_Normal3f(0.340801597f, -0.940135181f, 0.f),
    C_Vertex3f(0.213771582f, 0.331434727f, 0.f),
    C_Vertex3f(0.213771582f, 0.331434727f, -1.f),
    C_Normal3f(0.238820046f, -0.971063912f, 0.f),
    C_Vertex3f(0.200240016f, 0.327338457f, 0.f),
    C_Vertex3f(0.200240016f, 0.327338457f, -1.f),
    C_Normal3f(0.139143705f, -0.990272164f, 0.f),
    C_Vertex3f(0.18601191f, 0.324607611f, 0.f),
    C_Vertex3f(0.18601191f, 0.324607611f, -1.f),
    C_Normal3f(0.0444981493f, -0.99900943f, 0.f),
    C_Vertex3f(0.171087265f, 0.323242188f, 0.f),
    C_Vertex3f(0.171087265f, 0.323242188f, -1.f),
    C_Normal3f(-0.0433378592f, -0.999060452f, 0.f),
    C_Vertex3f(0.155357361f, 0.323242188f, 0.f),
    C_Vertex3f(0.155357361f, 0.323242188f, -1.f),
    C_Normal3f(-0.134798601f, -0.990872979f, 0.f),
    C_Vertex3f(0.140242457f, 0.324580193f, 0.f),
    C_Vertex3f(0.140242457f, 0.324580193f, -1.f),
    C_Normal3f(-0.231928751f, -0.972732782f, 0.f),
    C_Vertex3f(0.125851274f, 0.327256203f, 0.f),
    C_Vertex3f(0.125851274f, 0.327256203f, -1.f),
    C_Normal3f(-0.331995517f, -0.943280876f, 0.f),
    C_Vertex3f(0.112183809f, 0.331270218f, 0.f),
    C_Vertex3f(0.112183809f, 0.331270218f, -1.f),
    C_Normal3f(-0.43166551f, -0.902033746f, 0.f),
    C_Vertex3f(0.0992400646f, 0.336622238f, 0.f),
    C_Vertex3f(0.0992400646f, 0.336622238f, -1.f),
    C_Normal3f(-0.527300119f, -0.849679112f, 0.f),
    C_Vertex3f(0.0870200396f, 0.343312263f, 0.f),
    C_Vertex3f(0.0870200396f, 0.343312263f, -1.f),
    C_Normal3f(-0.615567029f, -0.788084567f, 0.f),
    C_Vertex3f(0.0755237341f, 0.351340294f, 0.f),
    C_Vertex3f(0.0755237341f, 0.351340294f, -1.f),
    C_Normal3f(-0.693996608f, -0.719978213f, 0.f),
    C_Vertex3f(0.0647511482f, 0.360706329f, 0.f),
    C_Vertex3f(0.0647511482f, 0.360706329f, -1.f),
    C_Normal3f(-0.766113937f, -0.642704725f, 0.f),
    C_Vertex3f(0.054702282f, 0.37141037f, 0.f),
    C_Vertex3f(0.054702282f, 0.37141037f, -1.f),
    C_Normal3f(-0.832761526f, -0.553631783f, 0.f),
    C_Vertex3f(0.045909524f, 0.383165836f, 0.f),
    C_Vertex3f(0.045909524f, 0.383165836f, -1.f),
    C_Normal3f(-0.887879908f, -0.460075349f, 0.f),
    C_Vertex3f(0.0383728743f, 0.395972729f, 0.f),
    C_Vertex3f(0.0383728743f, 0.395972729f, -1.f),
    C_Normal3f(-0.930715799f, -0.365743041f, 0.f),
    C_Vertex3f(0.0320923328f, 0.409831047f, 0.f),
    C_Vertex3f(0.0320923328f, 0.409831047f, -1.f),
    C_Normal3f(-0.961749911f, -0.273928851f, 0.f),
    C_Vertex3f(0.0270678997f, 0.424740791f, 0.f),
    C_Vertex3f(0.0270678997f, 0.424740791f, -1.f),
    C_Normal3f(-0.982339561f, -0.187107056f, 0.f),
    C_Vertex3f(0.0232995749f, 0.440701962f, 0.f),
    C_Vertex3f(0.0232995749f, 0.440701962f, -1.f),
    C_Normal3f(-0.994278789f, -0.106816605f, 0.f),
    C_Vertex3f(0.0207873583f, 0.457714558f, 0.f),
    C_Vertex3f(0.0207873583f, 0.457714558f, -1.f),
    C_Normal3f(-0.999437988f, -0.0335201025f, 0.f),
    C_Vertex3f(0.01953125f, 0.47577858f, 0.f),
    C_Vertex3f(0.01953125f, 0.47577858f, -1.f),
    C_Normal3f(-0.999458671f, 0.032896135f, 0.f),
    C_Vertex3f(0.01953125f, 0.495166779f, 0.f),
    C_Vertex3f(0.01953125f, 0.495166779f, -1.f),
    C_Normal3f(-0.994637668f, 0.103421003f, 0.f),
    C_Vertex3f(0.0207805634f, 0.513735533f, 0.f),
    C_Vertex3f(0.0207805634f, 0.513735533f, -1.f),
    C_Normal3f(-0.983406365f, 0.18141602f, 0.f),
    C_Vertex3f(0.0232791901f, 0.531212091f, 0.f),
    C_Vertex3f(0.0232791901f, 0.531212091f, -1.f),
    C_Normal3f(-0.96393925f, 0.266122431f, 0.f),
    C_Vertex3f(0.0270271301f, 0.547596455f, 0.f),
    C_Vertex3f(0.0270271301f, 0.547596455f, -1.f),
    C_Normal3f(-0.93439585f, 0.356236577f, 0.f),
    C_Vertex3f(0.0320243835f, 0.562888622f, 0.f),
    C_Vertex3f(0.0320243835f, 0.562888622f, -1.f),
    C_Normal3f(-0.893265307f, 0.449529946f, 0.f),
    C_Vertex3f(0.0382709503f, 0.577088594f, 0.f),
    C_Vertex3f(0.0382709503f, 0.577088594f, -1.f),
    C_Normal3f(-0.839802086f, 0.542892694f, 0.f),
    C_Vertex3f(0.0457668304f, 0.590196371f, 0.f),
    C_Vertex3f(0.0457668304f, 0.590196371f, -1.f),
    C_Normal3f(-0.774429381f, 0.632660329f, 0.f),
    C_Vertex3f(0.0545120239f, 0.602211952f, 0.f),
    C_Vertex3f(0.0545120239f, 0.602211952f, -1.f),
    C_Normal3f(-0.703007281f, 0.711182654f, 0.f),
    C_Vertex3f(0.0645065308f, 0.613135338f, 0.f),
    C_Vertex3f(0.0645065308f, 0.613135338f, -1.f),
    C_Normal3f(-0.624825001f, 0.780764759f, 0.f),
    C_Vertex3f(0.0752315521f, 0.6226933f, 0.f),
    C_Vertex3f(0.0752315521f, 0.6226933f, -1.f),
    C_Normal3f(-0.53624028f, 0.844065368f, 0.f),
    C_Vertex3f(0.086687088f, 0.630885839f, 0.f),
    C_Vertex3f(0.086687088f, 0.630885839f, -1.f),
    C_Normal3f(-0.439684182f, 0.898152411f, 0.f),
    C_Vertex3f(0.0988731384f, 0.637712955f, 0.f),
    C_Vertex3f(0.0988731384f, 0.637712955f, -1.f),
    C_Normal3f(-0.338564277f, 0.940943301f, 0.f),
    C_Vertex3f(0.111789703f, 0.643174648f, 0.f),
    C_Vertex3f(0.111789703f, 0.643174648f, -1.f),
    C_Normal3f(-0.236688837f, 0.971585572f, 0.f),
    C_Vertex3f(0.125436783f, 0.647270918f, 0.f),
    C_Vertex3f(0.125436783f, 0.647270918f, -1.f),
    C_Normal3f(-0.137602419f, 0.990487516f, 0.f),
    C_Vertex3f(0.139814377f, 0.650001764f, 0.f),
    C_Vertex3f(0.139814377f, 0.650001764f, -1.f),
    C_Normal3f(-0.0440788455f, 0.999028087f, 0.f),
    C_Vertex3f(0.154922485f, 0.651367188f, 0.f),
    C_Vertex3f(0.154922485f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.0950326473f, -0.9954741f, 0.f),
    C_Vertex3f(0.17149353f, 0.59375f, 0.f),
    C_Vertex3f(0.17149353f, 0.59375f, -1.f),
    C_Normal3f(0.0926491618f, -0.995698869f, 0.f),
    C_Vertex3f(0.153970718f, 0.59375f, 0.f),
    C_Vertex3f(0.153970718f, 0.59375f, -1.f),
    C_Normal3f(0.295136631f, -0.955455124f, 0.f),
    C_Vertex3f(0.137953281f, 0.590629101f, 0.f),
    C_Vertex3f(0.137953281f, 0.590629101f, -1.f),
    C_Normal3f(0.502326965f, -0.864677787f, 0.f),
    C_Vertex3f(0.123660564f, 0.584387302f, 0.f),
    C_Vertex3f(0.123660564f, 0.584387302f, -1.f),
    C_Normal3f(0.670842648f, -0.741599739f, 0.f),
    C_Vertex3f(0.111092567f, 0.575024605f, 0.f),
    C_Vertex3f(0.111092567f, 0.575024605f, -1.f),
    C_Normal3f(0.789977729f, -0.613135457f, 0.f),
    C_Vertex3f(0.10296011f, 0.565661907f, 0.f),
    C_Vertex3f(0.10296011f, 0.565661907f, -1.f),
    C_Normal3f(0.868561924f, -0.495580614f, 0.f),
    C_Vertex3f(0.098216176f, 0.558434367f, 0.f),
    C_Vertex3f(0.098216176f, 0.558434367f, -1.f),
    C_Normal3f(0.919291437f, -0.393577456f, 0.f),
    C_Vertex3f(0.0941499472f, 0.550221086f, 0.f),
    C_Vertex3f(0.0941499472f, 0.550221086f, -1.f),
    C_Normal3f(0.953890979f, -0.300153345f, 0.f),
    C_Vertex3f(0.0907614231f, 0.541022062f, 0.f),
    C_Vertex3f(0.0907614231f, 0.541022062f, -1.f),
    C_Normal3f(0.976203382f, -0.21685718f, 0.f),
    C_Vertex3f(0.0880506039f, 0.530837297f, 0.f),
    C_Vertex3f(0.0880506039f, 0.530837297f, -1.f),
    C_Normal3f(0.989613354f, -0.143754616f, 0.f),
    C_Vertex3f(0.0860174894f, 0.519666791f, 0.f),
    C_Vertex3f(0.0860174894f, 0.519666791f, -1.f),
    C_Normal3f(0.996786237f, -0.0801075771f, 0.f),
    C_Vertex3f(0.0846620798f, 0.507510543f, 0.f),
    C_Vertex3f(0.0846620798f, 0.507510543f, -1.f),
    C_Normal3f(0.999687612f, -0.0249953773f, 0.f),
    C_Vertex3f(0.083984375f, 0.494368553f, 0.f),
    C_Vertex3f(0.083984375f, 0.494368553f, -1.f),
    C_Normal3f(0.9996683f, 0.025755262f, 0.f),
    C_Vertex3f(0.083984375f, 0.480405807f, 0.f),
    C_Vertex3f(0.083984375f, 0.480405807f, -1.f),
    C_Normal3f(0.996527255f, 0.0832680613f, 0.f),
    C_Vertex3f(0.0846757889f, 0.467531919f, 0.f),
    C_Vertex3f(0.0846757889f, 0.467531919f, -1.f),
    C_Normal3f(0.988868952f, 0.148788512f, 0.f),
    C_Vertex3f(0.0860586166f, 0.455581903f, 0.f),
    C_Vertex3f(0.0860586166f, 0.455581903f, -1.f),
    C_Normal3f(0.974752665f, 0.223287091f, 0.f),
    C_Vertex3f(0.0881328583f, 0.444555759f, 0.f),
    C_Vertex3f(0.0881328583f, 0.444555759f, -1.f),
    C_Normal3f(0.951663077f, 0.307143748f, 0.f),
    C_Vertex3f(0.0908985138f, 0.434453487f, 0.f),
    C_Vertex3f(0.0908985138f, 0.434453487f, -1.f),
    C_Normal3f(0.916569114f, 0.399876297f, 0.f),
    C_Vertex3f(0.0943555832f, 0.425275087f, 0.f),
    C_Vertex3f(0.0943555832f, 0.425275087f, -1.f),
    C_Normal3f(0.866259456f, 0.499594331f, 0.f),
    C_Vertex3f(0.0985040665f, 0.417020559f, 0.f),
    C_Vertex3f(0.0985040665f, 0.417020559f, -1.f),
    C_Normal3f(0.790240705f, 0.612796605f, 0.f),
    C_Vertex3f(0.103343964f, 0.409689903f, 0.f),
    C_Vertex3f(0.103343964f, 0.409689903f, -1.f),
    C_Normal3f(0.676563323f, 0.736384451f, 0.f),
    C_Vertex3f(0.11164093f, 0.400079727f, 0.f),
    C_Vertex3f(0.11164093f, 0.400079727f, -1.f),
    C_Normal3f(0.512927473f, 0.858431995f, 0.f),
    C_Vertex3f(0.124263763f, 0.390469551f, 0.f),
    C_Vertex3f(0.124263763f, 0.390469551f, -1.f),
    C_Normal3f(0.305749893f, 0.9521119f, 0.f),
    C_Vertex3f(0.138446808f, 0.384062767f, 0.f),
    C_Vertex3f(0.138446808f, 0.384062767f, -1.f),
    C_Normal3f(0.0961640775f, 0.995365441f, 0.f),
    C_Vertex3f(0.154190063f, 0.380859375f, 0.f),
    C_Vertex3f(0.154190063f, 0.380859375f, -1.f),
    C_Normal3f(-0.0957683548f, 0.995403588f, 0.f),
    C_Vertex3f(0.171604156f, 0.380859375f, 0.f),
    C_Vertex3f(0.171604156f, 0.380859375f, -1.f),
    C_Normal3f(-0.30396679f, 0.952682614f, 0.f),
    C_Vertex3f(0.187485695f, 0.384062767f, 0.f),
    C_Vertex3f(0.187485695f, 0.384062767f, -1.f),
    C_Normal3f(-0.512537718f, 0.858664691f, 0.f),
    C_Vertex3f(0.201724052f, 0.390469551f, 0.f),
    C_Vertex3f(0.201724052f, 0.390469551f, -1.f),
    C_Normal3f(-0.678506553f, 0.734594285f, 0.f),
    C_Vertex3f(0.214319229f, 0.400079727f, 0.f),
    C_Vertex3f(0.214319229f, 0.400079727f, -1.f),
    C_Normal3f(-0.79439348f, 0.607403338f, 0.f),
    C_Vertex3f(0.222533226f, 0.409689903f, 0.f),
    C_Vertex3f(0.222533226f, 0.409689903f, -1.f),
    C_Normal3f(-0.871189535f, 0.4909468f, 0.f),
    C_Vertex3f(0.227324724f, 0.417088985f, 0.f),
    C_Vertex3f(0.227324724f, 0.417088985f, -1.f),
    C_Normal3f(-0.920794666f, 0.390047699f, 0.f),
    C_Vertex3f(0.231431723f, 0.425480366f, 0.f),
    C_Vertex3f(0.231431723f, 0.425480366f, -1.f),
    C_Normal3f(-0.954677761f, 0.297641337f, 0.f),
    C_Vertex3f(0.234854221f, 0.434864044f, 0.f),
    C_Vertex3f(0.234854221f, 0.434864044f, -1.f),
    C_Normal3f(-0.976571262f, 0.215194315f, 0.f),
    C_Vertex3f(0.23759222f, 0.445240021f, 0.f),
    C_Vertex3f(0.23759222f, 0.445240021f, -1.f),
    C_Normal3f(-0.989757836f, 0.142756402f, 0.f),
    C_Vertex3f(0.23964572f, 0.456608295f, 0.f),
    C_Vertex3f(0.23964572f, 0.456608295f, -1.f),
    C_Normal3f(-0.996826291f, 0.0796074569f, 0.f),
    C_Vertex3f(0.241014719f, 0.468968868f, 0.f),
    C_Vertex3f(0.241014719f, 0.468968868f, -1.f),
    C_Normal3f(-0.999683619f, 0.0251519736f, 0.f),
    C_Vertex3f(0.241699219f, 0.482321739f, 0.f),
    C_Vertex3f(0.241699219f, 0.482321739f, -1.f),
    C_Normal3f(-0.999655604f, -0.0262426678f, 0.f),
    C_Vertex3f(0.241699219f, 0.496174812f, 0.f),
    C_Vertex3f(0.241699219f, 0.496174812f, -1.f),
    C_Normal3f(-0.996314585f, -0.085774444f, 0.f),
    C_Vertex3f(0.241007805f, 0.508659601f, 0.f),
    C_Vertex3f(0.241007805f, 0.508659601f, -1.f),
    C_Normal3f(-0.988238454f, -0.152920946f, 0.f),
    C_Vertex3f(0.239624977f, 0.520268202f, 0.f),
    C_Vertex3f(0.239624977f, 0.520268202f, -1.f),
    C_Normal3f(-0.973464489f, -0.228838265f, 0.f),
    C_Vertex3f(0.237550735f, 0.531000614f, 0.f),
    C_Vertex3f(0.237550735f, 0.531000614f, -1.f),
    C_Normal3f(-0.949527204f, -0.313684702f, 0.f),
    C_Vertex3f(0.23478508f, 0.540856838f, 0.f),
    C_Vertex3f(0.23478508f, 0.540856838f, -1.f),
    C_Normal3f(-0.91355902f, -0.406706184f, 0.f),
    C_Vertex3f(0.231328011f, 0.549836874f, 0.f),
    C_Vertex3f(0.231328011f, 0.549836874f, -1.f),
    C_Normal3f(-0.862674713f, -0.50575906f, 0.f),
    C_Vertex3f(0.227179527f, 0.557940722f, 0.f),
    C_Vertex3f(0.227179527f, 0.557940722f, -1.f),
    C_Normal3f(-0.786949337f, -0.617017567f, 0.f),
    C_Vertex3f(0.22233963f, 0.565168381f, 0.f),
    C_Vertex3f(0.22233963f, 0.565168381f, -1.f),
    C_Normal3f(-0.673377037f, -0.739299238f, 0.f),
    C_Vertex3f(0.214042664f, 0.574695587f, 0.f),
    C_Vertex3f(0.214042664f, 0.574695587f, -1.f),
    C_Normal3f(-0.509653032f, -0.860379994f, 0.f),
    C_Vertex3f(0.20141983f, 0.584222794f, 0.f),
    C_Vertex3f(0.20141983f, 0.584222794f, -1.f),
    C_Normal3f(-0.303354055f, -0.952877879f, 0.f),
    C_Vertex3f(0.187236786f, 0.590574265f, 0.f),
    C_Vertex3f(0.187236786f, 0.590574265f, -1.f),
    C_Normal3f(-0.0950326473f, -0.9954741f, 0.f),
    C_Vertex3f(0.17149353f, 0.59375f, 0.f),
    C_Vertex3f(0.17149353f, 0.59375f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241699219f, 0.496174812f, 0.f),
    C_Vertex3f(0.250187039f, 0.351916075f, 0.f),
    C_Vertex3f(0.250452042f, 0.6226933f, 0.f),
    C_Vertex3f(0.260932446f, 0.361474037f, 0.f),
    C_Vertex3f(0.261177063f, 0.613135338f, 0.f),
    C_Vertex3f(0.270981312f, 0.372397423f, 0.f),
    C_Vertex3f(0.27117157f, 0.602211952f, 0.f),
    C_Vertex3f(0.27977407f, 0.384344578f, 0.f),
    C_Vertex3f(0.279916763f, 0.590250969f, 0.f),
    C_Vertex3f(0.287310719f, 0.397315502f, 0.f),
    C_Vertex3f(0.287412643f, 0.577252388f, 0.f),
    C_Vertex3f(0.293591261f, 0.411310196f, 0.f),
    C_Vertex3f(0.29365921f, 0.563216209f, 0.f),
    C_Vertex3f(0.298615694f, 0.426328659f, 0.f),
    C_Vertex3f(0.298656464f, 0.548142433f, 0.f),
    C_Vertex3f(0.302384019f, 0.442370892f, 0.f),
    C_Vertex3f(0.302404404f, 0.532031059f, 0.f),
    C_Vertex3f(0.304896235f, 0.459436893f, 0.f),
    C_Vertex3f(0.30490303f, 0.514882088f, 0.f),
    C_Vertex3f(0.306152344f, 0.477526665f, 0.f),
    C_Vertex3f(0.306152344f, 0.496695518f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.23478508f, 0.540856838f, 0.f),
    C_Vertex3f(0.237550735f, 0.531000614f, 0.f),
    C_Vertex3f(0.238996506f, 0.630885839f, 0.f),
    C_Vertex3f(0.239624977f, 0.520268202f, 0.f),
    C_Vertex3f(0.250452042f, 0.6226933f, 0.f),
    C_Vertex3f(0.241007805f, 0.508659601f, 0.f),
    C_Vertex3f(0.241699219f, 0.496174812f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.214042664f, 0.574695587f, 0.f),
    C_Vertex3f(0.22233963f, 0.565168381f, 0.f),
    C_Vertex3f(0.226810455f, 0.637712955f, 0.f),
    C_Vertex3f(0.227179527f, 0.557940722f, 0.f),
    C_Vertex3f(0.238996506f, 0.630885839f, 0.f),
    C_Vertex3f(0.231328011f, 0.549836874f, 0.f),
    C_Vertex3f(0.23478508f, 0.540856838f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.170761108f, 0.651367188f, 0.f),
    C_Vertex3f(0.17149353f, 0.59375f, 0.f),
    C_Vertex3f(0.185869217f, 0.650001764f, 0.f),
    C_Vertex3f(0.187236786f, 0.590574265f, 0.f),
    C_Vertex3f(0.200246811f, 0.647270918f, 0.f),
    C_Vertex3f(0.20141983f, 0.584222794f, 0.f),
    C_Vertex3f(0.21389389f, 0.643174648f, 0.f),
    C_Vertex3f(0.214042664f, 0.574695587f, 0.f),
    C_Vertex3f(0.226810455f, 0.637712955f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.10296011f, 0.565661907f, 0.f),
    C_Vertex3f(0.111092567f, 0.575024605f, 0.f),
    C_Vertex3f(0.111789703f, 0.643174648f, 0.f),
    C_Vertex3f(0.123660564f, 0.584387302f, 0.f),
    C_Vertex3f(0.125436783f, 0.647270918f, 0.f),
    C_Vertex3f(0.137953281f, 0.590629101f, 0.f),
    C_Vertex3f(0.139814377f, 0.650001764f, 0.f),
    C_Vertex3f(0.153970718f, 0.59375f, 0.f),
    C_Vertex3f(0.154922485f, 0.651367188f, 0.f),
    C_Vertex3f(0.17149353f, 0.59375f, 0.f),
    C_Vertex3f(0.170761108f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0988731384f, 0.637712955f, 0.f),
    C_Vertex3f(0.086687088f, 0.630885839f, 0.f),
    C_Vertex3f(0.0880506039f, 0.530837297f, 0.f),
    C_Vertex3f(0.0907614231f, 0.541022062f, 0.f),
    C_Vertex3f(0.0941499472f, 0.550221086f, 0.f),
    C_Vertex3f(0.098216176f, 0.558434367f, 0.f),
    C_Vertex3f(0.10296011f, 0.565661907f, 0.f),
    C_Vertex3f(0.111789703f, 0.643174648f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.086687088f, 0.630885839f, 0.f),
    C_Vertex3f(0.0752315521f, 0.6226933f, 0.f),
    C_Vertex3f(0.0755237341f, 0.351340294f, 0.f),
    C_Vertex3f(0.083984375f, 0.494368553f, 0.f),
    C_Vertex3f(0.0846620798f, 0.507510543f, 0.f),
    C_Vertex3f(0.0860174894f, 0.519666791f, 0.f),
    C_Vertex3f(0.0880506039f, 0.530837297f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0860586166f, 0.455581903f, 0.f),
    C_Vertex3f(0.0846757889f, 0.467531919f, 0.f),
    C_Vertex3f(0.0870200396f, 0.343312263f, 0.f),
    C_Vertex3f(0.083984375f, 0.480405807f, 0.f),
    C_Vertex3f(0.0755237341f, 0.351340294f, 0.f),
    C_Vertex3f(0.083984375f, 0.494368553f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0755237341f, 0.351340294f, 0.f),
    C_Vertex3f(0.0752315521f, 0.6226933f, 0.f),
    C_Vertex3f(0.0647511482f, 0.360706329f, 0.f),
    C_Vertex3f(0.0645065308f, 0.613135338f, 0.f),
    C_Vertex3f(0.054702282f, 0.37141037f, 0.f),
    C_Vertex3f(0.0545120239f, 0.602211952f, 0.f),
    C_Vertex3f(0.045909524f, 0.383165836f, 0.f),
    C_Vertex3f(0.0457668304f, 0.590196371f, 0.f),
    C_Vertex3f(0.0383728743f, 0.395972729f, 0.f),
    C_Vertex3f(0.0382709503f, 0.577088594f, 0.f),
    C_Vertex3f(0.0320923328f, 0.409831047f, 0.f),
    C_Vertex3f(0.0320243835f, 0.562888622f, 0.f),
    C_Vertex3f(0.0270678997f, 0.424740791f, 0.f),
    C_Vertex3f(0.0270271301f, 0.547596455f, 0.f),
    C_Vertex3f(0.0232995749f, 0.440701962f, 0.f),
    C_Vertex3f(0.0232791901f, 0.531212091f, 0.f),
    C_Vertex3f(0.0207873583f, 0.457714558f, 0.f),
    C_Vertex3f(0.0207805634f, 0.513735533f, 0.f),
    C_Vertex3f(0.01953125f, 0.495166779f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.250187039f, 0.351916075f, 0.f),
    C_Vertex3f(0.241699219f, 0.496174812f, 0.f),
    C_Vertex3f(0.241699219f, 0.482321739f, 0.f),
    C_Vertex3f(0.241014719f, 0.468968868f, 0.f),
    C_Vertex3f(0.23964572f, 0.456608295f, 0.f),
    C_Vertex3f(0.238745093f, 0.343723536f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.238745093f, 0.343723536f, 0.f),
    C_Vertex3f(0.23964572f, 0.456608295f, 0.f),
    C_Vertex3f(0.23759222f, 0.445240021f, 0.f),
    C_Vertex3f(0.234854221f, 0.434864044f, 0.f),
    C_Vertex3f(0.231431723f, 0.425480366f, 0.f),
    C_Vertex3f(0.227324724f, 0.417088985f, 0.f),
    C_Vertex3f(0.226606607f, 0.33689642f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.155357361f, 0.323242188f, 0.f),
    C_Vertex3f(0.171087265f, 0.323242188f, 0.f),
    C_Vertex3f(0.171604156f, 0.380859375f, 0.f),
    C_Vertex3f(0.18601191f, 0.324607611f, 0.f),
    C_Vertex3f(0.187485695f, 0.384062767f, 0.f),
    C_Vertex3f(0.200240016f, 0.327338457f, 0.f),
    C_Vertex3f(0.201724052f, 0.390469551f, 0.f),
    C_Vertex3f(0.213771582f, 0.331434727f, 0.f),
    C_Vertex3f(0.214319229f, 0.400079727f, 0.f),
    C_Vertex3f(0.226606607f, 0.33689642f, 0.f),
    C_Vertex3f(0.222533226f, 0.409689903f, 0.f),
    C_Vertex3f(0.227324724f, 0.417088985f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11164093f, 0.400079727f, 0.f),
    C_Vertex3f(0.112183809f, 0.331270218f, 0.f),
    C_Vertex3f(0.124263763f, 0.390469551f, 0.f),
    C_Vertex3f(0.125851274f, 0.327256203f, 0.f),
    C_Vertex3f(0.138446808f, 0.384062767f, 0.f),
    C_Vertex3f(0.140242457f, 0.324580193f, 0.f),
    C_Vertex3f(0.154190063f, 0.380859375f, 0.f),
    C_Vertex3f(0.155357361f, 0.323242188f, 0.f),
    C_Vertex3f(0.171604156f, 0.380859375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.103343964f, 0.409689903f, 0.f),
    C_Vertex3f(0.0985040665f, 0.417020559f, 0.f),
    C_Vertex3f(0.0992400646f, 0.336622238f, 0.f),
    C_Vertex3f(0.112183809f, 0.331270218f, 0.f),
    C_Vertex3f(0.11164093f, 0.400079727f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0992400646f, 0.336622238f, 0.f),
    C_Vertex3f(0.0985040665f, 0.417020559f, 0.f),
    C_Vertex3f(0.0943555832f, 0.425275087f, 0.f),
    C_Vertex3f(0.0908985138f, 0.434453487f, 0.f),
    C_Vertex3f(0.0881328583f, 0.444555759f, 0.f),
    C_Vertex3f(0.0870200396f, 0.343312263f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0860586166f, 0.455581903f, 0.f),
    C_Vertex3f(0.0870200396f, 0.343312263f, 0.f),
    C_Vertex3f(0.0881328583f, 0.444555759f, 0.f),
    C_Vertex3f(0.0207873583f, 0.457714558f, 0.f),
    C_Vertex3f(0.01953125f, 0.495166779f, 0.f),
    C_Vertex3f(0.01953125f, 0.47577858f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241699219f, 0.496174812f, -1.f),
    C_Vertex3f(0.250187039f, 0.351916075f, -1.f),
    C_Vertex3f(0.250452042f, 0.6226933f, -1.f),
    C_Vertex3f(0.260932446f, 0.361474037f, -1.f),
    C_Vertex3f(0.261177063f, 0.613135338f, -1.f),
    C_Vertex3f(0.270981312f, 0.372397423f, -1.f),
    C_Vertex3f(0.27117157f, 0.602211952f, -1.f),
    C_Vertex3f(0.27977407f, 0.384344578f, -1.f),
    C_Vertex3f(0.279916763f, 0.590250969f, -1.f),
    C_Vertex3f(0.287310719f, 0.397315502f, -1.f),
    C_Vertex3f(0.287412643f, 0.577252388f, -1.f),
    C_Vertex3f(0.293591261f, 0.411310196f, -1.f),
    C_Vertex3f(0.29365921f, 0.563216209f, -1.f),
    C_Vertex3f(0.298615694f, 0.426328659f, -1.f),
    C_Vertex3f(0.298656464f, 0.548142433f, -1.f),
    C_Vertex3f(0.302384019f, 0.442370892f, -1.f),
    C_Vertex3f(0.302404404f, 0.532031059f, -1.f),
    C_Vertex3f(0.304896235f, 0.459436893f, -1.f),
    C_Vertex3f(0.30490303f, 0.514882088f, -1.f),
    C_Vertex3f(0.306152344f, 0.477526665f, -1.f),
    C_Vertex3f(0.306152344f, 0.496695518f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.23478508f, 0.540856838f, -1.f),
    C_Vertex3f(0.237550735f, 0.531000614f, -1.f),
    C_Vertex3f(0.238996506f, 0.630885839f, -1.f),
    C_Vertex3f(0.239624977f, 0.520268202f, -1.f),
    C_Vertex3f(0.250452042f, 0.6226933f, -1.f),
    C_Vertex3f(0.241007805f, 0.508659601f, -1.f),
    C_Vertex3f(0.241699219f, 0.496174812f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.214042664f, 0.574695587f, -1.f),
    C_Vertex3f(0.22233963f, 0.565168381f, -1.f),
    C_Vertex3f(0.226810455f, 0.637712955f, -1.f),
    C_Vertex3f(0.227179527f, 0.557940722f, -1.f),
    C_Vertex3f(0.238996506f, 0.630885839f, -1.f),
    C_Vertex3f(0.231328011f, 0.549836874f, -1.f),
    C_Vertex3f(0.23478508f, 0.540856838f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.170761108f, 0.651367188f, -1.f),
    C_Vertex3f(0.17149353f, 0.59375f, -1.f),
    C_Vertex3f(0.185869217f, 0.650001764f, -1.f),
    C_Vertex3f(0.187236786f, 0.590574265f, -1.f),
    C_Vertex3f(0.200246811f, 0.647270918f, -1.f),
    C_Vertex3f(0.20141983f, 0.584222794f, -1.f),
    C_Vertex3f(0.21389389f, 0.643174648f, -1.f),
    C_Vertex3f(0.214042664f, 0.574695587f, -1.f),
    C_Vertex3f(0.226810455f, 0.637712955f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.10296011f, 0.565661907f, -1.f),
    C_Vertex3f(0.111092567f, 0.575024605f, -1.f),
    C_Vertex3f(0.111789703f, 0.643174648f, -1.f),
    C_Vertex3f(0.123660564f, 0.584387302f, -1.f),
    C_Vertex3f(0.125436783f, 0.647270918f, -1.f),
    C_Vertex3f(0.137953281f, 0.590629101f, -1.f),
    C_Vertex3f(0.139814377f, 0.650001764f, -1.f),
    C_Vertex3f(0.153970718f, 0.59375f, -1.f),
    C_Vertex3f(0.154922485f, 0.651367188f, -1.f),
    C_Vertex3f(0.17149353f, 0.59375f, -1.f),
    C_Vertex3f(0.170761108f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0988731384f, 0.637712955f, -1.f),
    C_Vertex3f(0.086687088f, 0.630885839f, -1.f),
    C_Vertex3f(0.0880506039f, 0.530837297f, -1.f),
    C_Vertex3f(0.0907614231f, 0.541022062f, -1.f),
    C_Vertex3f(0.0941499472f, 0.550221086f, -1.f),
    C_Vertex3f(0.098216176f, 0.558434367f, -1.f),
    C_Vertex3f(0.10296011f, 0.565661907f, -1.f),
    C_Vertex3f(0.111789703f, 0.643174648f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.086687088f, 0.630885839f, -1.f),
    C_Vertex3f(0.0752315521f, 0.6226933f, -1.f),
    C_Vertex3f(0.0755237341f, 0.351340294f, -1.f),
    C_Vertex3f(0.083984375f, 0.494368553f, -1.f),
    C_Vertex3f(0.0846620798f, 0.507510543f, -1.f),
    C_Vertex3f(0.0860174894f, 0.519666791f, -1.f),
    C_Vertex3f(0.0880506039f, 0.530837297f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0860586166f, 0.455581903f, -1.f),
    C_Vertex3f(0.0846757889f, 0.467531919f, -1.f),
    C_Vertex3f(0.0870200396f, 0.343312263f, -1.f),
    C_Vertex3f(0.083984375f, 0.480405807f, -1.f),
    C_Vertex3f(0.0755237341f, 0.351340294f, -1.f),
    C_Vertex3f(0.083984375f, 0.494368553f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0755237341f, 0.351340294f, -1.f),
    C_Vertex3f(0.0752315521f, 0.6226933f, -1.f),
    C_Vertex3f(0.0647511482f, 0.360706329f, -1.f),
    C_Vertex3f(0.0645065308f, 0.613135338f, -1.f),
    C_Vertex3f(0.054702282f, 0.37141037f, -1.f),
    C_Vertex3f(0.0545120239f, 0.602211952f, -1.f),
    C_Vertex3f(0.045909524f, 0.383165836f, -1.f),
    C_Vertex3f(0.0457668304f, 0.590196371f, -1.f),
    C_Vertex3f(0.0383728743f, 0.395972729f, -1.f),
    C_Vertex3f(0.0382709503f, 0.577088594f, -1.f),
    C_Vertex3f(0.0320923328f, 0.409831047f, -1.f),
    C_Vertex3f(0.0320243835f, 0.562888622f, -1.f),
    C_Vertex3f(0.0270678997f, 0.424740791f, -1.f),
    C_Vertex3f(0.0270271301f, 0.547596455f, -1.f),
    C_Vertex3f(0.0232995749f, 0.440701962f, -1.f),
    C_Vertex3f(0.0232791901f, 0.531212091f, -1.f),
    C_Vertex3f(0.0207873583f, 0.457714558f, -1.f),
    C_Vertex3f(0.0207805634f, 0.513735533f, -1.f),
    C_Vertex3f(0.01953125f, 0.495166779f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.250187039f, 0.351916075f, -1.f),
    C_Vertex3f(0.241699219f, 0.496174812f, -1.f),
    C_Vertex3f(0.241699219f, 0.482321739f, -1.f),
    C_Vertex3f(0.241014719f, 0.468968868f, -1.f),
    C_Vertex3f(0.23964572f, 0.456608295f, -1.f),
    C_Vertex3f(0.238745093f, 0.343723536f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.238745093f, 0.343723536f, -1.f),
    C_Vertex3f(0.23964572f, 0.456608295f, -1.f),
    C_Vertex3f(0.23759222f, 0.445240021f, -1.f),
    C_Vertex3f(0.234854221f, 0.434864044f, -1.f),
    C_Vertex3f(0.231431723f, 0.425480366f, -1.f),
    C_Vertex3f(0.227324724f, 0.417088985f, -1.f),
    C_Vertex3f(0.226606607f, 0.33689642f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.155357361f, 0.323242188f, -1.f),
    C_Vertex3f(0.171087265f, 0.323242188f, -1.f),
    C_Vertex3f(0.171604156f, 0.380859375f, -1.f),
    C_Vertex3f(0.18601191f, 0.324607611f, -1.f),
    C_Vertex3f(0.187485695f, 0.384062767f, -1.f),
    C_Vertex3f(0.200240016f, 0.327338457f, -1.f),
    C_Vertex3f(0.201724052f, 0.390469551f, -1.f),
    C_Vertex3f(0.213771582f, 0.331434727f, -1.f),
    C_Vertex3f(0.214319229f, 0.400079727f, -1.f),
    C_Vertex3f(0.226606607f, 0.33689642f, -1.f),
    C_Vertex3f(0.222533226f, 0.409689903f, -1.f),
    C_Vertex3f(0.227324724f, 0.417088985f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11164093f, 0.400079727f, -1.f),
    C_Vertex3f(0.112183809f, 0.331270218f, -1.f),
    C_Vertex3f(0.124263763f, 0.390469551f, -1.f),
    C_Vertex3f(0.125851274f, 0.327256203f, -1.f),
    C_Vertex3f(0.138446808f, 0.384062767f, -1.f),
    C_Vertex3f(0.140242457f, 0.324580193f, -1.f),
    C_Vertex3f(0.154190063f, 0.380859375f, -1.f),
    C_Vertex3f(0.155357361f, 0.323242188f, -1.f),
    C_Vertex3f(0.171604156f, 0.380859375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.103343964f, 0.409689903f, -1.f),
    C_Vertex3f(0.0985040665f, 0.417020559f, -1.f),
    C_Vertex3f(0.0992400646f, 0.336622238f, -1.f),
    C_Vertex3f(0.112183809f, 0.331270218f, -1.f),
    C_Vertex3f(0.11164093f, 0.400079727f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0992400646f, 0.336622238f, -1.f),
    C_Vertex3f(0.0985040665f, 0.417020559f, -1.f),
    C_Vertex3f(0.0943555832f, 0.425275087f, -1.f),
    C_Vertex3f(0.0908985138f, 0.434453487f, -1.f),
    C_Vertex3f(0.0881328583f, 0.444555759f, -1.f),
    C_Vertex3f(0.0870200396f, 0.343312263f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0860586166f, 0.455581903f, -1.f),
    C_Vertex3f(0.0870200396f, 0.343312263f, -1.f),
    C_Vertex3f(0.0881328583f, 0.444555759f, -1.f),
    C_Vertex3f(0.0207873583f, 0.457714558f, -1.f),
    C_Vertex3f(0.01953125f, 0.495166779f, -1.f),
    C_Vertex3f(0.01953125f, 0.47577858f, -1.f),
    C_End(),
    C_Translatef(0.326660156f, 0.f, 0.f),
};
static const Call char_187[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.866538942f, -0.499109447f, 0.f),
    C_Vertex3f(0.345703125f, 0.23046875f, 0.f),
    C_Vertex3f(0.345703125f, 0.23046875f, -1.f),
    C_Vertex3f(0.230957031f, 0.4296875f, -1.f),
    C_Vertex3f(0.230957031f, 0.4296875f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.230957031f, 0.4296875f, 0.f),
    C_Vertex3f(0.230957031f, 0.4296875f, -1.f),
    C_Vertex3f(0.295410156f, 0.4296875f, -1.f),
    C_Vertex3f(0.295410156f, 0.4296875f, 0.f),
    C_Normal3f(0.817911267f, 0.575344443f, 0.f),
    C_Vertex3f(0.295410156f, 0.4296875f, 0.f),
    C_Vertex3f(0.295410156f, 0.4296875f, -1.f),
    C_Vertex3f(0.435546875f, 0.23046875f, -1.f),
    C_Vertex3f(0.435546875f, 0.23046875f, 0.f),
    C_Normal3f(0.81885463f, -0.574001014f, 0.f),
    C_Vertex3f(0.435546875f, 0.23046875f, 0.f),
    C_Vertex3f(0.435546875f, 0.23046875f, -1.f),
    C_Vertex3f(0.295898438f, 0.03125f, -1.f),
    C_Vertex3f(0.295898438f, 0.03125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.295898438f, 0.03125f, 0.f),
    C_Vertex3f(0.295898438f, 0.03125f, -1.f),
    C_Vertex3f(0.231445313f, 0.03125f, -1.f),
    C_Vertex3f(0.231445313f, 0.03125f, 0.f),
    C_Normal3f(-0.867457092f, 0.497512132f, 0.f),
    C_Vertex3f(0.231445313f, 0.03125f, 0.f),
    C_Vertex3f(0.231445313f, 0.03125f, -1.f),
    C_Vertex3f(0.345703125f, 0.23046875f, -1.f),
    C_Vertex3f(0.345703125f, 0.23046875f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.864699841f, -0.502288878f, 0.f),
    C_Vertex3f(0.176757813f, 0.23046875f, 0.f),
    C_Vertex3f(0.176757813f, 0.23046875f, -1.f),
    C_Vertex3f(0.0610351563f, 0.4296875f, -1.f),
    C_Vertex3f(0.0610351563f, 0.4296875f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0610351563f, 0.4296875f, 0.f),
    C_Vertex3f(0.0610351563f, 0.4296875f, -1.f),
    C_Vertex3f(0.126953125f, 0.4296875f, -1.f),
    C_Vertex3f(0.126953125f, 0.4296875f, 0.f),
    C_Normal3f(0.822627425f, 0.568580747f, 0.f),
    C_Vertex3f(0.126953125f, 0.4296875f, 0.f),
    C_Vertex3f(0.126953125f, 0.4296875f, -1.f),
    C_Vertex3f(0.264648438f, 0.23046875f, -1.f),
    C_Vertex3f(0.264648438f, 0.23046875f, 0.f),
    C_Normal3f(0.822627425f, -0.568580747f, 0.f),
    C_Vertex3f(0.264648438f, 0.23046875f, 0.f),
    C_Vertex3f(0.264648438f, 0.23046875f, -1.f),
    C_Vertex3f(0.126953125f, 0.03125f, -1.f),
    C_Vertex3f(0.126953125f, 0.03125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.126953125f, 0.03125f, 0.f),
    C_Vertex3f(0.126953125f, 0.03125f, -1.f),
    C_Vertex3f(0.0610351563f, 0.03125f, -1.f),
    C_Vertex3f(0.0610351563f, 0.03125f, 0.f),
    C_Normal3f(-0.864699841f, 0.502288878f, 0.f),
    C_Vertex3f(0.0610351563f, 0.03125f, 0.f),
    C_Vertex3f(0.0610351563f, 0.03125f, -1.f),
    C_Vertex3f(0.176757813f, 0.23046875f, -1.f),
    C_Vertex3f(0.176757813f, 0.23046875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.345703125f, 0.23046875f, 0.f),
    C_Vertex3f(0.231445313f, 0.03125f, 0.f),
    C_Vertex3f(0.295898438f, 0.03125f, 0.f),
    C_Vertex3f(0.435546875f, 0.23046875f, 0.f),
    C_Vertex3f(0.295410156f, 0.4296875f, 0.f),
    C_Vertex3f(0.230957031f, 0.4296875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.176757813f, 0.23046875f, 0.f),
    C_Vertex3f(0.0610351563f, 0.03125f, 0.f),
    C_Vertex3f(0.126953125f, 0.03125f, 0.f),
    C_Vertex3f(0.264648438f, 0.23046875f, 0.f),
    C_Vertex3f(0.126953125f, 0.4296875f, 0.f),
    C_Vertex3f(0.0610351563f, 0.4296875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.345703125f, 0.23046875f, -1.f),
    C_Vertex3f(0.231445313f, 0.03125f, -1.f),
    C_Vertex3f(0.295898438f, 0.03125f, -1.f),
    C_Vertex3f(0.435546875f, 0.23046875f, -1.f),
    C_Vertex3f(0.295410156f, 0.4296875f, -1.f),
    C_Vertex3f(0.230957031f, 0.4296875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.176757813f, 0.23046875f, -1.f),
    C_Vertex3f(0.0610351563f, 0.03125f, -1.f),
    C_Vertex3f(0.126953125f, 0.03125f, -1.f),
    C_Vertex3f(0.264648438f, 0.23046875f, -1.f),
    C_Vertex3f(0.126953125f, 0.4296875f, -1.f),
    C_Vertex3f(0.0610351563f, 0.4296875f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_188[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.81474638f, 0.579817414f, 0.f),
    C_Vertex3f(0.110160828f, -0.0249023438f, 0.f),
    C_Vertex3f(0.110160828f, -0.0249023438f, -1.f),
    C_Vertex3f(0.591430664f, 0.651367188f, -1.f),
    C_Vertex3f(0.591430664f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.591430664f, 0.651367188f, 0.f),
    C_Vertex3f(0.591430664f, 0.651367188f, -1.f),
    C_Vertex3f(0.66003418f, 0.651367188f, -1.f),
    C_Vertex3f(0.66003418f, 0.651367188f, 0.f),
    C_Normal3f(0.814755082f, -0.579805255f, 0.f),
    C_Vertex3f(0.66003418f, 0.651367188f, 0.f),
    C_Vertex3f(0.66003418f, 0.651367188f, -1.f),
    C_Vertex3f(0.178779602f, -0.0249023438f, -1.f),
    C_Vertex3f(0.178779602f, -0.0249023438f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.178779602f, -0.0249023438f, 0.f),
    C_Vertex3f(0.178779602f, -0.0249023438f, -1.f),
    C_Vertex3f(0.110160828f, -0.0249023438f, -1.f),
    C_Vertex3f(0.110160828f, -0.0249023438f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.14453125f, 0.3203125f, 0.f),
    C_Vertex3f(0.14453125f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.62246567f, -0.782647192f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Normal3f(0.569342554f, -0.822100401f, 0.f),
    C_Vertex3f(0.133413315f, 0.553688049f, 0.f),
    C_Vertex3f(0.133413315f, 0.553688049f, -1.f),
    C_Normal3f(0.497072011f, -0.867709279f, 0.f),
    C_Vertex3f(0.110088348f, 0.538677216f, 0.f),
    C_Vertex3f(0.110088348f, 0.538677216f, -1.f),
    C_Normal3f(0.403288066f, -0.915073037f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, -1.f),
    C_Normal3f(0.321139812f, -0.947031856f, 0.f),
    C_Vertex3f(0.0601711273f, 0.516677856f, 0.f),
    C_Vertex3f(0.0601711273f, 0.516677856f, -1.f),
    C_Normal3f(0.254172325f, -0.967158973f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, -1.f),
    C_Vertex3f(0.046875f, 0.566894531f, -1.f),
    C_Vertex3f(0.046875f, 0.566894531f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.306869417f, 0.951751649f, 0.f),
    C_Vertex3f(0.046875f, 0.566894531f, 0.f),
    C_Vertex3f(0.046875f, 0.566894531f, -1.f),
    C_Normal3f(-0.368349254f, 0.92968744f, 0.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, 0.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, -1.f),
    C_Normal3f(-0.433635086f, 0.901088595f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, -1.f),
    C_Normal3f(-0.497827768f, 0.867275894f, 0.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, 0.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, -1.f),
    C_Normal3f(-0.549703896f, 0.835359514f, 0.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, 0.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, -1.f),
    C_Normal3f(-0.597880721f, 0.801585138f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, -1.f),
    C_Normal3f(-0.651854753f, 0.758343875f, 0.f),
    C_Vertex3f(0.130358219f, 0.61649847f, 0.f),
    C_Vertex3f(0.130358219f, 0.61649847f, -1.f),
    C_Normal3f(-0.707733631f, 0.706479251f, 0.f),
    C_Vertex3f(0.143771648f, 0.6289258f, 0.f),
    C_Vertex3f(0.143771648f, 0.6289258f, -1.f),
    C_Normal3f(-0.754794896f, 0.655960798f, 0.f),
    C_Vertex3f(0.155603409f, 0.641788483f, 0.f),
    C_Vertex3f(0.155603409f, 0.641788483f, -1.f),
    C_Normal3f(-0.792024016f, 0.610490084f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.6484375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.208007813f, 0.6484375f, 0.f),
    C_Vertex3f(0.208007813f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.3203125f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.625f, -0.013671875f, 0.f),
    C_Vertex3f(0.625f, -0.013671875f, -1.f),
    C_Vertex3f(0.625f, 0.0537109375f, -1.f),
    C_Vertex3f(0.625f, 0.0537109375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.625f, 0.0537109375f, 0.f),
    C_Vertex3f(0.625f, 0.0537109375f, -1.f),
    C_Vertex3f(0.457519531f, 0.0537109375f, -1.f),
    C_Vertex3f(0.457519531f, 0.0537109375f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.457519531f, 0.0537109375f, 0.f),
    C_Vertex3f(0.457519531f, 0.0537109375f, -1.f),
    C_Vertex3f(0.457519531f, 0.107620239f, -1.f),
    C_Vertex3f(0.457519531f, 0.107620239f, 0.f),
    C_Normal3f(-0.759590805f, 0.650401354f, 0.f),
    C_Vertex3f(0.457519531f, 0.107620239f, 0.f),
    C_Vertex3f(0.457519531f, 0.107620239f, -1.f),
    C_Vertex3f(0.634620667f, 0.314453125f, -1.f),
    C_Vertex3f(0.634620667f, 0.314453125f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.634620667f, 0.314453125f, 0.f),
    C_Vertex3f(0.634620667f, 0.314453125f, -1.f),
    C_Vertex3f(0.687988281f, 0.314453125f, -1.f),
    C_Vertex3f(0.687988281f, 0.314453125f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.687988281f, 0.314453125f, 0.f),
    C_Vertex3f(0.687988281f, 0.314453125f, -1.f),
    C_Vertex3f(0.687988281f, 0.101074219f, -1.f),
    C_Vertex3f(0.687988281f, 0.101074219f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.687988281f, 0.101074219f, 0.f),
    C_Vertex3f(0.687988281f, 0.101074219f, -1.f),
    C_Vertex3f(0.733398438f, 0.101074219f, -1.f),
    C_Vertex3f(0.733398438f, 0.101074219f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.733398438f, 0.101074219f, 0.f),
    C_Vertex3f(0.733398438f, 0.101074219f, -1.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, -1.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, 0.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, -1.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, -1.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, 0.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, -1.f),
    C_Vertex3f(0.687988281f, -0.013671875f, -1.f),
    C_Vertex3f(0.687988281f, -0.013671875f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.687988281f, -0.013671875f, 0.f),
    C_Vertex3f(0.687988281f, -0.013671875f, -1.f),
    C_Vertex3f(0.625f, -0.013671875f, -1.f),
    C_Vertex3f(0.625f, -0.013671875f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.625f, 0.101074219f, 0.f),
    C_Vertex3f(0.625f, 0.101074219f, -1.f),
    C_Vertex3f(0.625f, 0.215835571f, -1.f),
    C_Vertex3f(0.625f, 0.215835571f, 0.f),
    C_Normal3f(0.752049923f, -0.659106076f, 0.f),
    C_Vertex3f(0.625f, 0.215835571f, 0.f),
    C_Vertex3f(0.625f, 0.215835571f, -1.f),
    C_Vertex3f(0.524421692f, 0.101074219f, -1.f),
    C_Vertex3f(0.524421692f, 0.101074219f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.524421692f, 0.101074219f, 0.f),
    C_Vertex3f(0.524421692f, 0.101074219f, -1.f),
    C_Vertex3f(0.625f, 0.101074219f, -1.f),
    C_Vertex3f(0.625f, 0.101074219f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.591430664f, 0.651367188f, 0.f),
    C_Vertex3f(0.110160828f, -0.0249023438f, 0.f),
    C_Vertex3f(0.178779602f, -0.0249023438f, 0.f),
    C_Vertex3f(0.66003418f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.208007813f, 0.6484375f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, 0.f),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Vertex3f(0.155603409f, 0.641788483f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_Vertex3f(0.143771648f, 0.6289258f, 0.f),
    C_Vertex3f(0.133413315f, 0.553688049f, 0.f),
    C_Vertex3f(0.130358219f, 0.61649847f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133413315f, 0.553688049f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, 0.f),
    C_Vertex3f(0.110088348f, 0.538677216f, 0.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, 0.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0601711273f, 0.516677856f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, 0.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, 0.f),
    C_Vertex3f(0.046875f, 0.566894531f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.457519531f, 0.0537109375f, 0.f),
    C_Vertex3f(0.625f, 0.0537109375f, 0.f),
    C_Vertex3f(0.457519531f, 0.107620239f, 0.f),
    C_Vertex3f(0.524421692f, 0.101074219f, 0.f),
    C_Vertex3f(0.634620667f, 0.314453125f, 0.f),
    C_Vertex3f(0.625f, 0.215835571f, 0.f),
    C_Vertex3f(0.687988281f, -0.013671875f, 0.f),
    C_Vertex3f(0.625f, 0.101074219f, 0.f),
    C_Vertex3f(0.625f, 0.0537109375f, 0.f),
    C_Vertex3f(0.524421692f, 0.101074219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.634620667f, 0.314453125f, 0.f),
    C_Vertex3f(0.687988281f, -0.013671875f, 0.f),
    C_Vertex3f(0.687988281f, 0.314453125f, 0.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, 0.f),
    C_Vertex3f(0.687988281f, 0.101074219f, 0.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, 0.f),
    C_Vertex3f(0.733398438f, 0.101074219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.687988281f, -0.013671875f, 0.f),
    C_Vertex3f(0.625f, 0.0537109375f, 0.f),
    C_Vertex3f(0.625f, -0.013671875f, 0.f),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_Vertex3f(0.14453125f, 0.3203125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.591430664f, 0.651367188f, -1.f),
    C_Vertex3f(0.110160828f, -0.0249023438f, -1.f),
    C_Vertex3f(0.178779602f, -0.0249023438f, -1.f),
    C_Vertex3f(0.66003418f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.208007813f, 0.6484375f, -1.f),
    C_Vertex3f(0.160728455f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.155603409f, 0.641788483f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Vertex3f(0.143771648f, 0.6289258f, -1.f),
    C_Vertex3f(0.133413315f, 0.553688049f, -1.f),
    C_Vertex3f(0.130358219f, 0.61649847f, -1.f),
    C_Vertex3f(0.115363121f, 0.604506493f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133413315f, 0.553688049f, -1.f),
    C_Vertex3f(0.115363121f, 0.604506493f, -1.f),
    C_Vertex3f(0.110088348f, 0.538677216f, -1.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, -1.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, -1.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, -1.f),
    C_Vertex3f(0.067923069f, 0.575233936f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0601711273f, 0.516677856f, -1.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, -1.f),
    C_Vertex3f(0.067923069f, 0.575233936f, -1.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, -1.f),
    C_Vertex3f(0.046875f, 0.566894531f, -1.f),
    C_Vertex3f(0.046875f, 0.513183594f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.457519531f, 0.0537109375f, -1.f),
    C_Vertex3f(0.625f, 0.0537109375f, -1.f),
    C_Vertex3f(0.457519531f, 0.107620239f, -1.f),
    C_Vertex3f(0.524421692f, 0.101074219f, -1.f),
    C_Vertex3f(0.634620667f, 0.314453125f, -1.f),
    C_Vertex3f(0.625f, 0.215835571f, -1.f),
    C_Vertex3f(0.687988281f, -0.013671875f, -1.f),
    C_Vertex3f(0.625f, 0.101074219f, -1.f),
    C_Vertex3f(0.625f, 0.0537109375f, -1.f),
    C_Vertex3f(0.524421692f, 0.101074219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.634620667f, 0.314453125f, -1.f),
    C_Vertex3f(0.687988281f, -0.013671875f, -1.f),
    C_Vertex3f(0.687988281f, 0.314453125f, -1.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, -1.f),
    C_Vertex3f(0.687988281f, 0.101074219f, -1.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, -1.f),
    C_Vertex3f(0.733398438f, 0.101074219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.687988281f, -0.013671875f, -1.f),
    C_Vertex3f(0.625f, 0.0537109375f, -1.f),
    C_Vertex3f(0.625f, -0.013671875f, -1.f),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Vertex3f(0.14453125f, 0.3203125f, -1.f),
    C_End(),
    C_Translatef(0.746582031f, 0.f, 0.f),
};
static const Call char_189[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.814755082f, 0.579805255f, 0.f),
    C_Vertex3f(0.0996627808f, -0.0249023438f, 0.f),
    C_Vertex3f(0.0996627808f, -0.0249023438f, -1.f),
    C_Vertex3f(0.580917358f, 0.651367188f, -1.f),
    C_Vertex3f(0.580917358f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.580917358f, 0.651367188f, 0.f),
    C_Vertex3f(0.580917358f, 0.651367188f, -1.f),
    C_Vertex3f(0.649551392f, 0.651367188f, -1.f),
    C_Vertex3f(0.649551392f, 0.651367188f, 0.f),
    C_Normal3f(0.81474638f, -0.579817414f, 0.f),
    C_Vertex3f(0.649551392f, 0.651367188f, 0.f),
    C_Vertex3f(0.649551392f, 0.651367188f, -1.f),
    C_Vertex3f(0.168281555f, -0.0249023438f, -1.f),
    C_Vertex3f(0.168281555f, -0.0249023438f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.168281555f, -0.0249023438f, 0.f),
    C_Vertex3f(0.168281555f, -0.0249023438f, -1.f),
    C_Vertex3f(0.0996627808f, -0.0249023438f, -1.f),
    C_Vertex3f(0.0996627808f, -0.0249023438f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.14453125f, 0.3203125f, 0.f),
    C_Vertex3f(0.14453125f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.62246567f, -0.782647192f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Normal3f(0.569342554f, -0.822100401f, 0.f),
    C_Vertex3f(0.133413315f, 0.553688049f, 0.f),
    C_Vertex3f(0.133413315f, 0.553688049f, -1.f),
    C_Normal3f(0.497072011f, -0.867709279f, 0.f),
    C_Vertex3f(0.110088348f, 0.538677216f, 0.f),
    C_Vertex3f(0.110088348f, 0.538677216f, -1.f),
    C_Normal3f(0.403288066f, -0.915073037f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, -1.f),
    C_Normal3f(0.321139812f, -0.947031856f, 0.f),
    C_Vertex3f(0.0601711273f, 0.516677856f, 0.f),
    C_Vertex3f(0.0601711273f, 0.516677856f, -1.f),
    C_Normal3f(0.254172325f, -0.967158973f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, -1.f),
    C_Vertex3f(0.046875f, 0.566894531f, -1.f),
    C_Vertex3f(0.046875f, 0.566894531f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.306869417f, 0.951751649f, 0.f),
    C_Vertex3f(0.046875f, 0.566894531f, 0.f),
    C_Vertex3f(0.046875f, 0.566894531f, -1.f),
    C_Normal3f(-0.368349254f, 0.92968744f, 0.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, 0.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, -1.f),
    C_Normal3f(-0.433635086f, 0.901088595f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, -1.f),
    C_Normal3f(-0.497827768f, 0.867275894f, 0.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, 0.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, -1.f),
    C_Normal3f(-0.549703896f, 0.835359514f, 0.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, 0.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, -1.f),
    C_Normal3f(-0.597880721f, 0.801585138f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, -1.f),
    C_Normal3f(-0.651854753f, 0.758343875f, 0.f),
    C_Vertex3f(0.130358219f, 0.61649847f, 0.f),
    C_Vertex3f(0.130358219f, 0.61649847f, -1.f),
    C_Normal3f(-0.707733631f, 0.706479251f, 0.f),
    C_Vertex3f(0.143771648f, 0.6289258f, 0.f),
    C_Vertex3f(0.143771648f, 0.6289258f, -1.f),
    C_Normal3f(-0.754794896f, 0.655960798f, 0.f),
    C_Vertex3f(0.155603409f, 0.641788483f, 0.f),
    C_Vertex3f(0.155603409f, 0.641788483f, -1.f),
    C_Normal3f(-0.792024016f, 0.610490084f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.6484375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.208007813f, 0.6484375f, 0.f),
    C_Vertex3f(0.208007813f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.3203125f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.994285583f, 0.106752262f, 0.f),
    C_Vertex3f(0.458984375f, -0.013671875f, 0.f),
    C_Vertex3f(0.458984375f, -0.013671875f, -1.f),
    C_Normal3f(-0.977850556f, 0.209303945f, 0.f),
    C_Vertex3f(0.459640503f, -0.00756072998f, 0.f),
    C_Vertex3f(0.459640503f, -0.00756072998f, -1.f),
    C_Normal3f(-0.944881916f, 0.327411115f, 0.f),
    C_Vertex3f(0.462920189f, 0.00471591949f, 0.f),
    C_Vertex3f(0.462920189f, 0.00471591949f, -1.f),
    C_Normal3f(-0.892926931f, 0.45020166f, 0.f),
    C_Vertex3f(0.468167305f, 0.0170469284f, 0.f),
    C_Vertex3f(0.468167305f, 0.0170469284f, -1.f),
    C_Normal3f(-0.844362855f, 0.535771608f, 0.f),
    C_Vertex3f(0.475381851f, 0.0294322968f, 0.f),
    C_Vertex3f(0.475381851f, 0.0294322968f, -1.f),
    C_Normal3f(-0.80865705f, 0.588280439f, 0.f),
    C_Vertex3f(0.479972839f, 0.0356521606f, 0.f),
    C_Vertex3f(0.479972839f, 0.0356521606f, -1.f),
    C_Normal3f(-0.788625956f, 0.61487323f, 0.f),
    C_Vertex3f(0.483360291f, 0.0403995514f, 0.f),
    C_Vertex3f(0.483360291f, 0.0403995514f, -1.f),
    C_Normal3f(-0.758400083f, 0.651789308f, 0.f),
    C_Vertex3f(0.491257906f, 0.050126195f, 0.f),
    C_Vertex3f(0.491257906f, 0.050126195f, -1.f),
    C_Normal3f(-0.724605322f, 0.689164102f, 0.f),
    C_Vertex3f(0.500278234f, 0.0600847006f, 0.f),
    C_Vertex3f(0.500278234f, 0.0600847006f, -1.f),
    C_Normal3f(-0.693584383f, 0.720375299f, 0.f),
    C_Vertex3f(0.510421276f, 0.0702750683f, 0.f),
    C_Vertex3f(0.510421276f, 0.0702750683f, -1.f),
    C_Normal3f(-0.665251136f, 0.746619761f, 0.f),
    C_Vertex3f(0.521687031f, 0.080697298f, 0.f),
    C_Vertex3f(0.521687031f, 0.080697298f, -1.f),
    C_Normal3f(-0.639430881f, 0.768848658f, 0.f),
    C_Vertex3f(0.534075499f, 0.0913513899f, 0.f),
    C_Vertex3f(0.534075499f, 0.0913513899f, -1.f),
    C_Normal3f(-0.615912199f, 0.787814796f, 0.f),
    C_Vertex3f(0.547586679f, 0.102237344f, 0.f),
    C_Vertex3f(0.547586679f, 0.102237344f, -1.f),
    C_Normal3f(-0.594475746f, 0.804113567f, 0.f),
    C_Vertex3f(0.562220573f, 0.11335516f, 0.f),
    C_Vertex3f(0.562220573f, 0.11335516f, -1.f),
    C_Normal3f(-0.58726275f, 0.809396267f, 0.f),
    C_Vertex3f(0.57797718f, 0.124704838f, 0.f),
    C_Vertex3f(0.57797718f, 0.124704838f, -1.f),
    C_Normal3f(-0.593766391f, 0.804637492f, 0.f),
    C_Vertex3f(0.592133284f, 0.135058522f, 0.f),
    C_Vertex3f(0.592133284f, 0.135058522f, -1.f),
    C_Normal3f(-0.601875901f, 0.798589706f, 0.f),
    C_Vertex3f(0.604688883f, 0.144416213f, 0.f),
    C_Vertex3f(0.604688883f, 0.144416213f, -1.f),
    C_Normal3f(-0.612264991f, 0.790652633f, 0.f),
    C_Vertex3f(0.615643978f, 0.15277791f, 0.f),
    C_Vertex3f(0.615643978f, 0.15277791f, -1.f),
    C_Normal3f(-0.626041174f, 0.779789925f, 0.f),
    C_Vertex3f(0.624998569f, 0.160143614f, 0.f),
    C_Vertex3f(0.624998569f, 0.160143614f, -1.f),
    C_Normal3f(-0.645154238f, 0.764052391f, 0.f),
    C_Vertex3f(0.632752657f, 0.166513324f, 0.f),
    C_Vertex3f(0.632752657f, 0.166513324f, -1.f),
    C_Normal3f(-0.673357129f, 0.739317358f, 0.f),
    C_Vertex3f(0.63890624f, 0.17188704f, 0.f),
    C_Vertex3f(0.63890624f, 0.17188704f, -1.f),
    C_Normal3f(-0.709395409f, 0.704810739f, 0.f),
    C_Vertex3f(0.64345932f, 0.176264763f, 0.f),
    C_Vertex3f(0.64345932f, 0.176264763f, -1.f),
    C_Normal3f(-0.744695783f, 0.667404056f, 0.f),
    C_Vertex3f(0.644935608f, 0.177955627f, 0.f),
    C_Vertex3f(0.644935608f, 0.177955627f, -1.f),
    C_Normal3f(-0.800764501f, 0.598979294f, 0.f),
    C_Vertex3f(0.648984909f, 0.182430267f, 0.f),
    C_Vertex3f(0.648984909f, 0.182430267f, -1.f),
    C_Normal3f(-0.87422204f, 0.485526443f, 0.f),
    C_Vertex3f(0.655058861f, 0.19148922f, 0.f),
    C_Vertex3f(0.655058861f, 0.19148922f, -1.f),
    C_Normal3f(-0.949835777f, 0.312748998f, 0.f),
    C_Vertex3f(0.659108162f, 0.200657845f, 0.f),
    C_Vertex3f(0.659108162f, 0.200657845f, -1.f),
    C_Normal3f(-0.994491637f, 0.104815856f, 0.f),
    C_Vertex3f(0.661132813f, 0.209936142f, 0.f),
    C_Vertex3f(0.661132813f, 0.209936142f, -1.f),
    C_Normal3f(-0.995323837f, -0.0965947658f, 0.f),
    C_Vertex3f(0.661132813f, 0.219867706f, 0.f),
    C_Vertex3f(0.661132813f, 0.219867706f, -1.f),
    C_Normal3f(-0.954308867f, -0.298822194f, 0.f),
    C_Vertex3f(0.659245014f, 0.229388237f, 0.f),
    C_Vertex3f(0.659245014f, 0.229388237f, -1.f),
    C_Normal3f(-0.863725305f, -0.503962815f, 0.f),
    C_Vertex3f(0.655469418f, 0.23795414f, 0.f),
    C_Vertex3f(0.655469418f, 0.23795414f, -1.f),
    C_Normal3f(-0.733669639f, -0.679506302f, 0.f),
    C_Vertex3f(0.649806023f, 0.245565414f, 0.f),
    C_Vertex3f(0.649806023f, 0.245565414f, -1.f),
    C_Normal3f(-0.565486073f, -0.824757814f, 0.f),
    C_Vertex3f(0.642254829f, 0.252222061f, 0.f),
    C_Vertex3f(0.642254829f, 0.252222061f, -1.f),
    C_Normal3f(-0.371949226f, -0.928253114f, 0.f),
    C_Vertex3f(0.632815838f, 0.257214546f, 0.f),
    C_Vertex3f(0.632815838f, 0.257214546f, -1.f),
    C_Normal3f(-0.199348196f, -0.979928732f, 0.f),
    C_Vertex3f(0.621489048f, 0.26054287f, 0.f),
    C_Vertex3f(0.621489048f, 0.26054287f, -1.f),
    C_Normal3f(-0.0593530759f, -0.998237133f, 0.f),
    C_Vertex3f(0.60827446f, 0.262207031f, 0.f),
    C_Vertex3f(0.60827446f, 0.262207031f, -1.f),
    C_Normal3f(0.046766106f, -0.998905897f, 0.f),
    C_Vertex3f(0.593500137f, 0.262207031f, 0.f),
    C_Vertex3f(0.593500137f, 0.262207031f, -1.f),
    C_Normal3f(0.162252486f, -0.986749351f, 0.f),
    C_Vertex3f(0.580886841f, 0.260924816f, 0.f),
    C_Vertex3f(0.580886841f, 0.260924816f, -1.f),
    C_Normal3f(0.309067369f, -0.951040149f, 0.f),
    C_Vertex3f(0.570106506f, 0.258360386f, 0.f),
    C_Vertex3f(0.570106506f, 0.258360386f, -1.f),
    C_Normal3f(0.487813175f, -0.872947931f, 0.f),
    C_Vertex3f(0.561159134f, 0.254513741f, 0.f),
    C_Vertex3f(0.561159134f, 0.254513741f, -1.f),
    C_Normal3f(0.678930998f, -0.734202087f, 0.f),
    C_Vertex3f(0.554044724f, 0.24938488f, 0.f),
    C_Vertex3f(0.554044724f, 0.24938488f, -1.f),
    C_Normal3f(0.812690079f, -0.582696199f, 0.f),
    C_Vertex3f(0.547559738f, 0.241938114f, 0.f),
    C_Vertex3f(0.547559738f, 0.241938114f, -1.f),
    C_Normal3f(0.891831219f, -0.452368289f, 0.f),
    C_Vertex3f(0.541704178f, 0.232173443f, 0.f),
    C_Vertex3f(0.541704178f, 0.232173443f, -1.f),
    C_Normal3f(0.931585968f, -0.363521039f, 0.f),
    C_Vertex3f(0.536478043f, 0.220090866f, 0.f),
    C_Vertex3f(0.536478043f, 0.220090866f, -1.f),
    C_Normal3f(0.952643692f, -0.304088891f, 0.f),
    C_Vertex3f(0.534179688f, 0.212890625f, 0.f),
    C_Vertex3f(0.534179688f, 0.212890625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.0988194942f, -0.995105445f, 0.f),
    C_Vertex3f(0.534179688f, 0.212890625f, 0.f),
    C_Vertex3f(0.534179688f, 0.212890625f, -1.f),
    C_Vertex3f(0.468261719f, 0.219436646f, -1.f),
    C_Vertex3f(0.468261719f, 0.219436646f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.968234956f, 0.250042021f, 0.f),
    C_Vertex3f(0.468261719f, 0.219436646f, 0.f),
    C_Vertex3f(0.468261719f, 0.219436646f, -1.f),
    C_Normal3f(-0.956868351f, 0.290522158f, 0.f),
    C_Vertex3f(0.469844818f, 0.225566864f, 0.f),
    C_Vertex3f(0.469844818f, 0.225566864f, -1.f),
    C_Normal3f(-0.938894093f, 0.344206095f, 0.f),
    C_Vertex3f(0.473598242f, 0.237013102f, 0.f),
    C_Vertex3f(0.473598242f, 0.237013102f, -1.f),
    C_Normal3f(-0.910816848f, 0.412810624f, 0.f),
    C_Vertex3f(0.47793889f, 0.24764514f, 0.f),
    C_Vertex3f(0.47793889f, 0.24764514f, -1.f),
    C_Normal3f(-0.874421477f, 0.485166967f, 0.f),
    C_Vertex3f(0.482866764f, 0.257462978f, 0.f),
    C_Vertex3f(0.482866764f, 0.257462978f, -1.f),
    C_Normal3f(-0.828578293f, 0.559873164f, 0.f),
    C_Vertex3f(0.488381863f, 0.266466618f, 0.f),
    C_Vertex3f(0.488381863f, 0.266466618f, -1.f),
    C_Normal3f(-0.772565603f, 0.634934962f, 0.f),
    C_Vertex3f(0.494484186f, 0.274656057f, 0.f),
    C_Vertex3f(0.494484186f, 0.274656057f, -1.f),
    C_Normal3f(-0.706345081f, 0.707867622f, 0.f),
    C_Vertex3f(0.501173735f, 0.282031298f, 0.f),
    C_Vertex3f(0.501173735f, 0.282031298f, -1.f),
    C_Normal3f(-0.630777955f, 0.775963366f, 0.f),
    C_Vertex3f(0.508450508f, 0.288592339f, 0.f),
    C_Vertex3f(0.508450508f, 0.288592339f, -1.f),
    C_Normal3f(-0.545378447f, 0.83818984f, 0.f),
    C_Vertex3f(0.516314507f, 0.29433918f, 0.f),
    C_Vertex3f(0.516314507f, 0.29433918f, -1.f),
    C_Normal3f(-0.456011921f, 0.889973581f, 0.f),
    C_Vertex3f(0.525011063f, 0.299367666f, 0.f),
    C_Vertex3f(0.525011063f, 0.299367666f, -1.f),
    C_Normal3f(-0.369198233f, 0.929350674f, 0.f),
    C_Vertex3f(0.534540176f, 0.303677797f, 0.f),
    C_Vertex3f(0.534540176f, 0.303677797f, -1.f),
    C_Normal3f(-0.287283719f, 0.957845509f, 0.f),
    C_Vertex3f(0.544901848f, 0.307269573f, 0.f),
    C_Vertex3f(0.544901848f, 0.307269573f, -1.f),
    C_Normal3f(-0.211643443f, 0.977346957f, 0.f),
    C_Vertex3f(0.556096077f, 0.310142994f, 0.f),
    C_Vertex3f(0.556096077f, 0.310142994f, -1.f),
    C_Normal3f(-0.142848253f, 0.989744544f, 0.f),
    C_Vertex3f(0.568122864f, 0.312298059f, 0.f),
    C_Vertex3f(0.568122864f, 0.312298059f, -1.f),
    C_Normal3f(-0.0809002146f, 0.996722162f, 0.f),
    C_Vertex3f(0.580982208f, 0.31373477f, 0.f),
    C_Vertex3f(0.580982208f, 0.31373477f, -1.f),
    C_Normal3f(-0.0245941021f, 0.999697566f, 0.f),
    C_Vertex3f(0.59467411f, 0.314453125f, 0.f),
    C_Vertex3f(0.59467411f, 0.314453125f, -1.f),
    C_Normal3f(0.0260747336f, 0.999660015f, 0.f),
    C_Vertex3f(0.610181808f, 0.314453125f, 0.f),
    C_Vertex3f(0.610181808f, 0.314453125f, -1.f),
    C_Normal3f(0.0811840594f, 0.996699095f, 0.f),
    C_Vertex3f(0.625573754f, 0.313647151f, 0.f),
    C_Vertex3f(0.625573754f, 0.313647151f, -1.f),
    C_Normal3f(0.145059749f, 0.989422917f, 0.f),
    C_Vertex3f(0.63986671f, 0.312035203f, 0.f),
    C_Vertex3f(0.63986671f, 0.312035203f, -1.f),
    C_Normal3f(0.217741475f, 0.976006508f, 0.f),
    C_Vertex3f(0.653060675f, 0.309617281f, 0.f),
    C_Vertex3f(0.653060675f, 0.309617281f, -1.f),
    C_Normal3f(0.299699038f, 0.954033852f, 0.f),
    C_Vertex3f(0.665155649f, 0.306393385f, 0.f),
    C_Vertex3f(0.665155649f, 0.306393385f, -1.f),
    C_Normal3f(0.390625507f, 0.920549631f, 0.f),
    C_Vertex3f(0.676151633f, 0.302363515f, 0.f),
    C_Vertex3f(0.676151633f, 0.302363515f, -1.f),
    C_Normal3f(0.488904029f, 0.872337639f, 0.f),
    C_Vertex3f(0.686048627f, 0.297527671f, 0.f),
    C_Vertex3f(0.686048627f, 0.297527671f, -1.f),
    C_Normal3f(0.591102839f, 0.80659616f, 0.f),
    C_Vertex3f(0.69484663f, 0.291885853f, 0.f),
    C_Vertex3f(0.69484663f, 0.291885853f, -1.f),
    C_Normal3f(0.67622447f, 0.736695647f, 0.f),
    C_Vertex3f(0.702545643f, 0.285438061f, 0.f),
    C_Vertex3f(0.702545643f, 0.285438061f, -1.f),
    C_Normal3f(0.744723797f, 0.667372704f, 0.f),
    C_Vertex3f(0.709282279f, 0.278635144f, 0.f),
    C_Vertex3f(0.709282279f, 0.278635144f, -1.f),
    C_Normal3f(0.810934365f, 0.585137069f, 0.f),
    C_Vertex3f(0.715056539f, 0.271477103f, 0.f),
    C_Vertex3f(0.715056539f, 0.271477103f, -1.f),
    C_Normal3f(0.871349812f, 0.490662277f, 0.f),
    C_Vertex3f(0.719868422f, 0.263963938f, 0.f),
    C_Vertex3f(0.719868422f, 0.263963938f, -1.f),
    C_Normal3f(0.922429264f, 0.386166096f, 0.f),
    C_Vertex3f(0.723717928f, 0.256095648f, 0.f),
    C_Vertex3f(0.723717928f, 0.256095648f, -1.f),
    C_Normal3f(0.961352587f, 0.275320143f, 0.f),
    C_Vertex3f(0.726605058f, 0.247872233f, 0.f),
    C_Vertex3f(0.726605058f, 0.247872233f, -1.f),
    C_Normal3f(0.986680925f, 0.162668049f, 0.f),
    C_Vertex3f(0.728529811f, 0.239293694f, 0.f),
    C_Vertex3f(0.728529811f, 0.239293694f, -1.f),
    C_Normal3f(0.999107063f, 0.0422494039f, 0.f),
    C_Vertex3f(0.729492188f, 0.230360031f, 0.f),
    C_Vertex3f(0.729492188f, 0.230360031f, -1.f),
    C_Normal3f(0.994867444f, -0.101187162f, 0.f),
    C_Vertex3f(0.729492188f, 0.216535568f, 0.f),
    C_Vertex3f(0.729492188f, 0.216535568f, -1.f),
    C_Normal3f(0.964358032f, -0.264600873f, 0.f),
    C_Vertex3f(0.726268768f, 0.198667526f, 0.f),
    C_Vertex3f(0.726268768f, 0.198667526f, -1.f),
    C_Normal3f(0.904871285f, -0.425685465f, 0.f),
    C_Vertex3f(0.71982193f, 0.18129158f, 0.f),
    C_Vertex3f(0.71982193f, 0.18129158f, -1.f),
    C_Normal3f(0.841263711f, -0.540625036f, 0.f),
    C_Vertex3f(0.710151672f, 0.16440773f, 0.f),
    C_Vertex3f(0.710151672f, 0.16440773f, -1.f),
    C_Normal3f(0.784215569f, -0.620488465f, 0.f),
    C_Vertex3f(0.703704834f, 0.156211853f, 0.f),
    C_Vertex3f(0.703704834f, 0.156211853f, -1.f),
    C_Normal3f(0.745725334f, -0.666253626f, 0.f),
    C_Vertex3f(0.701246262f, 0.153152466f, 0.f),
    C_Vertex3f(0.701246262f, 0.153152466f, -1.f),
    C_Normal3f(0.711750269f, -0.702432632f, 0.f),
    C_Vertex3f(0.694772243f, 0.14621377f, 0.f),
    C_Vertex3f(0.694772243f, 0.14621377f, -1.f),
    C_Normal3f(0.679936349f, -0.733271122f, 0.f),
    C_Vertex3f(0.686741352f, 0.138455153f, 0.f),
    C_Vertex3f(0.686741352f, 0.138455153f, -1.f),
    C_Normal3f(0.655116022f, -0.755528212f, 0.f),
    C_Vertex3f(0.677153587f, 0.129876614f, 0.f),
    C_Vertex3f(0.677153587f, 0.129876614f, -1.f),
    C_Normal3f(0.635298729f, -0.772266448f, 0.f),
    C_Vertex3f(0.666008949f, 0.120478153f, 0.f),
    C_Vertex3f(0.666008949f, 0.120478153f, -1.f),
    C_Normal3f(0.619151771f, -0.785271406f, 0.f),
    C_Vertex3f(0.653307438f, 0.110259771f, 0.f),
    C_Vertex3f(0.653307438f, 0.110259771f, -1.f),
    C_Normal3f(0.605763257f, -0.795644939f, 0.f),
    C_Vertex3f(0.639049053f, 0.099221468f, 0.f),
    C_Vertex3f(0.639049053f, 0.099221468f, -1.f),
    C_Normal3f(0.593926787f, -0.804519176f, 0.f),
    C_Vertex3f(0.623233795f, 0.0873632431f, 0.f),
    C_Vertex3f(0.623233795f, 0.0873632431f, -1.f),
    C_Normal3f(0.596301734f, -0.802760422f, 0.f),
    C_Vertex3f(0.605369568f, 0.0743579865f, 0.f),
    C_Vertex3f(0.605369568f, 0.0743579865f, -1.f),
    C_Normal3f(0.615544856f, -0.788101792f, 0.f),
    C_Vertex3f(0.589690208f, 0.0624465942f, 0.f),
    C_Vertex3f(0.589690208f, 0.0624465942f, -1.f),
    C_Normal3f(0.642507851f, -0.766279161f, 0.f),
    C_Vertex3f(0.576687813f, 0.0519561768f, 0.f),
    C_Vertex3f(0.576687813f, 0.0519561768f, -1.f),
    C_Normal3f(0.673540235f, -0.739150584f, 0.f),
    C_Vertex3f(0.566362381f, 0.042886734f, 0.f),
    C_Vertex3f(0.566362381f, 0.042886734f, -1.f),
    C_Normal3f(0.707106709f, -0.707106709f, 0.f),
    C_Vertex3f(0.562538147f, 0.0390625f, 0.f),
    C_Vertex3f(0.562538147f, 0.0390625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.562538147f, 0.0390625f, 0.f),
    C_Vertex3f(0.562538147f, 0.0390625f, -1.f),
    C_Vertex3f(0.730957031f, 0.0390625f, -1.f),
    C_Vertex3f(0.730957031f, 0.0390625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.730957031f, 0.0390625f, 0.f),
    C_Vertex3f(0.730957031f, 0.0390625f, -1.f),
    C_Vertex3f(0.730957031f, -0.013671875f, -1.f),
    C_Vertex3f(0.730957031f, -0.013671875f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.730957031f, -0.013671875f, 0.f),
    C_Vertex3f(0.730957031f, -0.013671875f, -1.f),
    C_Vertex3f(0.458984375f, -0.013671875f, -1.f),
    C_Vertex3f(0.458984375f, -0.013671875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.580917358f, 0.651367188f, 0.f),
    C_Vertex3f(0.0996627808f, -0.0249023438f, 0.f),
    C_Vertex3f(0.168281555f, -0.0249023438f, 0.f),
    C_Vertex3f(0.649551392f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.208007813f, 0.6484375f, 0.f),
    C_Vertex3f(0.160728455f, 0.6484375f, 0.f),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Vertex3f(0.155603409f, 0.641788483f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_Vertex3f(0.143771648f, 0.6289258f, 0.f),
    C_Vertex3f(0.133413315f, 0.553688049f, 0.f),
    C_Vertex3f(0.130358219f, 0.61649847f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133413315f, 0.553688049f, 0.f),
    C_Vertex3f(0.115363121f, 0.604506493f, 0.f),
    C_Vertex3f(0.110088348f, 0.538677216f, 0.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, 0.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0601711273f, 0.516677856f, 0.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, 0.f),
    C_Vertex3f(0.067923069f, 0.575233936f, 0.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, 0.f),
    C_Vertex3f(0.046875f, 0.566894531f, 0.f),
    C_Vertex3f(0.046875f, 0.513183594f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.562538147f, 0.0390625f, 0.f),
    C_Vertex3f(0.566362381f, 0.042886734f, 0.f),
    C_Vertex3f(0.57797718f, 0.124704838f, 0.f),
    C_Vertex3f(0.562220573f, 0.11335516f, 0.f),
    C_Vertex3f(0.547586679f, 0.102237344f, 0.f),
    C_Vertex3f(0.534075499f, 0.0913513899f, 0.f),
    C_Vertex3f(0.521687031f, 0.080697298f, 0.f),
    C_Vertex3f(0.510421276f, 0.0702750683f, 0.f),
    C_Vertex3f(0.500278234f, 0.0600847006f, 0.f),
    C_Vertex3f(0.491257906f, 0.050126195f, 0.f),
    C_Vertex3f(0.730957031f, -0.013671875f, 0.f),
    C_Vertex3f(0.730957031f, 0.0390625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.730957031f, -0.013671875f, 0.f),
    C_Vertex3f(0.491257906f, 0.050126195f, 0.f),
    C_Vertex3f(0.483360291f, 0.0403995514f, 0.f),
    C_Vertex3f(0.479972839f, 0.0356521606f, 0.f),
    C_Vertex3f(0.475381851f, 0.0294322968f, 0.f),
    C_Vertex3f(0.468167305f, 0.0170469284f, 0.f),
    C_Vertex3f(0.462920189f, 0.00471591949f, 0.f),
    C_Vertex3f(0.459640503f, -0.00756072998f, 0.f),
    C_Vertex3f(0.458984375f, -0.013671875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.659245014f, 0.229388237f, 0.f),
    C_Vertex3f(0.661132813f, 0.219867706f, 0.f),
    C_Vertex3f(0.665155649f, 0.306393385f, 0.f),
    C_Vertex3f(0.666008949f, 0.120478153f, 0.f),
    C_Vertex3f(0.676151633f, 0.302363515f, 0.f),
    C_Vertex3f(0.677153587f, 0.129876614f, 0.f),
    C_Vertex3f(0.686048627f, 0.297527671f, 0.f),
    C_Vertex3f(0.686741352f, 0.138455153f, 0.f),
    C_Vertex3f(0.69484663f, 0.291885853f, 0.f),
    C_Vertex3f(0.694772243f, 0.14621377f, 0.f),
    C_Vertex3f(0.701246262f, 0.153152466f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.642254829f, 0.252222061f, 0.f),
    C_Vertex3f(0.649806023f, 0.245565414f, 0.f),
    C_Vertex3f(0.653060675f, 0.309617281f, 0.f),
    C_Vertex3f(0.655469418f, 0.23795414f, 0.f),
    C_Vertex3f(0.665155649f, 0.306393385f, 0.f),
    C_Vertex3f(0.659245014f, 0.229388237f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.570106506f, 0.258360386f, 0.f),
    C_Vertex3f(0.580886841f, 0.260924816f, 0.f),
    C_Vertex3f(0.580982208f, 0.31373477f, 0.f),
    C_Vertex3f(0.593500137f, 0.262207031f, 0.f),
    C_Vertex3f(0.59467411f, 0.314453125f, 0.f),
    C_Vertex3f(0.60827446f, 0.262207031f, 0.f),
    C_Vertex3f(0.610181808f, 0.314453125f, 0.f),
    C_Vertex3f(0.621489048f, 0.26054287f, 0.f),
    C_Vertex3f(0.625573754f, 0.313647151f, 0.f),
    C_Vertex3f(0.632815838f, 0.257214546f, 0.f),
    C_Vertex3f(0.63986671f, 0.312035203f, 0.f),
    C_Vertex3f(0.642254829f, 0.252222061f, 0.f),
    C_Vertex3f(0.653060675f, 0.309617281f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.547559738f, 0.241938114f, 0.f),
    C_Vertex3f(0.554044724f, 0.24938488f, 0.f),
    C_Vertex3f(0.556096077f, 0.310142994f, 0.f),
    C_Vertex3f(0.561159134f, 0.254513741f, 0.f),
    C_Vertex3f(0.568122864f, 0.312298059f, 0.f),
    C_Vertex3f(0.570106506f, 0.258360386f, 0.f),
    C_Vertex3f(0.580982208f, 0.31373477f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.544901848f, 0.307269573f, 0.f),
    C_Vertex3f(0.534540176f, 0.303677797f, 0.f),
    C_Vertex3f(0.536478043f, 0.220090866f, 0.f),
    C_Vertex3f(0.541704178f, 0.232173443f, 0.f),
    C_Vertex3f(0.547559738f, 0.241938114f, 0.f),
    C_Vertex3f(0.556096077f, 0.310142994f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534179688f, 0.212890625f, 0.f),
    C_Vertex3f(0.536478043f, 0.220090866f, 0.f),
    C_Vertex3f(0.534540176f, 0.303677797f, 0.f),
    C_Vertex3f(0.525011063f, 0.299367666f, 0.f),
    C_Vertex3f(0.516314507f, 0.29433918f, 0.f),
    C_Vertex3f(0.508450508f, 0.288592339f, 0.f),
    C_Vertex3f(0.501173735f, 0.282031298f, 0.f),
    C_Vertex3f(0.494484186f, 0.274656057f, 0.f),
    C_Vertex3f(0.488381863f, 0.266466618f, 0.f),
    C_Vertex3f(0.482866764f, 0.257462978f, 0.f),
    C_Vertex3f(0.47793889f, 0.24764514f, 0.f),
    C_Vertex3f(0.473598242f, 0.237013102f, 0.f),
    C_Vertex3f(0.469844818f, 0.225566864f, 0.f),
    C_Vertex3f(0.468261719f, 0.219436646f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.729492188f, 0.230360031f, 0.f),
    C_Vertex3f(0.728529811f, 0.239293694f, 0.f),
    C_Vertex3f(0.729492188f, 0.216535568f, 0.f),
    C_Vertex3f(0.726605058f, 0.247872233f, 0.f),
    C_Vertex3f(0.726268768f, 0.198667526f, 0.f),
    C_Vertex3f(0.723717928f, 0.256095648f, 0.f),
    C_Vertex3f(0.719868422f, 0.263963938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.69484663f, 0.291885853f, 0.f),
    C_Vertex3f(0.701246262f, 0.153152466f, 0.f),
    C_Vertex3f(0.702545643f, 0.285438061f, 0.f),
    C_Vertex3f(0.703704834f, 0.156211853f, 0.f),
    C_Vertex3f(0.709282279f, 0.278635144f, 0.f),
    C_Vertex3f(0.710151672f, 0.16440773f, 0.f),
    C_Vertex3f(0.715056539f, 0.271477103f, 0.f),
    C_Vertex3f(0.71982193f, 0.18129158f, 0.f),
    C_Vertex3f(0.719868422f, 0.263963938f, 0.f),
    C_Vertex3f(0.726268768f, 0.198667526f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.666008949f, 0.120478153f, 0.f),
    C_Vertex3f(0.661132813f, 0.219867706f, 0.f),
    C_Vertex3f(0.661132813f, 0.209936142f, 0.f),
    C_Vertex3f(0.659108162f, 0.200657845f, 0.f),
    C_Vertex3f(0.655058861f, 0.19148922f, 0.f),
    C_Vertex3f(0.653307438f, 0.110259771f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.653307438f, 0.110259771f, 0.f),
    C_Vertex3f(0.655058861f, 0.19148922f, 0.f),
    C_Vertex3f(0.648984909f, 0.182430267f, 0.f),
    C_Vertex3f(0.644935608f, 0.177955627f, 0.f),
    C_Vertex3f(0.64345932f, 0.176264763f, 0.f),
    C_Vertex3f(0.639049053f, 0.099221468f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.639049053f, 0.099221468f, 0.f),
    C_Vertex3f(0.64345932f, 0.176264763f, 0.f),
    C_Vertex3f(0.63890624f, 0.17188704f, 0.f),
    C_Vertex3f(0.632752657f, 0.166513324f, 0.f),
    C_Vertex3f(0.624998569f, 0.160143614f, 0.f),
    C_Vertex3f(0.623233795f, 0.0873632431f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.615643978f, 0.15277791f, 0.f),
    C_Vertex3f(0.604688883f, 0.144416213f, 0.f),
    C_Vertex3f(0.605369568f, 0.0743579865f, 0.f),
    C_Vertex3f(0.623233795f, 0.0873632431f, 0.f),
    C_Vertex3f(0.624998569f, 0.160143614f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.566362381f, 0.042886734f, 0.f),
    C_Vertex3f(0.576687813f, 0.0519561768f, 0.f),
    C_Vertex3f(0.57797718f, 0.124704838f, 0.f),
    C_Vertex3f(0.589690208f, 0.0624465942f, 0.f),
    C_Vertex3f(0.592133284f, 0.135058522f, 0.f),
    C_Vertex3f(0.605369568f, 0.0743579865f, 0.f),
    C_Vertex3f(0.604688883f, 0.144416213f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.208007813f, 0.3203125f, 0.f),
    C_Vertex3f(0.14453125f, 0.562530518f, 0.f),
    C_Vertex3f(0.14453125f, 0.3203125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.580917358f, 0.651367188f, -1.f),
    C_Vertex3f(0.0996627808f, -0.0249023438f, -1.f),
    C_Vertex3f(0.168281555f, -0.0249023438f, -1.f),
    C_Vertex3f(0.649551392f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.208007813f, 0.6484375f, -1.f),
    C_Vertex3f(0.160728455f, 0.6484375f, -1.f),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.155603409f, 0.641788483f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Vertex3f(0.143771648f, 0.6289258f, -1.f),
    C_Vertex3f(0.133413315f, 0.553688049f, -1.f),
    C_Vertex3f(0.130358219f, 0.61649847f, -1.f),
    C_Vertex3f(0.115363121f, 0.604506493f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133413315f, 0.553688049f, -1.f),
    C_Vertex3f(0.115363121f, 0.604506493f, -1.f),
    C_Vertex3f(0.110088348f, 0.538677216f, -1.f),
    C_Vertex3f(0.0987863541f, 0.592949867f, -1.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, -1.f),
    C_Vertex3f(0.0829730034f, 0.583192348f, -1.f),
    C_Vertex3f(0.067923069f, 0.575233936f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0601711273f, 0.516677856f, -1.f),
    C_Vertex3f(0.0856742859f, 0.526340485f, -1.f),
    C_Vertex3f(0.067923069f, 0.575233936f, -1.f),
    C_Vertex3f(0.0536365509f, 0.569074631f, -1.f),
    C_Vertex3f(0.046875f, 0.566894531f, -1.f),
    C_Vertex3f(0.046875f, 0.513183594f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.562538147f, 0.0390625f, -1.f),
    C_Vertex3f(0.566362381f, 0.042886734f, -1.f),
    C_Vertex3f(0.57797718f, 0.124704838f, -1.f),
    C_Vertex3f(0.562220573f, 0.11335516f, -1.f),
    C_Vertex3f(0.547586679f, 0.102237344f, -1.f),
    C_Vertex3f(0.534075499f, 0.0913513899f, -1.f),
    C_Vertex3f(0.521687031f, 0.080697298f, -1.f),
    C_Vertex3f(0.510421276f, 0.0702750683f, -1.f),
    C_Vertex3f(0.500278234f, 0.0600847006f, -1.f),
    C_Vertex3f(0.491257906f, 0.050126195f, -1.f),
    C_Vertex3f(0.730957031f, -0.013671875f, -1.f),
    C_Vertex3f(0.730957031f, 0.0390625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.730957031f, -0.013671875f, -1.f),
    C_Vertex3f(0.491257906f, 0.050126195f, -1.f),
    C_Vertex3f(0.483360291f, 0.0403995514f, -1.f),
    C_Vertex3f(0.479972839f, 0.0356521606f, -1.f),
    C_Vertex3f(0.475381851f, 0.0294322968f, -1.f),
    C_Vertex3f(0.468167305f, 0.0170469284f, -1.f),
    C_Vertex3f(0.462920189f, 0.00471591949f, -1.f),
    C_Vertex3f(0.459640503f, -0.00756072998f, -1.f),
    C_Vertex3f(0.458984375f, -0.013671875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.659245014f, 0.229388237f, -1.f),
    C_Vertex3f(0.661132813f, 0.219867706f, -1.f),
    C_Vertex3f(0.665155649f, 0.306393385f, -1.f),
    C_Vertex3f(0.666008949f, 0.120478153f, -1.f),
    C_Vertex3f(0.676151633f, 0.302363515f, -1.f),
    C_Vertex3f(0.677153587f, 0.129876614f, -1.f),
    C_Vertex3f(0.686048627f, 0.297527671f, -1.f),
    C_Vertex3f(0.686741352f, 0.138455153f, -1.f),
    C_Vertex3f(0.69484663f, 0.291885853f, -1.f),
    C_Vertex3f(0.694772243f, 0.14621377f, -1.f),
    C_Vertex3f(0.701246262f, 0.153152466f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.642254829f, 0.252222061f, -1.f),
    C_Vertex3f(0.649806023f, 0.245565414f, -1.f),
    C_Vertex3f(0.653060675f, 0.309617281f, -1.f),
    C_Vertex3f(0.655469418f, 0.23795414f, -1.f),
    C_Vertex3f(0.665155649f, 0.306393385f, -1.f),
    C_Vertex3f(0.659245014f, 0.229388237f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.570106506f, 0.258360386f, -1.f),
    C_Vertex3f(0.580886841f, 0.260924816f, -1.f),
    C_Vertex3f(0.580982208f, 0.31373477f, -1.f),
    C_Vertex3f(0.593500137f, 0.262207031f, -1.f),
    C_Vertex3f(0.59467411f, 0.314453125f, -1.f),
    C_Vertex3f(0.60827446f, 0.262207031f, -1.f),
    C_Vertex3f(0.610181808f, 0.314453125f, -1.f),
    C_Vertex3f(0.621489048f, 0.26054287f, -1.f),
    C_Vertex3f(0.625573754f, 0.313647151f, -1.f),
    C_Vertex3f(0.632815838f, 0.257214546f, -1.f),
    C_Vertex3f(0.63986671f, 0.312035203f, -1.f),
    C_Vertex3f(0.642254829f, 0.252222061f, -1.f),
    C_Vertex3f(0.653060675f, 0.309617281f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.547559738f, 0.241938114f, -1.f),
    C_Vertex3f(0.554044724f, 0.24938488f, -1.f),
    C_Vertex3f(0.556096077f, 0.310142994f, -1.f),
    C_Vertex3f(0.561159134f, 0.254513741f, -1.f),
    C_Vertex3f(0.568122864f, 0.312298059f, -1.f),
    C_Vertex3f(0.570106506f, 0.258360386f, -1.f),
    C_Vertex3f(0.580982208f, 0.31373477f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.544901848f, 0.307269573f, -1.f),
    C_Vertex3f(0.534540176f, 0.303677797f, -1.f),
    C_Vertex3f(0.536478043f, 0.220090866f, -1.f),
    C_Vertex3f(0.541704178f, 0.232173443f, -1.f),
    C_Vertex3f(0.547559738f, 0.241938114f, -1.f),
    C_Vertex3f(0.556096077f, 0.310142994f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534179688f, 0.212890625f, -1.f),
    C_Vertex3f(0.536478043f, 0.220090866f, -1.f),
    C_Vertex3f(0.534540176f, 0.303677797f, -1.f),
    C_Vertex3f(0.525011063f, 0.299367666f, -1.f),
    C_Vertex3f(0.516314507f, 0.29433918f, -1.f),
    C_Vertex3f(0.508450508f, 0.288592339f, -1.f),
    C_Vertex3f(0.501173735f, 0.282031298f, -1.f),
    C_Vertex3f(0.494484186f, 0.274656057f, -1.f),
    C_Vertex3f(0.488381863f, 0.266466618f, -1.f),
    C_Vertex3f(0.482866764f, 0.257462978f, -1.f),
    C_Vertex3f(0.47793889f, 0.24764514f, -1.f),
    C_Vertex3f(0.473598242f, 0.237013102f, -1.f),
    C_Vertex3f(0.469844818f, 0.225566864f, -1.f),
    C_Vertex3f(0.468261719f, 0.219436646f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.729492188f, 0.230360031f, -1.f),
    C_Vertex3f(0.728529811f, 0.239293694f, -1.f),
    C_Vertex3f(0.729492188f, 0.216535568f, -1.f),
    C_Vertex3f(0.726605058f, 0.247872233f, -1.f),
    C_Vertex3f(0.726268768f, 0.198667526f, -1.f),
    C_Vertex3f(0.723717928f, 0.256095648f, -1.f),
    C_Vertex3f(0.719868422f, 0.263963938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.69484663f, 0.291885853f, -1.f),
    C_Vertex3f(0.701246262f, 0.153152466f, -1.f),
    C_Vertex3f(0.702545643f, 0.285438061f, -1.f),
    C_Vertex3f(0.703704834f, 0.156211853f, -1.f),
    C_Vertex3f(0.709282279f, 0.278635144f, -1.f),
    C_Vertex3f(0.710151672f, 0.16440773f, -1.f),
    C_Vertex3f(0.715056539f, 0.271477103f, -1.f),
    C_Vertex3f(0.71982193f, 0.18129158f, -1.f),
    C_Vertex3f(0.719868422f, 0.263963938f, -1.f),
    C_Vertex3f(0.726268768f, 0.198667526f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.666008949f, 0.120478153f, -1.f),
    C_Vertex3f(0.661132813f, 0.219867706f, -1.f),
    C_Vertex3f(0.661132813f, 0.209936142f, -1.f),
    C_Vertex3f(0.659108162f, 0.200657845f, -1.f),
    C_Vertex3f(0.655058861f, 0.19148922f, -1.f),
    C_Vertex3f(0.653307438f, 0.110259771f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.653307438f, 0.110259771f, -1.f),
    C_Vertex3f(0.655058861f, 0.19148922f, -1.f),
    C_Vertex3f(0.648984909f, 0.182430267f, -1.f),
    C_Vertex3f(0.644935608f, 0.177955627f, -1.f),
    C_Vertex3f(0.64345932f, 0.176264763f, -1.f),
    C_Vertex3f(0.639049053f, 0.099221468f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.639049053f, 0.099221468f, -1.f),
    C_Vertex3f(0.64345932f, 0.176264763f, -1.f),
    C_Vertex3f(0.63890624f, 0.17188704f, -1.f),
    C_Vertex3f(0.632752657f, 0.166513324f, -1.f),
    C_Vertex3f(0.624998569f, 0.160143614f, -1.f),
    C_Vertex3f(0.623233795f, 0.0873632431f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.615643978f, 0.15277791f, -1.f),
    C_Vertex3f(0.604688883f, 0.144416213f, -1.f),
    C_Vertex3f(0.605369568f, 0.0743579865f, -1.f),
    C_Vertex3f(0.623233795f, 0.0873632431f, -1.f),
    C_Vertex3f(0.624998569f, 0.160143614f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.566362381f, 0.042886734f, -1.f),
    C_Vertex3f(0.576687813f, 0.0519561768f, -1.f),
    C_Vertex3f(0.57797718f, 0.124704838f, -1.f),
    C_Vertex3f(0.589690208f, 0.0624465942f, -1.f),
    C_Vertex3f(0.592133284f, 0.135058522f, -1.f),
    C_Vertex3f(0.605369568f, 0.0743579865f, -1.f),
    C_Vertex3f(0.604688883f, 0.144416213f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.208007813f, 0.3203125f, -1.f),
    C_Vertex3f(0.14453125f, 0.562530518f, -1.f),
    C_Vertex3f(0.14453125f, 0.3203125f, -1.f),
    C_End(),
    C_Translatef(0.746582031f, 0.f, 0.f),
};
static const Call char_190[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.814755082f, 0.579805255f, 0.f),
    C_Vertex3f(0.110153198f, -0.0249023438f, 0.f),
    C_Vertex3f(0.110153198f, -0.0249023438f, -1.f),
    C_Vertex3f(0.591407776f, 0.651367188f, -1.f),
    C_Vertex3f(0.591407776f, 0.651367188f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.591407776f, 0.651367188f, 0.f),
    C_Vertex3f(0.591407776f, 0.651367188f, -1.f),
    C_Vertex3f(0.660041809f, 0.651367188f, -1.f),
    C_Vertex3f(0.660041809f, 0.651367188f, 0.f),
    C_Normal3f(0.81474638f, -0.579817414f, 0.f),
    C_Vertex3f(0.660041809f, 0.651367188f, 0.f),
    C_Vertex3f(0.660041809f, 0.651367188f, -1.f),
    C_Vertex3f(0.178771973f, -0.0249023438f, -1.f),
    C_Vertex3f(0.178771973f, -0.0249023438f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.178771973f, -0.0249023438f, 0.f),
    C_Vertex3f(0.178771973f, -0.0249023438f, -1.f),
    C_Vertex3f(0.110153198f, -0.0249023438f, -1.f),
    C_Vertex3f(0.110153198f, -0.0249023438f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.102158323f, 0.994768143f, 0.f),
    C_Vertex3f(0.0146484375f, 0.40625f, 0.f),
    C_Vertex3f(0.0146484375f, 0.40625f, -1.f),
    C_Vertex3f(0.0786132813f, 0.412818909f, -1.f),
    C_Vertex3f(0.0786132813f, 0.412818909f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.949013472f, 0.315235585f, 0.f),
    C_Vertex3f(0.0786132813f, 0.412818909f, 0.f),
    C_Vertex3f(0.0786132813f, 0.412818909f, -1.f),
    C_Normal3f(0.925726354f, 0.378194004f, 0.f),
    C_Vertex3f(0.0807952881f, 0.40625f, 0.f),
    C_Vertex3f(0.0807952881f, 0.40625f, -1.f),
    C_Normal3f(0.882209897f, 0.470856279f, 0.f),
    C_Vertex3f(0.085814476f, 0.395192146f, 0.f),
    C_Vertex3f(0.085814476f, 0.395192146f, -1.f),
    C_Normal3f(0.797653735f, 0.603115737f, 0.f),
    C_Vertex3f(0.0914888382f, 0.386214256f, 0.f),
    C_Vertex3f(0.0914888382f, 0.386214256f, -1.f),
    C_Normal3f(0.655304611f, 0.755364716f, 0.f),
    C_Vertex3f(0.0978183746f, 0.37931633f, 0.f),
    C_Vertex3f(0.0978183746f, 0.37931633f, -1.f),
    C_Normal3f(0.499051124f, 0.866572499f, 0.f),
    C_Vertex3f(0.106002808f, 0.373622894f, 0.f),
    C_Vertex3f(0.106002808f, 0.373622894f, -1.f),
    C_Normal3f(0.358435005f, 0.93355459f, 0.f),
    C_Vertex3f(0.116259575f, 0.368696213f, 0.f),
    C_Vertex3f(0.116259575f, 0.368696213f, -1.f),
    C_Normal3f(0.208314478f, 0.978061914f, 0.f),
    C_Vertex3f(0.127388954f, 0.365411758f, 0.f),
    C_Vertex3f(0.127388954f, 0.365411758f, -1.f),
    C_Normal3f(0.062588729f, 0.998039484f, 0.f),
    C_Vertex3f(0.139390945f, 0.363769531f, 0.f),
    C_Vertex3f(0.139390945f, 0.363769531f, -1.f),
    C_Normal3f(-0.0755289048f, 0.997143626f, 0.f),
    C_Vertex3f(0.153575897f, 0.363769531f, 0.f),
    C_Vertex3f(0.153575897f, 0.363769531f, -1.f),
    C_Normal3f(-0.235443294f, 0.971888006f, 0.f),
    C_Vertex3f(0.167543411f, 0.365901947f, 0.f),
    C_Vertex3f(0.167543411f, 0.365901947f, -1.f),
    C_Normal3f(-0.415341794f, 0.909665406f, 0.f),
    C_Vertex3f(0.179983139f, 0.370166779f, 0.f),
    C_Vertex3f(0.179983139f, 0.370166779f, -1.f),
    C_Normal3f(-0.592475772f, 0.805588365f, 0.f),
    C_Vertex3f(0.190895081f, 0.376564026f, 0.f),
    C_Vertex3f(0.190895081f, 0.376564026f, -1.f),
    C_Normal3f(-0.737504065f, 0.675342739f, 0.f),
    C_Vertex3f(0.200279236f, 0.385093689f, 0.f),
    C_Vertex3f(0.200279236f, 0.385093689f, -1.f),
    C_Normal3f(-0.859020531f, 0.511941075f, 0.f),
    C_Vertex3f(0.207317352f, 0.394497871f, 0.f),
    C_Vertex3f(0.207317352f, 0.394497871f, -1.f),
    C_Normal3f(-0.950081706f, 0.312001199f, 0.f),
    C_Vertex3f(0.21200943f, 0.404776573f, 0.f),
    C_Vertex3f(0.21200943f, 0.404776573f, -1.f),
    C_Normal3f(-0.994671762f, 0.103092164f, 0.f),
    C_Vertex3f(0.214355469f, 0.415929794f, 0.f),
    C_Vertex3f(0.214355469f, 0.415929794f, -1.f),
    C_Normal3f(-0.99584192f, -0.0910978988f, 0.f),
    C_Vertex3f(0.214355469f, 0.427412033f, 0.f),
    C_Vertex3f(0.214355469f, 0.427412033f, -1.f),
    C_Normal3f(-0.953854024f, -0.300270617f, 0.f),
    C_Vertex3f(0.212404728f, 0.437254429f, 0.f),
    C_Vertex3f(0.212404728f, 0.437254429f, -1.f),
    C_Normal3f(-0.859505892f, -0.511125743f, 0.f),
    C_Vertex3f(0.208503246f, 0.446002483f, 0.f),
    C_Vertex3f(0.208503246f, 0.446002483f, -1.f),
    C_Normal3f(-0.721117973f, -0.692812264f, 0.f),
    C_Vertex3f(0.202651024f, 0.453656197f, 0.f),
    C_Vertex3f(0.202651024f, 0.453656197f, -1.f),
    C_Normal3f(-0.53331846f, -0.845914483f, 0.f),
    C_Vertex3f(0.194848061f, 0.460215569f, 0.f),
    C_Vertex3f(0.194848061f, 0.460215569f, -1.f),
    C_Normal3f(-0.330563635f, -0.9437837f, 0.f),
    C_Vertex3f(0.184443951f, 0.465135098f, 0.f),
    C_Vertex3f(0.184443951f, 0.465135098f, -1.f),
    C_Normal3f(-0.169454783f, -0.985538006f, 0.f),
    C_Vertex3f(0.171438694f, 0.468414783f, 0.f),
    C_Vertex3f(0.171438694f, 0.468414783f, -1.f),
    C_Normal3f(-0.0662174001f, -0.997805119f, 0.f),
    C_Vertex3f(0.155832291f, 0.470054626f, 0.f),
    C_Vertex3f(0.155832291f, 0.470054626f, -1.f),
    C_Normal3f(0.0224529654f, -0.999747872f, 0.f),
    C_Vertex3f(0.146728516f, 0.470054626f, 0.f),
    C_Vertex3f(0.146728516f, 0.470054626f, -1.f),
    C_Normal3f(0.058223024f, -0.998303592f, 0.f),
    C_Vertex3f(0.141309738f, 0.46972847f, 0.f),
    C_Vertex3f(0.141309738f, 0.46972847f, -1.f),
    C_Normal3f(0.054830689f, -0.998495698f, 0.f),
    C_Vertex3f(0.13274765f, 0.469239235f, 0.f),
    C_Vertex3f(0.13274765f, 0.469239235f, -1.f),
    C_Normal3f(0.0474552698f, -0.998873293f, 0.f),
    C_Vertex3f(0.126461029f, 0.468913078f, 0.f),
    C_Vertex3f(0.126461029f, 0.468913078f, -1.f),
    C_Normal3f(0.0334058963f, -0.999441862f, 0.f),
    C_Vertex3f(0.122449875f, 0.46875f, 0.f),
    C_Vertex3f(0.122449875f, 0.46875f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.121582031f, 0.46875f, 0.f),
    C_Vertex3f(0.121582031f, 0.46875f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.980853558f, 0.194746494f, 0.f),
    C_Vertex3f(0.121582031f, 0.46875f, 0.f),
    C_Vertex3f(0.121582031f, 0.46875f, -1.f),
    C_Vertex3f(0.13117981f, 0.517089844f, -1.f),
    C_Vertex3f(0.13117981f, 0.517089844f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.012307425f, 0.999924302f, 0.f),
    C_Vertex3f(0.13117981f, 0.517089844f, 0.f),
    C_Vertex3f(0.13117981f, 0.517089844f, -1.f),
    C_Normal3f(-0.0561989658f, 0.998419583f, 0.f),
    C_Vertex3f(0.140012741f, 0.516981125f, 0.f),
    C_Vertex3f(0.140012741f, 0.516981125f, -1.f),
    C_Normal3f(-0.161512494f, 0.986870706f, 0.f),
    C_Vertex3f(0.155306339f, 0.518447876f, 0.f),
    C_Vertex3f(0.155306339f, 0.518447876f, -1.f),
    C_Normal3f(-0.322124451f, 0.946697354f, 0.f),
    C_Vertex3f(0.168227673f, 0.521598816f, 0.f),
    C_Vertex3f(0.168227673f, 0.521598816f, -1.f),
    C_Normal3f(-0.518483162f, 0.855087817f, 0.f),
    C_Vertex3f(0.178776741f, 0.526433945f, 0.f),
    C_Vertex3f(0.178776741f, 0.526433945f, -1.f),
    C_Normal3f(-0.69272387f, 0.72120285f, 0.f),
    C_Vertex3f(0.186953545f, 0.532953262f, 0.f),
    C_Vertex3f(0.186953545f, 0.532953262f, -1.f),
    C_Normal3f(-0.82908088f, 0.559128761f, 0.f),
    C_Vertex3f(0.193086147f, 0.540178299f, 0.f),
    C_Vertex3f(0.193086147f, 0.540178299f, -1.f),
    C_Normal3f(-0.937811553f, 0.347144663f, 0.f),
    C_Vertex3f(0.197174549f, 0.548109055f, 0.f),
    C_Vertex3f(0.197174549f, 0.548109055f, -1.f),
    C_Normal3f(-0.993113458f, 0.117156781f, 0.f),
    C_Vertex3f(0.19921875f, 0.556745529f, 0.f),
    C_Vertex3f(0.19921875f, 0.556745529f, -1.f),
    C_Normal3f(-0.994821727f, -0.101634219f, 0.f),
    C_Vertex3f(0.19921875f, 0.565437317f, 0.f),
    C_Vertex3f(0.19921875f, 0.565437317f, -1.f),
    C_Normal3f(-0.945004106f, -0.327058583f, 0.f),
    C_Vertex3f(0.197570324f, 0.572880745f, 0.f),
    C_Vertex3f(0.197570324f, 0.572880745f, -1.f),
    C_Normal3f(-0.846279621f, -0.532738864f, 0.f),
    C_Vertex3f(0.194273472f, 0.579726219f, 0.f),
    C_Vertex3f(0.194273472f, 0.579726219f, -1.f),
    C_Normal3f(-0.717827618f, -0.696220875f, 0.f),
    C_Vertex3f(0.189328194f, 0.58597374f, 0.f),
    C_Vertex3f(0.189328194f, 0.58597374f, -1.f),
    C_Normal3f(-0.555961907f, -0.831207812f, 0.f),
    C_Vertex3f(0.182734489f, 0.591623306f, 0.f),
    C_Vertex3f(0.182734489f, 0.591623306f, -1.f),
    C_Normal3f(-0.365763187f, -0.930707991f, 0.f),
    C_Vertex3f(0.174546719f, 0.595860481f, 0.f),
    C_Vertex3f(0.174546719f, 0.595860481f, -1.f),
    C_Normal3f(-0.196366847f, -0.980530441f, 0.f),
    C_Vertex3f(0.164764881f, 0.598685265f, 0.f),
    C_Vertex3f(0.164764881f, 0.598685265f, -1.f),
    C_Normal3f(-0.0587092414f, -0.998275161f, 0.f),
    C_Vertex3f(0.153388977f, 0.600097656f, 0.f),
    C_Vertex3f(0.153388977f, 0.600097656f, -1.f),
    C_Normal3f(0.0512446761f, -0.998686135f, 0.f),
    C_Vertex3f(0.140748978f, 0.600097656f, 0.f),
    C_Vertex3f(0.140748978f, 0.600097656f, -1.f),
    C_Normal3f(0.173851714f, -0.984771848f, 0.f),
    C_Vertex3f(0.129785061f, 0.59888649f, 0.f),
    C_Vertex3f(0.129785061f, 0.59888649f, -1.f),
    C_Normal3f(0.320640087f, -0.947201133f, 0.f),
    C_Vertex3f(0.120167255f, 0.596464157f, 0.f),
    C_Vertex3f(0.120167255f, 0.596464157f, -1.f),
    C_Normal3f(0.487189293f, -0.87329638f, 0.f),
    C_Vertex3f(0.111895561f, 0.592830658f, 0.f),
    C_Vertex3f(0.111895561f, 0.592830658f, -1.f),
    C_Normal3f(0.657247066f, -0.753675222f, 0.f),
    C_Vertex3f(0.104969978f, 0.587985992f, 0.f),
    C_Vertex3f(0.104969978f, 0.587985992f, -1.f),
    C_Normal3f(0.782326162f, -0.622868896f, 0.f),
    C_Vertex3f(0.0985121727f, 0.581159592f, 0.f),
    C_Vertex3f(0.0985121727f, 0.581159592f, -1.f),
    C_Normal3f(0.862240911f, -0.506498337f, 0.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, 0.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, -1.f),
    C_Normal3f(0.905491531f, -0.424364328f, 0.f),
    C_Vertex3f(0.0869998932f, 0.561561584f, 0.f),
    C_Vertex3f(0.0869998932f, 0.561561584f, -1.f),
    C_Normal3f(0.929830372f, -0.367988437f, 0.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, 0.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.161898106f, -0.986807525f, 0.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, 0.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, -1.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, -1.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.932396173f, 0.361437947f, 0.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, 0.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, -1.f),
    C_Normal3f(-0.905484498f, 0.424379319f, 0.f),
    C_Vertex3f(0.0264606476f, 0.577007294f, 0.f),
    C_Vertex3f(0.0264606476f, 0.577007294f, -1.f),
    C_Normal3f(-0.857638955f, 0.514252305f, 0.f),
    C_Vertex3f(0.0366129875f, 0.59666729f, 0.f),
    C_Vertex3f(0.0366129875f, 0.59666729f, -1.f),
    C_Normal3f(-0.770776927f, 0.637105107f, 0.f),
    C_Vertex3f(0.0479416847f, 0.612832069f, 0.f),
    C_Vertex3f(0.0479416847f, 0.612832069f, -1.f),
    C_Normal3f(-0.640562534f, 0.767905951f, 0.f),
    C_Vertex3f(0.0604467392f, 0.625501633f, 0.f),
    C_Vertex3f(0.0604467392f, 0.625501633f, -1.f),
    C_Normal3f(-0.465689421f, 0.884948194f, 0.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, 0.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, -1.f),
    C_Normal3f(-0.297496587f, 0.954722881f, 0.f),
    C_Vertex3f(0.0909562111f, 0.64155674f, 0.f),
    C_Vertex3f(0.0909562111f, 0.64155674f, -1.f),
    C_Normal3f(-0.157664075f, 0.9874928f, 0.f),
    C_Vertex3f(0.11093092f, 0.646143913f, 0.f),
    C_Vertex3f(0.11093092f, 0.646143913f, -1.f),
    C_Normal3f(-0.0518968254f, 0.998652458f, 0.f),
    C_Vertex3f(0.134052277f, 0.6484375f, 0.f),
    C_Vertex3f(0.134052277f, 0.6484375f, -1.f),
    C_Normal3f(0.0203911792f, 0.999792099f, 0.f),
    C_Vertex3f(0.15506649f, 0.6484375f, 0.f),
    C_Vertex3f(0.15506649f, 0.6484375f, -1.f),
    C_Normal3f(0.076960817f, 0.997034073f, 0.f),
    C_Vertex3f(0.169752955f, 0.64770937f, 0.f),
    C_Vertex3f(0.169752955f, 0.64770937f, -1.f),
    C_Normal3f(0.137886435f, 0.990448117f, 0.f),
    C_Vertex3f(0.183365464f, 0.646253109f, 0.f),
    C_Vertex3f(0.183365464f, 0.646253109f, -1.f),
    C_Normal3f(0.207712278f, 0.978190005f, 0.f),
    C_Vertex3f(0.195904016f, 0.644068718f, 0.f),
    C_Vertex3f(0.195904016f, 0.644068718f, -1.f),
    C_Normal3f(0.287211269f, 0.957867205f, 0.f),
    C_Vertex3f(0.207368612f, 0.641156197f, 0.f),
    C_Vertex3f(0.207368612f, 0.641156197f, -1.f),
    C_Normal3f(0.376511544f, 0.926411867f, 0.f),
    C_Vertex3f(0.217759252f, 0.637515545f, 0.f),
    C_Vertex3f(0.217759252f, 0.637515545f, -1.f),
    C_Normal3f(0.474512786f, 0.880248606f, 0.f),
    C_Vertex3f(0.227075934f, 0.633146763f, 0.f),
    C_Vertex3f(0.227075934f, 0.633146763f, -1.f),
    C_Normal3f(0.58870846f, 0.808345437f, 0.f),
    C_Vertex3f(0.235318661f, 0.62804985f, 0.f),
    C_Vertex3f(0.235318661f, 0.62804985f, -1.f),
    C_Normal3f(0.712377429f, 0.701796532f, 0.f),
    C_Vertex3f(0.246071815f, 0.619312286f, 0.f),
    C_Vertex3f(0.246071815f, 0.619312286f, -1.f),
    C_Normal3f(0.838954687f, 0.544201314f, 0.f),
    C_Vertex3f(0.25682497f, 0.606219292f, 0.f),
    C_Vertex3f(0.25682497f, 0.606219292f, -1.f),
    C_Normal3f(0.943152666f, 0.33235985f, 0.f),
    C_Vertex3f(0.26399374f, 0.591683388f, 0.f),
    C_Vertex3f(0.26399374f, 0.591683388f, -1.f),
    C_Normal3f(0.993242919f, 0.116053827f, 0.f),
    C_Vertex3f(0.267578125f, 0.575704575f, 0.f),
    C_Vertex3f(0.267578125f, 0.575704575f, -1.f),
    C_Normal3f(0.997228563f, -0.0743987635f, 0.f),
    C_Vertex3f(0.267578125f, 0.561006546f, 0.f),
    C_Vertex3f(0.267578125f, 0.561006546f, -1.f),
    C_Normal3f(0.966676414f, -0.256001443f, 0.f),
    C_Vertex3f(0.265636921f, 0.549685001f, 0.f),
    C_Vertex3f(0.265636921f, 0.549685001f, -1.f),
    C_Normal3f(0.90528971f, -0.424794763f, 0.f),
    C_Vertex3f(0.261754513f, 0.539016247f, 0.f),
    C_Vertex3f(0.261754513f, 0.539016247f, -1.f),
    C_Normal3f(0.818774521f, -0.574115276f, 0.f),
    C_Vertex3f(0.255930901f, 0.529000282f, 0.f),
    C_Vertex3f(0.255930901f, 0.529000282f, -1.f),
    C_Normal3f(0.710521519f, -0.703675508f, 0.f),
    C_Vertex3f(0.248166084f, 0.519637108f, 0.f),
    C_Vertex3f(0.248166084f, 0.519637108f, -1.f),
    C_Normal3f(0.581858337f, -0.813290119f, 0.f),
    C_Vertex3f(0.238678455f, 0.51157999f, 0.f),
    C_Vertex3f(0.238678455f, 0.51157999f, -1.f),
    C_Normal3f(0.450883567f, -0.892582715f, 0.f),
    C_Vertex3f(0.227468014f, 0.50482893f, 0.f),
    C_Vertex3f(0.227468014f, 0.50482893f, -1.f),
    C_Normal3f(0.347733289f, -0.93759346f, 0.f),
    C_Vertex3f(0.21453476f, 0.499383926f, 0.f),
    C_Vertex3f(0.21453476f, 0.499383926f, -1.f),
    C_Normal3f(0.271775454f, -0.96236068f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.240741223f, 0.97058934f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, -1.f),
    C_Normal3f(0.313972771f, 0.949431956f, 0.f),
    C_Vertex3f(0.216918945f, 0.494905472f, 0.f),
    C_Vertex3f(0.216918945f, 0.494905472f, -1.f),
    C_Normal3f(0.417560071f, 0.908649325f, 0.f),
    C_Vertex3f(0.233778477f, 0.488527298f, 0.f),
    C_Vertex3f(0.233778477f, 0.488527298f, -1.f),
    C_Normal3f(0.556630671f, 0.830760002f, 0.f),
    C_Vertex3f(0.248073101f, 0.480588913f, 0.f),
    C_Vertex3f(0.248073101f, 0.480588913f, -1.f),
    C_Normal3f(0.701332808f, 0.71283406f, 0.f),
    C_Vertex3f(0.259802818f, 0.471090317f, 0.f),
    C_Vertex3f(0.259802818f, 0.471090317f, -1.f),
    C_Normal3f(0.821135819f, 0.570732832f, 0.f),
    C_Vertex3f(0.268967628f, 0.460031509f, 0.f),
    C_Vertex3f(0.268967628f, 0.460031509f, -1.f),
    C_Normal3f(0.909034133f, 0.416721582f, 0.f),
    C_Vertex3f(0.275841236f, 0.448015213f, 0.f),
    C_Vertex3f(0.275841236f, 0.448015213f, -1.f),
    C_Normal3f(0.968881249f, 0.247526288f, 0.f),
    C_Vertex3f(0.280423641f, 0.435041428f, 0.f),
    C_Vertex3f(0.280423641f, 0.435041428f, -1.f),
    C_Normal3f(0.996273696f, 0.0862477496f, 0.f),
    C_Vertex3f(0.282714844f, 0.421110153f, 0.f),
    C_Vertex3f(0.282714844f, 0.421110153f, -1.f),
    C_Normal3f(0.998839438f, -0.04816401f, 0.f),
    C_Vertex3f(0.282714844f, 0.408575058f, 0.f),
    C_Vertex3f(0.282714844f, 0.408575058f, -1.f),
    C_Normal3f(0.986319363f, -0.164845675f, 0.f),
    C_Vertex3f(0.281635284f, 0.398721933f, 0.f),
    C_Vertex3f(0.281635284f, 0.398721933f, -1.f),
    C_Normal3f(0.960858941f, -0.277037889f, 0.f),
    C_Vertex3f(0.279476166f, 0.389197111f, 0.f),
    C_Vertex3f(0.279476166f, 0.389197111f, -1.f),
    C_Normal3f(0.922533095f, -0.385917842f, 0.f),
    C_Vertex3f(0.276237488f, 0.380000591f, 0.f),
    C_Vertex3f(0.276237488f, 0.380000591f, -1.f),
    C_Normal3f(0.873200536f, -0.487361073f, 0.f),
    C_Vertex3f(0.27191925f, 0.371132374f, 0.f),
    C_Vertex3f(0.27191925f, 0.371132374f, -1.f),
    C_Normal3f(0.815806508f, -0.578324974f, 0.f),
    C_Vertex3f(0.266521454f, 0.362592459f, 0.f),
    C_Vertex3f(0.266521454f, 0.362592459f, -1.f),
    C_Normal3f(0.753707826f, -0.657209754f, 0.f),
    C_Vertex3f(0.260044098f, 0.354380846f, 0.f),
    C_Vertex3f(0.260044098f, 0.354380846f, -1.f),
    C_Normal3f(0.690031052f, -0.723779738f, 0.f),
    C_Vertex3f(0.252487183f, 0.346497536f, 0.f),
    C_Vertex3f(0.252487183f, 0.346497536f, -1.f),
    C_Normal3f(0.614450276f, -0.788955569f, 0.f),
    C_Vertex3f(0.243850708f, 0.338942528f, 0.f),
    C_Vertex3f(0.243850708f, 0.338942528f, -1.f),
    C_Normal3f(0.52276057f, -0.852479517f, 0.f),
    C_Vertex3f(0.234298468f, 0.332331896f, 0.f),
    C_Vertex3f(0.234298468f, 0.332331896f, -1.f),
    C_Normal3f(0.429344833f, -0.903140724f, 0.f),
    C_Vertex3f(0.223830462f, 0.32666564f, 0.f),
    C_Vertex3f(0.223830462f, 0.32666564f, -1.f),
    C_Normal3f(0.337783158f, -0.941223919f, 0.f),
    C_Vertex3f(0.21244669f, 0.32194376f, 0.f),
    C_Vertex3f(0.21244669f, 0.32194376f, -1.f),
    C_Normal3f(0.250808328f, -0.968036711f, 0.f),
    C_Vertex3f(0.200147152f, 0.318166256f, 0.f),
    C_Vertex3f(0.200147152f, 0.318166256f, -1.f),
    C_Normal3f(0.170151472f, -0.985417962f, 0.f),
    C_Vertex3f(0.186931849f, 0.315333128f, 0.f),
    C_Vertex3f(0.186931849f, 0.315333128f, -1.f),
    C_Normal3f(0.0966438875f, -0.995319009f, 0.f),
    C_Vertex3f(0.172800779f, 0.313444376f, 0.f),
    C_Vertex3f(0.172800779f, 0.313444376f, -1.f),
    C_Normal3f(0.0307664964f, -0.99952662f, 0.f),
    C_Vertex3f(0.157753944f, 0.3125f, 0.f),
    C_Vertex3f(0.157753944f, 0.3125f, -1.f),
    C_Normal3f(-0.0252892748f, -0.999680161f, 0.f),
    C_Vertex3f(0.142120361f, 0.3125f, 0.f),
    C_Vertex3f(0.142120361f, 0.3125f, -1.f),
    C_Normal3f(-0.0814212263f, -0.996679723f, 0.f),
    C_Vertex3f(0.127717614f, 0.31325984f, 0.f),
    C_Vertex3f(0.127717614f, 0.31325984f, -1.f),
    C_Normal3f(-0.144044951f, -0.989571154f, 0.f),
    C_Vertex3f(0.114216685f, 0.31477952f, 0.f),
    C_Vertex3f(0.114216685f, 0.31477952f, -1.f),
    C_Normal3f(-0.213851959f, -0.976866007f, 0.f),
    C_Vertex3f(0.101617575f, 0.31705904f, 0.f),
    C_Vertex3f(0.101617575f, 0.31705904f, -1.f),
    C_Normal3f(-0.290886581f, -0.956757545f, 0.f),
    C_Vertex3f(0.0899202824f, 0.3200984f, 0.f),
    C_Vertex3f(0.0899202824f, 0.3200984f, -1.f),
    C_Normal3f(-0.374579161f, -0.927194834f, 0.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, 0.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, -1.f),
    C_Normal3f(-0.463473737f, -0.886110604f, 0.f),
    C_Vertex3f(0.0692311525f, 0.32845664f, 0.f),
    C_Vertex3f(0.0692311525f, 0.32845664f, -1.f),
    C_Normal3f(-0.564197719f, -0.825639844f, 0.f),
    C_Vertex3f(0.060239315f, 0.33377552f, 0.f),
    C_Vertex3f(0.060239315f, 0.33377552f, -1.f),
    C_Normal3f(-0.684651256f, -0.728870809f, 0.f),
    C_Vertex3f(0.0481042862f, 0.3428936f, 0.f),
    C_Vertex3f(0.0481042862f, 0.3428936f, -1.f),
    C_Normal3f(-0.805000246f, -0.593274474f, 0.f),
    C_Vertex3f(0.0348210335f, 0.357651711f, 0.f),
    C_Vertex3f(0.0348210335f, 0.357651711f, -1.f),
    C_Normal3f(-0.901867867f, -0.432012022f, 0.f),
    C_Vertex3f(0.0244345665f, 0.37501049f, 0.f),
    C_Vertex3f(0.0244345665f, 0.37501049f, -1.f),
    C_Normal3f(-0.954272866f, -0.298936754f, 0.f),
    C_Vertex3f(0.0169448853f, 0.39496994f, 0.f),
    C_Vertex3f(0.0169448853f, 0.39496994f, -1.f),
    C_Normal3f(-0.979899347f, -0.199492529f, 0.f),
    C_Vertex3f(0.0146484375f, 0.40625f, 0.f),
    C_Vertex3f(0.0146484375f, 0.40625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.624511719f, -0.013671875f, 0.f),
    C_Vertex3f(0.624511719f, -0.013671875f, -1.f),
    C_Vertex3f(0.624511719f, 0.0537109375f, -1.f),
    C_Vertex3f(0.624511719f, 0.0537109375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.624511719f, 0.0537109375f, 0.f),
    C_Vertex3f(0.624511719f, 0.0537109375f, -1.f),
    C_Vertex3f(0.457519531f, 0.0537109375f, -1.f),
    C_Vertex3f(0.457519531f, 0.0537109375f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.457519531f, 0.0537109375f, 0.f),
    C_Vertex3f(0.457519531f, 0.0537109375f, -1.f),
    C_Vertex3f(0.457519531f, 0.107429504f, -1.f),
    C_Vertex3f(0.457519531f, 0.107429504f, 0.f),
    C_Normal3f(-0.760758638f, 0.649034917f, 0.f),
    C_Vertex3f(0.457519531f, 0.107429504f, 0.f),
    C_Vertex3f(0.457519531f, 0.107429504f, -1.f),
    C_Vertex3f(0.634140015f, 0.314453125f, -1.f),
    C_Vertex3f(0.634140015f, 0.314453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.634140015f, 0.314453125f, 0.f),
    C_Vertex3f(0.634140015f, 0.314453125f, -1.f),
    C_Vertex3f(0.687988281f, 0.314453125f, -1.f),
    C_Vertex3f(0.687988281f, 0.314453125f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.687988281f, 0.314453125f, 0.f),
    C_Vertex3f(0.687988281f, 0.314453125f, -1.f),
    C_Vertex3f(0.687988281f, 0.101074219f, -1.f),
    C_Vertex3f(0.687988281f, 0.101074219f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.687988281f, 0.101074219f, 0.f),
    C_Vertex3f(0.687988281f, 0.101074219f, -1.f),
    C_Vertex3f(0.733398438f, 0.101074219f, -1.f),
    C_Vertex3f(0.733398438f, 0.101074219f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.733398438f, 0.101074219f, 0.f),
    C_Vertex3f(0.733398438f, 0.101074219f, -1.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, -1.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, 0.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, -1.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, -1.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, 0.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, -1.f),
    C_Vertex3f(0.687988281f, -0.013671875f, -1.f),
    C_Vertex3f(0.687988281f, -0.013671875f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.687988281f, -0.013671875f, 0.f),
    C_Vertex3f(0.687988281f, -0.013671875f, -1.f),
    C_Vertex3f(0.624511719f, -0.013671875f, -1.f),
    C_Vertex3f(0.624511719f, -0.013671875f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.624511719f, 0.101074219f, 0.f),
    C_Vertex3f(0.624511719f, 0.101074219f, -1.f),
    C_Vertex3f(0.624511719f, 0.215744019f, -1.f),
    C_Vertex3f(0.624511719f, 0.215744019f, 0.f),
    C_Normal3f(0.753352404f, -0.657616973f, 0.f),
    C_Vertex3f(0.624511719f, 0.215744019f, 0.f),
    C_Vertex3f(0.624511719f, 0.215744019f, -1.f),
    C_Vertex3f(0.524414063f, 0.101074219f, -1.f),
    C_Vertex3f(0.524414063f, 0.101074219f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.524414063f, 0.101074219f, 0.f),
    C_Vertex3f(0.524414063f, 0.101074219f, -1.f),
    C_Vertex3f(0.624511719f, 0.101074219f, -1.f),
    C_Vertex3f(0.624511719f, 0.101074219f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.591407776f, 0.651367188f, 0.f),
    C_Vertex3f(0.110153198f, -0.0249023438f, 0.f),
    C_Vertex3f(0.178771973f, -0.0249023438f, 0.f),
    C_Vertex3f(0.660041809f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.248166084f, 0.519637108f, 0.f),
    C_Vertex3f(0.255930901f, 0.529000282f, 0.f),
    C_Vertex3f(0.25682497f, 0.606219292f, 0.f),
    C_Vertex3f(0.261754513f, 0.539016247f, 0.f),
    C_Vertex3f(0.26399374f, 0.591683388f, 0.f),
    C_Vertex3f(0.265636921f, 0.549685001f, 0.f),
    C_Vertex3f(0.267578125f, 0.575704575f, 0.f),
    C_Vertex3f(0.267578125f, 0.561006546f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227075934f, 0.633146763f, 0.f),
    C_Vertex3f(0.227468014f, 0.50482893f, 0.f),
    C_Vertex3f(0.235318661f, 0.62804985f, 0.f),
    C_Vertex3f(0.238678455f, 0.51157999f, 0.f),
    C_Vertex3f(0.246071815f, 0.619312286f, 0.f),
    C_Vertex3f(0.248166084f, 0.519637108f, 0.f),
    C_Vertex3f(0.25682497f, 0.606219292f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.202651024f, 0.453656197f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, 0.f),
    C_Vertex3f(0.207368612f, 0.641156197f, 0.f),
    C_Vertex3f(0.21453476f, 0.499383926f, 0.f),
    C_Vertex3f(0.217759252f, 0.637515545f, 0.f),
    C_Vertex3f(0.227468014f, 0.50482893f, 0.f),
    C_Vertex3f(0.227075934f, 0.633146763f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.189328194f, 0.58597374f, 0.f),
    C_Vertex3f(0.194273472f, 0.579726219f, 0.f),
    C_Vertex3f(0.195904016f, 0.644068718f, 0.f),
    C_Vertex3f(0.197570324f, 0.572880745f, 0.f),
    C_Vertex3f(0.207368612f, 0.641156197f, 0.f),
    C_Vertex3f(0.19921875f, 0.565437317f, 0.f),
    C_Vertex3f(0.202651024f, 0.453656197f, 0.f),
    C_Vertex3f(0.19921875f, 0.556745529f, 0.f),
    C_Vertex3f(0.197174549f, 0.548109055f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.183365464f, 0.646253109f, 0.f),
    C_Vertex3f(0.169752955f, 0.64770937f, 0.f),
    C_Vertex3f(0.174546719f, 0.595860481f, 0.f),
    C_Vertex3f(0.182734489f, 0.591623306f, 0.f),
    C_Vertex3f(0.189328194f, 0.58597374f, 0.f),
    C_Vertex3f(0.195904016f, 0.644068718f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.140748978f, 0.600097656f, 0.f),
    C_Vertex3f(0.153388977f, 0.600097656f, 0.f),
    C_Vertex3f(0.15506649f, 0.6484375f, 0.f),
    C_Vertex3f(0.164764881f, 0.598685265f, 0.f),
    C_Vertex3f(0.169752955f, 0.64770937f, 0.f),
    C_Vertex3f(0.174546719f, 0.595860481f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052277f, 0.6484375f, 0.f),
    C_Vertex3f(0.11093092f, 0.646143913f, 0.f),
    C_Vertex3f(0.111895561f, 0.592830658f, 0.f),
    C_Vertex3f(0.120167255f, 0.596464157f, 0.f),
    C_Vertex3f(0.129785061f, 0.59888649f, 0.f),
    C_Vertex3f(0.140748978f, 0.600097656f, 0.f),
    C_Vertex3f(0.15506649f, 0.6484375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.11093092f, 0.646143913f, 0.f),
    C_Vertex3f(0.0909562111f, 0.64155674f, 0.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, 0.f),
    C_Vertex3f(0.0985121727f, 0.581159592f, 0.f),
    C_Vertex3f(0.104969978f, 0.587985992f, 0.f),
    C_Vertex3f(0.111895561f, 0.592830658f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0909562111f, 0.64155674f, 0.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, 0.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, 0.f),
    C_Vertex3f(0.0869998932f, 0.561561584f, 0.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0844726563f, 0.555175781f, 0.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, 0.f),
    C_Vertex3f(0.0604467392f, 0.625501633f, 0.f),
    C_Vertex3f(0.0479416847f, 0.612832069f, 0.f),
    C_Vertex3f(0.0366129875f, 0.59666729f, 0.f),
    C_Vertex3f(0.0264606476f, 0.577007294f, 0.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.212404728f, 0.437254429f, 0.f),
    C_Vertex3f(0.214355469f, 0.427412033f, 0.f),
    C_Vertex3f(0.216918945f, 0.494905472f, 0.f),
    C_Vertex3f(0.223830462f, 0.32666564f, 0.f),
    C_Vertex3f(0.233778477f, 0.488527298f, 0.f),
    C_Vertex3f(0.234298468f, 0.332331896f, 0.f),
    C_Vertex3f(0.248073101f, 0.480588913f, 0.f),
    C_Vertex3f(0.243850708f, 0.338942528f, 0.f),
    C_Vertex3f(0.252487183f, 0.346497536f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.208503246f, 0.446002483f, 0.f),
    C_Vertex3f(0.212404728f, 0.437254429f, 0.f),
    C_Vertex3f(0.216918945f, 0.494905472f, 0.f),
    C_Vertex3f(0.207206726f, 0.497314453f, 0.f),
    C_Vertex3f(0.202651024f, 0.453656197f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276237488f, 0.380000591f, 0.f),
    C_Vertex3f(0.279476166f, 0.389197111f, 0.f),
    C_Vertex3f(0.280423641f, 0.435041428f, 0.f),
    C_Vertex3f(0.281635284f, 0.398721933f, 0.f),
    C_Vertex3f(0.282714844f, 0.421110153f, 0.f),
    C_Vertex3f(0.282714844f, 0.408575058f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.260044098f, 0.354380846f, 0.f),
    C_Vertex3f(0.266521454f, 0.362592459f, 0.f),
    C_Vertex3f(0.268967628f, 0.460031509f, 0.f),
    C_Vertex3f(0.27191925f, 0.371132374f, 0.f),
    C_Vertex3f(0.275841236f, 0.448015213f, 0.f),
    C_Vertex3f(0.276237488f, 0.380000591f, 0.f),
    C_Vertex3f(0.280423641f, 0.435041428f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259802818f, 0.471090317f, 0.f),
    C_Vertex3f(0.248073101f, 0.480588913f, 0.f),
    C_Vertex3f(0.252487183f, 0.346497536f, 0.f),
    C_Vertex3f(0.260044098f, 0.354380846f, 0.f),
    C_Vertex3f(0.268967628f, 0.460031509f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214355469f, 0.415929794f, 0.f),
    C_Vertex3f(0.21200943f, 0.404776573f, 0.f),
    C_Vertex3f(0.21244669f, 0.32194376f, 0.f),
    C_Vertex3f(0.223830462f, 0.32666564f, 0.f),
    C_Vertex3f(0.214355469f, 0.427412033f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.21200943f, 0.404776573f, 0.f),
    C_Vertex3f(0.207317352f, 0.394497871f, 0.f),
    C_Vertex3f(0.21244669f, 0.32194376f, 0.f),
    C_Vertex3f(0.200279236f, 0.385093689f, 0.f),
    C_Vertex3f(0.200147152f, 0.318166256f, 0.f),
    C_Vertex3f(0.190895081f, 0.376564026f, 0.f),
    C_Vertex3f(0.186931849f, 0.315333128f, 0.f),
    C_Vertex3f(0.179983139f, 0.370166779f, 0.f),
    C_Vertex3f(0.172800779f, 0.313444376f, 0.f),
    C_Vertex3f(0.167543411f, 0.365901947f, 0.f),
    C_Vertex3f(0.157753944f, 0.3125f, 0.f),
    C_Vertex3f(0.153575897f, 0.363769531f, 0.f),
    C_Vertex3f(0.142120361f, 0.3125f, 0.f),
    C_Vertex3f(0.139390945f, 0.363769531f, 0.f),
    C_Vertex3f(0.127717614f, 0.31325984f, 0.f),
    C_Vertex3f(0.127388954f, 0.365411758f, 0.f),
    C_Vertex3f(0.116259575f, 0.368696213f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127717614f, 0.31325984f, 0.f),
    C_Vertex3f(0.116259575f, 0.368696213f, 0.f),
    C_Vertex3f(0.114216685f, 0.31477952f, 0.f),
    C_Vertex3f(0.106002808f, 0.373622894f, 0.f),
    C_Vertex3f(0.101617575f, 0.31705904f, 0.f),
    C_Vertex3f(0.0978183746f, 0.37931633f, 0.f),
    C_Vertex3f(0.0914888382f, 0.386214256f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0899202824f, 0.3200984f, 0.f),
    C_Vertex3f(0.101617575f, 0.31705904f, 0.f),
    C_Vertex3f(0.0914888382f, 0.386214256f, 0.f),
    C_Vertex3f(0.085814476f, 0.395192146f, 0.f),
    C_Vertex3f(0.0807952881f, 0.40625f, 0.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0786132813f, 0.412818909f, 0.f),
    C_Vertex3f(0.0146484375f, 0.40625f, 0.f),
    C_Vertex3f(0.0169448853f, 0.39496994f, 0.f),
    C_Vertex3f(0.0244345665f, 0.37501049f, 0.f),
    C_Vertex3f(0.0348210335f, 0.357651711f, 0.f),
    C_Vertex3f(0.0481042862f, 0.3428936f, 0.f),
    C_Vertex3f(0.060239315f, 0.33377552f, 0.f),
    C_Vertex3f(0.0692311525f, 0.32845664f, 0.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, 0.f),
    C_Vertex3f(0.0807952881f, 0.40625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.194848061f, 0.460215569f, 0.f),
    C_Vertex3f(0.202651024f, 0.453656197f, 0.f),
    C_Vertex3f(0.197174549f, 0.548109055f, 0.f),
    C_Vertex3f(0.193086147f, 0.540178299f, 0.f),
    C_Vertex3f(0.186953545f, 0.532953262f, 0.f),
    C_Vertex3f(0.184443951f, 0.465135098f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.141309738f, 0.46972847f, 0.f),
    C_Vertex3f(0.146728516f, 0.470054626f, 0.f),
    C_Vertex3f(0.155306339f, 0.518447876f, 0.f),
    C_Vertex3f(0.155832291f, 0.470054626f, 0.f),
    C_Vertex3f(0.168227673f, 0.521598816f, 0.f),
    C_Vertex3f(0.171438694f, 0.468414783f, 0.f),
    C_Vertex3f(0.178776741f, 0.526433945f, 0.f),
    C_Vertex3f(0.184443951f, 0.465135098f, 0.f),
    C_Vertex3f(0.186953545f, 0.532953262f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122449875f, 0.46875f, 0.f),
    C_Vertex3f(0.126461029f, 0.468913078f, 0.f),
    C_Vertex3f(0.13117981f, 0.517089844f, 0.f),
    C_Vertex3f(0.13274765f, 0.469239235f, 0.f),
    C_Vertex3f(0.140012741f, 0.516981125f, 0.f),
    C_Vertex3f(0.141309738f, 0.46972847f, 0.f),
    C_Vertex3f(0.155306339f, 0.518447876f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.457519531f, 0.0537109375f, 0.f),
    C_Vertex3f(0.624511719f, 0.0537109375f, 0.f),
    C_Vertex3f(0.457519531f, 0.107429504f, 0.f),
    C_Vertex3f(0.524414063f, 0.101074219f, 0.f),
    C_Vertex3f(0.634140015f, 0.314453125f, 0.f),
    C_Vertex3f(0.624511719f, 0.215744019f, 0.f),
    C_Vertex3f(0.687988281f, -0.013671875f, 0.f),
    C_Vertex3f(0.624511719f, 0.101074219f, 0.f),
    C_Vertex3f(0.624511719f, 0.0537109375f, 0.f),
    C_Vertex3f(0.524414063f, 0.101074219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.634140015f, 0.314453125f, 0.f),
    C_Vertex3f(0.687988281f, -0.013671875f, 0.f),
    C_Vertex3f(0.687988281f, 0.314453125f, 0.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, 0.f),
    C_Vertex3f(0.687988281f, 0.101074219f, 0.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, 0.f),
    C_Vertex3f(0.733398438f, 0.101074219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.687988281f, -0.013671875f, 0.f),
    C_Vertex3f(0.624511719f, 0.0537109375f, 0.f),
    C_Vertex3f(0.624511719f, -0.013671875f, 0.f),
    C_Vertex3f(0.122449875f, 0.46875f, 0.f),
    C_Vertex3f(0.13117981f, 0.517089844f, 0.f),
    C_Vertex3f(0.121582031f, 0.46875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.591407776f, 0.651367188f, -1.f),
    C_Vertex3f(0.110153198f, -0.0249023438f, -1.f),
    C_Vertex3f(0.178771973f, -0.0249023438f, -1.f),
    C_Vertex3f(0.660041809f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.248166084f, 0.519637108f, -1.f),
    C_Vertex3f(0.255930901f, 0.529000282f, -1.f),
    C_Vertex3f(0.25682497f, 0.606219292f, -1.f),
    C_Vertex3f(0.261754513f, 0.539016247f, -1.f),
    C_Vertex3f(0.26399374f, 0.591683388f, -1.f),
    C_Vertex3f(0.265636921f, 0.549685001f, -1.f),
    C_Vertex3f(0.267578125f, 0.575704575f, -1.f),
    C_Vertex3f(0.267578125f, 0.561006546f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227075934f, 0.633146763f, -1.f),
    C_Vertex3f(0.227468014f, 0.50482893f, -1.f),
    C_Vertex3f(0.235318661f, 0.62804985f, -1.f),
    C_Vertex3f(0.238678455f, 0.51157999f, -1.f),
    C_Vertex3f(0.246071815f, 0.619312286f, -1.f),
    C_Vertex3f(0.248166084f, 0.519637108f, -1.f),
    C_Vertex3f(0.25682497f, 0.606219292f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.202651024f, 0.453656197f, -1.f),
    C_Vertex3f(0.207206726f, 0.497314453f, -1.f),
    C_Vertex3f(0.207368612f, 0.641156197f, -1.f),
    C_Vertex3f(0.21453476f, 0.499383926f, -1.f),
    C_Vertex3f(0.217759252f, 0.637515545f, -1.f),
    C_Vertex3f(0.227468014f, 0.50482893f, -1.f),
    C_Vertex3f(0.227075934f, 0.633146763f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.189328194f, 0.58597374f, -1.f),
    C_Vertex3f(0.194273472f, 0.579726219f, -1.f),
    C_Vertex3f(0.195904016f, 0.644068718f, -1.f),
    C_Vertex3f(0.197570324f, 0.572880745f, -1.f),
    C_Vertex3f(0.207368612f, 0.641156197f, -1.f),
    C_Vertex3f(0.19921875f, 0.565437317f, -1.f),
    C_Vertex3f(0.202651024f, 0.453656197f, -1.f),
    C_Vertex3f(0.19921875f, 0.556745529f, -1.f),
    C_Vertex3f(0.197174549f, 0.548109055f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.183365464f, 0.646253109f, -1.f),
    C_Vertex3f(0.169752955f, 0.64770937f, -1.f),
    C_Vertex3f(0.174546719f, 0.595860481f, -1.f),
    C_Vertex3f(0.182734489f, 0.591623306f, -1.f),
    C_Vertex3f(0.189328194f, 0.58597374f, -1.f),
    C_Vertex3f(0.195904016f, 0.644068718f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.140748978f, 0.600097656f, -1.f),
    C_Vertex3f(0.153388977f, 0.600097656f, -1.f),
    C_Vertex3f(0.15506649f, 0.6484375f, -1.f),
    C_Vertex3f(0.164764881f, 0.598685265f, -1.f),
    C_Vertex3f(0.169752955f, 0.64770937f, -1.f),
    C_Vertex3f(0.174546719f, 0.595860481f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052277f, 0.6484375f, -1.f),
    C_Vertex3f(0.11093092f, 0.646143913f, -1.f),
    C_Vertex3f(0.111895561f, 0.592830658f, -1.f),
    C_Vertex3f(0.120167255f, 0.596464157f, -1.f),
    C_Vertex3f(0.129785061f, 0.59888649f, -1.f),
    C_Vertex3f(0.140748978f, 0.600097656f, -1.f),
    C_Vertex3f(0.15506649f, 0.6484375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.11093092f, 0.646143913f, -1.f),
    C_Vertex3f(0.0909562111f, 0.64155674f, -1.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, -1.f),
    C_Vertex3f(0.0985121727f, 0.581159592f, -1.f),
    C_Vertex3f(0.104969978f, 0.587985992f, -1.f),
    C_Vertex3f(0.111895561f, 0.592830658f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0909562111f, 0.64155674f, -1.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, -1.f),
    C_Vertex3f(0.0844726563f, 0.555175781f, -1.f),
    C_Vertex3f(0.0869998932f, 0.561561584f, -1.f),
    C_Vertex3f(0.0925221443f, 0.572351456f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0844726563f, 0.555175781f, -1.f),
    C_Vertex3f(0.0741281509f, 0.63467598f, -1.f),
    C_Vertex3f(0.0604467392f, 0.625501633f, -1.f),
    C_Vertex3f(0.0479416847f, 0.612832069f, -1.f),
    C_Vertex3f(0.0366129875f, 0.59666729f, -1.f),
    C_Vertex3f(0.0264606476f, 0.577007294f, -1.f),
    C_Vertex3f(0.0219726563f, 0.565429688f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.212404728f, 0.437254429f, -1.f),
    C_Vertex3f(0.214355469f, 0.427412033f, -1.f),
    C_Vertex3f(0.216918945f, 0.494905472f, -1.f),
    C_Vertex3f(0.223830462f, 0.32666564f, -1.f),
    C_Vertex3f(0.233778477f, 0.488527298f, -1.f),
    C_Vertex3f(0.234298468f, 0.332331896f, -1.f),
    C_Vertex3f(0.248073101f, 0.480588913f, -1.f),
    C_Vertex3f(0.243850708f, 0.338942528f, -1.f),
    C_Vertex3f(0.252487183f, 0.346497536f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.208503246f, 0.446002483f, -1.f),
    C_Vertex3f(0.212404728f, 0.437254429f, -1.f),
    C_Vertex3f(0.216918945f, 0.494905472f, -1.f),
    C_Vertex3f(0.207206726f, 0.497314453f, -1.f),
    C_Vertex3f(0.202651024f, 0.453656197f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276237488f, 0.380000591f, -1.f),
    C_Vertex3f(0.279476166f, 0.389197111f, -1.f),
    C_Vertex3f(0.280423641f, 0.435041428f, -1.f),
    C_Vertex3f(0.281635284f, 0.398721933f, -1.f),
    C_Vertex3f(0.282714844f, 0.421110153f, -1.f),
    C_Vertex3f(0.282714844f, 0.408575058f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.260044098f, 0.354380846f, -1.f),
    C_Vertex3f(0.266521454f, 0.362592459f, -1.f),
    C_Vertex3f(0.268967628f, 0.460031509f, -1.f),
    C_Vertex3f(0.27191925f, 0.371132374f, -1.f),
    C_Vertex3f(0.275841236f, 0.448015213f, -1.f),
    C_Vertex3f(0.276237488f, 0.380000591f, -1.f),
    C_Vertex3f(0.280423641f, 0.435041428f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259802818f, 0.471090317f, -1.f),
    C_Vertex3f(0.248073101f, 0.480588913f, -1.f),
    C_Vertex3f(0.252487183f, 0.346497536f, -1.f),
    C_Vertex3f(0.260044098f, 0.354380846f, -1.f),
    C_Vertex3f(0.268967628f, 0.460031509f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214355469f, 0.415929794f, -1.f),
    C_Vertex3f(0.21200943f, 0.404776573f, -1.f),
    C_Vertex3f(0.21244669f, 0.32194376f, -1.f),
    C_Vertex3f(0.223830462f, 0.32666564f, -1.f),
    C_Vertex3f(0.214355469f, 0.427412033f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.21200943f, 0.404776573f, -1.f),
    C_Vertex3f(0.207317352f, 0.394497871f, -1.f),
    C_Vertex3f(0.21244669f, 0.32194376f, -1.f),
    C_Vertex3f(0.200279236f, 0.385093689f, -1.f),
    C_Vertex3f(0.200147152f, 0.318166256f, -1.f),
    C_Vertex3f(0.190895081f, 0.376564026f, -1.f),
    C_Vertex3f(0.186931849f, 0.315333128f, -1.f),
    C_Vertex3f(0.179983139f, 0.370166779f, -1.f),
    C_Vertex3f(0.172800779f, 0.313444376f, -1.f),
    C_Vertex3f(0.167543411f, 0.365901947f, -1.f),
    C_Vertex3f(0.157753944f, 0.3125f, -1.f),
    C_Vertex3f(0.153575897f, 0.363769531f, -1.f),
    C_Vertex3f(0.142120361f, 0.3125f, -1.f),
    C_Vertex3f(0.139390945f, 0.363769531f, -1.f),
    C_Vertex3f(0.127717614f, 0.31325984f, -1.f),
    C_Vertex3f(0.127388954f, 0.365411758f, -1.f),
    C_Vertex3f(0.116259575f, 0.368696213f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127717614f, 0.31325984f, -1.f),
    C_Vertex3f(0.116259575f, 0.368696213f, -1.f),
    C_Vertex3f(0.114216685f, 0.31477952f, -1.f),
    C_Vertex3f(0.106002808f, 0.373622894f, -1.f),
    C_Vertex3f(0.101617575f, 0.31705904f, -1.f),
    C_Vertex3f(0.0978183746f, 0.37931633f, -1.f),
    C_Vertex3f(0.0914888382f, 0.386214256f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0899202824f, 0.3200984f, -1.f),
    C_Vertex3f(0.101617575f, 0.31705904f, -1.f),
    C_Vertex3f(0.0914888382f, 0.386214256f, -1.f),
    C_Vertex3f(0.085814476f, 0.395192146f, -1.f),
    C_Vertex3f(0.0807952881f, 0.40625f, -1.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0786132813f, 0.412818909f, -1.f),
    C_Vertex3f(0.0146484375f, 0.40625f, -1.f),
    C_Vertex3f(0.0169448853f, 0.39496994f, -1.f),
    C_Vertex3f(0.0244345665f, 0.37501049f, -1.f),
    C_Vertex3f(0.0348210335f, 0.357651711f, -1.f),
    C_Vertex3f(0.0481042862f, 0.3428936f, -1.f),
    C_Vertex3f(0.060239315f, 0.33377552f, -1.f),
    C_Vertex3f(0.0692311525f, 0.32845664f, -1.f),
    C_Vertex3f(0.0791248083f, 0.3238976f, -1.f),
    C_Vertex3f(0.0807952881f, 0.40625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.194848061f, 0.460215569f, -1.f),
    C_Vertex3f(0.202651024f, 0.453656197f, -1.f),
    C_Vertex3f(0.197174549f, 0.548109055f, -1.f),
    C_Vertex3f(0.193086147f, 0.540178299f, -1.f),
    C_Vertex3f(0.186953545f, 0.532953262f, -1.f),
    C_Vertex3f(0.184443951f, 0.465135098f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.141309738f, 0.46972847f, -1.f),
    C_Vertex3f(0.146728516f, 0.470054626f, -1.f),
    C_Vertex3f(0.155306339f, 0.518447876f, -1.f),
    C_Vertex3f(0.155832291f, 0.470054626f, -1.f),
    C_Vertex3f(0.168227673f, 0.521598816f, -1.f),
    C_Vertex3f(0.171438694f, 0.468414783f, -1.f),
    C_Vertex3f(0.178776741f, 0.526433945f, -1.f),
    C_Vertex3f(0.184443951f, 0.465135098f, -1.f),
    C_Vertex3f(0.186953545f, 0.532953262f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122449875f, 0.46875f, -1.f),
    C_Vertex3f(0.126461029f, 0.468913078f, -1.f),
    C_Vertex3f(0.13117981f, 0.517089844f, -1.f),
    C_Vertex3f(0.13274765f, 0.469239235f, -1.f),
    C_Vertex3f(0.140012741f, 0.516981125f, -1.f),
    C_Vertex3f(0.141309738f, 0.46972847f, -1.f),
    C_Vertex3f(0.155306339f, 0.518447876f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.457519531f, 0.0537109375f, -1.f),
    C_Vertex3f(0.624511719f, 0.0537109375f, -1.f),
    C_Vertex3f(0.457519531f, 0.107429504f, -1.f),
    C_Vertex3f(0.524414063f, 0.101074219f, -1.f),
    C_Vertex3f(0.634140015f, 0.314453125f, -1.f),
    C_Vertex3f(0.624511719f, 0.215744019f, -1.f),
    C_Vertex3f(0.687988281f, -0.013671875f, -1.f),
    C_Vertex3f(0.624511719f, 0.101074219f, -1.f),
    C_Vertex3f(0.624511719f, 0.0537109375f, -1.f),
    C_Vertex3f(0.524414063f, 0.101074219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.634140015f, 0.314453125f, -1.f),
    C_Vertex3f(0.687988281f, -0.013671875f, -1.f),
    C_Vertex3f(0.687988281f, 0.314453125f, -1.f),
    C_Vertex3f(0.687988281f, 0.0537109375f, -1.f),
    C_Vertex3f(0.687988281f, 0.101074219f, -1.f),
    C_Vertex3f(0.733398438f, 0.0537109375f, -1.f),
    C_Vertex3f(0.733398438f, 0.101074219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.687988281f, -0.013671875f, -1.f),
    C_Vertex3f(0.624511719f, 0.0537109375f, -1.f),
    C_Vertex3f(0.624511719f, -0.013671875f, -1.f),
    C_Vertex3f(0.122449875f, 0.46875f, -1.f),
    C_Vertex3f(0.13117981f, 0.517089844f, -1.f),
    C_Vertex3f(0.121582031f, 0.46875f, -1.f),
    C_End(),
    C_Translatef(0.746582031f, 0.f, 0.f),
};
static const Call char_191[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.3203125f, 0.464355469f, 0.f),
    C_Vertex3f(0.3203125f, 0.464355469f, -1.f),
    C_Vertex3f(0.3203125f, 0.375f, -1.f),
    C_Vertex3f(0.3203125f, 0.375f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.3203125f, 0.375f, 0.f),
    C_Vertex3f(0.3203125f, 0.375f, -1.f),
    C_Vertex3f(0.239257813f, 0.375f, -1.f),
    C_Vertex3f(0.239257813f, 0.375f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.239257813f, 0.375f, 0.f),
    C_Vertex3f(0.239257813f, 0.375f, -1.f),
    C_Vertex3f(0.239257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.239257813f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.239257813f, 0.464355469f, 0.f),
    C_Vertex3f(0.239257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.3203125f, 0.464355469f, -1.f),
    C_Vertex3f(0.3203125f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999573529f, 0.0292033255f, 0.f),
    C_Vertex3f(0.314941406f, 0.306640625f, 0.f),
    C_Vertex3f(0.314941406f, 0.306640625f, -1.f),
    C_Normal3f(0.999726951f, 0.0233686436f, 0.f),
    C_Vertex3f(0.315158844f, 0.299198151f, 0.f),
    C_Vertex3f(0.315158844f, 0.299198151f, -1.f),
    C_Normal3f(0.999967337f, 0.00807682332f, 0.f),
    C_Vertex3f(0.315376282f, 0.288036346f, 0.f),
    C_Vertex3f(0.315376282f, 0.288036346f, -1.f),
    C_Normal3f(0.999074399f, -0.0430164672f, 0.f),
    C_Vertex3f(0.315376282f, 0.272277832f, 0.f),
    C_Vertex3f(0.315376282f, 0.272277832f, -1.f),
    C_Normal3f(0.993458927f, -0.114189915f, 0.f),
    C_Vertex3f(0.313738823f, 0.250005722f, 0.f),
    C_Vertex3f(0.313738823f, 0.250005722f, -1.f),
    C_Normal3f(0.978799403f, -0.204821154f, 0.f),
    C_Vertex3f(0.310463905f, 0.229539871f, 0.f),
    C_Vertex3f(0.310463905f, 0.229539871f, -1.f),
    C_Normal3f(0.957226932f, -0.289338529f, 0.f),
    C_Vertex3f(0.305551529f, 0.21088028f, 0.f),
    C_Vertex3f(0.305551529f, 0.21088028f, -1.f),
    C_Normal3f(0.933475971f, -0.358639866f, 0.f),
    C_Vertex3f(0.302276611f, 0.202453613f, 0.f),
    C_Vertex3f(0.302276611f, 0.202453613f, -1.f),
    C_Normal3f(0.910422564f, -0.413679659f, 0.f),
    C_Vertex3f(0.299875259f, 0.196105957f, 0.f),
    C_Vertex3f(0.299875259f, 0.196105957f, -1.f),
    C_Normal3f(0.876633465f, -0.481158733f, 0.f),
    C_Vertex3f(0.293599129f, 0.183356285f, 0.f),
    C_Vertex3f(0.293599129f, 0.183356285f, -1.f),
    C_Normal3f(0.834080875f, -0.55164212f, 0.f),
    C_Vertex3f(0.285849571f, 0.170552254f, 0.f),
    C_Vertex3f(0.285849571f, 0.170552254f, -1.f),
    C_Normal3f(0.792253196f, -0.610192537f, 0.f),
    C_Vertex3f(0.276626587f, 0.157693863f, 0.f),
    C_Vertex3f(0.276626587f, 0.157693863f, -1.f),
    C_Normal3f(0.745937347f, -0.666016042f, 0.f),
    C_Vertex3f(0.267349243f, 0.146532059f, 0.f),
    C_Vertex3f(0.267349243f, 0.146532059f, -1.f),
    C_Normal3f(0.708404601f, -0.705806553f, 0.f),
    C_Vertex3f(0.256352901f, 0.134987354f, 0.f),
    C_Vertex3f(0.256352901f, 0.134987354f, -1.f),
    C_Normal3f(0.684521496f, -0.728992641f, 0.f),
    C_Vertex3f(0.242218494f, 0.121308804f, 0.f),
    C_Vertex3f(0.242218494f, 0.121308804f, -1.f),
    C_Normal3f(0.668330491f, -0.743864417f, 0.f),
    C_Vertex3f(0.224946022f, 0.105496407f, 0.f),
    C_Vertex3f(0.224946022f, 0.105496407f, -1.f),
    C_Normal3f(0.666131258f, -0.745834529f, 0.f),
    C_Vertex3f(0.209638119f, 0.0920367241f, 0.f),
    C_Vertex3f(0.209638119f, 0.0920367241f, -1.f),
    C_Normal3f(0.683786392f, -0.729682267f, 0.f),
    C_Vertex3f(0.200333238f, 0.0835138559f, 0.f),
    C_Vertex3f(0.200333238f, 0.0835138559f, -1.f),
    C_Normal3f(0.702311099f, -0.711870193f, 0.f),
    C_Vertex3f(0.191928744f, 0.0754412413f, 0.f),
    C_Vertex3f(0.191928744f, 0.0754412413f, -1.f),
    C_Normal3f(0.723701596f, -0.690112948f, 0.f),
    C_Vertex3f(0.184424639f, 0.0678188801f, 0.f),
    C_Vertex3f(0.184424639f, 0.0678188801f, -1.f),
    C_Normal3f(0.748562336f, -0.66306448f, 0.f),
    C_Vertex3f(0.177820921f, 0.0606467724f, 0.f),
    C_Vertex3f(0.177820921f, 0.0606467724f, -1.f),
    C_Normal3f(0.777602911f, -0.628755629f, 0.f),
    C_Vertex3f(0.172117591f, 0.0539249182f, 0.f),
    C_Vertex3f(0.172117591f, 0.0539249182f, -1.f),
    C_Normal3f(0.811579704f, -0.584241688f, 0.f),
    C_Vertex3f(0.167314649f, 0.0476533175f, 0.f),
    C_Vertex3f(0.167314649f, 0.0476533175f, -1.f),
    C_Normal3f(0.855337024f, -0.518071949f, 0.f),
    C_Vertex3f(0.163412094f, 0.0418319702f, 0.f),
    C_Vertex3f(0.163412094f, 0.0418319702f, -1.f),
    C_Normal3f(0.906125903f, -0.423008054f, 0.f),
    C_Vertex3f(0.158908844f, 0.0337753296f, 0.f),
    C_Vertex3f(0.158908844f, 0.0337753296f, -1.f),
    C_Normal3f(0.950529873f, -0.310633242f, 0.f),
    C_Vertex3f(0.154405594f, 0.0225391388f, 0.f),
    C_Vertex3f(0.154405594f, 0.0225391388f, -1.f),
    C_Normal3f(0.98278439f, -0.184756264f, 0.f),
    C_Vertex3f(0.151403427f, 0.0108089447f, 0.f),
    C_Vertex3f(0.151403427f, 0.0108089447f, -1.f),
    C_Normal3f(0.998111904f, -0.0614210702f, 0.f),
    C_Vertex3f(0.149902344f, -0.00141525269f, 0.f),
    C_Vertex3f(0.149902344f, -0.00141525269f, -1.f),
    C_Normal3f(0.998861015f, 0.0477137156f, 0.f),
    C_Vertex3f(0.149902344f, -0.013584137f, 0.f),
    C_Vertex3f(0.149902344f, -0.013584137f, -1.f),
    C_Normal3f(0.988695502f, 0.149938077f, 0.f),
    C_Vertex3f(0.151021004f, -0.0248337984f, 0.f),
    C_Vertex3f(0.151021004f, -0.0248337984f, -1.f),
    C_Normal3f(0.967469037f, 0.252989471f, 0.f),
    C_Vertex3f(0.153258324f, -0.0357135534f, 0.f),
    C_Vertex3f(0.153258324f, -0.0357135534f, -1.f),
    C_Normal3f(0.935028434f, 0.354572684f, 0.f),
    C_Vertex3f(0.156614304f, -0.046223402f, 0.f),
    C_Vertex3f(0.156614304f, -0.046223402f, -1.f),
    C_Normal3f(0.892392993f, 0.451259106f, 0.f),
    C_Vertex3f(0.161088943f, -0.0563633442f, 0.f),
    C_Vertex3f(0.161088943f, -0.0563633442f, -1.f),
    C_Normal3f(0.841546059f, 0.540185392f, 0.f),
    C_Vertex3f(0.166682243f, -0.0661333799f, 0.f),
    C_Vertex3f(0.166682243f, -0.0661333799f, -1.f),
    C_Normal3f(0.785042226f, 0.619442225f, 0.f),
    C_Vertex3f(0.173394203f, -0.0755335093f, 0.f),
    C_Vertex3f(0.173394203f, -0.0755335093f, -1.f),
    C_Normal3f(0.725534022f, 0.688186347f, 0.f),
    C_Vertex3f(0.181224823f, -0.0845637321f, 0.f),
    C_Vertex3f(0.181224823f, -0.0845637321f, -1.f),
    C_Normal3f(0.661614418f, 0.749844253f, 0.f),
    C_Vertex3f(0.190174103f, -0.0932240486f, 0.f),
    C_Vertex3f(0.190174103f, -0.0932240486f, -1.f),
    C_Normal3f(0.586919665f, 0.809645176f, 0.f),
    C_Vertex3f(0.199628353f, -0.100801826f, 0.f),
    C_Vertex3f(0.199628353f, -0.100801826f, -1.f),
    C_Normal3f(0.503690422f, 0.863884211f, 0.f),
    C_Vertex3f(0.209587574f, -0.107297063f, 0.f),
    C_Vertex3f(0.209587574f, -0.107297063f, -1.f),
    C_Normal3f(0.413817823f, 0.91035974f, 0.f),
    C_Vertex3f(0.220051765f, -0.112709761f, 0.f),
    C_Vertex3f(0.220051765f, -0.112709761f, -1.f),
    C_Normal3f(0.319898605f, 0.94745183f, 0.f),
    C_Vertex3f(0.231020927f, -0.117039919f, 0.f),
    C_Vertex3f(0.231020927f, -0.117039919f, -1.f),
    C_Normal3f(0.224875778f, 0.974387467f, 0.f),
    C_Vertex3f(0.24249506f, -0.120287538f, 0.f),
    C_Vertex3f(0.24249506f, -0.120287538f, -1.f),
    C_Normal3f(0.131600782f, 0.991302729f, 0.f),
    C_Vertex3f(0.254474163f, -0.122452617f, 0.f),
    C_Vertex3f(0.254474163f, -0.122452617f, -1.f),
    C_Normal3f(0.042919334f, 0.999078512f, 0.f),
    C_Vertex3f(0.266958237f, -0.123535156f, 0.f),
    C_Vertex3f(0.266958237f, -0.123535156f, -1.f),
    C_Normal3f(-0.0400054604f, 0.99919939f, 0.f),
    C_Vertex3f(0.279673576f, -0.123535156f, 0.f),
    C_Vertex3f(0.279673576f, -0.123535156f, -1.f),
    C_Normal3f(-0.125627682f, 0.99207747f, 0.f),
    C_Vertex3f(0.291617393f, -0.122547865f, 0.f),
    C_Vertex3f(0.291617393f, -0.122547865f, -1.f),
    C_Normal3f(-0.215266764f, 0.976555288f, 0.f),
    C_Vertex3f(0.303063393f, -0.120573282f, 0.f),
    C_Vertex3f(0.303063393f, -0.120573282f, -1.f),
    C_Normal3f(-0.30733642f, 0.951600969f, 0.f),
    C_Vertex3f(0.314011574f, -0.117611408f, 0.f),
    C_Vertex3f(0.314011574f, -0.117611408f, -1.f),
    C_Normal3f(-0.399285078f, 0.916826844f, 0.f),
    C_Vertex3f(0.324461937f, -0.113662243f, 0.f),
    C_Vertex3f(0.324461937f, -0.113662243f, -1.f),
    C_Normal3f(-0.48833403f, 0.872656822f, 0.f),
    C_Vertex3f(0.334414482f, -0.108725786f, 0.f),
    C_Vertex3f(0.334414482f, -0.108725786f, -1.f),
    C_Normal3f(-0.571865678f, 0.82034725f, 0.f),
    C_Vertex3f(0.343869209f, -0.102802038f, 0.f),
    C_Vertex3f(0.343869209f, -0.102802038f, -1.f),
    C_Normal3f(-0.647794902f, 0.761814773f, 0.f),
    C_Vertex3f(0.352826118f, -0.0958909988f, 0.f),
    C_Vertex3f(0.352826118f, -0.0958909988f, -1.f),
    C_Normal3f(-0.722021997f, 0.691870213f, 0.f),
    C_Vertex3f(0.36128521f, -0.0879926682f, 0.f),
    C_Vertex3f(0.36128521f, -0.0879926682f, -1.f),
    C_Normal3f(-0.791714311f, 0.610891461f, 0.f),
    C_Vertex3f(0.369014263f, -0.0789973736f, 0.f),
    C_Vertex3f(0.369014263f, -0.0789973736f, -1.f),
    C_Normal3f(-0.84858793f, 0.529054344f, 0.f),
    C_Vertex3f(0.376013279f, -0.0689051151f, 0.f),
    C_Vertex3f(0.376013279f, -0.0689051151f, -1.f),
    C_Normal3f(-0.893355846f, 0.44934994f, 0.f),
    C_Vertex3f(0.382282257f, -0.0577158928f, 0.f),
    C_Vertex3f(0.382282257f, -0.0577158928f, -1.f),
    C_Normal3f(-0.927475274f, 0.373884559f, 0.f),
    C_Vertex3f(0.387821198f, -0.0454297066f, 0.f),
    C_Vertex3f(0.387821198f, -0.0454297066f, -1.f),
    C_Normal3f(-0.952706397f, 0.303892374f, 0.f),
    C_Vertex3f(0.3926301f, -0.0320465565f, 0.f),
    C_Vertex3f(0.3926301f, -0.0320465565f, -1.f),
    C_Normal3f(-0.970797062f, 0.239902049f, 0.f),
    C_Vertex3f(0.396708965f, -0.0175664425f, 0.f),
    C_Vertex3f(0.396708965f, -0.0175664425f, -1.f),
    C_Normal3f(-0.981551766f, 0.191196263f, 0.f),
    C_Vertex3f(0.400057793f, -0.00198936462f, 0.f),
    C_Vertex3f(0.400057793f, -0.00198936462f, -1.f),
    C_Normal3f(-0.987889886f, 0.155155897f, 0.f),
    C_Vertex3f(0.401367188f, 0.00634765625f, 0.f),
    C_Vertex3f(0.401367188f, 0.00634765625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.11371538f, 0.993513346f, 0.f),
    C_Vertex3f(0.401367188f, 0.00634765625f, 0.f),
    C_Vertex3f(0.401367188f, 0.00634765625f, -1.f),
    C_Vertex3f(0.482421875f, -0.0029296875f, -1.f),
    C_Vertex3f(0.482421875f, -0.0029296875f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.992497981f, -0.122261159f, 0.f),
    C_Vertex3f(0.482421875f, -0.0029296875f, 0.f),
    C_Vertex3f(0.482421875f, -0.0029296875f, -1.f),
    C_Normal3f(0.985834599f, -0.16772069f, 0.f),
    C_Vertex3f(0.481054306f, -0.0140314102f, 0.f),
    C_Vertex3f(0.481054306f, -0.0140314102f, -1.f),
    C_Normal3f(0.973626614f, -0.228147238f, 0.f),
    C_Vertex3f(0.476971984f, -0.0349632502f, 0.f),
    C_Vertex3f(0.476971984f, -0.0349632502f, -1.f),
    C_Normal3f(0.952201903f, -0.305469394f, 0.f),
    C_Vertex3f(0.471542478f, -0.0546234846f, 0.f),
    C_Vertex3f(0.471542478f, -0.0546234846f, -1.f),
    C_Normal3f(0.922091544f, -0.386971772f, 0.f),
    C_Vertex3f(0.464765787f, -0.0730121136f, 0.f),
    C_Vertex3f(0.464765787f, -0.0730121136f, -1.f),
    C_Normal3f(0.882186532f, -0.470899999f, 0.f),
    C_Vertex3f(0.456641912f, -0.090129137f, 0.f),
    C_Vertex3f(0.456641912f, -0.090129137f, -1.f),
    C_Normal3f(0.831926346f, -0.554885983f, 0.f),
    C_Vertex3f(0.447170854f, -0.105974555f, 0.f),
    C_Vertex3f(0.447170854f, -0.105974555f, -1.f),
    C_Normal3f(0.771564066f, -0.636151552f, 0.f),
    C_Vertex3f(0.436352611f, -0.120548368f, 0.f),
    C_Vertex3f(0.436352611f, -0.120548368f, -1.f),
    C_Normal3f(0.702305138f, -0.711876035f, 0.f),
    C_Vertex3f(0.424187183f, -0.133850574f, 0.f),
    C_Vertex3f(0.424187183f, -0.133850574f, -1.f),
    C_Normal3f(0.626021147f, -0.779806137f, 0.f),
    C_Vertex3f(0.410674572f, -0.145881176f, 0.f),
    C_Vertex3f(0.410674572f, -0.145881176f, -1.f),
    C_Normal3f(0.542840242f, -0.839835882f, 0.f),
    C_Vertex3f(0.396088481f, -0.156407952f, 0.f),
    C_Vertex3f(0.396088481f, -0.156407952f, -1.f),
    C_Normal3f(0.454802513f, -0.890592277f, 0.f),
    C_Vertex3f(0.38042891f, -0.165430903f, 0.f),
    C_Vertex3f(0.38042891f, -0.165430903f, -1.f),
    C_Normal3f(0.364842027f, -0.931069374f, 0.f),
    C_Vertex3f(0.36369586f, -0.172950029f, 0.f),
    C_Vertex3f(0.36369586f, -0.172950029f, -1.f),
    C_Normal3f(0.275808781f, -0.961212575f, 0.f),
    C_Vertex3f(0.34588933f, -0.17896533f, 0.f),
    C_Vertex3f(0.34588933f, -0.17896533f, -1.f),
    C_Normal3f(0.190094158f, -0.981765926f, 0.f),
    C_Vertex3f(0.32700932f, -0.183476806f, 0.f),
    C_Vertex3f(0.32700932f, -0.183476806f, -1.f),
    C_Normal3f(0.109427363f, -0.993994832f, 0.f),
    C_Vertex3f(0.307055831f, -0.186484456f, 0.f),
    C_Vertex3f(0.307055831f, -0.186484456f, -1.f),
    C_Normal3f(0.0342409015f, -0.99941361f, 0.f),
    C_Vertex3f(0.286028862f, -0.187988281f, 0.f),
    C_Vertex3f(0.286028862f, -0.187988281f, -1.f),
    C_Normal3f(-0.0353158303f, -0.999376237f, 0.f),
    C_Vertex3f(0.263162613f, -0.187988281f, 0.f),
    C_Vertex3f(0.263162613f, -0.187988281f, -1.f),
    C_Normal3f(-0.109373637f, -0.994000673f, 0.f),
    C_Vertex3f(0.240740776f, -0.186387897f, 0.f),
    C_Vertex3f(0.240740776f, -0.186387897f, -1.f),
    C_Normal3f(-0.190600604f, -0.981667638f, 0.f),
    C_Vertex3f(0.219529152f, -0.183187127f, 0.f),
    C_Vertex3f(0.219529152f, -0.183187127f, -1.f),
    C_Normal3f(-0.277446836f, -0.960740983f, 0.f),
    C_Vertex3f(0.19952774f, -0.178385973f, 0.f),
    C_Vertex3f(0.19952774f, -0.178385973f, -1.f),
    C_Normal3f(-0.368184119f, -0.929752886f, 0.f),
    C_Vertex3f(0.180736542f, -0.171984434f, 0.f),
    C_Vertex3f(0.180736542f, -0.171984434f, -1.f),
    C_Normal3f(-0.460309207f, -0.887758553f, 0.f),
    C_Vertex3f(0.163155556f, -0.163982511f, 0.f),
    C_Vertex3f(0.163155556f, -0.163982511f, -1.f),
    C_Normal3f(-0.550737441f, -0.834678471f, 0.f),
    C_Vertex3f(0.146784782f, -0.154380202f, 0.f),
    C_Vertex3f(0.146784782f, -0.154380202f, -1.f),
    C_Normal3f(-0.636213779f, -0.771512866f, 0.f),
    C_Vertex3f(0.131624222f, -0.143177509f, 0.f),
    C_Vertex3f(0.131624222f, -0.143177509f, -1.f),
    C_Normal3f(-0.708717406f, -0.705492496f, 0.f),
    C_Vertex3f(0.117673874f, -0.130374432f, 0.f),
    C_Vertex3f(0.117673874f, -0.130374432f, -1.f),
    C_Normal3f(-0.772948027f, -0.63446939f, 0.f),
    C_Vertex3f(0.105467319f, -0.11690104f, 0.f),
    C_Vertex3f(0.105467319f, -0.11690104f, -1.f),
    C_Normal3f(-0.83368665f, -0.552237809f, 0.f),
    C_Vertex3f(0.0950045586f, -0.102757335f, 0.f),
    C_Vertex3f(0.0950045586f, -0.102757335f, -1.f),
    C_Normal3f(-0.88794744f, -0.459944904f, 0.f),
    C_Vertex3f(0.0862855911f, -0.0879433155f, 0.f),
    C_Vertex3f(0.0862855911f, -0.0879433155f, -1.f),
    C_Normal3f(-0.932972491f, -0.359947741f, 0.f),
    C_Vertex3f(0.0793104172f, -0.0724589825f, 0.f),
    C_Vertex3f(0.0793104172f, -0.0724589825f, -1.f),
    C_Normal3f(-0.966784358f, -0.255593151f, 0.f),
    C_Vertex3f(0.0740790367f, -0.0563043356f, 0.f),
    C_Vertex3f(0.0740790367f, -0.0563043356f, -1.f),
    C_Normal3f(-0.988581419f, -0.150687933f, 0.f),
    C_Vertex3f(0.0705914497f, -0.0394793749f, 0.f),
    C_Vertex3f(0.0705914497f, -0.0394793749f, -1.f),
    C_Normal3f(-0.998902619f, -0.0468342155f, 0.f),
    C_Vertex3f(0.0688476563f, -0.0219841003f, 0.f),
    C_Vertex3f(0.0688476563f, -0.0219841003f, -1.f),
    C_Normal3f(-0.998158693f, 0.0606556609f, 0.f),
    C_Vertex3f(0.0688476563f, -0.00228691101f, 0.f),
    C_Vertex3f(0.0688476563f, -0.00228691101f, -1.f),
    C_Normal3f(-0.983522475f, 0.180785969f, 0.f),
    C_Vertex3f(0.0712780952f, 0.01801157f, 0.f),
    C_Vertex3f(0.0712780952f, 0.01801157f, -1.f),
    C_Normal3f(-0.952026665f, 0.306015015f, 0.f),
    C_Vertex3f(0.0761389732f, 0.0373797417f, 0.f),
    C_Vertex3f(0.0761389732f, 0.0373797417f, -1.f),
    C_Normal3f(-0.907822013f, 0.419355541f, 0.f),
    C_Vertex3f(0.0834302902f, 0.0558176041f, 0.f),
    C_Vertex3f(0.0834302902f, 0.0558176041f, -1.f),
    C_Normal3f(-0.859058976f, 0.511876702f, 0.f),
    C_Vertex3f(0.0907216072f, 0.0689482689f, 0.f),
    C_Vertex3f(0.0907216072f, 0.0689482689f, -1.f),
    C_Normal3f(-0.819717467f, 0.572768152f, 0.f),
    C_Vertex3f(0.09676373f, 0.078194499f, 0.f),
    C_Vertex3f(0.09676373f, 0.078194499f, -1.f),
    C_Normal3f(-0.787515998f, 0.616294205f, 0.f),
    C_Vertex3f(0.103987098f, 0.0879331827f, 0.f),
    C_Vertex3f(0.103987098f, 0.0879331827f, -1.f),
    C_Normal3f(-0.758732617f, 0.651402235f, 0.f),
    C_Vertex3f(0.11239171f, 0.09816432f, 0.f),
    C_Vertex3f(0.11239171f, 0.09816432f, -1.f),
    C_Normal3f(-0.733119309f, 0.680100083f, 0.f),
    C_Vertex3f(0.121977568f, 0.108887911f, 0.f),
    C_Vertex3f(0.121977568f, 0.108887911f, -1.f),
    C_Normal3f(-0.710338891f, 0.703859925f, 0.f),
    C_Vertex3f(0.13274467f, 0.120103955f, 0.f),
    C_Vertex3f(0.13274467f, 0.120103955f, -1.f),
    C_Normal3f(-0.690043986f, 0.7237674f, 0.f),
    C_Vertex3f(0.144693017f, 0.131812453f, 0.f),
    C_Vertex3f(0.144693017f, 0.131812453f, -1.f),
    C_Normal3f(-0.672846079f, 0.739782453f, 0.f),
    C_Vertex3f(0.157822609f, 0.144013405f, 0.f),
    C_Vertex3f(0.157822609f, 0.144013405f, -1.f),
    C_Normal3f(-0.667757869f, 0.744378507f, 0.f),
    C_Vertex3f(0.169838905f, 0.154683113f, 0.f),
    C_Vertex3f(0.169838905f, 0.154683113f, -1.f),
    C_Normal3f(-0.67744714f, 0.735571384f, 0.f),
    C_Vertex3f(0.178714156f, 0.162754536f, 0.f),
    C_Vertex3f(0.178714156f, 0.162754536f, -1.f),
    C_Normal3f(-0.687962592f, 0.725746095f, 0.f),
    C_Vertex3f(0.186742902f, 0.170251369f, 0.f),
    C_Vertex3f(0.186742902f, 0.170251369f, -1.f),
    C_Normal3f(-0.700528443f, 0.713624477f, 0.f),
    C_Vertex3f(0.193925142f, 0.177173615f, 0.f),
    C_Vertex3f(0.193925142f, 0.177173615f, -1.f),
    C_Normal3f(-0.715785801f, 0.698319912f, 0.f),
    C_Vertex3f(0.200260878f, 0.183521271f, 0.f),
    C_Vertex3f(0.200260878f, 0.183521271f, -1.f),
    C_Normal3f(-0.734657466f, 0.678438246f, 0.f),
    C_Vertex3f(0.205750108f, 0.189294338f, 0.f),
    C_Vertex3f(0.205750108f, 0.189294338f, -1.f),
    C_Normal3f(-0.758501649f, 0.651671112f, 0.f),
    C_Vertex3f(0.210392833f, 0.194492817f, 0.f),
    C_Vertex3f(0.210392833f, 0.194492817f, -1.f),
    C_Normal3f(-0.792921841f, 0.609323442f, 0.f),
    C_Vertex3f(0.214189053f, 0.199116707f, 0.f),
    C_Vertex3f(0.214189053f, 0.199116707f, -1.f),
    C_Normal3f(-0.84030503f, 0.5421139f, 0.f),
    C_Vertex3f(0.218613625f, 0.205190659f, 0.f),
    C_Vertex3f(0.218613625f, 0.205190659f, -1.f),
    C_Normal3f(-0.88600409f, 0.463677198f, 0.f),
    C_Vertex3f(0.223720551f, 0.213891029f, 0.f),
    C_Vertex3f(0.223720551f, 0.213891029f, -1.f),
    C_Normal3f(-0.92589587f, 0.377778769f, 0.f),
    C_Vertex3f(0.228034973f, 0.223193169f, 0.f),
    C_Vertex3f(0.228034973f, 0.223193169f, -1.f),
    C_Normal3f(-0.956154048f, 0.292864323f, 0.f),
    C_Vertex3f(0.231556892f, 0.233097076f, 0.f),
    C_Vertex3f(0.231556892f, 0.233097076f, -1.f),
    C_Normal3f(-0.98036921f, 0.19717063f, 0.f),
    C_Vertex3f(0.234286308f, 0.243602753f, 0.f),
    C_Vertex3f(0.234286308f, 0.243602753f, -1.f),
    C_Normal3f(-0.992733955f, 0.120329998f, 0.f),
    C_Vertex3f(0.236443043f, 0.25739193f, 0.f),
    C_Vertex3f(0.236443043f, 0.25739193f, -1.f),
    C_Normal3f(-0.99760443f, 0.0691767633f, 0.f),
    C_Vertex3f(0.238027096f, 0.274464607f, 0.f),
    C_Vertex3f(0.238027096f, 0.274464607f, -1.f),
    C_Normal3f(-0.999269247f, 0.038221553f, 0.f),
    C_Vertex3f(0.239038467f, 0.294820786f, 0.f),
    C_Vertex3f(0.239038467f, 0.294820786f, -1.f),
    C_Normal3f(-0.999827802f, 0.0185541715f, 0.f),
    C_Vertex3f(0.239257813f, 0.306640625f, 0.f),
    C_Vertex3f(0.239257813f, 0.306640625f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.239257813f, 0.306640625f, 0.f),
    C_Vertex3f(0.239257813f, 0.306640625f, -1.f),
    C_Vertex3f(0.314941406f, 0.306640625f, -1.f),
    C_Vertex3f(0.314941406f, 0.306640625f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.239257813f, 0.464355469f, 0.f),
    C_Vertex3f(0.239257813f, 0.375f, 0.f),
    C_Vertex3f(0.3203125f, 0.375f, 0.f),
    C_Vertex3f(0.3203125f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310463905f, 0.229539871f, 0.f),
    C_Vertex3f(0.313738823f, 0.250005722f, 0.f),
    C_Vertex3f(0.314941406f, 0.306640625f, 0.f),
    C_Vertex3f(0.315376282f, 0.272277832f, 0.f),
    C_Vertex3f(0.315158844f, 0.299198151f, 0.f),
    C_Vertex3f(0.315376282f, 0.288036346f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.314941406f, 0.306640625f, 0.f),
    C_Vertex3f(0.239257813f, 0.306640625f, 0.f),
    C_Vertex3f(0.242218494f, 0.121308804f, 0.f),
    C_Vertex3f(0.256352901f, 0.134987354f, 0.f),
    C_Vertex3f(0.267349243f, 0.146532059f, 0.f),
    C_Vertex3f(0.276626587f, 0.157693863f, 0.f),
    C_Vertex3f(0.285849571f, 0.170552254f, 0.f),
    C_Vertex3f(0.293599129f, 0.183356285f, 0.f),
    C_Vertex3f(0.299875259f, 0.196105957f, 0.f),
    C_Vertex3f(0.302276611f, 0.202453613f, 0.f),
    C_Vertex3f(0.305551529f, 0.21088028f, 0.f),
    C_Vertex3f(0.310463905f, 0.229539871f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.242218494f, 0.121308804f, 0.f),
    C_Vertex3f(0.239257813f, 0.306640625f, 0.f),
    C_Vertex3f(0.239038467f, 0.294820786f, 0.f),
    C_Vertex3f(0.238027096f, 0.274464607f, 0.f),
    C_Vertex3f(0.236443043f, 0.25739193f, 0.f),
    C_Vertex3f(0.234286308f, 0.243602753f, 0.f),
    C_Vertex3f(0.231556892f, 0.233097076f, 0.f),
    C_Vertex3f(0.228034973f, 0.223193169f, 0.f),
    C_Vertex3f(0.224946022f, 0.105496407f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.224946022f, 0.105496407f, 0.f),
    C_Vertex3f(0.228034973f, 0.223193169f, 0.f),
    C_Vertex3f(0.223720551f, 0.213891029f, 0.f),
    C_Vertex3f(0.218613625f, 0.205190659f, 0.f),
    C_Vertex3f(0.214189053f, 0.199116707f, 0.f),
    C_Vertex3f(0.210392833f, 0.194492817f, 0.f),
    C_Vertex3f(0.209638119f, 0.0920367241f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.205750108f, 0.189294338f, 0.f),
    C_Vertex3f(0.200260878f, 0.183521271f, 0.f),
    C_Vertex3f(0.200333238f, 0.0835138559f, 0.f),
    C_Vertex3f(0.209638119f, 0.0920367241f, 0.f),
    C_Vertex3f(0.210392833f, 0.194492817f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172117591f, 0.0539249182f, 0.f),
    C_Vertex3f(0.177820921f, 0.0606467724f, 0.f),
    C_Vertex3f(0.178714156f, 0.162754536f, 0.f),
    C_Vertex3f(0.184424639f, 0.0678188801f, 0.f),
    C_Vertex3f(0.186742902f, 0.170251369f, 0.f),
    C_Vertex3f(0.191928744f, 0.0754412413f, 0.f),
    C_Vertex3f(0.193925142f, 0.177173615f, 0.f),
    C_Vertex3f(0.200333238f, 0.0835138559f, 0.f),
    C_Vertex3f(0.200260878f, 0.183521271f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169838905f, 0.154683113f, 0.f),
    C_Vertex3f(0.157822609f, 0.144013405f, 0.f),
    C_Vertex3f(0.158908844f, 0.0337753296f, 0.f),
    C_Vertex3f(0.163412094f, 0.0418319702f, 0.f),
    C_Vertex3f(0.167314649f, 0.0476533175f, 0.f),
    C_Vertex3f(0.172117591f, 0.0539249182f, 0.f),
    C_Vertex3f(0.178714156f, 0.162754536f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.157822609f, 0.144013405f, 0.f),
    C_Vertex3f(0.144693017f, 0.131812453f, 0.f),
    C_Vertex3f(0.149902344f, -0.00141525269f, 0.f),
    C_Vertex3f(0.151403427f, 0.0108089447f, 0.f),
    C_Vertex3f(0.154405594f, 0.0225391388f, 0.f),
    C_Vertex3f(0.158908844f, 0.0337753296f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.146784782f, -0.154380202f, 0.f),
    C_Vertex3f(0.163155556f, -0.163982511f, 0.f),
    C_Vertex3f(0.149902344f, -0.013584137f, 0.f),
    C_Vertex3f(0.149902344f, -0.00141525269f, 0.f),
    C_Vertex3f(0.144693017f, 0.131812453f, 0.f),
    C_Vertex3f(0.13274467f, 0.120103955f, 0.f),
    C_Vertex3f(0.131624222f, -0.143177509f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.103987098f, 0.0879331827f, 0.f),
    C_Vertex3f(0.105467319f, -0.11690104f, 0.f),
    C_Vertex3f(0.11239171f, 0.09816432f, 0.f),
    C_Vertex3f(0.117673874f, -0.130374432f, 0.f),
    C_Vertex3f(0.121977568f, 0.108887911f, 0.f),
    C_Vertex3f(0.131624222f, -0.143177509f, 0.f),
    C_Vertex3f(0.13274467f, 0.120103955f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0688476563f, -0.00228691101f, 0.f),
    C_Vertex3f(0.0705914497f, -0.0394793749f, 0.f),
    C_Vertex3f(0.0712780952f, 0.01801157f, 0.f),
    C_Vertex3f(0.0740790367f, -0.0563043356f, 0.f),
    C_Vertex3f(0.0761389732f, 0.0373797417f, 0.f),
    C_Vertex3f(0.0793104172f, -0.0724589825f, 0.f),
    C_Vertex3f(0.0834302902f, 0.0558176041f, 0.f),
    C_Vertex3f(0.0862855911f, -0.0879433155f, 0.f),
    C_Vertex3f(0.0907216072f, 0.0689482689f, 0.f),
    C_Vertex3f(0.0950045586f, -0.102757335f, 0.f),
    C_Vertex3f(0.09676373f, 0.078194499f, 0.f),
    C_Vertex3f(0.105467319f, -0.11690104f, 0.f),
    C_Vertex3f(0.103987098f, 0.0879331827f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.482421875f, -0.0029296875f, 0.f),
    C_Vertex3f(0.401367188f, 0.00634765625f, 0.f),
    C_Vertex3f(0.410674572f, -0.145881176f, 0.f),
    C_Vertex3f(0.424187183f, -0.133850574f, 0.f),
    C_Vertex3f(0.436352611f, -0.120548368f, 0.f),
    C_Vertex3f(0.447170854f, -0.105974555f, 0.f),
    C_Vertex3f(0.456641912f, -0.090129137f, 0.f),
    C_Vertex3f(0.464765787f, -0.0730121136f, 0.f),
    C_Vertex3f(0.471542478f, -0.0546234846f, 0.f),
    C_Vertex3f(0.476971984f, -0.0349632502f, 0.f),
    C_Vertex3f(0.481054306f, -0.0140314102f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.401367188f, 0.00634765625f, 0.f),
    C_Vertex3f(0.400057793f, -0.00198936462f, 0.f),
    C_Vertex3f(0.410674572f, -0.145881176f, 0.f),
    C_Vertex3f(0.396708965f, -0.0175664425f, 0.f),
    C_Vertex3f(0.396088481f, -0.156407952f, 0.f),
    C_Vertex3f(0.3926301f, -0.0320465565f, 0.f),
    C_Vertex3f(0.387821198f, -0.0454297066f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.382282257f, -0.0577158928f, 0.f),
    C_Vertex3f(0.376013279f, -0.0689051151f, 0.f),
    C_Vertex3f(0.38042891f, -0.165430903f, 0.f),
    C_Vertex3f(0.396088481f, -0.156407952f, 0.f),
    C_Vertex3f(0.387821198f, -0.0454297066f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.369014263f, -0.0789973736f, 0.f),
    C_Vertex3f(0.36128521f, -0.0879926682f, 0.f),
    C_Vertex3f(0.36369586f, -0.172950029f, 0.f),
    C_Vertex3f(0.38042891f, -0.165430903f, 0.f),
    C_Vertex3f(0.376013279f, -0.0689051151f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.352826118f, -0.0958909988f, 0.f),
    C_Vertex3f(0.343869209f, -0.102802038f, 0.f),
    C_Vertex3f(0.34588933f, -0.17896533f, 0.f),
    C_Vertex3f(0.36369586f, -0.172950029f, 0.f),
    C_Vertex3f(0.36128521f, -0.0879926682f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.334414482f, -0.108725786f, 0.f),
    C_Vertex3f(0.324461937f, -0.113662243f, 0.f),
    C_Vertex3f(0.32700932f, -0.183476806f, 0.f),
    C_Vertex3f(0.34588933f, -0.17896533f, 0.f),
    C_Vertex3f(0.343869209f, -0.102802038f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.314011574f, -0.117611408f, 0.f),
    C_Vertex3f(0.303063393f, -0.120573282f, 0.f),
    C_Vertex3f(0.307055831f, -0.186484456f, 0.f),
    C_Vertex3f(0.32700932f, -0.183476806f, 0.f),
    C_Vertex3f(0.324461937f, -0.113662243f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.291617393f, -0.122547865f, 0.f),
    C_Vertex3f(0.279673576f, -0.123535156f, 0.f),
    C_Vertex3f(0.286028862f, -0.187988281f, 0.f),
    C_Vertex3f(0.307055831f, -0.186484456f, 0.f),
    C_Vertex3f(0.303063393f, -0.120573282f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.266958237f, -0.123535156f, 0.f),
    C_Vertex3f(0.254474163f, -0.122452617f, 0.f),
    C_Vertex3f(0.263162613f, -0.187988281f, 0.f),
    C_Vertex3f(0.286028862f, -0.187988281f, 0.f),
    C_Vertex3f(0.279673576f, -0.123535156f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.24249506f, -0.120287538f, 0.f),
    C_Vertex3f(0.231020927f, -0.117039919f, 0.f),
    C_Vertex3f(0.240740776f, -0.186387897f, 0.f),
    C_Vertex3f(0.263162613f, -0.187988281f, 0.f),
    C_Vertex3f(0.254474163f, -0.122452617f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.220051765f, -0.112709761f, 0.f),
    C_Vertex3f(0.209587574f, -0.107297063f, 0.f),
    C_Vertex3f(0.219529152f, -0.183187127f, 0.f),
    C_Vertex3f(0.240740776f, -0.186387897f, 0.f),
    C_Vertex3f(0.231020927f, -0.117039919f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.199628353f, -0.100801826f, 0.f),
    C_Vertex3f(0.190174103f, -0.0932240486f, 0.f),
    C_Vertex3f(0.19952774f, -0.178385973f, 0.f),
    C_Vertex3f(0.219529152f, -0.183187127f, 0.f),
    C_Vertex3f(0.209587574f, -0.107297063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.181224823f, -0.0845637321f, 0.f),
    C_Vertex3f(0.173394203f, -0.0755335093f, 0.f),
    C_Vertex3f(0.180736542f, -0.171984434f, 0.f),
    C_Vertex3f(0.19952774f, -0.178385973f, 0.f),
    C_Vertex3f(0.190174103f, -0.0932240486f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.166682243f, -0.0661333799f, 0.f),
    C_Vertex3f(0.161088943f, -0.0563633442f, 0.f),
    C_Vertex3f(0.163155556f, -0.163982511f, 0.f),
    C_Vertex3f(0.180736542f, -0.171984434f, 0.f),
    C_Vertex3f(0.173394203f, -0.0755335093f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.163155556f, -0.163982511f, 0.f),
    C_Vertex3f(0.161088943f, -0.0563633442f, 0.f),
    C_Vertex3f(0.156614304f, -0.046223402f, 0.f),
    C_Vertex3f(0.153258324f, -0.0357135534f, 0.f),
    C_Vertex3f(0.151021004f, -0.0248337984f, 0.f),
    C_Vertex3f(0.149902344f, -0.013584137f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0705914497f, -0.0394793749f, 0.f),
    C_Vertex3f(0.0688476563f, -0.00228691101f, 0.f),
    C_Vertex3f(0.0688476563f, -0.0219841003f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.239257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.239257813f, 0.375f, -1.f),
    C_Vertex3f(0.3203125f, 0.375f, -1.f),
    C_Vertex3f(0.3203125f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310463905f, 0.229539871f, -1.f),
    C_Vertex3f(0.313738823f, 0.250005722f, -1.f),
    C_Vertex3f(0.314941406f, 0.306640625f, -1.f),
    C_Vertex3f(0.315376282f, 0.272277832f, -1.f),
    C_Vertex3f(0.315158844f, 0.299198151f, -1.f),
    C_Vertex3f(0.315376282f, 0.288036346f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.314941406f, 0.306640625f, -1.f),
    C_Vertex3f(0.239257813f, 0.306640625f, -1.f),
    C_Vertex3f(0.242218494f, 0.121308804f, -1.f),
    C_Vertex3f(0.256352901f, 0.134987354f, -1.f),
    C_Vertex3f(0.267349243f, 0.146532059f, -1.f),
    C_Vertex3f(0.276626587f, 0.157693863f, -1.f),
    C_Vertex3f(0.285849571f, 0.170552254f, -1.f),
    C_Vertex3f(0.293599129f, 0.183356285f, -1.f),
    C_Vertex3f(0.299875259f, 0.196105957f, -1.f),
    C_Vertex3f(0.302276611f, 0.202453613f, -1.f),
    C_Vertex3f(0.305551529f, 0.21088028f, -1.f),
    C_Vertex3f(0.310463905f, 0.229539871f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.242218494f, 0.121308804f, -1.f),
    C_Vertex3f(0.239257813f, 0.306640625f, -1.f),
    C_Vertex3f(0.239038467f, 0.294820786f, -1.f),
    C_Vertex3f(0.238027096f, 0.274464607f, -1.f),
    C_Vertex3f(0.236443043f, 0.25739193f, -1.f),
    C_Vertex3f(0.234286308f, 0.243602753f, -1.f),
    C_Vertex3f(0.231556892f, 0.233097076f, -1.f),
    C_Vertex3f(0.228034973f, 0.223193169f, -1.f),
    C_Vertex3f(0.224946022f, 0.105496407f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.224946022f, 0.105496407f, -1.f),
    C_Vertex3f(0.228034973f, 0.223193169f, -1.f),
    C_Vertex3f(0.223720551f, 0.213891029f, -1.f),
    C_Vertex3f(0.218613625f, 0.205190659f, -1.f),
    C_Vertex3f(0.214189053f, 0.199116707f, -1.f),
    C_Vertex3f(0.210392833f, 0.194492817f, -1.f),
    C_Vertex3f(0.209638119f, 0.0920367241f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.205750108f, 0.189294338f, -1.f),
    C_Vertex3f(0.200260878f, 0.183521271f, -1.f),
    C_Vertex3f(0.200333238f, 0.0835138559f, -1.f),
    C_Vertex3f(0.209638119f, 0.0920367241f, -1.f),
    C_Vertex3f(0.210392833f, 0.194492817f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172117591f, 0.0539249182f, -1.f),
    C_Vertex3f(0.177820921f, 0.0606467724f, -1.f),
    C_Vertex3f(0.178714156f, 0.162754536f, -1.f),
    C_Vertex3f(0.184424639f, 0.0678188801f, -1.f),
    C_Vertex3f(0.186742902f, 0.170251369f, -1.f),
    C_Vertex3f(0.191928744f, 0.0754412413f, -1.f),
    C_Vertex3f(0.193925142f, 0.177173615f, -1.f),
    C_Vertex3f(0.200333238f, 0.0835138559f, -1.f),
    C_Vertex3f(0.200260878f, 0.183521271f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169838905f, 0.154683113f, -1.f),
    C_Vertex3f(0.157822609f, 0.144013405f, -1.f),
    C_Vertex3f(0.158908844f, 0.0337753296f, -1.f),
    C_Vertex3f(0.163412094f, 0.0418319702f, -1.f),
    C_Vertex3f(0.167314649f, 0.0476533175f, -1.f),
    C_Vertex3f(0.172117591f, 0.0539249182f, -1.f),
    C_Vertex3f(0.178714156f, 0.162754536f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.157822609f, 0.144013405f, -1.f),
    C_Vertex3f(0.144693017f, 0.131812453f, -1.f),
    C_Vertex3f(0.149902344f, -0.00141525269f, -1.f),
    C_Vertex3f(0.151403427f, 0.0108089447f, -1.f),
    C_Vertex3f(0.154405594f, 0.0225391388f, -1.f),
    C_Vertex3f(0.158908844f, 0.0337753296f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.146784782f, -0.154380202f, -1.f),
    C_Vertex3f(0.163155556f, -0.163982511f, -1.f),
    C_Vertex3f(0.149902344f, -0.013584137f, -1.f),
    C_Vertex3f(0.149902344f, -0.00141525269f, -1.f),
    C_Vertex3f(0.144693017f, 0.131812453f, -1.f),
    C_Vertex3f(0.13274467f, 0.120103955f, -1.f),
    C_Vertex3f(0.131624222f, -0.143177509f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.103987098f, 0.0879331827f, -1.f),
    C_Vertex3f(0.105467319f, -0.11690104f, -1.f),
    C_Vertex3f(0.11239171f, 0.09816432f, -1.f),
    C_Vertex3f(0.117673874f, -0.130374432f, -1.f),
    C_Vertex3f(0.121977568f, 0.108887911f, -1.f),
    C_Vertex3f(0.131624222f, -0.143177509f, -1.f),
    C_Vertex3f(0.13274467f, 0.120103955f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0688476563f, -0.00228691101f, -1.f),
    C_Vertex3f(0.0705914497f, -0.0394793749f, -1.f),
    C_Vertex3f(0.0712780952f, 0.01801157f, -1.f),
    C_Vertex3f(0.0740790367f, -0.0563043356f, -1.f),
    C_Vertex3f(0.0761389732f, 0.0373797417f, -1.f),
    C_Vertex3f(0.0793104172f, -0.0724589825f, -1.f),
    C_Vertex3f(0.0834302902f, 0.0558176041f, -1.f),
    C_Vertex3f(0.0862855911f, -0.0879433155f, -1.f),
    C_Vertex3f(0.0907216072f, 0.0689482689f, -1.f),
    C_Vertex3f(0.0950045586f, -0.102757335f, -1.f),
    C_Vertex3f(0.09676373f, 0.078194499f, -1.f),
    C_Vertex3f(0.105467319f, -0.11690104f, -1.f),
    C_Vertex3f(0.103987098f, 0.0879331827f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.482421875f, -0.0029296875f, -1.f),
    C_Vertex3f(0.401367188f, 0.00634765625f, -1.f),
    C_Vertex3f(0.410674572f, -0.145881176f, -1.f),
    C_Vertex3f(0.424187183f, -0.133850574f, -1.f),
    C_Vertex3f(0.436352611f, -0.120548368f, -1.f),
    C_Vertex3f(0.447170854f, -0.105974555f, -1.f),
    C_Vertex3f(0.456641912f, -0.090129137f, -1.f),
    C_Vertex3f(0.464765787f, -0.0730121136f, -1.f),
    C_Vertex3f(0.471542478f, -0.0546234846f, -1.f),
    C_Vertex3f(0.476971984f, -0.0349632502f, -1.f),
    C_Vertex3f(0.481054306f, -0.0140314102f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.401367188f, 0.00634765625f, -1.f),
    C_Vertex3f(0.400057793f, -0.00198936462f, -1.f),
    C_Vertex3f(0.410674572f, -0.145881176f, -1.f),
    C_Vertex3f(0.396708965f, -0.0175664425f, -1.f),
    C_Vertex3f(0.396088481f, -0.156407952f, -1.f),
    C_Vertex3f(0.3926301f, -0.0320465565f, -1.f),
    C_Vertex3f(0.387821198f, -0.0454297066f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.382282257f, -0.0577158928f, -1.f),
    C_Vertex3f(0.376013279f, -0.0689051151f, -1.f),
    C_Vertex3f(0.38042891f, -0.165430903f, -1.f),
    C_Vertex3f(0.396088481f, -0.156407952f, -1.f),
    C_Vertex3f(0.387821198f, -0.0454297066f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.369014263f, -0.0789973736f, -1.f),
    C_Vertex3f(0.36128521f, -0.0879926682f, -1.f),
    C_Vertex3f(0.36369586f, -0.172950029f, -1.f),
    C_Vertex3f(0.38042891f, -0.165430903f, -1.f),
    C_Vertex3f(0.376013279f, -0.0689051151f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.352826118f, -0.0958909988f, -1.f),
    C_Vertex3f(0.343869209f, -0.102802038f, -1.f),
    C_Vertex3f(0.34588933f, -0.17896533f, -1.f),
    C_Vertex3f(0.36369586f, -0.172950029f, -1.f),
    C_Vertex3f(0.36128521f, -0.0879926682f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.334414482f, -0.108725786f, -1.f),
    C_Vertex3f(0.324461937f, -0.113662243f, -1.f),
    C_Vertex3f(0.32700932f, -0.183476806f, -1.f),
    C_Vertex3f(0.34588933f, -0.17896533f, -1.f),
    C_Vertex3f(0.343869209f, -0.102802038f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.314011574f, -0.117611408f, -1.f),
    C_Vertex3f(0.303063393f, -0.120573282f, -1.f),
    C_Vertex3f(0.307055831f, -0.186484456f, -1.f),
    C_Vertex3f(0.32700932f, -0.183476806f, -1.f),
    C_Vertex3f(0.324461937f, -0.113662243f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.291617393f, -0.122547865f, -1.f),
    C_Vertex3f(0.279673576f, -0.123535156f, -1.f),
    C_Vertex3f(0.286028862f, -0.187988281f, -1.f),
    C_Vertex3f(0.307055831f, -0.186484456f, -1.f),
    C_Vertex3f(0.303063393f, -0.120573282f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.266958237f, -0.123535156f, -1.f),
    C_Vertex3f(0.254474163f, -0.122452617f, -1.f),
    C_Vertex3f(0.263162613f, -0.187988281f, -1.f),
    C_Vertex3f(0.286028862f, -0.187988281f, -1.f),
    C_Vertex3f(0.279673576f, -0.123535156f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.24249506f, -0.120287538f, -1.f),
    C_Vertex3f(0.231020927f, -0.117039919f, -1.f),
    C_Vertex3f(0.240740776f, -0.186387897f, -1.f),
    C_Vertex3f(0.263162613f, -0.187988281f, -1.f),
    C_Vertex3f(0.254474163f, -0.122452617f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.220051765f, -0.112709761f, -1.f),
    C_Vertex3f(0.209587574f, -0.107297063f, -1.f),
    C_Vertex3f(0.219529152f, -0.183187127f, -1.f),
    C_Vertex3f(0.240740776f, -0.186387897f, -1.f),
    C_Vertex3f(0.231020927f, -0.117039919f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.199628353f, -0.100801826f, -1.f),
    C_Vertex3f(0.190174103f, -0.0932240486f, -1.f),
    C_Vertex3f(0.19952774f, -0.178385973f, -1.f),
    C_Vertex3f(0.219529152f, -0.183187127f, -1.f),
    C_Vertex3f(0.209587574f, -0.107297063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.181224823f, -0.0845637321f, -1.f),
    C_Vertex3f(0.173394203f, -0.0755335093f, -1.f),
    C_Vertex3f(0.180736542f, -0.171984434f, -1.f),
    C_Vertex3f(0.19952774f, -0.178385973f, -1.f),
    C_Vertex3f(0.190174103f, -0.0932240486f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.166682243f, -0.0661333799f, -1.f),
    C_Vertex3f(0.161088943f, -0.0563633442f, -1.f),
    C_Vertex3f(0.163155556f, -0.163982511f, -1.f),
    C_Vertex3f(0.180736542f, -0.171984434f, -1.f),
    C_Vertex3f(0.173394203f, -0.0755335093f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.163155556f, -0.163982511f, -1.f),
    C_Vertex3f(0.161088943f, -0.0563633442f, -1.f),
    C_Vertex3f(0.156614304f, -0.046223402f, -1.f),
    C_Vertex3f(0.153258324f, -0.0357135534f, -1.f),
    C_Vertex3f(0.151021004f, -0.0248337984f, -1.f),
    C_Vertex3f(0.149902344f, -0.013584137f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0705914497f, -0.0394793749f, -1.f),
    C_Vertex3f(0.0688476563f, -0.00228691101f, -1.f),
    C_Vertex3f(0.0688476563f, -0.0219841003f, -1.f),
    C_End(),
    C_Translatef(0.546875f, 0.f, 0.f),
};
static const Call char_192[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.933440149f, 0.358733237f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Normal3f(0.92576915f, 0.378089279f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Normal3f(-0.933976114f, -0.357335538f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Normal3f(0.940247476f, -0.340491831f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.935599327f, -0.353063881f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Normal3f(-0.935954869f, -0.352120042f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Normal3f(-0.940465152f, -0.339890033f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Normal3f(-0.942786753f, -0.333396405f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Normal3f(-0.945318818f, -0.326147795f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Normal3f(-0.948088706f, -0.318005919f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Normal3f(-0.951127946f, -0.30879733f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Normal3f(-0.954471946f, -0.298300862f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Normal3f(-0.957520068f, -0.288366556f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Normal3f(-0.9601475f, -0.279493541f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.97945708f, -0.201652929f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Normal3f(0.974358559f, -0.225001022f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_Normal3f(0.967271984f, -0.253741771f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Normal3f(0.957781792f, -0.287495911f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Normal3f(0.949109674f, -0.314945877f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.352539063f, 0.680175781f, 0.f),
    C_Vertex3f(0.352539063f, 0.680175781f, -1.f),
    C_Vertex3f(0.2890625f, 0.680175781f, -1.f),
    C_Vertex3f(0.2890625f, 0.680175781f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.2890625f, 0.680175781f, 0.f),
    C_Vertex3f(0.2890625f, 0.680175781f, -1.f),
    C_Vertex3f(0.187988281f, 0.802734375f, -1.f),
    C_Vertex3f(0.187988281f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.187988281f, 0.802734375f, 0.f),
    C_Vertex3f(0.187988281f, 0.802734375f, -1.f),
    C_Vertex3f(0.293457031f, 0.802734375f, -1.f),
    C_Vertex3f(0.293457031f, 0.802734375f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.293457031f, 0.802734375f, 0.f),
    C_Vertex3f(0.293457031f, 0.802734375f, -1.f),
    C_Vertex3f(0.352539063f, 0.680175781f, -1.f),
    C_Vertex3f(0.352539063f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.293457031f, 0.802734375f, 0.f),
    C_Vertex3f(0.187988281f, 0.802734375f, 0.f),
    C_Vertex3f(0.2890625f, 0.680175781f, 0.f),
    C_Vertex3f(0.352539063f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.293457031f, 0.802734375f, -1.f),
    C_Vertex3f(0.187988281f, 0.802734375f, -1.f),
    C_Vertex3f(0.2890625f, 0.680175781f, -1.f),
    C_Vertex3f(0.352539063f, 0.680175781f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_193[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.933440149f, 0.358733237f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Normal3f(0.92576915f, 0.378089279f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Normal3f(-0.933976114f, -0.357335538f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Normal3f(0.940247476f, -0.340491831f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.935599327f, -0.353063881f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Normal3f(-0.935954869f, -0.352120042f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Normal3f(-0.940465152f, -0.339890033f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Normal3f(-0.942786753f, -0.333396405f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Normal3f(-0.945318818f, -0.326147795f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Normal3f(-0.948088706f, -0.318005919f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Normal3f(-0.951127946f, -0.30879733f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Normal3f(-0.954471946f, -0.298300862f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Normal3f(-0.957520068f, -0.288366556f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Normal3f(-0.9601475f, -0.279493541f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.97945708f, -0.201652929f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Normal3f(0.974358559f, -0.225001022f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_Normal3f(0.967271984f, -0.253741771f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Normal3f(0.957781792f, -0.287495911f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Normal3f(0.949109674f, -0.314945877f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.224121094f, 0.680175781f, 0.f),
    C_Vertex3f(0.224121094f, 0.680175781f, -1.f),
    C_Vertex3f(0.281738281f, 0.802734375f, -1.f),
    C_Vertex3f(0.281738281f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.281738281f, 0.802734375f, 0.f),
    C_Vertex3f(0.281738281f, 0.802734375f, -1.f),
    C_Vertex3f(0.385253906f, 0.802734375f, -1.f),
    C_Vertex3f(0.385253906f, 0.802734375f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.385253906f, 0.802734375f, 0.f),
    C_Vertex3f(0.385253906f, 0.802734375f, -1.f),
    C_Vertex3f(0.2890625f, 0.680175781f, -1.f),
    C_Vertex3f(0.2890625f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.2890625f, 0.680175781f, 0.f),
    C_Vertex3f(0.2890625f, 0.680175781f, -1.f),
    C_Vertex3f(0.224121094f, 0.680175781f, -1.f),
    C_Vertex3f(0.224121094f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.281738281f, 0.802734375f, 0.f),
    C_Vertex3f(0.224121094f, 0.680175781f, 0.f),
    C_Vertex3f(0.2890625f, 0.680175781f, 0.f),
    C_Vertex3f(0.385253906f, 0.802734375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.281738281f, 0.802734375f, -1.f),
    C_Vertex3f(0.224121094f, 0.680175781f, -1.f),
    C_Vertex3f(0.2890625f, 0.680175781f, -1.f),
    C_Vertex3f(0.385253906f, 0.802734375f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_194[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.933440149f, 0.358733237f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Normal3f(0.92576915f, 0.378089279f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Normal3f(-0.933976114f, -0.357335538f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Normal3f(0.940247476f, -0.340491831f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.935599327f, -0.353063881f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Normal3f(-0.935954869f, -0.352120042f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Normal3f(-0.940465152f, -0.339890033f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Normal3f(-0.942786753f, -0.333396405f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Normal3f(-0.945318818f, -0.326147795f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Normal3f(-0.948088706f, -0.318005919f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Normal3f(-0.951127946f, -0.30879733f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Normal3f(-0.954471946f, -0.298300862f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Normal3f(-0.957520068f, -0.288366556f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Normal3f(-0.9601475f, -0.279493541f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.97945708f, -0.201652929f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Normal3f(0.974358559f, -0.225001022f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_Normal3f(0.967271984f, -0.253741771f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Normal3f(0.957781792f, -0.287495911f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Normal3f(0.949109674f, -0.314945877f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.289550781f, 0.754882813f, 0.f),
    C_Vertex3f(0.289550781f, 0.754882813f, -1.f),
    C_Vertex3f(0.240478516f, 0.680175781f, -1.f),
    C_Vertex3f(0.240478516f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.240478516f, 0.680175781f, 0.f),
    C_Vertex3f(0.240478516f, 0.680175781f, -1.f),
    C_Vertex3f(0.151123047f, 0.680175781f, -1.f),
    C_Vertex3f(0.151123047f, 0.680175781f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(0.151123047f, 0.680175781f, 0.f),
    C_Vertex3f(0.151123047f, 0.680175781f, -1.f),
    C_Vertex3f(0.244628906f, 0.802734375f, -1.f),
    C_Vertex3f(0.244628906f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.244628906f, 0.802734375f, 0.f),
    C_Vertex3f(0.244628906f, 0.802734375f, -1.f),
    C_Vertex3f(0.328613281f, 0.802734375f, -1.f),
    C_Vertex3f(0.328613281f, 0.802734375f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.328613281f, 0.802734375f, 0.f),
    C_Vertex3f(0.328613281f, 0.802734375f, -1.f),
    C_Vertex3f(0.427001953f, 0.680175781f, -1.f),
    C_Vertex3f(0.427001953f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.427001953f, 0.680175781f, 0.f),
    C_Vertex3f(0.427001953f, 0.680175781f, -1.f),
    C_Vertex3f(0.337646484f, 0.680175781f, -1.f),
    C_Vertex3f(0.337646484f, 0.680175781f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.337646484f, 0.680175781f, 0.f),
    C_Vertex3f(0.337646484f, 0.680175781f, -1.f),
    C_Vertex3f(0.289550781f, 0.754882813f, -1.f),
    C_Vertex3f(0.289550781f, 0.754882813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151123047f, 0.680175781f, 0.f),
    C_Vertex3f(0.240478516f, 0.680175781f, 0.f),
    C_Vertex3f(0.244628906f, 0.802734375f, 0.f),
    C_Vertex3f(0.289550781f, 0.754882813f, 0.f),
    C_Vertex3f(0.328613281f, 0.802734375f, 0.f),
    C_Vertex3f(0.337646484f, 0.680175781f, 0.f),
    C_Vertex3f(0.427001953f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151123047f, 0.680175781f, -1.f),
    C_Vertex3f(0.240478516f, 0.680175781f, -1.f),
    C_Vertex3f(0.244628906f, 0.802734375f, -1.f),
    C_Vertex3f(0.289550781f, 0.754882813f, -1.f),
    C_Vertex3f(0.328613281f, 0.802734375f, -1.f),
    C_Vertex3f(0.337646484f, 0.680175781f, -1.f),
    C_Vertex3f(0.427001953f, 0.680175781f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_195[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.933440149f, 0.358733237f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Normal3f(0.92576915f, 0.378089279f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Normal3f(-0.933976114f, -0.357335538f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Normal3f(0.940247476f, -0.340491831f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.935599327f, -0.353063881f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Normal3f(-0.935954869f, -0.352120042f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Normal3f(-0.940465152f, -0.339890033f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Normal3f(-0.942786753f, -0.333396405f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Normal3f(-0.945318818f, -0.326147795f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Normal3f(-0.948088706f, -0.318005919f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Normal3f(-0.951127946f, -0.30879733f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Normal3f(-0.954471946f, -0.298300862f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Normal3f(-0.957520068f, -0.288366556f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Normal3f(-0.9601475f, -0.279493541f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.97945708f, -0.201652929f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Normal3f(0.974358559f, -0.225001022f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_Normal3f(0.967271984f, -0.253741771f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Normal3f(0.957781792f, -0.287495911f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Normal3f(0.949109674f, -0.314945877f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999953806f, -0.00961818546f, 0.f),
    C_Vertex3f(0.146484375f, 0.678710938f, 0.f),
    C_Vertex3f(0.146484375f, 0.678710938f, -1.f),
    C_Normal3f(-0.999230683f, 0.0392165966f, 0.f),
    C_Vertex3f(0.146430016f, 0.684362411f, 0.f),
    C_Vertex3f(0.146430016f, 0.684362411f, -1.f),
    C_Normal3f(-0.994498312f, 0.104752719f, 0.f),
    C_Vertex3f(0.14712739f, 0.695094824f, 0.f),
    C_Vertex3f(0.14712739f, 0.695094824f, -1.f),
    C_Normal3f(-0.981874108f, 0.189534381f, 0.f),
    C_Vertex3f(0.148630857f, 0.705256701f, 0.f),
    C_Vertex3f(0.148630857f, 0.705256701f, -1.f),
    C_Normal3f(-0.960046053f, 0.279841781f, 0.f),
    C_Vertex3f(0.150940418f, 0.714848042f, 0.f),
    C_Vertex3f(0.150940418f, 0.714848042f, -1.f),
    C_Normal3f(-0.927577436f, 0.373630971f, 0.f),
    C_Vertex3f(0.154056072f, 0.723868847f, 0.f),
    C_Vertex3f(0.154056072f, 0.723868847f, -1.f),
    C_Normal3f(-0.883691847f, 0.468069077f, 0.f),
    C_Vertex3f(0.157977819f, 0.732319117f, 0.f),
    C_Vertex3f(0.157977819f, 0.732319117f, -1.f),
    C_Normal3f(-0.828614712f, 0.559819281f, 0.f),
    C_Vertex3f(0.16270566f, 0.740198851f, 0.f),
    C_Vertex3f(0.16270566f, 0.740198851f, -1.f),
    C_Normal3f(-0.759081304f, 0.650995851f, 0.f),
    C_Vertex3f(0.168239594f, 0.747508049f, 0.f),
    C_Vertex3f(0.168239594f, 0.747508049f, -1.f),
    C_Normal3f(-0.650275886f, 0.759698212f, 0.f),
    C_Vertex3f(0.177642822f, 0.757616043f, 0.f),
    C_Vertex3f(0.177642822f, 0.757616043f, -1.f),
    C_Normal3f(-0.487445921f, 0.87315321f, 0.f),
    C_Vertex3f(0.191857338f, 0.767724037f, 0.f),
    C_Vertex3f(0.191857338f, 0.767724037f, -1.f),
    C_Normal3f(-0.287504673f, 0.957779288f, 0.f),
    C_Vertex3f(0.207819939f, 0.7744627f, 0.f),
    C_Vertex3f(0.207819939f, 0.7744627f, -1.f),
    C_Normal3f(-0.0979905054f, 0.995187461f, 0.f),
    C_Vertex3f(0.225530624f, 0.777832031f, 0.f),
    C_Vertex3f(0.225530624f, 0.777832031f, -1.f),
    C_Normal3f(0.0889112428f, 0.99603951f, 0.f),
    C_Vertex3f(0.242038727f, 0.777832031f, 0.f),
    C_Vertex3f(0.242038727f, 0.777832031f, -1.f),
    C_Normal3f(0.244741365f, 0.969588399f, 0.f),
    C_Vertex3f(0.258062363f, 0.774928093f, 0.f),
    C_Vertex3f(0.258062363f, 0.774928093f, -1.f),
    C_Normal3f(0.345433027f, 0.938443422f, 0.f),
    C_Vertex3f(0.2765522f, 0.769120216f, 0.f),
    C_Vertex3f(0.2765522f, 0.769120216f, -1.f),
    C_Normal3f(0.413061887f, 0.910703003f, 0.f),
    C_Vertex3f(0.29750824f, 0.760408401f, 0.f),
    C_Vertex3f(0.29750824f, 0.760408401f, -1.f),
    C_Normal3f(0.42238301f, 0.906417429f, 0.f),
    C_Vertex3f(0.315675735f, 0.751375198f, 0.f),
    C_Vertex3f(0.315675735f, 0.751375198f, -1.f),
    C_Normal3f(0.345709145f, 0.938341737f, 0.f),
    C_Vertex3f(0.327275276f, 0.746537209f, 0.f),
    C_Vertex3f(0.327275276f, 0.746537209f, -1.f),
    C_Normal3f(0.243631572f, 0.969867826f, 0.f),
    C_Vertex3f(0.337561607f, 0.743311882f, 0.f),
    C_Vertex3f(0.337561607f, 0.743311882f, -1.f),
    C_Normal3f(0.0984259844f, 0.995144367f, 0.f),
    C_Vertex3f(0.346534729f, 0.741699219f, 0.f),
    C_Vertex3f(0.346534729f, 0.741699219f, -1.f),
    C_Normal3f(-0.070669286f, 0.997499824f, 0.f),
    C_Vertex3f(0.353866577f, 0.741699219f, 0.f),
    C_Vertex3f(0.353866577f, 0.741699219f, -1.f),
    C_Normal3f(-0.24143213f, 0.970417738f, 0.f),
    C_Vertex3f(0.360049725f, 0.742656708f, 0.f),
    C_Vertex3f(0.360049725f, 0.742656708f, -1.f),
    C_Normal3f(-0.435192645f, 0.900337279f, 0.f),
    C_Vertex3f(0.365412235f, 0.744571686f, 0.f),
    C_Vertex3f(0.365412235f, 0.744571686f, -1.f),
    C_Normal3f(-0.62994957f, 0.776636064f, 0.f),
    C_Vertex3f(0.369954109f, 0.747444153f, 0.f),
    C_Vertex3f(0.369954109f, 0.747444153f, -1.f),
    C_Normal3f(-0.810694158f, 0.585469842f, 0.f),
    C_Vertex3f(0.373675346f, 0.751274109f, 0.f),
    C_Vertex3f(0.373675346f, 0.751274109f, -1.f),
    C_Normal3f(-0.9168607f, 0.399207145f, 0.f),
    C_Vertex3f(0.376849651f, 0.75699234f, 0.f),
    C_Vertex3f(0.376849651f, 0.75699234f, -1.f),
    C_Normal3f(-0.964133918f, 0.265416414f, 0.f),
    C_Vertex3f(0.379477024f, 0.764598846f, 0.f),
    C_Vertex3f(0.379477024f, 0.764598846f, -1.f),
    C_Normal3f(-0.982875049f, 0.18427363f, 0.f),
    C_Vertex3f(0.381557465f, 0.774093628f, 0.f),
    C_Vertex3f(0.381557465f, 0.774093628f, -1.f),
    C_Normal3f(-0.991047204f, 0.133512393f, 0.f),
    C_Vertex3f(0.382324219f, 0.779785156f, 0.f),
    C_Vertex3f(0.382324219f, 0.779785156f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.382324219f, 0.779785156f, 0.f),
    C_Vertex3f(0.382324219f, 0.779785156f, -1.f),
    C_Vertex3f(0.438964844f, 0.779785156f, -1.f),
    C_Vertex3f(0.438964844f, 0.779785156f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999653041f, -0.0263388976f, 0.f),
    C_Vertex3f(0.438964844f, 0.779785156f, 0.f),
    C_Vertex3f(0.438964844f, 0.779785156f, -1.f),
    C_Normal3f(0.99777931f, -0.066607669f, 0.f),
    C_Vertex3f(0.438800812f, 0.77355957f, 0.f),
    C_Vertex3f(0.438800812f, 0.77355957f, -1.f),
    C_Normal3f(0.992581308f, -0.121582195f, 0.f),
    C_Vertex3f(0.437770009f, 0.761886597f, 0.f),
    C_Vertex3f(0.437770009f, 0.761886597f, -1.f),
    C_Normal3f(0.980870843f, -0.194659367f, 0.f),
    C_Vertex3f(0.436036468f, 0.750991821f, 0.f),
    C_Vertex3f(0.436036468f, 0.750991821f, -1.f),
    C_Normal3f(0.961305082f, -0.275485754f, 0.f),
    C_Vertex3f(0.433600187f, 0.740875244f, 0.f),
    C_Vertex3f(0.433600187f, 0.740875244f, -1.f),
    C_Normal3f(0.931648552f, -0.363360554f, 0.f),
    C_Vertex3f(0.430461168f, 0.731536865f, 0.f),
    C_Vertex3f(0.430461168f, 0.731536865f, -1.f),
    C_Normal3f(0.889692426f, -0.456560522f, 0.f),
    C_Vertex3f(0.426619411f, 0.722976685f, 0.f),
    C_Vertex3f(0.426619411f, 0.722976685f, -1.f),
    C_Normal3f(0.833749712f, -0.552142739f, 0.f),
    C_Vertex3f(0.422074914f, 0.715194702f, 0.f),
    C_Vertex3f(0.422074914f, 0.715194702f, -1.f),
    C_Normal3f(0.755483389f, -0.655167818f, 0.f),
    C_Vertex3f(0.416827679f, 0.708190918f, 0.f),
    C_Vertex3f(0.416827679f, 0.708190918f, -1.f),
    C_Normal3f(0.639779389f, -0.768558562f, 0.f),
    C_Vertex3f(0.407902718f, 0.698852539f, 0.f),
    C_Vertex3f(0.407902718f, 0.698852539f, -1.f),
    C_Normal3f(0.477566123f, -0.878595829f, 0.f),
    C_Vertex3f(0.394391537f, 0.68951416f, 0.f),
    C_Vertex3f(0.394391537f, 0.68951416f, -1.f),
    C_Normal3f(0.281305313f, -0.95961827f, 0.f),
    C_Vertex3f(0.379269123f, 0.683288574f, 0.f),
    C_Vertex3f(0.379269123f, 0.683288574f, -1.f),
    C_Normal3f(0.0945234597f, -0.995522618f, 0.f),
    C_Vertex3f(0.362535477f, 0.680175781f, 0.f),
    C_Vertex3f(0.362535477f, 0.680175781f, -1.f),
    C_Normal3f(-0.0947503f, -0.995501101f, 0.f),
    C_Vertex3f(0.346485138f, 0.680175781f, 0.f),
    C_Vertex3f(0.346485138f, 0.680175781f, -1.f),
    C_Normal3f(-0.258822501f, -0.965924919f, 0.f),
    C_Vertex3f(0.330532074f, 0.683221817f, 0.f),
    C_Vertex3f(0.330532074f, 0.683221817f, -1.f),
    C_Normal3f(-0.367869198f, -0.92987752f, 0.f),
    C_Vertex3f(0.312381744f, 0.689313889f, 0.f),
    C_Vertex3f(0.312381744f, 0.689313889f, -1.f),
    C_Normal3f(-0.441156626f, -0.897430122f, 0.f),
    C_Vertex3f(0.292034149f, 0.698451996f, 0.f),
    C_Vertex3f(0.292034149f, 0.698451996f, -1.f),
    C_Normal3f(-0.452606946f, -0.891710103f, 0.f),
    C_Vertex3f(0.273429871f, 0.708461761f, 0.f),
    C_Vertex3f(0.273429871f, 0.708461761f, -1.f),
    C_Normal3f(-0.385828793f, -0.922570407f, 0.f),
    C_Vertex3f(0.260735512f, 0.714338303f, 0.f),
    C_Vertex3f(0.260735512f, 0.714338303f, -1.f),
    C_Normal3f(-0.288805217f, -0.957387865f, 0.f),
    C_Vertex3f(0.25001049f, 0.718255997f, 0.f),
    C_Vertex3f(0.25001049f, 0.718255997f, -1.f),
    C_Normal3f(-0.122500464f, -0.992468476f, 0.f),
    C_Vertex3f(0.241254807f, 0.720214844f, 0.f),
    C_Vertex3f(0.241254807f, 0.720214844f, -1.f),
    C_Normal3f(0.0942201167f, -0.995551467f, 0.f),
    C_Vertex3f(0.234140396f, 0.720214844f, 0.f),
    C_Vertex3f(0.234140396f, 0.720214844f, -1.f),
    C_Normal3f(0.294618219f, -0.955615103f, 0.f),
    C_Vertex3f(0.227409363f, 0.718904495f, 0.f),
    C_Vertex3f(0.227409363f, 0.718904495f, -1.f),
    C_Normal3f(0.500668526f, -0.865639091f, 0.f),
    C_Vertex3f(0.221389771f, 0.716283798f, 0.f),
    C_Vertex3f(0.221389771f, 0.716283798f, -1.f),
    C_Normal3f(0.652620137f, -0.757685304f, 0.f),
    C_Vertex3f(0.216081619f, 0.712352753f, 0.f),
    C_Vertex3f(0.216081619f, 0.712352753f, -1.f),
    C_Normal3f(0.744249165f, -0.667902052f, 0.f),
    C_Vertex3f(0.213783264f, 0.709732056f, 0.f),
    C_Vertex3f(0.213783264f, 0.709732056f, -1.f),
    C_Normal3f(0.837037444f, -0.547145605f, 0.f),
    C_Vertex3f(0.21137619f, 0.707109451f, 0.f),
    C_Vertex3f(0.21137619f, 0.707109451f, -1.f),
    C_Normal3f(0.925027549f, -0.379899919f, 0.f),
    C_Vertex3f(0.207819939f, 0.700609207f, 0.f),
    C_Vertex3f(0.207819939f, 0.700609207f, -1.f),
    C_Normal3f(0.980741322f, -0.195311442f, 0.f),
    C_Vertex3f(0.205521584f, 0.692853928f, 0.f),
    C_Vertex3f(0.205521584f, 0.692853928f, -1.f),
    C_Normal3f(0.997837007f, -0.065737471f, 0.f),
    C_Vertex3f(0.204481125f, 0.683843613f, 0.f),
    C_Vertex3f(0.204481125f, 0.683843613f, -1.f),
    C_Normal3f(0.999775767f, 0.0211769659f, 0.f),
    C_Vertex3f(0.204589844f, 0.678710938f, 0.f),
    C_Vertex3f(0.204589844f, 0.678710938f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.204589844f, 0.678710938f, 0.f),
    C_Vertex3f(0.204589844f, 0.678710938f, -1.f),
    C_Vertex3f(0.146484375f, 0.678710938f, -1.f),
    C_Vertex3f(0.146484375f, 0.678710938f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.204589844f, 0.678710938f, 0.f),
    C_Vertex3f(0.204481125f, 0.683843613f, 0.f),
    C_Vertex3f(0.191857338f, 0.767724037f, 0.f),
    C_Vertex3f(0.177642822f, 0.757616043f, 0.f),
    C_Vertex3f(0.168239594f, 0.747508049f, 0.f),
    C_Vertex3f(0.16270566f, 0.740198851f, 0.f),
    C_Vertex3f(0.157977819f, 0.732319117f, 0.f),
    C_Vertex3f(0.154056072f, 0.723868847f, 0.f),
    C_Vertex3f(0.150940418f, 0.714848042f, 0.f),
    C_Vertex3f(0.148630857f, 0.705256701f, 0.f),
    C_Vertex3f(0.14712739f, 0.695094824f, 0.f),
    C_Vertex3f(0.146484375f, 0.678710938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.438964844f, 0.779785156f, 0.f),
    C_Vertex3f(0.382324219f, 0.779785156f, 0.f),
    C_Vertex3f(0.394391537f, 0.68951416f, 0.f),
    C_Vertex3f(0.407902718f, 0.698852539f, 0.f),
    C_Vertex3f(0.416827679f, 0.708190918f, 0.f),
    C_Vertex3f(0.422074914f, 0.715194702f, 0.f),
    C_Vertex3f(0.426619411f, 0.722976685f, 0.f),
    C_Vertex3f(0.430461168f, 0.731536865f, 0.f),
    C_Vertex3f(0.433600187f, 0.740875244f, 0.f),
    C_Vertex3f(0.436036468f, 0.750991821f, 0.f),
    C_Vertex3f(0.437770009f, 0.761886597f, 0.f),
    C_Vertex3f(0.438800812f, 0.77355957f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.382324219f, 0.779785156f, 0.f),
    C_Vertex3f(0.381557465f, 0.774093628f, 0.f),
    C_Vertex3f(0.394391537f, 0.68951416f, 0.f),
    C_Vertex3f(0.379477024f, 0.764598846f, 0.f),
    C_Vertex3f(0.379269123f, 0.683288574f, 0.f),
    C_Vertex3f(0.376849651f, 0.75699234f, 0.f),
    C_Vertex3f(0.373675346f, 0.751274109f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.373675346f, 0.751274109f, 0.f),
    C_Vertex3f(0.369954109f, 0.747444153f, 0.f),
    C_Vertex3f(0.379269123f, 0.683288574f, 0.f),
    C_Vertex3f(0.365412235f, 0.744571686f, 0.f),
    C_Vertex3f(0.362535477f, 0.680175781f, 0.f),
    C_Vertex3f(0.360049725f, 0.742656708f, 0.f),
    C_Vertex3f(0.353866577f, 0.741699219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.327275276f, 0.746537209f, 0.f),
    C_Vertex3f(0.330532074f, 0.683221817f, 0.f),
    C_Vertex3f(0.337561607f, 0.743311882f, 0.f),
    C_Vertex3f(0.346485138f, 0.680175781f, 0.f),
    C_Vertex3f(0.346534729f, 0.741699219f, 0.f),
    C_Vertex3f(0.362535477f, 0.680175781f, 0.f),
    C_Vertex3f(0.353866577f, 0.741699219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.260735512f, 0.714338303f, 0.f),
    C_Vertex3f(0.273429871f, 0.708461761f, 0.f),
    C_Vertex3f(0.2765522f, 0.769120216f, 0.f),
    C_Vertex3f(0.292034149f, 0.698451996f, 0.f),
    C_Vertex3f(0.29750824f, 0.760408401f, 0.f),
    C_Vertex3f(0.312381744f, 0.689313889f, 0.f),
    C_Vertex3f(0.315675735f, 0.751375198f, 0.f),
    C_Vertex3f(0.330532074f, 0.683221817f, 0.f),
    C_Vertex3f(0.327275276f, 0.746537209f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.234140396f, 0.720214844f, 0.f),
    C_Vertex3f(0.241254807f, 0.720214844f, 0.f),
    C_Vertex3f(0.242038727f, 0.777832031f, 0.f),
    C_Vertex3f(0.25001049f, 0.718255997f, 0.f),
    C_Vertex3f(0.258062363f, 0.774928093f, 0.f),
    C_Vertex3f(0.260735512f, 0.714338303f, 0.f),
    C_Vertex3f(0.2765522f, 0.769120216f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.216081619f, 0.712352753f, 0.f),
    C_Vertex3f(0.221389771f, 0.716283798f, 0.f),
    C_Vertex3f(0.225530624f, 0.777832031f, 0.f),
    C_Vertex3f(0.227409363f, 0.718904495f, 0.f),
    C_Vertex3f(0.242038727f, 0.777832031f, 0.f),
    C_Vertex3f(0.234140396f, 0.720214844f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.205521584f, 0.692853928f, 0.f),
    C_Vertex3f(0.207819939f, 0.700609207f, 0.f),
    C_Vertex3f(0.207819939f, 0.7744627f, 0.f),
    C_Vertex3f(0.21137619f, 0.707109451f, 0.f),
    C_Vertex3f(0.225530624f, 0.777832031f, 0.f),
    C_Vertex3f(0.213783264f, 0.709732056f, 0.f),
    C_Vertex3f(0.216081619f, 0.712352753f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.207819939f, 0.7744627f, 0.f),
    C_Vertex3f(0.191857338f, 0.767724037f, 0.f),
    C_Vertex3f(0.204481125f, 0.683843613f, 0.f),
    C_Vertex3f(0.205521584f, 0.692853928f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.146484375f, 0.678710938f, 0.f),
    C_Vertex3f(0.14712739f, 0.695094824f, 0.f),
    C_Vertex3f(0.146430016f, 0.684362411f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.204589844f, 0.678710938f, -1.f),
    C_Vertex3f(0.204481125f, 0.683843613f, -1.f),
    C_Vertex3f(0.191857338f, 0.767724037f, -1.f),
    C_Vertex3f(0.177642822f, 0.757616043f, -1.f),
    C_Vertex3f(0.168239594f, 0.747508049f, -1.f),
    C_Vertex3f(0.16270566f, 0.740198851f, -1.f),
    C_Vertex3f(0.157977819f, 0.732319117f, -1.f),
    C_Vertex3f(0.154056072f, 0.723868847f, -1.f),
    C_Vertex3f(0.150940418f, 0.714848042f, -1.f),
    C_Vertex3f(0.148630857f, 0.705256701f, -1.f),
    C_Vertex3f(0.14712739f, 0.695094824f, -1.f),
    C_Vertex3f(0.146484375f, 0.678710938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.438964844f, 0.779785156f, -1.f),
    C_Vertex3f(0.382324219f, 0.779785156f, -1.f),
    C_Vertex3f(0.394391537f, 0.68951416f, -1.f),
    C_Vertex3f(0.407902718f, 0.698852539f, -1.f),
    C_Vertex3f(0.416827679f, 0.708190918f, -1.f),
    C_Vertex3f(0.422074914f, 0.715194702f, -1.f),
    C_Vertex3f(0.426619411f, 0.722976685f, -1.f),
    C_Vertex3f(0.430461168f, 0.731536865f, -1.f),
    C_Vertex3f(0.433600187f, 0.740875244f, -1.f),
    C_Vertex3f(0.436036468f, 0.750991821f, -1.f),
    C_Vertex3f(0.437770009f, 0.761886597f, -1.f),
    C_Vertex3f(0.438800812f, 0.77355957f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.382324219f, 0.779785156f, -1.f),
    C_Vertex3f(0.381557465f, 0.774093628f, -1.f),
    C_Vertex3f(0.394391537f, 0.68951416f, -1.f),
    C_Vertex3f(0.379477024f, 0.764598846f, -1.f),
    C_Vertex3f(0.379269123f, 0.683288574f, -1.f),
    C_Vertex3f(0.376849651f, 0.75699234f, -1.f),
    C_Vertex3f(0.373675346f, 0.751274109f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.373675346f, 0.751274109f, -1.f),
    C_Vertex3f(0.369954109f, 0.747444153f, -1.f),
    C_Vertex3f(0.379269123f, 0.683288574f, -1.f),
    C_Vertex3f(0.365412235f, 0.744571686f, -1.f),
    C_Vertex3f(0.362535477f, 0.680175781f, -1.f),
    C_Vertex3f(0.360049725f, 0.742656708f, -1.f),
    C_Vertex3f(0.353866577f, 0.741699219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.327275276f, 0.746537209f, -1.f),
    C_Vertex3f(0.330532074f, 0.683221817f, -1.f),
    C_Vertex3f(0.337561607f, 0.743311882f, -1.f),
    C_Vertex3f(0.346485138f, 0.680175781f, -1.f),
    C_Vertex3f(0.346534729f, 0.741699219f, -1.f),
    C_Vertex3f(0.362535477f, 0.680175781f, -1.f),
    C_Vertex3f(0.353866577f, 0.741699219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.260735512f, 0.714338303f, -1.f),
    C_Vertex3f(0.273429871f, 0.708461761f, -1.f),
    C_Vertex3f(0.2765522f, 0.769120216f, -1.f),
    C_Vertex3f(0.292034149f, 0.698451996f, -1.f),
    C_Vertex3f(0.29750824f, 0.760408401f, -1.f),
    C_Vertex3f(0.312381744f, 0.689313889f, -1.f),
    C_Vertex3f(0.315675735f, 0.751375198f, -1.f),
    C_Vertex3f(0.330532074f, 0.683221817f, -1.f),
    C_Vertex3f(0.327275276f, 0.746537209f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.234140396f, 0.720214844f, -1.f),
    C_Vertex3f(0.241254807f, 0.720214844f, -1.f),
    C_Vertex3f(0.242038727f, 0.777832031f, -1.f),
    C_Vertex3f(0.25001049f, 0.718255997f, -1.f),
    C_Vertex3f(0.258062363f, 0.774928093f, -1.f),
    C_Vertex3f(0.260735512f, 0.714338303f, -1.f),
    C_Vertex3f(0.2765522f, 0.769120216f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.216081619f, 0.712352753f, -1.f),
    C_Vertex3f(0.221389771f, 0.716283798f, -1.f),
    C_Vertex3f(0.225530624f, 0.777832031f, -1.f),
    C_Vertex3f(0.227409363f, 0.718904495f, -1.f),
    C_Vertex3f(0.242038727f, 0.777832031f, -1.f),
    C_Vertex3f(0.234140396f, 0.720214844f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.205521584f, 0.692853928f, -1.f),
    C_Vertex3f(0.207819939f, 0.700609207f, -1.f),
    C_Vertex3f(0.207819939f, 0.7744627f, -1.f),
    C_Vertex3f(0.21137619f, 0.707109451f, -1.f),
    C_Vertex3f(0.225530624f, 0.777832031f, -1.f),
    C_Vertex3f(0.213783264f, 0.709732056f, -1.f),
    C_Vertex3f(0.216081619f, 0.712352753f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.207819939f, 0.7744627f, -1.f),
    C_Vertex3f(0.191857338f, 0.767724037f, -1.f),
    C_Vertex3f(0.204481125f, 0.683843613f, -1.f),
    C_Vertex3f(0.205521584f, 0.692853928f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.146484375f, 0.678710938f, -1.f),
    C_Vertex3f(0.14712739f, 0.695094824f, -1.f),
    C_Vertex3f(0.146430016f, 0.684362411f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_196[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.933440149f, 0.358733237f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Normal3f(0.92576915f, 0.378089279f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Normal3f(-0.933976114f, -0.357335538f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Normal3f(0.940247476f, -0.340491831f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.935599327f, -0.353063881f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Normal3f(-0.935954869f, -0.352120042f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Normal3f(-0.940465152f, -0.339890033f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Normal3f(-0.942786753f, -0.333396405f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Normal3f(-0.945318818f, -0.326147795f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Normal3f(-0.948088706f, -0.318005919f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Normal3f(-0.951127946f, -0.30879733f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Normal3f(-0.954471946f, -0.298300862f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Normal3f(-0.957520068f, -0.288366556f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Normal3f(-0.9601475f, -0.279493541f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.97945708f, -0.201652929f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Normal3f(0.974358559f, -0.225001022f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_Normal3f(0.967271984f, -0.253741771f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Normal3f(0.957781792f, -0.287495911f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Normal3f(0.949109674f, -0.314945877f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.169433594f, 0.680175781f, 0.f),
    C_Vertex3f(0.169433594f, 0.680175781f, -1.f),
    C_Vertex3f(0.169433594f, 0.76953125f, -1.f),
    C_Vertex3f(0.169433594f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.169433594f, 0.76953125f, 0.f),
    C_Vertex3f(0.169433594f, 0.76953125f, -1.f),
    C_Vertex3f(0.251464844f, 0.76953125f, -1.f),
    C_Vertex3f(0.251464844f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.251464844f, 0.76953125f, 0.f),
    C_Vertex3f(0.251464844f, 0.76953125f, -1.f),
    C_Vertex3f(0.251464844f, 0.680175781f, -1.f),
    C_Vertex3f(0.251464844f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.251464844f, 0.680175781f, 0.f),
    C_Vertex3f(0.251464844f, 0.680175781f, -1.f),
    C_Vertex3f(0.169433594f, 0.680175781f, -1.f),
    C_Vertex3f(0.169433594f, 0.680175781f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.332519531f, 0.680175781f, 0.f),
    C_Vertex3f(0.332519531f, 0.680175781f, -1.f),
    C_Vertex3f(0.332519531f, 0.76953125f, -1.f),
    C_Vertex3f(0.332519531f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.332519531f, 0.76953125f, 0.f),
    C_Vertex3f(0.332519531f, 0.76953125f, -1.f),
    C_Vertex3f(0.414550781f, 0.76953125f, -1.f),
    C_Vertex3f(0.414550781f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.414550781f, 0.76953125f, 0.f),
    C_Vertex3f(0.414550781f, 0.76953125f, -1.f),
    C_Vertex3f(0.414550781f, 0.680175781f, -1.f),
    C_Vertex3f(0.414550781f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.414550781f, 0.680175781f, 0.f),
    C_Vertex3f(0.414550781f, 0.680175781f, -1.f),
    C_Vertex3f(0.332519531f, 0.680175781f, -1.f),
    C_Vertex3f(0.332519531f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169433594f, 0.76953125f, 0.f),
    C_Vertex3f(0.169433594f, 0.680175781f, 0.f),
    C_Vertex3f(0.251464844f, 0.680175781f, 0.f),
    C_Vertex3f(0.251464844f, 0.76953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.332519531f, 0.76953125f, 0.f),
    C_Vertex3f(0.332519531f, 0.680175781f, 0.f),
    C_Vertex3f(0.414550781f, 0.680175781f, 0.f),
    C_Vertex3f(0.414550781f, 0.76953125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169433594f, 0.76953125f, -1.f),
    C_Vertex3f(0.169433594f, 0.680175781f, -1.f),
    C_Vertex3f(0.251464844f, 0.680175781f, -1.f),
    C_Vertex3f(0.251464844f, 0.76953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.332519531f, 0.76953125f, -1.f),
    C_Vertex3f(0.332519531f, 0.680175781f, -1.f),
    C_Vertex3f(0.414550781f, 0.680175781f, -1.f),
    C_Vertex3f(0.414550781f, 0.76953125f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_197[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.933440149f, 0.358733237f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Normal3f(0.92576915f, 0.378089279f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Normal3f(-0.933976114f, -0.357335538f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Normal3f(0.940247476f, -0.340491831f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.935599327f, -0.353063881f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Normal3f(-0.935954869f, -0.352120042f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Normal3f(-0.940465152f, -0.339890033f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Normal3f(-0.942786753f, -0.333396405f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Normal3f(-0.945318818f, -0.326147795f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Normal3f(-0.948088706f, -0.318005919f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Normal3f(-0.951127946f, -0.30879733f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Normal3f(-0.954471946f, -0.298300862f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Normal3f(-0.957520068f, -0.288366556f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Normal3f(-0.9601475f, -0.279493541f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.97945708f, -0.201652929f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Normal3f(0.974358559f, -0.225001022f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_Normal3f(0.967271984f, -0.253741771f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Normal3f(0.957781792f, -0.287495911f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Normal3f(0.949109674f, -0.314945877f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.93592304f, -0.352204621f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.996150136f, -0.0876631811f, 0.f),
    C_Vertex3f(0.209960938f, 0.680898666f, 0.f),
    C_Vertex3f(0.209960938f, 0.680898666f, -1.f),
    C_Normal3f(-0.995711446f, 0.0925136656f, 0.f),
    C_Vertex3f(0.209960938f, 0.697637558f, 0.f),
    C_Vertex3f(0.209960938f, 0.697637558f, -1.f),
    C_Normal3f(-0.955586612f, 0.294710487f, 0.f),
    C_Vertex3f(0.212890625f, 0.712430477f, 0.f),
    C_Vertex3f(0.212890625f, 0.712430477f, -1.f),
    C_Normal3f(-0.873809934f, 0.486267775f, 0.f),
    C_Vertex3f(0.21875f, 0.726135731f, 0.f),
    C_Vertex3f(0.21875f, 0.726135731f, -1.f),
    C_Normal3f(-0.762212932f, 0.64732635f, 0.f),
    C_Vertex3f(0.227539063f, 0.738753319f, 0.f),
    C_Vertex3f(0.227539063f, 0.738753319f, -1.f),
    C_Normal3f(-0.635637224f, 0.771987855f, 0.f),
    C_Vertex3f(0.239257813f, 0.750283241f, 0.f),
    C_Vertex3f(0.239257813f, 0.750283241f, -1.f),
    C_Normal3f(-0.475808382f, 0.879548907f, 0.f),
    C_Vertex3f(0.252044678f, 0.758930683f, 0.f),
    C_Vertex3f(0.252044678f, 0.758930683f, -1.f),
    C_Normal3f(-0.287775874f, 0.957697809f, 0.f),
    C_Vertex3f(0.265899658f, 0.764695644f, 0.f),
    C_Vertex3f(0.265899658f, 0.764695644f, -1.f),
    C_Normal3f(-0.0926625952f, 0.995697558f, 0.f),
    C_Vertex3f(0.280822754f, 0.767578125f, 0.f),
    C_Vertex3f(0.280822754f, 0.767578125f, -1.f),
    C_Normal3f(0.0936417803f, 0.995605886f, 0.f),
    C_Vertex3f(0.296873093f, 0.767578125f, 0.f),
    C_Vertex3f(0.296873093f, 0.767578125f, -1.f),
    C_Normal3f(0.290848225f, 0.956769228f, 0.f),
    C_Vertex3f(0.311839581f, 0.764660835f, 0.f),
    C_Vertex3f(0.311839581f, 0.764660835f, -1.f),
    C_Normal3f(0.482152909f, 0.876087129f, 0.f),
    C_Vertex3f(0.32566309f, 0.758826256f, 0.f),
    C_Vertex3f(0.32566309f, 0.758826256f, -1.f),
    C_Normal3f(0.644631505f, 0.764493287f, 0.f),
    C_Vertex3f(0.33834362f, 0.750074387f, 0.f),
    C_Vertex3f(0.33834362f, 0.750074387f, -1.f),
    C_Normal3f(0.77254498f, 0.634960055f, 0.f),
    C_Vertex3f(0.349881172f, 0.738405228f, 0.f),
    C_Vertex3f(0.349881172f, 0.738405228f, -1.f),
    C_Normal3f(0.882203341f, 0.470868587f, 0.f),
    C_Vertex3f(0.358534336f, 0.72550869f, 0.f),
    C_Vertex3f(0.358534336f, 0.72550869f, -1.f),
    C_Normal3f(0.959506512f, 0.281686425f, 0.f),
    C_Vertex3f(0.364303112f, 0.711384773f, 0.f),
    C_Vertex3f(0.364303112f, 0.711384773f, -1.f),
    C_Normal3f(0.995984793f, 0.0895228982f, 0.f),
    C_Vertex3f(0.3671875f, 0.696033478f, 0.f),
    C_Vertex3f(0.3671875f, 0.696033478f, -1.f),
    C_Normal3f(0.996016741f, -0.0891657025f, 0.f),
    C_Vertex3f(0.3671875f, 0.679294586f, 0.f),
    C_Vertex3f(0.3671875f, 0.679294586f, -1.f),
    C_Normal3f(0.960216522f, -0.279256493f, 0.f),
    C_Vertex3f(0.364294052f, 0.663712502f, 0.f),
    C_Vertex3f(0.364294052f, 0.663712502f, -1.f),
    C_Normal3f(0.882977486f, -0.469415396f, 0.f),
    C_Vertex3f(0.358507156f, 0.649447441f, 0.f),
    C_Vertex3f(0.358507156f, 0.649447441f, -1.f),
    C_Normal3f(0.771736562f, -0.63594234f, 0.f),
    C_Vertex3f(0.349826813f, 0.636499405f, 0.f),
    C_Vertex3f(0.349826813f, 0.636499405f, -1.f),
    C_Normal3f(0.643063486f, -0.765812874f, 0.f),
    C_Vertex3f(0.338253021f, 0.624868393f, 0.f),
    C_Vertex3f(0.338253021f, 0.624868393f, -1.f),
    C_Normal3f(0.482068062f, -0.8761338f, 0.f),
    C_Vertex3f(0.325587273f, 0.616145134f, 0.f),
    C_Vertex3f(0.325587273f, 0.616145134f, -1.f),
    C_Normal3f(0.291671664f, -0.956518531f, 0.f),
    C_Vertex3f(0.311829567f, 0.610329628f, 0.f),
    C_Vertex3f(0.311829567f, 0.610329628f, -1.f),
    C_Normal3f(0.0933625251f, -0.995632172f, 0.f),
    C_Vertex3f(0.296979904f, 0.607421875f, 0.f),
    C_Vertex3f(0.296979904f, 0.607421875f, -1.f),
    C_Normal3f(-0.0934955925f, -0.995619774f, 0.f),
    C_Vertex3f(0.280820847f, 0.607421875f, 0.f),
    C_Vertex3f(0.280820847f, 0.607421875f, -1.f),
    C_Normal3f(-0.289828807f, -0.957078516f, 0.f),
    C_Vertex3f(0.265645027f, 0.610364437f, 0.f),
    C_Vertex3f(0.265645027f, 0.610364437f, -1.f),
    C_Normal3f(-0.481931925f, -0.876208663f, 0.f),
    C_Vertex3f(0.251669884f, 0.616249561f, 0.f),
    C_Vertex3f(0.251669884f, 0.616249561f, -1.f),
    C_Normal3f(-0.645861149f, -0.763454914f, 0.f),
    C_Vertex3f(0.238895416f, 0.625077248f, 0.f),
    C_Vertex3f(0.238895416f, 0.625077248f, -1.f),
    C_Normal3f(-0.776966989f, -0.629541397f, 0.f),
    C_Vertex3f(0.227321625f, 0.636847496f, 0.f),
    C_Vertex3f(0.227321625f, 0.636847496f, -1.f),
    C_Normal3f(-0.88781929f, -0.460192323f, 0.f),
    C_Vertex3f(0.218641281f, 0.650074482f, 0.f),
    C_Vertex3f(0.218641281f, 0.650074482f, -1.f),
    C_Normal3f(-0.962561071f, -0.271065116f, 0.f),
    C_Vertex3f(0.212854385f, 0.664758205f, 0.f),
    C_Vertex3f(0.212854385f, 0.664758205f, -1.f),
    C_Normal3f(-0.996150136f, -0.0876631811f, 0.f),
    C_Vertex3f(0.209960938f, 0.680898666f, 0.f),
    C_Vertex3f(0.209960938f, 0.680898666f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.99585247f, -0.0909829065f, 0.f),
    C_Vertex3f(0.243164063f, 0.693906784f, 0.f),
    C_Vertex3f(0.243164063f, 0.693906784f, -1.f),
    C_Normal3f(0.996069252f, 0.0885776132f, 0.f),
    C_Vertex3f(0.243164063f, 0.683971405f, 0.f),
    C_Vertex3f(0.243164063f, 0.683971405f, -1.f),
    C_Normal3f(0.961418629f, 0.275089741f, 0.f),
    C_Vertex3f(0.244887352f, 0.674528122f, 0.f),
    C_Vertex3f(0.244887352f, 0.674528122f, -1.f),
    C_Normal3f(0.885625243f, 0.464400709f, 0.f),
    C_Vertex3f(0.248333931f, 0.665903091f, 0.f),
    C_Vertex3f(0.248333931f, 0.665903091f, -1.f),
    C_Normal3f(0.77504015f, 0.631911993f, 0.f),
    C_Vertex3f(0.253503799f, 0.658096313f, 0.f),
    C_Vertex3f(0.253503799f, 0.658096313f, -1.f),
    C_Normal3f(0.646952569f, 0.762530267f, 0.f),
    C_Vertex3f(0.260396957f, 0.651107788f, 0.f),
    C_Vertex3f(0.260396957f, 0.651107788f, -1.f),
    C_Normal3f(0.486883491f, 0.873466849f, 0.f),
    C_Vertex3f(0.267918587f, 0.645866394f, 0.f),
    C_Vertex3f(0.267918587f, 0.645866394f, -1.f),
    C_Normal3f(0.295764416f, 0.955260932f, 0.f),
    C_Vertex3f(0.276068687f, 0.642372131f, 0.f),
    C_Vertex3f(0.276068687f, 0.642372131f, -1.f),
    C_Normal3f(0.0956318378f, 0.99541682f, 0.f),
    C_Vertex3f(0.28484726f, 0.640625f, 0.f),
    C_Vertex3f(0.28484726f, 0.640625f, -1.f),
    C_Normal3f(-0.0956318378f, 0.99541682f, 0.f),
    C_Vertex3f(0.294254303f, 0.640625f, 0.f),
    C_Vertex3f(0.294254303f, 0.640625f, -1.f),
    C_Normal3f(-0.295764416f, 0.955260932f, 0.f),
    C_Vertex3f(0.303032875f, 0.642372131f, 0.f),
    C_Vertex3f(0.303032875f, 0.642372131f, -1.f),
    C_Normal3f(-0.486883491f, 0.873466849f, 0.f),
    C_Vertex3f(0.311182976f, 0.645866394f, 0.f),
    C_Vertex3f(0.311182976f, 0.645866394f, -1.f),
    C_Normal3f(-0.646952569f, 0.762530267f, 0.f),
    C_Vertex3f(0.318704605f, 0.651107788f, 0.f),
    C_Vertex3f(0.318704605f, 0.651107788f, -1.f),
    C_Normal3f(-0.772750914f, 0.634709477f, 0.f),
    C_Vertex3f(0.325597763f, 0.658096313f, 0.f),
    C_Vertex3f(0.325597763f, 0.658096313f, -1.f),
    C_Normal3f(-0.881743789f, 0.471728742f, 0.f),
    C_Vertex3f(0.330767632f, 0.665794373f, 0.f),
    C_Vertex3f(0.330767632f, 0.665794373f, -1.f),
    C_Normal3f(-0.959135056f, 0.282949001f, 0.f),
    C_Vertex3f(0.334214211f, 0.674201965f, 0.f),
    C_Vertex3f(0.334214211f, 0.674201965f, -1.f),
    C_Normal3f(-0.995887816f, 0.0905945078f, 0.f),
    C_Vertex3f(0.3359375f, 0.683319092f, 0.f),
    C_Vertex3f(0.3359375f, 0.683319092f, -1.f),
    C_Normal3f(-0.996017992f, -0.0891522393f, 0.f),
    C_Vertex3f(0.3359375f, 0.693145752f, 0.f),
    C_Vertex3f(0.3359375f, 0.693145752f, -1.f),
    C_Normal3f(-0.960359752f, -0.278763533f, 0.f),
    C_Vertex3f(0.334241867f, 0.702262878f, 0.f),
    C_Vertex3f(0.334241867f, 0.702262878f, -1.f),
    C_Normal3f(-0.884884119f, -0.465811223f, 0.f),
    C_Vertex3f(0.330850601f, 0.710670471f, 0.f),
    C_Vertex3f(0.330850601f, 0.710670471f, -1.f),
    C_Normal3f(-0.777755082f, -0.628567517f, 0.f),
    C_Vertex3f(0.325763702f, 0.71836853f, 0.f),
    C_Vertex3f(0.325763702f, 0.71836853f, -1.f),
    C_Normal3f(-0.650574088f, -0.759442806f, 0.f),
    C_Vertex3f(0.318981171f, 0.725357056f, 0.f),
    C_Vertex3f(0.318981171f, 0.725357056f, -1.f),
    C_Normal3f(-0.486228704f, -0.87383157f, 0.f),
    C_Vertex3f(0.311487198f, 0.73059845f, 0.f),
    C_Vertex3f(0.311487198f, 0.73059845f, -1.f),
    C_Normal3f(-0.292708308f, -0.956201792f, 0.f),
    C_Vertex3f(0.303281784f, 0.734092712f, 0.f),
    C_Vertex3f(0.303281784f, 0.734092712f, -1.f),
    C_Normal3f(-0.0943521634f, -0.99553895f, 0.f),
    C_Vertex3f(0.294364929f, 0.735839844f, 0.f),
    C_Vertex3f(0.294364929f, 0.735839844f, -1.f),
    C_Normal3f(0.0935930908f, -0.995610476f, 0.f),
    C_Vertex3f(0.28484726f, 0.735839844f, 0.f),
    C_Vertex3f(0.28484726f, 0.735839844f, -1.f),
    C_Normal3f(0.291557223f, -0.9565534f, 0.f),
    C_Vertex3f(0.276068687f, 0.734119892f, 0.f),
    C_Vertex3f(0.276068687f, 0.734119892f, -1.f),
    C_Normal3f(0.481072754f, -0.876680732f, 0.f),
    C_Vertex3f(0.267918587f, 0.730679989f, 0.f),
    C_Vertex3f(0.267918587f, 0.730679989f, -1.f),
    C_Normal3f(0.641043007f, -0.767504871f, 0.f),
    C_Vertex3f(0.260396957f, 0.725520134f, 0.f),
    C_Vertex3f(0.260396957f, 0.725520134f, -1.f),
    C_Normal3f(0.767487586f, -0.641063869f, 0.f),
    C_Vertex3f(0.253503799f, 0.718640327f, 0.f),
    C_Vertex3f(0.253503799f, 0.718640327f, -1.f),
    C_Normal3f(0.878021121f, -0.478622019f, 0.f),
    C_Vertex3f(0.248333931f, 0.711078167f, 0.f),
    C_Vertex3f(0.248333931f, 0.711078167f, -1.f),
    C_Normal3f(0.957542598f, -0.288291812f, 0.f),
    C_Vertex3f(0.244887352f, 0.702833652f, 0.f),
    C_Vertex3f(0.244887352f, 0.702833652f, -1.f),
    C_Normal3f(0.99585247f, -0.0909829065f, 0.f),
    C_Vertex3f(0.243164063f, 0.693906784f, 0.f),
    C_Vertex3f(0.243164063f, 0.693906784f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, 0.f),
    C_Vertex3f(0.338253021f, 0.624868393f, 0.f),
    C_Vertex3f(0.341849983f, 0.628483117f, 0.f),
    C_Vertex3f(0.33834362f, 0.750074387f, 0.f),
    C_Vertex3f(0.3359375f, 0.693145752f, 0.f),
    C_Vertex3f(0.3359375f, 0.683319092f, 0.f),
    C_Vertex3f(0.334214211f, 0.674201965f, 0.f),
    C_Vertex3f(0.330767632f, 0.665794373f, 0.f),
    C_Vertex3f(0.325597763f, 0.658096313f, 0.f),
    C_Vertex3f(0.318704605f, 0.651107788f, 0.f),
    C_Vertex3f(0.311182976f, 0.645866394f, 0.f),
    C_Vertex3f(0.303032875f, 0.642372131f, 0.f),
    C_Vertex3f(0.294254303f, 0.640625f, 0.f),
    C_Vertex3f(0.296979904f, 0.607421875f, 0.f),
    C_Vertex3f(0.311829567f, 0.610329628f, 0.f),
    C_Vertex3f(0.325587273f, 0.616145134f, 0.f),
    C_Vertex3f(0.338253021f, 0.624868393f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.296979904f, 0.607421875f, 0.f),
    C_Vertex3f(0.294254303f, 0.640625f, 0.f),
    C_Vertex3f(0.28484726f, 0.640625f, 0.f),
    C_Vertex3f(0.280820847f, 0.607421875f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_Vertex3f(0.290698051f, 0.566784859f, 0.f),
    C_Vertex3f(0.294928789f, 0.552952528f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.311829567f, 0.610329628f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.209960938f, 0.697637558f, 0.f),
    C_Vertex3f(0.212854385f, 0.664758205f, 0.f),
    C_Vertex3f(0.212890625f, 0.712430477f, 0.f),
    C_Vertex3f(0.218641281f, 0.650074482f, 0.f),
    C_Vertex3f(0.21875f, 0.726135731f, 0.f),
    C_Vertex3f(0.227321625f, 0.636847496f, 0.f),
    C_Vertex3f(0.227539063f, 0.738753319f, 0.f),
    C_Vertex3f(0.238895416f, 0.625077248f, 0.f),
    C_Vertex3f(0.239257813f, 0.750283241f, 0.f),
    C_Vertex3f(0.239096269f, 0.624938428f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.251669884f, 0.616249561f, 0.f),
    C_Vertex3f(0.28484726f, 0.640625f, 0.f),
    C_Vertex3f(0.265645027f, 0.610364437f, 0.f),
    C_Vertex3f(0.280820847f, 0.607421875f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.341849983f, 0.628483117f, 0.f),
    C_Vertex3f(0.349826813f, 0.636499405f, 0.f),
    C_Vertex3f(0.349881172f, 0.738405228f, 0.f),
    C_Vertex3f(0.358507156f, 0.649447441f, 0.f),
    C_Vertex3f(0.358534336f, 0.72550869f, 0.f),
    C_Vertex3f(0.364294052f, 0.663712502f, 0.f),
    C_Vertex3f(0.364303112f, 0.711384773f, 0.f),
    C_Vertex3f(0.3671875f, 0.679294586f, 0.f),
    C_Vertex3f(0.3671875f, 0.696033478f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33834362f, 0.750074387f, 0.f),
    C_Vertex3f(0.32566309f, 0.758826256f, 0.f),
    C_Vertex3f(0.325763702f, 0.71836853f, 0.f),
    C_Vertex3f(0.330850601f, 0.710670471f, 0.f),
    C_Vertex3f(0.334241867f, 0.702262878f, 0.f),
    C_Vertex3f(0.3359375f, 0.693145752f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.303281784f, 0.734092712f, 0.f),
    C_Vertex3f(0.311487198f, 0.73059845f, 0.f),
    C_Vertex3f(0.311839581f, 0.764660835f, 0.f),
    C_Vertex3f(0.318981171f, 0.725357056f, 0.f),
    C_Vertex3f(0.32566309f, 0.758826256f, 0.f),
    C_Vertex3f(0.325763702f, 0.71836853f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.296873093f, 0.767578125f, 0.f),
    C_Vertex3f(0.280822754f, 0.767578125f, 0.f),
    C_Vertex3f(0.28484726f, 0.735839844f, 0.f),
    C_Vertex3f(0.294364929f, 0.735839844f, 0.f),
    C_Vertex3f(0.303281784f, 0.734092712f, 0.f),
    C_Vertex3f(0.311839581f, 0.764660835f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253503799f, 0.718640327f, 0.f),
    C_Vertex3f(0.260396957f, 0.725520134f, 0.f),
    C_Vertex3f(0.265899658f, 0.764695644f, 0.f),
    C_Vertex3f(0.267918587f, 0.730679989f, 0.f),
    C_Vertex3f(0.280822754f, 0.767578125f, 0.f),
    C_Vertex3f(0.276068687f, 0.734119892f, 0.f),
    C_Vertex3f(0.28484726f, 0.735839844f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.252044678f, 0.758930683f, 0.f),
    C_Vertex3f(0.239257813f, 0.750283241f, 0.f),
    C_Vertex3f(0.243164063f, 0.693906784f, 0.f),
    C_Vertex3f(0.244887352f, 0.702833652f, 0.f),
    C_Vertex3f(0.248333931f, 0.711078167f, 0.f),
    C_Vertex3f(0.253503799f, 0.718640327f, 0.f),
    C_Vertex3f(0.265899658f, 0.764695644f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.248333931f, 0.665903091f, 0.f),
    C_Vertex3f(0.244887352f, 0.674528122f, 0.f),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.243164063f, 0.683971405f, 0.f),
    C_Vertex3f(0.239257813f, 0.750283241f, 0.f),
    C_Vertex3f(0.243164063f, 0.693906784f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.427787781f, 0.193847656f, 0.f),
    C_Vertex3f(0.501953125f, 0.f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.323152781f, 0.470025778f, 0.f),
    C_Vertex3f(0.325587273f, 0.616145134f, 0.f),
    C_Vertex3f(0.330211639f, 0.450687408f, 0.f),
    C_Vertex3f(0.338253021f, 0.624868393f, 0.f),
    C_Vertex3f(0.401153564f, 0.262695313f, 0.f),
    C_Vertex3f(0.341849983f, 0.628483117f, 0.f),
    C_Vertex3f(0.598526001f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.311829567f, 0.610329628f, 0.f),
    C_Vertex3f(0.29963088f, 0.538202524f, 0.f),
    C_Vertex3f(0.304804325f, 0.522534847f, 0.f),
    C_Vertex3f(0.310449123f, 0.505949497f, 0.f),
    C_Vertex3f(0.316565275f, 0.488446474f, 0.f),
    C_Vertex3f(0.325587273f, 0.616145134f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.280820847f, 0.607421875f, 0.f),
    C_Vertex3f(0.278186798f, 0.527202606f, 0.f),
    C_Vertex3f(0.285654068f, 0.557872772f, 0.f),
    C_Vertex3f(0.288818359f, 0.573242188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.265645027f, 0.610364437f, 0.f),
    C_Vertex3f(0.251669884f, 0.616249561f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.259836197f, 0.466068268f, 0.f),
    C_Vertex3f(0.269580841f, 0.496601105f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.239096269f, 0.624938428f, 0.f),
    C_Vertex3f(-0.00107574463f, 0.f, 0.f),
    C_Vertex3f(0.0888671875f, 0.f, 0.f),
    C_Vertex3f(0.159065247f, 0.193847656f, 0.f),
    C_Vertex3f(0.18359375f, 0.262695313f, 0.f),
    C_Vertex3f(0.254394531f, 0.450836182f, 0.f),
    C_Vertex3f(0.251669884f, 0.616249561f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, 0.f),
    C_Vertex3f(0.28484726f, 0.640625f, 0.f),
    C_Vertex3f(0.276068687f, 0.642372131f, 0.f),
    C_Vertex3f(0.267918587f, 0.645866394f, 0.f),
    C_Vertex3f(0.260396957f, 0.651107788f, 0.f),
    C_Vertex3f(0.253503799f, 0.658096313f, 0.f),
    C_Vertex3f(0.248333931f, 0.665903091f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.212854385f, 0.664758205f, 0.f),
    C_Vertex3f(0.209960938f, 0.697637558f, 0.f),
    C_Vertex3f(0.209960938f, 0.680898666f, 0.f),
    C_Vertex3f(0.33834362f, 0.750074387f, 0.f),
    C_Vertex3f(0.341849983f, 0.628483117f, 0.f),
    C_Vertex3f(0.349881172f, 0.738405228f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.336891174f, 0.640625f, -1.f),
    C_Vertex3f(0.338253021f, 0.624868393f, -1.f),
    C_Vertex3f(0.341849983f, 0.628483117f, -1.f),
    C_Vertex3f(0.33834362f, 0.750074387f, -1.f),
    C_Vertex3f(0.3359375f, 0.693145752f, -1.f),
    C_Vertex3f(0.3359375f, 0.683319092f, -1.f),
    C_Vertex3f(0.334214211f, 0.674201965f, -1.f),
    C_Vertex3f(0.330767632f, 0.665794373f, -1.f),
    C_Vertex3f(0.325597763f, 0.658096313f, -1.f),
    C_Vertex3f(0.318704605f, 0.651107788f, -1.f),
    C_Vertex3f(0.311182976f, 0.645866394f, -1.f),
    C_Vertex3f(0.303032875f, 0.642372131f, -1.f),
    C_Vertex3f(0.294254303f, 0.640625f, -1.f),
    C_Vertex3f(0.296979904f, 0.607421875f, -1.f),
    C_Vertex3f(0.311829567f, 0.610329628f, -1.f),
    C_Vertex3f(0.325587273f, 0.616145134f, -1.f),
    C_Vertex3f(0.338253021f, 0.624868393f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.296979904f, 0.607421875f, -1.f),
    C_Vertex3f(0.294254303f, 0.640625f, -1.f),
    C_Vertex3f(0.28484726f, 0.640625f, -1.f),
    C_Vertex3f(0.280820847f, 0.607421875f, -1.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_Vertex3f(0.290698051f, 0.566784859f, -1.f),
    C_Vertex3f(0.294928789f, 0.552952528f, -1.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Vertex3f(0.311829567f, 0.610329628f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.209960938f, 0.697637558f, -1.f),
    C_Vertex3f(0.212854385f, 0.664758205f, -1.f),
    C_Vertex3f(0.212890625f, 0.712430477f, -1.f),
    C_Vertex3f(0.218641281f, 0.650074482f, -1.f),
    C_Vertex3f(0.21875f, 0.726135731f, -1.f),
    C_Vertex3f(0.227321625f, 0.636847496f, -1.f),
    C_Vertex3f(0.227539063f, 0.738753319f, -1.f),
    C_Vertex3f(0.238895416f, 0.625077248f, -1.f),
    C_Vertex3f(0.239257813f, 0.750283241f, -1.f),
    C_Vertex3f(0.239096269f, 0.624938428f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.251669884f, 0.616249561f, -1.f),
    C_Vertex3f(0.28484726f, 0.640625f, -1.f),
    C_Vertex3f(0.265645027f, 0.610364437f, -1.f),
    C_Vertex3f(0.280820847f, 0.607421875f, -1.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.341849983f, 0.628483117f, -1.f),
    C_Vertex3f(0.349826813f, 0.636499405f, -1.f),
    C_Vertex3f(0.349881172f, 0.738405228f, -1.f),
    C_Vertex3f(0.358507156f, 0.649447441f, -1.f),
    C_Vertex3f(0.358534336f, 0.72550869f, -1.f),
    C_Vertex3f(0.364294052f, 0.663712502f, -1.f),
    C_Vertex3f(0.364303112f, 0.711384773f, -1.f),
    C_Vertex3f(0.3671875f, 0.679294586f, -1.f),
    C_Vertex3f(0.3671875f, 0.696033478f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33834362f, 0.750074387f, -1.f),
    C_Vertex3f(0.32566309f, 0.758826256f, -1.f),
    C_Vertex3f(0.325763702f, 0.71836853f, -1.f),
    C_Vertex3f(0.330850601f, 0.710670471f, -1.f),
    C_Vertex3f(0.334241867f, 0.702262878f, -1.f),
    C_Vertex3f(0.3359375f, 0.693145752f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.303281784f, 0.734092712f, -1.f),
    C_Vertex3f(0.311487198f, 0.73059845f, -1.f),
    C_Vertex3f(0.311839581f, 0.764660835f, -1.f),
    C_Vertex3f(0.318981171f, 0.725357056f, -1.f),
    C_Vertex3f(0.32566309f, 0.758826256f, -1.f),
    C_Vertex3f(0.325763702f, 0.71836853f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.296873093f, 0.767578125f, -1.f),
    C_Vertex3f(0.280822754f, 0.767578125f, -1.f),
    C_Vertex3f(0.28484726f, 0.735839844f, -1.f),
    C_Vertex3f(0.294364929f, 0.735839844f, -1.f),
    C_Vertex3f(0.303281784f, 0.734092712f, -1.f),
    C_Vertex3f(0.311839581f, 0.764660835f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253503799f, 0.718640327f, -1.f),
    C_Vertex3f(0.260396957f, 0.725520134f, -1.f),
    C_Vertex3f(0.265899658f, 0.764695644f, -1.f),
    C_Vertex3f(0.267918587f, 0.730679989f, -1.f),
    C_Vertex3f(0.280822754f, 0.767578125f, -1.f),
    C_Vertex3f(0.276068687f, 0.734119892f, -1.f),
    C_Vertex3f(0.28484726f, 0.735839844f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.252044678f, 0.758930683f, -1.f),
    C_Vertex3f(0.239257813f, 0.750283241f, -1.f),
    C_Vertex3f(0.243164063f, 0.693906784f, -1.f),
    C_Vertex3f(0.244887352f, 0.702833652f, -1.f),
    C_Vertex3f(0.248333931f, 0.711078167f, -1.f),
    C_Vertex3f(0.253503799f, 0.718640327f, -1.f),
    C_Vertex3f(0.265899658f, 0.764695644f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.248333931f, 0.665903091f, -1.f),
    C_Vertex3f(0.244887352f, 0.674528122f, -1.f),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.243164063f, 0.683971405f, -1.f),
    C_Vertex3f(0.239257813f, 0.750283241f, -1.f),
    C_Vertex3f(0.243164063f, 0.693906784f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.427787781f, 0.193847656f, -1.f),
    C_Vertex3f(0.501953125f, 0.f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Vertex3f(0.323152781f, 0.470025778f, -1.f),
    C_Vertex3f(0.325587273f, 0.616145134f, -1.f),
    C_Vertex3f(0.330211639f, 0.450687408f, -1.f),
    C_Vertex3f(0.338253021f, 0.624868393f, -1.f),
    C_Vertex3f(0.401153564f, 0.262695313f, -1.f),
    C_Vertex3f(0.341849983f, 0.628483117f, -1.f),
    C_Vertex3f(0.598526001f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.311829567f, 0.610329628f, -1.f),
    C_Vertex3f(0.29963088f, 0.538202524f, -1.f),
    C_Vertex3f(0.304804325f, 0.522534847f, -1.f),
    C_Vertex3f(0.310449123f, 0.505949497f, -1.f),
    C_Vertex3f(0.316565275f, 0.488446474f, -1.f),
    C_Vertex3f(0.325587273f, 0.616145134f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.280820847f, 0.607421875f, -1.f),
    C_Vertex3f(0.278186798f, 0.527202606f, -1.f),
    C_Vertex3f(0.285654068f, 0.557872772f, -1.f),
    C_Vertex3f(0.288818359f, 0.573242188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.265645027f, 0.610364437f, -1.f),
    C_Vertex3f(0.251669884f, 0.616249561f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.259836197f, 0.466068268f, -1.f),
    C_Vertex3f(0.269580841f, 0.496601105f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.239096269f, 0.624938428f, -1.f),
    C_Vertex3f(-0.00107574463f, 0.f, -1.f),
    C_Vertex3f(0.0888671875f, 0.f, -1.f),
    C_Vertex3f(0.159065247f, 0.193847656f, -1.f),
    C_Vertex3f(0.18359375f, 0.262695313f, -1.f),
    C_Vertex3f(0.254394531f, 0.450836182f, -1.f),
    C_Vertex3f(0.251669884f, 0.616249561f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245124817f, 0.640625f, -1.f),
    C_Vertex3f(0.28484726f, 0.640625f, -1.f),
    C_Vertex3f(0.276068687f, 0.642372131f, -1.f),
    C_Vertex3f(0.267918587f, 0.645866394f, -1.f),
    C_Vertex3f(0.260396957f, 0.651107788f, -1.f),
    C_Vertex3f(0.253503799f, 0.658096313f, -1.f),
    C_Vertex3f(0.248333931f, 0.665903091f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.212854385f, 0.664758205f, -1.f),
    C_Vertex3f(0.209960938f, 0.697637558f, -1.f),
    C_Vertex3f(0.209960938f, 0.680898666f, -1.f),
    C_Vertex3f(0.33834362f, 0.750074387f, -1.f),
    C_Vertex3f(0.341849983f, 0.628483117f, -1.f),
    C_Vertex3f(0.349881172f, 0.738405228f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_198[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.901203096f, 0.433397025f, 0.f),
    C_Vertex3f(0.000244140625f, 0.f, 0.f),
    C_Vertex3f(0.000244140625f, 0.f, -1.f),
    C_Vertex3f(0.308326721f, 0.640625f, -1.f),
    C_Vertex3f(0.308326721f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.308326721f, 0.640625f, 0.f),
    C_Vertex3f(0.308326721f, 0.640625f, -1.f),
    C_Vertex3f(0.834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.834960938f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.834960938f, 0.564941406f, -1.f),
    C_Vertex3f(0.834960938f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.834960938f, 0.564941406f, 0.f),
    C_Vertex3f(0.834960938f, 0.564941406f, -1.f),
    C_Vertex3f(0.51171875f, 0.564941406f, -1.f),
    C_Vertex3f(0.51171875f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.51171875f, 0.564941406f, 0.f),
    C_Vertex3f(0.51171875f, 0.564941406f, -1.f),
    C_Vertex3f(0.51171875f, 0.368652344f, -1.f),
    C_Vertex3f(0.51171875f, 0.368652344f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.51171875f, 0.368652344f, 0.f),
    C_Vertex3f(0.51171875f, 0.368652344f, -1.f),
    C_Vertex3f(0.812011719f, 0.368652344f, -1.f),
    C_Vertex3f(0.812011719f, 0.368652344f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.812011719f, 0.368652344f, 0.f),
    C_Vertex3f(0.812011719f, 0.368652344f, -1.f),
    C_Vertex3f(0.812011719f, 0.29296875f, -1.f),
    C_Vertex3f(0.812011719f, 0.29296875f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.812011719f, 0.29296875f, 0.f),
    C_Vertex3f(0.812011719f, 0.29296875f, -1.f),
    C_Vertex3f(0.51171875f, 0.29296875f, -1.f),
    C_Vertex3f(0.51171875f, 0.29296875f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.51171875f, 0.29296875f, 0.f),
    C_Vertex3f(0.51171875f, 0.29296875f, -1.f),
    C_Vertex3f(0.51171875f, 0.0756835938f, -1.f),
    C_Vertex3f(0.51171875f, 0.0756835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.51171875f, 0.0756835938f, 0.f),
    C_Vertex3f(0.51171875f, 0.0756835938f, -1.f),
    C_Vertex3f(0.846679688f, 0.0756835938f, -1.f),
    C_Vertex3f(0.846679688f, 0.0756835938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.846679688f, 0.0756835938f, 0.f),
    C_Vertex3f(0.846679688f, 0.0756835938f, -1.f),
    C_Vertex3f(0.846679688f, 0.f, -1.f),
    C_Vertex3f(0.846679688f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.846679688f, 0.f, 0.f),
    C_Vertex3f(0.846679688f, 0.f, -1.f),
    C_Vertex3f(0.426757813f, 0.f, -1.f),
    C_Vertex3f(0.426757813f, 0.f, 0.f),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.426757813f, 0.f, 0.f),
    C_Vertex3f(0.426757813f, 0.f, -1.f),
    C_Vertex3f(0.426757813f, 0.184570313f, -1.f),
    C_Vertex3f(0.426757813f, 0.184570313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.426757813f, 0.184570313f, 0.f),
    C_Vertex3f(0.426757813f, 0.184570313f, -1.f),
    C_Vertex3f(0.180786133f, 0.184570313f, -1.f),
    C_Vertex3f(0.180786133f, 0.184570313f, 0.f),
    C_Normal3f(0.901190817f, -0.433422416f, 0.f),
    C_Vertex3f(0.180786133f, 0.184570313f, 0.f),
    C_Vertex3f(0.180786133f, 0.184570313f, -1.f),
    C_Vertex3f(0.0920181274f, 0.f, -1.f),
    C_Vertex3f(0.0920181274f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0920181274f, 0.f, 0.f),
    C_Vertex3f(0.0920181274f, 0.f, -1.f),
    C_Vertex3f(0.000244140625f, 0.f, -1.f),
    C_Vertex3f(0.000244140625f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.217185974f, 0.260253906f, 0.f),
    C_Vertex3f(0.217185974f, 0.260253906f, -1.f),
    C_Vertex3f(0.426757813f, 0.260253906f, -1.f),
    C_Vertex3f(0.426757813f, 0.260253906f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.426757813f, 0.260253906f, 0.f),
    C_Vertex3f(0.426757813f, 0.260253906f, -1.f),
    C_Vertex3f(0.426757813f, 0.564941406f, -1.f),
    C_Vertex3f(0.426757813f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.426757813f, 0.564941406f, 0.f),
    C_Vertex3f(0.426757813f, 0.564941406f, -1.f),
    C_Vertex3f(0.363716125f, 0.564941406f, -1.f),
    C_Vertex3f(0.363716125f, 0.564941406f, 0.f),
    C_Normal3f(0.901199579f, -0.433404416f, 0.f),
    C_Vertex3f(0.363716125f, 0.564941406f, 0.f),
    C_Vertex3f(0.363716125f, 0.564941406f, -1.f),
    C_Vertex3f(0.217185974f, 0.260253906f, -1.f),
    C_Vertex3f(0.217185974f, 0.260253906f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.834960938f, 0.564941406f, 0.f),
    C_Vertex3f(0.834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.51171875f, 0.564941406f, 0.f),
    C_Vertex3f(0.426757813f, 0.564941406f, 0.f),
    C_Vertex3f(0.51171875f, 0.0756835938f, 0.f),
    C_Vertex3f(0.426757813f, 0.260253906f, 0.f),
    C_Vertex3f(0.426757813f, 0.184570313f, 0.f),
    C_Vertex3f(0.217185974f, 0.260253906f, 0.f),
    C_Vertex3f(0.180786133f, 0.184570313f, 0.f),
    C_Vertex3f(0.308326721f, 0.640625f, 0.f),
    C_Vertex3f(0.0920181274f, 0.f, 0.f),
    C_Vertex3f(0.000244140625f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.51171875f, 0.29296875f, 0.f),
    C_Vertex3f(0.812011719f, 0.29296875f, 0.f),
    C_Vertex3f(0.51171875f, 0.368652344f, 0.f),
    C_Vertex3f(0.51171875f, 0.564941406f, 0.f),
    C_Vertex3f(0.51171875f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.363716125f, 0.564941406f, 0.f),
    C_Vertex3f(0.426757813f, 0.564941406f, 0.f),
    C_Vertex3f(0.834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.308326721f, 0.640625f, 0.f),
    C_Vertex3f(0.217185974f, 0.260253906f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.51171875f, 0.0756835938f, 0.f),
    C_Vertex3f(0.426757813f, 0.184570313f, 0.f),
    C_Vertex3f(0.426757813f, 0.f, 0.f),
    C_Vertex3f(0.846679688f, 0.f, 0.f),
    C_Vertex3f(0.846679688f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.51171875f, 0.368652344f, 0.f),
    C_Vertex3f(0.812011719f, 0.29296875f, 0.f),
    C_Vertex3f(0.812011719f, 0.368652344f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.834960938f, 0.564941406f, -1.f),
    C_Vertex3f(0.834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.51171875f, 0.564941406f, -1.f),
    C_Vertex3f(0.426757813f, 0.564941406f, -1.f),
    C_Vertex3f(0.51171875f, 0.0756835938f, -1.f),
    C_Vertex3f(0.426757813f, 0.260253906f, -1.f),
    C_Vertex3f(0.426757813f, 0.184570313f, -1.f),
    C_Vertex3f(0.217185974f, 0.260253906f, -1.f),
    C_Vertex3f(0.180786133f, 0.184570313f, -1.f),
    C_Vertex3f(0.308326721f, 0.640625f, -1.f),
    C_Vertex3f(0.0920181274f, 0.f, -1.f),
    C_Vertex3f(0.000244140625f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.51171875f, 0.29296875f, -1.f),
    C_Vertex3f(0.812011719f, 0.29296875f, -1.f),
    C_Vertex3f(0.51171875f, 0.368652344f, -1.f),
    C_Vertex3f(0.51171875f, 0.564941406f, -1.f),
    C_Vertex3f(0.51171875f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.363716125f, 0.564941406f, -1.f),
    C_Vertex3f(0.426757813f, 0.564941406f, -1.f),
    C_Vertex3f(0.834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.308326721f, 0.640625f, -1.f),
    C_Vertex3f(0.217185974f, 0.260253906f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.51171875f, 0.0756835938f, -1.f),
    C_Vertex3f(0.426757813f, 0.184570313f, -1.f),
    C_Vertex3f(0.426757813f, 0.f, -1.f),
    C_Vertex3f(0.846679688f, 0.f, -1.f),
    C_Vertex3f(0.846679688f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.51171875f, 0.368652344f, -1.f),
    C_Vertex3f(0.812011719f, 0.29296875f, -1.f),
    C_Vertex3f(0.812011719f, 0.368652344f, -1.f),
    C_End(),
    C_Translatef(0.895019531f, 0.f, 0.f),
};
static const Call char_199[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.244256333f, 0.969710708f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, -1.f),
    C_Vertex3f(0.610839844f, 0.203514099f, -1.f),
    C_Vertex3f(0.610839844f, 0.203514099f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.96897471f, -0.247160003f, 0.f),
    C_Vertex3f(0.610839844f, 0.203514099f, 0.f),
    C_Vertex3f(0.610839844f, 0.203514099f, -1.f),
    C_Normal3f(0.957571566f, -0.28819555f, 0.f),
    C_Vertex3f(0.607507706f, 0.190450668f, 0.f),
    C_Vertex3f(0.607507706f, 0.190450668f, -1.f),
    C_Normal3f(0.939702749f, -0.341992259f, 0.f),
    C_Vertex3f(0.599511743f, 0.165874839f, 0.f),
    C_Vertex3f(0.599511743f, 0.165874839f, -1.f),
    C_Normal3f(0.912207901f, -0.409727544f, 0.f),
    C_Vertex3f(0.590184093f, 0.142850041f, 0.f),
    C_Vertex3f(0.590184093f, 0.142850041f, -1.f),
    C_Normal3f(0.877267182f, -0.480002373f, 0.f),
    C_Vertex3f(0.579524755f, 0.121376276f, 0.f),
    C_Vertex3f(0.579524755f, 0.121376276f, -1.f),
    C_Normal3f(0.834214509f, -0.55144006f, 0.f),
    C_Vertex3f(0.567533731f, 0.101453543f, 0.f),
    C_Vertex3f(0.567533731f, 0.101453543f, -1.f),
    C_Normal3f(0.78278327f, -0.622294366f, 0.f),
    C_Vertex3f(0.55421102f, 0.0830818415f, 0.f),
    C_Vertex3f(0.55421102f, 0.0830818415f, -1.f),
    C_Normal3f(0.72325623f, -0.690579891f, 0.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, 0.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, -1.f),
    C_Normal3f(0.656544447f, -0.754287302f, 0.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, 0.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, -1.f),
    C_Normal3f(0.585676551f, -0.810544848f, 0.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, 0.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, -1.f),
    C_Normal3f(0.510755539f, -0.859726012f, 0.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, 0.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, -1.f),
    C_Normal3f(0.431376964f, -0.902171791f, 0.f),
    C_Vertex3f(0.468728662f, 0.014980197f, 0.f),
    C_Vertex3f(0.468728662f, 0.014980197f, -1.f),
    C_Normal3f(0.349487215f, -0.936941028f, 0.f),
    C_Vertex3f(0.448522329f, 0.0064060688f, 0.f),
    C_Vertex3f(0.448522329f, 0.0064060688f, -1.f),
    C_Normal3f(0.267140955f, -0.963657379f, 0.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, 0.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, -1.f),
    C_Normal3f(0.186262488f, -0.982499957f, 0.f),
    C_Vertex3f(0.405221105f, -0.00559771061f, 0.f),
    C_Vertex3f(0.405221105f, -0.00559771061f, -1.f),
    C_Normal3f(0.108459026f, -0.994100928f, 0.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, 0.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, -1.f),
    C_Normal3f(0.034610942f, -0.999400854f, 0.f),
    C_Vertex3f(0.358068466f, -0.0107421875f, 0.f),
    C_Vertex3f(0.358068466f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0262763742f, -0.99965477f, 0.f),
    C_Vertex3f(0.33261013f, -0.0107421875f, 0.f),
    C_Vertex3f(0.33261013f, -0.0107421875f, -1.f),
    C_Normal3f(-0.08182282f, -0.996646881f, 0.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, 0.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, -1.f),
    C_Normal3f(-0.142589718f, -0.989781857f, 0.f),
    C_Vertex3f(0.28444469f, -0.00678789616f, 0.f),
    C_Vertex3f(0.28444469f, -0.00678789616f, -1.f),
    C_Normal3f(-0.208271027f, -0.978071153f, 0.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, 0.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, -1.f),
    C_Normal3f(-0.27848947f, -0.960439324f, 0.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, 0.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, -1.f),
    C_Normal3f(-0.352481455f, -0.935818851f, 0.f),
    C_Vertex3f(0.221263289f, 0.00902926922f, 0.f),
    C_Vertex3f(0.221263289f, 0.00902926922f, -1.f),
    C_Normal3f(-0.429035544f, -0.90328753f, 0.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, 0.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, -1.f),
    C_Normal3f(-0.506492198f, -0.862244546f, 0.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, 0.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, -1.f),
    C_Normal3f(-0.577542961f, -0.816360354f, 0.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, 0.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, -1.f),
    C_Normal3f(-0.64043498f, -0.768012404f, 0.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, 0.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, -1.f),
    C_Normal3f(-0.699227035f, -0.714899778f, 0.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, 0.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, -1.f),
    C_Normal3f(-0.752920151f, -0.65811193f, 0.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, 0.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, -1.f),
    C_Normal3f(-0.800829113f, -0.598892927f, 0.f),
    C_Vertex3f(0.113759041f, 0.0907020569f, 0.f),
    C_Vertex3f(0.113759041f, 0.0907020569f, -1.f),
    C_Normal3f(-0.842613876f, -0.53851819f, 0.f),
    C_Vertex3f(0.102382421f, 0.107153654f, 0.f),
    C_Vertex3f(0.102382421f, 0.107153654f, -1.f),
    C_Normal3f(-0.878260493f, -0.478182614f, 0.f),
    C_Vertex3f(0.0919754505f, 0.124786615f, 0.f),
    C_Vertex3f(0.0919754505f, 0.124786615f, -1.f),
    C_Normal3f(-0.90802449f, -0.41891703f, 0.f),
    C_Vertex3f(0.0825381279f, 0.143600941f, 0.f),
    C_Vertex3f(0.0825381279f, 0.143600941f, -1.f),
    C_Normal3f(-0.930585563f, -0.366074562f, 0.f),
    C_Vertex3f(0.0740704536f, 0.16359663f, 0.f),
    C_Vertex3f(0.0740704536f, 0.16359663f, -1.f),
    C_Normal3f(-0.948261619f, -0.317489982f, 0.f),
    C_Vertex3f(0.0666612387f, 0.183961034f, 0.f),
    C_Vertex3f(0.0666612387f, 0.183961034f, -1.f),
    C_Normal3f(-0.963385701f, -0.268119097f, 0.f),
    C_Vertex3f(0.060310483f, 0.204694152f, 0.f),
    C_Vertex3f(0.060310483f, 0.204694152f, -1.f),
    C_Normal3f(-0.975867569f, -0.218363285f, 0.f),
    C_Vertex3f(0.0550181866f, 0.225795984f, 0.f),
    C_Vertex3f(0.0550181866f, 0.225795984f, -1.f),
    C_Normal3f(-0.985680282f, -0.168625012f, 0.f),
    C_Vertex3f(0.0507843494f, 0.247266531f, 0.f),
    C_Vertex3f(0.0507843494f, 0.247266531f, -1.f),
    C_Normal3f(-0.992859125f, -0.119292498f, 0.f),
    C_Vertex3f(0.0476089716f, 0.269105792f, 0.f),
    C_Vertex3f(0.0476089716f, 0.269105792f, -1.f),
    C_Normal3f(-0.99749589f, -0.0707256868f, 0.f),
    C_Vertex3f(0.045492053f, 0.291313767f, 0.f),
    C_Vertex3f(0.045492053f, 0.291313767f, -1.f),
    C_Normal3f(-0.999741733f, -0.0227270462f, 0.f),
    C_Vertex3f(0.0444335938f, 0.313890457f, 0.f),
    C_Vertex3f(0.0444335938f, 0.313890457f, -1.f),
    C_Normal3f(-0.999692678f, 0.0247900393f, 0.f),
    C_Vertex3f(0.0444335938f, 0.337874413f, 0.f),
    C_Vertex3f(0.0444335938f, 0.337874413f, -1.f),
    C_Normal3f(-0.997179449f, 0.0750543475f, 0.f),
    C_Vertex3f(0.0456290245f, 0.362097859f, 0.f),
    C_Vertex3f(0.0456290245f, 0.362097859f, -1.f),
    C_Normal3f(-0.991681218f, 0.128718078f, 0.f),
    C_Vertex3f(0.048019886f, 0.385522246f, 0.f),
    C_Vertex3f(0.048019886f, 0.385522246f, -1.f),
    C_Normal3f(-0.982738376f, 0.185000554f, 0.f),
    C_Vertex3f(0.0516061783f, 0.408147573f, 0.f),
    C_Vertex3f(0.0516061783f, 0.408147573f, -1.f),
    C_Normal3f(-0.969899654f, 0.243504897f, 0.f),
    C_Vertex3f(0.0563879013f, 0.429973841f, 0.f),
    C_Vertex3f(0.0563879013f, 0.429973841f, -1.f),
    C_Normal3f(-0.952771962f, 0.303686678f, 0.f),
    C_Vertex3f(0.0623650551f, 0.451001048f, 0.f),
    C_Vertex3f(0.0623650551f, 0.451001048f, -1.f),
    C_Normal3f(-0.931062996f, 0.364858449f, 0.f),
    C_Vertex3f(0.0695376396f, 0.471229196f, 0.f),
    C_Vertex3f(0.0695376396f, 0.471229196f, -1.f),
    C_Normal3f(-0.904623687f, 0.426211268f, 0.f),
    C_Vertex3f(0.0779056549f, 0.490658283f, 0.f),
    C_Vertex3f(0.0779056549f, 0.490658283f, -1.f),
    C_Normal3f(-0.873540044f, 0.486752242f, 0.f),
    C_Vertex3f(0.087469101f, 0.509288311f, 0.f),
    C_Vertex3f(0.087469101f, 0.509288311f, -1.f),
    C_Normal3f(-0.837032259f, 0.547153473f, 0.f),
    C_Vertex3f(0.098043561f, 0.526798368f, 0.f),
    C_Vertex3f(0.098043561f, 0.526798368f, -1.f),
    C_Normal3f(-0.794706285f, 0.606994212f, 0.f),
    C_Vertex3f(0.109629035f, 0.543188453f, 0.f),
    C_Vertex3f(0.109629035f, 0.543188453f, -1.f),
    C_Normal3f(-0.746745348f, 0.665110052f, 0.f),
    C_Vertex3f(0.122225523f, 0.558458567f, 0.f),
    C_Vertex3f(0.122225523f, 0.558458567f, -1.f),
    C_Normal3f(-0.693637788f, 0.72032398f, 0.f),
    C_Vertex3f(0.135833025f, 0.572608709f, 0.f),
    C_Vertex3f(0.135833025f, 0.572608709f, -1.f),
    C_Normal3f(-0.636166453f, 0.771551907f, 0.f),
    C_Vertex3f(0.150451541f, 0.585638881f, 0.f),
    C_Vertex3f(0.150451541f, 0.585638881f, -1.f),
    C_Normal3f(-0.575355411f, 0.817903519f, 0.f),
    C_Vertex3f(0.166081071f, 0.597549081f, 0.f),
    C_Vertex3f(0.166081071f, 0.597549081f, -1.f),
    C_Normal3f(-0.512380719f, 0.85875833f, 0.f),
    C_Vertex3f(0.182721615f, 0.60833931f, 0.f),
    C_Vertex3f(0.182721615f, 0.60833931f, -1.f),
    C_Normal3f(-0.452358484f, 0.891836226f, 0.f),
    C_Vertex3f(0.200373173f, 0.618009567f, 0.f),
    C_Vertex3f(0.200373173f, 0.618009567f, -1.f),
    C_Normal3f(-0.394201905f, 0.919023931f, 0.f),
    C_Vertex3f(0.218468785f, 0.626471043f, 0.f),
    C_Vertex3f(0.218468785f, 0.626471043f, -1.f),
    C_Normal3f(-0.334004939f, 0.942571342f, 0.f),
    C_Vertex3f(0.237008452f, 0.633723736f, 0.f),
    C_Vertex3f(0.237008452f, 0.633723736f, -1.f),
    C_Normal3f(-0.272504777f, 0.962154448f, 0.f),
    C_Vertex3f(0.255992174f, 0.639767647f, 0.f),
    C_Vertex3f(0.255992174f, 0.639767647f, -1.f),
    C_Normal3f(-0.210483432f, 0.977597356f, 0.f),
    C_Vertex3f(0.27541995f, 0.644602776f, 0.f),
    C_Vertex3f(0.27541995f, 0.644602776f, -1.f),
    C_Normal3f(-0.148719549f, 0.988879383f, 0.f),
    C_Vertex3f(0.295291781f, 0.648229122f, 0.f),
    C_Vertex3f(0.295291781f, 0.648229122f, -1.f),
    C_Normal3f(-0.0879421532f, 0.996125579f, 0.f),
    C_Vertex3f(0.315607667f, 0.650646687f, 0.f),
    C_Vertex3f(0.315607667f, 0.650646687f, -1.f),
    C_Normal3f(-0.027851617f, 0.999612153f, 0.f),
    C_Vertex3f(0.336367607f, 0.651855469f, 0.f),
    C_Vertex3f(0.336367607f, 0.651855469f, -1.f),
    C_Normal3f(0.0334659219f, 0.999439895f, 0.f),
    C_Vertex3f(0.358991623f, 0.651855469f, 0.f),
    C_Vertex3f(0.358991623f, 0.651855469f, -1.f),
    C_Normal3f(0.101028174f, 0.994883597f, 0.f),
    C_Vertex3f(0.382079482f, 0.650324821f, 0.f),
    C_Vertex3f(0.382079482f, 0.650324821f, -1.f),
    C_Normal3f(0.174023315f, 0.984741509f, 0.f),
    C_Vertex3f(0.404211164f, 0.647263527f, 0.f),
    C_Vertex3f(0.404211164f, 0.647263527f, -1.f),
    C_Normal3f(0.250579476f, 0.968096018f, 0.f),
    C_Vertex3f(0.425386667f, 0.642671585f, 0.f),
    C_Vertex3f(0.425386667f, 0.642671585f, -1.f),
    C_Normal3f(0.329433322f, 0.94417882f, 0.f),
    C_Vertex3f(0.445605993f, 0.636548996f, 0.f),
    C_Vertex3f(0.445605993f, 0.636548996f, -1.f),
    C_Normal3f(0.408961713f, 0.912551582f, 0.f),
    C_Vertex3f(0.464869142f, 0.62889576f, 0.f),
    C_Vertex3f(0.464869142f, 0.62889576f, -1.f),
    C_Normal3f(0.487299085f, 0.873235166f, 0.f),
    C_Vertex3f(0.483176112f, 0.619711876f, 0.f),
    C_Vertex3f(0.483176112f, 0.619711876f, -1.f),
    C_Normal3f(0.562525034f, 0.826780319f, 0.f),
    C_Vertex3f(0.500526905f, 0.608997345f, 0.f),
    C_Vertex3f(0.500526905f, 0.608997345f, -1.f),
    C_Normal3f(0.632112801f, 0.774876416f, 0.f),
    C_Vertex3f(0.51692152f, 0.596752167f, 0.f),
    C_Vertex3f(0.51692152f, 0.596752167f, -1.f),
    C_Normal3f(0.696682513f, 0.717379689f, 0.f),
    C_Vertex3f(0.532072902f, 0.583263397f, 0.f),
    C_Vertex3f(0.532072902f, 0.583263397f, -1.f),
    C_Normal3f(0.756186187f, 0.65435642f, 0.f),
    C_Vertex3f(0.54598105f, 0.568531036f, 0.f),
    C_Vertex3f(0.54598105f, 0.568531036f, -1.f),
    C_Normal3f(0.809380293f, 0.587284982f, 0.f),
    C_Vertex3f(0.558645964f, 0.552555084f, 0.f),
    C_Vertex3f(0.558645964f, 0.552555084f, -1.f),
    C_Normal3f(0.855471253f, 0.517850339f, 0.f),
    C_Vertex3f(0.570067644f, 0.535335541f, 0.f),
    C_Vertex3f(0.570067644f, 0.535335541f, -1.f),
    C_Normal3f(0.894157887f, 0.447751939f, 0.f),
    C_Vertex3f(0.580246091f, 0.516872406f, 0.f),
    C_Vertex3f(0.580246091f, 0.516872406f, -1.f),
    C_Normal3f(0.925588667f, 0.378530741f, 0.f),
    C_Vertex3f(0.589181304f, 0.49716568f, 0.f),
    C_Vertex3f(0.589181304f, 0.49716568f, -1.f),
    C_Normal3f(0.946590006f, 0.322439611f, 0.f),
    C_Vertex3f(0.596873283f, 0.476215363f, 0.f),
    C_Vertex3f(0.596873283f, 0.476215363f, -1.f),
    C_Normal3f(0.960284233f, 0.279023796f, 0.f),
    C_Vertex3f(0.600097656f, 0.465118408f, 0.f),
    C_Vertex3f(0.600097656f, 0.465118408f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.229118645f, -0.973398507f, 0.f),
    C_Vertex3f(0.600097656f, 0.465118408f, 0.f),
    C_Vertex3f(0.600097656f, 0.465118408f, -1.f),
    C_Vertex3f(0.516601563f, 0.445465088f, -1.f),
    C_Vertex3f(0.516601563f, 0.445465088f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.952661753f, -0.304032117f, 0.f),
    C_Vertex3f(0.516601563f, 0.445465088f, 0.f),
    C_Vertex3f(0.516601563f, 0.445465088f, -1.f),
    C_Normal3f(-0.941064239f, -0.338227957f, 0.f),
    C_Vertex3f(0.513813972f, 0.454199791f, 0.f),
    C_Vertex3f(0.513813972f, 0.454199791f, -1.f),
    C_Normal3f(-0.923471034f, -0.383668244f, 0.f),
    C_Vertex3f(0.507610321f, 0.470481753f, 0.f),
    C_Vertex3f(0.507610321f, 0.470481753f, -1.f),
    C_Normal3f(-0.896986663f, -0.442057669f, 0.f),
    C_Vertex3f(0.500778198f, 0.485576272f, 0.f),
    C_Vertex3f(0.500778198f, 0.485576272f, -1.f),
    C_Normal3f(-0.86353302f, -0.504292369f, 0.f),
    C_Vertex3f(0.493317604f, 0.499483347f, 0.f),
    C_Vertex3f(0.493317604f, 0.499483347f, -1.f),
    C_Normal3f(-0.821924388f, -0.569596708f, 0.f),
    C_Vertex3f(0.485228539f, 0.512202978f, 0.f),
    C_Vertex3f(0.485228539f, 0.512202978f, -1.f),
    C_Normal3f(-0.771112323f, -0.63669908f, 0.f),
    C_Vertex3f(0.476511002f, 0.523735166f, 0.f),
    C_Vertex3f(0.476511002f, 0.523735166f, -1.f),
    C_Normal3f(-0.710405827f, -0.703792214f, 0.f),
    C_Vertex3f(0.467164993f, 0.534079909f, 0.f),
    C_Vertex3f(0.467164993f, 0.534079909f, -1.f),
    C_Normal3f(-0.63972795f, -0.768601477f, 0.f),
    C_Vertex3f(0.457190514f, 0.543237209f, 0.f),
    C_Vertex3f(0.457190514f, 0.543237209f, -1.f),
    C_Normal3f(-0.563830793f, -0.825890362f, 0.f),
    C_Vertex3f(0.446587563f, 0.551207066f, 0.f),
    C_Vertex3f(0.446587563f, 0.551207066f, -1.f),
    C_Normal3f(-0.486547291f, -0.873654306f, 0.f),
    C_Vertex3f(0.435301542f, 0.55818069f, 0.f),
    C_Vertex3f(0.435301542f, 0.55818069f, -1.f),
    C_Normal3f(-0.406627208f, -0.913594127f, 0.f),
    C_Vertex3f(0.423332453f, 0.564158082f, 0.f),
    C_Vertex3f(0.423332453f, 0.564158082f, -1.f),
    C_Normal3f(-0.326150656f, -0.945317864f, 0.f),
    C_Vertex3f(0.410680294f, 0.569139242f, 0.f),
    C_Vertex3f(0.410680294f, 0.569139242f, -1.f),
    C_Normal3f(-0.247042209f, -0.96900475f, 0.f),
    C_Vertex3f(0.397345066f, 0.57312417f, 0.f),
    C_Vertex3f(0.397345066f, 0.57312417f, -1.f),
    C_Normal3f(-0.170889482f, -0.98529017f, 0.f),
    C_Vertex3f(0.383326769f, 0.576112866f, 0.f),
    C_Vertex3f(0.383326769f, 0.576112866f, -1.f),
    C_Normal3f(-0.0988525376f, -0.995102167f, 0.f),
    C_Vertex3f(0.368625402f, 0.57810533f, 0.f),
    C_Vertex3f(0.368625402f, 0.57810533f, -1.f),
    C_Normal3f(-0.0304940492f, -0.999534905f, 0.f),
    C_Vertex3f(0.353240967f, 0.579101563f, 0.f),
    C_Vertex3f(0.353240967f, 0.579101563f, -1.f),
    C_Normal3f(0.0315818004f, -0.999501169f, 0.f),
    C_Vertex3f(0.335970879f, 0.579101563f, 0.f),
    C_Vertex3f(0.335970879f, 0.579101563f, -1.f),
    C_Normal3f(0.0952159464f, -0.995456576f, 0.f),
    C_Vertex3f(0.318256378f, 0.577996135f, 0.f),
    C_Vertex3f(0.318256378f, 0.577996135f, -1.f),
    C_Normal3f(0.164439261f, -0.986387312f, 0.f),
    C_Vertex3f(0.301300049f, 0.575785279f, 0.f),
    C_Vertex3f(0.301300049f, 0.575785279f, -1.f),
    C_Normal3f(0.237575501f, -0.971369028f, 0.f),
    C_Vertex3f(0.285101891f, 0.572468996f, 0.f),
    C_Vertex3f(0.285101891f, 0.572468996f, -1.f),
    C_Normal3f(0.313623488f, -0.94954741f, 0.f),
    C_Vertex3f(0.269661903f, 0.568047285f, 0.f),
    C_Vertex3f(0.269661903f, 0.568047285f, -1.f),
    C_Normal3f(0.391205609f, -0.920303345f, 0.f),
    C_Vertex3f(0.254980087f, 0.562520146f, 0.f),
    C_Vertex3f(0.254980087f, 0.562520146f, -1.f),
    C_Normal3f(0.468632936f, -0.883393049f, 0.f),
    C_Vertex3f(0.241056442f, 0.55588758f, 0.f),
    C_Vertex3f(0.241056442f, 0.55588758f, -1.f),
    C_Normal3f(0.544044912f, -0.839056134f, 0.f),
    C_Vertex3f(0.227890968f, 0.548149586f, 0.f),
    C_Vertex3f(0.227890968f, 0.548149586f, -1.f),
    C_Normal3f(0.61108011f, -0.791568756f, 0.f),
    C_Vertex3f(0.215483665f, 0.539306164f, 0.f),
    C_Vertex3f(0.215483665f, 0.539306164f, -1.f),
    C_Normal3f(0.671135724f, -0.741334558f, 0.f),
    C_Vertex3f(0.203998804f, 0.529705167f, 0.f),
    C_Vertex3f(0.203998804f, 0.529705167f, -1.f),
    C_Normal3f(0.728354037f, -0.685200989f, 0.f),
    C_Vertex3f(0.193436384f, 0.519346595f, 0.f),
    C_Vertex3f(0.193436384f, 0.519346595f, -1.f),
    C_Normal3f(0.781438589f, -0.623982251f, 0.f),
    C_Vertex3f(0.183796406f, 0.508230448f, 0.f),
    C_Vertex3f(0.183796406f, 0.508230448f, -1.f),
    C_Normal3f(0.829291821f, -0.558815777f, 0.f),
    C_Vertex3f(0.175078869f, 0.496356726f, 0.f),
    C_Vertex3f(0.175078869f, 0.496356726f, -1.f),
    C_Normal3f(0.871125877f, -0.491059691f, 0.f),
    C_Vertex3f(0.167283773f, 0.483725429f, 0.f),
    C_Vertex3f(0.167283773f, 0.483725429f, -1.f),
    C_Normal3f(0.90652287f, -0.422156781f, 0.f),
    C_Vertex3f(0.160411119f, 0.470336556f, 0.f),
    C_Vertex3f(0.160411119f, 0.470336556f, -1.f),
    C_Normal3f(0.937979281f, -0.34669134f, 0.f),
    C_Vertex3f(0.154460907f, 0.456190109f, 0.f),
    C_Vertex3f(0.154460907f, 0.456190109f, -1.f),
    C_Normal3f(0.961297512f, -0.275512487f, 0.f),
    C_Vertex3f(0.14691925f, 0.433834076f, 0.f),
    C_Vertex3f(0.14691925f, 0.433834076f, -1.f),
    C_Normal3f(0.979427636f, -0.201795474f, 0.f),
    C_Vertex3f(0.139377594f, 0.403562546f, 0.f),
    C_Vertex3f(0.139377594f, 0.403562546f, -1.f),
    C_Normal3f(0.992667437f, -0.120877311f, 0.f),
    C_Vertex3f(0.134349823f, 0.37282753f, 0.f),
    C_Vertex3f(0.134349823f, 0.37282753f, -1.f),
    C_Normal3f(0.999036968f, -0.0438771285f, 0.f),
    C_Vertex3f(0.131835938f, 0.341629028f, 0.f),
    C_Vertex3f(0.131835938f, 0.341629028f, -1.f),
    C_Normal3f(0.999868095f, 0.0162386186f, 0.f),
    C_Vertex3f(0.131835938f, 0.315588951f, 0.f),
    C_Vertex3f(0.131835938f, 0.315588951f, -1.f),
    C_Normal3f(0.998357296f, 0.0572959892f, 0.f),
    C_Vertex3f(0.132579923f, 0.295819283f, 0.f),
    C_Vertex3f(0.132579923f, 0.295819283f, -1.f),
    C_Normal3f(0.995140135f, 0.0984695256f, 0.f),
    C_Vertex3f(0.134067893f, 0.276698112f, 0.f),
    C_Vertex3f(0.134067893f, 0.276698112f, -1.f),
    C_Normal3f(0.989862859f, 0.142026201f, 0.f),
    C_Vertex3f(0.136299849f, 0.258225441f, 0.f),
    C_Vertex3f(0.136299849f, 0.258225441f, -1.f),
    C_Normal3f(0.98218751f, 0.187903568f, 0.f),
    C_Vertex3f(0.139275789f, 0.240401268f, 0.f),
    C_Vertex3f(0.139275789f, 0.240401268f, -1.f),
    C_Normal3f(0.971761227f, 0.23596625f, 0.f),
    C_Vertex3f(0.142995715f, 0.223225594f, 0.f),
    C_Vertex3f(0.142995715f, 0.223225594f, -1.f),
    C_Normal3f(0.958231807f, 0.285992563f, 0.f),
    C_Vertex3f(0.147459626f, 0.206698418f, 0.f),
    C_Vertex3f(0.147459626f, 0.206698418f, -1.f),
    C_Normal3f(0.941266835f, 0.337663591f, 0.f),
    C_Vertex3f(0.152667522f, 0.19081974f, 0.f),
    C_Vertex3f(0.152667522f, 0.19081974f, -1.f),
    C_Normal3f(0.918154716f, 0.396222055f, 0.f),
    C_Vertex3f(0.158619404f, 0.175589561f, 0.f),
    C_Vertex3f(0.158619404f, 0.175589561f, -1.f),
    C_Normal3f(0.886648476f, 0.462444037f, 0.f),
    C_Vertex3f(0.165397286f, 0.161321402f, 0.f),
    C_Vertex3f(0.165397286f, 0.161321402f, -1.f),
    C_Normal3f(0.847963274f, 0.530054867f, 0.f),
    C_Vertex3f(0.17300117f, 0.148015261f, 0.f),
    C_Vertex3f(0.17300117f, 0.148015261f, -1.f),
    C_Normal3f(0.801762938f, 0.597642303f, 0.f),
    C_Vertex3f(0.181431055f, 0.135671139f, 0.f),
    C_Vertex3f(0.181431055f, 0.135671139f, -1.f),
    C_Normal3f(0.748123646f, 0.663559377f, 0.f),
    C_Vertex3f(0.190686941f, 0.124289036f, 0.f),
    C_Vertex3f(0.190686941f, 0.124289036f, -1.f),
    C_Normal3f(0.687618852f, 0.726071894f, 0.f),
    C_Vertex3f(0.200768828f, 0.113868952f, 0.f),
    C_Vertex3f(0.200768828f, 0.113868952f, -1.f),
    C_Normal3f(0.621327579f, 0.783550918f, 0.f),
    C_Vertex3f(0.211676717f, 0.104410887f, 0.f),
    C_Vertex3f(0.211676717f, 0.104410887f, -1.f),
    C_Normal3f(0.550750792f, 0.834669828f, 0.f),
    C_Vertex3f(0.223410606f, 0.0959148407f, 0.f),
    C_Vertex3f(0.223410606f, 0.0959148407f, -1.f),
    C_Normal3f(0.486346662f, 0.873765886f, 0.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, 0.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, -1.f),
    C_Normal3f(0.427382827f, 0.904070735f, 0.f),
    C_Vertex3f(0.248789787f, 0.0817885399f, 0.f),
    C_Vertex3f(0.248789787f, 0.0817885399f, -1.f),
    C_Normal3f(0.365080386f, 0.930975974f, 0.f),
    C_Vertex3f(0.261868477f, 0.0761380196f, 0.f),
    C_Vertex3f(0.261868477f, 0.0761380196f, -1.f),
    C_Normal3f(0.300158978f, 0.953889191f, 0.f),
    C_Vertex3f(0.275206566f, 0.0714292526f, 0.f),
    C_Vertex3f(0.275206566f, 0.0714292526f, -1.f),
    C_Normal3f(0.233480901f, 0.972361386f, 0.f),
    C_Vertex3f(0.288804054f, 0.0676622391f, 0.f),
    C_Vertex3f(0.288804054f, 0.0676622391f, -1.f),
    C_Normal3f(0.165996134f, 0.986126363f, 0.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, 0.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, -1.f),
    C_Normal3f(0.098675929f, 0.995119691f, 0.f),
    C_Vertex3f(0.316777229f, 0.0629534721f, 0.f),
    C_Vertex3f(0.316777229f, 0.0629534721f, -1.f),
    C_Normal3f(0.0307678469f, 0.99952662f, 0.f),
    C_Vertex3f(0.331152916f, 0.0620117188f, 0.f),
    C_Vertex3f(0.331152916f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0384552181f, 0.999260366f, 0.f),
    C_Vertex3f(0.347371101f, 0.0620117188f, 0.f),
    C_Vertex3f(0.347371101f, 0.0620117188f, -1.f),
    C_Normal3f(-0.113953315f, 0.993486106f, 0.f),
    C_Vertex3f(0.364489913f, 0.0632946491f, 0.f),
    C_Vertex3f(0.364489913f, 0.0632946491f, -1.f),
    C_Normal3f(-0.19543092f, 0.98071748f, 0.f),
    C_Vertex3f(0.380926251f, 0.0658605099f, 0.f),
    C_Vertex3f(0.380926251f, 0.0658605099f, -1.f),
    C_Normal3f(-0.279707551f, 0.960085154f, 0.f),
    C_Vertex3f(0.396680117f, 0.069709301f, 0.f),
    C_Vertex3f(0.396680117f, 0.069709301f, -1.f),
    C_Normal3f(-0.364904284f, 0.931044996f, 0.f),
    C_Vertex3f(0.411751509f, 0.0748410225f, 0.f),
    C_Vertex3f(0.411751509f, 0.0748410225f, -1.f),
    C_Normal3f(-0.44885537f, 0.893604457f, 0.f),
    C_Vertex3f(0.426140428f, 0.0812556744f, 0.f),
    C_Vertex3f(0.426140428f, 0.0812556744f, -1.f),
    C_Normal3f(-0.529350221f, 0.848403394f, 0.f),
    C_Vertex3f(0.439846873f, 0.0889532566f, 0.f),
    C_Vertex3f(0.439846873f, 0.0889532566f, -1.f),
    C_Normal3f(-0.604410052f, 0.796673357f, 0.f),
    C_Vertex3f(0.452870846f, 0.0979337692f, 0.f),
    C_Vertex3f(0.452870846f, 0.0979337692f, -1.f),
    C_Normal3f(-0.676676273f, 0.73628056f, 0.f),
    C_Vertex3f(0.465212345f, 0.108197212f, 0.f),
    C_Vertex3f(0.465212345f, 0.108197212f, -1.f),
    C_Normal3f(-0.745346248f, 0.666677594f, 0.f),
    C_Vertex3f(0.476557136f, 0.119702578f, 0.f),
    C_Vertex3f(0.476557136f, 0.119702578f, -1.f),
    C_Normal3f(-0.80507195f, 0.59317708f, 0.f),
    C_Vertex3f(0.486905217f, 0.132449865f, 0.f),
    C_Vertex3f(0.486905217f, 0.132449865f, -1.f),
    C_Normal3f(-0.855240166f, 0.518231869f, 0.f),
    C_Vertex3f(0.49625659f, 0.146439075f, 0.f),
    C_Vertex3f(0.49625659f, 0.146439075f, -1.f),
    C_Normal3f(-0.895998299f, 0.444057435f, 0.f),
    C_Vertex3f(0.504611254f, 0.161670208f, 0.f),
    C_Vertex3f(0.504611254f, 0.161670208f, -1.f),
    C_Normal3f(-0.928064346f, 0.372419834f, 0.f),
    C_Vertex3f(0.511969209f, 0.178143263f, 0.f),
    C_Vertex3f(0.511969209f, 0.178143263f, -1.f),
    C_Normal3f(-0.952493608f, 0.304558545f, 0.f),
    C_Vertex3f(0.518330455f, 0.19585824f, 0.f),
    C_Vertex3f(0.518330455f, 0.19585824f, -1.f),
    C_Normal3f(-0.967872739f, 0.251440734f, 0.f),
    C_Vertex3f(0.523694992f, 0.21481514f, 0.f),
    C_Vertex3f(0.523694992f, 0.21481514f, -1.f),
    C_Normal3f(-0.977408946f, 0.21135661f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.959265828f, 0.282505304f, 0.f),
    C_Vertex3f(0.277084351f, -0.0771484375f, 0.f),
    C_Vertex3f(0.277084351f, -0.0771484375f, -1.f),
    C_Vertex3f(0.299804688f, 0.f, -1.f),
    C_Vertex3f(0.299804688f, 0.f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.299804688f, 0.f, 0.f),
    C_Vertex3f(0.299804688f, 0.f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_Normal3f(0.955515504f, -0.294941097f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_Vertex3f(0.343978882f, -0.0467147827f, -1.f),
    C_Vertex3f(0.343978882f, -0.0467147827f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.116983637f, 0.993133843f, 0.f),
    C_Vertex3f(0.343978882f, -0.0467147827f, 0.f),
    C_Vertex3f(0.343978882f, -0.0467147827f, -1.f),
    C_Normal3f(0.194189295f, 0.980964124f, 0.f),
    C_Vertex3f(0.353240967f, -0.0478057861f, 0.f),
    C_Vertex3f(0.353240967f, -0.0478057861f, -1.f),
    C_Normal3f(0.305636048f, 0.952148438f, 0.f),
    C_Vertex3f(0.369473457f, -0.0517616272f, 0.f),
    C_Vertex3f(0.369473457f, -0.0517616272f, -1.f),
    C_Normal3f(0.45934099f, 0.888260067f, 0.f),
    C_Vertex3f(0.383414268f, -0.0574913025f, 0.f),
    C_Vertex3f(0.383414268f, -0.0574913025f, -1.f),
    C_Normal3f(0.624141216f, 0.781311631f, 0.f),
    C_Vertex3f(0.3950634f, -0.064994812f, 0.f),
    C_Vertex3f(0.3950634f, -0.064994812f, -1.f),
    C_Normal3f(0.758728862f, 0.651406646f, 0.f),
    C_Vertex3f(0.404420853f, -0.0742721558f, 0.f),
    C_Vertex3f(0.404420853f, -0.0742721558f, -1.f),
    C_Normal3f(0.864426374f, 0.502759397f, 0.f),
    C_Vertex3f(0.411438942f, -0.0840682983f, 0.f),
    C_Vertex3f(0.411438942f, -0.0840682983f, -1.f),
    C_Normal3f(0.949106157f, 0.314956367f, 0.f),
    C_Vertex3f(0.416117668f, -0.0943832397f, 0.f),
    C_Vertex3f(0.416117668f, -0.0943832397f, -1.f),
    C_Normal3f(0.993615568f, 0.112818882f, 0.f),
    C_Vertex3f(0.418457031f, -0.10521698f, 0.f),
    C_Vertex3f(0.418457031f, -0.10521698f, -1.f),
    C_Normal3f(0.998467624f, -0.0553379245f, 0.f),
    C_Vertex3f(0.418457031f, -0.11498642f, 0.f),
    C_Vertex3f(0.418457031f, -0.11498642f, -1.f),
    C_Normal3f(0.98275584f, -0.184907913f, 0.f),
    C_Vertex3f(0.417475104f, -0.122933984f, 0.f),
    C_Vertex3f(0.417475104f, -0.122933984f, -1.f),
    C_Normal3f(0.951465666f, -0.307754815f, 0.f),
    C_Vertex3f(0.41551125f, -0.130642772f, 0.f),
    C_Vertex3f(0.41551125f, -0.130642772f, -1.f),
    C_Normal3f(0.905878186f, -0.423538417f, 0.f),
    C_Vertex3f(0.41256547f, -0.138112783f, 0.f),
    C_Vertex3f(0.41256547f, -0.138112783f, -1.f),
    C_Normal3f(0.84940356f, -0.527743995f, 0.f),
    C_Vertex3f(0.408637762f, -0.145344019f, 0.f),
    C_Vertex3f(0.408637762f, -0.145344019f, -1.f),
    C_Normal3f(0.786300123f, -0.61784476f, 0.f),
    C_Vertex3f(0.403728127f, -0.152336478f, 0.f),
    C_Vertex3f(0.403728127f, -0.152336478f, -1.f),
    C_Normal3f(0.720648766f, -0.693300307f, 0.f),
    C_Vertex3f(0.397836566f, -0.159090161f, 0.f),
    C_Vertex3f(0.397836566f, -0.159090161f, -1.f),
    C_Normal3f(0.655690968f, -0.75502938f, 0.f),
    C_Vertex3f(0.390963078f, -0.165605068f, 0.f),
    C_Vertex3f(0.390963078f, -0.165605068f, -1.f),
    C_Normal3f(0.575867295f, -0.817543089f, 0.f),
    C_Vertex3f(0.383107662f, -0.171881199f, 0.f),
    C_Vertex3f(0.383107662f, -0.171881199f, -1.f),
    C_Normal3f(0.478857249f, -0.877892852f, 0.f),
    C_Vertex3f(0.37425673f, -0.177372813f, 0.f),
    C_Vertex3f(0.37425673f, -0.177372813f, -1.f),
    C_Normal3f(0.384976745f, -0.922926307f, 0.f),
    C_Vertex3f(0.364410281f, -0.182079911f, 0.f),
    C_Vertex3f(0.364410281f, -0.182079911f, -1.f),
    C_Normal3f(0.297251523f, -0.954799294f, 0.f),
    C_Vertex3f(0.353568316f, -0.186002493f, 0.f),
    C_Vertex3f(0.353568316f, -0.186002493f, -1.f),
    C_Normal3f(0.217280537f, -0.976109207f, 0.f),
    C_Vertex3f(0.341730833f, -0.189140558f, 0.f),
    C_Vertex3f(0.341730833f, -0.189140558f, -1.f),
    C_Normal3f(0.145558298f, -0.989349663f, 0.f),
    C_Vertex3f(0.328897834f, -0.191494107f, 0.f),
    C_Vertex3f(0.328897834f, -0.191494107f, -1.f),
    C_Normal3f(0.0818652809f, -0.996643364f, 0.f),
    C_Vertex3f(0.315069318f, -0.19306314f, 0.f),
    C_Vertex3f(0.315069318f, -0.19306314f, -1.f),
    C_Normal3f(0.0247552618f, -0.999693513f, 0.f),
    C_Vertex3f(0.300245285f, -0.193847656f, 0.f),
    C_Vertex3f(0.300245285f, -0.193847656f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.283388138f, -0.193847656f, 0.f),
    C_Vertex3f(0.283388138f, -0.193847656f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.26658535f, -0.193847656f, 0.f),
    C_Vertex3f(0.26658535f, -0.193847656f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.250874519f, -0.193847656f, 0.f),
    C_Vertex3f(0.250874519f, -0.193847656f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.236255646f, -0.193847656f, 0.f),
    C_Vertex3f(0.236255646f, -0.193847656f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.229492188f, -0.193847656f, 0.f),
    C_Vertex3f(0.229492188f, -0.193847656f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.996240556f, 0.0866296142f, 0.f),
    C_Vertex3f(0.229492188f, -0.193847656f, 0.f),
    C_Vertex3f(0.229492188f, -0.193847656f, -1.f),
    C_Vertex3f(0.234375f, -0.137695313f, -1.f),
    C_Vertex3f(0.234375f, -0.137695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.0523584522f, 0.998628378f, 0.f),
    C_Vertex3f(0.234375f, -0.137695313f, 0.f),
    C_Vertex3f(0.234375f, -0.137695313f, -1.f),
    C_Normal3f(0.0478803068f, 0.998853087f, 0.f),
    C_Vertex3f(0.241359711f, -0.138061523f, 0.f),
    C_Vertex3f(0.241359711f, -0.138061523f, -1.f),
    C_Normal3f(0.0408847593f, 0.999163866f, 0.f),
    C_Vertex3f(0.253474236f, -0.13861084f, 0.f),
    C_Vertex3f(0.253474236f, -0.13861084f, -1.f),
    C_Normal3f(0.0294185635f, 0.999567151f, 0.f),
    C_Vertex3f(0.263733864f, -0.138977051f, 0.f),
    C_Vertex3f(0.263733864f, -0.138977051f, -1.f),
    C_Normal3f(0.00834572129f, 0.999965191f, 0.f),
    C_Vertex3f(0.272138596f, -0.139160156f, 0.f),
    C_Vertex3f(0.272138596f, -0.139160156f, -1.f),
    C_Normal3f(-0.04573109f, 0.998953879f, 0.f),
    C_Vertex3f(0.285673141f, -0.139160156f, 0.f),
    C_Vertex3f(0.285673141f, -0.139160156f, -1.f),
    C_Normal3f(-0.132396847f, 0.991196811f, 0.f),
    C_Vertex3f(0.303136826f, -0.137741089f, 0.f),
    C_Vertex3f(0.303136826f, -0.137741089f, -1.f),
    C_Normal3f(-0.265544683f, 0.964098513f, 0.f),
    C_Vertex3f(0.317544937f, -0.134902954f, 0.f),
    C_Vertex3f(0.317544937f, -0.134902954f, -1.f),
    C_Normal3f(-0.445162714f, 0.895449579f, 0.f),
    C_Vertex3f(0.328897476f, -0.130645752f, 0.f),
    C_Vertex3f(0.328897476f, -0.130645752f, -1.f),
    C_Normal3f(-0.633790612f, 0.773504674f, 0.f),
    C_Vertex3f(0.336210251f, -0.125623703f, 0.f),
    C_Vertex3f(0.336210251f, -0.125623703f, -1.f),
    C_Normal3f(-0.790464282f, 0.612508297f, 0.f),
    C_Vertex3f(0.340956688f, -0.120764732f, 0.f),
    C_Vertex3f(0.340956688f, -0.120764732f, -1.f),
    C_Normal3f(-0.92062366f, 0.390451133f, 0.f),
    C_Vertex3f(0.344120979f, -0.115414619f, 0.f),
    C_Vertex3f(0.344120979f, -0.115414619f, -1.f),
    C_Normal3f(-0.992604256f, 0.121395536f, 0.f),
    C_Vertex3f(0.345703125f, -0.109573364f, 0.f),
    C_Vertex3f(0.345703125f, -0.109573364f, -1.f),
    C_Normal3f(-0.984436631f, -0.175739869f, 0.f),
    C_Vertex3f(0.345703125f, -0.102478027f, 0.f),
    C_Vertex3f(0.345703125f, -0.102478027f, -1.f),
    C_Normal3f(-0.878779709f, -0.477227718f, 0.f),
    C_Vertex3f(0.34318924f, -0.0954914093f, 0.f),
    C_Vertex3f(0.34318924f, -0.0954914093f, -1.f),
    C_Normal3f(-0.710703135f, -0.703492045f, 0.f),
    C_Vertex3f(0.339418411f, -0.0909051895f, 0.f),
    C_Vertex3f(0.339418411f, -0.0909051895f, -1.f),
    C_Normal3f(-0.553116202f, -0.833104074f, 0.f),
    C_Vertex3f(0.335839272f, -0.0880661011f, 0.f),
    C_Vertex3f(0.335839272f, -0.0880661011f, -1.f),
    C_Normal3f(-0.4364748f, -0.899716496f, 0.f),
    C_Vertex3f(0.331194878f, -0.0854454041f, 0.f),
    C_Vertex3f(0.331194878f, -0.0854454041f, -1.f),
    C_Normal3f(-0.344821334f, -0.938668311f, 0.f),
    C_Vertex3f(0.325485229f, -0.0830430984f, 0.f),
    C_Vertex3f(0.325485229f, -0.0830430984f, -1.f),
    C_Normal3f(-0.238420397f, -0.971162021f, 0.f),
    C_Vertex3f(0.318710327f, -0.0808591843f, 0.f),
    C_Vertex3f(0.318710327f, -0.0808591843f, -1.f),
    C_Normal3f(-0.141307235f, -0.989965796f, 0.f),
    C_Vertex3f(0.309696198f, -0.0791668892f, 0.f),
    C_Vertex3f(0.309696198f, -0.0791668892f, -1.f),
    C_Normal3f(-0.0769447833f, -0.997035325f, 0.f),
    C_Vertex3f(0.298442841f, -0.0779662132f, 0.f),
    C_Vertex3f(0.298442841f, -0.0779662132f, -1.f),
    C_Normal3f(-0.0382600538f, -0.999267817f, 0.f),
    C_Vertex3f(0.284950256f, -0.0772571564f, 0.f),
    C_Vertex3f(0.284950256f, -0.0772571564f, -1.f),
    C_Normal3f(-0.0138202142f, -0.999904573f, 0.f),
    C_Vertex3f(0.277084351f, -0.0771484375f, 0.f),
    C_Vertex3f(0.277084351f, -0.0771484375f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.600097656f, 0.465118408f, 0.f),
    C_Vertex3f(0.596873283f, 0.476215363f, 0.f),
    C_Vertex3f(0.589181304f, 0.49716568f, 0.f),
    C_Vertex3f(0.580246091f, 0.516872406f, 0.f),
    C_Vertex3f(0.570067644f, 0.535335541f, 0.f),
    C_Vertex3f(0.558645964f, 0.552555084f, 0.f),
    C_Vertex3f(0.54598105f, 0.568531036f, 0.f),
    C_Vertex3f(0.532072902f, 0.583263397f, 0.f),
    C_Vertex3f(0.51692152f, 0.596752167f, 0.f),
    C_Vertex3f(0.516601563f, 0.445465088f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.51692152f, 0.596752167f, 0.f),
    C_Vertex3f(0.500526905f, 0.608997345f, 0.f),
    C_Vertex3f(0.500778198f, 0.485576272f, 0.f),
    C_Vertex3f(0.507610321f, 0.470481753f, 0.f),
    C_Vertex3f(0.513813972f, 0.454199791f, 0.f),
    C_Vertex3f(0.516601563f, 0.445465088f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.467164993f, 0.534079909f, 0.f),
    C_Vertex3f(0.476511002f, 0.523735166f, 0.f),
    C_Vertex3f(0.483176112f, 0.619711876f, 0.f),
    C_Vertex3f(0.485228539f, 0.512202978f, 0.f),
    C_Vertex3f(0.500526905f, 0.608997345f, 0.f),
    C_Vertex3f(0.493317604f, 0.499483347f, 0.f),
    C_Vertex3f(0.500778198f, 0.485576272f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.464869142f, 0.62889576f, 0.f),
    C_Vertex3f(0.445605993f, 0.636548996f, 0.f),
    C_Vertex3f(0.446587563f, 0.551207066f, 0.f),
    C_Vertex3f(0.457190514f, 0.543237209f, 0.f),
    C_Vertex3f(0.467164993f, 0.534079909f, 0.f),
    C_Vertex3f(0.483176112f, 0.619711876f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.410680294f, 0.569139242f, 0.f),
    C_Vertex3f(0.423332453f, 0.564158082f, 0.f),
    C_Vertex3f(0.425386667f, 0.642671585f, 0.f),
    C_Vertex3f(0.435301542f, 0.55818069f, 0.f),
    C_Vertex3f(0.445605993f, 0.636548996f, 0.f),
    C_Vertex3f(0.446587563f, 0.551207066f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.404211164f, 0.647263527f, 0.f),
    C_Vertex3f(0.382079482f, 0.650324821f, 0.f),
    C_Vertex3f(0.383326769f, 0.576112866f, 0.f),
    C_Vertex3f(0.397345066f, 0.57312417f, 0.f),
    C_Vertex3f(0.410680294f, 0.569139242f, 0.f),
    C_Vertex3f(0.425386667f, 0.642671585f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318256378f, 0.577996135f, 0.f),
    C_Vertex3f(0.335970879f, 0.579101563f, 0.f),
    C_Vertex3f(0.336367607f, 0.651855469f, 0.f),
    C_Vertex3f(0.353240967f, 0.579101563f, 0.f),
    C_Vertex3f(0.358991623f, 0.651855469f, 0.f),
    C_Vertex3f(0.368625402f, 0.57810533f, 0.f),
    C_Vertex3f(0.382079482f, 0.650324821f, 0.f),
    C_Vertex3f(0.383326769f, 0.576112866f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241056442f, 0.55588758f, 0.f),
    C_Vertex3f(0.254980087f, 0.562520146f, 0.f),
    C_Vertex3f(0.255992174f, 0.639767647f, 0.f),
    C_Vertex3f(0.269661903f, 0.568047285f, 0.f),
    C_Vertex3f(0.27541995f, 0.644602776f, 0.f),
    C_Vertex3f(0.285101891f, 0.572468996f, 0.f),
    C_Vertex3f(0.295291781f, 0.648229122f, 0.f),
    C_Vertex3f(0.301300049f, 0.575785279f, 0.f),
    C_Vertex3f(0.315607667f, 0.650646687f, 0.f),
    C_Vertex3f(0.318256378f, 0.577996135f, 0.f),
    C_Vertex3f(0.336367607f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.203998804f, 0.529705167f, 0.f),
    C_Vertex3f(0.215483665f, 0.539306164f, 0.f),
    C_Vertex3f(0.218468785f, 0.626471043f, 0.f),
    C_Vertex3f(0.227890968f, 0.548149586f, 0.f),
    C_Vertex3f(0.237008452f, 0.633723736f, 0.f),
    C_Vertex3f(0.241056442f, 0.55588758f, 0.f),
    C_Vertex3f(0.255992174f, 0.639767647f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.200373173f, 0.618009567f, 0.f),
    C_Vertex3f(0.182721615f, 0.60833931f, 0.f),
    C_Vertex3f(0.183796406f, 0.508230448f, 0.f),
    C_Vertex3f(0.193436384f, 0.519346595f, 0.f),
    C_Vertex3f(0.203998804f, 0.529705167f, 0.f),
    C_Vertex3f(0.218468785f, 0.626471043f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154460907f, 0.456190109f, 0.f),
    C_Vertex3f(0.160411119f, 0.470336556f, 0.f),
    C_Vertex3f(0.166081071f, 0.597549081f, 0.f),
    C_Vertex3f(0.167283773f, 0.483725429f, 0.f),
    C_Vertex3f(0.182721615f, 0.60833931f, 0.f),
    C_Vertex3f(0.175078869f, 0.496356726f, 0.f),
    C_Vertex3f(0.183796406f, 0.508230448f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150451541f, 0.585638881f, 0.f),
    C_Vertex3f(0.135833025f, 0.572608709f, 0.f),
    C_Vertex3f(0.139377594f, 0.403562546f, 0.f),
    C_Vertex3f(0.14691925f, 0.433834076f, 0.f),
    C_Vertex3f(0.154460907f, 0.456190109f, 0.f),
    C_Vertex3f(0.166081071f, 0.597549081f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.134067893f, 0.276698112f, 0.f),
    C_Vertex3f(0.132579923f, 0.295819283f, 0.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, 0.f),
    C_Vertex3f(0.131835938f, 0.315588951f, 0.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, 0.f),
    C_Vertex3f(0.131835938f, 0.341629028f, 0.f),
    C_Vertex3f(0.135833025f, 0.572608709f, 0.f),
    C_Vertex3f(0.134349823f, 0.37282753f, 0.f),
    C_Vertex3f(0.139377594f, 0.403562546f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0444335938f, 0.337874413f, 0.f),
    C_Vertex3f(0.045492053f, 0.291313767f, 0.f),
    C_Vertex3f(0.0456290245f, 0.362097859f, 0.f),
    C_Vertex3f(0.0476089716f, 0.269105792f, 0.f),
    C_Vertex3f(0.048019886f, 0.385522246f, 0.f),
    C_Vertex3f(0.0507843494f, 0.247266531f, 0.f),
    C_Vertex3f(0.0516061783f, 0.408147573f, 0.f),
    C_Vertex3f(0.0550181866f, 0.225795984f, 0.f),
    C_Vertex3f(0.0563879013f, 0.429973841f, 0.f),
    C_Vertex3f(0.060310483f, 0.204694152f, 0.f),
    C_Vertex3f(0.0623650551f, 0.451001048f, 0.f),
    C_Vertex3f(0.0666612387f, 0.183961034f, 0.f),
    C_Vertex3f(0.0695376396f, 0.471229196f, 0.f),
    C_Vertex3f(0.0740704536f, 0.16359663f, 0.f),
    C_Vertex3f(0.0779056549f, 0.490658283f, 0.f),
    C_Vertex3f(0.0825381279f, 0.143600941f, 0.f),
    C_Vertex3f(0.087469101f, 0.509288311f, 0.f),
    C_Vertex3f(0.0919754505f, 0.124786615f, 0.f),
    C_Vertex3f(0.098043561f, 0.526798368f, 0.f),
    C_Vertex3f(0.102382421f, 0.107153654f, 0.f),
    C_Vertex3f(0.109629035f, 0.543188453f, 0.f),
    C_Vertex3f(0.113759041f, 0.0907020569f, 0.f),
    C_Vertex3f(0.122225523f, 0.558458567f, 0.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, 0.f),
    C_Vertex3f(0.135833025f, 0.572608709f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339418411f, -0.0909051895f, 0.f),
    C_Vertex3f(0.34318924f, -0.0954914093f, 0.f),
    C_Vertex3f(0.343978882f, -0.0467147827f, 0.f),
    C_Vertex3f(0.345703125f, -0.102478027f, 0.f),
    C_Vertex3f(0.353240967f, -0.0478057861f, 0.f),
    C_Vertex3f(0.353568316f, -0.186002493f, 0.f),
    C_Vertex3f(0.369473457f, -0.0517616272f, 0.f),
    C_Vertex3f(0.364410281f, -0.182079911f, 0.f),
    C_Vertex3f(0.37425673f, -0.177372813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.358068466f, -0.0107421875f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_Vertex3f(0.355082631f, -0.0107421875f, 0.f),
    C_Vertex3f(0.33261013f, -0.0107421875f, 0.f),
    C_Vertex3f(0.343978882f, -0.0467147827f, 0.f),
    C_Vertex3f(0.335839272f, -0.0880661011f, 0.f),
    C_Vertex3f(0.339418411f, -0.0909051895f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33261013f, -0.0107421875f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, 0.f),
    C_Vertex3f(0.309696198f, -0.0791668892f, 0.f),
    C_Vertex3f(0.318710327f, -0.0808591843f, 0.f),
    C_Vertex3f(0.325485229f, -0.0830430984f, 0.f),
    C_Vertex3f(0.331194878f, -0.0854454041f, 0.f),
    C_Vertex3f(0.335839272f, -0.0880661011f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.331152916f, 0.0620117188f, 0.f),
    C_Vertex3f(0.316777229f, 0.0629534721f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, 0.f),
    C_Vertex3f(0.299804688f, 0.f, 0.f),
    C_Vertex3f(0.297377974f, -0.00824007485f, 0.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, 0.f),
    C_Vertex3f(0.298442841f, -0.0779662132f, 0.f),
    C_Vertex3f(0.309696198f, -0.0791668892f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.297377974f, -0.00824007485f, 0.f),
    C_Vertex3f(0.277084351f, -0.0771484375f, 0.f),
    C_Vertex3f(0.284950256f, -0.0772571564f, 0.f),
    C_Vertex3f(0.298442841f, -0.0779662132f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.396680117f, 0.069709301f, 0.f),
    C_Vertex3f(0.380926251f, 0.0658605099f, 0.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, 0.f),
    C_Vertex3f(0.364489913f, 0.0632946491f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_Vertex3f(0.347371101f, 0.0620117188f, 0.f),
    C_Vertex3f(0.331152916f, 0.0620117188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.261868477f, 0.0761380196f, 0.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, 0.f),
    C_Vertex3f(0.275206566f, 0.0714292526f, 0.f),
    C_Vertex3f(0.28444469f, -0.00678789616f, 0.f),
    C_Vertex3f(0.288804054f, 0.0676622391f, 0.f),
    C_Vertex3f(0.297377974f, -0.00824007485f, 0.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.248789787f, 0.0817885399f, 0.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, 0.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, 0.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, 0.f),
    C_Vertex3f(0.261868477f, 0.0761380196f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200768828f, 0.113868952f, 0.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, 0.f),
    C_Vertex3f(0.211676717f, 0.104410887f, 0.f),
    C_Vertex3f(0.221263289f, 0.00902926922f, 0.f),
    C_Vertex3f(0.223410606f, 0.0959148407f, 0.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, 0.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190686941f, 0.124289036f, 0.f),
    C_Vertex3f(0.181431055f, 0.135671139f, 0.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, 0.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, 0.f),
    C_Vertex3f(0.200768828f, 0.113868952f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.17300117f, 0.148015261f, 0.f),
    C_Vertex3f(0.165397286f, 0.161321402f, 0.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, 0.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, 0.f),
    C_Vertex3f(0.181431055f, 0.135671139f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158619404f, 0.175589561f, 0.f),
    C_Vertex3f(0.152667522f, 0.19081974f, 0.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, 0.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, 0.f),
    C_Vertex3f(0.165397286f, 0.161321402f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152667522f, 0.19081974f, 0.f),
    C_Vertex3f(0.147459626f, 0.206698418f, 0.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, 0.f),
    C_Vertex3f(0.142995715f, 0.223225594f, 0.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, 0.f),
    C_Vertex3f(0.139275789f, 0.240401268f, 0.f),
    C_Vertex3f(0.136299849f, 0.258225441f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.610839844f, 0.203514099f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, 0.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, 0.f),
    C_Vertex3f(0.55421102f, 0.0830818415f, 0.f),
    C_Vertex3f(0.567533731f, 0.101453543f, 0.f),
    C_Vertex3f(0.579524755f, 0.121376276f, 0.f),
    C_Vertex3f(0.590184093f, 0.142850041f, 0.f),
    C_Vertex3f(0.599511743f, 0.165874839f, 0.f),
    C_Vertex3f(0.607507706f, 0.190450668f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.523694992f, 0.21481514f, 0.f),
    C_Vertex3f(0.518330455f, 0.19585824f, 0.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, 0.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, 0.f),
    C_Vertex3f(0.525878906f, 0.224914551f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.511969209f, 0.178143263f, 0.f),
    C_Vertex3f(0.504611254f, 0.161670208f, 0.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, 0.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, 0.f),
    C_Vertex3f(0.518330455f, 0.19585824f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.49625659f, 0.146439075f, 0.f),
    C_Vertex3f(0.486905217f, 0.132449865f, 0.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, 0.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, 0.f),
    C_Vertex3f(0.504611254f, 0.161670208f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.476557136f, 0.119702578f, 0.f),
    C_Vertex3f(0.465212345f, 0.108197212f, 0.f),
    C_Vertex3f(0.468728662f, 0.014980197f, 0.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, 0.f),
    C_Vertex3f(0.486905217f, 0.132449865f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426140428f, 0.0812556744f, 0.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, 0.f),
    C_Vertex3f(0.439846873f, 0.0889532566f, 0.f),
    C_Vertex3f(0.448522329f, 0.0064060688f, 0.f),
    C_Vertex3f(0.452870846f, 0.0979337692f, 0.f),
    C_Vertex3f(0.468728662f, 0.014980197f, 0.f),
    C_Vertex3f(0.465212345f, 0.108197212f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.411751509f, 0.0748410225f, 0.f),
    C_Vertex3f(0.396680117f, 0.069709301f, 0.f),
    C_Vertex3f(0.405221105f, -0.00559771061f, 0.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, 0.f),
    C_Vertex3f(0.426140428f, 0.0812556744f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.41256547f, -0.138112783f, 0.f),
    C_Vertex3f(0.41551125f, -0.130642772f, 0.f),
    C_Vertex3f(0.416117668f, -0.0943832397f, 0.f),
    C_Vertex3f(0.417475104f, -0.122933984f, 0.f),
    C_Vertex3f(0.418457031f, -0.10521698f, 0.f),
    C_Vertex3f(0.418457031f, -0.11498642f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.397836566f, -0.159090161f, 0.f),
    C_Vertex3f(0.403728127f, -0.152336478f, 0.f),
    C_Vertex3f(0.404420853f, -0.0742721558f, 0.f),
    C_Vertex3f(0.408637762f, -0.145344019f, 0.f),
    C_Vertex3f(0.411438942f, -0.0840682983f, 0.f),
    C_Vertex3f(0.41256547f, -0.138112783f, 0.f),
    C_Vertex3f(0.416117668f, -0.0943832397f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.37425673f, -0.177372813f, 0.f),
    C_Vertex3f(0.383107662f, -0.171881199f, 0.f),
    C_Vertex3f(0.383414268f, -0.0574913025f, 0.f),
    C_Vertex3f(0.390963078f, -0.165605068f, 0.f),
    C_Vertex3f(0.3950634f, -0.064994812f, 0.f),
    C_Vertex3f(0.397836566f, -0.159090161f, 0.f),
    C_Vertex3f(0.404420853f, -0.0742721558f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.345703125f, -0.102478027f, 0.f),
    C_Vertex3f(0.345703125f, -0.109573364f, 0.f),
    C_Vertex3f(0.353568316f, -0.186002493f, 0.f),
    C_Vertex3f(0.344120979f, -0.115414619f, 0.f),
    C_Vertex3f(0.341730833f, -0.189140558f, 0.f),
    C_Vertex3f(0.340956688f, -0.120764732f, 0.f),
    C_Vertex3f(0.336210251f, -0.125623703f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.328897834f, -0.191494107f, 0.f),
    C_Vertex3f(0.341730833f, -0.189140558f, 0.f),
    C_Vertex3f(0.336210251f, -0.125623703f, 0.f),
    C_Vertex3f(0.328897476f, -0.130645752f, 0.f),
    C_Vertex3f(0.317544937f, -0.134902954f, 0.f),
    C_Vertex3f(0.315069318f, -0.19306314f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.263733864f, -0.138977051f, 0.f),
    C_Vertex3f(0.26658535f, -0.193847656f, 0.f),
    C_Vertex3f(0.272138596f, -0.139160156f, 0.f),
    C_Vertex3f(0.283388138f, -0.193847656f, 0.f),
    C_Vertex3f(0.285673141f, -0.139160156f, 0.f),
    C_Vertex3f(0.300245285f, -0.193847656f, 0.f),
    C_Vertex3f(0.303136826f, -0.137741089f, 0.f),
    C_Vertex3f(0.315069318f, -0.19306314f, 0.f),
    C_Vertex3f(0.317544937f, -0.134902954f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.234375f, -0.137695313f, 0.f),
    C_Vertex3f(0.236255646f, -0.193847656f, 0.f),
    C_Vertex3f(0.241359711f, -0.138061523f, 0.f),
    C_Vertex3f(0.250874519f, -0.193847656f, 0.f),
    C_Vertex3f(0.253474236f, -0.13861084f, 0.f),
    C_Vertex3f(0.26658535f, -0.193847656f, 0.f),
    C_Vertex3f(0.263733864f, -0.138977051f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.236255646f, -0.193847656f, 0.f),
    C_Vertex3f(0.234375f, -0.137695313f, 0.f),
    C_Vertex3f(0.229492188f, -0.193847656f, 0.f),
    C_Vertex3f(0.369473457f, -0.0517616272f, 0.f),
    C_Vertex3f(0.37425673f, -0.177372813f, 0.f),
    C_Vertex3f(0.383414268f, -0.0574913025f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_Vertex3f(0.358068466f, -0.0107421875f, 0.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, 0.f),
    C_Vertex3f(0.396680117f, 0.069709301f, 0.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, 0.f),
    C_Vertex3f(0.405221105f, -0.00559771061f, 0.f),
    C_Vertex3f(0.134067893f, 0.276698112f, 0.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, 0.f),
    C_Vertex3f(0.136299849f, 0.258225441f, 0.f),
    C_Vertex3f(0.299804688f, 0.f, 0.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, 0.f),
    C_Vertex3f(0.358398438f, 0.f, 0.f),
    C_Vertex3f(0.045492053f, 0.291313767f, 0.f),
    C_Vertex3f(0.0444335938f, 0.337874413f, 0.f),
    C_Vertex3f(0.0444335938f, 0.313890457f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.600097656f, 0.465118408f, -1.f),
    C_Vertex3f(0.596873283f, 0.476215363f, -1.f),
    C_Vertex3f(0.589181304f, 0.49716568f, -1.f),
    C_Vertex3f(0.580246091f, 0.516872406f, -1.f),
    C_Vertex3f(0.570067644f, 0.535335541f, -1.f),
    C_Vertex3f(0.558645964f, 0.552555084f, -1.f),
    C_Vertex3f(0.54598105f, 0.568531036f, -1.f),
    C_Vertex3f(0.532072902f, 0.583263397f, -1.f),
    C_Vertex3f(0.51692152f, 0.596752167f, -1.f),
    C_Vertex3f(0.516601563f, 0.445465088f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.51692152f, 0.596752167f, -1.f),
    C_Vertex3f(0.500526905f, 0.608997345f, -1.f),
    C_Vertex3f(0.500778198f, 0.485576272f, -1.f),
    C_Vertex3f(0.507610321f, 0.470481753f, -1.f),
    C_Vertex3f(0.513813972f, 0.454199791f, -1.f),
    C_Vertex3f(0.516601563f, 0.445465088f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.467164993f, 0.534079909f, -1.f),
    C_Vertex3f(0.476511002f, 0.523735166f, -1.f),
    C_Vertex3f(0.483176112f, 0.619711876f, -1.f),
    C_Vertex3f(0.485228539f, 0.512202978f, -1.f),
    C_Vertex3f(0.500526905f, 0.608997345f, -1.f),
    C_Vertex3f(0.493317604f, 0.499483347f, -1.f),
    C_Vertex3f(0.500778198f, 0.485576272f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.464869142f, 0.62889576f, -1.f),
    C_Vertex3f(0.445605993f, 0.636548996f, -1.f),
    C_Vertex3f(0.446587563f, 0.551207066f, -1.f),
    C_Vertex3f(0.457190514f, 0.543237209f, -1.f),
    C_Vertex3f(0.467164993f, 0.534079909f, -1.f),
    C_Vertex3f(0.483176112f, 0.619711876f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.410680294f, 0.569139242f, -1.f),
    C_Vertex3f(0.423332453f, 0.564158082f, -1.f),
    C_Vertex3f(0.425386667f, 0.642671585f, -1.f),
    C_Vertex3f(0.435301542f, 0.55818069f, -1.f),
    C_Vertex3f(0.445605993f, 0.636548996f, -1.f),
    C_Vertex3f(0.446587563f, 0.551207066f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.404211164f, 0.647263527f, -1.f),
    C_Vertex3f(0.382079482f, 0.650324821f, -1.f),
    C_Vertex3f(0.383326769f, 0.576112866f, -1.f),
    C_Vertex3f(0.397345066f, 0.57312417f, -1.f),
    C_Vertex3f(0.410680294f, 0.569139242f, -1.f),
    C_Vertex3f(0.425386667f, 0.642671585f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318256378f, 0.577996135f, -1.f),
    C_Vertex3f(0.335970879f, 0.579101563f, -1.f),
    C_Vertex3f(0.336367607f, 0.651855469f, -1.f),
    C_Vertex3f(0.353240967f, 0.579101563f, -1.f),
    C_Vertex3f(0.358991623f, 0.651855469f, -1.f),
    C_Vertex3f(0.368625402f, 0.57810533f, -1.f),
    C_Vertex3f(0.382079482f, 0.650324821f, -1.f),
    C_Vertex3f(0.383326769f, 0.576112866f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.241056442f, 0.55588758f, -1.f),
    C_Vertex3f(0.254980087f, 0.562520146f, -1.f),
    C_Vertex3f(0.255992174f, 0.639767647f, -1.f),
    C_Vertex3f(0.269661903f, 0.568047285f, -1.f),
    C_Vertex3f(0.27541995f, 0.644602776f, -1.f),
    C_Vertex3f(0.285101891f, 0.572468996f, -1.f),
    C_Vertex3f(0.295291781f, 0.648229122f, -1.f),
    C_Vertex3f(0.301300049f, 0.575785279f, -1.f),
    C_Vertex3f(0.315607667f, 0.650646687f, -1.f),
    C_Vertex3f(0.318256378f, 0.577996135f, -1.f),
    C_Vertex3f(0.336367607f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.203998804f, 0.529705167f, -1.f),
    C_Vertex3f(0.215483665f, 0.539306164f, -1.f),
    C_Vertex3f(0.218468785f, 0.626471043f, -1.f),
    C_Vertex3f(0.227890968f, 0.548149586f, -1.f),
    C_Vertex3f(0.237008452f, 0.633723736f, -1.f),
    C_Vertex3f(0.241056442f, 0.55588758f, -1.f),
    C_Vertex3f(0.255992174f, 0.639767647f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.200373173f, 0.618009567f, -1.f),
    C_Vertex3f(0.182721615f, 0.60833931f, -1.f),
    C_Vertex3f(0.183796406f, 0.508230448f, -1.f),
    C_Vertex3f(0.193436384f, 0.519346595f, -1.f),
    C_Vertex3f(0.203998804f, 0.529705167f, -1.f),
    C_Vertex3f(0.218468785f, 0.626471043f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.154460907f, 0.456190109f, -1.f),
    C_Vertex3f(0.160411119f, 0.470336556f, -1.f),
    C_Vertex3f(0.166081071f, 0.597549081f, -1.f),
    C_Vertex3f(0.167283773f, 0.483725429f, -1.f),
    C_Vertex3f(0.182721615f, 0.60833931f, -1.f),
    C_Vertex3f(0.175078869f, 0.496356726f, -1.f),
    C_Vertex3f(0.183796406f, 0.508230448f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150451541f, 0.585638881f, -1.f),
    C_Vertex3f(0.135833025f, 0.572608709f, -1.f),
    C_Vertex3f(0.139377594f, 0.403562546f, -1.f),
    C_Vertex3f(0.14691925f, 0.433834076f, -1.f),
    C_Vertex3f(0.154460907f, 0.456190109f, -1.f),
    C_Vertex3f(0.166081071f, 0.597549081f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.134067893f, 0.276698112f, -1.f),
    C_Vertex3f(0.132579923f, 0.295819283f, -1.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, -1.f),
    C_Vertex3f(0.131835938f, 0.315588951f, -1.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, -1.f),
    C_Vertex3f(0.131835938f, 0.341629028f, -1.f),
    C_Vertex3f(0.135833025f, 0.572608709f, -1.f),
    C_Vertex3f(0.134349823f, 0.37282753f, -1.f),
    C_Vertex3f(0.139377594f, 0.403562546f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0444335938f, 0.337874413f, -1.f),
    C_Vertex3f(0.045492053f, 0.291313767f, -1.f),
    C_Vertex3f(0.0456290245f, 0.362097859f, -1.f),
    C_Vertex3f(0.0476089716f, 0.269105792f, -1.f),
    C_Vertex3f(0.048019886f, 0.385522246f, -1.f),
    C_Vertex3f(0.0507843494f, 0.247266531f, -1.f),
    C_Vertex3f(0.0516061783f, 0.408147573f, -1.f),
    C_Vertex3f(0.0550181866f, 0.225795984f, -1.f),
    C_Vertex3f(0.0563879013f, 0.429973841f, -1.f),
    C_Vertex3f(0.060310483f, 0.204694152f, -1.f),
    C_Vertex3f(0.0623650551f, 0.451001048f, -1.f),
    C_Vertex3f(0.0666612387f, 0.183961034f, -1.f),
    C_Vertex3f(0.0695376396f, 0.471229196f, -1.f),
    C_Vertex3f(0.0740704536f, 0.16359663f, -1.f),
    C_Vertex3f(0.0779056549f, 0.490658283f, -1.f),
    C_Vertex3f(0.0825381279f, 0.143600941f, -1.f),
    C_Vertex3f(0.087469101f, 0.509288311f, -1.f),
    C_Vertex3f(0.0919754505f, 0.124786615f, -1.f),
    C_Vertex3f(0.098043561f, 0.526798368f, -1.f),
    C_Vertex3f(0.102382421f, 0.107153654f, -1.f),
    C_Vertex3f(0.109629035f, 0.543188453f, -1.f),
    C_Vertex3f(0.113759041f, 0.0907020569f, -1.f),
    C_Vertex3f(0.122225523f, 0.558458567f, -1.f),
    C_Vertex3f(0.126105309f, 0.0754318237f, -1.f),
    C_Vertex3f(0.135833025f, 0.572608709f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339418411f, -0.0909051895f, -1.f),
    C_Vertex3f(0.34318924f, -0.0954914093f, -1.f),
    C_Vertex3f(0.343978882f, -0.0467147827f, -1.f),
    C_Vertex3f(0.345703125f, -0.102478027f, -1.f),
    C_Vertex3f(0.353240967f, -0.0478057861f, -1.f),
    C_Vertex3f(0.353568316f, -0.186002493f, -1.f),
    C_Vertex3f(0.369473457f, -0.0517616272f, -1.f),
    C_Vertex3f(0.364410281f, -0.182079911f, -1.f),
    C_Vertex3f(0.37425673f, -0.177372813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.358068466f, -0.0107421875f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_Vertex3f(0.355082631f, -0.0107421875f, -1.f),
    C_Vertex3f(0.33261013f, -0.0107421875f, -1.f),
    C_Vertex3f(0.343978882f, -0.0467147827f, -1.f),
    C_Vertex3f(0.335839272f, -0.0880661011f, -1.f),
    C_Vertex3f(0.339418411f, -0.0909051895f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33261013f, -0.0107421875f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, -1.f),
    C_Vertex3f(0.309696198f, -0.0791668892f, -1.f),
    C_Vertex3f(0.318710327f, -0.0808591843f, -1.f),
    C_Vertex3f(0.325485229f, -0.0830430984f, -1.f),
    C_Vertex3f(0.331194878f, -0.0854454041f, -1.f),
    C_Vertex3f(0.335839272f, -0.0880661011f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.331152916f, 0.0620117188f, -1.f),
    C_Vertex3f(0.316777229f, 0.0629534721f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, -1.f),
    C_Vertex3f(0.299804688f, 0.f, -1.f),
    C_Vertex3f(0.297377974f, -0.00824007485f, -1.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, -1.f),
    C_Vertex3f(0.298442841f, -0.0779662132f, -1.f),
    C_Vertex3f(0.309696198f, -0.0791668892f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.297377974f, -0.00824007485f, -1.f),
    C_Vertex3f(0.277084351f, -0.0771484375f, -1.f),
    C_Vertex3f(0.284950256f, -0.0772571564f, -1.f),
    C_Vertex3f(0.298442841f, -0.0779662132f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.396680117f, 0.069709301f, -1.f),
    C_Vertex3f(0.380926251f, 0.0658605099f, -1.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, -1.f),
    C_Vertex3f(0.364489913f, 0.0632946491f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_Vertex3f(0.347371101f, 0.0620117188f, -1.f),
    C_Vertex3f(0.331152916f, 0.0620117188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.261868477f, 0.0761380196f, -1.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, -1.f),
    C_Vertex3f(0.275206566f, 0.0714292526f, -1.f),
    C_Vertex3f(0.28444469f, -0.00678789616f, -1.f),
    C_Vertex3f(0.288804054f, 0.0676622391f, -1.f),
    C_Vertex3f(0.297377974f, -0.00824007485f, -1.f),
    C_Vertex3f(0.302660942f, 0.0648369789f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.248789787f, 0.0817885399f, -1.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, -1.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, -1.f),
    C_Vertex3f(0.262175322f, -0.00283360481f, -1.f),
    C_Vertex3f(0.261868477f, 0.0761380196f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200768828f, 0.113868952f, -1.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, -1.f),
    C_Vertex3f(0.211676717f, 0.104410887f, -1.f),
    C_Vertex3f(0.221263289f, 0.00902926922f, -1.f),
    C_Vertex3f(0.223410606f, 0.0959148407f, -1.f),
    C_Vertex3f(0.241114855f, 0.00243878365f, -1.f),
    C_Vertex3f(0.235970497f, 0.0883808136f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190686941f, 0.124289036f, -1.f),
    C_Vertex3f(0.181431055f, 0.135671139f, -1.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, -1.f),
    C_Vertex3f(0.202620625f, 0.0169378519f, -1.f),
    C_Vertex3f(0.200768828f, 0.113868952f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.17300117f, 0.148015261f, -1.f),
    C_Vertex3f(0.165397286f, 0.161321402f, -1.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, -1.f),
    C_Vertex3f(0.185186863f, 0.0261645317f, -1.f),
    C_Vertex3f(0.181431055f, 0.135671139f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158619404f, 0.175589561f, -1.f),
    C_Vertex3f(0.152667522f, 0.19081974f, -1.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, -1.f),
    C_Vertex3f(0.168962002f, 0.0367093086f, -1.f),
    C_Vertex3f(0.165397286f, 0.161321402f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152667522f, 0.19081974f, -1.f),
    C_Vertex3f(0.147459626f, 0.206698418f, -1.f),
    C_Vertex3f(0.153706789f, 0.0484354496f, -1.f),
    C_Vertex3f(0.142995715f, 0.223225594f, -1.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, -1.f),
    C_Vertex3f(0.139275789f, 0.240401268f, -1.f),
    C_Vertex3f(0.136299849f, 0.258225441f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.610839844f, 0.203514099f, -1.f),
    C_Vertex3f(0.525878906f, 0.224914551f, -1.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, -1.f),
    C_Vertex3f(0.55421102f, 0.0830818415f, -1.f),
    C_Vertex3f(0.567533731f, 0.101453543f, -1.f),
    C_Vertex3f(0.579524755f, 0.121376276f, -1.f),
    C_Vertex3f(0.590184093f, 0.142850041f, -1.f),
    C_Vertex3f(0.599511743f, 0.165874839f, -1.f),
    C_Vertex3f(0.607507706f, 0.190450668f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.523694992f, 0.21481514f, -1.f),
    C_Vertex3f(0.518330455f, 0.19585824f, -1.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, -1.f),
    C_Vertex3f(0.539556623f, 0.0662611723f, -1.f),
    C_Vertex3f(0.525878906f, 0.224914551f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.511969209f, 0.178143263f, -1.f),
    C_Vertex3f(0.504611254f, 0.161670208f, -1.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, -1.f),
    C_Vertex3f(0.523570538f, 0.0509915352f, -1.f),
    C_Vertex3f(0.518330455f, 0.19585824f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.49625659f, 0.146439075f, -1.f),
    C_Vertex3f(0.486905217f, 0.132449865f, -1.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, -1.f),
    C_Vertex3f(0.506252766f, 0.0372729301f, -1.f),
    C_Vertex3f(0.504611254f, 0.161670208f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.476557136f, 0.119702578f, -1.f),
    C_Vertex3f(0.465212345f, 0.108197212f, -1.f),
    C_Vertex3f(0.468728662f, 0.014980197f, -1.f),
    C_Vertex3f(0.48797214f, 0.0252691507f, -1.f),
    C_Vertex3f(0.486905217f, 0.132449865f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426140428f, 0.0812556744f, -1.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, -1.f),
    C_Vertex3f(0.439846873f, 0.0889532566f, -1.f),
    C_Vertex3f(0.448522329f, 0.0064060688f, -1.f),
    C_Vertex3f(0.452870846f, 0.0979337692f, -1.f),
    C_Vertex3f(0.468728662f, 0.014980197f, -1.f),
    C_Vertex3f(0.465212345f, 0.108197212f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.411751509f, 0.0748410225f, -1.f),
    C_Vertex3f(0.396680117f, 0.069709301f, -1.f),
    C_Vertex3f(0.405221105f, -0.00559771061f, -1.f),
    C_Vertex3f(0.427353144f, -0.000453233719f, -1.f),
    C_Vertex3f(0.426140428f, 0.0812556744f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.41256547f, -0.138112783f, -1.f),
    C_Vertex3f(0.41551125f, -0.130642772f, -1.f),
    C_Vertex3f(0.416117668f, -0.0943832397f, -1.f),
    C_Vertex3f(0.417475104f, -0.122933984f, -1.f),
    C_Vertex3f(0.418457031f, -0.10521698f, -1.f),
    C_Vertex3f(0.418457031f, -0.11498642f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.397836566f, -0.159090161f, -1.f),
    C_Vertex3f(0.403728127f, -0.152336478f, -1.f),
    C_Vertex3f(0.404420853f, -0.0742721558f, -1.f),
    C_Vertex3f(0.408637762f, -0.145344019f, -1.f),
    C_Vertex3f(0.411438942f, -0.0840682983f, -1.f),
    C_Vertex3f(0.41256547f, -0.138112783f, -1.f),
    C_Vertex3f(0.416117668f, -0.0943832397f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.37425673f, -0.177372813f, -1.f),
    C_Vertex3f(0.383107662f, -0.171881199f, -1.f),
    C_Vertex3f(0.383414268f, -0.0574913025f, -1.f),
    C_Vertex3f(0.390963078f, -0.165605068f, -1.f),
    C_Vertex3f(0.3950634f, -0.064994812f, -1.f),
    C_Vertex3f(0.397836566f, -0.159090161f, -1.f),
    C_Vertex3f(0.404420853f, -0.0742721558f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.345703125f, -0.102478027f, -1.f),
    C_Vertex3f(0.345703125f, -0.109573364f, -1.f),
    C_Vertex3f(0.353568316f, -0.186002493f, -1.f),
    C_Vertex3f(0.344120979f, -0.115414619f, -1.f),
    C_Vertex3f(0.341730833f, -0.189140558f, -1.f),
    C_Vertex3f(0.340956688f, -0.120764732f, -1.f),
    C_Vertex3f(0.336210251f, -0.125623703f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.328897834f, -0.191494107f, -1.f),
    C_Vertex3f(0.341730833f, -0.189140558f, -1.f),
    C_Vertex3f(0.336210251f, -0.125623703f, -1.f),
    C_Vertex3f(0.328897476f, -0.130645752f, -1.f),
    C_Vertex3f(0.317544937f, -0.134902954f, -1.f),
    C_Vertex3f(0.315069318f, -0.19306314f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.263733864f, -0.138977051f, -1.f),
    C_Vertex3f(0.26658535f, -0.193847656f, -1.f),
    C_Vertex3f(0.272138596f, -0.139160156f, -1.f),
    C_Vertex3f(0.283388138f, -0.193847656f, -1.f),
    C_Vertex3f(0.285673141f, -0.139160156f, -1.f),
    C_Vertex3f(0.300245285f, -0.193847656f, -1.f),
    C_Vertex3f(0.303136826f, -0.137741089f, -1.f),
    C_Vertex3f(0.315069318f, -0.19306314f, -1.f),
    C_Vertex3f(0.317544937f, -0.134902954f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.234375f, -0.137695313f, -1.f),
    C_Vertex3f(0.236255646f, -0.193847656f, -1.f),
    C_Vertex3f(0.241359711f, -0.138061523f, -1.f),
    C_Vertex3f(0.250874519f, -0.193847656f, -1.f),
    C_Vertex3f(0.253474236f, -0.13861084f, -1.f),
    C_Vertex3f(0.26658535f, -0.193847656f, -1.f),
    C_Vertex3f(0.263733864f, -0.138977051f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.236255646f, -0.193847656f, -1.f),
    C_Vertex3f(0.234375f, -0.137695313f, -1.f),
    C_Vertex3f(0.229492188f, -0.193847656f, -1.f),
    C_Vertex3f(0.369473457f, -0.0517616272f, -1.f),
    C_Vertex3f(0.37425673f, -0.177372813f, -1.f),
    C_Vertex3f(0.383414268f, -0.0574913025f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_Vertex3f(0.358068466f, -0.0107421875f, -1.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, -1.f),
    C_Vertex3f(0.396680117f, 0.069709301f, -1.f),
    C_Vertex3f(0.382126212f, -0.00902736187f, -1.f),
    C_Vertex3f(0.405221105f, -0.00559771061f, -1.f),
    C_Vertex3f(0.134067893f, 0.276698112f, -1.f),
    C_Vertex3f(0.139421225f, 0.0613429546f, -1.f),
    C_Vertex3f(0.136299849f, 0.258225441f, -1.f),
    C_Vertex3f(0.299804688f, 0.f, -1.f),
    C_Vertex3f(0.307922959f, -0.00942409039f, -1.f),
    C_Vertex3f(0.358398438f, 0.f, -1.f),
    C_Vertex3f(0.045492053f, 0.291313767f, -1.f),
    C_Vertex3f(0.0444335938f, 0.337874413f, -1.f),
    C_Vertex3f(0.0444335938f, 0.313890457f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_200[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.372070313f, 0.680175781f, 0.f),
    C_Vertex3f(0.372070313f, 0.680175781f, -1.f),
    C_Vertex3f(0.30859375f, 0.680175781f, -1.f),
    C_Vertex3f(0.30859375f, 0.680175781f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.30859375f, 0.680175781f, 0.f),
    C_Vertex3f(0.30859375f, 0.680175781f, -1.f),
    C_Vertex3f(0.207519531f, 0.802734375f, -1.f),
    C_Vertex3f(0.207519531f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.207519531f, 0.802734375f, 0.f),
    C_Vertex3f(0.207519531f, 0.802734375f, -1.f),
    C_Vertex3f(0.312988281f, 0.802734375f, -1.f),
    C_Vertex3f(0.312988281f, 0.802734375f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.312988281f, 0.802734375f, 0.f),
    C_Vertex3f(0.312988281f, 0.802734375f, -1.f),
    C_Vertex3f(0.372070313f, 0.680175781f, -1.f),
    C_Vertex3f(0.372070313f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.312988281f, 0.802734375f, 0.f),
    C_Vertex3f(0.207519531f, 0.802734375f, 0.f),
    C_Vertex3f(0.30859375f, 0.680175781f, 0.f),
    C_Vertex3f(0.372070313f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.312988281f, 0.802734375f, -1.f),
    C_Vertex3f(0.207519531f, 0.802734375f, -1.f),
    C_Vertex3f(0.30859375f, 0.680175781f, -1.f),
    C_Vertex3f(0.372070313f, 0.680175781f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_201[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.247070313f, 0.680175781f, 0.f),
    C_Vertex3f(0.247070313f, 0.680175781f, -1.f),
    C_Vertex3f(0.3046875f, 0.802734375f, -1.f),
    C_Vertex3f(0.3046875f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.3046875f, 0.802734375f, 0.f),
    C_Vertex3f(0.3046875f, 0.802734375f, -1.f),
    C_Vertex3f(0.408203125f, 0.802734375f, -1.f),
    C_Vertex3f(0.408203125f, 0.802734375f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.408203125f, 0.802734375f, 0.f),
    C_Vertex3f(0.408203125f, 0.802734375f, -1.f),
    C_Vertex3f(0.312011719f, 0.680175781f, -1.f),
    C_Vertex3f(0.312011719f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.312011719f, 0.680175781f, 0.f),
    C_Vertex3f(0.312011719f, 0.680175781f, -1.f),
    C_Vertex3f(0.247070313f, 0.680175781f, -1.f),
    C_Vertex3f(0.247070313f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.3046875f, 0.802734375f, 0.f),
    C_Vertex3f(0.247070313f, 0.680175781f, 0.f),
    C_Vertex3f(0.312011719f, 0.680175781f, 0.f),
    C_Vertex3f(0.408203125f, 0.802734375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.3046875f, 0.802734375f, -1.f),
    C_Vertex3f(0.247070313f, 0.680175781f, -1.f),
    C_Vertex3f(0.312011719f, 0.680175781f, -1.f),
    C_Vertex3f(0.408203125f, 0.802734375f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_202[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.310546875f, 0.754882813f, 0.f),
    C_Vertex3f(0.310546875f, 0.754882813f, -1.f),
    C_Vertex3f(0.261474609f, 0.680175781f, -1.f),
    C_Vertex3f(0.261474609f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.261474609f, 0.680175781f, 0.f),
    C_Vertex3f(0.261474609f, 0.680175781f, -1.f),
    C_Vertex3f(0.172119141f, 0.680175781f, -1.f),
    C_Vertex3f(0.172119141f, 0.680175781f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(0.172119141f, 0.680175781f, 0.f),
    C_Vertex3f(0.172119141f, 0.680175781f, -1.f),
    C_Vertex3f(0.265625f, 0.802734375f, -1.f),
    C_Vertex3f(0.265625f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.265625f, 0.802734375f, 0.f),
    C_Vertex3f(0.265625f, 0.802734375f, -1.f),
    C_Vertex3f(0.349609375f, 0.802734375f, -1.f),
    C_Vertex3f(0.349609375f, 0.802734375f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.349609375f, 0.802734375f, 0.f),
    C_Vertex3f(0.349609375f, 0.802734375f, -1.f),
    C_Vertex3f(0.447998047f, 0.680175781f, -1.f),
    C_Vertex3f(0.447998047f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.447998047f, 0.680175781f, 0.f),
    C_Vertex3f(0.447998047f, 0.680175781f, -1.f),
    C_Vertex3f(0.358642578f, 0.680175781f, -1.f),
    C_Vertex3f(0.358642578f, 0.680175781f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.358642578f, 0.680175781f, 0.f),
    C_Vertex3f(0.358642578f, 0.680175781f, -1.f),
    C_Vertex3f(0.310546875f, 0.754882813f, -1.f),
    C_Vertex3f(0.310546875f, 0.754882813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172119141f, 0.680175781f, 0.f),
    C_Vertex3f(0.261474609f, 0.680175781f, 0.f),
    C_Vertex3f(0.265625f, 0.802734375f, 0.f),
    C_Vertex3f(0.310546875f, 0.754882813f, 0.f),
    C_Vertex3f(0.349609375f, 0.802734375f, 0.f),
    C_Vertex3f(0.358642578f, 0.680175781f, 0.f),
    C_Vertex3f(0.447998047f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172119141f, 0.680175781f, -1.f),
    C_Vertex3f(0.261474609f, 0.680175781f, -1.f),
    C_Vertex3f(0.265625f, 0.802734375f, -1.f),
    C_Vertex3f(0.310546875f, 0.754882813f, -1.f),
    C_Vertex3f(0.349609375f, 0.802734375f, -1.f),
    C_Vertex3f(0.358642578f, 0.680175781f, -1.f),
    C_Vertex3f(0.447998047f, 0.680175781f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_203[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.191894531f, 0.680175781f, 0.f),
    C_Vertex3f(0.191894531f, 0.680175781f, -1.f),
    C_Vertex3f(0.191894531f, 0.76953125f, -1.f),
    C_Vertex3f(0.191894531f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.191894531f, 0.76953125f, 0.f),
    C_Vertex3f(0.191894531f, 0.76953125f, -1.f),
    C_Vertex3f(0.273925781f, 0.76953125f, -1.f),
    C_Vertex3f(0.273925781f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.273925781f, 0.76953125f, 0.f),
    C_Vertex3f(0.273925781f, 0.76953125f, -1.f),
    C_Vertex3f(0.273925781f, 0.680175781f, -1.f),
    C_Vertex3f(0.273925781f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.273925781f, 0.680175781f, 0.f),
    C_Vertex3f(0.273925781f, 0.680175781f, -1.f),
    C_Vertex3f(0.191894531f, 0.680175781f, -1.f),
    C_Vertex3f(0.191894531f, 0.680175781f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354980469f, 0.680175781f, 0.f),
    C_Vertex3f(0.354980469f, 0.680175781f, -1.f),
    C_Vertex3f(0.354980469f, 0.76953125f, -1.f),
    C_Vertex3f(0.354980469f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.354980469f, 0.76953125f, 0.f),
    C_Vertex3f(0.354980469f, 0.76953125f, -1.f),
    C_Vertex3f(0.437011719f, 0.76953125f, -1.f),
    C_Vertex3f(0.437011719f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.437011719f, 0.76953125f, 0.f),
    C_Vertex3f(0.437011719f, 0.76953125f, -1.f),
    C_Vertex3f(0.437011719f, 0.680175781f, -1.f),
    C_Vertex3f(0.437011719f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.437011719f, 0.680175781f, 0.f),
    C_Vertex3f(0.437011719f, 0.680175781f, -1.f),
    C_Vertex3f(0.354980469f, 0.680175781f, -1.f),
    C_Vertex3f(0.354980469f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.0756835938f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.f, 0.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534667969f, 0.640625f, 0.f),
    C_Vertex3f(0.0708007813f, 0.640625f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.534667969f, 0.564941406f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.369140625f, 0.f),
    C_Vertex3f(0.155761719f, 0.564941406f, 0.f),
    C_Vertex3f(0.155761719f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.293457031f, 0.f),
    C_Vertex3f(0.510742188f, 0.369140625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191894531f, 0.76953125f, 0.f),
    C_Vertex3f(0.191894531f, 0.680175781f, 0.f),
    C_Vertex3f(0.273925781f, 0.680175781f, 0.f),
    C_Vertex3f(0.273925781f, 0.76953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.354980469f, 0.76953125f, 0.f),
    C_Vertex3f(0.354980469f, 0.680175781f, 0.f),
    C_Vertex3f(0.437011719f, 0.680175781f, 0.f),
    C_Vertex3f(0.437011719f, 0.76953125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.0756835938f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.f, -1.f),
    C_Vertex3f(0.549316406f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.534667969f, 0.640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.640625f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.534667969f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155761719f, 0.369140625f, -1.f),
    C_Vertex3f(0.155761719f, 0.564941406f, -1.f),
    C_Vertex3f(0.155761719f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.293457031f, -1.f),
    C_Vertex3f(0.510742188f, 0.369140625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.191894531f, 0.76953125f, -1.f),
    C_Vertex3f(0.191894531f, 0.680175781f, -1.f),
    C_Vertex3f(0.273925781f, 0.680175781f, -1.f),
    C_Vertex3f(0.273925781f, 0.76953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.354980469f, 0.76953125f, -1.f),
    C_Vertex3f(0.354980469f, 0.680175781f, -1.f),
    C_Vertex3f(0.437011719f, 0.680175781f, -1.f),
    C_Vertex3f(0.437011719f, 0.76953125f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_204[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.192382813f, 0.680175781f, 0.f),
    C_Vertex3f(0.192382813f, 0.680175781f, -1.f),
    C_Vertex3f(0.12890625f, 0.680175781f, -1.f),
    C_Vertex3f(0.12890625f, 0.680175781f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.12890625f, 0.680175781f, 0.f),
    C_Vertex3f(0.12890625f, 0.680175781f, -1.f),
    C_Vertex3f(0.0278320313f, 0.802734375f, -1.f),
    C_Vertex3f(0.0278320313f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0278320313f, 0.802734375f, 0.f),
    C_Vertex3f(0.0278320313f, 0.802734375f, -1.f),
    C_Vertex3f(0.133300781f, 0.802734375f, -1.f),
    C_Vertex3f(0.133300781f, 0.802734375f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.133300781f, 0.802734375f, 0.f),
    C_Vertex3f(0.133300781f, 0.802734375f, -1.f),
    C_Vertex3f(0.192382813f, 0.680175781f, -1.f),
    C_Vertex3f(0.192382813f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.133300781f, 0.802734375f, 0.f),
    C_Vertex3f(0.0278320313f, 0.802734375f, 0.f),
    C_Vertex3f(0.12890625f, 0.680175781f, 0.f),
    C_Vertex3f(0.192382813f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.133300781f, 0.802734375f, -1.f),
    C_Vertex3f(0.0278320313f, 0.802734375f, -1.f),
    C_Vertex3f(0.12890625f, 0.680175781f, -1.f),
    C_Vertex3f(0.192382813f, 0.680175781f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_205[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.0581054688f, 0.680175781f, 0.f),
    C_Vertex3f(0.0581054688f, 0.680175781f, -1.f),
    C_Vertex3f(0.115722656f, 0.802734375f, -1.f),
    C_Vertex3f(0.115722656f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.115722656f, 0.802734375f, 0.f),
    C_Vertex3f(0.115722656f, 0.802734375f, -1.f),
    C_Vertex3f(0.219238281f, 0.802734375f, -1.f),
    C_Vertex3f(0.219238281f, 0.802734375f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.219238281f, 0.802734375f, 0.f),
    C_Vertex3f(0.219238281f, 0.802734375f, -1.f),
    C_Vertex3f(0.123046875f, 0.680175781f, -1.f),
    C_Vertex3f(0.123046875f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.123046875f, 0.680175781f, 0.f),
    C_Vertex3f(0.123046875f, 0.680175781f, -1.f),
    C_Vertex3f(0.0581054688f, 0.680175781f, -1.f),
    C_Vertex3f(0.0581054688f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115722656f, 0.802734375f, 0.f),
    C_Vertex3f(0.0581054688f, 0.680175781f, 0.f),
    C_Vertex3f(0.123046875f, 0.680175781f, 0.f),
    C_Vertex3f(0.219238281f, 0.802734375f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115722656f, 0.802734375f, -1.f),
    C_Vertex3f(0.0581054688f, 0.680175781f, -1.f),
    C_Vertex3f(0.123046875f, 0.680175781f, -1.f),
    C_Vertex3f(0.219238281f, 0.802734375f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_206[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.125976563f, 0.754882813f, 0.f),
    C_Vertex3f(0.125976563f, 0.754882813f, -1.f),
    C_Vertex3f(0.0769042969f, 0.680175781f, -1.f),
    C_Vertex3f(0.0769042969f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0769042969f, 0.680175781f, 0.f),
    C_Vertex3f(0.0769042969f, 0.680175781f, -1.f),
    C_Vertex3f(-0.0124511719f, 0.680175781f, -1.f),
    C_Vertex3f(-0.0124511719f, 0.680175781f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(-0.0124511719f, 0.680175781f, 0.f),
    C_Vertex3f(-0.0124511719f, 0.680175781f, -1.f),
    C_Vertex3f(0.0810546875f, 0.802734375f, -1.f),
    C_Vertex3f(0.0810546875f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0810546875f, 0.802734375f, 0.f),
    C_Vertex3f(0.0810546875f, 0.802734375f, -1.f),
    C_Vertex3f(0.165039063f, 0.802734375f, -1.f),
    C_Vertex3f(0.165039063f, 0.802734375f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.165039063f, 0.802734375f, 0.f),
    C_Vertex3f(0.165039063f, 0.802734375f, -1.f),
    C_Vertex3f(0.263427734f, 0.680175781f, -1.f),
    C_Vertex3f(0.263427734f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.263427734f, 0.680175781f, 0.f),
    C_Vertex3f(0.263427734f, 0.680175781f, -1.f),
    C_Vertex3f(0.174072266f, 0.680175781f, -1.f),
    C_Vertex3f(0.174072266f, 0.680175781f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.174072266f, 0.680175781f, 0.f),
    C_Vertex3f(0.174072266f, 0.680175781f, -1.f),
    C_Vertex3f(0.125976563f, 0.754882813f, -1.f),
    C_Vertex3f(0.125976563f, 0.754882813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(-0.0124511719f, 0.680175781f, 0.f),
    C_Vertex3f(0.0769042969f, 0.680175781f, 0.f),
    C_Vertex3f(0.0810546875f, 0.802734375f, 0.f),
    C_Vertex3f(0.125976563f, 0.754882813f, 0.f),
    C_Vertex3f(0.165039063f, 0.802734375f, 0.f),
    C_Vertex3f(0.174072266f, 0.680175781f, 0.f),
    C_Vertex3f(0.263427734f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(-0.0124511719f, 0.680175781f, -1.f),
    C_Vertex3f(0.0769042969f, 0.680175781f, -1.f),
    C_Vertex3f(0.0810546875f, 0.802734375f, -1.f),
    C_Vertex3f(0.125976563f, 0.754882813f, -1.f),
    C_Vertex3f(0.165039063f, 0.802734375f, -1.f),
    C_Vertex3f(0.174072266f, 0.680175781f, -1.f),
    C_Vertex3f(0.263427734f, 0.680175781f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_207[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.00537109375f, 0.680175781f, 0.f),
    C_Vertex3f(0.00537109375f, 0.680175781f, -1.f),
    C_Vertex3f(0.00537109375f, 0.76953125f, -1.f),
    C_Vertex3f(0.00537109375f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.00537109375f, 0.76953125f, 0.f),
    C_Vertex3f(0.00537109375f, 0.76953125f, -1.f),
    C_Vertex3f(0.0874023438f, 0.76953125f, -1.f),
    C_Vertex3f(0.0874023438f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.0874023438f, 0.76953125f, 0.f),
    C_Vertex3f(0.0874023438f, 0.76953125f, -1.f),
    C_Vertex3f(0.0874023438f, 0.680175781f, -1.f),
    C_Vertex3f(0.0874023438f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0874023438f, 0.680175781f, 0.f),
    C_Vertex3f(0.0874023438f, 0.680175781f, -1.f),
    C_Vertex3f(0.00537109375f, 0.680175781f, -1.f),
    C_Vertex3f(0.00537109375f, 0.680175781f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.680175781f, 0.f),
    C_Vertex3f(0.168457031f, 0.680175781f, -1.f),
    C_Vertex3f(0.168457031f, 0.76953125f, -1.f),
    C_Vertex3f(0.168457031f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.168457031f, 0.76953125f, 0.f),
    C_Vertex3f(0.168457031f, 0.76953125f, -1.f),
    C_Vertex3f(0.250488281f, 0.76953125f, -1.f),
    C_Vertex3f(0.250488281f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.250488281f, 0.76953125f, 0.f),
    C_Vertex3f(0.250488281f, 0.76953125f, -1.f),
    C_Vertex3f(0.250488281f, 0.680175781f, -1.f),
    C_Vertex3f(0.250488281f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.250488281f, 0.680175781f, 0.f),
    C_Vertex3f(0.250488281f, 0.680175781f, -1.f),
    C_Vertex3f(0.168457031f, 0.680175781f, -1.f),
    C_Vertex3f(0.168457031f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0834960938f, 0.640625f, 0.f),
    C_Vertex3f(0.0834960938f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.f, 0.f),
    C_Vertex3f(0.168457031f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.00537109375f, 0.76953125f, 0.f),
    C_Vertex3f(0.00537109375f, 0.680175781f, 0.f),
    C_Vertex3f(0.0874023438f, 0.680175781f, 0.f),
    C_Vertex3f(0.0874023438f, 0.76953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.168457031f, 0.76953125f, 0.f),
    C_Vertex3f(0.168457031f, 0.680175781f, 0.f),
    C_Vertex3f(0.250488281f, 0.680175781f, 0.f),
    C_Vertex3f(0.250488281f, 0.76953125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0834960938f, 0.640625f, -1.f),
    C_Vertex3f(0.0834960938f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.f, -1.f),
    C_Vertex3f(0.168457031f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.00537109375f, 0.76953125f, -1.f),
    C_Vertex3f(0.00537109375f, 0.680175781f, -1.f),
    C_Vertex3f(0.0874023438f, 0.680175781f, -1.f),
    C_Vertex3f(0.0874023438f, 0.76953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.168457031f, 0.76953125f, -1.f),
    C_Vertex3f(0.168457031f, 0.680175781f, -1.f),
    C_Vertex3f(0.250488281f, 0.680175781f, -1.f),
    C_Vertex3f(0.250488281f, 0.76953125f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_208[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.291503906f, -1.f),
    C_Vertex3f(0.0688476563f, 0.291503906f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.291503906f, 0.f),
    C_Vertex3f(0.0688476563f, 0.291503906f, -1.f),
    C_Vertex3f(-0.00146484375f, 0.291503906f, -1.f),
    C_Vertex3f(-0.00146484375f, 0.291503906f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(-0.00146484375f, 0.291503906f, 0.f),
    C_Vertex3f(-0.00146484375f, 0.291503906f, -1.f),
    C_Vertex3f(-0.00146484375f, 0.349121094f, -1.f),
    C_Vertex3f(-0.00146484375f, 0.349121094f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(-0.00146484375f, 0.349121094f, 0.f),
    C_Vertex3f(-0.00146484375f, 0.349121094f, -1.f),
    C_Vertex3f(0.0688476563f, 0.349121094f, -1.f),
    C_Vertex3f(0.0688476563f, 0.349121094f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.349121094f, 0.f),
    C_Vertex3f(0.0688476563f, 0.349121094f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Normal3f(0.00115672115f, 0.999999344f, 0.f),
    C_Vertex3f(0.299320221f, 0.640625f, 0.f),
    C_Vertex3f(0.299320221f, 0.640625f, -1.f),
    C_Normal3f(0.0253915619f, 0.999677539f, 0.f),
    C_Vertex3f(0.316804886f, 0.640338182f, 0.f),
    C_Vertex3f(0.316804886f, 0.640338182f, -1.f),
    C_Normal3f(0.0452065207f, 0.998977661f, 0.f),
    C_Vertex3f(0.33319664f, 0.639764547f, 0.f),
    C_Vertex3f(0.33319664f, 0.639764547f, -1.f),
    C_Normal3f(0.0678903908f, 0.997692823f, 0.f),
    C_Vertex3f(0.348495483f, 0.638904095f, 0.f),
    C_Vertex3f(0.348495483f, 0.638904095f, -1.f),
    C_Normal3f(0.0940706208f, 0.995565534f, 0.f),
    C_Vertex3f(0.362701416f, 0.637756824f, 0.f),
    C_Vertex3f(0.362701416f, 0.637756824f, -1.f),
    C_Normal3f(0.12455371f, 0.992212951f, 0.f),
    C_Vertex3f(0.375814438f, 0.636322737f, 0.f),
    C_Vertex3f(0.375814438f, 0.636322737f, -1.f),
    C_Normal3f(0.160383075f, 0.987054765f, 0.f),
    C_Vertex3f(0.387834549f, 0.634601831f, 0.f),
    C_Vertex3f(0.387834549f, 0.634601831f, -1.f),
    C_Normal3f(0.208813891f, 0.977955461f, 0.f),
    C_Vertex3f(0.398761749f, 0.632594109f, 0.f),
    C_Vertex3f(0.398761749f, 0.632594109f, -1.f),
    C_Normal3f(0.286513388f, 0.958076298f, 0.f),
    C_Vertex3f(0.417448044f, 0.628278732f, 0.f),
    C_Vertex3f(0.417448044f, 0.628278732f, -1.f),
    C_Normal3f(0.383859992f, 0.923391342f, 0.f),
    C_Vertex3f(0.442964554f, 0.619375229f, 0.f),
    C_Vertex3f(0.442964554f, 0.619375229f, -1.f),
    C_Normal3f(0.493410945f, 0.869796336f, 0.f),
    C_Vertex3f(0.466459274f, 0.607904434f, 0.f),
    C_Vertex3f(0.466459274f, 0.607904434f, -1.f),
    C_Normal3f(0.593687773f, 0.804695487f, 0.f),
    C_Vertex3f(0.487932205f, 0.593866348f, 0.f),
    C_Vertex3f(0.487932205f, 0.593866348f, -1.f),
    C_Normal3f(0.668363988f, 0.743834376f, 0.f),
    C_Vertex3f(0.503995895f, 0.580210686f, 0.f),
    C_Vertex3f(0.503995895f, 0.580210686f, -1.f),
    C_Normal3f(0.717492461f, 0.696566164f, 0.f),
    C_Vertex3f(0.515873075f, 0.568760395f, 0.f),
    C_Vertex3f(0.515873075f, 0.568760395f, -1.f),
    C_Normal3f(0.762034953f, 0.64753598f, 0.f),
    C_Vertex3f(0.526951194f, 0.556565762f, 0.f),
    C_Vertex3f(0.526951194f, 0.556565762f, -1.f),
    C_Normal3f(0.802996516f, 0.595983744f, 0.f),
    C_Vertex3f(0.537230253f, 0.543626785f, 0.f),
    C_Vertex3f(0.537230253f, 0.543626785f, -1.f),
    C_Normal3f(0.839958608f, 0.542650461f, 0.f),
    C_Vertex3f(0.546710253f, 0.529943466f, 0.f),
    C_Vertex3f(0.546710253f, 0.529943466f, -1.f),
    C_Normal3f(0.87266928f, 0.488311708f, 0.f),
    C_Vertex3f(0.555391192f, 0.515515804f, 0.f),
    C_Vertex3f(0.555391192f, 0.515515804f, -1.f),
    C_Normal3f(0.901044369f, 0.433727026f, 0.f),
    C_Vertex3f(0.563273072f, 0.5003438f, 0.f),
    C_Vertex3f(0.563273072f, 0.5003438f, -1.f),
    C_Normal3f(0.925152183f, 0.379596233f, 0.f),
    C_Vertex3f(0.570355892f, 0.484427452f, 0.f),
    C_Vertex3f(0.570355892f, 0.484427452f, -1.f),
    C_Normal3f(0.944622636f, 0.328158587f, 0.f),
    C_Vertex3f(0.576639652f, 0.467766762f, 0.f),
    C_Vertex3f(0.576639652f, 0.467766762f, -1.f),
    C_Normal3f(0.960200369f, 0.279311985f, 0.f),
    C_Vertex3f(0.582137942f, 0.450512171f, 0.f),
    C_Vertex3f(0.582137942f, 0.450512171f, -1.f),
    C_Normal3f(0.972809255f, 0.231607407f, 0.f),
    C_Vertex3f(0.586850762f, 0.432663679f, 0.f),
    C_Vertex3f(0.586850762f, 0.432663679f, -1.f),
    C_Normal3f(0.982672215f, 0.185351655f, 0.f),
    C_Vertex3f(0.590778112f, 0.414221287f, 0.f),
    C_Vertex3f(0.590778112f, 0.414221287f, -1.f),
    C_Normal3f(0.99004072f, 0.140781626f, 0.f),
    C_Vertex3f(0.593919992f, 0.395184994f, 0.f),
    C_Vertex3f(0.593919992f, 0.395184994f, -1.f),
    C_Normal3f(0.995179772f, 0.0980677232f, 0.f),
    C_Vertex3f(0.596276402f, 0.3755548f, 0.f),
    C_Vertex3f(0.596276402f, 0.3755548f, -1.f),
    C_Normal3f(0.998355806f, 0.057320077f, 0.f),
    C_Vertex3f(0.597847342f, 0.355330706f, 0.f),
    C_Vertex3f(0.597847342f, 0.355330706f, -1.f),
    C_Normal3f(0.999839008f, 0.0179467779f, 0.f),
    C_Vertex3f(0.598632813f, 0.334512711f, 0.f),
    C_Vertex3f(0.598632813f, 0.334512711f, -1.f),
    C_Normal3f(0.999783099f, -0.0208266824f, 0.f),
    C_Vertex3f(0.598632813f, 0.311571121f, 0.f),
    C_Vertex3f(0.598632813f, 0.311571121f, -1.f),
    C_Normal3f(0.998038113f, -0.0626095235f, 0.f),
    C_Vertex3f(0.597662926f, 0.287953377f, 0.f),
    C_Vertex3f(0.597662926f, 0.287953377f, -1.f),
    C_Normal3f(0.994159997f, -0.107915543f, 0.f),
    C_Vertex3f(0.595723152f, 0.265189171f, 0.f),
    C_Vertex3f(0.595723152f, 0.265189171f, -1.f),
    C_Normal3f(0.987745941f, -0.15607065f, 0.f),
    C_Vertex3f(0.592813492f, 0.243278503f, 0.f),
    C_Vertex3f(0.592813492f, 0.243278503f, -1.f),
    C_Normal3f(0.97834605f, -0.206975654f, 0.f),
    C_Vertex3f(0.588933945f, 0.222221375f, 0.f),
    C_Vertex3f(0.588933945f, 0.222221375f, -1.f),
    C_Normal3f(0.96549499f, -0.260421753f, 0.f),
    C_Vertex3f(0.584084511f, 0.202017784f, 0.f),
    C_Vertex3f(0.584084511f, 0.202017784f, -1.f),
    C_Normal3f(0.948735952f, -0.31606999f, 0.f),
    C_Vertex3f(0.57826519f, 0.182667732f, 0.f),
    C_Vertex3f(0.57826519f, 0.182667732f, -1.f),
    C_Normal3f(0.927655578f, -0.373436749f, 0.f),
    C_Vertex3f(0.571475983f, 0.164171219f, 0.f),
    C_Vertex3f(0.571475983f, 0.164171219f, -1.f),
    C_Normal3f(0.906205952f, -0.422836721f, 0.f),
    C_Vertex3f(0.563716888f, 0.146528244f, 0.f),
    C_Vertex3f(0.563716888f, 0.146528244f, -1.f),
    C_Normal3f(0.884895802f, -0.46578899f, 0.f),
    C_Vertex3f(0.555575371f, 0.130093694f, 0.f),
    C_Vertex3f(0.555575371f, 0.130093694f, -1.f),
    C_Normal3f(0.858992398f, -0.511988401f, 0.f),
    C_Vertex3f(0.54705143f, 0.114867568f, 0.f),
    C_Vertex3f(0.54705143f, 0.114867568f, -1.f),
    C_Normal3f(0.827602863f, -0.561314046f, 0.f),
    C_Vertex3f(0.538145065f, 0.100849867f, 0.f),
    C_Vertex3f(0.538145065f, 0.100849867f, -1.f),
    C_Normal3f(0.789754331f, -0.613423288f, 0.f),
    C_Vertex3f(0.528856277f, 0.0880405903f, 0.f),
    C_Vertex3f(0.528856277f, 0.0880405903f, -1.f),
    C_Normal3f(0.744456172f, -0.667671382f, 0.f),
    C_Vertex3f(0.519185066f, 0.0764397383f, 0.f),
    C_Vertex3f(0.519185066f, 0.0764397383f, -1.f),
    C_Normal3f(0.690809667f, -0.723036706f, 0.f),
    C_Vertex3f(0.509131432f, 0.0660473108f, 0.f),
    C_Vertex3f(0.509131432f, 0.0660473108f, -1.f),
    C_Normal3f(0.621398926f, -0.783494294f, 0.f),
    C_Vertex3f(0.498695374f, 0.056863308f, 0.f),
    C_Vertex3f(0.498695374f, 0.056863308f, -1.f),
    C_Normal3f(0.544120252f, -0.839007318f, 0.f),
    C_Vertex3f(0.482467651f, 0.0448999405f, 0.f),
    C_Vertex3f(0.482467651f, 0.0448999405f, -1.f),
    C_Normal3f(0.460735172f, -0.887537599f, 0.f),
    C_Vertex3f(0.459191322f, 0.0312438011f, 0.f),
    C_Vertex3f(0.459191322f, 0.0312438011f, -1.f),
    C_Normal3f(0.368881762f, -0.929476321f, 0.f),
    C_Vertex3f(0.434275627f, 0.019882679f, 0.f),
    C_Vertex3f(0.434275627f, 0.019882679f, -1.f),
    C_Normal3f(0.287761837f, -0.9577021f, 0.f),
    C_Vertex3f(0.407720566f, 0.0108165741f, 0.f),
    C_Vertex3f(0.407720566f, 0.0108165741f, -1.f),
    C_Normal3f(0.217011601f, -0.97616899f, 0.f),
    C_Vertex3f(0.389743805f, 0.00650215149f, 0.f),
    C_Vertex3f(0.389743805f, 0.00650215149f, -1.f),
    C_Normal3f(0.161317781f, -0.986902475f, 0.f),
    C_Vertex3f(0.381001234f, 0.00487661362f, 0.f),
    C_Vertex3f(0.381001234f, 0.00487661362f, -1.f),
    C_Normal3f(0.12403331f, -0.992278039f, 0.f),
    C_Vertex3f(0.371275187f, 0.00348329544f, 0.f),
    C_Vertex3f(0.371275187f, 0.00348329544f, -1.f),
    C_Normal3f(0.092888698f, -0.995676517f, 0.f),
    C_Vertex3f(0.360565662f, 0.00232219696f, 0.f),
    C_Vertex3f(0.360565662f, 0.00232219696f, -1.f),
    C_Normal3f(0.0665559396f, -0.997782707f, 0.f),
    C_Vertex3f(0.348872662f, 0.00139331818f, 0.f),
    C_Vertex3f(0.348872662f, 0.00139331818f, -1.f),
    C_Normal3f(0.0440443754f, -0.999029577f, 0.f),
    C_Vertex3f(0.336196184f, 0.000696659088f, 0.f),
    C_Vertex3f(0.336196184f, 0.000696659088f, -1.f),
    C_Normal3f(0.0246065333f, -0.999697268f, 0.f),
    C_Vertex3f(0.32253623f, 0.000232219696f, 0.f),
    C_Vertex3f(0.32253623f, 0.000232219696f, -1.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.307892799f, 0.f, 0.f),
    C_Vertex3f(0.307892799f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.307892799f, 0.f, 0.f),
    C_Vertex3f(0.307892799f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, 0.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, -1.f),
    C_Normal3f(-0.00814025756f, 0.999966979f, 0.f),
    C_Vertex3f(0.306522369f, 0.0756835938f, 0.f),
    C_Vertex3f(0.306522369f, 0.0756835938f, -1.f),
    C_Normal3f(-0.082688421f, 0.996575534f, 0.f),
    C_Vertex3f(0.334924698f, 0.0771579742f, 0.f),
    C_Vertex3f(0.334924698f, 0.0771579742f, -1.f),
    C_Normal3f(-0.15718627f, 0.987568974f, 0.f),
    C_Vertex3f(0.359830856f, 0.0801067352f, 0.f),
    C_Vertex3f(0.359830856f, 0.0801067352f, -1.f),
    C_Normal3f(-0.253855884f, 0.967242062f, 0.f),
    C_Vertex3f(0.381240845f, 0.0845298767f, 0.f),
    C_Vertex3f(0.381240845f, 0.0845298767f, -1.f),
    C_Normal3f(-0.355997503f, 0.934486866f, 0.f),
    C_Vertex3f(0.399154663f, 0.0904273987f, 0.f),
    C_Vertex3f(0.399154663f, 0.0904273987f, -1.f),
    C_Normal3f(-0.449700534f, 0.893179417f, 0.f),
    C_Vertex3f(0.415211678f, 0.0974712372f, 0.f),
    C_Vertex3f(0.415211678f, 0.0974712372f, -1.f),
    C_Normal3f(-0.551013887f, 0.834496081f, 0.f),
    C_Vertex3f(0.429411888f, 0.105661392f, 0.f),
    C_Vertex3f(0.429411888f, 0.105661392f, -1.f),
    C_Normal3f(-0.660698533f, 0.7506513f, 0.f),
    C_Vertex3f(0.441755295f, 0.114997864f, 0.f),
    C_Vertex3f(0.441755295f, 0.114997864f, -1.f),
    C_Normal3f(-0.757438898f, 0.652906001f, 0.f),
    C_Vertex3f(0.454536438f, 0.127775192f, 0.f),
    C_Vertex3f(0.454536438f, 0.127775192f, -1.f),
    C_Normal3f(-0.831987381f, 0.55479455f, 0.f),
    C_Vertex3f(0.467973232f, 0.145413399f, 0.f),
    C_Vertex3f(0.467973232f, 0.145413399f, -1.f),
    C_Normal3f(-0.89050442f, 0.454974771f, 0.f),
    C_Vertex3f(0.479771137f, 0.165617943f, 0.f),
    C_Vertex3f(0.479771137f, 0.165617943f, -1.f),
    C_Normal3f(-0.932198048f, 0.361948609f, 0.f),
    C_Vertex3f(0.489930153f, 0.188388824f, 0.f),
    C_Vertex3f(0.489930153f, 0.188388824f, -1.f),
    C_Normal3f(-0.963236451f, 0.268654913f, 0.f),
    C_Vertex3f(0.498450279f, 0.213726044f, 0.f),
    C_Vertex3f(0.498450279f, 0.213726044f, -1.f),
    C_Normal3f(-0.984110534f, 0.177556783f, 0.f),
    C_Vertex3f(0.504840374f, 0.241847992f, 0.f),
    C_Vertex3f(0.504840374f, 0.241847992f, -1.f),
    C_Normal3f(-0.995142937f, 0.0984403491f, 0.f),
    C_Vertex3f(0.509100437f, 0.272754669f, 0.f),
    C_Vertex3f(0.509100437f, 0.272754669f, -1.f),
    C_Normal3f(-0.999397278f, 0.0347146802f, 0.f),
    C_Vertex3f(0.511230469f, 0.306446075f, 0.f),
    C_Vertex3f(0.511230469f, 0.306446075f, -1.f),
    C_Normal3f(-0.999933124f, -0.0115605872f, 0.f),
    C_Vertex3f(0.511230469f, 0.334075928f, 0.f),
    C_Vertex3f(0.511230469f, 0.334075928f, -1.f),
    C_Normal3f(-0.998971105f, -0.0453496613f, 0.f),
    C_Vertex3f(0.510704756f, 0.351917624f, 0.f),
    C_Vertex3f(0.510704756f, 0.351917624f, -1.f),
    C_Normal3f(-0.996815503f, -0.0797435045f, 0.f),
    C_Vertex3f(0.50965333f, 0.368817449f, 0.f),
    C_Vertex3f(0.50965333f, 0.368817449f, -1.f),
    C_Normal3f(-0.993016124f, -0.117979161f, 0.f),
    C_Vertex3f(0.508076191f, 0.3847754f, 0.f),
    C_Vertex3f(0.508076191f, 0.3847754f, -1.f),
    C_Normal3f(-0.98702997f, -0.160536408f, 0.f),
    C_Vertex3f(0.505973339f, 0.399791479f, 0.f),
    C_Vertex3f(0.505973339f, 0.399791479f, -1.f),
    C_Normal3f(-0.978148222f, -0.207908735f, 0.f),
    C_Vertex3f(0.503344774f, 0.413865685f, 0.f),
    C_Vertex3f(0.503344774f, 0.413865685f, -1.f),
    C_Normal3f(-0.965456784f, -0.260563284f, 0.f),
    C_Vertex3f(0.500190496f, 0.426998019f, 0.f),
    C_Vertex3f(0.500190496f, 0.426998019f, -1.f),
    C_Normal3f(-0.945706367f, -0.325022191f, 0.f),
    C_Vertex3f(0.496510506f, 0.43918848f, 0.f),
    C_Vertex3f(0.496510506f, 0.43918848f, -1.f),
    C_Normal3f(-0.918335497f, -0.395802975f, 0.f),
    C_Vertex3f(0.49020195f, 0.456061363f, 0.f),
    C_Vertex3f(0.49020195f, 0.456061363f, -1.f),
    C_Normal3f(-0.87987107f, -0.475212574f, 0.f),
    C_Vertex3f(0.480506897f, 0.47631979f, 0.f),
    C_Vertex3f(0.480506897f, 0.47631979f, -1.f),
    C_Normal3f(-0.824006319f, -0.566580594f, 0.f),
    C_Vertex3f(0.469528198f, 0.494339466f, 0.f),
    C_Vertex3f(0.469528198f, 0.494339466f, -1.f),
    C_Normal3f(-0.750662684f, -0.660685718f, 0.f),
    C_Vertex3f(0.457265854f, 0.510120392f, 0.f),
    C_Vertex3f(0.457265854f, 0.510120392f, -1.f),
    C_Normal3f(-0.668055713f, -0.74411124f, 0.f),
    C_Vertex3f(0.443719864f, 0.523662567f, 0.f),
    C_Vertex3f(0.443719864f, 0.523662567f, -1.f),
    C_Normal3f(-0.578836143f, -0.815443933f, 0.f),
    C_Vertex3f(0.429409504f, 0.535129547f, 0.f),
    C_Vertex3f(0.429409504f, 0.535129547f, -1.f),
    C_Normal3f(-0.475477308f, -0.879728019f, 0.f),
    C_Vertex3f(0.414334774f, 0.544521332f, 0.f),
    C_Vertex3f(0.414334774f, 0.544521332f, -1.f),
    C_Normal3f(-0.369434804f, -0.929256618f, 0.f),
    C_Vertex3f(0.398495674f, 0.551837921f, 0.f),
    C_Vertex3f(0.398495674f, 0.551837921f, -1.f),
    C_Normal3f(-0.273841292f, -0.961774945f, 0.f),
    C_Vertex3f(0.386043072f, 0.555768967f, 0.f),
    C_Vertex3f(0.386043072f, 0.555768967f, -1.f),
    C_Normal3f(-0.209050685f, -0.977904737f, 0.f),
    C_Vertex3f(0.376635432f, 0.558062077f, 0.f),
    C_Vertex3f(0.376635432f, 0.558062077f, -1.f),
    C_Normal3f(-0.160693556f, -0.987004399f, 0.f),
    C_Vertex3f(0.366121888f, 0.560027599f, 0.f),
    C_Vertex3f(0.366121888f, 0.560027599f, -1.f),
    C_Normal3f(-0.120226838f, -0.992746472f, 0.f),
    C_Vertex3f(0.354502439f, 0.561665535f, 0.f),
    C_Vertex3f(0.354502439f, 0.561665535f, -1.f),
    C_Normal3f(-0.0860278681f, -0.99629271f, 0.f),
    C_Vertex3f(0.341777086f, 0.562975883f, 0.f),
    C_Vertex3f(0.341777086f, 0.562975883f, -1.f),
    C_Normal3f(-0.0568431392f, -0.998383164f, 0.f),
    C_Vertex3f(0.327945828f, 0.563958645f, 0.f),
    C_Vertex3f(0.327945828f, 0.563958645f, -1.f),
    C_Normal3f(-0.0317062661f, -0.999497294f, 0.f),
    C_Vertex3f(0.313008666f, 0.564613819f, 0.f),
    C_Vertex3f(0.313008666f, 0.564613819f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.296965599f, 0.564941406f, 0.f),
    C_Vertex3f(0.296965599f, 0.564941406f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.296965599f, 0.564941406f, 0.f),
    C_Vertex3f(0.296965599f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.349121094f, -1.f),
    C_Vertex3f(0.153808594f, 0.349121094f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.153808594f, 0.349121094f, 0.f),
    C_Vertex3f(0.153808594f, 0.349121094f, -1.f),
    C_Vertex3f(0.330566406f, 0.349121094f, -1.f),
    C_Vertex3f(0.330566406f, 0.349121094f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.330566406f, 0.349121094f, 0.f),
    C_Vertex3f(0.330566406f, 0.349121094f, -1.f),
    C_Vertex3f(0.330566406f, 0.291503906f, -1.f),
    C_Vertex3f(0.330566406f, 0.291503906f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.330566406f, 0.291503906f, 0.f),
    C_Vertex3f(0.330566406f, 0.291503906f, -1.f),
    C_Vertex3f(0.153808594f, 0.291503906f, -1.f),
    C_Vertex3f(0.153808594f, 0.291503906f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.291503906f, 0.f),
    C_Vertex3f(0.153808594f, 0.291503906f, -1.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, -1.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(-0.00146484375f, 0.349121094f, 0.f),
    C_Vertex3f(0.0688476563f, 0.291503906f, 0.f),
    C_Vertex3f(0.0688476563f, 0.349121094f, 0.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.299320221f, 0.640625f, 0.f),
    C_Vertex3f(0.296965599f, 0.564941406f, 0.f),
    C_Vertex3f(0.313008666f, 0.564613819f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.510704756f, 0.351917624f, 0.f),
    C_Vertex3f(0.511230469f, 0.334075928f, 0.f),
    C_Vertex3f(0.515873075f, 0.568760395f, 0.f),
    C_Vertex3f(0.519185066f, 0.0764397383f, 0.f),
    C_Vertex3f(0.526951194f, 0.556565762f, 0.f),
    C_Vertex3f(0.528856277f, 0.0880405903f, 0.f),
    C_Vertex3f(0.537230253f, 0.543626785f, 0.f),
    C_Vertex3f(0.538145065f, 0.100849867f, 0.f),
    C_Vertex3f(0.546710253f, 0.529943466f, 0.f),
    C_Vertex3f(0.54705143f, 0.114867568f, 0.f),
    C_Vertex3f(0.555391192f, 0.515515804f, 0.f),
    C_Vertex3f(0.555575371f, 0.130093694f, 0.f),
    C_Vertex3f(0.563273072f, 0.5003438f, 0.f),
    C_Vertex3f(0.563716888f, 0.146528244f, 0.f),
    C_Vertex3f(0.570355892f, 0.484427452f, 0.f),
    C_Vertex3f(0.571475983f, 0.164171219f, 0.f),
    C_Vertex3f(0.576639652f, 0.467766762f, 0.f),
    C_Vertex3f(0.57826519f, 0.182667732f, 0.f),
    C_Vertex3f(0.582137942f, 0.450512171f, 0.f),
    C_Vertex3f(0.584084511f, 0.202017784f, 0.f),
    C_Vertex3f(0.586850762f, 0.432663679f, 0.f),
    C_Vertex3f(0.588933945f, 0.222221375f, 0.f),
    C_Vertex3f(0.590778112f, 0.414221287f, 0.f),
    C_Vertex3f(0.592813492f, 0.243278503f, 0.f),
    C_Vertex3f(0.593919992f, 0.395184994f, 0.f),
    C_Vertex3f(0.595723152f, 0.265189171f, 0.f),
    C_Vertex3f(0.596276402f, 0.3755548f, 0.f),
    C_Vertex3f(0.597662926f, 0.287953377f, 0.f),
    C_Vertex3f(0.597847342f, 0.355330706f, 0.f),
    C_Vertex3f(0.598632813f, 0.311571121f, 0.f),
    C_Vertex3f(0.598632813f, 0.334512711f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.515873075f, 0.568760395f, 0.f),
    C_Vertex3f(0.503995895f, 0.580210686f, 0.f),
    C_Vertex3f(0.505973339f, 0.399791479f, 0.f),
    C_Vertex3f(0.508076191f, 0.3847754f, 0.f),
    C_Vertex3f(0.50965333f, 0.368817449f, 0.f),
    C_Vertex3f(0.510704756f, 0.351917624f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.503995895f, 0.580210686f, 0.f),
    C_Vertex3f(0.487932205f, 0.593866348f, 0.f),
    C_Vertex3f(0.49020195f, 0.456061363f, 0.f),
    C_Vertex3f(0.496510506f, 0.43918848f, 0.f),
    C_Vertex3f(0.500190496f, 0.426998019f, 0.f),
    C_Vertex3f(0.503344774f, 0.413865685f, 0.f),
    C_Vertex3f(0.505973339f, 0.399791479f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.443719864f, 0.523662567f, 0.f),
    C_Vertex3f(0.457265854f, 0.510120392f, 0.f),
    C_Vertex3f(0.466459274f, 0.607904434f, 0.f),
    C_Vertex3f(0.469528198f, 0.494339466f, 0.f),
    C_Vertex3f(0.487932205f, 0.593866348f, 0.f),
    C_Vertex3f(0.480506897f, 0.47631979f, 0.f),
    C_Vertex3f(0.49020195f, 0.456061363f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.376635432f, 0.558062077f, 0.f),
    C_Vertex3f(0.386043072f, 0.555768967f, 0.f),
    C_Vertex3f(0.387834549f, 0.634601831f, 0.f),
    C_Vertex3f(0.398495674f, 0.551837921f, 0.f),
    C_Vertex3f(0.398761749f, 0.632594109f, 0.f),
    C_Vertex3f(0.414334774f, 0.544521332f, 0.f),
    C_Vertex3f(0.417448044f, 0.628278732f, 0.f),
    C_Vertex3f(0.429409504f, 0.535129547f, 0.f),
    C_Vertex3f(0.442964554f, 0.619375229f, 0.f),
    C_Vertex3f(0.443719864f, 0.523662567f, 0.f),
    C_Vertex3f(0.466459274f, 0.607904434f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.299320221f, 0.640625f, 0.f),
    C_Vertex3f(0.313008666f, 0.564613819f, 0.f),
    C_Vertex3f(0.316804886f, 0.640338182f, 0.f),
    C_Vertex3f(0.327945828f, 0.563958645f, 0.f),
    C_Vertex3f(0.33319664f, 0.639764547f, 0.f),
    C_Vertex3f(0.341777086f, 0.562975883f, 0.f),
    C_Vertex3f(0.348495483f, 0.638904095f, 0.f),
    C_Vertex3f(0.354502439f, 0.561665535f, 0.f),
    C_Vertex3f(0.362701416f, 0.637756824f, 0.f),
    C_Vertex3f(0.366121888f, 0.560027599f, 0.f),
    C_Vertex3f(0.375814438f, 0.636322737f, 0.f),
    C_Vertex3f(0.376635432f, 0.558062077f, 0.f),
    C_Vertex3f(0.387834549f, 0.634601831f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.153808594f, 0.291503906f, 0.f),
    C_Vertex3f(0.330566406f, 0.291503906f, 0.f),
    C_Vertex3f(0.153808594f, 0.349121094f, 0.f),
    C_Vertex3f(0.153808594f, 0.564941406f, 0.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.511230469f, 0.306446075f, 0.f),
    C_Vertex3f(0.509100437f, 0.272754669f, 0.f),
    C_Vertex3f(0.509131432f, 0.0660473108f, 0.f),
    C_Vertex3f(0.519185066f, 0.0764397383f, 0.f),
    C_Vertex3f(0.511230469f, 0.334075928f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.504840374f, 0.241847992f, 0.f),
    C_Vertex3f(0.498450279f, 0.213726044f, 0.f),
    C_Vertex3f(0.498695374f, 0.056863308f, 0.f),
    C_Vertex3f(0.509131432f, 0.0660473108f, 0.f),
    C_Vertex3f(0.509100437f, 0.272754669f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.489930153f, 0.188388824f, 0.f),
    C_Vertex3f(0.479771137f, 0.165617943f, 0.f),
    C_Vertex3f(0.482467651f, 0.0448999405f, 0.f),
    C_Vertex3f(0.498695374f, 0.056863308f, 0.f),
    C_Vertex3f(0.498450279f, 0.213726044f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.467973232f, 0.145413399f, 0.f),
    C_Vertex3f(0.454536438f, 0.127775192f, 0.f),
    C_Vertex3f(0.459191322f, 0.0312438011f, 0.f),
    C_Vertex3f(0.482467651f, 0.0448999405f, 0.f),
    C_Vertex3f(0.479771137f, 0.165617943f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.441755295f, 0.114997864f, 0.f),
    C_Vertex3f(0.429411888f, 0.105661392f, 0.f),
    C_Vertex3f(0.434275627f, 0.019882679f, 0.f),
    C_Vertex3f(0.459191322f, 0.0312438011f, 0.f),
    C_Vertex3f(0.454536438f, 0.127775192f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.371275187f, 0.00348329544f, 0.f),
    C_Vertex3f(0.381001234f, 0.00487661362f, 0.f),
    C_Vertex3f(0.381240845f, 0.0845298767f, 0.f),
    C_Vertex3f(0.389743805f, 0.00650215149f, 0.f),
    C_Vertex3f(0.399154663f, 0.0904273987f, 0.f),
    C_Vertex3f(0.407720566f, 0.0108165741f, 0.f),
    C_Vertex3f(0.415211678f, 0.0974712372f, 0.f),
    C_Vertex3f(0.434275627f, 0.019882679f, 0.f),
    C_Vertex3f(0.429411888f, 0.105661392f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.336196184f, 0.000696659088f, 0.f),
    C_Vertex3f(0.348872662f, 0.00139331818f, 0.f),
    C_Vertex3f(0.359830856f, 0.0801067352f, 0.f),
    C_Vertex3f(0.360565662f, 0.00232219696f, 0.f),
    C_Vertex3f(0.381240845f, 0.0845298767f, 0.f),
    C_Vertex3f(0.371275187f, 0.00348329544f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.334924698f, 0.0771579742f, 0.f),
    C_Vertex3f(0.306522369f, 0.0756835938f, 0.f),
    C_Vertex3f(0.307892799f, 0.f, 0.f),
    C_Vertex3f(0.32253623f, 0.000232219696f, 0.f),
    C_Vertex3f(0.336196184f, 0.000696659088f, 0.f),
    C_Vertex3f(0.359830856f, 0.0801067352f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.153808594f, 0.0756835938f, 0.f),
    C_Vertex3f(0.0688476563f, 0.291503906f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_Vertex3f(0.307892799f, 0.f, 0.f),
    C_Vertex3f(0.306522369f, 0.0756835938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.153808594f, 0.349121094f, 0.f),
    C_Vertex3f(0.330566406f, 0.291503906f, 0.f),
    C_Vertex3f(0.330566406f, 0.349121094f, 0.f),
    C_Vertex3f(0.0688476563f, 0.291503906f, 0.f),
    C_Vertex3f(-0.00146484375f, 0.349121094f, 0.f),
    C_Vertex3f(-0.00146484375f, 0.291503906f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(-0.00146484375f, 0.349121094f, -1.f),
    C_Vertex3f(0.0688476563f, 0.291503906f, -1.f),
    C_Vertex3f(0.0688476563f, 0.349121094f, -1.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.299320221f, 0.640625f, -1.f),
    C_Vertex3f(0.296965599f, 0.564941406f, -1.f),
    C_Vertex3f(0.313008666f, 0.564613819f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.510704756f, 0.351917624f, -1.f),
    C_Vertex3f(0.511230469f, 0.334075928f, -1.f),
    C_Vertex3f(0.515873075f, 0.568760395f, -1.f),
    C_Vertex3f(0.519185066f, 0.0764397383f, -1.f),
    C_Vertex3f(0.526951194f, 0.556565762f, -1.f),
    C_Vertex3f(0.528856277f, 0.0880405903f, -1.f),
    C_Vertex3f(0.537230253f, 0.543626785f, -1.f),
    C_Vertex3f(0.538145065f, 0.100849867f, -1.f),
    C_Vertex3f(0.546710253f, 0.529943466f, -1.f),
    C_Vertex3f(0.54705143f, 0.114867568f, -1.f),
    C_Vertex3f(0.555391192f, 0.515515804f, -1.f),
    C_Vertex3f(0.555575371f, 0.130093694f, -1.f),
    C_Vertex3f(0.563273072f, 0.5003438f, -1.f),
    C_Vertex3f(0.563716888f, 0.146528244f, -1.f),
    C_Vertex3f(0.570355892f, 0.484427452f, -1.f),
    C_Vertex3f(0.571475983f, 0.164171219f, -1.f),
    C_Vertex3f(0.576639652f, 0.467766762f, -1.f),
    C_Vertex3f(0.57826519f, 0.182667732f, -1.f),
    C_Vertex3f(0.582137942f, 0.450512171f, -1.f),
    C_Vertex3f(0.584084511f, 0.202017784f, -1.f),
    C_Vertex3f(0.586850762f, 0.432663679f, -1.f),
    C_Vertex3f(0.588933945f, 0.222221375f, -1.f),
    C_Vertex3f(0.590778112f, 0.414221287f, -1.f),
    C_Vertex3f(0.592813492f, 0.243278503f, -1.f),
    C_Vertex3f(0.593919992f, 0.395184994f, -1.f),
    C_Vertex3f(0.595723152f, 0.265189171f, -1.f),
    C_Vertex3f(0.596276402f, 0.3755548f, -1.f),
    C_Vertex3f(0.597662926f, 0.287953377f, -1.f),
    C_Vertex3f(0.597847342f, 0.355330706f, -1.f),
    C_Vertex3f(0.598632813f, 0.311571121f, -1.f),
    C_Vertex3f(0.598632813f, 0.334512711f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.515873075f, 0.568760395f, -1.f),
    C_Vertex3f(0.503995895f, 0.580210686f, -1.f),
    C_Vertex3f(0.505973339f, 0.399791479f, -1.f),
    C_Vertex3f(0.508076191f, 0.3847754f, -1.f),
    C_Vertex3f(0.50965333f, 0.368817449f, -1.f),
    C_Vertex3f(0.510704756f, 0.351917624f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.503995895f, 0.580210686f, -1.f),
    C_Vertex3f(0.487932205f, 0.593866348f, -1.f),
    C_Vertex3f(0.49020195f, 0.456061363f, -1.f),
    C_Vertex3f(0.496510506f, 0.43918848f, -1.f),
    C_Vertex3f(0.500190496f, 0.426998019f, -1.f),
    C_Vertex3f(0.503344774f, 0.413865685f, -1.f),
    C_Vertex3f(0.505973339f, 0.399791479f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.443719864f, 0.523662567f, -1.f),
    C_Vertex3f(0.457265854f, 0.510120392f, -1.f),
    C_Vertex3f(0.466459274f, 0.607904434f, -1.f),
    C_Vertex3f(0.469528198f, 0.494339466f, -1.f),
    C_Vertex3f(0.487932205f, 0.593866348f, -1.f),
    C_Vertex3f(0.480506897f, 0.47631979f, -1.f),
    C_Vertex3f(0.49020195f, 0.456061363f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.376635432f, 0.558062077f, -1.f),
    C_Vertex3f(0.386043072f, 0.555768967f, -1.f),
    C_Vertex3f(0.387834549f, 0.634601831f, -1.f),
    C_Vertex3f(0.398495674f, 0.551837921f, -1.f),
    C_Vertex3f(0.398761749f, 0.632594109f, -1.f),
    C_Vertex3f(0.414334774f, 0.544521332f, -1.f),
    C_Vertex3f(0.417448044f, 0.628278732f, -1.f),
    C_Vertex3f(0.429409504f, 0.535129547f, -1.f),
    C_Vertex3f(0.442964554f, 0.619375229f, -1.f),
    C_Vertex3f(0.443719864f, 0.523662567f, -1.f),
    C_Vertex3f(0.466459274f, 0.607904434f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.299320221f, 0.640625f, -1.f),
    C_Vertex3f(0.313008666f, 0.564613819f, -1.f),
    C_Vertex3f(0.316804886f, 0.640338182f, -1.f),
    C_Vertex3f(0.327945828f, 0.563958645f, -1.f),
    C_Vertex3f(0.33319664f, 0.639764547f, -1.f),
    C_Vertex3f(0.341777086f, 0.562975883f, -1.f),
    C_Vertex3f(0.348495483f, 0.638904095f, -1.f),
    C_Vertex3f(0.354502439f, 0.561665535f, -1.f),
    C_Vertex3f(0.362701416f, 0.637756824f, -1.f),
    C_Vertex3f(0.366121888f, 0.560027599f, -1.f),
    C_Vertex3f(0.375814438f, 0.636322737f, -1.f),
    C_Vertex3f(0.376635432f, 0.558062077f, -1.f),
    C_Vertex3f(0.387834549f, 0.634601831f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.153808594f, 0.291503906f, -1.f),
    C_Vertex3f(0.330566406f, 0.291503906f, -1.f),
    C_Vertex3f(0.153808594f, 0.349121094f, -1.f),
    C_Vertex3f(0.153808594f, 0.564941406f, -1.f),
    C_Vertex3f(0.153808594f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.511230469f, 0.306446075f, -1.f),
    C_Vertex3f(0.509100437f, 0.272754669f, -1.f),
    C_Vertex3f(0.509131432f, 0.0660473108f, -1.f),
    C_Vertex3f(0.519185066f, 0.0764397383f, -1.f),
    C_Vertex3f(0.511230469f, 0.334075928f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.504840374f, 0.241847992f, -1.f),
    C_Vertex3f(0.498450279f, 0.213726044f, -1.f),
    C_Vertex3f(0.498695374f, 0.056863308f, -1.f),
    C_Vertex3f(0.509131432f, 0.0660473108f, -1.f),
    C_Vertex3f(0.509100437f, 0.272754669f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.489930153f, 0.188388824f, -1.f),
    C_Vertex3f(0.479771137f, 0.165617943f, -1.f),
    C_Vertex3f(0.482467651f, 0.0448999405f, -1.f),
    C_Vertex3f(0.498695374f, 0.056863308f, -1.f),
    C_Vertex3f(0.498450279f, 0.213726044f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.467973232f, 0.145413399f, -1.f),
    C_Vertex3f(0.454536438f, 0.127775192f, -1.f),
    C_Vertex3f(0.459191322f, 0.0312438011f, -1.f),
    C_Vertex3f(0.482467651f, 0.0448999405f, -1.f),
    C_Vertex3f(0.479771137f, 0.165617943f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.441755295f, 0.114997864f, -1.f),
    C_Vertex3f(0.429411888f, 0.105661392f, -1.f),
    C_Vertex3f(0.434275627f, 0.019882679f, -1.f),
    C_Vertex3f(0.459191322f, 0.0312438011f, -1.f),
    C_Vertex3f(0.454536438f, 0.127775192f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.371275187f, 0.00348329544f, -1.f),
    C_Vertex3f(0.381001234f, 0.00487661362f, -1.f),
    C_Vertex3f(0.381240845f, 0.0845298767f, -1.f),
    C_Vertex3f(0.389743805f, 0.00650215149f, -1.f),
    C_Vertex3f(0.399154663f, 0.0904273987f, -1.f),
    C_Vertex3f(0.407720566f, 0.0108165741f, -1.f),
    C_Vertex3f(0.415211678f, 0.0974712372f, -1.f),
    C_Vertex3f(0.434275627f, 0.019882679f, -1.f),
    C_Vertex3f(0.429411888f, 0.105661392f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.336196184f, 0.000696659088f, -1.f),
    C_Vertex3f(0.348872662f, 0.00139331818f, -1.f),
    C_Vertex3f(0.359830856f, 0.0801067352f, -1.f),
    C_Vertex3f(0.360565662f, 0.00232219696f, -1.f),
    C_Vertex3f(0.381240845f, 0.0845298767f, -1.f),
    C_Vertex3f(0.371275187f, 0.00348329544f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.334924698f, 0.0771579742f, -1.f),
    C_Vertex3f(0.306522369f, 0.0756835938f, -1.f),
    C_Vertex3f(0.307892799f, 0.f, -1.f),
    C_Vertex3f(0.32253623f, 0.000232219696f, -1.f),
    C_Vertex3f(0.336196184f, 0.000696659088f, -1.f),
    C_Vertex3f(0.359830856f, 0.0801067352f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.153808594f, 0.0756835938f, -1.f),
    C_Vertex3f(0.0688476563f, 0.291503906f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.307892799f, 0.f, -1.f),
    C_Vertex3f(0.306522369f, 0.0756835938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.153808594f, 0.349121094f, -1.f),
    C_Vertex3f(0.330566406f, 0.291503906f, -1.f),
    C_Vertex3f(0.330566406f, 0.349121094f, -1.f),
    C_Vertex3f(0.0688476563f, 0.291503906f, -1.f),
    C_Vertex3f(-0.00146484375f, 0.349121094f, -1.f),
    C_Vertex3f(-0.00146484375f, 0.291503906f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_209[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.068359375f, 0.f, 0.f),
    C_Vertex3f(0.068359375f, 0.f, -1.f),
    C_Vertex3f(0.068359375f, 0.640625f, -1.f),
    C_Vertex3f(0.068359375f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.068359375f, 0.640625f, 0.f),
    C_Vertex3f(0.068359375f, 0.640625f, -1.f),
    C_Vertex3f(0.155105591f, 0.640625f, -1.f),
    C_Vertex3f(0.155105591f, 0.640625f, 0.f),
    C_Normal3f(0.830788314f, 0.556588352f, 0.f),
    C_Vertex3f(0.155105591f, 0.640625f, 0.f),
    C_Vertex3f(0.155105591f, 0.640625f, -1.f),
    C_Vertex3f(0.4921875f, 0.137481689f, -1.f),
    C_Vertex3f(0.4921875f, 0.137481689f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.4921875f, 0.137481689f, 0.f),
    C_Vertex3f(0.4921875f, 0.137481689f, -1.f),
    C_Vertex3f(0.4921875f, 0.640625f, -1.f),
    C_Vertex3f(0.4921875f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 0.99999994f, 0.f),
    C_Vertex3f(0.4921875f, 0.640625f, 0.f),
    C_Vertex3f(0.4921875f, 0.640625f, -1.f),
    C_Vertex3f(0.573242188f, 0.640625f, -1.f),
    C_Vertex3f(0.573242188f, 0.640625f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.573242188f, 0.640625f, 0.f),
    C_Vertex3f(0.573242188f, 0.640625f, -1.f),
    C_Vertex3f(0.573242188f, 0.f, -1.f),
    C_Vertex3f(0.573242188f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.573242188f, 0.f, 0.f),
    C_Vertex3f(0.573242188f, 0.f, -1.f),
    C_Vertex3f(0.486495972f, 0.f, -1.f),
    C_Vertex3f(0.486495972f, 0.f, 0.f),
    C_Normal3f(-0.830800116f, -0.556570947f, 0.f),
    C_Vertex3f(0.486495972f, 0.f, 0.f),
    C_Vertex3f(0.486495972f, 0.f, -1.f),
    C_Vertex3f(0.149414063f, 0.503166199f, -1.f),
    C_Vertex3f(0.149414063f, 0.503166199f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.149414063f, 0.503166199f, 0.f),
    C_Vertex3f(0.149414063f, 0.503166199f, -1.f),
    C_Vertex3f(0.149414063f, 0.f, -1.f),
    C_Vertex3f(0.149414063f, 0.f, 0.f),
    C_Normal3f(-0.f, -0.99999994f, 0.f),
    C_Vertex3f(0.149414063f, 0.f, 0.f),
    C_Vertex3f(0.149414063f, 0.f, -1.f),
    C_Vertex3f(0.068359375f, 0.f, -1.f),
    C_Vertex3f(0.068359375f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999953806f, -0.00961818546f, 0.f),
    C_Vertex3f(0.1875f, 0.678710938f, 0.f),
    C_Vertex3f(0.1875f, 0.678710938f, -1.f),
    C_Normal3f(-0.999230683f, 0.0392165966f, 0.f),
    C_Vertex3f(0.187445641f, 0.684362411f, 0.f),
    C_Vertex3f(0.187445641f, 0.684362411f, -1.f),
    C_Normal3f(-0.994498312f, 0.104752719f, 0.f),
    C_Vertex3f(0.188143015f, 0.695094824f, 0.f),
    C_Vertex3f(0.188143015f, 0.695094824f, -1.f),
    C_Normal3f(-0.981874108f, 0.189534381f, 0.f),
    C_Vertex3f(0.189646482f, 0.705256701f, 0.f),
    C_Vertex3f(0.189646482f, 0.705256701f, -1.f),
    C_Normal3f(-0.960046053f, 0.279841781f, 0.f),
    C_Vertex3f(0.191956043f, 0.714848042f, 0.f),
    C_Vertex3f(0.191956043f, 0.714848042f, -1.f),
    C_Normal3f(-0.927577436f, 0.373630971f, 0.f),
    C_Vertex3f(0.195071697f, 0.723868847f, 0.f),
    C_Vertex3f(0.195071697f, 0.723868847f, -1.f),
    C_Normal3f(-0.883691847f, 0.468069077f, 0.f),
    C_Vertex3f(0.198993444f, 0.732319117f, 0.f),
    C_Vertex3f(0.198993444f, 0.732319117f, -1.f),
    C_Normal3f(-0.828614712f, 0.559819281f, 0.f),
    C_Vertex3f(0.203721285f, 0.740198851f, 0.f),
    C_Vertex3f(0.203721285f, 0.740198851f, -1.f),
    C_Normal3f(-0.759081304f, 0.650995851f, 0.f),
    C_Vertex3f(0.209255219f, 0.747508049f, 0.f),
    C_Vertex3f(0.209255219f, 0.747508049f, -1.f),
    C_Normal3f(-0.650275886f, 0.759698212f, 0.f),
    C_Vertex3f(0.218658447f, 0.757616043f, 0.f),
    C_Vertex3f(0.218658447f, 0.757616043f, -1.f),
    C_Normal3f(-0.487445921f, 0.87315321f, 0.f),
    C_Vertex3f(0.232872963f, 0.767724037f, 0.f),
    C_Vertex3f(0.232872963f, 0.767724037f, -1.f),
    C_Normal3f(-0.287504673f, 0.957779288f, 0.f),
    C_Vertex3f(0.248835564f, 0.7744627f, 0.f),
    C_Vertex3f(0.248835564f, 0.7744627f, -1.f),
    C_Normal3f(-0.0979905054f, 0.995187461f, 0.f),
    C_Vertex3f(0.266546249f, 0.777832031f, 0.f),
    C_Vertex3f(0.266546249f, 0.777832031f, -1.f),
    C_Normal3f(0.0889112428f, 0.99603951f, 0.f),
    C_Vertex3f(0.283054352f, 0.777832031f, 0.f),
    C_Vertex3f(0.283054352f, 0.777832031f, -1.f),
    C_Normal3f(0.244741365f, 0.969588399f, 0.f),
    C_Vertex3f(0.299077988f, 0.774928093f, 0.f),
    C_Vertex3f(0.299077988f, 0.774928093f, -1.f),
    C_Normal3f(0.345433027f, 0.938443422f, 0.f),
    C_Vertex3f(0.317567825f, 0.769120216f, 0.f),
    C_Vertex3f(0.317567825f, 0.769120216f, -1.f),
    C_Normal3f(0.413061887f, 0.910703003f, 0.f),
    C_Vertex3f(0.338523865f, 0.760408401f, 0.f),
    C_Vertex3f(0.338523865f, 0.760408401f, -1.f),
    C_Normal3f(0.42238301f, 0.906417429f, 0.f),
    C_Vertex3f(0.35669136f, 0.751375198f, 0.f),
    C_Vertex3f(0.35669136f, 0.751375198f, -1.f),
    C_Normal3f(0.345709145f, 0.938341737f, 0.f),
    C_Vertex3f(0.368290901f, 0.746537209f, 0.f),
    C_Vertex3f(0.368290901f, 0.746537209f, -1.f),
    C_Normal3f(0.243631572f, 0.969867826f, 0.f),
    C_Vertex3f(0.378577232f, 0.743311882f, 0.f),
    C_Vertex3f(0.378577232f, 0.743311882f, -1.f),
    C_Normal3f(0.0984259844f, 0.995144367f, 0.f),
    C_Vertex3f(0.387550354f, 0.741699219f, 0.f),
    C_Vertex3f(0.387550354f, 0.741699219f, -1.f),
    C_Normal3f(-0.070669286f, 0.997499824f, 0.f),
    C_Vertex3f(0.394882202f, 0.741699219f, 0.f),
    C_Vertex3f(0.394882202f, 0.741699219f, -1.f),
    C_Normal3f(-0.24143213f, 0.970417738f, 0.f),
    C_Vertex3f(0.40106535f, 0.742656708f, 0.f),
    C_Vertex3f(0.40106535f, 0.742656708f, -1.f),
    C_Normal3f(-0.435192645f, 0.900337279f, 0.f),
    C_Vertex3f(0.40642786f, 0.744571686f, 0.f),
    C_Vertex3f(0.40642786f, 0.744571686f, -1.f),
    C_Normal3f(-0.62994957f, 0.776636064f, 0.f),
    C_Vertex3f(0.410969734f, 0.747444153f, 0.f),
    C_Vertex3f(0.410969734f, 0.747444153f, -1.f),
    C_Normal3f(-0.810694158f, 0.585469842f, 0.f),
    C_Vertex3f(0.414690971f, 0.751274109f, 0.f),
    C_Vertex3f(0.414690971f, 0.751274109f, -1.f),
    C_Normal3f(-0.9168607f, 0.399207145f, 0.f),
    C_Vertex3f(0.417865276f, 0.75699234f, 0.f),
    C_Vertex3f(0.417865276f, 0.75699234f, -1.f),
    C_Normal3f(-0.964133918f, 0.265416414f, 0.f),
    C_Vertex3f(0.420492649f, 0.764598846f, 0.f),
    C_Vertex3f(0.420492649f, 0.764598846f, -1.f),
    C_Normal3f(-0.982875049f, 0.18427363f, 0.f),
    C_Vertex3f(0.42257309f, 0.774093628f, 0.f),
    C_Vertex3f(0.42257309f, 0.774093628f, -1.f),
    C_Normal3f(-0.991047204f, 0.133512393f, 0.f),
    C_Vertex3f(0.423339844f, 0.779785156f, 0.f),
    C_Vertex3f(0.423339844f, 0.779785156f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.423339844f, 0.779785156f, 0.f),
    C_Vertex3f(0.423339844f, 0.779785156f, -1.f),
    C_Vertex3f(0.479980469f, 0.779785156f, -1.f),
    C_Vertex3f(0.479980469f, 0.779785156f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999653041f, -0.0263388976f, 0.f),
    C_Vertex3f(0.479980469f, 0.779785156f, 0.f),
    C_Vertex3f(0.479980469f, 0.779785156f, -1.f),
    C_Normal3f(0.99777931f, -0.066607669f, 0.f),
    C_Vertex3f(0.479816437f, 0.77355957f, 0.f),
    C_Vertex3f(0.479816437f, 0.77355957f, -1.f),
    C_Normal3f(0.992581308f, -0.121582195f, 0.f),
    C_Vertex3f(0.478785634f, 0.761886597f, 0.f),
    C_Vertex3f(0.478785634f, 0.761886597f, -1.f),
    C_Normal3f(0.980870843f, -0.194659367f, 0.f),
    C_Vertex3f(0.477052093f, 0.750991821f, 0.f),
    C_Vertex3f(0.477052093f, 0.750991821f, -1.f),
    C_Normal3f(0.961305082f, -0.275485754f, 0.f),
    C_Vertex3f(0.474615812f, 0.740875244f, 0.f),
    C_Vertex3f(0.474615812f, 0.740875244f, -1.f),
    C_Normal3f(0.931648552f, -0.363360554f, 0.f),
    C_Vertex3f(0.471476793f, 0.731536865f, 0.f),
    C_Vertex3f(0.471476793f, 0.731536865f, -1.f),
    C_Normal3f(0.889692426f, -0.456560522f, 0.f),
    C_Vertex3f(0.467635036f, 0.722976685f, 0.f),
    C_Vertex3f(0.467635036f, 0.722976685f, -1.f),
    C_Normal3f(0.833749712f, -0.552142739f, 0.f),
    C_Vertex3f(0.463090539f, 0.715194702f, 0.f),
    C_Vertex3f(0.463090539f, 0.715194702f, -1.f),
    C_Normal3f(0.755483389f, -0.655167818f, 0.f),
    C_Vertex3f(0.457843304f, 0.708190918f, 0.f),
    C_Vertex3f(0.457843304f, 0.708190918f, -1.f),
    C_Normal3f(0.639779389f, -0.768558562f, 0.f),
    C_Vertex3f(0.448918343f, 0.698852539f, 0.f),
    C_Vertex3f(0.448918343f, 0.698852539f, -1.f),
    C_Normal3f(0.477566123f, -0.878595829f, 0.f),
    C_Vertex3f(0.435407162f, 0.68951416f, 0.f),
    C_Vertex3f(0.435407162f, 0.68951416f, -1.f),
    C_Normal3f(0.281305313f, -0.95961827f, 0.f),
    C_Vertex3f(0.420284748f, 0.683288574f, 0.f),
    C_Vertex3f(0.420284748f, 0.683288574f, -1.f),
    C_Normal3f(0.0945234597f, -0.995522618f, 0.f),
    C_Vertex3f(0.403551102f, 0.680175781f, 0.f),
    C_Vertex3f(0.403551102f, 0.680175781f, -1.f),
    C_Normal3f(-0.0947503f, -0.995501101f, 0.f),
    C_Vertex3f(0.387500763f, 0.680175781f, 0.f),
    C_Vertex3f(0.387500763f, 0.680175781f, -1.f),
    C_Normal3f(-0.258822501f, -0.965924919f, 0.f),
    C_Vertex3f(0.371547699f, 0.683221817f, 0.f),
    C_Vertex3f(0.371547699f, 0.683221817f, -1.f),
    C_Normal3f(-0.367869198f, -0.92987752f, 0.f),
    C_Vertex3f(0.353397369f, 0.689313889f, 0.f),
    C_Vertex3f(0.353397369f, 0.689313889f, -1.f),
    C_Normal3f(-0.441156626f, -0.897430122f, 0.f),
    C_Vertex3f(0.333049774f, 0.698451996f, 0.f),
    C_Vertex3f(0.333049774f, 0.698451996f, -1.f),
    C_Normal3f(-0.452606946f, -0.891710103f, 0.f),
    C_Vertex3f(0.314445496f, 0.708461761f, 0.f),
    C_Vertex3f(0.314445496f, 0.708461761f, -1.f),
    C_Normal3f(-0.385828793f, -0.922570407f, 0.f),
    C_Vertex3f(0.301751137f, 0.714338303f, 0.f),
    C_Vertex3f(0.301751137f, 0.714338303f, -1.f),
    C_Normal3f(-0.288805217f, -0.957387865f, 0.f),
    C_Vertex3f(0.291026115f, 0.718255997f, 0.f),
    C_Vertex3f(0.291026115f, 0.718255997f, -1.f),
    C_Normal3f(-0.122500464f, -0.992468476f, 0.f),
    C_Vertex3f(0.282270432f, 0.720214844f, 0.f),
    C_Vertex3f(0.282270432f, 0.720214844f, -1.f),
    C_Normal3f(0.0942201167f, -0.995551467f, 0.f),
    C_Vertex3f(0.275156021f, 0.720214844f, 0.f),
    C_Vertex3f(0.275156021f, 0.720214844f, -1.f),
    C_Normal3f(0.294618219f, -0.955615103f, 0.f),
    C_Vertex3f(0.268424988f, 0.718904495f, 0.f),
    C_Vertex3f(0.268424988f, 0.718904495f, -1.f),
    C_Normal3f(0.500668526f, -0.865639091f, 0.f),
    C_Vertex3f(0.262405396f, 0.716283798f, 0.f),
    C_Vertex3f(0.262405396f, 0.716283798f, -1.f),
    C_Normal3f(0.652620137f, -0.757685304f, 0.f),
    C_Vertex3f(0.257097244f, 0.712352753f, 0.f),
    C_Vertex3f(0.257097244f, 0.712352753f, -1.f),
    C_Normal3f(0.744249165f, -0.667902052f, 0.f),
    C_Vertex3f(0.254798889f, 0.709732056f, 0.f),
    C_Vertex3f(0.254798889f, 0.709732056f, -1.f),
    C_Normal3f(0.837037444f, -0.547145605f, 0.f),
    C_Vertex3f(0.252391815f, 0.707109451f, 0.f),
    C_Vertex3f(0.252391815f, 0.707109451f, -1.f),
    C_Normal3f(0.925027549f, -0.379899919f, 0.f),
    C_Vertex3f(0.248835564f, 0.700609207f, 0.f),
    C_Vertex3f(0.248835564f, 0.700609207f, -1.f),
    C_Normal3f(0.980741322f, -0.195311442f, 0.f),
    C_Vertex3f(0.246537209f, 0.692853928f, 0.f),
    C_Vertex3f(0.246537209f, 0.692853928f, -1.f),
    C_Normal3f(0.997837007f, -0.065737471f, 0.f),
    C_Vertex3f(0.24549675f, 0.683843613f, 0.f),
    C_Vertex3f(0.24549675f, 0.683843613f, -1.f),
    C_Normal3f(0.999775767f, 0.0211769659f, 0.f),
    C_Vertex3f(0.245605469f, 0.678710938f, 0.f),
    C_Vertex3f(0.245605469f, 0.678710938f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.245605469f, 0.678710938f, 0.f),
    C_Vertex3f(0.245605469f, 0.678710938f, -1.f),
    C_Vertex3f(0.1875f, 0.678710938f, -1.f),
    C_Vertex3f(0.1875f, 0.678710938f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.068359375f, 0.f, 0.f),
    C_Vertex3f(0.149414063f, 0.f, 0.f),
    C_Vertex3f(0.068359375f, 0.640625f, 0.f),
    C_Vertex3f(0.149414063f, 0.503166199f, 0.f),
    C_Vertex3f(0.155105591f, 0.640625f, 0.f),
    C_Vertex3f(0.486495972f, 0.f, 0.f),
    C_Vertex3f(0.4921875f, 0.137481689f, 0.f),
    C_Vertex3f(0.573242188f, 0.f, 0.f),
    C_Vertex3f(0.4921875f, 0.640625f, 0.f),
    C_Vertex3f(0.573242188f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245605469f, 0.678710938f, 0.f),
    C_Vertex3f(0.24549675f, 0.683843613f, 0.f),
    C_Vertex3f(0.232872963f, 0.767724037f, 0.f),
    C_Vertex3f(0.218658447f, 0.757616043f, 0.f),
    C_Vertex3f(0.209255219f, 0.747508049f, 0.f),
    C_Vertex3f(0.203721285f, 0.740198851f, 0.f),
    C_Vertex3f(0.198993444f, 0.732319117f, 0.f),
    C_Vertex3f(0.195071697f, 0.723868847f, 0.f),
    C_Vertex3f(0.191956043f, 0.714848042f, 0.f),
    C_Vertex3f(0.189646482f, 0.705256701f, 0.f),
    C_Vertex3f(0.188143015f, 0.695094824f, 0.f),
    C_Vertex3f(0.1875f, 0.678710938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.479980469f, 0.779785156f, 0.f),
    C_Vertex3f(0.423339844f, 0.779785156f, 0.f),
    C_Vertex3f(0.435407162f, 0.68951416f, 0.f),
    C_Vertex3f(0.448918343f, 0.698852539f, 0.f),
    C_Vertex3f(0.457843304f, 0.708190918f, 0.f),
    C_Vertex3f(0.463090539f, 0.715194702f, 0.f),
    C_Vertex3f(0.467635036f, 0.722976685f, 0.f),
    C_Vertex3f(0.471476793f, 0.731536865f, 0.f),
    C_Vertex3f(0.474615812f, 0.740875244f, 0.f),
    C_Vertex3f(0.477052093f, 0.750991821f, 0.f),
    C_Vertex3f(0.478785634f, 0.761886597f, 0.f),
    C_Vertex3f(0.479816437f, 0.77355957f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.423339844f, 0.779785156f, 0.f),
    C_Vertex3f(0.42257309f, 0.774093628f, 0.f),
    C_Vertex3f(0.435407162f, 0.68951416f, 0.f),
    C_Vertex3f(0.420492649f, 0.764598846f, 0.f),
    C_Vertex3f(0.420284748f, 0.683288574f, 0.f),
    C_Vertex3f(0.417865276f, 0.75699234f, 0.f),
    C_Vertex3f(0.414690971f, 0.751274109f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.414690971f, 0.751274109f, 0.f),
    C_Vertex3f(0.410969734f, 0.747444153f, 0.f),
    C_Vertex3f(0.420284748f, 0.683288574f, 0.f),
    C_Vertex3f(0.40642786f, 0.744571686f, 0.f),
    C_Vertex3f(0.403551102f, 0.680175781f, 0.f),
    C_Vertex3f(0.40106535f, 0.742656708f, 0.f),
    C_Vertex3f(0.394882202f, 0.741699219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.368290901f, 0.746537209f, 0.f),
    C_Vertex3f(0.371547699f, 0.683221817f, 0.f),
    C_Vertex3f(0.378577232f, 0.743311882f, 0.f),
    C_Vertex3f(0.387500763f, 0.680175781f, 0.f),
    C_Vertex3f(0.387550354f, 0.741699219f, 0.f),
    C_Vertex3f(0.403551102f, 0.680175781f, 0.f),
    C_Vertex3f(0.394882202f, 0.741699219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.301751137f, 0.714338303f, 0.f),
    C_Vertex3f(0.314445496f, 0.708461761f, 0.f),
    C_Vertex3f(0.317567825f, 0.769120216f, 0.f),
    C_Vertex3f(0.333049774f, 0.698451996f, 0.f),
    C_Vertex3f(0.338523865f, 0.760408401f, 0.f),
    C_Vertex3f(0.353397369f, 0.689313889f, 0.f),
    C_Vertex3f(0.35669136f, 0.751375198f, 0.f),
    C_Vertex3f(0.371547699f, 0.683221817f, 0.f),
    C_Vertex3f(0.368290901f, 0.746537209f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.275156021f, 0.720214844f, 0.f),
    C_Vertex3f(0.282270432f, 0.720214844f, 0.f),
    C_Vertex3f(0.283054352f, 0.777832031f, 0.f),
    C_Vertex3f(0.291026115f, 0.718255997f, 0.f),
    C_Vertex3f(0.299077988f, 0.774928093f, 0.f),
    C_Vertex3f(0.301751137f, 0.714338303f, 0.f),
    C_Vertex3f(0.317567825f, 0.769120216f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.257097244f, 0.712352753f, 0.f),
    C_Vertex3f(0.262405396f, 0.716283798f, 0.f),
    C_Vertex3f(0.266546249f, 0.777832031f, 0.f),
    C_Vertex3f(0.268424988f, 0.718904495f, 0.f),
    C_Vertex3f(0.283054352f, 0.777832031f, 0.f),
    C_Vertex3f(0.275156021f, 0.720214844f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.246537209f, 0.692853928f, 0.f),
    C_Vertex3f(0.248835564f, 0.700609207f, 0.f),
    C_Vertex3f(0.248835564f, 0.7744627f, 0.f),
    C_Vertex3f(0.252391815f, 0.707109451f, 0.f),
    C_Vertex3f(0.266546249f, 0.777832031f, 0.f),
    C_Vertex3f(0.254798889f, 0.709732056f, 0.f),
    C_Vertex3f(0.257097244f, 0.712352753f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.248835564f, 0.7744627f, 0.f),
    C_Vertex3f(0.232872963f, 0.767724037f, 0.f),
    C_Vertex3f(0.24549675f, 0.683843613f, 0.f),
    C_Vertex3f(0.246537209f, 0.692853928f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.1875f, 0.678710938f, 0.f),
    C_Vertex3f(0.188143015f, 0.695094824f, 0.f),
    C_Vertex3f(0.187445641f, 0.684362411f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.068359375f, 0.f, -1.f),
    C_Vertex3f(0.149414063f, 0.f, -1.f),
    C_Vertex3f(0.068359375f, 0.640625f, -1.f),
    C_Vertex3f(0.149414063f, 0.503166199f, -1.f),
    C_Vertex3f(0.155105591f, 0.640625f, -1.f),
    C_Vertex3f(0.486495972f, 0.f, -1.f),
    C_Vertex3f(0.4921875f, 0.137481689f, -1.f),
    C_Vertex3f(0.573242188f, 0.f, -1.f),
    C_Vertex3f(0.4921875f, 0.640625f, -1.f),
    C_Vertex3f(0.573242188f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.245605469f, 0.678710938f, -1.f),
    C_Vertex3f(0.24549675f, 0.683843613f, -1.f),
    C_Vertex3f(0.232872963f, 0.767724037f, -1.f),
    C_Vertex3f(0.218658447f, 0.757616043f, -1.f),
    C_Vertex3f(0.209255219f, 0.747508049f, -1.f),
    C_Vertex3f(0.203721285f, 0.740198851f, -1.f),
    C_Vertex3f(0.198993444f, 0.732319117f, -1.f),
    C_Vertex3f(0.195071697f, 0.723868847f, -1.f),
    C_Vertex3f(0.191956043f, 0.714848042f, -1.f),
    C_Vertex3f(0.189646482f, 0.705256701f, -1.f),
    C_Vertex3f(0.188143015f, 0.695094824f, -1.f),
    C_Vertex3f(0.1875f, 0.678710938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.479980469f, 0.779785156f, -1.f),
    C_Vertex3f(0.423339844f, 0.779785156f, -1.f),
    C_Vertex3f(0.435407162f, 0.68951416f, -1.f),
    C_Vertex3f(0.448918343f, 0.698852539f, -1.f),
    C_Vertex3f(0.457843304f, 0.708190918f, -1.f),
    C_Vertex3f(0.463090539f, 0.715194702f, -1.f),
    C_Vertex3f(0.467635036f, 0.722976685f, -1.f),
    C_Vertex3f(0.471476793f, 0.731536865f, -1.f),
    C_Vertex3f(0.474615812f, 0.740875244f, -1.f),
    C_Vertex3f(0.477052093f, 0.750991821f, -1.f),
    C_Vertex3f(0.478785634f, 0.761886597f, -1.f),
    C_Vertex3f(0.479816437f, 0.77355957f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.423339844f, 0.779785156f, -1.f),
    C_Vertex3f(0.42257309f, 0.774093628f, -1.f),
    C_Vertex3f(0.435407162f, 0.68951416f, -1.f),
    C_Vertex3f(0.420492649f, 0.764598846f, -1.f),
    C_Vertex3f(0.420284748f, 0.683288574f, -1.f),
    C_Vertex3f(0.417865276f, 0.75699234f, -1.f),
    C_Vertex3f(0.414690971f, 0.751274109f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.414690971f, 0.751274109f, -1.f),
    C_Vertex3f(0.410969734f, 0.747444153f, -1.f),
    C_Vertex3f(0.420284748f, 0.683288574f, -1.f),
    C_Vertex3f(0.40642786f, 0.744571686f, -1.f),
    C_Vertex3f(0.403551102f, 0.680175781f, -1.f),
    C_Vertex3f(0.40106535f, 0.742656708f, -1.f),
    C_Vertex3f(0.394882202f, 0.741699219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.368290901f, 0.746537209f, -1.f),
    C_Vertex3f(0.371547699f, 0.683221817f, -1.f),
    C_Vertex3f(0.378577232f, 0.743311882f, -1.f),
    C_Vertex3f(0.387500763f, 0.680175781f, -1.f),
    C_Vertex3f(0.387550354f, 0.741699219f, -1.f),
    C_Vertex3f(0.403551102f, 0.680175781f, -1.f),
    C_Vertex3f(0.394882202f, 0.741699219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.301751137f, 0.714338303f, -1.f),
    C_Vertex3f(0.314445496f, 0.708461761f, -1.f),
    C_Vertex3f(0.317567825f, 0.769120216f, -1.f),
    C_Vertex3f(0.333049774f, 0.698451996f, -1.f),
    C_Vertex3f(0.338523865f, 0.760408401f, -1.f),
    C_Vertex3f(0.353397369f, 0.689313889f, -1.f),
    C_Vertex3f(0.35669136f, 0.751375198f, -1.f),
    C_Vertex3f(0.371547699f, 0.683221817f, -1.f),
    C_Vertex3f(0.368290901f, 0.746537209f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.275156021f, 0.720214844f, -1.f),
    C_Vertex3f(0.282270432f, 0.720214844f, -1.f),
    C_Vertex3f(0.283054352f, 0.777832031f, -1.f),
    C_Vertex3f(0.291026115f, 0.718255997f, -1.f),
    C_Vertex3f(0.299077988f, 0.774928093f, -1.f),
    C_Vertex3f(0.301751137f, 0.714338303f, -1.f),
    C_Vertex3f(0.317567825f, 0.769120216f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.257097244f, 0.712352753f, -1.f),
    C_Vertex3f(0.262405396f, 0.716283798f, -1.f),
    C_Vertex3f(0.266546249f, 0.777832031f, -1.f),
    C_Vertex3f(0.268424988f, 0.718904495f, -1.f),
    C_Vertex3f(0.283054352f, 0.777832031f, -1.f),
    C_Vertex3f(0.275156021f, 0.720214844f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.246537209f, 0.692853928f, -1.f),
    C_Vertex3f(0.248835564f, 0.700609207f, -1.f),
    C_Vertex3f(0.248835564f, 0.7744627f, -1.f),
    C_Vertex3f(0.252391815f, 0.707109451f, -1.f),
    C_Vertex3f(0.266546249f, 0.777832031f, -1.f),
    C_Vertex3f(0.254798889f, 0.709732056f, -1.f),
    C_Vertex3f(0.257097244f, 0.712352753f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.248835564f, 0.7744627f, -1.f),
    C_Vertex3f(0.232872963f, 0.767724037f, -1.f),
    C_Vertex3f(0.24549675f, 0.683843613f, -1.f),
    C_Vertex3f(0.246537209f, 0.692853928f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.1875f, 0.678710938f, -1.f),
    C_Vertex3f(0.188143015f, 0.695094824f, -1.f),
    C_Vertex3f(0.187445641f, 0.684362411f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_210[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_Normal3f(-0.999229491f, 0.0392487273f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Normal3f(-0.994037926f, 0.109034933f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Normal3f(-0.981642663f, 0.190729454f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Normal3f(-0.960366726f, 0.278739631f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Normal3f(-0.928485572f, 0.371368498f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Normal3f(-0.884783864f, 0.46600157f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Normal3f(-0.828982711f, 0.559274256f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Normal3f(-0.762047708f, 0.647521019f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Normal3f(-0.689660072f, 0.724133313f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Normal3f(-0.610743105f, 0.791828811f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Normal3f(-0.522311151f, 0.85275501f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Normal3f(-0.426912457f, 0.904293001f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Normal3f(-0.327885866f, 0.944717348f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Normal3f(-0.228800729f, 0.97347337f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Normal3f(-0.132870927f, 0.991133392f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Normal3f(-0.0466583073f, 0.998910844f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Normal3f(0.0266486872f, 0.999644876f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Normal3f(0.0925619975f, 0.995706856f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Normal3f(0.157029793f, 0.98759383f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Normal3f(0.222855464f, 0.974851489f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Normal3f(0.289159536f, 0.957280874f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Normal3f(0.354979903f, 0.934873939f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Normal3f(0.419335544f, 0.907831371f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Normal3f(0.481297553f, 0.876557291f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Normal3f(0.540897012f, 0.841088831f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Normal3f(0.599702179f, 0.800223231f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Normal3f(0.656618118f, 0.754223228f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Normal3f(0.710595846f, 0.703600347f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Normal3f(0.760694981f, 0.649109423f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Normal3f(0.806162775f, 0.591693759f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Normal3f(0.846489131f, 0.532405972f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Normal3f(0.881427884f, 0.47231859f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Normal3f(0.910308123f, 0.41393134f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Normal3f(0.934312224f, 0.356455833f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Normal3f(0.954386652f, 0.298573464f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Normal3f(0.970532179f, 0.240971416f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Normal3f(0.982873797f, 0.184280097f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Normal3f(0.991638541f, 0.129045784f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Normal3f(0.997129619f, 0.0757144392f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Normal3f(0.999698818f, 0.0245408304f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_Normal3f(0.999670804f, -0.0256569553f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Normal3f(0.996906638f, -0.0785944462f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Normal3f(0.990996599f, -0.133886874f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Normal3f(0.981580496f, -0.191049173f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Normal3f(0.968356967f, -0.249568939f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Normal3f(0.951117396f, -0.308829606f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Normal3f(0.929773211f, -0.36813283f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Normal3f(0.90437752f, -0.426733196f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Normal3f(0.874280751f, -0.485420704f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Normal3f(0.838072717f, -0.545558572f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Normal3f(0.795999944f, -0.605296671f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Normal3f(0.74819994f, -0.663473368f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Normal3f(0.695117056f, -0.718896568f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Normal3f(0.637500465f, -0.770449877f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Normal3f(0.576356828f, -0.817198098f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Normal3f(0.512864292f, -0.858469784f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Normal3f(0.45348832f, -0.891262233f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Normal3f(0.396994501f, -0.91782105f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Normal3f(0.33799234f, -0.941148937f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Normal3f(0.277123004f, -0.960834503f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Normal3f(0.215111941f, -0.976589382f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Normal3f(0.152726561f, -0.988268495f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Normal3f(0.0907294229f, -0.995875537f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Normal3f(0.0292172581f, -0.999573112f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0311744045f, -0.999514043f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0938582048f, -0.995585501f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Normal3f(-0.159526795f, -0.987193584f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Normal3f(-0.226795033f, -0.973942518f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Normal3f(-0.294730932f, -0.955580294f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Normal3f(-0.362296075f, -0.932063103f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Normal3f(-0.428416908f, -0.903581202f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Normal3f(-0.492066324f, -0.870557725f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Normal3f(-0.551469922f, -0.834194779f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Normal3f(-0.60842973f, -0.793607712f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Normal3f(-0.663708389f, -0.747991323f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Normal3f(-0.716302752f, -0.697789609f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Normal3f(-0.765292704f, -0.64368248f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Normal3f(-0.809918106f, -0.586542904f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Normal3f(-0.84963721f, -0.527367532f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Normal3f(-0.884154499f, -0.467194706f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Normal3f(-0.911144316f, -0.412087172f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Normal3f(-0.932779849f, -0.360446692f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Normal3f(-0.951756537f, -0.306854159f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Normal3f(-0.967785478f, -0.251776159f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Normal3f(-0.980656683f, -0.195735648f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Normal3f(-0.990251899f, -0.139287636f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Normal3f(-0.996550143f, -0.0829921961f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Normal3f(0.999472976f, 0.0324606337f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Normal3f(0.994043112f, 0.108987279f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Normal3f(0.98171258f, 0.190368786f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Normal3f(0.960640907f, 0.277793169f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Normal3f(0.929206252f, 0.369561613f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Normal3f(0.886297405f, 0.463116556f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Normal3f(0.831705272f, 0.555217385f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Normal3f(0.766394138f, 0.642370582f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Normal3f(0.696807623f, 0.717258096f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Normal3f(0.621382415f, 0.783507407f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Normal3f(0.535601974f, 0.84447062f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Normal3f(0.441416115f, 0.897302508f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Normal3f(0.341795832f, 0.939774215f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Normal3f(0.240293324f, 0.970700264f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Normal3f(0.140439287f, 0.990089297f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Normal3f(0.0449865349f, 0.998987615f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0450702235f, 0.9989838f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Normal3f(-0.139715239f, 0.990191758f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Normal3f(-0.239666194f, 0.970855415f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Normal3f(-0.341789991f, 0.939776421f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Normal3f(-0.442493498f, 0.896771669f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Normal3f(-0.538056731f, 0.842908621f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Normal3f(-0.625280082f, 0.780400395f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Normal3f(-0.701993346f, 0.712183535f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Normal3f(-0.773547888f, 0.633737862f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Normal3f(-0.840062797f, 0.542489231f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Normal3f(-0.894041598f, 0.44798404f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Normal3f(-0.935230792f, 0.354038715f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Normal3f(-0.964579821f, 0.263791323f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Normal3f(-0.983775556f, 0.179403409f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Normal3f(-0.994777799f, 0.102064326f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Normal3f(-0.999368489f, 0.0355346389f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Normal3f(-0.999824762f, -0.0187195782f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Normal3f(-0.997788966f, -0.066463232f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Normal3f(-0.993469477f, -0.114098631f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Normal3f(-0.986415029f, -0.164272755f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Normal3f(-0.976221919f, -0.216773391f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Normal3f(-0.962501347f, -0.2712771f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Normal3f(-0.944906473f, -0.327340305f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Normal3f(-0.923166513f, -0.384400129f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Normal3f(-0.896669626f, -0.442700326f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Normal3f(-0.864540637f, -0.50256294f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Normal3f(-0.826793492f, -0.562505424f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Normal3f(-0.783447027f, -0.62145859f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Normal3f(-0.734794199f, -0.678290129f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Normal3f(-0.681415319f, -0.731896937f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Normal3f(-0.62415117f, -0.781303585f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Normal3f(-0.564038873f, -0.825748205f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Normal3f(-0.503067136f, -0.864247262f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Normal3f(-0.439818412f, -0.898086727f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Normal3f(-0.373481601f, -0.927637637f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Normal3f(-0.305047989f, -0.952336967f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Normal3f(-0.235612959f, -0.971846938f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Normal3f(-0.166289985f, -0.986076891f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Normal3f(-0.098126255f, -0.995173991f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Normal3f(-0.0289524645f, -0.999580801f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Normal3f(0.0471588001f, -0.99888742f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Normal3f(0.134547606f, -0.990907192f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Normal3f(0.228727072f, -0.973490655f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Normal3f(0.323609829f, -0.946190596f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Normal3f(0.416355193f, -0.909202039f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Normal3f(0.504215479f, -0.863577902f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Normal3f(0.584928811f, -0.811084628f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Normal3f(0.656986952f, -0.753901958f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Normal3f(0.740095019f, -0.672502279f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Normal3f(0.826100826f, -0.56352222f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Normal3f(0.891164541f, -0.453680247f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Normal3f(0.937078595f, -0.349118531f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Normal3f(0.967324972f, -0.253539503f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Normal3f(0.985708833f, -0.16845797f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Normal3f(0.995581388f, -0.0939031243f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.413574219f, 0.680175781f, 0.f),
    C_Vertex3f(0.413574219f, 0.680175781f, -1.f),
    C_Vertex3f(0.350097656f, 0.680175781f, -1.f),
    C_Vertex3f(0.350097656f, 0.680175781f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.350097656f, 0.680175781f, 0.f),
    C_Vertex3f(0.350097656f, 0.680175781f, -1.f),
    C_Vertex3f(0.249023438f, 0.802734375f, -1.f),
    C_Vertex3f(0.249023438f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.249023438f, 0.802734375f, 0.f),
    C_Vertex3f(0.249023438f, 0.802734375f, -1.f),
    C_Vertex3f(0.354492188f, 0.802734375f, -1.f),
    C_Vertex3f(0.354492188f, 0.802734375f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.354492188f, 0.802734375f, 0.f),
    C_Vertex3f(0.354492188f, 0.802734375f, -1.f),
    C_Vertex3f(0.413574219f, 0.680175781f, -1.f),
    C_Vertex3f(0.413574219f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.354492188f, 0.802734375f, 0.f),
    C_Vertex3f(0.249023438f, 0.802734375f, 0.f),
    C_Vertex3f(0.350097656f, 0.680175781f, 0.f),
    C_Vertex3f(0.413574219f, 0.680175781f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.354492188f, 0.802734375f, -1.f),
    C_Vertex3f(0.249023438f, 0.802734375f, -1.f),
    C_Vertex3f(0.350097656f, 0.680175781f, -1.f),
    C_Vertex3f(0.413574219f, 0.680175781f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Translatef(0.696289063f, 0.f, 0.f),
};
static const Call char_211[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_Normal3f(-0.999229491f, 0.0392487273f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Normal3f(-0.994037926f, 0.109034933f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Normal3f(-0.981642663f, 0.190729454f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Normal3f(-0.960366726f, 0.278739631f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Normal3f(-0.928485572f, 0.371368498f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Normal3f(-0.884783864f, 0.46600157f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Normal3f(-0.828982711f, 0.559274256f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Normal3f(-0.762047708f, 0.647521019f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Normal3f(-0.689660072f, 0.724133313f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Normal3f(-0.610743105f, 0.791828811f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Normal3f(-0.522311151f, 0.85275501f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Normal3f(-0.426912457f, 0.904293001f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Normal3f(-0.327885866f, 0.944717348f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Normal3f(-0.228800729f, 0.97347337f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Normal3f(-0.132870927f, 0.991133392f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Normal3f(-0.0466583073f, 0.998910844f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Normal3f(0.0266486872f, 0.999644876f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Normal3f(0.0925619975f, 0.995706856f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Normal3f(0.157029793f, 0.98759383f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Normal3f(0.222855464f, 0.974851489f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Normal3f(0.289159536f, 0.957280874f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Normal3f(0.354979903f, 0.934873939f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Normal3f(0.419335544f, 0.907831371f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Normal3f(0.481297553f, 0.876557291f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Normal3f(0.540897012f, 0.841088831f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Normal3f(0.599702179f, 0.800223231f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Normal3f(0.656618118f, 0.754223228f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Normal3f(0.710595846f, 0.703600347f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Normal3f(0.760694981f, 0.649109423f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Normal3f(0.806162775f, 0.591693759f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Normal3f(0.846489131f, 0.532405972f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Normal3f(0.881427884f, 0.47231859f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Normal3f(0.910308123f, 0.41393134f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Normal3f(0.934312224f, 0.356455833f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Normal3f(0.954386652f, 0.298573464f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Normal3f(0.970532179f, 0.240971416f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Normal3f(0.982873797f, 0.184280097f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Normal3f(0.991638541f, 0.129045784f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Normal3f(0.997129619f, 0.0757144392f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Normal3f(0.999698818f, 0.0245408304f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_Normal3f(0.999670804f, -0.0256569553f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Normal3f(0.996906638f, -0.0785944462f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Normal3f(0.990996599f, -0.133886874f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Normal3f(0.981580496f, -0.191049173f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Normal3f(0.968356967f, -0.249568939f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Normal3f(0.951117396f, -0.308829606f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Normal3f(0.929773211f, -0.36813283f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Normal3f(0.90437752f, -0.426733196f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Normal3f(0.874280751f, -0.485420704f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Normal3f(0.838072717f, -0.545558572f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Normal3f(0.795999944f, -0.605296671f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Normal3f(0.74819994f, -0.663473368f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Normal3f(0.695117056f, -0.718896568f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Normal3f(0.637500465f, -0.770449877f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Normal3f(0.576356828f, -0.817198098f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Normal3f(0.512864292f, -0.858469784f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Normal3f(0.45348832f, -0.891262233f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Normal3f(0.396994501f, -0.91782105f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Normal3f(0.33799234f, -0.941148937f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Normal3f(0.277123004f, -0.960834503f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Normal3f(0.215111941f, -0.976589382f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Normal3f(0.152726561f, -0.988268495f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Normal3f(0.0907294229f, -0.995875537f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Normal3f(0.0292172581f, -0.999573112f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0311744045f, -0.999514043f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0938582048f, -0.995585501f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Normal3f(-0.159526795f, -0.987193584f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Normal3f(-0.226795033f, -0.973942518f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Normal3f(-0.294730932f, -0.955580294f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Normal3f(-0.362296075f, -0.932063103f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Normal3f(-0.428416908f, -0.903581202f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Normal3f(-0.492066324f, -0.870557725f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Normal3f(-0.551469922f, -0.834194779f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Normal3f(-0.60842973f, -0.793607712f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Normal3f(-0.663708389f, -0.747991323f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Normal3f(-0.716302752f, -0.697789609f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Normal3f(-0.765292704f, -0.64368248f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Normal3f(-0.809918106f, -0.586542904f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Normal3f(-0.84963721f, -0.527367532f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Normal3f(-0.884154499f, -0.467194706f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Normal3f(-0.911144316f, -0.412087172f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Normal3f(-0.932779849f, -0.360446692f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Normal3f(-0.951756537f, -0.306854159f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Normal3f(-0.967785478f, -0.251776159f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Normal3f(-0.980656683f, -0.195735648f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Normal3f(-0.990251899f, -0.139287636f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Normal3f(-0.996550143f, -0.0829921961f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Normal3f(0.999472976f, 0.0324606337f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Normal3f(0.994043112f, 0.108987279f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Normal3f(0.98171258f, 0.190368786f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Normal3f(0.960640907f, 0.277793169f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Normal3f(0.929206252f, 0.369561613f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Normal3f(0.886297405f, 0.463116556f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Normal3f(0.831705272f, 0.555217385f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Normal3f(0.766394138f, 0.642370582f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Normal3f(0.696807623f, 0.717258096f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Normal3f(0.621382415f, 0.783507407f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Normal3f(0.535601974f, 0.84447062f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Normal3f(0.441416115f, 0.897302508f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Normal3f(0.341795832f, 0.939774215f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Normal3f(0.240293324f, 0.970700264f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Normal3f(0.140439287f, 0.990089297f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Normal3f(0.0449865349f, 0.998987615f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0450702235f, 0.9989838f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Normal3f(-0.139715239f, 0.990191758f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Normal3f(-0.239666194f, 0.970855415f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Normal3f(-0.341789991f, 0.939776421f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Normal3f(-0.442493498f, 0.896771669f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Normal3f(-0.538056731f, 0.842908621f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Normal3f(-0.625280082f, 0.780400395f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Normal3f(-0.701993346f, 0.712183535f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Normal3f(-0.773547888f, 0.633737862f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Normal3f(-0.840062797f, 0.542489231f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Normal3f(-0.894041598f, 0.44798404f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Normal3f(-0.935230792f, 0.354038715f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Normal3f(-0.964579821f, 0.263791323f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Normal3f(-0.983775556f, 0.179403409f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Normal3f(-0.994777799f, 0.102064326f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Normal3f(-0.999368489f, 0.0355346389f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Normal3f(-0.999824762f, -0.0187195782f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Normal3f(-0.997788966f, -0.066463232f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Normal3f(-0.993469477f, -0.114098631f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Normal3f(-0.986415029f, -0.164272755f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Normal3f(-0.976221919f, -0.216773391f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Normal3f(-0.962501347f, -0.2712771f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Normal3f(-0.944906473f, -0.327340305f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Normal3f(-0.923166513f, -0.384400129f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Normal3f(-0.896669626f, -0.442700326f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Normal3f(-0.864540637f, -0.50256294f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Normal3f(-0.826793492f, -0.562505424f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Normal3f(-0.783447027f, -0.62145859f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Normal3f(-0.734794199f, -0.678290129f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Normal3f(-0.681415319f, -0.731896937f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Normal3f(-0.62415117f, -0.781303585f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Normal3f(-0.564038873f, -0.825748205f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Normal3f(-0.503067136f, -0.864247262f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Normal3f(-0.439818412f, -0.898086727f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Normal3f(-0.373481601f, -0.927637637f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Normal3f(-0.305047989f, -0.952336967f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Normal3f(-0.235612959f, -0.971846938f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Normal3f(-0.166289985f, -0.986076891f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Normal3f(-0.098126255f, -0.995173991f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Normal3f(-0.0289524645f, -0.999580801f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Normal3f(0.0471588001f, -0.99888742f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Normal3f(0.134547606f, -0.990907192f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Normal3f(0.228727072f, -0.973490655f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Normal3f(0.323609829f, -0.946190596f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Normal3f(0.416355193f, -0.909202039f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Normal3f(0.504215479f, -0.863577902f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Normal3f(0.584928811f, -0.811084628f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Normal3f(0.656986952f, -0.753901958f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Normal3f(0.740095019f, -0.672502279f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Normal3f(0.826100826f, -0.56352222f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Normal3f(0.891164541f, -0.453680247f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Normal3f(0.937078595f, -0.349118531f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Normal3f(0.967324972f, -0.253539503f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Normal3f(0.985708833f, -0.16845797f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Normal3f(0.995581388f, -0.0939031243f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.28515625f, 0.680175781f, 0.f),
    C_Vertex3f(0.28515625f, 0.680175781f, -1.f),
    C_Vertex3f(0.342773438f, 0.802734375f, -1.f),
    C_Vertex3f(0.342773438f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.342773438f, 0.802734375f, 0.f),
    C_Vertex3f(0.342773438f, 0.802734375f, -1.f),
    C_Vertex3f(0.446289063f, 0.802734375f, -1.f),
    C_Vertex3f(0.446289063f, 0.802734375f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.446289063f, 0.802734375f, 0.f),
    C_Vertex3f(0.446289063f, 0.802734375f, -1.f),
    C_Vertex3f(0.350097656f, 0.680175781f, -1.f),
    C_Vertex3f(0.350097656f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.350097656f, 0.680175781f, 0.f),
    C_Vertex3f(0.350097656f, 0.680175781f, -1.f),
    C_Vertex3f(0.28515625f, 0.680175781f, -1.f),
    C_Vertex3f(0.28515625f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342773438f, 0.802734375f, 0.f),
    C_Vertex3f(0.28515625f, 0.680175781f, 0.f),
    C_Vertex3f(0.350097656f, 0.680175781f, 0.f),
    C_Vertex3f(0.446289063f, 0.802734375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342773438f, 0.802734375f, -1.f),
    C_Vertex3f(0.28515625f, 0.680175781f, -1.f),
    C_Vertex3f(0.350097656f, 0.680175781f, -1.f),
    C_Vertex3f(0.446289063f, 0.802734375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Translatef(0.696289063f, 0.f, 0.f),
};
static const Call char_212[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_Normal3f(-0.999229491f, 0.0392487273f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Normal3f(-0.994037926f, 0.109034933f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Normal3f(-0.981642663f, 0.190729454f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Normal3f(-0.960366726f, 0.278739631f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Normal3f(-0.928485572f, 0.371368498f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Normal3f(-0.884783864f, 0.46600157f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Normal3f(-0.828982711f, 0.559274256f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Normal3f(-0.762047708f, 0.647521019f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Normal3f(-0.689660072f, 0.724133313f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Normal3f(-0.610743105f, 0.791828811f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Normal3f(-0.522311151f, 0.85275501f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Normal3f(-0.426912457f, 0.904293001f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Normal3f(-0.327885866f, 0.944717348f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Normal3f(-0.228800729f, 0.97347337f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Normal3f(-0.132870927f, 0.991133392f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Normal3f(-0.0466583073f, 0.998910844f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Normal3f(0.0266486872f, 0.999644876f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Normal3f(0.0925619975f, 0.995706856f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Normal3f(0.157029793f, 0.98759383f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Normal3f(0.222855464f, 0.974851489f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Normal3f(0.289159536f, 0.957280874f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Normal3f(0.354979903f, 0.934873939f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Normal3f(0.419335544f, 0.907831371f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Normal3f(0.481297553f, 0.876557291f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Normal3f(0.540897012f, 0.841088831f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Normal3f(0.599702179f, 0.800223231f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Normal3f(0.656618118f, 0.754223228f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Normal3f(0.710595846f, 0.703600347f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Normal3f(0.760694981f, 0.649109423f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Normal3f(0.806162775f, 0.591693759f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Normal3f(0.846489131f, 0.532405972f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Normal3f(0.881427884f, 0.47231859f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Normal3f(0.910308123f, 0.41393134f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Normal3f(0.934312224f, 0.356455833f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Normal3f(0.954386652f, 0.298573464f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Normal3f(0.970532179f, 0.240971416f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Normal3f(0.982873797f, 0.184280097f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Normal3f(0.991638541f, 0.129045784f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Normal3f(0.997129619f, 0.0757144392f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Normal3f(0.999698818f, 0.0245408304f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_Normal3f(0.999670804f, -0.0256569553f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Normal3f(0.996906638f, -0.0785944462f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Normal3f(0.990996599f, -0.133886874f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Normal3f(0.981580496f, -0.191049173f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Normal3f(0.968356967f, -0.249568939f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Normal3f(0.951117396f, -0.308829606f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Normal3f(0.929773211f, -0.36813283f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Normal3f(0.90437752f, -0.426733196f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Normal3f(0.874280751f, -0.485420704f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Normal3f(0.838072717f, -0.545558572f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Normal3f(0.795999944f, -0.605296671f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Normal3f(0.74819994f, -0.663473368f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Normal3f(0.695117056f, -0.718896568f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Normal3f(0.637500465f, -0.770449877f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Normal3f(0.576356828f, -0.817198098f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Normal3f(0.512864292f, -0.858469784f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Normal3f(0.45348832f, -0.891262233f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Normal3f(0.396994501f, -0.91782105f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Normal3f(0.33799234f, -0.941148937f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Normal3f(0.277123004f, -0.960834503f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Normal3f(0.215111941f, -0.976589382f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Normal3f(0.152726561f, -0.988268495f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Normal3f(0.0907294229f, -0.995875537f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Normal3f(0.0292172581f, -0.999573112f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0311744045f, -0.999514043f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0938582048f, -0.995585501f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Normal3f(-0.159526795f, -0.987193584f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Normal3f(-0.226795033f, -0.973942518f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Normal3f(-0.294730932f, -0.955580294f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Normal3f(-0.362296075f, -0.932063103f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Normal3f(-0.428416908f, -0.903581202f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Normal3f(-0.492066324f, -0.870557725f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Normal3f(-0.551469922f, -0.834194779f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Normal3f(-0.60842973f, -0.793607712f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Normal3f(-0.663708389f, -0.747991323f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Normal3f(-0.716302752f, -0.697789609f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Normal3f(-0.765292704f, -0.64368248f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Normal3f(-0.809918106f, -0.586542904f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Normal3f(-0.84963721f, -0.527367532f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Normal3f(-0.884154499f, -0.467194706f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Normal3f(-0.911144316f, -0.412087172f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Normal3f(-0.932779849f, -0.360446692f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Normal3f(-0.951756537f, -0.306854159f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Normal3f(-0.967785478f, -0.251776159f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Normal3f(-0.980656683f, -0.195735648f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Normal3f(-0.990251899f, -0.139287636f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Normal3f(-0.996550143f, -0.0829921961f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Normal3f(0.999472976f, 0.0324606337f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Normal3f(0.994043112f, 0.108987279f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Normal3f(0.98171258f, 0.190368786f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Normal3f(0.960640907f, 0.277793169f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Normal3f(0.929206252f, 0.369561613f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Normal3f(0.886297405f, 0.463116556f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Normal3f(0.831705272f, 0.555217385f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Normal3f(0.766394138f, 0.642370582f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Normal3f(0.696807623f, 0.717258096f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Normal3f(0.621382415f, 0.783507407f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Normal3f(0.535601974f, 0.84447062f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Normal3f(0.441416115f, 0.897302508f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Normal3f(0.341795832f, 0.939774215f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Normal3f(0.240293324f, 0.970700264f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Normal3f(0.140439287f, 0.990089297f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Normal3f(0.0449865349f, 0.998987615f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0450702235f, 0.9989838f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Normal3f(-0.139715239f, 0.990191758f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Normal3f(-0.239666194f, 0.970855415f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Normal3f(-0.341789991f, 0.939776421f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Normal3f(-0.442493498f, 0.896771669f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Normal3f(-0.538056731f, 0.842908621f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Normal3f(-0.625280082f, 0.780400395f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Normal3f(-0.701993346f, 0.712183535f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Normal3f(-0.773547888f, 0.633737862f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Normal3f(-0.840062797f, 0.542489231f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Normal3f(-0.894041598f, 0.44798404f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Normal3f(-0.935230792f, 0.354038715f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Normal3f(-0.964579821f, 0.263791323f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Normal3f(-0.983775556f, 0.179403409f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Normal3f(-0.994777799f, 0.102064326f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Normal3f(-0.999368489f, 0.0355346389f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Normal3f(-0.999824762f, -0.0187195782f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Normal3f(-0.997788966f, -0.066463232f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Normal3f(-0.993469477f, -0.114098631f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Normal3f(-0.986415029f, -0.164272755f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Normal3f(-0.976221919f, -0.216773391f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Normal3f(-0.962501347f, -0.2712771f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Normal3f(-0.944906473f, -0.327340305f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Normal3f(-0.923166513f, -0.384400129f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Normal3f(-0.896669626f, -0.442700326f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Normal3f(-0.864540637f, -0.50256294f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Normal3f(-0.826793492f, -0.562505424f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Normal3f(-0.783447027f, -0.62145859f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Normal3f(-0.734794199f, -0.678290129f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Normal3f(-0.681415319f, -0.731896937f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Normal3f(-0.62415117f, -0.781303585f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Normal3f(-0.564038873f, -0.825748205f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Normal3f(-0.503067136f, -0.864247262f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Normal3f(-0.439818412f, -0.898086727f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Normal3f(-0.373481601f, -0.927637637f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Normal3f(-0.305047989f, -0.952336967f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Normal3f(-0.235612959f, -0.971846938f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Normal3f(-0.166289985f, -0.986076891f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Normal3f(-0.098126255f, -0.995173991f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Normal3f(-0.0289524645f, -0.999580801f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Normal3f(0.0471588001f, -0.99888742f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Normal3f(0.134547606f, -0.990907192f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Normal3f(0.228727072f, -0.973490655f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Normal3f(0.323609829f, -0.946190596f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Normal3f(0.416355193f, -0.909202039f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Normal3f(0.504215479f, -0.863577902f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Normal3f(0.584928811f, -0.811084628f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Normal3f(0.656986952f, -0.753901958f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Normal3f(0.740095019f, -0.672502279f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Normal3f(0.826100826f, -0.56352222f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Normal3f(0.891164541f, -0.453680247f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Normal3f(0.937078595f, -0.349118531f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Normal3f(0.967324972f, -0.253539503f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Normal3f(0.985708833f, -0.16845797f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Normal3f(0.995581388f, -0.0939031243f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.348144531f, 0.754882813f, 0.f),
    C_Vertex3f(0.348144531f, 0.754882813f, -1.f),
    C_Vertex3f(0.299072266f, 0.680175781f, -1.f),
    C_Vertex3f(0.299072266f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.299072266f, 0.680175781f, 0.f),
    C_Vertex3f(0.299072266f, 0.680175781f, -1.f),
    C_Vertex3f(0.209716797f, 0.680175781f, -1.f),
    C_Vertex3f(0.209716797f, 0.680175781f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(0.209716797f, 0.680175781f, 0.f),
    C_Vertex3f(0.209716797f, 0.680175781f, -1.f),
    C_Vertex3f(0.303222656f, 0.802734375f, -1.f),
    C_Vertex3f(0.303222656f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.303222656f, 0.802734375f, 0.f),
    C_Vertex3f(0.303222656f, 0.802734375f, -1.f),
    C_Vertex3f(0.387207031f, 0.802734375f, -1.f),
    C_Vertex3f(0.387207031f, 0.802734375f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.387207031f, 0.802734375f, 0.f),
    C_Vertex3f(0.387207031f, 0.802734375f, -1.f),
    C_Vertex3f(0.485595703f, 0.680175781f, -1.f),
    C_Vertex3f(0.485595703f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.485595703f, 0.680175781f, 0.f),
    C_Vertex3f(0.485595703f, 0.680175781f, -1.f),
    C_Vertex3f(0.396240234f, 0.680175781f, -1.f),
    C_Vertex3f(0.396240234f, 0.680175781f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.396240234f, 0.680175781f, 0.f),
    C_Vertex3f(0.396240234f, 0.680175781f, -1.f),
    C_Vertex3f(0.348144531f, 0.754882813f, -1.f),
    C_Vertex3f(0.348144531f, 0.754882813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.209716797f, 0.680175781f, 0.f),
    C_Vertex3f(0.299072266f, 0.680175781f, 0.f),
    C_Vertex3f(0.303222656f, 0.802734375f, 0.f),
    C_Vertex3f(0.348144531f, 0.754882813f, 0.f),
    C_Vertex3f(0.387207031f, 0.802734375f, 0.f),
    C_Vertex3f(0.396240234f, 0.680175781f, 0.f),
    C_Vertex3f(0.485595703f, 0.680175781f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.209716797f, 0.680175781f, -1.f),
    C_Vertex3f(0.299072266f, 0.680175781f, -1.f),
    C_Vertex3f(0.303222656f, 0.802734375f, -1.f),
    C_Vertex3f(0.348144531f, 0.754882813f, -1.f),
    C_Vertex3f(0.387207031f, 0.802734375f, -1.f),
    C_Vertex3f(0.396240234f, 0.680175781f, -1.f),
    C_Vertex3f(0.485595703f, 0.680175781f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Translatef(0.696289063f, 0.f, 0.f),
};
static const Call char_213[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_Normal3f(-0.999229491f, 0.0392487273f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Normal3f(-0.994037926f, 0.109034933f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Normal3f(-0.981642663f, 0.190729454f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Normal3f(-0.960366726f, 0.278739631f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Normal3f(-0.928485572f, 0.371368498f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Normal3f(-0.884783864f, 0.46600157f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Normal3f(-0.828982711f, 0.559274256f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Normal3f(-0.762047708f, 0.647521019f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Normal3f(-0.689660072f, 0.724133313f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Normal3f(-0.610743105f, 0.791828811f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Normal3f(-0.522311151f, 0.85275501f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Normal3f(-0.426912457f, 0.904293001f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Normal3f(-0.327885866f, 0.944717348f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Normal3f(-0.228800729f, 0.97347337f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Normal3f(-0.132870927f, 0.991133392f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Normal3f(-0.0466583073f, 0.998910844f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Normal3f(0.0266486872f, 0.999644876f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Normal3f(0.0925619975f, 0.995706856f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Normal3f(0.157029793f, 0.98759383f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Normal3f(0.222855464f, 0.974851489f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Normal3f(0.289159536f, 0.957280874f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Normal3f(0.354979903f, 0.934873939f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Normal3f(0.419335544f, 0.907831371f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Normal3f(0.481297553f, 0.876557291f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Normal3f(0.540897012f, 0.841088831f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Normal3f(0.599702179f, 0.800223231f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Normal3f(0.656618118f, 0.754223228f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Normal3f(0.710595846f, 0.703600347f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Normal3f(0.760694981f, 0.649109423f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Normal3f(0.806162775f, 0.591693759f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Normal3f(0.846489131f, 0.532405972f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Normal3f(0.881427884f, 0.47231859f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Normal3f(0.910308123f, 0.41393134f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Normal3f(0.934312224f, 0.356455833f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Normal3f(0.954386652f, 0.298573464f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Normal3f(0.970532179f, 0.240971416f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Normal3f(0.982873797f, 0.184280097f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Normal3f(0.991638541f, 0.129045784f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Normal3f(0.997129619f, 0.0757144392f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Normal3f(0.999698818f, 0.0245408304f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_Normal3f(0.999670804f, -0.0256569553f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Normal3f(0.996906638f, -0.0785944462f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Normal3f(0.990996599f, -0.133886874f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Normal3f(0.981580496f, -0.191049173f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Normal3f(0.968356967f, -0.249568939f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Normal3f(0.951117396f, -0.308829606f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Normal3f(0.929773211f, -0.36813283f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Normal3f(0.90437752f, -0.426733196f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Normal3f(0.874280751f, -0.485420704f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Normal3f(0.838072717f, -0.545558572f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Normal3f(0.795999944f, -0.605296671f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Normal3f(0.74819994f, -0.663473368f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Normal3f(0.695117056f, -0.718896568f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Normal3f(0.637500465f, -0.770449877f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Normal3f(0.576356828f, -0.817198098f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Normal3f(0.512864292f, -0.858469784f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Normal3f(0.45348832f, -0.891262233f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Normal3f(0.396994501f, -0.91782105f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Normal3f(0.33799234f, -0.941148937f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Normal3f(0.277123004f, -0.960834503f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Normal3f(0.215111941f, -0.976589382f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Normal3f(0.152726561f, -0.988268495f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Normal3f(0.0907294229f, -0.995875537f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Normal3f(0.0292172581f, -0.999573112f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0311744045f, -0.999514043f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0938582048f, -0.995585501f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Normal3f(-0.159526795f, -0.987193584f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Normal3f(-0.226795033f, -0.973942518f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Normal3f(-0.294730932f, -0.955580294f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Normal3f(-0.362296075f, -0.932063103f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Normal3f(-0.428416908f, -0.903581202f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Normal3f(-0.492066324f, -0.870557725f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Normal3f(-0.551469922f, -0.834194779f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Normal3f(-0.60842973f, -0.793607712f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Normal3f(-0.663708389f, -0.747991323f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Normal3f(-0.716302752f, -0.697789609f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Normal3f(-0.765292704f, -0.64368248f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Normal3f(-0.809918106f, -0.586542904f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Normal3f(-0.84963721f, -0.527367532f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Normal3f(-0.884154499f, -0.467194706f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Normal3f(-0.911144316f, -0.412087172f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Normal3f(-0.932779849f, -0.360446692f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Normal3f(-0.951756537f, -0.306854159f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Normal3f(-0.967785478f, -0.251776159f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Normal3f(-0.980656683f, -0.195735648f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Normal3f(-0.990251899f, -0.139287636f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Normal3f(-0.996550143f, -0.0829921961f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Normal3f(0.999472976f, 0.0324606337f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Normal3f(0.994043112f, 0.108987279f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Normal3f(0.98171258f, 0.190368786f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Normal3f(0.960640907f, 0.277793169f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Normal3f(0.929206252f, 0.369561613f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Normal3f(0.886297405f, 0.463116556f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Normal3f(0.831705272f, 0.555217385f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Normal3f(0.766394138f, 0.642370582f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Normal3f(0.696807623f, 0.717258096f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Normal3f(0.621382415f, 0.783507407f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Normal3f(0.535601974f, 0.84447062f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Normal3f(0.441416115f, 0.897302508f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Normal3f(0.341795832f, 0.939774215f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Normal3f(0.240293324f, 0.970700264f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Normal3f(0.140439287f, 0.990089297f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Normal3f(0.0449865349f, 0.998987615f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0450702235f, 0.9989838f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Normal3f(-0.139715239f, 0.990191758f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Normal3f(-0.239666194f, 0.970855415f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Normal3f(-0.341789991f, 0.939776421f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Normal3f(-0.442493498f, 0.896771669f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Normal3f(-0.538056731f, 0.842908621f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Normal3f(-0.625280082f, 0.780400395f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Normal3f(-0.701993346f, 0.712183535f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Normal3f(-0.773547888f, 0.633737862f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Normal3f(-0.840062797f, 0.542489231f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Normal3f(-0.894041598f, 0.44798404f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Normal3f(-0.935230792f, 0.354038715f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Normal3f(-0.964579821f, 0.263791323f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Normal3f(-0.983775556f, 0.179403409f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Normal3f(-0.994777799f, 0.102064326f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Normal3f(-0.999368489f, 0.0355346389f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Normal3f(-0.999824762f, -0.0187195782f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Normal3f(-0.997788966f, -0.066463232f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Normal3f(-0.993469477f, -0.114098631f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Normal3f(-0.986415029f, -0.164272755f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Normal3f(-0.976221919f, -0.216773391f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Normal3f(-0.962501347f, -0.2712771f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Normal3f(-0.944906473f, -0.327340305f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Normal3f(-0.923166513f, -0.384400129f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Normal3f(-0.896669626f, -0.442700326f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Normal3f(-0.864540637f, -0.50256294f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Normal3f(-0.826793492f, -0.562505424f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Normal3f(-0.783447027f, -0.62145859f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Normal3f(-0.734794199f, -0.678290129f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Normal3f(-0.681415319f, -0.731896937f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Normal3f(-0.62415117f, -0.781303585f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Normal3f(-0.564038873f, -0.825748205f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Normal3f(-0.503067136f, -0.864247262f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Normal3f(-0.439818412f, -0.898086727f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Normal3f(-0.373481601f, -0.927637637f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Normal3f(-0.305047989f, -0.952336967f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Normal3f(-0.235612959f, -0.971846938f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Normal3f(-0.166289985f, -0.986076891f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Normal3f(-0.098126255f, -0.995173991f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Normal3f(-0.0289524645f, -0.999580801f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Normal3f(0.0471588001f, -0.99888742f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Normal3f(0.134547606f, -0.990907192f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Normal3f(0.228727072f, -0.973490655f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Normal3f(0.323609829f, -0.946190596f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Normal3f(0.416355193f, -0.909202039f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Normal3f(0.504215479f, -0.863577902f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Normal3f(0.584928811f, -0.811084628f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Normal3f(0.656986952f, -0.753901958f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Normal3f(0.740095019f, -0.672502279f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Normal3f(0.826100826f, -0.56352222f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Normal3f(0.891164541f, -0.453680247f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Normal3f(0.937078595f, -0.349118531f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Normal3f(0.967324972f, -0.253539503f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Normal3f(0.985708833f, -0.16845797f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Normal3f(0.995581388f, -0.0939031243f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999953806f, -0.00961818546f, 0.f),
    C_Vertex3f(0.203125f, 0.678710938f, 0.f),
    C_Vertex3f(0.203125f, 0.678710938f, -1.f),
    C_Normal3f(-0.999230683f, 0.0392165966f, 0.f),
    C_Vertex3f(0.203070641f, 0.684362411f, 0.f),
    C_Vertex3f(0.203070641f, 0.684362411f, -1.f),
    C_Normal3f(-0.994498312f, 0.104752719f, 0.f),
    C_Vertex3f(0.203768015f, 0.695094824f, 0.f),
    C_Vertex3f(0.203768015f, 0.695094824f, -1.f),
    C_Normal3f(-0.981874108f, 0.189534381f, 0.f),
    C_Vertex3f(0.205271482f, 0.705256701f, 0.f),
    C_Vertex3f(0.205271482f, 0.705256701f, -1.f),
    C_Normal3f(-0.960046053f, 0.279841781f, 0.f),
    C_Vertex3f(0.207581043f, 0.714848042f, 0.f),
    C_Vertex3f(0.207581043f, 0.714848042f, -1.f),
    C_Normal3f(-0.927577436f, 0.373630971f, 0.f),
    C_Vertex3f(0.210696697f, 0.723868847f, 0.f),
    C_Vertex3f(0.210696697f, 0.723868847f, -1.f),
    C_Normal3f(-0.883691847f, 0.468069077f, 0.f),
    C_Vertex3f(0.214618444f, 0.732319117f, 0.f),
    C_Vertex3f(0.214618444f, 0.732319117f, -1.f),
    C_Normal3f(-0.828614712f, 0.559819281f, 0.f),
    C_Vertex3f(0.219346285f, 0.740198851f, 0.f),
    C_Vertex3f(0.219346285f, 0.740198851f, -1.f),
    C_Normal3f(-0.759081304f, 0.650995851f, 0.f),
    C_Vertex3f(0.224880219f, 0.747508049f, 0.f),
    C_Vertex3f(0.224880219f, 0.747508049f, -1.f),
    C_Normal3f(-0.650275886f, 0.759698212f, 0.f),
    C_Vertex3f(0.234283447f, 0.757616043f, 0.f),
    C_Vertex3f(0.234283447f, 0.757616043f, -1.f),
    C_Normal3f(-0.487445921f, 0.87315321f, 0.f),
    C_Vertex3f(0.248497963f, 0.767724037f, 0.f),
    C_Vertex3f(0.248497963f, 0.767724037f, -1.f),
    C_Normal3f(-0.287504673f, 0.957779288f, 0.f),
    C_Vertex3f(0.264460564f, 0.7744627f, 0.f),
    C_Vertex3f(0.264460564f, 0.7744627f, -1.f),
    C_Normal3f(-0.0979905054f, 0.995187461f, 0.f),
    C_Vertex3f(0.282171249f, 0.777832031f, 0.f),
    C_Vertex3f(0.282171249f, 0.777832031f, -1.f),
    C_Normal3f(0.0889112428f, 0.99603951f, 0.f),
    C_Vertex3f(0.298679352f, 0.777832031f, 0.f),
    C_Vertex3f(0.298679352f, 0.777832031f, -1.f),
    C_Normal3f(0.244741365f, 0.969588399f, 0.f),
    C_Vertex3f(0.314702988f, 0.774928093f, 0.f),
    C_Vertex3f(0.314702988f, 0.774928093f, -1.f),
    C_Normal3f(0.345433027f, 0.938443422f, 0.f),
    C_Vertex3f(0.333192825f, 0.769120216f, 0.f),
    C_Vertex3f(0.333192825f, 0.769120216f, -1.f),
    C_Normal3f(0.413061887f, 0.910703003f, 0.f),
    C_Vertex3f(0.354148865f, 0.760408401f, 0.f),
    C_Vertex3f(0.354148865f, 0.760408401f, -1.f),
    C_Normal3f(0.42238301f, 0.906417429f, 0.f),
    C_Vertex3f(0.37231636f, 0.751375198f, 0.f),
    C_Vertex3f(0.37231636f, 0.751375198f, -1.f),
    C_Normal3f(0.345709145f, 0.938341737f, 0.f),
    C_Vertex3f(0.383915901f, 0.746537209f, 0.f),
    C_Vertex3f(0.383915901f, 0.746537209f, -1.f),
    C_Normal3f(0.243631572f, 0.969867826f, 0.f),
    C_Vertex3f(0.394202232f, 0.743311882f, 0.f),
    C_Vertex3f(0.394202232f, 0.743311882f, -1.f),
    C_Normal3f(0.0984259844f, 0.995144367f, 0.f),
    C_Vertex3f(0.403175354f, 0.741699219f, 0.f),
    C_Vertex3f(0.403175354f, 0.741699219f, -1.f),
    C_Normal3f(-0.070669286f, 0.997499824f, 0.f),
    C_Vertex3f(0.410507202f, 0.741699219f, 0.f),
    C_Vertex3f(0.410507202f, 0.741699219f, -1.f),
    C_Normal3f(-0.24143213f, 0.970417738f, 0.f),
    C_Vertex3f(0.41669035f, 0.742656708f, 0.f),
    C_Vertex3f(0.41669035f, 0.742656708f, -1.f),
    C_Normal3f(-0.435192645f, 0.900337279f, 0.f),
    C_Vertex3f(0.42205286f, 0.744571686f, 0.f),
    C_Vertex3f(0.42205286f, 0.744571686f, -1.f),
    C_Normal3f(-0.62994957f, 0.776636064f, 0.f),
    C_Vertex3f(0.426594734f, 0.747444153f, 0.f),
    C_Vertex3f(0.426594734f, 0.747444153f, -1.f),
    C_Normal3f(-0.810694158f, 0.585469842f, 0.f),
    C_Vertex3f(0.430315971f, 0.751274109f, 0.f),
    C_Vertex3f(0.430315971f, 0.751274109f, -1.f),
    C_Normal3f(-0.9168607f, 0.399207145f, 0.f),
    C_Vertex3f(0.433490276f, 0.75699234f, 0.f),
    C_Vertex3f(0.433490276f, 0.75699234f, -1.f),
    C_Normal3f(-0.964133918f, 0.265416414f, 0.f),
    C_Vertex3f(0.436117649f, 0.764598846f, 0.f),
    C_Vertex3f(0.436117649f, 0.764598846f, -1.f),
    C_Normal3f(-0.982875049f, 0.18427363f, 0.f),
    C_Vertex3f(0.43819809f, 0.774093628f, 0.f),
    C_Vertex3f(0.43819809f, 0.774093628f, -1.f),
    C_Normal3f(-0.991047204f, 0.133512393f, 0.f),
    C_Vertex3f(0.438964844f, 0.779785156f, 0.f),
    C_Vertex3f(0.438964844f, 0.779785156f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.438964844f, 0.779785156f, 0.f),
    C_Vertex3f(0.438964844f, 0.779785156f, -1.f),
    C_Vertex3f(0.495605469f, 0.779785156f, -1.f),
    C_Vertex3f(0.495605469f, 0.779785156f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999653041f, -0.0263388976f, 0.f),
    C_Vertex3f(0.495605469f, 0.779785156f, 0.f),
    C_Vertex3f(0.495605469f, 0.779785156f, -1.f),
    C_Normal3f(0.99777931f, -0.066607669f, 0.f),
    C_Vertex3f(0.495441437f, 0.77355957f, 0.f),
    C_Vertex3f(0.495441437f, 0.77355957f, -1.f),
    C_Normal3f(0.992581308f, -0.121582195f, 0.f),
    C_Vertex3f(0.494410634f, 0.761886597f, 0.f),
    C_Vertex3f(0.494410634f, 0.761886597f, -1.f),
    C_Normal3f(0.980870843f, -0.194659367f, 0.f),
    C_Vertex3f(0.492677093f, 0.750991821f, 0.f),
    C_Vertex3f(0.492677093f, 0.750991821f, -1.f),
    C_Normal3f(0.961305082f, -0.275485754f, 0.f),
    C_Vertex3f(0.490240812f, 0.740875244f, 0.f),
    C_Vertex3f(0.490240812f, 0.740875244f, -1.f),
    C_Normal3f(0.931648552f, -0.363360554f, 0.f),
    C_Vertex3f(0.487101793f, 0.731536865f, 0.f),
    C_Vertex3f(0.487101793f, 0.731536865f, -1.f),
    C_Normal3f(0.889692426f, -0.456560522f, 0.f),
    C_Vertex3f(0.483260036f, 0.722976685f, 0.f),
    C_Vertex3f(0.483260036f, 0.722976685f, -1.f),
    C_Normal3f(0.833749712f, -0.552142739f, 0.f),
    C_Vertex3f(0.478715539f, 0.715194702f, 0.f),
    C_Vertex3f(0.478715539f, 0.715194702f, -1.f),
    C_Normal3f(0.755483389f, -0.655167818f, 0.f),
    C_Vertex3f(0.473468304f, 0.708190918f, 0.f),
    C_Vertex3f(0.473468304f, 0.708190918f, -1.f),
    C_Normal3f(0.639779389f, -0.768558562f, 0.f),
    C_Vertex3f(0.464543343f, 0.698852539f, 0.f),
    C_Vertex3f(0.464543343f, 0.698852539f, -1.f),
    C_Normal3f(0.477566123f, -0.878595829f, 0.f),
    C_Vertex3f(0.451032162f, 0.68951416f, 0.f),
    C_Vertex3f(0.451032162f, 0.68951416f, -1.f),
    C_Normal3f(0.281305313f, -0.95961827f, 0.f),
    C_Vertex3f(0.435909748f, 0.683288574f, 0.f),
    C_Vertex3f(0.435909748f, 0.683288574f, -1.f),
    C_Normal3f(0.0945234597f, -0.995522618f, 0.f),
    C_Vertex3f(0.419176102f, 0.680175781f, 0.f),
    C_Vertex3f(0.419176102f, 0.680175781f, -1.f),
    C_Normal3f(-0.0947503f, -0.995501101f, 0.f),
    C_Vertex3f(0.403125763f, 0.680175781f, 0.f),
    C_Vertex3f(0.403125763f, 0.680175781f, -1.f),
    C_Normal3f(-0.258822501f, -0.965924919f, 0.f),
    C_Vertex3f(0.387172699f, 0.683221817f, 0.f),
    C_Vertex3f(0.387172699f, 0.683221817f, -1.f),
    C_Normal3f(-0.367869198f, -0.92987752f, 0.f),
    C_Vertex3f(0.369022369f, 0.689313889f, 0.f),
    C_Vertex3f(0.369022369f, 0.689313889f, -1.f),
    C_Normal3f(-0.441156626f, -0.897430122f, 0.f),
    C_Vertex3f(0.348674774f, 0.698451996f, 0.f),
    C_Vertex3f(0.348674774f, 0.698451996f, -1.f),
    C_Normal3f(-0.452606946f, -0.891710103f, 0.f),
    C_Vertex3f(0.330070496f, 0.708461761f, 0.f),
    C_Vertex3f(0.330070496f, 0.708461761f, -1.f),
    C_Normal3f(-0.385828793f, -0.922570407f, 0.f),
    C_Vertex3f(0.317376137f, 0.714338303f, 0.f),
    C_Vertex3f(0.317376137f, 0.714338303f, -1.f),
    C_Normal3f(-0.288805217f, -0.957387865f, 0.f),
    C_Vertex3f(0.306651115f, 0.718255997f, 0.f),
    C_Vertex3f(0.306651115f, 0.718255997f, -1.f),
    C_Normal3f(-0.122500464f, -0.992468476f, 0.f),
    C_Vertex3f(0.297895432f, 0.720214844f, 0.f),
    C_Vertex3f(0.297895432f, 0.720214844f, -1.f),
    C_Normal3f(0.0942201167f, -0.995551467f, 0.f),
    C_Vertex3f(0.290781021f, 0.720214844f, 0.f),
    C_Vertex3f(0.290781021f, 0.720214844f, -1.f),
    C_Normal3f(0.294618219f, -0.955615103f, 0.f),
    C_Vertex3f(0.284049988f, 0.718904495f, 0.f),
    C_Vertex3f(0.284049988f, 0.718904495f, -1.f),
    C_Normal3f(0.500668526f, -0.865639091f, 0.f),
    C_Vertex3f(0.278030396f, 0.716283798f, 0.f),
    C_Vertex3f(0.278030396f, 0.716283798f, -1.f),
    C_Normal3f(0.652620137f, -0.757685304f, 0.f),
    C_Vertex3f(0.272722244f, 0.712352753f, 0.f),
    C_Vertex3f(0.272722244f, 0.712352753f, -1.f),
    C_Normal3f(0.744249165f, -0.667902052f, 0.f),
    C_Vertex3f(0.270423889f, 0.709732056f, 0.f),
    C_Vertex3f(0.270423889f, 0.709732056f, -1.f),
    C_Normal3f(0.837037444f, -0.547145605f, 0.f),
    C_Vertex3f(0.268016815f, 0.707109451f, 0.f),
    C_Vertex3f(0.268016815f, 0.707109451f, -1.f),
    C_Normal3f(0.925027549f, -0.379899919f, 0.f),
    C_Vertex3f(0.264460564f, 0.700609207f, 0.f),
    C_Vertex3f(0.264460564f, 0.700609207f, -1.f),
    C_Normal3f(0.980741322f, -0.195311442f, 0.f),
    C_Vertex3f(0.262162209f, 0.692853928f, 0.f),
    C_Vertex3f(0.262162209f, 0.692853928f, -1.f),
    C_Normal3f(0.997837007f, -0.065737471f, 0.f),
    C_Vertex3f(0.26112175f, 0.683843613f, 0.f),
    C_Vertex3f(0.26112175f, 0.683843613f, -1.f),
    C_Normal3f(0.999775767f, 0.0211769659f, 0.f),
    C_Vertex3f(0.261230469f, 0.678710938f, 0.f),
    C_Vertex3f(0.261230469f, 0.678710938f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.261230469f, 0.678710938f, 0.f),
    C_Vertex3f(0.261230469f, 0.678710938f, -1.f),
    C_Vertex3f(0.203125f, 0.678710938f, -1.f),
    C_Vertex3f(0.203125f, 0.678710938f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.261230469f, 0.678710938f, 0.f),
    C_Vertex3f(0.26112175f, 0.683843613f, 0.f),
    C_Vertex3f(0.248497963f, 0.767724037f, 0.f),
    C_Vertex3f(0.234283447f, 0.757616043f, 0.f),
    C_Vertex3f(0.224880219f, 0.747508049f, 0.f),
    C_Vertex3f(0.219346285f, 0.740198851f, 0.f),
    C_Vertex3f(0.214618444f, 0.732319117f, 0.f),
    C_Vertex3f(0.210696697f, 0.723868847f, 0.f),
    C_Vertex3f(0.207581043f, 0.714848042f, 0.f),
    C_Vertex3f(0.205271482f, 0.705256701f, 0.f),
    C_Vertex3f(0.203768015f, 0.695094824f, 0.f),
    C_Vertex3f(0.203125f, 0.678710938f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.495605469f, 0.779785156f, 0.f),
    C_Vertex3f(0.438964844f, 0.779785156f, 0.f),
    C_Vertex3f(0.451032162f, 0.68951416f, 0.f),
    C_Vertex3f(0.464543343f, 0.698852539f, 0.f),
    C_Vertex3f(0.473468304f, 0.708190918f, 0.f),
    C_Vertex3f(0.478715539f, 0.715194702f, 0.f),
    C_Vertex3f(0.483260036f, 0.722976685f, 0.f),
    C_Vertex3f(0.487101793f, 0.731536865f, 0.f),
    C_Vertex3f(0.490240812f, 0.740875244f, 0.f),
    C_Vertex3f(0.492677093f, 0.750991821f, 0.f),
    C_Vertex3f(0.494410634f, 0.761886597f, 0.f),
    C_Vertex3f(0.495441437f, 0.77355957f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.438964844f, 0.779785156f, 0.f),
    C_Vertex3f(0.43819809f, 0.774093628f, 0.f),
    C_Vertex3f(0.451032162f, 0.68951416f, 0.f),
    C_Vertex3f(0.436117649f, 0.764598846f, 0.f),
    C_Vertex3f(0.435909748f, 0.683288574f, 0.f),
    C_Vertex3f(0.433490276f, 0.75699234f, 0.f),
    C_Vertex3f(0.430315971f, 0.751274109f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.430315971f, 0.751274109f, 0.f),
    C_Vertex3f(0.426594734f, 0.747444153f, 0.f),
    C_Vertex3f(0.435909748f, 0.683288574f, 0.f),
    C_Vertex3f(0.42205286f, 0.744571686f, 0.f),
    C_Vertex3f(0.419176102f, 0.680175781f, 0.f),
    C_Vertex3f(0.41669035f, 0.742656708f, 0.f),
    C_Vertex3f(0.410507202f, 0.741699219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.383915901f, 0.746537209f, 0.f),
    C_Vertex3f(0.387172699f, 0.683221817f, 0.f),
    C_Vertex3f(0.394202232f, 0.743311882f, 0.f),
    C_Vertex3f(0.403125763f, 0.680175781f, 0.f),
    C_Vertex3f(0.403175354f, 0.741699219f, 0.f),
    C_Vertex3f(0.419176102f, 0.680175781f, 0.f),
    C_Vertex3f(0.410507202f, 0.741699219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.317376137f, 0.714338303f, 0.f),
    C_Vertex3f(0.330070496f, 0.708461761f, 0.f),
    C_Vertex3f(0.333192825f, 0.769120216f, 0.f),
    C_Vertex3f(0.348674774f, 0.698451996f, 0.f),
    C_Vertex3f(0.354148865f, 0.760408401f, 0.f),
    C_Vertex3f(0.369022369f, 0.689313889f, 0.f),
    C_Vertex3f(0.37231636f, 0.751375198f, 0.f),
    C_Vertex3f(0.387172699f, 0.683221817f, 0.f),
    C_Vertex3f(0.383915901f, 0.746537209f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.290781021f, 0.720214844f, 0.f),
    C_Vertex3f(0.297895432f, 0.720214844f, 0.f),
    C_Vertex3f(0.298679352f, 0.777832031f, 0.f),
    C_Vertex3f(0.306651115f, 0.718255997f, 0.f),
    C_Vertex3f(0.314702988f, 0.774928093f, 0.f),
    C_Vertex3f(0.317376137f, 0.714338303f, 0.f),
    C_Vertex3f(0.333192825f, 0.769120216f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.272722244f, 0.712352753f, 0.f),
    C_Vertex3f(0.278030396f, 0.716283798f, 0.f),
    C_Vertex3f(0.282171249f, 0.777832031f, 0.f),
    C_Vertex3f(0.284049988f, 0.718904495f, 0.f),
    C_Vertex3f(0.298679352f, 0.777832031f, 0.f),
    C_Vertex3f(0.290781021f, 0.720214844f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.262162209f, 0.692853928f, 0.f),
    C_Vertex3f(0.264460564f, 0.700609207f, 0.f),
    C_Vertex3f(0.264460564f, 0.7744627f, 0.f),
    C_Vertex3f(0.268016815f, 0.707109451f, 0.f),
    C_Vertex3f(0.282171249f, 0.777832031f, 0.f),
    C_Vertex3f(0.270423889f, 0.709732056f, 0.f),
    C_Vertex3f(0.272722244f, 0.712352753f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.264460564f, 0.7744627f, 0.f),
    C_Vertex3f(0.248497963f, 0.767724037f, 0.f),
    C_Vertex3f(0.26112175f, 0.683843613f, 0.f),
    C_Vertex3f(0.262162209f, 0.692853928f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.203125f, 0.678710938f, 0.f),
    C_Vertex3f(0.203768015f, 0.695094824f, 0.f),
    C_Vertex3f(0.203070641f, 0.684362411f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.261230469f, 0.678710938f, -1.f),
    C_Vertex3f(0.26112175f, 0.683843613f, -1.f),
    C_Vertex3f(0.248497963f, 0.767724037f, -1.f),
    C_Vertex3f(0.234283447f, 0.757616043f, -1.f),
    C_Vertex3f(0.224880219f, 0.747508049f, -1.f),
    C_Vertex3f(0.219346285f, 0.740198851f, -1.f),
    C_Vertex3f(0.214618444f, 0.732319117f, -1.f),
    C_Vertex3f(0.210696697f, 0.723868847f, -1.f),
    C_Vertex3f(0.207581043f, 0.714848042f, -1.f),
    C_Vertex3f(0.205271482f, 0.705256701f, -1.f),
    C_Vertex3f(0.203768015f, 0.695094824f, -1.f),
    C_Vertex3f(0.203125f, 0.678710938f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.495605469f, 0.779785156f, -1.f),
    C_Vertex3f(0.438964844f, 0.779785156f, -1.f),
    C_Vertex3f(0.451032162f, 0.68951416f, -1.f),
    C_Vertex3f(0.464543343f, 0.698852539f, -1.f),
    C_Vertex3f(0.473468304f, 0.708190918f, -1.f),
    C_Vertex3f(0.478715539f, 0.715194702f, -1.f),
    C_Vertex3f(0.483260036f, 0.722976685f, -1.f),
    C_Vertex3f(0.487101793f, 0.731536865f, -1.f),
    C_Vertex3f(0.490240812f, 0.740875244f, -1.f),
    C_Vertex3f(0.492677093f, 0.750991821f, -1.f),
    C_Vertex3f(0.494410634f, 0.761886597f, -1.f),
    C_Vertex3f(0.495441437f, 0.77355957f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.438964844f, 0.779785156f, -1.f),
    C_Vertex3f(0.43819809f, 0.774093628f, -1.f),
    C_Vertex3f(0.451032162f, 0.68951416f, -1.f),
    C_Vertex3f(0.436117649f, 0.764598846f, -1.f),
    C_Vertex3f(0.435909748f, 0.683288574f, -1.f),
    C_Vertex3f(0.433490276f, 0.75699234f, -1.f),
    C_Vertex3f(0.430315971f, 0.751274109f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.430315971f, 0.751274109f, -1.f),
    C_Vertex3f(0.426594734f, 0.747444153f, -1.f),
    C_Vertex3f(0.435909748f, 0.683288574f, -1.f),
    C_Vertex3f(0.42205286f, 0.744571686f, -1.f),
    C_Vertex3f(0.419176102f, 0.680175781f, -1.f),
    C_Vertex3f(0.41669035f, 0.742656708f, -1.f),
    C_Vertex3f(0.410507202f, 0.741699219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.383915901f, 0.746537209f, -1.f),
    C_Vertex3f(0.387172699f, 0.683221817f, -1.f),
    C_Vertex3f(0.394202232f, 0.743311882f, -1.f),
    C_Vertex3f(0.403125763f, 0.680175781f, -1.f),
    C_Vertex3f(0.403175354f, 0.741699219f, -1.f),
    C_Vertex3f(0.419176102f, 0.680175781f, -1.f),
    C_Vertex3f(0.410507202f, 0.741699219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.317376137f, 0.714338303f, -1.f),
    C_Vertex3f(0.330070496f, 0.708461761f, -1.f),
    C_Vertex3f(0.333192825f, 0.769120216f, -1.f),
    C_Vertex3f(0.348674774f, 0.698451996f, -1.f),
    C_Vertex3f(0.354148865f, 0.760408401f, -1.f),
    C_Vertex3f(0.369022369f, 0.689313889f, -1.f),
    C_Vertex3f(0.37231636f, 0.751375198f, -1.f),
    C_Vertex3f(0.387172699f, 0.683221817f, -1.f),
    C_Vertex3f(0.383915901f, 0.746537209f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.290781021f, 0.720214844f, -1.f),
    C_Vertex3f(0.297895432f, 0.720214844f, -1.f),
    C_Vertex3f(0.298679352f, 0.777832031f, -1.f),
    C_Vertex3f(0.306651115f, 0.718255997f, -1.f),
    C_Vertex3f(0.314702988f, 0.774928093f, -1.f),
    C_Vertex3f(0.317376137f, 0.714338303f, -1.f),
    C_Vertex3f(0.333192825f, 0.769120216f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.272722244f, 0.712352753f, -1.f),
    C_Vertex3f(0.278030396f, 0.716283798f, -1.f),
    C_Vertex3f(0.282171249f, 0.777832031f, -1.f),
    C_Vertex3f(0.284049988f, 0.718904495f, -1.f),
    C_Vertex3f(0.298679352f, 0.777832031f, -1.f),
    C_Vertex3f(0.290781021f, 0.720214844f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.262162209f, 0.692853928f, -1.f),
    C_Vertex3f(0.264460564f, 0.700609207f, -1.f),
    C_Vertex3f(0.264460564f, 0.7744627f, -1.f),
    C_Vertex3f(0.268016815f, 0.707109451f, -1.f),
    C_Vertex3f(0.282171249f, 0.777832031f, -1.f),
    C_Vertex3f(0.270423889f, 0.709732056f, -1.f),
    C_Vertex3f(0.272722244f, 0.712352753f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.264460564f, 0.7744627f, -1.f),
    C_Vertex3f(0.248497963f, 0.767724037f, -1.f),
    C_Vertex3f(0.26112175f, 0.683843613f, -1.f),
    C_Vertex3f(0.262162209f, 0.692853928f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.203125f, 0.678710938f, -1.f),
    C_Vertex3f(0.203768015f, 0.695094824f, -1.f),
    C_Vertex3f(0.203070641f, 0.684362411f, -1.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Translatef(0.696289063f, 0.f, 0.f),
};
static const Call char_214[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_Normal3f(-0.999229491f, 0.0392487273f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Normal3f(-0.994037926f, 0.109034933f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Normal3f(-0.981642663f, 0.190729454f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Normal3f(-0.960366726f, 0.278739631f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Normal3f(-0.928485572f, 0.371368498f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Normal3f(-0.884783864f, 0.46600157f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Normal3f(-0.828982711f, 0.559274256f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Normal3f(-0.762047708f, 0.647521019f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Normal3f(-0.689660072f, 0.724133313f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Normal3f(-0.610743105f, 0.791828811f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Normal3f(-0.522311151f, 0.85275501f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Normal3f(-0.426912457f, 0.904293001f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Normal3f(-0.327885866f, 0.944717348f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Normal3f(-0.228800729f, 0.97347337f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Normal3f(-0.132870927f, 0.991133392f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Normal3f(-0.0466583073f, 0.998910844f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Normal3f(0.0266486872f, 0.999644876f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Normal3f(0.0925619975f, 0.995706856f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Normal3f(0.157029793f, 0.98759383f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Normal3f(0.222855464f, 0.974851489f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Normal3f(0.289159536f, 0.957280874f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Normal3f(0.354979903f, 0.934873939f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Normal3f(0.419335544f, 0.907831371f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Normal3f(0.481297553f, 0.876557291f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Normal3f(0.540897012f, 0.841088831f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Normal3f(0.599702179f, 0.800223231f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Normal3f(0.656618118f, 0.754223228f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Normal3f(0.710595846f, 0.703600347f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Normal3f(0.760694981f, 0.649109423f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Normal3f(0.806162775f, 0.591693759f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Normal3f(0.846489131f, 0.532405972f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Normal3f(0.881427884f, 0.47231859f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Normal3f(0.910308123f, 0.41393134f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Normal3f(0.934312224f, 0.356455833f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Normal3f(0.954386652f, 0.298573464f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Normal3f(0.970532179f, 0.240971416f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Normal3f(0.982873797f, 0.184280097f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Normal3f(0.991638541f, 0.129045784f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Normal3f(0.997129619f, 0.0757144392f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Normal3f(0.999698818f, 0.0245408304f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_Normal3f(0.999670804f, -0.0256569553f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Normal3f(0.996906638f, -0.0785944462f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Normal3f(0.990996599f, -0.133886874f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Normal3f(0.981580496f, -0.191049173f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Normal3f(0.968356967f, -0.249568939f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Normal3f(0.951117396f, -0.308829606f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Normal3f(0.929773211f, -0.36813283f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Normal3f(0.90437752f, -0.426733196f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Normal3f(0.874280751f, -0.485420704f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Normal3f(0.838072717f, -0.545558572f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Normal3f(0.795999944f, -0.605296671f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Normal3f(0.74819994f, -0.663473368f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Normal3f(0.695117056f, -0.718896568f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Normal3f(0.637500465f, -0.770449877f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Normal3f(0.576356828f, -0.817198098f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Normal3f(0.512864292f, -0.858469784f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Normal3f(0.45348832f, -0.891262233f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Normal3f(0.396994501f, -0.91782105f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Normal3f(0.33799234f, -0.941148937f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Normal3f(0.277123004f, -0.960834503f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Normal3f(0.215111941f, -0.976589382f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Normal3f(0.152726561f, -0.988268495f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Normal3f(0.0907294229f, -0.995875537f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Normal3f(0.0292172581f, -0.999573112f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0311744045f, -0.999514043f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0938582048f, -0.995585501f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Normal3f(-0.159526795f, -0.987193584f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Normal3f(-0.226795033f, -0.973942518f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Normal3f(-0.294730932f, -0.955580294f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Normal3f(-0.362296075f, -0.932063103f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Normal3f(-0.428416908f, -0.903581202f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Normal3f(-0.492066324f, -0.870557725f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Normal3f(-0.551469922f, -0.834194779f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Normal3f(-0.60842973f, -0.793607712f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Normal3f(-0.663708389f, -0.747991323f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Normal3f(-0.716302752f, -0.697789609f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Normal3f(-0.765292704f, -0.64368248f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Normal3f(-0.809918106f, -0.586542904f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Normal3f(-0.84963721f, -0.527367532f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Normal3f(-0.884154499f, -0.467194706f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Normal3f(-0.911144316f, -0.412087172f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Normal3f(-0.932779849f, -0.360446692f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Normal3f(-0.951756537f, -0.306854159f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Normal3f(-0.967785478f, -0.251776159f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Normal3f(-0.980656683f, -0.195735648f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Normal3f(-0.990251899f, -0.139287636f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Normal3f(-0.996550143f, -0.0829921961f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Normal3f(-0.99974966f, -0.0223741643f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Normal3f(0.999472976f, 0.0324606337f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Normal3f(0.994043112f, 0.108987279f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Normal3f(0.98171258f, 0.190368786f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Normal3f(0.960640907f, 0.277793169f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Normal3f(0.929206252f, 0.369561613f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Normal3f(0.886297405f, 0.463116556f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Normal3f(0.831705272f, 0.555217385f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Normal3f(0.766394138f, 0.642370582f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Normal3f(0.696807623f, 0.717258096f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Normal3f(0.621382415f, 0.783507407f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Normal3f(0.535601974f, 0.84447062f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Normal3f(0.441416115f, 0.897302508f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Normal3f(0.341795832f, 0.939774215f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Normal3f(0.240293324f, 0.970700264f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Normal3f(0.140439287f, 0.990089297f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Normal3f(0.0449865349f, 0.998987615f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0450702235f, 0.9989838f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Normal3f(-0.139715239f, 0.990191758f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Normal3f(-0.239666194f, 0.970855415f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Normal3f(-0.341789991f, 0.939776421f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Normal3f(-0.442493498f, 0.896771669f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Normal3f(-0.538056731f, 0.842908621f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Normal3f(-0.625280082f, 0.780400395f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Normal3f(-0.701993346f, 0.712183535f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Normal3f(-0.773547888f, 0.633737862f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Normal3f(-0.840062797f, 0.542489231f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Normal3f(-0.894041598f, 0.44798404f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Normal3f(-0.935230792f, 0.354038715f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Normal3f(-0.964579821f, 0.263791323f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Normal3f(-0.983775556f, 0.179403409f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Normal3f(-0.994777799f, 0.102064326f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Normal3f(-0.999368489f, 0.0355346389f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Normal3f(-0.999824762f, -0.0187195782f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Normal3f(-0.997788966f, -0.066463232f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Normal3f(-0.993469477f, -0.114098631f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Normal3f(-0.986415029f, -0.164272755f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Normal3f(-0.976221919f, -0.216773391f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Normal3f(-0.962501347f, -0.2712771f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Normal3f(-0.944906473f, -0.327340305f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Normal3f(-0.923166513f, -0.384400129f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Normal3f(-0.896669626f, -0.442700326f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Normal3f(-0.864540637f, -0.50256294f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Normal3f(-0.826793492f, -0.562505424f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Normal3f(-0.783447027f, -0.62145859f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Normal3f(-0.734794199f, -0.678290129f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Normal3f(-0.681415319f, -0.731896937f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Normal3f(-0.62415117f, -0.781303585f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Normal3f(-0.564038873f, -0.825748205f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Normal3f(-0.503067136f, -0.864247262f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Normal3f(-0.439818412f, -0.898086727f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Normal3f(-0.373481601f, -0.927637637f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Normal3f(-0.305047989f, -0.952336967f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Normal3f(-0.235612959f, -0.971846938f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Normal3f(-0.166289985f, -0.986076891f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Normal3f(-0.098126255f, -0.995173991f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Normal3f(-0.0289524645f, -0.999580801f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Normal3f(0.0471588001f, -0.99888742f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Normal3f(0.134547606f, -0.990907192f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Normal3f(0.228727072f, -0.973490655f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Normal3f(0.323609829f, -0.946190596f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Normal3f(0.416355193f, -0.909202039f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Normal3f(0.504215479f, -0.863577902f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Normal3f(0.584928811f, -0.811084628f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Normal3f(0.656986952f, -0.753901958f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Normal3f(0.740095019f, -0.672502279f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Normal3f(0.826100826f, -0.56352222f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Normal3f(0.891164541f, -0.453680247f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Normal3f(0.937078595f, -0.349118531f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Normal3f(0.967324972f, -0.253539503f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Normal3f(0.985708833f, -0.16845797f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Normal3f(0.995581388f, -0.0939031243f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Normal3f(0.999529362f, -0.0306785274f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.230957031f, 0.680175781f, 0.f),
    C_Vertex3f(0.230957031f, 0.680175781f, -1.f),
    C_Vertex3f(0.230957031f, 0.76953125f, -1.f),
    C_Vertex3f(0.230957031f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.230957031f, 0.76953125f, 0.f),
    C_Vertex3f(0.230957031f, 0.76953125f, -1.f),
    C_Vertex3f(0.312988281f, 0.76953125f, -1.f),
    C_Vertex3f(0.312988281f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.312988281f, 0.76953125f, 0.f),
    C_Vertex3f(0.312988281f, 0.76953125f, -1.f),
    C_Vertex3f(0.312988281f, 0.680175781f, -1.f),
    C_Vertex3f(0.312988281f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.312988281f, 0.680175781f, 0.f),
    C_Vertex3f(0.312988281f, 0.680175781f, -1.f),
    C_Vertex3f(0.230957031f, 0.680175781f, -1.f),
    C_Vertex3f(0.230957031f, 0.680175781f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.394042969f, 0.680175781f, 0.f),
    C_Vertex3f(0.394042969f, 0.680175781f, -1.f),
    C_Vertex3f(0.394042969f, 0.76953125f, -1.f),
    C_Vertex3f(0.394042969f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.394042969f, 0.76953125f, 0.f),
    C_Vertex3f(0.394042969f, 0.76953125f, -1.f),
    C_Vertex3f(0.476074219f, 0.76953125f, -1.f),
    C_Vertex3f(0.476074219f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.476074219f, 0.76953125f, 0.f),
    C_Vertex3f(0.476074219f, 0.76953125f, -1.f),
    C_Vertex3f(0.476074219f, 0.680175781f, -1.f),
    C_Vertex3f(0.476074219f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.476074219f, 0.680175781f, 0.f),
    C_Vertex3f(0.476074219f, 0.680175781f, -1.f),
    C_Vertex3f(0.394042969f, 0.680175781f, -1.f),
    C_Vertex3f(0.394042969f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.589253664f, 0.100642443f, 0.f),
    C_Vertex3f(0.592344165f, 0.535572052f, 0.f),
    C_Vertex3f(0.601136923f, 0.11752677f, 0.f),
    C_Vertex3f(0.603804231f, 0.518714905f, 0.f),
    C_Vertex3f(0.611995697f, 0.135578156f, 0.f),
    C_Vertex3f(0.614205837f, 0.500813484f, 0.f),
    C_Vertex3f(0.621829987f, 0.1547966f, 0.f),
    C_Vertex3f(0.623548985f, 0.48186779f, 0.f),
    C_Vertex3f(0.63043499f, 0.17465663f, 0.f),
    C_Vertex3f(0.631724238f, 0.462287426f, 0.f),
    C_Vertex3f(0.637810707f, 0.195158243f, 0.f),
    C_Vertex3f(0.638731599f, 0.442072392f, 0.f),
    C_Vertex3f(0.643957138f, 0.216301441f, 0.f),
    C_Vertex3f(0.644571066f, 0.421222687f, 0.f),
    C_Vertex3f(0.648874283f, 0.238086224f, 0.f),
    C_Vertex3f(0.64924264f, 0.399738312f, 0.f),
    C_Vertex3f(0.652562141f, 0.26051259f, 0.f),
    C_Vertex3f(0.65274632f, 0.377619267f, 0.f),
    C_Vertex3f(0.655020714f, 0.283580542f, 0.f),
    C_Vertex3f(0.655082107f, 0.354865551f, 0.f),
    C_Vertex3f(0.65625f, 0.307290077f, 0.f),
    C_Vertex3f(0.65625f, 0.331477165f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_Vertex3f(0.563888788f, 0.384796143f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.566368222f, 0.367115974f, 0.f),
    C_Vertex3f(0.57982564f, 0.551384926f, 0.f),
    C_Vertex3f(0.568021178f, 0.348814964f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_Vertex3f(0.551491618f, 0.434111595f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.556450486f, 0.418293953f, 0.f),
    C_Vertex3f(0.566248655f, 0.566153526f, 0.f),
    C_Vertex3f(0.560582876f, 0.401855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.531717777f, 0.477641463f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_Vertex3f(0.539094448f, 0.463884354f, 0.f),
    C_Vertex3f(0.551613212f, 0.579877853f, 0.f),
    C_Vertex3f(0.545706272f, 0.449308395f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_Vertex3f(0.514669895f, 0.502699137f, 0.f),
    C_Vertex3f(0.52357626f, 0.490579724f, 0.f),
    C_Vertex3f(0.535919309f, 0.592557907f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.458665371f, 0.55101347f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_Vertex3f(0.471395969f, 0.5429883f, 0.f),
    C_Vertex3f(0.482998848f, 0.624052763f, 0.f),
    C_Vertex3f(0.483361721f, 0.534144282f, 0.f),
    C_Vertex3f(0.501356125f, 0.614785194f, 0.f),
    C_Vertex3f(0.494562626f, 0.524481416f, 0.f),
    C_Vertex3f(0.504998684f, 0.513999701f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.403850794f, 0.573082685f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_Vertex3f(0.418138266f, 0.569070101f, 0.f),
    C_Vertex3f(0.424648285f, 0.643911839f, 0.f),
    C_Vertex3f(0.432036519f, 0.56405437f, 0.f),
    C_Vertex3f(0.444644928f, 0.638616085f, 0.f),
    C_Vertex3f(0.445545554f, 0.558035493f, 0.f),
    C_Vertex3f(0.464095116f, 0.631996393f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.358653069f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_Vertex3f(0.374108195f, 0.578098416f, 0.f),
    C_Vertex3f(0.383015633f, 0.65053153f, 0.f),
    C_Vertex3f(0.389174104f, 0.576092124f, 0.f),
    C_Vertex3f(0.404105186f, 0.647883654f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.296806574f, 0.573307991f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_Vertex3f(0.317747831f, 0.577170372f, 0.f),
    C_Vertex3f(0.333352089f, 0.651855469f, 0.f),
    C_Vertex3f(0.339474678f, 0.579101563f, 0.f),
    C_Vertex3f(0.361379623f, 0.651855469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.238696337f, 0.550133705f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_Vertex3f(0.257280827f, 0.559789658f, 0.f),
    C_Vertex3f(0.2702775f, 0.643399715f, 0.f),
    C_Vertex3f(0.276650906f, 0.56751442f, 0.f),
    C_Vertex3f(0.301036358f, 0.649036884f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_Vertex3f(0.203884125f, 0.525028229f, 0.f),
    C_Vertex3f(0.220897436f, 0.538546562f, 0.f),
    C_Vertex3f(0.241075516f, 0.634943962f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_Vertex3f(0.161286354f, 0.470921278f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.173457146f, 0.491543055f, 0.f),
    C_Vertex3f(0.187342167f, 0.609576702f, 0.f),
    C_Vertex3f(0.187656403f, 0.509578705f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_Vertex3f(0.136944771f, 0.39353919f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.143030167f, 0.421919346f, 0.f),
    C_Vertex3f(0.162810802f, 0.592665195f, 0.f),
    C_Vertex3f(0.151144028f, 0.447713375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.13669908f, 0.243221641f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.132805943f, 0.26908505f, 0.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, 0.f),
    C_Vertex3f(0.130859375f, 0.296483994f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.130859375f, 0.3290205f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_Vertex3f(0.13288784f, 0.362572908f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_Vertex3f(0.106358171f, 0.107939005f, 0.f),
    C_Vertex3f(0.118418694f, 0.550386429f, 0.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, 0.f),
    C_Vertex3f(0.139836311f, 0.572935104f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, 0.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, 0.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, 0.f),
    C_Vertex3f(0.099678278f, 0.525674343f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, 0.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, 0.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, 0.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, 0.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, 0.f),
    C_Vertex3f(0.568847656f, 0.329893112f, 0.f),
    C_Vertex3f(0.568847656f, 0.304672241f, 0.f),
    C_Vertex3f(0.566914797f, 0.275533795f, 0.f),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, 0.f),
    C_Vertex3f(0.5572505f, 0.22253871f, 0.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, 0.f),
    C_Vertex3f(0.549519062f, 0.19868207f, 0.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, 0.f),
    C_Vertex3f(0.539854765f, 0.176586032f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_Vertex3f(0.528257608f, 0.156250596f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, 0.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, 0.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, 0.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, 0.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, 0.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, 0.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, 0.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, 0.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, 0.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, 0.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, 0.f),
    C_Vertex3f(0.482743263f, 0.106149077f, 0.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, 0.f),
    C_Vertex3f(0.499264717f, 0.12086153f, 0.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, 0.f),
    C_Vertex3f(0.514727592f, 0.137675762f, 0.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, 0.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, 0.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, 0.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, 0.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, 0.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, 0.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, 0.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, 0.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.171737313f, 0.155123353f, 0.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, 0.f),
    C_Vertex3f(0.185363293f, 0.136937618f, 0.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, 0.f),
    C_Vertex3f(0.200935841f, 0.120287418f, 0.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, 0.f),
    C_Vertex3f(0.21749866f, 0.105718493f, 0.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, 0.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, 0.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, 0.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, 0.f),
    C_Vertex3f(0.142538786f, 0.218893766f, 0.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, 0.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, 0.f),
    C_Vertex3f(0.160057902f, 0.174844623f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.230957031f, 0.76953125f, 0.f),
    C_Vertex3f(0.230957031f, 0.680175781f, 0.f),
    C_Vertex3f(0.312988281f, 0.680175781f, 0.f),
    C_Vertex3f(0.312988281f, 0.76953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.394042969f, 0.76953125f, 0.f),
    C_Vertex3f(0.394042969f, 0.680175781f, 0.f),
    C_Vertex3f(0.476074219f, 0.680175781f, 0.f),
    C_Vertex3f(0.476074219f, 0.76953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0446044207f, 0.28070116f, 0.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, 0.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.589253664f, 0.100642443f, -1.f),
    C_Vertex3f(0.592344165f, 0.535572052f, -1.f),
    C_Vertex3f(0.601136923f, 0.11752677f, -1.f),
    C_Vertex3f(0.603804231f, 0.518714905f, -1.f),
    C_Vertex3f(0.611995697f, 0.135578156f, -1.f),
    C_Vertex3f(0.614205837f, 0.500813484f, -1.f),
    C_Vertex3f(0.621829987f, 0.1547966f, -1.f),
    C_Vertex3f(0.623548985f, 0.48186779f, -1.f),
    C_Vertex3f(0.63043499f, 0.17465663f, -1.f),
    C_Vertex3f(0.631724238f, 0.462287426f, -1.f),
    C_Vertex3f(0.637810707f, 0.195158243f, -1.f),
    C_Vertex3f(0.638731599f, 0.442072392f, -1.f),
    C_Vertex3f(0.643957138f, 0.216301441f, -1.f),
    C_Vertex3f(0.644571066f, 0.421222687f, -1.f),
    C_Vertex3f(0.648874283f, 0.238086224f, -1.f),
    C_Vertex3f(0.64924264f, 0.399738312f, -1.f),
    C_Vertex3f(0.652562141f, 0.26051259f, -1.f),
    C_Vertex3f(0.65274632f, 0.377619267f, -1.f),
    C_Vertex3f(0.655020714f, 0.283580542f, -1.f),
    C_Vertex3f(0.655082107f, 0.354865551f, -1.f),
    C_Vertex3f(0.65625f, 0.307290077f, -1.f),
    C_Vertex3f(0.65625f, 0.331477165f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_Vertex3f(0.563888788f, 0.384796143f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.566368222f, 0.367115974f, -1.f),
    C_Vertex3f(0.57982564f, 0.551384926f, -1.f),
    C_Vertex3f(0.568021178f, 0.348814964f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_Vertex3f(0.551491618f, 0.434111595f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.556450486f, 0.418293953f, -1.f),
    C_Vertex3f(0.566248655f, 0.566153526f, -1.f),
    C_Vertex3f(0.560582876f, 0.401855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.531717777f, 0.477641463f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_Vertex3f(0.539094448f, 0.463884354f, -1.f),
    C_Vertex3f(0.551613212f, 0.579877853f, -1.f),
    C_Vertex3f(0.545706272f, 0.449308395f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.519166946f, 0.604193687f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_Vertex3f(0.514669895f, 0.502699137f, -1.f),
    C_Vertex3f(0.52357626f, 0.490579724f, -1.f),
    C_Vertex3f(0.535919309f, 0.592557907f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.458665371f, 0.55101347f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_Vertex3f(0.471395969f, 0.5429883f, -1.f),
    C_Vertex3f(0.482998848f, 0.624052763f, -1.f),
    C_Vertex3f(0.483361721f, 0.534144282f, -1.f),
    C_Vertex3f(0.501356125f, 0.614785194f, -1.f),
    C_Vertex3f(0.494562626f, 0.524481416f, -1.f),
    C_Vertex3f(0.504998684f, 0.513999701f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.403850794f, 0.573082685f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_Vertex3f(0.418138266f, 0.569070101f, -1.f),
    C_Vertex3f(0.424648285f, 0.643911839f, -1.f),
    C_Vertex3f(0.432036519f, 0.56405437f, -1.f),
    C_Vertex3f(0.444644928f, 0.638616085f, -1.f),
    C_Vertex3f(0.445545554f, 0.558035493f, -1.f),
    C_Vertex3f(0.464095116f, 0.631996393f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.358653069f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_Vertex3f(0.374108195f, 0.578098416f, -1.f),
    C_Vertex3f(0.383015633f, 0.65053153f, -1.f),
    C_Vertex3f(0.389174104f, 0.576092124f, -1.f),
    C_Vertex3f(0.404105186f, 0.647883654f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.296806574f, 0.573307991f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_Vertex3f(0.317747831f, 0.577170372f, -1.f),
    C_Vertex3f(0.333352089f, 0.651855469f, -1.f),
    C_Vertex3f(0.339474678f, 0.579101563f, -1.f),
    C_Vertex3f(0.361379623f, 0.651855469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.238696337f, 0.550133705f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_Vertex3f(0.257280827f, 0.559789658f, -1.f),
    C_Vertex3f(0.2702775f, 0.643399715f, -1.f),
    C_Vertex3f(0.276650906f, 0.56751442f, -1.f),
    C_Vertex3f(0.301036358f, 0.649036884f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.213430405f, 0.623669624f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_Vertex3f(0.203884125f, 0.525028229f, -1.f),
    C_Vertex3f(0.220897436f, 0.538546562f, -1.f),
    C_Vertex3f(0.241075516f, 0.634943962f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_Vertex3f(0.161286354f, 0.470921278f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.173457146f, 0.491543055f, -1.f),
    C_Vertex3f(0.187342167f, 0.609576702f, -1.f),
    C_Vertex3f(0.187656403f, 0.509578705f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_Vertex3f(0.136944771f, 0.39353919f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.143030167f, 0.421919346f, -1.f),
    C_Vertex3f(0.162810802f, 0.592665195f, -1.f),
    C_Vertex3f(0.151144028f, 0.447713375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.13669908f, 0.243221641f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.132805943f, 0.26908505f, -1.f),
    C_Vertex3f(0.132269144f, 0.0771327019f, -1.f),
    C_Vertex3f(0.130859375f, 0.296483994f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.130859375f, 0.3290205f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_Vertex3f(0.13288784f, 0.362572908f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.0950211287f, 0.124836683f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_Vertex3f(0.106358171f, 0.107939005f, -1.f),
    C_Vertex3f(0.118418694f, 0.550386429f, -1.f),
    C_Vertex3f(0.118774176f, 0.0920376778f, -1.f),
    C_Vertex3f(0.139836311f, 0.572935104f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0675522089f, 0.180784702f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_Vertex3f(0.0755839348f, 0.161621094f, -1.f),
    C_Vertex3f(0.0836150646f, 0.498798847f, -1.f),
    C_Vertex3f(0.0847630501f, 0.142730713f, -1.f),
    C_Vertex3f(0.099678278f, 0.525674343f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0503413677f, 0.239914894f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_Vertex3f(0.0549309254f, 0.219931602f, -1.f),
    C_Vertex3f(0.0595202446f, 0.438557625f, -1.f),
    C_Vertex3f(0.0606678724f, 0.200221539f, -1.f),
    C_Vertex3f(0.0702290535f, 0.469759941f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0461342335f, 0.369662762f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0468991995f, 0.260171413f, -1.f),
    C_Vertex3f(0.0514886379f, 0.405191898f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.576345921f, 0.0849251747f, -1.f),
    C_Vertex3f(0.568847656f, 0.329893112f, -1.f),
    C_Vertex3f(0.568847656f, 0.304672241f, -1.f),
    C_Vertex3f(0.566914797f, 0.275533795f, -1.f),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.563049078f, 0.248155951f, -1.f),
    C_Vertex3f(0.5572505f, 0.22253871f, -1.f),
    C_Vertex3f(0.562413692f, 0.0703749657f, -1.f),
    C_Vertex3f(0.549519062f, 0.19868207f, -1.f),
    C_Vertex3f(0.54745698f, 0.0569918156f, -1.f),
    C_Vertex3f(0.539854765f, 0.176586032f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_Vertex3f(0.528257608f, 0.156250596f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.36028862f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_Vertex3f(0.380803823f, -0.00950694084f, -1.f),
    C_Vertex3f(0.384257317f, 0.0641134977f, -1.f),
    C_Vertex3f(0.400964022f, -0.00703644753f, -1.f),
    C_Vertex3f(0.406071663f, 0.0683170557f, -1.f),
    C_Vertex3f(0.420769215f, -0.00333070755f, -1.f),
    C_Vertex3f(0.426827431f, 0.0746223927f, -1.f),
    C_Vertex3f(0.440219402f, 0.00161027908f, -1.f),
    C_Vertex3f(0.44652462f, 0.0830295086f, -1.f),
    C_Vertex3f(0.459314585f, 0.00778651237f, -1.f),
    C_Vertex3f(0.465163231f, 0.0935384035f, -1.f),
    C_Vertex3f(0.478054762f, 0.0151979923f, -1.f),
    C_Vertex3f(0.482743263f, 0.106149077f, -1.f),
    C_Vertex3f(0.496439934f, 0.0238447189f, -1.f),
    C_Vertex3f(0.499264717f, 0.12086153f, -1.f),
    C_Vertex3f(0.5144701f, 0.0337266922f, -1.f),
    C_Vertex3f(0.514727592f, 0.137675762f, -1.f),
    C_Vertex3f(0.531475782f, 0.0447757244f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.273128748f, 0.0744993687f, -1.f),
    C_Vertex3f(0.274292469f, -0.00255250931f, -1.f),
    C_Vertex3f(0.293652654f, 0.0682555437f, -1.f),
    C_Vertex3f(0.295108557f, -0.0066473484f, -1.f),
    C_Vertex3f(0.315166831f, 0.0640929937f, -1.f),
    C_Vertex3f(0.316525698f, -0.00937724113f, -1.f),
    C_Vertex3f(0.33767128f, 0.0620117188f, -1.f),
    C_Vertex3f(0.338543892f, -0.0107421875f, -1.f),
    C_Vertex3f(0.361384392f, 0.0620117188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.171737313f, 0.155123353f, -1.f),
    C_Vertex3f(0.179227829f, 0.0383958817f, -1.f),
    C_Vertex3f(0.185363293f, 0.136937618f, -1.f),
    C_Vertex3f(0.197038651f, 0.0274763107f, -1.f),
    C_Vertex3f(0.200935841f, 0.120287418f, -1.f),
    C_Vertex3f(0.215450525f, 0.0179216862f, -1.f),
    C_Vertex3f(0.21749866f, 0.105718493f, -1.f),
    C_Vertex3f(0.234463453f, 0.00973200798f, -1.f),
    C_Vertex3f(0.235051751f, 0.0932308435f, -1.f),
    C_Vertex3f(0.254077435f, 0.00290727615f, -1.f),
    C_Vertex3f(0.253595114f, 0.0828244686f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.15032506f, 0.196101427f, -1.f),
    C_Vertex3f(0.142538786f, 0.218893766f, -1.f),
    C_Vertex3f(0.146843076f, 0.0632240772f, -1.f),
    C_Vertex3f(0.162495971f, 0.0503118038f, -1.f),
    C_Vertex3f(0.160057902f, 0.174844623f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.230957031f, 0.76953125f, -1.f),
    C_Vertex3f(0.230957031f, 0.680175781f, -1.f),
    C_Vertex3f(0.312988281f, 0.680175781f, -1.f),
    C_Vertex3f(0.312988281f, 0.76953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.394042969f, 0.76953125f, -1.f),
    C_Vertex3f(0.394042969f, 0.680175781f, -1.f),
    C_Vertex3f(0.476074219f, 0.680175781f, -1.f),
    C_Vertex3f(0.476074219f, 0.76953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0446044207f, 0.28070116f, -1.f),
    C_Vertex3f(0.0434570313f, 0.331970215f, -1.f),
    C_Vertex3f(0.0434570313f, 0.301504135f, -1.f),
    C_End(),
    C_Translatef(0.696289063f, 0.f, 0.f),
};
static const Call char_215[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.707106829f, 0.707106829f, 0.f),
    C_Vertex3f(0.0703582764f, 0.178741455f, 0.f),
    C_Vertex3f(0.0703582764f, 0.178741455f, -1.f),
    C_Vertex3f(0.208023071f, 0.31640625f, -1.f),
    C_Vertex3f(0.208023071f, 0.31640625f, 0.f),
    C_Normal3f(-0.707106769f, -0.707106769f, 0.f),
    C_Vertex3f(0.208023071f, 0.31640625f, 0.f),
    C_Vertex3f(0.208023071f, 0.31640625f, -1.f),
    C_Vertex3f(0.0708007813f, 0.45362854f, -1.f),
    C_Vertex3f(0.0708007813f, 0.45362854f, 0.f),
    C_Normal3f(-0.707106769f, 0.707106769f, 0.f),
    C_Vertex3f(0.0708007813f, 0.45362854f, 0.f),
    C_Vertex3f(0.0708007813f, 0.45362854f, -1.f),
    C_Vertex3f(0.12411499f, 0.506942749f, -1.f),
    C_Vertex3f(0.12411499f, 0.506942749f, 0.f),
    C_Normal3f(0.707106769f, 0.707106769f, 0.f),
    C_Vertex3f(0.12411499f, 0.506942749f, 0.f),
    C_Vertex3f(0.12411499f, 0.506942749f, -1.f),
    C_Vertex3f(0.26133728f, 0.369720459f, -1.f),
    C_Vertex3f(0.26133728f, 0.369720459f, 0.f),
    C_Normal3f(-0.707106709f, 0.707106709f, 0.f),
    C_Vertex3f(0.26133728f, 0.369720459f, 0.f),
    C_Vertex3f(0.26133728f, 0.369720459f, -1.f),
    C_Vertex3f(0.398124695f, 0.506507874f, -1.f),
    C_Vertex3f(0.398124695f, 0.506507874f, 0.f),
    C_Normal3f(0.710020363f, 0.704181194f, 0.f),
    C_Vertex3f(0.398124695f, 0.506507874f, 0.f),
    C_Vertex3f(0.398124695f, 0.506507874f, -1.f),
    C_Vertex3f(0.450569153f, 0.45362854f, -1.f),
    C_Vertex3f(0.450569153f, 0.45362854f, 0.f),
    C_Normal3f(0.708231688f, -0.705980062f, 0.f),
    C_Vertex3f(0.450569153f, 0.45362854f, 0.f),
    C_Vertex3f(0.450569153f, 0.45362854f, -1.f),
    C_Vertex3f(0.314216614f, 0.316841125f, -1.f),
    C_Vertex3f(0.314216614f, 0.316841125f, 0.f),
    C_Normal3f(0.707087159f, 0.707126439f, 0.f),
    C_Vertex3f(0.314216614f, 0.316841125f, 0.f),
    C_Vertex3f(0.314216614f, 0.316841125f, -1.f),
    C_Vertex3f(0.451446533f, 0.179618835f, -1.f),
    C_Vertex3f(0.451446533f, 0.179618835f, 0.f),
    C_Normal3f(0.707106769f, -0.707106769f, 0.f),
    C_Vertex3f(0.451446533f, 0.179618835f, 0.f),
    C_Vertex3f(0.451446533f, 0.179618835f, -1.f),
    C_Vertex3f(0.398124695f, 0.126296997f, -1.f),
    C_Vertex3f(0.398124695f, 0.126296997f, 0.f),
    C_Normal3f(-0.707126439f, -0.707087159f, 0.f),
    C_Vertex3f(0.398124695f, 0.126296997f, 0.f),
    C_Vertex3f(0.398124695f, 0.126296997f, -1.f),
    C_Vertex3f(0.260902405f, 0.263526917f, -1.f),
    C_Vertex3f(0.260902405f, 0.263526917f, 0.f),
    C_Normal3f(0.707106829f, -0.707106829f, 0.f),
    C_Vertex3f(0.260902405f, 0.263526917f, 0.f),
    C_Vertex3f(0.260902405f, 0.263526917f, -1.f),
    C_Vertex3f(0.12323761f, 0.125862122f, -1.f),
    C_Vertex3f(0.12323761f, 0.125862122f, 0.f),
    C_Normal3f(-0.707106769f, -0.707106769f, 0.f),
    C_Vertex3f(0.12323761f, 0.125862122f, 0.f),
    C_Vertex3f(0.12323761f, 0.125862122f, -1.f),
    C_Vertex3f(0.0703582764f, 0.178741455f, -1.f),
    C_Vertex3f(0.0703582764f, 0.178741455f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0703582764f, 0.178741455f, 0.f),
    C_Vertex3f(0.12323761f, 0.125862122f, 0.f),
    C_Vertex3f(0.208023071f, 0.31640625f, 0.f),
    C_Vertex3f(0.260902405f, 0.263526917f, 0.f),
    C_Vertex3f(0.26133728f, 0.369720459f, 0.f),
    C_Vertex3f(0.398124695f, 0.126296997f, 0.f),
    C_Vertex3f(0.314216614f, 0.316841125f, 0.f),
    C_Vertex3f(0.451446533f, 0.179618835f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.12411499f, 0.506942749f, 0.f),
    C_Vertex3f(0.0708007813f, 0.45362854f, 0.f),
    C_Vertex3f(0.208023071f, 0.31640625f, 0.f),
    C_Vertex3f(0.26133728f, 0.369720459f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.398124695f, 0.506507874f, 0.f),
    C_Vertex3f(0.26133728f, 0.369720459f, 0.f),
    C_Vertex3f(0.314216614f, 0.316841125f, 0.f),
    C_Vertex3f(0.450569153f, 0.45362854f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0703582764f, 0.178741455f, -1.f),
    C_Vertex3f(0.12323761f, 0.125862122f, -1.f),
    C_Vertex3f(0.208023071f, 0.31640625f, -1.f),
    C_Vertex3f(0.260902405f, 0.263526917f, -1.f),
    C_Vertex3f(0.26133728f, 0.369720459f, -1.f),
    C_Vertex3f(0.398124695f, 0.126296997f, -1.f),
    C_Vertex3f(0.314216614f, 0.316841125f, -1.f),
    C_Vertex3f(0.451446533f, 0.179618835f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.12411499f, 0.506942749f, -1.f),
    C_Vertex3f(0.0708007813f, 0.45362854f, -1.f),
    C_Vertex3f(0.208023071f, 0.31640625f, -1.f),
    C_Vertex3f(0.26133728f, 0.369720459f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.398124695f, 0.506507874f, -1.f),
    C_Vertex3f(0.26133728f, 0.369720459f, -1.f),
    C_Vertex3f(0.314216614f, 0.316841125f, -1.f),
    C_Vertex3f(0.450569153f, 0.45362854f, -1.f),
    C_End(),
    C_Translatef(0.522460938f, 0.f, 0.f),
};
static const Call char_216[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.746266544f, 0.665647268f, 0.f),
    C_Vertex3f(0.546005249f, 0.582359314f, 0.f),
    C_Vertex3f(0.546005249f, 0.582359314f, -1.f),
    C_Vertex3f(0.619140625f, 0.664352417f, -1.f),
    C_Vertex3f(0.619140625f, 0.664352417f, 0.f),
    C_Normal3f(0.646068633f, 0.763279378f, 0.f),
    C_Vertex3f(0.619140625f, 0.664352417f, 0.f),
    C_Vertex3f(0.619140625f, 0.664352417f, -1.f),
    C_Vertex3f(0.662513733f, 0.627639771f, -1.f),
    C_Vertex3f(0.662513733f, 0.627639771f, 0.f),
    C_Normal3f(0.745866179f, -0.666095853f, 0.f),
    C_Vertex3f(0.662513733f, 0.627639771f, 0.f),
    C_Vertex3f(0.662513733f, 0.627639771f, -1.f),
    C_Vertex3f(0.58555603f, 0.541465759f, -1.f),
    C_Vertex3f(0.58555603f, 0.541465759f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.829850674f, 0.557985485f, 0.f),
    C_Vertex3f(0.58555603f, 0.541465759f, 0.f),
    C_Vertex3f(0.58555603f, 0.541465759f, -1.f),
    C_Normal3f(0.847823203f, 0.530278981f, 0.f),
    C_Vertex3f(0.594963074f, 0.527475357f, 0.f),
    C_Vertex3f(0.594963074f, 0.527475357f, -1.f),
    C_Normal3f(0.872235715f, 0.489085734f, 0.f),
    C_Vertex3f(0.610713959f, 0.501242638f, 0.f),
    C_Vertex3f(0.610713959f, 0.501242638f, -1.f),
    C_Normal3f(0.904150486f, 0.427214056f, 0.f),
    C_Vertex3f(0.623401642f, 0.476758003f, 0.f),
    C_Vertex3f(0.623401642f, 0.476758003f, -1.f),
    C_Normal3f(0.938579977f, 0.345061958f, 0.f),
    C_Vertex3f(0.633026123f, 0.454021454f, 0.f),
    C_Vertex3f(0.633026123f, 0.454021454f, -1.f),
    C_Normal3f(0.966389835f, 0.257080942f, 0.f),
    C_Vertex3f(0.64068222f, 0.429754257f, 0.f),
    C_Vertex3f(0.64068222f, 0.429754257f, -1.f),
    C_Normal3f(0.983788192f, 0.179333955f, 0.f),
    C_Vertex3f(0.647245407f, 0.400569916f, 0.f),
    C_Vertex3f(0.647245407f, 0.400569916f, -1.f),
    C_Normal3f(0.994665086f, 0.103156887f, 0.f),
    C_Vertex3f(0.651620865f, 0.369747162f, 0.f),
    C_Vertex3f(0.651620865f, 0.369747162f, -1.f),
    C_Normal3f(0.999374211f, 0.0353730768f, 0.f),
    C_Vertex3f(0.653808594f, 0.337285995f, 0.f),
    C_Vertex3f(0.653808594f, 0.337285995f, -1.f),
    C_Normal3f(0.999728084f, -0.0233195163f, 0.f),
    C_Vertex3f(0.653808594f, 0.307938576f, 0.f),
    C_Vertex3f(0.653808594f, 0.307938576f, -1.f),
    C_Normal3f(0.996893048f, -0.0787669048f, 0.f),
    C_Vertex3f(0.652565956f, 0.284013033f, 0.f),
    C_Vertex3f(0.652565956f, 0.284013033f, -1.f),
    C_Normal3f(0.990939796f, -0.134307072f, 0.f),
    C_Vertex3f(0.650080681f, 0.260757208f, 0.f),
    C_Vertex3f(0.650080681f, 0.260757208f, -1.f),
    C_Normal3f(0.981428146f, -0.191830069f, 0.f),
    C_Vertex3f(0.646352768f, 0.238171101f, 0.f),
    C_Vertex3f(0.646352768f, 0.238171101f, -1.f),
    C_Normal3f(0.968033314f, -0.25082168f, 0.f),
    C_Vertex3f(0.641382217f, 0.216254711f, 0.f),
    C_Vertex3f(0.641382217f, 0.216254711f, -1.f),
    C_Normal3f(0.95052278f, -0.310654789f, 0.f),
    C_Vertex3f(0.635169029f, 0.195008039f, 0.f),
    C_Vertex3f(0.635169029f, 0.195008039f, -1.f),
    C_Normal3f(0.928787947f, -0.370611578f, 0.f),
    C_Vertex3f(0.627713203f, 0.174431086f, 0.f),
    C_Vertex3f(0.627713203f, 0.174431086f, -1.f),
    C_Normal3f(0.902867675f, -0.429918647f, 0.f),
    C_Vertex3f(0.61901474f, 0.154523849f, 0.f),
    C_Vertex3f(0.61901474f, 0.154523849f, -1.f),
    C_Normal3f(0.872526467f, -0.488566846f, 0.f),
    C_Vertex3f(0.609073639f, 0.135286331f, 0.f),
    C_Vertex3f(0.609073639f, 0.135286331f, -1.f),
    C_Normal3f(0.836513519f, -0.547946155f, 0.f),
    C_Vertex3f(0.598128796f, 0.117223859f, 0.f),
    C_Vertex3f(0.598128796f, 0.117223859f, -1.f),
    C_Normal3f(0.794691145f, -0.607013881f, 0.f),
    C_Vertex3f(0.58618021f, 0.100336432f, 0.f),
    C_Vertex3f(0.58618021f, 0.100336432f, -1.f),
    C_Normal3f(0.747166872f, -0.664636493f, 0.f),
    C_Vertex3f(0.573227882f, 0.084624052f, 0.f),
    C_Vertex3f(0.573227882f, 0.084624052f, -1.f),
    C_Normal3f(0.69434768f, -0.719639659f, 0.f),
    C_Vertex3f(0.559271812f, 0.0700867176f, 0.f),
    C_Vertex3f(0.559271812f, 0.0700867176f, -1.f),
    C_Normal3f(0.636942804f, -0.770911038f, 0.f),
    C_Vertex3f(0.544312f, 0.0567244291f, 0.f),
    C_Vertex3f(0.544312f, 0.0567244291f, -1.f),
    C_Normal3f(0.575922966f, -0.817503989f, 0.f),
    C_Vertex3f(0.528348446f, 0.0445371866f, 0.f),
    C_Vertex3f(0.528348446f, 0.0445371866f, -1.f),
    C_Normal3f(0.512440324f, -0.858722925f, 0.f),
    C_Vertex3f(0.511381149f, 0.0335249901f, 0.f),
    C_Vertex3f(0.511381149f, 0.0335249901f, -1.f),
    C_Normal3f(0.453768522f, -0.89111954f, 0.f),
    C_Vertex3f(0.49341011f, 0.0236878395f, 0.f),
    C_Vertex3f(0.49341011f, 0.0236878395f, -1.f),
    C_Normal3f(0.398585379f, -0.917131245f, 0.f),
    C_Vertex3f(0.475159168f, 0.0150803328f, 0.f),
    C_Vertex3f(0.475159168f, 0.0150803328f, -1.f),
    C_Normal3f(0.340571284f, -0.940218687f, 0.f),
    C_Vertex3f(0.456628323f, 0.00770246983f, 0.f),
    C_Vertex3f(0.456628323f, 0.00770246983f, -1.f),
    C_Normal3f(0.28028518f, -0.95991677f, 0.f),
    C_Vertex3f(0.437817574f, 0.00155425072f, 0.f),
    C_Vertex3f(0.437817574f, 0.00155425072f, -1.f),
    C_Normal3f(0.218394727f, -0.975860417f, 0.f),
    C_Vertex3f(0.418726921f, -0.00336432457f, 0.f),
    C_Vertex3f(0.418726921f, -0.00336432457f, -1.f),
    C_Normal3f(0.15564163f, -0.987813592f, 0.f),
    C_Vertex3f(0.399356365f, -0.00705325603f, 0.f),
    C_Vertex3f(0.399356365f, -0.00705325603f, -1.f),
    C_Normal3f(0.0927978158f, -0.995685041f, 0.f),
    C_Vertex3f(0.379705906f, -0.00951254368f, 0.f),
    C_Vertex3f(0.379705906f, -0.00951254368f, -1.f),
    C_Normal3f(0.0272479523f, -0.999628723f, 0.f),
    C_Vertex3f(0.359775543f, -0.0107421875f, 0.f),
    C_Vertex3f(0.359775543f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0395755582f, -0.999216557f, 0.f),
    C_Vertex3f(0.334594727f, -0.0107421875f, 0.f),
    C_Vertex3f(0.334594727f, -0.0107421875f, -1.f),
    C_Normal3f(-0.113563277f, -0.993530691f, 0.f),
    C_Vertex3f(0.30586338f, -0.00860691071f, 0.f),
    C_Vertex3f(0.30586338f, -0.00860691071f, -1.f),
    C_Normal3f(-0.196751505f, -0.980453432f, 0.f),
    C_Vertex3f(0.278552055f, -0.00433635712f, 0.f),
    C_Vertex3f(0.278552055f, -0.00433635712f, -1.f),
    C_Normal3f(-0.281861931f, -0.959455013f, 0.f),
    C_Vertex3f(0.252660751f, 0.00206947327f, 0.f),
    C_Vertex3f(0.252660751f, 0.00206947327f, -1.f),
    C_Normal3f(-0.377692163f, -0.925931215f, 0.f),
    C_Vertex3f(0.23102951f, 0.0096244812f, 0.f),
    C_Vertex3f(0.23102951f, 0.0096244812f, -1.f),
    C_Normal3f(-0.463921994f, -0.88587606f, 0.f),
    C_Vertex3f(0.210654259f, 0.0192041397f, 0.f),
    C_Vertex3f(0.210654259f, 0.0192041397f, -1.f),
    C_Normal3f(-0.526668668f, -0.850070715f, 0.f),
    C_Vertex3f(0.188694954f, 0.031794548f, 0.f),
    C_Vertex3f(0.188694954f, 0.031794548f, -1.f),
    C_Normal3f(-0.567820847f, -0.823152184f, 0.f),
    C_Vertex3f(0.165151596f, 0.0473957062f, 0.f),
    C_Vertex3f(0.165151596f, 0.0473957062f, -1.f),
    C_Normal3f(-0.595207155f, -0.803572357f, 0.f),
    C_Vertex3f(0.152587891f, 0.0567016602f, 0.f),
    C_Vertex3f(0.152587891f, 0.0567016602f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.745864093f, -0.666098058f, 0.f),
    C_Vertex3f(0.152587891f, 0.0567016602f, 0.f),
    C_Vertex3f(0.152587891f, 0.0567016602f, -1.f),
    C_Vertex3f(0.0790023804f, -0.0256958008f, -1.f),
    C_Vertex3f(0.0790023804f, -0.0256958008f, 0.f),
    C_Normal3f(-0.645261049f, -0.763962209f, 0.f),
    C_Vertex3f(0.0790023804f, -0.0256958008f, 0.f),
    C_Vertex3f(0.0790023804f, -0.0256958008f, -1.f),
    C_Vertex3f(0.0356445313f, 0.010925293f, -1.f),
    C_Vertex3f(0.0356445313f, 0.010925293f, 0.f),
    C_Normal3f(-0.745852351f, 0.66611129f, 0.f),
    C_Vertex3f(0.0356445313f, 0.010925293f, 0.f),
    C_Vertex3f(0.0356445313f, 0.010925293f, -1.f),
    C_Vertex3f(0.112823486f, 0.0973434448f, -1.f),
    C_Vertex3f(0.112823486f, 0.0973434448f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.834988892f, -0.550266862f, 0.f),
    C_Vertex3f(0.112823486f, 0.0973434448f, 0.f),
    C_Vertex3f(0.112823486f, 0.0973434448f, -1.f),
    C_Normal3f(-0.842993021f, -0.537924647f, 0.f),
    C_Vertex3f(0.107460022f, 0.105482101f, 0.f),
    C_Vertex3f(0.107460022f, 0.105482101f, -1.f),
    C_Normal3f(-0.85352242f, -0.521056056f, 0.f),
    C_Vertex3f(0.0976223946f, 0.121165395f, 0.f),
    C_Vertex3f(0.0976223946f, 0.121165395f, -1.f),
    C_Normal3f(-0.86695379f, -0.498388499f, 0.f),
    C_Vertex3f(0.0886740685f, 0.136254668f, 0.f),
    C_Vertex3f(0.0886740685f, 0.136254668f, -1.f),
    C_Normal3f(-0.881268799f, -0.47261551f, 0.f),
    C_Vertex3f(0.0806150436f, 0.150749922f, 0.f),
    C_Vertex3f(0.0806150436f, 0.150749922f, -1.f),
    C_Normal3f(-0.896448493f, -0.443147779f, 0.f),
    C_Vertex3f(0.0734453201f, 0.164651155f, 0.f),
    C_Vertex3f(0.0734453201f, 0.164651155f, -1.f),
    C_Normal3f(-0.912414789f, -0.40926677f, 0.f),
    C_Vertex3f(0.0671648979f, 0.177958369f, 0.f),
    C_Vertex3f(0.0671648979f, 0.177958369f, -1.f),
    C_Normal3f(-0.928991854f, -0.370100081f, 0.f),
    C_Vertex3f(0.061773777f, 0.190671563f, 0.f),
    C_Vertex3f(0.061773777f, 0.190671563f, -1.f),
    C_Normal3f(-0.947533607f, -0.319656193f, 0.f),
    C_Vertex3f(0.0572719574f, 0.202790737f, 0.f),
    C_Vertex3f(0.0572719574f, 0.202790737f, -1.f),
    C_Normal3f(-0.969211459f, -0.246230111f, 0.f),
    C_Vertex3f(0.0518531799f, 0.220078468f, 0.f),
    C_Vertex3f(0.0518531799f, 0.220078468f, -1.f),
    C_Normal3f(-0.985823274f, -0.167786881f, 0.f),
    C_Vertex3f(0.0464344025f, 0.245449543f, 0.f),
    C_Vertex3f(0.0464344025f, 0.245449543f, -1.f),
    C_Normal3f(-0.995619774f, -0.0934943855f, 0.f),
    C_Vertex3f(0.0428218842f, 0.273141384f, 0.f),
    C_Vertex3f(0.0428218842f, 0.273141384f, -1.f),
    C_Normal3f(-0.999524057f, -0.0308490563f, 0.f),
    C_Vertex3f(0.041015625f, 0.303153992f, 0.f),
    C_Vertex3f(0.041015625f, 0.303153992f, -1.f),
    C_Normal3f(-0.999726057f, 0.0234047379f, 0.f),
    C_Vertex3f(0.041015625f, 0.331665039f, 0.f),
    C_Vertex3f(0.041015625f, 0.331665039f, -1.f),
    C_Normal3f(-0.996978819f, 0.0776728764f, 0.f),
    C_Vertex3f(0.0422452688f, 0.355677843f, 0.f),
    C_Vertex3f(0.0422452688f, 0.355677843f, -1.f),
    C_Normal3f(-0.991185009f, 0.132484928f, 0.f),
    C_Vertex3f(0.0447045565f, 0.37901473f, 0.f),
    C_Vertex3f(0.0447045565f, 0.37901473f, -1.f),
    C_Normal3f(-0.981917739f, 0.189308316f, 0.f),
    C_Vertex3f(0.0483934879f, 0.401675701f, 0.f),
    C_Vertex3f(0.0483934879f, 0.401675701f, -1.f),
    C_Normal3f(-0.968848169f, 0.247655466f, 0.f),
    C_Vertex3f(0.0533120632f, 0.423660755f, 0.f),
    C_Vertex3f(0.0533120632f, 0.423660755f, -1.f),
    C_Normal3f(-0.951733172f, 0.306926489f, 0.f),
    C_Vertex3f(0.0594602823f, 0.444969893f, 0.f),
    C_Vertex3f(0.0594602823f, 0.444969893f, -1.f),
    C_Normal3f(-0.930446148f, 0.366428524f, 0.f),
    C_Vertex3f(0.0668381453f, 0.465603113f, 0.f),
    C_Vertex3f(0.0668381453f, 0.465603113f, -1.f),
    C_Normal3f(-0.90500164f, 0.425407797f, 0.f),
    C_Vertex3f(0.075445652f, 0.485560417f, 0.f),
    C_Vertex3f(0.075445652f, 0.485560417f, -1.f),
    C_Normal3f(-0.875250041f, 0.483670712f, 0.f),
    C_Vertex3f(0.0852828026f, 0.504841805f, 0.f),
    C_Vertex3f(0.0852828026f, 0.504841805f, -1.f),
    C_Normal3f(-0.840004802f, 0.542578876f, 0.f),
    C_Vertex3f(0.0961105824f, 0.522955656f, 0.f),
    C_Vertex3f(0.0961105824f, 0.522955656f, -1.f),
    C_Normal3f(-0.79901886f, 0.601305962f, 0.f),
    C_Vertex3f(0.107928991f, 0.539901972f, 0.f),
    C_Vertex3f(0.107928991f, 0.539901972f, -1.f),
    C_Normal3f(-0.752362728f, 0.658749044f, 0.f),
    C_Vertex3f(0.120738029f, 0.555680752f, 0.f),
    C_Vertex3f(0.120738029f, 0.555680752f, -1.f),
    C_Normal3f(-0.700396955f, 0.713753641f, 0.f),
    C_Vertex3f(0.134537697f, 0.570291996f, 0.f),
    C_Vertex3f(0.134537697f, 0.570291996f, -1.f),
    C_Normal3f(-0.643779755f, 0.765210867f, 0.f),
    C_Vertex3f(0.149327993f, 0.583735704f, 0.f),
    C_Vertex3f(0.149327993f, 0.583735704f, -1.f),
    C_Normal3f(-0.583433926f, 0.812160611f, 0.f),
    C_Vertex3f(0.165108919f, 0.596011877f, 0.f),
    C_Vertex3f(0.165108919f, 0.596011877f, -1.f),
    C_Normal3f(-0.520473659f, 0.853877723f, 0.f),
    C_Vertex3f(0.181880474f, 0.607120514f, 0.f),
    C_Vertex3f(0.181880474f, 0.607120514f, -1.f),
    C_Normal3f(-0.459558487f, 0.888147593f, 0.f),
    C_Vertex3f(0.199642658f, 0.617061615f, 0.f),
    C_Vertex3f(0.199642658f, 0.617061615f, -1.f),
    C_Normal3f(-0.399943769f, 0.916539729f, 0.f),
    C_Vertex3f(0.217903495f, 0.625760078f, 0.f),
    C_Vertex3f(0.217903495f, 0.625760078f, -1.f),
    C_Normal3f(-0.338339657f, 0.941024125f, 0.f),
    C_Vertex3f(0.236662984f, 0.633215904f, 0.f),
    C_Vertex3f(0.236662984f, 0.633215904f, -1.f),
    C_Normal3f(-0.275555104f, 0.961285353f, 0.f),
    C_Vertex3f(0.255921125f, 0.639429092f, 0.f),
    C_Vertex3f(0.255921125f, 0.639429092f, -1.f),
    C_Normal3f(-0.212433144f, 0.977175593f, 0.f),
    C_Vertex3f(0.275677919f, 0.644399643f, 0.f),
    C_Vertex3f(0.275677919f, 0.644399643f, -1.f),
    C_Normal3f(-0.149796441f, 0.988716841f, 0.f),
    C_Vertex3f(0.295933366f, 0.648127556f, 0.f),
    C_Vertex3f(0.295933366f, 0.648127556f, -1.f),
    C_Normal3f(-0.0883979425f, 0.996085286f, 0.f),
    C_Vertex3f(0.316687465f, 0.650612831f, 0.f),
    C_Vertex3f(0.316687465f, 0.650612831f, -1.f),
    C_Normal3f(-0.0265558362f, 0.999647379f, 0.f),
    C_Vertex3f(0.337940216f, 0.651855469f, 0.f),
    C_Vertex3f(0.337940216f, 0.651855469f, -1.f),
    C_Normal3f(0.0365825929f, 0.99933064f, 0.f),
    C_Vertex3f(0.363464355f, 0.651855469f, 0.f),
    C_Vertex3f(0.363464355f, 0.651855469f, -1.f),
    C_Normal3f(0.108944096f, 0.99404794f, 0.f),
    C_Vertex3f(0.390929222f, 0.649915695f, 0.f),
    C_Vertex3f(0.390929222f, 0.649915695f, -1.f),
    C_Normal3f(0.192528844f, 0.981291294f, 0.f),
    C_Vertex3f(0.41656208f, 0.646036148f, 0.f),
    C_Vertex3f(0.41656208f, 0.646036148f, -1.f),
    C_Normal3f(0.284416139f, 0.958700895f, 0.f),
    C_Vertex3f(0.44036293f, 0.640216827f, 0.f),
    C_Vertex3f(0.44036293f, 0.640216827f, -1.f),
    C_Normal3f(0.37977469f, 0.925079048f, 0.f),
    C_Vertex3f(0.462331772f, 0.632457733f, 0.f),
    C_Vertex3f(0.462331772f, 0.632457733f, -1.f),
    C_Normal3f(0.460617483f, 0.887598753f, 0.f),
    C_Vertex3f(0.48514843f, 0.62183094f, 0.f),
    C_Vertex3f(0.48514843f, 0.62183094f, -1.f),
    C_Normal3f(0.526775122f, 0.850004733f, 0.f),
    C_Vertex3f(0.508812904f, 0.608336449f, 0.f),
    C_Vertex3f(0.508812904f, 0.608336449f, -1.f),
    C_Normal3f(0.572611272f, 0.81982702f, 0.f),
    C_Vertex3f(0.533325195f, 0.591974258f, 0.f),
    C_Vertex3f(0.533325195f, 0.591974258f, -1.f),
    C_Normal3f(0.604210675f, 0.796824574f, 0.f),
    C_Vertex3f(0.546005249f, 0.582359314f, 0.f),
    C_Vertex3f(0.546005249f, 0.582359314f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.618047833f, -0.786140501f, 0.f),
    C_Vertex3f(0.493278503f, 0.52331543f, 0.f),
    C_Vertex3f(0.493278503f, 0.52331543f, -1.f),
    C_Normal3f(-0.591401577f, -0.806377172f, 0.f),
    C_Vertex3f(0.483037949f, 0.531366348f, 0.f),
    C_Vertex3f(0.483037949f, 0.531366348f, -1.f),
    C_Normal3f(-0.551238656f, -0.834347665f, 0.f),
    C_Vertex3f(0.463825226f, 0.54491663f, 0.f),
    C_Vertex3f(0.463825226f, 0.54491663f, -1.f),
    C_Normal3f(-0.489725232f, -0.871876836f, 0.f),
    C_Vertex3f(0.44588089f, 0.555915356f, 0.f),
    C_Vertex3f(0.44588089f, 0.555915356f, -1.f),
    C_Normal3f(-0.408120334f, -0.912928224f, 0.f),
    C_Vertex3f(0.429204941f, 0.564362526f, 0.f),
    C_Vertex3f(0.429204941f, 0.564362526f, -1.f),
    C_Normal3f(-0.305528283f, -0.952183008f, 0.f),
    C_Vertex3f(0.413797379f, 0.570258141f, 0.f),
    C_Vertex3f(0.413797379f, 0.570258141f, -1.f),
    C_Normal3f(-0.206991151f, -0.978342891f, 0.f),
    C_Vertex3f(0.397050858f, 0.574679852f, 0.f),
    C_Vertex3f(0.397050858f, 0.574679852f, -1.f),
    C_Normal3f(-0.117070518f, -0.993123591f, 0.f),
    C_Vertex3f(0.378965378f, 0.577627659f, 0.f),
    C_Vertex3f(0.378965378f, 0.577627659f, -1.f),
    C_Normal3f(-0.0352626368f, -0.999378085f, 0.f),
    C_Vertex3f(0.359540939f, 0.579101563f, 0.f),
    C_Vertex3f(0.359540939f, 0.579101563f, -1.f),
    C_Normal3f(0.0447525941f, -0.998998165f, 0.f),
    C_Vertex3f(0.337193489f, 0.579101563f, 0.f),
    C_Vertex3f(0.337193489f, 0.579101563f, -1.f),
    C_Normal3f(0.134633288f, -0.99089551f, 0.f),
    C_Vertex3f(0.31430006f, 0.577074885f, 0.f),
    C_Vertex3f(0.31430006f, 0.577074885f, -1.f),
    C_Normal3f(0.231042981f, -0.972943604f, 0.f),
    C_Vertex3f(0.292444706f, 0.573021531f, 0.f),
    C_Vertex3f(0.292444706f, 0.573021531f, -1.f),
    C_Normal3f(0.329894006f, -0.944017947f, 0.f),
    C_Vertex3f(0.271627426f, 0.5669415f, 0.f),
    C_Vertex3f(0.271627426f, 0.5669415f, -1.f),
    C_Normal3f(0.427964061f, -0.903795719f, 0.f),
    C_Vertex3f(0.251848221f, 0.558834791f, 0.f),
    C_Vertex3f(0.251848221f, 0.558834791f, -1.f),
    C_Normal3f(0.521824896f, -0.853052616f, 0.f),
    C_Vertex3f(0.23310709f, 0.548701406f, 0.f),
    C_Vertex3f(0.23310709f, 0.548701406f, -1.f),
    C_Normal3f(0.608402073f, -0.793628931f, 0.f),
    C_Vertex3f(0.215404034f, 0.536541343f, 0.f),
    C_Vertex3f(0.215404034f, 0.536541343f, -1.f),
    C_Normal3f(0.685457885f, -0.72811234f, 0.f),
    C_Vertex3f(0.198739052f, 0.522354603f, 0.f),
    C_Vertex3f(0.198739052f, 0.522354603f, -1.f),
    C_Normal3f(0.761523604f, -0.648137212f, 0.f),
    C_Vertex3f(0.183112144f, 0.506141186f, 0.f),
    C_Vertex3f(0.183112144f, 0.506141186f, -1.f),
    C_Normal3f(0.834279776f, -0.551341355f, 0.f),
    C_Vertex3f(0.169438601f, 0.487928271f, 0.f),
    C_Vertex3f(0.169438601f, 0.487928271f, -1.f),
    C_Normal3f(0.892103493f, -0.451831043f, 0.f),
    C_Vertex3f(0.15771842f, 0.467715859f, 0.f),
    C_Vertex3f(0.15771842f, 0.467715859f, -1.f),
    C_Normal3f(0.935188234f, -0.354150981f, 0.f),
    C_Vertex3f(0.147951603f, 0.44550395f, 0.f),
    C_Vertex3f(0.147951603f, 0.44550395f, -1.f),
    C_Normal3f(0.965141773f, -0.261727571f, 0.f),
    C_Vertex3f(0.140138149f, 0.421292543f, 0.f),
    C_Vertex3f(0.140138149f, 0.421292543f, -1.f),
    C_Normal3f(0.984274685f, -0.176644608f, 0.f),
    C_Vertex3f(0.134278059f, 0.395081639f, 0.f),
    C_Vertex3f(0.134278059f, 0.395081639f, -1.f),
    C_Normal3f(0.99500674f, -0.0998082757f, 0.f),
    C_Vertex3f(0.130371332f, 0.366871238f, 0.f),
    C_Vertex3f(0.130371332f, 0.366871238f, -1.f),
    C_Normal3f(0.999451995f, -0.0331012756f, 0.f),
    C_Vertex3f(0.128417969f, 0.336661339f, 0.f),
    C_Vertex3f(0.128417969f, 0.336661339f, -1.f),
    C_Normal3f(0.999739945f, 0.0228040405f, 0.f),
    C_Vertex3f(0.128417969f, 0.307891846f, 0.f),
    C_Vertex3f(0.128417969f, 0.307891846f, -1.f),
    C_Normal3f(0.996995926f, 0.0774548799f, 0.f),
    C_Vertex3f(0.129619598f, 0.283981323f, 0.f),
    C_Vertex3f(0.129619598f, 0.283981323f, -1.f),
    C_Normal3f(0.990623176f, 0.136622727f, 0.f),
    C_Vertex3f(0.132022858f, 0.261489868f, 0.f),
    C_Vertex3f(0.132022858f, 0.261489868f, -1.f),
    C_Normal3f(0.980012178f, 0.198937342f, 0.f),
    C_Vertex3f(0.135627747f, 0.24041748f, 0.f),
    C_Vertex3f(0.135627747f, 0.24041748f, -1.f),
    C_Normal3f(0.961453617f, 0.274967134f, 0.f),
    C_Vertex3f(0.139780045f, 0.223276138f, 0.f),
    C_Vertex3f(0.139780045f, 0.223276138f, -1.f),
    C_Normal3f(0.937428832f, 0.348176718f, 0.f),
    C_Vertex3f(0.145200729f, 0.206944466f, 0.f),
    C_Vertex3f(0.145200729f, 0.206944466f, -1.f),
    C_Normal3f(0.915401936f, 0.402541101f, 0.f),
    C_Vertex3f(0.152544022f, 0.188910484f, 0.f),
    C_Vertex3f(0.152544022f, 0.188910484f, -1.f),
    C_Normal3f(0.898724616f, 0.438513458f, 0.f),
    C_Vertex3f(0.161809921f, 0.169174194f, 0.f),
    C_Vertex3f(0.161809921f, 0.169174194f, -1.f),
    C_Normal3f(0.886531413f, 0.462668449f, 0.f),
    C_Vertex3f(0.167404175f, 0.158454895f, 0.f),
    C_Vertex3f(0.167404175f, 0.158454895f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.745830059f, 0.666136384f, 0.f),
    C_Vertex3f(0.167404175f, 0.158454895f, 0.f),
    C_Vertex3f(0.167404175f, 0.158454895f, -1.f),
    C_Vertex3f(0.493278503f, 0.52331543f, -1.f),
    C_Vertex3f(0.493278503f, 0.52331543f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.745831788f, -0.666134298f, 0.f),
    C_Vertex3f(0.529151917f, 0.478309631f, 0.f),
    C_Vertex3f(0.529151917f, 0.478309631f, -1.f),
    C_Vertex3f(0.20450592f, 0.114822388f, -1.f),
    C_Vertex3f(0.20450592f, 0.114822388f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.632692575f, 0.774403036f, 0.f),
    C_Vertex3f(0.20450592f, 0.114822388f, 0.f),
    C_Vertex3f(0.20450592f, 0.114822388f, -1.f),
    C_Normal3f(0.602297068f, 0.798272014f, 0.f),
    C_Vertex3f(0.212917328f, 0.107950211f, 0.f),
    C_Vertex3f(0.212917328f, 0.107950211f, -1.f),
    C_Normal3f(0.557482898f, 0.830188394f, 0.f),
    C_Vertex3f(0.229084015f, 0.0962781906f, 0.f),
    C_Vertex3f(0.229084015f, 0.0962781906f, -1.f),
    C_Normal3f(0.491279125f, 0.871002138f, 0.f),
    C_Vertex3f(0.244594574f, 0.0866785049f, 0.f),
    C_Vertex3f(0.244594574f, 0.0866785049f, -1.f),
    C_Normal3f(0.405462623f, 0.914111614f, 0.f),
    C_Vertex3f(0.259449005f, 0.0791511536f, 0.f),
    C_Vertex3f(0.259449005f, 0.0791511536f, -1.f),
    C_Normal3f(0.312032938f, 0.950071216f, 0.f),
    C_Vertex3f(0.275836945f, 0.0728206635f, 0.f),
    C_Vertex3f(0.275836945f, 0.0728206635f, -1.f),
    C_Normal3f(0.222598895f, 0.97491014f, 0.f),
    C_Vertex3f(0.295179367f, 0.0674161911f, 0.f),
    C_Vertex3f(0.295179367f, 0.0674161911f, -1.f),
    C_Normal3f(0.130750537f, 0.991415322f, 0.f),
    C_Vertex3f(0.315286636f, 0.0638132095f, 0.f),
    C_Vertex3f(0.315286636f, 0.0638132095f, -1.f),
    C_Normal3f(0.0412814878f, 0.999147534f, 0.f),
    C_Vertex3f(0.336158752f, 0.0620117188f, 0.f),
    C_Vertex3f(0.336158752f, 0.0620117188f, -1.f),
    C_Normal3f(-0.0461291075f, 0.998935461f, 0.f),
    C_Vertex3f(0.358888626f, 0.0620117188f, 0.f),
    C_Vertex3f(0.358888626f, 0.0620117188f, -1.f),
    C_Normal3f(-0.140195951f, 0.990123749f, 0.f),
    C_Vertex3f(0.381673217f, 0.0641134977f, 0.f),
    C_Vertex3f(0.381673217f, 0.0641134977f, -1.f),
    C_Normal3f(-0.240280688f, 0.970703483f, 0.f),
    C_Vertex3f(0.403419614f, 0.0683170557f, 0.f),
    C_Vertex3f(0.403419614f, 0.0683170557f, -1.f),
    C_Normal3f(-0.342348576f, 0.93957299f, 0.f),
    C_Vertex3f(0.424127817f, 0.0746223927f, 0.f),
    C_Vertex3f(0.424127817f, 0.0746223927f, -1.f),
    C_Normal3f(-0.442809105f, 0.896615922f, 0.f),
    C_Vertex3f(0.443797827f, 0.0830295086f, 0.f),
    C_Vertex3f(0.443797827f, 0.0830295086f, -1.f),
    C_Normal3f(-0.537984967f, 0.842954338f, 0.f),
    C_Vertex3f(0.462429643f, 0.0935384035f, 0.f),
    C_Vertex3f(0.462429643f, 0.0935384035f, -1.f),
    C_Normal3f(-0.624749243f, 0.780825436f, 0.f),
    C_Vertex3f(0.480023265f, 0.106149077f, 0.f),
    C_Vertex3f(0.480023265f, 0.106149077f, -1.f),
    C_Normal3f(-0.701010704f, 0.713150799f, 0.f),
    C_Vertex3f(0.496578693f, 0.12086153f, 0.f),
    C_Vertex3f(0.496578693f, 0.12086153f, -1.f),
    C_Normal3f(-0.771852732f, 0.635801435f, 0.f),
    C_Vertex3f(0.512095928f, 0.137675762f, 0.f),
    C_Vertex3f(0.512095928f, 0.137675762f, -1.f),
    C_Normal3f(-0.837876737f, 0.545859456f, 0.f),
    C_Vertex3f(0.525673509f, 0.15618217f, 0.f),
    C_Vertex3f(0.525673509f, 0.15618217f, -1.f),
    C_Normal3f(-0.891942322f, 0.452149093f, 0.f),
    C_Vertex3f(0.537311435f, 0.176380754f, 0.f),
    C_Vertex3f(0.537311435f, 0.176380754f, -1.f),
    C_Normal3f(-0.933572412f, 0.35838896f, 0.f),
    C_Vertex3f(0.547009706f, 0.198271513f, 0.f),
    C_Vertex3f(0.547009706f, 0.198271513f, -1.f),
    C_Normal3f(-0.963487923f, 0.267751902f, 0.f),
    C_Vertex3f(0.554768324f, 0.221854448f, 0.f),
    C_Vertex3f(0.554768324f, 0.221854448f, -1.f),
    C_Normal3f(-0.983201802f, 0.182521448f, 0.f),
    C_Vertex3f(0.560587287f, 0.24712956f, 0.f),
    C_Vertex3f(0.560587287f, 0.24712956f, -1.f),
    C_Normal3f(-0.994573116f, 0.104039595f, 0.f),
    C_Vertex3f(0.564466596f, 0.274096847f, 0.f),
    C_Vertex3f(0.564466596f, 0.274096847f, -1.f),
    C_Normal3f(-0.999383092f, 0.0351229161f, 0.f),
    C_Vertex3f(0.56640625f, 0.30275631f, 0.f),
    C_Vertex3f(0.56640625f, 0.30275631f, -1.f),
    C_Normal3f(-0.999712944f, -0.0239595883f, 0.f),
    C_Vertex3f(0.56640625f, 0.329287529f, 0.f),
    C_Vertex3f(0.56640625f, 0.329287529f, -1.f),
    C_Normal3f(-0.996781111f, -0.0801713392f, 0.f),
    C_Vertex3f(0.565242052f, 0.351332426f, 0.f),
    C_Vertex3f(0.565242052f, 0.351332426f, -1.f),
    C_Normal3f(-0.990572691f, -0.136988059f, 0.f),
    C_Vertex3f(0.562913656f, 0.37271142f, 0.f),
    C_Vertex3f(0.562913656f, 0.37271142f, -1.f),
    C_Normal3f(-0.980593026f, -0.196054354f, 0.f),
    C_Vertex3f(0.559421062f, 0.393424511f, 0.f),
    C_Vertex3f(0.559421062f, 0.393424511f, -1.f),
    C_Normal3f(-0.966457129f, -0.256827772f, 0.f),
    C_Vertex3f(0.554764271f, 0.413471699f, 0.f),
    C_Vertex3f(0.554764271f, 0.413471699f, -1.f),
    C_Normal3f(-0.947879493f, -0.318629265f, 0.f),
    C_Vertex3f(0.548943281f, 0.432852983f, 0.f),
    C_Vertex3f(0.548943281f, 0.432852983f, -1.f),
    C_Normal3f(-0.924712658f, -0.380665958f, 0.f),
    C_Vertex3f(0.541958094f, 0.451568365f, 0.f),
    C_Vertex3f(0.541958094f, 0.451568365f, -1.f),
    C_Normal3f(-0.90191251f, -0.431918651f, 0.f),
    C_Vertex3f(0.533808708f, 0.469617844f, 0.f),
    C_Vertex3f(0.533808708f, 0.469617844f, -1.f),
    C_Normal3f(-0.881459832f, -0.472258985f, 0.f),
    C_Vertex3f(0.529151917f, 0.478309631f, 0.f),
    C_Vertex3f(0.529151917f, 0.478309631f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.112823486f, 0.0973434448f, 0.f),
    C_Vertex3f(0.0356445313f, 0.010925293f, 0.f),
    C_Vertex3f(0.0790023804f, -0.0256958008f, 0.f),
    C_Vertex3f(0.152587891f, 0.0567016602f, 0.f),
    C_Vertex3f(0.129619598f, 0.283981323f, 0.f),
    C_Vertex3f(0.128417969f, 0.307891846f, 0.f),
    C_Vertex3f(0.120738029f, 0.555680752f, 0.f),
    C_Vertex3f(0.107928991f, 0.539901972f, 0.f),
    C_Vertex3f(0.107460022f, 0.105482101f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0671648979f, 0.177958369f, 0.f),
    C_Vertex3f(0.0734453201f, 0.164651155f, 0.f),
    C_Vertex3f(0.075445652f, 0.485560417f, 0.f),
    C_Vertex3f(0.0806150436f, 0.150749922f, 0.f),
    C_Vertex3f(0.0852828026f, 0.504841805f, 0.f),
    C_Vertex3f(0.0886740685f, 0.136254668f, 0.f),
    C_Vertex3f(0.0961105824f, 0.522955656f, 0.f),
    C_Vertex3f(0.0976223946f, 0.121165395f, 0.f),
    C_Vertex3f(0.107928991f, 0.539901972f, 0.f),
    C_Vertex3f(0.107460022f, 0.105482101f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0422452688f, 0.355677843f, 0.f),
    C_Vertex3f(0.0428218842f, 0.273141384f, 0.f),
    C_Vertex3f(0.0447045565f, 0.37901473f, 0.f),
    C_Vertex3f(0.0464344025f, 0.245449543f, 0.f),
    C_Vertex3f(0.0483934879f, 0.401675701f, 0.f),
    C_Vertex3f(0.0518531799f, 0.220078468f, 0.f),
    C_Vertex3f(0.0533120632f, 0.423660755f, 0.f),
    C_Vertex3f(0.0572719574f, 0.202790737f, 0.f),
    C_Vertex3f(0.0594602823f, 0.444969893f, 0.f),
    C_Vertex3f(0.061773777f, 0.190671563f, 0.f),
    C_Vertex3f(0.0668381453f, 0.465603113f, 0.f),
    C_Vertex3f(0.0671648979f, 0.177958369f, 0.f),
    C_Vertex3f(0.075445652f, 0.485560417f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.041015625f, 0.331665039f, 0.f),
    C_Vertex3f(0.041015625f, 0.303153992f, 0.f),
    C_Vertex3f(0.0428218842f, 0.273141384f, 0.f),
    C_Vertex3f(0.0422452688f, 0.355677843f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.463825226f, 0.54491663f, 0.f),
    C_Vertex3f(0.483037949f, 0.531366348f, 0.f),
    C_Vertex3f(0.48514843f, 0.62183094f, 0.f),
    C_Vertex3f(0.493278503f, 0.52331543f, 0.f),
    C_Vertex3f(0.508812904f, 0.608336449f, 0.f),
    C_Vertex3f(0.529151917f, 0.478309631f, 0.f),
    C_Vertex3f(0.533325195f, 0.591974258f, 0.f),
    C_Vertex3f(0.533808708f, 0.469617844f, 0.f),
    C_Vertex3f(0.546005249f, 0.582359314f, 0.f),
    C_Vertex3f(0.541958094f, 0.451568365f, 0.f),
    C_Vertex3f(0.548943281f, 0.432852983f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.397050858f, 0.574679852f, 0.f),
    C_Vertex3f(0.413797379f, 0.570258141f, 0.f),
    C_Vertex3f(0.41656208f, 0.646036148f, 0.f),
    C_Vertex3f(0.429204941f, 0.564362526f, 0.f),
    C_Vertex3f(0.44036293f, 0.640216827f, 0.f),
    C_Vertex3f(0.44588089f, 0.555915356f, 0.f),
    C_Vertex3f(0.462331772f, 0.632457733f, 0.f),
    C_Vertex3f(0.463825226f, 0.54491663f, 0.f),
    C_Vertex3f(0.48514843f, 0.62183094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.183112144f, 0.506141186f, 0.f),
    C_Vertex3f(0.198739052f, 0.522354603f, 0.f),
    C_Vertex3f(0.199642658f, 0.617061615f, 0.f),
    C_Vertex3f(0.215404034f, 0.536541343f, 0.f),
    C_Vertex3f(0.217903495f, 0.625760078f, 0.f),
    C_Vertex3f(0.23310709f, 0.548701406f, 0.f),
    C_Vertex3f(0.236662984f, 0.633215904f, 0.f),
    C_Vertex3f(0.251848221f, 0.558834791f, 0.f),
    C_Vertex3f(0.255921125f, 0.639429092f, 0.f),
    C_Vertex3f(0.271627426f, 0.5669415f, 0.f),
    C_Vertex3f(0.275677919f, 0.644399643f, 0.f),
    C_Vertex3f(0.292444706f, 0.573021531f, 0.f),
    C_Vertex3f(0.295933366f, 0.648127556f, 0.f),
    C_Vertex3f(0.31430006f, 0.577074885f, 0.f),
    C_Vertex3f(0.316687465f, 0.650612831f, 0.f),
    C_Vertex3f(0.337193489f, 0.579101563f, 0.f),
    C_Vertex3f(0.337940216f, 0.651855469f, 0.f),
    C_Vertex3f(0.359540939f, 0.579101563f, 0.f),
    C_Vertex3f(0.363464355f, 0.651855469f, 0.f),
    C_Vertex3f(0.378965378f, 0.577627659f, 0.f),
    C_Vertex3f(0.390929222f, 0.649915695f, 0.f),
    C_Vertex3f(0.397050858f, 0.574679852f, 0.f),
    C_Vertex3f(0.41656208f, 0.646036148f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.140138149f, 0.421292543f, 0.f),
    C_Vertex3f(0.147951603f, 0.44550395f, 0.f),
    C_Vertex3f(0.149327993f, 0.583735704f, 0.f),
    C_Vertex3f(0.15771842f, 0.467715859f, 0.f),
    C_Vertex3f(0.165108919f, 0.596011877f, 0.f),
    C_Vertex3f(0.169438601f, 0.487928271f, 0.f),
    C_Vertex3f(0.181880474f, 0.607120514f, 0.f),
    C_Vertex3f(0.183112144f, 0.506141186f, 0.f),
    C_Vertex3f(0.199642658f, 0.617061615f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134537697f, 0.570291996f, 0.f),
    C_Vertex3f(0.120738029f, 0.555680752f, 0.f),
    C_Vertex3f(0.128417969f, 0.336661339f, 0.f),
    C_Vertex3f(0.130371332f, 0.366871238f, 0.f),
    C_Vertex3f(0.134278059f, 0.395081639f, 0.f),
    C_Vertex3f(0.140138149f, 0.421292543f, 0.f),
    C_Vertex3f(0.149327993f, 0.583735704f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.58555603f, 0.541465759f, 0.f),
    C_Vertex3f(0.662513733f, 0.627639771f, 0.f),
    C_Vertex3f(0.619140625f, 0.664352417f, 0.f),
    C_Vertex3f(0.548943281f, 0.432852983f, 0.f),
    C_Vertex3f(0.554764271f, 0.413471699f, 0.f),
    C_Vertex3f(0.559421062f, 0.393424511f, 0.f),
    C_Vertex3f(0.562913656f, 0.37271142f, 0.f),
    C_Vertex3f(0.565242052f, 0.351332426f, 0.f),
    C_Vertex3f(0.56640625f, 0.329287529f, 0.f),
    C_Vertex3f(0.573227882f, 0.084624052f, 0.f),
    C_Vertex3f(0.58618021f, 0.100336432f, 0.f),
    C_Vertex3f(0.594963074f, 0.527475357f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.20450592f, 0.114822388f, 0.f),
    C_Vertex3f(0.529151917f, 0.478309631f, 0.f),
    C_Vertex3f(0.493278503f, 0.52331543f, 0.f),
    C_Vertex3f(0.167404175f, 0.158454895f, 0.f),
    C_Vertex3f(0.188694954f, 0.031794548f, 0.f),
    C_Vertex3f(0.210654259f, 0.0192041397f, 0.f),
    C_Vertex3f(0.212917328f, 0.107950211f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.188694954f, 0.031794548f, 0.f),
    C_Vertex3f(0.167404175f, 0.158454895f, 0.f),
    C_Vertex3f(0.165151596f, 0.0473957062f, 0.f),
    C_Vertex3f(0.161809921f, 0.169174194f, 0.f),
    C_Vertex3f(0.152587891f, 0.0567016602f, 0.f),
    C_Vertex3f(0.152544022f, 0.188910484f, 0.f),
    C_Vertex3f(0.145200729f, 0.206944466f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.152587891f, 0.0567016602f, 0.f),
    C_Vertex3f(0.145200729f, 0.206944466f, 0.f),
    C_Vertex3f(0.139780045f, 0.223276138f, 0.f),
    C_Vertex3f(0.135627747f, 0.24041748f, 0.f),
    C_Vertex3f(0.132022858f, 0.261489868f, 0.f),
    C_Vertex3f(0.129619598f, 0.283981323f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.641382217f, 0.216254711f, 0.f),
    C_Vertex3f(0.646352768f, 0.238171101f, 0.f),
    C_Vertex3f(0.647245407f, 0.400569916f, 0.f),
    C_Vertex3f(0.650080681f, 0.260757208f, 0.f),
    C_Vertex3f(0.651620865f, 0.369747162f, 0.f),
    C_Vertex3f(0.652565956f, 0.284013033f, 0.f),
    C_Vertex3f(0.653808594f, 0.337285995f, 0.f),
    C_Vertex3f(0.653808594f, 0.307938576f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.598128796f, 0.117223859f, 0.f),
    C_Vertex3f(0.609073639f, 0.135286331f, 0.f),
    C_Vertex3f(0.610713959f, 0.501242638f, 0.f),
    C_Vertex3f(0.61901474f, 0.154523849f, 0.f),
    C_Vertex3f(0.623401642f, 0.476758003f, 0.f),
    C_Vertex3f(0.627713203f, 0.174431086f, 0.f),
    C_Vertex3f(0.633026123f, 0.454021454f, 0.f),
    C_Vertex3f(0.635169029f, 0.195008039f, 0.f),
    C_Vertex3f(0.64068222f, 0.429754257f, 0.f),
    C_Vertex3f(0.641382217f, 0.216254711f, 0.f),
    C_Vertex3f(0.647245407f, 0.400569916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.594963074f, 0.527475357f, 0.f),
    C_Vertex3f(0.58618021f, 0.100336432f, 0.f),
    C_Vertex3f(0.598128796f, 0.117223859f, 0.f),
    C_Vertex3f(0.610713959f, 0.501242638f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.573227882f, 0.084624052f, 0.f),
    C_Vertex3f(0.56640625f, 0.329287529f, 0.f),
    C_Vertex3f(0.56640625f, 0.30275631f, 0.f),
    C_Vertex3f(0.564466596f, 0.274096847f, 0.f),
    C_Vertex3f(0.560587287f, 0.24712956f, 0.f),
    C_Vertex3f(0.559271812f, 0.0700867176f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560587287f, 0.24712956f, 0.f),
    C_Vertex3f(0.554768324f, 0.221854448f, 0.f),
    C_Vertex3f(0.559271812f, 0.0700867176f, 0.f),
    C_Vertex3f(0.547009706f, 0.198271513f, 0.f),
    C_Vertex3f(0.544312f, 0.0567244291f, 0.f),
    C_Vertex3f(0.537311435f, 0.176380754f, 0.f),
    C_Vertex3f(0.528348446f, 0.0445371866f, 0.f),
    C_Vertex3f(0.525673509f, 0.15618217f, 0.f),
    C_Vertex3f(0.512095928f, 0.137675762f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.359775543f, -0.0107421875f, 0.f),
    C_Vertex3f(0.379705906f, -0.00951254368f, 0.f),
    C_Vertex3f(0.381673217f, 0.0641134977f, 0.f),
    C_Vertex3f(0.399356365f, -0.00705325603f, 0.f),
    C_Vertex3f(0.403419614f, 0.0683170557f, 0.f),
    C_Vertex3f(0.418726921f, -0.00336432457f, 0.f),
    C_Vertex3f(0.424127817f, 0.0746223927f, 0.f),
    C_Vertex3f(0.437817574f, 0.00155425072f, 0.f),
    C_Vertex3f(0.443797827f, 0.0830295086f, 0.f),
    C_Vertex3f(0.456628323f, 0.00770246983f, 0.f),
    C_Vertex3f(0.462429643f, 0.0935384035f, 0.f),
    C_Vertex3f(0.475159168f, 0.0150803328f, 0.f),
    C_Vertex3f(0.480023265f, 0.106149077f, 0.f),
    C_Vertex3f(0.49341011f, 0.0236878395f, 0.f),
    C_Vertex3f(0.496578693f, 0.12086153f, 0.f),
    C_Vertex3f(0.511381149f, 0.0335249901f, 0.f),
    C_Vertex3f(0.512095928f, 0.137675762f, 0.f),
    C_Vertex3f(0.528348446f, 0.0445371866f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.381673217f, 0.0641134977f, 0.f),
    C_Vertex3f(0.358888626f, 0.0620117188f, 0.f),
    C_Vertex3f(0.359775543f, -0.0107421875f, 0.f),
    C_Vertex3f(0.336158752f, 0.0620117188f, 0.f),
    C_Vertex3f(0.334594727f, -0.0107421875f, 0.f),
    C_Vertex3f(0.315286636f, 0.0638132095f, 0.f),
    C_Vertex3f(0.30586338f, -0.00860691071f, 0.f),
    C_Vertex3f(0.295179367f, 0.0674161911f, 0.f),
    C_Vertex3f(0.278552055f, -0.00433635712f, 0.f),
    C_Vertex3f(0.275836945f, 0.0728206635f, 0.f),
    C_Vertex3f(0.259449005f, 0.0791511536f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278552055f, -0.00433635712f, 0.f),
    C_Vertex3f(0.259449005f, 0.0791511536f, 0.f),
    C_Vertex3f(0.252660751f, 0.00206947327f, 0.f),
    C_Vertex3f(0.244594574f, 0.0866785049f, 0.f),
    C_Vertex3f(0.23102951f, 0.0096244812f, 0.f),
    C_Vertex3f(0.229084015f, 0.0962781906f, 0.f),
    C_Vertex3f(0.212917328f, 0.107950211f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.212917328f, 0.107950211f, 0.f),
    C_Vertex3f(0.210654259f, 0.0192041397f, 0.f),
    C_Vertex3f(0.23102951f, 0.0096244812f, 0.f),
    C_Vertex3f(0.546005249f, 0.582359314f, 0.f),
    C_Vertex3f(0.548943281f, 0.432852983f, 0.f),
    C_Vertex3f(0.619140625f, 0.664352417f, 0.f),
    C_Vertex3f(0.128417969f, 0.336661339f, 0.f),
    C_Vertex3f(0.120738029f, 0.555680752f, 0.f),
    C_Vertex3f(0.128417969f, 0.307891846f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.112823486f, 0.0973434448f, -1.f),
    C_Vertex3f(0.0356445313f, 0.010925293f, -1.f),
    C_Vertex3f(0.0790023804f, -0.0256958008f, -1.f),
    C_Vertex3f(0.152587891f, 0.0567016602f, -1.f),
    C_Vertex3f(0.129619598f, 0.283981323f, -1.f),
    C_Vertex3f(0.128417969f, 0.307891846f, -1.f),
    C_Vertex3f(0.120738029f, 0.555680752f, -1.f),
    C_Vertex3f(0.107928991f, 0.539901972f, -1.f),
    C_Vertex3f(0.107460022f, 0.105482101f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0671648979f, 0.177958369f, -1.f),
    C_Vertex3f(0.0734453201f, 0.164651155f, -1.f),
    C_Vertex3f(0.075445652f, 0.485560417f, -1.f),
    C_Vertex3f(0.0806150436f, 0.150749922f, -1.f),
    C_Vertex3f(0.0852828026f, 0.504841805f, -1.f),
    C_Vertex3f(0.0886740685f, 0.136254668f, -1.f),
    C_Vertex3f(0.0961105824f, 0.522955656f, -1.f),
    C_Vertex3f(0.0976223946f, 0.121165395f, -1.f),
    C_Vertex3f(0.107928991f, 0.539901972f, -1.f),
    C_Vertex3f(0.107460022f, 0.105482101f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0422452688f, 0.355677843f, -1.f),
    C_Vertex3f(0.0428218842f, 0.273141384f, -1.f),
    C_Vertex3f(0.0447045565f, 0.37901473f, -1.f),
    C_Vertex3f(0.0464344025f, 0.245449543f, -1.f),
    C_Vertex3f(0.0483934879f, 0.401675701f, -1.f),
    C_Vertex3f(0.0518531799f, 0.220078468f, -1.f),
    C_Vertex3f(0.0533120632f, 0.423660755f, -1.f),
    C_Vertex3f(0.0572719574f, 0.202790737f, -1.f),
    C_Vertex3f(0.0594602823f, 0.444969893f, -1.f),
    C_Vertex3f(0.061773777f, 0.190671563f, -1.f),
    C_Vertex3f(0.0668381453f, 0.465603113f, -1.f),
    C_Vertex3f(0.0671648979f, 0.177958369f, -1.f),
    C_Vertex3f(0.075445652f, 0.485560417f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.041015625f, 0.331665039f, -1.f),
    C_Vertex3f(0.041015625f, 0.303153992f, -1.f),
    C_Vertex3f(0.0428218842f, 0.273141384f, -1.f),
    C_Vertex3f(0.0422452688f, 0.355677843f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.463825226f, 0.54491663f, -1.f),
    C_Vertex3f(0.483037949f, 0.531366348f, -1.f),
    C_Vertex3f(0.48514843f, 0.62183094f, -1.f),
    C_Vertex3f(0.493278503f, 0.52331543f, -1.f),
    C_Vertex3f(0.508812904f, 0.608336449f, -1.f),
    C_Vertex3f(0.529151917f, 0.478309631f, -1.f),
    C_Vertex3f(0.533325195f, 0.591974258f, -1.f),
    C_Vertex3f(0.533808708f, 0.469617844f, -1.f),
    C_Vertex3f(0.546005249f, 0.582359314f, -1.f),
    C_Vertex3f(0.541958094f, 0.451568365f, -1.f),
    C_Vertex3f(0.548943281f, 0.432852983f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.397050858f, 0.574679852f, -1.f),
    C_Vertex3f(0.413797379f, 0.570258141f, -1.f),
    C_Vertex3f(0.41656208f, 0.646036148f, -1.f),
    C_Vertex3f(0.429204941f, 0.564362526f, -1.f),
    C_Vertex3f(0.44036293f, 0.640216827f, -1.f),
    C_Vertex3f(0.44588089f, 0.555915356f, -1.f),
    C_Vertex3f(0.462331772f, 0.632457733f, -1.f),
    C_Vertex3f(0.463825226f, 0.54491663f, -1.f),
    C_Vertex3f(0.48514843f, 0.62183094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.183112144f, 0.506141186f, -1.f),
    C_Vertex3f(0.198739052f, 0.522354603f, -1.f),
    C_Vertex3f(0.199642658f, 0.617061615f, -1.f),
    C_Vertex3f(0.215404034f, 0.536541343f, -1.f),
    C_Vertex3f(0.217903495f, 0.625760078f, -1.f),
    C_Vertex3f(0.23310709f, 0.548701406f, -1.f),
    C_Vertex3f(0.236662984f, 0.633215904f, -1.f),
    C_Vertex3f(0.251848221f, 0.558834791f, -1.f),
    C_Vertex3f(0.255921125f, 0.639429092f, -1.f),
    C_Vertex3f(0.271627426f, 0.5669415f, -1.f),
    C_Vertex3f(0.275677919f, 0.644399643f, -1.f),
    C_Vertex3f(0.292444706f, 0.573021531f, -1.f),
    C_Vertex3f(0.295933366f, 0.648127556f, -1.f),
    C_Vertex3f(0.31430006f, 0.577074885f, -1.f),
    C_Vertex3f(0.316687465f, 0.650612831f, -1.f),
    C_Vertex3f(0.337193489f, 0.579101563f, -1.f),
    C_Vertex3f(0.337940216f, 0.651855469f, -1.f),
    C_Vertex3f(0.359540939f, 0.579101563f, -1.f),
    C_Vertex3f(0.363464355f, 0.651855469f, -1.f),
    C_Vertex3f(0.378965378f, 0.577627659f, -1.f),
    C_Vertex3f(0.390929222f, 0.649915695f, -1.f),
    C_Vertex3f(0.397050858f, 0.574679852f, -1.f),
    C_Vertex3f(0.41656208f, 0.646036148f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.140138149f, 0.421292543f, -1.f),
    C_Vertex3f(0.147951603f, 0.44550395f, -1.f),
    C_Vertex3f(0.149327993f, 0.583735704f, -1.f),
    C_Vertex3f(0.15771842f, 0.467715859f, -1.f),
    C_Vertex3f(0.165108919f, 0.596011877f, -1.f),
    C_Vertex3f(0.169438601f, 0.487928271f, -1.f),
    C_Vertex3f(0.181880474f, 0.607120514f, -1.f),
    C_Vertex3f(0.183112144f, 0.506141186f, -1.f),
    C_Vertex3f(0.199642658f, 0.617061615f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134537697f, 0.570291996f, -1.f),
    C_Vertex3f(0.120738029f, 0.555680752f, -1.f),
    C_Vertex3f(0.128417969f, 0.336661339f, -1.f),
    C_Vertex3f(0.130371332f, 0.366871238f, -1.f),
    C_Vertex3f(0.134278059f, 0.395081639f, -1.f),
    C_Vertex3f(0.140138149f, 0.421292543f, -1.f),
    C_Vertex3f(0.149327993f, 0.583735704f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.58555603f, 0.541465759f, -1.f),
    C_Vertex3f(0.662513733f, 0.627639771f, -1.f),
    C_Vertex3f(0.619140625f, 0.664352417f, -1.f),
    C_Vertex3f(0.548943281f, 0.432852983f, -1.f),
    C_Vertex3f(0.554764271f, 0.413471699f, -1.f),
    C_Vertex3f(0.559421062f, 0.393424511f, -1.f),
    C_Vertex3f(0.562913656f, 0.37271142f, -1.f),
    C_Vertex3f(0.565242052f, 0.351332426f, -1.f),
    C_Vertex3f(0.56640625f, 0.329287529f, -1.f),
    C_Vertex3f(0.573227882f, 0.084624052f, -1.f),
    C_Vertex3f(0.58618021f, 0.100336432f, -1.f),
    C_Vertex3f(0.594963074f, 0.527475357f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.20450592f, 0.114822388f, -1.f),
    C_Vertex3f(0.529151917f, 0.478309631f, -1.f),
    C_Vertex3f(0.493278503f, 0.52331543f, -1.f),
    C_Vertex3f(0.167404175f, 0.158454895f, -1.f),
    C_Vertex3f(0.188694954f, 0.031794548f, -1.f),
    C_Vertex3f(0.210654259f, 0.0192041397f, -1.f),
    C_Vertex3f(0.212917328f, 0.107950211f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.188694954f, 0.031794548f, -1.f),
    C_Vertex3f(0.167404175f, 0.158454895f, -1.f),
    C_Vertex3f(0.165151596f, 0.0473957062f, -1.f),
    C_Vertex3f(0.161809921f, 0.169174194f, -1.f),
    C_Vertex3f(0.152587891f, 0.0567016602f, -1.f),
    C_Vertex3f(0.152544022f, 0.188910484f, -1.f),
    C_Vertex3f(0.145200729f, 0.206944466f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.152587891f, 0.0567016602f, -1.f),
    C_Vertex3f(0.145200729f, 0.206944466f, -1.f),
    C_Vertex3f(0.139780045f, 0.223276138f, -1.f),
    C_Vertex3f(0.135627747f, 0.24041748f, -1.f),
    C_Vertex3f(0.132022858f, 0.261489868f, -1.f),
    C_Vertex3f(0.129619598f, 0.283981323f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.641382217f, 0.216254711f, -1.f),
    C_Vertex3f(0.646352768f, 0.238171101f, -1.f),
    C_Vertex3f(0.647245407f, 0.400569916f, -1.f),
    C_Vertex3f(0.650080681f, 0.260757208f, -1.f),
    C_Vertex3f(0.651620865f, 0.369747162f, -1.f),
    C_Vertex3f(0.652565956f, 0.284013033f, -1.f),
    C_Vertex3f(0.653808594f, 0.337285995f, -1.f),
    C_Vertex3f(0.653808594f, 0.307938576f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.598128796f, 0.117223859f, -1.f),
    C_Vertex3f(0.609073639f, 0.135286331f, -1.f),
    C_Vertex3f(0.610713959f, 0.501242638f, -1.f),
    C_Vertex3f(0.61901474f, 0.154523849f, -1.f),
    C_Vertex3f(0.623401642f, 0.476758003f, -1.f),
    C_Vertex3f(0.627713203f, 0.174431086f, -1.f),
    C_Vertex3f(0.633026123f, 0.454021454f, -1.f),
    C_Vertex3f(0.635169029f, 0.195008039f, -1.f),
    C_Vertex3f(0.64068222f, 0.429754257f, -1.f),
    C_Vertex3f(0.641382217f, 0.216254711f, -1.f),
    C_Vertex3f(0.647245407f, 0.400569916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.594963074f, 0.527475357f, -1.f),
    C_Vertex3f(0.58618021f, 0.100336432f, -1.f),
    C_Vertex3f(0.598128796f, 0.117223859f, -1.f),
    C_Vertex3f(0.610713959f, 0.501242638f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.573227882f, 0.084624052f, -1.f),
    C_Vertex3f(0.56640625f, 0.329287529f, -1.f),
    C_Vertex3f(0.56640625f, 0.30275631f, -1.f),
    C_Vertex3f(0.564466596f, 0.274096847f, -1.f),
    C_Vertex3f(0.560587287f, 0.24712956f, -1.f),
    C_Vertex3f(0.559271812f, 0.0700867176f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.560587287f, 0.24712956f, -1.f),
    C_Vertex3f(0.554768324f, 0.221854448f, -1.f),
    C_Vertex3f(0.559271812f, 0.0700867176f, -1.f),
    C_Vertex3f(0.547009706f, 0.198271513f, -1.f),
    C_Vertex3f(0.544312f, 0.0567244291f, -1.f),
    C_Vertex3f(0.537311435f, 0.176380754f, -1.f),
    C_Vertex3f(0.528348446f, 0.0445371866f, -1.f),
    C_Vertex3f(0.525673509f, 0.15618217f, -1.f),
    C_Vertex3f(0.512095928f, 0.137675762f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.359775543f, -0.0107421875f, -1.f),
    C_Vertex3f(0.379705906f, -0.00951254368f, -1.f),
    C_Vertex3f(0.381673217f, 0.0641134977f, -1.f),
    C_Vertex3f(0.399356365f, -0.00705325603f, -1.f),
    C_Vertex3f(0.403419614f, 0.0683170557f, -1.f),
    C_Vertex3f(0.418726921f, -0.00336432457f, -1.f),
    C_Vertex3f(0.424127817f, 0.0746223927f, -1.f),
    C_Vertex3f(0.437817574f, 0.00155425072f, -1.f),
    C_Vertex3f(0.443797827f, 0.0830295086f, -1.f),
    C_Vertex3f(0.456628323f, 0.00770246983f, -1.f),
    C_Vertex3f(0.462429643f, 0.0935384035f, -1.f),
    C_Vertex3f(0.475159168f, 0.0150803328f, -1.f),
    C_Vertex3f(0.480023265f, 0.106149077f, -1.f),
    C_Vertex3f(0.49341011f, 0.0236878395f, -1.f),
    C_Vertex3f(0.496578693f, 0.12086153f, -1.f),
    C_Vertex3f(0.511381149f, 0.0335249901f, -1.f),
    C_Vertex3f(0.512095928f, 0.137675762f, -1.f),
    C_Vertex3f(0.528348446f, 0.0445371866f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.381673217f, 0.0641134977f, -1.f),
    C_Vertex3f(0.358888626f, 0.0620117188f, -1.f),
    C_Vertex3f(0.359775543f, -0.0107421875f, -1.f),
    C_Vertex3f(0.336158752f, 0.0620117188f, -1.f),
    C_Vertex3f(0.334594727f, -0.0107421875f, -1.f),
    C_Vertex3f(0.315286636f, 0.0638132095f, -1.f),
    C_Vertex3f(0.30586338f, -0.00860691071f, -1.f),
    C_Vertex3f(0.295179367f, 0.0674161911f, -1.f),
    C_Vertex3f(0.278552055f, -0.00433635712f, -1.f),
    C_Vertex3f(0.275836945f, 0.0728206635f, -1.f),
    C_Vertex3f(0.259449005f, 0.0791511536f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278552055f, -0.00433635712f, -1.f),
    C_Vertex3f(0.259449005f, 0.0791511536f, -1.f),
    C_Vertex3f(0.252660751f, 0.00206947327f, -1.f),
    C_Vertex3f(0.244594574f, 0.0866785049f, -1.f),
    C_Vertex3f(0.23102951f, 0.0096244812f, -1.f),
    C_Vertex3f(0.229084015f, 0.0962781906f, -1.f),
    C_Vertex3f(0.212917328f, 0.107950211f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.212917328f, 0.107950211f, -1.f),
    C_Vertex3f(0.210654259f, 0.0192041397f, -1.f),
    C_Vertex3f(0.23102951f, 0.0096244812f, -1.f),
    C_Vertex3f(0.546005249f, 0.582359314f, -1.f),
    C_Vertex3f(0.548943281f, 0.432852983f, -1.f),
    C_Vertex3f(0.619140625f, 0.664352417f, -1.f),
    C_Vertex3f(0.128417969f, 0.336661339f, -1.f),
    C_Vertex3f(0.120738029f, 0.555680752f, -1.f),
    C_Vertex3f(0.128417969f, 0.307891846f, -1.f),
    C_End(),
    C_Translatef(0.696289063f, 0.f, 0.f),
};
static const Call char_217[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Normal3f(0.999998629f, -0.00168604392f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, -1.f),
    C_Normal3f(0.998944402f, -0.0459359773f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, -1.f),
    C_Normal3f(0.996721923f, -0.0809027106f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, -1.f),
    C_Normal3f(0.992785931f, -0.119900197f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, -1.f),
    C_Normal3f(0.986551166f, -0.163452342f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, -1.f),
    C_Normal3f(0.977247715f, -0.212101072f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, -1.f),
    C_Normal3f(0.963873327f, -0.266360849f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, -1.f),
    C_Normal3f(0.945149362f, -0.32663849f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, -1.f),
    C_Normal3f(0.917468131f, -0.397809178f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, -1.f),
    C_Normal3f(0.88042134f, -0.474192262f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, -1.f),
    C_Normal3f(0.836648703f, -0.547739744f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, -1.f),
    C_Normal3f(0.787085712f, -0.616843581f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, -1.f),
    C_Normal3f(0.733018816f, -0.680208266f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, -1.f),
    C_Normal3f(0.675931513f, -0.736964524f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, -1.f),
    C_Normal3f(0.617332876f, -0.786702096f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Normal3f(0.558609724f, -0.82943064f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_Normal3f(0.490787864f, -0.871279061f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Normal3f(0.413442761f, -0.91053015f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, -1.f),
    C_Normal3f(0.337979287f, -0.941153526f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_Normal3f(0.265874416f, -0.964007616f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, -1.f),
    C_Normal3f(0.198111668f, -0.980179429f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, -1.f),
    C_Normal3f(0.135230228f, -0.990814209f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, -1.f),
    C_Normal3f(0.0774159282f, -0.996998906f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, -1.f),
    C_Normal3f(0.024781188f, -0.999692857f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0218659379f, -0.999760926f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0691669211f, -0.997605145f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, -1.f),
    C_Normal3f(-0.120602004f, -0.992700875f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, -1.f),
    C_Normal3f(-0.17647697f, -0.984304726f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, -1.f),
    C_Normal3f(-0.236778811f, -0.971563518f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, -1.f),
    C_Normal3f(-0.301268518f, -0.953539252f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Normal3f(-0.369405627f, -0.929268181f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Normal3f(-0.440283626f, -0.897858739f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, -1.f),
    C_Normal3f(-0.510246277f, -0.860028386f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Normal3f(-0.5782184f, -0.815882027f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_Normal3f(-0.645087063f, -0.764109015f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_Normal3f(-0.709094107f, -0.705113769f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, -1.f),
    C_Normal3f(-0.768517613f, -0.639828682f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, -1.f),
    C_Normal3f(-0.821880043f, -0.569660723f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, -1.f),
    C_Normal3f(-0.868125141f, -0.496345371f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, -1.f),
    C_Normal3f(-0.906716228f, -0.421741188f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, -1.f),
    C_Normal3f(-0.936816335f, -0.349821717f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, -1.f),
    C_Normal3f(-0.958463192f, -0.285216063f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, -1.f),
    C_Normal3f(-0.973905027f, -0.226956055f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, -1.f),
    C_Normal3f(-0.984618485f, -0.174717858f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, -1.f),
    C_Normal3f(-0.991773546f, -0.128004253f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, -1.f),
    C_Normal3f(-0.99627322f, -0.0862539262f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, -1.f),
    C_Normal3f(-0.998803377f, -0.048905924f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Normal3f(0.999999225f, 0.00121280376f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_Normal3f(0.99925375f, 0.0386249162f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, -1.f),
    C_Normal3f(0.997593999f, 0.0693271309f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Normal3f(0.99447f, 0.105021358f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, -1.f),
    C_Normal3f(0.989158213f, 0.14685446f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, -1.f),
    C_Normal3f(0.980551183f, 0.196263924f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, -1.f),
    C_Normal3f(0.966938674f, 0.255009204f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, -1.f),
    C_Normal3f(0.942981362f, 0.332845628f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Normal3f(0.892347693f, 0.451348603f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, -1.f),
    C_Normal3f(0.808162808f, 0.588959098f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, -1.f),
    C_Normal3f(0.691974759f, 0.721921682f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_Normal3f(0.560325563f, 0.828272521f, 0.f),
    C_Vertex3f(0.214595795f, 0.0926685333f, 0.f),
    C_Vertex3f(0.214595795f, 0.0926685333f, -1.f),
    C_Normal3f(0.425149351f, 0.905123174f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_Normal3f(0.290447056f, 0.95689106f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, -1.f),
    C_Normal3f(0.164019689f, 0.986457109f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, -1.f),
    C_Normal3f(0.0528943539f, 0.998600125f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, -1.f),
    C_Normal3f(-0.0281555988f, 0.999603629f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, -1.f),
    C_Normal3f(-0.0948538408f, 0.995491266f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, -1.f),
    C_Normal3f(-0.170193285f, 0.98541069f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, -1.f),
    C_Normal3f(-0.256163627f, 0.966633379f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, -1.f),
    C_Normal3f(-0.352678686f, 0.935744524f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, -1.f),
    C_Normal3f(-0.458108366f, 0.888896406f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_Normal3f(-0.568576157f, 0.822630644f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, -1.f),
    C_Normal3f(-0.677751124f, 0.735291302f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, -1.f),
    C_Normal3f(-0.785317659f, 0.619092882f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_Normal3f(-0.870217264f, 0.492668152f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, -1.f),
    C_Normal3f(-0.92558831f, 0.378531873f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, -1.f),
    C_Normal3f(-0.959939063f, 0.280209035f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, -1.f),
    C_Normal3f(-0.980327249f, 0.197379053f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Normal3f(-0.991757751f, 0.12812756f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, -1.f),
    C_Normal3f(-0.997533739f, 0.0701881126f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.379882813f, 0.680175781f, 0.f),
    C_Vertex3f(0.379882813f, 0.680175781f, -1.f),
    C_Vertex3f(0.31640625f, 0.680175781f, -1.f),
    C_Vertex3f(0.31640625f, 0.680175781f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.31640625f, 0.680175781f, 0.f),
    C_Vertex3f(0.31640625f, 0.680175781f, -1.f),
    C_Vertex3f(0.215332031f, 0.802734375f, -1.f),
    C_Vertex3f(0.215332031f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.215332031f, 0.802734375f, 0.f),
    C_Vertex3f(0.215332031f, 0.802734375f, -1.f),
    C_Vertex3f(0.320800781f, 0.802734375f, -1.f),
    C_Vertex3f(0.320800781f, 0.802734375f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.320800781f, 0.802734375f, 0.f),
    C_Vertex3f(0.320800781f, 0.802734375f, -1.f),
    C_Vertex3f(0.379882813f, 0.680175781f, -1.f),
    C_Vertex3f(0.379882813f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214595795f, 0.0926685333f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.320800781f, 0.802734375f, 0.f),
    C_Vertex3f(0.215332031f, 0.802734375f, 0.f),
    C_Vertex3f(0.31640625f, 0.680175781f, 0.f),
    C_Vertex3f(0.379882813f, 0.680175781f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, -1.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, -1.f),
    C_Vertex3f(0.077960968f, 0.169532061f, -1.f),
    C_Vertex3f(0.081785202f, 0.150808215f, -1.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, -1.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, -1.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, -1.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, -1.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, -1.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, -1.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, -1.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, -1.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, -1.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, -1.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, -1.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, -1.f),
    C_Vertex3f(0.549636841f, 0.110161304f, -1.f),
    C_Vertex3f(0.555099487f, 0.124358654f, -1.f),
    C_Vertex3f(0.559879303f, 0.139798522f, -1.f),
    C_Vertex3f(0.563976288f, 0.156480908f, -1.f),
    C_Vertex3f(0.567390442f, 0.174405813f, -1.f),
    C_Vertex3f(0.570121765f, 0.193573236f, -1.f),
    C_Vertex3f(0.572170258f, 0.213983178f, -1.f),
    C_Vertex3f(0.573535919f, 0.235635638f, -1.f),
    C_Vertex3f(0.57421875f, 0.258530617f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_Vertex3f(0.488016248f, 0.228415847f, -1.f),
    C_Vertex3f(0.485533118f, 0.203168988f, -1.f),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Vertex3f(0.476842165f, 0.160003424f, -1.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_Vertex3f(0.470634341f, 0.142084718f, -1.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Vertex3f(0.463184953f, 0.126608729f, -1.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, -1.f),
    C_Vertex3f(0.444561481f, 0.102984905f, -1.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, -1.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, -1.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, -1.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, -1.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, -1.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, -1.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, -1.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, -1.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, -1.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, -1.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, -1.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, -1.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, -1.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, -1.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, -1.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, -1.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, -1.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, -1.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, -1.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214595795f, 0.0926685333f, -1.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Vertex3f(0.174647331f, 0.13835907f, -1.f),
    C_Vertex3f(0.183670044f, 0.016397953f, -1.f),
    C_Vertex3f(0.185180187f, 0.120863914f, -1.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Vertex3f(0.165444732f, 0.164430857f, -1.f),
    C_Vertex3f(0.162538648f, 0.177054167f, -1.f),
    C_Vertex3f(0.160116911f, 0.191049099f, -1.f),
    C_Vertex3f(0.158179522f, 0.206415653f, -1.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Vertex3f(0.155757785f, 0.241263628f, -1.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.320800781f, 0.802734375f, -1.f),
    C_Vertex3f(0.215332031f, 0.802734375f, -1.f),
    C_Vertex3f(0.31640625f, 0.680175781f, -1.f),
    C_Vertex3f(0.379882813f, 0.680175781f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_218[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Normal3f(0.999998629f, -0.00168604392f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, -1.f),
    C_Normal3f(0.998944402f, -0.0459359773f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, -1.f),
    C_Normal3f(0.996721923f, -0.0809027106f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, -1.f),
    C_Normal3f(0.992785931f, -0.119900197f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, -1.f),
    C_Normal3f(0.986551166f, -0.163452342f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, -1.f),
    C_Normal3f(0.977247715f, -0.212101072f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, -1.f),
    C_Normal3f(0.963873327f, -0.266360849f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, -1.f),
    C_Normal3f(0.945149362f, -0.32663849f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, -1.f),
    C_Normal3f(0.917468131f, -0.397809178f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, -1.f),
    C_Normal3f(0.88042134f, -0.474192262f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, -1.f),
    C_Normal3f(0.836648703f, -0.547739744f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, -1.f),
    C_Normal3f(0.787085712f, -0.616843581f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, -1.f),
    C_Normal3f(0.733018816f, -0.680208266f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, -1.f),
    C_Normal3f(0.675931513f, -0.736964524f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, -1.f),
    C_Normal3f(0.617332876f, -0.786702096f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Normal3f(0.558609724f, -0.82943064f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_Normal3f(0.490787864f, -0.871279061f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Normal3f(0.413442761f, -0.91053015f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, -1.f),
    C_Normal3f(0.337979287f, -0.941153526f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_Normal3f(0.265874416f, -0.964007616f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, -1.f),
    C_Normal3f(0.198111668f, -0.980179429f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, -1.f),
    C_Normal3f(0.135230228f, -0.990814209f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, -1.f),
    C_Normal3f(0.0774159282f, -0.996998906f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, -1.f),
    C_Normal3f(0.024781188f, -0.999692857f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0218659379f, -0.999760926f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0691669211f, -0.997605145f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, -1.f),
    C_Normal3f(-0.120602004f, -0.992700875f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, -1.f),
    C_Normal3f(-0.17647697f, -0.984304726f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, -1.f),
    C_Normal3f(-0.236778811f, -0.971563518f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, -1.f),
    C_Normal3f(-0.301268518f, -0.953539252f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Normal3f(-0.369405627f, -0.929268181f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Normal3f(-0.440283626f, -0.897858739f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, -1.f),
    C_Normal3f(-0.510246277f, -0.860028386f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Normal3f(-0.5782184f, -0.815882027f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_Normal3f(-0.645087063f, -0.764109015f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_Normal3f(-0.709094107f, -0.705113769f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, -1.f),
    C_Normal3f(-0.768517613f, -0.639828682f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, -1.f),
    C_Normal3f(-0.821880043f, -0.569660723f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, -1.f),
    C_Normal3f(-0.868125141f, -0.496345371f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, -1.f),
    C_Normal3f(-0.906716228f, -0.421741188f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, -1.f),
    C_Normal3f(-0.936816335f, -0.349821717f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, -1.f),
    C_Normal3f(-0.958463192f, -0.285216063f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, -1.f),
    C_Normal3f(-0.973905027f, -0.226956055f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, -1.f),
    C_Normal3f(-0.984618485f, -0.174717858f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, -1.f),
    C_Normal3f(-0.991773546f, -0.128004253f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, -1.f),
    C_Normal3f(-0.99627322f, -0.0862539262f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, -1.f),
    C_Normal3f(-0.998803377f, -0.048905924f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Normal3f(0.999999225f, 0.00121280376f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_Normal3f(0.99925375f, 0.0386249162f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, -1.f),
    C_Normal3f(0.997593999f, 0.0693271309f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Normal3f(0.99447f, 0.105021358f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, -1.f),
    C_Normal3f(0.989158213f, 0.14685446f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, -1.f),
    C_Normal3f(0.980551183f, 0.196263924f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, -1.f),
    C_Normal3f(0.966938674f, 0.255009204f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, -1.f),
    C_Normal3f(0.942981362f, 0.332845628f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Normal3f(0.892347693f, 0.451348603f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, -1.f),
    C_Normal3f(0.808162808f, 0.588959098f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, -1.f),
    C_Normal3f(0.691974759f, 0.721921682f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_Normal3f(0.560325563f, 0.828272521f, 0.f),
    C_Vertex3f(0.214595795f, 0.0926685333f, 0.f),
    C_Vertex3f(0.214595795f, 0.0926685333f, -1.f),
    C_Normal3f(0.425149351f, 0.905123174f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_Normal3f(0.290447056f, 0.95689106f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, -1.f),
    C_Normal3f(0.164019689f, 0.986457109f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, -1.f),
    C_Normal3f(0.0528943539f, 0.998600125f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, -1.f),
    C_Normal3f(-0.0281555988f, 0.999603629f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, -1.f),
    C_Normal3f(-0.0948538408f, 0.995491266f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, -1.f),
    C_Normal3f(-0.170193285f, 0.98541069f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, -1.f),
    C_Normal3f(-0.256163627f, 0.966633379f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, -1.f),
    C_Normal3f(-0.352678686f, 0.935744524f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, -1.f),
    C_Normal3f(-0.458108366f, 0.888896406f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_Normal3f(-0.568576157f, 0.822630644f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, -1.f),
    C_Normal3f(-0.677751124f, 0.735291302f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, -1.f),
    C_Normal3f(-0.785317659f, 0.619092882f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_Normal3f(-0.870217264f, 0.492668152f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, -1.f),
    C_Normal3f(-0.92558831f, 0.378531873f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, -1.f),
    C_Normal3f(-0.959939063f, 0.280209035f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, -1.f),
    C_Normal3f(-0.980327249f, 0.197379053f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Normal3f(-0.991757751f, 0.12812756f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, -1.f),
    C_Normal3f(-0.997533739f, 0.0701881126f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.251464844f, 0.680175781f, 0.f),
    C_Vertex3f(0.251464844f, 0.680175781f, -1.f),
    C_Vertex3f(0.309082031f, 0.802734375f, -1.f),
    C_Vertex3f(0.309082031f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.309082031f, 0.802734375f, 0.f),
    C_Vertex3f(0.309082031f, 0.802734375f, -1.f),
    C_Vertex3f(0.412597656f, 0.802734375f, -1.f),
    C_Vertex3f(0.412597656f, 0.802734375f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.412597656f, 0.802734375f, 0.f),
    C_Vertex3f(0.412597656f, 0.802734375f, -1.f),
    C_Vertex3f(0.31640625f, 0.680175781f, -1.f),
    C_Vertex3f(0.31640625f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.31640625f, 0.680175781f, 0.f),
    C_Vertex3f(0.31640625f, 0.680175781f, -1.f),
    C_Vertex3f(0.251464844f, 0.680175781f, -1.f),
    C_Vertex3f(0.251464844f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214595795f, 0.0926685333f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.309082031f, 0.802734375f, 0.f),
    C_Vertex3f(0.251464844f, 0.680175781f, 0.f),
    C_Vertex3f(0.31640625f, 0.680175781f, 0.f),
    C_Vertex3f(0.412597656f, 0.802734375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, -1.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, -1.f),
    C_Vertex3f(0.077960968f, 0.169532061f, -1.f),
    C_Vertex3f(0.081785202f, 0.150808215f, -1.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, -1.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, -1.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, -1.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, -1.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, -1.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, -1.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, -1.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, -1.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, -1.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, -1.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, -1.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, -1.f),
    C_Vertex3f(0.549636841f, 0.110161304f, -1.f),
    C_Vertex3f(0.555099487f, 0.124358654f, -1.f),
    C_Vertex3f(0.559879303f, 0.139798522f, -1.f),
    C_Vertex3f(0.563976288f, 0.156480908f, -1.f),
    C_Vertex3f(0.567390442f, 0.174405813f, -1.f),
    C_Vertex3f(0.570121765f, 0.193573236f, -1.f),
    C_Vertex3f(0.572170258f, 0.213983178f, -1.f),
    C_Vertex3f(0.573535919f, 0.235635638f, -1.f),
    C_Vertex3f(0.57421875f, 0.258530617f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_Vertex3f(0.488016248f, 0.228415847f, -1.f),
    C_Vertex3f(0.485533118f, 0.203168988f, -1.f),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Vertex3f(0.476842165f, 0.160003424f, -1.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_Vertex3f(0.470634341f, 0.142084718f, -1.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Vertex3f(0.463184953f, 0.126608729f, -1.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, -1.f),
    C_Vertex3f(0.444561481f, 0.102984905f, -1.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, -1.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, -1.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, -1.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, -1.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, -1.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, -1.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, -1.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, -1.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, -1.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, -1.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, -1.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, -1.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, -1.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, -1.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, -1.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, -1.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, -1.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, -1.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, -1.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214595795f, 0.0926685333f, -1.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Vertex3f(0.174647331f, 0.13835907f, -1.f),
    C_Vertex3f(0.183670044f, 0.016397953f, -1.f),
    C_Vertex3f(0.185180187f, 0.120863914f, -1.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Vertex3f(0.165444732f, 0.164430857f, -1.f),
    C_Vertex3f(0.162538648f, 0.177054167f, -1.f),
    C_Vertex3f(0.160116911f, 0.191049099f, -1.f),
    C_Vertex3f(0.158179522f, 0.206415653f, -1.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Vertex3f(0.155757785f, 0.241263628f, -1.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.309082031f, 0.802734375f, -1.f),
    C_Vertex3f(0.251464844f, 0.680175781f, -1.f),
    C_Vertex3f(0.31640625f, 0.680175781f, -1.f),
    C_Vertex3f(0.412597656f, 0.802734375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_219[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Normal3f(0.999998629f, -0.00168604392f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, -1.f),
    C_Normal3f(0.998944402f, -0.0459359773f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, -1.f),
    C_Normal3f(0.996721923f, -0.0809027106f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, -1.f),
    C_Normal3f(0.992785931f, -0.119900197f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, -1.f),
    C_Normal3f(0.986551166f, -0.163452342f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, -1.f),
    C_Normal3f(0.977247715f, -0.212101072f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, -1.f),
    C_Normal3f(0.963873327f, -0.266360849f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, -1.f),
    C_Normal3f(0.945149362f, -0.32663849f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, -1.f),
    C_Normal3f(0.917468131f, -0.397809178f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, -1.f),
    C_Normal3f(0.88042134f, -0.474192262f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, -1.f),
    C_Normal3f(0.836648703f, -0.547739744f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, -1.f),
    C_Normal3f(0.787085712f, -0.616843581f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, -1.f),
    C_Normal3f(0.733018816f, -0.680208266f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, -1.f),
    C_Normal3f(0.675931513f, -0.736964524f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, -1.f),
    C_Normal3f(0.617332876f, -0.786702096f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Normal3f(0.558609724f, -0.82943064f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_Normal3f(0.490787864f, -0.871279061f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Normal3f(0.413442761f, -0.91053015f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, -1.f),
    C_Normal3f(0.337979287f, -0.941153526f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_Normal3f(0.265874416f, -0.964007616f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, -1.f),
    C_Normal3f(0.198111668f, -0.980179429f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, -1.f),
    C_Normal3f(0.135230228f, -0.990814209f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, -1.f),
    C_Normal3f(0.0774159282f, -0.996998906f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, -1.f),
    C_Normal3f(0.024781188f, -0.999692857f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0218659379f, -0.999760926f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0691669211f, -0.997605145f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, -1.f),
    C_Normal3f(-0.120602004f, -0.992700875f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, -1.f),
    C_Normal3f(-0.17647697f, -0.984304726f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, -1.f),
    C_Normal3f(-0.236778811f, -0.971563518f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, -1.f),
    C_Normal3f(-0.301268518f, -0.953539252f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Normal3f(-0.369405627f, -0.929268181f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Normal3f(-0.440283626f, -0.897858739f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, -1.f),
    C_Normal3f(-0.510246277f, -0.860028386f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Normal3f(-0.5782184f, -0.815882027f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_Normal3f(-0.645087063f, -0.764109015f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_Normal3f(-0.709094107f, -0.705113769f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, -1.f),
    C_Normal3f(-0.768517613f, -0.639828682f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, -1.f),
    C_Normal3f(-0.821880043f, -0.569660723f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, -1.f),
    C_Normal3f(-0.868125141f, -0.496345371f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, -1.f),
    C_Normal3f(-0.906716228f, -0.421741188f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, -1.f),
    C_Normal3f(-0.936816335f, -0.349821717f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, -1.f),
    C_Normal3f(-0.958463192f, -0.285216063f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, -1.f),
    C_Normal3f(-0.973905027f, -0.226956055f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, -1.f),
    C_Normal3f(-0.984618485f, -0.174717858f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, -1.f),
    C_Normal3f(-0.991773546f, -0.128004253f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, -1.f),
    C_Normal3f(-0.99627322f, -0.0862539262f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, -1.f),
    C_Normal3f(-0.998803377f, -0.048905924f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Normal3f(0.999999225f, 0.00121280376f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_Normal3f(0.99925375f, 0.0386249162f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, -1.f),
    C_Normal3f(0.997593999f, 0.0693271309f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Normal3f(0.99447f, 0.105021358f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, -1.f),
    C_Normal3f(0.989158213f, 0.14685446f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, -1.f),
    C_Normal3f(0.980551183f, 0.196263924f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, -1.f),
    C_Normal3f(0.966938674f, 0.255009204f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, -1.f),
    C_Normal3f(0.942981362f, 0.332845628f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Normal3f(0.892347693f, 0.451348603f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, -1.f),
    C_Normal3f(0.808162808f, 0.588959098f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, -1.f),
    C_Normal3f(0.691974759f, 0.721921682f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_Normal3f(0.560325563f, 0.828272521f, 0.f),
    C_Vertex3f(0.214595795f, 0.0926685333f, 0.f),
    C_Vertex3f(0.214595795f, 0.0926685333f, -1.f),
    C_Normal3f(0.425149351f, 0.905123174f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_Normal3f(0.290447056f, 0.95689106f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, -1.f),
    C_Normal3f(0.164019689f, 0.986457109f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, -1.f),
    C_Normal3f(0.0528943539f, 0.998600125f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, -1.f),
    C_Normal3f(-0.0281555988f, 0.999603629f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, -1.f),
    C_Normal3f(-0.0948538408f, 0.995491266f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, -1.f),
    C_Normal3f(-0.170193285f, 0.98541069f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, -1.f),
    C_Normal3f(-0.256163627f, 0.966633379f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, -1.f),
    C_Normal3f(-0.352678686f, 0.935744524f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, -1.f),
    C_Normal3f(-0.458108366f, 0.888896406f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_Normal3f(-0.568576157f, 0.822630644f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, -1.f),
    C_Normal3f(-0.677751124f, 0.735291302f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, -1.f),
    C_Normal3f(-0.785317659f, 0.619092882f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_Normal3f(-0.870217264f, 0.492668152f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, -1.f),
    C_Normal3f(-0.92558831f, 0.378531873f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, -1.f),
    C_Normal3f(-0.959939063f, 0.280209035f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, -1.f),
    C_Normal3f(-0.980327249f, 0.197379053f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Normal3f(-0.991757751f, 0.12812756f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, -1.f),
    C_Normal3f(-0.997533739f, 0.0701881126f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.325195313f, 0.754882813f, 0.f),
    C_Vertex3f(0.325195313f, 0.754882813f, -1.f),
    C_Vertex3f(0.276123047f, 0.680175781f, -1.f),
    C_Vertex3f(0.276123047f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.276123047f, 0.680175781f, 0.f),
    C_Vertex3f(0.276123047f, 0.680175781f, -1.f),
    C_Vertex3f(0.186767578f, 0.680175781f, -1.f),
    C_Vertex3f(0.186767578f, 0.680175781f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(0.186767578f, 0.680175781f, 0.f),
    C_Vertex3f(0.186767578f, 0.680175781f, -1.f),
    C_Vertex3f(0.280273438f, 0.802734375f, -1.f),
    C_Vertex3f(0.280273438f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.280273438f, 0.802734375f, 0.f),
    C_Vertex3f(0.280273438f, 0.802734375f, -1.f),
    C_Vertex3f(0.364257813f, 0.802734375f, -1.f),
    C_Vertex3f(0.364257813f, 0.802734375f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.364257813f, 0.802734375f, 0.f),
    C_Vertex3f(0.364257813f, 0.802734375f, -1.f),
    C_Vertex3f(0.462646484f, 0.680175781f, -1.f),
    C_Vertex3f(0.462646484f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.462646484f, 0.680175781f, 0.f),
    C_Vertex3f(0.462646484f, 0.680175781f, -1.f),
    C_Vertex3f(0.373291016f, 0.680175781f, -1.f),
    C_Vertex3f(0.373291016f, 0.680175781f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.373291016f, 0.680175781f, 0.f),
    C_Vertex3f(0.373291016f, 0.680175781f, -1.f),
    C_Vertex3f(0.325195313f, 0.754882813f, -1.f),
    C_Vertex3f(0.325195313f, 0.754882813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214595795f, 0.0926685333f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186767578f, 0.680175781f, 0.f),
    C_Vertex3f(0.276123047f, 0.680175781f, 0.f),
    C_Vertex3f(0.280273438f, 0.802734375f, 0.f),
    C_Vertex3f(0.325195313f, 0.754882813f, 0.f),
    C_Vertex3f(0.364257813f, 0.802734375f, 0.f),
    C_Vertex3f(0.373291016f, 0.680175781f, 0.f),
    C_Vertex3f(0.462646484f, 0.680175781f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, -1.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, -1.f),
    C_Vertex3f(0.077960968f, 0.169532061f, -1.f),
    C_Vertex3f(0.081785202f, 0.150808215f, -1.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, -1.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, -1.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, -1.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, -1.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, -1.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, -1.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, -1.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, -1.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, -1.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, -1.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, -1.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, -1.f),
    C_Vertex3f(0.549636841f, 0.110161304f, -1.f),
    C_Vertex3f(0.555099487f, 0.124358654f, -1.f),
    C_Vertex3f(0.559879303f, 0.139798522f, -1.f),
    C_Vertex3f(0.563976288f, 0.156480908f, -1.f),
    C_Vertex3f(0.567390442f, 0.174405813f, -1.f),
    C_Vertex3f(0.570121765f, 0.193573236f, -1.f),
    C_Vertex3f(0.572170258f, 0.213983178f, -1.f),
    C_Vertex3f(0.573535919f, 0.235635638f, -1.f),
    C_Vertex3f(0.57421875f, 0.258530617f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_Vertex3f(0.488016248f, 0.228415847f, -1.f),
    C_Vertex3f(0.485533118f, 0.203168988f, -1.f),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Vertex3f(0.476842165f, 0.160003424f, -1.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_Vertex3f(0.470634341f, 0.142084718f, -1.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Vertex3f(0.463184953f, 0.126608729f, -1.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, -1.f),
    C_Vertex3f(0.444561481f, 0.102984905f, -1.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, -1.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, -1.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, -1.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, -1.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, -1.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, -1.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, -1.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, -1.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, -1.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, -1.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, -1.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, -1.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, -1.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, -1.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, -1.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, -1.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, -1.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, -1.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, -1.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214595795f, 0.0926685333f, -1.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Vertex3f(0.174647331f, 0.13835907f, -1.f),
    C_Vertex3f(0.183670044f, 0.016397953f, -1.f),
    C_Vertex3f(0.185180187f, 0.120863914f, -1.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Vertex3f(0.165444732f, 0.164430857f, -1.f),
    C_Vertex3f(0.162538648f, 0.177054167f, -1.f),
    C_Vertex3f(0.160116911f, 0.191049099f, -1.f),
    C_Vertex3f(0.158179522f, 0.206415653f, -1.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Vertex3f(0.155757785f, 0.241263628f, -1.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186767578f, 0.680175781f, -1.f),
    C_Vertex3f(0.276123047f, 0.680175781f, -1.f),
    C_Vertex3f(0.280273438f, 0.802734375f, -1.f),
    C_Vertex3f(0.325195313f, 0.754882813f, -1.f),
    C_Vertex3f(0.364257813f, 0.802734375f, -1.f),
    C_Vertex3f(0.373291016f, 0.680175781f, -1.f),
    C_Vertex3f(0.462646484f, 0.680175781f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_220[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Normal3f(0.999998629f, -0.00168604392f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, -1.f),
    C_Normal3f(0.998944402f, -0.0459359773f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, -1.f),
    C_Normal3f(0.996721923f, -0.0809027106f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, -1.f),
    C_Normal3f(0.992785931f, -0.119900197f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, -1.f),
    C_Normal3f(0.986551166f, -0.163452342f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, -1.f),
    C_Normal3f(0.977247715f, -0.212101072f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, -1.f),
    C_Normal3f(0.963873327f, -0.266360849f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, -1.f),
    C_Normal3f(0.945149362f, -0.32663849f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, -1.f),
    C_Normal3f(0.917468131f, -0.397809178f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, -1.f),
    C_Normal3f(0.88042134f, -0.474192262f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, -1.f),
    C_Normal3f(0.836648703f, -0.547739744f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, -1.f),
    C_Normal3f(0.787085712f, -0.616843581f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, -1.f),
    C_Normal3f(0.733018816f, -0.680208266f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, -1.f),
    C_Normal3f(0.675931513f, -0.736964524f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, -1.f),
    C_Normal3f(0.617332876f, -0.786702096f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Normal3f(0.558609724f, -0.82943064f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_Normal3f(0.490787864f, -0.871279061f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Normal3f(0.413442761f, -0.91053015f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, -1.f),
    C_Normal3f(0.337979287f, -0.941153526f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_Normal3f(0.265874416f, -0.964007616f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, -1.f),
    C_Normal3f(0.198111668f, -0.980179429f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, -1.f),
    C_Normal3f(0.135230228f, -0.990814209f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, -1.f),
    C_Normal3f(0.0774159282f, -0.996998906f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, -1.f),
    C_Normal3f(0.024781188f, -0.999692857f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0218659379f, -0.999760926f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0691669211f, -0.997605145f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, -1.f),
    C_Normal3f(-0.120602004f, -0.992700875f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, -1.f),
    C_Normal3f(-0.17647697f, -0.984304726f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, -1.f),
    C_Normal3f(-0.236778811f, -0.971563518f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, -1.f),
    C_Normal3f(-0.301268518f, -0.953539252f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Normal3f(-0.369405627f, -0.929268181f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Normal3f(-0.440283626f, -0.897858739f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, -1.f),
    C_Normal3f(-0.510246277f, -0.860028386f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Normal3f(-0.5782184f, -0.815882027f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_Normal3f(-0.645087063f, -0.764109015f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_Normal3f(-0.709094107f, -0.705113769f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, -1.f),
    C_Normal3f(-0.768517613f, -0.639828682f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, -1.f),
    C_Normal3f(-0.821880043f, -0.569660723f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, -1.f),
    C_Normal3f(-0.868125141f, -0.496345371f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, -1.f),
    C_Normal3f(-0.906716228f, -0.421741188f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, -1.f),
    C_Normal3f(-0.936816335f, -0.349821717f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, -1.f),
    C_Normal3f(-0.958463192f, -0.285216063f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, -1.f),
    C_Normal3f(-0.973905027f, -0.226956055f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, -1.f),
    C_Normal3f(-0.984618485f, -0.174717858f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, -1.f),
    C_Normal3f(-0.991773546f, -0.128004253f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, -1.f),
    C_Normal3f(-0.99627322f, -0.0862539262f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, -1.f),
    C_Normal3f(-0.998803377f, -0.048905924f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Normal3f(0.999999225f, 0.00121280376f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_Normal3f(0.99925375f, 0.0386249162f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, -1.f),
    C_Normal3f(0.997593999f, 0.0693271309f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Normal3f(0.99447f, 0.105021358f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, -1.f),
    C_Normal3f(0.989158213f, 0.14685446f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, -1.f),
    C_Normal3f(0.980551183f, 0.196263924f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, -1.f),
    C_Normal3f(0.966938674f, 0.255009204f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, -1.f),
    C_Normal3f(0.942981362f, 0.332845628f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Normal3f(0.892347693f, 0.451348603f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, -1.f),
    C_Normal3f(0.808162808f, 0.588959098f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, -1.f),
    C_Normal3f(0.691974759f, 0.721921682f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_Normal3f(0.560325563f, 0.828272521f, 0.f),
    C_Vertex3f(0.214595795f, 0.0926685333f, 0.f),
    C_Vertex3f(0.214595795f, 0.0926685333f, -1.f),
    C_Normal3f(0.425149351f, 0.905123174f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_Normal3f(0.290447056f, 0.95689106f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, -1.f),
    C_Normal3f(0.164019689f, 0.986457109f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, -1.f),
    C_Normal3f(0.0528943539f, 0.998600125f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, -1.f),
    C_Normal3f(-0.0281555988f, 0.999603629f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, -1.f),
    C_Normal3f(-0.0948538408f, 0.995491266f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, -1.f),
    C_Normal3f(-0.170193285f, 0.98541069f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, -1.f),
    C_Normal3f(-0.256163627f, 0.966633379f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, -1.f),
    C_Normal3f(-0.352678686f, 0.935744524f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, -1.f),
    C_Normal3f(-0.458108366f, 0.888896406f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_Normal3f(-0.568576157f, 0.822630644f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, -1.f),
    C_Normal3f(-0.677751124f, 0.735291302f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, -1.f),
    C_Normal3f(-0.785317659f, 0.619092882f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_Normal3f(-0.870217264f, 0.492668152f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, -1.f),
    C_Normal3f(-0.92558831f, 0.378531873f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, -1.f),
    C_Normal3f(-0.959939063f, 0.280209035f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, -1.f),
    C_Normal3f(-0.980327249f, 0.197379053f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Normal3f(-0.991757751f, 0.12812756f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, -1.f),
    C_Normal3f(-0.997533739f, 0.0701881126f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.198730469f, 0.680175781f, 0.f),
    C_Vertex3f(0.198730469f, 0.680175781f, -1.f),
    C_Vertex3f(0.198730469f, 0.76953125f, -1.f),
    C_Vertex3f(0.198730469f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.198730469f, 0.76953125f, 0.f),
    C_Vertex3f(0.198730469f, 0.76953125f, -1.f),
    C_Vertex3f(0.280761719f, 0.76953125f, -1.f),
    C_Vertex3f(0.280761719f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.280761719f, 0.76953125f, 0.f),
    C_Vertex3f(0.280761719f, 0.76953125f, -1.f),
    C_Vertex3f(0.280761719f, 0.680175781f, -1.f),
    C_Vertex3f(0.280761719f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.280761719f, 0.680175781f, 0.f),
    C_Vertex3f(0.280761719f, 0.680175781f, -1.f),
    C_Vertex3f(0.198730469f, 0.680175781f, -1.f),
    C_Vertex3f(0.198730469f, 0.680175781f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.361816406f, 0.680175781f, 0.f),
    C_Vertex3f(0.361816406f, 0.680175781f, -1.f),
    C_Vertex3f(0.361816406f, 0.76953125f, -1.f),
    C_Vertex3f(0.361816406f, 0.76953125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.361816406f, 0.76953125f, 0.f),
    C_Vertex3f(0.361816406f, 0.76953125f, -1.f),
    C_Vertex3f(0.443847656f, 0.76953125f, -1.f),
    C_Vertex3f(0.443847656f, 0.76953125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.443847656f, 0.76953125f, 0.f),
    C_Vertex3f(0.443847656f, 0.76953125f, -1.f),
    C_Vertex3f(0.443847656f, 0.680175781f, -1.f),
    C_Vertex3f(0.443847656f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.443847656f, 0.680175781f, 0.f),
    C_Vertex3f(0.443847656f, 0.680175781f, -1.f),
    C_Vertex3f(0.361816406f, 0.680175781f, -1.f),
    C_Vertex3f(0.361816406f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155273438f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, 0.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, 0.f),
    C_Vertex3f(0.077960968f, 0.169532061f, 0.f),
    C_Vertex3f(0.081785202f, 0.150808215f, 0.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, 0.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, 0.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, 0.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, 0.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, 0.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, 0.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.57421875f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, 0.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, 0.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, 0.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, 0.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, 0.f),
    C_Vertex3f(0.549636841f, 0.110161304f, 0.f),
    C_Vertex3f(0.555099487f, 0.124358654f, 0.f),
    C_Vertex3f(0.559879303f, 0.139798522f, 0.f),
    C_Vertex3f(0.563976288f, 0.156480908f, 0.f),
    C_Vertex3f(0.567390442f, 0.174405813f, 0.f),
    C_Vertex3f(0.570121765f, 0.193573236f, 0.f),
    C_Vertex3f(0.572170258f, 0.213983178f, 0.f),
    C_Vertex3f(0.573535919f, 0.235635638f, 0.f),
    C_Vertex3f(0.57421875f, 0.258530617f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493772149f, 0.0388785601f, 0.f),
    C_Vertex3f(0.489257813f, 0.640625f, 0.f),
    C_Vertex3f(0.489257813f, 0.256105423f, 0.f),
    C_Vertex3f(0.488016248f, 0.228415847f, 0.f),
    C_Vertex3f(0.485533118f, 0.203168988f, 0.f),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481808424f, 0.180364847f, 0.f),
    C_Vertex3f(0.476842165f, 0.160003424f, 0.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, 0.f),
    C_Vertex3f(0.470634341f, 0.142084718f, 0.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.463184953f, 0.126608729f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466354847f, 0.0204133987f, 0.f),
    C_Vertex3f(0.454493999f, 0.113575459f, 0.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, 0.f),
    C_Vertex3f(0.444561481f, 0.102984905f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, 0.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, 0.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, 0.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, 0.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, 0.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, 0.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, 0.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, 0.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, 0.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, 0.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, 0.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, 0.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, 0.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, 0.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, 0.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, 0.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, 0.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, 0.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214595795f, 0.0926685333f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, 0.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.174647331f, 0.13835907f, 0.f),
    C_Vertex3f(0.183670044f, 0.016397953f, 0.f),
    C_Vertex3f(0.185180187f, 0.120863914f, 0.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, 0.f),
    C_Vertex3f(0.198496342f, 0.105633736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1693573f, 0.0241522789f, 0.f),
    C_Vertex3f(0.168835163f, 0.153179169f, 0.f),
    C_Vertex3f(0.165444732f, 0.164430857f, 0.f),
    C_Vertex3f(0.162538648f, 0.177054167f, 0.f),
    C_Vertex3f(0.160116911f, 0.191049099f, 0.f),
    C_Vertex3f(0.158179522f, 0.206415653f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156068802f, 0.0327734947f, 0.f),
    C_Vertex3f(0.15672648f, 0.22315383f, 0.f),
    C_Vertex3f(0.155757785f, 0.241263628f, 0.f),
    C_Vertex3f(0.155273438f, 0.260745049f, 0.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.198730469f, 0.76953125f, 0.f),
    C_Vertex3f(0.198730469f, 0.680175781f, 0.f),
    C_Vertex3f(0.280761719f, 0.680175781f, 0.f),
    C_Vertex3f(0.280761719f, 0.76953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361816406f, 0.76953125f, 0.f),
    C_Vertex3f(0.361816406f, 0.680175781f, 0.f),
    C_Vertex3f(0.443847656f, 0.680175781f, 0.f),
    C_Vertex3f(0.443847656f, 0.76953125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0710773468f, 0.233772397f, 0.f),
    C_Vertex3f(0.0703125f, 0.640625f, 0.f),
    C_Vertex3f(0.0703125f, 0.257875443f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.155273438f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Vertex3f(0.0726070404f, 0.211014152f, -1.f),
    C_Vertex3f(0.0749015808f, 0.189600706f, -1.f),
    C_Vertex3f(0.077960968f, 0.169532061f, -1.f),
    C_Vertex3f(0.081785202f, 0.150808215f, -1.f),
    C_Vertex3f(0.0863742828f, 0.13342917f, -1.f),
    C_Vertex3f(0.0917282104f, 0.117394924f, -1.f),
    C_Vertex3f(0.0978469849f, 0.102705479f, -1.f),
    C_Vertex3f(0.104990005f, 0.0888829231f, -1.f),
    C_Vertex3f(0.113157272f, 0.0759272575f, -1.f),
    C_Vertex3f(0.122348785f, 0.0638384819f, -1.f),
    C_Vertex3f(0.132564545f, 0.0526165962f, -1.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.57421875f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Vertex3f(0.505831599f, 0.0491040945f, -1.f),
    C_Vertex3f(0.516791582f, 0.0599915981f, -1.f),
    C_Vertex3f(0.526652098f, 0.0715410709f, -1.f),
    C_Vertex3f(0.535413146f, 0.0837525129f, -1.f),
    C_Vertex3f(0.543074727f, 0.0966259241f, -1.f),
    C_Vertex3f(0.549636841f, 0.110161304f, -1.f),
    C_Vertex3f(0.555099487f, 0.124358654f, -1.f),
    C_Vertex3f(0.559879303f, 0.139798522f, -1.f),
    C_Vertex3f(0.563976288f, 0.156480908f, -1.f),
    C_Vertex3f(0.567390442f, 0.174405813f, -1.f),
    C_Vertex3f(0.570121765f, 0.193573236f, -1.f),
    C_Vertex3f(0.572170258f, 0.213983178f, -1.f),
    C_Vertex3f(0.573535919f, 0.235635638f, -1.f),
    C_Vertex3f(0.57421875f, 0.258530617f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.493772149f, 0.0388785601f, -1.f),
    C_Vertex3f(0.489257813f, 0.640625f, -1.f),
    C_Vertex3f(0.489257813f, 0.256105423f, -1.f),
    C_Vertex3f(0.488016248f, 0.228415847f, -1.f),
    C_Vertex3f(0.485533118f, 0.203168988f, -1.f),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481808424f, 0.180364847f, -1.f),
    C_Vertex3f(0.476842165f, 0.160003424f, -1.f),
    C_Vertex3f(0.480613232f, 0.0293149948f, -1.f),
    C_Vertex3f(0.470634341f, 0.142084718f, -1.f),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Vertex3f(0.463184953f, 0.126608729f, -1.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.466354847f, 0.0204133987f, -1.f),
    C_Vertex3f(0.454493999f, 0.113575459f, -1.f),
    C_Vertex3f(0.450983167f, 0.0126245022f, -1.f),
    C_Vertex3f(0.444561481f, 0.102984905f, -1.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_Vertex3f(0.432950854f, 0.0937181711f, -1.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Vertex3f(0.232753754f, -0.00104928017f, -1.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_Vertex3f(0.251163483f, -0.0049264431f, -1.f),
    C_Vertex3f(0.254461288f, 0.0739431381f, -1.f),
    C_Vertex3f(0.270597458f, -0.0078343153f, -1.f),
    C_Vertex3f(0.277544022f, 0.0685930252f, -1.f),
    C_Vertex3f(0.291055679f, -0.00977289677f, -1.f),
    C_Vertex3f(0.302726746f, 0.0659179688f, -1.f),
    C_Vertex3f(0.312538147f, -0.0107421875f, -1.f),
    C_Vertex3f(0.328046799f, 0.0659179688f, -1.f),
    C_Vertex3f(0.335373878f, -0.0107421875f, -1.f),
    C_Vertex3f(0.349726081f, 0.0672417879f, -1.f),
    C_Vertex3f(0.357425332f, -0.00962948799f, -1.f),
    C_Vertex3f(0.369727254f, 0.0698894262f, -1.f),
    C_Vertex3f(0.37836349f, -0.00740408897f, -1.f),
    C_Vertex3f(0.388050318f, 0.0738608837f, -1.f),
    C_Vertex3f(0.398188353f, -0.00406599045f, -1.f),
    C_Vertex3f(0.404695272f, 0.0791561604f, -1.f),
    C_Vertex3f(0.41689992f, 0.000384807587f, -1.f),
    C_Vertex3f(0.419662118f, 0.0857752562f, -1.f),
    C_Vertex3f(0.434498191f, 0.00594830513f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.214595795f, 0.0926685333f, -1.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Vertex3f(0.215368271f, 0.0037971735f, -1.f),
    C_Vertex3f(0.233478546f, 0.0819683075f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Vertex3f(0.174647331f, 0.13835907f, -1.f),
    C_Vertex3f(0.183670044f, 0.016397953f, -1.f),
    C_Vertex3f(0.185180187f, 0.120863914f, -1.f),
    C_Vertex3f(0.199007034f, 0.0096129179f, -1.f),
    C_Vertex3f(0.198496342f, 0.105633736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1693573f, 0.0241522789f, -1.f),
    C_Vertex3f(0.168835163f, 0.153179169f, -1.f),
    C_Vertex3f(0.165444732f, 0.164430857f, -1.f),
    C_Vertex3f(0.162538648f, 0.177054167f, -1.f),
    C_Vertex3f(0.160116911f, 0.191049099f, -1.f),
    C_Vertex3f(0.158179522f, 0.206415653f, -1.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.156068802f, 0.0327734947f, -1.f),
    C_Vertex3f(0.15672648f, 0.22315383f, -1.f),
    C_Vertex3f(0.155757785f, 0.241263628f, -1.f),
    C_Vertex3f(0.155273438f, 0.260745049f, -1.f),
    C_Vertex3f(0.14380455f, 0.0422616005f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.198730469f, 0.76953125f, -1.f),
    C_Vertex3f(0.198730469f, 0.680175781f, -1.f),
    C_Vertex3f(0.280761719f, 0.680175781f, -1.f),
    C_Vertex3f(0.280761719f, 0.76953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361816406f, 0.76953125f, -1.f),
    C_Vertex3f(0.361816406f, 0.680175781f, -1.f),
    C_Vertex3f(0.443847656f, 0.680175781f, -1.f),
    C_Vertex3f(0.443847656f, 0.76953125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0710773468f, 0.233772397f, -1.f),
    C_Vertex3f(0.0703125f, 0.640625f, -1.f),
    C_Vertex3f(0.0703125f, 0.257875443f, -1.f),
    C_End(),
    C_Translatef(0.646484375f, 0.f, 0.f),
};
static const Call char_221[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.249542236f, 0.f, 0.f),
    C_Vertex3f(0.249542236f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Normal3f(-0.831113875f, -0.556102216f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.00247192383f, 0.640625f, -1.f),
    C_Vertex3f(0.00247192383f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.00247192383f, 0.640625f, 0.f),
    C_Vertex3f(0.00247192383f, 0.640625f, -1.f),
    C_Vertex3f(0.105628967f, 0.640625f, -1.f),
    C_Vertex3f(0.105628967f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.836778283f, 0.547541976f, 0.f),
    C_Vertex3f(0.105628967f, 0.640625f, 0.f),
    C_Vertex3f(0.105628967f, 0.640625f, -1.f),
    C_Normal3f(0.840595007f, 0.541664064f, 0.f),
    C_Vertex3f(0.249668121f, 0.420497894f, 0.f),
    C_Vertex3f(0.249668121f, 0.420497894f, -1.f),
    C_Normal3f(0.862715006f, 0.505690515f, 0.f),
    C_Vertex3f(0.282409668f, 0.366283417f, 0.f),
    C_Vertex3f(0.282409668f, 0.366283417f, -1.f),
    C_Normal3f(0.875902712f, 0.482487768f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.86717093f, 0.498010665f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_Normal3f(-0.86124754f, 0.508185625f, 0.f),
    C_Vertex3f(0.304641724f, 0.351566315f, 0.f),
    C_Vertex3f(0.304641724f, 0.351566315f, -1.f),
    C_Normal3f(-0.854441464f, 0.519547701f, 0.f),
    C_Vertex3f(0.320591927f, 0.37826252f, 0.f),
    C_Vertex3f(0.320591927f, 0.37826252f, -1.f),
    C_Normal3f(-0.847048819f, 0.531515121f, 0.f),
    C_Vertex3f(0.338072777f, 0.406546593f, 0.f),
    C_Vertex3f(0.338072777f, 0.406546593f, -1.f),
    C_Normal3f(-0.835170507f, 0.549991071f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.835170507f, 0.549991071f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, -1.f),
    C_Vertex3f(0.49156189f, 0.640625f, -1.f),
    C_Vertex3f(0.49156189f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.49156189f, 0.640625f, 0.f),
    C_Vertex3f(0.49156189f, 0.640625f, -1.f),
    C_Vertex3f(0.590362549f, 0.640625f, -1.f),
    C_Vertex3f(0.590362549f, 0.640625f, 0.f),
    C_Normal3f(0.821961641f, -0.569542885f, 0.f),
    C_Vertex3f(0.590362549f, 0.640625f, 0.f),
    C_Vertex3f(0.590362549f, 0.640625f, -1.f),
    C_Vertex3f(0.334503174f, 0.271369934f, -1.f),
    C_Vertex3f(0.334503174f, 0.271369934f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.334503174f, 0.271369934f, 0.f),
    C_Vertex3f(0.334503174f, 0.271369934f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.243652344f, 0.680175781f, 0.f),
    C_Vertex3f(0.243652344f, 0.680175781f, -1.f),
    C_Vertex3f(0.301269531f, 0.802734375f, -1.f),
    C_Vertex3f(0.301269531f, 0.802734375f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.301269531f, 0.802734375f, 0.f),
    C_Vertex3f(0.301269531f, 0.802734375f, -1.f),
    C_Vertex3f(0.404785156f, 0.802734375f, -1.f),
    C_Vertex3f(0.404785156f, 0.802734375f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.404785156f, 0.802734375f, 0.f),
    C_Vertex3f(0.404785156f, 0.802734375f, -1.f),
    C_Vertex3f(0.30859375f, 0.680175781f, -1.f),
    C_Vertex3f(0.30859375f, 0.680175781f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.30859375f, 0.680175781f, 0.f),
    C_Vertex3f(0.30859375f, 0.680175781f, -1.f),
    C_Vertex3f(0.243652344f, 0.680175781f, -1.f),
    C_Vertex3f(0.243652344f, 0.680175781f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_Vertex3f(0.282409668f, 0.366283417f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.249668121f, 0.420497894f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Vertex3f(0.105628967f, 0.640625f, 0.f),
    C_Vertex3f(0.00247192383f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.49156189f, 0.640625f, 0.f),
    C_Vertex3f(0.357084274f, 0.436418533f, 0.f),
    C_Vertex3f(0.590362549f, 0.640625f, 0.f),
    C_Vertex3f(0.338072777f, 0.406546593f, 0.f),
    C_Vertex3f(0.334503174f, 0.271369934f, 0.f),
    C_Vertex3f(0.320591927f, 0.37826252f, 0.f),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.304641724f, 0.351566315f, 0.f),
    C_Vertex3f(0.297431946f, 0.339012146f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.301269531f, 0.802734375f, 0.f),
    C_Vertex3f(0.243652344f, 0.680175781f, 0.f),
    C_Vertex3f(0.30859375f, 0.680175781f, 0.f),
    C_Vertex3f(0.404785156f, 0.802734375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.334503174f, 0.f, 0.f),
    C_Vertex3f(0.249542236f, 0.271369934f, 0.f),
    C_Vertex3f(0.249542236f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_Vertex3f(0.282409668f, 0.366283417f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.249668121f, 0.420497894f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.105628967f, 0.640625f, -1.f),
    C_Vertex3f(0.00247192383f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.49156189f, 0.640625f, -1.f),
    C_Vertex3f(0.357084274f, 0.436418533f, -1.f),
    C_Vertex3f(0.590362549f, 0.640625f, -1.f),
    C_Vertex3f(0.338072777f, 0.406546593f, -1.f),
    C_Vertex3f(0.334503174f, 0.271369934f, -1.f),
    C_Vertex3f(0.320591927f, 0.37826252f, -1.f),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.304641724f, 0.351566315f, -1.f),
    C_Vertex3f(0.297431946f, 0.339012146f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.301269531f, 0.802734375f, -1.f),
    C_Vertex3f(0.243652344f, 0.680175781f, -1.f),
    C_Vertex3f(0.30859375f, 0.680175781f, -1.f),
    C_Vertex3f(0.404785156f, 0.802734375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.334503174f, 0.f, -1.f),
    C_Vertex3f(0.249542236f, 0.271369934f, -1.f),
    C_Vertex3f(0.249542236f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_222[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Vertex3f(0.153808594f, 0.640625f, -1.f),
    C_Vertex3f(0.153808594f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.640625f, 0.f),
    C_Vertex3f(0.153808594f, 0.640625f, -1.f),
    C_Vertex3f(0.153808594f, 0.510253906f, -1.f),
    C_Vertex3f(0.153808594f, 0.510253906f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.153808594f, 0.510253906f, 0.f),
    C_Vertex3f(0.153808594f, 0.510253906f, -1.f),
    C_Normal3f(0.00380395097f, 0.999992728f, 0.f),
    C_Vertex3f(0.326465607f, 0.510253906f, 0.f),
    C_Vertex3f(0.326465607f, 0.510253906f, -1.f),
    C_Normal3f(0.0435236804f, 0.999052346f, 0.f),
    C_Vertex3f(0.354622841f, 0.509490013f, 0.f),
    C_Vertex3f(0.354622841f, 0.509490013f, -1.f),
    C_Normal3f(0.08423426f, 0.996446073f, 0.f),
    C_Vertex3f(0.379069328f, 0.507962227f, 0.f),
    C_Vertex3f(0.379069328f, 0.507962227f, -1.f),
    C_Normal3f(0.142187804f, 0.989839673f, 0.f),
    C_Vertex3f(0.399805069f, 0.505670547f, 0.f),
    C_Vertex3f(0.399805069f, 0.505670547f, -1.f),
    C_Normal3f(0.220173106f, 0.975460827f, 0.f),
    C_Vertex3f(0.42010498f, 0.502067566f, 0.f),
    C_Vertex3f(0.42010498f, 0.502067566f, -1.f),
    C_Normal3f(0.308985561f, 0.951066732f, 0.f),
    C_Vertex3f(0.4417696f, 0.496198654f, 0.f),
    C_Vertex3f(0.4417696f, 0.496198654f, -1.f),
    C_Normal3f(0.409840256f, 0.912157178f, 0.f),
    C_Vertex3f(0.46152401f, 0.488611221f, 0.f),
    C_Vertex3f(0.46152401f, 0.488611221f, -1.f),
    C_Normal3f(0.515679836f, 0.856781363f, 0.f),
    C_Vertex3f(0.47936821f, 0.479305267f, 0.f),
    C_Vertex3f(0.47936821f, 0.479305267f, -1.f),
    C_Normal3f(0.625186443f, 0.780475438f, 0.f),
    C_Vertex3f(0.4953022f, 0.468280792f, 0.f),
    C_Vertex3f(0.4953022f, 0.468280792f, -1.f),
    C_Normal3f(0.726599574f, 0.687061071f, 0.f),
    C_Vertex3f(0.509653568f, 0.4550457f, 0.f),
    C_Vertex3f(0.509653568f, 0.4550457f, -1.f),
    C_Normal3f(0.810123086f, 0.586259782f, 0.f),
    C_Vertex3f(0.522422314f, 0.439599991f, 0.f),
    C_Vertex3f(0.522422314f, 0.439599991f, -1.f),
    C_Normal3f(0.874717236f, 0.484633714f, 0.f),
    C_Vertex3f(0.533608437f, 0.421943665f, 0.f),
    C_Vertex3f(0.533608437f, 0.421943665f, -1.f),
    C_Normal3f(0.924356103f, 0.381531f, 0.f),
    C_Vertex3f(0.543211937f, 0.402076721f, 0.f),
    C_Vertex3f(0.543211937f, 0.402076721f, -1.f),
    C_Normal3f(0.96265316f, 0.270737469f, 0.f),
    C_Vertex3f(0.550414562f, 0.38122654f, 0.f),
    C_Vertex3f(0.550414562f, 0.38122654f, -1.f),
    C_Normal3f(0.987237632f, 0.159253985f, 0.f),
    C_Vertex3f(0.555216312f, 0.35939312f, 0.f),
    C_Vertex3f(0.555216312f, 0.35939312f, -1.f),
    C_Normal3f(0.998571455f, 0.0534327403f, 0.f),
    C_Vertex3f(0.557617188f, 0.336576462f, 0.f),
    C_Vertex3f(0.557617188f, 0.336576462f, -1.f),
    C_Normal3f(0.999238908f, -0.0390085764f, 0.f),
    C_Vertex3f(0.557617188f, 0.314524651f, 0.f),
    C_Vertex3f(0.557617188f, 0.314524651f, -1.f),
    C_Normal3f(0.992009878f, -0.12616019f, 0.f),
    C_Vertex3f(0.555993795f, 0.294991851f, 0.f),
    C_Vertex3f(0.555993795f, 0.294991851f, -1.f),
    C_Normal3f(0.976469398f, -0.215655938f, 0.f),
    C_Vertex3f(0.552747011f, 0.276229978f, 0.f),
    C_Vertex3f(0.552747011f, 0.276229978f, -1.f),
    C_Normal3f(0.951666236f, -0.307134211f, 0.f),
    C_Vertex3f(0.547876835f, 0.258239031f, 0.f),
    C_Vertex3f(0.547876835f, 0.258239031f, -1.f),
    C_Normal3f(0.917350888f, -0.398079425f, 0.f),
    C_Vertex3f(0.541383266f, 0.241019011f, 0.f),
    C_Vertex3f(0.541383266f, 0.241019011f, -1.f),
    C_Normal3f(0.874055207f, -0.485826582f, 0.f),
    C_Vertex3f(0.533266306f, 0.224569917f, 0.f),
    C_Vertex3f(0.533266306f, 0.224569917f, -1.f),
    C_Normal3f(0.823078036f, -0.567928255f, 0.f),
    C_Vertex3f(0.523525953f, 0.208891749f, 0.f),
    C_Vertex3f(0.523525953f, 0.208891749f, -1.f),
    C_Normal3f(0.766298413f, -0.642484784f, 0.f),
    C_Vertex3f(0.512162209f, 0.193984509f, 0.f),
    C_Vertex3f(0.512162209f, 0.193984509f, -1.f),
    C_Normal3f(0.679823399f, -0.733375788f, 0.f),
    C_Vertex3f(0.499175072f, 0.179848194f, 0.f),
    C_Vertex3f(0.499175072f, 0.179848194f, -1.f),
    C_Normal3f(0.561736345f, -0.827316284f, 0.f),
    C_Vertex3f(0.483568668f, 0.167478919f, 0.f),
    C_Vertex3f(0.483568668f, 0.167478919f, -1.f),
    C_Normal3f(0.445418507f, -0.895322502f, 0.f),
    C_Vertex3f(0.465342999f, 0.156876683f, 0.f),
    C_Vertex3f(0.465342999f, 0.156876683f, -1.f),
    C_Normal3f(0.337817878f, -0.941211462f, 0.f),
    C_Vertex3f(0.444498062f, 0.148041487f, 0.f),
    C_Vertex3f(0.444498062f, 0.148041487f, -1.f),
    C_Normal3f(0.242210463f, -0.970223725f, 0.f),
    C_Vertex3f(0.421033859f, 0.14097333f, 0.f),
    C_Vertex3f(0.421033859f, 0.14097333f, -1.f),
    C_Normal3f(0.159209818f, -0.987244844f, 0.f),
    C_Vertex3f(0.39495039f, 0.135672212f, 0.f),
    C_Vertex3f(0.39495039f, 0.135672212f, -1.f),
    C_Normal3f(0.0879731402f, -0.996122837f, 0.f),
    C_Vertex3f(0.366247654f, 0.132138133f, 0.f),
    C_Vertex3f(0.366247654f, 0.132138133f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.334925652f, 0.130371094f, 0.f),
    C_Vertex3f(0.334925652f, 0.130371094f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.334925652f, 0.130371094f, 0.f),
    C_Vertex3f(0.334925652f, 0.130371094f, -1.f),
    C_Vertex3f(0.153808594f, 0.130371094f, -1.f),
    C_Vertex3f(0.153808594f, 0.130371094f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.130371094f, 0.f),
    C_Vertex3f(0.153808594f, 0.130371094f, -1.f),
    C_Vertex3f(0.153808594f, 0.f, -1.f),
    C_Vertex3f(0.153808594f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.153808594f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.153808594f, 0.206054688f, 0.f),
    C_Vertex3f(0.153808594f, 0.206054688f, -1.f),
    C_Normal3f(-0.00489968667f, 0.99998796f, 0.f),
    C_Vertex3f(0.329469681f, 0.206054688f, 0.f),
    C_Vertex3f(0.329469681f, 0.206054688f, -1.f),
    C_Normal3f(-0.0780287087f, 0.996951163f, 0.f),
    C_Vertex3f(0.348494768f, 0.2070086f, 0.f),
    C_Vertex3f(0.348494768f, 0.2070086f, -1.f),
    C_Normal3f(-0.140581116f, 0.990069151f, 0.f),
    C_Vertex3f(0.366033316f, 0.208916426f, 0.f),
    C_Vertex3f(0.366033316f, 0.208916426f, -1.f),
    C_Normal3f(-0.213082165f, 0.977034211f, 0.f),
    C_Vertex3f(0.382085323f, 0.211778164f, 0.f),
    C_Vertex3f(0.382085323f, 0.211778164f, -1.f),
    C_Normal3f(-0.296585709f, 0.955006182f, 0.f),
    C_Vertex3f(0.396650791f, 0.215593815f, 0.f),
    C_Vertex3f(0.396650791f, 0.215593815f, -1.f),
    C_Normal3f(-0.39138487f, 0.92022711f, 0.f),
    C_Vertex3f(0.409729719f, 0.220363379f, 0.f),
    C_Vertex3f(0.409729719f, 0.220363379f, -1.f),
    C_Normal3f(-0.496195406f, 0.868210793f, 0.f),
    C_Vertex3f(0.421322107f, 0.226086855f, 0.f),
    C_Vertex3f(0.421322107f, 0.226086855f, -1.f),
    C_Normal3f(-0.607167661f, 0.794573724f, 0.f),
    C_Vertex3f(0.431427956f, 0.232764244f, 0.f),
    C_Vertex3f(0.431427956f, 0.232764244f, -1.f),
    C_Normal3f(-0.704130828f, 0.710070372f, 0.f),
    C_Vertex3f(0.440047264f, 0.240395546f, 0.f),
    C_Vertex3f(0.440047264f, 0.240395546f, -1.f),
    C_Normal3f(-0.781648219f, 0.623719633f, 0.f),
    C_Vertex3f(0.447589159f, 0.248790264f, 0.f),
    C_Vertex3f(0.447589159f, 0.248790264f, -1.f),
    C_Normal3f(-0.84946084f, 0.527651608f, 0.f),
    C_Vertex3f(0.45405364f, 0.257948399f, 0.f),
    C_Vertex3f(0.45405364f, 0.257948399f, -1.f),
    C_Normal3f(-0.904826522f, 0.425780386f, 0.f),
    C_Vertex3f(0.459440708f, 0.267869949f, 0.f),
    C_Vertex3f(0.459440708f, 0.267869949f, -1.f),
    C_Normal3f(-0.946556687f, 0.322537303f, 0.f),
    C_Vertex3f(0.463750362f, 0.278554916f, 0.f),
    C_Vertex3f(0.463750362f, 0.278554916f, -1.f),
    C_Normal3f(-0.975045979f, 0.222003296f, 0.f),
    C_Vertex3f(0.466982603f, 0.2900033f, 0.f),
    C_Vertex3f(0.466982603f, 0.2900033f, -1.f),
    C_Normal3f(-0.991866112f, 0.127285823f, 0.f),
    C_Vertex3f(0.46913743f, 0.302215099f, 0.f),
    C_Vertex3f(0.46913743f, 0.302215099f, -1.f),
    C_Normal3f(-0.999345481f, 0.0361740626f, 0.f),
    C_Vertex3f(0.470214844f, 0.315190315f, 0.f),
    C_Vertex3f(0.470214844f, 0.315190315f, -1.f),
    C_Normal3f(-0.997477829f, -0.0709796473f, 0.f),
    C_Vertex3f(0.470214844f, 0.331979752f, 0.f),
    C_Vertex3f(0.470214844f, 0.331979752f, -1.f),
    C_Normal3f(-0.978353083f, -0.206942424f, 0.f),
    C_Vertex3f(0.46770525f, 0.350457668f, 0.f),
    C_Vertex3f(0.46770525f, 0.350457668f, -1.f),
    C_Normal3f(-0.934236944f, -0.356652975f, 0.f),
    C_Vertex3f(0.462686062f, 0.367573261f, 0.f),
    C_Vertex3f(0.462686062f, 0.367573261f, -1.f),
    C_Normal3f(-0.863990545f, -0.503508151f, 0.f),
    C_Vertex3f(0.45515728f, 0.38332653f, 0.f),
    C_Vertex3f(0.45515728f, 0.38332653f, -1.f),
    C_Normal3f(-0.771503747f, -0.636224747f, 0.f),
    C_Vertex3f(0.445118904f, 0.397717476f, 0.f),
    C_Vertex3f(0.445118904f, 0.397717476f, -1.f),
    C_Normal3f(-0.650952399f, -0.759118497f, 0.f),
    C_Vertex3f(0.4334445f, 0.409656048f, 0.f),
    C_Vertex3f(0.4334445f, 0.409656048f, -1.f),
    C_Normal3f(-0.504710436f, -0.863288641f, 0.f),
    C_Vertex3f(0.420134068f, 0.419142246f, 0.f),
    C_Vertex3f(0.420134068f, 0.419142246f, -1.f),
    C_Normal3f(-0.362791359f, -0.931870461f, 0.f),
    C_Vertex3f(0.405187607f, 0.426176071f, 0.f),
    C_Vertex3f(0.405187607f, 0.426176071f, -1.f),
    C_Normal3f(-0.243828818f, -0.969818234f, 0.f),
    C_Vertex3f(0.394223213f, 0.429229736f, 0.f),
    C_Vertex3f(0.394223213f, 0.429229736f, -1.f),
    C_Normal3f(-0.172847569f, -0.984948635f, 0.f),
    C_Vertex3f(0.387731314f, 0.43056488f, 0.f),
    C_Vertex3f(0.387731314f, 0.43056488f, -1.f),
    C_Normal3f(-0.126741052f, -0.991935909f, 0.f),
    C_Vertex3f(0.380093813f, 0.43170929f, 0.f),
    C_Vertex3f(0.380093813f, 0.43170929f, -1.f),
    C_Normal3f(-0.0913559869f, -0.995818317f, 0.f),
    C_Vertex3f(0.371310711f, 0.432662964f, 0.f),
    C_Vertex3f(0.371310711f, 0.432662964f, -1.f),
    C_Normal3f(-0.0634411275f, -0.997985601f, 0.f),
    C_Vertex3f(0.361382008f, 0.433425903f, 0.f),
    C_Vertex3f(0.361382008f, 0.433425903f, -1.f),
    C_Normal3f(-0.0409061313f, -0.999162972f, 0.f),
    C_Vertex3f(0.350307703f, 0.433998108f, 0.f),
    C_Vertex3f(0.350307703f, 0.433998108f, -1.f),
    C_Normal3f(-0.0223588943f, -0.999750018f, 0.f),
    C_Vertex3f(0.338087797f, 0.434379578f, 0.f),
    C_Vertex3f(0.338087797f, 0.434379578f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.32472229f, 0.434570313f, 0.f),
    C_Vertex3f(0.32472229f, 0.434570313f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.32472229f, 0.434570313f, 0.f),
    C_Vertex3f(0.32472229f, 0.434570313f, -1.f),
    C_Vertex3f(0.153808594f, 0.434570313f, -1.f),
    C_Vertex3f(0.153808594f, 0.434570313f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.434570313f, 0.f),
    C_Vertex3f(0.153808594f, 0.434570313f, -1.f),
    C_Vertex3f(0.153808594f, 0.206054688f, -1.f),
    C_Vertex3f(0.153808594f, 0.206054688f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153808594f, 0.130371094f, 0.f),
    C_Vertex3f(0.153808594f, 0.206054688f, 0.f),
    C_Vertex3f(0.153808594f, 0.640625f, 0.f),
    C_Vertex3f(0.153808594f, 0.434570313f, 0.f),
    C_Vertex3f(0.153808594f, 0.510253906f, 0.f),
    C_Vertex3f(0.32472229f, 0.434570313f, 0.f),
    C_Vertex3f(0.326465607f, 0.510253906f, 0.f),
    C_Vertex3f(0.338087797f, 0.434379578f, 0.f),
    C_Vertex3f(0.354622841f, 0.509490013f, 0.f),
    C_Vertex3f(0.350307703f, 0.433998108f, 0.f),
    C_Vertex3f(0.361382008f, 0.433425903f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.153808594f, 0.f, 0.f),
    C_Vertex3f(0.153808594f, 0.130371094f, 0.f),
    C_Vertex3f(0.153808594f, 0.640625f, 0.f),
    C_Vertex3f(0.0688476563f, 0.640625f, 0.f),
    C_Vertex3f(0.0688476563f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.46770525f, 0.350457668f, 0.f),
    C_Vertex3f(0.470214844f, 0.331979752f, 0.f),
    C_Vertex3f(0.47936821f, 0.479305267f, 0.f),
    C_Vertex3f(0.483568668f, 0.167478919f, 0.f),
    C_Vertex3f(0.4953022f, 0.468280792f, 0.f),
    C_Vertex3f(0.499175072f, 0.179848194f, 0.f),
    C_Vertex3f(0.509653568f, 0.4550457f, 0.f),
    C_Vertex3f(0.512162209f, 0.193984509f, 0.f),
    C_Vertex3f(0.522422314f, 0.439599991f, 0.f),
    C_Vertex3f(0.523525953f, 0.208891749f, 0.f),
    C_Vertex3f(0.533608437f, 0.421943665f, 0.f),
    C_Vertex3f(0.533266306f, 0.224569917f, 0.f),
    C_Vertex3f(0.541383266f, 0.241019011f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445118904f, 0.397717476f, 0.f),
    C_Vertex3f(0.45515728f, 0.38332653f, 0.f),
    C_Vertex3f(0.46152401f, 0.488611221f, 0.f),
    C_Vertex3f(0.462686062f, 0.367573261f, 0.f),
    C_Vertex3f(0.47936821f, 0.479305267f, 0.f),
    C_Vertex3f(0.46770525f, 0.350457668f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.4417696f, 0.496198654f, 0.f),
    C_Vertex3f(0.42010498f, 0.502067566f, 0.f),
    C_Vertex3f(0.420134068f, 0.419142246f, 0.f),
    C_Vertex3f(0.4334445f, 0.409656048f, 0.f),
    C_Vertex3f(0.445118904f, 0.397717476f, 0.f),
    C_Vertex3f(0.46152401f, 0.488611221f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.387731314f, 0.43056488f, 0.f),
    C_Vertex3f(0.394223213f, 0.429229736f, 0.f),
    C_Vertex3f(0.399805069f, 0.505670547f, 0.f),
    C_Vertex3f(0.405187607f, 0.426176071f, 0.f),
    C_Vertex3f(0.42010498f, 0.502067566f, 0.f),
    C_Vertex3f(0.420134068f, 0.419142246f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.361382008f, 0.433425903f, 0.f),
    C_Vertex3f(0.371310711f, 0.432662964f, 0.f),
    C_Vertex3f(0.379069328f, 0.507962227f, 0.f),
    C_Vertex3f(0.380093813f, 0.43170929f, 0.f),
    C_Vertex3f(0.399805069f, 0.505670547f, 0.f),
    C_Vertex3f(0.387731314f, 0.43056488f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.533608437f, 0.421943665f, 0.f),
    C_Vertex3f(0.541383266f, 0.241019011f, 0.f),
    C_Vertex3f(0.543211937f, 0.402076721f, 0.f),
    C_Vertex3f(0.547876835f, 0.258239031f, 0.f),
    C_Vertex3f(0.550414562f, 0.38122654f, 0.f),
    C_Vertex3f(0.552747011f, 0.276229978f, 0.f),
    C_Vertex3f(0.555216312f, 0.35939312f, 0.f),
    C_Vertex3f(0.555993795f, 0.294991851f, 0.f),
    C_Vertex3f(0.557617188f, 0.336576462f, 0.f),
    C_Vertex3f(0.557617188f, 0.314524651f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.483568668f, 0.167478919f, 0.f),
    C_Vertex3f(0.470214844f, 0.331979752f, 0.f),
    C_Vertex3f(0.470214844f, 0.315190315f, 0.f),
    C_Vertex3f(0.46913743f, 0.302215099f, 0.f),
    C_Vertex3f(0.466982603f, 0.2900033f, 0.f),
    C_Vertex3f(0.465342999f, 0.156876683f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.465342999f, 0.156876683f, 0.f),
    C_Vertex3f(0.466982603f, 0.2900033f, 0.f),
    C_Vertex3f(0.463750362f, 0.278554916f, 0.f),
    C_Vertex3f(0.459440708f, 0.267869949f, 0.f),
    C_Vertex3f(0.45405364f, 0.257948399f, 0.f),
    C_Vertex3f(0.447589159f, 0.248790264f, 0.f),
    C_Vertex3f(0.444498062f, 0.148041487f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.444498062f, 0.148041487f, 0.f),
    C_Vertex3f(0.447589159f, 0.248790264f, 0.f),
    C_Vertex3f(0.440047264f, 0.240395546f, 0.f),
    C_Vertex3f(0.431427956f, 0.232764244f, 0.f),
    C_Vertex3f(0.421322107f, 0.226086855f, 0.f),
    C_Vertex3f(0.421033859f, 0.14097333f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.421322107f, 0.226086855f, 0.f),
    C_Vertex3f(0.409729719f, 0.220363379f, 0.f),
    C_Vertex3f(0.421033859f, 0.14097333f, 0.f),
    C_Vertex3f(0.396650791f, 0.215593815f, 0.f),
    C_Vertex3f(0.39495039f, 0.135672212f, 0.f),
    C_Vertex3f(0.382085323f, 0.211778164f, 0.f),
    C_Vertex3f(0.366247654f, 0.132138133f, 0.f),
    C_Vertex3f(0.366033316f, 0.208916426f, 0.f),
    C_Vertex3f(0.348494768f, 0.2070086f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.334925652f, 0.130371094f, 0.f),
    C_Vertex3f(0.366247654f, 0.132138133f, 0.f),
    C_Vertex3f(0.348494768f, 0.2070086f, 0.f),
    C_Vertex3f(0.329469681f, 0.206054688f, 0.f),
    C_Vertex3f(0.153808594f, 0.206054688f, 0.f),
    C_Vertex3f(0.153808594f, 0.130371094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.354622841f, 0.509490013f, 0.f),
    C_Vertex3f(0.361382008f, 0.433425903f, 0.f),
    C_Vertex3f(0.379069328f, 0.507962227f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153808594f, 0.130371094f, -1.f),
    C_Vertex3f(0.153808594f, 0.206054688f, -1.f),
    C_Vertex3f(0.153808594f, 0.640625f, -1.f),
    C_Vertex3f(0.153808594f, 0.434570313f, -1.f),
    C_Vertex3f(0.153808594f, 0.510253906f, -1.f),
    C_Vertex3f(0.32472229f, 0.434570313f, -1.f),
    C_Vertex3f(0.326465607f, 0.510253906f, -1.f),
    C_Vertex3f(0.338087797f, 0.434379578f, -1.f),
    C_Vertex3f(0.354622841f, 0.509490013f, -1.f),
    C_Vertex3f(0.350307703f, 0.433998108f, -1.f),
    C_Vertex3f(0.361382008f, 0.433425903f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.153808594f, 0.f, -1.f),
    C_Vertex3f(0.153808594f, 0.130371094f, -1.f),
    C_Vertex3f(0.153808594f, 0.640625f, -1.f),
    C_Vertex3f(0.0688476563f, 0.640625f, -1.f),
    C_Vertex3f(0.0688476563f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.46770525f, 0.350457668f, -1.f),
    C_Vertex3f(0.470214844f, 0.331979752f, -1.f),
    C_Vertex3f(0.47936821f, 0.479305267f, -1.f),
    C_Vertex3f(0.483568668f, 0.167478919f, -1.f),
    C_Vertex3f(0.4953022f, 0.468280792f, -1.f),
    C_Vertex3f(0.499175072f, 0.179848194f, -1.f),
    C_Vertex3f(0.509653568f, 0.4550457f, -1.f),
    C_Vertex3f(0.512162209f, 0.193984509f, -1.f),
    C_Vertex3f(0.522422314f, 0.439599991f, -1.f),
    C_Vertex3f(0.523525953f, 0.208891749f, -1.f),
    C_Vertex3f(0.533608437f, 0.421943665f, -1.f),
    C_Vertex3f(0.533266306f, 0.224569917f, -1.f),
    C_Vertex3f(0.541383266f, 0.241019011f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.445118904f, 0.397717476f, -1.f),
    C_Vertex3f(0.45515728f, 0.38332653f, -1.f),
    C_Vertex3f(0.46152401f, 0.488611221f, -1.f),
    C_Vertex3f(0.462686062f, 0.367573261f, -1.f),
    C_Vertex3f(0.47936821f, 0.479305267f, -1.f),
    C_Vertex3f(0.46770525f, 0.350457668f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.4417696f, 0.496198654f, -1.f),
    C_Vertex3f(0.42010498f, 0.502067566f, -1.f),
    C_Vertex3f(0.420134068f, 0.419142246f, -1.f),
    C_Vertex3f(0.4334445f, 0.409656048f, -1.f),
    C_Vertex3f(0.445118904f, 0.397717476f, -1.f),
    C_Vertex3f(0.46152401f, 0.488611221f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.387731314f, 0.43056488f, -1.f),
    C_Vertex3f(0.394223213f, 0.429229736f, -1.f),
    C_Vertex3f(0.399805069f, 0.505670547f, -1.f),
    C_Vertex3f(0.405187607f, 0.426176071f, -1.f),
    C_Vertex3f(0.42010498f, 0.502067566f, -1.f),
    C_Vertex3f(0.420134068f, 0.419142246f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.361382008f, 0.433425903f, -1.f),
    C_Vertex3f(0.371310711f, 0.432662964f, -1.f),
    C_Vertex3f(0.379069328f, 0.507962227f, -1.f),
    C_Vertex3f(0.380093813f, 0.43170929f, -1.f),
    C_Vertex3f(0.399805069f, 0.505670547f, -1.f),
    C_Vertex3f(0.387731314f, 0.43056488f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.533608437f, 0.421943665f, -1.f),
    C_Vertex3f(0.541383266f, 0.241019011f, -1.f),
    C_Vertex3f(0.543211937f, 0.402076721f, -1.f),
    C_Vertex3f(0.547876835f, 0.258239031f, -1.f),
    C_Vertex3f(0.550414562f, 0.38122654f, -1.f),
    C_Vertex3f(0.552747011f, 0.276229978f, -1.f),
    C_Vertex3f(0.555216312f, 0.35939312f, -1.f),
    C_Vertex3f(0.555993795f, 0.294991851f, -1.f),
    C_Vertex3f(0.557617188f, 0.336576462f, -1.f),
    C_Vertex3f(0.557617188f, 0.314524651f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.483568668f, 0.167478919f, -1.f),
    C_Vertex3f(0.470214844f, 0.331979752f, -1.f),
    C_Vertex3f(0.470214844f, 0.315190315f, -1.f),
    C_Vertex3f(0.46913743f, 0.302215099f, -1.f),
    C_Vertex3f(0.466982603f, 0.2900033f, -1.f),
    C_Vertex3f(0.465342999f, 0.156876683f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.465342999f, 0.156876683f, -1.f),
    C_Vertex3f(0.466982603f, 0.2900033f, -1.f),
    C_Vertex3f(0.463750362f, 0.278554916f, -1.f),
    C_Vertex3f(0.459440708f, 0.267869949f, -1.f),
    C_Vertex3f(0.45405364f, 0.257948399f, -1.f),
    C_Vertex3f(0.447589159f, 0.248790264f, -1.f),
    C_Vertex3f(0.444498062f, 0.148041487f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.444498062f, 0.148041487f, -1.f),
    C_Vertex3f(0.447589159f, 0.248790264f, -1.f),
    C_Vertex3f(0.440047264f, 0.240395546f, -1.f),
    C_Vertex3f(0.431427956f, 0.232764244f, -1.f),
    C_Vertex3f(0.421322107f, 0.226086855f, -1.f),
    C_Vertex3f(0.421033859f, 0.14097333f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.421322107f, 0.226086855f, -1.f),
    C_Vertex3f(0.409729719f, 0.220363379f, -1.f),
    C_Vertex3f(0.421033859f, 0.14097333f, -1.f),
    C_Vertex3f(0.396650791f, 0.215593815f, -1.f),
    C_Vertex3f(0.39495039f, 0.135672212f, -1.f),
    C_Vertex3f(0.382085323f, 0.211778164f, -1.f),
    C_Vertex3f(0.366247654f, 0.132138133f, -1.f),
    C_Vertex3f(0.366033316f, 0.208916426f, -1.f),
    C_Vertex3f(0.348494768f, 0.2070086f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.334925652f, 0.130371094f, -1.f),
    C_Vertex3f(0.366247654f, 0.132138133f, -1.f),
    C_Vertex3f(0.348494768f, 0.2070086f, -1.f),
    C_Vertex3f(0.329469681f, 0.206054688f, -1.f),
    C_Vertex3f(0.153808594f, 0.206054688f, -1.f),
    C_Vertex3f(0.153808594f, 0.130371094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.354622841f, 0.509490013f, -1.f),
    C_Vertex3f(0.361382008f, 0.433425903f, -1.f),
    C_Vertex3f(0.379069328f, 0.507962227f, -1.f),
    C_End(),
    C_Translatef(0.597167969f, 0.f, 0.f),
};
static const Call char_223[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0673828125f, 0.f, 0.f),
    C_Vertex3f(0.0673828125f, 0.f, -1.f),
    C_Normal3f(-0.999999225f, 0.00130494183f, 0.f),
    C_Vertex3f(0.0673828125f, 0.446424484f, 0.f),
    C_Vertex3f(0.0673828125f, 0.446424484f, -1.f),
    C_Normal3f(-0.998757958f, 0.0498252288f, 0.f),
    C_Vertex3f(0.0679899454f, 0.465256333f, 0.f),
    C_Vertex3f(0.0679899454f, 0.465256333f, -1.f),
    C_Normal3f(-0.99608463f, 0.0884051546f, 0.f),
    C_Vertex3f(0.0692042112f, 0.482934833f, 0.f),
    C_Vertex3f(0.0692042112f, 0.482934833f, -1.f),
    C_Normal3f(-0.991240025f, 0.132072225f, 0.f),
    C_Vertex3f(0.07102561f, 0.499459982f, 0.f),
    C_Vertex3f(0.07102561f, 0.499459982f, -1.f),
    C_Normal3f(-0.98337394f, 0.181591853f, 0.f),
    C_Vertex3f(0.0734541416f, 0.514831781f, 0.f),
    C_Vertex3f(0.0734541416f, 0.514831781f, -1.f),
    C_Normal3f(-0.971323848f, 0.237760469f, 0.f),
    C_Vertex3f(0.0764898062f, 0.529050231f, 0.f),
    C_Vertex3f(0.0764898062f, 0.529050231f, -1.f),
    C_Normal3f(-0.95352459f, 0.301315427f, 0.f),
    C_Vertex3f(0.0801326036f, 0.542115331f, 0.f),
    C_Vertex3f(0.0801326036f, 0.542115331f, -1.f),
    C_Normal3f(-0.927926183f, 0.372764081f, 0.f),
    C_Vertex3f(0.084382534f, 0.554027081f, 0.f),
    C_Vertex3f(0.084382534f, 0.554027081f, -1.f),
    C_Normal3f(-0.893442214f, 0.449178189f, 0.f),
    C_Vertex3f(0.0892395973f, 0.56478548f, 0.f),
    C_Vertex3f(0.0892395973f, 0.56478548f, -1.f),
    C_Normal3f(-0.852053344f, 0.523455143f, 0.f),
    C_Vertex3f(0.0949083567f, 0.57496357f, 0.f),
    C_Vertex3f(0.0949083567f, 0.57496357f, -1.f),
    C_Normal3f(-0.803897142f, 0.594768345f, 0.f),
    C_Vertex3f(0.101388812f, 0.584561348f, 0.f),
    C_Vertex3f(0.101388812f, 0.584561348f, -1.f),
    C_Normal3f(-0.749947369f, 0.661497414f, 0.f),
    C_Vertex3f(0.108680964f, 0.593578815f, 0.f),
    C_Vertex3f(0.108680964f, 0.593578815f, -1.f),
    C_Normal3f(-0.691545844f, 0.722332597f, 0.f),
    C_Vertex3f(0.116784811f, 0.602015972f, 0.f),
    C_Vertex3f(0.116784811f, 0.602015972f, -1.f),
    C_Normal3f(-0.630244315f, 0.776396751f, 0.f),
    C_Vertex3f(0.125700355f, 0.609872818f, 0.f),
    C_Vertex3f(0.125700355f, 0.609872818f, -1.f),
    C_Normal3f(-0.567625225f, 0.82328707f, 0.f),
    C_Vertex3f(0.135427594f, 0.617149353f, 0.f),
    C_Vertex3f(0.135427594f, 0.617149353f, -1.f),
    C_Normal3f(-0.498950362f, 0.866630554f, 0.f),
    C_Vertex3f(0.14596653f, 0.623845577f, 0.f),
    C_Vertex3f(0.14596653f, 0.623845577f, -1.f),
    C_Normal3f(-0.407267541f, 0.913308918f, 0.f),
    C_Vertex3f(0.162992477f, 0.633019447f, 0.f),
    C_Vertex3f(0.162992477f, 0.633019447f, -1.f),
    C_Normal3f(-0.294274807f, 0.955720901f, 0.f),
    C_Vertex3f(0.187111855f, 0.642193317f, 0.f),
    C_Vertex3f(0.187111855f, 0.642193317f, -1.f),
    C_Normal3f(-0.172154158f, 0.98507005f, 0.f),
    C_Vertex3f(0.212649345f, 0.648309231f, 0.f),
    C_Vertex3f(0.212649345f, 0.648309231f, -1.f),
    C_Normal3f(-0.0603442043f, 0.998177648f, 0.f),
    C_Vertex3f(0.23960495f, 0.651367188f, 0.f),
    C_Vertex3f(0.23960495f, 0.651367188f, -1.f),
    C_Normal3f(0.028421659f, 0.999596f, 0.f),
    C_Vertex3f(0.263232231f, 0.651367188f, 0.f),
    C_Vertex3f(0.263232231f, 0.651367188f, -1.f),
    C_Normal3f(0.101363599f, 0.994849443f, 0.f),
    C_Vertex3f(0.281107664f, 0.650187135f, 0.f),
    C_Vertex3f(0.281107664f, 0.650187135f, -1.f),
    C_Normal3f(0.177386835f, 0.984141231f, 0.f),
    C_Vertex3f(0.297977686f, 0.647827029f, 0.f),
    C_Vertex3f(0.297977686f, 0.647827029f, -1.f),
    C_Normal3f(0.259638608f, 0.965705872f, 0.f),
    C_Vertex3f(0.313842297f, 0.644286871f, 0.f),
    C_Vertex3f(0.313842297f, 0.644286871f, -1.f),
    C_Normal3f(0.346913159f, 0.937897205f, 0.f),
    C_Vertex3f(0.328701496f, 0.63956666f, 0.f),
    C_Vertex3f(0.328701496f, 0.63956666f, -1.f),
    C_Normal3f(0.437202543f, 0.899363041f, 0.f),
    C_Vertex3f(0.342555285f, 0.633666396f, 0.f),
    C_Vertex3f(0.342555285f, 0.633666396f, -1.f),
    C_Normal3f(0.527735889f, 0.849408448f, 0.f),
    C_Vertex3f(0.355403662f, 0.62658608f, 0.f),
    C_Vertex3f(0.355403662f, 0.62658608f, -1.f),
    C_Normal3f(0.615247548f, 0.788333952f, 0.f),
    C_Vertex3f(0.367246628f, 0.61832571f, 0.f),
    C_Vertex3f(0.367246628f, 0.61832571f, -1.f),
    C_Normal3f(0.687835753f, 0.725866437f, 0.f),
    C_Vertex3f(0.378084183f, 0.608885288f, 0.f),
    C_Vertex3f(0.378084183f, 0.608885288f, -1.f),
    C_Normal3f(0.750599742f, 0.660757124f, 0.f),
    C_Vertex3f(0.387567043f, 0.599069953f, 0.f),
    C_Vertex3f(0.387567043f, 0.599069953f, -1.f),
    C_Normal3f(0.812317789f, 0.583214998f, 0.f),
    C_Vertex3f(0.39569521f, 0.588879704f, 0.f),
    C_Vertex3f(0.39569521f, 0.588879704f, -1.f),
    C_Normal3f(0.869919002f, 0.493194729f, 0.f),
    C_Vertex3f(0.402468681f, 0.578314543f, 0.f),
    C_Vertex3f(0.402468681f, 0.578314543f, -1.f),
    C_Normal3f(0.919966578f, 0.391996861f, 0.f),
    C_Vertex3f(0.407887459f, 0.567374468f, 0.f),
    C_Vertex3f(0.407887459f, 0.567374468f, -1.f),
    C_Normal3f(0.959282756f, 0.282447547f, 0.f),
    C_Vertex3f(0.411951542f, 0.55605948f, 0.f),
    C_Vertex3f(0.411951542f, 0.55605948f, -1.f),
    C_Normal3f(0.985678554f, 0.16863513f, 0.f),
    C_Vertex3f(0.414660931f, 0.544369578f, 0.f),
    C_Vertex3f(0.414660931f, 0.544369578f, -1.f),
    C_Normal3f(0.998390973f, 0.0567039102f, 0.f),
    C_Vertex3f(0.416015625f, 0.532304764f, 0.f),
    C_Vertex3f(0.416015625f, 0.532304764f, -1.f),
    C_Normal3f(0.999091625f, -0.0426150858f, 0.f),
    C_Vertex3f(0.416015625f, 0.520517349f, 0.f),
    C_Vertex3f(0.416015625f, 0.520517349f, -1.f),
    C_Normal3f(0.991680861f, -0.128720909f, 0.f),
    C_Vertex3f(0.415030956f, 0.509219646f, 0.f),
    C_Vertex3f(0.415030956f, 0.509219646f, -1.f),
    C_Normal3f(0.978002131f, -0.208594933f, 0.f),
    C_Vertex3f(0.413061619f, 0.497759342f, 0.f),
    C_Vertex3f(0.413061619f, 0.497759342f, -1.f),
    C_Normal3f(0.959278286f, -0.282462507f, 0.f),
    C_Vertex3f(0.410107613f, 0.486136436f, 0.f),
    C_Vertex3f(0.410107613f, 0.486136436f, -1.f),
    C_Normal3f(0.93602097f, -0.351944298f, 0.f),
    C_Vertex3f(0.406168938f, 0.474350929f, 0.f),
    C_Vertex3f(0.406168938f, 0.474350929f, -1.f),
    C_Normal3f(0.913089335f, -0.407759458f, 0.f),
    C_Vertex3f(0.400699139f, 0.46111393f, 0.f),
    C_Vertex3f(0.400699139f, 0.46111393f, -1.f),
    C_Normal3f(0.893047154f, -0.449963033f, 0.f),
    C_Vertex3f(0.393698215f, 0.446425438f, 0.f),
    C_Vertex3f(0.393698215f, 0.446425438f, -1.f),
    C_Normal3f(0.875744998f, -0.48277393f, 0.f),
    C_Vertex3f(0.385166168f, 0.430285454f, 0.f),
    C_Vertex3f(0.385166168f, 0.430285454f, -1.f),
    C_Normal3f(0.871775627f, -0.489905447f, 0.f),
    C_Vertex3f(0.375102997f, 0.412693977f, 0.f),
    C_Vertex3f(0.375102997f, 0.412693977f, -1.f),
    C_Normal3f(0.88143909f, -0.472297907f, 0.f),
    C_Vertex3f(0.366899014f, 0.397779465f, 0.f),
    C_Vertex3f(0.366899014f, 0.397779465f, -1.f),
    C_Normal3f(0.895552695f, -0.444955349f, 0.f),
    C_Vertex3f(0.360554218f, 0.385541916f, 0.f),
    C_Vertex3f(0.360554218f, 0.385541916f, -1.f),
    C_Normal3f(0.920769095f, -0.390108109f, 0.f),
    C_Vertex3f(0.356068611f, 0.375981331f, 0.f),
    C_Vertex3f(0.356068611f, 0.375981331f, -1.f),
    C_Normal3f(0.96041882f, -0.278560191f, 0.f),
    C_Vertex3f(0.352128983f, 0.365655899f, 0.f),
    C_Vertex3f(0.352128983f, 0.365655899f, -1.f),
    C_Normal3f(0.989647448f, -0.143520057f, 0.f),
    C_Vertex3f(0.349555969f, 0.353527069f, 0.f),
    C_Vertex3f(0.349555969f, 0.353527069f, -1.f),
    C_Normal3f(0.999999702f, 0.000872790464f, 0.f),
    C_Vertex3f(0.349609375f, 0.34828186f, 0.f),
    C_Vertex3f(0.349609375f, 0.34828186f, -1.f),
    C_Normal3f(0.994606078f, 0.103724115f, 0.f),
    C_Vertex3f(0.349563599f, 0.34478569f, 0.f),
    C_Vertex3f(0.349563599f, 0.34478569f, -1.f),
    C_Normal3f(0.969513595f, 0.245037481f, 0.f),
    C_Vertex3f(0.350697517f, 0.33784771f, 0.f),
    C_Vertex3f(0.350697517f, 0.33784771f, -1.f),
    C_Normal3f(0.917506754f, 0.397720277f, 0.f),
    C_Vertex3f(0.353056908f, 0.330964088f, 0.f),
    C_Vertex3f(0.353056908f, 0.330964088f, -1.f),
    C_Normal3f(0.862664044f, 0.505777359f, 0.f),
    C_Vertex3f(0.356641769f, 0.324134827f, 0.f),
    C_Vertex3f(0.356641769f, 0.324134827f, -1.f),
    C_Normal3f(0.822412074f, 0.5688923f, 0.f),
    C_Vertex3f(0.359046936f, 0.320747375f, 0.f),
    C_Vertex3f(0.359046936f, 0.320747375f, -1.f),
    C_Normal3f(0.78344655f, 0.621459305f, 0.f),
    C_Vertex3f(0.360194206f, 0.31899929f, 0.f),
    C_Vertex3f(0.360194206f, 0.31899929f, -1.f),
    C_Normal3f(0.729886055f, 0.683568835f, 0.f),
    C_Vertex3f(0.363566399f, 0.315049887f, 0.f),
    C_Vertex3f(0.363566399f, 0.315049887f, -1.f),
    C_Normal3f(0.680207908f, 0.733019352f, 0.f),
    C_Vertex3f(0.368016243f, 0.310647249f, 0.f),
    C_Vertex3f(0.368016243f, 0.310647249f, -1.f),
    C_Normal3f(0.642211914f, 0.766526997f, 0.f),
    C_Vertex3f(0.373543739f, 0.305791378f, 0.f),
    C_Vertex3f(0.373543739f, 0.305791378f, -1.f),
    C_Normal3f(0.61251235f, 0.790461063f, 0.f),
    C_Vertex3f(0.380148888f, 0.300482273f, 0.f),
    C_Vertex3f(0.380148888f, 0.300482273f, -1.f),
    C_Normal3f(0.588788211f, 0.808287323f, 0.f),
    C_Vertex3f(0.387831688f, 0.294719934f, 0.f),
    C_Vertex3f(0.387831688f, 0.294719934f, -1.f),
    C_Normal3f(0.569462538f, 0.822017252f, 0.f),
    C_Vertex3f(0.39659214f, 0.288504362f, 0.f),
    C_Vertex3f(0.39659214f, 0.288504362f, -1.f),
    C_Normal3f(0.552829325f, 0.83329457f, 0.f),
    C_Vertex3f(0.406430244f, 0.281835556f, 0.f),
    C_Vertex3f(0.406430244f, 0.281835556f, -1.f),
    C_Normal3f(0.553775847f, 0.832665741f, 0.f),
    C_Vertex3f(0.419325829f, 0.273422241f, 0.f),
    C_Vertex3f(0.419325829f, 0.273422241f, -1.f),
    C_Normal3f(0.568545341f, 0.822651982f, 0.f),
    C_Vertex3f(0.432956696f, 0.264193773f, 0.f),
    C_Vertex3f(0.432956696f, 0.264193773f, -1.f),
    C_Normal3f(0.586306095f, 0.810089588f, 0.f),
    C_Vertex3f(0.445343018f, 0.255441427f, 0.f),
    C_Vertex3f(0.445343018f, 0.255441427f, -1.f),
    C_Normal3f(0.607156336f, 0.794582486f, 0.f),
    C_Vertex3f(0.456484795f, 0.247165203f, 0.f),
    C_Vertex3f(0.456484795f, 0.247165203f, -1.f),
    C_Normal3f(0.631907284f, 0.775043964f, 0.f),
    C_Vertex3f(0.466382027f, 0.239365101f, 0.f),
    C_Vertex3f(0.466382027f, 0.239365101f, -1.f),
    C_Normal3f(0.66163522f, 0.749825835f, 0.f),
    C_Vertex3f(0.475034714f, 0.232041121f, 0.f),
    C_Vertex3f(0.475034714f, 0.232041121f, -1.f),
    C_Normal3f(0.697752357f, 0.716339111f, 0.f),
    C_Vertex3f(0.482442856f, 0.225193262f, 0.f),
    C_Vertex3f(0.482442856f, 0.225193262f, -1.f),
    C_Normal3f(0.750275731f, 0.661125124f, 0.f),
    C_Vertex3f(0.488606453f, 0.218821526f, 0.f),
    C_Vertex3f(0.488606453f, 0.218821526f, -1.f),
    C_Normal3f(0.830076039f, 0.557650268f, 0.f),
    C_Vertex3f(0.498060226f, 0.20746994f, 0.f),
    C_Vertex3f(0.498060226f, 0.20746994f, -1.f),
    C_Normal3f(0.909286797f, 0.41617009f, 0.f),
    C_Vertex3f(0.508551598f, 0.18913269f, 0.f),
    C_Vertex3f(0.508551598f, 0.18913269f, -1.f),
    C_Normal3f(0.969165027f, 0.246412694f, 0.f),
    C_Vertex3f(0.515545845f, 0.169265747f, 0.f),
    C_Vertex3f(0.515545845f, 0.169265747f, -1.f),
    C_Normal3f(0.996282935f, 0.0861420259f, 0.f),
    C_Vertex3f(0.519042969f, 0.14786911f, 0.f),
    C_Vertex3f(0.519042969f, 0.14786911f, -1.f),
    C_Normal3f(0.999125123f, -0.041821491f, 0.f),
    C_Vertex3f(0.519042969f, 0.128819466f, 0.f),
    C_Vertex3f(0.519042969f, 0.128819466f, -1.f),
    C_Normal3f(0.989323974f, -0.145732895f, 0.f),
    C_Vertex3f(0.517633319f, 0.114192247f, 0.f),
    C_Vertex3f(0.517633319f, 0.114192247f, -1.f),
    C_Normal3f(0.968942523f, -0.247285903f, 0.f),
    C_Vertex3f(0.514814019f, 0.100110769f, 0.f),
    C_Vertex3f(0.514814019f, 0.100110769f, -1.f),
    C_Normal3f(0.937250614f, -0.348656356f, 0.f),
    C_Vertex3f(0.51058507f, 0.0865750313f, 0.f),
    C_Vertex3f(0.51058507f, 0.0865750313f, -1.f),
    C_Normal3f(0.894852877f, -0.446361125f, 0.f),
    C_Vertex3f(0.50494647f, 0.0735850334f, 0.f),
    C_Vertex3f(0.50494647f, 0.0735850334f, -1.f),
    C_Normal3f(0.8434214f, -0.537252486f, 0.f),
    C_Vertex3f(0.497898221f, 0.0611407757f, 0.f),
    C_Vertex3f(0.497898221f, 0.0611407757f, -1.f),
    C_Normal3f(0.785389185f, -0.619002461f, 0.f),
    C_Vertex3f(0.489440322f, 0.0492422581f, 0.f),
    C_Vertex3f(0.489440322f, 0.0492422581f, -1.f),
    C_Normal3f(0.723482728f, -0.690342426f, 0.f),
    C_Vertex3f(0.479572773f, 0.0378894806f, 0.f),
    C_Vertex3f(0.479572773f, 0.0378894806f, -1.f),
    C_Normal3f(0.655808032f, -0.754927754f, 0.f),
    C_Vertex3f(0.468295574f, 0.0270824432f, 0.f),
    C_Vertex3f(0.468295574f, 0.0270824432f, -1.f),
    C_Normal3f(0.576839387f, -0.816857636f, 0.f),
    C_Vertex3f(0.456246972f, 0.0176262856f, 0.f),
    C_Vertex3f(0.456246972f, 0.0176262856f, -1.f),
    C_Normal3f(0.490342855f, -0.871529639f, 0.f),
    C_Vertex3f(0.443426967f, 0.00952100754f, 0.f),
    C_Vertex3f(0.443426967f, 0.00952100754f, -1.f),
    C_Normal3f(0.39883393f, -0.917023182f, 0.f),
    C_Vertex3f(0.429835558f, 0.00276660919f, 0.f),
    C_Vertex3f(0.429835558f, 0.00276660919f, -1.f),
    C_Normal3f(0.305276722f, -0.952263713f, 0.f),
    C_Vertex3f(0.415472746f, -0.00263690948f, 0.f),
    C_Vertex3f(0.415472746f, -0.00263690948f, -1.f),
    C_Normal3f(0.212628275f, -0.977133155f, 0.f),
    C_Vertex3f(0.400338531f, -0.00668954849f, 0.f),
    C_Vertex3f(0.400338531f, -0.00668954849f, -1.f),
    C_Normal3f(0.123429224f, -0.992353439f, 0.f),
    C_Vertex3f(0.384432912f, -0.00939130783f, 0.f),
    C_Vertex3f(0.384432912f, -0.00939130783f, -1.f),
    C_Normal3f(0.0418232046f, -0.999125063f, 0.f),
    C_Vertex3f(0.36775589f, -0.0107421875f, 0.f),
    C_Vertex3f(0.36775589f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0318018645f, -0.999494195f, 0.f),
    C_Vertex3f(0.352161407f, -0.0107421875f, 0.f),
    C_Vertex3f(0.352161407f, -0.0107421875f, -1.f),
    C_Normal3f(-0.10470292f, -0.994503617f, 0.f),
    C_Vertex3f(0.338843346f, -0.00982224941f, 0.f),
    C_Vertex3f(0.338843346f, -0.00982224941f, -1.f),
    C_Normal3f(-0.178404883f, -0.983957171f, 0.f),
    C_Vertex3f(0.325947762f, -0.00798237324f, 0.f),
    C_Vertex3f(0.325947762f, -0.00798237324f, -1.f),
    C_Normal3f(-0.253975093f, -0.96721077f, 0.f),
    C_Vertex3f(0.313474655f, -0.00522255898f, 0.f),
    C_Vertex3f(0.313474655f, -0.00522255898f, -1.f),
    C_Normal3f(-0.330061674f, -0.943959355f, 0.f),
    C_Vertex3f(0.301424026f, -0.00154280663f, 0.f),
    C_Vertex3f(0.301424026f, -0.00154280663f, -1.f),
    C_Normal3f(-0.405167043f, -0.914242625f, 0.f),
    C_Vertex3f(0.289795876f, 0.00305688381f, 0.f),
    C_Vertex3f(0.289795876f, 0.00305688381f, -1.f),
    C_Normal3f(-0.477782309f, -0.878478229f, 0.f),
    C_Vertex3f(0.278590202f, 0.00857651234f, 0.f),
    C_Vertex3f(0.278590202f, 0.00857651234f, -1.f),
    C_Normal3f(-0.553150475f, -0.833081424f, 0.f),
    C_Vertex3f(0.267807007f, 0.0150160789f, 0.f),
    C_Vertex3f(0.267807007f, 0.0150160789f, -1.f),
    C_Normal3f(-0.637743771f, -0.770248592f, 0.f),
    C_Vertex3f(0.25226593f, 0.026055336f, 0.f),
    C_Vertex3f(0.25226593f, 0.026055336f, -1.f),
    C_Normal3f(-0.728630841f, -0.684906542f, 0.f),
    C_Vertex3f(0.234162331f, 0.0428729057f, 0.f),
    C_Vertex3f(0.234162331f, 0.0428729057f, -1.f),
    C_Normal3f(-0.815355897f, -0.578960061f, 0.f),
    C_Vertex3f(0.218676567f, 0.0617890358f, 0.f),
    C_Vertex3f(0.218676567f, 0.0617890358f, -1.f),
    C_Normal3f(-0.875320733f, -0.48354274f, 0.f),
    C_Vertex3f(0.20580864f, 0.0828037262f, 0.f),
    C_Vertex3f(0.20580864f, 0.0828037262f, -1.f),
    C_Normal3f(-0.914141059f, -0.405396432f, 0.f),
    C_Vertex3f(0.200683594f, 0.0943603516f, 0.f),
    C_Vertex3f(0.200683594f, 0.0943603516f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.420583904f, 0.907253742f, 0.f),
    C_Vertex3f(0.200683594f, 0.0943603516f, 0.f),
    C_Vertex3f(0.200683594f, 0.0943603516f, -1.f),
    C_Vertex3f(0.268554688f, 0.125823975f, -1.f),
    C_Vertex3f(0.268554688f, 0.125823975f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.883456945f, 0.468512475f, 0.f),
    C_Vertex3f(0.268554688f, 0.125823975f, 0.f),
    C_Vertex3f(0.268554688f, 0.125823975f, -1.f),
    C_Normal3f(0.860579729f, 0.50931567f, 0.f),
    C_Vertex3f(0.273942947f, 0.115663528f, 0.f),
    C_Vertex3f(0.273942947f, 0.115663528f, -1.f),
    C_Normal3f(0.822441459f, 0.568849623f, 0.f),
    C_Vertex3f(0.284751892f, 0.0984559059f, 0.f),
    C_Vertex3f(0.284751892f, 0.0984559059f, -1.f),
    C_Normal3f(0.755940914f, 0.6546399f, 0.f),
    C_Vertex3f(0.295593262f, 0.0843615532f, 0.f),
    C_Vertex3f(0.295593262f, 0.0843615532f, -1.f),
    C_Normal3f(0.654392719f, 0.756154776f, 0.f),
    C_Vertex3f(0.306467056f, 0.0733804703f, 0.f),
    C_Vertex3f(0.306467056f, 0.0733804703f, -1.f),
    C_Normal3f(0.529043376f, 0.848594785f, 0.f),
    C_Vertex3f(0.317373276f, 0.0655126572f, 0.f),
    C_Vertex3f(0.317373276f, 0.0655126572f, -1.f),
    C_Normal3f(0.39856714f, 0.917139173f, 0.f),
    C_Vertex3f(0.328552246f, 0.0596117973f, 0.f),
    C_Vertex3f(0.328552246f, 0.0596117973f, -1.f),
    C_Normal3f(0.24673681f, 0.969082534f, 0.f),
    C_Vertex3f(0.340003967f, 0.0556778908f, 0.f),
    C_Vertex3f(0.340003967f, 0.0556778908f, -1.f),
    C_Normal3f(0.0754061714f, 0.997152865f, 0.f),
    C_Vertex3f(0.351728439f, 0.0537109375f, 0.f),
    C_Vertex3f(0.351728439f, 0.0537109375f, -1.f),
    C_Normal3f(-0.095407173f, 0.995438278f, 0.f),
    C_Vertex3f(0.366014481f, 0.0537109375f, 0.f),
    C_Vertex3f(0.366014481f, 0.0537109375f, -1.f),
    C_Normal3f(-0.277283698f, 0.960788131f, 0.f),
    C_Vertex3f(0.381390095f, 0.0565538406f, 0.f),
    C_Vertex3f(0.381390095f, 0.0565538406f, -1.f),
    C_Normal3f(-0.463783175f, 0.885948777f, 0.f),
    C_Vertex3f(0.395566463f, 0.0622396469f, 0.f),
    C_Vertex3f(0.395566463f, 0.0622396469f, -1.f),
    C_Normal3f(-0.626542389f, 0.779387295f, 0.f),
    C_Vertex3f(0.408543587f, 0.0707683563f, 0.f),
    C_Vertex3f(0.408543587f, 0.0707683563f, -1.f),
    C_Normal3f(-0.754940212f, 0.655793607f, 0.f),
    C_Vertex3f(0.420321465f, 0.0821399689f, 0.f),
    C_Vertex3f(0.420321465f, 0.0821399689f, -1.f),
    C_Normal3f(-0.867674768f, 0.497132182f, 0.f),
    C_Vertex3f(0.429154873f, 0.0944957733f, 0.f),
    C_Vertex3f(0.429154873f, 0.0944957733f, -1.f),
    C_Normal3f(-0.952615201f, 0.304178089f, 0.f),
    C_Vertex3f(0.435043812f, 0.10783577f, 0.f),
    C_Vertex3f(0.435043812f, 0.10783577f, -1.f),
    C_Normal3f(-0.991144836f, 0.132784858f, 0.f),
    C_Vertex3f(0.437988281f, 0.122159958f, 0.f),
    C_Vertex3f(0.437988281f, 0.122159958f, -1.f),
    C_Normal3f(-0.999937177f, 0.0112134721f, 0.f),
    C_Vertex3f(0.437988281f, 0.129814148f, 0.f),
    C_Vertex3f(0.437988281f, 0.129814148f, -1.f),
    C_Normal3f(-0.996983469f, -0.0776150227f, 0.f),
    C_Vertex3f(0.438142776f, 0.135936737f, 0.f),
    C_Vertex3f(0.438142776f, 0.135936737f, -1.f),
    C_Normal3f(-0.97619909f, -0.216876373f, 0.f),
    C_Vertex3f(0.436626434f, 0.147307396f, 0.f),
    C_Vertex3f(0.436626434f, 0.147307396f, -1.f),
    C_Normal3f(-0.921011269f, -0.389536023f, 0.f),
    C_Vertex3f(0.43328476f, 0.157803535f, 0.f),
    C_Vertex3f(0.43328476f, 0.157803535f, -1.f),
    C_Normal3f(-0.843837678f, -0.536598504f, 0.f),
    C_Vertex3f(0.428117752f, 0.167425156f, 0.f),
    C_Vertex3f(0.428117752f, 0.167425156f, -1.f),
    C_Normal3f(-0.757808983f, -0.65247643f, 0.f),
    C_Vertex3f(0.423480988f, 0.173220634f, 0.f),
    C_Vertex3f(0.423480988f, 0.173220634f, -1.f),
    C_Normal3f(-0.695984244f, -0.718056977f, 0.f),
    C_Vertex3f(0.420248985f, 0.176564217f, 0.f),
    C_Vertex3f(0.420248985f, 0.176564217f, -1.f),
    C_Normal3f(-0.660076439f, -0.751198411f, 0.f),
    C_Vertex3f(0.41606617f, 0.180407524f, 0.f),
    C_Vertex3f(0.41606617f, 0.180407524f, -1.f),
    C_Normal3f(-0.63354218f, -0.773708165f, 0.f),
    C_Vertex3f(0.410932541f, 0.184750557f, 0.f),
    C_Vertex3f(0.410932541f, 0.184750557f, -1.f),
    C_Normal3f(-0.613228798f, -0.78990531f, 0.f),
    C_Vertex3f(0.404848099f, 0.189593315f, 0.f),
    C_Vertex3f(0.404848099f, 0.189593315f, -1.f),
    C_Normal3f(-0.597218037f, -0.802078903f, 0.f),
    C_Vertex3f(0.397812843f, 0.194935799f, 0.f),
    C_Vertex3f(0.397812843f, 0.194935799f, -1.f),
    C_Normal3f(-0.58429271f, -0.811543047f, 0.f),
    C_Vertex3f(0.389826775f, 0.200778008f, 0.f),
    C_Vertex3f(0.389826775f, 0.200778008f, -1.f),
    C_Normal3f(-0.575290203f, -0.817949474f, 0.f),
    C_Vertex3f(0.380889893f, 0.207119942f, 0.f),
    C_Vertex3f(0.380889893f, 0.207119942f, -1.f),
    C_Normal3f(-0.564284086f, -0.825580657f, 0.f),
    C_Vertex3f(0.375946045f, 0.210540771f, 0.f),
    C_Vertex3f(0.375946045f, 0.210540771f, -1.f),
    C_Normal3f(-0.570820034f, -0.821075201f, 0.f),
    C_Vertex3f(0.366839409f, 0.216723442f, 0.f),
    C_Vertex3f(0.366839409f, 0.216723442f, -1.f),
    C_Normal3f(-0.583764136f, -0.811923265f, 0.f),
    C_Vertex3f(0.35036993f, 0.228321552f, 0.f),
    C_Vertex3f(0.35036993f, 0.228321552f, -1.f),
    C_Normal3f(-0.602097511f, -0.798422575f, 0.f),
    C_Vertex3f(0.335644245f, 0.239152431f, 0.f),
    C_Vertex3f(0.335644245f, 0.239152431f, -1.f),
    C_Normal3f(-0.62438184f, -0.781119347f, 0.f),
    C_Vertex3f(0.322662354f, 0.24921608f, 0.f),
    C_Vertex3f(0.322662354f, 0.24921608f, -1.f),
    C_Normal3f(-0.651949644f, -0.758262277f, 0.f),
    C_Vertex3f(0.311424255f, 0.258512497f, 0.f),
    C_Vertex3f(0.311424255f, 0.258512497f, -1.f),
    C_Normal3f(-0.686720848f, -0.726921141f, 0.f),
    C_Vertex3f(0.301929951f, 0.267041683f, 0.f),
    C_Vertex3f(0.301929951f, 0.267041683f, -1.f),
    C_Normal3f(-0.73144263f, -0.681902945f, 0.f),
    C_Vertex3f(0.29417944f, 0.274803638f, 0.f),
    C_Vertex3f(0.29417944f, 0.274803638f, -1.f),
    C_Normal3f(-0.778932929f, -0.627107322f, 0.f),
    C_Vertex3f(0.288172722f, 0.281798363f, 0.f),
    C_Vertex3f(0.288172722f, 0.281798363f, -1.f),
    C_Normal3f(-0.820705354f, -0.571351707f, 0.f),
    C_Vertex3f(0.28604126f, 0.284912109f, 0.f),
    C_Vertex3f(0.28604126f, 0.284912109f, -1.f),
    C_Normal3f(-0.865068436f, -0.50165391f, 0.f),
    C_Vertex3f(0.281669617f, 0.291139603f, 0.f),
    C_Vertex3f(0.281669617f, 0.291139603f, -1.f),
    C_Normal3f(-0.918587565f, -0.395217508f, 0.f),
    C_Vertex3f(0.275112152f, 0.303758621f, 0.f),
    C_Vertex3f(0.275112152f, 0.303758621f, -1.f),
    C_Normal3f(-0.969025373f, -0.2469614f, 0.f),
    C_Vertex3f(0.270740509f, 0.316541672f, 0.f),
    C_Vertex3f(0.270740509f, 0.316541672f, -1.f),
    C_Normal3f(-0.996972084f, -0.07776016f, 0.f),
    C_Vertex3f(0.268554688f, 0.329488754f, 0.f),
    C_Vertex3f(0.268554688f, 0.329488754f, -1.f),
    C_Normal3f(-0.995853126f, 0.0909752548f, 0.f),
    C_Vertex3f(0.268554688f, 0.344566345f, 0.f),
    C_Vertex3f(0.268554688f, 0.344566345f, -1.f),
    C_Normal3f(-0.97168535f, 0.2362784f, 0.f),
    C_Vertex3f(0.271509171f, 0.361829758f, 0.f),
    C_Vertex3f(0.271509171f, 0.361829758f, -1.f),
    C_Normal3f(-0.932698309f, 0.360657662f, 0.f),
    C_Vertex3f(0.275940895f, 0.374941826f, 0.f),
    C_Vertex3f(0.275940895f, 0.374941826f, -1.f),
    C_Normal3f(-0.901862085f, 0.432023853f, 0.f),
    C_Vertex3f(0.280973434f, 0.386305332f, 0.f),
    C_Vertex3f(0.280973434f, 0.386305332f, -1.f),
    C_Normal3f(-0.882559299f, 0.470201254f, 0.f),
    C_Vertex3f(0.28808403f, 0.400290966f, 0.f),
    C_Vertex3f(0.28808403f, 0.400290966f, -1.f),
    C_Normal3f(-0.868486226f, 0.495713264f, 0.f),
    C_Vertex3f(0.297272682f, 0.416898727f, 0.f),
    C_Vertex3f(0.297272682f, 0.416898727f, -1.f),
    C_Normal3f(-0.868364692f, 0.495926201f, 0.f),
    C_Vertex3f(0.308539391f, 0.436128616f, 0.f),
    C_Vertex3f(0.308539391f, 0.436128616f, -1.f),
    C_Normal3f(-0.881854832f, 0.471521109f, 0.f),
    C_Vertex3f(0.317864418f, 0.452954769f, 0.f),
    C_Vertex3f(0.317864418f, 0.452954769f, -1.f),
    C_Normal3f(-0.89974612f, 0.436413616f, 0.f),
    C_Vertex3f(0.325247765f, 0.467377186f, 0.f),
    C_Vertex3f(0.325247765f, 0.467377186f, -1.f),
    C_Normal3f(-0.924170554f, 0.381979913f, 0.f),
    C_Vertex3f(0.33068943f, 0.479395866f, 0.f),
    C_Vertex3f(0.33068943f, 0.479395866f, -1.f),
    C_Normal3f(-0.950400949f, 0.311027527f, 0.f),
    C_Vertex3f(0.334189415f, 0.489010811f, 0.f),
    C_Vertex3f(0.334189415f, 0.489010811f, -1.f),
    C_Normal3f(-0.970845759f, 0.239704937f, 0.f),
    C_Vertex3f(0.336803913f, 0.498079777f, 0.f),
    C_Vertex3f(0.336803913f, 0.498079777f, -1.f),
    C_Normal3f(-0.988062978f, 0.154050216f, 0.f),
    C_Vertex3f(0.338532925f, 0.506602764f, 0.f),
    C_Vertex3f(0.338532925f, 0.506602764f, -1.f),
    C_Normal3f(-0.99900955f, 0.0444961525f, 0.f),
    C_Vertex3f(0.33937645f, 0.514579773f, 0.f),
    C_Vertex3f(0.33937645f, 0.514579773f, -1.f),
    C_Normal3f(-0.991803288f, -0.127774462f, 0.f),
    C_Vertex3f(0.339355469f, 0.52507019f, 0.f),
    C_Vertex3f(0.339355469f, 0.52507019f, -1.f),
    C_Normal3f(-0.940618634f, -0.339465231f, 0.f),
    C_Vertex3f(0.336381435f, 0.537827492f, 0.f),
    C_Vertex3f(0.336381435f, 0.537827492f, -1.f),
    C_Normal3f(-0.841239274f, -0.540662944f, 0.f),
    C_Vertex3f(0.330433369f, 0.54979229f, 0.f),
    C_Vertex3f(0.330433369f, 0.54979229f, -1.f),
    C_Normal3f(-0.71924603f, -0.694755554f, 0.f),
    C_Vertex3f(0.321511269f, 0.560964584f, 0.f),
    C_Vertex3f(0.321511269f, 0.560964584f, -1.f),
    C_Normal3f(-0.575924337f, -0.817502975f, 0.f),
    C_Vertex3f(0.309615135f, 0.571344376f, 0.f),
    C_Vertex3f(0.309615135f, 0.571344376f, -1.f),
    C_Normal3f(-0.399563611f, -0.916705489f, 0.f),
    C_Vertex3f(0.295727253f, 0.579129219f, 0.f),
    C_Vertex3f(0.295727253f, 0.579129219f, -1.f),
    C_Normal3f(-0.224753946f, -0.9744156f, 0.f),
    C_Vertex3f(0.279847622f, 0.584319115f, 0.f),
    C_Vertex3f(0.279847622f, 0.584319115f, -1.f),
    C_Normal3f(-0.0769050643f, -0.997038424f, 0.f),
    C_Vertex3f(0.261976242f, 0.586914063f, 0.f),
    C_Vertex3f(0.261976242f, 0.586914063f, -1.f),
    C_Normal3f(0.0334717892f, -0.999439657f, 0.f),
    C_Vertex3f(0.24620533f, 0.586914063f, 0.f),
    C_Vertex3f(0.24620533f, 0.586914063f, -1.f),
    C_Normal3f(0.123338126f, -0.992364645f, 0.f),
    C_Vertex3f(0.235059381f, 0.586012602f, 0.f),
    C_Vertex3f(0.235059381f, 0.586012602f, -1.f),
    C_Normal3f(0.212821037f, -0.977091193f, 0.f),
    C_Vertex3f(0.224446177f, 0.584209681f, 0.f),
    C_Vertex3f(0.224446177f, 0.584209681f, -1.f),
    C_Normal3f(0.306060642f, -0.952012002f, 0.f),
    C_Vertex3f(0.214365721f, 0.581505299f, 0.f),
    C_Vertex3f(0.214365721f, 0.581505299f, -1.f),
    C_Normal3f(0.40048635f, -0.916302741f, 0.f),
    C_Vertex3f(0.20481801f, 0.577899456f, 0.f),
    C_Vertex3f(0.20481801f, 0.577899456f, -1.f),
    C_Normal3f(0.493050128f, -0.870000958f, 0.f),
    C_Vertex3f(0.195803046f, 0.573392153f, 0.f),
    C_Vertex3f(0.195803046f, 0.573392153f, -1.f),
    C_Normal3f(0.580649734f, -0.814153492f, 0.f),
    C_Vertex3f(0.187320828f, 0.567983389f, 0.f),
    C_Vertex3f(0.187320828f, 0.567983389f, -1.f),
    C_Normal3f(0.660618424f, -0.750721812f, 0.f),
    C_Vertex3f(0.179371357f, 0.561673164f, 0.f),
    C_Vertex3f(0.179371357f, 0.561673164f, -1.f),
    C_Normal3f(0.765154302f, -0.643846929f, 0.f),
    C_Vertex3f(0.171954632f, 0.554461479f, 0.f),
    C_Vertex3f(0.171954632f, 0.554461479f, -1.f),
    C_Normal3f(0.864537716f, -0.502568007f, 0.f),
    C_Vertex3f(0.165464997f, 0.545146704f, 0.f),
    C_Vertex3f(0.165464997f, 0.545146704f, -1.f),
    C_Normal3f(0.925602138f, -0.37849775f, 0.f),
    C_Vertex3f(0.159902453f, 0.533728838f, 0.f),
    C_Vertex3f(0.159902453f, 0.533728838f, -1.f),
    C_Normal3f(0.961378515f, -0.275229514f, 0.f),
    C_Vertex3f(0.155267f, 0.520207882f, 0.f),
    C_Vertex3f(0.155267f, 0.520207882f, -1.f),
    C_Normal3f(0.981589556f, -0.191002443f, 0.f),
    C_Vertex3f(0.151558638f, 0.504583836f, 0.f),
    C_Vertex3f(0.151558638f, 0.504583836f, -1.f),
    C_Normal3f(0.992469251f, -0.122493818f, 0.f),
    C_Vertex3f(0.148777366f, 0.486856699f, 0.f),
    C_Vertex3f(0.148777366f, 0.486856699f, -1.f),
    C_Normal3f(0.997789919f, -0.0664485022f, 0.f),
    C_Vertex3f(0.146923184f, 0.467026472f, 0.f),
    C_Vertex3f(0.146923184f, 0.467026472f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.145996094f, 0.445093155f, 0.f),
    C_Vertex3f(0.145996094f, 0.445093155f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.145996094f, 0.445093155f, 0.f),
    C_Vertex3f(0.145996094f, 0.445093155f, -1.f),
    C_Vertex3f(0.145996094f, 0.f, -1.f),
    C_Vertex3f(0.145996094f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.145996094f, 0.f, 0.f),
    C_Vertex3f(0.145996094f, 0.f, -1.f),
    C_Vertex3f(0.0673828125f, 0.f, -1.f),
    C_Vertex3f(0.0673828125f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342555285f, 0.633666396f, 0.f),
    C_Vertex3f(0.328701496f, 0.63956666f, 0.f),
    C_Vertex3f(0.330433369f, 0.54979229f, 0.f),
    C_Vertex3f(0.336381435f, 0.537827492f, 0.f),
    C_Vertex3f(0.339355469f, 0.52507019f, 0.f),
    C_Vertex3f(0.33937645f, 0.514579773f, 0.f),
    C_Vertex3f(0.35036993f, 0.228321552f, 0.f),
    C_Vertex3f(0.349563599f, 0.34478569f, 0.f),
    C_Vertex3f(0.349555969f, 0.353527069f, 0.f),
    C_Vertex3f(0.355403662f, 0.62658608f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.24620533f, 0.586914063f, 0.f),
    C_Vertex3f(0.261976242f, 0.586914063f, 0.f),
    C_Vertex3f(0.263232231f, 0.651367188f, 0.f),
    C_Vertex3f(0.279847622f, 0.584319115f, 0.f),
    C_Vertex3f(0.281107664f, 0.650187135f, 0.f),
    C_Vertex3f(0.295727253f, 0.579129219f, 0.f),
    C_Vertex3f(0.297977686f, 0.647827029f, 0.f),
    C_Vertex3f(0.309615135f, 0.571344376f, 0.f),
    C_Vertex3f(0.313842297f, 0.644286871f, 0.f),
    C_Vertex3f(0.321511269f, 0.560964584f, 0.f),
    C_Vertex3f(0.328701496f, 0.63956666f, 0.f),
    C_Vertex3f(0.330433369f, 0.54979229f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.23960495f, 0.651367188f, 0.f),
    C_Vertex3f(0.212649345f, 0.648309231f, 0.f),
    C_Vertex3f(0.214365721f, 0.581505299f, 0.f),
    C_Vertex3f(0.224446177f, 0.584209681f, 0.f),
    C_Vertex3f(0.235059381f, 0.586012602f, 0.f),
    C_Vertex3f(0.24620533f, 0.586914063f, 0.f),
    C_Vertex3f(0.263232231f, 0.651367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.212649345f, 0.648309231f, 0.f),
    C_Vertex3f(0.187111855f, 0.642193317f, 0.f),
    C_Vertex3f(0.187320828f, 0.567983389f, 0.f),
    C_Vertex3f(0.195803046f, 0.573392153f, 0.f),
    C_Vertex3f(0.20481801f, 0.577899456f, 0.f),
    C_Vertex3f(0.214365721f, 0.581505299f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.187111855f, 0.642193317f, 0.f),
    C_Vertex3f(0.162992477f, 0.633019447f, 0.f),
    C_Vertex3f(0.165464997f, 0.545146704f, 0.f),
    C_Vertex3f(0.171954632f, 0.554461479f, 0.f),
    C_Vertex3f(0.179371357f, 0.561673164f, 0.f),
    C_Vertex3f(0.187320828f, 0.567983389f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162992477f, 0.633019447f, 0.f),
    C_Vertex3f(0.14596653f, 0.623845577f, 0.f),
    C_Vertex3f(0.148777366f, 0.486856699f, 0.f),
    C_Vertex3f(0.151558638f, 0.504583836f, 0.f),
    C_Vertex3f(0.155267f, 0.520207882f, 0.f),
    C_Vertex3f(0.159902453f, 0.533728838f, 0.f),
    C_Vertex3f(0.165464997f, 0.545146704f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14596653f, 0.623845577f, 0.f),
    C_Vertex3f(0.135427594f, 0.617149353f, 0.f),
    C_Vertex3f(0.145996094f, 0.f, 0.f),
    C_Vertex3f(0.145996094f, 0.445093155f, 0.f),
    C_Vertex3f(0.146923184f, 0.467026472f, 0.f),
    C_Vertex3f(0.148777366f, 0.486856699f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145996094f, 0.f, 0.f),
    C_Vertex3f(0.135427594f, 0.617149353f, 0.f),
    C_Vertex3f(0.125700355f, 0.609872818f, 0.f),
    C_Vertex3f(0.116784811f, 0.602015972f, 0.f),
    C_Vertex3f(0.108680964f, 0.593578815f, 0.f),
    C_Vertex3f(0.101388812f, 0.584561348f, 0.f),
    C_Vertex3f(0.0949083567f, 0.57496357f, 0.f),
    C_Vertex3f(0.0892395973f, 0.56478548f, 0.f),
    C_Vertex3f(0.084382534f, 0.554027081f, 0.f),
    C_Vertex3f(0.0801326036f, 0.542115331f, 0.f),
    C_Vertex3f(0.0764898062f, 0.529050231f, 0.f),
    C_Vertex3f(0.0734541416f, 0.514831781f, 0.f),
    C_Vertex3f(0.07102561f, 0.499459982f, 0.f),
    C_Vertex3f(0.0692042112f, 0.482934833f, 0.f),
    C_Vertex3f(0.0679899454f, 0.465256333f, 0.f),
    C_Vertex3f(0.0673828125f, 0.446424484f, 0.f),
    C_Vertex3f(0.0673828125f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.51058507f, 0.0865750313f, 0.f),
    C_Vertex3f(0.514814019f, 0.100110769f, 0.f),
    C_Vertex3f(0.515545845f, 0.169265747f, 0.f),
    C_Vertex3f(0.517633319f, 0.114192247f, 0.f),
    C_Vertex3f(0.519042969f, 0.14786911f, 0.f),
    C_Vertex3f(0.519042969f, 0.128819466f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.489440322f, 0.0492422581f, 0.f),
    C_Vertex3f(0.497898221f, 0.0611407757f, 0.f),
    C_Vertex3f(0.498060226f, 0.20746994f, 0.f),
    C_Vertex3f(0.50494647f, 0.0735850334f, 0.f),
    C_Vertex3f(0.508551598f, 0.18913269f, 0.f),
    C_Vertex3f(0.51058507f, 0.0865750313f, 0.f),
    C_Vertex3f(0.515545845f, 0.169265747f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.498060226f, 0.20746994f, 0.f),
    C_Vertex3f(0.488606453f, 0.218821526f, 0.f),
    C_Vertex3f(0.489440322f, 0.0492422581f, 0.f),
    C_Vertex3f(0.482442856f, 0.225193262f, 0.f),
    C_Vertex3f(0.479572773f, 0.0378894806f, 0.f),
    C_Vertex3f(0.475034714f, 0.232041121f, 0.f),
    C_Vertex3f(0.468295574f, 0.0270824432f, 0.f),
    C_Vertex3f(0.466382027f, 0.239365101f, 0.f),
    C_Vertex3f(0.456484795f, 0.247165203f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.468295574f, 0.0270824432f, 0.f),
    C_Vertex3f(0.456484795f, 0.247165203f, 0.f),
    C_Vertex3f(0.456246972f, 0.0176262856f, 0.f),
    C_Vertex3f(0.445343018f, 0.255441427f, 0.f),
    C_Vertex3f(0.443426967f, 0.00952100754f, 0.f),
    C_Vertex3f(0.438142776f, 0.135936737f, 0.f),
    C_Vertex3f(0.437988281f, 0.129814148f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.423480988f, 0.173220634f, 0.f),
    C_Vertex3f(0.428117752f, 0.167425156f, 0.f),
    C_Vertex3f(0.432956696f, 0.264193773f, 0.f),
    C_Vertex3f(0.43328476f, 0.157803535f, 0.f),
    C_Vertex3f(0.445343018f, 0.255441427f, 0.f),
    C_Vertex3f(0.436626434f, 0.147307396f, 0.f),
    C_Vertex3f(0.438142776f, 0.135936737f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.410932541f, 0.184750557f, 0.f),
    C_Vertex3f(0.41606617f, 0.180407524f, 0.f),
    C_Vertex3f(0.419325829f, 0.273422241f, 0.f),
    C_Vertex3f(0.420248985f, 0.176564217f, 0.f),
    C_Vertex3f(0.432956696f, 0.264193773f, 0.f),
    C_Vertex3f(0.423480988f, 0.173220634f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.406430244f, 0.281835556f, 0.f),
    C_Vertex3f(0.39659214f, 0.288504362f, 0.f),
    C_Vertex3f(0.397812843f, 0.194935799f, 0.f),
    C_Vertex3f(0.404848099f, 0.189593315f, 0.f),
    C_Vertex3f(0.410932541f, 0.184750557f, 0.f),
    C_Vertex3f(0.419325829f, 0.273422241f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.397812843f, 0.194935799f, 0.f),
    C_Vertex3f(0.39659214f, 0.288504362f, 0.f),
    C_Vertex3f(0.389826775f, 0.200778008f, 0.f),
    C_Vertex3f(0.387831688f, 0.294719934f, 0.f),
    C_Vertex3f(0.380889893f, 0.207119942f, 0.f),
    C_Vertex3f(0.380148888f, 0.300482273f, 0.f),
    C_Vertex3f(0.375946045f, 0.210540771f, 0.f),
    C_Vertex3f(0.373543739f, 0.305791378f, 0.f),
    C_Vertex3f(0.368016243f, 0.310647249f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.366839409f, 0.216723442f, 0.f),
    C_Vertex3f(0.375946045f, 0.210540771f, 0.f),
    C_Vertex3f(0.368016243f, 0.310647249f, 0.f),
    C_Vertex3f(0.363566399f, 0.315049887f, 0.f),
    C_Vertex3f(0.360194206f, 0.31899929f, 0.f),
    C_Vertex3f(0.359046936f, 0.320747375f, 0.f),
    C_Vertex3f(0.356641769f, 0.324134827f, 0.f),
    C_Vertex3f(0.353056908f, 0.330964088f, 0.f),
    C_Vertex3f(0.350697517f, 0.33784771f, 0.f),
    C_Vertex3f(0.35036993f, 0.228321552f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.33937645f, 0.514579773f, 0.f),
    C_Vertex3f(0.338532925f, 0.506602764f, 0.f),
    C_Vertex3f(0.35036993f, 0.228321552f, 0.f),
    C_Vertex3f(0.336803913f, 0.498079777f, 0.f),
    C_Vertex3f(0.335644245f, 0.239152431f, 0.f),
    C_Vertex3f(0.334189415f, 0.489010811f, 0.f),
    C_Vertex3f(0.33068943f, 0.479395866f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.288172722f, 0.281798363f, 0.f),
    C_Vertex3f(0.29417944f, 0.274803638f, 0.f),
    C_Vertex3f(0.297272682f, 0.416898727f, 0.f),
    C_Vertex3f(0.301929951f, 0.267041683f, 0.f),
    C_Vertex3f(0.308539391f, 0.436128616f, 0.f),
    C_Vertex3f(0.311424255f, 0.258512497f, 0.f),
    C_Vertex3f(0.317864418f, 0.452954769f, 0.f),
    C_Vertex3f(0.322662354f, 0.24921608f, 0.f),
    C_Vertex3f(0.325247765f, 0.467377186f, 0.f),
    C_Vertex3f(0.335644245f, 0.239152431f, 0.f),
    C_Vertex3f(0.33068943f, 0.479395866f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.28808403f, 0.400290966f, 0.f),
    C_Vertex3f(0.280973434f, 0.386305332f, 0.f),
    C_Vertex3f(0.281669617f, 0.291139603f, 0.f),
    C_Vertex3f(0.28604126f, 0.284912109f, 0.f),
    C_Vertex3f(0.288172722f, 0.281798363f, 0.f),
    C_Vertex3f(0.297272682f, 0.416898727f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268554688f, 0.344566345f, 0.f),
    C_Vertex3f(0.270740509f, 0.316541672f, 0.f),
    C_Vertex3f(0.271509171f, 0.361829758f, 0.f),
    C_Vertex3f(0.275112152f, 0.303758621f, 0.f),
    C_Vertex3f(0.275940895f, 0.374941826f, 0.f),
    C_Vertex3f(0.281669617f, 0.291139603f, 0.f),
    C_Vertex3f(0.280973434f, 0.386305332f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.437988281f, 0.129814148f, 0.f),
    C_Vertex3f(0.437988281f, 0.122159958f, 0.f),
    C_Vertex3f(0.443426967f, 0.00952100754f, 0.f),
    C_Vertex3f(0.435043812f, 0.10783577f, 0.f),
    C_Vertex3f(0.429835558f, 0.00276660919f, 0.f),
    C_Vertex3f(0.429154873f, 0.0944957733f, 0.f),
    C_Vertex3f(0.420321465f, 0.0821399689f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.429835558f, 0.00276660919f, 0.f),
    C_Vertex3f(0.420321465f, 0.0821399689f, 0.f),
    C_Vertex3f(0.415472746f, -0.00263690948f, 0.f),
    C_Vertex3f(0.408543587f, 0.0707683563f, 0.f),
    C_Vertex3f(0.400338531f, -0.00668954849f, 0.f),
    C_Vertex3f(0.395566463f, 0.0622396469f, 0.f),
    C_Vertex3f(0.384432912f, -0.00939130783f, 0.f),
    C_Vertex3f(0.381390095f, 0.0565538406f, 0.f),
    C_Vertex3f(0.36775589f, -0.0107421875f, 0.f),
    C_Vertex3f(0.366014481f, 0.0537109375f, 0.f),
    C_Vertex3f(0.352161407f, -0.0107421875f, 0.f),
    C_Vertex3f(0.351728439f, 0.0537109375f, 0.f),
    C_Vertex3f(0.340003967f, 0.0556778908f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.352161407f, -0.0107421875f, 0.f),
    C_Vertex3f(0.340003967f, 0.0556778908f, 0.f),
    C_Vertex3f(0.338843346f, -0.00982224941f, 0.f),
    C_Vertex3f(0.328552246f, 0.0596117973f, 0.f),
    C_Vertex3f(0.325947762f, -0.00798237324f, 0.f),
    C_Vertex3f(0.317373276f, 0.0655126572f, 0.f),
    C_Vertex3f(0.313474655f, -0.00522255898f, 0.f),
    C_Vertex3f(0.306467056f, 0.0733804703f, 0.f),
    C_Vertex3f(0.301424026f, -0.00154280663f, 0.f),
    C_Vertex3f(0.295593262f, 0.0843615532f, 0.f),
    C_Vertex3f(0.289795876f, 0.00305688381f, 0.f),
    C_Vertex3f(0.284751892f, 0.0984559059f, 0.f),
    C_Vertex3f(0.278590202f, 0.00857651234f, 0.f),
    C_Vertex3f(0.273942947f, 0.115663528f, 0.f),
    C_Vertex3f(0.268554688f, 0.125823975f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.268554688f, 0.125823975f, 0.f),
    C_Vertex3f(0.200683594f, 0.0943603516f, 0.f),
    C_Vertex3f(0.20580864f, 0.0828037262f, 0.f),
    C_Vertex3f(0.218676567f, 0.0617890358f, 0.f),
    C_Vertex3f(0.234162331f, 0.0428729057f, 0.f),
    C_Vertex3f(0.25226593f, 0.026055336f, 0.f),
    C_Vertex3f(0.267807007f, 0.0150160789f, 0.f),
    C_Vertex3f(0.278590202f, 0.00857651234f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.360554218f, 0.385541916f, 0.f),
    C_Vertex3f(0.366899014f, 0.397779465f, 0.f),
    C_Vertex3f(0.367246628f, 0.61832571f, 0.f),
    C_Vertex3f(0.375102997f, 0.412693977f, 0.f),
    C_Vertex3f(0.378084183f, 0.608885288f, 0.f),
    C_Vertex3f(0.385166168f, 0.430285454f, 0.f),
    C_Vertex3f(0.387567043f, 0.599069953f, 0.f),
    C_Vertex3f(0.393698215f, 0.446425438f, 0.f),
    C_Vertex3f(0.39569521f, 0.588879704f, 0.f),
    C_Vertex3f(0.400699139f, 0.46111393f, 0.f),
    C_Vertex3f(0.402468681f, 0.578314543f, 0.f),
    C_Vertex3f(0.406168938f, 0.474350929f, 0.f),
    C_Vertex3f(0.407887459f, 0.567374468f, 0.f),
    C_Vertex3f(0.410107613f, 0.486136436f, 0.f),
    C_Vertex3f(0.411951542f, 0.55605948f, 0.f),
    C_Vertex3f(0.413061619f, 0.497759342f, 0.f),
    C_Vertex3f(0.414660931f, 0.544369578f, 0.f),
    C_Vertex3f(0.415030956f, 0.509219646f, 0.f),
    C_Vertex3f(0.416015625f, 0.532304764f, 0.f),
    C_Vertex3f(0.416015625f, 0.520517349f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.349555969f, 0.353527069f, 0.f),
    C_Vertex3f(0.352128983f, 0.365655899f, 0.f),
    C_Vertex3f(0.355403662f, 0.62658608f, 0.f),
    C_Vertex3f(0.356068611f, 0.375981331f, 0.f),
    C_Vertex3f(0.367246628f, 0.61832571f, 0.f),
    C_Vertex3f(0.360554218f, 0.385541916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.349609375f, 0.34828186f, 0.f),
    C_Vertex3f(0.349555969f, 0.353527069f, 0.f),
    C_Vertex3f(0.349563599f, 0.34478569f, 0.f),
    C_Vertex3f(0.270740509f, 0.316541672f, 0.f),
    C_Vertex3f(0.268554688f, 0.344566345f, 0.f),
    C_Vertex3f(0.268554688f, 0.329488754f, 0.f),
    C_Vertex3f(0.349563599f, 0.34478569f, 0.f),
    C_Vertex3f(0.35036993f, 0.228321552f, 0.f),
    C_Vertex3f(0.350697517f, 0.33784771f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342555285f, 0.633666396f, -1.f),
    C_Vertex3f(0.328701496f, 0.63956666f, -1.f),
    C_Vertex3f(0.330433369f, 0.54979229f, -1.f),
    C_Vertex3f(0.336381435f, 0.537827492f, -1.f),
    C_Vertex3f(0.339355469f, 0.52507019f, -1.f),
    C_Vertex3f(0.33937645f, 0.514579773f, -1.f),
    C_Vertex3f(0.35036993f, 0.228321552f, -1.f),
    C_Vertex3f(0.349563599f, 0.34478569f, -1.f),
    C_Vertex3f(0.349555969f, 0.353527069f, -1.f),
    C_Vertex3f(0.355403662f, 0.62658608f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.24620533f, 0.586914063f, -1.f),
    C_Vertex3f(0.261976242f, 0.586914063f, -1.f),
    C_Vertex3f(0.263232231f, 0.651367188f, -1.f),
    C_Vertex3f(0.279847622f, 0.584319115f, -1.f),
    C_Vertex3f(0.281107664f, 0.650187135f, -1.f),
    C_Vertex3f(0.295727253f, 0.579129219f, -1.f),
    C_Vertex3f(0.297977686f, 0.647827029f, -1.f),
    C_Vertex3f(0.309615135f, 0.571344376f, -1.f),
    C_Vertex3f(0.313842297f, 0.644286871f, -1.f),
    C_Vertex3f(0.321511269f, 0.560964584f, -1.f),
    C_Vertex3f(0.328701496f, 0.63956666f, -1.f),
    C_Vertex3f(0.330433369f, 0.54979229f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.23960495f, 0.651367188f, -1.f),
    C_Vertex3f(0.212649345f, 0.648309231f, -1.f),
    C_Vertex3f(0.214365721f, 0.581505299f, -1.f),
    C_Vertex3f(0.224446177f, 0.584209681f, -1.f),
    C_Vertex3f(0.235059381f, 0.586012602f, -1.f),
    C_Vertex3f(0.24620533f, 0.586914063f, -1.f),
    C_Vertex3f(0.263232231f, 0.651367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.212649345f, 0.648309231f, -1.f),
    C_Vertex3f(0.187111855f, 0.642193317f, -1.f),
    C_Vertex3f(0.187320828f, 0.567983389f, -1.f),
    C_Vertex3f(0.195803046f, 0.573392153f, -1.f),
    C_Vertex3f(0.20481801f, 0.577899456f, -1.f),
    C_Vertex3f(0.214365721f, 0.581505299f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.187111855f, 0.642193317f, -1.f),
    C_Vertex3f(0.162992477f, 0.633019447f, -1.f),
    C_Vertex3f(0.165464997f, 0.545146704f, -1.f),
    C_Vertex3f(0.171954632f, 0.554461479f, -1.f),
    C_Vertex3f(0.179371357f, 0.561673164f, -1.f),
    C_Vertex3f(0.187320828f, 0.567983389f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162992477f, 0.633019447f, -1.f),
    C_Vertex3f(0.14596653f, 0.623845577f, -1.f),
    C_Vertex3f(0.148777366f, 0.486856699f, -1.f),
    C_Vertex3f(0.151558638f, 0.504583836f, -1.f),
    C_Vertex3f(0.155267f, 0.520207882f, -1.f),
    C_Vertex3f(0.159902453f, 0.533728838f, -1.f),
    C_Vertex3f(0.165464997f, 0.545146704f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14596653f, 0.623845577f, -1.f),
    C_Vertex3f(0.135427594f, 0.617149353f, -1.f),
    C_Vertex3f(0.145996094f, 0.f, -1.f),
    C_Vertex3f(0.145996094f, 0.445093155f, -1.f),
    C_Vertex3f(0.146923184f, 0.467026472f, -1.f),
    C_Vertex3f(0.148777366f, 0.486856699f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145996094f, 0.f, -1.f),
    C_Vertex3f(0.135427594f, 0.617149353f, -1.f),
    C_Vertex3f(0.125700355f, 0.609872818f, -1.f),
    C_Vertex3f(0.116784811f, 0.602015972f, -1.f),
    C_Vertex3f(0.108680964f, 0.593578815f, -1.f),
    C_Vertex3f(0.101388812f, 0.584561348f, -1.f),
    C_Vertex3f(0.0949083567f, 0.57496357f, -1.f),
    C_Vertex3f(0.0892395973f, 0.56478548f, -1.f),
    C_Vertex3f(0.084382534f, 0.554027081f, -1.f),
    C_Vertex3f(0.0801326036f, 0.542115331f, -1.f),
    C_Vertex3f(0.0764898062f, 0.529050231f, -1.f),
    C_Vertex3f(0.0734541416f, 0.514831781f, -1.f),
    C_Vertex3f(0.07102561f, 0.499459982f, -1.f),
    C_Vertex3f(0.0692042112f, 0.482934833f, -1.f),
    C_Vertex3f(0.0679899454f, 0.465256333f, -1.f),
    C_Vertex3f(0.0673828125f, 0.446424484f, -1.f),
    C_Vertex3f(0.0673828125f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.51058507f, 0.0865750313f, -1.f),
    C_Vertex3f(0.514814019f, 0.100110769f, -1.f),
    C_Vertex3f(0.515545845f, 0.169265747f, -1.f),
    C_Vertex3f(0.517633319f, 0.114192247f, -1.f),
    C_Vertex3f(0.519042969f, 0.14786911f, -1.f),
    C_Vertex3f(0.519042969f, 0.128819466f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.489440322f, 0.0492422581f, -1.f),
    C_Vertex3f(0.497898221f, 0.0611407757f, -1.f),
    C_Vertex3f(0.498060226f, 0.20746994f, -1.f),
    C_Vertex3f(0.50494647f, 0.0735850334f, -1.f),
    C_Vertex3f(0.508551598f, 0.18913269f, -1.f),
    C_Vertex3f(0.51058507f, 0.0865750313f, -1.f),
    C_Vertex3f(0.515545845f, 0.169265747f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.498060226f, 0.20746994f, -1.f),
    C_Vertex3f(0.488606453f, 0.218821526f, -1.f),
    C_Vertex3f(0.489440322f, 0.0492422581f, -1.f),
    C_Vertex3f(0.482442856f, 0.225193262f, -1.f),
    C_Vertex3f(0.479572773f, 0.0378894806f, -1.f),
    C_Vertex3f(0.475034714f, 0.232041121f, -1.f),
    C_Vertex3f(0.468295574f, 0.0270824432f, -1.f),
    C_Vertex3f(0.466382027f, 0.239365101f, -1.f),
    C_Vertex3f(0.456484795f, 0.247165203f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.468295574f, 0.0270824432f, -1.f),
    C_Vertex3f(0.456484795f, 0.247165203f, -1.f),
    C_Vertex3f(0.456246972f, 0.0176262856f, -1.f),
    C_Vertex3f(0.445343018f, 0.255441427f, -1.f),
    C_Vertex3f(0.443426967f, 0.00952100754f, -1.f),
    C_Vertex3f(0.438142776f, 0.135936737f, -1.f),
    C_Vertex3f(0.437988281f, 0.129814148f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.423480988f, 0.173220634f, -1.f),
    C_Vertex3f(0.428117752f, 0.167425156f, -1.f),
    C_Vertex3f(0.432956696f, 0.264193773f, -1.f),
    C_Vertex3f(0.43328476f, 0.157803535f, -1.f),
    C_Vertex3f(0.445343018f, 0.255441427f, -1.f),
    C_Vertex3f(0.436626434f, 0.147307396f, -1.f),
    C_Vertex3f(0.438142776f, 0.135936737f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.410932541f, 0.184750557f, -1.f),
    C_Vertex3f(0.41606617f, 0.180407524f, -1.f),
    C_Vertex3f(0.419325829f, 0.273422241f, -1.f),
    C_Vertex3f(0.420248985f, 0.176564217f, -1.f),
    C_Vertex3f(0.432956696f, 0.264193773f, -1.f),
    C_Vertex3f(0.423480988f, 0.173220634f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.406430244f, 0.281835556f, -1.f),
    C_Vertex3f(0.39659214f, 0.288504362f, -1.f),
    C_Vertex3f(0.397812843f, 0.194935799f, -1.f),
    C_Vertex3f(0.404848099f, 0.189593315f, -1.f),
    C_Vertex3f(0.410932541f, 0.184750557f, -1.f),
    C_Vertex3f(0.419325829f, 0.273422241f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.397812843f, 0.194935799f, -1.f),
    C_Vertex3f(0.39659214f, 0.288504362f, -1.f),
    C_Vertex3f(0.389826775f, 0.200778008f, -1.f),
    C_Vertex3f(0.387831688f, 0.294719934f, -1.f),
    C_Vertex3f(0.380889893f, 0.207119942f, -1.f),
    C_Vertex3f(0.380148888f, 0.300482273f, -1.f),
    C_Vertex3f(0.375946045f, 0.210540771f, -1.f),
    C_Vertex3f(0.373543739f, 0.305791378f, -1.f),
    C_Vertex3f(0.368016243f, 0.310647249f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.366839409f, 0.216723442f, -1.f),
    C_Vertex3f(0.375946045f, 0.210540771f, -1.f),
    C_Vertex3f(0.368016243f, 0.310647249f, -1.f),
    C_Vertex3f(0.363566399f, 0.315049887f, -1.f),
    C_Vertex3f(0.360194206f, 0.31899929f, -1.f),
    C_Vertex3f(0.359046936f, 0.320747375f, -1.f),
    C_Vertex3f(0.356641769f, 0.324134827f, -1.f),
    C_Vertex3f(0.353056908f, 0.330964088f, -1.f),
    C_Vertex3f(0.350697517f, 0.33784771f, -1.f),
    C_Vertex3f(0.35036993f, 0.228321552f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.33937645f, 0.514579773f, -1.f),
    C_Vertex3f(0.338532925f, 0.506602764f, -1.f),
    C_Vertex3f(0.35036993f, 0.228321552f, -1.f),
    C_Vertex3f(0.336803913f, 0.498079777f, -1.f),
    C_Vertex3f(0.335644245f, 0.239152431f, -1.f),
    C_Vertex3f(0.334189415f, 0.489010811f, -1.f),
    C_Vertex3f(0.33068943f, 0.479395866f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.288172722f, 0.281798363f, -1.f),
    C_Vertex3f(0.29417944f, 0.274803638f, -1.f),
    C_Vertex3f(0.297272682f, 0.416898727f, -1.f),
    C_Vertex3f(0.301929951f, 0.267041683f, -1.f),
    C_Vertex3f(0.308539391f, 0.436128616f, -1.f),
    C_Vertex3f(0.311424255f, 0.258512497f, -1.f),
    C_Vertex3f(0.317864418f, 0.452954769f, -1.f),
    C_Vertex3f(0.322662354f, 0.24921608f, -1.f),
    C_Vertex3f(0.325247765f, 0.467377186f, -1.f),
    C_Vertex3f(0.335644245f, 0.239152431f, -1.f),
    C_Vertex3f(0.33068943f, 0.479395866f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.28808403f, 0.400290966f, -1.f),
    C_Vertex3f(0.280973434f, 0.386305332f, -1.f),
    C_Vertex3f(0.281669617f, 0.291139603f, -1.f),
    C_Vertex3f(0.28604126f, 0.284912109f, -1.f),
    C_Vertex3f(0.288172722f, 0.281798363f, -1.f),
    C_Vertex3f(0.297272682f, 0.416898727f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268554688f, 0.344566345f, -1.f),
    C_Vertex3f(0.270740509f, 0.316541672f, -1.f),
    C_Vertex3f(0.271509171f, 0.361829758f, -1.f),
    C_Vertex3f(0.275112152f, 0.303758621f, -1.f),
    C_Vertex3f(0.275940895f, 0.374941826f, -1.f),
    C_Vertex3f(0.281669617f, 0.291139603f, -1.f),
    C_Vertex3f(0.280973434f, 0.386305332f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.437988281f, 0.129814148f, -1.f),
    C_Vertex3f(0.437988281f, 0.122159958f, -1.f),
    C_Vertex3f(0.443426967f, 0.00952100754f, -1.f),
    C_Vertex3f(0.435043812f, 0.10783577f, -1.f),
    C_Vertex3f(0.429835558f, 0.00276660919f, -1.f),
    C_Vertex3f(0.429154873f, 0.0944957733f, -1.f),
    C_Vertex3f(0.420321465f, 0.0821399689f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.429835558f, 0.00276660919f, -1.f),
    C_Vertex3f(0.420321465f, 0.0821399689f, -1.f),
    C_Vertex3f(0.415472746f, -0.00263690948f, -1.f),
    C_Vertex3f(0.408543587f, 0.0707683563f, -1.f),
    C_Vertex3f(0.400338531f, -0.00668954849f, -1.f),
    C_Vertex3f(0.395566463f, 0.0622396469f, -1.f),
    C_Vertex3f(0.384432912f, -0.00939130783f, -1.f),
    C_Vertex3f(0.381390095f, 0.0565538406f, -1.f),
    C_Vertex3f(0.36775589f, -0.0107421875f, -1.f),
    C_Vertex3f(0.366014481f, 0.0537109375f, -1.f),
    C_Vertex3f(0.352161407f, -0.0107421875f, -1.f),
    C_Vertex3f(0.351728439f, 0.0537109375f, -1.f),
    C_Vertex3f(0.340003967f, 0.0556778908f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.352161407f, -0.0107421875f, -1.f),
    C_Vertex3f(0.340003967f, 0.0556778908f, -1.f),
    C_Vertex3f(0.338843346f, -0.00982224941f, -1.f),
    C_Vertex3f(0.328552246f, 0.0596117973f, -1.f),
    C_Vertex3f(0.325947762f, -0.00798237324f, -1.f),
    C_Vertex3f(0.317373276f, 0.0655126572f, -1.f),
    C_Vertex3f(0.313474655f, -0.00522255898f, -1.f),
    C_Vertex3f(0.306467056f, 0.0733804703f, -1.f),
    C_Vertex3f(0.301424026f, -0.00154280663f, -1.f),
    C_Vertex3f(0.295593262f, 0.0843615532f, -1.f),
    C_Vertex3f(0.289795876f, 0.00305688381f, -1.f),
    C_Vertex3f(0.284751892f, 0.0984559059f, -1.f),
    C_Vertex3f(0.278590202f, 0.00857651234f, -1.f),
    C_Vertex3f(0.273942947f, 0.115663528f, -1.f),
    C_Vertex3f(0.268554688f, 0.125823975f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.268554688f, 0.125823975f, -1.f),
    C_Vertex3f(0.200683594f, 0.0943603516f, -1.f),
    C_Vertex3f(0.20580864f, 0.0828037262f, -1.f),
    C_Vertex3f(0.218676567f, 0.0617890358f, -1.f),
    C_Vertex3f(0.234162331f, 0.0428729057f, -1.f),
    C_Vertex3f(0.25226593f, 0.026055336f, -1.f),
    C_Vertex3f(0.267807007f, 0.0150160789f, -1.f),
    C_Vertex3f(0.278590202f, 0.00857651234f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.360554218f, 0.385541916f, -1.f),
    C_Vertex3f(0.366899014f, 0.397779465f, -1.f),
    C_Vertex3f(0.367246628f, 0.61832571f, -1.f),
    C_Vertex3f(0.375102997f, 0.412693977f, -1.f),
    C_Vertex3f(0.378084183f, 0.608885288f, -1.f),
    C_Vertex3f(0.385166168f, 0.430285454f, -1.f),
    C_Vertex3f(0.387567043f, 0.599069953f, -1.f),
    C_Vertex3f(0.393698215f, 0.446425438f, -1.f),
    C_Vertex3f(0.39569521f, 0.588879704f, -1.f),
    C_Vertex3f(0.400699139f, 0.46111393f, -1.f),
    C_Vertex3f(0.402468681f, 0.578314543f, -1.f),
    C_Vertex3f(0.406168938f, 0.474350929f, -1.f),
    C_Vertex3f(0.407887459f, 0.567374468f, -1.f),
    C_Vertex3f(0.410107613f, 0.486136436f, -1.f),
    C_Vertex3f(0.411951542f, 0.55605948f, -1.f),
    C_Vertex3f(0.413061619f, 0.497759342f, -1.f),
    C_Vertex3f(0.414660931f, 0.544369578f, -1.f),
    C_Vertex3f(0.415030956f, 0.509219646f, -1.f),
    C_Vertex3f(0.416015625f, 0.532304764f, -1.f),
    C_Vertex3f(0.416015625f, 0.520517349f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.349555969f, 0.353527069f, -1.f),
    C_Vertex3f(0.352128983f, 0.365655899f, -1.f),
    C_Vertex3f(0.355403662f, 0.62658608f, -1.f),
    C_Vertex3f(0.356068611f, 0.375981331f, -1.f),
    C_Vertex3f(0.367246628f, 0.61832571f, -1.f),
    C_Vertex3f(0.360554218f, 0.385541916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.349609375f, 0.34828186f, -1.f),
    C_Vertex3f(0.349555969f, 0.353527069f, -1.f),
    C_Vertex3f(0.349563599f, 0.34478569f, -1.f),
    C_Vertex3f(0.270740509f, 0.316541672f, -1.f),
    C_Vertex3f(0.268554688f, 0.344566345f, -1.f),
    C_Vertex3f(0.268554688f, 0.329488754f, -1.f),
    C_Vertex3f(0.349563599f, 0.34478569f, -1.f),
    C_Vertex3f(0.35036993f, 0.228321552f, -1.f),
    C_Vertex3f(0.350697517f, 0.33784771f, -1.f),
    C_End(),
    C_Translatef(0.546875f, 0.f, 0.f),
};
static const Call char_224[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.644008815f, -0.765018106f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Normal3f(0.609832764f, -0.792530119f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Normal3f(0.560195386f, -0.828360498f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Normal3f(0.488698423f, -0.872452796f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Normal3f(0.402021319f, -0.915630281f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Normal3f(0.307879895f, -0.951425254f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Normal3f(0.217794061f, -0.975994766f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Normal3f(0.128345996f, -0.991729438f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Normal3f(0.0436385125f, -0.999047339f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0297879688f, -0.999556243f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0997409225f, -0.995013535f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Normal3f(-0.175596207f, -0.984462261f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Normal3f(-0.2587111f, -0.965954781f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Normal3f(-0.348068923f, -0.937468946f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Normal3f(-0.441677898f, -0.897173643f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Normal3f(-0.536508918f, -0.84389472f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Normal3f(-0.628730416f, -0.777623355f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Normal3f(-0.708205581f, -0.706006348f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Normal3f(-0.777196348f, -0.629258096f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Normal3f(-0.840526164f, -0.541770935f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Normal3f(-0.895173013f, -0.445718706f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Normal3f(-0.93884182f, -0.344348639f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Normal3f(-0.970410466f, -0.241460919f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Normal3f(-0.990046978f, -0.140737399f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Normal3f(-0.999063611f, -0.043263685f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_Normal3f(-0.998176992f, 0.0603547506f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Normal3f(-0.983728349f, 0.179661945f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Normal3f(-0.952656448f, 0.304048836f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Normal3f(-0.90515995f, 0.425071001f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Normal3f(-0.848285139f, 0.529539704f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Normal3f(-0.782067299f, 0.62319386f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Normal3f(-0.701901376f, 0.712274075f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Normal3f(-0.610135138f, 0.792297363f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Normal3f(-0.525150001f, 0.851009607f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Normal3f(-0.451541305f, 0.89225018f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Normal3f(-0.377094716f, 0.926174641f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Normal3f(-0.315602541f, 0.948891461f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Normal3f(-0.262250215f, 0.964999855f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Normal3f(-0.217653468f, 0.976026177f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Normal3f(-0.182780772f, 0.983153641f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Normal3f(-0.152309522f, 0.988332868f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Normal3f(-0.129105836f, 0.991630852f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Normal3f(-0.125778303f, 0.992058396f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Normal3f(-0.140358314f, 0.990100741f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Normal3f(-0.157823548f, 0.987467349f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Normal3f(-0.178095639f, 0.98401314f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Normal3f(-0.20187664f, 0.979411006f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Normal3f(-0.230107605f, 0.973165154f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Normal3f(-0.264068544f, 0.964503944f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Normal3f(-0.297998816f, 0.954566181f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Normal3f(-0.724053323f, 0.689743996f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Normal3f(-0.999676347f, 0.0254404638f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Normal3f(-0.999733686f, 0.0230763834f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Normal3f(-0.999828994f, 0.0184933227f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Normal3f(-0.999987364f, 0.00503883278f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Normal3f(-0.999277949f, -0.0379956067f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Normal3f(-0.995299578f, -0.0968433693f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Normal3f(-0.98471868f, -0.174152732f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Normal3f(-0.964879274f, -0.262693614f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Normal3f(-0.932045639f, -0.362340957f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Normal3f(-0.882021129f, -0.471209705f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Normal3f(-0.811133981f, -0.584860384f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Normal3f(-0.713830113f, -0.700318813f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Normal3f(-0.61406064f, -0.789258838f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Normal3f(-0.521080196f, -0.853507817f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Normal3f(-0.424242347f, -0.905548692f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Normal3f(-0.330907196f, -0.943663299f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Normal3f(-0.243749335f, -0.969838262f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Normal3f(-0.164196789f, -0.986427605f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Normal3f(-0.0926971957f, -0.995694339f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Normal3f(-0.0295525081f, -0.999563217f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Normal3f(0.0223024357f, -0.99975127f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Normal3f(0.0729011372f, -0.997339189f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Normal3f(0.130242616f, -0.991482139f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Normal3f(0.195666403f, -0.980670512f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Normal3f(0.269929171f, -0.962880194f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Normal3f(0.35332042f, -0.93550241f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Normal3f(0.44521296f, -0.895424783f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Normal3f(0.543500543f, -0.839408755f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Normal3f(0.641533256f, -0.767095149f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Normal3f(0.72543782f, -0.688287675f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Normal3f(0.793855846f, -0.608106017f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Normal3f(0.847895145f, -0.530164063f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Normal3f(0.889552832f, -0.45683226f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Normal3f(0.921084702f, -0.389362156f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Normal3f(0.944615781f, -0.328178316f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Normal3f(0.959435046f, -0.281929553f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Normal3f(0.968791068f, -0.247878939f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.131752744f, -0.991282582f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.978250861f, 0.207425416f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Normal3f(-0.964124143f, 0.265451789f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Normal3f(-0.93853426f, 0.345186114f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Normal3f(-0.892804325f, 0.450444669f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Normal3f(-0.82656163f, 0.562846184f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Normal3f(-0.731744051f, 0.68157953f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Normal3f(-0.617528915f, 0.786548138f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Normal3f(-0.501185894f, 0.865339637f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Normal3f(-0.390982777f, 0.920397997f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Normal3f(-0.290884405f, 0.956758201f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Normal3f(-0.197783709f, 0.98024565f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Normal3f(-0.11233186f, 0.993670762f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Normal3f(-0.0353869088f, 0.999373674f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Normal3f(0.0308220927f, 0.999524832f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Normal3f(0.102260895f, 0.994757652f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Normal3f(0.188853115f, 0.982005298f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Normal3f(0.293315113f, 0.956015766f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Normal3f(0.394779444f, 0.918775976f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Normal3f(0.489459455f, 0.872026086f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Normal3f(0.597415328f, 0.801932037f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Normal3f(0.71381259f, 0.700336814f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Normal3f(0.813518882f, 0.581538498f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Normal3f(0.882400215f, 0.470499665f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Normal3f(0.93244344f, 0.361315906f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Normal3f(0.96134758f, 0.275337785f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Normal3f(0.978440464f, 0.206529349f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Normal3f(0.987760723f, 0.155976713f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Normal3f(0.992642164f, 0.121084802f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Normal3f(0.996025085f, 0.0890726f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Normal3f(0.997923374f, 0.0644121915f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Normal3f(0.998993039f, 0.0448651426f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Normal3f(0.999579132f, 0.0290074293f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Normal3f(0.999873698f, 0.015893871f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Normal3f(0.999999583f, 0.000895282777f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Normal3f(0.999999404f, 0.00104705233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Normal3f(0.999950409f, 0.00996590592f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Normal3f(0.999827266f, 0.0185867213f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Normal3f(0.999563813f, 0.0295336749f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Normal3f(0.999036491f, 0.0438889116f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Normal3f(0.997980356f, 0.0635225922f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Normal3f(0.995762765f, 0.0919589549f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Normal3f(0.98978281f, 0.142583564f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Normal3f(0.975923538f, 0.218112797f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Normal3f(0.955474317f, 0.295074373f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Normal3f(0.927006185f, 0.375045985f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Normal3f(0.898835897f, 0.438285291f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Normal3f(0.873886287f, 0.486130446f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.895264685f, -0.445534706f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Normal3f(-0.923641562f, -0.383257538f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Normal3f(-0.95153141f, -0.307551742f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Normal3f(-0.975133479f, -0.221618384f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Normal3f(-0.987897873f, -0.155105904f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Normal3f(-0.994414866f, -0.105542049f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.377323061f, -0.926081717f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Normal3f(0.344282955f, -0.93886584f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Normal3f(0.30806917f, -0.951363921f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Normal3f(0.27083087f, -0.962626934f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Normal3f(0.23911649f, -0.970990896f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Normal3f(0.211867765f, -0.977298319f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Normal3f(0.188257366f, -0.982119739f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Normal3f(0.167636946f, -0.985848784f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Normal3f(0.149529129f, -0.988757372f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Normal3f(0.158242494f, -0.987400293f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Normal3f(0.197629482f, -0.980276704f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Normal3f(0.253822058f, -0.967250884f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Normal3f(0.338508695f, -0.940963209f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Normal3f(0.456908435f, -0.889513731f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Normal3f(0.576411605f, -0.817159474f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Normal3f(0.691899657f, -0.721993685f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Normal3f(0.794279575f, -0.607552409f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Normal3f(0.874713659f, -0.484639972f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Normal3f(0.936157763f, -0.351580352f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Normal3f(0.977780998f, -0.209628955f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Normal3f(0.998205423f, -0.0598839708f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Normal3f(0.99415952f, 0.107920907f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Normal3f(0.951207876f, 0.308550656f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Normal3f(0.858417392f, 0.51295191f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Normal3f(0.730436027f, 0.682981133f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Normal3f(0.568286002f, 0.822831035f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Normal3f(0.378173947f, 0.92573458f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Normal3f(0.204722017f, 0.978820205f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Normal3f(0.0608421154f, 0.998147368f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Normal3f(-0.056054458f, 0.998427689f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Normal3f(-0.176223293f, 0.984350204f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Normal3f(-0.302620173f, 0.953111231f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Normal3f(-0.429032564f, 0.90328896f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Normal3f(-0.549294412f, 0.835628867f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Normal3f(-0.66699791f, 0.74505949f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Normal3f(-0.775314748f, 0.631575108f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Normal3f(-0.863186479f, 0.504885197f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Normal3f(-0.934871018f, 0.354987651f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_Normal3f(-0.976849318f, 0.213928372f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Normal3f(-0.993807793f, 0.111112691f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Normal3f(-0.999024689f, 0.0441556387f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.322753906f, 0.521972656f, 0.f),
    C_Vertex3f(0.322753906f, 0.521972656f, -1.f),
    C_Vertex3f(0.259277344f, 0.521972656f, -1.f),
    C_Vertex3f(0.259277344f, 0.521972656f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.259277344f, 0.521972656f, 0.f),
    C_Vertex3f(0.259277344f, 0.521972656f, -1.f),
    C_Vertex3f(0.158203125f, 0.64453125f, -1.f),
    C_Vertex3f(0.158203125f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.158203125f, 0.64453125f, 0.f),
    C_Vertex3f(0.158203125f, 0.64453125f, -1.f),
    C_Vertex3f(0.263671875f, 0.64453125f, -1.f),
    C_Vertex3f(0.263671875f, 0.64453125f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.263671875f, 0.64453125f, 0.f),
    C_Vertex3f(0.263671875f, 0.64453125f, -1.f),
    C_Vertex3f(0.322753906f, 0.521972656f, -1.f),
    C_Vertex3f(0.322753906f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.263671875f, 0.64453125f, 0.f),
    C_Vertex3f(0.158203125f, 0.64453125f, 0.f),
    C_Vertex3f(0.259277344f, 0.521972656f, 0.f),
    C_Vertex3f(0.322753906f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.263671875f, 0.64453125f, -1.f),
    C_Vertex3f(0.158203125f, 0.64453125f, -1.f),
    C_Vertex3f(0.259277344f, 0.521972656f, -1.f),
    C_Vertex3f(0.322753906f, 0.521972656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_225[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.644008815f, -0.765018106f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Normal3f(0.609832764f, -0.792530119f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Normal3f(0.560195386f, -0.828360498f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Normal3f(0.488698423f, -0.872452796f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Normal3f(0.402021319f, -0.915630281f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Normal3f(0.307879895f, -0.951425254f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Normal3f(0.217794061f, -0.975994766f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Normal3f(0.128345996f, -0.991729438f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Normal3f(0.0436385125f, -0.999047339f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0297879688f, -0.999556243f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0997409225f, -0.995013535f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Normal3f(-0.175596207f, -0.984462261f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Normal3f(-0.2587111f, -0.965954781f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Normal3f(-0.348068923f, -0.937468946f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Normal3f(-0.441677898f, -0.897173643f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Normal3f(-0.536508918f, -0.84389472f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Normal3f(-0.628730416f, -0.777623355f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Normal3f(-0.708205581f, -0.706006348f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Normal3f(-0.777196348f, -0.629258096f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Normal3f(-0.840526164f, -0.541770935f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Normal3f(-0.895173013f, -0.445718706f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Normal3f(-0.93884182f, -0.344348639f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Normal3f(-0.970410466f, -0.241460919f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Normal3f(-0.990046978f, -0.140737399f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Normal3f(-0.999063611f, -0.043263685f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_Normal3f(-0.998176992f, 0.0603547506f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Normal3f(-0.983728349f, 0.179661945f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Normal3f(-0.952656448f, 0.304048836f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Normal3f(-0.90515995f, 0.425071001f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Normal3f(-0.848285139f, 0.529539704f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Normal3f(-0.782067299f, 0.62319386f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Normal3f(-0.701901376f, 0.712274075f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Normal3f(-0.610135138f, 0.792297363f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Normal3f(-0.525150001f, 0.851009607f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Normal3f(-0.451541305f, 0.89225018f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Normal3f(-0.377094716f, 0.926174641f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Normal3f(-0.315602541f, 0.948891461f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Normal3f(-0.262250215f, 0.964999855f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Normal3f(-0.217653468f, 0.976026177f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Normal3f(-0.182780772f, 0.983153641f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Normal3f(-0.152309522f, 0.988332868f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Normal3f(-0.129105836f, 0.991630852f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Normal3f(-0.125778303f, 0.992058396f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Normal3f(-0.140358314f, 0.990100741f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Normal3f(-0.157823548f, 0.987467349f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Normal3f(-0.178095639f, 0.98401314f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Normal3f(-0.20187664f, 0.979411006f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Normal3f(-0.230107605f, 0.973165154f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Normal3f(-0.264068544f, 0.964503944f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Normal3f(-0.297998816f, 0.954566181f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Normal3f(-0.724053323f, 0.689743996f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Normal3f(-0.999676347f, 0.0254404638f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Normal3f(-0.999733686f, 0.0230763834f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Normal3f(-0.999828994f, 0.0184933227f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Normal3f(-0.999987364f, 0.00503883278f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Normal3f(-0.999277949f, -0.0379956067f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Normal3f(-0.995299578f, -0.0968433693f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Normal3f(-0.98471868f, -0.174152732f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Normal3f(-0.964879274f, -0.262693614f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Normal3f(-0.932045639f, -0.362340957f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Normal3f(-0.882021129f, -0.471209705f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Normal3f(-0.811133981f, -0.584860384f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Normal3f(-0.713830113f, -0.700318813f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Normal3f(-0.61406064f, -0.789258838f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Normal3f(-0.521080196f, -0.853507817f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Normal3f(-0.424242347f, -0.905548692f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Normal3f(-0.330907196f, -0.943663299f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Normal3f(-0.243749335f, -0.969838262f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Normal3f(-0.164196789f, -0.986427605f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Normal3f(-0.0926971957f, -0.995694339f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Normal3f(-0.0295525081f, -0.999563217f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Normal3f(0.0223024357f, -0.99975127f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Normal3f(0.0729011372f, -0.997339189f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Normal3f(0.130242616f, -0.991482139f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Normal3f(0.195666403f, -0.980670512f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Normal3f(0.269929171f, -0.962880194f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Normal3f(0.35332042f, -0.93550241f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Normal3f(0.44521296f, -0.895424783f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Normal3f(0.543500543f, -0.839408755f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Normal3f(0.641533256f, -0.767095149f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Normal3f(0.72543782f, -0.688287675f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Normal3f(0.793855846f, -0.608106017f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Normal3f(0.847895145f, -0.530164063f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Normal3f(0.889552832f, -0.45683226f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Normal3f(0.921084702f, -0.389362156f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Normal3f(0.944615781f, -0.328178316f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Normal3f(0.959435046f, -0.281929553f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Normal3f(0.968791068f, -0.247878939f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.131752744f, -0.991282582f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.978250861f, 0.207425416f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Normal3f(-0.964124143f, 0.265451789f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Normal3f(-0.93853426f, 0.345186114f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Normal3f(-0.892804325f, 0.450444669f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Normal3f(-0.82656163f, 0.562846184f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Normal3f(-0.731744051f, 0.68157953f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Normal3f(-0.617528915f, 0.786548138f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Normal3f(-0.501185894f, 0.865339637f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Normal3f(-0.390982777f, 0.920397997f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Normal3f(-0.290884405f, 0.956758201f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Normal3f(-0.197783709f, 0.98024565f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Normal3f(-0.11233186f, 0.993670762f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Normal3f(-0.0353869088f, 0.999373674f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Normal3f(0.0308220927f, 0.999524832f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Normal3f(0.102260895f, 0.994757652f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Normal3f(0.188853115f, 0.982005298f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Normal3f(0.293315113f, 0.956015766f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Normal3f(0.394779444f, 0.918775976f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Normal3f(0.489459455f, 0.872026086f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Normal3f(0.597415328f, 0.801932037f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Normal3f(0.71381259f, 0.700336814f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Normal3f(0.813518882f, 0.581538498f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Normal3f(0.882400215f, 0.470499665f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Normal3f(0.93244344f, 0.361315906f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Normal3f(0.96134758f, 0.275337785f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Normal3f(0.978440464f, 0.206529349f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Normal3f(0.987760723f, 0.155976713f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Normal3f(0.992642164f, 0.121084802f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Normal3f(0.996025085f, 0.0890726f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Normal3f(0.997923374f, 0.0644121915f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Normal3f(0.998993039f, 0.0448651426f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Normal3f(0.999579132f, 0.0290074293f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Normal3f(0.999873698f, 0.015893871f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Normal3f(0.999999583f, 0.000895282777f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Normal3f(0.999999404f, 0.00104705233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Normal3f(0.999950409f, 0.00996590592f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Normal3f(0.999827266f, 0.0185867213f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Normal3f(0.999563813f, 0.0295336749f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Normal3f(0.999036491f, 0.0438889116f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Normal3f(0.997980356f, 0.0635225922f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Normal3f(0.995762765f, 0.0919589549f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Normal3f(0.98978281f, 0.142583564f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Normal3f(0.975923538f, 0.218112797f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Normal3f(0.955474317f, 0.295074373f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Normal3f(0.927006185f, 0.375045985f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Normal3f(0.898835897f, 0.438285291f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Normal3f(0.873886287f, 0.486130446f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.895264685f, -0.445534706f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Normal3f(-0.923641562f, -0.383257538f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Normal3f(-0.95153141f, -0.307551742f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Normal3f(-0.975133479f, -0.221618384f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Normal3f(-0.987897873f, -0.155105904f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Normal3f(-0.994414866f, -0.105542049f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.377323061f, -0.926081717f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Normal3f(0.344282955f, -0.93886584f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Normal3f(0.30806917f, -0.951363921f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Normal3f(0.27083087f, -0.962626934f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Normal3f(0.23911649f, -0.970990896f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Normal3f(0.211867765f, -0.977298319f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Normal3f(0.188257366f, -0.982119739f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Normal3f(0.167636946f, -0.985848784f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Normal3f(0.149529129f, -0.988757372f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Normal3f(0.158242494f, -0.987400293f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Normal3f(0.197629482f, -0.980276704f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Normal3f(0.253822058f, -0.967250884f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Normal3f(0.338508695f, -0.940963209f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Normal3f(0.456908435f, -0.889513731f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Normal3f(0.576411605f, -0.817159474f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Normal3f(0.691899657f, -0.721993685f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Normal3f(0.794279575f, -0.607552409f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Normal3f(0.874713659f, -0.484639972f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Normal3f(0.936157763f, -0.351580352f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Normal3f(0.977780998f, -0.209628955f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Normal3f(0.998205423f, -0.0598839708f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Normal3f(0.99415952f, 0.107920907f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Normal3f(0.951207876f, 0.308550656f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Normal3f(0.858417392f, 0.51295191f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Normal3f(0.730436027f, 0.682981133f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Normal3f(0.568286002f, 0.822831035f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Normal3f(0.378173947f, 0.92573458f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Normal3f(0.204722017f, 0.978820205f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Normal3f(0.0608421154f, 0.998147368f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Normal3f(-0.056054458f, 0.998427689f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Normal3f(-0.176223293f, 0.984350204f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Normal3f(-0.302620173f, 0.953111231f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Normal3f(-0.429032564f, 0.90328896f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Normal3f(-0.549294412f, 0.835628867f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Normal3f(-0.66699791f, 0.74505949f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Normal3f(-0.775314748f, 0.631575108f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Normal3f(-0.863186479f, 0.504885197f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Normal3f(-0.934871018f, 0.354987651f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_Normal3f(-0.976849318f, 0.213928372f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Normal3f(-0.993807793f, 0.111112691f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Normal3f(-0.999024689f, 0.0441556387f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.189941406f, 0.521972656f, 0.f),
    C_Vertex3f(0.189941406f, 0.521972656f, -1.f),
    C_Vertex3f(0.247558594f, 0.64453125f, -1.f),
    C_Vertex3f(0.247558594f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.247558594f, 0.64453125f, 0.f),
    C_Vertex3f(0.247558594f, 0.64453125f, -1.f),
    C_Vertex3f(0.351074219f, 0.64453125f, -1.f),
    C_Vertex3f(0.351074219f, 0.64453125f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.351074219f, 0.64453125f, 0.f),
    C_Vertex3f(0.351074219f, 0.64453125f, -1.f),
    C_Vertex3f(0.254882813f, 0.521972656f, -1.f),
    C_Vertex3f(0.254882813f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.254882813f, 0.521972656f, 0.f),
    C_Vertex3f(0.254882813f, 0.521972656f, -1.f),
    C_Vertex3f(0.189941406f, 0.521972656f, -1.f),
    C_Vertex3f(0.189941406f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.247558594f, 0.64453125f, 0.f),
    C_Vertex3f(0.189941406f, 0.521972656f, 0.f),
    C_Vertex3f(0.254882813f, 0.521972656f, 0.f),
    C_Vertex3f(0.351074219f, 0.64453125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.247558594f, 0.64453125f, -1.f),
    C_Vertex3f(0.189941406f, 0.521972656f, -1.f),
    C_Vertex3f(0.254882813f, 0.521972656f, -1.f),
    C_Vertex3f(0.351074219f, 0.64453125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_226[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.644008815f, -0.765018106f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Normal3f(0.609832764f, -0.792530119f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Normal3f(0.560195386f, -0.828360498f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Normal3f(0.488698423f, -0.872452796f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Normal3f(0.402021319f, -0.915630281f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Normal3f(0.307879895f, -0.951425254f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Normal3f(0.217794061f, -0.975994766f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Normal3f(0.128345996f, -0.991729438f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Normal3f(0.0436385125f, -0.999047339f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0297879688f, -0.999556243f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0997409225f, -0.995013535f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Normal3f(-0.175596207f, -0.984462261f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Normal3f(-0.2587111f, -0.965954781f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Normal3f(-0.348068923f, -0.937468946f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Normal3f(-0.441677898f, -0.897173643f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Normal3f(-0.536508918f, -0.84389472f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Normal3f(-0.628730416f, -0.777623355f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Normal3f(-0.708205581f, -0.706006348f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Normal3f(-0.777196348f, -0.629258096f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Normal3f(-0.840526164f, -0.541770935f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Normal3f(-0.895173013f, -0.445718706f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Normal3f(-0.93884182f, -0.344348639f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Normal3f(-0.970410466f, -0.241460919f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Normal3f(-0.990046978f, -0.140737399f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Normal3f(-0.999063611f, -0.043263685f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_Normal3f(-0.998176992f, 0.0603547506f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Normal3f(-0.983728349f, 0.179661945f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Normal3f(-0.952656448f, 0.304048836f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Normal3f(-0.90515995f, 0.425071001f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Normal3f(-0.848285139f, 0.529539704f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Normal3f(-0.782067299f, 0.62319386f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Normal3f(-0.701901376f, 0.712274075f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Normal3f(-0.610135138f, 0.792297363f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Normal3f(-0.525150001f, 0.851009607f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Normal3f(-0.451541305f, 0.89225018f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Normal3f(-0.377094716f, 0.926174641f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Normal3f(-0.315602541f, 0.948891461f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Normal3f(-0.262250215f, 0.964999855f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Normal3f(-0.217653468f, 0.976026177f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Normal3f(-0.182780772f, 0.983153641f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Normal3f(-0.152309522f, 0.988332868f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Normal3f(-0.129105836f, 0.991630852f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Normal3f(-0.125778303f, 0.992058396f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Normal3f(-0.140358314f, 0.990100741f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Normal3f(-0.157823548f, 0.987467349f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Normal3f(-0.178095639f, 0.98401314f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Normal3f(-0.20187664f, 0.979411006f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Normal3f(-0.230107605f, 0.973165154f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Normal3f(-0.264068544f, 0.964503944f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Normal3f(-0.297998816f, 0.954566181f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Normal3f(-0.724053323f, 0.689743996f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Normal3f(-0.999676347f, 0.0254404638f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Normal3f(-0.999733686f, 0.0230763834f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Normal3f(-0.999828994f, 0.0184933227f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Normal3f(-0.999987364f, 0.00503883278f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Normal3f(-0.999277949f, -0.0379956067f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Normal3f(-0.995299578f, -0.0968433693f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Normal3f(-0.98471868f, -0.174152732f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Normal3f(-0.964879274f, -0.262693614f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Normal3f(-0.932045639f, -0.362340957f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Normal3f(-0.882021129f, -0.471209705f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Normal3f(-0.811133981f, -0.584860384f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Normal3f(-0.713830113f, -0.700318813f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Normal3f(-0.61406064f, -0.789258838f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Normal3f(-0.521080196f, -0.853507817f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Normal3f(-0.424242347f, -0.905548692f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Normal3f(-0.330907196f, -0.943663299f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Normal3f(-0.243749335f, -0.969838262f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Normal3f(-0.164196789f, -0.986427605f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Normal3f(-0.0926971957f, -0.995694339f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Normal3f(-0.0295525081f, -0.999563217f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Normal3f(0.0223024357f, -0.99975127f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Normal3f(0.0729011372f, -0.997339189f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Normal3f(0.130242616f, -0.991482139f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Normal3f(0.195666403f, -0.980670512f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Normal3f(0.269929171f, -0.962880194f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Normal3f(0.35332042f, -0.93550241f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Normal3f(0.44521296f, -0.895424783f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Normal3f(0.543500543f, -0.839408755f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Normal3f(0.641533256f, -0.767095149f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Normal3f(0.72543782f, -0.688287675f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Normal3f(0.793855846f, -0.608106017f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Normal3f(0.847895145f, -0.530164063f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Normal3f(0.889552832f, -0.45683226f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Normal3f(0.921084702f, -0.389362156f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Normal3f(0.944615781f, -0.328178316f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Normal3f(0.959435046f, -0.281929553f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Normal3f(0.968791068f, -0.247878939f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.131752744f, -0.991282582f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.978250861f, 0.207425416f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Normal3f(-0.964124143f, 0.265451789f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Normal3f(-0.93853426f, 0.345186114f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Normal3f(-0.892804325f, 0.450444669f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Normal3f(-0.82656163f, 0.562846184f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Normal3f(-0.731744051f, 0.68157953f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Normal3f(-0.617528915f, 0.786548138f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Normal3f(-0.501185894f, 0.865339637f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Normal3f(-0.390982777f, 0.920397997f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Normal3f(-0.290884405f, 0.956758201f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Normal3f(-0.197783709f, 0.98024565f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Normal3f(-0.11233186f, 0.993670762f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Normal3f(-0.0353869088f, 0.999373674f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Normal3f(0.0308220927f, 0.999524832f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Normal3f(0.102260895f, 0.994757652f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Normal3f(0.188853115f, 0.982005298f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Normal3f(0.293315113f, 0.956015766f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Normal3f(0.394779444f, 0.918775976f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Normal3f(0.489459455f, 0.872026086f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Normal3f(0.597415328f, 0.801932037f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Normal3f(0.71381259f, 0.700336814f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Normal3f(0.813518882f, 0.581538498f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Normal3f(0.882400215f, 0.470499665f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Normal3f(0.93244344f, 0.361315906f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Normal3f(0.96134758f, 0.275337785f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Normal3f(0.978440464f, 0.206529349f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Normal3f(0.987760723f, 0.155976713f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Normal3f(0.992642164f, 0.121084802f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Normal3f(0.996025085f, 0.0890726f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Normal3f(0.997923374f, 0.0644121915f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Normal3f(0.998993039f, 0.0448651426f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Normal3f(0.999579132f, 0.0290074293f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Normal3f(0.999873698f, 0.015893871f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Normal3f(0.999999583f, 0.000895282777f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Normal3f(0.999999404f, 0.00104705233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Normal3f(0.999950409f, 0.00996590592f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Normal3f(0.999827266f, 0.0185867213f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Normal3f(0.999563813f, 0.0295336749f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Normal3f(0.999036491f, 0.0438889116f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Normal3f(0.997980356f, 0.0635225922f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Normal3f(0.995762765f, 0.0919589549f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Normal3f(0.98978281f, 0.142583564f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Normal3f(0.975923538f, 0.218112797f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Normal3f(0.955474317f, 0.295074373f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Normal3f(0.927006185f, 0.375045985f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Normal3f(0.898835897f, 0.438285291f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Normal3f(0.873886287f, 0.486130446f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.895264685f, -0.445534706f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Normal3f(-0.923641562f, -0.383257538f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Normal3f(-0.95153141f, -0.307551742f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Normal3f(-0.975133479f, -0.221618384f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Normal3f(-0.987897873f, -0.155105904f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Normal3f(-0.994414866f, -0.105542049f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.377323061f, -0.926081717f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Normal3f(0.344282955f, -0.93886584f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Normal3f(0.30806917f, -0.951363921f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Normal3f(0.27083087f, -0.962626934f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Normal3f(0.23911649f, -0.970990896f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Normal3f(0.211867765f, -0.977298319f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Normal3f(0.188257366f, -0.982119739f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Normal3f(0.167636946f, -0.985848784f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Normal3f(0.149529129f, -0.988757372f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Normal3f(0.158242494f, -0.987400293f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Normal3f(0.197629482f, -0.980276704f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Normal3f(0.253822058f, -0.967250884f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Normal3f(0.338508695f, -0.940963209f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Normal3f(0.456908435f, -0.889513731f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Normal3f(0.576411605f, -0.817159474f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Normal3f(0.691899657f, -0.721993685f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Normal3f(0.794279575f, -0.607552409f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Normal3f(0.874713659f, -0.484639972f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Normal3f(0.936157763f, -0.351580352f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Normal3f(0.977780998f, -0.209628955f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Normal3f(0.998205423f, -0.0598839708f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Normal3f(0.99415952f, 0.107920907f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Normal3f(0.951207876f, 0.308550656f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Normal3f(0.858417392f, 0.51295191f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Normal3f(0.730436027f, 0.682981133f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Normal3f(0.568286002f, 0.822831035f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Normal3f(0.378173947f, 0.92573458f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Normal3f(0.204722017f, 0.978820205f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Normal3f(0.0608421154f, 0.998147368f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Normal3f(-0.056054458f, 0.998427689f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Normal3f(-0.176223293f, 0.984350204f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Normal3f(-0.302620173f, 0.953111231f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Normal3f(-0.429032564f, 0.90328896f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Normal3f(-0.549294412f, 0.835628867f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Normal3f(-0.66699791f, 0.74505949f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Normal3f(-0.775314748f, 0.631575108f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Normal3f(-0.863186479f, 0.504885197f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Normal3f(-0.934871018f, 0.354987651f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_Normal3f(-0.976849318f, 0.213928372f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Normal3f(-0.993807793f, 0.111112691f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Normal3f(-0.999024689f, 0.0441556387f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.247070313f, 0.596679688f, 0.f),
    C_Vertex3f(0.247070313f, 0.596679688f, -1.f),
    C_Vertex3f(0.197998047f, 0.521972656f, -1.f),
    C_Vertex3f(0.197998047f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.197998047f, 0.521972656f, 0.f),
    C_Vertex3f(0.197998047f, 0.521972656f, -1.f),
    C_Vertex3f(0.108642578f, 0.521972656f, -1.f),
    C_Vertex3f(0.108642578f, 0.521972656f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(0.108642578f, 0.521972656f, 0.f),
    C_Vertex3f(0.108642578f, 0.521972656f, -1.f),
    C_Vertex3f(0.202148438f, 0.64453125f, -1.f),
    C_Vertex3f(0.202148438f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.202148438f, 0.64453125f, 0.f),
    C_Vertex3f(0.202148438f, 0.64453125f, -1.f),
    C_Vertex3f(0.286132813f, 0.64453125f, -1.f),
    C_Vertex3f(0.286132813f, 0.64453125f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.286132813f, 0.64453125f, 0.f),
    C_Vertex3f(0.286132813f, 0.64453125f, -1.f),
    C_Vertex3f(0.384521484f, 0.521972656f, -1.f),
    C_Vertex3f(0.384521484f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.384521484f, 0.521972656f, 0.f),
    C_Vertex3f(0.384521484f, 0.521972656f, -1.f),
    C_Vertex3f(0.295166016f, 0.521972656f, -1.f),
    C_Vertex3f(0.295166016f, 0.521972656f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.295166016f, 0.521972656f, 0.f),
    C_Vertex3f(0.295166016f, 0.521972656f, -1.f),
    C_Vertex3f(0.247070313f, 0.596679688f, -1.f),
    C_Vertex3f(0.247070313f, 0.596679688f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.108642578f, 0.521972656f, 0.f),
    C_Vertex3f(0.197998047f, 0.521972656f, 0.f),
    C_Vertex3f(0.202148438f, 0.64453125f, 0.f),
    C_Vertex3f(0.247070313f, 0.596679688f, 0.f),
    C_Vertex3f(0.286132813f, 0.64453125f, 0.f),
    C_Vertex3f(0.295166016f, 0.521972656f, 0.f),
    C_Vertex3f(0.384521484f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.108642578f, 0.521972656f, -1.f),
    C_Vertex3f(0.197998047f, 0.521972656f, -1.f),
    C_Vertex3f(0.202148438f, 0.64453125f, -1.f),
    C_Vertex3f(0.247070313f, 0.596679688f, -1.f),
    C_Vertex3f(0.286132813f, 0.64453125f, -1.f),
    C_Vertex3f(0.295166016f, 0.521972656f, -1.f),
    C_Vertex3f(0.384521484f, 0.521972656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_227[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.644008815f, -0.765018106f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Normal3f(0.609832764f, -0.792530119f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Normal3f(0.560195386f, -0.828360498f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Normal3f(0.488698423f, -0.872452796f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Normal3f(0.402021319f, -0.915630281f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Normal3f(0.307879895f, -0.951425254f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Normal3f(0.217794061f, -0.975994766f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Normal3f(0.128345996f, -0.991729438f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Normal3f(0.0436385125f, -0.999047339f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0297879688f, -0.999556243f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0997409225f, -0.995013535f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Normal3f(-0.175596207f, -0.984462261f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Normal3f(-0.2587111f, -0.965954781f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Normal3f(-0.348068923f, -0.937468946f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Normal3f(-0.441677898f, -0.897173643f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Normal3f(-0.536508918f, -0.84389472f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Normal3f(-0.628730416f, -0.777623355f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Normal3f(-0.708205581f, -0.706006348f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Normal3f(-0.777196348f, -0.629258096f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Normal3f(-0.840526164f, -0.541770935f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Normal3f(-0.895173013f, -0.445718706f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Normal3f(-0.93884182f, -0.344348639f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Normal3f(-0.970410466f, -0.241460919f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Normal3f(-0.990046978f, -0.140737399f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Normal3f(-0.999063611f, -0.043263685f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_Normal3f(-0.998176992f, 0.0603547506f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Normal3f(-0.983728349f, 0.179661945f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Normal3f(-0.952656448f, 0.304048836f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Normal3f(-0.90515995f, 0.425071001f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Normal3f(-0.848285139f, 0.529539704f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Normal3f(-0.782067299f, 0.62319386f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Normal3f(-0.701901376f, 0.712274075f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Normal3f(-0.610135138f, 0.792297363f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Normal3f(-0.525150001f, 0.851009607f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Normal3f(-0.451541305f, 0.89225018f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Normal3f(-0.377094716f, 0.926174641f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Normal3f(-0.315602541f, 0.948891461f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Normal3f(-0.262250215f, 0.964999855f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Normal3f(-0.217653468f, 0.976026177f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Normal3f(-0.182780772f, 0.983153641f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Normal3f(-0.152309522f, 0.988332868f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Normal3f(-0.129105836f, 0.991630852f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Normal3f(-0.125778303f, 0.992058396f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Normal3f(-0.140358314f, 0.990100741f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Normal3f(-0.157823548f, 0.987467349f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Normal3f(-0.178095639f, 0.98401314f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Normal3f(-0.20187664f, 0.979411006f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Normal3f(-0.230107605f, 0.973165154f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Normal3f(-0.264068544f, 0.964503944f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Normal3f(-0.297998816f, 0.954566181f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Normal3f(-0.724053323f, 0.689743996f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Normal3f(-0.999676347f, 0.0254404638f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Normal3f(-0.999733686f, 0.0230763834f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Normal3f(-0.999828994f, 0.0184933227f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Normal3f(-0.999987364f, 0.00503883278f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Normal3f(-0.999277949f, -0.0379956067f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Normal3f(-0.995299578f, -0.0968433693f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Normal3f(-0.98471868f, -0.174152732f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Normal3f(-0.964879274f, -0.262693614f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Normal3f(-0.932045639f, -0.362340957f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Normal3f(-0.882021129f, -0.471209705f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Normal3f(-0.811133981f, -0.584860384f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Normal3f(-0.713830113f, -0.700318813f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Normal3f(-0.61406064f, -0.789258838f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Normal3f(-0.521080196f, -0.853507817f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Normal3f(-0.424242347f, -0.905548692f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Normal3f(-0.330907196f, -0.943663299f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Normal3f(-0.243749335f, -0.969838262f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Normal3f(-0.164196789f, -0.986427605f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Normal3f(-0.0926971957f, -0.995694339f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Normal3f(-0.0295525081f, -0.999563217f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Normal3f(0.0223024357f, -0.99975127f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Normal3f(0.0729011372f, -0.997339189f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Normal3f(0.130242616f, -0.991482139f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Normal3f(0.195666403f, -0.980670512f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Normal3f(0.269929171f, -0.962880194f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Normal3f(0.35332042f, -0.93550241f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Normal3f(0.44521296f, -0.895424783f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Normal3f(0.543500543f, -0.839408755f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Normal3f(0.641533256f, -0.767095149f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Normal3f(0.72543782f, -0.688287675f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Normal3f(0.793855846f, -0.608106017f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Normal3f(0.847895145f, -0.530164063f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Normal3f(0.889552832f, -0.45683226f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Normal3f(0.921084702f, -0.389362156f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Normal3f(0.944615781f, -0.328178316f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Normal3f(0.959435046f, -0.281929553f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Normal3f(0.968791068f, -0.247878939f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.131752744f, -0.991282582f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.978250861f, 0.207425416f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Normal3f(-0.964124143f, 0.265451789f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Normal3f(-0.93853426f, 0.345186114f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Normal3f(-0.892804325f, 0.450444669f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Normal3f(-0.82656163f, 0.562846184f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Normal3f(-0.731744051f, 0.68157953f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Normal3f(-0.617528915f, 0.786548138f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Normal3f(-0.501185894f, 0.865339637f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Normal3f(-0.390982777f, 0.920397997f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Normal3f(-0.290884405f, 0.956758201f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Normal3f(-0.197783709f, 0.98024565f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Normal3f(-0.11233186f, 0.993670762f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Normal3f(-0.0353869088f, 0.999373674f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Normal3f(0.0308220927f, 0.999524832f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Normal3f(0.102260895f, 0.994757652f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Normal3f(0.188853115f, 0.982005298f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Normal3f(0.293315113f, 0.956015766f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Normal3f(0.394779444f, 0.918775976f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Normal3f(0.489459455f, 0.872026086f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Normal3f(0.597415328f, 0.801932037f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Normal3f(0.71381259f, 0.700336814f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Normal3f(0.813518882f, 0.581538498f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Normal3f(0.882400215f, 0.470499665f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Normal3f(0.93244344f, 0.361315906f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Normal3f(0.96134758f, 0.275337785f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Normal3f(0.978440464f, 0.206529349f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Normal3f(0.987760723f, 0.155976713f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Normal3f(0.992642164f, 0.121084802f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Normal3f(0.996025085f, 0.0890726f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Normal3f(0.997923374f, 0.0644121915f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Normal3f(0.998993039f, 0.0448651426f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Normal3f(0.999579132f, 0.0290074293f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Normal3f(0.999873698f, 0.015893871f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Normal3f(0.999999583f, 0.000895282777f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Normal3f(0.999999404f, 0.00104705233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Normal3f(0.999950409f, 0.00996590592f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Normal3f(0.999827266f, 0.0185867213f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Normal3f(0.999563813f, 0.0295336749f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Normal3f(0.999036491f, 0.0438889116f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Normal3f(0.997980356f, 0.0635225922f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Normal3f(0.995762765f, 0.0919589549f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Normal3f(0.98978281f, 0.142583564f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Normal3f(0.975923538f, 0.218112797f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Normal3f(0.955474317f, 0.295074373f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Normal3f(0.927006185f, 0.375045985f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Normal3f(0.898835897f, 0.438285291f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Normal3f(0.873886287f, 0.486130446f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.895264685f, -0.445534706f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Normal3f(-0.923641562f, -0.383257538f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Normal3f(-0.95153141f, -0.307551742f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Normal3f(-0.975133479f, -0.221618384f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Normal3f(-0.987897873f, -0.155105904f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Normal3f(-0.994414866f, -0.105542049f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.377323061f, -0.926081717f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Normal3f(0.344282955f, -0.93886584f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Normal3f(0.30806917f, -0.951363921f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Normal3f(0.27083087f, -0.962626934f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Normal3f(0.23911649f, -0.970990896f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Normal3f(0.211867765f, -0.977298319f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Normal3f(0.188257366f, -0.982119739f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Normal3f(0.167636946f, -0.985848784f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Normal3f(0.149529129f, -0.988757372f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Normal3f(0.158242494f, -0.987400293f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Normal3f(0.197629482f, -0.980276704f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Normal3f(0.253822058f, -0.967250884f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Normal3f(0.338508695f, -0.940963209f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Normal3f(0.456908435f, -0.889513731f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Normal3f(0.576411605f, -0.817159474f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Normal3f(0.691899657f, -0.721993685f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Normal3f(0.794279575f, -0.607552409f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Normal3f(0.874713659f, -0.484639972f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Normal3f(0.936157763f, -0.351580352f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Normal3f(0.977780998f, -0.209628955f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Normal3f(0.998205423f, -0.0598839708f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Normal3f(0.99415952f, 0.107920907f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Normal3f(0.951207876f, 0.308550656f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Normal3f(0.858417392f, 0.51295191f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Normal3f(0.730436027f, 0.682981133f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Normal3f(0.568286002f, 0.822831035f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Normal3f(0.378173947f, 0.92573458f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Normal3f(0.204722017f, 0.978820205f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Normal3f(0.0608421154f, 0.998147368f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Normal3f(-0.056054458f, 0.998427689f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Normal3f(-0.176223293f, 0.984350204f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Normal3f(-0.302620173f, 0.953111231f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Normal3f(-0.429032564f, 0.90328896f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Normal3f(-0.549294412f, 0.835628867f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Normal3f(-0.66699791f, 0.74505949f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Normal3f(-0.775314748f, 0.631575108f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Normal3f(-0.863186479f, 0.504885197f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Normal3f(-0.934871018f, 0.354987651f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_Normal3f(-0.976849318f, 0.213928372f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Normal3f(-0.993807793f, 0.111112691f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Normal3f(-0.999024689f, 0.0441556387f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999953806f, -0.00961818546f, 0.f),
    C_Vertex3f(0.099609375f, 0.520507813f, 0.f),
    C_Vertex3f(0.099609375f, 0.520507813f, -1.f),
    C_Normal3f(-0.999230683f, 0.0392165966f, 0.f),
    C_Vertex3f(0.0995550156f, 0.526159286f, 0.f),
    C_Vertex3f(0.0995550156f, 0.526159286f, -1.f),
    C_Normal3f(-0.994498312f, 0.104752719f, 0.f),
    C_Vertex3f(0.10025239f, 0.536891699f, 0.f),
    C_Vertex3f(0.10025239f, 0.536891699f, -1.f),
    C_Normal3f(-0.981874108f, 0.189534381f, 0.f),
    C_Vertex3f(0.101755857f, 0.547053576f, 0.f),
    C_Vertex3f(0.101755857f, 0.547053576f, -1.f),
    C_Normal3f(-0.960046053f, 0.279841781f, 0.f),
    C_Vertex3f(0.104065418f, 0.556644917f, 0.f),
    C_Vertex3f(0.104065418f, 0.556644917f, -1.f),
    C_Normal3f(-0.927577436f, 0.373630971f, 0.f),
    C_Vertex3f(0.107181072f, 0.565665722f, 0.f),
    C_Vertex3f(0.107181072f, 0.565665722f, -1.f),
    C_Normal3f(-0.883691847f, 0.468069077f, 0.f),
    C_Vertex3f(0.111102819f, 0.574115992f, 0.f),
    C_Vertex3f(0.111102819f, 0.574115992f, -1.f),
    C_Normal3f(-0.828614712f, 0.559819281f, 0.f),
    C_Vertex3f(0.11583066f, 0.581995726f, 0.f),
    C_Vertex3f(0.11583066f, 0.581995726f, -1.f),
    C_Normal3f(-0.759081304f, 0.650995851f, 0.f),
    C_Vertex3f(0.121364594f, 0.589304924f, 0.f),
    C_Vertex3f(0.121364594f, 0.589304924f, -1.f),
    C_Normal3f(-0.650275886f, 0.759698212f, 0.f),
    C_Vertex3f(0.130767822f, 0.599412918f, 0.f),
    C_Vertex3f(0.130767822f, 0.599412918f, -1.f),
    C_Normal3f(-0.487445921f, 0.87315321f, 0.f),
    C_Vertex3f(0.144982338f, 0.609520912f, 0.f),
    C_Vertex3f(0.144982338f, 0.609520912f, -1.f),
    C_Normal3f(-0.287504673f, 0.957779288f, 0.f),
    C_Vertex3f(0.160944939f, 0.616259575f, 0.f),
    C_Vertex3f(0.160944939f, 0.616259575f, -1.f),
    C_Normal3f(-0.0979905054f, 0.995187461f, 0.f),
    C_Vertex3f(0.178655624f, 0.619628906f, 0.f),
    C_Vertex3f(0.178655624f, 0.619628906f, -1.f),
    C_Normal3f(0.0889112428f, 0.99603951f, 0.f),
    C_Vertex3f(0.195163727f, 0.619628906f, 0.f),
    C_Vertex3f(0.195163727f, 0.619628906f, -1.f),
    C_Normal3f(0.244741365f, 0.969588399f, 0.f),
    C_Vertex3f(0.211187363f, 0.616724968f, 0.f),
    C_Vertex3f(0.211187363f, 0.616724968f, -1.f),
    C_Normal3f(0.345433027f, 0.938443422f, 0.f),
    C_Vertex3f(0.2296772f, 0.610917091f, 0.f),
    C_Vertex3f(0.2296772f, 0.610917091f, -1.f),
    C_Normal3f(0.413061887f, 0.910703003f, 0.f),
    C_Vertex3f(0.25063324f, 0.602205276f, 0.f),
    C_Vertex3f(0.25063324f, 0.602205276f, -1.f),
    C_Normal3f(0.42238301f, 0.906417429f, 0.f),
    C_Vertex3f(0.268800735f, 0.593172073f, 0.f),
    C_Vertex3f(0.268800735f, 0.593172073f, -1.f),
    C_Normal3f(0.345709145f, 0.938341737f, 0.f),
    C_Vertex3f(0.280400276f, 0.588334084f, 0.f),
    C_Vertex3f(0.280400276f, 0.588334084f, -1.f),
    C_Normal3f(0.243631572f, 0.969867826f, 0.f),
    C_Vertex3f(0.290686607f, 0.585108757f, 0.f),
    C_Vertex3f(0.290686607f, 0.585108757f, -1.f),
    C_Normal3f(0.0984259844f, 0.995144367f, 0.f),
    C_Vertex3f(0.299659729f, 0.583496094f, 0.f),
    C_Vertex3f(0.299659729f, 0.583496094f, -1.f),
    C_Normal3f(-0.070669286f, 0.997499824f, 0.f),
    C_Vertex3f(0.306991577f, 0.583496094f, 0.f),
    C_Vertex3f(0.306991577f, 0.583496094f, -1.f),
    C_Normal3f(-0.24143213f, 0.970417738f, 0.f),
    C_Vertex3f(0.313174725f, 0.584453583f, 0.f),
    C_Vertex3f(0.313174725f, 0.584453583f, -1.f),
    C_Normal3f(-0.435192645f, 0.900337279f, 0.f),
    C_Vertex3f(0.318537235f, 0.586368561f, 0.f),
    C_Vertex3f(0.318537235f, 0.586368561f, -1.f),
    C_Normal3f(-0.62994957f, 0.776636064f, 0.f),
    C_Vertex3f(0.323079109f, 0.589241028f, 0.f),
    C_Vertex3f(0.323079109f, 0.589241028f, -1.f),
    C_Normal3f(-0.810694158f, 0.585469842f, 0.f),
    C_Vertex3f(0.326800346f, 0.593070984f, 0.f),
    C_Vertex3f(0.326800346f, 0.593070984f, -1.f),
    C_Normal3f(-0.9168607f, 0.399207145f, 0.f),
    C_Vertex3f(0.329974651f, 0.598789215f, 0.f),
    C_Vertex3f(0.329974651f, 0.598789215f, -1.f),
    C_Normal3f(-0.964133918f, 0.265416414f, 0.f),
    C_Vertex3f(0.332602024f, 0.606395721f, 0.f),
    C_Vertex3f(0.332602024f, 0.606395721f, -1.f),
    C_Normal3f(-0.982875049f, 0.18427363f, 0.f),
    C_Vertex3f(0.334682465f, 0.615890503f, 0.f),
    C_Vertex3f(0.334682465f, 0.615890503f, -1.f),
    C_Normal3f(-0.991047204f, 0.133512393f, 0.f),
    C_Vertex3f(0.335449219f, 0.621582031f, 0.f),
    C_Vertex3f(0.335449219f, 0.621582031f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.335449219f, 0.621582031f, 0.f),
    C_Vertex3f(0.335449219f, 0.621582031f, -1.f),
    C_Vertex3f(0.392089844f, 0.621582031f, -1.f),
    C_Vertex3f(0.392089844f, 0.621582031f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999653041f, -0.0263388976f, 0.f),
    C_Vertex3f(0.392089844f, 0.621582031f, 0.f),
    C_Vertex3f(0.392089844f, 0.621582031f, -1.f),
    C_Normal3f(0.99777931f, -0.066607669f, 0.f),
    C_Vertex3f(0.391925812f, 0.615356445f, 0.f),
    C_Vertex3f(0.391925812f, 0.615356445f, -1.f),
    C_Normal3f(0.992581308f, -0.121582195f, 0.f),
    C_Vertex3f(0.390895009f, 0.603683472f, 0.f),
    C_Vertex3f(0.390895009f, 0.603683472f, -1.f),
    C_Normal3f(0.980870843f, -0.194659367f, 0.f),
    C_Vertex3f(0.389161468f, 0.592788696f, 0.f),
    C_Vertex3f(0.389161468f, 0.592788696f, -1.f),
    C_Normal3f(0.961305082f, -0.275485754f, 0.f),
    C_Vertex3f(0.386725187f, 0.582672119f, 0.f),
    C_Vertex3f(0.386725187f, 0.582672119f, -1.f),
    C_Normal3f(0.931648552f, -0.363360554f, 0.f),
    C_Vertex3f(0.383586168f, 0.57333374f, 0.f),
    C_Vertex3f(0.383586168f, 0.57333374f, -1.f),
    C_Normal3f(0.889692426f, -0.456560522f, 0.f),
    C_Vertex3f(0.379744411f, 0.56477356f, 0.f),
    C_Vertex3f(0.379744411f, 0.56477356f, -1.f),
    C_Normal3f(0.833749712f, -0.552142739f, 0.f),
    C_Vertex3f(0.375199914f, 0.556991577f, 0.f),
    C_Vertex3f(0.375199914f, 0.556991577f, -1.f),
    C_Normal3f(0.755483389f, -0.655167818f, 0.f),
    C_Vertex3f(0.369952679f, 0.549987793f, 0.f),
    C_Vertex3f(0.369952679f, 0.549987793f, -1.f),
    C_Normal3f(0.639779389f, -0.768558562f, 0.f),
    C_Vertex3f(0.361027718f, 0.540649414f, 0.f),
    C_Vertex3f(0.361027718f, 0.540649414f, -1.f),
    C_Normal3f(0.477566123f, -0.878595829f, 0.f),
    C_Vertex3f(0.347516537f, 0.531311035f, 0.f),
    C_Vertex3f(0.347516537f, 0.531311035f, -1.f),
    C_Normal3f(0.281305313f, -0.95961827f, 0.f),
    C_Vertex3f(0.332394123f, 0.525085449f, 0.f),
    C_Vertex3f(0.332394123f, 0.525085449f, -1.f),
    C_Normal3f(0.0945234597f, -0.995522618f, 0.f),
    C_Vertex3f(0.315660477f, 0.521972656f, 0.f),
    C_Vertex3f(0.315660477f, 0.521972656f, -1.f),
    C_Normal3f(-0.0947503f, -0.995501101f, 0.f),
    C_Vertex3f(0.299610138f, 0.521972656f, 0.f),
    C_Vertex3f(0.299610138f, 0.521972656f, -1.f),
    C_Normal3f(-0.258822501f, -0.965924919f, 0.f),
    C_Vertex3f(0.283657074f, 0.525018692f, 0.f),
    C_Vertex3f(0.283657074f, 0.525018692f, -1.f),
    C_Normal3f(-0.367869198f, -0.92987752f, 0.f),
    C_Vertex3f(0.265506744f, 0.531110764f, 0.f),
    C_Vertex3f(0.265506744f, 0.531110764f, -1.f),
    C_Normal3f(-0.441156626f, -0.897430122f, 0.f),
    C_Vertex3f(0.245159149f, 0.540248871f, 0.f),
    C_Vertex3f(0.245159149f, 0.540248871f, -1.f),
    C_Normal3f(-0.452606946f, -0.891710103f, 0.f),
    C_Vertex3f(0.226554871f, 0.550258636f, 0.f),
    C_Vertex3f(0.226554871f, 0.550258636f, -1.f),
    C_Normal3f(-0.385828793f, -0.922570407f, 0.f),
    C_Vertex3f(0.213860512f, 0.556135178f, 0.f),
    C_Vertex3f(0.213860512f, 0.556135178f, -1.f),
    C_Normal3f(-0.288805217f, -0.957387865f, 0.f),
    C_Vertex3f(0.20313549f, 0.560052872f, 0.f),
    C_Vertex3f(0.20313549f, 0.560052872f, -1.f),
    C_Normal3f(-0.122500464f, -0.992468476f, 0.f),
    C_Vertex3f(0.194379807f, 0.562011719f, 0.f),
    C_Vertex3f(0.194379807f, 0.562011719f, -1.f),
    C_Normal3f(0.0942201167f, -0.995551467f, 0.f),
    C_Vertex3f(0.187265396f, 0.562011719f, 0.f),
    C_Vertex3f(0.187265396f, 0.562011719f, -1.f),
    C_Normal3f(0.294618219f, -0.955615103f, 0.f),
    C_Vertex3f(0.180534363f, 0.56070137f, 0.f),
    C_Vertex3f(0.180534363f, 0.56070137f, -1.f),
    C_Normal3f(0.500668526f, -0.865639091f, 0.f),
    C_Vertex3f(0.174514771f, 0.558080673f, 0.f),
    C_Vertex3f(0.174514771f, 0.558080673f, -1.f),
    C_Normal3f(0.652620137f, -0.757685304f, 0.f),
    C_Vertex3f(0.169206619f, 0.554149628f, 0.f),
    C_Vertex3f(0.169206619f, 0.554149628f, -1.f),
    C_Normal3f(0.744249165f, -0.667902052f, 0.f),
    C_Vertex3f(0.166908264f, 0.551528931f, 0.f),
    C_Vertex3f(0.166908264f, 0.551528931f, -1.f),
    C_Normal3f(0.837037444f, -0.547145605f, 0.f),
    C_Vertex3f(0.16450119f, 0.548906326f, 0.f),
    C_Vertex3f(0.16450119f, 0.548906326f, -1.f),
    C_Normal3f(0.925027549f, -0.379899919f, 0.f),
    C_Vertex3f(0.160944939f, 0.542406082f, 0.f),
    C_Vertex3f(0.160944939f, 0.542406082f, -1.f),
    C_Normal3f(0.980741322f, -0.195311442f, 0.f),
    C_Vertex3f(0.158646584f, 0.534650803f, 0.f),
    C_Vertex3f(0.158646584f, 0.534650803f, -1.f),
    C_Normal3f(0.997837007f, -0.065737471f, 0.f),
    C_Vertex3f(0.157606125f, 0.525640488f, 0.f),
    C_Vertex3f(0.157606125f, 0.525640488f, -1.f),
    C_Normal3f(0.999775767f, 0.0211769659f, 0.f),
    C_Vertex3f(0.157714844f, 0.520507813f, 0.f),
    C_Vertex3f(0.157714844f, 0.520507813f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.157714844f, 0.520507813f, 0.f),
    C_Vertex3f(0.157714844f, 0.520507813f, -1.f),
    C_Vertex3f(0.099609375f, 0.520507813f, -1.f),
    C_Vertex3f(0.099609375f, 0.520507813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.157714844f, 0.520507813f, 0.f),
    C_Vertex3f(0.157606125f, 0.525640488f, 0.f),
    C_Vertex3f(0.144982338f, 0.609520912f, 0.f),
    C_Vertex3f(0.130767822f, 0.599412918f, 0.f),
    C_Vertex3f(0.121364594f, 0.589304924f, 0.f),
    C_Vertex3f(0.11583066f, 0.581995726f, 0.f),
    C_Vertex3f(0.111102819f, 0.574115992f, 0.f),
    C_Vertex3f(0.107181072f, 0.565665722f, 0.f),
    C_Vertex3f(0.104065418f, 0.556644917f, 0.f),
    C_Vertex3f(0.101755857f, 0.547053576f, 0.f),
    C_Vertex3f(0.10025239f, 0.536891699f, 0.f),
    C_Vertex3f(0.099609375f, 0.520507813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.392089844f, 0.621582031f, 0.f),
    C_Vertex3f(0.335449219f, 0.621582031f, 0.f),
    C_Vertex3f(0.347516537f, 0.531311035f, 0.f),
    C_Vertex3f(0.361027718f, 0.540649414f, 0.f),
    C_Vertex3f(0.369952679f, 0.549987793f, 0.f),
    C_Vertex3f(0.375199914f, 0.556991577f, 0.f),
    C_Vertex3f(0.379744411f, 0.56477356f, 0.f),
    C_Vertex3f(0.383586168f, 0.57333374f, 0.f),
    C_Vertex3f(0.386725187f, 0.582672119f, 0.f),
    C_Vertex3f(0.389161468f, 0.592788696f, 0.f),
    C_Vertex3f(0.390895009f, 0.603683472f, 0.f),
    C_Vertex3f(0.391925812f, 0.615356445f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335449219f, 0.621582031f, 0.f),
    C_Vertex3f(0.334682465f, 0.615890503f, 0.f),
    C_Vertex3f(0.347516537f, 0.531311035f, 0.f),
    C_Vertex3f(0.332602024f, 0.606395721f, 0.f),
    C_Vertex3f(0.332394123f, 0.525085449f, 0.f),
    C_Vertex3f(0.329974651f, 0.598789215f, 0.f),
    C_Vertex3f(0.326800346f, 0.593070984f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.326800346f, 0.593070984f, 0.f),
    C_Vertex3f(0.323079109f, 0.589241028f, 0.f),
    C_Vertex3f(0.332394123f, 0.525085449f, 0.f),
    C_Vertex3f(0.318537235f, 0.586368561f, 0.f),
    C_Vertex3f(0.315660477f, 0.521972656f, 0.f),
    C_Vertex3f(0.313174725f, 0.584453583f, 0.f),
    C_Vertex3f(0.306991577f, 0.583496094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.280400276f, 0.588334084f, 0.f),
    C_Vertex3f(0.283657074f, 0.525018692f, 0.f),
    C_Vertex3f(0.290686607f, 0.585108757f, 0.f),
    C_Vertex3f(0.299610138f, 0.521972656f, 0.f),
    C_Vertex3f(0.299659729f, 0.583496094f, 0.f),
    C_Vertex3f(0.315660477f, 0.521972656f, 0.f),
    C_Vertex3f(0.306991577f, 0.583496094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213860512f, 0.556135178f, 0.f),
    C_Vertex3f(0.226554871f, 0.550258636f, 0.f),
    C_Vertex3f(0.2296772f, 0.610917091f, 0.f),
    C_Vertex3f(0.245159149f, 0.540248871f, 0.f),
    C_Vertex3f(0.25063324f, 0.602205276f, 0.f),
    C_Vertex3f(0.265506744f, 0.531110764f, 0.f),
    C_Vertex3f(0.268800735f, 0.593172073f, 0.f),
    C_Vertex3f(0.283657074f, 0.525018692f, 0.f),
    C_Vertex3f(0.280400276f, 0.588334084f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.187265396f, 0.562011719f, 0.f),
    C_Vertex3f(0.194379807f, 0.562011719f, 0.f),
    C_Vertex3f(0.195163727f, 0.619628906f, 0.f),
    C_Vertex3f(0.20313549f, 0.560052872f, 0.f),
    C_Vertex3f(0.211187363f, 0.616724968f, 0.f),
    C_Vertex3f(0.213860512f, 0.556135178f, 0.f),
    C_Vertex3f(0.2296772f, 0.610917091f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.169206619f, 0.554149628f, 0.f),
    C_Vertex3f(0.174514771f, 0.558080673f, 0.f),
    C_Vertex3f(0.178655624f, 0.619628906f, 0.f),
    C_Vertex3f(0.180534363f, 0.56070137f, 0.f),
    C_Vertex3f(0.195163727f, 0.619628906f, 0.f),
    C_Vertex3f(0.187265396f, 0.562011719f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.158646584f, 0.534650803f, 0.f),
    C_Vertex3f(0.160944939f, 0.542406082f, 0.f),
    C_Vertex3f(0.160944939f, 0.616259575f, 0.f),
    C_Vertex3f(0.16450119f, 0.548906326f, 0.f),
    C_Vertex3f(0.178655624f, 0.619628906f, 0.f),
    C_Vertex3f(0.166908264f, 0.551528931f, 0.f),
    C_Vertex3f(0.169206619f, 0.554149628f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.160944939f, 0.616259575f, 0.f),
    C_Vertex3f(0.144982338f, 0.609520912f, 0.f),
    C_Vertex3f(0.157606125f, 0.525640488f, 0.f),
    C_Vertex3f(0.158646584f, 0.534650803f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.099609375f, 0.520507813f, 0.f),
    C_Vertex3f(0.10025239f, 0.536891699f, 0.f),
    C_Vertex3f(0.0995550156f, 0.526159286f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.157714844f, 0.520507813f, -1.f),
    C_Vertex3f(0.157606125f, 0.525640488f, -1.f),
    C_Vertex3f(0.144982338f, 0.609520912f, -1.f),
    C_Vertex3f(0.130767822f, 0.599412918f, -1.f),
    C_Vertex3f(0.121364594f, 0.589304924f, -1.f),
    C_Vertex3f(0.11583066f, 0.581995726f, -1.f),
    C_Vertex3f(0.111102819f, 0.574115992f, -1.f),
    C_Vertex3f(0.107181072f, 0.565665722f, -1.f),
    C_Vertex3f(0.104065418f, 0.556644917f, -1.f),
    C_Vertex3f(0.101755857f, 0.547053576f, -1.f),
    C_Vertex3f(0.10025239f, 0.536891699f, -1.f),
    C_Vertex3f(0.099609375f, 0.520507813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.392089844f, 0.621582031f, -1.f),
    C_Vertex3f(0.335449219f, 0.621582031f, -1.f),
    C_Vertex3f(0.347516537f, 0.531311035f, -1.f),
    C_Vertex3f(0.361027718f, 0.540649414f, -1.f),
    C_Vertex3f(0.369952679f, 0.549987793f, -1.f),
    C_Vertex3f(0.375199914f, 0.556991577f, -1.f),
    C_Vertex3f(0.379744411f, 0.56477356f, -1.f),
    C_Vertex3f(0.383586168f, 0.57333374f, -1.f),
    C_Vertex3f(0.386725187f, 0.582672119f, -1.f),
    C_Vertex3f(0.389161468f, 0.592788696f, -1.f),
    C_Vertex3f(0.390895009f, 0.603683472f, -1.f),
    C_Vertex3f(0.391925812f, 0.615356445f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335449219f, 0.621582031f, -1.f),
    C_Vertex3f(0.334682465f, 0.615890503f, -1.f),
    C_Vertex3f(0.347516537f, 0.531311035f, -1.f),
    C_Vertex3f(0.332602024f, 0.606395721f, -1.f),
    C_Vertex3f(0.332394123f, 0.525085449f, -1.f),
    C_Vertex3f(0.329974651f, 0.598789215f, -1.f),
    C_Vertex3f(0.326800346f, 0.593070984f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.326800346f, 0.593070984f, -1.f),
    C_Vertex3f(0.323079109f, 0.589241028f, -1.f),
    C_Vertex3f(0.332394123f, 0.525085449f, -1.f),
    C_Vertex3f(0.318537235f, 0.586368561f, -1.f),
    C_Vertex3f(0.315660477f, 0.521972656f, -1.f),
    C_Vertex3f(0.313174725f, 0.584453583f, -1.f),
    C_Vertex3f(0.306991577f, 0.583496094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.280400276f, 0.588334084f, -1.f),
    C_Vertex3f(0.283657074f, 0.525018692f, -1.f),
    C_Vertex3f(0.290686607f, 0.585108757f, -1.f),
    C_Vertex3f(0.299610138f, 0.521972656f, -1.f),
    C_Vertex3f(0.299659729f, 0.583496094f, -1.f),
    C_Vertex3f(0.315660477f, 0.521972656f, -1.f),
    C_Vertex3f(0.306991577f, 0.583496094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213860512f, 0.556135178f, -1.f),
    C_Vertex3f(0.226554871f, 0.550258636f, -1.f),
    C_Vertex3f(0.2296772f, 0.610917091f, -1.f),
    C_Vertex3f(0.245159149f, 0.540248871f, -1.f),
    C_Vertex3f(0.25063324f, 0.602205276f, -1.f),
    C_Vertex3f(0.265506744f, 0.531110764f, -1.f),
    C_Vertex3f(0.268800735f, 0.593172073f, -1.f),
    C_Vertex3f(0.283657074f, 0.525018692f, -1.f),
    C_Vertex3f(0.280400276f, 0.588334084f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.187265396f, 0.562011719f, -1.f),
    C_Vertex3f(0.194379807f, 0.562011719f, -1.f),
    C_Vertex3f(0.195163727f, 0.619628906f, -1.f),
    C_Vertex3f(0.20313549f, 0.560052872f, -1.f),
    C_Vertex3f(0.211187363f, 0.616724968f, -1.f),
    C_Vertex3f(0.213860512f, 0.556135178f, -1.f),
    C_Vertex3f(0.2296772f, 0.610917091f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.169206619f, 0.554149628f, -1.f),
    C_Vertex3f(0.174514771f, 0.558080673f, -1.f),
    C_Vertex3f(0.178655624f, 0.619628906f, -1.f),
    C_Vertex3f(0.180534363f, 0.56070137f, -1.f),
    C_Vertex3f(0.195163727f, 0.619628906f, -1.f),
    C_Vertex3f(0.187265396f, 0.562011719f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.158646584f, 0.534650803f, -1.f),
    C_Vertex3f(0.160944939f, 0.542406082f, -1.f),
    C_Vertex3f(0.160944939f, 0.616259575f, -1.f),
    C_Vertex3f(0.16450119f, 0.548906326f, -1.f),
    C_Vertex3f(0.178655624f, 0.619628906f, -1.f),
    C_Vertex3f(0.166908264f, 0.551528931f, -1.f),
    C_Vertex3f(0.169206619f, 0.554149628f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.160944939f, 0.616259575f, -1.f),
    C_Vertex3f(0.144982338f, 0.609520912f, -1.f),
    C_Vertex3f(0.157606125f, 0.525640488f, -1.f),
    C_Vertex3f(0.158646584f, 0.534650803f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.099609375f, 0.520507813f, -1.f),
    C_Vertex3f(0.10025239f, 0.536891699f, -1.f),
    C_Vertex3f(0.0995550156f, 0.526159286f, -1.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_228[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.644008815f, -0.765018106f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Normal3f(0.609832764f, -0.792530119f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Normal3f(0.560195386f, -0.828360498f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Normal3f(0.488698423f, -0.872452796f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Normal3f(0.402021319f, -0.915630281f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Normal3f(0.307879895f, -0.951425254f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Normal3f(0.217794061f, -0.975994766f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Normal3f(0.128345996f, -0.991729438f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Normal3f(0.0436385125f, -0.999047339f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0297879688f, -0.999556243f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0997409225f, -0.995013535f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Normal3f(-0.175596207f, -0.984462261f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Normal3f(-0.2587111f, -0.965954781f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Normal3f(-0.348068923f, -0.937468946f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Normal3f(-0.441677898f, -0.897173643f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Normal3f(-0.536508918f, -0.84389472f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Normal3f(-0.628730416f, -0.777623355f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Normal3f(-0.708205581f, -0.706006348f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Normal3f(-0.777196348f, -0.629258096f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Normal3f(-0.840526164f, -0.541770935f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Normal3f(-0.895173013f, -0.445718706f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Normal3f(-0.93884182f, -0.344348639f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Normal3f(-0.970410466f, -0.241460919f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Normal3f(-0.990046978f, -0.140737399f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Normal3f(-0.999063611f, -0.043263685f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_Normal3f(-0.998176992f, 0.0603547506f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Normal3f(-0.983728349f, 0.179661945f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Normal3f(-0.952656448f, 0.304048836f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Normal3f(-0.90515995f, 0.425071001f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Normal3f(-0.848285139f, 0.529539704f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Normal3f(-0.782067299f, 0.62319386f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Normal3f(-0.701901376f, 0.712274075f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Normal3f(-0.610135138f, 0.792297363f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Normal3f(-0.525150001f, 0.851009607f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Normal3f(-0.451541305f, 0.89225018f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Normal3f(-0.377094716f, 0.926174641f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Normal3f(-0.315602541f, 0.948891461f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Normal3f(-0.262250215f, 0.964999855f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Normal3f(-0.217653468f, 0.976026177f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Normal3f(-0.182780772f, 0.983153641f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Normal3f(-0.152309522f, 0.988332868f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Normal3f(-0.129105836f, 0.991630852f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Normal3f(-0.125778303f, 0.992058396f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Normal3f(-0.140358314f, 0.990100741f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Normal3f(-0.157823548f, 0.987467349f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Normal3f(-0.178095639f, 0.98401314f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Normal3f(-0.20187664f, 0.979411006f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Normal3f(-0.230107605f, 0.973165154f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Normal3f(-0.264068544f, 0.964503944f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Normal3f(-0.297998816f, 0.954566181f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Normal3f(-0.724053323f, 0.689743996f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Normal3f(-0.999676347f, 0.0254404638f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Normal3f(-0.999733686f, 0.0230763834f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Normal3f(-0.999828994f, 0.0184933227f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Normal3f(-0.999987364f, 0.00503883278f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Normal3f(-0.999277949f, -0.0379956067f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Normal3f(-0.995299578f, -0.0968433693f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Normal3f(-0.98471868f, -0.174152732f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Normal3f(-0.964879274f, -0.262693614f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Normal3f(-0.932045639f, -0.362340957f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Normal3f(-0.882021129f, -0.471209705f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Normal3f(-0.811133981f, -0.584860384f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Normal3f(-0.713830113f, -0.700318813f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Normal3f(-0.61406064f, -0.789258838f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Normal3f(-0.521080196f, -0.853507817f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Normal3f(-0.424242347f, -0.905548692f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Normal3f(-0.330907196f, -0.943663299f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Normal3f(-0.243749335f, -0.969838262f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Normal3f(-0.164196789f, -0.986427605f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Normal3f(-0.0926971957f, -0.995694339f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Normal3f(-0.0295525081f, -0.999563217f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Normal3f(0.0223024357f, -0.99975127f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Normal3f(0.0729011372f, -0.997339189f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Normal3f(0.130242616f, -0.991482139f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Normal3f(0.195666403f, -0.980670512f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Normal3f(0.269929171f, -0.962880194f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Normal3f(0.35332042f, -0.93550241f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Normal3f(0.44521296f, -0.895424783f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Normal3f(0.543500543f, -0.839408755f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Normal3f(0.641533256f, -0.767095149f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Normal3f(0.72543782f, -0.688287675f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Normal3f(0.793855846f, -0.608106017f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Normal3f(0.847895145f, -0.530164063f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Normal3f(0.889552832f, -0.45683226f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Normal3f(0.921084702f, -0.389362156f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Normal3f(0.944615781f, -0.328178316f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Normal3f(0.959435046f, -0.281929553f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Normal3f(0.968791068f, -0.247878939f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.131752744f, -0.991282582f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.978250861f, 0.207425416f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Normal3f(-0.964124143f, 0.265451789f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Normal3f(-0.93853426f, 0.345186114f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Normal3f(-0.892804325f, 0.450444669f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Normal3f(-0.82656163f, 0.562846184f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Normal3f(-0.731744051f, 0.68157953f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Normal3f(-0.617528915f, 0.786548138f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Normal3f(-0.501185894f, 0.865339637f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Normal3f(-0.390982777f, 0.920397997f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Normal3f(-0.290884405f, 0.956758201f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Normal3f(-0.197783709f, 0.98024565f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Normal3f(-0.11233186f, 0.993670762f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Normal3f(-0.0353869088f, 0.999373674f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Normal3f(0.0308220927f, 0.999524832f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Normal3f(0.102260895f, 0.994757652f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Normal3f(0.188853115f, 0.982005298f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Normal3f(0.293315113f, 0.956015766f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Normal3f(0.394779444f, 0.918775976f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Normal3f(0.489459455f, 0.872026086f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Normal3f(0.597415328f, 0.801932037f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Normal3f(0.71381259f, 0.700336814f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Normal3f(0.813518882f, 0.581538498f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Normal3f(0.882400215f, 0.470499665f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Normal3f(0.93244344f, 0.361315906f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Normal3f(0.96134758f, 0.275337785f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Normal3f(0.978440464f, 0.206529349f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Normal3f(0.987760723f, 0.155976713f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Normal3f(0.992642164f, 0.121084802f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Normal3f(0.996025085f, 0.0890726f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Normal3f(0.997923374f, 0.0644121915f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Normal3f(0.998993039f, 0.0448651426f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Normal3f(0.999579132f, 0.0290074293f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Normal3f(0.999873698f, 0.015893871f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Normal3f(0.999999583f, 0.000895282777f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Normal3f(0.999999404f, 0.00104705233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Normal3f(0.999950409f, 0.00996590592f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Normal3f(0.999827266f, 0.0185867213f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Normal3f(0.999563813f, 0.0295336749f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Normal3f(0.999036491f, 0.0438889116f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Normal3f(0.997980356f, 0.0635225922f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Normal3f(0.995762765f, 0.0919589549f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Normal3f(0.98978281f, 0.142583564f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Normal3f(0.975923538f, 0.218112797f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Normal3f(0.955474317f, 0.295074373f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Normal3f(0.927006185f, 0.375045985f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Normal3f(0.898835897f, 0.438285291f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Normal3f(0.873886287f, 0.486130446f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.895264685f, -0.445534706f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Normal3f(-0.923641562f, -0.383257538f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Normal3f(-0.95153141f, -0.307551742f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Normal3f(-0.975133479f, -0.221618384f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Normal3f(-0.987897873f, -0.155105904f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Normal3f(-0.994414866f, -0.105542049f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.377323061f, -0.926081717f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Normal3f(0.344282955f, -0.93886584f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Normal3f(0.30806917f, -0.951363921f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Normal3f(0.27083087f, -0.962626934f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Normal3f(0.23911649f, -0.970990896f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Normal3f(0.211867765f, -0.977298319f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Normal3f(0.188257366f, -0.982119739f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Normal3f(0.167636946f, -0.985848784f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Normal3f(0.149529129f, -0.988757372f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Normal3f(0.158242494f, -0.987400293f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Normal3f(0.197629482f, -0.980276704f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Normal3f(0.253822058f, -0.967250884f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Normal3f(0.338508695f, -0.940963209f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Normal3f(0.456908435f, -0.889513731f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Normal3f(0.576411605f, -0.817159474f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Normal3f(0.691899657f, -0.721993685f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Normal3f(0.794279575f, -0.607552409f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Normal3f(0.874713659f, -0.484639972f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Normal3f(0.936157763f, -0.351580352f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Normal3f(0.977780998f, -0.209628955f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Normal3f(0.998205423f, -0.0598839708f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Normal3f(0.99415952f, 0.107920907f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Normal3f(0.951207876f, 0.308550656f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Normal3f(0.858417392f, 0.51295191f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Normal3f(0.730436027f, 0.682981133f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Normal3f(0.568286002f, 0.822831035f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Normal3f(0.378173947f, 0.92573458f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Normal3f(0.204722017f, 0.978820205f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Normal3f(0.0608421154f, 0.998147368f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Normal3f(-0.056054458f, 0.998427689f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Normal3f(-0.176223293f, 0.984350204f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Normal3f(-0.302620173f, 0.953111231f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Normal3f(-0.429032564f, 0.90328896f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Normal3f(-0.549294412f, 0.835628867f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Normal3f(-0.66699791f, 0.74505949f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Normal3f(-0.775314748f, 0.631575108f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Normal3f(-0.863186479f, 0.504885197f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Normal3f(-0.934871018f, 0.354987651f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_Normal3f(-0.976849318f, 0.213928372f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Normal3f(-0.993807793f, 0.111112691f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Normal3f(-0.999024689f, 0.0441556387f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.134765625f, 0.521972656f, 0.f),
    C_Vertex3f(0.134765625f, 0.521972656f, -1.f),
    C_Vertex3f(0.134765625f, 0.611328125f, -1.f),
    C_Vertex3f(0.134765625f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.134765625f, 0.611328125f, 0.f),
    C_Vertex3f(0.134765625f, 0.611328125f, -1.f),
    C_Vertex3f(0.216796875f, 0.611328125f, -1.f),
    C_Vertex3f(0.216796875f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.216796875f, 0.611328125f, 0.f),
    C_Vertex3f(0.216796875f, 0.611328125f, -1.f),
    C_Vertex3f(0.216796875f, 0.521972656f, -1.f),
    C_Vertex3f(0.216796875f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.216796875f, 0.521972656f, 0.f),
    C_Vertex3f(0.216796875f, 0.521972656f, -1.f),
    C_Vertex3f(0.134765625f, 0.521972656f, -1.f),
    C_Vertex3f(0.134765625f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.297851563f, 0.521972656f, 0.f),
    C_Vertex3f(0.297851563f, 0.521972656f, -1.f),
    C_Vertex3f(0.297851563f, 0.611328125f, -1.f),
    C_Vertex3f(0.297851563f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.297851563f, 0.611328125f, 0.f),
    C_Vertex3f(0.297851563f, 0.611328125f, -1.f),
    C_Vertex3f(0.379882813f, 0.611328125f, -1.f),
    C_Vertex3f(0.379882813f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.379882813f, 0.611328125f, 0.f),
    C_Vertex3f(0.379882813f, 0.611328125f, -1.f),
    C_Vertex3f(0.379882813f, 0.521972656f, -1.f),
    C_Vertex3f(0.379882813f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.379882813f, 0.521972656f, 0.f),
    C_Vertex3f(0.379882813f, 0.521972656f, -1.f),
    C_Vertex3f(0.297851563f, 0.521972656f, -1.f),
    C_Vertex3f(0.297851563f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134765625f, 0.611328125f, 0.f),
    C_Vertex3f(0.134765625f, 0.521972656f, 0.f),
    C_Vertex3f(0.216796875f, 0.521972656f, 0.f),
    C_Vertex3f(0.216796875f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.297851563f, 0.611328125f, 0.f),
    C_Vertex3f(0.297851563f, 0.521972656f, 0.f),
    C_Vertex3f(0.379882813f, 0.521972656f, 0.f),
    C_Vertex3f(0.379882813f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134765625f, 0.611328125f, -1.f),
    C_Vertex3f(0.134765625f, 0.521972656f, -1.f),
    C_Vertex3f(0.216796875f, 0.521972656f, -1.f),
    C_Vertex3f(0.216796875f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.297851563f, 0.611328125f, -1.f),
    C_Vertex3f(0.297851563f, 0.521972656f, -1.f),
    C_Vertex3f(0.379882813f, 0.521972656f, -1.f),
    C_Vertex3f(0.379882813f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_229[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.644008815f, -0.765018106f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Normal3f(0.609832764f, -0.792530119f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Normal3f(0.560195386f, -0.828360498f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Normal3f(0.488698423f, -0.872452796f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Normal3f(0.402021319f, -0.915630281f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Normal3f(0.307879895f, -0.951425254f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Normal3f(0.217794061f, -0.975994766f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Normal3f(0.128345996f, -0.991729438f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Normal3f(0.0436385125f, -0.999047339f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0297879688f, -0.999556243f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0997409225f, -0.995013535f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Normal3f(-0.175596207f, -0.984462261f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Normal3f(-0.2587111f, -0.965954781f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Normal3f(-0.348068923f, -0.937468946f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Normal3f(-0.441677898f, -0.897173643f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Normal3f(-0.536508918f, -0.84389472f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Normal3f(-0.628730416f, -0.777623355f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Normal3f(-0.708205581f, -0.706006348f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Normal3f(-0.777196348f, -0.629258096f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Normal3f(-0.840526164f, -0.541770935f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Normal3f(-0.895173013f, -0.445718706f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Normal3f(-0.93884182f, -0.344348639f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Normal3f(-0.970410466f, -0.241460919f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Normal3f(-0.990046978f, -0.140737399f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Normal3f(-0.999063611f, -0.043263685f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_Normal3f(-0.998176992f, 0.0603547506f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Normal3f(-0.983728349f, 0.179661945f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Normal3f(-0.952656448f, 0.304048836f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Normal3f(-0.90515995f, 0.425071001f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Normal3f(-0.848285139f, 0.529539704f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Normal3f(-0.782067299f, 0.62319386f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Normal3f(-0.701901376f, 0.712274075f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Normal3f(-0.610135138f, 0.792297363f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Normal3f(-0.525150001f, 0.851009607f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Normal3f(-0.451541305f, 0.89225018f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Normal3f(-0.377094716f, 0.926174641f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Normal3f(-0.315602541f, 0.948891461f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Normal3f(-0.262250215f, 0.964999855f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Normal3f(-0.217653468f, 0.976026177f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Normal3f(-0.182780772f, 0.983153641f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Normal3f(-0.152309522f, 0.988332868f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Normal3f(-0.129105836f, 0.991630852f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Normal3f(-0.125778303f, 0.992058396f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Normal3f(-0.140358314f, 0.990100741f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Normal3f(-0.157823548f, 0.987467349f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Normal3f(-0.178095639f, 0.98401314f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Normal3f(-0.20187664f, 0.979411006f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Normal3f(-0.230107605f, 0.973165154f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Normal3f(-0.264068544f, 0.964503944f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Normal3f(-0.297998816f, 0.954566181f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Normal3f(-0.724053323f, 0.689743996f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Normal3f(-0.999676347f, 0.0254404638f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Normal3f(-0.999733686f, 0.0230763834f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Normal3f(-0.999828994f, 0.0184933227f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Normal3f(-0.999987364f, 0.00503883278f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Normal3f(-0.999277949f, -0.0379956067f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Normal3f(-0.995299578f, -0.0968433693f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Normal3f(-0.98471868f, -0.174152732f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Normal3f(-0.964879274f, -0.262693614f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Normal3f(-0.932045639f, -0.362340957f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Normal3f(-0.882021129f, -0.471209705f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Normal3f(-0.811133981f, -0.584860384f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Normal3f(-0.713830113f, -0.700318813f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Normal3f(-0.61406064f, -0.789258838f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Normal3f(-0.521080196f, -0.853507817f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Normal3f(-0.424242347f, -0.905548692f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Normal3f(-0.330907196f, -0.943663299f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Normal3f(-0.243749335f, -0.969838262f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Normal3f(-0.164196789f, -0.986427605f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Normal3f(-0.0926971957f, -0.995694339f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Normal3f(-0.0295525081f, -0.999563217f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Normal3f(0.0223024357f, -0.99975127f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Normal3f(0.0729011372f, -0.997339189f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Normal3f(0.130242616f, -0.991482139f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Normal3f(0.195666403f, -0.980670512f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Normal3f(0.269929171f, -0.962880194f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Normal3f(0.35332042f, -0.93550241f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Normal3f(0.44521296f, -0.895424783f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Normal3f(0.543500543f, -0.839408755f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Normal3f(0.641533256f, -0.767095149f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Normal3f(0.72543782f, -0.688287675f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Normal3f(0.793855846f, -0.608106017f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Normal3f(0.847895145f, -0.530164063f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Normal3f(0.889552832f, -0.45683226f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Normal3f(0.921084702f, -0.389362156f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Normal3f(0.944615781f, -0.328178316f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Normal3f(0.959435046f, -0.281929553f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Normal3f(0.968791068f, -0.247878939f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.131752744f, -0.991282582f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.978250861f, 0.207425416f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_Normal3f(-0.964124143f, 0.265451789f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Normal3f(-0.93853426f, 0.345186114f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Normal3f(-0.892804325f, 0.450444669f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Normal3f(-0.82656163f, 0.562846184f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Normal3f(-0.731744051f, 0.68157953f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Normal3f(-0.617528915f, 0.786548138f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Normal3f(-0.501185894f, 0.865339637f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Normal3f(-0.390982777f, 0.920397997f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Normal3f(-0.290884405f, 0.956758201f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Normal3f(-0.197783709f, 0.98024565f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Normal3f(-0.11233186f, 0.993670762f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Normal3f(-0.0353869088f, 0.999373674f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Normal3f(0.0308220927f, 0.999524832f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Normal3f(0.102260895f, 0.994757652f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Normal3f(0.188853115f, 0.982005298f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Normal3f(0.293315113f, 0.956015766f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Normal3f(0.394779444f, 0.918775976f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Normal3f(0.489459455f, 0.872026086f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Normal3f(0.597415328f, 0.801932037f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Normal3f(0.71381259f, 0.700336814f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Normal3f(0.813518882f, 0.581538498f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Normal3f(0.882400215f, 0.470499665f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Normal3f(0.93244344f, 0.361315906f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Normal3f(0.96134758f, 0.275337785f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Normal3f(0.978440464f, 0.206529349f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Normal3f(0.987760723f, 0.155976713f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Normal3f(0.992642164f, 0.121084802f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Normal3f(0.996025085f, 0.0890726f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Normal3f(0.997923374f, 0.0644121915f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Normal3f(0.998993039f, 0.0448651426f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Normal3f(0.999579132f, 0.0290074293f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Normal3f(0.999873698f, 0.015893871f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Normal3f(0.999999583f, 0.000895282777f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Normal3f(0.999999404f, 0.00104705233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Normal3f(0.999950409f, 0.00996590592f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Normal3f(0.999827266f, 0.0185867213f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Normal3f(0.999563813f, 0.0295336749f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Normal3f(0.999036491f, 0.0438889116f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Normal3f(0.997980356f, 0.0635225922f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Normal3f(0.995762765f, 0.0919589549f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Normal3f(0.98978281f, 0.142583564f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Normal3f(0.975923538f, 0.218112797f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Normal3f(0.955474317f, 0.295074373f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Normal3f(0.927006185f, 0.375045985f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Normal3f(0.898835897f, 0.438285291f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Normal3f(0.873886287f, 0.486130446f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.895264685f, -0.445534706f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Normal3f(-0.923641562f, -0.383257538f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Normal3f(-0.95153141f, -0.307551742f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Normal3f(-0.975133479f, -0.221618384f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Normal3f(-0.987897873f, -0.155105904f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Normal3f(-0.994414866f, -0.105542049f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.377323061f, -0.926081717f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Normal3f(0.344282955f, -0.93886584f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Normal3f(0.30806917f, -0.951363921f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Normal3f(0.27083087f, -0.962626934f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Normal3f(0.23911649f, -0.970990896f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Normal3f(0.211867765f, -0.977298319f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Normal3f(0.188257366f, -0.982119739f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Normal3f(0.167636946f, -0.985848784f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Normal3f(0.149529129f, -0.988757372f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Normal3f(0.158242494f, -0.987400293f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Normal3f(0.197629482f, -0.980276704f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Normal3f(0.253822058f, -0.967250884f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Normal3f(0.338508695f, -0.940963209f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Normal3f(0.456908435f, -0.889513731f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Normal3f(0.576411605f, -0.817159474f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Normal3f(0.691899657f, -0.721993685f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Normal3f(0.794279575f, -0.607552409f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Normal3f(0.874713659f, -0.484639972f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Normal3f(0.936157763f, -0.351580352f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Normal3f(0.977780998f, -0.209628955f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Normal3f(0.998205423f, -0.0598839708f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Normal3f(0.99415952f, 0.107920907f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Normal3f(0.951207876f, 0.308550656f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Normal3f(0.858417392f, 0.51295191f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Normal3f(0.730436027f, 0.682981133f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Normal3f(0.568286002f, 0.822831035f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Normal3f(0.378173947f, 0.92573458f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Normal3f(0.204722017f, 0.978820205f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Normal3f(0.0608421154f, 0.998147368f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Normal3f(-0.056054458f, 0.998427689f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Normal3f(-0.176223293f, 0.984350204f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Normal3f(-0.302620173f, 0.953111231f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Normal3f(-0.429032564f, 0.90328896f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Normal3f(-0.549294412f, 0.835628867f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Normal3f(-0.66699791f, 0.74505949f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Normal3f(-0.775314748f, 0.631575108f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Normal3f(-0.863186479f, 0.504885197f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Normal3f(-0.934871018f, 0.354987651f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_Normal3f(-0.976849318f, 0.213928372f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Normal3f(-0.993807793f, 0.111112691f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Normal3f(-0.999024689f, 0.0441556387f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.999886155f, 0.0150896236f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.996150136f, -0.0876631811f, 0.f),
    C_Vertex3f(0.176078796f, 0.595449448f, 0.f),
    C_Vertex3f(0.176078796f, 0.595449448f, -1.f),
    C_Normal3f(-0.995711446f, 0.0925136656f, 0.f),
    C_Vertex3f(0.176078796f, 0.612188339f, 0.f),
    C_Vertex3f(0.176078796f, 0.612188339f, -1.f),
    C_Normal3f(-0.955586612f, 0.294710487f, 0.f),
    C_Vertex3f(0.179008484f, 0.626981258f, 0.f),
    C_Vertex3f(0.179008484f, 0.626981258f, -1.f),
    C_Normal3f(-0.873809934f, 0.486267775f, 0.f),
    C_Vertex3f(0.184867859f, 0.640686512f, 0.f),
    C_Vertex3f(0.184867859f, 0.640686512f, -1.f),
    C_Normal3f(-0.762212932f, 0.64732635f, 0.f),
    C_Vertex3f(0.193656921f, 0.6533041f, 0.f),
    C_Vertex3f(0.193656921f, 0.6533041f, -1.f),
    C_Normal3f(-0.635637224f, 0.771987855f, 0.f),
    C_Vertex3f(0.205375671f, 0.664834023f, 0.f),
    C_Vertex3f(0.205375671f, 0.664834023f, -1.f),
    C_Normal3f(-0.475808382f, 0.879548907f, 0.f),
    C_Vertex3f(0.218162537f, 0.673481464f, 0.f),
    C_Vertex3f(0.218162537f, 0.673481464f, -1.f),
    C_Normal3f(-0.287775874f, 0.957697809f, 0.f),
    C_Vertex3f(0.232017517f, 0.679246426f, 0.f),
    C_Vertex3f(0.232017517f, 0.679246426f, -1.f),
    C_Normal3f(-0.0926625952f, 0.995697558f, 0.f),
    C_Vertex3f(0.246940613f, 0.682128906f, 0.f),
    C_Vertex3f(0.246940613f, 0.682128906f, -1.f),
    C_Normal3f(0.0936417803f, 0.995605886f, 0.f),
    C_Vertex3f(0.262990952f, 0.682128906f, 0.f),
    C_Vertex3f(0.262990952f, 0.682128906f, -1.f),
    C_Normal3f(0.290848225f, 0.956769228f, 0.f),
    C_Vertex3f(0.277957439f, 0.679211617f, 0.f),
    C_Vertex3f(0.277957439f, 0.679211617f, -1.f),
    C_Normal3f(0.482152909f, 0.876087129f, 0.f),
    C_Vertex3f(0.291780949f, 0.673377037f, 0.f),
    C_Vertex3f(0.291780949f, 0.673377037f, -1.f),
    C_Normal3f(0.644631505f, 0.764493287f, 0.f),
    C_Vertex3f(0.304461479f, 0.664625168f, 0.f),
    C_Vertex3f(0.304461479f, 0.664625168f, -1.f),
    C_Normal3f(0.77254498f, 0.634960055f, 0.f),
    C_Vertex3f(0.315999031f, 0.652956009f, 0.f),
    C_Vertex3f(0.315999031f, 0.652956009f, -1.f),
    C_Normal3f(0.882203341f, 0.470868587f, 0.f),
    C_Vertex3f(0.324652195f, 0.640059471f, 0.f),
    C_Vertex3f(0.324652195f, 0.640059471f, -1.f),
    C_Normal3f(0.959506512f, 0.281686425f, 0.f),
    C_Vertex3f(0.330420971f, 0.625935555f, 0.f),
    C_Vertex3f(0.330420971f, 0.625935555f, -1.f),
    C_Normal3f(0.995984793f, 0.0895228982f, 0.f),
    C_Vertex3f(0.333305359f, 0.610584259f, 0.f),
    C_Vertex3f(0.333305359f, 0.610584259f, -1.f),
    C_Normal3f(0.996016741f, -0.0891657025f, 0.f),
    C_Vertex3f(0.333305359f, 0.593845367f, 0.f),
    C_Vertex3f(0.333305359f, 0.593845367f, -1.f),
    C_Normal3f(0.960216522f, -0.279256493f, 0.f),
    C_Vertex3f(0.330411911f, 0.578263283f, 0.f),
    C_Vertex3f(0.330411911f, 0.578263283f, -1.f),
    C_Normal3f(0.882977486f, -0.469415396f, 0.f),
    C_Vertex3f(0.324625015f, 0.563998222f, 0.f),
    C_Vertex3f(0.324625015f, 0.563998222f, -1.f),
    C_Normal3f(0.771736562f, -0.63594234f, 0.f),
    C_Vertex3f(0.315944672f, 0.551050186f, 0.f),
    C_Vertex3f(0.315944672f, 0.551050186f, -1.f),
    C_Normal3f(0.643063486f, -0.765812874f, 0.f),
    C_Vertex3f(0.30437088f, 0.539419174f, 0.f),
    C_Vertex3f(0.30437088f, 0.539419174f, -1.f),
    C_Normal3f(0.482068062f, -0.8761338f, 0.f),
    C_Vertex3f(0.291705132f, 0.530695915f, 0.f),
    C_Vertex3f(0.291705132f, 0.530695915f, -1.f),
    C_Normal3f(0.291671664f, -0.956518531f, 0.f),
    C_Vertex3f(0.277947426f, 0.524880409f, 0.f),
    C_Vertex3f(0.277947426f, 0.524880409f, -1.f),
    C_Normal3f(0.0933625251f, -0.995632172f, 0.f),
    C_Vertex3f(0.263097763f, 0.521972656f, 0.f),
    C_Vertex3f(0.263097763f, 0.521972656f, -1.f),
    C_Normal3f(-0.0934955925f, -0.995619774f, 0.f),
    C_Vertex3f(0.246938705f, 0.521972656f, 0.f),
    C_Vertex3f(0.246938705f, 0.521972656f, -1.f),
    C_Normal3f(-0.289828807f, -0.957078516f, 0.f),
    C_Vertex3f(0.231762886f, 0.524915218f, 0.f),
    C_Vertex3f(0.231762886f, 0.524915218f, -1.f),
    C_Normal3f(-0.481931925f, -0.876208663f, 0.f),
    C_Vertex3f(0.217787743f, 0.530800343f, 0.f),
    C_Vertex3f(0.217787743f, 0.530800343f, -1.f),
    C_Normal3f(-0.645861149f, -0.763454914f, 0.f),
    C_Vertex3f(0.205013275f, 0.539628029f, 0.f),
    C_Vertex3f(0.205013275f, 0.539628029f, -1.f),
    C_Normal3f(-0.776966989f, -0.629541397f, 0.f),
    C_Vertex3f(0.193439484f, 0.551398277f, 0.f),
    C_Vertex3f(0.193439484f, 0.551398277f, -1.f),
    C_Normal3f(-0.88781929f, -0.460192323f, 0.f),
    C_Vertex3f(0.18475914f, 0.564625263f, 0.f),
    C_Vertex3f(0.18475914f, 0.564625263f, -1.f),
    C_Normal3f(-0.962561071f, -0.271065116f, 0.f),
    C_Vertex3f(0.178972244f, 0.579308987f, 0.f),
    C_Vertex3f(0.178972244f, 0.579308987f, -1.f),
    C_Normal3f(-0.996150136f, -0.0876631811f, 0.f),
    C_Vertex3f(0.176078796f, 0.595449448f, 0.f),
    C_Vertex3f(0.176078796f, 0.595449448f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.99585247f, -0.0909829065f, 0.f),
    C_Vertex3f(0.209281921f, 0.608457565f, 0.f),
    C_Vertex3f(0.209281921f, 0.608457565f, -1.f),
    C_Normal3f(0.996069252f, 0.0885776132f, 0.f),
    C_Vertex3f(0.209281921f, 0.598522186f, 0.f),
    C_Vertex3f(0.209281921f, 0.598522186f, -1.f),
    C_Normal3f(0.961418629f, 0.275089741f, 0.f),
    C_Vertex3f(0.211005211f, 0.589078903f, 0.f),
    C_Vertex3f(0.211005211f, 0.589078903f, -1.f),
    C_Normal3f(0.885625243f, 0.464400709f, 0.f),
    C_Vertex3f(0.21445179f, 0.580453873f, 0.f),
    C_Vertex3f(0.21445179f, 0.580453873f, -1.f),
    C_Normal3f(0.77504015f, 0.631911993f, 0.f),
    C_Vertex3f(0.219621658f, 0.572647095f, 0.f),
    C_Vertex3f(0.219621658f, 0.572647095f, -1.f),
    C_Normal3f(0.646952569f, 0.762530267f, 0.f),
    C_Vertex3f(0.226514816f, 0.565658569f, 0.f),
    C_Vertex3f(0.226514816f, 0.565658569f, -1.f),
    C_Normal3f(0.486883491f, 0.873466849f, 0.f),
    C_Vertex3f(0.234036446f, 0.560417175f, 0.f),
    C_Vertex3f(0.234036446f, 0.560417175f, -1.f),
    C_Normal3f(0.295764416f, 0.955260932f, 0.f),
    C_Vertex3f(0.242186546f, 0.556922913f, 0.f),
    C_Vertex3f(0.242186546f, 0.556922913f, -1.f),
    C_Normal3f(0.0956318378f, 0.99541682f, 0.f),
    C_Vertex3f(0.250965118f, 0.555175781f, 0.f),
    C_Vertex3f(0.250965118f, 0.555175781f, -1.f),
    C_Normal3f(-0.0956318378f, 0.99541682f, 0.f),
    C_Vertex3f(0.260372162f, 0.555175781f, 0.f),
    C_Vertex3f(0.260372162f, 0.555175781f, -1.f),
    C_Normal3f(-0.295764416f, 0.955260932f, 0.f),
    C_Vertex3f(0.269150734f, 0.556922913f, 0.f),
    C_Vertex3f(0.269150734f, 0.556922913f, -1.f),
    C_Normal3f(-0.486883491f, 0.873466849f, 0.f),
    C_Vertex3f(0.277300835f, 0.560417175f, 0.f),
    C_Vertex3f(0.277300835f, 0.560417175f, -1.f),
    C_Normal3f(-0.646952569f, 0.762530267f, 0.f),
    C_Vertex3f(0.284822464f, 0.565658569f, 0.f),
    C_Vertex3f(0.284822464f, 0.565658569f, -1.f),
    C_Normal3f(-0.772750914f, 0.634709477f, 0.f),
    C_Vertex3f(0.291715622f, 0.572647095f, 0.f),
    C_Vertex3f(0.291715622f, 0.572647095f, -1.f),
    C_Normal3f(-0.881743789f, 0.471728742f, 0.f),
    C_Vertex3f(0.29688549f, 0.580345154f, 0.f),
    C_Vertex3f(0.29688549f, 0.580345154f, -1.f),
    C_Normal3f(-0.959135056f, 0.282949001f, 0.f),
    C_Vertex3f(0.300332069f, 0.588752747f, 0.f),
    C_Vertex3f(0.300332069f, 0.588752747f, -1.f),
    C_Normal3f(-0.995887816f, 0.0905945078f, 0.f),
    C_Vertex3f(0.302055359f, 0.597869873f, 0.f),
    C_Vertex3f(0.302055359f, 0.597869873f, -1.f),
    C_Normal3f(-0.996017992f, -0.0891522393f, 0.f),
    C_Vertex3f(0.302055359f, 0.607696533f, 0.f),
    C_Vertex3f(0.302055359f, 0.607696533f, -1.f),
    C_Normal3f(-0.960359752f, -0.278763533f, 0.f),
    C_Vertex3f(0.300359726f, 0.61681366f, 0.f),
    C_Vertex3f(0.300359726f, 0.61681366f, -1.f),
    C_Normal3f(-0.884884119f, -0.465811223f, 0.f),
    C_Vertex3f(0.29696846f, 0.625221252f, 0.f),
    C_Vertex3f(0.29696846f, 0.625221252f, -1.f),
    C_Normal3f(-0.777755082f, -0.628567517f, 0.f),
    C_Vertex3f(0.291881561f, 0.632919312f, 0.f),
    C_Vertex3f(0.291881561f, 0.632919312f, -1.f),
    C_Normal3f(-0.650574088f, -0.759442806f, 0.f),
    C_Vertex3f(0.28509903f, 0.639907837f, 0.f),
    C_Vertex3f(0.28509903f, 0.639907837f, -1.f),
    C_Normal3f(-0.486228704f, -0.87383157f, 0.f),
    C_Vertex3f(0.277605057f, 0.645149231f, 0.f),
    C_Vertex3f(0.277605057f, 0.645149231f, -1.f),
    C_Normal3f(-0.292708308f, -0.956201792f, 0.f),
    C_Vertex3f(0.269399643f, 0.648643494f, 0.f),
    C_Vertex3f(0.269399643f, 0.648643494f, -1.f),
    C_Normal3f(-0.0943521634f, -0.99553895f, 0.f),
    C_Vertex3f(0.260482788f, 0.650390625f, 0.f),
    C_Vertex3f(0.260482788f, 0.650390625f, -1.f),
    C_Normal3f(0.0935930908f, -0.995610476f, 0.f),
    C_Vertex3f(0.250965118f, 0.650390625f, 0.f),
    C_Vertex3f(0.250965118f, 0.650390625f, -1.f),
    C_Normal3f(0.291557223f, -0.9565534f, 0.f),
    C_Vertex3f(0.242186546f, 0.648670673f, 0.f),
    C_Vertex3f(0.242186546f, 0.648670673f, -1.f),
    C_Normal3f(0.481072754f, -0.876680732f, 0.f),
    C_Vertex3f(0.234036446f, 0.64523077f, 0.f),
    C_Vertex3f(0.234036446f, 0.64523077f, -1.f),
    C_Normal3f(0.641043007f, -0.767504871f, 0.f),
    C_Vertex3f(0.226514816f, 0.640070915f, 0.f),
    C_Vertex3f(0.226514816f, 0.640070915f, -1.f),
    C_Normal3f(0.767487586f, -0.641063869f, 0.f),
    C_Vertex3f(0.219621658f, 0.633191109f, 0.f),
    C_Vertex3f(0.219621658f, 0.633191109f, -1.f),
    C_Normal3f(0.878021121f, -0.478622019f, 0.f),
    C_Vertex3f(0.21445179f, 0.625628948f, 0.f),
    C_Vertex3f(0.21445179f, 0.625628948f, -1.f),
    C_Normal3f(0.957542598f, -0.288291812f, 0.f),
    C_Vertex3f(0.211005211f, 0.617384434f, 0.f),
    C_Vertex3f(0.211005211f, 0.617384434f, -1.f),
    C_Normal3f(0.99585247f, -0.0909829065f, 0.f),
    C_Vertex3f(0.209281921f, 0.608457565f, 0.f),
    C_Vertex3f(0.209281921f, 0.608457565f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, 0.f),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.355414391f, 0.310261726f, 0.f),
    C_Vertex3f(0.355305672f, 0.305113792f, 0.f),
    C_Vertex3f(0.355142593f, 0.29848671f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.163232803f, 0.189683914f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_Vertex3f(0.176846504f, 0.193832397f, 0.f),
    C_Vertex3f(0.18270874f, 0.26306057f, 0.f),
    C_Vertex3f(0.194014549f, 0.197761536f, 0.f),
    C_Vertex3f(0.203128815f, 0.265953064f, 0.f),
    C_Vertex3f(0.214736938f, 0.201471329f, 0.f),
    C_Vertex3f(0.226716995f, 0.268790245f, 0.f),
    C_Vertex3f(0.237591743f, 0.204745293f, 0.f),
    C_Vertex3f(0.248954535f, 0.271763086f, 0.f),
    C_Vertex3f(0.257684946f, 0.207966328f, 0.f),
    C_Vertex3f(0.269622087f, 0.274872541f, 0.f),
    C_Vertex3f(0.276438475f, 0.211350918f, 0.f),
    C_Vertex3f(0.288719654f, 0.27811861f, 0.f),
    C_Vertex3f(0.293852329f, 0.214899063f, 0.f),
    C_Vertex3f(0.306247234f, 0.281501293f, 0.f),
    C_Vertex3f(0.30992651f, 0.218610764f, 0.f),
    C_Vertex3f(0.322204828f, 0.28502059f, 0.f),
    C_Vertex3f(0.324661016f, 0.222486019f, 0.f),
    C_Vertex3f(0.336592436f, 0.2886765f, 0.f),
    C_Vertex3f(0.338055849f, 0.22652483f, 0.f),
    C_Vertex3f(0.349410057f, 0.292469025f, 0.f),
    C_Vertex3f(0.350111008f, 0.230727196f, 0.f),
    C_Vertex3f(0.355033875f, 0.294433594f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_Vertex3f(0.136497974f, 0.173553467f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.144261837f, 0.17993927f, 0.f),
    C_Vertex3f(0.144927979f, 0.255584717f, 0.f),
    C_Vertex3f(0.153173447f, 0.185316086f, 0.f),
    C_Vertex3f(0.150714874f, 0.257112503f, 0.f),
    C_Vertex3f(0.165237427f, 0.260113716f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.120312214f, 0.148886681f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.12441349f, 0.157754898f, 0.f),
    C_Vertex3f(0.13706398f, 0.25340271f, 0.f),
    C_Vertex3f(0.129881859f, 0.166158676f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, 0.f),
    C_Vertex3f(0.11757803f, 0.139554024f, 0.f),
    C_Vertex3f(0.122128487f, 0.248001575f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, 0.f),
    C_Vertex3f(0.060564518f, 0.204376221f, 0.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, 0.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, 0.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, 0.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, 0.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, 0.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, 0.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, 0.f),
    C_Vertex3f(0.107985497f, 0.24156332f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, 0.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, 0.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, 0.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, 0.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, 0.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, 0.f),
    C_Vertex3f(0.051609993f, 0.191689491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, 0.f),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_Vertex3f(0.348518848f, 0.360346794f, 0.f),
    C_Vertex3f(0.351298809f, 0.351857185f, 0.f),
    C_Vertex3f(0.35338378f, 0.342477798f, 0.f),
    C_Vertex3f(0.35477376f, 0.332208633f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_Vertex3f(0.329467773f, 0.386282921f, 0.f),
    C_Vertex3f(0.336009026f, 0.380476952f, 0.f),
    C_Vertex3f(0.340873957f, 0.374656677f, 0.f),
    C_Vertex3f(0.345043898f, 0.367946625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.290248394f, 0.404656887f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_Vertex3f(0.301422596f, 0.401316166f, 0.f),
    C_Vertex3f(0.320971966f, 0.469362259f, 0.f),
    C_Vertex3f(0.311683893f, 0.397140265f, 0.f),
    C_Vertex3f(0.321032286f, 0.392129183f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_Vertex3f(0.265161276f, 0.408832788f, 0.f),
    C_Vertex3f(0.278161287f, 0.407162428f, 0.f),
    C_Vertex3f(0.296843052f, 0.472860336f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_Vertex3f(0.211940885f, 0.407842636f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.223973632f, 0.409059525f, 0.f),
    C_Vertex3f(0.240123749f, 0.474609375f, 0.f),
    C_Vertex3f(0.236912727f, 0.409667969f, 0.f),
    C_Vertex3f(0.25124836f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_Vertex3f(0.181280732f, 0.400541306f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.190594435f, 0.403583527f, 0.f),
    C_Vertex3f(0.212010384f, 0.472558498f, 0.f),
    C_Vertex3f(0.200814486f, 0.406017303f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_Vertex3f(0.158777714f, 0.387763977f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.165372372f, 0.392631531f, 0.f),
    C_Vertex3f(0.185698509f, 0.468456745f, 0.f),
    C_Vertex3f(0.172873378f, 0.39689064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_Vertex3f(0.136689425f, 0.35892868f, 0.f),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.141567945f, 0.367542267f, 0.f),
    C_Vertex3f(0.161188126f, 0.462304115f, 0.f),
    C_Vertex3f(0.146875501f, 0.375219345f, 0.f),
    C_Vertex3f(0.15261209f, 0.381959915f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.124628067f, 0.327468872f, 0.f),
    C_Vertex3f(0.128219485f, 0.338891983f, 0.f),
    C_Vertex3f(0.132239938f, 0.349378586f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, 0.f),
    C_Vertex3f(0.118445396f, 0.44414711f, 0.f),
    C_Vertex3f(0.101086617f, 0.432443619f, 0.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, 0.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, 0.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, 0.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, 0.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, 0.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, 0.f),
    C_Vertex3f(0.434570313f, 0.3058815f, 0.f),
    C_Vertex3f(0.434447527f, 0.317972422f, 0.f),
    C_Vertex3f(0.434201956f, 0.329054594f, 0.f),
    C_Vertex3f(0.433833599f, 0.339128017f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, 0.f),
    C_Vertex3f(0.433342457f, 0.348192692f, 0.f),
    C_Vertex3f(0.432728529f, 0.356248617f, 0.f),
    C_Vertex3f(0.431991816f, 0.363295794f, 0.f),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, 0.f),
    C_Vertex3f(0.430641174f, 0.37184906f, 0.f),
    C_Vertex3f(0.435041308f, 0.133566856f, 0.f),
    C_Vertex3f(0.428894043f, 0.379938126f, 0.f),
    C_Vertex3f(0.435512304f, 0.113711357f, 0.f),
    C_Vertex3f(0.424089909f, 0.394722939f, 0.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, 0.f),
    C_Vertex3f(0.417975903f, 0.408114433f, 0.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, 0.f),
    C_Vertex3f(0.410552025f, 0.42011261f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_Vertex3f(0.401818275f, 0.430717468f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_Vertex3f(0.374403f, 0.00610542297f, 0.f),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.390627861f, 0.440420151f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, 0.f),
    C_Vertex3f(0.459472656f, 0.f, 0.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, 0.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, 0.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, 0.f),
    C_Vertex3f(0.440850258f, 0.048532486f, 0.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, 0.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, 0.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, 0.f),
    C_Vertex3f(0.355033875f, 0.191106796f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.353451729f, 0.168297768f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.350287437f, 0.148653984f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, 0.f),
    C_Vertex3f(0.345541f, 0.132175446f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, 0.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, 0.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, 0.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, 0.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, 0.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, 0.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, 0.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, 0.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, 0.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, 0.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, 0.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, 0.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, 0.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, 0.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, 0.f),
    C_Vertex3f(0.30825758f, 0.019092083f, 0.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, 0.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, 0.f),
    C_Vertex3f(0.327584743f, 0.101476192f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, 0.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, 0.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, 0.f),
    C_Vertex3f(0.119185448f, 0.102355957f, 0.f),
    C_Vertex3f(0.116210938f, 0.116765976f, 0.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.302055359f, 0.607696533f, 0.f),
    C_Vertex3f(0.30437088f, 0.539419174f, 0.f),
    C_Vertex3f(0.304461479f, 0.664625168f, 0.f),
    C_Vertex3f(0.315944672f, 0.551050186f, 0.f),
    C_Vertex3f(0.315999031f, 0.652956009f, 0.f),
    C_Vertex3f(0.324625015f, 0.563998222f, 0.f),
    C_Vertex3f(0.324652195f, 0.640059471f, 0.f),
    C_Vertex3f(0.330411911f, 0.578263283f, 0.f),
    C_Vertex3f(0.330420971f, 0.625935555f, 0.f),
    C_Vertex3f(0.333305359f, 0.593845367f, 0.f),
    C_Vertex3f(0.333305359f, 0.610584259f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304461479f, 0.664625168f, 0.f),
    C_Vertex3f(0.291780949f, 0.673377037f, 0.f),
    C_Vertex3f(0.291881561f, 0.632919312f, 0.f),
    C_Vertex3f(0.29696846f, 0.625221252f, 0.f),
    C_Vertex3f(0.300359726f, 0.61681366f, 0.f),
    C_Vertex3f(0.302055359f, 0.607696533f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.269399643f, 0.648643494f, 0.f),
    C_Vertex3f(0.277605057f, 0.645149231f, 0.f),
    C_Vertex3f(0.277957439f, 0.679211617f, 0.f),
    C_Vertex3f(0.28509903f, 0.639907837f, 0.f),
    C_Vertex3f(0.291780949f, 0.673377037f, 0.f),
    C_Vertex3f(0.291881561f, 0.632919312f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262990952f, 0.682128906f, 0.f),
    C_Vertex3f(0.246940613f, 0.682128906f, 0.f),
    C_Vertex3f(0.250965118f, 0.650390625f, 0.f),
    C_Vertex3f(0.260482788f, 0.650390625f, 0.f),
    C_Vertex3f(0.269399643f, 0.648643494f, 0.f),
    C_Vertex3f(0.277957439f, 0.679211617f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.219621658f, 0.633191109f, 0.f),
    C_Vertex3f(0.226514816f, 0.640070915f, 0.f),
    C_Vertex3f(0.232017517f, 0.679246426f, 0.f),
    C_Vertex3f(0.234036446f, 0.64523077f, 0.f),
    C_Vertex3f(0.246940613f, 0.682128906f, 0.f),
    C_Vertex3f(0.242186546f, 0.648670673f, 0.f),
    C_Vertex3f(0.250965118f, 0.650390625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.218162537f, 0.673481464f, 0.f),
    C_Vertex3f(0.205375671f, 0.664834023f, 0.f),
    C_Vertex3f(0.209281921f, 0.608457565f, 0.f),
    C_Vertex3f(0.211005211f, 0.617384434f, 0.f),
    C_Vertex3f(0.21445179f, 0.625628948f, 0.f),
    C_Vertex3f(0.219621658f, 0.633191109f, 0.f),
    C_Vertex3f(0.232017517f, 0.679246426f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.176078796f, 0.612188339f, 0.f),
    C_Vertex3f(0.178972244f, 0.579308987f, 0.f),
    C_Vertex3f(0.179008484f, 0.626981258f, 0.f),
    C_Vertex3f(0.18475914f, 0.564625263f, 0.f),
    C_Vertex3f(0.184867859f, 0.640686512f, 0.f),
    C_Vertex3f(0.193439484f, 0.551398277f, 0.f),
    C_Vertex3f(0.193656921f, 0.6533041f, 0.f),
    C_Vertex3f(0.205013275f, 0.539628029f, 0.f),
    C_Vertex3f(0.205375671f, 0.664834023f, 0.f),
    C_Vertex3f(0.209281921f, 0.598522186f, 0.f),
    C_Vertex3f(0.209281921f, 0.608457565f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.30437088f, 0.539419174f, 0.f),
    C_Vertex3f(0.302055359f, 0.607696533f, 0.f),
    C_Vertex3f(0.302055359f, 0.597869873f, 0.f),
    C_Vertex3f(0.300332069f, 0.588752747f, 0.f),
    C_Vertex3f(0.29688549f, 0.580345154f, 0.f),
    C_Vertex3f(0.291715622f, 0.572647095f, 0.f),
    C_Vertex3f(0.291705132f, 0.530695915f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.284822464f, 0.565658569f, 0.f),
    C_Vertex3f(0.277300835f, 0.560417175f, 0.f),
    C_Vertex3f(0.277947426f, 0.524880409f, 0.f),
    C_Vertex3f(0.291705132f, 0.530695915f, 0.f),
    C_Vertex3f(0.291715622f, 0.572647095f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269150734f, 0.556922913f, 0.f),
    C_Vertex3f(0.260372162f, 0.555175781f, 0.f),
    C_Vertex3f(0.263097763f, 0.521972656f, 0.f),
    C_Vertex3f(0.277947426f, 0.524880409f, 0.f),
    C_Vertex3f(0.277300835f, 0.560417175f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.250965118f, 0.555175781f, 0.f),
    C_Vertex3f(0.242186546f, 0.556922913f, 0.f),
    C_Vertex3f(0.246938705f, 0.521972656f, 0.f),
    C_Vertex3f(0.263097763f, 0.521972656f, 0.f),
    C_Vertex3f(0.260372162f, 0.555175781f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.234036446f, 0.560417175f, 0.f),
    C_Vertex3f(0.226514816f, 0.565658569f, 0.f),
    C_Vertex3f(0.231762886f, 0.524915218f, 0.f),
    C_Vertex3f(0.246938705f, 0.521972656f, 0.f),
    C_Vertex3f(0.242186546f, 0.556922913f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219621658f, 0.572647095f, 0.f),
    C_Vertex3f(0.21445179f, 0.580453873f, 0.f),
    C_Vertex3f(0.217787743f, 0.530800343f, 0.f),
    C_Vertex3f(0.231762886f, 0.524915218f, 0.f),
    C_Vertex3f(0.226514816f, 0.565658569f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.217787743f, 0.530800343f, 0.f),
    C_Vertex3f(0.21445179f, 0.580453873f, 0.f),
    C_Vertex3f(0.211005211f, 0.589078903f, 0.f),
    C_Vertex3f(0.209281921f, 0.598522186f, 0.f),
    C_Vertex3f(0.205013275f, 0.539628029f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.178972244f, 0.579308987f, 0.f),
    C_Vertex3f(0.176078796f, 0.612188339f, 0.f),
    C_Vertex3f(0.176078796f, 0.595449448f, 0.f),
    C_Vertex3f(0.35546875f, 0.313930511f, 0.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, 0.f),
    C_Vertex3f(0.35546875f, 0.32104969f, 0.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, 0.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, 0.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.35546875f, 0.232910156f, -1.f),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.355414391f, 0.310261726f, -1.f),
    C_Vertex3f(0.355305672f, 0.305113792f, -1.f),
    C_Vertex3f(0.355142593f, 0.29848671f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.163232803f, 0.189683914f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_Vertex3f(0.176846504f, 0.193832397f, -1.f),
    C_Vertex3f(0.18270874f, 0.26306057f, -1.f),
    C_Vertex3f(0.194014549f, 0.197761536f, -1.f),
    C_Vertex3f(0.203128815f, 0.265953064f, -1.f),
    C_Vertex3f(0.214736938f, 0.201471329f, -1.f),
    C_Vertex3f(0.226716995f, 0.268790245f, -1.f),
    C_Vertex3f(0.237591743f, 0.204745293f, -1.f),
    C_Vertex3f(0.248954535f, 0.271763086f, -1.f),
    C_Vertex3f(0.257684946f, 0.207966328f, -1.f),
    C_Vertex3f(0.269622087f, 0.274872541f, -1.f),
    C_Vertex3f(0.276438475f, 0.211350918f, -1.f),
    C_Vertex3f(0.288719654f, 0.27811861f, -1.f),
    C_Vertex3f(0.293852329f, 0.214899063f, -1.f),
    C_Vertex3f(0.306247234f, 0.281501293f, -1.f),
    C_Vertex3f(0.30992651f, 0.218610764f, -1.f),
    C_Vertex3f(0.322204828f, 0.28502059f, -1.f),
    C_Vertex3f(0.324661016f, 0.222486019f, -1.f),
    C_Vertex3f(0.336592436f, 0.2886765f, -1.f),
    C_Vertex3f(0.338055849f, 0.22652483f, -1.f),
    C_Vertex3f(0.349410057f, 0.292469025f, -1.f),
    C_Vertex3f(0.350111008f, 0.230727196f, -1.f),
    C_Vertex3f(0.355033875f, 0.294433594f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_Vertex3f(0.136497974f, 0.173553467f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.144261837f, 0.17993927f, -1.f),
    C_Vertex3f(0.144927979f, 0.255584717f, -1.f),
    C_Vertex3f(0.153173447f, 0.185316086f, -1.f),
    C_Vertex3f(0.150714874f, 0.257112503f, -1.f),
    C_Vertex3f(0.165237427f, 0.260113716f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.120312214f, 0.148886681f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.12441349f, 0.157754898f, -1.f),
    C_Vertex3f(0.13706398f, 0.25340271f, -1.f),
    C_Vertex3f(0.129881859f, 0.166158676f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.116210938f, 0.129756927f, -1.f),
    C_Vertex3f(0.11757803f, 0.139554024f, -1.f),
    C_Vertex3f(0.122128487f, 0.248001575f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.0514758825f, 0.0523393154f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_Vertex3f(0.0591756105f, 0.0417129993f, -1.f),
    C_Vertex3f(0.060564518f, 0.204376221f, -1.f),
    C_Vertex3f(0.0681586266f, 0.0317344666f, -1.f),
    C_Vertex3f(0.0707201958f, 0.215671539f, -1.f),
    C_Vertex3f(0.0784249306f, 0.022403717f, -1.f),
    C_Vertex3f(0.0820770264f, 0.225575447f, -1.f),
    C_Vertex3f(0.0897969007f, 0.0142393112f, -1.f),
    C_Vertex3f(0.0946350098f, 0.234087944f, -1.f),
    C_Vertex3f(0.102274537f, 0.00724124908f, -1.f),
    C_Vertex3f(0.107985497f, 0.24156332f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0341649055f, 0.14724493f, -1.f),
    C_Vertex3f(0.0360764265f, 0.0881049633f, -1.f),
    C_Vertex3f(0.0380415916f, 0.162796497f, -1.f),
    C_Vertex3f(0.0399262905f, 0.0755352974f, -1.f),
    C_Vertex3f(0.0438566208f, 0.177611351f, -1.f),
    C_Vertex3f(0.0450594425f, 0.0636134148f, -1.f),
    C_Vertex3f(0.051609993f, 0.191689491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360877037f, 0.457118988f, -1.f),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_Vertex3f(0.348518848f, 0.360346794f, -1.f),
    C_Vertex3f(0.351298809f, 0.351857185f, -1.f),
    C_Vertex3f(0.35338378f, 0.342477798f, -1.f),
    C_Vertex3f(0.35477376f, 0.332208633f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.342316628f, 0.464115143f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_Vertex3f(0.329467773f, 0.386282921f, -1.f),
    C_Vertex3f(0.336009026f, 0.380476952f, -1.f),
    C_Vertex3f(0.340873957f, 0.374656677f, -1.f),
    C_Vertex3f(0.345043898f, 0.367946625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.290248394f, 0.404656887f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_Vertex3f(0.301422596f, 0.401316166f, -1.f),
    C_Vertex3f(0.320971966f, 0.469362259f, -1.f),
    C_Vertex3f(0.311683893f, 0.397140265f, -1.f),
    C_Vertex3f(0.321032286f, 0.392129183f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269929886f, 0.474609375f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_Vertex3f(0.265161276f, 0.408832788f, -1.f),
    C_Vertex3f(0.278161287f, 0.407162428f, -1.f),
    C_Vertex3f(0.296843052f, 0.472860336f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_Vertex3f(0.211940885f, 0.407842636f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.223973632f, 0.409059525f, -1.f),
    C_Vertex3f(0.240123749f, 0.474609375f, -1.f),
    C_Vertex3f(0.236912727f, 0.409667969f, -1.f),
    C_Vertex3f(0.25124836f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_Vertex3f(0.181280732f, 0.400541306f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.190594435f, 0.403583527f, -1.f),
    C_Vertex3f(0.212010384f, 0.472558498f, -1.f),
    C_Vertex3f(0.200814486f, 0.406017303f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_Vertex3f(0.158777714f, 0.387763977f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.165372372f, 0.392631531f, -1.f),
    C_Vertex3f(0.185698509f, 0.468456745f, -1.f),
    C_Vertex3f(0.172873378f, 0.39689064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_Vertex3f(0.136689425f, 0.35892868f, -1.f),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.141567945f, 0.367542267f, -1.f),
    C_Vertex3f(0.161188126f, 0.462304115f, -1.f),
    C_Vertex3f(0.146875501f, 0.375219345f, -1.f),
    C_Vertex3f(0.15261209f, 0.381959915f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.138479233f, 0.454100609f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.124628067f, 0.327468872f, -1.f),
    C_Vertex3f(0.128219485f, 0.338891983f, -1.f),
    C_Vertex3f(0.132239938f, 0.349378586f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.123046875f, 0.321289063f, -1.f),
    C_Vertex3f(0.118445396f, 0.44414711f, -1.f),
    C_Vertex3f(0.101086617f, 0.432443619f, -1.f),
    C_Vertex3f(0.0864028931f, 0.418990135f, -1.f),
    C_Vertex3f(0.0743942261f, 0.403786659f, -1.f),
    C_Vertex3f(0.0640773773f, 0.386204243f, -1.f),
    C_Vertex3f(0.0554523468f, 0.366242886f, -1.f),
    C_Vertex3f(0.0485191345f, 0.343902588f, -1.f),
    C_Vertex3f(0.0458984375f, 0.331542969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434570313f, 0.180825233f, -1.f),
    C_Vertex3f(0.434570313f, 0.3058815f, -1.f),
    C_Vertex3f(0.434447527f, 0.317972422f, -1.f),
    C_Vertex3f(0.434201956f, 0.329054594f, -1.f),
    C_Vertex3f(0.433833599f, 0.339128017f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.434727311f, 0.155938148f, -1.f),
    C_Vertex3f(0.433342457f, 0.348192692f, -1.f),
    C_Vertex3f(0.432728529f, 0.356248617f, -1.f),
    C_Vertex3f(0.431991816f, 0.363295794f, -1.f),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.431132317f, 0.369334221f, -1.f),
    C_Vertex3f(0.430641174f, 0.37184906f, -1.f),
    C_Vertex3f(0.435041308f, 0.133566856f, -1.f),
    C_Vertex3f(0.428894043f, 0.379938126f, -1.f),
    C_Vertex3f(0.435512304f, 0.113711357f, -1.f),
    C_Vertex3f(0.424089909f, 0.394722939f, -1.f),
    C_Vertex3f(0.436140299f, 0.0963716507f, -1.f),
    C_Vertex3f(0.417975903f, 0.408114433f, -1.f),
    C_Vertex3f(0.436925292f, 0.0815477371f, -1.f),
    C_Vertex3f(0.410552025f, 0.42011261f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_Vertex3f(0.401818275f, 0.430717468f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_Vertex3f(0.374403f, 0.00610542297f, -1.f),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.390627861f, 0.440420151f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_Vertex3f(0.437867284f, 0.0692396164f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.377441406f, 0.f, -1.f),
    C_Vertex3f(0.459472656f, 0.f, -1.f),
    C_Vertex3f(0.455759048f, 0.00667572021f, -1.f),
    C_Vertex3f(0.449560642f, 0.0203275681f, -1.f),
    C_Vertex3f(0.444591045f, 0.0342798233f, -1.f),
    C_Vertex3f(0.440850258f, 0.048532486f, -1.f),
    C_Vertex3f(0.438966274f, 0.0594472885f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376980782f, 0.449220657f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.36268425f, 0.0489521027f, -1.f),
    C_Vertex3f(0.365505219f, 0.0336341858f, -1.f),
    C_Vertex3f(0.369411469f, 0.0193519592f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.355033875f, 0.204093933f, -1.f),
    C_Vertex3f(0.355033875f, 0.191106796f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.353451729f, 0.168297768f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.350287437f, 0.148653984f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.338228226f, 0.116895676f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.35084343f, 0.0478916168f, -1.f),
    C_Vertex3f(0.345541f, 0.132175446f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.160610199f, 0.0591287613f, -1.f),
    C_Vertex3f(0.163241744f, -0.00908756256f, -1.f),
    C_Vertex3f(0.178020477f, 0.0538892746f, -1.f),
    C_Vertex3f(0.181247711f, -0.0102539063f, -1.f),
    C_Vertex3f(0.198186874f, 0.0512695313f, -1.f),
    C_Vertex3f(0.202379227f, -0.0102539063f, -1.f),
    C_Vertex3f(0.220998764f, 0.0512695313f, -1.f),
    C_Vertex3f(0.224793911f, -0.00835180283f, -1.f),
    C_Vertex3f(0.242445469f, 0.0537543297f, -1.f),
    C_Vertex3f(0.246471882f, -0.00454759598f, -1.f),
    C_Vertex3f(0.262637615f, 0.0587239265f, -1.f),
    C_Vertex3f(0.267413139f, 0.00115871429f, -1.f),
    C_Vertex3f(0.281575203f, 0.0661783218f, -1.f),
    C_Vertex3f(0.287617683f, 0.00876712799f, -1.f),
    C_Vertex3f(0.299258232f, 0.0761175156f, -1.f),
    C_Vertex3f(0.30825758f, 0.019092083f, -1.f),
    C_Vertex3f(0.314594746f, 0.0878834724f, -1.f),
    C_Vertex3f(0.329332829f, 0.0321335793f, -1.f),
    C_Vertex3f(0.327584743f, 0.101476192f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134057999f, 0.0774669647f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.146341443f, -0.00675487518f, -1.f),
    C_Vertex3f(0.145956039f, 0.0669879913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130546808f, -0.00325584412f, -1.f),
    C_Vertex3f(0.125134468f, 0.0892562866f, -1.f),
    C_Vertex3f(0.119185448f, 0.102355957f, -1.f),
    C_Vertex3f(0.116210938f, 0.116765976f, -1.f),
    C_Vertex3f(0.11585784f, 0.00140953064f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.302055359f, 0.607696533f, -1.f),
    C_Vertex3f(0.30437088f, 0.539419174f, -1.f),
    C_Vertex3f(0.304461479f, 0.664625168f, -1.f),
    C_Vertex3f(0.315944672f, 0.551050186f, -1.f),
    C_Vertex3f(0.315999031f, 0.652956009f, -1.f),
    C_Vertex3f(0.324625015f, 0.563998222f, -1.f),
    C_Vertex3f(0.324652195f, 0.640059471f, -1.f),
    C_Vertex3f(0.330411911f, 0.578263283f, -1.f),
    C_Vertex3f(0.330420971f, 0.625935555f, -1.f),
    C_Vertex3f(0.333305359f, 0.593845367f, -1.f),
    C_Vertex3f(0.333305359f, 0.610584259f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304461479f, 0.664625168f, -1.f),
    C_Vertex3f(0.291780949f, 0.673377037f, -1.f),
    C_Vertex3f(0.291881561f, 0.632919312f, -1.f),
    C_Vertex3f(0.29696846f, 0.625221252f, -1.f),
    C_Vertex3f(0.300359726f, 0.61681366f, -1.f),
    C_Vertex3f(0.302055359f, 0.607696533f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.269399643f, 0.648643494f, -1.f),
    C_Vertex3f(0.277605057f, 0.645149231f, -1.f),
    C_Vertex3f(0.277957439f, 0.679211617f, -1.f),
    C_Vertex3f(0.28509903f, 0.639907837f, -1.f),
    C_Vertex3f(0.291780949f, 0.673377037f, -1.f),
    C_Vertex3f(0.291881561f, 0.632919312f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262990952f, 0.682128906f, -1.f),
    C_Vertex3f(0.246940613f, 0.682128906f, -1.f),
    C_Vertex3f(0.250965118f, 0.650390625f, -1.f),
    C_Vertex3f(0.260482788f, 0.650390625f, -1.f),
    C_Vertex3f(0.269399643f, 0.648643494f, -1.f),
    C_Vertex3f(0.277957439f, 0.679211617f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.219621658f, 0.633191109f, -1.f),
    C_Vertex3f(0.226514816f, 0.640070915f, -1.f),
    C_Vertex3f(0.232017517f, 0.679246426f, -1.f),
    C_Vertex3f(0.234036446f, 0.64523077f, -1.f),
    C_Vertex3f(0.246940613f, 0.682128906f, -1.f),
    C_Vertex3f(0.242186546f, 0.648670673f, -1.f),
    C_Vertex3f(0.250965118f, 0.650390625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.218162537f, 0.673481464f, -1.f),
    C_Vertex3f(0.205375671f, 0.664834023f, -1.f),
    C_Vertex3f(0.209281921f, 0.608457565f, -1.f),
    C_Vertex3f(0.211005211f, 0.617384434f, -1.f),
    C_Vertex3f(0.21445179f, 0.625628948f, -1.f),
    C_Vertex3f(0.219621658f, 0.633191109f, -1.f),
    C_Vertex3f(0.232017517f, 0.679246426f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.176078796f, 0.612188339f, -1.f),
    C_Vertex3f(0.178972244f, 0.579308987f, -1.f),
    C_Vertex3f(0.179008484f, 0.626981258f, -1.f),
    C_Vertex3f(0.18475914f, 0.564625263f, -1.f),
    C_Vertex3f(0.184867859f, 0.640686512f, -1.f),
    C_Vertex3f(0.193439484f, 0.551398277f, -1.f),
    C_Vertex3f(0.193656921f, 0.6533041f, -1.f),
    C_Vertex3f(0.205013275f, 0.539628029f, -1.f),
    C_Vertex3f(0.205375671f, 0.664834023f, -1.f),
    C_Vertex3f(0.209281921f, 0.598522186f, -1.f),
    C_Vertex3f(0.209281921f, 0.608457565f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.30437088f, 0.539419174f, -1.f),
    C_Vertex3f(0.302055359f, 0.607696533f, -1.f),
    C_Vertex3f(0.302055359f, 0.597869873f, -1.f),
    C_Vertex3f(0.300332069f, 0.588752747f, -1.f),
    C_Vertex3f(0.29688549f, 0.580345154f, -1.f),
    C_Vertex3f(0.291715622f, 0.572647095f, -1.f),
    C_Vertex3f(0.291705132f, 0.530695915f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.284822464f, 0.565658569f, -1.f),
    C_Vertex3f(0.277300835f, 0.560417175f, -1.f),
    C_Vertex3f(0.277947426f, 0.524880409f, -1.f),
    C_Vertex3f(0.291705132f, 0.530695915f, -1.f),
    C_Vertex3f(0.291715622f, 0.572647095f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.269150734f, 0.556922913f, -1.f),
    C_Vertex3f(0.260372162f, 0.555175781f, -1.f),
    C_Vertex3f(0.263097763f, 0.521972656f, -1.f),
    C_Vertex3f(0.277947426f, 0.524880409f, -1.f),
    C_Vertex3f(0.277300835f, 0.560417175f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.250965118f, 0.555175781f, -1.f),
    C_Vertex3f(0.242186546f, 0.556922913f, -1.f),
    C_Vertex3f(0.246938705f, 0.521972656f, -1.f),
    C_Vertex3f(0.263097763f, 0.521972656f, -1.f),
    C_Vertex3f(0.260372162f, 0.555175781f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.234036446f, 0.560417175f, -1.f),
    C_Vertex3f(0.226514816f, 0.565658569f, -1.f),
    C_Vertex3f(0.231762886f, 0.524915218f, -1.f),
    C_Vertex3f(0.246938705f, 0.521972656f, -1.f),
    C_Vertex3f(0.242186546f, 0.556922913f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219621658f, 0.572647095f, -1.f),
    C_Vertex3f(0.21445179f, 0.580453873f, -1.f),
    C_Vertex3f(0.217787743f, 0.530800343f, -1.f),
    C_Vertex3f(0.231762886f, 0.524915218f, -1.f),
    C_Vertex3f(0.226514816f, 0.565658569f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.217787743f, 0.530800343f, -1.f),
    C_Vertex3f(0.21445179f, 0.580453873f, -1.f),
    C_Vertex3f(0.211005211f, 0.589078903f, -1.f),
    C_Vertex3f(0.209281921f, 0.598522186f, -1.f),
    C_Vertex3f(0.205013275f, 0.539628029f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.178972244f, 0.579308987f, -1.f),
    C_Vertex3f(0.176078796f, 0.612188339f, -1.f),
    C_Vertex3f(0.176078796f, 0.595449448f, -1.f),
    C_Vertex3f(0.35546875f, 0.313930511f, -1.f),
    C_Vertex3f(0.361816406f, 0.0571289063f, -1.f),
    C_Vertex3f(0.35546875f, 0.32104969f, -1.f),
    C_Vertex3f(0.0335098505f, 0.101322412f, -1.f),
    C_Vertex3f(0.0322265625f, 0.13095665f, -1.f),
    C_Vertex3f(0.0322265625f, 0.115187645f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_230[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.957895219f, -0.287117928f, 0.f),
    C_Vertex3f(0.7578125f, 0.142089844f, 0.f),
    C_Vertex3f(0.7578125f, 0.142089844f, -1.f),
    C_Normal3f(0.944476187f, -0.328580081f, 0.f),
    C_Vertex3f(0.755081177f, 0.132977486f, 0.f),
    C_Vertex3f(0.755081177f, 0.132977486f, -1.f),
    C_Normal3f(0.924043834f, -0.382286638f, 0.f),
    C_Vertex3f(0.74866271f, 0.115789533f, 0.f),
    C_Vertex3f(0.74866271f, 0.115789533f, -1.f),
    C_Normal3f(0.893607259f, -0.448849618f, 0.f),
    C_Vertex3f(0.741288424f, 0.0996383429f, 0.f),
    C_Vertex3f(0.741288424f, 0.0996383429f, -1.f),
    C_Normal3f(0.856186152f, -0.516667485f, 0.f),
    C_Vertex3f(0.732958317f, 0.0845239162f, 0.f),
    C_Vertex3f(0.732958317f, 0.0845239162f, -1.f),
    C_Normal3f(0.811501563f, -0.584350228f, 0.f),
    C_Vertex3f(0.72367239f, 0.0704462528f, 0.f),
    C_Vertex3f(0.72367239f, 0.0704462528f, -1.f),
    C_Normal3f(0.759674013f, -0.650304079f, 0.f),
    C_Vertex3f(0.713430643f, 0.0574053526f, 0.f),
    C_Vertex3f(0.713430643f, 0.0574053526f, -1.f),
    C_Normal3f(0.701291084f, -0.712875009f, 0.f),
    C_Vertex3f(0.702233076f, 0.0454012156f, 0.f),
    C_Vertex3f(0.702233076f, 0.0454012156f, -1.f),
    C_Normal3f(0.637404978f, -0.770529032f, 0.f),
    C_Vertex3f(0.690079689f, 0.0344338417f, 0.f),
    C_Vertex3f(0.690079689f, 0.0344338417f, -1.f),
    C_Normal3f(0.567564428f, -0.823329031f, 0.f),
    C_Vertex3f(0.676970482f, 0.024503231f, 0.f),
    C_Vertex3f(0.676970482f, 0.024503231f, -1.f),
    C_Normal3f(0.491473943f, -0.870892286f, 0.f),
    C_Vertex3f(0.66306901f, 0.0158139467f, 0.f),
    C_Vertex3f(0.66306901f, 0.0158139467f, -1.f),
    C_Normal3f(0.41221422f, -0.911086917f, 0.f),
    C_Vertex3f(0.648375273f, 0.00836598873f, 0.f),
    C_Vertex3f(0.648375273f, 0.00836598873f, -1.f),
    C_Normal3f(0.331790477f, -0.943353117f, 0.f),
    C_Vertex3f(0.632889271f, 0.00215935707f, 0.f),
    C_Vertex3f(0.632889271f, 0.00215935707f, -1.f),
    C_Normal3f(0.2521393f, -0.967691004f, 0.f),
    C_Vertex3f(0.616611004f, -0.00280594826f, 0.f),
    C_Vertex3f(0.616611004f, -0.00280594826f, -1.f),
    C_Normal3f(0.174931213f, -0.984580636f, 0.f),
    C_Vertex3f(0.599540472f, -0.00652992725f, 0.f),
    C_Vertex3f(0.599540472f, -0.00652992725f, -1.f),
    C_Normal3f(0.101450786f, -0.994840562f, 0.f),
    C_Vertex3f(0.581677675f, -0.00901257992f, 0.f),
    C_Vertex3f(0.581677675f, -0.00901257992f, -1.f),
    C_Normal3f(0.0293659046f, -0.99956882f, 0.f),
    C_Vertex3f(0.563022614f, -0.0102539063f, 0.f),
    C_Vertex3f(0.563022614f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0527648143f, -0.99860698f, 0.f),
    C_Vertex3f(0.539424896f, -0.0102539063f, 0.f),
    C_Vertex3f(0.539424896f, -0.0102539063f, -1.f),
    C_Normal3f(-0.154406577f, -0.988007426f, 0.f),
    C_Vertex3f(0.513397217f, -0.00763177872f, 0.f),
    C_Vertex3f(0.513397217f, -0.00763177872f, -1.f),
    C_Normal3f(-0.269254714f, -0.963069081f, 0.f),
    C_Vertex3f(0.489089966f, -0.00238752365f, 0.f),
    C_Vertex3f(0.489089966f, -0.00238752365f, -1.f),
    C_Normal3f(-0.389115572f, -0.921188891f, 0.f),
    C_Vertex3f(0.466503143f, 0.00547885895f, 0.f),
    C_Vertex3f(0.466503143f, 0.00547885895f, -1.f),
    C_Normal3f(-0.506597817f, -0.862182438f, 0.f),
    C_Vertex3f(0.445636749f, 0.0159673691f, 0.f),
    C_Vertex3f(0.445636749f, 0.0159673691f, -1.f),
    C_Normal3f(-0.610459924f, -0.792047143f, 0.f),
    C_Vertex3f(0.42578125f, 0.0294060707f, 0.f),
    C_Vertex3f(0.42578125f, 0.0294060707f, -1.f),
    C_Normal3f(-0.697765887f, -0.716325879f, 0.f),
    C_Vertex3f(0.406936646f, 0.0457949638f, 0.f),
    C_Vertex3f(0.406936646f, 0.0457949638f, -1.f),
    C_Normal3f(-0.757827282f, -0.65245527f, 0.f),
    C_Vertex3f(0.389102936f, 0.0651340485f, 0.f),
    C_Vertex3f(0.389102936f, 0.0651340485f, -1.f),
    C_Normal3f(-0.798176825f, -0.60242337f, 0.f),
    C_Vertex3f(0.380691528f, 0.0762786865f, 0.f),
    C_Vertex3f(0.380691528f, 0.0762786865f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.700242758f, -0.713904858f, 0.f),
    C_Vertex3f(0.380691528f, 0.0762786865f, 0.f),
    C_Vertex3f(0.380691528f, 0.0762786865f, -1.f),
    C_Normal3f(0.665139079f, -0.74671948f, 0.f),
    C_Vertex3f(0.369329453f, 0.0651340485f, 0.f),
    C_Vertex3f(0.369329453f, 0.0651340485f, -1.f),
    C_Normal3f(0.614595175f, -0.788842738f, 0.f),
    C_Vertex3f(0.346468925f, 0.0457949638f, 0.f),
    C_Vertex3f(0.346468925f, 0.0457949638f, -1.f),
    C_Normal3f(0.542977393f, -0.839747429f, 0.f),
    C_Vertex3f(0.323472023f, 0.0294060707f, 0.f),
    C_Vertex3f(0.323472023f, 0.0294060707f, -1.f),
    C_Normal3f(0.458299696f, -0.888797641f, 0.f),
    C_Vertex3f(0.300338745f, 0.0159673691f, 0.f),
    C_Vertex3f(0.300338745f, 0.0159673691f, -1.f),
    C_Normal3f(0.360837102f, -0.93262887f, 0.f),
    C_Vertex3f(0.277069092f, 0.00547885895f, 0.f),
    C_Vertex3f(0.277069092f, 0.00547885895f, -1.f),
    C_Normal3f(0.257281154f, -0.966336608f, 0.f),
    C_Vertex3f(0.252898216f, -0.00238752365f, 0.f),
    C_Vertex3f(0.252898216f, -0.00238752365f, -1.f),
    C_Normal3f(0.15230763f, -0.988333166f, 0.f),
    C_Vertex3f(0.227826118f, -0.00763177872f, 0.f),
    C_Vertex3f(0.227826118f, -0.00763177872f, -1.f),
    C_Normal3f(0.0536479354f, -0.998559892f, 0.f),
    C_Vertex3f(0.201852798f, -0.0102539063f, 0.f),
    C_Vertex3f(0.201852798f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0297602583f, -0.999557078f, 0.f),
    C_Vertex3f(0.179019928f, -0.0102539063f, 0.f),
    C_Vertex3f(0.179019928f, -0.0102539063f, -1.f),
    C_Normal3f(-0.104640171f, -0.994510055f, 0.f),
    C_Vertex3f(0.161273479f, -0.0090457201f, 0.f),
    C_Vertex3f(0.161273479f, -0.0090457201f, -1.f),
    C_Normal3f(-0.183516204f, -0.983016729f, 0.f),
    C_Vertex3f(0.144571781f, -0.0066293478f, 0.f),
    C_Vertex3f(0.144571781f, -0.0066293478f, -1.f),
    C_Normal3f(-0.26909709f, -0.96311307f, 0.f),
    C_Vertex3f(0.128914833f, -0.00300478935f, 0.f),
    C_Vertex3f(0.128914833f, -0.00300478935f, -1.f),
    C_Normal3f(-0.359998405f, -0.93295294f, 0.f),
    C_Vertex3f(0.114302635f, 0.00182795525f, 0.f),
    C_Vertex3f(0.114302635f, 0.00182795525f, -1.f),
    C_Normal3f(-0.453894377f, -0.891055524f, 0.f),
    C_Vertex3f(0.100735188f, 0.00786888599f, 0.f),
    C_Vertex3f(0.100735188f, 0.00786888599f, -1.f),
    C_Normal3f(-0.547583699f, -0.836750925f, 0.f),
    C_Vertex3f(0.0882124901f, 0.0151180029f, 0.f),
    C_Vertex3f(0.0882124901f, 0.0151180029f, -1.f),
    C_Normal3f(-0.637347102f, -0.770576894f, 0.f),
    C_Vertex3f(0.0767345428f, 0.0235753059f, 0.f),
    C_Vertex3f(0.0767345428f, 0.0235753059f, -1.f),
    C_Normal3f(-0.712366641f, -0.701807439f, 0.f),
    C_Vertex3f(0.0663013458f, 0.0332407951f, 0.f),
    C_Vertex3f(0.0663013458f, 0.0332407951f, -1.f),
    C_Normal3f(-0.776724756f, -0.629840255f, 0.f),
    C_Vertex3f(0.0571722984f, 0.0434318781f, 0.f),
    C_Vertex3f(0.0571722984f, 0.0434318781f, -1.f),
    C_Normal3f(-0.83718133f, -0.546925426f, 0.f),
    C_Vertex3f(0.0493474007f, 0.0541485548f, 0.f),
    C_Vertex3f(0.0493474007f, 0.0541485548f, -1.f),
    C_Normal3f(-0.890801311f, -0.45439297f, 0.f),
    C_Vertex3f(0.0428266525f, 0.0653908253f, 0.f),
    C_Vertex3f(0.0428266525f, 0.0653908253f, -1.f),
    C_Normal3f(-0.934967399f, -0.354733676f, 0.f),
    C_Vertex3f(0.037610054f, 0.0771586895f, 0.f),
    C_Vertex3f(0.037610054f, 0.0771586895f, -1.f),
    C_Normal3f(-0.967902243f, -0.251326799f, 0.f),
    C_Vertex3f(0.0336976051f, 0.0894521475f, 0.f),
    C_Vertex3f(0.0336976051f, 0.0894521475f, -1.f),
    C_Normal3f(-0.98900342f, -0.147892907f, 0.f),
    C_Vertex3f(0.0310893059f, 0.102271199f, 0.f),
    C_Vertex3f(0.0310893059f, 0.102271199f, -1.f),
    C_Normal3f(-0.999035478f, -0.0439104736f, 0.f),
    C_Vertex3f(0.0297851563f, 0.115615845f, 0.f),
    C_Vertex3f(0.0297851563f, 0.115615845f, -1.f),
    C_Normal3f(-0.996950984f, 0.0780300274f, 0.f),
    C_Vertex3f(0.0297851563f, 0.131942749f, 0.f),
    C_Vertex3f(0.0297851563f, 0.131942749f, -1.f),
    C_Normal3f(-0.975353599f, 0.220647499f, 0.f),
    C_Vertex3f(0.0324883461f, 0.15015316f, 0.f),
    C_Vertex3f(0.0324883461f, 0.15015316f, -1.f),
    C_Normal3f(-0.931814253f, 0.362935603f, 0.f),
    C_Vertex3f(0.0378947258f, 0.167790413f, 0.f),
    C_Vertex3f(0.0378947258f, 0.167790413f, -1.f),
    C_Normal3f(-0.871046782f, 0.49120006f, 0.f),
    C_Vertex3f(0.0460042953f, 0.184854507f, 0.f),
    C_Vertex3f(0.0460042953f, 0.184854507f, -1.f),
    C_Normal3f(-0.79363656f, 0.608392f, 0.f),
    C_Vertex3f(0.0568170547f, 0.201345444f, 0.f),
    C_Vertex3f(0.0568170547f, 0.201345444f, -1.f),
    C_Normal3f(-0.69396615f, 0.720007658f, 0.f),
    C_Vertex3f(0.0697598457f, 0.215843201f, 0.f),
    C_Vertex3f(0.0697598457f, 0.215843201f, -1.f),
    C_Normal3f(-0.58060199f, 0.814187586f, 0.f),
    C_Vertex3f(0.0848326683f, 0.228347778f, 0.f),
    C_Vertex3f(0.0848326683f, 0.228347778f, -1.f),
    C_Normal3f(-0.470416039f, 0.882444859f, 0.f),
    C_Vertex3f(0.102035522f, 0.238859177f, 0.f),
    C_Vertex3f(0.102035522f, 0.238859177f, -1.f),
    C_Normal3f(-0.378051132f, 0.925784767f, 0.f),
    C_Vertex3f(0.116535187f, 0.245247841f, 0.f),
    C_Vertex3f(0.116535187f, 0.245247841f, -1.f),
    C_Normal3f(-0.318391144f, 0.947959423f, 0.f),
    C_Vertex3f(0.127641916f, 0.249315739f, 0.f),
    C_Vertex3f(0.127641916f, 0.249315739f, -1.f),
    C_Normal3f(-0.274081945f, 0.961706281f, 0.f),
    C_Vertex3f(0.140188932f, 0.253192425f, 0.f),
    C_Vertex3f(0.140188932f, 0.253192425f, -1.f),
    C_Normal3f(-0.237123668f, 0.971479475f, 0.f),
    C_Vertex3f(0.154176235f, 0.256877899f, 0.f),
    C_Vertex3f(0.154176235f, 0.256877899f, -1.f),
    C_Normal3f(-0.205960229f, 0.978560328f, 0.f),
    C_Vertex3f(0.169603825f, 0.260372162f, 0.f),
    C_Vertex3f(0.169603825f, 0.260372162f, -1.f),
    C_Normal3f(-0.179406464f, 0.98377502f, 0.f),
    C_Vertex3f(0.186471701f, 0.263675213f, 0.f),
    C_Vertex3f(0.186471701f, 0.263675213f, -1.f),
    C_Normal3f(-0.156558305f, 0.987668693f, 0.f),
    C_Vertex3f(0.204779863f, 0.266787052f, 0.f),
    C_Vertex3f(0.204779863f, 0.266787052f, -1.f),
    C_Normal3f(-0.137185022f, 0.990545452f, 0.f),
    C_Vertex3f(0.224528313f, 0.26970768f, 0.f),
    C_Vertex3f(0.224528313f, 0.26970768f, -1.f),
    C_Normal3f(-0.145702749f, 0.989328444f, 0.f),
    C_Vertex3f(0.251504898f, 0.273258209f, 0.f),
    C_Vertex3f(0.251504898f, 0.273258209f, -1.f),
    C_Normal3f(-0.17440398f, 0.984674156f, 0.f),
    C_Vertex3f(0.281648636f, 0.278120041f, 0.f),
    C_Vertex3f(0.281648636f, 0.278120041f, -1.f),
    C_Normal3f(-0.208819509f, 0.977954209f, 0.f),
    C_Vertex3f(0.309171677f, 0.283472061f, 0.f),
    C_Vertex3f(0.309171677f, 0.283472061f, -1.f),
    C_Normal3f(-0.242475212f, 0.970157564f, 0.f),
    C_Vertex3f(0.33407402f, 0.28931427f, 0.f),
    C_Vertex3f(0.33407402f, 0.28931427f, -1.f),
    C_Normal3f(-0.605100274f, 0.796149373f, 0.f),
    C_Vertex3f(0.345214844f, 0.292480469f, 0.f),
    C_Vertex3f(0.345214844f, 0.292480469f, -1.f),
    C_Normal3f(-0.99954623f, 0.0301211402f, 0.f),
    C_Vertex3f(0.345432281f, 0.29794693f, 0.f),
    C_Vertex3f(0.345432281f, 0.29794693f, -1.f),
    C_Normal3f(-0.999936819f, 0.0112369061f, 0.f),
    C_Vertex3f(0.345649719f, 0.306911469f, 0.f),
    C_Vertex3f(0.345649719f, 0.306911469f, -1.f),
    C_Normal3f(-0.999521077f, -0.0309447348f, 0.f),
    C_Vertex3f(0.345649719f, 0.317296028f, 0.f),
    C_Vertex3f(0.345649719f, 0.317296028f, -1.f),
    C_Normal3f(-0.996163905f, -0.0875072777f, 0.f),
    C_Vertex3f(0.344933748f, 0.330037475f, 0.f),
    C_Vertex3f(0.344933748f, 0.330037475f, -1.f),
    C_Normal3f(-0.987456322f, -0.15789248f, 0.f),
    C_Vertex3f(0.343501806f, 0.341747403f, 0.f),
    C_Vertex3f(0.343501806f, 0.341747403f, -1.f),
    C_Normal3f(-0.970918953f, -0.239408419f, 0.f),
    C_Vertex3f(0.341353893f, 0.352425814f, 0.f),
    C_Vertex3f(0.341353893f, 0.352425814f, -1.f),
    C_Normal3f(-0.943019032f, -0.332738996f, 0.f),
    C_Vertex3f(0.338490009f, 0.362072706f, 0.f),
    C_Vertex3f(0.338490009f, 0.362072706f, -1.f),
    C_Normal3f(-0.899344862f, -0.437240094f, 0.f),
    C_Vertex3f(0.334910154f, 0.370688081f, 0.f),
    C_Vertex3f(0.334910154f, 0.370688081f, -1.f),
    C_Normal3f(-0.835213661f, -0.549925506f, 0.f),
    C_Vertex3f(0.330614328f, 0.378271937f, 0.f),
    C_Vertex3f(0.330614328f, 0.378271937f, -1.f),
    C_Normal3f(-0.747167468f, -0.664635777f, 0.f),
    C_Vertex3f(0.325602531f, 0.384824276f, 0.f),
    C_Vertex3f(0.325602531f, 0.384824276f, -1.f),
    C_Normal3f(-0.637018561f, -0.770848572f, 0.f),
    C_Vertex3f(0.319874763f, 0.390345097f, 0.f),
    C_Vertex3f(0.319874763f, 0.390345097f, -1.f),
    C_Normal3f(-0.521776378f, -0.85308224f, 0.f),
    C_Vertex3f(0.313076258f, 0.395175815f, 0.f),
    C_Vertex3f(0.313076258f, 0.395175815f, -1.f),
    C_Normal3f(-0.411580652f, -0.911373317f, 0.f),
    C_Vertex3f(0.305207014f, 0.39931643f, 0.f),
    C_Vertex3f(0.305207014f, 0.39931643f, -1.f),
    C_Normal3f(-0.311441034f, -0.950265467f, 0.f),
    C_Vertex3f(0.296267033f, 0.402766943f, 0.f),
    C_Vertex3f(0.296267033f, 0.402766943f, -1.f),
    C_Normal3f(-0.223248571f, -0.974761605f, 0.f),
    C_Vertex3f(0.286256313f, 0.405527353f, 0.f),
    C_Vertex3f(0.286256313f, 0.405527353f, -1.f),
    C_Normal3f(-0.146902353f, -0.989150941f, 0.f),
    C_Vertex3f(0.275174856f, 0.407597661f, 0.f),
    C_Vertex3f(0.275174856f, 0.407597661f, -1.f),
    C_Normal3f(-0.0813178718f, -0.996688247f, 0.f),
    C_Vertex3f(0.263022661f, 0.408977866f, 0.f),
    C_Vertex3f(0.263022661f, 0.408977866f, -1.f),
    C_Normal3f(-0.0231030602f, -0.99973315f, 0.f),
    C_Vertex3f(0.249799728f, 0.409667969f, 0.f),
    C_Vertex3f(0.249799728f, 0.409667969f, -1.f),
    C_Normal3f(0.0336159095f, -0.999434888f, 0.f),
    C_Vertex3f(0.233160019f, 0.409667969f, 0.f),
    C_Vertex3f(0.233160019f, 0.409667969f, -1.f),
    C_Normal3f(0.102174148f, -0.994766474f, 0.f),
    C_Vertex3f(0.215647697f, 0.408519268f, 0.f),
    C_Vertex3f(0.215647697f, 0.408519268f, -1.f),
    C_Normal3f(0.184449404f, -0.982841969f, 0.f),
    C_Vertex3f(0.199608803f, 0.406221867f, 0.f),
    C_Vertex3f(0.199608803f, 0.406221867f, -1.f),
    C_Normal3f(0.279172182f, -0.960241079f, 0.f),
    C_Vertex3f(0.185043335f, 0.402775764f, 0.f),
    C_Vertex3f(0.185043335f, 0.402775764f, -1.f),
    C_Normal3f(0.383084714f, -0.923713207f, 0.f),
    C_Vertex3f(0.171951294f, 0.398180962f, 0.f),
    C_Vertex3f(0.171951294f, 0.398180962f, -1.f),
    C_Normal3f(0.497491866f, -0.867468655f, 0.f),
    C_Vertex3f(0.160577297f, 0.392629147f, 0.f),
    C_Vertex3f(0.160577297f, 0.392629147f, -1.f),
    C_Normal3f(0.621965945f, -0.783044279f, 0.f),
    C_Vertex3f(0.150921345f, 0.386120319f, 0.f),
    C_Vertex3f(0.150921345f, 0.386120319f, -1.f),
    C_Normal3f(0.746603966f, -0.665268719f, 0.f),
    C_Vertex3f(0.142983437f, 0.37865448f, 0.f),
    C_Vertex3f(0.142983437f, 0.37865448f, -1.f),
    C_Normal3f(0.851850271f, -0.523785233f, 0.f),
    C_Vertex3f(0.136763573f, 0.370231628f, 0.f),
    C_Vertex3f(0.136763573f, 0.370231628f, -1.f),
    C_Normal3f(0.91370362f, -0.406381279f, 0.f),
    C_Vertex3f(0.131062031f, 0.359266281f, 0.f),
    C_Vertex3f(0.131062031f, 0.359266281f, -1.f),
    C_Normal3f(0.948727727f, -0.316094637f, 0.f),
    C_Vertex3f(0.125878811f, 0.345758438f, 0.f),
    C_Vertex3f(0.125878811f, 0.345758438f, -1.f),
    C_Normal3f(0.966433525f, -0.256916851f, 0.f),
    C_Vertex3f(0.121213913f, 0.329708099f, 0.f),
    C_Vertex3f(0.121213913f, 0.329708099f, -1.f),
    C_Normal3f(0.976021826f, -0.217672497f, 0.f),
    C_Vertex3f(0.119140625f, 0.320411682f, 0.f),
    C_Vertex3f(0.119140625f, 0.320411682f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.118876353f, -0.992909074f, 0.f),
    C_Vertex3f(0.119140625f, 0.320411682f, 0.f),
    C_Vertex3f(0.119140625f, 0.320411682f, -1.f),
    C_Vertex3f(0.0424804688f, 0.329589844f, -1.f),
    C_Vertex3f(0.0424804688f, 0.329589844f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.972511053f, 0.232856795f, 0.f),
    C_Vertex3f(0.0424804688f, 0.329589844f, 0.f),
    C_Vertex3f(0.0424804688f, 0.329589844f, -1.f),
    C_Normal3f(-0.959204674f, 0.28271243f, 0.f),
    C_Vertex3f(0.045539856f, 0.342367172f, 0.f),
    C_Vertex3f(0.045539856f, 0.342367172f, -1.f),
    C_Normal3f(-0.935840964f, 0.352422655f, 0.f),
    C_Vertex3f(0.0529975891f, 0.365272999f, 0.f),
    C_Vertex3f(0.0529975891f, 0.365272999f, -1.f),
    C_Normal3f(-0.894429445f, 0.44720909f, 0.f),
    C_Vertex3f(0.061794281f, 0.385529995f, 0.f),
    C_Vertex3f(0.061794281f, 0.385529995f, -1.f),
    C_Normal3f(-0.833245814f, 0.552902699f, 0.f),
    C_Vertex3f(0.0719299316f, 0.403138161f, 0.f),
    C_Vertex3f(0.0719299316f, 0.403138161f, -1.f),
    C_Normal3f(-0.743850231f, 0.668346286f, 0.f),
    C_Vertex3f(0.083404541f, 0.418097496f, 0.f),
    C_Vertex3f(0.083404541f, 0.418097496f, -1.f),
    C_Normal3f(-0.636420548f, 0.771342218f, 0.f),
    C_Vertex3f(0.0973649025f, 0.431446552f, 0.f),
    C_Vertex3f(0.0973649025f, 0.431446552f, -1.f),
    C_Normal3f(-0.5257743f, 0.850624204f, 0.f),
    C_Vertex3f(0.113811016f, 0.443185329f, 0.f),
    C_Vertex3f(0.113811016f, 0.443185329f, -1.f),
    C_Normal3f(-0.419500977f, 0.907754838f, 0.f),
    C_Vertex3f(0.132742882f, 0.453313828f, 0.f),
    C_Vertex3f(0.132742882f, 0.453313828f, -1.f),
    C_Normal3f(-0.316753596f, 0.948507905f, 0.f),
    C_Vertex3f(0.1541605f, 0.461832047f, 0.f),
    C_Vertex3f(0.1541605f, 0.461832047f, -1.f),
    C_Normal3f(-0.21551995f, 0.976499379f, 0.f),
    C_Vertex3f(0.177381039f, 0.468220711f, 0.f),
    C_Vertex3f(0.177381039f, 0.468220711f, -1.f),
    C_Normal3f(-0.122289933f, 0.992494464f, 0.f),
    C_Vertex3f(0.202404499f, 0.47247982f, 0.f),
    C_Vertex3f(0.202404499f, 0.47247982f, -1.f),
    C_Normal3f(-0.0423279181f, 0.999103785f, 0.f),
    C_Vertex3f(0.229230881f, 0.474609375f, 0.f),
    C_Vertex3f(0.229230881f, 0.474609375f, -1.f),
    C_Normal3f(0.0167332757f, 0.999859929f, 0.f),
    C_Vertex3f(0.252670288f, 0.474609375f, 0.f),
    C_Vertex3f(0.252670288f, 0.474609375f, -1.f),
    C_Normal3f(0.0604321994f, 0.998172224f, 0.f),
    C_Vertex3f(0.270031929f, 0.473926544f, 0.f),
    C_Vertex3f(0.270031929f, 0.473926544f, -1.f),
    C_Normal3f(0.10589496f, 0.994377375f, 0.f),
    C_Vertex3f(0.286505699f, 0.472560883f, 0.f),
    C_Vertex3f(0.286505699f, 0.472560883f, -1.f),
    C_Normal3f(0.155903473f, 0.987772286f, 0.f),
    C_Vertex3f(0.302091599f, 0.47051239f, 0.f),
    C_Vertex3f(0.302091599f, 0.47051239f, -1.f),
    C_Normal3f(0.210728183f, 0.977544665f, 0.f),
    C_Vertex3f(0.316789627f, 0.467781067f, 0.f),
    C_Vertex3f(0.316789627f, 0.467781067f, -1.f),
    C_Normal3f(0.270499915f, 0.962720037f, 0.f),
    C_Vertex3f(0.330599785f, 0.464366913f, 0.f),
    C_Vertex3f(0.330599785f, 0.464366913f, -1.f),
    C_Normal3f(0.335120499f, 0.942175329f, 0.f),
    C_Vertex3f(0.343522072f, 0.460269928f, 0.f),
    C_Vertex3f(0.343522072f, 0.460269928f, -1.f),
    C_Normal3f(0.405768603f, 0.913975894f, 0.f),
    C_Vertex3f(0.355556488f, 0.455490112f, 0.f),
    C_Vertex3f(0.355556488f, 0.455490112f, -1.f),
    C_Normal3f(0.509773195f, 0.860308826f, 0.f),
    C_Vertex3f(0.367576599f, 0.449590683f, 0.f),
    C_Vertex3f(0.367576599f, 0.449590683f, -1.f),
    C_Normal3f(0.630297542f, 0.776353657f, 0.f),
    C_Vertex3f(0.379432678f, 0.441342354f, 0.f),
    C_Vertex3f(0.379432678f, 0.441342354f, -1.f),
    C_Normal3f(0.73354274f, 0.679643273f, 0.f),
    C_Vertex3f(0.39025116f, 0.431181908f, 0.f),
    C_Vertex3f(0.39025116f, 0.431181908f, -1.f),
    C_Normal3f(0.802944899f, 0.596053362f, 0.f),
    C_Vertex3f(0.400032043f, 0.419109344f, 0.f),
    C_Vertex3f(0.400032043f, 0.419109344f, -1.f),
    C_Normal3f(0.847920299f, 0.530123651f, 0.f),
    C_Vertex3f(0.404403687f, 0.412117004f, 0.f),
    C_Vertex3f(0.404403687f, 0.412117004f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.73805505f, 0.674740613f, 0.f),
    C_Vertex3f(0.404403687f, 0.412117004f, 0.f),
    C_Vertex3f(0.404403687f, 0.412117004f, -1.f),
    C_Normal3f(-0.693971395f, 0.720002592f, 0.f),
    C_Vertex3f(0.411396027f, 0.419765472f, 0.f),
    C_Vertex3f(0.411396027f, 0.419765472f, -1.f),
    C_Normal3f(-0.632268667f, 0.774749219f, 0.f),
    C_Vertex3f(0.426310539f, 0.433231831f, 0.f),
    C_Vertex3f(0.426310539f, 0.433231831f, -1.f),
    C_Normal3f(-0.549284399f, 0.835635424f, 0.f),
    C_Vertex3f(0.442154884f, 0.444867611f, 0.f),
    C_Vertex3f(0.442154884f, 0.444867611f, -1.f),
    C_Normal3f(-0.458330065f, 0.888782024f, 0.f),
    C_Vertex3f(0.458929062f, 0.454672813f, 0.f),
    C_Vertex3f(0.458929062f, 0.454672813f, -1.f),
    C_Normal3f(-0.350711077f, 0.936483741f, 0.f),
    C_Vertex3f(0.476633072f, 0.462647438f, 0.f),
    C_Vertex3f(0.476633072f, 0.462647438f, -1.f),
    C_Normal3f(-0.236363977f, 0.971664548f, 0.f),
    C_Vertex3f(0.496193886f, 0.468628407f, 0.f),
    C_Vertex3f(0.496193886f, 0.468628407f, -1.f),
    C_Normal3f(-0.132643744f, 0.99116385f, 0.f),
    C_Vertex3f(0.517611504f, 0.472615719f, 0.f),
    C_Vertex3f(0.517611504f, 0.472615719f, -1.f),
    C_Normal3f(-0.0445644334f, 0.99900645f, 0.f),
    C_Vertex3f(0.540885925f, 0.474609375f, 0.f),
    C_Vertex3f(0.540885925f, 0.474609375f, -1.f),
    C_Normal3f(0.0279449597f, 0.999609411f, 0.f),
    C_Vertex3f(0.562303543f, 0.474609375f, 0.f),
    C_Vertex3f(0.562303543f, 0.474609375f, -1.f),
    C_Normal3f(0.0966945365f, 0.995314062f, 0.f),
    C_Vertex3f(0.579242349f, 0.473537087f, 0.f),
    C_Vertex3f(0.579242349f, 0.473537087f, -1.f),
    C_Normal3f(0.167367861f, 0.985894442f, 0.f),
    C_Vertex3f(0.59541595f, 0.471392512f, 0.f),
    C_Vertex3f(0.59541595f, 0.471392512f, -1.f),
    C_Normal3f(0.242326438f, 0.970194817f, 0.f),
    C_Vertex3f(0.610824347f, 0.46817565f, 0.f),
    C_Vertex3f(0.610824347f, 0.46817565f, -1.f),
    C_Normal3f(0.320514679f, 0.947243512f, 0.f),
    C_Vertex3f(0.625467539f, 0.463886499f, 0.f),
    C_Vertex3f(0.625467539f, 0.463886499f, -1.f),
    C_Normal3f(0.400439829f, 0.916323006f, 0.f),
    C_Vertex3f(0.639345527f, 0.458525062f, 0.f),
    C_Vertex3f(0.639345527f, 0.458525062f, -1.f),
    C_Normal3f(0.480239362f, 0.877137601f, 0.f),
    C_Vertex3f(0.65245831f, 0.452091336f, 0.f),
    C_Vertex3f(0.65245831f, 0.452091336f, -1.f),
    C_Normal3f(0.557841897f, 0.829947233f, 0.f),
    C_Vertex3f(0.664805889f, 0.444585323f, 0.f),
    C_Vertex3f(0.664805889f, 0.444585323f, -1.f),
    C_Normal3f(0.623358667f, 0.781936049f, 0.f),
    C_Vertex3f(0.676388264f, 0.436007023f, 0.f),
    C_Vertex3f(0.676388264f, 0.436007023f, -1.f),
    C_Normal3f(0.678254604f, 0.734826982f, 0.f),
    C_Vertex3f(0.687192202f, 0.426738977f, 0.f),
    C_Vertex3f(0.687192202f, 0.426738977f, -1.f),
    C_Normal3f(0.730331242f, 0.683093131f, 0.f),
    C_Vertex3f(0.697217703f, 0.416781187f, 0.f),
    C_Vertex3f(0.697217703f, 0.416781187f, -1.f),
    C_Normal3f(0.778655231f, 0.627452016f, 0.f),
    C_Vertex3f(0.706464767f, 0.406133652f, 0.f),
    C_Vertex3f(0.706464767f, 0.406133652f, -1.f),
    C_Normal3f(0.822463393f, 0.568817973f, 0.f),
    C_Vertex3f(0.714933395f, 0.394796371f, 0.f),
    C_Vertex3f(0.714933395f, 0.394796371f, -1.f),
    C_Normal3f(0.86122185f, 0.508229017f, 0.f),
    C_Vertex3f(0.722623587f, 0.382769346f, 0.f),
    C_Vertex3f(0.722623587f, 0.382769346f, -1.f),
    C_Normal3f(0.894654036f, 0.446759552f, 0.f),
    C_Vertex3f(0.729535341f, 0.370052576f, 0.f),
    C_Vertex3f(0.729535341f, 0.370052576f, -1.f),
    C_Normal3f(0.925254107f, 0.37934798f, 0.f),
    C_Vertex3f(0.735668659f, 0.356646061f, 0.f),
    C_Vertex3f(0.735668659f, 0.356646061f, -1.f),
    C_Normal3f(0.950456321f, 0.31085813f, 0.f),
    C_Vertex3f(0.743700981f, 0.335501671f, 0.f),
    C_Vertex3f(0.743700981f, 0.335501671f, -1.f),
    C_Normal3f(0.972338259f, 0.233577162f, 0.f),
    C_Vertex3f(0.751733303f, 0.307528019f, 0.f),
    C_Vertex3f(0.751733303f, 0.307528019f, -1.f),
    C_Normal3f(0.989611864f, 0.143765241f, 0.f),
    C_Vertex3f(0.757088184f, 0.279773235f, 0.f),
    C_Vertex3f(0.757088184f, 0.279773235f, -1.f),
    C_Normal3f(0.998091221f, 0.0617572181f, 0.f),
    C_Vertex3f(0.759765625f, 0.25223732f, 0.f),
    C_Vertex3f(0.759765625f, 0.25223732f, -1.f),
    C_Normal3f(0.999986529f, -0.00518641807f, 0.f),
    C_Vertex3f(0.759765625f, 0.236501694f, 0.f),
    C_Vertex3f(0.759765625f, 0.236501694f, -1.f),
    C_Normal3f(0.999629021f, -0.0272346213f, 0.f),
    C_Vertex3f(0.759655952f, 0.231091499f, 0.f),
    C_Vertex3f(0.759655952f, 0.231091499f, -1.f),
    C_Normal3f(0.999294281f, -0.0375625417f, 0.f),
    C_Vertex3f(0.759436607f, 0.224425316f, 0.f),
    C_Vertex3f(0.759436607f, 0.224425316f, -1.f),
    C_Normal3f(0.999040842f, -0.0437875204f, 0.f),
    C_Vertex3f(0.75910759f, 0.216503143f, 0.f),
    C_Vertex3f(0.75910759f, 0.216503143f, -1.f),
    C_Normal3f(0.998859644f, -0.0477426685f, 0.f),
    C_Vertex3f(0.758888245f, 0.211914063f, 0.f),
    C_Vertex3f(0.758888245f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.758888245f, 0.211914063f, 0.f),
    C_Vertex3f(0.758888245f, 0.211914063f, -1.f),
    C_Vertex3f(0.423828125f, 0.211914063f, -1.f),
    C_Vertex3f(0.423828125f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999904394f, 0.0138235632f, 0.f),
    C_Vertex3f(0.423828125f, 0.211914063f, 0.f),
    C_Vertex3f(0.423828125f, 0.211914063f, -1.f),
    C_Normal3f(0.999436498f, 0.0335675105f, 0.f),
    C_Vertex3f(0.423936844f, 0.204050064f, 0.f),
    C_Vertex3f(0.423936844f, 0.204050064f, -1.f),
    C_Normal3f(0.998137414f, 0.0610069372f, 0.f),
    C_Vertex3f(0.42458427f, 0.189400673f, 0.f),
    C_Vertex3f(0.42458427f, 0.189400673f, -1.f),
    C_Normal3f(0.995118141f, 0.0986916199f, 0.f),
    C_Vertex3f(0.425661683f, 0.175829887f, 0.f),
    C_Vertex3f(0.425661683f, 0.175829887f, -1.f),
    C_Normal3f(0.989776671f, 0.142625645f, 0.f),
    C_Vertex3f(0.427169085f, 0.163337708f, 0.f),
    C_Vertex3f(0.427169085f, 0.163337708f, -1.f),
    C_Normal3f(0.980968595f, 0.194166616f, 0.f),
    C_Vertex3f(0.429106474f, 0.151924133f, 0.f),
    C_Vertex3f(0.429106474f, 0.151924133f, -1.f),
    C_Normal3f(0.966963589f, 0.254914671f, 0.f),
    C_Vertex3f(0.431473851f, 0.141589165f, 0.f),
    C_Vertex3f(0.431473851f, 0.141589165f, -1.f),
    C_Normal3f(0.945156395f, 0.326618195f, 0.f),
    C_Vertex3f(0.434271216f, 0.132332802f, 0.f),
    C_Vertex3f(0.434271216f, 0.132332802f, -1.f),
    C_Normal3f(0.907514632f, 0.420020312f, 0.f),
    C_Vertex3f(0.437498569f, 0.124155045f, 0.f),
    C_Vertex3f(0.437498569f, 0.124155045f, -1.f),
    C_Normal3f(0.844961107f, 0.534827709f, 0.f),
    C_Vertex3f(0.442984581f, 0.113506317f, 0.f),
    C_Vertex3f(0.442984581f, 0.113506317f, -1.f),
    C_Normal3f(0.761392355f, 0.64829129f, 0.f),
    C_Vertex3f(0.4524827f, 0.100481987f, 0.f),
    C_Vertex3f(0.4524827f, 0.100481987f, -1.f),
    C_Normal3f(0.66138494f, 0.750046551f, 0.f),
    C_Vertex3f(0.464164257f, 0.0886316299f, 0.f),
    C_Vertex3f(0.464164257f, 0.0886316299f, -1.f),
    C_Normal3f(0.559229791f, 0.829012752f, 0.f),
    C_Vertex3f(0.478029251f, 0.077955246f, 0.f),
    C_Vertex3f(0.478029251f, 0.077955246f, -1.f),
    C_Normal3f(0.451235682f, 0.892404795f, 0.f),
    C_Vertex3f(0.494077682f, 0.0684528351f, 0.f),
    C_Vertex3f(0.494077682f, 0.0684528351f, -1.f),
    C_Normal3f(0.325801194f, 0.945438206f, 0.f),
    C_Vertex3f(0.51091671f, 0.0613260269f, 0.f),
    C_Vertex3f(0.51091671f, 0.0613260269f, -1.f),
    C_Normal3f(0.193939671f, 0.981013417f, 0.f),
    C_Vertex3f(0.528546333f, 0.0565748215f, 0.f),
    C_Vertex3f(0.528546333f, 0.0565748215f, -1.f),
    C_Normal3f(0.0603845455f, 0.998175204f, 0.f),
    C_Vertex3f(0.546966553f, 0.0541992188f, 0.f),
    C_Vertex3f(0.546966553f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0772555545f, 0.997011304f, 0.f),
    C_Vertex3f(0.567815781f, 0.0541992188f, 0.f),
    C_Vertex3f(0.567815781f, 0.0541992188f, -1.f),
    C_Normal3f(-0.236098573f, 0.9717291f, 0.f),
    C_Vertex3f(0.588584423f, 0.0574240685f, 0.f),
    C_Vertex3f(0.588584423f, 0.0574240685f, -1.f),
    C_Normal3f(-0.405198783f, 0.914228618f, 0.f),
    C_Vertex3f(0.607634068f, 0.0638737679f, 0.f),
    C_Vertex3f(0.607634068f, 0.0638737679f, -1.f),
    C_Normal3f(-0.565271914f, 0.824904621f, 0.f),
    C_Vertex3f(0.624964714f, 0.073548317f, 0.f),
    C_Vertex3f(0.624964714f, 0.073548317f, -1.f),
    C_Normal3f(-0.702495396f, 0.711688399f, 0.f),
    C_Vertex3f(0.640576363f, 0.0864477158f, 0.f),
    C_Vertex3f(0.640576363f, 0.0864477158f, -1.f),
    C_Normal3f(-0.815033913f, 0.579413176f, 0.f),
    C_Vertex3f(0.653758526f, 0.101970196f, 0.f),
    C_Vertex3f(0.653758526f, 0.101970196f, -1.f),
    C_Normal3f(-0.897918165f, 0.44016245f, 0.f),
    C_Vertex3f(0.664511204f, 0.120115757f, 0.f),
    C_Vertex3f(0.664511204f, 0.120115757f, -1.f),
    C_Normal3f(-0.944694877f, 0.327950627f, 0.f),
    C_Vertex3f(0.672834396f, 0.140884399f, 0.f),
    C_Vertex3f(0.672834396f, 0.140884399f, -1.f),
    C_Normal3f(-0.969694316f, 0.244321212f, 0.f),
    C_Vertex3f(0.67578125f, 0.152580261f, 0.f),
    C_Vertex3f(0.67578125f, 0.152580261f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.126850128f, 0.991921902f, 0.f),
    C_Vertex3f(0.67578125f, 0.152580261f, 0.f),
    C_Vertex3f(0.67578125f, 0.152580261f, -1.f),
    C_Vertex3f(0.7578125f, 0.142089844f, -1.f),
    C_Vertex3f(0.7578125f, 0.142089844f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.423828125f, 0.276367188f, 0.f),
    C_Vertex3f(0.423828125f, 0.276367188f, -1.f),
    C_Vertex3f(0.678710938f, 0.276367188f, -1.f),
    C_Vertex3f(0.678710938f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.996132433f, -0.0878650025f, 0.f),
    C_Vertex3f(0.678710938f, 0.276367188f, 0.f),
    C_Vertex3f(0.678710938f, 0.276367188f, -1.f),
    C_Normal3f(-0.992103696f, -0.125420645f, 0.f),
    C_Vertex3f(0.677947044f, 0.285027504f, 0.f),
    C_Vertex3f(0.677947044f, 0.285027504f, -1.f),
    C_Normal3f(-0.984261572f, -0.176717967f, 0.f),
    C_Vertex3f(0.675573707f, 0.301183343f, 0.f),
    C_Vertex3f(0.675573707f, 0.301183343f, -1.f),
    C_Normal3f(-0.969513834f, -0.245036498f, 0.f),
    C_Vertex3f(0.672354817f, 0.316174388f, 0.f),
    C_Vertex3f(0.672354817f, 0.316174388f, -1.f),
    C_Normal3f(-0.947114229f, -0.320896506f, 0.f),
    C_Vertex3f(0.668290377f, 0.330000639f, 0.f),
    C_Vertex3f(0.668290377f, 0.330000639f, -1.f),
    C_Normal3f(-0.914812326f, -0.403879255f, 0.f),
    C_Vertex3f(0.663380384f, 0.342662096f, 0.f),
    C_Vertex3f(0.663380384f, 0.342662096f, -1.f),
    C_Normal3f(-0.870242f, -0.492624342f, 0.f),
    C_Vertex3f(0.657624841f, 0.354158759f, 0.f),
    C_Vertex3f(0.657624841f, 0.354158759f, -1.f),
    C_Normal3f(-0.811366558f, -0.584537685f, 0.f),
    C_Vertex3f(0.651023746f, 0.364490628f, 0.f),
    C_Vertex3f(0.651023746f, 0.364490628f, -1.f),
    C_Normal3f(-0.737148464f, -0.675730944f, 0.f),
    C_Vertex3f(0.643577099f, 0.373657703f, 0.f),
    C_Vertex3f(0.643577099f, 0.373657703f, -1.f),
    C_Normal3f(-0.657113314f, -0.75379169f, 0.f),
    C_Vertex3f(0.635284901f, 0.381659985f, 0.f),
    C_Vertex3f(0.635284901f, 0.381659985f, -1.f),
    C_Normal3f(-0.575750947f, -0.817625165f, 0.f),
    C_Vertex3f(0.626365304f, 0.388661981f, 0.f),
    C_Vertex3f(0.626365304f, 0.388661981f, -1.f),
    C_Normal3f(-0.487226069f, -0.873275936f, 0.f),
    C_Vertex3f(0.616818309f, 0.394663692f, 0.f),
    C_Vertex3f(0.616818309f, 0.394663692f, -1.f),
    C_Normal3f(-0.394391984f, -0.918942332f, 0.f),
    C_Vertex3f(0.606643915f, 0.399665117f, 0.f),
    C_Vertex3f(0.606643915f, 0.399665117f, -1.f),
    C_Normal3f(-0.300416797f, -0.953808069f, 0.f),
    C_Vertex3f(0.595842123f, 0.403666258f, 0.f),
    C_Vertex3f(0.595842123f, 0.403666258f, -1.f),
    C_Normal3f(-0.208284989f, -0.978068173f, 0.f),
    C_Vertex3f(0.584412932f, 0.406667113f, 0.f),
    C_Vertex3f(0.584412932f, 0.406667113f, -1.f),
    C_Normal3f(-0.120410368f, -0.99272424f, 0.f),
    C_Vertex3f(0.572356343f, 0.408667684f, 0.f),
    C_Vertex3f(0.572356343f, 0.408667684f, -1.f),
    C_Normal3f(-0.0381310247f, -0.999272823f, 0.f),
    C_Vertex3f(0.559672356f, 0.409667969f, 0.f),
    C_Vertex3f(0.559672356f, 0.409667969f, -1.f),
    C_Normal3f(0.041088976f, -0.999155521f, 0.f),
    C_Vertex3f(0.546142578f, 0.409667969f, 0.f),
    C_Vertex3f(0.546142578f, 0.409667969f, -1.f),
    C_Normal3f(0.127250656f, -0.991870642f, 0.f),
    C_Vertex3f(0.533015013f, 0.40857172f, 0.f),
    C_Vertex3f(0.533015013f, 0.40857172f, -1.f),
    C_Normal3f(0.219236031f, -0.975671828f, 0.f),
    C_Vertex3f(0.520508051f, 0.406379223f, 0.f),
    C_Vertex3f(0.520508051f, 0.406379223f, -1.f),
    C_Normal3f(0.314617366f, -0.949218512f, 0.f),
    C_Vertex3f(0.508621693f, 0.403090477f, 0.f),
    C_Vertex3f(0.508621693f, 0.403090477f, -1.f),
    C_Normal3f(0.410583735f, -0.911822855f, 0.f),
    C_Vertex3f(0.497355938f, 0.398705482f, 0.f),
    C_Vertex3f(0.497355938f, 0.398705482f, -1.f),
    C_Normal3f(0.503915012f, -0.8637532f, 0.f),
    C_Vertex3f(0.486710787f, 0.393224239f, 0.f),
    C_Vertex3f(0.486710787f, 0.393224239f, -1.f),
    C_Normal3f(0.591462731f, -0.80633229f, 0.f),
    C_Vertex3f(0.476686239f, 0.386646748f, 0.f),
    C_Vertex3f(0.476686239f, 0.386646748f, -1.f),
    C_Normal3f(0.670656919f, -0.741767764f, 0.f),
    C_Vertex3f(0.467282295f, 0.378973007f, 0.f),
    C_Vertex3f(0.467282295f, 0.378973007f, -1.f),
    C_Normal3f(0.742965996f, -0.669329226f, 0.f),
    C_Vertex3f(0.458498955f, 0.370203018f, 0.f),
    C_Vertex3f(0.458498955f, 0.370203018f, -1.f),
    C_Normal3f(0.809275925f, -0.587428629f, 0.f),
    C_Vertex3f(0.450690746f, 0.360556126f, 0.f),
    C_Vertex3f(0.450690746f, 0.360556126f, -1.f),
    C_Normal3f(0.865462422f, -0.50097394f, 0.f),
    C_Vertex3f(0.44385767f, 0.35003233f, 0.f),
    C_Vertex3f(0.44385767f, 0.35003233f, -1.f),
    C_Normal3f(0.910686493f, -0.413098097f, 0.f),
    C_Vertex3f(0.437999725f, 0.33863163f, 0.f),
    C_Vertex3f(0.437999725f, 0.33863163f, -1.f),
    C_Normal3f(0.945134521f, -0.326681435f, 0.f),
    C_Vertex3f(0.433116913f, 0.326354027f, 0.f),
    C_Vertex3f(0.433116913f, 0.326354027f, -1.f),
    C_Normal3f(0.969774187f, -0.24400419f, 0.f),
    C_Vertex3f(0.429209232f, 0.31319952f, 0.f),
    C_Vertex3f(0.429209232f, 0.31319952f, -1.f),
    C_Normal3f(0.986022949f, -0.166608974f, 0.f),
    C_Vertex3f(0.426276684f, 0.29916811f, 0.f),
    C_Vertex3f(0.426276684f, 0.29916811f, -1.f),
    C_Normal3f(0.994283259f, -0.106774673f, 0.f),
    C_Vertex3f(0.424319267f, 0.284259796f, 0.f),
    C_Vertex3f(0.424319267f, 0.284259796f, -1.f),
    C_Normal3f(0.998069465f, -0.0621079952f, 0.f),
    C_Vertex3f(0.423828125f, 0.276367188f, 0.f),
    C_Vertex3f(0.423828125f, 0.276367188f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.394825608f, -0.918756127f, 0.f),
    C_Vertex3f(0.346679688f, 0.232910156f, 0.f),
    C_Vertex3f(0.346679688f, 0.232910156f, -1.f),
    C_Normal3f(0.347726077f, -0.937596142f, 0.f),
    C_Vertex3f(0.342363358f, 0.23105526f, 0.f),
    C_Vertex3f(0.342363358f, 0.23105526f, -1.f),
    C_Normal3f(0.302959263f, -0.953003466f, 0.f),
    C_Vertex3f(0.331490755f, 0.22727704f, 0.f),
    C_Vertex3f(0.331490755f, 0.22727704f, -1.f),
    C_Normal3f(0.263070911f, -0.964776456f, 0.f),
    C_Vertex3f(0.31837821f, 0.223430395f, 0.f),
    C_Vertex3f(0.31837821f, 0.223430395f, -1.f),
    C_Normal3f(0.233143777f, -0.972442329f, 0.f),
    C_Vertex3f(0.303025723f, 0.219515324f, 0.f),
    C_Vertex3f(0.303025723f, 0.219515324f, -1.f),
    C_Normal3f(0.209924251f, -0.977717698f, 0.f),
    C_Vertex3f(0.285433292f, 0.215531826f, 0.f),
    C_Vertex3f(0.285433292f, 0.215531826f, -1.f),
    C_Normal3f(0.191414446f, -0.981509328f, 0.f),
    C_Vertex3f(0.26560092f, 0.211479902f, 0.f),
    C_Vertex3f(0.26560092f, 0.211479902f, -1.f),
    C_Normal3f(0.176328748f, -0.984331369f, 0.f),
    C_Vertex3f(0.243528605f, 0.207359552f, 0.f),
    C_Vertex3f(0.243528605f, 0.207359552f, -1.f),
    C_Normal3f(0.16287981f, -0.986645937f, 0.f),
    C_Vertex3f(0.219216347f, 0.203170776f, 0.f),
    C_Vertex3f(0.219216347f, 0.203170776f, -1.f),
    C_Normal3f(0.186306894f, -0.982491612f, 0.f),
    C_Vertex3f(0.194030762f, 0.199188232f, 0.f),
    C_Vertex3f(0.194030762f, 0.199188232f, -1.f),
    C_Normal3f(0.26591149f, -0.963997483f, 0.f),
    C_Vertex3f(0.173762321f, 0.194551468f, 0.f),
    C_Vertex3f(0.173762321f, 0.194551468f, -1.f),
    C_Normal3f(0.374225736f, -0.927337646f, 0.f),
    C_Vertex3f(0.157044411f, 0.188985825f, 0.f),
    C_Vertex3f(0.157044411f, 0.188985825f, -1.f),
    C_Normal3f(0.493742287f, -0.869608343f, 0.f),
    C_Vertex3f(0.143877029f, 0.182491302f, 0.f),
    C_Vertex3f(0.143877029f, 0.182491302f, -1.f),
    C_Normal3f(0.617697954f, -0.786415458f, 0.f),
    C_Vertex3f(0.139068604f, 0.178779602f, 0.f),
    C_Vertex3f(0.139068604f, 0.178779602f, -1.f),
    C_Normal3f(0.725067616f, -0.688677728f, 0.f),
    C_Vertex3f(0.132621765f, 0.173650742f, 0.f),
    C_Vertex3f(0.132621765f, 0.173650742f, -1.f),
    C_Normal3f(0.843035817f, -0.537857413f, 0.f),
    C_Vertex3f(0.122951508f, 0.161810875f, 0.f),
    C_Vertex3f(0.122951508f, 0.161810875f, -1.f),
    C_Normal3f(0.946718335f, -0.322062731f, 0.f),
    C_Vertex3f(0.116504669f, 0.148388863f, 0.f),
    C_Vertex3f(0.116504669f, 0.148388863f, -1.f),
    C_Normal3f(0.994652808f, -0.103275478f, 0.f),
    C_Vertex3f(0.11328125f, 0.133384705f, 0.f),
    C_Vertex3f(0.11328125f, 0.133384705f, -1.f),
    C_Normal3f(0.995466709f, 0.0951113775f, 0.f),
    C_Vertex3f(0.11328125f, 0.117343903f, 0.f),
    C_Vertex3f(0.11328125f, 0.117343903f, -1.f),
    C_Normal3f(0.952941477f, 0.3031542f, 0.f),
    C_Vertex3f(0.116177082f, 0.103075981f, 0.f),
    C_Vertex3f(0.116177082f, 0.103075981f, -1.f),
    C_Normal3f(0.86406374f, 0.503382385f, 0.f),
    C_Vertex3f(0.121968746f, 0.0900354385f, 0.f),
    C_Vertex3f(0.121968746f, 0.0900354385f, -1.f),
    C_Normal3f(0.741451621f, 0.671006322f, 0.f),
    C_Vertex3f(0.130656242f, 0.0782222748f, 0.f),
    C_Vertex3f(0.130656242f, 0.0782222748f, -1.f),
    C_Normal3f(0.583416641f, 0.812173009f, 0.f),
    C_Vertex3f(0.142239571f, 0.0676364899f, 0.f),
    C_Vertex3f(0.142239571f, 0.0676364899f, -1.f),
    C_Normal3f(0.392224729f, 0.919869483f, 0.f),
    C_Vertex3f(0.156445026f, 0.0596971512f, 0.f),
    C_Vertex3f(0.156445026f, 0.0596971512f, -1.f),
    C_Normal3f(0.21379146f, 0.976879299f, 0.f),
    C_Vertex3f(0.17327261f, 0.0544042587f, 0.f),
    C_Vertex3f(0.17327261f, 0.0544042587f, -1.f),
    C_Normal3f(0.0613561869f, 0.998115957f, 0.f),
    C_Vertex3f(0.192722321f, 0.0517578125f, 0.f),
    C_Vertex3f(0.192722321f, 0.0517578125f, -1.f),
    C_Normal3f(-0.0658555105f, 0.997829139f, 0.f),
    C_Vertex3f(0.216323853f, 0.0517578125f, 0.f),
    C_Vertex3f(0.216323853f, 0.0517578125f, -1.f),
    C_Normal3f(-0.201179415f, 0.979554355f, 0.f),
    C_Vertex3f(0.239843369f, 0.0548677444f, 0.f),
    C_Vertex3f(0.239843369f, 0.0548677444f, -1.f),
    C_Normal3f(-0.345722049f, 0.938336909f, 0.f),
    C_Vertex3f(0.261751175f, 0.0610876083f, 0.f),
    C_Vertex3f(0.261751175f, 0.0610876083f, -1.f),
    C_Normal3f(-0.48758769f, 0.873073995f, 0.f),
    C_Vertex3f(0.282047272f, 0.0704174042f, 0.f),
    C_Vertex3f(0.282047272f, 0.0704174042f, -1.f),
    C_Normal3f(-0.61208272f, 0.790793777f, 0.f),
    C_Vertex3f(0.300731659f, 0.082857132f, 0.f),
    C_Vertex3f(0.300731659f, 0.082857132f, -1.f),
    C_Normal3f(-0.729607165f, 0.683866501f, 0.f),
    C_Vertex3f(0.316165924f, 0.0968255997f, 0.f),
    C_Vertex3f(0.316165924f, 0.0968255997f, -1.f),
    C_Normal3f(-0.838701427f, 0.544591486f, 0.f),
    C_Vertex3f(0.328350067f, 0.112322807f, 0.f),
    C_Vertex3f(0.328350067f, 0.112322807f, -1.f),
    C_Normal3f(-0.912705541f, 0.408617824f, 0.f),
    C_Vertex3f(0.337284088f, 0.129348755f, 0.f),
    C_Vertex3f(0.337284088f, 0.129348755f, -1.f),
    C_Normal3f(-0.957241535f, 0.289289951f, 0.f),
    C_Vertex3f(0.340126038f, 0.138626099f, 0.f),
    C_Vertex3f(0.340126038f, 0.138626099f, -1.f),
    C_Normal3f(-0.975735188f, 0.218954161f, 0.f),
    C_Vertex3f(0.340945244f, 0.14146328f, 0.f),
    C_Vertex3f(0.340945244f, 0.14146328f, -1.f),
    C_Normal3f(-0.986322761f, 0.164825007f, 0.f),
    C_Vertex3f(0.342378855f, 0.148665428f, 0.f),
    C_Vertex3f(0.342378855f, 0.148665428f, -1.f),
    C_Normal3f(-0.993034959f, 0.117820032f, 0.f),
    C_Vertex3f(0.343607664f, 0.157395363f, 0.f),
    C_Vertex3f(0.343607664f, 0.157395363f, -1.f),
    C_Normal3f(-0.996522188f, 0.0833273456f, 0.f),
    C_Vertex3f(0.344631672f, 0.167653084f, 0.f),
    C_Vertex3f(0.344631672f, 0.167653084f, -1.f),
    C_Normal3f(-0.998372734f, 0.057025753f, 0.f),
    C_Vertex3f(0.345450878f, 0.179438591f, 0.f),
    C_Vertex3f(0.345450878f, 0.179438591f, -1.f),
    C_Normal3f(-0.999339223f, 0.0363471471f, 0.f),
    C_Vertex3f(0.346065283f, 0.192751884f, 0.f),
    C_Vertex3f(0.346065283f, 0.192751884f, -1.f),
    C_Normal3f(-0.999806225f, 0.0196823664f, 0.f),
    C_Vertex3f(0.346474886f, 0.207592964f, 0.f),
    C_Vertex3f(0.346474886f, 0.207592964f, -1.f),
    C_Normal3f(-0.999967277f, 0.00808916148f, 0.f),
    C_Vertex3f(0.346679688f, 0.22396183f, 0.f),
    C_Vertex3f(0.346679688f, 0.22396183f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.346679688f, 0.232910156f, 0.f),
    C_Vertex3f(0.346679688f, 0.232910156f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.346679688f, 0.232910156f, 0.f),
    C_Vertex3f(0.346679688f, 0.22396183f, 0.f),
    C_Vertex3f(0.369329453f, 0.0651340485f, 0.f),
    C_Vertex3f(0.355556488f, 0.455490112f, 0.f),
    C_Vertex3f(0.345649719f, 0.317296028f, 0.f),
    C_Vertex3f(0.345649719f, 0.306911469f, 0.f),
    C_Vertex3f(0.345432281f, 0.29794693f, 0.f),
    C_Vertex3f(0.345214844f, 0.292480469f, 0.f),
    C_Vertex3f(0.342363358f, 0.23105526f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.341353893f, 0.352425814f, 0.f),
    C_Vertex3f(0.343501806f, 0.341747403f, 0.f),
    C_Vertex3f(0.343522072f, 0.460269928f, 0.f),
    C_Vertex3f(0.344933748f, 0.330037475f, 0.f),
    C_Vertex3f(0.355556488f, 0.455490112f, 0.f),
    C_Vertex3f(0.345649719f, 0.317296028f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.343522072f, 0.460269928f, 0.f),
    C_Vertex3f(0.330599785f, 0.464366913f, 0.f),
    C_Vertex3f(0.330614328f, 0.378271937f, 0.f),
    C_Vertex3f(0.334910154f, 0.370688081f, 0.f),
    C_Vertex3f(0.338490009f, 0.362072706f, 0.f),
    C_Vertex3f(0.341353893f, 0.352425814f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.305207014f, 0.39931643f, 0.f),
    C_Vertex3f(0.313076258f, 0.395175815f, 0.f),
    C_Vertex3f(0.316789627f, 0.467781067f, 0.f),
    C_Vertex3f(0.319874763f, 0.390345097f, 0.f),
    C_Vertex3f(0.330599785f, 0.464366913f, 0.f),
    C_Vertex3f(0.325602531f, 0.384824276f, 0.f),
    C_Vertex3f(0.330614328f, 0.378271937f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.275174856f, 0.407597661f, 0.f),
    C_Vertex3f(0.286256313f, 0.405527353f, 0.f),
    C_Vertex3f(0.286505699f, 0.472560883f, 0.f),
    C_Vertex3f(0.296267033f, 0.402766943f, 0.f),
    C_Vertex3f(0.302091599f, 0.47051239f, 0.f),
    C_Vertex3f(0.305207014f, 0.39931643f, 0.f),
    C_Vertex3f(0.316789627f, 0.467781067f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.233160019f, 0.409667969f, 0.f),
    C_Vertex3f(0.249799728f, 0.409667969f, 0.f),
    C_Vertex3f(0.252670288f, 0.474609375f, 0.f),
    C_Vertex3f(0.263022661f, 0.408977866f, 0.f),
    C_Vertex3f(0.270031929f, 0.473926544f, 0.f),
    C_Vertex3f(0.275174856f, 0.407597661f, 0.f),
    C_Vertex3f(0.286505699f, 0.472560883f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.185043335f, 0.402775764f, 0.f),
    C_Vertex3f(0.199608803f, 0.406221867f, 0.f),
    C_Vertex3f(0.202404499f, 0.47247982f, 0.f),
    C_Vertex3f(0.215647697f, 0.408519268f, 0.f),
    C_Vertex3f(0.229230881f, 0.474609375f, 0.f),
    C_Vertex3f(0.233160019f, 0.409667969f, 0.f),
    C_Vertex3f(0.252670288f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.177381039f, 0.468220711f, 0.f),
    C_Vertex3f(0.1541605f, 0.461832047f, 0.f),
    C_Vertex3f(0.160577297f, 0.392629147f, 0.f),
    C_Vertex3f(0.171951294f, 0.398180962f, 0.f),
    C_Vertex3f(0.185043335f, 0.402775764f, 0.f),
    C_Vertex3f(0.202404499f, 0.47247982f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1541605f, 0.461832047f, 0.f),
    C_Vertex3f(0.132742882f, 0.453313828f, 0.f),
    C_Vertex3f(0.136763573f, 0.370231628f, 0.f),
    C_Vertex3f(0.142983437f, 0.37865448f, 0.f),
    C_Vertex3f(0.150921345f, 0.386120319f, 0.f),
    C_Vertex3f(0.160577297f, 0.392629147f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.132742882f, 0.453313828f, 0.f),
    C_Vertex3f(0.113811016f, 0.443185329f, 0.f),
    C_Vertex3f(0.119140625f, 0.320411682f, 0.f),
    C_Vertex3f(0.121213913f, 0.329708099f, 0.f),
    C_Vertex3f(0.125878811f, 0.345758438f, 0.f),
    C_Vertex3f(0.131062031f, 0.359266281f, 0.f),
    C_Vertex3f(0.136763573f, 0.370231628f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.119140625f, 0.320411682f, 0.f),
    C_Vertex3f(0.113811016f, 0.443185329f, 0.f),
    C_Vertex3f(0.0973649025f, 0.431446552f, 0.f),
    C_Vertex3f(0.083404541f, 0.418097496f, 0.f),
    C_Vertex3f(0.0719299316f, 0.403138161f, 0.f),
    C_Vertex3f(0.061794281f, 0.385529995f, 0.f),
    C_Vertex3f(0.0529975891f, 0.365272999f, 0.f),
    C_Vertex3f(0.045539856f, 0.342367172f, 0.f),
    C_Vertex3f(0.0424804688f, 0.329589844f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.7578125f, 0.142089844f, 0.f),
    C_Vertex3f(0.67578125f, 0.152580261f, 0.f),
    C_Vertex3f(0.676970482f, 0.024503231f, 0.f),
    C_Vertex3f(0.690079689f, 0.0344338417f, 0.f),
    C_Vertex3f(0.702233076f, 0.0454012156f, 0.f),
    C_Vertex3f(0.713430643f, 0.0574053526f, 0.f),
    C_Vertex3f(0.72367239f, 0.0704462528f, 0.f),
    C_Vertex3f(0.732958317f, 0.0845239162f, 0.f),
    C_Vertex3f(0.741288424f, 0.0996383429f, 0.f),
    C_Vertex3f(0.74866271f, 0.115789533f, 0.f),
    C_Vertex3f(0.755081177f, 0.132977486f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.67578125f, 0.152580261f, 0.f),
    C_Vertex3f(0.672834396f, 0.140884399f, 0.f),
    C_Vertex3f(0.676970482f, 0.024503231f, 0.f),
    C_Vertex3f(0.664511204f, 0.120115757f, 0.f),
    C_Vertex3f(0.66306901f, 0.0158139467f, 0.f),
    C_Vertex3f(0.653758526f, 0.101970196f, 0.f),
    C_Vertex3f(0.648375273f, 0.00836598873f, 0.f),
    C_Vertex3f(0.640576363f, 0.0864477158f, 0.f),
    C_Vertex3f(0.632889271f, 0.00215935707f, 0.f),
    C_Vertex3f(0.624964714f, 0.073548317f, 0.f),
    C_Vertex3f(0.616611004f, -0.00280594826f, 0.f),
    C_Vertex3f(0.607634068f, 0.0638737679f, 0.f),
    C_Vertex3f(0.599540472f, -0.00652992725f, 0.f),
    C_Vertex3f(0.588584423f, 0.0574240685f, 0.f),
    C_Vertex3f(0.581677675f, -0.00901257992f, 0.f),
    C_Vertex3f(0.567815781f, 0.0541992188f, 0.f),
    C_Vertex3f(0.563022614f, -0.0102539063f, 0.f),
    C_Vertex3f(0.546966553f, 0.0541992188f, 0.f),
    C_Vertex3f(0.539424896f, -0.0102539063f, 0.f),
    C_Vertex3f(0.528546333f, 0.0565748215f, 0.f),
    C_Vertex3f(0.513397217f, -0.00763177872f, 0.f),
    C_Vertex3f(0.51091671f, 0.0613260269f, 0.f),
    C_Vertex3f(0.494077682f, 0.0684528351f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.513397217f, -0.00763177872f, 0.f),
    C_Vertex3f(0.494077682f, 0.0684528351f, 0.f),
    C_Vertex3f(0.489089966f, -0.00238752365f, 0.f),
    C_Vertex3f(0.478029251f, 0.077955246f, 0.f),
    C_Vertex3f(0.466503143f, 0.00547885895f, 0.f),
    C_Vertex3f(0.464164257f, 0.0886316299f, 0.f),
    C_Vertex3f(0.4524827f, 0.100481987f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.445636749f, 0.0159673691f, 0.f),
    C_Vertex3f(0.466503143f, 0.00547885895f, 0.f),
    C_Vertex3f(0.4524827f, 0.100481987f, 0.f),
    C_Vertex3f(0.442984581f, 0.113506317f, 0.f),
    C_Vertex3f(0.437498569f, 0.124155045f, 0.f),
    C_Vertex3f(0.434271216f, 0.132332802f, 0.f),
    C_Vertex3f(0.431473851f, 0.141589165f, 0.f),
    C_Vertex3f(0.429106474f, 0.151924133f, 0.f),
    C_Vertex3f(0.427169085f, 0.163337708f, 0.f),
    C_Vertex3f(0.42578125f, 0.0294060707f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.42578125f, 0.0294060707f, 0.f),
    C_Vertex3f(0.427169085f, 0.163337708f, 0.f),
    C_Vertex3f(0.425661683f, 0.175829887f, 0.f),
    C_Vertex3f(0.42458427f, 0.189400673f, 0.f),
    C_Vertex3f(0.423936844f, 0.204050064f, 0.f),
    C_Vertex3f(0.423828125f, 0.211914063f, 0.f),
    C_Vertex3f(0.411396027f, 0.419765472f, 0.f),
    C_Vertex3f(0.406936646f, 0.0457949638f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.406936646f, 0.0457949638f, 0.f),
    C_Vertex3f(0.411396027f, 0.419765472f, 0.f),
    C_Vertex3f(0.404403687f, 0.412117004f, 0.f),
    C_Vertex3f(0.400032043f, 0.419109344f, 0.f),
    C_Vertex3f(0.39025116f, 0.431181908f, 0.f),
    C_Vertex3f(0.389102936f, 0.0651340485f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389102936f, 0.0651340485f, 0.f),
    C_Vertex3f(0.39025116f, 0.431181908f, 0.f),
    C_Vertex3f(0.380691528f, 0.0762786865f, 0.f),
    C_Vertex3f(0.379432678f, 0.441342354f, 0.f),
    C_Vertex3f(0.369329453f, 0.0651340485f, 0.f),
    C_Vertex3f(0.367576599f, 0.449590683f, 0.f),
    C_Vertex3f(0.355556488f, 0.455490112f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.346474886f, 0.207592964f, 0.f),
    C_Vertex3f(0.346065283f, 0.192751884f, 0.f),
    C_Vertex3f(0.346468925f, 0.0457949638f, 0.f),
    C_Vertex3f(0.369329453f, 0.0651340485f, 0.f),
    C_Vertex3f(0.346679688f, 0.22396183f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.346468925f, 0.0457949638f, 0.f),
    C_Vertex3f(0.346065283f, 0.192751884f, 0.f),
    C_Vertex3f(0.345450878f, 0.179438591f, 0.f),
    C_Vertex3f(0.344631672f, 0.167653084f, 0.f),
    C_Vertex3f(0.343607664f, 0.157395363f, 0.f),
    C_Vertex3f(0.342378855f, 0.148665428f, 0.f),
    C_Vertex3f(0.340945244f, 0.14146328f, 0.f),
    C_Vertex3f(0.340126038f, 0.138626099f, 0.f),
    C_Vertex3f(0.337284088f, 0.129348755f, 0.f),
    C_Vertex3f(0.328350067f, 0.112322807f, 0.f),
    C_Vertex3f(0.323472023f, 0.0294060707f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.328350067f, 0.112322807f, 0.f),
    C_Vertex3f(0.316165924f, 0.0968255997f, 0.f),
    C_Vertex3f(0.323472023f, 0.0294060707f, 0.f),
    C_Vertex3f(0.300731659f, 0.082857132f, 0.f),
    C_Vertex3f(0.300338745f, 0.0159673691f, 0.f),
    C_Vertex3f(0.282047272f, 0.0704174042f, 0.f),
    C_Vertex3f(0.277069092f, 0.00547885895f, 0.f),
    C_Vertex3f(0.261751175f, 0.0610876083f, 0.f),
    C_Vertex3f(0.252898216f, -0.00238752365f, 0.f),
    C_Vertex3f(0.239843369f, 0.0548677444f, 0.f),
    C_Vertex3f(0.227826118f, -0.00763177872f, 0.f),
    C_Vertex3f(0.216323853f, 0.0517578125f, 0.f),
    C_Vertex3f(0.201852798f, -0.0102539063f, 0.f),
    C_Vertex3f(0.192722321f, 0.0517578125f, 0.f),
    C_Vertex3f(0.179019928f, -0.0102539063f, 0.f),
    C_Vertex3f(0.17327261f, 0.0544042587f, 0.f),
    C_Vertex3f(0.161273479f, -0.0090457201f, 0.f),
    C_Vertex3f(0.156445026f, 0.0596971512f, 0.f),
    C_Vertex3f(0.144571781f, -0.0066293478f, 0.f),
    C_Vertex3f(0.142239571f, 0.0676364899f, 0.f),
    C_Vertex3f(0.130656242f, 0.0782222748f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.128914833f, -0.00300478935f, 0.f),
    C_Vertex3f(0.144571781f, -0.0066293478f, 0.f),
    C_Vertex3f(0.130656242f, 0.0782222748f, 0.f),
    C_Vertex3f(0.121968746f, 0.0900354385f, 0.f),
    C_Vertex3f(0.116177082f, 0.103075981f, 0.f),
    C_Vertex3f(0.114302635f, 0.00182795525f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116504669f, 0.148388863f, 0.f),
    C_Vertex3f(0.116535187f, 0.245247841f, 0.f),
    C_Vertex3f(0.11328125f, 0.133384705f, 0.f),
    C_Vertex3f(0.102035522f, 0.238859177f, 0.f),
    C_Vertex3f(0.11328125f, 0.117343903f, 0.f),
    C_Vertex3f(0.114302635f, 0.00182795525f, 0.f),
    C_Vertex3f(0.116177082f, 0.103075981f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0571722984f, 0.0434318781f, 0.f),
    C_Vertex3f(0.0663013458f, 0.0332407951f, 0.f),
    C_Vertex3f(0.0697598457f, 0.215843201f, 0.f),
    C_Vertex3f(0.0767345428f, 0.0235753059f, 0.f),
    C_Vertex3f(0.0848326683f, 0.228347778f, 0.f),
    C_Vertex3f(0.0882124901f, 0.0151180029f, 0.f),
    C_Vertex3f(0.102035522f, 0.238859177f, 0.f),
    C_Vertex3f(0.100735188f, 0.00786888599f, 0.f),
    C_Vertex3f(0.114302635f, 0.00182795525f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0336976051f, 0.0894521475f, 0.f),
    C_Vertex3f(0.037610054f, 0.0771586895f, 0.f),
    C_Vertex3f(0.0378947258f, 0.167790413f, 0.f),
    C_Vertex3f(0.0428266525f, 0.0653908253f, 0.f),
    C_Vertex3f(0.0460042953f, 0.184854507f, 0.f),
    C_Vertex3f(0.0493474007f, 0.0541485548f, 0.f),
    C_Vertex3f(0.0568170547f, 0.201345444f, 0.f),
    C_Vertex3f(0.0571722984f, 0.0434318781f, 0.f),
    C_Vertex3f(0.0697598457f, 0.215843201f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0324883461f, 0.15015316f, 0.f),
    C_Vertex3f(0.0297851563f, 0.131942749f, 0.f),
    C_Vertex3f(0.0310893059f, 0.102271199f, 0.f),
    C_Vertex3f(0.0336976051f, 0.0894521475f, 0.f),
    C_Vertex3f(0.0378947258f, 0.167790413f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.672354817f, 0.316174388f, 0.f),
    C_Vertex3f(0.675573707f, 0.301183343f, 0.f),
    C_Vertex3f(0.676388264f, 0.436007023f, 0.f),
    C_Vertex3f(0.677947044f, 0.285027504f, 0.f),
    C_Vertex3f(0.687192202f, 0.426738977f, 0.f),
    C_Vertex3f(0.678710938f, 0.276367188f, 0.f),
    C_Vertex3f(0.758888245f, 0.211914063f, 0.f),
    C_Vertex3f(0.423828125f, 0.276367188f, 0.f),
    C_Vertex3f(0.423828125f, 0.211914063f, 0.f),
    C_Vertex3f(0.411396027f, 0.419765472f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.657624841f, 0.354158759f, 0.f),
    C_Vertex3f(0.663380384f, 0.342662096f, 0.f),
    C_Vertex3f(0.664805889f, 0.444585323f, 0.f),
    C_Vertex3f(0.668290377f, 0.330000639f, 0.f),
    C_Vertex3f(0.676388264f, 0.436007023f, 0.f),
    C_Vertex3f(0.672354817f, 0.316174388f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.65245831f, 0.452091336f, 0.f),
    C_Vertex3f(0.639345527f, 0.458525062f, 0.f),
    C_Vertex3f(0.643577099f, 0.373657703f, 0.f),
    C_Vertex3f(0.651023746f, 0.364490628f, 0.f),
    C_Vertex3f(0.657624841f, 0.354158759f, 0.f),
    C_Vertex3f(0.664805889f, 0.444585323f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.595842123f, 0.403666258f, 0.f),
    C_Vertex3f(0.606643915f, 0.399665117f, 0.f),
    C_Vertex3f(0.610824347f, 0.46817565f, 0.f),
    C_Vertex3f(0.616818309f, 0.394663692f, 0.f),
    C_Vertex3f(0.625467539f, 0.463886499f, 0.f),
    C_Vertex3f(0.626365304f, 0.388661981f, 0.f),
    C_Vertex3f(0.639345527f, 0.458525062f, 0.f),
    C_Vertex3f(0.635284901f, 0.381659985f, 0.f),
    C_Vertex3f(0.643577099f, 0.373657703f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.546142578f, 0.409667969f, 0.f),
    C_Vertex3f(0.559672356f, 0.409667969f, 0.f),
    C_Vertex3f(0.562303543f, 0.474609375f, 0.f),
    C_Vertex3f(0.572356343f, 0.408667684f, 0.f),
    C_Vertex3f(0.579242349f, 0.473537087f, 0.f),
    C_Vertex3f(0.584412932f, 0.406667113f, 0.f),
    C_Vertex3f(0.59541595f, 0.471392512f, 0.f),
    C_Vertex3f(0.595842123f, 0.403666258f, 0.f),
    C_Vertex3f(0.610824347f, 0.46817565f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.540885925f, 0.474609375f, 0.f),
    C_Vertex3f(0.517611504f, 0.472615719f, 0.f),
    C_Vertex3f(0.520508051f, 0.406379223f, 0.f),
    C_Vertex3f(0.533015013f, 0.40857172f, 0.f),
    C_Vertex3f(0.546142578f, 0.409667969f, 0.f),
    C_Vertex3f(0.562303543f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.476686239f, 0.386646748f, 0.f),
    C_Vertex3f(0.486710787f, 0.393224239f, 0.f),
    C_Vertex3f(0.496193886f, 0.468628407f, 0.f),
    C_Vertex3f(0.497355938f, 0.398705482f, 0.f),
    C_Vertex3f(0.517611504f, 0.472615719f, 0.f),
    C_Vertex3f(0.508621693f, 0.403090477f, 0.f),
    C_Vertex3f(0.520508051f, 0.406379223f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.450690746f, 0.360556126f, 0.f),
    C_Vertex3f(0.458498955f, 0.370203018f, 0.f),
    C_Vertex3f(0.458929062f, 0.454672813f, 0.f),
    C_Vertex3f(0.467282295f, 0.378973007f, 0.f),
    C_Vertex3f(0.476633072f, 0.462647438f, 0.f),
    C_Vertex3f(0.476686239f, 0.386646748f, 0.f),
    C_Vertex3f(0.496193886f, 0.468628407f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.433116913f, 0.326354027f, 0.f),
    C_Vertex3f(0.437999725f, 0.33863163f, 0.f),
    C_Vertex3f(0.442154884f, 0.444867611f, 0.f),
    C_Vertex3f(0.44385767f, 0.35003233f, 0.f),
    C_Vertex3f(0.458929062f, 0.454672813f, 0.f),
    C_Vertex3f(0.450690746f, 0.360556126f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.424319267f, 0.284259796f, 0.f),
    C_Vertex3f(0.426276684f, 0.29916811f, 0.f),
    C_Vertex3f(0.426310539f, 0.433231831f, 0.f),
    C_Vertex3f(0.429209232f, 0.31319952f, 0.f),
    C_Vertex3f(0.442154884f, 0.444867611f, 0.f),
    C_Vertex3f(0.433116913f, 0.326354027f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.426310539f, 0.433231831f, 0.f),
    C_Vertex3f(0.411396027f, 0.419765472f, 0.f),
    C_Vertex3f(0.423828125f, 0.276367188f, 0.f),
    C_Vertex3f(0.424319267f, 0.284259796f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.759765625f, 0.25223732f, 0.f),
    C_Vertex3f(0.757088184f, 0.279773235f, 0.f),
    C_Vertex3f(0.758888245f, 0.211914063f, 0.f),
    C_Vertex3f(0.75910759f, 0.216503143f, 0.f),
    C_Vertex3f(0.759436607f, 0.224425316f, 0.f),
    C_Vertex3f(0.759655952f, 0.231091499f, 0.f),
    C_Vertex3f(0.759765625f, 0.236501694f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.758888245f, 0.211914063f, 0.f),
    C_Vertex3f(0.757088184f, 0.279773235f, 0.f),
    C_Vertex3f(0.751733303f, 0.307528019f, 0.f),
    C_Vertex3f(0.743700981f, 0.335501671f, 0.f),
    C_Vertex3f(0.735668659f, 0.356646061f, 0.f),
    C_Vertex3f(0.729535341f, 0.370052576f, 0.f),
    C_Vertex3f(0.722623587f, 0.382769346f, 0.f),
    C_Vertex3f(0.714933395f, 0.394796371f, 0.f),
    C_Vertex3f(0.706464767f, 0.406133652f, 0.f),
    C_Vertex3f(0.697217703f, 0.416781187f, 0.f),
    C_Vertex3f(0.687192202f, 0.426738977f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33407402f, 0.28931427f, 0.f),
    C_Vertex3f(0.309171677f, 0.283472061f, 0.f),
    C_Vertex3f(0.31837821f, 0.223430395f, 0.f),
    C_Vertex3f(0.331490755f, 0.22727704f, 0.f),
    C_Vertex3f(0.342363358f, 0.23105526f, 0.f),
    C_Vertex3f(0.345214844f, 0.292480469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132621765f, 0.173650742f, 0.f),
    C_Vertex3f(0.139068604f, 0.178779602f, 0.f),
    C_Vertex3f(0.140188932f, 0.253192425f, 0.f),
    C_Vertex3f(0.143877029f, 0.182491302f, 0.f),
    C_Vertex3f(0.154176235f, 0.256877899f, 0.f),
    C_Vertex3f(0.157044411f, 0.188985825f, 0.f),
    C_Vertex3f(0.169603825f, 0.260372162f, 0.f),
    C_Vertex3f(0.173762321f, 0.194551468f, 0.f),
    C_Vertex3f(0.186471701f, 0.263675213f, 0.f),
    C_Vertex3f(0.194030762f, 0.199188232f, 0.f),
    C_Vertex3f(0.204779863f, 0.266787052f, 0.f),
    C_Vertex3f(0.219216347f, 0.203170776f, 0.f),
    C_Vertex3f(0.224528313f, 0.26970768f, 0.f),
    C_Vertex3f(0.243528605f, 0.207359552f, 0.f),
    C_Vertex3f(0.251504898f, 0.273258209f, 0.f),
    C_Vertex3f(0.26560092f, 0.211479902f, 0.f),
    C_Vertex3f(0.281648636f, 0.278120041f, 0.f),
    C_Vertex3f(0.285433292f, 0.215531826f, 0.f),
    C_Vertex3f(0.309171677f, 0.283472061f, 0.f),
    C_Vertex3f(0.303025723f, 0.219515324f, 0.f),
    C_Vertex3f(0.31837821f, 0.223430395f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.127641916f, 0.249315739f, 0.f),
    C_Vertex3f(0.116535187f, 0.245247841f, 0.f),
    C_Vertex3f(0.122951508f, 0.161810875f, 0.f),
    C_Vertex3f(0.132621765f, 0.173650742f, 0.f),
    C_Vertex3f(0.140188932f, 0.253192425f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.116535187f, 0.245247841f, 0.f),
    C_Vertex3f(0.116504669f, 0.148388863f, 0.f),
    C_Vertex3f(0.122951508f, 0.161810875f, 0.f),
    C_Vertex3f(0.0310893059f, 0.102271199f, 0.f),
    C_Vertex3f(0.0297851563f, 0.131942749f, 0.f),
    C_Vertex3f(0.0297851563f, 0.115615845f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.346679688f, 0.232910156f, -1.f),
    C_Vertex3f(0.346679688f, 0.22396183f, -1.f),
    C_Vertex3f(0.369329453f, 0.0651340485f, -1.f),
    C_Vertex3f(0.355556488f, 0.455490112f, -1.f),
    C_Vertex3f(0.345649719f, 0.317296028f, -1.f),
    C_Vertex3f(0.345649719f, 0.306911469f, -1.f),
    C_Vertex3f(0.345432281f, 0.29794693f, -1.f),
    C_Vertex3f(0.345214844f, 0.292480469f, -1.f),
    C_Vertex3f(0.342363358f, 0.23105526f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.341353893f, 0.352425814f, -1.f),
    C_Vertex3f(0.343501806f, 0.341747403f, -1.f),
    C_Vertex3f(0.343522072f, 0.460269928f, -1.f),
    C_Vertex3f(0.344933748f, 0.330037475f, -1.f),
    C_Vertex3f(0.355556488f, 0.455490112f, -1.f),
    C_Vertex3f(0.345649719f, 0.317296028f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.343522072f, 0.460269928f, -1.f),
    C_Vertex3f(0.330599785f, 0.464366913f, -1.f),
    C_Vertex3f(0.330614328f, 0.378271937f, -1.f),
    C_Vertex3f(0.334910154f, 0.370688081f, -1.f),
    C_Vertex3f(0.338490009f, 0.362072706f, -1.f),
    C_Vertex3f(0.341353893f, 0.352425814f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.305207014f, 0.39931643f, -1.f),
    C_Vertex3f(0.313076258f, 0.395175815f, -1.f),
    C_Vertex3f(0.316789627f, 0.467781067f, -1.f),
    C_Vertex3f(0.319874763f, 0.390345097f, -1.f),
    C_Vertex3f(0.330599785f, 0.464366913f, -1.f),
    C_Vertex3f(0.325602531f, 0.384824276f, -1.f),
    C_Vertex3f(0.330614328f, 0.378271937f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.275174856f, 0.407597661f, -1.f),
    C_Vertex3f(0.286256313f, 0.405527353f, -1.f),
    C_Vertex3f(0.286505699f, 0.472560883f, -1.f),
    C_Vertex3f(0.296267033f, 0.402766943f, -1.f),
    C_Vertex3f(0.302091599f, 0.47051239f, -1.f),
    C_Vertex3f(0.305207014f, 0.39931643f, -1.f),
    C_Vertex3f(0.316789627f, 0.467781067f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.233160019f, 0.409667969f, -1.f),
    C_Vertex3f(0.249799728f, 0.409667969f, -1.f),
    C_Vertex3f(0.252670288f, 0.474609375f, -1.f),
    C_Vertex3f(0.263022661f, 0.408977866f, -1.f),
    C_Vertex3f(0.270031929f, 0.473926544f, -1.f),
    C_Vertex3f(0.275174856f, 0.407597661f, -1.f),
    C_Vertex3f(0.286505699f, 0.472560883f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.185043335f, 0.402775764f, -1.f),
    C_Vertex3f(0.199608803f, 0.406221867f, -1.f),
    C_Vertex3f(0.202404499f, 0.47247982f, -1.f),
    C_Vertex3f(0.215647697f, 0.408519268f, -1.f),
    C_Vertex3f(0.229230881f, 0.474609375f, -1.f),
    C_Vertex3f(0.233160019f, 0.409667969f, -1.f),
    C_Vertex3f(0.252670288f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.177381039f, 0.468220711f, -1.f),
    C_Vertex3f(0.1541605f, 0.461832047f, -1.f),
    C_Vertex3f(0.160577297f, 0.392629147f, -1.f),
    C_Vertex3f(0.171951294f, 0.398180962f, -1.f),
    C_Vertex3f(0.185043335f, 0.402775764f, -1.f),
    C_Vertex3f(0.202404499f, 0.47247982f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.1541605f, 0.461832047f, -1.f),
    C_Vertex3f(0.132742882f, 0.453313828f, -1.f),
    C_Vertex3f(0.136763573f, 0.370231628f, -1.f),
    C_Vertex3f(0.142983437f, 0.37865448f, -1.f),
    C_Vertex3f(0.150921345f, 0.386120319f, -1.f),
    C_Vertex3f(0.160577297f, 0.392629147f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.132742882f, 0.453313828f, -1.f),
    C_Vertex3f(0.113811016f, 0.443185329f, -1.f),
    C_Vertex3f(0.119140625f, 0.320411682f, -1.f),
    C_Vertex3f(0.121213913f, 0.329708099f, -1.f),
    C_Vertex3f(0.125878811f, 0.345758438f, -1.f),
    C_Vertex3f(0.131062031f, 0.359266281f, -1.f),
    C_Vertex3f(0.136763573f, 0.370231628f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.119140625f, 0.320411682f, -1.f),
    C_Vertex3f(0.113811016f, 0.443185329f, -1.f),
    C_Vertex3f(0.0973649025f, 0.431446552f, -1.f),
    C_Vertex3f(0.083404541f, 0.418097496f, -1.f),
    C_Vertex3f(0.0719299316f, 0.403138161f, -1.f),
    C_Vertex3f(0.061794281f, 0.385529995f, -1.f),
    C_Vertex3f(0.0529975891f, 0.365272999f, -1.f),
    C_Vertex3f(0.045539856f, 0.342367172f, -1.f),
    C_Vertex3f(0.0424804688f, 0.329589844f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.7578125f, 0.142089844f, -1.f),
    C_Vertex3f(0.67578125f, 0.152580261f, -1.f),
    C_Vertex3f(0.676970482f, 0.024503231f, -1.f),
    C_Vertex3f(0.690079689f, 0.0344338417f, -1.f),
    C_Vertex3f(0.702233076f, 0.0454012156f, -1.f),
    C_Vertex3f(0.713430643f, 0.0574053526f, -1.f),
    C_Vertex3f(0.72367239f, 0.0704462528f, -1.f),
    C_Vertex3f(0.732958317f, 0.0845239162f, -1.f),
    C_Vertex3f(0.741288424f, 0.0996383429f, -1.f),
    C_Vertex3f(0.74866271f, 0.115789533f, -1.f),
    C_Vertex3f(0.755081177f, 0.132977486f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.67578125f, 0.152580261f, -1.f),
    C_Vertex3f(0.672834396f, 0.140884399f, -1.f),
    C_Vertex3f(0.676970482f, 0.024503231f, -1.f),
    C_Vertex3f(0.664511204f, 0.120115757f, -1.f),
    C_Vertex3f(0.66306901f, 0.0158139467f, -1.f),
    C_Vertex3f(0.653758526f, 0.101970196f, -1.f),
    C_Vertex3f(0.648375273f, 0.00836598873f, -1.f),
    C_Vertex3f(0.640576363f, 0.0864477158f, -1.f),
    C_Vertex3f(0.632889271f, 0.00215935707f, -1.f),
    C_Vertex3f(0.624964714f, 0.073548317f, -1.f),
    C_Vertex3f(0.616611004f, -0.00280594826f, -1.f),
    C_Vertex3f(0.607634068f, 0.0638737679f, -1.f),
    C_Vertex3f(0.599540472f, -0.00652992725f, -1.f),
    C_Vertex3f(0.588584423f, 0.0574240685f, -1.f),
    C_Vertex3f(0.581677675f, -0.00901257992f, -1.f),
    C_Vertex3f(0.567815781f, 0.0541992188f, -1.f),
    C_Vertex3f(0.563022614f, -0.0102539063f, -1.f),
    C_Vertex3f(0.546966553f, 0.0541992188f, -1.f),
    C_Vertex3f(0.539424896f, -0.0102539063f, -1.f),
    C_Vertex3f(0.528546333f, 0.0565748215f, -1.f),
    C_Vertex3f(0.513397217f, -0.00763177872f, -1.f),
    C_Vertex3f(0.51091671f, 0.0613260269f, -1.f),
    C_Vertex3f(0.494077682f, 0.0684528351f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.513397217f, -0.00763177872f, -1.f),
    C_Vertex3f(0.494077682f, 0.0684528351f, -1.f),
    C_Vertex3f(0.489089966f, -0.00238752365f, -1.f),
    C_Vertex3f(0.478029251f, 0.077955246f, -1.f),
    C_Vertex3f(0.466503143f, 0.00547885895f, -1.f),
    C_Vertex3f(0.464164257f, 0.0886316299f, -1.f),
    C_Vertex3f(0.4524827f, 0.100481987f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.445636749f, 0.0159673691f, -1.f),
    C_Vertex3f(0.466503143f, 0.00547885895f, -1.f),
    C_Vertex3f(0.4524827f, 0.100481987f, -1.f),
    C_Vertex3f(0.442984581f, 0.113506317f, -1.f),
    C_Vertex3f(0.437498569f, 0.124155045f, -1.f),
    C_Vertex3f(0.434271216f, 0.132332802f, -1.f),
    C_Vertex3f(0.431473851f, 0.141589165f, -1.f),
    C_Vertex3f(0.429106474f, 0.151924133f, -1.f),
    C_Vertex3f(0.427169085f, 0.163337708f, -1.f),
    C_Vertex3f(0.42578125f, 0.0294060707f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.42578125f, 0.0294060707f, -1.f),
    C_Vertex3f(0.427169085f, 0.163337708f, -1.f),
    C_Vertex3f(0.425661683f, 0.175829887f, -1.f),
    C_Vertex3f(0.42458427f, 0.189400673f, -1.f),
    C_Vertex3f(0.423936844f, 0.204050064f, -1.f),
    C_Vertex3f(0.423828125f, 0.211914063f, -1.f),
    C_Vertex3f(0.411396027f, 0.419765472f, -1.f),
    C_Vertex3f(0.406936646f, 0.0457949638f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.406936646f, 0.0457949638f, -1.f),
    C_Vertex3f(0.411396027f, 0.419765472f, -1.f),
    C_Vertex3f(0.404403687f, 0.412117004f, -1.f),
    C_Vertex3f(0.400032043f, 0.419109344f, -1.f),
    C_Vertex3f(0.39025116f, 0.431181908f, -1.f),
    C_Vertex3f(0.389102936f, 0.0651340485f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389102936f, 0.0651340485f, -1.f),
    C_Vertex3f(0.39025116f, 0.431181908f, -1.f),
    C_Vertex3f(0.380691528f, 0.0762786865f, -1.f),
    C_Vertex3f(0.379432678f, 0.441342354f, -1.f),
    C_Vertex3f(0.369329453f, 0.0651340485f, -1.f),
    C_Vertex3f(0.367576599f, 0.449590683f, -1.f),
    C_Vertex3f(0.355556488f, 0.455490112f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.346474886f, 0.207592964f, -1.f),
    C_Vertex3f(0.346065283f, 0.192751884f, -1.f),
    C_Vertex3f(0.346468925f, 0.0457949638f, -1.f),
    C_Vertex3f(0.369329453f, 0.0651340485f, -1.f),
    C_Vertex3f(0.346679688f, 0.22396183f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.346468925f, 0.0457949638f, -1.f),
    C_Vertex3f(0.346065283f, 0.192751884f, -1.f),
    C_Vertex3f(0.345450878f, 0.179438591f, -1.f),
    C_Vertex3f(0.344631672f, 0.167653084f, -1.f),
    C_Vertex3f(0.343607664f, 0.157395363f, -1.f),
    C_Vertex3f(0.342378855f, 0.148665428f, -1.f),
    C_Vertex3f(0.340945244f, 0.14146328f, -1.f),
    C_Vertex3f(0.340126038f, 0.138626099f, -1.f),
    C_Vertex3f(0.337284088f, 0.129348755f, -1.f),
    C_Vertex3f(0.328350067f, 0.112322807f, -1.f),
    C_Vertex3f(0.323472023f, 0.0294060707f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.328350067f, 0.112322807f, -1.f),
    C_Vertex3f(0.316165924f, 0.0968255997f, -1.f),
    C_Vertex3f(0.323472023f, 0.0294060707f, -1.f),
    C_Vertex3f(0.300731659f, 0.082857132f, -1.f),
    C_Vertex3f(0.300338745f, 0.0159673691f, -1.f),
    C_Vertex3f(0.282047272f, 0.0704174042f, -1.f),
    C_Vertex3f(0.277069092f, 0.00547885895f, -1.f),
    C_Vertex3f(0.261751175f, 0.0610876083f, -1.f),
    C_Vertex3f(0.252898216f, -0.00238752365f, -1.f),
    C_Vertex3f(0.239843369f, 0.0548677444f, -1.f),
    C_Vertex3f(0.227826118f, -0.00763177872f, -1.f),
    C_Vertex3f(0.216323853f, 0.0517578125f, -1.f),
    C_Vertex3f(0.201852798f, -0.0102539063f, -1.f),
    C_Vertex3f(0.192722321f, 0.0517578125f, -1.f),
    C_Vertex3f(0.179019928f, -0.0102539063f, -1.f),
    C_Vertex3f(0.17327261f, 0.0544042587f, -1.f),
    C_Vertex3f(0.161273479f, -0.0090457201f, -1.f),
    C_Vertex3f(0.156445026f, 0.0596971512f, -1.f),
    C_Vertex3f(0.144571781f, -0.0066293478f, -1.f),
    C_Vertex3f(0.142239571f, 0.0676364899f, -1.f),
    C_Vertex3f(0.130656242f, 0.0782222748f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.128914833f, -0.00300478935f, -1.f),
    C_Vertex3f(0.144571781f, -0.0066293478f, -1.f),
    C_Vertex3f(0.130656242f, 0.0782222748f, -1.f),
    C_Vertex3f(0.121968746f, 0.0900354385f, -1.f),
    C_Vertex3f(0.116177082f, 0.103075981f, -1.f),
    C_Vertex3f(0.114302635f, 0.00182795525f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116504669f, 0.148388863f, -1.f),
    C_Vertex3f(0.116535187f, 0.245247841f, -1.f),
    C_Vertex3f(0.11328125f, 0.133384705f, -1.f),
    C_Vertex3f(0.102035522f, 0.238859177f, -1.f),
    C_Vertex3f(0.11328125f, 0.117343903f, -1.f),
    C_Vertex3f(0.114302635f, 0.00182795525f, -1.f),
    C_Vertex3f(0.116177082f, 0.103075981f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0571722984f, 0.0434318781f, -1.f),
    C_Vertex3f(0.0663013458f, 0.0332407951f, -1.f),
    C_Vertex3f(0.0697598457f, 0.215843201f, -1.f),
    C_Vertex3f(0.0767345428f, 0.0235753059f, -1.f),
    C_Vertex3f(0.0848326683f, 0.228347778f, -1.f),
    C_Vertex3f(0.0882124901f, 0.0151180029f, -1.f),
    C_Vertex3f(0.102035522f, 0.238859177f, -1.f),
    C_Vertex3f(0.100735188f, 0.00786888599f, -1.f),
    C_Vertex3f(0.114302635f, 0.00182795525f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0336976051f, 0.0894521475f, -1.f),
    C_Vertex3f(0.037610054f, 0.0771586895f, -1.f),
    C_Vertex3f(0.0378947258f, 0.167790413f, -1.f),
    C_Vertex3f(0.0428266525f, 0.0653908253f, -1.f),
    C_Vertex3f(0.0460042953f, 0.184854507f, -1.f),
    C_Vertex3f(0.0493474007f, 0.0541485548f, -1.f),
    C_Vertex3f(0.0568170547f, 0.201345444f, -1.f),
    C_Vertex3f(0.0571722984f, 0.0434318781f, -1.f),
    C_Vertex3f(0.0697598457f, 0.215843201f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0324883461f, 0.15015316f, -1.f),
    C_Vertex3f(0.0297851563f, 0.131942749f, -1.f),
    C_Vertex3f(0.0310893059f, 0.102271199f, -1.f),
    C_Vertex3f(0.0336976051f, 0.0894521475f, -1.f),
    C_Vertex3f(0.0378947258f, 0.167790413f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.672354817f, 0.316174388f, -1.f),
    C_Vertex3f(0.675573707f, 0.301183343f, -1.f),
    C_Vertex3f(0.676388264f, 0.436007023f, -1.f),
    C_Vertex3f(0.677947044f, 0.285027504f, -1.f),
    C_Vertex3f(0.687192202f, 0.426738977f, -1.f),
    C_Vertex3f(0.678710938f, 0.276367188f, -1.f),
    C_Vertex3f(0.758888245f, 0.211914063f, -1.f),
    C_Vertex3f(0.423828125f, 0.276367188f, -1.f),
    C_Vertex3f(0.423828125f, 0.211914063f, -1.f),
    C_Vertex3f(0.411396027f, 0.419765472f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.657624841f, 0.354158759f, -1.f),
    C_Vertex3f(0.663380384f, 0.342662096f, -1.f),
    C_Vertex3f(0.664805889f, 0.444585323f, -1.f),
    C_Vertex3f(0.668290377f, 0.330000639f, -1.f),
    C_Vertex3f(0.676388264f, 0.436007023f, -1.f),
    C_Vertex3f(0.672354817f, 0.316174388f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.65245831f, 0.452091336f, -1.f),
    C_Vertex3f(0.639345527f, 0.458525062f, -1.f),
    C_Vertex3f(0.643577099f, 0.373657703f, -1.f),
    C_Vertex3f(0.651023746f, 0.364490628f, -1.f),
    C_Vertex3f(0.657624841f, 0.354158759f, -1.f),
    C_Vertex3f(0.664805889f, 0.444585323f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.595842123f, 0.403666258f, -1.f),
    C_Vertex3f(0.606643915f, 0.399665117f, -1.f),
    C_Vertex3f(0.610824347f, 0.46817565f, -1.f),
    C_Vertex3f(0.616818309f, 0.394663692f, -1.f),
    C_Vertex3f(0.625467539f, 0.463886499f, -1.f),
    C_Vertex3f(0.626365304f, 0.388661981f, -1.f),
    C_Vertex3f(0.639345527f, 0.458525062f, -1.f),
    C_Vertex3f(0.635284901f, 0.381659985f, -1.f),
    C_Vertex3f(0.643577099f, 0.373657703f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.546142578f, 0.409667969f, -1.f),
    C_Vertex3f(0.559672356f, 0.409667969f, -1.f),
    C_Vertex3f(0.562303543f, 0.474609375f, -1.f),
    C_Vertex3f(0.572356343f, 0.408667684f, -1.f),
    C_Vertex3f(0.579242349f, 0.473537087f, -1.f),
    C_Vertex3f(0.584412932f, 0.406667113f, -1.f),
    C_Vertex3f(0.59541595f, 0.471392512f, -1.f),
    C_Vertex3f(0.595842123f, 0.403666258f, -1.f),
    C_Vertex3f(0.610824347f, 0.46817565f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.540885925f, 0.474609375f, -1.f),
    C_Vertex3f(0.517611504f, 0.472615719f, -1.f),
    C_Vertex3f(0.520508051f, 0.406379223f, -1.f),
    C_Vertex3f(0.533015013f, 0.40857172f, -1.f),
    C_Vertex3f(0.546142578f, 0.409667969f, -1.f),
    C_Vertex3f(0.562303543f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.476686239f, 0.386646748f, -1.f),
    C_Vertex3f(0.486710787f, 0.393224239f, -1.f),
    C_Vertex3f(0.496193886f, 0.468628407f, -1.f),
    C_Vertex3f(0.497355938f, 0.398705482f, -1.f),
    C_Vertex3f(0.517611504f, 0.472615719f, -1.f),
    C_Vertex3f(0.508621693f, 0.403090477f, -1.f),
    C_Vertex3f(0.520508051f, 0.406379223f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.450690746f, 0.360556126f, -1.f),
    C_Vertex3f(0.458498955f, 0.370203018f, -1.f),
    C_Vertex3f(0.458929062f, 0.454672813f, -1.f),
    C_Vertex3f(0.467282295f, 0.378973007f, -1.f),
    C_Vertex3f(0.476633072f, 0.462647438f, -1.f),
    C_Vertex3f(0.476686239f, 0.386646748f, -1.f),
    C_Vertex3f(0.496193886f, 0.468628407f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.433116913f, 0.326354027f, -1.f),
    C_Vertex3f(0.437999725f, 0.33863163f, -1.f),
    C_Vertex3f(0.442154884f, 0.444867611f, -1.f),
    C_Vertex3f(0.44385767f, 0.35003233f, -1.f),
    C_Vertex3f(0.458929062f, 0.454672813f, -1.f),
    C_Vertex3f(0.450690746f, 0.360556126f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.424319267f, 0.284259796f, -1.f),
    C_Vertex3f(0.426276684f, 0.29916811f, -1.f),
    C_Vertex3f(0.426310539f, 0.433231831f, -1.f),
    C_Vertex3f(0.429209232f, 0.31319952f, -1.f),
    C_Vertex3f(0.442154884f, 0.444867611f, -1.f),
    C_Vertex3f(0.433116913f, 0.326354027f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.426310539f, 0.433231831f, -1.f),
    C_Vertex3f(0.411396027f, 0.419765472f, -1.f),
    C_Vertex3f(0.423828125f, 0.276367188f, -1.f),
    C_Vertex3f(0.424319267f, 0.284259796f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.759765625f, 0.25223732f, -1.f),
    C_Vertex3f(0.757088184f, 0.279773235f, -1.f),
    C_Vertex3f(0.758888245f, 0.211914063f, -1.f),
    C_Vertex3f(0.75910759f, 0.216503143f, -1.f),
    C_Vertex3f(0.759436607f, 0.224425316f, -1.f),
    C_Vertex3f(0.759655952f, 0.231091499f, -1.f),
    C_Vertex3f(0.759765625f, 0.236501694f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.758888245f, 0.211914063f, -1.f),
    C_Vertex3f(0.757088184f, 0.279773235f, -1.f),
    C_Vertex3f(0.751733303f, 0.307528019f, -1.f),
    C_Vertex3f(0.743700981f, 0.335501671f, -1.f),
    C_Vertex3f(0.735668659f, 0.356646061f, -1.f),
    C_Vertex3f(0.729535341f, 0.370052576f, -1.f),
    C_Vertex3f(0.722623587f, 0.382769346f, -1.f),
    C_Vertex3f(0.714933395f, 0.394796371f, -1.f),
    C_Vertex3f(0.706464767f, 0.406133652f, -1.f),
    C_Vertex3f(0.697217703f, 0.416781187f, -1.f),
    C_Vertex3f(0.687192202f, 0.426738977f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.33407402f, 0.28931427f, -1.f),
    C_Vertex3f(0.309171677f, 0.283472061f, -1.f),
    C_Vertex3f(0.31837821f, 0.223430395f, -1.f),
    C_Vertex3f(0.331490755f, 0.22727704f, -1.f),
    C_Vertex3f(0.342363358f, 0.23105526f, -1.f),
    C_Vertex3f(0.345214844f, 0.292480469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.132621765f, 0.173650742f, -1.f),
    C_Vertex3f(0.139068604f, 0.178779602f, -1.f),
    C_Vertex3f(0.140188932f, 0.253192425f, -1.f),
    C_Vertex3f(0.143877029f, 0.182491302f, -1.f),
    C_Vertex3f(0.154176235f, 0.256877899f, -1.f),
    C_Vertex3f(0.157044411f, 0.188985825f, -1.f),
    C_Vertex3f(0.169603825f, 0.260372162f, -1.f),
    C_Vertex3f(0.173762321f, 0.194551468f, -1.f),
    C_Vertex3f(0.186471701f, 0.263675213f, -1.f),
    C_Vertex3f(0.194030762f, 0.199188232f, -1.f),
    C_Vertex3f(0.204779863f, 0.266787052f, -1.f),
    C_Vertex3f(0.219216347f, 0.203170776f, -1.f),
    C_Vertex3f(0.224528313f, 0.26970768f, -1.f),
    C_Vertex3f(0.243528605f, 0.207359552f, -1.f),
    C_Vertex3f(0.251504898f, 0.273258209f, -1.f),
    C_Vertex3f(0.26560092f, 0.211479902f, -1.f),
    C_Vertex3f(0.281648636f, 0.278120041f, -1.f),
    C_Vertex3f(0.285433292f, 0.215531826f, -1.f),
    C_Vertex3f(0.309171677f, 0.283472061f, -1.f),
    C_Vertex3f(0.303025723f, 0.219515324f, -1.f),
    C_Vertex3f(0.31837821f, 0.223430395f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.127641916f, 0.249315739f, -1.f),
    C_Vertex3f(0.116535187f, 0.245247841f, -1.f),
    C_Vertex3f(0.122951508f, 0.161810875f, -1.f),
    C_Vertex3f(0.132621765f, 0.173650742f, -1.f),
    C_Vertex3f(0.140188932f, 0.253192425f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.116535187f, 0.245247841f, -1.f),
    C_Vertex3f(0.116504669f, 0.148388863f, -1.f),
    C_Vertex3f(0.122951508f, 0.161810875f, -1.f),
    C_Vertex3f(0.0310893059f, 0.102271199f, -1.f),
    C_Vertex3f(0.0297851563f, 0.131942749f, -1.f),
    C_Vertex3f(0.0297851563f, 0.115615845f, -1.f),
    C_End(),
    C_Translatef(0.795898438f, 0.f, 0.f),
};
static const Call char_231[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.131752744f, 0.991282582f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, -1.f),
    C_Vertex3f(0.439453125f, 0.159667969f, -1.f),
    C_Vertex3f(0.439453125f, 0.159667969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.987604678f, -0.156962141f, 0.f),
    C_Vertex3f(0.439453125f, 0.159667969f, 0.f),
    C_Vertex3f(0.439453125f, 0.159667969f, -1.f),
    C_Normal3f(0.97908473f, -0.203453019f, 0.f),
    C_Vertex3f(0.437869072f, 0.149701118f, 0.f),
    C_Vertex3f(0.437869072f, 0.149701118f, -1.f),
    C_Normal3f(0.964377761f, -0.26452899f, 0.f),
    C_Vertex3f(0.433464766f, 0.130849957f, 0.f),
    C_Vertex3f(0.433464766f, 0.130849957f, -1.f),
    C_Normal3f(0.939896464f, -0.341459483f, 0.f),
    C_Vertex3f(0.427824259f, 0.113081336f, 0.f),
    C_Vertex3f(0.427824259f, 0.113081336f, -1.f),
    C_Normal3f(0.907030344f, -0.421065122f, 0.f),
    C_Vertex3f(0.420947552f, 0.0963952541f, 0.f),
    C_Vertex3f(0.420947552f, 0.0963952541f, -1.f),
    C_Normal3f(0.865158498f, -0.501498401f, 0.f),
    C_Vertex3f(0.412834644f, 0.0807917118f, 0.f),
    C_Vertex3f(0.412834644f, 0.0807917118f, -1.f),
    C_Normal3f(0.814237773f, -0.580531597f, 0.f),
    C_Vertex3f(0.403485537f, 0.066270709f, 0.f),
    C_Vertex3f(0.403485537f, 0.066270709f, -1.f),
    C_Normal3f(0.754939914f, -0.655793846f, 0.f),
    C_Vertex3f(0.392900229f, 0.0528322458f, 0.f),
    C_Vertex3f(0.392900229f, 0.0528322458f, -1.f),
    C_Normal3f(0.688658357f, -0.725086033f, 0.f),
    C_Vertex3f(0.38107872f, 0.0404763222f, 0.f),
    C_Vertex3f(0.38107872f, 0.0404763222f, -1.f),
    C_Normal3f(0.618423939f, -0.785844684f, 0.f),
    C_Vertex3f(0.368021011f, 0.0292029381f, 0.f),
    C_Vertex3f(0.368021011f, 0.0292029381f, -1.f),
    C_Normal3f(0.542743206f, -0.839898705f, 0.f),
    C_Vertex3f(0.354218721f, 0.019338727f, 0.f),
    C_Vertex3f(0.354218721f, 0.019338727f, -1.f),
    C_Normal3f(0.461000979f, -0.887399673f, 0.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, 0.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, -1.f),
    C_Normal3f(0.375253379f, -0.926922321f, 0.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, 0.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, -1.f),
    C_Normal3f(0.287862331f, -0.957671821f, 0.f),
    C_Vertex3f(0.308344364f, -0.00179886818f, 0.f),
    C_Vertex3f(0.308344364f, -0.00179886818f, -1.f),
    C_Normal3f(0.201184601f, -0.979553342f, 0.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, 0.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, -1.f),
    C_Normal3f(0.117288537f, -0.993097842f, 0.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, 0.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, -1.f),
    C_Normal3f(0.0354844518f, -0.999370158f, 0.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, 0.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, -1.f),
    C_Normal3f(-0.044093933f, -0.999027431f, 0.f),
    C_Vertex3f(0.234351158f, -0.0102539063f, 0.f),
    C_Vertex3f(0.234351158f, -0.0102539063f, -1.f),
    C_Normal3f(-0.131067663f, -0.99137342f, 0.f),
    C_Vertex3f(0.211698174f, -0.00830876827f, 0.f),
    C_Vertex3f(0.211698174f, -0.00830876827f, -1.f),
    C_Normal3f(-0.226606935f, -0.973986268f, 0.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, 0.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, -1.f),
    C_Normal3f(-0.326136827f, -0.945322514f, 0.f),
    C_Vertex3f(0.169895887f, 0.00141692162f, 0.f),
    C_Vertex3f(0.169895887f, 0.00141692162f, -1.f),
    C_Normal3f(-0.426454365f, -0.904509068f, 0.f),
    C_Vertex3f(0.150746584f, 0.00919747353f, 0.f),
    C_Vertex3f(0.150746584f, 0.00919747353f, -1.f),
    C_Normal3f(-0.523817837f, -0.851830363f, 0.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, 0.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, -1.f),
    C_Normal3f(-0.61455822f, -0.788871467f, 0.f),
    C_Vertex3f(0.115951657f, 0.0305939913f, 0.f),
    C_Vertex3f(0.115951657f, 0.0305939913f, -1.f),
    C_Normal3f(-0.69573462f, -0.718298972f, 0.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, 0.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, -1.f),
    C_Normal3f(-0.770461857f, -0.637486041f, 0.f),
    C_Vertex3f(0.0858283043f, 0.0597710609f, 0.f),
    C_Vertex3f(0.0858283043f, 0.0597710609f, -1.f),
    C_Normal3f(-0.838346004f, -0.545138538f, 0.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, 0.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, -1.f),
    C_Normal3f(-0.893242121f, -0.449575752f, 0.f),
    C_Vertex3f(0.0623019934f, 0.0959511995f, 0.f),
    C_Vertex3f(0.0623019934f, 0.0959511995f, -1.f),
    C_Normal3f(-0.934954882f, -0.354766816f, 0.f),
    C_Vertex3f(0.0532534122f, 0.116570234f, 0.f),
    C_Vertex3f(0.0532534122f, 0.116570234f, -1.f),
    C_Normal3f(-0.96454227f, -0.263928503f, 0.f),
    C_Vertex3f(0.0460145473f, 0.138875246f, 0.f),
    C_Vertex3f(0.0460145473f, 0.138875246f, -1.f),
    C_Normal3f(-0.983806968f, -0.179231405f, 0.f),
    C_Vertex3f(0.0405853987f, 0.162866235f, 0.f),
    C_Vertex3f(0.0405853987f, 0.162866235f, -1.f),
    C_Normal3f(-0.994802058f, -0.101827636f, 0.f),
    C_Vertex3f(0.0369659662f, 0.1885432f, 0.f),
    C_Vertex3f(0.0369659662f, 0.1885432f, -1.f),
    C_Normal3f(-0.999555469f, -0.0298151094f, 0.f),
    C_Vertex3f(0.03515625f, 0.215906143f, 0.f),
    C_Vertex3f(0.03515625f, 0.215906143f, -1.f),
    C_Normal3f(-0.998968303f, 0.0454123914f, 0.f),
    C_Vertex3f(0.03515625f, 0.249214172f, 0.f),
    C_Vertex3f(0.03515625f, 0.249214172f, -1.f),
    C_Normal3f(-0.990711629f, 0.135979757f, 0.f),
    C_Vertex3f(0.0382714272f, 0.284432888f, 0.f),
    C_Vertex3f(0.0382714272f, 0.284432888f, -1.f),
    C_Normal3f(-0.97111553f, 0.238609701f, 0.f),
    C_Vertex3f(0.0445017815f, 0.317303181f, 0.f),
    C_Vertex3f(0.0445017815f, 0.317303181f, -1.f),
    C_Normal3f(-0.940327108f, 0.34027198f, 0.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, 0.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, -1.f),
    C_Normal3f(-0.902472794f, 0.430746824f, 0.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, 0.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, -1.f),
    C_Normal3f(-0.863703251f, 0.504000664f, 0.f),
    C_Vertex3f(0.0702359676f, 0.382161498f, 0.f),
    C_Vertex3f(0.0702359676f, 0.382161498f, -1.f),
    C_Normal3f(-0.821103454f, 0.570779502f, 0.f),
    C_Vertex3f(0.0780918598f, 0.3944875f, 0.f),
    C_Vertex3f(0.0780918598f, 0.3944875f, -1.f),
    C_Normal3f(-0.771526754f, 0.636196911f, 0.f),
    C_Vertex3f(0.0867605209f, 0.405933142f, 0.f),
    C_Vertex3f(0.0867605209f, 0.405933142f, -1.f),
    C_Normal3f(-0.715440392f, 0.698673785f, 0.f),
    C_Vertex3f(0.096241951f, 0.416498423f, 0.f),
    C_Vertex3f(0.096241951f, 0.416498423f, -1.f),
    C_Normal3f(-0.653754652f, 0.756706655f, 0.f),
    C_Vertex3f(0.10653615f, 0.426183343f, 0.f),
    C_Vertex3f(0.10653615f, 0.426183343f, -1.f),
    C_Normal3f(-0.587759495f, 0.809035778f, 0.f),
    C_Vertex3f(0.117643118f, 0.434987903f, 0.f),
    C_Vertex3f(0.117643118f, 0.434987903f, -1.f),
    C_Normal3f(-0.512026966f, 0.85896939f, 0.f),
    C_Vertex3f(0.129562855f, 0.442912102f, 0.f),
    C_Vertex3f(0.129562855f, 0.442912102f, -1.f),
    C_Normal3f(-0.419815004f, 0.907609761f, 0.f),
    C_Vertex3f(0.148661613f, 0.453477859f, 0.f),
    C_Vertex3f(0.148661613f, 0.453477859f, -1.f),
    C_Normal3f(-0.308520645f, 0.951217651f, 0.f),
    C_Vertex3f(0.175247669f, 0.464043617f, 0.f),
    C_Vertex3f(0.175247669f, 0.464043617f, -1.f),
    C_Normal3f(-0.183707342f, 0.982981026f, 0.f),
    C_Vertex3f(0.202954769f, 0.471087456f, 0.f),
    C_Vertex3f(0.202954769f, 0.471087456f, -1.f),
    C_Normal3f(-0.0659859255f, 0.997820497f, 0.f),
    C_Vertex3f(0.231782913f, 0.474609375f, 0.f),
    C_Vertex3f(0.231782913f, 0.474609375f, -1.f),
    C_Normal3f(0.0281452518f, 0.999603868f, 0.f),
    C_Vertex3f(0.256212234f, 0.474609375f, 0.f),
    C_Vertex3f(0.256212234f, 0.474609375f, -1.f),
    C_Normal3f(0.10130471f, 0.994855404f, 0.f),
    C_Vertex3f(0.274260879f, 0.473413348f, 0.f),
    C_Vertex3f(0.274260879f, 0.473413348f, -1.f),
    C_Normal3f(0.175657928f, 0.984451234f, 0.f),
    C_Vertex3f(0.291448712f, 0.471021295f, 0.f),
    C_Vertex3f(0.291448712f, 0.471021295f, -1.f),
    C_Normal3f(0.25465107f, 0.967033029f, 0.f),
    C_Vertex3f(0.307775736f, 0.467433214f, 0.f),
    C_Vertex3f(0.307775736f, 0.467433214f, -1.f),
    C_Normal3f(0.337013245f, 0.941499949f, 0.f),
    C_Vertex3f(0.323241949f, 0.462649107f, 0.f),
    C_Vertex3f(0.323241949f, 0.462649107f, -1.f),
    C_Normal3f(0.420948029f, 0.907084763f, 0.f),
    C_Vertex3f(0.337847352f, 0.456668973f, 0.f),
    C_Vertex3f(0.337847352f, 0.456668973f, -1.f),
    C_Normal3f(0.504237056f, 0.863565207f, 0.f),
    C_Vertex3f(0.351591945f, 0.449492812f, 0.f),
    C_Vertex3f(0.351591945f, 0.449492812f, -1.f),
    C_Normal3f(0.584466457f, 0.811417818f, 0.f),
    C_Vertex3f(0.364475727f, 0.441120625f, 0.f),
    C_Vertex3f(0.364475727f, 0.441120625f, -1.f),
    C_Normal3f(0.659348786f, 0.751837194f, 0.f),
    C_Vertex3f(0.376498699f, 0.43155241f, 0.f),
    C_Vertex3f(0.376498699f, 0.43155241f, -1.f),
    C_Normal3f(0.728860795f, 0.684661984f, 0.f),
    C_Vertex3f(0.387442231f, 0.420979381f, 0.f),
    C_Vertex3f(0.387442231f, 0.420979381f, -1.f),
    C_Normal3f(0.791613638f, 0.611021996f, 0.f),
    C_Vertex3f(0.397306323f, 0.409401536f, 0.f),
    C_Vertex3f(0.397306323f, 0.409401536f, -1.f),
    C_Normal3f(0.846052468f, 0.533099651f, 0.f),
    C_Vertex3f(0.406090975f, 0.396818876f, 0.f),
    C_Vertex3f(0.406090975f, 0.396818876f, -1.f),
    C_Normal3f(0.891355872f, 0.453304201f, 0.f),
    C_Vertex3f(0.413796186f, 0.383231401f, 0.f),
    C_Vertex3f(0.413796186f, 0.383231401f, -1.f),
    C_Normal3f(0.927452505f, 0.373940825f, 0.f),
    C_Vertex3f(0.420421958f, 0.368639112f, 0.f),
    C_Vertex3f(0.420421958f, 0.368639112f, -1.f),
    C_Normal3f(0.954892576f, 0.296951592f, 0.f),
    C_Vertex3f(0.425968289f, 0.353042006f, 0.f),
    C_Vertex3f(0.425968289f, 0.353042006f, -1.f),
    C_Normal3f(0.971834719f, 0.235663593f, 0.f),
    C_Vertex3f(0.430435181f, 0.336440086f, 0.f),
    C_Vertex3f(0.430435181f, 0.336440086f, -1.f),
    C_Normal3f(0.981990516f, 0.188930243f, 0.f),
    C_Vertex3f(0.432128906f, 0.327636719f, 0.f),
    C_Vertex3f(0.432128906f, 0.327636719f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.151110858f, -0.988516808f, 0.f),
    C_Vertex3f(0.432128906f, 0.327636719f, 0.f),
    C_Vertex3f(0.432128906f, 0.327636719f, -1.f),
    C_Vertex3f(0.35546875f, 0.315917969f, -1.f),
    C_Vertex3f(0.35546875f, 0.315917969f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.973956823f, -0.226733729f, 0.f),
    C_Vertex3f(0.35546875f, 0.315917969f, 0.f),
    C_Vertex3f(0.35546875f, 0.315917969f, -1.f),
    C_Normal3f(-0.952746391f, -0.303766668f, 0.f),
    C_Vertex3f(0.352739334f, 0.327642441f, 0.f),
    C_Vertex3f(0.352739334f, 0.327642441f, -1.f),
    C_Normal3f(-0.911611378f, -0.41105324f, 0.f),
    C_Vertex3f(0.345180035f, 0.348187923f, 0.f),
    C_Vertex3f(0.345180035f, 0.348187923f, -1.f),
    C_Normal3f(-0.834030807f, -0.551717758f, 0.f),
    C_Vertex3f(0.335520267f, 0.365829945f, 0.f),
    C_Vertex3f(0.335520267f, 0.365829945f, -1.f),
    C_Normal3f(-0.719893813f, -0.694084167f, 0.f),
    C_Vertex3f(0.323760033f, 0.380568504f, 0.f),
    C_Vertex3f(0.323760033f, 0.380568504f, -1.f),
    C_Normal3f(-0.578953743f, -0.815360427f, 0.f),
    C_Vertex3f(0.30989933f, 0.392403603f, 0.f),
    C_Vertex3f(0.30989933f, 0.392403603f, -1.f),
    C_Normal3f(-0.418954581f, -0.908007264f, 0.f),
    C_Vertex3f(0.294591427f, 0.401279926f, 0.f),
    C_Vertex3f(0.294591427f, 0.401279926f, -1.f),
    C_Normal3f(-0.246108323f, -0.969242394f, 0.f),
    C_Vertex3f(0.277836323f, 0.407197475f, 0.f),
    C_Vertex3f(0.277836323f, 0.407197475f, -1.f),
    C_Normal3f(-0.0831752717f, -0.996534884f, 0.f),
    C_Vertex3f(0.259634018f, 0.41015625f, 0.f),
    C_Vertex3f(0.259634018f, 0.41015625f, -1.f),
    C_Normal3f(0.0424002297f, -0.999100745f, 0.f),
    C_Vertex3f(0.242386818f, 0.41015625f, 0.f),
    C_Vertex3f(0.242386818f, 0.41015625f, -1.f),
    C_Normal3f(0.143321022f, -0.989676297f, 0.f),
    C_Vertex3f(0.228237867f, 0.408823848f, 0.f),
    C_Vertex3f(0.228237867f, 0.408823848f, -1.f),
    C_Normal3f(0.246344402f, -0.969182372f, 0.f),
    C_Vertex3f(0.214784861f, 0.406159043f, 0.f),
    C_Vertex3f(0.214784861f, 0.406159043f, -1.f),
    C_Normal3f(0.351784617f, -0.936080992f, 0.f),
    C_Vertex3f(0.202027798f, 0.402161837f, 0.f),
    C_Vertex3f(0.202027798f, 0.402161837f, -1.f),
    C_Normal3f(0.455658734f, -0.8901546f, 0.f),
    C_Vertex3f(0.189966679f, 0.396832228f, 0.f),
    C_Vertex3f(0.189966679f, 0.396832228f, -1.f),
    C_Normal3f(0.553831875f, -0.832628489f, 0.f),
    C_Vertex3f(0.178601503f, 0.390170217f, 0.f),
    C_Vertex3f(0.178601503f, 0.390170217f, -1.f),
    C_Normal3f(0.642789781f, -0.76604265f, 0.f),
    C_Vertex3f(0.167932272f, 0.382175803f, 0.f),
    C_Vertex3f(0.167932272f, 0.382175803f, -1.f),
    C_Normal3f(0.720233381f, -0.693731844f, 0.f),
    C_Vertex3f(0.157958984f, 0.372848988f, 0.f),
    C_Vertex3f(0.157958984f, 0.372848988f, -1.f),
    C_Normal3f(0.79592365f, -0.605397046f, 0.f),
    C_Vertex3f(0.148681641f, 0.36218977f, 0.f),
    C_Vertex3f(0.148681641f, 0.36218977f, -1.f),
    C_Normal3f(0.864854872f, -0.502021909f, 0.f),
    C_Vertex3f(0.140563965f, 0.34997952f, 0.f),
    C_Vertex3f(0.140563965f, 0.34997952f, -1.f),
    C_Normal3f(0.915732861f, -0.40178749f, 0.f),
    C_Vertex3f(0.133605957f, 0.336218238f, 0.f),
    C_Vertex3f(0.133605957f, 0.336218238f, -1.f),
    C_Normal3f(0.951208353f, -0.308549166f, 0.f),
    C_Vertex3f(0.127807617f, 0.320905924f, 0.f),
    C_Vertex3f(0.127807617f, 0.320905924f, -1.f),
    C_Normal3f(0.974532247f, -0.224247366f, 0.f),
    C_Vertex3f(0.123168945f, 0.304042578f, 0.f),
    C_Vertex3f(0.123168945f, 0.304042578f, -1.f),
    C_Normal3f(0.988779366f, -0.14938277f, 0.f),
    C_Vertex3f(0.119689941f, 0.2856282f, 0.f),
    C_Vertex3f(0.119689941f, 0.2856282f, -1.f),
    C_Normal3f(0.996501565f, -0.0835746899f, 0.f),
    C_Vertex3f(0.117370605f, 0.265662789f, 0.f),
    C_Vertex3f(0.117370605f, 0.265662789f, -1.f),
    C_Normal3f(0.999664426f, -0.025906859f, 0.f),
    C_Vertex3f(0.116210938f, 0.244146347f, 0.f),
    C_Vertex3f(0.116210938f, 0.244146347f, -1.f),
    C_Normal3f(0.999690711f, 0.0248676781f, 0.f),
    C_Vertex3f(0.116210938f, 0.220914841f, 0.f),
    C_Vertex3f(0.116210938f, 0.220914841f, -1.f),
    C_Normal3f(0.996813893f, 0.0797625408f, 0.f),
    C_Vertex3f(0.117331028f, 0.199118257f, 0.f),
    C_Vertex3f(0.117331028f, 0.199118257f, -1.f),
    C_Normal3f(0.989740729f, 0.14287433f, 0.f),
    C_Vertex3f(0.119571209f, 0.178920627f, 0.f),
    C_Vertex3f(0.119571209f, 0.178920627f, -1.f),
    C_Normal3f(0.97659272f, 0.215096921f, 0.f),
    C_Vertex3f(0.12293148f, 0.160321951f, 0.f),
    C_Vertex3f(0.12293148f, 0.160321951f, -1.f),
    C_Normal3f(0.95485121f, 0.297084332f, 0.f),
    C_Vertex3f(0.127411842f, 0.143322229f, 0.f),
    C_Vertex3f(0.127411842f, 0.143322229f, -1.f),
    C_Normal3f(0.921351075f, 0.38873148f, 0.f),
    C_Vertex3f(0.133012295f, 0.127921462f, 0.f),
    C_Vertex3f(0.133012295f, 0.127921462f, -1.f),
    C_Normal3f(0.872526586f, 0.488566548f, 0.f),
    C_Vertex3f(0.139732838f, 0.114119649f, 0.f),
    C_Vertex3f(0.139732838f, 0.114119649f, -1.f),
    C_Normal3f(0.805116951f, 0.593116045f, 0.f),
    C_Vertex3f(0.147573471f, 0.10191679f, 0.f),
    C_Vertex3f(0.147573471f, 0.10191679f, -1.f),
    C_Normal3f(0.730181158f, 0.683253646f, 0.f),
    C_Vertex3f(0.156534195f, 0.0913128853f, 0.f),
    C_Vertex3f(0.156534195f, 0.0913128853f, -1.f),
    C_Normal3f(0.653271496f, 0.757123709f, 0.f),
    C_Vertex3f(0.166177988f, 0.0820344687f, 0.f),
    C_Vertex3f(0.166177988f, 0.0820344687f, -1.f),
    C_Normal3f(0.564196467f, 0.825640559f, 0.f),
    C_Vertex3f(0.17650485f, 0.0740815401f, 0.f),
    C_Vertex3f(0.17650485f, 0.0740815401f, -1.f),
    C_Normal3f(0.465150356f, 0.885231733f, 0.f),
    C_Vertex3f(0.187514782f, 0.0674540997f, 0.f),
    C_Vertex3f(0.187514782f, 0.0674540997f, -1.f),
    C_Normal3f(0.359690905f, 0.933071494f, 0.f),
    C_Vertex3f(0.199207783f, 0.0621521473f, 0.f),
    C_Vertex3f(0.199207783f, 0.0621521473f, -1.f),
    C_Normal3f(0.252142906f, 0.967689991f, 0.f),
    C_Vertex3f(0.211583853f, 0.058175683f, 0.f),
    C_Vertex3f(0.211583853f, 0.058175683f, -1.f),
    C_Normal3f(0.146761522f, 0.989171922f, 0.f),
    C_Vertex3f(0.224642992f, 0.0555247068f, 0.f),
    C_Vertex3f(0.224642992f, 0.0555247068f, -1.f),
    C_Normal3f(0.0494966768f, 0.99877435f, 0.f),
    C_Vertex3f(0.238385201f, 0.0541992188f, 0.f),
    C_Vertex3f(0.238385201f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0367671363f, 0.999323905f, 0.f),
    C_Vertex3f(0.251389503f, 0.0541992188f, 0.f),
    C_Vertex3f(0.251389503f, 0.0541992188f, -1.f),
    C_Normal3f(-0.121542171f, 0.992586255f, 0.f),
    C_Vertex3f(0.262494683f, 0.0550862551f, 0.f),
    C_Vertex3f(0.262494683f, 0.0550862551f, -1.f),
    C_Normal3f(-0.208772928f, 0.977964103f, 0.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, 0.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, -1.f),
    C_Normal3f(-0.298960239f, 0.954265594f, 0.f),
    C_Vertex3f(0.283270597f, 0.0595214367f, 0.f),
    C_Vertex3f(0.283270597f, 0.0595214367f, -1.f),
    C_Normal3f(-0.389751405f, 0.920920014f, 0.f),
    C_Vertex3f(0.292941332f, 0.063069582f, 0.f),
    C_Vertex3f(0.292941332f, 0.063069582f, -1.f),
    C_Normal3f(-0.478471577f, 0.878103018f, 0.f),
    C_Vertex3f(0.302133918f, 0.0675047636f, 0.f),
    C_Vertex3f(0.302133918f, 0.0675047636f, -1.f),
    C_Normal3f(-0.562476099f, 0.826813638f, 0.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, 0.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, -1.f),
    C_Normal3f(-0.639527619f, 0.768768132f, 0.f),
    C_Vertex3f(0.319084644f, 0.0790362358f, 0.f),
    C_Vertex3f(0.319084644f, 0.0790362358f, -1.f),
    C_Normal3f(-0.715125501f, 0.698996067f, 0.f),
    C_Vertex3f(0.326842785f, 0.0861325264f, 0.f),
    C_Vertex3f(0.326842785f, 0.0861325264f, -1.f),
    C_Normal3f(-0.786699533f, 0.617336094f, 0.f),
    C_Vertex3f(0.333890796f, 0.094184041f, 0.f),
    C_Vertex3f(0.333890796f, 0.094184041f, -1.f),
    C_Normal3f(-0.845782697f, 0.533527553f, 0.f),
    C_Vertex3f(0.340228677f, 0.10319078f, 0.f),
    C_Vertex3f(0.340228677f, 0.10319078f, -1.f),
    C_Normal3f(-0.892610729f, 0.450828373f, 0.f),
    C_Vertex3f(0.345856428f, 0.113152742f, 0.f),
    C_Vertex3f(0.345856428f, 0.113152742f, -1.f),
    C_Normal3f(-0.928346336f, 0.37171635f, 0.f),
    C_Vertex3f(0.35077405f, 0.124069929f, 0.f),
    C_Vertex3f(0.35077405f, 0.124069929f, -1.f),
    C_Normal3f(-0.954632878f, 0.297785163f, 0.f),
    C_Vertex3f(0.354981542f, 0.13594234f, 0.f),
    C_Vertex3f(0.354981542f, 0.13594234f, -1.f),
    C_Normal3f(-0.973226845f, 0.229846776f, 0.f),
    C_Vertex3f(0.358478904f, 0.148769975f, 0.f),
    C_Vertex3f(0.358478904f, 0.148769975f, -1.f),
    C_Normal3f(-0.984033346f, 0.177983955f, 0.f),
    C_Vertex3f(0.361266136f, 0.162552834f, 0.f),
    C_Vertex3f(0.361266136f, 0.162552834f, -1.f),
    C_Normal3f(-0.990214288f, 0.139555246f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.959265828f, 0.282505304f, 0.f),
    C_Vertex3f(0.178451538f, -0.0771484375f, 0.f),
    C_Vertex3f(0.178451538f, -0.0771484375f, -1.f),
    C_Vertex3f(0.201171875f, 0.f, -1.f),
    C_Vertex3f(0.201171875f, 0.f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.201171875f, 0.f, 0.f),
    C_Vertex3f(0.201171875f, 0.f, -1.f),
    C_Vertex3f(0.259765625f, 0.f, -1.f),
    C_Vertex3f(0.259765625f, 0.f, 0.f),
    C_Normal3f(0.955515504f, -0.294941097f, 0.f),
    C_Vertex3f(0.259765625f, 0.f, 0.f),
    C_Vertex3f(0.259765625f, 0.f, -1.f),
    C_Vertex3f(0.245346069f, -0.0467147827f, -1.f),
    C_Vertex3f(0.245346069f, -0.0467147827f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.116983637f, 0.993133843f, 0.f),
    C_Vertex3f(0.245346069f, -0.0467147827f, 0.f),
    C_Vertex3f(0.245346069f, -0.0467147827f, -1.f),
    C_Normal3f(0.194189295f, 0.980964124f, 0.f),
    C_Vertex3f(0.254608154f, -0.0478057861f, 0.f),
    C_Vertex3f(0.254608154f, -0.0478057861f, -1.f),
    C_Normal3f(0.305636048f, 0.952148438f, 0.f),
    C_Vertex3f(0.270840645f, -0.0517616272f, 0.f),
    C_Vertex3f(0.270840645f, -0.0517616272f, -1.f),
    C_Normal3f(0.45934099f, 0.888260067f, 0.f),
    C_Vertex3f(0.284781456f, -0.0574913025f, 0.f),
    C_Vertex3f(0.284781456f, -0.0574913025f, -1.f),
    C_Normal3f(0.624141216f, 0.781311631f, 0.f),
    C_Vertex3f(0.296430588f, -0.064994812f, 0.f),
    C_Vertex3f(0.296430588f, -0.064994812f, -1.f),
    C_Normal3f(0.758728862f, 0.651406646f, 0.f),
    C_Vertex3f(0.30578804f, -0.0742721558f, 0.f),
    C_Vertex3f(0.30578804f, -0.0742721558f, -1.f),
    C_Normal3f(0.864426374f, 0.502759397f, 0.f),
    C_Vertex3f(0.312806129f, -0.0840682983f, 0.f),
    C_Vertex3f(0.312806129f, -0.0840682983f, -1.f),
    C_Normal3f(0.949106157f, 0.314956367f, 0.f),
    C_Vertex3f(0.317484856f, -0.0943832397f, 0.f),
    C_Vertex3f(0.317484856f, -0.0943832397f, -1.f),
    C_Normal3f(0.993615568f, 0.112818882f, 0.f),
    C_Vertex3f(0.319824219f, -0.10521698f, 0.f),
    C_Vertex3f(0.319824219f, -0.10521698f, -1.f),
    C_Normal3f(0.998467624f, -0.0553379245f, 0.f),
    C_Vertex3f(0.319824219f, -0.11498642f, 0.f),
    C_Vertex3f(0.319824219f, -0.11498642f, -1.f),
    C_Normal3f(0.98275584f, -0.184907913f, 0.f),
    C_Vertex3f(0.318842292f, -0.122933984f, 0.f),
    C_Vertex3f(0.318842292f, -0.122933984f, -1.f),
    C_Normal3f(0.951465666f, -0.307754815f, 0.f),
    C_Vertex3f(0.316878438f, -0.130642772f, 0.f),
    C_Vertex3f(0.316878438f, -0.130642772f, -1.f),
    C_Normal3f(0.905878186f, -0.423538417f, 0.f),
    C_Vertex3f(0.313932657f, -0.138112783f, 0.f),
    C_Vertex3f(0.313932657f, -0.138112783f, -1.f),
    C_Normal3f(0.84940356f, -0.527743995f, 0.f),
    C_Vertex3f(0.31000495f, -0.145344019f, 0.f),
    C_Vertex3f(0.31000495f, -0.145344019f, -1.f),
    C_Normal3f(0.786300123f, -0.61784476f, 0.f),
    C_Vertex3f(0.305095315f, -0.152336478f, 0.f),
    C_Vertex3f(0.305095315f, -0.152336478f, -1.f),
    C_Normal3f(0.720648766f, -0.693300307f, 0.f),
    C_Vertex3f(0.299203753f, -0.159090161f, 0.f),
    C_Vertex3f(0.299203753f, -0.159090161f, -1.f),
    C_Normal3f(0.655690968f, -0.75502938f, 0.f),
    C_Vertex3f(0.292330265f, -0.165605068f, 0.f),
    C_Vertex3f(0.292330265f, -0.165605068f, -1.f),
    C_Normal3f(0.575867295f, -0.817543089f, 0.f),
    C_Vertex3f(0.28447485f, -0.171881199f, 0.f),
    C_Vertex3f(0.28447485f, -0.171881199f, -1.f),
    C_Normal3f(0.478857249f, -0.877892852f, 0.f),
    C_Vertex3f(0.275623918f, -0.177372813f, 0.f),
    C_Vertex3f(0.275623918f, -0.177372813f, -1.f),
    C_Normal3f(0.384976745f, -0.922926307f, 0.f),
    C_Vertex3f(0.265777469f, -0.182079911f, 0.f),
    C_Vertex3f(0.265777469f, -0.182079911f, -1.f),
    C_Normal3f(0.297251523f, -0.954799294f, 0.f),
    C_Vertex3f(0.254935503f, -0.186002493f, 0.f),
    C_Vertex3f(0.254935503f, -0.186002493f, -1.f),
    C_Normal3f(0.217280537f, -0.976109207f, 0.f),
    C_Vertex3f(0.243098021f, -0.189140558f, 0.f),
    C_Vertex3f(0.243098021f, -0.189140558f, -1.f),
    C_Normal3f(0.145558298f, -0.989349663f, 0.f),
    C_Vertex3f(0.230265021f, -0.191494107f, 0.f),
    C_Vertex3f(0.230265021f, -0.191494107f, -1.f),
    C_Normal3f(0.0818652809f, -0.996643364f, 0.f),
    C_Vertex3f(0.216436505f, -0.19306314f, 0.f),
    C_Vertex3f(0.216436505f, -0.19306314f, -1.f),
    C_Normal3f(0.0247552618f, -0.999693513f, 0.f),
    C_Vertex3f(0.201612473f, -0.193847656f, 0.f),
    C_Vertex3f(0.201612473f, -0.193847656f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.184755325f, -0.193847656f, 0.f),
    C_Vertex3f(0.184755325f, -0.193847656f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.167952538f, -0.193847656f, 0.f),
    C_Vertex3f(0.167952538f, -0.193847656f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.152241707f, -0.193847656f, 0.f),
    C_Vertex3f(0.152241707f, -0.193847656f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.137622833f, -0.193847656f, 0.f),
    C_Vertex3f(0.137622833f, -0.193847656f, -1.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.130859375f, -0.193847656f, 0.f),
    C_Vertex3f(0.130859375f, -0.193847656f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.996240556f, 0.0866296142f, 0.f),
    C_Vertex3f(0.130859375f, -0.193847656f, 0.f),
    C_Vertex3f(0.130859375f, -0.193847656f, -1.f),
    C_Vertex3f(0.135742188f, -0.137695313f, -1.f),
    C_Vertex3f(0.135742188f, -0.137695313f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.0523584522f, 0.998628378f, 0.f),
    C_Vertex3f(0.135742188f, -0.137695313f, 0.f),
    C_Vertex3f(0.135742188f, -0.137695313f, -1.f),
    C_Normal3f(0.0478803068f, 0.998853087f, 0.f),
    C_Vertex3f(0.142726898f, -0.138061523f, 0.f),
    C_Vertex3f(0.142726898f, -0.138061523f, -1.f),
    C_Normal3f(0.0408847593f, 0.999163866f, 0.f),
    C_Vertex3f(0.154841423f, -0.13861084f, 0.f),
    C_Vertex3f(0.154841423f, -0.13861084f, -1.f),
    C_Normal3f(0.0294185635f, 0.999567151f, 0.f),
    C_Vertex3f(0.165101051f, -0.138977051f, 0.f),
    C_Vertex3f(0.165101051f, -0.138977051f, -1.f),
    C_Normal3f(0.00834572129f, 0.999965191f, 0.f),
    C_Vertex3f(0.173505783f, -0.139160156f, 0.f),
    C_Vertex3f(0.173505783f, -0.139160156f, -1.f),
    C_Normal3f(-0.04573109f, 0.998953879f, 0.f),
    C_Vertex3f(0.187040329f, -0.139160156f, 0.f),
    C_Vertex3f(0.187040329f, -0.139160156f, -1.f),
    C_Normal3f(-0.132396847f, 0.991196811f, 0.f),
    C_Vertex3f(0.204504013f, -0.137741089f, 0.f),
    C_Vertex3f(0.204504013f, -0.137741089f, -1.f),
    C_Normal3f(-0.265544683f, 0.964098513f, 0.f),
    C_Vertex3f(0.218912125f, -0.134902954f, 0.f),
    C_Vertex3f(0.218912125f, -0.134902954f, -1.f),
    C_Normal3f(-0.445162714f, 0.895449579f, 0.f),
    C_Vertex3f(0.230264664f, -0.130645752f, 0.f),
    C_Vertex3f(0.230264664f, -0.130645752f, -1.f),
    C_Normal3f(-0.633790612f, 0.773504674f, 0.f),
    C_Vertex3f(0.237577438f, -0.125623703f, 0.f),
    C_Vertex3f(0.237577438f, -0.125623703f, -1.f),
    C_Normal3f(-0.790464282f, 0.612508297f, 0.f),
    C_Vertex3f(0.242323875f, -0.120764732f, 0.f),
    C_Vertex3f(0.242323875f, -0.120764732f, -1.f),
    C_Normal3f(-0.92062366f, 0.390451133f, 0.f),
    C_Vertex3f(0.245488167f, -0.115414619f, 0.f),
    C_Vertex3f(0.245488167f, -0.115414619f, -1.f),
    C_Normal3f(-0.992604256f, 0.121395536f, 0.f),
    C_Vertex3f(0.247070313f, -0.109573364f, 0.f),
    C_Vertex3f(0.247070313f, -0.109573364f, -1.f),
    C_Normal3f(-0.984436631f, -0.175739869f, 0.f),
    C_Vertex3f(0.247070313f, -0.102478027f, 0.f),
    C_Vertex3f(0.247070313f, -0.102478027f, -1.f),
    C_Normal3f(-0.878779709f, -0.477227718f, 0.f),
    C_Vertex3f(0.244556427f, -0.0954914093f, 0.f),
    C_Vertex3f(0.244556427f, -0.0954914093f, -1.f),
    C_Normal3f(-0.710703135f, -0.703492045f, 0.f),
    C_Vertex3f(0.240785599f, -0.0909051895f, 0.f),
    C_Vertex3f(0.240785599f, -0.0909051895f, -1.f),
    C_Normal3f(-0.553116202f, -0.833104074f, 0.f),
    C_Vertex3f(0.237206459f, -0.0880661011f, 0.f),
    C_Vertex3f(0.237206459f, -0.0880661011f, -1.f),
    C_Normal3f(-0.4364748f, -0.899716496f, 0.f),
    C_Vertex3f(0.232562065f, -0.0854454041f, 0.f),
    C_Vertex3f(0.232562065f, -0.0854454041f, -1.f),
    C_Normal3f(-0.344821334f, -0.938668311f, 0.f),
    C_Vertex3f(0.226852417f, -0.0830430984f, 0.f),
    C_Vertex3f(0.226852417f, -0.0830430984f, -1.f),
    C_Normal3f(-0.238420397f, -0.971162021f, 0.f),
    C_Vertex3f(0.220077515f, -0.0808591843f, 0.f),
    C_Vertex3f(0.220077515f, -0.0808591843f, -1.f),
    C_Normal3f(-0.141307235f, -0.989965796f, 0.f),
    C_Vertex3f(0.211063385f, -0.0791668892f, 0.f),
    C_Vertex3f(0.211063385f, -0.0791668892f, -1.f),
    C_Normal3f(-0.0769447833f, -0.997035325f, 0.f),
    C_Vertex3f(0.199810028f, -0.0779662132f, 0.f),
    C_Vertex3f(0.199810028f, -0.0779662132f, -1.f),
    C_Normal3f(-0.0382600538f, -0.999267817f, 0.f),
    C_Vertex3f(0.186317444f, -0.0772571564f, 0.f),
    C_Vertex3f(0.186317444f, -0.0772571564f, -1.f),
    C_Normal3f(-0.0138202142f, -0.999904573f, 0.f),
    C_Vertex3f(0.178451538f, -0.0771484375f, 0.f),
    C_Vertex3f(0.178451538f, -0.0771484375f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.432128906f, 0.327636719f, 0.f),
    C_Vertex3f(0.430435181f, 0.336440086f, 0.f),
    C_Vertex3f(0.425968289f, 0.353042006f, 0.f),
    C_Vertex3f(0.420421958f, 0.368639112f, 0.f),
    C_Vertex3f(0.413796186f, 0.383231401f, 0.f),
    C_Vertex3f(0.406090975f, 0.396818876f, 0.f),
    C_Vertex3f(0.397306323f, 0.409401536f, 0.f),
    C_Vertex3f(0.387442231f, 0.420979381f, 0.f),
    C_Vertex3f(0.376498699f, 0.43155241f, 0.f),
    C_Vertex3f(0.364475727f, 0.441120625f, 0.f),
    C_Vertex3f(0.35546875f, 0.315917969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323760033f, 0.380568504f, 0.f),
    C_Vertex3f(0.335520267f, 0.365829945f, 0.f),
    C_Vertex3f(0.337847352f, 0.456668973f, 0.f),
    C_Vertex3f(0.345180035f, 0.348187923f, 0.f),
    C_Vertex3f(0.351591945f, 0.449492812f, 0.f),
    C_Vertex3f(0.352739334f, 0.327642441f, 0.f),
    C_Vertex3f(0.364475727f, 0.441120625f, 0.f),
    C_Vertex3f(0.35546875f, 0.315917969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.214784861f, 0.406159043f, 0.f),
    C_Vertex3f(0.228237867f, 0.408823848f, 0.f),
    C_Vertex3f(0.231782913f, 0.474609375f, 0.f),
    C_Vertex3f(0.242386818f, 0.41015625f, 0.f),
    C_Vertex3f(0.256212234f, 0.474609375f, 0.f),
    C_Vertex3f(0.259634018f, 0.41015625f, 0.f),
    C_Vertex3f(0.274260879f, 0.473413348f, 0.f),
    C_Vertex3f(0.277836323f, 0.407197475f, 0.f),
    C_Vertex3f(0.291448712f, 0.471021295f, 0.f),
    C_Vertex3f(0.294591427f, 0.401279926f, 0.f),
    C_Vertex3f(0.307775736f, 0.467433214f, 0.f),
    C_Vertex3f(0.30989933f, 0.392403603f, 0.f),
    C_Vertex3f(0.323241949f, 0.462649107f, 0.f),
    C_Vertex3f(0.323760033f, 0.380568504f, 0.f),
    C_Vertex3f(0.337847352f, 0.456668973f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.202954769f, 0.471087456f, 0.f),
    C_Vertex3f(0.175247669f, 0.464043617f, 0.f),
    C_Vertex3f(0.178601503f, 0.390170217f, 0.f),
    C_Vertex3f(0.189966679f, 0.396832228f, 0.f),
    C_Vertex3f(0.202027798f, 0.402161837f, 0.f),
    C_Vertex3f(0.214784861f, 0.406159043f, 0.f),
    C_Vertex3f(0.231782913f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.175247669f, 0.464043617f, 0.f),
    C_Vertex3f(0.148661613f, 0.453477859f, 0.f),
    C_Vertex3f(0.148681641f, 0.36218977f, 0.f),
    C_Vertex3f(0.157958984f, 0.372848988f, 0.f),
    C_Vertex3f(0.167932272f, 0.382175803f, 0.f),
    C_Vertex3f(0.178601503f, 0.390170217f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.123168945f, 0.304042578f, 0.f),
    C_Vertex3f(0.127807617f, 0.320905924f, 0.f),
    C_Vertex3f(0.129562855f, 0.442912102f, 0.f),
    C_Vertex3f(0.133605957f, 0.336218238f, 0.f),
    C_Vertex3f(0.148661613f, 0.453477859f, 0.f),
    C_Vertex3f(0.140563965f, 0.34997952f, 0.f),
    C_Vertex3f(0.148681641f, 0.36218977f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116210938f, 0.244146347f, 0.f),
    C_Vertex3f(0.117370605f, 0.265662789f, 0.f),
    C_Vertex3f(0.117643118f, 0.434987903f, 0.f),
    C_Vertex3f(0.119689941f, 0.2856282f, 0.f),
    C_Vertex3f(0.129562855f, 0.442912102f, 0.f),
    C_Vertex3f(0.123168945f, 0.304042578f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115951657f, 0.0305939913f, 0.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, 0.f),
    C_Vertex3f(0.117331028f, 0.199118257f, 0.f),
    C_Vertex3f(0.116210938f, 0.220914841f, 0.f),
    C_Vertex3f(0.116210938f, 0.244146347f, 0.f),
    C_Vertex3f(0.10653615f, 0.426183343f, 0.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.10653615f, 0.426183343f, 0.f),
    C_Vertex3f(0.096241951f, 0.416498423f, 0.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, 0.f),
    C_Vertex3f(0.0867605209f, 0.405933142f, 0.f),
    C_Vertex3f(0.0858283043f, 0.0597710609f, 0.f),
    C_Vertex3f(0.0780918598f, 0.3944875f, 0.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, 0.f),
    C_Vertex3f(0.0702359676f, 0.382161498f, 0.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0460145473f, 0.138875246f, 0.f),
    C_Vertex3f(0.0532534122f, 0.116570234f, 0.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, 0.f),
    C_Vertex3f(0.0623019934f, 0.0959511995f, 0.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, 0.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.03515625f, 0.249214172f, 0.f),
    C_Vertex3f(0.0369659662f, 0.1885432f, 0.f),
    C_Vertex3f(0.0382714272f, 0.284432888f, 0.f),
    C_Vertex3f(0.0405853987f, 0.162866235f, 0.f),
    C_Vertex3f(0.0445017815f, 0.317303181f, 0.f),
    C_Vertex3f(0.0460145473f, 0.138875246f, 0.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232562065f, -0.0854454041f, 0.f),
    C_Vertex3f(0.237206459f, -0.0880661011f, 0.f),
    C_Vertex3f(0.234351158f, -0.0102539063f, 0.f),
    C_Vertex3f(0.245346069f, -0.0467147827f, 0.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, 0.f),
    C_Vertex3f(0.256620824f, -0.0101881875f, 0.f),
    C_Vertex3f(0.259765625f, 0.f, 0.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, 0.f),
    C_Vertex3f(0.262494683f, 0.0550862551f, 0.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.240785599f, -0.0909051895f, 0.f),
    C_Vertex3f(0.244556427f, -0.0954914093f, 0.f),
    C_Vertex3f(0.245346069f, -0.0467147827f, 0.f),
    C_Vertex3f(0.247070313f, -0.102478027f, 0.f),
    C_Vertex3f(0.254608154f, -0.0478057861f, 0.f),
    C_Vertex3f(0.254935503f, -0.186002493f, 0.f),
    C_Vertex3f(0.270840645f, -0.0517616272f, 0.f),
    C_Vertex3f(0.265777469f, -0.182079911f, 0.f),
    C_Vertex3f(0.275623918f, -0.177372813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.234351158f, -0.0102539063f, 0.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, 0.f),
    C_Vertex3f(0.259765625f, 0.f, 0.f),
    C_Vertex3f(0.211698174f, -0.00830876827f, 0.f),
    C_Vertex3f(0.220077515f, -0.0808591843f, 0.f),
    C_Vertex3f(0.226852417f, -0.0830430984f, 0.f),
    C_Vertex3f(0.232562065f, -0.0854454041f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.211698174f, -0.00830876827f, 0.f),
    C_Vertex3f(0.259765625f, 0.f, 0.f),
    C_Vertex3f(0.201171875f, 0.f, 0.f),
    C_Vertex3f(0.199381709f, -0.00607864186f, 0.f),
    C_Vertex3f(0.199810028f, -0.0779662132f, 0.f),
    C_Vertex3f(0.211063385f, -0.0791668892f, 0.f),
    C_Vertex3f(0.220077515f, -0.0808591843f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.199381709f, -0.00607864186f, 0.f),
    C_Vertex3f(0.178451538f, -0.0771484375f, 0.f),
    C_Vertex3f(0.186317444f, -0.0772571564f, 0.f),
    C_Vertex3f(0.199810028f, -0.0779662132f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259765625f, 0.f, 0.f),
    C_Vertex3f(0.262494683f, 0.0550862551f, 0.f),
    C_Vertex3f(0.251389503f, 0.0541992188f, 0.f),
    C_Vertex3f(0.238385201f, 0.0541992188f, 0.f),
    C_Vertex3f(0.224642992f, 0.0555247068f, 0.f),
    C_Vertex3f(0.211583853f, 0.058175683f, 0.f),
    C_Vertex3f(0.201171875f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.201171875f, 0.f, 0.f),
    C_Vertex3f(0.211583853f, 0.058175683f, 0.f),
    C_Vertex3f(0.199381709f, -0.00607864186f, 0.f),
    C_Vertex3f(0.199207783f, 0.0621521473f, 0.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, 0.f),
    C_Vertex3f(0.187514782f, 0.0674540997f, 0.f),
    C_Vertex3f(0.17650485f, 0.0740815401f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169895887f, 0.00141692162f, 0.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, 0.f),
    C_Vertex3f(0.17650485f, 0.0740815401f, 0.f),
    C_Vertex3f(0.166177988f, 0.0820344687f, 0.f),
    C_Vertex3f(0.156534195f, 0.0913128853f, 0.f),
    C_Vertex3f(0.150746584f, 0.00919747353f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150746584f, 0.00919747353f, 0.f),
    C_Vertex3f(0.156534195f, 0.0913128853f, 0.f),
    C_Vertex3f(0.147573471f, 0.10191679f, 0.f),
    C_Vertex3f(0.139732838f, 0.114119649f, 0.f),
    C_Vertex3f(0.133012295f, 0.127921462f, 0.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.132765174f, 0.0189231634f, 0.f),
    C_Vertex3f(0.133012295f, 0.127921462f, 0.f),
    C_Vertex3f(0.127411842f, 0.143322229f, 0.f),
    C_Vertex3f(0.12293148f, 0.160321951f, 0.f),
    C_Vertex3f(0.119571209f, 0.178920627f, 0.f),
    C_Vertex3f(0.117331028f, 0.199118257f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.439453125f, 0.159667969f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, 0.f),
    C_Vertex3f(0.368021011f, 0.0292029381f, 0.f),
    C_Vertex3f(0.38107872f, 0.0404763222f, 0.f),
    C_Vertex3f(0.392900229f, 0.0528322458f, 0.f),
    C_Vertex3f(0.403485537f, 0.066270709f, 0.f),
    C_Vertex3f(0.412834644f, 0.0807917118f, 0.f),
    C_Vertex3f(0.420947552f, 0.0963952541f, 0.f),
    C_Vertex3f(0.427824259f, 0.113081336f, 0.f),
    C_Vertex3f(0.433464766f, 0.130849957f, 0.f),
    C_Vertex3f(0.437869072f, 0.149701118f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.368021011f, 0.0292029381f, 0.f),
    C_Vertex3f(0.362304688f, 0.169921875f, 0.f),
    C_Vertex3f(0.361266136f, 0.162552834f, 0.f),
    C_Vertex3f(0.358478904f, 0.148769975f, 0.f),
    C_Vertex3f(0.354981542f, 0.13594234f, 0.f),
    C_Vertex3f(0.354218721f, 0.019338727f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.354218721f, 0.019338727f, 0.f),
    C_Vertex3f(0.354981542f, 0.13594234f, 0.f),
    C_Vertex3f(0.35077405f, 0.124069929f, 0.f),
    C_Vertex3f(0.345856428f, 0.113152742f, 0.f),
    C_Vertex3f(0.340228677f, 0.10319078f, 0.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.340228677f, 0.10319078f, 0.f),
    C_Vertex3f(0.333890796f, 0.094184041f, 0.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, 0.f),
    C_Vertex3f(0.326842785f, 0.0861325264f, 0.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, 0.f),
    C_Vertex3f(0.319084644f, 0.0790362358f, 0.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.308344364f, -0.00179886818f, 0.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, 0.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, 0.f),
    C_Vertex3f(0.302133918f, 0.0675047636f, 0.f),
    C_Vertex3f(0.292941332f, 0.063069582f, 0.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.283270597f, 0.0595214367f, 0.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, 0.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, 0.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, 0.f),
    C_Vertex3f(0.292941332f, 0.063069582f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.313932657f, -0.138112783f, 0.f),
    C_Vertex3f(0.316878438f, -0.130642772f, 0.f),
    C_Vertex3f(0.317484856f, -0.0943832397f, 0.f),
    C_Vertex3f(0.318842292f, -0.122933984f, 0.f),
    C_Vertex3f(0.319824219f, -0.10521698f, 0.f),
    C_Vertex3f(0.319824219f, -0.11498642f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.299203753f, -0.159090161f, 0.f),
    C_Vertex3f(0.305095315f, -0.152336478f, 0.f),
    C_Vertex3f(0.30578804f, -0.0742721558f, 0.f),
    C_Vertex3f(0.31000495f, -0.145344019f, 0.f),
    C_Vertex3f(0.312806129f, -0.0840682983f, 0.f),
    C_Vertex3f(0.313932657f, -0.138112783f, 0.f),
    C_Vertex3f(0.317484856f, -0.0943832397f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.275623918f, -0.177372813f, 0.f),
    C_Vertex3f(0.28447485f, -0.171881199f, 0.f),
    C_Vertex3f(0.284781456f, -0.0574913025f, 0.f),
    C_Vertex3f(0.292330265f, -0.165605068f, 0.f),
    C_Vertex3f(0.296430588f, -0.064994812f, 0.f),
    C_Vertex3f(0.299203753f, -0.159090161f, 0.f),
    C_Vertex3f(0.30578804f, -0.0742721558f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.247070313f, -0.102478027f, 0.f),
    C_Vertex3f(0.247070313f, -0.109573364f, 0.f),
    C_Vertex3f(0.254935503f, -0.186002493f, 0.f),
    C_Vertex3f(0.245488167f, -0.115414619f, 0.f),
    C_Vertex3f(0.243098021f, -0.189140558f, 0.f),
    C_Vertex3f(0.242323875f, -0.120764732f, 0.f),
    C_Vertex3f(0.237577438f, -0.125623703f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.230265021f, -0.191494107f, 0.f),
    C_Vertex3f(0.243098021f, -0.189140558f, 0.f),
    C_Vertex3f(0.237577438f, -0.125623703f, 0.f),
    C_Vertex3f(0.230264664f, -0.130645752f, 0.f),
    C_Vertex3f(0.218912125f, -0.134902954f, 0.f),
    C_Vertex3f(0.216436505f, -0.19306314f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.165101051f, -0.138977051f, 0.f),
    C_Vertex3f(0.167952538f, -0.193847656f, 0.f),
    C_Vertex3f(0.173505783f, -0.139160156f, 0.f),
    C_Vertex3f(0.184755325f, -0.193847656f, 0.f),
    C_Vertex3f(0.187040329f, -0.139160156f, 0.f),
    C_Vertex3f(0.201612473f, -0.193847656f, 0.f),
    C_Vertex3f(0.204504013f, -0.137741089f, 0.f),
    C_Vertex3f(0.216436505f, -0.19306314f, 0.f),
    C_Vertex3f(0.218912125f, -0.134902954f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.135742188f, -0.137695313f, 0.f),
    C_Vertex3f(0.137622833f, -0.193847656f, 0.f),
    C_Vertex3f(0.142726898f, -0.138061523f, 0.f),
    C_Vertex3f(0.152241707f, -0.193847656f, 0.f),
    C_Vertex3f(0.154841423f, -0.13861084f, 0.f),
    C_Vertex3f(0.167952538f, -0.193847656f, 0.f),
    C_Vertex3f(0.165101051f, -0.138977051f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.137622833f, -0.193847656f, 0.f),
    C_Vertex3f(0.135742188f, -0.137695313f, 0.f),
    C_Vertex3f(0.130859375f, -0.193847656f, 0.f),
    C_Vertex3f(0.270840645f, -0.0517616272f, 0.f),
    C_Vertex3f(0.275623918f, -0.177372813f, 0.f),
    C_Vertex3f(0.284781456f, -0.0574913025f, 0.f),
    C_Vertex3f(0.245346069f, -0.0467147827f, 0.f),
    C_Vertex3f(0.237206459f, -0.0880661011f, 0.f),
    C_Vertex3f(0.240785599f, -0.0909051895f, 0.f),
    C_Vertex3f(0.0369659662f, 0.1885432f, 0.f),
    C_Vertex3f(0.03515625f, 0.249214172f, 0.f),
    C_Vertex3f(0.03515625f, 0.215906143f, 0.f),
    C_Vertex3f(0.10653615f, 0.426183343f, 0.f),
    C_Vertex3f(0.116210938f, 0.244146347f, 0.f),
    C_Vertex3f(0.117643118f, 0.434987903f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.432128906f, 0.327636719f, -1.f),
    C_Vertex3f(0.430435181f, 0.336440086f, -1.f),
    C_Vertex3f(0.425968289f, 0.353042006f, -1.f),
    C_Vertex3f(0.420421958f, 0.368639112f, -1.f),
    C_Vertex3f(0.413796186f, 0.383231401f, -1.f),
    C_Vertex3f(0.406090975f, 0.396818876f, -1.f),
    C_Vertex3f(0.397306323f, 0.409401536f, -1.f),
    C_Vertex3f(0.387442231f, 0.420979381f, -1.f),
    C_Vertex3f(0.376498699f, 0.43155241f, -1.f),
    C_Vertex3f(0.364475727f, 0.441120625f, -1.f),
    C_Vertex3f(0.35546875f, 0.315917969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.323760033f, 0.380568504f, -1.f),
    C_Vertex3f(0.335520267f, 0.365829945f, -1.f),
    C_Vertex3f(0.337847352f, 0.456668973f, -1.f),
    C_Vertex3f(0.345180035f, 0.348187923f, -1.f),
    C_Vertex3f(0.351591945f, 0.449492812f, -1.f),
    C_Vertex3f(0.352739334f, 0.327642441f, -1.f),
    C_Vertex3f(0.364475727f, 0.441120625f, -1.f),
    C_Vertex3f(0.35546875f, 0.315917969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.214784861f, 0.406159043f, -1.f),
    C_Vertex3f(0.228237867f, 0.408823848f, -1.f),
    C_Vertex3f(0.231782913f, 0.474609375f, -1.f),
    C_Vertex3f(0.242386818f, 0.41015625f, -1.f),
    C_Vertex3f(0.256212234f, 0.474609375f, -1.f),
    C_Vertex3f(0.259634018f, 0.41015625f, -1.f),
    C_Vertex3f(0.274260879f, 0.473413348f, -1.f),
    C_Vertex3f(0.277836323f, 0.407197475f, -1.f),
    C_Vertex3f(0.291448712f, 0.471021295f, -1.f),
    C_Vertex3f(0.294591427f, 0.401279926f, -1.f),
    C_Vertex3f(0.307775736f, 0.467433214f, -1.f),
    C_Vertex3f(0.30989933f, 0.392403603f, -1.f),
    C_Vertex3f(0.323241949f, 0.462649107f, -1.f),
    C_Vertex3f(0.323760033f, 0.380568504f, -1.f),
    C_Vertex3f(0.337847352f, 0.456668973f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.202954769f, 0.471087456f, -1.f),
    C_Vertex3f(0.175247669f, 0.464043617f, -1.f),
    C_Vertex3f(0.178601503f, 0.390170217f, -1.f),
    C_Vertex3f(0.189966679f, 0.396832228f, -1.f),
    C_Vertex3f(0.202027798f, 0.402161837f, -1.f),
    C_Vertex3f(0.214784861f, 0.406159043f, -1.f),
    C_Vertex3f(0.231782913f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.175247669f, 0.464043617f, -1.f),
    C_Vertex3f(0.148661613f, 0.453477859f, -1.f),
    C_Vertex3f(0.148681641f, 0.36218977f, -1.f),
    C_Vertex3f(0.157958984f, 0.372848988f, -1.f),
    C_Vertex3f(0.167932272f, 0.382175803f, -1.f),
    C_Vertex3f(0.178601503f, 0.390170217f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.123168945f, 0.304042578f, -1.f),
    C_Vertex3f(0.127807617f, 0.320905924f, -1.f),
    C_Vertex3f(0.129562855f, 0.442912102f, -1.f),
    C_Vertex3f(0.133605957f, 0.336218238f, -1.f),
    C_Vertex3f(0.148661613f, 0.453477859f, -1.f),
    C_Vertex3f(0.140563965f, 0.34997952f, -1.f),
    C_Vertex3f(0.148681641f, 0.36218977f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116210938f, 0.244146347f, -1.f),
    C_Vertex3f(0.117370605f, 0.265662789f, -1.f),
    C_Vertex3f(0.117643118f, 0.434987903f, -1.f),
    C_Vertex3f(0.119689941f, 0.2856282f, -1.f),
    C_Vertex3f(0.129562855f, 0.442912102f, -1.f),
    C_Vertex3f(0.123168945f, 0.304042578f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.115951657f, 0.0305939913f, -1.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, -1.f),
    C_Vertex3f(0.117331028f, 0.199118257f, -1.f),
    C_Vertex3f(0.116210938f, 0.220914841f, -1.f),
    C_Vertex3f(0.116210938f, 0.244146347f, -1.f),
    C_Vertex3f(0.10653615f, 0.426183343f, -1.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.10653615f, 0.426183343f, -1.f),
    C_Vertex3f(0.096241951f, 0.416498423f, -1.f),
    C_Vertex3f(0.100306034f, 0.0442099571f, -1.f),
    C_Vertex3f(0.0867605209f, 0.405933142f, -1.f),
    C_Vertex3f(0.0858283043f, 0.0597710609f, -1.f),
    C_Vertex3f(0.0780918598f, 0.3944875f, -1.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, -1.f),
    C_Vertex3f(0.0702359676f, 0.382161498f, -1.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0460145473f, 0.138875246f, -1.f),
    C_Vertex3f(0.0532534122f, 0.116570234f, -1.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, -1.f),
    C_Vertex3f(0.0623019934f, 0.0959511995f, -1.f),
    C_Vertex3f(0.0631928444f, 0.368955135f, -1.f),
    C_Vertex3f(0.0731602907f, 0.0770181417f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.03515625f, 0.249214172f, -1.f),
    C_Vertex3f(0.0369659662f, 0.1885432f, -1.f),
    C_Vertex3f(0.0382714272f, 0.284432888f, -1.f),
    C_Vertex3f(0.0405853987f, 0.162866235f, -1.f),
    C_Vertex3f(0.0445017815f, 0.317303181f, -1.f),
    C_Vertex3f(0.0460145473f, 0.138875246f, -1.f),
    C_Vertex3f(0.0538473129f, 0.34782505f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232562065f, -0.0854454041f, -1.f),
    C_Vertex3f(0.237206459f, -0.0880661011f, -1.f),
    C_Vertex3f(0.234351158f, -0.0102539063f, -1.f),
    C_Vertex3f(0.245346069f, -0.0467147827f, -1.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, -1.f),
    C_Vertex3f(0.256620824f, -0.0101881875f, -1.f),
    C_Vertex3f(0.259765625f, 0.f, -1.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, -1.f),
    C_Vertex3f(0.262494683f, 0.0550862551f, -1.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.240785599f, -0.0909051895f, -1.f),
    C_Vertex3f(0.244556427f, -0.0954914093f, -1.f),
    C_Vertex3f(0.245346069f, -0.0467147827f, -1.f),
    C_Vertex3f(0.247070313f, -0.102478027f, -1.f),
    C_Vertex3f(0.254608154f, -0.0478057861f, -1.f),
    C_Vertex3f(0.254935503f, -0.186002493f, -1.f),
    C_Vertex3f(0.270840645f, -0.0517616272f, -1.f),
    C_Vertex3f(0.265777469f, -0.182079911f, -1.f),
    C_Vertex3f(0.275623918f, -0.177372813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.234351158f, -0.0102539063f, -1.f),
    C_Vertex3f(0.255768776f, -0.0102539063f, -1.f),
    C_Vertex3f(0.259765625f, 0.f, -1.f),
    C_Vertex3f(0.211698174f, -0.00830876827f, -1.f),
    C_Vertex3f(0.220077515f, -0.0808591843f, -1.f),
    C_Vertex3f(0.226852417f, -0.0830430984f, -1.f),
    C_Vertex3f(0.232562065f, -0.0854454041f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.211698174f, -0.00830876827f, -1.f),
    C_Vertex3f(0.259765625f, 0.f, -1.f),
    C_Vertex3f(0.201171875f, 0.f, -1.f),
    C_Vertex3f(0.199381709f, -0.00607864186f, -1.f),
    C_Vertex3f(0.199810028f, -0.0779662132f, -1.f),
    C_Vertex3f(0.211063385f, -0.0791668892f, -1.f),
    C_Vertex3f(0.220077515f, -0.0808591843f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.199381709f, -0.00607864186f, -1.f),
    C_Vertex3f(0.178451538f, -0.0771484375f, -1.f),
    C_Vertex3f(0.186317444f, -0.0772571564f, -1.f),
    C_Vertex3f(0.199810028f, -0.0779662132f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259765625f, 0.f, -1.f),
    C_Vertex3f(0.262494683f, 0.0550862551f, -1.f),
    C_Vertex3f(0.251389503f, 0.0541992188f, -1.f),
    C_Vertex3f(0.238385201f, 0.0541992188f, -1.f),
    C_Vertex3f(0.224642992f, 0.0555247068f, -1.f),
    C_Vertex3f(0.211583853f, 0.058175683f, -1.f),
    C_Vertex3f(0.201171875f, 0.f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.201171875f, 0.f, -1.f),
    C_Vertex3f(0.211583853f, 0.058175683f, -1.f),
    C_Vertex3f(0.199381709f, -0.00607864186f, -1.f),
    C_Vertex3f(0.199207783f, 0.0621521473f, -1.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, -1.f),
    C_Vertex3f(0.187514782f, 0.0674540997f, -1.f),
    C_Vertex3f(0.17650485f, 0.0740815401f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169895887f, 0.00141692162f, -1.f),
    C_Vertex3f(0.190213084f, -0.00441849232f, -1.f),
    C_Vertex3f(0.17650485f, 0.0740815401f, -1.f),
    C_Vertex3f(0.166177988f, 0.0820344687f, -1.f),
    C_Vertex3f(0.156534195f, 0.0913128853f, -1.f),
    C_Vertex3f(0.150746584f, 0.00919747353f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150746584f, 0.00919747353f, -1.f),
    C_Vertex3f(0.156534195f, 0.0913128853f, -1.f),
    C_Vertex3f(0.147573471f, 0.10191679f, -1.f),
    C_Vertex3f(0.139732838f, 0.114119649f, -1.f),
    C_Vertex3f(0.133012295f, 0.127921462f, -1.f),
    C_Vertex3f(0.132765174f, 0.0189231634f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.132765174f, 0.0189231634f, -1.f),
    C_Vertex3f(0.133012295f, 0.127921462f, -1.f),
    C_Vertex3f(0.127411842f, 0.143322229f, -1.f),
    C_Vertex3f(0.12293148f, 0.160321951f, -1.f),
    C_Vertex3f(0.119571209f, 0.178920627f, -1.f),
    C_Vertex3f(0.117331028f, 0.199118257f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.439453125f, 0.159667969f, -1.f),
    C_Vertex3f(0.362304688f, 0.169921875f, -1.f),
    C_Vertex3f(0.368021011f, 0.0292029381f, -1.f),
    C_Vertex3f(0.38107872f, 0.0404763222f, -1.f),
    C_Vertex3f(0.392900229f, 0.0528322458f, -1.f),
    C_Vertex3f(0.403485537f, 0.066270709f, -1.f),
    C_Vertex3f(0.412834644f, 0.0807917118f, -1.f),
    C_Vertex3f(0.420947552f, 0.0963952541f, -1.f),
    C_Vertex3f(0.427824259f, 0.113081336f, -1.f),
    C_Vertex3f(0.433464766f, 0.130849957f, -1.f),
    C_Vertex3f(0.437869072f, 0.149701118f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.368021011f, 0.0292029381f, -1.f),
    C_Vertex3f(0.362304688f, 0.169921875f, -1.f),
    C_Vertex3f(0.361266136f, 0.162552834f, -1.f),
    C_Vertex3f(0.358478904f, 0.148769975f, -1.f),
    C_Vertex3f(0.354981542f, 0.13594234f, -1.f),
    C_Vertex3f(0.354218721f, 0.019338727f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.354218721f, 0.019338727f, -1.f),
    C_Vertex3f(0.354981542f, 0.13594234f, -1.f),
    C_Vertex3f(0.35077405f, 0.124069929f, -1.f),
    C_Vertex3f(0.345856428f, 0.113152742f, -1.f),
    C_Vertex3f(0.340228677f, 0.10319078f, -1.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.340228677f, 0.10319078f, -1.f),
    C_Vertex3f(0.333890796f, 0.094184041f, -1.f),
    C_Vertex3f(0.33967185f, 0.0108836889f, -1.f),
    C_Vertex3f(0.326842785f, 0.0861325264f, -1.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, -1.f),
    C_Vertex3f(0.319084644f, 0.0790362358f, -1.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.308344364f, -0.00179886818f, -1.f),
    C_Vertex3f(0.324380398f, 0.00383782387f, -1.f),
    C_Vertex3f(0.310848355f, 0.0728269815f, -1.f),
    C_Vertex3f(0.302133918f, 0.0675047636f, -1.f),
    C_Vertex3f(0.292941332f, 0.063069582f, -1.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.283270597f, 0.0595214367f, -1.f),
    C_Vertex3f(0.273121715f, 0.0568603277f, -1.f),
    C_Vertex3f(0.274038553f, -0.00884473324f, -1.f),
    C_Vertex3f(0.291563749f, -0.00602638721f, -1.f),
    C_Vertex3f(0.292941332f, 0.063069582f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.313932657f, -0.138112783f, -1.f),
    C_Vertex3f(0.316878438f, -0.130642772f, -1.f),
    C_Vertex3f(0.317484856f, -0.0943832397f, -1.f),
    C_Vertex3f(0.318842292f, -0.122933984f, -1.f),
    C_Vertex3f(0.319824219f, -0.10521698f, -1.f),
    C_Vertex3f(0.319824219f, -0.11498642f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.299203753f, -0.159090161f, -1.f),
    C_Vertex3f(0.305095315f, -0.152336478f, -1.f),
    C_Vertex3f(0.30578804f, -0.0742721558f, -1.f),
    C_Vertex3f(0.31000495f, -0.145344019f, -1.f),
    C_Vertex3f(0.312806129f, -0.0840682983f, -1.f),
    C_Vertex3f(0.313932657f, -0.138112783f, -1.f),
    C_Vertex3f(0.317484856f, -0.0943832397f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.275623918f, -0.177372813f, -1.f),
    C_Vertex3f(0.28447485f, -0.171881199f, -1.f),
    C_Vertex3f(0.284781456f, -0.0574913025f, -1.f),
    C_Vertex3f(0.292330265f, -0.165605068f, -1.f),
    C_Vertex3f(0.296430588f, -0.064994812f, -1.f),
    C_Vertex3f(0.299203753f, -0.159090161f, -1.f),
    C_Vertex3f(0.30578804f, -0.0742721558f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.247070313f, -0.102478027f, -1.f),
    C_Vertex3f(0.247070313f, -0.109573364f, -1.f),
    C_Vertex3f(0.254935503f, -0.186002493f, -1.f),
    C_Vertex3f(0.245488167f, -0.115414619f, -1.f),
    C_Vertex3f(0.243098021f, -0.189140558f, -1.f),
    C_Vertex3f(0.242323875f, -0.120764732f, -1.f),
    C_Vertex3f(0.237577438f, -0.125623703f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.230265021f, -0.191494107f, -1.f),
    C_Vertex3f(0.243098021f, -0.189140558f, -1.f),
    C_Vertex3f(0.237577438f, -0.125623703f, -1.f),
    C_Vertex3f(0.230264664f, -0.130645752f, -1.f),
    C_Vertex3f(0.218912125f, -0.134902954f, -1.f),
    C_Vertex3f(0.216436505f, -0.19306314f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.165101051f, -0.138977051f, -1.f),
    C_Vertex3f(0.167952538f, -0.193847656f, -1.f),
    C_Vertex3f(0.173505783f, -0.139160156f, -1.f),
    C_Vertex3f(0.184755325f, -0.193847656f, -1.f),
    C_Vertex3f(0.187040329f, -0.139160156f, -1.f),
    C_Vertex3f(0.201612473f, -0.193847656f, -1.f),
    C_Vertex3f(0.204504013f, -0.137741089f, -1.f),
    C_Vertex3f(0.216436505f, -0.19306314f, -1.f),
    C_Vertex3f(0.218912125f, -0.134902954f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.135742188f, -0.137695313f, -1.f),
    C_Vertex3f(0.137622833f, -0.193847656f, -1.f),
    C_Vertex3f(0.142726898f, -0.138061523f, -1.f),
    C_Vertex3f(0.152241707f, -0.193847656f, -1.f),
    C_Vertex3f(0.154841423f, -0.13861084f, -1.f),
    C_Vertex3f(0.167952538f, -0.193847656f, -1.f),
    C_Vertex3f(0.165101051f, -0.138977051f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.137622833f, -0.193847656f, -1.f),
    C_Vertex3f(0.135742188f, -0.137695313f, -1.f),
    C_Vertex3f(0.130859375f, -0.193847656f, -1.f),
    C_Vertex3f(0.270840645f, -0.0517616272f, -1.f),
    C_Vertex3f(0.275623918f, -0.177372813f, -1.f),
    C_Vertex3f(0.284781456f, -0.0574913025f, -1.f),
    C_Vertex3f(0.245346069f, -0.0467147827f, -1.f),
    C_Vertex3f(0.237206459f, -0.0880661011f, -1.f),
    C_Vertex3f(0.240785599f, -0.0909051895f, -1.f),
    C_Vertex3f(0.0369659662f, 0.1885432f, -1.f),
    C_Vertex3f(0.03515625f, 0.249214172f, -1.f),
    C_Vertex3f(0.03515625f, 0.215906143f, -1.f),
    C_Vertex3f(0.10653615f, 0.426183343f, -1.f),
    C_Vertex3f(0.116210938f, 0.244146347f, -1.f),
    C_Vertex3f(0.117643118f, 0.434987903f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_232[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.125192568f, 0.992132485f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.965454936f, -0.260570467f, 0.f),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Normal3f(0.952376783f, -0.304923922f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, -1.f),
    C_Normal3f(0.932039082f, -0.362357676f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, -1.f),
    C_Normal3f(0.901194572f, -0.433414668f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, -1.f),
    C_Normal3f(0.862806261f, -0.505534768f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, -1.f),
    C_Normal3f(0.816684961f, -0.577083826f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, -1.f),
    C_Normal3f(0.763127804f, -0.646247566f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, -1.f),
    C_Normal3f(0.702967107f, -0.711222291f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, -1.f),
    C_Normal3f(0.637524903f, -0.77042979f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, -1.f),
    C_Normal3f(0.565409184f, -0.824810505f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Normal3f(0.48646915f, -0.873697758f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, -1.f),
    C_Normal3f(0.405260772f, -0.9142012f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, -1.f),
    C_Normal3f(0.323988765f, -0.946060896f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Normal3f(0.244618788f, -0.969619274f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, -1.f),
    C_Normal3f(0.168704703f, -0.985666752f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, -1.f),
    C_Normal3f(0.0973226056f, -0.995252907f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, -1.f),
    C_Normal3f(0.029160779f, -0.999574721f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0416765399f, -0.999131143f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Normal3f(-0.124204665f, -0.992256582f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Normal3f(-0.21582f, -0.976433158f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, -1.f),
    C_Normal3f(-0.312524736f, -0.949909627f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Normal3f(-0.411543161f, -0.911390245f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Normal3f(-0.509326518f, -0.860573351f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Normal3f(-0.602056086f, -0.798453748f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Normal3f(-0.686321557f, -0.727298141f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, -1.f),
    C_Normal3f(-0.762221575f, -0.647316158f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, -1.f),
    C_Normal3f(-0.830126584f, -0.557574868f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, -1.f),
    C_Normal3f(-0.886253357f, -0.463200867f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, -1.f),
    C_Normal3f(-0.929815829f, -0.368024975f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, -1.f),
    C_Normal3f(-0.961317182f, -0.275443733f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, -1.f),
    C_Normal3f(-0.982170284f, -0.187993705f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, -1.f),
    C_Normal3f(-0.994233549f, -0.107235931f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Normal3f(-0.999436319f, -0.0335720852f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, -1.f),
    C_Normal3f(-0.999442458f, 0.0333874822f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Normal3f(-0.994488776f, 0.104842685f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, -1.f),
    C_Normal3f(-0.982936144f, 0.183947176f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, -1.f),
    C_Normal3f(-0.962905705f, 0.269838274f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, -1.f),
    C_Normal3f(-0.932518303f, 0.361122698f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, -1.f),
    C_Normal3f(-0.890263319f, 0.455446035f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, -1.f),
    C_Normal3f(-0.835456192f, 0.54955703f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, -1.f),
    C_Normal3f(-0.768646836f, 0.639673412f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, -1.f),
    C_Normal3f(-0.695958257f, 0.718082249f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Normal3f(-0.616909444f, 0.787034094f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Normal3f(-0.527987719f, 0.849251986f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Normal3f(-0.431770295f, 0.901983559f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Normal3f(-0.33168745f, 0.943389297f, 0.f),
    C_Vertex3f(0.172035694f, 0.462487936f, 0.f),
    C_Vertex3f(0.172035694f, 0.462487936f, -1.f),
    C_Normal3f(-0.231436506f, 0.972849965f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_Normal3f(-0.134358883f, 0.990932763f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Normal3f(-0.0433523022f, 0.999059916f, 0.f),
    C_Vertex3f(0.237405777f, 0.474609375f, 0.f),
    C_Vertex3f(0.237405777f, 0.474609375f, -1.f),
    C_Normal3f(0.043157652f, 0.99906826f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_Normal3f(0.135833368f, 0.990731657f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Normal3f(0.233665943f, 0.97231698f, 0.f),
    C_Vertex3f(0.304389715f, 0.468667746f, 0.f),
    C_Vertex3f(0.304389715f, 0.468667746f, -1.f),
    C_Normal3f(0.334365368f, 0.94244355f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_Normal3f(0.434524238f, 0.900660098f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Normal3f(0.530448139f, 0.847717404f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Normal3f(0.618784487f, 0.785560787f, 0.f),
    C_Vertex3f(0.378282785f, 0.433017969f, 0.f),
    C_Vertex3f(0.378282785f, 0.433017969f, -1.f),
    C_Normal3f(0.697084367f, 0.7169891f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_Normal3f(0.769382715f, 0.638788223f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, -1.f),
    C_Normal3f(0.836126029f, 0.548537314f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, -1.f),
    C_Normal3f(0.890802264f, 0.454391062f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, -1.f),
    C_Normal3f(0.932900369f, 0.360134482f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, -1.f),
    C_Normal3f(0.963139415f, 0.269002616f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, -1.f),
    C_Normal3f(0.983052552f, 0.183323756f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, -1.f),
    C_Normal3f(0.994528711f, 0.104463689f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Normal3f(0.999162018f, 0.0409296043f, 0.f),
    C_Vertex3f(0.460449219f, 0.247264862f, 0.f),
    C_Vertex3f(0.460449219f, 0.247264862f, -1.f),
    C_Normal3f(0.999970615f, -0.0076672486f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, -1.f),
    C_Normal3f(0.999690831f, -0.0248668455f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, -1.f),
    C_Normal3f(0.999516904f, -0.0310815405f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.998369277f, 0.0570866764f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Normal3f(0.995165646f, 0.0982105583f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, -1.f),
    C_Normal3f(0.988114417f, 0.153720722f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, -1.f),
    C_Normal3f(0.97403872f, 0.226381436f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, -1.f),
    C_Normal3f(0.95226264f, 0.305279851f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_Normal3f(0.921066403f, 0.389405489f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, -1.f),
    C_Normal3f(0.878937125f, 0.476937532f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, -1.f),
    C_Normal3f(0.824944496f, 0.565213859f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, -1.f),
    C_Normal3f(0.759140372f, 0.650927007f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Normal3f(0.688533545f, 0.725204468f, 0.f),
    C_Vertex3f(0.161923409f, 0.0897507668f, 0.f),
    C_Vertex3f(0.161923409f, 0.0897507668f, -1.f),
    C_Normal3f(0.613223195f, 0.78990972f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_Normal3f(0.528003633f, 0.849242151f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, -1.f),
    C_Normal3f(0.434818447f, 0.900518119f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, -1.f),
    C_Normal3f(0.336550295f, 0.94166553f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_Normal3f(0.236599728f, 0.971607208f, 0.f),
    C_Vertex3f(0.219586134f, 0.0580083132f, 0.f),
    C_Vertex3f(0.219586134f, 0.0580083132f, -1.f),
    C_Normal3f(0.138325498f, 0.990386844f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_Normal3f(0.0396123156f, 0.999215126f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, -1.f),
    C_Normal3f(-0.074584201f, 0.997214675f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, -1.f),
    C_Normal3f(-0.216858819f, 0.976202905f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, -1.f),
    C_Normal3f(-0.374131411f, 0.927375793f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, -1.f),
    C_Normal3f(-0.527194917f, 0.84974438f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, -1.f),
    C_Normal3f(-0.670490623f, 0.741917968f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_Normal3f(-0.787500262f, 0.616314411f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, -1.f),
    C_Normal3f(-0.869012713f, 0.49478972f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, -1.f),
    C_Normal3f(-0.915187478f, 0.40302819f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, -1.f),
    C_Normal3f(-0.9414168f, 0.337245345f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.996035993f, -0.088950254f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Normal3f(-0.992682815f, -0.120750621f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, -1.f),
    C_Normal3f(-0.986461818f, -0.163990855f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, -1.f),
    C_Normal3f(-0.975184321f, -0.221394569f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, -1.f),
    C_Normal3f(-0.958473325f, -0.285182297f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, -1.f),
    C_Normal3f(-0.934688568f, -0.355467528f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_Normal3f(-0.901933372f, -0.431875169f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, -1.f),
    C_Normal3f(-0.85821414f, -0.513291836f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, -1.f),
    C_Normal3f(-0.79895848f, -0.601386249f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_Normal3f(-0.734940171f, -0.678131938f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, -1.f),
    C_Normal3f(-0.663570166f, -0.74811399f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, -1.f),
    C_Normal3f(-0.575440705f, -0.817843437f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, -1.f),
    C_Normal3f(-0.476377249f, -0.87924099f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Normal3f(-0.369776905f, -0.929120481f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, -1.f),
    C_Normal3f(-0.260044605f, -0.965596616f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_Normal3f(-0.151727274f, -0.988422394f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, -1.f),
    C_Normal3f(-0.0498818606f, -0.998755097f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, -1.f),
    C_Normal3f(0.0416725017f, -0.999131382f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Normal3f(0.132727861f, -0.991152465f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, -1.f),
    C_Normal3f(0.226974145f, -0.973900795f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_Normal3f(0.32306844f, -0.946375608f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, -1.f),
    C_Normal3f(0.418056071f, -0.908421159f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, -1.f),
    C_Normal3f(0.50887841f, -0.860838413f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, -1.f),
    C_Normal3f(0.59283942f, -0.80532068f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Normal3f(0.667988896f, -0.744171143f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, -1.f),
    C_Normal3f(0.736174107f, -0.676792264f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_Normal3f(0.80003792f, -0.59994936f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, -1.f),
    C_Normal3f(0.855952084f, -0.517055154f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, -1.f),
    C_Normal3f(0.902495027f, -0.430700183f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, -1.f),
    C_Normal3f(0.939107299f, -0.343623996f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_Normal3f(0.966056883f, -0.258329391f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, -1.f),
    C_Normal3f(0.984245181f, -0.176808938f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, -1.f),
    C_Normal3f(0.993621826f, -0.112763919f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Normal3f(0.997922182f, -0.0644304082f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.312988281f, 0.521972656f, 0.f),
    C_Vertex3f(0.312988281f, 0.521972656f, -1.f),
    C_Vertex3f(0.249511719f, 0.521972656f, -1.f),
    C_Vertex3f(0.249511719f, 0.521972656f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.249511719f, 0.521972656f, 0.f),
    C_Vertex3f(0.249511719f, 0.521972656f, -1.f),
    C_Vertex3f(0.1484375f, 0.64453125f, -1.f),
    C_Vertex3f(0.1484375f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.1484375f, 0.64453125f, 0.f),
    C_Vertex3f(0.1484375f, 0.64453125f, -1.f),
    C_Vertex3f(0.25390625f, 0.64453125f, -1.f),
    C_Vertex3f(0.25390625f, 0.64453125f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.25390625f, 0.64453125f, 0.f),
    C_Vertex3f(0.25390625f, 0.64453125f, -1.f),
    C_Vertex3f(0.312988281f, 0.521972656f, -1.f),
    C_Vertex3f(0.312988281f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219586134f, 0.0580083132f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161923409f, 0.0897507668f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378282785f, 0.433017969f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304389715f, 0.468667746f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237405777f, 0.474609375f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172035694f, 0.462487936f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460449219f, 0.247264862f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25390625f, 0.64453125f, 0.f),
    C_Vertex3f(0.1484375f, 0.64453125f, 0.f),
    C_Vertex3f(0.249511719f, 0.521972656f, 0.f),
    C_Vertex3f(0.312988281f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, -1.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, -1.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, -1.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, -1.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, -1.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, -1.f),
    C_Vertex3f(0.449591041f, 0.113932014f, -1.f),
    C_Vertex3f(0.455421448f, 0.13074398f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, -1.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, -1.f),
    C_Vertex3f(0.361532211f, 0.115768433f, -1.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, -1.f),
    C_Vertex3f(0.372081757f, 0.137390137f, -1.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, -1.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, -1.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, -1.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, -1.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, -1.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, -1.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, -1.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, -1.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, -1.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219586134f, 0.0580083132f, -1.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, -1.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, -1.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, -1.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161923409f, 0.0897507668f, -1.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Vertex3f(0.143544197f, 0.111185431f, -1.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Vertex3f(0.13597393f, 0.123582006f, -1.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Vertex3f(0.129483223f, 0.137098074f, -1.f),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_Vertex3f(0.119740486f, 0.167488694f, -1.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Vertex3f(0.116488457f, 0.184363246f, -1.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Vertex3f(0.114315987f, 0.202357292f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, -1.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, -1.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, -1.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, -1.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, -1.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, -1.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, -1.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, -1.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, -1.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, -1.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, -1.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, -1.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, -1.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, -1.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378282785f, 0.433017969f, -1.f),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_Vertex3f(0.369242191f, 0.322288036f, -1.f),
    C_Vertex3f(0.372357607f, 0.310348272f, -1.f),
    C_Vertex3f(0.374871969f, 0.297490358f, -1.f),
    C_Vertex3f(0.376785278f, 0.283714294f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_Vertex3f(0.356289625f, 0.352598429f, -1.f),
    C_Vertex3f(0.3612082f, 0.343413115f, -1.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Vertex3f(0.322116375f, 0.38877368f, -1.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_Vertex3f(0.332911015f, 0.380220652f, -1.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Vertex3f(0.343008995f, 0.370242119f, -1.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304389715f, 0.468667746f, -1.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_Vertex3f(0.298437119f, 0.401603222f, -1.f),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Vertex3f(0.257693291f, 0.41015625f, -1.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_Vertex3f(0.271971226f, 0.408730745f, -1.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237405777f, 0.474609375f, -1.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_Vertex3f(0.230415225f, 0.409018517f, -1.f),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Vertex3f(0.183755994f, 0.393090248f, -1.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_Vertex3f(0.194611311f, 0.398778915f, -1.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Vertex3f(0.206006289f, 0.403329849f, -1.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172035694f, 0.462487936f, -1.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_Vertex3f(0.163664341f, 0.378299713f, -1.f),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_Vertex3f(0.132908106f, 0.337285995f, -1.f),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Vertex3f(0.139049888f, 0.348690987f, -1.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Vertex3f(0.146223187f, 0.35932827f, -1.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Vertex3f(0.120671868f, 0.298464775f, -1.f),
    C_Vertex3f(0.123719096f, 0.31217289f, -1.f),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460449219f, 0.247264862f, -1.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.460231781f, 0.218906403f, -1.f),
    C_Vertex3f(0.460449219f, 0.22939682f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Vertex3f(0.454919577f, 0.299908876f, -1.f),
    C_Vertex3f(0.449389935f, 0.323812962f, -1.f),
    C_Vertex3f(0.442017078f, 0.346105099f, -1.f),
    C_Vertex3f(0.432801008f, 0.366785288f, -1.f),
    C_Vertex3f(0.421741724f, 0.385853529f, -1.f),
    C_Vertex3f(0.408839226f, 0.403309822f, -1.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25390625f, 0.64453125f, -1.f),
    C_Vertex3f(0.1484375f, 0.64453125f, -1.f),
    C_Vertex3f(0.249511719f, 0.521972656f, -1.f),
    C_Vertex3f(0.312988281f, 0.521972656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_233[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.125192568f, 0.992132485f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.965454936f, -0.260570467f, 0.f),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Normal3f(0.952376783f, -0.304923922f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, -1.f),
    C_Normal3f(0.932039082f, -0.362357676f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, -1.f),
    C_Normal3f(0.901194572f, -0.433414668f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, -1.f),
    C_Normal3f(0.862806261f, -0.505534768f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, -1.f),
    C_Normal3f(0.816684961f, -0.577083826f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, -1.f),
    C_Normal3f(0.763127804f, -0.646247566f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, -1.f),
    C_Normal3f(0.702967107f, -0.711222291f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, -1.f),
    C_Normal3f(0.637524903f, -0.77042979f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, -1.f),
    C_Normal3f(0.565409184f, -0.824810505f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Normal3f(0.48646915f, -0.873697758f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, -1.f),
    C_Normal3f(0.405260772f, -0.9142012f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, -1.f),
    C_Normal3f(0.323988765f, -0.946060896f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Normal3f(0.244618788f, -0.969619274f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, -1.f),
    C_Normal3f(0.168704703f, -0.985666752f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, -1.f),
    C_Normal3f(0.0973226056f, -0.995252907f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, -1.f),
    C_Normal3f(0.029160779f, -0.999574721f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0416765399f, -0.999131143f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Normal3f(-0.124204665f, -0.992256582f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Normal3f(-0.21582f, -0.976433158f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, -1.f),
    C_Normal3f(-0.312524736f, -0.949909627f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Normal3f(-0.411543161f, -0.911390245f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Normal3f(-0.509326518f, -0.860573351f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Normal3f(-0.602056086f, -0.798453748f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Normal3f(-0.686321557f, -0.727298141f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, -1.f),
    C_Normal3f(-0.762221575f, -0.647316158f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, -1.f),
    C_Normal3f(-0.830126584f, -0.557574868f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, -1.f),
    C_Normal3f(-0.886253357f, -0.463200867f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, -1.f),
    C_Normal3f(-0.929815829f, -0.368024975f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, -1.f),
    C_Normal3f(-0.961317182f, -0.275443733f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, -1.f),
    C_Normal3f(-0.982170284f, -0.187993705f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, -1.f),
    C_Normal3f(-0.994233549f, -0.107235931f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Normal3f(-0.999436319f, -0.0335720852f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, -1.f),
    C_Normal3f(-0.999442458f, 0.0333874822f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Normal3f(-0.994488776f, 0.104842685f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, -1.f),
    C_Normal3f(-0.982936144f, 0.183947176f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, -1.f),
    C_Normal3f(-0.962905705f, 0.269838274f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, -1.f),
    C_Normal3f(-0.932518303f, 0.361122698f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, -1.f),
    C_Normal3f(-0.890263319f, 0.455446035f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, -1.f),
    C_Normal3f(-0.835456192f, 0.54955703f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, -1.f),
    C_Normal3f(-0.768646836f, 0.639673412f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, -1.f),
    C_Normal3f(-0.695958257f, 0.718082249f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Normal3f(-0.616909444f, 0.787034094f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Normal3f(-0.527987719f, 0.849251986f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Normal3f(-0.431770295f, 0.901983559f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Normal3f(-0.33168745f, 0.943389297f, 0.f),
    C_Vertex3f(0.172035694f, 0.462487936f, 0.f),
    C_Vertex3f(0.172035694f, 0.462487936f, -1.f),
    C_Normal3f(-0.231436506f, 0.972849965f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_Normal3f(-0.134358883f, 0.990932763f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Normal3f(-0.0433523022f, 0.999059916f, 0.f),
    C_Vertex3f(0.237405777f, 0.474609375f, 0.f),
    C_Vertex3f(0.237405777f, 0.474609375f, -1.f),
    C_Normal3f(0.043157652f, 0.99906826f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_Normal3f(0.135833368f, 0.990731657f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Normal3f(0.233665943f, 0.97231698f, 0.f),
    C_Vertex3f(0.304389715f, 0.468667746f, 0.f),
    C_Vertex3f(0.304389715f, 0.468667746f, -1.f),
    C_Normal3f(0.334365368f, 0.94244355f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_Normal3f(0.434524238f, 0.900660098f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Normal3f(0.530448139f, 0.847717404f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Normal3f(0.618784487f, 0.785560787f, 0.f),
    C_Vertex3f(0.378282785f, 0.433017969f, 0.f),
    C_Vertex3f(0.378282785f, 0.433017969f, -1.f),
    C_Normal3f(0.697084367f, 0.7169891f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_Normal3f(0.769382715f, 0.638788223f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, -1.f),
    C_Normal3f(0.836126029f, 0.548537314f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, -1.f),
    C_Normal3f(0.890802264f, 0.454391062f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, -1.f),
    C_Normal3f(0.932900369f, 0.360134482f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, -1.f),
    C_Normal3f(0.963139415f, 0.269002616f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, -1.f),
    C_Normal3f(0.983052552f, 0.183323756f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, -1.f),
    C_Normal3f(0.994528711f, 0.104463689f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Normal3f(0.999162018f, 0.0409296043f, 0.f),
    C_Vertex3f(0.460449219f, 0.247264862f, 0.f),
    C_Vertex3f(0.460449219f, 0.247264862f, -1.f),
    C_Normal3f(0.999970615f, -0.0076672486f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, -1.f),
    C_Normal3f(0.999690831f, -0.0248668455f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, -1.f),
    C_Normal3f(0.999516904f, -0.0310815405f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.998369277f, 0.0570866764f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Normal3f(0.995165646f, 0.0982105583f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, -1.f),
    C_Normal3f(0.988114417f, 0.153720722f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, -1.f),
    C_Normal3f(0.97403872f, 0.226381436f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, -1.f),
    C_Normal3f(0.95226264f, 0.305279851f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_Normal3f(0.921066403f, 0.389405489f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, -1.f),
    C_Normal3f(0.878937125f, 0.476937532f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, -1.f),
    C_Normal3f(0.824944496f, 0.565213859f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, -1.f),
    C_Normal3f(0.759140372f, 0.650927007f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Normal3f(0.688533545f, 0.725204468f, 0.f),
    C_Vertex3f(0.161923409f, 0.0897507668f, 0.f),
    C_Vertex3f(0.161923409f, 0.0897507668f, -1.f),
    C_Normal3f(0.613223195f, 0.78990972f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_Normal3f(0.528003633f, 0.849242151f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, -1.f),
    C_Normal3f(0.434818447f, 0.900518119f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, -1.f),
    C_Normal3f(0.336550295f, 0.94166553f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_Normal3f(0.236599728f, 0.971607208f, 0.f),
    C_Vertex3f(0.219586134f, 0.0580083132f, 0.f),
    C_Vertex3f(0.219586134f, 0.0580083132f, -1.f),
    C_Normal3f(0.138325498f, 0.990386844f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_Normal3f(0.0396123156f, 0.999215126f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, -1.f),
    C_Normal3f(-0.074584201f, 0.997214675f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, -1.f),
    C_Normal3f(-0.216858819f, 0.976202905f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, -1.f),
    C_Normal3f(-0.374131411f, 0.927375793f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, -1.f),
    C_Normal3f(-0.527194917f, 0.84974438f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, -1.f),
    C_Normal3f(-0.670490623f, 0.741917968f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_Normal3f(-0.787500262f, 0.616314411f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, -1.f),
    C_Normal3f(-0.869012713f, 0.49478972f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, -1.f),
    C_Normal3f(-0.915187478f, 0.40302819f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, -1.f),
    C_Normal3f(-0.9414168f, 0.337245345f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.996035993f, -0.088950254f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Normal3f(-0.992682815f, -0.120750621f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, -1.f),
    C_Normal3f(-0.986461818f, -0.163990855f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, -1.f),
    C_Normal3f(-0.975184321f, -0.221394569f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, -1.f),
    C_Normal3f(-0.958473325f, -0.285182297f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, -1.f),
    C_Normal3f(-0.934688568f, -0.355467528f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_Normal3f(-0.901933372f, -0.431875169f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, -1.f),
    C_Normal3f(-0.85821414f, -0.513291836f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, -1.f),
    C_Normal3f(-0.79895848f, -0.601386249f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_Normal3f(-0.734940171f, -0.678131938f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, -1.f),
    C_Normal3f(-0.663570166f, -0.74811399f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, -1.f),
    C_Normal3f(-0.575440705f, -0.817843437f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, -1.f),
    C_Normal3f(-0.476377249f, -0.87924099f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Normal3f(-0.369776905f, -0.929120481f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, -1.f),
    C_Normal3f(-0.260044605f, -0.965596616f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_Normal3f(-0.151727274f, -0.988422394f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, -1.f),
    C_Normal3f(-0.0498818606f, -0.998755097f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, -1.f),
    C_Normal3f(0.0416725017f, -0.999131382f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Normal3f(0.132727861f, -0.991152465f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, -1.f),
    C_Normal3f(0.226974145f, -0.973900795f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_Normal3f(0.32306844f, -0.946375608f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, -1.f),
    C_Normal3f(0.418056071f, -0.908421159f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, -1.f),
    C_Normal3f(0.50887841f, -0.860838413f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, -1.f),
    C_Normal3f(0.59283942f, -0.80532068f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Normal3f(0.667988896f, -0.744171143f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, -1.f),
    C_Normal3f(0.736174107f, -0.676792264f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_Normal3f(0.80003792f, -0.59994936f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, -1.f),
    C_Normal3f(0.855952084f, -0.517055154f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, -1.f),
    C_Normal3f(0.902495027f, -0.430700183f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, -1.f),
    C_Normal3f(0.939107299f, -0.343623996f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_Normal3f(0.966056883f, -0.258329391f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, -1.f),
    C_Normal3f(0.984245181f, -0.176808938f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, -1.f),
    C_Normal3f(0.993621826f, -0.112763919f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Normal3f(0.997922182f, -0.0644304082f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.184570313f, 0.521972656f, 0.f),
    C_Vertex3f(0.184570313f, 0.521972656f, -1.f),
    C_Vertex3f(0.2421875f, 0.64453125f, -1.f),
    C_Vertex3f(0.2421875f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.2421875f, 0.64453125f, 0.f),
    C_Vertex3f(0.2421875f, 0.64453125f, -1.f),
    C_Vertex3f(0.345703125f, 0.64453125f, -1.f),
    C_Vertex3f(0.345703125f, 0.64453125f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.345703125f, 0.64453125f, 0.f),
    C_Vertex3f(0.345703125f, 0.64453125f, -1.f),
    C_Vertex3f(0.249511719f, 0.521972656f, -1.f),
    C_Vertex3f(0.249511719f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.249511719f, 0.521972656f, 0.f),
    C_Vertex3f(0.249511719f, 0.521972656f, -1.f),
    C_Vertex3f(0.184570313f, 0.521972656f, -1.f),
    C_Vertex3f(0.184570313f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219586134f, 0.0580083132f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161923409f, 0.0897507668f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378282785f, 0.433017969f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304389715f, 0.468667746f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237405777f, 0.474609375f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172035694f, 0.462487936f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460449219f, 0.247264862f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.2421875f, 0.64453125f, 0.f),
    C_Vertex3f(0.184570313f, 0.521972656f, 0.f),
    C_Vertex3f(0.249511719f, 0.521972656f, 0.f),
    C_Vertex3f(0.345703125f, 0.64453125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, -1.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, -1.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, -1.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, -1.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, -1.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, -1.f),
    C_Vertex3f(0.449591041f, 0.113932014f, -1.f),
    C_Vertex3f(0.455421448f, 0.13074398f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, -1.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, -1.f),
    C_Vertex3f(0.361532211f, 0.115768433f, -1.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, -1.f),
    C_Vertex3f(0.372081757f, 0.137390137f, -1.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, -1.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, -1.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, -1.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, -1.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, -1.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, -1.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, -1.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, -1.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, -1.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219586134f, 0.0580083132f, -1.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, -1.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, -1.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, -1.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161923409f, 0.0897507668f, -1.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Vertex3f(0.143544197f, 0.111185431f, -1.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Vertex3f(0.13597393f, 0.123582006f, -1.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Vertex3f(0.129483223f, 0.137098074f, -1.f),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_Vertex3f(0.119740486f, 0.167488694f, -1.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Vertex3f(0.116488457f, 0.184363246f, -1.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Vertex3f(0.114315987f, 0.202357292f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, -1.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, -1.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, -1.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, -1.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, -1.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, -1.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, -1.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, -1.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, -1.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, -1.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, -1.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, -1.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, -1.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, -1.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378282785f, 0.433017969f, -1.f),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_Vertex3f(0.369242191f, 0.322288036f, -1.f),
    C_Vertex3f(0.372357607f, 0.310348272f, -1.f),
    C_Vertex3f(0.374871969f, 0.297490358f, -1.f),
    C_Vertex3f(0.376785278f, 0.283714294f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_Vertex3f(0.356289625f, 0.352598429f, -1.f),
    C_Vertex3f(0.3612082f, 0.343413115f, -1.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Vertex3f(0.322116375f, 0.38877368f, -1.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_Vertex3f(0.332911015f, 0.380220652f, -1.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Vertex3f(0.343008995f, 0.370242119f, -1.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304389715f, 0.468667746f, -1.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_Vertex3f(0.298437119f, 0.401603222f, -1.f),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Vertex3f(0.257693291f, 0.41015625f, -1.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_Vertex3f(0.271971226f, 0.408730745f, -1.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237405777f, 0.474609375f, -1.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_Vertex3f(0.230415225f, 0.409018517f, -1.f),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Vertex3f(0.183755994f, 0.393090248f, -1.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_Vertex3f(0.194611311f, 0.398778915f, -1.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Vertex3f(0.206006289f, 0.403329849f, -1.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172035694f, 0.462487936f, -1.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_Vertex3f(0.163664341f, 0.378299713f, -1.f),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_Vertex3f(0.132908106f, 0.337285995f, -1.f),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Vertex3f(0.139049888f, 0.348690987f, -1.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Vertex3f(0.146223187f, 0.35932827f, -1.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Vertex3f(0.120671868f, 0.298464775f, -1.f),
    C_Vertex3f(0.123719096f, 0.31217289f, -1.f),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460449219f, 0.247264862f, -1.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.460231781f, 0.218906403f, -1.f),
    C_Vertex3f(0.460449219f, 0.22939682f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Vertex3f(0.454919577f, 0.299908876f, -1.f),
    C_Vertex3f(0.449389935f, 0.323812962f, -1.f),
    C_Vertex3f(0.442017078f, 0.346105099f, -1.f),
    C_Vertex3f(0.432801008f, 0.366785288f, -1.f),
    C_Vertex3f(0.421741724f, 0.385853529f, -1.f),
    C_Vertex3f(0.408839226f, 0.403309822f, -1.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.2421875f, 0.64453125f, -1.f),
    C_Vertex3f(0.184570313f, 0.521972656f, -1.f),
    C_Vertex3f(0.249511719f, 0.521972656f, -1.f),
    C_Vertex3f(0.345703125f, 0.64453125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_234[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.125192568f, 0.992132485f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.965454936f, -0.260570467f, 0.f),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Normal3f(0.952376783f, -0.304923922f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, -1.f),
    C_Normal3f(0.932039082f, -0.362357676f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, -1.f),
    C_Normal3f(0.901194572f, -0.433414668f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, -1.f),
    C_Normal3f(0.862806261f, -0.505534768f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, -1.f),
    C_Normal3f(0.816684961f, -0.577083826f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, -1.f),
    C_Normal3f(0.763127804f, -0.646247566f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, -1.f),
    C_Normal3f(0.702967107f, -0.711222291f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, -1.f),
    C_Normal3f(0.637524903f, -0.77042979f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, -1.f),
    C_Normal3f(0.565409184f, -0.824810505f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Normal3f(0.48646915f, -0.873697758f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, -1.f),
    C_Normal3f(0.405260772f, -0.9142012f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, -1.f),
    C_Normal3f(0.323988765f, -0.946060896f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Normal3f(0.244618788f, -0.969619274f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, -1.f),
    C_Normal3f(0.168704703f, -0.985666752f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, -1.f),
    C_Normal3f(0.0973226056f, -0.995252907f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, -1.f),
    C_Normal3f(0.029160779f, -0.999574721f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0416765399f, -0.999131143f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Normal3f(-0.124204665f, -0.992256582f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Normal3f(-0.21582f, -0.976433158f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, -1.f),
    C_Normal3f(-0.312524736f, -0.949909627f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Normal3f(-0.411543161f, -0.911390245f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Normal3f(-0.509326518f, -0.860573351f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Normal3f(-0.602056086f, -0.798453748f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Normal3f(-0.686321557f, -0.727298141f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, -1.f),
    C_Normal3f(-0.762221575f, -0.647316158f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, -1.f),
    C_Normal3f(-0.830126584f, -0.557574868f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, -1.f),
    C_Normal3f(-0.886253357f, -0.463200867f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, -1.f),
    C_Normal3f(-0.929815829f, -0.368024975f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, -1.f),
    C_Normal3f(-0.961317182f, -0.275443733f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, -1.f),
    C_Normal3f(-0.982170284f, -0.187993705f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, -1.f),
    C_Normal3f(-0.994233549f, -0.107235931f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Normal3f(-0.999436319f, -0.0335720852f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, -1.f),
    C_Normal3f(-0.999442458f, 0.0333874822f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Normal3f(-0.994488776f, 0.104842685f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, -1.f),
    C_Normal3f(-0.982936144f, 0.183947176f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, -1.f),
    C_Normal3f(-0.962905705f, 0.269838274f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, -1.f),
    C_Normal3f(-0.932518303f, 0.361122698f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, -1.f),
    C_Normal3f(-0.890263319f, 0.455446035f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, -1.f),
    C_Normal3f(-0.835456192f, 0.54955703f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, -1.f),
    C_Normal3f(-0.768646836f, 0.639673412f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, -1.f),
    C_Normal3f(-0.695958257f, 0.718082249f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Normal3f(-0.616909444f, 0.787034094f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Normal3f(-0.527987719f, 0.849251986f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Normal3f(-0.431770295f, 0.901983559f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Normal3f(-0.33168745f, 0.943389297f, 0.f),
    C_Vertex3f(0.172035694f, 0.462487936f, 0.f),
    C_Vertex3f(0.172035694f, 0.462487936f, -1.f),
    C_Normal3f(-0.231436506f, 0.972849965f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_Normal3f(-0.134358883f, 0.990932763f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Normal3f(-0.0433523022f, 0.999059916f, 0.f),
    C_Vertex3f(0.237405777f, 0.474609375f, 0.f),
    C_Vertex3f(0.237405777f, 0.474609375f, -1.f),
    C_Normal3f(0.043157652f, 0.99906826f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_Normal3f(0.135833368f, 0.990731657f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Normal3f(0.233665943f, 0.97231698f, 0.f),
    C_Vertex3f(0.304389715f, 0.468667746f, 0.f),
    C_Vertex3f(0.304389715f, 0.468667746f, -1.f),
    C_Normal3f(0.334365368f, 0.94244355f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_Normal3f(0.434524238f, 0.900660098f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Normal3f(0.530448139f, 0.847717404f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Normal3f(0.618784487f, 0.785560787f, 0.f),
    C_Vertex3f(0.378282785f, 0.433017969f, 0.f),
    C_Vertex3f(0.378282785f, 0.433017969f, -1.f),
    C_Normal3f(0.697084367f, 0.7169891f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_Normal3f(0.769382715f, 0.638788223f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, -1.f),
    C_Normal3f(0.836126029f, 0.548537314f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, -1.f),
    C_Normal3f(0.890802264f, 0.454391062f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, -1.f),
    C_Normal3f(0.932900369f, 0.360134482f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, -1.f),
    C_Normal3f(0.963139415f, 0.269002616f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, -1.f),
    C_Normal3f(0.983052552f, 0.183323756f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, -1.f),
    C_Normal3f(0.994528711f, 0.104463689f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Normal3f(0.999162018f, 0.0409296043f, 0.f),
    C_Vertex3f(0.460449219f, 0.247264862f, 0.f),
    C_Vertex3f(0.460449219f, 0.247264862f, -1.f),
    C_Normal3f(0.999970615f, -0.0076672486f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, -1.f),
    C_Normal3f(0.999690831f, -0.0248668455f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, -1.f),
    C_Normal3f(0.999516904f, -0.0310815405f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.998369277f, 0.0570866764f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Normal3f(0.995165646f, 0.0982105583f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, -1.f),
    C_Normal3f(0.988114417f, 0.153720722f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, -1.f),
    C_Normal3f(0.97403872f, 0.226381436f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, -1.f),
    C_Normal3f(0.95226264f, 0.305279851f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_Normal3f(0.921066403f, 0.389405489f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, -1.f),
    C_Normal3f(0.878937125f, 0.476937532f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, -1.f),
    C_Normal3f(0.824944496f, 0.565213859f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, -1.f),
    C_Normal3f(0.759140372f, 0.650927007f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Normal3f(0.688533545f, 0.725204468f, 0.f),
    C_Vertex3f(0.161923409f, 0.0897507668f, 0.f),
    C_Vertex3f(0.161923409f, 0.0897507668f, -1.f),
    C_Normal3f(0.613223195f, 0.78990972f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_Normal3f(0.528003633f, 0.849242151f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, -1.f),
    C_Normal3f(0.434818447f, 0.900518119f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, -1.f),
    C_Normal3f(0.336550295f, 0.94166553f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_Normal3f(0.236599728f, 0.971607208f, 0.f),
    C_Vertex3f(0.219586134f, 0.0580083132f, 0.f),
    C_Vertex3f(0.219586134f, 0.0580083132f, -1.f),
    C_Normal3f(0.138325498f, 0.990386844f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_Normal3f(0.0396123156f, 0.999215126f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, -1.f),
    C_Normal3f(-0.074584201f, 0.997214675f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, -1.f),
    C_Normal3f(-0.216858819f, 0.976202905f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, -1.f),
    C_Normal3f(-0.374131411f, 0.927375793f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, -1.f),
    C_Normal3f(-0.527194917f, 0.84974438f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, -1.f),
    C_Normal3f(-0.670490623f, 0.741917968f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_Normal3f(-0.787500262f, 0.616314411f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, -1.f),
    C_Normal3f(-0.869012713f, 0.49478972f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, -1.f),
    C_Normal3f(-0.915187478f, 0.40302819f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, -1.f),
    C_Normal3f(-0.9414168f, 0.337245345f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.996035993f, -0.088950254f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Normal3f(-0.992682815f, -0.120750621f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, -1.f),
    C_Normal3f(-0.986461818f, -0.163990855f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, -1.f),
    C_Normal3f(-0.975184321f, -0.221394569f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, -1.f),
    C_Normal3f(-0.958473325f, -0.285182297f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, -1.f),
    C_Normal3f(-0.934688568f, -0.355467528f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_Normal3f(-0.901933372f, -0.431875169f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, -1.f),
    C_Normal3f(-0.85821414f, -0.513291836f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, -1.f),
    C_Normal3f(-0.79895848f, -0.601386249f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_Normal3f(-0.734940171f, -0.678131938f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, -1.f),
    C_Normal3f(-0.663570166f, -0.74811399f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, -1.f),
    C_Normal3f(-0.575440705f, -0.817843437f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, -1.f),
    C_Normal3f(-0.476377249f, -0.87924099f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Normal3f(-0.369776905f, -0.929120481f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, -1.f),
    C_Normal3f(-0.260044605f, -0.965596616f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_Normal3f(-0.151727274f, -0.988422394f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, -1.f),
    C_Normal3f(-0.0498818606f, -0.998755097f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, -1.f),
    C_Normal3f(0.0416725017f, -0.999131382f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Normal3f(0.132727861f, -0.991152465f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, -1.f),
    C_Normal3f(0.226974145f, -0.973900795f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_Normal3f(0.32306844f, -0.946375608f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, -1.f),
    C_Normal3f(0.418056071f, -0.908421159f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, -1.f),
    C_Normal3f(0.50887841f, -0.860838413f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, -1.f),
    C_Normal3f(0.59283942f, -0.80532068f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Normal3f(0.667988896f, -0.744171143f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, -1.f),
    C_Normal3f(0.736174107f, -0.676792264f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_Normal3f(0.80003792f, -0.59994936f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, -1.f),
    C_Normal3f(0.855952084f, -0.517055154f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, -1.f),
    C_Normal3f(0.902495027f, -0.430700183f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, -1.f),
    C_Normal3f(0.939107299f, -0.343623996f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_Normal3f(0.966056883f, -0.258329391f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, -1.f),
    C_Normal3f(0.984245181f, -0.176808938f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, -1.f),
    C_Normal3f(0.993621826f, -0.112763919f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Normal3f(0.997922182f, -0.0644304082f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.247558594f, 0.596679688f, 0.f),
    C_Vertex3f(0.247558594f, 0.596679688f, -1.f),
    C_Vertex3f(0.198486328f, 0.521972656f, -1.f),
    C_Vertex3f(0.198486328f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.198486328f, 0.521972656f, 0.f),
    C_Vertex3f(0.198486328f, 0.521972656f, -1.f),
    C_Vertex3f(0.109130859f, 0.521972656f, -1.f),
    C_Vertex3f(0.109130859f, 0.521972656f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(0.109130859f, 0.521972656f, 0.f),
    C_Vertex3f(0.109130859f, 0.521972656f, -1.f),
    C_Vertex3f(0.202636719f, 0.64453125f, -1.f),
    C_Vertex3f(0.202636719f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.202636719f, 0.64453125f, 0.f),
    C_Vertex3f(0.202636719f, 0.64453125f, -1.f),
    C_Vertex3f(0.286621094f, 0.64453125f, -1.f),
    C_Vertex3f(0.286621094f, 0.64453125f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.286621094f, 0.64453125f, 0.f),
    C_Vertex3f(0.286621094f, 0.64453125f, -1.f),
    C_Vertex3f(0.385009766f, 0.521972656f, -1.f),
    C_Vertex3f(0.385009766f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.385009766f, 0.521972656f, 0.f),
    C_Vertex3f(0.385009766f, 0.521972656f, -1.f),
    C_Vertex3f(0.295654297f, 0.521972656f, -1.f),
    C_Vertex3f(0.295654297f, 0.521972656f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.295654297f, 0.521972656f, 0.f),
    C_Vertex3f(0.295654297f, 0.521972656f, -1.f),
    C_Vertex3f(0.247558594f, 0.596679688f, -1.f),
    C_Vertex3f(0.247558594f, 0.596679688f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219586134f, 0.0580083132f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161923409f, 0.0897507668f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378282785f, 0.433017969f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304389715f, 0.468667746f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237405777f, 0.474609375f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172035694f, 0.462487936f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460449219f, 0.247264862f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.109130859f, 0.521972656f, 0.f),
    C_Vertex3f(0.198486328f, 0.521972656f, 0.f),
    C_Vertex3f(0.202636719f, 0.64453125f, 0.f),
    C_Vertex3f(0.247558594f, 0.596679688f, 0.f),
    C_Vertex3f(0.286621094f, 0.64453125f, 0.f),
    C_Vertex3f(0.295654297f, 0.521972656f, 0.f),
    C_Vertex3f(0.385009766f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, -1.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, -1.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, -1.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, -1.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, -1.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, -1.f),
    C_Vertex3f(0.449591041f, 0.113932014f, -1.f),
    C_Vertex3f(0.455421448f, 0.13074398f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, -1.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, -1.f),
    C_Vertex3f(0.361532211f, 0.115768433f, -1.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, -1.f),
    C_Vertex3f(0.372081757f, 0.137390137f, -1.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, -1.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, -1.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, -1.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, -1.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, -1.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, -1.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, -1.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, -1.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, -1.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219586134f, 0.0580083132f, -1.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, -1.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, -1.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, -1.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161923409f, 0.0897507668f, -1.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Vertex3f(0.143544197f, 0.111185431f, -1.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Vertex3f(0.13597393f, 0.123582006f, -1.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Vertex3f(0.129483223f, 0.137098074f, -1.f),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_Vertex3f(0.119740486f, 0.167488694f, -1.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Vertex3f(0.116488457f, 0.184363246f, -1.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Vertex3f(0.114315987f, 0.202357292f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, -1.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, -1.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, -1.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, -1.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, -1.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, -1.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, -1.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, -1.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, -1.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, -1.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, -1.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, -1.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, -1.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, -1.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378282785f, 0.433017969f, -1.f),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_Vertex3f(0.369242191f, 0.322288036f, -1.f),
    C_Vertex3f(0.372357607f, 0.310348272f, -1.f),
    C_Vertex3f(0.374871969f, 0.297490358f, -1.f),
    C_Vertex3f(0.376785278f, 0.283714294f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_Vertex3f(0.356289625f, 0.352598429f, -1.f),
    C_Vertex3f(0.3612082f, 0.343413115f, -1.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Vertex3f(0.322116375f, 0.38877368f, -1.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_Vertex3f(0.332911015f, 0.380220652f, -1.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Vertex3f(0.343008995f, 0.370242119f, -1.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304389715f, 0.468667746f, -1.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_Vertex3f(0.298437119f, 0.401603222f, -1.f),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Vertex3f(0.257693291f, 0.41015625f, -1.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_Vertex3f(0.271971226f, 0.408730745f, -1.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237405777f, 0.474609375f, -1.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_Vertex3f(0.230415225f, 0.409018517f, -1.f),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Vertex3f(0.183755994f, 0.393090248f, -1.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_Vertex3f(0.194611311f, 0.398778915f, -1.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Vertex3f(0.206006289f, 0.403329849f, -1.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172035694f, 0.462487936f, -1.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_Vertex3f(0.163664341f, 0.378299713f, -1.f),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_Vertex3f(0.132908106f, 0.337285995f, -1.f),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Vertex3f(0.139049888f, 0.348690987f, -1.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Vertex3f(0.146223187f, 0.35932827f, -1.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Vertex3f(0.120671868f, 0.298464775f, -1.f),
    C_Vertex3f(0.123719096f, 0.31217289f, -1.f),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460449219f, 0.247264862f, -1.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.460231781f, 0.218906403f, -1.f),
    C_Vertex3f(0.460449219f, 0.22939682f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Vertex3f(0.454919577f, 0.299908876f, -1.f),
    C_Vertex3f(0.449389935f, 0.323812962f, -1.f),
    C_Vertex3f(0.442017078f, 0.346105099f, -1.f),
    C_Vertex3f(0.432801008f, 0.366785288f, -1.f),
    C_Vertex3f(0.421741724f, 0.385853529f, -1.f),
    C_Vertex3f(0.408839226f, 0.403309822f, -1.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.109130859f, 0.521972656f, -1.f),
    C_Vertex3f(0.198486328f, 0.521972656f, -1.f),
    C_Vertex3f(0.202636719f, 0.64453125f, -1.f),
    C_Vertex3f(0.247558594f, 0.596679688f, -1.f),
    C_Vertex3f(0.286621094f, 0.64453125f, -1.f),
    C_Vertex3f(0.295654297f, 0.521972656f, -1.f),
    C_Vertex3f(0.385009766f, 0.521972656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_235[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(0.125192568f, 0.992132485f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.965454936f, -0.260570467f, 0.f),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Normal3f(0.952376783f, -0.304923922f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, -1.f),
    C_Normal3f(0.932039082f, -0.362357676f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, -1.f),
    C_Normal3f(0.901194572f, -0.433414668f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, -1.f),
    C_Normal3f(0.862806261f, -0.505534768f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, -1.f),
    C_Normal3f(0.816684961f, -0.577083826f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, -1.f),
    C_Normal3f(0.763127804f, -0.646247566f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, -1.f),
    C_Normal3f(0.702967107f, -0.711222291f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, -1.f),
    C_Normal3f(0.637524903f, -0.77042979f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, -1.f),
    C_Normal3f(0.565409184f, -0.824810505f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Normal3f(0.48646915f, -0.873697758f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, -1.f),
    C_Normal3f(0.405260772f, -0.9142012f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, -1.f),
    C_Normal3f(0.323988765f, -0.946060896f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Normal3f(0.244618788f, -0.969619274f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, -1.f),
    C_Normal3f(0.168704703f, -0.985666752f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, -1.f),
    C_Normal3f(0.0973226056f, -0.995252907f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, -1.f),
    C_Normal3f(0.029160779f, -0.999574721f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0416765399f, -0.999131143f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Normal3f(-0.124204665f, -0.992256582f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Normal3f(-0.21582f, -0.976433158f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, -1.f),
    C_Normal3f(-0.312524736f, -0.949909627f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Normal3f(-0.411543161f, -0.911390245f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Normal3f(-0.509326518f, -0.860573351f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Normal3f(-0.602056086f, -0.798453748f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Normal3f(-0.686321557f, -0.727298141f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, -1.f),
    C_Normal3f(-0.762221575f, -0.647316158f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, -1.f),
    C_Normal3f(-0.830126584f, -0.557574868f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, -1.f),
    C_Normal3f(-0.886253357f, -0.463200867f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, -1.f),
    C_Normal3f(-0.929815829f, -0.368024975f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, -1.f),
    C_Normal3f(-0.961317182f, -0.275443733f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, -1.f),
    C_Normal3f(-0.982170284f, -0.187993705f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, -1.f),
    C_Normal3f(-0.994233549f, -0.107235931f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Normal3f(-0.999436319f, -0.0335720852f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, -1.f),
    C_Normal3f(-0.999442458f, 0.0333874822f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Normal3f(-0.994488776f, 0.104842685f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, -1.f),
    C_Normal3f(-0.982936144f, 0.183947176f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, -1.f),
    C_Normal3f(-0.962905705f, 0.269838274f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, -1.f),
    C_Normal3f(-0.932518303f, 0.361122698f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, -1.f),
    C_Normal3f(-0.890263319f, 0.455446035f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, -1.f),
    C_Normal3f(-0.835456192f, 0.54955703f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, -1.f),
    C_Normal3f(-0.768646836f, 0.639673412f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, -1.f),
    C_Normal3f(-0.695958257f, 0.718082249f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Normal3f(-0.616909444f, 0.787034094f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Normal3f(-0.527987719f, 0.849251986f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Normal3f(-0.431770295f, 0.901983559f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Normal3f(-0.33168745f, 0.943389297f, 0.f),
    C_Vertex3f(0.172035694f, 0.462487936f, 0.f),
    C_Vertex3f(0.172035694f, 0.462487936f, -1.f),
    C_Normal3f(-0.231436506f, 0.972849965f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_Normal3f(-0.134358883f, 0.990932763f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Normal3f(-0.0433523022f, 0.999059916f, 0.f),
    C_Vertex3f(0.237405777f, 0.474609375f, 0.f),
    C_Vertex3f(0.237405777f, 0.474609375f, -1.f),
    C_Normal3f(0.043157652f, 0.99906826f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_Normal3f(0.135833368f, 0.990731657f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Normal3f(0.233665943f, 0.97231698f, 0.f),
    C_Vertex3f(0.304389715f, 0.468667746f, 0.f),
    C_Vertex3f(0.304389715f, 0.468667746f, -1.f),
    C_Normal3f(0.334365368f, 0.94244355f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_Normal3f(0.434524238f, 0.900660098f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Normal3f(0.530448139f, 0.847717404f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Normal3f(0.618784487f, 0.785560787f, 0.f),
    C_Vertex3f(0.378282785f, 0.433017969f, 0.f),
    C_Vertex3f(0.378282785f, 0.433017969f, -1.f),
    C_Normal3f(0.697084367f, 0.7169891f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_Normal3f(0.769382715f, 0.638788223f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, -1.f),
    C_Normal3f(0.836126029f, 0.548537314f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, -1.f),
    C_Normal3f(0.890802264f, 0.454391062f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, -1.f),
    C_Normal3f(0.932900369f, 0.360134482f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, -1.f),
    C_Normal3f(0.963139415f, 0.269002616f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, -1.f),
    C_Normal3f(0.983052552f, 0.183323756f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, -1.f),
    C_Normal3f(0.994528711f, 0.104463689f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Normal3f(0.999162018f, 0.0409296043f, 0.f),
    C_Vertex3f(0.460449219f, 0.247264862f, 0.f),
    C_Vertex3f(0.460449219f, 0.247264862f, -1.f),
    C_Normal3f(0.999970615f, -0.0076672486f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, -1.f),
    C_Normal3f(0.999690831f, -0.0248668455f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, -1.f),
    C_Normal3f(0.999516904f, -0.0310815405f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.998369277f, 0.0570866764f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Normal3f(0.995165646f, 0.0982105583f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, -1.f),
    C_Normal3f(0.988114417f, 0.153720722f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, -1.f),
    C_Normal3f(0.97403872f, 0.226381436f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, -1.f),
    C_Normal3f(0.95226264f, 0.305279851f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_Normal3f(0.921066403f, 0.389405489f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, -1.f),
    C_Normal3f(0.878937125f, 0.476937532f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, -1.f),
    C_Normal3f(0.824944496f, 0.565213859f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, -1.f),
    C_Normal3f(0.759140372f, 0.650927007f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Normal3f(0.688533545f, 0.725204468f, 0.f),
    C_Vertex3f(0.161923409f, 0.0897507668f, 0.f),
    C_Vertex3f(0.161923409f, 0.0897507668f, -1.f),
    C_Normal3f(0.613223195f, 0.78990972f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_Normal3f(0.528003633f, 0.849242151f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, -1.f),
    C_Normal3f(0.434818447f, 0.900518119f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, -1.f),
    C_Normal3f(0.336550295f, 0.94166553f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_Normal3f(0.236599728f, 0.971607208f, 0.f),
    C_Vertex3f(0.219586134f, 0.0580083132f, 0.f),
    C_Vertex3f(0.219586134f, 0.0580083132f, -1.f),
    C_Normal3f(0.138325498f, 0.990386844f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_Normal3f(0.0396123156f, 0.999215126f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, -1.f),
    C_Normal3f(-0.074584201f, 0.997214675f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, -1.f),
    C_Normal3f(-0.216858819f, 0.976202905f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, -1.f),
    C_Normal3f(-0.374131411f, 0.927375793f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, -1.f),
    C_Normal3f(-0.527194917f, 0.84974438f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, -1.f),
    C_Normal3f(-0.670490623f, 0.741917968f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_Normal3f(-0.787500262f, 0.616314411f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, -1.f),
    C_Normal3f(-0.869012713f, 0.49478972f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, -1.f),
    C_Normal3f(-0.915187478f, 0.40302819f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, -1.f),
    C_Normal3f(-0.9414168f, 0.337245345f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.996035993f, -0.088950254f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Normal3f(-0.992682815f, -0.120750621f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, -1.f),
    C_Normal3f(-0.986461818f, -0.163990855f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, -1.f),
    C_Normal3f(-0.975184321f, -0.221394569f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, -1.f),
    C_Normal3f(-0.958473325f, -0.285182297f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, -1.f),
    C_Normal3f(-0.934688568f, -0.355467528f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_Normal3f(-0.901933372f, -0.431875169f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, -1.f),
    C_Normal3f(-0.85821414f, -0.513291836f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, -1.f),
    C_Normal3f(-0.79895848f, -0.601386249f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_Normal3f(-0.734940171f, -0.678131938f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, -1.f),
    C_Normal3f(-0.663570166f, -0.74811399f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, -1.f),
    C_Normal3f(-0.575440705f, -0.817843437f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, -1.f),
    C_Normal3f(-0.476377249f, -0.87924099f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Normal3f(-0.369776905f, -0.929120481f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, -1.f),
    C_Normal3f(-0.260044605f, -0.965596616f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_Normal3f(-0.151727274f, -0.988422394f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, -1.f),
    C_Normal3f(-0.0498818606f, -0.998755097f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, -1.f),
    C_Normal3f(0.0416725017f, -0.999131382f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Normal3f(0.132727861f, -0.991152465f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, -1.f),
    C_Normal3f(0.226974145f, -0.973900795f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_Normal3f(0.32306844f, -0.946375608f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, -1.f),
    C_Normal3f(0.418056071f, -0.908421159f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, -1.f),
    C_Normal3f(0.50887841f, -0.860838413f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, -1.f),
    C_Normal3f(0.59283942f, -0.80532068f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Normal3f(0.667988896f, -0.744171143f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, -1.f),
    C_Normal3f(0.736174107f, -0.676792264f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_Normal3f(0.80003792f, -0.59994936f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, -1.f),
    C_Normal3f(0.855952084f, -0.517055154f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, -1.f),
    C_Normal3f(0.902495027f, -0.430700183f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, -1.f),
    C_Normal3f(0.939107299f, -0.343623996f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_Normal3f(0.966056883f, -0.258329391f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, -1.f),
    C_Normal3f(0.984245181f, -0.176808938f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, -1.f),
    C_Normal3f(0.993621826f, -0.112763919f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Normal3f(0.997922182f, -0.0644304082f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.129882813f, 0.521972656f, 0.f),
    C_Vertex3f(0.129882813f, 0.521972656f, -1.f),
    C_Vertex3f(0.129882813f, 0.611328125f, -1.f),
    C_Vertex3f(0.129882813f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.129882813f, 0.611328125f, 0.f),
    C_Vertex3f(0.129882813f, 0.611328125f, -1.f),
    C_Vertex3f(0.211914063f, 0.611328125f, -1.f),
    C_Vertex3f(0.211914063f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.211914063f, 0.611328125f, 0.f),
    C_Vertex3f(0.211914063f, 0.611328125f, -1.f),
    C_Vertex3f(0.211914063f, 0.521972656f, -1.f),
    C_Vertex3f(0.211914063f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.211914063f, 0.521972656f, 0.f),
    C_Vertex3f(0.211914063f, 0.521972656f, -1.f),
    C_Vertex3f(0.129882813f, 0.521972656f, -1.f),
    C_Vertex3f(0.129882813f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.29296875f, 0.521972656f, 0.f),
    C_Vertex3f(0.29296875f, 0.521972656f, -1.f),
    C_Vertex3f(0.29296875f, 0.611328125f, -1.f),
    C_Vertex3f(0.29296875f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.29296875f, 0.611328125f, 0.f),
    C_Vertex3f(0.29296875f, 0.611328125f, -1.f),
    C_Vertex3f(0.375f, 0.611328125f, -1.f),
    C_Vertex3f(0.375f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.375f, 0.611328125f, 0.f),
    C_Vertex3f(0.375f, 0.611328125f, -1.f),
    C_Vertex3f(0.375f, 0.521972656f, -1.f),
    C_Vertex3f(0.375f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.375f, 0.521972656f, 0.f),
    C_Vertex3f(0.375f, 0.521972656f, -1.f),
    C_Vertex3f(0.29296875f, 0.521972656f, -1.f),
    C_Vertex3f(0.29296875f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.457824707f, 0.139648438f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, 0.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, 0.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, 0.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, 0.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, 0.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, 0.f),
    C_Vertex3f(0.449591041f, 0.113932014f, 0.f),
    C_Vertex3f(0.455421448f, 0.13074398f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, 0.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, 0.f),
    C_Vertex3f(0.361532211f, 0.115768433f, 0.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, 0.f),
    C_Vertex3f(0.372081757f, 0.137390137f, 0.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, 0.f),
    C_Vertex3f(0.376564026f, 0.149902344f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, 0.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, 0.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, 0.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, 0.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, 0.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, 0.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, 0.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, 0.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, 0.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, 0.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219586134f, 0.0580083132f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, 0.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, 0.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, 0.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, 0.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, 0.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161923409f, 0.0897507668f, 0.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, 0.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152194023f, 0.0999083519f, 0.f),
    C_Vertex3f(0.143544197f, 0.111185431f, 0.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, 0.f),
    C_Vertex3f(0.13597393f, 0.123582006f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.129483223f, 0.137098074f, 0.f),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.124072075f, 0.151733637f, 0.f),
    C_Vertex3f(0.119740486f, 0.167488694f, 0.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, 0.f),
    C_Vertex3f(0.116488457f, 0.184363246f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.114315987f, 0.202357292f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, 0.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, 0.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, 0.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, 0.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, 0.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, 0.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, 0.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, 0.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, 0.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, 0.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, 0.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, 0.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, 0.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378282785f, 0.433017969f, 0.f),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_Vertex3f(0.369242191f, 0.322288036f, 0.f),
    C_Vertex3f(0.372357607f, 0.310348272f, 0.f),
    C_Vertex3f(0.374871969f, 0.297490358f, 0.f),
    C_Vertex3f(0.376785278f, 0.283714294f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361407042f, 0.444901228f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_Vertex3f(0.356289625f, 0.352598429f, 0.f),
    C_Vertex3f(0.3612082f, 0.343413115f, 0.f),
    C_Vertex3f(0.365525723f, 0.33330965f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.322116375f, 0.38877368f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_Vertex3f(0.332911015f, 0.380220652f, 0.f),
    C_Vertex3f(0.343466282f, 0.454803944f, 0.f),
    C_Vertex3f(0.343008995f, 0.370242119f, 0.f),
    C_Vertex3f(0.350769997f, 0.360865593f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304389715f, 0.468667746f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_Vertex3f(0.298437119f, 0.401603222f, 0.f),
    C_Vertex3f(0.310625076f, 0.395901203f, 0.f),
    C_Vertex3f(0.324460506f, 0.462726116f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.257693291f, 0.41015625f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_Vertex3f(0.271971226f, 0.408730745f, 0.f),
    C_Vertex3f(0.283253908f, 0.472628832f, 0.f),
    C_Vertex3f(0.285552502f, 0.405879736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237405777f, 0.474609375f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_Vertex3f(0.230415225f, 0.409018517f, 0.f),
    C_Vertex3f(0.243429184f, 0.41015625f, 0.f),
    C_Vertex3f(0.261053085f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.183755994f, 0.393090248f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_Vertex3f(0.194611311f, 0.398778915f, 0.f),
    C_Vertex3f(0.214496374f, 0.472589135f, 0.f),
    C_Vertex3f(0.206006289f, 0.403329849f, 0.f),
    C_Vertex3f(0.217940927f, 0.40674305f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172035694f, 0.462487936f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_Vertex3f(0.163664341f, 0.378299713f, 0.f),
    C_Vertex3f(0.173440337f, 0.386263847f, 0.f),
    C_Vertex3f(0.192706347f, 0.468548656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_Vertex3f(0.132908106f, 0.337285995f, 0.f),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.139049888f, 0.348690987f, 0.f),
    C_Vertex3f(0.152484417f, 0.454406977f, 0.f),
    C_Vertex3f(0.146223187f, 0.35932827f, 0.f),
    C_Vertex3f(0.154428005f, 0.369197845f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052515f, 0.444305778f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.120671868f, 0.298464775f, 0.f),
    C_Vertex3f(0.123719096f, 0.31217289f, 0.f),
    C_Vertex3f(0.127797842f, 0.325113297f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118164063f, 0.276367188f, 0.f),
    C_Vertex3f(0.118656158f, 0.283988953f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.377441406f, 0.276367188f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460449219f, 0.247264862f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.460231781f, 0.218906403f, 0.f),
    C_Vertex3f(0.460449219f, 0.22939682f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460014343f, 0.211914063f, 0.f),
    C_Vertex3f(0.458606005f, 0.274392843f, 0.f),
    C_Vertex3f(0.454919577f, 0.299908876f, 0.f),
    C_Vertex3f(0.449389935f, 0.323812962f, 0.f),
    C_Vertex3f(0.442017078f, 0.346105099f, 0.f),
    C_Vertex3f(0.432801008f, 0.366785288f, 0.f),
    C_Vertex3f(0.421741724f, 0.385853529f, 0.f),
    C_Vertex3f(0.408839226f, 0.403309822f, 0.f),
    C_Vertex3f(0.394093513f, 0.419154167f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.129882813f, 0.611328125f, 0.f),
    C_Vertex3f(0.129882813f, 0.521972656f, 0.f),
    C_Vertex3f(0.211914063f, 0.521972656f, 0.f),
    C_Vertex3f(0.211914063f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.29296875f, 0.611328125f, 0.f),
    C_Vertex3f(0.29296875f, 0.521972656f, 0.f),
    C_Vertex3f(0.375f, 0.521972656f, 0.f),
    C_Vertex3f(0.375f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.113769531f, 0.211914063f, 0.f),
    C_Vertex3f(0.116739988f, 0.432184339f, 0.f),
    C_Vertex3f(0.100546837f, 0.41804266f, 0.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, 0.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, 0.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.457824707f, 0.139648438f, -1.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Vertex3f(0.393106937f, 0.0339961052f, -1.f),
    C_Vertex3f(0.405080676f, 0.0448263884f, -1.f),
    C_Vertex3f(0.416030526f, 0.0566536188f, -1.f),
    C_Vertex3f(0.425956488f, 0.0694777966f, -1.f),
    C_Vertex3f(0.434858561f, 0.0832989216f, -1.f),
    C_Vertex3f(0.442736745f, 0.0981169939f, -1.f),
    C_Vertex3f(0.449591041f, 0.113932014f, -1.f),
    C_Vertex3f(0.455421448f, 0.13074398f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Vertex3f(0.349397659f, 0.0975494385f, -1.f),
    C_Vertex3f(0.351410747f, 0.00818359852f, -1.f),
    C_Vertex3f(0.361532211f, 0.115768433f, -1.f),
    C_Vertex3f(0.36621058f, 0.0155586004f, -1.f),
    C_Vertex3f(0.372081757f, 0.137390137f, -1.f),
    C_Vertex3f(0.38010931f, 0.0241627693f, -1.f),
    C_Vertex3f(0.376564026f, 0.149902344f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Vertex3f(0.246858597f, 0.0541992188f, -1.f),
    C_Vertex3f(0.263895035f, -0.0102539063f, -1.f),
    C_Vertex3f(0.264949799f, 0.0541992188f, -1.f),
    C_Vertex3f(0.283200383f, -0.00902473927f, -1.f),
    C_Vertex3f(0.285009384f, 0.0570526123f, -1.f),
    C_Vertex3f(0.301604629f, -0.0065664053f, -1.f),
    C_Vertex3f(0.303483963f, 0.0627593994f, -1.f),
    C_Vertex3f(0.319107771f, -0.00287890434f, -1.f),
    C_Vertex3f(0.320373535f, 0.0713195801f, -1.f),
    C_Vertex3f(0.33570981f, 0.0020377636f, -1.f),
    C_Vertex3f(0.335678101f, 0.0827331543f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219586134f, 0.0580083132f, -1.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Vertex3f(0.241066933f, -0.0102539063f, -1.f),
    C_Vertex3f(0.232921839f, 0.0554689169f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Vertex3f(0.183185339f, 0.0732446909f, -1.f),
    C_Vertex3f(0.19411993f, -0.00437736511f, -1.f),
    C_Vertex3f(0.194717884f, 0.0668962002f, -1.f),
    C_Vertex3f(0.216934681f, -0.0082950592f, -1.f),
    C_Vertex3f(0.206851482f, 0.0618174076f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161923409f, 0.0897507668f, -1.f),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Vertex3f(0.172622681f, 0.00149917603f, -1.f),
    C_Vertex3f(0.172253847f, 0.0808628798f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.152194023f, 0.0999083519f, -1.f),
    C_Vertex3f(0.143544197f, 0.111185431f, -1.f),
    C_Vertex3f(0.152442932f, 0.00933456421f, -1.f),
    C_Vertex3f(0.13597393f, 0.123582006f, -1.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Vertex3f(0.129483223f, 0.137098074f, -1.f),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.124072075f, 0.151733637f, -1.f),
    C_Vertex3f(0.119740486f, 0.167488694f, -1.f),
    C_Vertex3f(0.133580685f, 0.0191287994f, -1.f),
    C_Vertex3f(0.116488457f, 0.184363246f, -1.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Vertex3f(0.114315987f, 0.202357292f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Vertex3f(0.0345990658f, 0.270510077f, -1.f),
    C_Vertex3f(0.0383059978f, 0.1622684f, -1.f),
    C_Vertex3f(0.0383675098f, 0.296500325f, -1.f),
    C_Vertex3f(0.0438971519f, 0.138714314f, -1.f),
    C_Vertex3f(0.0440201759f, 0.320852518f, -1.f),
    C_Vertex3f(0.0513520241f, 0.116736889f, -1.f),
    C_Vertex3f(0.0515570641f, 0.343566656f, -1.f),
    C_Vertex3f(0.0606706142f, 0.0963361263f, -1.f),
    C_Vertex3f(0.0609781742f, 0.364642739f, -1.f),
    C_Vertex3f(0.0718529224f, 0.0775120258f, -1.f),
    C_Vertex3f(0.0722835064f, 0.384080768f, -1.f),
    C_Vertex3f(0.0848989487f, 0.0602645874f, -1.f),
    C_Vertex3f(0.0854730606f, 0.401880741f, -1.f),
    C_Vertex3f(0.0998086929f, 0.044593811f, -1.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Vertex3f(0.116035938f, 0.0308818817f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.378282785f, 0.433017969f, -1.f),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_Vertex3f(0.369242191f, 0.322288036f, -1.f),
    C_Vertex3f(0.372357607f, 0.310348272f, -1.f),
    C_Vertex3f(0.374871969f, 0.297490358f, -1.f),
    C_Vertex3f(0.376785278f, 0.283714294f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.361407042f, 0.444901228f, -1.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_Vertex3f(0.356289625f, 0.352598429f, -1.f),
    C_Vertex3f(0.3612082f, 0.343413115f, -1.f),
    C_Vertex3f(0.365525723f, 0.33330965f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Vertex3f(0.322116375f, 0.38877368f, -1.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_Vertex3f(0.332911015f, 0.380220652f, -1.f),
    C_Vertex3f(0.343466282f, 0.454803944f, -1.f),
    C_Vertex3f(0.343008995f, 0.370242119f, -1.f),
    C_Vertex3f(0.350769997f, 0.360865593f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.304389715f, 0.468667746f, -1.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_Vertex3f(0.298437119f, 0.401603222f, -1.f),
    C_Vertex3f(0.310625076f, 0.395901203f, -1.f),
    C_Vertex3f(0.324460506f, 0.462726116f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Vertex3f(0.257693291f, 0.41015625f, -1.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_Vertex3f(0.271971226f, 0.408730745f, -1.f),
    C_Vertex3f(0.283253908f, 0.472628832f, -1.f),
    C_Vertex3f(0.285552502f, 0.405879736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.237405777f, 0.474609375f, -1.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_Vertex3f(0.230415225f, 0.409018517f, -1.f),
    C_Vertex3f(0.243429184f, 0.41015625f, -1.f),
    C_Vertex3f(0.261053085f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Vertex3f(0.183755994f, 0.393090248f, -1.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_Vertex3f(0.194611311f, 0.398778915f, -1.f),
    C_Vertex3f(0.214496374f, 0.472589135f, -1.f),
    C_Vertex3f(0.206006289f, 0.403329849f, -1.f),
    C_Vertex3f(0.217940927f, 0.40674305f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.172035694f, 0.462487936f, -1.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_Vertex3f(0.163664341f, 0.378299713f, -1.f),
    C_Vertex3f(0.173440337f, 0.386263847f, -1.f),
    C_Vertex3f(0.192706347f, 0.468548656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_Vertex3f(0.132908106f, 0.337285995f, -1.f),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Vertex3f(0.139049888f, 0.348690987f, -1.f),
    C_Vertex3f(0.152484417f, 0.454406977f, -1.f),
    C_Vertex3f(0.146223187f, 0.35932827f, -1.f),
    C_Vertex3f(0.154428005f, 0.369197845f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.134052515f, 0.444305778f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Vertex3f(0.120671868f, 0.298464775f, -1.f),
    C_Vertex3f(0.123719096f, 0.31217289f, -1.f),
    C_Vertex3f(0.127797842f, 0.325113297f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118164063f, 0.276367188f, -1.f),
    C_Vertex3f(0.118656158f, 0.283988953f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.377441406f, 0.276367188f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460449219f, 0.247264862f, -1.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.460231781f, 0.218906403f, -1.f),
    C_Vertex3f(0.460449219f, 0.22939682f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.460014343f, 0.211914063f, -1.f),
    C_Vertex3f(0.458606005f, 0.274392843f, -1.f),
    C_Vertex3f(0.454919577f, 0.299908876f, -1.f),
    C_Vertex3f(0.449389935f, 0.323812962f, -1.f),
    C_Vertex3f(0.442017078f, 0.346105099f, -1.f),
    C_Vertex3f(0.432801008f, 0.366785288f, -1.f),
    C_Vertex3f(0.421741724f, 0.385853529f, -1.f),
    C_Vertex3f(0.408839226f, 0.403309822f, -1.f),
    C_Vertex3f(0.394093513f, 0.419154167f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.129882813f, 0.611328125f, -1.f),
    C_Vertex3f(0.129882813f, 0.521972656f, -1.f),
    C_Vertex3f(0.211914063f, 0.521972656f, -1.f),
    C_Vertex3f(0.211914063f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.29296875f, 0.611328125f, -1.f),
    C_Vertex3f(0.29296875f, 0.521972656f, -1.f),
    C_Vertex3f(0.375f, 0.521972656f, -1.f),
    C_Vertex3f(0.375f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.113769531f, 0.211914063f, -1.f),
    C_Vertex3f(0.116739988f, 0.432184339f, -1.f),
    C_Vertex3f(0.100546837f, 0.41804266f, -1.f),
    C_Vertex3f(0.0345785618f, 0.187399149f, -1.f),
    C_Vertex3f(0.0327148438f, 0.242881775f, -1.f),
    C_Vertex3f(0.0327148438f, 0.21410656f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_236[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.188964844f, 0.521972656f, 0.f),
    C_Vertex3f(0.188964844f, 0.521972656f, -1.f),
    C_Vertex3f(0.125488281f, 0.521972656f, -1.f),
    C_Vertex3f(0.125488281f, 0.521972656f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.125488281f, 0.521972656f, 0.f),
    C_Vertex3f(0.125488281f, 0.521972656f, -1.f),
    C_Vertex3f(0.0244140625f, 0.64453125f, -1.f),
    C_Vertex3f(0.0244140625f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0244140625f, 0.64453125f, 0.f),
    C_Vertex3f(0.0244140625f, 0.64453125f, -1.f),
    C_Vertex3f(0.129882813f, 0.64453125f, -1.f),
    C_Vertex3f(0.129882813f, 0.64453125f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.129882813f, 0.64453125f, 0.f),
    C_Vertex3f(0.129882813f, 0.64453125f, -1.f),
    C_Vertex3f(0.188964844f, 0.521972656f, -1.f),
    C_Vertex3f(0.188964844f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.129882813f, 0.64453125f, 0.f),
    C_Vertex3f(0.0244140625f, 0.64453125f, 0.f),
    C_Vertex3f(0.125488281f, 0.521972656f, 0.f),
    C_Vertex3f(0.188964844f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.129882813f, 0.64453125f, -1.f),
    C_Vertex3f(0.0244140625f, 0.64453125f, -1.f),
    C_Vertex3f(0.125488281f, 0.521972656f, -1.f),
    C_Vertex3f(0.188964844f, 0.521972656f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_237[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.0610351563f, 0.521972656f, 0.f),
    C_Vertex3f(0.0610351563f, 0.521972656f, -1.f),
    C_Vertex3f(0.118652344f, 0.64453125f, -1.f),
    C_Vertex3f(0.118652344f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.118652344f, 0.64453125f, 0.f),
    C_Vertex3f(0.118652344f, 0.64453125f, -1.f),
    C_Vertex3f(0.222167969f, 0.64453125f, -1.f),
    C_Vertex3f(0.222167969f, 0.64453125f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.222167969f, 0.64453125f, 0.f),
    C_Vertex3f(0.222167969f, 0.64453125f, -1.f),
    C_Vertex3f(0.125976563f, 0.521972656f, -1.f),
    C_Vertex3f(0.125976563f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.125976563f, 0.521972656f, 0.f),
    C_Vertex3f(0.125976563f, 0.521972656f, -1.f),
    C_Vertex3f(0.0610351563f, 0.521972656f, -1.f),
    C_Vertex3f(0.0610351563f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118652344f, 0.64453125f, 0.f),
    C_Vertex3f(0.0610351563f, 0.521972656f, 0.f),
    C_Vertex3f(0.125976563f, 0.521972656f, 0.f),
    C_Vertex3f(0.222167969f, 0.64453125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.118652344f, 0.64453125f, -1.f),
    C_Vertex3f(0.0610351563f, 0.521972656f, -1.f),
    C_Vertex3f(0.125976563f, 0.521972656f, -1.f),
    C_Vertex3f(0.222167969f, 0.64453125f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_238[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.130371094f, 0.596679688f, 0.f),
    C_Vertex3f(0.130371094f, 0.596679688f, -1.f),
    C_Vertex3f(0.0812988281f, 0.521972656f, -1.f),
    C_Vertex3f(0.0812988281f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.0812988281f, 0.521972656f, 0.f),
    C_Vertex3f(0.0812988281f, 0.521972656f, -1.f),
    C_Vertex3f(-0.00805664063f, 0.521972656f, -1.f),
    C_Vertex3f(-0.00805664063f, 0.521972656f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(-0.00805664063f, 0.521972656f, 0.f),
    C_Vertex3f(-0.00805664063f, 0.521972656f, -1.f),
    C_Vertex3f(0.0854492188f, 0.64453125f, -1.f),
    C_Vertex3f(0.0854492188f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0854492188f, 0.64453125f, 0.f),
    C_Vertex3f(0.0854492188f, 0.64453125f, -1.f),
    C_Vertex3f(0.169433594f, 0.64453125f, -1.f),
    C_Vertex3f(0.169433594f, 0.64453125f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.169433594f, 0.64453125f, 0.f),
    C_Vertex3f(0.169433594f, 0.64453125f, -1.f),
    C_Vertex3f(0.267822266f, 0.521972656f, -1.f),
    C_Vertex3f(0.267822266f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.267822266f, 0.521972656f, 0.f),
    C_Vertex3f(0.267822266f, 0.521972656f, -1.f),
    C_Vertex3f(0.178466797f, 0.521972656f, -1.f),
    C_Vertex3f(0.178466797f, 0.521972656f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.178466797f, 0.521972656f, 0.f),
    C_Vertex3f(0.178466797f, 0.521972656f, -1.f),
    C_Vertex3f(0.130371094f, 0.596679688f, -1.f),
    C_Vertex3f(0.130371094f, 0.596679688f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(-0.00805664063f, 0.521972656f, 0.f),
    C_Vertex3f(0.0812988281f, 0.521972656f, 0.f),
    C_Vertex3f(0.0854492188f, 0.64453125f, 0.f),
    C_Vertex3f(0.130371094f, 0.596679688f, 0.f),
    C_Vertex3f(0.169433594f, 0.64453125f, 0.f),
    C_Vertex3f(0.178466797f, 0.521972656f, 0.f),
    C_Vertex3f(0.267822266f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(-0.00805664063f, 0.521972656f, -1.f),
    C_Vertex3f(0.0812988281f, 0.521972656f, -1.f),
    C_Vertex3f(0.0854492188f, 0.64453125f, -1.f),
    C_Vertex3f(0.130371094f, 0.596679688f, -1.f),
    C_Vertex3f(0.169433594f, 0.64453125f, -1.f),
    C_Vertex3f(0.178466797f, 0.521972656f, -1.f),
    C_Vertex3f(0.267822266f, 0.521972656f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_239[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.001953125f, 0.521972656f, 0.f),
    C_Vertex3f(0.001953125f, 0.521972656f, -1.f),
    C_Vertex3f(0.001953125f, 0.611328125f, -1.f),
    C_Vertex3f(0.001953125f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.001953125f, 0.611328125f, 0.f),
    C_Vertex3f(0.001953125f, 0.611328125f, -1.f),
    C_Vertex3f(0.083984375f, 0.611328125f, -1.f),
    C_Vertex3f(0.083984375f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.083984375f, 0.611328125f, 0.f),
    C_Vertex3f(0.083984375f, 0.611328125f, -1.f),
    C_Vertex3f(0.083984375f, 0.521972656f, -1.f),
    C_Vertex3f(0.083984375f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.083984375f, 0.521972656f, 0.f),
    C_Vertex3f(0.083984375f, 0.521972656f, -1.f),
    C_Vertex3f(0.001953125f, 0.521972656f, -1.f),
    C_Vertex3f(0.001953125f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.521972656f, 0.f),
    C_Vertex3f(0.165039063f, 0.521972656f, -1.f),
    C_Vertex3f(0.165039063f, 0.611328125f, -1.f),
    C_Vertex3f(0.165039063f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.165039063f, 0.611328125f, 0.f),
    C_Vertex3f(0.165039063f, 0.611328125f, -1.f),
    C_Vertex3f(0.247070313f, 0.611328125f, -1.f),
    C_Vertex3f(0.247070313f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.247070313f, 0.611328125f, 0.f),
    C_Vertex3f(0.247070313f, 0.611328125f, -1.f),
    C_Vertex3f(0.247070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.247070313f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.247070313f, 0.521972656f, 0.f),
    C_Vertex3f(0.247070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.165039063f, 0.521972656f, -1.f),
    C_Vertex3f(0.165039063f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0864257813f, 0.464355469f, 0.f),
    C_Vertex3f(0.0864257813f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.f, 0.f),
    C_Vertex3f(0.165039063f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.165039063f, 0.611328125f, 0.f),
    C_Vertex3f(0.165039063f, 0.521972656f, 0.f),
    C_Vertex3f(0.247070313f, 0.521972656f, 0.f),
    C_Vertex3f(0.247070313f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.001953125f, 0.611328125f, 0.f),
    C_Vertex3f(0.001953125f, 0.521972656f, 0.f),
    C_Vertex3f(0.083984375f, 0.521972656f, 0.f),
    C_Vertex3f(0.083984375f, 0.611328125f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0864257813f, 0.464355469f, -1.f),
    C_Vertex3f(0.0864257813f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.f, -1.f),
    C_Vertex3f(0.165039063f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.165039063f, 0.611328125f, -1.f),
    C_Vertex3f(0.165039063f, 0.521972656f, -1.f),
    C_Vertex3f(0.247070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.247070313f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.001953125f, 0.611328125f, -1.f),
    C_Vertex3f(0.001953125f, 0.521972656f, -1.f),
    C_Vertex3f(0.083984375f, 0.521972656f, -1.f),
    C_Vertex3f(0.083984375f, 0.611328125f, -1.f),
    C_End(),
    C_Translatef(0.248535156f, 0.f, 0.f),
};
static const Call char_240[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.135055542f, 0.640625f, 0.f),
    C_Vertex3f(0.135055542f, 0.640625f, -1.f),
    C_Vertex3f(0.229774475f, 0.640625f, -1.f),
    C_Vertex3f(0.229774475f, 0.640625f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.229774475f, 0.640625f, 0.f),
    C_Vertex3f(0.229774475f, 0.640625f, -1.f),
    C_Normal3f(0.632156551f, 0.774840713f, 0.f),
    C_Vertex3f(0.237379074f, 0.634420395f, 0.f),
    C_Vertex3f(0.237379074f, 0.634420395f, -1.f),
    C_Normal3f(0.632125914f, 0.774865746f, 0.f),
    C_Vertex3f(0.251707077f, 0.622731209f, 0.f),
    C_Vertex3f(0.251707077f, 0.622731209f, -1.f),
    C_Normal3f(0.632084489f, 0.774899423f, 0.f),
    C_Vertex3f(0.265153885f, 0.611762047f, 0.f),
    C_Vertex3f(0.265153885f, 0.611762047f, -1.f),
    C_Normal3f(0.632045567f, 0.774931252f, 0.f),
    C_Vertex3f(0.277719498f, 0.601512909f, 0.f),
    C_Vertex3f(0.277719498f, 0.601512909f, -1.f),
    C_Normal3f(0.632010818f, 0.774959624f, 0.f),
    C_Vertex3f(0.283561707f, 0.596748352f, 0.f),
    C_Vertex3f(0.283561707f, 0.596748352f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.422254294f, 0.906477392f, 0.f),
    C_Vertex3f(0.283561707f, 0.596748352f, 0.f),
    C_Vertex3f(0.283561707f, 0.596748352f, -1.f),
    C_Vertex3f(0.377754211f, 0.640625f, -1.f),
    C_Vertex3f(0.377754211f, 0.640625f, 0.f),
    C_Normal3f(0.915048897f, 0.403342903f, 0.f),
    C_Vertex3f(0.377754211f, 0.640625f, 0.f),
    C_Vertex3f(0.377754211f, 0.640625f, -1.f),
    C_Vertex3f(0.397377014f, 0.596107483f, -1.f),
    C_Vertex3f(0.397377014f, 0.596107483f, 0.f),
    C_Normal3f(0.42840445f, -0.903587103f, 0.f),
    C_Vertex3f(0.397377014f, 0.596107483f, 0.f),
    C_Vertex3f(0.397377014f, 0.596107483f, -1.f),
    C_Vertex3f(0.321922302f, 0.560333252f, -1.f),
    C_Vertex3f(0.321922302f, 0.560333252f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.754461467f, 0.656344295f, 0.f),
    C_Vertex3f(0.321922302f, 0.560333252f, 0.f),
    C_Vertex3f(0.321922302f, 0.560333252f, -1.f),
    C_Normal3f(0.768935144f, 0.639326751f, 0.f),
    C_Vertex3f(0.330671787f, 0.550275803f, 0.f),
    C_Vertex3f(0.330671787f, 0.550275803f, -1.f),
    C_Normal3f(0.787124753f, 0.616793871f, 0.f),
    C_Vertex3f(0.347077072f, 0.530078948f, 0.f),
    C_Vertex3f(0.347077072f, 0.530078948f, -1.f),
    C_Normal3f(0.808987379f, 0.587826014f, 0.f),
    C_Vertex3f(0.36238867f, 0.509800136f, 0.f),
    C_Vertex3f(0.36238867f, 0.509800136f, -1.f),
    C_Normal3f(0.830734193f, 0.556669295f, 0.f),
    C_Vertex3f(0.376606584f, 0.489439368f, 0.f),
    C_Vertex3f(0.376606584f, 0.489439368f, -1.f),
    C_Normal3f(0.852177322f, 0.523253143f, 0.f),
    C_Vertex3f(0.389730811f, 0.468996644f, 0.f),
    C_Vertex3f(0.389730811f, 0.468996644f, -1.f),
    C_Normal3f(0.873103857f, 0.487534195f, 0.f),
    C_Vertex3f(0.401761353f, 0.448471963f, 0.f),
    C_Vertex3f(0.401761353f, 0.448471963f, -1.f),
    C_Normal3f(0.89327848f, 0.449503511f, 0.f),
    C_Vertex3f(0.412698209f, 0.427865326f, 0.f),
    C_Vertex3f(0.412698209f, 0.427865326f, -1.f),
    C_Normal3f(0.912447512f, 0.409193754f, 0.f),
    C_Vertex3f(0.42254138f, 0.407176733f, 0.f),
    C_Vertex3f(0.42254138f, 0.407176733f, -1.f),
    C_Normal3f(0.930344939f, 0.366685569f, 0.f),
    C_Vertex3f(0.431290865f, 0.386406183f, 0.f),
    C_Vertex3f(0.431290865f, 0.386406183f, -1.f),
    C_Normal3f(0.946701348f, 0.322112799f, 0.f),
    C_Vertex3f(0.438946664f, 0.365553677f, 0.f),
    C_Vertex3f(0.438946664f, 0.365553677f, -1.f),
    C_Normal3f(0.961253703f, 0.275665313f, 0.f),
    C_Vertex3f(0.445508778f, 0.344619215f, 0.f),
    C_Vertex3f(0.445508778f, 0.344619215f, -1.f),
    C_Normal3f(0.973757267f, 0.227588892f, 0.f),
    C_Vertex3f(0.450977206f, 0.323602796f, 0.f),
    C_Vertex3f(0.450977206f, 0.323602796f, -1.f),
    C_Normal3f(0.983997643f, 0.178181618f, 0.f),
    C_Vertex3f(0.455351949f, 0.30250442f, 0.f),
    C_Vertex3f(0.455351949f, 0.30250442f, -1.f),
    C_Normal3f(0.991801739f, 0.127786562f, 0.f),
    C_Vertex3f(0.458633006f, 0.281324089f, 0.f),
    C_Vertex3f(0.458633006f, 0.281324089f, -1.f),
    C_Normal3f(0.997047961f, 0.0767809749f, 0.f),
    C_Vertex3f(0.460820377f, 0.2600618f, 0.f),
    C_Vertex3f(0.460820377f, 0.2600618f, -1.f),
    C_Normal3f(0.999716043f, 0.0238282718f, 0.f),
    C_Vertex3f(0.461914063f, 0.238717556f, 0.f),
    C_Vertex3f(0.461914063f, 0.238717556f, -1.f),
    C_Normal3f(0.99930042f, -0.0373979658f, 0.f),
    C_Vertex3f(0.461914063f, 0.214176178f, 0.f),
    C_Vertex3f(0.461914063f, 0.214176178f, -1.f),
    C_Normal3f(0.993806124f, -0.111128516f, 0.f),
    C_Vertex3f(0.460015059f, 0.187974811f, 0.f),
    C_Vertex3f(0.460015059f, 0.187974811f, -1.f),
    C_Normal3f(0.981019735f, -0.193908215f, 0.f),
    C_Vertex3f(0.456217051f, 0.163228631f, 0.f),
    C_Vertex3f(0.456217051f, 0.163228631f, -1.f),
    C_Normal3f(0.959247708f, -0.28256619f, 0.f),
    C_Vertex3f(0.450520039f, 0.139937639f, 0.f),
    C_Vertex3f(0.450520039f, 0.139937639f, -1.f),
    C_Normal3f(0.926920533f, -0.375257581f, 0.f),
    C_Vertex3f(0.442924023f, 0.118101835f, 0.f),
    C_Vertex3f(0.442924023f, 0.118101835f, -1.f),
    C_Normal3f(0.883044004f, -0.469290316f, 0.f),
    C_Vertex3f(0.433429003f, 0.0977212191f, 0.f),
    C_Vertex3f(0.433429003f, 0.0977212191f, -1.f),
    C_Normal3f(0.827581525f, -0.561345577f, 0.f),
    C_Vertex3f(0.422034979f, 0.0787957907f, 0.f),
    C_Vertex3f(0.422034979f, 0.0787957907f, -1.f),
    C_Normal3f(0.761686027f, -0.647946298f, 0.f),
    C_Vertex3f(0.408741951f, 0.0613255501f, 0.f),
    C_Vertex3f(0.408741951f, 0.0613255501f, -1.f),
    C_Normal3f(0.690889359f, -0.722960532f, 0.f),
    C_Vertex3f(0.393549919f, 0.0453104973f, 0.f),
    C_Vertex3f(0.393549919f, 0.0453104973f, -1.f),
    C_Normal3f(0.613503337f, -0.789692104f, 0.f),
    C_Vertex3f(0.377319813f, 0.0312973261f, 0.f),
    C_Vertex3f(0.377319813f, 0.0312973261f, -1.f),
    C_Normal3f(0.526327431f, -0.850281954f, 0.f),
    C_Vertex3f(0.360051632f, 0.0192860365f, 0.f),
    C_Vertex3f(0.360051632f, 0.0192860365f, -1.f),
    C_Normal3f(0.431653053f, -0.902039707f, 0.f),
    C_Vertex3f(0.341745377f, 0.00927662849f, 0.f),
    C_Vertex3f(0.341745377f, 0.00927662849f, -1.f),
    C_Normal3f(0.332650632f, -0.943050146f, 0.f),
    C_Vertex3f(0.322401047f, 0.0012691021f, 0.f),
    C_Vertex3f(0.322401047f, 0.0012691021f, -1.f),
    C_Normal3f(0.232860744f, -0.972510099f, 0.f),
    C_Vertex3f(0.302018642f, -0.0047365427f, 0.f),
    C_Vertex3f(0.302018642f, -0.0047365427f, -1.f),
    C_Normal3f(0.135603979f, -0.990763128f, 0.f),
    C_Vertex3f(0.280598164f, -0.0087403059f, 0.f),
    C_Vertex3f(0.280598164f, -0.0087403059f, -1.f),
    C_Normal3f(0.0415469855f, -0.999136567f, 0.f),
    C_Vertex3f(0.25813961f, -0.0107421875f, 0.f),
    C_Vertex3f(0.25813961f, -0.0107421875f, -1.f),
    C_Normal3f(-0.0508137308f, -0.998708129f, 0.f),
    C_Vertex3f(0.232456207f, -0.0107421875f, 0.f),
    C_Vertex3f(0.232456207f, -0.0107421875f, -1.f),
    C_Normal3f(-0.153407589f, -0.988162994f, 0.f),
    C_Vertex3f(0.206116199f, -0.00809526443f, 0.f),
    C_Vertex3f(0.206116199f, -0.00809526443f, -1.f),
    C_Normal3f(-0.265342891f, -0.964154065f, 0.f),
    C_Vertex3f(0.181306362f, -0.0028014183f, 0.f),
    C_Vertex3f(0.181306362f, -0.0028014183f, -1.f),
    C_Normal3f(-0.380522251f, -0.924771786f, 0.f),
    C_Vertex3f(0.158026695f, 0.00513935089f, 0.f),
    C_Vertex3f(0.158026695f, 0.00513935089f, -1.f),
    C_Normal3f(-0.493639439f, -0.869666636f, 0.f),
    C_Vertex3f(0.136277199f, 0.0157270432f, 0.f),
    C_Vertex3f(0.136277199f, 0.0157270432f, -1.f),
    C_Normal3f(-0.599140882f, -0.800643623f, 0.f),
    C_Vertex3f(0.116057873f, 0.0289616585f, 0.f),
    C_Vertex3f(0.116057873f, 0.0289616585f, -1.f),
    C_Normal3f(-0.692487657f, -0.721429706f, 0.f),
    C_Vertex3f(0.0973687172f, 0.0448431969f, 0.f),
    C_Vertex3f(0.0973687172f, 0.0448431969f, -1.f),
    C_Normal3f(-0.768062174f, -0.640375257f, 0.f),
    C_Vertex3f(0.0802097321f, 0.0633716583f, 0.f),
    C_Vertex3f(0.0802097321f, 0.0633716583f, -1.f),
    C_Normal3f(-0.831903577f, -0.554920137f, 0.f),
    C_Vertex3f(0.0673131943f, 0.0808916092f, 0.f),
    C_Vertex3f(0.0673131943f, 0.0808916092f, -1.f),
    C_Normal3f(-0.886093438f, -0.463506639f, 0.f),
    C_Vertex3f(0.058419466f, 0.0960383415f, 0.f),
    C_Vertex3f(0.058419466f, 0.0960383415f, -1.f),
    C_Normal3f(-0.92545712f, -0.378852278f, 0.f),
    C_Vertex3f(0.0507962704f, 0.112467289f, 0.f),
    C_Vertex3f(0.0507962704f, 0.112467289f, -1.f),
    C_Normal3f(-0.954741657f, -0.297436595f, 0.f),
    C_Vertex3f(0.0444436073f, 0.130178452f, 0.f),
    C_Vertex3f(0.0444436073f, 0.130178452f, -1.f),
    C_Normal3f(-0.975299239f, -0.220888048f, 0.f),
    C_Vertex3f(0.0393614769f, 0.149171829f, 0.f),
    C_Vertex3f(0.0393614769f, 0.149171829f, -1.f),
    C_Normal3f(-0.988665462f, -0.150135025f, 0.f),
    C_Vertex3f(0.0355498791f, 0.169447422f, 0.f),
    C_Vertex3f(0.0355498791f, 0.169447422f, -1.f),
    C_Normal3f(-0.996335089f, -0.0855363607f, 0.f),
    C_Vertex3f(0.0330088139f, 0.19100523f, 0.f),
    C_Vertex3f(0.0330088139f, 0.19100523f, -1.f),
    C_Normal3f(-0.999666035f, -0.0258434825f, 0.f),
    C_Vertex3f(0.0317382813f, 0.213845253f, 0.f),
    C_Vertex3f(0.0317382813f, 0.213845253f, -1.f),
    C_Normal3f(-0.999450088f, 0.0331601687f, 0.f),
    C_Vertex3f(0.0317382813f, 0.240151405f, 0.f),
    C_Vertex3f(0.0317382813f, 0.240151405f, -1.f),
    C_Normal3f(-0.99488163f, 0.101046987f, 0.f),
    C_Vertex3f(0.0335024595f, 0.267017722f, 0.f),
    C_Vertex3f(0.0335024595f, 0.267017722f, -1.f),
    C_Normal3f(-0.984092474f, 0.177657127f, 0.f),
    C_Vertex3f(0.0370308161f, 0.292260289f, 0.f),
    C_Vertex3f(0.0370308161f, 0.292260289f, -1.f),
    C_Normal3f(-0.965250552f, 0.261326522f, 0.f),
    C_Vertex3f(0.0423233509f, 0.315879107f, 0.f),
    C_Vertex3f(0.0423233509f, 0.315879107f, -1.f),
    C_Normal3f(-0.936400354f, 0.350933731f, 0.f),
    C_Vertex3f(0.049380064f, 0.337874174f, 0.f),
    C_Vertex3f(0.049380064f, 0.337874174f, -1.f),
    C_Normal3f(-0.895828664f, 0.444399565f, 0.f),
    C_Vertex3f(0.0582009554f, 0.358245492f, 0.f),
    C_Vertex3f(0.0582009554f, 0.358245492f, -1.f),
    C_Normal3f(-0.842520535f, 0.538664281f, 0.f),
    C_Vertex3f(0.068786025f, 0.37699306f, 0.f),
    C_Vertex3f(0.068786025f, 0.37699306f, -1.f),
    C_Normal3f(-0.77662915f, 0.629958153f, 0.f),
    C_Vertex3f(0.081135273f, 0.394116879f, 0.f),
    C_Vertex3f(0.081135273f, 0.394116879f, -1.f),
    C_Normal3f(-0.706460297f, 0.707752645f, 0.f),
    C_Vertex3f(0.0952486992f, 0.409616947f, 0.f),
    C_Vertex3f(0.0952486992f, 0.409616947f, -1.f),
    C_Normal3f(-0.631898761f, 0.775050879f, 0.f),
    C_Vertex3f(0.110251069f, 0.423179507f, 0.f),
    C_Vertex3f(0.110251069f, 0.423179507f, -1.f),
    C_Normal3f(-0.546358585f, 0.837551355f, 0.f),
    C_Vertex3f(0.126142383f, 0.434804559f, 0.f),
    C_Vertex3f(0.126142383f, 0.434804559f, -1.f),
    C_Normal3f(-0.451618254f, 0.892211318f, 0.f),
    C_Vertex3f(0.14292264f, 0.444492102f, 0.f),
    C_Vertex3f(0.14292264f, 0.444492102f, -1.f),
    C_Normal3f(-0.350609064f, 0.936521947f, 0.f),
    C_Vertex3f(0.160591841f, 0.452242136f, 0.f),
    C_Vertex3f(0.160591841f, 0.452242136f, -1.f),
    C_Normal3f(-0.24700214f, 0.969014943f, 0.f),
    C_Vertex3f(0.179149985f, 0.458054662f, 0.f),
    C_Vertex3f(0.179149985f, 0.458054662f, -1.f),
    C_Normal3f(-0.144570425f, 0.989494562f, 0.f),
    C_Vertex3f(0.198597074f, 0.461929679f, 0.f),
    C_Vertex3f(0.198597074f, 0.461929679f, -1.f),
    C_Normal3f(-0.0518819429f, 0.998653173f, 0.f),
    C_Vertex3f(0.218933105f, 0.463867188f, 0.f),
    C_Vertex3f(0.218933105f, 0.463867188f, -1.f),
    C_Normal3f(0.0227765162f, 0.999740541f, 0.f),
    C_Vertex3f(0.235891342f, 0.463867188f, 0.f),
    C_Vertex3f(0.235891342f, 0.463867188f, -1.f),
    C_Normal3f(0.0852067545f, 0.996363282f, 0.f),
    C_Vertex3f(0.247816563f, 0.463209152f, 0.f),
    C_Vertex3f(0.247816563f, 0.463209152f, -1.f),
    C_Normal3f(0.150916457f, 0.98854655f, 0.f),
    C_Vertex3f(0.258975506f, 0.461893082f, 0.f),
    C_Vertex3f(0.258975506f, 0.461893082f, -1.f),
    C_Normal3f(0.224233896f, 0.974535346f, 0.f),
    C_Vertex3f(0.269368172f, 0.459918976f, 0.f),
    C_Vertex3f(0.269368172f, 0.459918976f, -1.f),
    C_Normal3f(0.30392313f, 0.952696621f, 0.f),
    C_Vertex3f(0.27899456f, 0.457286835f, 0.f),
    C_Vertex3f(0.27899456f, 0.457286835f, -1.f),
    C_Normal3f(0.368004769f, 0.929823875f, 0.f),
    C_Vertex3f(0.289824963f, 0.453392982f, 0.f),
    C_Vertex3f(0.289824963f, 0.453392982f, -1.f),
    C_Normal3f(0.416341037f, 0.909208536f, 0.f),
    C_Vertex3f(0.301859379f, 0.448237419f, 0.f),
    C_Vertex3f(0.301859379f, 0.448237419f, -1.f),
    C_Normal3f(0.448155075f, 0.893955827f, 0.f),
    C_Vertex3f(0.315097809f, 0.441820145f, 0.f),
    C_Vertex3f(0.315097809f, 0.441820145f, -1.f),
    C_Normal3f(0.469462126f, 0.882952631f, 0.f),
    C_Vertex3f(0.322319031f, 0.437980652f, 0.f),
    C_Vertex3f(0.322319031f, 0.437980652f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.898275852f, -0.439432025f, 0.f),
    C_Vertex3f(0.322319031f, 0.437980652f, 0.f),
    C_Vertex3f(0.322319031f, 0.437980652f, -1.f),
    C_Normal3f(-0.892134726f, -0.451769322f, 0.f),
    C_Vertex3f(0.318389893f, 0.446012497f, 0.f),
    C_Vertex3f(0.318389893f, 0.446012497f, -1.f),
    C_Normal3f(-0.882769704f, -0.469805986f, 0.f),
    C_Vertex3f(0.311022758f, 0.460288048f, 0.f),
    C_Vertex3f(0.311022758f, 0.460288048f, -1.f),
    C_Normal3f(-0.868060768f, -0.496458054f, 0.f),
    C_Vertex3f(0.304146767f, 0.472775459f, 0.f),
    C_Vertex3f(0.304146767f, 0.472775459f, -1.f),
    C_Normal3f(-0.847569585f, -0.530684173f, 0.f),
    C_Vertex3f(0.297761917f, 0.483474731f, 0.f),
    C_Vertex3f(0.297761917f, 0.483474731f, -1.f),
    C_Normal3f(-0.819560647f, -0.572992444f, 0.f),
    C_Vertex3f(0.29186821f, 0.492385864f, 0.f),
    C_Vertex3f(0.29186821f, 0.492385864f, -1.f),
    C_Normal3f(-0.79488194f, -0.606764078f, 0.f),
    C_Vertex3f(0.284568787f, 0.502345085f, 0.f),
    C_Vertex3f(0.284568787f, 0.502345085f, -1.f),
    C_Normal3f(-0.774837375f, -0.632160664f, 0.f),
    C_Vertex3f(0.275863647f, 0.513352394f, 0.f),
    C_Vertex3f(0.275863647f, 0.513352394f, -1.f),
    C_Normal3f(-0.760867417f, -0.648907483f, 0.f),
    C_Vertex3f(0.265752792f, 0.525407791f, 0.f),
    C_Vertex3f(0.265752792f, 0.525407791f, -1.f),
    C_Normal3f(-0.751125515f, -0.660159528f, 0.f),
    C_Vertex3f(0.259994507f, 0.531959534f, 0.f),
    C_Vertex3f(0.259994507f, 0.531959534f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.423027754f, -0.906116784f, 0.f),
    C_Vertex3f(0.259994507f, 0.531959534f, 0.f),
    C_Vertex3f(0.259994507f, 0.531959534f, -1.f),
    C_Vertex3f(0.140289307f, 0.476074219f, -1.f),
    C_Vertex3f(0.140289307f, 0.476074219f, 0.f),
    C_Normal3f(-0.91962868f, -0.392788976f, 0.f),
    C_Vertex3f(0.140289307f, 0.476074219f, 0.f),
    C_Vertex3f(0.140289307f, 0.476074219f, -1.f),
    C_Vertex3f(0.121086121f, 0.521034241f, -1.f),
    C_Vertex3f(0.121086121f, 0.521034241f, 0.f),
    C_Normal3f(-0.422322124f, 0.906445861f, 0.f),
    C_Vertex3f(0.121086121f, 0.521034241f, 0.f),
    C_Vertex3f(0.121086121f, 0.521034241f, -1.f),
    C_Vertex3f(0.224594116f, 0.569259644f, -1.f),
    C_Vertex3f(0.224594116f, 0.569259644f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.623270392f, -0.782006443f, 0.f),
    C_Vertex3f(0.224594116f, 0.569259644f, 0.f),
    C_Vertex3f(0.224594116f, 0.569259644f, -1.f),
    C_Normal3f(-0.623280108f, -0.781998694f, 0.f),
    C_Vertex3f(0.180130005f, 0.604698181f, 0.f),
    C_Vertex3f(0.180130005f, 0.604698181f, -1.f),
    C_Normal3f(-0.623289704f, -0.781990945f, 0.f),
    C_Vertex3f(0.135055542f, 0.640625f, 0.f),
    C_Vertex3f(0.135055542f, 0.640625f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999574602f, 0.0291639771f, 0.f),
    C_Vertex3f(0.380371094f, 0.220233917f, 0.f),
    C_Vertex3f(0.380371094f, 0.220233917f, -1.f),
    C_Normal3f(-0.999548137f, -0.0300593283f, 0.f),
    C_Vertex3f(0.380371094f, 0.241594315f, 0.f),
    C_Vertex3f(0.380371094f, 0.241594315f, -1.f),
    C_Normal3f(-0.995320559f, -0.096628055f, 0.f),
    C_Vertex3f(0.379136562f, 0.261285186f, 0.f),
    C_Vertex3f(0.379136562f, 0.261285186f, -1.f),
    C_Normal3f(-0.985365987f, -0.170451999f, 0.f),
    C_Vertex3f(0.3766675f, 0.279743314f, 0.f),
    C_Vertex3f(0.3766675f, 0.279743314f, -1.f),
    C_Normal3f(-0.967786789f, -0.251771212f, 0.f),
    C_Vertex3f(0.372963905f, 0.296968699f, 0.f),
    C_Vertex3f(0.372963905f, 0.296968699f, -1.f),
    C_Normal3f(-0.940498114f, -0.339798599f, 0.f),
    C_Vertex3f(0.36802578f, 0.31296134f, 0.f),
    C_Vertex3f(0.36802578f, 0.31296134f, -1.f),
    C_Normal3f(-0.901497483f, -0.432784438f, 0.f),
    C_Vertex3f(0.361853123f, 0.327721238f, 0.f),
    C_Vertex3f(0.361853123f, 0.327721238f, -1.f),
    C_Normal3f(-0.849319756f, -0.527878702f, 0.f),
    C_Vertex3f(0.354445934f, 0.341248393f, 0.f),
    C_Vertex3f(0.354445934f, 0.341248393f, -1.f),
    C_Normal3f(-0.783592701f, -0.621274948f, 0.f),
    C_Vertex3f(0.345804214f, 0.353542805f, 0.f),
    C_Vertex3f(0.345804214f, 0.353542805f, -1.f),
    C_Normal3f(-0.714338601f, -0.699800193f, 0.f),
    C_Vertex3f(0.335927963f, 0.364604473f, 0.f),
    C_Vertex3f(0.335927963f, 0.364604473f, -1.f),
    C_Normal3f(-0.642286062f, -0.766464949f, 0.f),
    C_Vertex3f(0.325485706f, 0.374283433f, 0.f),
    C_Vertex3f(0.325485706f, 0.374283433f, -1.f),
    C_Normal3f(-0.558629453f, -0.829417348f, 0.f),
    C_Vertex3f(0.314477444f, 0.382579684f, 0.f),
    C_Vertex3f(0.314477444f, 0.382579684f, -1.f),
    C_Normal3f(-0.464665353f, -0.885486364f, 0.f),
    C_Vertex3f(0.302903175f, 0.389493227f, 0.f),
    C_Vertex3f(0.302903175f, 0.389493227f, -1.f),
    C_Normal3f(-0.362980783f, -0.931796551f, 0.f),
    C_Vertex3f(0.290762901f, 0.395024061f, 0.f),
    C_Vertex3f(0.290762901f, 0.395024061f, -1.f),
    C_Normal3f(-0.257173836f, -0.966365159f, 0.f),
    C_Vertex3f(0.278056622f, 0.399172187f, 0.f),
    C_Vertex3f(0.278056622f, 0.399172187f, -1.f),
    C_Normal3f(-0.151247442f, -0.988495946f, 0.f),
    C_Vertex3f(0.264784336f, 0.401937604f, 0.f),
    C_Vertex3f(0.264784336f, 0.401937604f, -1.f),
    C_Normal3f(-0.049088221f, -0.998794436f, 0.f),
    C_Vertex3f(0.250946045f, 0.403320313f, 0.f),
    C_Vertex3f(0.250946045f, 0.403320313f, -1.f),
    C_Normal3f(0.0495544933f, -0.998771429f, 0.f),
    C_Vertex3f(0.236650467f, 0.403320313f, 0.f),
    C_Vertex3f(0.236650467f, 0.403320313f, -1.f),
    C_Normal3f(0.154468149f, -0.98799777f, 0.f),
    C_Vertex3f(0.223077536f, 0.401937604f, 0.f),
    C_Vertex3f(0.223077536f, 0.401937604f, -1.f),
    C_Normal3f(0.263558418f, -0.964643419f, 0.f),
    C_Vertex3f(0.210118532f, 0.399172187f, 0.f),
    C_Vertex3f(0.210118532f, 0.399172187f, -1.f),
    C_Normal3f(0.373000383f, -0.927831173f, 0.f),
    C_Vertex3f(0.197773457f, 0.395024061f, 0.f),
    C_Vertex3f(0.197773457f, 0.395024061f, -1.f),
    C_Normal3f(0.478307903f, -0.878192186f, 0.f),
    C_Vertex3f(0.186042309f, 0.389493227f, 0.f),
    C_Vertex3f(0.186042309f, 0.389493227f, -1.f),
    C_Normal3f(0.575376034f, -0.817889094f, 0.f),
    C_Vertex3f(0.174925089f, 0.382579684f, 0.f),
    C_Vertex3f(0.174925089f, 0.382579684f, -1.f),
    C_Normal3f(0.661242843f, -0.75017184f, 0.f),
    C_Vertex3f(0.164421797f, 0.374283433f, 0.f),
    C_Vertex3f(0.164421797f, 0.374283433f, -1.f),
    C_Normal3f(0.734457314f, -0.67865485f, 0.f),
    C_Vertex3f(0.154532433f, 0.364604473f, 0.f),
    C_Vertex3f(0.154532433f, 0.364604473f, -1.f),
    C_Normal3f(0.80487752f, -0.59344101f, 0.f),
    C_Vertex3f(0.145256996f, 0.353542805f, 0.f),
    C_Vertex3f(0.145256996f, 0.353542805f, -1.f),
    C_Normal3f(0.869368911f, -0.494163632f, 0.f),
    C_Vertex3f(0.137140989f, 0.34101665f, 0.f),
    C_Vertex3f(0.137140989f, 0.34101665f, -1.f),
    C_Normal3f(0.91762501f, -0.397447169f, 0.f),
    C_Vertex3f(0.130184412f, 0.32702601f, 0.f),
    C_Vertex3f(0.130184412f, 0.32702601f, -1.f),
    C_Normal3f(0.951782405f, -0.306773782f, 0.f),
    C_Vertex3f(0.124387264f, 0.311570883f, 0.f),
    C_Vertex3f(0.124387264f, 0.311570883f, -1.f),
    C_Normal3f(0.97457844f, -0.224046841f, 0.f),
    C_Vertex3f(0.119749546f, 0.29465127f, 0.f),
    C_Vertex3f(0.119749546f, 0.29465127f, -1.f),
    C_Normal3f(0.988698959f, -0.149914503f, 0.f),
    C_Vertex3f(0.116271257f, 0.276267171f, 0.f),
    C_Vertex3f(0.116271257f, 0.276267171f, -1.f),
    C_Normal3f(0.996448636f, -0.0842030197f, 0.f),
    C_Vertex3f(0.113952398f, 0.256418586f, 0.f),
    C_Vertex3f(0.113952398f, 0.256418586f, -1.f),
    C_Normal3f(0.99964124f, -0.0267841462f, 0.f),
    C_Vertex3f(0.112792969f, 0.235105515f, 0.f),
    C_Vertex3f(0.112792969f, 0.235105515f, -1.f),
    C_Normal3f(0.999592423f, 0.0285465047f, 0.f),
    C_Vertex3f(0.112792969f, 0.21314621f, 0.f),
    C_Vertex3f(0.112792969f, 0.21314621f, -1.f),
    C_Normal3f(0.995671868f, 0.0929389894f, 0.f),
    C_Vertex3f(0.113986492f, 0.193312764f, 0.f),
    C_Vertex3f(0.113986492f, 0.193312764f, -1.f),
    C_Normal3f(0.986347854f, 0.164674759f, 0.f),
    C_Vertex3f(0.116373539f, 0.174786925f, 0.f),
    C_Vertex3f(0.116373539f, 0.174786925f, -1.f),
    C_Normal3f(0.969641566f, 0.244530901f, 0.f),
    C_Vertex3f(0.119954109f, 0.157568693f, 0.f),
    C_Vertex3f(0.119954109f, 0.157568693f, -1.f),
    C_Normal3f(0.943259835f, 0.332055539f, 0.f),
    C_Vertex3f(0.124728203f, 0.141658068f, 0.f),
    C_Vertex3f(0.124728203f, 0.141658068f, -1.f),
    C_Normal3f(0.904817641f, 0.425799251f, 0.f),
    C_Vertex3f(0.13069582f, 0.127055049f, 0.f),
    C_Vertex3f(0.13069582f, 0.127055049f, -1.f),
    C_Normal3f(0.852306187f, 0.523043275f, 0.f),
    C_Vertex3f(0.13785696f, 0.113759637f, 0.f),
    C_Vertex3f(0.13785696f, 0.113759637f, -1.f),
    C_Normal3f(0.784763873f, 0.619794786f, 0.f),
    C_Vertex3f(0.146211624f, 0.101771832f, 0.f),
    C_Vertex3f(0.146211624f, 0.101771832f, -1.f),
    C_Normal3f(0.713056087f, 0.701107085f, 0.f),
    C_Vertex3f(0.155759811f, 0.0910916328f, 0.f),
    C_Vertex3f(0.155759811f, 0.0910916328f, -1.f),
    C_Normal3f(0.639273524f, 0.76897943f, 0.f),
    C_Vertex3f(0.165901423f, 0.081746459f, 0.f),
    C_Vertex3f(0.165901423f, 0.081746459f, -1.f),
    C_Normal3f(0.554080367f, 0.832463205f, 0.f),
    C_Vertex3f(0.176636457f, 0.07373631f, 0.f),
    C_Vertex3f(0.176636457f, 0.07373631f, -1.f),
    C_Normal3f(0.459096879f, 0.88838619f, 0.f),
    C_Vertex3f(0.187964916f, 0.0670611858f, 0.f),
    C_Vertex3f(0.187964916f, 0.0670611858f, -1.f),
    C_Normal3f(0.35718891f, 0.934032202f, 0.f),
    C_Vertex3f(0.199886799f, 0.0617210865f, 0.f),
    C_Vertex3f(0.199886799f, 0.0617210865f, -1.f),
    C_Normal3f(0.252089262f, 0.967704058f, 0.f),
    C_Vertex3f(0.212402105f, 0.057716012f, 0.f),
    C_Vertex3f(0.212402105f, 0.057716012f, -1.f),
    C_Normal3f(0.147743061f, 0.989025772f, 0.f),
    C_Vertex3f(0.225510836f, 0.0550459623f, 0.f),
    C_Vertex3f(0.225510836f, 0.0550459623f, -1.f),
    C_Normal3f(0.0479076132f, 0.998851717f, 0.f),
    C_Vertex3f(0.23921299f, 0.0537109375f, 0.f),
    C_Vertex3f(0.23921299f, 0.0537109375f, -1.f),
    C_Normal3f(-0.0510596298f, 0.998695672f, 0.f),
    C_Vertex3f(0.25334549f, 0.0537109375f, 0.f),
    C_Vertex3f(0.25334549f, 0.0537109375f, -1.f),
    C_Normal3f(-0.158414841f, 0.987372577f, 0.f),
    C_Vertex3f(0.266789556f, 0.0551208258f, 0.f),
    C_Vertex3f(0.266789556f, 0.0551208258f, -1.f),
    C_Normal3f(-0.268290788f, 0.963337958f, 0.f),
    C_Vertex3f(0.279708266f, 0.0579406023f, 0.f),
    C_Vertex3f(0.279708266f, 0.0579406023f, -1.f),
    C_Normal3f(-0.376804411f, 0.926292837f, 0.f),
    C_Vertex3f(0.292101622f, 0.0621702671f, 0.f),
    C_Vertex3f(0.292101622f, 0.0621702671f, -1.f),
    C_Normal3f(-0.479686439f, 0.877439976f, 0.f),
    C_Vertex3f(0.303969622f, 0.0678098202f, 0.f),
    C_Vertex3f(0.303969622f, 0.0678098202f, -1.f),
    C_Normal3f(-0.573383331f, 0.819287181f, 0.f),
    C_Vertex3f(0.315312266f, 0.0748592615f, 0.f),
    C_Vertex3f(0.315312266f, 0.0748592615f, -1.f),
    C_Normal3f(-0.655622482f, 0.755088866f, 0.f),
    C_Vertex3f(0.326129556f, 0.0833185911f, 0.f),
    C_Vertex3f(0.326129556f, 0.0833185911f, -1.f),
    C_Normal3f(-0.725554168f, 0.688165069f, 0.f),
    C_Vertex3f(0.33642149f, 0.093187809f, 0.f),
    C_Vertex3f(0.33642149f, 0.093187809f, -1.f),
    C_Normal3f(-0.793450296f, 0.608635068f, 0.f),
    C_Vertex3f(0.346188068f, 0.104466915f, 0.f),
    C_Vertex3f(0.346188068f, 0.104466915f, -1.f),
    C_Normal3f(-0.857932508f, 0.513762295f, 0.f),
    C_Vertex3f(0.354733825f, 0.117060781f, 0.f),
    C_Vertex3f(0.354733825f, 0.117060781f, -1.f),
    C_Normal3f(-0.908154964f, 0.418634176f, 0.f),
    C_Vertex3f(0.362058759f, 0.130969405f, 0.f),
    C_Vertex3f(0.362058759f, 0.130969405f, -1.f),
    C_Normal3f(-0.94505024f, 0.326925278f, 0.f),
    C_Vertex3f(0.36816287f, 0.146192789f, 0.f),
    C_Vertex3f(0.36816287f, 0.146192789f, -1.f),
    C_Normal3f(-0.97048682f, 0.241154164f, 0.f),
    C_Vertex3f(0.37304616f, 0.162730932f, 0.f),
    C_Vertex3f(0.37304616f, 0.162730932f, -1.f),
    C_Normal3f(-0.986677647f, 0.162687704f, 0.f),
    C_Vertex3f(0.376708627f, 0.180583835f, 0.f),
    C_Vertex3f(0.376708627f, 0.180583835f, -1.f),
    C_Normal3f(-0.995761096f, 0.09197817f, 0.f),
    C_Vertex3f(0.379150271f, 0.199751496f, 0.f),
    C_Vertex3f(0.379150271f, 0.199751496f, -1.f),
    C_Normal3f(-0.999574602f, 0.0291639771f, 0.f),
    C_Vertex3f(0.380371094f, 0.220233917f, 0.f),
    C_Vertex3f(0.380371094f, 0.220233917f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.229774475f, 0.640625f, 0.f),
    C_Vertex3f(0.135055542f, 0.640625f, 0.f),
    C_Vertex3f(0.180130005f, 0.604698181f, 0.f),
    C_Vertex3f(0.224594116f, 0.569259644f, 0.f),
    C_Vertex3f(0.259994507f, 0.531959534f, 0.f),
    C_Vertex3f(0.237379074f, 0.634420395f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.321922302f, 0.560333252f, 0.f),
    C_Vertex3f(0.397377014f, 0.596107483f, 0.f),
    C_Vertex3f(0.377754211f, 0.640625f, 0.f),
    C_Vertex3f(0.283561707f, 0.596748352f, 0.f),
    C_Vertex3f(0.284568787f, 0.502345085f, 0.f),
    C_Vertex3f(0.29186821f, 0.492385864f, 0.f),
    C_Vertex3f(0.297761917f, 0.483474731f, 0.f),
    C_Vertex3f(0.304146767f, 0.472775459f, 0.f),
    C_Vertex3f(0.311022758f, 0.460288048f, 0.f),
    C_Vertex3f(0.318389893f, 0.446012497f, 0.f),
    C_Vertex3f(0.322319031f, 0.437980652f, 0.f),
    C_Vertex3f(0.330671787f, 0.550275803f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.277719498f, 0.601512909f, 0.f),
    C_Vertex3f(0.265153885f, 0.611762047f, 0.f),
    C_Vertex3f(0.265752792f, 0.525407791f, 0.f),
    C_Vertex3f(0.275863647f, 0.513352394f, 0.f),
    C_Vertex3f(0.284568787f, 0.502345085f, 0.f),
    C_Vertex3f(0.283561707f, 0.596748352f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259994507f, 0.531959534f, 0.f),
    C_Vertex3f(0.265752792f, 0.525407791f, 0.f),
    C_Vertex3f(0.265153885f, 0.611762047f, 0.f),
    C_Vertex3f(0.251707077f, 0.622731209f, 0.f),
    C_Vertex3f(0.237379074f, 0.634420395f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.224594116f, 0.569259644f, 0.f),
    C_Vertex3f(0.121086121f, 0.521034241f, 0.f),
    C_Vertex3f(0.140289307f, 0.476074219f, 0.f),
    C_Vertex3f(0.259994507f, 0.531959534f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.438946664f, 0.365553677f, 0.f),
    C_Vertex3f(0.431290865f, 0.386406183f, 0.f),
    C_Vertex3f(0.433429003f, 0.0977212191f, 0.f),
    C_Vertex3f(0.42254138f, 0.407176733f, 0.f),
    C_Vertex3f(0.422034979f, 0.0787957907f, 0.f),
    C_Vertex3f(0.412698209f, 0.427865326f, 0.f),
    C_Vertex3f(0.408741951f, 0.0613255501f, 0.f),
    C_Vertex3f(0.401761353f, 0.448471963f, 0.f),
    C_Vertex3f(0.393549919f, 0.0453104973f, 0.f),
    C_Vertex3f(0.389730811f, 0.468996644f, 0.f),
    C_Vertex3f(0.380371094f, 0.241594315f, 0.f),
    C_Vertex3f(0.379136562f, 0.261285186f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.36802578f, 0.31296134f, 0.f),
    C_Vertex3f(0.372963905f, 0.296968699f, 0.f),
    C_Vertex3f(0.376606584f, 0.489439368f, 0.f),
    C_Vertex3f(0.3766675f, 0.279743314f, 0.f),
    C_Vertex3f(0.389730811f, 0.468996644f, 0.f),
    C_Vertex3f(0.379136562f, 0.261285186f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36238867f, 0.509800136f, 0.f),
    C_Vertex3f(0.347077072f, 0.530078948f, 0.f),
    C_Vertex3f(0.354445934f, 0.341248393f, 0.f),
    C_Vertex3f(0.361853123f, 0.327721238f, 0.f),
    C_Vertex3f(0.36802578f, 0.31296134f, 0.f),
    C_Vertex3f(0.376606584f, 0.489439368f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.322319031f, 0.437980652f, 0.f),
    C_Vertex3f(0.325485706f, 0.374283433f, 0.f),
    C_Vertex3f(0.330671787f, 0.550275803f, 0.f),
    C_Vertex3f(0.335927963f, 0.364604473f, 0.f),
    C_Vertex3f(0.347077072f, 0.530078948f, 0.f),
    C_Vertex3f(0.345804214f, 0.353542805f, 0.f),
    C_Vertex3f(0.354445934f, 0.341248393f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.455351949f, 0.30250442f, 0.f),
    C_Vertex3f(0.456217051f, 0.163228631f, 0.f),
    C_Vertex3f(0.458633006f, 0.281324089f, 0.f),
    C_Vertex3f(0.460015059f, 0.187974811f, 0.f),
    C_Vertex3f(0.460820377f, 0.2600618f, 0.f),
    C_Vertex3f(0.461914063f, 0.214176178f, 0.f),
    C_Vertex3f(0.461914063f, 0.238717556f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.438946664f, 0.365553677f, 0.f),
    C_Vertex3f(0.442924023f, 0.118101835f, 0.f),
    C_Vertex3f(0.445508778f, 0.344619215f, 0.f),
    C_Vertex3f(0.450520039f, 0.139937639f, 0.f),
    C_Vertex3f(0.450977206f, 0.323602796f, 0.f),
    C_Vertex3f(0.456217051f, 0.163228631f, 0.f),
    C_Vertex3f(0.455351949f, 0.30250442f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.380371094f, 0.241594315f, 0.f),
    C_Vertex3f(0.380371094f, 0.220233917f, 0.f),
    C_Vertex3f(0.393549919f, 0.0453104973f, 0.f),
    C_Vertex3f(0.379150271f, 0.199751496f, 0.f),
    C_Vertex3f(0.377319813f, 0.0312973261f, 0.f),
    C_Vertex3f(0.376708627f, 0.180583835f, 0.f),
    C_Vertex3f(0.37304616f, 0.162730932f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.37304616f, 0.162730932f, 0.f),
    C_Vertex3f(0.36816287f, 0.146192789f, 0.f),
    C_Vertex3f(0.377319813f, 0.0312973261f, 0.f),
    C_Vertex3f(0.362058759f, 0.130969405f, 0.f),
    C_Vertex3f(0.360051632f, 0.0192860365f, 0.f),
    C_Vertex3f(0.354733825f, 0.117060781f, 0.f),
    C_Vertex3f(0.346188068f, 0.104466915f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.341745377f, 0.00927662849f, 0.f),
    C_Vertex3f(0.360051632f, 0.0192860365f, 0.f),
    C_Vertex3f(0.346188068f, 0.104466915f, 0.f),
    C_Vertex3f(0.33642149f, 0.093187809f, 0.f),
    C_Vertex3f(0.326129556f, 0.0833185911f, 0.f),
    C_Vertex3f(0.322401047f, 0.0012691021f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.326129556f, 0.0833185911f, 0.f),
    C_Vertex3f(0.315312266f, 0.0748592615f, 0.f),
    C_Vertex3f(0.322401047f, 0.0012691021f, 0.f),
    C_Vertex3f(0.303969622f, 0.0678098202f, 0.f),
    C_Vertex3f(0.302018642f, -0.0047365427f, 0.f),
    C_Vertex3f(0.292101622f, 0.0621702671f, 0.f),
    C_Vertex3f(0.280598164f, -0.0087403059f, 0.f),
    C_Vertex3f(0.279708266f, 0.0579406023f, 0.f),
    C_Vertex3f(0.266789556f, 0.0551208258f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25813961f, -0.0107421875f, 0.f),
    C_Vertex3f(0.280598164f, -0.0087403059f, 0.f),
    C_Vertex3f(0.266789556f, 0.0551208258f, 0.f),
    C_Vertex3f(0.25334549f, 0.0537109375f, 0.f),
    C_Vertex3f(0.23921299f, 0.0537109375f, 0.f),
    C_Vertex3f(0.232456207f, -0.0107421875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.23921299f, 0.0537109375f, 0.f),
    C_Vertex3f(0.225510836f, 0.0550459623f, 0.f),
    C_Vertex3f(0.232456207f, -0.0107421875f, 0.f),
    C_Vertex3f(0.212402105f, 0.057716012f, 0.f),
    C_Vertex3f(0.206116199f, -0.00809526443f, 0.f),
    C_Vertex3f(0.199886799f, 0.0617210865f, 0.f),
    C_Vertex3f(0.187964916f, 0.0670611858f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.181306362f, -0.0028014183f, 0.f),
    C_Vertex3f(0.206116199f, -0.00809526443f, 0.f),
    C_Vertex3f(0.187964916f, 0.0670611858f, 0.f),
    C_Vertex3f(0.176636457f, 0.07373631f, 0.f),
    C_Vertex3f(0.165901423f, 0.081746459f, 0.f),
    C_Vertex3f(0.158026695f, 0.00513935089f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158026695f, 0.00513935089f, 0.f),
    C_Vertex3f(0.165901423f, 0.081746459f, 0.f),
    C_Vertex3f(0.155759811f, 0.0910916328f, 0.f),
    C_Vertex3f(0.146211624f, 0.101771832f, 0.f),
    C_Vertex3f(0.13785696f, 0.113759637f, 0.f),
    C_Vertex3f(0.136277199f, 0.0157270432f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136277199f, 0.0157270432f, 0.f),
    C_Vertex3f(0.13785696f, 0.113759637f, 0.f),
    C_Vertex3f(0.13069582f, 0.127055049f, 0.f),
    C_Vertex3f(0.124728203f, 0.141658068f, 0.f),
    C_Vertex3f(0.119954109f, 0.157568693f, 0.f),
    C_Vertex3f(0.116373539f, 0.174786925f, 0.f),
    C_Vertex3f(0.116057873f, 0.0289616585f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116373539f, 0.174786925f, 0.f),
    C_Vertex3f(0.113986492f, 0.193312764f, 0.f),
    C_Vertex3f(0.116057873f, 0.0289616585f, 0.f),
    C_Vertex3f(0.112792969f, 0.21314621f, 0.f),
    C_Vertex3f(0.0973687172f, 0.0448431969f, 0.f),
    C_Vertex3f(0.110251069f, 0.423179507f, 0.f),
    C_Vertex3f(0.0952486992f, 0.409616947f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.058419466f, 0.0960383415f, 0.f),
    C_Vertex3f(0.0673131943f, 0.0808916092f, 0.f),
    C_Vertex3f(0.068786025f, 0.37699306f, 0.f),
    C_Vertex3f(0.0802097321f, 0.0633716583f, 0.f),
    C_Vertex3f(0.081135273f, 0.394116879f, 0.f),
    C_Vertex3f(0.0973687172f, 0.0448431969f, 0.f),
    C_Vertex3f(0.0952486992f, 0.409616947f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0317382813f, 0.240151405f, 0.f),
    C_Vertex3f(0.0330088139f, 0.19100523f, 0.f),
    C_Vertex3f(0.0335024595f, 0.267017722f, 0.f),
    C_Vertex3f(0.0355498791f, 0.169447422f, 0.f),
    C_Vertex3f(0.0370308161f, 0.292260289f, 0.f),
    C_Vertex3f(0.0393614769f, 0.149171829f, 0.f),
    C_Vertex3f(0.0423233509f, 0.315879107f, 0.f),
    C_Vertex3f(0.0444436073f, 0.130178452f, 0.f),
    C_Vertex3f(0.049380064f, 0.337874174f, 0.f),
    C_Vertex3f(0.0507962704f, 0.112467289f, 0.f),
    C_Vertex3f(0.0582009554f, 0.358245492f, 0.f),
    C_Vertex3f(0.058419466f, 0.0960383415f, 0.f),
    C_Vertex3f(0.068786025f, 0.37699306f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.315097809f, 0.441820145f, 0.f),
    C_Vertex3f(0.301859379f, 0.448237419f, 0.f),
    C_Vertex3f(0.302903175f, 0.389493227f, 0.f),
    C_Vertex3f(0.314477444f, 0.382579684f, 0.f),
    C_Vertex3f(0.325485706f, 0.374283433f, 0.f),
    C_Vertex3f(0.322319031f, 0.437980652f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.290762901f, 0.395024061f, 0.f),
    C_Vertex3f(0.302903175f, 0.389493227f, 0.f),
    C_Vertex3f(0.301859379f, 0.448237419f, 0.f),
    C_Vertex3f(0.289824963f, 0.453392982f, 0.f),
    C_Vertex3f(0.27899456f, 0.457286835f, 0.f),
    C_Vertex3f(0.278056622f, 0.399172187f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186042309f, 0.389493227f, 0.f),
    C_Vertex3f(0.197773457f, 0.395024061f, 0.f),
    C_Vertex3f(0.198597074f, 0.461929679f, 0.f),
    C_Vertex3f(0.210118532f, 0.399172187f, 0.f),
    C_Vertex3f(0.218933105f, 0.463867188f, 0.f),
    C_Vertex3f(0.223077536f, 0.401937604f, 0.f),
    C_Vertex3f(0.235891342f, 0.463867188f, 0.f),
    C_Vertex3f(0.236650467f, 0.403320313f, 0.f),
    C_Vertex3f(0.247816563f, 0.463209152f, 0.f),
    C_Vertex3f(0.250946045f, 0.403320313f, 0.f),
    C_Vertex3f(0.258975506f, 0.461893082f, 0.f),
    C_Vertex3f(0.264784336f, 0.401937604f, 0.f),
    C_Vertex3f(0.269368172f, 0.459918976f, 0.f),
    C_Vertex3f(0.278056622f, 0.399172187f, 0.f),
    C_Vertex3f(0.27899456f, 0.457286835f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179149985f, 0.458054662f, 0.f),
    C_Vertex3f(0.160591841f, 0.452242136f, 0.f),
    C_Vertex3f(0.164421797f, 0.374283433f, 0.f),
    C_Vertex3f(0.174925089f, 0.382579684f, 0.f),
    C_Vertex3f(0.186042309f, 0.389493227f, 0.f),
    C_Vertex3f(0.198597074f, 0.461929679f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.130184412f, 0.32702601f, 0.f),
    C_Vertex3f(0.137140989f, 0.34101665f, 0.f),
    C_Vertex3f(0.14292264f, 0.444492102f, 0.f),
    C_Vertex3f(0.145256996f, 0.353542805f, 0.f),
    C_Vertex3f(0.160591841f, 0.452242136f, 0.f),
    C_Vertex3f(0.154532433f, 0.364604473f, 0.f),
    C_Vertex3f(0.164421797f, 0.374283433f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.126142383f, 0.434804559f, 0.f),
    C_Vertex3f(0.110251069f, 0.423179507f, 0.f),
    C_Vertex3f(0.113952398f, 0.256418586f, 0.f),
    C_Vertex3f(0.116271257f, 0.276267171f, 0.f),
    C_Vertex3f(0.119749546f, 0.29465127f, 0.f),
    C_Vertex3f(0.124387264f, 0.311570883f, 0.f),
    C_Vertex3f(0.130184412f, 0.32702601f, 0.f),
    C_Vertex3f(0.14292264f, 0.444492102f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.112792969f, 0.235105515f, 0.f),
    C_Vertex3f(0.113952398f, 0.256418586f, 0.f),
    C_Vertex3f(0.110251069f, 0.423179507f, 0.f),
    C_Vertex3f(0.112792969f, 0.21314621f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0330088139f, 0.19100523f, 0.f),
    C_Vertex3f(0.0317382813f, 0.240151405f, 0.f),
    C_Vertex3f(0.0317382813f, 0.213845253f, 0.f),
    C_Vertex3f(0.438946664f, 0.365553677f, 0.f),
    C_Vertex3f(0.433429003f, 0.0977212191f, 0.f),
    C_Vertex3f(0.442924023f, 0.118101835f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.229774475f, 0.640625f, -1.f),
    C_Vertex3f(0.135055542f, 0.640625f, -1.f),
    C_Vertex3f(0.180130005f, 0.604698181f, -1.f),
    C_Vertex3f(0.224594116f, 0.569259644f, -1.f),
    C_Vertex3f(0.259994507f, 0.531959534f, -1.f),
    C_Vertex3f(0.237379074f, 0.634420395f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.321922302f, 0.560333252f, -1.f),
    C_Vertex3f(0.397377014f, 0.596107483f, -1.f),
    C_Vertex3f(0.377754211f, 0.640625f, -1.f),
    C_Vertex3f(0.283561707f, 0.596748352f, -1.f),
    C_Vertex3f(0.284568787f, 0.502345085f, -1.f),
    C_Vertex3f(0.29186821f, 0.492385864f, -1.f),
    C_Vertex3f(0.297761917f, 0.483474731f, -1.f),
    C_Vertex3f(0.304146767f, 0.472775459f, -1.f),
    C_Vertex3f(0.311022758f, 0.460288048f, -1.f),
    C_Vertex3f(0.318389893f, 0.446012497f, -1.f),
    C_Vertex3f(0.322319031f, 0.437980652f, -1.f),
    C_Vertex3f(0.330671787f, 0.550275803f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.277719498f, 0.601512909f, -1.f),
    C_Vertex3f(0.265153885f, 0.611762047f, -1.f),
    C_Vertex3f(0.265752792f, 0.525407791f, -1.f),
    C_Vertex3f(0.275863647f, 0.513352394f, -1.f),
    C_Vertex3f(0.284568787f, 0.502345085f, -1.f),
    C_Vertex3f(0.283561707f, 0.596748352f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.259994507f, 0.531959534f, -1.f),
    C_Vertex3f(0.265752792f, 0.525407791f, -1.f),
    C_Vertex3f(0.265153885f, 0.611762047f, -1.f),
    C_Vertex3f(0.251707077f, 0.622731209f, -1.f),
    C_Vertex3f(0.237379074f, 0.634420395f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.224594116f, 0.569259644f, -1.f),
    C_Vertex3f(0.121086121f, 0.521034241f, -1.f),
    C_Vertex3f(0.140289307f, 0.476074219f, -1.f),
    C_Vertex3f(0.259994507f, 0.531959534f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.438946664f, 0.365553677f, -1.f),
    C_Vertex3f(0.431290865f, 0.386406183f, -1.f),
    C_Vertex3f(0.433429003f, 0.0977212191f, -1.f),
    C_Vertex3f(0.42254138f, 0.407176733f, -1.f),
    C_Vertex3f(0.422034979f, 0.0787957907f, -1.f),
    C_Vertex3f(0.412698209f, 0.427865326f, -1.f),
    C_Vertex3f(0.408741951f, 0.0613255501f, -1.f),
    C_Vertex3f(0.401761353f, 0.448471963f, -1.f),
    C_Vertex3f(0.393549919f, 0.0453104973f, -1.f),
    C_Vertex3f(0.389730811f, 0.468996644f, -1.f),
    C_Vertex3f(0.380371094f, 0.241594315f, -1.f),
    C_Vertex3f(0.379136562f, 0.261285186f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.36802578f, 0.31296134f, -1.f),
    C_Vertex3f(0.372963905f, 0.296968699f, -1.f),
    C_Vertex3f(0.376606584f, 0.489439368f, -1.f),
    C_Vertex3f(0.3766675f, 0.279743314f, -1.f),
    C_Vertex3f(0.389730811f, 0.468996644f, -1.f),
    C_Vertex3f(0.379136562f, 0.261285186f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36238867f, 0.509800136f, -1.f),
    C_Vertex3f(0.347077072f, 0.530078948f, -1.f),
    C_Vertex3f(0.354445934f, 0.341248393f, -1.f),
    C_Vertex3f(0.361853123f, 0.327721238f, -1.f),
    C_Vertex3f(0.36802578f, 0.31296134f, -1.f),
    C_Vertex3f(0.376606584f, 0.489439368f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.322319031f, 0.437980652f, -1.f),
    C_Vertex3f(0.325485706f, 0.374283433f, -1.f),
    C_Vertex3f(0.330671787f, 0.550275803f, -1.f),
    C_Vertex3f(0.335927963f, 0.364604473f, -1.f),
    C_Vertex3f(0.347077072f, 0.530078948f, -1.f),
    C_Vertex3f(0.345804214f, 0.353542805f, -1.f),
    C_Vertex3f(0.354445934f, 0.341248393f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.455351949f, 0.30250442f, -1.f),
    C_Vertex3f(0.456217051f, 0.163228631f, -1.f),
    C_Vertex3f(0.458633006f, 0.281324089f, -1.f),
    C_Vertex3f(0.460015059f, 0.187974811f, -1.f),
    C_Vertex3f(0.460820377f, 0.2600618f, -1.f),
    C_Vertex3f(0.461914063f, 0.214176178f, -1.f),
    C_Vertex3f(0.461914063f, 0.238717556f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.438946664f, 0.365553677f, -1.f),
    C_Vertex3f(0.442924023f, 0.118101835f, -1.f),
    C_Vertex3f(0.445508778f, 0.344619215f, -1.f),
    C_Vertex3f(0.450520039f, 0.139937639f, -1.f),
    C_Vertex3f(0.450977206f, 0.323602796f, -1.f),
    C_Vertex3f(0.456217051f, 0.163228631f, -1.f),
    C_Vertex3f(0.455351949f, 0.30250442f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.380371094f, 0.241594315f, -1.f),
    C_Vertex3f(0.380371094f, 0.220233917f, -1.f),
    C_Vertex3f(0.393549919f, 0.0453104973f, -1.f),
    C_Vertex3f(0.379150271f, 0.199751496f, -1.f),
    C_Vertex3f(0.377319813f, 0.0312973261f, -1.f),
    C_Vertex3f(0.376708627f, 0.180583835f, -1.f),
    C_Vertex3f(0.37304616f, 0.162730932f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.37304616f, 0.162730932f, -1.f),
    C_Vertex3f(0.36816287f, 0.146192789f, -1.f),
    C_Vertex3f(0.377319813f, 0.0312973261f, -1.f),
    C_Vertex3f(0.362058759f, 0.130969405f, -1.f),
    C_Vertex3f(0.360051632f, 0.0192860365f, -1.f),
    C_Vertex3f(0.354733825f, 0.117060781f, -1.f),
    C_Vertex3f(0.346188068f, 0.104466915f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.341745377f, 0.00927662849f, -1.f),
    C_Vertex3f(0.360051632f, 0.0192860365f, -1.f),
    C_Vertex3f(0.346188068f, 0.104466915f, -1.f),
    C_Vertex3f(0.33642149f, 0.093187809f, -1.f),
    C_Vertex3f(0.326129556f, 0.0833185911f, -1.f),
    C_Vertex3f(0.322401047f, 0.0012691021f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.326129556f, 0.0833185911f, -1.f),
    C_Vertex3f(0.315312266f, 0.0748592615f, -1.f),
    C_Vertex3f(0.322401047f, 0.0012691021f, -1.f),
    C_Vertex3f(0.303969622f, 0.0678098202f, -1.f),
    C_Vertex3f(0.302018642f, -0.0047365427f, -1.f),
    C_Vertex3f(0.292101622f, 0.0621702671f, -1.f),
    C_Vertex3f(0.280598164f, -0.0087403059f, -1.f),
    C_Vertex3f(0.279708266f, 0.0579406023f, -1.f),
    C_Vertex3f(0.266789556f, 0.0551208258f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25813961f, -0.0107421875f, -1.f),
    C_Vertex3f(0.280598164f, -0.0087403059f, -1.f),
    C_Vertex3f(0.266789556f, 0.0551208258f, -1.f),
    C_Vertex3f(0.25334549f, 0.0537109375f, -1.f),
    C_Vertex3f(0.23921299f, 0.0537109375f, -1.f),
    C_Vertex3f(0.232456207f, -0.0107421875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.23921299f, 0.0537109375f, -1.f),
    C_Vertex3f(0.225510836f, 0.0550459623f, -1.f),
    C_Vertex3f(0.232456207f, -0.0107421875f, -1.f),
    C_Vertex3f(0.212402105f, 0.057716012f, -1.f),
    C_Vertex3f(0.206116199f, -0.00809526443f, -1.f),
    C_Vertex3f(0.199886799f, 0.0617210865f, -1.f),
    C_Vertex3f(0.187964916f, 0.0670611858f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.181306362f, -0.0028014183f, -1.f),
    C_Vertex3f(0.206116199f, -0.00809526443f, -1.f),
    C_Vertex3f(0.187964916f, 0.0670611858f, -1.f),
    C_Vertex3f(0.176636457f, 0.07373631f, -1.f),
    C_Vertex3f(0.165901423f, 0.081746459f, -1.f),
    C_Vertex3f(0.158026695f, 0.00513935089f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158026695f, 0.00513935089f, -1.f),
    C_Vertex3f(0.165901423f, 0.081746459f, -1.f),
    C_Vertex3f(0.155759811f, 0.0910916328f, -1.f),
    C_Vertex3f(0.146211624f, 0.101771832f, -1.f),
    C_Vertex3f(0.13785696f, 0.113759637f, -1.f),
    C_Vertex3f(0.136277199f, 0.0157270432f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.136277199f, 0.0157270432f, -1.f),
    C_Vertex3f(0.13785696f, 0.113759637f, -1.f),
    C_Vertex3f(0.13069582f, 0.127055049f, -1.f),
    C_Vertex3f(0.124728203f, 0.141658068f, -1.f),
    C_Vertex3f(0.119954109f, 0.157568693f, -1.f),
    C_Vertex3f(0.116373539f, 0.174786925f, -1.f),
    C_Vertex3f(0.116057873f, 0.0289616585f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.116373539f, 0.174786925f, -1.f),
    C_Vertex3f(0.113986492f, 0.193312764f, -1.f),
    C_Vertex3f(0.116057873f, 0.0289616585f, -1.f),
    C_Vertex3f(0.112792969f, 0.21314621f, -1.f),
    C_Vertex3f(0.0973687172f, 0.0448431969f, -1.f),
    C_Vertex3f(0.110251069f, 0.423179507f, -1.f),
    C_Vertex3f(0.0952486992f, 0.409616947f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.058419466f, 0.0960383415f, -1.f),
    C_Vertex3f(0.0673131943f, 0.0808916092f, -1.f),
    C_Vertex3f(0.068786025f, 0.37699306f, -1.f),
    C_Vertex3f(0.0802097321f, 0.0633716583f, -1.f),
    C_Vertex3f(0.081135273f, 0.394116879f, -1.f),
    C_Vertex3f(0.0973687172f, 0.0448431969f, -1.f),
    C_Vertex3f(0.0952486992f, 0.409616947f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0317382813f, 0.240151405f, -1.f),
    C_Vertex3f(0.0330088139f, 0.19100523f, -1.f),
    C_Vertex3f(0.0335024595f, 0.267017722f, -1.f),
    C_Vertex3f(0.0355498791f, 0.169447422f, -1.f),
    C_Vertex3f(0.0370308161f, 0.292260289f, -1.f),
    C_Vertex3f(0.0393614769f, 0.149171829f, -1.f),
    C_Vertex3f(0.0423233509f, 0.315879107f, -1.f),
    C_Vertex3f(0.0444436073f, 0.130178452f, -1.f),
    C_Vertex3f(0.049380064f, 0.337874174f, -1.f),
    C_Vertex3f(0.0507962704f, 0.112467289f, -1.f),
    C_Vertex3f(0.0582009554f, 0.358245492f, -1.f),
    C_Vertex3f(0.058419466f, 0.0960383415f, -1.f),
    C_Vertex3f(0.068786025f, 0.37699306f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.315097809f, 0.441820145f, -1.f),
    C_Vertex3f(0.301859379f, 0.448237419f, -1.f),
    C_Vertex3f(0.302903175f, 0.389493227f, -1.f),
    C_Vertex3f(0.314477444f, 0.382579684f, -1.f),
    C_Vertex3f(0.325485706f, 0.374283433f, -1.f),
    C_Vertex3f(0.322319031f, 0.437980652f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.290762901f, 0.395024061f, -1.f),
    C_Vertex3f(0.302903175f, 0.389493227f, -1.f),
    C_Vertex3f(0.301859379f, 0.448237419f, -1.f),
    C_Vertex3f(0.289824963f, 0.453392982f, -1.f),
    C_Vertex3f(0.27899456f, 0.457286835f, -1.f),
    C_Vertex3f(0.278056622f, 0.399172187f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.186042309f, 0.389493227f, -1.f),
    C_Vertex3f(0.197773457f, 0.395024061f, -1.f),
    C_Vertex3f(0.198597074f, 0.461929679f, -1.f),
    C_Vertex3f(0.210118532f, 0.399172187f, -1.f),
    C_Vertex3f(0.218933105f, 0.463867188f, -1.f),
    C_Vertex3f(0.223077536f, 0.401937604f, -1.f),
    C_Vertex3f(0.235891342f, 0.463867188f, -1.f),
    C_Vertex3f(0.236650467f, 0.403320313f, -1.f),
    C_Vertex3f(0.247816563f, 0.463209152f, -1.f),
    C_Vertex3f(0.250946045f, 0.403320313f, -1.f),
    C_Vertex3f(0.258975506f, 0.461893082f, -1.f),
    C_Vertex3f(0.264784336f, 0.401937604f, -1.f),
    C_Vertex3f(0.269368172f, 0.459918976f, -1.f),
    C_Vertex3f(0.278056622f, 0.399172187f, -1.f),
    C_Vertex3f(0.27899456f, 0.457286835f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.179149985f, 0.458054662f, -1.f),
    C_Vertex3f(0.160591841f, 0.452242136f, -1.f),
    C_Vertex3f(0.164421797f, 0.374283433f, -1.f),
    C_Vertex3f(0.174925089f, 0.382579684f, -1.f),
    C_Vertex3f(0.186042309f, 0.389493227f, -1.f),
    C_Vertex3f(0.198597074f, 0.461929679f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.130184412f, 0.32702601f, -1.f),
    C_Vertex3f(0.137140989f, 0.34101665f, -1.f),
    C_Vertex3f(0.14292264f, 0.444492102f, -1.f),
    C_Vertex3f(0.145256996f, 0.353542805f, -1.f),
    C_Vertex3f(0.160591841f, 0.452242136f, -1.f),
    C_Vertex3f(0.154532433f, 0.364604473f, -1.f),
    C_Vertex3f(0.164421797f, 0.374283433f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.126142383f, 0.434804559f, -1.f),
    C_Vertex3f(0.110251069f, 0.423179507f, -1.f),
    C_Vertex3f(0.113952398f, 0.256418586f, -1.f),
    C_Vertex3f(0.116271257f, 0.276267171f, -1.f),
    C_Vertex3f(0.119749546f, 0.29465127f, -1.f),
    C_Vertex3f(0.124387264f, 0.311570883f, -1.f),
    C_Vertex3f(0.130184412f, 0.32702601f, -1.f),
    C_Vertex3f(0.14292264f, 0.444492102f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.112792969f, 0.235105515f, -1.f),
    C_Vertex3f(0.113952398f, 0.256418586f, -1.f),
    C_Vertex3f(0.110251069f, 0.423179507f, -1.f),
    C_Vertex3f(0.112792969f, 0.21314621f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0330088139f, 0.19100523f, -1.f),
    C_Vertex3f(0.0317382813f, 0.240151405f, -1.f),
    C_Vertex3f(0.0317382813f, 0.213845253f, -1.f),
    C_Vertex3f(0.438946664f, 0.365553677f, -1.f),
    C_Vertex3f(0.433429003f, 0.0977212191f, -1.f),
    C_Vertex3f(0.442924023f, 0.118101835f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_241[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.129882813f, 0.464355469f, -1.f),
    C_Vertex3f(0.129882813f, 0.464355469f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.129882813f, 0.464355469f, 0.f),
    C_Vertex3f(0.129882813f, 0.464355469f, -1.f),
    C_Vertex3f(0.129882813f, 0.398361206f, -1.f),
    C_Vertex3f(0.129882813f, 0.398361206f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.830692351f, 0.556731761f, 0.f),
    C_Vertex3f(0.129882813f, 0.398361206f, 0.f),
    C_Vertex3f(0.129882813f, 0.398361206f, -1.f),
    C_Normal3f(-0.786487639f, 0.617606044f, 0.f),
    C_Vertex3f(0.136270523f, 0.407892227f, 0.f),
    C_Vertex3f(0.136270523f, 0.407892227f, -1.f),
    C_Normal3f(-0.720558703f, 0.693393946f, 0.f),
    C_Vertex3f(0.150465012f, 0.424571514f, 0.f),
    C_Vertex3f(0.150465012f, 0.424571514f, -1.f),
    C_Normal3f(-0.626053929f, 0.779779792f, 0.f),
    C_Vertex3f(0.166078568f, 0.438868046f, 0.f),
    C_Vertex3f(0.166078568f, 0.438868046f, -1.f),
    C_Normal3f(-0.517227948f, 0.855847657f, 0.f),
    C_Vertex3f(0.183111191f, 0.450781822f, 0.f),
    C_Vertex3f(0.183111191f, 0.450781822f, -1.f),
    C_Normal3f(-0.39907527f, 0.916918218f, 0.f),
    C_Vertex3f(0.201562881f, 0.460312843f, 0.f),
    C_Vertex3f(0.201562881f, 0.460312843f, -1.f),
    C_Normal3f(-0.278033763f, 0.960571289f, 0.f),
    C_Vertex3f(0.22143364f, 0.467461109f, 0.f),
    C_Vertex3f(0.22143364f, 0.467461109f, -1.f),
    C_Normal3f(-0.16036272f, 0.987058163f, 0.f),
    C_Vertex3f(0.242723465f, 0.47222662f, 0.f),
    C_Vertex3f(0.242723465f, 0.47222662f, -1.f),
    C_Normal3f(-0.0525781661f, 0.998616874f, 0.f),
    C_Vertex3f(0.265432358f, 0.474609375f, 0.f),
    C_Vertex3f(0.265432358f, 0.474609375f, -1.f),
    C_Normal3f(0.0441252515f, 0.99902606f, 0.f),
    C_Vertex3f(0.287979126f, 0.474609375f, 0.f),
    C_Vertex3f(0.287979126f, 0.474609375f, -1.f),
    C_Normal3f(0.142065495f, 0.989857197f, 0.f),
    C_Vertex3f(0.308097839f, 0.472724915f, 0.f),
    C_Vertex3f(0.308097839f, 0.472724915f, -1.f),
    C_Normal3f(0.242489442f, 0.970154107f, 0.f),
    C_Vertex3f(0.32736969f, 0.468955994f, 0.f),
    C_Vertex3f(0.32736969f, 0.468955994f, -1.f),
    C_Normal3f(0.344026625f, 0.938959897f, 0.f),
    C_Vertex3f(0.345794678f, 0.463302612f, 0.f),
    C_Vertex3f(0.345794678f, 0.463302612f, -1.f),
    C_Normal3f(0.442368001f, 0.896833658f, 0.f),
    C_Vertex3f(0.363372803f, 0.455764771f, 0.f),
    C_Vertex3f(0.363372803f, 0.455764771f, -1.f),
    C_Normal3f(0.546218216f, 0.837642908f, 0.f),
    C_Vertex3f(0.378740311f, 0.447052002f, 0.f),
    C_Vertex3f(0.378740311f, 0.447052002f, -1.f),
    C_Normal3f(0.656018972f, 0.754744351f, 0.f),
    C_Vertex3f(0.391897202f, 0.437164307f, 0.f),
    C_Vertex3f(0.391897202f, 0.437164307f, -1.f),
    C_Normal3f(0.763929188f, 0.645300031f, 0.f),
    C_Vertex3f(0.402843475f, 0.426101685f, 0.f),
    C_Vertex3f(0.402843475f, 0.426101685f, -1.f),
    C_Normal3f(0.845810473f, 0.533483624f, 0.f),
    C_Vertex3f(0.411579132f, 0.413864136f, 0.f),
    C_Vertex3f(0.411579132f, 0.413864136f, -1.f),
    C_Normal3f(0.899753392f, 0.436398685f, 0.f),
    C_Vertex3f(0.41900444f, 0.400479317f, 0.f),
    C_Vertex3f(0.41900444f, 0.400479317f, -1.f),
    C_Normal3f(0.940455556f, 0.339916557f, 0.f),
    C_Vertex3f(0.4251194f, 0.385947227f, 0.f),
    C_Vertex3f(0.4251194f, 0.385947227f, -1.f),
    C_Normal3f(0.966733932f, 0.255783916f, 0.f),
    C_Vertex3f(0.429924011f, 0.370267868f, 0.f),
    C_Vertex3f(0.429924011f, 0.370267868f, -1.f),
    C_Normal3f(0.983459711f, 0.181127235f, 0.f),
    C_Vertex3f(0.432216644f, 0.35912323f, 0.f),
    C_Vertex3f(0.432216644f, 0.35912323f, -1.f),
    C_Normal3f(0.992074847f, 0.125648394f, 0.f),
    C_Vertex3f(0.433171272f, 0.352636337f, 0.f),
    C_Vertex3f(0.433171272f, 0.352636337f, -1.f),
    C_Normal3f(0.995658576f, 0.0930804089f, 0.f),
    C_Vertex3f(0.433989525f, 0.345125198f, 0.f),
    C_Vertex3f(0.433989525f, 0.345125198f, -1.f),
    C_Normal3f(0.997707486f, 0.06767416f, 0.f),
    C_Vertex3f(0.434671402f, 0.336589813f, 0.f),
    C_Vertex3f(0.434671402f, 0.336589813f, -1.f),
    C_Normal3f(0.998878896f, 0.0473382138f, 0.f),
    C_Vertex3f(0.435216904f, 0.327030182f, 0.f),
    C_Vertex3f(0.435216904f, 0.327030182f, -1.f),
    C_Normal3f(0.999528348f, 0.0307117645f, 0.f),
    C_Vertex3f(0.43562603f, 0.316446304f, 0.f),
    C_Vertex3f(0.43562603f, 0.316446304f, -1.f),
    C_Normal3f(0.999857664f, 0.0168754011f, 0.f),
    C_Vertex3f(0.435898781f, 0.304838181f, 0.f),
    C_Vertex3f(0.435898781f, 0.304838181f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.436035156f, 0.292205811f, 0.f),
    C_Vertex3f(0.436035156f, 0.292205811f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.436035156f, 0.292205811f, 0.f),
    C_Vertex3f(0.436035156f, 0.292205811f, -1.f),
    C_Vertex3f(0.436035156f, 0.f, -1.f),
    C_Vertex3f(0.436035156f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.436035156f, 0.f, 0.f),
    C_Vertex3f(0.436035156f, 0.f, -1.f),
    C_Vertex3f(0.357421875f, 0.f, -1.f),
    C_Vertex3f(0.357421875f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.357421875f, 0.f, 0.f),
    C_Vertex3f(0.357421875f, 0.f, -1.f),
    C_Normal3f(-0.999993324f, -0.00363833085f, 0.f),
    C_Vertex3f(0.357421875f, 0.294200897f, 0.f),
    C_Vertex3f(0.357421875f, 0.294200897f, -1.f),
    C_Normal3f(-0.996123493f, -0.0879653245f, 0.f),
    C_Vertex3f(0.356275082f, 0.31519556f, 0.f),
    C_Vertex3f(0.356275082f, 0.31519556f, -1.f),
    C_Normal3f(-0.98514843f, -0.171704516f, 0.f),
    C_Vertex3f(0.353981495f, 0.333159924f, 0.f),
    C_Vertex3f(0.353981495f, 0.333159924f, -1.f),
    C_Normal3f(-0.958059728f, -0.286568522f, 0.f),
    C_Vertex3f(0.350541115f, 0.348093987f, 0.f),
    C_Vertex3f(0.350541115f, 0.348093987f, -1.f),
    C_Normal3f(-0.900001645f, -0.435886383f, 0.f),
    C_Vertex3f(0.345953941f, 0.359997749f, 0.f),
    C_Vertex3f(0.345953941f, 0.359997749f, -1.f),
    C_Normal3f(-0.812479973f, -0.582989037f, 0.f),
    C_Vertex3f(0.339592457f, 0.370700359f, 0.f),
    C_Vertex3f(0.339592457f, 0.370700359f, -1.f),
    C_Normal3f(-0.70227474f, -0.711906016f, 0.f),
    C_Vertex3f(0.331456661f, 0.380201817f, 0.f),
    C_Vertex3f(0.331456661f, 0.380201817f, -1.f),
    C_Normal3f(-0.580608785f, -0.814182699f, 0.f),
    C_Vertex3f(0.321546555f, 0.388502121f, 0.f),
    C_Vertex3f(0.321546555f, 0.388502121f, -1.f),
    C_Normal3f(-0.454031795f, -0.890985548f, 0.f),
    C_Vertex3f(0.309862137f, 0.395601273f, 0.f),
    C_Vertex3f(0.309862137f, 0.395601273f, -1.f),
    C_Normal3f(-0.318599582f, -0.947889447f, 0.f),
    C_Vertex3f(0.297166824f, 0.400925636f, 0.f),
    C_Vertex3f(0.297166824f, 0.400925636f, -1.f),
    C_Normal3f(-0.184121281f, -0.98290354f, 0.f),
    C_Vertex3f(0.283460617f, 0.404475212f, 0.f),
    C_Vertex3f(0.283460617f, 0.404475212f, -1.f),
    C_Normal3f(-0.0613803118f, -0.998114467f, 0.f),
    C_Vertex3f(0.268743515f, 0.40625f, 0.f),
    C_Vertex3f(0.268743515f, 0.40625f, -1.f),
    C_Normal3f(0.0379030257f, -0.999281406f, 0.f),
    C_Vertex3f(0.254600525f, 0.40625f, 0.f),
    C_Vertex3f(0.254600525f, 0.40625f, -1.f),
    C_Normal3f(0.124490522f, -0.992220819f, 0.f),
    C_Vertex3f(0.24247241f, 0.40525353f, 0.f),
    C_Vertex3f(0.24247241f, 0.40525353f, -1.f),
    C_Normal3f(0.212007388f, -0.9772681f, 0.f),
    C_Vertex3f(0.230774164f, 0.403260589f, 0.f),
    C_Vertex3f(0.230774164f, 0.403260589f, -1.f),
    C_Normal3f(0.300902724f, -0.953654885f, 0.f),
    C_Vertex3f(0.219505787f, 0.400271177f, 0.f),
    C_Vertex3f(0.219505787f, 0.400271177f, -1.f),
    C_Normal3f(0.388869584f, -0.921292841f, 0.f),
    C_Vertex3f(0.208667278f, 0.396285295f, 0.f),
    C_Vertex3f(0.208667278f, 0.396285295f, -1.f),
    C_Normal3f(0.473541707f, -0.880771399f, 0.f),
    C_Vertex3f(0.198258638f, 0.391302943f, 0.f),
    C_Vertex3f(0.198258638f, 0.391302943f, -1.f),
    C_Normal3f(0.552798092f, -0.833315134f, 0.f),
    C_Vertex3f(0.188279867f, 0.385324121f, 0.f),
    C_Vertex3f(0.188279867f, 0.385324121f, -1.f),
    C_Normal3f(0.625019014f, -0.780609548f, 0.f),
    C_Vertex3f(0.178730965f, 0.378348827f, 0.f),
    C_Vertex3f(0.178730965f, 0.378348827f, -1.f),
    C_Normal3f(0.719850719f, -0.694128811f, 0.f),
    C_Vertex3f(0.169611931f, 0.370377064f, 0.f),
    C_Vertex3f(0.169611931f, 0.370377064f, -1.f),
    C_Normal3f(0.820990741f, -0.570941508f, 0.f),
    C_Vertex3f(0.161632776f, 0.360617042f, 0.f),
    C_Vertex3f(0.161632776f, 0.360617042f, -1.f),
    C_Normal3f(0.893041074f, -0.449975073f, 0.f),
    C_Vertex3f(0.154793501f, 0.349068761f, 0.f),
    C_Vertex3f(0.154793501f, 0.349068761f, -1.f),
    C_Normal3f(0.940752149f, -0.339094907f, 0.f),
    C_Vertex3f(0.149094105f, 0.335732222f, 0.f),
    C_Vertex3f(0.149094105f, 0.335732222f, -1.f),
    C_Normal3f(0.970358133f, -0.241671532f, 0.f),
    C_Vertex3f(0.144534588f, 0.320607424f, 0.f),
    C_Vertex3f(0.144534588f, 0.320607424f, -1.f),
    C_Normal3f(0.987435877f, -0.158020154f, 0.f),
    C_Vertex3f(0.14111495f, 0.303694367f, 0.f),
    C_Vertex3f(0.14111495f, 0.303694367f, -1.f),
    C_Normal3f(0.996214747f, -0.0869256482f, 0.f),
    C_Vertex3f(0.138835192f, 0.284993052f, 0.f),
    C_Vertex3f(0.138835192f, 0.284993052f, -1.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.264503479f, 0.f),
    C_Vertex3f(0.137695313f, 0.264503479f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.264503479f, 0.f),
    C_Vertex3f(0.137695313f, 0.264503479f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999953806f, -0.00961818546f, 0.f),
    C_Vertex3f(0.11328125f, 0.520507813f, 0.f),
    C_Vertex3f(0.11328125f, 0.520507813f, -1.f),
    C_Normal3f(-0.999230683f, 0.0392165966f, 0.f),
    C_Vertex3f(0.113226891f, 0.526159286f, 0.f),
    C_Vertex3f(0.113226891f, 0.526159286f, -1.f),
    C_Normal3f(-0.994498312f, 0.104752719f, 0.f),
    C_Vertex3f(0.113924265f, 0.536891699f, 0.f),
    C_Vertex3f(0.113924265f, 0.536891699f, -1.f),
    C_Normal3f(-0.981874108f, 0.189534381f, 0.f),
    C_Vertex3f(0.115427732f, 0.547053576f, 0.f),
    C_Vertex3f(0.115427732f, 0.547053576f, -1.f),
    C_Normal3f(-0.960046053f, 0.279841781f, 0.f),
    C_Vertex3f(0.117737293f, 0.556644917f, 0.f),
    C_Vertex3f(0.117737293f, 0.556644917f, -1.f),
    C_Normal3f(-0.927577436f, 0.373630971f, 0.f),
    C_Vertex3f(0.120852947f, 0.565665722f, 0.f),
    C_Vertex3f(0.120852947f, 0.565665722f, -1.f),
    C_Normal3f(-0.883691847f, 0.468069077f, 0.f),
    C_Vertex3f(0.124774694f, 0.574115992f, 0.f),
    C_Vertex3f(0.124774694f, 0.574115992f, -1.f),
    C_Normal3f(-0.828614712f, 0.559819281f, 0.f),
    C_Vertex3f(0.129502535f, 0.581995726f, 0.f),
    C_Vertex3f(0.129502535f, 0.581995726f, -1.f),
    C_Normal3f(-0.759081304f, 0.650995851f, 0.f),
    C_Vertex3f(0.135036469f, 0.589304924f, 0.f),
    C_Vertex3f(0.135036469f, 0.589304924f, -1.f),
    C_Normal3f(-0.650275886f, 0.759698212f, 0.f),
    C_Vertex3f(0.144439697f, 0.599412918f, 0.f),
    C_Vertex3f(0.144439697f, 0.599412918f, -1.f),
    C_Normal3f(-0.487445921f, 0.87315321f, 0.f),
    C_Vertex3f(0.158654213f, 0.609520912f, 0.f),
    C_Vertex3f(0.158654213f, 0.609520912f, -1.f),
    C_Normal3f(-0.287504673f, 0.957779288f, 0.f),
    C_Vertex3f(0.174616814f, 0.616259575f, 0.f),
    C_Vertex3f(0.174616814f, 0.616259575f, -1.f),
    C_Normal3f(-0.0979905054f, 0.995187461f, 0.f),
    C_Vertex3f(0.192327499f, 0.619628906f, 0.f),
    C_Vertex3f(0.192327499f, 0.619628906f, -1.f),
    C_Normal3f(0.0889112428f, 0.99603951f, 0.f),
    C_Vertex3f(0.208835602f, 0.619628906f, 0.f),
    C_Vertex3f(0.208835602f, 0.619628906f, -1.f),
    C_Normal3f(0.244741365f, 0.969588399f, 0.f),
    C_Vertex3f(0.224859238f, 0.616724968f, 0.f),
    C_Vertex3f(0.224859238f, 0.616724968f, -1.f),
    C_Normal3f(0.345433027f, 0.938443422f, 0.f),
    C_Vertex3f(0.243349075f, 0.610917091f, 0.f),
    C_Vertex3f(0.243349075f, 0.610917091f, -1.f),
    C_Normal3f(0.413061887f, 0.910703003f, 0.f),
    C_Vertex3f(0.264305115f, 0.602205276f, 0.f),
    C_Vertex3f(0.264305115f, 0.602205276f, -1.f),
    C_Normal3f(0.42238301f, 0.906417429f, 0.f),
    C_Vertex3f(0.28247261f, 0.593172073f, 0.f),
    C_Vertex3f(0.28247261f, 0.593172073f, -1.f),
    C_Normal3f(0.345709145f, 0.938341737f, 0.f),
    C_Vertex3f(0.294072151f, 0.588334084f, 0.f),
    C_Vertex3f(0.294072151f, 0.588334084f, -1.f),
    C_Normal3f(0.243631572f, 0.969867826f, 0.f),
    C_Vertex3f(0.304358482f, 0.585108757f, 0.f),
    C_Vertex3f(0.304358482f, 0.585108757f, -1.f),
    C_Normal3f(0.0984259844f, 0.995144367f, 0.f),
    C_Vertex3f(0.313331604f, 0.583496094f, 0.f),
    C_Vertex3f(0.313331604f, 0.583496094f, -1.f),
    C_Normal3f(-0.070669286f, 0.997499824f, 0.f),
    C_Vertex3f(0.320663452f, 0.583496094f, 0.f),
    C_Vertex3f(0.320663452f, 0.583496094f, -1.f),
    C_Normal3f(-0.24143213f, 0.970417738f, 0.f),
    C_Vertex3f(0.3268466f, 0.584453583f, 0.f),
    C_Vertex3f(0.3268466f, 0.584453583f, -1.f),
    C_Normal3f(-0.435192645f, 0.900337279f, 0.f),
    C_Vertex3f(0.33220911f, 0.586368561f, 0.f),
    C_Vertex3f(0.33220911f, 0.586368561f, -1.f),
    C_Normal3f(-0.62994957f, 0.776636064f, 0.f),
    C_Vertex3f(0.336750984f, 0.589241028f, 0.f),
    C_Vertex3f(0.336750984f, 0.589241028f, -1.f),
    C_Normal3f(-0.810694158f, 0.585469842f, 0.f),
    C_Vertex3f(0.340472221f, 0.593070984f, 0.f),
    C_Vertex3f(0.340472221f, 0.593070984f, -1.f),
    C_Normal3f(-0.9168607f, 0.399207145f, 0.f),
    C_Vertex3f(0.343646526f, 0.598789215f, 0.f),
    C_Vertex3f(0.343646526f, 0.598789215f, -1.f),
    C_Normal3f(-0.964133918f, 0.265416414f, 0.f),
    C_Vertex3f(0.346273899f, 0.606395721f, 0.f),
    C_Vertex3f(0.346273899f, 0.606395721f, -1.f),
    C_Normal3f(-0.982875049f, 0.18427363f, 0.f),
    C_Vertex3f(0.34835434f, 0.615890503f, 0.f),
    C_Vertex3f(0.34835434f, 0.615890503f, -1.f),
    C_Normal3f(-0.991047204f, 0.133512393f, 0.f),
    C_Vertex3f(0.349121094f, 0.621582031f, 0.f),
    C_Vertex3f(0.349121094f, 0.621582031f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.349121094f, 0.621582031f, 0.f),
    C_Vertex3f(0.349121094f, 0.621582031f, -1.f),
    C_Vertex3f(0.405761719f, 0.621582031f, -1.f),
    C_Vertex3f(0.405761719f, 0.621582031f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999653041f, -0.0263388976f, 0.f),
    C_Vertex3f(0.405761719f, 0.621582031f, 0.f),
    C_Vertex3f(0.405761719f, 0.621582031f, -1.f),
    C_Normal3f(0.99777931f, -0.066607669f, 0.f),
    C_Vertex3f(0.405597687f, 0.615356445f, 0.f),
    C_Vertex3f(0.405597687f, 0.615356445f, -1.f),
    C_Normal3f(0.992581308f, -0.121582195f, 0.f),
    C_Vertex3f(0.404566884f, 0.603683472f, 0.f),
    C_Vertex3f(0.404566884f, 0.603683472f, -1.f),
    C_Normal3f(0.980870843f, -0.194659367f, 0.f),
    C_Vertex3f(0.402833343f, 0.592788696f, 0.f),
    C_Vertex3f(0.402833343f, 0.592788696f, -1.f),
    C_Normal3f(0.961305082f, -0.275485754f, 0.f),
    C_Vertex3f(0.400397062f, 0.582672119f, 0.f),
    C_Vertex3f(0.400397062f, 0.582672119f, -1.f),
    C_Normal3f(0.931648552f, -0.363360554f, 0.f),
    C_Vertex3f(0.397258043f, 0.57333374f, 0.f),
    C_Vertex3f(0.397258043f, 0.57333374f, -1.f),
    C_Normal3f(0.889692426f, -0.456560522f, 0.f),
    C_Vertex3f(0.393416286f, 0.56477356f, 0.f),
    C_Vertex3f(0.393416286f, 0.56477356f, -1.f),
    C_Normal3f(0.833749712f, -0.552142739f, 0.f),
    C_Vertex3f(0.388871789f, 0.556991577f, 0.f),
    C_Vertex3f(0.388871789f, 0.556991577f, -1.f),
    C_Normal3f(0.755483389f, -0.655167818f, 0.f),
    C_Vertex3f(0.383624554f, 0.549987793f, 0.f),
    C_Vertex3f(0.383624554f, 0.549987793f, -1.f),
    C_Normal3f(0.639779389f, -0.768558562f, 0.f),
    C_Vertex3f(0.374699593f, 0.540649414f, 0.f),
    C_Vertex3f(0.374699593f, 0.540649414f, -1.f),
    C_Normal3f(0.477566123f, -0.878595829f, 0.f),
    C_Vertex3f(0.361188412f, 0.531311035f, 0.f),
    C_Vertex3f(0.361188412f, 0.531311035f, -1.f),
    C_Normal3f(0.281305313f, -0.95961827f, 0.f),
    C_Vertex3f(0.346065998f, 0.525085449f, 0.f),
    C_Vertex3f(0.346065998f, 0.525085449f, -1.f),
    C_Normal3f(0.0945234597f, -0.995522618f, 0.f),
    C_Vertex3f(0.329332352f, 0.521972656f, 0.f),
    C_Vertex3f(0.329332352f, 0.521972656f, -1.f),
    C_Normal3f(-0.0947503f, -0.995501101f, 0.f),
    C_Vertex3f(0.313282013f, 0.521972656f, 0.f),
    C_Vertex3f(0.313282013f, 0.521972656f, -1.f),
    C_Normal3f(-0.258822501f, -0.965924919f, 0.f),
    C_Vertex3f(0.297328949f, 0.525018692f, 0.f),
    C_Vertex3f(0.297328949f, 0.525018692f, -1.f),
    C_Normal3f(-0.367869198f, -0.92987752f, 0.f),
    C_Vertex3f(0.279178619f, 0.531110764f, 0.f),
    C_Vertex3f(0.279178619f, 0.531110764f, -1.f),
    C_Normal3f(-0.441156626f, -0.897430122f, 0.f),
    C_Vertex3f(0.258831024f, 0.540248871f, 0.f),
    C_Vertex3f(0.258831024f, 0.540248871f, -1.f),
    C_Normal3f(-0.452606946f, -0.891710103f, 0.f),
    C_Vertex3f(0.240226746f, 0.550258636f, 0.f),
    C_Vertex3f(0.240226746f, 0.550258636f, -1.f),
    C_Normal3f(-0.385828793f, -0.922570407f, 0.f),
    C_Vertex3f(0.227532387f, 0.556135178f, 0.f),
    C_Vertex3f(0.227532387f, 0.556135178f, -1.f),
    C_Normal3f(-0.288805217f, -0.957387865f, 0.f),
    C_Vertex3f(0.216807365f, 0.560052872f, 0.f),
    C_Vertex3f(0.216807365f, 0.560052872f, -1.f),
    C_Normal3f(-0.122500464f, -0.992468476f, 0.f),
    C_Vertex3f(0.208051682f, 0.562011719f, 0.f),
    C_Vertex3f(0.208051682f, 0.562011719f, -1.f),
    C_Normal3f(0.0942201167f, -0.995551467f, 0.f),
    C_Vertex3f(0.200937271f, 0.562011719f, 0.f),
    C_Vertex3f(0.200937271f, 0.562011719f, -1.f),
    C_Normal3f(0.294618219f, -0.955615103f, 0.f),
    C_Vertex3f(0.194206238f, 0.56070137f, 0.f),
    C_Vertex3f(0.194206238f, 0.56070137f, -1.f),
    C_Normal3f(0.500668526f, -0.865639091f, 0.f),
    C_Vertex3f(0.188186646f, 0.558080673f, 0.f),
    C_Vertex3f(0.188186646f, 0.558080673f, -1.f),
    C_Normal3f(0.652620137f, -0.757685304f, 0.f),
    C_Vertex3f(0.182878494f, 0.554149628f, 0.f),
    C_Vertex3f(0.182878494f, 0.554149628f, -1.f),
    C_Normal3f(0.744249165f, -0.667902052f, 0.f),
    C_Vertex3f(0.180580139f, 0.551528931f, 0.f),
    C_Vertex3f(0.180580139f, 0.551528931f, -1.f),
    C_Normal3f(0.837037444f, -0.547145605f, 0.f),
    C_Vertex3f(0.178173065f, 0.548906326f, 0.f),
    C_Vertex3f(0.178173065f, 0.548906326f, -1.f),
    C_Normal3f(0.925027549f, -0.379899919f, 0.f),
    C_Vertex3f(0.174616814f, 0.542406082f, 0.f),
    C_Vertex3f(0.174616814f, 0.542406082f, -1.f),
    C_Normal3f(0.980741322f, -0.195311442f, 0.f),
    C_Vertex3f(0.172318459f, 0.534650803f, 0.f),
    C_Vertex3f(0.172318459f, 0.534650803f, -1.f),
    C_Normal3f(0.997837007f, -0.065737471f, 0.f),
    C_Vertex3f(0.171278f, 0.525640488f, 0.f),
    C_Vertex3f(0.171278f, 0.525640488f, -1.f),
    C_Normal3f(0.999775767f, 0.0211769659f, 0.f),
    C_Vertex3f(0.171386719f, 0.520507813f, 0.f),
    C_Vertex3f(0.171386719f, 0.520507813f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.171386719f, 0.520507813f, 0.f),
    C_Vertex3f(0.171386719f, 0.520507813f, -1.f),
    C_Vertex3f(0.11328125f, 0.520507813f, -1.f),
    C_Vertex3f(0.11328125f, 0.520507813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.363372803f, 0.455764771f, 0.f),
    C_Vertex3f(0.345794678f, 0.463302612f, 0.f),
    C_Vertex3f(0.345953941f, 0.359997749f, 0.f),
    C_Vertex3f(0.350541115f, 0.348093987f, 0.f),
    C_Vertex3f(0.353981495f, 0.333159924f, 0.f),
    C_Vertex3f(0.356275082f, 0.31519556f, 0.f),
    C_Vertex3f(0.357421875f, 0.294200897f, 0.f),
    C_Vertex3f(0.436035156f, 0.f, 0.f),
    C_Vertex3f(0.378740311f, 0.447052002f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.309862137f, 0.395601273f, 0.f),
    C_Vertex3f(0.321546555f, 0.388502121f, 0.f),
    C_Vertex3f(0.32736969f, 0.468955994f, 0.f),
    C_Vertex3f(0.331456661f, 0.380201817f, 0.f),
    C_Vertex3f(0.345794678f, 0.463302612f, 0.f),
    C_Vertex3f(0.339592457f, 0.370700359f, 0.f),
    C_Vertex3f(0.345953941f, 0.359997749f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268743515f, 0.40625f, 0.f),
    C_Vertex3f(0.283460617f, 0.404475212f, 0.f),
    C_Vertex3f(0.287979126f, 0.474609375f, 0.f),
    C_Vertex3f(0.297166824f, 0.400925636f, 0.f),
    C_Vertex3f(0.308097839f, 0.472724915f, 0.f),
    C_Vertex3f(0.309862137f, 0.395601273f, 0.f),
    C_Vertex3f(0.32736969f, 0.468955994f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230774164f, 0.403260589f, 0.f),
    C_Vertex3f(0.24247241f, 0.40525353f, 0.f),
    C_Vertex3f(0.242723465f, 0.47222662f, 0.f),
    C_Vertex3f(0.254600525f, 0.40625f, 0.f),
    C_Vertex3f(0.265432358f, 0.474609375f, 0.f),
    C_Vertex3f(0.268743515f, 0.40625f, 0.f),
    C_Vertex3f(0.287979126f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.22143364f, 0.467461109f, 0.f),
    C_Vertex3f(0.201562881f, 0.460312843f, 0.f),
    C_Vertex3f(0.208667278f, 0.396285295f, 0.f),
    C_Vertex3f(0.219505787f, 0.400271177f, 0.f),
    C_Vertex3f(0.230774164f, 0.403260589f, 0.f),
    C_Vertex3f(0.242723465f, 0.47222662f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.169611931f, 0.370377064f, 0.f),
    C_Vertex3f(0.178730965f, 0.378348827f, 0.f),
    C_Vertex3f(0.183111191f, 0.450781822f, 0.f),
    C_Vertex3f(0.188279867f, 0.385324121f, 0.f),
    C_Vertex3f(0.201562881f, 0.460312843f, 0.f),
    C_Vertex3f(0.198258638f, 0.391302943f, 0.f),
    C_Vertex3f(0.208667278f, 0.396285295f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.166078568f, 0.438868046f, 0.f),
    C_Vertex3f(0.150465012f, 0.424571514f, 0.f),
    C_Vertex3f(0.154793501f, 0.349068761f, 0.f),
    C_Vertex3f(0.161632776f, 0.360617042f, 0.f),
    C_Vertex3f(0.169611931f, 0.370377064f, 0.f),
    C_Vertex3f(0.183111191f, 0.450781822f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150465012f, 0.424571514f, 0.f),
    C_Vertex3f(0.136270523f, 0.407892227f, 0.f),
    C_Vertex3f(0.138835192f, 0.284993052f, 0.f),
    C_Vertex3f(0.14111495f, 0.303694367f, 0.f),
    C_Vertex3f(0.144534588f, 0.320607424f, 0.f),
    C_Vertex3f(0.149094105f, 0.335732222f, 0.f),
    C_Vertex3f(0.154793501f, 0.349068761f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129882813f, 0.464355469f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Vertex3f(0.129882813f, 0.398361206f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Vertex3f(0.136270523f, 0.407892227f, 0.f),
    C_Vertex3f(0.137695313f, 0.264503479f, 0.f),
    C_Vertex3f(0.138835192f, 0.284993052f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.436035156f, 0.f, 0.f),
    C_Vertex3f(0.436035156f, 0.292205811f, 0.f),
    C_Vertex3f(0.435898781f, 0.304838181f, 0.f),
    C_Vertex3f(0.43562603f, 0.316446304f, 0.f),
    C_Vertex3f(0.435216904f, 0.327030182f, 0.f),
    C_Vertex3f(0.434671402f, 0.336589813f, 0.f),
    C_Vertex3f(0.433989525f, 0.345125198f, 0.f),
    C_Vertex3f(0.433171272f, 0.352636337f, 0.f),
    C_Vertex3f(0.432216644f, 0.35912323f, 0.f),
    C_Vertex3f(0.429924011f, 0.370267868f, 0.f),
    C_Vertex3f(0.4251194f, 0.385947227f, 0.f),
    C_Vertex3f(0.41900444f, 0.400479317f, 0.f),
    C_Vertex3f(0.411579132f, 0.413864136f, 0.f),
    C_Vertex3f(0.402843475f, 0.426101685f, 0.f),
    C_Vertex3f(0.391897202f, 0.437164307f, 0.f),
    C_Vertex3f(0.378740311f, 0.447052002f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.171386719f, 0.520507813f, 0.f),
    C_Vertex3f(0.171278f, 0.525640488f, 0.f),
    C_Vertex3f(0.158654213f, 0.609520912f, 0.f),
    C_Vertex3f(0.144439697f, 0.599412918f, 0.f),
    C_Vertex3f(0.135036469f, 0.589304924f, 0.f),
    C_Vertex3f(0.129502535f, 0.581995726f, 0.f),
    C_Vertex3f(0.124774694f, 0.574115992f, 0.f),
    C_Vertex3f(0.120852947f, 0.565665722f, 0.f),
    C_Vertex3f(0.117737293f, 0.556644917f, 0.f),
    C_Vertex3f(0.115427732f, 0.547053576f, 0.f),
    C_Vertex3f(0.113924265f, 0.536891699f, 0.f),
    C_Vertex3f(0.11328125f, 0.520507813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.405761719f, 0.621582031f, 0.f),
    C_Vertex3f(0.349121094f, 0.621582031f, 0.f),
    C_Vertex3f(0.361188412f, 0.531311035f, 0.f),
    C_Vertex3f(0.374699593f, 0.540649414f, 0.f),
    C_Vertex3f(0.383624554f, 0.549987793f, 0.f),
    C_Vertex3f(0.388871789f, 0.556991577f, 0.f),
    C_Vertex3f(0.393416286f, 0.56477356f, 0.f),
    C_Vertex3f(0.397258043f, 0.57333374f, 0.f),
    C_Vertex3f(0.400397062f, 0.582672119f, 0.f),
    C_Vertex3f(0.402833343f, 0.592788696f, 0.f),
    C_Vertex3f(0.404566884f, 0.603683472f, 0.f),
    C_Vertex3f(0.405597687f, 0.615356445f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.349121094f, 0.621582031f, 0.f),
    C_Vertex3f(0.34835434f, 0.615890503f, 0.f),
    C_Vertex3f(0.361188412f, 0.531311035f, 0.f),
    C_Vertex3f(0.346273899f, 0.606395721f, 0.f),
    C_Vertex3f(0.346065998f, 0.525085449f, 0.f),
    C_Vertex3f(0.343646526f, 0.598789215f, 0.f),
    C_Vertex3f(0.340472221f, 0.593070984f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.340472221f, 0.593070984f, 0.f),
    C_Vertex3f(0.336750984f, 0.589241028f, 0.f),
    C_Vertex3f(0.346065998f, 0.525085449f, 0.f),
    C_Vertex3f(0.33220911f, 0.586368561f, 0.f),
    C_Vertex3f(0.329332352f, 0.521972656f, 0.f),
    C_Vertex3f(0.3268466f, 0.584453583f, 0.f),
    C_Vertex3f(0.320663452f, 0.583496094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.294072151f, 0.588334084f, 0.f),
    C_Vertex3f(0.297328949f, 0.525018692f, 0.f),
    C_Vertex3f(0.304358482f, 0.585108757f, 0.f),
    C_Vertex3f(0.313282013f, 0.521972656f, 0.f),
    C_Vertex3f(0.313331604f, 0.583496094f, 0.f),
    C_Vertex3f(0.329332352f, 0.521972656f, 0.f),
    C_Vertex3f(0.320663452f, 0.583496094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227532387f, 0.556135178f, 0.f),
    C_Vertex3f(0.240226746f, 0.550258636f, 0.f),
    C_Vertex3f(0.243349075f, 0.610917091f, 0.f),
    C_Vertex3f(0.258831024f, 0.540248871f, 0.f),
    C_Vertex3f(0.264305115f, 0.602205276f, 0.f),
    C_Vertex3f(0.279178619f, 0.531110764f, 0.f),
    C_Vertex3f(0.28247261f, 0.593172073f, 0.f),
    C_Vertex3f(0.297328949f, 0.525018692f, 0.f),
    C_Vertex3f(0.294072151f, 0.588334084f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200937271f, 0.562011719f, 0.f),
    C_Vertex3f(0.208051682f, 0.562011719f, 0.f),
    C_Vertex3f(0.208835602f, 0.619628906f, 0.f),
    C_Vertex3f(0.216807365f, 0.560052872f, 0.f),
    C_Vertex3f(0.224859238f, 0.616724968f, 0.f),
    C_Vertex3f(0.227532387f, 0.556135178f, 0.f),
    C_Vertex3f(0.243349075f, 0.610917091f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.182878494f, 0.554149628f, 0.f),
    C_Vertex3f(0.188186646f, 0.558080673f, 0.f),
    C_Vertex3f(0.192327499f, 0.619628906f, 0.f),
    C_Vertex3f(0.194206238f, 0.56070137f, 0.f),
    C_Vertex3f(0.208835602f, 0.619628906f, 0.f),
    C_Vertex3f(0.200937271f, 0.562011719f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172318459f, 0.534650803f, 0.f),
    C_Vertex3f(0.174616814f, 0.542406082f, 0.f),
    C_Vertex3f(0.174616814f, 0.616259575f, 0.f),
    C_Vertex3f(0.178173065f, 0.548906326f, 0.f),
    C_Vertex3f(0.192327499f, 0.619628906f, 0.f),
    C_Vertex3f(0.180580139f, 0.551528931f, 0.f),
    C_Vertex3f(0.182878494f, 0.554149628f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.174616814f, 0.616259575f, 0.f),
    C_Vertex3f(0.158654213f, 0.609520912f, 0.f),
    C_Vertex3f(0.171278f, 0.525640488f, 0.f),
    C_Vertex3f(0.172318459f, 0.534650803f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.11328125f, 0.520507813f, 0.f),
    C_Vertex3f(0.113924265f, 0.536891699f, 0.f),
    C_Vertex3f(0.113226891f, 0.526159286f, 0.f),
    C_Vertex3f(0.436035156f, 0.f, 0.f),
    C_Vertex3f(0.357421875f, 0.294200897f, 0.f),
    C_Vertex3f(0.357421875f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, 0.f),
    C_Vertex3f(0.0590820313f, 0.f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.363372803f, 0.455764771f, -1.f),
    C_Vertex3f(0.345794678f, 0.463302612f, -1.f),
    C_Vertex3f(0.345953941f, 0.359997749f, -1.f),
    C_Vertex3f(0.350541115f, 0.348093987f, -1.f),
    C_Vertex3f(0.353981495f, 0.333159924f, -1.f),
    C_Vertex3f(0.356275082f, 0.31519556f, -1.f),
    C_Vertex3f(0.357421875f, 0.294200897f, -1.f),
    C_Vertex3f(0.436035156f, 0.f, -1.f),
    C_Vertex3f(0.378740311f, 0.447052002f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.309862137f, 0.395601273f, -1.f),
    C_Vertex3f(0.321546555f, 0.388502121f, -1.f),
    C_Vertex3f(0.32736969f, 0.468955994f, -1.f),
    C_Vertex3f(0.331456661f, 0.380201817f, -1.f),
    C_Vertex3f(0.345794678f, 0.463302612f, -1.f),
    C_Vertex3f(0.339592457f, 0.370700359f, -1.f),
    C_Vertex3f(0.345953941f, 0.359997749f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268743515f, 0.40625f, -1.f),
    C_Vertex3f(0.283460617f, 0.404475212f, -1.f),
    C_Vertex3f(0.287979126f, 0.474609375f, -1.f),
    C_Vertex3f(0.297166824f, 0.400925636f, -1.f),
    C_Vertex3f(0.308097839f, 0.472724915f, -1.f),
    C_Vertex3f(0.309862137f, 0.395601273f, -1.f),
    C_Vertex3f(0.32736969f, 0.468955994f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.230774164f, 0.403260589f, -1.f),
    C_Vertex3f(0.24247241f, 0.40525353f, -1.f),
    C_Vertex3f(0.242723465f, 0.47222662f, -1.f),
    C_Vertex3f(0.254600525f, 0.40625f, -1.f),
    C_Vertex3f(0.265432358f, 0.474609375f, -1.f),
    C_Vertex3f(0.268743515f, 0.40625f, -1.f),
    C_Vertex3f(0.287979126f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.22143364f, 0.467461109f, -1.f),
    C_Vertex3f(0.201562881f, 0.460312843f, -1.f),
    C_Vertex3f(0.208667278f, 0.396285295f, -1.f),
    C_Vertex3f(0.219505787f, 0.400271177f, -1.f),
    C_Vertex3f(0.230774164f, 0.403260589f, -1.f),
    C_Vertex3f(0.242723465f, 0.47222662f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.169611931f, 0.370377064f, -1.f),
    C_Vertex3f(0.178730965f, 0.378348827f, -1.f),
    C_Vertex3f(0.183111191f, 0.450781822f, -1.f),
    C_Vertex3f(0.188279867f, 0.385324121f, -1.f),
    C_Vertex3f(0.201562881f, 0.460312843f, -1.f),
    C_Vertex3f(0.198258638f, 0.391302943f, -1.f),
    C_Vertex3f(0.208667278f, 0.396285295f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.166078568f, 0.438868046f, -1.f),
    C_Vertex3f(0.150465012f, 0.424571514f, -1.f),
    C_Vertex3f(0.154793501f, 0.349068761f, -1.f),
    C_Vertex3f(0.161632776f, 0.360617042f, -1.f),
    C_Vertex3f(0.169611931f, 0.370377064f, -1.f),
    C_Vertex3f(0.183111191f, 0.450781822f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.150465012f, 0.424571514f, -1.f),
    C_Vertex3f(0.136270523f, 0.407892227f, -1.f),
    C_Vertex3f(0.138835192f, 0.284993052f, -1.f),
    C_Vertex3f(0.14111495f, 0.303694367f, -1.f),
    C_Vertex3f(0.144534588f, 0.320607424f, -1.f),
    C_Vertex3f(0.149094105f, 0.335732222f, -1.f),
    C_Vertex3f(0.154793501f, 0.349068761f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.129882813f, 0.464355469f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.129882813f, 0.398361206f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.136270523f, 0.407892227f, -1.f),
    C_Vertex3f(0.137695313f, 0.264503479f, -1.f),
    C_Vertex3f(0.138835192f, 0.284993052f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.436035156f, 0.f, -1.f),
    C_Vertex3f(0.436035156f, 0.292205811f, -1.f),
    C_Vertex3f(0.435898781f, 0.304838181f, -1.f),
    C_Vertex3f(0.43562603f, 0.316446304f, -1.f),
    C_Vertex3f(0.435216904f, 0.327030182f, -1.f),
    C_Vertex3f(0.434671402f, 0.336589813f, -1.f),
    C_Vertex3f(0.433989525f, 0.345125198f, -1.f),
    C_Vertex3f(0.433171272f, 0.352636337f, -1.f),
    C_Vertex3f(0.432216644f, 0.35912323f, -1.f),
    C_Vertex3f(0.429924011f, 0.370267868f, -1.f),
    C_Vertex3f(0.4251194f, 0.385947227f, -1.f),
    C_Vertex3f(0.41900444f, 0.400479317f, -1.f),
    C_Vertex3f(0.411579132f, 0.413864136f, -1.f),
    C_Vertex3f(0.402843475f, 0.426101685f, -1.f),
    C_Vertex3f(0.391897202f, 0.437164307f, -1.f),
    C_Vertex3f(0.378740311f, 0.447052002f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.171386719f, 0.520507813f, -1.f),
    C_Vertex3f(0.171278f, 0.525640488f, -1.f),
    C_Vertex3f(0.158654213f, 0.609520912f, -1.f),
    C_Vertex3f(0.144439697f, 0.599412918f, -1.f),
    C_Vertex3f(0.135036469f, 0.589304924f, -1.f),
    C_Vertex3f(0.129502535f, 0.581995726f, -1.f),
    C_Vertex3f(0.124774694f, 0.574115992f, -1.f),
    C_Vertex3f(0.120852947f, 0.565665722f, -1.f),
    C_Vertex3f(0.117737293f, 0.556644917f, -1.f),
    C_Vertex3f(0.115427732f, 0.547053576f, -1.f),
    C_Vertex3f(0.113924265f, 0.536891699f, -1.f),
    C_Vertex3f(0.11328125f, 0.520507813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.405761719f, 0.621582031f, -1.f),
    C_Vertex3f(0.349121094f, 0.621582031f, -1.f),
    C_Vertex3f(0.361188412f, 0.531311035f, -1.f),
    C_Vertex3f(0.374699593f, 0.540649414f, -1.f),
    C_Vertex3f(0.383624554f, 0.549987793f, -1.f),
    C_Vertex3f(0.388871789f, 0.556991577f, -1.f),
    C_Vertex3f(0.393416286f, 0.56477356f, -1.f),
    C_Vertex3f(0.397258043f, 0.57333374f, -1.f),
    C_Vertex3f(0.400397062f, 0.582672119f, -1.f),
    C_Vertex3f(0.402833343f, 0.592788696f, -1.f),
    C_Vertex3f(0.404566884f, 0.603683472f, -1.f),
    C_Vertex3f(0.405597687f, 0.615356445f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.349121094f, 0.621582031f, -1.f),
    C_Vertex3f(0.34835434f, 0.615890503f, -1.f),
    C_Vertex3f(0.361188412f, 0.531311035f, -1.f),
    C_Vertex3f(0.346273899f, 0.606395721f, -1.f),
    C_Vertex3f(0.346065998f, 0.525085449f, -1.f),
    C_Vertex3f(0.343646526f, 0.598789215f, -1.f),
    C_Vertex3f(0.340472221f, 0.593070984f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.340472221f, 0.593070984f, -1.f),
    C_Vertex3f(0.336750984f, 0.589241028f, -1.f),
    C_Vertex3f(0.346065998f, 0.525085449f, -1.f),
    C_Vertex3f(0.33220911f, 0.586368561f, -1.f),
    C_Vertex3f(0.329332352f, 0.521972656f, -1.f),
    C_Vertex3f(0.3268466f, 0.584453583f, -1.f),
    C_Vertex3f(0.320663452f, 0.583496094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.294072151f, 0.588334084f, -1.f),
    C_Vertex3f(0.297328949f, 0.525018692f, -1.f),
    C_Vertex3f(0.304358482f, 0.585108757f, -1.f),
    C_Vertex3f(0.313282013f, 0.521972656f, -1.f),
    C_Vertex3f(0.313331604f, 0.583496094f, -1.f),
    C_Vertex3f(0.329332352f, 0.521972656f, -1.f),
    C_Vertex3f(0.320663452f, 0.583496094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227532387f, 0.556135178f, -1.f),
    C_Vertex3f(0.240226746f, 0.550258636f, -1.f),
    C_Vertex3f(0.243349075f, 0.610917091f, -1.f),
    C_Vertex3f(0.258831024f, 0.540248871f, -1.f),
    C_Vertex3f(0.264305115f, 0.602205276f, -1.f),
    C_Vertex3f(0.279178619f, 0.531110764f, -1.f),
    C_Vertex3f(0.28247261f, 0.593172073f, -1.f),
    C_Vertex3f(0.297328949f, 0.525018692f, -1.f),
    C_Vertex3f(0.294072151f, 0.588334084f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.200937271f, 0.562011719f, -1.f),
    C_Vertex3f(0.208051682f, 0.562011719f, -1.f),
    C_Vertex3f(0.208835602f, 0.619628906f, -1.f),
    C_Vertex3f(0.216807365f, 0.560052872f, -1.f),
    C_Vertex3f(0.224859238f, 0.616724968f, -1.f),
    C_Vertex3f(0.227532387f, 0.556135178f, -1.f),
    C_Vertex3f(0.243349075f, 0.610917091f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.182878494f, 0.554149628f, -1.f),
    C_Vertex3f(0.188186646f, 0.558080673f, -1.f),
    C_Vertex3f(0.192327499f, 0.619628906f, -1.f),
    C_Vertex3f(0.194206238f, 0.56070137f, -1.f),
    C_Vertex3f(0.208835602f, 0.619628906f, -1.f),
    C_Vertex3f(0.200937271f, 0.562011719f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.172318459f, 0.534650803f, -1.f),
    C_Vertex3f(0.174616814f, 0.542406082f, -1.f),
    C_Vertex3f(0.174616814f, 0.616259575f, -1.f),
    C_Vertex3f(0.178173065f, 0.548906326f, -1.f),
    C_Vertex3f(0.192327499f, 0.619628906f, -1.f),
    C_Vertex3f(0.180580139f, 0.551528931f, -1.f),
    C_Vertex3f(0.182878494f, 0.554149628f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.174616814f, 0.616259575f, -1.f),
    C_Vertex3f(0.158654213f, 0.609520912f, -1.f),
    C_Vertex3f(0.171278f, 0.525640488f, -1.f),
    C_Vertex3f(0.172318459f, 0.534650803f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.11328125f, 0.520507813f, -1.f),
    C_Vertex3f(0.113924265f, 0.536891699f, -1.f),
    C_Vertex3f(0.113226891f, 0.526159286f, -1.f),
    C_Vertex3f(0.436035156f, 0.f, -1.f),
    C_Vertex3f(0.357421875f, 0.294200897f, -1.f),
    C_Vertex3f(0.357421875f, 0.f, -1.f),
    C_Vertex3f(0.137695313f, 0.f, -1.f),
    C_Vertex3f(0.0590820313f, 0.464355469f, -1.f),
    C_Vertex3f(0.0590820313f, 0.f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_242[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_Normal3f(-0.999325812f, 0.0367154181f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Normal3f(-0.993402183f, 0.114683464f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Normal3f(-0.979182839f, 0.20298022f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Normal3f(-0.953921318f, 0.300056934f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Normal3f(-0.914828122f, 0.403843552f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Normal3f(-0.859796643f, 0.510636568f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Normal3f(-0.788298845f, 0.615292549f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Normal3f(-0.703678906f, 0.710518122f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Normal3f(-0.620272636f, 0.784386277f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Normal3f(-0.543085396f, 0.839677453f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Normal3f(-0.461994827f, 0.886882663f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Normal3f(-0.376675963f, 0.92634511f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Normal3f(-0.289430827f, 0.957198918f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Normal3f(-0.202602044f, 0.97926122f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Normal3f(-0.118287019f, 0.992979467f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Normal3f(-0.0370853581f, 0.999312103f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Normal3f(0.0429813564f, 0.99907589f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Normal3f(0.130778968f, 0.991411507f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Normal3f(0.22524552f, 0.974301994f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Normal3f(0.322965264f, 0.946410835f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Normal3f(0.420876533f, 0.907118022f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Normal3f(0.515538931f, 0.856866181f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Normal3f(0.603679597f, 0.797226965f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Normal3f(0.682734966f, 0.730666041f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Normal3f(0.755707204f, 0.65490967f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Normal3f(0.823695898f, 0.56703198f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Normal3f(0.880775928f, 0.473533183f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Normal3f(0.925762534f, 0.378105432f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Normal3f(0.958753943f, 0.284237236f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Normal3f(0.980860114f, 0.194713563f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Normal3f(0.993775964f, 0.111397475f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Normal3f(0.999312639f, 0.0370702483f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Normal3f(0.999835014f, -0.0181650277f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_Normal3f(0.99813658f, -0.0610187873f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Normal3f(0.994201541f, -0.107533179f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Normal3f(0.987230599f, -0.159297675f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Normal3f(0.976227641f, -0.216747552f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Normal3f(0.95995146f, -0.280166566f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Normal3f(0.936915934f, -0.349554807f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Normal3f(0.90544951f, -0.424453944f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Normal3f(0.869160771f, -0.494529665f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Normal3f(0.831279218f, -0.555855095f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Normal3f(0.787997842f, -0.615678072f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Normal3f(0.739750624f, -0.672881126f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Normal3f(0.687237859f, -0.726432562f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Normal3f(0.631377041f, -0.775475979f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Normal3f(0.573225498f, -0.819397628f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Normal3f(0.507929087f, -0.861398876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Normal3f(0.419512093f, -0.907749832f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Normal3f(0.30787161f, -0.951427937f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Normal3f(0.183039665f, -0.9831056f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Normal3f(0.0626576319f, -0.998035133f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Normal3f(-0.038599968f, -0.999254763f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Normal3f(-0.128493756f, -0.991710246f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Normal3f(-0.222021356f, -0.975041866f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Normal3f(-0.319475144f, -0.947594643f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Normal3f(-0.417872697f, -0.908505559f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Normal3f(-0.513709545f, -0.857964098f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Normal3f(-0.60350287f, -0.797360897f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Normal3f(-0.684390008f, -0.729116023f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Normal3f(-0.760332525f, -0.649533927f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Normal3f(-0.830618083f, -0.556842566f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Normal3f(-0.887843728f, -0.460145056f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Normal3f(-0.931549191f, -0.363615453f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Normal3f(-0.962654054f, -0.270734549f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Normal3f(-0.982942224f, -0.183914542f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Normal3f(-0.994526625f, -0.104483135f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Normal3f(0.999606252f, 0.0280574113f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Normal3f(0.995970607f, 0.0896810889f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Normal3f(0.987246752f, 0.159197494f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Normal3f(0.971514463f, 0.236980483f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Normal3f(0.946460783f, 0.322818667f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Normal3f(0.909567118f, 0.415557086f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Normal3f(0.858534515f, 0.51275593f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Normal3f(0.791965246f, 0.610566258f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Normal3f(0.720430911f, 0.693526745f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Normal3f(0.646434367f, 0.762969673f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Normal3f(0.5606305f, 0.828066111f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Normal3f(0.464660376f, 0.885488927f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Normal3f(0.361490935f, 0.932375669f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Normal3f(0.25501439f, 0.966937244f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Normal3f(0.149346635f, 0.988784969f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Normal3f(0.0481923707f, 0.998838007f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0485911854f, 0.998818815f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Normal3f(-0.151025206f, 0.988529861f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Normal3f(-0.257574052f, 0.966258526f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Normal3f(-0.36460194f, 0.93116349f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Normal3f(-0.467931896f, 0.883764625f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Normal3f(-0.563691318f, 0.825985491f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Normal3f(-0.649005592f, 0.760783732f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Normal3f(-0.72236073f, 0.69151628f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Normal3f(-0.794082761f, 0.607809663f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Normal3f(-0.861015081f, 0.508579373f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_Normal3f(-0.911770999f, 0.410699069f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Normal3f(-0.948089004f, 0.318005204f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Normal3f(-0.972527146f, 0.232789084f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Normal3f(-0.987754703f, 0.156014875f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Normal3f(-0.99614507f, 0.0877208859f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Normal3f(-0.999613106f, 0.0278154202f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Normal3f(-0.999587059f, -0.0287333243f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Normal3f(-0.995663822f, -0.0930251554f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Normal3f(-0.986366749f, -0.164561793f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Normal3f(-0.969791949f, -0.243933469f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Normal3f(-0.943758905f, -0.330634564f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Normal3f(-0.906032026f, -0.423209161f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Normal3f(-0.854762733f, -0.519018829f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Normal3f(-0.789101183f, -0.614263177f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Normal3f(-0.719142795f, -0.694862306f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Normal3f(-0.64632684f, -0.763060749f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Normal3f(-0.561695635f, -0.827343881f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Normal3f(-0.466670901f, -0.884431064f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Normal3f(-0.364006966f, -0.931396246f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Normal3f(-0.257463753f, -0.966287971f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Normal3f(-0.151149184f, -0.988511026f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Normal3f(-0.0485963188f, -0.998818517f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Normal3f(0.0480433144f, -0.99884522f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Normal3f(0.148618564f, -0.988894641f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Normal3f(0.253825158f, -0.967250109f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Normal3f(0.359922737f, -0.932982147f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Normal3f(0.462841064f, -0.886441231f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Normal3f(0.558709502f, -0.829363465f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Normal3f(0.644552469f, -0.764560103f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Normal3f(0.718696654f, -0.695323706f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Normal3f(0.790661871f, -0.612253129f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Normal3f(0.857762754f, -0.514045715f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Normal3f(0.909156442f, -0.416454464f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Normal3f(0.946268499f, -0.32338205f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Normal3f(0.971438348f, -0.237291992f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Normal3f(0.987223685f, -0.159340307f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Normal3f(0.995966375f, -0.0897267833f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.310546875f, 0.521972656f, 0.f),
    C_Vertex3f(0.310546875f, 0.521972656f, -1.f),
    C_Vertex3f(0.247070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.247070313f, 0.521972656f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.247070313f, 0.521972656f, 0.f),
    C_Vertex3f(0.247070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.145996094f, 0.64453125f, -1.f),
    C_Vertex3f(0.145996094f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.145996094f, 0.64453125f, 0.f),
    C_Vertex3f(0.145996094f, 0.64453125f, -1.f),
    C_Vertex3f(0.251464844f, 0.64453125f, -1.f),
    C_Vertex3f(0.251464844f, 0.64453125f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.251464844f, 0.64453125f, 0.f),
    C_Vertex3f(0.251464844f, 0.64453125f, -1.f),
    C_Vertex3f(0.310546875f, 0.521972656f, -1.f),
    C_Vertex3f(0.310546875f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.251464844f, 0.64453125f, 0.f),
    C_Vertex3f(0.145996094f, 0.64453125f, 0.f),
    C_Vertex3f(0.247070313f, 0.521972656f, 0.f),
    C_Vertex3f(0.310546875f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.251464844f, 0.64453125f, -1.f),
    C_Vertex3f(0.145996094f, 0.64453125f, -1.f),
    C_Vertex3f(0.247070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.310546875f, 0.521972656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_243[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_Normal3f(-0.999325812f, 0.0367154181f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Normal3f(-0.993402183f, 0.114683464f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Normal3f(-0.979182839f, 0.20298022f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Normal3f(-0.953921318f, 0.300056934f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Normal3f(-0.914828122f, 0.403843552f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Normal3f(-0.859796643f, 0.510636568f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Normal3f(-0.788298845f, 0.615292549f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Normal3f(-0.703678906f, 0.710518122f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Normal3f(-0.620272636f, 0.784386277f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Normal3f(-0.543085396f, 0.839677453f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Normal3f(-0.461994827f, 0.886882663f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Normal3f(-0.376675963f, 0.92634511f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Normal3f(-0.289430827f, 0.957198918f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Normal3f(-0.202602044f, 0.97926122f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Normal3f(-0.118287019f, 0.992979467f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Normal3f(-0.0370853581f, 0.999312103f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Normal3f(0.0429813564f, 0.99907589f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Normal3f(0.130778968f, 0.991411507f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Normal3f(0.22524552f, 0.974301994f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Normal3f(0.322965264f, 0.946410835f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Normal3f(0.420876533f, 0.907118022f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Normal3f(0.515538931f, 0.856866181f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Normal3f(0.603679597f, 0.797226965f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Normal3f(0.682734966f, 0.730666041f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Normal3f(0.755707204f, 0.65490967f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Normal3f(0.823695898f, 0.56703198f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Normal3f(0.880775928f, 0.473533183f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Normal3f(0.925762534f, 0.378105432f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Normal3f(0.958753943f, 0.284237236f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Normal3f(0.980860114f, 0.194713563f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Normal3f(0.993775964f, 0.111397475f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Normal3f(0.999312639f, 0.0370702483f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Normal3f(0.999835014f, -0.0181650277f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_Normal3f(0.99813658f, -0.0610187873f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Normal3f(0.994201541f, -0.107533179f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Normal3f(0.987230599f, -0.159297675f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Normal3f(0.976227641f, -0.216747552f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Normal3f(0.95995146f, -0.280166566f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Normal3f(0.936915934f, -0.349554807f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Normal3f(0.90544951f, -0.424453944f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Normal3f(0.869160771f, -0.494529665f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Normal3f(0.831279218f, -0.555855095f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Normal3f(0.787997842f, -0.615678072f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Normal3f(0.739750624f, -0.672881126f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Normal3f(0.687237859f, -0.726432562f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Normal3f(0.631377041f, -0.775475979f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Normal3f(0.573225498f, -0.819397628f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Normal3f(0.507929087f, -0.861398876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Normal3f(0.419512093f, -0.907749832f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Normal3f(0.30787161f, -0.951427937f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Normal3f(0.183039665f, -0.9831056f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Normal3f(0.0626576319f, -0.998035133f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Normal3f(-0.038599968f, -0.999254763f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Normal3f(-0.128493756f, -0.991710246f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Normal3f(-0.222021356f, -0.975041866f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Normal3f(-0.319475144f, -0.947594643f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Normal3f(-0.417872697f, -0.908505559f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Normal3f(-0.513709545f, -0.857964098f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Normal3f(-0.60350287f, -0.797360897f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Normal3f(-0.684390008f, -0.729116023f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Normal3f(-0.760332525f, -0.649533927f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Normal3f(-0.830618083f, -0.556842566f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Normal3f(-0.887843728f, -0.460145056f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Normal3f(-0.931549191f, -0.363615453f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Normal3f(-0.962654054f, -0.270734549f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Normal3f(-0.982942224f, -0.183914542f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Normal3f(-0.994526625f, -0.104483135f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Normal3f(0.999606252f, 0.0280574113f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Normal3f(0.995970607f, 0.0896810889f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Normal3f(0.987246752f, 0.159197494f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Normal3f(0.971514463f, 0.236980483f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Normal3f(0.946460783f, 0.322818667f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Normal3f(0.909567118f, 0.415557086f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Normal3f(0.858534515f, 0.51275593f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Normal3f(0.791965246f, 0.610566258f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Normal3f(0.720430911f, 0.693526745f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Normal3f(0.646434367f, 0.762969673f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Normal3f(0.5606305f, 0.828066111f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Normal3f(0.464660376f, 0.885488927f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Normal3f(0.361490935f, 0.932375669f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Normal3f(0.25501439f, 0.966937244f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Normal3f(0.149346635f, 0.988784969f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Normal3f(0.0481923707f, 0.998838007f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0485911854f, 0.998818815f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Normal3f(-0.151025206f, 0.988529861f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Normal3f(-0.257574052f, 0.966258526f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Normal3f(-0.36460194f, 0.93116349f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Normal3f(-0.467931896f, 0.883764625f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Normal3f(-0.563691318f, 0.825985491f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Normal3f(-0.649005592f, 0.760783732f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Normal3f(-0.72236073f, 0.69151628f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Normal3f(-0.794082761f, 0.607809663f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Normal3f(-0.861015081f, 0.508579373f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_Normal3f(-0.911770999f, 0.410699069f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Normal3f(-0.948089004f, 0.318005204f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Normal3f(-0.972527146f, 0.232789084f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Normal3f(-0.987754703f, 0.156014875f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Normal3f(-0.99614507f, 0.0877208859f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Normal3f(-0.999613106f, 0.0278154202f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Normal3f(-0.999587059f, -0.0287333243f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Normal3f(-0.995663822f, -0.0930251554f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Normal3f(-0.986366749f, -0.164561793f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Normal3f(-0.969791949f, -0.243933469f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Normal3f(-0.943758905f, -0.330634564f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Normal3f(-0.906032026f, -0.423209161f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Normal3f(-0.854762733f, -0.519018829f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Normal3f(-0.789101183f, -0.614263177f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Normal3f(-0.719142795f, -0.694862306f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Normal3f(-0.64632684f, -0.763060749f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Normal3f(-0.561695635f, -0.827343881f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Normal3f(-0.466670901f, -0.884431064f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Normal3f(-0.364006966f, -0.931396246f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Normal3f(-0.257463753f, -0.966287971f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Normal3f(-0.151149184f, -0.988511026f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Normal3f(-0.0485963188f, -0.998818517f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Normal3f(0.0480433144f, -0.99884522f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Normal3f(0.148618564f, -0.988894641f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Normal3f(0.253825158f, -0.967250109f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Normal3f(0.359922737f, -0.932982147f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Normal3f(0.462841064f, -0.886441231f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Normal3f(0.558709502f, -0.829363465f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Normal3f(0.644552469f, -0.764560103f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Normal3f(0.718696654f, -0.695323706f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Normal3f(0.790661871f, -0.612253129f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Normal3f(0.857762754f, -0.514045715f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Normal3f(0.909156442f, -0.416454464f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Normal3f(0.946268499f, -0.32338205f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Normal3f(0.971438348f, -0.237291992f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Normal3f(0.987223685f, -0.159340307f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Normal3f(0.995966375f, -0.0897267833f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.182128906f, 0.521972656f, 0.f),
    C_Vertex3f(0.182128906f, 0.521972656f, -1.f),
    C_Vertex3f(0.239746094f, 0.64453125f, -1.f),
    C_Vertex3f(0.239746094f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.239746094f, 0.64453125f, 0.f),
    C_Vertex3f(0.239746094f, 0.64453125f, -1.f),
    C_Vertex3f(0.343261719f, 0.64453125f, -1.f),
    C_Vertex3f(0.343261719f, 0.64453125f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.343261719f, 0.64453125f, 0.f),
    C_Vertex3f(0.343261719f, 0.64453125f, -1.f),
    C_Vertex3f(0.247070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.247070313f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.247070313f, 0.521972656f, 0.f),
    C_Vertex3f(0.247070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.182128906f, 0.521972656f, -1.f),
    C_Vertex3f(0.182128906f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.239746094f, 0.64453125f, 0.f),
    C_Vertex3f(0.182128906f, 0.521972656f, 0.f),
    C_Vertex3f(0.247070313f, 0.521972656f, 0.f),
    C_Vertex3f(0.343261719f, 0.64453125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.239746094f, 0.64453125f, -1.f),
    C_Vertex3f(0.182128906f, 0.521972656f, -1.f),
    C_Vertex3f(0.247070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.343261719f, 0.64453125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_244[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_Normal3f(-0.999325812f, 0.0367154181f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Normal3f(-0.993402183f, 0.114683464f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Normal3f(-0.979182839f, 0.20298022f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Normal3f(-0.953921318f, 0.300056934f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Normal3f(-0.914828122f, 0.403843552f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Normal3f(-0.859796643f, 0.510636568f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Normal3f(-0.788298845f, 0.615292549f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Normal3f(-0.703678906f, 0.710518122f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Normal3f(-0.620272636f, 0.784386277f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Normal3f(-0.543085396f, 0.839677453f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Normal3f(-0.461994827f, 0.886882663f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Normal3f(-0.376675963f, 0.92634511f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Normal3f(-0.289430827f, 0.957198918f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Normal3f(-0.202602044f, 0.97926122f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Normal3f(-0.118287019f, 0.992979467f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Normal3f(-0.0370853581f, 0.999312103f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Normal3f(0.0429813564f, 0.99907589f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Normal3f(0.130778968f, 0.991411507f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Normal3f(0.22524552f, 0.974301994f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Normal3f(0.322965264f, 0.946410835f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Normal3f(0.420876533f, 0.907118022f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Normal3f(0.515538931f, 0.856866181f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Normal3f(0.603679597f, 0.797226965f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Normal3f(0.682734966f, 0.730666041f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Normal3f(0.755707204f, 0.65490967f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Normal3f(0.823695898f, 0.56703198f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Normal3f(0.880775928f, 0.473533183f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Normal3f(0.925762534f, 0.378105432f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Normal3f(0.958753943f, 0.284237236f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Normal3f(0.980860114f, 0.194713563f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Normal3f(0.993775964f, 0.111397475f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Normal3f(0.999312639f, 0.0370702483f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Normal3f(0.999835014f, -0.0181650277f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_Normal3f(0.99813658f, -0.0610187873f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Normal3f(0.994201541f, -0.107533179f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Normal3f(0.987230599f, -0.159297675f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Normal3f(0.976227641f, -0.216747552f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Normal3f(0.95995146f, -0.280166566f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Normal3f(0.936915934f, -0.349554807f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Normal3f(0.90544951f, -0.424453944f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Normal3f(0.869160771f, -0.494529665f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Normal3f(0.831279218f, -0.555855095f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Normal3f(0.787997842f, -0.615678072f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Normal3f(0.739750624f, -0.672881126f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Normal3f(0.687237859f, -0.726432562f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Normal3f(0.631377041f, -0.775475979f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Normal3f(0.573225498f, -0.819397628f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Normal3f(0.507929087f, -0.861398876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Normal3f(0.419512093f, -0.907749832f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Normal3f(0.30787161f, -0.951427937f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Normal3f(0.183039665f, -0.9831056f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Normal3f(0.0626576319f, -0.998035133f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Normal3f(-0.038599968f, -0.999254763f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Normal3f(-0.128493756f, -0.991710246f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Normal3f(-0.222021356f, -0.975041866f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Normal3f(-0.319475144f, -0.947594643f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Normal3f(-0.417872697f, -0.908505559f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Normal3f(-0.513709545f, -0.857964098f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Normal3f(-0.60350287f, -0.797360897f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Normal3f(-0.684390008f, -0.729116023f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Normal3f(-0.760332525f, -0.649533927f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Normal3f(-0.830618083f, -0.556842566f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Normal3f(-0.887843728f, -0.460145056f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Normal3f(-0.931549191f, -0.363615453f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Normal3f(-0.962654054f, -0.270734549f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Normal3f(-0.982942224f, -0.183914542f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Normal3f(-0.994526625f, -0.104483135f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Normal3f(0.999606252f, 0.0280574113f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Normal3f(0.995970607f, 0.0896810889f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Normal3f(0.987246752f, 0.159197494f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Normal3f(0.971514463f, 0.236980483f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Normal3f(0.946460783f, 0.322818667f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Normal3f(0.909567118f, 0.415557086f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Normal3f(0.858534515f, 0.51275593f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Normal3f(0.791965246f, 0.610566258f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Normal3f(0.720430911f, 0.693526745f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Normal3f(0.646434367f, 0.762969673f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Normal3f(0.5606305f, 0.828066111f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Normal3f(0.464660376f, 0.885488927f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Normal3f(0.361490935f, 0.932375669f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Normal3f(0.25501439f, 0.966937244f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Normal3f(0.149346635f, 0.988784969f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Normal3f(0.0481923707f, 0.998838007f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0485911854f, 0.998818815f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Normal3f(-0.151025206f, 0.988529861f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Normal3f(-0.257574052f, 0.966258526f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Normal3f(-0.36460194f, 0.93116349f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Normal3f(-0.467931896f, 0.883764625f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Normal3f(-0.563691318f, 0.825985491f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Normal3f(-0.649005592f, 0.760783732f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Normal3f(-0.72236073f, 0.69151628f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Normal3f(-0.794082761f, 0.607809663f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Normal3f(-0.861015081f, 0.508579373f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_Normal3f(-0.911770999f, 0.410699069f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Normal3f(-0.948089004f, 0.318005204f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Normal3f(-0.972527146f, 0.232789084f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Normal3f(-0.987754703f, 0.156014875f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Normal3f(-0.99614507f, 0.0877208859f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Normal3f(-0.999613106f, 0.0278154202f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Normal3f(-0.999587059f, -0.0287333243f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Normal3f(-0.995663822f, -0.0930251554f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Normal3f(-0.986366749f, -0.164561793f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Normal3f(-0.969791949f, -0.243933469f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Normal3f(-0.943758905f, -0.330634564f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Normal3f(-0.906032026f, -0.423209161f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Normal3f(-0.854762733f, -0.519018829f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Normal3f(-0.789101183f, -0.614263177f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Normal3f(-0.719142795f, -0.694862306f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Normal3f(-0.64632684f, -0.763060749f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Normal3f(-0.561695635f, -0.827343881f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Normal3f(-0.466670901f, -0.884431064f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Normal3f(-0.364006966f, -0.931396246f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Normal3f(-0.257463753f, -0.966287971f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Normal3f(-0.151149184f, -0.988511026f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Normal3f(-0.0485963188f, -0.998818517f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Normal3f(0.0480433144f, -0.99884522f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Normal3f(0.148618564f, -0.988894641f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Normal3f(0.253825158f, -0.967250109f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Normal3f(0.359922737f, -0.932982147f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Normal3f(0.462841064f, -0.886441231f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Normal3f(0.558709502f, -0.829363465f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Normal3f(0.644552469f, -0.764560103f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Normal3f(0.718696654f, -0.695323706f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Normal3f(0.790661871f, -0.612253129f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Normal3f(0.857762754f, -0.514045715f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Normal3f(0.909156442f, -0.416454464f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Normal3f(0.946268499f, -0.32338205f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Normal3f(0.971438348f, -0.237291992f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Normal3f(0.987223685f, -0.159340307f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Normal3f(0.995966375f, -0.0897267833f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.247558594f, 0.596679688f, 0.f),
    C_Vertex3f(0.247558594f, 0.596679688f, -1.f),
    C_Vertex3f(0.198486328f, 0.521972656f, -1.f),
    C_Vertex3f(0.198486328f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.198486328f, 0.521972656f, 0.f),
    C_Vertex3f(0.198486328f, 0.521972656f, -1.f),
    C_Vertex3f(0.109130859f, 0.521972656f, -1.f),
    C_Vertex3f(0.109130859f, 0.521972656f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(0.109130859f, 0.521972656f, 0.f),
    C_Vertex3f(0.109130859f, 0.521972656f, -1.f),
    C_Vertex3f(0.202636719f, 0.64453125f, -1.f),
    C_Vertex3f(0.202636719f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.202636719f, 0.64453125f, 0.f),
    C_Vertex3f(0.202636719f, 0.64453125f, -1.f),
    C_Vertex3f(0.286621094f, 0.64453125f, -1.f),
    C_Vertex3f(0.286621094f, 0.64453125f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.286621094f, 0.64453125f, 0.f),
    C_Vertex3f(0.286621094f, 0.64453125f, -1.f),
    C_Vertex3f(0.385009766f, 0.521972656f, -1.f),
    C_Vertex3f(0.385009766f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.385009766f, 0.521972656f, 0.f),
    C_Vertex3f(0.385009766f, 0.521972656f, -1.f),
    C_Vertex3f(0.295654297f, 0.521972656f, -1.f),
    C_Vertex3f(0.295654297f, 0.521972656f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.295654297f, 0.521972656f, 0.f),
    C_Vertex3f(0.295654297f, 0.521972656f, -1.f),
    C_Vertex3f(0.247558594f, 0.596679688f, -1.f),
    C_Vertex3f(0.247558594f, 0.596679688f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.109130859f, 0.521972656f, 0.f),
    C_Vertex3f(0.198486328f, 0.521972656f, 0.f),
    C_Vertex3f(0.202636719f, 0.64453125f, 0.f),
    C_Vertex3f(0.247558594f, 0.596679688f, 0.f),
    C_Vertex3f(0.286621094f, 0.64453125f, 0.f),
    C_Vertex3f(0.295654297f, 0.521972656f, 0.f),
    C_Vertex3f(0.385009766f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.109130859f, 0.521972656f, -1.f),
    C_Vertex3f(0.198486328f, 0.521972656f, -1.f),
    C_Vertex3f(0.202636719f, 0.64453125f, -1.f),
    C_Vertex3f(0.247558594f, 0.596679688f, -1.f),
    C_Vertex3f(0.286621094f, 0.64453125f, -1.f),
    C_Vertex3f(0.295654297f, 0.521972656f, -1.f),
    C_Vertex3f(0.385009766f, 0.521972656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_245[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_Normal3f(-0.999325812f, 0.0367154181f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Normal3f(-0.993402183f, 0.114683464f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Normal3f(-0.979182839f, 0.20298022f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Normal3f(-0.953921318f, 0.300056934f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Normal3f(-0.914828122f, 0.403843552f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Normal3f(-0.859796643f, 0.510636568f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Normal3f(-0.788298845f, 0.615292549f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Normal3f(-0.703678906f, 0.710518122f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Normal3f(-0.620272636f, 0.784386277f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Normal3f(-0.543085396f, 0.839677453f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Normal3f(-0.461994827f, 0.886882663f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Normal3f(-0.376675963f, 0.92634511f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Normal3f(-0.289430827f, 0.957198918f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Normal3f(-0.202602044f, 0.97926122f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Normal3f(-0.118287019f, 0.992979467f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Normal3f(-0.0370853581f, 0.999312103f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Normal3f(0.0429813564f, 0.99907589f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Normal3f(0.130778968f, 0.991411507f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Normal3f(0.22524552f, 0.974301994f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Normal3f(0.322965264f, 0.946410835f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Normal3f(0.420876533f, 0.907118022f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Normal3f(0.515538931f, 0.856866181f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Normal3f(0.603679597f, 0.797226965f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Normal3f(0.682734966f, 0.730666041f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Normal3f(0.755707204f, 0.65490967f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Normal3f(0.823695898f, 0.56703198f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Normal3f(0.880775928f, 0.473533183f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Normal3f(0.925762534f, 0.378105432f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Normal3f(0.958753943f, 0.284237236f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Normal3f(0.980860114f, 0.194713563f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Normal3f(0.993775964f, 0.111397475f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Normal3f(0.999312639f, 0.0370702483f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Normal3f(0.999835014f, -0.0181650277f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_Normal3f(0.99813658f, -0.0610187873f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Normal3f(0.994201541f, -0.107533179f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Normal3f(0.987230599f, -0.159297675f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Normal3f(0.976227641f, -0.216747552f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Normal3f(0.95995146f, -0.280166566f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Normal3f(0.936915934f, -0.349554807f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Normal3f(0.90544951f, -0.424453944f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Normal3f(0.869160771f, -0.494529665f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Normal3f(0.831279218f, -0.555855095f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Normal3f(0.787997842f, -0.615678072f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Normal3f(0.739750624f, -0.672881126f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Normal3f(0.687237859f, -0.726432562f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Normal3f(0.631377041f, -0.775475979f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Normal3f(0.573225498f, -0.819397628f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Normal3f(0.507929087f, -0.861398876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Normal3f(0.419512093f, -0.907749832f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Normal3f(0.30787161f, -0.951427937f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Normal3f(0.183039665f, -0.9831056f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Normal3f(0.0626576319f, -0.998035133f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Normal3f(-0.038599968f, -0.999254763f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Normal3f(-0.128493756f, -0.991710246f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Normal3f(-0.222021356f, -0.975041866f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Normal3f(-0.319475144f, -0.947594643f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Normal3f(-0.417872697f, -0.908505559f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Normal3f(-0.513709545f, -0.857964098f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Normal3f(-0.60350287f, -0.797360897f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Normal3f(-0.684390008f, -0.729116023f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Normal3f(-0.760332525f, -0.649533927f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Normal3f(-0.830618083f, -0.556842566f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Normal3f(-0.887843728f, -0.460145056f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Normal3f(-0.931549191f, -0.363615453f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Normal3f(-0.962654054f, -0.270734549f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Normal3f(-0.982942224f, -0.183914542f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Normal3f(-0.994526625f, -0.104483135f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Normal3f(0.999606252f, 0.0280574113f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Normal3f(0.995970607f, 0.0896810889f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Normal3f(0.987246752f, 0.159197494f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Normal3f(0.971514463f, 0.236980483f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Normal3f(0.946460783f, 0.322818667f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Normal3f(0.909567118f, 0.415557086f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Normal3f(0.858534515f, 0.51275593f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Normal3f(0.791965246f, 0.610566258f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Normal3f(0.720430911f, 0.693526745f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Normal3f(0.646434367f, 0.762969673f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Normal3f(0.5606305f, 0.828066111f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Normal3f(0.464660376f, 0.885488927f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Normal3f(0.361490935f, 0.932375669f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Normal3f(0.25501439f, 0.966937244f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Normal3f(0.149346635f, 0.988784969f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Normal3f(0.0481923707f, 0.998838007f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0485911854f, 0.998818815f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Normal3f(-0.151025206f, 0.988529861f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Normal3f(-0.257574052f, 0.966258526f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Normal3f(-0.36460194f, 0.93116349f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Normal3f(-0.467931896f, 0.883764625f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Normal3f(-0.563691318f, 0.825985491f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Normal3f(-0.649005592f, 0.760783732f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Normal3f(-0.72236073f, 0.69151628f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Normal3f(-0.794082761f, 0.607809663f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Normal3f(-0.861015081f, 0.508579373f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_Normal3f(-0.911770999f, 0.410699069f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Normal3f(-0.948089004f, 0.318005204f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Normal3f(-0.972527146f, 0.232789084f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Normal3f(-0.987754703f, 0.156014875f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Normal3f(-0.99614507f, 0.0877208859f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Normal3f(-0.999613106f, 0.0278154202f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Normal3f(-0.999587059f, -0.0287333243f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Normal3f(-0.995663822f, -0.0930251554f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Normal3f(-0.986366749f, -0.164561793f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Normal3f(-0.969791949f, -0.243933469f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Normal3f(-0.943758905f, -0.330634564f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Normal3f(-0.906032026f, -0.423209161f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Normal3f(-0.854762733f, -0.519018829f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Normal3f(-0.789101183f, -0.614263177f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Normal3f(-0.719142795f, -0.694862306f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Normal3f(-0.64632684f, -0.763060749f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Normal3f(-0.561695635f, -0.827343881f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Normal3f(-0.466670901f, -0.884431064f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Normal3f(-0.364006966f, -0.931396246f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Normal3f(-0.257463753f, -0.966287971f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Normal3f(-0.151149184f, -0.988511026f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Normal3f(-0.0485963188f, -0.998818517f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Normal3f(0.0480433144f, -0.99884522f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Normal3f(0.148618564f, -0.988894641f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Normal3f(0.253825158f, -0.967250109f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Normal3f(0.359922737f, -0.932982147f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Normal3f(0.462841064f, -0.886441231f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Normal3f(0.558709502f, -0.829363465f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Normal3f(0.644552469f, -0.764560103f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Normal3f(0.718696654f, -0.695323706f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Normal3f(0.790661871f, -0.612253129f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Normal3f(0.857762754f, -0.514045715f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Normal3f(0.909156442f, -0.416454464f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Normal3f(0.946268499f, -0.32338205f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Normal3f(0.971438348f, -0.237291992f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Normal3f(0.987223685f, -0.159340307f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Normal3f(0.995966375f, -0.0897267833f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999953806f, -0.00961818546f, 0.f),
    C_Vertex3f(0.104492188f, 0.520507813f, 0.f),
    C_Vertex3f(0.104492188f, 0.520507813f, -1.f),
    C_Normal3f(-0.999230683f, 0.0392165966f, 0.f),
    C_Vertex3f(0.104437828f, 0.526159286f, 0.f),
    C_Vertex3f(0.104437828f, 0.526159286f, -1.f),
    C_Normal3f(-0.994498312f, 0.104752719f, 0.f),
    C_Vertex3f(0.105135202f, 0.536891699f, 0.f),
    C_Vertex3f(0.105135202f, 0.536891699f, -1.f),
    C_Normal3f(-0.981874108f, 0.189534381f, 0.f),
    C_Vertex3f(0.10663867f, 0.547053576f, 0.f),
    C_Vertex3f(0.10663867f, 0.547053576f, -1.f),
    C_Normal3f(-0.960046053f, 0.279841781f, 0.f),
    C_Vertex3f(0.108948231f, 0.556644917f, 0.f),
    C_Vertex3f(0.108948231f, 0.556644917f, -1.f),
    C_Normal3f(-0.927577436f, 0.373630971f, 0.f),
    C_Vertex3f(0.112063885f, 0.565665722f, 0.f),
    C_Vertex3f(0.112063885f, 0.565665722f, -1.f),
    C_Normal3f(-0.883691847f, 0.468069077f, 0.f),
    C_Vertex3f(0.115985632f, 0.574115992f, 0.f),
    C_Vertex3f(0.115985632f, 0.574115992f, -1.f),
    C_Normal3f(-0.828614712f, 0.559819281f, 0.f),
    C_Vertex3f(0.120713472f, 0.581995726f, 0.f),
    C_Vertex3f(0.120713472f, 0.581995726f, -1.f),
    C_Normal3f(-0.759081304f, 0.650995851f, 0.f),
    C_Vertex3f(0.126247406f, 0.589304924f, 0.f),
    C_Vertex3f(0.126247406f, 0.589304924f, -1.f),
    C_Normal3f(-0.650275886f, 0.759698212f, 0.f),
    C_Vertex3f(0.135650635f, 0.599412918f, 0.f),
    C_Vertex3f(0.135650635f, 0.599412918f, -1.f),
    C_Normal3f(-0.487445921f, 0.87315321f, 0.f),
    C_Vertex3f(0.14986515f, 0.609520912f, 0.f),
    C_Vertex3f(0.14986515f, 0.609520912f, -1.f),
    C_Normal3f(-0.287504673f, 0.957779288f, 0.f),
    C_Vertex3f(0.165827751f, 0.616259575f, 0.f),
    C_Vertex3f(0.165827751f, 0.616259575f, -1.f),
    C_Normal3f(-0.0979905054f, 0.995187461f, 0.f),
    C_Vertex3f(0.183538437f, 0.619628906f, 0.f),
    C_Vertex3f(0.183538437f, 0.619628906f, -1.f),
    C_Normal3f(0.0889112428f, 0.99603951f, 0.f),
    C_Vertex3f(0.200046539f, 0.619628906f, 0.f),
    C_Vertex3f(0.200046539f, 0.619628906f, -1.f),
    C_Normal3f(0.244741365f, 0.969588399f, 0.f),
    C_Vertex3f(0.216070175f, 0.616724968f, 0.f),
    C_Vertex3f(0.216070175f, 0.616724968f, -1.f),
    C_Normal3f(0.345433027f, 0.938443422f, 0.f),
    C_Vertex3f(0.234560013f, 0.610917091f, 0.f),
    C_Vertex3f(0.234560013f, 0.610917091f, -1.f),
    C_Normal3f(0.413061887f, 0.910703003f, 0.f),
    C_Vertex3f(0.255516052f, 0.602205276f, 0.f),
    C_Vertex3f(0.255516052f, 0.602205276f, -1.f),
    C_Normal3f(0.42238301f, 0.906417429f, 0.f),
    C_Vertex3f(0.273683548f, 0.593172073f, 0.f),
    C_Vertex3f(0.273683548f, 0.593172073f, -1.f),
    C_Normal3f(0.345709145f, 0.938341737f, 0.f),
    C_Vertex3f(0.285283089f, 0.588334084f, 0.f),
    C_Vertex3f(0.285283089f, 0.588334084f, -1.f),
    C_Normal3f(0.243631572f, 0.969867826f, 0.f),
    C_Vertex3f(0.29556942f, 0.585108757f, 0.f),
    C_Vertex3f(0.29556942f, 0.585108757f, -1.f),
    C_Normal3f(0.0984259844f, 0.995144367f, 0.f),
    C_Vertex3f(0.304542542f, 0.583496094f, 0.f),
    C_Vertex3f(0.304542542f, 0.583496094f, -1.f),
    C_Normal3f(-0.070669286f, 0.997499824f, 0.f),
    C_Vertex3f(0.31187439f, 0.583496094f, 0.f),
    C_Vertex3f(0.31187439f, 0.583496094f, -1.f),
    C_Normal3f(-0.24143213f, 0.970417738f, 0.f),
    C_Vertex3f(0.318057537f, 0.584453583f, 0.f),
    C_Vertex3f(0.318057537f, 0.584453583f, -1.f),
    C_Normal3f(-0.435192645f, 0.900337279f, 0.f),
    C_Vertex3f(0.323420048f, 0.586368561f, 0.f),
    C_Vertex3f(0.323420048f, 0.586368561f, -1.f),
    C_Normal3f(-0.62994957f, 0.776636064f, 0.f),
    C_Vertex3f(0.327961922f, 0.589241028f, 0.f),
    C_Vertex3f(0.327961922f, 0.589241028f, -1.f),
    C_Normal3f(-0.810694158f, 0.585469842f, 0.f),
    C_Vertex3f(0.331683159f, 0.593070984f, 0.f),
    C_Vertex3f(0.331683159f, 0.593070984f, -1.f),
    C_Normal3f(-0.9168607f, 0.399207145f, 0.f),
    C_Vertex3f(0.334857464f, 0.598789215f, 0.f),
    C_Vertex3f(0.334857464f, 0.598789215f, -1.f),
    C_Normal3f(-0.964133918f, 0.265416414f, 0.f),
    C_Vertex3f(0.337484837f, 0.606395721f, 0.f),
    C_Vertex3f(0.337484837f, 0.606395721f, -1.f),
    C_Normal3f(-0.982875049f, 0.18427363f, 0.f),
    C_Vertex3f(0.339565277f, 0.615890503f, 0.f),
    C_Vertex3f(0.339565277f, 0.615890503f, -1.f),
    C_Normal3f(-0.991047204f, 0.133512393f, 0.f),
    C_Vertex3f(0.340332031f, 0.621582031f, 0.f),
    C_Vertex3f(0.340332031f, 0.621582031f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.340332031f, 0.621582031f, 0.f),
    C_Vertex3f(0.340332031f, 0.621582031f, -1.f),
    C_Vertex3f(0.396972656f, 0.621582031f, -1.f),
    C_Vertex3f(0.396972656f, 0.621582031f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999653041f, -0.0263388976f, 0.f),
    C_Vertex3f(0.396972656f, 0.621582031f, 0.f),
    C_Vertex3f(0.396972656f, 0.621582031f, -1.f),
    C_Normal3f(0.99777931f, -0.066607669f, 0.f),
    C_Vertex3f(0.396808624f, 0.615356445f, 0.f),
    C_Vertex3f(0.396808624f, 0.615356445f, -1.f),
    C_Normal3f(0.992581308f, -0.121582195f, 0.f),
    C_Vertex3f(0.395777822f, 0.603683472f, 0.f),
    C_Vertex3f(0.395777822f, 0.603683472f, -1.f),
    C_Normal3f(0.980870843f, -0.194659367f, 0.f),
    C_Vertex3f(0.39404428f, 0.592788696f, 0.f),
    C_Vertex3f(0.39404428f, 0.592788696f, -1.f),
    C_Normal3f(0.961305082f, -0.275485754f, 0.f),
    C_Vertex3f(0.391608f, 0.582672119f, 0.f),
    C_Vertex3f(0.391608f, 0.582672119f, -1.f),
    C_Normal3f(0.931648552f, -0.363360554f, 0.f),
    C_Vertex3f(0.388468981f, 0.57333374f, 0.f),
    C_Vertex3f(0.388468981f, 0.57333374f, -1.f),
    C_Normal3f(0.889692426f, -0.456560522f, 0.f),
    C_Vertex3f(0.384627223f, 0.56477356f, 0.f),
    C_Vertex3f(0.384627223f, 0.56477356f, -1.f),
    C_Normal3f(0.833749712f, -0.552142739f, 0.f),
    C_Vertex3f(0.380082726f, 0.556991577f, 0.f),
    C_Vertex3f(0.380082726f, 0.556991577f, -1.f),
    C_Normal3f(0.755483389f, -0.655167818f, 0.f),
    C_Vertex3f(0.374835491f, 0.549987793f, 0.f),
    C_Vertex3f(0.374835491f, 0.549987793f, -1.f),
    C_Normal3f(0.639779389f, -0.768558562f, 0.f),
    C_Vertex3f(0.36591053f, 0.540649414f, 0.f),
    C_Vertex3f(0.36591053f, 0.540649414f, -1.f),
    C_Normal3f(0.477566123f, -0.878595829f, 0.f),
    C_Vertex3f(0.352399349f, 0.531311035f, 0.f),
    C_Vertex3f(0.352399349f, 0.531311035f, -1.f),
    C_Normal3f(0.281305313f, -0.95961827f, 0.f),
    C_Vertex3f(0.337276936f, 0.525085449f, 0.f),
    C_Vertex3f(0.337276936f, 0.525085449f, -1.f),
    C_Normal3f(0.0945234597f, -0.995522618f, 0.f),
    C_Vertex3f(0.320543289f, 0.521972656f, 0.f),
    C_Vertex3f(0.320543289f, 0.521972656f, -1.f),
    C_Normal3f(-0.0947503f, -0.995501101f, 0.f),
    C_Vertex3f(0.30449295f, 0.521972656f, 0.f),
    C_Vertex3f(0.30449295f, 0.521972656f, -1.f),
    C_Normal3f(-0.258822501f, -0.965924919f, 0.f),
    C_Vertex3f(0.288539886f, 0.525018692f, 0.f),
    C_Vertex3f(0.288539886f, 0.525018692f, -1.f),
    C_Normal3f(-0.367869198f, -0.92987752f, 0.f),
    C_Vertex3f(0.270389557f, 0.531110764f, 0.f),
    C_Vertex3f(0.270389557f, 0.531110764f, -1.f),
    C_Normal3f(-0.441156626f, -0.897430122f, 0.f),
    C_Vertex3f(0.250041962f, 0.540248871f, 0.f),
    C_Vertex3f(0.250041962f, 0.540248871f, -1.f),
    C_Normal3f(-0.452606946f, -0.891710103f, 0.f),
    C_Vertex3f(0.231437683f, 0.550258636f, 0.f),
    C_Vertex3f(0.231437683f, 0.550258636f, -1.f),
    C_Normal3f(-0.385828793f, -0.922570407f, 0.f),
    C_Vertex3f(0.218743324f, 0.556135178f, 0.f),
    C_Vertex3f(0.218743324f, 0.556135178f, -1.f),
    C_Normal3f(-0.288805217f, -0.957387865f, 0.f),
    C_Vertex3f(0.208018303f, 0.560052872f, 0.f),
    C_Vertex3f(0.208018303f, 0.560052872f, -1.f),
    C_Normal3f(-0.122500464f, -0.992468476f, 0.f),
    C_Vertex3f(0.199262619f, 0.562011719f, 0.f),
    C_Vertex3f(0.199262619f, 0.562011719f, -1.f),
    C_Normal3f(0.0942201167f, -0.995551467f, 0.f),
    C_Vertex3f(0.192148209f, 0.562011719f, 0.f),
    C_Vertex3f(0.192148209f, 0.562011719f, -1.f),
    C_Normal3f(0.294618219f, -0.955615103f, 0.f),
    C_Vertex3f(0.185417175f, 0.56070137f, 0.f),
    C_Vertex3f(0.185417175f, 0.56070137f, -1.f),
    C_Normal3f(0.500668526f, -0.865639091f, 0.f),
    C_Vertex3f(0.179397583f, 0.558080673f, 0.f),
    C_Vertex3f(0.179397583f, 0.558080673f, -1.f),
    C_Normal3f(0.652620137f, -0.757685304f, 0.f),
    C_Vertex3f(0.174089432f, 0.554149628f, 0.f),
    C_Vertex3f(0.174089432f, 0.554149628f, -1.f),
    C_Normal3f(0.744249165f, -0.667902052f, 0.f),
    C_Vertex3f(0.171791077f, 0.551528931f, 0.f),
    C_Vertex3f(0.171791077f, 0.551528931f, -1.f),
    C_Normal3f(0.837037444f, -0.547145605f, 0.f),
    C_Vertex3f(0.169384003f, 0.548906326f, 0.f),
    C_Vertex3f(0.169384003f, 0.548906326f, -1.f),
    C_Normal3f(0.925027549f, -0.379899919f, 0.f),
    C_Vertex3f(0.165827751f, 0.542406082f, 0.f),
    C_Vertex3f(0.165827751f, 0.542406082f, -1.f),
    C_Normal3f(0.980741322f, -0.195311442f, 0.f),
    C_Vertex3f(0.163529396f, 0.534650803f, 0.f),
    C_Vertex3f(0.163529396f, 0.534650803f, -1.f),
    C_Normal3f(0.997837007f, -0.065737471f, 0.f),
    C_Vertex3f(0.162488937f, 0.525640488f, 0.f),
    C_Vertex3f(0.162488937f, 0.525640488f, -1.f),
    C_Normal3f(0.999775767f, 0.0211769659f, 0.f),
    C_Vertex3f(0.162597656f, 0.520507813f, 0.f),
    C_Vertex3f(0.162597656f, 0.520507813f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.162597656f, 0.520507813f, 0.f),
    C_Vertex3f(0.162597656f, 0.520507813f, -1.f),
    C_Vertex3f(0.104492188f, 0.520507813f, -1.f),
    C_Vertex3f(0.104492188f, 0.520507813f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162597656f, 0.520507813f, 0.f),
    C_Vertex3f(0.162488937f, 0.525640488f, 0.f),
    C_Vertex3f(0.14986515f, 0.609520912f, 0.f),
    C_Vertex3f(0.135650635f, 0.599412918f, 0.f),
    C_Vertex3f(0.126247406f, 0.589304924f, 0.f),
    C_Vertex3f(0.120713472f, 0.581995726f, 0.f),
    C_Vertex3f(0.115985632f, 0.574115992f, 0.f),
    C_Vertex3f(0.112063885f, 0.565665722f, 0.f),
    C_Vertex3f(0.108948231f, 0.556644917f, 0.f),
    C_Vertex3f(0.10663867f, 0.547053576f, 0.f),
    C_Vertex3f(0.105135202f, 0.536891699f, 0.f),
    C_Vertex3f(0.104492188f, 0.520507813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.396972656f, 0.621582031f, 0.f),
    C_Vertex3f(0.340332031f, 0.621582031f, 0.f),
    C_Vertex3f(0.352399349f, 0.531311035f, 0.f),
    C_Vertex3f(0.36591053f, 0.540649414f, 0.f),
    C_Vertex3f(0.374835491f, 0.549987793f, 0.f),
    C_Vertex3f(0.380082726f, 0.556991577f, 0.f),
    C_Vertex3f(0.384627223f, 0.56477356f, 0.f),
    C_Vertex3f(0.388468981f, 0.57333374f, 0.f),
    C_Vertex3f(0.391608f, 0.582672119f, 0.f),
    C_Vertex3f(0.39404428f, 0.592788696f, 0.f),
    C_Vertex3f(0.395777822f, 0.603683472f, 0.f),
    C_Vertex3f(0.396808624f, 0.615356445f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.340332031f, 0.621582031f, 0.f),
    C_Vertex3f(0.339565277f, 0.615890503f, 0.f),
    C_Vertex3f(0.352399349f, 0.531311035f, 0.f),
    C_Vertex3f(0.337484837f, 0.606395721f, 0.f),
    C_Vertex3f(0.337276936f, 0.525085449f, 0.f),
    C_Vertex3f(0.334857464f, 0.598789215f, 0.f),
    C_Vertex3f(0.331683159f, 0.593070984f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.331683159f, 0.593070984f, 0.f),
    C_Vertex3f(0.327961922f, 0.589241028f, 0.f),
    C_Vertex3f(0.337276936f, 0.525085449f, 0.f),
    C_Vertex3f(0.323420048f, 0.586368561f, 0.f),
    C_Vertex3f(0.320543289f, 0.521972656f, 0.f),
    C_Vertex3f(0.318057537f, 0.584453583f, 0.f),
    C_Vertex3f(0.31187439f, 0.583496094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.285283089f, 0.588334084f, 0.f),
    C_Vertex3f(0.288539886f, 0.525018692f, 0.f),
    C_Vertex3f(0.29556942f, 0.585108757f, 0.f),
    C_Vertex3f(0.30449295f, 0.521972656f, 0.f),
    C_Vertex3f(0.304542542f, 0.583496094f, 0.f),
    C_Vertex3f(0.320543289f, 0.521972656f, 0.f),
    C_Vertex3f(0.31187439f, 0.583496094f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.218743324f, 0.556135178f, 0.f),
    C_Vertex3f(0.231437683f, 0.550258636f, 0.f),
    C_Vertex3f(0.234560013f, 0.610917091f, 0.f),
    C_Vertex3f(0.250041962f, 0.540248871f, 0.f),
    C_Vertex3f(0.255516052f, 0.602205276f, 0.f),
    C_Vertex3f(0.270389557f, 0.531110764f, 0.f),
    C_Vertex3f(0.273683548f, 0.593172073f, 0.f),
    C_Vertex3f(0.288539886f, 0.525018692f, 0.f),
    C_Vertex3f(0.285283089f, 0.588334084f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.192148209f, 0.562011719f, 0.f),
    C_Vertex3f(0.199262619f, 0.562011719f, 0.f),
    C_Vertex3f(0.200046539f, 0.619628906f, 0.f),
    C_Vertex3f(0.208018303f, 0.560052872f, 0.f),
    C_Vertex3f(0.216070175f, 0.616724968f, 0.f),
    C_Vertex3f(0.218743324f, 0.556135178f, 0.f),
    C_Vertex3f(0.234560013f, 0.610917091f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.174089432f, 0.554149628f, 0.f),
    C_Vertex3f(0.179397583f, 0.558080673f, 0.f),
    C_Vertex3f(0.183538437f, 0.619628906f, 0.f),
    C_Vertex3f(0.185417175f, 0.56070137f, 0.f),
    C_Vertex3f(0.200046539f, 0.619628906f, 0.f),
    C_Vertex3f(0.192148209f, 0.562011719f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163529396f, 0.534650803f, 0.f),
    C_Vertex3f(0.165827751f, 0.542406082f, 0.f),
    C_Vertex3f(0.165827751f, 0.616259575f, 0.f),
    C_Vertex3f(0.169384003f, 0.548906326f, 0.f),
    C_Vertex3f(0.183538437f, 0.619628906f, 0.f),
    C_Vertex3f(0.171791077f, 0.551528931f, 0.f),
    C_Vertex3f(0.174089432f, 0.554149628f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.165827751f, 0.616259575f, 0.f),
    C_Vertex3f(0.14986515f, 0.609520912f, 0.f),
    C_Vertex3f(0.162488937f, 0.525640488f, 0.f),
    C_Vertex3f(0.163529396f, 0.534650803f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.104492188f, 0.520507813f, 0.f),
    C_Vertex3f(0.105135202f, 0.536891699f, 0.f),
    C_Vertex3f(0.104437828f, 0.526159286f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162597656f, 0.520507813f, -1.f),
    C_Vertex3f(0.162488937f, 0.525640488f, -1.f),
    C_Vertex3f(0.14986515f, 0.609520912f, -1.f),
    C_Vertex3f(0.135650635f, 0.599412918f, -1.f),
    C_Vertex3f(0.126247406f, 0.589304924f, -1.f),
    C_Vertex3f(0.120713472f, 0.581995726f, -1.f),
    C_Vertex3f(0.115985632f, 0.574115992f, -1.f),
    C_Vertex3f(0.112063885f, 0.565665722f, -1.f),
    C_Vertex3f(0.108948231f, 0.556644917f, -1.f),
    C_Vertex3f(0.10663867f, 0.547053576f, -1.f),
    C_Vertex3f(0.105135202f, 0.536891699f, -1.f),
    C_Vertex3f(0.104492188f, 0.520507813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.396972656f, 0.621582031f, -1.f),
    C_Vertex3f(0.340332031f, 0.621582031f, -1.f),
    C_Vertex3f(0.352399349f, 0.531311035f, -1.f),
    C_Vertex3f(0.36591053f, 0.540649414f, -1.f),
    C_Vertex3f(0.374835491f, 0.549987793f, -1.f),
    C_Vertex3f(0.380082726f, 0.556991577f, -1.f),
    C_Vertex3f(0.384627223f, 0.56477356f, -1.f),
    C_Vertex3f(0.388468981f, 0.57333374f, -1.f),
    C_Vertex3f(0.391608f, 0.582672119f, -1.f),
    C_Vertex3f(0.39404428f, 0.592788696f, -1.f),
    C_Vertex3f(0.395777822f, 0.603683472f, -1.f),
    C_Vertex3f(0.396808624f, 0.615356445f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.340332031f, 0.621582031f, -1.f),
    C_Vertex3f(0.339565277f, 0.615890503f, -1.f),
    C_Vertex3f(0.352399349f, 0.531311035f, -1.f),
    C_Vertex3f(0.337484837f, 0.606395721f, -1.f),
    C_Vertex3f(0.337276936f, 0.525085449f, -1.f),
    C_Vertex3f(0.334857464f, 0.598789215f, -1.f),
    C_Vertex3f(0.331683159f, 0.593070984f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.331683159f, 0.593070984f, -1.f),
    C_Vertex3f(0.327961922f, 0.589241028f, -1.f),
    C_Vertex3f(0.337276936f, 0.525085449f, -1.f),
    C_Vertex3f(0.323420048f, 0.586368561f, -1.f),
    C_Vertex3f(0.320543289f, 0.521972656f, -1.f),
    C_Vertex3f(0.318057537f, 0.584453583f, -1.f),
    C_Vertex3f(0.31187439f, 0.583496094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.285283089f, 0.588334084f, -1.f),
    C_Vertex3f(0.288539886f, 0.525018692f, -1.f),
    C_Vertex3f(0.29556942f, 0.585108757f, -1.f),
    C_Vertex3f(0.30449295f, 0.521972656f, -1.f),
    C_Vertex3f(0.304542542f, 0.583496094f, -1.f),
    C_Vertex3f(0.320543289f, 0.521972656f, -1.f),
    C_Vertex3f(0.31187439f, 0.583496094f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.218743324f, 0.556135178f, -1.f),
    C_Vertex3f(0.231437683f, 0.550258636f, -1.f),
    C_Vertex3f(0.234560013f, 0.610917091f, -1.f),
    C_Vertex3f(0.250041962f, 0.540248871f, -1.f),
    C_Vertex3f(0.255516052f, 0.602205276f, -1.f),
    C_Vertex3f(0.270389557f, 0.531110764f, -1.f),
    C_Vertex3f(0.273683548f, 0.593172073f, -1.f),
    C_Vertex3f(0.288539886f, 0.525018692f, -1.f),
    C_Vertex3f(0.285283089f, 0.588334084f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.192148209f, 0.562011719f, -1.f),
    C_Vertex3f(0.199262619f, 0.562011719f, -1.f),
    C_Vertex3f(0.200046539f, 0.619628906f, -1.f),
    C_Vertex3f(0.208018303f, 0.560052872f, -1.f),
    C_Vertex3f(0.216070175f, 0.616724968f, -1.f),
    C_Vertex3f(0.218743324f, 0.556135178f, -1.f),
    C_Vertex3f(0.234560013f, 0.610917091f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.174089432f, 0.554149628f, -1.f),
    C_Vertex3f(0.179397583f, 0.558080673f, -1.f),
    C_Vertex3f(0.183538437f, 0.619628906f, -1.f),
    C_Vertex3f(0.185417175f, 0.56070137f, -1.f),
    C_Vertex3f(0.200046539f, 0.619628906f, -1.f),
    C_Vertex3f(0.192148209f, 0.562011719f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163529396f, 0.534650803f, -1.f),
    C_Vertex3f(0.165827751f, 0.542406082f, -1.f),
    C_Vertex3f(0.165827751f, 0.616259575f, -1.f),
    C_Vertex3f(0.169384003f, 0.548906326f, -1.f),
    C_Vertex3f(0.183538437f, 0.619628906f, -1.f),
    C_Vertex3f(0.171791077f, 0.551528931f, -1.f),
    C_Vertex3f(0.174089432f, 0.554149628f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.165827751f, 0.616259575f, -1.f),
    C_Vertex3f(0.14986515f, 0.609520912f, -1.f),
    C_Vertex3f(0.162488937f, 0.525640488f, -1.f),
    C_Vertex3f(0.163529396f, 0.534650803f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.104492188f, 0.520507813f, -1.f),
    C_Vertex3f(0.105135202f, 0.536891699f, -1.f),
    C_Vertex3f(0.104437828f, 0.526159286f, -1.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_246[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_Normal3f(-0.999325812f, 0.0367154181f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Normal3f(-0.993402183f, 0.114683464f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Normal3f(-0.979182839f, 0.20298022f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Normal3f(-0.953921318f, 0.300056934f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Normal3f(-0.914828122f, 0.403843552f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Normal3f(-0.859796643f, 0.510636568f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Normal3f(-0.788298845f, 0.615292549f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Normal3f(-0.703678906f, 0.710518122f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Normal3f(-0.620272636f, 0.784386277f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Normal3f(-0.543085396f, 0.839677453f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Normal3f(-0.461994827f, 0.886882663f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Normal3f(-0.376675963f, 0.92634511f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Normal3f(-0.289430827f, 0.957198918f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Normal3f(-0.202602044f, 0.97926122f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Normal3f(-0.118287019f, 0.992979467f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Normal3f(-0.0370853581f, 0.999312103f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Normal3f(0.0429813564f, 0.99907589f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Normal3f(0.130778968f, 0.991411507f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Normal3f(0.22524552f, 0.974301994f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Normal3f(0.322965264f, 0.946410835f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Normal3f(0.420876533f, 0.907118022f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Normal3f(0.515538931f, 0.856866181f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Normal3f(0.603679597f, 0.797226965f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Normal3f(0.682734966f, 0.730666041f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Normal3f(0.755707204f, 0.65490967f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Normal3f(0.823695898f, 0.56703198f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Normal3f(0.880775928f, 0.473533183f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Normal3f(0.925762534f, 0.378105432f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Normal3f(0.958753943f, 0.284237236f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Normal3f(0.980860114f, 0.194713563f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Normal3f(0.993775964f, 0.111397475f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Normal3f(0.999312639f, 0.0370702483f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Normal3f(0.999835014f, -0.0181650277f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_Normal3f(0.99813658f, -0.0610187873f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Normal3f(0.994201541f, -0.107533179f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Normal3f(0.987230599f, -0.159297675f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Normal3f(0.976227641f, -0.216747552f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Normal3f(0.95995146f, -0.280166566f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Normal3f(0.936915934f, -0.349554807f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Normal3f(0.90544951f, -0.424453944f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Normal3f(0.869160771f, -0.494529665f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Normal3f(0.831279218f, -0.555855095f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Normal3f(0.787997842f, -0.615678072f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Normal3f(0.739750624f, -0.672881126f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Normal3f(0.687237859f, -0.726432562f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Normal3f(0.631377041f, -0.775475979f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Normal3f(0.573225498f, -0.819397628f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Normal3f(0.507929087f, -0.861398876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Normal3f(0.419512093f, -0.907749832f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Normal3f(0.30787161f, -0.951427937f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Normal3f(0.183039665f, -0.9831056f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Normal3f(0.0626576319f, -0.998035133f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Normal3f(-0.038599968f, -0.999254763f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Normal3f(-0.128493756f, -0.991710246f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Normal3f(-0.222021356f, -0.975041866f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Normal3f(-0.319475144f, -0.947594643f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Normal3f(-0.417872697f, -0.908505559f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Normal3f(-0.513709545f, -0.857964098f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Normal3f(-0.60350287f, -0.797360897f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Normal3f(-0.684390008f, -0.729116023f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Normal3f(-0.760332525f, -0.649533927f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Normal3f(-0.830618083f, -0.556842566f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Normal3f(-0.887843728f, -0.460145056f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Normal3f(-0.931549191f, -0.363615453f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Normal3f(-0.962654054f, -0.270734549f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Normal3f(-0.982942224f, -0.183914542f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Normal3f(-0.994526625f, -0.104483135f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Normal3f(-0.999485016f, -0.0320912637f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Normal3f(0.999606252f, 0.0280574113f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Normal3f(0.995970607f, 0.0896810889f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Normal3f(0.987246752f, 0.159197494f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Normal3f(0.971514463f, 0.236980483f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Normal3f(0.946460783f, 0.322818667f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Normal3f(0.909567118f, 0.415557086f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Normal3f(0.858534515f, 0.51275593f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Normal3f(0.791965246f, 0.610566258f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Normal3f(0.720430911f, 0.693526745f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Normal3f(0.646434367f, 0.762969673f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Normal3f(0.5606305f, 0.828066111f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Normal3f(0.464660376f, 0.885488927f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Normal3f(0.361490935f, 0.932375669f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Normal3f(0.25501439f, 0.966937244f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Normal3f(0.149346635f, 0.988784969f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Normal3f(0.0481923707f, 0.998838007f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0485911854f, 0.998818815f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Normal3f(-0.151025206f, 0.988529861f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Normal3f(-0.257574052f, 0.966258526f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Normal3f(-0.36460194f, 0.93116349f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Normal3f(-0.467931896f, 0.883764625f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Normal3f(-0.563691318f, 0.825985491f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Normal3f(-0.649005592f, 0.760783732f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Normal3f(-0.72236073f, 0.69151628f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Normal3f(-0.794082761f, 0.607809663f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Normal3f(-0.861015081f, 0.508579373f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_Normal3f(-0.911770999f, 0.410699069f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Normal3f(-0.948089004f, 0.318005204f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Normal3f(-0.972527146f, 0.232789084f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Normal3f(-0.987754703f, 0.156014875f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Normal3f(-0.99614507f, 0.0877208859f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Normal3f(-0.999613106f, 0.0278154202f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Normal3f(-0.999587059f, -0.0287333243f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Normal3f(-0.995663822f, -0.0930251554f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Normal3f(-0.986366749f, -0.164561793f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Normal3f(-0.969791949f, -0.243933469f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Normal3f(-0.943758905f, -0.330634564f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Normal3f(-0.906032026f, -0.423209161f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Normal3f(-0.854762733f, -0.519018829f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Normal3f(-0.789101183f, -0.614263177f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Normal3f(-0.719142795f, -0.694862306f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Normal3f(-0.64632684f, -0.763060749f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Normal3f(-0.561695635f, -0.827343881f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Normal3f(-0.466670901f, -0.884431064f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Normal3f(-0.364006966f, -0.931396246f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Normal3f(-0.257463753f, -0.966287971f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Normal3f(-0.151149184f, -0.988511026f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Normal3f(-0.0485963188f, -0.998818517f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Normal3f(0.0480433144f, -0.99884522f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Normal3f(0.148618564f, -0.988894641f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Normal3f(0.253825158f, -0.967250109f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Normal3f(0.359922737f, -0.932982147f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Normal3f(0.462841064f, -0.886441231f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Normal3f(0.558709502f, -0.829363465f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Normal3f(0.644552469f, -0.764560103f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Normal3f(0.718696654f, -0.695323706f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Normal3f(0.790661871f, -0.612253129f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Normal3f(0.857762754f, -0.514045715f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Normal3f(0.909156442f, -0.416454464f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Normal3f(0.946268499f, -0.32338205f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Normal3f(0.971438348f, -0.237291992f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Normal3f(0.987223685f, -0.159340307f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Normal3f(0.995966375f, -0.0897267833f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Normal3f(0.999606252f, -0.0280619022f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.132324219f, 0.521972656f, 0.f),
    C_Vertex3f(0.132324219f, 0.521972656f, -1.f),
    C_Vertex3f(0.132324219f, 0.611328125f, -1.f),
    C_Vertex3f(0.132324219f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.132324219f, 0.611328125f, 0.f),
    C_Vertex3f(0.132324219f, 0.611328125f, -1.f),
    C_Vertex3f(0.214355469f, 0.611328125f, -1.f),
    C_Vertex3f(0.214355469f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.214355469f, 0.611328125f, 0.f),
    C_Vertex3f(0.214355469f, 0.611328125f, -1.f),
    C_Vertex3f(0.214355469f, 0.521972656f, -1.f),
    C_Vertex3f(0.214355469f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.214355469f, 0.521972656f, 0.f),
    C_Vertex3f(0.214355469f, 0.521972656f, -1.f),
    C_Vertex3f(0.132324219f, 0.521972656f, -1.f),
    C_Vertex3f(0.132324219f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.295410156f, 0.521972656f, 0.f),
    C_Vertex3f(0.295410156f, 0.521972656f, -1.f),
    C_Vertex3f(0.295410156f, 0.611328125f, -1.f),
    C_Vertex3f(0.295410156f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.295410156f, 0.611328125f, 0.f),
    C_Vertex3f(0.295410156f, 0.611328125f, -1.f),
    C_Vertex3f(0.377441406f, 0.611328125f, -1.f),
    C_Vertex3f(0.377441406f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.377441406f, 0.611328125f, 0.f),
    C_Vertex3f(0.377441406f, 0.611328125f, -1.f),
    C_Vertex3f(0.377441406f, 0.521972656f, -1.f),
    C_Vertex3f(0.377441406f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.377441406f, 0.521972656f, 0.f),
    C_Vertex3f(0.377441406f, 0.521972656f, -1.f),
    C_Vertex3f(0.295410156f, 0.521972656f, -1.f),
    C_Vertex3f(0.295410156f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.461837411f, 0.186348081f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_Vertex3f(0.463516116f, 0.206345916f, 0.f),
    C_Vertex3f(0.464355469f, 0.252545357f, 0.f),
    C_Vertex3f(0.464355469f, 0.227538109f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.45176518f, 0.133520722f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_Vertex3f(0.455961943f, 0.149935484f, 0.f),
    C_Vertex3f(0.458664179f, 0.303317308f, 0.f),
    C_Vertex3f(0.459319353f, 0.167544603f, 0.f),
    C_Vertex3f(0.462458372f, 0.278678656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.434138775f, 0.091442585f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_Vertex3f(0.440853596f, 0.104274273f, 0.f),
    C_Vertex3f(0.445384502f, 0.348110676f, 0.f),
    C_Vertex3f(0.446729064f, 0.118300319f, 0.f),
    C_Vertex3f(0.452972889f, 0.326461315f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, 0.f),
    C_Vertex3f(0.424516439f, 0.386925459f, 0.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, 0.f),
    C_Vertex3f(0.435899019f, 0.36826539f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, 0.f),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, 0.f),
    C_Vertex3f(0.411236763f, 0.404090881f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_Vertex3f(0.371108055f, 0.318627357f, 0.f),
    C_Vertex3f(0.375985146f, 0.302313805f, 0.f),
    C_Vertex3f(0.379642963f, 0.284695625f, 0.f),
    C_Vertex3f(0.382081509f, 0.26577282f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.339406967f, 0.370835304f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_Vertex3f(0.349161148f, 0.359740257f, 0.f),
    C_Vertex3f(0.362430573f, 0.445226669f, 0.f),
    C_Vertex3f(0.357696056f, 0.347340584f, 0.f),
    C_Vertex3f(0.365011692f, 0.333636284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_Vertex3f(0.318121195f, 0.388864756f, 0.f),
    C_Vertex3f(0.329060316f, 0.38054347f, 0.f),
    C_Vertex3f(0.343977928f, 0.455020905f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.28174901f, 0.405507326f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_Vertex3f(0.294465542f, 0.401346684f, 0.f),
    C_Vertex3f(0.303796768f, 0.468732834f, 0.f),
    C_Vertex3f(0.306589603f, 0.39579916f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_Vertex3f(0.254538536f, 0.409667969f, 0.f),
    C_Vertex3f(0.268440008f, 0.408281088f, 0.f),
    C_Vertex3f(0.282068253f, 0.472650528f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.212388396f, 0.405528069f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_Vertex3f(0.225848317f, 0.408288002f, 0.f),
    C_Vertex3f(0.236536026f, 0.474609375f, 0.f),
    C_Vertex3f(0.239934921f, 0.409667969f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.175768733f, 0.388968468f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_Vertex3f(0.187348604f, 0.395868301f, 0.f),
    C_Vertex3f(0.195970774f, 0.469777107f, 0.f),
    C_Vertex3f(0.199555159f, 0.401388168f, 0.f),
    C_Vertex3f(0.215843916f, 0.472998619f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_Vertex3f(0.15448904f, 0.3710289f, 0.f),
    C_Vertex3f(0.164815545f, 0.380688667f, 0.f),
    C_Vertex3f(0.176916599f, 0.464944839f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_Vertex3f(0.122964621f, 0.318425655f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.129027009f, 0.333687544f, 0.f),
    C_Vertex3f(0.141265154f, 0.450448036f, 0.f),
    C_Vertex3f(0.136301875f, 0.347542048f, 0.f),
    C_Vertex3f(0.144789219f, 0.359989166f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.110839844f, 0.243304253f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.112052321f, 0.264195681f, 0.f),
    C_Vertex3f(0.124667883f, 0.440783501f, 0.f),
    C_Vertex3f(0.114477277f, 0.283679724f, 0.f),
    C_Vertex3f(0.11811471f, 0.301756382f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, 0.f),
    C_Vertex3f(0.383300781f, 0.245545387f, 0.f),
    C_Vertex3f(0.383300781f, 0.223356247f, 0.f),
    C_Vertex3f(0.382088304f, 0.201972127f, 0.f),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, 0.f),
    C_Vertex3f(0.376025915f, 0.163590193f, 0.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, 0.f),
    C_Vertex3f(0.371176004f, 0.146592379f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.365113616f, 0.131056666f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, 0.f),
    C_Vertex3f(0.35783875f, 0.116983056f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, 0.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, 0.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, 0.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, 0.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, 0.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, 0.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, 0.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, 0.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, 0.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, 0.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, 0.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, 0.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, 0.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, 0.f),
    C_Vertex3f(0.199555159f, 0.062520504f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, 0.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, 0.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, 0.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, 0.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, 0.f),
    C_Vertex3f(0.144789219f, 0.10412693f, 0.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, 0.f),
    C_Vertex3f(0.136301875f, 0.116622448f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.129027009f, 0.130518436f, 0.f),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, 0.f),
    C_Vertex3f(0.11811471f, 0.162511826f, 0.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, 0.f),
    C_Vertex3f(0.114477277f, 0.180609226f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.112052321f, 0.200107098f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, 0.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, 0.f),
    C_Vertex3f(0.110839844f, 0.22100544f, 0.f),
    C_Vertex3f(0.10888958f, 0.429508209f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, 0.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, 0.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, 0.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, 0.f),
    C_Vertex3f(0.043214798f, 0.332399368f, 0.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, 0.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, 0.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, 0.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, 0.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, 0.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, 0.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, 0.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.132324219f, 0.611328125f, 0.f),
    C_Vertex3f(0.132324219f, 0.521972656f, 0.f),
    C_Vertex3f(0.214355469f, 0.521972656f, 0.f),
    C_Vertex3f(0.214355469f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.295410156f, 0.611328125f, 0.f),
    C_Vertex3f(0.295410156f, 0.521972656f, 0.f),
    C_Vertex3f(0.377441406f, 0.521972656f, 0.f),
    C_Vertex3f(0.377441406f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0316618681f, 0.189834952f, 0.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, 0.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.461837411f, 0.186348081f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_Vertex3f(0.463516116f, 0.206345916f, -1.f),
    C_Vertex3f(0.464355469f, 0.252545357f, -1.f),
    C_Vertex3f(0.464355469f, 0.227538109f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.45176518f, 0.133520722f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_Vertex3f(0.455961943f, 0.149935484f, -1.f),
    C_Vertex3f(0.458664179f, 0.303317308f, -1.f),
    C_Vertex3f(0.459319353f, 0.167544603f, -1.f),
    C_Vertex3f(0.462458372f, 0.278678656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.434138775f, 0.091442585f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_Vertex3f(0.440853596f, 0.104274273f, -1.f),
    C_Vertex3f(0.445384502f, 0.348110676f, -1.f),
    C_Vertex3f(0.446729064f, 0.118300319f, -1.f),
    C_Vertex3f(0.452972889f, 0.326461315f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.409408092f, 0.0572476387f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_Vertex3f(0.418416023f, 0.0679292679f, -1.f),
    C_Vertex3f(0.424516439f, 0.386925459f, -1.f),
    C_Vertex3f(0.426659584f, 0.0793275833f, -1.f),
    C_Vertex3f(0.435899019f, 0.36826539f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.39605999f, 0.419761658f, -1.f),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.399635792f, 0.0472826958f, -1.f),
    C_Vertex3f(0.411236763f, 0.404090881f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.37979126f, 0.433473587f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_Vertex3f(0.371108055f, 0.318627357f, -1.f),
    C_Vertex3f(0.375985146f, 0.302313805f, -1.f),
    C_Vertex3f(0.379642963f, 0.284695625f, -1.f),
    C_Vertex3f(0.382081509f, 0.26577282f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.339406967f, 0.370835304f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_Vertex3f(0.349161148f, 0.359740257f, -1.f),
    C_Vertex3f(0.362430573f, 0.445226669f, -1.f),
    C_Vertex3f(0.357696056f, 0.347340584f, -1.f),
    C_Vertex3f(0.365011692f, 0.333636284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.324433327f, 0.462856293f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_Vertex3f(0.318121195f, 0.388864756f, -1.f),
    C_Vertex3f(0.329060316f, 0.38054347f, -1.f),
    C_Vertex3f(0.343977928f, 0.455020905f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.28174901f, 0.405507326f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_Vertex3f(0.294465542f, 0.401346684f, -1.f),
    C_Vertex3f(0.303796768f, 0.468732834f, -1.f),
    C_Vertex3f(0.306589603f, 0.39579916f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.25924778f, 0.474609375f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_Vertex3f(0.254538536f, 0.409667969f, -1.f),
    C_Vertex3f(0.268440008f, 0.408281088f, -1.f),
    C_Vertex3f(0.282068253f, 0.472650528f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.212388396f, 0.405528069f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_Vertex3f(0.225848317f, 0.408288002f, -1.f),
    C_Vertex3f(0.236536026f, 0.474609375f, -1.f),
    C_Vertex3f(0.239934921f, 0.409667969f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.175768733f, 0.388968468f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_Vertex3f(0.187348604f, 0.395868301f, -1.f),
    C_Vertex3f(0.195970774f, 0.469777107f, -1.f),
    C_Vertex3f(0.199555159f, 0.401388168f, -1.f),
    C_Vertex3f(0.215843916f, 0.472998619f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.158681393f, 0.458501816f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_Vertex3f(0.15448904f, 0.3710289f, -1.f),
    C_Vertex3f(0.164815545f, 0.380688667f, -1.f),
    C_Vertex3f(0.176916599f, 0.464944839f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_Vertex3f(0.122964621f, 0.318425655f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.129027009f, 0.333687544f, -1.f),
    C_Vertex3f(0.141265154f, 0.450448036f, -1.f),
    C_Vertex3f(0.136301875f, 0.347542048f, -1.f),
    C_Vertex3f(0.144789219f, 0.359989166f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.110839844f, 0.243304253f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.112052321f, 0.264195681f, -1.f),
    C_Vertex3f(0.124667883f, 0.440783501f, -1.f),
    C_Vertex3f(0.114477277f, 0.283679724f, -1.f),
    C_Vertex3f(0.11811471f, 0.301756382f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.389099121f, 0.0380344391f, -1.f),
    C_Vertex3f(0.383300781f, 0.245545387f, -1.f),
    C_Vertex3f(0.383300781f, 0.223356247f, -1.f),
    C_Vertex3f(0.382088304f, 0.201972127f, -1.f),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379663348f, 0.182050109f, -1.f),
    C_Vertex3f(0.376025915f, 0.163590193f, -1.f),
    C_Vertex3f(0.37779808f, 0.0295028687f, -1.f),
    C_Vertex3f(0.371176004f, 0.146592379f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.365113616f, 0.131056666f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.349351406f, 0.104371548f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.36573267f, 0.0216879845f, -1.f),
    C_Vertex3f(0.35783875f, 0.116983056f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.329338908f, 0.0834664106f, -1.f),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.346487999f, 0.0110406876f, -1.f),
    C_Vertex3f(0.339651585f, 0.0932221413f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.318413377f, 0.0751043558f, -1.f),
    C_Vertex3f(0.30687499f, 0.0681359768f, -1.f),
    C_Vertex3f(0.319654942f, 0.000393390656f, -1.f),
    C_Vertex3f(0.294723749f, 0.0625612736f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.281959653f, 0.0583802462f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.262468338f, -0.0102539063f, -1.f),
    C_Vertex3f(0.291648388f, -0.00670480728f, -1.f),
    C_Vertex3f(0.268582702f, 0.0555928946f, -1.f),
    C_Vertex3f(0.254592896f, 0.0541992188f, -1.f),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239934921f, 0.0541992188f, -1.f),
    C_Vertex3f(0.225848317f, 0.0555860996f, -1.f),
    C_Vertex3f(0.235116959f, -0.0102539063f, -1.f),
    C_Vertex3f(0.212388396f, 0.0583598614f, -1.f),
    C_Vertex3f(0.211934686f, -0.00830185413f, -1.f),
    C_Vertex3f(0.199555159f, 0.062520504f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.187348604f, 0.0680680275f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.169070959f, 0.00145840645f, -1.f),
    C_Vertex3f(0.189919353f, -0.0043977499f, -1.f),
    C_Vertex3f(0.175768733f, 0.0750024319f, -1.f),
    C_Vertex3f(0.164815545f, 0.0833237171f, -1.f),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.15448904f, 0.0930318832f, -1.f),
    C_Vertex3f(0.144789219f, 0.10412693f, -1.f),
    C_Vertex3f(0.149389505f, 0.00926661491f, -1.f),
    C_Vertex3f(0.136301875f, 0.116622448f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.129027009f, 0.130518436f, -1.f),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.122964621f, 0.145814896f, -1.f),
    C_Vertex3f(0.11811471f, 0.162511826f, -1.f),
    C_Vertex3f(0.130874991f, 0.0190268755f, -1.f),
    C_Vertex3f(0.114477277f, 0.180609226f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.112052321f, 0.200107098f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0973467827f, 0.044403553f, -1.f),
    C_Vertex3f(0.113527417f, 0.0307391882f, -1.f),
    C_Vertex3f(0.110839844f, 0.22100544f, -1.f),
    C_Vertex3f(0.10888958f, 0.429508209f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0320234299f, 0.278411865f, -1.f),
    C_Vertex3f(0.0354152918f, 0.163899541f, -1.f),
    C_Vertex3f(0.0364999771f, 0.30644989f, -1.f),
    C_Vertex3f(0.0410454273f, 0.139683962f, -1.f),
    C_Vertex3f(0.043214798f, 0.332399368f, -1.f),
    C_Vertex3f(0.0485522747f, 0.117188215f, -1.f),
    C_Vertex3f(0.0521678925f, 0.3562603f, -1.f),
    C_Vertex3f(0.0579358339f, 0.0964123011f, -1.f),
    C_Vertex3f(0.0633592606f, 0.378032684f, -1.f),
    C_Vertex3f(0.069196105f, 0.0773562193f, -1.f),
    C_Vertex3f(0.0767889023f, 0.397716522f, -1.f),
    C_Vertex3f(0.0823330879f, 0.0600199699f, -1.f),
    C_Vertex3f(0.0924568176f, 0.415311813f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.132324219f, 0.611328125f, -1.f),
    C_Vertex3f(0.132324219f, 0.521972656f, -1.f),
    C_Vertex3f(0.214355469f, 0.521972656f, -1.f),
    C_Vertex3f(0.214355469f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.295410156f, 0.611328125f, -1.f),
    C_Vertex3f(0.295410156f, 0.521972656f, -1.f),
    C_Vertex3f(0.377441406f, 0.521972656f, -1.f),
    C_Vertex3f(0.377441406f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0316618681f, 0.189834952f, -1.f),
    C_Vertex3f(0.0297851563f, 0.248285294f, -1.f),
    C_Vertex3f(0.0297851563f, 0.217490196f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_247[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.200683594f, 0.403320313f, 0.f),
    C_Vertex3f(0.200683594f, 0.403320313f, -1.f),
    C_Vertex3f(0.200683594f, 0.492675781f, -1.f),
    C_Vertex3f(0.200683594f, 0.492675781f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.200683594f, 0.492675781f, 0.f),
    C_Vertex3f(0.200683594f, 0.492675781f, -1.f),
    C_Vertex3f(0.290039063f, 0.492675781f, -1.f),
    C_Vertex3f(0.290039063f, 0.492675781f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.290039063f, 0.492675781f, 0.f),
    C_Vertex3f(0.290039063f, 0.492675781f, -1.f),
    C_Vertex3f(0.290039063f, 0.403320313f, -1.f),
    C_Vertex3f(0.290039063f, 0.403320313f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.290039063f, 0.403320313f, 0.f),
    C_Vertex3f(0.290039063f, 0.403320313f, -1.f),
    C_Vertex3f(0.200683594f, 0.403320313f, -1.f),
    C_Vertex3f(0.200683594f, 0.403320313f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.45703125f, 0.280273438f, 0.f),
    C_Vertex3f(0.45703125f, 0.280273438f, -1.f),
    C_Vertex3f(0.0341796875f, 0.280273438f, -1.f),
    C_Vertex3f(0.0341796875f, 0.280273438f, 0.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0341796875f, 0.280273438f, 0.f),
    C_Vertex3f(0.0341796875f, 0.280273438f, -1.f),
    C_Vertex3f(0.0341796875f, 0.353515625f, -1.f),
    C_Vertex3f(0.0341796875f, 0.353515625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0341796875f, 0.353515625f, 0.f),
    C_Vertex3f(0.0341796875f, 0.353515625f, -1.f),
    C_Vertex3f(0.45703125f, 0.353515625f, -1.f),
    C_Vertex3f(0.45703125f, 0.353515625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.45703125f, 0.353515625f, 0.f),
    C_Vertex3f(0.45703125f, 0.353515625f, -1.f),
    C_Vertex3f(0.45703125f, 0.280273438f, -1.f),
    C_Vertex3f(0.45703125f, 0.280273438f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.200683594f, 0.141113281f, 0.f),
    C_Vertex3f(0.200683594f, 0.141113281f, -1.f),
    C_Vertex3f(0.200683594f, 0.23046875f, -1.f),
    C_Vertex3f(0.200683594f, 0.23046875f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.200683594f, 0.23046875f, 0.f),
    C_Vertex3f(0.200683594f, 0.23046875f, -1.f),
    C_Vertex3f(0.290039063f, 0.23046875f, -1.f),
    C_Vertex3f(0.290039063f, 0.23046875f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.290039063f, 0.23046875f, 0.f),
    C_Vertex3f(0.290039063f, 0.23046875f, -1.f),
    C_Vertex3f(0.290039063f, 0.141113281f, -1.f),
    C_Vertex3f(0.290039063f, 0.141113281f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.290039063f, 0.141113281f, 0.f),
    C_Vertex3f(0.290039063f, 0.141113281f, -1.f),
    C_Vertex3f(0.200683594f, 0.141113281f, -1.f),
    C_Vertex3f(0.200683594f, 0.141113281f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.200683594f, 0.492675781f, 0.f),
    C_Vertex3f(0.200683594f, 0.403320313f, 0.f),
    C_Vertex3f(0.290039063f, 0.403320313f, 0.f),
    C_Vertex3f(0.290039063f, 0.492675781f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0341796875f, 0.353515625f, 0.f),
    C_Vertex3f(0.0341796875f, 0.280273438f, 0.f),
    C_Vertex3f(0.45703125f, 0.280273438f, 0.f),
    C_Vertex3f(0.45703125f, 0.353515625f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.200683594f, 0.23046875f, 0.f),
    C_Vertex3f(0.200683594f, 0.141113281f, 0.f),
    C_Vertex3f(0.290039063f, 0.141113281f, 0.f),
    C_Vertex3f(0.290039063f, 0.23046875f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.200683594f, 0.492675781f, -1.f),
    C_Vertex3f(0.200683594f, 0.403320313f, -1.f),
    C_Vertex3f(0.290039063f, 0.403320313f, -1.f),
    C_Vertex3f(0.290039063f, 0.492675781f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0341796875f, 0.353515625f, -1.f),
    C_Vertex3f(0.0341796875f, 0.280273438f, -1.f),
    C_Vertex3f(0.45703125f, 0.280273438f, -1.f),
    C_Vertex3f(0.45703125f, 0.353515625f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.200683594f, 0.23046875f, -1.f),
    C_Vertex3f(0.200683594f, 0.141113281f, -1.f),
    C_Vertex3f(0.290039063f, 0.141113281f, -1.f),
    C_Vertex3f(0.290039063f, 0.23046875f, -1.f),
    C_End(),
    C_Translatef(0.491210938f, 0.f, 0.f),
};
static const Call char_248[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.792088211f, 0.610406756f, 0.f),
    C_Vertex3f(0.401763916f, 0.436027527f, 0.f),
    C_Vertex3f(0.401763916f, 0.436027527f, -1.f),
    C_Vertex3f(0.445083618f, 0.492240906f, -1.f),
    C_Vertex3f(0.445083618f, 0.492240906f, 0.f),
    C_Normal3f(0.585782528f, 0.810468316f, 0.f),
    C_Vertex3f(0.445083618f, 0.492240906f, 0.f),
    C_Vertex3f(0.445083618f, 0.492240906f, -1.f),
    C_Vertex3f(0.486831665f, 0.46206665f, -1.f),
    C_Vertex3f(0.486831665f, 0.46206665f, 0.f),
    C_Normal3f(0.792061269f, -0.610441506f, 0.f),
    C_Vertex3f(0.486831665f, 0.46206665f, 0.f),
    C_Vertex3f(0.486831665f, 0.46206665f, -1.f),
    C_Vertex3f(0.440032959f, 0.401344299f, -1.f),
    C_Vertex3f(0.440032959f, 0.401344299f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.806092381f, 0.591789782f, 0.f),
    C_Vertex3f(0.440032959f, 0.401344299f, 0.f),
    C_Vertex3f(0.440032959f, 0.401344299f, -1.f),
    C_Normal3f(0.827899635f, 0.56087631f, 0.f),
    C_Vertex3f(0.446922302f, 0.391960144f, 0.f),
    C_Vertex3f(0.446922302f, 0.391960144f, -1.f),
    C_Normal3f(0.856996894f, 0.515321612f, 0.f),
    C_Vertex3f(0.458514214f, 0.374064445f, 0.f),
    C_Vertex3f(0.458514214f, 0.374064445f, -1.f),
    C_Normal3f(0.894030988f, 0.448004991f, 0.f),
    C_Vertex3f(0.46791935f, 0.357041359f, 0.f),
    C_Vertex3f(0.46791935f, 0.357041359f, -1.f),
    C_Normal3f(0.925517976f, 0.378703624f, 0.f),
    C_Vertex3f(0.475137711f, 0.340890884f, 0.f),
    C_Vertex3f(0.475137711f, 0.340890884f, -1.f),
    C_Normal3f(0.952887118f, 0.303324968f, 0.f),
    C_Vertex3f(0.477653503f, 0.333251953f, 0.f),
    C_Vertex3f(0.477653503f, 0.333251953f, -1.f),
    C_Normal3f(0.971196473f, 0.238280237f, 0.f),
    C_Vertex3f(0.481042862f, 0.322340012f, 0.f),
    C_Vertex3f(0.481042862f, 0.322340012f, -1.f),
    C_Normal3f(0.985404253f, 0.170230925f, 0.f),
    C_Vertex3f(0.4861269f, 0.298715591f, 0.f),
    C_Vertex3f(0.4861269f, 0.298715591f, -1.f),
    C_Normal3f(0.995370269f, 0.0961150378f, 0.f),
    C_Vertex3f(0.489516258f, 0.273290634f, 0.f),
    C_Vertex3f(0.489516258f, 0.273290634f, -1.f),
    C_Normal3f(0.999575019f, 0.0291526988f, 0.f),
    C_Vertex3f(0.491210938f, 0.24606514f, 0.f),
    C_Vertex3f(0.491210938f, 0.24606514f, -1.f),
    C_Normal3f(0.999293506f, -0.0375824086f, 0.f),
    C_Vertex3f(0.491210938f, 0.215184212f, 0.f),
    C_Vertex3f(0.491210938f, 0.215184212f, -1.f),
    C_Normal3f(0.993158579f, -0.116773441f, 0.f),
    C_Vertex3f(0.488899946f, 0.184617281f, 0.f),
    C_Vertex3f(0.488899946f, 0.184617281f, -1.f),
    C_Normal3f(0.978352547f, -0.206944987f, 0.f),
    C_Vertex3f(0.484277964f, 0.156219244f, 0.f),
    C_Vertex3f(0.484277964f, 0.156219244f, -1.f),
    C_Normal3f(0.95195961f, -0.306223482f, 0.f),
    C_Vertex3f(0.47734499f, 0.129990101f, 0.f),
    C_Vertex3f(0.47734499f, 0.129990101f, -1.f),
    C_Normal3f(0.911023915f, -0.412353724f, 0.f),
    C_Vertex3f(0.468101025f, 0.105929852f, 0.f),
    C_Vertex3f(0.468101025f, 0.105929852f, -1.f),
    C_Normal3f(0.853370905f, -0.52130425f, 0.f),
    C_Vertex3f(0.456546068f, 0.084038496f, 0.f),
    C_Vertex3f(0.456546068f, 0.084038496f, -1.f),
    C_Normal3f(0.77860105f, -0.627519131f, 0.f),
    C_Vertex3f(0.44268012f, 0.0643160343f, 0.f),
    C_Vertex3f(0.44268012f, 0.0643160343f, -1.f),
    C_Normal3f(0.690401852f, -0.723426104f, 0.f),
    C_Vertex3f(0.426503181f, 0.0467624664f, 0.f),
    C_Vertex3f(0.426503181f, 0.0467624664f, -1.f),
    C_Normal3f(0.604395866f, -0.796684206f, 0.f),
    C_Vertex3f(0.409766197f, 0.0329046249f, 0.f),
    C_Vertex3f(0.409766197f, 0.0329046249f, -1.f),
    C_Normal3f(0.52751708f, -0.849544406f, 0.f),
    C_Vertex3f(0.39401412f, 0.0221149921f, 0.f),
    C_Vertex3f(0.39401412f, 0.0221149921f, -1.f),
    C_Normal3f(0.448355645f, -0.893855274f, 0.f),
    C_Vertex3f(0.377496004f, 0.0128667355f, 0.f),
    C_Vertex3f(0.377496004f, 0.0128667355f, -1.f),
    C_Normal3f(0.36544773f, -0.93083185f, 0.f),
    C_Vertex3f(0.360211849f, 0.00515985489f, 0.f),
    C_Vertex3f(0.360211849f, 0.00515985489f, -1.f),
    C_Normal3f(0.280885786f, -0.959741175f, 0.f),
    C_Vertex3f(0.342161655f, -0.00100564957f, 0.f),
    C_Vertex3f(0.342161655f, -0.00100564957f, -1.f),
    C_Normal3f(0.196783409f, -0.980447054f, 0.f),
    C_Vertex3f(0.323345423f, -0.00562977791f, 0.f),
    C_Vertex3f(0.323345423f, -0.00562977791f, -1.f),
    C_Normal3f(0.115035407f, -0.993361413f, 0.f),
    C_Vertex3f(0.303763151f, -0.00871253014f, 0.f),
    C_Vertex3f(0.303763151f, -0.00871253014f, -1.f),
    C_Normal3f(0.0388921835f, -0.999243379f, 0.f),
    C_Vertex3f(0.283414841f, -0.0102539063f, 0.f),
    C_Vertex3f(0.283414841f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0304043144f, -0.999537706f, 0.f),
    C_Vertex3f(0.26416111f, -0.0102539063f, 0.f),
    C_Vertex3f(0.26416111f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0986507237f, -0.995122135f, 0.f),
    C_Vertex3f(0.247273922f, -0.00915455818f, 0.f),
    C_Vertex3f(0.247273922f, -0.00915455818f, -1.f),
    C_Normal3f(-0.167985126f, -0.985789597f, 0.f),
    C_Vertex3f(0.230892658f, -0.00695586205f, 0.f),
    C_Vertex3f(0.230892658f, -0.00695586205f, -1.f),
    C_Normal3f(-0.239148572f, -0.970983028f, 0.f),
    C_Vertex3f(0.215017319f, -0.00365781784f, 0.f),
    C_Vertex3f(0.215017319f, -0.00365781784f, -1.f),
    C_Normal3f(-0.311031282f, -0.950399578f, 0.f),
    C_Vertex3f(0.199647903f, 0.000739574432f, 0.f),
    C_Vertex3f(0.199647903f, 0.000739574432f, -1.f),
    C_Normal3f(-0.382388413f, -0.924001694f, 0.f),
    C_Vertex3f(0.184784412f, 0.00623631477f, 0.f),
    C_Vertex3f(0.184784412f, 0.00623631477f, -1.f),
    C_Normal3f(-0.451936096f, -0.892050326f, 0.f),
    C_Vertex3f(0.170426846f, 0.0128324032f, 0.f),
    C_Vertex3f(0.170426846f, 0.0128324032f, -1.f),
    C_Normal3f(-0.507631004f, -0.86157459f, 0.f),
    C_Vertex3f(0.156575203f, 0.0205278397f, 0.f),
    C_Vertex3f(0.156575203f, 0.0205278397f, -1.f),
    C_Normal3f(-0.550258517f, -0.834994376f, 0.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, 0.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.792042136f, -0.61046648f, 0.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, 0.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, -1.f),
    C_Vertex3f(0.104347229f, -0.0341796875f, -1.f),
    C_Vertex3f(0.104347229f, -0.0341796875f, 0.f),
    C_Normal3f(-0.600805879f, -0.799394906f, 0.f),
    C_Vertex3f(0.104347229f, -0.0341796875f, 0.f),
    C_Vertex3f(0.104347229f, -0.0341796875f, -1.f),
    C_Vertex3f(0.0631942749f, -0.00325012207f, -1.f),
    C_Vertex3f(0.0631942749f, -0.00325012207f, 0.f),
    C_Normal3f(-0.792001247f, 0.610519409f, 0.f),
    C_Vertex3f(0.0631942749f, -0.00325012207f, 0.f),
    C_Vertex3f(0.0631942749f, -0.00325012207f, -1.f),
    C_Vertex3f(0.110496521f, 0.0581130981f, -1.f),
    C_Vertex3f(0.110496521f, 0.0581130981f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.754409492f, -0.656404018f, 0.f),
    C_Vertex3f(0.110496521f, 0.0581130981f, 0.f),
    C_Vertex3f(0.110496521f, 0.0581130981f, -1.f),
    C_Normal3f(-0.786022902f, -0.618197203f, 0.f),
    C_Vertex3f(0.104032516f, 0.0655422211f, 0.f),
    C_Vertex3f(0.104032516f, 0.0655422211f, -1.f),
    C_Normal3f(-0.826159477f, -0.563436389f, 0.f),
    C_Vertex3f(0.0929670334f, 0.0804014206f, 0.f),
    C_Vertex3f(0.0929670334f, 0.0804014206f, -1.f),
    C_Normal3f(-0.873762906f, -0.486352175f, 0.f),
    C_Vertex3f(0.0837640762f, 0.0952615738f, 0.f),
    C_Vertex3f(0.0837640762f, 0.0952615738f, -1.f),
    C_Normal3f(-0.921031356f, -0.38948831f, 0.f),
    C_Vertex3f(0.076423645f, 0.110122681f, 0.f),
    C_Vertex3f(0.076423645f, 0.110122681f, -1.f),
    C_Normal3f(-0.956194937f, -0.292730629f, 0.f),
    C_Vertex3f(0.0693016052f, 0.129461288f, 0.f),
    C_Vertex3f(0.0693016052f, 0.129461288f, -1.f),
    C_Normal3f(-0.97881788f, -0.204733357f, 0.f),
    C_Vertex3f(0.0627269745f, 0.154862404f, 0.f),
    C_Vertex3f(0.0627269745f, 0.154862404f, -1.f),
    C_Normal3f(-0.993071198f, -0.117513739f, 0.f),
    C_Vertex3f(0.0583438873f, 0.18184948f, 0.f),
    C_Vertex3f(0.0583438873f, 0.18184948f, -1.f),
    C_Normal3f(-0.999335706f, -0.0364432819f, 0.f),
    C_Vertex3f(0.0561523438f, 0.210422516f, 0.f),
    C_Vertex3f(0.0561523438f, 0.210422516f, -1.f),
    C_Normal3f(-0.999477327f, 0.0323258862f, 0.f),
    C_Vertex3f(0.0561523438f, 0.241945267f, 0.f),
    C_Vertex3f(0.0561523438f, 0.241945267f, -1.f),
    C_Normal3f(-0.994811177f, 0.101738416f, 0.f),
    C_Vertex3f(0.058161974f, 0.272557855f, 0.f),
    C_Vertex3f(0.058161974f, 0.272557855f, -1.f),
    C_Normal3f(-0.98336786f, 0.181625128f, 0.f),
    C_Vertex3f(0.0621812344f, 0.300896525f, 0.f),
    C_Vertex3f(0.0621812344f, 0.300896525f, -1.f),
    C_Normal3f(-0.962421596f, 0.271559924f, 0.f),
    C_Vertex3f(0.068210125f, 0.326961279f, 0.f),
    C_Vertex3f(0.068210125f, 0.326961279f, -1.f),
    C_Normal3f(-0.928734004f, 0.370746821f, 0.f),
    C_Vertex3f(0.0762486458f, 0.350752115f, 0.f),
    C_Vertex3f(0.0762486458f, 0.350752115f, -1.f),
    C_Normal3f(-0.879042149f, 0.476743966f, 0.f),
    C_Vertex3f(0.0862967968f, 0.372269034f, 0.f),
    C_Vertex3f(0.0862967968f, 0.372269034f, -1.f),
    C_Normal3f(-0.810976565f, 0.585078657f, 0.f),
    C_Vertex3f(0.098354578f, 0.391512036f, 0.f),
    C_Vertex3f(0.098354578f, 0.391512036f, -1.f),
    C_Normal3f(-0.724274695f, 0.689511478f, 0.f),
    C_Vertex3f(0.112421989f, 0.408481121f, 0.f),
    C_Vertex3f(0.112421989f, 0.408481121f, -1.f),
    C_Normal3f(-0.642620504f, 0.766184568f, 0.f),
    C_Vertex3f(0.128499031f, 0.423176289f, 0.f),
    C_Vertex3f(0.128499031f, 0.423176289f, -1.f),
    C_Normal3f(-0.571971297f, 0.820273697f, 0.f),
    C_Vertex3f(0.145273447f, 0.43603456f, 0.f),
    C_Vertex3f(0.145273447f, 0.43603456f, -1.f),
    C_Normal3f(-0.493184656f, 0.869924664f, 0.f),
    C_Vertex3f(0.162745237f, 0.447055936f, 0.f),
    C_Vertex3f(0.162745237f, 0.447055936f, -1.f),
    C_Normal3f(-0.407615215f, 0.913153827f, 0.f),
    C_Vertex3f(0.180914402f, 0.456240416f, 0.f),
    C_Vertex3f(0.180914402f, 0.456240416f, -1.f),
    C_Normal3f(-0.317296267f, 0.948326409f, 0.f),
    C_Vertex3f(0.199780941f, 0.463587999f, 0.f),
    C_Vertex3f(0.199780941f, 0.463587999f, -1.f),
    C_Normal3f(-0.224721253f, 0.974423051f, 0.f),
    C_Vertex3f(0.219344854f, 0.469098687f, 0.f),
    C_Vertex3f(0.219344854f, 0.469098687f, -1.f),
    C_Normal3f(-0.132510886f, 0.991181552f, 0.f),
    C_Vertex3f(0.239606142f, 0.472772479f, 0.f),
    C_Vertex3f(0.239606142f, 0.472772479f, -1.f),
    C_Normal3f(-0.0450259261f, 0.998985887f, 0.f),
    C_Vertex3f(0.260564804f, 0.474609375f, 0.f),
    C_Vertex3f(0.260564804f, 0.474609375f, -1.f),
    C_Normal3f(0.0314240493f, 0.999506176f, 0.f),
    C_Vertex3f(0.280361176f, 0.474609375f, 0.f),
    C_Vertex3f(0.280361176f, 0.474609375f, -1.f),
    C_Normal3f(0.10294231f, 0.994687259f, 0.f),
    C_Vertex3f(0.297586918f, 0.473445415f, 0.f),
    C_Vertex3f(0.297586918f, 0.473445415f, -1.f),
    C_Normal3f(0.177225277f, 0.984170258f, 0.f),
    C_Vertex3f(0.314101696f, 0.471117496f, 0.f),
    C_Vertex3f(0.314101696f, 0.471117496f, -1.f),
    C_Normal3f(0.254991293f, 0.966943324f, 0.f),
    C_Vertex3f(0.32990551f, 0.467625618f, 0.f),
    C_Vertex3f(0.32990551f, 0.467625618f, -1.f),
    C_Normal3f(0.327746481f, 0.944765687f, 0.f),
    C_Vertex3f(0.34499836f, 0.46296978f, 0.f),
    C_Vertex3f(0.34499836f, 0.46296978f, -1.f),
    C_Normal3f(0.387371033f, 0.921923876f, 0.f),
    C_Vertex3f(0.360583782f, 0.456983089f, 0.f),
    C_Vertex3f(0.360583782f, 0.456983089f, -1.f),
    C_Normal3f(0.439308226f, 0.898336351f, 0.f),
    C_Vertex3f(0.376661777f, 0.449665546f, 0.f),
    C_Vertex3f(0.376661777f, 0.449665546f, -1.f),
    C_Normal3f(0.47739318f, 0.878689826f, 0.f),
    C_Vertex3f(0.393232346f, 0.441017151f, 0.f),
    C_Vertex3f(0.393232346f, 0.441017151f, -1.f),
    C_Normal3f(0.504842639f, 0.863211393f, 0.f),
    C_Vertex3f(0.401763916f, 0.436027527f, 0.f),
    C_Vertex3f(0.401763916f, 0.436027527f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.616593361f, -0.787281871f, 0.f),
    C_Vertex3f(0.357444763f, 0.37852478f, 0.f),
    C_Vertex3f(0.357444763f, 0.37852478f, -1.f),
    C_Normal3f(-0.580032706f, -0.814593196f, 0.f),
    C_Vertex3f(0.352474213f, 0.382417679f, 0.f),
    C_Vertex3f(0.352474213f, 0.382417679f, -1.f),
    C_Normal3f(-0.529844582f, -0.848094821f, 0.f),
    C_Vertex3f(0.342410088f, 0.389230251f, 0.f),
    C_Vertex3f(0.342410088f, 0.389230251f, -1.f),
    C_Normal3f(-0.46294722f, -0.886385858f, 0.f),
    C_Vertex3f(0.332222939f, 0.395069599f, 0.f),
    C_Vertex3f(0.332222939f, 0.395069599f, -1.f),
    C_Normal3f(-0.388998777f, -0.921238244f, 0.f),
    C_Vertex3f(0.321912766f, 0.399935722f, 0.f),
    C_Vertex3f(0.321912766f, 0.399935722f, -1.f),
    C_Normal3f(-0.308717638f, -0.951153696f, 0.f),
    C_Vertex3f(0.311479568f, 0.403828621f, 0.f),
    C_Vertex3f(0.311479568f, 0.403828621f, -1.f),
    C_Normal3f(-0.223361433f, -0.974735737f, 0.f),
    C_Vertex3f(0.300923347f, 0.406748295f, 0.f),
    C_Vertex3f(0.300923347f, 0.406748295f, -1.f),
    C_Normal3f(-0.134677395f, -0.99088949f, 0.f),
    C_Vertex3f(0.290244102f, 0.408694744f, 0.f),
    C_Vertex3f(0.290244102f, 0.408694744f, -1.f),
    C_Normal3f(-0.0413273424f, -0.999145687f, 0.f),
    C_Vertex3f(0.279441833f, 0.409667969f, 0.f),
    C_Vertex3f(0.279441833f, 0.409667969f, -1.f),
    C_Normal3f(0.0522066765f, -0.998636305f, 0.f),
    C_Vertex3f(0.26671505f, 0.409667969f, 0.f),
    C_Vertex3f(0.26671505f, 0.409667969f, -1.f),
    C_Normal3f(0.151460588f, -0.988463223f, 0.f),
    C_Vertex3f(0.252773762f, 0.408273816f, 0.f),
    C_Vertex3f(0.252773762f, 0.408273816f, -1.f),
    C_Normal3f(0.257835269f, -0.966188908f, 0.f),
    C_Vertex3f(0.23941946f, 0.405485511f, 0.f),
    C_Vertex3f(0.23941946f, 0.405485511f, -1.f),
    C_Normal3f(0.364300549f, -0.931281507f, 0.f),
    C_Vertex3f(0.226652145f, 0.401303053f, 0.f),
    C_Vertex3f(0.226652145f, 0.401303053f, -1.f),
    C_Normal3f(0.466763228f, -0.884382248f, 0.f),
    C_Vertex3f(0.214471817f, 0.395726442f, 0.f),
    C_Vertex3f(0.214471817f, 0.395726442f, -1.f),
    C_Normal3f(0.561505675f, -0.827472866f, 0.f),
    C_Vertex3f(0.202878475f, 0.388755679f, 0.f),
    C_Vertex3f(0.202878475f, 0.388755679f, -1.f),
    C_Normal3f(0.645828545f, -0.763482451f, 0.f),
    C_Vertex3f(0.19187212f, 0.380390763f, 0.f),
    C_Vertex3f(0.19187212f, 0.380390763f, -1.f),
    C_Normal3f(0.718360901f, -0.695670664f, 0.f),
    C_Vertex3f(0.181452751f, 0.370631695f, 0.f),
    C_Vertex3f(0.181452751f, 0.370631695f, -1.f),
    C_Normal3f(0.78897661f, -0.614423215f, 0.f),
    C_Vertex3f(0.171620369f, 0.359478474f, 0.f),
    C_Vertex3f(0.171620369f, 0.359478474f, -1.f),
    C_Normal3f(0.855576754f, -0.517675817f, 0.f),
    C_Vertex3f(0.163017035f, 0.346958518f, 0.f),
    C_Vertex3f(0.163017035f, 0.346958518f, -1.f),
    C_Normal3f(0.907125711f, -0.420860022f, 0.f),
    C_Vertex3f(0.155642748f, 0.333071828f, 0.f),
    C_Vertex3f(0.155642748f, 0.333071828f, -1.f),
    C_Normal3f(0.944726765f, -0.327858597f, 0.f),
    C_Vertex3f(0.149497509f, 0.317818403f, 0.f),
    C_Vertex3f(0.149497509f, 0.317818403f, -1.f),
    C_Normal3f(0.97046119f, -0.24125725f, 0.f),
    C_Vertex3f(0.144581318f, 0.301198244f, 0.f),
    C_Vertex3f(0.144581318f, 0.301198244f, -1.f),
    C_Normal3f(0.98672688f, -0.162388533f, 0.f),
    C_Vertex3f(0.140894175f, 0.28321135f, 0.f),
    C_Vertex3f(0.140894175f, 0.28321135f, -1.f),
    C_Normal3f(0.995793879f, -0.0916213915f, 0.f),
    C_Vertex3f(0.138436079f, 0.263857722f, 0.f),
    C_Vertex3f(0.138436079f, 0.263857722f, -1.f),
    C_Normal3f(0.999527454f, -0.0307381786f, 0.f),
    C_Vertex3f(0.137207031f, 0.24313736f, 0.f),
    C_Vertex3f(0.137207031f, 0.24313736f, -1.f),
    C_Normal3f(0.99979347f, 0.020321615f, 0.f),
    C_Vertex3f(0.137207031f, 0.223892212f, 0.f),
    C_Vertex3f(0.137207031f, 0.223892212f, -1.f),
    C_Normal3f(0.997610331f, 0.0690920651f, 0.f),
    C_Vertex3f(0.137917995f, 0.20815897f, 0.f),
    C_Vertex3f(0.137917995f, 0.20815897f, -1.f),
    C_Normal3f(0.992796719f, 0.119810596f, 0.f),
    C_Vertex3f(0.139339924f, 0.193095684f, 0.f),
    C_Vertex3f(0.139339924f, 0.193095684f, -1.f),
    C_Normal3f(0.984693587f, 0.174294025f, 0.f),
    C_Vertex3f(0.141472816f, 0.178702354f, 0.f),
    C_Vertex3f(0.141472816f, 0.178702354f, -1.f),
    C_Normal3f(0.972615659f, 0.232419729f, 0.f),
    C_Vertex3f(0.144316673f, 0.164978981f, 0.f),
    C_Vertex3f(0.144316673f, 0.164978981f, -1.f),
    C_Normal3f(0.955843687f, 0.293875575f, 0.f),
    C_Vertex3f(0.147871494f, 0.151925564f, 0.f),
    C_Vertex3f(0.147871494f, 0.151925564f, -1.f),
    C_Normal3f(0.9336766f, 0.358117223f, 0.f),
    C_Vertex3f(0.15213728f, 0.139542103f, 0.f),
    C_Vertex3f(0.15213728f, 0.139542103f, -1.f),
    C_Normal3f(0.9106372f, 0.413206816f, 0.f),
    C_Vertex3f(0.157114029f, 0.127828598f, 0.f),
    C_Vertex3f(0.157114029f, 0.127828598f, -1.f),
    C_Normal3f(0.889020085f, 0.457868367f, 0.f),
    C_Vertex3f(0.159957886f, 0.122306824f, 0.f),
    C_Vertex3f(0.159957886f, 0.122306824f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.792031825f, 0.610479832f, 0.f),
    C_Vertex3f(0.159957886f, 0.122306824f, 0.f),
    C_Vertex3f(0.159957886f, 0.122306824f, -1.f),
    C_Vertex3f(0.357444763f, 0.37852478f, -1.f),
    C_Vertex3f(0.357444763f, 0.37852478f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.792052567f, -0.61045295f, 0.f),
    C_Vertex3f(0.389907837f, 0.33631897f, 0.f),
    C_Vertex3f(0.389907837f, 0.33631897f, -1.f),
    C_Vertex3f(0.193763733f, 0.0818252563f, -1.f),
    C_Vertex3f(0.193763733f, 0.0818252563f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.630304217f, 0.776348233f, 0.f),
    C_Vertex3f(0.193763733f, 0.0818252563f, 0.f),
    C_Vertex3f(0.193763733f, 0.0818252563f, -1.f),
    C_Normal3f(0.544299185f, 0.838891149f, 0.f),
    C_Vertex3f(0.202270508f, 0.0749187469f, 0.f),
    C_Vertex3f(0.202270508f, 0.0749187469f, -1.f),
    C_Normal3f(0.420077562f, 0.907488167f, 0.f),
    C_Vertex3f(0.220375061f, 0.0645589828f, 0.f),
    C_Vertex3f(0.220375061f, 0.0645589828f, -1.f),
    C_Normal3f(0.253799766f, 0.967256784f, 0.f),
    C_Vertex3f(0.239570618f, 0.0576524734f, 0.f),
    C_Vertex3f(0.239570618f, 0.0576524734f, -1.f),
    C_Normal3f(0.0892073363f, 0.996013105f, 0.f),
    C_Vertex3f(0.259857178f, 0.0541992188f, 0.f),
    C_Vertex3f(0.259857178f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0424402282f, 0.999098957f, 0.f),
    C_Vertex3f(0.278126717f, 0.0541992188f, 0.f),
    C_Vertex3f(0.278126717f, 0.0541992188f, -1.f),
    C_Normal3f(-0.1457856f, 0.989316165f, 0.f),
    C_Vertex3f(0.292626858f, 0.0555912256f, 0.f),
    C_Vertex3f(0.292626858f, 0.0555912256f, -1.f),
    C_Normal3f(-0.249478802f, 0.968380272f, 0.f),
    C_Vertex3f(0.306465626f, 0.0583752394f, 0.f),
    C_Vertex3f(0.306465626f, 0.0583752394f, -1.f),
    C_Normal3f(-0.354599178f, 0.93501842f, 0.f),
    C_Vertex3f(0.319643021f, 0.06255126f, 0.f),
    C_Vertex3f(0.319643021f, 0.06255126f, -1.f),
    C_Normal3f(-0.457191437f, 0.889368355f, 0.f),
    C_Vertex3f(0.332159042f, 0.0681192875f, 0.f),
    C_Vertex3f(0.332159042f, 0.0681192875f, -1.f),
    C_Normal3f(-0.553368211f, 0.832936645f, 0.f),
    C_Vertex3f(0.344013691f, 0.0750793219f, 0.f),
    C_Vertex3f(0.344013691f, 0.0750793219f, -1.f),
    C_Normal3f(-0.640012681f, 0.76836431f, 0.f),
    C_Vertex3f(0.355206966f, 0.0834313631f, 0.f),
    C_Vertex3f(0.355206966f, 0.0834313631f, -1.f),
    C_Normal3f(-0.715240717f, 0.698878229f, 0.f),
    C_Vertex3f(0.365738869f, 0.0931754112f, 0.f),
    C_Vertex3f(0.365738869f, 0.0931754112f, -1.f),
    C_Normal3f(-0.787709951f, 0.616046309f, 0.f),
    C_Vertex3f(0.375609398f, 0.104311466f, 0.f),
    C_Vertex3f(0.375609398f, 0.104311466f, -1.f),
    C_Normal3f(-0.855054438f, 0.518538296f, 0.f),
    C_Vertex3f(0.384246111f, 0.116839767f, 0.f),
    C_Vertex3f(0.384246111f, 0.116839767f, -1.f),
    C_Normal3f(-0.907016695f, 0.421094924f, 0.f),
    C_Vertex3f(0.391649008f, 0.130760312f, 0.f),
    C_Vertex3f(0.391649008f, 0.130760312f, -1.f),
    C_Normal3f(-0.944792032f, 0.327670485f, 0.f),
    C_Vertex3f(0.397818089f, 0.146073103f, 0.f),
    C_Vertex3f(0.397818089f, 0.146073103f, -1.f),
    C_Normal3f(-0.970560133f, 0.240858942f, 0.f),
    C_Vertex3f(0.402753353f, 0.162778139f, 0.f),
    C_Vertex3f(0.402753353f, 0.162778139f, -1.f),
    C_Normal3f(-0.986796975f, 0.161961883f, 0.f),
    C_Vertex3f(0.406454802f, 0.180875421f, 0.f),
    C_Vertex3f(0.406454802f, 0.180875421f, -1.f),
    C_Normal3f(-0.995823205f, 0.0913021937f, 0.f),
    C_Vertex3f(0.408922434f, 0.200364947f, 0.f),
    C_Vertex3f(0.408922434f, 0.200364947f, -1.f),
    C_Normal3f(-0.999531984f, 0.0305895042f, 0.f),
    C_Vertex3f(0.41015625f, 0.221246719f, 0.f),
    C_Vertex3f(0.41015625f, 0.221246719f, -1.f),
    C_Normal3f(-0.999829352f, -0.0184754934f, 0.f),
    C_Vertex3f(0.41015625f, 0.240680695f, 0.f),
    C_Vertex3f(0.41015625f, 0.240680695f, -1.f),
    C_Normal3f(-0.997926891f, -0.0643578842f, 0.f),
    C_Vertex3f(0.409509659f, 0.256237984f, 0.f),
    C_Vertex3f(0.409509659f, 0.256237984f, -1.f),
    C_Normal3f(-0.993402302f, -0.114681892f, 0.f),
    C_Vertex3f(0.408216476f, 0.270758629f, 0.f),
    C_Vertex3f(0.408216476f, 0.270758629f, -1.f),
    C_Normal3f(-0.986810207f, -0.16188176f, 0.f),
    C_Vertex3f(0.406276703f, 0.28424263f, 0.f),
    C_Vertex3f(0.406276703f, 0.28424263f, -1.f),
    C_Normal3f(-0.980726838f, -0.195383906f, 0.f),
    C_Vertex3f(0.404983521f, 0.290466309f, 0.f),
    C_Vertex3f(0.404983521f, 0.290466309f, -1.f),
    C_Normal3f(-0.971965611f, -0.235122874f, 0.f),
    C_Vertex3f(0.404123306f, 0.295051575f, 0.f),
    C_Vertex3f(0.404123306f, 0.295051575f, -1.f),
    C_Normal3f(-0.95807302f, -0.286524147f, 0.f),
    C_Vertex3f(0.401378632f, 0.305368423f, 0.f),
    C_Vertex3f(0.401378632f, 0.305368423f, -1.f),
    C_Normal3f(-0.942178488f, -0.33511126f, 0.f),
    C_Vertex3f(0.397609711f, 0.316831589f, 0.f),
    C_Vertex3f(0.397609711f, 0.316831589f, -1.f),
    C_Normal3f(-0.930000484f, -0.367558181f, 0.f),
    C_Vertex3f(0.392816544f, 0.329441071f, 0.f),
    C_Vertex3f(0.392816544f, 0.329441071f, -1.f),
    C_Normal3f(-0.921023607f, -0.389506638f, 0.f),
    C_Vertex3f(0.389907837f, 0.33631897f, 0.f),
    C_Vertex3f(0.389907837f, 0.33631897f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.076423645f, 0.110122681f, 0.f),
    C_Vertex3f(0.0837640762f, 0.0952615738f, 0.f),
    C_Vertex3f(0.0862967968f, 0.372269034f, 0.f),
    C_Vertex3f(0.0929670334f, 0.0804014206f, 0.f),
    C_Vertex3f(0.098354578f, 0.391512036f, 0.f),
    C_Vertex3f(0.104032516f, 0.0655422211f, 0.f),
    C_Vertex3f(0.112421989f, 0.408481121f, 0.f),
    C_Vertex3f(0.110496521f, 0.0581130981f, 0.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, 0.f),
    C_Vertex3f(0.104347229f, -0.0341796875f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.058161974f, 0.272557855f, 0.f),
    C_Vertex3f(0.0583438873f, 0.18184948f, 0.f),
    C_Vertex3f(0.0621812344f, 0.300896525f, 0.f),
    C_Vertex3f(0.0627269745f, 0.154862404f, 0.f),
    C_Vertex3f(0.068210125f, 0.326961279f, 0.f),
    C_Vertex3f(0.0693016052f, 0.129461288f, 0.f),
    C_Vertex3f(0.0762486458f, 0.350752115f, 0.f),
    C_Vertex3f(0.076423645f, 0.110122681f, 0.f),
    C_Vertex3f(0.0862967968f, 0.372269034f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0561523438f, 0.241945267f, 0.f),
    C_Vertex3f(0.0561523438f, 0.210422516f, 0.f),
    C_Vertex3f(0.0583438873f, 0.18184948f, 0.f),
    C_Vertex3f(0.058161974f, 0.272557855f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.332222939f, 0.395069599f, 0.f),
    C_Vertex3f(0.342410088f, 0.389230251f, 0.f),
    C_Vertex3f(0.34499836f, 0.46296978f, 0.f),
    C_Vertex3f(0.352474213f, 0.382417679f, 0.f),
    C_Vertex3f(0.360583782f, 0.456983089f, 0.f),
    C_Vertex3f(0.357444763f, 0.37852478f, 0.f),
    C_Vertex3f(0.389907837f, 0.33631897f, 0.f),
    C_Vertex3f(0.193763733f, 0.0818252563f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.300923347f, 0.406748295f, 0.f),
    C_Vertex3f(0.311479568f, 0.403828621f, 0.f),
    C_Vertex3f(0.314101696f, 0.471117496f, 0.f),
    C_Vertex3f(0.321912766f, 0.399935722f, 0.f),
    C_Vertex3f(0.32990551f, 0.467625618f, 0.f),
    C_Vertex3f(0.332222939f, 0.395069599f, 0.f),
    C_Vertex3f(0.34499836f, 0.46296978f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.26671505f, 0.409667969f, 0.f),
    C_Vertex3f(0.279441833f, 0.409667969f, 0.f),
    C_Vertex3f(0.280361176f, 0.474609375f, 0.f),
    C_Vertex3f(0.290244102f, 0.408694744f, 0.f),
    C_Vertex3f(0.297586918f, 0.473445415f, 0.f),
    C_Vertex3f(0.300923347f, 0.406748295f, 0.f),
    C_Vertex3f(0.314101696f, 0.471117496f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.226652145f, 0.401303053f, 0.f),
    C_Vertex3f(0.23941946f, 0.405485511f, 0.f),
    C_Vertex3f(0.239606142f, 0.472772479f, 0.f),
    C_Vertex3f(0.252773762f, 0.408273816f, 0.f),
    C_Vertex3f(0.260564804f, 0.474609375f, 0.f),
    C_Vertex3f(0.26671505f, 0.409667969f, 0.f),
    C_Vertex3f(0.280361176f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219344854f, 0.469098687f, 0.f),
    C_Vertex3f(0.199780941f, 0.463587999f, 0.f),
    C_Vertex3f(0.202878475f, 0.388755679f, 0.f),
    C_Vertex3f(0.214471817f, 0.395726442f, 0.f),
    C_Vertex3f(0.226652145f, 0.401303053f, 0.f),
    C_Vertex3f(0.239606142f, 0.472772479f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163017035f, 0.346958518f, 0.f),
    C_Vertex3f(0.171620369f, 0.359478474f, 0.f),
    C_Vertex3f(0.180914402f, 0.456240416f, 0.f),
    C_Vertex3f(0.181452751f, 0.370631695f, 0.f),
    C_Vertex3f(0.199780941f, 0.463587999f, 0.f),
    C_Vertex3f(0.19187212f, 0.380390763f, 0.f),
    C_Vertex3f(0.202878475f, 0.388755679f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162745237f, 0.447055936f, 0.f),
    C_Vertex3f(0.145273447f, 0.43603456f, 0.f),
    C_Vertex3f(0.149497509f, 0.317818403f, 0.f),
    C_Vertex3f(0.155642748f, 0.333071828f, 0.f),
    C_Vertex3f(0.163017035f, 0.346958518f, 0.f),
    C_Vertex3f(0.180914402f, 0.456240416f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145273447f, 0.43603456f, 0.f),
    C_Vertex3f(0.128499031f, 0.423176289f, 0.f),
    C_Vertex3f(0.137207031f, 0.24313736f, 0.f),
    C_Vertex3f(0.138436079f, 0.263857722f, 0.f),
    C_Vertex3f(0.140894175f, 0.28321135f, 0.f),
    C_Vertex3f(0.144581318f, 0.301198244f, 0.f),
    C_Vertex3f(0.149497509f, 0.317818403f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.141472816f, 0.178702354f, 0.f),
    C_Vertex3f(0.139339924f, 0.193095684f, 0.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, 0.f),
    C_Vertex3f(0.137917995f, 0.20815897f, 0.f),
    C_Vertex3f(0.112421989f, 0.408481121f, 0.f),
    C_Vertex3f(0.137207031f, 0.223892212f, 0.f),
    C_Vertex3f(0.128499031f, 0.423176289f, 0.f),
    C_Vertex3f(0.137207031f, 0.24313736f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.445083618f, 0.492240906f, 0.f),
    C_Vertex3f(0.401763916f, 0.436027527f, 0.f),
    C_Vertex3f(0.404123306f, 0.295051575f, 0.f),
    C_Vertex3f(0.404983521f, 0.290466309f, 0.f),
    C_Vertex3f(0.406276703f, 0.28424263f, 0.f),
    C_Vertex3f(0.408216476f, 0.270758629f, 0.f),
    C_Vertex3f(0.409509659f, 0.256237984f, 0.f),
    C_Vertex3f(0.41015625f, 0.240680695f, 0.f),
    C_Vertex3f(0.440032959f, 0.401344299f, 0.f),
    C_Vertex3f(0.486831665f, 0.46206665f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.475137711f, 0.340890884f, 0.f),
    C_Vertex3f(0.46791935f, 0.357041359f, 0.f),
    C_Vertex3f(0.468101025f, 0.105929852f, 0.f),
    C_Vertex3f(0.458514214f, 0.374064445f, 0.f),
    C_Vertex3f(0.456546068f, 0.084038496f, 0.f),
    C_Vertex3f(0.446922302f, 0.391960144f, 0.f),
    C_Vertex3f(0.44268012f, 0.0643160343f, 0.f),
    C_Vertex3f(0.440032959f, 0.401344299f, 0.f),
    C_Vertex3f(0.426503181f, 0.0467624664f, 0.f),
    C_Vertex3f(0.41015625f, 0.240680695f, 0.f),
    C_Vertex3f(0.41015625f, 0.221246719f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389907837f, 0.33631897f, 0.f),
    C_Vertex3f(0.392816544f, 0.329441071f, 0.f),
    C_Vertex3f(0.393232346f, 0.441017151f, 0.f),
    C_Vertex3f(0.397609711f, 0.316831589f, 0.f),
    C_Vertex3f(0.401763916f, 0.436027527f, 0.f),
    C_Vertex3f(0.401378632f, 0.305368423f, 0.f),
    C_Vertex3f(0.404123306f, 0.295051575f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376661777f, 0.449665546f, 0.f),
    C_Vertex3f(0.360583782f, 0.456983089f, 0.f),
    C_Vertex3f(0.389907837f, 0.33631897f, 0.f),
    C_Vertex3f(0.393232346f, 0.441017151f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.332159042f, 0.0681192875f, 0.f),
    C_Vertex3f(0.319643021f, 0.06255126f, 0.f),
    C_Vertex3f(0.323345423f, -0.00562977791f, 0.f),
    C_Vertex3f(0.306465626f, 0.0583752394f, 0.f),
    C_Vertex3f(0.303763151f, -0.00871253014f, 0.f),
    C_Vertex3f(0.292626858f, 0.0555912256f, 0.f),
    C_Vertex3f(0.283414841f, -0.0102539063f, 0.f),
    C_Vertex3f(0.278126717f, 0.0541992188f, 0.f),
    C_Vertex3f(0.26416111f, -0.0102539063f, 0.f),
    C_Vertex3f(0.259857178f, 0.0541992188f, 0.f),
    C_Vertex3f(0.247273922f, -0.00915455818f, 0.f),
    C_Vertex3f(0.239570618f, 0.0576524734f, 0.f),
    C_Vertex3f(0.230892658f, -0.00695586205f, 0.f),
    C_Vertex3f(0.220375061f, 0.0645589828f, 0.f),
    C_Vertex3f(0.215017319f, -0.00365781784f, 0.f),
    C_Vertex3f(0.202270508f, 0.0749187469f, 0.f),
    C_Vertex3f(0.199647903f, 0.000739574432f, 0.f),
    C_Vertex3f(0.193763733f, 0.0818252563f, 0.f),
    C_Vertex3f(0.184784412f, 0.00623631477f, 0.f),
    C_Vertex3f(0.170426846f, 0.0128324032f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147871494f, 0.151925564f, 0.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, 0.f),
    C_Vertex3f(0.15213728f, 0.139542103f, 0.f),
    C_Vertex3f(0.156575203f, 0.0205278397f, 0.f),
    C_Vertex3f(0.157114029f, 0.127828598f, 0.f),
    C_Vertex3f(0.170426846f, 0.0128324032f, 0.f),
    C_Vertex3f(0.159957886f, 0.122306824f, 0.f),
    C_Vertex3f(0.193763733f, 0.0818252563f, 0.f),
    C_Vertex3f(0.357444763f, 0.37852478f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.144316673f, 0.164978981f, 0.f),
    C_Vertex3f(0.141472816f, 0.178702354f, 0.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, 0.f),
    C_Vertex3f(0.147871494f, 0.151925564f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481042862f, 0.322340012f, 0.f),
    C_Vertex3f(0.484277964f, 0.156219244f, 0.f),
    C_Vertex3f(0.4861269f, 0.298715591f, 0.f),
    C_Vertex3f(0.488899946f, 0.184617281f, 0.f),
    C_Vertex3f(0.489516258f, 0.273290634f, 0.f),
    C_Vertex3f(0.491210938f, 0.215184212f, 0.f),
    C_Vertex3f(0.491210938f, 0.24606514f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.477653503f, 0.333251953f, 0.f),
    C_Vertex3f(0.475137711f, 0.340890884f, 0.f),
    C_Vertex3f(0.47734499f, 0.129990101f, 0.f),
    C_Vertex3f(0.484277964f, 0.156219244f, 0.f),
    C_Vertex3f(0.481042862f, 0.322340012f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.409766197f, 0.0329046249f, 0.f),
    C_Vertex3f(0.426503181f, 0.0467624664f, 0.f),
    C_Vertex3f(0.41015625f, 0.221246719f, 0.f),
    C_Vertex3f(0.408922434f, 0.200364947f, 0.f),
    C_Vertex3f(0.406454802f, 0.180875421f, 0.f),
    C_Vertex3f(0.402753353f, 0.162778139f, 0.f),
    C_Vertex3f(0.397818089f, 0.146073103f, 0.f),
    C_Vertex3f(0.39401412f, 0.0221149921f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.397818089f, 0.146073103f, 0.f),
    C_Vertex3f(0.391649008f, 0.130760312f, 0.f),
    C_Vertex3f(0.39401412f, 0.0221149921f, 0.f),
    C_Vertex3f(0.384246111f, 0.116839767f, 0.f),
    C_Vertex3f(0.377496004f, 0.0128667355f, 0.f),
    C_Vertex3f(0.375609398f, 0.104311466f, 0.f),
    C_Vertex3f(0.365738869f, 0.0931754112f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360211849f, 0.00515985489f, 0.f),
    C_Vertex3f(0.377496004f, 0.0128667355f, 0.f),
    C_Vertex3f(0.365738869f, 0.0931754112f, 0.f),
    C_Vertex3f(0.355206966f, 0.0834313631f, 0.f),
    C_Vertex3f(0.344013691f, 0.0750793219f, 0.f),
    C_Vertex3f(0.342161655f, -0.00100564957f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.332159042f, 0.0681192875f, 0.f),
    C_Vertex3f(0.323345423f, -0.00562977791f, 0.f),
    C_Vertex3f(0.342161655f, -0.00100564957f, 0.f),
    C_Vertex3f(0.344013691f, 0.0750793219f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.475137711f, 0.340890884f, 0.f),
    C_Vertex3f(0.468101025f, 0.105929852f, 0.f),
    C_Vertex3f(0.47734499f, 0.129990101f, 0.f),
    C_Vertex3f(0.104347229f, -0.0341796875f, 0.f),
    C_Vertex3f(0.110496521f, 0.0581130981f, 0.f),
    C_Vertex3f(0.0631942749f, -0.00325012207f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.076423645f, 0.110122681f, -1.f),
    C_Vertex3f(0.0837640762f, 0.0952615738f, -1.f),
    C_Vertex3f(0.0862967968f, 0.372269034f, -1.f),
    C_Vertex3f(0.0929670334f, 0.0804014206f, -1.f),
    C_Vertex3f(0.098354578f, 0.391512036f, -1.f),
    C_Vertex3f(0.104032516f, 0.0655422211f, -1.f),
    C_Vertex3f(0.112421989f, 0.408481121f, -1.f),
    C_Vertex3f(0.110496521f, 0.0581130981f, -1.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, -1.f),
    C_Vertex3f(0.104347229f, -0.0341796875f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.058161974f, 0.272557855f, -1.f),
    C_Vertex3f(0.0583438873f, 0.18184948f, -1.f),
    C_Vertex3f(0.0621812344f, 0.300896525f, -1.f),
    C_Vertex3f(0.0627269745f, 0.154862404f, -1.f),
    C_Vertex3f(0.068210125f, 0.326961279f, -1.f),
    C_Vertex3f(0.0693016052f, 0.129461288f, -1.f),
    C_Vertex3f(0.0762486458f, 0.350752115f, -1.f),
    C_Vertex3f(0.076423645f, 0.110122681f, -1.f),
    C_Vertex3f(0.0862967968f, 0.372269034f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0561523438f, 0.241945267f, -1.f),
    C_Vertex3f(0.0561523438f, 0.210422516f, -1.f),
    C_Vertex3f(0.0583438873f, 0.18184948f, -1.f),
    C_Vertex3f(0.058161974f, 0.272557855f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.332222939f, 0.395069599f, -1.f),
    C_Vertex3f(0.342410088f, 0.389230251f, -1.f),
    C_Vertex3f(0.34499836f, 0.46296978f, -1.f),
    C_Vertex3f(0.352474213f, 0.382417679f, -1.f),
    C_Vertex3f(0.360583782f, 0.456983089f, -1.f),
    C_Vertex3f(0.357444763f, 0.37852478f, -1.f),
    C_Vertex3f(0.389907837f, 0.33631897f, -1.f),
    C_Vertex3f(0.193763733f, 0.0818252563f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.300923347f, 0.406748295f, -1.f),
    C_Vertex3f(0.311479568f, 0.403828621f, -1.f),
    C_Vertex3f(0.314101696f, 0.471117496f, -1.f),
    C_Vertex3f(0.321912766f, 0.399935722f, -1.f),
    C_Vertex3f(0.32990551f, 0.467625618f, -1.f),
    C_Vertex3f(0.332222939f, 0.395069599f, -1.f),
    C_Vertex3f(0.34499836f, 0.46296978f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.26671505f, 0.409667969f, -1.f),
    C_Vertex3f(0.279441833f, 0.409667969f, -1.f),
    C_Vertex3f(0.280361176f, 0.474609375f, -1.f),
    C_Vertex3f(0.290244102f, 0.408694744f, -1.f),
    C_Vertex3f(0.297586918f, 0.473445415f, -1.f),
    C_Vertex3f(0.300923347f, 0.406748295f, -1.f),
    C_Vertex3f(0.314101696f, 0.471117496f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.226652145f, 0.401303053f, -1.f),
    C_Vertex3f(0.23941946f, 0.405485511f, -1.f),
    C_Vertex3f(0.239606142f, 0.472772479f, -1.f),
    C_Vertex3f(0.252773762f, 0.408273816f, -1.f),
    C_Vertex3f(0.260564804f, 0.474609375f, -1.f),
    C_Vertex3f(0.26671505f, 0.409667969f, -1.f),
    C_Vertex3f(0.280361176f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.219344854f, 0.469098687f, -1.f),
    C_Vertex3f(0.199780941f, 0.463587999f, -1.f),
    C_Vertex3f(0.202878475f, 0.388755679f, -1.f),
    C_Vertex3f(0.214471817f, 0.395726442f, -1.f),
    C_Vertex3f(0.226652145f, 0.401303053f, -1.f),
    C_Vertex3f(0.239606142f, 0.472772479f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.163017035f, 0.346958518f, -1.f),
    C_Vertex3f(0.171620369f, 0.359478474f, -1.f),
    C_Vertex3f(0.180914402f, 0.456240416f, -1.f),
    C_Vertex3f(0.181452751f, 0.370631695f, -1.f),
    C_Vertex3f(0.199780941f, 0.463587999f, -1.f),
    C_Vertex3f(0.19187212f, 0.380390763f, -1.f),
    C_Vertex3f(0.202878475f, 0.388755679f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.162745237f, 0.447055936f, -1.f),
    C_Vertex3f(0.145273447f, 0.43603456f, -1.f),
    C_Vertex3f(0.149497509f, 0.317818403f, -1.f),
    C_Vertex3f(0.155642748f, 0.333071828f, -1.f),
    C_Vertex3f(0.163017035f, 0.346958518f, -1.f),
    C_Vertex3f(0.180914402f, 0.456240416f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.145273447f, 0.43603456f, -1.f),
    C_Vertex3f(0.128499031f, 0.423176289f, -1.f),
    C_Vertex3f(0.137207031f, 0.24313736f, -1.f),
    C_Vertex3f(0.138436079f, 0.263857722f, -1.f),
    C_Vertex3f(0.140894175f, 0.28321135f, -1.f),
    C_Vertex3f(0.144581318f, 0.301198244f, -1.f),
    C_Vertex3f(0.149497509f, 0.317818403f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.141472816f, 0.178702354f, -1.f),
    C_Vertex3f(0.139339924f, 0.193095684f, -1.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, -1.f),
    C_Vertex3f(0.137917995f, 0.20815897f, -1.f),
    C_Vertex3f(0.112421989f, 0.408481121f, -1.f),
    C_Vertex3f(0.137207031f, 0.223892212f, -1.f),
    C_Vertex3f(0.128499031f, 0.423176289f, -1.f),
    C_Vertex3f(0.137207031f, 0.24313736f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.445083618f, 0.492240906f, -1.f),
    C_Vertex3f(0.401763916f, 0.436027527f, -1.f),
    C_Vertex3f(0.404123306f, 0.295051575f, -1.f),
    C_Vertex3f(0.404983521f, 0.290466309f, -1.f),
    C_Vertex3f(0.406276703f, 0.28424263f, -1.f),
    C_Vertex3f(0.408216476f, 0.270758629f, -1.f),
    C_Vertex3f(0.409509659f, 0.256237984f, -1.f),
    C_Vertex3f(0.41015625f, 0.240680695f, -1.f),
    C_Vertex3f(0.440032959f, 0.401344299f, -1.f),
    C_Vertex3f(0.486831665f, 0.46206665f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.475137711f, 0.340890884f, -1.f),
    C_Vertex3f(0.46791935f, 0.357041359f, -1.f),
    C_Vertex3f(0.468101025f, 0.105929852f, -1.f),
    C_Vertex3f(0.458514214f, 0.374064445f, -1.f),
    C_Vertex3f(0.456546068f, 0.084038496f, -1.f),
    C_Vertex3f(0.446922302f, 0.391960144f, -1.f),
    C_Vertex3f(0.44268012f, 0.0643160343f, -1.f),
    C_Vertex3f(0.440032959f, 0.401344299f, -1.f),
    C_Vertex3f(0.426503181f, 0.0467624664f, -1.f),
    C_Vertex3f(0.41015625f, 0.240680695f, -1.f),
    C_Vertex3f(0.41015625f, 0.221246719f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.389907837f, 0.33631897f, -1.f),
    C_Vertex3f(0.392816544f, 0.329441071f, -1.f),
    C_Vertex3f(0.393232346f, 0.441017151f, -1.f),
    C_Vertex3f(0.397609711f, 0.316831589f, -1.f),
    C_Vertex3f(0.401763916f, 0.436027527f, -1.f),
    C_Vertex3f(0.401378632f, 0.305368423f, -1.f),
    C_Vertex3f(0.404123306f, 0.295051575f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.376661777f, 0.449665546f, -1.f),
    C_Vertex3f(0.360583782f, 0.456983089f, -1.f),
    C_Vertex3f(0.389907837f, 0.33631897f, -1.f),
    C_Vertex3f(0.393232346f, 0.441017151f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.332159042f, 0.0681192875f, -1.f),
    C_Vertex3f(0.319643021f, 0.06255126f, -1.f),
    C_Vertex3f(0.323345423f, -0.00562977791f, -1.f),
    C_Vertex3f(0.306465626f, 0.0583752394f, -1.f),
    C_Vertex3f(0.303763151f, -0.00871253014f, -1.f),
    C_Vertex3f(0.292626858f, 0.0555912256f, -1.f),
    C_Vertex3f(0.283414841f, -0.0102539063f, -1.f),
    C_Vertex3f(0.278126717f, 0.0541992188f, -1.f),
    C_Vertex3f(0.26416111f, -0.0102539063f, -1.f),
    C_Vertex3f(0.259857178f, 0.0541992188f, -1.f),
    C_Vertex3f(0.247273922f, -0.00915455818f, -1.f),
    C_Vertex3f(0.239570618f, 0.0576524734f, -1.f),
    C_Vertex3f(0.230892658f, -0.00695586205f, -1.f),
    C_Vertex3f(0.220375061f, 0.0645589828f, -1.f),
    C_Vertex3f(0.215017319f, -0.00365781784f, -1.f),
    C_Vertex3f(0.202270508f, 0.0749187469f, -1.f),
    C_Vertex3f(0.199647903f, 0.000739574432f, -1.f),
    C_Vertex3f(0.193763733f, 0.0818252563f, -1.f),
    C_Vertex3f(0.184784412f, 0.00623631477f, -1.f),
    C_Vertex3f(0.170426846f, 0.0128324032f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.147871494f, 0.151925564f, -1.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, -1.f),
    C_Vertex3f(0.15213728f, 0.139542103f, -1.f),
    C_Vertex3f(0.156575203f, 0.0205278397f, -1.f),
    C_Vertex3f(0.157114029f, 0.127828598f, -1.f),
    C_Vertex3f(0.170426846f, 0.0128324032f, -1.f),
    C_Vertex3f(0.159957886f, 0.122306824f, -1.f),
    C_Vertex3f(0.193763733f, 0.0818252563f, -1.f),
    C_Vertex3f(0.357444763f, 0.37852478f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.144316673f, 0.164978981f, -1.f),
    C_Vertex3f(0.141472816f, 0.178702354f, -1.f),
    C_Vertex3f(0.149902344f, 0.0249252319f, -1.f),
    C_Vertex3f(0.147871494f, 0.151925564f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.481042862f, 0.322340012f, -1.f),
    C_Vertex3f(0.484277964f, 0.156219244f, -1.f),
    C_Vertex3f(0.4861269f, 0.298715591f, -1.f),
    C_Vertex3f(0.488899946f, 0.184617281f, -1.f),
    C_Vertex3f(0.489516258f, 0.273290634f, -1.f),
    C_Vertex3f(0.491210938f, 0.215184212f, -1.f),
    C_Vertex3f(0.491210938f, 0.24606514f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.477653503f, 0.333251953f, -1.f),
    C_Vertex3f(0.475137711f, 0.340890884f, -1.f),
    C_Vertex3f(0.47734499f, 0.129990101f, -1.f),
    C_Vertex3f(0.484277964f, 0.156219244f, -1.f),
    C_Vertex3f(0.481042862f, 0.322340012f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.409766197f, 0.0329046249f, -1.f),
    C_Vertex3f(0.426503181f, 0.0467624664f, -1.f),
    C_Vertex3f(0.41015625f, 0.221246719f, -1.f),
    C_Vertex3f(0.408922434f, 0.200364947f, -1.f),
    C_Vertex3f(0.406454802f, 0.180875421f, -1.f),
    C_Vertex3f(0.402753353f, 0.162778139f, -1.f),
    C_Vertex3f(0.397818089f, 0.146073103f, -1.f),
    C_Vertex3f(0.39401412f, 0.0221149921f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.397818089f, 0.146073103f, -1.f),
    C_Vertex3f(0.391649008f, 0.130760312f, -1.f),
    C_Vertex3f(0.39401412f, 0.0221149921f, -1.f),
    C_Vertex3f(0.384246111f, 0.116839767f, -1.f),
    C_Vertex3f(0.377496004f, 0.0128667355f, -1.f),
    C_Vertex3f(0.375609398f, 0.104311466f, -1.f),
    C_Vertex3f(0.365738869f, 0.0931754112f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.360211849f, 0.00515985489f, -1.f),
    C_Vertex3f(0.377496004f, 0.0128667355f, -1.f),
    C_Vertex3f(0.365738869f, 0.0931754112f, -1.f),
    C_Vertex3f(0.355206966f, 0.0834313631f, -1.f),
    C_Vertex3f(0.344013691f, 0.0750793219f, -1.f),
    C_Vertex3f(0.342161655f, -0.00100564957f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.332159042f, 0.0681192875f, -1.f),
    C_Vertex3f(0.323345423f, -0.00562977791f, -1.f),
    C_Vertex3f(0.342161655f, -0.00100564957f, -1.f),
    C_Vertex3f(0.344013691f, 0.0750793219f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.475137711f, 0.340890884f, -1.f),
    C_Vertex3f(0.468101025f, 0.105929852f, -1.f),
    C_Vertex3f(0.47734499f, 0.129990101f, -1.f),
    C_Vertex3f(0.104347229f, -0.0341796875f, -1.f),
    C_Vertex3f(0.110496521f, 0.0581130981f, -1.f),
    C_Vertex3f(0.0631942749f, -0.00325012207f, -1.f),
    C_End(),
    C_Translatef(0.546875f, 0.f, 0.f),
};
static const Call char_249[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.824172258f, -0.566339135f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Normal3f(0.782976985f, -0.622050703f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Normal3f(0.721546888f, -0.692365587f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_Normal3f(0.632779717f, -0.774331868f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, -1.f),
    C_Normal3f(0.528753817f, -0.848775268f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Normal3f(0.413056374f, -0.910705447f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, -1.f),
    C_Normal3f(0.291300356f, -0.95663166f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, -1.f),
    C_Normal3f(0.169845596f, -0.985470653f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, -1.f),
    C_Normal3f(0.0559700541f, -0.998432457f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0470732637f, -0.998891473f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, -1.f),
    C_Normal3f(-0.149569377f, -0.988751233f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, -1.f),
    C_Normal3f(-0.252776206f, -0.967524767f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Normal3f(-0.354831725f, -0.934930205f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Normal3f(-0.449254483f, -0.893403828f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Normal3f(-0.547086418f, -0.837076187f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, -1.f),
    C_Normal3f(-0.652715385f, -0.757603228f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, -1.f),
    C_Normal3f(-0.759380996f, -0.65064621f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, -1.f),
    C_Normal3f(-0.844308019f, -0.535858274f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, -1.f),
    C_Normal3f(-0.901026189f, -0.433764547f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, -1.f),
    C_Normal3f(-0.942368746f, -0.3345761f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, -1.f),
    C_Normal3f(-0.969121754f, -0.246582538f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, -1.f),
    C_Normal3f(-0.987194002f, -0.1595245f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, -1.f),
    C_Normal3f(-0.995961547f, -0.0897801891f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, -1.f),
    C_Normal3f(-0.99899143f, -0.0449032933f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Normal3f(0.999999881f, 0.000538040244f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_Normal3f(0.999861121f, 0.0166672915f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, -1.f),
    C_Normal3f(0.999531269f, 0.0306145567f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, -1.f),
    C_Normal3f(0.998860717f, 0.0477204137f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, -1.f),
    C_Normal3f(0.997604668f, 0.0691730455f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Normal3f(0.995301366f, 0.0968251824f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, -1.f),
    C_Normal3f(0.991019964f, 0.133714229f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, -1.f),
    C_Normal3f(0.984530509f, 0.175213054f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, -1.f),
    C_Normal3f(0.97341013f, 0.229069084f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, -1.f),
    C_Normal3f(0.942804158f, 0.333346993f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Normal3f(0.885687292f, 0.464282304f, 0.f),
    C_Vertex3f(0.14818573f, 0.102423668f, 0.f),
    C_Vertex3f(0.14818573f, 0.102423668f, -1.f),
    C_Normal3f(0.786936522f, 0.617033899f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_Normal3f(0.661581755f, 0.749873102f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, -1.f),
    C_Normal3f(0.519911349f, 0.854220212f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, -1.f),
    C_Normal3f(0.364257187f, 0.931298375f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, -1.f),
    C_Normal3f(0.208589152f, 0.978003383f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_Normal3f(0.065011777f, 0.997884512f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, -1.f),
    C_Normal3f(-0.0649489835f, 0.997888565f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, -1.f),
    C_Normal3f(-0.197600305f, 0.980282605f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, -1.f),
    C_Normal3f(-0.327999711f, 0.94467783f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, -1.f),
    C_Normal3f(-0.449308246f, 0.893376768f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_Normal3f(-0.566825271f, 0.823837996f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, -1.f),
    C_Normal3f(-0.689566791f, 0.724222064f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, -1.f),
    C_Normal3f(-0.800973952f, 0.598699152f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, -1.f),
    C_Normal3f(-0.890218019f, 0.455534756f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, -1.f),
    C_Normal3f(-0.951235354f, 0.308466077f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Normal3f(-0.981952488f, 0.189127818f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, -1.f),
    C_Normal3f(-0.995145023f, 0.0984196588f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.294921875f, 0.521972656f, 0.f),
    C_Vertex3f(0.294921875f, 0.521972656f, -1.f),
    C_Vertex3f(0.231445313f, 0.521972656f, -1.f),
    C_Vertex3f(0.231445313f, 0.521972656f, 0.f),
    C_Normal3f(-0.771486402f, -0.636245787f, 0.f),
    C_Vertex3f(0.231445313f, 0.521972656f, 0.f),
    C_Vertex3f(0.231445313f, 0.521972656f, -1.f),
    C_Vertex3f(0.130371094f, 0.64453125f, -1.f),
    C_Vertex3f(0.130371094f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.130371094f, 0.64453125f, 0.f),
    C_Vertex3f(0.130371094f, 0.64453125f, -1.f),
    C_Vertex3f(0.235839844f, 0.64453125f, -1.f),
    C_Vertex3f(0.235839844f, 0.64453125f, 0.f),
    C_Normal3f(0.900793791f, 0.434247196f, 0.f),
    C_Vertex3f(0.235839844f, 0.64453125f, 0.f),
    C_Vertex3f(0.235839844f, 0.64453125f, -1.f),
    C_Vertex3f(0.294921875f, 0.521972656f, -1.f),
    C_Vertex3f(0.294921875f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14818573f, 0.102423668f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235839844f, 0.64453125f, 0.f),
    C_Vertex3f(0.130371094f, 0.64453125f, 0.f),
    C_Vertex3f(0.231445313f, 0.521972656f, 0.f),
    C_Vertex3f(0.294921875f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, -1.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, -1.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, -1.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, -1.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, -1.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, -1.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, -1.f),
    C_Vertex3f(0.101396084f, 0.028175354f, -1.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, -1.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_Vertex3f(0.352881908f, 0.174730778f, -1.f),
    C_Vertex3f(0.34966135f, 0.152068615f, -1.f),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Vertex3f(0.338389397f, 0.117308617f, -1.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_Vertex3f(0.329683781f, 0.103327751f, -1.f),
    C_Vertex3f(0.325308323f, 0.0266819f, -1.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, -1.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, -1.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, -1.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, -1.f),
    C_Vertex3f(0.270124912f, -0.002866745f, -1.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, -1.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, -1.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, -1.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, -1.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, -1.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, -1.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, -1.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, -1.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, -1.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14818573f, 0.102423668f, -1.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Vertex3f(0.140548706f, 0.124023438f, -1.f),
    C_Vertex3f(0.139947891f, 0.126699448f, -1.f),
    C_Vertex3f(0.138896465f, 0.133307457f, -1.f),
    C_Vertex3f(0.137995243f, 0.141171455f, -1.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Vertex3f(0.13664341f, 0.160667419f, -1.f),
    C_Vertex3f(0.136192799f, 0.172299385f, -1.f),
    C_Vertex3f(0.135892391f, 0.18518734f, -1.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235839844f, 0.64453125f, -1.f),
    C_Vertex3f(0.130371094f, 0.64453125f, -1.f),
    C_Vertex3f(0.231445313f, 0.521972656f, -1.f),
    C_Vertex3f(0.294921875f, 0.521972656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_250[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.824172258f, -0.566339135f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Normal3f(0.782976985f, -0.622050703f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Normal3f(0.721546888f, -0.692365587f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_Normal3f(0.632779717f, -0.774331868f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, -1.f),
    C_Normal3f(0.528753817f, -0.848775268f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Normal3f(0.413056374f, -0.910705447f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, -1.f),
    C_Normal3f(0.291300356f, -0.95663166f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, -1.f),
    C_Normal3f(0.169845596f, -0.985470653f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, -1.f),
    C_Normal3f(0.0559700541f, -0.998432457f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0470732637f, -0.998891473f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, -1.f),
    C_Normal3f(-0.149569377f, -0.988751233f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, -1.f),
    C_Normal3f(-0.252776206f, -0.967524767f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Normal3f(-0.354831725f, -0.934930205f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Normal3f(-0.449254483f, -0.893403828f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Normal3f(-0.547086418f, -0.837076187f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, -1.f),
    C_Normal3f(-0.652715385f, -0.757603228f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, -1.f),
    C_Normal3f(-0.759380996f, -0.65064621f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, -1.f),
    C_Normal3f(-0.844308019f, -0.535858274f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, -1.f),
    C_Normal3f(-0.901026189f, -0.433764547f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, -1.f),
    C_Normal3f(-0.942368746f, -0.3345761f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, -1.f),
    C_Normal3f(-0.969121754f, -0.246582538f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, -1.f),
    C_Normal3f(-0.987194002f, -0.1595245f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, -1.f),
    C_Normal3f(-0.995961547f, -0.0897801891f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, -1.f),
    C_Normal3f(-0.99899143f, -0.0449032933f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Normal3f(0.999999881f, 0.000538040244f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_Normal3f(0.999861121f, 0.0166672915f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, -1.f),
    C_Normal3f(0.999531269f, 0.0306145567f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, -1.f),
    C_Normal3f(0.998860717f, 0.0477204137f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, -1.f),
    C_Normal3f(0.997604668f, 0.0691730455f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Normal3f(0.995301366f, 0.0968251824f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, -1.f),
    C_Normal3f(0.991019964f, 0.133714229f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, -1.f),
    C_Normal3f(0.984530509f, 0.175213054f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, -1.f),
    C_Normal3f(0.97341013f, 0.229069084f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, -1.f),
    C_Normal3f(0.942804158f, 0.333346993f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Normal3f(0.885687292f, 0.464282304f, 0.f),
    C_Vertex3f(0.14818573f, 0.102423668f, 0.f),
    C_Vertex3f(0.14818573f, 0.102423668f, -1.f),
    C_Normal3f(0.786936522f, 0.617033899f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_Normal3f(0.661581755f, 0.749873102f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, -1.f),
    C_Normal3f(0.519911349f, 0.854220212f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, -1.f),
    C_Normal3f(0.364257187f, 0.931298375f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, -1.f),
    C_Normal3f(0.208589152f, 0.978003383f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_Normal3f(0.065011777f, 0.997884512f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, -1.f),
    C_Normal3f(-0.0649489835f, 0.997888565f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, -1.f),
    C_Normal3f(-0.197600305f, 0.980282605f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, -1.f),
    C_Normal3f(-0.327999711f, 0.94467783f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, -1.f),
    C_Normal3f(-0.449308246f, 0.893376768f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_Normal3f(-0.566825271f, 0.823837996f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, -1.f),
    C_Normal3f(-0.689566791f, 0.724222064f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, -1.f),
    C_Normal3f(-0.800973952f, 0.598699152f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, -1.f),
    C_Normal3f(-0.890218019f, 0.455534756f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, -1.f),
    C_Normal3f(-0.951235354f, 0.308466077f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Normal3f(-0.981952488f, 0.189127818f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, -1.f),
    C_Normal3f(-0.995145023f, 0.0984196588f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.192871094f, 0.521972656f, 0.f),
    C_Vertex3f(0.192871094f, 0.521972656f, -1.f),
    C_Vertex3f(0.250488281f, 0.64453125f, -1.f),
    C_Vertex3f(0.250488281f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.250488281f, 0.64453125f, 0.f),
    C_Vertex3f(0.250488281f, 0.64453125f, -1.f),
    C_Vertex3f(0.354003906f, 0.64453125f, -1.f),
    C_Vertex3f(0.354003906f, 0.64453125f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.354003906f, 0.64453125f, 0.f),
    C_Vertex3f(0.354003906f, 0.64453125f, -1.f),
    C_Vertex3f(0.2578125f, 0.521972656f, -1.f),
    C_Vertex3f(0.2578125f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.2578125f, 0.521972656f, 0.f),
    C_Vertex3f(0.2578125f, 0.521972656f, -1.f),
    C_Vertex3f(0.192871094f, 0.521972656f, -1.f),
    C_Vertex3f(0.192871094f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14818573f, 0.102423668f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.250488281f, 0.64453125f, 0.f),
    C_Vertex3f(0.192871094f, 0.521972656f, 0.f),
    C_Vertex3f(0.2578125f, 0.521972656f, 0.f),
    C_Vertex3f(0.354003906f, 0.64453125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, -1.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, -1.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, -1.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, -1.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, -1.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, -1.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, -1.f),
    C_Vertex3f(0.101396084f, 0.028175354f, -1.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, -1.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_Vertex3f(0.352881908f, 0.174730778f, -1.f),
    C_Vertex3f(0.34966135f, 0.152068615f, -1.f),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Vertex3f(0.338389397f, 0.117308617f, -1.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_Vertex3f(0.329683781f, 0.103327751f, -1.f),
    C_Vertex3f(0.325308323f, 0.0266819f, -1.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, -1.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, -1.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, -1.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, -1.f),
    C_Vertex3f(0.270124912f, -0.002866745f, -1.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, -1.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, -1.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, -1.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, -1.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, -1.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, -1.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, -1.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, -1.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, -1.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14818573f, 0.102423668f, -1.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Vertex3f(0.140548706f, 0.124023438f, -1.f),
    C_Vertex3f(0.139947891f, 0.126699448f, -1.f),
    C_Vertex3f(0.138896465f, 0.133307457f, -1.f),
    C_Vertex3f(0.137995243f, 0.141171455f, -1.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Vertex3f(0.13664341f, 0.160667419f, -1.f),
    C_Vertex3f(0.136192799f, 0.172299385f, -1.f),
    C_Vertex3f(0.135892391f, 0.18518734f, -1.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.250488281f, 0.64453125f, -1.f),
    C_Vertex3f(0.192871094f, 0.521972656f, -1.f),
    C_Vertex3f(0.2578125f, 0.521972656f, -1.f),
    C_Vertex3f(0.354003906f, 0.64453125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_251[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.824172258f, -0.566339135f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Normal3f(0.782976985f, -0.622050703f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Normal3f(0.721546888f, -0.692365587f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_Normal3f(0.632779717f, -0.774331868f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, -1.f),
    C_Normal3f(0.528753817f, -0.848775268f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Normal3f(0.413056374f, -0.910705447f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, -1.f),
    C_Normal3f(0.291300356f, -0.95663166f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, -1.f),
    C_Normal3f(0.169845596f, -0.985470653f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, -1.f),
    C_Normal3f(0.0559700541f, -0.998432457f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0470732637f, -0.998891473f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, -1.f),
    C_Normal3f(-0.149569377f, -0.988751233f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, -1.f),
    C_Normal3f(-0.252776206f, -0.967524767f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Normal3f(-0.354831725f, -0.934930205f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Normal3f(-0.449254483f, -0.893403828f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Normal3f(-0.547086418f, -0.837076187f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, -1.f),
    C_Normal3f(-0.652715385f, -0.757603228f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, -1.f),
    C_Normal3f(-0.759380996f, -0.65064621f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, -1.f),
    C_Normal3f(-0.844308019f, -0.535858274f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, -1.f),
    C_Normal3f(-0.901026189f, -0.433764547f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, -1.f),
    C_Normal3f(-0.942368746f, -0.3345761f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, -1.f),
    C_Normal3f(-0.969121754f, -0.246582538f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, -1.f),
    C_Normal3f(-0.987194002f, -0.1595245f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, -1.f),
    C_Normal3f(-0.995961547f, -0.0897801891f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, -1.f),
    C_Normal3f(-0.99899143f, -0.0449032933f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Normal3f(0.999999881f, 0.000538040244f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_Normal3f(0.999861121f, 0.0166672915f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, -1.f),
    C_Normal3f(0.999531269f, 0.0306145567f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, -1.f),
    C_Normal3f(0.998860717f, 0.0477204137f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, -1.f),
    C_Normal3f(0.997604668f, 0.0691730455f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Normal3f(0.995301366f, 0.0968251824f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, -1.f),
    C_Normal3f(0.991019964f, 0.133714229f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, -1.f),
    C_Normal3f(0.984530509f, 0.175213054f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, -1.f),
    C_Normal3f(0.97341013f, 0.229069084f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, -1.f),
    C_Normal3f(0.942804158f, 0.333346993f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Normal3f(0.885687292f, 0.464282304f, 0.f),
    C_Vertex3f(0.14818573f, 0.102423668f, 0.f),
    C_Vertex3f(0.14818573f, 0.102423668f, -1.f),
    C_Normal3f(0.786936522f, 0.617033899f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_Normal3f(0.661581755f, 0.749873102f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, -1.f),
    C_Normal3f(0.519911349f, 0.854220212f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, -1.f),
    C_Normal3f(0.364257187f, 0.931298375f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, -1.f),
    C_Normal3f(0.208589152f, 0.978003383f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_Normal3f(0.065011777f, 0.997884512f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, -1.f),
    C_Normal3f(-0.0649489835f, 0.997888565f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, -1.f),
    C_Normal3f(-0.197600305f, 0.980282605f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, -1.f),
    C_Normal3f(-0.327999711f, 0.94467783f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, -1.f),
    C_Normal3f(-0.449308246f, 0.893376768f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_Normal3f(-0.566825271f, 0.823837996f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, -1.f),
    C_Normal3f(-0.689566791f, 0.724222064f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, -1.f),
    C_Normal3f(-0.800973952f, 0.598699152f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, -1.f),
    C_Normal3f(-0.890218019f, 0.455534756f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, -1.f),
    C_Normal3f(-0.951235354f, 0.308466077f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Normal3f(-0.981952488f, 0.189127818f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, -1.f),
    C_Normal3f(-0.995145023f, 0.0984196588f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(0.835812926f, -0.54901439f, 0.f),
    C_Vertex3f(0.246582031f, 0.596679688f, 0.f),
    C_Vertex3f(0.246582031f, 0.596679688f, -1.f),
    C_Vertex3f(0.197509766f, 0.521972656f, -1.f),
    C_Vertex3f(0.197509766f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.197509766f, 0.521972656f, 0.f),
    C_Vertex3f(0.197509766f, 0.521972656f, -1.f),
    C_Vertex3f(0.108154297f, 0.521972656f, -1.f),
    C_Vertex3f(0.108154297f, 0.521972656f, 0.f),
    C_Normal3f(-0.795031667f, 0.606567979f, 0.f),
    C_Vertex3f(0.108154297f, 0.521972656f, 0.f),
    C_Vertex3f(0.108154297f, 0.521972656f, -1.f),
    C_Vertex3f(0.201660156f, 0.64453125f, -1.f),
    C_Vertex3f(0.201660156f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.201660156f, 0.64453125f, 0.f),
    C_Vertex3f(0.201660156f, 0.64453125f, -1.f),
    C_Vertex3f(0.285644531f, 0.64453125f, -1.f),
    C_Vertex3f(0.285644531f, 0.64453125f, 0.f),
    C_Normal3f(0.779806852f, 0.626020253f, 0.f),
    C_Vertex3f(0.285644531f, 0.64453125f, 0.f),
    C_Vertex3f(0.285644531f, 0.64453125f, -1.f),
    C_Vertex3f(0.384033203f, 0.521972656f, -1.f),
    C_Vertex3f(0.384033203f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.384033203f, 0.521972656f, 0.f),
    C_Vertex3f(0.384033203f, 0.521972656f, -1.f),
    C_Vertex3f(0.294677734f, 0.521972656f, -1.f),
    C_Vertex3f(0.294677734f, 0.521972656f, 0.f),
    C_Normal3f(-0.840821147f, -0.541312993f, 0.f),
    C_Vertex3f(0.294677734f, 0.521972656f, 0.f),
    C_Vertex3f(0.294677734f, 0.521972656f, -1.f),
    C_Vertex3f(0.246582031f, 0.596679688f, -1.f),
    C_Vertex3f(0.246582031f, 0.596679688f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14818573f, 0.102423668f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.108154297f, 0.521972656f, 0.f),
    C_Vertex3f(0.197509766f, 0.521972656f, 0.f),
    C_Vertex3f(0.201660156f, 0.64453125f, 0.f),
    C_Vertex3f(0.246582031f, 0.596679688f, 0.f),
    C_Vertex3f(0.285644531f, 0.64453125f, 0.f),
    C_Vertex3f(0.294677734f, 0.521972656f, 0.f),
    C_Vertex3f(0.384033203f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, -1.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, -1.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, -1.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, -1.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, -1.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, -1.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, -1.f),
    C_Vertex3f(0.101396084f, 0.028175354f, -1.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, -1.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_Vertex3f(0.352881908f, 0.174730778f, -1.f),
    C_Vertex3f(0.34966135f, 0.152068615f, -1.f),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Vertex3f(0.338389397f, 0.117308617f, -1.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_Vertex3f(0.329683781f, 0.103327751f, -1.f),
    C_Vertex3f(0.325308323f, 0.0266819f, -1.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, -1.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, -1.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, -1.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, -1.f),
    C_Vertex3f(0.270124912f, -0.002866745f, -1.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, -1.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, -1.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, -1.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, -1.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, -1.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, -1.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, -1.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, -1.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, -1.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14818573f, 0.102423668f, -1.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Vertex3f(0.140548706f, 0.124023438f, -1.f),
    C_Vertex3f(0.139947891f, 0.126699448f, -1.f),
    C_Vertex3f(0.138896465f, 0.133307457f, -1.f),
    C_Vertex3f(0.137995243f, 0.141171455f, -1.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Vertex3f(0.13664341f, 0.160667419f, -1.f),
    C_Vertex3f(0.136192799f, 0.172299385f, -1.f),
    C_Vertex3f(0.135892391f, 0.18518734f, -1.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.108154297f, 0.521972656f, -1.f),
    C_Vertex3f(0.197509766f, 0.521972656f, -1.f),
    C_Vertex3f(0.201660156f, 0.64453125f, -1.f),
    C_Vertex3f(0.246582031f, 0.596679688f, -1.f),
    C_Vertex3f(0.285644531f, 0.64453125f, -1.f),
    C_Vertex3f(0.294677734f, 0.521972656f, -1.f),
    C_Vertex3f(0.384033203f, 0.521972656f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_252[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.824172258f, -0.566339135f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Normal3f(0.782976985f, -0.622050703f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Normal3f(0.721546888f, -0.692365587f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_Normal3f(0.632779717f, -0.774331868f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, -1.f),
    C_Normal3f(0.528753817f, -0.848775268f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Normal3f(0.413056374f, -0.910705447f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, -1.f),
    C_Normal3f(0.291300356f, -0.95663166f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, -1.f),
    C_Normal3f(0.169845596f, -0.985470653f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, -1.f),
    C_Normal3f(0.0559700541f, -0.998432457f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0470732637f, -0.998891473f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, -1.f),
    C_Normal3f(-0.149569377f, -0.988751233f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, -1.f),
    C_Normal3f(-0.252776206f, -0.967524767f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Normal3f(-0.354831725f, -0.934930205f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Normal3f(-0.449254483f, -0.893403828f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Normal3f(-0.547086418f, -0.837076187f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, -1.f),
    C_Normal3f(-0.652715385f, -0.757603228f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, -1.f),
    C_Normal3f(-0.759380996f, -0.65064621f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, -1.f),
    C_Normal3f(-0.844308019f, -0.535858274f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, -1.f),
    C_Normal3f(-0.901026189f, -0.433764547f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, -1.f),
    C_Normal3f(-0.942368746f, -0.3345761f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, -1.f),
    C_Normal3f(-0.969121754f, -0.246582538f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, -1.f),
    C_Normal3f(-0.987194002f, -0.1595245f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, -1.f),
    C_Normal3f(-0.995961547f, -0.0897801891f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, -1.f),
    C_Normal3f(-0.99899143f, -0.0449032933f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Normal3f(0.999999881f, 0.000538040244f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_Normal3f(0.999861121f, 0.0166672915f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, -1.f),
    C_Normal3f(0.999531269f, 0.0306145567f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, -1.f),
    C_Normal3f(0.998860717f, 0.0477204137f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, -1.f),
    C_Normal3f(0.997604668f, 0.0691730455f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Normal3f(0.995301366f, 0.0968251824f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, -1.f),
    C_Normal3f(0.991019964f, 0.133714229f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, -1.f),
    C_Normal3f(0.984530509f, 0.175213054f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, -1.f),
    C_Normal3f(0.97341013f, 0.229069084f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, -1.f),
    C_Normal3f(0.942804158f, 0.333346993f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Normal3f(0.885687292f, 0.464282304f, 0.f),
    C_Vertex3f(0.14818573f, 0.102423668f, 0.f),
    C_Vertex3f(0.14818573f, 0.102423668f, -1.f),
    C_Normal3f(0.786936522f, 0.617033899f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_Normal3f(0.661581755f, 0.749873102f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, -1.f),
    C_Normal3f(0.519911349f, 0.854220212f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, -1.f),
    C_Normal3f(0.364257187f, 0.931298375f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, -1.f),
    C_Normal3f(0.208589152f, 0.978003383f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_Normal3f(0.065011777f, 0.997884512f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, -1.f),
    C_Normal3f(-0.0649489835f, 0.997888565f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, -1.f),
    C_Normal3f(-0.197600305f, 0.980282605f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, -1.f),
    C_Normal3f(-0.327999711f, 0.94467783f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, -1.f),
    C_Normal3f(-0.449308246f, 0.893376768f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_Normal3f(-0.566825271f, 0.823837996f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, -1.f),
    C_Normal3f(-0.689566791f, 0.724222064f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, -1.f),
    C_Normal3f(-0.800973952f, 0.598699152f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, -1.f),
    C_Normal3f(-0.890218019f, 0.455534756f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, -1.f),
    C_Normal3f(-0.951235354f, 0.308466077f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Normal3f(-0.981952488f, 0.189127818f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, -1.f),
    C_Normal3f(-0.995145023f, 0.0984196588f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, -1.f),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Normal3f(0.99999994f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.122070313f, 0.521972656f, 0.f),
    C_Vertex3f(0.122070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.122070313f, 0.611328125f, -1.f),
    C_Vertex3f(0.122070313f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.122070313f, 0.611328125f, 0.f),
    C_Vertex3f(0.122070313f, 0.611328125f, -1.f),
    C_Vertex3f(0.204101563f, 0.611328125f, -1.f),
    C_Vertex3f(0.204101563f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.204101563f, 0.611328125f, 0.f),
    C_Vertex3f(0.204101563f, 0.611328125f, -1.f),
    C_Vertex3f(0.204101563f, 0.521972656f, -1.f),
    C_Vertex3f(0.204101563f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.204101563f, 0.521972656f, 0.f),
    C_Vertex3f(0.204101563f, 0.521972656f, -1.f),
    C_Vertex3f(0.122070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.122070313f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.28515625f, 0.521972656f, 0.f),
    C_Vertex3f(0.28515625f, 0.521972656f, -1.f),
    C_Vertex3f(0.28515625f, 0.611328125f, -1.f),
    C_Vertex3f(0.28515625f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.28515625f, 0.611328125f, 0.f),
    C_Vertex3f(0.28515625f, 0.611328125f, -1.f),
    C_Vertex3f(0.3671875f, 0.611328125f, -1.f),
    C_Vertex3f(0.3671875f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.3671875f, 0.611328125f, 0.f),
    C_Vertex3f(0.3671875f, 0.611328125f, -1.f),
    C_Vertex3f(0.3671875f, 0.521972656f, -1.f),
    C_Vertex3f(0.3671875f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.3671875f, 0.521972656f, 0.f),
    C_Vertex3f(0.3671875f, 0.521972656f, -1.f),
    C_Vertex3f(0.28515625f, 0.521972656f, -1.f),
    C_Vertex3f(0.28515625f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135742188f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, 0.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, 0.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, 0.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, 0.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, 0.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, 0.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, 0.f),
    C_Vertex3f(0.101396084f, 0.028175354f, 0.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362792969f, 0.f, 0.f),
    C_Vertex3f(0.433105469f, 0.f, 0.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, 0.f),
    C_Vertex3f(0.433105469f, 0.464355469f, 0.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.354492188f, 0.464355469f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356024742f, 0.0586929321f, 0.f),
    C_Vertex3f(0.354492188f, 0.200914383f, 0.f),
    C_Vertex3f(0.352881908f, 0.174730778f, 0.f),
    C_Vertex3f(0.34966135f, 0.152068615f, 0.f),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.344830513f, 0.132927895f, 0.f),
    C_Vertex3f(0.338389397f, 0.117308617f, 0.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, 0.f),
    C_Vertex3f(0.329683781f, 0.103327751f, 0.f),
    C_Vertex3f(0.325308323f, 0.0266819f, 0.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, 0.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308128357f, 0.0143699646f, 0.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, 0.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, 0.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, 0.f),
    C_Vertex3f(0.270124912f, -0.002866745f, 0.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, 0.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, 0.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, 0.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227263451f, -0.0102539063f, 0.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, 0.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, 0.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, 0.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, 0.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14818573f, 0.102423668f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, 0.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147900581f, 0.00154495239f, 0.f),
    C_Vertex3f(0.142402649f, 0.116268158f, 0.f),
    C_Vertex3f(0.140548706f, 0.124023438f, 0.f),
    C_Vertex3f(0.139947891f, 0.126699448f, 0.f),
    C_Vertex3f(0.138896465f, 0.133307457f, 0.f),
    C_Vertex3f(0.137995243f, 0.141171455f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130106926f, 0.00941085815f, 0.f),
    C_Vertex3f(0.137244225f, 0.150291443f, 0.f),
    C_Vertex3f(0.13664341f, 0.160667419f, 0.f),
    C_Vertex3f(0.136192799f, 0.172299385f, 0.f),
    C_Vertex3f(0.135892391f, 0.18518734f, 0.f),
    C_Vertex3f(0.135742188f, 0.199331284f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122070313f, 0.611328125f, 0.f),
    C_Vertex3f(0.122070313f, 0.521972656f, 0.f),
    C_Vertex3f(0.204101563f, 0.521972656f, 0.f),
    C_Vertex3f(0.204101563f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.28515625f, 0.611328125f, 0.f),
    C_Vertex3f(0.28515625f, 0.521972656f, 0.f),
    C_Vertex3f(0.3671875f, 0.521972656f, 0.f),
    C_Vertex3f(0.3671875f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0577297211f, 0.142486572f, 0.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, 0.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.135742188f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Vertex3f(0.0589313507f, 0.124130249f, -1.f),
    C_Vertex3f(0.0607337952f, 0.109161377f, -1.f),
    C_Vertex3f(0.0636806488f, 0.0947399139f, -1.f),
    C_Vertex3f(0.0684547424f, 0.0788164139f, -1.f),
    C_Vertex3f(0.0745124817f, 0.0642309189f, -1.f),
    C_Vertex3f(0.0818538666f, 0.050983429f, -1.f),
    C_Vertex3f(0.0904788971f, 0.0390739441f, -1.f),
    C_Vertex3f(0.101396084f, 0.028175354f, -1.f),
    C_Vertex3f(0.114605427f, 0.0182876587f, -1.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.362792969f, 0.f, -1.f),
    C_Vertex3f(0.433105469f, 0.f, -1.f),
    C_Vertex3f(0.362792969f, 0.0685424805f, -1.f),
    C_Vertex3f(0.433105469f, 0.464355469f, -1.f),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Vertex3f(0.354492188f, 0.464355469f, -1.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356024742f, 0.0586929321f, -1.f),
    C_Vertex3f(0.354492188f, 0.200914383f, -1.f),
    C_Vertex3f(0.352881908f, 0.174730778f, -1.f),
    C_Vertex3f(0.34966135f, 0.152068615f, -1.f),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.344830513f, 0.132927895f, -1.f),
    C_Vertex3f(0.338389397f, 0.117308617f, -1.f),
    C_Vertex3f(0.341273785f, 0.0414562225f, -1.f),
    C_Vertex3f(0.329683781f, 0.103327751f, -1.f),
    C_Vertex3f(0.325308323f, 0.0266819f, -1.f),
    C_Vertex3f(0.318713665f, 0.0909852982f, -1.f),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Vertex3f(0.30547905f, 0.0802812576f, -1.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.308128357f, 0.0143699646f, -1.f),
    C_Vertex3f(0.289979935f, 0.0712156296f, -1.f),
    C_Vertex3f(0.289733887f, 0.00452041626f, -1.f),
    C_Vertex3f(0.273934364f, 0.0644164085f, -1.f),
    C_Vertex3f(0.270124912f, -0.002866745f, -1.f),
    C_Vertex3f(0.257342339f, 0.0598835945f, -1.f),
    C_Vertex3f(0.249301434f, -0.00779151917f, -1.f),
    C_Vertex3f(0.240203857f, 0.0576171875f, -1.f),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Vertex3f(0.222520828f, 0.0576171875f, -1.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.227263451f, -0.0102539063f, -1.f),
    C_Vertex3f(0.20625782f, 0.0598287582f, -1.f),
    C_Vertex3f(0.205375671f, -0.0102539063f, -1.f),
    C_Vertex3f(0.191412926f, 0.0642518997f, -1.f),
    C_Vertex3f(0.185534954f, -0.00828742981f, -1.f),
    C_Vertex3f(0.177986145f, 0.0708866119f, -1.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Vertex3f(0.165977478f, 0.0797328949f, -1.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.14818573f, 0.102423668f, -1.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Vertex3f(0.166376591f, -0.00435447693f, -1.f),
    C_Vertex3f(0.156044006f, 0.0902452469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.147900581f, 0.00154495239f, -1.f),
    C_Vertex3f(0.142402649f, 0.116268158f, -1.f),
    C_Vertex3f(0.140548706f, 0.124023438f, -1.f),
    C_Vertex3f(0.139947891f, 0.126699448f, -1.f),
    C_Vertex3f(0.138896465f, 0.133307457f, -1.f),
    C_Vertex3f(0.137995243f, 0.141171455f, -1.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.130106926f, 0.00941085815f, -1.f),
    C_Vertex3f(0.137244225f, 0.150291443f, -1.f),
    C_Vertex3f(0.13664341f, 0.160667419f, -1.f),
    C_Vertex3f(0.136192799f, 0.172299385f, -1.f),
    C_Vertex3f(0.135892391f, 0.18518734f, -1.f),
    C_Vertex3f(0.135742188f, 0.199331284f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.122070313f, 0.611328125f, -1.f),
    C_Vertex3f(0.122070313f, 0.521972656f, -1.f),
    C_Vertex3f(0.204101563f, 0.521972656f, -1.f),
    C_Vertex3f(0.204101563f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.28515625f, 0.611328125f, -1.f),
    C_Vertex3f(0.28515625f, 0.521972656f, -1.f),
    C_Vertex3f(0.3671875f, 0.521972656f, -1.f),
    C_Vertex3f(0.3671875f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.0577297211f, 0.142486572f, -1.f),
    C_Vertex3f(0.0571289063f, 0.464355469f, -1.f),
    C_Vertex3f(0.0571289063f, 0.164230347f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_253[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.993022859f, -0.117922105f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_Vertex3f(0.046875f, -0.1043396f, -1.f),
    C_Vertex3f(0.046875f, -0.1043396f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.262003899f, 0.965066791f, 0.f),
    C_Vertex3f(0.046875f, -0.1043396f, 0.f),
    C_Vertex3f(0.046875f, -0.1043396f, -1.f),
    C_Normal3f(0.229912221f, 0.973211408f, 0.f),
    C_Vertex3f(0.0533103943f, -0.106086731f, 0.f),
    C_Vertex3f(0.0533103943f, -0.106086731f, -1.f),
    C_Normal3f(0.184329838f, 0.982864439f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, -1.f),
    C_Normal3f(0.120143346f, 0.992756546f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, -1.f),
    C_Normal3f(0.0401061773f, 0.999195457f, 0.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, 0.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, -1.f),
    C_Normal3f(-0.0471678302f, 0.998887002f, 0.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, 0.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, -1.f),
    C_Normal3f(-0.145652682f, 0.989335775f, 0.f),
    C_Vertex3f(0.110143661f, -0.110236168f, 0.f),
    C_Vertex3f(0.110143661f, -0.110236168f, -1.f),
    C_Normal3f(-0.267909348f, 0.963444173f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, -1.f),
    C_Normal3f(-0.409735411f, 0.912204444f, 0.f),
    C_Vertex3f(0.129777908f, -0.104776382f, 0.f),
    C_Vertex3f(0.129777908f, -0.104776382f, -1.f),
    C_Normal3f(-0.53943646f, 0.842026293f, 0.f),
    C_Vertex3f(0.13763237f, -0.100408554f, 0.f),
    C_Vertex3f(0.13763237f, -0.100408554f, -1.f),
    C_Normal3f(-0.641497552f, 0.76712507f, 0.f),
    C_Vertex3f(0.144777298f, -0.09516716f, 0.f),
    C_Vertex3f(0.144777298f, -0.09516716f, -1.f),
    C_Normal3f(-0.732968271f, 0.680262864f, 0.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, 0.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, -1.f),
    C_Normal3f(-0.804138839f, 0.594441473f, 0.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, 0.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, -1.f),
    C_Normal3f(-0.86107707f, 0.508474469f, 0.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, 0.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, -1.f),
    C_Normal3f(-0.895077646f, 0.445910394f, 0.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, 0.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, -1.f),
    C_Normal3f(-0.910975873f, 0.412459642f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, -1.f),
    C_Normal3f(-0.921322405f, 0.388799489f, 0.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, 0.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, -1.f),
    C_Normal3f(-0.928546607f, 0.37121591f, 0.f),
    C_Vertex3f(0.170659065f, -0.053943634f, 0.f),
    C_Vertex3f(0.170659065f, -0.053943634f, -1.f),
    C_Normal3f(-0.93385601f, 0.357649088f, 0.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, 0.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, -1.f),
    C_Normal3f(-0.937913001f, 0.346870661f, 0.f),
    C_Vertex3f(0.177517891f, -0.036034584f, 0.f),
    C_Vertex3f(0.177517891f, -0.036034584f, -1.f),
    C_Normal3f(-0.940624177f, 0.339449972f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, -1.f),
    C_Normal3f(-0.944727242f, 0.327857316f, 0.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, 0.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, -1.f),
    C_Normal3f(-0.942189276f, 0.335081249f, 0.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, 0.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, -1.f),
    C_Normal3f(-0.938051581f, 0.346495599f, 0.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, 0.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, -1.f),
    C_Normal3f(-0.935004234f, -0.354636669f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.935004234f, -0.354636669f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, -1.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, 0.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.94061166f, 0.339484304f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Normal3f(0.942783356f, 0.333405942f, 0.f),
    C_Vertex3f(0.20488739f, 0.170864105f, 0.f),
    C_Vertex3f(0.20488739f, 0.170864105f, -1.f),
    C_Normal3f(0.95868063f, 0.284484535f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, -1.f),
    C_Normal3f(0.966813982f, 0.255481392f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.968233764f, 0.250046819f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_Normal3f(-0.959497869f, 0.281715602f, 0.f),
    C_Vertex3f(0.235969543f, 0.115627289f, 0.f),
    C_Vertex3f(0.235969543f, 0.115627289f, -1.f),
    C_Normal3f(-0.938995123f, 0.343930542f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.938995123f, 0.343930542f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_Vertex3f(0.360839844f, 0.464355469f, -1.f),
    C_Vertex3f(0.360839844f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.360839844f, 0.464355469f, 0.f),
    C_Vertex3f(0.360839844f, 0.464355469f, -1.f),
    C_Vertex3f(0.439498901f, 0.464355469f, -1.f),
    C_Vertex3f(0.439498901f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.936978996f, -0.349385917f, 0.f),
    C_Vertex3f(0.439498901f, 0.464355469f, 0.f),
    C_Vertex3f(0.439498901f, 0.464355469f, -1.f),
    C_Normal3f(0.936883092f, -0.349642813f, 0.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, 0.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, -1.f),
    C_Normal3f(0.932654917f, -0.360769868f, 0.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, 0.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, -1.f),
    C_Normal3f(0.928989351f, -0.370106369f, 0.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, 0.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, -1.f),
    C_Normal3f(0.92445457f, -0.381292284f, 0.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, 0.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, -1.f),
    C_Normal3f(0.918711185f, -0.394930184f, 0.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, 0.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, -1.f),
    C_Normal3f(0.911223829f, -0.411911488f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, -1.f),
    C_Normal3f(0.901101232f, -0.433608681f, 0.f),
    C_Vertex3f(0.225468159f, -0.100803614f, 0.f),
    C_Vertex3f(0.225468159f, -0.100803614f, -1.f),
    C_Normal3f(0.88511163f, -0.465378761f, 0.f),
    C_Vertex3f(0.221137047f, -0.109474182f, 0.f),
    C_Vertex3f(0.221137047f, -0.109474182f, -1.f),
    C_Normal3f(0.852197766f, -0.523219883f, 0.f),
    C_Vertex3f(0.213924408f, -0.122758865f, 0.f),
    C_Vertex3f(0.213924408f, -0.122758865f, -1.f),
    C_Normal3f(0.799718082f, -0.600375772f, 0.f),
    C_Vertex3f(0.20266819f, -0.139555454f, 0.f),
    C_Vertex3f(0.20266819f, -0.139555454f, -1.f),
    C_Normal3f(0.721279263f, -0.692644298f, 0.f),
    C_Vertex3f(0.190646172f, -0.153766155f, 0.f),
    C_Vertex3f(0.190646172f, -0.153766155f, -1.f),
    C_Normal3f(0.617209971f, -0.786798477f, 0.f),
    C_Vertex3f(0.177858353f, -0.165390968f, 0.f),
    C_Vertex3f(0.177858353f, -0.165390968f, -1.f),
    C_Normal3f(0.486348569f, -0.873764873f, 0.f),
    C_Vertex3f(0.164304733f, -0.174429893f, 0.f),
    C_Vertex3f(0.164304733f, -0.174429893f, -1.f),
    C_Normal3f(0.342036247f, -0.939686716f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, -1.f),
    C_Normal3f(0.197416425f, -0.980319679f, 0.f),
    C_Vertex3f(0.133263588f, -0.18572855f, 0.f),
    C_Vertex3f(0.133263588f, -0.18572855f, -1.f),
    C_Normal3f(0.0692126006f, -0.997601986f, 0.f),
    C_Vertex3f(0.115776062f, -0.187988281f, 0.f),
    C_Vertex3f(0.115776062f, -0.187988281f, -1.f),
    C_Normal3f(-0.0443920419f, -0.999014199f, 0.f),
    C_Vertex3f(0.100692749f, -0.187988281f, 0.f),
    C_Vertex3f(0.100692749f, -0.187988281f, -1.f),
    C_Normal3f(-0.144747168f, -0.989468634f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, -1.f),
    C_Normal3f(-0.225563779f, -0.974228382f, 0.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, 0.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, -1.f),
    C_Normal3f(-0.284065366f, -0.958804965f, 0.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, 0.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, -1.f),
    C_Normal3f(-0.325648993f, -0.945490777f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.90498203f, 0.425449729f, 0.f),
    C_Vertex3f(0.179199219f, 0.521972656f, 0.f),
    C_Vertex3f(0.179199219f, 0.521972656f, -1.f),
    C_Vertex3f(0.236816406f, 0.64453125f, -1.f),
    C_Vertex3f(0.236816406f, 0.64453125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.236816406f, 0.64453125f, 0.f),
    C_Vertex3f(0.236816406f, 0.64453125f, -1.f),
    C_Vertex3f(0.340332031f, 0.64453125f, -1.f),
    C_Vertex3f(0.340332031f, 0.64453125f, 0.f),
    C_Normal3f(0.786644518f, -0.617406249f, 0.f),
    C_Vertex3f(0.340332031f, 0.64453125f, 0.f),
    C_Vertex3f(0.340332031f, 0.64453125f, -1.f),
    C_Vertex3f(0.244140625f, 0.521972656f, -1.f),
    C_Vertex3f(0.244140625f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.244140625f, 0.521972656f, 0.f),
    C_Vertex3f(0.244140625f, 0.521972656f, -1.f),
    C_Vertex3f(0.179199219f, 0.521972656f, -1.f),
    C_Vertex3f(0.179199219f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.240839005f, -0.0658898354f, 0.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, 0.f),
    C_Vertex3f(0.360839844f, 0.464355469f, 0.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, 0.f),
    C_Vertex3f(0.439498901f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235969543f, 0.115627289f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, 0.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, 0.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213924408f, -0.122758865f, 0.f),
    C_Vertex3f(0.221137047f, -0.109474182f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, 0.f),
    C_Vertex3f(0.225468159f, -0.100803614f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.20488739f, 0.170864105f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.190646172f, -0.153766155f, 0.f),
    C_Vertex3f(0.20266819f, -0.139555454f, 0.f),
    C_Vertex3f(0.213924408f, -0.122758865f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190646172f, -0.153766155f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, 0.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, 0.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, 0.f),
    C_Vertex3f(0.177858353f, -0.165390968f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.177858353f, -0.165390968f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, 0.f),
    C_Vertex3f(0.177517891f, -0.036034584f, 0.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, 0.f),
    C_Vertex3f(0.170659065f, -0.053943634f, 0.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, 0.f),
    C_Vertex3f(0.164304733f, -0.174429893f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.164304733f, -0.174429893f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, 0.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, 0.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, 0.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, 0.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151212692f, -0.0890522003f, 0.f),
    C_Vertex3f(0.144777298f, -0.09516716f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, 0.f),
    C_Vertex3f(0.13763237f, -0.100408554f, 0.f),
    C_Vertex3f(0.133263588f, -0.18572855f, 0.f),
    C_Vertex3f(0.129777908f, -0.104776382f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133263588f, -0.18572855f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, 0.f),
    C_Vertex3f(0.115776062f, -0.187988281f, 0.f),
    C_Vertex3f(0.110143661f, -0.110236168f, 0.f),
    C_Vertex3f(0.100692749f, -0.187988281f, 0.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, 0.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, 0.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0533103943f, -0.106086731f, 0.f),
    C_Vertex3f(0.046875f, -0.1043396f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.236816406f, 0.64453125f, 0.f),
    C_Vertex3f(0.179199219f, 0.521972656f, 0.f),
    C_Vertex3f(0.244140625f, 0.521972656f, 0.f),
    C_Vertex3f(0.340332031f, 0.64453125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.240839005f, -0.0658898354f, -1.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, -1.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, -1.f),
    C_Vertex3f(0.360839844f, 0.464355469f, -1.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, -1.f),
    C_Vertex3f(0.439498901f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235969543f, 0.115627289f, -1.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, -1.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, -1.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, -1.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213924408f, -0.122758865f, -1.f),
    C_Vertex3f(0.221137047f, -0.109474182f, -1.f),
    C_Vertex3f(0.22170639f, 0.117263794f, -1.f),
    C_Vertex3f(0.225468159f, -0.100803614f, -1.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.20488739f, 0.170864105f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.190646172f, -0.153766155f, -1.f),
    C_Vertex3f(0.20266819f, -0.139555454f, -1.f),
    C_Vertex3f(0.213924408f, -0.122758865f, -1.f),
    C_Vertex3f(0.22170639f, 0.117263794f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190646172f, -0.153766155f, -1.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, -1.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, -1.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, -1.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, -1.f),
    C_Vertex3f(0.177858353f, -0.165390968f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.177858353f, -0.165390968f, -1.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, -1.f),
    C_Vertex3f(0.177517891f, -0.036034584f, -1.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, -1.f),
    C_Vertex3f(0.170659065f, -0.053943634f, -1.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, -1.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, -1.f),
    C_Vertex3f(0.164304733f, -0.174429893f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.164304733f, -0.174429893f, -1.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, -1.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, -1.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, -1.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, -1.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, -1.f),
    C_Vertex3f(0.149439812f, -0.181209087f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151212692f, -0.0890522003f, -1.f),
    C_Vertex3f(0.144777298f, -0.09516716f, -1.f),
    C_Vertex3f(0.149439812f, -0.181209087f, -1.f),
    C_Vertex3f(0.13763237f, -0.100408554f, -1.f),
    C_Vertex3f(0.133263588f, -0.18572855f, -1.f),
    C_Vertex3f(0.129777908f, -0.104776382f, -1.f),
    C_Vertex3f(0.120615005f, -0.108052254f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133263588f, -0.18572855f, -1.f),
    C_Vertex3f(0.120615005f, -0.108052254f, -1.f),
    C_Vertex3f(0.115776062f, -0.187988281f, -1.f),
    C_Vertex3f(0.110143661f, -0.110236168f, -1.f),
    C_Vertex3f(0.100692749f, -0.187988281f, -1.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, -1.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, -1.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, -1.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, -1.f),
    C_Vertex3f(0.065363884f, -0.108707428f, -1.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, -1.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, -1.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0533103943f, -0.106086731f, -1.f),
    C_Vertex3f(0.046875f, -0.1043396f, -1.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_Vertex3f(0.065363884f, -0.108707428f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.236816406f, 0.64453125f, -1.f),
    C_Vertex3f(0.179199219f, 0.521972656f, -1.f),
    C_Vertex3f(0.244140625f, 0.521972656f, -1.f),
    C_Vertex3f(0.340332031f, 0.64453125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};
static const Call char_254[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, 0.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, -1.f),
    C_Vertex3f(0.0590820313f, 0.640625f, -1.f),
    C_Vertex3f(0.0590820313f, 0.640625f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0590820313f, 0.640625f, 0.f),
    C_Vertex3f(0.0590820313f, 0.640625f, -1.f),
    C_Vertex3f(0.137695313f, 0.640625f, -1.f),
    C_Vertex3f(0.137695313f, 0.640625f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.640625f, 0.f),
    C_Vertex3f(0.137695313f, 0.640625f, -1.f),
    C_Vertex3f(0.137695313f, 0.415267944f, -1.f),
    C_Vertex3f(0.137695313f, 0.415267944f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.725229681f, 0.68850708f, 0.f),
    C_Vertex3f(0.137695313f, 0.415267944f, 0.f),
    C_Vertex3f(0.137695313f, 0.415267944f, -1.f),
    C_Normal3f(-0.698044777f, 0.716054082f, 0.f),
    C_Vertex3f(0.145669937f, 0.423667908f, 0.f),
    C_Vertex3f(0.145669937f, 0.423667908f, -1.f),
    C_Normal3f(-0.654963732f, 0.755660355f, 0.f),
    C_Vertex3f(0.160635948f, 0.437631607f, 0.f),
    C_Vertex3f(0.160635948f, 0.437631607f, -1.f),
    C_Normal3f(-0.584135413f, 0.811656237f, 0.f),
    C_Vertex3f(0.174618721f, 0.448759079f, 0.f),
    C_Vertex3f(0.174618721f, 0.448759079f, -1.f),
    C_Normal3f(-0.501523852f, 0.865143836f, 0.f),
    C_Vertex3f(0.187618256f, 0.457050323f, 0.f),
    C_Vertex3f(0.187618256f, 0.457050323f, -1.f),
    C_Normal3f(-0.420829147f, 0.907139957f, 0.f),
    C_Vertex3f(0.193626404f, 0.459777832f, 0.f),
    C_Vertex3f(0.193626404f, 0.459777832f, -1.f),
    C_Normal3f(-0.351551473f, 0.936168551f, 0.f),
    C_Vertex3f(0.201490402f, 0.463485718f, 0.f),
    C_Vertex3f(0.201490402f, 0.463485718f, -1.f),
    C_Normal3f(-0.257847548f, 0.96618557f, 0.f),
    C_Vertex3f(0.21831131f, 0.469047546f, 0.f),
    C_Vertex3f(0.21831131f, 0.469047546f, -1.f),
    C_Normal3f(-0.148962423f, 0.988842845f, 0.f),
    C_Vertex3f(0.236225128f, 0.472755432f, 0.f),
    C_Vertex3f(0.236225128f, 0.472755432f, -1.f),
    C_Normal3f(-0.0507474281f, 0.998711467f, 0.f),
    C_Vertex3f(0.255231857f, 0.474609375f, 0.f),
    C_Vertex3f(0.255231857f, 0.474609375f, -1.f),
    C_Normal3f(0.0299421735f, 0.999551654f, 0.f),
    C_Vertex3f(0.2727108f, 0.474609375f, 0.f),
    C_Vertex3f(0.2727108f, 0.474609375f, -1.f),
    C_Normal3f(0.100414865f, 0.994945645f, 0.f),
    C_Vertex3f(0.287131786f, 0.473653793f, 0.f),
    C_Vertex3f(0.287131786f, 0.473653793f, -1.f),
    C_Normal3f(0.170995474f, 0.985271811f, 0.f),
    C_Vertex3f(0.301115513f, 0.47174263f, 0.f),
    C_Vertex3f(0.301115513f, 0.47174263f, -1.f),
    C_Normal3f(0.243396968f, 0.969926834f, 0.f),
    C_Vertex3f(0.31466198f, 0.468875885f, 0.f),
    C_Vertex3f(0.31466198f, 0.468875885f, -1.f),
    C_Normal3f(0.316443563f, 0.948611319f, 0.f),
    C_Vertex3f(0.327771187f, 0.465053558f, 0.f),
    C_Vertex3f(0.327771187f, 0.465053558f, -1.f),
    C_Normal3f(0.38882333f, 0.921312332f, 0.f),
    C_Vertex3f(0.340443134f, 0.46027565f, 0.f),
    C_Vertex3f(0.340443134f, 0.46027565f, -1.f),
    C_Normal3f(0.459194154f, 0.888335884f, 0.f),
    C_Vertex3f(0.352677822f, 0.45454216f, 0.f),
    C_Vertex3f(0.352677822f, 0.45454216f, -1.f),
    C_Normal3f(0.526303411f, 0.850296915f, 0.f),
    C_Vertex3f(0.36447525f, 0.447853088f, 0.f),
    C_Vertex3f(0.36447525f, 0.447853088f, -1.f),
    C_Normal3f(0.589528143f, 0.807747841f, 0.f),
    C_Vertex3f(0.375835419f, 0.440208435f, 0.f),
    C_Vertex3f(0.375835419f, 0.440208435f, -1.f),
    C_Normal3f(0.649930656f, 0.759993613f, 0.f),
    C_Vertex3f(0.386499286f, 0.431779027f, 0.f),
    C_Vertex3f(0.386499286f, 0.431779027f, -1.f),
    C_Normal3f(0.706633091f, 0.707580149f, 0.f),
    C_Vertex3f(0.396466851f, 0.422564864f, 0.f),
    C_Vertex3f(0.396466851f, 0.422564864f, -1.f),
    C_Normal3f(0.758668721f, 0.651476741f, 0.f),
    C_Vertex3f(0.405738115f, 0.412565947f, 0.f),
    C_Vertex3f(0.405738115f, 0.412565947f, -1.f),
    C_Normal3f(0.805336893f, 0.592817247f, 0.f),
    C_Vertex3f(0.414313078f, 0.401782274f, 0.f),
    C_Vertex3f(0.414313078f, 0.401782274f, -1.f),
    C_Normal3f(0.846244514f, 0.532794714f, 0.f),
    C_Vertex3f(0.422191739f, 0.390213847f, 0.f),
    C_Vertex3f(0.422191739f, 0.390213847f, -1.f),
    C_Normal3f(0.881300569f, 0.472556144f, 0.f),
    C_Vertex3f(0.429374099f, 0.377860665f, 0.f),
    C_Vertex3f(0.429374099f, 0.377860665f, -1.f),
    C_Normal3f(0.913314521f, 0.407254875f, 0.f),
    C_Vertex3f(0.435860157f, 0.364722729f, 0.f),
    C_Vertex3f(0.435860157f, 0.364722729f, -1.f),
    C_Normal3f(0.94476378f, 0.327752054f, 0.f),
    C_Vertex3f(0.444544792f, 0.343838692f, 0.f),
    C_Vertex3f(0.444544792f, 0.343838692f, -1.f),
    C_Normal3f(0.971849501f, 0.235602468f, 0.f),
    C_Vertex3f(0.453229427f, 0.314654827f, 0.f),
    C_Vertex3f(0.453229427f, 0.314654827f, -1.f),
    C_Normal3f(0.990448296f, 0.137884647f, 0.f),
    C_Vertex3f(0.459019184f, 0.284132481f, 0.f),
    C_Vertex3f(0.459019184f, 0.284132481f, -1.f),
    C_Normal3f(0.9990502f, 0.0435755253f, 0.f),
    C_Vertex3f(0.461914063f, 0.252271652f, 0.f),
    C_Vertex3f(0.461914063f, 0.252271652f, -1.f),
    C_Normal3f(0.99891293f, -0.0466142669f, 0.f),
    C_Vertex3f(0.461914063f, 0.217761993f, 0.f),
    C_Vertex3f(0.461914063f, 0.217761993f, -1.f),
    C_Normal3f(0.98963815f, -0.143583879f, 0.f),
    C_Vertex3f(0.458718777f, 0.18379879f, 0.f),
    C_Vertex3f(0.458718777f, 0.18379879f, -1.f),
    C_Normal3f(0.968709707f, -0.248196214f, 0.f),
    C_Vertex3f(0.452328205f, 0.15169239f, 0.f),
    C_Vertex3f(0.452328205f, 0.15169239f, -1.f),
    C_Normal3f(0.937266588f, -0.34861365f, 0.f),
    C_Vertex3f(0.442742348f, 0.121442795f, 0.f),
    C_Vertex3f(0.442742348f, 0.121442795f, -1.f),
    C_Normal3f(0.900904775f, -0.434016913f, 0.f),
    C_Vertex3f(0.43315649f, 0.100148201f, 0.f),
    C_Vertex3f(0.43315649f, 0.100148201f, -1.f),
    C_Normal3f(0.866167486f, -0.499753982f, 0.f),
    C_Vertex3f(0.426042199f, 0.0867776871f, 0.f),
    C_Vertex3f(0.426042199f, 0.0867776871f, -1.f),
    C_Normal3f(0.828500867f, -0.559987903f, 0.f),
    C_Vertex3f(0.418204188f, 0.0742330551f, 0.f),
    C_Vertex3f(0.418204188f, 0.0742330551f, -1.f),
    C_Normal3f(0.784951866f, -0.619556606f, 0.f),
    C_Vertex3f(0.409642458f, 0.0625143051f, 0.f),
    C_Vertex3f(0.409642458f, 0.0625143051f, -1.f),
    C_Normal3f(0.73572892f, -0.677276134f, 0.f),
    C_Vertex3f(0.400357008f, 0.0516214371f, 0.f),
    C_Vertex3f(0.400357008f, 0.0516214371f, -1.f),
    C_Normal3f(0.681352794f, -0.73195529f, 0.f),
    C_Vertex3f(0.390347838f, 0.041554451f, 0.f),
    C_Vertex3f(0.390347838f, 0.041554451f, -1.f),
    C_Normal3f(0.622644246f, -0.782505035f, 0.f),
    C_Vertex3f(0.379614949f, 0.0323133469f, 0.f),
    C_Vertex3f(0.379614949f, 0.0323133469f, -1.f),
    C_Normal3f(0.554335058f, -0.83229357f, 0.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, 0.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, -1.f),
    C_Normal3f(0.466209978f, -0.884674132f, 0.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, 0.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, -1.f),
    C_Normal3f(0.352067709f, -0.935974538f, 0.f),
    C_Vertex3f(0.324954033f, 0.00113010406f, 0.f),
    C_Vertex3f(0.324954033f, 0.00113010406f, -1.f),
    C_Normal3f(0.215490922f, -0.976505816f, 0.f),
    C_Vertex3f(0.29944706f, -0.00645923615f, 0.f),
    C_Vertex3f(0.29944706f, -0.00645923615f, -1.f),
    C_Normal3f(0.0780463666f, -0.996949673f, 0.f),
    C_Vertex3f(0.273366928f, -0.0102539063f, 0.f),
    C_Vertex3f(0.273366928f, -0.0102539063f, -1.f),
    C_Normal3f(-0.0450196117f, -0.998986125f, 0.f),
    C_Vertex3f(0.250974655f, -0.0102539063f, 0.f),
    C_Vertex3f(0.250974655f, -0.0102539063f, -1.f),
    C_Normal3f(-0.155243188f, -0.987876236f, 0.f),
    C_Vertex3f(0.23349762f, -0.00845718384f, 0.f),
    C_Vertex3f(0.23349762f, -0.00845718384f, -1.f),
    C_Normal3f(-0.262735039f, -0.964867949f, 0.f),
    C_Vertex3f(0.216674805f, -0.00486373901f, 0.f),
    C_Vertex3f(0.216674805f, -0.00486373901f, -1.f),
    C_Normal3f(-0.351516008f, -0.936181903f, 0.f),
    C_Vertex3f(0.20050621f, 0.000526428223f, 0.f),
    C_Vertex3f(0.20050621f, 0.000526428223f, -1.f),
    C_Normal3f(-0.425699621f, -0.90486449f, 0.f),
    C_Vertex3f(0.192749023f, 0.00411987305f, 0.f),
    C_Vertex3f(0.192749023f, 0.00411987305f, -1.f),
    C_Normal3f(-0.512902677f, -0.858446836f, 0.f),
    C_Vertex3f(0.186851501f, 0.00695037842f, 0.f),
    C_Vertex3f(0.186851501f, 0.00695037842f, -1.f),
    C_Normal3f(-0.588580966f, -0.808438301f, 0.f),
    C_Vertex3f(0.174072266f, 0.0152788162f, 0.f),
    C_Vertex3f(0.174072266f, 0.0152788162f, -1.f),
    C_Normal3f(-0.6541664f, -0.756350636f, 0.f),
    C_Vertex3f(0.160308838f, 0.0262746811f, 0.f),
    C_Vertex3f(0.160308838f, 0.0262746811f, -1.f),
    C_Normal3f(-0.694510639f, -0.719482541f, 0.f),
    C_Vertex3f(0.145561218f, 0.039937973f, 0.f),
    C_Vertex3f(0.145561218f, 0.039937973f, -1.f),
    C_Normal3f(-0.720189452f, -0.693777502f, 0.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, 0.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, 0.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, -1.f),
    C_Vertex3f(0.137695313f, -0.177734375f, -1.f),
    C_Vertex3f(0.137695313f, -0.177734375f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.137695313f, -0.177734375f, 0.f),
    C_Vertex3f(0.137695313f, -0.177734375f, -1.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, -1.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.999615848f, -0.0277152546f, 0.f),
    C_Vertex3f(0.130271912f, 0.240693092f, 0.f),
    C_Vertex3f(0.130271912f, 0.240693092f, -1.f),
    C_Normal3f(0.999660134f, 0.0260691009f, 0.f),
    C_Vertex3f(0.130271912f, 0.218290329f, 0.f),
    C_Vertex3f(0.130271912f, 0.218290329f, -1.f),
    C_Normal3f(0.996499777f, 0.0835965872f, 0.f),
    C_Vertex3f(0.131402254f, 0.197348356f, 0.f),
    C_Vertex3f(0.131402254f, 0.197348356f, -1.f),
    C_Normal3f(0.988832057f, 0.149034306f, 0.f),
    C_Vertex3f(0.133662939f, 0.177868128f, 0.f),
    C_Vertex3f(0.133662939f, 0.177868128f, -1.f),
    C_Normal3f(0.974800289f, 0.223079026f, 0.f),
    C_Vertex3f(0.137053967f, 0.159849644f, 0.f),
    C_Vertex3f(0.137053967f, 0.159849644f, -1.f),
    C_Normal3f(0.952034593f, 0.305990368f, 0.f),
    C_Vertex3f(0.141575336f, 0.143292904f, 0.f),
    C_Vertex3f(0.141575336f, 0.143292904f, -1.f),
    C_Normal3f(0.917732f, 0.397200227f, 0.f),
    C_Vertex3f(0.147227049f, 0.128197908f, 0.f),
    C_Vertex3f(0.147227049f, 0.128197908f, -1.f),
    C_Normal3f(0.868983746f, 0.494840533f, 0.f),
    C_Vertex3f(0.154009104f, 0.114564657f, 0.f),
    C_Vertex3f(0.154009104f, 0.114564657f, -1.f),
    C_Normal3f(0.803455591f, 0.59536469f, 0.f),
    C_Vertex3f(0.161921501f, 0.10239315f, 0.f),
    C_Vertex3f(0.161921501f, 0.10239315f, -1.f),
    C_Normal3f(0.734325945f, 0.678797007f, 0.f),
    C_Vertex3f(0.170964241f, 0.0916833878f, 0.f),
    C_Vertex3f(0.170964241f, 0.0916833878f, -1.f),
    C_Normal3f(0.665555418f, 0.746348441f, 0.f),
    C_Vertex3f(0.180483818f, 0.0823123455f, 0.f),
    C_Vertex3f(0.180483818f, 0.0823123455f, -1.f),
    C_Normal3f(0.58398515f, 0.811764359f, 0.f),
    C_Vertex3f(0.190480232f, 0.0742800236f, 0.f),
    C_Vertex3f(0.190480232f, 0.0742800236f, -1.f),
    C_Normal3f(0.490194947f, 0.871612847f, 0.f),
    C_Vertex3f(0.200953484f, 0.067586422f, 0.f),
    C_Vertex3f(0.200953484f, 0.067586422f, -1.f),
    C_Normal3f(0.386275589f, 0.922383428f, 0.f),
    C_Vertex3f(0.211903572f, 0.0622315407f, 0.f),
    C_Vertex3f(0.211903572f, 0.0622315407f, -1.f),
    C_Normal3f(0.275775731f, 0.961222053f, 0.f),
    C_Vertex3f(0.223330498f, 0.0582153797f, 0.f),
    C_Vertex3f(0.223330498f, 0.0582153797f, -1.f),
    C_Normal3f(0.163164258f, 0.986598909f, 0.f),
    C_Vertex3f(0.235234261f, 0.0555379391f, 0.f),
    C_Vertex3f(0.235234261f, 0.0555379391f, -1.f),
    C_Normal3f(0.0527447052f, 0.998607993f, 0.f),
    C_Vertex3f(0.247614861f, 0.0541992188f, 0.f),
    C_Vertex3f(0.247614861f, 0.0541992188f, -1.f),
    C_Normal3f(-0.0541482642f, 0.998532891f, 0.f),
    C_Vertex3f(0.260580063f, 0.0541992188f, 0.f),
    C_Vertex3f(0.260580063f, 0.0541992188f, -1.f),
    C_Normal3f(-0.165824533f, 0.986155331f, 0.f),
    C_Vertex3f(0.273183346f, 0.055585742f, 0.f),
    C_Vertex3f(0.273183346f, 0.055585742f, -1.f),
    C_Normal3f(-0.279690474f, 0.96009022f, 0.f),
    C_Vertex3f(0.285316944f, 0.0583587885f, 0.f),
    C_Vertex3f(0.285316944f, 0.0583587885f, -1.f),
    C_Normal3f(-0.390832007f, 0.920462012f, 0.f),
    C_Vertex3f(0.296980858f, 0.0625183582f, 0.f),
    C_Vertex3f(0.296980858f, 0.0625183582f, -1.f),
    C_Normal3f(-0.494754314f, 0.86903286f, 0.f),
    C_Vertex3f(0.308175087f, 0.0680644512f, 0.f),
    C_Vertex3f(0.308175087f, 0.0680644512f, -1.f),
    C_Normal3f(-0.588020563f, 0.808845997f, 0.f),
    C_Vertex3f(0.318899632f, 0.0749970675f, 0.f),
    C_Vertex3f(0.318899632f, 0.0749970675f, -1.f),
    C_Normal3f(-0.668734491f, 0.743501246f, 0.f),
    C_Vertex3f(0.329154491f, 0.0833162069f, 0.f),
    C_Vertex3f(0.329154491f, 0.0833162069f, -1.f),
    C_Normal3f(-0.736519337f, 0.676416576f, 0.f),
    C_Vertex3f(0.338939667f, 0.0930218697f, 0.f),
    C_Vertex3f(0.338939667f, 0.0930218697f, -1.f),
    C_Normal3f(-0.805090606f, 0.593151808f, 0.f),
    C_Vertex3f(0.348255157f, 0.104114056f, 0.f),
    C_Vertex3f(0.348255157f, 0.104114056f, -1.f),
    C_Normal3f(-0.870340228f, 0.492450923f, 0.f),
    C_Vertex3f(0.356406212f, 0.116729379f, 0.f),
    C_Vertex3f(0.356406212f, 0.116729379f, -1.f),
    C_Normal3f(-0.918727636f, 0.394891798f, 0.f),
    C_Vertex3f(0.36339283f, 0.130867839f, 0.f),
    C_Vertex3f(0.36339283f, 0.130867839f, -1.f),
    C_Normal3f(-0.952683985f, 0.303962678f, 0.f),
    C_Vertex3f(0.369215012f, 0.146529436f, 0.f),
    C_Vertex3f(0.369215012f, 0.146529436f, -1.f),
    C_Normal3f(-0.975170195f, 0.2214569f, 0.f),
    C_Vertex3f(0.373872757f, 0.16371417f, 0.f),
    C_Vertex3f(0.373872757f, 0.16371417f, -1.f),
    C_Normal3f(-0.989005744f, 0.147877157f, 0.f),
    C_Vertex3f(0.377366066f, 0.182422042f, 0.f),
    C_Vertex3f(0.377366066f, 0.182422042f, -1.f),
    C_Normal3f(-0.996556401f, 0.0829169154f, 0.f),
    C_Vertex3f(0.379694939f, 0.20265305f, 0.f),
    C_Vertex3f(0.379694939f, 0.20265305f, -1.f),
    C_Normal3f(-0.999657571f, 0.0261669848f, 0.f),
    C_Vertex3f(0.380859375f, 0.224407196f, 0.f),
    C_Vertex3f(0.380859375f, 0.224407196f, -1.f),
    C_Normal3f(-0.999658823f, -0.0261190329f, 0.f),
    C_Vertex3f(0.380859375f, 0.247138023f, 0.f),
    C_Vertex3f(0.380859375f, 0.247138023f, -1.f),
    C_Normal3f(-0.996416628f, -0.0845805034f, 0.f),
    C_Vertex3f(0.379722238f, 0.267929077f, 0.f),
    C_Vertex3f(0.379722238f, 0.267929077f, -1.f),
    C_Normal3f(-0.988642216f, -0.150288418f, 0.f),
    C_Vertex3f(0.377447963f, 0.287326813f, 0.f),
    C_Vertex3f(0.377447963f, 0.287326813f, -1.f),
    C_Normal3f(-0.974565089f, -0.224104732f, 0.f),
    C_Vertex3f(0.374036551f, 0.30533123f, 0.f),
    C_Vertex3f(0.374036551f, 0.30533123f, -1.f),
    C_Normal3f(-0.951997876f, -0.306104749f, 0.f),
    C_Vertex3f(0.369488001f, 0.321942329f, 0.f),
    C_Vertex3f(0.369488001f, 0.321942329f, -1.f),
    C_Normal3f(-0.918435454f, -0.395570695f, 0.f),
    C_Vertex3f(0.363802314f, 0.33716011f, 0.f),
    C_Vertex3f(0.363802314f, 0.33716011f, -1.f),
    C_Normal3f(-0.871376812f, -0.490614325f, 0.f),
    C_Vertex3f(0.356979489f, 0.350984573f, 0.f),
    C_Vertex3f(0.356979489f, 0.350984573f, -1.f),
    C_Normal3f(-0.808921039f, -0.587917387f, 0.f),
    C_Vertex3f(0.349019527f, 0.363415718f, 0.f),
    C_Vertex3f(0.349019527f, 0.363415718f, -1.f),
    C_Normal3f(-0.743106782f, -0.669172823f, 0.f),
    C_Vertex3f(0.339922428f, 0.374453545f, 0.f),
    C_Vertex3f(0.339922428f, 0.374453545f, -1.f),
    C_Normal3f(-0.676569343f, -0.736378789f, 0.f),
    C_Vertex3f(0.330382705f, 0.384111643f, 0.f),
    C_Vertex3f(0.330382705f, 0.384111643f, -1.f),
    C_Normal3f(-0.596761942f, -0.802418292f, 0.f),
    C_Vertex3f(0.320400357f, 0.392390013f, 0.f),
    C_Vertex3f(0.320400357f, 0.392390013f, -1.f),
    C_Normal3f(-0.503776908f, -0.863833845f, 0.f),
    C_Vertex3f(0.309975386f, 0.399288654f, 0.f),
    C_Vertex3f(0.309975386f, 0.399288654f, -1.f),
    C_Normal3f(-0.399267375f, -0.916834533f, 0.f),
    C_Vertex3f(0.29910779f, 0.404807568f, 0.f),
    C_Vertex3f(0.29910779f, 0.404807568f, -1.f),
    C_Normal3f(-0.286575019f, -0.95805788f, 0.f),
    C_Vertex3f(0.28779757f, 0.408946753f, 0.f),
    C_Vertex3f(0.28779757f, 0.408946753f, -1.f),
    C_Normal3f(-0.17031312f, -0.985390007f, 0.f),
    C_Vertex3f(0.276044726f, 0.411706209f, 0.f),
    C_Vertex3f(0.276044726f, 0.411706209f, -1.f),
    C_Normal3f(-0.0555947907f, -0.998453438f, 0.f),
    C_Vertex3f(0.263849258f, 0.413085938f, 0.f),
    C_Vertex3f(0.263849258f, 0.413085938f, -1.f),
    C_Normal3f(0.0592253618f, -0.998244703f, 0.f),
    C_Vertex3f(0.251265526f, 0.413085938f, 0.f),
    C_Vertex3f(0.251265526f, 0.413085938f, -1.f),
    C_Normal3f(0.18072392f, -0.983533859f, 0.f),
    C_Vertex3f(0.239096999f, 0.411617398f, 0.f),
    C_Vertex3f(0.239096999f, 0.411617398f, -1.f),
    C_Normal3f(0.301100224f, -0.95359242f, 0.f),
    C_Vertex3f(0.227289319f, 0.40868032f, 0.f),
    C_Vertex3f(0.227289319f, 0.40868032f, -1.f),
    C_Normal3f(0.415060222f, -0.909794033f, 0.f),
    C_Vertex3f(0.215842485f, 0.404274702f, 0.f),
    C_Vertex3f(0.215842485f, 0.404274702f, -1.f),
    C_Normal3f(0.518243968f, -0.855232775f, 0.f),
    C_Vertex3f(0.204756498f, 0.398400545f, 0.f),
    C_Vertex3f(0.204756498f, 0.398400545f, -1.f),
    C_Normal3f(0.6080845f, -0.793872297f, 0.f),
    C_Vertex3f(0.194031358f, 0.391057849f, 0.f),
    C_Vertex3f(0.194031358f, 0.391057849f, -1.f),
    C_Normal3f(0.68386966f, -0.729604185f, 0.f),
    C_Vertex3f(0.183667064f, 0.382246614f, 0.f),
    C_Vertex3f(0.183667064f, 0.382246614f, -1.f),
    C_Normal3f(0.746305943f, -0.665603042f, 0.f),
    C_Vertex3f(0.173663616f, 0.371966839f, 0.f),
    C_Vertex3f(0.173663616f, 0.371966839f, -1.f),
    C_Normal3f(0.808380961f, -0.588659704f, 0.f),
    C_Vertex3f(0.164021015f, 0.360218525f, 0.f),
    C_Vertex3f(0.164021015f, 0.360218525f, -1.f),
    C_Normal3f(0.868791342f, -0.495178312f, 0.f),
    C_Vertex3f(0.155583739f, 0.347138524f, 0.f),
    C_Vertex3f(0.155583739f, 0.347138524f, -1.f),
    C_Normal3f(0.915422916f, -0.402493119f, 0.f),
    C_Vertex3f(0.148351789f, 0.332726836f, 0.f),
    C_Vertex3f(0.148351789f, 0.332726836f, -1.f),
    C_Normal3f(0.949474931f, -0.313842922f, 0.f),
    C_Vertex3f(0.142325163f, 0.316983461f, 0.f),
    C_Vertex3f(0.142325163f, 0.316983461f, -1.f),
    C_Normal3f(0.972872734f, -0.231340945f, 0.f),
    C_Vertex3f(0.137503862f, 0.2999084f, 0.f),
    C_Vertex3f(0.137503862f, 0.2999084f, -1.f),
    C_Normal3f(0.987748206f, -0.156056091f, 0.f),
    C_Vertex3f(0.133887887f, 0.281501651f, 0.f),
    C_Vertex3f(0.133887887f, 0.281501651f, -1.f),
    C_Normal3f(0.996097267f, -0.0882624462f, 0.f),
    C_Vertex3f(0.131477237f, 0.261763215f, 0.f),
    C_Vertex3f(0.131477237f, 0.261763215f, -1.f),
    C_Normal3f(0.999615848f, -0.0277152546f, 0.f),
    C_Vertex3f(0.130271912f, 0.240693092f, 0.f),
    C_Vertex3f(0.130271912f, 0.240693092f, -1.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.131477237f, 0.261763215f, 0.f),
    C_Vertex3f(0.133887887f, 0.281501651f, 0.f),
    C_Vertex3f(0.137695313f, 0.640625f, 0.f),
    C_Vertex3f(0.137503862f, 0.2999084f, 0.f),
    C_Vertex3f(0.137695313f, 0.415267944f, 0.f),
    C_Vertex3f(0.142325163f, 0.316983461f, 0.f),
    C_Vertex3f(0.145669937f, 0.423667908f, 0.f),
    C_Vertex3f(0.148351789f, 0.332726836f, 0.f),
    C_Vertex3f(0.160635948f, 0.437631607f, 0.f),
    C_Vertex3f(0.155583739f, 0.347138524f, 0.f),
    C_Vertex3f(0.164021015f, 0.360218525f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133662939f, 0.177868128f, 0.f),
    C_Vertex3f(0.131402254f, 0.197348356f, 0.f),
    C_Vertex3f(0.137695313f, -0.177734375f, 0.f),
    C_Vertex3f(0.130271912f, 0.218290329f, 0.f),
    C_Vertex3f(0.0590820313f, 0.640625f, 0.f),
    C_Vertex3f(0.130271912f, 0.240693092f, 0.f),
    C_Vertex3f(0.137695313f, 0.640625f, 0.f),
    C_Vertex3f(0.131477237f, 0.261763215f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379722238f, 0.267929077f, 0.f),
    C_Vertex3f(0.380859375f, 0.247138023f, 0.f),
    C_Vertex3f(0.386499286f, 0.431779027f, 0.f),
    C_Vertex3f(0.390347838f, 0.041554451f, 0.f),
    C_Vertex3f(0.396466851f, 0.422564864f, 0.f),
    C_Vertex3f(0.400357008f, 0.0516214371f, 0.f),
    C_Vertex3f(0.405738115f, 0.412565947f, 0.f),
    C_Vertex3f(0.409642458f, 0.0625143051f, 0.f),
    C_Vertex3f(0.414313078f, 0.401782274f, 0.f),
    C_Vertex3f(0.418204188f, 0.0742330551f, 0.f),
    C_Vertex3f(0.422191739f, 0.390213847f, 0.f),
    C_Vertex3f(0.426042199f, 0.0867776871f, 0.f),
    C_Vertex3f(0.429374099f, 0.377860665f, 0.f),
    C_Vertex3f(0.43315649f, 0.100148201f, 0.f),
    C_Vertex3f(0.435860157f, 0.364722729f, 0.f),
    C_Vertex3f(0.442742348f, 0.121442795f, 0.f),
    C_Vertex3f(0.444544792f, 0.343838692f, 0.f),
    C_Vertex3f(0.452328205f, 0.15169239f, 0.f),
    C_Vertex3f(0.453229427f, 0.314654827f, 0.f),
    C_Vertex3f(0.458718777f, 0.18379879f, 0.f),
    C_Vertex3f(0.459019184f, 0.284132481f, 0.f),
    C_Vertex3f(0.461914063f, 0.217761993f, 0.f),
    C_Vertex3f(0.461914063f, 0.252271652f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369488001f, 0.321942329f, 0.f),
    C_Vertex3f(0.374036551f, 0.30533123f, 0.f),
    C_Vertex3f(0.375835419f, 0.440208435f, 0.f),
    C_Vertex3f(0.377447963f, 0.287326813f, 0.f),
    C_Vertex3f(0.386499286f, 0.431779027f, 0.f),
    C_Vertex3f(0.379722238f, 0.267929077f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36447525f, 0.447853088f, 0.f),
    C_Vertex3f(0.352677822f, 0.45454216f, 0.f),
    C_Vertex3f(0.356979489f, 0.350984573f, 0.f),
    C_Vertex3f(0.363802314f, 0.33716011f, 0.f),
    C_Vertex3f(0.369488001f, 0.321942329f, 0.f),
    C_Vertex3f(0.375835419f, 0.440208435f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.330382705f, 0.384111643f, 0.f),
    C_Vertex3f(0.339922428f, 0.374453545f, 0.f),
    C_Vertex3f(0.340443134f, 0.46027565f, 0.f),
    C_Vertex3f(0.349019527f, 0.363415718f, 0.f),
    C_Vertex3f(0.352677822f, 0.45454216f, 0.f),
    C_Vertex3f(0.356979489f, 0.350984573f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.28779757f, 0.408946753f, 0.f),
    C_Vertex3f(0.29910779f, 0.404807568f, 0.f),
    C_Vertex3f(0.301115513f, 0.47174263f, 0.f),
    C_Vertex3f(0.309975386f, 0.399288654f, 0.f),
    C_Vertex3f(0.31466198f, 0.468875885f, 0.f),
    C_Vertex3f(0.320400357f, 0.392390013f, 0.f),
    C_Vertex3f(0.327771187f, 0.465053558f, 0.f),
    C_Vertex3f(0.330382705f, 0.384111643f, 0.f),
    C_Vertex3f(0.340443134f, 0.46027565f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239096999f, 0.411617398f, 0.f),
    C_Vertex3f(0.251265526f, 0.413085938f, 0.f),
    C_Vertex3f(0.255231857f, 0.474609375f, 0.f),
    C_Vertex3f(0.263849258f, 0.413085938f, 0.f),
    C_Vertex3f(0.2727108f, 0.474609375f, 0.f),
    C_Vertex3f(0.276044726f, 0.411706209f, 0.f),
    C_Vertex3f(0.287131786f, 0.473653793f, 0.f),
    C_Vertex3f(0.28779757f, 0.408946753f, 0.f),
    C_Vertex3f(0.301115513f, 0.47174263f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.204756498f, 0.398400545f, 0.f),
    C_Vertex3f(0.215842485f, 0.404274702f, 0.f),
    C_Vertex3f(0.21831131f, 0.469047546f, 0.f),
    C_Vertex3f(0.227289319f, 0.40868032f, 0.f),
    C_Vertex3f(0.236225128f, 0.472755432f, 0.f),
    C_Vertex3f(0.239096999f, 0.411617398f, 0.f),
    C_Vertex3f(0.255231857f, 0.474609375f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.201490402f, 0.463485718f, 0.f),
    C_Vertex3f(0.193626404f, 0.459777832f, 0.f),
    C_Vertex3f(0.194031358f, 0.391057849f, 0.f),
    C_Vertex3f(0.204756498f, 0.398400545f, 0.f),
    C_Vertex3f(0.21831131f, 0.469047546f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164021015f, 0.360218525f, 0.f),
    C_Vertex3f(0.173663616f, 0.371966839f, 0.f),
    C_Vertex3f(0.174618721f, 0.448759079f, 0.f),
    C_Vertex3f(0.183667064f, 0.382246614f, 0.f),
    C_Vertex3f(0.187618256f, 0.457050323f, 0.f),
    C_Vertex3f(0.194031358f, 0.391057849f, 0.f),
    C_Vertex3f(0.193626404f, 0.459777832f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.380859375f, 0.247138023f, 0.f),
    C_Vertex3f(0.380859375f, 0.224407196f, 0.f),
    C_Vertex3f(0.390347838f, 0.041554451f, 0.f),
    C_Vertex3f(0.379694939f, 0.20265305f, 0.f),
    C_Vertex3f(0.379614949f, 0.0323133469f, 0.f),
    C_Vertex3f(0.377366066f, 0.182422042f, 0.f),
    C_Vertex3f(0.373872757f, 0.16371417f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.369215012f, 0.146529436f, 0.f),
    C_Vertex3f(0.36339283f, 0.130867839f, 0.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, 0.f),
    C_Vertex3f(0.379614949f, 0.0323133469f, 0.f),
    C_Vertex3f(0.373872757f, 0.16371417f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356406212f, 0.116729379f, 0.f),
    C_Vertex3f(0.348255157f, 0.104114056f, 0.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, 0.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, 0.f),
    C_Vertex3f(0.36339283f, 0.130867839f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.348255157f, 0.104114056f, 0.f),
    C_Vertex3f(0.338939667f, 0.0930218697f, 0.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, 0.f),
    C_Vertex3f(0.329154491f, 0.0833162069f, 0.f),
    C_Vertex3f(0.324954033f, 0.00113010406f, 0.f),
    C_Vertex3f(0.318899632f, 0.0749970675f, 0.f),
    C_Vertex3f(0.308175087f, 0.0680644512f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.29944706f, -0.00645923615f, 0.f),
    C_Vertex3f(0.324954033f, 0.00113010406f, 0.f),
    C_Vertex3f(0.308175087f, 0.0680644512f, 0.f),
    C_Vertex3f(0.296980858f, 0.0625183582f, 0.f),
    C_Vertex3f(0.285316944f, 0.0583587885f, 0.f),
    C_Vertex3f(0.273366928f, -0.0102539063f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.285316944f, 0.0583587885f, 0.f),
    C_Vertex3f(0.273183346f, 0.055585742f, 0.f),
    C_Vertex3f(0.273366928f, -0.0102539063f, 0.f),
    C_Vertex3f(0.260580063f, 0.0541992188f, 0.f),
    C_Vertex3f(0.250974655f, -0.0102539063f, 0.f),
    C_Vertex3f(0.247614861f, 0.0541992188f, 0.f),
    C_Vertex3f(0.235234261f, 0.0555379391f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.250974655f, -0.0102539063f, 0.f),
    C_Vertex3f(0.235234261f, 0.0555379391f, 0.f),
    C_Vertex3f(0.23349762f, -0.00845718384f, 0.f),
    C_Vertex3f(0.223330498f, 0.0582153797f, 0.f),
    C_Vertex3f(0.216674805f, -0.00486373901f, 0.f),
    C_Vertex3f(0.211903572f, 0.0622315407f, 0.f),
    C_Vertex3f(0.200953484f, 0.067586422f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.170964241f, 0.0916833878f, 0.f),
    C_Vertex3f(0.174072266f, 0.0152788162f, 0.f),
    C_Vertex3f(0.180483818f, 0.0823123455f, 0.f),
    C_Vertex3f(0.186851501f, 0.00695037842f, 0.f),
    C_Vertex3f(0.190480232f, 0.0742800236f, 0.f),
    C_Vertex3f(0.192749023f, 0.00411987305f, 0.f),
    C_Vertex3f(0.200953484f, 0.067586422f, 0.f),
    C_Vertex3f(0.20050621f, 0.000526428223f, 0.f),
    C_Vertex3f(0.216674805f, -0.00486373901f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161921501f, 0.10239315f, 0.f),
    C_Vertex3f(0.154009104f, 0.114564657f, 0.f),
    C_Vertex3f(0.160308838f, 0.0262746811f, 0.f),
    C_Vertex3f(0.174072266f, 0.0152788162f, 0.f),
    C_Vertex3f(0.170964241f, 0.0916833878f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133662939f, 0.177868128f, 0.f),
    C_Vertex3f(0.137695313f, -0.177734375f, 0.f),
    C_Vertex3f(0.137053967f, 0.159849644f, 0.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, 0.f),
    C_Vertex3f(0.141575336f, 0.143292904f, 0.f),
    C_Vertex3f(0.145561218f, 0.039937973f, 0.f),
    C_Vertex3f(0.147227049f, 0.128197908f, 0.f),
    C_Vertex3f(0.160308838f, 0.0262746811f, 0.f),
    C_Vertex3f(0.154009104f, 0.114564657f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.137695313f, -0.177734375f, 0.f),
    C_Vertex3f(0.0590820313f, 0.640625f, 0.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, 0.f),
    C_Vertex3f(0.160635948f, 0.437631607f, 0.f),
    C_Vertex3f(0.164021015f, 0.360218525f, 0.f),
    C_Vertex3f(0.174618721f, 0.448759079f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.131477237f, 0.261763215f, -1.f),
    C_Vertex3f(0.133887887f, 0.281501651f, -1.f),
    C_Vertex3f(0.137695313f, 0.640625f, -1.f),
    C_Vertex3f(0.137503862f, 0.2999084f, -1.f),
    C_Vertex3f(0.137695313f, 0.415267944f, -1.f),
    C_Vertex3f(0.142325163f, 0.316983461f, -1.f),
    C_Vertex3f(0.145669937f, 0.423667908f, -1.f),
    C_Vertex3f(0.148351789f, 0.332726836f, -1.f),
    C_Vertex3f(0.160635948f, 0.437631607f, -1.f),
    C_Vertex3f(0.155583739f, 0.347138524f, -1.f),
    C_Vertex3f(0.164021015f, 0.360218525f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133662939f, 0.177868128f, -1.f),
    C_Vertex3f(0.131402254f, 0.197348356f, -1.f),
    C_Vertex3f(0.137695313f, -0.177734375f, -1.f),
    C_Vertex3f(0.130271912f, 0.218290329f, -1.f),
    C_Vertex3f(0.0590820313f, 0.640625f, -1.f),
    C_Vertex3f(0.130271912f, 0.240693092f, -1.f),
    C_Vertex3f(0.137695313f, 0.640625f, -1.f),
    C_Vertex3f(0.131477237f, 0.261763215f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.379722238f, 0.267929077f, -1.f),
    C_Vertex3f(0.380859375f, 0.247138023f, -1.f),
    C_Vertex3f(0.386499286f, 0.431779027f, -1.f),
    C_Vertex3f(0.390347838f, 0.041554451f, -1.f),
    C_Vertex3f(0.396466851f, 0.422564864f, -1.f),
    C_Vertex3f(0.400357008f, 0.0516214371f, -1.f),
    C_Vertex3f(0.405738115f, 0.412565947f, -1.f),
    C_Vertex3f(0.409642458f, 0.0625143051f, -1.f),
    C_Vertex3f(0.414313078f, 0.401782274f, -1.f),
    C_Vertex3f(0.418204188f, 0.0742330551f, -1.f),
    C_Vertex3f(0.422191739f, 0.390213847f, -1.f),
    C_Vertex3f(0.426042199f, 0.0867776871f, -1.f),
    C_Vertex3f(0.429374099f, 0.377860665f, -1.f),
    C_Vertex3f(0.43315649f, 0.100148201f, -1.f),
    C_Vertex3f(0.435860157f, 0.364722729f, -1.f),
    C_Vertex3f(0.442742348f, 0.121442795f, -1.f),
    C_Vertex3f(0.444544792f, 0.343838692f, -1.f),
    C_Vertex3f(0.452328205f, 0.15169239f, -1.f),
    C_Vertex3f(0.453229427f, 0.314654827f, -1.f),
    C_Vertex3f(0.458718777f, 0.18379879f, -1.f),
    C_Vertex3f(0.459019184f, 0.284132481f, -1.f),
    C_Vertex3f(0.461914063f, 0.217761993f, -1.f),
    C_Vertex3f(0.461914063f, 0.252271652f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.369488001f, 0.321942329f, -1.f),
    C_Vertex3f(0.374036551f, 0.30533123f, -1.f),
    C_Vertex3f(0.375835419f, 0.440208435f, -1.f),
    C_Vertex3f(0.377447963f, 0.287326813f, -1.f),
    C_Vertex3f(0.386499286f, 0.431779027f, -1.f),
    C_Vertex3f(0.379722238f, 0.267929077f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.36447525f, 0.447853088f, -1.f),
    C_Vertex3f(0.352677822f, 0.45454216f, -1.f),
    C_Vertex3f(0.356979489f, 0.350984573f, -1.f),
    C_Vertex3f(0.363802314f, 0.33716011f, -1.f),
    C_Vertex3f(0.369488001f, 0.321942329f, -1.f),
    C_Vertex3f(0.375835419f, 0.440208435f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.330382705f, 0.384111643f, -1.f),
    C_Vertex3f(0.339922428f, 0.374453545f, -1.f),
    C_Vertex3f(0.340443134f, 0.46027565f, -1.f),
    C_Vertex3f(0.349019527f, 0.363415718f, -1.f),
    C_Vertex3f(0.352677822f, 0.45454216f, -1.f),
    C_Vertex3f(0.356979489f, 0.350984573f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.28779757f, 0.408946753f, -1.f),
    C_Vertex3f(0.29910779f, 0.404807568f, -1.f),
    C_Vertex3f(0.301115513f, 0.47174263f, -1.f),
    C_Vertex3f(0.309975386f, 0.399288654f, -1.f),
    C_Vertex3f(0.31466198f, 0.468875885f, -1.f),
    C_Vertex3f(0.320400357f, 0.392390013f, -1.f),
    C_Vertex3f(0.327771187f, 0.465053558f, -1.f),
    C_Vertex3f(0.330382705f, 0.384111643f, -1.f),
    C_Vertex3f(0.340443134f, 0.46027565f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.239096999f, 0.411617398f, -1.f),
    C_Vertex3f(0.251265526f, 0.413085938f, -1.f),
    C_Vertex3f(0.255231857f, 0.474609375f, -1.f),
    C_Vertex3f(0.263849258f, 0.413085938f, -1.f),
    C_Vertex3f(0.2727108f, 0.474609375f, -1.f),
    C_Vertex3f(0.276044726f, 0.411706209f, -1.f),
    C_Vertex3f(0.287131786f, 0.473653793f, -1.f),
    C_Vertex3f(0.28779757f, 0.408946753f, -1.f),
    C_Vertex3f(0.301115513f, 0.47174263f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.204756498f, 0.398400545f, -1.f),
    C_Vertex3f(0.215842485f, 0.404274702f, -1.f),
    C_Vertex3f(0.21831131f, 0.469047546f, -1.f),
    C_Vertex3f(0.227289319f, 0.40868032f, -1.f),
    C_Vertex3f(0.236225128f, 0.472755432f, -1.f),
    C_Vertex3f(0.239096999f, 0.411617398f, -1.f),
    C_Vertex3f(0.255231857f, 0.474609375f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.201490402f, 0.463485718f, -1.f),
    C_Vertex3f(0.193626404f, 0.459777832f, -1.f),
    C_Vertex3f(0.194031358f, 0.391057849f, -1.f),
    C_Vertex3f(0.204756498f, 0.398400545f, -1.f),
    C_Vertex3f(0.21831131f, 0.469047546f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.164021015f, 0.360218525f, -1.f),
    C_Vertex3f(0.173663616f, 0.371966839f, -1.f),
    C_Vertex3f(0.174618721f, 0.448759079f, -1.f),
    C_Vertex3f(0.183667064f, 0.382246614f, -1.f),
    C_Vertex3f(0.187618256f, 0.457050323f, -1.f),
    C_Vertex3f(0.194031358f, 0.391057849f, -1.f),
    C_Vertex3f(0.193626404f, 0.459777832f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.380859375f, 0.247138023f, -1.f),
    C_Vertex3f(0.380859375f, 0.224407196f, -1.f),
    C_Vertex3f(0.390347838f, 0.041554451f, -1.f),
    C_Vertex3f(0.379694939f, 0.20265305f, -1.f),
    C_Vertex3f(0.379614949f, 0.0323133469f, -1.f),
    C_Vertex3f(0.377366066f, 0.182422042f, -1.f),
    C_Vertex3f(0.373872757f, 0.16371417f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.369215012f, 0.146529436f, -1.f),
    C_Vertex3f(0.36339283f, 0.130867839f, -1.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, -1.f),
    C_Vertex3f(0.379614949f, 0.0323133469f, -1.f),
    C_Vertex3f(0.373872757f, 0.16371417f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.356406212f, 0.116729379f, -1.f),
    C_Vertex3f(0.348255157f, 0.104114056f, -1.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, -1.f),
    C_Vertex3f(0.36815834f, 0.0238981247f, -1.f),
    C_Vertex3f(0.36339283f, 0.130867839f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.348255157f, 0.104114056f, -1.f),
    C_Vertex3f(0.338939667f, 0.0930218697f, -1.f),
    C_Vertex3f(0.349887848f, 0.0125141144f, -1.f),
    C_Vertex3f(0.329154491f, 0.0833162069f, -1.f),
    C_Vertex3f(0.324954033f, 0.00113010406f, -1.f),
    C_Vertex3f(0.318899632f, 0.0749970675f, -1.f),
    C_Vertex3f(0.308175087f, 0.0680644512f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.29944706f, -0.00645923615f, -1.f),
    C_Vertex3f(0.324954033f, 0.00113010406f, -1.f),
    C_Vertex3f(0.308175087f, 0.0680644512f, -1.f),
    C_Vertex3f(0.296980858f, 0.0625183582f, -1.f),
    C_Vertex3f(0.285316944f, 0.0583587885f, -1.f),
    C_Vertex3f(0.273366928f, -0.0102539063f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.285316944f, 0.0583587885f, -1.f),
    C_Vertex3f(0.273183346f, 0.055585742f, -1.f),
    C_Vertex3f(0.273366928f, -0.0102539063f, -1.f),
    C_Vertex3f(0.260580063f, 0.0541992188f, -1.f),
    C_Vertex3f(0.250974655f, -0.0102539063f, -1.f),
    C_Vertex3f(0.247614861f, 0.0541992188f, -1.f),
    C_Vertex3f(0.235234261f, 0.0555379391f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.250974655f, -0.0102539063f, -1.f),
    C_Vertex3f(0.235234261f, 0.0555379391f, -1.f),
    C_Vertex3f(0.23349762f, -0.00845718384f, -1.f),
    C_Vertex3f(0.223330498f, 0.0582153797f, -1.f),
    C_Vertex3f(0.216674805f, -0.00486373901f, -1.f),
    C_Vertex3f(0.211903572f, 0.0622315407f, -1.f),
    C_Vertex3f(0.200953484f, 0.067586422f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.170964241f, 0.0916833878f, -1.f),
    C_Vertex3f(0.174072266f, 0.0152788162f, -1.f),
    C_Vertex3f(0.180483818f, 0.0823123455f, -1.f),
    C_Vertex3f(0.186851501f, 0.00695037842f, -1.f),
    C_Vertex3f(0.190480232f, 0.0742800236f, -1.f),
    C_Vertex3f(0.192749023f, 0.00411987305f, -1.f),
    C_Vertex3f(0.200953484f, 0.067586422f, -1.f),
    C_Vertex3f(0.20050621f, 0.000526428223f, -1.f),
    C_Vertex3f(0.216674805f, -0.00486373901f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.161921501f, 0.10239315f, -1.f),
    C_Vertex3f(0.154009104f, 0.114564657f, -1.f),
    C_Vertex3f(0.160308838f, 0.0262746811f, -1.f),
    C_Vertex3f(0.174072266f, 0.0152788162f, -1.f),
    C_Vertex3f(0.170964241f, 0.0916833878f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133662939f, 0.177868128f, -1.f),
    C_Vertex3f(0.137695313f, -0.177734375f, -1.f),
    C_Vertex3f(0.137053967f, 0.159849644f, -1.f),
    C_Vertex3f(0.137695313f, 0.0481033325f, -1.f),
    C_Vertex3f(0.141575336f, 0.143292904f, -1.f),
    C_Vertex3f(0.145561218f, 0.039937973f, -1.f),
    C_Vertex3f(0.147227049f, 0.128197908f, -1.f),
    C_Vertex3f(0.160308838f, 0.0262746811f, -1.f),
    C_Vertex3f(0.154009104f, 0.114564657f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.137695313f, -0.177734375f, -1.f),
    C_Vertex3f(0.0590820313f, 0.640625f, -1.f),
    C_Vertex3f(0.0590820313f, -0.177734375f, -1.f),
    C_Vertex3f(0.160635948f, 0.437631607f, -1.f),
    C_Vertex3f(0.164021015f, 0.360218525f, -1.f),
    C_Vertex3f(0.174618721f, 0.448759079f, -1.f),
    C_End(),
    C_Translatef(0.497558594f, 0.f, 0.f),
};
static const Call char_255[] = {
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_FrontFace(GL_CW),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.993022859f, -0.117922105f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_Vertex3f(0.046875f, -0.1043396f, -1.f),
    C_Vertex3f(0.046875f, -0.1043396f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.262003899f, 0.965066791f, 0.f),
    C_Vertex3f(0.046875f, -0.1043396f, 0.f),
    C_Vertex3f(0.046875f, -0.1043396f, -1.f),
    C_Normal3f(0.229912221f, 0.973211408f, 0.f),
    C_Vertex3f(0.0533103943f, -0.106086731f, 0.f),
    C_Vertex3f(0.0533103943f, -0.106086731f, -1.f),
    C_Normal3f(0.184329838f, 0.982864439f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, -1.f),
    C_Normal3f(0.120143346f, 0.992756546f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, -1.f),
    C_Normal3f(0.0401061773f, 0.999195457f, 0.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, 0.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, -1.f),
    C_Normal3f(-0.0471678302f, 0.998887002f, 0.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, 0.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, -1.f),
    C_Normal3f(-0.145652682f, 0.989335775f, 0.f),
    C_Vertex3f(0.110143661f, -0.110236168f, 0.f),
    C_Vertex3f(0.110143661f, -0.110236168f, -1.f),
    C_Normal3f(-0.267909348f, 0.963444173f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, -1.f),
    C_Normal3f(-0.409735411f, 0.912204444f, 0.f),
    C_Vertex3f(0.129777908f, -0.104776382f, 0.f),
    C_Vertex3f(0.129777908f, -0.104776382f, -1.f),
    C_Normal3f(-0.53943646f, 0.842026293f, 0.f),
    C_Vertex3f(0.13763237f, -0.100408554f, 0.f),
    C_Vertex3f(0.13763237f, -0.100408554f, -1.f),
    C_Normal3f(-0.641497552f, 0.76712507f, 0.f),
    C_Vertex3f(0.144777298f, -0.09516716f, 0.f),
    C_Vertex3f(0.144777298f, -0.09516716f, -1.f),
    C_Normal3f(-0.732968271f, 0.680262864f, 0.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, 0.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, -1.f),
    C_Normal3f(-0.804138839f, 0.594441473f, 0.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, 0.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, -1.f),
    C_Normal3f(-0.86107707f, 0.508474469f, 0.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, 0.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, -1.f),
    C_Normal3f(-0.895077646f, 0.445910394f, 0.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, 0.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, -1.f),
    C_Normal3f(-0.910975873f, 0.412459642f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, -1.f),
    C_Normal3f(-0.921322405f, 0.388799489f, 0.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, 0.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, -1.f),
    C_Normal3f(-0.928546607f, 0.37121591f, 0.f),
    C_Vertex3f(0.170659065f, -0.053943634f, 0.f),
    C_Vertex3f(0.170659065f, -0.053943634f, -1.f),
    C_Normal3f(-0.93385601f, 0.357649088f, 0.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, 0.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, -1.f),
    C_Normal3f(-0.937913001f, 0.346870661f, 0.f),
    C_Vertex3f(0.177517891f, -0.036034584f, 0.f),
    C_Vertex3f(0.177517891f, -0.036034584f, -1.f),
    C_Normal3f(-0.940624177f, 0.339449972f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, -1.f),
    C_Normal3f(-0.944727242f, 0.327857316f, 0.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, 0.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, -1.f),
    C_Normal3f(-0.942189276f, 0.335081249f, 0.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, 0.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, -1.f),
    C_Normal3f(-0.938051581f, 0.346495599f, 0.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, 0.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, -1.f),
    C_Normal3f(-0.935004234f, -0.354636669f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.935004234f, -0.354636669f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, -1.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, 0.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.94061166f, 0.339484304f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Normal3f(0.942783356f, 0.333405942f, 0.f),
    C_Vertex3f(0.20488739f, 0.170864105f, 0.f),
    C_Vertex3f(0.20488739f, 0.170864105f, -1.f),
    C_Normal3f(0.95868063f, 0.284484535f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, -1.f),
    C_Normal3f(0.966813982f, 0.255481392f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(-0.968233764f, 0.250046819f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_Normal3f(-0.959497869f, 0.281715602f, 0.f),
    C_Vertex3f(0.235969543f, 0.115627289f, 0.f),
    C_Vertex3f(0.235969543f, 0.115627289f, -1.f),
    C_Normal3f(-0.938995123f, 0.343930542f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-0.938995123f, 0.343930542f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_Vertex3f(0.360839844f, 0.464355469f, -1.f),
    C_Vertex3f(0.360839844f, 0.464355469f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.360839844f, 0.464355469f, 0.f),
    C_Vertex3f(0.360839844f, 0.464355469f, -1.f),
    C_Vertex3f(0.439498901f, 0.464355469f, -1.f),
    C_Vertex3f(0.439498901f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_QUAD_STRIP),
    C_Normal3f(0.936978996f, -0.349385917f, 0.f),
    C_Vertex3f(0.439498901f, 0.464355469f, 0.f),
    C_Vertex3f(0.439498901f, 0.464355469f, -1.f),
    C_Normal3f(0.936883092f, -0.349642813f, 0.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, 0.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, -1.f),
    C_Normal3f(0.932654917f, -0.360769868f, 0.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, 0.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, -1.f),
    C_Normal3f(0.928989351f, -0.370106369f, 0.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, 0.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, -1.f),
    C_Normal3f(0.92445457f, -0.381292284f, 0.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, 0.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, -1.f),
    C_Normal3f(0.918711185f, -0.394930184f, 0.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, 0.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, -1.f),
    C_Normal3f(0.911223829f, -0.411911488f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, -1.f),
    C_Normal3f(0.901101232f, -0.433608681f, 0.f),
    C_Vertex3f(0.225468159f, -0.100803614f, 0.f),
    C_Vertex3f(0.225468159f, -0.100803614f, -1.f),
    C_Normal3f(0.88511163f, -0.465378761f, 0.f),
    C_Vertex3f(0.221137047f, -0.109474182f, 0.f),
    C_Vertex3f(0.221137047f, -0.109474182f, -1.f),
    C_Normal3f(0.852197766f, -0.523219883f, 0.f),
    C_Vertex3f(0.213924408f, -0.122758865f, 0.f),
    C_Vertex3f(0.213924408f, -0.122758865f, -1.f),
    C_Normal3f(0.799718082f, -0.600375772f, 0.f),
    C_Vertex3f(0.20266819f, -0.139555454f, 0.f),
    C_Vertex3f(0.20266819f, -0.139555454f, -1.f),
    C_Normal3f(0.721279263f, -0.692644298f, 0.f),
    C_Vertex3f(0.190646172f, -0.153766155f, 0.f),
    C_Vertex3f(0.190646172f, -0.153766155f, -1.f),
    C_Normal3f(0.617209971f, -0.786798477f, 0.f),
    C_Vertex3f(0.177858353f, -0.165390968f, 0.f),
    C_Vertex3f(0.177858353f, -0.165390968f, -1.f),
    C_Normal3f(0.486348569f, -0.873764873f, 0.f),
    C_Vertex3f(0.164304733f, -0.174429893f, 0.f),
    C_Vertex3f(0.164304733f, -0.174429893f, -1.f),
    C_Normal3f(0.342036247f, -0.939686716f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, -1.f),
    C_Normal3f(0.197416425f, -0.980319679f, 0.f),
    C_Vertex3f(0.133263588f, -0.18572855f, 0.f),
    C_Vertex3f(0.133263588f, -0.18572855f, -1.f),
    C_Normal3f(0.0692126006f, -0.997601986f, 0.f),
    C_Vertex3f(0.115776062f, -0.187988281f, 0.f),
    C_Vertex3f(0.115776062f, -0.187988281f, -1.f),
    C_Normal3f(-0.0443920419f, -0.999014199f, 0.f),
    C_Vertex3f(0.100692749f, -0.187988281f, 0.f),
    C_Vertex3f(0.100692749f, -0.187988281f, -1.f),
    C_Normal3f(-0.144747168f, -0.989468634f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, -1.f),
    C_Normal3f(-0.225563779f, -0.974228382f, 0.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, 0.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, -1.f),
    C_Normal3f(-0.284065366f, -0.958804965f, 0.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, 0.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, -1.f),
    C_Normal3f(-0.325648993f, -0.945490777f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.108886719f, 0.521972656f, 0.f),
    C_Vertex3f(0.108886719f, 0.521972656f, -1.f),
    C_Vertex3f(0.108886719f, 0.611328125f, -1.f),
    C_Vertex3f(0.108886719f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.108886719f, 0.611328125f, 0.f),
    C_Vertex3f(0.108886719f, 0.611328125f, -1.f),
    C_Vertex3f(0.190917969f, 0.611328125f, -1.f),
    C_Vertex3f(0.190917969f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.190917969f, 0.611328125f, 0.f),
    C_Vertex3f(0.190917969f, 0.611328125f, -1.f),
    C_Vertex3f(0.190917969f, 0.521972656f, -1.f),
    C_Vertex3f(0.190917969f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.190917969f, 0.521972656f, 0.f),
    C_Vertex3f(0.190917969f, 0.521972656f, -1.f),
    C_Vertex3f(0.108886719f, 0.521972656f, -1.f),
    C_Vertex3f(0.108886719f, 0.521972656f, 0.f),
    C_End(),
    C_Begin(GL_QUADS),
    C_Normal3f(-1.f, 0.f, 0.f),
    C_Vertex3f(0.271972656f, 0.521972656f, 0.f),
    C_Vertex3f(0.271972656f, 0.521972656f, -1.f),
    C_Vertex3f(0.271972656f, 0.611328125f, -1.f),
    C_Vertex3f(0.271972656f, 0.611328125f, 0.f),
    C_Normal3f(-0.f, 1.f, 0.f),
    C_Vertex3f(0.271972656f, 0.611328125f, 0.f),
    C_Vertex3f(0.271972656f, 0.611328125f, -1.f),
    C_Vertex3f(0.354003906f, 0.611328125f, -1.f),
    C_Vertex3f(0.354003906f, 0.611328125f, 0.f),
    C_Normal3f(1.f, 0.f, 0.f),
    C_Vertex3f(0.354003906f, 0.611328125f, 0.f),
    C_Vertex3f(0.354003906f, 0.611328125f, -1.f),
    C_Vertex3f(0.354003906f, 0.521972656f, -1.f),
    C_Vertex3f(0.354003906f, 0.521972656f, 0.f),
    C_Normal3f(-0.f, -1.f, 0.f),
    C_Vertex3f(0.354003906f, 0.521972656f, 0.f),
    C_Vertex3f(0.354003906f, 0.521972656f, -1.f),
    C_Vertex3f(0.271972656f, 0.521972656f, -1.f),
    C_Vertex3f(0.271972656f, 0.521972656f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, 1.f),
    C_FrontFace(GL_CCW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.240839005f, -0.0658898354f, 0.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, 0.f),
    C_Vertex3f(0.360839844f, 0.464355469f, 0.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, 0.f),
    C_Vertex3f(0.439498901f, 0.464355469f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235969543f, 0.115627289f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, 0.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, 0.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, 0.f),
    C_Vertex3f(0.252243042f, 0.167865753f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213924408f, -0.122758865f, 0.f),
    C_Vertex3f(0.221137047f, -0.109474182f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, 0.f),
    C_Vertex3f(0.225468159f, -0.100803614f, 0.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, 0.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.20488739f, 0.170864105f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.190646172f, -0.153766155f, 0.f),
    C_Vertex3f(0.20266819f, -0.139555454f, 0.f),
    C_Vertex3f(0.213924408f, -0.122758865f, 0.f),
    C_Vertex3f(0.22170639f, 0.117263794f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190646172f, -0.153766155f, 0.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, 0.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, 0.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, 0.f),
    C_Vertex3f(0.177858353f, -0.165390968f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.177858353f, -0.165390968f, 0.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, 0.f),
    C_Vertex3f(0.177517891f, -0.036034584f, 0.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, 0.f),
    C_Vertex3f(0.170659065f, -0.053943634f, 0.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, 0.f),
    C_Vertex3f(0.164304733f, -0.174429893f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.164304733f, -0.174429893f, 0.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, 0.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, 0.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, 0.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, 0.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151212692f, -0.0890522003f, 0.f),
    C_Vertex3f(0.144777298f, -0.09516716f, 0.f),
    C_Vertex3f(0.149439812f, -0.181209087f, 0.f),
    C_Vertex3f(0.13763237f, -0.100408554f, 0.f),
    C_Vertex3f(0.133263588f, -0.18572855f, 0.f),
    C_Vertex3f(0.129777908f, -0.104776382f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133263588f, -0.18572855f, 0.f),
    C_Vertex3f(0.120615005f, -0.108052254f, 0.f),
    C_Vertex3f(0.115776062f, -0.187988281f, 0.f),
    C_Vertex3f(0.110143661f, -0.110236168f, 0.f),
    C_Vertex3f(0.100692749f, -0.187988281f, 0.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, 0.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, 0.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, 0.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, 0.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0533103943f, -0.106086731f, 0.f),
    C_Vertex3f(0.046875f, -0.1043396f, 0.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, 0.f),
    C_Vertex3f(0.065363884f, -0.108707428f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.108886719f, 0.611328125f, 0.f),
    C_Vertex3f(0.108886719f, 0.521972656f, 0.f),
    C_Vertex3f(0.190917969f, 0.521972656f, 0.f),
    C_Vertex3f(0.190917969f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.271972656f, 0.611328125f, 0.f),
    C_Vertex3f(0.271972656f, 0.521972656f, 0.f),
    C_Vertex3f(0.354003906f, 0.521972656f, 0.f),
    C_Vertex3f(0.354003906f, 0.611328125f, 0.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.190429688f, -0.000381469727f, 0.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, 0.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, 0.f),
    C_End(),
    C_Normal3f(0.f, 0.f, -1.f),
    C_FrontFace(GL_CW),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.240839005f, -0.0658898354f, -1.f),
    C_Vertex3f(0.246755123f, -0.0512845516f, -1.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_Vertex3f(0.253067493f, -0.0351955891f, -1.f),
    C_Vertex3f(0.360839844f, 0.464355469f, -1.f),
    C_Vertex3f(0.259776115f, -0.0176229477f, -1.f),
    C_Vertex3f(0.439498901f, 0.464355469f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.235969543f, 0.115627289f, -1.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, -1.f),
    C_Vertex3f(0.235319138f, -0.0790114403f, -1.f),
    C_Vertex3f(0.240839005f, -0.0658898354f, -1.f),
    C_Vertex3f(0.252243042f, 0.167865753f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.213924408f, -0.122758865f, -1.f),
    C_Vertex3f(0.221137047f, -0.109474182f, -1.f),
    C_Vertex3f(0.22170639f, 0.117263794f, -1.f),
    C_Vertex3f(0.225468159f, -0.100803614f, -1.f),
    C_Vertex3f(0.229133606f, 0.0891571045f, -1.f),
    C_Vertex3f(0.230195522f, -0.0906493664f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.20488739f, 0.170864105f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.190646172f, -0.153766155f, -1.f),
    C_Vertex3f(0.20266819f, -0.139555454f, -1.f),
    C_Vertex3f(0.213924408f, -0.122758865f, -1.f),
    C_Vertex3f(0.22170639f, 0.117263794f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.190646172f, -0.153766155f, -1.f),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.188030243f, -0.00671768188f, -1.f),
    C_Vertex3f(0.184539795f, -0.0163269043f, -1.f),
    C_Vertex3f(0.183448792f, -0.0195999146f, -1.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, -1.f),
    C_Vertex3f(0.177858353f, -0.165390968f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.177858353f, -0.165390968f, -1.f),
    C_Vertex3f(0.181376457f, -0.0254421234f, -1.f),
    C_Vertex3f(0.177517891f, -0.036034584f, -1.f),
    C_Vertex3f(0.173945427f, -0.0455350876f, -1.f),
    C_Vertex3f(0.170659065f, -0.053943634f, -1.f),
    C_Vertex3f(0.167658806f, -0.0612602234f, -1.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, -1.f),
    C_Vertex3f(0.164304733f, -0.174429893f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.164304733f, -0.174429893f, -1.f),
    C_Vertex3f(0.164944649f, -0.0674848557f, -1.f),
    C_Vertex3f(0.162516594f, -0.0726175308f, -1.f),
    C_Vertex3f(0.160374641f, -0.0766582489f, -1.f),
    C_Vertex3f(0.156938553f, -0.0820636749f, -1.f),
    C_Vertex3f(0.151212692f, -0.0890522003f, -1.f),
    C_Vertex3f(0.149439812f, -0.181209087f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.151212692f, -0.0890522003f, -1.f),
    C_Vertex3f(0.144777298f, -0.09516716f, -1.f),
    C_Vertex3f(0.149439812f, -0.181209087f, -1.f),
    C_Vertex3f(0.13763237f, -0.100408554f, -1.f),
    C_Vertex3f(0.133263588f, -0.18572855f, -1.f),
    C_Vertex3f(0.129777908f, -0.104776382f, -1.f),
    C_Vertex3f(0.120615005f, -0.108052254f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.133263588f, -0.18572855f, -1.f),
    C_Vertex3f(0.120615005f, -0.108052254f, -1.f),
    C_Vertex3f(0.115776062f, -0.187988281f, -1.f),
    C_Vertex3f(0.110143661f, -0.110236168f, -1.f),
    C_Vertex3f(0.100692749f, -0.187988281f, -1.f),
    C_Vertex3f(0.0983638763f, -0.111328125f, -1.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, -1.f),
    C_Vertex3f(0.0870189667f, -0.111328125f, -1.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_STRIP),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_Vertex3f(0.0626583099f, -0.180761337f, -1.f),
    C_Vertex3f(0.065363884f, -0.108707428f, -1.f),
    C_Vertex3f(0.0759916306f, -0.184374809f, -1.f),
    C_Vertex3f(0.0766000748f, -0.110454559f, -1.f),
    C_Vertex3f(0.0886697769f, -0.186783791f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.0533103943f, -0.106086731f, -1.f),
    C_Vertex3f(0.046875f, -0.1043396f, -1.f),
    C_Vertex3f(0.0556640625f, -0.178352356f, -1.f),
    C_Vertex3f(0.065363884f, -0.108707428f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.108886719f, 0.611328125f, -1.f),
    C_Vertex3f(0.108886719f, 0.521972656f, -1.f),
    C_Vertex3f(0.190917969f, 0.521972656f, -1.f),
    C_Vertex3f(0.190917969f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLE_FAN),
    C_Vertex3f(0.271972656f, 0.611328125f, -1.f),
    C_Vertex3f(0.271972656f, 0.521972656f, -1.f),
    C_Vertex3f(0.354003906f, 0.521972656f, -1.f),
    C_Vertex3f(0.354003906f, 0.611328125f, -1.f),
    C_End(),
    C_Begin(GL_TRIANGLES),
    C_Vertex3f(0.190429688f, -0.000381469727f, -1.f),
    C_Vertex3f(0.0989608765f, 0.464355469f, -1.f),
    C_Vertex3f(0.0141601563f, 0.464355469f, -1.f),
    C_End(),
    C_Translatef(0.447753906f, 0.f, 0.f),
};


#define CL(i) { char_##i, sizeof char_##i / sizeof char_##i[0] }

static const CallList font[] = {
    CL(  0), CL(  1), CL(  2), CL(  3), CL(  4), CL(  5), CL(  6), CL(  7),
    CL(  8), CL(  9), CL( 10), CL( 11), CL( 12), CL( 13), CL( 14), CL( 15),
    CL( 16), CL( 17), CL( 18), CL( 19), CL( 20), CL( 21), CL( 22), CL( 23),
    CL( 24), CL( 25), CL( 26), CL( 27), CL( 28), CL( 29), CL( 30), CL( 31),
    CL( 32), CL( 33), CL( 34), CL( 35), CL( 36), CL( 37), CL( 38), CL( 39),
    CL( 40), CL( 41), CL( 42), CL( 43), CL( 44), CL( 45), CL( 46), CL( 47),
    CL( 48), CL( 49), CL( 50), CL( 51), CL( 52), CL( 53), CL( 54), CL( 55),
    CL( 56), CL( 57), CL( 58), CL( 59), CL( 60), CL( 61), CL( 62), CL( 63),
    CL( 64), CL( 65), CL( 66), CL( 67), CL( 68), CL( 69), CL( 70), CL( 71),
    CL( 72), CL( 73), CL( 74), CL( 75), CL( 76), CL( 77), CL( 78), CL( 79),
    CL( 80), CL( 81), CL( 82), CL( 83), CL( 84), CL( 85), CL( 86), CL( 87),
    CL( 88), CL( 89), CL( 90), CL( 91), CL( 92), CL( 93), CL( 94), CL( 95),
    CL( 96), CL( 97), CL( 98), CL( 99), CL(100), CL(101), CL(102), CL(103),
    CL(104), CL(105), CL(106), CL(107), CL(108), CL(109), CL(110), CL(111),
    CL(112), CL(113), CL(114), CL(115), CL(116), CL(117), CL(118), CL(119),
    CL(120), CL(121), CL(122), CL(123), CL(124), CL(125), CL(126), CL(127),
    CL(128), CL(129), CL(130), CL(131), CL(132), CL(133), CL(134), CL(135),
    CL(136), CL(137), CL(138), CL(139), CL(140), CL(141), CL(142), CL(143),
    CL(144), CL(145), CL(146), CL(147), CL(148), CL(149), CL(150), CL(151),
    CL(152), CL(153), CL(154), CL(155), CL(156), CL(157), CL(158), CL(159),
    CL(160), CL(161), CL(162), CL(163), CL(164), CL(165), CL(166), CL(167),
    CL(168), CL(169), CL(170), CL(171), CL(172), CL(173), CL(174), CL(175),
    CL(176), CL(177), CL(178), CL(179), CL(180), CL(181), CL(182), CL(183),
    CL(184), CL(185), CL(186), CL(187), CL(188), CL(189), CL(190), CL(191),
    CL(192), CL(193), CL(194), CL(195), CL(196), CL(197), CL(198), CL(199),
    CL(200), CL(201), CL(202), CL(203), CL(204), CL(205), CL(206), CL(207),
    CL(208), CL(209), CL(210), CL(211), CL(212), CL(213), CL(214), CL(215),
    CL(216), CL(217), CL(218), CL(219), CL(220), CL(221), CL(222), CL(223),
    CL(224), CL(225), CL(226), CL(227), CL(228), CL(229), CL(230), CL(231),
    CL(232), CL(233), CL(234), CL(235), CL(236), CL(237), CL(238), CL(239),
    CL(240), CL(241), CL(242), CL(243), CL(244), CL(245), CL(246), CL(247),
    CL(248), CL(249), CL(250), CL(251), CL(252), CL(253), CL(254), CL(255)
};


void
wglSystemFontOutlines(unsigned c, float extrude)
{
    if (c < sizeof font / sizeof font[0] ) {
        const CallList *callList = &font[c];

        for (size_t i = 0; i < callList->length; ++i) {
            const Call *call = &callList->calls[i];
            const float *args = call->args;

            switch (call->function) {
            case F_BEGIN:
                glBegin(args[0]);
                break;
            case F_FRONTFACE:
                glFrontFace(args[0]);
                break;
            case F_VERTEX3F:
                glVertex3f(args[0], args[1], args[2] * extrude);
                break;
            case F_NORMAL3F:
                glNormal3f(args[0], args[1], args[2]);
                break;
            case F_END:
                glEnd();
                break;
            case F_TRANSLATEF:
                glTranslatef(args[0], args[1], args[2]);
                break;
            }
        }

    }
}
